!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods to include the effect of an external potential during an MD
!>        or energy calculation
!> \author Teodoro Laino (03.2008) [tlaino]
! **************************************************************************************************
MODULE external_potential_methods
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_set,&
                                              force_env_type
   USE force_fields_util,               ONLY: get_generic_info
   USE fparser,                         ONLY: evalf,&
                                              evalfd,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_list_types,             ONLY: particle_list_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'external_potential_methods'
   PUBLIC :: add_external_potential

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \date 03.2008
!> \author Teodoro Laino - University of Zurich [tlaino]
! **************************************************************************************************
   SUBROUTINE add_external_potential(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'add_external_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length)               :: def_error, this_error
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: my_par
      INTEGER                                            :: a_var, handle, i, iatom, j, k, n_var, &
                                                            natom, rep
      INTEGER, DIMENSION(:), POINTER                     :: iatms, nparticle
      LOGICAL                                            :: useall
      REAL(KIND=dp)                                      :: dedf, dx, energy, err, lerr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_val
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: ext_pot_section

      useall = .FALSE.
      CALL timeset(routineN, handle)
      NULLIFY (my_par, my_val, logger, subsys, particles, ext_pot_section, nparticle)
      ext_pot_section => section_vals_get_subs_vals(force_env%force_env_section, &
                                                    "EXTERNAL_POTENTIAL")
      CALL section_vals_get(ext_pot_section, n_repetition=n_var)
      DO rep = 1, n_var
         natom = 0
         logger => cp_get_default_logger()
         CALL section_vals_val_get(ext_pot_section, "DX", r_val=dx, i_rep_section=rep)
         CALL section_vals_val_get(ext_pot_section, "ERROR_LIMIT", r_val=lerr, i_rep_section=rep)
         CALL get_generic_info(ext_pot_section, "FUNCTION", coupling_function, my_par, my_val, &
                               input_variables=(/"X", "Y", "Z"/), i_rep_sec=rep)
         CALL initf(1)
         CALL parsef(1, TRIM(coupling_function), my_par)

         ! Apply potential on all atoms, computing energy and forces
         NULLIFY (particles, subsys)
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         CALL force_env_get(force_env, additional_potential=energy)
         CALL section_vals_val_get(ext_pot_section, "ATOMS_LIST", n_rep_val=a_var, i_rep_section=rep)
         DO k = 1, a_var
            CALL section_vals_val_get(ext_pot_section, "ATOMS_LIST", i_rep_val=k, i_vals=iatms, i_rep_section=rep)
            CALL reallocate(nparticle, 1, natom+SIZE(iatms))
            nparticle(natom+1:natom+SIZE(iatms)) = iatms
            natom = natom+SIZE(iatms)
         END DO
         IF (a_var == 0) THEN
            natom = particles%n_els
            useall = .TRUE.
         END IF
         DO i = 1, natom
            IF (useall) THEN
               iatom = i
            ELSE
               iatom = nparticle(i)
            END IF
            my_val(1) = particles%els(iatom)%r(1)
            my_val(2) = particles%els(iatom)%r(2)
            my_val(3) = particles%els(iatom)%r(3)

            energy = energy+evalf(1, my_val)
            DO j = 1, 3
               dedf = evalfd(1, j, my_val, dx, err)
               IF (ABS(err) > lerr) THEN
                  WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
                  WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
                  CALL compress(this_error, .TRUE.)
                  CALL compress(def_error, .TRUE.)
                  CALL cp_warn(cp__l("external_potential_methods.F",123), &
                               'ASSERTION (cond) failed at line '//cp_to_string(124)// &
                               ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                               TRIM(def_error)//' .')
               END IF
               particles%els(iatom)%f(j) = particles%els(iatom)%f(j)-dedf
            END DO
         END DO
         CALL force_env_set(force_env, additional_potential=energy)
         DEALLOCATE (my_par)
         DEALLOCATE (my_val)
         IF (a_var /= 0) THEN
            DEALLOCATE (nparticle)
         END IF
         CALL finalizef()
      END DO
      CALL timestop(handle)
   END SUBROUTINE add_external_potential

END MODULE external_potential_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface for the force calculations
!> \par History
!>      cjm, FEB-20-2001: pass variable box_ref
!>      cjm, SEPT-12-2002: major reorganization
!>      fawzi, APR-12-2003: introduced force_env (based on the work by CJM&JGH)
!>      fawzi, NOV-3-2004: reorganized interface for f77 interface
!> \author fawzi
! **************************************************************************************************
MODULE force_env_methods
   USE atprop_types,                    ONLY: atprop_init,&
                                              atprop_type
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              init_cell,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE constraint_fxd,                  ONLY: fix_atom_control
   USE constraint_vsite,                ONLY: vsite_force_control
   USE cp_iter_types,                   ONLY: cp_iteration_info_copy_iter
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_rm_default_logger,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              low_print_level
   USE cp_para_env,                     ONLY: cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              cp_results_mp_bcast,&
                                              get_results,&
                                              test_for_result
   USE cp_result_types,                 ONLY: cp_result_copy,&
                                              cp_result_create,&
                                              cp_result_p_type,&
                                              cp_result_release,&
                                              cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_p_type,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE eip_environment_types,           ONLY: eip_env_retain,&
                                              eip_environment_type
   USE eip_silicon,                     ONLY: eip_bazant,&
                                              eip_lenosky
   USE external_potential_methods,      ONLY: add_external_potential
   USE fist_environment_types,          ONLY: fist_env_retain,&
                                              fist_environment_type
   USE fist_force,                      ONLY: fist_calc_energy_force
   USE force_env_types,                 ONLY: &
        force_env_get, force_env_get_natom, force_env_p_type, force_env_set, force_env_type, &
        use_eip_force, use_fist_force, use_mixed_force, use_prog_name, use_qmmm, use_qmmmx, &
        use_qs_force
   USE force_env_utils,                 ONLY: rescale_forces,&
                                              write_forces,&
                                              write_stress_tensor
   USE force_fields_util,               ONLY: get_generic_info
   USE fp_methods,                      ONLY: fp_eval
   USE fparser,                         ONLY: EvalErrType,&
                                              evalf,&
                                              evalfd,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE global_types,                    ONLY: global_environment_type,&
                                              globenv_retain
   USE input_constants,                 ONLY: debug_run,&
                                              mix_coupled,&
                                              mix_generic,&
                                              mix_linear_combination,&
                                              mix_minimum,&
                                              mix_restrained,&
                                              use_bazant_eip,&
                                              use_lenosky_eip
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_memory
   USE mathlib,                         ONLY: abnormal_value
   USE message_passing,                 ONLY: mp_sum,&
                                              mp_sync
   USE metadynamics_types,              ONLY: meta_env_retain,&
                                              meta_env_type
   USE mixed_energy_types,              ONLY: mixed_energy_type,&
                                              mixed_force_type
   USE mixed_environment_types,         ONLY: get_mixed_env,&
                                              mixed_env_retain,&
                                              mixed_environment_type
   USE mixed_environment_utils,         ONLY: get_subsys_map_index,&
                                              mixed_map_forces
   USE particle_list_types,             ONLY: particle_list_p_type,&
                                              particle_list_type
   USE physcon,                         ONLY: debye
   USE qmmm_force,                      ONLY: qmmm_calc_energy_force
   USE qmmm_types,                      ONLY: qmmm_env_retain,&
                                              qmmm_env_type
   USE qmmmx_force,                     ONLY: qmmmx_calc_energy_force
   USE qmmmx_types,                     ONLY: qmmmx_env_retain,&
                                              qmmmx_env_type
   USE qs_environment_types,            ONLY: qs_env_retain,&
                                              qs_environment_type
   USE qs_force,                        ONLY: qs_calc_energy_force
   USE restraint,                       ONLY: restraint_control
   USE string_utilities,                ONLY: compress
   USE virial_types,                    ONLY: cp_virial,&
                                              sym_virial,&
                                              virial_create,&
                                              virial_p_type,&
                                              virial_release,&
                                              virial_type,&
                                              zero_virial
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_env_methods'

   PUBLIC :: force_env_create, &
             force_env_calc_energy_force, &
             force_env_calc_num_pressure

   INTEGER, SAVE, PRIVATE :: last_force_env_id = 0

CONTAINS

! **************************************************************************************************
!> \brief Interface routine for force and energy calculations
!> \param force_env the force_env of which you want the energy and forces
!> \param calc_force if false the forces *might* be left unchanged
!>        or be unvalid, no guarantee on them is done.Defaults to true
!> \param consistent_energies Performs an additional qs_ks_update_qs_env, so
!>          that the energies are appropriate to the forces, they are in the
!>          non-selfconsistent case not consistent to each other! [08.2005, TdK]
!> \param skip_external_control ...
!> \param eval_energy_forces ...
!> \param require_consistent_energy_force ...
!> \param linres ...
!> \param calc_stress_tensor ...
!> \author CJM & fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE force_env_calc_energy_force(force_env, calc_force, &
                                                    consistent_energies, skip_external_control, eval_energy_forces, &
                                                    require_consistent_energy_force, linres, calc_stress_tensor)

      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL, INTENT(IN), OPTIONAL :: calc_force, consistent_energies, skip_external_control, &
         eval_energy_forces, require_consistent_energy_force, linres, calc_stress_tensor

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_calc_energy_force', &
         routineP = moduleN//':'//routineN
      REAL(kind=dp), PARAMETER                           :: ateps = 1.0E-6_dp

      INTEGER                                            :: i, j, nat, ndigits, output_unit, &
                                                            print_forces
      LOGICAL                                            :: calculate_forces, &
                                                            calculate_stress_tensor, &
                                                            energy_consistency, eval_ef, &
                                                            linres_run, my_skip
      REAL(KIND=dp)                                      :: checksum, e_pot, sum_energy, &
                                                            sum_pv_virial, sum_stress_tensor
      REAL(KIND=dp), DIMENSION(3)                        :: grand_total_force, total_force
      REAL(KIND=dp), DIMENSION(3, 3)                     :: atomic_stress_tensor, diff_stress_tensor
      TYPE(atprop_type), POINTER                         :: atprop_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (logger, virial, subsys, atprop_env, cell)
      logger => cp_get_default_logger()
      eval_ef = .TRUE.
      my_skip = .FALSE.
      calculate_forces = .TRUE.
      energy_consistency = .FALSE.
      linres_run = .FALSE.
      IF (PRESENT(eval_energy_forces)) eval_ef = eval_energy_forces
      IF (PRESENT(skip_external_control)) my_skip = skip_external_control
      IF (PRESENT(calc_force)) calculate_forces = calc_force
      IF (PRESENT(calc_stress_tensor)) THEN
         calculate_stress_tensor = calc_stress_tensor
      ELSE
         calculate_stress_tensor = calculate_forces
      END IF
      IF (PRESENT(consistent_energies)) energy_consistency = consistent_energies
      IF (PRESENT(linres)) linres_run = linres

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_methods.F",205)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_methods.F",206)
      CALL force_env_get(force_env, subsys=subsys)
      CALL force_env_set(force_env, additional_potential=0.0_dp)
      CALL cp_subsys_get(subsys, virial=virial, atprop=atprop_env, cell=cell)
      IF (virial%pv_availability) CALL zero_virial(virial, reset=.FALSE.)

      nat = force_env_get_natom(force_env)
      CALL atprop_init(atprop_env, nat)
      IF (eval_ef) THEN
         SELECT CASE (force_env%in_use)
         CASE (use_fist_force)
            CALL fist_calc_energy_force(force_env%fist_env)
         CASE (use_qs_force)
            CALL qs_calc_energy_force(force_env%qs_env, calculate_forces, energy_consistency, linres_run)
         CASE (use_eip_force)
            IF (force_env%eip_env%eip_model == use_lenosky_eip) THEN
               CALL eip_lenosky(force_env%eip_env)
            ELSE IF (force_env%eip_env%eip_model == use_bazant_eip) THEN
               CALL eip_bazant(force_env%eip_env)
            END IF
         CASE (use_qmmm)
            CALL qmmm_calc_energy_force(force_env%qmmm_env, &
                                        calculate_forces, energy_consistency, linres=linres_run)
         CASE (use_qmmmx)
            CALL qmmmx_calc_energy_force(force_env%qmmmx_env, &
                                         calculate_forces, energy_consistency, linres=linres_run, &
                                         require_consistent_energy_force=require_consistent_energy_force)
         CASE (use_mixed_force)
            CALL mixed_energy_forces(force_env, calculate_forces)
         CASE default
            CALL cp__b("force_env_methods.F",236,"")
         END SELECT
      END IF
      ! In case it is requested, we evaluate the stress tensor numerically
      IF (virial%pv_availability) THEN
         IF (virial%pv_numer .AND. calculate_stress_tensor) THEN
            ! Compute the numerical stress tensor
            CALL force_env_calc_num_pressure(force_env)
         ELSE
            IF (calculate_forces) THEN
               ! Symmetrize analytical stress tensor
               CALL sym_virial(virial)
            ELSE
               IF (calculate_stress_tensor) THEN
                  CALL cp_warn(cp__l("force_env_methods.F",250), "The calculation of the stress tensor "// &
                               "requires the calculation of the forces")
               END IF
            END IF
         END IF
      END IF

      !sample peak memory
      CALL m_memory()

      ! Some additional tasks..
      IF (.NOT. my_skip) THEN
         ! Flexible Partitioning
         IF (ASSOCIATED(force_env%fp_env)) THEN
            IF (force_env%fp_env%use_fp) THEN
               CALL fp_eval(force_env%fp_env, subsys, cell)
            ENDIF
         ENDIF
         ! Constraints ONLY of Fixed Atom type
         CALL fix_atom_control(force_env)
         ! All Restraints
         CALL restraint_control(force_env)
         ! Virtual Sites
         CALL vsite_force_control(force_env)
         ! External Potential
         CALL add_external_potential(force_env)
         ! Rescale forces if requested
         CALL rescale_forces(force_env)
      END IF

      CALL force_env_get(force_env, potential_energy=e_pot)

      ! Print always Energy in the same format for all methods
      output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".Log")
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(/,T2,"ENERGY| Total FORCE_EVAL ( ",A," ) energy (a.u.): ",T55,F26.15,/)') &
            ADJUSTR(TRIM(use_prog_name(force_env%in_use))), e_pot
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      ! terminate the run if the value of the potential is abnormal
      IF (abnormal_value(e_pot)) &
         CALL cp__b("force_env_methods.F",294,"Potential energy is an abnormal value (NaN/Inf).")

      ! Print forces, if requested
      print_forces = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%FORCES", &
                                          extension=".xyz")
      IF ((print_forces > 0) .AND. calculate_forces) THEN
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, &
                            core_particles=core_particles, &
                            particles=particles, &
                            shell_particles=shell_particles)
         ! Variable precision output of the forces
         CALL section_vals_val_get(force_env%force_env_section, "PRINT%FORCES%NDIGITS", &
                                   i_val=ndigits)
         CALL write_forces(particles, print_forces, "ATOMIC", ndigits, total_force)
         grand_total_force(:) = total_force(:)
         IF (ASSOCIATED(core_particles)) THEN
            CALL write_forces(core_particles, print_forces, "CORE", ndigits, total_force)
            grand_total_force(:) = grand_total_force(:)+total_force(:)
         END IF
         IF (ASSOCIATED(shell_particles)) THEN
            CALL write_forces(shell_particles, print_forces, "SHELL", ndigits, total_force, &
                              grand_total_force)
         END IF
      END IF
      CALL cp_print_key_finished_output(print_forces, logger, force_env%force_env_section, "PRINT%FORCES")

      ! Write stress tensor
      IF (virial%pv_availability) THEN
         ! If the virial is defined but we are not computing forces let's zero the
         ! virial for consistency
         IF (calculate_forces .AND. calculate_stress_tensor) THEN
            output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%STRESS_TENSOR", &
                                               extension=".stress_tensor")
            IF (output_unit > 0) THEN
               CALL section_vals_val_get(force_env%force_env_section, "PRINT%STRESS_TENSOR%NDIGITS", &
                                         i_val=ndigits)
               CALL write_stress_tensor(virial%pv_virial, output_unit, cell, ndigits, virial%pv_numer)
            END IF
            CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                              "PRINT%STRESS_TENSOR")
         ELSE
            CALL zero_virial(virial, reset=.FALSE.)
         END IF
      ELSE
         output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%STRESS_TENSOR", &
                                            extension=".stress_tensor")
         IF (output_unit > 0) THEN
            CALL cp_warn(cp__l("force_env_methods.F",342), "To print the stress tensor switch on the "// &
                         "virial evaluation with the keyword: STRESS_TENSOR")
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                           "PRINT%STRESS_TENSOR")
      END IF

      ! Atomic energy
      output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".Log")
      IF (atprop_env%energy) THEN
         CALL mp_sum(atprop_env%atener, force_env%para_env%group)
         CALL force_env_get(force_env, potential_energy=e_pot)
         IF (output_unit > 0) THEN
            IF (logger%iter_info%print_level > low_print_level) THEN
               WRITE (UNIT=output_unit, FMT="(/,T6,A,T15,A)") "Atom", "Potential energy"
               WRITE (UNIT=output_unit, FMT="(T2,I8,1X,F20.10)") &
                  (i, atprop_env%atener(i), i=1, SIZE(atprop_env%atener))
            END IF
            sum_energy = accurate_sum(atprop_env%atener(:))
            checksum = ABS(e_pot-sum_energy)
            WRITE (UNIT=output_unit, FMT="(/,(T2,A,T56,F25.13))") &
               "Potential energy (Atomic):", sum_energy, &
               "Potential energy (Total) :", e_pot, &
               "Difference               :", checksum
            IF(.NOT.((checksum < ateps*ABS(e_pot))))CALL cp__a("force_env_methods.F",367)
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                           "PRINT%PROGRAM_RUN_INFO")
      END IF

      ! Atomic stress
      output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".Log")

      IF (atprop_env%stress) THEN
         CALL mp_sum(atprop_env%atstress, force_env%para_env%group)
         ! symmetrize (same as pv_virial)
         DO i = 1, SIZE(atprop_env%atstress, 3)
            atprop_env%atstress(:, :, i) = 0.5_dp*(atprop_env%atstress(:, :, i) &
                                                   +TRANSPOSE(atprop_env%atstress(:, :, i)))
         END DO
         IF (output_unit > 0) THEN
            IF (logger%iter_info%print_level > low_print_level) THEN
               DO i = 1, SIZE(atprop_env%atstress, 3)
                  WRITE (UNIT=output_unit, FMT="(/,T2,I0,T16,A1,2(19X,A1))") i, "X", "Y", "Z"
                  WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "X", (atprop_env%atstress(1, j, i), j=1, 3)
                  WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Y", (atprop_env%atstress(2, j, i), j=1, 3)
                  WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Z", (atprop_env%atstress(3, j, i), j=1, 3)
                  WRITE (UNIT=output_unit, FMT="(T2,A,F20.13)") "1/3 Trace(Atomic stress tensor):", &
                     (atprop_env%atstress(1, 1, i)+atprop_env%atstress(2, 2, i)+atprop_env%atstress(3, 3, i))/3.0_dp
               END DO
            END IF
            atomic_stress_tensor(:, :) = 0.0_dp
            DO i = 1, 3
               atomic_stress_tensor(i, i) = accurate_sum(atprop_env%atstress(i, i, :))
               DO j = i+1, 3
                  atomic_stress_tensor(i, j) = accurate_sum(atprop_env%atstress(i, j, :))
                  atomic_stress_tensor(j, i) = atomic_stress_tensor(i, j)
               END DO
            END DO
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T15,A1,2(19X,A1))") "Atomic", "X", "Y", "Z"
            WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "X", (atomic_stress_tensor(1, i), i=1, 3)
            WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Y", (atomic_stress_tensor(2, i), i=1, 3)
            WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Z", (atomic_stress_tensor(3, i), i=1, 3)
            WRITE (UNIT=output_unit, FMT="(T2,A,10X,F20.13)") "1/3 Trace(Atomic stress tensor):", &
               (atomic_stress_tensor(1, 1)+atomic_stress_tensor(2, 2)+atomic_stress_tensor(3, 3))/3.0_dp
            sum_stress_tensor = accurate_sum(atomic_stress_tensor(:, :))
            IF (virial%pv_availability .AND. calculate_forces) THEN
               WRITE (UNIT=output_unit, FMT="(/,T2,A,T16,A1,2(19X,A1))") "Total", "X", "Y", "Z"
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "X", (virial%pv_virial(1, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Y", (virial%pv_virial(2, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Z", (virial%pv_virial(3, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(T2,A,10X,F20.13)") "1/3 Trace(Total stress tensor): ", &
                  (virial%pv_virial(1, 1)+virial%pv_virial(2, 2)+virial%pv_virial(3, 3))/3.0_dp
               sum_pv_virial = SUM(virial%pv_virial(:, :))
               diff_stress_tensor(:, :) = ABS(virial%pv_virial(:, :)-atomic_stress_tensor(:, :))
               WRITE (UNIT=output_unit, FMT="(/,T2,A,T16,A1,2(19X,A1))") "Diff", "X", "Y", "Z"
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "X", (diff_stress_tensor(1, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Y", (diff_stress_tensor(2, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(A3,3F20.13)") "Z", (diff_stress_tensor(3, i), i=1, 3)
               WRITE (UNIT=output_unit, FMT="(T2,A,10X,F20.13)") "1/3 Trace(Diff)               : ", &
                  (diff_stress_tensor(1, 1)+diff_stress_tensor(2, 2)+diff_stress_tensor(3, 3))/3.0_dp
               checksum = accurate_sum(diff_stress_tensor(:, :))
               WRITE (UNIT=output_unit, FMT="(/,(T2,A,11X,F25.13))") &
                  "Checksum stress (Atomic) :", sum_stress_tensor, &
                  "Checksum stress (Total)  :", sum_pv_virial, &
                  "Difference               :", checksum
               IF(.NOT.(checksum < ateps))CALL cp__a("force_env_methods.F",430)
            END IF
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                           "PRINT%PROGRAM_RUN_INFO")
      END IF

   END SUBROUTINE force_env_calc_energy_force

! **************************************************************************************************
!> \brief Evaluates the stress tensor and pressure numerically
!> \param force_env ...
!> \param dx ...
!> \par History
!>      10.2005 created [JCS]
!>      05.2009 Teodoro Laino [tlaino] - rewriting for general force_env
!>
!> \author JCS
! **************************************************************************************************
   SUBROUTINE force_env_calc_num_pressure(force_env, dx)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: dx

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_calc_num_pressure', &
         routineP = moduleN//':'//routineN
      REAL(kind=dp), PARAMETER                           :: default_dx = 0.001_dp

      INTEGER                                            :: i, ip, iq, j, k, natom, ncore, ndigits, &
                                                            nshell, output_unit
      REAL(KIND=dp)                                      :: dx_w
      REAL(KIND=dp), DIMENSION(2)                        :: numer_energy
      REAL(KIND=dp), DIMENSION(3)                        :: s
      REAL(KIND=dp), DIMENSION(3, 3)                     :: numer_stress
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ref_pos_atom, ref_pos_core, ref_pos_shell
      TYPE(cell_type), POINTER                           :: cell, cell_local
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (cell_local)
      NULLIFY (core_particles)
      NULLIFY (particles)
      NULLIFY (shell_particles)
      NULLIFY (ref_pos_atom)
      NULLIFY (ref_pos_core)
      NULLIFY (ref_pos_shell)
      natom = 0
      ncore = 0
      nshell = 0
      numer_stress = 0.0_dp

      logger => cp_get_default_logger()

      dx_w = default_dx
      IF (PRESENT(dx)) dx_w = dx
      CALL force_env_get(force_env, subsys=subsys, globenv=globenv)
      CALL cp_subsys_get(subsys, &
                         core_particles=core_particles, &
                         particles=particles, &
                         shell_particles=shell_particles, &
                         virial=virial)
      output_unit = cp_print_key_unit_nr(logger, force_env%force_env_section, "PRINT%STRESS_TENSOR", &
                                         extension=".stress_tensor")
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(/A,A/)') ' **************************** ', &
            'NUMERICAL STRESS ********************************'
      END IF

      ! Save all original particle positions
      natom = particles%n_els
      ALLOCATE (ref_pos_atom(natom, 3))
      DO i = 1, natom
         ref_pos_atom(i, :) = particles%els(i)%r
      END DO
      IF (ASSOCIATED(core_particles)) THEN
         ncore = core_particles%n_els
         ALLOCATE (ref_pos_core(ncore, 3))
         DO i = 1, ncore
            ref_pos_core(i, :) = core_particles%els(i)%r
         END DO
      END IF
      IF (ASSOCIATED(shell_particles)) THEN
         nshell = shell_particles%n_els
         ALLOCATE (ref_pos_shell(nshell, 3))
         DO i = 1, nshell
            ref_pos_shell(i, :) = shell_particles%els(i)%r
         END DO
      END IF
      CALL force_env_get(force_env, cell=cell)
      CALL cell_create(cell_local)
      CALL cell_clone(cell, cell_local)
      ! First change box
      DO ip = 1, 3
         DO iq = 1, 3
            IF (virial%pv_diagonal .AND. (ip /= iq)) CYCLE
            DO k = 1, 2
               cell%hmat(ip, iq) = cell_local%hmat(ip, iq)-(-1.0_dp)**k*dx_w
               CALL init_cell(cell)
               ! Scale positions
               DO i = 1, natom
                  CALL real_to_scaled(s, ref_pos_atom(i, 1:3), cell_local)
                  CALL scaled_to_real(particles%els(i)%r, s, cell)
               END DO
               DO i = 1, ncore
                  CALL real_to_scaled(s, ref_pos_core(i, 1:3), cell_local)
                  CALL scaled_to_real(core_particles%els(i)%r, s, cell)
               END DO
               DO i = 1, nshell
                  CALL real_to_scaled(s, ref_pos_shell(i, 1:3), cell_local)
                  CALL scaled_to_real(shell_particles%els(i)%r, s, cell)
               END DO

               ! Compute energies
               CALL force_env_calc_energy_force(force_env, &
                                                calc_force=.FALSE., &
                                                consistent_energies=.TRUE., &
                                                calc_stress_tensor=.FALSE.)
               CALL force_env_get(force_env, potential_energy=numer_energy(k))
               ! Reset cell
               cell%hmat(ip, iq) = cell_local%hmat(ip, iq)
            END DO
            CALL init_cell(cell)
            numer_stress(ip, iq) = 0.5_dp*(numer_energy(1)-numer_energy(2))/dx_w
            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T7,A,F7.4,A,T27,A,F7.4,A,T49,A)") &
                  "E("//ACHAR(119+ip)//ACHAR(119+iq)//" +", dx_w, ")", &
                  "E("//ACHAR(119+ip)//ACHAR(119+iq)//" -", dx_w, ")", &
                  "f(numerical)"
               WRITE (UNIT=output_unit, FMT="(3(1X,F19.8))") &
                  numer_energy(1:2), numer_stress(ip, iq)
            END IF
         END DO
      END DO

      ! Reset positions and rebuild original environment
      DO i = 1, natom
         particles%els(i)%r = ref_pos_atom(i, :)
      END DO
      DO i = 1, ncore
         core_particles%els(i)%r = ref_pos_core(i, :)
      END DO
      DO i = 1, nshell
         shell_particles%els(i)%r = ref_pos_shell(i, :)
      END DO
      CALL force_env_calc_energy_force(force_env, &
                                       calc_force=.FALSE., &
                                       consistent_energies=.TRUE., &
                                       calc_stress_tensor=.FALSE.)

      ! Computing pv_test
      virial%pv_virial = 0.0_dp
      DO i = 1, 3
         DO j = 1, 3
            DO k = 1, 3
               virial%pv_virial(i, j) = virial%pv_virial(i, j)- &
                                        0.5_dp*(numer_stress(i, k)*cell_local%hmat(j, k)+ &
                                                numer_stress(j, k)*cell_local%hmat(i, k))
            END DO
         END DO
      END DO

      IF (output_unit > 0) THEN
         IF (globenv%run_type_id == debug_run) THEN
            CALL section_vals_val_get(force_env%force_env_section, "PRINT%STRESS_TENSOR%NDIGITS", &
                                      i_val=ndigits)
            CALL write_stress_tensor(virial%pv_virial, output_unit, cell, ndigits, virial%pv_numer)
         END IF
         WRITE (output_unit, '(/,A,/)') ' **************************** '// &
            'NUMERICAL STRESS END *****************************'
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, force_env%force_env_section, &
                                        "PRINT%STRESS_TENSOR")

      ! Release storage
      IF (ASSOCIATED(ref_pos_atom)) THEN
         DEALLOCATE (ref_pos_atom)
      END IF
      IF (ASSOCIATED(ref_pos_core)) THEN
         DEALLOCATE (ref_pos_core)
      END IF
      IF (ASSOCIATED(ref_pos_shell)) THEN
         DEALLOCATE (ref_pos_shell)
      END IF
      IF (ASSOCIATED(cell_local)) CALL cell_release(cell_local)

   END SUBROUTINE force_env_calc_num_pressure

! **************************************************************************************************
!> \brief creates and initializes a force environment
!> \param force_env the force env to create
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param fist_env , qs_env: exactly one of these should be
!>        associated, the one that is active
!> \param qs_env ...
!> \param meta_env ...
!> \param sub_force_env ...
!> \param qmmm_env ...
!> \param qmmmx_env ...
!> \param eip_env ...
!> \param force_env_section ...
!> \param mixed_env ...
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE force_env_create(force_env, root_section, para_env, globenv, fist_env, &
                               qs_env, meta_env, sub_force_env, qmmm_env, qmmmx_env, eip_env, force_env_section, &
                               mixed_env)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(fist_environment_type), OPTIONAL, POINTER     :: fist_env
      TYPE(qs_environment_type), OPTIONAL, POINTER       :: qs_env
      TYPE(meta_env_type), OPTIONAL, POINTER             :: meta_env
      TYPE(force_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: sub_force_env
      TYPE(qmmm_env_type), OPTIONAL, POINTER             :: qmmm_env
      TYPE(qmmmx_env_type), OPTIONAL, POINTER            :: qmmmx_env
      TYPE(eip_environment_type), OPTIONAL, POINTER      :: eip_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      TYPE(mixed_environment_type), OPTIONAL, POINTER    :: mixed_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (force_env)
      NULLIFY (force_env%fist_env, force_env%qs_env, &
               force_env%para_env, force_env%globenv, &
               force_env%meta_env, force_env%sub_force_env, &
               force_env%qmmm_env, force_env%qmmmx_env, force_env%fp_env, &
               force_env%force_env_section, force_env%eip_env, force_env%mixed_env, &
               force_env%root_section)
      last_force_env_id = last_force_env_id+1
      force_env%id_nr = last_force_env_id
      force_env%ref_count = 1
      force_env%in_use = 0
      force_env%additional_potential = 0.0_dp

      force_env%globenv => globenv
      CALL globenv_retain(force_env%globenv)

      force_env%root_section => root_section
      CALL section_vals_retain(root_section)

      force_env%para_env => para_env
      CALL cp_para_env_retain(force_env%para_env)

      CALL section_vals_retain(force_env_section)
      force_env%force_env_section => force_env_section

      IF (PRESENT(fist_env)) THEN
         IF(.NOT.(ASSOCIATED(fist_env)))CALL cp__a("force_env_methods.F",690)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",691)
         force_env%in_use = use_fist_force
         force_env%fist_env => fist_env
         CALL fist_env_retain(fist_env)
      END IF
      IF (PRESENT(eip_env)) THEN
         IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("force_env_methods.F",697)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",698)
         force_env%in_use = use_eip_force
         force_env%eip_env => eip_env
         CALL eip_env_retain(eip_env)
      END IF
      IF (PRESENT(qs_env)) THEN
         IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("force_env_methods.F",704)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",705)
         force_env%in_use = use_qs_force
         force_env%qs_env => qs_env
         CALL qs_env_retain(qs_env)
      END IF
      IF (PRESENT(qmmm_env)) THEN
         IF(.NOT.(ASSOCIATED(qmmm_env)))CALL cp__a("force_env_methods.F",711)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",712)
         force_env%in_use = use_qmmm
         force_env%qmmm_env => qmmm_env
         CALL qmmm_env_retain(qmmm_env)
      END IF
      IF (PRESENT(qmmmx_env)) THEN
         IF(.NOT.(ASSOCIATED(qmmmx_env)))CALL cp__a("force_env_methods.F",718)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",719)
         force_env%in_use = use_qmmmx
         force_env%qmmmx_env => qmmmx_env
         CALL qmmmx_env_retain(qmmmx_env)
      END IF
      IF (PRESENT(mixed_env)) THEN
         IF(.NOT.(ASSOCIATED(mixed_env)))CALL cp__a("force_env_methods.F",725)
         IF(.NOT.(force_env%in_use == 0))CALL cp__a("force_env_methods.F",726)
         force_env%in_use = use_mixed_force
         force_env%mixed_env => mixed_env
         CALL mixed_env_retain(mixed_env)
      END IF
      IF(.NOT.(force_env%in_use /= 0))CALL cp__a("force_env_methods.F",731)

      IF (PRESENT(sub_force_env)) THEN
         force_env%sub_force_env => sub_force_env
      END IF

      IF (PRESENT(meta_env)) THEN
         force_env%meta_env => meta_env
         CALL meta_env_retain(meta_env)
      ELSE
         NULLIFY (force_env%meta_env)
      END IF

   END SUBROUTINE force_env_create

! **************************************************************************************************
!> \brief ****f* force_env_methods/mixed_energy_forces  [1.0]
!>
!>     Computes energy and forces for a mixed force_env type
!> \param force_env ...
!> \param calculate_forces ...
!> \par History
!>       11.06  created [fschiff]
!>       04.07  generalization to an illimited number of force_eval [tlaino]
!>       04.07  further generalization to force_eval with different geometrical
!>              structures [tlaino]
!>       04.08  reorganizing the genmix structure (collecting common code)
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE mixed_energy_forces(force_env, calculate_forces)

      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_energy_forces', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length)               :: def_error, description, this_error
      INTEGER                                            :: iforce_eval, iparticle, jparticle, &
                                                            mixing_type, my_group, natom, &
                                                            nforce_eval, source, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: glob_natoms, map_index
      LOGICAL                                            :: dip_exists
      REAL(KIND=dp)                                      :: coupling_parameter, dedf, der_1, der_2, &
                                                            dx, energy, err, lambda, lerr, &
                                                            restraint_strength, restraint_target, &
                                                            sd
      REAL(KIND=dp), DIMENSION(3)                        :: dip_mix
      REAL(KIND=dp), DIMENSION(:), POINTER               :: energies
      TYPE(cell_type), POINTER                           :: cell_mix
      TYPE(cp_logger_type), POINTER                      :: logger, my_logger
      TYPE(cp_result_p_type), DIMENSION(:), POINTER      :: results
      TYPE(cp_result_type), POINTER                      :: loc_results, results_mix
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: subsystems
      TYPE(cp_subsys_type), POINTER                      :: subsys_mix
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy
      TYPE(mixed_force_type), DIMENSION(:), POINTER      :: global_forces
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles
      TYPE(particle_list_type), POINTER                  :: particles_mix
      TYPE(section_vals_type), POINTER                   :: force_env_section, gen_section, &
                                                            mapping_section, mixed_section, &
                                                            root_section
      TYPE(virial_p_type), DIMENSION(:), POINTER         :: virials
      TYPE(virial_type), POINTER                         :: loc_virial, virial_mix

      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_methods.F",798)
      ! Get infos about the mixed subsys
      CALL force_env_get(force_env=force_env, &
                         subsys=subsys_mix, &
                         force_env_section=force_env_section, &
                         root_section=root_section, &
                         cell=cell_mix)
      CALL cp_subsys_get(subsys=subsys_mix, &
                         particles=particles_mix, &
                         virial=virial_mix, &
                         results=results_mix)
      NULLIFY (map_index, glob_natoms, global_forces)

      nforce_eval = SIZE(force_env%sub_force_env)
      mixed_section => section_vals_get_subs_vals(force_env_section, "MIXED")
      mapping_section => section_vals_get_subs_vals(mixed_section, "MAPPING")
      ! Global Info
      ALLOCATE (subsystems(nforce_eval))
      ALLOCATE (particles(nforce_eval))
      ! Local Info to sync
      ALLOCATE (global_forces(nforce_eval))
      ALLOCATE (energies(nforce_eval))
      ALLOCATE (glob_natoms(nforce_eval))
      ALLOCATE (virials(nforce_eval))
      ALLOCATE (results(nforce_eval))
      energies = 0.0_dp
      glob_natoms = 0
      DO iforce_eval = 1, nforce_eval
         NULLIFY (subsystems(iforce_eval)%subsys, particles(iforce_eval)%list)
         NULLIFY (results(iforce_eval)%results, virials(iforce_eval)%virial)
         CALL virial_create(virials(iforce_eval)%virial)
         CALL cp_result_create(results(iforce_eval)%results)
         IF (.NOT. ASSOCIATED(force_env%sub_force_env(iforce_eval)%force_env)) CYCLE
         ! From this point on the error is the sub_error
         my_group = force_env%mixed_env%group_distribution(force_env%para_env%mepos)
         my_logger => force_env%mixed_env%sub_logger(my_group+1)%p
         ! Copy iterations info (they are updated only in the main mixed_env)
         CALL cp_iteration_info_copy_iter(logger%iter_info, my_logger%iter_info)
         CALL cp_add_default_logger(my_logger)

         ! Get all available subsys
         CALL force_env_get(force_env=force_env%sub_force_env(iforce_eval)%force_env, &
                            subsys=subsystems(iforce_eval)%subsys)

         ! all force_env share the same cell
         CALL cp_subsys_set(subsystems(iforce_eval)%subsys, cell=cell_mix)

         ! Get available particles
         CALL cp_subsys_get(subsys=subsystems(iforce_eval)%subsys, &
                            particles=particles(iforce_eval)%list)

         ! Get Mapping index array
         natom = SIZE(particles(iforce_eval)%list%els)
         CALL get_subsys_map_index(mapping_section, natom, iforce_eval, nforce_eval, &
                                   map_index)

         ! Mapping particles from iforce_eval environment to the mixed env
         DO iparticle = 1, natom
            jparticle = map_index(iparticle)
            particles(iforce_eval)%list%els(iparticle)%r = particles_mix%els(jparticle)%r
         END DO

         ! Calculate energy and forces for each sub_force_env
         CALL force_env_calc_energy_force(force_env%sub_force_env(iforce_eval)%force_env, &
                                          calc_force=calculate_forces, &
                                          skip_external_control=.TRUE.)
         ! Only the rank 0 process collect info for each computation
         IF (force_env%sub_force_env(iforce_eval)%force_env%para_env%mepos == &
             force_env%sub_force_env(iforce_eval)%force_env%para_env%source) THEN
            CALL force_env_get(force_env%sub_force_env(iforce_eval)%force_env, &
                               potential_energy=energy)
            CALL cp_subsys_get(subsystems(iforce_eval)%subsys, &
                               virial=loc_virial, results=loc_results)
            energies(iforce_eval) = energy
            glob_natoms(iforce_eval) = natom
            CALL cp_virial(loc_virial, virials(iforce_eval)%virial)
            CALL cp_result_copy(loc_results, results(iforce_eval)%results)
         END IF
         ! Deallocate map_index array
         IF (ASSOCIATED(map_index)) THEN
            DEALLOCATE (map_index)
         END IF
         CALL cp_rm_default_logger()
      END DO

      ! Handling Parallel execution
      CALL mp_sync(force_env%para_env%group)
      ! Let's transfer energy, natom, forces, virials
      CALL mp_sum(energies, force_env%para_env%group)
      CALL mp_sum(glob_natoms, force_env%para_env%group)
      ! Transfer forces
      DO iforce_eval = 1, nforce_eval
         ALLOCATE (global_forces(iforce_eval)%forces(3, glob_natoms(iforce_eval)))
         global_forces(iforce_eval)%forces = 0.0_dp
         IF (ASSOCIATED(force_env%sub_force_env(iforce_eval)%force_env)) THEN
            IF (force_env%sub_force_env(iforce_eval)%force_env%para_env%mepos == &
                force_env%sub_force_env(iforce_eval)%force_env%para_env%source) THEN
               ! Forces
               DO iparticle = 1, glob_natoms(iforce_eval)
                  global_forces(iforce_eval)%forces(:, iparticle) = &
                     particles(iforce_eval)%list%els(iparticle)%f
               END DO
            END IF
         END IF
         CALL mp_sum(global_forces(iforce_eval)%forces, force_env%para_env%group)
         !Transfer only the relevant part of the virial..
         CALL mp_sum(virials(iforce_eval)%virial%pv_total, force_env%para_env%group)
         CALL mp_sum(virials(iforce_eval)%virial%pv_kinetic, force_env%para_env%group)
         CALL mp_sum(virials(iforce_eval)%virial%pv_virial, force_env%para_env%group)
         CALL mp_sum(virials(iforce_eval)%virial%pv_xc, force_env%para_env%group)
         CALL mp_sum(virials(iforce_eval)%virial%pv_fock_4c, force_env%para_env%group)
         CALL mp_sum(virials(iforce_eval)%virial%pv_constraint, force_env%para_env%group)
         !Transfer results
         source = 0
         IF (ASSOCIATED(force_env%sub_force_env(iforce_eval)%force_env)) THEN
            IF (force_env%sub_force_env(iforce_eval)%force_env%para_env%mepos == &
                force_env%sub_force_env(iforce_eval)%force_env%para_env%source) &
               source = force_env%para_env%mepos
         ENDIF
         CALL mp_sum(source, force_env%para_env%group)
         CALL cp_results_mp_bcast(results(iforce_eval)%results, source, force_env%para_env)
      END DO

      force_env%mixed_env%energies = energies
      ! Start combining the different sub_force_env
      CALL get_mixed_env(mixed_env=force_env%mixed_env, &
                         mixed_energy=mixed_energy)

      !NB: do this for all MIXING_TYPE values, since some need it (e.g. linear mixing
      !NB if the first system has fewer atoms than the second)
      DO iparticle = 1, SIZE(particles_mix%els)
         particles_mix%els(iparticle)%f(:) = 0.0_dp
      END DO

      CALL section_vals_val_get(mixed_section, "MIXING_TYPE", i_val=mixing_type)
      SELECT CASE (mixing_type)
      CASE (mix_linear_combination)
         ! Support offered only 2 force_eval
         IF(.NOT.(nforce_eval == 2))CALL cp__a("force_env_methods.F",936)
         CALL section_vals_val_get(mixed_section, "LINEAR%LAMBDA", r_val=lambda)
         mixed_energy%pot = lambda*energies(1)+(1.0_dp-lambda)*energies(2)
         ! General Mapping of forces...
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               lambda, 1, nforce_eval, map_index, mapping_section, .TRUE.)
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               (1.0_dp-lambda), 2, nforce_eval, map_index, mapping_section, .FALSE.)
      CASE (mix_minimum)
         ! Support offered only 2 force_eval
         IF(.NOT.(nforce_eval == 2))CALL cp__a("force_env_methods.F",946)
         IF (energies(1) < energies(2)) THEN
            mixed_energy%pot = energies(1)
            CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                                  1.0_dp, 1, nforce_eval, map_index, mapping_section, .TRUE.)
         ELSE
            mixed_energy%pot = energies(2)
            CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                                  1.0_dp, 2, nforce_eval, map_index, mapping_section, .TRUE.)
         ENDIF
      CASE (mix_coupled)
         ! Support offered only 2 force_eval
         IF(.NOT.(nforce_eval == 2))CALL cp__a("force_env_methods.F",958)
         CALL section_vals_val_get(mixed_section, "COUPLING%COUPLING_PARAMETER", &
                                   r_val=coupling_parameter)
         sd = SQRT((energies(1)-energies(2))**2+4.0_dp*coupling_parameter**2)
         der_1 = (1.0_dp-(1.0_dp/(2.0_dp*sd))*2.0_dp*(energies(1)-energies(2)))/2.0_dp
         der_2 = (1.0_dp+(1.0_dp/(2.0_dp*sd))*2.0_dp*(energies(1)-energies(2)))/2.0_dp
         mixed_energy%pot = (energies(1)+energies(2)-sd)/2.0_dp
         ! General Mapping of forces...
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               der_1, 1, nforce_eval, map_index, mapping_section, .TRUE.)
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               der_2, 2, nforce_eval, map_index, mapping_section, .FALSE.)
      CASE (mix_restrained)
         ! Support offered only 2 force_eval
         IF(.NOT.(nforce_eval == 2))CALL cp__a("force_env_methods.F",972)
         CALL section_vals_val_get(mixed_section, "RESTRAINT%RESTRAINT_TARGET", &
                                   r_val=restraint_target)
         CALL section_vals_val_get(mixed_section, "RESTRAINT%RESTRAINT_STRENGTH", &
                                   r_val=restraint_strength)
         mixed_energy%pot = energies(1)+restraint_strength*(energies(1)-energies(2)-restraint_target)**2
         der_2 = -2.0_dp*restraint_strength*(energies(1)-energies(2)-restraint_target)
         der_1 = 1.0_dp-der_2
         ! General Mapping of forces...
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               der_1, 1, nforce_eval, map_index, mapping_section, .TRUE.)
         CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                               der_2, 2, nforce_eval, map_index, mapping_section, .FALSE.)
      CASE (mix_generic)
         ! Support any number of force_eval sections
         gen_section => section_vals_get_subs_vals(mixed_section, "GENERIC")
         CALL get_generic_info(gen_section, "MIXING_FUNCTION", coupling_function, force_env%mixed_env%par, &
                               force_env%mixed_env%val, energies)
         CALL initf(1)
         CALL parsef(1, TRIM(coupling_function), force_env%mixed_env%par)
         ! Now the hardest part.. map energy with corresponding force_eval
         mixed_energy%pot = evalf(1, force_env%mixed_env%val)
         IF(.NOT.(EvalErrType <= 0))CALL cp__a("force_env_methods.F",994)
         CALL zero_virial(virial_mix, reset=.FALSE.)
         CALL cp_results_erase(results_mix)
         DO iforce_eval = 1, nforce_eval
            CALL section_vals_val_get(gen_section, "DX", r_val=dx)
            CALL section_vals_val_get(gen_section, "ERROR_LIMIT", r_val=lerr)
            dedf = evalfd(1, iforce_eval, force_env%mixed_env%val, dx, err)
            IF (ABS(err) > lerr) THEN
               WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
               WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
               CALL compress(this_error, .TRUE.)
               CALL compress(def_error, .TRUE.)
               CALL cp_warn(cp__l("force_env_methods.F",1006), &
                            'ASSERTION (cond) failed at line '//cp_to_string(1007)// &
                            ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                            TRIM(def_error)//' .')
            END IF
            ! General Mapping of forces...
            CALL mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, virials, results, &
                                  dedf, iforce_eval, nforce_eval, map_index, mapping_section, .FALSE.)
            force_env%mixed_env%val(iforce_eval) = energies(iforce_eval)
         END DO
         ! Let's store the needed information..
         force_env%mixed_env%dx = dx
         force_env%mixed_env%lerr = lerr
         force_env%mixed_env%coupling_function = TRIM(coupling_function)
         CALL finalizef()
      CASE DEFAULT
         CALL cp__b("force_env_methods.F",1022,"")
      END SELECT
      !Simply deallocate and loose the pointer references..
      DO iforce_eval = 1, nforce_eval
         DEALLOCATE (global_forces(iforce_eval)%forces)
         CALL virial_release(virials(iforce_eval)%virial)
         CALL cp_result_release(results(iforce_eval)%results)
      END DO
      DEALLOCATE (global_forces)
      DEALLOCATE (subsystems)
      DEALLOCATE (particles)
      DEALLOCATE (energies)
      DEALLOCATE (glob_natoms)
      DEALLOCATE (virials)
      DEALLOCATE (results)
      ! Print Section
      unit_nr = cp_print_key_unit_nr(logger, mixed_section, "PRINT%DIPOLE", &
                                     extension=".data", middle_name="MIXED_DIPOLE", log_filename=.FALSE.)
      IF (unit_nr > 0) THEN
         description = '[DIPOLE]'
         dip_exists = test_for_result(results=results_mix, description=description)
         IF (dip_exists) THEN
            CALL get_results(results=results_mix, description=description, values=dip_mix)
            WRITE (unit_nr, '(/,1X,A,T48,3F11.6)') "MIXED ENV| DIPOLE  ( A.U.)|", dip_mix
            WRITE (unit_nr, '(  1X,A,T48,3F11.6)') "MIXED ENV| DIPOLE  (Debye)|", dip_mix*debye
         ELSE
            WRITE (unit_nr, *) "NO FORCE_EVAL section calculated the dipole"
         END IF
      END IF
      CALL cp_print_key_finished_output(unit_nr, logger, mixed_section, "PRINT%DIPOLE")
   END SUBROUTINE mixed_energy_forces

END MODULE force_env_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief creates the qmmm section of the input
!> \note
!>      moved out of input_cp2k
!> \par History
!>      10.2005 split out of input_cp2k
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_qmmm
   USE bibliography,                    ONLY: Bernstein2009,&
                                              Bernstein2012,&
                                              Golze2013,&
                                              Laino2005,&
                                              Laino2006
   USE cell_types,                      ONLY: use_perd_none
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE eri_mme_types,                   ONLY: create_eri_mme_section
   USE input_constants,                 ONLY: &
        alpha_imomm_default, charge_scale_factor, do_eri_gpw, do_eri_mme, &
        do_fm_mom_conserv_buffer, do_fm_mom_conserv_core, do_fm_mom_conserv_equal_a, &
        do_fm_mom_conserv_equal_f, do_fm_mom_conserv_none, do_fm_mom_conserv_qm, &
        do_multipole_section_off, do_multipole_section_on, do_par_atom, do_par_grid, &
        do_qmmm_center_every_step, do_qmmm_center_max_minus_min, do_qmmm_center_never, &
        do_qmmm_center_pbc_aware, do_qmmm_center_setup_only, do_qmmm_coulomb, do_qmmm_gauss, &
        do_qmmm_image_calcmatrix, do_qmmm_image_iter, do_qmmm_link_gho, do_qmmm_link_imomm, &
        do_qmmm_link_pseudo, do_qmmm_none, do_qmmm_pcharge, do_qmmm_swave, do_qmmm_wall_none, &
        do_qmmm_wall_quadratic, do_qmmm_wall_reflective, gaussian, radius_qmmm_default, &
        spline3_nopbc_interp
   USE input_cp2k_mm,                   ONLY: create_GENPOT_section,&
                                              create_Goodwin_section,&
                                              create_LJ_section,&
                                              create_NONBONDED14_section,&
                                              create_Williams_section
   USE input_cp2k_poisson,              ONLY: create_gspace_interp_section,&
                                              create_poisson_section
   USE input_cp2k_subsys,               ONLY: create_cell_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              logical_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE pw_spline_utils,                 ONLY: no_precond,&
                                              precond_spl3_1,&
                                              precond_spl3_2,&
                                              precond_spl3_3,&
                                              precond_spl3_aint,&
                                              precond_spl3_aint2
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_qmmm'

   PUBLIC :: create_qmmm_section

!***
CONTAINS

! **************************************************************************************************
!> \brief Creates the QM/MM section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_qmmm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",97)
      CALL section_create(section, name="qmmm", &
                          description="Input for QM/MM calculations.", &
                          n_keywords=6, n_subsections=3, repeats=.FALSE., &
                          citations=(/Laino2005, Laino2006/))

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="E_COUPL", &
                          variants=s2a("QMMM_COUPLING", "ECOUPL"), &
                          description="Specifies the type of the QM - MM electrostatic coupling.", &
                          usage="E_COUPL GAUSS", &
                          enum_c_vals=s2a("NONE", "COULOMB", "GAUSS", "S-WAVE", "POINT_CHARGE"), &
                          enum_i_vals=(/do_qmmm_none, do_qmmm_coulomb, do_qmmm_gauss, do_qmmm_swave, do_qmmm_pcharge/), &
                          enum_desc=s2a("Mechanical coupling (i.e. classical point charge based)", &
                                        "Using analytical 1/r potential (Coulomb) - not available for GPW/GAPW", &
                                        "Using fast gaussian expansion of the electrostatic potential (Erf(r/rc)/r)", &
                                        "Using fast gaussian expansion of the s-wave electrostatic potential", &
                                        "Using quantum mechanics derived point charges interacting with MM charges"), &
                          default_i_val=do_qmmm_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MM_POTENTIAL_FILE_NAME", &
                          description="Name of the file containing the potential expansion in gaussians. See the "// &
                          "USE_GEEP_LIB keyword.", &
                          usage="MM_POTENTIAL_FILE_NAME {filename}", &
                          default_lc_val="MM_POTENTIAL")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="use_geep_lib", &
                          description=" This keyword enables the use of the internal GEEP library to generate the "// &
                          "gaussian expansion of the MM potential. Using this keyword there's no need to provide "// &
                          "the MM_POTENTIAL_FILENAME. It expects a number from 2 to 15 (the number of gaussian funtions"// &
                          " to be used in the expansion.", &
                          usage="use_geep_lib INTEGER", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="nocompatibility", &
                          description="This keyword disables the compatibility of QM/MM "// &
                          "potential between CPMD and CP2K implementations. The compatibility"// &
                          " is achieved using an MM potential of the form: Erf[x/rc]/x + (1/rc -2/(pi^1/2*rc))*Exp[-(x/rc)^2] ."// &
                          "This keyword has effect only selecting GAUSS E_COUPLING type.", &
                          usage="nocompatibility LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_mm_rspace", &
                          description="Set the threshold for the collocation of the GEEP gaussian functions."// &
                          "this keyword affects only the GAUSS E_COUPLING.", &
                          usage="eps_mm_rspace real", &
                          default_r_val=1.0E-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SPHERICAL_CUTOFF", &
                          description="Set the spherical cutoff for the QMMM electrostatic interaction. "// &
                          "This acts like a charge multiplicative factor dependent on cutoff. For MM atoms "// &
                          "farther than the SPHERICAL_CUTOFF(1) their charge is zero. The switch is performed "// &
                          "with a smooth function: 0.5*(1-TANH((r-[SPH_CUT(1)-20*SPH_CUT(2)])/(SPH_CUT(2)))). "// &
                          "Two values are required: the first one is the distance cutoff. The second one controls "// &
                          "the stiffness of the smoothing.", &
                          usage="SPHERICAL_CUTOFF  <REAL>", default_r_vals=(/-1.0_dp, 0.0_dp/), n_var=2, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="parallel_scheme", &
                          description="Chooses the parallel_scheme for the long range Potential ", &
                          usage="parallel_scheme (ATOM|GRID)", &
                          enum_c_vals=s2a("ATOM", "GRID"), &
                          enum_desc=s2a("parallelizes on atoms. grids replicated. "// &
                                        "Replication of the grids can be quite expensive memory wise if running on a system "// &
                                        "with limited memory per core. The grid option may be preferred in this case.", &
                                        "parallelizes on grid slices. atoms replicated."), &
                          enum_i_vals=(/do_par_atom, do_par_grid/), &
                          default_i_val=do_par_atom)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Centering keywords
      CALL keyword_create(keyword, name="CENTER", &
                          description="This keyword sets when the QM system is automatically "// &
                          "centered.  Default is EVERY_STEP.", &
                          usage="center (EVERY_STEP|SETUP_ONLY|NEVER)", &
                          enum_c_vals=s2a("EVERY_STEP", "SETUP_ONLY", "NEVER"), &
                          enum_desc=s2a("Re-center every step", &
                                        "Center at first step only", &
                                        "Never center"), &
                          enum_i_vals=(/do_qmmm_center_every_step, do_qmmm_center_setup_only, do_qmmm_center_never/), &
                          default_i_val=do_qmmm_center_every_step)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CENTER_TYPE", &
                          description="This keyword specifies how to do the QM system centering.", &
                          usage="center_type (MAX_MINUS_MIN|PBC_AWARE_MAX_MINUS_MIN)", &
                          enum_c_vals=s2a("MAX_MINUS_MIN", "PBC_AWARE_MAX_MINUS_MIN"), &
                          enum_desc=s2a("Center of box defined by maximum coordinate minus minimum coordinate", &
                                        "PBC-aware centering (useful for &QMMM&FORCE_MIXING)"), &
                          enum_i_vals=(/do_qmmm_center_max_minus_min, do_qmmm_center_pbc_aware/), &
                          default_i_val=do_qmmm_center_max_minus_min)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CENTER_GRID", &
                          description="This keyword specifies whether the QM system is centered in units of the grid spacing.", &
                          usage="grid_center LOGICAL", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="initial_translation_vector", &
                          description="This keyword specify the initial translation vector to be applied to the system.", &
                          usage="initial_translation_vector <REAL> <REAL> <REAL>", &
                          n_var=3, default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="DELTA_CHARGE", &
         description="Additional net charge relative to that specified in DFT section.  Used automatically by force mixing", &
         usage="DELTA_CHARGE q", default_i_val=0, &
         n_var=1, type_of_var=integer_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! NB: remember to create these
      CALL create_qmmm_force_mixing_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_qm_kinds(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_mm_kinds(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cell_section(subsection, periodic=use_perd_none)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_periodic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_link_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_forcefield_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_walls_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_image_charge_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_qmmm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_section

! **************************************************************************************************
!> \brief Input section to create MM kinds sections
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_mm_kinds(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_mm_kinds', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",288)
      CALL section_create(section, name="MM_KIND", &
                          description="Information about the MM kind in the QM/MM scheme", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The MM  kind", usage="O", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADIUS", &
                          description="Specifies the radius of the atomic kinds", &
                          usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom", &
                          default_r_val=cp_unit_to_cp2k(RADIUS_QMMM_DEFAULT, "angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORR_RADIUS", &
                          description="Specifies the correction radius of the atomic kinds"// &
                          " The correction radius is connected to the use of the compatibility keyword.", &
                          usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_qmmm_mm_kinds

! **************************************************************************************************
!> \brief Input section to create FORCE_MIXING sections
!> \param section ...
!> \author noam
! **************************************************************************************************
   SUBROUTINE create_qmmm_force_mixing_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_force_mixing_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: link_subsection, print_key, &
                                                            qm_kinds_subsection, subsection

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",330)
      CALL section_create(section, name="FORCE_MIXING", &
                          description="This section enables and defines parameters for force-mixing based QM/MM,"// &
                          " which actually does two conventional QM/MM calculations, on a small "// &
                          " and a large QM region, and combines the MM forces from one and QM "// &
                          " forces from the other to create a complete set of forces.  Energy is "// &
                          " not conserved (although the QM/MM energy from the large QM region calculation is reported) "// &
                          " so a proper thermostat (i.e. massive, and able to handle dissipation, such as "// &
                          " Adaptive Langevin (AD_LANGEVIN)) must be used. For some propagation algorithms "// &
                          " (NVT and REFTRAJ MD ensembles) algorithm is adaptive, "// &
                          " including molecules hysteretically based on their instantaneous distance from the core region. "// &
                          " Information on core/QM/buffer labels can be written in PDB file using "// &
                          " MOTION&PRINT&FORCE_MIXING_LABELS.  Will fail if calculation requires a "// &
                          " meaningfull stress, or an energy that is consistent with the forces. "// &
                          " For GEO_OPT this means "// &
                          " only MOTION&GEO_OPT&TYPE CG, MOTION&GEO_OPT&CG&LINE_SEARCH&TYPE 2PNT, and "// &
                          " MOTION&GEO_OPT&CG&LINE_SEARCH&2PNT&LINMIN_GRAD_ONLY T", &
                          n_keywords=5, n_subsections=3, repeats=.FALSE., &
                          citations=(/Bernstein2009, Bernstein2012/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Enables force-mixing", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOMENTUM_CONSERVATION_TYPE", &
                          description="How to apply force to get momentum conservation", &
                          usage="MOMENTUM_CONSERVATION_TYPE <type>", &
                          enum_c_vals=s2a("NONE", "EQUAL_F", "EQUAL_A"), &
                          enum_i_vals=(/do_fm_mom_conserv_none, do_fm_mom_conserv_equal_f, do_fm_mom_conserv_equal_a/), &
                          enum_desc=s2a("No momentum conservation", &
                                        "Equal force on each atom", &
                                        "Equal acceleration on each atom"), &
                          default_i_val=do_fm_mom_conserv_equal_a)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOMENTUM_CONSERVATION_REGION", &
                          description="Region to apply correction force to for momentum conservation", &
                          usage="MOMENTUM_CONSERVATION_REGION <label>", &
                          enum_c_vals=s2a("CORE", "QM", "BUFFER"), &
                          enum_i_vals=(/do_fm_mom_conserv_core, do_fm_mom_conserv_QM, do_fm_mom_conserv_buffer/), &
                          enum_desc=s2a("Apply to QM core region", &
                                        "Apply to full QM (dynamics) region", &
                                        "Apply to QM+buffer regions"), &
                          default_i_val=do_fm_mom_conserv_QM)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R_CORE", &
                          description="Specify the inner and outer radii of core QM region. "// &
                          " All molecules with any atoms within this distance (hysteretically) of any atoms "// &
                          " specified as QM in enclosing QM/MM section  will be core QM atoms in the force-mixing calculation.", &
                          usage="R_CORE <real> <real>", n_var=2, type_of_var=real_t, &
                          default_r_vals=(/cp_unit_to_cp2k(0.0_dp, "angstrom"), &
                                           cp_unit_to_cp2k(0.0_dp, "angstrom")/), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R_QM", &
                          description="Specify the inner and outer radii of QM dynamics region. "// &
                          " All molecules with atoms within this distance (hysteretically) of any atoms in "// &
                          " core will follow QM dynamics in the force-mixing calculation.", &
                          usage="R_QM <real> <real>", n_var=2, type_of_var=real_t, &
                          default_r_vals=(/cp_unit_to_cp2k(0.5_dp, "angstrom"), &
                                           cp_unit_to_cp2k(1.0_dp, "angstrom")/), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QM_EXTENDED_SEED_IS_ONLY_CORE_LIST", &
                          description="Makes the extended QM zone be defined hysterestically "// &
                          " by distance from QM core list (i.e. atoms specified explicitly by "// &
                          " user) instead of from full QM core region (specified by user + hysteretic "// &
                          " selection + unbreakable bonds)", &
                          usage="QM_EXTENDED_SEED_IS_ONLY_CORE_LIST <logical>", n_var=1, type_of_var=logical_t, &
                          default_l_val=.FALSE., repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R_BUF", &
                          description="Specify the inner and outer radii of buffer region. "// &
                          " All atoms within this distance (hysteretically) of any QM atoms "// &
                          " will be buffer atoms in the force-mixing calculation.", &
                          usage="R_BUF <real> <real>", n_var=2, type_of_var=real_t, &
                          default_r_vals=(/cp_unit_to_cp2k(0.5_dp, "angstrom"), &
                                           cp_unit_to_cp2k(1.0_dp, "angstrom")/), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QM_KIND_ELEMENT_MAPPING", &
                          description="Mapping from elements to QM_KINDs for adaptively included atoms.", &
                          usage="QM_KIND_ELEMENT_MAPPING {El} {QM_KIND}", &
                          n_var=2, type_of_var=char_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_N_QM", &
                          description="Maximum number of QM atoms, for detection of runaway adaptive selection.", &
                          usage="MAX_N_QM int", default_i_val=300, &
                          n_var=1, type_of_var=integer_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ADAPTIVE_EXCLUDE_MOLECULES", &
                          description="List of molecule names to exclude from adaptive regions (e.g. big things like proteins)", &
                          usage="ADAPTIVE_EXCLUDE_MOLECULES molec1 molec2 ...", &
                          n_var=-1, type_of_var=char_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXTENDED_DELTA_CHARGE", &
                          description="Additional net charge in extended region relative to core (core charge is "// &
                          " specified in DFT section, as usual for a convetional QM/MM calculation)", &
                          usage="EXTENDED_DELTA_CHARGE q", default_i_val=0, &
                          n_var=1, type_of_var=integer_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! QM_NON_ADAPTIVE subsection
      NULLIFY (subsection)
      CALL section_create(subsection, name="QM_NON_ADAPTIVE", &
                          description="List of atoms always in QM region, non-adaptively", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      NULLIFY (qm_kinds_subsection)
      CALL create_qmmm_qm_kinds(qm_kinds_subsection)
      CALL section_add_subsection(subsection, qm_kinds_subsection)
      CALL section_release(qm_kinds_subsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! BUFFER_NON_ADAPTIVE subsection
      NULLIFY (subsection)
      CALL section_create(subsection, name="BUFFER_NON_ADAPTIVE", &
                          description="List of atoms always in buffer region, non-adaptively, and any needed LINK sections", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      NULLIFY (qm_kinds_subsection)
      CALL create_qmmm_qm_kinds(qm_kinds_subsection)
      CALL section_add_subsection(subsection, qm_kinds_subsection)
      CALL section_release(qm_kinds_subsection)
      NULLIFY (link_subsection)
      CALL create_qmmm_link_section(link_subsection)
      CALL section_add_subsection(subsection, link_subsection)
      CALL section_release(link_subsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ![NB] also need a list?
      ![NB] maybe not list+links , but some sort of link template
      ![NB] also, breakable bonds?
      ! BUFFER_LINKS subsection
      NULLIFY (subsection)
      CALL section_create( &
         subsection, name="BUFFER_LINKS", &
         description="Information about possible links for automatic covalent bond breaking for the buffer QM/MM calculation."// &
         "Ignored - need to implement buffer selection by atom and walking of connectivity data.", &
         n_keywords=0, n_subsections=1, repeats=.TRUE.)

      NULLIFY (link_subsection)
      CALL create_qmmm_link_section(link_subsection)
      CALL section_add_subsection(subsection, link_subsection)
      CALL section_release(link_subsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! RESTART_INFO subsection
      NULLIFY (subsection)
      CALL section_create(subsection, name="RESTART_INFO", &
                          description="This section provides information about old force-mixing indices and labels, "// &
                          "for restarts.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="INDICES", &
                          description="Indices of atoms in previous step QM regions.", &
                          usage="INDICES 1 2 ...", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LABELS", &
                          description="Labels of atoms in previous step QM regions.", &
                          usage="LABELS 1 1 ...", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! PRINT subsection, with keys for neighbor list
      CALL section_create(subsection, name="print", &
                          description="Section of possible print options in FORCE_MIXING.", &
                          n_keywords=0, n_subsections=2, repeats=.FALSE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "SUBCELL", &
                                       description="Activates the printing of the subcells used for the"// &
                                       "generation of neighbor lists.", unit_str="angstrom", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "NEIGHBOR_LISTS", &
                                       description="Activates the printing of the neighbor lists used"// &
                                       " for the hysteretic region calculations.", &
                                       print_level=high_print_level, filename="", unit_str="angstrom")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_force_mixing_section

! **************************************************************************************************
!> \brief Input section to create QM kinds sections
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_qm_kinds(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_qm_kinds', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",565)
      CALL section_create(section, name="QM_KIND", &
                          description="Information about the QM kind in the QM/MM scheme", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The QM kind", usage="O", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MM_INDEX", &
                          description="The indexes of the MM atoms that have this kind. This keyword can be"// &
                          " repeated several times (useful if you have to specify many indexes).", &
                          usage="MM_INDEX 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_qmmm_qm_kinds

! **************************************************************************************************
!> \brief Input section to set QM/MM periodic boundary conditions
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_walls_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_walls_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",599)
      CALL section_create(section, name="WALLS", &
                          description="Enables Walls for the QM box. This can be used to avoid that QM "// &
                          " atoms move out of the QM box.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="WALL_SKIN", &
                          description="Specify the value of the skin of the Wall in each dimension. "// &
                          "The wall's effect is felt when atoms fall within the skin of the Wall.", &
                          usage="WALL_SKIN <real> <real> <real>", n_var=3, type_of_var=real_t, &
                          default_r_vals=(/cp_unit_to_cp2k(0.5_dp, "angstrom"), &
                                           cp_unit_to_cp2k(0.5_dp, "angstrom"), &
                                           cp_unit_to_cp2k(0.5_dp, "angstrom")/), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE", &
                          description="Specifies the type of wall", &
                          usage="TYPE REFLECTIVE", &
                          enum_c_vals=s2a("NONE", "REFLECTIVE", "QUADRATIC"), &
                          enum_i_vals=(/do_qmmm_wall_none, do_qmmm_wall_reflective, do_qmmm_wall_quadratic/), &
                          enum_desc=s2a("No Wall around QM box", &
                                        "Reflective Wall around QM box", &
                                        "Quadratic Wall around QM box"), &
                          default_i_val=do_qmmm_wall_reflective)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Specify the value of the the force constant for the quadratic wall", &
                          usage="K <real>", unit_str='internal_cp2k', &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_qmmm_walls_section

! ****************************************************************************
!> \brief Input section for QM/MM image charge calculations
!> \param section ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_qmmm_image_charge_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_image_charge_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",652)
      CALL section_create(section, name="IMAGE_CHARGE", &
                          description="Inclusion of polarization effects within the image charge "// &
                          "approach for systems where QM molecules are physisorbed on e.g. metal "// &
                          "surfaces described by MM. QM box size has to be equal to MM box size.", &
                          n_keywords=3, n_subsections=1, repeats=.FALSE., &
                          citations=(/Golze2013/))

      CALL keyword_create(keyword, name="MM_ATOM_LIST", &
                          description="List of MM atoms carrying an induced Gaussian charge. "// &
                          "If this keyword is not given, all MM atoms will carry an image charge.", &
                          usage="MM_ATOM_LIST 1 2 3 or 1..3 ", n_var=-1, type_of_var=integer_t, &
                          repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WIDTH", &
                          description="Specifies the value of the width of the (induced) Gaussian "// &
                          "charge distribution carried by each MM atom.", &
                          usage="WIDTH <real> ", n_var=1, type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=3.0_dp, unit_str="angstrom^-2"), &
                          unit_str="angstrom^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXT_POTENTIAL", &
                          description="External potential applied to the metal electrode ", &
                          usage="EXT_POTENTIAL <real> ", n_var=1, type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="volt")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DETERM_COEFF", &
                          description="Specifies how the coefficients are determined.", &
                          usage="DETERM_COEFF ITERATIVE", &
                          enum_c_vals=s2a("CALC_MATRIX", "ITERATIVE"), &
                          enum_i_vals=(/do_qmmm_image_calcmatrix, do_qmmm_image_iter/), &
                          enum_desc=s2a("Calculates image matrix and solves linear set of equations", &
                                        "Uses an iterative scheme to calculate the coefficients"), &
                          default_i_val=do_qmmm_image_calcmatrix)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_IMAGE_MATRIX", &
                          description="Restart the image matrix. Useful when "// &
                          "calculating coefficients iteratively (the image matrix "// &
                          "is used as preconditioner in that case)", &
                          usage="RESTART_IMAGE_MATRIX", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IMAGE_RESTART_FILE_NAME", &
                          description="File name where to read the image matrix used "// &
                          "as preconditioner in the iterative scheme", &
                          usage="IMAGE_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IMAGE_MATRIX_METHOD", &
                          description="Method for calculating the image matrix. The MME method is "// &
                          "much faster than GPW s.t. the overhead of the IC approach compared to "// &
                          "standard QM/MM becomes negligible in total computation time. Note that MME "// &
                          "is an approximative method that has limited accuracy in the current implementation.", &
                          usage="IMAGE_MATRIX_METHOD MME", &
                          enum_c_vals=s2a("GPW", "MME"), &
                          enum_i_vals=(/do_eri_gpw, do_eri_mme/), &
                          enum_desc=s2a("Uses Gaussian Plane Wave method [Golze2013]", &
                                        "Uses MiniMax-Ewald method (ERI_MME subsection)"), &
                          default_i_val=do_eri_gpw)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_eri_mme_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_image_charge_section

! **************************************************************************************************
!> \brief Input section to set QM/MM periodic boundary conditions
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_periodic_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_periodic_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",748)
      CALL section_create(section, name="PERIODIC", &
                          description="Specify parameters for QM/MM periodic boundary conditions calculations", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Laino2006/))

      CALL keyword_create( &
         keyword, name="GMAX", &
         description="Specifies the maximum value of G in the reciprocal space over which perform the Ewald sum.", &
         usage="GMAX <real>", n_var=1, default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REPLICA", &
                          description="Specifies the number of replica to take into consideration for the real part of the "// &
                          "calculation. Default is letting the qmmm module decide how many replica you really need.", &
                          usage="REPLICA <integer>", n_var=1, default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGRIDS", &
                          description="Specifies the number of grid points used for the Interpolation of the G-space term", &
                          usage="NGRIDS <integer> <iteger> <integer> ", n_var=3, default_i_vals=(/50, 50, 50/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_multipole_qmmm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_gspace_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_poisson_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, "check_spline", &
                                       description="Controls the checking of the G-space term Spline Interpolation.", &
                                       print_level=medium_print_level, filename="GSpace-SplInterp")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_periodic_section

! **************************************************************************************************
!> \brief Section to set-up parameters for decoupling using the Bloechl scheme
!> \param section the section to create
!> \par History
!>      Dorothea Golze [04.2014] copied from input_cp2k_poisson.F and
!>      enabled switch-on/off
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_multipole_qmmm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multipole_qmmm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",811)

      CALL section_create(section, name="MULTIPOLE", &
                          description="This section is used to set up the decoupling of QM periodic images with "// &
                          "the use of density derived atomic point charges. Switched on by default even if not "// &
                          "explicitly given. Can be switched off if e.g. QM and MM box are of the same size.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Defines the usage of the multipole section", &
                          usage="ON", &
                          enum_c_vals=s2a("ON", "OFF"), &
                          enum_i_vals=(/do_multipole_section_on, do_multipole_section_off/), &
                          enum_desc=s2a("switch on MULTIPOLE section", &
                                        "switch off MULTIPOLE section"), &
                          default_i_val=do_multipole_section_on, lone_keyword_i_val=do_multipole_section_on)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Real space cutoff for the Ewald sum.", &
                          usage="RCUT {real}", n_var=1, type_of_var=real_t, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EWALD_PRECISION", &
                          description="Precision achieved in the Ewald sum.", &
                          usage="EWALD_PRECISION {real}", n_var=1, type_of_var=real_t, &
                          unit_str="hartree", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANALYTICAL_GTERM", &
                          description="Evaluates the Gterm in the Ewald Scheme analytically instead of using Splines.", &
                          usage="ANALYTICAL_GTERM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGRIDS", &
                          description="Specifies the number of grid points used for the Interpolation of the G-space term", &
                          usage="NGRIDS <integer> <iteger> <integer> ", n_var=3, default_i_vals=(/50, 50, 50/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_gspace_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, "check_spline", &
                                       description="Controls the checking of the G-space term Spline Interpolation.", &
                                       print_level=medium_print_level, filename="GSpace-SplInterp")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, "program_run_info", &
                                       description="Controls the printing of basic information during the run", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_multipole_qmmm_section

! **************************************************************************************************
!> \brief creates the qm/mm forcefield section to override to the FF specification
!>      given in the FIST input
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_forcefield_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_forcefield_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection, keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",892)
      CALL section_create(section, name="FORCEFIELD", &
                          description="Specify information on the QM/MM forcefield", &
                          n_keywords=0, n_subsections=2, repeats=.TRUE.)

      CALL keyword_create(keyword, name="MULTIPLE_POTENTIAL", &
                          description="Enables the possibility to define NONBONDED and NONBONDED14 as a"// &
                          " sum of different kinds of potential. Useful for piecewise defined potentials.", &
                          usage="MULTIPLE_POTENTIAL T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_qmmm_ff_nb_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_NONBONDED14_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_forcefield_section

! **************************************************************************************************
!> \brief creates the qm/mm forcefield section to override to the FF specification
!>      given in the FIST input - NONBONDED PART
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_ff_nb_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_ff_nb_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",929)
      CALL section_create(section, name="NONBONDED", &
                          description="Specify information on the QM/MM non-bonded forcefield", &
                          n_keywords=0, n_subsections=2, repeats=.TRUE.)

      CALL create_LJ_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Williams_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Goodwin_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_GENPOT_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qmmm_ff_nb_section

! **************************************************************************************************
!> \brief creates the qm/mm link section
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_link_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_link_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",967)
      CALL section_create(section, name="LINK", &
                          description="Specify information on the QM/MM link treatment", &
                          n_keywords=7, n_subsections=2, repeats=.TRUE.)

      CALL keyword_create(keyword, name="QM_INDEX", &
                          variants=(/"QM"/), &
                          description="Specifies the index of the QM atom involved in the QM/MM link", &
                          usage="QM_INDEX integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QM_KIND", &
                          description="Specifies the element of the QM capping atom involved in the QM/MM link", &
                          usage="QM_KIND char", n_var=1, type_of_var=char_t, &
                          default_c_val="H")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MM_INDEX", &
                          variants=(/"MM"/), &
                          description="Specifies the index of the MM atom involved in the QM/MM link, Default hydrogen.", &
                          usage="MM_INDEX integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADIUS", &
                          description="Overwrite the specification of the radius only for the MM atom involved in the link."// &
                          "Default is to use the same radius as for the specified type.", &
                          usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="CORR_RADIUS", &
         description="Overwrite the specification of the correction radius only for the MM atom involved in the link."// &
         "Default is to use the same correction radius as for the specified type.", &
         usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LINK_TYPE", &
                          variants=(/"LINK ", "TYPE ", "LTYPE"/), &
                          description="Specifies the method to use to treat the defined QM/MM link", &
                          usage="LINK_TYPE char", &
                          enum_c_vals=s2a("IMOMM", "GHO", "PSEUDO"), &
                          enum_i_vals=(/do_qmmm_link_imomm, do_qmmm_link_gho, do_qmmm_link_pseudo/), &
                          enum_desc=s2a("Use Generalized Hybrid Orbital method", &
                                        "Use Integrated Molecular Orbital Molecular Mechanics method", &
                                        "Use a monovalent pseudo-potential"), &
                          default_i_val=do_qmmm_link_imomm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA_IMOMM", &
                          variants=s2a("ALPHA"), &
                          description="Specifies the scaling factor to be used for projecting the forces "// &
                          "on the capping hydrogen in the IMOMM QM/MM link scheme to the MM atom of the link. "// &
                          "A good guess can be derived from the bond distances of the forcefield: "// &
                          "alpha = r_eq(QM-MM) / r_eq(QM-H).", &
                          usage="ALPHA_IMOMM real", n_var=1, type_of_var=real_t, &
                          default_r_val=ALPHA_IMOMM_DEFAULT)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QMMM_SCALE_FACTOR", &
                          variants=(/"QMMM_CHARGE_SCALE ", &
                                     "QMMM_CHARGE_FACTOR", &
                                     "QMMM_SCALE_CHARGE "/), &
                          description="Specifies the scaling factor for the MM charge involved in the link QM/MM."// &
                          " This keyword affects only the QM/MM potential, it doesn't affect the electrostatic in "// &
                          " the classical part of the code. "// &
                          "Default 1.0 i.e. no charge rescaling of the MM atom of the QM/MM link bond.", &
                          usage="SCALE_FACTOR real", n_var=1, type_of_var=real_t, &
                          default_r_val=CHARGE_SCALE_FACTOR)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIST_SCALE_FACTOR", &
                          variants=(/"FIST_CHARGE_SCALE ", &
                                     "FIST_CHARGE_FACTOR", &
                                     "FIST_SCALE_CHARGE "/), &
                          description="Specifies the scaling factor for the MM charge involved in the link QM/MM."// &
                          " This keyword modifies the MM charge in FIST. The modified charge will be used then also"// &
                          " for the generation of the QM/MM potential. "// &
                          "Default 1.0 i.e. no charge rescaling of the MM atom of the QM/MM link bond.", &
                          usage="SCALE_FACTOR real", n_var=1, type_of_var=real_t, &
                          default_r_val=CHARGE_SCALE_FACTOR)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="MOVE_MM_CHARGE", &
                          description="Specify information to move a classical charge before the"// &
                          " QM/MM energies and forces evaluation", &
                          n_keywords=4, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="ATOM_INDEX_1", &
                          variants=(/"MM1"/), &
                          description="Specifies the index of the MM atom involved in the QM/MM link to be moved", &
                          usage="ATOM_INDEX_1 integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_INDEX_2", &
                          variants=(/"MM2"/), &
                          description="Specifies the index of the second atom defining the direction along which "// &
                          " the atom will be moved", &
                          usage="ATOM_INDEX_2 integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA", &
                          description="Specifies the scaling factor that defines the movement along the defined direction", &
                          usage="ALPHA real", n_var=1, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADIUS", &
                          description="Specifies the radius used for the QM/MM electrostatic coupling after movement", &
                          usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORR_RADIUS", &
                          description="Specifies the correction radius used for the QM/MM electrostatic coupling after movement", &
                          usage="RADIUS real", n_var=1, type_of_var=real_t, unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="ADD_MM_CHARGE", &
                          description="Specify information to add a classical charge before the"// &
                          " QM/MM energies and forces evaluation", &
                          n_keywords=5, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="ATOM_INDEX_1", &
                          variants=(/"MM1"/), &
                          description="Specifies the index of the first atom defining the direction along which"// &
                          " the atom will be added", &
                          usage="ATOM_INDEX_1 integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_INDEX_2", &
                          variants=(/"MM2"/), &
                          description="Specifies the index of the second atom defining the direction along which "// &
                          " the atom will be added", &
                          usage="ATOM_INDEX_2 integer", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA", &
                          description="Specifies the scaling factor that defines the movement along the defined direction", &
                          usage="ALPHA real", n_var=1, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADIUS", &
                          description="Specifies the radius used for the QM/MM electrostatic coupling for the added source", &
                          usage="RADIUS real", n_var=1, unit_str="angstrom", &
                          default_r_val=cp_unit_to_cp2k(RADIUS_QMMM_DEFAULT, "angstrom"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="CORR_RADIUS", &
         description="Specifies the correction radius used for the QM/MM electrostatic coupling for the added source", &
         usage="RADIUS real", n_var=1, unit_str="angstrom", &
         default_r_val=cp_unit_to_cp2k(RADIUS_QMMM_DEFAULT, "angstrom"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE", &
                          description="Specifies the charge for the added source of QM/MM potential", &
                          usage="CHARGE real", default_r_val=0.0_dp, n_var=1, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_qmmm_link_section

! **************************************************************************************************
!> \brief creates the interpolation section
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_qmmm_interp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qmmm_interp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",1165)
      CALL section_create(section, name="interpolator", &
                          description="kind of interpolation used between the multigrids", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="kind", &
                          description="the interpolator to use", &
                          usage="kind spline3", &
                          default_i_val=spline3_nopbc_interp, &
                          enum_c_vals=s2a("spline3_nopbc"), &
                          enum_i_vals=(/spline3_nopbc_interp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="safe_computation", &
                          description="if a non unrolled calculation is to be performed in parallel", &
                          usage="safe_computation OFF", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="aint_precond", &
                          description="the approximate inverse to use to get the starting point"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_aint, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="precond", &
                          description="The preconditioner used"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_3, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_x", &
                          description="accuracy on the solution for spline3 the interpolators", &
                          usage="eps_x 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_r", &
                          description="accuracy on the residual for spline3 the interpolators", &
                          usage="eps_r 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="max_iter", &
                          variants=(/'maxiter'/), &
                          description="the maximum number of iterations", &
                          usage="max_iter 200", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "conv_info", &
                                       description="if convergence information about the linear solver"// &
                                       " of the spline methods should be printed", &
                                       print_level=medium_print_level, each_iter_names=s2a("SPLINE_FIND_COEFFS"), &
                                       each_iter_values=(/10/), filename="__STD_OUT__", &
                                       add_last=add_last_numeric)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "spl_coeffs", &
                                       description="outputs a cube with the coefficents calculated for "// &
                                       "the spline interpolation", &
                                       print_level=debug_print_level)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_qmmm_interp_section

! **************************************************************************************************
!> \brief Create the print qmmm section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_print_qmmm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_qmmm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_qmmm.F",1264)
      NULLIFY (keyword, print_key)
      CALL section_create(section, name="print", &
                          description="Section of possible print options specific of the QMMM code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "DIPOLE", &
                                       description="Controls the printing of the DIPOLE in a QM/MM calculations."// &
                                       " It requires that the DIPOLE calculations is "// &
                                       " requested both for the QS  and for the MM  part.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PGF", &
                                       description="Controls the printing of the gaussian expansion basis set of the"// &
                                       " electrostatic potential", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "POTENTIAL", &
                                       description="Controls the printing of the QMMM  potential", &
                                       print_level=high_print_level, filename="MM_ELPOT_QMMM", &
                                       common_iter_levels=1)

      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "MM_POTENTIAL", &
                                       description="Controls the printing of the MM unidimensional potential on file", &
                                       print_level=high_print_level, filename="MM_ELPOT", &
                                       common_iter_levels=1)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "QMMM_MATRIX", &
                                       description="Controls the printing of the QMMM 1 electron Hamiltonian Matrix"// &
                                       " for methods like semiempirical and DFTB", &
                                       print_level=high_print_level, filename="__STD_OUT__", &
                                       common_iter_levels=1)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_BANNER", &
                                       description="Controls the printing of the banner of the MM program", &
                                       print_level=silent_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of information regarding the run.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "PERIODIC_INFO", &
         description="Controls the printing of information regarding the periodic boundary condition.", &
         print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "GRID_INFORMATION", &
                                       description="Controls the printing of information regarding the PW grid structures"// &
                                       " for PERIODIC QM/MM calculations.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "derivatives", &
                                       description="Print all derivatives after QM/MM calculation", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "qmmm_charges", &
                                       description="Print all charges generating the QM/MM potential", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "qmmm_link_info", &
                                       description="Print all information on QM/MM links", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "qs_derivatives", &
                                       description="Print QM derivatives after QS calculation", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "image_charge_info", &
                                       description="Prints image charge coefficients and detailed energy info", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "image_charge_restart", &
                                       description="Controls the printing of the restart file for "// &
                                       "the image matrix when using the iterative scheme", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="RESTART", &
                                       common_iter_levels=3)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_print_qmmm_section

END MODULE input_cp2k_qmmm
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Creates the EIP section of the input
!> \par History
!>      03.2006 created
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
MODULE input_cp2k_eip
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              high_print_level,&
                                              medium_print_level
   USE input_constants,                 ONLY: use_bazant_eip,&
                                              use_lenosky_eip
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: enum_t
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_eip'

   PUBLIC :: create_eip_section

CONTAINS

! **************************************************************************************************
!> \brief Create the input section for EIP
!> \param section the section to create
!> \par History
!>      03.2006 created
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE create_eip_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eip_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

!   ------------------------------------------------------------------------

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_eip.F",58)
      CALL section_create(section, name="EIP", &
                          description="This section contains all information to run an "// &
                          "Empirical Interatomic Potential (EIP) calculation.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="EIP_MODEL", &
                          description="Selects the empirical interaction potential model", &
                          usage="EIP_MODEL BAZANT", type_of_var=enum_t, &
                          n_var=1, repeats=.FALSE., variants=(/"EIP-MODEL"/), &
                          enum_c_vals=s2a("BAZANT", "EDIP", "LENOSKY"), &
                          enum_i_vals=(/use_bazant_eip, use_bazant_eip, use_lenosky_eip/), &
                          enum_desc=s2a("Bazant potentials", &
                                        "Environment-Dependent Interatomic Potential", &
                                        "Lenosky potentials"), &
                          default_i_val=use_lenosky_eip)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_eip_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_eip_section

! **************************************************************************************************
!> \brief Creates the print section for the eip subsection
!> \param section the section to create
!> \par History
!>      03.2006 created
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE create_eip_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eip_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

!   ------------------------------------------------------------------------

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_eip.F",102)
      CALL section_create(section, name="PRINT", &
                          description="Section of possible print options in EIP code.", &
                          n_keywords=0, n_subsections=6, repeats=.FALSE.)

      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "ENERGIES", &
                                       description="Controls the printing of the EIP energies.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ENERGIES_VAR", &
                                       description="Controls the printing of the variance of the EIP energies.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FORCES", &
                                       description="Controls the printing of the EIP forces.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COORD_AVG", &
                                       description="Controls the printing of the average coordination number.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COORD_VAR", &
                                       description="Controls the printing of the variance of the coordination number.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COUNT", &
                                       description="Controls the printing of the number of function calls.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_eip_print_section

END MODULE input_cp2k_eip
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for the MIXED environment
!> \par History
!>      10.2008 created [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
MODULE input_cp2k_mixed
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              low_print_level,&
                                              medium_print_level
   USE input_constants,                 ONLY: mix_coupled,&
                                              mix_generic,&
                                              mix_linear_combination,&
                                              mix_minimum,&
                                              mix_restrained
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_mixed'

   PUBLIC :: create_mix_section

CONTAINS

! **************************************************************************************************
!> \brief Create the input section for MIXED.
!> \param section the section to create
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE create_mix_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mix_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: sub2section, sub3section, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mixed.F",62)
      CALL section_create(section, name="MIXED", &
                          description="This section contains all information to run with a hamiltonian "// &
                          "defined by a mixing of force_evals", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create( &
         keyword, name="MIXING_TYPE", &
         description="The type of mixing to be employed", &
         usage="MIXING_TYPE LINEAR_COMBINATION", &
         default_i_val=mix_linear_combination, &
         enum_c_vals=s2a("LINEAR_COMBINATION", &
                         "MINIMUM", &
                         "COUPLED", &
                         "RESTRAINT", &
                         "GENMIX"), &
         enum_desc=s2a("Linear combination of force envs (support only 2 force_evals)", &
                       "Use the force env with the minimum energy (support only 2 force_evals)", &
                       "Consider the force envs as a two state system with a given"// &
                       " coupling matrix element (support only 2 force_evals)", &
                       "Use the difference between the energy of the force envs as a"// &
                       " restraint on the first (support only 2 force_evals)", &
                       "Defines a user-driven generica coupling (support for an unlimited number of force_eval)"), &
         enum_i_vals=(/mix_linear_combination, mix_minimum, mix_coupled, mix_restrained, mix_generic/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GROUP_PARTITION", &
                          description="gives the exact number of processors for each group."// &
                          " If not specified processors allocated will be equally distributed for"// &
                          " the specified subforce_eval, trying to build a number of groups equal to the"// &
                          " number of subforce_eval specified.", &
                          usage="group_partition  2 2 4 2 4 ", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGROUPS", variants=(/"NGROUP"/), &
                          description="Gives the wanted number of groups. If not specified the number"// &
                          " of groups is set to the number of subforce_eval defined.", &
                          usage="ngroups 4", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Double force_eval
      CALL section_create(subsection, name="LINEAR", &
                          description="Linear combination between two force_eval:  F= lambda F1 + (1-lambda) F2", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Specify the mixing parameter lambda in the formula.", &
                          usage="lambda <REAL>", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="COUPLING", &
                          description="Coupling between two force_eval: E=(E1+E2 - sqrt((E1-E2)**2+4*H12**2))/2", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="COUPLING_PARAMETER", &
                          description="Coupling parameter H12 used in the coupling", &
                          usage="COUPLING_PARAMETER <REAL>", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="RESTRAINT", &
                          description="Restraint between two force_eval: E = E1 + k*(E1-E2-t)**2", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="RESTRAINT_TARGET", &
                          description="Target value of the restraint (t) ", &
                          usage="RESTRAINT_TARGET <REAL>", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTRAINT_STRENGTH", &
                          description="Strength of the restraint (k) in "// &
                          "k*(E1-E2-t)**2", &
                          usage="RESTRAINT_STRENGTH <REAL>", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Multiple force_eval
      CALL section_create(subsection, name="GENERIC", &
                          description="User driven coupling between two or more force_eval.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="MIXING_FUNCTION", &
                          description="Specifies the mixing functional form in mathematical notation.", &
                          usage="MIXING_FUNCTION (E1+E2-LOG(E1/E2))", type_of_var=lchar_t, &
                          n_var=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VARIABLES", &
                          description="Defines the variables of the functional form. To allow an efficient"// &
                          " mapping the order of the energy variables will be considered identical to the"// &
                          " order of the force_eval in the force_eval_order list.", &
                          usage="VARIABLES x", type_of_var=char_t, &
                          n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNITS", &
                          description="Optionally, allows to define valid CP2K unit strings for each parameter value. "// &
                          "It is assumed that the corresponding parameter value is specified in this unit.", &
                          usage="UNITS angstrom eV*angstrom^-1 angstrom^1 K", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Parameter used for computing the derivative with the Ridders' method.", &
                          usage="DX <REAL>", default_r_val=0.1_dp, unit_str="bohr")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERROR_LIMIT", &
                          description="Checks that the error in computing the derivative is not larger than "// &
                          "the value set; in case error is larger a warning message is printed.", &
                          usage="ERROR_LIMIT <REAL>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Mapping of atoms
      NULLIFY (sub2section, sub3section)
      CALL section_create(subsection, name="MAPPING", &
                          description="Defines the mapping of atoms for the different force_eval with the mixed force_eval."// &
                          " The default is to have a mapping 1-1 between atom index (i.e. all force_eval share the same"// &
                          " geometrical structure). The mapping is based on defining fragments and the mapping the "// &
                          " fragments between the several force_eval and the mixed force_eval", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      ! Mixed force_eval
      CALL section_create(sub2section, name="FORCE_EVAL_MIXED", &
                          description="Defines the fragments for the mixed force_eval (reference)", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL section_create(sub3section, name="FRAGMENT", &
                          description="Fragment definition", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Defines the index of the fragment defined", &
                          usage="<INTEGER>", type_of_var=integer_t, n_var=1)
      CALL section_add_keyword(sub3section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Starting and ending atomic index defining one fragment must be provided", &
                          usage="{Integer} {Integer}", type_of_var=integer_t, n_var=2, repeats=.TRUE.)
      CALL section_add_keyword(sub3section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(sub2section, sub3section)
      CALL section_release(sub3section)
      CALL section_add_subsection(subsection, sub2section)
      CALL section_release(sub2section)

      ! All other force_eval
      CALL section_create(sub2section, name="FORCE_EVAL", &
                          description="Defines the fragments and the mapping for each force_eval (an integer index (ID) "// &
                          "needs to be provided as parameter)", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create( &
         keyword, name="DEFINE_FRAGMENTS", &
         description="Specify the fragments definition of the force_eval through the fragments of the"// &
         " force_eval_mixed. This avoids the pedantic definition of the fragments for the force_eval,"// &
         " assuming the order of the fragments for the specified force_eval is the same as the sequence "// &
         " of integers provided. Easier to USE should be preferred to the specification of the single fragments.", &
         usage="DEFINE_FRAGMENTS <INTEGER> .. <INTEGER>", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(sub2section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Defines the index of the force_eval for which fragments and mappings are provided", &
                          usage="<INTEGER>", type_of_var=integer_t, n_var=1)
      CALL section_add_keyword(sub2section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(sub3section, name="FRAGMENT", &
                          description="Fragment definition", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Defines the index of the fragment defined", &
                          usage="<INTEGER>", type_of_var=integer_t, n_var=1)
      CALL section_add_keyword(sub3section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Starting and ending atomic index defining one fragment must be provided", &
                          usage="{Integer} {Integer}", type_of_var=integer_t, n_var=2, repeats=.FALSE.)
      CALL section_add_keyword(sub3section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAP", &
                          description="Provides the index of the fragment of the MIXED force_eval mapped on the"// &
                          " locally defined fragment.", &
                          usage="MAP <INTEGER>", type_of_var=integer_t, n_var=1, repeats=.FALSE.)
      CALL section_add_keyword(sub3section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(sub2section, sub3section)
      CALL section_release(sub3section)
      CALL section_add_subsection(subsection, sub2section)
      CALL section_release(sub2section)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_mix_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_mix_section

! **************************************************************************************************
!> \brief Create the print section for mixed
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_print_mix_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_mix_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mixed.F",310)
      CALL section_create(section, name="print", &
                          description="Section of possible print options in MIXED env.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of information during the evaluation of "// &
                                       "the mixed environment. ", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DIPOLE", &
                                       description="Controls the printing of dipole information. "// &
                                       "Requires the DIPOLE calculation be active for all subforce_eval.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_print_mix_section

END MODULE input_cp2k_mixed
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that builds the resp section of the input
!> \par History
!>      02.2007 created
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE input_cp2k_resp
   USE bibliography,                    ONLY: Campana2009,&
                                              Golze2015,&
                                              Rappe1992
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        do_resp_minus_x_dir, do_resp_minus_y_dir, do_resp_minus_z_dir, do_resp_x_dir, &
        do_resp_y_dir, do_resp_z_dir, gaussian, use_cambridge_vdw_radii, use_uff_vdw_radii
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_resp'

   PUBLIC :: create_resp_section

CONTAINS

! **************************************************************************************************
!> \brief Creates the RESP section
!> \param section the section to create
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_resp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_resp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",63)
      CALL section_create(section, name="RESP", &
                          description="Requests a RESP fit of charges. When using a periodic "// &
                          "Poisson solver and a periodic cell, the periodic RESP routines are "// &
                          "used. If the Hartree potential matches with the one of an isolated "// &
                          "system (i.e. isolated Poisson solver and big, nonperiodic cells), "// &
                          "the nonperiodic RESP routines are automatically used. All restraints "// &
                          "are harmonic!", &
                          n_keywords=2, n_subsections=2, repeats=.FALSE., citations=(/Golze2015/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide "// &
                          "3 numbers (for X,Y,Z) or 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTEGER_TOTAL_CHARGE", &
                          description="Forces the total charge to be integer", &
                          usage="INTEGER_TOTAL_CHARGE TRUE", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTRAIN_HEAVIES_TO_ZERO", &
                          description="Restrain non-hydrogen atoms to zero.", &
                          usage="RESTRAIN_HEAVIES_TO_ZERO FALSE", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTRAIN_HEAVIES_STRENGTH", &
                          description="If defined, enforce the restraint of non-hydrogen "// &
                          "atoms to zero. Its value is the strength of the restraint on "// &
                          "the heavy atoms.", &
                          usage="RESTRAIN_HEAVIES_STRENGTH 0.0001 ", &
                          default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WIDTH", &
                          description="Specifies the value of the width of the Gaussian "// &
                          "charge distribution carried by each atom. Needs only "// &
                          "to be specified when using a periodic Poisson solver.", &
                          usage="WIDTH <real> ", n_var=1, type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=11.249_dp, unit_str="angstrom^-2"), &
                          unit_str="angstrom^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_REPEAT_METHOD", &
                          description="Fits the variance of the potential, i.e. the deviation "// &
                          "from the mean value of the potential within the selected "// &
                          "range. The evaluation of the potentials is still treated "// &
                          "within the GPW approach as described in [Golze2015]. "// &
                          "When used in conjunction with INTEGER_TOTAL_CHARGE = T "// &
                          "and SPHERE_SAMPLING, the results will be very similar to "// &
                          "the REPEAT charges given in [Campana2009]. In most "// &
                          "cases switching on this option gives reasonable "// &
                          "atomic charges without the need to define any "// &
                          "restraints. Note that by switching on this option, "// &
                          "RESTRAIN_HEAVIES_TO_ZERO will be switched off. ", &
                          usage="USE_REPEAT_METHOD", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE., &
                          citations=(/Campana2009/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_constraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_restraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_sphere_sampling_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_slab_sampling_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_resp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_resp_section

! **************************************************************************************************
!> \brief specifies constraints to be satisfied in a resp fit
!> \param section the section to create
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_constraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_constraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",169)
      CALL section_create(section, name="CONSTRAINT", &
                          description="specifies a linear constraint on the fitted charges."// &
                          "This can be used to give equal values to equivalent atoms."// &
                          "sum over atom_list c_i * q_i = t", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="the target value for the constraint", &
                          usage="TARGET 0.0", &
                          n_var=1, default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EQUAL_CHARGES", &
                          description="All atoms in ATOM_LIST are constrained to have the "// &
                          "same charges. When using this keyword, TARGET and ATOM_COEF do "// &
                          "not need to be set and will be ignored. Instead of using this "// &
                          "keyword, the constraint section could be repeated.", &
                          usage="EQUAL_CHARGES", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_LIST", &
                          description="Defines the list of atoms involved in this constraint", &
                          usage="ATOM_LIST 3 4", &
                          type_of_var=integer_t, n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_COEF", &
                          description="Defines the coefficient of the atom in this "// &
                          "linear constraint", &
                          usage="ATOM_COEF 1.0 -1.0", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_constraint_section

! **************************************************************************************************
!> \brief specifies restraints to be added to a resp fit
!> \param section the section to create
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_restraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_restraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",225)
      CALL section_create(section, name="RESTRAINT", &
                          description="specifies a restraint on the fitted charges."// &
                          "This can be used to restrain values to zero."// &
                          "s*(sum over atom_list q_i - t)**2", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="the target value for the restraint", &
                          usage="TARGET 0.0", &
                          n_var=1, default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="the target value for the constraint", &
                          usage="STRENGTH 0.001", &
                          n_var=1, default_r_val=0.001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_LIST", &
                          description="Defines the list of atoms involved in this restraint", &
                          usage="ATOM_LIST 3 4", &
                          type_of_var=integer_t, n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_COEF", &
                          description="Defines the coefficient of the atom in this "// &
                          "linear restraint. If given, the restraint will be: "// &
                          "s*(sum over atom_list c_i * q_i - t)**2 ", &
                          usage="ATOM_COEF 1.0 -1.0", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_restraint_section

! **************************************************************************************************
!> \brief specifies the parameter for sampling the resp fitting points for
!>        molecular structures; sampling in spheres around the atoms
!> \param section the section to create
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_sphere_sampling_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_sphere_sampling_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",280)
      CALL section_create(section, name="SPHERE_SAMPLING", &
                          description="Specifies the parameter for sampling the RESP fitting points "// &
                          "for molecular structures, i.e. systems that do not involve "// &
                          "surfaces. Fitting points are sampled in spheres around the "// &
                          "atom. All grid points in the shell defined by rmin and rmax "// &
                          "are accepted for fitting. Default is that rmin is the vdW "// &
                          "radius and rmax=100.0*vdW_radius, which can be overwritten "// &
                          "by the keywords below.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="X_LOW", &
                          description="Specifies the lower boundary of the box along X used to "// &
                          "sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="X_LOW  -15.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_HI", &
                          description="Specifies the upper boundary of the box along X used to "// &
                          " sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="X_HI  5.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Y_LOW", &
                          description="Specifies the lower boundary of the box along Y used to "// &
                          "sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="Y_LOW  -15.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Y_HI", &
                          description="Specifies the upper boundary of the box along Y used to "// &
                          "sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="Y_HI  5.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Z_LOW", &
                          description="Specifies the lower boundary of the box along Z used to "// &
                          "sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="Z_LOW  -15.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Z_HI", &
                          description="Specifies the upper boundary of the box along Z used to "// &
                          "sample the potential. Only valid for nonperiodic RESP fitting.", &
                          usage="Z_HI  5.", type_of_var=real_t, n_var=1, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AUTO_VDW_RADII_TABLE", &
                          description="Select which vdW radii table to use for automatic "// &
                          "determination of RMIN_KIND and RMAX_KIND if those "// &
                          "are not declared explicitly", &
                          usage="AUTO_VDW_RADII_TABLE UFF", &
                          default_i_val=use_cambridge_vdw_radii, &
                          enum_c_vals=s2a("CAMBRIDGE", &
                                          "UFF"), &
                          enum_desc=s2a("Cambridge Structural Database", &
                                        "Universal Force Field: "// &
                                        "Rappe et al. J. Am. Chem. Soc. 114, 10024 (1992)"), &
                          enum_i_vals=(/use_cambridge_vdw_radii, &
                                        use_uff_vdw_radii/), &
                          citations=(/Rappe1992/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AUTO_RMAX_SCALE", &
                          description="IF RMAX or RMAX_KIND keywords are not present, defines the "// &
                          "maximumn distance a fit point is away from an atom based on "// &
                          "the formula: rmax(kind) = AUTO_RMAX_SCALE * vdW_radius(kind). "// &
                          "The van der Waals radiii of the elements are based on data from "// &
                          "table chosen by AUTO_VDW_RADII_TABLE.", &
                          usage="AUTO_RMAX_SCALE 60.0", &
                          default_r_val=100.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AUTO_RMIN_SCALE", &
                          description="IF RMIN or RMIN_KIND keywords are not present, defines the "// &
                          "minimum distance a fit point is away from an atom based on "// &
                          "the formula: rmin(kind) = AUTO_RMIN_SCALE * vdW_radius(kind). "// &
                          "The van der Waals radii of the elements are based on data from "// &
                          "table chosen by AUTO_VDW_RADII_TABLE.", &
                          usage="AUTO_RMIN_SCALE 1.5", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Specifies the maximum distance a fit point is away from an atom. "// &
                          "Valid for all atomic kinds for which no RMAX_KIND are specified.", &
                          usage="RMAX 2.5", &
                          default_r_val=cp_unit_to_cp2k(value=2.5_dp, unit_str="angstrom"), &
                          unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Specifies the minimum distance a fit point is away from an atom. "// &
                          "Valid for all atomic kinds for which no RMIN_KIND are specified.", &
                          usage="RMIN 2.1", &
                          default_r_val=cp_unit_to_cp2k(value=2.1_dp, unit_str="angstrom"), &
                          unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX_KIND", &
                          description="Specifies the maximum distance a fit point is away from an atom "// &
                          "of a given kind", &
                          usage="RMAX 2.5 Br", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN_KIND", &
                          description="Specifies the minimum distance a fit point is away from an atom "// &
                          "of a given kind", &
                          usage="RMIN 2.1 Br", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_sphere_sampling_section

! **************************************************************************************************
!> \brief specifies the parameter for sampling the resp fitting points for
!>        slab-like periodic systems, i.e. systems that involve surfaces
!> \param section the section to create
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_slab_sampling_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_slab_sampling_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",424)
      CALL section_create(section, name="SLAB_SAMPLING", &
                          description="Specifies the parameter for sampling the RESP fitting "// &
                          "points for slab-like periodic systems, i.e. systems that "// &
                          "involve surfaces. This section can only be used with periodic  "// &
                          "Poisson solver and cell. To see, which grid points were "// &
                          "used, switch on COORD_FIT_POINTS in the PRINT section.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOM_LIST", &
                          description="Specifies the list of indexes of atoms used to define "// &
                          "the region for the RESP fitting. The list should "// &
                          "contain indexes of atoms of the first surface layer.", &
                          usage="ATOM_LIST 1 2 3 or 1..3", &
                          type_of_var=integer_t, n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANGE", &
                          description="Range where the fitting points are sampled. A range of "// &
                          "3 to 5 Angstroms means that the fitting points are sampled in the region "// &
                          "of 3 to 5 Angstroms above the surface which is defined by atom indexes given "// &
                          "in ATOM_LIST.", &
                          usage="RANGE <real> <real>", unit_str="angstrom", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LENGTH", &
                          description="Length of the sampling box, i.e. a box of this length and "// &
                          "the height specified by RANGE is defined above each surface atom given "// &
                          "in ATOM_LIST. The grid points in the boxes are accepted as fitting point. "// &
                          "Should be in the range of the nearest neighbour distance (a bit larger to be "// &
                          "on the safe side). Allows for a refined sampling of grid points in case of "// &
                          "corrugated surfaces.", &
                          usage="LENGTH <real> ", unit_str="angstrom", n_var=1, type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=3.0_dp, unit_str="angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SURF_DIRECTION", &
                          description="Specifies what above the surface means. Defines the direction.", &
                          usage="SURF_DIRECTION Z", &
                          enum_c_vals=s2a("X", "Y", "Z", "-X", "-Y", "-Z"), &
                          enum_i_vals=(/do_resp_x_dir, do_resp_y_dir, do_resp_z_dir, &
                                        do_resp_minus_x_dir, do_resp_minus_y_dir, do_resp_minus_z_dir/), &
                          enum_desc=s2a("surface layers are piled up in x-direction", &
                                        "surface layers are piled up in y-direction", &
                                        "surface layers are piled up in z-direction", &
                                        "surface layers are piled up in -x-direction", &
                                        "surface layers are piled up in -y-direction", &
                                        "surface layers are piled up in -z-direction"), &
                          default_i_val=do_resp_z_dir)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_slab_sampling_section

! **************************************************************************************************
!> \brief create the resp print section
!> \param section the section to create
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_print_resp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_resp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_resp.F",497)
      NULLIFY (print_key, keyword)
      CALL section_create(section, name="print", &
                          description="Section of possible print options specific for the RESP code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of information regarding the run.", &
                                       print_level=low_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COORD_FIT_POINTS", &
                                       description="Controls the printing of the coordinates of the "// &
                                       "grid points used for periodic RESP fitting. This section "// &
                                       "is intended to be only used for testing (you can get large files).", &
                                       print_level=high_print_level, add_last=add_last_numeric, &
                                       filename="RESP_FIT_POINTS", &
                                       common_iter_levels=3)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESP_CHARGES_TO_FILE", &
                                       description="Controls the printing of the RESP charges "// &
                                       "to a file.", &
                                       print_level=high_print_level, add_last=add_last_numeric, &
                                       filename="RESP_CHARGES", &
                                       common_iter_levels=3)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "V_RESP_CUBE", &
                                       description="Controls the printing of the potential generated "// &
                                       "by the RESP CHARGES to a cube file. Prints the relative "// &
                                       "root-mean-square (RRMS) and root-mean-square (RMS) errors.", &
                                       print_level=high_print_level, add_last=add_last_numeric, &
                                       filename="RESP_POTENTIAL", &
                                       common_iter_levels=3)
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_print_resp_section

END MODULE input_cp2k_resp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief input section for atomic properties
!> \par History
!>      07.2011 created
!> \author JHU
! **************************************************************************************************
MODULE input_cp2k_atprop
   USE bibliography,                    ONLY: Kikuchi2009
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_atprop'

   PUBLIC :: create_atprop_section

CONTAINS

! **************************************************************************************************
!> \brief Creates the ATOMIC section
!> \param section the section to create
!> \author JHU
! **************************************************************************************************
   SUBROUTINE create_atprop_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atprop_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atprop.F",45)
      CALL section_create(section, name="ATOMIC", &
                          description="Controls the calculation of atomic properties. "// &
                          "Printing is controled by FORCE_EVAL / PRINT / PROGRAM_RUN_INFO", &
                          repeats=.FALSE., &
                          citations=(/Kikuchi2009/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ENERGY", &
                          description="Calculate atomic energies ", &
                          usage="ENERGY {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRESSURE", &
                          description="Calculate atomic pressure tensors ", &
                          usage="PRESSURE {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_atprop_section

END MODULE input_cp2k_atprop
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that build the dft section of the input
!> \par History
!>      01.2013 moved out of input_cp2k_dft [MI]
!> \author MI
! **************************************************************************************************
MODULE input_cp2k_properties_dft

   USE bibliography,                    ONLY: Luber2014,&
                                              Putrino2000,&
                                              Putrino2002,&
                                              Sebastiani2001,&
                                              Weber2009
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        current_gauge_atom, current_gauge_r, current_gauge_r_and_step_func, &
        current_orb_center_atom, current_orb_center_box, current_orb_center_common, &
        current_orb_center_wannier, do_ddapc_constraint, do_ddapc_restraint, do_et_becke, &
        do_et_ddapc, do_full_density, do_no_et, do_spin_density, ot_precond_full_all, &
        ot_precond_full_kinetic, ot_precond_full_single, ot_precond_full_single_inverse, &
        ot_precond_none, ot_precond_s_inverse
   USE input_cp2k_atprop,               ONLY: create_atprop_section
   USE input_cp2k_dft,                  ONLY: create_interp_section,&
                                              create_localize_section
   USE input_cp2k_resp,                 ONLY: create_resp_section
   USE input_cp2k_xc,                   ONLY: create_xc_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_properties_dft'

   PUBLIC :: create_properties_section

CONTAINS

! **************************************************************************************************
!> \brief Create the PROPERTIES section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_properties_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_properties_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",75)
      CALL section_create(section, name="PROPERTIES", &
                          description="This section is used to set up the PROPERTIES calculation.", &
                          n_keywords=4, n_subsections=1, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL create_linres_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_et_coupling_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_resp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_atprop_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, name="FIT_CHARGE", &
                                       description="This section is used to print the density derived atomic point charges."// &
                                       "The fit of the charges is controlled through the DENSITY_FITTING section", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="TYPE_OF_DENSITY", &
                          description="Specifies the type of density used for the fitting", &
                          usage="TYPE_OF_DENSITY (FULL|SPIN)", &
                          enum_c_vals=s2a("FULL", "SPIN"), &
                          enum_i_vals=(/do_full_density, do_spin_density/), &
                          enum_desc=s2a("Full density", "Spin density"), &
                          default_i_val=do_full_density)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_properties_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>      a linear response calculation
!>      Available properties : none
!> \param section the section to create
!> \author MI
! **************************************************************************************************
   SUBROUTINE create_linres_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_linres_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, subsection, print_key)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",134)
      CALL section_create(section, name="linres", &
                          description="The linear response is used to calculate one of the "// &
                          " following properties: nmr, epr, raman, ... ", &
                          n_keywords=5, n_subsections=2, repeats=.FALSE., &
                          citations=(/Putrino2000/))

      CALL keyword_create(keyword, name="EPS", &
                          description="target accuracy for the convergence of the conjugate gradient.", &
                          usage="EPS 1.e-6", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of conjugate gradient iteration to be performed for one optimization.", &
                          usage="MAX_ITER 200", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_EVERY", &
                          description="Restart the conjugate gradient after the specified number of iterations.", &
                          usage="RESTART_EVERY 200", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PRECONDITIONER", &
         description="Type of preconditioner to be used with all minimization schemes. "// &
         "They differ in effectiveness, cost of construction, cost of application. "// &
         "Properly preconditioned minimization can be orders of magnitude faster than doing nothing.", &
         usage="PRECONDITIONER FULL_ALL", &
         default_i_val=ot_precond_none, &
         enum_c_vals=s2a("FULL_ALL", "FULL_SINGLE_INVERSE", "FULL_SINGLE", "FULL_KINETIC", "FULL_S_INVERSE", &
                         "NONE"), &
         enum_desc=s2a("Most effective state selective preconditioner based on diagonalization, "// &
                       "requires the ENERGY_GAP parameter to be an underestimate of the HOMO-LUMO gap. "// &
                       "This preconditioner is recommended for almost all systems, except very large systems where "// &
                       "make_preconditioner would dominate the total computational cost.", &
                       "Based on H-eS cholesky inversion, similar to FULL_SINGLE in preconditioning efficiency "// &
                       "but cheaper to construct, "// &
                       "might be somewhat less robust. Recommended for large systems.", &
                       "Based on H-eS diagonalisation, not as good as FULL_ALL, but somewhat cheaper to apply. ", &
                       "Cholesky inversion of S and T, fast construction, robust, and relatively good, "// &
                       "use for very large systems.", &
                       "Cholesky inversion of S, not as good as FULL_KINETIC, yet equally expensive.", &
                       "skip preconditioning"), &
         enum_i_vals=(/ot_precond_full_all, ot_precond_full_single_inverse, ot_precond_full_single, &
                       ot_precond_full_kinetic, ot_precond_s_inverse, ot_precond_none/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY_GAP", &
                          description="Energy gap estimate [a.u.] for preconditioning", &
                          usage="ENERGY_GAP 0.1", &
                          default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART", &
                          description="Restart the response calculation if the restart file exists", &
                          usage="RESTART", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WFN_RESTART_FILE_NAME", &
                          variants=(/"RESTART_FILE_NAME"/), &
                          description="Root of the file names where to read the response functions from"// &
                          "which to restart the calculation of the linear response", &
                          usage="WFN_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_localize_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_current_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_nmr_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_spin_spin_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_epr_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_polarizability_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, "PRINT", "printing of information during the linear response calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create( &
         print_key, "program_run_info", &
         description="Controls the printing of basic iteration information during the LINRES calculation", &
         print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of restart file of the response wavefunction."// &
                                       "For each set of response functions, i.e. for each perturbation,"// &
                                       "one different restart file is dumped. These restart files should be"// &
                                       "employed only to restart the same type of LINRES calculation, "// &
                                       "i.e. with the same perturbation.", &
                                       print_level=low_print_level, common_iter_levels=3, each_iter_names=s2a("ITER"), &
                                       add_last=add_last_numeric, each_iter_values=(/3/), filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_linres_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>      calculation of induced current  DFPT
!>      Available properties : none
!> \param section the section to create
!> \author  MI/VW
! **************************************************************************************************
   SUBROUTINE create_current_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_current_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, print_key, subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",276)
      CALL section_create(section, name="current", &
                          description="The induced current density is calculated by DFPT.", &
                          n_keywords=4, n_subsections=1, repeats=.FALSE., &
                          citations=(/Sebastiani2001, Weber2009/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the induced current calculation", &
                          usage="&CURRENT T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GAUGE", &
                          description="The gauge used to compute the induced current within GAPW.", &
                          usage="GAUGE R", &
                          default_i_val=current_gauge_r_and_step_func, &
                          enum_c_vals=s2a("R", "R_AND_STEP_FUNCTION", "ATOM"), &
                          enum_desc=s2a("Position gauge (doesnt work well).", &
                                        "Position and step function for the soft and the local parts, respectively.", &
                                        "Atoms."), &
                          enum_i_vals=(/current_gauge_r, current_gauge_r_and_step_func, current_gauge_atom/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GAUGE_ATOM_RADIUS", &
                          description="Build the gauge=atom using only the atoms within this radius.", &
                          usage="GAUGE_ATOM_RADIUS 10.0", &
                          type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=4.0_dp, unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_OLD_GAUGE_ATOM", &
                          description="Use the old way to compute the gauge.", &
                          usage="USE_OLD_GAUGE_ATOM T", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORBITAL_CENTER", &
                          description="The orbital center.", &
                          usage="ORBITAL_CENTER WANNIER", &
                          default_i_val=current_orb_center_wannier, &
                          enum_c_vals=s2a("WANNIER", "COMMON", "ATOM", "BOX"), &
                          enum_desc=s2a("Use the Wannier centers.", &
                                        "Use a common center (works only for an isolate molecule).", &
                                        "Use the atoms as center.", &
                                        "Boxing."), &
                          enum_i_vals=(/current_orb_center_wannier, current_orb_center_common, &
                                        current_orb_center_atom, current_orb_center_box/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COMMON_CENTER", &
                          description="The common center ", usage="COMMON_CENTER 0.0 1.0 0.0", &
                          n_var=3, default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp/), type_of_var=real_t, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NBOX", &
                          description="How many boxes along each directions ", usage="NBOX 6 6 5", &
                          n_var=3, default_i_vals=(/4, 4, 4/), type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHI_PBC", &
                          description="Calculate the succeptibility correction to the shift with PBC", &
                          usage="CHI_PBC T", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SELECTED_STATES_ON_ATOM_LIST", &
                          description="Indexes of the atoms for selecting"// &
                          " the states to be used for the response calculations.", &
                          usage="SELECTED_STATES_ON_ATOM_LIST 1 2 10", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SELECTED_STATES_ATOM_RADIUS", &
                          description="Select all the states included in the given radius arround each atoms "// &
                          "in SELECTED_STATES_ON_ATOM_LIST.", &
                          usage="SELECTED_STATES_ATOM_RADIUS 2.0", &
                          type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=4.0_dp, unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_CURRENT", &
                          description="Restart the induced current density calculation"// &
                          " from a previous run (not working yet).", &
                          usage="RESTART_CURRENT", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="print results of induced current density calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "CURRENT_CUBES", &
                                       description="Controls the printing of the induced current density (not working yet).", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components (not working yet).", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESPONSE_FUNCTION_CUBES", &
                                       description="Controls the printing of the response functions (not working yet).", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components (not working yet).", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LU_BOUNDS", &
                          variants=(/"CUBES_LU"/), &
                          description="The lower and upper index of the states to be printed as cube (not working yet).", &
                          usage="CUBES_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LIST", &
                          description="Indexes of the states to be printed as cube files"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes) (not working yet).", &
                          usage="CUBES_LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_current_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>         calculation of NMR chemical shift using
!>         the induced current obtained from DFPT
!>      Available properties : none
!> \param section the section to create
!> \author  MI/VW
! **************************************************************************************************
   SUBROUTINE create_nmr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_nmr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, print_key, subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",467)
      CALL section_create(section, name="nmr", &
                          description="The chemical shift is calculated by DFPT.", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE., &
                          citations=(/Weber2009/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the nmr calculation", &
                          usage="&NMR T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERPOLATE_SHIFT", &
                          description="Calculate the soft part of the chemical shift by interpolation ", &
                          usage="INTERPOLATE_SHIFT T", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NICS", &
                          description="Calculate the chemical shift in a set of points  "// &
                          " given from an external file", usage="NICS", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NICS_FILE_NAME", &
                          description="Name of the file with the NICS points coordinates", &
                          usage="NICS_FILE_NAME nics_file", &
                          default_lc_val="nics_file")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_NMR", &
                          description="Restart the NMR calculation from a previous run (NOT WORKING YET)", &
                          usage="RESTART_NMR", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT_GAPW_RADIUS", &
                          description="While computing the local part of the shift (GAPW), "// &
                          "the integration is restricted to nuclei that are within this radius.", &
                          usage="SHIFT_GAPW_RADIUS 20.0", &
                          type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=60.0_dp, unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="print results of nmr calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "RESPONSE_FUNCTION_CUBES", &
                                       description="Controls the printing of the response functions ", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LU_BOUNDS", &
                          variants=(/"CUBES_LU"/), &
                          description="The lower and upper index of the states to be printed as cube", &
                          usage="CUBES_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LIST", &
                          description="Indexes of the states to be printed as cube files"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="CUBES_LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CHI_TENSOR", &
                                       description="Controls the printing of susceptibility", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SHIELDING_TENSOR", &
                                       description="Controls the printing of the chemical shift", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="")

      CALL keyword_create(keyword, name="ATOMS_LU_BOUNDS", &
                          variants=(/"ATOMS_LU"/), &
                          description="The lower and upper atomic index for which the tensor is printed", &
                          usage="ATOMS_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_LIST", &
                          description="list of atoms for which the shift is printed into a file ", &
                          usage="LIST_ATOMS 1 2", n_var=-1, &
                          type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_nmr_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>      calculation of NMR spin-spin coupling (implementation not operating)
!>      Available properties : none
!> \param section the section to create
!> \author  VW
! **************************************************************************************************
   SUBROUTINE create_spin_spin_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_spin_spin_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, print_key, subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",616)
      CALL section_create(section, name="spinspin", &
                          description="Compute indirect spin-spin coupling constants.", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the nmr calculation", &
                          usage="&SPINSPIN T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_SPINSPIN", &
                          description="Restart the spin-spin calculation from a previous run (NOT WORKING YET)", &
                          usage="RESTART_SPINSPIN", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ISSC_ON_ATOM_LIST", &
                          description="Atoms for which the issc is computed.", &
                          usage="ISSC_ON_ATOM_LIST 1 2 10", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_FC", &
                          description="Compute the Fermi contact contribution", &
                          usage="DO_FC F", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_SD", &
                          description="Compute the spin-dipolar contribution", &
                          usage="DO_SD F", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_PSO", &
                          description="Compute the paramagnetic spin-orbit contribution", &
                          usage="DO_PSO F", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_DSO", &
                          description="Compute the diamagnetic spin-orbit contribution (NOT YET IMPLEMENTED)", &
                          usage="DO_DSO F", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="print results of the indirect spin-spin calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "K_MATRIX", &
                                       description="Controls the printing of the indirect spin-spin matrix", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="")

      CALL keyword_create(keyword, name="ATOMS_LIST", &
                          description="list of atoms for which the indirect spin-spin is printed into a file ", &
                          usage="LIST_ATOMS 1 2", n_var=-1, &
                          type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_spin_spin_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>         calculation of EPR using
!>         the induced current obtained from DFPT
!>      Available properties : none
!> \param section the section to create
!> \author  VW
! **************************************************************************************************
   SUBROUTINE create_epr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_epr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsubsection

      NULLIFY (keyword, print_key, subsection, subsubsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",719)
      CALL section_create(section, name="EPR", &
                          description="The g tensor is calculated by DFPT ", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE., &
                          citations=(/Weber2009/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the epr calculation", &
                          usage="&EPR T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_EPR", &
                          description="Restart the EPR calculation from a previous run (NOT WORKING)", &
                          usage="RESTART_EPR", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="print results of epr calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "NABLAVKS_CUBES", &
                                       description="Controls the printing of the components of nabla v_ks ", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "G_TENSOR", &
                                       description="Controls the printing of the g tensor", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL create_xc_section(subsubsection)
      CALL section_add_subsection(print_key, subsubsection)
      CALL section_release(subsubsection)

      CALL keyword_create(keyword, name="GAPW_MAX_ALPHA", &
                          description="Maximum alpha of GTH potentials allowed on the soft grids ", &
                          usage="GAPW_MAX_ALPHA real", default_r_val=5.0_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SOO_RHO_HARD", &
                          description="Whether or not to include the atomic parts of the density "// &
                          "in the SOO part of the g tensor", usage="SOO_RHO_HARD", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESPONSE_FUNCTION_CUBES", &
                                       description="Controls the printing of the response functions ", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LU_BOUNDS", &
                          variants=(/"CUBES_LU"/), &
                          description="The lower and upper index of the states to be printed as cube", &
                          usage="CUBES_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LIST", &
                          description="Indexes of the states to be printed as cube files"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="CUBES_LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_epr_section

! **************************************************************************************************
!> \brief creates the input structure used to activate
!>      calculation of polarizability tensor DFPT
!>      Available properties : none
!> \param section the section to create
!> \author SL
! **************************************************************************************************
   SUBROUTINE create_polarizability_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_polarizability_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, print_key, subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",852)
      CALL section_create(section, name="POLAR", &
                          description="Compute polarizabilities.", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE., &
                          citations=(/Putrino2002/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the polarizability calculation", &
                          usage="&POLAR T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_RAMAN", &
                          description="Compute the electric-dipole--electric-dipole polarizability", &
                          usage="DO_RAMAN F", &
                          citations=(/Luber2014/), &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="print results of the polarizability calculation", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "POLAR_MATRIX", &
                                       description="Controls the printing of the polarizabilities", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="")

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_polarizability_section

! **************************************************************************************************
!> \brief creates the section for electron transfer coupling
!> \param section ...
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE create_et_coupling_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_et_coupling_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",910)
      CALL section_create(section, "ET_COUPLING", &
                          description="specifies the two constraints/restraints for extracting ET coupling elements", &
                          n_keywords=1, n_subsections=4, repeats=.FALSE.)

      NULLIFY (subsection)
      CALL create_restraint_A(subsection, "DDAPC_RESTRAINT_A")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_restraint_A(subsection, "DDAPC_RESTRAINT_B")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_restraint_A(subsection, "BECKE_RESTRAINT_A")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_restraint_A(subsection, "BECKE_RESTRAINT_B")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="TYPE_OF_CONSTRAINT", &
                          description="Specifies the type of constraint", &
                          usage="TYPE_OF_CONSTRAINT DDAPC", &
                          enum_c_vals=s2a("NONE", "DDAPC", "BECKE"), &
                          enum_i_vals=(/do_no_et, do_et_ddapc, do_et_becke/), &
                          enum_desc=s2a("NONE", "ddapc_restraint", "Sperical potential"), &
                          default_i_val=do_no_et)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_et_coupling_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param section_name ...
! **************************************************************************************************
   SUBROUTINE create_restraint_A(section, section_name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: section_name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_restraint_A', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_properties_dft.F",970)
      CALL section_create(section, TRIM(ADJUSTL(section_name)), &
                          description="Use DDAPC charges in a restraint (check code for details),"// &
                          " section can be repeated, but only one constraint is possible at the moment.", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="force constant of the restraint", &
                          usage="STRENGTH {real} ", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE_OF_DENSITY", &
                          description="Specifies the type of density used for the fitting", &
                          usage="TYPE_OF_DENSITY (FULL|SPIN)", &
                          enum_c_vals=s2a("FULL", "SPIN"), &
                          enum_i_vals=(/do_full_density, do_spin_density/), &
                          enum_desc=s2a("Full density", "Spin density"), &
                          default_i_val=do_full_density)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="target value of the restraint", &
                          usage="TARGET {real} ", default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the list of atoms that is summed in the restraint", &
                          usage="ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="COEFF", &
         description="Defines the the coefficient of the atom in the atom list (default is one), currently DDAPC only ", &
         usage="COEFF 1.0 -1.0", &
         type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTIONAL_FORM", &
                          description="Specifies the functional form of the term added", &
                          usage="FUNCTIONAL_FORM RESTRAINT", &
                          enum_c_vals=s2a("RESTRAINT", "CONSTRAINT"), &
                          enum_i_vals=(/do_ddapc_restraint, do_ddapc_constraint/), &
                          enum_desc=s2a("Harmonic potential: s*(q-t)**2", "Constraint form: s*(q-t)"), &
                          default_i_val=do_ddapc_restraint)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_restraint_A

END MODULE input_cp2k_properties_dft
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for the FORCE_EVAL section of cp2k
!> \par History
!>      06.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_force_eval
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        do_eip, do_fist, do_mixed, do_qmmm, do_qs, do_stress_analytical, do_stress_diagonal_anal, &
        do_stress_diagonal_numer, do_stress_none, do_stress_numerical, numerical
   USE input_cp2k_dft,                  ONLY: create_bsse_section,&
                                              create_dft_section
   USE input_cp2k_eip,                  ONLY: create_eip_section
   USE input_cp2k_mixed,                ONLY: create_mix_section
   USE input_cp2k_mm,                   ONLY: create_mm_section
   USE input_cp2k_properties_dft,       ONLY: create_properties_section
   USE input_cp2k_qmmm,                 ONLY: create_qmmm_section
   USE input_cp2k_subsys,               ONLY: create_subsys_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_force_eval'

   PUBLIC :: create_force_eval_section

CONTAINS

! **************************************************************************************************
!> \brief creates the force_eval section
!> \param section the section to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_force_eval_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_force_eval_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_force_eval.F",70)
      CALL section_create(section, name="force_eval", &
                          description="parameters needed to calculate energy and forces and"// &
                          " describe the system you want to analyze.", &
                          n_keywords=1, n_subsections=9, repeats=.TRUE.)

      NULLIFY (subsection)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="METHOD", &
                          description="Which method should be used to compute forces", &
                          usage="METHOD <STRING>", &
                          enum_c_vals=s2a("QS", &
                                          "FIST", &
                                          "QMMM", &
                                          "EIP", &
                                          "QUICKSTEP", &
                                          "MIXED"), &
                          enum_desc=s2a("Alias for QUICKSTEP", &
                                        "Molecular Mechanics", &
                                        "Hybrid quantum classical", &
                                        "Empirical Interatomic Potential", &
                                        "Electronic structure methods (DFT, ...)", &
                                        "Use a combination of two of the above"), &
                          enum_i_vals=(/do_qs, do_fist, do_qmmm, do_eip, do_qs, do_mixed/), &
                          default_i_val=do_qs)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRESS_TENSOR", &
                          description="Controls the calculation of the stress tensor. The combinations defined below"// &
                          " are not implemented for all methods.", &
                          usage="stress_tensor (NONE|ANALYTICAL|NUMERICAL|DIAGONAL_ANA|DIAGONAL_NUM)", &
                          default_i_val=do_stress_none, &
                          enum_c_vals=s2a("NONE", "ANALYTICAL", "NUMERICAL", "DIAGONAL_ANALYTICAL", "DIAGONAL_NUMERICAL"), &
                          enum_i_vals=(/do_stress_none, do_stress_analytical, do_stress_numerical, &
                                        do_stress_diagonal_anal, do_stress_diagonal_numer/), &
                          enum_desc=s2a("Do not compute stress tensor", &
                                        "Compute the stress tensor analytically (if available).", &
                                        "Compute the stress tensor numerically.", &
                                        "Compute the diagonal part only of the stress tensor analytically (if available).", &
                                        "Compute the diagonal part only of the stress tensor numerically"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_ext_pot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_rescale_force_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mix_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_dft_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qmmm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eip_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_bsse_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_subsys_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_properties_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_f_env_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_force_eval_section

! **************************************************************************************************
!> \brief Creates the section for applying an external potential
!> \param section ...
!> \date 03.2008
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_ext_pot_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ext_pot_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_force_eval.F",174)
      CALL section_create(section, name="EXTERNAL_POTENTIAL", &
                          description="Section controlling the presence of an external potential dependent "// &
                          " on the atomic positions (X,Y,Z)", &
                          n_keywords=7, n_subsections=0, repeats=.TRUE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS_LIST", &
                          description="Specifies the atoms on which the external potential will act", &
                          usage="ATOMS_LIST {INT} {INT} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTION", &
                          description="Specifies the functional form in mathematical notation. Variables must be the atomic "// &
                          "coordinates (X,Y,Z).", usage="FUNCTION  X^2+Y^2+Z^2+LOG(ABS(X+Y))", &
                          type_of_var=lchar_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of  parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNITS", &
                          description="Optionally, allows to define valid CP2K unit strings for each parameter value. "// &
                          "It is assumed that the corresponding parameter value is specified in this unit.", &
                          usage="UNITS angstrom eV*angstrom^-1 angstrom^1 K", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Parameter used for computing the derivative with the Ridders' method.", &
                          usage="DX <REAL>", default_r_val=0.1_dp, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERROR_LIMIT", &
                          description="Checks that the error in computing the derivative is not larger than "// &
                          "the value set; in case error is larger a warning message is printed.", &
                          usage="ERROR_LIMIT <REAL>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ext_pot_section

! **************************************************************************************************
!> \brief Creates the section controlling the rescaling of forces
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_rescale_force_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_rescale_force_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_force_eval.F",245)
      CALL section_create(section, name="RESCALE_FORCES", &
                          description="Section controlling the rescaling of forces. Useful when"// &
                          " starting from quite bad geometries with unphysically large forces.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="MAX_FORCE", &
                          description="Specify the Maximum Values of the force. If the force"// &
                          " of one atom exceed this value it's rescaled to the MAX_FORCE"// &
                          " value.", &
                          default_r_val=cp_unit_to_cp2k(value=50.0_dp, &
                                                        unit_str="kcalmol*angstrom^-1"), &
                          unit_str="hartree*bohr^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_rescale_force_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_f_env_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_f_env_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword)
      NULLIFY (print_key)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_force_eval.F",281)

      CALL section_create(section, name="PRINT", &
                          description="Properties that you want to output and that are common to all methods", &
                          n_keywords=0, n_subsections=5, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of basic information generated by force_eval", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FORCES", &
                                       description="Controls the printing of the forces after each force evaluation", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword=keyword, &
                          name="NDIGITS", &
                          description="Specifies the number of digits used "// &
                          "for the printing of the forces", &
                          usage="NDIGITS 6", &
                          default_i_val=8, &
                          repeats=.FALSE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "GRID_INFORMATION", &
         description="Controls the printing of information regarding the PW and RS grid structures.", &
         print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "TOTAL_NUMBERS", &
                                       description="Controls the printing of the total number of atoms, kinds,...", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DISTRIBUTION", &
                                       description="Controls the printing of the distribution of molecules, atoms, ...", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DISTRIBUTION2D", &
                                       description="Controls the printing of the distribution of matrix blocks,...", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DISTRIBUTION1D", &
                                       description="Each node prints out its distribution info ...", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "STRESS_TENSOR", &
                                       description="Controls the printing of the stress tensor", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword=keyword, &
                          name="NDIGITS", &
                          description="Specifies the number of digits used "// &
                          "for the printing of the stress tensor", &
                          usage="NDIGITS 6", &
                          default_i_val=8, &
                          repeats=.FALSE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_f_env_print_section

END MODULE input_cp2k_force_eval
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module that contains the routines for error handling
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_error_handling
   USE base_hooks,                      ONLY: cp_abort_hook,&
                                              cp_warn_hook
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: default_output_unit,&
                                              m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_abort
   USE termination,                     ONLY: print_message
   USE timings,                         ONLY: print_stack

!$ USE OMP_LIB,                        ONLY: omp_get_thread_num

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_error_handling'

   !API public routines
   PUBLIC :: cp_error_handling_setup

   !API (via pointer assignment to hook, PR67982, not meant to be called directly)
   PUBLIC :: cp_abort_handler, cp_warn_handler

   INTEGER, PUBLIC, SAVE :: warning_counter = 0

CONTAINS

! **************************************************************************************************
!> \brief Registers handlers with base_hooks.F
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_error_handling_setup()
      cp_abort_hook => cp_abort_handler
      cp_warn_hook => cp_warn_handler
   END SUBROUTINE cp_error_handling_setup

! **************************************************************************************************
!> \brief Abort program with error message
!> \param location ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_abort_handler(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      INTEGER                                            :: unit_nr

      CALL delay_non_master() ! cleaner output if all ranks abort simultaneously

      unit_nr = cp_logger_get_default_io_unit()
      IF (unit_nr <= 0) &
         unit_nr = default_output_unit ! fall back to stdout

      CALL print_abort_message(message, location, unit_nr)
      CALL print_stack(unit_nr)
      CALL m_flush(unit_nr)

      CALL mp_abort()
   END SUBROUTINE cp_abort_handler

! **************************************************************************************************
!> \brief Signal a warning
!> \param location ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_warn_handler(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      INTEGER                                            :: unit_nr

!$OMP MASTER
      warning_counter = warning_counter+1
!$OMP END MASTER

      unit_nr = cp_logger_get_default_io_unit()
      IF (unit_nr > 0) THEN
         CALL print_message("WARNING in "//TRIM(location)//' :: '//TRIM(ADJUSTL(message)), unit_nr, 1, 1, 1)
         CALL m_flush(unit_nr)
      ENDIF
   END SUBROUTINE cp_warn_handler

! **************************************************************************************************
!> \brief Delay non-master ranks/threads, used by cp_abort_handler()
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE delay_non_master()
      INTEGER                                            :: unit_nr
      REAL(KIND=dp)                                      :: t1, wait_time

      wait_time = 0.0_dp

      ! we (ab)use the logger to determine the first MPI rank
      unit_nr = cp_logger_get_default_io_unit()
      IF (unit_nr <= 0) &
         wait_time = wait_time+1.0_dp ! rank-0 gets a head start of one second.

!$    IF (omp_get_thread_num() /= 0) &
!$       wait_time = wait_time+1.0_dp ! master threads gets another second

      ! sleep
      IF (wait_time > 0.0_dp) THEN
         t1 = m_walltime()
         DO
            IF (m_walltime()-t1 > wait_time .OR. t1 < 0) EXIT
         ENDDO
      ENDIF

   END SUBROUTINE delay_non_master

! **************************************************************************************************
!> \brief Prints a nicely formatted abort message box
!> \param message ...
!> \param location ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE print_abort_message(message, location, output_unit)
      CHARACTER(LEN=*), INTENT(IN)                       :: message, location
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: img = "   ___   "//"  /   \  "//" [ABORT] "//"  \___/  "// &
         "    |    "//"  O/|    "//" /| |    "//" / \     "
      INTEGER, PARAMETER :: img_height = 8, img_width = 9, screen_width = 80, &
         txt_width = screen_width-img_width-5

      CHARACTER(LEN=screen_width)                        :: msg_line
      INTEGER                                            :: a, b, c, fill, i, img_start, indent, &
                                                            msg_height, msg_start

! count message lines

      a = 1; b = -1; msg_height = 0
      DO WHILE (b < LEN_TRIM(message))
         b = next_linebreak(message, a, txt_width)
         a = b+1
         msg_height = msg_height+1
      ENDDO

      ! calculate message and image starting lines
      IF (img_height > msg_height) THEN
         msg_start = (img_height-msg_height)/2+1
         img_start = 1
      ELSE
         msg_start = 1
         img_start = msg_height-img_height+2
      ENDIF

      ! print empty line
      WRITE (UNIT=output_unit, FMT="(A)") ""

      ! print opening line
      WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", screen_width-1)

      ! print body
      a = 1; b = -1; c = 1
      DO i = 1, MAX(img_height-1, msg_height)
         WRITE (UNIT=output_unit, FMT="(A)", advance='no') " *"
         IF (i < img_start) THEN
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') REPEAT(" ", img_width)
         ELSE
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') img(c:c+img_width-1)
            c = c+img_width
         ENDIF
         IF (i < msg_start) THEN
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') REPEAT(" ", txt_width+2)
         ELSE
            b = next_linebreak(message, a, txt_width)
            msg_line = message(a:b)
            a = b+1
            fill = (txt_width-LEN_TRIM(msg_line))/2+1
            indent = txt_width-LEN_TRIM(msg_line)-fill+2
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') REPEAT(" ", indent)
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') TRIM(msg_line)
            WRITE (UNIT=output_unit, FMT="(A)", advance='no') REPEAT(" ", fill)
         ENDIF
         WRITE (UNIT=output_unit, FMT="(A)", advance='yes') "*"
      ENDDO

      ! print location line
      WRITE (UNIT=output_unit, FMT="(A)", advance='no') " *"
      WRITE (UNIT=output_unit, FMT="(A)", advance='no') img(c:c+img_width-1)
      indent = txt_width-LEN_TRIM(location)+1
      WRITE (UNIT=output_unit, FMT="(A)", advance='no') REPEAT(" ", indent)
      WRITE (UNIT=output_unit, FMT="(A)", advance='no') TRIM(location)
      WRITE (UNIT=output_unit, FMT="(A)", advance='yes') " *"

      ! print closing line
      WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", screen_width-1)

      ! print empty line
      WRITE (UNIT=output_unit, FMT="(A)") ""

   END SUBROUTINE print_abort_message

! **************************************************************************************************
!> \brief Helper routine for print_abort_message()
!> \param message ...
!> \param pos ...
!> \param rowlen ...
!> \retval ibreak ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION next_linebreak(message, pos, rowlen) RESULT(ibreak)
      CHARACTER(LEN=*), INTENT(IN)                       :: message
      INTEGER, INTENT(IN)                                :: pos, rowlen
      INTEGER                                            :: ibreak

      INTEGER                                            :: i, n

      n = LEN_TRIM(message)
      IF (n-pos <= rowlen) THEN
         ibreak = n ! remaining message shorter than line
      ELSE
         i = INDEX(message(pos+1:pos+1+rowlen), " ", BACK=.TRUE.)
         IF (i == 0) THEN
            ibreak = pos+rowlen-1 ! no space found, break mid-word
         ELSE
            ibreak = pos+i ! break at space closest to rowlen
         ENDIF
      ENDIF
   END FUNCTION next_linebreak

END MODULE cp_error_handling
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Timing routines for accounting
!> \par History
!>      02.2004 made a stacked version (of stacks...) [Joost VandeVondele]
!>      11.2004 storable timer_envs (for f77 interface) [fawzi]
!>      10.2005 binary search to speed up lookup in timeset [fawzi]
!>      12.2012 Complete rewrite based on dictionaries. [ole]
!>      01.2014 Collect statistics from all MPI ranks. [ole]
!> \author JGH
! **************************************************************************************************
MODULE timings_report
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dict,                            ONLY: dict_get,&
                                              dict_haskey,&
                                              dict_items
   USE dict_i4tuple_callstat,           ONLY: dict_i4tuple_callstat_item_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE list,                            ONLY: list_destroy,&
                                              list_get,&
                                              list_init,&
                                              list_isready,&
                                              list_pop,&
                                              list_push,&
                                              list_size
   USE list_routinereport,              ONLY: list_routinereport_type
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_max,&
                                              mp_sum
   USE timings,                         ONLY: get_timer_env
   USE timings_base_type,               ONLY: call_stat_type,&
                                              routine_report_type,&
                                              routine_stat_type
   USE timings_types,                   ONLY: timer_env_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   INTEGER, PUBLIC, PARAMETER :: cost_type_time = 17, cost_type_energy = 18

   PUBLIC :: timings_report_print, timings_report_callgraph

CONTAINS

! **************************************************************************************************
!> \brief Print accumulated information on timers
!> \param iw ...
!> \param r_timings ...
!> \param sort_by_self_time ...
!> \param cost_type ...
!> \param para_env is needed to collect statistics from other nodes.
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE timings_report_print(iw, r_timings, sort_by_self_time, cost_type, para_env)
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: r_timings
      LOGICAL, INTENT(IN)                                :: sort_by_self_time
      INTEGER, INTENT(IN)                                :: cost_type
      TYPE(cp_para_env_type), INTENT(IN)                 :: para_env

      TYPE(list_routinereport_type)                      :: reports
      TYPE(routine_report_type), POINTER                 :: r_report

      CALL list_init(reports)
      CALL collect_reports_from_ranks(reports, cost_type, para_env)

      IF (list_size(reports) > 0 .AND. iw > 0) &
         CALL print_reports(reports, iw, r_timings, sort_by_self_time, cost_type, para_env)

      ! deallocate reports
      DO WHILE (list_size(reports) > 0)
         r_report => list_pop(reports)
         DEALLOCATE (r_report)
      END DO
      CALL list_destroy(reports)

   END SUBROUTINE timings_report_print

! **************************************************************************************************
!> \brief Collects the timing or energy reports from all MPI ranks.
!> \param reports ...
!> \param cost_type ...
!> \param para_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE collect_reports_from_ranks(reports, cost_type, para_env)
      TYPE(list_routinereport_type), INTENT(INOUT)       :: reports
      INTEGER, INTENT(IN)                                :: cost_type
      TYPE(cp_para_env_type), INTENT(IN)                 :: para_env

      CHARACTER(LEN=default_string_length)               :: routineN
      INTEGER                                            :: local_routine_id, sending_rank
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: collected
      REAL(KIND=dp)                                      :: foobar
      TYPE(routine_report_type), POINTER                 :: r_report
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      NULLIFY (r_stat, r_report, timer_env)
      IF (.NOT. list_isready(reports)) &
         CALL cp__b("common/timings_report.F",113,"BUG")

      timer_env => get_timer_env()

      ! make sure all functions have been called so that list_size(timer_env%routine_stats)
      ! and the actual dictionary are consistent in the loop below, preventing out of bounds.
      ! this hack makes sure they are called before
      routineN = ""
      CALL mp_bcast(routineN, 0, para_env%group)
      sending_rank = 0
      CALL mp_max(sending_rank, para_env%group)
      CALL mp_sum(sending_rank, para_env%group)
      foobar = 0.0_dp
      CALL mp_max(foobar, para_env%group)
      CALL mp_sum(foobar, para_env%group)
      ! end hack

      ! Array collected is used as a bit field.
      ! It's of type integer in order to use the convenient MINLOC routine.
      ALLOCATE (collected(list_size(timer_env%routine_stats)))
      collected(:) = 0

      DO
         ! does any rank have uncollected stats?
         sending_rank = -1
         IF (.NOT. ALL(collected == 1)) sending_rank = para_env%mepos
         CALL mp_max(sending_rank, para_env%group)
         IF (sending_rank < 0) EXIT ! every rank got all routines collected
         IF (sending_rank == para_env%mepos) THEN
            local_routine_id = MINLOC(collected, dim=1)
            r_stat => list_get(timer_env%routine_stats, local_routine_id)
            routineN = r_stat%routineN
         ENDIF
         CALL mp_bcast(routineN, sending_rank, para_env%group)

         ! Create new report for routineN
         ALLOCATE (r_report)
         CALL list_push(reports, r_report)
         r_report%routineN = routineN

         ! If routineN was called on local node, add local stats
         IF (dict_haskey(timer_env%routine_names, routineN)) THEN
            local_routine_id = dict_get(timer_env%routine_names, routineN)
            collected(local_routine_id) = 1
            r_stat => list_get(timer_env%routine_stats, local_routine_id)
            r_report%max_total_calls = r_stat%total_calls
            r_report%sum_total_calls = r_stat%total_calls
            r_report%sum_stackdepth = r_stat%stackdepth_accu
            SELECT CASE (cost_type)
            CASE (cost_type_energy)
               r_report%max_icost = r_stat%incl_energy_accu
               r_report%sum_icost = r_stat%incl_energy_accu
               r_report%max_ecost = r_stat%excl_energy_accu
               r_report%sum_ecost = r_stat%excl_energy_accu
            CASE (cost_type_time)
               r_report%max_icost = r_stat%incl_walltime_accu
               r_report%sum_icost = r_stat%incl_walltime_accu
               r_report%max_ecost = r_stat%excl_walltime_accu
               r_report%sum_ecost = r_stat%excl_walltime_accu
            CASE DEFAULT
               CALL cp__b("common/timings_report.F",173,"BUG")
            END SELECT
         END IF

         ! collect stats of routineN via MPI
         CALL mp_max(r_report%max_total_calls, para_env%group)
         CALL mp_sum(r_report%sum_total_calls, para_env%group)
         CALL mp_sum(r_report%sum_stackdepth, para_env%group)
         CALL mp_max(r_report%max_icost, para_env%group)
         CALL mp_sum(r_report%sum_icost, para_env%group)
         CALL mp_max(r_report%max_ecost, para_env%group)
         CALL mp_sum(r_report%sum_ecost, para_env%group)
      ENDDO

   END SUBROUTINE collect_reports_from_ranks

! **************************************************************************************************
!> \brief Print the collected reports
!> \param reports ...
!> \param iw ...
!> \param threshold ...
!> \param sort_by_exclusiv_cost ...
!> \param cost_type ...
!> \param para_env ...
!> \par History
!>      01.2014 Refactored (Ole Schuett)
!> \author JGH
! **************************************************************************************************
   SUBROUTINE print_reports(reports, iw, threshold, sort_by_exclusiv_cost, cost_type, para_env)
      TYPE(list_routinereport_type), INTENT(IN)          :: reports
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      LOGICAL, INTENT(IN)                                :: sort_by_exclusiv_cost
      INTEGER, INTENT(IN)                                :: cost_type
      TYPE(cp_para_env_type), INTENT(IN)                 :: para_env

      CHARACTER(LEN=4)                                   :: label
      CHARACTER(LEN=default_string_length)               :: fmt, title
      INTEGER                                            :: decimals, i, j, num_routines
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indices
      REAL(KIND=dp)                                      :: asd, maxcost, mincost
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: max_costs
      TYPE(routine_report_type), POINTER                 :: r_report_i, r_report_j

      NULLIFY (r_report_i, r_report_j)
      IF (.NOT. list_isready(reports)) &
         CALL cp__b("common/timings_report.F",219,"BUG")

      ! are we printing timing or energy ?
      SELECT CASE (cost_type)
      CASE (cost_type_energy)
         title = "E N E R G Y"
         label = "ENER"
      CASE (cost_type_time)
         title = "T I M I N G"
         label = "TIME"
      CASE DEFAULT
         CALL cp__b("common/timings_report.F",230,"BUG")
      END SELECT

      ! write banner
      WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
      WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
      WRITE (UNIT=iw, FMT="(T2,A,T35,A,T80,A)") "-", TRIM(title), "-"
      WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
      WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
      WRITE (UNIT=iw, FMT="(T2,A,T35,A,T41,A,T45,2A18)") &
         "SUBROUTINE", "CALLS", " ASD", "SELF "//label, "TOTAL "//label
      WRITE (UNIT=iw, FMT="(T33,A)") &
         "MAXIMUM       AVERAGE  MAXIMUM  AVERAGE  MAXIMUM"

      ! sort statistics
      num_routines = list_size(reports)
      ALLOCATE (max_costs(num_routines))
      DO i = 1, num_routines
         r_report_i => list_get(reports, i)
         IF (sort_by_exclusiv_cost) THEN
            max_costs(i) = r_report_i%max_ecost
         ELSE
            max_costs(i) = r_report_i%max_icost
         END IF
      ENDDO
      ALLOCATE (indices(num_routines))
      CALL sort(max_costs, num_routines, indices)

      maxcost = MAXVAL(max_costs)
      mincost = maxcost*threshold

      ! adjust fmt dynamically based on the max walltime.
      ! few clocks have more than 3 digits resolution, so stop there
      decimals = 3
      IF (maxcost >= 10000) decimals = 2
      IF (maxcost >= 100000) decimals = 1
      IF (maxcost >= 1000000) decimals = 0
      WRITE (UNIT=fmt, FMT="(A,I0,A)") &
         "(T2,A30,1X,I7,1X,F4.1,4(1X,F8.", decimals, "))"

      !write output
      DO i = num_routines, 1, -1
         IF (max_costs(i) >= mincost) THEN
            j = indices(i)
            r_report_j => list_get(reports, j)
            ! average stack depth
            asd = REAL(r_report_j%sum_stackdepth, KIND=dp)/ &
                  REAL(MAX(1_int_8, r_report_j%sum_total_calls), KIND=dp)
            WRITE (UNIT=iw, FMT=fmt) &
               ADJUSTL(r_report_j%routineN(1:31)), &
               r_report_j%max_total_calls, &
               asd, &
               r_report_j%sum_ecost/para_env%num_pe, &
               r_report_j%max_ecost, &
               r_report_j%sum_icost/para_env%num_pe, &
               r_report_j%max_icost
         END IF
      END DO
      WRITE (UNIT=iw, FMT="(T2,A,/)") REPEAT("-", 79)

   END SUBROUTINE print_reports

! **************************************************************************************************
!> \brief Write accumulated callgraph information as cachegrind-file.
!> http://kcachegrind.sourceforge.net/cgi-bin/show.cgi/KcacheGrindCalltreeFormat
!> \param filename ...
!> \par History
!>     12.2012  intial version[ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE timings_report_callgraph(filename)
      CHARACTER(len=*), INTENT(in)                       :: filename

      INTEGER, PARAMETER                                 :: E = 1000, T = 100000

      INTEGER                                            :: i, unit
      TYPE(call_stat_type), POINTER                      :: c_stat
      TYPE(dict_i4tuple_callstat_item_type), &
         DIMENSION(:), POINTER                           :: ct_items
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      CALL open_file(file_name=filename, file_status="REPLACE", file_action="WRITE", &
                     file_form="FORMATTED", unit_number=unit)
      timer_env => get_timer_env()

      ! use outermost routine as total runtime
      r_stat => list_get(timer_env%routine_stats, 1)
      WRITE (UNIT=unit, FMT="(A)") "events: Walltime Energy"
      WRITE (UNIT=unit, FMT="(A,I0,1X,I0)") "summary: ", &
         INT(T*r_stat%incl_walltime_accu, KIND=int_8), &
         INT(E*r_stat%incl_energy_accu, KIND=int_8)

      DO i = 1, list_size(timer_env%routine_stats)
         r_stat => list_get(timer_env%routine_stats, i)
         WRITE (UNIT=unit, FMT="(A,I0,A,A)") "fn=(", r_stat%routine_id, ") ", r_stat%routineN
         WRITE (UNIT=unit, FMT="(A,I0,1X,I0)") "1 ", &
            INT(T*r_stat%excl_walltime_accu, KIND=int_8), &
            INT(E*r_stat%excl_energy_accu, KIND=int_8)
      END DO

      ct_items => dict_items(timer_env%callgraph)
      DO i = 1, SIZE(ct_items)
         c_stat => ct_items(i)%value
         WRITE (UNIT=unit, FMT="(A,I0,A)") "fn=(", ct_items(i)%key(1), ")"
         WRITE (UNIT=unit, FMT="(A,I0,A)") "cfn=(", ct_items(i)%key(2), ")"
         WRITE (UNIT=unit, FMT="(A,I0,A)") "calls=", c_stat%total_calls, " 1"
         WRITE (UNIT=unit, FMT="(A,I0,1X,I0)") "1 ", &
            INT(T*c_stat%incl_walltime_accu, KIND=int_8), &
            INT(E*c_stat%incl_energy_accu, KIND=int_8)
      END DO
      DEALLOCATE (ct_items)

      CALL close_file(unit_number=unit, file_status="KEEP")

   END SUBROUTINE timings_report_callgraph
END MODULE timings_report

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the global input section for cp2k
!> \par History
!>      06.2004 created [fawzi]
!>      03.2014 moved to seperate file [Ole Schuett]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_global
   USE bibliography,                    ONLY: Borstnik2014,&
                                              Ceriotti2014,&
                                              Frigo2005,&
                                              Schonherr2014
   USE cp_blacs_env,                    ONLY: BLACS_GRID_COL,&
                                              BLACS_GRID_ROW,&
                                              BLACS_GRID_SQUARE
   USE cp_dbcsr_interface,              ONLY: &
        accdrv_avoid_after_busy, accdrv_binning_binsize, accdrv_binning_nbins, &
        accdrv_do_inhomogenous, accdrv_min_flop_process, accdrv_min_flop_sort, &
        accdrv_posterior_buffers, accdrv_posterior_streams, accdrv_priority_buffers, &
        accdrv_priority_streams, dbcsr_get_conf_avg_elements_images, dbcsr_get_conf_max_ele_block, &
        dbcsr_get_conf_randmat_seed, dbcsr_get_conf_use_comm_thread, dbcsr_get_conf_use_mpi_exp, &
        dbcsr_get_conf_use_mpi_filtering, mm_driver_auto, mm_driver_blas, mm_driver_matmul, &
        mm_driver_smm, mm_driver_xsmm, mm_name_auto, mm_name_blas, mm_name_matmul, mm_name_smm, &
        mm_name_xsmm, multrec_limit
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE input_constants,                 ONLY: &
        bsse_run, callgraph_all, callgraph_master, callgraph_none, cell_opt_run, debug_run, &
        do_atom, do_band, do_cp2k, do_dbcsr, do_diag_elpa, do_diag_sl, do_diag_sl2, do_farming, &
        do_fft_fftw3, do_fft_sg, do_opt_basis, do_optimize_input, do_pdgemm, do_swarm, do_tamc, &
        do_test, do_tree_mc, do_tree_mc_ana, driver_run, ehrenfest, electronic_spectra_run, &
        elpa_avx_b2, elpa_avx_b4, elpa_avx_b6, elpa_bgp, elpa_bgq, elpa_gen, elpa_gen_sim, &
        elpa_sse, energy_force_run, energy_run, fftw_plan_estimate, fftw_plan_exhaustive, &
        fftw_plan_measure, fftw_plan_patient, gaussian, geo_opt_run, linear_response_run, &
        mol_dyn_run, mon_car_run, none_run, pint_run, real_time_propagation, tree_mc_run, vib_anal
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_global'

   PUBLIC :: create_global_section

CONTAINS

! **************************************************************************************************
!> \brief section to hold global settings for the whole program
!> \param section the section to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_global_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_global_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, sub_section

      NULLIFY (print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_global.F",84)
      CALL section_create(section, name="GLOBAL", &
                          description="Section with general information regarding which kind "// &
                          "of simulation to perform an parameters for the whole PROGRAM", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="BLACS_GRID", &
                          description="how to distribute the processors on the 2d grid needed "// &
                          "by BLACS (and thus SCALAPACK)", usage="BLACS_GRID SQUARE", &
                          default_i_val=BLACS_GRID_SQUARE, enum_c_vals=s2a("SQUARE", "ROW", "COLUMN"), &
                          enum_desc=s2a("Distribution by matrix blocks", "Distribution by matrix rows", &
                                        "Distribution by matrix columns"), &
                          enum_i_vals=(/BLACS_GRID_SQUARE, BLACS_GRID_ROW, BLACS_GRID_COL/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BLACS_REPEATABLE", &
                          description="Use a topology for BLACS collectives that is guaranteed to be repeatable "// &
                          "on homegeneous architectures", &
                          usage="BLACS_REPEATABLE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PREFERRED_DIAG_LIBRARY", &
         description="Specifies the DIAGONALIZATION library to be used. If not availabe, the standard scalapack is used", &
         usage="PREFERRED_DIAG_LIBRARY ELPA", &
         default_i_val=do_diag_sl, &
         enum_i_vals=(/do_diag_sl, do_diag_sl2, do_diag_elpa/), &
         enum_c_vals=s2a("SL", "SL2", "ELPA"), &
         enum_desc=s2a("Standard scalapack: syevd", "Scalapack 2.0: syevr", "ELPA"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="ELPA_KERNEL", &
         description="Specifies the kernel to be used when ELPA is in use", &
         usage="ELPA_KERNEL SSE", &
         default_i_val=elpa_gen, &
         enum_i_vals=(/elpa_gen, elpa_gen_sim, elpa_bgp, elpa_bgq, elpa_sse, elpa_avx_b2, elpa_avx_b4, elpa_avx_b6/), &
         enum_c_vals=s2a("GENERIC", "GENERIC_SIMPLE", "BGP", "BGQ", "SSE", "AVX_BLOCK2", "AVX_BLOCK4", "AVX_BLOCK6"), &
         enum_desc=s2a( &
         "Generic kernel", "Simplified generic kernel", "Kernel optimized for IBM BGP", "Kernel optimized for IBM BGQ", &
         "Kernel wiwth assembler for SSE vectorization", "Kernel optimized for x86_64 using SSE2/SSE3 (Intel)", &
         "Kernel optimized for x86_64 using SSE2/SSE4 (AMD Bulldozer) ", &
         "Kernel optimized for x86_64 using SSE2/SSE4 (block6)"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PREFERRED_FFT_LIBRARY", &
         description="Specifies the FFT library which should be preferred. "// &
         "If it is not available, use FFTW3 if this is linked in, if FFTW3 is not available use FFTSG. "// &
         "Improved performance with FFTW3 can be obtained specifying a proper value for FFTW_PLAN_TYPE. "// &
         "Contrary to earlier CP2K versions, all libraries will result in the same grids, "// &
         "i.e. the subset of grids which all FFT libraries can transform. "// &
         "See EXTENDED_FFT_LENGTHS if larger FFTs or grids that more precisely match a given cutoff are needed, "// &
         "or older results need to be reproduced. "// &
         "FFTW3 is often (close to) optimal, and well tested with CP2K.", &
         usage="PREFERRED_FFT_LIBRARY FFTW3", &
         citations=(/Frigo2005/), &
         default_i_val=do_fft_fftw3, &
         enum_i_vals=(/do_fft_sg, do_fft_fftw3, do_fft_fftw3/), &
         enum_c_vals=s2a("FFTSG", "FFTW3", "FFTW"), &
         enum_desc=s2a("Stefan Goedecker''s FFT (FFTSG), always available,"// &
                       "will be used in case a FFT lib is specified and not available", &
                       "a fast portable FFT library. Recommended."// &
                       "See also the FFTW_PLAN_TYPE, and FFTW_WISDOM_FILE_NAME keywords.", &
                       "Same as FFTW3 (for compatability with CP2K 2.3)"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FFTW_WISDOM_FILE_NAME", &
                          description="The name of the file that contains wisdom (pre-planned FFTs) for use with FFTW3. "// &
                          "Using wisdom can significantly speed up the FFTs (see the FFTW homepage for details). "// &
                          "Note that wisdom is not transferable between different computer (architectures). "// &
                          "Wisdom can be generated using the fftw-wisdom tool that is part of the fftw installation. "// &
                          "cp2k/tools/cp2k-wisdom is a script that contains some additional info, and can help "// &
                          "to generate a useful default for /etc/fftw/wisdom or particular values for a given simulation.", &
                          usage="FFTW_WISDOM_FILE_NAME wisdom.dat", default_lc_val="/etc/fftw/wisdom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FFTW_PLAN_TYPE", &
                          description="FFTW can have improved performance if it is allowed to plan with "// &
                          "explicit measurements which strategy is best for a given FFT. "// &
                          "While a plan based on measurements is generally faster, "// &
                          "differences in machine load will lead to different plans for the same input file, "// &
                          "and thus numerics for the FFTs will be slightly different from run to run."// &
                          "PATIENT planning is recommended for long ab initio MD runs.", &
                          usage="FFTW_PLAN_TYPE PATIENT", &
                          citations=(/Frigo2005/), &
                          default_i_val=fftw_plan_estimate, &
                          enum_i_vals=(/fftw_plan_estimate, fftw_plan_measure, fftw_plan_patient, fftw_plan_exhaustive/), &
                          enum_c_vals=s2a("ESTIMATE", &
                                          "MEASURE", &
                                          "PATIENT", &
                                          "EXHAUSTIVE"), &
                          enum_desc=s2a("Quick estimate, no runtime measurements.", &
                                        "Quick measurement, somewhat faster FFTs.", &
                                        "Measurements trying a wider range of possibilities.", &
                                        "Measurements trying all possibilities - use with caution."))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXTENDED_FFT_LENGTHS", &
                          description="Use fft library specific values for the allows number of points in FFTs. "// &
                          "The default is to use the internal FFT lengths. For external fft libraries this may "// &
                          "create an error at the external library level, because the length provided by cp2k is "// &
                          "not supported by the external library. In this case switch on this keyword "// &
                          "to obtain, with certain fft libraries, lengths matching the external fft library lengths, or "// &
                          "larger allowed grids, or grids that more precisely match a given cutoff. "// &
                          "IMPORTANT NOTE: in this case, the actual grids used in CP2K depends on the FFT library. "// &
                          "A change of FFT library must therefore be considered equivalent to a change of basis, "// &
                          "which implies a change of total energy. ", &
                          usage="EXTENDED_FFT_LENGTHS", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FFT_POOL_SCRATCH_LIMIT", &
                          description="Limits the memory usage of the FFT scratch pool, potentially reducing efficiency a bit", &
                          usage="FFT_POOL_SCRATCH_LIMIT <INTEGER>", default_i_val=15)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALLTOALL_SGL", &
                          description="All-to-all communication (FFT) should use single precision", &
                          usage="ALLTOALL_SGL YES", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRINT_LEVEL", &
                          variants=(/"IOLEVEL"/), &
                          description="How much output is written out.", &
                          usage="PRINT_LEVEL HIGH", &
                          default_i_val=medium_print_level, enum_c_vals= &
                          s2a("SILENT", "LOW", "MEDIUM", "HIGH", "DEBUG"), &
                          enum_desc=s2a("Almost no output", &
                                        "Little output", "Quite some output", "Lots of output", &
                                        "Everything is written out, useful for debugging purposes only"), &
                          enum_i_vals=(/silent_print_level, low_print_level, medium_print_level, &
                                        high_print_level, debug_print_level/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PROGRAM_NAME", &
         variants=(/"PROGRAM"/), &
         description="Which program should be run", &
         usage="PROGRAM_NAME <STRING>", &
         enum_c_vals=s2a("ATOM", "FARMING", "TEST", "CP2K", "OPTIMIZE_INPUT", "OPTIMIZE_BASIS", "TMC", "MC_ANALYSIS", "SWARM"), &
         enum_desc=s2a("Runs single atom calculations", &
                       "Runs N independent jobs in a single run", &
                       "Do some benchmarking and testing", &
                       "Runs one of the CP2K package", &
                       "A tool to optimize parameters in a CP2K input", &
                       "A tool to create a MOLOPT or ADMM basis for a given set"// &
                       " of training structures", &
                       "Runs Tree Monte Carlo algorithm using additional input file(s)", &
                       "Runs (Tree) Monte Carlo trajectory file analysis", &
                       "Runs swarm based calculation"), &
         enum_i_vals=(/do_atom, do_farming, do_test, do_cp2k, do_optimize_input, &
                       do_opt_basis, do_tree_mc, do_tree_mc_ana, do_swarm/), &
         default_i_val=do_cp2k)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROJECT_NAME", &
                          variants=(/"PROJECT"/), &
                          description="Name of the project (used to build the name of the "// &
                          "trajectory, and other files generated by the program)", &
                          usage="PROJECT_NAME <STRING>", &
                          default_c_val="PROJECT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OUTPUT_FILE_NAME", &
                          description="Name of the output file. "// &
                          "Relevant only if automatically started (through farming for example). "// &
                          "If empty uses the project name as basis for it.", &
                          usage="OUTPUT_FILE_NAME {filename}", default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="RUN_TYPE", &
         description="Type of run that you want to perform Geometry "// &
         "optimization, md, montecarlo,...", &
         usage="RUN_TYPE MD", &
         default_i_val=energy_force_run, &
         citations=(/Ceriotti2014, Schonherr2014/), &
         enum_c_vals=s2a("NONE", "ENERGY", "ENERGY_FORCE", "MD", "GEO_OPT", &
                         "MC", "SPECTRA", "DEBUG", "BSSE", "LR", "PINT", "VIBRATIONAL_ANALYSIS", &
                         "BAND", "CELL_OPT", "WFN_OPT", "WAVEFUNCTION_OPTIMIZATION", &
                         "MOLECULAR_DYNAMICS", "GEOMETRY_OPTIMIZATION", "MONTECARLO", &
                         "ELECTRONIC_SPECTRA", "LINEAR_RESPONSE", "NORMAL_MODES", "RT_PROPAGATION", &
                         "EHRENFEST_DYN", "TAMC", "TMC", "DRIVER"), &
         enum_i_vals=(/none_run, energy_run, energy_force_run, mol_dyn_run, &
                       geo_opt_run, mon_car_run, electronic_spectra_run, debug_run, &
                       bsse_run, linear_response_run, pint_run, vib_anal, do_band, &
                       cell_opt_run, energy_run, energy_run, mol_dyn_run, geo_opt_run, &
                       mon_car_run, electronic_spectra_run, linear_response_run, &
                       vib_anal, real_time_propagation, ehrenfest, do_tamc, tree_mc_run, driver_run/), &
         enum_desc=s2a("Perform no tasks", "Computes energy", "Computes energy and forces", &
                       "Molecular Dynamics", "Geometry Optimization", "Monte Carlo", "Computes absorption Spectra", &
                       "Performs a Debug analysis", "Basis set superposition error", "Linear Response", &
                       "Path integral", "Vibrational analysis", "Band methods", "Cell optimization", &
                       "Alias for ENERGY", "Alias for ENERGY", "Alias for MD", "Alias for GEO_OPT", &
                       "Alias for MC", "Alias for SPECTRA", "Alias for LR", "Alias for VIBRATIONAL_ANALYSIS", &
                       "Real Time propagation run (fixed ionic positions)", &
                       "Ehrenfest dynamics (using real time propagation of the wavefunction)", &
                       "Temperature Accelerated Monte Carlo (TAMC)", &
                       "Tree Monte Carlo (TMC), a pre-sampling MC algorithm", &
                       "i-PI driver mode"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WALLTIME", &
                          variants=(/"WALLTI"/), &
                          description="Maximum execution time for this run. Time in seconds or in HH:MM:SS.", &
                          usage="WALLTIME {real} or {HH:MM:SS}", default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ECHO_INPUT", &
                          description="If the input should be echoed to the output with all the "// &
                          "defaults made explicit", &
                          usage="ECHO_INPUT NO", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ECHO_ALL_HOSTS", &
                          description="Echo a list of hostname and pid for all MPI processes.", &
                          usage="ECHO_ALL_HOSTS NO", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRACE", &
                          description="If a debug trace of the execution of the program should be written ", &
                          usage="TRACE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRACE_MASTER", &
                          description="For parallel TRACEd runs: only the master node writes output.", &
                          usage="TRACE_MASTER", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="TRACE_MAX", &
         description="Limit the total number a given subroutine is printed in the trace. Accounting is not influenced.", &
         usage="TRACE_MAX 100", default_i_val=HUGE(0))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="TRACE_ROUTINES", &
         description="A list of routines to trace. If left empty all routines are traced. Accounting is not influenced.", &
         usage="TRACE_ROUTINES {routine_name1} {routine_name2} ...", type_of_var=char_t, &
         n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="FLUSH_SHOULD_FLUSH", &
         description="Flush output regularly, enabling this option might degrade performance significantly on certain machines.", &
         usage="FLUSH_SHOULD_FLUSH", &
         default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CALLGRAPH", &
                          description="At the end of the run write a callgraph to file, "// &
                          "which contains detailed timing informations. "// &
                          "This callgraph can be viewed e.g. with the open-source program kcachegrind.", &
                          usage="CALLGRAPH <NONE|MASTER|ALL>", &
                          default_i_val=CALLGRAPH_NONE, lone_keyword_i_val=CALLGRAPH_MASTER, &
                          enum_c_vals=s2a("NONE", "MASTER", "ALL"), &
                          enum_desc=s2a("No callgraph gets written", &
                                        "Only the master process writes his callgraph", &
                                        "All processes write their callgraph (into a separate files)."), &
                          enum_i_vals=(/CALLGRAPH_NONE, CALLGRAPH_MASTER, CALLGRAPH_ALL/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CALLGRAPH_FILE_NAME", &
                          description="Name of the callgraph file, which is writte a the end of the run. "// &
                          "If not specified the project name will be used as filename.", &
                          usage="CALLGRAPH_FILE_NAME {filename}", default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SEED", &
                          description="Initial seed for the global (pseudo)random number "// &
                          "generator to create a stream of normally Gaussian "// &
                          "distributed random numbers.", &
                          usage="SEED <INTEGER>", default_i_val=2000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SAVE_MEM", &
                          description="Some sections of the input structure are deallocated when not needed,"// &
                          " and reallocated only when used. This reduces the required maximum memory  ", &
                          usage="SAVE_MEM", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "TIMINGS", description= &
                                       "Controls the printing of the timing report at the end of CP2K execution", &
                                       print_level=silent_print_level, filename="__STD_OUT__")

      CALL keyword_create(keyword, name="THRESHOLD", &
                          description="Specify % of CPUTIME above which the contribution will be inserted in the"// &
                          " final timing report (e.g. 0.02 = 2%)", &
                          usage="THRESHOLD <real>", &
                          default_r_val=0.02_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SORT_BY_SELF_TIME", &
                          description="Sort the final timing report by the average self (exclusive) time instead of the "// &
                          "total (inclusive) time of a routine", &
                          usage="SORT_BY_SELF_TIME on", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TIME_MPI", &
                          description="Include message_passing calls in the timing report (useful with CALLGRAPH).", &
                          usage="TIME_MPI .TRUE.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "REFERENCES", description= &
                                       "Controls the printing of the references relevant to the calculations performed", &
                                       print_level=silent_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="controls the printing of  initialization controlled by the global section", &
                                       print_level=silent_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PRINT", description= &
                                       "controls the printing of physical and mathematical constants", &
                                       print_level=medium_print_level, filename="__STD_OUT__")

      CALL keyword_create(keyword, "BASIC_DATA_TYPES", &
                          description="Controls the printing of the basic data types.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "physcon", &
                          description="if the printkey is active prints the physical constants", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "spherical_harmonics", &
                          description="if the printkey is active prints the spherical harmonics", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "RNG_MATRICES", &
                          description="Prints the transformation matrices used by the "// &
                          " random number generator", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "RNG_CHECK", &
                          description="Performs a check of the global (pseudo)random "// &
                          "number generator (RNG) and prints the result", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "GLOBAL_GAUSSIAN_RNG", &
                          description="Prints the initial status of the global Gaussian "// &
                          "(pseudo)random number stream which is mostly used for "// &
                          "the velocity initialization", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
      NULLIFY (sub_section)
      ! FM section
      CALL create_fm_section(sub_section)
      CALL section_add_subsection(section, sub_section)
      CALL section_release(sub_section)
      ! DBCSR options
      CALL create_dbcsr_section(sub_section)
      CALL section_add_subsection(section, sub_section)
      CALL section_release(sub_section)
   END SUBROUTINE create_global_section

! **************************************************************************************************
!> \brief   Creates the dbcsr section for configuring DBCSR
!> \param section ...
!> \date    2011-04-05
!> \author  Urban Borstnik
! **************************************************************************************************
   SUBROUTINE create_dbcsr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dbcsr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_global.F",511)
      CALL section_create(section, name="DBCSR", &
                          description="Configuration options for the DBCSR library.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Borstnik2014/))

      NULLIFY (keyword)
      !
      CALL keyword_create(keyword, name="mm_stack_size", &
                          description="Size of multiplication parameter stack." &
                          //" A negative value leaves the decision up to DBCSR.", &
                          usage="mm_stack_size 1000", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="mm_driver", &
                          description="Select which backend to use preferably "// &
                          "for matrix block multiplications on the host.", &
                          usage="mm_driver blas", &
                          default_i_val=mm_driver_auto, &
                          enum_c_vals=s2a(mm_name_auto, mm_name_blas, mm_name_matmul, mm_name_smm, &
                                          mm_name_xsmm), &
                          enum_i_vals=(/mm_driver_auto, mm_driver_blas, mm_driver_matmul, mm_driver_smm, &
                                        mm_driver_xsmm/), &
                          enum_desc=s2a("Choose automatically the best available driver", &
                                        "BLAS (requires the BLAS library at link time)", &
                                        "Fortran MATMUL", &
                                        "Library optimised for Small Matrix Multiplies "// &
                                        "(requires the SMM library at link time)", &
                                        "Intel's libxsmm"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="avg_elements_images", &
                          description="Average number of elements (dense limit)" &
                          //" for each image, which also corresponds to" &
                          //" the average number of elements exchanged between MPI processes" &
                          //" during the operations." &
                          //" A negative or zero value means unlimited.", &
                          usage="avg_elements_images 10000", &
                          default_i_val=dbcsr_get_conf_avg_elements_images())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="randmat_seed", &
                          description="Seed value used to create random matrices in testing", &
                          usage="randmat_seed 3141592", &
                          default_i_val=dbcsr_get_conf_randmat_seed())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="use_mpi_filtering", &
                          description="Use filtering for MPI communications" &
                          //" in RMA algorithm for each image, which also corresponds to" &
                          //" the number of elements exchanged between MPI processes" &
                          //" during the operations.", &
                          usage="use_mpi_filtering F", &
                          default_l_val=dbcsr_get_conf_use_mpi_filtering())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="use_mpi_exp", &
                          description="Use RMA for MPI communications" &
                          //" for each image, which also corresponds to" &
                          //" the number of elements exchanged between MPI processes" &
                          //" during the operations.", &
                          usage="use_mpi_exp F", &
                          default_l_val=dbcsr_get_conf_use_mpi_exp())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="n_size_mnk_stacks", &
                          description="Number of stacks to use for distinct atomic sizes" &
                          //" (e.g., 2 for a system of mostly waters). " &
                          //"A negative value leaves the decision up to DBCSR.", &
                          usage="n_size_mnk_stacks 2", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="use_comm_thread", &
                          description="During multiplication, use a thread to periodically poll" &
                          //" MPI to progress outstanding message completions.  This is" &
                          //" beneficial on systems without a DMA-capable network adapter" &
                          //" e.g. Cray XE6.", &
                          usage="use_comm_thread T", &
                          default_l_val=dbcsr_get_conf_use_comm_thread())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="MAX_ELEMENTS_PER_BLOCK", &
                          description="Default block size for turning dense matrices in blocked ones", &
                          usage="MAX_ELEMENTS_PER_BLOCK 32", &
                          default_i_val=dbcsr_get_conf_max_ele_block())
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="comm_thread_load", &
                          description="If a communications thread is used, specify how much " &
                          //"multiplication workload (%) the thread should perform in " &
                          //"addition to communication tasks. " &
                          //"A negative value leaves the decision up to DBCSR.", &
                          usage="comm_thread_load 50", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="multrec_limit", &
                          description="Recursion limit of cache oblivious multrec algorithm.", &
                          default_i_val=multrec_limit)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !---------------------------------------------------------------------------
      NULLIFY (subsection)
      CALL section_create(subsection, name="ACC", &
                          description="Configuration options for the ACC-Driver.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="priority_buffers", &
                          description="Number of transfer-buffers associated with high priority streams.", &
                          default_i_val=accdrv_priority_buffers)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="posterior_buffers", &
                          description="Number of transfer-buffers associated with low priority streams.", &
                          default_i_val=accdrv_posterior_buffers)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="priority_streams", &
                          description="Number of acc streams created with high priority.", &
                          default_i_val=accdrv_priority_streams)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="posterior_streams", &
                          description="Number of acc streams created with low priority.", &
                          default_i_val=accdrv_posterior_streams)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="avoid_after_busy", &
                          description="If enabled, stacks are not processed by the acc-driver " &
                          //"after it has signaled congestion during a round of flushing. " &
                          //"For the next round of flusing the driver is used again.", &
                          default_l_val=accdrv_avoid_after_busy)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="min_flop_process", &
                          description="Only process stacks with more than the given number of " &
                          //"floating-point operations per stack-entry (2*m*n*k).", &
                          default_i_val=accdrv_min_flop_process)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="min_flop_sort", &
                          description="Only sort stacks with more than the given number of " &
                          //"floating-point operations per stack-entry (2*m*n*k). " &
                          //"Alternatively, the stacks are roughly ordered through a " &
                          //"binning-scheme by Peter Messmer.", &
                          default_i_val=accdrv_min_flop_sort)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="process_inhomogenous", &
                          description="If enabled, inhomogenous stacks are also processed by the acc driver.", &
                          default_l_val=accdrv_do_inhomogenous)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="binning_nbins", &
                          description="Number of bins used when ordering " &
                          //"the stacks with the binning-scheme.", &
                          default_i_val=accdrv_binning_nbins)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="binning_binsize", &
                          description="Size of bins used when ordering " &
                          //"the stacks with the binning-scheme.", &
                          default_i_val=accdrv_binning_binsize)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
      !
   END SUBROUTINE create_dbcsr_section

! **************************************************************************************************
!> \brief   Creates the dbcsr section for configuring FM
!> \param section ...
!> \date    2011-04-05
!> \author  Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE create_fm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_fm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_global.F",718)
      CALL section_create(section, name="FM", &
                          description="Configuration options for the full matrices.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="NROW_BLOCKS", &
                          description="Defines the number of rows per scalapack block in "// &
                          "the creation of block cyclic dense matrices ", &
                          default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NCOL_BLOCKS", &
                          description="Defines the number of columns per scalapack block in "// &
                          "the creation of vlock cyclic dense matrices ", &
                          default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FORCE_BLOCK_SIZE", &
                          description="Ensure for small matrices that the layout is compatible "// &
                          "with bigger ones, i.e. no subdivision is performed (can break LAPACK!!!).", &
                          usage="FORCE_BLOCK_SIZE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE_OF_MATRIX_MULTIPLICATION", &
                          description="Allows to switch between scalapack pdgemm and dbcsr_multiply. "// &
                          "On normal systems pdgemm is recommended on system with GPU "// &
                          "is optimized and can give better performance. NOTE: if DBCSR is employed "// &
                          "FORCE_BLOCK_SIZE should be set. The perfomance on GPU's dependes "// &
                          "crucially on the BLOCK_SIZES. Make sure optimized kernels are available.", &
                          usage="TYPE_OF_MATRIX_MULTIPLICATION ELPA", &
                          default_i_val=do_pdgemm, &
                          enum_i_vals=(/do_pdgemm, do_dbcsr/), &
                          enum_c_vals=s2a("PDGEMM", "DBCSR_MM"), &
                          enum_desc=s2a("Standard scalapack: pdgemm", &
                                        "DBCSR_MM is employed. This needs local transformation of the matrices"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !
   END SUBROUTINE create_fm_section

END MODULE input_cp2k_global
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets up and terminates the global environment variables
!> \par History
!>      - Merged with Quickstep MODULE start_program_run (17.01.2002,MK)
!>      - Compile information added (16.01.2002,MK)
!>      - Merged with MODULE cp2k_input, some rearrangements (30.10.2002,MK)
!> \author JGH,MK
! **************************************************************************************************
MODULE environment
   USE bibliography,                    ONLY: Frigo2005,&
                                              cite_reference
   USE cp2k_info,                       ONLY: &
        compile_arch, compile_date, compile_host, compile_revision, cp2k_flags, cp2k_home, &
        cp2k_version, cp2k_year, get_runtime_info, id_cp2k_version, r_host_name, r_pid, r_user_name
   USE cp_error_handling,               ONLY: warning_counter
   USE cp_files,                        ONLY: close_file,&
                                              get_data_dir,&
                                              open_file
   USE cp_fm_diag,                      ONLY: diag_init
   USE cp_fm_struct,                    ONLY: cp_fm_struct_config
   USE cp_fm_types,                     ONLY: cp_fm_setup
   USE cp_log_handling,                 ONLY: &
        cp_add_default_logger, cp_get_default_logger, cp_logger_create, &
        cp_logger_get_default_unit_nr, cp_logger_release, cp_logger_set, cp_logger_type, &
        cp_rm_default_logger, cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE fft_tools,                       ONLY: fft3d,&
                                              finalize_fft,&
                                              init_fft
   USE force_env_types,                 ONLY: multiple_fe_list
   USE gamma,                           ONLY: deallocate_md_ftable
   USE global_types,                    ONLY: global_environment_type
   USE header,                          ONLY: cp2k_footer,&
                                              cp2k_header
   USE input_constants,                 ONLY: &
        callgraph_all, callgraph_none, do_cp2k, do_diag_elpa, do_diag_sl, do_diag_sl2, do_eip, &
        do_farming, do_fft_fftw3, do_fft_sg, do_fist, do_qs, do_test, energy_run, &
        id_development_version, mol_dyn_run, none_run
   USE input_cp2k_global,               ONLY: create_global_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: &
        section_get_keyword, section_get_lval, section_get_rval, section_release, section_type, &
        section_vals_get, section_vals_get_subs_vals, section_vals_get_subs_vals3, &
        section_vals_type, section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              int_8,&
                                              int_size,&
                                              print_kind_info
   USE machine,                         ONLY: flush_should_flush,&
                                              m_cpuinfo,&
                                              m_energy,&
                                              m_flush_internal,&
                                              m_memory_details,&
                                              m_procrun
   USE message_passing,                 ONLY: add_mp_perf_env,&
                                              describe_mp_perf_env,&
                                              mp_collect_timings,&
                                              mp_max,&
                                              mp_sum,&
                                              rm_mp_perf_env
   USE orbital_pointers,                ONLY: deallocate_orbital_pointers
   USE orbital_transformation_matrices, ONLY: deallocate_spherical_harmonics
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              check_rng,&
                                              create_rng_stream,&
                                              init_rng,&
                                              write_rng_matrices,&
                                              write_rng_stream
   USE physcon,                         ONLY: write_physcon
   USE reference_manager,               ONLY: collect_citations_from_ranks,&
                                              print_all_references,&
                                              print_format_journal
   USE string_utilities,                ONLY: ascii_to_string,&
                                              integer_to_string,&
                                              string_to_ascii
   USE termination,                     ONLY: stop_memory
   USE timings,                         ONLY: add_timer_env,&
                                              rm_timer_env,&
                                              timings_setup_tracing
   USE timings_report,                  ONLY: cost_type_energy,&
                                              cost_type_time,&
                                              timings_report_callgraph,&
                                              timings_report_print

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'environment'

   ! *** Public subroutines ***
   PUBLIC :: cp2k_finalize, cp2k_init, cp2k_read, cp2k_setup, cp2k_get_walltime

CONTAINS

! **************************************************************************************************
!> \brief Initializes a CP2K run (setting of the global environment variables)
!> \param para_env ...
!> \param output_unit ...
!> \param globenv ...
!> \param input_file_name ...
!> \param wdir ...
!> \par History
!>      JGH (28.11.2001) : default for pp_library_path
!>      - print keys added (17.01.2002, MK)
!>      - merged with cp2k_input (30.10.2002,MK)
!> \author JGH,MK
! **************************************************************************************************
   SUBROUTINE cp2k_init(para_env, output_unit, globenv, input_file_name, wdir)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: output_unit
      TYPE(global_environment_type), POINTER             :: globenv
      CHARACTER(LEN=*)                                   :: input_file_name
      CHARACTER(LEN=*), OPTIONAL                         :: wdir

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_init', routineP = moduleN//':'//routineN

      CHARACTER(LEN=10*default_string_length)            :: cp_flags
      CHARACTER(LEN=default_string_length)               :: dev_flag
      INTEGER                                            :: ilen, my_output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

! create a timer_env

      CALL add_timer_env()

      ! message passing performance
      CALL add_mp_perf_env()

      ! Set flag if this is a development version
      dev_flag = ""
      IF (id_cp2k_version == id_development_version) dev_flag = " (Development Version)"

      ! Init the default logger
      IF (para_env%source == para_env%mepos) THEN
         my_output_unit = output_unit
      ELSE
         my_output_unit = -1
      END IF
      NULLIFY (logger)
      CALL cp_logger_create(logger, para_env=para_env, &
                            default_global_unit_nr=output_unit, &
                            close_global_unit_on_dealloc=.FALSE.)
      CALL cp_add_default_logger(logger)
      CALL cp_logger_release(logger)

      ! Initialize timing
      CALL timeset("CP2K", globenv%handle)

      ! Print header
      CALL cp2k_header(my_output_unit, wdir)

      IF (my_output_unit > 0) THEN
         WRITE (UNIT=my_output_unit, FMT="(A)") " "
	 WRITE (UNIT=my_output_unit, FMT="(A)") &
	    "Special version for Android (arm, pie)"
	 WRITE (UNIT=my_output_unit, FMT="(A)") &
	    "compiled on June 27, 2019 by Alan Liska & Veronika Ruzickova."
	 WRITE (UNIT=my_output_unit, FMT="(A)") " "
         WRITE (UNIT=my_output_unit, FMT="(/,T2,A,T31,A50)") &
            "CP2K| version string: ", &
            ADJUSTR(TRIM(cp2k_version)//TRIM(dev_flag))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T41,A40)") &
            "CP2K| source code revision number:", &
            ADJUSTR(compile_revision)
         cp_flags = cp2k_flags()
         ilen = LEN_TRIM(cp_flags)
         WRITE (UNIT=my_output_unit, FMT="(T2,A)") &
            "CP2K| "//cp_flags(1:73)
         IF (ilen > 73) THEN
            WRITE (UNIT=my_output_unit, FMT="(T2,A)") &
               "CP2K|            "//TRIM(cp_flags(74:))
         END IF
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T41,A40)") &
            "CP2K| is freely available from ", &
            ADJUSTR(TRIM(cp2k_home))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T31,A50)") &
            "CP2K| Program compiled at", &
            ADJUSTR(compile_date(1:MIN(50, LEN(compile_date))))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T31,A50)") &
            "CP2K| Program compiled on", &
            ADJUSTR(compile_host(1:MIN(50, LEN(compile_host))))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T31,A50)") &
            "CP2K| Program compiled for", &
            ADJUSTR(compile_arch(1:MIN(50, LEN(compile_arch))))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T31,A50)") &
            "CP2K| Data directory path", &
            ADJUSTR(TRIM(get_data_dir()))
         WRITE (UNIT=my_output_unit, FMT="(T2,A,T31,A50)") &
            "CP2K| Input file name", &
            ADJUSTR(TRIM(input_file_name))
         CALL m_flush_internal(my_output_unit)
      END IF

   END SUBROUTINE cp2k_init

! **************************************************************************************************
!> \brief echos the list of host names and pids
!> \param para_env ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE echo_all_hosts(para_env, output_unit)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'echo_all_hosts', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: ipe, istat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: all_pid
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: all_host

!   *** Print a list of all started processes ***

      ALLOCATE (all_pid(para_env%num_pe), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 232, "all_pid", &
                          para_env%num_pe*int_size)
      END IF
      all_pid(:) = 0
      all_pid(para_env%mepos+1) = r_pid

      CALL mp_sum(all_pid, para_env%group)
      ALLOCATE (all_host(30, para_env%num_pe), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 241, "all_host", &
                          30*para_env%num_pe*int_size)
      END IF
      all_host(:, :) = 0
      CALL string_to_ascii(r_host_name, all_host(:, para_env%mepos+1))
      CALL mp_sum(all_host, para_env%group)
      IF (output_unit > 0) THEN

         WRITE (UNIT=output_unit, FMT="(T2,A)") ""
         DO ipe = 1, para_env%num_pe
            CALL ascii_to_string(all_host(:, ipe), string)
            WRITE (UNIT=output_unit, FMT="(T2,A,T63,I8,T71,I10)") &
               TRIM(r_user_name)//"@"//TRIM(string)// &
               " has created process number", ipe-1, all_pid(ipe)
         END DO
         WRITE (UNIT=output_unit, FMT="(T2,A)") ""
      END IF
      DEALLOCATE (all_pid)
      DEALLOCATE (all_host)

   END SUBROUTINE echo_all_hosts

! **************************************************************************************************
!> \brief echos the list the number of process per host
!> \param para_env ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE echo_all_process_host(para_env, output_unit)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'echo_all_process_host', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string, string_sec
      INTEGER                                            :: ipe, istat, jpe, nr_occu
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: all_pid
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: all_host

      ALLOCATE (all_host(30, para_env%num_pe), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 282, "all_host", &
                          30*para_env%num_pe*int_size)
      END IF
      all_host(:, :) = 0

      IF (m_procrun(r_pid) .EQ. 1) THEN
         CALL string_to_ascii(r_host_name, all_host(:, para_env%mepos+1))
         CALL mp_sum(all_host, para_env%group)
      ENDIF

      IF (output_unit > 0) THEN
         ALLOCATE (all_pid(para_env%num_pe), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 295, "all_pid", &
                             para_env%num_pe*int_size)
         END IF
         all_pid(:) = 0

         WRITE (UNIT=output_unit, FMT="(T2,A)") ""
         DO ipe = 1, para_env%num_pe
            nr_occu = 0
            IF (all_pid(ipe) .NE. -1) THEN
               CALL ascii_to_string(all_host(:, ipe), string)
               DO jpe = 1, para_env%num_pe
                  CALL ascii_to_string(all_host(:, jpe), string_sec)
                  IF (string .EQ. string_sec) THEN
                     nr_occu = nr_occu+1
                     all_pid(jpe) = -1
                  ENDIF
               END DO
               WRITE (UNIT=output_unit, FMT="(T2,A,T63,I8,A)") &
                  TRIM(r_user_name)//"@"//TRIM(string)// &
                  " is running ", nr_occu, " processes"
               WRITE (UNIT=output_unit, FMT="(T2,A)") ""
            END IF
         END DO
         DEALLOCATE (all_pid)

      END IF

      DEALLOCATE (all_host)

   END SUBROUTINE echo_all_process_host

! **************************************************************************************************
!> \brief read part of cp2k_init
!> \param root_section ...
!> \param para_env ...
!> \param globenv the globenv
!> \author fawzi
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   SUBROUTINE cp2k_read(root_section, para_env, globenv)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_read', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: c_val
      INTEGER                                            :: iw
      TYPE(cp_logger_type), POINTER                      :: logger

!   *** Read the input/output section ***

      logger => cp_get_default_logger()

      ! try to use better names for the local log if it is not too late
      CALL section_vals_val_get(root_section, "GLOBAL%OUTPUT_FILE_NAME", &
                                c_val=c_val)
      IF (c_val /= "") THEN
         CALL cp_logger_set(logger, &
                            local_filename=TRIM(c_val)//"_localLog")
      END IF
      CALL section_vals_val_get(root_section, "GLOBAL%PROJECT", c_val=c_val)
      IF (c_val /= "") THEN
         CALL cp_logger_set(logger, local_filename=TRIM(c_val)//"_localLog")
      END IF
      logger%iter_info%project_name = c_val
      CALL section_vals_val_get(root_section, "GLOBAL%PRINT_LEVEL", i_val=logger%iter_info%print_level)

      !   *** Read the CP2K section ***
      CALL read_cp2k_section(root_section, para_env, globenv)

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/BASIC_DATA_TYPES", &
                                extension=".Log")
      IF (iw > 0) CALL print_kind_info(iw)
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PRINT/BASIC_DATA_TYPES")

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/PHYSCON", &
                                extension=".Log")
      IF (iw > 0) CALL write_physcon(iw)
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PRINT/PHYSCON")

   END SUBROUTINE cp2k_read

! **************************************************************************************************
!> \brief globenv initializations that need the input and error
!> \param root_section ...
!> \param para_env ...
!> \param globenv the global environment to initialize
!> \author fawzi
!> \note
!>      if possible do the initializations here as the environement
!>      (error,...) is setup, instaed of cp2k_init
! **************************************************************************************************
   SUBROUTINE cp2k_setup(root_section, para_env, globenv)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp2k_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: input_seed, iw
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! Initialize the parallel random number generator

      CALL init_rng()
      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/RNG_MATRICES", &
                                extension=".Log")
      IF (iw > 0) THEN
         CALL write_rng_matrices(iw)
      END IF
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PRINT/RNG_MATRICES")

      ! Initialize a global normally Gaussian distributed (pseudo)random number stream

      CALL section_vals_val_get(root_section, "GLOBAL%SEED", i_val=input_seed)
      initial_seed(:, :) = REAL(input_seed, KIND=dp)
      CALL create_rng_stream(rng_stream=globenv%gaussian_rng_stream, &
                             name="Global Gaussian random numbers", &
                             distribution_type=GAUSSIAN, &
                             seed=initial_seed, &
                             extended_precision=.TRUE.)

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/RNG_CHECK", &
                                extension=".Log")
      IF (iw > 0) THEN
         CALL check_rng(iw, para_env%ionode)
      END IF
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PRINT/RNG_CHECK")

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/GLOBAL_GAUSSIAN_RNG", &
                                extension=".Log")
      IF (iw > 0) THEN
         CALL write_rng_stream(globenv%gaussian_rng_stream, iw, write_all=.TRUE.)
      END IF
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PRINT/GLOBAL_GAUSSIAN_RNG")

   END SUBROUTINE cp2k_setup

! **************************************************************************************************
!> \brief read the global sectionof new input
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \par History
!>      06-2005 [created]
!> \author MI
!> \note
!>      Should not be required anymore once everything is converted
!>      to get information directly from the input structure
! **************************************************************************************************
   SUBROUTINE read_global_section(root_section, para_env, globenv)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'read_global_section', &
         routineP = moduleN//':'//routineN, start_section_label = "GLOBAL"

      CHARACTER(len=13)                                  :: tracing_string
      CHARACTER(len=6)                                   :: print_level_string
      CHARACTER(len=default_path_length)                 :: basis_set_file_name, coord_file_name, &
                                                            mm_potential_file_name, &
                                                            potential_file_name
      CHARACTER(len=default_string_length)               :: env_num, model_name, project_name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: trace_routines
      INTEGER :: i_diag, i_fft, iforce_eval, method_name_id, n_rep_val, nforce_eval, num_threads, &
         output_unit, print_level, trace_max, unit_nr
      INTEGER(kind=int_8) :: Buffers, Buffers_avr, Buffers_max, Buffers_min, Cached, Cached_avr, &
         Cached_max, Cached_min, MemFree, MemFree_avr, MemFree_max, MemFree_min, MemLikelyFree, &
         MemLikelyFree_avr, MemLikelyFree_max, MemLikelyFree_min, MemTotal, MemTotal_avr, &
         MemTotal_max, MemTotal_min, Slab, Slab_avr, Slab_max, Slab_min, SReclaimable, &
         SReclaimable_avr, SReclaimable_max, SReclaimable_min
      INTEGER, DIMENSION(:), POINTER                     :: i_force_eval
      LOGICAL                                            :: ata, do_echo_all_hosts, efl, explicit, &
                                                            trace, trace_master
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum1, enum2
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: dft_section, force_env_sections, &
                                                            global_section, qmmm_section, &
                                                            subsys_section

      NULLIFY (dft_section, global_section, i_force_eval)

      global_section => section_vals_get_subs_vals(root_section, "GLOBAL")
      CALL section_vals_val_get(global_section, "BLACS_GRID", i_val=globenv%blacs_grid_layout)
      CALL section_vals_val_get(global_section, "BLACS_REPEATABLE", l_val=globenv%blacs_repeatable)
      CALL section_vals_val_get(global_section, "PREFERRED_DIAG_LIBRARY", i_val=i_diag)
      CALL section_vals_val_get(global_section, "ELPA_KERNEL", i_val=globenv%k_elpa)
      CALL section_vals_val_get(global_section, "PREFERRED_FFT_LIBRARY", i_val=i_fft)

      CALL section_vals_val_get(global_section, "PRINT_LEVEL", i_val=print_level)
      CALL section_vals_val_get(global_section, "PROGRAM_NAME", i_val=globenv%prog_name_id)
      CALL section_vals_val_get(global_section, "FFT_POOL_SCRATCH_LIMIT", i_val=globenv%fft_pool_scratch_limit)
      CALL section_vals_val_get(global_section, "FFTW_PLAN_TYPE", i_val=globenv%fftw_plan_type)
      CALL section_vals_val_get(global_section, "PROJECT_NAME", c_val=project_name)
      CALL section_vals_val_get(global_section, "FFTW_WISDOM_FILE_NAME", c_val=globenv%fftw_wisdom_file_name)
      CALL section_vals_val_get(global_section, "RUN_TYPE", i_val=globenv%run_type_id)
      CALL cp2k_get_walltime(section=global_section, keyword_name="WALLTIME", &
                             walltime=globenv%cp2k_target_time)
      CALL section_vals_val_get(global_section, "TRACE", l_val=trace)
      CALL section_vals_val_get(global_section, "TRACE_MASTER", l_val=trace_MASTER)
      CALL section_vals_val_get(global_section, "TRACE_MAX", i_val=trace_max)
      CALL section_vals_val_get(global_section, "TRACE_ROUTINES", explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(global_section, "TRACE_ROUTINES", c_vals=trace_routines)
      ELSE
         NULLIFY (trace_routines)
      ENDIF
      CALL section_vals_val_get(global_section, "FLUSH_SHOULD_FLUSH", l_val=flush_should_flush)
      CALL section_vals_val_get(global_section, "ECHO_ALL_HOSTS", l_val=do_echo_all_hosts)
      force_env_sections => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
      CALL section_vals_get(force_env_sections, n_repetition=nforce_eval)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, global_section, "PROGRAM_RUN_INFO", &
                                         extension=".log")

      CALL fm_setup(global_section)

      IF (trace .AND. (.NOT. trace_master .OR. para_env%mepos == 0)) THEN
         unit_nr = -1
         IF (logger%para_env%mepos == logger%para_env%source .OR. .NOT. trace_master) &
            unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
         WRITE (tracing_string, "(I6.6,A1,I6.6)") para_env%mepos, ":", para_env%num_pe
         IF (ASSOCIATED(trace_routines)) THEN
            CALL timings_setup_tracing(trace_max, unit_nr, tracing_string, trace_routines)
         ELSE
            CALL timings_setup_tracing(trace_max, unit_nr, tracing_string)
         END IF
      ENDIF

      CALL section_vals_val_get(global_section, "TIMINGS%TIME_MPI", l_val=mp_collect_timings)

      SELECT CASE (i_diag)
      CASE (do_diag_sl)
         globenv%diag_library = "SL"
      CASE (do_diag_sl2)
         globenv%diag_library = "SL2"
      CASE (do_diag_elpa)
         globenv%diag_library = "ELPA"
      CASE DEFAULT
         CALL cp__b("environment.F",556,"Unknown DIAG type")
      END SELECT

      SELECT CASE (i_fft)
      CASE (do_fft_sg)
         globenv%default_fft_library = "FFTSG"
      CASE (do_fft_fftw3)
         globenv%default_fft_library = "FFTW3"
         CALL cite_reference(Frigo2005)
      CASE DEFAULT
         CALL cp__b("environment.F",566,"Unknown FFT type")
      END SELECT

      IF (globenv%run_type_id == 0) THEN
         SELECT CASE (globenv%prog_name_id)
         CASE (do_farming, do_test)
            globenv%run_type_id = none_run
         CASE (do_cp2k)
            IF (nforce_eval /= 1) THEN
               ! multiple force_eval corresponds at the moment to RESPA calculations only
               ! default MD
               globenv%run_type_id = mol_dyn_run
            ELSE
               CALL section_vals_val_get(force_env_sections, "METHOD", i_val=method_name_id)
               SELECT CASE (method_name_id)
               CASE (do_fist)
                  globenv%run_type_id = mol_dyn_run
               CASE (do_eip)
                  globenv%run_type_id = mol_dyn_run
               CASE (do_qs)
                  globenv%run_type_id = energy_run
               END SELECT
            END IF
         END SELECT
      END IF

      IF (globenv%prog_name_id == do_farming .AND. globenv%run_type_id /= none_run) THEN
         CALL cp__b("environment.F",593,"FARMING program supports only NONE as run type")
      ENDIF

      IF (globenv%prog_name_id == do_test .AND. globenv%run_type_id /= none_run) &
         CALL cp__b("environment.F",597,"TEST program supports only NONE as run type")

      CALL m_memory_details(MemTotal, MemFree, Buffers, Cached, Slab, SReclaimable, MemLikelyFree)
      MemTotal_avr = MemTotal
      MemFree_avr = MemFree
      Buffers_avr = Buffers
      Cached_avr = Cached
      Slab_avr = Slab
      SReclaimable_avr = SReclaimable
      MemLikelyFree_avr = MemLikelyFree
      CALL mp_sum(MemTotal_avr, para_env%group); MemTotal_avr = MemTotal_avr/para_env%num_pe/1024
      CALL mp_sum(MemFree_avr, para_env%group); MemFree_avr = MemFree_avr/para_env%num_pe/1024
      CALL mp_sum(Buffers_avr, para_env%group); Buffers_avr = Buffers_avr/para_env%num_pe/1024
      CALL mp_sum(Cached_avr, para_env%group); Cached_avr = Cached_avr/para_env%num_pe/1024
      CALL mp_sum(Slab_avr, para_env%group); Slab_avr = Slab_avr/para_env%num_pe/1024
      CALL mp_sum(SReclaimable_avr, para_env%group); SReclaimable_avr = SReclaimable_avr/para_env%num_pe/1024
      CALL mp_sum(MemLikelyFree_avr, para_env%group); MemLikelyFree_avr = MemLikelyFree_avr/para_env%num_pe/1024

      MemTotal_min = -MemTotal
      MemFree_min = -MemFree
      Buffers_min = -Buffers
      Cached_min = -Cached
      Slab_min = -Slab
      SReclaimable_min = -SReclaimable
      MemLikelyFree_min = -MemLikelyFree
      CALL mp_max(MemTotal_min, para_env%group); MemTotal_min = -MemTotal_min/1024
      CALL mp_max(MemFree_min, para_env%group); MemFree_min = -MemFree_min/1024
      CALL mp_max(Buffers_min, para_env%group); Buffers_min = -Buffers_min/1024
      CALL mp_max(Cached_min, para_env%group); Cached_min = -Cached_min/1024
      CALL mp_max(Slab_min, para_env%group); Slab_min = -Slab_min/1024
      CALL mp_max(SReclaimable_min, para_env%group); SReclaimable_min = -SReclaimable_min/1024
      CALL mp_max(MemLikelyFree_min, para_env%group); MemLikelyFree_min = -MemLikelyFree_min/1024

      MemTotal_max = MemTotal
      MemFree_max = MemFree
      Buffers_max = Buffers
      Cached_max = Cached
      Slab_max = Slab
      SReclaimable_max = SReclaimable
      MemLikelyFree_max = MemLikelyFree
      CALL mp_max(MemTotal_max, para_env%group); MemTotal_max = MemTotal_max/1024
      CALL mp_max(MemFree_max, para_env%group); MemFree_max = MemFree_max/1024
      CALL mp_max(Buffers_max, para_env%group); Buffers_max = Buffers_max/1024
      CALL mp_max(Cached_max, para_env%group); Cached_max = Cached_max/1024
      CALL mp_max(Slab_max, para_env%group); Slab_max = Slab_max/1024
      CALL mp_max(SReclaimable_max, para_env%group); SReclaimable_max = SReclaimable_max/1024
      CALL mp_max(MemLikelyFree_max, para_env%group); MemLikelyFree_max = MemLikelyFree_max/1024

      MemTotal = MemTotal/1024
      MemFree = MemFree/1024
      Buffers = Buffers/1024
      Cached = Cached/1024
      Slab = Slab/1024
      SReclaimable = SReclaimable/1024
      MemLikelyFree = MemLikelyFree/1024

      !   *** Print a list of all started processes ***
      IF (do_echo_all_hosts) THEN
         CALL echo_all_hosts(para_env, output_unit)

         ! *** Print the number of processes per host ***
         CALL echo_all_process_host(para_env, output_unit)
      ENDIF

      num_threads = 1
!$    num_threads = omp_get_max_threads()
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT=*)
         CALL multiple_fe_list(force_env_sections, root_section, i_force_eval, nforce_eval)
         DO iforce_eval = 1, nforce_eval
            dft_section => section_vals_get_subs_vals3(force_env_sections, "DFT", &
                                                       i_rep_section=i_force_eval(iforce_eval))
            qmmm_section => section_vals_get_subs_vals3(force_env_sections, "QMMM", &
                                                        i_rep_section=i_force_eval(iforce_eval))
            CALL section_vals_val_get(dft_section, "BASIS_SET_FILE_NAME", &
                                      c_val=basis_set_file_name)
            CALL section_vals_val_get(dft_section, "POTENTIAL_FILE_NAME", &
                                      c_val=potential_file_name)

            CALL section_vals_val_get(qmmm_section, "MM_POTENTIAL_FILE_NAME", &
                                      c_val=mm_potential_file_name)
            ! SUBSYS - If any
            subsys_section => section_vals_get_subs_vals3(force_env_sections, "SUBSYS", &
                                                          i_rep_section=i_force_eval(iforce_eval))
            CALL section_vals_get(subsys_section, explicit=explicit)
            coord_file_name = "__STD_INPUT__"
            IF (explicit) THEN
               CALL section_vals_val_get(subsys_section, "TOPOLOGY%COORD_FILE_NAME", &
                                         n_rep_val=n_rep_val)
               IF (n_rep_val == 1) THEN
                  CALL section_vals_val_get(subsys_section, "TOPOLOGY%COORD_FILE_NAME", &
                                            c_val=coord_file_name)
               END IF
            END IF
            CALL integer_to_string(i_force_eval(iforce_eval), env_num)

            WRITE (UNIT=output_unit, FMT="(T2,A,T41,A)") &
               start_section_label//"| Force Environment number", &
               ADJUSTR(env_num(:40)), &
               start_section_label//"| Basis set file name", &
               ADJUSTR(potential_file_name(:40)), &
               start_section_label//"| MM Potential file name", &
               ADJUSTR(mm_potential_file_name(:40)), &
               start_section_label//"| Coordinate file name", &
               ADJUSTR(coord_file_name(:40))
         END DO
         DEALLOCATE (i_force_eval)

         NULLIFY (enum1, enum2, keyword, section)
         CALL create_global_section(section)
         keyword => section_get_keyword(section, "PROGRAM_NAME")
         CALL keyword_get(keyword, enum=enum1)
         keyword => section_get_keyword(section, "RUN_TYPE")
         CALL keyword_get(keyword, enum=enum2)

         WRITE (UNIT=output_unit, FMT="(T2,A,T41,A40)") &
            start_section_label//"| Method name", &
            ADJUSTR(TRIM(enum_i2c(enum1, globenv%prog_name_id))), &
            start_section_label//"| Project name", &
            ADJUSTR(project_name(:40)), &
            start_section_label//"| Preferred FFT library", &
            ADJUSTR(globenv%default_fft_library(:40)), &
            start_section_label//"| Preferred diagonalization lib.", &
            ADJUSTR(globenv%diag_library(:40)), &
            start_section_label//"| Run type", &
            ADJUSTR(TRIM(enum_i2c(enum2, globenv%run_type_id)))

         CALL section_release(section)

         CALL section_vals_val_get(global_section, "ALLTOALL_SGL", l_val=ata)
         WRITE (UNIT=output_unit, FMT="(T2,A,T80,L1)") &
            start_section_label//"| All-to-all communication in single precision", ata
         CALL section_vals_val_get(global_section, "EXTENDED_FFT_LENGTHS", l_val=efl)
         WRITE (UNIT=output_unit, FMT="(T2,A,T80,L1)") &
            start_section_label//"| FFTs using library dependent lengths", efl

         SELECT CASE (print_level)
         CASE (silent_print_level)
            print_level_string = "SILENT"
         CASE (low_print_level)
            print_level_string = "   LOW"
         CASE (medium_print_level)
            print_level_string = "MEDIUM"
         CASE (high_print_level)
            print_level_string = "  HIGH"
         CASE (debug_print_level)
            print_level_string = " DEBUG"
         CASE DEFAULT
            CALL cp__b("environment.F",745,"Unknown print_level")
         END SELECT

         WRITE (UNIT=output_unit, FMT="(T2,A,T75,A6)") &
            start_section_label//"| Global print level", print_level_string
         WRITE (UNIT=output_unit, FMT="(T2,A,T75,I6)") &
            start_section_label//"| Total number of message passing processes", &
            para_env%num_pe, &
            start_section_label//"| Number of threads for this process", &
            num_threads, &
            start_section_label//"| This output is from process", para_env%mepos

         CALL m_cpuinfo(model_name)
         WRITE (UNIT=output_unit, FMT="(T2,A)") &
            start_section_label//"| CPU model name : "//TRIM(model_name)

         WRITE (UNIT=output_unit, FMT="()")
         WRITE (UNIT=output_unit, FMT="(T2,A)") "MEMORY| system memory details [Kb]"
         WRITE (UNIT=output_unit, FMT="(T2,A23,4A14)") "MEMORY|                ", "rank 0", "min", "max", "average"
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") "MEMORY| MemTotal       ", memtotal, memtotal_min, memtotal_max, memtotal_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") "MEMORY| MemFree        ", memFree, memfree_min, memfree_max, memfree_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") "MEMORY| Buffers        ", Buffers, Buffers_min, Buffers_max, Buffers_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") "MEMORY| Cached         ", Cached, Cached_min, Cached_max, Cached_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") "MEMORY| Slab           ", Slab, Slab_min, Slab_max, Slab_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") &
            "MEMORY| SReclaimable   ", SReclaimable, SReclaimable_min, SReclaimable_max, &
            SReclaimable_avr
         WRITE (UNIT=output_unit, FMT="(T2,A23,4I14)") &
            "MEMORY| MemLikelyFree  ", MemLikelyFree, MemLikelyFree_min, MemLikelyFree_max, &
            MemLikelyFree_avr
         WRITE (UNIT=output_unit, FMT='()')

      END IF

      CALL cp_print_key_finished_output(output_unit, logger, global_section, &
                                        "PROGRAM_RUN_INFO")

   END SUBROUTINE read_global_section

! **************************************************************************************************
!> \brief ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \par History
!>      2-Dec-2000 (JGH) added default fft library
!> \author JGH,MK
! **************************************************************************************************
   SUBROUTINE read_cp2k_section(root_section, para_env, globenv)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'read_cp2k_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: global_section

      global_section => section_vals_get_subs_vals(root_section, "GLOBAL")
      CALL read_global_section(root_section, para_env, globenv)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, global_section, "PROGRAM_RUN_INFO", &
                                         extension=".log")

      CALL fft_setup_library(globenv, global_section, output_unit)
      CALL diag_setup_library(globenv, output_unit)

      CALL cp_print_key_finished_output(output_unit, logger, global_section, &
                                        "PROGRAM_RUN_INFO")

   END SUBROUTINE read_cp2k_section

! **************************************************************************************************
!> \brief check FFT preferred library availability, if not switch
!> \param globenv ...
!> \param global_section ...
!> \param output_unit ...
!> \par History
!>      2-Dec-2000 (JGH) added default fft library
!>      Nov-2013 (MI) refactoring
!> \author JGH,MK
! **************************************************************************************************
   SUBROUTINE fft_setup_library(globenv, global_section, output_unit)

      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: global_section
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'fft_setup_library', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(4, 4, 4)               :: zz
      INTEGER                                            :: stat
      INTEGER, DIMENSION(3)                              :: n
      LOGICAL                                            :: try_fftw

      n(:) = 4
      zz(:, :, :) = 0.0_dp

      !
      ! Setup the FFT library
      ! If the user has specified PREFERRED_FFT_LIBRARY try that first (default FFTW3)
      ! If that one is not available, try FFTW3 (unless it has been tried already)
      ! If FFTW3 is not available use FFTSG
      !
      IF (globenv%default_fft_library .EQ. "FFTW3") THEN
         try_fftw = .FALSE.
      ELSE
         try_fftw = .TRUE.
      ENDIF

      !   *** Initialize FFT library with the user's prefered FFT library ***
      CALL init_fft(fftlib=TRIM(globenv%default_fft_library), &
                    alltoall=section_get_lval(global_section, "ALLTOALL_SGL"), &
                    fftsg_sizes=.NOT. section_get_lval(global_section, "EXTENDED_FFT_LENGTHS"), &
                    pool_limit=globenv%fft_pool_scratch_limit, &
                    wisdom_file=globenv%fftw_wisdom_file_name, &
                    plan_style=globenv%fftw_plan_type)

      !   *** Check for FFT library ***
      CALL fft3d(1, n, zz, status=stat)
      IF (stat /= 0) THEN

         IF (try_fftw) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(A,A,T55,A)') &
                  " WARNING : FFT library "//TRIM(globenv%default_fft_library)// &
                  " is not available ", " Trying FFTW3"
            ENDIF
            globenv%default_fft_library = "FFTW3"
            CALL init_fft(fftlib=TRIM(globenv%default_fft_library), &
                          alltoall=section_get_lval(global_section, "ALLTOALL_SGL"), &
                          fftsg_sizes=.NOT. section_get_lval(global_section, "EXTENDED_FFT_LENGTHS"), &
                          pool_limit=globenv%fft_pool_scratch_limit, &
                          wisdom_file=globenv%fftw_wisdom_file_name, &
                          plan_style=globenv%fftw_plan_type)

            CALL fft3d(1, n, zz, status=stat)
         ENDIF

         IF (stat /= 0) THEN

            IF (output_unit > 0) THEN
               WRITE (output_unit, '(A,A,T55,A)') &
                  " WARNING : FFT library "//TRIM(globenv%default_fft_library)// &
                  " is not available ", " Trying FFTSG as a default"
            ENDIF

            globenv%default_fft_library = "FFTSG"
            CALL init_fft(fftlib=TRIM(globenv%default_fft_library), &
                          alltoall=section_get_lval(global_section, "ALLTOALL_SGL"), &
                          fftsg_sizes=.NOT. section_get_lval(global_section, "EXTENDED_FFT_LENGTHS"), &
                          pool_limit=globenv%fft_pool_scratch_limit, &
                          wisdom_file=globenv%fftw_wisdom_file_name, &
                          plan_style=globenv%fftw_plan_type)

            CALL fft3d(1, n, zz, status=stat)
            IF (stat /= 0) THEN
               CALL cp__b("environment.F",906,"FFTSG not functional....")
            ENDIF

         ENDIF

      END IF

   END SUBROUTINE fft_setup_library

! **************************************************************************************************
!> \brief availability diagonalizatioon library
!>
!> \param globenv ...
!> \param output_unit ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE diag_setup_library(globenv, output_unit)
      TYPE(global_environment_type), POINTER             :: globenv
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'diag_setup_library', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: switched

      switched = .FALSE.

      CALL diag_init(diag_lib=TRIM(globenv%diag_library), switched=switched, k_elpa=globenv%k_elpa)

      IF (switched) THEN

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(A,A,T55,A)') &
               " WARNING : DIAGONALIZATION library "//TRIM(globenv%diag_library)// &
               " is not available ", " Trying SCALAPACK"
         ENDIF

      END IF

   END SUBROUTINE diag_setup_library

! **************************************************************************************************
!> \brief ...
!> \param glob_section ...
! **************************************************************************************************
   SUBROUTINE fm_setup(glob_section)
      TYPE(section_vals_type), POINTER                   :: glob_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fm_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: multiplication_type, ncb, nrb
      LOGICAL                                            :: force_me
      TYPE(section_vals_type), POINTER                   :: fm_section

      fm_section => section_vals_get_subs_vals(glob_section, "FM")

      CALL section_vals_val_get(fm_section, "NROW_BLOCKS", i_val=nrb)
      CALL section_vals_val_get(fm_section, "NCOL_BLOCKS", i_val=ncb)
      CALL section_vals_val_get(fm_section, "FORCE_BLOCK_SIZE", l_val=force_me)

      CALL cp_fm_struct_config(nrow_block=nrb, ncol_block=ncb, force_block=force_me)

      CALL section_vals_val_get(fm_section, "TYPE_OF_MATRIX_MULTIPLICATION", &
                                i_val=multiplication_type)

      CALL cp_fm_setup(multiplication_type)

   END SUBROUTINE fm_setup
! **************************************************************************************************
!> \brief reads the Walltime also in format HH:MM:SS
!> \param section ...
!> \param keyword_name ...
!> \param walltime ...
!> \par History
!>      none
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE cp2k_get_walltime(section, keyword_name, walltime)
      TYPE(section_vals_type), POINTER                   :: section
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      REAL(KIND=dp), INTENT(out)                         :: walltime

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_get_walltime', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: c1, c2
      CHARACTER(LEN=100)                                 :: txt
      INTEGER                                            :: hours, ierr, minutes, n, seconds

      CALL section_vals_val_get(section, keyword_name, c_val=txt)
      n = LEN_TRIM(txt)

      IF (n == 0) THEN
         walltime = -1.0_dp
      ELSE IF (INDEX(txt, ":") == 0) THEN
         READ (txt(1:n), FMT=*, IOSTAT=ierr) walltime
         IF (ierr /= 0) CALL cp__b("environment.F",1002,'Could not parse WALLTIME: "'//txt(1:n)//'"')
      ELSE
         READ (txt(1:n), FMT="(I2,A1,I2,A1,I2)", IOSTAT=ierr) hours, c1, minutes, c2, seconds
         IF (n /= 8 .OR. ierr /= 0 .OR. c1 .NE. ":" .OR. c2 .NE. ":") &
            CALL cp__b("environment.F",1006,'Could not parse WALLTIME: "'//txt(1:n)//'"')
         walltime = 3600.0_dp*REAL(hours, dp)+60.0_dp*REAL(minutes, dp)+REAL(seconds, dp)
      END IF
   END SUBROUTINE cp2k_get_walltime

! **************************************************************************************************
!> \brief Writes final timings and banner for CP2K
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param wdir ...
!> \par History
!>      none
!> \author JGH,MK
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   SUBROUTINE cp2k_finalize(root_section, para_env, globenv, wdir)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      CHARACTER(LEN=*), OPTIONAL                         :: wdir

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_finalize', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: cg_filename
      CHARACTER(LEN=default_string_length)               :: dev_flag
      INTEGER                                            :: cg_mode, iw, unit_exit
      LOGICAL                                            :: delete_it, sort_by_self_time
      REAL(KIND=dp)                                      :: r_timings
      TYPE(cp_logger_type), POINTER                      :: logger

! look if we inherited a failure, more care is needed if so
! i.e. the input is most likely not available
! Set flag if this is a development version

      dev_flag = ""
      IF (id_cp2k_version == id_development_version) dev_flag = " (Development Version)"

      ! Clean up
      NULLIFY (logger)
      logger => cp_get_default_logger()
      CALL deallocate_spherical_harmonics()
      CALL deallocate_orbital_pointers()
      CALL deallocate_md_ftable()

      ! finalize the fft (i.e. writes the wisdom if FFTW3 )
      CALL finalize_fft(para_env, globenv%fftw_wisdom_file_name)

      ! Write message passing performance info

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PROGRAM_RUN_INFO", &
                                extension=".log")
      CALL describe_mp_perf_env(iw)
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%PROGRAM_RUN_INFO")

      CALL collect_citations_from_ranks(para_env)
      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%REFERENCES", &
                                extension=".Log")
      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T30,A,T80,A)") "-", "R E F E R E N C E S", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)

         WRITE (UNIT=iw, FMT="(T2,A)") ""
         WRITE (UNIT=iw, FMT="(T2,A)") TRIM(cp2k_version)//TRIM(dev_flag)//", the CP2K developers group ("//TRIM(cp2k_year)//")."
         WRITE (UNIT=iw, FMT="(T2,A)") "CP2K is freely available from "//TRIM(cp2k_home)//" ."

         CALL print_all_references(sorted=.TRUE., cited_only=.TRUE., &
                                   FORMAT=print_format_journal, unit=iw)
      ENDIF
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%REFERENCES")

      CALL timestop(globenv%handle) ! corresponding the "CP2K" in cp2k_init

      iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%TIMINGS", &
                                extension=".Log")
      r_timings = section_get_rval(root_section, "GLOBAL%TIMINGS%THRESHOLD")
      sort_by_self_time = section_get_lval(root_section, "GLOBAL%TIMINGS%SORT_BY_SELF_TIME")
      IF (m_energy() .NE. 0.0_dp) THEN
         CALL timings_report_print(iw, r_timings, sort_by_self_time, cost_type_energy, para_env)
      ENDIF
      CALL timings_report_print(iw, r_timings, sort_by_self_time, cost_type_time, para_env)

      !Write the callgraph, if desired by user
      CALL section_vals_val_get(root_section, "GLOBAL%CALLGRAPH", i_val=cg_mode)
      IF (cg_mode /= CALLGRAPH_NONE) THEN
         CALL section_vals_val_get(root_section, "GLOBAL%CALLGRAPH_FILE_NAME", c_val=cg_filename)
         IF (LEN_TRIM(cg_filename) == 0) cg_filename = TRIM(logger%iter_info%project_name)
         IF (cg_mode == CALLGRAPH_ALL) & !incorporate mpi-rank into filename
            cg_filename = TRIM(cg_filename)//"_"//TRIM(ADJUSTL(cp_to_string(para_env%mepos)))
         IF (iw > 0) THEN
            WRITE (UNIT=iw, FMT="(T2,3X,A)") "Writing callgraph to: "//TRIM(cg_filename)//".callgraph"
            WRITE (UNIT=iw, FMT="()")
            WRITE (UNIT=iw, FMT="(T2,A)") "-------------------------------------------------------------------------------"
         ENDIF
         IF (cg_mode == CALLGRAPH_ALL .OR. para_env%mepos == para_env%source) &
            CALL timings_report_callgraph(TRIM(cg_filename)//".callgraph")
      END IF

      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "GLOBAL%TIMINGS")

      CALL rm_mp_perf_env()
      CALL rm_timer_env()

      IF (para_env%ionode) THEN
         iw = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PROGRAM_RUN_INFO", &
                                   extension=".log")

         ! Deleting (if existing) the external EXIT files
         delete_it = .FALSE.
         INQUIRE (FILE="/mnt/sdcard/cp2k/work/EXIT", EXIST=delete_it)
         IF (delete_it) THEN
            CALL open_file(file_name="/mnt/sdcard/cp2k/work/EXIT", unit_number=unit_exit)
            CALL close_file(unit_number=unit_exit, file_status="DELETE")
         END IF

         delete_it = .FALSE.
         INQUIRE (FILE=TRIM(logger%iter_info%project_name)//".EXIT", EXIST=delete_it)
         IF (delete_it) THEN
            CALL open_file(file_name=TRIM(logger%iter_info%project_name)//".EXIT", unit_number=unit_exit)
            CALL close_file(unit_number=unit_exit, file_status="DELETE")
         END IF

         ! print warning counter
         IF (iw > 0) THEN
            WRITE (iw, "(T2,A,I0)") "The number of warnings for this run is : ", warning_counter
            WRITE (iw, *) ""
            WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
         ENDIF

         ! update the runtime enviroment variables
         CALL get_runtime_info()

         ! Just a choice, do not print the CP2K footer if there is a failure
         CALL cp2k_footer(iw, wdir)
         IF (iw > 0) CALL m_flush_internal(iw)

         CALL cp_print_key_finished_output(iw, logger, root_section, &
                                           "GLOBAL%PROGRAM_RUN_INFO")
      END IF
      ! Release message passing environment
      CALL cp_rm_default_logger()

   END SUBROUTINE cp2k_finalize

END MODULE environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief parse cp2k input files
!> \par History
!>      06.2004 created [fawzi]
!>      03.2014 moved into separate module [Ole Schuett]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_read
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              empty_initial_variables,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_set_create,&
                                              cp_unit_set_release,&
                                              cp_unit_set_type
   USE input_parsing,                   ONLY: section_vals_parse
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_create,&
                                              section_vals_type,&
                                              typo_match_section
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_read'

   PUBLIC :: read_input, empty_initial_variables

CONTAINS

! **************************************************************************************************
!> \brief reads the cp2k input from the given filepath and returns a section_vals
!>      containing the input
!> \param input_declaration ...
!> \param file_path path where the input should be read
!> \param initial_variables ...
!> \param para_env ...
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION read_input(input_declaration, file_path, initial_variables, para_env) RESULT(res)
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(len=*), INTENT(in)                       :: file_path
      CHARACTER(len=*), DIMENSION(:, :)                  :: initial_variables
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'read_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_parser_type), POINTER                      :: cpparser
      TYPE(cp_unit_set_type), POINTER                    :: default_units

      CALL timeset(routineN, handle)
      NULLIFY (res)
      NULLIFY (cpparser, default_units)
      CALL section_vals_create(res, input_declaration)
      CALL parser_create(cpparser, initial_variables=initial_variables, file_name=file_path, &
                         para_env=para_env)
      CALL cp_unit_set_create(default_units, "OUTPUT")
      typo_match_section => input_declaration
      CALL section_vals_parse(res, cpparser, root_section=.FALSE., &
                              default_units=default_units)
      typo_match_section => NULL()
      CALL cp_unit_set_release(default_units)
      CALL parser_release(cpparser)
      CALL timestop(handle)
   END FUNCTION read_input

END MODULE input_cp2k_read
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief GEEP coefficients for fixed radii. THIS IS THE GEEP LIB.
!>      They can be used together with the "omotetia" rule for erf function
!>      to obtain the expansion for whatever radius one desires
!>      Radius value are in Angstrom
!> \par History
!>      7.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gaussian_data

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER, PARAMETER, PUBLIC       :: min_geep_lib_gauss = 2
   INTEGER, PARAMETER, PUBLIC       :: max_geep_lib_gauss = 18
   ! G2
   REAL(KIND=dp), PARAMETER, PUBLIC :: g2_rc = 1.30_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g2_a1 = 0.121492_dp, g2_b1 = 3.85695_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g2_a2 = 0.104509_dp, g2_b2 = 2.72971_dp
   ! G3
   REAL(KIND=dp), PARAMETER, PUBLIC :: g3_rc = 1.17_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g3_a1 = 0.1142400_dp, g3_b1 = 3.08635_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g3_a2 = 0.0867876_dp, g3_b2 = 2.38345_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g3_a3 = 0.0971514_dp, g3_b3 = 4.35065_dp
   ! G4
   REAL(KIND=dp), PARAMETER, PUBLIC :: g4_rc = 1.05_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g4_a1 = 0.0776927_dp, g4_b1 = 2.10429_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g4_a2 = 0.1123350_dp, g4_b2 = 2.59523_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g4_a3 = 0.0812506_dp, g4_b3 = 4.73860_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g4_a4 = 0.0989222_dp, g4_b4 = 3.44579_dp
   ! G5
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_rc = 0.9_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_a1 = 0.0728205_dp, g5_b1 = 4.98419_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_a2 = 0.0797713_dp, g5_b2 = 1.78985_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_a3 = 0.0885763_dp, g5_b3 = 3.68614_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_a4 = 0.1104310_dp, g5_b4 = 2.78214_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g5_a5 = 0.1211280_dp, g5_b5 = 2.15508_dp
   ! G6
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_rc = 0.9_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a1 = 0.0863850_dp, g6_b1 = 3.20872_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a2 = 0.0704031_dp, g6_b2 = 4.12508_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a3 = 0.0604464_dp, g6_b3 = 5.40108_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a4 = 0.1029610_dp, g6_b4 = 2.04751_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a5 = 0.1020950_dp, g6_b5 = 2.52820_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g6_a6 = 0.0621064_dp, g6_b6 = 1.76831_dp
   ! G7
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_rc = 0.85_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a1 = 0.0969005_dp, g7_b1 = 1.88076_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a2 = 0.0556821_dp, g7_b2 = 1.65955_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a3 = 0.0607934_dp, g7_b3 = 4.42485_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a4 = 0.0888440_dp, g7_b4 = 2.79941_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a5 = 0.0535919_dp, g7_b5 = 5.68373_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a6 = 0.0736222_dp, g7_b6 = 3.50534_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g7_a7 = 0.1011310_dp, g7_b7 = 2.26261_dp
   ! G8
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_rc = 0.80_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a1 = 0.0557370_dp, g8_b1 = 1.55883_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a2 = 0.0450800_dp, g8_b2 = 6.02024_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a3 = 0.0496262_dp, g8_b3 = 4.82212_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a4 = 0.0757863_dp, g8_b4 = 3.16120_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a5 = 0.0608617_dp, g8_b5 = 3.90652_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a6 = 0.1040810_dp, g8_b6 = 2.09075_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a7 = 0.0983610_dp, g8_b7 = 1.75423_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g8_a8 = 0.0920273_dp, g8_b8 = 2.55933_dp
   ! G9
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_rc = 0.70_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a1 = 0.1152140_dp, g9_b1 = 1.54224_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a2 = 0.0280838_dp, g9_b2 = 5.91022_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a3 = 0.0718097_dp, g9_b3 = 3.54224_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a4 = 0.0585113_dp, g9_b4 = 4.44860_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a5 = 0.1068470_dp, g9_b5 = 2.27491_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a6 = 0.0883974_dp, g9_b6 = 2.83230_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a7 = 0.1211100_dp, g9_b7 = 1.84735_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a8 = 0.0273353_dp, g9_b8 = 5.52297_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g9_a9 = 0.0657478_dp, g9_b9 = 1.36539_dp
   ! G10
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_rc = 0.60_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a1 = 0.1005460_dp, g10_b1 = 2.49407_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a2 = 0.1246370_dp, g10_b2 = 1.99718_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a3 = 0.0492434_dp, g10_b3 = 4.70767_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a4 = 0.0785904_dp, g10_b4 = 3.10735_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a5 = 0.0242791_dp, g10_b5 = 5.81268_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a6 = 0.0229056_dp, g10_b6 = 5.92886_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a7 = 0.0606229_dp, g10_b7 = 3.83892_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a8 = 0.1442920_dp, g10_b8 = 1.61113_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a9 = 0.0814189_dp, g10_b9 = 1.17276_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g10_a10 = 0.1401800_dp, g10_b10 = 1.33402_dp
   ! G11
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_rc = 0.50_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a1 = 0.1041690_dp, g11_b1 = 0.979627_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a2 = 0.0271666_dp, g11_b2 = 5.799740_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a3 = 0.0200111_dp, g11_b3 = 5.799660_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a4 = 0.0483906_dp, g11_b4 = 4.203330_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a5 = 0.1777430_dp, g11_b5 = 1.370190_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a6 = 0.0731972_dp, g11_b6 = 3.415900_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a7 = 0.0950076_dp, g11_b7 = 2.724470_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a8 = 0.0295802_dp, g11_b8 = 4.825220_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a9 = 0.1212380_dp, g11_b9 = 2.162170_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a10 = 0.1513510_dp, g11_b10 = 1.714860_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g11_a11 = 0.1760780_dp, g11_b11 = 1.123400_dp
   ! G12
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_rc = 0.40_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a1 = 0.1420750_dp, g12_b1 = 0.786470_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a2 = 0.0880114_dp, g12_b2 = 2.942550_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a3 = 0.0645509_dp, g12_b3 = 3.669640_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a4 = 0.1505330_dp, g12_b4 = 1.826690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a5 = 0.0183100_dp, g12_b5 = 5.965520_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a6 = 0.0182645_dp, g12_b6 = 5.965520_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a7 = 0.1160340_dp, g12_b7 = 2.326450_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a8 = 0.0449121_dp, g12_b8 = 4.458350_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a9 = 0.0264801_dp, g12_b9 = 5.182710_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a10 = 0.2338230_dp, g12_b10 = 0.912631_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a11 = 0.2295300_dp, g12_b11 = 1.128850_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g12_a12 = 0.1911980_dp, g12_b12 = 1.431550_dp
   ! G13
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_rc = 0.30_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a1 = 0.2001960_dp, g13_b1 = 0.591285_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a2 = 0.0884682_dp, g13_b2 = 2.958380_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a3 = 0.1999450_dp, g13_b3 = 1.420640_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a4 = 0.3235490_dp, g13_b4 = 0.691636_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a5 = 0.0636926_dp, g13_b5 = 3.693610_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a6 = 0.0409795_dp, g13_b6 = 4.450810_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a7 = 0.1172330_dp, g13_b7 = 2.332480_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a8 = 0.0278787_dp, g13_b8 = 5.119550_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a9 = 0.0187190_dp, g13_b9 = 5.952110_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a10 = 0.0187159_dp, g13_b10 = 5.952110_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a11 = 0.3118340_dp, g13_b11 = 0.863435_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a12 = 0.2562320_dp, g13_b12 = 1.104410_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g13_a13 = 0.1535690_dp, g13_b13 = 1.824620_dp
   ! G14
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_rc = 0.20_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a1 = 0.3778350_dp, g14_b1 = 0.786779_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a2 = 0.0944711_dp, g14_b2 = 2.781090_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a3 = 0.0204159_dp, g14_b3 = 5.876050_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a4 = 0.0204239_dp, g14_b4 = 5.876050_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a5 = 0.2884320_dp, g14_b5 = 1.023350_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a6 = 0.2186240_dp, g14_b6 = 1.326440_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a7 = 0.0700271_dp, g14_b7 = 3.483220_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a8 = 0.1656890_dp, g14_b8 = 1.710300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a9 = 0.0493301_dp, g14_b9 = 4.268870_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a10 = 0.1254840_dp, g14_b10 = 2.190730_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a11 = 0.0303071_dp, g14_b11 = 4.999930_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a12 = 0.5269670_dp, g14_b12 = 0.475269_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a13 = 0.3496170_dp, g14_b13 = 0.397232_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g14_a14 = 0.4776640_dp, g14_b14 = 0.605916_dp
   ! G15
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_rc = 0.10_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a1 = 1.3606800_dp, g15_b1 = 0.310940_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a2 = 0.0266233_dp, g15_b2 = 5.801190_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a3 = 0.3896280_dp, g15_b3 = 0.813967_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a4 = 0.0155094_dp, g15_b4 = 5.800280_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a5 = 0.5574040_dp, g15_b5 = 0.609044_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a6 = 0.8507370_dp, g15_b6 = 0.443840_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a7 = 0.1590220_dp, g15_b7 = 1.808170_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a8 = 0.2837530_dp, g15_b8 = 1.071380_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a9 = 0.0645391_dp, g15_b9 = 3.723390_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a10 = 0.0241980_dp, g15_b10 = 4.946500_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a11 = 0.0344342_dp, g15_b11 = 4.462030_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a12 = 0.2112170_dp, g15_b12 = 1.396840_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a13 = 0.1202950_dp, g15_b13 = 2.325120_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a14 = 0.0904425_dp, g15_b14 = 2.965550_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g15_a15 = 1.6040000_dp, g15_b15 = 0.216434_dp
   ! G16
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_rc = 0.10_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a1 = 1.5843100_dp, g16_b1 = 0.215959_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a2 = 0.0253725_dp, g16_b2 = 5.737070_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a3 = 0.1191710_dp, g16_b3 = 2.233620_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a4 = 0.2848860_dp, g16_b4 = 1.050490_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a5 = 0.0134257_dp, g16_b5 = 5.720440_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a6 = 0.0269327_dp, g16_b6 = 4.542090_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a7 = 0.0888076_dp, g16_b7 = 2.807370_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a8 = 0.3907140_dp, g16_b8 = 0.801210_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a9 = 0.5584020_dp, g16_b9 = 0.601540_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a10 = 0.0674428_dp, g16_b10 = 3.477020_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a11 = 0.0167448_dp, g16_b11 = 5.320770_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a12 = 0.2120830_dp, g16_b12 = 1.363230_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a13 = 0.0363796_dp, g16_b13 = 4.204950_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a14 = 0.8512990_dp, g16_b14 = 0.439692_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a15 = 0.1591170_dp, g16_b15 = 1.753690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g16_a16 = 1.3566500_dp, g16_b16 = 0.308985_dp
   ! G17
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_rc = 0.10_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a1 = 0.0286263_dp, g17_b1 = 5.897570_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a2 = 1.3621700_dp, g17_b2 = 0.210930_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a3 = 0.8582340_dp, g17_b3 = 0.396804_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a4 = 0.0918589_dp, g17_b4 = 2.514930_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a5 = 0.2314390_dp, g17_b5 = 1.077620_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a6 = 1.3030900_dp, g17_b6 = 0.288252_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a7 = 0.1437920_dp, g17_b7 = 1.658970_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a8 = 0.3010060_dp, g17_b8 = 0.860535_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a9 = 0.0274321_dp, g17_b9 = 5.099500_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a10 = 0.1814220_dp, g17_b10 = 1.340220_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a11 = 0.4046220_dp, g17_b11 = 0.679621_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a12 = 0.0735338_dp, g17_b12 = 3.080580_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a13 = 0.0577419_dp, g17_b13 = 3.752150_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a14 = 0.0388557_dp, g17_b14 = 4.481620_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a15 = 0.1147140_dp, g17_b15 = 2.045960_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a16 = 0.5719720_dp, g17_b16 = 0.526925_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g17_a17 = 0.0222331_dp, g17_b17 = 6.674390_dp
   ! G18
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_rc = 0.10_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a1 = 0.0326473_dp, g18_b1 = 5.962800_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a2 = 0.0181199_dp, g18_b2 = 6.674950_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a3 = 0.8582970_dp, g18_b3 = 0.396103_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a4 = 0.0880221_dp, g18_b4 = 2.468750_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a5 = 0.1812460_dp, g18_b5 = 1.332910_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a6 = 0.2314380_dp, g18_b6 = 1.072960_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a7 = 0.0449632_dp, g18_b7 = 3.435210_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a8 = 1.3583000_dp, g18_b8 = 0.210847_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a9 = 0.1132310_dp, g18_b9 = 2.024200_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a10 = 0.0251085_dp, g18_b10 = 4.226690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a11 = 0.4047630_dp, g18_b11 = 0.677704_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a12 = 0.3010870_dp, g18_b12 = 0.857535_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a13 = 1.3019600_dp, g18_b13 = 0.287909_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a14 = 0.1432190_dp, g18_b14 = 1.646940_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a15 = 0.0634725_dp, g18_b15 = 2.959960_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a16 = 0.5721690_dp, g18_b16 = 0.525730_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a17 = 0.0309731_dp, g18_b17 = 3.918400_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: g18_a18 = 0.0429091_dp, g18_b18 = 4.928750_dp
   !
   ! S-WAVE expansion
   ! S2
   REAL(KIND=dp), PARAMETER, PUBLIC :: s2_rc = 1.2_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s2_a1 = 0.1167720_dp, s2_b1 = 3.15760_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s2_a2 = 0.0537459_dp, s2_b2 = 1.66713_dp
   ! S3
   REAL(KIND=dp), PARAMETER, PUBLIC :: s3_rc = 1.05_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s3_a1 = 0.1258370_dp, s3_b1 = 3.54518_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s3_a2 = 0.0920562_dp, s3_b2 = 2.17201_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s3_a3 = 0.0354098_dp, s3_b3 = 1.16797_dp
   ! S4
   REAL(KIND=dp), PARAMETER, PUBLIC :: s4_rc = 0.90_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s4_a1 = 0.1237040_dp, s4_b1 = 3.82356_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s4_a2 = 0.0263957_dp, s4_b2 = 0.84445_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s4_a3 = 0.0773548_dp, s4_b3 = 1.56300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s4_a4 = 0.1243300_dp, s4_b4 = 2.49568_dp
   ! S5
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_rc = 0.80_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_a1 = 0.0601509_dp, s5_b1 = 1.169940_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_a2 = 0.1133640_dp, s5_b2 = 1.853660_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_a3 = 0.1120710_dp, s5_b3 = 4.084220_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_a4 = 0.1334100_dp, s5_b4 = 2.762410_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s5_a5 = 0.0186655_dp, s5_b5 = 0.633305_dp
   ! S6
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_rc = 0.70_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a1 = 0.0494163_dp, s6_b1 = 0.889897_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a2 = 0.1415590_dp, s6_b2 = 2.073620_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a3 = 0.1311110_dp, s6_b3 = 2.977960_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a4 = 0.1012530_dp, s6_b4 = 4.306440_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a5 = 0.1037040_dp, s6_b5 = 1.405500_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s6_a6 = 0.0144409_dp, s6_b6 = 0.482098_dp
   ! S7
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_rc = 0.65_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a1 = 0.08156290_dp, s7_b1 = 1.104000_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a2 = 0.00971659_dp, s7_b2 = 0.379602_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a3 = 0.09067690_dp, s7_b3 = 4.537130_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a4 = 0.14334400_dp, s7_b4 = 2.290480_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a5 = 0.13012600_dp, s7_b5 = 1.619190_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a6 = 0.11794100_dp, s7_b6 = 3.205850_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s7_a7 = 0.03510960_dp, s7_b7 = 0.700361_dp
   ! S8
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_rc = 0.60_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a1 = 0.14676300_dp, s8_b1 = 1.802170_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a2 = 0.00686211_dp, s8_b2 = 0.302201_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a3 = 0.08248590_dp, s8_b3 = 4.741690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a4 = 0.10588700_dp, s8_b4 = 3.410740_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a5 = 0.06409090_dp, s8_b5 = 0.877859_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a6 = 0.11461200_dp, s8_b6 = 1.283730_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a7 = 0.13708200_dp, s8_b7 = 2.482170_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s8_a8 = 0.02571900_dp, s8_b8 = 0.557431_dp
   ! S9
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_rc = 0.55_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a1 = 0.0760147_dp, s9_b1 = 4.924130_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a2 = 0.0050867_dp, s9_b2 = 0.242656_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a3 = 0.1524530_dp, s9_b3 = 1.962780_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a4 = 0.0993944_dp, s9_b4 = 1.028540_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a5 = 0.1433380_dp, s9_b5 = 1.438020_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a6 = 0.1267950_dp, s9_b6 = 2.655630_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a7 = 0.0958376_dp, s9_b7 = 3.595590_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a8 = 0.0511698_dp, s9_b8 = 0.704455_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s9_a9 = 0.0195500_dp, s9_b9 = 0.447544_dp
   ! S10
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_rc = 0.5_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a1 = 0.11585200_dp, s10_b1 = 2.814420_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a2 = 0.01551000_dp, s10_b2 = 0.361841_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a3 = 0.08772520_dp, s10_b3 = 3.762340_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a4 = 0.16279600_dp, s10_b4 = 1.572690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a5 = 0.08664120_dp, s10_b5 = 0.830482_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a6 = 0.13675500_dp, s10_b6 = 1.158370_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a7 = 0.04201910_dp, s10_b7 = 0.569340_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a8 = 0.14880800_dp, s10_b8 = 2.107180_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a9 = 0.07080240_dp, s10_b9 = 5.087590_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s10_a10 = 0.00396733_dp, s10_b10 = 0.196230_dp
   ! S11
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_rc = 0.45_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a1 = 0.16967500_dp, s11_b1 = 1.274890_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a2 = 0.01316070_dp, s11_b2 = 0.296680_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a3 = 0.06640200_dp, s11_b3 = 5.239300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a4 = 0.10578100_dp, s11_b4 = 2.964600_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a5 = 0.07769660_dp, s11_b5 = 0.677602_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a6 = 0.08097080_dp, s11_b6 = 3.917730_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a7 = 0.16973700_dp, s11_b7 = 1.696540_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a8 = 0.13064000_dp, s11_b8 = 0.942777_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a9 = 0.03627480_dp, s11_b9 = 0.465541_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a10 = 0.13917200_dp, s11_b10 = 2.244230_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s11_a11 = 0.00336822_dp, s11_b11 = 0.161548_dp
   ! S12
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_rc = 0.40_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a1 = 0.09642700_dp, s12_b1 = 3.118890_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a2 = 0.17691800_dp, s12_b2 = 1.052170_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a3 = 0.06232170_dp, s12_b3 = 5.393160_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a4 = 0.01269890_dp, s12_b4 = 0.252183_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a5 = 0.07481830_dp, s12_b5 = 4.075740_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a6 = 0.12917000_dp, s12_b6 = 0.782380_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a7 = 0.03450160_dp, s12_b7 = 0.391473_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a8 = 0.12664700_dp, s12_b8 = 2.387640_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a9 = 0.07450870_dp, s12_b9 = 0.565577_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a10 = 0.18916800_dp, s12_b10 = 1.390920_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a11 = 0.16350600_dp, s12_b11 = 1.824470_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s12_a12 = 0.00339586_dp, s12_b12 = 0.139492_dp
   ! S13
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_rc = 0.35_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a1 = 0.0827110_dp, s13_b1 = 3.406540_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a2 = 0.1088730_dp, s13_b2 = 2.646630_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a3 = 0.0548105_dp, s13_b3 = 5.651890_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a4 = 0.1851320_dp, s13_b4 = 1.572120_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a5 = 0.0153281_dp, s13_b5 = 0.227432_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a6 = 0.1447980_dp, s13_b6 = 0.684903_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a7 = 0.0400841_dp, s13_b7 = 0.348420_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a8 = 0.1976780_dp, s13_b8 = 0.915974_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a9 = 0.0845825_dp, s13_b9 = 0.498603_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a10 = 0.2124150_dp, s13_b10 = 1.204860_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a11 = 0.0643854_dp, s13_b11 = 4.368000_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a12 = 0.1438660_dp, s13_b12 = 2.044290_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s13_a13 = 0.0043540_dp, s13_b13 = 0.127923_dp
   ! S14
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_rc = 0.3_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a1 = 0.1185130_dp, s14_b1 = 2.483340_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a2 = 0.2451880_dp, s14_b2 = 1.119610_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a3 = 0.0585112_dp, s14_b3 = 0.327881_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a4 = 0.2441530_dp, s14_b4 = 0.851146_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a5 = 0.1561530_dp, s14_b5 = 1.909900_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a6 = 0.1178640_dp, s14_b6 = 0.466300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a7 = 0.0710939_dp, s14_b7 = 4.176560_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a8 = 0.0907698_dp, s14_b8 = 3.219580_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a9 = 0.0298147_dp, s14_b9 = 5.495980_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a10 = 0.0233818_dp, s14_b10 = 0.215814_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a11 = 0.2037940_dp, s14_b11 = 1.464080_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a12 = 0.1909540_dp, s14_b12 = 0.637845_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a13 = 0.0300539_dp, s14_b13 = 5.481830_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s14_a14 = 0.0069580_dp, s14_b14 = 0.122528_dp
   ! S15
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_rc = 0.25_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a1 = 0.0945048_dp, s15_b1 = 2.945510_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a2 = 0.2601310_dp, s15_b2 = 0.589258_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a3 = 0.1267470_dp, s15_b3 = 2.288550_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a4 = 0.0389633_dp, s15_b4 = 0.203996_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a5 = 0.1688290_dp, s15_b5 = 1.763470_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a6 = 0.2822050_dp, s15_b6 = 1.031520_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a7 = 0.3048940_dp, s15_b7 = 0.784165_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a8 = 0.0262863_dp, s15_b8 = 5.598990_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a9 = 0.0231427_dp, s15_b9 = 5.844390_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a10 = 0.0484052_dp, s15_b10 = 4.606440_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a11 = 0.0680845_dp, s15_b11 = 3.734880_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a12 = 0.0918188_dp, s15_b12 = 0.306903_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a13 = 0.2236530_dp, s15_b13 = 1.351070_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a14 = 0.1732960_dp, s15_b14 = 0.433086_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s15_a15 = 0.0122787_dp, s15_b15 = 0.116976_dp
   ! S16
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_rc = 0.20_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a1 = 0.0355363_dp, s16_b1 = 5.119830_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a2 = 0.0250461_dp, s16_b2 = 0.111228_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a3 = 0.0555360_dp, s16_b3 = 4.232680_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a4 = 0.0747956_dp, s16_b4 = 3.406690_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a5 = 0.0739676_dp, s16_b5 = 0.192075_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a6 = 0.1016150_dp, s16_b6 = 2.694530_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a7 = 0.1600940_dp, s16_b7 = 0.285880_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a8 = 0.3823200_dp, s16_b8 = 0.717807_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a9 = 0.0174934_dp, s16_b9 = 5.965260_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a10 = 0.1369610_dp, s16_b10 = 2.100230_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a11 = 0.1835200_dp, s16_b11 = 1.620280_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a12 = 0.2453470_dp, s16_b12 = 1.240770_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a13 = 0.3219940_dp, s16_b13 = 0.945350_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a14 = 0.0174978_dp, s16_b14 = 5.965260_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a15 = 0.3677700_dp, s16_b15 = 0.540682_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s16_a16 = 0.2742210_dp, s16_b16 = 0.399796_dp
   ! S17
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_rc = 0.15_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a1 = 0.0844127_dp, s17_b1 = 3.242020_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a2 = 0.5321910_dp, s17_b2 = 0.495959_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a3 = 0.4682070_dp, s17_b3 = 0.659162_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a4 = 0.0316333_dp, s17_b4 = 4.520730_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a5 = 0.1497940_dp, s17_b5 = 1.957400_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a6 = 0.0181484_dp, s17_b6 = 5.592050_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a7 = 0.0181805_dp, s17_b7 = 5.592050_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a8 = 0.4758870_dp, s17_b8 = 0.368353_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a9 = 0.0474720_dp, s17_b9 = 4.043220_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a10 = 0.2678960_dp, s17_b10 = 1.148850_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a11 = 0.1998770_dp, s17_b11 = 1.504430_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a12 = 0.0181970_dp, s17_b12 = 5.592050_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a13 = 0.3261780_dp, s17_b13 = 0.265656_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a14 = 0.1720050_dp, s17_b14 = 0.180302_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a15 = 0.1127310_dp, s17_b15 = 2.529320_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a16 = 0.3595840_dp, s17_b16 = 0.872121_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s17_a17 = 0.0642967_dp, s17_b17 = 0.105311_dp
   ! S18
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_rc = 0.05_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a1 = 0.1013430_dp, s18_b1 = 0.0270641_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a2 = 0.0583885_dp, s18_b2 = 4.2001100_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a3 = 1.1587600_dp, s18_b3 = 0.3072560_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a4 = 0.1255030_dp, s18_b4 = 2.3069300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a5 = 0.0469195_dp, s18_b5 = 3.5331200_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a6 = 0.3728020_dp, s18_b6 = 0.0498868_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a7 = 0.8990960_dp, s18_b7 = 0.0785500_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a8 = 0.3146440_dp, s18_b8 = 1.0197600_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a9 = 1.5250400_dp, s18_b9 = 0.1149420_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a10 = 0.4308660_dp, s18_b10 = 0.7644820_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a11 = 0.8278010_dp, s18_b11 = 0.4197150_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a12 = 0.5947340_dp, s18_b12 = 0.5687230_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a13 = 1.5771300_dp, s18_b13 = 0.2235160_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a14 = 0.1707930_dp, s18_b14 = 1.7733300_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a15 = 1.8160800_dp, s18_b15 = 0.1616990_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a16 = 0.0857559_dp, s18_b16 = 2.9441700_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a17 = 0.2313250_dp, s18_b17 = 1.3500600_dp
   REAL(KIND=dp), PARAMETER, PUBLIC :: s18_a18 = 0.0593302_dp, s18_b18 = 5.4023800_dp

END MODULE qmmm_gaussian_data

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set the QMMM Gaussian Input Environment
!> \par History
!>      6.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gaussian_input
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: rootpi
   USE physcon,                         ONLY: bohr
   USE qmmm_gaussian_data,              ONLY: &
        g10_a1, g10_a10, g10_a2, g10_a3, g10_a4, g10_a5, g10_a6, g10_a7, g10_a8, g10_a9, g10_b1, &
        g10_b10, g10_b2, g10_b3, g10_b4, g10_b5, g10_b6, g10_b7, g10_b8, g10_b9, g10_rc, g11_a1, &
        g11_a10, g11_a11, g11_a2, g11_a3, g11_a4, g11_a5, g11_a6, g11_a7, g11_a8, g11_a9, g11_b1, &
        g11_b10, g11_b11, g11_b2, g11_b3, g11_b4, g11_b5, g11_b6, g11_b7, g11_b8, g11_b9, g11_rc, &
        g12_a1, g12_a10, g12_a11, g12_a12, g12_a2, g12_a3, g12_a4, g12_a5, g12_a6, g12_a7, g12_a8, &
        g12_a9, g12_b1, g12_b10, g12_b11, g12_b12, g12_b2, g12_b3, g12_b4, g12_b5, g12_b6, g12_b7, &
        g12_b8, g12_b9, g12_rc, g13_a1, g13_a10, g13_a11, g13_a12, g13_a13, g13_a2, g13_a3, &
        g13_a4, g13_a5, g13_a6, g13_a7, g13_a8, g13_a9, g13_b1, g13_b10, g13_b11, g13_b12, &
        g13_b13, g13_b2, g13_b3, g13_b4, g13_b5, g13_b6, g13_b7, g13_b8, g13_b9, g13_rc, g14_a1, &
        g14_a10, g14_a11, g14_a12, g14_a13, g14_a14, g14_a2, g14_a3, g14_a4, g14_a5, g14_a6, &
        g14_a7, g14_a8, g14_a9, g14_b1, g14_b10, g14_b11, g14_b12, g14_b13, g14_b14, g14_b2, &
        g14_b3, g14_b4, g14_b5, g14_b6, g14_b7, g14_b8, g14_b9, g14_rc, g15_a1, g15_a10, g15_a11, &
        g15_a12, g15_a13, g15_a14, g15_a15, g15_a2, g15_a3, g15_a4, g15_a5, g15_a6, g15_a7, &
        g15_a8, g15_a9, g15_b1, g15_b10, g15_b11, g15_b12, g15_b13, g15_b14, g15_b15, g15_b2, &
        g15_b3, g15_b4, g15_b5, g15_b6, g15_b7, g15_b8, g15_b9, g15_rc, g16_a1, g16_a10, g16_a11, &
        g16_a12, g16_a13, g16_a14, g16_a15, g16_a16, g16_a2, g16_a3, g16_a4, g16_a5, g16_a6, &
        g16_a7, g16_a8, g16_a9, g16_b1, g16_b10, g16_b11, g16_b12, g16_b13, g16_b14, g16_b15, &
        g16_b16, g16_b2, g16_b3, g16_b4, g16_b5, g16_b6, g16_b7, g16_b8, g16_b9, g16_rc, g17_a1, &
        g17_a10, g17_a11, g17_a12, g17_a13, g17_a14, g17_a15, g17_a16, g17_a17, g17_a2, g17_a3, &
        g17_a4, g17_a5, g17_a6, g17_a7, g17_a8, g17_a9, g17_b1, g17_b10, g17_b11, g17_b12, &
        g17_b13, g17_b14, g17_b15, g17_b16, g17_b17, g17_b2, g17_b3, g17_b4, g17_b5, g17_b6, &
        g17_b7, g17_b8, g17_b9, g18_a1, g18_a10, g18_a11, g18_a12, g18_a13, g18_a14, g18_a15, &
        g18_a16, g18_a17, g18_a18, g18_a2, g18_a3, g18_a4, g18_a5, g18_a6, g18_a7, g18_a8, g18_a9, &
        g18_b1, g18_b10, g18_b11, g18_b12, g18_b13, g18_b14, g18_b15, g18_b16, g18_b17, g18_b18, &
        g18_b2, g18_b3, g18_b4, g18_b5, g18_b6, g18_b7, g18_b8, g18_b9, g2_a1, g2_a2, g2_b1, &
        g2_b2, g2_rc, g3_a1, g3_a2, g3_a3, g3_b1, g3_b2, g3_b3, g3_rc, g4_a1, g4_a2, g4_a3, g4_a4, &
        g4_b1, g4_b2, g4_b3, g4_b4, g4_rc, g5_a1, g5_a2, g5_a3, g5_a4, g5_a5, g5_b1, g5_b2, g5_b3, &
        g5_b4, g5_b5, g5_rc, g6_a1, g6_a2, g6_a3, g6_a4, g6_a5, g6_a6, g6_b1, g6_b2, g6_b3, g6_b4, &
        g6_b5, g6_b6, g6_rc, g7_a1, g7_a2, g7_a3, g7_a4, g7_a5, g7_a6, g7_a7, g7_b1, g7_b2, g7_b3, &
        g7_b4, g7_b5, g7_b6, g7_b7, g7_rc, g8_a1, g8_a2, g8_a3, g8_a4, g8_a5, g8_a6, g8_a7, g8_a8, &
        g8_b1, g8_b2, g8_b3, g8_b4, g8_b5, g8_b6, g8_b7, g8_b8, g8_rc, g9_a1, g9_a2, g9_a3, g9_a4, &
        g9_a5, g9_a6, g9_a7, g9_a8, g9_a9, g9_b1, g9_b2, g9_b3, g9_b4, g9_b5, g9_b6, g9_b7, g9_b8, &
        g9_b9, g9_rc, s10_a1, s10_a10, s10_a2, s10_a3, s10_a4, s10_a5, s10_a6, s10_a7, s10_a8, &
        s10_a9, s10_b1, s10_b10, s10_b2, s10_b3, s10_b4, s10_b5, s10_b6, s10_b7, s10_b8, s10_b9, &
        s10_rc, s11_a1, s11_a10, s11_a11, s11_a2, s11_a3, s11_a4, s11_a5, s11_a6, s11_a7, s11_a8, &
        s11_a9, s11_b1, s11_b10, s11_b11, s11_b2, s11_b3, s11_b4, s11_b5, s11_b6, s11_b7, s11_b8, &
        s11_b9, s11_rc, s12_a1, s12_a10, s12_a11, s12_a12, s12_a2, s12_a3, s12_a4, s12_a5, s12_a6, &
        s12_a7, s12_a8, s12_a9, s12_b1, s12_b10, s12_b11, s12_b12, s12_b2, s12_b3, s12_b4, s12_b5, &
        s12_b6, s12_b7, s12_b8, s12_b9, s12_rc, s13_a1, s13_a10, s13_a11, s13_a12, s13_a13, &
        s13_a2, s13_a3, s13_a4, s13_a5, s13_a6, s13_a7, s13_a8, s13_a9, s13_b1, s13_b10, s13_b11, &
        s13_b12, s13_b13, s13_b2, s13_b3, s13_b4, s13_b5, s13_b6, s13_b7, s13_b8, s13_b9, s13_rc, &
        s14_a1, s14_a10, s14_a11, s14_a12, s14_a13, s14_a14, s14_a2, s14_a3, s14_a4, s14_a5, &
        s14_a6, s14_a7, s14_a8, s14_a9, s14_b1, s14_b10, s14_b11, s14_b12, s14_b13, s14_b14, &
        s14_b2, s14_b3, s14_b4, s14_b5, s14_b6, s14_b7, s14_b8, s14_b9, s14_rc, s15_a1, s15_a10, &
        s15_a11, s15_a12, s15_a13, s15_a14, s15_a15, s15_a2, s15_a3, s15_a4, s15_a5, s15_a6, &
        s15_a7, s15_a8, s15_a9, s15_b1, s15_b10, s15_b11, s15_b12, s15_b13, s15_b14, s15_b15, &
        s15_b2, s15_b3, s15_b4, s15_b5, s15_b6, s15_b7, s15_b8, s15_b9, s15_rc, s16_a1, s16_a10, &
        s16_a11, s16_a12, s16_a13, s16_a14, s16_a15, s16_a16, s16_a2, s16_a3, s16_a4, s16_a5, &
        s16_a6, s16_a7, s16_a8, s16_a9, s16_b1, s16_b10, s16_b11, s16_b12, s16_b13, s16_b14, &
        s16_b15, s16_b16, s16_b2, s16_b3, s16_b4, s16_b5, s16_b6, s16_b7, s16_b8, s16_b9, s16_rc, &
        s17_a1, s17_a10, s17_a11, s17_a12, s17_a13, s17_a14, s17_a15, s17_a16, s17_a17, s17_a2, &
        s17_a3, s17_a4, s17_a5, s17_a6, s17_a7, s17_a8, s17_a9, s17_b1, s17_b10, s17_b11, s17_b12, &
        s17_b13, s17_b14, s17_b15, s17_b16, s17_b17, s17_b2, s17_b3, s17_b4, s17_b5, s17_b6, &
        s17_b7, s17_b8, s17_b9, s17_rc, s18_a1, s18_a10, s18_a11, s18_a12, s18_a13, s18_a14, &
        s18_a15, s18_a16, s18_a17, s18_a18, s18_a2, s18_a3, s18_a4, s18_a5, s18_a6, s18_a7, &
        s18_a8, s18_a9, s18_b1, s18_b10, s18_b11, s18_b12, s18_b13, s18_b14, s18_b15, s18_b16, &
        s18_b17, s18_b18, s18_b2, s18_b3, s18_b4, s18_b5, s18_b6, s18_b7, s18_b8, s18_b9, s18_rc, &
        s2_a1, s2_a2, s2_b1, s2_b2, s2_rc, s3_a1, s3_a2, s3_a3, s3_b1, s3_b2, s3_b3, s3_rc, s4_a1, &
        s4_a2, s4_a3, s4_a4, s4_b1, s4_b2, s4_b3, s4_b4, s4_rc, s5_a1, s5_a2, s5_a3, s5_a4, s5_a5, &
        s5_b1, s5_b2, s5_b3, s5_b4, s5_b5, s5_rc, s6_a1, s6_a2, s6_a3, s6_a4, s6_a5, s6_a6, s6_b1, &
        s6_b2, s6_b3, s6_b4, s6_b5, s6_b6, s6_rc, s7_a1, s7_a2, s7_a3, s7_a4, s7_a5, s7_a6, s7_a7, &
        s7_b1, s7_b2, s7_b3, s7_b4, s7_b5, s7_b6, s7_b7, s7_rc, s8_a1, s8_a2, s8_a3, s8_a4, s8_a5, &
        s8_a6, s8_a7, s8_a8, s8_b1, s8_b2, s8_b3, s8_b4, s8_b5, s8_b6, s8_b7, s8_b8, s8_rc, s9_a1, &
        s9_a2, s9_a3, s9_a4, s9_a5, s9_a6, s9_a7, s9_a8, s9_a9, s9_b1, s9_b2, s9_b3, s9_b4, s9_b5, &
        s9_b6, s9_b7, s9_b8, s9_b9, s9_rc
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_gaussian_input'

   PUBLIC  :: read_mm_potential, &
              set_mm_potential_swave, &
              set_mm_potential_erf
!***
CONTAINS

! **************************************************************************************************
!> \brief read MM_POTENTIAL file
!> \param para_env ...
!> \param qmmm_gaussian_fns ...
!> \param compatibility ...
!> \param qmmm_section ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE read_mm_potential(para_env, qmmm_gaussian_fns, &
                                compatibility, qmmm_section)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: qmmm_gaussian_fns
      LOGICAL, INTENT(IN)                                :: compatibility
      TYPE(section_vals_type), POINTER                   :: qmmm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_mm_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=default_string_length)               :: Ftarget, mm_potential_file_name, Units
      INTEGER                                            :: IRad, istart, Ival, j, Nog, Nval, &
                                                            output_unit
      LOGICAL                                            :: Found, Found_Radius
      REAL(KIND=dp)                                      :: fconv, my_radius, Radius
      TYPE(cp_parser_type), POINTER                      :: parser

      output_unit = cp_logger_get_default_io_unit()
      Nval = SIZE(qmmm_gaussian_fns)
      Ival = 0
      CALL section_vals_val_get(qmmm_section, "MM_POTENTIAL_FILE_NAME", &
                                c_val=mm_potential_file_name)

      NULLIFY (parser)
      CALL parser_create(parser, mm_potential_file_name, para_env=para_env)

      search_loop: DO
         Ftarget = "&MM_FIT_POT"
         IF (Ival .EQ. Nval) EXIT search_loop
         CALL parser_search_string(parser, Ftarget, .TRUE., found, line)
         IF (Found) THEN
!
! Structure example of the MM fit potential file:
!
!           &MM_FIT_POT
!           RADIUS  0.4 Angstrom
!           7
!           0.223396   0.811453  Bohr
!           0.306814   1.01988   Bohr
!           0.254879   1.37404   Bohr
!           0.188293   1.87929   Bohr
!           0.136391   2.56745   Bohr
!           0.100305   3.50033   Bohr
!           0.0790169  4.82046   Bohr
!           &END
!
            CALL parser_get_object(parser, Ftarget, newline=.TRUE.)
            IF(.NOT.(TRIM(Ftarget) == "RADIUS"))CALL cp__a("qmmm_gaussian_input.F",166)
            CALL parser_get_object(parser, radius)
            CALL parser_get_object(parser, units)
            CALL uppercase(units)
            fconv = 1.0_dp
            IF (TRIM(units) .EQ. "ANGSTROM") fconv = bohr
            Found_Radius = .FALSE.
            radius = radius*fconv
            Radius_Loop: DO J = 1, SIZE(qmmm_gaussian_fns)
               IF (ABS(radius-qmmm_gaussian_fns(J)%pgf%Elp_Radius) .LT. EPSILON(0.0_dp)) THEN
                  Found_Radius = .TRUE.
                  EXIT Radius_Loop
               END IF
            END DO Radius_Loop
            IF (.NOT. Found_Radius) THEN
               CYCLE search_loop
            END IF
            Ival = Ival+1
            IRad = J
            ! Read  Rmin, Rmax
            CALL parser_get_object(parser, qmmm_gaussian_fns(J)%pgf%Number_of_Gaussians, newline=.TRUE.)
            ! Allocate Vectors
            istart = 1
            IF (compatibility) THEN
               qmmm_gaussian_fns(J)%pgf%Number_of_Gaussians = qmmm_gaussian_fns(J)%pgf%Number_of_Gaussians+1
               istart = 2
            END IF
            NOG = qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians
            ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Ak(NOG))
            ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Gk(NOG))
            IF (compatibility) THEN
               my_radius = qmmm_gaussian_fns(J)%pgf%Elp_Radius_corr
               qmmm_gaussian_fns(IRad)%pgf%Ak(1) = 1.0_dp/radius-2.0_dp/(rootpi*radius)
               qmmm_gaussian_fns(IRad)%pgf%Gk(1) = my_radius
            END IF
            DO J = istart, qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians
               CALL parser_get_object(parser, qmmm_gaussian_fns(IRad)%pgf%Ak(J), newline=.TRUE.)
               CALL parser_get_object(parser, qmmm_gaussian_fns(IRad)%pgf%Gk(J))
               CALL parser_get_object(parser, units)
               CALL uppercase(units)
               fconv = 1.0_dp
               IF (TRIM(units) .EQ. "ANGSTROM") fconv = bohr
               qmmm_gaussian_fns(IRad)%pgf%Ak(J) = qmmm_gaussian_fns(IRad)%pgf%Ak(J)*fconv
               qmmm_gaussian_fns(IRad)%pgf%Gk(J) = qmmm_gaussian_fns(IRad)%pgf%Gk(J)*fconv
            END DO
         ELSE
!       *** Stop program, if the end of file is reached ***
            IF (output_unit > 0) WRITE (output_unit, '(A,F12.6,A)') "Discrepancies in Radius: ", Radius, " definition."
            CALL cp__b("qmmm_gaussian_input.F",214,"Radius Value not found in MM_POTENTIAL file")
         END IF

      END DO search_loop

      CALL parser_release(parser)

   END SUBROUTINE read_mm_potential

! **************************************************************************************************
!> \brief set the GEEP information for Erf(r/rc)/r
!> \param qmmm_gaussian_fns ...
!> \param compatibility ...
!> \param num_geep_gauss ...
!> \par History
!>      07.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE set_mm_potential_erf(qmmm_gaussian_fns, &
                                   compatibility, num_geep_gauss)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: qmmm_gaussian_fns
      LOGICAL, INTENT(IN)                                :: compatibility
      INTEGER, INTENT(IN)                                :: num_geep_gauss

      CHARACTER(len=*), PARAMETER :: routineN = 'set_mm_potential_erf', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: IRad, istart, Nog, Nval
      REAL(KIND=dp)                                      :: my_radius, radius, rc

      Nval = SIZE(qmmm_gaussian_fns)
      DO IRad = 1, Nval
         qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians = num_geep_gauss
         radius = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius
         istart = 0
         ! Allocate Vectors
         IF (compatibility) THEN
            qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians = qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians+1
            istart = 1
         END IF
         SELECT CASE (num_geep_gauss)
         CASE (2)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g2_rc*bohr)
         CASE (3)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g3_rc*bohr)
         CASE (4)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g4_rc*bohr)
         CASE (5)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g5_rc*bohr)
         CASE (6)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g6_rc*bohr)
         CASE (7)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g7_rc*bohr)
         CASE (8)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g8_rc*bohr)
         CASE (9)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g9_rc*bohr)
         CASE (10)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g10_rc*bohr)
         CASE (11)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g11_rc*bohr)
         CASE (12)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g12_rc*bohr)
         CASE (13)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g13_rc*bohr)
         CASE (14)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g14_rc*bohr)
         CASE (15)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g15_rc*bohr)
         CASE (16)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g16_rc*bohr)
         CASE (17)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g16_rc*bohr)
         CASE (18)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(g16_rc*bohr)
         END SELECT
         NOG = qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians
         ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Ak(NOG))
         ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Gk(NOG))
         IF (compatibility) THEN
            my_radius = qmmm_gaussian_fns(IRad)%pgf%Elp_Radius_corr
            qmmm_gaussian_fns(IRad)%pgf%Ak(1) = 1.0_dp/radius-2.0_dp/(rootpi*radius)
            qmmm_gaussian_fns(IRad)%pgf%Gk(1) = my_radius
         END IF
         SELECT CASE (num_geep_gauss)
         CASE (2)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g2_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g2_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g2_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g2_b2
         CASE (3)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g3_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g3_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g3_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g3_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g3_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g3_b3
         CASE (4)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g4_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g4_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g4_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g4_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g4_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g4_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g4_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g4_b4
         CASE (5)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g5_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g5_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g5_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g5_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g5_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g5_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g5_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g5_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g5_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g5_b5
         CASE (6)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g6_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g6_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g6_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g6_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g6_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g6_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g6_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g6_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g6_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g6_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g6_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g6_b6
         CASE (7)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g7_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g7_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g7_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g7_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g7_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g7_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g7_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g7_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g7_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g7_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g7_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g7_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g7_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g7_b7
         CASE (8)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g8_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g8_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g8_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g8_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g8_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g8_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g8_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g8_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g8_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g8_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g8_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g8_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g8_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g8_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g8_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g8_b8
         CASE (9)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g9_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g9_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g9_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g9_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g9_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g9_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g9_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g9_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g9_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g9_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g9_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g9_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g9_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g9_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g9_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g9_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g9_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g9_b9
         CASE (10)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g10_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g10_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g10_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g10_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g10_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g10_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g10_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g10_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g10_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g10_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g10_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g10_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g10_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g10_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g10_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g10_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g10_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g10_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g10_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g10_b10
         CASE (11)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g11_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g11_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g11_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g11_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g11_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g11_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g11_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g11_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g11_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g11_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g11_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g11_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g11_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g11_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g11_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g11_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g11_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g11_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g11_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g11_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g11_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g11_b11
         CASE (12)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g12_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g12_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g12_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g12_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g12_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g12_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g12_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g12_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g12_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g12_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g12_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g12_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g12_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g12_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g12_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g12_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g12_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g12_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g12_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g12_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g12_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g12_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g12_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g12_b12
         CASE (13)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g13_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g13_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g13_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g13_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g13_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g13_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g13_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g13_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g13_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g13_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g13_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g13_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g13_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g13_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g13_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g13_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g13_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g13_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g13_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g13_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g13_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g13_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g13_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g13_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g13_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g13_b13
         CASE (14)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g14_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g14_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g14_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g14_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g14_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g14_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g14_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g14_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g14_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g14_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g14_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g14_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g14_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g14_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g14_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g14_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g14_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g14_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g14_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g14_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g14_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g14_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g14_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g14_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g14_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g14_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = g14_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = g14_b14
         CASE (15)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g15_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g15_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g15_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g15_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g15_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g15_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g15_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g15_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g15_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g15_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g15_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g15_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g15_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g15_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g15_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g15_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g15_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g15_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g15_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g15_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g15_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g15_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g15_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g15_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g15_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g15_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = g15_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = g15_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = g15_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = g15_b15
         CASE (16)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g16_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g16_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g16_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g16_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g16_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g16_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g16_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g16_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g16_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g16_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g16_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g16_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g16_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g16_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g16_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g16_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g16_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g16_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g16_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g16_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g16_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g16_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g16_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g16_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g16_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g16_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = g16_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = g16_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = g16_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = g16_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = g16_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = g16_b16
         CASE (17)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g17_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g17_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g17_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g17_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g17_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g17_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g17_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g17_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g17_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g17_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g17_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g17_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g17_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g17_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g17_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g17_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g17_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g17_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g17_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g17_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g17_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g17_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g17_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g17_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g17_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g17_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = g17_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = g17_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = g17_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = g17_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = g17_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = g17_b16
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+17) = g17_a17
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+17) = g17_b17
         CASE (18)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = g18_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = g18_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = g18_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = g18_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = g18_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = g18_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = g18_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = g18_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = g18_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = g18_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = g18_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = g18_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = g18_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = g18_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = g18_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = g18_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = g18_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = g18_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = g18_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = g18_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = g18_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = g18_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = g18_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = g18_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = g18_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = g18_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = g18_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = g18_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = g18_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = g18_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = g18_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = g18_b16
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+17) = g18_a17
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+17) = g18_b17
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+18) = g18_a18
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+18) = g18_b18
         END SELECT
         qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1:) = qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1:)/rc
         qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1:) = qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1:)*rc
      END DO
   END SUBROUTINE set_mm_potential_erf

! **************************************************************************************************
!> \brief set the GEEP information for the S-WAVE expansion
!> \param qmmm_gaussian_fns ...
!> \param num_geep_gauss ...
!> \par History
!>      02.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE set_mm_potential_swave(qmmm_gaussian_fns, &
                                     num_geep_gauss)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: qmmm_gaussian_fns
      INTEGER, INTENT(IN)                                :: num_geep_gauss

      CHARACTER(len=*), PARAMETER :: routineN = 'set_mm_potential_swave', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: IRad, istart, Nog, Nval
      REAL(KIND=dp)                                      :: radius, rc

      Nval = SIZE(qmmm_gaussian_fns)
      DO IRad = 1, Nval
         qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians = num_geep_gauss
         radius = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius
         istart = 0
         ! Allocate Vectors
         SELECT CASE (num_geep_gauss)
         CASE (2)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s2_rc*bohr)
         CASE (3)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s3_rc*bohr)
         CASE (4)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s4_rc*bohr)
         CASE (5)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s5_rc*bohr)
         CASE (6)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s6_rc*bohr)
         CASE (7)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s7_rc*bohr)
         CASE (8)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s8_rc*bohr)
         CASE (9)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s9_rc*bohr)
         CASE (10)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s10_rc*bohr)
         CASE (11)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s11_rc*bohr)
         CASE (12)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s12_rc*bohr)
         CASE (13)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s13_rc*bohr)
         CASE (14)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s14_rc*bohr)
         CASE (15)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s15_rc*bohr)
         CASE (16)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s16_rc*bohr)
         CASE (17)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s17_rc*bohr)
         CASE (18)
            rc = qmmm_gaussian_fns(Irad)%pgf%Elp_Radius/(s18_rc*bohr)
         END SELECT
         NOG = qmmm_gaussian_fns(IRad)%pgf%Number_of_Gaussians
         ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Ak(NOG))
         ALLOCATE (qmmm_gaussian_fns(IRad)%pgf%Gk(NOG))
         SELECT CASE (num_geep_gauss)
         CASE (2)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s2_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s2_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s2_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s2_b2
         CASE (3)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s3_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s3_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s3_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s3_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s3_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s3_b3
         CASE (4)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s4_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s4_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s4_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s4_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s4_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s4_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s4_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s4_b4
         CASE (5)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s5_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s5_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s5_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s5_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s5_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s5_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s5_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s5_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s5_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s5_b5
         CASE (6)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s6_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s6_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s6_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s6_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s6_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s6_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s6_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s6_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s6_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s6_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s6_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s6_b6
         CASE (7)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s7_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s7_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s7_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s7_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s7_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s7_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s7_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s7_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s7_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s7_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s7_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s7_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s7_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s7_b7
         CASE (8)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s8_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s8_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s8_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s8_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s8_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s8_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s8_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s8_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s8_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s8_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s8_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s8_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s8_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s8_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s8_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s8_b8
         CASE (9)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s9_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s9_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s9_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s9_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s9_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s9_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s9_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s9_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s9_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s9_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s9_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s9_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s9_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s9_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s9_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s9_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s9_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s9_b9
         CASE (10)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s10_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s10_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s10_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s10_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s10_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s10_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s10_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s10_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s10_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s10_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s10_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s10_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s10_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s10_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s10_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s10_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s10_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s10_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s10_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s10_b10
         CASE (11)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s11_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s11_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s11_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s11_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s11_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s11_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s11_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s11_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s11_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s11_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s11_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s11_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s11_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s11_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s11_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s11_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s11_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s11_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s11_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s11_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s11_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s11_b11
         CASE (12)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s12_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s12_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s12_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s12_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s12_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s12_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s12_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s12_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s12_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s12_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s12_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s12_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s12_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s12_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s12_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s12_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s12_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s12_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s12_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s12_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s12_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s12_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s12_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s12_b12
         CASE (13)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s13_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s13_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s13_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s13_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s13_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s13_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s13_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s13_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s13_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s13_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s13_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s13_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s13_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s13_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s13_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s13_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s13_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s13_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s13_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s13_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s13_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s13_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s13_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s13_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s13_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s13_b13
         CASE (14)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s14_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s14_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s14_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s14_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s14_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s14_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s14_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s14_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s14_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s14_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s14_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s14_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s14_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s14_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s14_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s14_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s14_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s14_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s14_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s14_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s14_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s14_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s14_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s14_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s14_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s14_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = s14_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = s14_b14
         CASE (15)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s15_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s15_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s15_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s15_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s15_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s15_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s15_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s15_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s15_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s15_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s15_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s15_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s15_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s15_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s15_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s15_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s15_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s15_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s15_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s15_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s15_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s15_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s15_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s15_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s15_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s15_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = s15_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = s15_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = s15_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = s15_b15
         CASE (16)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s16_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s16_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s16_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s16_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s16_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s16_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s16_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s16_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s16_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s16_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s16_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s16_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s16_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s16_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s16_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s16_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s16_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s16_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s16_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s16_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s16_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s16_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s16_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s16_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s16_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s16_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = s16_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = s16_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = s16_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = s16_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = s16_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = s16_b16
         CASE (17)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s17_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s17_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s17_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s17_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s17_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s17_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s17_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s17_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s17_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s17_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s17_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s17_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s17_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s17_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s17_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s17_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s17_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s17_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s17_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s17_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s17_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s17_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s17_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s17_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s17_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s17_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = s17_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = s17_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = s17_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = s17_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = s17_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = s17_b16
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+17) = s17_a17
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+17) = s17_b17
         CASE (18)
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1) = s18_a1
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1) = s18_b1
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+2) = s18_a2
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+2) = s18_b2
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+3) = s18_a3
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+3) = s18_b3
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+4) = s18_a4
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+4) = s18_b4
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+5) = s18_a5
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+5) = s18_b5
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+6) = s18_a6
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+6) = s18_b6
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+7) = s18_a7
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+7) = s18_b7
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+8) = s18_a8
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+8) = s18_b8
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+9) = s18_a9
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+9) = s18_b9
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+10) = s18_a10
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+10) = s18_b10
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+11) = s18_a11
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+11) = s18_b11
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+12) = s18_a12
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+12) = s18_b12
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+13) = s18_a13
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+13) = s18_b13
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+14) = s18_a14
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+14) = s18_b14
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+15) = s18_a15
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+15) = s18_b15
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+16) = s18_a16
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+16) = s18_b16
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+17) = s18_a17
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+17) = s18_b17
            qmmm_gaussian_fns(IRad)%pgf%Ak(istart+18) = s18_a18
            qmmm_gaussian_fns(IRad)%pgf%Gk(istart+18) = s18_b18
         END SELECT
         qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1:) = qmmm_gaussian_fns(IRad)%pgf%Ak(istart+1:)/rc
         qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1:) = qmmm_gaussian_fns(IRad)%pgf%Gk(istart+1:)*rc
      END DO
   END SUBROUTINE set_mm_potential_swave

END MODULE qmmm_gaussian_input

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize the use of the gaussians to treat the QMMM
!>      coupling potential
!> \par History
!>      6.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gaussian_init
   USE ao_util,                         ONLY: exp_radius
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel,&
                                              gridlevel_info_type
   USE input_constants,                 ONLY: do_qmmm_gauss,&
                                              do_qmmm_swave
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE memory_utilities,                ONLY: reallocate
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_p_type
   USE qmmm_gaussian_data,              ONLY: max_geep_lib_gauss,&
                                              min_geep_lib_gauss
   USE qmmm_gaussian_input,             ONLY: read_mm_potential,&
                                              set_mm_potential_erf,&
                                              set_mm_potential_swave
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type,&
                                              qmmm_gaussian_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_gaussian_init'
   PUBLIC :: qmmm_gaussian_initialize

CONTAINS

! **************************************************************************************************
!> \brief Initialize the Gaussian QMMM Environment
!> \param qmmm_gaussian_fns ...
!> \param para_env ...
!> \param pw_env ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param qmmm_coupl_type ...
!> \param eps_mm_rspace ...
!> \param maxradius ...
!> \param maxchrg ...
!> \param compatibility ...
!> \param print_section ...
!> \param qmmm_section ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_gaussian_initialize(qmmm_gaussian_fns, para_env, pw_env, &
                                       mm_el_pot_radius, mm_el_pot_radius_corr, &
                                       qmmm_coupl_type, eps_mm_rspace, maxradius, maxchrg, compatibility, &
                                       print_section, qmmm_section)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: qmmm_gaussian_fns
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_el_pot_radius, mm_el_pot_radius_corr
      INTEGER, INTENT(IN)                                :: qmmm_coupl_type
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      REAL(KIND=dp), DIMENSION(:), POINTER               :: maxradius
      REAL(KIND=dp), INTENT(IN)                          :: maxchrg
      LOGICAL, INTENT(IN)                                :: compatibility
      TYPE(section_vals_type), POINTER                   :: print_section, qmmm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_gaussian_initialize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ilevel, j, Ndim, num_geep_gauss, &
                                                            output_unit
      LOGICAL                                            :: Found, use_geep_lib
      REAL(KIND=dp)                                      :: alpha, mymaxradius, Prefactor
      REAL(KIND=dp), DIMENSION(:), POINTER               :: c_radius, radius
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pools
      TYPE(qmmm_gaussian_type), POINTER                  :: mypgf

! Statements

      NULLIFY (mypgf, gridlevel_info, radius, c_radius, logger)
      logger => cp_get_default_logger()
      CALL section_vals_val_get(qmmm_section, "USE_GEEP_LIB", i_val=num_geep_gauss)
      IF (num_geep_gauss == 0) THEN
         use_geep_lib = .FALSE.
      ELSE
         use_geep_lib = .TRUE.
         IF(.NOT.(num_geep_gauss >= min_geep_lib_gauss))CALL cp__a("qmmm_gaussian_init.F",104)
         IF(.NOT.(num_geep_gauss <= max_geep_lib_gauss))CALL cp__a("qmmm_gaussian_init.F",105)
      END IF
      SELECT CASE (qmmm_coupl_type)
      CASE (do_qmmm_gauss, do_qmmm_swave)
         !
         ! Preprocessing...
         !
         ALLOCATE (radius(1))
         ALLOCATE (c_radius(1))
         Ndim = SIZE(radius)
         Loop_on_all_values: DO I = 1, SIZE(mm_el_pot_radius)
            Found = .FALSE.
            Loop_on_found_values: DO J = 1, SIZE(radius)-1
               IF (mm_el_pot_radius(i) .EQ. radius(j)) THEN
                  Found = .TRUE.
                  EXIT Loop_on_found_values
               END IF
            END DO Loop_on_found_values
            IF (.NOT. Found) THEN
               Ndim = SIZE(radius)
               radius(Ndim) = mm_el_pot_radius(i)
               c_radius(Ndim) = mm_el_pot_radius_corr(i)
               Ndim = Ndim+1
               CALL REALLOCATE(radius, 1, Ndim)
               CALL REALLOCATE(c_radius, 1, Ndim)
            END IF
         END DO Loop_on_all_values
         !
         IF (Ndim-1 > 0) THEN
            CALL REALLOCATE(radius, 1, Ndim-1)
            CALL REALLOCATE(c_radius, 1, Ndim-1)
         ELSE IF (Ndim-1 == 0) THEN
            DEALLOCATE (radius)
            DEALLOCATE (c_radius)
         ELSE
            CALL cp__b("qmmm_gaussian_init.F",140,"")
         END IF
         !
         ALLOCATE (qmmm_gaussian_fns(Ndim-1))
         DO I = 1, Ndim-1
            NULLIFY (qmmm_gaussian_fns(I)%pgf)
            ALLOCATE (qmmm_gaussian_fns(I)%pgf)
            NULLIFY (qmmm_gaussian_fns(I)%pgf%Ak)
            NULLIFY (qmmm_gaussian_fns(I)%pgf%Gk)
            NULLIFY (qmmm_gaussian_fns(I)%pgf%grid_level)
            !
            ! Default Values
            !
            qmmm_gaussian_fns(I)%pgf%Elp_Radius = radius(I)
            qmmm_gaussian_fns(I)%pgf%Elp_Radius_corr = c_radius(I)
         END DO
         IF (ASSOCIATED(radius)) THEN
            DEALLOCATE (radius)
         END IF
         IF (ASSOCIATED(c_radius)) THEN
            DEALLOCATE (c_radius)
         END IF
         !
         IF (use_geep_lib) THEN
            IF (qmmm_coupl_type == do_qmmm_gauss) THEN
               CALL set_mm_potential_erf(qmmm_gaussian_fns, &
                                         compatibility, num_geep_gauss)
            ELSEIF (qmmm_coupl_type == do_qmmm_swave) THEN
               CALL set_mm_potential_swave(qmmm_gaussian_fns, &
                                           num_geep_gauss)
            END IF
         ELSE
            CALL read_mm_potential(para_env, qmmm_gaussian_fns, &
                                   (compatibility .AND. (qmmm_coupl_type == do_qmmm_gauss)), qmmm_section)
         END IF
         !
         CALL pw_env_get(pw_env, pw_pools=pools, gridlevel_info=gridlevel_info)
         ALLOCATE (maxradius(SIZE(pools)))
         maxradius = 0.0_dp
         DO J = 1, SIZE(qmmm_gaussian_fns)
            mypgf => qmmm_gaussian_fns(J)%pgf
            ALLOCATE (mypgf%grid_level(SIZE(mypgf%Ak)))
            mypgf%grid_level = 0
            mymaxradius = 0.0_dp
            DO I = 1, mypgf%number_of_gaussians
               IF (mypgf%Gk(I) /= 0.0_dp) THEN
                  alpha = 1.0_dp/mypgf%Gk(I)
                  alpha = alpha*alpha
                  ilevel = gaussian_gridlevel(gridlevel_info, alpha)
                  Prefactor = mypgf%Ak(I)*maxchrg
                  mymaxradius = exp_radius(0, alpha, eps_mm_rspace, Prefactor)
                  maxradius(ilevel) = MAX(maxradius(ilevel), mymaxradius)
                  mypgf%grid_level(i) = ilevel
               END IF
            END DO
         END DO
         !
         ! End of gaussian initialization...
      CASE DEFAULT
         output_unit = cp_print_key_unit_nr(logger, print_section, "PROGRAM_RUN_INFO", &
                                            extension=".qmmmLog")
         IF (output_unit > 0) WRITE (output_unit, '(A)') " QMMM Gaussian Data Not Initialized!"
         CALL cp_print_key_finished_output(output_unit, logger, print_section, "PROGRAM_RUN_INFO")
      END SELECT
   END SUBROUTINE qmmm_gaussian_initialize

END MODULE qmmm_gaussian_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qmmm_per_elpot

! **************************************************************************************************
!> \brief Setting up the potential for QM/MM periodic boundary conditions calculations
!> \par History
!>      07.2005 created [tlaino]
!> \author Teodoro Laino
   USE ao_util,                         ONLY: exp_radius
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE ewald_environment_types,         ONLY: ewald_env_create,&
                                              ewald_env_get,&
                                              ewald_env_set,&
                                              ewald_environment_type,&
                                              read_ewald_section
   USE ewald_pw_types,                  ONLY: ewald_pw_create,&
                                              ewald_pw_type
   USE ewald_spline_util,               ONLY: Setup_Ewald_Spline
   USE input_constants,                 ONLY: do_qmmm_coulomb,&
                                              do_qmmm_gauss,&
                                              do_qmmm_pcharge,&
                                              do_qmmm_swave
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type,&
                                              qmmm_gaussian_type
   USE qmmm_types_low,                  ONLY: qmmm_per_pot_p_type,&
                                              qmmm_per_pot_type,&
                                              qmmm_pot_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_per_elpot'
   PUBLIC :: qmmm_per_potential_init, qmmm_ewald_potential_init

CONTAINS

! **************************************************************************************************
!> \brief Initialize the QMMM potential stored on vector,
!>      according the qmmm_coupl_type
!> \param qmmm_coupl_type ...
!> \param per_potentials ...
!> \param potentials ...
!> \param pgfs ...
!> \param qm_cell_small ...
!> \param mm_cell ...
!> \param para_env ...
!> \param compatibility ...
!> \param qmmm_periodic ...
!> \param print_section ...
!> \param eps_mm_rspace ...
!> \param maxchrg ...
!> \param ncp ...
!> \param ncpl ...
!> \par History
!>      09.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_per_potential_init(qmmm_coupl_type, per_potentials, potentials, &
                                      pgfs, qm_cell_small, mm_cell, para_env, compatibility, qmmm_periodic, print_section, &
                                      eps_mm_rspace, maxchrg, ncp, ncpl)
      INTEGER, INTENT(IN)                                :: qmmm_coupl_type
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(cell_type), POINTER                           :: qm_cell_small, mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: compatibility
      TYPE(section_vals_type), POINTER                   :: qmmm_periodic, print_section
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace, maxchrg
      INTEGER, INTENT(IN)                                :: ncp(3), ncpl(3)

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_per_potential_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, idim, ig, ig_start, iw, ix, iy, iz, &
                                                            K, Kmax(3), n_rep_real(3), &
                                                            n_rep_real_val, ncoarsel, ncoarset, &
                                                            Ndim, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      REAL(KIND=dp)                                      :: Ak, alpha, box(3), Fac(3), fs, g, g2, &
                                                            Gk, Gmax, mymaxradius, npl, npt, &
                                                            Prefactor, rc, rc2, Rmax, tmp, vec(3), &
                                                            vol
      REAL(KIND=dp), DIMENSION(:), POINTER               :: gx, gy, gz, Lg
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qmmm_gaussian_type), POINTER                  :: pgf

      NULLIFY (Lg, gx, gy, gz)
      ncoarset = PRODUCT(ncp)
      ncoarsel = PRODUCT(ncpl)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      Rmax = SQRT(mm_cell%hmat(1, 1)**2+ &
                  mm_cell%hmat(2, 2)**2+ &
                  mm_cell%hmat(3, 3)**2)
      CALL section_vals_val_get(qmmm_periodic, "GMAX", r_val=Gmax)
      CALL section_vals_val_get(qmmm_periodic, "REPLICA", i_val=n_rep_real_val)
      fac = 2.0e0_dp*Pi/(/mm_cell%hmat(1, 1), mm_cell%hmat(2, 2), mm_cell%hmat(3, 3)/)
      Kmax = CEILING(Gmax/Fac)
      Vol = mm_cell%hmat(1, 1)* &
            mm_cell%hmat(2, 2)* &
            mm_cell%hmat(3, 3)
      Ndim = (Kmax(1)+1)*(2*Kmax(2)+1)*(2*Kmax(3)+1)
      ig_start = 1
      n_rep_real = n_rep_real_val
      IF (compatibility .AND. (qmmm_coupl_type == do_qmmm_gauss)) ig_start = 2

      IF(.NOT.(.NOT. ASSOCIATED(per_potentials)))CALL cp__a("qmmm_per_elpot.F",130)
      ALLOCATE (per_potentials(SIZE(pgfs)))
      IF(.NOT.(SIZE(pgfs) == SIZE(potentials)))CALL cp__a("qmmm_per_elpot.F",132)
      Potential_Type: DO K = 1, SIZE(pgfs)

         rc = pgfs(K)%pgf%Elp_Radius
         ALLOCATE (per_potentials(K)%Pot)
         SELECT CASE (qmmm_coupl_type)
         CASE (do_qmmm_coulomb, do_qmmm_pcharge)
            ! Not yet implemented for this case
            CALL cp__b("qmmm_per_elpot.F",140,"")
         CASE (do_qmmm_gauss, do_qmmm_swave)
            ALLOCATE (Lg(Ndim))
            ALLOCATE (gx(Ndim))
            ALLOCATE (gy(Ndim))
            ALLOCATE (gz(Ndim))
         END SELECT

         LG = 0.0_dp
         gx = 0.0_dp
         gy = 0.0_dp
         gz = 0.0_dp

         SELECT CASE (qmmm_coupl_type)
         CASE (do_qmmm_coulomb, do_qmmm_pcharge)
            ! Not yet implemented for this case
            CALL cp__b("qmmm_per_elpot.F",156,"")
         CASE (do_qmmm_gauss, do_qmmm_swave)
            pgf => pgfs(K)%pgf
            idim = 0
            DO ix = 0, kmax(1)
               DO iy = -kmax(2), kmax(2)
                  DO iz = -kmax(3), kmax(3)
                     idim = idim+1
                     IF (ix == 0 .AND. iy == 0 .AND. iz == 0) THEN
                        DO Ig = ig_start, pgf%number_of_gaussians
                           Gk = pgf%Gk(Ig)
                           Ak = pgf%Ak(Ig)*Pi**(3.0_dp/2.0_dp)*Gk**3.0_dp
                           LG(idim) = LG(idim)-Ak
                        END DO
                     ELSE
                        fs = 2.0_dp; IF (ix == 0) fs = 1.0_dp
                        vec = fac*(/REAL(ix, KIND=dp), REAL(iy, KIND=dp), REAL(iz, KIND=dp)/)
                        g2 = DOT_PRODUCT(vec, vec)
                        rc2 = rc*rc
                        g = SQRT(g2)
                        IF (qmmm_coupl_type == do_qmmm_gauss) THEN
                           LG(idim) = 4.0_dp*Pi/g2*EXP(-(g2*rc2)/4.0_dp)
                        ELSEIF (qmmm_coupl_type == do_qmmm_swave) THEN
                           tmp = 4.0_dp/rc2
                           LG(idim) = 4.0_dp*Pi*tmp**2/(g2*(g2+tmp)**2)
                        END IF
                        DO Ig = ig_start, pgf%number_of_gaussians
                           Gk = pgf%Gk(Ig)
                           Ak = pgf%Ak(Ig)*Pi**(3.0_dp/2.0_dp)*Gk**3.0_dp
                           LG(idim) = LG(idim)-Ak*EXP(-(g*Gk)**2.0_dp/4.0_dp)
                        END DO
                     ENDIF
                     LG(idim) = fs*LG(idim)*1.0_dp/Vol
                     gx(idim) = fac(1)*REAL(ix, KIND=dp)
                     gy(idim) = fac(2)*REAL(iy, KIND=dp)
                     gz(idim) = fac(3)*REAL(iz, KIND=dp)
                  END DO
               END DO
            END DO

            IF (ALL(n_rep_real == -1)) THEN
               mymaxradius = 0.0_dp
               DO I = 1, pgf%number_of_gaussians
                  IF (pgf%Gk(I) /= 0.0_dp) THEN
                     alpha = 1.0_dp/pgf%Gk(I)
                     alpha = alpha*alpha
                     Prefactor = pgf%Ak(I)*maxchrg
                     mymaxradius = MAX(mymaxradius, exp_radius(0, alpha, eps_mm_rspace, Prefactor))
                  END IF
               END DO
               box(1) = (qm_cell_small%hmat(1, 1)-mm_cell%hmat(1, 1))/2.0_dp
               box(2) = (qm_cell_small%hmat(2, 2)-mm_cell%hmat(2, 2))/2.0_dp
               box(3) = (qm_cell_small%hmat(3, 3)-mm_cell%hmat(3, 3))/2.0_dp
               IF (ANY(box > 0.0_dp)) THEN
                  CALL cp__b("qmmm_per_elpot.F",210,"")
               END IF
               n_rep_real(1) = CEILING((box(1)+mymaxradius)/mm_cell%hmat(1, 1))
               n_rep_real(2) = CEILING((box(2)+mymaxradius)/mm_cell%hmat(2, 2))
               n_rep_real(3) = CEILING((box(3)+mymaxradius)/mm_cell%hmat(3, 3))
            END IF

         CASE DEFAULT
            DEALLOCATE (per_potentials(K)%Pot)
            IF (output_unit > 0) WRITE (output_unit, '(A)') " QMMM Periodic Potential - not Initialized!"
            CYCLE Potential_Type
         END SELECT

         NULLIFY (mm_atom_index)
         ALLOCATE (mm_atom_index(SIZE(potentials(K)%pot%mm_atom_index)))
         mm_atom_index = potentials(K)%pot%mm_atom_index

         NULLIFY (per_potentials(K)%Pot%LG, per_potentials(K)%Pot%mm_atom_index, &
                  per_potentials(K)%Pot%gx, per_potentials(K)%Pot%gy, per_potentials(K)%Pot%gz)
         CALL qmmm_per_pot_type_create(per_potentials(K)%Pot, LG=LG, gx=gx, gy=gy, gz=gz, &
                                       Gmax=Gmax, Kmax=Kmax, n_rep_real=n_rep_real, &
                                       Fac=Fac, mm_atom_index=mm_atom_index, &
                                       mm_cell=mm_cell, para_env=para_env, &
                                       qmmm_per_section=qmmm_periodic, print_section=print_section)

         iw = cp_print_key_unit_nr(logger, print_section, "PERIODIC_INFO", &
                                   extension=".log")
         IF (iw > 0) THEN
            npt = REAL(ncoarset, KIND=dp)*REAL(ndim, KIND=dp)*REAL(SIZE(mm_atom_index), KIND=dp)
            npl = REAL(ncoarsel, KIND=dp)*REAL(ndim, KIND=dp)*REAL(SIZE(mm_atom_index), KIND=dp)
            WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
            WRITE (UNIT=iw, FMT="(T2,A,T20,A,T80,A)") "-", "QMMM PERIODIC BOUNDARY CONDITION INFO", "-"
            WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
            WRITE (UNIT=iw, FMT="(T2,A,T10,A,F15.6,T50,A,3I5,T80,A)") "-", "RADIUS  =", rc, "REPLICA =", n_rep_real, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T10,A,F15.6,T50,A,I15,T80,A)") "-", "MINGVAL =", MINVAL(ABS(Lg)), &
               "GPOINTS =", ndim, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T10,A,3I5,T50,A,3I5,T80,A)") "-", "NCOARSL =", ncpl, &
               "NCOARST =", ncp, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T10,A,F15.0,T50,A,F15.0,T80,A)") "-", "NFLOP-L ~", npl, &
               "NFLOP-T ~", npt, "-"
            WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
         END IF
         CALL cp_print_key_finished_output(iw, logger, print_section, &
                                           "PERIODIC_INFO")

      END DO Potential_Type

   END SUBROUTINE qmmm_per_potential_init

! **************************************************************************************************
!> \brief Creates the qmmm_pot_type structure
!> \param Pot ...
!> \param LG ...
!> \param gx ...
!> \param gy ...
!> \param gz ...
!> \param GMax ...
!> \param Kmax ...
!> \param n_rep_real ...
!> \param Fac ...
!> \param mm_atom_index ...
!> \param mm_cell ...
!> \param para_env ...
!> \param qmmm_per_section ...
!> \param print_section ...
!> \par History
!>      09.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_per_pot_type_create(Pot, LG, gx, gy, gz, GMax, Kmax, n_rep_real, &
                                       Fac, mm_atom_index, mm_cell, para_env, qmmm_per_section, print_section)
      TYPE(qmmm_per_pot_type), POINTER                   :: Pot
      REAL(KIND=dp), DIMENSION(:), POINTER               :: LG, gx, gy, gz
      REAL(KIND=dp), INTENT(IN)                          :: Gmax
      INTEGER, INTENT(IN)                                :: Kmax(3), n_rep_real(3)
      REAL(KIND=dp), INTENT(IN)                          :: Fac(3)
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: qmmm_per_section, print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_per_pot_type_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: npts(3)
      INTEGER, DIMENSION(:), POINTER                     :: ngrids
      REAL(KIND=dp)                                      :: hmat(3, 3)
      TYPE(section_vals_type), POINTER                   :: grid_print_section

      Pot%LG => LG
      Pot%gx => gx
      Pot%gy => gy
      Pot%gz => gz
      Pot%mm_atom_index => mm_atom_index
      Pot%Gmax = Gmax
      Pot%Kmax = Kmax
      Pot%n_rep_real = n_rep_real
      Pot%Fac = Fac
      !
      ! Setting Up Fit Procedure
      !
      NULLIFY (Pot%pw_grid)
      NULLIFY (Pot%pw_pool)
      NULLIFY (Pot%TabLR, ngrids)
      CALL section_vals_val_get(qmmm_per_section, "ngrids", i_vals=ngrids)
      npts = ngrids
      hmat = mm_cell%hmat

      grid_print_section => section_vals_get_subs_vals(print_section, "GRID_INFORMATION")
      CALL Setup_Ewald_Spline(pw_grid=Pot%pw_grid, pw_pool=Pot%pw_pool, coeff=Pot%TabLR, &
                              LG=LG, gx=gx, gy=gy, gz=gz, hmat=hmat, npts=npts, param_section=qmmm_per_section, &
                              tag="qmmm", para_env=para_env, print_section=grid_print_section)

   END SUBROUTINE qmmm_per_pot_type_create

! **************************************************************************************************
!> \brief Initialize the QMMM Ewald potential needed for QM-QM Coupling using
!>      point charges
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param qmmm_coupl_type ...
!> \param mm_cell ...
!> \param para_env ...
!> \param qmmm_periodic ...
!> \param print_section ...
!> \par History
!>      10.2014 created [JGH]
!> \author JGH
! **************************************************************************************************
   SUBROUTINE qmmm_ewald_potential_init(ewald_env, ewald_pw, qmmm_coupl_type, mm_cell, para_env, &
                                        qmmm_periodic, print_section)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      INTEGER, INTENT(IN)                                :: qmmm_coupl_type
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: qmmm_periodic, print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_ewald_potential_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, gmax(3), iw, o_spline, ounit
      LOGICAL                                            :: do_multipoles
      REAL(KIND=dp)                                      :: alpha, rcut
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: ewald_print_section, ewald_section, &
                                                            poisson_section

      logger => cp_get_default_logger()
      ounit = cp_logger_get_default_io_unit(logger)
      IF(.NOT.(.NOT. ASSOCIATED(ewald_env)))CALL cp__a("qmmm_per_elpot.F",360)
      IF(.NOT.(.NOT. ASSOCIATED(ewald_pw)))CALL cp__a("qmmm_per_elpot.F",361)

      ! Create Ewald environments
      poisson_section => section_vals_get_subs_vals(qmmm_periodic, "POISSON")
      CALL ewald_env_create(ewald_env, para_env)
      CALL ewald_env_set(ewald_env, poisson_section=poisson_section)
      ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD")
      CALL read_ewald_section(ewald_env, ewald_section)
      ewald_print_section => section_vals_get_subs_vals(print_section, "GRID_INFORMATION")
      CALL ewald_pw_create(ewald_pw, ewald_env, mm_cell, mm_cell, print_section=ewald_print_section)

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, do_multipoles=do_multipoles, &
                         gmax=gmax, o_spline=o_spline, alpha=alpha, rcut=rcut)
      IF (do_multipoles) &
         CALL cp__b("qmmm_per_elpot.F",375,"No multipole force fields allowed in QM-QM Ewald long range correction")

      SELECT CASE (qmmm_coupl_type)
      CASE (do_qmmm_coulomb)
         CALL cp__b("qmmm_per_elpot.F",379,"QM-QM long range correction not possible with COULOMB coupling")
      CASE (do_qmmm_pcharge)
         ! OK
      CASE (do_qmmm_gauss, do_qmmm_swave)
         CALL cp__b("qmmm_per_elpot.F",383,"QM-QM long range correction not possible with GAUSS/SWAVE coupling")
      CASE DEFAULT
         ! We should never get to this point
         CALL cp__b("qmmm_per_elpot.F",386,"")
      END SELECT

      iw = cp_print_key_unit_nr(logger, print_section, "PERIODIC_INFO", extension=".log")
      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=iw, FMT="(T2,A,T20,A,T80,A)") "-", "QMMM PERIODIC BOUNDARY CONDITION INFO", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T25,A,T80,A)") "-", "QM-QM Long Range Correction", "-"
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
         SELECT CASE (ewald_type)
         CASE (do_ewald_none)
            CALL cp__b("qmmm_per_elpot.F",397,"QM-QM long range correction not compatible with Ewald=NONE")
         CASE (do_ewald_pme)
            CALL cp__b("qmmm_per_elpot.F",399,"QM-QM long range correction not possible with Ewald=PME")
         CASE (do_ewald_ewald)
            CALL cp__b("qmmm_per_elpot.F",401,"QM-QM long range correction not possible with Ewald method")
         CASE (do_ewald_spme)
            WRITE (UNIT=iw, FMT="(T2,A,T35,A,T75,A,T80,A)") "-", "Ewald type", "SPME", "-"
            WRITE (UNIT=iw, FMT="(T2,A,T35,A,T61,3I6,T80,A)") "-", "GMAX values", gmax, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T35,A,T73,I6,T80,A)") "-", "Spline order", o_spline, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T35,A,T67,F12.4,T80,A)") "-", "Alpha value", alpha, "-"
            WRITE (UNIT=iw, FMT="(T2,A,T35,A,T67,F12.4,T80,A)") "-", "Real space cutoff value", rcut, "-"
         END SELECT
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
      END IF
      CALL cp_print_key_finished_output(iw, logger, print_section, "PERIODIC_INFO")

   END SUBROUTINE qmmm_ewald_potential_init

! **************************************************************************************************

END MODULE qmmm_per_elpot

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      09.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_elpot
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE erf_fn,                          ONLY: erf
   USE input_constants,                 ONLY: do_qmmm_coulomb,&
                                              do_qmmm_gauss,&
                                              do_qmmm_pcharge,&
                                              do_qmmm_swave
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: rootpi
   USE memory_utilities,                ONLY: reallocate
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type,&
                                              qmmm_gaussian_type
   USE qmmm_types_low,                  ONLY: qmmm_Pot_Type,&
                                              qmmm_pot_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_elpot'
   PUBLIC :: qmmm_potential_init

CONTAINS

! **************************************************************************************************
!> \brief Initialize the QMMM potential stored on vector,
!>      according the qmmm_coupl_type
!> \param qmmm_coupl_type ...
!> \param mm_el_pot_radius ...
!> \param potentials ...
!> \param pgfs ...
!> \param mm_cell ...
!> \param compatibility ...
!> \param print_section ...
!> \par History
!>      09.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_potential_init(qmmm_coupl_type, mm_el_pot_radius, potentials, &
                                  pgfs, mm_cell, compatibility, print_section)
      INTEGER, INTENT(IN)                                :: qmmm_coupl_type
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_el_pot_radius
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(cell_type), POINTER                           :: mm_cell
      LOGICAL, INTENT(IN)                                :: compatibility
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_potential_init', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: dx = 0.05_dp

      CHARACTER(LEN=default_path_length)                 :: myFormat
      CHARACTER(LEN=default_string_length)               :: rc_s
      INTEGER                                            :: I, ig, ig_start, J, K, myind, ndim, Np, &
                                                            output_unit, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: A, G, rc, Rmax, Rmin, t, t1, t2, x
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radius
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pot0_2
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qmmm_gaussian_type), POINTER                  :: pgf

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      Rmin = 0.0_dp
      Rmax = SQRT(mm_cell%hmat(1, 1)**2+ &
                  mm_cell%hmat(2, 2)**2+ &
                  mm_cell%hmat(3, 3)**2)
      np = CEILING(rmax/dx)+1
      !
      ! Preprocessing
      !
      IF (SIZE(mm_el_pot_radius) /= 0) THEN
         ALLOCATE (radius(1))
         radius(1) = mm_el_pot_radius(1)
      ELSE
         ALLOCATE (radius(0))
      END IF
      Loop_on_all_values: DO I = 2, SIZE(mm_el_pot_radius)
         Found = .FALSE.
         Loop_on_found_values: DO J = 1, SIZE(radius)
            IF (mm_el_pot_radius(i) .EQ. radius(j)) THEN
               Found = .TRUE.
               EXIT Loop_on_found_values
            END IF
         END DO Loop_on_found_values
         IF (.NOT. Found) THEN
            Ndim = SIZE(radius)
            Ndim = Ndim+1
            CALL REALLOCATE(radius, 1, Ndim)
            radius(Ndim) = mm_el_pot_radius(i)
         END IF
      END DO Loop_on_all_values
      !
      IF(.NOT.(.NOT. ASSOCIATED(potentials)))CALL cp__a("qmmm_elpot.F",118)
      ALLOCATE (potentials(SIZE(radius)))

      Potential_Type: DO K = 1, SIZE(radius)

         rc = radius(K)
         ALLOCATE (potentials(K)%Pot)
         SELECT CASE (qmmm_coupl_type)
         CASE (do_qmmm_coulomb)
            NULLIFY (pot0_2)
         CASE (do_qmmm_pcharge)
            NULLIFY (pot0_2)
         CASE (do_qmmm_gauss, do_qmmm_swave)
            ALLOCATE (pot0_2(2, np))
         END SELECT

         SELECT CASE (qmmm_coupl_type)
         CASE (do_qmmm_coulomb, do_qmmm_pcharge)
            ! Do Nothing
         CASE (do_qmmm_gauss, do_qmmm_swave)
            IF (qmmm_coupl_type == do_qmmm_gauss) THEN
               ! Smooth Coulomb Potential ::  Erf(x/rc)/x
               pot0_2(1, 1) = 2.0_dp/(rootpi*rc)
               pot0_2(2, 1) = 0.0_dp
               x = 0.0_dp
               DO i = 2, np
                  x = x+dx
                  pot0_2(1, i) = erf(x/rc)/x
                  t = 2._dp/(rootpi*x*rc)*EXP(-(x/rc)**2)
                  pot0_2(2, i) = (t-pot0_2(1, i)/x)*dx
               END DO
            ELSEIF (qmmm_coupl_type == do_qmmm_swave) THEN
               ! S-wave expansion :: 1/x - exp(-2*x/rc) * ( 1/x - 1/rc )
               pot0_2(1, 1) = 1.0_dp/rc
               pot0_2(2, 1) = 0.0_dp
               x = 0.0_dp
               DO i = 2, np
                  x = x+dx
                  t = EXP(-2.0_dp*x/rc)/rc
                  pot0_2(1, i) = (1.0_dp-t*(rc+x))/x
                  pot0_2(2, i) = ((t*(rc**2+2.0_dp*rc*x+2.0_dp*x**2)/rc-1.0_dp)/x**2)*dx
               END DO
            END IF
            pgf => pgfs(K)%pgf
            IF(.NOT.(pgf%Elp_Radius == rc))CALL cp__a("qmmm_elpot.F",162)
            ig_start = 1
            IF (compatibility .AND. (qmmm_coupl_type == do_qmmm_gauss)) ig_start = 2
            DO Ig = ig_start, pgf%number_of_gaussians
               A = pgf%Ak(Ig)
               G = pgf%Gk(Ig)
               pot0_2(1, 1) = pot0_2(1, 1)-A
               x = 0.0_dp
               DO i = 2, np
                  x = x+dx
                  t = EXP(-(x/G)**2)*A
                  t1 = 1/G**2
                  t2 = t1*t
                  pot0_2(1, i) = pot0_2(1, i)-t
                  pot0_2(2, i) = pot0_2(2, i)+2.0_dp*x*t2*dx
               END DO
            END DO

            ! Print info on the unidimensional MM electrostatic potential
            IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, "MM_POTENTIAL") &
                      , cp_p_file)) THEN
               WRITE (rc_s, '(F6.3)') rc
               unit_nr = cp_print_key_unit_nr(logger, print_section, "MM_POTENTIAL", &
                                              extension="_rc="//TRIM(ADJUSTL(rc_s))//".data")
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, '(A)') "# MM ELECTROSTATIC POTENTIAL - UNIDIMENSIONAL - ATOMIC UNITS"
                  WRITE (unit_nr, '(A,I5)') "# MM ELECTROSTATIC POTENTIAL - Nr. of Gaussians:", pgf%number_of_gaussians
                  WRITE (unit_nr, '(A,T10,A,T30,A,T300,A)') "#", "Xval", "Gaussians", "LongRange"
                  myFormat = "T10,F15.9,T30,"
                  DO Ig = 1, pgf%number_of_gaussians
                     myind = INDEX(myFormat, " ")
                     WRITE (myFormat(myind:), '(A6)') "F12.9,"
                  END DO
                  myind = INDEX(myFormat, " ")-1
                  myFormat = myFormat(1:myind)//"T300,F15.9"
                  myind = INDEX(myFormat, " ")-1
                  x = 0.0_dp
                  DO i = 1, np
                     WRITE (unit_nr, '('//myFormat(1:myind)//')') &
                        x, (EXP(-(x/pgf%Gk(Ig))**2)*pgf%Ak(Ig), Ig=1, pgf%number_of_gaussians), pot0_2(1, i)
                     x = x+dx
                  END DO
               END IF
               CALL cp_print_key_finished_output(unit_nr, logger, print_section, &
                                                 "MM_POTENTIAL")
            END IF
         CASE DEFAULT
            DEALLOCATE (potentials(K)%Pot)
            IF (output_unit > 0) WRITE (output_unit, '(A)') " QMMM Potential - Spline Interpolation - not Initialized!"
            CYCLE Potential_Type
         END SELECT
         NULLIFY (mm_atom_index)
         ALLOCATE (mm_atom_index(1))
         ! Build mm_atom_index List
         DO J = 1, SIZE(mm_el_pot_radius)
            IF (rc .EQ. mm_el_pot_radius(J)) THEN
               Ndim = SIZE(mm_atom_index)
               mm_atom_index(Ndim) = J
               CALL reallocate(mm_atom_index, 1, Ndim+1)
            ENDIF
         END DO
         CALL reallocate(mm_atom_index, 1, Ndim)

         NULLIFY (potentials(K)%Pot%pot0_2)
         CALL qmmm_pot_type_create(potentials(K)%Pot, pot0_2=pot0_2, &
                                   Rmax=Rmax, Rmin=Rmin, dx=dx, Rc=rc, npts=np, &
                                   mm_atom_index=mm_atom_index)

      END DO Potential_Type
      DEALLOCATE (radius)
   END SUBROUTINE qmmm_potential_init

! **************************************************************************************************
!> \brief Creates the qmmm_pot_type structure
!> \param Pot ...
!> \param pot0_2 ...
!> \param Rmax ...
!> \param Rmin ...
!> \param dx ...
!> \param npts ...
!> \param rc ...
!> \param mm_atom_index ...
!> \par History
!>      09.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_pot_type_create(Pot, pot0_2, Rmax, Rmin, dx, npts, rc, &
                                   mm_atom_index)
      TYPE(qmmm_Pot_Type), POINTER                       :: Pot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pot0_2
      REAL(KIND=dp), INTENT(IN)                          :: Rmax, Rmin, dx
      INTEGER, INTENT(IN)                                :: npts
      REAL(KIND=dp), INTENT(IN)                          :: Rc
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index

      Pot%pot0_2 => pot0_2
      Pot%mm_atom_index => mm_atom_index
      Pot%Rmax = Rmax
      Pot%Rmin = Rmin
      Pot%Rc = rc
      Pot%dx = dx
      Pot%npts = npts

   END SUBROUTINE qmmm_pot_type_create

END MODULE qmmm_elpot
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize a QM/MM calculation
!> \par History
!>      5.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qmmm_init
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              set_atomic_kind
   USE cell_methods,                    ONLY: read_cell
   USE cell_types,                      ONLY: cell_type,&
                                              use_perd_xyz
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_eri_mme_interface,            ONLY: eri_mme_set_params
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE eri_mme_types,                   ONLY: eri_mme_init_read_input
   USE external_potential_types,        ONLY: fist_potential_type,&
                                              get_potential,&
                                              set_potential
   USE force_field_types,               ONLY: input_info_type
   USE force_fields_input,              ONLY: read_gd_section,&
                                              read_gp_section,&
                                              read_lj_section,&
                                              read_wl_section
   USE input_constants,                 ONLY: &
        RADIUS_QMMM_DEFAULT, calc_always, calc_once, do_eri_mme, do_qmmm_gauss, &
        do_qmmm_image_calcmatrix, do_qmmm_image_iter, do_qmmm_link_gho, do_qmmm_link_imomm, &
        do_qmmm_link_pseudo, do_qmmm_pcharge, do_qmmm_swave
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE pair_potential_types,            ONLY: pair_potential_reallocate
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_type
   USE qmmm_elpot,                      ONLY: qmmm_potential_init
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qmmm_gaussian_init,              ONLY: qmmm_gaussian_initialize
   USE qmmm_per_elpot,                  ONLY: qmmm_ewald_potential_init,&
                                              qmmm_per_potential_init
   USE qmmm_types_low,                  ONLY: add_set_type,&
                                              add_shell_type,&
                                              create_add_set_type,&
                                              create_add_shell_type,&
                                              qmmm_env_mm_type,&
                                              qmmm_env_qm_type,&
                                              qmmm_links_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE shell_potential_types,           ONLY: get_shell,&
                                              shell_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_init'

   PUBLIC :: assign_mm_charges_and_radius, &
             print_qmmm_charges, &
             print_qmmm_links, &
             print_image_charge_info, &
             qmmm_init_gaussian_type, &
             qmmm_init_potential, &
             qmmm_init_periodic_potential, &
             setup_qmmm_vars_qm, &
             setup_qmmm_vars_mm, &
             setup_qmmm_links, &
             move_or_add_atoms, &
             setup_origin_mm_cell

CONTAINS

! **************************************************************************************************
!> \brief Assigns charges and radius to evaluate the MM electrostatic potential
!> \param subsys the subsys containing the MM charges
!> \param charges ...
!> \param mm_atom_chrg ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param mm_atom_index ...
!> \param mm_link_atoms ...
!> \param mm_link_scale_factor ...
!> \param added_shells ...
!> \param shell_model ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE assign_mm_charges_and_radius(subsys, charges, mm_atom_chrg, mm_el_pot_radius, &
                                           mm_el_pot_radius_corr, mm_atom_index, mm_link_atoms, &
                                           mm_link_scale_factor, added_shells, shell_model)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      REAL(dp), DIMENSION(:), POINTER                    :: mm_atom_chrg, mm_el_pot_radius, &
                                                            mm_el_pot_radius_corr
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index, mm_link_atoms
      REAL(dp), DIMENSION(:), POINTER                    :: mm_link_scale_factor
      TYPE(add_shell_type), OPTIONAL, POINTER            :: added_shells
      LOGICAL                                            :: shell_model

      CHARACTER(LEN=*), PARAMETER :: routineN = 'assign_mm_charges_and_radius', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, ilink, IndMM, IndShell, ishell
      LOGICAL                                            :: is_shell
      REAL(dp)                                           :: qcore, qi, qshell, rc, ri
      TYPE(atomic_kind_type), POINTER                    :: my_kind
      TYPE(fist_potential_type), POINTER                 :: my_potential
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_set, particle_set, shell_set
      TYPE(shell_kind_type), POINTER                     :: shell_kind

      NULLIFY (particle_set, my_kind, added_shells)
      CALL cp_subsys_get(subsys=subsys, particles=particles, core_particles=core_particles, &
                         shell_particles=shell_particles)
      particle_set => particles%els

      IF (ALL(particle_set(:)%shell_index .EQ. 0)) THEN
         shell_model = .FALSE.
         CALL create_add_shell_type(added_shells, ndim=0)
      ELSE
         shell_model = .TRUE.
      END IF

      IF (shell_model) THEN
         shell_set => shell_particles%els
         core_set => core_particles%els
         ishell = SIZE(shell_set)
         CALL create_add_shell_type(added_shells, ndim=ishell)
         added_shells%added_particles => shell_set
         added_shells%added_cores => core_set
      END IF

      DO I = 1, SIZE(mm_atom_index)
         IndMM = mm_atom_index(I)
         my_kind => particle_set(IndMM)%atomic_kind
         CALL get_atomic_kind(atomic_kind=my_kind, fist_potential=my_potential, &
                              shell_active=is_shell, shell=shell_kind)
         CALL get_potential(potential=my_potential, &
                            qeff=qi, &
                            qmmm_radius=ri, &
                            qmmm_corr_radius=rc)
         IF (ASSOCIATED(charges)) qi = charges(IndMM)
         mm_atom_chrg(I) = qi
         mm_el_pot_radius(I) = ri
         mm_el_pot_radius_corr(I) = rc
         IF (is_shell) THEN
            IndShell = particle_set(IndMM)%shell_index
            IF (ASSOCIATED(shell_kind)) THEN
               CALL get_shell(shell=shell_kind, &
                              charge_core=qcore, &
                              charge_shell=qshell)
               mm_atom_chrg(I) = qcore
            END IF
            added_shells%mm_core_index(IndShell) = IndMM
            added_shells%mm_core_chrg(IndShell) = qshell
            added_shells%mm_el_pot_radius(Indshell) = ri*1.0_dp
            added_shells%mm_el_pot_radius_corr(Indshell) = rc*1.0_dp
         END IF
      END DO

      IF (ASSOCIATED(mm_link_atoms)) THEN
         DO ilink = 1, SIZE(mm_link_atoms)
            DO i = 1, SIZE(mm_atom_index)
               IF (mm_atom_index(i) == mm_link_atoms(ilink)) EXIT
            END DO
            IndMM = mm_atom_index(I)
            mm_atom_chrg(i) = mm_atom_chrg(i)*mm_link_scale_factor(ilink)
         END DO
      END IF

   END SUBROUTINE assign_mm_charges_and_radius

! **************************************************************************************************
!> \brief Print info on charges generating the qmmm potential..
!> \param mm_atom_index ...
!> \param mm_atom_chrg ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param added_charges ...
!> \param added_shells ...
!> \param qmmm_section ...
!> \param nocompatibility ...
!> \param shell_model ...
!> \par History
!>      01.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE print_qmmm_charges(mm_atom_index, mm_atom_chrg, mm_el_pot_radius, mm_el_pot_radius_corr, &
                                 added_charges, added_shells, qmmm_section, nocompatibility, shell_model)
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      REAL(dp), DIMENSION(:), POINTER                    :: mm_atom_chrg, mm_el_pot_radius, &
                                                            mm_el_pot_radius_corr
      TYPE(add_set_type), POINTER                        :: added_charges
      TYPE(add_shell_type), POINTER                      :: added_shells
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      LOGICAL, INTENT(IN)                                :: nocompatibility, shell_model

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_qmmm_charges', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, ind1, ind2, IndMM, iw
      REAL(KIND=dp)                                      :: qi, qtot, rc, ri
      TYPE(cp_logger_type), POINTER                      :: logger

      qtot = 0.0_dp
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, qmmm_section, "PRINT%QMMM_CHARGES", &
                                extension=".log")
      IF (iw > 0) THEN
         WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (iw, FMT='(/5X,A)') "MM    POINT CHARGES GENERATING THE QM/MM ELECTROSTATIC POTENTIAL"
         WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 79)
         DO I = 1, SIZE(mm_atom_index)
            IndMM = mm_atom_index(I)
            qi = mm_atom_chrg(I)
            qtot = qtot+qi
            ri = mm_el_pot_radius(I)
            rc = mm_el_pot_radius_corr(I)
            IF (nocompatibility) THEN
               WRITE (iw, '(5X,A9,T15,I5,T28,A8,T38,F12.6,T60,A8,T69,F12.6)') ' MM ATOM:', IndMM, ' RADIUS:', ri, &
                  ' CHARGE:', qi
            ELSE
               WRITE (iw, '(5X,A9,T15,I5,T28,A8,T38,F12.6,T60,A8,T69,F12.6,/,T56,A12,T69,F12.6)') &
                  ' MM ATOM:', IndMM, ' RADIUS:', ri, ' CHARGE:', qi, 'CORR. RADIUS', rc
            END IF
         END DO
         IF (added_charges%num_mm_atoms /= 0) THEN
            WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 79)
            WRITE (iw, '(/5X,A)') "ADDED POINT CHARGES GENERATING THE QM/MM ELECTROSTATIC POTENTIAL"
            WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 79)
            DO I = 1, SIZE(added_charges%mm_atom_index)
               IndMM = added_charges%mm_atom_index(I)
               qi = added_charges%mm_atom_chrg(I)
               qtot = qtot+qi
               ri = added_charges%mm_el_pot_radius(I)
               ind1 = added_charges%add_env(I)%Index1
               ind2 = added_charges%add_env(I)%Index2
               IF (nocompatibility) THEN
                  WRITE (iw, '(5X,A9,I5,T25,A8,T35,F12.6,T50,A8,T59,F12.6,I5,I5)') 'MM POINT:', IndMM, ' RADIUS:', ri, &
                     ' CHARGE:', qi, ind1, ind2
               ELSE
                  WRITE (iw, '(5X,A9,I5,T25,A8,T35,F12.6,T50,A8,T59,F12.6,I5,I5,/,T56,A12,T69,F12.6)') &
                     'MM POINT:', IndMM, ' RADIUS:', ri, ' CHARGE:', qi, ind1, ind2, 'CORR. RADIUS', rc
               END IF
            END DO

         END IF

         IF (shell_model) THEN
            WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 73)
            WRITE (iw, '(/5X,A)') "ADDED SHELL CHARGES GENERATING THE QM/MM ELECTROSTATIC POTENTIAL"
            WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 73)

            DO I = 1, SIZE(added_shells%mm_core_index)
               IndMM = added_shells%mm_core_index(I)
               qi = added_shells%mm_core_chrg(I)
               qtot = qtot+qi
               ri = added_shells%mm_el_pot_radius(I)
               IF (nocompatibility) THEN
                  WRITE (iw, '(7X,A,I5,A8,F12.6,A8,F12.6,3F12.6)') 'SHELL:', IndMM, ' RADIUS:', ri, &
                     ' CHARGE:', qi, added_shells%added_particles(I)%r
               ELSE
                  WRITE (iw, '(7X,A,I5,A8,F12.6,A8,F12.6,A,F12.6)') 'SHELL:', IndMM, ' RADIUS:', ri, &
                     ' CHARGE:', qi, ' CORR. RADIUS', rc
               END IF

            END DO

         END IF

         WRITE (iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (iw, '(/,T50,A,T69,F12.6)') ' TOTAL CHARGE:', qtot
         WRITE (iw, FMT="(/,T2,A,/)") REPEAT("-", 79)
      END IF
      CALL cp_print_key_finished_output(iw, logger, qmmm_section, &
                                        "PRINT%QMMM_CHARGES")
   END SUBROUTINE print_qmmm_charges

! **************************************************************************************************
!> \brief Print info on qm/mm links
!> \param qmmm_section ...
!> \param qmmm_links ...
!> \par History
!>      01.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE print_qmmm_links(qmmm_section, qmmm_links)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_qmmm_links', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iw, mm_index, qm_index
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, qmmm_section, "PRINT%qmmm_link_info", extension=".log")
      IF (iw > 0) THEN
         IF (ASSOCIATED(qmmm_links)) THEN
            WRITE (iw, FMT="(/,T2, A)") REPEAT("-", 73)
            WRITE (iw, FMT="(/,T31,A)") " QM/MM LINKS "
            WRITE (iw, FMT="(/,T2, A)") REPEAT("-", 73)
            IF (ASSOCIATED(qmmm_links%imomm)) THEN
               WRITE (iw, FMT="(/,T31,A)") " IMOMM TYPE LINK "
               DO I = 1, SIZE(qmmm_links%imomm)
                  qm_index = qmmm_links%imomm(I)%link%qm_index
                  mm_index = qmmm_links%imomm(I)%link%mm_index
                  alpha = qmmm_links%imomm(I)%link%alpha
                  WRITE (iw, FMT="(T2,A,T20,A9,I8,1X,A9,I8,T62,A6,F12.6)") "TYPE: IMOMM", &
                     "QM INDEX:", qm_index, "MM INDEX:", mm_index, "ALPHA:", alpha
               END DO
            END IF
            IF (ASSOCIATED(qmmm_links%pseudo)) THEN
               WRITE (iw, FMT="(/,T31,A)") " PSEUDO TYPE LINK "
               DO I = 1, SIZE(qmmm_links%pseudo)
                  qm_index = qmmm_links%pseudo(I)%link%qm_index
                  mm_index = qmmm_links%pseudo(I)%link%mm_index
                  WRITE (iw, FMT="(T2,A,T20,A9,I8,1X,A9,I8)") "TYPE: PSEUDO", &
                     "QM INDEX:", qm_index, "MM INDEX:", mm_index
               END DO
            END IF
            WRITE (iw, FMT="(/,T2,A,/)") REPEAT("-", 73)
         ELSE
            WRITE (iw, FMT="(/,T2, A)") REPEAT("-", 73)
            WRITE (iw, FMT="(/,T26,A)") " NO QM/MM LINKS DETECTED"
            WRITE (iw, FMT="(/,T2, A)") REPEAT("-", 73)
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, qmmm_section, &
                                        "PRINT%qmmm_link_info")
   END SUBROUTINE print_qmmm_links

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env_qm ...
!> \param para_env ...
!> \param mm_atom_chrg ...
!> \param qs_env ...
!> \param added_charges ...
!> \param added_shells ...
!> \param print_section ...
!> \param qmmm_section ...
!> \par History
!>      1.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_init_gaussian_type(qmmm_env_qm, para_env, &
                                      mm_atom_chrg, qs_env, added_charges, added_shells, &
                                      print_section, qmmm_section)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env_qm
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_atom_chrg
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(add_set_type), POINTER                        :: added_charges
      TYPE(add_shell_type), POINTER                      :: added_shells
      TYPE(section_vals_type), POINTER                   :: print_section, qmmm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_init_gaussian_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: maxchrg
      REAL(KIND=dp), DIMENSION(:), POINTER               :: maxradius, maxradius2
      TYPE(pw_env_type), POINTER                         :: pw_env

      NULLIFY (maxradius, maxradius2, pw_env)

      maxchrg = MAXVAL(ABS(mm_atom_chrg(:)))
      CALL get_qs_env(qs_env, pw_env=pw_env)
      IF (qmmm_env_qm%add_mm_charges) maxchrg = MAX(maxchrg, MAXVAL(ABS(added_charges%mm_atom_chrg(:))))
      CALL qmmm_gaussian_initialize(qmmm_gaussian_fns=qmmm_env_qm%pgfs, &
                                    para_env=para_env, &
                                    pw_env=pw_env, &
                                    mm_el_pot_radius=qmmm_env_qm%mm_el_pot_radius, &
                                    mm_el_pot_radius_corr=qmmm_env_qm%mm_el_pot_radius_corr, &
                                    qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                    eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                    maxradius=maxradius, &
                                    maxchrg=maxchrg, &
                                    compatibility=qmmm_env_qm%compatibility, &
                                    print_section=print_section, &
                                    qmmm_section=qmmm_section)

      IF (qmmm_env_qm%move_mm_charges .OR. qmmm_env_qm%add_mm_charges) THEN
         CALL qmmm_gaussian_initialize(qmmm_gaussian_fns=added_charges%pgfs, &
                                       para_env=para_env, &
                                       pw_env=pw_env, &
                                       mm_el_pot_radius=added_charges%mm_el_pot_radius, &
                                       mm_el_pot_radius_corr=added_charges%mm_el_pot_radius_corr, &
                                       qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                       eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                       maxradius=maxradius2, &
                                       maxchrg=maxchrg, &
                                       compatibility=qmmm_env_qm%compatibility, &
                                       print_section=print_section, &
                                       qmmm_section=qmmm_section)

         SELECT CASE (qmmm_env_qm%qmmm_coupl_type)
         CASE (do_qmmm_gauss, do_qmmm_swave, do_qmmm_pcharge)
            DO i = 1, SIZE(maxradius)
               maxradius(i) = MAX(maxradius(i), maxradius2(i))
            END DO
         END SELECT

         IF (ASSOCIATED(maxradius2)) DEALLOCATE (maxradius2)
      END IF

      IF (qmmm_env_qm%added_shells%num_mm_atoms .GT. 0) THEN

         maxchrg = MAXVAL(ABS(added_shells%mm_core_chrg(:)))
         CALL qmmm_gaussian_initialize(qmmm_gaussian_fns=added_shells%pgfs, &
                                       para_env=para_env, &
                                       pw_env=pw_env, &
                                       mm_el_pot_radius=added_shells%mm_el_pot_radius, &
                                       mm_el_pot_radius_corr=added_shells%mm_el_pot_radius_corr, &
                                       qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                       eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                       maxradius=maxradius2, &
                                       maxchrg=maxchrg, &
                                       compatibility=qmmm_env_qm%compatibility, &
                                       print_section=print_section, &
                                       qmmm_section=qmmm_section)

         SELECT CASE (qmmm_env_qm%qmmm_coupl_type)
         CASE (do_qmmm_gauss, do_qmmm_swave, do_qmmm_pcharge)
            DO i = 1, SIZE(maxradius)
               maxradius(i) = MAX(maxradius(i), maxradius2(i))
            END DO
         END SELECT

         IF (ASSOCIATED(maxradius2)) DEALLOCATE (maxradius2)

      END IF

      qmmm_env_qm%maxradius => maxradius

   END SUBROUTINE qmmm_init_gaussian_type

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env_qm ...
!> \param mm_cell ...
!> \param added_charges ...
!> \param added_shells ...
!> \param print_section ...
!> \par History
!>      1.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_init_potential(qmmm_env_qm, mm_cell, &
                                  added_charges, added_shells, print_section)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env_qm
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(add_set_type), POINTER                        :: added_charges
      TYPE(add_shell_type), POINTER                      :: added_shells
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qmmm_init_potential', &
         routineP = moduleN//':'//routineN

      CALL qmmm_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                               mm_el_pot_radius=qmmm_env_qm%mm_el_pot_radius, &
                               potentials=qmmm_env_qm%potentials, &
                               pgfs=qmmm_env_qm%pgfs, &
                               mm_cell=mm_cell, &
                               compatibility=qmmm_env_qm%compatibility, &
                               print_section=print_section)

      IF (qmmm_env_qm%move_mm_charges .OR. qmmm_env_qm%add_mm_charges) THEN

         CALL qmmm_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                  mm_el_pot_radius=added_charges%mm_el_pot_radius, &
                                  potentials=added_charges%potentials, &
                                  pgfs=added_charges%pgfs, &
                                  mm_cell=mm_cell, &
                                  compatibility=qmmm_env_qm%compatibility, &
                                  print_section=print_section)
      END IF

      IF (qmmm_env_qm%added_shells%num_mm_atoms .GT. 0) THEN

         CALL qmmm_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                  mm_el_pot_radius=added_shells%mm_el_pot_radius, &
                                  potentials=added_shells%potentials, &
                                  pgfs=added_shells%pgfs, &
                                  mm_cell=mm_cell, &
                                  compatibility=qmmm_env_qm%compatibility, &
                                  print_section=print_section)
      END IF

   END SUBROUTINE qmmm_init_potential

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env_qm ...
!> \param qm_cell_small ...
!> \param mm_cell ...
!> \param para_env ...
!> \param qs_env ...
!> \param added_charges ...
!> \param added_shells ...
!> \param qmmm_periodic ...
!> \param print_section ...
!> \param mm_atom_chrg ...
!> \par History
!>      7.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_init_periodic_potential(qmmm_env_qm, qm_cell_small, mm_cell, para_env, qs_env, &
                                           added_charges, added_shells, qmmm_periodic, print_section, mm_atom_chrg)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env_qm
      TYPE(cell_type), POINTER                           :: qm_cell_small, mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(add_set_type), POINTER                        :: added_charges
      TYPE(add_shell_type), POINTER                      :: added_shells
      TYPE(section_vals_type), POINTER                   :: qmmm_periodic, print_section
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_atom_chrg

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qmmm_init_periodic_potential', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: maxchrg
      TYPE(dft_control_type), POINTER                    :: dft_control

      IF (qmmm_env_qm%periodic) THEN

         NULLIFY (dft_control)
         CALL get_qs_env(qs_env, dft_control=dft_control)

         IF (dft_control%qs_control%semi_empirical) THEN
            CALL cp__b("qmmm_init.F",557,"QM/MM periodic calculations not implemented for semi empirical methods")
         ELSE IF (dft_control%qs_control%dftb) THEN
            CALL qmmm_ewald_potential_init(qmmm_env_qm%ewald_env, qmmm_env_qm%ewald_pw, &
                                           qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, mm_cell=mm_cell, &
                                           para_env=para_env, qmmm_periodic=qmmm_periodic, print_section=print_section)
         ELSE IF (dft_control%qs_control%scptb) THEN
            CALL cp__b("qmmm_init.F",563,"QM/MM periodic calculations not implemented for SCPTB method")
         ELSE

            ! setup for GPW/GPAW
            maxchrg = MAXVAL(ABS(mm_atom_chrg(:)))
            IF (qmmm_env_qm%add_mm_charges) maxchrg = MAX(maxchrg, MAXVAL(ABS(added_charges%mm_atom_chrg(:))))

            CALL qmmm_per_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                         per_potentials=qmmm_env_qm%per_potentials, &
                                         potentials=qmmm_env_qm%potentials, &
                                         pgfs=qmmm_env_qm%pgfs, &
                                         qm_cell_small=qm_cell_small, &
                                         mm_cell=mm_cell, &
                                         para_env=para_env, &
                                         compatibility=qmmm_env_qm%compatibility, &
                                         qmmm_periodic=qmmm_periodic, &
                                         print_section=print_section, &
                                         eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                         maxchrg=maxchrg, &
                                         ncp=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts, &
                                         ncpl=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts_local)

            IF (qmmm_env_qm%move_mm_charges .OR. qmmm_env_qm%add_mm_charges) THEN

               CALL qmmm_per_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                            per_potentials=added_charges%per_potentials, &
                                            potentials=added_charges%potentials, &
                                            pgfs=added_charges%pgfs, &
                                            qm_cell_small=qm_cell_small, &
                                            mm_cell=mm_cell, &
                                            para_env=para_env, &
                                            compatibility=qmmm_env_qm%compatibility, &
                                            qmmm_periodic=qmmm_periodic, &
                                            print_section=print_section, &
                                            eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                            maxchrg=maxchrg, &
                                            ncp=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts, &
                                            ncpl=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts_local)
            END IF

            IF (qmmm_env_qm%added_shells%num_mm_atoms .GT. 0) THEN

               CALL qmmm_per_potential_init(qmmm_coupl_type=qmmm_env_qm%qmmm_coupl_type, &
                                            per_potentials=added_shells%per_potentials, &
                                            potentials=added_shells%potentials, &
                                            pgfs=added_shells%pgfs, &
                                            qm_cell_small=qm_cell_small, &
                                            mm_cell=mm_cell, &
                                            para_env=para_env, &
                                            compatibility=qmmm_env_qm%compatibility, &
                                            qmmm_periodic=qmmm_periodic, &
                                            print_section=print_section, &
                                            eps_mm_rspace=qmmm_env_qm%eps_mm_rspace, &
                                            maxchrg=maxchrg, &
                                            ncp=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts, &
                                            ncpl=qmmm_env_qm%aug_pools(SIZE(qmmm_env_qm%aug_pools))%pool%pw_grid%npts_local)
            END IF

         END IF

      END IF

   END SUBROUTINE qmmm_init_periodic_potential

! **************************************************************************************************
!> \brief ...
!> \param qmmm_section ...
!> \param qmmm_env ...
!> \param subsys_mm ...
!> \param qm_atom_type ...
!> \param qm_atom_index ...
!> \param mm_atom_index ...
!> \param qm_cell_small ...
!> \param qmmm_coupl_type ...
!> \param eps_mm_rspace ...
!> \param qmmm_link ...
!> \param para_env ...
!> \par History
!>      11.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE setup_qmmm_vars_qm(qmmm_section, qmmm_env, subsys_mm, qm_atom_type, &
                                 qm_atom_index, mm_atom_index, qm_cell_small, qmmm_coupl_type, eps_mm_rspace, &
                                 qmmm_link, para_env)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(cp_subsys_type), POINTER                      :: subsys_mm
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: qm_atom_type
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index, mm_atom_index
      TYPE(cell_type), POINTER                           :: qm_cell_small
      INTEGER, INTENT(OUT)                               :: qmmm_coupl_type
      REAL(KIND=dp), INTENT(OUT)                         :: eps_mm_rspace
      LOGICAL, INTENT(OUT)                               :: qmmm_link
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_qmmm_vars_qm', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: atmname, mm_atom_kind
      INTEGER                                            :: i, icount, ikind, ikindr, my_type, &
                                                            n_rep_val, nkind, size_mm_system
      INTEGER, DIMENSION(:), POINTER                     :: mm_link_atoms
      LOGICAL                                            :: explicit, is_mm, is_qm
      REAL(KIND=dp)                                      :: tmp_radius, tmp_radius_c
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp_sph_cut
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(section_vals_type), POINTER                   :: cell_section, eri_mme_section, &
                                                            image_charge_section, mm_kinds

      NULLIFY (mm_link_atoms, cell_section, tmp_sph_cut)
      NULLIFY (image_charge_section)
      qmmm_link = .FALSE.

      CALL section_vals_get(qmmm_section, explicit=explicit)
      IF (explicit) THEN
         !
         ! QM_CELL
         !
         cell_section => section_vals_get_subs_vals(qmmm_section, "CELL")
         CALL read_cell(qm_cell_small, qm_cell_small, cell_section=cell_section, &
                        check_for_ref=.FALSE., para_env=para_env)
         CALL section_vals_val_get(qmmm_section, "E_COUPL", i_val=qmmm_coupl_type)
         CALL section_vals_val_get(qmmm_section, "EPS_MM_RSPACE", r_val=eps_mm_rspace)
         CALL section_vals_val_get(qmmm_section, "SPHERICAL_CUTOFF", r_vals=tmp_sph_cut)
         IF(.NOT.(SIZE(tmp_sph_cut) == 2))CALL cp__a("qmmm_init.F",690)
         qmmm_env%spherical_cutoff = tmp_sph_cut
         IF (qmmm_env%spherical_cutoff(1) <= 0.0_dp) THEN
            qmmm_env%spherical_cutoff(2) = 0.0_dp
         ELSE
            IF (qmmm_env%spherical_cutoff(2) <= 0.0_dp) qmmm_env%spherical_cutoff(2) = EPSILON(0.0_dp)
            tmp_radius = qmmm_env%spherical_cutoff(1)-20.0_dp*qmmm_env%spherical_cutoff(2)
            IF (tmp_radius <= 0.0_dp) &
               CALL cp_abort(cp__l("qmmm_init.F",698), &
                             "SPHERICAL_CUTOFF(1) > 20*SPHERICAL_CUTOFF(1)! Please correct parameters for "// &
                             "the Spherical Cutoff in order to satisfy the previous condition!")
         END IF
         !
         ! Initialization of arrays and core_charge_radius...
         !
         tmp_radius = 0.0_dp
         CALL cp_subsys_get(subsys=subsys_mm, atomic_kinds=atomic_kinds)
         DO Ikind = 1, SIZE(atomic_kinds%els)
            atomic_kind => atomic_kinds%els(Ikind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 fist_potential=fist_potential)
            CALL set_potential(potential=fist_potential, &
                               qmmm_radius=tmp_radius, &
                               qmmm_corr_radius=tmp_radius)
            CALL set_atomic_kind(atomic_kind=atomic_kind, &
                                 fist_potential=fist_potential)
         END DO
         CALL setup_qm_atom_list(qmmm_section=qmmm_section, &
                                 qm_atom_index=qm_atom_index, &
                                 qm_atom_type=qm_atom_type, &
                                 mm_link_atoms=mm_link_atoms, &
                                 qmmm_link=qmmm_link)
         !
         ! MM_KINDS
         !
         mm_kinds => section_vals_get_subs_vals(qmmm_section, "MM_KIND")
         CALL section_vals_get(mm_kinds, explicit=explicit, n_repetition=nkind)
         !
         ! Default
         !
         tmp_radius = cp_unit_to_cp2k(RADIUS_QMMM_DEFAULT, "angstrom")
         Set_Radius_Pot_0: DO IkindR = 1, SIZE(atomic_kinds%els)
            atomic_kind => atomic_kinds%els(IkindR)
            CALL get_atomic_kind(atomic_kind=atomic_kind, name=atmname)
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 fist_potential=fist_potential)
            CALL set_potential(potential=fist_potential, qmmm_radius=tmp_radius, &
                               qmmm_corr_radius=tmp_radius)
            CALL set_atomic_kind(atomic_kind=atomic_kind, &
                                 fist_potential=fist_potential)
         END DO Set_Radius_Pot_0
         !
         ! If present overwrite the kind specified in input file...
         !
         IF (explicit) THEN
            DO ikind = 1, nkind
               CALL section_vals_val_get(mm_kinds, "_SECTION_PARAMETERS_", i_rep_section=ikind, &
                                         c_val=mm_atom_kind)
               CALL section_vals_val_get(mm_kinds, "RADIUS", i_rep_section=ikind, r_val=tmp_radius)
               tmp_radius_c = tmp_radius
               CALL section_vals_val_get(mm_kinds, "CORR_RADIUS", i_rep_section=ikind, n_rep_val=n_rep_val)
               IF (n_rep_val == 1) CALL section_vals_val_get(mm_kinds, "CORR_RADIUS", i_rep_section=ikind, &
                                                             r_val=tmp_radius_c)
               Set_Radius_Pot_1: DO IkindR = 1, SIZE(atomic_kinds%els)
                  atomic_kind => atomic_kinds%els(IkindR)
                  CALL get_atomic_kind(atomic_kind=atomic_kind, name=atmname)
                  is_qm = qmmm_ff_precond_only_qm(atmname)
                  IF (TRIM(mm_atom_kind) == atmname) THEN
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          fist_potential=fist_potential)
                     CALL set_potential(potential=fist_potential, &
                                        qmmm_radius=tmp_radius, &
                                        qmmm_corr_radius=tmp_radius_c)
                     CALL set_atomic_kind(atomic_kind=atomic_kind, &
                                          fist_potential=fist_potential)
                  END IF
               END DO Set_Radius_Pot_1
            END DO
         END IF

         !Image charge section

         image_charge_section => section_vals_get_subs_vals(qmmm_section, "IMAGE_CHARGE")
         CALL section_vals_get(image_charge_section, explicit=qmmm_env%image_charge)

      ELSE
         CALL cp__b("qmmm_init.F",776,"QMMM section not present in input file!")
      ENDIF
      !
      ! Build MM atoms list
      !
      size_mm_system = SIZE(subsys_mm%particles%els)-SIZE(qm_atom_index)
      IF (qmmm_link .AND. ASSOCIATED(mm_link_atoms)) size_mm_system = size_mm_system+SIZE(mm_link_atoms)
      ALLOCATE (mm_atom_index(size_mm_system))
      icount = 0

      DO i = 1, SIZE(subsys_mm%particles%els)
         is_mm = .TRUE.
         IF (ANY(qm_atom_index == i)) THEN
            is_mm = .FALSE.
         END IF
         IF (ASSOCIATED(mm_link_atoms)) THEN
            IF (ANY(mm_link_atoms == i) .AND. qmmm_link) is_mm = .TRUE.
         END IF
         IF (is_mm) THEN
            icount = icount+1
            IF (icount <= size_mm_system) mm_atom_index(icount) = i
         END IF
      END DO
      IF(.NOT.(icount == size_mm_system))CALL cp__a("qmmm_init.F",799)
      IF (ASSOCIATED(mm_link_atoms)) THEN
         DEALLOCATE (mm_link_atoms)
      END IF

      ! Build image charge atom list + set up variables
      !
      IF (qmmm_env%image_charge) THEN
         CALL section_vals_val_get(image_charge_section, "MM_ATOM_LIST", &
                                   explicit=explicit)
         IF (explicit) qmmm_env%image_charge_pot%all_mm = .FALSE.

         IF (qmmm_env%image_charge_pot%all_mm) THEN
            qmmm_env%image_charge_pot%image_mm_list => mm_atom_index
         ELSE
            CALL setup_image_atom_list(image_charge_section, qmmm_env, &
                                       qm_atom_index, subsys_mm)
         END IF

         qmmm_env%image_charge_pot%particles_all => subsys_mm%particles%els

         CALL section_vals_val_get(image_charge_section, "EXT_POTENTIAL", &
                                   r_val=qmmm_env%image_charge_pot%V0)
         CALL section_vals_val_get(image_charge_section, "WIDTH", &
                                   r_val=qmmm_env%image_charge_pot%eta)
         CALL section_vals_val_get(image_charge_section, "DETERM_COEFF", &
                                   i_val=my_type)
         SELECT CASE (my_type)
         CASE (do_qmmm_image_calcmatrix)
            qmmm_env%image_charge_pot%coeff_iterative = .FALSE.
         CASE (do_qmmm_image_iter)
            qmmm_env%image_charge_pot%coeff_iterative = .TRUE.
         END SELECT

         CALL section_vals_val_get(image_charge_section, "RESTART_IMAGE_MATRIX", &
                                   l_val=qmmm_env%image_charge_pot%image_restart)

         CALL section_vals_val_get(image_charge_section, "IMAGE_MATRIX_METHOD", &
                                   i_val=qmmm_env%image_charge_pot%image_matrix_method)

         IF (qmmm_env%image_charge_pot%image_matrix_method .EQ. do_eri_mme) THEN
            eri_mme_section => section_vals_get_subs_vals(image_charge_section, "ERI_MME")
            CALL eri_mme_init_read_input(eri_mme_section, qmmm_env%image_charge_pot%eri_mme_param)
            CALL eri_mme_set_params(qmmm_env%image_charge_pot%eri_mme_param, &
                                    qm_cell_small, &
                                    zet_err_minimax=qmmm_env%image_charge_pot%eta, &
                                    zet_err_cutoff=[qmmm_env%image_charge_pot%eta], &
                                    l_max=0, &
                                    para_env=para_env)

         ENDIF
      END IF

   END SUBROUTINE setup_qmmm_vars_qm

! **************************************************************************************************
!> \brief ...
!> \param qmmm_section ...
!> \param qmmm_env ...
!> \param qm_atom_index ...
!> \param mm_link_atoms ...
!> \param mm_link_scale_factor ...
!> \param fist_scale_charge_link ...
!> \param qmmm_coupl_type ...
!> \param qmmm_link ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE setup_qmmm_vars_mm(qmmm_section, qmmm_env, qm_atom_index, &
                                 mm_link_atoms, mm_link_scale_factor, &
                                 fist_scale_charge_link, qmmm_coupl_type, &
                                 qmmm_link)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index, mm_link_atoms
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_link_scale_factor, &
                                                            fist_scale_charge_link
      INTEGER, INTENT(OUT)                               :: qmmm_coupl_type
      LOGICAL, INTENT(OUT)                               :: qmmm_link

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_qmmm_vars_mm', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: qmmm_ff_section

      NULLIFY (qmmm_ff_section)
      qmmm_link = .FALSE.
      CALL section_vals_get(qmmm_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(qmmm_section, "E_COUPL", i_val=qmmm_coupl_type)
         CALL setup_qm_atom_list(qmmm_section, qm_atom_index=qm_atom_index, qmmm_link=qmmm_link, &
                                 mm_link_atoms=mm_link_atoms, mm_link_scale_factor=mm_link_scale_factor, &
                                 fist_scale_charge_link=fist_scale_charge_link)
         !
         ! Do we want to use a different FF for the non-bonded QM/MM interactions?
         !
         qmmm_ff_section => section_vals_get_subs_vals(qmmm_section, "FORCEFIELD")
         CALL section_vals_get(qmmm_ff_section, explicit=qmmm_env%use_qmmm_ff)
         IF (qmmm_env%use_qmmm_ff) THEN
            CALL section_vals_val_get(qmmm_ff_section, "MULTIPLE_POTENTIAL", &
                                      l_val=qmmm_env%multiple_potential)
            CALL read_qmmm_ff_section(qmmm_ff_section, qmmm_env%inp_info)
         END IF
      END IF
   END SUBROUTINE setup_qmmm_vars_mm

! **************************************************************************************************
!> \brief reads information regarding the forcefield specific for the QM/MM
!>      interactions
!> \param qmmm_ff_section ...
!> \param inp_info ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE read_qmmm_ff_section(qmmm_ff_section, inp_info)
      TYPE(section_vals_type), POINTER                   :: qmmm_ff_section
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_qmmm_ff_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n_gd, n_gp, n_lj, n_wl, np
      TYPE(section_vals_type), POINTER                   :: gd_section, gp_section, lj_section, &
                                                            wl_section

!
! NONBONDED
!

      lj_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED%LENNARD-JONES")
      wl_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED%WILLIAMS")
      gd_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED%GOODWIN")
      gp_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED%GENPOT")
      CALL section_vals_get(lj_section, n_repetition=n_lj)
      np = n_lj
      IF (n_lj /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded, 1, np, lj_charmm=.TRUE.)
         CALL read_lj_section(inp_info%nonbonded, lj_section, start=0)
      END IF
      CALL section_vals_get(wl_section, n_repetition=n_wl)
      np = n_lj+n_wl
      IF (n_wl /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded, 1, np, williams=.TRUE.)
         CALL read_wl_section(inp_info%nonbonded, wl_section, start=n_lj)
      END IF
      CALL section_vals_get(gd_section, n_repetition=n_gd)
      np = n_lj+n_wl+n_gd
      IF (n_gd /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded, 1, np, goodwin=.TRUE.)
         CALL read_gd_section(inp_info%nonbonded, gd_section, start=n_lj+n_wl)
      END IF
      CALL section_vals_get(gp_section, n_repetition=n_gp)
      np = n_lj+n_wl+n_gd+n_gp
      IF (n_gp /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded, 1, np, gp=.TRUE.)
         CALL read_gp_section(inp_info%nonbonded, gp_section, start=n_lj+n_wl+n_gd)
      END IF
      !
      ! NONBONDED14
      !
      lj_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED14%LENNARD-JONES")
      wl_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED14%WILLIAMS")
      gd_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED14%GOODWIN")
      gp_section => section_vals_get_subs_vals(qmmm_ff_section, "NONBONDED14%GENPOT")
      CALL section_vals_get(lj_section, n_repetition=n_lj)
      np = n_lj
      IF (n_lj /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded14, 1, np, lj_charmm=.TRUE.)
         CALL read_lj_section(inp_info%nonbonded14, lj_section, start=0)
      END IF
      CALL section_vals_get(wl_section, n_repetition=n_wl)
      np = n_lj+n_wl
      IF (n_wl /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded14, 1, np, williams=.TRUE.)
         CALL read_wl_section(inp_info%nonbonded14, wl_section, start=n_lj)
      END IF
      CALL section_vals_get(gd_section, n_repetition=n_gd)
      np = n_lj+n_wl+n_gd
      IF (n_gd /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded14, 1, np, goodwin=.TRUE.)
         CALL read_gd_section(inp_info%nonbonded14, gd_section, start=n_lj+n_wl)
      END IF
      CALL section_vals_get(gp_section, n_repetition=n_gp)
      np = n_lj+n_wl+n_gd+n_gp
      IF (n_gp /= 0) THEN
         CALL pair_potential_reallocate(inp_info%nonbonded14, 1, np, gp=.TRUE.)
         CALL read_gp_section(inp_info%nonbonded14, gp_section, start=n_lj+n_wl+n_gd)
      END IF

   END SUBROUTINE read_qmmm_ff_section

! **************************************************************************************************
!> \brief ...
!> \param qmmm_section ...
!> \param qm_atom_index ...
!> \param qm_atom_type ...
!> \param mm_link_atoms ...
!> \param mm_link_scale_factor ...
!> \param qmmm_link ...
!> \param fist_scale_charge_link ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE setup_qm_atom_list(qmmm_section, qm_atom_index, qm_atom_type, &
                                 mm_link_atoms, mm_link_scale_factor, qmmm_link, fist_scale_charge_link)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: qm_atom_index
      CHARACTER(len=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: qm_atom_type
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: mm_link_atoms
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: mm_link_scale_factor
      LOGICAL, INTENT(OUT), OPTIONAL                     :: qmmm_link
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: fist_scale_charge_link

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_qm_atom_list', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: qm_atom_kind, qm_link_element
      INTEGER                                            :: ikind, k, link_involv_mm, link_type, &
                                                            mm_index, n_var, nkind, nlinks, &
                                                            num_qm_atom_tot
      INTEGER, DIMENSION(:), POINTER                     :: mm_indexes
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: scale_f
      TYPE(section_vals_type), POINTER                   :: qm_kinds, qmmm_links

      num_qm_atom_tot = 0
      link_involv_mm = 0
      nlinks = 0
      !
      ! QM_KINDS
      !
      qm_kinds => section_vals_get_subs_vals(qmmm_section, "QM_KIND")
      CALL section_vals_get(qm_kinds, n_repetition=nkind)
      DO ikind = 1, nkind
         CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, n_rep_val=n_var)
         DO k = 1, n_var
            CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, i_rep_val=k, &
                                      i_vals=mm_indexes)
            num_qm_atom_tot = num_qm_atom_tot+SIZE(mm_indexes)
         END DO
      END DO
      !
      ! QM/MM LINKS
      !
      qmmm_links => section_vals_get_subs_vals(qmmm_section, "LINK")
      CALL section_vals_get(qmmm_links, explicit=explicit)
      IF (explicit) THEN
         qmmm_link = .TRUE.
         CALL section_vals_get(qmmm_links, n_repetition=nlinks)
         ! Take care of the various link types
         DO ikind = 1, nlinks
            CALL section_vals_val_get(qmmm_links, "LINK_TYPE", i_rep_section=ikind, &
                                      i_val=link_type)
            SELECT CASE (link_type)
            CASE (do_qmmm_link_imomm)
               num_qm_atom_tot = num_qm_atom_tot+1
               link_involv_mm = link_involv_mm+1
            CASE (do_qmmm_link_pseudo)
               num_qm_atom_tot = num_qm_atom_tot+1
            CASE (do_qmmm_link_gho)
               ! do nothing for the moment
            CASE DEFAULT
               CALL cp__b("qmmm_init.F",1066,"")
            END SELECT
         END DO
      END IF
      IF (PRESENT(mm_link_scale_factor) .AND. (link_involv_mm /= 0)) &
         ALLOCATE (mm_link_scale_factor(link_involv_mm))
      IF (PRESENT(fist_scale_charge_link) .AND. (link_involv_mm /= 0)) &
         ALLOCATE (fist_scale_charge_link(link_involv_mm))
      IF (PRESENT(mm_link_atoms) .AND. (link_involv_mm /= 0)) &
         ALLOCATE (mm_link_atoms(link_involv_mm))
      IF (PRESENT(qm_atom_index)) ALLOCATE (qm_atom_index(num_qm_atom_tot))
      IF (PRESENT(qm_atom_type)) ALLOCATE (qm_atom_type(num_qm_atom_tot))
      IF (PRESENT(qm_atom_index)) qm_atom_index = 0
      IF (PRESENT(qm_atom_type)) qm_atom_type = " "
      num_qm_atom_tot = 1
      DO ikind = 1, nkind
         CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, n_rep_val=n_var)
         DO k = 1, n_var
            CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, i_rep_val=k, &
                                      i_vals=mm_indexes)
            IF (PRESENT(qm_atom_index)) THEN
               qm_atom_index(num_qm_atom_tot:num_qm_atom_tot+SIZE(mm_indexes)-1) = mm_indexes(:)
            END IF
            IF (PRESENT(qm_atom_type)) THEN
               CALL section_vals_val_get(qm_kinds, "_SECTION_PARAMETERS_", i_rep_section=ikind, &
                                         c_val=qm_atom_kind)
               qm_atom_type(num_qm_atom_tot:num_qm_atom_tot+SIZE(mm_indexes)-1) = qm_atom_kind
            END IF
            num_qm_atom_tot = num_qm_atom_tot+SIZE(mm_indexes)
         END DO
      END DO
      IF (PRESENT(mm_link_scale_factor) .AND. (link_involv_mm /= 0)) mm_link_scale_factor = 0.0_dp
      IF (PRESENT(fist_scale_charge_link) .AND. (link_involv_mm /= 0)) fist_scale_charge_link = 0.0_dp
      IF (PRESENT(mm_link_atoms) .AND. (link_involv_mm /= 0)) mm_link_atoms = 0
      IF (explicit) THEN
         DO ikind = 1, nlinks
            IF (PRESENT(qm_atom_type)) THEN
               CALL section_vals_val_get(qmmm_links, "QM_KIND", i_rep_section=ikind, c_val=qm_link_element)
               qm_atom_type(num_qm_atom_tot:num_qm_atom_tot) = TRIM(qm_link_element)//"_LINK"
            END IF
            IF (PRESENT(qm_atom_index)) THEN
               CALL section_vals_val_get(qmmm_links, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               IF(.NOT.(ALL(qm_atom_index /= mm_index)))CALL cp__a("qmmm_init.F",1108)
               qm_atom_index(num_qm_atom_tot:num_qm_atom_tot) = mm_index
               num_qm_atom_tot = num_qm_atom_tot+1
            END IF
            IF (PRESENT(mm_link_atoms) .AND. (link_involv_mm /= 0)) THEN
               CALL section_vals_val_get(qmmm_links, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               mm_link_atoms(ikind) = mm_index
            END IF
            IF (PRESENT(mm_link_scale_factor) .AND. (link_involv_mm /= 0)) THEN
               CALL section_vals_val_get(qmmm_links, "QMMM_SCALE_FACTOR", i_rep_section=ikind, r_val=scale_f)
               mm_link_scale_factor(ikind) = scale_f
            END IF
            IF (PRESENT(fist_scale_charge_link) .AND. (link_involv_mm /= 0)) THEN
               CALL section_vals_val_get(qmmm_links, "FIST_SCALE_FACTOR", i_rep_section=ikind, r_val=scale_f)
               fist_scale_charge_link(ikind) = scale_f
            END IF
         END DO
      END IF
      IF(.NOT.(num_qm_atom_tot-1 == SIZE(qm_atom_index)))CALL cp__a("qmmm_init.F",1126)

   END SUBROUTINE setup_qm_atom_list

! **************************************************************************************************
!> \brief this routine sets up all variables to treat qmmm links
!> \param qmmm_section ...
!> \param qmmm_links ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param mm_atom_index ...
!> \param iw ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE setup_qmmm_links(qmmm_section, qmmm_links, mm_el_pot_radius, mm_el_pot_radius_corr, &
                               mm_atom_index, iw)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_el_pot_radius, mm_el_pot_radius_corr
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_qmmm_links', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, link_type, mm_index, n_gho, &
                                                            n_imomm, n_pseudo, n_rep_val, n_tot, &
                                                            nlinks, qm_index
      INTEGER, DIMENSION(:), POINTER                     :: wrk_tmp
      REAL(KIND=dp)                                      :: alpha, my_radius
      TYPE(section_vals_type), POINTER                   :: qmmm_link_section

      NULLIFY (wrk_tmp)
      n_imomm = 0
      n_gho = 0
      n_pseudo = 0
      qmmm_link_section => section_vals_get_subs_vals(qmmm_section, "LINK")
      CALL section_vals_get(qmmm_link_section, n_repetition=nlinks)
      IF(.NOT.(nlinks /= 0))CALL cp__a("qmmm_init.F",1166)
      DO ikind = 1, nlinks
         CALL section_vals_val_get(qmmm_link_section, "LINK_TYPE", i_rep_section=ikind, i_val=link_type)
         IF (link_type == do_qmmm_link_imomm) n_imomm = n_imomm+1
         IF (link_type == do_qmmm_link_gho) n_gho = n_gho+1
         IF (link_type == do_qmmm_link_pseudo) n_pseudo = n_pseudo+1
      END DO
      n_tot = n_imomm+n_gho+n_pseudo
      IF(.NOT.(n_tot /= 0))CALL cp__a("qmmm_init.F",1174)
      ALLOCATE (qmmm_links)
      NULLIFY (qmmm_links%imomm, &
               qmmm_links%pseudo)
      ! IMOMM
      IF (n_imomm /= 0) THEN
         ALLOCATE (qmmm_links%imomm(n_imomm))
         ALLOCATE (wrk_tmp(n_imomm))
         DO ikind = 1, n_imomm
            NULLIFY (qmmm_links%imomm(ikind)%link)
            ALLOCATE (qmmm_links%imomm(ikind)%link)
         END DO
         n_imomm = 0
         DO ikind = 1, nlinks
            CALL section_vals_val_get(qmmm_link_section, "LINK_TYPE", i_rep_section=ikind, i_val=link_type)
            IF (link_type == do_qmmm_link_imomm) THEN
               n_imomm = n_imomm+1
               CALL section_vals_val_get(qmmm_link_section, "QM_INDEX", i_rep_section=ikind, i_val=qm_index)
               CALL section_vals_val_get(qmmm_link_section, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               CALL section_vals_val_get(qmmm_link_section, "ALPHA_IMOMM", i_rep_section=ikind, r_val=alpha)
               CALL section_vals_val_get(qmmm_link_section, "RADIUS", i_rep_section=ikind, n_rep_val=n_rep_val)
               qmmm_links%imomm(n_imomm)%link%qm_index = qm_index
               qmmm_links%imomm(n_imomm)%link%mm_index = mm_index
               qmmm_links%imomm(n_imomm)%link%alpha = alpha
               wrk_tmp(n_imomm) = mm_index
               IF (n_rep_val == 1) THEN
                  CALL section_vals_val_get(qmmm_link_section, "RADIUS", i_rep_section=ikind, r_val=my_radius)
                  WHERE (mm_atom_index == mm_index) mm_el_pot_radius = my_radius
                  WHERE (mm_atom_index == mm_index) mm_el_pot_radius_corr = my_radius
               END IF
               CALL section_vals_val_get(qmmm_link_section, "CORR_RADIUS", i_rep_section=ikind, n_rep_val=n_rep_val)
               IF (n_rep_val == 1) THEN
                  CALL section_vals_val_get(qmmm_link_section, "CORR_RADIUS", i_rep_section=ikind, r_val=my_radius)
                  WHERE (mm_atom_index == mm_index) mm_el_pot_radius_corr = my_radius
               END IF
            END IF
         END DO
         !
         ! Checking the link structure
         !
         DO ikind = 1, SIZE(wrk_tmp)
            IF (COUNT(wrk_tmp == wrk_tmp(ikind)) > 1) THEN
               WRITE (iw, '(/A)') "In the IMOMM scheme no more than one QM atom can be bounded to the same MM atom."
               WRITE (iw, '(A)') "Multiple link MM atom not allowed. Check your link sections."
               CALL cp__b("qmmm_init.F",1218,"")
            END IF
         END DO
         DEALLOCATE (wrk_tmp)
      END IF
      ! PSEUDO
      IF (n_pseudo /= 0) THEN
         ALLOCATE (qmmm_links%pseudo(n_pseudo))
         DO ikind = 1, n_pseudo
            NULLIFY (qmmm_links%pseudo(ikind)%link)
            ALLOCATE (qmmm_links%pseudo(ikind)%link)
         END DO
         n_pseudo = 0
         DO ikind = 1, nlinks
            CALL section_vals_val_get(qmmm_link_section, "LINK_TYPE", i_rep_section=ikind, i_val=link_type)
            IF (link_type == do_qmmm_link_pseudo) THEN
               n_pseudo = n_pseudo+1
               CALL section_vals_val_get(qmmm_link_section, "QM_INDEX", i_rep_section=ikind, i_val=qm_index)
               CALL section_vals_val_get(qmmm_link_section, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               qmmm_links%pseudo(n_pseudo)%link%qm_index = qm_index
               qmmm_links%pseudo(n_pseudo)%link%mm_index = mm_index
            END IF
         END DO
      END IF
      ! GHO
      IF (n_gho /= 0) THEN
         ! not yet implemented
         ! still to define : type, implementation into QS
         CALL cp__b("qmmm_init.F",1246,"")
      END IF
   END SUBROUTINE setup_qmmm_links

! **************************************************************************************************
!> \brief this routine sets up all variables to treat qmmm links
!> \param qmmm_section ...
!> \param move_mm_charges ...
!> \param add_mm_charges ...
!> \param mm_atom_chrg ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param added_charges ...
!> \param mm_atom_index ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE move_or_add_atoms(qmmm_section, move_mm_charges, add_mm_charges, &
                                mm_atom_chrg, mm_el_pot_radius, mm_el_pot_radius_corr, &
                                added_charges, mm_atom_index)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      LOGICAL, INTENT(OUT)                               :: move_mm_charges, add_mm_charges
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_atom_chrg, mm_el_pot_radius, &
                                                            mm_el_pot_radius_corr
      TYPE(add_set_type), POINTER                        :: added_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index

      CHARACTER(len=*), PARAMETER :: routineN = 'move_or_add_atoms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_add, icount, ikind, ind1, Index1, &
                                                            Index2, n_add_tot, n_adds, n_move_tot, &
                                                            n_moves, n_rep_val, nlinks
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: alpha, c_radius, charge, radius
      TYPE(section_vals_type), POINTER                   :: add_section, move_section, &
                                                            qmmm_link_section

      explicit = .FALSE.
      move_mm_charges = .FALSE.
      add_mm_charges = .FALSE.
      NULLIFY (qmmm_link_section, move_section, add_section)
      qmmm_link_section => section_vals_get_subs_vals(qmmm_section, "LINK")
      CALL section_vals_get(qmmm_link_section, n_repetition=nlinks)
      IF(.NOT.(nlinks /= 0))CALL cp__a("qmmm_init.F",1291)
      icount = 0
      n_move_tot = 0
      n_add_tot = 0
      DO ikind = 1, nlinks
         move_section => section_vals_get_subs_vals(qmmm_link_section, "MOVE_MM_CHARGE", &
                                                    i_rep_section=ikind)
         CALL section_vals_get(move_section, n_repetition=n_moves)
         add_section => section_vals_get_subs_vals(qmmm_link_section, "ADD_MM_CHARGE", &
                                                   i_rep_section=ikind)
         CALL section_vals_get(add_section, n_repetition=n_adds)
         n_move_tot = n_move_tot+n_moves
         n_add_tot = n_add_tot+n_adds
      END DO
      icount = n_move_tot+n_add_tot
      IF (n_add_tot /= 0) add_mm_charges = .TRUE.
      IF (n_move_tot /= 0) move_mm_charges = .TRUE.
      !
      ! create add_set_type
      !
      CALL create_add_set_type(added_charges, ndim=icount)
      !
      ! Fill in structures
      !
      icount = 0
      DO ikind = 1, nlinks
         move_section => section_vals_get_subs_vals(qmmm_link_section, "MOVE_MM_CHARGE", &
                                                    i_rep_section=ikind)
         CALL section_vals_get(move_section, explicit=explicit, n_repetition=n_moves)
         !
         ! Moving charge atoms
         !
         IF (explicit) THEN
            DO i_add = 1, n_moves
               icount = icount+1
               CALL section_vals_val_get(move_section, "ATOM_INDEX_1", i_val=Index1, i_rep_section=i_add)
               CALL section_vals_val_get(move_section, "ATOM_INDEX_2", i_val=Index2, i_rep_section=i_add)
               CALL section_vals_val_get(move_section, "ALPHA", r_val=alpha, i_rep_section=i_add)
               CALL section_vals_val_get(move_section, "RADIUS", r_val=radius, i_rep_section=i_add)
               CALL section_vals_val_get(move_section, "CORR_RADIUS", n_rep_val=n_rep_val, i_rep_section=i_add)
               c_radius = radius
               IF (n_rep_val == 1) &
                  CALL section_vals_val_get(move_section, "CORR_RADIUS", r_val=c_radius, i_rep_section=i_add)

               CALL set_add_set_type(added_charges, icount, Index1, Index2, alpha, radius, c_radius, &
                                     mm_atom_chrg=mm_atom_chrg, mm_el_pot_radius=mm_el_pot_radius, &
                                     mm_el_pot_radius_corr=mm_el_pot_radius_corr, &
                                     mm_atom_index=mm_atom_index, move=n_moves, Ind1=ind1)
            END DO
            mm_atom_chrg(ind1) = 0.0_dp
         END IF

         add_section => section_vals_get_subs_vals(qmmm_link_section, "ADD_MM_CHARGE", &
                                                   i_rep_section=ikind)
         CALL section_vals_get(add_section, explicit=explicit, n_repetition=n_adds)
         !
         ! Adding charge atoms
         !
         IF (explicit) THEN
            DO i_add = 1, n_adds
               icount = icount+1
               CALL section_vals_val_get(add_section, "ATOM_INDEX_1", i_val=Index1, i_rep_section=i_add)
               CALL section_vals_val_get(add_section, "ATOM_INDEX_2", i_val=Index2, i_rep_section=i_add)
               CALL section_vals_val_get(add_section, "ALPHA", r_val=alpha, i_rep_section=i_add)
               CALL section_vals_val_get(add_section, "RADIUS", r_val=radius, i_rep_section=i_add)
               CALL section_vals_val_get(add_section, "CHARGE", r_val=charge, i_rep_section=i_add)
               CALL section_vals_val_get(add_section, "CORR_RADIUS", n_rep_val=n_rep_val, i_rep_section=i_add)
               c_radius = radius
               IF (n_rep_val == 1) &
                  CALL section_vals_val_get(add_section, "CORR_RADIUS", r_val=c_radius, i_rep_section=i_add)

               CALL set_add_set_type(added_charges, icount, Index1, Index2, alpha, radius, c_radius, charge, &
                                     mm_atom_chrg=mm_atom_chrg, mm_el_pot_radius=mm_el_pot_radius, &
                                     mm_el_pot_radius_corr=mm_el_pot_radius_corr, &
                                     mm_atom_index=mm_atom_index)
            END DO
         END IF
      END DO

   END SUBROUTINE move_or_add_atoms

! **************************************************************************************************
!> \brief this routine sets up all variables of the add_set_type type
!> \param added_charges ...
!> \param icount ...
!> \param Index1 ...
!> \param Index2 ...
!> \param alpha ...
!> \param radius ...
!> \param c_radius ...
!> \param charge ...
!> \param mm_atom_chrg ...
!> \param mm_el_pot_radius ...
!> \param mm_el_pot_radius_corr ...
!> \param mm_atom_index ...
!> \param move ...
!> \param ind1 ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE set_add_set_type(added_charges, icount, Index1, Index2, alpha, radius, c_radius, charge, &
                               mm_atom_chrg, mm_el_pot_radius, mm_el_pot_radius_corr, mm_atom_index, move, ind1)
      TYPE(add_set_type), POINTER                        :: added_charges
      INTEGER, INTENT(IN)                                :: icount, Index1, Index2
      REAL(KIND=dp), INTENT(IN)                          :: alpha, radius, c_radius
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: charge
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_atom_chrg, mm_el_pot_radius, &
                                                            mm_el_pot_radius_corr
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      INTEGER, INTENT(in), OPTIONAL                      :: move
      INTEGER, INTENT(OUT), OPTIONAL                     :: ind1

      CHARACTER(len=*), PARAMETER :: routineN = 'set_add_set_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_move
      REAL(KIND=dp)                                      :: my_c_radius, my_charge, my_radius

      my_move = 0
      my_radius = radius
      my_c_radius = c_radius
      IF (PRESENT(charge)) my_charge = charge
      IF (PRESENT(move)) my_move = move
      i = 1
      GetId: DO WHILE (i <= SIZE(mm_atom_index))
         IF (Index1 == mm_atom_index(i)) EXIT GetId
         i = i+1
      END DO GetId
      IF (PRESENT(ind1)) ind1 = i
      IF(.NOT.(i <= SIZE(mm_atom_index)))CALL cp__a("qmmm_init.F",1421)
      IF (.NOT. PRESENT(charge)) my_charge = mm_atom_chrg(i)/REAL(my_move, KIND=dp)
      IF (my_radius == 0.0_dp) my_radius = mm_el_pot_radius(i)
      IF (my_c_radius == 0.0_dp) my_c_radius = mm_el_pot_radius_corr(i)

      added_charges%add_env(icount)%Index1 = Index1
      added_charges%add_env(icount)%Index2 = Index2
      added_charges%add_env(icount)%alpha = alpha
      added_charges%mm_atom_index(icount) = icount
      added_charges%mm_atom_chrg(icount) = my_charge
      added_charges%mm_el_pot_radius(icount) = my_radius
      added_charges%mm_el_pot_radius_corr(icount) = my_c_radius
   END SUBROUTINE set_add_set_type

! **************************************************************************************************
!> \brief this routine sets up the origin of the MM cell respect to the
!>      origin of the QM cell. The origin of the QM cell is assumed to be
!>      in (0.0,0.0,0.0)...
!> \param qmmm_section ...
!> \param qmmm_env ...
!> \param qm_cell_small ...
!> \param dr ...
!> \par History
!>      02.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE setup_origin_mm_cell(qmmm_section, qmmm_env, qm_cell_small, &
                                   dr)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(cell_type), POINTER                           :: qm_cell_small
      REAL(KIND=dp), DIMENSION(3), INTENT(in)            :: dr

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_origin_mm_cell', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: center_grid
      REAL(KIND=dp), DIMENSION(3)                        :: tmp
      REAL(KINd=dp), DIMENSION(:), POINTER               :: vec

! This is the vector that corrects position to apply properly the PBC

      tmp(1) = qm_cell_small%hmat(1, 1)
      tmp(2) = qm_cell_small%hmat(2, 2)
      tmp(3) = qm_cell_small%hmat(3, 3)
      IF(.NOT.(ALL(tmp > 0)))CALL cp__a("qmmm_init.F",1466)
      qmmm_env%dOmmOqm = tmp/2.0_dp
      ! This is unit vector to translate the QM system in order to center it
      ! in QM cell
      CALL section_vals_val_get(qmmm_section, "CENTER_GRID", l_val=center_grid)
      IF (center_grid) THEN
         qmmm_env%utrasl = dr
      ELSE
         qmmm_env%utrasl = 1.0_dp
      ENDIF
      CALL section_vals_val_get(qmmm_section, "INITIAL_TRANSLATION_VECTOR", r_vals=vec)
      qmmm_env%transl_v = vec
   END SUBROUTINE setup_origin_mm_cell

! **************************************************************************************************
!> \brief this routine sets up list of MM atoms carrying an image charge
!> \param image_charge_section ...
!> \param qmmm_env ...
!> \param qm_atom_index ...
!> \param subsys_mm ...
!> \par History
!>      02.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE setup_image_atom_list(image_charge_section, qmmm_env, &
                                    qm_atom_index, subsys_mm)

      TYPE(section_vals_type), POINTER                   :: image_charge_section
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index
      TYPE(cp_subsys_type), POINTER                      :: subsys_mm

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_image_atom_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, i, j, k, max_index, &
                                                            n_var, num_const_atom, &
                                                            num_image_mm_atom
      INTEGER, DIMENSION(:), POINTER                     :: mm_indexes
      LOGICAL                                            :: fix_xyz, imageind_in_range
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind

      NULLIFY (mm_indexes, molecule_kind)
      imageind_in_range = .FALSE.
      num_image_mm_atom = 0
      max_index = 0

      CALL section_vals_val_get(image_charge_section, "MM_ATOM_LIST", &
                                n_rep_val=n_var)
      DO i = 1, n_var
         CALL section_vals_val_get(image_charge_section, "MM_ATOM_LIST", &
                                   i_rep_val=i, i_vals=mm_indexes)
         num_image_mm_atom = num_image_mm_atom+SIZE(mm_indexes)
      END DO

      ALLOCATE (qmmm_env%image_charge_pot%image_mm_list(num_image_mm_atom))

      qmmm_env%image_charge_pot%image_mm_list = 0
      num_image_mm_atom = 1

      DO i = 1, n_var
         CALL section_vals_val_get(image_charge_section, "MM_ATOM_LIST", &
                                   i_rep_val=i, i_vals=mm_indexes)
         qmmm_env%image_charge_pot%image_mm_list(num_image_mm_atom:num_image_mm_atom &
                                                 +SIZE(mm_indexes)-1) = mm_indexes(:)
         num_image_mm_atom = num_image_mm_atom+SIZE(mm_indexes)
      END DO

      ! checking, if in range, if list contains QM atoms or any atoms doubled
      num_image_mm_atom = num_image_mm_atom-1

      max_index = SIZE(subsys_mm%particles%els)

      IF(.NOT.(SIZE(qmmm_env%image_charge_pot%image_mm_list) /= 0))CALL cp__a("qmmm_init.F",1539)
      imageind_in_range = (MAXVAL(qmmm_env%image_charge_pot%image_mm_list) <= max_index) &
                          .AND. (MINVAL(qmmm_env%image_charge_pot%image_mm_list) > 0)
      IF(.NOT.(imageind_in_range))CALL cp__a("qmmm_init.F",1542)

      DO i = 1, num_image_mm_atom
         atom_a = qmmm_env%image_charge_pot%image_mm_list(i)
         IF (ANY(qm_atom_index == atom_a)) THEN
            CALL cp__b("qmmm_init.F",1547,"Image atom list must only contain MM atoms")
         ENDIF
         DO j = i+1, num_image_mm_atom
            atom_b = qmmm_env%image_charge_pot%image_mm_list(j)
            IF (atom_a == atom_b) &
               CALL cp__b("qmmm_init.F",1552,"There are atoms doubled in image list.")
         ENDDO
      ENDDO

      ! check if molecules in list carry constraints
      num_const_atom = 0
      fix_xyz = .TRUE.
      IF (ASSOCIATED(subsys_mm%molecule_kinds_new)) THEN
         IF (ASSOCIATED(subsys_mm%molecule_kinds_new%els)) THEN
            molecule_kind => subsys_mm%molecule_kinds_new%els
            DO i = 1, SIZE(molecule_kind)
               IF (.NOT. ASSOCIATED(molecule_kind(i)%fixd_list)) EXIT
               IF (.NOT. fix_xyz) EXIT
               DO j = 1, SIZE(molecule_kind(i)%fixd_list)
                  IF (.NOT. fix_xyz) EXIT
                  DO k = 1, num_image_mm_atom
                     atom_a = qmmm_env%image_charge_pot%image_mm_list(k)
                     IF (atom_a == molecule_kind(i)%fixd_list(j)%fixd) THEN
                        num_const_atom = num_const_atom+1
                        IF (molecule_kind(i)%fixd_list(j)%itype /= use_perd_xyz) THEN
                           fix_xyz = .FALSE.
                           EXIT
                        ENDIF
                     ENDIF
                  ENDDO
               ENDDO
            ENDDO
         ENDIF
      ENDIF

      ! if all image atoms are constrained, calculate image matrix only
      ! once for the first MD or GEO_OPT step (for non-iterative case)
      IF (num_const_atom == num_image_mm_atom .AND. fix_xyz) THEN
         qmmm_env%image_charge_pot%state_image_matrix = calc_once
      ELSE
         qmmm_env%image_charge_pot%state_image_matrix = calc_always
      ENDIF

   END SUBROUTINE setup_image_atom_list

! **************************************************************************************************
!> \brief Print info on image charges
!> \param qmmm_env ...
!> \param qmmm_section ...
!> \par History
!>      03.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE print_image_charge_info(qmmm_env, qmmm_section)

      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: qmmm_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_image_charge_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw
      REAL(KIND=dp)                                      :: eta, eta_conv, V0, V0_conv
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, qmmm_section, "PRINT%PROGRAM_RUN_INFO", &
                                extension=".log")
      eta = qmmm_env%image_charge_pot%eta
      eta_conv = cp_unit_from_cp2k(eta, "angstrom", power=-2)
      V0 = qmmm_env%image_charge_pot%V0
      V0_conv = cp_unit_from_cp2k(V0, "volt")

      IF (iw > 0) THEN
         WRITE (iw, FMT="(T25,A)") "IMAGE CHARGE PARAMETERS"
         WRITE (iw, FMT="(T25,A)") REPEAT("-", 23)
         WRITE (iw, FMT="(/)")
         WRITE (iw, FMT="(T2,A)") "INDEX OF MM ATOMS CARRYING AN IMAGE CHARGE:"
         WRITE (iw, FMT="(/)")

         WRITE (iw, "(7X,10I6)") qmmm_env%image_charge_pot%image_mm_list
         WRITE (iw, FMT="(/)")
         WRITE (iw, "(T2,A52,T69,F12.8)") &
            "WIDTH OF GAUSSIAN CHARGE DISTRIBUTION [angstrom^-2]:", eta_conv
         WRITE (iw, "(T2,A26,T69,F12.8)") "EXTERNAL POTENTIAL [volt]:", V0_conv
         WRITE (iw, FMT="(/,T2,A,/)") REPEAT("-", 79)
      END IF
      CALL cp_print_key_finished_output(iw, logger, qmmm_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE print_image_charge_info

END MODULE qmmm_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Subroutine input_torsions changed (DG) 05-Dec-2000
!>      Output formats changed (DG) 05-Dec-2000
!>      JGH (26-01-2002) : force field parameters stored in tables, not in
!>        matrices. Input changed to have parameters labeled by the position
!>        and not atom pairs (triples etc)
!>      Teo (11.2005) : Moved all information on force field  pair_potential to
!>                      a much lighter memory structure
!>      Teo 09.2006   : Splitted all routines force_field I/O in a separate file
!>      10.2008 Teodoro Laino [tlaino] :  splitted from force_fields_input in order
!>              to collect in a unique module only the functions to read external FF
!> \author CJM
! **************************************************************************************************
MODULE force_fields_ext
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE force_field_kind_types,          ONLY: do_ff_g96
   USE force_field_types,               ONLY: amber_info_type,&
                                              charmm_info_type,&
                                              force_field_type,&
                                              gromos_info_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: particle_type
   USE string_utilities,                ONLY: uppercase
   USE topology_amber,                  ONLY: rdparm_amber_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_fields_ext'

   PRIVATE
   PUBLIC :: read_force_field_charmm, &
             read_force_field_amber, &
             read_force_field_gromos

CONTAINS

! **************************************************************************************************
!> \brief Reads the GROMOS force_field
!> \param ff_type ...
!> \param para_env ...
!> \param mm_section ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE read_force_field_gromos(ff_type, para_env, mm_section)

      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_force_field_gromos', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: ekt = 2.5_dp

      CHARACTER(LEN=default_string_length)               :: label
      CHARACTER(LEN=default_string_length), &
         DIMENSION(21)                                   :: avail_section
      CHARACTER(LEN=default_string_length), POINTER      :: namearray(:)
      INTEGER                                            :: handle, iatom, icon, itemp, itype, iw, &
                                                            jatom, ncon, ntype, offset
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: cosphi0, cost2, csq, sdet
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(gromos_info_type), POINTER                    :: gro_info

      CALL timeset(routineN, handle)
      NULLIFY (logger, parser)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")

      avail_section(1) = "TITLE"
      avail_section(2) = "TOPPHYSCON"
      avail_section(3) = "TOPVERSION"
      avail_section(4) = "ATOMTYPENAME"
      avail_section(5) = "RESNAME"
      avail_section(6) = "SOLUTEATOM"
      avail_section(7) = "BONDTYPE"
      avail_section(8) = "BONDH"
      avail_section(9) = "BOND"
      avail_section(10) = "BONDANGLETYPE"
      avail_section(11) = "BONDANGLEH"
      avail_section(12) = "BONDANGLE"
      avail_section(13) = "IMPDIHEDRALTYPE"
      avail_section(14) = "IMPDIHEDRALH"
      avail_section(15) = "IMPDIHEDRAL"
      avail_section(16) = "DIHEDRALTYPE"
      avail_section(17) = "DIHEDRALH"
      avail_section(18) = "DIHEDRAL"
      avail_section(19) = "LJPARAMETERS"
      avail_section(20) = "SOLVENTATOM"
      avail_section(21) = "SOLVENTCONSTR"

      gro_info => ff_type%gro_info
      gro_info%ff_gromos_type = ff_type%ff_type
      NULLIFY (namearray)
      ! ATOMTYPENAME SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the ATOMTYPENAME section'
      CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
      label = TRIM(avail_section(4))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(namearray, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, namearray(itype), lower_to_upper=.TRUE.)
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO|  ", TRIM(namearray(itype))
         END DO
      END IF
      CALL parser_release(parser)

      ! SOLVENTCONSTR SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the SOLVENTATOM section'
      CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)

      label = TRIM(avail_section(21))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ncon)
         CALL reallocate(gro_info%solvent_k, 1, ncon)
         CALL reallocate(gro_info%solvent_r0, 1, ncon)
         DO icon = 1, ncon
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, itemp)
            CALL parser_get_object(parser, itemp)
            CALL parser_get_object(parser, gro_info%solvent_r0(icon))
            gro_info%solvent_k(icon) = 0.0_dp
         END DO
      END IF
      CALL parser_release(parser)

      CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
      ! BONDTYPE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the BONDTYPE section'
      label = TRIM(avail_section(7))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(gro_info%bond_k, 1, ntype)
         CALL reallocate(gro_info%bond_r0, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, gro_info%bond_k(itype))
            CALL parser_get_object(parser, gro_info%bond_r0(itype))
            IF (ff_type%ff_type == do_ff_g96) THEN
               gro_info%bond_k(itype) = cp_unit_to_cp2k(gro_info%bond_k(itype), "kjmol*nm^-4")
            ELSE ! Assume its G87
               gro_info%bond_k(itype) = (2.0_dp)*gro_info%bond_k(itype)*gro_info%bond_r0(itype)**2
               gro_info%bond_k(itype) = cp_unit_to_cp2k(gro_info%bond_k(itype), "kjmol*nm^-2")
            END IF
            gro_info%bond_r0(itype) = cp_unit_to_cp2k(gro_info%bond_r0(itype), "nm")
            IF (iw > 0) WRITE (iw, *) "GROMOS_FF| PUT BONDTYPE INFO HERE!!!!"
         END DO
      END IF

      ! BONDANGLETYPE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the BONDANGLETYPE section'
      label = TRIM(avail_section(10))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(gro_info%bend_k, 1, ntype)
         CALL reallocate(gro_info%bend_theta0, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, gro_info%bend_k(itype))
            CALL parser_get_object(parser, gro_info%bend_theta0(itype))
            gro_info%bend_theta0(itype) = cp_unit_to_cp2k(gro_info%bend_theta0(itype), "deg")
            IF (ff_type%ff_type == do_ff_g96) THEN
               gro_info%bend_theta0(itype) = COS(gro_info%bend_theta0(itype))
            ELSE ! Assume its G87
               cost2 = COS(gro_info%bend_theta0(itype))*COS(gro_info%bend_theta0(itype))
               sdet = cost2*cost2-(2.0_dp*cost2-1.0_dp)*(1.0_dp-ekt/gro_info%bend_k(itype))
               csq = (cost2-SQRT(sdet))/(2.0_dp*cost2-1.0_dp)
               gro_info%bend_k(itype) = ekt/ACOS(csq)**2
            END IF
            gro_info%bend_k(itype) = cp_unit_to_cp2k(gro_info%bend_k(itype), "kjmol")
            IF (iw > 0) WRITE (iw, *) "GROMOS_FF| PUT BONDANGLETYPE INFO HERE!!!!"
         END DO
      END IF

      ! IMPDIHEDRALTYPE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the IMPDIHEDRALTYPE section'
      label = TRIM(avail_section(13))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(gro_info%impr_k, 1, ntype)
         CALL reallocate(gro_info%impr_phi0, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, gro_info%impr_k(itype))
            CALL parser_get_object(parser, gro_info%impr_phi0(itype))
            gro_info%impr_phi0(itype) = cp_unit_to_cp2k(gro_info%impr_phi0(itype), "deg")
            gro_info%impr_k(itype) = cp_unit_to_cp2k(gro_info%impr_k(itype), "kjmol*deg^-2")
            IF (iw > 0) WRITE (iw, *) "GROMOS_FF| PUT IMPDIHEDRALTYPE INFO HERE!!!!"
         END DO
      END IF

      ! DIHEDRALTYPE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the DIHEDRALTYPE section'
      label = TRIM(avail_section(16))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(gro_info%torsion_k, 1, ntype)
         CALL reallocate(gro_info%torsion_m, 1, ntype)
         CALL reallocate(gro_info%torsion_phi0, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, gro_info%torsion_k(itype))
            CALL parser_get_object(parser, cosphi0)
            CALL parser_get_object(parser, gro_info%torsion_m(itype))
            gro_info%torsion_phi0(itype) = ACOS(cosphi0)
            gro_info%torsion_k(itype) = cp_unit_to_cp2k(gro_info%torsion_k(itype), "kjmol")
            IF (iw > 0) WRITE (iw, *) "GROMOS_FF| PUT DIHEDRALTYPE INFO HERE!!!!"
         END DO
      END IF

      CALL parser_release(parser)

      ! LJPARAMETERS SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GROMOS_FF| Parsing the LJPARAMETERS section'
      CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
      label = TRIM(avail_section(19))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(gro_info%nonbond_a)) offset = SIZE(gro_info%nonbond_a)
         ntype = SIZE(namearray)
         CALL reallocate(gro_info%nonbond_a, 1, ntype)
         CALL reallocate(gro_info%nonbond_a_14, 1, ntype)
         CALL reallocate(gro_info%nonbond_c6, 1, ntype, 1, ntype)
         CALL reallocate(gro_info%nonbond_c12, 1, ntype, 1, ntype)
         CALL reallocate(gro_info%nonbond_c6_14, 1, ntype, 1, ntype)
         CALL reallocate(gro_info%nonbond_c12_14, 1, ntype, 1, ntype)

         gro_info%nonbond_c12 = 0._dp
         gro_info%nonbond_c6 = 0._dp
         gro_info%nonbond_c12_14 = 0._dp
         gro_info%nonbond_c6_14 = 0._dp

         DO itype = 1, ntype*(ntype+1)/2
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, iatom)
            CALL parser_get_object(parser, jatom)
            IF (iatom == jatom) THEN
               gro_info%nonbond_a(iatom) = namearray(iatom)
               gro_info%nonbond_a_14(iatom) = namearray(iatom)
            END IF
            CALL parser_get_object(parser, gro_info%nonbond_c12(iatom, jatom))
            CALL parser_get_object(parser, gro_info%nonbond_c6(iatom, jatom))
            CALL parser_get_object(parser, gro_info%nonbond_c12_14(iatom, jatom))
            CALL parser_get_object(parser, gro_info%nonbond_c6_14(iatom, jatom))
            gro_info%nonbond_c6(iatom, jatom) = cp_unit_to_cp2k(gro_info%nonbond_c6(iatom, jatom), "kjmol*nm^6")
            gro_info%nonbond_c12(iatom, jatom) = cp_unit_to_cp2k(gro_info%nonbond_c12(iatom, jatom), "kjmol*nm^12")
            gro_info%nonbond_c6_14(iatom, jatom) = cp_unit_to_cp2k(gro_info%nonbond_c6_14(iatom, jatom), "kjmol*nm^6")
            gro_info%nonbond_c12_14(iatom, jatom) = cp_unit_to_cp2k(gro_info%nonbond_c12_14(iatom, jatom), "kjmol*nm^12")

            gro_info%nonbond_c6_14(jatom, iatom) = gro_info%nonbond_c6_14(iatom, jatom)
            gro_info%nonbond_c12_14(jatom, iatom) = gro_info%nonbond_c12_14(iatom, jatom)
            gro_info%nonbond_c6(jatom, iatom) = gro_info%nonbond_c6(iatom, jatom)
            gro_info%nonbond_c12(jatom, iatom) = gro_info%nonbond_c12(iatom, jatom)
            IF (iw > 0) WRITE (iw, *) "GROMOS_FF| PUT LJPARAMETERS INFO HERE!!!!"
         END DO
      END IF
      CALL parser_release(parser)

      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%FF_INFO")
      CALL timestop(handle)

      DEALLOCATE (namearray)
   END SUBROUTINE read_force_field_gromos

! **************************************************************************************************
!> \brief Reads the charmm force_field
!> \param ff_type ...
!> \param para_env ...
!> \param mm_section ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE read_force_field_charmm(ff_type, para_env, mm_section)

      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_force_field_charmm', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: label, string, string2, string3, string4
      CHARACTER(LEN=default_string_length), DIMENSION(1) :: bond_section
      CHARACTER(LEN=default_string_length), &
         DIMENSION(19)                                   :: avail_section
      CHARACTER(LEN=default_string_length), DIMENSION(2) :: angl_section, impr_section, &
                                                            nbon_section, thet_section
      INTEGER                                            :: dummy, handle, ilab, iw, nbend, nbond, &
                                                            nimpr, nnonbond, nonfo, ntorsion, nub
      LOGICAL                                            :: found
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      CALL timeset(routineN, handle)
      NULLIFY (logger, parser)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")

      avail_section(1) = "BOND"; bond_section(1) = avail_section(1)
      avail_section(11) = "BONDS"
      avail_section(2) = "ANGL"; angl_section(1) = avail_section(2)
      avail_section(3) = "THETA"; angl_section(2) = avail_section(3)
      avail_section(12) = "THETAS"
      avail_section(13) = "ANGLE"
      avail_section(14) = "ANGLES"
      avail_section(4) = "DIHEDRAL"; thet_section(1) = avail_section(4)
      avail_section(15) = "DIHEDRALS"
      avail_section(5) = "PHI"; thet_section(2) = avail_section(5)
      avail_section(6) = "IMPROPER"; impr_section(1) = avail_section(6)
      avail_section(7) = "IMPH"; impr_section(2) = avail_section(7)
      avail_section(16) = "IMPHI"
      avail_section(8) = "NONBONDED"; nbon_section(1) = avail_section(8)
      avail_section(9) = "NBOND"; nbon_section(2) = avail_section(9)
      avail_section(10) = "HBOND"
      avail_section(17) = "NBFIX"
      avail_section(18) = "CMAP"
      avail_section(19) = "END"

      chm_info => ff_type%chm_info
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Read in all the Bonds info from the param file here
      !      Vbond = Kb(b-b0)^2
      !      UNITS for Kb: [(kcal/mol)/(A^2)] to [Eh/(AU^2)]
      !-----------------------------------------------------------------------------
      nbond = 0
      DO ilab = 1, SIZE(bond_section)
         CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
         label = TRIM(bond_section(ilab))
         DO
            CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
            IF (found) THEN
               CALL parser_get_object(parser, string)
               IF (INDEX(string, TRIM(label)) /= 1) CYCLE
               CALL charmm_get_next_line(parser, 1)
               DO
                  CALL parser_get_object(parser, string)
                  CALL uppercase(string)
                  IF (ANY(string == avail_section)) EXIT
                  CALL parser_get_object(parser, string2)
                  CALL uppercase(string2)
                  nbond = nbond+1
                  CALL reallocate(chm_info%bond_a, 1, nbond)
                  CALL reallocate(chm_info%bond_b, 1, nbond)
                  CALL reallocate(chm_info%bond_k, 1, nbond)
                  CALL reallocate(chm_info%bond_r0, 1, nbond)
                  chm_info%bond_a(nbond) = string
                  chm_info%bond_b(nbond) = string2
                  CALL parser_get_object(parser, chm_info%bond_k(nbond))
                  CALL parser_get_object(parser, chm_info%bond_r0(nbond))
                  IF (iw > 0) WRITE (iw, *) "    CHM BOND ", nbond, &
                     TRIM(chm_info%bond_a(nbond)), " ", &
                     TRIM(chm_info%bond_b(nbond)), " ", &
                     chm_info%bond_k(nbond), &
                     chm_info%bond_r0(nbond)
                  ! Do some units conversion into internal atomic units
                  chm_info%bond_r0(nbond) = cp_unit_to_cp2k(chm_info%bond_r0(nbond), "angstrom")
                  chm_info%bond_k(nbond) = cp_unit_to_cp2k(chm_info%bond_k(nbond), "kcalmol*angstrom^-2")
                  CALL charmm_get_next_line(parser, 1)
               END DO
            ELSE
               EXIT
            END IF
         END DO
         CALL parser_release(parser)
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Read in all the Bends and UB info from the param file here
      !      Vangle = Ktheta(theta-theta0)^2
      !      UNITS for Ktheta: [(kcal/mol)/(rad^2)] to [Eh/(rad^2)]
      !      FACTOR of "2" rolled into Ktheta
      !      Vub = Kub(S-S0)^2
      !      UNITS for Kub: [(kcal/mol)/(A^2)] to [Eh/(AU^2)]
      !-----------------------------------------------------------------------------
      nbend = 0
      nub = 0
      DO ilab = 1, SIZE(angl_section)
         CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
         label = TRIM(angl_section(ilab))
         DO
            CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
            IF (found) THEN
               CALL parser_get_object(parser, string)
               IF (INDEX(string, TRIM(label)) /= 1) CYCLE
               CALL charmm_get_next_line(parser, 1)
               DO
                  CALL parser_get_object(parser, string)
                  CALL uppercase(string)
                  IF (ANY(string == avail_section)) EXIT
                  CALL parser_get_object(parser, string2)
                  CALL parser_get_object(parser, string3)
                  CALL uppercase(string2)
                  CALL uppercase(string3)
                  nbend = nbend+1
                  CALL reallocate(chm_info%bend_a, 1, nbend)
                  CALL reallocate(chm_info%bend_b, 1, nbend)
                  CALL reallocate(chm_info%bend_c, 1, nbend)
                  CALL reallocate(chm_info%bend_k, 1, nbend)
                  CALL reallocate(chm_info%bend_theta0, 1, nbend)
                  chm_info%bend_a(nbend) = string
                  chm_info%bend_b(nbend) = string2
                  chm_info%bend_c(nbend) = string3
                  CALL parser_get_object(parser, chm_info%bend_k(nbend))
                  CALL parser_get_object(parser, chm_info%bend_theta0(nbend))
                  IF (iw > 0) WRITE (iw, *) "    CHM BEND ", nbend, &
                     TRIM(chm_info%bend_a(nbend)), " ", &
                     TRIM(chm_info%bend_b(nbend)), " ", &
                     TRIM(chm_info%bend_c(nbend)), " ", &
                     chm_info%bend_k(nbend), &
                     chm_info%bend_theta0(nbend)
                  ! Do some units conversion into internal atomic units
                  chm_info%bend_theta0(nbend) = cp_unit_to_cp2k(chm_info%bend_theta0(nbend), "deg")
                  chm_info%bend_k(nbend) = cp_unit_to_cp2k(chm_info%bend_k(nbend), "kcalmol*rad^-2")
                  IF (parser_test_next_token(parser) == "FLT") THEN
                     nub = nub+1
                     CALL reallocate(chm_info%ub_a, 1, nub)
                     CALL reallocate(chm_info%ub_b, 1, nub)
                     CALL reallocate(chm_info%ub_c, 1, nub)
                     CALL reallocate(chm_info%ub_k, 1, nub)
                     CALL reallocate(chm_info%ub_r0, 1, nub)
                     chm_info%ub_a(nub) = string
                     chm_info%ub_b(nub) = string2
                     chm_info%ub_c(nub) = string3
                     CALL parser_get_object(parser, chm_info%ub_k(nub))
                     CALL parser_get_object(parser, chm_info%ub_r0(nub))
                     IF (iw > 0) WRITE (iw, *) "    CHM UB ", nub, &
                        TRIM(chm_info%ub_a(nub)), " ", &
                        TRIM(chm_info%ub_b(nub)), " ", &
                        TRIM(chm_info%ub_c(nub)), " ", &
                        chm_info%ub_k(nub), &
                        chm_info%ub_r0(nub)
                     ! Do some units conversion into internal atomic units
                     chm_info%ub_r0(nub) = cp_unit_to_cp2k(chm_info%ub_r0(nub), "angstrom")
                     chm_info%ub_k(nub) = cp_unit_to_cp2k(chm_info%ub_k(nub), "kcalmol*angstrom^-2")
                  END IF
                  CALL charmm_get_next_line(parser, 1)
               END DO
            ELSE
               EXIT
            END IF
         END DO
         CALL parser_release(parser)
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Read in all the Dihedrals info from the param file here
      !      Vtorsion = Kphi(1+COS(n(phi)-delta))
      !      UNITS for Kphi: [(kcal/mol)] to [Eh]
      !-----------------------------------------------------------------------------
      ntorsion = 0
      DO ilab = 1, SIZE(thet_section)
         CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
         label = TRIM(thet_section(ilab))
         DO
            CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
            IF (found) THEN
               CALL parser_get_object(parser, string)
               IF (INDEX(string, TRIM(label)) /= 1) CYCLE
               CALL charmm_get_next_line(parser, 1)
               DO
                  CALL parser_get_object(parser, string)
                  CALL uppercase(string)
                  IF (ANY(string == avail_section)) EXIT
                  CALL parser_get_object(parser, string2)
                  CALL parser_get_object(parser, string3)
                  CALL parser_get_object(parser, string4)
                  CALL uppercase(string2)
                  CALL uppercase(string3)
                  CALL uppercase(string4)
                  ntorsion = ntorsion+1
                  CALL reallocate(chm_info%torsion_a, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_b, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_c, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_d, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_k, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_m, 1, ntorsion)
                  CALL reallocate(chm_info%torsion_phi0, 1, ntorsion)
                  chm_info%torsion_a(ntorsion) = string
                  chm_info%torsion_b(ntorsion) = string2
                  chm_info%torsion_c(ntorsion) = string3
                  chm_info%torsion_d(ntorsion) = string4
                  CALL parser_get_object(parser, chm_info%torsion_k(ntorsion))
                  CALL parser_get_object(parser, chm_info%torsion_m(ntorsion))
                  CALL parser_get_object(parser, chm_info%torsion_phi0(ntorsion))
                  IF (iw > 0) WRITE (iw, *) "    CHM TORSION ", ntorsion, &
                     TRIM(chm_info%torsion_a(ntorsion)), " ", &
                     TRIM(chm_info%torsion_b(ntorsion)), " ", &
                     TRIM(chm_info%torsion_c(ntorsion)), " ", &
                     TRIM(chm_info%torsion_d(ntorsion)), " ", &
                     chm_info%torsion_k(ntorsion), &
                     chm_info%torsion_m(ntorsion), &
                     chm_info%torsion_phi0(ntorsion)
                  ! Do some units conversion into internal atomic units
                  chm_info%torsion_phi0(ntorsion) = cp_unit_to_cp2k(chm_info%torsion_phi0(ntorsion), &
                                                                    "deg")
                  chm_info%torsion_k(ntorsion) = cp_unit_to_cp2k(chm_info%torsion_k(ntorsion), "kcalmol")
                  CALL charmm_get_next_line(parser, 1)
               END DO
            ELSE
               EXIT
            END IF
         END DO
         CALL parser_release(parser)
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Read in all the Improper info from the param file here
      !      Vimpr = Kpsi(psi-psi0)^2
      !      UNITS for Kpsi: [(kcal/mol)/(rad^2)] to [Eh/(rad^2)]
      !-----------------------------------------------------------------------------
      nimpr = 0
      DO ilab = 1, SIZE(impr_section)
         CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
         label = TRIM(impr_section(ilab))
         DO
            CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
            IF (found) THEN
               CALL parser_get_object(parser, string)
               IF (INDEX(string, TRIM(label)) /= 1) CYCLE
               CALL charmm_get_next_line(parser, 1)
               DO
                  CALL parser_get_object(parser, string)
                  CALL uppercase(string)
                  IF (ANY(string == avail_section)) EXIT
                  CALL parser_get_object(parser, string2)
                  CALL parser_get_object(parser, string3)
                  CALL parser_get_object(parser, string4)
                  CALL uppercase(string2)
                  CALL uppercase(string3)
                  CALL uppercase(string4)
                  nimpr = nimpr+1
                  CALL reallocate(chm_info%impr_a, 1, nimpr)
                  CALL reallocate(chm_info%impr_b, 1, nimpr)
                  CALL reallocate(chm_info%impr_c, 1, nimpr)
                  CALL reallocate(chm_info%impr_d, 1, nimpr)
                  CALL reallocate(chm_info%impr_k, 1, nimpr)
                  CALL reallocate(chm_info%impr_phi0, 1, nimpr)
                  chm_info%impr_a(nimpr) = string
                  chm_info%impr_b(nimpr) = string2
                  chm_info%impr_c(nimpr) = string3
                  chm_info%impr_d(nimpr) = string4
                  CALL parser_get_object(parser, chm_info%impr_k(nimpr))
                  CALL parser_get_object(parser, dummy)
                  CALL parser_get_object(parser, chm_info%impr_phi0(nimpr))
                  IF (iw > 0) WRITE (iw, *) "    CHM IMPROPERS ", nimpr, &
                     TRIM(chm_info%impr_a(nimpr)), " ", &
                     TRIM(chm_info%impr_b(nimpr)), " ", &
                     TRIM(chm_info%impr_c(nimpr)), " ", &
                     TRIM(chm_info%impr_d(nimpr)), " ", &
                     chm_info%impr_k(nimpr), &
                     chm_info%impr_phi0(nimpr)
                  ! Do some units conversion into internal atomic units
                  chm_info%impr_phi0(nimpr) = cp_unit_to_cp2k(chm_info%impr_phi0(nimpr), "deg")
                  chm_info%impr_k(nimpr) = cp_unit_to_cp2k(chm_info%impr_k(nimpr), "kcalmol")
                  CALL charmm_get_next_line(parser, 1)
               END DO
            ELSE
               EXIT
            END IF
         END DO
         CALL parser_release(parser)
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Read in all the Nonbonded info from the param file here
      !-----------------------------------------------------------------------------
      nnonbond = 0
      nonfo = 0
      DO ilab = 1, SIZE(nbon_section)
         CALL parser_create(parser, ff_type%ff_file_name, para_env=para_env)
         label = TRIM(nbon_section(ilab))
         DO
            CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
            IF (found) THEN
               CALL parser_get_object(parser, string)
               IF (INDEX(string, TRIM(label)) /= 1) CYCLE
               CALL charmm_get_next_line(parser, 1)
               DO
                  CALL parser_get_object(parser, string)
                  CALL uppercase(string)
                  IF (ANY(string == avail_section)) EXIT
                  nnonbond = nnonbond+1
                  CALL reallocate(chm_info%nonbond_a, 1, nnonbond)
                  CALL reallocate(chm_info%nonbond_eps, 1, nnonbond)
                  CALL reallocate(chm_info%nonbond_rmin2, 1, nnonbond)
                  chm_info%nonbond_a(nnonbond) = string
                  CALL parser_get_object(parser, chm_info%nonbond_eps(nnonbond))
                  CALL parser_get_object(parser, chm_info%nonbond_eps(nnonbond))
                  CALL parser_get_object(parser, chm_info%nonbond_rmin2(nnonbond))
                  IF (iw > 0) WRITE (iw, *) "    CHM NONBOND ", nnonbond, &
                     TRIM(chm_info%nonbond_a(nnonbond)), " ", &
                     chm_info%nonbond_eps(nnonbond), &
                     chm_info%nonbond_rmin2(nnonbond)
                  chm_info%nonbond_rmin2(nnonbond) = cp_unit_to_cp2k(chm_info%nonbond_rmin2(nnonbond), &
                                                                     "angstrom")
                  chm_info%nonbond_eps(nnonbond) = cp_unit_to_cp2k(chm_info%nonbond_eps(nnonbond), &
                                                                   "kcalmol")
                  IF (parser_test_next_token(parser) == "FLT") THEN
                     nonfo = nonfo+1
                     CALL reallocate(chm_info%nonbond_a_14, 1, nonfo)
                     CALL reallocate(chm_info%nonbond_eps_14, 1, nonfo)
                     CALL reallocate(chm_info%nonbond_rmin2_14, 1, nonfo)
                     chm_info%nonbond_a_14(nonfo) = chm_info%nonbond_a(nnonbond)
                     CALL parser_get_object(parser, chm_info%nonbond_eps_14(nonfo))
                     CALL parser_get_object(parser, chm_info%nonbond_eps_14(nonfo))
                     CALL parser_get_object(parser, chm_info%nonbond_rmin2_14(nonfo))
                     IF (iw > 0) WRITE (iw, *) "    CHM ONFO ", nonfo, &
                        TRIM(chm_info%nonbond_a_14(nonfo)), " ", &
                        chm_info%nonbond_eps_14(nonfo), &
                        chm_info%nonbond_rmin2_14(nonfo)
                     chm_info%nonbond_rmin2_14(nonfo) = cp_unit_to_cp2k(chm_info%nonbond_rmin2_14(nonfo), &
                                                                        "angstrom")
                     chm_info%nonbond_eps_14(nonfo) = cp_unit_to_cp2k(chm_info%nonbond_eps_14(nonfo), &
                                                                      "kcalmol")
                  END IF
                  CALL charmm_get_next_line(parser, 1)
               END DO
            ELSE
               EXIT
            END IF
         END DO
         CALL parser_release(parser)
      END DO
      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%FF_INFO")
      CALL timestop(handle)

   END SUBROUTINE read_force_field_charmm

! **************************************************************************************************
!> \brief Reads the AMBER force_field
!> \param ff_type ...
!> \param para_env ...
!> \param mm_section ...
!> \param particle_set ...
!> \author Teodoro Laino [tlaino, teodoro.laino-AT-gmail.com] - 11.2008
! **************************************************************************************************
   SUBROUTINE read_force_field_amber(ff_type, para_env, mm_section, particle_set)

      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'read_force_field_amber', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw
      TYPE(amber_info_type), POINTER                     :: amb_info
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")

      amb_info => ff_type%amb_info

      ! Read the Amber topology file to gather the forcefield information
      CALL rdparm_amber_8(ff_type%ff_file_name, iw, para_env, do_connectivity=.FALSE., &
                          do_forcefield=.TRUE., particle_set=particle_set, amb_info=amb_info)

      !-----------------------------------------------------------------------------
      ! 1. Converts all the Bonds info from the param file here
      !      Vbond = Kb(b-b0)^2
      !      UNITS for Kb: [(kcal/mol)/(A^2)] to [Eh/(AU^2)]
      !-----------------------------------------------------------------------------
      DO i = 1, SIZE(amb_info%bond_a)
         IF (iw > 0) WRITE (iw, *) "    AMB BOND ", i, &
            TRIM(amb_info%bond_a(i)), " ", &
            TRIM(amb_info%bond_b(i)), " ", &
            amb_info%bond_k(i), &
            amb_info%bond_r0(i)

         ! Do some units conversion into internal atomic units
         amb_info%bond_r0(i) = cp_unit_to_cp2k(amb_info%bond_r0(i), "angstrom")
         amb_info%bond_k(i) = cp_unit_to_cp2k(amb_info%bond_k(i), "kcalmol*angstrom^-2")
      END DO

      !-----------------------------------------------------------------------------
      ! 2. Converts all the Bends info from the param file here
      !      Vangle = Ktheta(theta-theta0)^2
      !      UNITS for Ktheta: [(kcal/mol)/(rad^2)] to [Eh/(rad^2)]
      !      FACTOR of "2" rolled into Ktheta
      !      Vub = Kub(S-S0)^2
      !      UNITS for Kub: [(kcal/mol)/(A^2)] to [Eh/(AU^2)]
      !-----------------------------------------------------------------------------
      DO i = 1, SIZE(amb_info%bend_a)
         IF (iw > 0) WRITE (iw, *) "    AMB BEND ", i, &
            TRIM(amb_info%bend_a(i)), " ", &
            TRIM(amb_info%bend_b(i)), " ", &
            TRIM(amb_info%bend_c(i)), " ", &
            amb_info%bend_k(i), &
            amb_info%bend_theta0(i)

         ! Do some units conversion into internal atomic units
         amb_info%bend_theta0(i) = cp_unit_to_cp2k(amb_info%bend_theta0(i), "rad")
         amb_info%bend_k(i) = cp_unit_to_cp2k(amb_info%bend_k(i), "kcalmol*rad^-2")
      END DO

      !-----------------------------------------------------------------------------
      ! 3. Converts all the Dihedrals info from the param file here
      !      Vtorsion = Kphi(1+COS(n(phi)-delta))
      !      UNITS for Kphi: [(kcal/mol)] to [Eh]
      !-----------------------------------------------------------------------------
      DO i = 1, SIZE(amb_info%torsion_a)
         IF (iw > 0) WRITE (iw, *) "    AMB TORSION ", i, &
            TRIM(amb_info%torsion_a(i)), " ", &
            TRIM(amb_info%torsion_b(i)), " ", &
            TRIM(amb_info%torsion_c(i)), " ", &
            TRIM(amb_info%torsion_d(i)), " ", &
            amb_info%torsion_k(i), &
            amb_info%torsion_m(i), &
            amb_info%torsion_phi0(i)

         ! Do some units conversion into internal atomic units
         amb_info%torsion_phi0(i) = cp_unit_to_cp2k(amb_info%torsion_phi0(i), "rad")
         amb_info%torsion_k(i) = cp_unit_to_cp2k(amb_info%torsion_k(i), "kcalmol")
      END DO

      !-----------------------------------------------------------------------------
      ! 4. Converts all the Nonbonded info from the param file here
      !-----------------------------------------------------------------------------
      DO i = 1, SIZE(amb_info%nonbond_eps)
         IF (iw > 0) WRITE (iw, *) "    AMB NONBOND ", i, &
            TRIM(amb_info%nonbond_a(i)), " ", &
            amb_info%nonbond_eps(i), &
            amb_info%nonbond_rmin2(i)

         ! Do some units conversion into internal atomic units
         amb_info%nonbond_rmin2(i) = cp_unit_to_cp2k(amb_info%nonbond_rmin2(i), "angstrom")
         amb_info%nonbond_eps(i) = cp_unit_to_cp2k(amb_info%nonbond_eps(i), "kcalmol")
      END DO
      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%FF_INFO")
      CALL timestop(handle)
   END SUBROUTINE read_force_field_amber

! **************************************************************************************************
!> \brief This function is simply a wrap to the parser_get_next_line..
!>      Comments: This routine would not be necessary if the continuation
!>                char for CHARMM would not be the "-".. How can you choose this
!>                character in a file of numbers as a continuation char????
!>                This sounds simply crazy....
!> \param parser ...
!> \param nline ...
!> \author Teodoro Laino - Zurich University - 06.2007
! **************************************************************************************************
   SUBROUTINE charmm_get_next_line(parser, nline)
      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN)                                :: nline

      CHARACTER(len=*), PARAMETER :: routineN = 'charmm_get_next_line', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), PARAMETER                        :: continuation_char = "-"

      INTEGER                                            :: i, len_line

      DO i = 1, nline
         len_line = LEN_TRIM(parser%input_line)
         DO WHILE (parser%input_line(len_line:len_line) == continuation_char)
            CALL parser_get_next_line(parser, 1)
            len_line = LEN_TRIM(parser%input_line)
         END DO
         CALL parser_get_next_line(parser, 1)
      END DO

   END SUBROUTINE charmm_get_next_line

END MODULE force_fields_ext
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Subroutine input_torsions changed (DG) 05-Dec-2000
!>      Output formats changed (DG) 05-Dec-2000
!>      JGH (26-01-2002) : force field parameters stored in tables, not in
!>        matrices. Input changed to have parameters labeled by the position
!>        and not atom pairs (triples etc)
!>      Teo (11.2005) : Moved all information on force field  pair_potential to
!>                      a much lighter memory structure
!> \author CJM
! **************************************************************************************************
MODULE force_fields
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE ewald_environment_types,         ONLY: ewald_environment_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
   USE force_field_kind_types,          ONLY: do_ff_amber,&
                                              do_ff_charmm,&
                                              do_ff_g87,&
                                              do_ff_g96,&
                                              do_ff_undef
   USE force_field_types,               ONLY: deallocate_ff_type,&
                                              force_field_type,&
                                              init_ff_type
   USE force_fields_ext,                ONLY: read_force_field_amber,&
                                              read_force_field_charmm,&
                                              read_force_field_gromos
   USE force_fields_input,              ONLY: read_force_field_section
   USE force_fields_util,               ONLY: clean_intra_force_kind,&
                                              force_field_pack,&
                                              force_field_qeff_output
   USE input_constants,                 ONLY: do_skip_14
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_fields'

  PRIVATE
  PUBLIC :: force_field_control

CONTAINS

! **************************************************************************************************
!> \brief 1. If reading in from external file, make sure its there first
!>      2. Read in the force_field from the corresponding locations
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ewald_env ...
!> \param fist_nonbond_env ...
!> \param root_section ...
!> \param para_env ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param subsys_section ...
!> \param mm_section ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param cell ...
! **************************************************************************************************
  SUBROUTINE force_field_control(atomic_kind_set, particle_set, &
       molecule_kind_set, molecule_set, ewald_env, fist_nonbond_env, &
       root_section, para_env, qmmm, qmmm_env, subsys_section, mm_section, &
       shell_particle_set, core_particle_set, cell)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section, mm_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_particle_set, core_particle_set
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: exclude_ei, exclude_vdw, handle, iw
      LOGICAL                                            :: found
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(force_field_type)                             :: ff_type
      TYPE(section_vals_type), POINTER                   :: topology_section

    CALL timeset(routineN,handle)
    logger => cp_get_default_logger()

    iw = cp_print_key_unit_nr(logger,mm_section,"PRINT%FF_INFO",&
         extension=".mmLog")

    !-----------------------------------------------------------------------------
    ! 1. Initialize the ff_type structure type
    !-----------------------------------------------------------------------------
    CALL init_ff_type(ff_type)

    !-----------------------------------------------------------------------------
    ! 2. Read in the force field section in the input file if any
    !-----------------------------------------------------------------------------
    CALL read_force_field_section(ff_type,para_env,mm_section)

    !-----------------------------------------------------------------------------
    ! 2.1 In case exclusion 1-4 was requested, we need to modify the values of
    !     the scale factors setting them to zero..
    !-----------------------------------------------------------------------------
    topology_section => section_vals_get_subs_vals(subsys_section,"TOPOLOGY")
    CALL section_vals_val_get(topology_section,"EXCLUDE_VDW",i_val=exclude_vdw)
    CALL section_vals_val_get(topology_section,"EXCLUDE_EI",i_val=exclude_ei)
    IF (exclude_vdw==do_skip_14) ff_type%vdw_scale14 = 0.0_dp
    IF (exclude_ei ==do_skip_14) ff_type%ei_scale14  = 0.0_dp

    !-----------------------------------------------------------------------------
    ! 3. If reading in from external file, make sure its there first
    !-----------------------------------------------------------------------------
    SELECT CASE (ff_type%ff_type)
    CASE (do_ff_charmm,do_ff_amber,do_ff_g96,do_ff_g87)
       INQUIRE(FILE=ff_type%ff_file_name,EXIST=found)
       IF (.NOT.found) THEN
          CALL cp__b("force_fields.F",145,"Force field file missing")
       END IF
    CASE (do_ff_undef)
       ! Do Nothing
    CASE DEFAULT
       CALL cp__b("force_fields.F",150,"Force field type not implemented")
    END SELECT

    !-----------------------------------------------------------------------------
    ! 4. Read in the force field from the corresponding locations
    !-----------------------------------------------------------------------------
    SELECT CASE (ff_type%ff_type)
    CASE (do_ff_charmm)
       CALL read_force_field_charmm(ff_type,para_env,mm_section)
    CASE (do_ff_amber)
       CALL read_force_field_amber(ff_type,para_env,mm_section,particle_set)
    CASE (do_ff_g87,do_ff_g96)
       CALL read_force_field_gromos(ff_type,para_env,mm_section)
    CASE (do_ff_undef)
       ! Do Nothing
    CASE DEFAULT
       CALL cp__b("force_fields.F",166,"Force field type not implemented")
    END SELECT

    !-----------------------------------------------------------------------------
    ! 5. Possibly print the top file
    !-----------------------------------------------------------------------------
    CALL print_pot_parameter_file(ff_type, mm_section)

    !-----------------------------------------------------------------------------
    ! 6. Pack all force field info into different structures
    !-----------------------------------------------------------------------------
    CALL force_field_pack (particle_set,atomic_kind_set, molecule_kind_set,molecule_set,&
         ewald_env,fist_nonbond_env,ff_type,root_section, qmmm, qmmm_env, mm_section,&
         subsys_section, shell_particle_set=shell_particle_set, &
         core_particle_set=core_particle_set, cell=cell)

    !-----------------------------------------------------------------------------
    ! 7. Output total system charge assigned to qeff
    !-----------------------------------------------------------------------------
    CALL force_field_qeff_output (particle_set, molecule_kind_set,&
         molecule_set, mm_section, fist_nonbond_env%charges)

    !-----------------------------------------------------------------------------
    ! 8. Clean up "UNSET" bond,bend,UB,TORSION,IMPR,ONFO kinds
    !-----------------------------------------------------------------------------
    CALL clean_intra_force_kind (molecule_kind_set,mm_section)

    !-----------------------------------------------------------------------------
    ! 9. Cleanup the ff_type structure type
    !-----------------------------------------------------------------------------
    CALL deallocate_ff_type(ff_type)

    CALL cp_print_key_finished_output(iw,logger,mm_section,&
         "PRINT%FF_INFO")
    CALL timestop(handle)

  END SUBROUTINE force_field_control

! **************************************************************************************************
!> \brief Prints force field information in a pot file
!> \param ff_type ...
!> \param mm_section ...
!> \author Teodoro Laino [tlaino, teodoro.laino-AT-gmail.com] - 11.2008
! **************************************************************************************************
  SUBROUTINE print_pot_parameter_file(ff_type, mm_section)

      TYPE(force_field_type)                             :: ff_type
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'print_pot_parameter_file', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw, m
      REAL(KIND=dp)                                      :: eps, k, phi0, r0, sigma, theta0
      TYPE(cp_logger_type), POINTER                      :: logger

    CALL timeset(routineN,handle)
    logger => cp_get_default_logger()
    IF (BTEST(cp_print_key_should_output(logger%iter_info,mm_section,"PRINT%FF_PARAMETER_FILE")&
         ,cp_p_file)) THEN
       iw = cp_print_key_unit_nr(logger,mm_section,"PRINT%FF_PARAMETER_FILE",&
            middle_name="force_field",extension=".pot")
       IF (iw>0) THEN
          ! Header
          WRITE(iw,1000)"Force Field Parameter File dumped into CHARMM FF style"
       END IF
       SELECT CASE(ff_type%ff_type)
       CASE (do_ff_charmm)
          CALL cp__w("force_fields.F",234,"Dumping FF parameter file for CHARMM FF  not implemented!")
       CASE (do_ff_amber)
          IF (iw>0) THEN
             ! Bonds
             WRITE(iw,1001)
             DO i = 1, SIZE(ff_type%amb_info%bond_a)
                k = cp_unit_from_cp2k(ff_type%amb_info%bond_k(i),"kcalmol*angstrom^-2")
                r0= cp_unit_from_cp2k(ff_type%amb_info%bond_r0(i),"angstrom")
                WRITE(iw, 2001) ff_type%amb_info%bond_a(i),&
                                ff_type%amb_info%bond_b(i),&
                                k, r0
             END DO
             ! Angles
             WRITE(iw,1002)
             DO i = 1, SIZE(ff_type%amb_info%bend_a)
                k      = cp_unit_from_cp2k(ff_type%amb_info%bend_k(i),"kcalmol*rad^-2")
                theta0 = cp_unit_from_cp2k(ff_type%amb_info%bend_theta0(i),"deg")
                WRITE(iw, 2002) ff_type%amb_info%bend_a(i),&
                                ff_type%amb_info%bend_b(i),&
                                ff_type%amb_info%bend_c(i),&
                                k, theta0
             END DO
             ! Torsions
             WRITE(iw,1003)
             DO i = 1, SIZE(ff_type%amb_info%torsion_a)
                k    = cp_unit_from_cp2k(ff_type%amb_info%torsion_k(i),"kcalmol")
                m    = ff_type%amb_info%torsion_m(i)
                phi0 = cp_unit_from_cp2k(ff_type%amb_info%torsion_phi0(i),"deg")
                WRITE(iw, 2003) ff_type%amb_info%torsion_a(i),&
                                ff_type%amb_info%torsion_b(i),&
                                ff_type%amb_info%torsion_c(i),&
                                ff_type%amb_info%torsion_d(i),&
                                k, m, phi0
             END DO
             ! Lennard-Jones
             WRITE(iw,1005)
             DO i = 1, SIZE(ff_type%amb_info%nonbond_a)
                eps   = cp_unit_from_cp2k(ff_type%amb_info%nonbond_eps(i),"kcalmol")
                sigma = cp_unit_from_cp2k(ff_type%amb_info%nonbond_rmin2(i),"angstrom")
                WRITE(iw, 2005) ff_type%amb_info%nonbond_a(i),&
                                eps, sigma
             END DO
          END IF
       CASE (do_ff_g87,do_ff_g96)
          CALL cp__w("force_fields.F",278,"Dumping FF parameter file for GROMOS FF not implemented!")
       CASE (do_ff_undef)
          CALL cp__w("force_fields.F",280,"Dumping FF parameter file for INPUT FF  not implemented!")
       END SELECT
       IF (iw>0) THEN
          WRITE(iw,'(/,A)')"END"
       END IF
       CALL cp_print_key_finished_output(iw,logger,mm_section,&
            "PRINT%FF_PARAMETER_FILE")
    END IF
    CALL timestop(handle)
    RETURN
1000 FORMAT("*>>>>>>>",T12,A,T73,"<<<<<<<")
1001 FORMAT(/,"BONDS",/,"!",/,"!V(bond) = Kb(b - b0)**2",/,"!",/,"!Kb: kcal/mole/A**2",/, &
         "!b0: A",/,"!",/,"! atom type           Kb              b0",/,"!")
1002 FORMAT(/,"ANGLES",/,"!",/,"!V(angle) = Ktheta(Theta - Theta0)**2",/,"!",/,           &
         "!V(Urey-Bradley) = Kub(S - S0)**2",/,"!",/,"!Ktheta: kcal/mole/rad**2",/,  &
         "!Theta0: degrees",/,"!Kub: kcal/mole/A**2 (Urey-Bradley)",/,"!S0: A",/,    &
         "!",/,"!   atom types              Ktheta          Theta0       Kub        S0",/,"!")
1003 FORMAT(/,"DIHEDRALS",/,"!",/,"!V(dihedral) = Kchi(1 + cos(n(chi) - delta))",/,       &
         "!",/,"!Kchi: kcal/mole",/,"!n: multiplicity",/,"!delta: degrees",/,        &
         "!",/,"!     atom types                    Kchi       n       delta",/,"!")
1005 FORMAT(/,"NONBONDED",/,"!",/,                                                        &
         "!V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]",/, &
         "!",/,"!epsilon: kcal/mole, Eps,i,j = sqrt(eps,i * eps,j)",/,               &
         "!Rmin/2: A, Rmin,i,j = Rmin/2,i + Rmin/2,j",/,"!",/,                       &
         "!atom         ignored        epsilon       Rmin/2      ignored   eps,1-4       "//&
         "Rmin/2,1-4",/,"!")

2001 FORMAT(A6,1X,A6,1X,2F15.9)                     ! bond
2002 FORMAT(A6,1X,A6,1X,A6,1X,2F15.9)               ! angle
2003 FORMAT(A6,1X,A6,1X,A6,1X,A6,1X,F15.9,I5,F15.9) ! torsion
2005 FORMAT(A6,1X,"    0.000000000",2F15.9)         ! nonbond
  END SUBROUTINE print_pot_parameter_file

END MODULE force_fields
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief initialize fist environment
!> \author CJM
! **************************************************************************************************
MODULE fist_environment
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE bibliography,                    ONLY: Devynck2012,&
                                              Dick1958,&
                                              Mitchell1993,&
                                              cite_reference
   USE cell_methods,                    ONLY: read_cell,&
                                              write_cell
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: cp_subsys_create
   USE cp_subsys_types,                 ONLY: cp_subsys_release,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE cp_symmetry,                     ONLY: write_symmetry
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_type
   USE distribution_methods,            ONLY: distribute_molecules_1d
   USE ewald_environment_types,         ONLY: ewald_env_create,&
                                              ewald_env_get,&
                                              ewald_env_release,&
                                              ewald_env_set,&
                                              ewald_environment_type,&
                                              read_ewald_section
   USE ewald_pw_methods,                ONLY: ewald_pw_grid_update
   USE ewald_pw_types,                  ONLY: ewald_pw_create,&
                                              ewald_pw_release,&
                                              ewald_pw_type
   USE exclusion_types,                 ONLY: exclusion_type
   USE fist_energy_types,               ONLY: allocate_fist_energy,&
                                              fist_energy_type
   USE fist_environment_types,          ONLY: fist_env_get,&
                                              fist_env_set,&
                                              fist_environment_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_release,&
                                              fist_nonbond_env_type
   USE force_fields,                    ONLY: force_field_control
   USE header,                          ONLY: fist_header
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type,&
                                              write_molecule_kind_set
   USE molecule_types_new,              ONLY: molecule_type
   USE multipole_types,                 ONLY: create_multipole_type,&
                                              multipole_type,&
                                              release_multipole_type
   USE particle_list_types,             ONLY: particle_list_create,&
                                              particle_list_release,&
                                              particle_list_type
   USE particle_methods,                ONLY: write_fist_particle_coordinates,&
                                              write_particle_distances,&
                                              write_structure_data
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_environment'
   PUBLIC :: fist_init

CONTAINS
! **************************************************************************************************
!> \brief reads the input and database file for fist
!> \param fist_env ...
!> \param root_section ...
!> \param para_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param prev_subsys ...
!> \par Used By
!>      fist_main
! **************************************************************************************************
   SUBROUTINE fist_init(fist_env, root_section, para_env, force_env_section, &
                        subsys_section, use_motion_section, prev_subsys)

      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: prev_subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: qmmm, shell_adiabatic, shell_present
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(exclusion_type), DIMENSION(:), POINTER        :: exclusions
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(particle_list_type), POINTER                  :: core_particles, shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, shell_particle_set
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: cell_section, ewald_section, mm_section, &
                                                            poisson_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      NULLIFY (subsys, cell, cell_ref)
      NULLIFY (ewald_env, fist_nonbond_env, qmmm_env, cell_section, &
               poisson_section, shell_particle_set, shell_particles, &
               core_particle_set, core_particles, exclusions)
      IF (.NOT. ASSOCIATED(subsys_section)) THEN
         subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      END IF
      mm_section => section_vals_get_subs_vals(force_env_section, "MM")
      cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
      poisson_section => section_vals_get_subs_vals(mm_section, "POISSON")
      ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD")

      CALL fist_env_set(fist_env, input=force_env_section)

      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%PROGRAM_BANNER", &
                                extension=".mmLog")
      CALL fist_header(iw)
      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%PROGRAM_BANNER")

      CALL read_cell(cell, cell_ref, cell_section=cell_section, para_env=para_env)
      CALL get_cell(cell, abc=abc)

      ! Print the cell parameters
      CALL write_cell(cell, subsys_section, cell_ref)

      ! Create the ewald environment
      CALL ewald_env_create(ewald_env, para_env)

      ! Read the input section and set the ewald_env
      CALL read_ewald_section(ewald_env, ewald_section)
      CALL ewald_env_set(ewald_env, poisson_section=poisson_section)

      ! Topology
      CALL fist_env_get(fist_env, qmmm=qmmm, qmmm_env=qmmm_env)
      CALL cp_subsys_create(subsys, para_env=para_env, root_section=root_section, &
                            force_env_section=force_env_section, subsys_section=subsys_section, &
                            qmmm=qmmm, qmmm_env=qmmm_env, exclusions=exclusions, &
                            use_motion_section=use_motion_section)
      CALL fist_env_set(fist_env, subsys=subsys, exclusions=exclusions)

      CALL force_field_control(subsys%atomic_kinds%els, subsys%particles%els, &
                               subsys%molecule_kinds_new%els, subsys%molecules_new%els, &
                               ewald_env, fist_nonbond_env, root_section, para_env, qmmm=qmmm, &
                               qmmm_env=qmmm_env, subsys_section=subsys_section, &
                               mm_section=mm_section, shell_particle_set=shell_particle_set, &
                               core_particle_set=core_particle_set, cell=cell)

      NULLIFY (shell_particles, core_particles)
      IF (ASSOCIATED(shell_particle_set)) THEN
         CALL cite_reference(Devynck2012)
         CALL cite_reference(Mitchell1993)
         CALL cite_reference(Dick1958)
         CALL particle_list_create(shell_particles, els_ptr=shell_particle_set)
      END IF
      IF (ASSOCIATED(core_particle_set)) THEN
         CALL particle_list_create(core_particles, els_ptr=core_particle_set)
      END IF
      CALL get_atomic_kind_set(atomic_kind_set=subsys%atomic_kinds%els, &
                               shell_present=shell_present, shell_adiabatic=shell_adiabatic)
      CALL fist_env_set(fist_env, shell_model=shell_present, &
                        shell_model_ad=shell_adiabatic)
      CALL cp_subsys_set(subsys, shell_particles=shell_particles, &
                         core_particles=core_particles)
      CALL particle_list_release(shell_particles)
      CALL particle_list_release(core_particles)

      CALL fist_init_subsys(fist_env, subsys, cell, cell_ref, fist_nonbond_env, ewald_env, &
                            force_env_section, subsys_section, prev_subsys)

      CALL cell_release(cell)
      CALL cell_release(cell_ref)
      CALL ewald_env_release(ewald_env)
      CALL fist_nonbond_env_release(fist_nonbond_env)
      CALL cp_subsys_release(subsys)

      CALL timestop(handle)

   END SUBROUTINE fist_init

! **************************************************************************************************
!> \brief   Read the input and the database files for the setup of the
!>          FIST environment.
!> \param fist_env ...
!> \param subsys ...
!> \param cell ...
!> \param cell_ref ...
!> \param fist_nonbond_env ...
!> \param ewald_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param prev_subsys ...
!> \date    22.05.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE fist_init_subsys(fist_env, subsys, cell, cell_ref, fist_nonbond_env, &
                               ewald_env, force_env_section, subsys_section, &
                               prev_subsys)

      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: prev_subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_init_subsys', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, max_multipole
      LOGICAL                                            :: do_multipoles
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles, &
                                                            prev_local_molecules
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_energy_type), POINTER                    :: thermo
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set, prev_molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: grid_print_section

      CALL timeset(routineN, handle)
      NULLIFY (thermo, ewald_pw, local_molecules, local_particles, multipoles)
      particle_set => subsys%particles%els
      atomic_kind_set => subsys%atomic_kinds%els
      molecule_set => subsys%molecules_new%els
      molecule_kind_set => subsys%molecule_kinds_new%els

      IF (PRESENT(prev_subsys)) THEN
         prev_molecule_kind_set => prev_subsys%molecule_kinds_new%els
         prev_local_molecules => prev_subsys%local_molecules_new
      ELSE
         NULLIFY (prev_molecule_kind_set)
         NULLIFY (prev_local_molecules)
      ENDIF

      ! Create the fist_energy_type
      CALL allocate_fist_energy(thermo)

      ! Print the molecule kind set
      CALL write_molecule_kind_set(molecule_kind_set, subsys_section)

      ! Print the atomic coordinates
      CALL write_fist_particle_coordinates(particle_set, subsys_section, &
                                           fist_nonbond_env%charges)
      CALL write_particle_distances(particle_set, cell, subsys_section)
      CALL write_structure_data(particle_set, cell=cell, input_section=subsys_section)

      ! Print symmetry information
      CALL write_symmetry(particle_set, cell, subsys_section)

      ! Distribute molecules and atoms using the new data structures ***
      CALL distribute_molecules_1d(atomic_kind_set=atomic_kind_set, &
                                   particle_set=particle_set, &
                                   local_particles=local_particles, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   local_molecules=local_molecules, &
                                   prev_molecule_kind_set=prev_molecule_kind_set, &
                                   prev_local_molecules=prev_local_molecules, &
                                   force_env_section=force_env_section)

      ! Create ewald grids
      grid_print_section => section_vals_get_subs_vals(force_env_section, &
                                                       "PRINT%GRID_INFORMATION")
      CALL ewald_pw_create(ewald_pw, ewald_env, cell, cell_ref, grid_print_section)

      ! Initialize ewald grids
      CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)

      ! Possibly Initialize the multipole environment
      CALL ewald_env_get(ewald_env, do_multipoles=do_multipoles, &
                         max_multipole=max_multipole)
      IF (do_multipoles) &
         CALL create_multipole_type(multipoles, particle_set, subsys_section, max_multipole)
      CALL cp_subsys_set(subsys, multipoles=multipoles, cell=cell)

      ! Set the fist_env
      CALL fist_env_set(fist_env=fist_env, &
                        cell_ref=cell_ref, &
                        local_molecules=local_molecules, &
                        local_particles=local_particles, &
                        ewald_env=ewald_env, ewald_pw=ewald_pw, &
                        fist_nonbond_env=fist_nonbond_env, &
                        thermo=thermo)

      CALL distribution_1d_release(local_particles)
      CALL distribution_1d_release(local_molecules)
      CALL ewald_pw_release(ewald_pw)
      CALL release_multipole_type(multipoles)
      CALL timestop(handle)

   END SUBROUTINE fist_init_subsys
END MODULE fist_environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief sets variables for the qmmm pool of pw_types
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_pw_grid
   USE input_constants,                 ONLY: do_par_atom,&
                                              do_qmmm_gauss,&
                                              do_qmmm_swave
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_comm_dup,&
                                              mp_environ
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grid_types,                   ONLY: FULLSPACE,&
                                              PW_MODE_DISTRIBUTED,&
                                              PW_MODE_LOCAL,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_release
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_p_type,&
                                              pw_pool_type,&
                                              pw_pools_dealloc
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: qmmm_pw_grid_init
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_pw_grid'
   INTEGER :: qmmm_grid_tag = 0

CONTAINS

! **************************************************************************************************
!> \brief Initialize the qmmm pool of pw_type.
!>      Then Main difference w.r.t. QS pw_type pools is that this pool
!>      has [0,L] as boundaries.
!> \param qmmm_env ...
!> \param pw_env ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_pw_grid_init(qmmm_env, pw_env)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(pw_env_type), POINTER                         :: pw_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_pw_grid_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: auxbas_grid, Ilevel, pw_mode
      REAL(KIND=dp), DIMENSION(3)                        :: Maxdr, Mindr
      TYPE(pw_grid_type), POINTER                        :: el_struct
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: pool

      NULLIFY (el_struct)
      Maxdr = TINY(0.0_dp)
      Mindr = HUGE(0.0_dp)
      IF ((qmmm_env%qmmm_coupl_type == do_qmmm_gauss) .OR. (qmmm_env%qmmm_coupl_type == do_qmmm_swave)) THEN
         CALL pw_env_get(pw_env=pw_env, &
                         pw_pools=pw_pools, &
                         auxbas_grid=auxbas_grid)
         !
         IF (ASSOCIATED(qmmm_env%aug_pools)) THEN
            CALL pw_pools_dealloc(qmmm_env%aug_pools)
         END IF
         ALLOCATE (qmmm_env%aug_pools(SIZE(pw_pools)))
         !
         DO Ilevel = 1, SIZE(pw_pools)
            NULLIFY (pool, qmmm_env%aug_pools(Ilevel)%pool)
            pool => pw_pools(Ilevel)%pool
            NULLIFY (el_struct)
            pw_mode = PW_MODE_DISTRIBUTED
            ! Parallelization scheme
            IF (qmmm_env%par_scheme == do_par_atom) THEN
               pw_mode = PW_MODE_LOCAL
            END IF

            CALL pw_grid_create_copy_no_pbc(pool%pw_grid, el_struct, &
                                            pw_mode=pw_mode)
            CALL pw_pool_create(qmmm_env%aug_pools(Ilevel)%pool, &
                                pw_grid=el_struct)

            Maxdr = MAX(Maxdr, el_struct%dr)
            Mindr = MIN(Mindr, el_struct%dr)
            IF (ALL(Maxdr .EQ. el_struct%dr)) qmmm_env%gridlevel_info%coarser_grid = Ilevel
            IF (ALL(Mindr .EQ. el_struct%dr)) qmmm_env%gridlevel_info%auxbas_grid = Ilevel

            CALL pw_grid_release(el_struct)

         END DO
      END IF

   END SUBROUTINE qmmm_pw_grid_init

! **************************************************************************************************
!> \brief creates a copy of pw_grid_in in which the pbc have been removed
!>      (by adding a point for the upper boundary)
!> \param pw_grid_in the pw grid to duplicate
!> \param pw_grid_out the output pw_grid_type
!> \param pw_mode ...
!> \par History
!>      08.2004 created [tlaino]
!>      04.2005 completly rewritten the duplicate routine, fixed parallel
!>              behaviour, narrowed scope to copy to non pbc and renamed
!>              accordingly [fawzi]
!>      06.2007 moved to new module [jgh]
!> \author Fawzi, Teo
! **************************************************************************************************
   SUBROUTINE pw_grid_create_copy_no_pbc(pw_grid_in, pw_grid_out, pw_mode)
      TYPE(pw_grid_type), POINTER                        :: pw_grid_in, pw_grid_out
      INTEGER, INTENT(IN), OPTIONAL                      :: pw_mode

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_create_copy_no_pbc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: pw_mode_loc
      INTEGER, DIMENSION(:), POINTER                     :: pos_of_x

      IF(.NOT.(pw_grid_in%ngpts_cut > 0))CALL cp__a("qmmm_pw_grid.F",129)
      IF(.NOT.(.NOT. ASSOCIATED(pw_grid_out)))CALL cp__a("qmmm_pw_grid.F",130)
      pw_mode_loc = pw_grid_in%para%mode
      IF (PRESENT(pw_mode)) pw_mode_loc = pw_mode
      CALL pw_grid_create(pw_grid_out, pw_grid_in%para%group)
      qmmm_grid_tag = qmmm_grid_tag+1
      pw_grid_out%id_nr = qmmm_grid_tag
      pw_grid_out%ref_count = 1
      pw_grid_out%reference = 0
      pw_grid_out%bounds = pw_grid_in%bounds
      pw_grid_out%bounds(2, :) = pw_grid_out%bounds(2, :)+1
      IF (pw_mode_loc == PW_MODE_DISTRIBUTED) THEN
         pw_grid_out%bounds_local = pw_grid_in%bounds_local
         IF (pw_grid_in%bounds_local(2, 1) == pw_grid_in%bounds(2, 1) .AND. &
             pw_grid_in%bounds_local(1, 1) <= pw_grid_in%bounds(2, 1)) THEN
            pw_grid_out%bounds_local(2, 1) = pw_grid_out%bounds_local(2, 1)+1
         END IF
         pw_grid_out%bounds_local(2, 2) = pw_grid_out%bounds_local(2, 2)+1
         pw_grid_out%bounds_local(2, 3) = pw_grid_out%bounds_local(2, 3)+1
      ELSE
         pw_grid_out%bounds_local = pw_grid_out%bounds
      END IF
      pw_grid_out%npts = pw_grid_in%npts+1
      pw_grid_out%ngpts = PRODUCT(INT(pw_grid_out%npts, KIND=int_8))
      pw_grid_out%ngpts_cut = 0
      pw_grid_out%npts_local = pw_grid_out%bounds_local(2, :)-pw_grid_out%bounds_local(1, :)+1
      pw_grid_out%ngpts_local = PRODUCT(pw_grid_out%npts_local)
      pw_grid_out%ngpts_cut_local = 0
      pw_grid_out%dr = pw_grid_in%dr
      pw_grid_out%dh = pw_grid_in%dh
      pw_grid_out%dh_inv = pw_grid_in%dh_inv
      pw_grid_out%orthorhombic = pw_grid_in%orthorhombic
      pw_grid_out%dvol = pw_grid_in%dvol
      pw_grid_out%vol = pw_grid_in%vol*REAL(pw_grid_out%ngpts, dp) &
                        /REAL(pw_grid_in%ngpts, dp) !FM do not modify?
      pw_grid_out%cutoff = pw_grid_in%cutoff
      NULLIFY (pw_grid_out%mapl%pos, pw_grid_out%mapl%neg, &
               pw_grid_out%mapm%pos, pw_grid_out%mapm%neg, &
               pw_grid_out%mapn%pos, pw_grid_out%mapn%neg)

      !para
      CALL mp_environ(pw_grid_out%para%group_size, &
                      pw_grid_out%para%my_pos, &
                      pw_grid_out%para%group)
      pw_grid_out%para%group_head_id = pw_grid_in%para%group_head_id
      pw_grid_out%para%group_head = &
         (pw_grid_out%para%group_head_id == pw_grid_out%para%my_pos)
      pw_grid_out%para%mode = pw_mode_loc
      NULLIFY (pw_grid_out%para%yzp, &
               pw_grid_out%para%yzq, &
               pw_grid_out%para%nyzray, &
               pw_grid_out%para%bo)
      ALLOCATE (pos_of_x(pw_grid_out%bounds(1, 1):pw_grid_out%bounds(2, 1)))
      pos_of_x(:pw_grid_out%bounds(2, 1)-1) = pw_grid_in%para%pos_of_x
      pos_of_x(pw_grid_out%bounds(2, 1)) = pos_of_x(pw_grid_out%bounds(2, 1)-1)
      pw_grid_out%para%pos_of_x => pos_of_x
      pw_grid_out%para%rs_dims = pw_grid_in%para%rs_dims
      IF (PRODUCT(pw_grid_in%para%rs_dims) /= 0) THEN
         CALL mp_comm_dup(pw_grid_in%para%rs_group, &
                          pw_grid_out%para%rs_group)
      END IF
      pw_grid_out%para%rs_pos = pw_grid_in%para%rs_pos
      pw_grid_out%para%rs_mpo = pw_grid_in%para%rs_mpo

      NULLIFY (pw_grid_out%g, pw_grid_out%gsq, pw_grid_out%g_hat)
      IF(.NOT.(pw_grid_in%grid_span == FULLSPACE))CALL cp__a("qmmm_pw_grid.F",194)
      pw_grid_out%grid_span = pw_grid_in%grid_span
      pw_grid_out%have_g0 = .FALSE.
      pw_grid_out%first_gne0 = HUGE(0)
      NULLIFY (pw_grid_out%gidx)
      pw_grid_out%spherical = .FALSE.
      pw_grid_out%para%ray_distribution = .FALSE.
      pw_grid_out%para%blocked = .FALSE.
   END SUBROUTINE pw_grid_create_copy_no_pbc
END MODULE qmmm_pw_grid
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize a QM/MM calculation
!> \par History
!>      5.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qmmm_create
   USE bibliography,                    ONLY: Golze2013,&
                                              Laino2005,&
                                              Laino2006,&
                                              cite_reference
   USE cell_methods,                    ONLY: write_cell
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: create_small_subsys
   USE cp_subsys_types,                 ONLY: cp_subsys_release,&
                                              cp_subsys_type
   USE fist_environment,                ONLY: fist_init
   USE fist_environment_types,          ONLY: fist_env_create,&
                                              fist_env_get,&
                                              fist_env_set,&
                                              fist_environment_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
   USE global_types,                    ONLY: global_environment_type
   USE header,                          ONLY: qmmm_header
   USE input_constants,                 ONLY: &
        do_fist, do_multipole_section_off, do_multipole_section_on, do_qmmm, &
        do_qmmm_center_every_step, do_qmmm_center_never, do_qmmm_center_pbc_aware, &
        do_qmmm_center_setup_only, do_qmmm_none, do_qs
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE pw_env_types,                    ONLY: pw_env_type
   USE qmmm_init,                       ONLY: &
        assign_mm_charges_and_radius, move_or_add_atoms, print_image_charge_info, &
        print_qmmm_charges, print_qmmm_links, qmmm_init_gaussian_type, &
        qmmm_init_periodic_potential, qmmm_init_potential, setup_origin_mm_cell, setup_qmmm_links, &
        setup_qmmm_vars_mm, setup_qmmm_vars_qm
   USE qmmm_links_methods,              ONLY: qmmm_link_Imomm_coord
   USE qmmm_pw_grid,                    ONLY: qmmm_pw_grid_init
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE qmmm_types_low,                  ONLY: &
        add_set_release, add_set_type, add_shell_type, qmmm_env_mm_create, qmmm_env_mm_release, &
        qmmm_env_mm_type, qmmm_env_qm_create, qmmm_env_qm_type, qmmm_links_type
   USE qs_environment,                  ONLY: qs_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_env_create,&
                                              qs_environment_type,&
                                              set_qs_env
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_create'

   PUBLIC :: qmmm_env_create

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param force_env_section ...
!> \param qmmm_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param prev_subsys ...
!> \param ignore_outside_box ...
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qmmm_env_create(qmmm_env, root_section, para_env, globenv, &
                              force_env_section, qmmm_section, subsys_section, use_motion_section, prev_subsys, &
                              ignore_outside_box)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: force_env_section, qmmm_section, &
                                                            subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: prev_subsys
      LOGICAL, INTENT(in), OPTIONAL                      :: ignore_outside_box

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: qm_atom_type
      INTEGER                                            :: center_i, delta_charge, handle, iw, iw2, &
                                                            orig_charge, qmmm_coupl_type, &
                                                            use_multipole
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index, mm_link_atoms, &
                                                            qm_atom_index
      LOGICAL                                            :: add_mm_charges, explicit, &
                                                            move_mm_charges, nocompatibility, &
                                                            qmmm_link, qmmm_link_Imomm, shell_model
      REAL(dp), DIMENSION(:), POINTER                    :: mm_atom_chrg, mm_el_pot_radius, &
                                                            mm_el_pot_radius_corr
      REAL(KIND=dp)                                      :: eps_mm_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: abc_mm, abc_qm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: fist_scale_charge_link, &
                                                            mm_link_scale_factor
      TYPE(add_set_type), POINTER                        :: added_charges
      TYPE(add_shell_type), POINTER                      :: added_shells
      TYPE(cell_type), POINTER                           :: mm_cell, qm_cell_small, super_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys_mm, subsys_qm
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env_mm
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env_qm
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: multipole_section, print_gen, &
                                                            print_section, qmmm_periodic

      CALL timeset(routineN, handle)

      NULLIFY (qm_atom_index, mm_atom_index, qm_atom_type)
      NULLIFY (qmmm_env_qm, subsys_mm, subsys_qm, mm_cell, qm_cell_small)
      NULLIFY (mm_atom_chrg, mm_el_pot_radius, qmmm_env_mm, fist_env)
      NULLIFY (qmmm_env)
      NULLIFY (mm_link_atoms, mm_link_scale_factor, qmmm_links, added_charges, added_shells)
      NULLIFY (fist_scale_charge_link, print_section, fist_nonbond_env)
      NULLIFY (print_gen, logger, mm_el_pot_radius_corr, super_cell, pw_env)

      logger => cp_get_default_logger()

      ! citations
      CALL cite_reference(Laino2005)

      ! Input section...
      IF (.NOT. ASSOCIATED(subsys_section)) THEN
         subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      END IF
      qmmm_periodic => section_vals_get_subs_vals(qmmm_section, "PERIODIC")
      multipole_section => section_vals_get_subs_vals(qmmm_section, "PERIODIC%MULTIPOLE")
      print_section => section_vals_get_subs_vals(qmmm_section, "PRINT")
      print_gen => section_vals_get_subs_vals(print_section, "PROGRAM_RUN_INFO")
      iw = cp_print_key_unit_nr(logger, print_gen, "", extension=".log")

      ! Create QM/MM Environments..
      CALL qmmm_env_qm_create(qmmm_env_qm)
      CALL qmmm_env_mm_create(qmmm_env_mm)

      ! Set up QM/MM Options
      CALL setup_qmmm_vars_mm(qmmm_section, &
                              qmmm_env_mm, &
                              qm_atom_index, &
                              mm_link_atoms, &
                              mm_link_scale_factor, &
                              fist_scale_charge_link, &
                              qmmm_coupl_type, &
                              qmmm_link)

      qmmm_env_mm%qm_atom_index => qm_atom_index
      qmmm_env_mm%mm_link_atoms => mm_link_atoms
      qmmm_env_mm%mm_link_scale_factor => mm_link_scale_factor
      qmmm_env_mm%fist_scale_charge_link => fist_scale_charge_link
      qmmm_env_mm%qmmm_coupl_type = qmmm_coupl_type
      qmmm_env_mm%qmmm_link = qmmm_link
      ! Center the qm subsys into the qm box
      CALL section_vals_val_get(qmmm_section, "CENTER", i_val=center_i)
      IF (center_i == do_qmmm_center_never) THEN
         qmmm_env_qm%center_qm_subsys = .FALSE.
         qmmm_env_qm%center_qm_subsys0 = .FALSE.
      ELSE IF (center_i == do_qmmm_center_setup_only) THEN
         qmmm_env_qm%center_qm_subsys = .FALSE.
         qmmm_env_qm%center_qm_subsys0 = .TRUE.
      ELSE IF (center_i == do_qmmm_center_every_step) THEN
         qmmm_env_qm%center_qm_subsys = .TRUE.
         qmmm_env_qm%center_qm_subsys0 = .TRUE.
      ELSE
         CALL cp__b("qmmm_create.F",198,"Unknown type of CENTER! ")
      ENDIF

      CALL section_vals_val_get(qmmm_section, "CENTER_TYPE", i_val=center_i)
      qmmm_env_qm%center_qm_subsys_pbc_aware = (center_i == do_qmmm_center_pbc_aware)

      ! Compatibility with the QM/MM in CPMD code
      CALL section_vals_val_get(qmmm_section, "NOCOMPATIBILITY", l_val=nocompatibility)
      qmmm_env_qm%compatibility = .NOT. nocompatibility

      ! Parallel scheme for the long range
      CALL section_vals_val_get(qmmm_section, "PARALLEL_SCHEME", &
                                i_val=qmmm_env_qm%par_scheme)

      ! Periodic boundary condition calculation
      CALL section_vals_get(qmmm_periodic, explicit=explicit)
      qmmm_env_qm%periodic = explicit
      !multipole section is switched on by default; switched off only if explicitly stated
      IF (qmmm_env_qm%periodic) qmmm_env_qm%multipole = .TRUE.
      CALL section_vals_get(multipole_section, explicit=explicit)
      CALL section_vals_val_get(multipole_section, "_SECTION_PARAMETERS_", i_val=use_multipole)
      IF (explicit .AND. use_multipole == do_multipole_section_off) qmmm_env_qm%multipole = .FALSE.
      IF (explicit .AND. use_multipole == do_multipole_section_on) qmmm_env_qm%multipole = .TRUE.
      IF (qmmm_env_qm%periodic .AND. qmmm_env_qm%multipole) CALL cite_reference(Laino2006)
      IF (qmmm_coupl_type == do_qmmm_none) THEN
         IF (qmmm_env_qm%periodic) &
            CALL cp_warn(cp__l("qmmm_create.F",224), &
                         "QMMM periodic calculation with coupling NONE was requested! "// &
                         "Switching off the periodic keyword since periodic and non-periodic "// &
                         "calculation with coupling NONE represent the same method! ")
         qmmm_env_qm%periodic = .FALSE.
      END IF

      ! First Initialize Fist...
      CALL section_vals_val_set(force_env_section, "METHOD", i_val=do_fist)
      CALL fist_env_create(fist_env, para_env=para_env)
      CALL fist_env_set(fist_env, qmmm=.TRUE., qmmm_env=qmmm_env_mm)
      CALL fist_init(fist_env, root_section, para_env, force_env_section, &
                     subsys_section, use_motion_section, prev_subsys=prev_subsys)
      CALL fist_env_get(fist_env, subsys=subsys_mm, cell=mm_cell)

      ! Set up QM/MM Options
      CALL setup_qmmm_vars_qm(qmmm_section, &
                              qmmm_env_qm, &
                              subsys_mm, &
                              qm_atom_type, &
                              qm_atom_index, &
                              mm_atom_index, &
                              qm_cell_small, &
                              qmmm_coupl_type, &
                              eps_mm_rspace, &
                              qmmm_link, &
                              para_env)

      qmmm_env_qm%qm_atom_index => qm_atom_index
      qmmm_env_qm%mm_atom_index => mm_atom_index
      qmmm_env_qm%eps_mm_rspace = eps_mm_rspace
      qmmm_env_qm%qmmm_coupl_type = qmmm_coupl_type
      qmmm_env_qm%qmmm_link = qmmm_link
      qmmm_env_qm%num_qm_atoms = SIZE(qm_atom_index)
      qmmm_env_qm%num_mm_atoms = SIZE(mm_atom_index)
      IF (qmmm_env_qm%image_charge) THEN
         qmmm_env_qm%num_image_mm_atoms = SIZE(qmmm_env_qm%image_charge_pot%image_mm_list)
         CALL cite_reference(Golze2013)
      END IF

      ! Duplicate structure for link atoms
      IF (qmmm_link) THEN
         IF (ASSOCIATED(mm_link_atoms)) THEN
            ALLOCATE (qmmm_env_qm%mm_link_atoms(SIZE(mm_link_atoms)))
            qmmm_env_qm%mm_link_atoms = mm_link_atoms
         END IF
      END IF
      IF (iw > 0) THEN
         WRITE (iw, '(A,I26)') " Number of QM atoms: ", qmmm_env_qm%num_qm_atoms
         WRITE (iw, '(A,I26)') " Number of MM atoms: ", qmmm_env_qm%num_mm_atoms
         IF (qmmm_env_qm%image_charge) THEN
            WRITE (iw, '(A,I8)') " Number of MM atoms with image charge: ", &
               qmmm_env_qm%num_image_mm_atoms
         ENDIF
         WRITE (iw, '(A)') " QM cell ::"
         CALL write_cell(qm_cell_small, subsys_section)
      END IF
      CALL get_cell(qm_cell_small, abc=abc_qm)
      CALL get_cell(mm_cell, abc=abc_mm)

      IF (qmmm_env_qm%image_charge) THEN
         IF (ANY(ABS(abc_mm-abc_qm) > 1.0E-12)) &
            CALL cp__b("qmmm_create.F",286,"QM and MM box need to have the same size when using image charges")
      ENDIF

      ! Assign charges and mm_el_pot_radius from fist_topology
      CALL fist_env_get(fist_env, fist_nonbond_env=fist_nonbond_env)
      ALLOCATE (mm_atom_chrg(SIZE(mm_atom_index)))
      ALLOCATE (mm_el_pot_radius(SIZE(mm_atom_index)))
      ALLOCATE (mm_el_pot_radius_corr(SIZE(mm_atom_index)))
      mm_atom_chrg = 0.0_dp
      mm_el_pot_radius = 0.0_dp
      mm_el_pot_radius_corr = 0.0_dp

      CALL assign_mm_charges_and_radius(subsys=subsys_mm, &
                                        charges=fist_nonbond_env%charges, &
                                        mm_atom_chrg=mm_atom_chrg, &
                                        mm_el_pot_radius=mm_el_pot_radius, &
                                        mm_el_pot_radius_corr=mm_el_pot_radius_corr, &
                                        mm_atom_index=mm_atom_index, &
                                        mm_link_atoms=mm_link_atoms, &
                                        mm_link_scale_factor=mm_link_scale_factor, &
                                        added_shells=added_shells, &
                                        shell_model=shell_model)

      qmmm_env_qm%mm_atom_chrg => mm_atom_chrg
      qmmm_env_qm%mm_el_pot_radius => mm_el_pot_radius
      qmmm_env_qm%mm_el_pot_radius_corr => mm_el_pot_radius_corr
      qmmm_env_qm%added_shells => added_shells

      qmmm_link_Imomm = .FALSE.
      IF (qmmm_link) THEN
         CALL setup_qmmm_links(qmmm_section, qmmm_links, mm_el_pot_radius, &
                               mm_el_pot_radius_corr, mm_atom_index, iw)
         qmmm_env_qm%qmmm_links => qmmm_links

         CALL print_qmmm_links(qmmm_section, qmmm_links)

         CALL add_set_release(qmmm_env_qm%added_charges)
         CALL move_or_add_atoms(qmmm_section, move_mm_charges, add_mm_charges, &
                                mm_atom_chrg, mm_el_pot_radius, mm_el_pot_radius_corr, &
                                added_charges, mm_atom_index)
         qmmm_env_qm%move_mm_charges = move_mm_charges
         qmmm_env_qm%add_mm_charges = add_mm_charges
         qmmm_env_qm%added_charges => added_charges
         IF (ASSOCIATED(qmmm_links%imomm)) qmmm_link_imomm = (SIZE(qmmm_links%imomm) /= 0)
      END IF

      CALL print_qmmm_charges(mm_atom_index, mm_atom_chrg, mm_el_pot_radius, &
                              mm_el_pot_radius_corr, qmmm_env_qm%added_charges, &
                              qmmm_env_qm%added_shells, qmmm_section, nocompatibility, shell_model)
      IF (qmmm_env_qm%image_charge) THEN
         CALL print_image_charge_info(qmmm_env_qm, qmmm_section)
      ENDIF

      CALL section_vals_val_get(qmmm_section, "DELTA_CHARGE", i_val=delta_charge)
      CALL section_vals_val_get(force_env_section, "DFT%CHARGE", i_val=orig_charge)
      CALL section_vals_val_set(force_env_section, "DFT%CHARGE", i_val=orig_charge+delta_charge)

      CALL section_vals_val_set(force_env_section, "METHOD", i_val=do_qs)
      CALL create_small_subsys(subsys_qm, &
                               big_subsys=subsys_mm, small_para_env=para_env, &
                               small_cell=qm_cell_small, sub_atom_index=qm_atom_index, &
                               sub_atom_kind_name=qm_atom_type, para_env=para_env, &
                               force_env_section=force_env_section, subsys_section=subsys_section, &
                               ignore_outside_box=ignore_outside_box)
      IF (qmmm_link_imomm) CALL qmmm_link_Imomm_coord(qmmm_links, subsys_qm%particles%els, &
                                                      qm_atom_index)
      CALL qs_env_create(qs_env, globenv)
      CALL qs_init(qs_env, para_env, globenv, root_section, cp_subsys=subsys_qm, &
                   cell=qm_cell_small, qmmm=.TRUE., qmmm_env_qm=qmmm_env_qm, &
                   force_env_section=force_env_section, subsys_section=subsys_section, &
                   use_motion_section=use_motion_section)
      CALL cp_subsys_release(subsys_qm)

      IF (qmmm_env_qm%periodic) THEN
         IF (.NOT. ASSOCIATED(super_cell)) THEN
            ALLOCATE (super_cell)
         END IF
         CALL cell_clone(mm_cell, super_cell)
         CALL set_qs_env(qs_env, super_cell=super_cell, qmmm_periodic=qmmm_env_qm%periodic)
         CALL cell_release(super_cell)
      END IF
      CALL section_vals_val_set(force_env_section, "DFT%CHARGE", i_val=orig_charge)
      CALL cp_print_key_finished_output(iw, logger, print_gen, "")
      iw2 = cp_print_key_unit_nr(logger, qmmm_section, "PRINT%PROGRAM_BANNER", &
                                 extension=".qmmmLog")
      CALL qmmm_header(iw2)
      CALL cp_print_key_finished_output(iw2, logger, qmmm_section, &
                                        "PRINT%PROGRAM_BANNER")
      !
      ! Initialize MM Potential fitted with Gaussian
      !
      CALL qmmm_init_gaussian_type(qmmm_env_qm=qmmm_env_qm, &
                                   para_env=para_env, &
                                   qs_env=qs_env, &
                                   mm_atom_chrg=mm_atom_chrg, &
                                   added_charges=qmmm_env_qm%added_charges, &
                                   added_shells=qmmm_env_qm%added_shells, &
                                   print_section=print_section, &
                                   qmmm_section=qmmm_section)
      !
      ! Initialize the MM potential stored on vector
      !
      CALL qmmm_init_potential(qmmm_env_qm=qmmm_env_qm, &
                               mm_cell=mm_cell, &
                               added_charges=qmmm_env_qm%added_charges, &
                               added_shells=qmmm_env_qm%added_shells, &
                               print_section=print_section)
      !
      ! Initialize the qmmm_pw_grid
      !
      CALL get_qs_env(qs_env, pw_env=pw_env)
      CALL qmmm_pw_grid_init(qmmm_env=qmmm_env_qm, &
                             pw_env=pw_env)
      !
      ! Initialize the MM periodic potential
      !
      CALL qmmm_init_periodic_potential(qmmm_env_qm=qmmm_env_qm, &
                                        qm_cell_small=qm_cell_small, &
                                        mm_cell=mm_cell, &
                                        para_env=para_env, &
                                        qs_env=qs_env, &
                                        added_charges=qmmm_env_qm%added_charges, &
                                        added_shells=qmmm_env_qm%added_shells, &
                                        qmmm_periodic=qmmm_periodic, &
                                        print_section=print_section, &
                                        mm_atom_chrg=mm_atom_chrg)
      !
      ! Preparing for PBC...
      !
      CALL setup_origin_mm_cell(qmmm_section, qmmm_env_qm, qm_cell_small, &
                                dr=pw_env%pw_pools(pw_env%auxbas_grid)%pool%pw_grid%dr)

      CALL cell_release(qm_cell_small)

      ! assemble the actuall qmmm_env
      ALLOCATE (qmmm_env)
      qmmm_env%qs_env => qs_env
      qmmm_env%fist_env => fist_env
      qmmm_env%qm => qmmm_env_qm

      ! The qmmm_env inherits our ref_cout for qmmm_env_qm, fist_env, qs_env
      ! An expection is qmmm_env_mm, because it's buried in the fist_env
      CALL qmmm_env_mm_release(qmmm_env_mm)

      CALL section_vals_val_set(force_env_section, "METHOD", i_val=do_qmmm)
      DEALLOCATE (qm_atom_type)

      CALL timestop(handle)

   END SUBROUTINE qmmm_env_create

END MODULE qmmm_create
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize a QM/MM calculation with Force-Mixing
!> \author Ole Schuett
! **************************************************************************************************
MODULE qmmmx_create
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_type
   USE qmmm_create,                     ONLY: qmmm_env_create
   USE qmmm_types,                      ONLY: qmmm_env_get,&
                                              qmmm_env_release,&
                                              qmmm_env_type
   USE qmmmx_types,                     ONLY: qmmmx_env_type
   USE qmmmx_util,                      ONLY: setup_force_mixing_qmmm_sections,&
                                              update_force_mixing_labels
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmmx_create'

   PUBLIC :: qmmmx_env_create

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qmmmx_env ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \par History
!>      02.2012 created [noam]
!> \author Noam Bernstein
! **************************************************************************************************
   SUBROUTINE qmmmx_env_create(qmmmx_env, root_section, para_env, globenv, &
                               force_env_section, subsys_section, use_motion_section)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_env_create', &
         routineP = moduleN//':'//routineN

      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(qmmm_env_type), POINTER                       :: dummy_qmmm_env
      TYPE(section_vals_type), POINTER                   :: qmmm_core_section, &
                                                            qmmm_extended_section, qmmm_section

      NULLIFY (dummy_qmmm_env)

      qmmm_section => section_vals_get_subs_vals(force_env_section, "QMMM")

      CALL qmmm_env_create(dummy_qmmm_env, root_section, para_env, globenv, &
                           force_env_section, qmmm_section, subsys_section, use_motion_section, &
                           ignore_outside_box=.TRUE.)
      CALL qmmm_env_get(dummy_qmmm_env, subsys=subsys)

      CALL update_force_mixing_labels(subsys, qmmm_section)

      ! using CUR_INDICES and CUR_LABELS, create appropriate QM_KIND sections for two QM/MM calculations
      CALL setup_force_mixing_qmmm_sections(subsys, qmmm_section, qmmm_core_section, qmmm_extended_section)

      ALLOCATE (qmmmx_env)
      CALL qmmm_env_create(qmmmx_env%core, root_section, para_env, globenv, &
                           force_env_section, qmmm_core_section, subsys_section, use_motion_section, &
                           ignore_outside_box=.TRUE.)

      CALL qmmm_env_create(qmmmx_env%ext, root_section, para_env, globenv, &
                           force_env_section, qmmm_extended_section, subsys_section, use_motion_section, &
                           ignore_outside_box=.TRUE.)

      CALL section_vals_release(qmmm_core_section)
      CALL section_vals_release(qmmm_extended_section)
      CALL qmmm_env_release(dummy_qmmm_env)

   END SUBROUTINE qmmmx_env_create

END MODULE qmmmx_create
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE acc_device
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PUBLIC :: acc_get_ndevices, acc_set_active_device

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'acc_device'


CONTAINS

! **************************************************************************************************
!> \brief Get number of accelerator devices
!> \retval n number of accelerator devices
! **************************************************************************************************
   FUNCTION acc_get_ndevices() RESULT(n)
      INTEGER                                  :: n


      n = 0
   END FUNCTION acc_get_ndevices

! **************************************************************************************************
!> \brief Set active accelerator device
!> \param dev_id device ID
! **************************************************************************************************
   SUBROUTINE acc_set_active_device(dev_id)
      INTEGER :: dev_id

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dev_id))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_device.F",80,"__ACC not compiled in")
   END SUBROUTINE acc_set_active_device

END MODULE acc_device
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief perform classical molecular dynamics and path integral simulations
!> \par History
!>      gt SEPT-23-2002: part is allocated/deallocated/initialized in
!>                       read_coord_vel
!>      CJM rewrite
!> \author CJM-Sept-01-02
! **************************************************************************************************
MODULE fist_main
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE fist_environment,                ONLY: fist_init
   USE fist_environment_types,          ONLY: fist_env_create,&
                                              fist_env_release,&
                                              fist_env_set,&
                                              fist_environment_type
   USE force_env_methods,               ONLY: force_env_create
   USE force_env_types,                 ONLY: force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_vals_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_main'

! *** Global variables ***
   PUBLIC :: fist_create_force_env

!!-----------------------------------------------------------------------------!

CONTAINS

!-----------------------------------------------------------------------------!
! FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST FIST  !
!-----------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Controls program flow for classical MD and path-integrals
!> \param force_env ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param prev_subsys ...
!> \par Used By
!>      cp2k
!> \author CJM
! **************************************************************************************************
   SUBROUTINE fist_create_force_env(force_env, root_section, para_env, globenv, &
                                    qmmm, qmmm_env, force_env_section, subsys_section, use_motion_section, prev_subsys)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      LOGICAL, OPTIONAL                                  :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: prev_subsys

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fist_create_force_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: myqmmm
      TYPE(fist_environment_type), POINTER               :: fist_env

      CALL timeset(routineN, handle)
      myqmmm = .FALSE.
      IF (PRESENT(qmmm)) THEN
         myqmmm = qmmm
      END IF

      CALL fist_env_create(fist_env, para_env=para_env)
      IF (PRESENT(qmmm_env)) THEN
         CALL fist_env_set(fist_env, qmmm=myqmmm, qmmm_env=qmmm_env)
      ELSE
         CALL fist_env_set(fist_env, qmmm=myqmmm)
      END IF
      ! *** Read the input and the database files and perform further  ***
      ! *** initializations for the setup of the FIST environment ***
      CALL fist_init(fist_env, root_section, para_env, force_env_section, &
                     subsys_section, use_motion_section, prev_subsys=prev_subsys)

      CALL force_env_create(force_env, root_section, fist_env=fist_env, &
                            para_env=para_env, globenv=globenv, &
                            force_env_section=force_env_section)

      CALL fist_env_release(fist_env)
      CALL timestop(handle)
   END SUBROUTINE fist_create_force_env

END MODULE fist_main
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief initialize mixed environment
!> \author fschiff
! **************************************************************************************************
MODULE mixed_environment
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_methods,                    ONLY: read_cell,&
                                              write_cell
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: cp_subsys_create
   USE cp_subsys_types,                 ONLY: cp_subsys_release,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_type
   USE distribution_methods,            ONLY: distribute_molecules_1d
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE mixed_energy_types,              ONLY: allocate_mixed_energy,&
                                              mixed_energy_type
   USE mixed_environment_types,         ONLY: mixed_environment_type,&
                                              set_mixed_env
   USE molecule_kind_types,             ONLY: molecule_kind_type,&
                                              write_molecule_kind_set
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mixed_environment'
   PUBLIC :: mixed_init

CONTAINS

! **************************************************************************************************
!> \brief reads the input and database file for mixed
!> \param mixed_env ...
!> \param root_section ...
!> \param para_env ...
!> \param force_env_section ...
!> \param use_motion_section ...
!> \par Used By
!>      mixed_main
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE mixed_init(mixed_env, root_section, para_env, force_env_section, &
                         use_motion_section)

      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      LOGICAL, INTENT(IN)                                :: use_motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: use_ref_cell
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: cell_section, subsys_section

      CALL timeset(routineN, handle)

      NULLIFY (subsys, cell, cell_ref)
      NULLIFY (cell_section)

      subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      cell_section => section_vals_get_subs_vals(subsys_section, "CELL")

      CALL set_mixed_env(mixed_env, input=force_env_section)
      CALL cp_subsys_create(subsys, para_env, root_section, &
                            force_env_section=force_env_section, &
                            use_motion_section=use_motion_section)

      CALL read_cell(cell, cell_ref, use_ref_cell=use_ref_cell, &
                     cell_section=cell_section, para_env=para_env)
      CALL get_cell(cell, abc=abc)

      ! Print the cell parameters ***
      CALL write_cell(cell, subsys_section, cell_ref)

      CALL mixed_init_subsys(mixed_env, subsys, cell, cell_ref, &
                             force_env_section, subsys_section)

      CALL cell_release(cell)
      CALL cell_release(cell_ref)
      CALL cp_subsys_release(subsys)

      CALL timestop(handle)

   END SUBROUTINE mixed_init

! **************************************************************************************************
!> \brief   Read the input and the database files for the setup of the
!>          mixed environment.
!> \param mixed_env ...
!> \param subsys ...
!> \param cell ...
!> \param cell_ref ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \date    11.06
!> \author  fschiff
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE mixed_init_subsys(mixed_env, subsys, cell, cell_ref, &
                                force_env_section, subsys_section)

      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_init_subsys', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      NULLIFY (mixed_energy, local_molecules, local_particles)
      particle_set => subsys%particles%els
      atomic_kind_set => subsys%atomic_kinds%els
      molecule_set => subsys%molecules_new%els
      molecule_kind_set => subsys%molecule_kinds_new%els

      ! Create the mixed_energy_type
      CALL allocate_mixed_energy(mixed_energy)

      ! Print the molecule kind set
      CALL write_molecule_kind_set(molecule_kind_set, subsys_section)

      ! Distribute molecules and atoms using the new data structures ***
      CALL distribute_molecules_1d(atomic_kind_set=atomic_kind_set, &
                                   particle_set=particle_set, &
                                   local_particles=local_particles, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   local_molecules=local_molecules, &
                                   force_env_section=force_env_section)

      CALL cp_subsys_set(subsys, cell=cell)

      ! set the mixed_env
      CALL set_mixed_env(mixed_env=mixed_env, subsys=subsys)
      CALL set_mixed_env(mixed_env=mixed_env, &
                         cell_ref=cell_ref, &
                         local_molecules=local_molecules, &
                         local_particles=local_particles, &
                         mixed_energy=mixed_energy)

      CALL distribution_1d_release(local_particles)
      CALL distribution_1d_release(local_molecules)

      CALL timestop(handle)

   END SUBROUTINE mixed_init_subsys

END MODULE mixed_environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief perform biased molecular dynamics (H= k H1 + (1-k) H2 [linear or general mixing)
!> \author fschiff 11.06
! **************************************************************************************************
MODULE mixed_main
   USE cp_files,                        ONLY: open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_create,&
                                              cp_logger_set,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_para_env_create
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length
   USE message_passing,                 ONLY: mp_comm_split
   USE mixed_environment,               ONLY: mixed_init
   USE mixed_environment_types,         ONLY: mixed_env_create,&
                                              mixed_environment_type,&
                                              set_mixed_env
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mixed_main'

   PUBLIC :: mixed_create_force_env

CONTAINS
! **************************************************************************************************
!> \brief Controls program flow for mixed calculations
!> \param mixed_env ...
!> \param root_section ...
!> \param para_env ...
!> \param force_env_section ...
!> \param n_subforce_eval ...
!> \param use_motion_section ...
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE mixed_create_force_env(mixed_env, root_section, para_env, &
                                     force_env_section, n_subforce_eval, use_motion_section)

      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      INTEGER, INTENT(IN)                                :: n_subforce_eval
      LOGICAL, INTENT(IN)                                :: use_motion_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mixed_create_force_env', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: c_val, input_file_path, output_file_path
      INTEGER                                            :: group_size_wish, handle, i, lp, &
                                                            n_rep_val, ngroup_wish, output_unit, &
                                                            unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: group_partition, i_vals
      LOGICAL                                            :: group_size_wish_set, ngroup_wish_set
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, force_env_section, "MIXED%PRINT%PROGRAM_RUN_INFO", &
                                         extension=".log")

      CALL mixed_env_create(mixed_env, para_env=para_env)
      ! Setup the new parallel env
      NULLIFY (group_partition)
      group_size_wish_set = .FALSE.
      ngroup_wish_set = .FALSE.
      CALL section_vals_val_get(force_env_section, "MIXED%GROUP_PARTITION", n_rep_val=n_rep_val)
      IF (n_rep_val > 0) THEN
         CALL section_vals_val_get(force_env_section, "MIXED%GROUP_PARTITION", i_vals=i_vals)
         ALLOCATE (group_partition(0:SIZE(i_vals)-1))
         group_partition(:) = i_vals
         ngroup_wish_set = .TRUE.
         ngroup_wish = SIZE(i_vals)
      ELSE
         CALL section_vals_val_get(force_env_section, "MIXED%NGROUPS", n_rep_val=n_rep_val)
         IF (n_rep_val > 0) THEN
            CALL section_vals_val_get(force_env_section, "MIXED%NGROUPS", i_val=ngroup_wish)
         ELSE
            ngroup_wish = n_subforce_eval
         END IF
         group_size_wish_set = .TRUE.
         group_size_wish = MAX(1, para_env%num_pe/ngroup_wish)
      END IF

      ! Split the current communicator
      ALLOCATE (mixed_env%group_distribution(0:para_env%num_pe-1))
      IF (group_size_wish_set) THEN
         CALL mp_comm_split(para_env%group, mixed_env%new_group, mixed_env%ngroups, mixed_env%group_distribution, &
                            subgroup_min_size=group_size_wish)
      ELSE IF (ngroup_wish_set) THEN
         CALL mp_comm_split(para_env%group, mixed_env%new_group, mixed_env%ngroups, mixed_env%group_distribution, &
                            n_subgroups=ngroup_wish, &
                            group_partition=group_partition)
      ENDIF
      IF (output_unit > 0) THEN
         WRITE (output_unit, FMT="(T2,A,T71,I10)") "MIXED_ENV| Number of created MPI groups:", mixed_env%ngroups
         WRITE (output_unit, FMT="(T2,A)", ADVANCE="NO") "MIXED_ENV| Task to group correspondence:"
         DO i = 0, para_env%num_pe-1
            IF (MODULO(i, 4) == 0) WRITE (output_unit, *)
            WRITE (output_unit, FMT='(A3,I4,A3,I4,A1)', ADVANCE="NO") &
               "  (", i, " : ", mixed_env%group_distribution(i), ")"
         END DO
         WRITE (output_unit, *)
      ENDIF
      IF (ASSOCIATED(group_partition)) THEN
         DEALLOCATE (group_partition)
      END IF
      ! Allocate para_env and handle the several loggers
      ALLOCATE (mixed_env%sub_para_env(mixed_env%ngroups))
      ALLOCATE (mixed_env%sub_logger(mixed_env%ngroups))
      ALLOCATE (mixed_env%energies(n_subforce_eval))
      !
      DO i = 1, mixed_env%ngroups
         NULLIFY (mixed_env%sub_para_env(i)%para_env, logger)
         IF (MODULO(i-1, mixed_env%ngroups) == mixed_env%group_distribution(para_env%mepos)) THEN
            ! Create sub_para_env
            CALL cp_para_env_create(mixed_env%sub_para_env(i)%para_env, &
                                    group=mixed_env%new_group, &
                                    owns_group=.TRUE.)
            ! Create sub_logger
            IF (mixed_env%sub_para_env(i)%para_env%mepos == mixed_env%sub_para_env(i)%para_env%source) THEN
               ! Redirecting output of subforce_eval to file..
               CALL section_vals_val_get(root_section, "GLOBAL%PROJECT_NAME", &
                                         c_val=input_file_path)
               lp = LEN_TRIM(input_file_path)
               input_file_path(lp+1:LEN(input_file_path)) = "-r-"// &
                                                            ADJUSTL(cp_to_string(i))
               lp = LEN_TRIM(input_file_path)
               output_file_path = input_file_path(1:lp)//".out"
               CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=unit_nr)
            ELSE
               unit_nr = -1
            END IF
            CALL cp_logger_create(mixed_env%sub_logger(i)%p, &
                                  para_env=mixed_env%sub_para_env(i)%para_env, &
                                  default_global_unit_nr=unit_nr, &
                                  close_global_unit_on_dealloc=.FALSE.)
            ! Try to use better names for the local log if it is not too late
            CALL section_vals_val_get(root_section, "GLOBAL%OUTPUT_FILE_NAME", &
                                      c_val=c_val)
            IF (c_val /= "") THEN
               CALL cp_logger_set(mixed_env%sub_logger(i)%p, &
                                  local_filename=TRIM(c_val)//"_localLog")
            END IF
            CALL section_vals_val_get(root_section, "GLOBAL%PROJECT", c_val=c_val)
            IF (c_val /= "") THEN
               CALL cp_logger_set(mixed_env%sub_logger(i)%p, &
                                  local_filename=TRIM(c_val)//"_localLog")
            END IF
            mixed_env%sub_logger(i)%p%iter_info%project_name = c_val
            CALL section_vals_val_get(root_section, "GLOBAL%PRINT_LEVEL", &
                                      i_val=mixed_env%sub_logger(i)%p%iter_info%print_level)

         END IF
      END DO

      ! *** initializations for the setup of the MIXED environment ***
      CALL set_mixed_env(mixed_env)
      CALL mixed_init(mixed_env, root_section, para_env, force_env_section, &
                      use_motion_section)
      CALL timestop(handle)

   END SUBROUTINE mixed_create_force_env

END MODULE mixed_main
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief interface to use cp2k as library
!> \note
!>      useful additions for the future would be:
!>      - string(path) based set/get of simple values (to change the new
!>        input during the run and extract more data (energy types for example).
!>      - set/get of a subset of atoms
!> \par History
!>      07.2004 created [fawzi]
!>      11.2004 parallel version [fawzi]
!> \author fawzi & Johanna
! **************************************************************************************************
MODULE f77_interface
   USE acc_device,                      ONLY: acc_get_ndevices,&
                                              acc_set_active_device
   USE bibliography,                    ONLY: add_all_references
   USE cell_types,                      ONLY: cell_type,&
                                              init_cell
   USE cp2k_info,                       ONLY: get_runtime_info
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_finalize_lib,&
                                              cp_dbcsr_init_lib
   USE cp_error_handling,               ONLY: cp_error_handling_setup
   USE cp_files,                        ONLY: init_preconnection_list,&
                                              open_file
   USE cp_log_handling,                 ONLY: &
        cp_add_default_logger, cp_default_logger_stack_size, cp_failure_level, &
        cp_get_default_logger, cp_logger_create, cp_logger_get_default_unit_nr, cp_logger_release, &
        cp_logger_retain, cp_logger_type, cp_rm_default_logger, cp_to_string
   USE cp_output_handling,              ONLY: cp_iterate
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: get_results,&
                                              test_for_result
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type,&
                                              unpack_subsys_particles
   USE eip_environment,                 ONLY: eip_init
   USE eip_environment_types,           ONLY: eip_env_create,&
                                              eip_env_release,&
                                              eip_environment_type
   USE environment,                     ONLY: cp2k_finalize,&
                                              cp2k_init,&
                                              cp2k_read,&
                                              cp2k_setup
   USE fist_main,                       ONLY: fist_create_force_env
   USE force_env_methods,               ONLY: force_env_calc_energy_force,&
                                              force_env_create
   USE force_env_types,                 ONLY: &
        force_env_get, force_env_get_frc, force_env_get_natom, force_env_get_nparticle, &
        force_env_get_pos, force_env_get_vel, force_env_release, force_env_retain, force_env_set, &
        force_env_type, multiple_fe_list
   USE force_env_utils,                 ONLY: force_env_shake
   USE fp_types,                        ONLY: fp_env_create,&
                                              fp_env_read,&
                                              fp_env_release,&
                                              fp_env_write,&
                                              fp_type
   USE global_types,                    ONLY: global_environment_type,&
                                              globenv_create,&
                                              globenv_release,&
                                              globenv_retain
   USE input_constants,                 ONLY: do_eip,&
                                              do_fist,&
                                              do_mixed,&
                                              do_qmmm,&
                                              do_qmmmx,&
                                              do_qs
   USE input_cp2k_check,                ONLY: check_cp2k_input
   USE input_cp2k_force_eval,           ONLY: create_force_eval_section
   USE input_cp2k_read,                 ONLY: empty_initial_variables,&
                                              read_input
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: &
        section_get_keyword, section_release, section_type, section_vals_duplicate, &
        section_vals_get, section_vals_get_subs_vals, section_vals_release, &
        section_vals_remove_values, section_vals_retain, section_vals_type, section_vals_val_get, &
        section_vals_write
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: default_output_unit,&
                                              m_chdir,&
                                              m_getcwd,&
                                              m_memory
   USE message_passing,                 ONLY: &
        MPI_COMM_WORLD, add_mp_perf_env, get_mp_perf_env, mp_max, mp_perf_env_release, &
        mp_perf_env_retain, mp_perf_env_type, mp_world_finalize, mp_world_init, rm_mp_perf_env
   USE metadynamics_types,              ONLY: meta_env_release,&
                                              meta_env_type
   USE metadynamics_utils,              ONLY: metadyn_read
   USE mixed_environment_types,         ONLY: mixed_env_release,&
                                              mixed_environment_type
   USE mixed_main,                      ONLY: mixed_create_force_env
   USE periodic_table,                  ONLY: init_periodic_table
   USE pw_cuda,                         ONLY: pw_cuda_finalize,&
                                              pw_cuda_init
   USE qmmm_create,                     ONLY: qmmm_env_create
   USE qmmm_types,                      ONLY: qmmm_env_release,&
                                              qmmm_env_type
   USE qmmmx_create,                    ONLY: qmmmx_env_create
   USE qmmmx_types,                     ONLY: qmmmx_env_release,&
                                              qmmmx_env_type
   USE qs_environment,                  ONLY: qs_init
   USE qs_environment_types,            ONLY: qs_env_create,&
                                              qs_env_release,&
                                              qs_environment_type
   USE reference_manager,               ONLY: remove_all_references
   USE string_table,                    ONLY: string_table_allocate,&
                                              string_table_deallocate
   USE timings,                         ONLY: add_timer_env,&
                                              get_timer_env,&
                                              rm_timer_env,&
                                              timer_env_release,&
                                              timer_env_retain,&
                                              timings_register_hooks
   USE timings_types,                   ONLY: timer_env_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'f77_interface'

! **************************************************************************************************
   TYPE f_env_p_type
      TYPE(f_env_type), POINTER :: f_env
   END TYPE f_env_p_type

! **************************************************************************************************
   TYPE f_env_type
      INTEGER :: id_nr
      TYPE(force_env_type), POINTER      :: force_env
      TYPE(cp_logger_type), POINTER      :: logger
      TYPE(timer_env_type), POINTER      :: timer_env
      TYPE(mp_perf_env_type), POINTER    :: mp_perf_env
      CHARACTER(len=default_path_length) :: my_path, old_path
   END TYPE f_env_type

   TYPE(f_env_p_type), DIMENSION(:), POINTER, SAVE :: f_envs
   TYPE(cp_para_env_type), POINTER, SAVE :: default_para_env
   LOGICAL, SAVE :: module_initialized = .FALSE.
   INTEGER, SAVE :: last_f_env_id = 0, n_f_envs = 0

   PUBLIC :: default_para_env
   PUBLIC :: init_cp2k, finalize_cp2k
   PUBLIC :: create_force_env, destroy_force_env, set_pos, get_pos, &
             get_force, calc_energy_force, get_energy, get_stress_tensor, &
             calc_energy, calc_force, check_input, get_natom, get_nparticle, &
             f_env_add_defaults, f_env_rm_defaults, f_env_type, &
             f_env_get_from_id, &
             set_vel, do_shake, set_cell, get_cell, get_result_r1
CONTAINS

! **************************************************************************************************
!> \brief returns the position of the force env corresponding to the given id
!> \param env_id the id of the requested environment
!> \retval res ...
!> \author fawzi
!> \note
!>      private utility function
! **************************************************************************************************
   FUNCTION get_pos_of_env(env_id) RESULT(res)
      INTEGER, INTENT(in)                                :: env_id
      INTEGER                                            :: res

      INTEGER                                            :: env_pos, isub

      env_pos = -1
      DO isub = 1, n_f_envs
         IF (f_envs(isub)%f_env%id_nr == env_id) THEN
            env_pos = isub
         END IF
      END DO
      res = env_pos
   END FUNCTION get_pos_of_env

! **************************************************************************************************
!> \brief initializes cp2k, needs to be called once before using any of the
!>      other functions when using cp2k as library
!> \param init_mpi if the mpi environment should be initialized
!> \param ierr returns a number different from 0 if there was an error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE init_cp2k(init_mpi, ierr)
      LOGICAL, INTENT(in)                                :: init_mpi
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'init_cp2k', routineP = moduleN//':'//routineN

      INTEGER                                            :: mpi_comm_default, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      IF (.NOT. module_initialized) THEN
         ! install error handler hooks
         CALL cp_error_handling_setup()

         ! install timming handler hooks
         CALL timings_register_hooks()

         ! Initialise preconnection list
         CALL init_preconnection_list()

         ! get runtime information
         CALL get_runtime_info

         IF (init_mpi) THEN
            ! get the default system wide communicator
            CALL mp_world_init(mpi_comm_default)
         ELSE
            mpi_comm_default = MPI_COMM_WORLD
         END IF

         CALL string_table_allocate()
         CALL add_mp_perf_env()
         CALL add_timer_env()

         ! re-create the para_env and log with correct (reordered) ranks
         NULLIFY (default_para_env)
         CALL cp_para_env_create(default_para_env, group=mpi_comm_default, &
                                 owns_group=.FALSE.)
         IF (default_para_env%source == default_para_env%mepos) THEN
            unit_nr = default_output_unit
         ELSE
            unit_nr = -1
         END IF
         NULLIFY (logger)

         CALL cp_logger_create(logger, para_env=default_para_env, &
                               default_global_unit_nr=unit_nr, &
                               close_global_unit_on_dealloc=.FALSE.)
         CALL cp_add_default_logger(logger)
         CALL cp_logger_release(logger)

         ALLOCATE (f_envs(0))
         module_initialized = .TRUE.
         ierr = 0

         !   *** Initialize mathematical constants ***
         CALL init_periodic_table()

         !   *** init the bibliography ***
         CALL add_all_references()

         IF (acc_get_ndevices() > 0) &
            CALL acc_set_active_device(MOD(default_para_env%mepos, acc_get_ndevices()))

         CALL pw_cuda_init()

         ! Initialize the DBCSR configuration
         CALL cp_dbcsr_init_lib()

      ELSE
         ierr = cp_failure_level
      END IF

      !sample peak memory
      CALL m_memory()

   END SUBROUTINE init_cp2k

! **************************************************************************************************
!> \brief cleanup after you have finished using this interface
!> \param finalize_mpi if the mpi environment should be finalized
!> \param ierr returns a number different from 0 if there was an error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE finalize_cp2k(finalize_mpi, ierr)
      LOGICAL, INTENT(in)                                :: finalize_mpi
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'finalize_cp2k', routineP = moduleN//':'//routineN

      INTEGER                                            :: ienv

!sample peak memory

      CALL m_memory()

      IF (.NOT. module_initialized) THEN
         ierr = cp_failure_level
      ELSE
         ! Finalize the DBCSR configuration
         CALL cp_dbcsr_finalize_lib(group=default_para_env%group, output_unit=0)
         CALL pw_cuda_finalize()
         DO ienv = n_f_envs, 1, -1
            CALL destroy_force_env(f_envs(ienv)%f_env%id_nr, ierr=ierr)
            IF(.NOT.(ierr == 0))CALL cp__a("f77_interface.F",301)
         END DO
         DEALLOCATE (f_envs)
         CALL cp_para_env_release(default_para_env)
         ierr = 0
         CALL cp_rm_default_logger()

         !   *** deallocate the bibliography ***
         CALL remove_all_references()
         CALL rm_timer_env()
         CALL rm_mp_perf_env()
         CALL string_table_deallocate(0)
         IF (finalize_mpi) THEN
            CALL mp_world_finalize()
         END IF
      END IF
   END SUBROUTINE finalize_cp2k

! **************************************************************************************************
!> \brief deallocates a f_env
!> \param f_env the f_env to deallocate
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE f_env_dealloc(f_env)
      TYPE(f_env_type), POINTER                          :: f_env

      CHARACTER(len=*), PARAMETER :: routineN = 'f_env_dealloc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr

      IF(.NOT.(ASSOCIATED(f_env)))CALL cp__a("f77_interface.F",331)
      CALL force_env_release(f_env%force_env)
      CALL cp_logger_release(f_env%logger)
      CALL timer_env_release(f_env%timer_env)
      CALL mp_perf_env_release(f_env%mp_perf_env)
      IF (f_env%old_path /= f_env%my_path) THEN
         CALL m_chdir(f_env%old_path, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("f77_interface.F",338)
      END IF
   END SUBROUTINE f_env_dealloc

! **************************************************************************************************
!> \brief createates a f_env
!> \param f_env the f_env to createate
!> \param force_env the force_environment to be stored
!> \param timer_env the timer env to be stored
!> \param mp_perf_env the mp performance environement to be stored
!> \param id_nr ...
!> \param logger ...
!> \param old_dir ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE f_env_create(f_env, force_env, timer_env, mp_perf_env, id_nr, logger, old_dir)
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(timer_env_type), POINTER                      :: timer_env
      TYPE(mp_perf_env_type), POINTER                    :: mp_perf_env
      INTEGER, INTENT(in)                                :: id_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: old_dir

      CHARACTER(len=*), PARAMETER :: routineN = 'f_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (f_env)
      f_env%force_env => force_env
      CALL force_env_retain(f_env%force_env)
      f_env%logger => logger
      CALL cp_logger_retain(logger)
      f_env%timer_env => timer_env
      CALL timer_env_retain(f_env%timer_env)
      f_env%mp_perf_env => mp_perf_env
      CALL mp_perf_env_retain(f_env%mp_perf_env)
      f_env%id_nr = id_nr
      CALL m_getcwd(f_env%my_path)
      f_env%old_path = f_env%my_path
      IF (PRESENT(old_dir)) f_env%old_path = old_dir
   END SUBROUTINE f_env_create

! **************************************************************************************************
!> \brief ...
!> \param f_env_id ...
!> \param f_env ...
! **************************************************************************************************
   SUBROUTINE f_env_get_from_id(f_env_id, f_env)
      INTEGER, INTENT(in)                                :: f_env_id
      TYPE(f_env_type), POINTER                          :: f_env

      CHARACTER(len=*), PARAMETER :: routineN = 'f_env_get_from_id', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: f_env_pos

      NULLIFY (f_env)
      f_env_pos = get_pos_of_env(f_env_id)
      IF (f_env_pos < 1) THEN
         CALL cp__b("f77_interface.F",396,"invalid env_id "//cp_to_string(f_env_id))
      ELSE
         f_env => f_envs(f_env_pos)%f_env
      END IF

   END SUBROUTINE f_env_get_from_id

! **************************************************************************************************
!> \brief adds the default environments of the f_env to the stack of the
!>      defaults, and returns a new error and sets failure to true if
!>      something went wrong
!> \param f_env_id the f_env from where to take the defaults
!> \param f_env will contain the f_env corresponding to f_env_id
!> \param handle ...
!> \author fawzi
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   SUBROUTINE f_env_add_defaults(f_env_id, f_env, handle)
      INTEGER, INTENT(in)                                :: f_env_id
      TYPE(f_env_type), POINTER                          :: f_env
      INTEGER, INTENT(out), OPTIONAL                     :: handle

      CHARACTER(len=*), PARAMETER :: routineN = 'f_env_add_defaults', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: f_env_pos, ierr
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (f_env)
      f_env_pos = get_pos_of_env(f_env_id)
      IF (f_env_pos < 1) THEN
         CALL cp__b("f77_interface.F",432,"invalid env_id "//cp_to_string(f_env_id))
      ELSE
         f_env => f_envs(f_env_pos)%f_env
         logger => f_env%logger
         IF(.NOT.(ASSOCIATED(logger)))CALL cp__a("f77_interface.F",436)
         CALL m_getcwd(f_env%old_path)
         IF (f_env%old_path /= f_env%my_path) THEN
            CALL m_chdir(TRIM(f_env%my_path), ierr)
            IF(.NOT.(ierr == 0))CALL cp__a("f77_interface.F",440)
         END IF
         CALL add_mp_perf_env(f_env%mp_perf_env)
         CALL add_timer_env(f_env%timer_env)
         CALL cp_add_default_logger(logger)
         IF (PRESENT(handle)) handle = cp_default_logger_stack_size()
      END IF
   END SUBROUTINE f_env_add_defaults

! **************************************************************************************************
!> \brief removes the default environments of the f_env to the stack of the
!>      defaults, and sets ierr accordingly to the failuers stored in error
!>      It also releases the error
!> \param f_env the f_env from where to take the defaults
!> \param ierr variable that will be set to a number different from 0 if
!>        error contains an error (otherwise it will be set to 0)
!> \param handle ...
!> \author fawzi
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   SUBROUTINE f_env_rm_defaults(f_env, ierr, handle)
      TYPE(f_env_type), POINTER                          :: f_env
      INTEGER, INTENT(out), OPTIONAL                     :: ierr
      INTEGER, INTENT(in), OPTIONAL                      :: handle

      CHARACTER(len=*), PARAMETER :: routineN = 'f_env_rm_defaults', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr2
      TYPE(cp_logger_type), POINTER                      :: d_logger, logger
      TYPE(mp_perf_env_type), POINTER                    :: d_mp_perf_env
      TYPE(timer_env_type), POINTER                      :: d_timer_env

      IF (ASSOCIATED(f_env)) THEN
         IF (PRESENT(handle)) THEN
            IF(.NOT.(handle == cp_default_logger_stack_size()))CALL cp__a("f77_interface.F",480)
         END IF

         logger => f_env%logger
         d_logger => cp_get_default_logger()
         d_timer_env => get_timer_env()
         d_mp_perf_env => get_mp_perf_env()
         IF(.NOT.(ASSOCIATED(logger)))CALL cp__a("f77_interface.F",487)
         IF(.NOT.(ASSOCIATED(d_logger)))CALL cp__a("f77_interface.F",488)
         IF(.NOT.(ASSOCIATED(d_timer_env)))CALL cp__a("f77_interface.F",489)
         IF(.NOT.(ASSOCIATED(d_mp_perf_env)))CALL cp__a("f77_interface.F",490)
         IF(.NOT.(logger%id_nr == d_logger%id_nr))CALL cp__a("f77_interface.F",491)
         ! IF(.NOT.(d_timer_env%id_nr==f_env%timer_env%id_nr))CALL cp__a("f77_interface.F",492)
         IF(.NOT.(d_mp_perf_env%id_nr == f_env%mp_perf_env%id_nr))CALL cp__a("f77_interface.F",493)
         IF (f_env%old_path /= f_env%my_path) THEN
            CALL m_chdir(TRIM(f_env%old_path), ierr2)
            IF(.NOT.(ierr2 == 0))CALL cp__a("f77_interface.F",496)
         END IF
         IF (PRESENT(ierr)) THEN
            ierr = 0
         ENDIF
         CALL cp_rm_default_logger()
         CALL rm_timer_env()
         CALL rm_mp_perf_env()
      ELSE
         IF (PRESENT(ierr)) THEN
            ierr = 0
         ENDIF
      END IF
   END SUBROUTINE f_env_rm_defaults

! **************************************************************************************************
!> \brief creates a new force environment using the given input, and writing
!>      the output to the given output unit
!> \param new_env_id will contain the id of the newly created environment
!> \param input_declaration ...
!> \param input_path where to read the input (if the input is given it can
!>        a virtual path)
!> \param output_path filename (or name of the unit) for the output
!> \param mpi_comm the mpi communicator to be used for this environment
!>        it will not be freed when you get rid of the force_env
!> \param output_unit if given it should be the unit for the output
!>        and no file is open(should be valid on the processor with rank 0)
!> \param owns_out_unit if the output unit should be closed upon destroing
!>        of the force_env (defaults to true if not default_output_unit)
!> \param input the parsed input, if given and valid it is used
!>        instead of parsing from file
!> \param ierr will return a number different from 0 if there was an error
!> \param work_dir ...
!> \param initial_variables ...
!> \author fawzi
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   RECURSIVE SUBROUTINE create_force_env(new_env_id, input_declaration, input_path, &
                                         output_path, mpi_comm, output_unit, owns_out_unit, &
                                         input, ierr, work_dir, initial_variables)
      INTEGER, INTENT(out)                               :: new_env_id
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(len=*), INTENT(in)                       :: input_path
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: output_path
      INTEGER, INTENT(in), OPTIONAL                      :: mpi_comm, output_unit
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_out_unit
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      INTEGER, INTENT(out), OPTIONAL                     :: ierr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: work_dir
      CHARACTER(len=*), DIMENSION(:, :), OPTIONAL        :: initial_variables

      CHARACTER(len=*), PARAMETER :: routineN = 'create_force_env', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: old_dir, wdir
      INTEGER :: handle, i, ierr2, iforce_eval, isubforce_eval, k, method_name_id, my_group, &
         nforce_eval, ngroups, nsubforce_size, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: group_distribution, i_force_eval, &
                                                            lgroup_distribution
      LOGICAL :: check, do_qmmm_force_mixing, multiple_subsys, my_echo, my_owns_out_unit, &
         use_motion_section, use_multiple_para_env
      TYPE(cp_logger_type), POINTER                      :: logger, my_logger
      TYPE(cp_para_env_type), POINTER                    :: my_para_env, para_env
      TYPE(eip_environment_type), POINTER                :: eip_env
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(f_env_p_type), DIMENSION(:), POINTER          :: f_envs_old
      TYPE(force_env_type), POINTER                      :: force_env, my_force_env
      TYPE(fp_type), POINTER                             :: fp_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(mp_perf_env_type), POINTER                    :: mp_perf_env
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER :: fe_section, force_env_section, force_env_sections, &
         fp_section, input_file, qmmm_section, qmmmx_section, root_section, subsys_section, &
         wrk_section
      TYPE(timer_env_type), POINTER                      :: timer_env

      IF(.NOT.(ASSOCIATED(input_declaration)))CALL cp__a("f77_interface.F",583)
      NULLIFY (para_env, force_env, timer_env, mp_perf_env, globenv, meta_env, &
               fp_env, eip_env, mixed_env, qs_env, qmmm_env)
      new_env_id = -1
      IF (PRESENT(mpi_comm)) THEN
         CALL cp_para_env_create(para_env, group=mpi_comm, owns_group=.FALSE.)
      ELSE
         para_env => default_para_env
         CALL cp_para_env_retain(para_env)
      END IF

      CALL timeset(routineN, handle)

      CALL m_getcwd(old_dir)
      wdir = old_dir
      IF (PRESENT(work_dir)) THEN
         IF (work_dir /= " ") THEN
            CALL m_chdir(work_dir, ierr2)
            IF (ierr2 /= 0) THEN
               IF (PRESENT(ierr)) ierr = ierr2
               RETURN
            END IF
            wdir = work_dir
         END IF
      END IF

      IF (PRESENT(output_unit)) THEN
         unit_nr = output_unit
      ELSE
         IF (para_env%mepos == para_env%source) THEN
            IF (output_path == "__STD_OUT__") THEN
               unit_nr = default_output_unit
            ELSE
               CALL open_file(file_name=output_path, file_status="UNKNOWN", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=unit_nr)
            END IF
         ELSE
            unit_nr = -1
         END IF
      END IF
      my_owns_out_unit = unit_nr /= default_output_unit
      IF (PRESENT(owns_out_unit)) my_owns_out_unit = owns_out_unit
      CALL globenv_create(globenv)
      CALL cp2k_init(para_env, output_unit=unit_nr, globenv=globenv, input_file_name=input_path, &
                     wdir=wdir)
      logger => cp_get_default_logger()
      ! warning this is dangerous, I did not check that all the subfunctions
      ! support it, the program might crash upon error

      NULLIFY (input_file)
      IF (PRESENT(input)) input_file => input
      IF (.NOT. ASSOCIATED(input_file)) THEN
         IF (PRESENT(initial_variables)) THEN
            input_file => read_input(input_declaration, input_path, initial_variables, para_env=para_env)
         ELSE
            input_file => read_input(input_declaration, input_path, empty_initial_variables, para_env=para_env)
         ENDIF
      ELSE
         CALL section_vals_retain(input_file)
      END IF
      CALL section_vals_val_get(input_file, "GLOBAL%ECHO_INPUT", &
                                l_val=my_echo)
      ! echo after check?
      IF (para_env%ionode .AND. my_echo) THEN
         CALL section_vals_write(input_file, unit_nr=cp_logger_get_default_unit_nr(logger), &
                                 hide_root=.TRUE., hide_defaults=.FALSE.)
      END IF
      ! XXXXXXXXXXXXXXXXXXXXXXXXXXX
      ! root_section => input_file
      ! XXXXXXXXXXXXXXXXXXXXXXXXXXX

      CALL check_cp2k_input(input_declaration, input_file, para_env=para_env, output_unit=unit_nr)
      ! XXXXXXXXXXXXXXXXXXXXXXXXXXX
      ! NULLIFY(input_file)
      ! XXXXXXXXXXXXXXXXXXXXXXXXXXX
      root_section => input_file
      CALL section_vals_retain(root_section)

      IF (n_f_envs+1 > SIZE(f_envs)) THEN
         f_envs_old => f_envs
         ALLOCATE (f_envs(n_f_envs+10))
         DO i = 1, n_f_envs
            f_envs(i)%f_env => f_envs_old(i)%f_env
         END DO
         DO i = n_f_envs+1, SIZE(f_envs)
            NULLIFY (f_envs(i)%f_env)
         END DO
         DEALLOCATE (f_envs_old)
      END IF

      CALL cp2k_read(root_section, para_env, globenv)

      CALL cp2k_setup(root_section, para_env, globenv)
      ! Group Distribution
      ALLOCATE (group_distribution(0:para_env%num_pe-1))
      group_distribution = 0
      lgroup_distribution => group_distribution
      ! Setup all possible force_env
      force_env_sections => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
      CALL section_vals_val_get(root_section, "MULTIPLE_FORCE_EVALS%MULTIPLE_SUBSYS", &
                                l_val=multiple_subsys)
      CALL multiple_fe_list(force_env_sections, root_section, i_force_eval, nforce_eval)
      ! Enforce the deletion of the subsys (unless not explicitly required)
      IF (.NOT. multiple_subsys) THEN
         DO iforce_eval = 2, nforce_eval
            wrk_section => section_vals_get_subs_vals(force_env_sections, "SUBSYS", &
                                                      i_rep_section=i_force_eval(iforce_eval))
            CALL section_vals_remove_values(wrk_section)
         END DO
      END IF
      nsubforce_size = nforce_eval-1
      use_multiple_para_env = .FALSE.
      use_motion_section = .TRUE.
      DO iforce_eval = 1, nforce_eval
         NULLIFY (force_env_section, my_force_env, subsys_section)
         ! Reference subsys from the first ordered force_eval
         IF (.NOT. multiple_subsys) THEN
            subsys_section => section_vals_get_subs_vals(force_env_sections, "SUBSYS", &
                                                         i_rep_section=i_force_eval(1))
         END IF
         ! Handling para_env in case of multiple force_eval
         IF (use_multiple_para_env) THEN
            ! Check that the order of the force_eval is the correct one
            CALL section_vals_val_get(force_env_sections, "METHOD", i_val=method_name_id, &
                                      i_rep_section=i_force_eval(1))
            IF (method_name_id /= do_mixed) &
               CALL cp_abort(cp__l("f77_interface.F",710), &
                             "In case of multiple force_eval the MAIN force_eval (the first in the list of FORCE_EVAL_ORDER or "// &
                             "the one omitted from that order list) must be a MIXED_ENV type calculation. Please check your "// &
                             "input file and possibly correct the MULTIPLE_FORCE_EVAL%FORCE_EVAL_ORDER. ")
            check = ASSOCIATED(force_env%mixed_env%sub_para_env)
            IF(.NOT.(check))CALL cp__a("f77_interface.F",715)
            ngroups = force_env%mixed_env%ngroups
            my_group = lgroup_distribution(para_env%mepos)
            isubforce_eval = iforce_eval-1
            ! If task not allocated on this procs skip setup..
            IF (MODULO(isubforce_eval-1, ngroups) /= my_group) CYCLE
            my_para_env => force_env%mixed_env%sub_para_env(my_group+1)%para_env
            my_logger => force_env%mixed_env%sub_logger(my_group+1)%p
            CALL cp_rm_default_logger()
            CALL cp_add_default_logger(my_logger)
         ELSE
            my_para_env => para_env
         END IF

         ! Initialize force_env_section
         ! No need to allocate one more force_env_section if only 1 force_eval
         ! is provided.. this is in order to save memory..
         IF (nforce_eval > 1) THEN
            CALL section_vals_duplicate(force_env_sections, force_env_section, &
                                        i_force_eval(iforce_eval), i_force_eval(iforce_eval))
            IF (iforce_eval /= 1) use_motion_section = .FALSE.
         ELSE
            force_env_section => force_env_sections
            use_motion_section = .TRUE.
         END IF
         CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)

         IF (method_name_id == do_qmmm) THEN
            qmmmx_section => section_vals_get_subs_vals(force_env_section, "QMMM%FORCE_MIXING")
            CALL section_vals_get(qmmmx_section, explicit=do_qmmm_force_mixing)
            IF (do_qmmm_force_mixing) &
               method_name_id = do_qmmmx ! QMMM Force-Mixing has its own (hidden) method_id
         ENDIF

         SELECT CASE (method_name_id)
         CASE (do_fist)
            CALL fist_create_force_env(my_force_env, root_section, my_para_env, globenv, &
                                       force_env_section=force_env_section, subsys_section=subsys_section, &
                                       use_motion_section=use_motion_section)

         CASE (do_qs)
            CALL qs_env_create(qs_env, globenv)
            CALL qs_init(qs_env, my_para_env, globenv, root_section, force_env_section=force_env_section, &
                         subsys_section=subsys_section, use_motion_section=use_motion_section)
            CALL force_env_create(my_force_env, root_section, qs_env=qs_env, para_env=my_para_env, globenv=globenv, &
                                  force_env_section=force_env_section)
            CALL qs_env_release(qs_env)

         CASE (do_qmmm)
            qmmm_section => section_vals_get_subs_vals(force_env_section, "QMMM")
            CALL qmmm_env_create(qmmm_env, root_section, my_para_env, globenv, &
                                 force_env_section, qmmm_section, subsys_section, use_motion_section)
            CALL force_env_create(my_force_env, root_section, qmmm_env=qmmm_env, para_env=my_para_env, &
                                  globenv=globenv, force_env_section=force_env_section)
            CALL qmmm_env_release(qmmm_env)

         CASE (do_qmmmx)
            CALL qmmmx_env_create(qmmmx_env, root_section, my_para_env, globenv, &
                                  force_env_section, subsys_section, use_motion_section)
            CALL force_env_create(my_force_env, root_section, qmmmx_env=qmmmx_env, para_env=my_para_env, &
                                  globenv=globenv, force_env_section=force_env_section)
            CALL qmmmx_env_release(qmmmx_env)

         CASE (do_eip)
            CALL eip_env_create(eip_env)
            CALL eip_init(eip_env, root_section, my_para_env, force_env_section=force_env_section, &
                          subsys_section=subsys_section)
            CALL force_env_create(my_force_env, root_section, eip_env=eip_env, para_env=my_para_env, &
                                  globenv=globenv, force_env_section=force_env_section)
            CALL eip_env_release(eip_env)

         CASE (do_mixed)
            CALL mixed_create_force_env(mixed_env, root_section, my_para_env, &
                                        force_env_section=force_env_section, n_subforce_eval=nsubforce_size, &
                                        use_motion_section=use_motion_section)
            CALL force_env_create(my_force_env, root_section, mixed_env=mixed_env, para_env=my_para_env, &
                                  globenv=globenv, force_env_section=force_env_section)
            CALL mixed_env_release(mixed_env)
            !TODO: the sub_force_envs should really be created via recursion
            use_multiple_para_env = .TRUE.
            CALL cp_add_default_logger(logger) ! just to get the logger swapping started
            lgroup_distribution => my_force_env%mixed_env%group_distribution

         CASE default
            CALL create_force_eval_section(section)
            keyword => section_get_keyword(section, "METHOD")
            CALL keyword_get(keyword, enum=enum)
            CALL cp_abort(cp__l("f77_interface.F",802), &
                          "Invalid METHOD <"//TRIM(enum_i2c(enum, method_name_id))// &
                          "> was specified, ")
            CALL section_release(section)
         END SELECT

         NULLIFY (meta_env, fp_env)
         IF (use_motion_section) THEN
            ! Metadynamics Setup
            fe_section => section_vals_get_subs_vals(root_section, "MOTION%FREE_ENERGY")
            CALL metadyn_read(meta_env, my_force_env, root_section, my_para_env, fe_section)
            CALL force_env_set(my_force_env, meta_env=meta_env)
            CALL meta_env_release(meta_env)
            ! Flexible Partition Setup
            fp_section => section_vals_get_subs_vals(root_section, "MOTION%FLEXIBLE_PARTITIONING")
            CALL fp_env_create(fp_env)
            CALL fp_env_read(fp_env, fp_section)
            CALL fp_env_write(fp_env, fp_section)
            CALL force_env_set(my_force_env, fp_env=fp_env)
            CALL fp_env_release(fp_env)
         END IF
         ! Handle multiple force_eval
         IF (nforce_eval > 1 .AND. iforce_eval == 1) THEN
            ALLOCATE (my_force_env%sub_force_env(nsubforce_size))
            ! Nullify subforce_env
            DO k = 1, nsubforce_size
               NULLIFY (my_force_env%sub_force_env(k)%force_env)
            END DO
         END IF
         ! Reference the right force_env
         IF (iforce_eval == 1) THEN
            force_env => my_force_env
         ELSE
            force_env%sub_force_env(iforce_eval-1)%force_env => my_force_env
         END IF
         ! Multiple para env for sub_force_eval
         IF (.NOT. use_multiple_para_env) THEN
            lgroup_distribution = iforce_eval
         END IF
         ! Release force_env_section
         IF (nforce_eval > 1) CALL section_vals_release(force_env_section)
      END DO
      IF (use_multiple_para_env) &
         CALL cp_rm_default_logger()
      DEALLOCATE (group_distribution)
      DEALLOCATE (i_force_eval)
      timer_env => get_timer_env()
      mp_perf_env => get_mp_perf_env()
      CALL mp_max(last_f_env_id, para_env%group)
      last_f_env_id = last_f_env_id+1
      new_env_id = last_f_env_id
      n_f_envs = n_f_envs+1
      CALL f_env_create(f_envs(n_f_envs)%f_env, logger=logger, &
                        timer_env=timer_env, mp_perf_env=mp_perf_env, force_env=force_env, &
                        id_nr=last_f_env_id, old_dir=old_dir)
      CALL force_env_release(force_env)
      CALL globenv_release(globenv)
      CALL section_vals_release(root_section)
      CALL cp_para_env_release(para_env)
      CALL f_env_rm_defaults(f_envs(n_f_envs)%f_env, ierr=ierr)
      CALL timestop(handle)

   END SUBROUTINE create_force_env

! **************************************************************************************************
!> \brief deallocates the force_env with the given id
!> \param env_id the id of the force_env to remove
!> \param ierr will contain a number different from 0 if
!> \author fawzi
!> \note
!>      The following routines need to be synchronized wrt. adding/removing
!>      of the default environments (logging, perormance,error):
!>      environment:cp2k_init, environment:cp2k_finalize,
!>      f77_interface:f_env_add_defaults, f77_interface:f_env_rm_defaults,
!>      f77_interface:create_force_env, f77_interface:destroy_force_env
! **************************************************************************************************
   RECURSIVE SUBROUTINE destroy_force_env(env_id, ierr)
      INTEGER, INTENT(in)                                :: env_id
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'destroy_force_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: env_pos, i
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: root_section

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      env_pos = get_pos_of_env(env_id)
      n_f_envs = n_f_envs-1
      DO i = env_pos, n_f_envs
         f_envs(i)%f_env => f_envs(i+1)%f_env
      END DO
      NULLIFY (f_envs(n_f_envs+1)%f_env)

      CALL force_env_get(f_env%force_env, globenv=globenv, &
                         root_section=root_section, para_env=para_env)

      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("f77_interface.F",903)
      CALL globenv_retain(globenv)
      CALL f_env_dealloc(f_env)
      CALL cp2k_finalize(root_section, para_env, globenv, f_env%old_path)
      CALL section_vals_release(root_section)
      CALL globenv_release(globenv)
      DEALLOCATE (f_env)
      ierr = 0
   END SUBROUTINE destroy_force_env

! **************************************************************************************************
!> \brief returns the number of atoms in the given force env
!> \param env_id id of the force_env
!> \param n_atom ...
!> \param ierr will return a number different from 0 if there was an error
!> \date   22.11.2010 (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE get_natom(env_id, n_atom, ierr)

      INTEGER, INTENT(IN)                                :: env_id
      INTEGER, INTENT(OUT)                               :: n_atom, ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_natom', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      n_atom = 0
      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      n_atom = force_env_get_natom(f_env%force_env)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_natom

! **************************************************************************************************
!> \brief returns the number of particles in the given force env
!> \param env_id id of the force_env
!> \param n_particle ...
!> \param ierr will return a number different from 0 if there was an error
!> \author Matthias Krack
!>
! **************************************************************************************************
   SUBROUTINE get_nparticle(env_id, n_particle, ierr)

      INTEGER, INTENT(IN)                                :: env_id
      INTEGER, INTENT(OUT)                               :: n_particle, ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_nparticle', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      n_particle = 0
      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      n_particle = force_env_get_nparticle(f_env%force_env)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_nparticle

! **************************************************************************************************
!> \brief gets a cell
!> \param env_id id of the force_env
!> \param cell the array with the cell matrix
!> \param per periodicity
!> \param ierr will return a number different from 0 if there was an error
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE get_cell(env_id, cell, per, ierr)

      INTEGER, INTENT(IN)                                :: env_id
      REAL(KIND=DP), DIMENSION(3, 3)                     :: cell
      INTEGER, DIMENSION(3), OPTIONAL                    :: per
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_cell', routineP = moduleN//':'//routineN

      TYPE(cell_type), POINTER                           :: cell_full
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      NULLIFY (cell_full)
      CALL force_env_get(f_env%force_env, cell=cell_full)
      IF(.NOT.(ASSOCIATED(cell_full)))CALL cp__a("f77_interface.F",987)
      cell = cell_full%hmat
      IF (PRESENT(per)) per(:) = cell_full%perd(:)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_cell

! **************************************************************************************************
!> \brief gets a result from CP2K that is a real 1D array
!> \param env_id id of the force_env
!> \param description the tag of the result
!> \param N ...
!> \param RESULT ...
!> \param res_exist ...
!> \param ierr will return a number different from 0 if there was an error
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE get_result_r1(env_id, description, N, RESULT, res_exist, ierr)
      INTEGER                                            :: env_id
      CHARACTER(LEN=default_string_length)               :: description
      INTEGER                                            :: N
      REAL(KIND=dp), DIMENSION(1:N)                      :: RESULT
      LOGICAL, OPTIONAL                                  :: res_exist
      INTEGER                                            :: ierr

      INTEGER                                            :: nres
      LOGICAL                                            :: exist_res
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env, subsys, results)
      CALL f_env_add_defaults(env_id, f_env)

      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, results=results)
      ! first test for the result
      IF (PRESENT(res_exist)) THEN
         res_exist = test_for_result(results, description=description)
         exist_res = res_exist
      ELSE
         exist_res = .TRUE.
      END IF
      ! if existing (or assuming the existance) read the results
      IF (exist_res) THEN
         CALL get_results(results, description=description, n_rep=nres)
         CALL get_results(results, description=description, values=RESULT, nval=nres)
      END IF

      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_result_r1

! **************************************************************************************************
!> \brief gets the forces of the particles
!> \param env_id id of the force_env
!> \param frc the array where to write the forces
!> \param n_el number of positions (3*nparticle) just to check
!> \param ierr will return a number different from 0 if there was an error
!> \date   22.11.2010 (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE get_force(env_id, frc, n_el, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(KIND=dp), DIMENSION(1:n_el)                   :: frc
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_force', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      CALL force_env_get_frc(f_env%force_env, frc, n_el)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_force

! **************************************************************************************************
!> \brief gets the stress tensor
!> \param env_id id of the force_env
!> \param stress_tensor the array where to write the stress tensor
!> \param ierr will return a number different from 0 if there was an error
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE get_stress_tensor(env_id, stress_tensor, ierr)

      INTEGER, INTENT(IN)                                :: env_id
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: stress_tensor
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_stress_tensor', &
         routineP = moduleN//':'//routineN

      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (f_env, subsys, virial, cell)
      stress_tensor(:, :) = 0.0_dp

      CALL f_env_add_defaults(env_id, f_env)
      CALL force_env_get(f_env%force_env, subsys=subsys, cell=cell)
      CALL cp_subsys_get(subsys, virial=virial)
      IF (virial%pv_availability) THEN
         stress_tensor(:, :) = virial%pv_virial(:, :)/cell%deth
      ENDIF
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_stress_tensor

! **************************************************************************************************
!> \brief gets the positions of the particles
!> \param env_id id of the force_env
!> \param pos the array where to write the positions
!> \param n_el number of positions (3*nparticle) just to check
!> \param ierr will return a number different from 0 if there was an error
!> \date   22.11.2010 (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE get_pos(env_id, pos, n_el, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(KIND=DP), DIMENSION(1:n_el)                   :: pos
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_pos', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      CALL force_env_get_pos(f_env%force_env, pos, n_el)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_pos

! **************************************************************************************************
!> \brief gets the velocities of the particles
!> \param env_id id of the force_env
!> \param vel the array where to write the velocities
!> \param n_el number of velocities (3*nparticle) just to check
!> \param ierr will return a number different from 0 if there was an error
!> \author fawzi
!> date    22.11.2010 (MK)
! **************************************************************************************************
   SUBROUTINE get_vel(env_id, vel, n_el, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(KIND=DP), DIMENSION(1:n_el)                   :: vel
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_vel', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      CALL force_env_get_vel(f_env%force_env, vel, n_el)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_vel

! **************************************************************************************************
!> \brief sets a new cell
!> \param env_id id of the force_env
!> \param new_cell the array with the cell matrix
!> \param ierr will return a number different from 0 if there was an error
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE set_cell(env_id, new_cell, ierr)

      INTEGER, INTENT(IN)                                :: env_id
      REAL(KIND=DP), DIMENSION(3, 3)                     :: new_cell
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_cell', routineP = moduleN//':'//routineN

      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env, cell, subsys)
      CALL f_env_add_defaults(env_id, f_env)
      NULLIFY (cell)
      CALL force_env_get(f_env%force_env, cell=cell)
      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("f77_interface.F",1175)
      cell%hmat = new_cell
      CALL init_cell(cell)
      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL cp_subsys_set(subsys, cell=cell)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE set_cell

! **************************************************************************************************
!> \brief sets the positions of the particles
!> \param env_id id of the force_env
!> \param new_pos the array with the new positions
!> \param n_el number of positions (3*nparticle) just to check
!> \param ierr will return a number different from 0 if there was an error
!> \date   22.11.2010 updated (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE set_pos(env_id, new_pos, n_el, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(KIND=dp), DIMENSION(1:n_el)                   :: new_pos
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_pos', routineP = moduleN//':'//routineN

      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      NULLIFY (subsys)
      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL unpack_subsys_particles(subsys=subsys, r=new_pos)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE set_pos

! **************************************************************************************************
!> \brief sets the velocities of the particles
!> \param env_id id of the force_env
!> \param new_vel the array with the new velocities
!> \param n_el number of velocities (3*nparticle) just to check
!> \param ierr will return a number different from 0 if there was an error
!> \date   22.11.2010 updated (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE set_vel(env_id, new_vel, n_el, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(kind=dp), DIMENSION(1:n_el)                   :: new_vel
      INTEGER, INTENT(OUT)                               :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_vel', routineP = moduleN//':'//routineN

      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      NULLIFY (subsys)
      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL unpack_subsys_particles(subsys=subsys, v=new_vel)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE set_vel

! **************************************************************************************************
!> \brief updates the energy and the forces of given force_env
!> \param env_id id of the force_env that you want to update
!> \param calc_force if the forces should be updated, if false the forces
!>        might be wrong.
!> \param ierr will return a number different from 0 if there was an error
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE calc_energy_force(env_id, calc_force, ierr)

      INTEGER, INTENT(in)                                :: env_id
      LOGICAL, INTENT(in)                                :: calc_force
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_energy_force', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      logger => cp_get_default_logger()
      CALL cp_iterate(logger%iter_info) ! add one to the iteration count
      CALL force_env_calc_energy_force(f_env%force_env, calc_force=calc_force)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE calc_energy_force

! **************************************************************************************************
!> \brief returns the energy of the last configuration calculated
!> \param env_id id of the force_env that you want to update
!> \param e_pot the potential energy of the system
!> \param ierr will return a number different from 0 if there was an error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE get_energy(env_id, e_pot, ierr)

      INTEGER, INTENT(in)                                :: env_id
      REAL(kind=dp), INTENT(out)                         :: e_pot
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'get_energy', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(env_id, f_env)
      CALL force_env_get(f_env%force_env, potential_energy=e_pot)
      CALL f_env_rm_defaults(f_env, ierr)

   END SUBROUTINE get_energy

! **************************************************************************************************
!> \brief returns the energy of the configuration given by the positions
!>      passed as argument
!> \param env_id id of the force_env that you want to update
!> \param pos array with the positions
!> \param n_el number of elements in pos (3*natom)
!> \param e_pot the potential energy of the system
!> \param ierr will return a number different from 0 if there was an error
!> \author fawzi
!> \note
!>      utility call
! **************************************************************************************************
   RECURSIVE SUBROUTINE calc_energy(env_id, pos, n_el, e_pot, ierr)

      INTEGER, INTENT(IN)                                :: env_id, n_el
      REAL(KIND=dp), DIMENSION(1:n_el), INTENT(IN)       :: pos
      REAL(KIND=dp), INTENT(OUT)                         :: e_pot
      INTEGER, INTENT(OUT)                               :: ierr

      REAL(KIND=dp), DIMENSION(1)                        :: dummy_f

      CALL calc_force(env_id, pos, n_el, e_pot, dummy_f, 0, ierr)

   END SUBROUTINE calc_energy

! **************************************************************************************************
!> \brief returns the energy of the configuration given by the positions
!>      passed as argument
!> \param env_id id of the force_env that you want to update
!> \param pos array with the positions
!> \param n_el_pos number of elements in pos (3*natom)
!> \param e_pot the potential energy of the system
!> \param force array that will contain the forces
!> \param n_el_force number of elements in force (3*natom). If 0 the
!>        forces are not calculated
!> \param ierr will return a number different from 0 if there was an error
!> \author fawzi
!> \note
!>      utility call, but actually it could be a better and more efficient
!>      interface to connect to other codes if cp2k would be deeply
!>      refactored
! **************************************************************************************************
   RECURSIVE SUBROUTINE calc_force(env_id, pos, n_el_pos, e_pot, force, n_el_force, ierr)

      INTEGER, INTENT(in)                                :: env_id, n_el_pos
      REAL(kind=dp), DIMENSION(1:n_el_pos), INTENT(in)   :: pos
      REAL(kind=dp), INTENT(out)                         :: e_pot
      INTEGER, INTENT(in)                                :: n_el_force
      REAL(kind=dp), DIMENSION(1:n_el_force), &
         INTENT(inout)                                   :: force
      INTEGER, INTENT(out)                               :: ierr

      LOGICAL                                            :: calc_f

      calc_f = (n_el_force /= 0)
      CALL set_pos(env_id, pos, n_el_pos, ierr)
      IF (ierr == 0) CALL calc_energy_force(env_id, calc_f, ierr)
      IF (ierr == 0) CALL get_energy(env_id, e_pot, ierr)
      IF (calc_f .AND. (ierr == 0)) CALL get_force(env_id, force, n_el_force, ierr)

   END SUBROUTINE calc_force

! **************************************************************************************************
!> \brief performs a check of the input
!> \param input_declaration ...
!> \param input_file_path the path of the input file to check
!> \param output_file_path path of the output file (to which it is appended)
!>        if it is "__STD_OUT__" the default_output_unit is used
!> \param echo_input if the parsed input should be written out with all the
!>        defaults made explicit
!> \param mpi_comm the mpi communicator (if not given it uses the default
!>        one)
!> \param ierr error control, if different from 0 there was an error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE check_input(input_declaration, input_file_path, output_file_path, &
                          echo_input, mpi_comm, ierr)
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(len=*), INTENT(in)                       :: input_file_path, output_file_path
      LOGICAL, INTENT(in), OPTIONAL                      :: echo_input
      INTEGER, INTENT(in), OPTIONAL                      :: mpi_comm
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'check_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr
      LOGICAL                                            :: my_echo_input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input_file

      my_echo_input = .FALSE.
      IF (PRESENT(echo_input)) my_echo_input = echo_input

      IF (PRESENT(mpi_comm)) THEN
         NULLIFY (para_env)
         CALL cp_para_env_create(para_env, group=mpi_comm)
      ELSE
         para_env => default_para_env
         CALL cp_para_env_retain(para_env)
      END IF
      IF (para_env%mepos == para_env%source) THEN
         IF (output_file_path == "__STD_OUT__") THEN
            unit_nr = default_output_unit
         ELSE
            CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                           file_action="WRITE", file_position="APPEND", &
                           unit_number=unit_nr)
         END IF
      ELSE
         unit_nr = -1
      END IF

      NULLIFY (logger)
      CALL cp_logger_create(logger, para_env=para_env, &
                            default_global_unit_nr=unit_nr, &
                            close_global_unit_on_dealloc=.FALSE.)
      CALL cp_add_default_logger(logger)
      CALL cp_logger_release(logger)

      input_file => read_input(input_declaration, input_file_path, initial_variables=empty_initial_variables, &
                               para_env=para_env)
      CALL check_cp2k_input(input_declaration, input_file, para_env=para_env, output_unit=unit_nr)
      IF (my_echo_input .AND. para_env%mepos == para_env%source) THEN
         CALL section_vals_write(input_file, &
                                 unit_nr=cp_logger_get_default_unit_nr(logger, local=.FALSE.), hide_root=.TRUE., &
                                 hide_defaults=.FALSE.)
      END IF
      CALL section_vals_release(input_file)

      CALL cp_logger_release(logger)
      CALL cp_para_env_release(para_env)
      ierr = 0
      CALL cp_rm_default_logger()
   END SUBROUTINE check_input

! **************************************************************************************************
!> \brief perform the shake procedure (enforce constraints)
!> \param f_env_id the force env on which to apply the shake
!> \param dt timestep
!> \param shake_tol tolerance for shake
!> \param ierr ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE do_shake(f_env_id, dt, shake_tol, ierr)
      INTEGER, INTENT(in)                                :: f_env_id
      REAL(kind=dp), INTENT(in)                          :: dt, shake_tol
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'do_shake', routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(f_env_id, f_env)
      CALL force_env_shake(f_env%force_env, &
                           dt=dt, shake_tol=shake_tol)
      CALL f_env_rm_defaults(f_env, ierr)
   END SUBROUTINE do_shake

END MODULE f77_interface
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types used to handle many replica of the same system that differ only
!>      in atom positions, and velocity.
!>      This is useful for things like path integrals or nudged elastic band
!> \note
!>      this is a stupid implementation that replicates all the information
!>      about the replicas, if you really want to do a *lot* of replicas on
!>      a lot of processors you should think about distributiong also that
!>      information
!> \par History
!>      09.2005 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE replica_types
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_rm_iter_level
   USE cp_para_env,                     ONLY: cp_cart_release,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_cart_type,&
                                              cp_para_env_type
   USE cp_result_methods,               ONLY: cp_results_mp_bcast
   USE cp_result_types,                 ONLY: cp_result_p_type,&
                                              cp_result_release
   USE f77_interface,                   ONLY: destroy_force_env,&
                                              f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE message_passing,                 ONLY: mp_sum
   USE qs_wf_history_types,             ONLY: qs_wf_history_p_type,&
                                              wfi_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   LOGICAL, SAVE, PRIVATE :: module_initialized = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'replica_types'

   PUBLIC :: replica_env_type
   PUBLIC :: rep_env_release
   PUBLIC :: rep_env_sync, rep_env_sync_results, rep_envs_add_rep_env
   PUBLIC :: rep_envs_get_rep_env

! **************************************************************************************************
!> \brief keeps replicated information about the replicas
!> \param ref_count reference count
!> \param id_nr identity number (unique or each replica_env)
!> \param nrep number of replicas
!> \param nparticle number of particles (usually atoms) in each replica
!> \param ndim = 3*nparticle
!> \param f_env_id id of the force env that will do the calculations for the
!>        replicas owned by this processor
!> \param r ,v,f: positions, velocities and forces of the replicas.
!>        the indexing is as follow (idir,iat,irep)
!> \param replica_owner which replica group number owns the replica irep
!> \param cart 2d distribution of the processors for the replicas,
!>        a column (or row if row_force was true in the rep_env_create call)
!>        work together on the same force_env (i.e. changing the
!>        row (column) you stay in the same replica), rows (columns) have
!>        different replicas
!> \param force_dim which dimension of cart works on forces together
!>        used to be hardcoded to 1. Default is still 1, will
!>        be 2 if row_force is true in the rep_env_create call.
!> \param para_env the global para env that contains all the replicas,
!>        this is just the cart as para_env
!> \param para_env_f parallel environment of the underlying force
!>        environment
!> \param inter_rep_rank mapping replica group number -> rank in para_env_inter_rep
!>        (this used to be col_rank)
!> \param para_env_inter_rep parallel environment between replica
!> \param force_rank mapping number of processor in force env -> rank in para_env_f
!>        (this used to be row_rank)
!> \param local_rep_indices indices of the local replicas, starting at 1
!> \param rep_is_local logical if specific replica is a local one.
!> \param my_rep_group which replica group number this process belongs to
!>        (this used to be just cart%mepos(2) but with transposing the cart
!>        (row_force=.true.) became cart%mepos(1), and to generalize this it
!>        is now a separate variable, so one does not need to know
!>        which way the cart is mapped.)
!> \param wf_history wavefunction history for the owned replicas
!> \param keep_wf_history if the wavefunction history for the owned replicas
!>        should be kept
!> \author fawzi
! **************************************************************************************************
   TYPE replica_env_type
      INTEGER                                           :: ref_count, id_nr, f_env_id, &
                                                           nrep, ndim, nparticle, &
                                                           my_rep_group, force_dim
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: r, v, f
      LOGICAL                                           :: sync_v, keep_wf_history
      CHARACTER(LEN=default_path_length)                :: original_project_name
      TYPE(qs_wf_history_p_type), DIMENSION(:), POINTER :: wf_history
      TYPE(cp_result_p_type), DIMENSION(:), POINTER       :: results
      INTEGER, DIMENSION(:), POINTER                    :: local_rep_indices
      INTEGER, DIMENSION(:), POINTER                    :: replica_owner, force_rank, &
                                                           inter_rep_rank
      LOGICAL, DIMENSION(:), POINTER                    :: rep_is_local
      TYPE(cp_para_cart_type), POINTER                  :: cart
      TYPE(cp_para_env_type), POINTER                   :: para_env, para_env_f, &
                                                           para_env_inter_rep
   END TYPE replica_env_type

! **************************************************************************************************
!> \brief ****s* replica_types/replica_env_p_type *
!>
!>      to build arrays of pointers to a replica_env_type
!> \param rep_env the pointer to the replica_env
!> \author fawzi
! **************************************************************************************************
   TYPE replica_env_p_type
      TYPE(replica_env_type), POINTER                   :: rep_env
   END TYPE replica_env_p_type

   TYPE(replica_env_p_type), POINTER, DIMENSION(:), PRIVATE :: rep_envs

CONTAINS

! **************************************************************************************************
!> \brief releases the given replica environment
!> \param rep_env the replica environment to release
!> \author fawzi
!> \note
!>      here and not in replica_types to allow the use of replica_env_type
!>      in a force_env (call to destroy_force_env gives circular dep)
! **************************************************************************************************
   SUBROUTINE rep_env_release(rep_env)
      TYPE(replica_env_type), POINTER                    :: rep_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ierr

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(rep_env)) THEN
         IF(.NOT.(rep_env%ref_count > 0))CALL cp__a("replica_types.F",146)
         rep_env%ref_count = rep_env%ref_count-1
         IF (rep_env%ref_count == 0) THEN
            CALL rep_env_destroy_low(rep_env%id_nr, ierr)
            IF (rep_env%f_env_id > 0) THEN
               CALL destroy_force_env(rep_env%f_env_id, ierr)
               IF(.NOT.(ierr == 0))CALL cp__a("replica_types.F",152)
            END IF
            IF (ASSOCIATED(rep_env%r)) THEN
               DEALLOCATE (rep_env%r)
            END IF
            IF (ASSOCIATED(rep_env%v)) THEN
               DEALLOCATE (rep_env%v)
            END IF
            IF (ASSOCIATED(rep_env%f)) THEN
               DEALLOCATE (rep_env%f)
            END IF
            IF (ASSOCIATED(rep_env%wf_history)) THEN
               DO i = 1, SIZE(rep_env%wf_history)
                  CALL wfi_release(rep_env%wf_history(i)%wf_history)
               END DO
               DEALLOCATE (rep_env%wf_history)
            END IF
            IF (ASSOCIATED(rep_env%results)) THEN
               DO i = 1, SIZE(rep_env%results)
                  CALL cp_result_release(rep_env%results(i)%results)
               END DO
               DEALLOCATE (rep_env%results)
            END IF
            DEALLOCATE (rep_env%local_rep_indices)
            DEALLOCATE (rep_env%rep_is_local)
            IF (ASSOCIATED(rep_env%replica_owner)) THEN
               DEALLOCATE (rep_env%replica_owner)
            END IF
            DEALLOCATE (rep_env%inter_rep_rank, rep_env%force_rank)
            CALL cp_cart_release(rep_env%cart)
            CALL cp_para_env_release(rep_env%para_env)
            CALL cp_para_env_release(rep_env%para_env_f)
            CALL cp_para_env_release(rep_env%para_env_inter_rep)
            CALL rep_envs_rm_rep_env(rep_env)
            DEALLOCATE (rep_env)
         END IF
      END IF
      NULLIFY (rep_env)
      CALL timestop(handle)
   END SUBROUTINE rep_env_release

! **************************************************************************************************
!> \brief initializes the destruction of the replica_env
!> \param rep_env_id id_nr of the replica environment that should be initialized
!> \param ierr will be non zero if there is an initialization error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rep_env_destroy_low(rep_env_id, ierr)
      INTEGER, INTENT(in)                                :: rep_env_id
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_destroy_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: stat
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(replica_env_type), POINTER                    :: rep_env

      rep_env => rep_envs_get_rep_env(rep_env_id, ierr=stat)
      IF (.NOT. ASSOCIATED(rep_env)) &
         CALL cp__b("replica_types.F",213,"could not find rep_env with id_nr"//cp_to_string(rep_env_id))
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
      logger => cp_get_default_logger()
      CALL cp_rm_iter_level(iteration_info=logger%iter_info, &
                            level_name="REPLICA_EVAL")
      CALL f_env_rm_defaults(f_env, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("replica_types.F",219)
   END SUBROUTINE rep_env_destroy_low

! **************************************************************************************************
!> \brief sends the data from each replica to all the other
!>      on replica j/=i data from replica i overwrites val(:,i)
!> \param rep_env replica environment
!> \param vals the values to synchronize (second index runs over replicas)
!> \author fawzi
!> \note
!>      could be optimized: bcast in inter_rep, all2all or shift vs sum
! **************************************************************************************************
   SUBROUTINE rep_env_sync(rep_env, vals)
      TYPE(replica_env_type), POINTER                    :: rep_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(inout)      :: vals

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_sync', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, irep

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(rep_env)))CALL cp__a("replica_types.F",240)
      IF(.NOT.(rep_env%ref_count > 0))CALL cp__a("replica_types.F",241)
      IF(.NOT.(SIZE(vals, 2) == rep_env%nrep))CALL cp__a("replica_types.F",242)
      DO irep = 1, rep_env%nrep
         IF (.NOT. rep_env%rep_is_local(irep)) THEN
            vals(:, irep) = 0._dp
         END IF
      END DO
      CALL mp_sum(vals, rep_env%para_env_inter_rep%group)
      CALL timestop(handle)
   END SUBROUTINE rep_env_sync

! **************************************************************************************************
!> \brief sends the data from each replica to all the other
!>      in this case the result type is passed
!> \param rep_env replica environment
!> \param results is an array of result_types
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE rep_env_sync_results(rep_env, results)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(cp_result_p_type), DIMENSION(:), POINTER      :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_sync_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, irep, nrep, source

      CALL timeset(routineN, handle)
      nrep = rep_env%nrep
      IF(.NOT.(ASSOCIATED(rep_env)))CALL cp__a("replica_types.F",270)
      IF(.NOT.(rep_env%ref_count > 0))CALL cp__a("replica_types.F",271)
      IF(.NOT.(SIZE(results) == rep_env%nrep))CALL cp__a("replica_types.F",272)
      DO irep = 1, nrep
!         source = 0
!         IF (rep_env%rep_is_local(irep)) source = rep_env%para_env_inter_rep%mepos
!         CALL mp_sum(source, rep_env%para_env_inter_rep%group)
! above three lines should be the same as just:
         source = rep_env%inter_rep_rank(rep_env%replica_owner(irep))
         CALL cp_results_mp_bcast(results(irep)%results, source, rep_env%para_env_inter_rep)
      END DO
      CALL timestop(handle)
   END SUBROUTINE rep_env_sync_results

! **************************************************************************************************
!> \brief returns the replica environment with the given id_nr
!> \param id_nr the id_nr of the requested rep_envs
!> \param ierr ...
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION rep_envs_get_rep_env(id_nr, ierr) RESULT(res)
      INTEGER, INTENT(in)                                :: id_nr
      INTEGER, INTENT(OUT)                               :: ierr
      TYPE(replica_env_type), POINTER                    :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_envs_get_rep_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      NULLIFY (res)
      ierr = -1
      IF (module_initialized) THEN
         IF (ASSOCIATED(rep_envs)) THEN
            DO i = 1, SIZE(rep_envs)
               IF (rep_envs(i)%rep_env%id_nr == id_nr) THEN
                  res => rep_envs(i)%rep_env
                  ierr = 0
                  EXIT
               END IF
            END DO
         END IF
      END IF
   END FUNCTION rep_envs_get_rep_env

! **************************************************************************************************
!> \brief adds the given rep_env to the list of controlled rep_envs.
!> \param rep_env the rep_env to add
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rep_envs_add_rep_env(rep_env)
      TYPE(replica_env_type), POINTER                    :: rep_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_envs_add_rep_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, stat
      TYPE(replica_env_p_type), DIMENSION(:), POINTER    :: new_rep_envs
      TYPE(replica_env_type), POINTER                    :: rep_env2

      IF (ASSOCIATED(rep_env)) THEN
         rep_env2 => rep_envs_get_rep_env(rep_env%id_nr, ierr=stat)
         IF (.NOT. ASSOCIATED(rep_env2)) THEN
            IF (module_initialized) THEN
               IF (.NOT. ASSOCIATED(rep_envs)) THEN
                  ALLOCATE (rep_envs(1))
               ELSE
                  ALLOCATE (new_rep_envs(SIZE(rep_envs)+1))
                  DO i = 1, SIZE(rep_envs)
                     new_rep_envs(i)%rep_env => rep_envs(i)%rep_env
                  END DO
                  DEALLOCATE (rep_envs)
                  rep_envs => new_rep_envs
               END IF
            ELSE
               ALLOCATE (rep_envs(1))
            END IF
            rep_envs(SIZE(rep_envs))%rep_env => rep_env
            module_initialized = .TRUE.
         END IF
      END IF
   END SUBROUTINE rep_envs_add_rep_env

! **************************************************************************************************
!> \brief removes the given rep_env to the list of controlled rep_envs.
!> \param rep_env the rep_env to remove
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rep_envs_rm_rep_env(rep_env)
      TYPE(replica_env_type), POINTER                    :: rep_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_envs_rm_rep_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii
      TYPE(replica_env_p_type), DIMENSION(:), POINTER    :: new_rep_envs

      IF (ASSOCIATED(rep_env)) THEN
         IF(.NOT.(module_initialized))CALL cp__a("replica_types.F",369)
         ALLOCATE (new_rep_envs(SIZE(rep_envs)-1))
         ii = 0
         DO i = 1, SIZE(rep_envs)
            IF (rep_envs(i)%rep_env%id_nr /= rep_env%id_nr) THEN
               ii = ii+1
               new_rep_envs(ii)%rep_env => rep_envs(i)%rep_env
            END IF
         END DO
         IF(.NOT.(ii == SIZE(new_rep_envs)))CALL cp__a("replica_types.F",378)
         DEALLOCATE (rep_envs)
         rep_envs => new_rep_envs
         IF (SIZE(rep_envs) == 0) THEN
            DEALLOCATE (rep_envs)
         END IF
      END IF
   END SUBROUTINE rep_envs_rm_rep_env

END MODULE replica_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Functions handling the MOLDEN format. Split from mode_selective.
!> \author Teodoro Laino, 03.2009
! **************************************************************************************************
MODULE molden_utils
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cp_log_handling,                 ONLY: cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'molden_utils'
   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.

   PUBLIC :: molden_out

CONTAINS
! **************************************************************************************************
!> \brief writes the output for vibrational analysis in MOLDEN format
!> \param input ...
!> \param particles ...
!> \param freq ...
!> \param eigen_vec ...
!> \param intensities ...
!> \param calc_intens ...
!> \param dump_only_positive ...
!> \param logger ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE molden_out(input, particles, freq, eigen_vec, intensities, calc_intens, &
                         dump_only_positive, logger)

      TYPE(section_vals_type), POINTER                   :: input
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(KIND=dp), DIMENSION(:)                        :: freq
      REAL(KIND=dp), DIMENSION(:, :)                     :: eigen_vec
      REAL(KIND=dp), DIMENSION(:), POINTER               :: intensities
      LOGICAL                                            :: calc_intens, dump_only_positive
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'molden_out', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: i, iw, j, k, l

      iw = cp_print_key_unit_nr(logger, input, "VIBRATIONAL_ANALYSIS%PRINT%MOLDEN_VIB", &
                                extension=".mol", file_status='REPLACE')

      IF (iw .GT. 0) THEN
         IF(.NOT.(MOD(SIZE(eigen_vec, 1), 3) == 0))CALL cp__a("molden_utils.F",61)
         IF(.NOT.(SIZE(particles) == SIZE(eigen_vec, 1)/3))CALL cp__a("molden_utils.F",62)
         IF(.NOT.(SIZE(freq, 1) == SIZE(eigen_vec, 2)))CALL cp__a("molden_utils.F",63)
         WRITE (iw, '(T2,A)') "[Molden Format]"
         WRITE (iw, '(T2,A)') "[FREQ]"
         DO i = 1, SIZE(freq, 1)
            IF ((.NOT. dump_only_positive) .OR. (freq(i) >= 0._dp)) WRITE (iw, '(T5,F12.6)') freq(i)
         END DO
         WRITE (iw, '(T2,A)') "[FR-COORD]"
         DO i = 1, SIZE(particles)
            CALL get_atomic_kind(atomic_kind=particles(i)%atomic_kind, &
                                 element_symbol=element_symbol)
            WRITE (iw, '(T2,A2,3X,3(F12.6,3X))') &
               element_symbol, particles((i))%r(:)
         END DO
         WRITE (iw, '(T2,A)') "[FR-NORM-COORD]"
         l = 0
         DO i = 1, SIZE(eigen_vec, 2)
            IF ((.NOT. dump_only_positive) .OR. (freq(i) >= 0._dp)) THEN
               l = l+1
               WRITE (iw, '(T2,A,1X,I6)') "vibration", l
               DO j = 1, SIZE(eigen_vec, 1)/3
                  k = (j-1)*3
                  WRITE (iw, '(T2,3(F12.6,3X))') eigen_vec(k+1, i), eigen_vec(k+2, i), eigen_vec(k+3, i)
               END DO
            END IF
         END DO
         IF (calc_intens) THEN
            WRITE (iw, '(T2,A)') "[INT]"
            DO i = 1, SIZE(intensities)
               IF ((.NOT. dump_only_positive) .OR. (freq(i) >= 0._dp)) WRITE (iw, '(3X,F18.6)') intensities(i)
            END DO
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, input, "VIBRATIONAL_ANALYSIS%PRINT%MOLDEN_VIB")
   END SUBROUTINE molden_out

END MODULE molden_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods to setup replicas of the same system differing only by atom
!>      positions and velocities (as used in path integral or nudged elastic
!>      band for example)
!> \par History
!>      09.2005 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE replica_methods
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level
   USE cp_para_env,                     ONLY: cp_cart_create,&
                                              cp_para_env_create
   USE cp_para_types,                   ONLY: cp_para_cart_type,&
                                              cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_create,&
                                              cp_result_retain
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE f77_interface,                   ONLY: calc_force,&
                                              create_force_env,&
                                              f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type,&
                                              get_nparticle,&
                                              get_pos,&
                                              set_vel
   USE force_env_types,                 ONLY: force_env_get,&
                                              use_qs_force
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set,&
                                              section_vals_write
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE message_passing,                 ONLY: MPI_COMM_NULL,&
                                              mp_cart_create,&
                                              mp_cart_sub,&
                                              mp_sum
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_wf_history_methods,           ONLY: wfi_create,&
                                              wfi_create_for_kp
   USE qs_wf_history_types,             ONLY: wfi_retain
   USE replica_types,                   ONLY: rep_env_sync,&
                                              rep_env_sync_results,&
                                              rep_envs_add_rep_env,&
                                              rep_envs_get_rep_env,&
                                              replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'replica_methods'
   INTEGER, SAVE, PRIVATE :: last_rep_env_id = 0

   PUBLIC :: rep_env_create, rep_env_calc_e_f

CONTAINS

! **************************************************************************************************
!> \brief creates a replica environment together with its force environment
!> \param rep_env the replica environment that will be created
!> \param para_env the parallel enviroment that will contain the replicas
!> \param input the input used to initialize the force environment
!> \param input_declaration ...
!> \param nrep the number of replicas to calculate
!> \param prep the number of processors for each replica
!> \param sync_v if the volocity should be synchronized (defaults to false)
!> \param keep_wf_history if wf history should be kept on a per replica
!>        basis (defaults to true for QS jobs)
!> \param row_force to use the new mapping to the cart with rows
!>        working on force instead of columns.
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rep_env_create(rep_env, para_env, input, input_declaration, nrep, prep, &
                             sync_v, keep_wf_history, row_force)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration
      INTEGER                                            :: nrep, prep
      LOGICAL, INTENT(in), OPTIONAL                      :: sync_v, keep_wf_history, row_force

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_create', routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: input_file_path, output_file_path
      INTEGER :: comm_cart, comm_f, comm_inter_rep, forcedim, i, i0, ierr, ip, ir, irep, lp, &
         my_prep, new_env_id, nparticle, nrep_local, unit_nr
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: gridinfo
      INTEGER, DIMENSION(2)                              :: dims, pos
      LOGICAL, DIMENSION(2)                              :: rdim
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_cart_type), POINTER                   :: cart
      TYPE(cp_para_env_type), POINTER                    :: para_env_f, para_env_full, &
                                                            para_env_inter_rep

      IF(.NOT.(.NOT. ASSOCIATED(rep_env)))CALL cp__a("replica_methods.F",114)
      IF(.NOT.(ASSOCIATED(input_declaration)))CALL cp__a("replica_methods.F",115)

      NULLIFY (cart, para_env_f, para_env_inter_rep)
      logger => cp_get_default_logger()
      unit_nr = cp_logger_get_default_io_unit(logger)
      new_env_id = -1
      forcedim = 1
      IF (PRESENT(row_force)) THEN
         IF (row_force) forcedim = 2
      END IF
      my_prep = MIN(prep, para_env%num_pe)
      dims(3-forcedim) = MIN(para_env%num_pe/my_prep, nrep)
      dims(forcedim) = my_prep
      IF ((dims(1)*dims(2) /= para_env%num_pe) .AND. (unit_nr > 0)) THEN
         WRITE (unit_nr, FMT="(T2,A)") "REPLICA| WARNING: number of processors is not divisible by the number of replicas"
         WRITE (unit_nr, FMT="(T2,A,I0,A)") "REPLICA| ", para_env%num_pe-dims(1)*dims(2), " MPI process(es) will be idle"
      END IF
      CALL mp_cart_create(comm_old=para_env%group, ndims=2, dims=dims, pos=pos, comm_cart=comm_cart)
      IF (comm_cart /= MPI_COMM_NULL) THEN
         CALL cp_cart_create(cart, comm_cart, ndims=2, owns_group=.TRUE.)
         NULLIFY (para_env_full)
         CALL cp_para_env_create(para_env_full, comm_cart, owns_group=.FALSE.)
         rdim(3-forcedim) = .FALSE.
         rdim(forcedim) = .TRUE.
         CALL mp_cart_sub(comm=comm_cart, rdim=rdim, sub_comm=comm_f)
         CALL cp_para_env_create(para_env_f, comm_f, owns_group=.TRUE.)
         rdim(3-forcedim) = .TRUE.
         rdim(forcedim) = .FALSE.
         CALL mp_cart_sub(comm=comm_cart, rdim=rdim, sub_comm=comm_inter_rep)
         CALL cp_para_env_create(para_env_inter_rep, comm_inter_rep, &
                                 owns_group=.TRUE.)
         ALLOCATE (rep_env)
      END IF
      ALLOCATE (gridinfo(2, 0:para_env%num_pe-1))
      gridinfo = 0
      gridinfo(:, para_env%mepos) = pos
      CALL mp_sum(gridinfo, para_env%group)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, FMT="(T2,A,T71,I10)") "REPLICA| layout of the replica grid, number of groups ", para_env_inter_rep%num_pe
         WRITE (unit_nr, FMT="(T2,A,T71,I10)") "REPLICA| layout of the replica grid, size of each group", para_env_f%num_pe
         WRITE (unit_nr, FMT="(T2,A)", ADVANCE="NO") "REPLICA| MPI process to grid (group,rank) correspondence:"
         DO i = 0, para_env%num_pe-1
            IF (MODULO(i, 4) == 0) WRITE (unit_nr, *)
            WRITE (unit_nr, FMT='(A3,I4,A3,I4,A1,I4,A1)', ADVANCE="NO") &
               "  (", i, " : ", gridinfo(3-forcedim, i), ",", &
               gridinfo(forcedim, i), ")"
         END DO
         WRITE (unit_nr, *)
      ENDIF
      DEALLOCATE (gridinfo)
      IF (ASSOCIATED(rep_env)) THEN
         last_rep_env_id = last_rep_env_id+1
         rep_env%id_nr = last_rep_env_id
         rep_env%ref_count = 1
         rep_env%nrep = nrep
         rep_env%sync_v = .FALSE.
         IF (PRESENT(sync_v)) rep_env%sync_v = sync_v
         rep_env%keep_wf_history = .TRUE.
         IF (PRESENT(keep_wf_history)) rep_env%keep_wf_history = keep_wf_history
         NULLIFY (rep_env%wf_history)
         NULLIFY (rep_env%results)

         rep_env%force_dim = forcedim
         rep_env%my_rep_group = cart%mepos(3-forcedim)
         ALLOCATE (rep_env%inter_rep_rank(0:para_env_inter_rep%num_pe-1), &
                   rep_env%force_rank(0:para_env_f%num_pe-1))
         rep_env%inter_rep_rank = 0
         rep_env%inter_rep_rank(rep_env%my_rep_group) = para_env_inter_rep%mepos
         CALL mp_sum(rep_env%inter_rep_rank, para_env_inter_rep%group)
         rep_env%force_rank = 0
         rep_env%force_rank(cart%mepos(forcedim)) = para_env_f%mepos
         CALL mp_sum(rep_env%force_rank, para_env_f%group)

         CALL section_vals_val_get(input, "GLOBAL%PROJECT_NAME", &
                                   c_val=input_file_path)
         rep_env%original_project_name = input_file_path
         ! By default replica_env handles files for each replica
         ! with the structure PROJECT_NAME-r-N where N is the
         ! number of the local replica..
         lp = LEN_TRIM(input_file_path)
         input_file_path(lp+1:LEN(input_file_path)) = "-r-"// &
                                                      ADJUSTL(cp_to_string(rep_env%my_rep_group))
         lp = LEN_TRIM(input_file_path)
         ! Setup new project name
         CALL section_vals_val_set(input, "GLOBAL%PROJECT_NAME", &
                                   c_val=input_file_path)
         ! Redirect the output of each replica on a same local file
         output_file_path = input_file_path(1:lp)//".out"
         CALL section_vals_val_set(input, "GLOBAL%OUTPUT_FILE_NAME", &
                                   c_val=TRIM(output_file_path))

         ! Dump an input file to warm-up new force_eval structures and
         ! delete them immediately afterwards..
         input_file_path(lp+1:LEN(input_file_path)) = ".inp"
         IF (para_env_f%source == para_env_f%mepos) THEN
            CALL open_file(file_name=TRIM(input_file_path), file_status="UNKNOWN", &
                           file_form="FORMATTED", file_action="WRITE", &
                           unit_number=unit_nr)
            CALL section_vals_write(input, unit_nr, hide_root=.TRUE.)
            CALL close_file(unit_nr)
         END IF
         CALL create_force_env(new_env_id, input_declaration, input_file_path, &
                               output_file_path, para_env_f%group, ierr=ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",218)

         ! Delete input files..
         IF (para_env_f%source == para_env_f%mepos) THEN
            CALL open_file(file_name=TRIM(input_file_path), file_status="OLD", &
                           file_form="FORMATTED", file_action="READ", unit_number=unit_nr)
            CALL close_file(unit_number=unit_nr, file_status="DELETE")
         END IF

         rep_env%f_env_id = new_env_id
         CALL get_nparticle(new_env_id, nparticle, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",229)
         rep_env%nparticle = nparticle
         rep_env%ndim = 3*nparticle
         ALLOCATE (rep_env%replica_owner(nrep))

         i0 = nrep/para_env_inter_rep%num_pe
         ir = MODULO(nrep, para_env_inter_rep%num_pe)
         DO ip = 0, para_env_inter_rep%num_pe-1
            DO i = i0*ip+MIN(ip, ir)+1, i0*(ip+1)+MIN(ip+1, ir)
               rep_env%replica_owner(i) = ip
            END DO
         END DO

         nrep_local = i0
         IF (rep_env%my_rep_group < ir) nrep_local = nrep_local+1
         ALLOCATE (rep_env%local_rep_indices(nrep_local), &
                   rep_env%rep_is_local(nrep))
         nrep_local = 0
         rep_env%rep_is_local = .FALSE.
         DO irep = 1, nrep
            IF (rep_env%replica_owner(irep) == rep_env%my_rep_group) THEN
               nrep_local = nrep_local+1
               rep_env%local_rep_indices(nrep_local) = irep
               rep_env%rep_is_local(irep) = .TRUE.
            END IF
         END DO
         IF(.NOT.(nrep_local == SIZE(rep_env%local_rep_indices)))CALL cp__a("replica_methods.F",255)

         rep_env%cart => cart
         rep_env%para_env => para_env_full
         rep_env%para_env_f => para_env_f
         rep_env%para_env_inter_rep => para_env_inter_rep

         ALLOCATE (rep_env%r(rep_env%ndim, nrep), rep_env%v(rep_env%ndim, nrep), &
                   rep_env%f(rep_env%ndim+1, nrep))

         rep_env%r = 0._dp
         rep_env%f = 0._dp
         rep_env%v = 0._dp
         CALL set_vel(rep_env%f_env_id, rep_env%v(:, 1), rep_env%ndim, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",269)
         DO i = 1, nrep
            IF (rep_env%rep_is_local(i)) THEN
               CALL get_pos(rep_env%f_env_id, rep_env%r(:, i), rep_env%ndim, ierr)
               IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",273)
            END IF
         END DO
      END IF
      IF (ASSOCIATED(rep_env)) THEN
         CALL rep_envs_add_rep_env(rep_env)
         CALL rep_env_init_low(rep_env%id_nr, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",280)
      END IF
   END SUBROUTINE rep_env_create

! **************************************************************************************************
!> \brief finishes the low level initialization of the replica env
!> \param rep_env_id id_nr of the replica environment that should be initialized
!> \param ierr will be non zero if there is an initialization error
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rep_env_init_low(rep_env_id, ierr)
      INTEGER, INTENT(in)                                :: rep_env_id
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_init_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, in_use, stat
      LOGICAL                                            :: do_kpoints, has_unit_metric
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(replica_env_type), POINTER                    :: rep_env

      rep_env => rep_envs_get_rep_env(rep_env_id, ierr=stat)
      IF (.NOT. ASSOCIATED(rep_env)) &
         CALL cp__b("replica_methods.F",308,"could not find rep_env with id_nr"//cp_to_string(rep_env_id))
      NULLIFY (qs_env, dft_control, subsys)
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
      logger => cp_get_default_logger()
      logger%iter_info%iteration(1) = rep_env%my_rep_group
      CALL cp_add_iter_level(iteration_info=logger%iter_info, &
                             level_name="REPLICA_EVAL")
      !wf interp
      IF (rep_env%keep_wf_history) THEN
         CALL force_env_get(f_env%force_env, in_use=in_use)
         IF (in_use == use_qs_force) THEN
            CALL force_env_get(f_env%force_env, qs_env=qs_env)
            CALL get_qs_env(qs_env, dft_control=dft_control)
            ALLOCATE (rep_env%wf_history(SIZE(rep_env%local_rep_indices)))
            DO i = 1, SIZE(rep_env%wf_history)
               NULLIFY (rep_env%wf_history(i)%wf_history)
               IF (i == 1) THEN
                  CALL get_qs_env(qs_env, &
                                  wf_history=rep_env%wf_history(i)%wf_history)
                  CALL wfi_retain(rep_env%wf_history(i)%wf_history)
               ELSE
                  CALL get_qs_env(qs_env, has_unit_metric=has_unit_metric, &
                                  do_kpoints=do_kpoints)
                  CALL wfi_create(rep_env%wf_history(i)%wf_history, &
                                  interpolation_method_nr= &
                                  dft_control%qs_control%wf_interpolation_method_nr, &
                                  extrapolation_order=dft_control%qs_control%wf_extrapolation_order, &
                                  has_unit_metric=has_unit_metric)
                  IF (do_kpoints) THEN
                     CALL wfi_create_for_kp(rep_env%wf_history(i)%wf_history)
                  END IF
               END IF
            END DO
         ELSE
            rep_env%keep_wf_history = .FALSE.
         END IF
      END IF
      ALLOCATE (rep_env%results(rep_env%nrep))
      DO i = 1, rep_env%nrep
         NULLIFY (rep_env%results(i)%results)
         IF (i == 1) THEN
            CALL force_env_get(f_env%force_env, subsys=subsys)
            CALL cp_subsys_get(subsys, results=rep_env%results(i)%results)
            CALL cp_result_retain(rep_env%results(i)%results)
         ELSE
            CALL cp_result_create(rep_env%results(i)%results)
         END IF
      END DO
      CALL rep_env_sync(rep_env, rep_env%r)
      CALL rep_env_sync(rep_env, rep_env%v)
      CALL rep_env_sync(rep_env, rep_env%f)

      CALL f_env_rm_defaults(f_env, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",361)
   END SUBROUTINE rep_env_init_low

! **************************************************************************************************
!> \brief evaluates the forces
!> \param rep_env the replica environment on which you want to evaluate the
!>        forces
!> \param calc_f if true calculates also the forces, if false only the
!>        energy
!> \author fawzi
!> \note
!>      indirect through f77_int_low to work around fortran madness
! **************************************************************************************************
   SUBROUTINE rep_env_calc_e_f(rep_env, calc_f)
      TYPE(replica_env_type), POINTER                    :: rep_env
      LOGICAL, OPTIONAL                                  :: calc_f

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_calc_e_f', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, my_calc_f

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(rep_env)))CALL cp__a("replica_methods.F",384)
      IF(.NOT.(rep_env%ref_count > 0))CALL cp__a("replica_methods.F",385)
      my_calc_f = 0
      IF (PRESENT(calc_f)) THEN
         IF (calc_f) my_calc_f = 1
      END IF
      CALL rep_env_calc_e_f_low(rep_env%id_nr, my_calc_f, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",391)
      CALL timestop(handle)
   END SUBROUTINE rep_env_calc_e_f

! **************************************************************************************************
!> \brief calculates energy and force, internal private method
!> \param rep_env_id the id if the replica environment in which energy and
!>        forces have to be evaluated
!> \param calc_f if nonzero calculates also the forces along with the
!>        energy
!> \param ierr if an error happens this will be nonzero
!> \author fawzi
!> \note
!>      low level wrapper to export this function in f77_int_low and work
!>      around the handling of circular dependecies in fortran
! **************************************************************************************************
   RECURSIVE SUBROUTINE rep_env_calc_e_f_low(rep_env_id, calc_f, ierr)
      INTEGER, INTENT(in)                                :: rep_env_id, calc_f
      INTEGER, INTENT(out)                               :: ierr

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_calc_e_f_low', &
         routineP = moduleN//':'//routineN

      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(replica_env_type), POINTER                    :: rep_env

      rep_env => rep_envs_get_rep_env(rep_env_id, ierr)
      IF (ASSOCIATED(rep_env)) THEN
         CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
         CALL rep_env_calc_e_f_int(rep_env, calc_f /= 0)
         CALL f_env_rm_defaults(f_env, ierr)
      ELSE
         ierr = 111
      END IF
   END SUBROUTINE rep_env_calc_e_f_low

! **************************************************************************************************
!> \brief calculates energy and force, internal private method
!> \param rep_env the replica env to update
!> \param calc_f if the force should be calculated as well (defaults to true)
!> \author fawzi
!> \note
!>      this is the where the real work is done
! **************************************************************************************************
   SUBROUTINE rep_env_calc_e_f_int(rep_env, calc_f)
      TYPE(replica_env_type), POINTER                    :: rep_env
      LOGICAL, OPTIONAL                                  :: calc_f

      CHARACTER(len=*), PARAMETER :: routineN = 'rep_env_calc_e_f_int', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ierr, irep, md_iter, my_calc_f, ndim
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      NULLIFY (f_env, qs_env, subsys)
      IF(.NOT.(ASSOCIATED(rep_env)))CALL cp__a("replica_methods.F",449)
      IF(.NOT.(rep_env%ref_count > 0))CALL cp__a("replica_methods.F",450)
      my_calc_f = 3*rep_env%nparticle
      IF (PRESENT(calc_f)) THEN
         IF (.NOT. calc_f) my_calc_f = 0
      END IF

      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
      logger => cp_get_default_logger()
      !     md_iter=logger%iter_info%iteration(2)+1
      md_iter = logger%iter_info%iteration(2)
      CALL f_env_rm_defaults(f_env, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",461)
      DO i = 1, SIZE(rep_env%local_rep_indices)
         irep = rep_env%local_rep_indices(i)
         ndim = 3*rep_env%nparticle
         IF (rep_env%sync_v) THEN
            CALL set_vel(rep_env%f_env_id, rep_env%v(:, irep), ndim, ierr)
            IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",467)
         END IF

         logger%iter_info%iteration(1) = irep
         logger%iter_info%iteration(2) = md_iter

         IF (rep_env%keep_wf_history) THEN
            CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
            CALL force_env_get(f_env%force_env, qs_env=qs_env)
            CALL set_qs_env(qs_env, &
                            wf_history=rep_env%wf_history(i)%wf_history)
            CALL f_env_rm_defaults(f_env, ierr)
            IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",479)
         END IF

         CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
         CALL force_env_get(f_env%force_env, subsys=subsys)
         CALL cp_subsys_set(subsys, results=rep_env%results(irep)%results)
         CALL f_env_rm_defaults(f_env, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",486)
         CALL calc_force(rep_env%f_env_id, rep_env%r(:, irep), ndim, &
                         rep_env%f(ndim+1, irep), rep_env%f(:ndim, irep), &
                         my_calc_f, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("replica_methods.F",490)
      END DO
      CALL rep_env_sync(rep_env, rep_env%f)
      CALL rep_env_sync_results(rep_env, rep_env%results)

   END SUBROUTINE rep_env_calc_e_f_int

END MODULE replica_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module performing a mdoe selective vibrational analysis
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 08.2006)
!> \author Florian Schiffmann 08.2006
! **************************************************************************************************
MODULE mode_selective
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: get_results
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: ms_guess_atomic,&
                                              ms_guess_bfgs,&
                                              ms_guess_molden,&
                                              ms_guess_restart,&
                                              ms_guess_restart_vec
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              max_line_length
   USE mathlib,                         ONLY: diamat_all
   USE message_passing,                 ONLY: mp_bcast
   USE molden_utils,                    ONLY: molden_out
   USE parallel_rng_types,              ONLY: next_random_number
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: massunit,&
                                              vibfac
   USE replica_methods,                 ONLY: rep_env_calc_e_f
   USE replica_types,                   ONLY: replica_env_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mode_selective'
   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.

   TYPE ms_vib_type
      INTEGER                                  :: mat_size
      INTEGER                                  :: select_id
      INTEGER, DIMENSION(:), POINTER           :: inv_atoms
      REAL(KIND=dp)                            :: eps(2)
      REAL(KIND=dp)                            :: sel_freq
      REAL(KIND=dp), POINTER, DIMENSION(:, :)    :: b_vec
      REAL(KIND=dp), POINTER, DIMENSION(:, :)    :: delta_vec
      REAL(KIND=dp), POINTER, DIMENSION(:, :)    :: ms_force
      REAL(KIND=dp), DIMENSION(:), POINTER     :: eig_bfgs
      REAL(KIND=dp), DIMENSION(:), POINTER     :: f_range
      REAL(KIND=dp), DIMENSION(:), POINTER     :: inv_range
      REAL(KIND=dp), POINTER, DIMENSION(:)     :: step_b
      REAL(KIND=dp), POINTER, DIMENSION(:)     :: step_r
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: b_mat
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: dip_deriv
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: hes_bfgs
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: s_mat
      INTEGER                                  :: initial_guess
   END TYPE ms_vib_type

   PUBLIC :: ms_vb_anal

CONTAINS
! **************************************************************************************************
!> \brief Module performing a vibrational analysis
!> \param input ...
!> \param rep_env ...
!> \param para_env ...
!> \param globenv ...
!> \param particles ...
!> \param nrep ...
!> \param calc_intens ...
!> \param dx ...
!> \param output_unit ...
!> \param logger ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE ms_vb_anal(input, rep_env, para_env, globenv, particles, &
                         nrep, calc_intens, dx, output_unit, logger)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      INTEGER                                            :: nrep
      LOGICAL                                            :: calc_intens
      REAL(KIND=dp)                                      :: dx
      INTEGER                                            :: output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'ms_vb_anal', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      INTEGER                                            :: handle, i, ip1, j, natoms, ncoord
      LOGICAL                                            :: converged
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mass, pos0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: tmp_deriv
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: tmp_dip
      TYPE(ms_vib_type)                                  :: ms_vib

      CALL timeset(routineN, handle)
      converged = .FALSE.
      natoms = SIZE(particles)
      ncoord = 3*natoms
      ALLOCATE (mass(3*natoms))
      DO i = 1, natoms
         DO j = 1, 3
            mass((i-1)*3+j) = particles(i)%atomic_kind%mass
            mass((i-1)*3+j) = SQRT(mass((i-1)*3+j))
         END DO
      END DO
      ! Allocate working arrays
      ALLOCATE (ms_vib%delta_vec(ncoord, nrep))
      ALLOCATE (ms_vib%b_vec(ncoord, nrep))
      ALLOCATE (ms_vib%step_r(nrep))
      ALLOCATE (ms_vib%step_b(nrep))
      IF (calc_intens) THEN
         description = '[DIPOLE]'
         ALLOCATE (tmp_dip(nrep, 3, 2))
         ALLOCATE (ms_vib%dip_deriv(3, nrep))
      END IF
      CALL MS_initial_moves(para_env, nrep, input, globenv, ms_vib, &
                            particles, &
                            mass, &
                            dx, &
                            calc_intens, logger)
      ncoord = 3*natoms
      ALLOCATE (pos0(ncoord))
      ALLOCATE (ms_vib%ms_force(ncoord, nrep))
      DO i = 1, natoms
         DO j = 1, 3
            pos0((i-1)*3+j) = particles((i))%r(j)
         END DO
      END DO
      ncoord = 3*natoms
      DO
         ms_vib%ms_force = HUGE(0.0_dp)
         DO i = 1, nrep
            DO j = 1, ncoord
               rep_env%r(j, i) = pos0(j)+ms_vib%step_r(i)*ms_vib%delta_vec(j, i)
            END DO
         END DO
         CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)

         DO i = 1, nrep
            IF (calc_intens) THEN
               CALL get_results(results=rep_env%results(i)%results, &
                                description=description, &
                                n_rep=ip1)
               CALL get_results(results=rep_env%results(i)%results, &
                                description=description, &
                                values=tmp_dip(i, :, 1), &
                                nval=ip1)
            END IF
            DO j = 1, ncoord
               ms_vib%ms_force(j, i) = rep_env%f(j, i)
            END DO
         END DO
         DO i = 1, nrep
            DO j = 1, ncoord
               rep_env%r(j, i) = pos0(j)-ms_vib%step_r(i)*ms_vib%delta_vec(j, i)
            END DO
         END DO
         CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)
         IF (calc_intens) THEN
            DO i = 1, nrep
               CALL get_results(results=rep_env%results(i)%results, &
                                description=description, &
                                n_rep=ip1)
               CALL get_results(results=rep_env%results(i)%results, &
                                description=description, &
                                values=tmp_dip(i, :, 2), &
                                nval=ip1)
               ms_vib%dip_deriv(:, ms_vib%mat_size+i) = (tmp_dip(i, :, 1)-tmp_dip(i, :, 2))/(2*ms_vib%step_b(i))
            END DO
         END IF

         CALL evaluate_H_update_b(rep_env, ms_vib, input, nrep, &
                                  particles, &
                                  mass, &
                                  converged, &
                                  dx, calc_intens, &
                                  output_unit, logger)
         IF (converged) EXIT
         IF (calc_intens) THEN
            ALLOCATE (tmp_deriv(3, ms_vib%mat_size))
            tmp_deriv = ms_vib%dip_deriv
            DEALLOCATE (ms_vib%dip_deriv)
            ALLOCATE (ms_vib%dip_deriv(3, ms_vib%mat_size+nrep))
            ms_vib%dip_deriv(:, 1:ms_vib%mat_size) = tmp_deriv(:, 1:ms_vib%mat_size)
            DEALLOCATE (tmp_deriv)
         END IF
      END DO
      DEALLOCATE (ms_vib%ms_force)
      DEALLOCATE (pos0)
      DEALLOCATE (ms_vib%step_r)
      DEALLOCATE (ms_vib%step_b)
      DEALLOCATE (ms_vib%b_vec)
      DEALLOCATE (ms_vib%delta_vec)
      DEALLOCATE (mass)
      DEALLOCATE (ms_vib%b_mat)
      DEALLOCATE (ms_vib%s_mat)
      IF (ms_vib%select_id == 3) THEN
         DEALLOCATE (ms_vib%inv_atoms)
      END IF
      IF (ASSOCIATED(ms_vib%eig_bfgs)) THEN
         DEALLOCATE (ms_vib%eig_bfgs)
      END IF
      IF (ASSOCIATED(ms_vib%hes_bfgs)) THEN
         DEALLOCATE (ms_vib%hes_bfgs)
      END IF
      IF (calc_intens) THEN
         DEALLOCATE (ms_vib%dip_deriv)
         DEALLOCATE (tmp_dip)
      END IF
      CALL timestop(handle)
   END SUBROUTINE ms_vb_anal
! **************************************************************************************************
!> \brief Generates the first displacement vector for a mode selctive vibrational
!>      analysis. At the moment this is a random number for selected atoms
!> \param para_env ...
!> \param nrep ...
!> \param input ...
!> \param globenv ...
!> \param ms_vib ...
!> \param particles ...
!> \param mass ...
!> \param dx ...
!> \param calc_intens ...
!> \param logger ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE MS_initial_moves(para_env, nrep, input, globenv, ms_vib, particles, &
                               mass, dx, &
                               calc_intens, logger)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: nrep
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(ms_vib_type)                                  :: ms_vib
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      REAL(KIND=dp)                                      :: dx
      LOGICAL                                            :: calc_intens
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'MS_initial_moves', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: guess, handle, i, j, jj, k, m, &
                                                            n_rep_val, natoms, ncoord
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map_atoms
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      LOGICAL                                            :: do_involved_atoms, ionode
      REAL(KIND=dp)                                      :: my_val, norm
      TYPE(section_vals_type), POINTER                   :: involved_at_section, ms_vib_section

      CALL timeset(routineN, handle)
      NULLIFY (ms_vib%eig_bfgs, ms_vib%f_range, ms_vib%hes_bfgs, ms_vib%inv_range)
      ms_vib_section => section_vals_get_subs_vals(input, "VIBRATIONAL_ANALYSIS%MODE_SELECTIVE")
      CALL section_vals_val_get(ms_vib_section, "INITIAL_GUESS", i_val=guess)
      CALL section_vals_val_get(ms_vib_section, "EPS_MAX_VAL", r_val=ms_vib%eps(1))
      CALL section_vals_val_get(ms_vib_section, "EPS_NORM", r_val=ms_vib%eps(2))
      CALL section_vals_val_get(ms_vib_section, "RANGE", n_rep_val=n_rep_val)
      ms_vib%select_id = 0
      IF (n_rep_val .NE. 0) THEN
         CALL section_vals_val_get(ms_vib_section, "RANGE", r_vals=ms_vib%f_range)
         IF (ms_vib%f_range(1) .GT. ms_vib%f_range(2)) THEN
            my_val = ms_vib%f_range(2)
            ms_vib%f_range(2) = ms_vib%f_range(1)
            ms_vib%f_range(1) = my_val
         END IF
         ms_vib%select_id = 2
      END IF
      CALL section_vals_val_get(ms_vib_section, "FREQUENCY", r_val=ms_vib%sel_freq)
      IF (ms_vib%sel_freq .GT. 0._dp) ms_vib%select_id = 1
      involved_at_section => section_vals_get_subs_vals(ms_vib_section, "INVOLVED_ATOMS")
      CALL section_vals_get(involved_at_section, explicit=do_involved_atoms)
      IF (do_involved_atoms) THEN
         CALL section_vals_val_get(involved_at_section, "INVOLVED_ATOMS", n_rep_val=n_rep_val)
         jj = 0
         DO k = 1, n_rep_val
            CALL section_vals_val_get(involved_at_section, "INVOLVED_ATOMS", i_rep_val=k, i_vals=tmplist)
            DO j = 1, SIZE(tmplist)
               jj = jj+1
            END DO
         END DO
         IF (jj .GE. 1) THEN
            natoms = jj
            ALLOCATE (ms_vib%inv_atoms(natoms))
            jj = 0
            DO m = 1, n_rep_val
               CALL section_vals_val_get(involved_at_section, "INVOLVED_ATOMS", i_rep_val=m, i_vals=tmplist)
               DO j = 1, SIZE(tmplist)
                  ms_vib%inv_atoms(j) = tmplist(j)
               END DO
            END DO
            ms_vib%select_id = 3
         END IF
         CALL section_vals_val_get(involved_at_section, "RANGE", n_rep_val=n_rep_val)
         IF (n_rep_val .NE. 0) THEN
            CALL section_vals_val_get(involved_at_section, "RANGE", r_vals=ms_vib%inv_range)
            IF (ms_vib%inv_range(1) .GT. ms_vib%inv_range(2)) THEN
               ms_vib%inv_range(2) = my_val
               ms_vib%inv_range(2) = ms_vib%inv_range(1)
               ms_vib%inv_range(1) = my_val
            END IF
         END IF
      END IF
      IF (ms_vib%select_id == 0) &
         CALL cp__b("mode_selective.F",331,"no frequency, range or involved atoms specified ")
      ionode = para_env%mepos == para_env%source
      SELECT CASE (guess)
      CASE (ms_guess_atomic)
         ms_vib%initial_guess = 1
         CALL section_vals_val_get(ms_vib_section, "ATOMS", n_rep_val=n_rep_val)
         jj = 0
         DO k = 1, n_rep_val
            CALL section_vals_val_get(ms_vib_section, "ATOMS", i_rep_val=k, i_vals=tmplist)
            DO j = 1, SIZE(tmplist)
               jj = jj+1
            END DO
         END DO
         IF (jj < 1) THEN
            natoms = SIZE(particles)
            ALLOCATE (map_atoms(natoms))
            DO j = 1, natoms
               map_atoms(j) = j
            END DO
         ELSE
            natoms = jj
            ALLOCATE (map_atoms(natoms))
            jj = 0
            DO m = 1, n_rep_val
               CALL section_vals_val_get(ms_vib_section, "ATOMS", i_rep_val=m, i_vals=tmplist)
               DO j = 1, SIZE(tmplist)
                  map_atoms(j) = tmplist(j)
               END DO
            END DO
         END IF

         ! apply random displacement along the mass weighted nuclear cartesian coordinates
         ms_vib%b_vec = 0._dp
         ms_vib%delta_vec = 0._dp
         jj = 0

         DO i = 1, nrep
            DO j = 1, natoms
               DO k = 1, 3
                  jj = (map_atoms(j)-1)*3+k
                  ms_vib%b_vec(jj, i) = ABS(next_random_number(globenv%gaussian_rng_stream))
               END DO
            END DO
            norm = SQRT(DOT_PRODUCT(ms_vib%b_vec(:, i), ms_vib%b_vec(:, i)))
            ms_vib%b_vec(:, i) = ms_vib%b_vec(:, i)/norm
         END DO

         IF (nrep .GT. 1) THEN
            DO k = 1, 10
               DO j = 1, nrep
                  DO i = 1, nrep
                     IF (i .NE. j) THEN
                        ms_vib%b_vec(:, j) = &
                           ms_vib%b_vec(:, j)-DOT_PRODUCT(ms_vib%b_vec(:, j), ms_vib%b_vec(:, i))*ms_vib%b_vec(:, i)
                        ms_vib%b_vec(:, j) = &
                           ms_vib%b_vec(:, j)/SQRT(DOT_PRODUCT(ms_vib%b_vec(:, j), ms_vib%b_vec(:, j)))
                     END IF
                  END DO
               END DO
            END DO
         END IF

         ms_vib%mat_size = 0
         DO i = 1, SIZE(ms_vib%b_vec, 1)
            ms_vib%delta_vec(i, :) = ms_vib%b_vec(i, :)/mass(i)
         END DO
      CASE (ms_guess_bfgs)

         ms_vib%initial_guess = 2
         CALL bfgs_guess(ms_vib_section, ms_vib, particles, mass, para_env, nrep)
         ms_vib%mat_size = 0

      CASE (ms_guess_restart_vec)

         ms_vib%initial_guess = 3
         ncoord = 3*SIZE(particles)
         CALL rest_guess(ms_vib_section, para_env, ms_vib, mass, ionode, particles, nrep, calc_intens)

         ms_vib%mat_size = 0
      CASE (ms_guess_restart)
         ms_vib%initial_guess = 4
         ncoord = 3*SIZE(particles)
         CALL rest_guess(ms_vib_section, para_env, ms_vib, mass, ionode, particles, nrep, calc_intens)

      CASE (ms_guess_molden)
         ms_vib%initial_guess = 5
         ncoord = 3*SIZE(particles)
         CALL molden_guess(ms_vib_section, input, para_env, ms_vib, mass, ncoord, nrep, logger)
         ms_vib%mat_size = 0
      END SELECT
      CALL mp_bcast(ms_vib%b_vec, para_env%source, para_env%group)
      CALL mp_bcast(ms_vib%delta_vec, para_env%source, para_env%group)
      DO i = 1, nrep
         ms_vib%step_r(i) = dx/SQRT(DOT_PRODUCT(ms_vib%delta_vec(:, i), ms_vib%delta_vec(:, i)))
         ms_vib%step_b(i) = SQRT(DOT_PRODUCT(ms_vib%step_r(i)*ms_vib%b_vec(:, i), ms_vib%step_r(i)*ms_vib%b_vec(:, i)))
      END DO
      CALL timestop(handle)

   END SUBROUTINE MS_initial_moves

! **************************************************************************************************
!> \brief ...
!> \param ms_vib_section ...
!> \param ms_vib ...
!> \param particles ...
!> \param mass ...
!> \param para_env ...
!> \param nrep ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE bfgs_guess(ms_vib_section, ms_vib, particles, mass, para_env, nrep)

      TYPE(section_vals_type), POINTER                   :: ms_vib_section
      TYPE(ms_vib_type)                                  :: ms_vib
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: nrep

      CHARACTER(len=*), PARAMETER :: routineN = 'bfgs_guess', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: hes_filename
      INTEGER                                            :: hesunit, i, j, jj, k, natoms, ncoord, &
                                                            output_unit, stat
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      REAL(KIND=dp)                                      :: my_val, norm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      natoms = SIZE(particles)
      ncoord = 3*natoms

      ALLOCATE (ms_vib%hes_bfgs(ncoord, ncoord))
      ALLOCATE (ms_vib%eig_bfgs(ncoord))

      IF (para_env%mepos == para_env%source) THEN
         CALL section_vals_val_get(ms_vib_section, "RESTART_FILE_NAME", c_val=hes_filename)
         IF (hes_filename == "") hes_filename = "HESSIAN"
         CALL open_file(file_name=hes_filename, file_status="OLD", &
                        file_form="UNFORMATTED", file_action="READ", unit_number=hesunit)
         ALLOCATE (tmp(ncoord))
         ALLOCATE (tmplist(ncoord))

         ! should use the cp_fm_read_unformatted...
         DO i = 1, ncoord
            READ (UNIT=hesunit, IOSTAT=stat) ms_vib%hes_bfgs(:, i)
         ENDDO
         CALL close_file(hesunit)
         IF (output_unit > 0) THEN
            IF (stat /= 0) THEN
               WRITE (output_unit, FMT="(/,T2,A)") "**  Error while reading HESSIAN **"
            ELSE
               WRITE (output_unit, FMT="(/,T2,A)") &
                  "*** Initial Hessian has been read successfully ***"
            END IF
         END IF
         DO i = 1, ncoord
            DO j = 1, ncoord
               ms_vib%hes_bfgs(i, j) = ms_vib%hes_bfgs(i, j)/(mass(i)*mass(j))
            END DO
         END DO

         CALL diamat_all(ms_vib%hes_bfgs, ms_vib%eig_bfgs)
         tmp(:) = 0._dp
         IF (ms_vib%select_id == 1) my_val = (ms_vib%sel_freq/vibfac)**2/massunit
         IF (ms_vib%select_id == 2) my_val = (((ms_vib%f_range(2)+ms_vib%f_range(1))*0.5_dp)/vibfac)**2/massunit
         IF (ms_vib%select_id == 1 .OR. ms_vib%select_id == 2) THEN
            DO i = 1, ncoord
               tmp(i) = ABS(my_val-ms_vib%eig_bfgs(i))
            END DO
         ELSE IF (ms_vib%select_id == 3) THEN
            DO i = 1, ncoord
               DO j = 1, SIZE(ms_vib%inv_atoms)
                  DO k = 1, 3
                     jj = (ms_vib%inv_atoms(j)-1)*3+k
                     tmp(i) = tmp(i)+SQRT(ms_vib%hes_bfgs(jj, i)**2)
                  END DO
               END DO
               IF ((SIGN(1._dp, ms_vib%eig_bfgs(i))*SQRT(ABS(ms_vib%eig_bfgs(i))*massunit)*vibfac) .LE. 400._dp) tmp(i) = 0._dp
            END DO
            tmp(:) = -tmp(:)
         END IF
         CALL sort(tmp, ncoord, tmplist)
         DO i = 1, nrep
            ms_vib%b_vec(:, i) = ms_vib%hes_bfgs(:, tmplist(i))
            norm = SQRT(DOT_PRODUCT(ms_vib%b_vec(:, i), ms_vib%b_vec(:, i)))
            ms_vib%b_vec(:, i) = ms_vib%b_vec(:, i)/norm
         END DO
         DO i = 1, SIZE(ms_vib%b_vec, 1)
            ms_vib%delta_vec(i, :) = ms_vib%b_vec(i, :)/mass(i)
         END DO
         DEALLOCATE (tmp)
         DEALLOCATE (tmplist)
      END IF

      CALL mp_bcast(ms_vib%b_vec, para_env%source, para_env%group)
      CALL mp_bcast(ms_vib%delta_vec, para_env%source, para_env%group)

      DEALLOCATE (ms_vib%hes_bfgs)
      DEALLOCATE (ms_vib%eig_bfgs)
      ms_vib%mat_size = 0

   END SUBROUTINE bfgs_guess

! **************************************************************************************************
!> \brief ...
!> \param ms_vib_section ...
!> \param para_env ...
!> \param ms_vib ...
!> \param mass ...
!> \param ionode ...
!> \param particles ...
!> \param nrep ...
!> \param calc_intens ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE rest_guess(ms_vib_section, para_env, ms_vib, mass, ionode, particles, nrep, calc_intens)

      TYPE(section_vals_type), POINTER                   :: ms_vib_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(ms_vib_type)                                  :: ms_vib
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      LOGICAL                                            :: ionode
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      INTEGER                                            :: nrep
      LOGICAL                                            :: calc_intens

      CHARACTER(len=*), PARAMETER :: routineN = 'rest_guess', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: ms_filename
      INTEGER                                            :: hesunit, i, j, mat, natoms, ncoord, &
                                                            output_unit, stat, statint
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ind
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenval
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: approx_H
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      natoms = SIZE(particles)
      ncoord = 3*natoms
      IF (calc_intens) THEN
         DEALLOCATE (ms_vib%dip_deriv)
      END IF

      IF (ionode) THEN

         CALL section_vals_val_get(ms_vib_section, "RESTART_FILE_NAME", c_val=ms_filename)
         IF (ms_filename == "") ms_filename = "MS_RESTART"
         CALL open_file(file_name=ms_filename, &
                        file_status="UNKNOWN", &
                        file_form="UNFORMATTED", &
                        file_action="READ", &
                        unit_number=hesunit)
         READ (UNIT=hesunit, IOSTAT=stat) mat
         IF(.NOT.(stat == 0))CALL cp__a("mode_selective.F",590)
         ms_vib%mat_size = mat
      END IF
      CALL mp_bcast(ms_vib%mat_size, para_env%source, para_env%group)
      ALLOCATE (ms_vib%b_mat(ncoord, ms_vib%mat_size))
      ALLOCATE (ms_vib%s_mat(ncoord, ms_vib%mat_size))
      IF (calc_intens) THEN
         ALLOCATE (ms_vib%dip_deriv(3, ms_vib%mat_size+nrep))
      END IF
      IF (ionode) THEN
         statint = 0
         READ (UNIT=hesunit, IOSTAT=stat) ms_vib%b_mat
         READ (UNIT=hesunit, IOSTAT=stat) ms_vib%s_mat
         IF (calc_intens) READ (UNIT=hesunit, IOSTAT=statint) ms_vib%dip_deriv(:, 1:ms_vib%mat_size)
         IF (statint /= 0 .AND. output_unit > 0) WRITE (output_unit, FMT="(/,T2,A)") "**  Error while reading MS_RESTART,", &
            "intensities are requested but not present in restart file **"
         CALL close_file(hesunit)
         IF (output_unit > 0) THEN
            IF (stat /= 0) THEN
               WRITE (output_unit, FMT="(/,T2,A)") "**  Error while reading MS_RESTART **"
            ELSE
               WRITE (output_unit, FMT="(/,T2,A)") "*** RESTART has been read successfully ***"
            END IF
         END IF
      END IF
      CALL mp_bcast(ms_vib%b_mat, para_env%source, para_env%group)
      CALL mp_bcast(ms_vib%s_mat, para_env%source, para_env%group)
      IF (calc_intens) CALL mp_bcast(ms_vib%dip_deriv, para_env%source, para_env%group)
      ALLOCATE (approx_H(ms_vib%mat_size, ms_vib%mat_size))
      ALLOCATE (eigenval(ms_vib%mat_size))
      ALLOCATE (ind(ms_vib%mat_size))

      CALL dgemm('T', 'N', ms_vib%mat_size, ms_vib%mat_size, SIZE(ms_vib%s_mat, 1), 1._dp, ms_vib%b_mat, SIZE(ms_vib%b_mat, 1), &
                 ms_vib%s_mat, SIZE(ms_vib%s_mat, 1), 0._dp, approx_H, ms_vib%mat_size)
      CALL diamat_all(approx_H, eigenval)

      CALL select_vector(ms_vib, nrep, mass, ncoord, approx_H, eigenval, ind, ms_vib%b_vec)
      IF (ms_vib%initial_guess .NE. 4) THEN

         ms_vib%b_vec = 0._dp
         DO i = 1, nrep
            DO j = 1, ms_vib%mat_size
               ms_vib%b_vec(:, i) = ms_vib%b_vec(:, i)+approx_H(j, ind(i))*ms_vib%b_mat(:, j)
            END DO
            ms_vib%b_vec(:, i) = ms_vib%b_vec(:, i)/SQRT(DOT_PRODUCT(ms_vib%b_vec(:, i), ms_vib%b_vec(:, i)))
         END DO

         DEALLOCATE (ms_vib%s_mat)
         DEALLOCATE (ms_vib%b_mat)
         IF (calc_intens) THEN
            DEALLOCATE (ms_vib%dip_deriv)
            ALLOCATE (ms_vib%dip_deriv(3, nrep))
         END IF
      END IF
      DEALLOCATE (approx_H)
      DEALLOCATE (eigenval)
      DEALLOCATE (ind)
      DO i = 1, nrep
         ms_vib%delta_vec(:, i) = ms_vib%b_vec(:, i)/mass(:)
      END DO

   END SUBROUTINE rest_guess

! **************************************************************************************************
!> \brief ...
!> \param ms_vib_section ...
!> \param input ...
!> \param para_env ...
!> \param ms_vib ...
!> \param mass ...
!> \param ncoord ...
!> \param nrep ...
!> \param logger ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE molden_guess(ms_vib_section, input, para_env, ms_vib, mass, ncoord, nrep, logger)
      TYPE(section_vals_type), POINTER                   :: ms_vib_section, input
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(ms_vib_type)                                  :: ms_vib
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      INTEGER                                            :: ncoord, nrep
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'molden_guess', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: at_name
      CHARACTER(LEN=default_path_length)                 :: ms_filename
      CHARACTER(LEN=max_line_length)                     :: info
      INTEGER                                            :: i, istat, iw, j, jj, k, nvibs, &
                                                            output_molden, output_unit, stat
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      REAL(KIND=dp)                                      :: my_val, norm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: freq, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: modes
      REAL(KIND=dp), DIMENSION(3, ncoord/3)              :: pos

      output_unit = cp_logger_get_default_io_unit(logger)

      CALL section_vals_val_get(ms_vib_section, "RESTART_FILE_NAME", c_val=ms_filename)
      IF (ms_filename == "") output_molden = &
         cp_print_key_unit_nr(logger, input, "VIBRATIONAL_ANALYSIS%PRINT%MOLDEN_VIB", &
                              extension=".mol", file_status='UNKNOWN', &
                              file_action="READ")
      IF (para_env%mepos == para_env%source) THEN

         IF (ms_filename == "") THEN
            iw = output_molden
         ELSE
            CALL open_file(file_name=TRIM(ms_filename), &
                           file_status="UNKNOWN", &
                           file_form="FORMATTED", &
                           file_action="READ", &
                           unit_number=iw)
         END IF
         info = ""
         READ (iw, *, IOSTAT=stat) info
         READ (iw, *, IOSTAT=stat) info
         istat = 0
         nvibs = 0
         DO
            READ (iw, *, IOSTAT=stat) info
            istat = istat+stat
            IF (TRIM(ADJUSTL(info)) == "[FR-COORD]") EXIT

            IF(.NOT.(stat == 0))CALL cp__a("mode_selective.F",714)

            nvibs = nvibs+1
         END DO
         REWIND (iw)
         istat = 0
         READ (iw, *, IOSTAT=stat) info
         istat = istat+stat
         READ (iw, *, IOSTAT=stat) info
         istat = istat+stat

         ALLOCATE (freq(nvibs))
         ALLOCATE (modes(ncoord, nvibs))

         DO i = 1, nvibs
            READ (iw, *, IOSTAT=stat) freq(i)
            istat = istat+stat
         END DO
         READ (iw, *) info
         DO i = 1, ncoord/3
            READ (iw, *, IOSTAT=stat) at_name, pos(:, i)
            istat = istat+stat
         END DO
         READ (iw, *) info
         DO i = 1, nvibs
            READ (iw, *) info
            istat = istat+stat
            DO j = 1, ncoord/3
               k = (j-1)*3+1
               READ (iw, *, IOSTAT=stat) modes(k:k+2, i)
               istat = istat+stat
            END DO
         END DO
         IF (ms_filename .NE. "") CALL close_file(iw)
         IF (output_unit > 0) THEN
            IF (istat /= 0) THEN
               WRITE (output_unit, FMT="(/,T2,A)") "**  Error while reading MOLDEN file **"
            ELSE
               WRITE (output_unit, FMT="(/,T2,A)") "*** MOLDEN file has been read successfully ***"
            END IF
         END IF
         !!!!!!!    select modes     !!!!!!
         ALLOCATE (tmp(nvibs))
         tmp(:) = 0.0_dp
         ALLOCATE (tmplist(nvibs))
         IF (ms_vib%select_id == 1) my_val = ms_vib%sel_freq
         IF (ms_vib%select_id == 2) my_val = (ms_vib%f_range(2)+ms_vib%f_range(1))*0.5_dp
         IF (ms_vib%select_id == 1 .OR. ms_vib%select_id == 2) THEN
            DO i = 1, nvibs
               tmp(i) = ABS(my_val-freq(i))
            END DO
         ELSE IF (ms_vib%select_id == 3) THEN
            DO i = 1, nvibs
               DO j = 1, SIZE(ms_vib%inv_atoms)
                  DO k = 1, 3
                     jj = (ms_vib%inv_atoms(j)-1)*3+k
                     tmp(i) = tmp(i)+SQRT(modes(jj, i)**2)
                  END DO
               END DO
               IF (freq(i) .LE. 400._dp) tmp(i) = 0._dp
            END DO
            tmp(:) = -tmp(:)
         END IF
         CALL sort(tmp, nvibs, tmplist)
         DO i = 1, nrep
            ms_vib%b_vec(:, i) = modes(:, tmplist(i))*mass(:)
            norm = SQRT(DOT_PRODUCT(ms_vib%b_vec(:, i), ms_vib%b_vec(:, i)))
            ms_vib%b_vec(:, i) = ms_vib%b_vec(:, i)/norm
         END DO
         DO i = 1, nrep
            ms_vib%delta_vec(:, i) = ms_vib%b_vec(:, i)/mass(:)
         END DO

         DEALLOCATE (freq)
         DEALLOCATE (modes)
         DEALLOCATE (tmp)
         DEALLOCATE (tmplist)

      END IF
      CALL mp_bcast(ms_vib%b_vec, para_env%source, para_env%group)
      CALL mp_bcast(ms_vib%delta_vec, para_env%source, para_env%group)

      IF (ms_filename == "") CALL cp_print_key_finished_output(output_molden, logger, input, &
                                                               "VIBRATIONAL_ANALYSIS%PRINT%MOLDEN_VIB")
   END SUBROUTINE molden_guess

! **************************************************************************************************
!> \brief Davidson algorithm for to generate a approximate Hessian for mode
!>      selective vibrational analysis
!> \param rep_env ...
!> \param ms_vib ...
!> \param input ...
!> \param nrep ...
!> \param particles ...
!> \param mass ...
!> \param converged ...
!> \param dx ...
!> \param calc_intens ...
!> \param output_unit_ms ...
!> \param logger ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE evaluate_H_update_b(rep_env, ms_vib, input, nrep, &
                                  particles, &
                                  mass, &
                                  converged, dx, &
                                  calc_intens, output_unit_ms, logger)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(ms_vib_type)                                  :: ms_vib
      TYPE(section_vals_type), POINTER                   :: input
      INTEGER                                            :: nrep
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      LOGICAL                                            :: converged
      REAL(KIND=dp)                                      :: dx
      LOGICAL                                            :: calc_intens
      INTEGER                                            :: output_unit_ms
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'evaluate_H_update_b', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, jj, k, natoms, ncoord
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ind
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenval, freq
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: approx_H, H_save, residuum, tmp_b, tmp_s
      REAL(KIND=dp), DIMENSION(2, nrep)                  :: criteria
      REAL(Kind=dp), DIMENSION(:), POINTER               :: intensities

      natoms = SIZE(particles)
      ncoord = 3*natoms
      nrep = SIZE(rep_env%f, 2)

      !!!!!!!!   reallocate and update the davidson matrices   !!!!!!!!!!
      IF (ms_vib%mat_size .NE. 0) THEN

         ALLOCATE (tmp_b(3*natoms, ms_vib%mat_size))
         ALLOCATE (tmp_s(3*natoms, ms_vib%mat_size))

         tmp_b(:, :) = ms_vib%b_mat
         tmp_s(:, :) = ms_vib%s_mat

         DEALLOCATE (ms_vib%b_mat)
         DEALLOCATE (ms_vib%s_mat)
      ENDIF

      ALLOCATE (ms_vib%b_mat(3*natoms, ms_vib%mat_size+nrep))
      ALLOCATE (ms_vib%s_mat(3*natoms, ms_vib%mat_size+nrep))

      ms_vib%s_mat = 0.0_dp

      DO i = 1, 3*natoms
         IF (ms_vib%mat_size .NE. 0) THEN
            DO j = 1, ms_vib%mat_size
               ms_vib%b_mat(i, j) = tmp_b(i, j)
               ms_vib%s_mat(i, j) = tmp_s(i, j)
            END DO
         END IF
         DO j = 1, nrep
            ms_vib%b_mat(i, ms_vib%mat_size+j) = ms_vib%b_vec(i, j)
         END DO
      END DO

      IF (ms_vib%mat_size .NE. 0) THEN
         DEALLOCATE (tmp_s)
         DEALLOCATE (tmp_b)
      END IF

      ms_vib%mat_size = ms_vib%mat_size+nrep

      ALLOCATE (approx_H(ms_vib%mat_size, ms_vib%mat_size))
      ALLOCATE (H_save(ms_vib%mat_size, ms_vib%mat_size))
      ALLOCATE (eigenval(ms_vib%mat_size))

      !!!!!!!!!!!!  calculate the new derivativ and the approximate hessian

      DO i = 1, nrep
         DO j = 1, 3*natoms
            ms_vib%s_mat(j, ms_vib%mat_size-nrep+i) = -(ms_vib%ms_force(j, i)-rep_env%f(j, i))/(2*ms_vib%step_b(i)*mass(j))
         END DO
      END DO

      CALL dgemm('T', 'N', ms_vib%mat_size, ms_vib%mat_size, SIZE(ms_vib%s_mat, 1), 1._dp, ms_vib%b_mat, SIZE(ms_vib%b_mat, 1), &
                 ms_vib%s_mat, SIZE(ms_vib%s_mat, 1), 0._dp, approx_H, ms_vib%mat_size)
      H_save(:, :) = approx_H

      CALL diamat_all(approx_H, eigenval)

      !!!!!!!!!!!! select eigenvalue(s) and vector(s) and calculate the new displacement vector
      ALLOCATE (ind(ms_vib%mat_size))
      ALLOCATE (residuum(SIZE(ms_vib%s_mat, 1), nrep))

      CALL select_vector(ms_vib, nrep, mass, ncoord, approx_H, eigenval, ind, residuum, criteria)

      DO i = 1, nrep
         DO j = 1, natoms
            DO k = 1, 3
               jj = (j-1)*3+k
               ms_vib%delta_vec(jj, i) = ms_vib%b_vec(jj, i)/mass(jj)
            END DO
         END DO
      END DO

      DO i = 1, nrep
         ms_vib%step_r(i) = dx/SQRT(DOT_PRODUCT(ms_vib%delta_vec(:, i), ms_vib%delta_vec(:, i)))
         ms_vib%step_b(i) = SQRT(DOT_PRODUCT(ms_vib%step_r(i)*ms_vib%b_vec(:, i), ms_vib%step_r(i)*ms_vib%b_vec(:, i)))
      END DO
      converged = .FALSE.
      IF (MAXVAL(criteria(1, :)) .LE. ms_vib%eps(1) .AND. MAXVAL(criteria(2, :)) &
          .LE. ms_vib%eps(2) .OR. ms_vib%mat_size .GE. ncoord) converged = .TRUE.
      ALLOCATE (freq(nrep))
      DO i = 1, nrep
         freq(i) = SQRT(ABS(eigenval(ind(i)))*massunit)*vibfac
      END DO

      !!!   write information and output   !!!
      IF (converged) THEN
         eigenval(:) = SIGN(1._dp, eigenval(:))*SQRT(ABS(eigenval(:))*massunit)*vibfac
         ALLOCATE (tmp_b(ncoord, ms_vib%mat_size))
         tmp_b = 0._dp
         ALLOCATE (tmp_s(3, ms_vib%mat_size))
         tmp_s = 0._dp
         IF (calc_intens) THEN
            ALLOCATE (intensities(ms_vib%mat_size))
            intensities = 0._dp
         END IF
         DO i = 1, ms_vib%mat_size
            DO j = 1, ms_vib%mat_size
               tmp_b(:, i) = tmp_b(:, i)+approx_H(j, i)*ms_vib%b_mat(:, j)/mass(:)
            END DO
            tmp_b(:, i) = tmp_b(:, i)/SQRT(DOT_PRODUCT(tmp_b(:, i), tmp_b(:, i)))
         END DO
         IF (calc_intens) THEN
            DO i = 1, ms_vib%mat_size
               DO j = 1, ms_vib%mat_size
                  tmp_s(:, i) = tmp_s(:, i)+ms_vib%dip_deriv(:, j)*approx_H(j, i)
               END DO
               IF (calc_intens) intensities(i) = SQRT(DOT_PRODUCT(tmp_s(:, i), tmp_s(:, i)))
            END DO
         END IF
         IF (calc_intens) THEN
            CALL ms_out(output_unit_ms, converged, freq, criteria, ms_vib, &
                        input, nrep, approx_H, eigenval, calc_intens, &
                        intensities=intensities, logger=logger)
         ELSE
            CALL ms_out(output_unit_ms, converged, freq, criteria, ms_vib, &
                        input, nrep, approx_H, eigenval, calc_intens, logger=logger)
         ENDIF
         CALL molden_out(input, particles, eigenval, tmp_b, intensities, calc_intens, &
                         dump_only_positive=.TRUE., logger=logger)
         IF (calc_intens) THEN
            DEALLOCATE (intensities)
         END IF
         DEALLOCATE (tmp_b)
         DEALLOCATE (tmp_s)
      ENDIF

      IF (.NOT. converged) CALL ms_out(output_unit_ms, converged, freq, criteria, &
                                       ms_vib, input, nrep, approx_H, eigenval, calc_intens, logger=logger)

      DEALLOCATE (freq)
      DEALLOCATE (approx_H)
      DEALLOCATE (eigenval)
      DEALLOCATE (residuum)
      DEALLOCATE (ind)

   END SUBROUTINE evaluate_H_update_b

! **************************************************************************************************
!> \brief writes the output for a mode tracking calculation
!> \param ms_vib ...
!> \param nrep ...
!> \param mass ...
!> \param ncoord ...
!> \param approx_H ...
!> \param eigenval ...
!> \param ind ...
!> \param residuum ...
!> \param criteria ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE select_vector(ms_vib, nrep, mass, ncoord, approx_H, eigenval, ind, residuum, criteria)

      TYPE(ms_vib_type)                                  :: ms_vib
      INTEGER                                            :: nrep
      REAL(Kind=dp), DIMENSION(:)                        :: mass
      INTEGER                                            :: ncoord
      REAL(KIND=dp), DIMENSION(:, :)                     :: approx_H
      REAL(Kind=dp), DIMENSION(:)                        :: eigenval
      INTEGER, DIMENSION(:)                              :: ind
      REAL(KIND=dp), DIMENSION(:, :)                     :: residuum
      REAL(KIND=dp), DIMENSION(2, nrep), OPTIONAL        :: criteria

      CHARACTER(len=*), PARAMETER :: routineN = 'select_vector', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, jj, k
      REAL(KIND=dp)                                      :: my_val, norm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: tmp_b

      ALLOCATE (tmp(ms_vib%mat_size))

      SELECT CASE (ms_vib%select_id)
      CASE (1)
         my_val = (ms_vib%sel_freq/(vibfac))**2/massunit
         DO i = 1, ms_vib%mat_size
            tmp(i) = ABS(my_val-eigenval(i))
         END DO
         CALL sort(tmp, (ms_vib%mat_size), ind)
         residuum = 0._dp
         DO j = 1, nrep
            DO i = 1, ms_vib%mat_size
               residuum(:, j) = residuum(:, j)+approx_H(i, ind(j))*(ms_vib%s_mat(:, i)-eigenval(ind(j))*ms_vib%b_mat(:, i))
            END DO
         END DO
      CASE (2)
         CALL get_vibs_in_range(ms_vib, approx_H, eigenval, residuum, nrep, ind)
      CASE (3)

         ALLOCATE (tmp_b(ncoord, ms_vib%mat_size))
         tmp_b = 0._dp

         DO i = 1, ms_vib%mat_size
            DO j = 1, ms_vib%mat_size
               tmp_b(:, i) = tmp_b(:, i)+approx_H(j, i)*ms_vib%b_mat(:, j)/mass(:)
            END DO
            tmp_b(:, i) = tmp_b(:, i)/SQRT(DOT_PRODUCT(tmp_b(:, i), tmp_b(:, i)))
         END DO
         tmp = 0._dp
         DO i = 1, ms_vib%mat_size
            DO j = 1, SIZE(ms_vib%inv_atoms)
               DO k = 1, 3
                  jj = (ms_vib%inv_atoms(j)-1)*3+k
                  tmp(i) = tmp(i)+SQRT(tmp_b(jj, i)**2)
               END DO
            END DO
            IF (.NOT. ASSOCIATED(ms_vib%inv_range)) THEN
               IF ((SIGN(1._dp, eigenval(i))*SQRT(ABS(eigenval(i))*massunit)*vibfac) .LE. 400._dp) tmp(i) = 0._dp
            ELSE
               IF ((SIGN(1._dp, eigenval(i))*SQRT(ABS(eigenval(i))*massunit)*vibfac) .LE. ms_vib%inv_range(1)) tmp(i) = 0._dp
               IF ((SIGN(1._dp, eigenval(i))*SQRT(ABS(eigenval(i))*massunit)*vibfac) .GE. ms_vib%inv_range(2)) tmp(i) = 0._dp
            END IF
         END DO
         tmp(:) = -tmp(:)
         CALL sort(tmp, (ms_vib%mat_size), ind)
         residuum(:, :) = 0._dp

         DO j = 1, nrep
            DO i = 1, ms_vib%mat_size
               residuum(:, j) = residuum(:, j)+approx_H(i, ind(j))*(ms_vib%s_mat(:, i)-eigenval(ind(j))*ms_vib%b_mat(:, i))
            END DO
         END DO
         DEALLOCATE (tmp_b)
      END SELECT

      DO j = 1, nrep
         DO i = 1, ms_vib%mat_size
            residuum(:, j) = residuum(:, j)-DOT_PRODUCT(residuum(:, j), ms_vib%b_mat(:, i))*ms_vib%b_mat(:, i)
         END DO
      END DO
      IF (PRESENT(criteria)) THEN
         DO i = 1, nrep
            criteria(1, i) = MAXVAL((residuum(:, i)))
            criteria(2, i) = SQRT(DOT_PRODUCT(residuum(:, i), residuum(:, i)))
         END DO
      END IF

      DO i = 1, nrep
         norm = SQRT(DOT_PRODUCT(residuum(:, i), residuum(:, i)))
         residuum(:, i) = residuum(:, i)/norm
      END DO

      DO k = 1, 10
         DO j = 1, nrep
            DO i = 1, ms_vib%mat_size
               residuum(:, j) = residuum(:, j)-DOT_PRODUCT(residuum(:, j), ms_vib%b_mat(:, i))*ms_vib%b_mat(:, i)
               residuum(:, j) = residuum(:, j)/SQRT(DOT_PRODUCT(residuum(:, j), residuum(:, j)))
            END DO
            IF (nrep .GT. 1) THEN
               DO i = 1, nrep
                  IF (i .NE. j) THEN
                     residuum(:, j) = residuum(:, j)-DOT_PRODUCT(residuum(:, j), residuum(:, i))*residuum(:, i)
                     residuum(:, j) = residuum(:, j)/SQRT(DOT_PRODUCT(residuum(:, j), residuum(:, j)))
                  END IF
               END DO
            END IF
         END DO
      END DO
      ms_vib%b_vec = residuum
      DEALLOCATE (tmp)
   END SUBROUTINE select_vector

! **************************************************************************************************
!> \brief writes the output for a mode tracking calculation
!> \param iw ...
!> \param converged ...
!> \param freq ...
!> \param criter ...
!> \param ms_vib ...
!> \param input ...
!> \param nrep ...
!> \param approx_H ...
!> \param eigenval ...
!> \param calc_intens ...
!> \param intensities ...
!> \param logger ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE ms_out(iw, converged, freq, criter, ms_vib, input, nrep, &
                     approx_H, eigenval, calc_intens, intensities, logger)

      INTEGER                                            :: iw
      LOGICAL                                            :: converged
      REAL(KIND=dp), DIMENSION(:)                        :: freq
      REAL(KIND=dp), DIMENSION(:, :)                     :: criter
      TYPE(ms_vib_type)                                  :: ms_vib
      TYPE(section_vals_type), POINTER                   :: input
      INTEGER                                            :: nrep
      REAL(KIND=dp), DIMENSION(:, :)                     :: approx_H
      REAL(KIND=dp), DIMENSION(:)                        :: eigenval
      LOGICAL                                            :: calc_intens
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: intensities
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'ms_out', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, msunit, stat
      REAL(KIND=dp)                                      :: crit_a, crit_b
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: residuum
      TYPE(section_vals_type), POINTER                   :: ms_vib_section

      ms_vib_section => section_vals_get_subs_vals(input, &
                                                   "VIBRATIONAL_ANALYSIS%MODE_SELECTIVE")

      IF (converged) THEN
         IF (iw .GT. 0) THEN
            WRITE (iw, '(T2,A)') "MS| DAVIDSON ALGORITHM CONVERGED"
            DO i = 1, nrep
               WRITE (iw, '(T2,"MS| TRACKED FREQUENCY (",I0,") IS:",F12.6,3X,A)') i, freq(i), 'cm-1'
            END DO
            ALLOCATE (residuum(SIZE(ms_vib%b_mat, 1)))
            WRITE (iw, '( /, 1X, 79("-") )')
            WRITE (iw, '( 25X, A)') 'FREQUENCY AND CONVERGENCE LIST'
            IF (PRESENT(intensities)) THEN
               WRITE (iw, '(3X,5(4X, A))') 'FREQUENCY', 'INTENSITY', 'MAXVAL CRITERIA', 'NORM CRITERIA', 'CONVERGENCE'
            ELSE
               WRITE (iw, '(3X,5(4X, A))') 'FREQUENCY', 'MAXVAL CRITERIA', 'NORM CRITERIA', 'CONVERGENCE'
            ENDIF
            DO i = 1, SIZE(ms_vib%b_mat, 2)
               residuum = 0._dp
               DO j = 1, SIZE(ms_vib%b_mat, 2)
                  residuum(:) = residuum(:)+approx_H(j, i)*(ms_vib%s_mat(:, j)-eigenval(i)*ms_vib%b_mat(:, j))
               END DO
               DO j = 1, ms_vib%mat_size
                  residuum(:) = residuum(:)-DOT_PRODUCT(residuum(:), ms_vib%b_mat(:, j))*ms_vib%b_mat(:, j)
               END DO
               crit_a = MAXVAL(residuum(:))
               crit_b = SQRT(DOT_PRODUCT(residuum, residuum))
               IF (PRESENT(intensities)) THEN
                  IF (crit_a .LE. ms_vib%eps(1) .AND. crit_b .LE. ms_vib%eps(2)) THEN
                     IF (eigenval(i) .GT. 0._dp) WRITE (iw, '(2X,A,2X,F8.3,1X,F12.6,3X,E12.3,7X,E12.3,11X,A)') &
                        'VIB|', eigenval(i), intensities(i), crit_a, crit_b, 'YES'
                  ELSE
                     IF (eigenval(i) .GT. 0._dp) WRITE (iw, '(2X,A,2X,F8.3,1X,F12.6,3X,E12.3,7X,E12.3,11X,A)') &
                        'VIB|', eigenval(i), intensities(i), crit_a, crit_b, 'NO'
                  ENDIF
               ELSE
                  IF (crit_a .LE. ms_vib%eps(1) .AND. crit_b .LE. ms_vib%eps(2)) THEN
                     IF (eigenval(i) .GT. 0._dp) WRITE (iw, '(2X,A,2X,F8.3,5X,E12.6,5X,E12.3,11X,A)') &
                        'VIB|', eigenval(i), crit_a, crit_b, 'YES'
                  ELSE
                     IF (eigenval(i) .GT. 0._dp) WRITE (iw, '(2X,A,2X,F8.3,5X,E12.6,5X,E12.3,11X,A)') &
                        'VIB|', eigenval(i), crit_a, crit_b, 'NO'
                  ENDIF
               END IF
            END DO
            DEALLOCATE (residuum)

            msunit = cp_print_key_unit_nr(logger, ms_vib_section, &
                                          "PRINT%MS_RESTART", extension=".bin", middle_name="MS_RESTART", &
                                          file_status="REPLACE", file_form="UNFORMATTED", &
                                          file_action="WRITE")

            IF (msunit > 0) THEN
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%mat_size
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%b_mat
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%s_mat
               IF (calc_intens) WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%dip_deriv
            END IF

            CALL cp_print_key_finished_output(msunit, logger, ms_vib_section, &
                                              "PRINT%MS_RESTART")
         END IF
      ELSE
         IF (iw .GT. 0) THEN
            msunit = cp_print_key_unit_nr(logger, ms_vib_section, &
                                          "PRINT%MS_RESTART", extension=".bin", middle_name="MS_RESTART", &
                                          file_status="REPLACE", file_form="UNFORMATTED", &
                                          file_action="WRITE")

            IF (msunit > 0) THEN
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%mat_size
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%b_mat
               WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%s_mat
               IF (calc_intens) WRITE (UNIT=msunit, IOSTAT=stat) ms_vib%dip_deriv
            END IF

            CALL cp_print_key_finished_output(msunit, logger, ms_vib_section, &
                                              "PRINT%MS_RESTART")

            WRITE (iw, '(T2,A,3X,I6)') "MS| ITERATION STEP", ms_vib%mat_size/nrep
            DO i = 1, nrep
               IF (criter(1, i) .LE. 1E-7 .AND. (criter(2, i)) .LE. 1E-6) THEN
                  WRITE (iw, '(T2,A,3X,F12.6,A)') "MS| TRACKED MODE ", freq(i), "cm-1  IS  CONVERGED"
               ELSE
                  WRITE (iw, '(T2,A,3X,F12.6,A)') "MS| TRACKED MODE ", freq(i), "cm-1  NOT  CONVERGED"
               END IF
            END DO
         END IF
      END IF

   END SUBROUTINE ms_out

! **************************************************************************************************
!> \brief ...
!> \param ms_vib ...
!> \param approx_H ...
!> \param eigenval ...
!> \param residuum ...
!> \param nrep ...
!> \param ind ...
!> \author Florian Schiffmann 11.2007
! **************************************************************************************************
   SUBROUTINE get_vibs_in_range(ms_vib, approx_H, eigenval, residuum, nrep, ind)

      TYPE(ms_vib_type)                                  :: ms_vib
      REAL(KIND=dp), DIMENSION(:, :)                     :: approx_H
      REAL(KIND=dp), DIMENSION(:)                        :: eigenval
      REAL(KIND=dp), DIMENSION(:, :)                     :: residuum
      INTEGER                                            :: nrep
      INTEGER, DIMENSION(:)                              :: ind

      INTEGER                                            :: count1, count2, i, j
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map2
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: map1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tmp, tmp1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: tmp_resid
      REAL(KIND=dp), DIMENSION(2)                        :: myrange

      myrange(:) = (ms_vib%f_range(:)/(vibfac))**2/massunit
      count1 = 0
      count2 = 0
      residuum = 0.0_dp
      ms_vib%mat_size = SIZE(ms_vib%b_mat, 2)
      ALLOCATE (map1(SIZE(eigenval), 2))
      ALLOCATE (tmp(SIZE(eigenval)))
      DO i = 1, SIZE(eigenval)
         IF (ABS(eigenval(i)-myrange(1))+ABS(eigenval(i)-myrange(2)) .LE. &
             ABS(myrange(1)-myrange(2))+myrange(1)*0.001_dp) THEN
            count1 = count1+1
            map1(count1, 1) = i
         ELSE
            count2 = count2+1
            map1(count2, 2) = i
            tmp(count2) = MIN(ABS(eigenval(i)-myrange(1)), ABS(eigenval(i)-myrange(2)))
         END IF
      END DO

      IF (count1 .EQ. nrep) THEN
         DO j = 1, count1
            DO i = 1, ms_vib%mat_size
               residuum(:, j) = residuum(:, j)+approx_H(i, map1(j, 1))*(ms_vib%s_mat(:, i)-eigenval(map1(j, 1))*ms_vib%b_mat(:, i))
               ind(j) = map1(j, 1)
            END DO
         END DO
      ELSE IF (count1 .GT. nrep) THEN
         ALLOCATE (tmp_resid(SIZE(ms_vib%b_mat, 1), count1))
         ALLOCATE (tmp1(count1))
         ALLOCATE (map2(count1))
         tmp_resid = 0._dp
         DO j = 1, count1
            DO i = 1, ms_vib%mat_size
               tmp_resid(:, j) = tmp_resid(:, j)+approx_H(i, map1(j, 1))* &
                                 (ms_vib%s_mat(:, i)-eigenval(map1(j, 1))*ms_vib%b_mat(:, i))
            END DO
         END DO

         DO j = 1, count1
            DO i = 1, ms_vib%mat_size
               tmp_resid(:, j) = tmp_resid(:, j)-DOT_PRODUCT(tmp_resid(:, j), ms_vib%b_mat(:, i))*ms_vib%b_mat(:, i)
            END DO
            tmp(j) = MAXVAL(tmp_resid(:, j))
         END DO
         CALL sort(tmp, count1, map2)
         DO j = 1, nrep
            residuum(:, j) = tmp_resid(:, map2(count1+1-j))
            ind(j) = map1(map2(count1+1-j), 1)
         END DO
         DEALLOCATE (tmp_resid)
         DEALLOCATE (tmp1)
         DEALLOCATE (map2)
      ELSE IF (count1 .LT. nrep) THEN

         ALLOCATE (map2(count2))
         IF (count1 .NE. 0) THEN
            DO j = 1, count1
               DO i = 1, ms_vib%mat_size
                  residuum(:, j) = residuum(:, j)+approx_H(i, map1(j, 1))* &
                                   (ms_vib%s_mat(:, i)-eigenval(map1(j, 1))*ms_vib%b_mat(:, i))
               END DO
               ind(j) = map1(j, 1)
            END DO
         END IF
         CALL sort(tmp, count2, map2)
         DO j = 1, nrep-count1
            DO i = 1, ms_vib%mat_size
               residuum(:, count1+j) = residuum(:, count1+j)+approx_H(i, map1(map2(j), 2)) &
                                       *(ms_vib%s_mat(:, i)-eigenval(map1(map2(j), 2))*ms_vib%b_mat(:, i))
            END DO
            ind(count1+j) = map1(map2(j), 2)
         END DO

         DEALLOCATE (map2)
      END IF

      DEALLOCATE (map1)
      DEALLOCATE (tmp)

   END SUBROUTINE get_vibs_in_range
END MODULE mode_selective
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module performing a vibrational analysis
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 08.2006)
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
MODULE vibrational_analysis
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: get_results
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE header,                          ONLY: vib_header
   USE input_constants,                 ONLY: do_rep_blocked
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: diamat_all
   USE mode_selective,                  ONLY: ms_vb_anal
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molden_utils,                    ONLY: molden_out
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE motion_utils,                    ONLY: rot_ana,&
                                              thrs_motion
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: write_particle_matrix
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: &
        a_bohr, boltzmann, e_mass, h_bar, hertz, kelvin, kjmol, massunit, n_avogadro, pascal, &
        vibfac, wavenumbers
   USE replica_methods,                 ONLY: rep_env_calc_e_f,&
                                              rep_env_create
   USE replica_types,                   ONLY: rep_env_release,&
                                              replica_env_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'vibrational_analysis'
   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.

   PUBLIC :: vb_anal

CONTAINS

! **************************************************************************************************
!> \brief Module performing a vibrational analysis
!> \param input ...
!> \param input_declaration ...
!> \param para_env ...
!> \param globenv ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE vb_anal(input, input_declaration, para_env, globenv)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'vb_anal', routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(3), PARAMETER          :: lab = (/"X", "Y", "Z"/)

      CHARACTER(LEN=default_string_length)               :: description
      INTEGER :: handle, i, icoord, icoordm, icoordp, ierr, imap, ip1, ip2, iparticle1, &
         iparticle2, iseq, iw, j, k, natoms, ncoord, nrep, nres, nRotTrM, nvib, output_unit, prep, &
         proc_dist_type
      INTEGER, DIMENSION(:), POINTER                     :: Clist, Mlist
      LOGICAL                                            :: calc_intens, calc_thchdata, &
                                                            do_mode_tracking, keep_rotations, &
                                                            row_force, something_frozen
      REAL(KIND=dp)                                      :: dx, inertia(3), minimum_energy, norm, &
                                                            tc_press, tc_temp, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: H_eigval1, H_eigval2, konst, mass, pos0, &
                                                            rmass
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Hessian, Hint1, Hint2
      REAL(KIND=dp), DIMENSION(3)                        :: D_deriv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: intensities
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: D, dip_deriv, RotTrM
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: tmp_dip
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, &
                                                            mode_tracking_section, print_section, &
                                                            vib_section

      CALL timeset(routineN, handle)
      NULLIFY (D, RotTrM, logger, subsys, f_env, particles, rep_env, intensities, &
               vib_section, print_section)
      logger => cp_get_default_logger()
      vib_section => section_vals_get_subs_vals(input, "VIBRATIONAL_ANALYSIS")
      print_section => section_vals_get_subs_vals(vib_section, "PRINT")
      output_unit = cp_print_key_unit_nr(logger, print_section, "PROGRAM_RUN_INFO", extension=".vibLog")

      CALL section_vals_val_get(vib_section, "DX", r_val=dx)
      CALL section_vals_val_get(vib_section, "NPROC_REP", i_val=prep)
      CALL section_vals_val_get(vib_section, "PROC_DIST_TYPE", i_val=proc_dist_type)
      row_force = (proc_dist_type == do_rep_blocked)
      CALL section_vals_val_get(vib_section, "FULLY_PERIODIC", l_val=keep_rotations)
      CALL section_vals_val_get(vib_section, "INTENSITIES", l_val=calc_intens)
      CALL section_vals_val_get(vib_section, "THERMOCHEMISTRY", l_val=calc_thchdata)
      CALL section_vals_val_get(vib_section, "TC_TEMPERATURE", r_val=tc_temp)
      CALL section_vals_val_get(vib_section, "TC_PRESSURE", r_val=tc_press)
      tc_temp = tc_temp*kelvin
      tc_press = tc_press*pascal

      mode_tracking_section => section_vals_get_subs_vals(vib_section, "MODE_SELECTIVE")
      CALL section_vals_get(mode_tracking_section, explicit=do_mode_tracking)
      nrep = MAX(1, para_env%num_pe/prep)
      prep = para_env%num_pe/nrep
      iw = cp_print_key_unit_nr(logger, print_section, "BANNER", extension=".vibLog")
      CALL vib_header(iw, nrep, prep)
      CALL cp_print_key_finished_output(iw, logger, print_section, "BANNER")
      ! Just one force_env allowed
      force_env_section => section_vals_get_subs_vals(input, "FORCE_EVAL")
      ! Create Replica Environments
      CALL rep_env_create(rep_env, para_env=para_env, input=input, &
                          input_declaration=input_declaration, nrep=nrep, prep=prep, row_force=row_force)
      IF (ASSOCIATED(rep_env)) THEN
         CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
         CALL force_env_get(f_env%force_env, subsys=subsys)
         particles => subsys%particles%els
         ! Decide which kind of Vibrational Analysis to perform
         IF (do_mode_tracking) THEN
            CALL ms_vb_anal(input, rep_env, para_env, globenv, particles, &
                            nrep, calc_intens, dx, output_unit, logger)
            CALL f_env_rm_defaults(f_env, ierr)
         ELSE
            CALL get_moving_atoms(force_env=f_env%force_env, Ilist=Mlist)
            something_frozen = SIZE(particles) .NE. SIZE(Mlist)
            natoms = SIZE(Mlist)
            ncoord = natoms*3
            ALLOCATE (Clist(ncoord))
            ALLOCATE (mass(natoms))
            ALLOCATE (pos0(ncoord))
            ALLOCATE (Hessian(ncoord, ncoord))
            IF (calc_intens) THEN
               description = '[DIPOLE]'
               ALLOCATE (tmp_dip(ncoord, 3, 2))
               tmp_dip = 0._dp
            END IF
            Clist = 0
            DO i = 1, natoms
               imap = Mlist(i)
               Clist((i-1)*3+1) = (imap-1)*3+1
               Clist((i-1)*3+2) = (imap-1)*3+2
               Clist((i-1)*3+3) = (imap-1)*3+3
               mass(i) = particles(imap)%atomic_kind%mass
               IF(.NOT.(mass(i) > 0.0_dp))CALL cp__a("motion/vibrational_analysis.F",181)
               mass(i) = SQRT(mass(i))
               pos0((i-1)*3+1) = particles(imap)%r(1)
               pos0((i-1)*3+2) = particles(imap)%r(2)
               pos0((i-1)*3+3) = particles(imap)%r(3)
            END DO
            !
            ! Determine the principal axes of inertia.
            ! Generation of coordinates in the rotating and translating frame
            !
            IF (something_frozen) THEN
               nRotTrM = 0
               ALLOCATE (RotTrM(natoms*3, nRotTrM))
            ELSE
               CALL rot_ana(particles, RotTrM, nRotTrM, print_section, &
                            keep_rotations, mass_weighted=.TRUE., natoms=natoms, inertia=inertia)
            END IF
            ! Generate the suitable rototranslating basis set
            CALL build_D_matrix(RotTrM, nRotTrM, D, full=.FALSE., &
                                natoms=natoms)
            !
            ! Loop on atoms and coordinates
            !
            Hessian = HUGE(0.0_dp)
            IF (output_unit > 0) WRITE (output_unit, '(/,T2,A)') "VIB| Vibrational Analysis Info"
            DO icoordp = 1, ncoord, nrep
               icoord = icoordp-1
               DO j = 1, nrep
                  DO i = 1, ncoord
                     imap = Clist(i)
                     rep_env%r(imap, j) = pos0(i)
                  END DO
                  IF (icoord+j <= ncoord) THEN
                     imap = Clist(icoord+j)
                     rep_env%r(imap, j) = rep_env%r(imap, j)+Dx
                  END IF
               END DO
               CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)

               DO j = 1, nrep
                  IF (calc_intens) THEN
                     IF (icoord+j <= ncoord) THEN
                        CALL get_results(results=rep_env%results(j)%results, &
                                         description=description, &
                                         n_rep=nres)
                        CALL get_results(results=rep_env%results(j)%results, &
                                         description=description, &
                                         values=tmp_dip(icoord+j, :, 1), &
                                         nval=nres)
                     END IF
                  END IF
                  IF (icoord+j <= ncoord) THEN
                     DO i = 1, ncoord
                        imap = Clist(i)
                        Hessian(i, icoord+j) = rep_env%f(imap, j)
                     END DO
                     imap = Clist(icoord+j)
                     ! Dump Info
                     IF (output_unit > 0) THEN
                        iparticle1 = imap/3
                        IF (MOD(imap, 3) /= 0) iparticle1 = iparticle1+1
                        WRITE (output_unit, '(T2,A,I5,A,I5,3A)') &
                           "VIB| REPLICA Nr.", j, "- Energy and Forces for particle:", &
                           iparticle1, "  coordinate: ", lab(imap-(iparticle1-1)*3), &
                           " + D"//TRIM(lab(imap-(iparticle1-1)*3))
                        !
                        WRITE (output_unit, '(T2,A,T42,A,9X,F15.9)') &
                           "VIB|", " Total Energy: ", rep_env%f(rep_env%ndim+1, j)
                        WRITE (output_unit, '(T2,"VIB|",T10,"ATOM",T33,3(9X,A,7X))') lab(1), lab(2), lab(3)
                        DO i = 1, natoms
                           imap = Mlist(i)
                           WRITE (output_unit, '(T2,"VIB|",T12,A,T30,3(2X,F15.9))') &
                              particles(imap)%atomic_kind%name, &
                              rep_env%f((imap-1)*3+1:(imap-1)*3+3, j)
                        END DO
                     END IF
                  END IF
               END DO
            END DO
            DO icoordm = 1, ncoord, nrep
               icoord = icoordm-1
               DO j = 1, nrep
                  DO i = 1, ncoord
                     imap = Clist(i)
                     rep_env%r(imap, j) = pos0(i)
                  END DO
                  IF (icoord+j <= ncoord) THEN
                     imap = Clist(icoord+j)
                     rep_env%r(imap, j) = rep_env%r(imap, j)-Dx
                  END IF
               END DO
               CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)

               DO j = 1, nrep
                  IF (calc_intens) THEN
                     IF (icoord+j <= ncoord) THEN
                        k = (icoord+j+2)/3
                        CALL get_results(results=rep_env%results(j)%results, &
                                         description=description, &
                                         n_rep=nres)
                        CALL get_results(results=rep_env%results(j)%results, &
                                         description=description, &
                                         values=tmp_dip(icoord+j, :, 2), &
                                         nval=nres)
                        tmp_dip(icoord+j, :, 1) = (tmp_dip(icoord+j, :, 1)-tmp_dip(icoord+j, :, 2))/(2.0_dp*Dx*mass(k))
                     END IF
                  END IF
                  IF (icoord+j <= ncoord) THEN
                     imap = Clist(icoord+j)
                     iparticle1 = imap/3
                     IF (MOD(imap, 3) /= 0) iparticle1 = iparticle1+1
                     ip1 = (icoord+j)/3
                     IF (MOD(icoord+j, 3) /= 0) ip1 = ip1+1
                     ! Dump Info
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, '(T2,A,I5,A,I5,3A)') &
                           "VIB| REPLICA Nr.", j, "- Energy and Forces for particle:", &
                           iparticle1, "  coordinate: ", lab(imap-(iparticle1-1)*3), &
                           " - D"//TRIM(lab(imap-(iparticle1-1)*3))
                        !
                        WRITE (output_unit, '(T2,A,T42,A,9X,F15.9)') &
                           "VIB|", " Total Energy: ", rep_env%f(rep_env%ndim+1, j)
                        WRITE (output_unit, '(T2,"VIB|",T10,"ATOM",T33,3(9X,A,7X))') lab(1), lab(2), lab(3)
                        DO i = 1, natoms
                           imap = Mlist(i)
                           WRITE (output_unit, '(T2,"VIB|",T12,A,T30,3(2X,F15.9))') &
                              particles(imap)%atomic_kind%name, &
                              rep_env%f((imap-1)*3+1:(imap-1)*3+3, j)
                        END DO
                     END IF
                     DO iseq = 1, ncoord
                        imap = Clist(iseq)
                        iparticle2 = imap/3
                        IF (MOD(imap, 3) /= 0) iparticle2 = iparticle2+1
                        ip2 = iseq/3
                        IF (MOD(iseq, 3) /= 0) ip2 = ip2+1
                        tmp = Hessian(iseq, icoord+j)-rep_env%f(imap, j)
                        tmp = -tmp/(2.0_dp*Dx*mass(ip1)*mass(ip2))*1E6_dp
                        ! Mass weighted Hessian
                        Hessian(iseq, icoord+j) = tmp

                     END DO
                  END IF
               END DO
            END DO

            ! restore original particle positions for output
            DO i = 1, natoms
               imap = Mlist(i)
               particles(imap)%r(1:3) = pos0((i-1)*3+1:(i-1)*3+3)
            ENDDO
            DO j = 1, nrep
               DO i = 1, ncoord
                  imap = Clist(i)
                  rep_env%r(imap, j) = pos0(i)
               END DO
               CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)
            ENDDO
            j = 1
            minimum_energy = rep_env%f(rep_env%ndim+1, j)
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,A)') &
                  "VIB| ", " Minimum Structure - Energy and Forces:"
               !
               WRITE (output_unit, '(T2,A,T42,A,9X,F15.9)') &
                  "VIB|", " Total Energy: ", rep_env%f(rep_env%ndim+1, j)
               WRITE (output_unit, '(T2,"VIB|",T10,"ATOM",T33,3(9X,A,7X))') lab(1), lab(2), lab(3)
               DO i = 1, natoms
                  imap = Mlist(i)
                  WRITE (output_unit, '(T2,"VIB|",T12,A,T30,3(2X,F15.9))') &
                     particles(imap)%atomic_kind%name, &
                     rep_env%f((imap-1)*3+1:(imap-1)*3+3, j)
               END DO
            END IF

            ! Dump Info
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,A)') "VIB| Hessian in cartesian coordinates"
               CALL write_particle_matrix(Hessian, particles, output_unit, el_per_part=3, &
                                          Ilist=Mlist)
            END IF
            ! Enforce symmetry in the Hessian
            DO i = 1, ncoord
               DO j = i, ncoord
                  ! Take the upper diagonal part
                  Hessian(j, i) = Hessian(i, j)
               END DO
            END DO
            nvib = ncoord-nRotTrM
            ALLOCATE (H_eigval1(ncoord))
            ALLOCATE (H_eigval2(SIZE(D, 2)))
            ALLOCATE (Hint1(ncoord, ncoord))
            ALLOCATE (Hint2(SIZE(D, 2), SIZE(D, 2)))
            ALLOCATE (rmass(SIZE(D, 2)))
            ALLOCATE (konst(SIZE(D, 2)))
            IF (calc_intens) THEN
               ALLOCATE (dip_deriv(3, SIZE(D, 2)))
               dip_deriv = 0.0_dp
            END IF
            ALLOCATE (intensities(SIZE(D, 2)))
            intensities = 0._dp
            Hint1(:, :) = Hessian
            CALL diamat_all(Hint1, H_eigval1)
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,"VIB| Cartesian Low frequencies ---",4G12.5)') &
                  (H_eigval1(i), i=1, MIN(9, ncoord))
            END IF
            IF (nvib /= 0) THEN
               Hint2(:, :) = MATMUL(TRANSPOSE(D), MATMUL(Hessian, D))
               IF (calc_intens) THEN
                  DO i = 1, 3
                     dip_deriv(i, :) = MATMUL(tmp_dip(:, i, 1), D)
                  END DO
               END IF
               CALL diamat_all(Hint2, H_eigval2)
               IF (output_unit > 0) THEN
                  WRITE (output_unit, '(T2,"VIB| Frequencies after removal of the rotations and translations")')
                  ! Frequency at the moment are in a.u.
                  WRITE (output_unit, '(T2,"VIB| Internal  Low frequencies ---",4G12.5)') H_eigval2
               END IF
               Hessian = 0.0_dp
               DO i = 1, natoms
                  DO j = 1, 3
                     Hessian((i-1)*3+j, (i-1)*3+j) = 1.0_dp/mass(i)
                  END DO
               END DO
               ! Cartesian displacements of the normal modes
               D = MATMUL(Hessian, MATMUL(D, Hint2))
               DO i = 1, nvib
                  norm = 1.0_dp/SUM(D(:, i)*D(:, i))
                  ! Reduced Masess
                  rmass(i) = norm/massunit
                  ! Renormalize displacements and convert in Angstrom
                  D(:, i) = SQRT(norm)*D(:, i)
                  ! Force constants
                  konst(i) = SIGN(1.0_dp, H_eigval2(i))*2.0_dp*pi**2*(ABS(H_eigval2(i))/massunit)**2*rmass(i)

                  IF (calc_intens) THEN
                     D_deriv = 0._dp
                     DO j = 1, nvib
                        D_deriv(:) = D_deriv(:)+dip_deriv(:, j)*Hint2(j, i)
                     END DO
                     intensities(i) = SQRT(DOT_PRODUCT(D_deriv, D_deriv))
                  END IF
                  ! Convert frequencies to cm^-1
                  H_eigval2(i) = SIGN(1.0_dp, H_eigval2(i))*SQRT(ABS(H_eigval2(i))*massunit)*vibfac/1000.0_dp
               END DO
               ! Dump Info
               iw = cp_logger_get_default_io_unit(logger)
               IF (iw > 0) THEN
                  CALL vib_out(iw, nvib, D, konst, rmass, H_eigval2, particles, Mlist, intensities)
               END IF
               IF (.NOT. something_frozen) THEN
                  IF (calc_thchdata) THEN
                     CALL get_thch_values(H_eigval2, iw, mass, nvib, inertia, 1, minimum_energy, tc_temp, tc_press)
                  ENDIF
                  CALL molden_out(input, particles, H_eigval2, D, intensities, calc_intens, &
                                  dump_only_positive=.FALSE., logger=logger)
               END IF
            ELSE
               IF (output_unit > 0) THEN
                  WRITE (output_unit, '(T2,"VIB| No further vibrational info. Detected a single atom")')
               END IF
            END IF
            ! Deallocate working arrays
            DEALLOCATE (Clist)
            DEALLOCATE (Mlist)
            DEALLOCATE (H_eigval1)
            DEALLOCATE (H_eigval2)
            DEALLOCATE (Hint1)
            DEALLOCATE (Hint2)
            DEALLOCATE (rmass)
            DEALLOCATE (konst)
            DEALLOCATE (mass)
            DEALLOCATE (pos0)
            DEALLOCATE (D)
            DEALLOCATE (Hessian)
            IF (calc_intens) THEN
               DEALLOCATE (dip_deriv)
               DEALLOCATE (tmp_dip)
            END IF
            DEALLOCATE (intensities)
            CALL f_env_rm_defaults(f_env, ierr)
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, print_section, "PROGRAM_RUN_INFO")
      CALL rep_env_release(rep_env)
      CALL timestop(handle)
   END SUBROUTINE vb_anal

! **************************************************************************************************
!> \brief give back a list of moving atoms
!> \param force_env ...
!> \param Ilist ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE get_moving_atoms(force_env, Ilist)
      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER, DIMENSION(:), POINTER                     :: Ilist

      CHARACTER(len=*), PARAMETER :: routineN = 'get_moving_atoms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, ikind, j, ndim, &
                                                            nfixed_atoms, nfixed_atoms_total, nkind
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ifixd_list, work
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      CALL force_env_get(force_env=force_env, subsys=subsys)

      CALL cp_subsys_get(subsys=subsys, particles=particles, &
                         molecule_kinds_new=molecule_kinds)

      nkind = molecule_kinds%n_els
      molecule_kind_set => molecule_kinds%els
      particle_set => particles%els

      ! Count the number of fixed atoms
      nfixed_atoms_total = 0
      DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, nfixd=nfixed_atoms)
         nfixed_atoms_total = nfixed_atoms_total+nfixed_atoms
      END DO
      ndim = SIZE(particle_set)-nfixed_atoms_total
      IF(.NOT.(ndim >= 0))CALL cp__a("motion/vibrational_analysis.F",513)
      ALLOCATE (Ilist(ndim))

      IF (nfixed_atoms_total /= 0) THEN
         ALLOCATE (ifixd_list(nfixed_atoms_total))
         ALLOCATE (work(nfixed_atoms_total))
         nfixed_atoms_total = 0
         DO ikind = 1, nkind
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, fixd_list=fixd_list)
            IF (ASSOCIATED(fixd_list)) THEN
               DO ii = 1, SIZE(fixd_list)
                  IF (.NOT. fixd_list(ii)%restraint%active) THEN
                     nfixed_atoms_total = nfixed_atoms_total+1
                     ifixd_list(nfixed_atoms_total) = fixd_list(ii)%fixd
                  END IF
               END DO
            END IF
         END DO
         CALL sort(ifixd_list, nfixed_atoms_total, work)

         ndim = 0
         j = 1
         Loop_count: DO i = 1, SIZE(particle_set)
            DO WHILE (i > ifixd_list(j))
               j = j+1
               IF (j > nfixed_atoms_total) EXIT Loop_count
            END DO
            IF (i /= ifixd_list(j)) THEN
               ndim = ndim+1
               Ilist(ndim) = i
            END IF
         END DO Loop_count
         DEALLOCATE (ifixd_list)
         DEALLOCATE (work)
      ELSE
         i = 1
         ndim = 0
      END IF
      DO j = i, SIZE(particle_set)
         ndim = ndim+1
         Ilist(ndim) = j
      END DO
      CALL timestop(handle)

   END SUBROUTINE get_moving_atoms

! **************************************************************************************************
!> \brief Dumps results of the vibrational analysis
!> \param iw ...
!> \param nvib ...
!> \param D ...
!> \param k ...
!> \param m ...
!> \param freq ...
!> \param particles ...
!> \param Mlist ...
!> \param intensities ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE vib_out(iw, nvib, D, k, m, freq, particles, Mlist, intensities)
      INTEGER, INTENT(IN)                                :: iw, nvib
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: D
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: k, m, freq
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      INTEGER, DIMENSION(:), POINTER                     :: Mlist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: intensities

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: from, iatom, icol, j, jatom, katom, &
                                                            natom, to

      natom = SIZE(D, 1)
      WRITE (UNIT=iw, FMT="(/,T2,'VIB|',T30,'NORMAL MODES - CARTESIAN DISPLACEMENTS')")
      WRITE (UNIT=iw, FMT="(T2,'VIB|')")
      DO jatom = 1, nvib, 3
         from = jatom
         to = MIN(from+2, nvib)
         WRITE (UNIT=iw, FMT="(T2,'VIB|',13X,3(8X,I5,8X))") &
            (icol, icol=from, to)
         WRITE (UNIT=iw, FMT="(T2,'VIB|Frequency (cm^-1)',3(1X,F12.6,8X))") &
            (freq(icol), icol=from, to)
         IF (ASSOCIATED(intensities)) THEN
            WRITE (UNIT=iw, FMT="(T2,'VIB|Intensities      ',3(1X,F12.6,8X))") &
               (intensities(icol), icol=from, to)
         END IF
         WRITE (UNIT=iw, FMT="(T2,'VIB|Red.Masses (a.u.)',3(1X,F12.6,8X))") &
            (m(icol), icol=from, to)
         WRITE (UNIT=iw, FMT="(T2,'VIB|Frc consts (a.u.)',3(1X,F12.6,8X))") &
            (k(icol), icol=from, to)
         WRITE (UNIT=iw, FMT="(T2,' ATOM',2X,'EL',7X,3(4X,'  X  ',1X,'  Y  ',1X,'  Z  '))")
         DO iatom = 1, natom, 3
            katom = iatom/3
            IF (MOD(iatom, 3) /= 0) katom = katom+1
            CALL get_atomic_kind(atomic_kind=particles(Mlist(katom))%atomic_kind, &
                                 element_symbol=element_symbol)
            WRITE (UNIT=iw, FMT="(T2,I5,2X,A2,7X,3(4X,2(F5.2,1X),F5.2))") &
               Mlist(katom), element_symbol, &
               ((D(iatom+j, icol), j=0, 2), icol=from, to)
         END DO
         WRITE (UNIT=iw, FMT="(/)")
      END DO

   END SUBROUTINE vib_out

! **************************************************************************************************
!> \brief Generates the transformation matrix from hessian in cartesian into
!>      internal coordinates (based on Gram-Schmidt orthogonalization)
!> \param mat ...
!> \param dof ...
!> \param Dout ...
!> \param full ...
!> \param natoms ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE build_D_matrix(mat, dof, Dout, full, natoms)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat
      INTEGER, INTENT(IN)                                :: dof
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Dout
      LOGICAL, OPTIONAL                                  :: full
      INTEGER, INTENT(IN)                                :: natoms

      CHARACTER(len=*), PARAMETER :: routineN = 'build_D_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ifound, iseq, j, nvib
      LOGICAL                                            :: my_full
      REAL(KIND=dp)                                      :: norm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: D

      CALL timeset(routineN, handle)
      my_full = .TRUE.
      IF (PRESENT(full)) my_full = full
      ! Generate the missing vectors of the orthogonal basis set
      nvib = 3*natoms-dof
      ALLOCATE (work(3*natoms))
      ALLOCATE (D(3*natoms, 3*natoms))
      ! Check First orthogonality in the first element of the basis set
      DO i = 1, dof
         D(:, i) = mat(:, i)
         DO j = i+1, dof
            norm = DOT_PRODUCT(mat(:, i), mat(:, j))
            IF(.NOT.(ABS(norm) < thrs_motion))CALL cp__a("motion/vibrational_analysis.F",655)
         END DO
      END DO
      ! Generate the nvib orthogonal vectors
      iseq = 0
      ifound = 0
      DO WHILE (ifound /= nvib)
         iseq = iseq+1
         IF(.NOT.(iseq <= 3*natoms))CALL cp__a("motion/vibrational_analysis.F",663)
         work = 0.0_dp
         work(iseq) = 1.0_dp
         ! Gram Schmidt orthogonalization
         DO i = 1, dof+ifound
            norm = DOT_PRODUCT(work, D(:, i))
            work(:) = work-norm*D(:, i)
         END DO
         ! Check norm of the new generated vector
         norm = SQRT(DOT_PRODUCT(work, work))
         IF (norm >= 10E4_dp*thrs_motion) THEN
            ! Accept new vector
            ifound = ifound+1
            D(:, dof+ifound) = work/norm
         END IF
      END DO
      IF(.NOT.(dof+ifound == 3*natoms))CALL cp__a("motion/vibrational_analysis.F",679)
      IF (my_full) THEN
         ALLOCATE (Dout(3*natoms, 3*natoms))
         Dout = D
      ELSE
         ALLOCATE (Dout(3*natoms, nvib))
         Dout = D(:, dof+1:)
      END IF
      DEALLOCATE (work)
      DEALLOCATE (D)
      DEALLOCATE (mat)
      CALL timestop(handle)
   END SUBROUTINE build_D_matrix

! **************************************************************************************************
!> \brief Calculate a few thermochemical  properties from vibrational analysis
!>         It is supposed to work for molecules in the gas phase and without constraints
!> \param freqs ...
!> \param iw ...
!> \param mass ...
!> \param nvib ...
!> \param inertia ...
!> \param spin ...
!> \param totene ...
!> \param temp ...
!> \param pressure ...
!> \author MI 10:2015
! **************************************************************************************************

   SUBROUTINE get_thch_values(freqs, iw, mass, nvib, inertia, spin, totene, temp, pressure)

      REAL(KIND=dp), DIMENSION(:)                        :: freqs
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), DIMENSION(:)                        :: mass
      INTEGER, INTENT(IN)                                :: nvib
      REAL(KIND=dp), INTENT(IN)                          :: inertia(3)
      INTEGER, INTENT(IN)                                :: spin
      REAL(KIND=dp), INTENT(IN)                          :: totene, temp, pressure

      CHARACTER(len=*), PARAMETER :: routineN = 'get_thch_values', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, natoms, sym_num
      REAL(KIND=dp) :: el_entropy, entropy, exp_min_one, fact, fact2, freq_arg, freq_arg2, &
         freqsum, Gibbs, heat_capacity, inertia_kg(3), mass_tot, one_min_exp, partition_function, &
         rot_cv, rot_energy, rot_entropy, rot_part_func, rotvibtra, tran_cv, tran_energy, &
         tran_enthalpy, tran_entropy, tran_part_func, vib_cv, vib_energy, vib_entropy, &
         vib_part_func, zpe
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mass_kg

!    temp = 273.150_dp ! in Kelvin
!    pressure = 101325.0_dp ! in Pascal

      freqsum = 0.0_dp
      DO i = 1, nvib
         freqsum = freqsum+freqs(i)
      ENDDO

!   ZPE
      zpe = 0.5_dp*(h_bar*2._dp*pi)*freqsum*(hertz/wavenumbers)*n_avogadro

      el_entropy = (n_avogadro*boltzmann)*LOG(REAL(spin))
!
      natoms = SIZE(mass)
      ALLOCATE (mass_kg(natoms))
      mass_kg(:) = mass(:)**2*e_mass
      mass_tot = SUM(mass_kg)
      inertia_kg = inertia*e_mass*(a_bohr**2)

!   ROTATIONAL: Partition function and Entropy
      sym_num = 1
      fact = temp*2.0_dp*boltzmann/(h_bar*h_bar)
      IF (inertia_kg(1)*inertia_kg(2)*inertia_kg(3) > 1.0_dp) THEN
         rot_part_func = fact*fact*fact*inertia_kg(1)*inertia_kg(2)*inertia_kg(3)*pi
         rot_part_func = SQRT(rot_part_func)
         rot_entropy = n_avogadro*boltzmann*(LOG(rot_part_func)+1.5_dp)
         rot_energy = 1.5_dp*n_avogadro*boltzmann*temp
         rot_cv = 1.5_dp*n_avogadro*boltzmann
      ELSE
         !linear molecule
         IF (inertia_kg(1) > 1.0_dp) THEN
            rot_part_func = fact*inertia_kg(1)
         ELSE IF (inertia_kg(2) > 1.0_dp) THEN
            rot_part_func = fact*inertia_kg(2)
         ELSE
            rot_part_func = fact*inertia_kg(3)
         END IF
         rot_entropy = n_avogadro*boltzmann*(LOG(rot_part_func)+1.0_dp)
         rot_energy = n_avogadro*boltzmann*temp
         rot_cv = n_avogadro*boltzmann
      END IF

!   TRANSLATIONAL: Partition function and Entropy
      tran_part_func = (boltzmann*temp)**2.5_dp/(pressure*(h_bar*2.0_dp*pi)**3.0_dp)*(2.0_dp*pi*mass_tot)**1.5_dp
      tran_entropy = n_avogadro*boltzmann*(LOG(tran_part_func)+2.5_dp)
      tran_energy = 1.5_dp*n_avogadro*boltzmann*temp
      tran_enthalpy = 2.5_dp*n_avogadro*boltzmann*temp
      tran_cv = 2.5_dp*n_avogadro*boltzmann

!   VIBRATIONAL:  Partition fuction and Entropy
      vib_part_func = 1.0_dp
      vib_energy = 0.0_dp
      vib_entropy = 0.0_dp
      vib_cv = 0.0_dp
      fact = 2.0_dp*pi*h_bar/boltzmann/temp*hertz/wavenumbers
      fact2 = 2.0_dp*pi*h_bar*hertz/wavenumbers
      DO i = 1, nvib
         freq_arg = fact*freqs(i)
         freq_arg2 = fact2*freqs(i)
         exp_min_one = EXP(freq_arg)-1.0_dp
         one_min_exp = 1.0_dp-EXP(-freq_arg)
!dbg
!  write(*,*) 'freq ', i, freqs(i), exp_min_one , one_min_exp
!      vib_part_func = vib_part_func*(1.0_dp/(1.0_dp - exp(-fact*freqs(i))))
         vib_part_func = vib_part_func*(1.0_dp/one_min_exp)
!      vib_energy = vib_energy + fact2*freqs(i)*0.5_dp+fact2*freqs(i)/(exp(fact*freqs(i))-1.0_dp)
         vib_energy = vib_energy+freq_arg2*0.5_dp+freq_arg2/exp_min_one
!      vib_entropy = vib_entropy +fact*freqs(i)/(exp(fact*freqs(i))-1.0_dp)-log(1.0_dp - exp(-fact*freqs(i)))
         vib_entropy = vib_entropy+freq_arg/exp_min_one-LOG(one_min_exp)
!      vib_cv = vib_cv + fact*fact*freqs(i)*freqs(i)*exp(fact*freqs(i))/(exp(fact*freqs(i))-1.0_dp)/(exp(fact*freqs(i))-1.0_dp)
         vib_cv = vib_cv+freq_arg*freq_arg*EXP(freq_arg)/exp_min_one/exp_min_one
      ENDDO
      vib_energy = vib_energy*n_avogadro ! it contains already ZPE
      vib_entropy = vib_entropy*(n_avogadro*boltzmann)
      vib_cv = vib_cv*(n_avogadro*boltzmann)

!   SUMMARY
!dbg
!    write(*,*) 'part ', rot_part_func,tran_part_func,vib_part_func
      partition_function = rot_part_func*tran_part_func*vib_part_func
!dbg
!    write(*,*) 'entropy ', el_entropy,rot_entropy,tran_entropy,vib_entropy

      entropy = el_entropy+rot_entropy+tran_entropy+vib_entropy
!dbg
!    write(*,*) 'energy ', rot_energy , tran_enthalpy , vib_energy, totene*kjmol*1000.0_dp

      rotvibtra = rot_energy+tran_enthalpy+vib_energy
!dbg
!    write(*,*) 'cv ', rot_cv, tran_cv, vib_cv
      heat_capacity = vib_cv+tran_cv+rot_cv

!   Free energy in J/mol: internal energy + PV - TS
      Gibbs = vib_energy+rot_energy+tran_enthalpy-temp*entropy

      DEALLOCATE (mass_kg)

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,T2,'VIB|',T30,'NORMAL MODES - THERMOCHEMICAL DATA')")
         WRITE (UNIT=iw, FMT="(T2,'VIB|')")

         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Symmetry number:',T70,I16)") sym_num
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Temperature [K]:',T70,F16.2)") temp
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Pressure [Pa]:',T70,F16.2)") pressure

         WRITE (UNIT=iw, FMT="(/)")

         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Electronic energy (U) [kJ/mol]:',T60,F26.8)") totene*kjmol
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Zero-point correction [kJ/mol]:',T60,F26.8)") zpe/1000.0_dp
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Entropy [kJ/(mol K)]:',T60,F26.8)") entropy/1000.0_dp
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Enthalpy correction (H-U) [kJ/mol]:',T60,F26.8)") rotvibtra/1000.0_dp
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Gibbs energy correction [kJ/mol]:',T60,F26.8)") Gibbs/1000.0_dp
         WRITE (UNIT=iw, FMT="(T2,'VIB|', T20, 'Heat capacity [kJ/(mol*K)]:',T70,F16.8)") heat_capacity/1000.0_dp
         WRITE (UNIT=iw, FMT="(/)")
      ENDIF

   END SUBROUTINE get_thch_values
END MODULE vibrational_analysis

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optimize_basis_types

   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE powell,                          ONLY: opt_state_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optimize_basis_types'

   PUBLIC :: basis_optimization_type, subset_type, flex_basis_type, &
             derived_basis_info, deallocate_basis_optimization_type

   ! constraint information for a single constraing. boundary is translateed into a fermi function
   ! like setting as for variational limited case
   TYPE exp_constraint_type
      INTEGER                                    :: const_type
      REAL(KIND=dp)                              :: llim, ulim
      REAL(KIND=dp)                              :: init, var_fac
   END TYPE

   ! Subset of a basis+ additional information on what to optimize.
   ! *_x_ind maps to the index in the optimization vector
   !  opt_* logical whether quantity ahould be optimized
   !  *_const information for exponents used to constrain them
   TYPE subset_type
      INTEGER                                    :: lmin, lmax, nexp
      INTEGER                                    :: n, ncon_tot, nl
      INTEGER, DIMENSION(:), ALLOCATABLE           :: l
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE   :: coeff
      LOGICAL, DIMENSION(:, :), ALLOCATABLE         :: opt_coeff
      INTEGER, DIMENSION(:, :), ALLOCATABLE         :: coeff_x_ind
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE     :: exps
      LOGICAL, DIMENSION(:), ALLOCATABLE           :: opt_exps
      INTEGER, DIMENSION(:), ALLOCATABLE           :: exp_x_ind
      LOGICAL, DIMENSION(:), ALLOCATABLE           :: exp_has_const
      TYPE(exp_constraint_type), DIMENSION(:), ALLOCATABLE :: exp_const
   END TYPE

   ! Top level information for basis sets+ vector subset with the real information
   TYPE flex_basis_type
      CHARACTER(LEN=default_string_length)       :: basis_name
      INTEGER                                    :: nopt
      INTEGER                                    :: nsets
      TYPE(subset_type), DIMENSION(:), ALLOCATABLE :: subset
   END TYPE

   ! information for optimization: whether coeff has to be optimized or not
   TYPE use_contr_type
      LOGICAL, DIMENSION(:), ALLOCATABLE           :: in_use
   END TYPE

   ! information about how to generate the derived basis sets
   TYPE derived_basis_info
      INTEGER                                    :: reference_set
      INTEGER, DIMENSION(:, :), ALLOCATABLE         :: remove_contr
      INTEGER                                    :: nsets, ncontr
      INTEGER, DIMENSION(:), ALLOCATABLE           :: remove_set
      LOGICAL, DIMENSION(:), ALLOCATABLE           :: in_use_set
      TYPE(use_contr_type), DIMENSION(:), ALLOCATABLE :: use_contr
   END TYPE

   ! some usual stuff for basis information and an info type containing the
   ! the translated input on how to genrate the derived basis sets
   ! a flexible basis type for every derived basis
   ! ATTENTION: both vectors go from 0:nbasis_deriv. entry 0 is the one specified
   !            in the template basis file
   TYPE kind_basis_type
      CHARACTER(LEN=default_string_length)       :: basis_name
      CHARACTER(LEN=default_string_length)       :: element
      INTEGER                                    :: nbasis_deriv
      TYPE(derived_basis_info), DIMENSION(:), ALLOCATABLE :: deriv_info
      TYPE(flex_basis_type), DIMENSION(:), ALLOCATABLE :: flex_basis
   END TYPE

   ! vector of length nparallel_groups containing the id's of the calculations in the group
   TYPE comp_group_type
      INTEGER, DIMENSION(:), ALLOCATABLE           :: member_list
   END TYPE

! **************************************************************************************************
!> \brief type containing all information needed for basis matching
!> \author Florian Schiffmann
! **************************************************************************************************
   TYPE basis_optimization_type
      TYPE(comp_group_type), DIMENSION(:), ALLOCATABLE  :: comp_group
      INTEGER :: ntraining_sets
      INTEGER :: ncombinations
      LOGICAL :: use_condition_number
      INTEGER, DIMENSION(:), POINTER   :: group_partition
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: combination
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: fval_weight
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: condition_weight
      INTEGER :: nkind
      INTEGER :: write_frequency
      INTEGER :: nbasis_deriv_types
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: x_opt
      TYPE(opt_state_type)  :: powell_param
      CHARACTER(LEN=default_path_length), DIMENSION(:), ALLOCATABLE :: training_input
      CHARACTER(LEN=default_path_length), DIMENSION(:), ALLOCATABLE :: training_dir
      CHARACTER(LEN=default_path_length) :: work_basis_file
      CHARACTER(LEN=default_path_length) :: output_basis_file
      CHARACTER(LEN=default_path_length) :: template_basis_file
      TYPE(kind_basis_type), DIMENSION(:), ALLOCATABLE :: kind_basis
      INTEGER :: opt_id
   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief Deallocate everything which was allocated before.
!>        Note not all arrays are used depending on the type of basis
!>        i.e derived or reference basis set
!> \param opt_bas ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE deallocate_basis_optimization_type(opt_bas)
      TYPE(basis_optimization_type)                      :: opt_bas

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_basis_optimization_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: igroup, ikind

      IF (ASSOCIATED(opt_bas%group_partition)) DEALLOCATE (opt_bas%group_partition)
      IF (ALLOCATED(opt_bas%combination)) DEALLOCATE (opt_bas%combination)
      IF (ALLOCATED(opt_bas%x_opt)) DEALLOCATE (opt_bas%x_opt)
      IF (ALLOCATED(opt_bas%training_input)) DEALLOCATE (opt_bas%training_input)
      IF (ALLOCATED(opt_bas%training_dir)) DEALLOCATE (opt_bas%training_dir)
      IF (ALLOCATED(opt_bas%fval_weight)) DEALLOCATE (opt_bas%fval_weight)
      IF (ALLOCATED(opt_bas%condition_weight)) DEALLOCATE (opt_bas%condition_weight)

      IF (ALLOCATED(opt_bas%comp_group)) THEN
         DO igroup = 1, SIZE(opt_bas%comp_group)
            IF (ALLOCATED(opt_bas%comp_group(igroup)%member_list)) DEALLOCATE (opt_bas%comp_group(igroup)%member_list)
         END DO
         DEALLOCATE (opt_bas%comp_group)
      END IF

      IF (ALLOCATED(opt_bas%kind_basis)) THEN
         DO ikind = 1, SIZE(opt_bas%kind_basis)
            CALL deallocate_kind_basis(opt_bas%kind_basis(ikind))
         END DO
         DEALLOCATE (opt_bas%kind_basis)
      END IF

   END SUBROUTINE deallocate_basis_optimization_type

! **************************************************************************************************
!> \brief Some more deallocation of the subtypes of optimize_absis type
!> \param kind ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE deallocate_kind_basis(kind)
      TYPE(kind_basis_type)                              :: kind

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_kind_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ibasis, icont, iinfo, iset

      IF (ALLOCATED(kind%deriv_info)) THEN
         DO iinfo = 0, SIZE(kind%deriv_info)-1
            IF (ALLOCATED(kind%deriv_info(iinfo)%remove_contr)) DEALLOCATE (kind%deriv_info(iinfo)%remove_contr)
            IF (ALLOCATED(kind%deriv_info(iinfo)%remove_set)) DEALLOCATE (kind%deriv_info(iinfo)%remove_set)
            IF (ALLOCATED(kind%deriv_info(iinfo)%in_use_set)) DEALLOCATE (kind%deriv_info(iinfo)%in_use_set)
            IF (ALLOCATED(kind%deriv_info(iinfo)%use_contr)) THEN
               DO icont = 1, SIZE(kind%deriv_info(iinfo)%use_contr)
                  IF (ALLOCATED(kind%deriv_info(iinfo)%use_contr(icont)%in_use)) &
                     DEALLOCATE (kind%deriv_info(iinfo)%use_contr(icont)%in_use)
               END DO
               DEALLOCATE (kind%deriv_info(iinfo)%use_contr)
            END IF
         END DO
         DEALLOCATE (kind%deriv_info)
      END IF

      IF (ALLOCATED(kind%flex_basis)) THEN
         DO ibasis = 0, SIZE(kind%flex_basis)-1
            IF (ALLOCATED(kind%flex_basis(ibasis)%subset)) THEN
               DO iset = 1, SIZE(kind%flex_basis(ibasis)%subset)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%l)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%l)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%coeff)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%coeff)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%opt_coeff)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%opt_coeff)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%coeff_x_ind)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%coeff_x_ind)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%exps)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%exps)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%opt_exps)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%opt_exps)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%exp_x_ind)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%exp_x_ind)
                  IF (ALLOCATED(kind%flex_basis(ibasis)%subset(iset)%exp_const)) &
                     DEALLOCATE (kind%flex_basis(ibasis)%subset(iset)%exp_const)
               END DO
               DEALLOCATE (kind%flex_basis(ibasis)%subset)
            END IF
         END DO
         DEALLOCATE (kind%flex_basis)
      END IF

   END SUBROUTINE deallocate_kind_basis

END MODULE optimize_basis_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optbas_opt_utils
   USE admm_types,                      ONLY: admm_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_basic_linalg,              ONLY: cp_fm_trace,&
                                              cp_fm_upper_to_full
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE kinds,                           ONLY: dp
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: evaluate_fval, evaluate_energy

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optbas_opt_utils'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mos ...
!> \param matrix_ks ...
!> \param S_inv_orb ...
!> \param Q ...
!> \param tmp1 ...
!> \param energy ...
! **************************************************************************************************
   SUBROUTINE evaluate_energy(mos, matrix_ks, S_inv_orb, Q, tmp1, energy)
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_fm_type), POINTER                          :: S_inv_orb, Q, tmp1
      REAL(KIND=dp)                                      :: energy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'evaluate_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, naux, nmo, norb
      REAL(KIND=dp)                                      :: tmp_energy
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, QS_inv, tmp, tmp2, work, &
                                                            work_orb

      CALL cp_fm_create(QS_inv, matrix_struct=Q%matrix_struct)
      CALL cp_fm_create(tmp, matrix_struct=Q%matrix_struct)
      CALL cp_fm_create(tmp2, matrix_struct=tmp1%matrix_struct)
      CALL cp_fm_create(work, matrix_struct=S_inv_orb%matrix_struct)
      CALL cp_fm_create(work_orb, matrix_struct=S_inv_orb%matrix_struct)
      CALL cp_fm_get_info(Q, nrow_global=naux, ncol_global=norb)
      CALL cp_gemm('N', 'N', naux, norb, norb, 1.0_dp, Q, S_inv_orb, 0.0_dp, QS_inv)
      energy = 0.0_dp
      DO ispin = 1, SIZE(matrix_ks)
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, work)
         CALL cp_fm_upper_to_full(work, work_orb)

         CALL get_mo_set(mos(ispin)%mo_set, nmo=nmo, mo_coeff=mo_coeff)
         CALL cp_gemm('N', 'N', naux, norb, norb, 1.0_dp, QS_inv, work, 0.0_dp, tmp)
         CALL cp_gemm('N', 'T', naux, naux, norb, 1.0_dp, tmp, QS_inv, 0.0_dp, tmp1)
         CALL cp_gemm('N', 'T', naux, naux, nmo, 1.0_dp, mo_coeff, mo_coeff, 0.0_dp, tmp2)
         CALL cp_fm_trace(tmp1, tmp2, tmp_energy)
         energy = energy+tmp_energy*(3.0_dp-REAL(SIZE(matrix_ks), dp))

      END DO

      CALL cp_fm_release(work_orb)
      CALL cp_fm_release(QS_inv)
      CALL cp_fm_release(tmp)
      CALL cp_fm_release(tmp2)
      CALL cp_fm_release(work)
      CALL cp_fm_release(work_orb)

   END SUBROUTINE evaluate_energy

! **************************************************************************************************
!> \brief ...
!> \param mos ...
!> \param mos_aux_fit ...
!> \param Q ...
!> \param Snew ...
!> \param admm_env ...
!> \param fval ...
!> \param S_cond_number ...
! **************************************************************************************************
   SUBROUTINE evaluate_fval(mos, mos_aux_fit, Q, Snew, admm_env, fval, S_cond_number)
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(cp_dbcsr_type), POINTER                       :: Q, Snew
      TYPE(admm_type), POINTER                           :: admm_env
      REAL(KIND=dp)                                      :: fval, S_cond_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'evaluate_fval', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nao_aux_fit, nao_orb, nmo, nspins
      REAL(KIND=dp)                                      :: trace
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_aux_fit

      nao_aux_fit = admm_env%nao_aux_fit
      nao_orb = admm_env%nao_orb
      nspins = SIZE(mos)

      CALL copy_dbcsr_to_fm(Q, admm_env%Q)
      fval = 0.0_dp
      DO ispin = 1, nspins
         nmo = admm_env%nmo(ispin)
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL get_mo_set(mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit)

         CALL cp_gemm('N', 'N', nao_aux_fit, nmo, nao_orb, -2.0_dp, admm_env%Q, mo_coeff, &
                      0.0_dp, admm_env%work_aux_nmo(ispin)%matrix)
         CALL cp_fm_trace(mo_coeff_aux_fit, admm_env%work_aux_nmo(ispin)%matrix, trace)
         fval = fval+trace+2.0_dp*nmo
      END DO

      ALLOCATE (eigenvalues(nao_aux_fit))
      CALL copy_dbcsr_to_fm(Snew, admm_env%work_aux_aux)
      CALL cp_fm_syevd(admm_env%work_aux_aux, admm_env%work_aux_aux2, eigenvalues)
      S_cond_number = MAXVAL(ABS(eigenvalues))/MAX(MINVAL(ABS(eigenvalues)), EPSILON(0.0_dp))
      DEALLOCATE (eigenvalues)

   END SUBROUTINE evaluate_fval

END MODULE optbas_opt_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optbas_fenv_manipulation
   USE admm_types,                      ONLY: admm_env_create,&
                                              admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_container_types,       ONLY: get_basis_from_container
   USE basis_set_types,                 ONLY: gto_basis_set_type,&
                                              init_orb_basis_set
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_info,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_basic_linalg,              ONLY: cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_invert
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_output_handling,              ONLY: debug_print_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_admm_basis_projection,&
                                              do_admm_purify_mo_diag
   USE input_section_types,             ONLY: section_vals_add_values,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length
   USE optimize_basis_types,            ONLY: basis_optimization_type,&
                                              flex_basis_type
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_utils,                 ONLY: qs_energies_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_interactions,                 ONLY: init_interaction_radii
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_did_change,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_mo_io,                        ONLY: read_mo_set
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_types,                     ONLY: init_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE scf_control_types,               ONLY: scf_control_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: modify_input_settings, &
             allocate_mo_sets, &
             calculate_overlap, &
             create_opt_admm_env, &
             update_basis_set, &
             calculate_ks_matrix, &
             calculate_overlap_inverse

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optbas_fenv_manipulation'

CONTAINS

! **************************************************************************************************
!> \brief change settings in the training input files to initialize
!>        all needed structures and adjust settings to basis optimization
!> \param basis_optimization ...
!> \param bas_id ...
!> \param input_file ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE modify_input_settings(basis_optimization, bas_id, input_file)
      TYPE(basis_optimization_type)                      :: basis_optimization
      INTEGER                                            :: bas_id
      TYPE(section_vals_type), POINTER                   :: input_file

      CHARACTER(len=*), PARAMETER :: routineN = 'modify_input_settings', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atom, orig_basis
      INTEGER                                            :: ibasis, ikind, jkind, nbasis, nkind
      TYPE(section_vals_type), POINTER                   :: admm_section, dft_section, &
                                                            feval_section, kind_section, &
                                                            subsys_section

      feval_section => section_vals_get_subs_vals(input_file, "FORCE_EVAL")
      dft_section => section_vals_get_subs_vals(feval_section, "DFT")
      admm_section => section_vals_get_subs_vals(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD")
      subsys_section => section_vals_get_subs_vals(feval_section, "SUBSYS")
      kind_section => section_vals_get_subs_vals(subsys_section, "KIND")

      CALL section_vals_val_set(feval_section, "PRINT%DISTRIBUTION%_SECTION_PARAMETERS_", &
                                i_val=debug_print_level)
      CALL section_vals_val_set(dft_section, "SCF%PRINT%TOTAL_DENSITIES%_SECTION_PARAMETERS_", &
                                i_val=debug_print_level)
      CALL section_vals_val_set(dft_section, "SCF%PRINT%DETAILED_ENERGY%_SECTION_PARAMETERS_", &
                                i_val=debug_print_level)
      CALL section_vals_add_values(admm_section)
      CALL section_vals_val_set(admm_section, "ADMM_PURIFICATION_METHOD", &
                                i_val=do_admm_purify_mo_diag)
      CALL section_vals_val_set(admm_section, "METHOD", i_val=do_admm_basis_projection)

      ! add the new basis file containing the templates to the basis file list
      CALL section_vals_val_get(dft_section, "BASIS_SET_FILE_NAME", n_rep_val=nbasis)
      CALL section_vals_val_set(dft_section, "BASIS_SET_FILE_NAME", i_rep_val=nbasis+1, &
                                c_val=basis_optimization%work_basis_file)

      ! Set the auxilarry basis in the kind sections
      CALL section_vals_get(kind_section, n_repetition=nkind)
      DO ikind = 1, nkind
         CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                   c_val=atom, i_rep_section=ikind)
         CALL uppercase(atom)
         CALL section_vals_val_get(kind_section, "BASIS_SET", &
                                   c_val=orig_basis, i_rep_section=ikind)
         CALL section_vals_val_set(kind_section, "AUX_FIT_BASIS_SET", &
                                   c_val=orig_basis, i_rep_section=ikind)

         DO jkind = 1, basis_optimization%nkind
            IF (atom == basis_optimization%kind_basis(jkind)%element) THEN

               ibasis = basis_optimization%combination(bas_id, jkind)
               CALL section_vals_val_set(kind_section, "AUX_FIT_BASIS_SET", &
                                         c_val=TRIM(ADJUSTL(basis_optimization%kind_basis(jkind)%flex_basis(ibasis)%basis_name)), &
                                         i_rep_section=ikind)
               EXIT
            END IF
         END DO
      END DO

   END SUBROUTINE modify_input_settings

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE allocate_mo_sets(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_mo_sets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      LOGICAL                                            :: natom_mismatch
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools, ao_mo_fm_pools_aux_fit
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      NULLIFY (para_env)
      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mos=mos, nelectron_spin=nelectron_spin, &
                      mos_aux_fit=mos_aux_fit, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      scf_control=scf_control, &
                      particle_set=particle_set, &
                      admm_env=admm_env, &
                      para_env=para_env)
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")

      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools)
      DO ispin = 1, dft_control%nspins
         IF (.NOT. ASSOCIATED(mos(ispin)%mo_set%mo_coeff)) THEN
            CALL init_mo_set(mos(ispin)%mo_set, &
                             fm_pool=ao_mo_fm_pools(ispin)%pool, &
                             name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                             "%mo"//TRIM(ADJUSTL(cp_to_string(ispin))))
         END IF
      END DO

      CALL read_mo_set(mos, atomic_kind_set, qs_kind_set, particle_set, para_env, &
                       id_nr=0, multiplicity=dft_control%multiplicity, dft_section=dft_section, &
                       natom_mismatch=natom_mismatch)

      CALL mpools_get(qs_env%mpools_aux_fit, ao_mo_fm_pools=ao_mo_fm_pools_aux_fit)
      DO ispin = 1, dft_control%nspins
         IF (.NOT. ASSOCIATED(mos_aux_fit(ispin)%mo_set%mo_coeff)) THEN
            CALL init_mo_set(mos_aux_fit(ispin)%mo_set, &
                             fm_pool=ao_mo_fm_pools_aux_fit(ispin)%pool, &
                             name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                             "%mo_aux_fit"//TRIM(ADJUSTL(cp_to_string(ispin))))
         END IF
      END DO

   END SUBROUTINE allocate_mo_sets

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE calculate_ks_matrix(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_ks_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (rho, dft_control, rho_ao)

      CALL qs_energies_init(qs_env, .FALSE.)
      CALL get_qs_env(qs_env, rho=rho, dft_control=dft_control)
      CALL qs_rho_get(rho, rho_ao=rho_ao)
      DO ispin = 1, dft_control%nspins
         CALL calculate_density_matrix(qs_env%mos(ispin)%mo_set, rho_ao(ispin)%matrix)
      END DO
      CALL qs_rho_update_rho(rho, qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      qs_env%requires_mo_derivs = .FALSE.
      dft_control%do_admm = .FALSE.
      dft_control%do_admm_mo = .FALSE.
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE.)
      dft_control%do_admm = .TRUE.
      dft_control%do_admm_mo = .TRUE.

   END SUBROUTINE calculate_ks_matrix

! **************************************************************************************************
!> \brief ...
!> \param matrix_s ...
!> \param matrix_s_inv ...
!> \param para_env ...
!> \param context ...
! **************************************************************************************************
   SUBROUTINE calculate_overlap_inverse(matrix_s, matrix_s_inv, para_env, context)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s
      TYPE(cp_fm_type), POINTER                          :: matrix_s_inv
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: context

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_overlap_inverse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nao
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: work1

      CALL cp_dbcsr_get_info(matrix_s, nfullrows_total=nao)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nao, &
                               para_env=para_env, context=context)

      CALL cp_fm_create(matrix_s_inv, matrix_struct=fm_struct_tmp)
      CALL cp_fm_create(work1, matrix_struct=fm_struct_tmp)
      CALL copy_dbcsr_to_fm(matrix_s, matrix_s_inv)
      CALL cp_fm_upper_to_full(matrix_s_inv, work1)
      CALL cp_fm_cholesky_decompose(matrix_s_inv)
      CALL cp_fm_cholesky_invert(matrix_s_inv)
      CALL cp_fm_upper_to_full(matrix_s_inv, work1)
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_release(work1)

   END SUBROUTINE calculate_overlap_inverse

! **************************************************************************************************
!> \brief ...
!> \param ks_env ...
!> \param S_type ...
!> \param qs_input ...
! **************************************************************************************************
   SUBROUTINE calculate_overlap(ks_env, S_type, qs_input)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      CHARACTER(LEN=*)                                   :: S_type
      TYPE(section_vals_type), POINTER                   :: qs_input

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_overlap', &
         routineP = moduleN//':'//routineN

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, matrix_s_aux_fit, &
                                                            matrix_s_aux_fit_vs_orb
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_aux_fit, sab_aux_fit_vs_orb, sab_orb

      IF (S_type == "S_AB") THEN
         CALL get_ks_env(ks_env, &
                         matrix_s=matrix_s, &
                         sab_orb=sab_orb)
         CALL build_overlap_matrix(ks_env, matrix_s=matrix_s, &
                                   matrix_name="OVERLAP", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, &
                                   qs_input=qs_input)
         CALL set_ks_env(ks_env, matrix_s=matrix_s)
      ELSE IF (S_type == "S_AB_AUX") THEN
         CALL get_ks_env(ks_env, &
                         matrix_s_aux_fit=matrix_s_aux_fit, &
                         matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                         sab_aux_fit=sab_aux_fit, &
                         sab_aux_fit_vs_orb=sab_aux_fit_vs_orb)
         CALL build_overlap_matrix(ks_env, matrix_s=matrix_s_aux_fit, &
                                   matrix_name="AUX_FIT_OVERLAP", &
                                   basis_type_a="AUX_FIT", &
                                   basis_type_b="AUX_FIT", &
                                   sab_nl=sab_aux_fit, &
                                   qs_input=qs_input)
         CALL build_overlap_matrix(ks_env, matrix_s=matrix_s_aux_fit_vs_orb, &
                                   matrix_name="MIXED_OVERLAP", &
                                   basis_type_a="AUX_FIT", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_aux_fit_vs_orb, &
                                   qs_input=qs_input)
         CALL set_ks_env(ks_env, matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                         matrix_s_aux_fit=matrix_s_aux_fit)
      END IF

   END SUBROUTINE calculate_overlap

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE create_opt_admm_env(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_opt_admm_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: natoms
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      NULLIFY (admm_env, dft_control, mos, mos_aux_fit, particle_set, para_env)
      CALL get_qs_env(qs_env, &
                      admm_env=admm_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      mos_aux_fit=mos_aux_fit, &
                      particle_set=particle_set, &
                      para_env=para_env)
      natoms = SIZE(particle_set, 1)
      CALL admm_env_create(qs_env%admm_env, dft_control%admm_control, mos, mos_aux_fit, &
                           para_env, natoms)

   END SUBROUTINE create_opt_admm_env

! **************************************************************************************************
!> \brief ...
!> \param opt_bas ...
!> \param bas_id ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE update_basis_set(opt_bas, bas_id, qs_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      INTEGER                                            :: bas_id
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'update_basis_set', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: elem
      INTEGER                                            :: ibasis, ikind, jkind
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set)
      DO ikind = 1, SIZE(qs_kind_set)
         DO jkind = 1, opt_bas%nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=elem)
            CALL uppercase(elem)
            IF (elem == opt_bas%kind_basis(jkind)%element) THEN
               ibasis = opt_bas%combination(bas_id, jkind)
               CALL get_basis_from_container(qs_kind_set(ikind)%basis_sets, basis_set=gto_basis, &
                                             basis_type="AUX_FIT")
               CALL transfer_data_to_gto(gto_basis, opt_bas%kind_basis(jkind)%flex_basis(ibasis))
               CALL init_orb_basis_set(gto_basis)
            END IF
         END DO
      END DO

      CALL init_interaction_radii(dft_control%qs_control, atomic_kind_set, qs_kind_set)

   END SUBROUTINE update_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE transfer_data_to_gto(gto_basis, basis)
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis
      TYPE(flex_basis_type)                              :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'transfer_data_to_gto', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, ishell

      DO iset = 1, basis%nsets
         DO ishell = 1, basis%subset(iset)%ncon_tot
            DO ipgf = 1, basis%subset(iset)%nexp
               gto_basis%gcc(ipgf, ishell, iset) = basis%subset(iset)%coeff(ipgf, ishell)
            END DO
         END DO
         DO ipgf = 1, basis%subset(iset)%nexp
            gto_basis%zet(ipgf, iset) = basis%subset(iset)%exps(ipgf)
         END DO
      END DO

   END SUBROUTINE transfer_data_to_gto

END MODULE optbas_fenv_manipulation

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optimize_basis_utils
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_constants,                 ONLY: do_opt_all,&
                                              do_opt_coeff,&
                                              do_opt_exps,&
                                              do_opt_none
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: default_output_unit,&
                                              m_getcwd
   USE optimize_basis_types,            ONLY: basis_optimization_type,&
                                              derived_basis_info,&
                                              flex_basis_type,&
                                              subset_type
   USE powell,                          ONLY: opt_state_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optimize_basis_utils'

   PUBLIC :: optimize_basis_init_read_input, get_set_and_basis_id, &
             update_derived_basis_sets

CONTAINS

! **************************************************************************************************
!> \brief initialize all parts of the optimization type and read input settings
!> \param opt_bas ...
!> \param root_section ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE optimize_basis_init_read_input(opt_bas, root_section, para_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'optimize_basis_init_read_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: main_dir
      INTEGER                                            :: iset, iweight, nrep
      TYPE(section_vals_type), POINTER                   :: kind_section, optbas_section, &
                                                            powell_section, train_section

      optbas_section => section_vals_get_subs_vals(root_section, "OPTIMIZE_BASIS")
      powell_section => section_vals_get_subs_vals(optbas_section, "OPTIMIZATION")
      train_section => section_vals_get_subs_vals(optbas_section, "TRAINING_FILES")
      kind_section => section_vals_get_subs_vals(optbas_section, "FIT_KIND")

      CALL section_vals_val_get(optbas_section, "BASIS_TEMPLATE_FILE", c_val=opt_bas%template_basis_file)
      CALL section_vals_val_get(optbas_section, "BASIS_WORK_FILE", c_val=opt_bas%work_basis_file)
      CALL section_vals_val_get(optbas_section, "BASIS_OUTPUT_FILE", c_val=opt_bas%output_basis_file)
      CALL m_getcwd(main_dir)
      opt_bas%work_basis_file = TRIM(ADJUSTL(main_dir))//"/"//TRIM(ADJUSTL(opt_bas%work_basis_file))

      CALL section_vals_val_get(optbas_section, "WRITE_FREQUENCY", i_val=opt_bas%write_frequency)
      CALL section_vals_val_get(optbas_section, "USE_CONDITION_NUMBER", l_val=opt_bas%use_condition_number)

      CALL generate_initial_basis(kind_section, opt_bas, para_env)

      CALL section_vals_get(train_section, n_repetition=opt_bas%ntraining_sets)
      IF (opt_bas%ntraining_sets == 0) &
         CALL cp__b("optimize_basis_utils.F",88,"No training set was specified in the Input")

      ALLOCATE (opt_bas%training_input(opt_bas%ntraining_sets))
      ALLOCATE (opt_bas%training_dir(opt_bas%ntraining_sets))
      DO iset = 1, opt_bas%ntraining_sets
         CALL section_vals_val_get(train_section, "DIRECTORY", c_val=opt_bas%training_dir(iset), &
                                   i_rep_section=iset)
         CALL section_vals_val_get(train_section, "INPUT_FILE_NAME", c_val=opt_bas%training_input(iset), &
                                   i_rep_section=iset)
      END DO

      CALL init_powell_var(opt_bas%powell_param, powell_section)
      opt_bas%powell_param%nvar = SIZE(opt_bas%x_opt)

      CALL generate_derived_basis_sets(opt_bas, para_env)

      CALL generate_basis_combinations(opt_bas, optbas_section)

      CALL section_vals_val_get(optbas_section, "RESIDUUM_WEIGHT", n_rep_val=nrep)
      ALLOCATE (opt_bas%fval_weight(0:opt_bas%ncombinations))
      opt_bas%fval_weight = 1.0_dp
      DO iweight = 1, nrep
         CALL section_vals_val_get(optbas_section, "RESIDUUM_WEIGHT", r_val=opt_bas%fval_weight(iweight-1), &
                                   i_rep_val=iweight)
      END DO

      CALL section_vals_val_get(optbas_section, "CONDITION_WEIGHT", n_rep_val=nrep)
      ALLOCATE (opt_bas%condition_weight(0:opt_bas%ncombinations))
      opt_bas%condition_weight = 1.0_dp
      DO iweight = 1, nrep
         CALL section_vals_val_get(optbas_section, "CONDITION_WEIGHT", r_val=opt_bas%condition_weight(iweight-1), &
                                   i_rep_val=iweight)
      END DO

      CALL generate_computation_groups(opt_bas, optbas_section, para_env)

      CALL print_opt_info(opt_bas)

   END SUBROUTINE optimize_basis_init_read_input

! **************************************************************************************************
!> \brief ...
!> \param opt_bas ...
! **************************************************************************************************
   SUBROUTINE print_opt_info(opt_bas)
      TYPE(basis_optimization_type)                      :: opt_bas

      CHARACTER(len=*), PARAMETER :: routineN = 'print_opt_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: icomb, ikind, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      unit_nr = -1
      IF (logger%para_env%ionode) &
         unit_nr = cp_logger_get_default_unit_nr(logger)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(1X,A,A)') "BASOPT| Total number of calculations ", &
            TRIM(cp_to_string(opt_bas%ncombinations*opt_bas%ntraining_sets))
         WRITE (unit_nr, '(A)') ""
         DO icomb = 1, opt_bas%ncombinations
            WRITE (unit_nr, '(1X,A,A)') "BASOPT| Content of basis combination ", TRIM(cp_to_string(icomb))
            DO ikind = 1, opt_bas%nkind
               WRITE (unit_nr, '(1X,A,A4,4X,A,3X,A)') "BASOPT|     Element: ", TRIM(opt_bas%kind_basis(ikind)%element), &
                  "Basis set: ", TRIM(opt_bas%kind_basis(ikind)%flex_basis(opt_bas%combination(icomb, ikind))%basis_name)
            END DO
            WRITE (unit_nr, '(A)') ""
         END DO
      END IF
   END SUBROUTINE print_opt_info

! **************************************************************************************************
!> \brief Generation of the requested basis set combinations if multiple kinds
!>        are fitted at the same time (if not specified create all possible)
!> \param opt_bas ...
!> \param optbas_section ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE generate_basis_combinations(opt_bas, optbas_section)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(section_vals_type), POINTER                   :: optbas_section

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_basis_combinations', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ikind, j, n_rep
      INTEGER, DIMENSION(:), POINTER                     :: i_vals, tmp_i, tmp_i2
      LOGICAL                                            :: explicit, raise

!setup the basis combinations to optimize

      CALL section_vals_val_get(optbas_section, "BASIS_COMBINATIONS", explicit=explicit, n_rep_val=n_rep)
      IF (.NOT. explicit) THEN
         opt_bas%ncombinations = 1
         ALLOCATE (tmp_i(opt_bas%nkind))
         ALLOCATE (tmp_i2(opt_bas%nkind))
         DO ikind = 1, opt_bas%nkind
            opt_bas%ncombinations = opt_bas%ncombinations*SIZE(opt_bas%kind_basis(ikind)%flex_basis)
            tmp_i(ikind) = opt_bas%kind_basis(ikind)%nbasis_deriv
         END DO
         ALLOCATE (opt_bas%combination(opt_bas%ncombinations, opt_bas%nkind))
         tmp_i2 = 0
         DO i = 1, opt_bas%ncombinations
            DO j = 1, opt_bas%nkind
               opt_bas%combination(i, j) = tmp_i2(j)
            END DO
            tmp_i2(opt_bas%nkind) = tmp_i2(opt_bas%nkind)+1
            raise = .FALSE.
            DO j = opt_bas%nkind, 1, -1
               IF (raise) tmp_i2(j) = tmp_i2(j)+1
               IF (tmp_i2(j) .GT. tmp_i(j)) THEN
                  tmp_i2(j) = 0
                  raise = .TRUE.
               END IF
            END DO
         END DO
         DEALLOCATE (tmp_i)
         DEALLOCATE (tmp_i2)
      ELSE
         opt_bas%ncombinations = n_rep
         ALLOCATE (opt_bas%combination(opt_bas%ncombinations, opt_bas%nkind))
         DO i = 1, n_rep
            CALL section_vals_val_get(optbas_section, "BASIS_COMBINATIONS", i_vals=i_vals, i_rep_val=i)
            opt_bas%combination(i, :) = i_vals(:)
         END DO
      END IF

   END SUBROUTINE generate_basis_combinations

! **************************************************************************************************
!> \brief returns a mapping from the calculation id to the trainings set id and
!>        basis combination id
!> \param calc_id ...
!> \param opt_bas ...
!> \param set_id ...
!> \param bas_id ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE get_set_and_basis_id(calc_id, opt_bas, set_id, bas_id)

      INTEGER                                            :: calc_id
      TYPE(basis_optimization_type)                      :: opt_bas
      INTEGER                                            :: set_id, bas_id

      CHARACTER(len=*), PARAMETER :: routineN = 'get_set_and_basis_id', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncom, nset

      ncom = opt_bas%ncombinations
      nset = opt_bas%ntraining_sets

      set_id = (calc_id)/ncom+1
      bas_id = MOD(calc_id, ncom)+1

   END SUBROUTINE

! **************************************************************************************************
!> \brief Pack calculations in groups. If less MPI tasks than systems are used
!>        multiple systems will be assigned to a single MPI task
!> \param opt_bas ...
!> \param optbas_section ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE generate_computation_groups(opt_bas, optbas_section, para_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(section_vals_type), POINTER                   :: optbas_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_computation_groups', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iadd1, iadd2, icount, igroup, isize, j, &
                                                            ncalc, nproc, nptot
      INTEGER, DIMENSION(:), POINTER                     :: i_vals
      LOGICAL                                            :: explicit

      nproc = para_env%num_pe
      ncalc = opt_bas%ncombinations*opt_bas%ntraining_sets
      CALL section_vals_val_get(optbas_section, "GROUP_PARTITION", explicit=explicit)

      ! No input information available, try to equally distribute
      IF (.NOT. explicit) THEN
         IF (nproc .GE. ncalc) THEN
            iadd1 = nproc/ncalc
            iadd2 = MOD(nproc, ncalc)
            ALLOCATE (opt_bas%comp_group(ncalc))
            ALLOCATE (opt_bas%group_partition(0:ncalc-1))
            DO igroup = 0, ncalc-1
               ALLOCATE (opt_bas%comp_group(igroup+1)%member_list(1))
               opt_bas%comp_group(igroup+1)%member_list(1) = igroup
               opt_bas%group_partition(igroup) = iadd1
               IF (igroup .LT. iadd2) opt_bas%group_partition(igroup) = opt_bas%group_partition(igroup)+1
            END DO
         ELSE
            iadd1 = ncalc/nproc
            iadd2 = MOD(ncalc, nproc)
            ALLOCATE (opt_bas%comp_group(nproc))
            ALLOCATE (opt_bas%group_partition(0:nproc-1))
            icount = 0
            DO igroup = 0, nproc-1
               opt_bas%group_partition(igroup) = 1
               isize = iadd1
               IF (igroup .LT. iadd2) isize = isize+1
               ALLOCATE (opt_bas%comp_group(igroup+1)%member_list(isize))
               DO j = 1, isize
                  opt_bas%comp_group(igroup+1)%member_list(j) = icount
                  icount = icount+1
               END DO
            END DO
         END IF
      ELSE

         ! Group partition from input. see if all systems can be assigned. If not add to existing group
         CALL section_vals_val_get(optbas_section, "GROUP_PARTITION", i_vals=i_vals)
         isize = SIZE(i_vals)
         nptot = SUM(i_vals)
         IF (nptot /= nproc) &
            CALL cp_abort(cp__l("optimize_basis_utils.F",310), &
                          "Number of processors in group distribution does not match number of MPI tasks."// &
                          " Please change input.")
         IF (.NOT. isize .LE. ncalc) &
            CALL cp_abort(cp__l("optimize_basis_utils.F",314), &
                          "Number of Groups larger than number of calculations"// &
                          " Please change input.")
         IF(.NOT.(nptot == nproc))CALL cp__a("optimize_basis_utils.F",317)
         ALLOCATE (opt_bas%comp_group(isize))
         ALLOCATE (opt_bas%group_partition(0:isize-1))
         IF (isize .LT. ncalc) THEN
            iadd1 = ncalc/isize
            iadd2 = MOD(ncalc, isize)
            icount = 0
            DO igroup = 0, isize-1
               opt_bas%group_partition(igroup) = i_vals(igroup+1)
               isize = iadd1
               IF (igroup .LT. iadd2) isize = isize+1
               ALLOCATE (opt_bas%comp_group(igroup+1)%member_list(isize))
               DO j = 1, isize
                  opt_bas%comp_group(igroup+1)%member_list(j) = icount
                  icount = icount+1
               END DO
            END DO
         ELSE
            DO igroup = 0, isize-1
               opt_bas%group_partition(igroup) = i_vals(igroup+1)
               ALLOCATE (opt_bas%comp_group(igroup+1)%member_list(1))
               opt_bas%comp_group(igroup+1)%member_list(1) = igroup
            END DO
         END IF
      END IF
   END SUBROUTINE generate_computation_groups

! **************************************************************************************************
!> \brief Regenerate the basis sets from reference 0 after an update from the
!>        optimizer to reference was performed, and print basis file if required
!> \param opt_bas ...
!> \param write_it ...
!> \param output_file ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE update_derived_basis_sets(opt_bas, write_it, output_file, para_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      LOGICAL                                            :: write_it
      CHARACTER(LEN=default_path_length)                 :: output_file
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'update_derived_basis_sets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ibasis, ikind, unit_nr

      DO ikind = 1, opt_bas%nkind
         DO ibasis = 1, opt_bas%kind_basis(ikind)%nbasis_deriv
            CALL update_used_parts(opt_bas%kind_basis(ikind)%deriv_info(ibasis), &
                                   opt_bas%kind_basis(ikind)%flex_basis(0), &
                                   opt_bas%kind_basis(ikind)%flex_basis(ibasis))
         END DO
      END DO

      IF (write_it) THEN
         IF (para_env%ionode) THEN
            CALL open_file(file_name=output_file, file_status="UNKNOWN", &
                           file_action="WRITE", unit_number=unit_nr)
         ELSE
            unit_nr = -999
         END IF
         DO ikind = 1, opt_bas%nkind
            DO ibasis = 0, opt_bas%kind_basis(ikind)%nbasis_deriv
               CALL write_basis(opt_bas%kind_basis(ikind)%flex_basis(ibasis), opt_bas%kind_basis(ikind)%element, &
                                unit_nr)
            END DO
         END DO
         IF (para_env%ionode) CALL close_file(unit_number=unit_nr)
      END IF

   END SUBROUTINE update_derived_basis_sets

! **************************************************************************************************
!> \brief Update the the information in a given basis set
!> \param info_new ...
!> \param basis ...
!> \param basis_new ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE update_used_parts(info_new, basis, basis_new)
      TYPE(derived_basis_info)                           :: info_new
      TYPE(flex_basis_type)                              :: basis, basis_new

      CHARACTER(len=*), PARAMETER :: routineN = 'update_used_parts', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icont, iset, jcont, jset

      jset = 0
      DO iset = 1, basis%nsets
         IF (info_new%in_use_set(iset)) THEN
            jset = jset+1
            basis_new%subset(jset)%exps(:) = basis%subset(iset)%exps
            jcont = 0
            DO icont = 1, basis%subset(iset)%ncon_tot
               IF (info_new%use_contr(iset)%in_use(icont)) THEN
                  jcont = jcont+1
                  basis_new%subset(jset)%coeff(:, jcont) = basis%subset(iset)%coeff(:, icont)
               END IF
            END DO
         END IF
      END DO

   END SUBROUTINE update_used_parts

! **************************************************************************************************
!> \brief Initial generation of the basis set from the file and DERIVED_SET
!> \param opt_bas ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE generate_derived_basis_sets(opt_bas, para_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_derived_basis_sets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ibasis, ikind, iref, jbasis, unit_nr

      DO ikind = 1, opt_bas%nkind
         CALL init_deriv_info_ref(opt_bas%kind_basis(ikind)%deriv_info(0), opt_bas%kind_basis(ikind)%flex_basis(0))
         ! initialize the reference set used as template for the rest
         DO ibasis = 1, opt_bas%kind_basis(ikind)%nbasis_deriv
            iref = opt_bas%kind_basis(ikind)%deriv_info(ibasis)%reference_set
            DO jbasis = 0, opt_bas%kind_basis(ikind)%nbasis_deriv
               IF (iref == jbasis) CALL setup_used_parts_init_basis(opt_bas%kind_basis(ikind)%deriv_info(ibasis), &
                                                                    opt_bas%kind_basis(ikind)%deriv_info(iref), &
                                                                    opt_bas%kind_basis(ikind)%flex_basis(0), &
                                                                    opt_bas%kind_basis(ikind)%flex_basis(ibasis))
            END DO
         END DO
      END DO

      IF (para_env%ionode) THEN
         CALL open_file(file_name=opt_bas%work_basis_file, file_status="UNKNOWN", &
                        file_action="WRITE", unit_number=unit_nr)
      ELSE
         unit_nr = -999
      END IF
      DO ikind = 1, opt_bas%nkind
         DO ibasis = 0, opt_bas%kind_basis(ikind)%nbasis_deriv
            opt_bas%kind_basis(ikind)%flex_basis(ibasis)%basis_name = &
               TRIM(ADJUSTL(opt_bas%kind_basis(ikind)%basis_name))//"-DERIVED_SET-"//ADJUSTL(cp_to_string(ibasis))
            CALL write_basis(opt_bas%kind_basis(ikind)%flex_basis(ibasis), opt_bas%kind_basis(ikind)%element, &
                             unit_nr)
         END DO
      END DO
      IF (para_env%ionode) CALL close_file(unit_number=unit_nr)

   END SUBROUTINE generate_derived_basis_sets

! **************************************************************************************************
!> \brief Write a basis set file which can be used from CP2K
!> \param basis ...
!> \param element ...
!> \param unit_nr ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE write_basis(basis, element, unit_nr)
      TYPE(flex_basis_type)                              :: basis
      CHARACTER(LEN=default_string_length)               :: element
      INTEGER                                            :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'write_basis', routineP = moduleN//':'//routineN

      INTEGER                                            :: iexp, iset

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(A)") TRIM(ADJUSTL(element))//" "//TRIM(ADJUSTL(basis%basis_name))
         WRITE (UNIT=unit_nr, FMT="(1X,I0)") basis%nsets
         DO iset = 1, basis%nsets
            WRITE (UNIT=unit_nr, FMT="(30(1X,I0))") basis%subset(iset)%n, basis%subset(iset)%lmin, basis%subset(iset)%lmax, &
               basis%subset(iset)%nexp, basis%subset(iset)%l
            DO iexp = 1, basis%subset(iset)%nexp
               WRITE (UNIT=unit_nr, FMT="(T2,F24.14,30(1X,ES24.14))") &
                  basis%subset(iset)%exps(iexp), basis%subset(iset)%coeff(iexp, :)
            END DO
         END DO
      END IF

   END SUBROUTINE write_basis

! **************************************************************************************************
!> \brief Initialize the derived basis sets and the vectors containing their
!>        assembly information ehich is used for regeneration of the sets.
!> \param info_new ...
!> \param info_ref ...
!> \param basis ...
!> \param basis_new ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE setup_used_parts_init_basis(info_new, info_ref, basis, basis_new)
      TYPE(derived_basis_info)                           :: info_new, info_ref
      TYPE(flex_basis_type)                              :: basis, basis_new

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_used_parts_init_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, jset, lind, nsets

! copy the reference information on the new set

      ALLOCATE (info_new%in_use_set(SIZE(info_ref%in_use_set)))
      info_new%in_use_set(:) = info_ref%in_use_set
      ALLOCATE (info_new%use_contr(SIZE(info_ref%in_use_set)))
      DO i = 1, SIZE(info_ref%in_use_set)
         ALLOCATE (info_new%use_contr(i)%in_use(SIZE(info_ref%use_contr(i)%in_use)))
         info_new%use_contr(i)%in_use(:) = info_ref%use_contr(i)%in_use
      END DO
      DO i = 1, info_new%nsets
         info_new%in_use_set(info_new%remove_set(i)) = .FALSE.
      END DO
      DO i = 1, info_new%ncontr
         lind = convert_l_contr_to_entry(basis%subset(info_new%remove_contr(i, 1))%lmin, &
                                         basis%subset(info_new%remove_contr(i, 1))%l, &
                                         info_new%remove_contr(i, 3), info_new%remove_contr(i, 2))

         info_new%use_contr(info_new%remove_contr(i, 1))%in_use(lind) = .FALSE.
      END DO

      nsets = 0
      DO i = 1, basis%nsets
         IF (info_new%in_use_set(i)) nsets = nsets+1
      END DO
      basis_new%nsets = nsets
      ALLOCATE (basis_new%subset(nsets))
      jset = 0
      DO i = 1, basis%nsets
         IF (info_new%in_use_set(i)) THEN
            jset = jset+1
            CALL create_new_subset(basis%subset(i), basis_new%subset(jset), info_new%use_contr(jset)%in_use)
         END IF
      END DO

   END SUBROUTINE setup_used_parts_init_basis

! **************************************************************************************************
!> \brief Fill the low level information of the derived basis set.
!> \param subset ...
!> \param subset_new ...
!> \param in_use ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE create_new_subset(subset, subset_new, in_use)
      TYPE(subset_type)                                  :: subset, subset_new
      LOGICAL, DIMENSION(:)                              :: in_use

      CHARACTER(len=*), PARAMETER :: routineN = 'create_new_subset', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icon, iind, il
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: tmp_l

      ALLOCATE (tmp_l(SIZE(subset%l)))
      tmp_l(:) = subset%l
      subset_new%lmin = subset%lmin
      subset_new%lmax = subset%lmin-1
      subset_new%nexp = subset%nexp
      subset_new%n = subset%n
      DO il = 1, SIZE(subset%l)
         DO icon = 1, subset%l(il)
            iind = convert_l_contr_to_entry(subset%lmin, subset%l, icon, subset%lmin+il-1)
            IF (.NOT. in_use(iind)) tmp_l(il) = tmp_l(il)-1
         END DO
         IF (tmp_l(il) .GT. 0) subset_new%lmax = subset_new%lmax+1
      END DO
      subset_new%nl = subset_new%lmax-subset_new%lmin+1
      subset_new%ncon_tot = SUM(tmp_l)
      ALLOCATE (subset_new%l(subset_new%nl))
      ALLOCATE (subset_new%coeff(subset_new%nexp, subset_new%ncon_tot))
      ALLOCATE (subset_new%exps(subset_new%nexp))
      subset_new%exps(:) = subset%exps
      DO il = 1, SIZE(subset%l)
         IF (tmp_l(il) == 0) EXIT
         subset_new%l(il) = tmp_l(il)
      END DO
      DEALLOCATE (tmp_l)
      iind = 0
      DO icon = 1, subset%ncon_tot
         IF (in_use(icon)) THEN
            iind = iind+1
            subset_new%coeff(:, iind) = subset%coeff(:, icon)
         END IF
      END DO

   END SUBROUTINE create_new_subset

! **************************************************************************************************
!> \brief for completeness generate the derived info for set 0(reference from file)
!> \param info ...
!> \param basis ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE init_deriv_info_ref(info, basis)
      TYPE(derived_basis_info)                           :: info
      TYPE(flex_basis_type)                              :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'init_deriv_info_ref', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      ALLOCATE (info%in_use_set(basis%nsets))
      info%in_use_set = .TRUE.
      ALLOCATE (info%use_contr(basis%nsets))
      DO i = 1, basis%nsets
         ALLOCATE (info%use_contr(i)%in_use(basis%subset(i)%ncon_tot))
         info%use_contr(i)%in_use = .TRUE.
      END DO
   END SUBROUTINE init_deriv_info_ref

! **************************************************************************************************
!> \brief get the general information for the basis sets. E.g. what to optimize,
!>        Basis set name, constraints upon optimization and read the reference basis
!> \param kind_section ...
!> \param opt_bas ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE generate_initial_basis(kind_section, opt_bas, para_env)
      TYPE(section_vals_type), POINTER                   :: kind_section
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_initial_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, variable_counter
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: set_section

      CALL section_vals_get(kind_section, n_repetition=opt_bas%nkind)
      ALLOCATE (opt_bas%kind_basis(opt_bas%nkind))

      ! counter to get the number of free variables in optimization
      variable_counter = 0
      DO ikind = 1, opt_bas%nkind
         CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", c_val=opt_bas%kind_basis(ikind)%element, &
                                   i_rep_section=ikind)
         CALL section_vals_val_get(kind_section, "BASIS_SET", c_val=opt_bas%kind_basis(ikind)%basis_name, &
                                   i_rep_section=ikind)
         set_section => section_vals_get_subs_vals(kind_section, "DERIVED_BASIS_SETS", &
                                                   i_rep_section=ikind)
         CALL section_vals_get(set_section, n_repetition=opt_bas%kind_basis(ikind)%nbasis_deriv, explicit=explicit)
         IF (.NOT. explicit) opt_bas%kind_basis(ikind)%nbasis_deriv = 0
         ALLOCATE (opt_bas%kind_basis(ikind)%flex_basis(0:opt_bas%kind_basis(ikind)%nbasis_deriv))
         ALLOCATE (opt_bas%kind_basis(ikind)%deriv_info(0:opt_bas%kind_basis(ikind)%nbasis_deriv))

         CALL fill_basis_template(kind_section, opt_bas%kind_basis(ikind)%flex_basis(0), opt_bas%template_basis_file, &
                                  opt_bas%kind_basis(ikind)%element, opt_bas%kind_basis(ikind)%basis_name, para_env, ikind)

         CALL setup_exp_constraints(kind_section, opt_bas%kind_basis(ikind)%flex_basis(0))

         CALL parse_derived_basis(kind_section, opt_bas%kind_basis(ikind)%deriv_info, ikind)

         variable_counter = variable_counter+opt_bas%kind_basis(ikind)%flex_basis(0)%nopt
      END DO

      ALLOCATE (opt_bas%x_opt(variable_counter))

      variable_counter = 0
      DO ikind = 1, opt_bas%nkind
         CALL assign_x_to_basis(opt_bas%x_opt, opt_bas%kind_basis(ikind)%flex_basis(0), variable_counter)
      END DO

      IF(.NOT.(variable_counter == SIZE(opt_bas%x_opt)))CALL cp__a("optimize_basis_utils.F",691)

   END SUBROUTINE generate_initial_basis

! **************************************************************************************************
!> \brief get low level information about how to construc new basis sets from reference
!> \param kind_section ...
!> \param deriv_info ...
!> \param ikind ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE parse_derived_basis(kind_section, deriv_info, ikind)
      TYPE(section_vals_type), POINTER                   :: kind_section
      TYPE(derived_basis_info), DIMENSION(:)             :: deriv_info
      INTEGER                                            :: ikind

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_derived_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_rep, iset, jset, n_rep, nsets
      INTEGER, DIMENSION(:), POINTER                     :: i_vals
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: set1_section

      nsets = SIZE(deriv_info)-1
      set1_section => section_vals_get_subs_vals(kind_section, "DERIVED_BASIS_SETS", &
                                                 i_rep_section=ikind)
      DO jset = 1, nsets
         ! stracnge but as derive info is allcated from 0 to n the count over here has to be shifted
         iset = jset+1
         CALL section_vals_val_get(set1_section, "REFERENCE_SET", i_vals=i_vals, i_rep_section=jset)
         deriv_info(iset)%reference_set = i_vals(1)
         CALL section_vals_val_get(set1_section, "REMOVE_CONTRACTION", explicit=explicit, n_rep_val=n_rep, &
                                   i_rep_section=jset)
         deriv_info(iset)%ncontr = n_rep
         IF (explicit) THEN
            ALLOCATE (deriv_info(iset)%remove_contr(n_rep, 3))
            DO i_rep = 1, n_rep
               CALL section_vals_val_get(set1_section, "REMOVE_CONTRACTION", i_rep_val=i_rep, i_vals=i_vals, &
                                         i_rep_section=jset)
               deriv_info(iset)%remove_contr(i_rep, :) = i_vals(:)
            END DO
         END IF
         CALL section_vals_val_get(set1_section, "REMOVE_SET", explicit=explicit, n_rep_val=n_rep, &
                                   i_rep_section=jset)
         deriv_info(iset)%nsets = n_rep
         IF (explicit) THEN
            ALLOCATE (deriv_info(iset)%remove_set(n_rep))
            DO i_rep = 1, n_rep
               CALL section_vals_val_get(set1_section, "REMOVE_SET", i_rep_val=i_rep, i_vals=i_vals, &
                                         i_rep_section=jset)
               deriv_info(iset)%remove_set(i_rep) = i_vals(1)
            END DO
         END IF
      END DO

   END SUBROUTINE parse_derived_basis

! **************************************************************************************************
!> \brief get low level information about constraint on exponents
!> \param kind1_section ...
!> \param flex_basis ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE setup_exp_constraints(kind1_section, flex_basis)
      TYPE(section_vals_type), POINTER                   :: kind1_section
      TYPE(flex_basis_type)                              :: flex_basis

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_exp_constraints', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, irep, iset, nrep
      INTEGER, DIMENSION(:), POINTER                     :: def_exp
      LOGICAL                                            :: is_bound, is_varlim
      TYPE(section_vals_type), POINTER                   :: const_section

      const_section => section_vals_get_subs_vals(kind1_section, "CONSTRAIN_EXPONENTS")
      CALL section_vals_get(const_section, n_repetition=nrep)
      DO irep = 1, nrep
         CALL section_vals_val_get(const_section, "USE_EXP", i_vals=def_exp, i_rep_section=irep)
         CALL section_vals_val_get(const_section, "BOUNDARIES", explicit=is_bound, i_rep_section=irep)
         CALL section_vals_val_get(const_section, "MAX_VAR_FRACTION", explicit=is_varlim, i_rep_section=irep)
         IF (is_bound .AND. is_varlim) &
            CALL cp_abort(cp__l("optimize_basis_utils.F",776), "Exponent has two constraints. "// &
                          "This is not possible at the moment. Please change input.")
         IF (.NOT. is_bound .AND. .NOT. is_varlim) &
            CALL cp_abort(cp__l("optimize_basis_utils.F",779), "Exponent is declared to be constraint but none is given"// &
                          " Please change input.")
         IF (def_exp(1) == -1) THEN
            DO iset = 1, flex_basis%nsets
               IF (def_exp(2) == -1) THEN
                  DO ipgf = 1, flex_basis%subset(iset)%nexp
                     CALL set_constraint(flex_basis, iset, ipgf, const_section, is_bound, is_varlim, irep)
                  END DO
               ELSE
                  IF (def_exp(2) .LE. flex_basis%subset(iset)%nexp) &
                     CALL cp_abort(cp__l("optimize_basis_utils.F",789), &
                                   "Exponent declared in constraint is larger than number of exponents in the set"// &
                                   " Please change input.")
                  CALL set_constraint(flex_basis, iset, def_exp(2), const_section, is_bound, is_varlim, irep)
               END IF
            END DO
         ELSE
            IF (.NOT. def_exp(1) .LE. flex_basis%nsets) &
               CALL cp_abort(cp__l("optimize_basis_utils.F",797), &
                             "Set number of constraint is larger than number of sets in the template basis set."// &
                             " Please change input.")
            IF (def_exp(2) == -1) THEN
               DO ipgf = 1, flex_basis%subset(iset)%nexp
                  CALL set_constraint(flex_basis, def_exp(1), ipgf, const_section, is_bound, is_varlim, irep)
               END DO
            ELSE
               IF (.NOT. def_exp(2) .LE. flex_basis%subset(def_exp(1))%nexp) &
                  CALL cp_abort(cp__l("optimize_basis_utils.F",806), &
                                "Exponent declared in constraint is larger than number of exponents in the set"// &
                                " Please change input.")
               CALL set_constraint(flex_basis, def_exp(1), def_exp(2), const_section, is_bound, is_varlim, irep)
            END IF
         END IF
      END DO

   END SUBROUTINE setup_exp_constraints

! **************************************************************************************************
!> \brief put the constraint information in type and process if requires
!>        BOUNDARIES constraint gets transformed into MAX_VAR_FRACTION constraint.
!> \param flex_basis ...
!> \param iset ...
!> \param ipgf ...
!> \param const_section ...
!> \param is_bound ...
!> \param is_varlim ...
!> \param irep ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE set_constraint(flex_basis, iset, ipgf, const_section, is_bound, is_varlim, irep)
      TYPE(flex_basis_type)                              :: flex_basis
      INTEGER                                            :: iset, ipgf
      TYPE(section_vals_type), POINTER                   :: const_section
      LOGICAL                                            :: is_bound, is_varlim
      INTEGER                                            :: irep

      CHARACTER(len=*), PARAMETER :: routineN = 'set_constraint', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: r_val
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_vals

      IF (flex_basis%subset(iset)%exp_has_const(ipgf)) &
         CALL cp_abort(cp__l("optimize_basis_utils.F",842), &
                       "Multiple constraints due to collision in CONSTRAIN_EXPONENTS."// &
                       " Please change input.")
      flex_basis%subset(iset)%exp_has_const(ipgf) = .TRUE.
      IF (is_bound) THEN
         flex_basis%subset(iset)%exp_const(ipgf)%const_type = 0
         CALL section_vals_val_get(const_section, "BOUNDARIES", r_vals=r_vals, i_rep_section=irep)
         flex_basis%subset(iset)%exp_const(ipgf)%llim = MINVAL(r_vals)
         flex_basis%subset(iset)%exp_const(ipgf)%ulim = MAXVAL(r_vals)
         r_val = flex_basis%subset(iset)%exps(ipgf)
         IF (flex_basis%subset(iset)%exps(ipgf) .GT. MAXVAL(r_vals) .OR. flex_basis%subset(iset)%exps(ipgf) .LT. MINVAL(r_vals)) &
            CALL cp_abort(cp__l("optimize_basis_utils.F",853), &
                          "Exponent "//cp_to_string(r_val)// &
                          " declared in constraint is out of bounds of constraint"//cp_to_string(MINVAL(r_vals))// &
                          " to"//cp_to_string(MAXVAL(r_vals))// &
                          " Please change input.")
         flex_basis%subset(iset)%exp_const(ipgf)%init = SUM(r_vals)/2.0_dp
         flex_basis%subset(iset)%exp_const(ipgf)%var_fac = MAXVAL(r_vals)/flex_basis%subset(iset)%exp_const(ipgf)%init-1.0_dp
      END IF
      IF (is_varlim) THEN
         flex_basis%subset(iset)%exp_const(ipgf)%const_type = 1
         CALL section_vals_val_get(const_section, "MAX_VAR_FRACTION", r_vals=r_vals, i_rep_section=irep)
         flex_basis%subset(iset)%exp_const(ipgf)%var_fac = r_vals(1)
         flex_basis%subset(iset)%exp_const(ipgf)%init = flex_basis%subset(iset)%exps(ipgf)
      END IF

   END SUBROUTINE set_constraint

! **************************************************************************************************
!> \brief Initialize the optimization vector with the values from the refernece sets
!> \param x ...
!> \param basis ...
!> \param x_ind ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE assign_x_to_basis(x, basis, x_ind)
      REAL(KIND=dp), DIMENSION(:)                        :: x
      TYPE(flex_basis_type)                              :: basis
      INTEGER                                            :: x_ind

      CHARACTER(len=*), PARAMETER :: routineN = 'assign_x_to_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icont, ipgf, iset

      DO iset = 1, basis%nsets
         DO ipgf = 1, basis%subset(iset)%nexp
            IF (basis%subset(iset)%opt_exps(ipgf)) THEN
               x_ind = x_ind+1
               basis%subset(iset)%exp_x_ind(ipgf) = x_ind
               x(x_ind) = basis%subset(iset)%exps(ipgf)
            END IF
            DO icont = 1, basis%subset(iset)%ncon_tot
               IF (basis%subset(iset)%opt_coeff(ipgf, icont)) THEN
                  x_ind = x_ind+1
                  basis%subset(iset)%coeff_x_ind(ipgf, icont) = x_ind
                  x(x_ind) = basis%subset(iset)%coeff(ipgf, icont)
               END IF
            END DO
         END DO
      END DO

   END SUBROUTINE assign_x_to_basis

! **************************************************************************************************
!> \brief Fill the reference set and get the free varialbles from input
!> \param kind1_section ...
!> \param flex_basis ...
!> \param template_basis_file ...
!> \param element ...
!> \param basis_name ...
!> \param para_env ...
!> \param ikind ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE fill_basis_template(kind1_section, flex_basis, template_basis_file, element, basis_name, para_env, ikind)
      TYPE(section_vals_type), POINTER                   :: kind1_section
      TYPE(flex_basis_type)                              :: flex_basis
      CHARACTER(LEN=default_path_length)                 :: template_basis_file
      CHARACTER(LEN=default_string_length)               :: element, basis_name
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: ikind

      CHARACTER(len=*), PARAMETER :: routineN = 'fill_basis_template', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icont, idof, ipgf, irep, iset, nrep
      INTEGER, DIMENSION(:), POINTER                     :: switch

      CALL parse_basis(flex_basis, template_basis_file, element, basis_name, para_env)

      ! get the optimizable parameters. Many way to modify them but in the end only logical matrix
      ! is either set or values get flipped according to the input
      CALL section_vals_val_get(kind1_section, "INITIAL_DEGREES_OF_FREEDOM", i_val=idof, &
                                i_rep_section=ikind)
      DO iset = 1, flex_basis%nsets
         SELECT CASE (idof)
         CASE (do_opt_none)
            ! initialization in parse subset did the job
         CASE (do_opt_all)
            flex_basis%subset(iset)%opt_coeff = .TRUE.
            flex_basis%subset(iset)%opt_exps = .TRUE.
         CASE (do_opt_coeff)
            flex_basis%subset(iset)%opt_coeff = .TRUE.
         CASE (do_opt_exps)
            flex_basis%subset(iset)%opt_exps = .TRUE.
         CASE DEFAULT
            CALL cp__b("optimize_basis_utils.F",951,"No initialization available?????")
         END SELECT
      END DO

      CALL section_vals_val_get(kind1_section, "SWITCH_CONTRACTION_STATE", n_rep_val=nrep, i_rep_section=ikind)
      DO irep = 1, nrep
         CALL section_vals_val_get(kind1_section, "SWITCH_CONTRACTION_STATE", i_rep_val=irep, &
                                   i_rep_section=ikind, i_vals=switch)
         icont = convert_l_contr_to_entry(flex_basis%subset(switch(1))%lmin, flex_basis%subset(switch(1))%l, switch(3), switch(2))
         DO ipgf = 1, flex_basis%subset(switch(1))%nexp
            flex_basis%subset(switch(1))%opt_coeff(ipgf, icont) = .NOT. flex_basis%subset(switch(1))%opt_coeff(ipgf, icont)
         END DO
      END DO

      CALL section_vals_val_get(kind1_section, "SWITCH_COEFF_STATE", n_rep_val=nrep, i_rep_section=ikind)
      DO irep = 1, nrep
         CALL section_vals_val_get(kind1_section, "SWITCH_COEFF_STATE", i_rep_val=irep, &
                                   i_rep_section=ikind, i_vals=switch)
         icont = convert_l_contr_to_entry(flex_basis%subset(switch(1))%lmin, flex_basis%subset(switch(1))%l, switch(3), switch(2))
         flex_basis%subset(switch(1))%opt_coeff(switch(4), icont) = &
            .NOT. flex_basis%subset(switch(1))%opt_coeff(switch(4), icont)
      END DO

      CALL section_vals_val_get(kind1_section, "SWITCH_EXP_STATE", n_rep_val=nrep, i_rep_section=ikind)
      DO irep = 1, nrep
         CALL section_vals_val_get(kind1_section, "SWITCH_EXP_STATE", i_rep_val=irep, &
                                   i_rep_section=ikind, i_vals=switch)
         flex_basis%subset(switch(1))%opt_exps(switch(2)) = .NOT. flex_basis%subset(switch(1))%opt_exps(switch(2))
      END DO

      CALL section_vals_val_get(kind1_section, "SWITCH_SET_STATE", n_rep_val=nrep, i_rep_section=ikind)
      DO irep = 1, nrep
         CALL section_vals_val_get(kind1_section, "SWITCH_SET_STATE", i_rep_val=irep, &
                                   i_rep_section=ikind, i_vals=switch)
         DO ipgf = 1, flex_basis%subset(switch(2))%nexp
            SELECT CASE (switch (1))
            CASE (0) ! switch all states in the set
               DO icont = 1, flex_basis%subset(switch(2))%ncon_tot
                  flex_basis%subset(switch(2))%opt_coeff(ipgf, icont) = &
                     .NOT. flex_basis%subset(switch(2))%opt_coeff(ipgf, icont)
               END DO
               flex_basis%subset(switch(2))%opt_exps(ipgf) = .NOT. flex_basis%subset(switch(2))%opt_exps(ipgf)
            CASE (1) ! switch only exp
               flex_basis%subset(switch(2))%opt_exps(ipgf) = .NOT. flex_basis%subset(switch(2))%opt_exps(ipgf)
            CASE (2) ! switch only coeff
               DO icont = 1, flex_basis%subset(switch(2))%ncon_tot
                  flex_basis%subset(switch(2))%opt_coeff(ipgf, icont) = &
                     .NOT. flex_basis%subset(switch(2))%opt_coeff(ipgf, icont)
               END DO
            CASE DEFAULT
               CALL cp__b("optimize_basis_utils.F",1001,"Invalid option in SWITCH_SET_STATE, 1st value has to be 0, 1 or 2")
            END SELECT
         END DO
      END DO

      ! perform a final modification. If basis set is uncontracted coefficient will never have to be optimized
      DO irep = 1, flex_basis%nsets
         IF (flex_basis%subset(irep)%nexp == 1) THEN
            DO ipgf = 1, flex_basis%subset(irep)%nexp
               flex_basis%subset(irep)%opt_coeff(ipgf, 1) = .FALSE.
            END DO
         END IF
      END DO

      ! finally count the total number of free parameters
      flex_basis%nopt = 0
      DO irep = 1, flex_basis%nsets
         DO ipgf = 1, flex_basis%subset(irep)%nexp
            DO icont = 1, flex_basis%subset(irep)%ncon_tot
               IF (flex_basis%subset(irep)%opt_coeff(ipgf, icont)) flex_basis%nopt = flex_basis%nopt+1
            END DO
            IF (flex_basis%subset(irep)%opt_exps(ipgf)) flex_basis%nopt = flex_basis%nopt+1
         END DO
      END DO

   END SUBROUTINE fill_basis_template

! **************************************************************************************************
!> \brief Helper function to parse input. Converts l and index position of
!>        a contraction to index in the contraction array of the set using lmin and nl
!> \param lmin ...
!> \param nl ...
!> \param icontr ...
!> \param l ...
!> \retval ientry ...
!> \author Florian Schiffmann
! **************************************************************************************************

   FUNCTION convert_l_contr_to_entry(lmin, nl, icontr, l) RESULT(ientry)
      INTEGER                                            :: lmin
      INTEGER, DIMENSION(:)                              :: nl
      INTEGER                                            :: icontr, l, ientry

      INTEGER                                            :: i, icon2l, iwork

      iwork = l-lmin
      icon2l = 0
      DO i = 1, iwork
         icon2l = icon2l+nl(i)
      END DO
      ientry = icon2l+icontr

   END FUNCTION convert_l_contr_to_entry

! **************************************************************************************************
!> \brief Read the reference basis sets from the template basis file
!> \param flex_basis ...
!> \param template_basis_file ...
!> \param element ...
!> \param basis_name ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE parse_basis(flex_basis, template_basis_file, element, basis_name, para_env)
      TYPE(flex_basis_type)                              :: flex_basis
      CHARACTER(LEN=default_path_length)                 :: template_basis_file
      CHARACTER(LEN=default_string_length)               :: element, basis_name
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_basis', routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(LEN=LEN(basis_name)+2)                   :: basis_name2
      CHARACTER(LEN=LEN(element)+2)                      :: element2
      INTEGER                                            :: iset, strlen1, strlen2
      LOGICAL                                            :: basis_found, found, match
      TYPE(cp_parser_type), POINTER                      :: parser

      basis_found = .FALSE.
      CALL uppercase(element)
      CALL uppercase(basis_name)
      NULLIFY (parser)
      CALL parser_create(parser, template_basis_file, para_env=para_env)

      search_loop: DO
         CALL parser_search_string(parser, TRIM(basis_name), .TRUE., found, line)
         IF (found) THEN
            match = .FALSE.
            CALL uppercase(line)
            ! Check both the element symbol and the basis set name
            line2 = " "//line//" "
            element2 = " "//TRIM(element)//" "
            basis_name2 = " "//TRIM(basis_name)//" "
            strlen1 = LEN_TRIM(element2)+1
            strlen2 = LEN_TRIM(basis_name2)+1
            IF ((INDEX(line2, element2(:strlen1)) > 0) .AND. &
                (INDEX(line2, basis_name2(:strlen2)) > 0)) match = .TRUE.
            IF (match) THEN
               CALL parser_get_object(parser, flex_basis%nsets, newline=.TRUE.)
               ALLOCATE (flex_basis%subset(flex_basis%nsets))
               DO iset = 1, flex_basis%nsets
                  CALL parse_subset(parser, flex_basis%subset(iset))
               END DO
               basis_found = .TRUE.
               EXIT
            END IF
         ELSE
            EXIT search_loop
         END IF
      END DO search_loop
      CALL parser_release(parser)

      IF (.NOT. basis_found) CALL cp_abort(cp__l("optimize_basis_utils.F",1115), &
                                           "The requested basis set <"//TRIM(basis_name)// &
                                           "> for element <"//TRIM(element)//"> was not "// &
                                           "found in the template basis set file ")

   END SUBROUTINE parse_basis

! **************************************************************************************************
!> \brief Read the subset information from the template basis file
!> \param parser ...
!> \param subset ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE parse_subset(parser, subset)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(subset_type)                                  :: subset

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_subset', routineP = moduleN//':'//routineN

      CHARACTER(len=20*default_string_length)            :: line_att
      INTEGER                                            :: icon1, icon2, il, ipgf, ishell, istart
      REAL(KIND=dp)                                      :: gs_scale
      REAL(KIND=dp), POINTER                             :: r_val

      line_att = ""
      CALL parser_get_object(parser, subset%n, newline=.TRUE.)
      CALL parser_get_object(parser, subset%lmin)
      CALL parser_get_object(parser, subset%lmax)
      CALL parser_get_object(parser, subset%nexp)
      subset%nl = subset%lmax-subset%lmin+1
      ALLOCATE (r_val)
      ALLOCATE (subset%l(subset%nl))
      ALLOCATE (subset%exps(subset%nexp))
      ALLOCATE (subset%exp_has_const(subset%nexp))
      subset%exp_has_const = .FALSE.
      ALLOCATE (subset%opt_exps(subset%nexp))
      subset%opt_exps = .FALSE.
      ALLOCATE (subset%exp_const(subset%nexp))
      ALLOCATE (subset%exp_x_ind(subset%nexp))
      DO ishell = 1, subset%nl
         CALL parser_get_object(parser, subset%l(ishell))
      END DO
      subset%ncon_tot = SUM(subset%l)
      ALLOCATE (subset%coeff(subset%nexp, subset%ncon_tot))
      ALLOCATE (subset%opt_coeff(subset%nexp, subset%ncon_tot))
      subset%opt_coeff = .FALSE.
      ALLOCATE (subset%coeff_x_ind(subset%nexp, subset%ncon_tot))
      DO ipgf = 1, subset%nexp
         CALL parser_get_object(parser, r_val, newline=.TRUE.)
         subset%exps(ipgf) = r_val
         DO ishell = 1, subset%ncon_tot
            CALL parser_get_object(parser, r_val)
            subset%coeff(ipgf, ishell) = r_val
         END DO
      END DO

      ! orthonormalize contraction coefficients using gram schmidt
      istart = 1
      DO il = 1, subset%nl
         DO icon1 = istart, istart+subset%l(il)-2
            DO icon2 = icon1+1, istart+subset%l(il)-1
               gs_scale = DOT_PRODUCT(subset%coeff(:, icon2), subset%coeff(:, icon1))/ &
                          DOT_PRODUCT(subset%coeff(:, icon1), subset%coeff(:, icon1))
               subset%coeff(:, icon2) = subset%coeff(:, icon2)-gs_scale*subset%coeff(:, icon1)
            END DO
         END DO
         istart = istart+subset%l(il)
      END DO

      ! just to get an understandable basis normalize coefficients
      DO icon1 = 1, subset%ncon_tot
         subset%coeff(:, icon1) = subset%coeff(:, icon1)/ &
                                  SQRT(DOT_PRODUCT(subset%coeff(:, icon1), subset%coeff(:, icon1)))
      END DO
      DEALLOCATE (r_val)

   END SUBROUTINE parse_subset

! **************************************************************************************************
!> \brief Initialize the variables for the powell optimizer
!> \param p_param ...
!> \param powell_section ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE init_powell_var(p_param, powell_section)
      TYPE(opt_state_type)                               :: p_param
      TYPE(section_vals_type), POINTER                   :: powell_section

      CHARACTER(len=*), PARAMETER :: routineN = 'init_powell_var', &
         routineP = moduleN//':'//routineN

      p_param%state = 0
      p_param%nvar = 0
      p_param%iprint = 0
      p_param%unit = default_output_unit
      CALL section_vals_val_get(powell_section, "ACCURACY", r_val=p_param%rhoend)
      CALL section_vals_val_get(powell_section, "STEP_SIZE", r_val=p_param%rhobeg)
      CALL section_vals_val_get(powell_section, "MAX_FUN", i_val=p_param%maxfun)

   END SUBROUTINE init_powell_var

END MODULE optimize_basis_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optimize_basis
   USE admm_methods,                    ONLY: admm_fit_mo_coeffs
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE f77_interface,                   ONLY: create_force_env,&
                                              destroy_force_env,&
                                              f_env_add_defaults,&
                                              f_env_get_from_id,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_cp2k_read,                 ONLY: empty_initial_variables,&
                                              read_input
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_release,&
                                              section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE machine,                         ONLY: m_chdir,&
                                              m_getcwd,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_comm_free,&
                                              mp_comm_split,&
                                              mp_sum,&
                                              mp_sync
   USE optbas_fenv_manipulation,        ONLY: allocate_mo_sets,&
                                              calculate_ks_matrix,&
                                              calculate_overlap,&
                                              calculate_overlap_inverse,&
                                              create_opt_admm_env,&
                                              modify_input_settings,&
                                              update_basis_set
   USE optbas_opt_utils,                ONLY: evaluate_energy,&
                                              evaluate_fval
   USE optimize_basis_types,            ONLY: basis_optimization_type,&
                                              deallocate_basis_optimization_type,&
                                              subset_type
   USE optimize_basis_utils,            ONLY: get_set_and_basis_id,&
                                              optimize_basis_init_read_input,&
                                              update_derived_basis_sets
   USE powell,                          ONLY: powell_optimize
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_neighbor_lists,               ONLY: build_qs_neighbor_lists
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: run_optimize_basis

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optimize_basis'

CONTAINS

! **************************************************************************************************
!> \brief main entry point for methods aimed at optimizing basis sets
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE run_optimize_basis(input_declaration, root_section, para_env)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'run_optimize_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(basis_optimization_type)                      :: opt_bas

      CALL timeset(routineN, handle)

      CALL optimize_basis_init_read_input(opt_bas, root_section, para_env)

      CALL driver_para_opt_basis(opt_bas, input_declaration, para_env)

      CALL deallocate_basis_optimization_type(opt_bas)
      CALL timestop(handle)

   END SUBROUTINE run_optimize_basis

! **************************************************************************************************
!> \brief driver routine for the parallel part of the method
!> \param opt_bas ...
!> \param input_declaration ...
!> \param para_env ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE driver_para_opt_basis(opt_bas, input_declaration, para_env)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'driver_para_opt_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n_groups_created, opt_group
      INTEGER, DIMENSION(:), POINTER                     :: group_distribution_p
      INTEGER, DIMENSION(0:para_env%num_pe-1), TARGET    :: group_distribution

      CALL timeset(routineN, handle)
      group_distribution_p => group_distribution
      CALL mp_comm_split(para_env%group, opt_group, n_groups_created, group_distribution_p, &
                         n_subgroups=SIZE(opt_bas%group_partition), group_partition=opt_bas%group_partition)
      opt_bas%opt_id = group_distribution(para_env%mepos)+1

      CALL driver_optimization_para_low(opt_bas, input_declaration, para_env, opt_group)

      CALL mp_comm_free(opt_group)
      CALL timestop(handle)

   END SUBROUTINE driver_para_opt_basis

! **************************************************************************************************
!> \brief low level optimization routine includes initialization of the subsytems
!>        powell optimizer and deallocation of the various force envs
!> \param opt_bas ...
!> \param input_declaration ...
!> \param para_env_top ...
!> \param mpi_comm_opt ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE driver_optimization_para_low(opt_bas, input_declaration, para_env_top, mpi_comm_opt)
      TYPE(basis_optimization_type)                      :: opt_bas
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env_top
      INTEGER                                            :: mpi_comm_opt

      CHARACTER(len=*), PARAMETER :: routineN = 'driver_optimization_para_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icalc, iopt, mp_id, stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: f_env_id
      LOGICAL                                            :: write_basis
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tot_time
      TYPE(cp_fm_p_type), ALLOCATABLE, DIMENSION(:)      :: matrix_S_inv
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)

      CALL timeset(routineN, handle)

      ! ======  initialize the f_env and precompute some matrices =====
      mp_id = opt_bas%opt_id
      NULLIFY (para_env, f_env)
      ALLOCATE (f_env_id(SIZE(opt_bas%comp_group(mp_id)%member_list)))
      ALLOCATE (tot_time(opt_bas%ncombinations*opt_bas%ntraining_sets))
      ALLOCATE (matrix_s_inv(SIZE(opt_bas%comp_group(mp_id)%member_list)))

      CALL cp_para_env_create(para_env, group=mpi_comm_opt, &
                              owns_group=.FALSE.)

      CALL init_training_force_envs(opt_bas, f_env_id, input_declaration, matrix_s_inv, para_env, mpi_comm_opt)

      CALL init_free_vars(opt_bas)
      tot_time = 0.0_dp

      ! ======= The real optimization loop  =======
      DO iopt = 0, opt_bas%powell_param%maxfun
         CALL compute_residuum_vectors(opt_bas, f_env_id, matrix_S_inv, tot_time, &
                                       para_env_top, para_env, iopt)
         IF (para_env_top%ionode) &
            CALL powell_optimize(opt_bas%powell_param%nvar, opt_bas%x_opt, opt_bas%powell_param)
         IF (.NOT. para_env_top%ionode) opt_bas%x_opt = 0.0_dp
         CALL mp_bcast(opt_bas%powell_param%state, para_env_top%source, para_env_top%group)
         CALL mp_sum(opt_bas%x_opt, para_env_top%group)
         CALL update_free_vars(opt_bas)
         write_basis = MOD(iopt, opt_bas%write_frequency) == 0
         CALL update_derived_basis_sets(opt_bas, write_basis, opt_bas%output_basis_file, &
                                        para_env_top)
         IF (opt_bas%powell_param%state == -1) EXIT
      END DO

      ! ======= Update the basis set and print the final basis  =======
      IF (para_env_top%ionode) THEN
         opt_bas%powell_param%state = 8
         CALL powell_optimize(opt_bas%powell_param%nvar, opt_bas%x_opt, opt_bas%powell_param)
      END IF

      IF (.NOT. para_env_top%ionode) opt_bas%x_opt = 0.0_dp
      CALL mp_sum(opt_bas%x_opt, para_env_top%group)
      CALL update_free_vars(opt_bas)
      CALL update_derived_basis_sets(opt_bas, .TRUE., opt_bas%output_basis_file, &
                                     para_env_top)

      ! ======  get rid of the f_env again =====

      DO icalc = SIZE(opt_bas%comp_group(mp_id)%member_list), 1, -1
         CALL f_env_get_from_id(f_env_id(icalc), f_env)
         CALL destroy_force_env(f_env_id(icalc), stat)
         CALL cp_fm_release(matrix_s_inv(icalc)%matrix)
      END DO
      DEALLOCATE (f_env_id); DEALLOCATE (tot_time); DEALLOCATE (matrix_S_inv)
      CALL cp_para_env_release(para_env)
      CALL timestop(handle)

   END SUBROUTINE driver_optimization_para_low

! **************************************************************************************************
!> \brief compute all ingredients for powell optimizer. Rho_diff,
!>        condition number, energy,... for all ttraining sets in
!>        the computational group
!> \param opt_bas ...
!> \param f_env_id ...
!> \param matrix_S_inv ...
!> \param tot_time ...
!> \param para_env_top ...
!> \param para_env ...
!> \param iopt ...
! **************************************************************************************************

   SUBROUTINE compute_residuum_vectors(opt_bas, f_env_id, matrix_S_inv, tot_time, &
                                       para_env_top, para_env, iopt)
      TYPE(basis_optimization_type)                      :: opt_bas
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: f_env_id
      TYPE(cp_fm_p_type), DIMENSION(:)                   :: matrix_S_inv
      REAL(KIND=dp), DIMENSION(:)                        :: tot_time
      TYPE(cp_para_env_type), POINTER                    :: para_env_top, para_env
      INTEGER                                            :: iopt

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_residuum_vectors', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bas_id, handle, icalc, icomb, mp_id, &
                                                            my_id, ncalc, set_id
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cond_vec, energy, f_vec, my_time, &
                                                            start_time
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s_aux_fit, &
                                                            matrix_s_aux_fit_vs_orb
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CALL timeset(routineN, handle)

      NULLIFY (matrix_ks, matrix_s_aux_fit_vs_orb, matrix_s_aux_fit, ks_env)

      ncalc = opt_bas%ncombinations*opt_bas%ntraining_sets
      ALLOCATE (f_vec(ncalc)); ALLOCATE (my_time(ncalc)); ALLOCATE (cond_vec(ncalc)); ALLOCATE (energy(ncalc))
      f_vec = 0.0_dp; cond_vec = 0.0_dp; my_time = 0.0_dp; energy = 0.0_dp
      mp_id = opt_bas%opt_id
      ALLOCATE (start_time(SIZE(opt_bas%comp_group(mp_id)%member_list)))

      DO icalc = 1, SIZE(opt_bas%comp_group(mp_id)%member_list)
         my_id = opt_bas%comp_group(mp_id)%member_list(icalc)+1
         ! setup timings
         start_time(icalc) = m_walltime()

         CALL get_set_and_basis_id(opt_bas%comp_group(mp_id)%member_list(icalc), opt_bas, set_id, bas_id)
         CALL f_env_get_from_id(f_env_id(icalc), f_env)
         force_env => f_env%force_env
         CALL force_env_get(force_env, qs_env=qs_env)
         CALL get_qs_env(qs_env, ks_env=ks_env)
         CALL update_basis_set(opt_bas, bas_id, qs_env)
         CALL build_qs_neighbor_lists(qs_env, para_env, molecular=.FALSE., force_env_section=qs_env%input)
         CALL calculate_overlap(ks_env, "S_AB_AUX", qs_env%input)
         CALL get_qs_env(qs_env, &
                         matrix_ks=matrix_ks, &
                         matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                         matrix_s_aux_fit=matrix_s_aux_fit)
         CALL admm_fit_mo_coeffs(qs_env%admm_env, matrix_s_aux_fit, &
                                 matrix_s_aux_fit_vs_orb, qs_env%mos, qs_env%mos_aux_fit, &
                                 geometry_did_change=.TRUE.)
         CALL evaluate_fval(qs_env%mos, qs_env%mos_aux_fit, matrix_s_aux_fit_vs_orb(1)%matrix, &
                            matrix_s_aux_fit(1)%matrix, qs_env%admm_env, f_vec(my_id), cond_vec(my_id))
         CALL evaluate_energy(qs_env%mos_aux_fit, matrix_ks, matrix_s_inv(icalc)%matrix, qs_env%admm_env%Q, &
                              qs_env%admm_env%work_aux_aux, energy(my_id))

         my_time(my_id) = m_walltime()-start_time(icalc)
         IF (.NOT. para_env%ionode) THEN
            f_vec = 0.0_dp; cond_vec = 0.0_dp; my_time = 0.0_dp; energy = 0.0_dp
         END IF
      END DO

      DEALLOCATE (start_time)

      CALL mp_sum(f_vec, para_env_top%group); CALL mp_sum(cond_vec, para_env_top%group); CALL mp_sum(my_time, para_env_top%group)
      opt_bas%powell_param%f = 0.0_dp
      DO icalc = 1, SIZE(f_vec)
         icomb = MOD(icalc-1, opt_bas%ncombinations)
         opt_bas%powell_param%f = opt_bas%powell_param%f+ &
                                  (f_vec(icalc)+energy(icalc))*opt_bas%fval_weight(icomb)
         IF (opt_bas%use_condition_number) &
            opt_bas%powell_param%f = opt_bas%powell_param%f+ &
                                     LOG(cond_vec(icalc))*opt_bas%condition_weight(icomb)
      END DO

      CALL mp_sync(para_env_top%group)

      ! output info if required
      CALL output_opt_info(f_vec, cond_vec, my_time, tot_time, opt_bas, iopt, para_env_top)
      DEALLOCATE (f_vec); DEALLOCATE (my_time); DEALLOCATE (cond_vec); DEALLOCATE (energy)

      CALL timestop(handle)

   END SUBROUTINE compute_residuum_vectors

! **************************************************************************************************
!> \brief create the force_envs for every input in the computational group
!> \param opt_bas ...
!> \param f_env_id ...
!> \param input_declaration ...
!> \param matrix_s_inv ...
!> \param para_env ...
!> \param mpi_comm_opt ...
! **************************************************************************************************

   SUBROUTINE init_training_force_envs(opt_bas, f_env_id, input_declaration, matrix_s_inv, para_env, mpi_comm_opt)

      TYPE(basis_optimization_type)                      :: opt_bas
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: f_env_id
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_fm_p_type), DIMENSION(:)                   :: matrix_S_inv
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: mpi_comm_opt

      CHARACTER(len=*), PARAMETER :: routineN = 'init_training_force_envs', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: main_dir
      INTEGER                                            :: bas_id, handle, icalc, ierr, mp_id, &
                                                            set_id, stat
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(section_vals_type), POINTER                   :: input_file

      CALL timeset(routineN, handle)

      NULLIFY (matrix_s, blacs_env, ks_env)

      mp_id = opt_bas%opt_id
      CALL m_getcwd(main_dir)

      ! ======= Create f_env for all calculations in MPI group =======
      DO icalc = 1, SIZE(opt_bas%comp_group(mp_id)%member_list)
         NULLIFY (input_file)
         ! parse the input of the training sets
         CALL get_set_and_basis_id(opt_bas%comp_group(mp_id)%member_list(icalc), opt_bas, set_id, bas_id)
         CALL m_chdir(TRIM(opt_bas%training_dir(set_id)), ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("optimize_basis.F",366)
         input_file => read_input(input_declaration, &
                                  opt_bas%training_input(set_id), &
                                  initial_variables=empty_initial_variables, &
                                  para_env=para_env)

         CALL modify_input_settings(opt_bas, bas_id, input_file)
         CALL create_force_env(f_env_id(icalc), &
                               input_declaration=input_declaration, &
                               input_path=opt_bas%training_input(set_id), &
                               input=input_file, &
                               output_path="scrap_information", &
                               mpi_comm=mpi_comm_opt, &
                               ierr=stat)

         ! some weirdness with the default stacks defaults have to be addded to get the
         ! correct default program name this causes trouble with the timer stack if kept
         CALL f_env_add_defaults(f_env_id(icalc), f_env)
         force_env => f_env%force_env
         CALL force_env_get(force_env, qs_env=qs_env)
         CALL allocate_mo_sets(qs_env)
         CALL f_env_rm_defaults(f_env, stat)
         CALL get_qs_env(qs_env, ks_env=ks_env)
         CALL build_qs_neighbor_lists(qs_env, para_env, molecular=.FALSE., &
                                      force_env_section=qs_env%input)
         CALL calculate_overlap(ks_env, "S_AB", qs_env%input)
         CALL get_qs_env(qs_env, matrix_s=matrix_s, blacs_env=blacs_env)
         CALL calculate_overlap_inverse(matrix_s(1)%matrix, matrix_s_inv(icalc)%matrix, &
                                        para_env, blacs_env)
         CALL calculate_ks_matrix(qs_env)

         CALL create_opt_admm_env(qs_env)
         CALL section_vals_release(input_file)
         CALL m_chdir(TRIM(ADJUSTL(main_dir)), ierr)
      END DO

      CALL timestop(handle)

   END SUBROUTINE init_training_force_envs

! **************************************************************************************************
!> \brief variable update from the powell vector for all sets
!> \param opt_bas ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE update_free_vars(opt_bas)
      TYPE(basis_optimization_type)                      :: opt_bas

      CHARACTER(len=*), PARAMETER :: routineN = 'update_free_vars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, iset, ix

      CALL timeset(routineN, handle)
      ix = 0
      DO ikind = 1, opt_bas%nkind
         DO iset = 1, opt_bas%kind_basis(ikind)%flex_basis(0)%nsets
            CALL update_subset_freevars(opt_bas%kind_basis(ikind)%flex_basis(0)%subset(iset), ix, opt_bas%x_opt)
         END DO
      END DO
      CALL timestop(handle)

   END SUBROUTINE update_free_vars

! **************************************************************************************************
!> \brief low level update for the basis sets. Exponents are transformed according to constraint
!> \param subset ...
!> \param ix ...
!> \param x ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE update_subset_freevars(subset, ix, x)
      TYPE(subset_type)                                  :: subset
      INTEGER                                            :: ix
      REAL(KIND=dp), DIMENSION(:)                        :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'update_subset_freevars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icon1, icon2, icont, iexp, il, &
                                                            istart
      REAL(KIND=dp)                                      :: fermi_f, gs_scale

      CALL timeset(routineN, handle)
      DO iexp = 1, subset%nexp
         IF (subset%opt_exps(iexp)) THEN
            ix = ix+1
            subset%exps(iexp) = ABS(x(ix))
            IF (subset%exp_has_const(iexp)) THEN
               !use a fermi function to keep expoenents in a given range around their initial value
               fermi_f = 1.0_dp/(EXP((x(ix)-1.0_dp)/0.5_dp)+1.0_dp)
               subset%exps(iexp) = (2.0_dp*fermi_f-1.0_dp)*subset%exp_const(iexp)%var_fac*subset%exp_const(iexp)%init+ &
                                   subset%exp_const(iexp)%init
            ELSE

            END IF
         END IF
         DO icont = 1, subset%ncon_tot
            IF (subset%opt_coeff(iexp, icont)) THEN
               ix = ix+1
               subset%coeff(iexp, icont) = x(ix)
            END IF
         END DO
      END DO

      ! orthonormalize contraction coefficients using gram schmidt
      istart = 1
      DO il = 1, subset%nl
         DO icon1 = istart, istart+subset%l(il)-2
            DO icon2 = icon1+1, istart+subset%l(il)-1
               gs_scale = DOT_PRODUCT(subset%coeff(:, icon2), subset%coeff(:, icon1))/ &
                          DOT_PRODUCT(subset%coeff(:, icon1), subset%coeff(:, icon1))
               subset%coeff(:, icon2) = subset%coeff(:, icon2)-gs_scale*subset%coeff(:, icon1)
            END DO
         END DO
         istart = istart+subset%l(il)
      END DO

      DO icon1 = 1, subset%ncon_tot
         subset%coeff(:, icon1) = subset%coeff(:, icon1)/ &
                                  SQRT(DOT_PRODUCT(subset%coeff(:, icon1), subset%coeff(:, icon1)))
      END DO
      CALL timestop(handle)

   END SUBROUTINE update_subset_freevars

! **************************************************************************************************
!> \brief variable initialization for the powell vector for all sets
!> \param opt_bas ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE init_free_vars(opt_bas)
      TYPE(basis_optimization_type)                      :: opt_bas

      CHARACTER(len=*), PARAMETER :: routineN = 'init_free_vars', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, iset, ix

      CALL timeset(routineN, handle)
      ix = 0
      DO ikind = 1, opt_bas%nkind
         DO iset = 1, opt_bas%kind_basis(ikind)%flex_basis(0)%nsets
            CALL init_subset_freevars(opt_bas%kind_basis(ikind)%flex_basis(0)%subset(iset), ix, opt_bas%x_opt)
         END DO
      END DO
      CALL timestop(handle)

   END SUBROUTINE init_free_vars

! **************************************************************************************************
!> \brief variable initialization for the powell vector from low level informations
!>        constraint exponents will be mapped on a fermi function
!> \param subset ...
!> \param ix ...
!> \param x ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE init_subset_freevars(subset, ix, x)
      TYPE(subset_type)                                  :: subset
      INTEGER                                            :: ix
      REAL(KIND=dp), DIMENSION(:)                        :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'init_subset_freevars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icont, iexp
      REAL(KIND=dp)                                      :: fract

      CALL timeset(routineN, handle)

      DO iexp = 1, subset%nexp
         IF (subset%opt_exps(iexp)) THEN
            ix = ix+1
            x(ix) = subset%exps(iexp)
            IF (subset%exp_has_const(iexp)) THEN
               IF (subset%exp_const(iexp)%const_type == 0) THEN
                  fract = 1.0_dp+(subset%exps(iexp)-subset%exp_const(iexp)%init)/ &
                          (subset%exp_const(iexp)%init*subset%exp_const(iexp)%var_fac)
                  x(ix) = 0.5_dp*LOG((2.0_dp/fract-1.0_dp))+1.0_dp
               END IF
               IF (subset%exp_const(iexp)%const_type == 1) THEN
                  x(ix) = 1.0_dp
               END IF
            END IF
         END IF
         DO icont = 1, subset%ncon_tot
            IF (subset%opt_coeff(iexp, icont)) THEN
               ix = ix+1
               x(ix) = subset%coeff(iexp, icont)
            END IF
         END DO
      END DO
      CALL timestop(handle)

   END SUBROUTINE init_subset_freevars

! **************************************************************************************************
!> \brief commuticates all info to the master and assembles the output
!> \param f_vec ...
!> \param cond_vec ...
!> \param my_time ...
!> \param tot_time ...
!> \param opt_bas ...
!> \param iopt ...
!> \param para_env_top ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE output_opt_info(f_vec, cond_vec, my_time, tot_time, opt_bas, iopt, para_env_top)
      REAL(KIND=dp), DIMENSION(:)                        :: f_vec, cond_vec, my_time, tot_time
      TYPE(basis_optimization_type)                      :: opt_bas
      INTEGER                                            :: iopt
      TYPE(cp_para_env_type), POINTER                    :: para_env_top

      CHARACTER(len=*), PARAMETER :: routineN = 'output_opt_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ibasis, icalc, iset, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      tot_time = tot_time+my_time

      unit_nr = -1
      IF (para_env_top%ionode .AND. (MOD(iopt, opt_bas%write_frequency) == 0 .OR. iopt == opt_bas%powell_param%maxfun)) &
         unit_nr = cp_logger_get_default_unit_nr(logger)

      IF (unit_nr .GT. 0) THEN
         WRITE (unit_nr, '(1X,A,I8)') "BASOPT| Information at iteration number:", iopt
         WRITE (unit_nr, '(1X,A)') "BASOPT| Training set | Combination | Rho difference | Condition num. | Time"
         WRITE (unit_nr, '(1X,A)') "BASOPT| -----------------------------------------------------------------------"
         icalc = 0
         DO iset = 1, opt_bas%ntraining_sets
            DO ibasis = 1, opt_bas%ncombinations
               icalc = icalc+1
               WRITE (unit_nr, '(1X,A,2(5X,I3,5X,A),2(1X,E14.8,1X,A),1X,F8.3)') &
                  'BASOPT| ', iset, "|", ibasis, "|", f_vec(icalc), "|", cond_vec(icalc), "|", tot_time(icalc)
            END DO
         END DO
         WRITE (unit_nr, '(1X,A)') "BASOPT| -----------------------------------------------------------------------"
         WRITE (unit_nr, '(1X,A,E14.8)') "BASOPT| Total residuum value: ", opt_bas%powell_param%f
         WRITE (unit_nr, '(A)') ""
      END IF
      CALL timestop(handle)
   END SUBROUTINE output_opt_info

END MODULE optimize_basis

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Subroutines to perform calculations on molecules from a bigger
!>        system. Useful to generate a high-quality MO guess for systems
!>        of many molecules with complex electronic structure, to bootstrap
!>        ALMO simulations, etc.
!> \par History
!>      10.2014 Rustam Z Khaliullin
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE mscfg_methods
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_create,&
                                              cp_dbcsr_init,&
                                              dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: create_small_subsys
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: almo_frz_crystal,&
                                              almo_frz_none,&
                                              do_qs,&
                                              molecular_guess
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length
   USE molecule_types_new,              ONLY: get_molecule_set_info,&
                                              molecule_type
   USE mscfg_types,                     ONLY: molecular_scf_guess_env_init,&
                                              molecular_scf_guess_env_type,&
                                              mscfg_max_moset_size
   USE particle_list_types,             ONLY: particle_list_type
   USE qs_energy,                       ONLY: qs_energies
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment,                  ONLY: qs_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_env_create,&
                                              qs_env_release,&
                                              qs_environment_type
   USE qs_mo_types,                     ONLY: mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mscfg_methods'

   PUBLIC :: loop_over_molecules, do_mol_loop

CONTAINS

! **************************************************************************************************
!> \brief Prepare data for calculations on isolated molecules.
!> \param globenv ...
!> \param force_env ...
!> \par   History
!>        10.2014 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE loop_over_molecules(globenv, force_env)

      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'loop_over_molecules', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nmols
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: charge_of_frag, first_atom_of_frag, &
                                                            last_atom_of_frag, multip_of_frag
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CALL force_env_get(force_env, qs_env=qs_env)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("mscfg_methods.F",90)
      CALL get_qs_env(qs_env, &
                      molecule_set=molecule_set)

      nmols = SIZE(molecule_set)

      ALLOCATE (first_atom_of_frag(nmols))
      ALLOCATE (last_atom_of_frag(nmols))
      ALLOCATE (charge_of_frag(nmols))
      ALLOCATE (multip_of_frag(nmols))

      CALL get_molecule_set_info(molecule_set, &
                                 mol_to_first_atom=first_atom_of_frag, &
                                 mol_to_last_atom=last_atom_of_frag, &
                                 mol_to_charge=charge_of_frag, &
                                 mol_to_multiplicity=multip_of_frag)

      CALL calcs_on_isolated_molecules(force_env, globenv, nmols, &
                                       first_atom_of_frag, last_atom_of_frag, charge_of_frag, multip_of_frag)

      DEALLOCATE (first_atom_of_frag)
      DEALLOCATE (last_atom_of_frag)
      DEALLOCATE (charge_of_frag)
      DEALLOCATE (multip_of_frag)

   END SUBROUTINE loop_over_molecules

! **************************************************************************************************
!> \brief Run calculations on isolated molecules. The ideas for setting up
!>        the calculations are borrowed from BSSE files
!> \param force_env ...
!> \param globenv ...
!> \param nfrags ...
!> \param first_atom_of_frag ...
!> \param last_atom_of_frag ...
!> \param charge_of_frag ...
!> \param multip_of_frag ...
!> \par   History
!>        10.2014 created
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE calcs_on_isolated_molecules(force_env, globenv, nfrags, &
                                          first_atom_of_frag, last_atom_of_frag, charge_of_frag, multip_of_frag)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      INTEGER, INTENT(IN)                                :: nfrags
      INTEGER, DIMENSION(:), INTENT(IN)                  :: first_atom_of_frag, last_atom_of_frag, &
                                                            charge_of_frag, multip_of_frag

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calcs_on_isolated_molecules', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_type
      INTEGER :: first_atom, force_method, global_charge, global_multpl, handle, i, ifrag, imo, &
         isize, j, k, last_atom, my_targ, nmosets_of_frag
      INTEGER, DIMENSION(:), POINTER                     :: atom_index, atom_list
      LOGICAL                                            :: global_almo_scf_keyword
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_loc
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_of_frag
      TYPE(molecular_scf_guess_env_type), POINTER        :: mscfg_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qs_energy_type), POINTER                      :: qs_energy
      TYPE(qs_environment_type), POINTER                 :: qs_env, qs_env_loc
      TYPE(section_vals_type), POINTER                   :: dft_section, force_env_section, &
                                                            qs_section, root_section, &
                                                            subsys_section

      CALL timeset(routineN, handle)

      NULLIFY (subsys_loc, subsys, particles, para_env, cell, atom_index, atom_type, &
               force_env_section, qs_env_loc, mscfg_env, qs_env, qs_energy)
      CALL force_env_get(force_env, force_env_section=force_env_section, &
                         qs_env=qs_env)
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=force_method)
      IF(.NOT.(force_method .EQ. do_qs))CALL cp__a("mscfg_methods.F",169)
      root_section => force_env%root_section
      subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      dft_section => section_vals_get_subs_vals(force_env_section, "DFT")
      !
      ! Save several global settings to restore them after the loop:
      !  charge, multiplicity, ALMO flag
      !
      CALL section_vals_val_get(dft_section, "CHARGE", i_val=global_charge)
      CALL section_vals_val_get(dft_section, "MULTIPLICITY", i_val=global_multpl)
      qs_section => section_vals_get_subs_vals(dft_section, "QS")
      CALL section_vals_val_get(qs_section, "ALMO_SCF", l_val=global_almo_scf_keyword)
      !
      ! Get access to critical data before the loop
      !
      CALL force_env_get(force_env=force_env, subsys=subsys, para_env=para_env, &
                         cell=cell)
      CALL cp_subsys_get(subsys, particles=particles)
      CALL get_qs_env(qs_env, mscfg_env=mscfg_env)
      IF(.NOT.(ASSOCIATED(mscfg_env)))CALL cp__a("mscfg_methods.F",188)
      !
      ! These flags determine the options of molecular runs (e.g. cell size)
      !
      !!!LATER is_fast_dirty = mscfg_env%is_fast_dirty - shrink the cell
      !!!LATER is_crystal = mscfg_env%is_crystal - remove periodicity
      !
      ! Prepare storage for the results
      ! Until molecular_scf_guess_env is destroyed it will keep
      ! the results of fragment calculations
      !
      CALL molecular_scf_guess_env_init(mscfg_env, nfrags)

      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      !
      ! Start the loop over molecules
      !
      ! Here is the list of modifications necessary to run isolated molecules:
      ! * Atom list of a subsystem and their names
      ! * Charge and multiplicity of a subsystem
      ! * ALMO SCF flag off (unless several levels of recursion is desired)
      ! * Smaller cell can be provided if a fast-and-dirty approach is ok
      ! * ... add your own and explain it here ...
      !
      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      DO ifrag = 1, nfrags
         !
         ! Turn ALMO SCF flag off
         !
         CALL section_vals_val_set(qs_section, "ALMO_SCF", l_val=.FALSE.)
         !
         ! Setup the charge and multiplicity of the molecule
         !
         CALL section_vals_val_set(dft_section, "CHARGE", i_val=charge_of_frag(ifrag))
         CALL section_vals_val_set(dft_section, "MULTIPLICITY", i_val=multip_of_frag(ifrag))
         !
         ! Create a list of atoms in the current molecule
         !
         ! Assume that atoms arranged consecutively (in ALMO SCF it is always the case)
         ! It is important to have a linear scaling procedure here
         first_atom = first_atom_of_frag(ifrag)
         last_atom = last_atom_of_frag(ifrag)
         isize = last_atom-first_atom+1
         ALLOCATE (atom_index(isize))
         atom_index(1:isize) = (/(i, i=first_atom, last_atom)/)
         !
         ! Get atom type names
         !
         ALLOCATE (atom_type(isize))
         DO j = 1, isize
            my_targ = atom_index(j)
            DO k = 1, SIZE(particles%els)
               CALL get_atomic_kind(particles%els(k)%atomic_kind, atom_list=atom_list, name=name)
               IF (ANY(atom_list == my_targ)) EXIT
            END DO
            atom_type(j) = name
         END DO
         !
         ! Create the environment of a subsystem
         !
         CALL create_small_subsys(subsys_loc, big_subsys=subsys, &
                                  small_para_env=para_env, small_cell=cell, sub_atom_index=atom_index, &
                                  sub_atom_kind_name=atom_type, para_env=para_env, &
                                  force_env_section=force_env_section, subsys_section=subsys_section)
         CALL qs_env_create(qs_env_loc, globenv)
         CALL qs_init(qs_env_loc, para_env, globenv, root_section, cp_subsys=subsys_loc, &
                      force_env_section=force_env_section, subsys_section=subsys_section, &
                      use_motion_section=.FALSE.)
         CALL cp_subsys_release(subsys_loc)

         !
         ! Print-out fragment info
         !
         CALL print_frag_info(atom_index, atom_type, ifrag, nfrags, &
                              charge_of_frag(ifrag), multip_of_frag(ifrag))
         !
         !  Run calculations on a subsystem
         !
         CALL qs_energies(qs_env_loc)
         !
         !  Get the desired results (energy and MOs) out
         !
         CALL get_qs_env(qs_env_loc, mos=mos_of_frag, energy=qs_energy)
         !
         ! Store all desired results of fragment calculations in the fragment_env
         ! of the qs_env to use them later as needed
         !
         mscfg_env%energy_of_frag(ifrag) = qs_energy%total
         nmosets_of_frag = SIZE(mos_of_frag)
         IF(.NOT.(nmosets_of_frag .LE. mscfg_max_moset_size))CALL cp__a("mscfg_methods.F",277)
         mscfg_env%nmosets_of_frag(ifrag) = nmosets_of_frag
         DO imo = 1, nmosets_of_frag
            ! the matrices have been allocated already - copy the results there
            CALL cp_dbcsr_init(mscfg_env%mos_of_frag(ifrag, imo))
            CALL cp_dbcsr_create(mscfg_env%mos_of_frag(ifrag, imo), &
                                 template=mos_of_frag(imo)%mo_set%mo_coeff_b, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_copy(mscfg_env%mos_of_frag(ifrag, imo), &
                               mos_of_frag(imo)%mo_set%mo_coeff_b)
         ENDDO
         !
         ! Clean up
         !
         NULLIFY (qs_energy)
         CALL qs_env_release(qs_env_loc)
         DEALLOCATE (atom_index)
         DEALLOCATE (atom_type)

      END DO

      CALL section_vals_val_set(dft_section, "CHARGE", i_val=global_charge)
      CALL section_vals_val_set(dft_section, "MULTIPLICITY", i_val=global_multpl)
      CALL section_vals_val_set(qs_section, "ALMO_SCF", l_val=global_almo_scf_keyword)

      CALL timestop(handle)

   END SUBROUTINE calcs_on_isolated_molecules

! **************************************************************************************************
!> \brief Print info about fragment
!> \param atom_index ...
!> \param atom_type ...
!> \param frag ...
!> \param nfrags ...
!> \param charge ...
!> \param multpl ...
!> \par History
!>      07.2005 created as a part of BSSE calculations [tlaino]
!>      10.2014 adapted to ALMO guess calculations [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE print_frag_info(atom_index, atom_type, frag, nfrags, charge, &
                              multpl)

      INTEGER, DIMENSION(:), POINTER                     :: atom_index
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_type
      INTEGER, INTENT(IN)                                :: frag, nfrags, charge, multpl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_frag_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=11)                                  :: charI
      INTEGER                                            :: i, iw
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         iw = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         iw = -1
      ENDIF

      IF (iw > 0) THEN

         WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T5,A,T25,A,T40,I11,T53,A,T67,I11,T80,A)") &
            "-", "MOLECULAR GUESS:", "FRAGMENT", frag, "OUT OF", nfrags, "-"
         WRITE (UNIT=iw, FMT="(T2,A,T25,A,T40,I11,T53,A,T67,I11,T80,A)") "-", "CHARGE", charge, "MULTIPLICITY", &
            multpl, "-"
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T25,A,T53,A,T80,A)") "-", "ATOM INDEX", "ATOM NAME", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T25,A,T53,A,T80,A)") "-", "----------", "---------", "-"
         DO i = 1, SIZE(atom_index)
            WRITE (charI, '(I11)') atom_index(i)
            WRITE (UNIT=iw, FMT="(T2,A,T25,A,T53,A,T80,A)") "-", ADJUSTL(charI), TRIM(atom_type(i)), "-"
         END DO
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
      END IF

   END SUBROUTINE print_frag_info

! **************************************************************************************************
!> \brief Is the loop over molecules requested?
!> \param force_env ...
!> \retval do_mol_loop ...
!> \par History
!>       10.2014 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   FUNCTION do_mol_loop(force_env)

      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL                                            :: do_mol_loop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_mol_loop', routineP = moduleN//':'//routineN

      INTEGER                                            :: almo_guess_type, frz_term_type, &
                                                            method_name_id, scf_guess_type
      LOGICAL                                            :: is_crystal, is_fast_dirty
      TYPE(molecular_scf_guess_env_type), POINTER        :: mscfg_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsection

      do_mol_loop = .FALSE.
      ! What kind of options are we using in the loop ?
      is_fast_dirty = .TRUE.
      is_crystal = .FALSE.

      NULLIFY (qs_env, mscfg_env, force_env_section, subsection)
      CALL force_env_get(force_env, force_env_section=force_env_section)
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)

      IF (method_name_id .EQ. do_qs) THEN

         CALL force_env_get(force_env, qs_env=qs_env)
         IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("mscfg_methods.F",396)

         CALL get_qs_env(qs_env, mscfg_env=mscfg_env)
         IF(.NOT.(ASSOCIATED(mscfg_env)))CALL cp__a("mscfg_methods.F",399)

         !!!! RZK-warning: All decisions are based on the values of input keywords
         !!!! The real danger is that many of these keywords might not be even
         !!!! in control of the job. They might be simply present in the input
         !!!! This section must be re-writen more accurately

         ! check ALMO SCF guess option
         NULLIFY (subsection)
         subsection => section_vals_get_subs_vals(force_env_section, "DFT%ALMO_SCF")
         CALL section_vals_val_get(subsection, "ALMO_SCF_GUESS", i_val=almo_guess_type)

         ! check SCF guess option
         NULLIFY (subsection)
         subsection => section_vals_get_subs_vals(force_env_section, "DFT%SCF")
         CALL section_vals_val_get(subsection, "SCF_GUESS", i_val=scf_guess_type)

         ! check ALMO EDA options
         NULLIFY (subsection)
         !!!LATER subsection    => section_vals_get_subs_vals(force_env_section,"DFT%ALMO_SCF%ALMO_DA")
         !!!LATER CALL section_vals_val_get(subsection,"FRZ_TERM",i_val=frz_term_type)
         frz_term_type = almo_frz_none

         ! Are we doing the loop ?
         IF (scf_guess_type .EQ. molecular_guess .OR. & ! SCF guess is molecular
             almo_guess_type .EQ. molecular_guess .OR. & ! ALMO SCF guess is molecular
             frz_term_type .NE. almo_frz_none) THEN ! ALMO FRZ term is requested

            do_mol_loop = .TRUE.

            ! If we are calculating molecular guess it is OK to do fast and dirty loop
            ! It is NOT ok to be sloppy with ALMO EDA calculations of the FRZ term
            IF (frz_term_type .NE. almo_frz_none) THEN
               is_fast_dirty = .FALSE.
               IF (frz_term_type .EQ. almo_frz_crystal) THEN
                  is_crystal = .TRUE.
               ENDIF
            ENDIF

         ENDIF

         mscfg_env%is_fast_dirty = is_fast_dirty
         mscfg_env%is_crystal = is_crystal

      END IF

      RETURN

   END FUNCTION do_mol_loop

END MODULE mscfg_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains typo and related routines to handle parameters controlling the
!>      GEO_OPT module
!> \par History
!>      none
! **************************************************************************************************
MODULE gopt_param_types

   USE input_constants,                 ONLY: default_bfgs_method_id,&
                                              default_cell_method_id,&
                                              default_cg_method_id,&
                                              default_lbfgs_method_id,&
                                              default_minimization_method_id,&
                                              default_ts_method_id,&
                                              none_ts_method_id
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gopt_param_types'
   INTEGER, PRIVATE, SAVE :: last_gopt_param_id = 0

   PUBLIC :: gopt_param_type
   PUBLIC :: gopt_param_read, &
             gopt_param_release

! **************************************************************************************************
!> \brief stores the parameters nedeed for a geometry optimization
!> \param method_id the optimization method
!> \param wanted_proj_gradient the wanted accuracy on the projected gradient
!> \param wanted_rel_f_error :
!> \param max_h_rank maximum rank of the hessian approximation
!> \param max_f_per_iter maximum number of function evaluations per iteration
!> \param max_iter maximum number of iterations
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cg_ls_param_type
      INTEGER         :: brent_max_iter, type_id
      REAL(KIND=dp) :: brent_tol, brack_limit, initial_step, max_step
      LOGICAL         :: grad_only
   END TYPE cg_ls_param_type

! **************************************************************************************************
   TYPE gopt_param_type
      LOGICAL                  :: Fletcher_Reeves
      INTEGER                  :: id_nr, ref_count
      INTEGER                  :: method_id, type_id
      INTEGER                  :: ts_method_id, cell_method_id, shellcore_method_id
      INTEGER                  :: max_f_per_iter, max_iter, max_h_rank
      INTEGER                  :: max_steep_steps
      REAL(KIND=dp)          :: restart_limit
      REAL(KIND=dp)          :: wanted_proj_gradient, wanted_rel_f_error
      REAL(KIND=dp)          :: max_dr, max_force, rms_dr, rms_force
      REAL(KIND=dp)          :: dimer_angle_tol
      TYPE(cg_ls_param_type)   :: cg_ls
   END TYPE gopt_param_type

CONTAINS

! **************************************************************************************************
!> \brief creates a new gopt_param object
!> \param gopt_param the object to create
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE gopt_param_create(gopt_param)
      TYPE(gopt_param_type), POINTER                     :: gopt_param

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_param_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (gopt_param)

      last_gopt_param_id = last_gopt_param_id+1
      gopt_param%id_nr = last_gopt_param_id
      gopt_param%ref_count = 1
   END SUBROUTINE gopt_param_create

! **************************************************************************************************
!> \brief reads the parameters of the geopmetry optimizer
!> \param gopt_param ...
!> \param gopt_section ...
!> \param type_id ...
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE gopt_param_read(gopt_param, gopt_section, type_id)
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(section_vals_type), POINTER                   :: gopt_section
      INTEGER, INTENT(IN), OPTIONAL                      :: type_id

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_param_read', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(gopt_param)))CALL cp__a("motion/gopt_param_types.F",109)
      IF(.NOT.(ASSOCIATED(gopt_section)))CALL cp__a("motion/gopt_param_types.F",110)
      CALL gopt_param_create(gopt_param)

      IF (PRESENT(type_id)) THEN
         gopt_param%type_id = type_id
      ELSE
         CALL section_vals_val_get(gopt_section, "TYPE", i_val=gopt_param%type_id)
      END IF
      CALL section_vals_val_get(gopt_section, "OPTIMIZER", i_val=gopt_param%method_id)

      CALL section_vals_val_get(gopt_section, "MAX_ITER", i_val=gopt_param%max_iter)
      CALL section_vals_val_get(gopt_section, "MAX_DR", r_val=gopt_param%max_dr)
      CALL section_vals_val_get(gopt_section, "MAX_FORCE", r_val=gopt_param%max_force)
      CALL section_vals_val_get(gopt_section, "RMS_DR", r_val=gopt_param%rms_dr)
      CALL section_vals_val_get(gopt_section, "RMS_FORCE", r_val=gopt_param%rms_force)

      SELECT CASE (gopt_param%method_id)
      CASE (default_lbfgs_method_id)
         CALL section_vals_val_get(gopt_section, "LBFGS%MAX_H_RANK", i_val=gopt_param%max_h_rank)
         CALL section_vals_val_get(gopt_section, "LBFGS%MAX_F_PER_ITER", i_val=gopt_param%max_f_per_iter)
         CALL section_vals_val_get(gopt_section, "LBFGS%WANTED_PROJ_GRADIENT", r_val=gopt_param%wanted_proj_gradient)
         CALL section_vals_val_get(gopt_section, "LBFGS%WANTED_REL_F_ERROR", r_val=gopt_param%wanted_rel_f_error)
      CASE (default_bfgs_method_id)
         ! Do nothing
      CASE (default_cg_method_id)
         CALL section_vals_val_get(gopt_section, "CG%FLETCHER_REEVES", l_val=gopt_param%Fletcher_Reeves)
         CALL section_vals_val_get(gopt_section, "CG%MAX_STEEP_STEPS", i_val=gopt_param%max_steep_steps)
         CALL section_vals_val_get(gopt_section, "CG%RESTART_LIMIT", r_val=gopt_param%restart_limit)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%TYPE", i_val=gopt_param%cg_ls%type_id)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%GOLD%INITIAL_STEP", &
                                   r_val=gopt_param%cg_ls%initial_step)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%GOLD%BRENT_TOL", &
                                   r_val=gopt_param%cg_ls%brent_tol)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%GOLD%BRENT_MAX_ITER", &
                                   i_val=gopt_param%cg_ls%brent_max_iter)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%GOLD%BRACK_LIMIT", &
                                   r_val=gopt_param%cg_ls%brack_limit)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%2PNT%MAX_ALLOWED_STEP", &
                                   r_val=gopt_param%cg_ls%max_step)
         CALL section_vals_val_get(gopt_section, "CG%LINE_SEARCH%2PNT%LINMIN_GRAD_ONLY", &
                                   l_val=gopt_param%cg_ls%grad_only)
      END SELECT

      SELECT CASE (gopt_param%type_id)
      CASE (default_minimization_method_id)
         ! Do Nothing
         gopt_param%ts_method_id = none_ts_method_id
      CASE (default_ts_method_id)
         CALL section_vals_val_get(gopt_section, "TRANSITION_STATE%METHOD", i_val=gopt_param%ts_method_id)
      CASE (default_cell_method_id)
         CALL section_vals_val_get(gopt_section, "TYPE", i_val=gopt_param%cell_method_id)
      END SELECT

   END SUBROUTINE gopt_param_read

! **************************************************************************************************
!> \brief ...
!> \param gopt_param the geometry optimization environment to retain
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE gopt_param_retain(gopt_param)
      TYPE(gopt_param_type), POINTER                     :: gopt_param

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_param_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(gopt_param)))CALL cp__a("motion/gopt_param_types.F",177)
      IF(.NOT.(gopt_param%ref_count > 0))CALL cp__a("motion/gopt_param_types.F",178)
      gopt_param%ref_count = gopt_param%ref_count+1
   END SUBROUTINE gopt_param_retain

! **************************************************************************************************
!> \brief ...
!> \param gopt_param the geometry optimization environment to release
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE gopt_param_release(gopt_param)
      TYPE(gopt_param_type), POINTER                     :: gopt_param

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_param_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(gopt_param)) THEN
         IF(.NOT.(gopt_param%ref_count > 0))CALL cp__a("motion/gopt_param_types.F",195)
         gopt_param%ref_count = gopt_param%ref_count-1
         IF (gopt_param%ref_count == 0) THEN
            DEALLOCATE (gopt_param)
         END IF
      END IF
   END SUBROUTINE gopt_param_release

END MODULE gopt_param_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains a functional that calculates the energy and its derivatives
!>      for the geometry optimizer
!> \par History
!>      03.2008 - Teodoro Laino [tlaino] - University of Zurich - Cell Optimization
! **************************************************************************************************
MODULE cell_opt_utils
   USE cell_types,                      ONLY: &
        cell_sym_cubic, cell_sym_hexagonal, cell_sym_monoclinic, cell_sym_orthorhombic, &
        cell_sym_rhombohedral, cell_sym_tetragonal_ab, cell_sym_tetragonal_ac, &
        cell_sym_tetragonal_bc, cell_sym_triclinic, cell_type, get_cell_param, set_cell_param
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_create,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_release,&
                                              cp_logger_set,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_units_rad
   USE input_cp2k_global,               ONLY: create_global_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: sqrt3
   USE mathlib,                         ONLY: angle
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cell_opt_utils'

   PUBLIC :: get_ut_cell_matrix, get_dg_dh, gopt_new_logger_create, &
             gopt_new_logger_release, read_external_press_tensor

CONTAINS
! **************************************************************************************************
!> \brief Transform a general cell matrix into an upper triangular one
!> \param cell ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE get_ut_cell_matrix(cell)
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'get_ut_cell_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(3)                              :: periodic
      REAL(KIND=dp), DIMENSION(3)                        :: cell_angle, cell_length

! orthorhombic cells are diagonal, and should stay exactly like this

      IF (.NOT. cell%orthorhombic) THEN
         CALL get_cell_param(cell, cell_length, cell_angle, cp_units_rad, periodic=periodic)
         CALL set_cell_param(cell, cell_length, cell_angle, periodic=periodic, &
                             do_init_cell=.TRUE.)
      END IF

   END SUBROUTINE get_ut_cell_matrix

! **************************************************************************************************
!> \brief creates a new logger used for cell optimization algorithm
!> \param new_logger ...
!> \param root_section ...
!> \param para_env ...
!> \param project_name ...
!> \param id_run ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE gopt_new_logger_create(new_logger, root_section, para_env, project_name, &
                                     id_run)
      TYPE(cp_logger_type), POINTER                      :: new_logger
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(len=default_string_length), INTENT(OUT)  :: project_name
      INTEGER, INTENT(IN)                                :: id_run

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gopt_new_logger_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: c_val, input_file_path, output_file_path
      CHARACTER(len=default_string_length)               :: label
      INTEGER                                            :: i, lp, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section

      NULLIFY (new_logger, logger, enum, keyword, section)
      logger => cp_get_default_logger()

      CALL create_global_section(section)
      keyword => section_get_keyword(section, "RUN_TYPE")
      CALL keyword_get(keyword, enum=enum)
      label = TRIM(enum_i2c(enum, id_run))
      CALL section_release(section)

      ! Redirecting output of the sub_calculation to a different file
      CALL section_vals_val_get(root_section, "GLOBAL%PROJECT_NAME", c_val=project_name)
      input_file_path = TRIM(project_name)
      lp = LEN_TRIM(input_file_path)
      i = logger%iter_info%iteration(logger%iter_info%n_rlevel)
      input_file_path(lp+1:LEN(input_file_path)) = "-"//TRIM(label)//"-"//ADJUSTL(cp_to_string(i))
      lp = LEN_TRIM(input_file_path)
      CALL section_vals_val_set(root_section, "GLOBAL%PROJECT_NAME", c_val=input_file_path(1:lp))
      CALL section_vals_val_set(root_section, "GLOBAL%RUN_TYPE", i_val=id_run)

      ! Redirecting output into a new file
      output_file_path = input_file_path(1:lp)//".out"
      IF (para_env%mepos == para_env%source) THEN
         CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                        file_action="WRITE", file_position="APPEND", unit_number=unit_nr)
      ELSE
         unit_nr = -1
      END IF
      CALL cp_logger_create(new_logger, para_env=para_env, default_global_unit_nr=unit_nr, &
                            close_global_unit_on_dealloc=.FALSE.)
      CALL section_vals_val_get(root_section, "GLOBAL%PROJECT", c_val=c_val)
      IF (c_val /= "") THEN
         CALL cp_logger_set(new_logger, local_filename=TRIM(c_val)//"_localLog")
      END IF
      new_logger%iter_info%project_name = c_val
      CALL section_vals_val_get(root_section, "GLOBAL%PRINT_LEVEL", &
                                i_val=new_logger%iter_info%print_level)

   END SUBROUTINE gopt_new_logger_create

! **************************************************************************************************
!> \brief releases a new logger used for cell optimization algorithm
!> \param new_logger ...
!> \param root_section ...
!> \param para_env ...
!> \param project_name ...
!> \param id_run ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE gopt_new_logger_release(new_logger, root_section, para_env, project_name, id_run)
      TYPE(cp_logger_type), POINTER                      :: new_logger
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(len=default_string_length), INTENT(IN)   :: project_name
      INTEGER, INTENT(IN)                                :: id_run

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gopt_new_logger_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr

      IF (para_env%mepos == para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(new_logger)
         CALL close_file(unit_number=unit_nr)
      END IF
      CALL cp_logger_release(new_logger)
      CALL section_vals_val_set(root_section, "GLOBAL%RUN_TYPE", i_val=id_run)
      CALL section_vals_val_set(root_section, "GLOBAL%PROJECT_NAME", c_val=project_name)

   END SUBROUTINE gopt_new_logger_release

! **************************************************************************************************
!> \brief Reads the external pressure tensor
!> \param geo_section ...
!> \param cell ...
!> \param pres_ext ...
!> \param mtrx ...
!> \param rot ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE read_external_press_tensor(geo_section, cell, pres_ext, mtrx, rot)
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(OUT)                         :: pres_ext
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: mtrx
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: rot

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_external_press_tensor', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind, j
      LOGICAL                                            :: check
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pres_ext_tens
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pvals

      NULLIFY (pvals)
      mtrx = 0.0_dp
      pres_ext_tens = 0.0_dp
      pres_ext = 0.0_dp
      CALL section_vals_val_get(geo_section, "EXTERNAL_PRESSURE", r_vals=pvals)
      check = (SIZE(pvals) == 1) .OR. (SIZE(pvals) == 9)
      IF (.NOT. check) &
         CALL cp__b("motion/cell_opt_utils.F",209,"EXTERNAL_PRESSURE can have 1 or 9 components only!")

      IF (SIZE(pvals) == 9) THEN
         ind = 0
         DO i = 1, 3
            DO j = 1, 3
               ind = ind+1
               pres_ext_tens(j, i) = pvals(ind)
            END DO
         END DO
         ! Also the pressure tensor must be oriented in the same canonical directions
         ! of the simulation cell
         pres_ext_tens = MATMUL(TRANSPOSE(rot), pres_ext_tens)
         DO i = 1, 3
            pres_ext = pres_ext+pres_ext_tens(i, i)
         ENDDO
         pres_ext = pres_ext/3.0_dp
         DO i = 1, 3
            pres_ext_tens(i, i) = pres_ext_tens(i, i)-pres_ext
         ENDDO
      ELSE
         pres_ext = pvals(1)
      END IF

      IF (ANY(pres_ext_tens > 1.0E-5_dp)) THEN
         mtrx = cell%deth*MATMUL(cell%h_inv, MATMUL(pres_ext_tens, TRANSPOSE(cell%h_inv)))
      END IF

   END SUBROUTINE read_external_press_tensor

! **************************************************************************************************
!> \brief Computes the derivatives for the cell
!> \param gradient ...
!> \param av_ptens ...
!> \param pres_ext ...
!> \param cell ...
!> \param mtrx ...
!> \param keep_angles ...
!> \param keep_symmetry ...
!> \param pres_int ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE get_dg_dh(gradient, av_ptens, pres_ext, cell, mtrx, keep_angles, &
                        keep_symmetry, pres_int)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: gradient
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: av_ptens
      REAL(KIND=dp), INTENT(IN)                          :: pres_ext
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: mtrx
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_angles, keep_symmetry
      REAL(KIND=dp), INTENT(OUT)                         :: pres_int

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_dg_dh', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_keep_angles, my_keep_symmetry
      REAL(KIND=dp)                                      :: a, cosa, cosah, g, norm, norm_b, norm_c, &
                                                            sina, sinah
      REAL(KIND=dp), DIMENSION(3, 3)                     :: correction, pten_hinv_old, ptens

      my_keep_angles = .FALSE.
      IF (PRESENT(keep_angles)) my_keep_angles = keep_angles
      my_keep_symmetry = .FALSE.
      IF (PRESENT(keep_symmetry)) my_keep_symmetry = keep_symmetry
      gradient = 0.0_dp
      ptens = av_ptens

      ! Evaluating the internal pressure
      pres_int = 0.0_dp
      DO i = 1, 3
         pres_int = pres_int+ptens(i, i)
      ENDDO
      pres_int = pres_int/3.0_dp

      ptens(1, 1) = av_ptens(1, 1)-pres_ext
      ptens(2, 2) = av_ptens(2, 2)-pres_ext
      ptens(3, 3) = av_ptens(3, 3)-pres_ext

      pten_hinv_old = cell%deth*MATMUL(cell%h_inv, ptens)
      correction = MATMUL(mtrx, cell%hmat)

      gradient(1) = pten_hinv_old(1, 1)-correction(1, 1)
      gradient(2) = pten_hinv_old(2, 1)-correction(2, 1)
      gradient(3) = pten_hinv_old(2, 2)-correction(2, 2)
      gradient(4) = pten_hinv_old(3, 1)-correction(3, 1)
      gradient(5) = pten_hinv_old(3, 2)-correction(3, 2)
      gradient(6) = pten_hinv_old(3, 3)-correction(3, 3)

      IF (my_keep_angles) THEN
         ! If we want to keep the angles constant we have to project out the
         ! components of the cell angles
         norm_b = DOT_PRODUCT(cell%hmat(:, 2), cell%hmat(:, 2))
         norm = DOT_PRODUCT(cell%hmat(1:2, 2), gradient(2:3))
         gradient(2:3) = cell%hmat(1:2, 2)/norm_b*norm
         norm_c = DOT_PRODUCT(cell%hmat(:, 3), cell%hmat(:, 3))
         norm = DOT_PRODUCT(cell%hmat(1:3, 3), gradient(4:6))
         gradient(4:6) = cell%hmat(1:3, 3)/norm_c*norm
         ! Retain an exact orthorhombic cell
         ! (off-diagonal elements must remain zero identically to keep QS fast)
         IF (cell%orthorhombic) THEN
            gradient(2) = 0.0_dp
            gradient(4) = 0.0_dp
            gradient(5) = 0.0_dp
         END IF
      END IF

      IF (my_keep_symmetry) THEN
         SELECT CASE (cell%symmetry_id)
         CASE (cell_sym_cubic, &
               cell_sym_tetragonal_ab, &
               cell_sym_tetragonal_ac, &
               cell_sym_tetragonal_bc, &
               cell_sym_orthorhombic)
            SELECT CASE (cell%symmetry_id)
            CASE (cell_sym_cubic)
               g = (gradient(1)+gradient(3)+gradient(6))/3.0_dp
               gradient(1) = g
               gradient(3) = g
               gradient(6) = g
            CASE (cell_sym_tetragonal_ab, &
                  cell_sym_tetragonal_ac, &
                  cell_sym_tetragonal_bc)
               SELECT CASE (cell%symmetry_id)
               CASE (cell_sym_tetragonal_ab)
                  g = 0.5_dp*(gradient(1)+gradient(3))
                  gradient(1) = g
                  gradient(3) = g
               CASE (cell_sym_tetragonal_ac)
                  g = 0.5_dp*(gradient(1)+gradient(6))
                  gradient(1) = g
                  gradient(6) = g
               CASE (cell_sym_tetragonal_bc)
                  g = 0.5_dp*(gradient(3)+gradient(6))
                  gradient(3) = g
                  gradient(6) = g
               END SELECT
            CASE (cell_sym_orthorhombic)
               ! Nothing else to do
            END SELECT
            gradient(2) = 0.0_dp
            gradient(4) = 0.0_dp
            gradient(5) = 0.0_dp
         CASE (cell_sym_hexagonal)
            g = 0.5_dp*(gradient(1)+0.5_dp*(gradient(2)+sqrt3*gradient(3)))
            gradient(1) = g
            gradient(2) = 0.5_dp*g
            gradient(3) = sqrt3*gradient(2)
            gradient(4) = 0.0_dp
            gradient(5) = 0.0_dp
         CASE (cell_sym_rhombohedral)
            a = (angle(cell%hmat(:, 3), cell%hmat(:, 2))+ &
                 angle(cell%hmat(:, 1), cell%hmat(:, 3))+ &
                 angle(cell%hmat(:, 1), cell%hmat(:, 2)))/3.0_dp
            cosa = COS(a)
            sina = SIN(a)
            cosah = COS(0.5_dp*a)
            sinah = SIN(0.5_dp*a)
            norm = cosa/cosah
            norm_c = SQRT(1.0_dp-norm*norm)
            g = (gradient(1)+gradient(2)*cosa+gradient(3)*sina+ &
                 gradient(4)*cosah*norm+gradient(5)*sinah*norm+gradient(6)*norm_c)/3.0_dp
            gradient(1) = g
            gradient(2) = g*cosa
            gradient(3) = g*sina
            gradient(4) = g*cosah*norm
            gradient(5) = g*sinah*norm
            gradient(6) = g*norm_c
         CASE (cell_sym_monoclinic)
            gradient(2) = 0.0_dp
            gradient(5) = 0.0_dp
         CASE (cell_sym_triclinic)
            ! Nothing to do
         END SELECT
      END IF

      gradient = -gradient

   END SUBROUTINE get_dg_dh

END MODULE cell_opt_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains type used for a Simulation Cell Optimization
!> \par History
!>      none
!> \author Teodoro Laino - created [tlaino] - 03.2008 - Zurich University
! **************************************************************************************************
MODULE cell_opt_types
   USE cell_opt_utils,                  ONLY: get_ut_cell_matrix,&
                                              read_external_press_tensor
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE particle_list_types,             ONLY: particle_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cell_opt_types'
   INTEGER, PRIVATE, SAVE :: last_cell_opt_env_id = 0

   PUBLIC :: cell_opt_env_type, &
             cell_opt_env_create, &
             cell_opt_env_release

! **************************************************************************************************
!> \brief Type containing all informations abour the simulation cell optimization
!> \par History
!>      none
!> \author Teodoro Laino - created [tlaino] - 03.2008 - Zurich University
! **************************************************************************************************
   TYPE cell_opt_env_type
      ! Simulation cell optimization parameters
      INTEGER                                    :: ref_count, id_nr
      LOGICAL                                    :: keep_angles, &
                                                    keep_symmetry
      REAL(KIND=dp)                              :: pres_ext, pres_int, pres_tol
      REAL(KIND=dp), DIMENSION(3, 3)              :: mtrx
      REAL(KIND=dp), DIMENSION(3, 3)              :: rot_matrix
      TYPE(cell_type), POINTER                   :: ref_cell
   END TYPE cell_opt_env_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cell_env ...
!> \param force_env ...
!> \param geo_section ...
!> \par History
!>      none
!> \author Teodoro Laino - created [tlaino] - 03.2008 - Zurich University
! **************************************************************************************************
   SUBROUTINE cell_opt_env_create(cell_env, force_env, geo_section)
      TYPE(cell_opt_env_type), POINTER                   :: cell_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: geo_section

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_opt_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp), DIMENSION(3)                        :: r
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles

      IF(.NOT.(.NOT. ASSOCIATED(cell_env)))CALL cp__a("motion/cell_opt_types.F",82)
      ALLOCATE (cell_env)
      NULLIFY (cell_env%ref_cell, cell, subsys, particles)
      cell_env%ref_count = 1
      last_cell_opt_env_id = last_cell_opt_env_id+1
      cell_env%id_nr = last_cell_opt_env_id
      CALL force_env_get(force_env, cell=cell, subsys=subsys)
      CALL cell_create(cell_env%ref_cell)
      CALL cell_clone(cell, cell_env%ref_cell)
      CALL section_vals_val_get(geo_section, "KEEP_ANGLES", l_val=cell_env%keep_angles)
      CALL section_vals_val_get(geo_section, "KEEP_SYMMETRY", l_val=cell_env%keep_symmetry)
      CALL section_vals_val_get(geo_section, "PRESSURE_TOLERANCE", r_val=cell_env%pres_tol)

      ! First let's rotate the cell vectors in order to have an upper triangular matrix.
      CALL get_ut_cell_matrix(cell)

      ! Compute the rotation matrix that give the cell vectors in the "canonical" orientation
      cell_env%rot_matrix = MATMUL(cell_env%ref_cell%hmat, cell%h_inv)

      ! Get the external pressure
      CALL read_external_press_tensor(geo_section, cell, cell_env%pres_ext, cell_env%mtrx, &
                                      cell_env%rot_matrix)

      ! Rotate particles accordingly
      CALL cp_subsys_get(subsys, particles=particles)
      DO ip = 1, particles%n_els
         r = MATMUL(TRANSPOSE(cell_env%rot_matrix), particles%els(ip)%r)
         particles%els(ip)%r = r
      END DO
   END SUBROUTINE cell_opt_env_create

! **************************************************************************************************
!> \brief ...
!> \param cell_env ...
!> \par History
!>      none
!> \author Teodoro Laino - created [tlaino] - 03.2008 - Zurich University
! **************************************************************************************************
   SUBROUTINE cell_opt_env_release(cell_env)
      TYPE(cell_opt_env_type), POINTER                   :: cell_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_opt_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(cell_env)) THEN
         IF(.NOT.(cell_env%ref_count > 0))CALL cp__a("motion/cell_opt_types.F",127)
         cell_env%ref_count = cell_env%ref_count-1
         IF (cell_env%ref_count == 0) THEN
            CALL cell_release(cell_env%ref_cell)
            DEALLOCATE (cell_env)
         END IF
      END IF
   END SUBROUTINE cell_opt_env_release

END MODULE cell_opt_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains types used for a Dimer Method calculations
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
MODULE dimer_types

   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: do_first_rotation_step
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE parallel_rng_types,              ONLY: random_numbers
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dimer_types'
   INTEGER, PRIVATE, SAVE :: last_dimer_id = 0

   PUBLIC :: dimer_env_type, &
             dimer_env_create, &
             dimer_env_retain, &
             dimer_env_release, &
             dimer_fixed_atom_control

! **************************************************************************************************
!> \brief Type containing all informations abour the rotation of the Dimer
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   TYPE dimer_rotational_type
      ! Rotational parameters
      INTEGER                                    :: rotation_step
      LOGICAL                                    :: interpolate_gradient
      REAL(KIND=dp)                              :: angle_tol, angle1, angle2, dCdp, curvature
      REAL(KIND=dp), POINTER, DIMENSION(:)       :: g0, g1, g1p
   END TYPE dimer_rotational_type

! **************************************************************************************************
!> \brief Type containing all informations abour the translation of the Dimer
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   TYPE dimer_translational_type
      ! Translational parameters
      REAL(KIND=dp), POINTER, DIMENSION(:)       :: tls_vec
   END TYPE dimer_translational_type

! **************************************************************************************************
!> \brief Conjugate Directions type
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   TYPE dimer_cg_rot_type
      REAL(KIND=dp)                              :: norm_theta, norm_theta_old, norm_h
      REAL(KIND=dp), POINTER, DIMENSION(:)       :: nvec_old
   END TYPE dimer_cg_rot_type

! **************************************************************************************************
!> \brief Defines the environment for a Dimer Method calculation
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   TYPE dimer_env_type
      INTEGER                                    :: ref_count, id_nr
      REAL(KIND=dp)                              :: dr
      REAL(KIND=dp), POINTER, DIMENSION(:)       :: nvec
      TYPE(dimer_rotational_type)                :: rot
      TYPE(dimer_translational_type)             :: tsl
      TYPE(dimer_cg_rot_type)                    :: cg_rot
   END TYPE dimer_env_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param dimer_env ...
!> \param subsys ...
!> \param globenv ...
!> \param dimer_section ...
!> \par History
!>      2016/03/03 [LTong] changed input natom to subsys
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE dimer_env_create(dimer_env, subsys, globenv, dimer_section)
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: dimer_section

      CHARACTER(len=*), PARAMETER :: routineN = 'dimer_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, isize, j, k, n_rep_val, natom, unit_nr
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: norm, xval(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array
      TYPE(section_vals_type), POINTER                   :: nvec_section

      unit_nr = cp_logger_get_default_io_unit()
      IF(.NOT.(.NOT. ASSOCIATED(dimer_env)))CALL cp__a("motion/dimer_types.F",130)
      ALLOCATE (dimer_env)
      dimer_env%ref_count = 1
      last_dimer_id = last_dimer_id+1
      dimer_env%id_nr = last_dimer_id
      ! Setup NVEC
      NULLIFY (dimer_env%nvec, dimer_env%rot%g0, dimer_env%rot%g1, dimer_env%rot%g1p, &
               dimer_env%tsl%tls_vec)
      ! get natom
      CALL cp_subsys_get(subsys=subsys, natom=natom)
      ! Allocate the working arrays
      ALLOCATE (dimer_env%nvec(natom*3))
      ALLOCATE (dimer_env%rot%g0(natom*3))
      ALLOCATE (dimer_env%rot%g1(natom*3))
      ALLOCATE (dimer_env%rot%g1p(natom*3))
      ! Check if the dimer vector is available in the input or not..
      nvec_section => section_vals_get_subs_vals(dimer_section, "DIMER_VECTOR")
      CALL section_vals_get(nvec_section, explicit=explicit)
      IF (explicit) THEN
         IF (unit_nr > 0) WRITE (unit_nr, *) "Reading Dimer Vector from file!"
         NULLIFY (array)
         CALL section_vals_val_get(nvec_section, "_DEFAULT_KEYWORD_", n_rep_val=n_rep_val)
         isize = 0
         DO i = 1, n_rep_val
            CALL section_vals_val_get(nvec_section, "_DEFAULT_KEYWORD_", r_vals=array, i_rep_val=i)
            DO j = 1, SIZE(array)
               isize = isize+1
               dimer_env%nvec(isize) = array(j)
            END DO
         END DO
         IF(.NOT.(isize == SIZE(dimer_env%nvec)))CALL cp__a("motion/dimer_types.F",160)
      ELSE
         CALL random_numbers(dimer_env%nvec, globenv%gaussian_rng_stream)
      END IF
      ! Check for translation in the dimer vector and remove them
      IF (natom > 1) THEN
         xval = 0.0_dp
         DO j = 1, natom
            DO k = 1, 3
               i = (j-1)*3+k
               xval(k) = xval(k)+dimer_env%nvec(i)
            END DO
         END DO
         ! Subtract net translations
         xval = xval/REAL(natom*3, KIND=dp)
         DO j = 1, natom
            DO k = 1, 3
               i = (j-1)*3+k
               dimer_env%nvec(i) = dimer_env%nvec(i)-xval(k)
            END DO
         END DO
      END IF
      ! set nvec components to zero for the corresponding constraints
      CALL dimer_fixed_atom_control(dimer_env%nvec, subsys)
      norm = SQRT(SUM(dimer_env%nvec**2))
      IF (norm <= EPSILON(0.0_dp)) &
         CALL cp__b("motion/dimer_types.F",186,"The norm of the dimer vector is 0! Calculation cannot proceed further.")
      dimer_env%nvec = dimer_env%nvec/norm
      dimer_env%rot%rotation_step = do_first_rotation_step
      CALL section_vals_val_get(dimer_section, "DR", r_val=dimer_env%dr)
      CALL section_vals_val_get(dimer_section, "INTERPOLATE_GRADIENT", &
                                l_val=dimer_env%rot%interpolate_gradient)
      CALL section_vals_val_get(dimer_section, "ANGLE_TOLERANCE", &
                                r_val=dimer_env%rot%angle_tol)
      ! initialise values
      dimer_env%cg_rot%norm_h = 1.0_dp
      dimer_env%cg_rot%norm_theta = 0.0_dp
      dimer_env%cg_rot%norm_theta_old = 0.0_dp
      dimer_env%rot%g0 = 0.0_dp
      dimer_env%rot%g1 = 0.0_dp
      dimer_env%rot%g1p = 0.0_dp
      ALLOCATE (dimer_env%cg_rot%nvec_old(natom*3))
   END SUBROUTINE dimer_env_create

! **************************************************************************************************
!> \brief ...
!> \param dimer_env ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE dimer_env_retain(dimer_env)
      TYPE(dimer_env_type), POINTER                      :: dimer_env

      CHARACTER(len=*), PARAMETER :: routineN = 'dimer_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dimer_env)))CALL cp__a("motion/dimer_types.F",217)
      IF(.NOT.(dimer_env%ref_count > 0))CALL cp__a("motion/dimer_types.F",218)
      dimer_env%ref_count = dimer_env%ref_count+1
   END SUBROUTINE dimer_env_retain

! **************************************************************************************************
!> \brief ...
!> \param dimer_env ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE dimer_env_release(dimer_env)
      TYPE(dimer_env_type), POINTER                      :: dimer_env

      CHARACTER(len=*), PARAMETER :: routineN = 'dimer_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dimer_env)) THEN
         IF(.NOT.(dimer_env%ref_count > 0))CALL cp__a("motion/dimer_types.F",236)
         dimer_env%ref_count = dimer_env%ref_count-1
         IF (dimer_env%ref_count == 0) THEN
            IF (ASSOCIATED(dimer_env%nvec)) THEN
               DEALLOCATE (dimer_env%nvec)
            END IF
            IF (ASSOCIATED(dimer_env%rot%g0)) THEN
               DEALLOCATE (dimer_env%rot%g0)
            END IF
            IF (ASSOCIATED(dimer_env%rot%g1)) THEN
               DEALLOCATE (dimer_env%rot%g1)
            END IF
            IF (ASSOCIATED(dimer_env%rot%g1p)) THEN
               DEALLOCATE (dimer_env%rot%g1p)
            END IF
            IF (ASSOCIATED(dimer_env%cg_rot%nvec_old)) THEN
               DEALLOCATE (dimer_env%cg_rot%nvec_old)
            END IF
            ! No need to deallocate tls_vec (just a pointer to aother local array)
            NULLIFY (dimer_env%tsl%tls_vec)
            DEALLOCATE (dimer_env)
         END IF
      END IF
   END SUBROUTINE dimer_env_release

! **************************************************************************************************
!> \brief Set parts of a given array vec to zero according to fixed atom constraints.
!>        When atoms are (partially) fixed then the relevant components of
!>        nvec should be set to zero.  Furthermore, the relevent components
!>        of the gradient in CG should also be set to zero.
!> \param vec : vector to be modified
!> \param subsys : subsys type object used by CP2k
!> \par History
!>      2016/03/03 [LTong] created
!> \author Lianheng Tong [LTong]
! **************************************************************************************************
   SUBROUTINE dimer_fixed_atom_control(vec, subsys)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: vec
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'dimer_fixed_atom_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, ikind, ind, iparticle, nfixed_atoms, &
                                                            nkinds
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (molecule_kinds, molecule_kind, fixd_list)

      ! need to get constraint information from molecule information
      CALL cp_subsys_get(subsys=subsys, &
                         molecule_kinds_new=molecule_kinds)
      molecule_kind_set => molecule_kinds%els

      ! get total number of fixed atoms
      ! nkinds is the kinds of molecules, not atoms
      nkinds = molecule_kinds%n_els
      DO ikind = 1, nkinds
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, &
                                nfixd=nfixed_atoms, &
                                fixd_list=fixd_list)
         IF (ASSOCIATED(fixd_list)) THEN
            DO ii = 1, nfixed_atoms
               IF (.NOT. fixd_list(ii)%restraint%active) THEN
                  iparticle = fixd_list(ii)%fixd
                  ind = (iparticle-1)*3
                  ! apply constraint to nvec
                  SELECT CASE (fixd_list (ii)%itype)
                  CASE (use_perd_x)
                     vec(ind+1) = 0.0_dp
                  CASE (use_perd_y)
                     vec(ind+2) = 0.0_dp
                  CASE (use_perd_z)
                     vec(ind+3) = 0.0_dp
                  CASE (use_perd_xy)
                     vec(ind+1) = 0.0_dp
                     vec(ind+2) = 0.0_dp
                  CASE (use_perd_xz)
                     vec(ind+1) = 0.0_dp
                     vec(ind+3) = 0.0_dp
                  CASE (use_perd_yz)
                     vec(ind+2) = 0.0_dp
                     vec(ind+3) = 0.0_dp
                  CASE (use_perd_xyz)
                     vec(ind+1) = 0.0_dp
                     vec(ind+2) = 0.0_dp
                     vec(ind+3) = 0.0_dp
                  END SELECT
               END IF ! .NOT.fixd_list(ii)%restraint%active
            END DO ! ii
         END IF ! ASSOCIATED(fixd_list)
      END DO ! ikind
   END SUBROUTINE dimer_fixed_atom_control

END MODULE dimer_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains a functional that calculates the energy and its derivatives
!>      for the geometry optimizer
!> \par History
!>      01.2008 - Luca Bellucci and Teodoro Laino - Generalizing for Dimer Method.
!>      03.2008 - Teodoro Laino [tlaino] - University of Zurich - Cell Optimization
! **************************************************************************************************
MODULE gopt_f_types
   USE cell_opt_types,                  ONLY: cell_opt_env_create,&
                                              cell_opt_env_release,&
                                              cell_opt_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE dimer_types,                     ONLY: dimer_env_create,&
                                              dimer_env_release,&
                                              dimer_env_retain,&
                                              dimer_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_get_natom,&
                                              force_env_release,&
                                              force_env_retain,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_param_types,                ONLY: gopt_param_read,&
                                              gopt_param_release,&
                                              gopt_param_type
   USE input_constants,                 ONLY: default_cell_method_id,&
                                              default_dimer_method_id,&
                                              default_minimization_method_id,&
                                              default_shellcore_method_id,&
                                              default_ts_method_id
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE particle_list_types,             ONLY: particle_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gopt_f_types'
   INTEGER, PRIVATE, SAVE :: last_gopt_f_id = 0

   PUBLIC :: gopt_f_type, gopt_f_create, gopt_f_retain, gopt_f_release

! **************************************************************************************************
!> \brief calculates the potential energy of a system, and its derivatives
!> \par History
!>      none
! **************************************************************************************************
   TYPE gopt_f_type
      INTEGER                                  :: ref_count, id_nr
      INTEGER                                  :: nfree
      INTEGER                                  :: type_id, ts_method_id, cell_method_id, shellcore_method_id
      LOGICAL                                  :: dimer_rotation, do_line_search, eval_opt_geo
      CHARACTER(LEN=default_string_length)     :: label, tag
      TYPE(force_env_type), POINTER            :: force_env
      TYPE(global_environment_type), POINTER   :: globenv
      ! Motion section must be references only for IO of the MOTION%PRINT..
      TYPE(section_vals_type), POINTER         :: motion_section, geo_section
      TYPE(dimer_env_type), POINTER            :: dimer_env
      TYPE(gopt_f_type), POINTER               :: gopt_dimer_env
      TYPE(gopt_param_type), POINTER           :: gopt_dimer_param
      TYPE(cell_opt_env_type), POINTER         :: cell_env
      REAL(KIND=dp), DIMENSION(3, 3)            :: h_ref
      LOGICAL                                  :: require_consistent_energy_force
   END TYPE gopt_f_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gopt_env the geometry optimization environment to be created
!>      force_env:
!> \param gopt_param ...
!> \param force_env ...
!> \param globenv ...
!> \param geo_opt_section ...
!> \param eval_opt_geo ...
!> \par History
!>      none
! **************************************************************************************************
   RECURSIVE SUBROUTINE gopt_f_create(gopt_env, gopt_param, force_env, globenv, geo_opt_section, &
                                      eval_opt_geo)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: geo_opt_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: eval_opt_geo

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: natom, nshell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles, shell_particles
      TYPE(section_vals_type), POINTER                   :: dimer_section, rot_opt_section

      IF(.NOT.(.NOT. ASSOCIATED(gopt_env)))CALL cp__a("motion/gopt_f_types.F",107)
      ALLOCATE (gopt_env)
      nshell = 0

      NULLIFY (gopt_env%dimer_env, gopt_env%gopt_dimer_env, gopt_env%gopt_dimer_param, gopt_env%cell_env)
      gopt_env%ref_count = 1
      last_gopt_f_id = last_gopt_f_id+1
      gopt_env%id_nr = last_gopt_f_id
      gopt_env%dimer_rotation = .FALSE.
      gopt_env%do_line_search = .FALSE.
      CALL force_env_retain(force_env)
      gopt_env%force_env => force_env
      gopt_env%motion_section => section_vals_get_subs_vals(force_env%root_section, "MOTION")
      gopt_env%geo_section => geo_opt_section
      gopt_env%globenv => globenv
      gopt_env%eval_opt_geo = .TRUE.
      IF (PRESENT(eval_opt_geo)) gopt_env%eval_opt_geo = eval_opt_geo
      gopt_env%require_consistent_energy_force = .TRUE.

      CALL force_env_get(force_env, subsys=subsys)
      gopt_env%type_id = gopt_param%type_id
      SELECT CASE (gopt_env%type_id)
      CASE (default_ts_method_id, default_minimization_method_id)
         CALL cp_subsys_get(subsys, &
                            particles=particles, &
                            shell_particles=shell_particles)
         IF (ASSOCIATED(shell_particles)) nshell = shell_particles%n_els
         ! The same number of shell and core particles is assumed
         gopt_env%nfree = particles%n_els+nshell
         gopt_env%label = "GEO_OPT"
         gopt_env%tag = "GEOMETRY"
         SELECT CASE (gopt_param%type_id)
         CASE (default_ts_method_id)
            gopt_env%ts_method_id = gopt_param%ts_method_id
            SELECT CASE (gopt_param%ts_method_id)
            CASE (default_dimer_method_id)
               ! For the Dimer method we use the same framework of geometry optimizers
               ! already defined for cp2k..
               natom = force_env_get_natom(force_env)
               dimer_section => section_vals_get_subs_vals(geo_opt_section, "TRANSITION_STATE%DIMER")
               CALL dimer_env_create(gopt_env%dimer_env, subsys, globenv, dimer_section)

               ! Setup the GEO_OPT environment for the rotation of the Dimer
               rot_opt_section => section_vals_get_subs_vals(dimer_section, "ROT_OPT")
               CALL gopt_param_read(gopt_env%gopt_dimer_param, rot_opt_section, &
                                    type_id=default_minimization_method_id)
               gopt_env%gopt_dimer_param%type_id = default_ts_method_id

               CALL gopt_f_create(gopt_env%gopt_dimer_env, gopt_env%gopt_dimer_param, force_env=force_env, &
                                  globenv=globenv, geo_opt_section=rot_opt_section, eval_opt_geo=eval_opt_geo)
               CALL dimer_env_retain(gopt_env%dimer_env)
               gopt_env%gopt_dimer_env%dimer_env => gopt_env%dimer_env
               gopt_env%gopt_dimer_env%label = "ROT_OPT"
               gopt_env%gopt_dimer_env%dimer_rotation = .TRUE.
            END SELECT
         END SELECT
      CASE (default_cell_method_id)
         gopt_env%nfree = 6
         gopt_env%label = "CELL_OPT"
         gopt_env%tag = "  CELL  "
         gopt_env%cell_method_id = gopt_param%cell_method_id
         CALL cell_opt_env_create(gopt_env%cell_env, force_env, gopt_env%geo_section)
      CASE (default_shellcore_method_id)
         gopt_env%nfree = subsys%shell_particles%n_els
         gopt_env%label = "SHELL_OPT"
         gopt_env%tag = "  SHELL-CORE  "
         gopt_env%shellcore_method_id = gopt_param%shellcore_method_id
      END SELECT
   END SUBROUTINE gopt_f_create

! **************************************************************************************************
!> \brief ...
!> \param gopt_env the geometry optimization environment to retain
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE gopt_f_retain(gopt_env)
      TYPE(gopt_f_type), POINTER                         :: gopt_env

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(gopt_env)))CALL cp__a("motion/gopt_f_types.F",188)
      IF(.NOT.(gopt_env%ref_count > 0))CALL cp__a("motion/gopt_f_types.F",189)
      gopt_env%ref_count = gopt_env%ref_count+1
   END SUBROUTINE gopt_f_retain

! **************************************************************************************************
!> \brief ...
!> \param gopt_env the geometry optimization environment to release
!> \par History
!>      none
! **************************************************************************************************
   RECURSIVE SUBROUTINE gopt_f_release(gopt_env)
      TYPE(gopt_f_type), POINTER                         :: gopt_env

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(gopt_env)) THEN
         IF(.NOT.(gopt_env%ref_count > 0))CALL cp__a("motion/gopt_f_types.F",205)
         gopt_env%ref_count = gopt_env%ref_count-1
         IF (gopt_env%ref_count == 0) THEN
            CALL force_env_release(gopt_env%force_env)
            NULLIFY (gopt_env%force_env, &
                     gopt_env%globenv, &
                     gopt_env%motion_section, &
                     gopt_env%geo_section)
            CALL cell_opt_env_release(gopt_env%cell_env)
            CALL dimer_env_release(gopt_env%dimer_env)
            CALL gopt_f_release(gopt_env%gopt_dimer_env)
            CALL gopt_param_release(gopt_env%gopt_dimer_param)
            DEALLOCATE (gopt_env)
         END IF
      END IF
   END SUBROUTINE gopt_f_release

END MODULE gopt_f_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains utilities for a Dimer Method calculations
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
MODULE dimer_utils
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE dimer_types,                     ONLY: dimer_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dimer_utils'

   PUBLIC :: rotate_dimer, update_dimer_vec, get_theta
   REAL(KIND=dp), PARAMETER, PUBLIC     :: dimer_thrs = EPSILON(0.0_dp)*1.0E4_dp

CONTAINS

! **************************************************************************************************
!> \brief Performs a rotation of the unit dimer vector
!> \param nvec ...
!> \param theta ...
!> \param dt ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE rotate_dimer(nvec, theta, dt)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: nvec, theta
      REAL(KIND=dp)                                      :: dt

      CHARACTER(len=*), PARAMETER :: routineN = 'rotate_dimer', routineP = moduleN//':'//routineN

      INTEGER                                            :: output_unit
      LOGICAL                                            :: check

      output_unit = cp_logger_get_default_io_unit()

      ! Orthogonality check for the rotation..
      check = ABS(DOT_PRODUCT(nvec, theta)) < MAX(1.0E-9_dp, dimer_thrs)
      IF (.NOT. check .AND. (output_unit > 0)) THEN
         WRITE (output_unit, *) "NVEC and THETA should be orthogonal! Residue: ", &
            ABS(DOT_PRODUCT(nvec, theta))
      END IF
      IF(.NOT.(check))CALL cp__a("motion/dimer_utils.F",60)
      nvec = nvec*COS(dt)+theta*SIN(dt)

   END SUBROUTINE rotate_dimer

! **************************************************************************************************
!> \brief Updates the orientation of the dimer vector in the input file
!> \param dimer_env ...
!> \param motion_section ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE update_dimer_vec(dimer_env, motion_section)
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      TYPE(section_vals_type), POINTER                   :: motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'update_dimer_vec', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i_rep_val, isize, j, size_array
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array
      TYPE(section_vals_type), POINTER                   :: nvec_section

      nvec_section => section_vals_get_subs_vals(motion_section, &
                                                 "GEO_OPT%TRANSITION_STATE%DIMER%DIMER_VECTOR")
      ! Clean the content of the section first..
      CALL section_vals_remove_values(nvec_section)
      ! Fill in the section with the present values..
      size_array = 6
      isize = 0
      i_rep_val = 0
      Main_Loop: DO i = 1, SIZE(dimer_env%nvec), size_array
         ALLOCATE (array(size_array))
         i_rep_val = i_rep_val+1
         DO j = 1, size_array
            isize = isize+1
            array(j) = dimer_env%nvec(isize)
            IF (isize == SIZE(dimer_env%nvec)) THEN
               CALL reallocate(array, 1, j)
               CALL section_vals_val_set(nvec_section, "_DEFAULT_KEYWORD_", r_vals_ptr=array, &
                                         i_rep_val=i_rep_val)
               EXIT Main_Loop
            END IF
         END DO
         CALL section_vals_val_set(nvec_section, "_DEFAULT_KEYWORD_", r_vals_ptr=array, &
                                   i_rep_val=i_rep_val)
      END DO Main_Loop
      IF(.NOT.(isize == SIZE(dimer_env%nvec)))CALL cp__a("motion/dimer_utils.F",108)
   END SUBROUTINE update_dimer_vec

! **************************************************************************************************
!> \brief This function orthonormalize the vector for the rotational search
!> \param gradient ...
!> \param dimer_env ...
!> \param norm ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE get_theta(gradient, dimer_env, norm)
      REAL(KIND=dp), DIMENSION(:)                        :: gradient
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      REAL(KIND=dp), INTENT(OUT)                         :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'get_theta', routineP = moduleN//':'//routineN

      gradient = gradient-DOT_PRODUCT(gradient, dimer_env%nvec)*dimer_env%nvec
      norm = SQRT(DOT_PRODUCT(gradient, gradient))
      IF (norm < EPSILON(0.0_dp)) THEN
         ! This means that NVEC is totally aligned with minimum curvature mode
         gradient = 0.0_dp
      ELSE
         gradient = gradient/norm
      END IF
   END SUBROUTINE get_theta

END MODULE dimer_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for Geometry optimization using  Conjugate Gradients
!> \author Teodoro Laino [teo]
!>      10.2005
! **************************************************************************************************
MODULE cg_utils
   USE cp_external_control,             ONLY: external_control
   USE dimer_types,                     ONLY: dimer_env_type
   USE dimer_utils,                     ONLY: dimer_thrs,&
                                              rotate_dimer
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_constants,                 ONLY: default_cell_method_id,&
                                              default_minimization_method_id,&
                                              default_shellcore_method_id,&
                                              default_ts_method_id,&
                                              ls_2pnt,&
                                              ls_fit,&
                                              ls_gold
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param error variable to control error logging, stopping,...
!>             see module cp_error_handling
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! *****************************************************************************
INTERFACE

  RECURSIVE SUBROUTINE cp_eval_at(gopt_env,x,f,gradient,master,&
                                  final_evaluation,para_env)

    USE cp_para_types,                   ONLY: cp_para_env_type
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE kinds,                           ONLY: dp

    TYPE(gopt_f_type), POINTER               :: gopt_env
    REAL(KIND=dp), DIMENSION(:), POINTER     :: x
    REAL(KIND=dp), INTENT(out), OPTIONAL     :: f
    REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
    INTEGER, INTENT(IN)                      :: master
    LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
    TYPE(cp_para_env_type), POINTER          :: para_env

  END SUBROUTINE cp_eval_at

END INTERFACE

   PUBLIC :: cg_linmin, get_conjugate_direction
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cg_utils'

CONTAINS

! **************************************************************************************************
!> \brief Main driver for line minimization routines for CG
!> \param gopt_env ...
!> \param xvec ...
!> \param xi ...
!> \param g ...
!> \param opt_energy ...
!> \param output_unit ...
!> \param gopt_param ...
!> \param globenv ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE cg_linmin(gopt_env, xvec, xi, g, opt_energy, output_unit, gopt_param, &
                                  globenv)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xvec, xi, g
      REAL(KIND=dp), INTENT(INOUT)                       :: opt_energy
      INTEGER                                            :: output_unit
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'cg_linmin', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      gopt_env%do_line_search = .TRUE.
      SELECT CASE (gopt_env%type_id)
      CASE (default_minimization_method_id)
         SELECT CASE (gopt_param%cg_ls%type_id)
         CASE (ls_2pnt)
            CALL linmin_2pnt(gopt_env, xvec, xi, g, opt_energy, gopt_param, use_only_grad=gopt_param%cg_ls%grad_only, &
                             output_unit=output_unit)
         CASE (ls_fit)
            CALL linmin_fit(gopt_env, xvec, xi, opt_energy, gopt_param%cg_ls%brack_limit, &
                            gopt_param%cg_ls%initial_step, output_unit, gopt_param, globenv)
         CASE (ls_gold)
            CALL linmin_gold(gopt_env, xvec, xi, opt_energy, gopt_param%cg_ls%brent_tol, &
                             gopt_param%cg_ls%brent_max_iter, gopt_param%cg_ls%brack_limit, &
                             gopt_param%cg_ls%initial_step, output_unit, globenv)
         CASE DEFAULT
            CALL cp__b("motion/cg_utils.F",85,"Line Search type not yet implemented in CG.")
         END SELECT
      CASE (default_ts_method_id)
         SELECT CASE (gopt_param%cg_ls%type_id)
         CASE (ls_2pnt)
            IF (gopt_env%dimer_rotation) THEN
               CALL rotmin_2pnt(gopt_env, gopt_env%dimer_env, xvec, xi, opt_energy)
            ELSE
               CALL tslmin_2pnt(gopt_env, gopt_env%dimer_env, xvec, xi, opt_energy, gopt_param, &
                                output_unit)
            END IF
         CASE DEFAULT
            CALL cp__b("motion/cg_utils.F",97,"Line Search type not yet implemented in CG for TS search.")
         END SELECT
      CASE (default_cell_method_id)
         SELECT CASE (gopt_param%cg_ls%type_id)
         CASE (ls_2pnt)
            CALL linmin_2pnt(gopt_env, xvec, xi, g, opt_energy, gopt_param, use_only_grad=.TRUE., &
                             output_unit=output_unit)
         CASE (ls_gold)
            CALL linmin_gold(gopt_env, xvec, xi, opt_energy, gopt_param%cg_ls%brent_tol, &
                             gopt_param%cg_ls%brent_max_iter, gopt_param%cg_ls%brack_limit, &
                             gopt_param%cg_ls%initial_step, output_unit, globenv)
         CASE DEFAULT
            CALL cp__b("motion/cg_utils.F",109,"Line Search type not yet implemented in CG for cell optimization.")
         END SELECT
      CASE (default_shellcore_method_id)
         SELECT CASE (gopt_param%cg_ls%type_id)
         CASE (ls_2pnt)
            CALL linmin_2pnt(gopt_env, xvec, xi, g, opt_energy, gopt_param, use_only_grad=.TRUE., &
                             output_unit=output_unit)
         CASE DEFAULT
            CALL cp__b("motion/cg_utils.F",117,"Line Search type not yet implemented in CG for shellcore optimization.")
         END SELECT

      END SELECT
      gopt_env%do_line_search = .FALSE.
      CALL timestop(handle)

   END SUBROUTINE cg_linmin

! **************************************************************************************************
!> \brief Line search subroutine based on 2 points (using gradients and energies
!>        or only gradients)
!> \param gopt_env ...
!> \param x0 ...
!> \param ls_vec ...
!> \param g ...
!> \param opt_energy ...
!> \param gopt_param ...
!> \param use_only_grad ...
!> \param output_unit ...
!> \author Teodoro Laino - created [tlaino] - 03.2008
! **************************************************************************************************
   RECURSIVE SUBROUTINE linmin_2pnt(gopt_env, x0, ls_vec, g, opt_energy, gopt_param, use_only_grad, &
                                    output_unit)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0, ls_vec, g
      REAL(KIND=dp), INTENT(INOUT)                       :: opt_energy
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_only_grad
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'linmin_2pnt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_use_only_grad, &
                                                            save_consistent_energy_force
      REAL(KIND=dp)                                      :: a, b, c, dx, dx_min, dx_min_save, &
                                                            dx_thrs, norm_grad1, norm_grad2, &
                                                            norm_ls_vec, opt_energy2, x_grad_zero
      REAL(KIND=dp), DIMENSION(:), POINTER               :: gradient2, ls_norm

      CALL timeset(routineN, handle)
      norm_ls_vec = SQRT(DOT_PRODUCT(ls_vec, ls_vec))
      my_use_only_grad = .FALSE.
      IF (PRESENT(use_only_grad)) my_use_only_grad = use_only_grad
      IF (norm_ls_vec /= 0.0_dp) THEN
         ALLOCATE (ls_norm(SIZE(ls_vec)))
         ALLOCATE (gradient2(SIZE(ls_vec)))
         ls_norm = ls_vec/norm_ls_vec
         dx = norm_ls_vec
         dx_thrs = gopt_param%cg_ls%max_step

         x0 = x0+dx*ls_norm
         ![NB] don't need consistent energies and forces if using only gradient
         save_consistent_energy_force = gopt_env%require_consistent_energy_force
         gopt_env%require_consistent_energy_force = .NOT. my_use_only_grad
         CALL cp_eval_at(gopt_env, x0, opt_energy2, gradient2, master=gopt_env%force_env%para_env%mepos, &
                         final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)
         gopt_env%require_consistent_energy_force = save_consistent_energy_force

         norm_grad1 = -DOT_PRODUCT(g, ls_norm)
         norm_grad2 = DOT_PRODUCT(gradient2, ls_norm)
         IF (my_use_only_grad) THEN
            ! a*x+b=y
            ! per x=0; b=norm_grad1
            b = norm_grad1
            ! per x=dx; a*dx+b=norm_grad2
            a = (norm_grad2-b)/dx
            x_grad_zero = -b/a
            dx_min = x_grad_zero
         ELSE
            ! ax**2+b*x+c=y
            ! per x=0 ; c=opt_energy
            c = opt_energy
            ! per x=dx;          a*dx**2 + b*dx + c = opt_energy2
            ! per x=dx;        2*a*dx    + b        = norm_grad2
            !
            !                  - a*dx**2        + c = (opt_energy2-norm_grad2*dx)
            !                    a*dx**2            = c - (opt_energy2-norm_grad2*dx)
            a = (c-(opt_energy2-norm_grad2*dx))/dx**2
            b = norm_grad2-2.0_dp*a*dx
            dx_min = 0.0_dp
            IF (a /= 0.0_dp) dx_min = -b/(2.0_dp*a)
            opt_energy = opt_energy2
         END IF
         dx_min_save = dx_min
         ! In case the solution is larger than the maximum threshold let's assume the maximum allowed
         ! step length
         IF (ABS(dx_min) > dx_thrs) dx_min = SIGN(1.0_dp, dx_min)*dx_thrs
         x0 = x0+(dx_min-dx)*ls_norm

         ! Print out LS info
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
            WRITE (UNIT=output_unit, FMT="(T2,A,T31,A,T78,A)") &
               "***", "2PNT LINE SEARCH INFO", "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,T78,A)") "***", "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,3X,A,F12.6,T45,A,F12.6,T78,A)") &
               "***", "DX (EVALUATED)=", dx, "DX (THRESHOLD)=", dx_thrs, "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,3X,A,F12.6,T45,A,F12.6,T78,A)") &
               "***", "DX (FITTED   )=", dx_min_save, "DX (ACCEPTED )=", dx_min, "***"
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
         END IF
         DEALLOCATE (ls_norm)
         DEALLOCATE (gradient2)
      ELSE
         ! Do Nothing, since.. if the effective force is 0 means that we are already
         ! in the saddle point..
      END IF
      CALL timestop(handle)
   END SUBROUTINE linmin_2pnt

! **************************************************************************************************
!> \brief Translational minimization for the Dimer Method - 2pnt LS
!> \param gopt_env ...
!> \param dimer_env ...
!> \param x0 ...
!> \param tls_vec ...
!> \param opt_energy ...
!> \param gopt_param ...
!> \param output_unit ...
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE tslmin_2pnt(gopt_env, dimer_env, x0, tls_vec, opt_energy, gopt_param, output_unit)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0, tls_vec
      REAL(KIND=dp), INTENT(INOUT)                       :: opt_energy
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'tslmin_2pnt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: dx, dx_min, dx_min_acc, dx_min_save, &
                                                            dx_thrs, norm_tls_vec, opt_energy2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tls_norm

      CALL timeset(routineN, handle)
      norm_tls_vec = SQRT(DOT_PRODUCT(tls_vec, tls_vec))
      IF (norm_tls_vec /= 0.0_dp) THEN
         ALLOCATE (tls_norm(SIZE(tls_vec)))

         tls_norm = tls_vec/norm_tls_vec
         dimer_env%tsl%tls_vec => tls_norm

         dx = norm_tls_vec
         dx_thrs = gopt_param%cg_ls%max_step
         ! If curvature is positive let's make the largest step allowed
         IF (dimer_env%rot%curvature > 0) dx = dx_thrs
         x0 = x0+dx*tls_norm
         CALL cp_eval_at(gopt_env, x0, opt_energy2, master=gopt_env%force_env%para_env%mepos, &
                         final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)
         IF (dimer_env%rot%curvature > 0) THEN
            dx_min = 0.0_dp
            dx_min_save = dx
            dx_min_acc = dx
         ELSE
            ! First let's try to interpolate the minimum
            dx_min = -opt_energy/(opt_energy2-opt_energy)*dx
            ! In case the solution is larger than the maximum threshold let's assume the maximum allowed
            ! step length
            dx_min_save = dx_min
            IF (ABS(dx_min) > dx_thrs) dx_min = SIGN(1.0_dp, dx_min)*dx_thrs
            dx_min_acc = dx_min
            dx_min = dx_min-dx
         END IF
         x0 = x0+dx_min*tls_norm

         ! Print out LS info
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
            WRITE (UNIT=output_unit, FMT="(T2,A,T24,A,T78,A)") &
               "***", "2PNT TRANSLATIONAL LINE SEARCH INFO", "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,T78,A)") "***", "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,3X,A,F12.6,T78,A)") &
               "***", "LOCAL CURVATURE =", dimer_env%rot%curvature, "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,3X,A,F12.6,T45,A,F12.6,T78,A)") &
               "***", "DX (EVALUATED)=", dx, "DX (THRESHOLD)=", dx_thrs, "***"
            WRITE (UNIT=output_unit, FMT="(T2,A,3X,A,F12.6,T45,A,F12.6,T78,A)") &
               "***", "DX (FITTED   )=", dx_min_save, "DX (ACCEPTED )=", dx_min_acc, "***"
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
         END IF

         ! Here we compute the value of the energy in point zero..
         CALL cp_eval_at(gopt_env, x0, opt_energy, master=gopt_env%force_env%para_env%mepos, &
                         final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)

         DEALLOCATE (tls_norm)
      ELSE
         ! Do Nothing, since.. if the effective force is 0 means that we are already
         ! in the saddle point..
      END IF
      CALL timestop(handle)

   END SUBROUTINE tslmin_2pnt

! **************************************************************************************************
!> \brief Rotational minimization for the Dimer Method - 2 pnt LS
!> \param gopt_env ...
!> \param dimer_env ...
!> \param x0 ...
!> \param theta ...
!> \param opt_energy ...
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE rotmin_2pnt(gopt_env, dimer_env, x0, theta, opt_energy)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0, theta
      REAL(KIND=dp), INTENT(INOUT)                       :: opt_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'rotmin_2pnt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: a0, a1, angle, b1, curvature0, &
                                                            curvature1, curvature2, dCdp, f
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work

      CALL timeset(routineN, handle)
      curvature0 = dimer_env%rot%curvature
      dCdp = dimer_env%rot%dCdp
      b1 = 0.5_dp*dCdp
      angle = -0.5_dp*ATAN(dCdp/(2.0_dp*ABS(curvature0)))
      dimer_env%rot%angle1 = angle
      dimer_env%cg_rot%nvec_old = dimer_env%nvec
      IF (angle > dimer_env%rot%angle_tol) THEN
         ! Rotating the dimer of dtheta degrees
         CALL rotate_dimer(dimer_env%nvec, theta, angle)
         ! Re-compute energy, gradients and rotation vector for new R1
         CALL cp_eval_at(gopt_env, x0, f, master=gopt_env%force_env%para_env%mepos, &
                         final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)

         curvature1 = dimer_env%rot%curvature
         a1 = (curvature0-curvature1+b1*SIN(2.0_dp*angle))/(1.0_dp-COS(2.0_dp*angle))
         a0 = 2.0_dp*(curvature0-a1)
         angle = 0.5_dp*ATAN(b1/a1)
         curvature2 = a0/2.0_dp+a1*COS(2.0_dp*angle)+b1*SIN(2.0_dp*angle)
         IF (curvature2 > curvature0) THEN
            angle = angle+pi/2.0_dp
            curvature2 = a0/2.0_dp+a1*COS(2.0_dp*angle)+b1*SIN(2.0_dp*angle)
         END IF
         dimer_env%rot%angle2 = angle
         dimer_env%rot%curvature = curvature2
         ! Rotating the dimer the optimized (in plane) vector position
         dimer_env%nvec = dimer_env%cg_rot%nvec_old
         CALL rotate_dimer(dimer_env%nvec, theta, angle)

         ! Evaluate (by interpolation) the norm of the rotational force in the
         ! minimum of the rotational search (this is for print-out only)
         ALLOCATE (work(SIZE(dimer_env%nvec)))
         work = dimer_env%rot%g1
         work = SIN(dimer_env%rot%angle1-dimer_env%rot%angle2)/SIN(dimer_env%rot%angle1)*dimer_env%rot%g1+ &
                SIN(dimer_env%rot%angle2)/SIN(dimer_env%rot%angle1)*dimer_env%rot%g1p+ &
                (1.0_dp-COS(dimer_env%rot%angle2)-SIN(dimer_env%rot%angle2)*TAN(dimer_env%rot%angle1/2.0_dp))* &
                dimer_env%rot%g0
         work = -2.0_dp*(work-dimer_env%rot%g0)
         work = work-DOT_PRODUCT(work, dimer_env%nvec)*dimer_env%nvec
         opt_energy = SQRT(DOT_PRODUCT(work, work))
         DEALLOCATE (work)
      END IF
      dimer_env%rot%angle2 = angle
      CALL timestop(handle)

   END SUBROUTINE rotmin_2pnt

! **************************************************************************************************
!> \brief Line Minimization - Fit
!> \param gopt_env ...
!> \param xvec ...
!> \param xi ...
!> \param opt_energy ...
!> \param brack_limit ...
!> \param step ...
!> \param output_unit ...
!> \param gopt_param ...
!> \param globenv ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
!> \note
!>      Given as input the vector XVEC and XI, finds the scalar
!>      xmin that minimizes the energy XVEC+xmin*XI. Replace step
!>      with the optimal value. Enhanced Version
! **************************************************************************************************
   SUBROUTINE linmin_fit(gopt_env, xvec, xi, opt_energy, &
                         brack_limit, step, output_unit, gopt_param, globenv)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xvec, xi
      REAL(KIND=dp)                                      :: opt_energy, brack_limit, step
      INTEGER                                            :: output_unit
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'linmin_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, loc_iter, odim
      LOGICAL                                            :: should_stop
      REAL(KIND=dp)                                      :: ax, bx, fprev, rms_dr, rms_force, scale, &
                                                            xmin, xx
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hist

      CALL timeset(routineN, handle)

      NULLIFY (pcom, xicom, hist)
      rms_dr = gopt_param%rms_dr
      rms_force = gopt_param%rms_force
      ALLOCATE (pcom(SIZE(xvec)))
      ALLOCATE (xicom(SIZE(xvec)))

      pcom = xvec
      xicom = xi
      xicom = xicom/SQRT(DOT_PRODUCT(xicom, xicom))
      step = step*0.8_dp ! target a little before the minimum for the first point
      ax = 0.0_dp
      xx = step
      CALL cg_mnbrak(gopt_env, ax, xx, bx, pcom, xicom, brack_limit, output_unit, &
                     histpoint=hist, globenv=globenv)
      !
      fprev = 0.0_dp
      opt_energy = MINVAL(hist(:, 2))
      odim = SIZE(hist, 1)
      scale = 0.25_dp
      loc_iter = 0
      DO WHILE (ABS(hist(odim, 3)) > rms_force*scale .OR. ABS(hist(odim, 1)-hist(odim-1, 1)) > scale*rms_dr)
         CALL external_control(should_stop, "LINFIT", globenv=globenv)
         IF (should_stop) EXIT
         !
         loc_iter = loc_iter+1
         fprev = opt_energy
         xmin = FindMin(hist(:, 1), hist(:, 2), hist(:, 3))
         CALL reallocate(hist, 1, odim+1, 1, 3)
         hist(odim+1, 1) = xmin
         hist(odim+1, 3) = cg_deval1d(gopt_env, xmin, pcom, xicom, opt_energy)
         hist(odim+1, 2) = opt_energy
         odim = SIZE(hist, 1)
      END DO
      !
      xicom = xmin*xicom
      step = xmin
      xvec = xvec+xicom
      DEALLOCATE (pcom)
      DEALLOCATE (xicom)
      DEALLOCATE (hist)
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T22,A,I7,T78,A)") &
            "***", "FIT LS  - NUMBER OF ENERGY EVALUATIONS : ", loc_iter, "***"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
      END IF
      CALL timestop(handle)

   END SUBROUTINE linmin_fit

! **************************************************************************************************
!> \brief Line Minimization routine - GOLD
!> \param gopt_env ...
!> \param xvec ...
!> \param xi ...
!> \param opt_energy ...
!> \param brent_tol ...
!> \param brent_max_iter ...
!> \param brack_limit ...
!> \param step ...
!> \param output_unit ...
!> \param globenv ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
!> \note
!>      Given as input the vector XVEC and XI, finds the scalar
!>      xmin that minimizes the energy XVEC+xmin*XI. Replaces XMIN
!>      with the optimal value
! **************************************************************************************************
   SUBROUTINE linmin_gold(gopt_env, xvec, xi, opt_energy, brent_tol, brent_max_iter, &
                          brack_limit, step, output_unit, globenv)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xvec, xi
      REAL(KIND=dp)                                      :: opt_energy, brent_tol
      INTEGER                                            :: brent_max_iter
      REAL(KIND=dp)                                      :: brack_limit, step
      INTEGER                                            :: output_unit
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'linmin_gold', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: ax, bx, xmin, xx
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom

      CALL timeset(routineN, handle)

      NULLIFY (pcom, xicom)
      ALLOCATE (pcom(SIZE(xvec)))
      ALLOCATE (xicom(SIZE(xvec)))

      pcom = xvec
      xicom = xi
      xicom = xicom/SQRT(DOT_PRODUCT(xicom, xicom))
      step = step*0.8_dp ! target a little before the minimum for the first point
      ax = 0.0_dp
      xx = step
      CALL cg_mnbrak(gopt_env, ax, xx, bx, pcom, xicom, brack_limit, output_unit, &
                     globenv=globenv)

      opt_energy = cg_dbrent(gopt_env, ax, xx, bx, brent_tol, brent_max_iter, &
                             xmin, pcom, xicom, output_unit, globenv)
      xicom = xmin*xicom
      step = xmin
      xvec = xvec+xicom
      DEALLOCATE (pcom)
      DEALLOCATE (xicom)
      CALL timestop(handle)
   END SUBROUTINE linmin_gold

! **************************************************************************************************
!> \brief Routine for intially bracketing a minimum based on the golden search
!>      minimum
!> \param gopt_env ...
!> \param ax ...
!> \param bx ...
!> \param cx ...
!> \param pcom ...
!> \param xicom ...
!> \param brack_limit ...
!> \param output_unit ...
!> \param histpoint ...
!> \param globenv ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
!> \note
!>      Given two distinct initial points ax and bx this routine searches
!>      in the downhill direction and returns new points ax, bx, cx that
!>      bracket the minimum of the function
! **************************************************************************************************
   SUBROUTINE cg_mnbrak(gopt_env, ax, bx, cx, pcom, xicom, brack_limit, output_unit, &
                        histpoint, globenv)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp)                                      :: ax, bx, cx
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom
      REAL(KIND=dp)                                      :: brack_limit
      INTEGER                                            :: output_unit
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: histpoint
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'cg_mnbrak', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, loc_iter, odim
      LOGICAL                                            :: hist, should_stop
      REAL(KIND=dp)                                      :: dum, fa, fb, fc, fu, gold, q, r, u, ulim

      CALL timeset(routineN, handle)
      hist = PRESENT(histpoint)
      IF (hist) THEN
         IF(.NOT.(.NOT. ASSOCIATED(histpoint)))CALL cp__a("motion/cg_utils.F",573)
         ALLOCATE (histpoint(3, 3))
      END IF
      gold = (1.0_dp+SQRT(5.0_dp))/2.0_dp
      IF (hist) THEN
         histpoint(1, 1) = ax
         histpoint(1, 3) = cg_deval1d(gopt_env, ax, pcom, xicom, fa)
         histpoint(1, 2) = fa
         histpoint(2, 1) = bx
         histpoint(2, 3) = cg_deval1d(gopt_env, bx, pcom, xicom, fb)
         histpoint(2, 2) = fb
      ELSE
         fa = cg_eval1d(gopt_env, ax, pcom, xicom)
         fb = cg_eval1d(gopt_env, bx, pcom, xicom)
      END IF
      IF (fb .GT. fa) THEN
         dum = ax
         ax = bx
         bx = dum
         dum = fb
         fb = fa
         fa = dum
      ENDIF
      cx = bx+gold*(bx-ax)
      IF (hist) THEN
         histpoint(3, 1) = cx
         histpoint(3, 3) = cg_deval1d(gopt_env, cx, pcom, xicom, fc)
         histpoint(3, 2) = fc
      ELSE
         fc = cg_eval1d(gopt_env, cx, pcom, xicom)
      END IF
      loc_iter = 3
      DO WHILE (fb .GE. fc)
         CALL external_control(should_stop, "MNBRACK", globenv=globenv)
         IF (should_stop) EXIT
         !
         r = (bx-ax)*(fb-fc)
         q = (bx-cx)*(fb-fa)
         u = bx-((bx-cx)*q-(bx-ax)*r)/(2.0_dp*SIGN(MAX(ABS(q-r), TINY(0.0_dp)), q-r))
         ulim = bx+brack_limit*(cx-bx)
         IF ((bx-u)*(u-cx) .GT. 0.0_dp) THEN
            IF (hist) THEN
               odim = SIZE(histpoint, 1)
               CALL reallocate(histpoint, 1, odim+1, 1, 3)
               histpoint(odim+1, 1) = u
               histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
               histpoint(odim+1, 2) = fu
            ELSE
               fu = cg_eval1d(gopt_env, u, pcom, xicom)
            END IF
            loc_iter = loc_iter+1
            IF (fu .LT. fc) THEN
               ax = bx
               fa = fb
               bx = u
               fb = fu
               EXIT
            ELSE IF (fu .GT. fb) THEN
               cx = u
               fc = fu
               EXIT
            ENDIF
            u = cx+gold*(cx-bx)
            IF (hist) THEN
               odim = SIZE(histpoint, 1)
               CALL reallocate(histpoint, 1, odim+1, 1, 3)
               histpoint(odim+1, 1) = u
               histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
               histpoint(odim+1, 2) = fu
            ELSE
               fu = cg_eval1d(gopt_env, u, pcom, xicom)
            END IF
            loc_iter = loc_iter+1
         ELSE IF ((cx-u)*(u-ulim) .GT. 0.) THEN
            IF (hist) THEN
               odim = SIZE(histpoint, 1)
               CALL reallocate(histpoint, 1, odim+1, 1, 3)
               histpoint(odim+1, 1) = u
               histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
               histpoint(odim+1, 2) = fu
            ELSE
               fu = cg_eval1d(gopt_env, u, pcom, xicom)
            END IF
            loc_iter = loc_iter+1
            IF (fu .LT. fc) THEN
               bx = cx
               cx = u
               u = cx+gold*(cx-bx)
               fb = fc
               fc = fu
               IF (hist) THEN
                  odim = SIZE(histpoint, 1)
                  CALL reallocate(histpoint, 1, odim+1, 1, 3)
                  histpoint(odim+1, 1) = u
                  histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
                  histpoint(odim+1, 2) = fu
               ELSE
                  fu = cg_eval1d(gopt_env, u, pcom, xicom)
               END IF
               loc_iter = loc_iter+1
            ENDIF
         ELSE IF ((u-ulim)*(ulim-cx) .GE. 0.) THEN
            u = ulim
            IF (hist) THEN
               odim = SIZE(histpoint, 1)
               CALL reallocate(histpoint, 1, odim+1, 1, 3)
               histpoint(odim+1, 1) = u
               histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
               histpoint(odim+1, 2) = fu
            ELSE
               fu = cg_eval1d(gopt_env, u, pcom, xicom)
            END IF
            loc_iter = loc_iter+1
         ELSE
            u = cx+gold*(cx-bx)
            IF (hist) THEN
               odim = SIZE(histpoint, 1)
               CALL reallocate(histpoint, 1, odim+1, 1, 3)
               histpoint(odim+1, 1) = u
               histpoint(odim+1, 3) = cg_deval1d(gopt_env, u, pcom, xicom, fu)
               histpoint(odim+1, 2) = fu
            ELSE
               fu = cg_eval1d(gopt_env, u, pcom, xicom)
            END IF
            loc_iter = loc_iter+1
         ENDIF
         ax = bx
         bx = cx
         cx = u
         fa = fb
         fb = fc
         fc = fu
      END DO
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T22,A,I7,T78,A)") &
            "***", "MNBRACK - NUMBER OF ENERGY EVALUATIONS : ", loc_iter, "***"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
      END IF
      CALL timestop(handle)
   END SUBROUTINE cg_mnbrak

! **************************************************************************************************
!> \brief Routine implementing the Brent Method
!>      Brent,R.P. Algorithm for Minimization without Derivatives, Chapt.5
!>      1973
!>      Extension in the use of derivatives
!> \param gopt_env ...
!> \param ax ...
!> \param bx ...
!> \param cx ...
!> \param tol ...
!> \param itmax ...
!> \param xmin ...
!> \param pcom ...
!> \param xicom ...
!> \param output_unit ...
!> \param globenv ...
!> \retval dbrent ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
!> \note
!>      Given a bracketing  triplet of abscissas ax, bx, cx (such that bx
!>      is between ax and cx and energy of bx is less than energy of ax and cx),
!>      this routine isolates the minimum to a precision of about tol using
!>      Brent method. This routine implements the extension of the Brent Method
!>      using derivatives
! **************************************************************************************************
   FUNCTION cg_dbrent(gopt_env, ax, bx, cx, tol, itmax, xmin, pcom, xicom, output_unit, &
                      globenv) RESULT(dbrent)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp)                                      :: ax, bx, cx, tol
      INTEGER                                            :: itmax
      REAL(KIND=dp)                                      :: xmin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom
      INTEGER                                            :: output_unit
      TYPE(global_environment_type), POINTER             :: globenv
      REAL(KIND=dp)                                      :: dbrent

      CHARACTER(len=*), PARAMETER :: routineN = 'cg_dbrent', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zeps = 1.0E-8_dp

      INTEGER                                            :: handle, iter, loc_iter
      LOGICAL                                            :: ok1, ok2, should_stop, skip0, skip1
      REAL(KIND=dp)                                      :: a, b, d, d1, d2, du, dv, dw, dx, e, fu, &
                                                            fv, fw, fx, olde, tol1, tol2, u, u1, &
                                                            u2, v, w, x, xm

      CALL timeset(routineN, handle)
      a = MIN(ax, cx)
      b = MAX(ax, cx)
      v = bx; w = v; x = v
      e = 0.0_dp
      dx = cg_deval1d(gopt_env, x, pcom, xicom, fx)
      fv = fx
      fw = fx
      dv = dx
      dw = dx
      loc_iter = 1
      DO iter = 1, itmax
         CALL external_control(should_stop, "BRENT", globenv=globenv)
         IF (should_stop) EXIT
         !
         xm = 0.5_dp*(a+b)
         tol1 = tol*ABS(x)+zeps
         tol2 = 2.0_dp*tol1
         skip0 = .FALSE.
         skip1 = .FALSE.
         IF (ABS(x-xm) .LE. (tol2-0.5_dp*(b-a))) EXIT
         IF (ABS(e) .GT. tol1) THEN
            d1 = 2.0_dp*(b-a)
            d2 = d1
            IF (dw .NE. dx) d1 = (w-x)*dx/(dx-dw)
            IF (dv .NE. dx) d2 = (v-x)*dx/(dx-dv)
            u1 = x+d1
            u2 = x+d2
            ok1 = ((a-u1)*(u1-b) .GT. 0.0_dp) .AND. (dx*d1 .LE. 0.0_dp)
            ok2 = ((a-u2)*(u2-b) .GT. 0.0_dp) .AND. (dx*d2 .LE. 0.0_dp)
            olde = e
            e = d
            IF (.NOT. (ok1 .OR. ok2)) THEN
               skip0 = .TRUE.
            ELSE IF (ok1 .AND. ok2) THEN
               IF (ABS(d1) .LT. ABS(d2)) THEN
                  d = d1
               ELSE
                  d = d2
               ENDIF
            ELSE IF (ok1) THEN
               d = d1
            ELSE
               d = d2
            ENDIF
            IF (.NOT. skip0) THEN
               IF (ABS(d) .GT. ABS(0.5_dp*olde)) skip0 = .TRUE.
               IF (.NOT. skip0) THEN
                  u = x+d
                  IF ((u-a) .LT. tol2 .OR. (b-u) .LT. tol2) d = SIGN(tol1, xm-x)
                  skip1 = .TRUE.
               END IF
            END IF
         ENDIF
         IF (.NOT. skip1) THEN
            IF (dx .GE. 0.0_dp) THEN
               e = a-x
            ELSE
               e = b-x
            ENDIF
            d = 0.5_dp*e
         END IF
         IF (ABS(d) .GE. tol1) THEN
            u = x+d
            du = cg_deval1d(gopt_env, u, pcom, xicom, fu)
            loc_iter = loc_iter+1
         ELSE
            u = x+SIGN(tol1, d)
            du = cg_deval1d(gopt_env, u, pcom, xicom, fu)
            loc_iter = loc_iter+1
            IF (fu .GT. fx) EXIT
         ENDIF
         IF (fu .LE. fx) THEN
            IF (u .GE. x) THEN
               a = x
            ELSE
               b = x
            ENDIF
            v = w; fv = fw; dv = dw; w = x
            fw = fx; dw = dx; x = u; fx = fu; dx = du
         ELSE
            IF (u .LT. x) THEN
               a = u
            ELSE
               b = u
            ENDIF
            IF (fu .LE. fw .OR. w .EQ. x) THEN
               v = w; fv = fw; dv = dw
               w = u; fw = fu; dw = du
            ELSE IF (fu .LE. fv .OR. v .EQ. x .OR. v .EQ. w) THEN
               v = u
               fv = fu
               dv = du
            ENDIF
         ENDIF
      END DO
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T22,A,I7,T78,A)") &
            "***", "BRENT   - NUMBER OF ENERGY EVALUATIONS : ", loc_iter, "***"
         IF (iter == itmax+1) &
            WRITE (UNIT=output_unit, FMT="(T2,A,T22,A,T78,A)") &
            "***", "BRENT - NUMBER OF ITERATIONS EXCEEDED ", "***"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
      END IF
      IF(.NOT.(iter /= itmax+1))CALL cp__a("motion/cg_utils.F",867)
      xmin = x
      dbrent = fx
      CALL timestop(handle)

   END FUNCTION cg_dbrent

! **************************************************************************************************
!> \brief Evaluates energy in one dimensional space defined by the point
!>      pcom and with direction xicom, position x
!> \param gopt_env ...
!> \param x ...
!> \param pcom ...
!> \param xicom ...
!> \retval my_val ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   FUNCTION cg_eval1d(gopt_env, x, pcom, xicom) RESULT(my_val)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp)                                      :: x
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom
      REAL(KIND=dp)                                      :: my_val

      CHARACTER(len=*), PARAMETER :: routineN = 'cg_eval1d', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xvec

      CALL timeset(routineN, handle)

      ALLOCATE (xvec(SIZE(pcom)))
      xvec = pcom+x*xicom
      CALL cp_eval_at(gopt_env, xvec, my_val, master=gopt_env%force_env%para_env%mepos, &
                      final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)
      DEALLOCATE (xvec)

      CALL timestop(handle)

   END FUNCTION cg_eval1d

! **************************************************************************************************
!> \brief Evaluates derivatives in one dimensional space defined by the point
!>      pcom and with direction xicom, position x
!> \param gopt_env ...
!> \param x ...
!> \param pcom ...
!> \param xicom ...
!> \param fval ...
!> \retval my_val ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   FUNCTION cg_deval1d(gopt_env, x, pcom, xicom, fval) RESULT(my_val)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp)                                      :: x
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pcom, xicom
      REAL(KIND=dp)                                      :: fval, my_val

      CHARACTER(len=*), PARAMETER :: routineN = 'cg_deval1d', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: grad, xvec

      CALL timeset(routineN, handle)

      ALLOCATE (xvec(SIZE(pcom)))
      ALLOCATE (grad(SIZE(pcom)))
      xvec = pcom+x*xicom
      CALL cp_eval_at(gopt_env, xvec, energy, grad, master=gopt_env%force_env%para_env%mepos, &
                      final_evaluation=.FALSE., para_env=gopt_env%force_env%para_env)
      my_val = DOT_PRODUCT(grad, xicom)
      fval = energy
      DEALLOCATE (xvec)
      DEALLOCATE (grad)
      CALL timestop(handle)

   END FUNCTION cg_deval1d

! **************************************************************************************************
!> \brief Find the minimum of a parabolic function obtained with a least square fit
!> \param x ...
!> \param y ...
!> \param dy ...
!> \retval res ...
!> \par History
!>      10.2005 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION FindMin(x, y, dy) RESULT(res)
      REAL(kind=dp), DIMENSION(:)                        :: x, y, dy
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'FindMin', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, iwork(8*3), lwork, min_pos, np
      REAL(kind=dp)                                      :: diag(3), res1(3), res2(3), res3(3), &
                                                            spread, sum_x, sum_xx, tmpw(1), &
                                                            vt(3, 3)
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: work
      REAL(kind=dp), DIMENSION(2*SIZE(x), 3)             :: f
      REAL(kind=dp), DIMENSION(2*SIZE(x))                :: b, w
      REAL(kind=dp)                                      :: u(2*SIZE(x), 3)

      np = SIZE(x)
      IF(.NOT.(np > 1))CALL cp__a("motion/cg_utils.F",975)
      sum_x = 0._dp
      sum_xx = 0._dp
      min_pos = 1
      DO i = 1, np
         sum_xx = sum_xx+x(i)**2
         sum_x = sum_x+x(i)
         IF (y(min_pos) > y(i)) min_pos = i
      END DO
      spread = SQRT(sum_xx/REAL(np, dp)-(sum_x/REAL(np, dp))**2)
      DO i = 1, np
         w(i) = EXP(-(REAL(np-i, dp))**2/(REAL(2*9, dp)))
         w(i+np) = 2._dp*w(i)
      END DO
      DO i = 1, np
         f(i, 1) = w(i)
         f(i, 2) = x(i)*w(i)
         f(i, 3) = x(i)**2*w(i)
         f(i+np, 1) = 0
         f(i+np, 2) = w(i+np)
         f(i+np, 3) = 2*x(i)*w(i+np)
      END DO
      DO i = 1, np
         b(i) = y(i)*w(i)
         b(i+np) = dy(i)*w(i+np)
      END DO
      lwork = -1
      CALL dgesdd('S', SIZE(f, 1), SIZE(f, 2), f, SIZE(f, 1), diag, u, SIZE(u, 1), vt, SIZE(vt, 1), tmpw, lwork, &
                  iwork, info)
      lwork = CEILING(tmpw(1))
      ALLOCATE (work(lwork))
      CALL dgesdd('S', SIZE(f, 1), SIZE(f, 2), f, SIZE(f, 1), diag, u, SIZE(u, 1), vt, SIZE(vt, 1), work, lwork, &
                  iwork, info)
      DEALLOCATE (work)
      CALL dgemv('T', SIZE(u, 1), SIZE(u, 2), 1._dp, u, SIZE(u, 1), b, 1, 0._dp, res1, 1)
      DO i = 1, 3
         res2(i) = res1(i)/diag(i)
      END DO
      CALL dgemv('T', 3, 3, 1._dp, vt, SIZE(vt, 1), res2, 1, 0._dp, res3, 1)
      res = -0.5*res3(2)/res3(3)
   END FUNCTION FindMin

! **************************************************************************************************
!> \brief Computes the Conjugate direction for the next search
!> \param gopt_env ...
!> \param Fletcher_Reeves ...
!> \param g contains the theta  of the previous step.. (norm 1.0 vector)
!> \param xi contains the -theta of the present step.. (norm 1.0 vector)
!> \param h contains the search direction of the previous step (must be orthogonal
!>            to nvec of the previous step (nvec_old))
!> \par   Info for DIMER method
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE get_conjugate_direction(gopt_env, Fletcher_Reeves, g, xi, h)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      LOGICAL, INTENT(IN)                                :: Fletcher_Reeves
      REAL(KIND=dp), DIMENSION(:), POINTER               :: g, xi, h

      CHARACTER(len=*), PARAMETER :: routineN = 'get_conjugate_direction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: check
      REAL(KIND=dp)                                      :: dgg, gam, gg, norm, norm_h
      TYPE(dimer_env_type), POINTER                      :: dimer_env

      CALL timeset(routineN, handle)
      NULLIFY (dimer_env)
      IF (.NOT. gopt_env%dimer_rotation) THEN
         gg = DOT_PRODUCT(g, g)
         IF (Fletcher_Reeves) THEN
            dgg = DOT_PRODUCT(xi, xi)
         ELSE
            dgg = DOT_PRODUCT((xi+g), xi)
         END IF
         gam = dgg/gg
         g = h
         h = -xi+gam*h
      ELSE
         dimer_env => gopt_env%dimer_env
         check = ABS(DOT_PRODUCT(g, g)-1.0_dp) < MAX(1.0E-9_dp, dimer_thrs)
         IF(.NOT.(check))CALL cp__a("motion/cg_utils.F",1058)

         check = ABS(DOT_PRODUCT(xi, xi)-1.0_dp) < MAX(1.0E-9_dp, dimer_thrs)
         IF(.NOT.(check))CALL cp__a("motion/cg_utils.F",1061)

         check = ABS(DOT_PRODUCT(h, dimer_env%cg_rot%nvec_old)) < MAX(1.0E-9_dp, dimer_thrs)
         IF(.NOT.(check))CALL cp__a("motion/cg_utils.F",1064)
         gg = dimer_env%cg_rot%norm_theta_old**2
         IF (Fletcher_Reeves) THEN
            dgg = dimer_env%cg_rot%norm_theta**2
         ELSE
            norm = dimer_env%cg_rot%norm_theta*dimer_env%cg_rot%norm_theta_old
            dgg = dimer_env%cg_rot%norm_theta**2+DOT_PRODUCT(g, xi)*norm
         END IF
         ! Compute Theta** and store it in nvec_old
         CALL rotate_dimer(dimer_env%cg_rot%nvec_old, g, dimer_env%rot%angle2+pi/2.0_dp)
         gam = dgg/gg
         g = h
         h = -xi*dimer_env%cg_rot%norm_theta+gam*dimer_env%cg_rot%norm_h*dimer_env%cg_rot%nvec_old
         h = h-DOT_PRODUCT(h, dimer_env%nvec)*dimer_env%nvec
         norm_h = SQRT(DOT_PRODUCT(h, h))
         IF (norm_h < EPSILON(0.0_dp)) THEN
            h = 0.0_dp
         ELSE
            h = h/norm_h
         END IF
         dimer_env%cg_rot%norm_h = norm_h
      END IF
      CALL timestop(handle)

   END SUBROUTINE get_conjugate_direction

END MODULE cg_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Split md_ener module from md_environment_type
!>  \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
MODULE md_ener_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE md_ener_type
      INTEGER       :: id_nr, ref_count
      INTEGER       :: nfree, nfree_shell
      REAL(KIND=dp) :: constant
      REAL(KIND=dp) :: delta_cons, delta_epot
      REAL(KIND=dp) :: epot
      REAL(KIND=dp) :: ekin, ekin_qm
      REAL(KIND=dp) :: temp_part, temp_qm
      REAL(KIND=dp) :: temp_baro
      REAL(KIND=dp) :: ekin_coefs
      REAL(KIND=dp) :: temp_coefs
      REAL(KIND=dp) :: ekin_shell, temp_shell
      REAL(KIND=dp) :: thermostat_part_kin, thermostat_part_pot
      REAL(KIND=dp) :: thermostat_fast_kin, thermostat_fast_pot
      REAL(KIND=dp) :: thermostat_slow_kin, thermostat_slow_pot
      REAL(KIND=dp) :: thermostat_baro_kin, thermostat_baro_pot
      REAL(KIND=dp) :: thermostat_coef_kin, thermostat_coef_pot
      REAL(KIND=dp) :: thermostat_shell_kin, thermostat_shell_pot
      REAL(KIND=dp) :: baro_kin, baro_pot
      REAL(KIND=dp) :: vcom(3), total_mass
      REAL(KIND=dp), DIMENSION(:), POINTER :: ekin_kind
      REAL(KIND=dp), DIMENSION(:), POINTER :: temp_kind
      INTEGER, DIMENSION(:), POINTER       :: nfree_kind
      REAL(KIND=dp), DIMENSION(:), POINTER :: ekin_shell_kind
      REAL(KIND=dp), DIMENSION(:), POINTER :: temp_shell_kind
      INTEGER, DIMENSION(:), POINTER       :: nfree_shell_kind
   END TYPE md_ener_type

! *** Public subroutines and data types ***
   PUBLIC :: create_md_ener, release_md_ener, retain_md_ener, md_ener_type, zero_md_ener

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_ener_types'
   INTEGER, SAVE, PRIVATE :: last_md_ener_id = 0

CONTAINS

! **************************************************************************************************
!> \brief retains the given md_ener structure
!> \param md_ener :
!> \par History
!>      10.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE create_md_ener(md_ener)
      TYPE(md_ener_type), POINTER                        :: md_ener

      CHARACTER(len=*), PARAMETER :: routineN = 'create_md_ener', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(md_ener)))CALL cp__a("motion/md_ener_types.F",70)
      ALLOCATE (md_ener)
      last_md_ener_id = last_md_ener_id+1
      md_ener%id_nr = last_md_ener_id

      md_ener%ref_count = 1

      NULLIFY (md_ener%ekin_kind)
      NULLIFY (md_ener%temp_kind)
      NULLIFY (md_ener%nfree_kind)
      NULLIFY (md_ener%ekin_shell_kind)
      NULLIFY (md_ener%temp_shell_kind)
      NULLIFY (md_ener%nfree_shell_kind)
   END SUBROUTINE create_md_ener

! **************************************************************************************************
!> \brief retains the given md_ener structure
!> \param md_ener :
!> \par History
!>      10.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE retain_md_ener(md_ener)
      TYPE(md_ener_type), POINTER                        :: md_ener

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_md_ener', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(md_ener)))CALL cp__a("motion/md_ener_types.F",97)
      IF(.NOT.(md_ener%ref_count > 0))CALL cp__a("motion/md_ener_types.F",98)
      md_ener%ref_count = md_ener%ref_count+1
   END SUBROUTINE retain_md_ener

! **************************************************************************************************
!> \brief releases the given md_ener structure
!> \param md_ener :
!> \par History
!>      10.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE release_md_ener(md_ener)
      TYPE(md_ener_type), POINTER                        :: md_ener

      CHARACTER(len=*), PARAMETER :: routineN = 'release_md_ener', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(md_ener)) THEN
         IF(.NOT.(md_ener%ref_count > 0))CALL cp__a("motion/md_ener_types.F",116)
         md_ener%ref_count = md_ener%ref_count-1
         IF (md_ener%ref_count == 0) THEN
            IF (ASSOCIATED(md_ener%temp_kind)) THEN
               DEALLOCATE (md_ener%temp_kind)
            END IF
            IF (ASSOCIATED(md_ener%ekin_kind)) THEN
               DEALLOCATE (md_ener%ekin_kind)
            END IF
            IF (ASSOCIATED(md_ener%nfree_kind)) THEN
               DEALLOCATE (md_ener%nfree_kind)
            END IF
            IF (ASSOCIATED(md_ener%temp_shell_kind)) THEN
               DEALLOCATE (md_ener%temp_shell_kind)
            END IF
            IF (ASSOCIATED(md_ener%ekin_shell_kind)) THEN
               DEALLOCATE (md_ener%ekin_shell_kind)
            END IF
            IF (ASSOCIATED(md_ener%nfree_shell_kind)) THEN
               DEALLOCATE (md_ener%nfree_shell_kind)
            END IF

            DEALLOCATE (md_ener)
         END IF
      END IF
   END SUBROUTINE release_md_ener

! **************************************************************************************************
!> \brief initialize to zero energies and temperatures
!> \param md_ener :
!> \param tkind ...
!> \param tshell ...
!> \par History
!>      10.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE zero_md_ener(md_ener, tkind, tshell)
      TYPE(md_ener_type), POINTER                        :: md_ener
      LOGICAL, INTENT(IN)                                :: tkind, tshell

      md_ener%ekin = 0.0_dp
      md_ener%temp_part = 0.0_dp
      md_ener%temp_baro = 0.0_dp
      md_ener%ekin_coefs = 0.0_dp
      md_ener%temp_coefs = 0.0_dp
      md_ener%ekin_qm = 0.0_dp
      md_ener%temp_qm = 0.0_dp
      md_ener%ekin_shell = 0.0_dp
      md_ener%temp_shell = 0.0_dp
      md_ener%constant = 0.0_dp
      md_ener%delta_cons = 0.0_dp
      md_ener%delta_epot = 0.0_dp
      md_ener%thermostat_part_kin = 0.0_dp
      md_ener%thermostat_part_pot = 0.0_dp
      md_ener%thermostat_fast_kin = 0.0_dp
      md_ener%thermostat_fast_pot = 0.0_dp
      md_ener%thermostat_slow_kin = 0.0_dp
      md_ener%thermostat_slow_pot = 0.0_dp
      md_ener%thermostat_coef_kin = 0.0_dp
      md_ener%thermostat_coef_pot = 0.0_dp
      md_ener%thermostat_baro_kin = 0.0_dp
      md_ener%thermostat_baro_pot = 0.0_dp
      md_ener%thermostat_shell_kin = 0.0_dp
      md_ener%thermostat_shell_pot = 0.0_dp
      md_ener%baro_kin = 0.0_dp
      md_ener%baro_pot = 0.0_dp
      IF (tkind) THEN
         md_ener%temp_kind = 0.0_dp
         md_ener%ekin_kind = 0.0_dp

         IF (tshell) THEN
            md_ener%temp_shell_kind = 0.0_dp
            md_ener%ekin_shell_kind = 0.0_dp
         END IF
      END IF
      md_ener%vcom(:) = 0.0_dp
      md_ener%total_mass = 0.0_dp
   END SUBROUTINE zero_md_ener

END MODULE md_ener_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles the type to compute averages during an MD
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
MODULE averages_types
   USE cell_types,                      ONLY: cell_type
   USE colvar_utils,                    ONLY: get_clv_force,&
                                              number_of_colvar
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE force_env_types,                 ONLY: force_env_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE md_ener_types,                   ONLY: md_ener_type
   USE virial_types,                    ONLY: virial_create,&
                                              virial_release,&
                                              virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE average_quantities_type
      INTEGER                                       :: id_nr, ref_count, itimes_start
      LOGICAL                                       :: do_averages
      TYPE(section_vals_type), POINTER              :: averages_section
      ! Real Scalar Quantities
      REAL(KIND=dp)                                 :: avetemp, avepot, avekin, &
                                                       avevol, aveca, avecb, avecc
      REAL(KIND=dp)                                 :: avetemp_baro, avehugoniot, avecpu
      REAL(KIND=dp)                                 :: aveal, avebe, avega, avepress, &
                                                       avekinc, avetempc, avepxx
      REAL(KIND=dp)                                 :: avetemp_qm, avekin_qm, econs
      ! Virial
      TYPE(virial_type), POINTER                    :: virial
      ! Colvar
      REAL(KIND=dp), POINTER, DIMENSION(:)          :: avecolvar
      REAL(KIND=dp), POINTER, DIMENSION(:)          :: aveMmatrix
   END TYPE average_quantities_type

! **************************************************************************************************
   INTERFACE get_averages
      MODULE PROCEDURE get_averages_rs, get_averages_rv, get_averages_rm
   END INTERFACE get_averages

! *** Public subroutines and data types ***
   PUBLIC :: average_quantities_type, create_averages, release_averages, &
             retain_averages, compute_averages

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'averages_types'
   INTEGER, SAVE, PRIVATE :: last_avg_env_id = 0

CONTAINS

! **************************************************************************************************
!> \brief Creates averages environment
!> \param averages ...
!> \param averages_section ...
!> \param virial_avg ...
!> \param force_env ...
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE create_averages(averages, averages_section, virial_avg, force_env)
      TYPE(average_quantities_type), POINTER             :: averages
      TYPE(section_vals_type), POINTER                   :: averages_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: virial_avg
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_averages', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nint
      LOGICAL                                            :: do_colvars

      ALLOCATE (averages)
      NULLIFY (averages%virial)
      NULLIFY (averages%avecolvar)
      NULLIFY (averages%aveMmatrix)
      ! Point to the averages section
      averages%averages_section => averages_section
      ! Initialize averages
      last_avg_env_id = last_avg_env_id+1
      averages%id_nr = last_avg_env_id
      averages%ref_count = 1
      averages%itimes_start = -1
      averages%avetemp = 0.0_dp
      averages%avepot = 0.0_dp
      averages%avekin = 0.0_dp
      averages%avevol = 0.0_dp
      averages%aveca = 0.0_dp
      averages%avecb = 0.0_dp
      averages%avecc = 0.0_dp
      averages%avetemp_baro = 0.0_dp
      averages%avehugoniot = 0.0_dp
      averages%avecpu = 0.0_dp
      averages%aveal = 0.0_dp
      averages%avebe = 0.0_dp
      averages%avega = 0.0_dp
      averages%avepress = 0.0_dp
      averages%avekinc = 0.0_dp
      averages%avetempc = 0.0_dp
      averages%avepxx = 0.0_dp
      averages%avetemp_qm = 0.0_dp
      averages%avekin_qm = 0.0_dp
      averages%econs = 0.0_dp
      CALL section_vals_val_get(averages_section, "_SECTION_PARAMETERS_", l_val=averages%do_averages)
      IF (averages%do_averages) THEN
         ! Setup Virial if requested
         IF (PRESENT(virial_avg)) THEN
            IF (virial_avg) CALL virial_create(averages%virial)
         END IF
         CALL section_vals_val_get(averages_section, "AVERAGE_COLVAR", l_val=do_colvars)
         ! Total number of COLVARs
         nint = 0
         IF (do_colvars) nint = number_of_colvar(force_env)
         ALLOCATE (averages%avecolvar(nint))
         ALLOCATE (averages%aveMmatrix(nint*nint))
         DO i = 1, nint
            averages%avecolvar(i) = 0.0_dp
         END DO
         DO i = 1, nint*nint
            averages%aveMmatrix(i) = 0.0_dp
         END DO
      END IF
   END SUBROUTINE create_averages

! **************************************************************************************************
!> \brief retains the given averages env
!> \param averages ...
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE retain_averages(averages)
      TYPE(average_quantities_type), POINTER             :: averages

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_averages', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(averages)))CALL cp__a("motion/averages_types.F",154)
      IF(.NOT.(averages%ref_count > 0))CALL cp__a("motion/averages_types.F",155)
      averages%ref_count = averages%ref_count+1
   END SUBROUTINE retain_averages

! **************************************************************************************************
!> \brief releases the given averages env
!> \param averages ...
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE release_averages(averages)
      TYPE(average_quantities_type), POINTER             :: averages

      CHARACTER(len=*), PARAMETER :: routineN = 'release_averages', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: work_section

      IF (ASSOCIATED(averages)) THEN
         IF(.NOT.(averages%ref_count > 0))CALL cp__a("motion/averages_types.F",173)
         averages%ref_count = averages%ref_count-1
         IF (averages%ref_count == 0) THEN
            CALL virial_release(averages%virial)
            IF (ASSOCIATED(averages%avecolvar)) THEN
               DEALLOCATE (averages%avecolvar)
            END IF
            IF (ASSOCIATED(averages%aveMmatrix)) THEN
               DEALLOCATE (averages%aveMmatrix)
            END IF
            ! Removes restart values from the corresponding restart section..
            work_section => section_vals_get_subs_vals(averages%averages_section, "RESTART_AVERAGES")
            CALL section_vals_remove_values(work_section)
            NULLIFY (averages%averages_section)
            !
            DEALLOCATE (averages)
         END IF
      END IF

   END SUBROUTINE release_averages

! **************************************************************************************************
!> \brief computes the averages
!> \param averages ...
!> \param force_env ...
!> \param md_ener ...
!> \param cell ...
!> \param virial ...
!> \param pv_scalar ...
!> \param pv_xx ...
!> \param used_time ...
!> \param hugoniot ...
!> \param abc ...
!> \param cell_angle ...
!> \param nat ...
!> \param itimes ...
!> \param time ...
!> \param my_pos ...
!> \param my_act ...
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE compute_averages(averages, force_env, md_ener, cell, virial, &
                               pv_scalar, pv_xx, used_time, hugoniot, abc, cell_angle, nat, itimes, &
                               time, my_pos, my_act)
      TYPE(average_quantities_type), POINTER             :: averages
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(cell_type), POINTER                           :: cell
      TYPE(virial_type), POINTER                         :: virial
      REAL(KIND=dp), INTENT(IN)                          :: pv_scalar, pv_xx
      REAL(KIND=dp), POINTER                             :: used_time
      REAL(KIND=dp), INTENT(IN)                          :: hugoniot
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: abc, cell_angle
      INTEGER, INTENT(IN)                                :: nat, itimes
      REAL(KIND=dp), INTENT(IN)                          :: time
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: my_pos, my_act

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_averages', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: ctmp
      INTEGER                                            :: delta_t, handle, i, nint, output_unit
      LOGICAL                                            :: restart_averages
      REAL(KIND=dp)                                      :: start_time
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cvalues, Mmatrix, tmp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: restart_section

      CALL timeset(routineN, handle)
      CALL section_vals_val_get(averages%averages_section, "ACQUISITION_START_TIME", &
                                r_val=start_time)
      IF (averages%do_averages) THEN
         NULLIFY (cvalues, Mmatrix, logger)
         logger => cp_get_default_logger()
         ! Determine the nr. of internal colvar (if any/requested)
         nint = 0
         IF (ASSOCIATED(averages%avecolvar)) nint = SIZE(averages%avecolvar)

         ! Evaluate averages if we collected enough statistics (user defined)
         IF (time >= start_time) THEN

            ! Handling properly the restart
            IF (averages%itimes_start == -1) THEN
               restart_section => section_vals_get_subs_vals(averages%averages_section, "RESTART_AVERAGES")
               CALL section_vals_get(restart_section, explicit=restart_averages)
               IF (restart_averages) THEN
                  CALL section_vals_val_get(restart_section, "ITIMES_START", i_val=averages%itimes_start)
                  CALL section_vals_val_get(restart_section, "AVECPU", r_val=averages%avecpu)
                  CALL section_vals_val_get(restart_section, "AVEHUGONIOT", r_val=averages%avehugoniot)
                  CALL section_vals_val_get(restart_section, "AVETEMP_BARO", r_val=averages%avetemp_baro)
                  CALL section_vals_val_get(restart_section, "AVEPOT", r_val=averages%avepot)
                  CALL section_vals_val_get(restart_section, "AVEKIN", r_val=averages%avekin)
                  CALL section_vals_val_get(restart_section, "AVETEMP", r_val=averages%avetemp)
                  CALL section_vals_val_get(restart_section, "AVEKIN_QM", r_val=averages%avekin_qm)
                  CALL section_vals_val_get(restart_section, "AVETEMP_QM", r_val=averages%avetemp_qm)
                  CALL section_vals_val_get(restart_section, "AVEVOL", r_val=averages%avevol)
                  CALL section_vals_val_get(restart_section, "AVECELL_A", r_val=averages%aveca)
                  CALL section_vals_val_get(restart_section, "AVECELL_B", r_val=averages%avecb)
                  CALL section_vals_val_get(restart_section, "AVECELL_C", r_val=averages%avecc)
                  CALL section_vals_val_get(restart_section, "AVEALPHA", r_val=averages%aveal)
                  CALL section_vals_val_get(restart_section, "AVEBETA", r_val=averages%avebe)
                  CALL section_vals_val_get(restart_section, "AVEGAMMA", r_val=averages%avega)
                  CALL section_vals_val_get(restart_section, "AVE_ECONS", r_val=averages%econs)
                  ! Virial
                  IF (virial%pv_availability) THEN
                     CALL section_vals_val_get(restart_section, "AVE_PRESS", r_val=averages%avepress)
                     CALL section_vals_val_get(restart_section, "AVE_PXX", r_val=averages%avepxx)
                     IF (ASSOCIATED(averages%virial)) THEN
                        CALL section_vals_val_get(restart_section, "AVE_PV_TOT", r_vals=tmp)
                        averages%virial%pv_total = RESHAPE(tmp, (/3, 3/))
                        CALL section_vals_val_get(restart_section, "AVE_PV_VIR", r_vals=tmp)
                        averages%virial%pv_virial = RESHAPE(tmp, (/3, 3/))
                        CALL section_vals_val_get(restart_section, "AVE_PV_KIN", r_vals=tmp)
                        averages%virial%pv_kinetic = RESHAPE(tmp, (/3, 3/))
                        CALL section_vals_val_get(restart_section, "AVE_PV_CNSTR", r_vals=tmp)
                        averages%virial%pv_constraint = RESHAPE(tmp, (/3, 3/))
                        CALL section_vals_val_get(restart_section, "AVE_PV_XC", r_vals=tmp)
                        averages%virial%pv_xc = RESHAPE(tmp, (/3, 3/))
                        CALL section_vals_val_get(restart_section, "AVE_PV_FOCK_4C", r_vals=tmp)
                        averages%virial%pv_fock_4c = RESHAPE(tmp, (/3, 3/))
                     END IF
                  END IF
                  ! Colvars
                  IF (nint > 0) THEN
                     CALL section_vals_val_get(restart_section, "AVE_COLVARS", r_vals=cvalues)
                     CALL section_vals_val_get(restart_section, "AVE_MMATRIX", r_vals=Mmatrix)
                     IF(.NOT.(nint == SIZE(cvalues)))CALL cp__a("motion/averages_types.F",299)
                     IF(.NOT.(nint*nint == SIZE(Mmatrix)))CALL cp__a("motion/averages_types.F",300)
                     averages%avecolvar = cvalues
                     averages%aveMmatrix = Mmatrix
                  END IF
               ELSE
                  averages%itimes_start = itimes
               END IF
            END IF
            delta_t = itimes-averages%itimes_start+1

            ! Perform averages
            SELECT CASE (delta_t)
            CASE (1)
               averages%avecpu = used_time
               averages%avehugoniot = hugoniot
               averages%avetemp_baro = md_ener%temp_baro
               averages%avepot = md_ener%epot
               averages%avekin = md_ener%ekin
               averages%avetemp = md_ener%temp_part
               averages%avekin_qm = md_ener%ekin_qm
               averages%avetemp_qm = md_ener%temp_qm
               averages%avevol = cell%deth
               averages%aveca = abc(1)
               averages%avecb = abc(2)
               averages%avecc = abc(3)
               averages%aveal = cell_angle(3)
               averages%avebe = cell_angle(2)
               averages%avega = cell_angle(1)
               averages%econs = 0._dp
               ! Virial
               IF (virial%pv_availability) THEN
                  averages%avepress = pv_scalar
                  averages%avepxx = pv_xx
                  IF (ASSOCIATED(averages%virial)) THEN
                     averages%virial%pv_total = virial%pv_total
                     averages%virial%pv_virial = virial%pv_virial
                     averages%virial%pv_kinetic = virial%pv_kinetic
                     averages%virial%pv_constraint = virial%pv_constraint
                     averages%virial%pv_xc = virial%pv_xc
                     averages%virial%pv_fock_4c = virial%pv_fock_4c
                  END IF
               END IF
               ! Colvars
               IF (nint > 0) THEN
                  CALL get_clv_force(force_env, nsize_xyz=nat*3, nsize_int=nint, &
                                     cvalues=averages%avecolvar, Mmatrix=averages%aveMmatrix)
               END IF
            CASE DEFAULT
               CALL get_averages(averages%avecpu, used_time, delta_t)
               CALL get_averages(averages%avehugoniot, hugoniot, delta_t)
               CALL get_averages(averages%avetemp_baro, md_ener%temp_baro, delta_t)
               CALL get_averages(averages%avepot, md_ener%epot, delta_t)
               CALL get_averages(averages%avekin, md_ener%ekin, delta_t)
               CALL get_averages(averages%avetemp, md_ener%temp_part, delta_t)
               CALL get_averages(averages%avekin_qm, md_ener%ekin_qm, delta_t)
               CALL get_averages(averages%avetemp_qm, md_ener%temp_qm, delta_t)
               CALL get_averages(averages%avevol, cell%deth, delta_t)
               CALL get_averages(averages%aveca, abc(1), delta_t)
               CALL get_averages(averages%avecb, abc(2), delta_t)
               CALL get_averages(averages%avecc, abc(3), delta_t)
               CALL get_averages(averages%aveal, cell_angle(3), delta_t)
               CALL get_averages(averages%avebe, cell_angle(2), delta_t)
               CALL get_averages(averages%avega, cell_angle(1), delta_t)
               CALL get_averages(averages%econs, md_ener%delta_cons, delta_t)
               ! Virial
               IF (virial%pv_availability) THEN
                  CALL get_averages(averages%avepress, pv_scalar, delta_t)
                  CALL get_averages(averages%avepxx, pv_xx, delta_t)
                  IF (ASSOCIATED(averages%virial)) THEN
                     CALL get_averages(averages%virial%pv_total, virial%pv_total, delta_t)
                     CALL get_averages(averages%virial%pv_virial, virial%pv_virial, delta_t)
                     CALL get_averages(averages%virial%pv_kinetic, virial%pv_kinetic, delta_t)
                     CALL get_averages(averages%virial%pv_constraint, virial%pv_constraint, delta_t)
                     CALL get_averages(averages%virial%pv_xc, virial%pv_xc, delta_t)
                     CALL get_averages(averages%virial%pv_fock_4c, virial%pv_fock_4c, delta_t)
                  END IF
               END IF
               ! Colvars
               IF (nint > 0) THEN
                  ALLOCATE (cvalues(nint))
                  ALLOCATE (Mmatrix(nint*nint))
                  CALL get_clv_force(force_env, nsize_xyz=nat*3, nsize_int=nint, cvalues=cvalues, &
                                     Mmatrix=Mmatrix)
                  CALL get_averages(averages%avecolvar, cvalues, delta_t)
                  CALL get_averages(averages%aveMmatrix, Mmatrix, delta_t)
                  DEALLOCATE (cvalues)
                  DEALLOCATE (Mmatrix)
               END IF
            END SELECT
         END IF

         ! Possibly print averages
         output_unit = cp_print_key_unit_nr(logger, averages%averages_section, "PRINT_AVERAGES", &
                                            extension=".avg", file_position=my_pos, file_action=my_act)
         IF (output_unit > 0) THEN
            WRITE (output_unit, FMT='(A15,1X,"=",1X,G15.9,"   NSTEP #",I15)') &
               "AVECPU", averages%avecpu, itimes, &
               "AVEHUGONIOT", averages%avehugoniot, itimes, &
               "AVETEMP_BARO", averages%avetemp_baro, itimes, &
               "AVEPOT", averages%avepot, itimes, &
               "AVEKIN", averages%avekin, itimes, &
               "AVETEMP", averages%avetemp, itimes, &
               "AVEKIN_QM", averages%avekin_qm, itimes, &
               "AVETEMP_QM", averages%avetemp_qm, itimes, &
               "AVEVOL", averages%avevol, itimes, &
               "AVECELL_A", averages%aveca, itimes, &
               "AVECELL_B", averages%avecb, itimes, &
               "AVECELL_C", averages%avecc, itimes, &
               "AVEALPHA", averages%aveal, itimes, &
               "AVEBETA", averages%avebe, itimes, &
               "AVEGAMMA", averages%avega, itimes, &
               "AVE_ECONS", averages%econs, itimes
            ! Print the virial
            IF (virial%pv_availability) THEN
               WRITE (output_unit, FMT='(A15,1X,"=",1X,G15.9,"   NSTEP #",I15)') &
                  "AVE_PRESS", averages%avepress, itimes, &
                  "AVE_PXX", averages%avepxx, itimes
               IF (ASSOCIATED(averages%virial)) THEN
                  WRITE (output_unit, FMT='(A15,1X,"=",1X,G15.9,"   NSTEP #",I15)') &
                     "AVE_PV_TOT", averages%virial%pv_total, itimes, &
                     "AVE_PV_VIR", averages%virial%pv_virial, itimes, &
                     "AVE_PV_KIN", averages%virial%pv_kinetic, itimes, &
                     "AVE_PV_CNSTR", averages%virial%pv_constraint, itimes, &
                     "AVE_PV_XC", averages%virial%pv_xc, itimes, &
                     "AVE_PV_FOCK_4C", averages%virial%pv_fock_4c, itimes
               END IF
            END IF
            DO i = 1, nint
               ctmp = cp_to_string(i)
               WRITE (output_unit, FMT='(A15,1X,"=",1X,G15.9,"   NSTEP #",I15)') &
                  TRIM("AVE_CV-"//ADJUSTL(ctmp)), averages%avecolvar(i), itimes
            END DO
            WRITE (output_unit, FMT='(/)')
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, averages%averages_section, &
                                           "PRINT_AVERAGES")
      END IF
      CALL timestop(handle)
   END SUBROUTINE compute_averages

! **************************************************************************************************
!> \brief computes the averages - low level for REAL
!> \param avg ...
!> \param add ...
!> \param delta_t ...
!> \author Teodoro Laino [tlaino] - 03.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_averages_rs(avg, add, delta_t)
      REAL(KIND=dp), INTENT(INOUT)                       :: avg
      REAL(KIND=dp), INTENT(IN)                          :: add
      INTEGER, INTENT(IN)                                :: delta_t

      CHARACTER(len=*), PARAMETER :: routineN = 'get_averages_rs', &
         routineP = moduleN//':'//routineN

      avg = (avg*REAL(delta_t-1, dp)+add)/REAL(delta_t, dp)
   END SUBROUTINE get_averages_rs

! **************************************************************************************************
!> \brief computes the averages - low level for REAL vector
!> \param avg ...
!> \param add ...
!> \param delta_t ...
!> \author Teodoro Laino [tlaino] - 10.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_averages_rv(avg, add, delta_t)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: avg
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: add
      INTEGER, INTENT(IN)                                :: delta_t

      CHARACTER(len=*), PARAMETER :: routineN = 'get_averages_rv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: check

      check = SIZE(avg) == SIZE(add)
      IF(.NOT.(check))CALL cp__a("motion/averages_types.F",477)
      DO i = 1, SIZE(avg)
         avg(i) = (avg(i)*REAL(delta_t-1, dp)+add(i))/REAL(delta_t, dp)
      END DO
   END SUBROUTINE get_averages_rv

! **************************************************************************************************
!> \brief computes the averages - low level for REAL matrix
!> \param avg ...
!> \param add ...
!> \param delta_t ...
!> \author Teodoro Laino [tlaino] - 10.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_averages_rm(avg, add, delta_t)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: avg
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: add
      INTEGER, INTENT(IN)                                :: delta_t

      CHARACTER(len=*), PARAMETER :: routineN = 'get_averages_rm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      LOGICAL                                            :: check

      check = SIZE(avg, 1) == SIZE(add, 1)
      IF(.NOT.(check))CALL cp__a("motion/averages_types.F",502)
      check = SIZE(avg, 2) == SIZE(add, 2)
      IF(.NOT.(check))CALL cp__a("motion/averages_types.F",504)
      DO i = 1, SIZE(avg, 2)
         DO j = 1, SIZE(avg, 1)
            avg(j, i) = (avg(j, i)*REAL(delta_t-1, dp)+add(j, i))/REAL(delta_t, dp)
         END DO
      END DO
   END SUBROUTINE get_averages_rm

END MODULE averages_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Typo for Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
MODULE neb_types

   USE force_env_types,                 ONLY: force_env_type
   USE input_constants,                 ONLY: do_band_cartesian,&
                                              do_band_collective
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_types'

   PUBLIC :: neb_type, &
             neb_var_type, &
             neb_var_create, &
             neb_var_release

! **************************************************************************************************
   TYPE neb_type
      ! NEB parameters
      INTEGER       :: id_type
      INTEGER       :: opt_type
      INTEGER       :: pot_type
      INTEGER       :: number_of_replica, nsize_xyz, nsize_int
      INTEGER       :: nsteps_it, istep
      INTEGER       :: nr_HE_image
      LOGICAL       :: rotate_frames, align_frames
      LOGICAL       :: optimize_end_points
      LOGICAL       :: use_colvar
      LOGICAL       :: reparametrize_frames
      INTEGER       :: spline_order
      REAL(KIND=dp) :: K, spring_energy, avg_distance
      REAL(KIND=dp) :: smoothing
      CHARACTER(LEN=default_string_length) :: opt_type_label
      ! Section used for restart
      TYPE(force_env_type), POINTER        :: force_env
      TYPE(section_vals_type), POINTER     :: root_section
      TYPE(section_vals_type), POINTER     :: motion_print_section
      TYPE(section_vals_type), POINTER     :: force_env_section
      TYPE(section_vals_type), POINTER     :: neb_section
   END TYPE neb_type

! **************************************************************************************************
   TYPE neb_var_type
      INTEGER :: in_use, size_wrk(2)
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: xyz, int, wrk
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER  :: Mmatrix
   END TYPE neb_var_type

CONTAINS
! **************************************************************************************************
!> \brief Creates a variable type for BAND calculation
!> \param neb_var ...
!> \param neb_env ...
!> \param full_allocation ...
!> \date   05.2007
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE neb_var_create(neb_var, neb_env, full_allocation)
      TYPE(neb_var_type), POINTER                        :: neb_var
      TYPE(neb_type), POINTER                            :: neb_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: full_allocation

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_var_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: neb_nr_replica
      LOGICAL                                            :: allocate_all

      IF(.NOT.(.NOT. ASSOCIATED(neb_var)))CALL cp__a("motion/neb_types.F",88)
      allocate_all = .FALSE.
      IF (PRESENT(full_allocation)) allocate_all = full_allocation
      neb_nr_replica = neb_env%number_of_replica
      ALLOCATE (neb_var)
      NULLIFY (neb_var%xyz, neb_var%int, neb_var%wrk, neb_var%Mmatrix)
      IF (allocate_all) THEN
         ALLOCATE (neb_var%xyz(neb_env%nsize_xyz, neb_nr_replica))
         neb_var%xyz = 0.0_dp
      END IF
      IF (neb_env%use_colvar) THEN
         neb_var%in_use = do_band_collective
         IF(.NOT.(neb_env%nsize_int > 0))CALL cp__a("motion/neb_types.F",100)
         ALLOCATE (neb_var%int(neb_env%nsize_int, neb_nr_replica))
         neb_var%int = 0.0_dp
         neb_var%wrk => neb_var%int
      ELSE
         neb_var%in_use = do_band_cartesian
         IF (.NOT. allocate_all) THEN
            ALLOCATE (neb_var%xyz(neb_env%nsize_xyz, neb_nr_replica))
            neb_var%xyz = 0.0_dp
         END IF
         neb_var%wrk => neb_var%xyz
      END IF
      neb_var%size_wrk(1) = SIZE(neb_var%wrk, 1)
      neb_var%size_wrk(2) = SIZE(neb_var%wrk, 2)

   END SUBROUTINE neb_var_create

! **************************************************************************************************
!> \brief Releases a variable type for BAND calculation
!> \param neb_var ...
!> \date   05.2007
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE neb_var_release(neb_var)
      TYPE(neb_var_type), POINTER                        :: neb_var

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_var_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(neb_var)))CALL cp__a("motion/neb_types.F",129)
      IF (ASSOCIATED(neb_var%xyz)) THEN
         DEALLOCATE (neb_var%xyz)
      END IF
      IF (neb_var%in_use == do_band_collective) THEN
         DEALLOCATE (neb_var%int)
      END IF
      NULLIFY (neb_var%wrk)
      DEALLOCATE (neb_var)

   END SUBROUTINE neb_var_release

END MODULE neb_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief
!> \par History
!> \author  MI 07.2009
! **************************************************************************************************
MODULE gle_system_types
   USE bibliography,                    ONLY: Ceriotti2009,&
                                              Ceriotti2009b,&
                                              cite_reference
   USE extended_system_types,           ONLY: create_map_info_type,&
                                              map_info_type,&
                                              release_map_info_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_rng_seed,&
                                              rng_stream_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: gle_dealloc, &
             gle_init, gle_thermo_create, &
             gle_type

!
   TYPE gle_thermo_type
      INTEGER                                 :: degrees_of_freedom
      REAL(KIND=dp)                           :: nkt, kin_energy, thermostat_energy
      REAL(KIND=dp), DIMENSION(:), POINTER    :: s
      TYPE(rng_stream_type), POINTER          :: gaussian_rng_stream
   END TYPE gle_thermo_type

! **************************************************************************************************
   TYPE gle_type
      INTEGER                                 :: ndim
      INTEGER                                 :: glob_num_gle, loc_num_gle, region
      INTEGER, DIMENSION(:), POINTER          :: mal
      REAL(dp)                                :: temp, dt, dt_fact
      REAL(dp), POINTER                       :: gle_s(:, :), gle_t(:, :)
      REAL(dp), POINTER                       :: a_mat(:, :), c_mat(:, :)
      TYPE(gle_thermo_type), POINTER          :: nvt(:)
      TYPE(map_info_type), POINTER            :: map_info
   END TYPE gle_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gle_system_types'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gle ...
!> \param dt ...
!> \param temp ...
!> \param section ...
!> \param
! **************************************************************************************************
   SUBROUTINE gle_init(gle, dt, temp, section)
      TYPE(gle_type), POINTER                            :: gle
      REAL(dp), INTENT(IN)                               :: dt, temp
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gle_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ir, j, k, n_rep
      REAL(dp), DIMENSION(:), POINTER                    :: list
      REAL(KIND=dp)                                      :: a_scale

      NULLIFY (gle%nvt)
      NULLIFY (gle%gle_s)
      NULLIFY (gle%gle_t)
      NULLIFY (gle%map_info)
      gle%loc_num_gle = 0
      gle%glob_num_gle = 0
      gle%temp = temp
      gle%dt = dt*0.5_dp

      CALL cite_reference(Ceriotti2009)
      CALL cite_reference(Ceriotti2009b)
      CALL section_vals_val_get(section, "NDIM", i_val=gle%ndim)
      CALL section_vals_val_get(section, "A_SCALE", r_val=a_scale)

      ALLOCATE (gle%a_mat(gle%ndim, gle%ndim))
      ALLOCATE (gle%c_mat(gle%ndim, gle%ndim))
      ALLOCATE (gle%gle_s(gle%ndim, gle%ndim))
      ALLOCATE (gle%gle_t(gle%ndim, gle%ndim))

      CALL section_vals_val_get(section, "A_LIST", n_rep_val=n_rep)

      j = 1
      k = 1
      DO ir = 1, n_rep
         NULLIFY (list)
         CALL section_vals_val_get(section, "A_LIST", &
                                   i_rep_val=ir, r_vals=list)

         IF (ASSOCIATED(list)) THEN
            DO i = 1, SIZE(list)
               IF (j > gle%ndim) THEN
                  CALL cp__b("gle_system_types.F",110,"GLE: Too many elements in A_LIST")
               END IF
               gle%a_mat(j, k) = list(i)
               k = k+1
               IF (k > gle%ndim) THEN
                  k = 1
                  j = j+1
               END IF
            END DO
         END IF
      END DO ! ir
      IF (j < gle%ndim+1) THEN
         CALL cp__b("gle_system_types.F",122,"GLE: Too few elements in A_LIST")
      END IF
      gle%a_mat = gle%a_mat*a_scale

      CALL section_vals_val_get(section, "C_LIST", n_rep_val=n_rep)
      IF (n_rep > 0) THEN
         j = 1
         k = 1
         DO ir = 1, n_rep
            NULLIFY (list)
            CALL section_vals_val_get(section, "C_LIST", &
                                      i_rep_val=ir, r_vals=list)

            IF (ASSOCIATED(list)) THEN
               DO i = 1, SIZE(list)
                  IF (j > gle%ndim) THEN
                     CALL cp__b("gle_system_types.F",138,"GLE: Too many elements in C_LIST")
                  END IF
                  gle%c_mat(j, k) = list(i)
                  k = k+1
                  IF (k > gle%ndim) THEN
                     k = 1
                     j = j+1
                  END IF
               END DO
            END IF
         END DO ! ir
         IF (j < gle%ndim+1) THEN
            CALL cp__b("gle_system_types.F",150,"GLE: Too few elements in C_LIST")
         END IF
      ELSE
         gle%c_mat = 0.0_dp
         DO i = 1, gle%ndim
            gle%c_mat(i, i) = gle%temp
         END DO
      END IF
      CALL create_map_info_type(gle%map_info)
   END SUBROUTINE gle_init

! **************************************************************************************************
!> \brief ...
!> \param gle ...
!> \param mal_size ...
!> \param
! **************************************************************************************************
   SUBROUTINE gle_thermo_create(gle, mal_size)
      TYPE(gle_type), POINTER                            :: gle
      INTEGER, INTENT(IN)                                :: mal_size

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gle_thermo_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=40)                                  :: name
      INTEGER                                            :: i, ithermo, my_index
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: seed
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed, my_seed

      IF(.NOT.(ASSOCIATED(gle)))CALL cp__a("gle_system_types.F",179)
      IF(.NOT.(.NOT. ASSOCIATED(gle%nvt)))CALL cp__a("gle_system_types.F",180)

      ALLOCATE (gle%nvt(gle%loc_num_gle))
      DO i = 1, gle%loc_num_gle
         NULLIFY (gle%nvt(i)%s)
         ALLOCATE (gle%nvt(i)%s(gle%ndim))
         gle%nvt(i)%kin_energy = 0.0_dp
         gle%nvt(i)%thermostat_energy = 0.0_dp
      END DO

      ALLOCATE (gle%mal(mal_size))
      gle%mal(:) = 0

      ! Initialize the gaussian stream random number
      initial_seed = next_rng_seed()
      ALLOCATE (seed(3, 2, gle%glob_num_gle))

      seed(:, :, 1) = initial_seed
      DO ithermo = 2, gle%glob_num_gle
         seed(:, :, ithermo) = next_rng_seed(seed(:, :, ithermo-1))
      END DO

      ! Update initial seed
      initial_seed = next_rng_seed(seed(:, :, gle%glob_num_gle))
      DO ithermo = 1, gle%loc_num_gle
         NULLIFY (gle%nvt(ithermo)%gaussian_rng_stream)
         my_index = gle%map_info%index(ithermo)
         my_seed = seed(:, :, my_index)
         WRITE (UNIT=name, FMT="(A,I8)") "Wiener process for Thermostat #", my_index
         CALL compress(name)
         CALL create_rng_stream(rng_stream=gle%nvt(ithermo)%gaussian_rng_stream, &
                                name=name, distribution_type=GAUSSIAN, extended_precision=.TRUE., &
                                seed=my_seed)
      END DO

      DEALLOCATE (seed)

   END SUBROUTINE gle_thermo_create

! **************************************************************************************************
!> \brief Deallocate type for GLE thermostat
!> \param gle ...
! **************************************************************************************************
   SUBROUTINE gle_dealloc(gle)
      TYPE(gle_type), POINTER                            :: gle

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gle_dealloc', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(gle)) THEN
         IF (ASSOCIATED(gle%a_mat)) THEN
            DEALLOCATE (gle%a_mat)
         END IF
         IF (ASSOCIATED(gle%c_mat)) THEN
            DEALLOCATE (gle%c_mat)
         END IF
         IF (ASSOCIATED(gle%gle_t)) THEN
            DEALLOCATE (gle%gle_t)
         END IF
         IF (ASSOCIATED(gle%gle_s)) THEN
            DEALLOCATE (gle%gle_s)
         END IF
         IF (ASSOCIATED(gle%nvt)) THEN
            DO i = 1, SIZE(gle%nvt)
               DEALLOCATE (gle%nvt(i)%s)
               IF (ASSOCIATED(gle%nvt(i)%gaussian_rng_stream)) THEN
                  CALL delete_rng_stream(gle%nvt(i)%gaussian_rng_stream)
               END IF
            END DO
            DEALLOCATE (gle%nvt)
         ENDIF
         IF (ASSOCIATED(gle%mal)) THEN
            DEALLOCATE (gle%mal)
         END IF

         CALL release_map_info_type(gle%map_info)
         DEALLOCATE (gle)
      ENDIF

   END SUBROUTINE gle_dealloc

END MODULE gle_system_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE pint_types

   USE cp_log_handling,                 ONLY: cp_logger_type
   USE gle_system_types,                ONLY: gle_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: rng_stream_type
   USE replica_types,                   ONLY: replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Energy contributions - symbolic names for indexing energy arrays
   INTEGER, PARAMETER, PUBLIC :: e_conserved_id = 1, &
                                 e_potential_id = 2, &
                                 e_kin_thermo_id = 3, &
                                 e_kin_virial_id = 4

   ! Number of energy contributions for static array allocation
   INTEGER, PARAMETER, PUBLIC :: e_num_ids = 4

   INTEGER, PARAMETER, PUBLIC :: thermostat_none = 0, &
                                 thermostat_nose = 1, &
                                 thermostat_gle = 2, &
                                 thermostat_pile = 3, &
                                 thermostat_piglet = 4

   PUBLIC :: pint_env_type
   PUBLIC :: normalmode_env_type
   PUBLIC :: staging_env_type
   PUBLIC :: pile_therm_type
   PUBLIC :: piglet_therm_type
   PUBLIC :: pint_propagator_type

   ! ***************************************************************************
   !> \brief environment for a path integral run
   !> \param ref_count reference count of this data structure
   !> \param id_nr identification number of this data structure
   !> \param p number of replicas/beads
   !> \param nnos nose hoover chain length
   !> \param nrespa number of respa steps
   !> \param nsteps - number of PIMD steps to be performed
   !> \param iter current iteration number
   !> \param ndim number of coordinates per replica/bead
   !> \param transform type of transform (normalmode or staging)
   !> \param t_tol temperature tolerance for rescaling
   !> \param v_tol velocity tolerance for rescaling
   !> \param kT boltzmann factor times temperature (simulation temperature
   !> \param    not necessarily the physical temperature)
   !> \param beta 1/kT (physical temperature)
   !> \param dt time step for dynamic
   !> \param e_pot_h potential energy in harmonic springs
   !> \param e_kin_beads (fictious) kinetic energy of the beads
   !> \param e_pot_t potential energy of thermostats
   !> \param e_kin_t kinetic energy of thermostats
   !> \param energy - energy contributions updated every step REAL(e_num_ids)
   !> \param    e_kin_virial_id - virial estimator of the (real) kinetic energy
   !> \param t current simulation time
   !> \param replicas replica environment for force calculations
   !> \param input input data structure
   !> \param staging_env description for the staging transformation
   !> \param normalmode_env description for the normal mode transformation
   !> \param randomG random number stream descriptor
   !> \param mass real masses
   !> \param e_pot_bead array with last energies from QS per replica
   !> \param x array with real space coordinates (P, 3*N)
   !> \param v array with real space velocities
   !> \param f array with real space forces
   !> \param mass_beads masses of the beads for harmonic forces (harmonic mass)
   !> \param mass_fict fictious mass of the beads for dynamics (kinetic mass)
   !> \param ux array with transformed space coordinates (P, 3*N)
   !> \param uv array with transformed velocities
   !> \param uv_t array with temporary transformed velocities
   !> \param uv_new array with new transformed velocities
   !> \param uf array with transformed accelerations (QS part)
   !> \param uf_h array with harmonic part transformed forces
   !> \param tx nose hoover chain positions (pint_env%nnos,pint_env%p,pint_env%ndim)
   !> \param tv nose hoover chain velocities
   !> \param tv_t nose hoover chain velocities (temporary)
   !> \param tv_old nose hoover chain velocities (older)
   !> \param tv_new nose hoover chain velocities (newer)
   !> \param tf nose hoover chain forces (?)
   !> \param Q nose hoover chain masses
   !> \param time_per_step - time per step in seconds (updated every step)
   !> \param pile_therm data used for the pile thermostat
   !> \param wsinex omega*sin(omega*deltat) for exact harminic integrator
   !> \param iwsinex 1/omega*sin(omega*deltat) for exact harminic integrator
   !> \param cosex cos(omega*deltat) for exact harminic integrator
   !> \param propagator contains propagator related constants
   !> \param harm_integrator selects between numeric and exact harmonic integrator scheme
   !> \param first_propagated_mode if 1 - propagate all normal modes,
   !>                              if 2 - keep centoid fixed
   !> \author fawzi
   !> \par History
   !>      Added some comments - hforbert
   !>      Added normal mode transformation - hforbert
   !>      2009-06-15 helium_solvent_type object is no longer a member of
   !>                   pint_env_type [lwalewski]
   !>      2014-10-23 added pile_therm [Felix Uhl]
   ! ***************************************************************************
   TYPE pint_env_type
      INTEGER :: ref_count, id_nr, p, nnos, nrespa, iter, ndim, transform
      INTEGER :: first_step, last_step, num_steps, first_propagated_mode
      INTEGER :: pimd_thermostat, harm_integrator, thermostat_rng_seed
      REAL(KIND=dp) :: t_tol, v_tol, kT, beta, dt, &
                       e_gle, e_pile, e_piglet, e_pot_h, e_kin_beads, e_pot_t, e_kin_t, t, time_per_step
      TYPE(cp_logger_type), POINTER :: logger
      TYPE(replica_env_type), POINTER :: replicas
      TYPE(section_vals_type), POINTER :: input
      TYPE(staging_env_type), POINTER :: staging_env
      TYPE(normalmode_env_type), POINTER :: normalmode_env
      TYPE(rng_stream_type), POINTER :: randomG
      TYPE(gle_type), POINTER        :: gle
      REAL(KIND=dp), DIMENSION(e_num_ids) :: energy
      REAL(KIND=dp), DIMENSION(:), POINTER :: mass, e_pot_bead
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: x, v, f, mass_beads, &
                                                 mass_fict, ux, ux_t, uv, uv_t, uv_new, uf, uf_h, external_f
      REAL(KIND=dp), DIMENSION(:), POINTER :: centroid
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: tx, tv, tv_t, tv_old, tv_new, tf
      REAL(KIND=dp), DIMENSION(:), POINTER :: Q ! dim p, make it (p,ndim)?
      REAL(KIND=dp), DIMENSION(:), POINTER :: rtmp_ndim, rtmp_natom
      REAL(KIND=dp), DIMENSION(:), POINTER :: iwsinex, wsinex, cosex
      TYPE(pile_therm_type), POINTER :: pile_therm
      TYPE(piglet_therm_type), POINTER :: piglet_therm
      TYPE(pint_propagator_type), POINTER :: propagator
   END TYPE pint_env_type

   ! ***************************************************************************
   !> \brief data to perform the normalmode transformation
   !> \note
   !>    ref_count     - reference count of this data structure
   !>    id_nr         - identification number of this data structure
   !>    p             - number of beads
   !>    Q_bead        - thermostat mass for a non-centroid bead
   !>    Q_centroid    - thermostat mass for a centroid degree of freedom
   !>    modefactor    - mass scale factor for non-centroid degrees of freedom
   !>    harm          - factor for harmonic potential ( w_p^2/modefactor )
   !>    x2u           - transformation matrix real coord to normal mode space
   !>    u2x           - transformation matrix normal mode coord to real space
   !>    lambda        - propagator frequencies of the ring polymer
   !>
   !>    This could be done via FFT calls as well, but for now...
   !> \author hforbert
   ! ***************************************************************************
   TYPE normalmode_env_type
      INTEGER :: id_nr, ref_count, p
      REAL(KIND=dp) :: Q_bead, Q_centroid, modefactor, harm
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: x2u, u2x
      REAL(KIND=dp), DIMENSION(:), POINTER :: lambda
   END TYPE normalmode_env_type

   ! ***************************************************************************
   !> \brief data to perform the staging transformation
   !> \note
   !>    ref_count     - reference count of this data structure
   !>    id_nr         - identification number of this data structure
   !>    nseg
   !>    j
   !>    p
   !>    w_p
   !>    w_j
   !>    Q_stage
   !>    Q_end
   !> \author fawzi
   ! ***************************************************************************
   TYPE staging_env_type
      INTEGER :: id_nr, ref_count, nseg, j, p
      REAL(KIND=dp) w_p, w_j, Q_stage, Q_end
   END TYPE staging_env_type

   ! ***************************************************************************
   !> \brief data to use the pile thermostat
   !> \note
   !>    lamb          - coupling constant of pile to the normal modes
   !>    tau           - time constant for centroid mode
   !>    thermostat_energy        - energy difference for conxerved quantity
   !>    c1            - scaling of the old momenta
   !>    c2            - scaling of the friction term
   !>    g_fric        - mode specific friction
   !>    massfact      - Mass prefactor to get units right
   !>    gaussian_rng_stream - random number generator
   !> \author Felix Uhl
   ! ***************************************************************************
   TYPE pile_therm_type
      INTEGER                              :: ref_count
      REAL(KIND=dp)                        :: lamb, tau, thermostat_energy
      REAL(KIND=dp), DIMENSION(:), POINTER :: c1
      REAL(KIND=dp), DIMENSION(:), POINTER :: c2
      REAL(KIND=dp), DIMENSION(:), POINTER :: g_fric
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: massfact
      TYPE(rng_stream_type), POINTER       :: gaussian_rng_stream
   END TYPE pile_therm_type

   ! ***************************************************************************
   !> \brief data to use the piglet thermostat
   !> \note
   !>    ndim          - number of degrees of freedom
   !>    p             - trotter number
   !>    nsp1          - number of additional degrees of freedom for Markovian
   !dynamics + 1
   !>    thermostat_energy        - energy difference for conxerved quantity
   !>    a_mat         - A matrices (9,9,P)
   !>    c_mat         - C matrices (9,9,P)
   !>    gle_t         - Deterministic part of propagator
   !>    gle_s         - Stochastic part of propagator
   !>    smalls        - Keeps a copy of momenta and additional degrees of
   !freedom
   !>                    to ensure Markovian dynamics
   !>    temp1         - Big storage array that is needed on the way
   !>    temp2         - vector to store the random numbers
   !>    sqrtmass      - contains the squareroot of the dynamical masses
   !>    gaussian_rng_stream - random number generator
   !> \author Felix Uhl
   ! ***************************************************************************
   TYPE piglet_therm_type
      INTEGER                              :: ref_count
      INTEGER                              :: ndim, p, nsp1
      REAL(KIND=dp)                        :: thermostat_energy
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: a_mat, c_mat
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: gle_s, gle_t
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: smalls
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: temp1
      REAL(KIND=dp), DIMENSION(:), POINTER   :: temp2
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: sqrtmass
      TYPE(rng_stream_type), POINTER       :: gaussian_rng_stream
   END TYPE piglet_therm_type

   ! ***************************************************************************
   !> \brief data for the use of different Path Integral propagators
   !> \note
   !>    prop_kind     - selects a hamiltonian for the equations of motion
   !>    temp_sim2phys - conversion factor for simulation to physical temperature
   !>    temp_phys2sim - conversion factor for physical to simulation temperature
   !>    physpotscale  - factor to scale the physical interaction potential
   !> \author Felix Uhl
   ! ***************************************************************************
   TYPE pint_propagator_type
      INTEGER                              :: prop_kind
      REAL(KIND=dp)                        :: temp_phys2sim
      REAL(KIND=dp)                        :: temp_sim2phys
      REAL(KIND=dp)                        :: physpotscale
   END TYPE pint_propagator_type

END MODULE pint_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Type for the canonical sampling through velocity rescaling
!> \author Teodoro Laino - 09.2007 University of Zurich [tlaino]
! **************************************************************************************************
MODULE csvr_system_types
   USE bibliography,                    ONLY: Bussi2007,&
                                              cite_reference
   USE extended_system_types,           ONLY: create_map_info_type,&
                                              map_info_type,&
                                              release_map_info_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_rng_seed,&
                                              rng_stream_type
   USE simpar_types,                    ONLY: simpar_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: csvr_system_type, &
             csvr_init, &
             csvr_dealloc, &
             csvr_thermo_create

! **************************************************************************************************
   TYPE csvr_thermo_type
      INTEGER                                 :: degrees_of_freedom
      REAL(KIND=dp)                           :: nkt
      REAL(KIND=dp)                           :: thermostat_energy
      REAL(KIND=dp)                           :: region_kin_energy
      TYPE(rng_stream_type), POINTER          :: gaussian_rng_stream
   END TYPE csvr_thermo_type

! **************************************************************************************************
   TYPE csvr_system_type
      INTEGER                                 :: region, glob_num_csvr, loc_num_csvr
      REAL(KIND=dp)                           :: tau_csvr, dt_fact
      TYPE(csvr_thermo_type), POINTER         :: nvt(:)
      TYPE(map_info_type), POINTER            :: map_info
   END TYPE csvr_system_type

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'csvr_system_types'

CONTAINS

! **************************************************************************************************
!> \brief Initialize type for Canonical Sampling through Velocity Rescaling (CSVR)
!> \param csvr ...
!> \param simpar ...
!> \param section ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_init(csvr, simpar, section)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_init', routineP = moduleN//':'//routineN

      NULLIFY (csvr%nvt)
      NULLIFY (csvr%map_info)
      csvr%loc_num_csvr = 0
      csvr%glob_num_csvr = 0
      csvr%dt_fact = 1.0_dp
      CALL cite_reference(Bussi2007)
      CALL section_vals_val_get(section, "TIMECON", r_val=csvr%tau_csvr)
      ! The CSVR library expects the tau_csv to be in unit of integration timestep
      ! if applied once.. divided by two if the process is applied both to the first
      ! and the second verlet step
      csvr%tau_csvr = csvr%tau_csvr/(0.5_dp*simpar%dt)
      CALL create_map_info_type(csvr%map_info)

   END SUBROUTINE csvr_init

! **************************************************************************************************
!> \brief Initialize NVT type for CSVR thermostat
!> \param csvr ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_thermo_create(csvr)
      TYPE(csvr_system_type), POINTER                    :: csvr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_thermo_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=40)                                  :: name
      INTEGER                                            :: i, ithermo, my_index
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: seed
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed, my_seed

      IF(.NOT.(ASSOCIATED(csvr)))CALL cp__a("csvr_system_types.F",103)
      IF(.NOT.(.NOT. ASSOCIATED(csvr%nvt)))CALL cp__a("csvr_system_types.F",104)

      ALLOCATE (csvr%nvt(csvr%loc_num_csvr))
      DO i = 1, csvr%loc_num_csvr
         csvr%nvt(i)%thermostat_energy = 0.0_dp
         NULLIFY (csvr%nvt(i)%gaussian_rng_stream)
      END DO
      ! Initialize the gaussian stream random number
      ALLOCATE (seed(3, 2, csvr%glob_num_csvr))
      initial_seed = next_rng_seed()

      seed(:, :, 1) = initial_seed
      DO ithermo = 2, csvr%glob_num_csvr
         seed(:, :, ithermo) = next_rng_seed(seed(:, :, ithermo-1))
      END DO
      ! Update initial seed
      initial_seed = next_rng_seed(seed(:, :, csvr%glob_num_csvr))
      DO ithermo = 1, csvr%loc_num_csvr
         my_index = csvr%map_info%index(ithermo)
         my_seed = seed(:, :, my_index)
         WRITE (UNIT=name, FMT="(A,I8)") "Wiener process for Thermostat #", my_index
         CALL compress(name)
         CALL create_rng_stream(rng_stream=csvr%nvt(ithermo)%gaussian_rng_stream, &
                                name=name, distribution_type=GAUSSIAN, extended_precision=.TRUE., &
                                seed=my_seed)
      END DO
      DEALLOCATE (seed)

   END SUBROUTINE csvr_thermo_create

! **************************************************************************************************
!> \brief Deallocate type for CSVR thermostat
!> \param csvr ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_dealloc(csvr)
      TYPE(csvr_system_type), POINTER                    :: csvr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_dealloc', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(csvr)) THEN
         CALL csvr_thermo_dealloc(csvr%nvt)
         CALL release_map_info_type(csvr%map_info)
         DEALLOCATE (csvr)
      ENDIF

   END SUBROUTINE csvr_dealloc

! **************************************************************************************************
!> \brief Deallocate NVT type for CSVR thermostat
!> \param nvt ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_thermo_dealloc(nvt)
      TYPE(csvr_thermo_type), DIMENSION(:), POINTER      :: nvt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_thermo_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(nvt)) THEN
         DO i = 1, SIZE(nvt)
            IF (ASSOCIATED(nvt(i)%gaussian_rng_stream)) THEN
               CALL delete_rng_stream(nvt(i)%gaussian_rng_stream)
            ENDIF
         END DO
         DEALLOCATE (nvt)
      ENDIF
   END SUBROUTINE csvr_thermo_dealloc

END MODULE csvr_system_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Type for the canonical sampling through velocity rescaling
!> \author Teodoro Laino - 09.2007 University of Zurich [tlaino]
! **************************************************************************************************
MODULE al_system_types
   USE bibliography,                    ONLY: Jones2011,&
                                              cite_reference
   USE extended_system_types,           ONLY: create_map_info_type,&
                                              map_info_type,&
                                              release_map_info_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: al_system_type, &
             al_init, &
             al_dealloc, &
             al_thermo_create

! **************************************************************************************************
   TYPE al_thermo_type
      INTEGER                                 :: degrees_of_freedom
      REAL(KIND=dp)                           :: nkt
      REAL(KIND=dp)                           :: chi
      REAL(KIND=dp)                           :: mass
      REAL(KIND=dp)                           :: region_kin_energy
   END TYPE al_thermo_type

! **************************************************************************************************
   TYPE al_system_type
      INTEGER                                 :: region, glob_num_al, loc_num_al
      REAL(KIND=dp)                           :: tau_nh, tau_langevin, dt_fact
      REAL(KIND=dp)                           :: dt
      TYPE(al_thermo_type), POINTER           :: nvt(:)
      TYPE(map_info_type), POINTER            :: map_info
   END TYPE al_system_type

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'al_system_types'

CONTAINS

! **************************************************************************************************
!> \brief Initialize type for Adaptive Langevin (AD_LANGEVIN)
!> \param al ...
!> \param simpar ...
!> \param section ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE al_init(al, simpar, section)
      TYPE(al_system_type), POINTER                      :: al
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'al_init', routineP = moduleN//':'//routineN

      NULLIFY (al%nvt)
      NULLIFY (al%map_info)
      al%loc_num_al = 0
      al%glob_num_al = 0
      al%dt_fact = 1.0_dp
      al%dt = simpar%dt
      CALL cite_reference(Jones2011)
      CALL section_vals_val_get(section, "TIMECON_NH", r_val=al%tau_nh)
      CALL section_vals_val_get(section, "TIMECON_LANGEVIN", r_val=al%tau_langevin)
      CALL create_map_info_type(al%map_info)

   END SUBROUTINE al_init

! **************************************************************************************************
!> \brief Initialize NVT type for AD_LANGEVIN thermostat
!> \param al ...
!> \author Noam Bernstein [noamb]  02.2012
! **************************************************************************************************
   SUBROUTINE al_thermo_create(al)
      TYPE(al_system_type), POINTER                      :: al

      CHARACTER(LEN=*), PARAMETER :: routineN = 'al_thermo_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: seed

      IF(.NOT.(ASSOCIATED(al)))CALL cp__a("al_system_types.F",94)
      IF(.NOT.(.NOT. ASSOCIATED(al%nvt)))CALL cp__a("al_system_types.F",95)

      ALLOCATE (al%nvt(al%loc_num_al))
      DO i = 1, al%loc_num_al
         al%nvt(i)%chi = 0.0_dp
      END DO
      ! Initialize the gaussian stream random number
      ALLOCATE (seed(3, 2, al%glob_num_al))

   END SUBROUTINE al_thermo_create

! **************************************************************************************************
!> \brief Deallocate type for AD_LANGEVIN thermostat
!> \param al ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE al_dealloc(al)
      TYPE(al_system_type), POINTER                      :: al

      CHARACTER(LEN=*), PARAMETER :: routineN = 'al_dealloc', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(al)) THEN
         CALL al_thermo_dealloc(al%nvt)
         CALL release_map_info_type(al%map_info)
         DEALLOCATE (al)
      ENDIF

   END SUBROUTINE al_dealloc

! **************************************************************************************************
!> \brief Deallocate NVT type for AD_LANGEVIN thermostat
!> \param nvt ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE al_thermo_dealloc(nvt)
      TYPE(al_thermo_type), DIMENSION(:), POINTER        :: nvt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'al_thermo_dealloc', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(nvt)) THEN
         DEALLOCATE (nvt)
      ENDIF
   END SUBROUTINE al_thermo_dealloc

END MODULE al_system_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Data types representing superfluid helium
!> \author hforbert
!> \date   2009-01-01
!> \par    History
!>         extracted helium_solvent_type from pint_types.F [lwalewski]
! **************************************************************************************************
MODULE helium_types

   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_logger_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE parallel_rng_types,              ONLY: rng_stream_type
   USE splines_types,                   ONLY: spline_data_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_types'

   REAL(KIND=dp), PARAMETER, PUBLIC :: he_mass = 4.00263037059764_dp !< 4He mass in [u]

   !> Energy contributions - symbolic names for indexing energy arrays
   INTEGER, PARAMETER, PUBLIC :: &
      e_id_total = 1, &
      e_id_potential = 2, &
      e_id_kinetic = 3, &
      e_id_interact = 4, &
      e_id_thermo = 5, &
      e_id_virial = 6

   !> Number of energy contributions for static array allocation
   INTEGER, PARAMETER, PUBLIC :: e_num_ids = 10

   !> number of density function identifiers
   INTEGER, PARAMETER, PUBLIC :: rho_num = 5

   !> density function identifier names
   INTEGER, PARAMETER, PUBLIC :: &
      rho_atom_number = 1, &
      rho_projected_area = 2, &
      rho_winding_number = 3, &
      rho_winding_cycle = 4, &
      rho_moment_of_inertia = 5

   !> number of atom type identifiers
   INTEGER, PARAMETER, PUBLIC :: hid_num = 4

   !> atom type identifier names (for helium-solute interactions)
   INTEGER, PARAMETER, PUBLIC :: &
      hid_chlorine = 1, &
      hid_oxygen = 2, &
      hid_hydrogen = 3, &
      hid_carbon = 4

   !> derived data types
   PUBLIC :: helium_solvent_type
   PUBLIC :: int_arr_ptr
   PUBLIC :: real_arr_ptr

   !> functions
   PUBLIC :: helium_destroy_int_arr_ptr
   PUBLIC :: helium_destroy_real_arr_ptr

! ***************************************************************************
!> \brief  Vector type useful for averaging
!> \author Lukasz Walewski
!> \date   2014-09-09
! ***************************************************************************
   TYPE helium_vector_type

      !> instantaneous value
      REAL(KIND=dp), DIMENSION(3)            :: inst

      !> accumulated value
      REAL(KIND=dp), DIMENSION(3)            :: accu

      !> running average
      REAL(KIND=dp), DIMENSION(3)            :: ravr

      !> restarted value
      REAL(KIND=dp), DIMENSION(3)            :: rstr

   END TYPE helium_vector_type

! ***************************************************************************
!> \brief data structure for solvent helium
!> \author hforbert
! ***************************************************************************
   TYPE helium_solvent_type

      INTEGER                           :: id_nr !< identification number of this data structure
      INTEGER                           :: ref_count !< reference count of this data structure
      TYPE(section_vals_type), POINTER  :: input !< input data structure (the whole tree)
      TYPE(cp_logger_type), POINTER     :: logger

      INTEGER       :: num_env !< number of He environments in runtime
      INTEGER       :: num_env_restart !< number of He environments in restart file

      INTEGER :: atoms !< number of atoms
      INTEGER :: beads !< number of beads per atom (needs to be an integer multiple of the solute's number of beads)
      INTEGER :: bead_ratio !< ratio of helium beads to system beads
      REAL(KIND=dp) :: density !< helium density for free bulk in box

      ! some useful constants
      !
      REAL(KIND=dp) :: hb2m !< hbar squared over m for 4He in CP2k units
      REAL(KIND=dp) :: tau !< 1/(k_B T p) with T - He temperature, p - number of beads
      REAL(KIND=dp) :: wpref !< prefactor for calculating superfluid fraction from <(M*W)^2>
      REAL(KIND=dp) :: apref !< prefactor for calculating superfluid fraction from <A^2/I_c>

      ! PBC related
      !
      LOGICAL                        :: periodic !< true if bulk liquid helium in periodic box
      INTEGER                        :: cell_shape !< unit cell shape for PBC calculations
      REAL(KIND=dp)                  :: cell_size !< size of the periodic box (helium only)
      REAL(KIND=dp)                  :: cell_size_inv !< 1/cell_size (inverse)
      REAL(KIND=dp), DIMENSION(3, 3)  :: cell_m !< the unit cell vectors' matrix
      REAL(KIND=dp), DIMENSION(3, 3)  :: cell_m_inv !< invrse  of the unit cell vectors' matrix
      REAL(KIND=dp), DIMENSION(3)    :: origin !< origin of the cell (first voxel position)
      REAL(KIND=dp)                  :: droplet_radius !< radius of the droplet

      REAL(KIND=dp), DIMENSION(3)    :: center !< COM of solute (if present) or center of periodic cell (if periodic) or COM of helium

      ! sampling parameters
      INTEGER :: sampling_method

      INTEGER :: iter_norot !< number of iterations to try for a given imaginary time slice rotation (num inner MC loop iters)
      INTEGER :: iter_rot !< number of rotations to try (total number of iterations is iter_norot*iter_rot) (num outer MC loop iters)
      !
      INTEGER       :: maxcycle !< maximum cyclic permutation change to attempt
      INTEGER       :: m_dist_type !< distribution from which the cycle length m is sampled
      INTEGER       :: m_value !< cycle length sampled with different probability than other lengths
      REAL(KIND=dp) :: m_ratio !< probability ratio betw m_value and other possible values of m
      !
      INTEGER :: relrot !< relative rotation in imaginary time wrt normal system/starting configuration
      INTEGER :: bisection !< power of 2 number for bisection algorithm
      INTEGER :: bisctlog2 !< log2(bisection)

      REAL(KIND=dp) :: e_corr !< potential correction energy due to finite box
      INTEGER :: pdx !< pair density expansion max exponent

      ! MC step counters
      !
      INTEGER :: num_steps !< number of iterations in the current run
      INTEGER :: first_step !< first step, restarted from MOTION%PINT%ITERATION (default value: 0)
      INTEGER :: last_step
      INTEGER :: current_step !< first_step + number of steps performed so far

      ! helium variables
      !
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: pos !< position of the helium atoms DIM(3,atoms,beads)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: work !< same dimensions as pos
      !
      INTEGER, DIMENSION(:), POINTER :: permutation !< current permutation state DIM(atoms)
      INTEGER, DIMENSION(:), POINTER :: iperm !< inverse of the current permutation state DIM(atoms)
      INTEGER, DIMENSION(:), POINTER :: ptable !< proposed cyclic permutation, DIM(max_cycle)
      INTEGER(KIND=int_8)              :: accepts !< number of accepted new configurations
      !
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: tmatrix !< ? permutation probability related
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: pmatrix !< ? permutation probability related [use might change/new ones added/etc]
      REAL(KIND=dp) :: pweight !< ? permutation probability related
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: ipmatrix
      INTEGER, DIMENSION(:, :), POINTER :: nmatrix

      TYPE(spline_data_p_type), DIMENSION(:, :), POINTER :: uij !< pair density matrix coefficients (action)
      TYPE(spline_data_p_type), DIMENSION(:, :), POINTER :: eij !< pair density matrix coefficients (energy)

      ! calculated properties
      !
      REAL(KIND=dp), DIMENSION(e_num_ids)    :: energy_inst !< energy contributions (instantaneous)
      REAL(KIND=dp), DIMENSION(e_num_ids)    :: energy_avrg !< energy contributions (averaged)
      TYPE(helium_vector_type)               :: wnumber !< winding number
      TYPE(helium_vector_type)               :: wnmber2 !< winding number squared
      TYPE(helium_vector_type)               :: proarea !< projected area
      TYPE(helium_vector_type)               :: prarea2 !< projected area squared
      TYPE(helium_vector_type)               :: mominer !< moment of inertia
      INTEGER                                :: averages_iweight !< weight for restarted averages
      LOGICAL                                :: averages_restarted !< flag indicating whether the averages have been restarted

      !
      INTEGER                                :: rdf_nbin !< number of bins for RDF
      INTEGER                                :: rdf_iweight !< weight for restarted RDF
      INTEGER                                :: rho_iweight !< weight for restarted RHO
      INTEGER                                :: rdf_num !< number of RDF estimators
      INTEGER                                :: rdf_num_ctr !< number of centers for RDF calc
      REAL(KIND=dp)                          :: rdf_delr !< delta r for RDF
      REAL(KIND=dp)                          :: rdf_maxr !< maximum r for RDF
      REAL(KIND=dp), DIMENSION(:), POINTER :: rdf_centers !< positions of RDF centers
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rdf_inst !< RDF (instantaneous/tmp array)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rdf_rstr !< RDF (restarted)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rdf_accu !< RDF (accumulated for one run)
      LOGICAL :: rdf_present
      !
      INTEGER :: rho_nbin
      INTEGER :: rho_num_act !< actual number of density estimators
      INTEGER :: rho_num_min_len_wdg !< number of optional estimators based on winding cycles
      INTEGER :: rho_num_min_len_non !< number of optional estimators based on non-winding cycles
      INTEGER :: rho_num_min_len_all !< number of optional estimators based on all cycles
      INTEGER, DIMENSION(:), POINTER :: rho_min_len_wdg_vals !< minimum lengths of winding cycles
      INTEGER, DIMENSION(:), POINTER :: rho_min_len_non_vals !< minimum lengths of non-winding cycles
      INTEGER, DIMENSION(:), POINTER :: rho_min_len_all_vals !< minimum lengths of all cycles
      REAL(KIND=dp) :: rho_delr, rho_maxr
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER :: rho_inst
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER :: rho_rstr
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER :: rho_accu
      LOGICAL :: rho_present
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER ::  rho_incr !< increment for density bining

      TYPE(density_properties_type), DIMENSION(:), POINTER  :: rho_property

      REAL(KIND=dp), DIMENSION(:, :), POINTER :: num_accepted !< average number of accepted permutations of a given length
      !! on a given Levy level, plus one additional level which
      !! counts # of trials, REAL(BISCTLOG2+2, MAX_PERM_CYCLE)
      !! num_accepted(1,l) - # of trials for perm length l
      !! num_accepted(2,l) - # of selected perms of length l
      !! num_accepted(3,l) - # of perms of length l accepted at level 1
      !! average over He environments/processors
      REAL(KIND=dp), DIMENSION(:), POINTER :: plength_avrg !< permutation length probability distribution DIM(atoms)
      REAL(KIND=dp), DIMENSION(:), POINTER :: plength_inst !< instantaneous permutation length probability DIM(atoms)
      INTEGER, DIMENSION(:), POINTER :: atom_plength !< length of the permutation cycle the atom belongs to DIM(atoms)

      TYPE(rng_stream_type), POINTER :: rng_stream_uniform !< random number stream with uniform distribution
      TYPE(rng_stream_type), POINTER :: rng_stream_gaussian !< random number stream with gaussian distribution

      ! variables related to solvated molecular system
      !
      LOGICAL :: solute_present !< switch the interactions with the solute on or off
      INTEGER :: solute_atoms !< number of solute atoms (pint_env%ndim/3)
      INTEGER :: solute_beads !< number of solute beads (pint_env%p)
      INTEGER :: get_helium_forces !< parameter to determine whether the average or last MC force should be taken to MD
      CHARACTER(LEN=2), DIMENSION(:), POINTER :: solute_element !< element names of solute atoms (pint_env%ndim/3)
      TYPE(int_arr_ptr), DIMENSION(:), POINTER :: solute_i !< indices of solute atoms sorted by atomic kinds
      TYPE(cell_type), POINTER  :: solute_cell !< dimensions of the solvated system cell (a,b,c) (should be removed at some point)
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: force_avrg !< averaged forces exerted by He solvent on the solute DIM(p,ndim)
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: force_inst !< instantaneous forces exerted by He on the solute (p,ndim)
      CHARACTER(LEN=2), DIMENSION(:), POINTER  :: ename
      INTEGER, DIMENSION(:), POINTER           :: eid
      INTEGER :: enum
      INTEGER :: solute_interaction

      LOGICAL :: interaction_pot_scan !< wheather to perform solute-helium interaction scan

      ! temporary arrays for optimization
      !
      INTEGER, DIMENSION(:), POINTER         :: itmp_atoms_1d !< DIM(atoms) - same as permutation
      INTEGER, DIMENSION(:), POINTER         :: itmp_atoms_np_1d !< DIM(atoms*num_env)
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rtmp_3_np_1d !< DIM(3*num_env)
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rtmp_p_ndim_1d !< DIM(p*ndim)
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rtmp_p_ndim_np_1d !< DIM(p*ndim*num_env)
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rtmp_3_atoms_beads_1d !< DIM(3*atoms*beads)
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rtmp_3_atoms_beads_np_1d
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: rtmp_p_ndim_2d !< DIM(p,ndim)
      LOGICAL, DIMENSION(:, :, :), POINTER     :: ltmp_3_atoms_beads_3d !< DIM(3,atoms,beads) - same as pos
      LOGICAL, DIMENSION(:), POINTER         :: ltmp_atoms_1d !< DIM(atoms) - for unpacking the permutation

   END TYPE helium_solvent_type

! ***************************************************************************
!> \brief  Container type for properties of a helium density function
!> \author Lukasz Walewski
!> \date   2014-09-09
! ***************************************************************************
   TYPE density_properties_type

      !> name of this density function
      CHARACTER(len=default_string_length) :: name

      !> flag indicating whether this function should be calculated
      LOGICAL :: is_calculated

      !> number of components that this function is composed of
      INTEGER :: num_components

      !> suffixes for the filenames storing components of this function
      CHARACTER(len=default_string_length), DIMENSION(:), POINTER :: filename_suffix

      !> component names
      CHARACTER(len=default_string_length), DIMENSION(:), POINTER :: component_name

      !> indices locating the components of this function in the global density arrays
      INTEGER, DIMENSION(:), POINTER :: component_index

   END TYPE density_properties_type

! ***************************************************************************
!> \brief  A pointer to an integer array, data type to be used in arrays of
!>         pointers.
!> \author Lukasz Walewski
!> \date   2013-12-11
! ***************************************************************************
   TYPE int_arr_ptr
      INTEGER, DIMENSION(:), POINTER :: iap
   END TYPE int_arr_ptr

! ***************************************************************************
!> \brief  A pointer to a real array, data type to be used in arrays of
!>         pointers.
!> \author Lukasz Walewski
!> \date   2013-12-11
! ***************************************************************************
   TYPE real_arr_ptr
      REAL(KIND=dp), DIMENSION(:), POINTER :: rap
   END TYPE real_arr_ptr

CONTAINS

! ***************************************************************************
!> \brief  Deallocate all arrays pointed to by the pointers stored in the
!>         integer pointer array
!> \param int_arr_p ...
!> \date   2013-12-12
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_destroy_int_arr_ptr(int_arr_p)

      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: int_arr_p

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_destroy_int_arr_ptr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip

! deallocate memory used by each component of the pointer array

      DO ip = 1, SIZE(int_arr_p)
         IF (ASSOCIATED(int_arr_p(ip)%iap)) THEN
            DEALLOCATE (int_arr_p(ip)%iap)
         END IF
      END DO

      ! deallocate the memory used for pointer array
      IF (ASSOCIATED(int_arr_p)) THEN
         DEALLOCATE (int_arr_p)
      END IF

      RETURN
   END SUBROUTINE helium_destroy_int_arr_ptr

! ***************************************************************************
!> \brief  Deallocate all arrays pointed to by the pointers stored in the
!>         real pointer array
!> \param real_arr_p ...
!> \date   2013-12-12
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_destroy_real_arr_ptr(real_arr_p)

      TYPE(real_arr_ptr), DIMENSION(:), POINTER          :: real_arr_p

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_destroy_real_arr_ptr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip

! do not attempt deallocation on null pointer

      IF (.NOT. ASSOCIATED(real_arr_p)) THEN
         RETURN
      END IF

      ! deallocate memory used by each component of the pointer array
      DO ip = 1, SIZE(real_arr_p)
         IF (ASSOCIATED(real_arr_p(ip)%rap)) THEN
            DEALLOCATE (real_arr_p(ip)%rap)
         END IF
      END DO

      ! deallocate the memory used for pointer array itself
      IF (ASSOCIATED(real_arr_p)) THEN
         DEALLOCATE (real_arr_p)
      END IF

      RETURN
   END SUBROUTINE helium_destroy_real_arr_ptr

END MODULE helium_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Thermal regions type: to initialize and control the temperature of
!>        different regions
!> \par History
!>   - Added support for langevin regions (2014/01/08, LT)
!> \author MI
! **************************************************************************************************
MODULE thermal_region_types

   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: thermal_regions_type, &
             thermal_region_type, &
             allocate_thermal_regions, &
             release_thermal_regions, &
             retain_thermal_regions

   TYPE thermal_regions_type
      INTEGER :: id_nr, ref_count, nregions
      LOGICAL :: force_rescaling
      REAL(KIND=dp) :: temp_reg0
      LOGICAL, DIMENSION(:), POINTER                   :: do_langevin
      TYPE(section_vals_type), POINTER                 :: section
      TYPE(thermal_region_type), DIMENSION(:), POINTER :: thermal_region
   END TYPE thermal_regions_type

   TYPE thermal_region_type
      INTEGER :: region_index, npart
      INTEGER, DIMENSION(:), POINTER :: part_index
      REAL(KIND=dp) :: ekin, temperature, temp_expected, temp_tol
   END TYPE thermal_region_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermal_region_types'
CONTAINS

! **************************************************************************************************
!> \brief allocate thermal_regions
!> \param thermal_regions ...
!> \author
! **************************************************************************************************
   SUBROUTINE allocate_thermal_regions(thermal_regions)
      TYPE(thermal_regions_type), POINTER                :: thermal_regions

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_thermal_regions', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = .NOT. ASSOCIATED(thermal_regions)
      IF(.NOT.(check))CALL cp__a("motion/thermal_region_types.F",60)

      ALLOCATE (thermal_regions)
      thermal_regions%ref_count = 1
      thermal_regions%nregions = 0
      NULLIFY (thermal_regions%thermal_region)
      NULLIFY (thermal_regions%do_langevin)

   END SUBROUTINE allocate_thermal_regions
! **************************************************************************************************
!> \brief retains  thermal_regions
!> \param thermal_regions ...
!> \author
! **************************************************************************************************
   SUBROUTINE retain_thermal_regions(thermal_regions)

      TYPE(thermal_regions_type), POINTER                :: thermal_regions

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_thermal_regions', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermal_regions)) THEN
         IF(.NOT.(thermal_regions%ref_count > 0))CALL cp__a("motion/thermal_region_types.F",82)
         thermal_regions%ref_count = thermal_regions%ref_count+1
      END IF

   END SUBROUTINE retain_thermal_regions

! **************************************************************************************************
!> \brief release thermal_regions
!> \param thermal_regions ...
!> \author
! **************************************************************************************************
   SUBROUTINE release_thermal_regions(thermal_regions)

      TYPE(thermal_regions_type), POINTER                :: thermal_regions

      CHARACTER(len=*), PARAMETER :: routineN = 'release_thermal_regions', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ireg
      LOGICAL                                            :: check

      check = ASSOCIATED(thermal_regions)
      IF (check) THEN
         check = thermal_regions%ref_count > 0
         IF(.NOT.(check))CALL cp__a("motion/thermal_region_types.F",106)
         thermal_regions%ref_count = thermal_regions%ref_count-1
         IF (thermal_regions%ref_count < 1) THEN
            IF (ASSOCIATED(thermal_regions%thermal_region)) THEN
               DO ireg = 1, SIZE(thermal_regions%thermal_region)
                  DEALLOCATE (thermal_regions%thermal_region(ireg)%part_index)
               END DO
               DEALLOCATE (thermal_regions%thermal_region)
            END IF
            IF (ASSOCIATED(thermal_regions%do_langevin)) THEN
               DEALLOCATE (thermal_regions%do_langevin)
            END IF
            DEALLOCATE (thermal_regions)
         END IF
      END IF

   END SUBROUTINE release_thermal_regions

END MODULE thermal_region_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief initialization of the reftraj structure used to analyse
!>     previously generated trajectories
!> \par History
!>      Created 10-07 [MI]
!> \author MI
! **************************************************************************************************
MODULE reftraj_types

   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: reftraj_type, reftraj_msd_type, &
             create_reftraj, release_reftraj, retain_reftraj

! **************************************************************************************************
!> \brief parameters related to the analysis of previously generated trajecorties
!> \author MI
! **************************************************************************************************
   TYPE reftraj_info_type
      INTEGER                                  :: first_snapshot
      INTEGER                                  :: last_snapshot
      INTEGER                                  :: stride
      LOGICAL                                  :: eval_ef
      LOGICAL                                  :: variable_volume
      LOGICAL                                  :: msd
      TYPE(cp_parser_type), POINTER            :: traj_parser
      TYPE(cp_parser_type), POINTER            :: cell_parser
   END TYPE reftraj_info_type

! **************************************************************************************************
   TYPE reftraj_msd_type
      LOGICAL                                  :: disp_atom, msd_kind, msd_molecule, msd_region
      INTEGER                                  :: num_disp_atom, ref0_unit
      INTEGER, POINTER, DIMENSION(:)           :: disp_atom_index
      REAL(KIND=dp)                            :: disp_atom_tol, drcom(3), ref0_com(3), total_mass
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: disp_atom_dr
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: ref0_pos
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: ref0_com_molecule
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: val_msd_kind
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: val_msd_molecule
      REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: val_msd_region
   END TYPE reftraj_msd_type

! **************************************************************************************************
   TYPE reftraj_type
      INTEGER                                  :: ref_count
      INTEGER                                  :: itimes
      INTEGER                                  :: itimes0
      INTEGER                                  :: isnap
      INTEGER                                  :: natom
      LOGICAL                                  :: init
      REAL(KIND=dp)                            :: epot, epot0, time, time0
      TYPE(reftraj_info_type), POINTER         :: info
      TYPE(reftraj_msd_type), POINTER          :: msd
   END TYPE reftraj_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'reftraj_types'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param reftraj ...
!> \param reftraj_section ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE create_reftraj(reftraj, reftraj_section, para_env)

      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(section_vals_type), POINTER                   :: reftraj_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_reftraj', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename

      IF(.NOT.(.NOT. ASSOCIATED(reftraj)))CALL cp__a("motion/reftraj_types.F",93)
      ALLOCATE (reftraj)
      reftraj%ref_count = 1

      NULLIFY (reftraj%info)
      NULLIFY (reftraj%msd)

      ALLOCATE (reftraj%info)
      NULLIFY (reftraj%info%traj_parser)
      NULLIFY (reftraj%info%cell_parser)

      ! Initialize parser for trajectory
      CALL section_vals_val_get(reftraj_section, "TRAJ_FILE_NAME", c_val=filename)
      CALL parser_create(reftraj%info%traj_parser, filename, para_env=para_env)

      CALL section_vals_val_get(reftraj_section, "VARIABLE_VOLUME", l_val=reftraj%info%variable_volume)
      IF (reftraj%info%variable_volume) THEN
         ! In case requested initialize parser for cell
         CALL section_vals_val_get(reftraj_section, "CELL_FILE_NAME", c_val=filename)
         CALL parser_create(reftraj%info%cell_parser, filename, para_env=para_env)
      END IF

      CALL section_vals_val_get(reftraj_section, "FIRST_SNAPSHOT", i_val=reftraj%info%first_snapshot)
      CALL section_vals_val_get(reftraj_section, "LAST_SNAPSHOT", i_val=reftraj%info%last_snapshot)
      CALL section_vals_val_get(reftraj_section, "STRIDE", i_val=reftraj%info%stride)
      CALL section_vals_val_get(reftraj_section, "EVAL_ENERGY_FORCES", l_val=reftraj%info%eval_ef)

      CALL section_vals_val_get(reftraj_section, "MSD%_SECTION_PARAMETERS_", &
                                l_val=reftraj%info%msd)

   END SUBROUTINE create_reftraj

! **************************************************************************************************
!> \brief ...
!> \param reftraj ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE retain_reftraj(reftraj)

      TYPE(reftraj_type), POINTER                        :: reftraj

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_reftraj', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(reftraj)) THEN
         IF(.NOT.(reftraj%ref_count > 0))CALL cp__a("motion/reftraj_types.F",139)
         reftraj%ref_count = reftraj%ref_count+1
      END IF

   END SUBROUTINE retain_reftraj

! **************************************************************************************************
!> \brief ...
!> \param reftraj ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE release_reftraj(reftraj)

      TYPE(reftraj_type), POINTER                        :: reftraj

      CHARACTER(len=*), PARAMETER :: routineN = 'release_reftraj', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(reftraj)) THEN
         IF(.NOT.(reftraj%ref_count > 0))CALL cp__a("motion/reftraj_types.F",160)
         reftraj%ref_count = reftraj%ref_count-1
         IF (reftraj%ref_count < 1) THEN
            CALL parser_release(reftraj%info%traj_parser)
            CALL parser_release(reftraj%info%cell_parser)
            IF (ASSOCIATED(reftraj%info)) THEN
               DEALLOCATE (reftraj%info)
            END IF
            IF (ASSOCIATED(reftraj%msd)) THEN
               DEALLOCATE (reftraj%msd%ref0_pos)
               IF (reftraj%msd%msd_kind) THEN
                  DEALLOCATE (reftraj%msd%val_msd_kind)
               END IF
               IF (reftraj%msd%msd_molecule) THEN
                  DEALLOCATE (reftraj%msd%val_msd_molecule)
                  DEALLOCATE (reftraj%msd%ref0_com_molecule)
               END IF
               IF (reftraj%msd%disp_atom) THEN
                  DEALLOCATE (reftraj%msd%disp_atom_index)
                  DEALLOCATE (reftraj%msd%disp_atom_dr)
               END IF

               DEALLOCATE (reftraj%msd)
            END IF

            DEALLOCATE (reftraj)
         END IF

      END IF
   END SUBROUTINE release_reftraj

END MODULE reftraj_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief defines types for metadynamics calculation
!> \par History
!>      01.2007 created [tlaino] Teodoro Laino
! **************************************************************************************************
MODULE free_energy_types

   USE input_constants,                 ONLY: do_fe_ac,&
                                              do_fe_ui
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'free_energy_types'

   PUBLIC :: free_energy_type, &
             fe_env_release, &
             fe_env_create, &
             ui_var_type

! **************************************************************************************************
   TYPE ui_var_type
      REAL(KIND=dp), DIMENSION(:), POINTER           :: ss
      INTEGER                                        :: icolvar
   END TYPE ui_var_type

! **************************************************************************************************
   TYPE ui_conv_type
      ! Specifying convergence parameters
      INTEGER                                        :: cg_width, max_cg_width
      INTEGER                                        :: cg_points
      REAL(KIND=dp)                                  :: eps_conv
      REAL(KIND=dp)                                  :: k_conf_lm
      REAL(KIND=dp)                                  :: sw_conf_lm
      REAL(KIND=dp)                                  :: vn_conf_lm
      LOGICAL                                        :: test_k, &
                                                        test_sw, &
                                                        test_vn
   END TYPE ui_conv_type

! **************************************************************************************************
   TYPE statistical_type
      ! Collecting coarse grained data
      REAL(KIND=dp), DIMENSION(:), POINTER           :: avg
      REAL(KIND=dp), DIMENSION(:, :), POINTER         :: var
   END TYPE statistical_type

! **************************************************************************************************
   TYPE free_energy_type
      INTEGER                                        :: ncolvar
      INTEGER                                        :: TYPE
      INTEGER                                        :: nr_points, &
                                                        nr_rejected
      TYPE(ui_conv_type), POINTER                    :: conv_par
      TYPE(ui_var_type), POINTER, DIMENSION(:)       :: uivar
      TYPE(statistical_type), DIMENSION(:), POINTER  :: cg_data
      ! Old data
      REAL(KIND=dp)                                  :: eps_conv
      REAL(KIND=dp), DIMENSION(:, :), POINTER         :: covmx

      CHARACTER(len=default_string_length)           :: plumed_input_file
   END TYPE free_energy_type

CONTAINS

! **************************************************************************************************
!> \brief creates the fe_env
!> \param fe_env ...
!> \param fe_section ...
!> \author Teodoro Laino 01.2007
! **************************************************************************************************
   SUBROUTINE fe_env_create(fe_env, fe_section)
      TYPE(free_energy_type), POINTER                    :: fe_env
      TYPE(section_vals_type), POINTER                   :: fe_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fe_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, id_method
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: ui_section, ui_var_section

      IF(.NOT.(.NOT. ASSOCIATED(fe_env)))CALL cp__a("free_energy_types.F",95)

      CALL section_vals_get(fe_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(fe_section, "METHOD", i_val=id_method)
         SELECT CASE (id_method)
         CASE (do_fe_ui)
            ALLOCATE (fe_env)
            NULLIFY (fe_env%covmx, fe_env%uivar, fe_env%conv_par, fe_env%cg_data)
            fe_env%type = id_method
            fe_env%nr_points = 0
            fe_env%nr_rejected = 0
            NULLIFY (fe_env%cg_data)
            ui_section => section_vals_get_subs_vals(fe_section, "UMBRELLA_INTEGRATION")
            ui_var_section => section_vals_get_subs_vals(ui_section, "UVAR")
            CALL section_vals_get(ui_var_section, n_repetition=fe_env%ncolvar)
            ! Convergence controlling parameters
            ALLOCATE (fe_env%conv_par)
            fe_env%conv_par%test_k = .FALSE.
            fe_env%conv_par%test_sw = .FALSE.
            fe_env%conv_par%test_vn = .FALSE.
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%COARSE_GRAINED_WIDTH", &
                                      i_val=fe_env%conv_par%cg_width)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%MAX_COARSE_GRAINED_WIDTH", &
                                      i_val=fe_env%conv_par%max_cg_width)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%COARSE_GRAINED_POINTS", &
                                      i_val=fe_env%conv_par%cg_points)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%EPS_CONV", &
                                      r_val=fe_env%conv_par%eps_conv)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%K_CONFIDENCE_LIMIT", &
                                      r_val=fe_env%conv_par%k_conf_lm)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%SW_CONFIDENCE_LIMIT", &
                                      r_val=fe_env%conv_par%sw_conf_lm)
            CALL section_vals_val_get(ui_section, "CONVERGENCE_CONTROL%VN_CONFIDENCE_LIMIT", &
                                      r_val=fe_env%conv_par%vn_conf_lm)
            ! Umbrella Integration variables
            ALLOCATE (fe_env%uivar(fe_env%ncolvar))
            DO i = 1, fe_env%ncolvar
               ! Read Umbrella Integration Variable definition
               CALL section_vals_val_get(ui_var_section, "COLVAR", &
                                         i_val=fe_env%uivar(i)%icolvar, i_rep_section=i)
               NULLIFY (fe_env%uivar(i)%ss)
            END DO
         CASE (do_fe_ac)
            ALLOCATE (fe_env)
            NULLIFY (fe_env%covmx, fe_env%uivar, fe_env%conv_par, fe_env%cg_data)
            ALLOCATE (fe_env%covmx(3, 0))
            fe_env%type = id_method
            CALL section_vals_val_get(fe_section, "ALCHEMICAL_CHANGE%EPS_CONV", r_val=fe_env%eps_conv)
         CASE DEFAULT
            ! Do Nothing
         END SELECT
      END IF
   END SUBROUTINE fe_env_create

! **************************************************************************************************
!> \brief releases the fe_env
!> \param fe_env ...
!> \author Laino Teodoro 01.2007
! **************************************************************************************************
   SUBROUTINE fe_env_release(fe_env)
      TYPE(free_energy_type), POINTER                    :: fe_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fe_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(fe_env)) THEN
         IF (ASSOCIATED(fe_env%covmx)) THEN
            DEALLOCATE (fe_env%covmx)
         END IF
         IF (ASSOCIATED(fe_env%cg_data)) THEN
            DO i = 1, SIZE(fe_env%cg_data)
               IF (ASSOCIATED(fe_env%cg_data(i)%avg)) THEN
                  DEALLOCATE (fe_env%cg_data(i)%avg)
               END IF
               IF (ASSOCIATED(fe_env%cg_data(i)%var)) THEN
                  DEALLOCATE (fe_env%cg_data(i)%var)
               END IF
            END DO
            DEALLOCATE (fe_env%cg_data)
         END IF
         IF (ASSOCIATED(fe_env%conv_par)) THEN
            DEALLOCATE (fe_env%conv_par)
         END IF
         IF (ASSOCIATED(fe_env%uivar)) THEN
            DO i = 1, SIZE(fe_env%uivar)
               IF (ASSOCIATED(fe_env%uivar(i)%ss)) THEN
                  DEALLOCATE (fe_env%uivar(i)%ss)
               END IF
            END DO
            DEALLOCATE (fe_env%uivar)
         END IF
         DEALLOCATE (fe_env)
      END IF
   END SUBROUTINE fe_env_release

END MODULE free_energy_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Thermostat structure: module containing thermostat available for MD
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
MODULE thermostat_types
   USE al_system_types,                 ONLY: al_dealloc,&
                                              al_init,&
                                              al_system_type
   USE csvr_system_types,               ONLY: csvr_dealloc,&
                                              csvr_init,&
                                              csvr_system_type
   USE extended_system_types,           ONLY: lnhc_dealloc,&
                                              lnhc_init,&
                                              lnhc_parameters_type
   USE gle_system_types,                ONLY: gle_dealloc,&
                                              gle_init,&
                                              gle_type
   USE input_constants,                 ONLY: do_region_global,&
                                              do_region_massive,&
                                              do_thermo_al,&
                                              do_thermo_csvr,&
                                              do_thermo_gle,&
                                              do_thermo_nose
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: thermostats_type, &
             thermostat_type, &
             allocate_thermostats, &
             release_thermostats, &
             retain_thermostats, &
             create_thermostat_type, &
             release_thermostat_type, &
             thermostat_info_type, &
             release_thermostat_info, &
             set_thermostats

! **************************************************************************************************
!> \brief Define thermostat types
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      10.2007 created [tlaino] - Teodoro Laino - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE thermostats_type
      INTEGER :: id_nr, ref_count
      TYPE(thermostat_info_type), POINTER :: thermostat_info_part
      TYPE(thermostat_info_type), POINTER :: thermostat_info_shell
! cjm
      TYPE(thermostat_info_type), POINTER :: thermostat_info_fast
      TYPE(thermostat_type), POINTER       :: thermostat_fast
      TYPE(thermostat_info_type), POINTER :: thermostat_info_slow
      TYPE(thermostat_type), POINTER       :: thermostat_slow
! cjm
      TYPE(thermostat_type), POINTER       :: thermostat_part
      TYPE(thermostat_type), POINTER       :: thermostat_coef
      TYPE(thermostat_type), POINTER       :: thermostat_shell
      TYPE(thermostat_type), POINTER       :: thermostat_baro
   END TYPE thermostats_type

   ! Single thermostat_type
! **************************************************************************************************
   TYPE thermostat_type
      INTEGER :: id_nr, ref_count, type_of_thermostat
      CHARACTER(LEN=default_string_length) :: label
      TYPE(lnhc_parameters_type), POINTER :: nhc
      TYPE(csvr_system_type), POINTER     :: csvr
      TYPE(al_system_type), POINTER     :: al
      TYPE(gle_type), POINTER      ::  gle
      TYPE(section_vals_type), POINTER    :: section
   END TYPE thermostat_type

   ! Global info type
! **************************************************************************************************
   TYPE thermostat_info_type
      INTEGER                              :: sum_of_thermostats
      INTEGER                              :: number_of_thermostats
      INTEGER                              :: dis_type
      INTEGER, POINTER, DIMENSION(:)       :: map_loc_thermo_gen
   END TYPE thermostat_info_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermostat_types'
   INTEGER, PRIVATE, SAVE :: last_thermostat_id_nr = 0, &
                             last_thermostats_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param thermostats ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE allocate_thermostats(thermostats)
      TYPE(thermostats_type), POINTER                    :: thermostats

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_thermostats', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = .NOT. ASSOCIATED(thermostats)
      IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_types.F",118)

      ! Preliminary allocation for thermostats
      ALLOCATE (thermostats)
      last_thermostats_id_nr = last_thermostats_id_nr+1
      thermostats%id_nr = last_thermostats_id_nr
      thermostats%ref_count = 1

      ! Thermostats Info
      ALLOCATE (thermostats%thermostat_info_part)
      ALLOCATE (thermostats%thermostat_info_shell)
!cjm
      ALLOCATE (thermostats%thermostat_info_fast)
      ALLOCATE (thermostats%thermostat_info_slow)
!cjm

      NULLIFY (thermostats%thermostat_info_part%map_loc_thermo_gen)
      NULLIFY (thermostats%thermostat_info_shell%map_loc_thermo_gen)
!cjm
      NULLIFY (thermostats%thermostat_info_fast%map_loc_thermo_gen)
      NULLIFY (thermostats%thermostat_fast)
      NULLIFY (thermostats%thermostat_info_slow%map_loc_thermo_gen)
      NULLIFY (thermostats%thermostat_slow)
!cjm
      NULLIFY (thermostats%thermostat_part)
      NULLIFY (thermostats%thermostat_shell)
      NULLIFY (thermostats%thermostat_coef)
      NULLIFY (thermostats%thermostat_baro)

   END SUBROUTINE allocate_thermostats

! **************************************************************************************************
!> \brief retains the  full set of thermostats
!> \param thermostats ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE retain_thermostats(thermostats)
      TYPE(thermostats_type), POINTER                    :: thermostats

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_thermostats', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostats)) THEN
         IF(.NOT.(thermostats%ref_count > 0))CALL cp__a("motion/thermostat/thermostat_types.F",163)
         thermostats%ref_count = thermostats%ref_count+1
      END IF

   END SUBROUTINE retain_thermostats

! **************************************************************************************************
!> \brief ...
!> \param thermostats ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_thermostats(thermostats)
      TYPE(thermostats_type), POINTER                    :: thermostats

      CHARACTER(len=*), PARAMETER :: routineN = 'release_thermostats', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = ASSOCIATED(thermostats)
      IF (check) THEN
         check = thermostats%ref_count > 0
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_types.F",187)
         thermostats%ref_count = thermostats%ref_count-1
         IF (thermostats%ref_count < 1) THEN
            CALL release_thermostat_info(thermostats%thermostat_info_part)
            CALL release_thermostat_info(thermostats%thermostat_info_shell)
!cjm
            CALL release_thermostat_info(thermostats%thermostat_info_fast)
            CALL release_thermostat_info(thermostats%thermostat_info_slow)
            IF (ASSOCIATED(thermostats%thermostat_fast)) THEN
               CALL release_thermostat_type(thermostats%thermostat_fast)
            END IF
            IF (ASSOCIATED(thermostats%thermostat_slow)) THEN
               CALL release_thermostat_type(thermostats%thermostat_slow)
            END IF
!cjm
            IF (ASSOCIATED(thermostats%thermostat_part)) THEN
               CALL release_thermostat_type(thermostats%thermostat_part)
            END IF
            IF (ASSOCIATED(thermostats%thermostat_shell)) THEN
               CALL release_thermostat_type(thermostats%thermostat_shell)
            END IF
            IF (ASSOCIATED(thermostats%thermostat_baro)) THEN
               CALL release_thermostat_type(thermostats%thermostat_baro)
            END IF
            IF (ASSOCIATED(thermostats%thermostat_coef)) THEN
               CALL release_thermostat_type(thermostats%thermostat_coef)
            END IF
            DEALLOCATE (thermostats)
         END IF
      END IF
   END SUBROUTINE release_thermostats

! **************************************************************************************************
!> \brief Create a thermostat type
!> \param thermostat ...
!> \param simpar ...
!> \param section ...
!> \param skip_region ...
!> \param label ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_thermostat_type(thermostat, simpar, section, skip_region, label)
      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(section_vals_type), POINTER                   :: section
      LOGICAL, INTENT(IN), OPTIONAL                      :: skip_region
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermostat_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: region
      LOGICAL                                            :: skip_region_loc
      TYPE(section_vals_type), POINTER                   :: al_section, csvr_section, gle_section, &
                                                            nose_section

      skip_region_loc = .FALSE.
      IF (PRESENT(skip_region)) skip_region_loc = skip_region
      ALLOCATE (thermostat)
      last_thermostat_id_nr = last_thermostat_id_nr+1
      thermostat%id_nr = last_thermostat_id_nr
      thermostat%ref_count = 1
      thermostat%section => section
      thermostat%label = label
      region = do_region_global

      NULLIFY (thermostat%nhc)
      NULLIFY (thermostat%csvr)
      NULLIFY (thermostat%al)
      NULLIFY (thermostat%gle)
      CALL section_vals_val_get(section, "TYPE", i_val=thermostat%type_of_thermostat)
      IF (.NOT. skip_region_loc) CALL section_vals_val_get(section, "REGION", i_val=region)
      IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
         nose_section => section_vals_get_subs_vals(section, "NOSE")
         ALLOCATE (thermostat%nhc)
         CALL lnhc_init(thermostat%nhc, nose_section)
         thermostat%nhc%region = region
      ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
         csvr_section => section_vals_get_subs_vals(section, "CSVR")
         ALLOCATE (thermostat%csvr)
         CALL csvr_init(thermostat%csvr, simpar, csvr_section)
         thermostat%csvr%region = region
      ELSE IF (thermostat%type_of_thermostat == do_thermo_al) THEN
         al_section => section_vals_get_subs_vals(section, "AD_LANGEVIN")
         ALLOCATE (thermostat%al)
         CALL al_init(thermostat%al, simpar, al_section)
         thermostat%al%region = region
      ELSE IF (thermostat%type_of_thermostat == do_thermo_gle) THEN
         gle_section => section_vals_get_subs_vals(section, "GLE")
         ALLOCATE (thermostat%gle)
         CALL gle_init(thermostat%gle, dt=simpar%dt, temp=simpar%temp_ext, &
                       section=gle_section)
         thermostat%gle%region = region
         IF(.NOT.(region == do_region_massive))CALL cp__a("motion/thermostat/thermostat_types.F",282)
      END IF

   END SUBROUTINE create_thermostat_type

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_thermostat_info(thermostat_info)
      TYPE(thermostat_info_type), POINTER                :: thermostat_info

      CHARACTER(len=*), PARAMETER :: routineN = 'release_thermostat_info', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostat_info)) THEN
         IF (ASSOCIATED(thermostat_info%map_loc_thermo_gen)) THEN
            DEALLOCATE (thermostat_info%map_loc_thermo_gen)
         END IF
         DEALLOCATE (thermostat_info)
      END IF
   END SUBROUTINE release_thermostat_info

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_thermostat_type(thermostat)
      TYPE(thermostat_type), POINTER                     :: thermostat

      CHARACTER(len=*), PARAMETER :: routineN = 'release_thermostat_type', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = ASSOCIATED(thermostat)
      IF (check) THEN
         check = thermostat%ref_count > 0
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_types.F",326)
         thermostat%ref_count = thermostat%ref_count-1
         IF (thermostat%ref_count < 1) THEN
            NULLIFY (thermostat%section)
            IF (ASSOCIATED(thermostat%nhc)) THEN
               CALL lnhc_dealloc(thermostat%nhc)
            END IF
            IF (ASSOCIATED(thermostat%csvr)) THEN
               CALL csvr_dealloc(thermostat%csvr)
            END IF
            IF (ASSOCIATED(thermostat%al)) THEN
               CALL al_dealloc(thermostat%al)
            END IF
            IF (ASSOCIATED(thermostat%gle)) THEN
               CALL gle_dealloc(thermostat%gle)
            END IF
            DEALLOCATE (thermostat)
         END IF
      END IF
   END SUBROUTINE release_thermostat_type

! **************************************************************************************************
!> \brief access internal structures of thermostats
!> \param thermostats ...
!> \param dt_fact ...
!> \par History
!>      10.2008 created [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE set_thermostats(thermostats, dt_fact)
      TYPE(thermostats_type), POINTER                    :: thermostats
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: dt_fact

      CHARACTER(len=*), PARAMETER :: routineN = 'set_thermostats', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostats)) THEN
         IF (PRESENT(dt_fact)) THEN
            ! Particles
!cjm
            IF (ASSOCIATED(thermostats%thermostat_fast)) THEN
               SELECT CASE (thermostats%thermostat_fast%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_fast%nhc%dt_fact = dt_fact
               END SELECT
            END IF
            IF (ASSOCIATED(thermostats%thermostat_slow)) THEN
               SELECT CASE (thermostats%thermostat_slow%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_slow%nhc%dt_fact = dt_fact
               END SELECT
            END IF
!cjm
            IF (ASSOCIATED(thermostats%thermostat_part)) THEN
               SELECT CASE (thermostats%thermostat_part%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_part%nhc%dt_fact = dt_fact
               CASE (do_thermo_csvr)
                  thermostats%thermostat_part%csvr%dt_fact = dt_fact
               CASE (do_thermo_al)
                  thermostats%thermostat_part%al%dt_fact = dt_fact
               CASE (do_thermo_gle)
                  thermostats%thermostat_part%gle%dt_fact = dt_fact
               END SELECT
            END IF
            ! Coefficients
            IF (ASSOCIATED(thermostats%thermostat_coef)) THEN
               SELECT CASE (thermostats%thermostat_coef%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_coef%nhc%dt_fact = dt_fact
               CASE (do_thermo_csvr)
                  thermostats%thermostat_coef%csvr%dt_fact = dt_fact
               END SELECT
            END IF
            ! Shell
            IF (ASSOCIATED(thermostats%thermostat_shell)) THEN
               SELECT CASE (thermostats%thermostat_shell%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_shell%nhc%dt_fact = dt_fact
               CASE (do_thermo_csvr)
                  thermostats%thermostat_shell%csvr%dt_fact = dt_fact
               END SELECT
            END IF
            ! Baro
            IF (ASSOCIATED(thermostats%thermostat_baro)) THEN
               SELECT CASE (thermostats%thermostat_baro%type_of_thermostat)
               CASE (do_thermo_nose)
                  thermostats%thermostat_baro%nhc%dt_fact = dt_fact
               CASE (do_thermo_csvr)
                  thermostats%thermostat_baro%csvr%dt_fact = dt_fact
               END SELECT
            END IF
         END IF
      END IF
   END SUBROUTINE set_thermostats

END MODULE thermostat_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
MODULE thermostat_mapping

   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: map_info_type
   USE input_constants,                 ONLY: do_region_defined,&
                                              do_region_global,&
                                              do_region_massive,&
                                              do_region_molecule,&
                                              do_thermo_communication,&
                                              do_thermo_no_communication
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_bcast,&
                                              mp_sum
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              g3x3_constraint_type,&
                                              g4x6_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              molecule_type
   USE simpar_types,                    ONLY: simpar_type
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: locate,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PUBLIC :: thermostat_mapping_region, &
             adiabatic_mapping_region, &
             init_baro_map_info

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermostat_mapping'

CONTAINS
! **************************************************************************************************
!> \brief Main general setup for adiabatic thermostat regions (Nose only)
!> \param map_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param number ...
!> \param region ...
!> \param gci ...
!> \param shell ...
!> \param map_loc_thermo_gen ...
!> \param sum_of_thermostats ...
!> \author CJM - PNNL
! **************************************************************************************************
   SUBROUTINE adiabatic_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                       molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, simpar, &
                                       number, region, gci, shell, map_loc_thermo_gen, sum_of_thermostats)

      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(INOUT)                             :: number
      INTEGER, INTENT(IN)                                :: region
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: shell
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen
      INTEGER, INTENT(INOUT)                             :: sum_of_thermostats

      CHARACTER(LEN=*), PARAMETER :: routineN = 'adiabatic_mapping_region', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nkind, nmol_local, nsize, &
                                                            number_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, tot_const
      INTEGER, DIMENSION(:, :), POINTER                  :: point

      CALL timeset(routineN, handle)

      NULLIFY (const_mol, tot_const, point)
      IF(.NOT.(.NOT. ASSOCIATED(deg_of_freedom)))CALL cp__a("motion/thermostat/thermostat_mapping.F",107)
      IF(.NOT.(.NOT. ASSOCIATED(massive_atom_list)))CALL cp__a("motion/thermostat/thermostat_mapping.F",108)

      nkind = SIZE(molecule_kind_set)
      CALL adiabatic_region_evaluate(map_info%dis_type, natoms_local, nmol_local, &
                                     const_mol, tot_const, point, local_molecules, molecule_kind_set, molecule_set, &
                                     simpar, shell)

      ! Now we can allocate the target array s_kin and p_kin..
      SELECT CASE (region)
      CASE (do_region_global, do_region_molecule, do_region_massive)
         nsize = number
      CASE DEFAULT
         ! STOP PROGRAM
      END SELECT
      ALLOCATE (map_info%s_kin(nsize))
      ALLOCATE (map_info%v_scale(nsize))
      ALLOCATE (map_info%p_kin(3, natoms_local))
      ALLOCATE (map_info%p_scale(3, natoms_local))
! nullify thermostat pointers
      ! Allocate index array to 1
      ALLOCATE (map_info%index(1))
      ALLOCATE (map_info%map_index(1))
      ALLOCATE (deg_of_freedom(1))

      CALL massive_list_generate(molecule_set, molecule_kind_set, &
                                 local_molecules, para_env, massive_atom_list, region, shell)

      CALL adiabatic_mapping_region_low(region, map_info, nkind, point, &
                                        deg_of_freedom, local_molecules, const_mol, massive_atom_list, &
                                        tot_const, molecule_set, number_of_thermostats, shell, gci, &
                                        map_loc_thermo_gen)

      number = number_of_thermostats
      sum_of_thermostats = number
      CALL mp_sum(sum_of_thermostats, para_env%group)

!    check = (number==number_of_thermostats)
!    CPPrecondition(check,cp_fatal_level,routineP,failure)
      DEALLOCATE (const_mol)
      DEALLOCATE (tot_const)
      DEALLOCATE (point)

      CALL timestop(handle)

   END SUBROUTINE adiabatic_mapping_region

! **************************************************************************************************
!> \brief Performs the real mapping for the thermostat region
!> \param region ...
!> \param map_info ...
!> \param nkind ...
!> \param point ...
!> \param deg_of_freedom ...
!> \param local_molecules ...
!> \param const_mol ...
!> \param massive_atom_list ...
!> \param tot_const ...
!> \param molecule_set ...
!> \param ntherm ...
!> \param shell ...
!> \param gci ...
!> \param map_loc_thermo_gen ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE adiabatic_mapping_region_low(region, map_info, nkind, point, &
                                           deg_of_freedom, local_molecules, const_mol, massive_atom_list, tot_const, &
                                           molecule_set, ntherm, shell, gci, map_loc_thermo_gen)

      INTEGER, INTENT(IN)                                :: region
      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER                                            :: nkind
      INTEGER, DIMENSION(:, :), POINTER                  :: point
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, massive_atom_list, tot_const
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(OUT)                               :: ntherm
      LOGICAL, INTENT(IN)                                :: shell
      TYPE(global_constraint_type), POINTER              :: gci
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen

      CHARACTER(LEN=*), PARAMETER :: routineN = 'adiabatic_mapping_region_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: first_atom, first_shell, glob_therm_num, handle, icount, ielement, ii, ikind, &
         imol, imol_local, ipart, jj, k, kk, last_atom, last_shell, nglob_cns, nmol_local, number
      LOGICAL                                            :: check, global_constraints, &
                                                            have_thermostat
      REAL(dp), SAVE, TARGET                             :: unity
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      unity = 1.0_dp
      global_constraints = ASSOCIATED(gci)
      deg_of_freedom = 0
      icount = 0
      number = 0
      ntherm = 0
      nglob_cns = 0
      IF (global_constraints) nglob_cns = gci%ntot-gci%nrestraint
      IF (region == do_region_global) THEN
         ! Global Region
         check = (map_info%dis_type == do_thermo_communication)
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",211)
         DO ikind = 1, nkind
            DO jj = point(1, ikind), point(2, ikind)
               IF (map_loc_thermo_gen(jj) /= HUGE(0)) THEN
                  DO ii = 1, 3
                     map_info%p_kin(ii, jj)%point => map_info%s_kin(1)
                     map_info%p_scale(ii, jj)%point => map_info%v_scale(1)
                  END DO
               ELSE
                  DO ii = 1, 3
                     NULLIFY (map_info%p_kin(ii, jj)%point)
                     map_info%p_scale(ii, jj)%point => unity
                  END DO
               ENDIF
            END DO
            deg_of_freedom(1) = deg_of_freedom(1)+tot_const(ikind)
            map_info%index(1) = 1
            map_info%map_index(1) = 1
            number = 1
         END DO
         deg_of_freedom(1) = deg_of_freedom(1)+nglob_cns
      ELSE IF (region == do_region_molecule) THEN
         ! Molecular Region
         IF (map_info%dis_type == do_thermo_no_communication) THEN
            ! This is the standard case..
            DO ikind = 1, nkind
               nmol_local = local_molecules%n_el(ikind)
               DO imol_local = 1, nmol_local
                  imol = local_molecules%list(ikind)%array(imol_local)
                  number = number+1
                  have_thermostat = .TRUE.
! determine if the local molecule belongs to a thermostat
                  DO kk = point(1, number), point(2, number)
                     !  WRITE ( *, * ) 'kk', size(map_loc_thermo_gen), kk, map_loc_thermo_gen ( kk )
                     IF (map_loc_thermo_gen(kk) == HUGE(0)) THEN
                        have_thermostat = .FALSE.
                        EXIT
                     ENDIF
                  END DO
! If molecule has a thermostat, then map
                  IF (have_thermostat) THEN
! glob_therm_num is the global thermostat number attached to the local molecule
! We can test to make sure all atoms in the local molecule belong to the same
! global thermostat as a way to detect errors.
                     glob_therm_num = map_loc_thermo_gen(point(1, number))
                     ntherm = ntherm+1
                     CALL reallocate(map_info%index, 1, ntherm)
                     CALL reallocate(map_info%map_index, 1, ntherm)
                     CALL reallocate(deg_of_freedom, 1, ntherm)
                     map_info%index(ntherm) = glob_therm_num
                     map_info%map_index(ntherm) = ntherm
                     deg_of_freedom(ntherm) = const_mol(number)
                     DO kk = point(1, number), point(2, number)
                        DO jj = 1, 3
                           map_info%p_kin(jj, kk)%point => map_info%s_kin(ntherm)
                           map_info%p_scale(jj, kk)%point => map_info%v_scale(ntherm)
                        END DO
                     END DO
! If molecule has no thermostat, then nullify pointers to that atom
                  ELSE
                     DO kk = point(1, number), point(2, number)
                        DO jj = 1, 3
                           NULLIFY (map_info%p_kin(jj, kk)%point)
                           map_info%p_scale(jj, kk)%point => unity
                        END DO
                     END DO
                  END IF
               END DO
            END DO
         ELSE IF (map_info%dis_type == do_thermo_communication) THEN
            ! This case is quite rare and happens only when we have one molecular
            ! kind and one molecule..
            IF(.NOT.(nkind == 1))CALL cp__a("motion/thermostat/thermostat_mapping.F",283)
            number = number+1
            ntherm = ntherm+1
            map_info%index(ntherm) = ntherm
            map_info%map_index(ntherm) = ntherm
            deg_of_freedom(ntherm) = deg_of_freedom(ntherm)+tot_const(nkind)
            DO kk = point(1, nkind), point(2, nkind)
               IF (map_loc_thermo_gen(kk) /= HUGE(0)) THEN
                  DO jj = 1, 3
                     map_info%p_kin(jj, kk)%point => map_info%s_kin(ntherm)
                     map_info%p_scale(jj, kk)%point => map_info%v_scale(ntherm)
                  END DO
               ELSE
               END IF
               DO jj = 1, 3
                  NULLIFY (map_info%p_kin(jj, kk)%point)
                  map_info%p_scale(jj, kk)%point => unity
               END DO
            END DO
         ELSE
            CALL cp__b("motion/thermostat/thermostat_mapping.F",303,"")
         END IF
         IF (nglob_cns /= 0) THEN
            CALL cp__b("motion/thermostat/thermostat_mapping.F",306,"Molecular thermostats with global constraints are impossible!")
         END IF
      ELSE IF (region == do_region_massive) THEN
         ! Massive Region
         check = (map_info%dis_type == do_thermo_no_communication)
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",311)
         DO ikind = 1, nkind
            nmol_local = local_molecules%n_el(ikind)
            DO imol_local = 1, nmol_local
               icount = icount+1
               imol = local_molecules%list(ikind)%array(imol_local)
               molecule => molecule_set(imol)
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom, &
                                 first_shell=first_shell, last_shell=last_shell)
               IF (shell) THEN
                  first_atom = first_shell
                  last_atom = last_shell
               ELSE
                  IF ((tot_const(icount) > 0) .OR. (nglob_cns /= 0)) THEN
                     CALL cp__b("motion/thermostat/thermostat_mapping.F",325,"Massive thermostats with constraints are impossible!")
                  END IF
               END IF
               k = 0

               have_thermostat = .TRUE.
               DO ii = point(1, icount), point(2, icount)
                  IF (map_loc_thermo_gen(ii) /= 1) THEN
                     have_thermostat = .FALSE.
                     EXIT
                  END IF
               END DO

               IF (have_thermostat) THEN
                  DO ii = point(1, icount), point(2, icount)
                     ipart = first_atom+k
                     ielement = locate(massive_atom_list, ipart)
                     k = k+1
                     DO jj = 1, 3
                        ntherm = ntherm+1
                        CALL reallocate(map_info%index, 1, ntherm)
                        CALL reallocate(map_info%map_index, 1, ntherm)
                        map_info%index(ntherm) = (ielement-1)*3+jj
                        map_info%map_index(ntherm) = ntherm
                        map_info%p_kin(jj, ii)%point => map_info%s_kin(ntherm)
                        map_info%p_scale(jj, ii)%point => map_info%v_scale(ntherm)
                     END DO
                  END DO
               ELSE
                  DO ii = point(1, icount), point(2, icount)
                     DO jj = 1, 3
                        NULLIFY (map_info%p_kin(jj, ii)%point)
                        map_info%p_scale(jj, ii)%point => unity
                     END DO
                  END DO
               END IF
               IF (first_atom+k-1 /= last_atom) THEN
                  CALL cp__b("motion/thermostat/thermostat_mapping.F",362,"Inconsistent mapping of particles")
               END IF
            END DO
         END DO
      ELSE
         CALL cp__b("motion/thermostat/thermostat_mapping.F",367,"Invalid region!")
      END IF

      CALL timestop(handle)

   END SUBROUTINE adiabatic_mapping_region_low

! **************************************************************************************************
!> \brief creates the mapping between the system and the thermostats
!> \param dis_type ...
!> \param natoms_local ...
!> \param nmol_local ...
!> \param const_mol ...
!> \param tot_const ...
!> \param point ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param simpar ...
!> \param shell ...
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE adiabatic_region_evaluate(dis_type, natoms_local, nmol_local, const_mol, &
                                        tot_const, point, local_molecules, molecule_kind_set, molecule_set, simpar, shell)
      INTEGER, INTENT(IN)                                :: dis_type
      INTEGER, INTENT(OUT)                               :: natoms_local, nmol_local
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, tot_const
      INTEGER, DIMENSION(:, :), POINTER                  :: point
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(simpar_type), POINTER                         :: simpar
      LOGICAL, INTENT(IN)                                :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'adiabatic_region_evaluate', &
         routineP = moduleN//':'//routineN

      INTEGER :: atm_offset, first_atom, handle, icount, ikind, ilist, imol, imol_local, katom, &
         last_atom, natom, nc, nfixd, nkind, nmol_per_kind, nmolecule, nshell
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)

      natoms_local = 0
      nmol_local = 0
      nkind = SIZE(molecule_kind_set)
      NULLIFY (fixd_list, molecule_kind, molecule)
      ! Compute the TOTAL number of molecules and atoms on THIS PROC and
      ! TOTAL number of molecules of IKIND on THIS PROC
      DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, natom=natom, nshell=nshell)
         IF (shell) THEN
            IF (nshell /= 0) THEN
               natoms_local = natoms_local+nshell*local_molecules%n_el(ikind)
               nmol_local = nmol_local+local_molecules%n_el(ikind)
            END IF
         ELSE
            natoms_local = natoms_local+natom*local_molecules%n_el(ikind)
            nmol_local = nmol_local+local_molecules%n_el(ikind)
         END IF
      END DO

      IF(.NOT.(.NOT. ASSOCIATED(const_mol)))CALL cp__a("motion/thermostat/thermostat_mapping.F",432)
      IF(.NOT.(.NOT. ASSOCIATED(tot_const)))CALL cp__a("motion/thermostat/thermostat_mapping.F",433)
      IF(.NOT.(.NOT. ASSOCIATED(point)))CALL cp__a("motion/thermostat/thermostat_mapping.F",434)
      IF (dis_type == do_thermo_no_communication) THEN
         ALLOCATE (const_mol(nmol_local))
         ALLOCATE (tot_const(nmol_local))
         ALLOCATE (point(2, nmol_local))

         point(:, :) = 0
         atm_offset = 0
         icount = 0
         DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, nconstraint=nc, natom=natom, &
                                   fixd_list=fixd_list, nshell=nshell)
            IF (shell) natom = nshell
            DO imol_local = 1, nmol_per_kind
               icount = icount+1
               point(1, icount) = atm_offset+1
               point(2, icount) = atm_offset+natom
               IF (.NOT. shell) THEN
                  ! nc keeps track of all constraints but not fixed ones..
                  ! Let's identify fixed atoms for this molecule
                  nfixd = 0
                  imol = local_molecules%list(ikind)%array(imol_local)
                  molecule => molecule_set(imol)
                  CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                  IF (ASSOCIATED(fixd_list)) THEN
                     DO katom = first_atom, last_atom
                        DO ilist = 1, SIZE(fixd_list)
                           IF ((katom == fixd_list(ilist)%fixd) .AND. &
                               (.NOT. fixd_list(ilist)%restraint%active)) THEN
                              SELECT CASE (fixd_list (ilist)%itype)
                              CASE (use_perd_x, use_perd_y, use_perd_z)
                                 nfixd = nfixd+1
                              CASE (use_perd_xy, use_perd_xz, use_perd_yz)
                                 nfixd = nfixd+2
                              CASE (use_perd_xyz)
                                 nfixd = nfixd+3
                              END SELECT
                           END IF
                        END DO
                     END DO
                  END IF
                  const_mol(icount) = nc+nfixd
                  tot_const(icount) = const_mol(icount)
               END IF
               atm_offset = point(2, icount)
            END DO
         END DO
      ELSE IF (dis_type == do_thermo_communication) THEN
         ALLOCATE (const_mol(nkind))
         ALLOCATE (tot_const(nkind))
         ALLOCATE (point(2, nkind))
         point(:, :) = 0
         atm_offset = 0
         ! nc keeps track of all constraints but not fixed ones..
         DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, nconstraint=nc, natom=natom, &
                                   nmolecule=nmolecule, nconstraint_fixd=nfixd, nshell=nshell)
            IF (shell) natom = nshell
            IF (.NOT. shell) THEN
               const_mol(ikind) = nc
               ! Let's consider the fixed atoms only for the total number of constraints
               ! in case we are in REPLICATED/INTERACTING thermostats
               tot_const(ikind) = const_mol(ikind)*nmolecule+nfixd
            END IF
            point(1, ikind) = atm_offset+1
            point(2, ikind) = atm_offset+natom*nmol_per_kind
            atm_offset = point(2, ikind)
         END DO
      ENDIF
      IF ((.NOT. simpar%constraint) .OR. shell) THEN
         const_mol = 0
         tot_const = 0
      END IF

      CALL timestop(handle)

   END SUBROUTINE adiabatic_region_evaluate

! **************************************************************************************************
!> \brief Main general setup thermostat regions (thermostat independent)
!> \param map_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param number ...
!> \param region ...
!> \param gci ...
!> \param shell ...
!> \param map_loc_thermo_gen ...
!> \param sum_of_thermostats ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE thermostat_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                        molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, simpar, &
                                        number, region, gci, shell, map_loc_thermo_gen, sum_of_thermostats)

      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(IN)                                :: number, region
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: shell
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen
      INTEGER, INTENT(IN)                                :: sum_of_thermostats

      CHARACTER(LEN=*), PARAMETER :: routineN = 'thermostat_mapping_region', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nkind, nmol_local, nsize, &
                                                            number_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, tot_const
      INTEGER, DIMENSION(:, :), POINTER                  :: point
      LOGICAL                                            :: check

      CALL timeset(routineN, handle)

      NULLIFY (const_mol, tot_const, point)
      IF(.NOT.(.NOT. ASSOCIATED(deg_of_freedom)))CALL cp__a("motion/thermostat/thermostat_mapping.F",565)
      IF(.NOT.(.NOT. ASSOCIATED(massive_atom_list)))CALL cp__a("motion/thermostat/thermostat_mapping.F",566)

      nkind = SIZE(molecule_kind_set)
      CALL mapping_region_evaluate(map_info%dis_type, natoms_local, nmol_local, &
                                   const_mol, tot_const, point, local_molecules, molecule_kind_set, molecule_set, &
                                   region, simpar, shell, map_loc_thermo_gen, sum_of_thermostats, para_env)

      ! Now we can allocate the target array s_kin and p_kin..
      SELECT CASE (region)
      CASE (do_region_global, do_region_molecule, do_region_massive)
         nsize = number
      CASE (do_region_defined)
         nsize = sum_of_thermostats
      END SELECT
      ALLOCATE (map_info%s_kin(nsize))
      ALLOCATE (map_info%v_scale(nsize))
      ALLOCATE (map_info%p_kin(3, natoms_local))
      ALLOCATE (map_info%p_scale(3, natoms_local))
      ! Allocate index array
      ALLOCATE (map_info%index(number))
      ALLOCATE (map_info%map_index(number))
      ALLOCATE (deg_of_freedom(number))

      CALL massive_list_generate(molecule_set, molecule_kind_set, &
                                 local_molecules, para_env, massive_atom_list, region, shell)

      CALL thermostat_mapping_region_low(region, map_info, nkind, point, &
                                         deg_of_freedom, local_molecules, const_mol, massive_atom_list, &
                                         tot_const, molecule_set, number_of_thermostats, shell, gci, &
                                         map_loc_thermo_gen)

      check = (number == number_of_thermostats)
      IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",598)
      DEALLOCATE (const_mol)
      DEALLOCATE (tot_const)
      DEALLOCATE (point)

      CALL timestop(handle)

   END SUBROUTINE thermostat_mapping_region

! **************************************************************************************************
!> \brief Performs the real mapping for the thermostat region
!> \param region ...
!> \param map_info ...
!> \param nkind ...
!> \param point ...
!> \param deg_of_freedom ...
!> \param local_molecules ...
!> \param const_mol ...
!> \param massive_atom_list ...
!> \param tot_const ...
!> \param molecule_set ...
!> \param number ...
!> \param shell ...
!> \param gci ...
!> \param map_loc_thermo_gen ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE thermostat_mapping_region_low(region, map_info, nkind, point, &
                                            deg_of_freedom, local_molecules, const_mol, massive_atom_list, tot_const, &
                                            molecule_set, number, shell, gci, map_loc_thermo_gen)

      INTEGER, INTENT(IN)                                :: region
      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER                                            :: nkind
      INTEGER, DIMENSION(:, :), POINTER                  :: point
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, massive_atom_list, tot_const
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(OUT)                               :: number
      LOGICAL, INTENT(IN)                                :: shell
      TYPE(global_constraint_type), POINTER              :: gci
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen

      CHARACTER(LEN=*), PARAMETER :: routineN = 'thermostat_mapping_region_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: first_atom, first_shell, handle, i, icount, ielement, ii, ikind, imap, imol, &
         imol_local, ipart, itmp, jj, k, kk, last_atom, last_shell, nglob_cns, nmol_local
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: tmp, wrk
      LOGICAL                                            :: check, global_constraints
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)

      global_constraints = ASSOCIATED(gci)
      deg_of_freedom = 0
      icount = 0
      number = 0
      nglob_cns = 0
      IF (global_constraints) nglob_cns = gci%ntot-gci%nrestraint
      IF (region == do_region_global) THEN
         ! Global Region
         check = (map_info%dis_type == do_thermo_communication)
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",662)
         DO ikind = 1, nkind
            DO jj = point(1, ikind), point(2, ikind)
               DO ii = 1, 3
                  map_info%p_kin(ii, jj)%point => map_info%s_kin(1)
                  map_info%p_scale(ii, jj)%point => map_info%v_scale(1)
               END DO
            END DO
            deg_of_freedom(1) = deg_of_freedom(1)+tot_const(ikind)
            map_info%index(1) = 1
            map_info%map_index(1) = 1
            number = 1
         END DO
         deg_of_freedom(1) = deg_of_freedom(1)+nglob_cns
      ELSE IF (region == do_region_defined) THEN
         ! User defined Region to thermostat
         check = (map_info%dis_type == do_thermo_communication)
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",679)
         ! Lets' identify the matching of the local thermostat w.r.t. the global one
         itmp = SIZE(map_loc_thermo_gen)
         ALLOCATE (tmp(itmp))
         ALLOCATE (wrk(itmp))
         tmp(:) = map_loc_thermo_gen
         CALL sort(tmp, itmp, wrk)
         number = 1
         map_info%index(number) = tmp(1)
         map_info%map_index(number) = tmp(1)
         deg_of_freedom(number) = tot_const(tmp(1))
         DO i = 2, itmp
            IF (tmp(i) /= tmp(i-1)) THEN
               number = number+1
               map_info%index(number) = tmp(i)
               map_info%map_index(number) = tmp(i)
               deg_of_freedom(number) = tot_const(tmp(i))
            END IF
         END DO
         DEALLOCATE (tmp)
         DEALLOCATE (wrk)
         DO jj = 1, SIZE(map_loc_thermo_gen)
            DO ii = 1, 3
               imap = map_loc_thermo_gen(jj)
               map_info%p_kin(ii, jj)%point => map_info%s_kin(imap)
               map_info%p_scale(ii, jj)%point => map_info%v_scale(imap)
            END DO
         END DO
         IF (nglob_cns /= 0) THEN
            CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",708), &
                          "User Defined thermostats with global constraints not implemented!")
         END IF
      ELSE IF (region == do_region_molecule) THEN
         ! Molecular Region
         IF (map_info%dis_type == do_thermo_no_communication) THEN
            ! This is the standard case..
            DO ikind = 1, nkind
               nmol_local = local_molecules%n_el(ikind)
               DO imol_local = 1, nmol_local
                  imol = local_molecules%list(ikind)%array(imol_local)
                  number = number+1
                  map_info%index(number) = imol
                  map_info%map_index(number) = number
                  deg_of_freedom(number) = const_mol(number)
                  DO kk = point(1, number), point(2, number)
                     DO jj = 1, 3
                        map_info%p_kin(jj, kk)%point => map_info%s_kin(number)
                        map_info%p_scale(jj, kk)%point => map_info%v_scale(number)
                     END DO
                  END DO
               END DO
            END DO
         ELSE IF (map_info%dis_type == do_thermo_communication) THEN
            ! This case is quite rare and happens only when we have one molecular
            ! kind and one molecule..
            IF(.NOT.(nkind == 1))CALL cp__a("motion/thermostat/thermostat_mapping.F",734)
            number = number+1
            map_info%index(number) = number
            map_info%map_index(number) = number
            deg_of_freedom(number) = deg_of_freedom(number)+tot_const(nkind)
            DO kk = point(1, nkind), point(2, nkind)
               DO jj = 1, 3
                  map_info%p_kin(jj, kk)%point => map_info%s_kin(number)
                  map_info%p_scale(jj, kk)%point => map_info%v_scale(number)
               END DO
            END DO
         ELSE
            CALL cp__b("motion/thermostat/thermostat_mapping.F",746,"")
         END IF
         IF (nglob_cns /= 0) THEN
            CALL cp__b("motion/thermostat/thermostat_mapping.F",749,"Molecular thermostats with global constraints are impossible!")
         END IF
      ELSE IF (region == do_region_massive) THEN
         ! Massive Region
         check = (map_info%dis_type == do_thermo_no_communication)
         IF(.NOT.(check))CALL cp__a("motion/thermostat/thermostat_mapping.F",754)
         DO ikind = 1, nkind
            nmol_local = local_molecules%n_el(ikind)
            DO imol_local = 1, nmol_local
               icount = icount+1
               imol = local_molecules%list(ikind)%array(imol_local)
               molecule => molecule_set(imol)
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom, &
                                 first_shell=first_shell, last_shell=last_shell)
               IF (shell) THEN
                  first_atom = first_shell
                  last_atom = last_shell
               ELSE
                  IF ((tot_const(icount) > 0) .OR. (nglob_cns /= 0)) THEN
                     CALL cp__b("motion/thermostat/thermostat_mapping.F",768,"Massive thermostats with constraints are impossible!")
                  END IF
               END IF
               k = 0
               DO ii = point(1, icount), point(2, icount)
                  ipart = first_atom+k
                  ielement = locate(massive_atom_list, ipart)
                  k = k+1
                  DO jj = 1, 3
                     number = number+1
                     map_info%index(number) = (ielement-1)*3+jj
                     map_info%map_index(number) = number
                     map_info%p_kin(jj, ii)%point => map_info%s_kin(number)
                     map_info%p_scale(jj, ii)%point => map_info%v_scale(number)
                  END DO
               END DO
               IF (first_atom+k-1 /= last_atom) THEN
                  CALL cp__b("motion/thermostat/thermostat_mapping.F",785,"Inconsistent mapping of particles")
               END IF
            END DO
         END DO
      ELSE
         CALL cp__b("motion/thermostat/thermostat_mapping.F",790,"Invalid region!")
      END IF

      CALL timestop(handle)

   END SUBROUTINE thermostat_mapping_region_low

! **************************************************************************************************
!> \brief creates the mapping between the system and the thermostats
!> \param dis_type ...
!> \param natoms_local ...
!> \param nmol_local ...
!> \param const_mol ...
!> \param tot_const ...
!> \param point ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param region ...
!> \param simpar ...
!> \param shell ...
!> \param map_loc_thermo_gen ...
!> \param sum_of_thermostats ...
!> \param para_env ...
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE mapping_region_evaluate(dis_type, natoms_local, nmol_local, const_mol, &
                                      tot_const, point, local_molecules, molecule_kind_set, molecule_set, region, &
                                      simpar, shell, map_loc_thermo_gen, sum_of_thermostats, para_env)
      INTEGER, INTENT(IN)                                :: dis_type
      INTEGER, INTENT(OUT)                               :: natoms_local, nmol_local
      INTEGER, DIMENSION(:), POINTER                     :: const_mol, tot_const
      INTEGER, DIMENSION(:, :), POINTER                  :: point
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(IN)                                :: region
      TYPE(simpar_type), POINTER                         :: simpar
      LOGICAL, INTENT(IN)                                :: shell
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen
      INTEGER, INTENT(IN)                                :: sum_of_thermostats
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mapping_region_evaluate', &
         routineP = moduleN//':'//routineN

      INTEGER :: atm_offset, first_atom, handle, i, iatm, icount, id_region, ikind, ilist, imol, &
         imol_local, j, jatm, katom, last_atom, natom, nc, nfixd, nkind, nmol_per_kind, nmolecule, &
         nshell
      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER  :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER  :: g4x6_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)

      natoms_local = 0
      nmol_local = 0
      nkind = SIZE(molecule_kind_set)
      NULLIFY (fixd_list, molecule_kind, molecule, colv_list, g3x3_list, g4x6_list)
      ! Compute the TOTAL number of molecules and atoms on THIS PROC and
      ! TOTAL number of molecules of IKIND on THIS PROC
      DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, natom=natom, nshell=nshell)
         IF (shell) THEN
            IF (nshell /= 0) THEN
               natoms_local = natoms_local+nshell*local_molecules%n_el(ikind)
               nmol_local = nmol_local+local_molecules%n_el(ikind)
            END IF
         ELSE
            natoms_local = natoms_local+natom*local_molecules%n_el(ikind)
            nmol_local = nmol_local+local_molecules%n_el(ikind)
         END IF
      END DO

      IF(.NOT.(.NOT. ASSOCIATED(const_mol)))CALL cp__a("motion/thermostat/thermostat_mapping.F",869)
      IF(.NOT.(.NOT. ASSOCIATED(tot_const)))CALL cp__a("motion/thermostat/thermostat_mapping.F",870)
      IF(.NOT.(.NOT. ASSOCIATED(point)))CALL cp__a("motion/thermostat/thermostat_mapping.F",871)
      IF (dis_type == do_thermo_no_communication) THEN
         ALLOCATE (const_mol(nmol_local))
         ALLOCATE (tot_const(nmol_local))
         ALLOCATE (point(2, nmol_local))

         point(:, :) = 0
         atm_offset = 0
         icount = 0
         DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, nconstraint=nc, natom=natom, &
                                   fixd_list=fixd_list, nshell=nshell)
            IF (shell) natom = nshell
            DO imol_local = 1, nmol_per_kind
               icount = icount+1
               point(1, icount) = atm_offset+1
               point(2, icount) = atm_offset+natom
               IF (.NOT. shell) THEN
                  ! nc keeps track of all constraints but not fixed ones..
                  ! Let's identify fixed atoms for this molecule
                  nfixd = 0
                  imol = local_molecules%list(ikind)%array(imol_local)
                  molecule => molecule_set(imol)
                  CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                  IF (ASSOCIATED(fixd_list)) THEN
                     DO katom = first_atom, last_atom
                        DO ilist = 1, SIZE(fixd_list)
                           IF ((katom == fixd_list(ilist)%fixd) .AND. &
                               (.NOT. fixd_list(ilist)%restraint%active)) THEN
                              SELECT CASE (fixd_list (ilist)%itype)
                              CASE (use_perd_x, use_perd_y, use_perd_z)
                                 nfixd = nfixd+1
                              CASE (use_perd_xy, use_perd_xz, use_perd_yz)
                                 nfixd = nfixd+2
                              CASE (use_perd_xyz)
                                 nfixd = nfixd+3
                              END SELECT
                           END IF
                        END DO
                     END DO
                  END IF
                  const_mol(icount) = nc+nfixd
                  tot_const(icount) = const_mol(icount)
               END IF
               atm_offset = point(2, icount)
            END DO
         END DO
      ELSE IF (dis_type == do_thermo_communication) THEN
         IF (region == do_region_defined) THEN
            ! Setup of the arbitrary region
            ALLOCATE (tot_const(sum_of_thermostats))
            ALLOCATE (point(2, 0))
            ALLOCATE (const_mol(0))
            atm_offset = 0
            tot_const = 0
            const_mol = 0
            point = 0
            DO ikind = 1, nkind
               nmol_per_kind = local_molecules%n_el(ikind)
               molecule_kind => molecule_kind_set(ikind)
               CALL get_molecule_kind(molecule_kind, nconstraint=nc, natom=natom, &
                                      fixd_list=fixd_list, colv_list=colv_list, g3x3_list=g3x3_list, &
                                      g4x6_list=g4x6_list, nshell=nshell)
               IF (shell) natom = nshell
               DO imol_local = 1, nmol_per_kind
                  IF (.NOT. shell) THEN
                     ! First if nc is not zero let's check if all atoms of a molecule
                     ! are in the same thermostatting region..
                     imol = local_molecules%list(ikind)%array(imol_local)
                     molecule => molecule_set(imol)
                     id_region = map_loc_thermo_gen(atm_offset+1)
                     IF (ALL(map_loc_thermo_gen(atm_offset+1:atm_offset+natom) == id_region)) THEN
                        ! All the atoms of a molecule are within the same thermostatting
                        ! region.. this is the easy case..
                        tot_const(id_region) = tot_const(id_region)+nc
                     ELSE
                        ! If not let's check the single constraints defined for this molecule
                        ! and continue only when atoms involved in the constraint belong to
                        ! the same thermostatting region
                        IF (ASSOCIATED(colv_list)) THEN
                           DO i = 1, SIZE(colv_list)
                              IF (.NOT. colv_list(i)%restraint%active) THEN
                                 iatm = atm_offset+colv_list(i)%i_atoms(1)
                                 DO j = 2, SIZE(colv_list(i)%i_atoms)
                                    jatm = atm_offset+colv_list(i)%i_atoms(j)
                                    IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                       CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",959), &
                                                     "User Defined Region: "// &
                                                     "A constraint (COLV) was defined between two thermostatting regions! "// &
                                                     "This is not allowed!")
                                    END IF
                                 END DO
                                 id_region = map_loc_thermo_gen(iatm)
                                 tot_const(id_region) = tot_const(id_region)+1
                              END IF
                           END DO
                        END IF
                        IF (ASSOCIATED(g3x3_list)) THEN
                           DO i = 1, SIZE(g3x3_list)
                              IF (.NOT. g3x3_list(i)%restraint%active) THEN
                                 iatm = atm_offset+g3x3_list(i)%a
                                 jatm = atm_offset+g3x3_list(i)%b
                                 IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                    CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",976), &
                                                  "User Defined Region: "// &
                                                  "A constraint (G3X3) was defined between two thermostatting regions! "// &
                                                  "This is not allowed!")
                                 END IF
                                 jatm = atm_offset+g3x3_list(i)%c
                                 IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                    CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",983), &
                                                  "User Defined Region: "// &
                                                  "A constraint (G3X3) was defined between two thermostatting regions! "// &
                                                  "This is not allowed!")
                                 END IF
                              END IF
                              id_region = map_loc_thermo_gen(iatm)
                              tot_const(id_region) = tot_const(id_region)+3
                           END DO
                        END IF
                        IF (ASSOCIATED(g4x6_list)) THEN
                           DO i = 1, SIZE(g4x6_list)
                              IF (.NOT. g4x6_list(i)%restraint%active) THEN
                                 iatm = atm_offset+g4x6_list(i)%a
                                 jatm = atm_offset+g4x6_list(i)%b
                                 IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                    CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",999), &
                                                  " User Defined Region: "// &
                                                  "A constraint (G4X6) was defined between two thermostatting regions! "// &
                                                  "This is not allowed!")
                                 END IF
                                 jatm = atm_offset+g4x6_list(i)%c
                                 IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                    CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",1006), &
                                                  " User Defined Region: "// &
                                                  "A constraint (G4X6) was defined between two thermostatting regions! "// &
                                                  "This is not allowed!")
                                 END IF
                                 jatm = atm_offset+g4x6_list(i)%d
                                 IF (map_loc_thermo_gen(iatm) /= map_loc_thermo_gen(jatm)) THEN
                                    CALL cp_abort(cp__l("motion/thermostat/thermostat_mapping.F",1013), &
                                                  " User Defined Region: "// &
                                                  "A constraint (G4X6) was defined between two thermostatting regions! "// &
                                                  "This is not allowed!")
                                 END IF
                              END IF
                              id_region = map_loc_thermo_gen(iatm)
                              tot_const(id_region) = tot_const(id_region)+6
                           END DO
                        END IF
                     END IF
                     ! Here we handle possibly fixed atoms
                     IF (ASSOCIATED(fixd_list)) THEN
                        CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                        iatm = 0
                        DO katom = first_atom, last_atom
                           iatm = iatm+1
                           DO ilist = 1, SIZE(fixd_list)
                              IF ((katom == fixd_list(ilist)%fixd) .AND. &
                                  (.NOT. fixd_list(ilist)%restraint%active)) THEN
                                 id_region = map_loc_thermo_gen(atm_offset+iatm)
                                 SELECT CASE (fixd_list (ilist)%itype)
                                 CASE (use_perd_x, use_perd_y, use_perd_z)
                                    tot_const(id_region) = tot_const(id_region)+1
                                 CASE (use_perd_xy, use_perd_xz, use_perd_yz)
                                    tot_const(id_region) = tot_const(id_region)+2
                                 CASE (use_perd_xyz)
                                    tot_const(id_region) = tot_const(id_region)+3
                                 END SELECT
                              END IF
                           END DO
                        END DO
                     END IF
                  END IF
                  atm_offset = atm_offset+natom
               END DO
            END DO
            CALL mp_sum(tot_const, para_env%group)
         ELSE
            ALLOCATE (const_mol(nkind))
            ALLOCATE (tot_const(nkind))
            ALLOCATE (point(2, nkind))
            point(:, :) = 0
            atm_offset = 0
            ! nc keeps track of all constraints but not fixed ones..
            DO ikind = 1, nkind
               nmol_per_kind = local_molecules%n_el(ikind)
               molecule_kind => molecule_kind_set(ikind)
               CALL get_molecule_kind(molecule_kind, nconstraint=nc, natom=natom, &
                                      nmolecule=nmolecule, nconstraint_fixd=nfixd, nshell=nshell)
               IF (shell) natom = nshell
               IF (.NOT. shell) THEN
                  const_mol(ikind) = nc
                  ! Let's consider the fixed atoms only for the total number of constraints
                  ! in case we are in REPLICATED/INTERACTING thermostats
                  tot_const(ikind) = const_mol(ikind)*nmolecule+nfixd
               END IF
               point(1, ikind) = atm_offset+1
               point(2, ikind) = atm_offset+natom*nmol_per_kind
               atm_offset = point(2, ikind)
            END DO
         END IF
      ENDIF
      IF ((.NOT. simpar%constraint) .OR. shell) THEN
         const_mol = 0
         tot_const = 0
      END IF

      CALL timestop(handle)

   END SUBROUTINE mapping_region_evaluate

! **************************************************************************************************
!> \brief ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param para_env ...
!> \param massive_atom_list ...
!> \param region ...
!> \param shell ...
! **************************************************************************************************
   SUBROUTINE massive_list_generate(molecule_set, molecule_kind_set, &
                                    local_molecules, para_env, massive_atom_list, region, shell)

      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, POINTER                                   :: massive_atom_list(:)
      INTEGER, INTENT(IN)                                :: region
      LOGICAL, INTENT(IN)                                :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'massive_list_generate', &
         routineP = moduleN//':'//routineN

      INTEGER :: first_atom, first_shell, handle, i, ikind, imol, iproc, j, natom, ncount, nkind, &
         nmol_per_kind, nshell, num_massive_atm, num_massive_atm_local, offset, stat
      INTEGER, DIMENSION(:), POINTER                     :: array_num_massive_atm, local_atm_list, &
                                                            work
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)

      num_massive_atm_local = 0
      NULLIFY (local_atm_list)
      CALL reallocate(local_atm_list, 1, num_massive_atm_local)

      nkind = SIZE(molecule_kind_set)
      DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, natom=natom, nshell=nshell)
            IF (region == do_region_massive) THEN
               IF (shell) THEN
                  natom = nshell
               END IF
               num_massive_atm_local = num_massive_atm_local+natom
               CALL reallocate(local_atm_list, 1, num_massive_atm_local)
               CALL get_molecule(molecule, first_atom=first_atom, first_shell=first_shell)
               IF (shell) THEN
                  first_atom = first_shell
               END IF
               DO j = 1, natom
                  local_atm_list(num_massive_atm_local-natom+j) = first_atom-1+j
               END DO
            END IF
         END DO
      END DO

      ALLOCATE (array_num_massive_atm(para_env%num_pe))
      CALL mp_allgather(num_massive_atm_local, array_num_massive_atm, para_env%group)

      num_massive_atm = SUM(array_num_massive_atm)
      ALLOCATE (massive_atom_list(num_massive_atm))

      offset = 0
      DO iproc = 1, para_env%num_pe
         ncount = array_num_massive_atm(iproc)
         ALLOCATE (work(ncount))
         IF (para_env%mepos == (iproc-1)) THEN
            DO i = 1, ncount
               work(i) = local_atm_list(i)
            END DO
         ELSE
            work(:) = 0
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         DO i = 1, ncount
            massive_atom_list(offset+i) = work(i)
         END DO
         DEALLOCATE (work)
         offset = offset+array_num_massive_atm(iproc)
      END DO

      ! Sort atom list
      ALLOCATE (work(num_massive_atm), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1174, &
                                      "work", int_size*num_massive_atm)
      CALL sort(massive_atom_list, num_massive_atm, work)
      DEALLOCATE (work)

      DEALLOCATE (local_atm_list)
      DEALLOCATE (array_num_massive_atm)

      CALL timestop(handle)

   END SUBROUTINE massive_list_generate

! **************************************************************************************************
!> \brief Initialize the map_info for barostat thermostat
!> \param map_info ...
!> \param ndeg ...
!> \param num_thermo ...
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE init_baro_map_info(map_info, ndeg, num_thermo)

      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER, INTENT(IN)                                :: ndeg, num_thermo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_baro_map_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      ALLOCATE (map_info%s_kin(num_thermo))
      ALLOCATE (map_info%v_scale(num_thermo))
      ALLOCATE (map_info%p_kin(1, ndeg))
      ALLOCATE (map_info%p_scale(1, ndeg))
      ! Allocate the index array
      ALLOCATE (map_info%index(1))
      ALLOCATE (map_info%map_index(1))

      ! Begin the mapping loop
      DO i = 1, ndeg
         map_info%p_kin(1, i)%point => map_info%s_kin(1)
         map_info%p_scale(1, i)%point => map_info%v_scale(1)
      END DO
      map_info%index(1) = 1
      map_info%map_index(1) = 1

      CALL timestop(handle)

   END SUBROUTINE init_baro_map_info

END MODULE thermostat_mapping

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      CJM, 20-Feb-01
!>      JGH (10-Mar-2001)
!>      CJM (10-Apr-2001)
!> \author CJM
! **************************************************************************************************
MODULE extended_system_mapping

   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: debug_isotropic_limit,&
                                              lnhc_parameters_type,&
                                              map_info_type
   USE input_constants,                 ONLY: &
        do_thermo_communication, do_thermo_no_communication, do_thermo_only_master, &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble, reftraj_ensemble
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_mapping,              ONLY: adiabatic_mapping_region,&
                                              init_baro_map_info,&
                                              thermostat_mapping_region
   USE thermostat_types,                ONLY: thermostat_info_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'extended_system_mapping'

   PUBLIC :: nhc_to_particle_mapping, nhc_to_barostat_mapping, &
             nhc_to_shell_mapping, nhc_to_particle_mapping_fast, &
             nhc_to_particle_mapping_slow

CONTAINS

! **************************************************************************************************
!> \brief Creates the thermostatting for the barostat
!> \param simpar ...
!> \param nhc ...
!> \par History
!>      CJM, 20-Feb-01  : nhc structure allocated to zero when not in use
!>      JGH (10-Mar-2001) : set nhc variables to zero when not in use
!> \author CJM
! **************************************************************************************************
   SUBROUTINE nhc_to_barostat_mapping(simpar, nhc)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(lnhc_parameters_type), POINTER                :: nhc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nhc_to_barostat_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, number
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_mapping.F",73,'Never reach this point!')
      CASE (npt_i_ensemble, npt_f_ensemble)
         map_info => nhc%map_info
         map_info%dis_type = do_thermo_only_master

         ! Counting the total number of thermostats ( 1 for both NPT_I and NPT_F )
         nhc%loc_num_nhc = 1
         nhc%glob_num_nhc = 1
         IF (simpar%ensemble == npt_f_ensemble) THEN
            number = 9
         ELSE
            number = 1
         ENDIF

         CALL init_baro_map_info(map_info, number, nhc%loc_num_nhc)

         ALLOCATE (nhc%nvt(nhc%nhc_len, nhc%loc_num_nhc))
         ! Now that we know how many there are stick this into nhc % nkt
         ! (number of degrees of freedom times k_B T )
         DO i = 1, nhc%loc_num_nhc
            nhc%nvt(1, i)%nkt = simpar%temp_ext*number
            nhc%nvt(1, i)%degrees_of_freedom = number
            IF (debug_isotropic_limit) THEN
               nhc%nvt(1, i)%nkt = simpar%temp_ext
            END IF
         END DO

         ! getting the number of degrees of freedom times k_B T for the rest of the chain
         DO i = 2, nhc%nhc_len
            nhc%nvt(i, :)%nkt = simpar%temp_ext
         END DO

         ! Let's clean the arrays
         map_info%s_kin = 0.0_dp
         map_info%v_scale = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE nhc_to_barostat_mapping

! **************************************************************************************************
!> \brief Creates the thermostatting maps
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param nhc ...
!> \param para_env ...
!> \param gci ...
!> \par History
!>      29-Nov-00 (JGH) correct counting of DOF if constraints are off
!>      CJM, 20-Feb-01  : nhc structure allocated to zero when not in use
!>      JGH (10-Mar-2001) : set nhc variables to zero when not in use
!>      CJM(10-NOV-2001) : New parallelization with new molecule structures
!>      Teodoro Laino 09.2007 [tlaino] - University of Zurich - cleaning and updating
!> \author CJM
! **************************************************************************************************
   SUBROUTINE nhc_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                      molecule_set, molecule_kind_set, nhc, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nhc_to_particle_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imap, j, natoms_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      NULLIFY (massive_atom_list, deg_of_freedom)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_mapping.F",159,'Unknown ensemble!')
      CASE (nve_ensemble, isokin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/extended_system_mapping.F",162,'Never reach this point!')
      CASE (nvt_ensemble, npt_i_ensemble, npt_f_ensemble)

         CALL setup_nhc_thermostat(nhc, thermostat_info, deg_of_freedom, massive_atom_list, &
                                   molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                   simpar, sum_of_thermostats, gci)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info => nhc%map_info
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! if thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/extended_system_mapping.F",187,'Zero degrees of freedom. Nothing to thermalize!')
            END IF
            nhc%nvt(1, 1)%nkt = simpar%temp_ext*fac
            nhc%nvt(1, 1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, nhc%loc_num_nhc
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               nhc%nvt(1, i)%nkt = simpar%temp_ext*fac
               nhc%nvt(1, i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF

         ! Getting the number of degrees of freedom times k_B T for the rest
         ! of the chain
         DO i = 2, nhc%nhc_len
            nhc%nvt(i, :)%nkt = simpar%temp_ext
            nhc%nvt(i, :)%degrees_of_freedom = 1
         END DO
         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)

         ! Let's clean the arrays
         map_info%s_kin = 0.0_dp
         map_info%v_scale = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE nhc_to_particle_mapping

! **************************************************************************************************
!> \brief Main general setup for Adiabatic Nose-Hoover thermostats
!> \param nhc ...
!> \param thermostat_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param sum_of_thermostats ...
!> \param gci ...
!> \param shell ...
!> \author CJM -PNNL -2011
! **************************************************************************************************
   SUBROUTINE setup_adiabatic_thermostat(nhc, thermostat_info, deg_of_freedom, &
                                         massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                         para_env, natoms_local, simpar, sum_of_thermostats, gci, shell)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(OUT)                               :: sum_of_thermostats
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_adiabatic_thermostat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nkind, number, region
      LOGICAL                                            :: do_shell
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      map_info => nhc%map_info

      nkind = SIZE(molecule_kind_set)
      sum_of_thermostats = thermostat_info%sum_of_thermostats
      map_info%dis_type = thermostat_info%dis_type
      number = thermostat_info%number_of_thermostats
      region = nhc%region

      CALL adiabatic_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                    molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                    simpar, number, region, gci, do_shell, thermostat_info%map_loc_thermo_gen, &
                                    sum_of_thermostats)
      ALLOCATE (nhc%nvt(nhc%nhc_len, number))

      ! Now that we know how many there are stick this into nhc%nkt
      ! (number of degrees of freedom times k_B T for the first thermostat
      !  on the chain)
      nhc%loc_num_nhc = number
      nhc%glob_num_nhc = sum_of_thermostats

      CALL timestop(handle)

   END SUBROUTINE setup_adiabatic_thermostat

! **************************************************************************************************
!> \brief Creates the thermostatting maps
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param nhc ...
!> \param para_env ...
!> \param gci ...
!> \par History
!> \author CJM
! **************************************************************************************************
   SUBROUTINE nhc_to_particle_mapping_slow(thermostat_info, simpar, local_molecules, &
                                           molecule_set, molecule_kind_set, nhc, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nhc_to_particle_mapping_slow', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imap, j, natoms_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      NULLIFY (massive_atom_list, deg_of_freedom)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_mapping.F",327,'Unknown ensemble!')
      CASE (nvt_adiabatic_ensemble)
         CALL setup_adiabatic_thermostat(nhc, thermostat_info, deg_of_freedom, massive_atom_list, &
                                         molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                         simpar, sum_of_thermostats, gci)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info => nhc%map_info
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               IF (ASSOCIATED(map_info%p_kin(i, j)%point)) &
                  map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! if thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/extended_system_mapping.F",352,'Zero degrees of freedom. Nothing to thermalize!')
            END IF
            nhc%nvt(1, 1)%nkt = simpar%temp_slow*fac
            nhc%nvt(1, 1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, nhc%loc_num_nhc
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               nhc%nvt(1, i)%nkt = simpar%temp_slow*fac
               nhc%nvt(1, i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF

         ! Getting the number of degrees of freedom times k_B T for the rest
         ! of the chain
         DO i = 2, nhc%nhc_len
            nhc%nvt(i, :)%nkt = simpar%temp_slow
            nhc%nvt(i, :)%degrees_of_freedom = 1
         END DO
         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)

         ! Let's clean the arrays
         map_info%s_kin = 0.0_dp
         map_info%v_scale = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE nhc_to_particle_mapping_slow

! **************************************************************************************************
!> \brief Creates the thermostatting maps
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param nhc ...
!> \param para_env ...
!> \param gci ...
!> \par History
!> \author CJM
! **************************************************************************************************
   SUBROUTINE nhc_to_particle_mapping_fast(thermostat_info, simpar, local_molecules, &
                                           molecule_set, molecule_kind_set, nhc, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nhc_to_particle_mapping_fast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imap, j, natoms_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      NULLIFY (massive_atom_list, deg_of_freedom)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_mapping.F",423,'Unknown ensemble!')
      CASE (nvt_adiabatic_ensemble)
         CALL setup_adiabatic_thermostat(nhc, thermostat_info, deg_of_freedom, massive_atom_list, &
                                         molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                         simpar, sum_of_thermostats, gci)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info => nhc%map_info
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               IF (ASSOCIATED(map_info%p_kin(i, j)%point)) &
                  map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! if thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/extended_system_mapping.F",448,'Zero degrees of freedom. Nothing to thermalize!')
            END IF
            nhc%nvt(1, 1)%nkt = simpar%temp_fast*fac
            nhc%nvt(1, 1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, nhc%loc_num_nhc
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               nhc%nvt(1, i)%nkt = simpar%temp_fast*fac
               nhc%nvt(1, i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF

         ! Getting the number of degrees of freedom times k_B T for the rest
         ! of the chain
         DO i = 2, nhc%nhc_len
            nhc%nvt(i, :)%nkt = simpar%temp_fast
            nhc%nvt(i, :)%degrees_of_freedom = 1
         END DO
         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)

         ! Let's clean the arrays
         map_info%s_kin = 0.0_dp
         map_info%v_scale = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE nhc_to_particle_mapping_fast

! **************************************************************************************************
!> \brief Main general setup for Nose-Hoover thermostats
!> \param nhc ...
!> \param thermostat_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param sum_of_thermostats ...
!> \param gci ...
!> \param shell ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE setup_nhc_thermostat(nhc, thermostat_info, deg_of_freedom, &
                                   massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                   para_env, natoms_local, simpar, sum_of_thermostats, gci, shell)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(OUT)                               :: sum_of_thermostats
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_nhc_thermostat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nkind, number, region
      LOGICAL                                            :: do_shell
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      map_info => nhc%map_info

      nkind = SIZE(molecule_kind_set)
      sum_of_thermostats = thermostat_info%sum_of_thermostats
      map_info%dis_type = thermostat_info%dis_type
      number = thermostat_info%number_of_thermostats
      region = nhc%region

      CALL thermostat_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                     molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                     simpar, number, region, gci, do_shell, thermostat_info%map_loc_thermo_gen, &
                                     sum_of_thermostats)

      ALLOCATE (nhc%nvt(nhc%nhc_len, number))

      ! Now that we know how many there are stick this into nhc%nkt
      ! (number of degrees of freedom times k_B T for the first thermostat
      !  on the chain)
      nhc%loc_num_nhc = number
      nhc%glob_num_nhc = sum_of_thermostats

      CALL timestop(handle)

   END SUBROUTINE setup_nhc_thermostat

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param nhc ...
!> \param para_env ...
!> \param gci ...
! **************************************************************************************************
   SUBROUTINE nhc_to_shell_mapping(thermostat_info, simpar, local_molecules, &
                                   molecule_set, molecule_kind_set, nhc, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nhc_to_shell_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imap, j, nshell_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_shell_list
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      NULLIFY (massive_shell_list, deg_of_freedom)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_mapping.F",586,'Unknown ensemble!')
      CASE (isokin_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/extended_system_mapping.F",589,'Never reach this point!')
      CASE (nve_ensemble, nvt_ensemble, npe_f_ensemble, npe_i_ensemble, npt_i_ensemble, npt_f_ensemble)

         CALL setup_nhc_thermostat(nhc, thermostat_info, deg_of_freedom, massive_shell_list, &
                                   molecule_kind_set, local_molecules, molecule_set, para_env, nshell_local, &
                                   simpar, sum_of_thermostats, gci, shell=.TRUE.)

         map_info => nhc%map_info
         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target, via p_kin init s_kin
         map_info%s_kin = 0.0_dp
         DO j = 1, nshell_local
            DO i = 1, 3
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! If thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! Now that we know how many there are stick this into nhc%nkt
         ! (number of degrees of freedom times k_B T )
         DO i = 1, nhc%loc_num_nhc
            imap = map_info%map_index(i)
            nhc%nvt(1, i)%nkt = simpar%temp_sh_ext*map_info%s_kin(imap)
            nhc%nvt(1, i)%degrees_of_freedom = INT(map_info%s_kin(imap))
         END DO

         ! Getting the number of degrees of freedom times k_B T for the rest of the chain
         DO i = 2, nhc%nhc_len
            nhc%nvt(i, :)%nkt = simpar%temp_sh_ext
            nhc%nvt(i, :)%degrees_of_freedom = 1
         END DO
         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_shell_list)

         ! Let's clean the arrays
         map_info%s_kin = 0.0_dp
         map_info%v_scale = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE nhc_to_shell_mapping

END MODULE extended_system_mapping
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for thermostats
!> \author teo [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
MODULE thermostat_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: lnhc_parameters_type,&
                                              map_info_type,&
                                              npt_info_type
   USE input_constants,                 ONLY: &
        do_constr_atomic, do_constr_molec, do_region_defined, do_region_global, do_region_massive, &
        do_region_molecule, do_thermo_al, do_thermo_communication, do_thermo_csvr, do_thermo_gle, &
        do_thermo_no_communication, do_thermo_nose, isokin_ensemble, langevin_ensemble, &
        npe_f_ensemble, npe_i_ensemble, nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, &
        npt_f_ensemble, npt_i_ensemble, nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble, &
        reftraj_ensemble
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE message_passing,                 ONLY: mp_sum
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              get_molecule_kind_set,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              molecule_type
   USE motion_utils,                    ONLY: rot_ana
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: femtoseconds
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE shell_potential_types,           ONLY: shell_kind_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: thermostat_info_type,&
                                              thermostat_type,&
                                              thermostats_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: compute_degrees_of_freedom, &
             compute_nfree, &
             setup_thermostat_info, &
             setup_adiabatic_thermostat_info, &
             ke_region_baro, &
             ke_region_particles, &
             ke_region_shells, &
             vel_rescale_baro, &
             vel_rescale_particles, &
             vel_rescale_shells, &
             get_thermostat_energies, &
             get_nhc_energies, &
             get_kin_energies, &
             communication_thermo_low2, &
             print_thermostats_status, &
             momentum_region_particles

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermostat_utils'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cell ...
!> \param simpar ...
!> \param molecule_kind_set ...
!> \param print_section ...
!> \param particles ...
!> \param gci ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE compute_nfree(cell, simpar, molecule_kind_set, &
                            print_section, particles, gci)

      TYPE(cell_type), POINTER                           :: cell
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(section_vals_type), POINTER                   :: print_section
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_nfree', routineP = moduleN//':'//routineN

      INTEGER                                            :: natom, nconstraint_ext, nconstraint_int, &
                                                            nrestraints_int, rot_dof, &
                                                            roto_trasl_dof

! Retrieve information on number of atoms, constraints (external and internal)

      CALL get_molecule_kind_set(molecule_kind_set=molecule_kind_set, &
                                 natom=natom, nconstraint=nconstraint_int, nrestraints=nrestraints_int)

      ! Compute degrees of freedom
      CALL rot_ana(particles%els, dof=roto_trasl_dof, rot_dof=rot_dof, &
                   print_section=print_section, keep_rotations=.FALSE., &
                   mass_weighted=.TRUE., natoms=natom)

      roto_trasl_dof = roto_trasl_dof-MIN(SUM(cell%perd(1:3)), rot_dof)

      ! Saving this value of simpar preliminar to the real count of constraints..
      simpar%nfree_rot_transl = roto_trasl_dof

      ! compute the total number of degrees of freedom for temperature
      nconstraint_ext = gci%ntot-gci%nrestraint
      simpar%nfree = 3*natom-nconstraint_int-nconstraint_ext-roto_trasl_dof

   END SUBROUTINE compute_nfree

! **************************************************************************************************
!> \brief ...
!> \param thermostats ...
!> \param cell ...
!> \param simpar ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecules ...
!> \param particles ...
!> \param print_section ...
!> \param region_sections ...
!> \param gci ...
!> \param region ...
!> \param qmmm_env ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE compute_degrees_of_freedom(thermostats, cell, simpar, molecule_kind_set, &
                                         local_molecules, molecules, particles, print_section, region_sections, gci, &
                                         region, qmmm_env)

      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(cell_type), POINTER                           :: cell
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: print_section, region_sections
      TYPE(global_constraint_type), POINTER              :: gci
      INTEGER, INTENT(IN)                                :: region
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_degrees_of_freedom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw, natom, nconstraint_ext, &
                                                            nconstraint_int, nrestraints_int, &
                                                            rot_dof, roto_trasl_dof
      TYPE(cp_logger_type), POINTER                      :: logger

! Retrieve information on number of atoms, constraints (external and internal)

      CALL get_molecule_kind_set(molecule_kind_set=molecule_kind_set, &
                                 natom=natom, nconstraint=nconstraint_int, nrestraints=nrestraints_int)

      ! Compute degrees of freedom
      CALL rot_ana(particles%els, dof=roto_trasl_dof, rot_dof=rot_dof, &
                   print_section=print_section, keep_rotations=.FALSE., &
                   mass_weighted=.TRUE., natoms=natom)

      roto_trasl_dof = roto_trasl_dof-MIN(SUM(cell%perd(1:3)), rot_dof)

      ! Collect info about thermostats
      CALL setup_thermostat_info(thermostats%thermostat_info_part, molecule_kind_set, &
                                 local_molecules, molecules, particles, region, simpar%ensemble, roto_trasl_dof, &
                                 region_sections=region_sections, qmmm_env=qmmm_env)

      ! Saving this value of simpar preliminar to the real count of constraints..
      simpar%nfree_rot_transl = roto_trasl_dof

      ! compute the total number of degrees of freedom for temperature
      nconstraint_ext = gci%ntot-gci%nrestraint
      simpar%nfree = 3*natom-nconstraint_int-nconstraint_ext-roto_trasl_dof

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, print_section, "PROGRAM_RUN_INFO", &
                                extension=".log")
      IF (iw > 0) THEN
         WRITE (iw, '( /, A )') ' Calculation of degrees of freedom'
         WRITE (iw, '( T48, A, T71, I10 )') '       Number of atoms:', natom
         WRITE (iw, '( T33, A, T71, I10 )') &
            ' Number of Intramolecular constraints:', nconstraint_int
         WRITE (iw, '( T33, A, T71, I10 )') &
            ' Number of Intermolecular constraints:', nconstraint_ext
         WRITE (iw, '( T34, A, T71, I10 )') &
            ' Invariants(translation + rotations):', roto_trasl_dof
         WRITE (iw, '( T48, A, T71, I10 )') '    Degrees of freedom:', simpar%nfree
         WRITE (iw, '()')
         WRITE (iw, '( /, A )') ' Restraints Information'
         WRITE (iw, '( T33, A, T71, I10 )') &
            '  Number of Intramolecular restraints:', nrestraints_int
         WRITE (iw, '( T33, A, T71, I10 )') &
            '  Number of Intermolecular restraints:', gci%nrestraint
      END IF
      CALL cp_print_key_finished_output(iw, logger, print_section, &
                                        "PROGRAM_RUN_INFO")

   END SUBROUTINE compute_degrees_of_freedom

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecules ...
!> \param particles ...
!> \param region ...
!> \param ensemble ...
!> \param nfree ...
!> \param shell ...
!> \param region_sections ...
!> \param qmmm_env ...
!> \author 10.2011  CJM - PNNL
! **************************************************************************************************
   SUBROUTINE setup_adiabatic_thermostat_info(thermostat_info, molecule_kind_set, local_molecules, &
                                              molecules, particles, region, ensemble, nfree, shell, region_sections, qmmm_env)
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: region, ensemble
      INTEGER, INTENT(INOUT), OPTIONAL                   :: nfree
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell
      TYPE(section_vals_type), POINTER                   :: region_sections
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_adiabatic_thermostat_info', &
         routineP = moduleN//':'//routineN

      INTEGER :: dis_type, first_atom, i, ikind, imol, imol_global, ipart, itherm, katom, &
         last_atom, natom, natom_local, nkind, nmol_local, nmol_per_kind, nmolecule, nshell, &
         number, stat, sum_of_thermostats
      INTEGER, POINTER                                   :: molecule_list(:), thermolist(:)
      LOGICAL                                            :: check, do_shell, nointer, on_therm
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule, molecule_set(:)

      NULLIFY (molecule_kind, molecule, thermostat_info%map_loc_thermo_gen, thermolist)
      nkind = SIZE(molecule_kind_set)
      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      ! Counting the global number of thermostats
      sum_of_thermostats = 0
      ! Variable to denote independent thermostats (no communication necessary)
      nointer = .TRUE.
      check = .TRUE.
      number = 0
      dis_type = do_thermo_no_communication

      CALL get_adiabatic_region_info(region_sections, sum_of_thermostats, &
                                     thermolist=thermolist, &
                                     molecule_kind_set=molecule_kind_set, &
                                     molecules=molecules, particles=particles, qmmm_env=qmmm_env)

!    map_loc_thermo_gen=>thermostat_info%map_loc_thermo_gen
      molecule_set => molecules%els
      SELECT CASE (ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/thermostat_utils.F",282,'Unknown ensemble')
      CASE (nvt_adiabatic_ensemble)
         SELECT CASE (region)
         CASE (do_region_global)
            ! Global Thermostat
            nointer = .FALSE.
            sum_of_thermostats = 1
         CASE (do_region_molecule)
            ! Molecular Thermostat
            itherm = 0
            DO ikind = 1, nkind
               molecule_kind => molecule_kind_set(ikind)
               nmol_per_kind = local_molecules%n_el(ikind)
               CALL get_molecule_kind(molecule_kind, natom=natom, &
                                      molecule_list=molecule_list)
! use thermolist ( ipart ) to get global indexing correct
               DO imol_global = 1, SIZE(molecule_list)
                  molecule => molecule_set(molecule_list(imol_global))
                  CALL get_molecule(molecule, first_atom=first_atom, &
                                    last_atom=last_atom)
                  on_therm = .TRUE.
                  DO katom = first_atom, last_atom
                     IF (thermolist(katom) == HUGE(0)) THEN
                        on_therm = .FALSE.
                        EXIT
                     END IF
                  END DO
                  IF (on_therm) THEN
                     itherm = itherm+1
                     DO katom = first_atom, last_atom
                        thermolist(katom) = itherm
                     END DO
                  END IF
               END DO
            END DO
            DO i = 1, nkind
               molecule_kind => molecule_kind_set(i)
               CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, nshell=nshell)
               IF ((do_shell) .AND. (nshell == 0)) nmolecule = 0
               sum_of_thermostats = sum_of_thermostats+nmolecule
            END DO
            ! If we have ONE kind and ONE molecule, then effectively we have a GLOBAL thermostat
            ! and the degrees of freedom will be computed correctly for this special case
            IF ((nmolecule == 1) .AND. (nkind == 1)) nointer = .FALSE.
         CASE (do_region_massive)
            ! Massive Thermostat
            DO i = 1, nkind
               molecule_kind => molecule_kind_set(i)
               CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, &
                                      natom=natom, nshell=nshell)
               IF (do_shell) natom = nshell
               sum_of_thermostats = sum_of_thermostats+3*natom*nmolecule
            END DO
         END SELECT

         natom_local = 0
         DO ikind = 1, SIZE(molecule_kind_set)
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
               DO ipart = first_atom, last_atom
                  natom_local = natom_local+1
               END DO
            END DO
         END DO

         ! Now map the local atoms with the corresponding thermostat
         ALLOCATE (thermostat_info%map_loc_thermo_gen(natom_local), stat=stat)
         thermostat_info%map_loc_thermo_gen = HUGE(0)
         IF(.NOT.(stat == 0))CALL cp__a("motion/thermostat/thermostat_utils.F",353)
         natom_local = 0
         DO ikind = 1, SIZE(molecule_kind_set)
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
               DO ipart = first_atom, last_atom
                  natom_local = natom_local+1
! only map the correct region to the thermostat
                  IF (thermolist(ipart) /= HUGE(0)) &
                     thermostat_info%map_loc_thermo_gen(natom_local) = thermolist(ipart)
               END DO
            END DO
         END DO
         ! Here we decide which parallel algorithm to use.
         ! if there are only massive and molecule type thermostats we can use
         ! a local scheme, in cases involving any combination with a
         ! global thermostat we assume a coupling of  degrees of freedom
         ! from different processors
         IF (nointer) THEN
            ! Distributed thermostats, no interaction
            dis_type = do_thermo_no_communication
            ! we only count thermostats on this processor
            number = 0
            DO ikind = 1, nkind
               nmol_local = local_molecules%n_el(ikind)
               molecule_kind => molecule_kind_set(ikind)
               CALL get_molecule_kind(molecule_kind, natom=natom, nshell=nshell)
               IF (do_shell) THEN
                  natom = nshell
                  IF (nshell == 0) nmol_local = 0
               END IF
               IF (region == do_region_molecule) THEN
                  number = number+nmol_local
               ELSE IF (region == do_region_massive) THEN
                  number = number+3*nmol_local*natom
               ELSE
                  CALL cp__b("motion/thermostat/thermostat_utils.F",392,'Invalid region setup')
               END IF
            END DO
         ELSE
            ! REPlicated thermostats, INTERacting via communication
            dis_type = do_thermo_communication
            IF ((region == do_region_global) .OR. (region == do_region_molecule)) number = 1
         END IF

         IF (PRESENT(nfree)) THEN
            ! re-initializing simpar%nfree to zero because of multiple thermostats in the adiabatic sampling
            nfree = 0
         END IF
      END SELECT

      ! Saving information about thermostats
      thermostat_info%sum_of_thermostats = sum_of_thermostats
      thermostat_info%number_of_thermostats = number
      thermostat_info%dis_type = dis_type

      DEALLOCATE (thermolist)

   END SUBROUTINE setup_adiabatic_thermostat_info

! **************************************************************************************************
!> \brief ...
!> \param region_sections ...
!> \param sum_of_thermostats ...
!> \param thermolist ...
!> \param molecule_kind_set ...
!> \param molecules ...
!> \param particles ...
!> \param qmmm_env ...
!> \author 10.2011 CJM -PNNL
! **************************************************************************************************
   SUBROUTINE get_adiabatic_region_info(region_sections, sum_of_thermostats, &
                                        thermolist, molecule_kind_set, molecules, particles, &
                                        qmmm_env)
      TYPE(section_vals_type), POINTER                   :: region_sections
      INTEGER, INTENT(INOUT), OPTIONAL                   :: sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: thermolist(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_adiabatic_region_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: first_atom, i, ig, ikind, ilist, imol, &
                                                            ipart, itherm, jg, last_atom, &
                                                            mregions, n_rep, nregions, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(molecule_type), POINTER                       :: molecule

      NULLIFY (tmplist, tmpstringlist, thermolist, molecule_kind, molecule, molecule_set)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      ! IF(.NOT.(.NOT.(ASSOCIATED(map_loc_thermo_gen))))CALL cp__a("motion/thermostat/thermostat_utils.F",456)
      CALL section_vals_get(region_sections, n_repetition=nregions)
      ALLOCATE (thermolist(particles%n_els))
      thermolist = HUGE(0)
      molecule_set => molecules%els
      mregions = nregions
      itherm = 0
      DO ig = 1, mregions
         CALL section_vals_val_get(region_sections, "LIST", i_rep_section=ig, n_rep_val=n_rep)
         DO jg = 1, n_rep
            CALL section_vals_val_get(region_sections, "LIST", i_rep_section=ig, i_rep_val=jg, i_vals=tmplist)
            DO i = 1, SIZE(tmplist)
               ipart = tmplist(i)
               IF(.NOT.(((ipart > 0) .AND. (ipart <= particles%n_els))))CALL cp__a("motion/thermostat/thermostat_utils.F",469)
               IF (thermolist(ipart) == HUGE(0)) THEN
                  itherm = itherm+1
                  thermolist(ipart) = itherm
               ELSE
                  CALL cp__b("motion/thermostat/thermostat_utils.F",474,"")
               END IF
            END DO
         END DO
         CALL section_vals_val_get(region_sections, "MOLNAME", i_rep_section=ig, n_rep_val=n_rep)
         DO jg = 1, n_rep
            CALL section_vals_val_get(region_sections, "MOLNAME", i_rep_section=ig, i_rep_val=jg, c_vals=tmpstringlist)
            DO ilist = 1, SIZE(tmpstringlist)
               DO ikind = 1, SIZE(molecule_kind_set)
                  molecule_kind => molecule_kind_set(ikind)
                  IF (molecule_kind%name == tmpstringlist(ilist)) THEN
                     DO imol = 1, SIZE(molecule_kind%molecule_list)
                        molecule => molecule_set(molecule_kind%molecule_list(imol))
                        CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                        DO ipart = first_atom, last_atom
                           IF (thermolist(ipart) == HUGE(0)) THEN
                              itherm = itherm+1
                              thermolist(ipart) = itherm
                           ELSE
                              CALL cp__b("motion/thermostat/thermostat_utils.F",493,"")
                           END IF
                        END DO
                     END DO
                  END IF
               END DO
            END DO
         END DO
         CALL setup_thermostat_subsys(region_sections, qmmm_env, thermolist, molecule_set, &
                                      subsys_qm=.FALSE., ig=ig, sum_of_thermostats=sum_of_thermostats, nregions=nregions)
         CALL setup_thermostat_subsys(region_sections, qmmm_env, thermolist, molecule_set, &
                                      subsys_qm=.TRUE., ig=ig, sum_of_thermostats=sum_of_thermostats, nregions=nregions)
      END DO

      IF(.NOT.(.NOT. ALL(thermolist == HUGE(0))))CALL cp__a("motion/thermostat/thermostat_utils.F",507)

!    natom_local = 0
!    DO ikind = 1, SIZE(molecule_kind_set)
!       nmol_per_kind = local_molecules%n_el(ikind)
!       DO imol = 1, nmol_per_kind
!          i = local_molecules%list(ikind)%array(imol)
!          molecule => molecule_set(i)
!          CALL get_molecule ( molecule, first_atom = first_atom, last_atom = last_atom )
!          DO ipart = first_atom, last_atom
!             natom_local = natom_local + 1
!          END DO
!       END DO
!    END DO

      ! Now map the local atoms with the corresponding thermostat
!    ALLOCATE(map_loc_thermo_gen(natom_local),stat=stat)
!    map_loc_thermo_gen  = HUGE ( 0 )
!    CPPostcondition(stat==0,cp_failure_level,routineP,failure)
!    natom_local = 0
!    DO ikind = 1, SIZE(molecule_kind_set)
!       nmol_per_kind = local_molecules%n_el(ikind)
!       DO imol = 1, nmol_per_kind
!          i = local_molecules%list(ikind)%array(imol)
!          molecule => molecule_set(i)
!          CALL get_molecule ( molecule, first_atom = first_atom, last_atom = last_atom )
!          DO ipart = first_atom, last_atom
!             natom_local = natom_local + 1
! only map the correct region to the thermostat
!             IF ( thermolist (ipart ) /= HUGE ( 0 ) ) &
!             map_loc_thermo_gen(natom_local) = thermolist(ipart)
!          END DO
!       END DO
!    END DO

!    DEALLOCATE(thermolist, stat=stat)
!    CPPostcondition(stat==0,cp_failure_level,routineP,failure)
   END SUBROUTINE get_adiabatic_region_info
! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecules ...
!> \param particles ...
!> \param region ...
!> \param ensemble ...
!> \param nfree ...
!> \param shell ...
!> \param region_sections ...
!> \param qmmm_env ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_thermostat_info(thermostat_info, molecule_kind_set, local_molecules, &
                                    molecules, particles, region, ensemble, nfree, shell, region_sections, qmmm_env)
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: region, ensemble
      INTEGER, INTENT(INOUT), OPTIONAL                   :: nfree
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell
      TYPE(section_vals_type), POINTER                   :: region_sections
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_thermostat_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dis_type, i, ikind, natom, nkind, &
                                                            nmol_local, nmolecule, nshell, number, &
                                                            sum_of_thermostats
      LOGICAL                                            :: check, do_shell, nointer
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (molecule_kind)
      nkind = SIZE(molecule_kind_set)
      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      ! Counting the global number of thermostats
      sum_of_thermostats = 0
      ! Variable to denote independent thermostats (no communication necessary)
      nointer = .TRUE.
      check = .TRUE.
      number = 0
      dis_type = do_thermo_no_communication

      SELECT CASE (ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/thermostat_utils.F",596,'Unknown ensemble')
      CASE (isokin_ensemble, nph_uniaxial_ensemble, nph_uniaxial_damped_ensemble, &
            reftraj_ensemble, langevin_ensemble)
         ! Do Nothing
      CASE (nve_ensemble, nvt_ensemble, nvt_adiabatic_ensemble, npt_i_ensemble, &
            npt_f_ensemble, npe_i_ensemble, npe_f_ensemble)
         IF (ensemble == nve_ensemble) check = do_shell
         IF (check) THEN
            SELECT CASE (region)
            CASE (do_region_global)
               ! Global Thermostat
               nointer = .FALSE.
               sum_of_thermostats = 1
            CASE (do_region_molecule)
               ! Molecular Thermostat
               DO i = 1, nkind
                  molecule_kind => molecule_kind_set(i)
                  CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, nshell=nshell)
                  IF ((do_shell) .AND. (nshell == 0)) nmolecule = 0
                  sum_of_thermostats = sum_of_thermostats+nmolecule
               END DO
               ! If we have ONE kind and ONE molecule, then effectively we have a GLOBAL thermostat
               ! and the degrees of freedom will be computed correctly for this special case
               IF ((nmolecule == 1) .AND. (nkind == 1)) nointer = .FALSE.
            CASE (do_region_massive)
               ! Massive Thermostat
               DO i = 1, nkind
                  molecule_kind => molecule_kind_set(i)
                  CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, &
                                         natom=natom, nshell=nshell)
                  IF (do_shell) natom = nshell
                  sum_of_thermostats = sum_of_thermostats+3*natom*nmolecule
               END DO
            CASE (do_region_defined)
               ! User defined region to thermostat..
               nointer = .FALSE.
               ! Determine the number of thermostats defined in the input
               CALL section_vals_get(region_sections, n_repetition=sum_of_thermostats)
               IF (sum_of_thermostats < 1) &
                  CALL cp_abort(cp__l("motion/thermostat/thermostat_utils.F",635), &
                                "Provide at least 1 region (&DEFINE_REGION) when using the thermostat type DEFINED")
            END SELECT

            ! Here we decide which parallel algorithm to use.
            ! if there are only massive and molecule type thermostats we can use
            ! a local scheme, in cases involving any combination with a
            ! global thermostat we assume a coupling of  degrees of freedom
            ! from different processors
            IF (nointer) THEN
               ! Distributed thermostats, no interaction
               dis_type = do_thermo_no_communication
               ! we only count thermostats on this processor
               number = 0
               DO ikind = 1, nkind
                  nmol_local = local_molecules%n_el(ikind)
                  molecule_kind => molecule_kind_set(ikind)
                  CALL get_molecule_kind(molecule_kind, natom=natom, nshell=nshell)
                  IF (do_shell) THEN
                     natom = nshell
                     IF (nshell == 0) nmol_local = 0
                  END IF
                  IF (region == do_region_molecule) THEN
                     number = number+nmol_local
                  ELSE IF (region == do_region_massive) THEN
                     number = number+3*nmol_local*natom
                  ELSE
                     CALL cp__b("motion/thermostat/thermostat_utils.F",662,'Invalid region setup')
                  END IF
               END DO
            ELSE
               ! REPlicated thermostats, INTERacting via communication
               dis_type = do_thermo_communication
               IF ((region == do_region_global) .OR. (region == do_region_molecule)) THEN
                  number = 1
               ELSE IF (region == do_region_defined) THEN
                  CALL get_defined_region_info(region_sections, number, sum_of_thermostats, &
                                               map_loc_thermo_gen=thermostat_info%map_loc_thermo_gen, &
                                               local_molecules=local_molecules, molecule_kind_set=molecule_kind_set, &
                                               molecules=molecules, particles=particles, qmmm_env=qmmm_env)
               END IF
            END IF

            IF (PRESENT(nfree)) THEN
               IF ((sum_of_thermostats > 1) .OR. (dis_type == do_thermo_no_communication)) THEN
                  ! re-initializing simpar%nfree to zero because of multiple thermostats
                  nfree = 0
               END IF
            END IF
         END IF
      END SELECT

      ! Saving information about thermostats
      thermostat_info%sum_of_thermostats = sum_of_thermostats
      thermostat_info%number_of_thermostats = number
      thermostat_info%dis_type = dis_type
   END SUBROUTINE setup_thermostat_info

! **************************************************************************************************
!> \brief ...
!> \param region_sections ...
!> \param number ...
!> \param sum_of_thermostats ...
!> \param map_loc_thermo_gen ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecules ...
!> \param particles ...
!> \param qmmm_env ...
!> \author 11.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_defined_region_info(region_sections, number, sum_of_thermostats, &
                                      map_loc_thermo_gen, local_molecules, molecule_kind_set, molecules, particles, &
                                      qmmm_env)
      TYPE(section_vals_type), POINTER                   :: region_sections
      INTEGER, INTENT(OUT), OPTIONAL                     :: number
      INTEGER, INTENT(INOUT), OPTIONAL                   :: sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: map_loc_thermo_gen
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_defined_region_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER :: first_atom, i, ig, ikind, ilist, imol, ipart, jg, last_atom, mregions, n_rep, &
         natom_local, nmol_per_kind, nregions, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: thermolist, tmp, tmplist
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(molecule_type), POINTER                       :: molecule

      NULLIFY (tmplist, tmpstringlist, thermolist, molecule_kind, molecule, molecule_set)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      IF(.NOT.(.NOT. (ASSOCIATED(map_loc_thermo_gen))))CALL cp__a("motion/thermostat/thermostat_utils.F",736)
      CALL section_vals_get(region_sections, n_repetition=nregions)
      ALLOCATE (thermolist(particles%n_els))
      thermolist = HUGE(0)
      molecule_set => molecules%els
      mregions = nregions
      DO ig = 1, mregions
         CALL section_vals_val_get(region_sections, "LIST", i_rep_section=ig, n_rep_val=n_rep)
         DO jg = 1, n_rep
            CALL section_vals_val_get(region_sections, "LIST", i_rep_section=ig, i_rep_val=jg, i_vals=tmplist)
            DO i = 1, SIZE(tmplist)
               ipart = tmplist(i)
               IF(.NOT.(((ipart > 0) .AND. (ipart <= particles%n_els))))CALL cp__a("motion/thermostat/thermostat_utils.F",748)
               IF (thermolist(ipart) == HUGE(0)) THEN
                  thermolist(ipart) = ig
               ELSE
                  CALL cp__b("motion/thermostat/thermostat_utils.F",752,"")
               END IF
            END DO
         END DO
         CALL section_vals_val_get(region_sections, "MOLNAME", i_rep_section=ig, n_rep_val=n_rep)
         DO jg = 1, n_rep
            CALL section_vals_val_get(region_sections, "MOLNAME", i_rep_section=ig, i_rep_val=jg, c_vals=tmpstringlist)
            DO ilist = 1, SIZE(tmpstringlist)
               DO ikind = 1, SIZE(molecule_kind_set)
                  molecule_kind => molecule_kind_set(ikind)
                  IF (molecule_kind%name == tmpstringlist(ilist)) THEN
                     DO imol = 1, SIZE(molecule_kind%molecule_list)
                        molecule => molecule_set(molecule_kind%molecule_list(imol))
                        CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                        DO ipart = first_atom, last_atom
                           IF (thermolist(ipart) == HUGE(0)) THEN
                              thermolist(ipart) = ig
                           ELSE
                              CALL cp__b("motion/thermostat/thermostat_utils.F",770,"")
                           END IF
                        END DO
                     END DO
                  END IF
               END DO
            END DO
         END DO
         CALL setup_thermostat_subsys(region_sections, qmmm_env, thermolist, molecule_set, &
                                      subsys_qm=.FALSE., ig=ig, sum_of_thermostats=sum_of_thermostats, nregions=nregions)
         CALL setup_thermostat_subsys(region_sections, qmmm_env, thermolist, molecule_set, &
                                      subsys_qm=.TRUE., ig=ig, sum_of_thermostats=sum_of_thermostats, nregions=nregions)
      END DO

      ! Dump IO warning for not thermalized particles
      IF (ANY(thermolist == HUGE(0))) THEN
         nregions = nregions+1
         sum_of_thermostats = sum_of_thermostats+1
         ALLOCATE (tmp(COUNT(thermolist == HUGE(0))))
         ilist = 0
         DO i = 1, SIZE(thermolist)
            IF (thermolist(i) == HUGE(0)) THEN
               ilist = ilist+1
               tmp(ilist) = i
               thermolist(i) = nregions
            END IF
         END DO
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(A)') " WARNING| No thermostats defined for the following atoms:"
            IF (ilist > 0) WRITE (output_unit, '(" WARNING|",9I8)') tmp
            WRITE (output_unit, '(A)') " WARNING| They will be included in a further unique thermostat!"
         END IF
         DEALLOCATE (tmp)
      END IF
      IF(.NOT.(ALL(thermolist /= HUGE(0))))CALL cp__a("motion/thermostat/thermostat_utils.F",804)

      ! Now identify the local number of thermostats
      ALLOCATE (tmp(nregions))
      tmp = 0
      natom_local = 0
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               natom_local = natom_local+1
               tmp(thermolist(ipart)) = 1
            END DO
         END DO
      END DO
      number = SUM(tmp)
      DEALLOCATE (tmp)

      ! Now map the local atoms with the corresponding thermostat
      ALLOCATE (map_loc_thermo_gen(natom_local))
      natom_local = 0
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               natom_local = natom_local+1
               map_loc_thermo_gen(natom_local) = thermolist(ipart)
            END DO
         END DO
      END DO

      DEALLOCATE (thermolist)
   END SUBROUTINE get_defined_region_info

! **************************************************************************************************
!> \brief ...
!> \param region_sections ...
!> \param qmmm_env ...
!> \param thermolist ...
!> \param molecule_set ...
!> \param subsys_qm ...
!> \param ig ...
!> \param sum_of_thermostats ...
!> \param nregions ...
!> \author 11.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_thermostat_subsys(region_sections, qmmm_env, thermolist, &
                                      molecule_set, subsys_qm, ig, sum_of_thermostats, nregions)
      TYPE(section_vals_type), POINTER                   :: region_sections
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      INTEGER, DIMENSION(:), POINTER                     :: thermolist
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      LOGICAL, INTENT(IN)                                :: subsys_qm
      INTEGER, INTENT(IN)                                :: ig
      INTEGER, INTENT(INOUT)                             :: sum_of_thermostats, nregions

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_thermostat_subsys', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: label1, label2
      INTEGER                                            :: first_atom, i, imolecule, ipart, &
                                                            last_atom, nrep, thermo1
      INTEGER, DIMENSION(:), POINTER                     :: atom_index1
      LOGICAL                                            :: explicit
      TYPE(molecule_type), POINTER                       :: molecule

      label1 = "MM_SUBSYS"
      label2 = "QM_SUBSYS"
      IF (subsys_qm) THEN
         label1 = "QM_SUBSYS"
         label2 = "MM_SUBSYS"
      END IF
      CALL section_vals_val_get(region_sections, TRIM(label1), i_rep_section=ig, &
                                n_rep_val=nrep, explicit=explicit)
      IF (nrep == 1 .AND. explicit) THEN
         IF (ASSOCIATED(qmmm_env)) THEN
            atom_index1 => qmmm_env%qm%mm_atom_index
            IF (subsys_qm) THEN
               atom_index1 => qmmm_env%qm%qm_atom_index
            END IF
            CALL section_vals_val_get(region_sections, TRIM(label1), i_val=thermo1, i_rep_section=ig)
            SELECT CASE (thermo1)
            CASE (do_constr_atomic)
               DO i = 1, SIZE(atom_index1)
                  ipart = atom_index1(i)
                  IF (subsys_qm .AND. qmmm_env%qm%qmmm_link .AND. ASSOCIATED(qmmm_env%qm%mm_link_atoms)) THEN
                     IF (ANY(ipart == qmmm_env%qm%mm_link_atoms)) CYCLE
                  END IF
                  IF (thermolist(ipart) == HUGE(0)) THEN
                     thermolist(ipart) = ig
                  ELSE
                     CALL cp_abort(cp__l("motion/thermostat/thermostat_utils.F",901), &
                                   'One atom ('//cp_to_string(ipart)//') of the '// &
                                   TRIM(label1)//' was already assigned to'// &
                                   ' the thermostatting region Nr.'//cp_to_string(thermolist(ipart))// &
                                   '. Please check the input for inconsistencies!')
                  END IF
               END DO
            CASE (do_constr_molec)
               DO imolecule = 1, SIZE(molecule_set)
                  molecule => molecule_set(imolecule)
                  CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
                  IF (ANY(atom_index1 >= first_atom .AND. atom_index1 <= last_atom)) THEN
                     DO ipart = first_atom, last_atom
                        IF (thermolist(ipart) == HUGE(0)) THEN
                           thermolist(ipart) = ig
                        ELSE
                           CALL cp_abort(cp__l("motion/thermostat/thermostat_utils.F",917), &
                                         'One atom ('//cp_to_string(ipart)//') of the '// &
                                         TRIM(label1)//' was already assigned to'// &
                                         ' the thermostatting region Nr.'//cp_to_string(thermolist(ipart))// &
                                         '. Please check the input for inconsistencies!')
                        END IF
                     END DO
                  END IF
               END DO
            END SELECT
         ELSE
            sum_of_thermostats = sum_of_thermostats-1
            nregions = nregions-1
         END IF
      END IF
   END SUBROUTINE setup_thermostat_subsys

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param npt ...
!> \param group ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE ke_region_baro(map_info, npt, group)
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(len=*), PARAMETER :: routineN = 'ke_region_baro', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, ncoef

      map_info%v_scale = 1.0_dp
      map_info%s_kin = 0.0_dp
      ncoef = 0
      DO i = 1, SIZE(npt, 1)
         DO j = 1, SIZE(npt, 2)
            ncoef = ncoef+1
            map_info%p_kin(1, ncoef)%point = map_info%p_kin(1, ncoef)%point &
                                             +npt(i, j)%mass*npt(i, j)%v**2
         END DO
      END DO

      IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, group)

   END SUBROUTINE ke_region_baro

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param npt ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE vel_rescale_baro(map_info, npt)
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt

      CHARACTER(len=*), PARAMETER :: routineN = 'vel_rescale_baro', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, ncoef

      ncoef = 0
      DO i = 1, SIZE(npt, 1)
         DO j = 1, SIZE(npt, 2)
            ncoef = ncoef+1
            npt(i, j)%v = npt(i, j)%v*map_info%p_scale(1, ncoef)%point
         END DO
      END DO

   END SUBROUTINE vel_rescale_baro

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param group ...
!> \param vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE ke_region_particles(map_info, particle_set, molecule_kind_set, &
                                  local_molecules, molecule_set, group, vel)

      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(IN)                                :: group
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'ke_region_particles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ii, ikind, imol, imol_local, &
                                                            ipart, last_atom, nmol_local
      LOGICAL                                            :: present_vel
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule

      map_info%v_scale = 1.0_dp
      map_info%s_kin = 0.0_dp
      present_vel = PRESENT(vel)
      ii = 0
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               ii = ii+1
               atomic_kind => particle_set(ipart)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
               IF (present_vel) THEN
                  IF (ASSOCIATED(map_info%p_kin(1, ii)%point)) &
                     map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+mass*vel(1, ipart)**2
                  IF (ASSOCIATED(map_info%p_kin(2, ii)%point)) &
                     map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+mass*vel(2, ipart)**2
                  IF (ASSOCIATED(map_info%p_kin(3, ii)%point)) &
                     map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+mass*vel(3, ipart)**2
               ELSE
                  IF (ASSOCIATED(map_info%p_kin(1, ii)%point)) &
                     map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+mass*particle_set(ipart)%v(1)**2
                  IF (ASSOCIATED(map_info%p_kin(2, ii)%point)) &
                     map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+mass*particle_set(ipart)%v(2)**2
                  IF (ASSOCIATED(map_info%p_kin(3, ii)%point)) &
                     map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+mass*particle_set(ipart)%v(3)**2
               END IF
            END DO
         END DO
      END DO

      IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, group)

   END SUBROUTINE ke_region_particles

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param group ...
!> \param vel ...
!> \author 07.2009 MI
! **************************************************************************************************
   SUBROUTINE momentum_region_particles(map_info, particle_set, molecule_kind_set, &
                                        local_molecules, molecule_set, group, vel)

      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(IN)                                :: group
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'momentum_region_particles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ii, ikind, imol, imol_local, &
                                                            ipart, last_atom, nmol_local
      LOGICAL                                            :: present_vel
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule

      map_info%v_scale = 1.0_dp
      map_info%s_kin = 0.0_dp
      present_vel = PRESENT(vel)
      ii = 0
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               ii = ii+1
               atomic_kind => particle_set(ipart)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
               IF (present_vel) THEN
                  map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+SQRT(mass)*vel(1, ipart)
                  map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+SQRT(mass)*vel(2, ipart)
                  map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+SQRT(mass)*vel(3, ipart)
               ELSE
                  map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+SQRT(mass)*particle_set(ipart)%v(1)
                  map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+SQRT(mass)*particle_set(ipart)%v(2)
                  map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+SQRT(mass)*particle_set(ipart)%v(3)
               END IF
            END DO
         END DO
      END DO

      IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, group)

   END SUBROUTINE momentum_region_particles

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param shell_adiabatic ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE vel_rescale_particles(map_info, molecule_kind_set, molecule_set, &
                                    particle_set, local_molecules, shell_adiabatic, shell_particle_set, &
                                    core_particle_set, vel, shell_vel, core_vel)

      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'vel_rescale_particles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ii, ikind, imol, imol_local, &
                                                            ipart, jj, last_atom, nmol_local, &
                                                            shell_index
      LOGICAL                                            :: present_vel
      REAL(KIND=dp)                                      :: fac_massc, fac_masss, mass, vc(3), vs(3)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(shell_kind_type), POINTER                     :: shell

      ii = 0
      jj = 0
      present_vel = PRESENT(vel)
      ! Just few checks for consistency
      IF (present_vel) THEN
         IF (shell_adiabatic) THEN
            IF(.NOT.(PRESENT(shell_vel)))CALL cp__a("motion/thermostat/thermostat_utils.F",1172)
            IF(.NOT.(PRESENT(core_vel)))CALL cp__a("motion/thermostat/thermostat_utils.F",1173)
         END IF
      ELSE
         IF (shell_adiabatic) THEN
            IF(.NOT.(PRESENT(shell_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1177)
            IF(.NOT.(PRESENT(core_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1178)
         END IF
      END IF
      Kind: DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         Mol_local: DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            Particle: DO ipart = first_atom, last_atom
               ii = ii+1
               IF (present_vel) THEN
                  vel(1, ipart) = vel(1, ipart)*map_info%p_scale(1, ii)%point
                  vel(2, ipart) = vel(2, ipart)*map_info%p_scale(2, ii)%point
                  vel(3, ipart) = vel(3, ipart)*map_info%p_scale(3, ii)%point
               ELSE
                  particle_set(ipart)%v(1) = particle_set(ipart)%v(1)*map_info%p_scale(1, ii)%point
                  particle_set(ipart)%v(2) = particle_set(ipart)%v(2)*map_info%p_scale(2, ii)%point
                  particle_set(ipart)%v(3) = particle_set(ipart)%v(3)*map_info%p_scale(3, ii)%point
               END IF
               ! If Shell Adiabatic then apply the NHC thermostat also to the Shells
               IF (shell_adiabatic) THEN
                  shell_index = particle_set(ipart)%shell_index
                  IF (shell_index /= 0) THEN
                     jj = jj+2
                     atomic_kind => particle_set(ipart)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, shell=shell)
                     fac_masss = shell%mass_shell/mass
                     fac_massc = shell%mass_core/mass
                     IF (present_vel) THEN
                        vs(1:3) = shell_vel(1:3, shell_index)
                        vc(1:3) = core_vel(1:3, shell_index)
                        shell_vel(1, shell_index) = vel(1, ipart)+fac_massc*(vs(1)-vc(1))
                        shell_vel(2, shell_index) = vel(2, ipart)+fac_massc*(vs(2)-vc(2))
                        shell_vel(3, shell_index) = vel(3, ipart)+fac_massc*(vs(3)-vc(3))
                        core_vel(1, shell_index) = vel(1, ipart)+fac_masss*(vc(1)-vs(1))
                        core_vel(2, shell_index) = vel(2, ipart)+fac_masss*(vc(2)-vs(2))
                        core_vel(3, shell_index) = vel(3, ipart)+fac_masss*(vc(3)-vs(3))
                     ELSE
                        vs(1:3) = shell_particle_set(shell_index)%v(1:3)
                        vc(1:3) = core_particle_set(shell_index)%v(1:3)
                        shell_particle_set(shell_index)%v(1) = particle_set(ipart)%v(1)+fac_massc*(vs(1)-vc(1))
                        shell_particle_set(shell_index)%v(2) = particle_set(ipart)%v(2)+fac_massc*(vs(2)-vc(2))
                        shell_particle_set(shell_index)%v(3) = particle_set(ipart)%v(3)+fac_massc*(vs(3)-vc(3))
                        core_particle_set(shell_index)%v(1) = particle_set(ipart)%v(1)+fac_masss*(vc(1)-vs(1))
                        core_particle_set(shell_index)%v(2) = particle_set(ipart)%v(2)+fac_masss*(vc(2)-vs(2))
                        core_particle_set(shell_index)%v(3) = particle_set(ipart)%v(3)+fac_masss*(vc(3)-vs(3))
                     END IF
                  END IF
               END IF
            END DO Particle
         END DO Mol_local
      END DO Kind

   END SUBROUTINE vel_rescale_particles

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param group ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param core_vel ...
!> \param shell_vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE ke_region_shells(map_info, particle_set, atomic_kind_set, &
                               local_particles, group, core_particle_set, shell_particle_set, &
                               core_vel, shell_vel)

      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      INTEGER, INTENT(IN)                                :: group
      TYPE(particle_type), OPTIONAL, POINTER             :: core_particle_set(:), &
                                                            shell_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: core_vel(:, :), shell_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'ke_region_shells', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local, shell_index
      LOGICAL                                            :: is_shell, present_vel
      REAL(dp)                                           :: mass, mu_mass, v_sc(3)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      present_vel = PRESENT(shell_vel)
      ! Preliminary checks for consistency usage
      IF (present_vel) THEN
         IF(.NOT.(PRESENT(core_vel)))CALL cp__a("motion/thermostat/thermostat_utils.F",1274)
      ELSE
         IF(.NOT.(PRESENT(shell_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1276)
         IF(.NOT.(PRESENT(core_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1277)
      END IF
      ! get force on first thermostat for all the chains in the system.
      map_info%v_scale = 1.0_dp
      map_info%s_kin = 0.0_dp
      ii = 0

      nparticle_kind = SIZE(atomic_kind_set)
      DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, shell_active=is_shell, shell=shell)
         IF (is_shell) THEN
            mu_mass = shell%mass_shell*shell%mass_core/mass
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               shell_index = particle_set(iparticle)%shell_index
               ii = ii+1
               IF (present_vel) THEN
                  v_sc(1) = core_vel(1, shell_index)-shell_vel(1, shell_index)
                  v_sc(2) = core_vel(2, shell_index)-shell_vel(2, shell_index)
                  v_sc(3) = core_vel(3, shell_index)-shell_vel(3, shell_index)
                  map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+mu_mass*v_sc(1)**2
                  map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+mu_mass*v_sc(2)**2
                  map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+mu_mass*v_sc(3)**2
               ELSE
                  v_sc(1) = core_particle_set(shell_index)%v(1)-shell_particle_set(shell_index)%v(1)
                  v_sc(2) = core_particle_set(shell_index)%v(2)-shell_particle_set(shell_index)%v(2)
                  v_sc(3) = core_particle_set(shell_index)%v(3)-shell_particle_set(shell_index)%v(3)
                  map_info%p_kin(1, ii)%point = map_info%p_kin(1, ii)%point+mu_mass*v_sc(1)**2
                  map_info%p_kin(2, ii)%point = map_info%p_kin(2, ii)%point+mu_mass*v_sc(2)**2
                  map_info%p_kin(3, ii)%point = map_info%p_kin(3, ii)%point+mu_mass*v_sc(3)**2
               END IF
            END DO
         END IF
      END DO
      IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, group)

   END SUBROUTINE ke_region_shells

! **************************************************************************************************
!> \brief ...
!> \param map_info ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param shell_vel ...
!> \param core_vel ...
!> \param vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE vel_rescale_shells(map_info, atomic_kind_set, particle_set, local_particles, &
                                 shell_particle_set, core_particle_set, shell_vel, core_vel, vel)

      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: shell_vel(:, :), core_vel(:, :), &
                                                            vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'vel_rescale_shells', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local, shell_index
      LOGICAL                                            :: is_shell, present_vel
      REAL(dp)                                           :: mass, massc, masss, umass, v(3), vc(3), &
                                                            vs(3)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      present_vel = PRESENT(vel)
      ! Preliminary checks for consistency usage
      IF (present_vel) THEN
         IF(.NOT.(PRESENT(shell_vel)))CALL cp__a("motion/thermostat/thermostat_utils.F",1357)
         IF(.NOT.(PRESENT(core_vel)))CALL cp__a("motion/thermostat/thermostat_utils.F",1358)
      ELSE
         IF(.NOT.(PRESENT(shell_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1360)
         IF(.NOT.(PRESENT(core_particle_set)))CALL cp__a("motion/thermostat/thermostat_utils.F",1361)
      END IF
      ii = 0
      nparticle_kind = SIZE(atomic_kind_set)
      ! now scale the core-shell velocities
      Kind: DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, shell_active=is_shell, shell=shell)
         IF (is_shell) THEN
            umass = 1.0_dp/mass
            masss = shell%mass_shell*umass
            massc = shell%mass_core*umass

            nparticle_local = local_particles%n_el(iparticle_kind)
            Particles: DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               shell_index = particle_set(iparticle)%shell_index
               ii = ii+1
               IF (present_vel) THEN
                  vc(1:3) = core_vel(1:3, shell_index)
                  vs(1:3) = shell_vel(1:3, shell_index)
                  v(1:3) = vel(1:3, iparticle)
                  shell_vel(1, shell_index) = v(1)+map_info%p_scale(1, ii)%point*massc*(vs(1)-vc(1))
                  shell_vel(2, shell_index) = v(2)+map_info%p_scale(2, ii)%point*massc*(vs(2)-vc(2))
                  shell_vel(3, shell_index) = v(3)+map_info%p_scale(3, ii)%point*massc*(vs(3)-vc(3))
                  core_vel(1, shell_index) = v(1)+map_info%p_scale(1, ii)%point*masss*(vc(1)-vs(1))
                  core_vel(2, shell_index) = v(2)+map_info%p_scale(2, ii)%point*masss*(vc(2)-vs(2))
                  core_vel(3, shell_index) = v(3)+map_info%p_scale(3, ii)%point*masss*(vc(3)-vs(3))
               ELSE
                  vc(1:3) = core_particle_set(shell_index)%v(1:3)
                  vs(1:3) = shell_particle_set(shell_index)%v(1:3)
                  v(1:3) = particle_set(iparticle)%v(1:3)
                  shell_particle_set(shell_index)%v(1) = v(1)+map_info%p_scale(1, ii)%point*massc*(vs(1)-vc(1))
                  shell_particle_set(shell_index)%v(2) = v(2)+map_info%p_scale(2, ii)%point*massc*(vs(2)-vc(2))
                  shell_particle_set(shell_index)%v(3) = v(3)+map_info%p_scale(3, ii)%point*massc*(vs(3)-vc(3))
                  core_particle_set(shell_index)%v(1) = v(1)+map_info%p_scale(1, ii)%point*masss*(vc(1)-vs(1))
                  core_particle_set(shell_index)%v(2) = v(2)+map_info%p_scale(2, ii)%point*masss*(vc(2)-vs(2))
                  core_particle_set(shell_index)%v(3) = v(3)+map_info%p_scale(3, ii)%point*masss*(vc(3)-vs(3))
               END IF
            END DO Particles
         END IF
      END DO Kind

   END SUBROUTINE vel_rescale_shells

! **************************************************************************************************
!> \brief Calculates kinetic energy and potential energy of the nhc variables
!> \param nhc ...
!> \param nhc_pot ...
!> \param nhc_kin ...
!> \param para_env ...
!> \param array_kin ...
!> \param array_pot ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_nhc_energies(nhc, nhc_pot, nhc_kin, para_env, array_kin, array_pot)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      REAL(KIND=dp), INTENT(OUT)                         :: nhc_pot, nhc_kin
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: array_kin, array_pot

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_nhc_energies', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imap, l, n, number
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: akin, vpot

      number = nhc%glob_num_nhc
      ALLOCATE (akin(number))
      ALLOCATE (vpot(number))
      akin = 0.0_dp
      vpot = 0.0_dp
      DO n = 1, nhc%loc_num_nhc
         imap = nhc%map_info%index(n)
         DO l = 1, nhc%nhc_len
            akin(imap) = akin(imap)+0.5_dp*nhc%nvt(l, n)%mass*nhc%nvt(l, n)%v**2
            vpot(imap) = vpot(imap)+nhc%nvt(l, n)%nkt*nhc%nvt(l, n)%eta
         END DO
      END DO

      ! Handle the thermostat distribution
      IF (nhc%map_info%dis_type == do_thermo_no_communication) THEN
         CALL mp_sum(akin, para_env%group)
         CALL mp_sum(vpot, para_env%group)
      ELSE IF (nhc%map_info%dis_type == do_thermo_communication) THEN
         CALL communication_thermo_low1(akin, number, para_env)
         CALL communication_thermo_low1(vpot, number, para_env)
      END IF
      nhc_kin = SUM(akin)
      nhc_pot = SUM(vpot)

      ! Possibly give back kinetic or potential energy arrays
      IF (PRESENT(array_pot)) THEN
         IF (ASSOCIATED(array_pot)) THEN
            IF(.NOT.(SIZE(array_pot) == number))CALL cp__a("motion/thermostat/thermostat_utils.F",1457)
         ELSE
            ALLOCATE (array_pot(number))
         ENDIF
         array_pot = vpot
      END IF
      IF (PRESENT(array_kin)) THEN
         IF (ASSOCIATED(array_kin)) THEN
            IF(.NOT.(SIZE(array_kin) == number))CALL cp__a("motion/thermostat/thermostat_utils.F",1465)
         ELSE
            ALLOCATE (array_kin(number))
         ENDIF
         array_kin = akin
      END IF
      DEALLOCATE (akin)
      DEALLOCATE (vpot)
   END SUBROUTINE get_nhc_energies

! **************************************************************************************************
!> \brief Calculates kinetic energy and potential energy
!>      of the csvr  and gle thermostats
!> \param map_info ...
!> \param loc_num ...
!> \param glob_num ...
!> \param thermo_energy ...
!> \param thermostat_kin ...
!> \param para_env ...
!> \param array_pot ...
!> \param array_kin ...
!> \par History generalized MI [07.2009]
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_kin_energies(map_info, loc_num, glob_num, thermo_energy, thermostat_kin, &
                               para_env, array_pot, array_kin)

      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER, INTENT(IN)                                :: loc_num, glob_num
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: thermo_energy
      REAL(KIND=dp), INTENT(OUT)                         :: thermostat_kin
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: array_pot, array_kin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_kin_energies', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imap, n, number
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: akin

      number = glob_num
      ALLOCATE (akin(number))
      akin = 0.0_dp
      DO n = 1, loc_num
         imap = map_info%index(n)
         akin(imap) = thermo_energy(n)
      END DO

      ! Handle the thermostat distribution
      IF (map_info%dis_type == do_thermo_no_communication) THEN
         CALL mp_sum(akin, para_env%group)
      ELSE IF (map_info%dis_type == do_thermo_communication) THEN
         CALL communication_thermo_low1(akin, number, para_env)
      END IF
      thermostat_kin = SUM(akin)

      ! Possibly give back kinetic or potential energy arrays
      IF (PRESENT(array_pot)) THEN
         IF (ASSOCIATED(array_pot)) THEN
            IF(.NOT.(SIZE(array_pot) == number))CALL cp__a("motion/thermostat/thermostat_utils.F",1524)
         ELSE
            ALLOCATE (array_pot(number))
         ENDIF
         array_pot = 0.0_dp
      END IF
      IF (PRESENT(array_kin)) THEN
         IF (ASSOCIATED(array_kin)) THEN
            IF(.NOT.(SIZE(array_kin) == number))CALL cp__a("motion/thermostat/thermostat_utils.F",1532)
         ELSE
            ALLOCATE (array_kin(number))
         ENDIF
         array_kin = akin
      END IF
      DEALLOCATE (akin)
   END SUBROUTINE get_kin_energies

! **************************************************************************************************
!> \brief Calculates the temperatures of the regions when a thermostat is
!>        applied
!> \param map_info ...
!> \param loc_num ...
!> \param glob_num ...
!> \param nkt ...
!> \param dof ...
!> \param para_env ...
!> \param temp_tot ...
!> \param array_temp ...
!> \par History generalized MI [07.2009]
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_temperatures(map_info, loc_num, glob_num, nkt, dof, para_env, &
                               temp_tot, array_temp)
      TYPE(map_info_type), POINTER                       :: map_info
      INTEGER, INTENT(IN)                                :: loc_num, glob_num
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: nkt, dof
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(OUT)                         :: temp_tot
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: array_temp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_temperatures', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, imap2, n, number
      REAL(KIND=dp)                                      :: fdeg_of_free
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: akin, deg_of_free

      number = glob_num
      ALLOCATE (akin(number))
      ALLOCATE (deg_of_free(number))
      akin = 0.0_dp
      deg_of_free = 0.0_dp
      DO n = 1, loc_num
         imap = map_info%index(n)
         imap2 = map_info%map_index(n)
         IF (nkt(n) == 0.0_dp) CYCLE
         deg_of_free(imap) = REAL(dof(n), KIND=dp)
         akin(imap) = map_info%s_kin(imap2)
      END DO

      ! Handle the thermostat distribution
      IF (map_info%dis_type == do_thermo_no_communication) THEN
         CALL mp_sum(akin, para_env%group)
         CALL mp_sum(deg_of_free, para_env%group)
      ELSE IF (map_info%dis_type == do_thermo_communication) THEN
         CALL communication_thermo_low1(akin, number, para_env)
         CALL communication_thermo_low1(deg_of_free, number, para_env)
      END IF
      temp_tot = SUM(akin)
      fdeg_of_free = SUM(deg_of_free)

      temp_tot = temp_tot/fdeg_of_free
      temp_tot = cp_unit_from_cp2k(temp_tot, "K_temp")
      ! Possibly give back temperatures of the full set of regions
      IF (PRESENT(array_temp)) THEN
         IF (ASSOCIATED(array_temp)) THEN
            IF(.NOT.(SIZE(array_temp) == number))CALL cp__a("motion/thermostat/thermostat_utils.F",1600)
         ELSE
            ALLOCATE (array_temp(number))
         ENDIF
         DO i = 1, number
            array_temp(i) = akin(i)/deg_of_free(i)
            array_temp(i) = cp_unit_from_cp2k(array_temp(i), "K_temp")
         END DO
      END IF
      DEALLOCATE (akin)
      DEALLOCATE (deg_of_free)
   END SUBROUTINE get_temperatures

! **************************************************************************************************
!> \brief Calculates energy associated with a thermostat
!> \param thermostat ...
!> \param thermostat_pot ...
!> \param thermostat_kin ...
!> \param para_env ...
!> \param array_pot ...
!> \param array_kin ...
!> \author Teodoro Laino [tlaino] - 10.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_thermostat_energies(thermostat, thermostat_pot, thermostat_kin, para_env, &
                                      array_pot, array_kin)
      TYPE(thermostat_type), POINTER                     :: thermostat
      REAL(KIND=dp), INTENT(OUT)                         :: thermostat_pot, thermostat_kin
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: array_pot, array_kin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_thermostat_energies', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: thermo_energy

      thermostat_pot = 0.0_dp
      thermostat_kin = 0.0_dp
      IF (ASSOCIATED(thermostat)) THEN
         IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
            ! Energy associated with the Nose-Hoover thermostat
            IF(.NOT.(ASSOCIATED(thermostat%nhc)))CALL cp__a("motion/thermostat/thermostat_utils.F",1641)
            CALL get_nhc_energies(thermostat%nhc, thermostat_pot, thermostat_kin, para_env, &
                                  array_pot, array_kin)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
            ! Energy associated with the CSVR thermostat
            IF(.NOT.(ASSOCIATED(thermostat%csvr)))CALL cp__a("motion/thermostat/thermostat_utils.F",1646)
            ALLOCATE (thermo_energy(thermostat%csvr%loc_num_csvr))
            DO i = 1, thermostat%csvr%loc_num_csvr
               thermo_energy(i) = thermostat%csvr%nvt(i)%thermostat_energy
            END DO
            CALL get_kin_energies(thermostat%csvr%map_info, thermostat%csvr%loc_num_csvr, &
                                  thermostat%csvr%glob_num_csvr, thermo_energy, &
                                  thermostat_kin, para_env, array_pot, array_kin)
            DEALLOCATE (thermo_energy)

         ELSE IF (thermostat%type_of_thermostat == do_thermo_gle) THEN
            ! Energy associated with the GLE thermostat
            IF(.NOT.(ASSOCIATED(thermostat%gle)))CALL cp__a("motion/thermostat/thermostat_utils.F",1658)
            ALLOCATE (thermo_energy(thermostat%gle%loc_num_gle))
            DO i = 1, thermostat%gle%loc_num_gle
               thermo_energy(i) = thermostat%gle%nvt(i)%thermostat_energy
            END DO
            CALL get_kin_energies(thermostat%gle%map_info, thermostat%gle%loc_num_gle, &
                                  thermostat%gle%glob_num_gle, thermo_energy, &
                                  thermostat_kin, para_env, array_pot, array_kin)
            DEALLOCATE (thermo_energy)

            ![NB] nothing to do for Ad-Langevin?

         END IF
      END IF

   END SUBROUTINE get_thermostat_energies

! **************************************************************************************************
!> \brief Calculates the temperatures for each region associated to a thermostat
!> \param thermostat ...
!> \param tot_temperature ...
!> \param para_env ...
!> \param array_temp ...
!> \author Teodoro Laino [tlaino] - 02.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_region_temperatures(thermostat, tot_temperature, para_env, array_temp)
      TYPE(thermostat_type), POINTER                     :: thermostat
      REAL(KIND=dp), INTENT(OUT)                         :: tot_temperature
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: array_temp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_region_temperatures', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dof, nkt

      IF (ASSOCIATED(thermostat)) THEN
         IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
            ! Energy associated with the Nose-Hoover thermostat
            IF(.NOT.(ASSOCIATED(thermostat%nhc)))CALL cp__a("motion/thermostat/thermostat_utils.F",1698)
            ALLOCATE (nkt(thermostat%nhc%loc_num_nhc))
            ALLOCATE (dof(thermostat%nhc%loc_num_nhc))
            DO i = 1, thermostat%nhc%loc_num_nhc
               nkt(i) = thermostat%nhc%nvt(1, i)%nkt
               dof(i) = REAL(thermostat%nhc%nvt(1, i)%degrees_of_freedom, KIND=dp)
            END DO
            CALL get_temperatures(thermostat%nhc%map_info, thermostat%nhc%loc_num_nhc, &
                                  thermostat%nhc%glob_num_nhc, nkt, dof, para_env, tot_temperature, array_temp)
            DEALLOCATE (nkt)
            DEALLOCATE (dof)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
            ! Energy associated with the CSVR thermostat
            IF(.NOT.(ASSOCIATED(thermostat%csvr)))CALL cp__a("motion/thermostat/thermostat_utils.F",1711)

            ALLOCATE (nkt(thermostat%csvr%loc_num_csvr))
            ALLOCATE (dof(thermostat%csvr%loc_num_csvr))
            DO i = 1, thermostat%csvr%loc_num_csvr
               nkt(i) = thermostat%csvr%nvt(i)%nkt
               dof(i) = REAL(thermostat%csvr%nvt(i)%degrees_of_freedom, KIND=dp)
            END DO
            CALL get_temperatures(thermostat%csvr%map_info, thermostat%csvr%loc_num_csvr, &
                                  thermostat%csvr%glob_num_csvr, nkt, dof, para_env, tot_temperature, array_temp)
            DEALLOCATE (nkt)
            DEALLOCATE (dof)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_al) THEN
            ! Energy associated with the AD_LANGEVIN thermostat
            IF(.NOT.(ASSOCIATED(thermostat%al)))CALL cp__a("motion/thermostat/thermostat_utils.F",1725)

            ALLOCATE (nkt(thermostat%al%loc_num_al))
            ALLOCATE (dof(thermostat%al%loc_num_al))
            DO i = 1, thermostat%al%loc_num_al
               nkt(i) = thermostat%al%nvt(i)%nkt
               dof(i) = REAL(thermostat%al%nvt(i)%degrees_of_freedom, KIND=dp)
            END DO
            CALL get_temperatures(thermostat%al%map_info, thermostat%al%loc_num_al, &
                                  thermostat%al%glob_num_al, nkt, dof, para_env, tot_temperature, array_temp)
            DEALLOCATE (nkt)
            DEALLOCATE (dof)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_gle) THEN
            ! Energy associated with the GLE thermostat
            IF(.NOT.(ASSOCIATED(thermostat%gle)))CALL cp__a("motion/thermostat/thermostat_utils.F",1739)

            ALLOCATE (nkt(thermostat%gle%loc_num_gle))
            ALLOCATE (dof(thermostat%gle%loc_num_gle))
            DO i = 1, thermostat%gle%loc_num_gle
               nkt(i) = thermostat%gle%nvt(i)%nkt
               dof(i) = REAL(thermostat%gle%nvt(i)%degrees_of_freedom, KIND=dp)
            END DO
            CALL get_temperatures(thermostat%gle%map_info, thermostat%gle%loc_num_gle, &
                                  thermostat%gle%glob_num_gle, nkt, dof, para_env, tot_temperature, array_temp)
            DEALLOCATE (nkt)
            DEALLOCATE (dof)
         END IF
      END IF

   END SUBROUTINE get_region_temperatures

! **************************************************************************************************
!> \brief Prints status of all thermostats during an MD run
!> \param thermostats ...
!> \param para_env ...
!> \param my_pos ...
!> \param my_act ...
!> \param itimes ...
!> \param time ...
!> \author Teodoro Laino [tlaino] - 02.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE print_thermostats_status(thermostats, para_env, my_pos, my_act, itimes, time)
      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(LEN=default_string_length)               :: my_pos, my_act
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_thermostats_status', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostats)) THEN
         IF (ASSOCIATED(thermostats%thermostat_part)) THEN
            CALL print_thermostat_status(thermostats%thermostat_part, para_env, my_pos, my_act, itimes, time)
         END IF
         IF (ASSOCIATED(thermostats%thermostat_shell)) THEN
            CALL print_thermostat_status(thermostats%thermostat_shell, para_env, my_pos, my_act, itimes, time)
         END IF
         IF (ASSOCIATED(thermostats%thermostat_coef)) THEN
            CALL print_thermostat_status(thermostats%thermostat_coef, para_env, my_pos, my_act, itimes, time)
         END IF
         IF (ASSOCIATED(thermostats%thermostat_baro)) THEN
            CALL print_thermostat_status(thermostats%thermostat_baro, para_env, my_pos, my_act, itimes, time)
         END IF
      END IF
   END SUBROUTINE print_thermostats_status

! **************************************************************************************************
!> \brief Prints status of a specific thermostat
!> \param thermostat ...
!> \param para_env ...
!> \param my_pos ...
!> \param my_act ...
!> \param itimes ...
!> \param time ...
!> \author Teodoro Laino [tlaino] - 02.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE print_thermostat_status(thermostat, para_env, my_pos, my_act, itimes, time)
      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(LEN=default_string_length)               :: my_pos, my_act
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_thermostat_status', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, unit
      LOGICAL                                            :: new_file
      REAL(KIND=dp)                                      :: thermo_kin, thermo_pot, tot_temperature
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array_kin, array_pot, array_temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger, print_key, array_pot, array_kin, array_temp)
      logger => cp_get_default_logger()

      IF (ASSOCIATED(thermostat)) THEN
         ! Print Energies
         print_key => section_vals_get_subs_vals(thermostat%section, "PRINT%ENERGY")
         IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
            CALL get_thermostat_energies(thermostat, thermo_pot, thermo_kin, para_env, array_pot, array_kin)
            unit = cp_print_key_unit_nr(logger, thermostat%section, "PRINT%ENERGY", &
                                        extension="."//TRIM(thermostat%label)//".tener", file_position=my_pos, &
                                        file_action=my_act, is_new_file=new_file)
            IF (unit > 0) THEN
               IF (new_file) THEN
                  WRITE (unit, '(A)') "# Thermostat Potential and Kinetic Energies - Total and per Region"
                  WRITE (unit, '("#",3X,A,2X,A,13X,A,10X,A)') "Step Nr.", "Time[fs]", "Kin.[a.u.]", "Pot.[a.u.]"
               END IF
               WRITE (UNIT=unit, FMT="(I8, F12.3,6X,2F20.10)") itimes, time*femtoseconds, thermo_kin, thermo_pot
               WRITE (unit, '(A,4F20.10)') "# KINETIC ENERGY REGIONS: ", array_kin(1:MIN(4, SIZE(array_kin)))
               DO i = 5, SIZE(array_kin), 4
                  WRITE (UNIT=unit, FMT='("#",25X,4F20.10)') array_kin(i:MIN(i+3, SIZE(array_kin)))
               END DO
               WRITE (unit, '(A,4F20.10)') "# POTENT. ENERGY REGIONS: ", array_pot(1:MIN(4, SIZE(array_pot)))
               DO i = 5, SIZE(array_pot), 4
                  WRITE (UNIT=unit, FMT='("#",25X,4F20.10)') array_pot(i:MIN(i+3, SIZE(array_pot)))
               END DO
               CALL m_flush(unit)
            END IF
            DEALLOCATE (array_kin)
            DEALLOCATE (array_pot)
            CALL cp_print_key_finished_output(unit, logger, thermostat%section, "PRINT%ENERGY")
         END IF
         ! Print Temperatures of the regions
         print_key => section_vals_get_subs_vals(thermostat%section, "PRINT%TEMPERATURE")
         IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
            CALL get_region_temperatures(thermostat, tot_temperature, para_env, array_temp)
            unit = cp_print_key_unit_nr(logger, thermostat%section, "PRINT%TEMPERATURE", &
                                        extension="."//TRIM(thermostat%label)//".temp", file_position=my_pos, &
                                        file_action=my_act, is_new_file=new_file)
            IF (unit > 0) THEN
               IF (new_file) THEN
                  WRITE (unit, '(A)') "# Temperature Total and per Region"
                  WRITE (unit, '("#",3X,A,2X,A,10X,A)') "Step Nr.", "Time[fs]", "Temp.[K]"
               END IF
               WRITE (UNIT=unit, FMT="(I8, F12.3,3X,F20.10)") itimes, time*femtoseconds, tot_temperature
               WRITE (unit, '(A,I10)') "# TEMPERATURE REGIONS: ", SIZE(array_temp)
               DO i = 1, SIZE(array_temp), 4
                  WRITE (UNIT=unit, FMT='("#",22X,4F20.10)') array_temp(i:MIN(i+3, SIZE(array_temp)))
               END DO
               CALL m_flush(unit)
            END IF
            DEALLOCATE (array_temp)
            CALL cp_print_key_finished_output(unit, logger, thermostat%section, "PRINT%TEMPERATURE")
         END IF
      END IF
   END SUBROUTINE print_thermostat_status

! **************************************************************************************************
!> \brief Handles the communication for thermostats (1D array)
!> \param array ...
!> \param number ...
!> \param para_env ...
!> \author Teodoro Laino [tlaino] - University of Zurich 11.2007
! **************************************************************************************************
   SUBROUTINE communication_thermo_low1(array, number, para_env)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: array
      INTEGER, INTENT(IN)                                :: number
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'communication_thermo_low1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, icheck, ncheck
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work, work2

      ALLOCATE (work(para_env%num_pe))
      DO i = 1, number
         work = 0.0_dp
         work(para_env%mepos+1) = array(i)
         CALL mp_sum(work, para_env%group)
         ncheck = COUNT(work /= 0.0_dp)
         array(i) = 0.0_dp
         IF (ncheck /= 0) THEN
            ALLOCATE (work2(ncheck))
            ncheck = 0
            DO icheck = 1, para_env%num_pe
               IF (work(icheck) /= 0.0_dp) THEN
                  ncheck = ncheck+1
                  work2(ncheck) = work(icheck)
               END IF
            END DO
            IF(.NOT.(ncheck == SIZE(work2)))CALL cp__a("motion/thermostat/thermostat_utils.F",1909)
            IF(.NOT.(ALL(work2 == work2(1))))CALL cp__a("motion/thermostat/thermostat_utils.F",1910)

            array(i) = work2(1)
            DEALLOCATE (work2)
         END IF
      END DO
      DEALLOCATE (work)
   END SUBROUTINE communication_thermo_low1

! **************************************************************************************************
!> \brief Handles the communication for thermostats (2D array)
!> \param array ...
!> \param number1 ...
!> \param number2 ...
!> \param para_env ...
!> \author Teodoro Laino [tlaino] - University of Zurich 11.2007
! **************************************************************************************************
   SUBROUTINE communication_thermo_low2(array, number1, number2, para_env)
      INTEGER, DIMENSION(:, :), INTENT(INOUT)            :: array
      INTEGER, INTENT(IN)                                :: number1, number2
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'communication_thermo_low2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, icheck, j, ncheck
      INTEGER, DIMENSION(:, :), POINTER                  :: work, work2

      ALLOCATE (work(number1, para_env%num_pe))
      DO i = 1, number2
         work = 0
         work(:, para_env%mepos+1) = array(:, i)
         CALL mp_sum(work, para_env%group)
         ncheck = 0
         DO j = 1, para_env%num_pe
            IF (ANY(work(:, j) /= 0)) THEN
               ncheck = ncheck+1
            END IF
         END DO
         array(:, i) = 0
         IF (ncheck /= 0) THEN
            ALLOCATE (work2(number1, ncheck))
            ncheck = 0
            DO icheck = 1, para_env%num_pe
               IF (ANY(work(:, icheck) /= 0)) THEN
                  ncheck = ncheck+1
                  work2(:, ncheck) = work(:, icheck)
               END IF
            END DO
            IF(.NOT.(ncheck == SIZE(work2, 2)))CALL cp__a("motion/thermostat/thermostat_utils.F",1959)
            DO j = 1, ncheck
               IF(.NOT.(ALL(work2(:, j) == work2(:, 1))))CALL cp__a("motion/thermostat/thermostat_utils.F",1961)
            END DO
            array(:, i) = work2(:, 1)
            DEALLOCATE (work2)
         END IF
      END DO
      DEALLOCATE (work)
   END SUBROUTINE communication_thermo_low2

END MODULE thermostat_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      cjm, FEB 20 2001:  added subroutine initialize_extended_parameters
!>      cjm, MAY 03 2001:  reorganized and added separtate routines for
!>                         nhc_part, nhc_baro, nhc_ao, npt
!> \author CJM
! **************************************************************************************************
MODULE extended_system_init

   USE cell_types,                      ONLY: cell_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_mapping,         ONLY: nhc_to_barostat_mapping,&
                                              nhc_to_particle_mapping,&
                                              nhc_to_particle_mapping_fast,&
                                              nhc_to_particle_mapping_slow,&
                                              nhc_to_shell_mapping
   USE extended_system_types,           ONLY: debug_isotropic_limit,&
                                              lnhc_parameters_type,&
                                              map_info_type,&
                                              npt_info_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: do_thermo_only_master,&
                                              npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble
   USE input_cp2k_binary_restarts,      ONLY: read_binary_thermostats_nose
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: next_random_number
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: thermostat_info_type
   USE thermostat_utils,                ONLY: get_nhc_energies
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'extended_system_init'

   PUBLIC :: initialize_nhc_part, initialize_nhc_baro, initialize_npt, &
             initialize_nhc_shell, initialize_nhc_slow, initialize_nhc_fast

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param simpar ...
!> \param globenv ...
!> \param npt_info ...
!> \param cell ...
!> \param work_section ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE initialize_npt(simpar, globenv, npt_info, cell, work_section)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(npt_info_type), DIMENSION(:, :), POINTER      :: npt_info
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: work_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_npt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ind, j
      LOGICAL                                            :: explicit, restart
      REAL(KIND=dp)                                      :: temp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buffer
      TYPE(section_vals_type), POINTER                   :: work_section2

      CALL timeset(routineN, handle)

      NULLIFY (work_section2)

      explicit = .FALSE.
      restart = .FALSE.

      IF(.NOT.(.NOT. ASSOCIATED(npt_info)))CALL cp__a("motion/thermostat/extended_system_init.F",94)

      ! first allocating the npt_info_type if requested
      SELECT CASE (simpar%ensemble)
      CASE (npt_i_ensemble, npe_i_ensemble)
         ALLOCATE (npt_info(1, 1))
         npt_info(:, :)%eps = LOG(cell%deth)/3.0_dp
         temp = simpar%temp_baro_ext

      CASE (npt_f_ensemble, npe_f_ensemble)
         ALLOCATE (npt_info(3, 3))
         temp = simpar%temp_baro_ext

      CASE (nph_uniaxial_ensemble)
         ALLOCATE (npt_info(1, 1))
         temp = simpar%temp_baro_ext

      CASE (nph_uniaxial_damped_ensemble)
         ALLOCATE (npt_info(1, 1))
         temp = simpar%temp_baro_ext

      CASE DEFAULT
         ! Do nothing..
         NULLIFY (npt_info)
      END SELECT

      IF (ASSOCIATED(npt_info)) THEN
         IF (ASSOCIATED(work_section)) THEN
            work_section2 => section_vals_get_subs_vals(work_section, "VELOCITY")
            CALL section_vals_get(work_section2, explicit=explicit)
            restart = explicit
            work_section2 => section_vals_get_subs_vals(work_section, "MASS")
            CALL section_vals_get(work_section2, explicit=explicit)
            IF (restart .NEQV. explicit) &
               CALL cp_abort(cp__l("motion/thermostat/extended_system_init.F",128), "You need to define both VELOCITY and "// &
                             "MASS section (or none) in the BAROSTAT section")
            restart = explicit .AND. restart
         END IF

         IF (restart) THEN
            CALL section_vals_val_get(work_section, "VELOCITY%_DEFAULT_KEYWORD_", r_vals=buffer)
            ind = 0
            DO i = 1, SIZE(npt_info, 1)
               DO j = 1, SIZE(npt_info, 2)
                  ind = ind+1
                  npt_info(i, j)%v = buffer(ind)
               END DO
            END DO
            CALL section_vals_val_get(work_section, "MASS%_DEFAULT_KEYWORD_", r_vals=buffer)
            ind = 0
            DO i = 1, SIZE(npt_info, 1)
               DO j = 1, SIZE(npt_info, 2)
                  ind = ind+1
                  npt_info(i, j)%mass = buffer(ind)
               END DO
            END DO
         ELSE
            CALL init_barostat_variables(npt_info, simpar%tau_cell, temp, &
                                         simpar%nfree, simpar%ensemble, simpar%cmass, &
                                         globenv)
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE initialize_npt

! **************************************************************************************************
!> \brief fire up the thermostats, if NPT
!> \param simpar ...
!> \param para_env ...
!> \param globenv ...
!> \param nhc ...
!> \param nose_section ...
!> \param save_mem ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE initialize_nhc_baro(simpar, para_env, globenv, nhc, nose_section, save_mem)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      LOGICAL, INTENT(IN)                                :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_nhc_baro', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: restart
      REAL(KIND=dp)                                      :: temp

      CALL timeset(routineN, handle)

      restart = .FALSE.

      CALL nhc_to_barostat_mapping(simpar, nhc)

      ! Set up the Yoshida weights
      IF (nhc%nyosh > 0) THEN
         ALLOCATE (nhc%dt_yosh(1:nhc%nyosh))
         CALL set_yoshida_coef(nhc, simpar%dt)
      END IF

      CALL restart_nose(nhc, nose_section, save_mem, restart, "", "", para_env)

      IF (.NOT. restart) THEN
         ! Initializing thermostat forces and velocities for the Nose-Hoover
         ! Chain variables
         SELECT CASE (simpar%ensemble)
         CASE DEFAULT
            temp = simpar%temp_baro_ext
         END SELECT
         IF (nhc%nhc_len /= 0) THEN
            CALL init_nhc_variables(nhc, temp, para_env, globenv)
         END IF
      END IF

      CALL init_nhc_forces(nhc)

      CALL timestop(handle)

   END SUBROUTINE initialize_nhc_baro

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param globenv ...
!> \param nhc ...
!> \param nose_section ...
!> \param gci ...
!> \param save_mem ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE initialize_nhc_slow(thermostat_info, simpar, local_molecules, &
                                  molecule, molecule_kind_set, para_env, globenv, nhc, nose_section, &
                                  gci, save_mem)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_nhc_slow', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: restart

      CALL timeset(routineN, handle)

      restart = .FALSE.
      ! fire up the thermostats, if not NVE

      CALL nhc_to_particle_mapping_slow(thermostat_info, simpar, local_molecules, &
                                        molecule, molecule_kind_set, nhc, para_env, gci)

      ! Set up the Yoshida weights
      IF (nhc%nyosh > 0) THEN
         ALLOCATE (nhc%dt_yosh(1:nhc%nyosh))
         CALL set_yoshida_coef(nhc, simpar%dt)
      END IF

      CALL restart_nose(nhc, nose_section, save_mem, restart, "", "", para_env)

      IF (.NOT. restart) THEN
         ! Initializing thermostat forces and velocities for the Nose-Hoover
         ! Chain variables
         IF (nhc%nhc_len /= 0) THEN
            CALL init_nhc_variables(nhc, simpar%temp_slow, para_env, globenv)
         END IF
      END IF

      CALL init_nhc_forces(nhc)

      CALL timestop(handle)

   END SUBROUTINE initialize_nhc_slow

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param globenv ...
!> \param nhc ...
!> \param nose_section ...
!> \param gci ...
!> \param save_mem ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE initialize_nhc_fast(thermostat_info, simpar, local_molecules, &
                                  molecule, molecule_kind_set, para_env, globenv, nhc, nose_section, &
                                  gci, save_mem)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_nhc_fast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: restart

      CALL timeset(routineN, handle)

      restart = .FALSE.
      ! fire up the thermostats, if not NVE

      CALL nhc_to_particle_mapping_fast(thermostat_info, simpar, local_molecules, &
                                        molecule, molecule_kind_set, nhc, para_env, gci)

      ! Set up the Yoshida weights
      IF (nhc%nyosh > 0) THEN
         ALLOCATE (nhc%dt_yosh(1:nhc%nyosh))
         CALL set_yoshida_coef(nhc, simpar%dt)
      END IF

      CALL restart_nose(nhc, nose_section, save_mem, restart, "", "", para_env)

      IF (.NOT. restart) THEN
         ! Initializing thermostat forces and velocities for the Nose-Hoover
         ! Chain variables
         IF (nhc%nhc_len /= 0) THEN
            CALL init_nhc_variables(nhc, simpar%temp_fast, para_env, globenv)
         END IF
      END IF

      CALL init_nhc_forces(nhc)

      CALL timestop(handle)

   END SUBROUTINE initialize_nhc_fast

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param globenv ...
!> \param nhc ...
!> \param nose_section ...
!> \param gci ...
!> \param save_mem ...
!> \param binary_restart_file_name ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE initialize_nhc_part(thermostat_info, simpar, local_molecules, &
                                  molecule, molecule_kind_set, para_env, globenv, nhc, nose_section, &
                                  gci, save_mem, binary_restart_file_name)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: save_mem
      CHARACTER(LEN=*), INTENT(IN)                       :: binary_restart_file_name

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_nhc_part', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: restart

      CALL timeset(routineN, handle)

      restart = .FALSE.
      ! fire up the thermostats, if not NVE

      CALL nhc_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                   molecule, molecule_kind_set, nhc, para_env, gci)

      ! Set up the Yoshida weights
      IF (nhc%nyosh > 0) THEN
         ALLOCATE (nhc%dt_yosh(1:nhc%nyosh))
         CALL set_yoshida_coef(nhc, simpar%dt)
      END IF

      CALL restart_nose(nhc, nose_section, save_mem, restart, binary_restart_file_name, &
                        "PARTICLE", para_env)

      IF (.NOT. restart) THEN
         ! Initializing thermostat forces and velocities for the Nose-Hoover
         ! Chain variables
         IF (nhc%nhc_len /= 0) THEN
            CALL init_nhc_variables(nhc, simpar%temp_ext, para_env, globenv)
         END IF
      END IF

      CALL init_nhc_forces(nhc)

      CALL timestop(handle)

   END SUBROUTINE initialize_nhc_part

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param globenv ...
!> \param nhc ...
!> \param nose_section ...
!> \param gci ...
!> \param save_mem ...
!> \param binary_restart_file_name ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE initialize_nhc_shell(thermostat_info, simpar, local_molecules, &
                                   molecule, molecule_kind_set, para_env, globenv, nhc, nose_section, &
                                   gci, save_mem, binary_restart_file_name)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: save_mem
      CHARACTER(LEN=*), INTENT(IN)                       :: binary_restart_file_name

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_nhc_shell', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: restart

      CALL timeset(routineN, handle)

      CALL nhc_to_shell_mapping(thermostat_info, simpar, local_molecules, &
                                molecule, molecule_kind_set, nhc, para_env, gci)

      restart = .FALSE.
      ! Set up the Yoshida weights
      IF (nhc%nyosh > 0) THEN
         ALLOCATE (nhc%dt_yosh(1:nhc%nyosh))
         CALL set_yoshida_coef(nhc, simpar%dt)
      END IF

      CALL restart_nose(nhc, nose_section, save_mem, restart, binary_restart_file_name, &
                        "SHELL", para_env)

      IF (.NOT. restart) THEN
         ! Initialize thermostat forces and velocities
         ! Chain variables
         IF (nhc%nhc_len /= 0) THEN
            CALL init_nhc_variables(nhc, simpar%temp_sh_ext, para_env, globenv)
         END IF
      END IF

      CALL init_nhc_forces(nhc)

      CALL timestop(handle)

   END SUBROUTINE initialize_nhc_shell

! **************************************************************************************************
!> \brief This lists the coefficients for the Yoshida method (higher
!>      order integrator used in NVT)
!> \param nhc ...
!> \param dt ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE set_yoshida_coef(nhc, dt)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      REAL(KIND=dp), INTENT(IN)                          :: dt

      CHARACTER(len=*), PARAMETER :: routineN = 'set_yoshida_coef', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(nhc%nyosh)                :: yosh_wt

      SELECT CASE (nhc%nyosh)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/extended_system_init.F",513,'Value not available.')
      CASE (1)
         yosh_wt(1) = 1.0_dp
      CASE (3)
         yosh_wt(1) = 1.0_dp/(2.0_dp-(2.0_dp)**(1.0_dp/3.0_dp))
         yosh_wt(2) = 1.0_dp-2.0_dp*yosh_wt(1)
         yosh_wt(3) = yosh_wt(1)
      CASE (5)
         yosh_wt(1) = 1.0_dp/(4.0_dp-(4.0_dp)**(1.0_dp/3.0_dp))
         yosh_wt(2) = yosh_wt(1)
         yosh_wt(4) = yosh_wt(1)
         yosh_wt(5) = yosh_wt(1)
         yosh_wt(3) = 1.0_dp-4.0_dp*yosh_wt(1)
      CASE (7)
         yosh_wt(1) = .78451361047756_dp
         yosh_wt(2) = .235573213359357_dp
         yosh_wt(3) = -1.17767998417887_dp
         yosh_wt(4) = 1.0_dp-2.0_dp*(yosh_wt(1)+yosh_wt(2)+yosh_wt(3))
         yosh_wt(5) = yosh_wt(3)
         yosh_wt(6) = yosh_wt(2)
         yosh_wt(7) = yosh_wt(1)
      CASE (9)
         yosh_wt(1) = 0.192_dp
         yosh_wt(2) = 0.554910818409783619692725006662999_dp
         yosh_wt(3) = 0.124659619941888644216504240951585_dp
         yosh_wt(4) = -0.843182063596933505315033808282941_dp
         yosh_wt(5) = 1.0_dp-2.0_dp*(yosh_wt(1)+yosh_wt(2)+ &
                                     yosh_wt(3)+yosh_wt(4))
         yosh_wt(6) = yosh_wt(4)
         yosh_wt(7) = yosh_wt(3)
         yosh_wt(8) = yosh_wt(2)
         yosh_wt(9) = yosh_wt(1)
      CASE (15)
         yosh_wt(1) = 0.102799849391985_dp
         yosh_wt(2) = -0.196061023297549e1_dp
         yosh_wt(3) = 0.193813913762276e1_dp
         yosh_wt(4) = -0.158240635368243_dp
         yosh_wt(5) = -0.144485223686048e1_dp
         yosh_wt(6) = 0.253693336566229_dp
         yosh_wt(7) = 0.914844246229740_dp
         yosh_wt(8) = 1.0_dp-2.0_dp*(yosh_wt(1)+yosh_wt(2)+ &
                                     yosh_wt(3)+yosh_wt(4)+yosh_wt(5)+yosh_wt(6)+yosh_wt(7))
         yosh_wt(9) = yosh_wt(7)
         yosh_wt(10) = yosh_wt(6)
         yosh_wt(11) = yosh_wt(5)
         yosh_wt(12) = yosh_wt(4)
         yosh_wt(13) = yosh_wt(3)
         yosh_wt(14) = yosh_wt(2)
         yosh_wt(15) = yosh_wt(1)
      END SELECT
      nhc%dt_yosh = dt*yosh_wt/REAL(nhc%nc, KIND=dp)

   END SUBROUTINE set_yoshida_coef

! **************************************************************************************************
!> \brief read coordinate, velocities, forces and masses of the
!>      thermostat from restart file
!> \param nhc ...
!> \param nose_section ...
!> \param save_mem ...
!> \param restart ...
!> \param binary_restart_file_name ...
!> \param thermostat_name ...
!> \param para_env ...
!> \par History
!>     24-07-07 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE restart_nose(nhc, nose_section, save_mem, restart, &
                           binary_restart_file_name, thermostat_name, &
                           para_env)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(section_vals_type), POINTER                   :: nose_section
      LOGICAL, INTENT(IN)                                :: save_mem
      LOGICAL, INTENT(OUT)                               :: restart
      CHARACTER(LEN=*), INTENT(IN)                       :: binary_restart_file_name, thermostat_name
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_nose', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ind, j
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buffer
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(section_vals_type), POINTER                   :: work_section

      CALL timeset(routineN, handle)

      NULLIFY (buffer)
      NULLIFY (work_section)

      IF (LEN_TRIM(binary_restart_file_name) > 0) THEN

         ! Read binary restart file, if available

         CALL read_binary_thermostats_nose(thermostat_name, nhc, binary_restart_file_name, &
                                           restart, para_env)

      ELSE

         ! Read the default restart file in ASCII format

         explicit = .FALSE.
         restart = .FALSE.

         IF (ASSOCIATED(nose_section)) THEN
            work_section => section_vals_get_subs_vals(nose_section, "VELOCITY")
            CALL section_vals_get(work_section, explicit=explicit)
            restart = explicit
            work_section => section_vals_get_subs_vals(nose_section, "COORD")
            CALL section_vals_get(work_section, explicit=explicit)
            IF (.NOT. restart .AND. explicit) &
               CALL cp_abort(cp__l("motion/thermostat/extended_system_init.F",626), "You need to define both VELOCITY and "// &
                             "COORD and MASS and FORCE section (or none) in the NOSE section")
            restart = explicit .AND. restart
            work_section => section_vals_get_subs_vals(nose_section, "MASS")
            CALL section_vals_get(work_section, explicit=explicit)
            IF (.NOT. restart .AND. explicit) &
               CALL cp_abort(cp__l("motion/thermostat/extended_system_init.F",632), "You need to define both VELOCITY and "// &
                             "COORD and MASS and FORCE section (or none) in the NOSE section")
            restart = explicit .AND. restart
            work_section => section_vals_get_subs_vals(nose_section, "FORCE")
            CALL section_vals_get(work_section, explicit=explicit)
            IF (.NOT. restart .AND. explicit) &
               CALL cp_abort(cp__l("motion/thermostat/extended_system_init.F",638), "You need to define both VELOCITY and "// &
                             "COORD and MASS and FORCE section (or none) in the NOSE section")
            restart = explicit .AND. restart
         END IF

         IF (restart) THEN
            map_info => nhc%map_info
            CALL section_vals_val_get(nose_section, "COORD%_DEFAULT_KEYWORD_", r_vals=buffer)
            DO i = 1, SIZE(nhc%nvt, 2)
               ind = map_info%index(i)
               ind = (ind-1)*nhc%nhc_len
               DO j = 1, SIZE(nhc%nvt, 1)
                  ind = ind+1
                  nhc%nvt(j, i)%eta = buffer(ind)
               END DO
            END DO
            CALL section_vals_val_get(nose_section, "VELOCITY%_DEFAULT_KEYWORD_", r_vals=buffer)
            DO i = 1, SIZE(nhc%nvt, 2)
               ind = map_info%index(i)
               ind = (ind-1)*nhc%nhc_len
               DO j = 1, SIZE(nhc%nvt, 1)
                  ind = ind+1
                  nhc%nvt(j, i)%v = buffer(ind)
               END DO
            END DO
            CALL section_vals_val_get(nose_section, "MASS%_DEFAULT_KEYWORD_", r_vals=buffer)
            DO i = 1, SIZE(nhc%nvt, 2)
               ind = map_info%index(i)
               ind = (ind-1)*nhc%nhc_len
               DO j = 1, SIZE(nhc%nvt, 1)
                  ind = ind+1
                  nhc%nvt(j, i)%mass = buffer(ind)
               END DO
            END DO
            CALL section_vals_val_get(nose_section, "FORCE%_DEFAULT_KEYWORD_", r_vals=buffer)
            DO i = 1, SIZE(nhc%nvt, 2)
               ind = map_info%index(i)
               ind = (ind-1)*nhc%nhc_len
               DO j = 1, SIZE(nhc%nvt, 1)
                  ind = ind+1
                  nhc%nvt(j, i)%f = buffer(ind)
               END DO
            END DO
         END IF

         IF (save_mem) THEN
            NULLIFY (work_section)
            work_section => section_vals_get_subs_vals(nose_section, "COORD")
            CALL section_vals_remove_values(work_section)
            NULLIFY (work_section)
            work_section => section_vals_get_subs_vals(nose_section, "VELOCITY")
            CALL section_vals_remove_values(work_section)
            NULLIFY (work_section)
            work_section => section_vals_get_subs_vals(nose_section, "FORCE")
            CALL section_vals_remove_values(work_section)
            NULLIFY (work_section)
            work_section => section_vals_get_subs_vals(nose_section, "MASS")
            CALL section_vals_remove_values(work_section)
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE restart_nose

! **************************************************************************************************
!> \brief Initializes the NHC velocities to the Maxwellian distribution
!> \param nhc ...
!> \param temp_ext ...
!> \param para_env ...
!> \param globenv ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE init_nhc_variables(nhc, temp_ext, para_env, globenv)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      REAL(KIND=dp), INTENT(IN)                          :: temp_ext
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'init_nhc_variables', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, icount, j, number, tot_rn
      REAL(KIND=dp)                                      :: akin, dum, temp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: array_of_rn
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)

      tot_rn = 0

      ! first initializing the mass of the nhc variables
      nhc%nvt(:, :)%mass = nhc%nvt(:, :)%nkt*nhc%tau_nhc**2
      nhc%nvt(:, :)%eta = 0._dp
      nhc%nvt(:, :)%v = 0._dp
      nhc%nvt(:, :)%f = 0._dp

      map_info => nhc%map_info
      SELECT CASE (map_info%dis_type)
      CASE (do_thermo_only_master) ! for NPT
      CASE DEFAULT
         tot_rn = nhc%glob_num_nhc*nhc%nhc_len

         ALLOCATE (array_of_rn(tot_rn))
         array_of_rn(:) = 0.0_dp
      END SELECT

      SELECT CASE (map_info%dis_type)
      CASE (do_thermo_only_master) ! for NPT
         ! Map deterministically determined random number to nhc % v
         DO i = 1, nhc%loc_num_nhc
            DO j = 1, nhc%nhc_len
               nhc%nvt(j, i)%v = next_random_number(globenv%gaussian_rng_stream)
            END DO
         END DO

         akin = 0.0_dp
         DO i = 1, nhc%loc_num_nhc
            DO j = 1, nhc%nhc_len
               akin = akin+0.5_dp*(nhc%nvt(j, i)%mass* &
                                   nhc%nvt(j, i)%v* &
                                   nhc%nvt(j, i)%v)
            END DO
         END DO
         number = nhc%loc_num_nhc

         ! scale velocities to get the correct initial temperature
         temp = 2.0_dp*akin/REAL(number)
         IF (temp > 0.0_dp) temp = SQRT(temp_ext/temp)
         DO i = 1, nhc%loc_num_nhc
            DO j = 1, nhc%nhc_len
               nhc%nvt(j, i)%v = temp*nhc%nvt(j, i)%v
               nhc%nvt(j, i)%eta = 0.0_dp
            END DO
         END DO

         ! initializing all of the forces on the thermostats
         DO i = 1, nhc%loc_num_nhc
            DO j = 2, nhc%nhc_len
               nhc%nvt(j, i)%f = nhc%nvt(j-1, i)%mass*nhc%nvt(j-1, i)%v* &
                                 nhc%nvt(j-1, i)%v-nhc%nvt(j, i)%nkt
               IF (nhc%nvt(j, i)%mass > 0.0_dp) THEN
                  nhc%nvt(j, i)%f = nhc%nvt(j, i)%f/nhc%nvt(j, i)%mass
               END IF
            END DO
         END DO

      CASE DEFAULT
         DO i = 1, tot_rn
            array_of_rn(i) = next_random_number(globenv%gaussian_rng_stream)
         END DO
         ! Map deterministically determined random number to nhc % v
         DO i = 1, nhc%loc_num_nhc
            icount = map_info%index(i)
            icount = (icount-1)*nhc%nhc_len
            DO j = 1, nhc%nhc_len
               icount = icount+1
               nhc%nvt(j, i)%v = array_of_rn(icount)
               ! WRITE ( *, * ) 'VEL', para_env%mepos, i,j, nhc%nvt(j,i)%v
               nhc%nvt(j, i)%eta = 0.0_dp
            END DO
         END DO
         DEALLOCATE (array_of_rn)

         number = nhc%glob_num_nhc
         CALL get_nhc_energies(nhc, dum, akin, para_env)

         ! scale velocities to get the correct initial temperature
         temp = 2.0_dp*akin/REAL(number)
         IF (temp > 0.0_dp) temp = SQRT(temp_ext/temp)
         DO i = 1, nhc%loc_num_nhc
            DO j = 1, nhc%nhc_len
               nhc%nvt(j, i)%v = temp*nhc%nvt(j, i)%v
            END DO
         END DO

         ! initializing all of the forces on the thermostats
         DO i = 1, nhc%loc_num_nhc
            DO j = 2, nhc%nhc_len
               nhc%nvt(j, i)%f = nhc%nvt(j-1, i)%mass*nhc%nvt(j-1, i)%v* &
                                 nhc%nvt(j-1, i)%v-nhc%nvt(j, i)%nkt
               IF (nhc%nvt(j, i)%mass > 0.0_dp) THEN
                  nhc%nvt(j, i)%f = nhc%nvt(j, i)%f/nhc%nvt(j, i)%mass
               END IF
            END DO
         END DO

      END SELECT

      CALL timestop(handle)

   END SUBROUTINE init_nhc_variables

! **************************************************************************************************
!> \brief Initializes the barostat velocities to the Maxwellian distribution
!> \param npt ...
!> \param tau_cell ...
!> \param temp_ext ...
!> \param nfree ...
!> \param ensemble ...
!> \param cmass ...
!> \param globenv ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE init_barostat_variables(npt, tau_cell, temp_ext, nfree, ensemble, &
                                      cmass, globenv)

      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt
      REAL(KIND=dp), INTENT(IN)                          :: tau_cell, temp_ext
      INTEGER, INTENT(IN)                                :: nfree, ensemble
      REAL(KIND=dp), INTENT(IN)                          :: cmass
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'init_barostat_variables', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, number
      REAL(KIND=dp)                                      :: akin, temp, v

      CALL timeset(routineN, handle)

      temp = 0.0_dp

      ! first initializing the mass of the nhc variables

      npt(:, :)%eps = 0.0_dp
      npt(:, :)%v = 0.0_dp
      npt(:, :)%f = 0.0_dp
      SELECT CASE (ensemble)
      CASE (npt_i_ensemble)
         npt(:, :)%mass = REAL(nfree+3, KIND=dp)*temp_ext*tau_cell**2
      CASE (npt_f_ensemble)
         npt(:, :)%mass = REAL(nfree+3, KIND=dp)*temp_ext*tau_cell**2/3.0_dp
      CASE (nph_uniaxial_ensemble, nph_uniaxial_damped_ensemble)
         npt(:, :)%mass = cmass
      CASE (npe_f_ensemble)
         npt(:, :)%mass = REAL(nfree+3, KIND=dp)*temp_ext*tau_cell**2/3.0_dp
      CASE (npe_i_ensemble)
         npt(:, :)%mass = REAL(nfree+3, KIND=dp)*temp_ext*tau_cell**2
      END SELECT
      ! initializing velocities
      DO i = 1, SIZE(npt, 1)
         DO j = i, SIZE(npt, 2)
            v = next_random_number(globenv%gaussian_rng_stream)
            ! Symmetrizing the initial barostat velocities to ensure
            ! no rotation of the cell under NPT_F
            npt(j, i)%v = v
            npt(i, j)%v = v
         END DO
      END DO

      akin = 0.0_dp
      DO i = 1, SIZE(npt, 1)
         DO j = 1, SIZE(npt, 2)
            akin = akin+0.5_dp*(npt(j, i)%mass*npt(j, i)%v*npt(j, i)%v)
         END DO
      END DO

      number = SIZE(npt, 1)*SIZE(npt, 2)

      ! scale velocities to get the correct initial temperature
      IF (number /= 0) THEN
         temp = 2.0_dp*akin/REAL(number, KIND=dp)
         IF (temp > 0.0_dp) temp = SQRT(temp_ext/temp)
      ENDIF
      DO i = 1, SIZE(npt, 1)
         DO j = i, SIZE(npt, 2)
            npt(j, i)%v = temp*npt(j, i)%v
            npt(i, j)%v = npt(j, i)%v
            IF (debug_isotropic_limit) THEN
               npt(j, i)%v = 0.0_dp
               npt(i, j)%v = 0.0_dp
               WRITE (*, *) 'DEBUG ISOTROPIC LIMIT| INTIAL v_eps', npt(j, i)%v
            END IF
         END DO
      END DO

      ! Zero barostat velocities for nph_uniaxial
      SELECT CASE (ensemble)
         ! Zero barostat velocities for nph_uniaxial
      CASE (nph_uniaxial_ensemble, nph_uniaxial_damped_ensemble)
         npt(:, :)%v = 0.0_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE init_barostat_variables

! **************************************************************************************************
!> \brief Assigns extended parameters from the restart file.
!> \param nhc ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE init_nhc_forces(nhc)

      TYPE(lnhc_parameters_type), POINTER                :: nhc

      CHARACTER(len=*), PARAMETER :: routineN = 'init_nhc_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(nhc)))CALL cp__a("motion/thermostat/extended_system_init.F",948)
      ! assign the forces
      DO i = 1, SIZE(nhc%nvt, 2)
         DO j = 2, SIZE(nhc%nvt, 1)
            nhc%nvt(j, i)%f = nhc%nvt(j-1, i)%mass* &
                              nhc%nvt(j-1, i)%v**2- &
                              nhc%nvt(j, i)%nkt
            IF (nhc%nvt(j, i)%mass > 0.0_dp) THEN
               nhc%nvt(j, i)%f = nhc%nvt(j, i)%f/nhc%nvt(j, i)%mass
            END IF
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE init_nhc_forces

END MODULE extended_system_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Barostat structure: module containing barostat available for MD
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
MODULE barostat_types
   USE cell_types,                      ONLY: cell_type
   USE extended_system_init,            ONLY: initialize_npt
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   INTEGER, PARAMETER, PUBLIC               :: do_clv_geo_center = 0, &
                                               do_clv_fix_point = 1, &
                                               do_clv_xyz = 0, &
                                               do_clv_x = 1, &
                                               do_clv_y = 2, &
                                               do_clv_z = 3, &
                                               do_clv_xy = 4, &
                                               do_clv_xz = 5, &
                                               do_clv_yz = 6

   PRIVATE
   PUBLIC :: barostat_type, &
             create_barostat_type, &
             release_barostat_type, &
             retain_barostat_type

! **************************************************************************************************
   TYPE barostat_type
      INTEGER                          :: id_nr, ref_count
      INTEGER                          :: type_of_barostat, virial_components
      REAL(KIND=dp)                    :: temp_ext
      TYPE(npt_info_type), POINTER  :: npt(:, :)
      TYPE(section_vals_type), POINTER :: section
   END TYPE barostat_type

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'barostat_types'
   INTEGER, PRIVATE, SAVE               :: last_barostat_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param barostat ...
!> \param md_section ...
!> \param force_env ...
!> \param simpar ...
!> \param globenv ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_barostat_type(barostat, md_section, force_env, simpar, &
                                   globenv)
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'create_barostat_type', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check, explicit
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: barostat_section

      check = .NOT. ASSOCIATED(barostat)
      IF(.NOT.(check))CALL cp__a("motion/thermostat/barostat_types.F",92)
      barostat_section => section_vals_get_subs_vals(md_section, "BAROSTAT")
      CALL section_vals_get(barostat_section, explicit=explicit)
      IF (simpar%ensemble == npt_i_ensemble .OR. &
          simpar%ensemble == npt_f_ensemble .OR. &
          simpar%ensemble == npe_f_ensemble .OR. &
          simpar%ensemble == npe_i_ensemble .OR. &
          simpar%ensemble == nph_uniaxial_ensemble .OR. &
          simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         ALLOCATE (barostat)
         last_barostat_id_nr = last_barostat_id_nr+1
         barostat%id_nr = last_barostat_id_nr
         barostat%ref_count = 1
         barostat%section => barostat_section
         NULLIFY (barostat%npt)
         CALL force_env_get(force_env, cell=cell)

         barostat%temp_ext = simpar%temp_baro_ext
         CALL section_vals_val_get(barostat_section, "TEMP_TOL", r_val=simpar%temp_baro_tol)
         ! Initialize or possibly restart Barostat
         CALL initialize_npt(simpar, globenv, barostat%npt, &
                             cell, work_section=barostat_section)

         ! If none of the possible barostat has been allocated let's deallocate
         ! the full structure
         IF (.NOT. ASSOCIATED(barostat%npt)) THEN
            CALL release_barostat_type(barostat)
         END IF

         ! User defined virial screening
         CALL section_vals_val_get(barostat_section, "VIRIAL", i_val=barostat%virial_components)
         check = barostat%virial_components == do_clv_xyz .OR. simpar%ensemble == npt_f_ensemble
         IF (.NOT. check) &
            CALL cp_abort(cp__l("motion/thermostat/barostat_types.F",125), "The screening of the components of "// &
                          "the virial is available only with the NPT_F ensemble!")
      ELSE
         IF (explicit) &
            CALL cp_warn(cp__l("motion/thermostat/barostat_types.F",129), &
                         "A barostat has been defined with an MD ensemble which does not support barostat! "// &
                         "It's definition will be ignored!")
      END IF

   END SUBROUTINE create_barostat_type

! **************************************************************************************************
!> \brief retains the given barostat
!> \param barostat ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE retain_barostat_type(barostat)
      TYPE(barostat_type), POINTER                       :: barostat

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_barostat_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(barostat)) THEN
         IF(.NOT.(barostat%ref_count > 0))CALL cp__a("motion/thermostat/barostat_types.F",150)
         barostat%ref_count = barostat%ref_count+1
      END IF
   END SUBROUTINE retain_barostat_type

! **************************************************************************************************
!> \brief ...
!> \param barostat ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_barostat_type(barostat)
      TYPE(barostat_type), POINTER                       :: barostat

      CHARACTER(len=*), PARAMETER :: routineN = 'release_barostat_type', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      IF (ASSOCIATED(barostat)) THEN
         check = barostat%ref_count > 0
         IF(.NOT.(check))CALL cp__a("motion/thermostat/barostat_types.F",172)
         barostat%ref_count = barostat%ref_count-1
         IF (barostat%ref_count < 1) THEN
            IF (ASSOCIATED(barostat%npt)) THEN
               DEALLOCATE (barostat%npt)
            END IF
            NULLIFY (barostat%section)
            DEALLOCATE (barostat)
         END IF
      END IF

   END SUBROUTINE release_barostat_type

END MODULE barostat_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      give the md_env its own para_env Joost VandeVondele 07.2003
!>      Teodoro Laino - 09.2007 - University of Zurich - generalizing thermostats
!>                                and barostats
!> \author CJM SEPT-12-02
! **************************************************************************************************
MODULE md_environment_types
   USE averages_types,                  ONLY: average_quantities_type,&
                                              create_averages,&
                                              release_averages,&
                                              retain_averages
   USE barostat_types,                  ONLY: barostat_type,&
                                              release_barostat_type,&
                                              retain_barostat_type
   USE cell_types,                      ONLY: cell_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_types,                 ONLY: force_env_release,&
                                              force_env_retain,&
                                              force_env_type
   USE free_energy_types,               ONLY: fe_env_release,&
                                              free_energy_type
   USE input_constants,                 ONLY: do_thermo_al,&
                                              langevin_ensemble
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE md_ener_types,                   ONLY: md_ener_type,&
                                              release_md_ener,&
                                              retain_md_ener
   USE reftraj_types,                   ONLY: reftraj_type,&
                                              release_reftraj,&
                                              retain_reftraj
   USE simpar_types,                    ONLY: simpar_type
   USE thermal_region_types,            ONLY: release_thermal_regions,&
                                              retain_thermal_regions,&
                                              thermal_regions_type
   USE thermostat_types,                ONLY: release_thermostats,&
                                              retain_thermostats,&
                                              thermostat_type,&
                                              thermostats_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE md_environment_type
      ! para_env is the parallel enviroment of the MD,  i.e. the systems
      ! that are dealt with by the integrator e.g in the PIMD this could
      ! be parent of every bead.
      PRIVATE
      INTEGER                                    :: id_nr, ref_count
      LOGICAL                                    :: init, first_time, ehrenfest_md
      INTEGER, POINTER                           :: itimes
      REAL(KIND=dp), POINTER                     :: used_time, t
      REAL(KIND=dp), POINTER                     :: constant
      TYPE(cp_para_env_type), POINTER            :: para_env
      TYPE(cell_type), POINTER                   :: cell
      TYPE(force_env_type), POINTER              :: force_env
      TYPE(md_ener_type), POINTER                :: md_ener
      TYPE(thermostats_type), POINTER            :: thermostats
      TYPE(barostat_type), POINTER               :: barostat
      TYPE(reftraj_type), POINTER                :: reftraj
      TYPE(free_energy_type), POINTER           :: fe_env
      TYPE(simpar_type), POINTER                 :: simpar
      TYPE(average_quantities_type), POINTER     :: averages
      TYPE(thermal_regions_type), POINTER        :: thermal_regions
   END TYPE md_environment_type

   ! *** Public subroutines and data types ***
   PUBLIC :: md_environment_type, set_md_env, get_md_env, md_env_create, &
             md_env_release, need_per_atom_wiener_process

   ! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_environment_types'
   INTEGER, SAVE, PRIVATE :: last_md_env_id = 0

CONTAINS

! **************************************************************************************************
!> \brief Creates MD environment
!>      Purpose: Initialise the integrator environment.
!>      retain the para_env for this environment (should be used for parallel
!>      communications)
!> \param md_env the force environment to retain
!> \param md_section ...
!> \param para_env ...
!> \param force_env ...
! **************************************************************************************************
   SUBROUTINE md_env_create(md_env, md_section, para_env, force_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'md_env_create', routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: averages_section

      ALLOCATE (md_env)
      last_md_env_id = last_md_env_id+1
      md_env%id_nr = last_md_env_id
      md_env%ref_count = 1

      NULLIFY (md_env%itimes)
      NULLIFY (md_env%constant)
      NULLIFY (md_env%t)
      NULLIFY (md_env%cell)
      NULLIFY (md_env%simpar)
      NULLIFY (md_env%thermostats)
      NULLIFY (md_env%barostat)
      NULLIFY (md_env%force_env)
      NULLIFY (md_env%fe_env)
      NULLIFY (md_env%md_ener)
      NULLIFY (md_env%reftraj)
      NULLIFY (md_env%averages)
      NULLIFY (md_env%thermal_regions)
      md_env%para_env => para_env
      CALL cp_para_env_retain(md_env%para_env)
      ALLOCATE (md_env%itimes)
      ALLOCATE (md_env%constant)
      ALLOCATE (md_env%used_time)
      ALLOCATE (md_env%t)
      md_env%itimes = -1
      md_env%constant = 0.0_dp
      md_env%used_time = 0.0_dp
      md_env%t = 0.0_dp
      md_env%init = .TRUE.
      md_env%first_time = .TRUE.
      md_env%ehrenfest_md = .FALSE.
      averages_section => section_vals_get_subs_vals(md_section, "AVERAGES")
      CALL create_averages(md_env%averages, averages_section, force_env=force_env)

   END SUBROUTINE md_env_create

! **************************************************************************************************
!> \brief releases the given md env
!> \param md_env the md environment to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE md_env_release(md_env)
      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'md_env_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(md_env)) THEN
         IF(.NOT.(md_env%ref_count > 0))CALL cp__a("motion/md_environment_types.F",161)
         md_env%ref_count = md_env%ref_count-1
         IF (md_env%ref_count == 0) THEN
            CALL fe_env_release(md_env%fe_env)
            CALL cp_para_env_release(md_env%para_env)
            DEALLOCATE (md_env%itimes)
            DEALLOCATE (md_env%constant)
            DEALLOCATE (md_env%used_time)
            DEALLOCATE (md_env%t)

            NULLIFY (md_env%cell)
            NULLIFY (md_env%simpar)
            CALL release_barostat_type(md_env%barostat)
            CALL release_thermostats(md_env%thermostats)
            CALL release_reftraj(md_env%reftraj)
            CALL release_md_ener(md_env%md_ener)
            CALL force_env_release(md_env%force_env)
            CALL release_averages(md_env%averages)
            CALL release_thermal_regions(md_env%thermal_regions)
            DEALLOCATE (md_env)
         END IF
      END IF
   END SUBROUTINE md_env_release

! **************************************************************************************************
!> \brief get components of MD environment type
!> \param md_env the force environment to retain
!> \param itimes ...
!> \param constant ...
!> \param used_time ...
!> \param cell ...
!> \param simpar ...
!> \param npt ...
!> \param force_env ...
!> \param para_env ...
!> \param reftraj ...
!> \param t ...
!> \param init ...
!> \param first_time ...
!> \param fe_env ...
!> \param thermostats ...
!> \param barostat ...
!> \param thermostat_coeff ...
!> \param thermostat_part ...
!> \param thermostat_shell ...
!> \param thermostat_baro ...
!> \param thermostat_fast ...
!> \param thermostat_slow ...
!> \param md_ener ...
!> \param averages ...
!> \param thermal_regions ...
!> \param ehrenfest_md ...
! **************************************************************************************************
   SUBROUTINE get_md_env(md_env, itimes, constant, used_time, cell, simpar, npt, &
                         force_env, para_env, reftraj, t, init, first_time, fe_env, thermostats, barostat, &
                         thermostat_coeff, thermostat_part, thermostat_shell, thermostat_baro, &
                         thermostat_fast, thermostat_slow, md_ener, averages, &
                         thermal_regions, ehrenfest_md)

      TYPE(md_environment_type), POINTER                 :: md_env
      INTEGER, OPTIONAL, POINTER                         :: itimes
      REAL(KIND=dp), OPTIONAL, POINTER                   :: constant, used_time
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      TYPE(simpar_type), OPTIONAL, POINTER               :: simpar
      TYPE(npt_info_type), OPTIONAL, POINTER             :: npt(:, :)
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(reftraj_type), OPTIONAL, POINTER              :: reftraj
      REAL(KIND=dp), OPTIONAL, POINTER                   :: t
      LOGICAL, OPTIONAL                                  :: init, first_time
      TYPE(free_energy_type), OPTIONAL, POINTER          :: fe_env
      TYPE(thermostats_type), OPTIONAL, POINTER          :: thermostats
      TYPE(barostat_type), OPTIONAL, POINTER             :: barostat
      TYPE(thermostat_type), OPTIONAL, POINTER           :: thermostat_coeff, thermostat_part, &
                                                            thermostat_shell, thermostat_baro, &
                                                            thermostat_fast, thermostat_slow
      TYPE(md_ener_type), OPTIONAL, POINTER              :: md_ener
      TYPE(average_quantities_type), OPTIONAL, POINTER   :: averages
      TYPE(thermal_regions_type), OPTIONAL, POINTER      :: thermal_regions
      LOGICAL, OPTIONAL                                  :: ehrenfest_md

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_md_env', routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = ASSOCIATED(md_env)
      IF(.NOT.(check))CALL cp__a("motion/md_environment_types.F",247)
      IF (PRESENT(itimes)) itimes => md_env%itimes
      IF (PRESENT(fe_env)) fe_env => md_env%fe_env
      IF (PRESENT(constant)) constant => md_env%constant
      IF (PRESENT(used_time)) used_time => md_env%used_time
      IF (PRESENT(t)) t => md_env%t
      IF (PRESENT(cell)) cell => md_env%cell
      IF (PRESENT(simpar)) simpar => md_env%simpar
      IF (PRESENT(thermostats)) thermostats => md_env%thermostats
      IF (PRESENT(barostat)) barostat => md_env%barostat
      IF (PRESENT(thermostat_part) .OR. PRESENT(thermostat_coeff) .OR. &
          PRESENT(thermostat_baro) .OR. PRESENT(thermostat_shell) .OR. &
          PRESENT(thermostat_fast) .OR. PRESENT(thermostat_slow)) THEN
         IF (ASSOCIATED(md_env%thermostats)) THEN
            IF (PRESENT(thermostat_part)) THEN
               thermostat_part => md_env%thermostats%thermostat_part
            END IF
            IF (PRESENT(thermostat_coeff)) THEN
               thermostat_coeff => md_env%thermostats%thermostat_coef
            END IF
            IF (PRESENT(thermostat_shell)) THEN
               thermostat_shell => md_env%thermostats%thermostat_shell
            END IF
            IF (PRESENT(thermostat_fast)) THEN
               thermostat_fast => md_env%thermostats%thermostat_fast
            END IF
            IF (PRESENT(thermostat_slow)) THEN
               thermostat_slow => md_env%thermostats%thermostat_slow
            END IF
            IF (PRESENT(thermostat_baro)) THEN
               thermostat_baro => md_env%thermostats%thermostat_baro
            END IF
         END IF
      END IF
      IF (PRESENT(npt)) THEN
         IF (ASSOCIATED(md_env%barostat)) THEN
            npt => md_env%barostat%npt
         END IF
      END IF
      IF (PRESENT(averages)) averages => md_env%averages
      IF (PRESENT(force_env)) force_env => md_env%force_env
      IF (PRESENT(para_env)) para_env => md_env%para_env
      IF (PRESENT(reftraj)) reftraj => md_env%reftraj
      IF (PRESENT(md_ener)) md_ener => md_env%md_ener
      IF (PRESENT(init)) init = md_env%init
      IF (PRESENT(first_time)) first_time = md_env%first_time
      IF (PRESENT(ehrenfest_md)) ehrenfest_md = md_env%ehrenfest_md
      IF (PRESENT(thermal_regions)) thermal_regions => md_env%thermal_regions

   END SUBROUTINE get_md_env

! **************************************************************************************************
!> \brief Set the integrator environment to the correct program.
!> \param md_env the force environment to retain
!> \param itimes ...
!> \param constant ...
!> \param cell ...
!> \param simpar ...
!> \param fe_env ...
!> \param force_env ...
!> \param para_env ...
!> \param init ...
!> \param first_time ...
!> \param thermostats ...
!> \param barostat ...
!> \param reftraj ...
!> \param md_ener ...
!> \param averages ...
!> \param thermal_regions ...
!> \param ehrenfest_md ...
! **************************************************************************************************
   SUBROUTINE set_md_env(md_env, itimes, constant, cell, simpar, fe_env, force_env, &
                         para_env, init, first_time, thermostats, barostat, reftraj, md_ener, averages, &
                         thermal_regions, ehrenfest_md)

      TYPE(md_environment_type), POINTER                 :: md_env
      INTEGER, OPTIONAL, POINTER                         :: itimes
      REAL(KIND=dp), OPTIONAL, POINTER                   :: constant
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      TYPE(simpar_type), OPTIONAL, POINTER               :: simpar
      TYPE(free_energy_type), OPTIONAL, POINTER          :: fe_env
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      LOGICAL, OPTIONAL                                  :: init, first_time
      TYPE(thermostats_type), OPTIONAL, POINTER          :: thermostats
      TYPE(barostat_type), OPTIONAL, POINTER             :: barostat
      TYPE(reftraj_type), OPTIONAL, POINTER              :: reftraj
      TYPE(md_ener_type), OPTIONAL, POINTER              :: md_ener
      TYPE(average_quantities_type), OPTIONAL, POINTER   :: averages
      TYPE(thermal_regions_type), OPTIONAL, POINTER      :: thermal_regions
      LOGICAL, OPTIONAL                                  :: ehrenfest_md

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_md_env', routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = ASSOCIATED(md_env)
      IF(.NOT.(check))CALL cp__a("motion/md_environment_types.F",344)
      IF (PRESENT(init)) md_env%init = init
      IF (PRESENT(first_time)) md_env%first_time = first_time
      IF (PRESENT(ehrenfest_md)) md_env%ehrenfest_md = ehrenfest_md
      IF (PRESENT(cell)) md_env%cell => cell
      IF (PRESENT(barostat)) THEN
         CALL release_barostat_type(md_env%barostat)
         CALL retain_barostat_type(barostat)
         md_env%barostat => barostat
      END IF
      IF (PRESENT(thermostats)) THEN
         CALL release_thermostats(md_env%thermostats)
         CALL retain_thermostats(thermostats)
         md_env%thermostats => thermostats
      END IF
      IF (PRESENT(simpar)) md_env%simpar => simpar
      IF (PRESENT(itimes)) md_env%itimes => itimes
      IF (PRESENT(fe_env)) md_env%fe_env => fe_env
      IF (PRESENT(constant)) md_env%constant => constant
      IF (PRESENT(para_env)) md_env%para_env => para_env
      IF (PRESENT(force_env)) THEN
         IF (ASSOCIATED(force_env)) THEN
            CALL force_env_retain(force_env)
         END IF
         IF (ASSOCIATED(md_env%force_env)) THEN
            CALL force_env_release(md_env%force_env)
         END IF
         md_env%force_env => force_env
      END IF
      IF (PRESENT(reftraj)) THEN
         CALL release_reftraj(md_env%reftraj)
         CALL retain_reftraj(reftraj)
         md_env%reftraj => reftraj
      END IF
      IF (PRESENT(md_ener)) THEN
         CALL release_md_ener(md_env%md_ener)
         CALL retain_md_ener(md_ener)
         md_env%md_ener => md_ener
      END IF
      IF (PRESENT(averages)) THEN
         CALL release_averages(md_env%averages)
         CALL retain_averages(averages)
         md_env%averages => averages
      END IF
      IF (PRESENT(thermal_regions)) THEN
         CALL release_thermal_regions(md_env%thermal_regions)
         CALL retain_thermal_regions(thermal_regions)
         md_env%thermal_regions => thermal_regions
      END IF

   END SUBROUTINE set_md_env

! **************************************************************************************************
!> \brief ...
!> \param md_env ...
!> \retval need_per_atom_wiener_process ...
!> \par History
!>      02.2012 created [noamb]
!> \author Noam Bernstein
! **************************************************************************************************
   FUNCTION need_per_atom_wiener_process(md_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      LOGICAL                                            :: need_per_atom_wiener_process

! return value
! check for Langevin ensemble

      need_per_atom_wiener_process = (md_env%simpar%ensemble == langevin_ensemble)
      IF (need_per_atom_wiener_process) RETURN

      ! check for adaptive-Langevin thermostat
      IF (.NOT. ASSOCIATED(md_env%thermostats)) RETURN
      IF (.NOT. ASSOCIATED(md_env%thermostats%thermostat_part)) RETURN
      need_per_atom_wiener_process = (md_env%thermostats%thermostat_part%type_of_thermostat == do_thermo_al)

   END FUNCTION need_per_atom_wiener_process

END MODULE md_environment_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE input_restart_rng
   USE cp_linked_list_val,              ONLY: cp_sll_val_create,&
                                              cp_sll_val_get_length,&
                                              cp_sll_val_type
   USE input_section_types,             ONLY: section_get_keyword_index,&
                                              section_type,&
                                              section_vals_add_values,&
                                              section_vals_type
   USE input_val_types,                 ONLY: val_create,&
                                              val_release,&
                                              val_type
   USE parallel_rng_types,              ONLY: rng_record_length
   USE string_utilities,                ONLY: ascii_to_string
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_restart_rng'

   PUBLIC :: section_rng_val_set

CONTAINS

! **************************************************************************************************
!> \brief routine to dump rngs.. fast implementation
!> \param rng_section ...
!> \param nsize ...
!> \param ascii ...
!> \par History
!>      02.2006 created [teo]
!>      - string dump (again) instead of integer ASCII code (07.03.06,MK)
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE section_rng_val_set(rng_section, nsize, ascii)

      TYPE(section_vals_type), POINTER                   :: rng_section
      INTEGER, INTENT(IN)                                :: nsize
      INTEGER, DIMENSION(:, :)                           :: ascii

      CHARACTER(LEN=*), PARAMETER :: routineN = 'section_rng_val_set', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: ik, irk, Nlist
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      IF(.NOT.(ASSOCIATED(rng_section)))CALL cp__a("input_restart_rng.F",56)
      IF(.NOT.(rng_section%ref_count > 0))CALL cp__a("input_restart_rng.F",57)

      NULLIFY (my_val, old_val, section, vals)

      section => rng_section%section

      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")

      IF (ik == -2) &
         CALL cp_abort(cp__l("input_restart_rng.F",66), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      DO
         IF (SIZE(rng_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(rng_section)
      END DO

      vals => rng_section%values(ik, 1)%list
      Nlist = 0

      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF

      DO irk = 1, nsize

         CALL ascii_to_string(ascii(:, irk), rng_record)
         CALL val_create(val=my_val, lc_val=rng_record)

         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF

         NULLIFY (my_val)

      END DO

      rng_section%values(ik, 1)%list => vals

   END SUBROUTINE section_rng_val_set

END MODULE input_restart_rng
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE input_restart_force_eval
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              real_to_scaled
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_linked_list_val,              ONLY: cp_sll_val_create,&
                                              cp_sll_val_get_length,&
                                              cp_sll_val_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE eip_environment_types,           ONLY: eip_env_get
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              multiple_fe_list,&
                                              use_eip_force,&
                                              use_qmmm,&
                                              use_qs_force
   USE input_constants,                 ONLY: do_coord_cp2k,&
                                              ehrenfest,&
                                              mol_dyn_run,&
                                              mon_car_run,&
                                              pint_run,&
                                              use_rt_restart
   USE input_cp2k_restarts_util,        ONLY: section_velocity_val_set
   USE input_restart_rng,               ONLY: section_rng_val_set
   USE input_section_types,             ONLY: &
        section_get_keyword_index, section_type, section_vals_add_values, section_vals_get, &
        section_vals_get_subs_vals, section_vals_get_subs_vals3, section_vals_remove_values, &
        section_vals_type, section_vals_val_get, section_vals_val_set, section_vals_val_unset
   USE input_val_types,                 ONLY: val_create,&
                                              val_release,&
                                              val_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_sum
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: get_molecule_kind
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE multipole_types,                 ONLY: multipole_type
   USE parallel_rng_types,              ONLY: dump_rng_stream,&
                                              rng_record_length
   USE particle_list_types,             ONLY: particle_list_type
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qs_environment_types,            ONLY: get_qs_env
   USE string_utilities,                ONLY: string_to_ascii
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_restart_force_eval'

   PUBLIC :: update_force_eval, update_subsys

CONTAINS

! **************************************************************************************************
!> \brief Updates the force_eval section of the input file
!> \param force_env ...
!> \param root_section ...
!> \param write_binary_restart_file ...
!> \param respa ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_force_eval(force_env, root_section, &
                                write_binary_restart_file, respa)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      LOGICAL, INTENT(IN)                                :: write_binary_restart_file
      LOGICAL, OPTIONAL                                  :: respa

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_force_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_subsys, iforce_eval, myid, nforce_eval
      INTEGER, DIMENSION(:), POINTER                     :: i_force_eval
      LOGICAL                                            :: multiple_subsys, skip_vel_section
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: cell_section, force_env_sections, &
                                                            qmmm_section, rng_section, &
                                                            subsys_section
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (rng_section, subsys_section, cell_section, virial, subsys, cell)
      ! If it's not a dynamical run don't update the velocity section
      CALL section_vals_val_get(root_section, "GLOBAL%RUN_TYPE", i_val=myid)
      skip_vel_section = ((myid /= mol_dyn_run) .AND. &
                          (myid /= mon_car_run) .AND. &
                          (myid /= pint_run) .AND. &
                          (myid /= ehrenfest))

      ! Go on updatig the force_env_section
      force_env_sections => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
      CALL multiple_fe_list(force_env_sections, root_section, i_force_eval, nforce_eval)
      ! The update of the input MUST be realized only on the main force_eval
      ! All the others will be left not updated because there is no real need to update them...
      iforce_eval = 1
      subsys_section => section_vals_get_subs_vals3(force_env_sections, "SUBSYS", &
                                                    i_rep_section=i_force_eval(iforce_eval))
      CALL update_subsys(subsys_section, force_env, skip_vel_section, &
                         write_binary_restart_file)

      ! For RESPA we need to update all subsystems
      IF (PRESENT(respa)) THEN
         IF (respa) THEN
            DO i_subsys = 1, 2
               iforce_eval = i_subsys
               subsys_section => section_vals_get_subs_vals3(force_env_sections, "SUBSYS", &
                                                             i_rep_section=i_force_eval(iforce_eval))
               CALL update_subsys(subsys_section, force_env, skip_vel_section, &
                                  write_binary_restart_file)
            ENDDO
         ENDIF
      END IF

      rng_section => section_vals_get_subs_vals(subsys_section, "RNG_INIT")
      CALL update_rng_particle(rng_section, force_env)

      qmmm_section => section_vals_get_subs_vals3(force_env_sections, "QMMM", &
                                                  i_rep_section=i_force_eval(iforce_eval))
      CALL update_qmmm(qmmm_section, force_env)

      ! And now update only the cells of all other force_eval
      ! This is to make consistent for cell variable runs..
      IF (nforce_eval > 1) THEN
         CALL force_env_get(force_env, subsys=subsys, cell=cell)
         CALL cp_subsys_get(subsys, virial=virial)
         CALL section_vals_val_get(root_section, "MULTIPLE_FORCE_EVALS%MULTIPLE_SUBSYS", l_val=multiple_subsys)
         IF (virial%pv_availability .AND. multiple_subsys) THEN
            DO iforce_eval = 2, nforce_eval
               subsys_section => section_vals_get_subs_vals3(force_env_sections, "SUBSYS", &
                                                             i_rep_section=i_force_eval(iforce_eval))
               cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
               CALL update_cell_section(cell, cell_section)
            END DO
         END IF
      END IF

      IF (myid == ehrenfest) CALL section_vals_val_set(root_section, "FORCE_EVAL%DFT%REAL_TIME_PROPAGATION%INITIAL_WFN", &
                                                       i_val=use_rt_restart)
      DEALLOCATE (i_force_eval)

   END SUBROUTINE update_force_eval

! **************************************************************************************************
!> \brief Updates the qmmm section if needed
!> \param qmmm_section ...
!> \param force_env ...
!> \par History
!>      08.2007 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_qmmm(qmmm_section, force_env)
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_qmmm', routineP = moduleN//':'//routineN

      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work

      SELECT CASE (force_env%in_use)
      CASE (use_qmmm)
         CALL section_vals_get(qmmm_section, explicit=explicit)
         IF(.NOT.(explicit))CALL cp__a("input_restart_force_eval.F",181)

         ALLOCATE (work(3))
         work = force_env%qmmm_env%qm%transl_v
         CALL section_vals_val_set(qmmm_section, "INITIAL_TRANSLATION_VECTOR", r_vals_ptr=work)
      END SELECT

   END SUBROUTINE update_qmmm

! **************************************************************************************************
!> \brief Updates the rng section of the input file
!>      Write current status of the parallel random number generator (RNG)
!> \param rng_section ...
!> \param force_env ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_rng_particle(rng_section, force_env)

      TYPE(section_vals_type), POINTER                   :: rng_section
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_rng_particle', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle, &
                                                            nparticle_kind, nparticle_local
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ascii
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_list_type), POINTER                  :: particles

      CALL force_env_get(force_env, subsys=subsys, para_env=para_env)
      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                         particles=particles)

      IF (ASSOCIATED(local_particles%local_particle_set)) THEN
         nparticle_kind = atomic_kinds%n_els
         nparticle = particles%n_els

         ALLOCATE (ascii(rng_record_length, nparticle))
         ascii = 0

         DO iparticle = 1, nparticle
            DO iparticle_kind = 1, nparticle_kind
               nparticle_local = local_particles%n_el(iparticle_kind)
               DO iparticle_local = 1, nparticle_local
                  IF (iparticle == local_particles%list(iparticle_kind)%array(iparticle_local)) THEN
                     CALL dump_rng_stream(rng_stream=local_particles%local_particle_set(iparticle_kind)% &
                                          rng(iparticle_local)%stream, &
                                          rng_record=rng_record)
                     CALL string_to_ascii(rng_record, ascii(:, iparticle))
                  END IF
               END DO
            END DO
         END DO

         CALL mp_sum(ascii, para_env%group)

         CALL section_rng_val_set(rng_section=rng_section, nsize=nparticle, ascii=ascii)

         DEALLOCATE (ascii)
      END IF

   END SUBROUTINE update_rng_particle

! **************************************************************************************************
!> \brief Updates the subsys section of the input file
!> \param subsys_section ...
!> \param force_env ...
!> \param skip_vel_section ...
!> \param write_binary_restart_file ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_subsys(subsys_section, force_env, skip_vel_section, &
                            write_binary_restart_file)
      TYPE(section_vals_type), POINTER                   :: subsys_section
      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL, INTENT(IN)                                :: skip_vel_section, &
                                                            write_binary_restart_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_subsys', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: coord_file_name, unit_str
      INTEGER                                            :: coord_file_format, handle, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: multiple_unit_cell
      LOGICAL                                            :: scale, use_ref_cell
      REAL(KIND=dp)                                      :: conv_factor
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(section_vals_type), POINTER                   :: work_section

      NULLIFY (work_section, core_particles, particles, shell_particles, &
               subsys, cell, para_env, multipoles)
      CALL timeset(routineN, handle)
      CALL force_env_get(force_env, subsys=subsys, cell=cell, para_env=para_env)

      CALL cp_subsys_get(subsys, particles=particles, molecules_new=molecules, &
                         shell_particles=shell_particles, core_particles=core_particles, &
                         multipoles=multipoles)

      ! Remove the multiple_unit_cell from the input structure.. at this point we have
      ! already all the information we need..
      ALLOCATE (multiple_unit_cell(3))
      multiple_unit_cell = 1
      CALL section_vals_val_set(subsys_section, "TOPOLOGY%MULTIPLE_UNIT_CELL", &
                                i_vals_ptr=multiple_unit_cell)

      ! Coordinates and Velocities
      work_section => section_vals_get_subs_vals(subsys_section, "COORD")
      CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
      CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
      conv_factor = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (.NOT. write_binary_restart_file) THEN
         CALL section_vals_val_get(subsys_section, "TOPOLOGY%COORD_FILE_FORMAT", &
                                   i_val=coord_file_format)
         IF (coord_file_format == do_coord_cp2k) THEN
            CALL section_vals_val_get(subsys_section, "TOPOLOGY%COORD_FILE_NAME", &
                                      c_val=coord_file_name)
            output_unit = 0
            IF (para_env%ionode) THEN
               CALL open_file(file_name=TRIM(ADJUSTL(coord_file_name)), &
                              file_action="READWRITE", &
                              file_form="FORMATTED", &
                              file_position="REWIND", &
                              file_status="UNKNOWN", &
                              unit_number=output_unit)
               CALL dump_coordinates_cp2k(particles, molecules, cell, conv_factor, &
                                          output_unit=output_unit, &
                                          core_or_shell=.FALSE., &
                                          scaled_coordinates=scale)
               CALL close_file(unit_number=output_unit)
            END IF
         ELSE
            CALL section_coord_val_set(work_section, particles, molecules, conv_factor, scale, &
                                       cell)
         END IF
      END IF
      CALL section_vals_val_set(subsys_section, "TOPOLOGY%NUMBER_OF_ATOMS", i_val=particles%n_els)
      work_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
      IF (.NOT. skip_vel_section) THEN
         IF (.NOT. write_binary_restart_file) THEN
            CALL section_velocity_val_set(work_section, particles, conv_factor=1.0_dp)
         END IF
      ELSE
         CALL section_vals_remove_values(work_section)
      END IF

      ! Write restart input for core-shell model
      IF (.NOT. write_binary_restart_file) THEN
         IF (ASSOCIATED(shell_particles)) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "SHELL_COORD")
            CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
            CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
            conv_factor = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
            CALL section_coord_val_set(work_section, shell_particles, molecules, &
                                       conv_factor, scale, cell, shell=.TRUE.)
            IF (.NOT. skip_vel_section) THEN
               work_section => section_vals_get_subs_vals(subsys_section, "SHELL_VELOCITY")
               CALL section_velocity_val_set(work_section, shell_particles, conv_factor=1.0_dp)
            END IF
         END IF
         IF (ASSOCIATED(core_particles)) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "CORE_COORD")
            CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
            CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
            conv_factor = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
            CALL section_coord_val_set(work_section, core_particles, molecules, &
                                       conv_factor, scale, cell, shell=.TRUE.)
            IF (.NOT. skip_vel_section) THEN
               work_section => section_vals_get_subs_vals(subsys_section, "CORE_VELOCITY")
               CALL section_velocity_val_set(work_section, core_particles, conv_factor=1.0_dp)
            END IF
         END IF
      END IF

      ! Updating cell info
      CALL force_env_get(force_env, cell=cell)
      work_section => section_vals_get_subs_vals(subsys_section, "CELL")
      CALL update_cell_section(cell, cell_section=work_section)

      ! Updating cell_ref info
      use_ref_cell = .FALSE.
      SELECT CASE (force_env%in_use)
      CASE (use_qs_force)
         CALL get_qs_env(force_env%qs_env, cell_ref=cell, use_ref_cell=use_ref_cell)
      CASE (use_eip_force)
         CALL eip_env_get(force_env%eip_env, cell_ref=cell, use_ref_cell=use_ref_cell)
      END SELECT
      IF (use_ref_cell) THEN
         work_section => section_vals_get_subs_vals(subsys_section, "CELL%CELL_REF")
         CALL update_cell_section(cell, cell_section=work_section)
      ENDIF

      ! Updating multipoles
      IF (ASSOCIATED(multipoles)) THEN
         work_section => section_vals_get_subs_vals(subsys_section, "MULTIPOLES")
         DO
            IF (SIZE(work_section%values, 2) == 1) EXIT
            CALL section_vals_add_values(work_section)
         END DO
         IF (ASSOCIATED(multipoles%dipoles)) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "MULTIPOLES%DIPOLES")
            CALL update_dipoles_section(multipoles%dipoles, work_section)
         END IF
         IF (ASSOCIATED(multipoles%quadrupoles)) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "MULTIPOLES%QUADRUPOLES")
            CALL update_quadrupoles_section(multipoles%quadrupoles, work_section)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE update_subsys

! **************************************************************************************************
!> \brief Routine to update a cell section
!> \param cell ...
!> \param cell_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE update_cell_section(cell, cell_section)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: cell_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_cell_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(:), POINTER                     :: iwork
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work

      NULLIFY (work, iwork)
      CALL timeset(routineN, handle)

      ! CELL VECTORS - A
      ALLOCATE (work(3))
      work(1:3) = cell%hmat(1:3, 1)
      CALL section_vals_val_set(cell_section, "A", r_vals_ptr=work)

      ! CELL VECTORS - B
      ALLOCATE (work(3))
      work(1:3) = cell%hmat(1:3, 2)
      CALL section_vals_val_set(cell_section, "B", r_vals_ptr=work)

      ! CELL VECTORS - C
      ALLOCATE (work(3))
      work(1:3) = cell%hmat(1:3, 3)
      CALL section_vals_val_set(cell_section, "C", r_vals_ptr=work)

      ! MULTIPLE_UNIT_CELL
      ALLOCATE (iwork(3))
      iwork = 1
      CALL section_vals_val_set(cell_section, "MULTIPLE_UNIT_CELL", i_vals_ptr=iwork)

      ! Unset unused or misleading information
      CALL section_vals_val_unset(cell_section, "ABC")
      CALL section_vals_val_unset(cell_section, "ALPHA_BETA_GAMMA")

      CALL timestop(handle)
   END SUBROUTINE update_cell_section

! **************************************************************************************************
!> \brief routine to dump coordinates.. fast implementation
!> \param coord_section ...
!> \param particles ...
!> \param molecules ...
!> \param conv_factor ...
!> \param scale ...
!> \param cell ...
!> \param shell ...
!> \par History
!>      02.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE section_coord_val_set(coord_section, particles, molecules, conv_factor, &
                                    scale, cell, shell)
      TYPE(section_vals_type), POINTER                   :: coord_section
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(mol_new_list_type), POINTER                   :: molecules
      REAL(KIND=dp)                                      :: conv_factor
      LOGICAL, INTENT(IN)                                :: scale
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'section_coord_val_set', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2*default_string_length)             :: line
      CHARACTER(LEN=default_string_length)               :: my_tag, name
      INTEGER                                            :: handle, ik, imol, irk, last_atom, Nlist
      LOGICAL                                            :: ldum, molname_generated, my_shell
      REAL(KIND=dp), DIMENSION(3)                        :: r, s
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(molecule_type), POINTER                       :: molecule_now
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      CALL timeset(routineN, handle)

      NULLIFY (my_val, old_val, section, vals)
      my_shell = .FALSE.
      IF (PRESENT(shell)) my_shell = shell
      IF(.NOT.(ASSOCIATED(coord_section)))CALL cp__a("input_restart_force_eval.F",494)
      IF(.NOT.(coord_section%ref_count > 0))CALL cp__a("input_restart_force_eval.F",495)
      section => coord_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("input_restart_force_eval.F",499), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      DO
         IF (SIZE(coord_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(coord_section)
      END DO
      vals => coord_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      imol = 0
      last_atom = 0
      DO irk = 1, particles%n_els
         CALL get_atomic_kind(particles%els(irk)%atomic_kind, name=name)
         IF (my_shell) THEN
            s = particles%els(irk)%r
            IF (scale) THEN
               r = s
               CALL real_to_scaled(s, r, cell)
            ELSE
               s = s*conv_factor
            END IF
            WRITE (UNIT=line, FMT="(A,3(1X,ES25.16),1X,I0)") &
               TRIM(name), s(1:3), particles%els(irk)%atom_index
            CALL val_create(my_val, lc_val=line)
            IF (Nlist /= 0) THEN
               IF (irk == 1) THEN
                  new_pos => vals
               ELSE
                  new_pos => new_pos%rest
               END IF
               old_val => new_pos%first_el
               CALL val_release(old_val)
               new_pos%first_el => my_val
            ELSE
               IF (irk == 1) THEN
                  NULLIFY (new_pos)
                  CALL cp_sll_val_create(new_pos, first_el=my_val)
                  vals => new_pos
               ELSE
                  NULLIFY (new_pos%rest)
                  CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
                  new_pos => new_pos%rest
               END IF
            END IF
            NULLIFY (my_val)
         ELSE
            IF (last_atom < irk) THEN
               imol = imol+1
               molecule_now => molecules%els(imol)
               CALL get_molecule(molecule_now, last_atom=last_atom)
               CALL get_molecule_kind(molecule_now%molecule_kind, molname_generated=molname_generated, &
                                      name=my_tag)
               IF (molname_generated) my_tag = ""
            END IF
            ldum = qmmm_ff_precond_only_qm(my_tag)
            ldum = qmmm_ff_precond_only_qm(name)
            s = particles%els(irk)%r
            IF (scale) THEN
               r = s
               CALL real_to_scaled(s, r, cell)
            ELSE
               s = s*conv_factor
            END IF
            IF (LEN_TRIM(my_tag) > 0) THEN
               WRITE (UNIT=line, FMT="(A,3(1X,ES25.16),1X,A,1X,I0)") &
                  TRIM(name), s(1:3), TRIM(my_tag), imol
            ELSE
               WRITE (UNIT=line, FMT="(A,3(1X,ES25.16))") &
                  TRIM(name), s(1:3)
            END IF
            CALL val_create(my_val, lc_val=line)

            IF (Nlist /= 0) THEN
               IF (irk == 1) THEN
                  new_pos => vals
               ELSE
                  new_pos => new_pos%rest
               END IF
               old_val => new_pos%first_el
               CALL val_release(old_val)
               new_pos%first_el => my_val
            ELSE
               IF (irk == 1) THEN
                  NULLIFY (new_pos)
                  CALL cp_sll_val_create(new_pos, first_el=my_val)
                  vals => new_pos
               ELSE
                  NULLIFY (new_pos%rest)
                  CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
                  new_pos => new_pos%rest
               END IF
            END IF
            NULLIFY (my_val)
         END IF
      END DO
      coord_section%values(ik, 1)%list => vals

      CALL timestop(handle)
   END SUBROUTINE section_coord_val_set

! **************************************************************************************************
!> \brief routine to dump dipoles.. fast implementation
!> \param dipoles ...
!> \param dipoles_section ...
!> \par History
!>      12.2007 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_dipoles_section(dipoles, dipoles_section)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dipoles
      TYPE(section_vals_type), POINTER                   :: dipoles_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_dipoles_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, irk, Nlist, nloop
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      CALL timeset(routineN, handle)
      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(dipoles_section)))CALL cp__a("input_restart_force_eval.F",627)
      IF(.NOT.(dipoles_section%ref_count > 0))CALL cp__a("input_restart_force_eval.F",628)
      section => dipoles_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("input_restart_force_eval.F",632), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      ! At least one of the two arguments must be present..
      nloop = SIZE(dipoles, 2)
      DO
         IF (SIZE(dipoles_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(dipoles_section)
      END DO
      vals => dipoles_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      DO irk = 1, nloop
         ALLOCATE (work(3))
         ! Always stored in A.U.
         work = dipoles(1:3, irk)
         CALL val_create(my_val, r_vals_ptr=work)

         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      dipoles_section%values(ik, 1)%list => vals

      CALL timestop(handle)
   END SUBROUTINE update_dipoles_section

! **************************************************************************************************
!> \brief routine to dump quadrupoles.. fast implementation
!> \param quadrupoles ...
!> \param quadrupoles_section ...
!> \par History
!>      12.2007 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_quadrupoles_section(quadrupoles, quadrupoles_section)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: quadrupoles
      TYPE(section_vals_type), POINTER                   :: quadrupoles_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_quadrupoles_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ik, ind, irk, j, Nlist, nloop
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      CALL timeset(routineN, handle)
      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(quadrupoles_section)))CALL cp__a("input_restart_force_eval.F",704)
      IF(.NOT.(quadrupoles_section%ref_count > 0))CALL cp__a("input_restart_force_eval.F",705)
      section => quadrupoles_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("input_restart_force_eval.F",709), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      ! At least one of the two arguments must be present..
      nloop = SIZE(quadrupoles, 2)
      DO
         IF (SIZE(quadrupoles_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(quadrupoles_section)
      END DO
      vals => quadrupoles_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      DO irk = 1, nloop
         ALLOCATE (work(6))
         ! Always stored in A.U.
         ind = 0
         DO i = 1, 3
            DO j = i, 3
               ind = ind+1
               work(ind) = quadrupoles(j, i, irk)
            END DO
         END DO
         CALL val_create(my_val, r_vals_ptr=work)

         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      quadrupoles_section%values(ik, 1)%list => vals

      CALL timestop(handle)
   END SUBROUTINE update_quadrupoles_section

! **************************************************************************************************
!> \brief   Dump atomic, core, or shell coordinates to a file in CP2K &COORD
!>          section format
!> \param particles ...
!> \param molecules ...
!> \param cell ...
!> \param conv_factor ...
!> \param output_unit ...
!> \param core_or_shell ...
!> \param scaled_coordinates ...
!> \par History
!>      07.02.2011 (Creation, MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dump_coordinates_cp2k(particles, molecules, cell, conv_factor, &
                                    output_unit, core_or_shell, &
                                    scaled_coordinates)

      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(IN)                          :: conv_factor
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: core_or_shell, scaled_coordinates

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_coordinates_cp2k', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: kind_name, tag
      INTEGER                                            :: handle, imolecule, iparticle, last_atom
      LOGICAL                                            :: ldum, molname_generated
      REAL(KIND=dp), DIMENSION(3)                        :: r, s
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(particles)))CALL cp__a("input_restart_force_eval.F",800)
      IF (.NOT. core_or_shell) THEN
         IF(.NOT.(ASSOCIATED(molecules)))CALL cp__a("input_restart_force_eval.F",802)
      END IF
      IF (scaled_coordinates) THEN
         IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("input_restart_force_eval.F",805)
      END IF

      kind_name = ""
      tag = ""
      imolecule = 0
      last_atom = 0
      DO iparticle = 1, particles%n_els
         CALL get_atomic_kind(particles%els(iparticle)%atomic_kind, name=kind_name)
         IF (.NOT. core_or_shell) THEN
            IF (iparticle > last_atom) THEN
               imolecule = imolecule+1
               molecule => molecules%els(imolecule)
               CALL get_molecule(molecule, last_atom=last_atom)
               CALL get_molecule_kind(molecule%molecule_kind, &
                                      molname_generated=molname_generated, &
                                      name=tag)
               IF (molname_generated) tag = ""
            END IF
            ldum = qmmm_ff_precond_only_qm(tag)
            ldum = qmmm_ff_precond_only_qm(kind_name)
         END IF
         IF (scaled_coordinates) THEN
            CALL real_to_scaled(s, particles%els(iparticle)%r, cell)
            r(1:3) = s(1:3)
         ELSE
            r(1:3) = particles%els(iparticle)%r(1:3)*conv_factor
         END IF
         IF (core_or_shell) THEN
            WRITE (UNIT=output_unit, FMT="(A,3(1X,ES25.16),1X,I0)") &
               TRIM(ADJUSTL(kind_name)), r(1:3), particles%els(iparticle)%atom_index
         ELSE
            IF (LEN_TRIM(tag) > 0) THEN
               WRITE (UNIT=output_unit, FMT="(A,3(1X,ES25.16),1X,A,1X,I0)") &
                  TRIM(ADJUSTL(kind_name)), r(1:3), TRIM(tag), imolecule
            ELSE
               WRITE (UNIT=output_unit, FMT="(A,3(1X,ES25.16))") &
                  TRIM(ADJUSTL(kind_name)), r(1:3)
            END IF
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE dump_coordinates_cp2k

END MODULE input_restart_force_eval
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Data type and methods dealing with PI calcs in staging coordinates
!> \author fawzi
!> \par    History
!>         2006-02 created
!>         2006-11 modified so it might actually work [hforbert]
!>         2009-04-07 moved from pint_types module to a separate file [lwalewski]
! **************************************************************************************************
MODULE pint_staging
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE pint_types,                      ONLY: staging_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_staging'

   INTEGER, SAVE, PRIVATE :: last_staging_id = 0

   PUBLIC :: staging_env_create, staging_release, &
             staging_init_masses, &
             staging_x2u, staging_u2x, staging_f2uf, &
             staging_calc_uf_h

CONTAINS

! ***************************************************************************
!> \brief creates the data needed for a staging transformation
!> \param staging_env ...
!> \param staging_section ...
!> \param p ...
!> \param kT ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE staging_env_create(staging_env, staging_section, p, kT)
      TYPE(staging_env_type), POINTER                    :: staging_env
      TYPE(section_vals_type), POINTER                   :: staging_section
      INTEGER, INTENT(in)                                :: p
      REAL(kind=dp), INTENT(in)                          :: kT

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(staging_env)))CALL cp__a("motion/pint_staging.F",53)
      ALLOCATE (staging_env)
      last_staging_id = last_staging_id+1
      staging_env%id_nr = last_staging_id
      staging_env%ref_count = 1

      CALL section_vals_val_get(staging_section, "j", i_val=staging_env%j)
      CALL section_vals_val_get(staging_section, "Q_end", i_val=staging_env%j)
      staging_env%p = p
      staging_env%nseg = staging_env%p/staging_env%j

      staging_env%w_p = SQRT(REAL(staging_env%p, dp))*kT
      staging_env%w_j = kT*SQRT(REAL(staging_env%nseg, dp))
      staging_env%Q_stage = kT/staging_env%w_p**2
      IF (staging_env%Q_end <= 0._dp) THEN
         staging_env%Q_end = staging_env%j*staging_env%Q_stage
      END IF
      RETURN
   END SUBROUTINE staging_env_create

! ***************************************************************************
!> \brief releases the staging environment
!> \param staging_env the staging_env to release
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE staging_release(staging_env)
      TYPE(staging_env_type), POINTER                    :: staging_env

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(staging_env)) THEN
         IF(.NOT.(staging_env%ref_count > 0))CALL cp__a("motion/pint_staging.F",85)
         staging_env%ref_count = staging_env%ref_count-1
         IF (staging_env%ref_count == 0) THEN
            DEALLOCATE (staging_env)
         END IF
      END IF
      NULLIFY (staging_env)
      RETURN
   END SUBROUTINE staging_release

! ***************************************************************************
!> \brief retains a staging_env
!> \param staging_env the staging_env to retain
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE staging_retain(staging_env)
      TYPE(staging_env_type), POINTER                    :: staging_env

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(staging_env)))CALL cp__a("motion/pint_staging.F",105)
      IF(.NOT.(staging_env%ref_count > 0))CALL cp__a("motion/pint_staging.F",106)
      staging_env%ref_count = staging_env%ref_count+1
      RETURN
   END SUBROUTINE staging_retain

! ***************************************************************************
!> \brief initializes the masses and fictitious masses compatibly with the
!>      staging information
!> \param staging_env the definition of the staging transformation
!> \param mass *input* the masses of the particles
!> \param mass_beads masses of the beads
!> \param mass_fict the fictitious masses
!> \param Q masses of the nose thermostats
!> \par History
!>      11.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE staging_init_masses(staging_env, mass, mass_beads, mass_fict, &
                                  Q)
      TYPE(staging_env_type), POINTER                    :: staging_env
      REAL(kind=dp), DIMENSION(:), INTENT(in)            :: mass
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL                                        :: mass_beads, mass_fict
      REAL(kind=dp), DIMENSION(:), INTENT(out), OPTIONAL :: Q

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_init_masses', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ib, iseg
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: scal

      IF (PRESENT(Q)) THEN
         Q = staging_env%Q_stage
         DO i = 1, staging_env%p, staging_env%j
            Q(i) = staging_env%Q_end
         END DO
      END IF
      IF (PRESENT(mass_beads) .OR. PRESENT(mass_fict)) THEN

         ALLOCATE (scal(staging_env%p))
         DO iseg = 1, staging_env%nseg
            DO i = 1, staging_env%j ! check order!!!
               scal(staging_env%j*(iseg-1)+i) = REAL(i, dp)/REAL(MAX(1, i-1), dp)
            END DO
         END DO
         !   scal=zeros(staging_env%j,Float64)
         !   divide(arange(2,staging_env%j+1,typecode=Float64),
         !          arange(1,staging_env%j,typecode=Float64),scal[1:])
         !   scal[0]=1.
         !   scal=outerproduct(ones(staging_env%nseg),scal)

         IF (PRESENT(mass_beads)) THEN
            DO iat = 1, SIZE(mass)
               DO ib = 1, staging_env%p
                  mass_beads(ib, iat) = scal(ib)*mass(iat)
               END DO
            END DO
         END IF
         IF (PRESENT(mass_fict)) THEN
            DO iat = 1, SIZE(mass)
               DO ib = 1, staging_env%p
                  mass_fict(ib, iat) = scal(ib)*mass(iat)
               END DO
            END DO
         END IF
      END IF
      RETURN
   END SUBROUTINE staging_init_masses

! ***************************************************************************
!> \brief Transforms from the x into the u variables using a staging
!>      transformation for the positions
!> \param staging_env the environment for the staging transformation
!> \param ux will contain the u variable
!> \param x the positions to transform
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE staging_x2u(staging_env, ux, x)
      TYPE(staging_env_type), POINTER                    :: staging_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_x2u', routineP = moduleN//':'//routineN

      INTEGER                                            :: k, s

      IF(.NOT.(ASSOCIATED(staging_env)))CALL cp__a("motion/pint_staging.F",192)
      IF(.NOT.(staging_env%ref_count > 0))CALL cp__a("motion/pint_staging.F",193)
      ux = x
      DO s = 0, staging_env%nseg-1
         DO k = 2, staging_env%j
            ux(staging_env%j*s+k, :) = ux(staging_env%j*s+k, :) &
                                       -((REAL(k-1, dp)/REAL(k, dp) &
                                          *x(MODULO((staging_env%j*s+k+1), staging_env%p), :)+ &
                                          x(staging_env%j*s+1, :)/REAL(k, dp)))
         END DO
      END DO
      RETURN
   END SUBROUTINE staging_x2u

! ***************************************************************************
!> \brief transform from the u variable to the x (back staging transformation
!>      for the positions)
!> \param staging_env the environment for the staging transformation
!> \param ux the u variable (positions to be backtransformed)
!> \param x will contain the positions
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE staging_u2x(staging_env, ux, x)
      TYPE(staging_env_type), POINTER                    :: staging_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_u2x', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ist, j
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii, jjj
      REAL(kind=dp)                                      :: const, const2

      IF(.NOT.(ASSOCIATED(staging_env)))CALL cp__a("motion/pint_staging.F",225)
      IF(.NOT.(staging_env%ref_count > 0))CALL cp__a("motion/pint_staging.F",226)
      j = staging_env%j
      const = REAL(j-1, dp)/REAL(j, dp)
      const2 = 1._dp/REAL(j, dp)
      ALLOCATE (iii(staging_env%nseg), jjj(staging_env%nseg))
      DO i = 1, staging_env%nseg
         iii(i) = staging_env%j*(i-1)+1 !first el
      END DO
      DO i = 1, staging_env%nseg-1
         jjj(i) = iii(i)+j ! next first el (pbc)
      END DO
      jjj(staging_env%nseg) = 1

      x = ux
      DO i = 1, staging_env%nseg
         x(j-1+iii(i), :) = x(j-1+iii(i), :)+ &
                            const*ux(jjj(i), :)+ux(iii(i), :)*const2
      END DO
      DO ist = 1, staging_env%nseg
         DO i = staging_env%j-2, 2, -1
            x(i+iii(ist), :) = x(i+iii(ist), :)+ &
                               REAL(i-1, dp)/REAL(i, dp)*x(i+iii(ist)+1, :) &
                               +ux(iii(ist), :)/REAL(i, dp)
         END DO
      END DO
      RETURN
   END SUBROUTINE staging_u2x

! ***************************************************************************
!> \brief staging transformation for the forces
!> \param staging_env the environment for the staging transformation
!> \param uf will contain the forces after for the transformed variable
!> \param f the forces to transform
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE staging_f2uf(staging_env, uf, f)
      TYPE(staging_env_type), POINTER                    :: staging_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: uf
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: f

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_f2uf', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idim, ij, ist, k
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii, jjj, kkk
      REAL(kind=dp)                                      :: const, sum_f

      IF(.NOT.(ASSOCIATED(staging_env)))CALL cp__a("motion/pint_staging.F",272)
      IF(.NOT.(staging_env%ref_count > 0))CALL cp__a("motion/pint_staging.F",273)
      const = REAL(staging_env%j-1, dp)/REAL(staging_env%j, dp)
      ALLOCATE (iii(staging_env%j), jjj(staging_env%j), &
                kkk(staging_env%j))
      DO ist = 1, staging_env%j-1
         iii(ist) = (ist-1)*staging_env%j+1 ! first el
         jjj(ist) = iii(ist)+staging_env%j-1 ! last el
         kkk(ist) = iii(ist)-1 ! prev el
      END DO
      kkk(1) = staging_env%p

      uf = f
      ! staging beads
      DO k = 1, staging_env%nseg
         DO i = 2, staging_env%j
            uf(i+iii(k), :) = uf(i+iii(k), :) &
                              +REAL(i-1, dp)/REAL(i, dp)*uf(i+iii(k)-1, :)
         END DO
      END DO
      ! end point beads
      DO idim = 1, SIZE(uf, 2)
         DO k = 1, staging_env%nseg
            sum_f = 0._dp
            DO ij = 2, staging_env%j-1
               sum_f = sum_f+uf((k-1)*staging_env%j+ij, idim)
            END DO
            uf(iii(k), idim) = uf(iii(k), idim)+ &
                               sum_f-const*(uf(jjj(k), idim)-uf(kkk(k), idim))
         END DO
      END DO
      RETURN
   END SUBROUTINE staging_f2uf

! ***************************************************************************
!> \brief calculates the harmonic force in the staging basis
!> \param staging_env the staging environment
!> \param mass_beads the masses of the beads
!> \param ux the positions of the beads in the staging basis
!> \param uf_h the harmonic forces (not accelerations)
!> \param e_h ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE staging_calc_uf_h(staging_env, mass_beads, ux, uf_h, e_h)
      TYPE(staging_env_type), POINTER                    :: staging_env
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: mass_beads, ux, uf_h
      REAL(KIND=dp), INTENT(OUT)                         :: e_h

      CHARACTER(len=*), PARAMETER :: routineN = 'staging_calc_uf_h', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idim, isg, ist
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii, jjj, kkk
      REAL(kind=dp)                                      :: d, f

      e_h = 0.0_dp

      ALLOCATE (iii(staging_env%nseg), jjj(staging_env%nseg), &
                kkk(staging_env%nseg))

      DO ist = 1, staging_env%nseg
         iii(ist) = (ist-1)*staging_env%j+1 ! first el
         jjj(ist) = iii(ist)+staging_env%j ! next fisrt (pbc)
         kkk(ist) = iii(ist)-staging_env%j ! prev first el (pbc)
      END DO
      jjj(staging_env%nseg) = 1
      kkk(1) = staging_env%p-staging_env%j

      DO idim = 1, SIZE(mass_beads, 2)
         DO ist = 1, staging_env%nseg
            e_h = e_h+0.5*mass_beads(1, idim)*staging_env%w_j**2* &
                  (ux(iii(ist), idim)-ux(jjj(ist), idim))**2
            uf_h(iii(ist), idim) = mass_beads(1, idim)*staging_env%w_j**2*( &
                                   2._dp*ux(iii(ist), idim) &
                                   -ux(jjj(ist), idim) &
                                   -ux(kkk(ist), idim) &
                                   )
            DO isg = 2, staging_env%j ! use 3 as start?
               d = ux((ist-1)*staging_env%j+isg, idim)
               f = mass_beads((ist-1)*staging_env%j+isg, idim)*staging_env%w_j**2*d
               e_h = e_h+0.5_dp*f*d
               uf_h((ist-1)*staging_env%j+isg, idim) = f
            END DO
         END DO
      END DO
      RETURN
   END SUBROUTINE staging_calc_uf_h

END MODULE pint_staging
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Data type and methods dealing with PI calcs in normal mode coords
!> \author fawzi
!> \par    History
!>         2006-02 created
!>         2006-11 modified so it might actually work [hforbert]
!>         2009-04-07 moved from pint_types module to a separate file [lwalewski]
!>         2015-10 added alternative normal mode transformation needed by RPMD
!>                 [Felix Uhl
! **************************************************************************************************
MODULE pint_normalmode
   USE input_constants,                 ONLY: propagator_pimd,&
                                              propagator_rpmd
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE pint_types,                      ONLY: normalmode_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_normalmode'

   INTEGER, SAVE, PRIVATE :: last_normalmode_id = 0

   PUBLIC :: normalmode_env_create
   PUBLIC :: normalmode_release
   PUBLIC :: normalmode_init_masses
   PUBLIC :: normalmode_x2u
   PUBLIC :: normalmode_u2x
   PUBLIC :: normalmode_f2uf
   PUBLIC :: normalmode_calc_uf_h

CONTAINS

! ***************************************************************************
!> \brief creates the data needed for a normal mode transformation
!> \param normalmode_env ...
!> \param normalmode_section ...
!> \param p ...
!> \param kT ...
!> \param propagator ...
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_env_create(normalmode_env, normalmode_section, p, kT, propagator)
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      TYPE(section_vals_type), POINTER                   :: normalmode_section
      INTEGER, INTENT(in)                                :: p
      REAL(kind=dp), INTENT(in)                          :: kT
      INTEGER, INTENT(in)                                :: propagator

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, li
      REAL(kind=dp)                                      :: invsqrtp, pip, sqrt2p, twopip

      IF(.NOT.(.NOT. ASSOCIATED(normalmode_env)))CALL cp__a("motion/pint_normalmode.F",67)
      ALLOCATE (normalmode_env)
      ALLOCATE (normalmode_env%x2u(p, p))
      ALLOCATE (normalmode_env%u2x(p, p))
      ALLOCATE (normalmode_env%lambda(p))
      last_normalmode_id = last_normalmode_id+1
      normalmode_env%id_nr = last_normalmode_id
      normalmode_env%ref_count = 1

      normalmode_env%p = p

      CALL section_vals_val_get(normalmode_section, "Q_CENTROID", &
                                r_val=normalmode_env%Q_centroid)
      CALL section_vals_val_get(normalmode_section, "Q_BEAD", &
                                r_val=normalmode_env%Q_bead)
      CALL section_vals_val_get(normalmode_section, "MODEFACTOR", &
                                r_val=normalmode_env%modefactor)

      IF (normalmode_env%Q_centroid < 0.0_dp) THEN
         normalmode_env%Q_centroid = -normalmode_env%Q_centroid/(kT*p)
      END IF
      IF (normalmode_env%Q_bead < 0.0_dp) THEN
         normalmode_env%Q_bead = -normalmode_env%Q_bead/(kT*p)
      END IF

      !Use different normal mode transformations depending on the propagator
      IF (propagator == propagator_pimd) THEN
         normalmode_env%harm = p*kT*kT/normalmode_env%modefactor

         ! set up the transformation matrices
         DO i = 1, p
            normalmode_env%lambda(i) = 2.0_dp*(1.0_dp-COS(pi*(i/2)*2.0_dp/p))
            DO j = 1, p
               k = ((i/2)*(j-1))/p
               k = (i/2)*(j-1)-k*p
               li = 2*(i-2*(i/2))*p-p
               normalmode_env%u2x(j, i) = SQRT(2.0_dp/p)*SIN(twopi*(k+0.125_dp*li)/p)
            END DO
         END DO
         normalmode_env%lambda(1) = 1.0_dp/(p*normalmode_env%modefactor)
         DO i = 1, p
            DO j = 1, p
               normalmode_env%x2u(i, j) = SQRT(normalmode_env%lambda(i)* &
                                               normalmode_env%modefactor)* &
                                          normalmode_env%u2x(j, i)
            END DO
         END DO
         DO i = 1, p
            DO j = 1, p
               normalmode_env%u2x(i, j) = normalmode_env%u2x(i, j)/ &
                                          SQRT(normalmode_env%lambda(j)* &
                                               normalmode_env%modefactor)
            END DO
         END DO
         normalmode_env%lambda(:) = normalmode_env%harm
      ELSE IF (propagator == propagator_rpmd) THEN
         normalmode_env%harm = kT/normalmode_env%modefactor
         sqrt2p = SQRT(2.0_dp/REAL(p, dp))
         twopip = twopi/REAL(p, dp)
         pip = pi/REAL(p, dp)
         invsqrtp = 1.0_dp/SQRT(REAL(p, dp))
         normalmode_env%x2u(:, :) = 0.0_dp
         normalmode_env%x2u(1, :) = invsqrtp
         DO j = 1, p
            DO i = 2, p/2+1
               normalmode_env%x2u(i, j) = sqrt2p*COS(twopip*(i-1)*(j-1))
            END DO
            DO i = p/2+2, p
               normalmode_env%x2u(i, j) = sqrt2p*SIN(twopip*(i-1)*(j-1))
            END DO
         END DO
         IF (MOD(p, 2) == 0) THEN
            DO i = 1, p-1, 2
               normalmode_env%x2u(p/2+1, i) = invsqrtp
               normalmode_env%x2u(p/2+1, i+1) = -1.0_dp*invsqrtp
            END DO
         END IF

         normalmode_env%u2x = TRANSPOSE(normalmode_env%x2u)

         ! Setting up propagator frequencies for rpmd
         normalmode_env%lambda(1) = 0.0_dp
         DO i = 2, p
            normalmode_env%lambda(i) = 2.0_dp*normalmode_env%harm*SIN((i-1)*pip)
            normalmode_env%lambda(i) = normalmode_env%lambda(i)*normalmode_env%lambda(i)
         END DO
         normalmode_env%harm = kT*kT
      ELSE
         CALL cp__b("motion/pint_normalmode.F",155,"UNKNOWN PROPAGATOR FOR PINT SELECTED")
      END IF

      RETURN
   END SUBROUTINE normalmode_env_create

! ***************************************************************************
!> \brief releases the normalmode environment
!> \param normalmode_env the normalmode_env to release
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_release(normalmode_env)

      TYPE(normalmode_env_type), POINTER                 :: normalmode_env

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(normalmode_env)) THEN
         IF(.NOT.(normalmode_env%ref_count > 0))CALL cp__a("motion/pint_normalmode.F",174)
         normalmode_env%ref_count = normalmode_env%ref_count-1
         IF (normalmode_env%ref_count == 0) THEN
            DEALLOCATE (normalmode_env%x2u)
            DEALLOCATE (normalmode_env%u2x)
            DEALLOCATE (normalmode_env%lambda)
            DEALLOCATE (normalmode_env)
         END IF
      END IF
      NULLIFY (normalmode_env)

      RETURN
   END SUBROUTINE normalmode_release

! ***************************************************************************
!> \brief retains a normalmode_env
!> \param normalmode_env the normalmode_env to retain
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_retain(normalmode_env)

      TYPE(normalmode_env_type), POINTER                 :: normalmode_env

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(normalmode_env)))CALL cp__a("motion/pint_normalmode.F",200)
      IF(.NOT.(normalmode_env%ref_count > 0))CALL cp__a("motion/pint_normalmode.F",201)
      normalmode_env%ref_count = normalmode_env%ref_count+1

      RETURN
   END SUBROUTINE normalmode_retain

! ***************************************************************************
!> \brief initializes the masses and fictitious masses compatible with the
!>      normal mode information
!> \param normalmode_env the definition of the normal mode transformation
!> \param mass *input* the masses of the particles
!> \param mass_beads masses of the beads
!> \param mass_fict the fictitious masses
!> \param Q masses of the nose thermostats
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_init_masses(normalmode_env, mass, mass_beads, mass_fict, &
                                     Q)

      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      REAL(kind=dp), DIMENSION(:), INTENT(in)            :: mass
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL                                        :: mass_beads, mass_fict
      REAL(kind=dp), DIMENSION(:), INTENT(out), OPTIONAL :: Q

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_init_masses', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, ib

      IF (PRESENT(Q)) THEN
         Q = normalmode_env%Q_bead
         Q(1) = normalmode_env%Q_centroid
      END IF
      IF (PRESENT(mass_beads) .OR. PRESENT(mass_fict)) THEN
         IF (PRESENT(mass_beads)) THEN
            DO iat = 1, SIZE(mass)
               mass_beads(1, iat) = 0.0_dp
               DO ib = 2, normalmode_env%p
                  mass_beads(ib, iat) = mass(iat)
               END DO
            END DO
         END IF
         IF (PRESENT(mass_fict)) THEN
            DO iat = 1, SIZE(mass)
               DO ib = 1, normalmode_env%p
                  mass_fict(ib, iat) = mass(iat)
               END DO
            END DO
         END IF
      END IF

      RETURN
   END SUBROUTINE normalmode_init_masses

! ***************************************************************************
!> \brief Transforms from the x into the u variables using a normal mode
!>      transformation for the positions
!> \param normalmode_env the environment for the normal mode transformation
!> \param ux will contain the u variable
!> \param x the positions to transform
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_x2u(normalmode_env, ux, x)
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_x2u', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(normalmode_env)))CALL cp__a("motion/pint_normalmode.F",271)
      IF(.NOT.(normalmode_env%ref_count > 0))CALL cp__a("motion/pint_normalmode.F",272)
      CALL DGEMM('N', 'N', normalmode_env%p, SIZE(x, 2), normalmode_env%p, 1.0_dp, &
                 normalmode_env%x2u(1, 1), SIZE(normalmode_env%x2u, 1), x(1, 1), SIZE(x, 1), &
                 0.0_dp, ux, SIZE(ux, 1))
      RETURN
   END SUBROUTINE normalmode_x2u

! ***************************************************************************
!> \brief transform from the u variable to the x (back normal mode
!>      transformation for the positions)
!> \param normalmode_env the environment for the normal mode transformation
!> \param ux the u variable (positions to be backtransformed)
!> \param x will contain the positions
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_u2x(normalmode_env, ux, x)
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_u2x', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(normalmode_env)))CALL cp__a("motion/pint_normalmode.F",294)
      IF(.NOT.(normalmode_env%ref_count > 0))CALL cp__a("motion/pint_normalmode.F",295)
      CALL DGEMM('N', 'N', normalmode_env%p, SIZE(ux, 2), normalmode_env%p, 1.0_dp, &
                 normalmode_env%u2x(1, 1), SIZE(normalmode_env%u2x, 1), ux(1, 1), SIZE(ux, 1), &
                 0.0_dp, x, SIZE(x, 1))
      RETURN
   END SUBROUTINE normalmode_u2x

! ***************************************************************************
!> \brief normalmode transformation for the forces
!> \param normalmode_env the environment for the normal mode transformation
!> \param uf will contain the forces for the transformed variables afterwards
!> \param f the forces to transform
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_f2uf(normalmode_env, uf, f)
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out)        :: uf
      REAL(kind=dp), DIMENSION(:, :), INTENT(in)         :: f

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_f2uf', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(normalmode_env)))CALL cp__a("motion/pint_normalmode.F",317)
      IF(.NOT.(normalmode_env%ref_count > 0))CALL cp__a("motion/pint_normalmode.F",318)
      CALL DGEMM('T', 'N', normalmode_env%p, SIZE(f, 2), normalmode_env%p, 1.0_dp, &
                 normalmode_env%u2x(1, 1), SIZE(normalmode_env%u2x, 1), f(1, 1), SIZE(f, 1), &
                 0.0_dp, uf, SIZE(uf, 1))
      RETURN
   END SUBROUTINE normalmode_f2uf

! ***************************************************************************
!> \brief calculates the harmonic force in the normal mode basis
!> \param normalmode_env the normal mode environment
!> \param mass_beads the masses of the beads
!> \param ux the positions of the beads in the staging basis
!> \param uf_h the harmonic forces (not accelerations)
!> \param e_h ...
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE normalmode_calc_uf_h(normalmode_env, mass_beads, ux, uf_h, e_h)
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: mass_beads, ux, uf_h
      REAL(KIND=dp), INTENT(OUT)                         :: e_h

      CHARACTER(len=*), PARAMETER :: routineN = 'normalmode_calc_uf_h', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ibead, idim
      REAL(kind=dp)                                      :: f

      e_h = 0.0_dp
      DO idim = 1, SIZE(mass_beads, 2)

         ! starting at 2 since the centroid is at 1 and it's mass_beads
         ! SHOULD be zero anyways:

         uf_h(1, idim) = 0.0_dp
         DO ibead = 2, normalmode_env%p
            f = -mass_beads(ibead, idim)*normalmode_env%lambda(ibead)*ux(ibead, idim)
            uf_h(ibead, idim) = f
            ! - to cancel the - in the force f.
            e_h = e_h-0.5_dp*ux(ibead, idim)*f
         END DO

      END DO
      RETURN
   END SUBROUTINE normalmode_calc_uf_h

END MODULE pint_normalmode
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE pint_transformations
   USE input_constants,                 ONLY: transformation_stage
   USE kinds,                           ONLY: dp
   USE pint_normalmode,                 ONLY: normalmode_f2uf,&
                                              normalmode_u2x,&
                                              normalmode_x2u
   USE pint_staging,                    ONLY: staging_f2uf,&
                                              staging_u2x,&
                                              staging_x2u
   USE pint_types,                      ONLY: pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_transformations'

   PUBLIC :: pint_x2u, &
             pint_u2x, &
             pint_f2uf

CONTAINS

! ***************************************************************************
!> \brief Transforms from the x into the u variables
!>      (at the moment a staging transformation for the positions)
!> \param pint_env the path integral environment
!> \param ux will contain the u variable (defaults to pint_env%ux)
!> \param x the positions to transform (defaults to pint_env%x)
!> \par History
!>      Added normal mode transformation [hforbert]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_x2u(pint_env, ux, x)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL, TARGET                                :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(in), &
         OPTIONAL, TARGET                                :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_x2u', routineP = moduleN//':'//routineN

      REAL(kind=dp), DIMENSION(:, :), POINTER            :: my_ux, my_x

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_transformations.F",51)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_transformations.F",52)
      my_x => pint_env%x
      my_ux => pint_env%ux
      IF (PRESENT(x)) my_x => x
      IF (PRESENT(ux)) my_ux => ux
      IF(.NOT.(ASSOCIATED(my_ux)))CALL cp__a("motion/pint_transformations.F",57)
      IF(.NOT.(ASSOCIATED(my_x)))CALL cp__a("motion/pint_transformations.F",58)

      IF (pint_env%transform == transformation_stage) THEN
         CALL staging_x2u(pint_env%staging_env, ux=my_ux, x=my_x)
      ELSE
         CALL normalmode_x2u(pint_env%normalmode_env, ux=my_ux, x=my_x)
      END IF
      RETURN
   END SUBROUTINE pint_x2u

! ***************************************************************************
!> \brief transform from the u variable to the x (inverse of x2u)
!> \param pint_env path integral environment
!> \param ux the u variable (positions to be backtransformed)
!> \param x will contain the positions
!> \par History
!>      Added normal mode transformation by hforbert
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_u2x(pint_env, ux, x)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(in), &
         OPTIONAL, TARGET                                :: ux
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL, TARGET                                :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_u2x', routineP = moduleN//':'//routineN

      REAL(kind=dp), DIMENSION(:, :), POINTER            :: my_ux, my_x

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_transformations.F",88)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_transformations.F",89)
      my_x => pint_env%x
      my_ux => pint_env%ux
      IF (PRESENT(x)) my_x => x
      IF (PRESENT(ux)) my_ux => ux
      IF(.NOT.(ASSOCIATED(my_ux)))CALL cp__a("motion/pint_transformations.F",94)
      IF(.NOT.(ASSOCIATED(my_x)))CALL cp__a("motion/pint_transformations.F",95)

      IF (pint_env%transform == transformation_stage) THEN
         CALL staging_u2x(pint_env%staging_env, ux=my_ux, x=my_x)
      ELSE
         CALL normalmode_u2x(pint_env%normalmode_env, ux=my_ux, x=my_x)
      END IF
      RETURN
   END SUBROUTINE pint_u2x

! ***************************************************************************
!> \brief transformation x to u for the forces
!> \param pint_env the path integral environment
!> \param uf will contain the accelerations for the transformed variables
!>        afterwards
!> \param f the forces to transform
!> \par History
!>      Added normal mode transformation [hforbert]
!>      Divide forces by the number of beads, since the replication
!>        environment (should) give raw forces [hforbert]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_f2uf(pint_env, uf, f)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL, TARGET                                :: uf
      REAL(kind=dp), DIMENSION(:, :), INTENT(in), &
         OPTIONAL, TARGET                                :: f

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_f2uf', routineP = moduleN//':'//routineN

      REAL(kind=dp), DIMENSION(:, :), POINTER            :: my_f, my_uf

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_transformations.F",128)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_transformations.F",129)
      my_f => pint_env%f
      my_uf => pint_env%uf
      IF (PRESENT(f)) my_f => f
      IF (PRESENT(uf)) my_uf => uf
      IF(.NOT.(ASSOCIATED(my_uf)))CALL cp__a("motion/pint_transformations.F",134)
      IF(.NOT.(ASSOCIATED(my_f)))CALL cp__a("motion/pint_transformations.F",135)

      IF (pint_env%transform == transformation_stage) THEN
         CALL staging_f2uf(pint_env%staging_env, uf=my_uf, f=my_f)
      ELSE
         CALL normalmode_f2uf(pint_env%normalmode_env, uf=my_uf, f=my_f)
      END IF

      my_uf = my_uf/pint_env%mass_fict*pint_env%propagator%physpotscale
      RETURN
   END SUBROUTINE pint_f2uf

END MODULE pint_transformations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set of routines to dump the restart file of CP2K
!> \par History
!>      01.2006 [created] Teodoro Laino
! **************************************************************************************************
MODULE input_cp2k_restarts
   USE al_system_types,                 ONLY: al_system_type
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE averages_types,                  ONLY: average_quantities_type
   USE cp2k_info,                       ONLY: write_restart_header
   USE cp_linked_list_val,              ONLY: cp_sll_val_create,&
                                              cp_sll_val_get_length,&
                                              cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE csvr_system_types,               ONLY: csvr_system_type
   USE extended_system_types,           ONLY: lnhc_parameters_type,&
                                              map_info_type,&
                                              npt_info_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              multiple_fe_list
   USE gle_system_types,                ONLY: gle_type
   USE helium_types,                    ONLY: helium_solvent_type
   USE input_constants,                 ONLY: &
        do_band_collective, do_thermo_al, do_thermo_csvr, do_thermo_gle, &
        do_thermo_no_communication, do_thermo_nose, mol_dyn_run, mon_car_run, pint_run
   USE input_restart_force_eval,        ONLY: update_force_eval
   USE input_restart_rng,               ONLY: section_rng_val_set
   USE input_section_types,             ONLY: &
        section_get_keyword_index, section_type, section_vals_add_values, section_vals_get, &
        section_vals_get_subs_vals, section_vals_get_subs_vals3, section_vals_remove_values, &
        section_vals_type, section_vals_val_get, section_vals_val_set, section_vals_val_unset, &
        section_vals_write
   USE input_val_types,                 ONLY: val_create,&
                                              val_release,&
                                              val_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_size
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_bcast,&
                                              mp_gather,&
                                              mp_sum
   USE metadynamics_types,              ONLY: meta_env_type
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE neb_types,                       ONLY: neb_var_type
   USE parallel_rng_types,              ONLY: dump_rng_stream,&
                                              get_rng_stream,&
                                              rng_record_length
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: get_particle_pos_or_vel,&
                                              particle_type
   USE physcon,                         ONLY: angstrom
   USE pint_transformations,            ONLY: pint_u2x
   USE pint_types,                      ONLY: pint_env_type,&
                                              thermostat_gle,&
                                              thermostat_nose,&
                                              thermostat_piglet,&
                                              thermostat_pile
   USE simpar_types,                    ONLY: simpar_type
   USE string_utilities,                ONLY: string_to_ascii
   USE termination,                     ONLY: stop_memory
   USE thermostat_types,                ONLY: thermostat_type
   USE thermostat_utils,                ONLY: communication_thermo_low2,&
                                              get_kin_energies
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_restarts'

   PUBLIC :: write_restart

CONTAINS

! **************************************************************************************************
!> \brief checks if a restart needs to be written and does so, updating all necessary fields
!>      in the input file. This is a relatively simple wrapper routine.
!> \param md_env ...
!> \param force_env ...
!> \param root_section ...
!> \param coords ...
!> \param vels ...
!> \param pint_env ...
!> \param helium_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE write_restart(md_env, force_env, root_section, &
                            coords, vels, pint_env, helium_env)
      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(neb_var_type), OPTIONAL, POINTER              :: coords, vels
      TYPE(pint_env_type), OPTIONAL, POINTER             :: pint_env
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_restart', routineP = moduleN//':'//routineN
      CHARACTER(LEN=30), DIMENSION(2), PARAMETER :: &
         keys = (/"PRINT%RESTART_HISTORY", "PRINT%RESTART        "/)

      INTEGER                                            :: handle, ikey, ires, log_unit, nforce_eval
      LOGICAL                                            :: save_mem, write_binary_restart_file
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: global_section, motion_section, sections

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      motion_section => section_vals_get_subs_vals(root_section, "MOTION")

      NULLIFY (global_section)
      global_section => section_vals_get_subs_vals(root_section, "GLOBAL")
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           motion_section, keys(1)), cp_p_file) .OR. &
          BTEST(cp_print_key_should_output(logger%iter_info, &
                                           motion_section, keys(2)), cp_p_file)) THEN

         sections => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
         CALL section_vals_get(sections, n_repetition=nforce_eval)
         CALL section_vals_val_get(motion_section, "PRINT%RESTART%SPLIT_RESTART_FILE", &
                                   l_val=write_binary_restart_file)

         IF (write_binary_restart_file) THEN
            CALL update_subsys_release(md_env, force_env, root_section)
            CALL update_motion_release(motion_section)
            DO ikey = 1, SIZE(keys)
               log_unit = cp_logger_get_default_io_unit(logger)
               IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                    motion_section, keys(ikey)), cp_p_file)) THEN
                  ires = cp_print_key_unit_nr(logger, motion_section, TRIM(keys(ikey)), &
                                              extension=".restart.bin", &
                                              file_action="READWRITE", &
                                              file_form="UNFORMATTED", &
                                              file_position="REWIND", &
                                              file_status="UNKNOWN", &
                                              do_backup=(ikey == 2))
                  CALL write_binary_restart(ires, log_unit, root_section, md_env, force_env)
                  CALL cp_print_key_finished_output(ires, logger, motion_section, &
                                                    TRIM(keys(ikey)))
               END IF
            END DO
         END IF

         CALL update_input(md_env, force_env, root_section, coords, vels, pint_env, helium_env, &
                           save_mem=save_mem, &
                           write_binary_restart_file=write_binary_restart_file)

         DO ikey = 1, SIZE(keys)
            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 motion_section, keys(ikey)), cp_p_file)) THEN
               ires = cp_print_key_unit_nr(logger, motion_section, TRIM(keys(ikey)), &
                                           extension=".restart", &
                                           file_position="REWIND", &
                                           do_backup=(ikey == 2))
               IF (ires > 0) THEN
                  CALL write_restart_header(ires)
                  CALL section_vals_write(root_section, unit_nr=ires, hide_root=.TRUE.)
               END IF
               CALL cp_print_key_finished_output(ires, logger, motion_section, TRIM(keys(ikey)))
            END IF
         END DO

         IF (save_mem) THEN
            CALL update_subsys_release(md_env, force_env, root_section)
            CALL update_motion_release(motion_section)
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE write_restart

! **************************************************************************************************
!> \brief deallocate some sub_sections of the section subsys to save some memory
!> \param md_env ...
!> \param force_env ...
!> \param root_section ...
!> \par History
!>      06.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE update_subsys_release(md_env, force_env, root_section)

      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_subsys_release', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: unit_str
      INTEGER                                            :: handle, iforce_eval, myid, nforce_eval
      INTEGER, DIMENSION(:), POINTER                     :: i_force_eval
      LOGICAL                                            :: explicit, scale, skip_vel_section
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: my_force_b, my_force_env
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(section_vals_type), POINTER                   :: force_env_sections, subsys_section, &
                                                            work_section

      CALL timeset(routineN, handle)

      NULLIFY (core_particles, my_force_env, my_force_b, particles, &
               shell_particles, subsys, work_section)

      IF (PRESENT(md_env)) THEN
         CALL get_md_env(md_env=md_env, force_env=my_force_env)
      ELSEIF (PRESENT(force_env)) THEN
         my_force_env => force_env
      END IF

      IF (ASSOCIATED(my_force_env)) THEN
         NULLIFY (subsys_section)
         CALL section_vals_val_get(root_section, "GLOBAL%RUN_TYPE", i_val=myid)
         skip_vel_section = ( &
                            (myid /= mol_dyn_run) .AND. &
                            (myid /= mon_car_run) .AND. &
                            (myid /= pint_run))

         force_env_sections => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
         CALL multiple_fe_list(force_env_sections, root_section, i_force_eval, nforce_eval)

         DO iforce_eval = 1, nforce_eval
            subsys_section => section_vals_get_subs_vals3(force_env_sections, "SUBSYS", &
                                                          i_rep_section=i_force_eval(iforce_eval))
            CALL section_vals_get(subsys_section, explicit=explicit)
            IF (.NOT. explicit) CYCLE ! Nothing to update...

            my_force_b => my_force_env
            IF (iforce_eval > 1) my_force_b => my_force_env%sub_force_env(iforce_eval-1)%force_env

            CALL force_env_get(my_force_b, subsys=subsys)

            CALL cp_subsys_get(subsys, particles=particles, shell_particles=shell_particles, &
                               core_particles=core_particles)

            work_section => section_vals_get_subs_vals(subsys_section, "COORD")
            CALL section_vals_get(work_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
               CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
            END IF
            CALL section_vals_remove_values(work_section)
            IF (explicit) THEN
               CALL section_vals_val_set(work_section, "UNIT", c_val=unit_str)
               CALL section_vals_val_set(work_section, "SCALED", l_val=scale)
            END IF

            work_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
            IF (.NOT. skip_vel_section) THEN
               CALL section_vals_remove_values(work_section)
            END IF

            IF (ASSOCIATED(shell_particles)) THEN
               work_section => section_vals_get_subs_vals(subsys_section, "SHELL_COORD")
               CALL section_vals_get(work_section, explicit=explicit)
               IF (explicit) THEN
                  CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
                  CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
               END IF
               CALL section_vals_remove_values(work_section)
               IF (explicit) THEN
                  CALL section_vals_val_set(work_section, "UNIT", c_val=unit_str)
                  CALL section_vals_val_set(work_section, "SCALED", l_val=scale)
               END IF

               work_section => section_vals_get_subs_vals(subsys_section, "SHELL_VELOCITY")
               IF (.NOT. skip_vel_section) THEN
                  CALL section_vals_remove_values(work_section)
               END IF
            END IF

            IF (ASSOCIATED(core_particles)) THEN
               work_section => section_vals_get_subs_vals(subsys_section, "CORE_COORD")
               CALL section_vals_get(work_section, explicit=explicit)
               IF (explicit) THEN
                  CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
                  CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
               END IF
               CALL section_vals_remove_values(work_section)
               IF (explicit) THEN
                  CALL section_vals_val_set(work_section, "UNIT", c_val=unit_str)
                  CALL section_vals_val_set(work_section, "SCALED", l_val=scale)
               END IF

               work_section => section_vals_get_subs_vals(subsys_section, "CORE_VELOCITY")
               IF (.NOT. skip_vel_section) THEN
                  CALL section_vals_remove_values(work_section)
               END IF
            END IF

         END DO

         DEALLOCATE (i_force_eval)

      END IF

      CALL timestop(handle)

   END SUBROUTINE update_subsys_release

! **************************************************************************************************
!> \brief deallocate the nose subsections (coord, vel, force, mass) in the md section
!> \param motion_section ...
!> \par History
!>      08.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE update_motion_release(motion_section)

      TYPE(section_vals_type), POINTER                   :: motion_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_motion_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: work_section

      CALL timeset(routineN, handle)

      NULLIFY (work_section)

      work_section => section_vals_get_subs_vals(motion_section, "MD%AVERAGES%RESTART_AVERAGES")
      CALL section_vals_remove_values(work_section)

      work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(work_section)

      work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(work_section)

      work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(work_section)
      work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(work_section)

      CALL timestop(handle)

   END SUBROUTINE update_motion_release

! **************************************************************************************************
!> \brief Updates the whole input file for the restart
!> \param md_env ...
!> \param force_env ...
!> \param root_section ...
!> \param coords ...
!> \param vels ...
!> \param pint_env ...
!> \param helium_env ...
!> \param save_mem ...
!> \param write_binary_restart_file ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_input(md_env, force_env, root_section, coords, vels, pint_env, &
                           helium_env, save_mem, write_binary_restart_file)

      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(neb_var_type), OPTIONAL, POINTER              :: coords, vels
      TYPE(pint_env_type), OPTIONAL, POINTER             :: pint_env
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: save_mem, write_binary_restart_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_respa, lcond, my_save_mem, &
                                                            my_write_binary_restart_file
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(force_env_type), POINTER                      :: my_force_env
      TYPE(section_vals_type), POINTER                   :: motion_section
      TYPE(simpar_type), POINTER                         :: simpar

      CALL timeset(routineN, handle)

      NULLIFY (logger, motion_section, my_force_env)

      IF (PRESENT(save_mem)) THEN
         my_save_mem = save_mem
      ELSE
         my_save_mem = .FALSE.
      END IF

      IF (PRESENT(write_binary_restart_file)) THEN
         my_write_binary_restart_file = write_binary_restart_file
      ELSE
         my_write_binary_restart_file = .FALSE.
      END IF

      logger => cp_get_default_logger()

      ! Can handle md_env or force_env
      lcond = PRESENT(md_env) .OR. PRESENT(force_env) .OR. PRESENT(pint_env) .OR. PRESENT(helium_env)
      IF (lcond) THEN
         IF (PRESENT(md_env)) THEN
            CALL get_md_env(md_env=md_env, force_env=my_force_env)
         ELSE IF (PRESENT(force_env)) THEN
            my_force_env => force_env
         END IF
         ! The real restart setting...
         motion_section => section_vals_get_subs_vals(root_section, "MOTION")
         CALL update_motion(motion_section, &
                            md_env=md_env, &
                            force_env=my_force_env, &
                            logger=logger, &
                            coords=coords, &
                            vels=vels, &
                            pint_env=pint_env, &
                            helium_env=helium_env, &
                            save_mem=my_save_mem, &
                            write_binary_restart_file=my_write_binary_restart_file)
         ! Update one force_env_section per time..
         IF (ASSOCIATED(my_force_env)) THEN
            do_respa = .FALSE.
            ! Do respa only in case of RESPA MD
            IF (PRESENT(md_env)) THEN
               CALL get_md_env(md_env=md_env, simpar=simpar)
               IF (simpar%do_respa) THEN
                  do_respa = .TRUE.
               ENDIF
            ENDIF

            CALL update_force_eval(force_env=my_force_env, &
                                   root_section=root_section, &
                                   write_binary_restart_file=my_write_binary_restart_file, &
                                   respa=do_respa)

         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE update_input

! **************************************************************************************************
!> \brief Updates the motion section of the input file
!> \param motion_section ...
!> \param md_env ...
!> \param force_env ...
!> \param logger ...
!> \param coords ...
!> \param vels ...
!> \param pint_env ...
!> \param helium_env ...
!> \param save_mem ...
!> \param write_binary_restart_file ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_motion(motion_section, md_env, force_env, logger, &
                            coords, vels, pint_env, helium_env, save_mem, &
                            write_binary_restart_file)
      TYPE(section_vals_type), POINTER                   :: motion_section
      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(neb_var_type), OPTIONAL, POINTER              :: coords, vels
      TYPE(pint_env_type), OPTIONAL, POINTER             :: pint_env
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: save_mem, write_binary_restart_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_motion', routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, handle, handle2, i, irep, isec, &
                                                            j, nhc_len, tot_nhcneed
      INTEGER, DIMENSION(:), POINTER                     :: walkers_status
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: my_save_mem, my_write_binary_restart_file
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buffer, eta, fnhc, mnhc, veta, wrk
      REAL(KIND=dp), POINTER                             :: constant, t
      TYPE(average_quantities_type), POINTER             :: averages
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(npt_info_type), POINTER                       :: npt(:, :)
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: replica_section, work_section
      TYPE(thermostat_type), POINTER                     :: thermostat_baro, thermostat_part, &
                                                            thermostat_shell

      CALL timeset(routineN, handle)
      NULLIFY (logger, thermostat_part, thermostat_baro, npt, para_env, nhc, &
               work_section, thermostat_shell, t, averages, constant, &
               walkers_status, itimes, meta_env)
      NULLIFY (particles)
      NULLIFY (subsys)
      IF (PRESENT(md_env)) THEN
         CALL get_md_env(md_env=md_env, &
                         thermostat_part=thermostat_part, &
                         thermostat_baro=thermostat_baro, &
                         thermostat_shell=thermostat_shell, &
                         npt=npt, &
                         t=t, &
                         constant=constant, &
                         itimes=itimes, &
                         averages=averages, &
                         para_env=para_env)
      ELSE
         IF (ASSOCIATED(force_env)) THEN
            para_env => force_env%para_env
         ELSEIF (PRESENT(pint_env)) THEN
            para_env => pint_env%logger%para_env
         ELSEIF (PRESENT(helium_env)) THEN
            para_env => helium_env%logger%para_env
         ELSE
            CALL cp__b("motion/input_cp2k_restarts.F",556,"")
         END IF
      END IF

      IF (ASSOCIATED(force_env)) THEN
         meta_env => force_env%meta_env
      END IF

      IF (PRESENT(save_mem)) THEN
         my_save_mem = save_mem
      ELSE
         my_save_mem = .FALSE.
      END IF

      IF (PRESENT(write_binary_restart_file)) THEN
         my_write_binary_restart_file = write_binary_restart_file
      ELSE
         my_write_binary_restart_file = .FALSE.
      END IF

      CALL timeset(routineN//"_COUNTERS", handle2)
      IF (ASSOCIATED(itimes)) THEN
         IF (itimes >= 0) THEN
            CALL section_vals_val_set(motion_section, "MD%STEP_START_VAL", i_val=itimes)
            IF(.NOT.(ASSOCIATED(t)))CALL cp__a("motion/input_cp2k_restarts.F",580)
            CALL section_vals_val_set(motion_section, "MD%TIME_START_VAL", r_val=t)
         END IF
      END IF
      IF (ASSOCIATED(constant)) THEN
         CALL section_vals_val_set(motion_section, "MD%ECONS_START_VAL", r_val=constant)
      END IF
      CALL timestop(handle2)
      ! AVERAGES
      CALL timeset(routineN//"_AVERAGES", handle2)
      IF (ASSOCIATED(averages)) THEN
         IF ((averages%do_averages) .AND. (averages%itimes_start /= -1)) THEN
            work_section => section_vals_get_subs_vals(motion_section, "MD%AVERAGES")
            CALL section_vals_val_set(work_section, "_SECTION_PARAMETERS_", l_val=averages%do_averages)
            work_section => section_vals_get_subs_vals(motion_section, "MD%AVERAGES%RESTART_AVERAGES")
            CALL section_vals_val_set(work_section, "ITIMES_START", i_val=averages%itimes_start)
            CALL section_vals_val_set(work_section, "AVECPU", r_val=averages%avecpu)
            CALL section_vals_val_set(work_section, "AVEHUGONIOT", r_val=averages%avehugoniot)
            CALL section_vals_val_set(work_section, "AVETEMP_BARO", r_val=averages%avetemp_baro)
            CALL section_vals_val_set(work_section, "AVEPOT", r_val=averages%avepot)
            CALL section_vals_val_set(work_section, "AVEKIN", r_val=averages%avekin)
            CALL section_vals_val_set(work_section, "AVETEMP", r_val=averages%avetemp)
            CALL section_vals_val_set(work_section, "AVEKIN_QM", r_val=averages%avekin_qm)
            CALL section_vals_val_set(work_section, "AVETEMP_QM", r_val=averages%avetemp_qm)
            CALL section_vals_val_set(work_section, "AVEVOL", r_val=averages%avevol)
            CALL section_vals_val_set(work_section, "AVECELL_A", r_val=averages%aveca)
            CALL section_vals_val_set(work_section, "AVECELL_B", r_val=averages%avecb)
            CALL section_vals_val_set(work_section, "AVECELL_C", r_val=averages%avecc)
            CALL section_vals_val_set(work_section, "AVEALPHA", r_val=averages%aveal)
            CALL section_vals_val_set(work_section, "AVEBETA", r_val=averages%avebe)
            CALL section_vals_val_set(work_section, "AVEGAMMA", r_val=averages%avega)
            CALL section_vals_val_set(work_section, "AVE_ECONS", r_val=averages%econs)
            CALL section_vals_val_set(work_section, "AVE_PRESS", r_val=averages%avepress)
            CALL section_vals_val_set(work_section, "AVE_PXX", r_val=averages%avepxx)
            ! Virial averages
            IF (ASSOCIATED(averages%virial)) THEN
               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_total, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_TOT", r_vals_ptr=buffer)

               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_virial, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_VIR", r_vals_ptr=buffer)

               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_kinetic, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_KIN", r_vals_ptr=buffer)

               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_constraint, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_CNSTR", r_vals_ptr=buffer)

               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_xc, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_XC", r_vals_ptr=buffer)

               ALLOCATE (buffer(9))
               buffer = RESHAPE(averages%virial%pv_fock_4c, (/9/))
               CALL section_vals_val_set(work_section, "AVE_PV_FOCK_4C", r_vals_ptr=buffer)
            END IF
            ! Colvars averages
            IF (SIZE(averages%avecolvar) > 0) THEN
               ALLOCATE (buffer(SIZE(averages%avecolvar)))
               buffer = averages%avecolvar
               CALL section_vals_val_set(work_section, "AVE_COLVARS", r_vals_ptr=buffer)
            END IF
            IF (SIZE(averages%aveMmatrix) > 0) THEN
               ALLOCATE (buffer(SIZE(averages%aveMmatrix)))
               buffer = averages%aveMmatrix
               CALL section_vals_val_set(work_section, "AVE_MMATRIX", r_vals_ptr=buffer)
            END IF
         END IF
      END IF
      CALL timestop(handle2)

      ! PARTICLE THERMOSTAT
      CALL timeset(routineN//"_THERMOSTAT_PARTICLES", handle2)
      IF (ASSOCIATED(thermostat_part)) THEN
         IF (thermostat_part%type_of_thermostat == do_thermo_nose) THEN
            ! Restart of Nose-Hoover Thermostat for Particles
            IF (.NOT. my_write_binary_restart_file) THEN
               nhc => thermostat_part%nhc
               CALL collect_nose_restart_info(nhc, para_env, eta, veta, fnhc, mnhc)
               work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%NOSE")
               CALL set_template_restart(work_section, eta, veta, fnhc, mnhc)
            END IF
         ELSE IF (thermostat_part%type_of_thermostat == do_thermo_csvr) THEN
            ! Restart of CSVR Thermostat for Particles
            work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%CSVR")
            CALL dump_csvr_restart_info(thermostat_part%csvr, para_env, work_section)
         ELSE IF (thermostat_part%type_of_thermostat == do_thermo_al) THEN
            ! Restart of AD_LANGEVIN Thermostat for Particles
            work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%AD_LANGEVIN")
            CALL dump_al_restart_info(thermostat_part%al, para_env, work_section)
         ELSE IF (thermostat_part%type_of_thermostat == do_thermo_gle) THEN
            ! Restart of GLE Thermostat for Particles
            work_section => section_vals_get_subs_vals(motion_section, "MD%THERMOSTAT%GLE")
            CALL dump_gle_restart_info(thermostat_part%gle, para_env, work_section)
         END IF
      END IF
      CALL timestop(handle2)

      ! BAROSTAT - THERMOSTAT
      CALL timeset(routineN//"_BAROSTAT", handle2)
      IF (ASSOCIATED(thermostat_baro)) THEN
         IF (thermostat_baro%type_of_thermostat == do_thermo_nose) THEN
            ! Restart of Nose-Hoover Thermostat for Barostat
            nhc => thermostat_baro%nhc
            nhc_len = SIZE(nhc%nvt, 1)
            tot_nhcneed = nhc%glob_num_nhc
            ALLOCATE (eta(tot_nhcneed*nhc_len))
            ALLOCATE (veta(tot_nhcneed*nhc_len))
            ALLOCATE (fnhc(tot_nhcneed*nhc_len))
            ALLOCATE (mnhc(tot_nhcneed*nhc_len))
            counter = 0
            DO i = 1, SIZE(nhc%nvt, 1)
               DO j = 1, SIZE(nhc%nvt, 2)
                  counter = counter+1
                  eta(counter) = nhc%nvt(i, j)%eta
                  veta(counter) = nhc%nvt(i, j)%v
                  fnhc(counter) = nhc%nvt(i, j)%f
                  mnhc(counter) = nhc%nvt(i, j)%mass
               END DO
            END DO
            work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%NOSE")
            CALL set_template_restart(work_section, eta, veta, fnhc, mnhc)
         ELSE IF (thermostat_baro%type_of_thermostat == do_thermo_csvr) THEN
            ! Restart of CSVR Thermostat for Barostat
            work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT%THERMOSTAT%CSVR")
            CALL dump_csvr_restart_info(thermostat_baro%csvr, para_env, work_section)
         END IF
      END IF
      CALL timestop(handle2)

      ! BAROSTAT
      CALL timeset(routineN//"_NPT", handle2)
      IF (ASSOCIATED(npt)) THEN
         ALLOCATE (veta(SIZE(npt, 1)*SIZE(npt, 2)))
         ALLOCATE (mnhc(SIZE(npt, 1)*SIZE(npt, 2)))
         counter = 0
         DO i = 1, SIZE(npt, 1)
            DO j = 1, SIZE(npt, 2)
               counter = counter+1
               veta(counter) = npt(i, j)%v
               mnhc(counter) = npt(i, j)%mass
            END DO
         END DO
         work_section => section_vals_get_subs_vals(motion_section, "MD%BAROSTAT")
         CALL set_template_restart(work_section, veta=veta, mnhc=mnhc)
      END IF
      CALL timestop(handle2)

      ! SHELL THERMOSTAT
      CALL timeset(routineN//"_THERMOSTAT_SHELL", handle2)
      IF (ASSOCIATED(thermostat_shell)) THEN
         IF (thermostat_shell%type_of_thermostat == do_thermo_nose) THEN
            ! Restart of Nose-Hoover Thermostat for Shell Particles
            IF (.NOT. my_write_binary_restart_file) THEN
               nhc => thermostat_shell%nhc
               CALL collect_nose_restart_info(nhc, para_env, eta, veta, fnhc, mnhc)
               work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%NOSE")
               CALL set_template_restart(work_section, eta, veta, fnhc, mnhc)
            END IF
         ELSE IF (thermostat_shell%type_of_thermostat == do_thermo_csvr) THEN
            work_section => section_vals_get_subs_vals(motion_section, "MD%SHELL%THERMOSTAT%CSVR")
            ! Restart of CSVR Thermostat for Shell Particles
            CALL dump_csvr_restart_info(thermostat_shell%csvr, para_env, work_section)
         END IF
      END IF
      CALL timestop(handle2)

      CALL timeset(routineN//"_META", handle2)
      IF (ASSOCIATED(meta_env)) THEN
         CALL section_vals_val_set(meta_env%metadyn_section, "STEP_START_VAL", &
                                   i_val=meta_env%n_steps)
         CALL section_vals_val_set(meta_env%metadyn_section, "NHILLS_START_VAL", &
                                   i_val=meta_env%hills_env%n_hills)
         !RG Adaptive hills
         CALL section_vals_val_set(meta_env%metadyn_section, "MIN_DISP", &
                                   r_val=meta_env%hills_env%min_disp)
         CALL section_vals_val_set(meta_env%metadyn_section, "OLD_HILL_NUMBER", &
                                   i_val=meta_env%hills_env%old_hill_number)
         CALL section_vals_val_set(meta_env%metadyn_section, "OLD_HILL_STEP", &
                                   i_val=meta_env%hills_env%old_hill_step)
         !RG Adaptive hills
         IF (meta_env%do_hills .AND. meta_env%hills_env%n_hills /= 0) THEN
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "SPAWNED_HILLS_POS")
            CALL meta_hills_val_set_ss(work_section, meta_env)
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "SPAWNED_HILLS_SCALE")
            CALL meta_hills_val_set_ds(work_section, meta_env)
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "SPAWNED_HILLS_HEIGHT")
            CALL meta_hills_val_set_ww(work_section, meta_env)
            IF (meta_env%well_tempered) THEN
               work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "SPAWNED_HILLS_INVDT")
               CALL meta_hills_val_set_dt(work_section, meta_env)
            END IF
         END IF
         IF (meta_env%extended_lagrange) THEN
            CALL section_vals_val_set(meta_env%metadyn_section, "COLVAR_AVG_TEMPERATURE_RESTART", &
                                      r_val=meta_env%avg_temp)
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_SS0")
            DO irep = 1, meta_env%n_colvar
               CALL section_vals_val_set(work_section, "_DEFAULT_KEYWORD_", r_val=meta_env%metavar(irep)%ss0, &
                                         i_rep_val=irep)
            END DO
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_VVP")
            DO irep = 1, meta_env%n_colvar
               CALL section_vals_val_set(work_section, "_DEFAULT_KEYWORD_", r_val=meta_env%metavar(irep)%vvp, &
                                         i_rep_val=irep)
            END DO

            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_SS")
            DO irep = 1, meta_env%n_colvar
               CALL section_vals_val_set(work_section, "_DEFAULT_KEYWORD_", r_val=meta_env%metavar(irep)%ss, &
                                         i_rep_val=irep)
            END DO
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_FS")
            DO irep = 1, meta_env%n_colvar
               CALL section_vals_val_set(work_section, "_DEFAULT_KEYWORD_", r_val=meta_env%metavar(irep)%ff_s, &
                                         i_rep_val=irep)
            END DO

         END IF
         ! Multiple Walkers
         IF (meta_env%do_multiple_walkers) THEN
            ALLOCATE (walkers_status(meta_env%multiple_walkers%walkers_tot_nr))
            walkers_status = meta_env%multiple_walkers%walkers_status
            work_section => section_vals_get_subs_vals(meta_env%metadyn_section, "MULTIPLE_WALKERS")
            CALL section_vals_val_set(work_section, "WALKERS_STATUS", i_vals_ptr=walkers_status)
         END IF
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_NEB", handle2)
      IF (PRESENT(coords) .OR. (PRESENT(vels))) THEN
         ! Update NEB section
         replica_section => section_vals_get_subs_vals(motion_section, "BAND%REPLICA")
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         IF (PRESENT(coords)) THEN
            ! Allocate possible missing sections
            DO
               IF (coords%size_wrk(2) <= SIZE(replica_section%values, 2)) EXIT
               CALL section_vals_add_values(replica_section)
            END DO
            ! Write Values
            DO isec = 1, coords%size_wrk(2)
               CALL section_vals_val_unset(replica_section, "COORD_FILE_NAME", i_rep_section=isec)
               work_section => section_vals_get_subs_vals3(replica_section, "COORD", i_rep_section=isec)
               CALL section_neb_coord_val_set(work_section, coords%xyz(:, isec), SIZE(coords%xyz, 1), 3*SIZE(particles%els), &
                                              3, particles%els, angstrom)
               ! Update Collective Variables
               IF (coords%in_use == do_band_collective) THEN
                  ALLOCATE (wrk(coords%size_wrk(1)))
                  wrk = coords%wrk(:, isec)
                  CALL section_vals_val_set(replica_section, "COLLECTIVE", r_vals_ptr=wrk, &
                                            i_rep_section=isec)
               END IF
            END DO
         END IF
         IF (PRESENT(vels)) THEN
            CALL force_env_get(force_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            ! Allocate possible missing sections
            DO
               IF (vels%size_wrk(2) <= SIZE(replica_section%values, 2)) EXIT
               CALL section_vals_add_values(replica_section)
            END DO
            ! Write Values
            DO isec = 1, vels%size_wrk(2)
               work_section => section_vals_get_subs_vals3(replica_section, "VELOCITY", i_rep_section=isec)
               IF (vels%in_use == do_band_collective) THEN
                  CALL section_neb_coord_val_set(work_section, vels%wrk(:, isec), SIZE(vels%wrk, 1), SIZE(vels%wrk, 1), &
                                                 1, particles%els, 1.0_dp)
               ELSE
                  CALL section_neb_coord_val_set(work_section, vels%wrk(:, isec), SIZE(vels%wrk, 1), 3*SIZE(particles%els), &
                                                 3, particles%els, 1.0_dp)
               END IF
            END DO
         END IF
      END IF
      CALL timestop(handle2)

      IF (PRESENT(pint_env)) THEN
         ! Update PINT section
         CALL update_motion_pint(motion_section, pint_env)
      END IF

      IF (PRESENT(helium_env)) THEN
         ! Update HELIUM section
         CALL update_motion_helium(helium_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE update_motion

! ***************************************************************************
!> \brief  Update PINT section in the input structure
!> \param motion_section ...
!> \param pint_env ...
!> \date   2010-10-13
!> \author Lukasz Walewski <Lukasz.Walewski@ruhr-uni-bochum.de>
! **************************************************************************************************
   SUBROUTINE update_motion_pint(motion_section, pint_env)

      TYPE(section_vals_type), POINTER                   :: motion_section
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_motion_pint', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: handle, i, iatom, ibead, inos, isp
      INTEGER, DIMENSION(rng_record_length, 1)           :: ascii
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_vals
      TYPE(section_vals_type), POINTER                   :: pint_section, tmpsec

      CALL timeset(routineN, handle)

      pint_section => section_vals_get_subs_vals(motion_section, "PINT")
      CALL section_vals_val_set(pint_section, "ITERATION", i_val=pint_env%iter)

      ! allocate memory for COORDs and VELOCITYs if the BEADS section was not
      ! explicitly given in the input (this is actually done only once since
      ! after section_vals_add_values section becomes explict)
      NULLIFY (tmpsec)
      tmpsec => section_vals_get_subs_vals(pint_section, "BEADS")
      CALL section_vals_get(tmpsec, explicit=explicit)
      IF (.NOT. explicit) THEN
         CALL section_vals_add_values(tmpsec)
      END IF

      ! update bead coordinates in the global input structure
      NULLIFY (r_vals)
      ALLOCATE (r_vals(pint_env%p*pint_env%ndim))

      i = 1
      CALL pint_u2x(pint_env)
      DO iatom = 1, pint_env%ndim
         DO ibead = 1, pint_env%p
            r_vals(i) = pint_env%x(ibead, iatom)
            i = i+1
         END DO
      END DO
      CALL section_vals_val_set(pint_section, "BEADS%COORD%_DEFAULT_KEYWORD_", &
                                r_vals_ptr=r_vals)

      ! update bead velocities in the global input structure
      NULLIFY (r_vals)
      ALLOCATE (r_vals(pint_env%p*pint_env%ndim))
      i = 1
      CALL pint_u2x(pint_env, ux=pint_env%uv, x=pint_env%v)
      DO iatom = 1, pint_env%ndim
         DO ibead = 1, pint_env%p
            r_vals(i) = pint_env%v(ibead, iatom)
            i = i+1
         END DO
      END DO
      CALL section_vals_val_set(pint_section, "BEADS%VELOCITY%_DEFAULT_KEYWORD_", &
                                r_vals_ptr=r_vals)

      IF (pint_env%pimd_thermostat == thermostat_nose) THEN

         ! allocate memory for COORDs and VELOCITYs if the NOSE section was not
         ! explicitly given in the input (this is actually done only once since
         ! after section_vals_add_values section becomes explict)
         NULLIFY (tmpsec)
         tmpsec => section_vals_get_subs_vals(pint_section, "NOSE")
         CALL section_vals_get(tmpsec, explicit=explicit)
         IF (.NOT. explicit) THEN
            CALL section_vals_add_values(tmpsec)
         END IF

         ! update thermostat coordinates in the global input structure
         NULLIFY (r_vals)
         ALLOCATE (r_vals(pint_env%p*pint_env%ndim*pint_env%nnos))
         i = 1
         DO iatom = 1, pint_env%ndim
            DO ibead = 1, pint_env%p
               DO inos = 1, pint_env%nnos
                  r_vals(i) = pint_env%tx(inos, ibead, iatom)
                  i = i+1
               END DO
            END DO
         END DO
         CALL section_vals_val_set(pint_section, "NOSE%COORD%_DEFAULT_KEYWORD_", &
                                   r_vals_ptr=r_vals)

         ! update thermostat velocities in the global input structure
         NULLIFY (r_vals)
         ALLOCATE (r_vals(pint_env%p*pint_env%ndim*pint_env%nnos))
         i = 1
         DO iatom = 1, pint_env%ndim
            DO ibead = 1, pint_env%p
               DO inos = 1, pint_env%nnos
                  r_vals(i) = pint_env%tv(inos, ibead, iatom)
                  i = i+1
               END DO
            END DO
         END DO
         CALL section_vals_val_set(pint_section, "NOSE%VELOCITY%_DEFAULT_KEYWORD_", &
                                   r_vals_ptr=r_vals)

      ELSEIF (pint_env%pimd_thermostat == thermostat_gle) THEN

         NULLIFY (tmpsec)
         tmpsec => section_vals_get_subs_vals(pint_section, "GLE")
         CALL dump_gle_restart_info(pint_env%gle, pint_env%replicas%para_env, tmpsec)

      ELSE IF (pint_env%pimd_thermostat == thermostat_pile) THEN
         tmpsec => section_vals_get_subs_vals(pint_section, &
                                              "PILE%RNG_INIT")
         CALL dump_rng_stream(rng_stream=pint_env%pile_therm%gaussian_rng_stream, &
                              rng_record=rng_record)
         CALL string_to_ascii(rng_record, ascii(:, 1))
         CALL section_rng_val_set(rng_section=tmpsec, nsize=1, &
                                  ascii=ascii)
      ELSE IF (pint_env%pimd_thermostat == thermostat_piglet) THEN
         tmpsec => section_vals_get_subs_vals(pint_section, &
                                              "PIGLET%RNG_INIT")
         CALL dump_rng_stream(rng_stream=pint_env%piglet_therm%gaussian_rng_stream, &
                              rng_record=rng_record)
         CALL string_to_ascii(rng_record, ascii(:, 1))
         CALL section_rng_val_set(rng_section=tmpsec, nsize=1, &
                                  ascii=ascii)
         ! update thermostat velocities in the global input structure
         NULLIFY (r_vals)
         ALLOCATE (r_vals((pint_env%piglet_therm%nsp1-1)* &
                          pint_env%piglet_therm%ndim* &
                          pint_env%piglet_therm%p))
         i = 1
         DO isp = 2, pint_env%piglet_therm%nsp1
            DO ibead = 1, pint_env%piglet_therm%p*pint_env%piglet_therm%ndim
               r_vals(i) = pint_env%piglet_therm%smalls(isp, ibead)
               i = i+1
            END DO
         END DO
         CALL section_vals_val_set(pint_section, "PIGLET%EXTRA_DOF%_DEFAULT_KEYWORD_", &
                                   r_vals_ptr=r_vals)
      END IF

      CALL timestop(handle)

   END SUBROUTINE update_motion_pint

! ***************************************************************************
!> \brief  Update HELIUM section in the input structure.
!> \param helium ...
!> \date   2009-11-12
!> \author Lukasz Walewski <Lukasz.Walewski@ruhr-uni-bochum.de>
!> \note Transfer the current helium state from the runtime environment
!>         to the input structure, so that it can be used for I/O, etc.
!> \note   Moved from the helium_io module directly, might be done better way
! **************************************************************************************************
   SUBROUTINE update_motion_helium(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_motion_helium', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: err_str, stmp
      INTEGER                                            :: handle, itmp, iweight, msglen, nsteps, &
                                                            offset, reqlen
      INTEGER, DIMENSION(:), POINTER                     :: int_msg_gather
      LOGICAL                                            :: lbf
      REAL(kind=dp)                                      :: bf, bu, invproc
      REAL(kind=dp), DIMENSION(3, 2)                     :: bg, cg, ig
      REAL(kind=dp), DIMENSION(:), POINTER               :: real_msg, real_msg_gather
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(helium)))CALL cp__a("motion/input_cp2k_restarts.F",1054)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (.NOT. helium%solute_present) THEN
         ! update iteration number
         itmp = logger%iter_info%iteration(2)
         CALL section_vals_val_set( &
            helium%input, &
            "MOTION%PINT%ITERATION", &
            i_val=itmp)
         ! else - PINT will do that
      END IF

      !
      ! save coordinates
      !
      ! allocate the buffer to be passed and fill it with local coords at each
      ! proc
      NULLIFY (real_msg)
      msglen = SIZE(helium%pos(:, :, 1:helium%beads))
      ALLOCATE (real_msg(msglen))
      real_msg(:) = PACK(helium%pos(:, :, 1:helium%beads), .TRUE.)

      ! allocate the buffer for message passing
      NULLIFY (real_msg_gather)
      msglen = SIZE(helium%pos(:, :, 1:helium%beads))*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))

      ! pass the message from all processors to logger%para_env%source
      real_msg_gather(:) = 0.0_dp
      CALL mp_gather(real_msg, real_msg_gather, &
                     logger%para_env%source, logger%para_env%group)

      ! update coordinates in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%COORD%_DEFAULT_KEYWORD_", &
                                r_vals_ptr=real_msg_gather)

      ! NULLIFY, but do not DEALLOCATE! - a new pointer to this array is silently
      ! assigned in section_vals_val_set - this memory will be used later on!
      ! "The val becomes the owner of the array" - from section_vals_val_set docu
      NULLIFY (real_msg_gather)

      ! DEALLOCATE since this array is only used locally
      DEALLOCATE (real_msg)

      !
      ! save permutation state
      !
      ! allocate the buffer for message passing
      NULLIFY (int_msg_gather)
      msglen = SIZE(helium%permutation)*logger%para_env%num_pe
      ALLOCATE (int_msg_gather(msglen))

      ! pass the message from all processors to logger%para_env%source
      int_msg_gather(:) = 0
      CALL mp_gather(helium%permutation, int_msg_gather, &
                     logger%para_env%source, logger%para_env%group)

      ! update permutation state in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%PERM%_DEFAULT_KEYWORD_", &
                                i_vals_ptr=int_msg_gather)

      ! NULLIFY, but do not DEALLOCATE! - a new pointer to this array is silently
      ! assigned in section_vals_val_set - this memory will be used later on!
      ! "The val becomes the owner of the array" - from section_vals_val_set docu
      NULLIFY (int_msg_gather)

      !
      ! save averages
      !
      ! update the weighting factor
      itmp = helium%averages_iweight
      IF (itmp .LT. 0) THEN
         itmp = helium%current_step-helium%first_step
      ELSE
         itmp = itmp+helium%current_step-helium%first_step
      END IF
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%IWEIGHT", &
                                i_val=itmp)

      ! allocate the buffer for message passing
      NULLIFY (real_msg_gather)
      msglen = 3*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))

      ! gather projected area from all processors
      CALL mp_gather(helium%proarea%ravr, &
                     real_msg_gather, &
                     logger%para_env%source, &
                     logger%para_env%group)

      ! update it in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA", &
                                r_vals_ptr=real_msg_gather)

      ! allocate the buffer for message passing
      NULLIFY (real_msg_gather)
      msglen = 3*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))

      ! gather projected area squared from all processors
      CALL mp_gather(helium%prarea2%ravr, &
                     real_msg_gather, &
                     logger%para_env%source, &
                     logger%para_env%group)

      ! update it in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA_2", &
                                r_vals_ptr=real_msg_gather)

      ! allocate the buffer for message passing
      NULLIFY (real_msg_gather)
      msglen = 3*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))

      ! gather winding number squared from all processors
      CALL mp_gather(helium%wnmber2%ravr, &
                     real_msg_gather, &
                     logger%para_env%source, &
                     logger%para_env%group)

      ! update it in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%WINDING_NUMBER_2", &
                                r_vals_ptr=real_msg_gather)

      ! allocate the buffer for message passing
      NULLIFY (real_msg_gather)
      msglen = 3*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))

      ! gather moment of inertia from all processors
      CALL mp_gather(helium%mominer%ravr, &
                     real_msg_gather, &
                     logger%para_env%source, &
                     logger%para_env%group)

      ! update it in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%MOMENT_OF_INERTIA", &
                                r_vals_ptr=real_msg_gather)

      ! NULLIFY, but do not DEALLOCATE! - a new pointer to this array is silently
      ! assigned in section_vals_val_set - this memory will be used later on!
      ! "The val becomes the owner of the array" - from section_vals_val_set docu
      NULLIFY (real_msg_gather)

      !
      ! save RNG state
      !
      ! pack RNG state on each processor to the local array
      NULLIFY (real_msg)
      msglen = 40
      ALLOCATE (real_msg(msglen))
      CALL get_rng_stream(helium%rng_stream_uniform, bg=bg, cg=cg, ig=ig, &
                          buffer=bu, buffer_filled=lbf)
      offset = 0
      real_msg(offset+1:offset+6) = PACK(bg, .TRUE.)
      real_msg(offset+7:offset+12) = PACK(cg, .TRUE.)
      real_msg(offset+13:offset+18) = PACK(ig, .TRUE.)
      IF (lbf) THEN
         bf = 1.0_dp
      ELSE
         bf = -1.0_dp
      END IF
      real_msg(offset+19) = bf
      real_msg(offset+20) = bu
      CALL get_rng_stream(helium%rng_stream_gaussian, bg=bg, cg=cg, ig=ig, &
                          buffer=bu, buffer_filled=lbf)
      offset = 20
      real_msg(offset+1:offset+6) = PACK(bg, .TRUE.)
      real_msg(offset+7:offset+12) = PACK(cg, .TRUE.)
      real_msg(offset+13:offset+18) = PACK(ig, .TRUE.)
      IF (lbf) THEN
         bf = 1.0_dp
      ELSE
         bf = -1.0_dp
      END IF
      real_msg(offset+19) = bf
      real_msg(offset+20) = bu

      ! Gather RNG state (in real_msg_gather vector) from all processors at
      ! logger%para_env%source
      NULLIFY (real_msg_gather)
      msglen = SIZE(real_msg)*logger%para_env%num_pe
      ALLOCATE (real_msg_gather(msglen))
      real_msg_gather(:) = 0.0_dp
      CALL mp_gather(real_msg, real_msg_gather, logger%para_env%source, logger%para_env%group)

      ! update the RNG state in the global input structure
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%RNG_STATE%_DEFAULT_KEYWORD_", &
                                r_vals_ptr=real_msg_gather)

      ! NULLIFY, but do not DEALLOCATE! - a new pointer to this array is silently
      ! assigned in section_vals_val_set - this memeory will be used later on!
      ! "The val becomes the owner of the array" - from section_vals_val_set docu
      NULLIFY (real_msg_gather)

      ! DEALLOCATE since this array is only used locally
      DEALLOCATE (real_msg)

      IF (helium%solute_present) THEN
         !
         ! save forces on the solute
         !
         ! check that the number of values match the current runtime
         reqlen = helium%solute_atoms*helium%solute_beads*3
         msglen = SIZE(helium%force_avrg)
         err_str = "Invalid size of HELIUM%FORCE: received '"
         stmp = ""
         WRITE (stmp, *) msglen
         err_str = TRIM(ADJUSTL(err_str))// &
                   TRIM(ADJUSTL(stmp))//"' but expected '"
         stmp = ""
         WRITE (stmp, *) reqlen
         err_str = TRIM(ADJUSTL(err_str))// &
                   TRIM(ADJUSTL(stmp))//"'."
         IF (msgLEN /= reqlen) &
            CALL cp__b("motion/input_cp2k_restarts.F",1280,err_str)

         ! allocate the buffer to be saved and fill it with forces
         ! forces should be the same on all processors, but we don't check that here
         NULLIFY (real_msg_gather)
         ALLOCATE (real_msg_gather(msglen))
         real_msg_gather(:) = PACK(helium%force_avrg, .TRUE.)

         ! update forces in the global input structure
         CALL section_vals_val_set(helium%input, &
                                   "MOTION%PINT%HELIUM%FORCE%_DEFAULT_KEYWORD_", &
                                   r_vals_ptr=real_msg_gather)

         ! NULLIFY, but do not DEALLOCATE! - a new pointer to this array is silently
         ! assigned in section_vals_val_set - this memeory will be used later on!
         ! "The val becomes the owner of the array" - from section_vals_val_set docu
         NULLIFY (real_msg_gather)
      END IF

      !
      ! save the RDFs
      !
      IF (helium%rdf_present) THEN

         ! work on the temporary array so that accumulated data remains intact
         helium%rdf_inst(:, :, :) = helium%rdf_accu(:, :, :)

         ! average over processors / helium environments
         CALL mp_sum(helium%rdf_inst, logger%para_env%group)
         itmp = logger%para_env%num_pe
         invproc = 1.0_dp/REAL(itmp, dp)
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)*invproc

         nsteps = helium%current_step-helium%first_step
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)/REAL(nsteps, dp)
         iweight = helium%rdf_iweight
         ! average over the old and the current density (observe the weights!)
         helium%rdf_inst(:, :, :) = nsteps*helium%rdf_inst(:, :, :)+ &
                                    iweight*helium%rdf_rstr(:, :, :)
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)/REAL(nsteps+iweight, dp)
         ! update in the global input structure
         NULLIFY (real_msg)
         msglen = SIZE(helium%rdf_inst)
         ALLOCATE (real_msg(msglen))
         real_msg(:) = PACK(helium%rdf_inst, .TRUE.)
         CALL section_vals_val_set( &
            helium%input, &
            "MOTION%PINT%HELIUM%AVERAGES%RDF", &
            r_vals_ptr=real_msg)
         NULLIFY (real_msg)

      END IF

      !
      ! save the densities
      !
      IF (helium%rho_present) THEN

         ! work on the temporary array so that accumulated data remains intact
         helium%rho_inst(:, :, :, :) = helium%rho_accu(:, :, :, :)

         ! average over processors / helium environments
         CALL mp_sum(helium%rho_inst, logger%para_env%group)
         itmp = logger%para_env%num_pe
         invproc = 1.0_dp/REAL(itmp, dp)
         helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)*invproc

         nsteps = helium%current_step-helium%first_step
         helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)/REAL(nsteps, dp)
         iweight = helium%averages_iweight
         ! average over the old and the current density (observe the weights!)
         helium%rho_inst(:, :, :, :) = nsteps*helium%rho_inst(:, :, :, :)+ &
                                       iweight*helium%rho_rstr(:, :, :, :)
         helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)/REAL(nsteps+iweight, dp)

         ! update the densities in the global input structure
         NULLIFY (real_msg)
         msglen = SIZE(helium%rho_inst)
         ALLOCATE (real_msg(msglen))
         real_msg(:) = PACK(helium%rho_inst, .TRUE.)
         CALL section_vals_val_set( &
            helium%input, &
            "MOTION%PINT%HELIUM%AVERAGES%RHO", &
            r_vals_ptr=real_msg)
         NULLIFY (real_msg)

      END IF

      CALL timestop(handle)

   END SUBROUTINE update_motion_helium

! **************************************************************************************************
!> \brief routine to dump thermostat CSVR energies
!> \param thermostat_energy ...
!> \param nsize ...
!> \param work_section ...
!> \par History
!>      10.2007 created [teo]
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE dump_csvr_energy_info(thermostat_energy, nsize, work_section)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: thermostat_energy
      INTEGER, INTENT(IN)                                :: nsize
      TYPE(section_vals_type), POINTER                   :: work_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_csvr_energy_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, Nlist
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      IF(.NOT.(ASSOCIATED(work_section)))CALL cp__a("motion/input_cp2k_restarts.F",1395)
      IF(.NOT.(work_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",1396)

      NULLIFY (my_val, old_val, section, vals)

      section => work_section%section

      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")

      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",1405), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      DO
         IF (SIZE(work_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(work_section)
      END DO

      vals => work_section%values(ik, 1)%list
      Nlist = 0

      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF

      DO irk = 1, nsize
         CALL val_create(val=my_val, r_val=thermostat_energy(irk))
         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      work_section%values(ik, 1)%list => vals

   END SUBROUTINE dump_csvr_energy_info

! **************************************************************************************************
!> \brief Collect all information needed to dump the restart for CSVR
!>      thermostat
!> \param csvr ...
!> \param para_env ...
!> \param csvr_section ...
!> \par History
!>      10.2007 created [tlaino] - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE dump_csvr_restart_info(csvr, para_env, csvr_section)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: csvr_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_csvr_restart_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: i, my_index
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: dwork
      REAL(KIND=dp)                                      :: dum
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: thermo_energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(section_vals_type), POINTER                   :: work_section

! Thermostat Energies

      ALLOCATE (work(csvr%glob_num_csvr))

      ALLOCATE (thermo_energy(csvr%loc_num_csvr))
      DO i = 1, csvr%loc_num_csvr
         thermo_energy(i) = csvr%nvt(i)%thermostat_energy
      END DO
      CALL get_kin_energies(csvr%map_info, csvr%loc_num_csvr, &
                            csvr%glob_num_csvr, thermo_energy, &
                            dum, para_env, array_kin=work)
      DEALLOCATE (thermo_energy)

      ! If check passes then let's dump the info on the restart file
      work_section => section_vals_get_subs_vals(csvr_section, "THERMOSTAT_ENERGY")
      CALL dump_csvr_energy_info(work, csvr%glob_num_csvr, work_section)
      DEALLOCATE (work)

      ! Thermostat Random Number info for restart
      work_section => section_vals_get_subs_vals(csvr_section, "RNG_INIT")
      ALLOCATE (dwork(rng_record_length, csvr%glob_num_csvr))
      dwork = 0
      DO i = 1, csvr%loc_num_csvr
         my_index = csvr%map_info%index(i)
         CALL dump_rng_stream(rng_stream=csvr%nvt(i)%gaussian_rng_stream, &
                              rng_record=rng_record)
         CALL string_to_ascii(rng_record, dwork(:, my_index))
      END DO

      !  Collect data if there was no communication in this thermostat
      IF (csvr%map_info%dis_type == do_thermo_no_communication) THEN
         ! Collect data if there was no communication in this thermostat
         CALL mp_sum(dwork, para_env%group)
      ELSE
         ! Perform some check and collect data in case of communicating thermostats
         CALL communication_thermo_low2(dwork, rng_record_length, csvr%glob_num_csvr, para_env)
      END IF
      CALL section_rng_val_set(rng_section=work_section, nsize=csvr%glob_num_csvr, ascii=dwork)
      DEALLOCATE (dwork)

   END SUBROUTINE dump_csvr_restart_info

! **************************************************************************************************
!> \brief Collect all information needed to dump the restart for AD_LANGEVIN
!>      thermostat
!> \param al ...
!> \param para_env ...
!> \param al_section ...
!> \par History
!>      10.2007 created [tlaino] - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE dump_al_restart_info(al, para_env, al_section)
      TYPE(al_system_type), POINTER                      :: al
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: al_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_al_restart_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: dum
      REAL(KIND=dp), DIMENSION(:), POINTER               :: t_array, work
      TYPE(section_vals_type), POINTER                   :: work_section

! chi and mass

      ALLOCATE (work(al%glob_num_al))
      ALLOCATE (t_array(al%loc_num_al))

      ! copy chi into temporary
      DO i = 1, al%loc_num_al
         t_array(i) = al%nvt(i)%chi
      END DO
      ! consolidate into work
      CALL get_kin_energies(al%map_info, al%loc_num_al, &
                            al%glob_num_al, t_array, &
                            dum, para_env, array_kin=work)

      ! If check passes then let's dump the info on the restart file
      work_section => section_vals_get_subs_vals(al_section, "CHI")
      CALL dump_csvr_energy_info(work, al%glob_num_al, work_section)

      ! copy mass into temporary
      DO i = 1, al%loc_num_al
         t_array(i) = al%nvt(i)%mass
      END DO
      ! consolidate into work
      CALL get_kin_energies(al%map_info, al%loc_num_al, &
                            al%glob_num_al, t_array, &
                            dum, para_env, array_kin=work)

      ! If check passes then let's dump the info on the restart file
      work_section => section_vals_get_subs_vals(al_section, "MASS")
      CALL dump_csvr_energy_info(work, al%glob_num_al, work_section)

      DEALLOCATE (t_array)
      DEALLOCATE (work)

   END SUBROUTINE dump_al_restart_info

! **************************************************************************************************
!> \brief Collect all information needed to dump the restart for GLE
!>      thermostat
!> \param gle ...
!> \param para_env ...
!> \param gle_section ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE dump_gle_restart_info(gle, para_env, gle_section)
      TYPE(gle_type), POINTER                            :: gle
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: gle_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_gle_restart_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: counter, glob_num, i, iproc, j, loc_num
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: dwork
      INTEGER, DIMENSION(:), POINTER                     :: gle_per_proc, index
      REAL(dp)                                           :: dum
      REAL(dp), DIMENSION(:), POINTER                    :: s_tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: thermo_energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(section_vals_type), POINTER                   :: work_section

! Thermostat Energies

      ALLOCATE (work(gle%glob_num_gle))
      ALLOCATE (thermo_energy(gle%loc_num_gle))
      DO i = 1, gle%loc_num_gle
         thermo_energy(i) = gle%nvt(i)%thermostat_energy
      END DO
      CALL get_kin_energies(gle%map_info, gle%loc_num_gle, &
                            gle%glob_num_gle, thermo_energy, &
                            dum, para_env, array_kin=work)
      DEALLOCATE (thermo_energy)

      ! If check passes then let's dump the info on the restart file
      work_section => section_vals_get_subs_vals(gle_section, "THERMOSTAT_ENERGY")
      CALL dump_csvr_energy_info(work, gle%glob_num_gle, work_section)
      DEALLOCATE (work)

      ! Thermostat Random Number info for restart
      work_section => section_vals_get_subs_vals(gle_section, "RNG_INIT")
      glob_num = gle%glob_num_gle
      loc_num = gle%loc_num_gle
      ALLOCATE (dwork(rng_record_length, glob_num))
      dwork = 0
      DO i = 1, loc_num
         j = gle%map_info%index(i)
         CALL dump_rng_stream(rng_stream=gle%nvt(i)%gaussian_rng_stream, &
                              rng_record=rng_record)
         CALL string_to_ascii(rng_record, dwork(:, j))
      END DO

      !  Collect data if there was no communication in this thermostat
      IF (gle%map_info%dis_type == do_thermo_no_communication) THEN
         ! Collect data if there was no communication in this thermostat
         CALL mp_sum(dwork, para_env%group)
      ELSE
         ! Perform some check and collect data in case of communicating thermostats
         CALL communication_thermo_low2(dwork, rng_record_length, glob_num, para_env)
      END IF
      CALL section_rng_val_set(rng_section=work_section, nsize=glob_num, ascii=dwork)
      DEALLOCATE (dwork)

      ALLOCATE (gle_per_proc(para_env%num_pe))
      gle_per_proc(:) = 0
      CALL mp_allgather(gle%loc_num_gle, gle_per_proc, para_env%group)

      ! Thermostat S variable info for restart
      NULLIFY (s_tmp)
      ALLOCATE (s_tmp((gle%ndim)*gle%glob_num_gle))
      s_tmp = 0.0_dp

      NULLIFY (work, index)
      DO iproc = 1, para_env%num_pe
         CALL reallocate(work, 1, gle_per_proc(iproc)*(gle%ndim))
         CALL reallocate(index, 1, gle_per_proc(iproc))
         IF (para_env%mepos == (iproc-1)) THEN
            INDEX(:) = 0
            counter = 0
            DO i = 1, gle%ndim
               DO j = 1, gle%loc_num_gle
                  counter = counter+1
                  work(counter) = gle%nvt(j)%s(i)
                  INDEX(j) = gle%map_info%index(j)
               END DO
            END DO
         ELSE
            work(:) = 0.0_dp
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         CALL mp_bcast(index, iproc-1, para_env%group)
         counter = 0
         DO i = 1, gle%ndim
            DO j = 1, gle_per_proc(iproc)
               counter = counter+1
               s_tmp((INDEX(j)-1)*(gle%ndim)+i) = work(counter)
            END DO
         END DO
      END DO

      IF (SIZE(s_tmp) > 0) THEN
         work_section => section_vals_get_subs_vals(gle_section, "S")
         CALL section_vals_val_set(work_section, "_DEFAULT_KEYWORD_", r_vals_ptr=s_tmp)
      ELSE
         DEALLOCATE (s_tmp)
      END IF

      DEALLOCATE (gle_per_proc)
      DEALLOCATE (work)
      DEALLOCATE (index)

   END SUBROUTINE dump_gle_restart_info

! **************************************************************************************************
!> \brief Collect all information needed to dump the restart for Nose-Hoover
!>      thermostat
!> \param nhc ...
!> \param para_env ...
!> \param eta ...
!> \param veta ...
!> \param fnhc ...
!> \param mnhc ...
!> \par History
!>      10.2007 created [tlaino] - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE collect_nose_restart_info(nhc, para_env, eta, veta, fnhc, mnhc)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eta, veta, fnhc, mnhc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'collect_nose_restart_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, i, iproc, j, nhc_len, num_nhc, &
                                                            numneed, stat, tot_nhcneed
      INTEGER, DIMENSION(:), POINTER                     :: index, nhc_per_proc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(map_info_type), POINTER                       :: map_info

      nhc_len = SIZE(nhc%nvt, 1)
      num_nhc = nhc%loc_num_nhc
      numneed = num_nhc
      map_info => nhc%map_info
      ALLOCATE (nhc_per_proc(para_env%num_pe), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1724, &
                                      "nhc_per_proc", int_size*para_env%num_pe)
      nhc_per_proc(:) = 0

      CALL mp_allgather(numneed, nhc_per_proc, para_env%group)
      tot_nhcneed = nhc%glob_num_nhc

      NULLIFY (work, index)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! nhc%eta
      !-----------------------------------------------------------------------------
      ALLOCATE (eta(tot_nhcneed*nhc_len), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1737, &
                                      "eta", dp_size*tot_nhcneed*nhc_len)
      DO iproc = 1, para_env%num_pe
         CALL reallocate(work, 1, nhc_per_proc(iproc)*nhc_len)
         CALL reallocate(index, 1, nhc_per_proc(iproc))
         IF (para_env%mepos == (iproc-1)) THEN
            INDEX(:) = 0
            counter = 0
            DO i = 1, nhc_len
               DO j = 1, num_nhc
                  counter = counter+1
                  work(counter) = nhc%nvt(i, j)%eta
                  INDEX(j) = map_info%index(j)
               END DO
            END DO
         ELSE
            work(:) = 0.0_dp
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         CALL mp_bcast(index, iproc-1, para_env%group)
         counter = 0
         DO i = 1, nhc_len
            DO j = 1, nhc_per_proc(iproc)
               counter = counter+1
               eta((INDEX(j)-1)*nhc_len+i) = work(counter)
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! nhc%veta
      !-----------------------------------------------------------------------------
      ALLOCATE (veta(tot_nhcneed*nhc_len), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1770, &
                                      "veta", dp_size*tot_nhcneed*nhc_len)
      DO iproc = 1, para_env%num_pe
         CALL reallocate(work, 1, nhc_per_proc(iproc)*nhc_len)
         CALL reallocate(index, 1, nhc_per_proc(iproc))
         IF (para_env%mepos == (iproc-1)) THEN
            INDEX(:) = 0
            counter = 0
            DO i = 1, nhc_len
               DO j = 1, num_nhc
                  counter = counter+1
                  work(counter) = nhc%nvt(i, j)%v
                  INDEX(j) = map_info%index(j)
               END DO
            END DO
         ELSE
            work(:) = 0.0_dp
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         CALL mp_bcast(index, iproc-1, para_env%group)
         counter = 0
         DO i = 1, nhc_len
            DO j = 1, nhc_per_proc(iproc)
               counter = counter+1
               veta((INDEX(j)-1)*nhc_len+i) = work(counter)
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! nhc%force
      !-----------------------------------------------------------------------------
      ALLOCATE (fnhc(tot_nhcneed*nhc_len), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1803, &
                                      "fnhc", dp_size*tot_nhcneed*nhc_len)
      DO iproc = 1, para_env%num_pe
         CALL reallocate(work, 1, nhc_per_proc(iproc)*nhc_len)
         CALL reallocate(index, 1, nhc_per_proc(iproc))
         IF (para_env%mepos == (iproc-1)) THEN
            INDEX(:) = 0
            counter = 0
            DO i = 1, nhc_len
               DO j = 1, num_nhc
                  counter = counter+1
                  work(counter) = nhc%nvt(i, j)%f
                  INDEX(j) = map_info%index(j)
               END DO
            END DO
         ELSE
            work(:) = 0.0_dp
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         CALL mp_bcast(index, iproc-1, para_env%group)
         counter = 0
         DO i = 1, nhc_len
            DO j = 1, nhc_per_proc(iproc)
               counter = counter+1
               fnhc((INDEX(j)-1)*nhc_len+i) = work(counter)
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! nhc%mass
      !-----------------------------------------------------------------------------
      ALLOCATE (mnhc(tot_nhcneed*nhc_len), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1836, &
                                      "mnhc", dp_size*tot_nhcneed*nhc_len)
      DO iproc = 1, para_env%num_pe
         CALL reallocate(work, 1, nhc_per_proc(iproc)*nhc_len)
         CALL reallocate(index, 1, nhc_per_proc(iproc))
         IF (para_env%mepos == (iproc-1)) THEN
            INDEX(:) = 0
            counter = 0
            DO i = 1, nhc_len
               DO j = 1, num_nhc
                  counter = counter+1
                  work(counter) = nhc%nvt(i, j)%mass
                  INDEX(j) = map_info%index(j)
               END DO
            END DO
         ELSE
            work(:) = 0.0_dp
         END IF
         CALL mp_bcast(work, iproc-1, para_env%group)
         CALL mp_bcast(index, iproc-1, para_env%group)
         counter = 0
         DO i = 1, nhc_len
            DO j = 1, nhc_per_proc(iproc)
               counter = counter+1
               mnhc((INDEX(j)-1)*nhc_len+i) = work(counter)
            END DO
         END DO
      END DO

      DEALLOCATE (work)
      DEALLOCATE (index)
      DEALLOCATE (nhc_per_proc)

   END SUBROUTINE collect_nose_restart_info

! **************************************************************************************************
!> \brief routine to dump NEB coordinates and velocities section.. fast implementation
!> \param coord_section ...
!> \param array ...
!> \param narray ...
!> \param nsize ...
!> \param nfield ...
!> \param particle_set ...
!> \param conv_factor ...
!> \par History
!>      12.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE section_neb_coord_val_set(coord_section, array, narray, nsize, nfield, &
                                        particle_set, conv_factor)

      TYPE(section_vals_type), POINTER                   :: coord_section
      REAL(KIND=dp), DIMENSION(*)                        :: array
      INTEGER, INTENT(IN)                                :: narray, nsize, nfield
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp)                                      :: conv_factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'section_neb_coord_val_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, Nlist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_c
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(coord_section)))CALL cp__a("motion/input_cp2k_restarts.F",1903)
      IF(.NOT.(coord_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",1904)
      section => coord_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",1908), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")
      DO
         IF (SIZE(coord_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(coord_section)
      END DO
      vals => coord_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      DO irk = 1, nsize/nfield
         ALLOCATE (my_c(nfield))
         IF (nfield == 3) THEN
            my_c(1:3) = get_particle_pos_or_vel(irk, particle_set, array(1:narray))
            my_c(1:3) = my_c(1:3)*conv_factor
         ELSE
            my_c(1) = array(irk)
         END IF
         CALL val_create(my_val, r_vals_ptr=my_c)

         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      coord_section%values(ik, 1)%list => vals
   END SUBROUTINE section_neb_coord_val_set

! **************************************************************************************************
!> \brief Set the nose structure like restart
!> \param work_section ...
!> \param eta ...
!> \param veta ...
!> \param fnhc ...
!> \param mnhc ...
!> \par History
!>      01.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE set_template_restart(work_section, eta, veta, fnhc, mnhc)
      TYPE(section_vals_type), POINTER                   :: work_section
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eta, veta, fnhc, mnhc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_template_restart', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: coord, force, mass, velocity

      NULLIFY (coord, force, velocity, mass)
      IF (PRESENT(eta)) THEN
         IF (SIZE(eta) > 0) THEN
            coord => section_vals_get_subs_vals(work_section, "COORD")
            CALL section_vals_val_set(coord, "_DEFAULT_KEYWORD_", r_vals_ptr=eta)
         ELSE
            DEALLOCATE (eta)
         END IF
      END IF
      IF (PRESENT(veta)) THEN
         IF (SIZE(veta) > 0) THEN
            velocity => section_vals_get_subs_vals(work_section, "VELOCITY")
            CALL section_vals_val_set(velocity, "_DEFAULT_KEYWORD_", r_vals_ptr=veta)
         ELSE
            DEALLOCATE (veta)
         END IF
      END IF
      IF (PRESENT(fnhc)) THEN
         IF (SIZE(fnhc) > 0) THEN
            force => section_vals_get_subs_vals(work_section, "FORCE")
            CALL section_vals_val_set(force, "_DEFAULT_KEYWORD_", r_vals_ptr=fnhc)
         ELSE
            DEALLOCATE (fnhc)
         END IF
      END IF
      IF (PRESENT(mnhc)) THEN
         IF (SIZE(mnhc) > 0) THEN
            mass => section_vals_get_subs_vals(work_section, "MASS")
            CALL section_vals_val_set(mass, "_DEFAULT_KEYWORD_", r_vals_ptr=mnhc)
         ELSE
            DEALLOCATE (mnhc)
         END IF
      END IF

   END SUBROUTINE set_template_restart

! **************************************************************************************************
!> \brief routine to dump hills information during metadynamics run
!> \param ss_section ...
!> \param meta_env ...
!> \par History
!>      02.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE meta_hills_val_set_ss(ss_section, meta_env)

      TYPE(section_vals_type), POINTER                   :: ss_section
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'meta_hills_val_set_ss', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, lsize, Nlist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ss_val
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(ss_section)))CALL cp__a("motion/input_cp2k_restarts.F",2033)
      IF(.NOT.(ss_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",2034)
      section => ss_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",2038), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")
      DO
         IF (SIZE(ss_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(ss_section)
      END DO
      vals => ss_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      lsize = SIZE(meta_env%hills_env%ss_history, 1)
      DO irk = 1, meta_env%hills_env%n_hills
         ALLOCATE (ss_val(lsize))
         ! Always stored in A.U.
         ss_val = meta_env%hills_env%ss_history(:, irk)
         CALL val_create(my_val, r_vals_ptr=ss_val)

         IF (irk <= Nlist) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      ss_section%values(ik, 1)%list => vals
   END SUBROUTINE meta_hills_val_set_ss

! **************************************************************************************************
!> \brief routine to dump hills information during metadynamics run
!> \param ds_section ...
!> \param meta_env ...
!> \par History
!>      02.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE meta_hills_val_set_ds(ds_section, meta_env)

      TYPE(section_vals_type), POINTER                   :: ds_section
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'meta_hills_val_set_ds', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, lsize, Nlist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ds_val
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(ds_section)))CALL cp__a("motion/input_cp2k_restarts.F",2104)
      IF(.NOT.(ds_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",2105)
      section => ds_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",2109), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")
      DO
         IF (SIZE(ds_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(ds_section)
      END DO
      vals => ds_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      lsize = SIZE(meta_env%hills_env%delta_s_history, 1)
      DO irk = 1, meta_env%hills_env%n_hills
         ALLOCATE (ds_val(lsize))
         ! Always stored in A.U.
         ds_val = meta_env%hills_env%delta_s_history(:, irk)
         CALL val_create(my_val, r_vals_ptr=ds_val)

         IF (irk <= Nlist) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      ds_section%values(ik, 1)%list => vals
   END SUBROUTINE meta_hills_val_set_ds

! **************************************************************************************************
!> \brief routine to dump hills information during metadynamics run
!> \param ww_section ...
!> \param meta_env ...
!> \par History
!>      02.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE meta_hills_val_set_ww(ww_section, meta_env)

      TYPE(section_vals_type), POINTER                   :: ww_section
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'meta_hills_val_set_ww', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, lsize, Nlist
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(ww_section)))CALL cp__a("motion/input_cp2k_restarts.F",2174)
      IF(.NOT.(ww_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",2175)
      section => ww_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",2179), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")
      DO
         IF (SIZE(ww_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(ww_section)
      END DO
      vals => ww_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      lsize = meta_env%hills_env%n_hills
      DO irk = 1, lsize
         CALL val_create(my_val, r_val=meta_env%hills_env%ww_history(irk))

         IF (irk <= Nlist) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      ww_section%values(ik, 1)%list => vals
   END SUBROUTINE meta_hills_val_set_ww

! **************************************************************************************************
!> \brief routine to dump hills information during metadynamics run
!> \param invdt_section ...
!> \param meta_env ...
!> \par History
!>      12.2009 created [seb]
!> \author SC
! **************************************************************************************************
   SUBROUTINE meta_hills_val_set_dt(invdt_section, meta_env)

      TYPE(section_vals_type), POINTER                   :: invdt_section
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'meta_hills_val_set_dt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, lsize, Nlist
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(invdt_section)))CALL cp__a("motion/input_cp2k_restarts.F",2241)
      IF(.NOT.(invdt_section%ref_count > 0))CALL cp__a("motion/input_cp2k_restarts.F",2242)
      section => invdt_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",2246), "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")
      DO
         IF (SIZE(invdt_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(invdt_section)
      END DO
      vals => invdt_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      lsize = meta_env%hills_env%n_hills
      DO irk = 1, lsize
         CALL val_create(my_val, r_val=meta_env%hills_env%invdt_history(irk))

         IF (irk <= Nlist) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      invdt_section%values(ik, 1)%list => vals
   END SUBROUTINE meta_hills_val_set_dt

! **************************************************************************************************
!> \brief   Write all input sections scaling in size with the number of atoms
!>          in the system to an external file in binary format
!> \param output_unit binary file to write to
!> \param log_unit unit for logging debug information
!> \param root_section ...
!> \param md_env ...
!> \param force_env ...
!> \par History
!>      - Creation (10.02.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_binary_restart(output_unit, log_unit, root_section, md_env, force_env)

      INTEGER, INTENT(IN)                                :: output_unit, log_unit
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_binary_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: binary_restart_file_name
      CHARACTER(LEN=default_string_length)               :: section_label
      INTEGER :: handle, iatom, icore, ikind, imolecule, ishell, istat, n_char_size, n_dp_size, &
         n_int_size, natom, natomkind, ncore, nhc_size, nmolecule, nmoleculekind, nshell, &
         print_level, run_type
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ibuf, imol
      LOGICAL                                            :: print_info, write_velocities
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rbuf
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: my_force_env
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(thermostat_type), POINTER                     :: thermostat_part, thermostat_shell

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kinds)
      NULLIFY (core_particles)
      NULLIFY (molecule_kinds)
      NULLIFY (molecules)
      NULLIFY (my_force_env)
      NULLIFY (para_env)
      NULLIFY (particles)
      NULLIFY (shell_particles)
      NULLIFY (subsys)
      NULLIFY (thermostat_part)
      NULLIFY (thermostat_shell)

      IF (PRESENT(md_env)) THEN
         CALL get_md_env(md_env=md_env, &
                         force_env=my_force_env, &
                         thermostat_part=thermostat_part, &
                         thermostat_shell=thermostat_shell)
      ELSE IF (PRESENT(force_env)) THEN
         my_force_env => force_env
      END IF

      IF (.NOT. ASSOCIATED(my_force_env)) THEN
         CALL timestop(handle)
         RETURN
      END IF

      CALL section_vals_val_get(root_section, "GLOBAL%PRINT_LEVEL", i_val=print_level)

      IF (print_level > 1) THEN
         print_info = .TRUE.
      ELSE
         print_info = .FALSE.
      END IF

      CALL section_vals_val_get(root_section, "GLOBAL%RUN_TYPE", i_val=run_type)
      write_velocities = ((run_type == mol_dyn_run) .OR. &
                          (run_type == mon_car_run) .OR. &
                          (run_type == pint_run))

      CALL force_env_get(force_env=my_force_env, &
                         para_env=para_env, &
                         subsys=subsys)
      CALL cp_subsys_get(subsys, &
                         atomic_kinds=atomic_kinds, &
                         particles=particles, &
                         natom=natom, &
                         core_particles=core_particles, &
                         ncore=ncore, &
                         shell_particles=shell_particles, &
                         nshell=nshell, &
                         molecule_kinds_new=molecule_kinds, &
                         molecules_new=molecules)

      natomkind = atomic_kinds%n_els
      IF (ASSOCIATED(molecule_kinds)) THEN
         nmoleculekind = molecule_kinds%n_els
      ELSE
         nmoleculekind = 0
      END IF

      IF (ASSOCIATED(molecules)) THEN
         nmolecule = molecules%n_els
      ELSE
         nmolecule = 0
      END IF

      n_char_size = 0 ! init
      n_int_size = 0 ! init
      n_dp_size = 0 ! init

      IF (output_unit > 0) THEN ! only ionode

         IF (print_info) THEN
            INQUIRE (UNIT=output_unit, NAME=binary_restart_file_name, IOSTAT=istat)
            IF (istat /= 0) THEN
               CALL cp_abort(cp__l("motion/input_cp2k_restarts.F",2405), &
                             "An error occurred inquiring logical unit <"// &
                             TRIM(ADJUSTL(cp_to_string(output_unit)))// &
                             "> which should be linked to the binary restart file")
            END IF
            IF (log_unit > 0) THEN
               WRITE (UNIT=log_unit, FMT="(T2,A,/,/,(T3,A,T71,I10))") &
                  "Writing binary restart file "//TRIM(ADJUSTL(binary_restart_file_name)), &
                  "Number of atomic kinds:", natomkind, &
                  "Number of atoms:", natom, &
                  "Number of cores (only core-shell model):", ncore, &
                  "Number of shells (only core-shell model):", nshell, &
                  "Number of molecule kinds:", nmoleculekind, &
                  "Number of molecules", nmolecule
            END IF

            n_int_size = n_int_size+6
         END IF

         WRITE (UNIT=output_unit, IOSTAT=istat) &
            natomkind, natom, ncore, nshell, nmoleculekind, nmolecule
         IF (istat /= 0) THEN
            CALL stop_write("natomkind,natom,ncore,nshell,nmoleculekind,nmolecule "// &
                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                            output_unit)
         END IF

         ! Write atomic kind names
         DO ikind = 1, natomkind
            WRITE (UNIT=output_unit, IOSTAT=istat) atomic_kinds%els(ikind)%name
            IF (istat /= 0) CALL stop_write("atomic_kinds%els(ikind)%name "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_char_size = n_char_size+LEN(atomic_kinds%els(ikind)%name)
         END DO

         ! Write atomic kind numbers of all atoms
         ALLOCATE (ibuf(natom), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2443, &
                                          "ibuf", int_size*natom)
         DO iatom = 1, natom
            ibuf(iatom) = particles%els(iatom)%atomic_kind%kind_number
         END DO
         WRITE (UNIT=output_unit, IOSTAT=istat) ibuf(1:natom)
         IF (istat /= 0) CALL stop_write("ibuf(1:natom) -> atomic kind numbers "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_int_size = n_int_size+natom
         ! Write atomic coordinates
         ALLOCATE (rbuf(3, natom), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2455, &
                                          "rbuf", dp_size*3*natom)
         DO iatom = 1, natom
            rbuf(1:3, iatom) = particles%els(iatom)%r(1:3)
         END DO
         WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:natom)
         IF (istat /= 0) CALL stop_write("rbuf(1:3,1:natom) -> atomic coordinates "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_dp_size = n_dp_size+3*natom
         DEALLOCATE (rbuf)

         ! Write molecule information if available
         IF (nmolecule > 0) THEN
            ! Write molecule kind names
            DO ikind = 1, nmoleculekind
               WRITE (UNIT=output_unit, IOSTAT=istat) molecule_kinds%els(ikind)%name
               IF (istat /= 0) CALL stop_write("molecule_kinds%els(ikind)%name "// &
                                               "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                               output_unit)
               n_char_size = n_char_size+LEN(molecule_kinds%els(ikind)%name)
            END DO
            ! Write molecule (kind) index numbers for all atoms
            ibuf(:) = 0
            ALLOCATE (imol(natom), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2480, &
                                             "imol", int_size*natom)
            imol(:) = 0
            DO imolecule = 1, nmolecule
               ikind = molecules%els(imolecule)%molecule_kind%kind_number
               DO iatom = molecules%els(imolecule)%first_atom, &
                          molecules%els(imolecule)%last_atom
                  ibuf(iatom) = ikind
                  imol(iatom) = imolecule
               END DO
            END DO
            ! Write molecule kind index number for each atom
            WRITE (UNIT=output_unit, IOSTAT=istat) ibuf(1:natom)
            IF (istat /= 0) CALL stop_write("ibuf(1:natom) -> molecule kind index numbers "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_int_size = n_int_size+natom
            ! Write molecule index number for each atom
            WRITE (UNIT=output_unit, IOSTAT=istat) imol(1:natom)
            IF (istat /= 0) CALL stop_write("imol(1:natom) -> molecule index numbers "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_int_size = n_int_size+natom
            DEALLOCATE (imol)
         END IF ! molecules

         DEALLOCATE (ibuf)

         ! Core-shell model only
         section_label = "SHELL COORDINATES"
         WRITE (UNIT=output_unit, IOSTAT=istat) section_label, nshell
         IF (istat /= 0) CALL stop_write("section_label, nshell "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (nshell > 0) THEN
            ! Write shell coordinates
            ALLOCATE (rbuf(3, nshell), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2519, &
                                             "rbuf", dp_size*3*nshell)
            DO ishell = 1, nshell
               rbuf(1:3, ishell) = shell_particles%els(ishell)%r(1:3)
            END DO
            WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:nshell)
            IF (istat /= 0) CALL stop_write("rbuf(1:3,1:nshell) -> shell coordinates "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_dp_size = n_dp_size+3*nshell
            DEALLOCATE (rbuf)
            ! Write atomic indices, i.e. number of the atom the shell belongs to
            ALLOCATE (ibuf(nshell), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2532, &
                                             "ibuf", int_size*nshell)
            DO ishell = 1, nshell
               ibuf(ishell) = shell_particles%els(ishell)%atom_index
            END DO
            WRITE (UNIT=output_unit, IOSTAT=istat) ibuf(1:nshell)
            IF (istat /= 0) CALL stop_write("ibuf(1:nshell) -> atomic indices "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_int_size = n_int_size+nshell
            DEALLOCATE (ibuf)
         END IF

         section_label = "CORE COORDINATES"
         WRITE (UNIT=output_unit, IOSTAT=istat) section_label, ncore
         IF (istat /= 0) CALL stop_write("section_label, ncore "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (ncore > 0) THEN
            ! Write core coordinates
            ALLOCATE (rbuf(3, ncore), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2555, &
                                             "rbuf", dp_size*3*ncore)
            DO icore = 1, ncore
               rbuf(1:3, icore) = core_particles%els(icore)%r(1:3)
            END DO
            WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:ncore)
            IF (istat /= 0) CALL stop_write("rbuf(1:3,1:ncore) -> core coordinates "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_dp_size = n_dp_size+3*ncore
            DEALLOCATE (rbuf)
            ! Write atomic indices, i.e. number of the atom the core belongs to
            ALLOCATE (ibuf(ncore), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2568, &
                                             "ibuf", int_size*ncore)
            DO icore = 1, ncore
               ibuf(icore) = core_particles%els(icore)%atom_index
            END DO
            WRITE (UNIT=output_unit, IOSTAT=istat) ibuf(1:ncore)
            IF (istat /= 0) CALL stop_write("ibuf(1:ncore) -> atomic indices "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_int_size = n_int_size+ncore
            DEALLOCATE (ibuf)
         END IF
      END IF ! ionode only

      ! Thermostat information

      ! Particle thermostats
      section_label = "PARTICLE THERMOSTATS"
      IF (ASSOCIATED(thermostat_part)) THEN
         ! Nose-Hoover thermostats
         IF (thermostat_part%type_of_thermostat == do_thermo_nose) THEN
            nhc => thermostat_part%nhc
            CALL write_binary_thermostats_nose(nhc, output_unit, log_unit, section_label, &
                                               n_char_size, n_dp_size, n_int_size, &
                                               print_info, para_env)
         END IF
      ELSE
         nhc_size = 0
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, IOSTAT=istat) section_label, nhc_size
            IF (istat /= 0) CALL stop_write(TRIM(section_label)//", nhc_size "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
         ENDIF
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (output_unit > 0 .AND. log_unit > 0) THEN ! only ionode
            IF (print_info) THEN
               WRITE (UNIT=log_unit, FMT="(T3,A,T71,I10)") &
                  "NHC size ("//TRIM(ADJUSTL(section_label))//")", nhc_size
            END IF
         END IF
      END IF

      ! Shell thermostats (only for core-shell models)
      section_label = "SHELL THERMOSTATS"
      IF (ASSOCIATED(thermostat_shell)) THEN
         ! Nose-Hoover thermostats
         IF (thermostat_shell%type_of_thermostat == do_thermo_nose) THEN
            nhc => thermostat_shell%nhc
            CALL write_binary_thermostats_nose(nhc, output_unit, log_unit, section_label, &
                                               n_char_size, n_dp_size, n_int_size, &
                                               print_info, para_env)
         END IF
      ELSE
         nhc_size = 0
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, IOSTAT=istat) section_label, nhc_size
            IF (istat /= 0) CALL stop_write("nhc_size "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
         ENDIF
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (output_unit > 0 .AND. log_unit > 0) THEN ! only ionode
            IF (print_info) THEN
               WRITE (UNIT=log_unit, FMT="(T3,A,T71,I10)") &
                  "NHC size ("//TRIM(ADJUSTL(section_label))//")", nhc_size
            END IF
         END IF
      END IF

      ! Particle velocities

      IF (output_unit > 0) THEN ! only ionode
         ! Write particle velocities if needed
         section_label = "VELOCITIES"
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, IOSTAT=istat) section_label, MERGE(natom, 0, write_velocities)
            IF (istat /= 0) CALL stop_write(TRIM(section_label)//", write_velocities "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
         ENDIF
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (print_info .AND. log_unit > 0) THEN
            WRITE (UNIT=log_unit, FMT="(T3,A,T78,A3)") &
               "Write "//TRIM(ADJUSTL(section_label))//" section", MERGE("YES", " NO", write_velocities)
         END IF
         IF (write_velocities) THEN
            ALLOCATE (rbuf(3, natom), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2659, &
                                             "rbuf", dp_size*3*natom)
            ! Write atomic velocities
            DO iatom = 1, natom
               rbuf(1:3, iatom) = particles%els(iatom)%v(1:3)
            END DO
            WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:natom)
            IF (istat /= 0) CALL stop_write("rbuf(1:3,1:natom) -> atomic velocities "// &
                                            "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                            output_unit)
            n_dp_size = n_dp_size+3*natom
            DEALLOCATE (rbuf)
         END IF
         ! Write shell velocities
         section_label = "SHELL VELOCITIES"
         WRITE (UNIT=output_unit, IOSTAT=istat) section_label, MERGE(nshell, 0, write_velocities)
         IF (istat /= 0) CALL stop_write(TRIM(section_label)//", write_velocities "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (print_info .AND. log_unit > 0) THEN
            WRITE (UNIT=log_unit, FMT="(T3,A,T78,A3)") &
               "Write "//TRIM(ADJUSTL(section_label))//" section", MERGE("YES", " NO", write_velocities)
         END IF
         IF (nshell > 0) THEN
            IF (write_velocities) THEN
               ALLOCATE (rbuf(3, nshell), STAT=istat)
               IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2687, &
                                                "rbuf", dp_size*3*nshell)
               DO ishell = 1, nshell
                  rbuf(1:3, ishell) = shell_particles%els(ishell)%v(1:3)
               END DO
               WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:nshell)
               IF (istat /= 0) CALL stop_write("rbuf(1:3,1:nshell) -> shell velocities "// &
                                               "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                               output_unit)
               n_dp_size = n_dp_size+3*nshell
               DEALLOCATE (rbuf)
            END IF
         END IF
         ! Write core velocities
         section_label = "CORE VELOCITIES"
         WRITE (UNIT=output_unit, IOSTAT=istat) section_label, MERGE(ncore, 0, write_velocities)
         IF (istat /= 0) CALL stop_write(TRIM(section_label)//", write_velocities "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (print_info .AND. log_unit > 0) THEN
            WRITE (UNIT=log_unit, FMT="(T3,A,T78,A3)") &
               "Write "//TRIM(ADJUSTL(section_label))//" section", MERGE("YES", " NO", write_velocities)
         END IF
         IF (ncore > 0) THEN
            IF (write_velocities) THEN
               ALLOCATE (rbuf(3, ncore), STAT=istat)
               IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2715, &
                                                "rbuf", dp_size*3*ncore)
               DO icore = 1, ncore
                  rbuf(1:3, icore) = core_particles%els(icore)%v(1:3)
               END DO
               WRITE (UNIT=output_unit, IOSTAT=istat) rbuf(1:3, 1:ncore)
               IF (istat /= 0) CALL stop_write("rbuf(1:3,1:ncore) -> core velocities "// &
                                               "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                               output_unit)
               n_dp_size = n_dp_size+3*ncore
               DEALLOCATE (rbuf)
            END IF
         END IF
      END IF ! ionode only

      ! Optionally, print a small I/O statistics
      IF (output_unit > 0) THEN ! only ionode
         IF (print_info .AND. log_unit > 0) THEN
            WRITE (UNIT=log_unit, FMT="(/,(T2,I10,1X,I0,A,T68,I10,A))") &
               n_char_size, int_size, "-byte characters written", n_char_size*int_size/1024, " KB", &
               n_dp_size, dp_size, "-byte floating point numbers written", n_dp_size*dp_size/1024, " KB", &
               n_int_size, int_size, "-byte integer numbers written", n_int_size*int_size/1024, " KB"
            WRITE (UNIT=log_unit, FMT="(/,T2,A)") &
               "Binary restart file "//TRIM(ADJUSTL(binary_restart_file_name))//" written"
         END IF
      END IF ! ionode only

      CALL timestop(handle)

   END SUBROUTINE write_binary_restart

! **************************************************************************************************
!> \brief   Write an input section for Nose thermostats to an external file in
!>          binary format
!> \param nhc ...
!> \param output_unit binary file to write to
!> \param log_unit unit for logging debug information
!> \param section_label ...
!> \param n_char_size ...
!> \param n_dp_size ...
!> \param n_int_size ...
!> \param print_info ...
!> \param para_env ...
!> \par History
!>      - Creation (23.03.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_binary_thermostats_nose(nhc, output_unit, log_unit, section_label, &
                                            n_char_size, n_dp_size, n_int_size, &
                                            print_info, para_env)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      INTEGER, INTENT(IN)                                :: output_unit, log_unit
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section_label
      INTEGER, INTENT(INOUT)                             :: n_char_size, n_dp_size, n_int_size
      LOGICAL, INTENT(IN)                                :: print_info
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_binary_thermostats_nose', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, istat, nhc_size
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eta, fnhc, mnhc, veta

      CALL timeset(routineN, handle)

      NULLIFY (eta)
      NULLIFY (fnhc)
      NULLIFY (mnhc)
      NULLIFY (veta)

      CALL collect_nose_restart_info(nhc, para_env, eta, veta, fnhc, mnhc)

      nhc_size = SIZE(eta)

      IF (output_unit > 0) THEN ! only ionode
         WRITE (UNIT=output_unit, IOSTAT=istat) section_label, nhc_size
         IF (istat /= 0) CALL stop_write("nhc_size "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_char_size = n_char_size+LEN(section_label)
         n_int_size = n_int_size+1
         IF (print_info .AND. log_unit > 0) THEN
            WRITE (UNIT=log_unit, FMT="(T3,A,T71,I10)") &
               "NHC size ("//TRIM(ADJUSTL(section_label))//")", nhc_size
         END IF
         ! eta
         WRITE (UNIT=output_unit, IOSTAT=istat) eta(1:nhc_size)
         IF (istat /= 0) CALL stop_write("eta(1:nhc_size) "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_dp_size = n_dp_size+nhc_size
      END IF ! ionode only

      DEALLOCATE (eta)

      ! veta
      IF (output_unit > 0) THEN ! only ionode
         WRITE (UNIT=output_unit, IOSTAT=istat) veta(1:nhc_size)
         IF (istat /= 0) CALL stop_write("veta(1:nhc_size) "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_dp_size = n_dp_size+nhc_size
      END IF ! ionode only

      DEALLOCATE (veta)

      ! mnhc
      IF (output_unit > 0) THEN ! only ionode
         WRITE (UNIT=output_unit, IOSTAT=istat) mnhc(1:nhc_size)
         IF (istat /= 0) CALL stop_write("mnhc(1:nhc_size) "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_dp_size = n_dp_size+nhc_size
      END IF ! ionode only

      DEALLOCATE (mnhc)

      ! fnhc
      IF (output_unit > 0) THEN ! only ionode
         WRITE (UNIT=output_unit, IOSTAT=istat) fnhc(1:nhc_size)
         IF (istat /= 0) CALL stop_write("fnhc(1:nhc_size) "// &
                                         "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                         output_unit)
         n_dp_size = n_dp_size+nhc_size
      END IF ! ionode only

      DEALLOCATE (fnhc)

      CALL timestop(handle)

   END SUBROUTINE write_binary_thermostats_nose

! **************************************************************************************************
!> \brief Print an error message and stop the program execution in case of a
!>        read error.
!> \param object ...
!> \param unit_number ...
!> \par History
!>      - Creation (15.02.2011,MK)
!> \author Matthias Krack (MK)
!> \note
!>      object     : Name of the data object for which I/O operation failed
!>      unit_number: Logical unit number of the file written to
! **************************************************************************************************
   SUBROUTINE stop_write(object, unit_number)
      CHARACTER(LEN=*), INTENT(IN)                       :: object
      INTEGER, INTENT(IN)                                :: unit_number

      CHARACTER(LEN=2*default_path_length)               :: message
      CHARACTER(LEN=default_path_length)                 :: file_name
      LOGICAL                                            :: file_exists

      IF (unit_number >= 0) THEN
         INQUIRE (UNIT=unit_number, EXIST=file_exists)
      ELSE
         file_exists = .FALSE.
      ENDIF
      IF (file_exists) THEN
         INQUIRE (UNIT=unit_number, NAME=file_name)
         WRITE (UNIT=message, FMT="(A)") &
            "An error occurred writing data object <"//TRIM(ADJUSTL(object))// &
            "> to file <"//TRIM(ADJUSTL(file_name))//">"
      ELSE
         WRITE (UNIT=message, FMT="(A,I0,A)") &
            "Could not write data object <"//TRIM(ADJUSTL(object))// &
            "> to logical unit ", unit_number, ". The I/O unit does not exist."
      END IF

      CALL cp__b("motion/input_cp2k_restarts.F",2885,message)

   END SUBROUTINE stop_write

END MODULE input_cp2k_restarts
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains a functional that calculates the energy and its derivatives
!>      for the geometry optimizer
!> \par History
!>      none
! **************************************************************************************************
MODULE gopt_f_methods
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_copy,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              init_cell,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type,&
                                              pack_subsys_particles
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE dimer_types,                     ONLY: dimer_env_type
   USE dimer_utils,                     ONLY: update_dimer_vec
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_get_natom,&
                                              force_env_get_nparticle,&
                                              force_env_type,&
                                              use_qmmm,&
                                              use_qmmmx
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_constants,                 ONLY: default_cell_direct_id,&
                                              default_cell_geo_opt_id,&
                                              default_cell_md_id,&
                                              default_cell_method_id,&
                                              default_minimization_method_id,&
                                              default_shellcore_method_id,&
                                              default_ts_method_id
   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE motion_utils,                    ONLY: write_simulation_cell,&
                                              write_stress_tensor,&
                                              write_trajectory
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: write_structure_data
   USE particle_types,                  ONLY: particle_type
   USE qmmm_util,                       ONLY: apply_qmmm_translate
   USE qmmmx_util,                      ONLY: apply_qmmmx_translate
   USE virial_methods,                  ONLY: virial_evaluate
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param error variable to control error logging, stopping,...
!>             see module cp_error_handling
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! *****************************************************************************
INTERFACE

  RECURSIVE SUBROUTINE cp_eval_at(gopt_env,x,f,gradient,master,&
                                  final_evaluation,para_env)

    USE cp_para_types,                   ONLY: cp_para_env_type
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE kinds,                           ONLY: dp

    TYPE(gopt_f_type), POINTER               :: gopt_env
    REAL(KIND=dp), DIMENSION(:), POINTER     :: x
    REAL(KIND=dp), INTENT(out), OPTIONAL     :: f
    REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
    INTEGER, INTENT(IN)                      :: master
    LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
    TYPE(cp_para_env_type), POINTER          :: para_env

  END SUBROUTINE cp_eval_at

END INTERFACE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gopt_f_methods'

   PUBLIC :: gopt_f_create_x0, &
             print_geo_opt_header, print_geo_opt_nc, &
             gopt_f_io_init, gopt_f_io, gopt_f_io_finalize, gopt_f_ii, &
             apply_cell_change

CONTAINS

! **************************************************************************************************
!> \brief returns the value of the parameters for the actual configuration
!> \param gopt_env the geometry optimization environment you want the info about
!>      x0: the parameter vector (is allocated by this routine)
!> \param x0 ...
!> \par History
!>      - Cell optimization revised (06.11.2012,MK)
! **************************************************************************************************
   SUBROUTINE gopt_f_create_x0(gopt_env, x0)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gopt_f_create_x0', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idg, j, nparticle
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys

      NULLIFY (cell)
      NULLIFY (subsys)

      SELECT CASE (gopt_env%type_id)
      CASE (default_minimization_method_id, default_ts_method_id)
         CALL force_env_get(gopt_env%force_env, subsys=subsys)
         ! before starting we handle the case of translating coordinates (QM/MM)
         IF (gopt_env%force_env%in_use == use_qmmm) &
            CALL apply_qmmm_translate(gopt_env%force_env%qmmm_env)
         IF (gopt_env%force_env%in_use == use_qmmmx) &
            CALL apply_qmmmx_translate(gopt_env%force_env%qmmmx_env)
         nparticle = force_env_get_nparticle(gopt_env%force_env)
         ALLOCATE (x0(3*nparticle))
         CALL pack_subsys_particles(subsys=subsys, r=x0)
      CASE (default_cell_method_id)
         SELECT CASE (gopt_env%cell_method_id)
         CASE (default_cell_direct_id)
            CALL force_env_get(gopt_env%force_env, subsys=subsys, cell=cell)
            ! Store reference cell
            gopt_env%h_ref = cell%hmat
            ! before starting we handle the case of translating coordinates (QM/MM)
            IF (gopt_env%force_env%in_use == use_qmmm) &
               CALL apply_qmmm_translate(gopt_env%force_env%qmmm_env)
            IF (gopt_env%force_env%in_use == use_qmmmx) &
               CALL apply_qmmmx_translate(gopt_env%force_env%qmmmx_env)
            nparticle = force_env_get_nparticle(gopt_env%force_env)
            ALLOCATE (x0(3*nparticle+6))
            CALL pack_subsys_particles(subsys=subsys, r=x0)
            idg = 3*nparticle
            DO i = 1, 3
               DO j = 1, i
                  idg = idg+1
                  x0(idg) = cell%hmat(j, i)
               END DO
            END DO
         CASE (default_cell_geo_opt_id, default_cell_md_id)
            CALL force_env_get(gopt_env%force_env, cell=cell)
            ALLOCATE (x0(6))
            idg = 0
            DO i = 1, 3
               DO j = 1, i
                  idg = idg+1
                  x0(idg) = cell%hmat(j, i)
               END DO
            END DO
         CASE DEFAULT
            CALL cp__b("motion/gopt_f_methods.F",145,"")
         END SELECT
      CASE DEFAULT
         CALL cp__b("motion/gopt_f_methods.F",148,"")
      END SELECT

   END SUBROUTINE gopt_f_create_x0

! **************************************************************************************************
!> \brief Prints iteration step of the optimization procedure on screen
!> \param its ...
!> \param output_unit ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE gopt_f_ii(its, output_unit)

      INTEGER, INTENT(IN)                                :: its, output_unit

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,26('-'))")
         WRITE (UNIT=output_unit, FMT="(T2,A,I6)") "OPTIMIZATION STEP: ", its
         WRITE (UNIT=output_unit, FMT="(T2,26('-'))")
         CALL m_flush(output_unit)
      END IF

   END SUBROUTINE gopt_f_ii

! **************************************************************************************************
!> \brief Handles the Output during an optimization run
!> \param gopt_env ...
!> \param output_unit ...
!> \param opt_energy ...
!> \param wildcard ...
!> \param its ...
!> \param used_time ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE gopt_f_io_init(gopt_env, output_unit, opt_energy, wildcard, its, used_time)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      INTEGER, INTENT(IN)                                :: output_unit
      REAL(KIND=dp)                                      :: opt_energy
      CHARACTER(LEN=5)                                   :: wildcard
      INTEGER, INTENT(IN)                                :: its
      REAL(KIND=dp)                                      :: used_time

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_io_init', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: pres_int

      SELECT CASE (gopt_env%type_id)
      CASE (default_ts_method_id, default_minimization_method_id)
         ! Geometry Optimization (Minimization and Transition State Search)
         IF (.NOT. gopt_env%dimer_rotation) THEN
            CALL write_cycle_infos(output_unit, it=its, etot=opt_energy, wildcard=wildcard, &
                                   used_time=used_time)
         ELSE
            CALL write_rot_cycle_infos(output_unit, it=its, etot=opt_energy, dimer_env=gopt_env%dimer_env, &
                                       wildcard=wildcard, used_time=used_time)
         END IF
      CASE (default_cell_method_id)
         ! Cell Optimization
         pres_int = gopt_env%cell_env%pres_int
         CALL write_cycle_infos(output_unit, it=its, etot=opt_energy, pres_int=pres_int, wildcard=wildcard, &
                                used_time=used_time)
      CASE (default_shellcore_method_id)
         CALL write_cycle_infos(output_unit, it=its, etot=opt_energy, wildcard=wildcard, &
                                used_time=used_time)
      END SELECT

   END SUBROUTINE gopt_f_io_init

! **************************************************************************************************
!> \brief Handles the Output during an optimization run
!> \param gopt_env ...
!> \param force_env ...
!> \param root_section ...
!> \param its ...
!> \param opt_energy ...
!> \param output_unit ...
!> \param eold ...
!> \param emin ...
!> \param wildcard ...
!> \param gopt_param ...
!> \param ndf ...
!> \param dx ...
!> \param xi ...
!> \param conv ...
!> \param pred ...
!> \param rat ...
!> \param step ...
!> \param rad ...
!> \param used_time ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE gopt_f_io(gopt_env, force_env, root_section, its, opt_energy, &
                        output_unit, eold, emin, wildcard, gopt_param, ndf, dx, xi, conv, pred, rat, &
                        step, rad, used_time)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: its
      REAL(KIND=dp), INTENT(IN)                          :: opt_energy
      INTEGER, INTENT(IN)                                :: output_unit
      REAL(KIND=dp)                                      :: eold, emin
      CHARACTER(LEN=5)                                   :: wildcard
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      INTEGER, INTENT(IN), OPTIONAL                      :: ndf
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: dx
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: xi
      LOGICAL, OPTIONAL                                  :: conv
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: pred, rat, step, rad
      REAL(KIND=dp)                                      :: used_time

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_io', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: pres_diff, pres_int, pres_tol

      SELECT CASE (gopt_env%type_id)
      CASE (default_ts_method_id, default_minimization_method_id)
         ! Geometry Optimization (Minimization and Transition State Search)
         IF (.NOT. gopt_env%dimer_rotation) THEN
            CALL geo_opt_io(force_env=force_env, root_section=root_section, &
                            motion_section=gopt_env%motion_section, its=its, opt_energy=opt_energy)
            CALL write_cycle_infos(output_unit, its, etot=opt_energy, ediff=opt_energy-eold, &
                                   pred=pred, rat=rat, step=step, rad=rad, emin=emin, wildcard=wildcard, used_time=used_time)
            ! Possibly check convergence
            IF (PRESENT(conv)) THEN
               IF(.NOT.(PRESENT(ndf)))CALL cp__a("motion/gopt_f_methods.F",272)
               IF(.NOT.(PRESENT(dx)))CALL cp__a("motion/gopt_f_methods.F",273)
               IF(.NOT.(PRESENT(xi)))CALL cp__a("motion/gopt_f_methods.F",274)
               CALL check_converg(ndf, dx, xi, output_unit, conv, gopt_param)
            END IF
         ELSE
            CALL update_dimer_vec(gopt_env%dimer_env, gopt_env%motion_section)
            CALL write_restart(force_env=force_env, root_section=root_section)
            CALL write_rot_cycle_infos(output_unit, its, opt_energy, opt_energy-eold, emin, gopt_env%dimer_env, &
                                       used_time=used_time, wildcard=wildcard)
            ! Possibly check convergence
            IF (PRESENT(conv)) THEN
               IF(.NOT.(ASSOCIATED(gopt_env%dimer_env)))CALL cp__a("motion/gopt_f_methods.F",284)
               CALL check_rot_conv(gopt_env%dimer_env, output_unit, conv)
            END IF
         END IF
      CASE (default_cell_method_id)
         ! Cell Optimization
         pres_diff = gopt_env%cell_env%pres_int-gopt_env%cell_env%pres_ext
         pres_int = gopt_env%cell_env%pres_int
         pres_tol = gopt_env%cell_env%pres_tol
         CALL geo_opt_io(force_env=force_env, root_section=root_section, &
                         motion_section=gopt_env%motion_section, its=its, opt_energy=opt_energy)
         CALL write_cycle_infos(output_unit, its, etot=opt_energy, ediff=opt_energy-eold, &
                                pred=pred, rat=rat, step=step, rad=rad, emin=emin, pres_int=pres_int, wildcard=wildcard, &
                                used_time=used_time)
         ! Possibly check convergence
         IF (PRESENT(conv)) THEN
            IF(.NOT.(PRESENT(ndf)))CALL cp__a("motion/gopt_f_methods.F",300)
            IF(.NOT.(PRESENT(dx)))CALL cp__a("motion/gopt_f_methods.F",301)
            IF(.NOT.(PRESENT(xi)))CALL cp__a("motion/gopt_f_methods.F",302)
            CALL check_converg(ndf, dx, xi, output_unit, conv, gopt_param, pres_diff, pres_tol)
         END IF
      CASE (default_shellcore_method_id)
         CALL write_cycle_infos(output_unit, its, etot=opt_energy, ediff=opt_energy-eold, &
                                pred=pred, rat=rat, step=step, rad=rad, emin=emin, wildcard=wildcard, used_time=used_time)
         ! Possibly check convergence
         IF (PRESENT(conv)) THEN
            IF(.NOT.(PRESENT(ndf)))CALL cp__a("motion/gopt_f_methods.F",310)
            IF(.NOT.(PRESENT(dx)))CALL cp__a("motion/gopt_f_methods.F",311)
            IF(.NOT.(PRESENT(xi)))CALL cp__a("motion/gopt_f_methods.F",312)
            CALL check_converg(ndf, dx, xi, output_unit, conv, gopt_param)
         END IF
      END SELECT
   END SUBROUTINE gopt_f_io

! **************************************************************************************************
!> \brief Handles the Output at the end of an optimization run
!> \param gopt_env ...
!> \param force_env ...
!> \param x0 ...
!> \param conv ...
!> \param its ...
!> \param root_section ...
!> \param para_env ...
!> \param master ...
!> \param output_unit ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   RECURSIVE SUBROUTINE gopt_f_io_finalize(gopt_env, force_env, x0, conv, its, root_section, &
                                           para_env, master, output_unit)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      LOGICAL                                            :: conv
      INTEGER                                            :: its
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: master, output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'gopt_f_io_finalize', &
         routineP = moduleN//':'//routineN

      IF (gopt_env%eval_opt_geo) THEN
         IF (.NOT. gopt_env%dimer_rotation) THEN
            CALL write_final_info(output_unit, conv, its, gopt_env, x0, master, &
                                  para_env, force_env, gopt_env%motion_section, root_section)
         ELSE
            CALL update_dimer_vec(gopt_env%dimer_env, gopt_env%motion_section)
            CALL write_restart(force_env=force_env, root_section=root_section)
         END IF
      END IF

   END SUBROUTINE gopt_f_io_finalize

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
!> \param it ...
!> \param etot ...
!> \param ediff ...
!> \param pred ...
!> \param rat ...
!> \param step ...
!> \param rad ...
!> \param emin ...
!> \param pres_int ...
!> \param wildcard ...
!> \param used_time ...
! **************************************************************************************************
   SUBROUTINE write_cycle_infos(output_unit, it, etot, ediff, pred, rat, step, rad, emin, &
                                pres_int, wildcard, used_time)

      INTEGER, INTENT(IN)                                :: output_unit, it
      REAL(KIND=dp), INTENT(IN)                          :: etot
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: ediff, pred, rat, step, rad, emin, &
                                                            pres_int
      CHARACTER(LEN=5), OPTIONAL                         :: wildcard
      REAL(KIND=dp), INTENT(IN)                          :: used_time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_cycle_infos', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: tmp_r1

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,8('-'),A,I5,1X,12('-'))") &
            "  Informations at step = ", it
         WRITE (UNIT=output_unit, FMT="(T2,A,T47,A)") &
            " Optimization Method        = ", wildcard
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Total Energy               = ", etot
         IF (PRESENT(pres_int)) THEN
            tmp_r1 = cp_unit_from_cp2k(pres_int, "bar")
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Internal Pressure [bar]    = ", tmp_r1
         END IF
         IF (PRESENT(ediff)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Real energy change         = ", ediff
         END IF
         IF (PRESENT(pred)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Predicted change in energy = ", pred
         END IF
         IF (PRESENT(rat)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Scaling factor             = ", rat
         END IF
         IF (PRESENT(step)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Step size                  = ", step
         END IF
         IF (PRESENT(rad)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Trust radius               = ", rad
         END IF
         IF (PRESENT(emin)) THEN
            IF (etot < emin) THEN
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Decrease in energy         =     ", &
                  "             YES"
            ELSE
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Decrease in energy         =     ", &
                  "              NO"
            END IF
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.3)") &
            " Used time                  = ", used_time
         IF (it == 0) WRITE (UNIT=output_unit, FMT="(T2,51('-'))")
      END IF
   END SUBROUTINE write_cycle_infos

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
!> \param it ...
!> \param etot ...
!> \param ediff ...
!> \param emin ...
!> \param dimer_env ...
!> \param used_time ...
!> \param wildcard ...
!> \date  01.2008
!> \author Luca Bellucci and Teodoro Laino - created [tlaino]
! **************************************************************************************************
   SUBROUTINE write_rot_cycle_infos(output_unit, it, etot, ediff, emin, dimer_env, used_time, wildcard)

      INTEGER, INTENT(IN)                                :: output_unit, it
      REAL(KIND=dp), INTENT(IN)                          :: etot
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: ediff, emin
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      REAL(KIND=dp)                                      :: used_time
      CHARACTER(LEN=5), OPTIONAL                         :: wildcard

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_rot_cycle_infos', &
         routineP = moduleN//':'//routineN

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,4('-'),A,I5,1X,5('-'))") &
            "  Informations at rotational step = ", it
         WRITE (UNIT=output_unit, FMT="(T2,A,T47,A)") &
            " Optimization Method        = ", wildcard
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Local Curvature            = ", dimer_env%rot%curvature
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Total Rotational Force     = ", etot
         IF (PRESENT(ediff)) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Real Force change          = ", ediff
         END IF
         IF (PRESENT(emin)) THEN
            IF (etot < emin) THEN
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Decrease in rotational force =   ", &
                  "             YES"
            ELSE
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Decrease in rotational force =   ", &
                  "              NO"
            END IF
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.3)") &
            " Used time                  = ", used_time
         IF (it == 0) WRITE (UNIT=output_unit, FMT="(T2,51('-'))")
      END IF
   END SUBROUTINE write_rot_cycle_infos

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param dr ...
!> \param g ...
!> \param output_unit ...
!> \param conv ...
!> \param gopt_param ...
!> \param pres_diff ...
!> \param pres_tol ...
! **************************************************************************************************
   SUBROUTINE check_converg(ndf, dr, g, output_unit, conv, gopt_param, pres_diff, pres_tol)

      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(IN)                          :: dr(ndf), g(ndf)
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(OUT)                               :: conv
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: pres_diff, pres_tol

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_converg', routineP = moduleN//':'//routineN

      INTEGER                                            :: indf
      LOGICAL                                            :: conv_dx, conv_g, conv_p, conv_rdx, &
                                                            conv_rg
      REAL(KIND=dp)                                      :: dumm, dxcon, gcon, maxdum(4), rmsgcon, &
                                                            rmsxcon, tmp_r1

      dxcon = gopt_param%max_dr
      gcon = gopt_param%max_force
      rmsgcon = gopt_param%rms_force
      rmsxcon = gopt_param%rms_dr

      conv = .FALSE.
      conv_dx = .TRUE.
      conv_rdx = .TRUE.
      conv_g = .TRUE.
      conv_rg = .TRUE.
      conv_p = .TRUE.

      dumm = 0.0_dp
      DO indf = 1, ndf
         IF (indf == 1) maxdum(1) = ABS(dr(indf))
         dumm = dumm+dr(indf)**2
         IF (ABS(dr(indf)) > dxcon) conv_dx = .FALSE.
         IF (ABS(dr(indf)) > maxdum(1)) maxdum(1) = ABS(dr(indf))
      END DO
      IF (SQRT(dumm/ndf) > rmsxcon) conv_rdx = .FALSE.
      maxdum(2) = SQRT(dumm/ndf)

      dumm = 0.0_dp
      DO indf = 1, ndf
         IF (indf == 1) maxdum(3) = ABS(g(indf))
         dumm = dumm+g(indf)**2
         IF (ABS(g(indf)) > gcon) conv_g = .FALSE.
         IF (ABS(g(indf)) > maxdum(3)) maxdum(3) = ABS(g(indf))
      END DO
      IF (SQRT(dumm/ndf) > rmsgcon) conv_rg = .FALSE.
      maxdum(4) = SQRT(dumm/ndf)

      IF (PRESENT(pres_diff) .AND. PRESENT(pres_tol)) conv_p = ABS(pres_diff) < ABS(pres_tol)

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            " Convergence check :"
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Max. step size             = ", maxdum(1)
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Conv. limit for step size  = ", dxcon
         IF (conv_dx) THEN
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in step size   =     ", &
               "             YES"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in step size   =     ", &
               "              NO"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " RMS step size              = ", maxdum(2)
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Conv. limit for RMS step   = ", rmsxcon
         IF (conv_rdx) THEN
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in RMS step    =     ", &
               "             YES"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in RMS step    =     ", &
               "              NO"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Max. gradient              = ", maxdum(3)
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Conv. limit for gradients  = ", gcon
         IF (conv_g) THEN
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Conv. in gradients         =     ", &
               "             YES"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Conv. for gradients        =     ", &
               "              NO"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " RMS gradient               = ", maxdum(4)
         WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
            " Conv. limit for RMS grad.  = ", rmsgcon
         IF (conv_rg) THEN
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Conv. in RMS gradients     =     ", &
               "             YES"

         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Conv. for gradients        =     ", &
               "              NO"
         END IF
         IF (PRESENT(pres_diff) .AND. PRESENT(pres_tol)) THEN
            tmp_r1 = cp_unit_from_cp2k(pres_diff, "bar")
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Pressure Deviation [bar]   = ", tmp_r1
            tmp_r1 = cp_unit_from_cp2k(pres_tol, "bar")
            WRITE (UNIT=output_unit, FMT="(T2,A,F20.10)") &
               " Pressure Tolerance [bar]   = ", tmp_r1
            IF (conv_p) THEN
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Conv. for  PRESSURE        =     ", &
                  "             YES"
            ELSE
               WRITE (UNIT=output_unit, FMT="(T2,2A)") &
                  " Conv. for  PRESSURE        =     ", &
                  "              NO"
            END IF
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,51('-'))")
      END IF

      IF (conv_dx .AND. conv_rdx .AND. conv_g .AND. conv_rg .AND. conv_p) conv = .TRUE.

      IF ((conv) .AND. (output_unit > 0)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T25,A,T78,A)") &
            "***", "GEOMETRY OPTIMIZATION COMPLETED", "***"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
      END IF

   END SUBROUTINE check_converg

! **************************************************************************************************
!> \brief ...
!> \param dimer_env ...
!> \param output_unit ...
!> \param conv ...
!> \date  01.2008
!> \author Luca Bellucci and Teodoro Laino - created [tlaino]
! **************************************************************************************************
   SUBROUTINE check_rot_conv(dimer_env, output_unit, conv)

      TYPE(dimer_env_type), POINTER                      :: dimer_env
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(OUT)                               :: conv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_rot_conv', routineP = moduleN//':'//routineN

      conv = (ABS(dimer_env%rot%angle2) < dimer_env%rot%angle_tol)
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            " Convergence check :"
         WRITE (UNIT=output_unit, FMT="(T2,A,F16.10)") &
            " Predicted angle step size  =     ", dimer_env%rot%angle1
         WRITE (UNIT=output_unit, FMT="(T2,A,F16.10)") &
            " Effective angle step size  =     ", dimer_env%rot%angle2
         WRITE (UNIT=output_unit, FMT="(T2,A,F16.10)") &
            " Conv. limit for angle step size =", dimer_env%rot%angle_tol
         IF (conv) THEN
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in angle step size   =", &
               "            YES"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,2A)") &
               " Convergence in angle step size   =", &
               "             NO"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,51('-'))")
      END IF
      IF ((conv) .AND. (output_unit > 0)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T25,A,T78,A)") &
            "***", "ROTATION OPTIMIZATION COMPLETED", "***"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
      END IF

   END SUBROUTINE check_rot_conv

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
!> \param conv ...
!> \param it ...
!> \param gopt_env ...
!> \param x0 ...
!> \param master ...
!> \param para_env ...
!> \param force_env ...
!> \param motion_section ...
!> \param root_section ...
!> \date  11.2007
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE write_final_info(output_unit, conv, it, gopt_env, x0, master, para_env, force_env, &
                                         motion_section, root_section)
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: conv
      INTEGER, INTENT(INOUT)                             :: it
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      INTEGER, INTENT(IN)                                :: master
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: motion_section, root_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_final_info', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: etot
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL force_env_get(force_env, cell=cell, subsys=subsys)
      CALL cp_subsys_get(subsys=subsys, particles=particles)
      particle_set => particles%els
      IF (conv) THEN
         it = it+1
         CALL write_structure_data(particle_set, cell, motion_section)
         CALL write_restart(force_env=force_env, root_section=root_section)

         IF (output_unit > 0) &
            WRITE (UNIT=output_unit, FMT="(/,T20,' Reevaluating energy at the minimum')")

         CALL cp_eval_at(gopt_env, x0, f=etot, master=master, final_evaluation=.TRUE., &
                         para_env=para_env)
         CALL write_geo_traj(force_env, root_section, it, etot)
      END IF

   END SUBROUTINE write_final_info

! **************************************************************************************************
!> \brief  Specific driver for dumping trajectory during a GEO_OPT
!> \param force_env ...
!> \param root_section ...
!> \param it ...
!> \param etot ...
!> \date   11.2007
!> \par    History
!>         09.2010: Output of core and shell positions and forces (MK)
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE write_geo_traj(force_env, root_section, it, etot)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: it
      REAL(KIND=dp), INTENT(IN)                          :: etot

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_geo_traj', routineP = moduleN//':'//routineN

      LOGICAL                                            :: shell_adiabatic, shell_present
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: core_particles, shell_particles

      NULLIFY (atomic_kinds)
      NULLIFY (atomic_kind_set)
      NULLIFY (core_particles)
      NULLIFY (shell_particles)
      NULLIFY (subsys)

      CALL write_trajectory(force_env, root_section, it, 0.0_dp, 0.0_dp, etot)
      CALL force_env_get(force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds)
      atomic_kind_set => atomic_kinds%els
      CALL get_atomic_kind_set(atomic_kind_set, &
                               shell_present=shell_present, &
                               shell_adiabatic=shell_adiabatic)
      IF (shell_present) THEN
         CALL cp_subsys_get(subsys, &
                            core_particles=core_particles, &
                            shell_particles=shell_particles)
         CALL write_trajectory(force_env, root_section, it=it, time=0.0_dp, dtime=0.0_dp, &
                               etot=etot, pk_name="SHELL_TRAJECTORY", middle_name="shpos", &
                               particles=shell_particles)
         IF (shell_adiabatic) THEN
            CALL write_trajectory(force_env, root_section, it=it, time=0.0_dp, dtime=0.0_dp, &
                                  etot=etot, pk_name="SHELL_FORCES", middle_name="shfrc", &
                                  particles=shell_particles)
            CALL write_trajectory(force_env, root_section, it=it, time=0.0_dp, dtime=0.0_dp, &
                                  etot=etot, pk_name="CORE_TRAJECTORY", middle_name="copos", &
                                  particles=core_particles)
            CALL write_trajectory(force_env, root_section, it=it, time=0.0_dp, dtime=0.0_dp, &
                                  etot=etot, pk_name="CORE_FORCES", middle_name="cofrc", &
                                  particles=core_particles)
         END IF
      END IF

   END SUBROUTINE write_geo_traj

! **************************************************************************************************
!> \brief ...
!> \param gopt_env ...
!> \param output_unit ...
!> \param label ...
!> \date  01.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE print_geo_opt_header(gopt_env, output_unit, label)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      INTEGER, INTENT(IN)                                :: output_unit
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_geo_opt_header', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_format, my_label
      INTEGER                                            :: ix

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("*", 79)
         IF (gopt_env%dimer_rotation) THEN
            my_label = "OPTIMIZING DIMER ROTATION"
         ELSE
            my_label = "STARTING "//gopt_env%tag(1:8)//" OPTIMIZATION"
         END IF

         ix = (80-7-LEN_TRIM(my_label))/2
         ix = ix+5
         my_format = "(T2,A,T"//cp_to_string(ix)//",A,T78,A)"
         WRITE (UNIT=output_unit, FMT=TRIM(my_format)) "***", TRIM(my_label), "***"

         ix = (80-7-LEN_TRIM(label))/2
         ix = ix+5
         my_format = "(T2,A,T"//cp_to_string(ix)//",A,T78,A)"
         WRITE (UNIT=output_unit, FMT=TRIM(my_format)) "***", TRIM(label), "***"

         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", 79)
         CALL m_flush(output_unit)
      END IF
   END SUBROUTINE print_geo_opt_header

! **************************************************************************************************
!> \brief ...
!> \param gopt_env ...
!> \param output_unit ...
!> \date  01.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE print_geo_opt_nc(gopt_env, output_unit)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      INTEGER, INTENT(IN)                                :: output_unit

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "*** MAXIMUM NUMBER OF OPTIMIZATION STEPS REACHED ***"
         IF (.NOT. gopt_env%dimer_rotation) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "***        EXITING GEOMETRY OPTIMIZATION         ***"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "***        EXITING ROTATION OPTIMIZATION         ***"
         END IF
         CALL m_flush(output_unit)
      END IF

   END SUBROUTINE print_geo_opt_nc

! **************************************************************************************************
!> \brief   Prints information during GEO_OPT common to all optimizers
!> \param force_env ...
!> \param root_section ...
!> \param motion_section ...
!> \param its ...
!> \param opt_energy ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE geo_opt_io(force_env, root_section, motion_section, its, opt_energy)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section, motion_section
      INTEGER, INTENT(IN)                                :: its
      REAL(KIND=dp), INTENT(IN)                          :: opt_energy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'geo_opt_io', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (para_env, atomic_kind_set, subsys, particle_set, &
               local_particles, atomic_kinds, particles)

      ! Write Restart File
      CALL write_restart(force_env=force_env, root_section=root_section)

      ! Write Trajectory
      CALL write_geo_traj(force_env, root_section, its, opt_energy)

      ! Write the stress Tensor
      CALL force_env_get(force_env, cell=cell, para_env=para_env, &
                         subsys=subsys)
      CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                         particles=particles, virial=virial)
      atomic_kind_set => atomic_kinds%els
      particle_set => particles%els
      CALL virial_evaluate(atomic_kind_set, particle_set, local_particles, &
                           virial, para_env%group)
      CALL write_stress_tensor(virial, cell, motion_section, its, 0.0_dp)

      ! Write the cell
      CALL write_simulation_cell(cell, motion_section, its, 0.0_dp)

   END SUBROUTINE geo_opt_io

! **************************************************************************************************
!> \brief   Apply coordinate transformations after cell (shape) change
!> \param gopt_env ...
!> \param cell ...
!> \param x ...
!> \param update_forces ...
!> \date    05.11.2012 (revised version of unbiase_coordinates moved here, MK)
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE apply_cell_change(gopt_env, cell, x, update_forces)

      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x
      LOGICAL, INTENT(IN)                                :: update_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_cell_change', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, idg, j, natom, nparticle, &
                                                            shell_index
      REAL(KIND=dp)                                      :: fc, fs, mass
      REAL(KIND=dp), DIMENSION(3)                        :: s
      TYPE(cell_type), POINTER                           :: cell_ref
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles

      NULLIFY (cell_ref)
      NULLIFY (core_particles)
      NULLIFY (particles)
      NULLIFY (shell_particles)
      NULLIFY (subsys)

      natom = force_env_get_natom(gopt_env%force_env)
      nparticle = force_env_get_nparticle(gopt_env%force_env)
      CALL force_env_get(gopt_env%force_env, &
                         subsys=subsys)
      CALL cp_subsys_get(subsys=subsys, &
                         core_particles=core_particles, &
                         particles=particles, &
                         shell_particles=shell_particles)

      ! Retrieve the reference cell
      CALL cell_create(cell_ref)
      CALL cell_copy(cell, cell_ref)

      ! Load the updated cell information
      SELECT CASE (gopt_env%cell_method_id)
      CASE (default_cell_direct_id)
         idg = 3*nparticle
         CALL init_cell(cell_ref, hmat=gopt_env%h_ref)
      CASE (default_cell_geo_opt_id, default_cell_md_id)
         idg = 0
      END SELECT
      IF(.NOT.((SIZE(x) == idg+6)))CALL cp__a("motion/gopt_f_methods.F",982)

      IF (update_forces) THEN

         ! Transform particle forces back to reference cell
         idg = 1
         DO iatom = 1, natom
            CALL real_to_scaled(s, x(idg:idg+2), cell)
            CALL scaled_to_real(x(idg:idg+2), s, cell_ref)
            idg = idg+3
         END DO

      ELSE

         ! Update cell
         DO i = 1, 3
            DO j = 1, i
               idg = idg+1
               cell%hmat(j, i) = x(idg)
            END DO
         END DO
         CALL init_cell(cell)
         CALL cp_subsys_set(subsys, cell=cell)

         ! Retrieve particle coordinates for the current cell
         SELECT CASE (gopt_env%cell_method_id)
         CASE (default_cell_direct_id)
            idg = 1
            DO iatom = 1, natom
               CALL real_to_scaled(s, x(idg:idg+2), cell_ref)
               shell_index = particles%els(iatom)%shell_index
               IF (shell_index == 0) THEN
                  CALL scaled_to_real(particles%els(iatom)%r, s, cell)
               ELSE
                  CALL scaled_to_real(core_particles%els(shell_index)%r, s, cell)
                  i = 3*(natom+shell_index-1)+1
                  CALL real_to_scaled(s, x(i:i+2), cell_ref)
                  CALL scaled_to_real(shell_particles%els(shell_index)%r, s, cell)
                  ! Update atomic position due to core and shell motion
                  mass = particles%els(iatom)%atomic_kind%mass
                  fc = core_particles%els(shell_index)%atomic_kind%shell%mass_core/mass
                  fs = shell_particles%els(shell_index)%atomic_kind%shell%mass_shell/mass
                  particles%els(iatom)%r(1:3) = fc*core_particles%els(shell_index)%r(1:3)+ &
                                                fs*shell_particles%els(shell_index)%r(1:3)
               END IF
               idg = idg+3
            END DO
         CASE (default_cell_geo_opt_id, default_cell_md_id)
            DO iatom = 1, natom
               shell_index = particles%els(iatom)%shell_index
               IF (shell_index == 0) THEN
                  CALL real_to_scaled(s, particles%els(iatom)%r, cell_ref)
                  CALL scaled_to_real(particles%els(iatom)%r, s, cell)
               ELSE
                  CALL real_to_scaled(s, core_particles%els(shell_index)%r, cell_ref)
                  CALL scaled_to_real(core_particles%els(shell_index)%r, s, cell)
                  i = 3*(natom+shell_index-1)+1
                  CALL real_to_scaled(s, shell_particles%els(shell_index)%r, cell_ref)
                  CALL scaled_to_real(shell_particles%els(shell_index)%r, s, cell)
                  ! Update atomic position due to core and shell motion
                  mass = particles%els(iatom)%atomic_kind%mass
                  fc = core_particles%els(shell_index)%atomic_kind%shell%mass_core/mass
                  fs = shell_particles%els(shell_index)%atomic_kind%shell%mass_shell/mass
                  particles%els(iatom)%r(1:3) = fc*core_particles%els(shell_index)%r(1:3)+ &
                                                fs*shell_particles%els(shell_index)%r(1:3)
               END IF
            END DO
         END SELECT

      END IF

      CALL cell_release(cell_ref)

   END SUBROUTINE apply_cell_change

END MODULE gopt_f_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for Geometry optimization using  Conjugate Gradients
!> \author Teodoro Laino [teo]
!>      10.2005
! **************************************************************************************************
MODULE cg_optimizer

   USE cell_types,                      ONLY: cell_type
   USE cg_utils,                        ONLY: cg_linmin,&
                                              get_conjugate_direction
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_methods,                  ONLY: gopt_f_ii,&
                                              gopt_f_io,&
                                              gopt_f_io_finalize,&
                                              gopt_f_io_init,&
                                              print_geo_opt_header,&
                                              print_geo_opt_nc
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param error variable to control error logging, stopping,...
!>             see module cp_error_handling
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! *****************************************************************************
INTERFACE

  RECURSIVE SUBROUTINE cp_eval_at(gopt_env,x,f,gradient,master,&
                                  final_evaluation,para_env)

    USE cp_para_types,                   ONLY: cp_para_env_type
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE kinds,                           ONLY: dp

    TYPE(gopt_f_type), POINTER               :: gopt_env
    REAL(KIND=dp), DIMENSION(:), POINTER     :: x
    REAL(KIND=dp), INTENT(out), OPTIONAL     :: f
    REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
    INTEGER, INTENT(IN)                      :: master
    LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
    TYPE(cp_para_env_type), POINTER          :: para_env

  END SUBROUTINE cp_eval_at

END INTERFACE

   PUBLIC :: geoopt_cg
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cg_optimizer'

CONTAINS

! **************************************************************************************************
!> \brief Driver for conjugate gradient optimization technique
!> \param force_env ...
!> \param gopt_param ...
!> \param globenv ...
!> \param geo_section ...
!> \param gopt_env ...
!> \param x0 ...
!> \param do_update ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE geoopt_cg(force_env, gopt_param, globenv, geo_section, &
                                  gopt_env, x0, do_update)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      LOGICAL, INTENT(OUT), OPTIONAL                     :: do_update

      CHARACTER(len=*), PARAMETER :: routineN = 'geoopt_cg', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      LOGICAL                                            :: my_do_update
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      output_unit = cp_print_key_unit_nr(logger, geo_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".geoLog")
      CALL print_geo_opt_header(gopt_env, output_unit, "CONJUGATE GRADIENTS")
      CALL cp_cg_main(force_env, x0, gopt_param, output_unit, globenv, &
                      gopt_env, do_update=my_do_update)
      CALL cp_print_key_finished_output(output_unit, logger, geo_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      IF (PRESENT(do_update)) do_update = my_do_update

      CALL timestop(handle)

   END SUBROUTINE geoopt_cg

! **************************************************************************************************
!> \brief This really performs the conjugate gradients optimization
!> \param force_env ...
!> \param x0 ...
!> \param gopt_param ...
!> \param output_unit ...
!> \param globenv ...
!> \param gopt_env ...
!> \param do_update ...
!> \par History
!>      10.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_cg_main(force_env, x0, gopt_param, output_unit, globenv, &
                                   gopt_env, do_update)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      LOGICAL, INTENT(OUT), OPTIONAL                     :: do_update

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cg_main', routineP = moduleN//':'//routineN

      CHARACTER(LEN=5)                                   :: wildcard
      INTEGER                                            :: handle, iter_nr, its, max_steep_steps, &
                                                            maxiter
      LOGICAL                                            :: conv, Fletcher_Reeves, &
                                                            save_consistent_energy_force, &
                                                            should_stop
      REAL(KIND=dp)                                      :: emin, eold, opt_energy, res_lim, t_diff, &
                                                            t_now, t_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: xold
      REAL(KIND=dp), DIMENSION(:), POINTER               :: g, h, xi
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: root_section

      CALL timeset(routineN, handle)
      t_old = m_walltime()
      NULLIFY (logger, g, h, xi)
      root_section => force_env%root_section
      logger => cp_get_default_logger()
      conv = .FALSE.
      maxiter = gopt_param%max_iter
      max_steep_steps = gopt_param%max_steep_steps
      Fletcher_Reeves = gopt_param%Fletcher_Reeves
      res_lim = gopt_param%restart_limit
      ALLOCATE (g(SIZE(x0)))
      ALLOCATE (h(SIZE(x0)))
      ALLOCATE (xi(SIZE(x0)))
      ALLOCATE (xold(SIZE(x0)))
      CALL force_env_get(force_env, cell=cell)

      ! Evaluate energy and forces at the first step
      ![NB] consistent energies and forces not required for CG, but some line minimizers might set it
      save_consistent_energy_force = gopt_env%require_consistent_energy_force
      gopt_env%require_consistent_energy_force = .FALSE.
      CALL cp_eval_at(gopt_env, x0, opt_energy, xi, gopt_env%force_env%para_env%mepos, &
                      .FALSE., gopt_env%force_env%para_env)
      gopt_env%require_consistent_energy_force = save_consistent_energy_force

      g = -xi
      h = g
      xi = h
      emin = HUGE(0.0_dp)
      CALL cp_iterate(logger%iter_info, increment=0, iter_nr_out=iter_nr)
      ! Main Loop
      wildcard = "   SD"
      t_now = m_walltime()
      t_diff = t_now-t_old
      t_old = t_now
      CALL gopt_f_io_init(gopt_env, output_unit, opt_energy, wildcard, used_time=t_diff, its=iter_nr)
      eold = opt_energy
      DO its = iter_nr+1, maxiter
         CALL cp_iterate(logger%iter_info, last=(its == maxiter))
         CALL section_vals_val_set(gopt_env%geo_section, "STEP_START_VAL", i_val=its)
         CALL gopt_f_ii(its, output_unit)

         xold(:) = x0

         ! Line minimization
         CALL cg_linmin(gopt_env, x0, xi, g, opt_energy, output_unit, gopt_param, globenv)

         ! Check for an external exit command
         CALL external_control(should_stop, "GEO", globenv=globenv)
         IF (should_stop) EXIT

         ! Some IO and Convergence check
         t_now = m_walltime()
         t_diff = t_now-t_old
         t_old = t_now
         CALL gopt_f_io(gopt_env, force_env, root_section, its, opt_energy, &
                        output_unit, eold, emin, wildcard, gopt_param, SIZE(x0), x0-xold, xi, conv, &
                        used_time=t_diff)
         eold = opt_energy
         emin = MIN(emin, opt_energy)

         IF (conv .OR. (its == maxiter)) EXIT
         ![NB] consistent energies and forces not required for CG, but some line minimizers might set it
         save_consistent_energy_force = gopt_env%require_consistent_energy_force
         gopt_env%require_consistent_energy_force = .FALSE.
         CALL cp_eval_at(gopt_env, x0, opt_energy, xi, gopt_env%force_env%para_env%mepos, &
                         .FALSE., gopt_env%force_env%para_env)
         gopt_env%require_consistent_energy_force = save_consistent_energy_force

         ! Get Conjugate Directions:  updates the searching direction (h)
         wildcard = "   CG"
         CALL get_conjugate_direction(gopt_env, Fletcher_Reeves, g, xi, h)

         ! Reset Condition or Steepest Descent Requested
         IF (ABS(DOT_PRODUCT(g, h))/SQRT((DOT_PRODUCT(g, g)*DOT_PRODUCT(h, h))) > res_lim &
             .OR. its+1 <= max_steep_steps) THEN
            ! Steepest Descent
            wildcard = "   SD"
            h = -xi
         END IF
         g = -xi
         xi = h
      END DO
      IF (its == maxiter .AND. (.NOT. conv)) THEN
         CALL print_geo_opt_nc(gopt_env, output_unit)
      END IF

      ! Write final particle information and restart, if converged
      IF (PRESENT(do_update)) do_update = conv
      CALL cp_iterate(logger%iter_info, last=.TRUE., increment=0)
      CALL gopt_f_io_finalize(gopt_env, force_env, x0, conv, its, root_section, &
                              gopt_env%force_env%para_env, gopt_env%force_env%para_env%mepos, output_unit)

      DEALLOCATE (xold)
      DEALLOCATE (g)
      DEALLOCATE (h)
      DEALLOCATE (xi)

      CALL timestop(handle)

   END SUBROUTINE cp_cg_main

END MODULE cg_optimizer
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for Geometry optimization using BFGS algorithm
! **************************************************************************************************
MODULE bfgs_optimizer

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: get_atomic_kind,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE constraint_fxd,                  ONLY: fix_atom_control
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_external_control,             ONLY: external_control
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_transpose
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_read_unformatted,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type,&
                                              cp_fm_write_unformatted
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_iterate,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_methods,                  ONLY: gopt_f_ii,&
                                              gopt_f_io,&
                                              gopt_f_io_finalize,&
                                              gopt_f_io_init,&
                                              print_geo_opt_header,&
                                              print_geo_opt_nc
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_constants,                 ONLY: default_ts_method_id
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_min,&
                                              mp_sum
   USE particle_list_types,             ONLY: particle_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE
  PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param error variable to control error logging, stopping,...
!>             see module cp_error_handling
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! *****************************************************************************
INTERFACE

  RECURSIVE SUBROUTINE cp_eval_at(gopt_env,x,f,gradient,master,&
                                  final_evaluation,para_env)

    USE cp_para_types,                   ONLY: cp_para_env_type
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE kinds,                           ONLY: dp

    TYPE(gopt_f_type), POINTER               :: gopt_env
    REAL(KIND=dp), DIMENSION(:), POINTER     :: x
    REAL(KIND=dp), INTENT(out), OPTIONAL     :: f
    REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
    INTEGER, INTENT(IN)                      :: master
    LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
    TYPE(cp_para_env_type), POINTER          :: para_env

  END SUBROUTINE cp_eval_at

END INTERFACE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'bfgs_optimizer'
  LOGICAL, PARAMETER                       :: debug_this_module=.TRUE.

  PUBLIC :: geoopt_bfgs

CONTAINS

! **************************************************************************************************
!> \brief Main driver for BFGS geometry optimizations
!> \param force_env ...
!> \param gopt_param ...
!> \param globenv ...
!> \param geo_section ...
!> \param gopt_env ...
!> \param x0 ...
! **************************************************************************************************
  RECURSIVE SUBROUTINE geoopt_bfgs(force_env,gopt_param,globenv,geo_section,gopt_env,x0)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0

      CHARACTER(len=*), PARAMETER :: routineN = 'geoopt_bfgs', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      CHARACTER(LEN=5)                                   :: wildcard
      CHARACTER(LEN=default_path_length)                 :: hes_filename
      INTEGER                                            :: handle, hesunit_read, indf, info, &
                                                            iter_nr, its, maxiter, ndf, nfree, &
                                                            output_unit
      LOGICAL                                            :: cell_opt, conv, hesrest, ionode, &
                                                            shell_present, should_stop, &
                                                            use_mod_hes, use_rfo
      REAL(KIND=dp)                                      :: ediff, emin, eold, etot, pred, rad, rat, &
                                                            step, t_diff, t_now, t_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dg, dr, dx, eigval, gold, work, xold
      REAL(KIND=dp), DIMENSION(:), POINTER               :: g
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_hes
      TYPE(cp_fm_type), POINTER                          :: eigvec_mat, hess_mat, hess_tmp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: print_key, root_section

    NULLIFY(logger, g,blacs_env)
    logger       => cp_get_default_logger()
    para_env     => force_env%para_env
    root_section => force_env%root_section
    t_old=m_walltime()

    CALL timeset(routineN,handle)
    CALL section_vals_val_get(geo_section,"BFGS%TRUST_RADIUS",r_val=rad)
    print_key    => section_vals_get_subs_vals(geo_section,"BFGS%RESTART")
    ionode      = para_env%mepos==para_env%source
    maxiter     = gopt_param%max_iter
    cell_opt    = .FALSE.
    conv        = .FALSE.
    rat         = 0.0_dp
    wildcard    = " BFGS"

    ! Stop if not yet implemented
    SELECT CASE (gopt_env%type_id)
    CASE (default_ts_method_id)
       CALL cp__b("motion/bfgs_optimizer.F",147,"BFGS method not yet working with DIMER")
    END SELECT

    CALL section_vals_val_get(geo_section,"BFGS%USE_RAT_FUN_OPT",l_val=use_rfo)
    CALL section_vals_val_get(geo_section,"BFGS%USE_MODEL_HESSIAN",l_val=use_mod_hes)
    CALL section_vals_val_get(geo_section,"BFGS%RESTART_HESSIAN",l_val=hesrest)
    output_unit = cp_print_key_unit_nr(logger,geo_section,"PRINT%PROGRAM_RUN_INFO",&
                                       extension=".geoLog")

    ndf   = SIZE(x0)
    nfree = gopt_env%nfree
    IF(ndf > 3000)&
       CALL cp_warn(cp__l("motion/bfgs_optimizer.F",159),&
            "The dimension of the Hessian matrix ("//&
            TRIM(ADJUSTL(cp_to_string(ndf)))//") is greater than 3000. "//&
            "The diagonalisation of the full Hessian  matrix needed for BFGS "//&
            "is computationally expensive. You should consider to use the linear "//&
            "scaling variant L-BFGS instead.")

    ! Initialize hessian (hes = unitary matrix or model hessian )
    CALL cp_blacs_env_create(blacs_env,para_env,globenv%blacs_grid_layout,&
                             globenv%blacs_repeatable)
    CALL cp_fm_struct_create(fm_struct_hes,para_env=para_env,context=blacs_env, &
         nrow_global=ndf,ncol_global=ndf)
    CALL cp_fm_create(hess_mat, fm_struct_hes,name="hess_mat")
    CALL cp_fm_create(hess_tmp, fm_struct_hes,name="hess_tmp")
    CALL cp_fm_create(eigvec_mat, fm_struct_hes,name="eigvec_mat")
    ALLOCATE (eigval(ndf))
    eigval(:) = 0.0_dp

    CALL force_env_get(force_env=force_env,subsys=subsys)
    CALL cp_subsys_get(subsys,atomic_kinds=atomic_kinds)
    CALL get_atomic_kind_set(atomic_kind_set=atomic_kinds%els,shell_present=shell_present)
    IF (shell_present) THEN
      IF(use_mod_hes)&
         CALL cp_warn(cp__l("motion/bfgs_optimizer.F",182),&
              "No model Hessian available for core-shell models. "//&
              "An unit matrix is used for the initial Hessian.")
      use_mod_hes = .FALSE.
    END IF

    IF (use_mod_hes) THEN
       CALL cp_fm_set_all(hess_mat,alpha=zero,beta=0.00_dp)
       CALL construct_initial_hess(gopt_env%force_env,hess_mat)
       CALL cp_fm_to_fm(hess_mat,hess_tmp)
       CALL choose_eigv_solver(hess_tmp,eigvec_mat,eigval,info=info)
       ! In rare cases the diagonalization of hess_mat fails (bug in scalapack?)
       IF(info/=0) THEN
          CALL cp_fm_set_all(hess_mat,alpha=zero,beta=one)
          IF(output_unit>0) WRITE(output_unit,*) &
             "BFGS: Matrix diagonalization failed, using unity as model Hessian."
       ELSE
          DO its=1,SIZE(eigval)
            IF(eigval(its).lt.0.1_dp)eigval(its)=0.1_dp
          END DO
          CALL cp_fm_to_fm(eigvec_mat,hess_tmp)
          CALL cp_fm_column_scale(eigvec_mat,eigval)
          CALL cp_gemm("N","T",ndf,ndf,ndf,one,hess_tmp,eigvec_mat,zero,hess_mat)
       ENDIF
    ELSE
       CALL cp_fm_set_all(hess_mat,alpha=zero,beta=one)
    END IF

    ALLOCATE (xold(ndf))
    xold(:) = x0(:)

    ALLOCATE (g(ndf))
    g(:) = 0.0_dp

    ALLOCATE (gold(ndf))
    gold(:) = 0.0_dp

    ALLOCATE (dx(ndf))
    dx(:) = 0.0_dp

    ALLOCATE (dg(ndf))
    dg(:) = 0.0_dp

    ALLOCATE (work(ndf))
    work(:) = 0.0_dp

    ALLOCATE (dr(ndf))
    dr(:) = 0.0_dp

    ! refernce cell needed for smooth cell optimizations
    CALL force_env_get(gopt_env%force_env,cell=cell)

    ! Geometry optimization starts now
    CALL cp_iterate(logger%iter_info,increment=0,iter_nr_out=iter_nr)
    CALL print_geo_opt_header(gopt_env, output_unit, wildcard)

    ! Calculate Energy & Gradients
    CALL cp_eval_at(gopt_env, x0, etot, g, gopt_env%force_env%para_env%mepos,&
                    .FALSE., gopt_env%force_env%para_env)

    ! Print info at time 0
    emin = etot
    t_now=m_walltime()
    t_diff=t_now-t_old
    t_old=t_now
    CALL gopt_f_io_init(gopt_env, output_unit, etot, wildcard=wildcard, its=iter_nr, used_time=t_diff)
    DO its = iter_nr+1, maxiter
       CALL cp_iterate(logger%iter_info,last=(its==maxiter))
       CALL section_vals_val_set(geo_section,"STEP_START_VAL",i_val=its)
       CALL gopt_f_ii(its, output_unit)

       ! Hessian update/restarting
       IF (((its-iter_nr) == 1).AND.hesrest) THEN
          IF(ionode)THEN
             CALL section_vals_val_get(geo_section,"BFGS%RESTART_FILE_NAME",c_val=hes_filename)
             CALL open_file(file_name=hes_filename,file_status="OLD",&
                  file_form="UNFORMATTED", file_action="READ", unit_number=hesunit_read)
          END IF
          CALL cp_fm_read_unformatted(hess_mat,hesunit_read)
          IF (ionode) CALL close_file(unit_number=hesunit_read)
       ELSE
          IF( (its-iter_nr) > 1 ) THEN
            DO indf = 1, ndf
               dx(indf) = x0(indf) - xold(indf)
               dg(indf) =  g(indf) - gold(indf)
            END DO

            CALL bfgs(ndf,dx,dg,hess_mat,work,para_env)

            !Possibly dump the Hessian file
            IF (BTEST(cp_print_key_should_output(logger%iter_info,print_key),cp_p_file))THEN
               CALL write_bfgs_hessian(geo_section,hess_mat,logger)
            ENDIF
          ENDIF
       END IF

       ! Setting the present positions & gradients as old
       xold(:) = x0
       gold(:) = g

       ! Copying hessian hes to (ndf x ndf) matrix hes_mat for diagonalization
       CALL cp_fm_to_fm(hess_mat,hess_tmp)

       CALL choose_eigv_solver(hess_tmp,eigvec_mat,eigval,info=info)

       ! In rare cases the diagonalization of hess_mat fails (bug in scalapack?)
       IF(info/=0) THEN
          IF(output_unit>0) WRITE(output_unit,*) &
             "BFGS: Matrix diagonalization failed, resetting Hessian to unity."
          CALL cp_fm_set_all(hess_mat,alpha=zero,beta=one)
          CALL cp_fm_to_fm(hess_mat,hess_tmp)
          CALL choose_eigv_solver(hess_tmp,eigvec_mat,eigval)
       END IF

       IF(use_rfo)THEN
          CALL set_hes_eig(ndf,eigval,work)
          dx(:) = eigval
          CALL rat_fun_opt(ndf,dg,eigval,work,eigvec_mat,g,para_env)
       END IF
       CALL geoopt_get_step(ndf,eigval,eigvec_mat,hess_tmp,dr,g,para_env,use_rfo)
       CALL trust_radius(ndf,step,rad,rat,dr,output_unit)

       ! Update the atomic positions
       x0 = x0 + dr

       CALL energy_predict(ndf,work,hess_mat,dr,g,conv,pred,para_env)
       eold = etot

       ! Energy & Gradients at new step
       CALL cp_eval_at(gopt_env, x0, etot, g, gopt_env%force_env%para_env%mepos,&
                       .FALSE., gopt_env%force_env%para_env)
       ediff = etot - eold

       ! check for an external exit command
       CALL external_control(should_stop,"GEO",globenv=globenv)
       IF(should_stop) EXIT

       ! Some IO and Convergence check
       t_now=m_walltime()
       t_diff=t_now-t_old
       t_old=t_now
       CALL gopt_f_io(gopt_env, force_env, root_section, its, etot, output_unit,&
            eold, emin, wildcard, gopt_param, ndf, dr, g, conv, pred, rat,&
            step, rad, used_time=t_diff)

       IF (conv.OR.(its==maxiter)) EXIT
       IF (etot < emin) emin = etot
       IF (use_rfo) CALL update_trust_rad(rat,rad,step,ediff)
    END DO

    IF (its == maxiter .AND. (.NOT.conv))THEN
       CALL print_geo_opt_nc(gopt_env, output_unit)
    END IF

    ! Write final  information, if converged
    CALL cp_iterate(logger%iter_info,last=.TRUE.,increment=0)
    CALL write_bfgs_hessian(geo_section,hess_mat,logger)
    CALL gopt_f_io_finalize(gopt_env, force_env, x0, conv, its, root_section,&
         gopt_env%force_env%para_env, gopt_env%force_env%para_env%mepos, output_unit)

    CALL cp_fm_struct_release(fm_struct_hes)
    CALL cp_fm_release(hess_mat)
    CALL cp_fm_release(eigvec_mat)
    CALL cp_fm_release(hess_tmp)

    CALL cp_blacs_env_release(blacs_env)
    DEALLOCATE (xold)
    DEALLOCATE (g)
    DEALLOCATE (gold)
    DEALLOCATE (dx)
    DEALLOCATE (dg)
    DEALLOCATE (eigval)
    DEALLOCATE (work)
    DEALLOCATE (dr)

    CALL cp_print_key_finished_output(output_unit,logger,geo_section,&
         "PRINT%PROGRAM_RUN_INFO")
    CALL timestop(handle)

  END SUBROUTINE geoopt_bfgs

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param dg ...
!> \param eigval ...
!> \param work ...
!> \param eigvec_mat ...
!> \param g ...
!> \param para_env ...
! **************************************************************************************************
  SUBROUTINE rat_fun_opt(ndf,dg,eigval,work,eigvec_mat,g,para_env)

      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: dg(ndf), eigval(ndf), work(ndf)
      TYPE(cp_fm_type), POINTER                          :: eigvec_mat
      REAL(KIND=dp), INTENT(INOUT)                       :: g(ndf)
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rat_fun_opt', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp

      INTEGER                                            :: handle, i, indf, iref, iter, j, k, l, &
                                                            maxit, ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: bisec, fail, set
      REAL(KIND=dp)                                      :: fun, fun1, fun2, fun3, fung, lam1, lam2, &
                                                            ln, lp, ssize, step, stol
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data

    CALL timeset(routineN,handle)

    stol  = 1.0E-8_dp
    ssize = 0.2_dp
    maxit = 999
    fail  = .FALSE.
    bisec = .FALSE.

    dg=0._dp

    CALL cp_fm_get_info(eigvec_mat,row_indices=row_indices,col_indices=col_indices, &
         local_data=local_data,nrow_local=nrow_local,ncol_local=ncol_local)

    DO i=1,nrow_local
       j=row_indices(i)
       DO k=1,ncol_local
          l=col_indices(k)
          dg(l)=dg(l)+local_data(i,k)*g(j)
       END DO
    END DO
    CALL mp_sum(dg,para_env%group)

    set = .FALSE.

    DO

!   calculating Lamda

      lp = 0.0_dp
      iref = 1
      ln = 0.0_dp
      IF(eigval(iref) < 0.0_dp )ln = eigval(iref) - 0.01_dp

      iter = 0
      DO
         iter = iter + 1
         fun = 0.0_dp
         fung = 0.0_dp
         DO indf = 1, ndf
            fun = fun + dg(indf)**2/(ln-eigval(indf))
            fung = fung - dg(indf)**2 /(ln-eigval(indf)**2)
         END DO
         fun = fun - ln
         fung = fung - one
         step = fun/fung
         ln = ln - step
         IF(ABS(step) < stol )GOTO 200
         IF(iter >= maxit)EXIT
      END DO
100   CONTINUE
      bisec = .TRUE.
      iter = 0
      maxit = 9999
      lam1 = 0.0_dp
      IF(eigval(iref) < 0.0_dp )lam1 = eigval(iref) - 0.01_dp
      fun1 = 0.0_dp
      DO indf = 1, ndf
        fun1 = fun1 + dg(indf)**2 / (lam1-eigval(indf))
      END DO
      fun1 = fun1 - lam1
      step = ABS(lam1)/1000.0_dp
      IF(step < ssize) step = ssize
      DO
         iter = iter + 1
         IF(iter > maxit)THEN
           ln = 0.0_dp
           lp = 0.0_dp
           fail = .TRUE.
           GOTO 300
         END IF
         fun2 = 0.0_dp
         lam2 = lam1 - iter * step
         DO indf = 1, ndf
           fun2 = fun2 + eigval(indf)**2 / (lam2 - eigval(indf))
         END DO
         fun2 = fun2 - lam2
         IF(fun2*fun1 < 0.0_dp )THEN
           iter = 0
           DO
             iter = iter + 1
             IF(iter > maxit)THEN
               ln = 0.0_dp
               lp = 0.0_dp
               fail = .TRUE.
               GOTO 300
             END IF
             step = (lam1+lam2)/2
             fun3 = 0.0_dp
             DO indf = 1, ndf
              fun3 = fun3 + dg(indf)**2/(step-eigval(indf))
             END DO
             fun3 = fun3 - step

             IF( ABS(step-lam2) < stol)THEN
               ln = step
               GOTO 200
             END IF

             IF( fun3 * fun1 < stol )THEN
                lam2 = step
             ELSE
                lam1 = step
             END IF
           END DO
         END IF
     END DO

200  CONTINUE
     IF( (ln > eigval(iref)).OR. ( (ln > 0.0_dp) .AND. &
              (eigval(iref) > 0.0_dp) ) )THEN

         IF(.NOT. bisec)GOTO 100
         ln = 0.0_dp
         lp = 0.0_dp
         fail = .TRUE.
     END IF

300  CONTINUE

     IF(fail .AND. .NOT. set )THEN
        set = .TRUE.
        DO indf = 1, ndf
           eigval(indf) = eigval(indf) * work(indf)
        END DO
        CYCLE
     END IF

     IF(.NOT. set) THEN
        work(1:ndf)=one
     ENDIF

     DO indf = 1, ndf
       eigval(indf) = eigval(indf) - ln
     END DO
     EXIT
     END DO

     CALL timestop(handle)

  END SUBROUTINE  rat_fun_opt

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param dx ...
!> \param dg ...
!> \param hess_mat ...
!> \param work ...
!> \param para_env ...
! **************************************************************************************************
  SUBROUTINE bfgs(ndf,dx,dg,hess_mat,work,para_env)
      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: dx(ndf), dg(ndf)
      TYPE(cp_fm_type), POINTER                          :: hess_mat
      REAL(KIND=dp), INTENT(INOUT)                       :: work(ndf)
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bfgs', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, j, k, l, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: DDOT, dxw, gdx
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_hes

    CALL timeset(routineN,handle)

    CALL cp_fm_get_info(hess_mat,row_indices=row_indices,col_indices=col_indices, &
         local_data=local_hes,nrow_local=nrow_local,ncol_local=ncol_local)

    work=zero
    DO i=1,nrow_local
       j=row_indices(i)
       DO k=1,ncol_local
          l=col_indices(k)
          work(j)=work(j)+local_hes(i,k)*dx(l)
       END DO
    END DO

    CALL mp_sum(work,para_env%group)

    gdx = DDOT(ndf,dg,1,dx,1)
    gdx = one/gdx
    dxw = DDOT(ndf,dx,1,work,1)
    dxw = one/dxw

    DO i=1,nrow_local
       j=row_indices(i)
       DO k=1,ncol_local
          l=col_indices(k)
          local_hes(i,k)=local_hes(i,k)+gdx*dg(j)*dg(l) -&
                                        dxw*work(j)*work(l)
       END DO
    END DO

    CALL timestop(handle)

  END SUBROUTINE bfgs

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param eigval ...
!> \param work ...
! **************************************************************************************************
  SUBROUTINE set_hes_eig(ndf,eigval,work)
      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: eigval(ndf), work(ndf)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_hes_eig', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: max_neg = -0.5_dp, max_pos = 5.0_dp, &
                                                            min_eig = 0.005_dp, one = 1.0_dp

      INTEGER                                            :: handle, indf
      LOGICAL                                            :: neg

    CALL timeset(routineN,handle)

     DO indf = 1, ndf
       IF(eigval(indf) < 0.0_dp )neg = .TRUE.
       IF(eigval(indf) > 1000.0_dp)eigval(indf) = 1000.0_dp
    END DO
    DO indf = 1, ndf
       IF( eigval(indf) < 0.0_dp ) THEN
          IF(eigval(indf) < max_neg)THEN
             eigval(indf) = max_neg
          ELSE IF(eigval(indf) > - min_eig )THEN
             eigval(indf) = - min_eig
          END IF
       ELSE IF(eigval(indf) < 1000.0_dp)THEN
          IF(eigval(indf) < min_eig)THEN
             eigval(indf) = min_eig
          ELSE IF(eigval(indf) > max_pos)THEN
             eigval(indf) = max_pos
          END IF
       END IF
     END DO

     DO indf = 1, ndf
       IF( eigval(indf) < 0.0_dp )THEN
         work(indf) = - one
       ELSE
         work(indf) =  one
       END IF
     END DO

     CALL timestop(handle)

  END SUBROUTINE set_hes_eig

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param eigval ...
!> \param eigvec_mat ...
!> \param hess_tmp ...
!> \param dr ...
!> \param g ...
!> \param para_env ...
!> \param use_rfo ...
! **************************************************************************************************
  SUBROUTINE geoopt_get_step(ndf,eigval,eigvec_mat,hess_tmp,dr,g,para_env,use_rfo)

      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: eigval(ndf)
      TYPE(cp_fm_type), POINTER                          :: eigvec_mat, hess_tmp
      REAL(KIND=dp), INTENT(INOUT)                       :: dr(ndf), g(ndf)
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      LOGICAL                                            :: use_rfo

      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: i, indf, j, k, l, ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      TYPE(cp_fm_type), POINTER                          :: tmp

    CALL cp_fm_to_fm(eigvec_mat,hess_tmp)
    IF(use_rfo)THEN
       DO indf=1,ndf
         eigval(indf) = one/eigval(indf)
       END DO
    ELSE
       DO indf=1,ndf
          eigval(indf) = one/MAX(0.0001_dp,eigval(indf))
       END DO
    END IF

    CALL cp_fm_column_scale(hess_tmp,eigval)
    CALL cp_fm_get_info(eigvec_mat,matrix_struct=matrix_struct)
    CALL cp_fm_create(tmp, matrix_struct ,name="tmp")

    CALL cp_gemm("N","T",ndf,ndf,ndf,one,hess_tmp,eigvec_mat,zero,tmp)

    CALL cp_fm_transpose(tmp,hess_tmp)
    CALL cp_fm_release(tmp)

!    ** New step **

    CALL cp_fm_get_info(hess_tmp,row_indices=row_indices,col_indices=col_indices, &
         local_data=local_data,nrow_local=nrow_local,ncol_local=ncol_local)

    dr=0.0_dp
    DO i=1,nrow_local
       j=row_indices(i)
       DO k=1,ncol_local
          l=col_indices(k)
          dr(j)=dr(j)-local_data(i,k)*g(l)
       END DO
    END DO

    CALL mp_sum(dr,para_env%group)

  END SUBROUTINE geoopt_get_step

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param step ...
!> \param rad ...
!> \param rat ...
!> \param dr ...
!> \param output_unit ...
! **************************************************************************************************
  SUBROUTINE trust_radius(ndf,step,rad,rat,dr,output_unit)
      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: step, rad, rat, dr(ndf)
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'trust_radius', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: scal

    CALL timeset(routineN,handle)

    step = MAXVAL(ABS(dr))
    scal = MAX(one,rad/step)

    IF( step > rad )THEN
       rat = rad / step
       CALL DSCAL(ndf,rat,dr,1)
       step = rad
       IF(output_unit>0) THEN
         WRITE(unit=output_unit,FMT="(/,T2,A,F8.5)") &
             " Step is scaled; Scaling factor = ", rat
         CALL m_flush(output_unit)
       ENDIF
    END IF
    CALL timestop(handle)

  END SUBROUTINE trust_radius

! **************************************************************************************************
!> \brief ...
!> \param ndf ...
!> \param work ...
!> \param hess_mat ...
!> \param dr ...
!> \param g ...
!> \param conv ...
!> \param pred ...
!> \param para_env ...
! **************************************************************************************************
  SUBROUTINE energy_predict(ndf,work,hess_mat,dr,g,conv,pred,para_env)

      INTEGER, INTENT(IN)                                :: ndf
      REAL(KIND=dp), INTENT(INOUT)                       :: work(ndf)
      TYPE(cp_fm_type), POINTER                          :: hess_mat
      REAL(KIND=dp), INTENT(INOUT)                       :: dr(ndf), g(ndf)
      LOGICAL, INTENT(INOUT)                             :: conv
      REAL(KIND=dp), INTENT(INOUT)                       :: pred
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'energy_predict', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: handle, i, j, k, l, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: DDOT, ener1, ener2
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data

    CALL timeset(routineN,handle)

    ener1 = DDOT(ndf,g,1,dr,1)

    CALL cp_fm_get_info(hess_mat,row_indices=row_indices,col_indices=col_indices, &
         local_data=local_data,nrow_local=nrow_local,ncol_local=ncol_local)

    work=zero
    DO i=1,nrow_local
       j=row_indices(i)
       DO k=1,ncol_local
          l=col_indices(k)
          work(j)=work(j)+local_data(i,k)*dr(l)
       END DO
    END DO

    CALL mp_sum(work,para_env%group)
    ener2 = DDOT(ndf,dr,1,work,1)
    pred  = ener1 + 0.5_dp * ener2
    conv = .FALSE.
    CALL timestop(handle)

  END SUBROUTINE energy_predict

! **************************************************************************************************
!> \brief ...
!> \param rat ...
!> \param rad ...
!> \param step ...
!> \param ediff ...
! **************************************************************************************************
   SUBROUTINE update_trust_rad(rat,rad,step,ediff)

      REAL(KIND=dp), INTENT(INOUT)                       :: rat, rad, step, ediff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_trust_rad', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: max_trust = 1.0_dp, min_trust = 0.1_dp

      INTEGER                                            :: handle

    CALL timeset(routineN,handle)

    IF( rat > 4.0_dp )THEN
      IF( ediff < 0.0_dp)THEN
        rad = step * 0.5_dp
      ELSE
        rad = step * 0.25_dp
      END IF
    ELSE IF ( rat > 2.0_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step * 0.75_dp
      ELSE
        rad = step * 0.5_dp
      END IF
    ELSE IF ( rat > 4.0_dp/3.0_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step
      ELSE
        rad = step * 0.75_dp
      END IF
    ELSE IF (rat > 10.0_dp/9.0_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step * 1.25_dp
      ELSE
        rad = step
      END IF
    ELSE IF( rat > 0.9_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step * 1.5_dp
      ELSE
        rad = step * 1.25_dp
      END IF
    ELSE IF( rat > 0.75_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step * 1.25_dp
      ELSE
        rad = step
      END IF
    ELSE IF( rat > 0.5_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step
      ELSE
        rad = step * 0.75_dp
      END IF
    ELSE IF( rat > 0.25_dp )THEN
      IF( ediff < 0.0_dp )THEN
        rad = step * 0.75_dp
      ELSE
        rad = step * 0.5_dp
      END IF
    ELSE IF( ediff < 0.0_dp )THEN
        rad = step * 0.5_dp
    ELSE
        rad = step * 0.25_dp
    END IF

    rad = MAX(rad,min_trust)
    rad = MIN(rad,max_trust)
    CALL timestop(handle)

  END SUBROUTINE update_trust_rad

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param geo_section ...
!> \param hess_mat ...
!> \param logger ...
! **************************************************************************************************
   SUBROUTINE write_bfgs_hessian(geo_section,hess_mat,logger)

      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(cp_fm_type), POINTER                          :: hess_mat
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_bfgs_hessian', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, hesunit

    CALL timeset(routineN,handle)

    hesunit = cp_print_key_unit_nr(logger,geo_section,"BFGS%RESTART",&
           extension=".Hessian",file_form="UNFORMATTED",file_action="WRITE",&
           file_position="REWIND")

    CALL cp_fm_write_unformatted(hess_mat,hesunit)

    CALL cp_print_key_finished_output(hesunit,logger,geo_section,"BFGS%RESTART")

    CALL timestop(handle)

  END SUBROUTINE write_bfgs_hessian
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \param hess_mat ...
! **************************************************************************************************
  SUBROUTINE construct_initial_hess(force_env,hess_mat)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(cp_fm_type), POINTER                          :: hess_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'construct_initial_hess', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat_col, iat_row, iglobal, iind, j, &
                                                            jat_row, jglobal, jind, k, natom, &
                                                            ncol_local, nrow_local, z
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: at_row
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: d_ij, rho_ij
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: r_ij
      REAL(KIND=dp), DIMENSION(3, 3)                     :: alpha, r0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fixed, local_data
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles

     CALL force_env_get(force_env=force_env,subsys=subsys,cell=cell)
     CALL cp_subsys_get(subsys,&
                 particles=particles)

     alpha(1,:)=(/1._dp,0.3949_dp,0.3949_dp/)
     alpha(2,:)=(/0.3494_dp,0.2800_dp,0.2800_dp/)
     alpha(3,:)=(/0.3494_dp,0.2800_dp,0.1800_dp/)

     r0(1,:)=(/1.35_dp,2.10_dp,2.53_dp/)
     r0(2,:)=(/2.10_dp,2.87_dp,3.40_dp/)
     r0(3,:)=(/2.53_dp,3.40_dp,3.40_dp/)

     CALL cp_fm_get_info(hess_mat,row_indices=row_indices,col_indices=col_indices, &
            local_data=local_data,nrow_local=nrow_local,ncol_local=ncol_local)
     natom=particles%n_els
     ALLOCATE(at_row(natom))
     ALLOCATE(rho_ij(natom,natom))
     ALLOCATE(d_ij(natom,natom))
     ALLOCATE(r_ij(natom,natom,3))
     ALLOCATE(fixed(3,natom))
     fixed=1.0_dp
     CALL fix_atom_control( force_env,fixed)
     DO i=1,3
        CALL mp_min(fixed(i,:),hess_mat%matrix_struct%para_env%group)
     END DO
     rho_ij=0
      !XXXX insert proper rows !XXX
     at_row=3
     DO i=1,natom
         CALL get_atomic_kind(atomic_kind=particles%els(i)%atomic_kind,z=z)
         IF(z.LE.10)at_row(i)=2
         IF(z.LE.2)at_row(i)=1
     END DO
    DO i=2,natom
        iat_row=at_row(i)
        DO j=1,i-1
           jat_row=at_row(j)
           !pbc for a distance vector
           r_ij(j,i,:)=pbc(particles%els(i)%r,particles%els(j)%r,cell)
           r_ij(i,j,:)=-r_ij(j,i,:)
           d_ij(j,i)=SQRT(DOT_PRODUCT(r_ij(j,i,:),r_ij(j,i,:)))
           d_ij(i,j)=d_ij(j,i)
           rho_ij(j,i)=EXP(alpha(jat_row,iat_row)*(r0(jat_row,iat_row)**2-d_ij(j,i)**2))
           rho_ij(i,j)=rho_ij(j,i)
        END DO
     END DO
     DO i=1,ncol_local
        iglobal=col_indices(i)
        iind=MOD(iglobal-1,3)+1
        iat_col=(iglobal+2)/3
        IF(iat_col.gt.natom)CYCLE
        DO j=1,nrow_local
           jglobal=row_indices(j)
           jind=MOD(jglobal-1,3)+1
           iat_row=(jglobal+2)/3 
           IF(iat_row.gt.natom)CYCLE
           IF(iat_row.ne.iat_col)THEN
               IF(d_ij(iat_row,iat_col).lt.6.0_dp)&
               local_data(j,i)=local_data(j,i)+&
                     angle_second_deriv(r_ij,d_ij,rho_ij,iind,jind,iat_col,iat_row,natom)
           ELSE
               local_data(j,i)=local_data(j,i)+&
                     angle_second_deriv(r_ij,d_ij,rho_ij,iind,jind,iat_col,iat_row,natom)
           END IF
           IF(iat_col.NE.iat_row)THEN
               IF(d_ij(iat_row,iat_col).lt.6.0_dp) &
               local_data(j,i)=local_data(j,i)-&
                               dist_second_deriv(r_ij(iat_col,iat_row,:),&
                               iind,jind,d_ij(iat_row,iat_col),rho_ij(iat_row,iat_col))
           ELSE
              DO k=1,natom
                 IF(k==iat_col)CYCLE
                 IF(d_ij(iat_row,k).lt.6.0_dp) &
                 local_data(j,i)=local_data(j,i)+&
                               dist_second_deriv(r_ij(iat_col,k,:),&
                               iind,jind,d_ij(iat_row,k),rho_ij(iat_row,k))
              END DO
           END IF
           IF(fixed(jind,iat_row).lt.0.5_dp.OR.fixed(iind,iat_col).lt.0.5_dp)THEN
              local_data(j,i)=0.0_dp
              IF(jind==iind.AND.iat_row==iat_col)local_data(j,i)=1.0_dp
           END IF 
        END DO
     END DO
     DEALLOCATE(fixed)
     DEALLOCATE(rho_ij)
     DEALLOCATE(d_ij)
     DEALLOCATE(r_ij)
     DEALLOCATE(at_row)

  END SUBROUTINE  construct_initial_hess

! **************************************************************************************************
!> \brief ...
!> \param r1 ...
!> \param i ...
!> \param j ...
!> \param d ...
!> \param rho ...
!> \retval deriv ...
! **************************************************************************************************
  FUNCTION dist_second_deriv(r1,i,j,d,rho)  RESULT(deriv)
      REAL(KIND=dp), DIMENSION(3)                        :: r1
      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: d, rho, deriv

      deriv=0.45_dp*rho*(r1(i)*r1(j))/d**2
  END FUNCTION

! **************************************************************************************************
!> \brief ...
!> \param r_ij ...
!> \param d_ij ...
!> \param rho_ij ...
!> \param idir ...
!> \param jdir ...
!> \param iat_der ...
!> \param jat_der ...
!> \param natom ...
!> \retval deriv ...
! **************************************************************************************************
  FUNCTION angle_second_deriv(r_ij,d_ij,rho_ij,idir,jdir,iat_der,jat_der,natom) RESULT(deriv)
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: r_ij
      REAL(KIND=dp), DIMENSION(:, :)                     :: d_ij, rho_ij
      INTEGER                                            :: idir, jdir, iat_der, jat_der, natom
      REAL(KIND=dp)                                      :: deriv

      INTEGER                                            :: i, iat, idr, j, jat, jdr
      REAL(KIND=dp)                                      :: d12, d23, d31, D_mat(3,2), denom1, &
                                                            denom2, denom3, ka1, ka2, ka3, rho12, &
                                                            rho23, rho31, rsst1, rsst2, rsst3
      REAL(KIND=dp), DIMENSION(3)                        :: r12, r23, r31

     deriv=0._dp
     IF(iat_der==jat_der)THEN
        DO i=1,natom-1
           IF(rho_ij(iat_der,i).lt.0.00001)CYCLE
           DO j=i+1,natom
              IF(rho_ij(iat_der,j).lt.0.00001)CYCLE
              IF(i==iat_der.OR.j==iat_der)CYCLE
              IF(iat_der.lt.i.OR.iat_der.gt.j)THEN
                 r12=r_ij(iat_der,i,:); r23=r_ij(i,j,:); r31=r_ij(j,iat_der,:)
                 d12=d_ij(iat_der,i); d23=d_ij(i,j); d31=d_ij(j,iat_der)
                 rho12=rho_ij(iat_der,i); rho23=rho_ij(i,j); rho31=rho_ij(j,iat_der)
              ELSE
                 r12=r_ij(iat_der,j,:); r23=r_ij(j,i,:); r31=r_ij(i,iat_der,:)
                 d12=d_ij(iat_der,j); d23=d_ij(j,i); d31=d_ij(i,iat_der)
                 rho12=rho_ij(iat_der,j); rho23=rho_ij(j,i); rho31=rho_ij(i,iat_der)
              END IF
              ka1=0.15_dp*rho12*rho23; ka2=0.15_dp*rho23*rho31; ka3=0.15_dp*rho31*rho12
              rsst1=DOT_PRODUCT(r12,r23);rsst2=DOT_PRODUCT(r23,r31);rsst3=DOT_PRODUCT(r31,r12)
              denom1=1.0_dp-rsst1**2/(d12**2*d23**2); denom2=1.0_dp-rsst2**2/(d23**2*d31**2)
              denom3=1.0_dp-rsst3**2/(d31**2*d12**2)
              denom1=SIGN(1.0_dp,denom1)*MAX(ABS(denom1),0.01_dp)
              denom2=SIGN(1.0_dp,denom2)*MAX(ABS(denom2),0.01_dp)
              denom3=SIGN(1.0_dp,denom3)*MAX(ABS(denom3),0.01_dp)
              D_mat(1,1)=r23(idir)/(d12*d23)-rsst1*r12(idir)/(d12**3*d23)
              D_mat(1,2)=r23(jdir)/(d12*d23)-rsst1*r12(jdir)/(d12**3*d23)
              D_mat(2,1)=-r23(idir)/(d23*d31)+rsst2*r31(idir)/(d23*d31**3)
              D_mat(2,2)=-r23(jdir)/(d23*d31)+rsst2*r31(jdir)/(d23*d31**3)
              D_mat(3,1)=(r31(idir)-r12(idir))/(d31*d12)+rsst3*r31(idir)/(d31**3*d12)-&
                                                         rsst3*r12(idir)/(d31*d12**3)
              D_mat(3,2)=(r31(jdir)-r12(jdir))/(d31*d12)+rsst3*r31(jdir)/(d31**3*d12)-&
                                                         rsst3*r12(jdir)/(d31*d12**3)
              IF(ABS(denom1).le.0.011_dp)D_mat(1,1)=0.0_dp
              IF(ABS(denom2).le.0.011_dp)D_mat(2,1)=0.0_dp
              IF(ABS(denom3).le.0.011_dp)D_mat(3,1)=0.0_dp
              deriv=deriv+ka1*D_mat(1,1)*D_mat(1,2)/denom1+&
                    ka2*D_mat(2,1)*D_mat(2,2)/denom2+&
                    ka3*D_mat(3,1)*D_mat(3,2)/denom3

           END DO
        END DO
     ELSE
        DO i=1,natom
           IF(i==iat_der.OR.i==jat_der)CYCLE
           IF(jat_der.lt.iat_der)THEN
              iat=jat_der; jat=iat_der; idr=jdir; jdr=idir
           ELSE
              iat=iat_der; jat=jat_der; idr=idir; jdr=jdir
           END IF
           IF(jat.lt.i.OR.iat.gt.i)THEN
              r12=r_ij(iat,jat,:); r23=r_ij(jat,i,:); r31=r_ij(i,iat,:)
              d12=d_ij(iat,jat); d23=d_ij(jat,i); d31=d_ij(i,iat)
              rho12=rho_ij(iat,jat); rho23=rho_ij(jat,i); rho31=rho_ij(i,iat)
           ELSE
              r12=r_ij(iat,i,:); r23=r_ij(i,jat,:); r31=r_ij(jat,iat,:)
              d12=d_ij(iat,i); d23=d_ij(i,jat); d31=d_ij(jat,iat)
              rho12=rho_ij(iat,i); rho23=rho_ij(i,jat); rho31=rho_ij(jat,iat)
           END IF
           ka1=0.15_dp*rho12*rho23; ka2=0.15_dp*rho23*rho31; ka3=0.15_dp*rho31*rho12
           rsst1=DOT_PRODUCT(r12,r23);rsst2=DOT_PRODUCT(r23,r31);rsst3=DOT_PRODUCT(r31,r12)
           denom1=1.0_dp-rsst1**2/(d12**2*d23**2); denom2=1.0_dp-rsst2**2/(d23**2*d31**2)
           denom3=1.0_dp-rsst3**2/(d31**2*d12**2)
           denom1=SIGN(1.0_dp,denom1)*MAX(ABS(denom1),0.01_dp)
           denom2=SIGN(1.0_dp,denom2)*MAX(ABS(denom2),0.01_dp)
           denom3=SIGN(1.0_dp,denom3)*MAX(ABS(denom3),0.01_dp)
           D_mat(1,1)=r23(idr)/(d12*d23)-rsst1*r12(idr)/(d12**3*d23)
           D_mat(2,1)=-r23(idr)/(d23*d31)+rsst2*r31(idr)/(d23*d31**3)
           D_mat(3,1)=(r31(idr)-r12(idr))/(d31*d12)+rsst3*r31(idr)/(d31**3*d12)-&
                                                         rsst3*r12(idr)/(d31*d12**3)
           IF(jat.lt.i.OR.iat.gt.i)THEN
              D_mat(1,2)=(r12(jdr)-r23(jdr))/(d12*d23)+rsst1*r12(jdr)/(d12**3*d23)-&
                                                       rsst1*r23(jdr)/(d12*d23**3)
              D_mat(2,2)=r31(jdr)/(d23*d31)-rsst2*r23(jdr)/(d23**3*d31)
              D_mat(3,2)=-r31(jdr)/(d31*d12)+rsst3*r12(jdr)/(d31*d12**3)
           ELSE
              D_mat(1,2)=-r12(jdr)/(d12*d23)+rsst1*r23(jdr)/(d12*d23**3)
              D_mat(2,2)=(r23(jdr)-r31(jdr))/(d23*d31)+rsst2*r23(jdr)/(d23**3*d31)-&
                                                       rsst2*r31(jdr)/(d23*d31**3)
              D_mat(3,2)=r12(jdr)/(d31*d12)-rsst3*r31(jdr)/(d31**3*d12)
           END IF
           IF(ABS(denom1).le.0.011_dp)D_mat(1,1)=0.0_dp
           IF(ABS(denom2).le.0.011_dp)D_mat(2,1)=0.0_dp
           IF(ABS(denom3).le.0.011_dp)D_mat(3,1)=0.0_dp

           deriv=deriv+ka1*D_mat(1,1)*D_mat(1,2)/denom1+&
                 ka2*D_mat(2,1)*D_mat(2,2)/denom2+&
                 ka3*D_mat(3,1)*D_mat(3,2)/denom3
         END DO
      END IF
      deriv=0.25_dp*deriv

  END FUNCTION angle_second_deriv

END MODULE bfgs_optimizer
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief LBFGS-B routine (version 3.0, April 25, 2011)
!> \note
!>      L-BFGS-B (version 3.0, April 25, 2011) converted to Fortran 90 module
!> \par History
!>      02.2005 Update to the new version 2.4 and deleting the blas part of
!>              the code (Teodoro Laino)
!>      11.2012 New version 3.0 converted to Fortran 90 (Matthias Krack)
!> \author Fawzi Mohamed (first version)
! **************************************************************************************************
MODULE cp_lbfgs
   USE bibliography,                    ONLY: Byrd1995,&
                                              cite_reference
   USE cp_files,                        ONLY: open_file
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_lbfgs'

  PUBLIC :: setulb

CONTAINS

!===========   L-BFGS-B (version 3.0.  April 25, 2011  =================
!
!     This is a modified version of L-BFGS-B.
!
!     Major changes are described in the accompanying paper:
!
!         Jorge Nocedal and Jose Luis Morales, Remark on "Algorithm 778:
!         L-BFGS-B: Fortran Subroutines for Large-Scale Bound Constraine
!         Optimization"  (2011). To appear in  ACM Transactions on
!         Mathematical Software,
!
!     The paper describes an improvement and a correction to Algorithm 7
!     It is shown that the performance of the algorithm can be improved
!     significantly by making a relatively simple modication to the subs
!     minimization phase. The correction concerns an error caused by the
!     of routine dpmeps to estimate machine precision.
!
!     The total work space **wa** required by the new version is
!
!                  2*m*n + 11m*m + 5*n + 8*m
!
!     the old version required
!
!                  2*m*n + 12m*m + 4*n + 12*m
!
!
!            J. Nocedal  Department of Electrical Engineering and
!                        Computer Science.
!                        Northwestern University. Evanston, IL. USA
!
!
!           J.L Morales  Departamento de Matematicas,
!                        Instituto Tecnologico Autonomo de Mexico
!                        Mexico D.F. Mexico.
!
!                        March  2011
!
!=======================================================================
! **************************************************************************************************
!> \brief          This subroutine partitions the working arrays wa and iwa, and
!>                 then uses the limited memory BFGS method to solve the bound
!>                 constrained optimization problem by calling mainlb.
!>                 (The direct method will be used in the subspace minimization.)
!> \param n        n is the dimension of the problem.
!> \param m        m is the maximum number of variable metric corrections
!>                 used to define the limited memory matrix.
!> \param x        On entry x is an approximation to the solution.
!>                 On exit x is the current approximation.
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd      nbd represents the type of bounds imposed on the
!>                 variables, and must be specified as follows:
!>                 nbd(i)=0 if x(i) is unbounded,
!>                        1 if x(i) has only a lower bound,
!>                        2 if x(i) has both lower and upper bounds, and
!>                        3 if x(i) has only an upper bound.
!> \param f        On first entry f is unspecified.
!>                 On final exit f is the value of the function at x.
!> \param g        On first entry g is unspecified.
!>                 On final exit g is the value of the gradient at x.
!> \param factr    factr >= 0 is specified by the user.  The iteration
!>                 will stop when
!>
!>                 (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch
!>
!>                 where epsmch is the machine precision, which is automatically
!>                 generated by the code. Typical values for factr: 1.d+12 for
!>                 low accuracy; 1.d+7 for moderate accuracy; 1.d+1 for extremely
!>                 high accuracy.
!> \param pgtol    pgtol >= 0 is specified by the user.  The iteration
!>                 will stop when
!>
!>                 max{|proj g_i | i = 1, ..., n} <= pgtol
!>
!>                 where pg_i is the ith component of the projected gradient.
!> \param wa       working array
!> \param iwa      integer working array
!> \param task     is a working string of characters of length 60 indicating
!>                 the current job when entering and quitting this subroutine.
!> \param iprint   iprint is a variable that must be set by the user.
!>                 It controls the frequency and type of output generated:
!>                 iprint<0    no output is generated;
!>                 iprint=0    print only one line at the last iteration;
!>                 0<iprint<99 print also f and |proj g| every iprint iterations;
!>                 iprint=99   print details of every iteration except n-vectors;
!>                 iprint=100  print also the changes of active set and final x;
!>                 iprint>100  print details of every iteration including x and g;
!>                 When iprint > 0, the file iterate.dat will be created to
!>                 summarize the iteration.
!> \param csave    is a working string of characters
!> \param lsave    lsave is a working array
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 If lsave(1) = .true.  then  the initial X has been replaced by
!>                               its projection in the feasible set
!>                 If lsave(2) = .true.  then  the problem is constrained;
!>                 If lsave(3) = .true.  then  each variable has upper and lower bounds;
!> \param isave    isave is a working array
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 isave(22) = the total number of intervals explored in the
!>                         search of Cauchy points;
!>                 isave(26) = the total number of skipped BFGS updates before the current iteration;
!>                 isave(30) = the number of current iteration;
!>                 isave(31) = the total number of BFGS updates prior the current iteration;
!>                 isave(33) = the number of intervals explored in the search of
!>                             Cauchy point in the current iteration;
!>                 isave(34) = the total number of function and gradient evaluations;
!>                 isave(36) = the number of function value or gradient
!>                             evaluations in the current iteration;
!>                 if isave(37) = 0  then the subspace argmin is within the box;
!>                 if isave(37) = 1  then the subspace argmin is beyond the box;
!>                 isave(38) = the number of free variables in the current iteration;
!>                 isave(39) = the number of active constraints in the current iteration;
!>                 n + 1 - isave(40) = the number of variables leaving the set of
!>                                     active constraints in the current iteration;
!>                 isave(41) = the number of variables entering the set of active
!>                             constraints in the current iteration.
!> \param dsave    dsave is a working array of dimension 29.
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 dsave(1) = current 'theta' in the BFGS matrix;
!>                 dsave(2) = f(x) in the previous iteration;
!>                 dsave(3) = factr*epsmch;
!>                 dsave(4) = 2-norm of the line search direction vector;
!>                 dsave(5) = the machine precision epsmch generated by the code;
!>                 dsave(7) = the accumulated time spent on searching for Cauchy points;
!>                 dsave(8) = the accumulated time spent on subspace minimization;
!>                 dsave(9) = the accumulated time spent on line search;
!>                 dsave(11) = the slope of the line search function at the current point of line search;
!>                 dsave(12) = the maximum relative step length imposed in line search;
!>                 dsave(13) = the infinity norm of the projected gradient;
!>                 dsave(14) = the relative step length in the line search;
!>                 dsave(15) = the slope of the line search function at the starting point of the line search;
!>                 dsave(16) = the square of the 2-norm of the line search direction vector.
!> \param trust_radius ...
!> \author         NEOS, November 1994. (Latest revision June 1996.)
!>                 Optimization Technology Center.
!>                 Argonne National Laboratory and Northwestern University.
!>                 Written by
!>                             Ciyou Zhu
!>                 in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE setulb(n, m, x, lower_bound, upper_bound, nbd, f, g, factr, pgtol, wa, iwa,&
                       task, iprint, csave, lsave, isave, dsave, trust_radius)

      INTEGER, INTENT(in)                                :: n, m
      REAL(KIND=dp), INTENT(inout)                       :: x(n)
      REAL(KIND=dp)                                      :: lower_bound(n), upper_bound(n)
      INTEGER                                            :: nbd(n)
      REAL(KIND=dp)                                      :: f, g(n)
      REAL(KIND=dp), INTENT(in)                          :: factr, pgtol
      REAL(KIND=dp)                                      :: wa(2*m*n + 5*n + 11*m*m + 8*m)
      INTEGER                                            :: iwa(3*n)
      CHARACTER(LEN=60)                                  :: task
      INTEGER                                            :: iprint
      CHARACTER(LEN=60)                                  :: csave
      LOGICAL                                            :: lsave(4)
      INTEGER                                            :: isave(44)
      REAL(KIND=dp)                                      :: dsave(29)
      REAL(KIND=dp), INTENT(in)                          :: trust_radius

      INTEGER                                            :: i, ld, lr, lsnd, lss, lsy, lt, lwa, lwn, &
                                                            lws, lwt, lwy, lxp, lz

!     References:
!
!       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
!       memory algorithm for bound constrained optimization'',
!       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
!
!       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: a
!       limited memory FORTRAN code for solving bound constrained
!       optimization problems'', Tech. Report, NAM-11, EECS Department,
!       Northwestern University, 1994.
!
!       (Postscript files of these papers are available via anonymous
!        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
!
!                           *  *  *

      IF (task == 'START') THEN
         CALL cite_reference(Byrd1995)
         isave(1)  = m*n
         isave(2)  = m**2
         isave(3)  = 4*m**2
                                            ! ws      m*n
         isave(4)  = 1
                                            ! wy      m*n
         isave(5)  = isave(4)  + isave(1)
                                            ! wsy     m**2
         isave(6)  = isave(5)  + isave(1)
                                            ! wss     m**2
         isave(7)  = isave(6)  + isave(2)
                                            ! wt      m**2
         isave(8)  = isave(7)  + isave(2)
                                            ! wn      4*m**2
         isave(9)  = isave(8)  + isave(2)
                                            ! wsnd    4*m**2
         isave(10) = isave(9)  + isave(3)
                                            ! wz      n
         isave(11) = isave(10) + isave(3)
                                            ! wr      n
         isave(12) = isave(11) + n
                                            ! wd      n
         isave(13) = isave(12) + n
                                            ! wt      n
         isave(14) = isave(13) + n
                                            ! wxp     n
         isave(15) = isave(14) + n
                                            ! wa      8*m
         isave(16) = isave(15) + n
      END IF
      lws  = isave(4)
      lwy  = isave(5)
      lsy  = isave(6)
      lss  = isave(7)
      lwt  = isave(8)
      lwn  = isave(9)
      lsnd = isave(10)
      lz   = isave(11)
      lr   = isave(12)
      ld   = isave(13)
      lt   = isave(14)
      lxp  = isave(15)
      lwa  = isave(16)

      !in case we use a trust radius we set the boundaries to be one times the trust radius away from the current positions
      !the original implementation only allowed for boundaries that remain constant during the optimization.
      !This way of including a trust radius seems to work,
      !but the change of the boundaries during optimization might introduce some not yet discovered problems.
      IF(trust_radius>0) THEN
        DO i=1,n
          lower_bound(i)=x(i)-trust_radius
          upper_bound(i)=x(i)+trust_radius
          nbd(i)=2
        END DO
      ENDIF

      CALL mainlb(n,m,x,lower_bound,upper_bound,nbd,f,g,factr,pgtol,                        &
        wa(lws),wa(lwy),wa(lsy),wa(lss), wa(lwt),                       &
        wa(lwn),wa(lsnd),wa(lz),wa(lr),wa(ld),wa(lt),wa(lxp),           &
        wa(lwa),                                                        &
        iwa(1),iwa(n+1),iwa(2*n+1),task,iprint,                         &
        csave,lsave,isave(22),dsave)

      RETURN

      END SUBROUTINE setulb

! **************************************************************************************************
!> \brief        This subroutine solves bound constrained optimization problems by
!>               using the compact formula of the limited memory BFGS updates.
!> \param n      n is the number of variables
!> \param m      m is the maximum number of variable metric
!>               corrections allowed in the limited memory matrix.
!> \param x      On entry x is an approximation to the solution.
!>               On exit x is the current approximation.
!> \param lower_bound  lower_bound is the lower bound of x.
!> \param upper_bound  upper_bound is the upper bound of x.
!> \param nbd    nbd represents the type of bounds imposed on the
!>               variables, and must be specified as follows:
!>               nbd(i)=0 if x(i) is unbounded,
!>               1 if x(i) has only a lower bound,
!>               2 if x(i) has both lower and upper bounds,
!>               3 if x(i) has only an upper bound.
!> \param f      On first entry f is unspecified.
!>               On final exit f is the value of the function at x.
!> \param g      On first entry g is unspecified.
!>               On final exit g is the value of the gradient at x.
!> \param factr  factr >= 0 is specified by the user.  The iteration
!>               will stop when
!>
!>               (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch
!>
!>               where epsmch is the machine precision, which is automatically
!>               generated by the code.
!> \param pgtol  pgtol >= 0 is specified by the user.  The iteration
!>               will stop when
!>
!>                 max{|proj g_i | i = 1, ..., n} <= pgtol
!>
!>               where pg_i is the ith component of the projected gradient.
!> \param ws     ws, wy, sy, and wt are working arrays used to store the following 
!>               information defining the limited memory BFGS matrix:
!>               ws stores S, the matrix of s-vectors;
!> \param wy     stores Y, the matrix of y-vectors;
!> \param sy     stores S'Y;
!> \param ss     stores S'S;
!> \param wt     stores the Cholesky factorization of (theta*S'S+LD^(-1)L');
!>               see eq. (2.26) in [3].
!> \param wn     wn is a working array of dimension 2m x 2m
!>               used to store the LEL^T factorization of the indefinite matrix
!>               K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!>                   [L_a -R_z           theta*S'AA'S ]
!>
!>               where     E = [-I  0]
!>                             [ 0  I]
!> \param snd    is a working array of dimension 2m x 2m
!>               used to store the lower triangular part of
!>               N = [Y' ZZ'Y   L_a'+R_z']
!>                   [L_a +R_z  S'AA'S   ]
!> \param z      z(n),r(n),d(n),t(n), xp(n),wa(8*m) are working arrays
!>               z  is used at different times to store the Cauchy point and
!>               the Newton point.
!> \param r      working array
!> \param d      working array
!> \param t      workign array
!> \param xp     xp is a workng array used to safeguard the projected Newton direction
!> \param wa     working array
!> \param index  In subroutine freev, index is used to store the free and fixed
!>               variables at the Generalized Cauchy Point (GCP).
!> \param iwhere iwhere is an integer working array of dimension n used to record
!>               the status of the vector x for GCP computation.
!>               iwhere(i)=0 or -3 if x(i) is free and has bounds,
!>                         1       if x(i) is fixed at l(i), and l(i) .ne. u(i)
!>                         2       if x(i) is fixed at u(i), and u(i) .ne. l(i)
!>                         3       if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
!>                        -1       if x(i) is always free, i.e., no bounds on it.
!> \param indx2  indx2 is a working array. Within subroutine cauchy, indx2 corresponds to the array iorder.
!>               In subroutine freev, a list of variables entering and leaving
!>               the free set is stored in indx2, and it is passed on to
!>               subroutine formk with this information
!> \param task   task is a working string of characters indicating
!>               the current job when entering and leaving this subroutine.
!> \param iprint is an variable that must be set by the user.
!>               It controls the frequency and type of output generated:
!>               iprint<0    no output is generated;
!>               iprint=0    print only one line at the last iteration;
!>               0<iprint<99 print also f and |proj g| every iprint iterations;
!>               iprint=99   print details of every iteration except n-vectors;
!>               iprint=100  print also the changes of active set and final x;
!>               iprint>100  print details of every iteration including x and g;
!>               When iprint > 0, the file iterate.dat will be created to summarize the iteration.
!> \param csave  csave is a working string of characters
!> \param lsave  lsave is a logical working array
!> \param isave  isave is an integer working array
!> \param dsave  is a double precision working array
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE mainlb(n, m, x, lower_bound, upper_bound, nbd, f, g, factr, pgtol, ws, wy, &
                        sy, ss, wt, wn, snd, z, r, d, t, xp, wa,        &
                        index, iwhere, indx2, task,                     &
                        iprint, csave, lsave, isave, dsave)
      INTEGER, INTENT(in)                                :: n, m
      REAL(KIND=dp), INTENT(inout)                       :: x(n)
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n)
      INTEGER                                            :: nbd(n)
      REAL(KIND=dp) :: f, g(n), factr, pgtol, ws(n, m), wy(n, m), sy(m, m), ss(m, m), wt(m, m), &
         wn(2*m, 2*m), snd(2*m, 2*m), z(n), r(n), d(n), t(n), xp(n), wa(8*m)
      INTEGER                                            :: INDEX(n), iwhere(n), indx2(n)
      CHARACTER(LEN=60)                                  :: task
      INTEGER                                            :: iprint
      CHARACTER(LEN=60)                                  :: csave
      LOGICAL                                            :: lsave(4)
      INTEGER                                            :: isave(23)
      REAL(KIND=dp)                                      :: dsave(29)

      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      CHARACTER(LEN=3)                                   :: word
      INTEGER                                            :: col, head, i, iback, ifun, ileave, info, &
                                                            itail, iter, itfile, iupdat, iword, k, &
                                                            nact, nenter, nfgv, nfree, nintol, &
                                                            nseg, nskip
      LOGICAL                                            :: boxed, constrained, first, updatd, wrk, &
                                                            x_projected
      REAL(KIND=dp) :: cachyt, cpu1, cpu2, ddot, ddum, dnorm, dr, dtd, epsmch, fold, g_inf_norm, &
         gd, gdold, lnscht, rr, sbtime, step_max, stp, theta, time, time1, time2, tol, xstep

!     References:
!
!       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
!       memory algorithm for bound constrained optimization'',
!       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
!
!       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
!       Subroutines for Large Scale Bound Constrained Optimization''
!       Tech. Report, NAM-11, EECS Department, Northwestern University,
!       1994.
!
!       [3] R. Byrd, J. Nocedal and R. Schnabel "Representations of
!       Quasi-Newton Matrices and their use in Limited Memory Methods'',
!       Mathematical Programming 63 (1994), no. 4, pp. 129-156.
!
!       (Postscript files of these papers are available via anonymous
!        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
!
!                           *  *  *

      IF (task == 'START') THEN

         epsmch = EPSILON(one)

         CALL timer(time1)

!        Initialize counters and scalars when task='START'.

!           for the limited memory BFGS matrices:
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .FALSE.
         iback  = 0
         itail  = 0
         iword  = 0
         nact   = 0
         ileave = 0
         nenter = 0
         fold   = zero
         dnorm  = zero
         cpu1   = zero
         gd     = zero
         step_max  = zero
         g_inf_norm = zero
         stp    = zero
         gdold  = zero
         dtd    = zero

!           for operation counts:
         iter   = 0
         nfgv   = 0
         nseg   = 0
         nintol = 0
         nskip  = 0
         nfree  = n
         ifun   = 0
!           for stopping tolerance:
         tol = factr*epsmch

!           for measuring running time:
         cachyt = 0
         sbtime = 0
         lnscht = 0

!           'word' records the status of subspace solutions.
         word = '---'

!           'info' records the termination information.
         info = 0

         itfile = 8
         IF (iprint >= 1) THEN
!                                open a summary file 'iterate.dat'
            CALL open_file(file_name='/mnt/sdcard/cp2k/work/iterate.dat',unit_number=itfile,file_action='WRITE', &
	    file_status='UNKNOWN')
         END IF

!        Check the input arguments for errors.

         CALL errclb(n,m,factr,lower_bound,upper_bound,nbd,task,info,k)
         IF (task(1:5) == 'ERROR') THEN
            CALL prn3lb(n,x,f,task,iprint,info,itfile,                  &
                        iter,nfgv,nintol,nskip,nact,g_inf_norm,             &
                        zero,nseg,word,iback,stp,xstep,k,               &
                        cachyt,sbtime,lnscht)
            RETURN
         END IF

         CALL prn1lb(n,m,lower_bound,upper_bound,x,iprint,itfile,epsmch)

!        Initialize iwhere & project x onto the feasible set.

         CALL active(n,lower_bound,upper_bound,nbd,x,iwhere,iprint,x_projected,constrained,boxed)

!        The end of the initialization.
         task = 'FG_START'
!        return to the driver to calculate f and g; reenter at 111.
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      ELSE
!          restore local variables.

         x_projected = lsave(1)
         constrained = lsave(2)
         boxed  = lsave(3)
         updatd = lsave(4)

         nintol = isave(1)
         itfile = isave(3)
         iback  = isave(4)
         nskip  = isave(5)
         head   = isave(6)
         col    = isave(7)
         itail  = isave(8)
         iter   = isave(9)
         iupdat = isave(10)
         nseg   = isave(12)
         nfgv   = isave(13)
         info   = isave(14)
         ifun   = isave(15)
         iword  = isave(16)
         nfree  = isave(17)
         nact   = isave(18)
         ileave = isave(19)
         nenter = isave(20)

         theta  = dsave(1)
         fold   = dsave(2)
         tol    = dsave(3)
         dnorm  = dsave(4)
         epsmch = dsave(5)
         cpu1   = dsave(6)
         cachyt = dsave(7)
         sbtime = dsave(8)
         lnscht = dsave(9)
         time1  = dsave(10)
         gd     = dsave(11)
         step_max  = dsave(12)
         g_inf_norm = dsave(13)
         stp    = dsave(14)
         gdold  = dsave(15)
         dtd    = dsave(16)

!        After returning from the driver go to the point where execution
!        is to resume.

         IF (task(1:4) == 'STOP') THEN
            IF (task(7:9) == 'CPU') THEN
!                                          restore the previous iterate.
               CALL dcopy(n,t,1,x,1)
               CALL dcopy(n,r,1,g,1)
               f = fold
            END IF
            CALL timer(time2)
            time = time2 - time1
            CALL prn3lb(n,x,f,task,iprint,info,itfile,                        &
                        iter,nfgv,nintol,nskip,nact,g_inf_norm,                   &
                        time,nseg,word,iback,stp,xstep,k,                     &
                        cachyt,sbtime,lnscht)
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
            RETURN
         END IF
      END IF

      IF(.NOT.(task(1:5) == 'FG_LN'.OR.task(1:5) == 'NEW_X')) THEN

!     Compute f0 and g0.
      nfgv = 1

!     Compute the infinity norm of the (-) projected gradient.

      CALL projgr(n,lower_bound,upper_bound,nbd,x,g,g_inf_norm)

      IF (iprint >= 1) THEN
         WRITE (*,1002) iter,f,g_inf_norm
         WRITE (itfile,1003) iter,nfgv,g_inf_norm,f
      END IF
      IF (g_inf_norm <= pgtol) THEN
!                                terminate the algorithm.
         task = 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
         CALL timer(time2)
         time = time2 - time1
         CALL prn3lb(n,x,f,task,iprint,info,itfile,                        &
                     iter,nfgv,nintol,nskip,nact,g_inf_norm,                   &
                     time,nseg,word,iback,stp,xstep,k,                     &
                     cachyt,sbtime,lnscht)
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      END IF
      END IF

      first=.TRUE.
      DO WHILE(.TRUE.)
      IF(.NOT.first.OR..NOT.(task(1:5) == 'FG_LN'.OR.task(1:5) == 'NEW_X')) THEN
      IF (iprint >= 99) WRITE (*,1001) iter + 1
      iword = -1
!
      IF (.NOT. constrained .AND. col > 0) THEN
!                                            skip the search for GCP.
         CALL dcopy(n,x,1,z,1)
         wrk = updatd
         nseg = 0
      ELSE

!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     Compute the Generalized Cauchy Point (GCP).
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

         CALL timer(cpu1)
         CALL cauchy(n,x,lower_bound,upper_bound,nbd,g,indx2,iwhere,t,d,z,                     &
                     m,wy,ws,sy,wt,theta,col,head,                         &
                     wa(1),wa(2*m+1),wa(4*m+1),wa(6*m+1),nseg,             &
                     iprint, g_inf_norm, info, epsmch)
         IF (info /= 0) THEN
!            singular triangular system detected; refresh the lbfgs memory.
            IF(iprint >= 1) WRITE (*, 1005)
            info   = 0
            col    = 0
            head   = 1
            theta  = one
            iupdat = 0
            updatd = .FALSE.
            CALL timer(cpu2)
            cachyt = cachyt + cpu2 - cpu1
            first=.FALSE.
            CYCLE
         END IF
         CALL timer(cpu2)
         cachyt = cachyt + cpu2 - cpu1
         nintol = nintol + nseg

!        Count the entering and leaving variables for iter > 0;
!        find the index set of free and active variables at the GCP.

         CALL freev(n,nfree,index,nenter,ileave,indx2,                     &
                    iwhere,wrk,updatd,constrained,iprint,iter)
         nact = n - nfree

      ENDIF

!     If there are no free variables or B=theta*I, then
!                                        skip the subspace minimization.

      IF (.NOT.(nfree == 0 .OR. col == 0)) THEN

!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     Subspace minimization.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      CALL timer(cpu1)

!     Form  the LEL^T factorization of the indefinite
!       matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!                     [L_a -R_z           theta*S'AA'S ]
!       where     E = [-I  0]
!                     [ 0  I]

      IF (wrk) CALL formk(n,nfree,index,nenter,ileave,indx2,iupdat,     &
                       updatd,wn,snd,m,ws,wy,sy,theta,col,head,info)
      IF (info /= 0) THEN
!          nonpositive definiteness in Cholesky factorization;
!          refresh the lbfgs memory and restart the iteration.
         IF(iprint >= 1) WRITE (*, 1006)
         info   = 0
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .FALSE.
         CALL timer(cpu2)
         sbtime = sbtime + cpu2 - cpu1
         first=.FALSE.
         CYCLE
      END IF

!        compute r=-Z'B(xcp-xk)-Z'g (using wa(2m+1)=W'(xcp-x)
!                                                   from 'cauchy').
      CALL cmprlb(n,m,x,g,ws,wy,sy,wt,z,r,wa,index,                     &
                 theta,col,head,nfree,constrained,info)
      IF (info == 0) THEN

!     call the direct method.

         CALL subsm( n, m, nfree, index, lower_bound, upper_bound, nbd, z, r, xp, ws, wy,      &
                    theta, x, g, col, head, iword, wa, wn, iprint, info)
      END IF
      IF (info /= 0) THEN
!          singular triangular system detected;
!          refresh the lbfgs memory and restart the iteration.
         IF(iprint >= 1) WRITE (*, 1005)
         info   = 0
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .FALSE.
         CALL timer(cpu2)
         sbtime = sbtime + cpu2 - cpu1
         first=.FALSE.
         CYCLE
      END IF

      CALL timer(cpu2)
      sbtime = sbtime + cpu2 - cpu1
      END IF

!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     Line search and optimality tests.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

!     Generate the search direction d:=z-x.

      DO i = 1, n
         d(i) = z(i) - x(i)
      END DO
      CALL timer(cpu1)
      ENDIF
      IF(.NOT.first.OR..NOT.(task(1:5) == 'NEW_X')) THEN
      CALL lnsrlb(n,lower_bound,upper_bound,nbd,x,f,fold,gd,gdold,g,d,r,t,z,stp,dnorm,      &
                  dtd,xstep,step_max,iter,ifun,iback,nfgv,info,task,       &
                  boxed,constrained,csave,isave(22),dsave(17))
      IF (info /= 0 .OR. iback >= 20) THEN
!          restore the previous iterate.
         CALL dcopy(n,t,1,x,1)
         CALL dcopy(n,r,1,g,1)
         f = fold
         IF (col == 0) THEN
!             abnormal termination.
            IF (info == 0) THEN
               info = -9
!                restore the actual number of f and g evaluations etc.
               nfgv = nfgv - 1
               ifun = ifun - 1
               iback = iback - 1
            END IF
            task = 'ABNORMAL_TERMINATION_IN_LNSRCH'
            iter = iter + 1
            CALL timer(time2)
            time = time2 - time1
            CALL prn3lb(n,x,f,task,iprint,info,itfile,                        &
                        iter,nfgv,nintol,nskip,nact,g_inf_norm,                   &
                        time,nseg,word,iback,stp,xstep,k,                     &
                        cachyt,sbtime,lnscht)
            CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
            RETURN
         ELSE
!             refresh the lbfgs memory and restart the iteration.
            IF(iprint >= 1) WRITE (*, 1008)
            IF (info == 0) nfgv = nfgv - 1
            info   = 0
            col    = 0
            head   = 1
            theta  = one
            iupdat = 0
            updatd = .FALSE.
            task   = 'RESTART_FROM_LNSRCH'
            CALL timer(cpu2)
            lnscht = lnscht + cpu2 - cpu1
            first=.FALSE.
            CYCLE
         END IF
      ELSE IF (task(1:5) == 'FG_LN') THEN
!          return to the driver for calculating f and g; reenter at 666.
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      ELSE
!          calculate and print out the quantities related to the new X.
         CALL timer(cpu2)
         lnscht = lnscht + cpu2 - cpu1
         iter = iter + 1

!        Compute the infinity norm of the projected (-)gradient.

         CALL projgr(n,lower_bound,upper_bound,nbd,x,g,g_inf_norm)

!        Print iteration information.

         CALL prn2lb(n,x,f,g,iprint,itfile,iter,nfgv,nact,              &
                     g_inf_norm,nseg,word,iword,iback,stp,xstep)
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      END IF
      ENDIF

!     Test for termination.

      IF (g_inf_norm <= pgtol) THEN
!                                terminate the algorithm.
         task = 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
         CALL timer(time2)
         time = time2 - time1
         CALL prn3lb(n,x,f,task,iprint,info,itfile,                        &
                     iter,nfgv,nintol,nskip,nact,g_inf_norm,                   &
                     time,nseg,word,iback,stp,xstep,k,                     &
                     cachyt,sbtime,lnscht)
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      END IF

      ddum = MAX(ABS(fold), ABS(f), one)
      IF ((fold - f) <= tol*ddum) THEN
!                                        terminate the algorithm.
         task = 'CONVERGENCE: REL_REDUCTION_OF_F_<=_FACTR*EPSMCH'
         IF (iback >= 10) info = -5
!           i.e., to issue a warning if iback>10 in the line search.
         CALL timer(time2)
         time = time2 - time1
         CALL prn3lb(n,x,f,task,iprint,info,itfile,                        &
                     iter,nfgv,nintol,nskip,nact,g_inf_norm,                   &
                     time,nseg,word,iback,stp,xstep,k,                     &
                     cachyt,sbtime,lnscht)
         CALL save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,&
                            cpu1,cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
         RETURN
      END IF

!     Compute d=newx-oldx, r=newg-oldg, rr=y'y and dr=y's.

      DO i = 1, n
         r(i) = g(i) - r(i)
      END DO
      rr = ddot(n,r,1,r,1)
      IF (stp == one) THEN
         dr = gd - gdold
         ddum = -gdold
      ELSE
         dr = (gd - gdold)*stp
         CALL dscal(n,stp,d,1)
         ddum = -gdold*stp
      END IF

      IF (dr <= epsmch*ddum) THEN
!                            skip the L-BFGS update.
         nskip = nskip + 1
         updatd = .FALSE.
         IF (iprint >= 1) WRITE (*,1004) dr, ddum
         first=.FALSE.
         CYCLE
      END IF

!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     Update the L-BFGS matrix.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      updatd = .TRUE.
      iupdat = iupdat + 1

!     Update matrices WS and WY and form the middle matrix in B.

      CALL matupd(n,m,ws,wy,sy,ss,d,r,itail,                            &
                  iupdat,col,head,theta,rr,dr,stp,dtd)

!     Form the upper half of the pds T = theta*SS + L*D^(-1)*L';
!        Store T in the upper triangular of the array wt;
!        Cholesky factorize T to J*J' with
!           J' stored in the upper triangular of wt.

      CALL formt(m,wt,sy,ss,col,theta,info)

      IF (info /= 0) THEN
!          nonpositive definiteness in Cholesky factorization;
!          refresh the lbfgs memory and restart the iteration.
         IF(iprint >= 1) WRITE (*, 1007)
         info = 0
         col = 0
         head = 1
         theta = one
         iupdat = 0
         updatd = .FALSE.
      END IF

!     Now the inverse of the middle matrix in B is

!       [  D^(1/2)      O ] [ -D^(1/2)  D^(-1/2)*L' ]
!       [ -L*D^(-1/2)   J ] [  0        J'          ]

      first=.FALSE.
      END DO

 1001 FORMAT (//,'ITERATION ',i5)
 1002 FORMAT                                                            &
     &  (/,'At iterate',i5,4x,'f= ',1p,d12.5,4x,'|proj g|= ',1p,d12.5)
 1003 FORMAT (2(1x,i4),5x,'-',5x,'-',3x,'-',5x,'-',5x,'-',8x,'-',3x,    &
              1p,2(1x,d10.3))
 1004 FORMAT ('  ys=',1p,e10.3,'  -gs=',1p,e10.3,' BFGS update SKIPPED')
 1005 FORMAT (/,                                                        &
     &' Singular triangular system detected;',/,                        &
     &'   refresh the lbfgs memory and restart the iteration.')
 1006 FORMAT (/,                                                        &
     &' Nonpositive definiteness in Cholesky factorization in formk;',/,&
     &'   refresh the lbfgs memory and restart the iteration.')
 1007 FORMAT (/,                                                        &
     &' Nonpositive definiteness in Cholesky factorization in formt;',/,&
     &'   refresh the lbfgs memory and restart the iteration.')
 1008 FORMAT (/,                                                        &
     &' Bad direction in the line search;',/,                           &
     &'   refresh the lbfgs memory and restart the iteration.')

      RETURN

      END SUBROUTINE mainlb

! **************************************************************************************************
!> \brief This subroutine initializes iwhere and projects the initial x to the feasible set if necessary.
!> \param n ...
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd ...
!> \param x ...
!> \param iwhere  iwhere(i)=-1  if x(i) has no bounds
!>                           3   if l(i)=u(i)
!>                           0   otherwise.
!>                In cauchy, iwhere is given finer gradations.
!> \param iprint ...
!> \param x_projected ...
!> \param constrained ...
!> \param boxed ...
!> \author        NEOS, November 1994. (Latest revision June 1996.)
!>                Optimization Technology Center.
!>                Argonne National Laboratory and Northwestern University.
!>                Written by
!>                            Ciyou Zhu
!>                in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE active(n, lower_bound,upper_bound, nbd, x, iwhere, iprint,                &
                        x_projected, constrained, boxed)

      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n)
      INTEGER                                            :: nbd(n)
      REAL(KIND=dp)                                      :: x(n)
      INTEGER, INTENT(out)                               :: iwhere(n)
      INTEGER                                            :: iprint
      LOGICAL                                            :: x_projected, constrained, boxed

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: i, nbdd

!     ************
!     Initialize nbdd, x_projected, constrained and boxed.

      nbdd = 0
      x_projected = .FALSE.
      constrained = .FALSE.
      boxed = .TRUE.

!     Project the initial x to the easible set if necessary.

      DO i = 1, n
         IF (nbd(i) > 0) THEN
            IF (nbd(i) <= 2 .AND. x(i) <= lower_bound(i)) THEN
               IF (x(i) < lower_bound(i)) THEN
                  x_projected = .TRUE.
                  x(i) = lower_bound(i)
               END IF
               nbdd = nbdd + 1
            ELSE IF (nbd(i) >= 2 .AND. x(i) >= upper_bound(i)) THEN
               IF (x(i) > upper_bound(i)) THEN
                  x_projected = .TRUE.
                  x(i) = upper_bound(i)
               END IF
               nbdd = nbdd + 1
            END IF
         END IF
      END DO

!     Initialize iwhere and assign values to constrained and boxed.

      DO i = 1, n
         IF (nbd(i) /= 2) boxed = .FALSE.
         IF (nbd(i) == 0) THEN
!                                this variable is always free
            iwhere(i) = -1

!           otherwise set x(i)=mid(x(i), u(i), l(i)).
         ELSE
            constrained= .TRUE.
            IF (nbd(i) == 2 .AND. upper_bound(i) - lower_bound(i) <= zero) THEN
!                   this variable is always fixed
               iwhere(i) = 3
            ELSE
               iwhere(i) = 0
            END IF
         END IF
      END DO

      IF (iprint >= 0) THEN
         IF (x_projected) WRITE (*,*)                                        &
     &   'The initial X is infeasible.  Restart with its projection.'
         IF (.NOT. constrained)                                              &
            WRITE (*,*) 'This problem is unconstrained.'
      END IF

      IF (iprint > 0) WRITE (*,1001) nbdd

 1001 FORMAT (/,'At X0 ',i9,' variables are exactly at the bounds')

      RETURN

      END SUBROUTINE active

! **************************************************************************************************
!> \brief       This subroutine computes the product of the 2m x 2m middle matrix
!>              in the compact L-BFGS formula of B and a 2m vector v;
!>              it returns the product in p.
!> \param m     m is the maximum number of variable metric corrections
!>              used to define the limited memory matrix.
!> \param sy    sy specifies the matrix S'Y.
!> \param wt    wt specifies the upper triangular matrix J' which is
!>              the Cholesky factor of (thetaS'S+LD^(-1)L').
!> \param col   col specifies the number of s-vectors (or y-vectors)
!>              stored in the compact L-BFGS formula.
!> \param v     v specifies vector v.
!> \param p     p is the product Mv.
!> \param info  info = 0 for normal return,
!>                   = nonzero for abnormal return when the system to be solved by dtrsl is singular.
!> \author      NEOS, November 1994. (Latest revision June 1996.)
!>              Optimization Technology Center.
!>              Argonne National Laboratory and Northwestern University.
!>              Written by
!>                          Ciyou Zhu
!>              in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE bmv(m, sy, wt, col, v, p, info)

      INTEGER                                            :: m
      REAL(KIND=dp)                                      :: sy(m, m), wt(m, m)
      INTEGER                                            :: col
      REAL(KIND=dp), INTENT(in)                          :: v(2*col)
      REAL(KIND=dp), INTENT(out)                         :: p(2*col)
      INTEGER, INTENT(out)                               :: info

      INTEGER                                            :: i, i2, k
      REAL(KIND=dp)                                      :: sum

      IF (col == 0) RETURN

!     PART I: solve [  D^(1/2)      O ] [ p1 ] = [ v1 ]
!                   [ -L*D^(-1/2)   J ] [ p2 ]   [ v2 ].

!       solve Jp2=v2+LD^(-1)v1.
      p(col + 1) = v(col + 1)
      DO i = 2, col
         i2 = col + i
         sum = 0.0_dp
         DO k = 1, i - 1
            sum = sum + sy(i,k)*v(k)/sy(k,k)
         END DO
         p(i2) = v(i2) + sum
      END DO
!     Solve the triangular system
      CALL dtrsl(wt,m,col,p(col+1),11,info)
      IF (info /= 0) RETURN

!       solve D^(1/2)p1=v1.
      DO i = 1, col
         p(i) = v(i)/SQRT(sy(i,i))
      END DO

!     PART II: solve [ -D^(1/2)   D^(-1/2)*L'  ] [ p1 ] = [ p1 ]
!                    [  0         J'           ] [ p2 ]   [ p2 ].

!       solve J^Tp2=p2.
      CALL dtrsl(wt,m,col,p(col+1),01,info)
      IF (info /= 0) RETURN

!       compute p1=-D^(-1/2)(p1-D^(-1/2)L'p2)
!                 =-D^(-1/2)p1+D^(-1)L'p2.
      DO i = 1, col
         p(i) = -p(i)/SQRT(sy(i,i))
      END DO
      DO i = 1, col
         sum = 0._dp
         DO k = i + 1, col
            sum = sum + sy(k,i)*p(col+k)/sy(i,i)
         END DO
         p(i) = p(i) + sum
      END DO

      RETURN

      END SUBROUTINE bmv

! **************************************************************************************************
!> \brief        For given x, l, u, g (with g_inf_norm > 0), and a limited memory
!>               BFGS matrix B defined in terms of matrices WY, WS, WT, and
!>               scalars head, col, and theta, this subroutine computes the
!>               generalized Cauchy point (GCP), defined as the first local
!>               minimizer of the quadratic
!>
!>                    Q(x + s) = g's + 1/2 s'Bs
!>
!>               along the projected gradient direction P(x-tg,l,u).
!>               The routine returns the GCP in xcp.
!> \param n      n is the dimension of the problem.
!> \param x      x is the starting point for the GCP computation.
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd    nbd represents the type of bounds imposed on the
!>               variables, and must be specified as follows:
!>               nbd(i)=0 if x(i) is unbounded,
!>                      1 if x(i) has only a lower bound,
!>                      2 if x(i) has both lower and upper bounds, and
!>                      3 if x(i) has only an upper bound.
!> \param g      g is the gradient of f(x).  g must be a nonzero vector.
!> \param iorder iorder will be used to store the breakpoints in the piecewise
!>               linear path and free variables encountered. On exit,
!>               iorder(1),...,iorder(nleft) are indices of breakpoints
!>                                which have not been encountered;
!>               iorder(nleft+1),...,iorder(nbreak) are indices of
!>                                     encountered breakpoints; and
!>               iorder(nfree),...,iorder(n) are indices of variables which
!>               have no bound constraits along the search direction.
!> \param iwhere On entry iwhere indicates only the permanently fixed (iwhere=3)
!>               or free (iwhere= -1) components of x.
!>               On exit iwhere records the status of the current x variables.
!>               iwhere(i)=-3  if x(i) is free and has bounds, but is not moved
!>                          0   if x(i) is free and has bounds, and is moved
!>                          1   if x(i) is fixed at l(i), and l(i) .ne. u(i)
!>                          2   if x(i) is fixed at u(i), and u(i) .ne. l(i)
!>                          3   if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
!>                         -1  if x(i) is always free, i.e., it has no bounds.
!> \param t      t will be used to store the break points.
!> \param d      d is used to store the Cauchy direction P(x-tg)-x.
!> \param xcp    is a double precision array of dimension n used to return the GCP on exit.
!> \param m      m is the maximum number of variable metric corrections used to define the limited memory matrix.
!> \param wy     ws, wy, sy, and wt are double precision arrays.
!>               On entry they store information that defines the limited memory BFGS matrix:
!>               wy(n,m) stores Y, a set of y-vectors;
!> \param ws     ws(n,m) stores S, a set of s-vectors;
!> \param sy     sy(m,m) stores S'Y;
!> \param wt     wt(m,m) stores the Cholesky factorization of (theta*S'S+LD^(-1)L').
!> \param theta  theta is the scaling factor specifying B_0 = theta I.
!> \param col    col is the actual number of variable metric corrections stored so far.
!> \param head   head is the location of the first s-vector (or y-vector in S (or Y)
!> \param p      p will be used to store the vector p = W^(T)d.
!> \param c      c will be used to store the vector c = W^(T)(xcp-x).
!> \param wbp    wbp will be used to store the row of W corresponding to a breakpoint.
!> \param v      v is a double precision working array.
!> \param nseg   On exit nseg records the number of quadratic segments explored in searching for the GCP.
!> \param iprint iprint is an INTEGER variable that must be set by the user.
!>               It controls the frequency and type of output generated:
!>               iprint<0    no output is generated;
!>               iprint=0    print only one line at the last iteration;
!>               0<iprint<99 print also f and |proj g| every iprint iterations;
!>               iprint=99   print details of every iteration except n-vectors;
!>               iprint=100  print also the changes of active set and final x;
!>               iprint>100  print details of every iteration including x and g;
!>               When iprint > 0, the file iterate.dat will be created to summarize the iteration.
!> \param g_inf_norm g_inf_norm is the norm of the projected gradient at x.
!> \param info   On entry info is 0.
!>               On exit info = 0       for normal return,
!>                            = nonzero for abnormal return when the the system
!>                              used in routine bmv is singular.
!> \param epsmch ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE cauchy(n, x, lower_bound, upper_bound, nbd, g, iorder, iwhere, t, d, xcp,  &
                        m, wy, ws, sy, wt, theta, col, head, p, c, wbp, &
                        v, nseg, iprint, g_inf_norm, info, epsmch)
      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: x(n), lower_bound(n), upper_bound(n)
      INTEGER, INTENT(in)                                :: nbd(n)
      REAL(KIND=dp), INTENT(in)                          :: g(n)
      INTEGER                                            :: iorder(n)
      INTEGER, INTENT(inout)                             :: iwhere(n)
      REAL(KIND=dp)                                      :: t(n), d(n), xcp(n)
      INTEGER, INTENT(in)                                :: m
      REAL(KIND=dp), INTENT(in)                          :: sy(m, m), wt(m, m), theta
      INTEGER, INTENT(in)                                :: col
      REAL(KIND=dp), INTENT(in)                          :: ws(n, col), wy(n, col)
      INTEGER, INTENT(in)                                :: head
      REAL(KIND=dp)                                      :: p(2*m), c(2*m), wbp(2*m), v(2*m)
      INTEGER                                            :: nseg, iprint
      REAL(KIND=dp), INTENT(in)                          :: g_inf_norm
      INTEGER, INTENT(inout)                             :: info
      REAL(KIND=dp)                                      :: epsmch

      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: col2, i, ibkmin, ibp, iter, j, nbreak, &
                                                            nfree, nleft, pointr
      LOGICAL                                            :: bnded, xlower, xupper
      REAL(KIND=dp)                                      :: bkmin, ddot, dibp, dibp2, dt, dtm, f1, &
                                                            f2, f2_org, neggi, tj, tj0, tl, tsum, &
                                                            tu, wmc, wmp, wmw, zibp

!     References:
!
!       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
!       memory algorithm for bound constrained optimization'',
!       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
!
!       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
!       Subroutines for Large Scale Bound Constrained Optimization''
!       Tech. Report, NAM-11, EECS Department, Northwestern University,
!       1994.
!
!       (Postscript files of these papers are available via anonymous
!        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
!
!                           *  *  *
!     Check the status of the variables, reset iwhere(i) if necessary;
!       compute the Cauchy direction d and the breakpoints t; initialize
!       the derivative f1 and the vector p = W'd (for theta = 1).

      IF (g_inf_norm <= zero) THEN
         IF (iprint >= 0) WRITE (*,*) 'Subgnorm = 0.  GCP = X.'
         CALL dcopy(n,x,1,xcp,1)
         RETURN
      END IF
      bnded = .TRUE.
      nfree = n + 1
      nbreak = 0
      ibkmin = 0
      bkmin = zero
      col2 = 2*col
      f1 = zero
      IF (iprint >= 99) WRITE (*,3010)

!     We set p to zero and build it up as we determine d.

      DO i = 1, col2
         p(i) = zero
      END DO

!     In the following loop we determine for each variable its bound
!        status and its breakpoint, and update p accordingly.
!        Smallest breakpoint is identified.

      DO i = 1, n
         neggi = -g(i)
         IF (iwhere(i) /= 3 .AND. iwhere(i) /= -1) THEN
!             if x(i) is not a constant and has bounds,
!             compute the difference between x(i) and its bounds.
            IF (nbd(i) <= 2) tl = x(i) - lower_bound(i)
            IF (nbd(i) >= 2) tu = upper_bound(i) - x(i)

!           If a variable is close enough to a bound
!             we treat it as at bound.
            xlower = nbd(i) <= 2 .AND. tl <= zero
            xupper = nbd(i) >= 2 .AND. tu <= zero

!              reset iwhere(i).
            iwhere(i) = 0
            IF (xlower) THEN
               IF (neggi <= zero) iwhere(i) = 1
            ELSE IF (xupper) THEN
               IF (neggi >= zero) iwhere(i) = 2
            ELSE
               IF (ABS(neggi) <= zero) iwhere(i) = -3
            END IF
         END IF
         pointr = head
         IF (iwhere(i) /= 0 .AND. iwhere(i) /= -1) THEN
            d(i) = zero
         ELSE
            d(i) = neggi
            f1 = f1 - neggi*neggi
!             calculate p := p - W'e_i* (g_i).
            DO j = 1, col
               p(j) = p(j) +  wy(i,pointr)* neggi
               p(col + j) = p(col + j) + ws(i,pointr)*neggi
               pointr = MOD(pointr,m) + 1
            END DO
            IF (nbd(i) <= 2 .AND. nbd(i) /= 0                       &
     &                        .AND. neggi < zero) THEN
!                                 x(i) + d(i) is bounded; compute t(i).
               nbreak = nbreak + 1
               iorder(nbreak) = i
               t(nbreak) = tl/(-neggi)
               IF (nbreak == 1 .OR. t(nbreak) < bkmin) THEN
                  bkmin = t(nbreak)
                  ibkmin = nbreak
               END IF
            ELSE IF (nbd(i) >= 2 .AND. neggi > zero) THEN
!                                 x(i) + d(i) is bounded; compute t(i).
               nbreak = nbreak + 1
               iorder(nbreak) = i
               t(nbreak) = tu/neggi
               IF (nbreak == 1 .OR. t(nbreak) < bkmin) THEN
                  bkmin = t(nbreak)
                  ibkmin = nbreak
               END IF
            ELSE
!                x(i) + d(i) is not bounded.
               nfree = nfree - 1
               iorder(nfree) = i
               IF (ABS(neggi) > zero) bnded = .FALSE.
            END IF
         END IF
      END DO

!     The indices of the nonzero components of d are now stored
!       in iorder(1),...,iorder(nbreak) and iorder(nfree),...,iorder(n).
!       The smallest of the nbreak breakpoints is in t(ibkmin)=bkmin.

      IF (theta /= one) THEN
!                   complete the initialization of p for theta not= one.
         CALL dscal(col,theta,p(col+1),1)
      END IF

!     Initialize GCP xcp = x.

      CALL dcopy(n,x,1,xcp,1)

      IF (nbreak == 0 .AND. nfree == n + 1) THEN
!                  is a zero vector, return with the initial xcp as GCP.
         IF (iprint > 100) WRITE (*,1010) (xcp(i), i = 1, n)
         RETURN
      END IF

!     Initialize c = W'(xcp - x) = 0.

      DO j = 1, col2
         c(j) = zero
      END DO

!     Initialize derivative f2.

      f2 =  -theta*f1
      f2_org  =  f2
      IF (col > 0) THEN
         CALL bmv(m,sy,wt,col,p,v,info)
         IF (info /= 0) RETURN
         f2 = f2 - ddot(col2,v,1,p,1)
      END IF
      dtm = -f1/f2
      tsum = zero
      nseg = 1
      IF (iprint >= 99)                                               &
         WRITE (*,*) 'There are ',nbreak,'  breakpoints '



      nleft = nbreak
      iter = 1


      tj = zero

!     If there are no breakpoints, locate the GCP and return.

      IF (nleft == 0) THEN
         IF (iprint >= 99) THEN
            WRITE (*,*)
            WRITE (*,*) 'GCP found in this segment'
            WRITE (*,4010) nseg,f1,f2
            WRITE (*,6010) dtm
         END IF
         IF (dtm <= zero) dtm = zero
         tsum = tsum + dtm

!        Move free variables (i.e., the ones w/o breakpoints) and
!          the variables whose breakpoints haven't been reached.

         CALL daxpy(n,tsum,d,1,xcp,1)
      END IF

      DO WHILE(nleft>0)

!     Find the next smallest breakpoint;
!       compute dt = t(nleft) - t(nleft + 1).

      tj0 = tj
      IF (iter == 1) THEN
!         Since we already have the smallest breakpoint we need not do
!         heapsort yet. Often only one breakpoint is used and the
!         cost of heapsort is avoided.
         tj = bkmin
         ibp = iorder(ibkmin)
      ELSE
         IF (iter == 2) THEN
!             Replace the already used smallest breakpoint with the
!             breakpoint numbered nbreak > nlast, before heapsort call.
            IF (ibkmin /= nbreak) THEN
               t(ibkmin) = t(nbreak)
               iorder(ibkmin) = iorder(nbreak)
            END IF
!        Update heap structure of breakpoints
!           (if iter=2, initialize heap).
         END IF
         CALL hpsolb(nleft,t,iorder,iter-2)
         tj = t(nleft)
         ibp = iorder(nleft)
      END IF

      dt = tj - tj0

      IF (dt /= zero .AND. iprint >= 100) THEN
         WRITE (*,4011) nseg,f1,f2
         WRITE (*,5010) dt
         WRITE (*,6010) dtm
      END IF

!     If a minimizer is within this interval, locate the GCP and return.

      IF (dtm < dt) THEN
         IF (iprint >= 99) THEN
            WRITE (*,*)
            WRITE (*,*) 'GCP found in this segment'
            WRITE (*,4010) nseg,f1,f2
            WRITE (*,6010) dtm
         END IF
         IF (dtm <= zero) dtm = zero
         tsum = tsum + dtm

!        Move free variables (i.e., the ones w/o breakpoints) and
!          the variables whose breakpoints haven't been reached.

         CALL daxpy(n,tsum,d,1,xcp,1)
         EXIT
      END IF

!     Otherwise fix one variable and
!       reset the corresponding component of d to zero.

      tsum = tsum + dt
      nleft = nleft - 1
      iter = iter + 1
      dibp = d(ibp)
      d(ibp) = zero
      IF (dibp > zero) THEN
         zibp = upper_bound(ibp) - x(ibp)
         xcp(ibp) = upper_bound(ibp)
         iwhere(ibp) = 2
      ELSE
         zibp = lower_bound(ibp) - x(ibp)
         xcp(ibp) = lower_bound(ibp)
         iwhere(ibp) = 1
      END IF
      IF (iprint >= 100) WRITE (*,*) 'Variable  ',ibp,'  is fixed.'
      IF (nleft == 0 .AND. nbreak == n) THEN
!                                             all n variables are fixed,
!                                                return with xcp as GCP.
         dtm = dt
         EXIT
      END IF

!     Update the derivative information.

      nseg = nseg + 1
      dibp2 = dibp**2

!     Update f1 and f2.

!        temporarily set f1 and f2 for col=0.
      f1 = f1 + dt*f2 + dibp2 - theta*dibp*zibp
      f2 = f2 - theta*dibp2

      IF (col > 0) THEN
!                          update c = c + dt*p.
         CALL daxpy(col2,dt,p,1,c,1)

!           choose wbp,
!           the row of W corresponding to the breakpoint encountered.
         pointr = head
         DO j = 1,col
            wbp(j) = wy(ibp,pointr)
            wbp(col + j) = theta*ws(ibp,pointr)
            pointr = MOD(pointr,m) + 1
         END DO

!           compute (wbp)Mc, (wbp)Mp, and (wbp)M(wbp)'.
         CALL bmv(m,sy,wt,col,wbp,v,info)
         IF (info /= 0) RETURN
         wmc = ddot(col2,c,1,v,1)
         wmp = ddot(col2,p,1,v,1)
         wmw = ddot(col2,wbp,1,v,1)

!           update p = p - dibp*wbp.
         CALL daxpy(col2,-dibp,wbp,1,p,1)

!           complete updating f1 and f2 while col > 0.
         f1 = f1 + dibp*wmc
         f2 = f2 + 2.0_dp*dibp*wmp - dibp2*wmw
      END IF

      f2 = MAX(epsmch*f2_org,f2)
      IF (nleft > 0) THEN
         dtm = -f1/f2
         CYCLE
!                 to repeat the loop for unsearched intervals.
      ELSE 
         IF(bnded) THEN
            f1 = zero
            f2 = zero
            dtm = zero
         ELSE
            dtm = -f1/f2
         END IF
         IF (iprint >= 99) THEN
            WRITE (*,*)
            WRITE (*,*) 'GCP found in this segment'
            WRITE (*,4010) nseg,f1,f2
            WRITE (*,6010) dtm
         END IF
         IF (dtm <= zero) dtm = zero
         tsum = tsum + dtm

!        Move free variables (i.e., the ones w/o breakpoints) and
!          the variables whose breakpoints haven't been reached.

         CALL daxpy(n,tsum,d,1,xcp,1)
         EXIT
      END IF
      END DO

!     Update c = c + dtm*p = W'(x^c - x)
!       which will be used in computing r = Z'(B(x^c - x) + g).

      IF (col > 0) CALL daxpy(col2,dtm,p,1,c,1)
      IF (iprint > 100) WRITE (*,1010) (xcp(i),i = 1,n)
      IF (iprint >= 99) WRITE (*,2010)

 1010 FORMAT ('Cauchy X =  ',/,(4x,1p,6(1x,d11.4)))
 2010 FORMAT (/,'---------------- exit CAUCHY----------------------',/)
 3010 FORMAT (/,'---------------- CAUCHY entered-------------------')
 4010 FORMAT ('Piece    ',i3,' --f1, f2 at start point ',1p,2(1x,d11.4))
 4011 FORMAT (/,'Piece    ',i3,' --f1, f2 at start point ',             &
              1p,2(1x,d11.4))
 5010 FORMAT ('Distance to the next break point =  ',1p,d11.4)
 6010 FORMAT ('Distance to the stationary point =  ',1p,d11.4)

      RETURN

      END SUBROUTINE cauchy

! **************************************************************************************************
!> \brief        This subroutine computes r=-Z'B(xcp-xk)-Z'g by using
!>               wa(2m+1)=W'(xcp-x) from subroutine cauchy.
!> \param n ...
!> \param m ...
!> \param x ...
!> \param g ...
!> \param ws ...
!> \param wy ...
!> \param sy ...
!> \param wt ...
!> \param z ...
!> \param r ...
!> \param wa ...
!> \param index ...
!> \param theta ...
!> \param col ...
!> \param head ...
!> \param nfree ...
!> \param constrained ...
!> \param info ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE cmprlb(n, m, x, g, ws, wy, sy, wt, z, r, wa, index,    &
                       theta, col, head, nfree, constrained, info)

      INTEGER, INTENT(in)                                :: n, m
      REAL(KIND=dp), INTENT(in)                          :: x(n), g(n), ws(n, m), wy(n, m), &
                                                            sy(m, m), wt(m, m), z(n)
      REAL(KIND=dp), INTENT(out)                         :: r(n), wa(4*m)
      INTEGER, INTENT(in)                                :: INDEX(n)
      REAL(KIND=dp), INTENT(in)                          :: theta
      INTEGER, INTENT(in)                                :: col, head, nfree
      LOGICAL, INTENT(in)                                :: constrained
      INTEGER                                            :: info

      INTEGER                                            :: i, j, k, pointr
      REAL(KIND=dp)                                      :: a1, a2

      IF (.NOT. constrained .AND. col > 0) THEN
         DO i = 1, n
            r(i) = -g(i)
         END DO
      ELSE
         DO i = 1, nfree
            k = INDEX(i)
            r(i) = -theta*(z(k) - x(k)) - g(k)
         END DO
         CALL bmv(m,sy,wt,col,wa(2*m+1),wa(1),info)
         IF (info /= 0) THEN
            info = -8
            RETURN
         END IF
         pointr = head
         DO j = 1, col
            a1 = wa(j)
            a2 = theta*wa(col + j)
            DO i = 1, nfree
               k = INDEX(i)
               r(i) = r(i) + wy(k,pointr)*a1 + ws(k,pointr)*a2
            END DO
            pointr = MOD(pointr,m) + 1
         END DO
      END IF

      RETURN

      END SUBROUTINE cmprlb

! **************************************************************************************************
!> \brief       This subroutine checks the validity of the input data.
!> \param n ...
!> \param m ...
!> \param factr ...
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd ...
!> \param task ...
!> \param info ...
!> \param k ...
!> \author      NEOS, November 1994. (Latest revision June 1996.)
!>              Optimization Technology Center.
!>              Argonne National Laboratory and Northwestern University.
!>              Written by
!>                          Ciyou Zhu
!>              in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE errclb(n, m, factr, lower_bound, upper_bound, nbd, task, info, k)

      INTEGER, INTENT(in)                                :: n, m
      REAL(KIND=dp), INTENT(in)                          :: factr, lower_bound(n), upper_bound(n)
      INTEGER                                            :: nbd(n)
      CHARACTER(LEN=60)                                  :: task
      INTEGER                                            :: info, k

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: i

!     Check the input arguments for errors.

      IF (n <= 0) task = 'ERROR: N <= 0'
      IF (m <= 0) task = 'ERROR: M <= 0'
      IF (factr < zero) task = 'ERROR: FACTR < 0'

!     Check the validity of the arrays nbd(i), u(i), and l(i).

      DO i = 1, n
         IF (nbd(i) < 0 .OR. nbd(i) > 3) THEN
!                                                   return
            task = 'ERROR: INVALID NBD'
            info = -6
            k = i
         END IF
         IF (nbd(i) == 2) THEN
            IF (lower_bound(i) > upper_bound(i)) THEN
!                                    return
               task = 'ERROR: NO FEASIBLE SOLUTION'
               info = -7
               k = i
            END IF
         END IF
      END DO

      RETURN

      END SUBROUTINE errclb

! **************************************************************************************************
!> \brief        This subroutine forms  the LEL^T factorization of the indefinite
!>               matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!>                             [L_a -R_z           theta*S'AA'S ]
!>               where     E = [-I  0]
!>                             [ 0  I]
!>               The matrix K can be shown to be equal to the matrix M^[-1]N
!>               occurring in section 5.1 of [1], as well as to the matrix
!>               Mbar^[-1] Nbar in section 5.3.
!> \param n      n is the dimension of the problem.
!> \param nsub   nsub is the number of subspace variables in free set.
!> \param ind    ind specifies the indices of subspace variables.
!> \param nenter nenter is the number of variables entering the free set.
!> \param ileave indx2(ileave),...,indx2(n) are the variables leaving the free set.
!> \param indx2  indx2(1),...,indx2(nenter) are the variables entering the free set,
!>               while indx2(ileave),...,indx2(n) are the variables leaving the free set.
!> \param iupdat iupdat is the total number of BFGS updates made so far.
!> \param updatd 'updatd' is true if the L-BFGS matrix is updatd.
!> \param wn     the upper triangle of wn stores the LEL^T factorization
!>               of the 2*col x 2*col indefinite matrix
!>                     [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!>                     [L_a -R_z           theta*S'AA'S ]
!> \param wn1    On entry wn1 stores the lower triangular part of
!>                     [Y' ZZ'Y   L_a'+R_z']
!>                     [L_a+R_z   S'AA'S   ]
!>               in the previous iteration.
!>               On exit wn1 stores the corresponding updated matrices.
!>               The purpose of wn1 is just to store these inner products
!>               so they can be easily updated and inserted into wn.
!> \param m      m is the maximum number of variable metric corrections
!>               used to define the limited memory matrix.
!> \param ws     ws(n,m) stores S, a set of s-vectors;
!> \param wy     wy(n,m) stores Y, a set of y-vectors;
!> \param sy     sy(m,m) stores S'Y;
!> \param theta  is the scaling factor specifying B_0 = theta I;
!> \param col    is the number of variable metric corrections stored;
!> \param head   is the location of the 1st s- (or y-) vector in S (or Y).
!> \param info   info =  0 for normal return;
!>                    = -1 when the 1st Cholesky factorization failed;
!>                    = -2 when the 2st Cholesky factorization failed.
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE formk(n, nsub, ind, nenter, ileave, indx2, iupdat,     &
                       updatd, wn, wn1, m, ws, wy, sy, theta, col,      &
                       head, info)

      INTEGER, INTENT(in)                                :: n, nsub, ind(n), nenter, ileave, &
                                                            indx2(n), iupdat
      LOGICAL                                            :: updatd
      INTEGER, INTENT(in)                                :: m
      REAL(KIND=dp)                                      :: wn1(2*m, 2*m)
      REAL(KIND=dp), INTENT(out)                         :: wn(2*m, 2*m)
      REAL(KIND=dp), INTENT(in)                          :: ws(n, m), wy(n, m), sy(m, m), theta
      INTEGER, INTENT(in)                                :: col, head
      INTEGER, INTENT(out)                               :: info

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: col2, dbegin, dend, i, ipntr, is, is1, &
                                                            iy, jpntr, js, js1, jy, k, k1, m2, &
                                                            pbegin, pend, upcl
      REAL(KIND=dp)                                      :: ddot, temp1, temp2, temp3, temp4

!     References:
!       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
!       memory algorithm for bound constrained optimization'',
!       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
!
!       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: a
!       limited memory FORTRAN code for solving bound constrained
!       optimization problems'', Tech. Report, NAM-11, EECS Department,
!       Northwestern University, 1994.
!
!       (Postscript files of these papers are available via anonymous
!        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
!
!                           *  *  *
!     Form the lower triangular part of
!               WN1 = [Y' ZZ'Y   L_a'+R_z']
!                     [L_a+R_z   S'AA'S   ]
!        where L_a is the strictly lower triangular part of S'AA'Y
!              R_z is the upper triangular part of S'ZZ'Y.

      IF (updatd) THEN
         IF (iupdat > m) THEN
!                                 shift old part of WN1.
            DO jy = 1, m - 1
               js = m + jy
               CALL dcopy(m-jy,wn1(jy+1,jy+1),1,wn1(jy,jy),1)
               CALL dcopy(m-jy,wn1(js+1,js+1),1,wn1(js,js),1)
               CALL dcopy(m-1,wn1(m+2,jy+1),1,wn1(m+1,jy),1)
            END DO
         END IF

!          put new rows in blocks (1,1), (2,1) and (2,2).
         pbegin = 1
         pend = nsub
         dbegin = nsub + 1
         dend = n
         iy = col
         is = m + col
         ipntr = head + col - 1
         IF (ipntr > m) ipntr = ipntr - m
         jpntr = head
         DO jy = 1, col
            js = m + jy
            temp1 = zero
            temp2 = zero
            temp3 = zero
!             compute element jy of row 'col' of Y'ZZ'Y
            DO k = pbegin, pend
               k1 = ind(k)
               temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
            END DO
!             compute elements jy of row 'col' of L_a and S'AA'S
            DO k = dbegin, dend
               k1 = ind(k)
               temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
            END DO
            wn1(iy,jy) = temp1
            wn1(is,js) = temp2
            wn1(is,jy) = temp3
            jpntr = MOD(jpntr,m) + 1
         END DO

!          put new column in block (2,1).
         jy = col
         jpntr = head + col - 1
         IF (jpntr > m) jpntr = jpntr - m
         ipntr = head
         DO i = 1, col
            is = m + i
            temp3 = zero
!             compute element i of column 'col' of R_z
            DO k = pbegin, pend
               k1 = ind(k)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
            END DO
            ipntr = MOD(ipntr,m) + 1
            wn1(is,jy) = temp3
         END DO
         upcl = col - 1
      ELSE
         upcl = col
      END IF

!       modify the old parts in blocks (1,1) and (2,2) due to changes
!       in the set of free variables.
      ipntr = head
      DO iy = 1, upcl
         is = m + iy
         jpntr = head
         DO jy = 1, iy
            js = m + jy
            temp1 = zero
            temp2 = zero
            temp3 = zero
            temp4 = zero
            DO k = 1, nenter
               k1 = indx2(k)
               temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
               temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
            END DO
            DO k = ileave, n
               k1 = indx2(k)
               temp3 = temp3 + wy(k1,ipntr)*wy(k1,jpntr)
               temp4 = temp4 + ws(k1,ipntr)*ws(k1,jpntr)
            END DO
            wn1(iy,jy) = wn1(iy,jy) + temp1 - temp3
            wn1(is,js) = wn1(is,js) - temp2 + temp4
            jpntr = MOD(jpntr,m) + 1
         END DO
         ipntr = MOD(ipntr,m) + 1
      END DO

!       modify the old parts in block (2,1).
      ipntr = head
      DO is = m + 1, m + upcl
         jpntr = head
         DO jy = 1, upcl
            temp1 = zero
            temp3 = zero
            DO k = 1, nenter
               k1 = indx2(k)
               temp1 = temp1 + ws(k1,ipntr)*wy(k1,jpntr)
            END DO
            DO k = ileave, n
               k1 = indx2(k)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
            END DO
         IF (is <= jy + m) THEN
               wn1(is,jy) = wn1(is,jy) + temp1 - temp3
            ELSE
               wn1(is,jy) = wn1(is,jy) - temp1 + temp3
            END IF
            jpntr = MOD(jpntr,m) + 1
         END DO
         ipntr = MOD(ipntr,m) + 1
      END DO

!     Form the upper triangle of WN = [D+Y' ZZ'Y/theta   -L_a'+R_z' ]
!                                     [-L_a +R_z        S'AA'S*theta]

      m2 = 2*m
      DO iy = 1, col
         is = col + iy
         is1 = m + iy
         DO jy = 1, iy
            js = col + jy
            js1 = m + jy
            wn(jy,iy) = wn1(iy,jy)/theta
            wn(js,is) = wn1(is1,js1)*theta
         END DO
         DO jy = 1, iy - 1
            wn(jy,is) = -wn1(is1,jy)
         END DO
         DO jy = iy, col
            wn(jy,is) = wn1(is1,jy)
         END DO
         wn(iy,iy) = wn(iy,iy) + sy(iy,iy)
      END DO

!     Form the upper triangle of WN= [  LL'            L^-1(-L_a'+R_z')]
!                                    [(-L_a +R_z)L'^-1   S'AA'S*theta  ]

!        first Cholesky factor (1,1) block of wn to get LL'
!                          with L' stored in the upper triangle of wn.
      CALL dpofa(wn,m2,col,info)
      IF (info /= 0) THEN
         info = -1
         RETURN
      END IF
!        then form L^-1(-L_a'+R_z') in the (1,2) block.
      col2 = 2*col
      DO js = col+1 ,col2
         CALL dtrsl(wn,m2,col,wn(1,js),11,info)
      END DO

!     Form S'AA'S*theta + (L^-1(-L_a'+R_z'))'L^-1(-L_a'+R_z') in the
!        upper triangle of (2,2) block of wn.


      DO is = col+1, col2
         DO js = is, col2
               wn(is,js) = wn(is,js) + ddot(col,wn(1,is),1,wn(1,js),1)
         END DO
      END DO

!     Cholesky factorization of (2,2) block of wn.

      CALL dpofa(wn(col+1,col+1),m2,col,info)
      IF (info /= 0) THEN
         info = -2
         RETURN
      END IF

      RETURN

      END SUBROUTINE formk

! **************************************************************************************************
!> \brief       This subroutine forms the upper half of the pos. def. and symm.
!>              T = theta*SS + L*D^(-1)*L', stores T in the upper triangle
!>              of the array wt, and performs the Cholesky factorization of T
!>              to produce J*J', with J' stored in the upper triangle of wt.
!> \param m ...
!> \param wt ...
!> \param sy ...
!> \param ss ...
!> \param col ...
!> \param theta ...
!> \param info ...
!> \author      NEOS, November 1994. (Latest revision June 1996.)
!>              Optimization Technology Center.
!>              Argonne National Laboratory and Northwestern University.
!>              Written by
!>                          Ciyou Zhu
!>              in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE formt(m, wt, sy, ss, col, theta, info)

      INTEGER                                            :: m
      REAL(KIND=dp)                                      :: wt(m, m), sy(m, m), ss(m, m)
      INTEGER                                            :: col
      REAL(KIND=dp)                                      :: theta
      INTEGER                                            :: info

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: i, j, k, k1
      REAL(KIND=dp)                                      :: ddum

!     Form the upper half of  T = theta*SS + L*D^(-1)*L',
!        store T in the upper triangle of the array wt.

      DO j = 1, col
         wt(1,j) = theta*ss(1,j)
      END DO
      DO i = 2, col
         DO j = i, col
            k1 = MIN(i,j) - 1
            ddum  = zero
            DO k = 1, k1
               ddum  = ddum + sy(i,k)*sy(j,k)/sy(k,k)
            END DO
            wt(i,j) = ddum + theta*ss(i,j)
         END DO
      END DO

!     Cholesky factorize T to J*J' with
!        J' stored in the upper triangle of wt.

      CALL dpofa(wt,m,col,info)
      IF (info /= 0) THEN
         info = -3
      END IF

      RETURN

      END SUBROUTINE formt

! **************************************************************************************************
!> \brief        This subroutine counts the entering and leaving variables when
!>               iter > 0, and finds the index set of free and active variables
!>               at the GCP.
!> \param n ...
!> \param nfree ...
!> \param index  for i=1,...,nfree, index(i) are the indices of free variables
!>               for i=nfree+1,...,n, index(i) are the indices of bound variables
!>               On entry after the first iteration, index gives
!>               the free variables at the previous iteration.
!>               On exit it gives the free variables based on the determination
!>               in cauchy using the array iwhere.
!> \param nenter ...
!> \param ileave ...
!> \param indx2  On exit with iter>0, indx2 indicates which variables
!>               have changed status since the previous iteration.
!>               For i= 1,...,nenter, indx2(i) have changed from bound to free.
!>               For i= ileave+1,...,n, indx2(i) have changed from free to bound.
!> \param iwhere ...
!> \param wrk ...
!> \param updatd ...
!> \param constrained     A variable indicating whether bounds are present
!> \param iprint ...
!> \param iter ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE freev(n, nfree, index, nenter, ileave, indx2,          &
                       iwhere, wrk, updatd, constrained, iprint, iter)

      INTEGER                                            :: n, nfree
      INTEGER, INTENT(inout)                             :: INDEX(n)
      INTEGER                                            :: nenter, ileave
      INTEGER, INTENT(out)                               :: indx2(n)
      INTEGER                                            :: iwhere(n)
      LOGICAL                                            :: wrk, updatd, constrained
      INTEGER                                            :: iprint, iter

      INTEGER                                            :: i, iact, k

      nenter = 0
      ileave = n + 1
      IF (iter > 0 .AND. constrained) THEN
!                           count the entering and leaving variables.
         DO i = 1, nfree
            k = INDEX(i)

!            write(*,*) ' k  = index(i) ', k
!            write(*,*) ' index = ', i

            IF (iwhere(k) > 0) THEN
               ileave = ileave - 1
               indx2(ileave) = k
               IF (iprint >= 100) WRITE (*,*)                         &
     &             'Variable ',k,' leaves the set of free variables'
            END IF
         END DO
         DO i = 1 + nfree, n
            k = INDEX(i)
            IF (iwhere(k) <= 0) THEN
               nenter = nenter + 1
               indx2(nenter) = k
               IF (iprint >= 100) WRITE (*,*)                         &
     &             'Variable ',k,' enters the set of free variables'
            END IF
         END DO
         IF (iprint >= 99) WRITE (*,*)                                &
             n+1-ileave,' variables leave; ',nenter,' variables enter'
      END IF
      wrk = (ileave < n+1) .OR. (nenter > 0) .OR. updatd

!     Find the index set of free and active variables at the GCP.

      nfree = 0
      iact = n + 1
      DO i = 1, n
         IF (iwhere(i) <= 0) THEN
            nfree = nfree + 1
            INDEX(nfree) = i
         ELSE
            iact = iact - 1
            INDEX(iact) = i
         END IF
      END DO
      IF (iprint >= 99) WRITE (*,*)                                   &
            nfree,' variables are free at GCP ',iter + 1

      RETURN

      END SUBROUTINE freev

! **************************************************************************************************
!> \brief        This subroutine sorts out the least element of t, and puts the
!>               remaining elements of t in a heap.
!> \param n      n is the dimension of the arrays t and iorder.
!> \param t      On entry t stores the elements to be sorted,
!>               On exit t(n) stores the least elements of t, and t(1) to t(n-1)
!>               stores the remaining elements in the form of a heap.
!> \param iorder On entry iorder(i) is the index of t(i).
!>               On exit iorder(i) is still the index of t(i), but iorder may be
!>               permuted in accordance with t.
!> \param iheap  iheap should be set as follows:
!>               iheap .eq. 0 if t(1) to t(n) is not in the form of a heap,
!>               iheap .ne. 0 if otherwise.
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE hpsolb(n, t, iorder, iheap)
      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(inout)                       :: t(n)
      INTEGER, INTENT(inout)                             :: iorder(n)
      INTEGER, INTENT(in)                                :: iheap

      INTEGER                                            :: i, indxin, indxou, j, k
      REAL(KIND=dp)                                      :: ddum, out

!
!     References:
!       Algorithm 232 of CACM (J. W. J. Williams): HEAPSORT.
!
!                           *  *  *

      IF (iheap == 0) THEN

!        Rearrange the elements t(1) to t(n) to form a heap.

         DO k = 2, n
            ddum  = t(k)
            indxin = iorder(k)

!           Add ddum to the heap.
            i = k
            DO WHILE (i> 1)
               j = i/2
               IF (ddum < t(j)) THEN
                  t(i) = t(j)
                  iorder(i) = iorder(j)
                  i = j
               ELSE
                  EXIT
               END IF
            END DO
            t(i) = ddum
            iorder(i) = indxin
         END DO
      END IF

!     Assign to 'out' the value of t(1), the least member of the heap,
!        and rearrange the remaining members to form a heap as
!        elements 1 to n-1 of t.

      IF (n > 1) THEN
         i = 1
         out = t(1)
         indxou = iorder(1)
         ddum  = t(n)
         indxin  = iorder(n)

!        Restore the heap
         j = 2*i
         DO WHILE (j<= n-1)
            IF (t(j+1) < t(j)) j = j+1
            IF (t(j) < ddum ) THEN
               t(i) = t(j)
               iorder(i) = iorder(j)
               i = j
            ELSE
               EXIT
            END IF
            j=2*i
         END DO
         t(i) = ddum
         iorder(i) = indxin

!     Put the least member in t(n).

         t(n) = out
         iorder(n) = indxou
      END IF

      RETURN

      END SUBROUTINE hpsolb

! **************************************************************************************************
!> \brief        This subroutine calls subroutine dcsrch from the Minpack2 library
!>               to perform the line search.  Subroutine dscrch is safeguarded so
!>               that all trial points lie within the feasible region.
!> \param n ...
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd ...
!> \param x ...
!> \param f ...
!> \param fold ...
!> \param gd ...
!> \param gdold ...
!> \param g ...
!> \param d ...
!> \param r ...
!> \param t ...
!> \param z ...
!> \param stp ...
!> \param dnorm ...
!> \param dtd ...
!> \param xstep ...
!> \param step_max ...
!> \param iter ...
!> \param ifun ...
!> \param iback ...
!> \param nfgv ...
!> \param info ...
!> \param task ...
!> \param boxed ...
!> \param constrained ...
!> \param csave ...
!> \param isave ...
!> \param dsave ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE lnsrlb(n, lower_bound, upper_bound, nbd, x, f, fold, gd, gdold, g, d, r, t,&
                        z, stp, dnorm, dtd, xstep, step_max, iter, ifun,   &
                        iback, nfgv, info, task, boxed, constrained, csave,  &
                        isave, dsave)

      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n)
      INTEGER                                            :: nbd(n)
      REAL(KIND=dp)                                      :: x(n), f, fold, gd, gdold, g(n), d(n), &
                                                            r(n), t(n), z(n), stp, dnorm, dtd, &
                                                            xstep, step_max
      INTEGER                                            :: iter, ifun, iback, nfgv, info
      CHARACTER(LEN=60)                                  :: task
      LOGICAL                                            :: boxed, constrained
      CHARACTER(LEN=60)                                  :: csave
      INTEGER                                            :: isave(2)
      REAL(KIND=dp)                                      :: dsave(13)

      REAL(KIND=dp), PARAMETER                           :: big = 1.0E10_dp, ftol = 1.0E-3_dp, &
                                                            gtol = 0.9_dp, one = 1.0_dp, &
                                                            xtol = 0.1_dp, zero = 0.0_dp

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: a1, a2, ddot

      IF (.NOT.(task(1:5) == 'FG_LN')) THEN

      dtd = ddot(n,d,1,d,1)
      dnorm = SQRT(dtd)

!     Determine the maximum step length.

      step_max = big
      IF (constrained) THEN
         IF (iter == 0) THEN
            step_max = one
         ELSE
            DO i = 1, n
               a1 = d(i)
               IF (nbd(i) /= 0) THEN
                  IF (a1 < zero .AND. nbd(i) <= 2) THEN
                     a2 = lower_bound(i) - x(i)
                     IF (a2 >= zero) THEN
                        step_max = zero
                     ELSE IF (a1*step_max < a2) THEN
                        step_max = a2/a1
                     END IF
                  ELSE IF (a1 > zero .AND. nbd(i) >= 2) THEN
                     a2 = upper_bound(i) - x(i)
                     IF (a2 <= zero) THEN
                        step_max = zero
                     ELSE IF (a1*step_max > a2) THEN
                        step_max = a2/a1
                     END IF
                  END IF
               END IF
           END DO
         END IF
      END IF

      IF (iter == 0 .AND. .NOT. boxed) THEN
         stp = MIN(one/dnorm, step_max)
      ELSE
         stp = one
      END IF

      CALL dcopy(n,x,1,t,1)
      CALL dcopy(n,g,1,r,1)
      fold = f
      ifun = 0
      iback = 0
      csave = 'START'
      ENDIF
      gd = ddot(n,g,1,d,1)
      IF (ifun == 0) THEN
         gdold=gd
         IF (gd >= zero) THEN
!                               the directional derivative >=0.
!                               Line search is impossible.
            WRITE(*,*)' ascent direction in projection gd = ', gd
            info = -4
            RETURN
         END IF
      END IF

      CALL dcsrch(f,gd,stp,ftol,gtol,xtol,zero,step_max,csave,isave,dsave)

      xstep = stp*dnorm
      IF (csave(1:4) /= 'CONV' .AND. csave(1:4) /= 'WARN') THEN
         task = 'FG_LNSRCH'
         ifun = ifun + 1
         nfgv = nfgv + 1
         iback = ifun - 1
         IF (stp == one) THEN
            CALL dcopy(n,z,1,x,1)
         ELSE
            DO i = 1, n
               x(i) = stp*d(i) + t(i)
            END DO
         END IF
      ELSE
         task = 'NEW_X'
      END IF

      RETURN

      END SUBROUTINE lnsrlb

! **************************************************************************************************
!> \brief        This subroutine updates matrices WS and WY, and forms the middle matrix in B.
!> \param n ...
!> \param m ...
!> \param ws ...
!> \param wy ...
!> \param sy ...
!> \param ss ...
!> \param d ...
!> \param r ...
!> \param itail ...
!> \param iupdat ...
!> \param col ...
!> \param head ...
!> \param theta ...
!> \param rr ...
!> \param dr ...
!> \param stp ...
!> \param dtd ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE matupd(n, m, ws, wy, sy, ss, d, r, itail,              &
                        iupdat, col, head, theta, rr, dr, stp, dtd)

      INTEGER                                            :: n, m
      REAL(KIND=dp)                                      :: ws(n, m), wy(n, m), sy(m, m), ss(m, m), &
                                                            d(n), r(n)
      INTEGER                                            :: itail, iupdat, col, head
      REAL(KIND=dp)                                      :: theta, rr, dr, stp, dtd

      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp

      INTEGER                                            :: j, pointr
      REAL(KIND=dp)                                      :: ddot

!     ************
!     Set pointers for matrices WS and WY.

      IF (iupdat <= m) THEN
         col = iupdat
         itail = MOD(head+iupdat-2,m) + 1
      ELSE
         itail = MOD(itail,m) + 1
         head = MOD(head,m) + 1
      END IF

!     Update matrices WS and WY.

      CALL dcopy(n,d,1,ws(1,itail),1)
      CALL dcopy(n,r,1,wy(1,itail),1)

!     Set theta=yy/ys.

      theta = rr/dr

!     Form the middle matrix in B.

!        update the upper triangle of SS,
!                                         and the lower triangle of SY:
      IF (iupdat > m) THEN
!                              move old information
         DO j = 1, col - 1
            CALL dcopy(j,ss(2,j+1),1,ss(1,j),1)
            CALL dcopy(col-j,sy(j+1,j+1),1,sy(j,j),1)
         END DO
      END IF
!        add new information: the last row of SY
!                                             and the last column of SS:
      pointr = head
      DO j = 1, col - 1
         sy(col,j) = ddot(n,d,1,wy(1,pointr),1)
         ss(j,col) = ddot(n,ws(1,pointr),1,d,1)
         pointr = MOD(pointr,m) + 1
      END DO
      IF (stp == one) THEN
         ss(col,col) = dtd
      ELSE
         ss(col,col) = stp*stp*dtd
      END IF
      sy(col,col) = dr

      RETURN

      END SUBROUTINE matupd

! **************************************************************************************************
!> \brief        This subroutine prints the input data, initial point, upper and
!>               lower bounds of each variable, machine precision, as well as
!>               the headings of the output.
!>
!> \param n ...
!> \param m ...
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param x ...
!> \param iprint ...
!> \param itfile ...
!> \param epsmch ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE prn1lb(n, m, lower_bound, upper_bound, x, iprint, itfile, epsmch)

      INTEGER, INTENT(in)                                :: n, m
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n), x(n)
      INTEGER                                            :: iprint, itfile
      REAL(KIND=dp)                                      :: epsmch

      INTEGER                                            :: i

      IF (iprint >= 0) THEN
         WRITE (*,7001) epsmch
         WRITE (*,*) 'N = ',n,'    M = ',m
         IF (iprint >= 1) THEN
            WRITE (itfile,2001) epsmch
            WRITE (itfile,*)'N = ',n,'    M = ',m
            WRITE (itfile,9001)
            IF (iprint > 100) THEN
               WRITE (*,1004) 'L =',(lower_bound(i),i = 1,n)
               WRITE (*,1004) 'X0 =',(x(i),i = 1,n)
               WRITE (*,1004) 'U =',(upper_bound(i),i = 1,n)
            END IF
         END IF
      END IF

 1004 FORMAT (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 2001 FORMAT ('RUNNING THE L-BFGS-B CODE',/,/,                          &
     & 'it    = iteration number',/,                                    &
     & 'nf    = number of function evaluations',/,                      &
     & 'nseg  = number of segments explored during the Cauchy search',/,&
     & 'nact  = number of active bounds at the generalized Cauchy point'&
     & ,/,                                                              &
     & 'sub   = manner in which the subspace minimization terminated:'  &
     & ,/,'        con = converged, bnd = a bound was reached',/,       &
     & 'itls  = number of iterations performed in the line search',/,   &
     & 'stepl = step length used',/,                                    &
     & 'tstep = norm of the displacement (total step)',/,               &
     & 'projg = norm of the projected gradient',/,                      &
     & 'f     = function value',/,/,                                    &
     & '           * * *',/,/,                                          &
     & 'Machine precision =',1p,d10.3)
 7001 FORMAT ('RUNNING THE L-BFGS-B CODE',/,/,                          &
     & '           * * *',/,/,                                          &
     & 'Machine precision =',1p,d10.3)
 9001 FORMAT (/,3x,'it',3x,'nf',2x,'nseg',2x,'nact',2x,'sub',2x,'itls', &
              2x,'stepl',4x,'tstep',5x,'projg',8x,'f')

      RETURN

      END SUBROUTINE prn1lb

! **************************************************************************************************
!> \brief        This subroutine prints out new information after a successful line search.
!> \param n ...
!> \param x ...
!> \param f ...
!> \param g ...
!> \param iprint ...
!> \param itfile ...
!> \param iter ...
!> \param nfgv ...
!> \param nact ...
!> \param g_inf_norm ...
!> \param nseg ...
!> \param word ...
!> \param iword ...
!> \param iback ...
!> \param stp ...
!> \param xstep ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE prn2lb(n, x, f, g, iprint, itfile, iter, nfgv, nact,   &
                        g_inf_norm, nseg, word, iword, iback, stp, xstep)

      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: x(n), f, g(n)
      INTEGER, INTENT(in)                                :: iprint, itfile, iter, nfgv, nact
      REAL(KIND=dp), INTENT(in)                          :: g_inf_norm
      INTEGER, INTENT(in)                                :: nseg
      CHARACTER(LEN=3)                                   :: word
      INTEGER                                            :: iword, iback
      REAL(KIND=dp)                                      :: stp, xstep

      INTEGER                                            :: i, imod

!           'word' records the status of subspace solutions.

      IF (iword == 0) THEN
!                            the subspace minimization converged.
         word = 'con'
      ELSE IF (iword == 1) THEN
!                          the subspace minimization stopped at a bound.
         word = 'bnd'
      ELSE IF (iword == 5) THEN
!                             the truncated Newton step has been used.
         word = 'TNT'
      ELSE
         word = '---'
      END IF
      IF (iprint >= 99) THEN
         WRITE (*,*) 'LINE SEARCH',iback,' times; norm of step = ',xstep
         WRITE (*,2001) iter,f,g_inf_norm
         IF (iprint > 100) THEN
            WRITE (*,1004) 'X =',(x(i), i = 1, n)
            WRITE (*,1004) 'G =',(g(i), i = 1, n)
         END IF
      ELSE IF (iprint > 0) THEN
         imod = MOD(iter,iprint)
         IF (imod == 0) WRITE (*,2001) iter,f,g_inf_norm
      END IF
      IF (iprint >= 1) WRITE (itfile,3001)                            &
                iter,nfgv,nseg,nact,word,iback,stp,xstep,g_inf_norm,f

 1004 FORMAT (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 2001 FORMAT                                                            &
     &  (/,'At iterate',i5,4x,'f= ',1p,d12.5,4x,'|proj g|= ',1p,d12.5)
 3001 FORMAT(2(1x,i4),2(1x,i5),2x,a3,1x,i4,1p,2(2x,d7.1),1p,2(1x,d10.3))

      RETURN

      END SUBROUTINE prn2lb

! **************************************************************************************************
!> \brief        This subroutine prints out information when either a built-in
!>               convergence test is satisfied or when an error message is
!>               generated.
!> \param n ...
!> \param x ...
!> \param f ...
!> \param task ...
!> \param iprint ...
!> \param info ...
!> \param itfile ...
!> \param iter ...
!> \param nfgv ...
!> \param nintol ...
!> \param nskip ...
!> \param nact ...
!> \param g_inf_norm ...
!> \param time ...
!> \param nseg ...
!> \param word ...
!> \param iback ...
!> \param stp ...
!> \param xstep ...
!> \param k ...
!> \param cachyt ...
!> \param sbtime ...
!> \param lnscht ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE prn3lb(n, x, f, task, iprint, info, itfile,            &
                        iter, nfgv, nintol, nskip, nact, g_inf_norm,        &
                        time, nseg, word, iback, stp, xstep, k,         &
                        cachyt, sbtime, lnscht)

      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: x(n), f
      CHARACTER(LEN=60), INTENT(in)                      :: task
      INTEGER, INTENT(in)                                :: iprint, info, itfile, iter, nfgv, &
                                                            nintol, nskip, nact
      REAL(KIND=dp), INTENT(in)                          :: g_inf_norm, time
      INTEGER, INTENT(in)                                :: nseg
      CHARACTER(LEN=3)                                   :: word
      INTEGER                                            :: iback
      REAL(KIND=dp)                                      :: stp, xstep
      INTEGER                                            :: k
      REAL(KIND=dp)                                      :: cachyt, sbtime, lnscht

      INTEGER                                            :: i

      IF (iprint >= 0.AND..NOT.(task(1:5) == 'ERROR')) THEN
         WRITE(*,3003)
         WRITE(*,3004)
         WRITE(*,3005) n,iter,nfgv,nintol,nskip,nact,g_inf_norm,f
         IF (iprint >= 100) THEN
            WRITE (*,1004) 'X =',(x(i),i = 1,n)
         END IF
         IF (iprint >= 1) WRITE (*,*) ' F =',f
      END IF
      IF (iprint >= 0) THEN
         WRITE (*,3009) task
         IF (info /= 0) THEN
            IF (info == -1) WRITE (*,9011)
            IF (info == -2) WRITE (*,9012)
            IF (info == -3) WRITE (*,9013)
            IF (info == -4) WRITE (*,9014)
            IF (info == -5) WRITE (*,9015)
            IF (info == -6) WRITE (*,*)' Input nbd(',k,') is invalid.'
            IF (info == -7)                                           &
            WRITE (*,*)' l(',k,') > u(',k,').  No feasible solution.'
            IF (info == -8) WRITE (*,9018)
            IF (info == -9) WRITE (*,9019)
         END IF
         IF (iprint >= 1) WRITE (*,3007) cachyt,sbtime,lnscht
         WRITE (*,3008) time
         IF (iprint >= 1) THEN
            IF (info == -4 .OR. info == -9) THEN
               WRITE (itfile,3002)                                      &
                   iter,nfgv,nseg,nact,word,iback,stp,xstep
            END IF
            WRITE (itfile,3009) task
            IF (info /= 0) THEN
               IF (info == -1) WRITE (itfile,9011)
               IF (info == -2) WRITE (itfile,9012)
               IF (info == -3) WRITE (itfile,9013)
               IF (info == -4) WRITE (itfile,9014)
               IF (info == -5) WRITE (itfile,9015)
               IF (info == -8) WRITE (itfile,9018)
               IF (info == -9) WRITE (itfile,9019)
            END IF
            WRITE (itfile,3008) time
         END IF
      END IF

 1004 FORMAT (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 3002 FORMAT(2(1x,i4),2(1x,i5),2x,a3,1x,i4,1p,2(2x,d7.1),6x,'-',10x,'-')
 3003 FORMAT (/,                                                        &
     & '           * * *',/,/,                                          &
     & 'Tit   = total number of iterations',/,                          &
     & 'Tnf   = total number of function evaluations',/,                &
     & 'Tnint = total number of segments explored during',              &
     &           ' Cauchy searches',/,                                  &
     & 'Skip  = number of BFGS updates skipped',/,                      &
     & 'Nact  = number of active bounds at final generalized',          &
     &          ' Cauchy point',/,                                      &
     & 'Projg = norm of the final projected gradient',/,                &
     & 'F     = final function value',/,/,                              &
     & '           * * *')
 3004 FORMAT (/,3x,'N',4x,'Tit',5x,'Tnf',2x,'Tnint',2x,                 &
     &       'Skip',2x,'Nact',5x,'Projg',8x,'F')
 3005 FORMAT (i5,2(1x,i6),(1x,i6),(2x,i4),(1x,i5),1p,2(2x,d10.3))
 3007 FORMAT (/,' Cauchy                time',1p,e10.3,' seconds.',/    &
     &        ' Subspace minimization time',1p,e10.3,' seconds.',/      &
     &        ' Line search           time',1p,e10.3,' seconds.')
 3008 FORMAT (/,' Total User time',1p,e10.3,' seconds.',/)
 3009 FORMAT (/,a60)
 9011 FORMAT (/,                                                        &
     &' Matrix in 1st Cholesky factorization in formk is not Pos. Def.')
 9012 FORMAT (/,                                                        &
     &' Matrix in 2st Cholesky factorization in formk is not Pos. Def.')
 9013 FORMAT (/,                                                        &
     &' Matrix in the Cholesky factorization in formt is not Pos. Def.')
 9014 FORMAT (/,                                                        &
     &' Derivative >= 0, backtracking line search impossible.',/,       &
     &'   Previous x, f and g restored.',/,                             &
     &' Possible causes: 1 error in function or gradient evaluation;',/,&
     &'                  2 rounding errors dominate computation.')
 9015 FORMAT (/,                                                        &
     &' Warning:  more than 10 function and gradient',/,                &
     &'   evaluations in the last line search.  Termination',/,         &
     &'   may possibly be caused by a bad search direction.')
 9018 FORMAT (/,' The triangular system is singular.')
 9019 FORMAT (/,                                                        &
     &' Line search cannot locate an adequate point after 20 function',/&
     &,'  and gradient evaluations.  Previous x, f and g restored.',/,  &
     &' Possible causes: 1 error in function or gradient evaluation;',/,&
     &'                  2 rounding error dominate computation.')

      RETURN

      END SUBROUTINE prn3lb

! **************************************************************************************************
!> \brief        This subroutine computes the infinity norm of the projected  gradient.
!> \param n ...
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd ...
!> \param x ...
!> \param g ...
!> \param g_inf_norm ...
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE projgr(n, lower_bound, upper_bound, nbd, x, g, g_inf_norm)

      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n)
      INTEGER, INTENT(in)                                :: nbd(n)
      REAL(KIND=dp), INTENT(in)                          :: x(n), g(n)
      REAL(KIND=dp)                                      :: g_inf_norm

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: gi

      g_inf_norm=zero
      DO i = 1, n
        gi = g(i)
        IF (nbd(i) /= 0) THEN
           IF (gi < zero) THEN
              IF (nbd(i) >= 2) gi = MAX((x(i)-upper_bound(i)),gi)
           ELSE
              IF (nbd(i) <= 2) gi = MIN((x(i)-lower_bound(i)),gi)
           END IF
        END IF
        g_inf_norm = MAX(g_inf_norm,ABS(gi))
      END DO

      RETURN

      END SUBROUTINE projgr

! **************************************************************************************************
!> \brief        This routine contains the major changes in the updated version.
!>               The changes are described in the accompanying paper
!>
!>               Jose Luis Morales, Jorge Nocedal
!>               "Remark On Algorithm 788: L-BFGS-B: Fortran Subroutines for Large
!>               Bound Constrained Optimization". Decemmber 27, 2010.
!>
!>               J.L. Morales  Departamento de Matematicas,
!>                             Instituto Tecnologico Autonomo de Mexico
!>                             Mexico D.F.
!>
!>               J, Nocedal    Department of Electrical Engineering and
!>                            Computer Science.
!>                             Northwestern University. Evanston, IL. USA
!>
!>                             January 17, 2011
!>
!>      *****************************************************************
!>
!>               Given xcp, l, u, r, an index set that specifies
!>               the active set at xcp, and an l-BFGS matrix B
!>               (in terms of WY, WS, SY, WT, head, col, and theta),
!>               this subroutine computes an approximate solution
!>               of the subspace problem
!>
!>               (P)   min Q(x) = r'(x-xcp) + 1/2 (x-xcp)' B (x-xcp)
!>
!>               subject to l<=x<=u
!>                       x_i=xcp_i for all i in A(xcp)
!>
!>               along the subspace unconstrained Newton direction
!>
!>               d = -(Z'BZ)^(-1) r.
!>
!>               The formula for the Newton direction, given the L-BFGS matrix
!>               and the Sherman-Morrison formula, is
!>
!>               d = (1/theta)r + (1/theta*2) Z'WK^(-1)W'Z r.
!>
!>               where
!>                 K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!>                     [L_a -R_z           theta*S'AA'S ]
!>
!>               Note that this procedure for computing d differs
!>               from that described in [1]. One can show that the matrix K is
!>               equal to the matrix M^[-1]N in that paper.
!> \param n      n is the dimension of the problem.
!> \param m      m is the maximum number of variable metric corrections
!>               used to define the limited memory matrix.
!> \param nsub   nsub is the number of free variables.
!> \param ind    ind specifies the coordinate indices of free variables.
!> \param lower_bound  the lower bound on x.
!> \param upper_bound  the upper bound on x.
!> \param nbd    nbd represents the type of bounds imposed on the
!>               variables, and must be specified as follows:
!>               nbd(i)=0 if x(i) is unbounded,
!>                      1 if x(i) has only a lower bound,
!>                      2 if x(i) has both lower and upper bounds, and
!>                      3 if x(i) has only an upper bound.
!> \param x      x is a double precision array of dimension n.
!>               On entry x specifies the Cauchy point xcp.
!>               On exit x(i) is the minimizer of Q over the subspace of free variables.
!> \param d      On entry d is the reduced gradient of Q at xcp.
!>               On exit d is the Newton direction of Q.
!> \param xp     xp is a double precision array of dimension n.
!>               used to safeguard the projected Newton direction
!> \param ws     ws and wy are double precision arrays;
!>               On entry they store the information defining the limited memory BFGS matrix:
!>               ws(n,m) stores S, a set of s-vectors;
!> \param wy     wy(n,m) stores Y, a set of y-vectors;
!> \param theta  theta is the scaling factor specifying B_0 = theta I;
!> \param xx     xx holds the current iterate
!> \param gg     gg holds the gradient at the current iterate
!> \param col    is the number of variable metric corrections stored;
!> \param head   head is the location of the 1st s- (or y-) vector in S (or Y).
!> \param iword  iword specifies the status of the subspace solution.
!>               iword = 0 if the solution is in the box,
!>                       1 if some bound is encountered.
!> \param wv     wv is a working array
!> \param wn     the upper triangle of wn stores the LEL^T factorization
!>               of the indefinite matrix
!>
!>               K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
!>                   [L_a -R_z           theta*S'AA'S ]
!>               where E = [-I  0]
!>                         [ 0  I]
!> \param iprint iprint is an INTEGER variable that must be set by the user.
!>               It controls the frequency and type of output generated:
!>               iprint<0    no output is generated;
!>               iprint=0    print only one line at the last iteration;
!>               0<iprint<99 print also f and |proj g| every iprint iterations;
!>               iprint=99   print details of every iteration except n-vectors;
!>               iprint=100  print also the changes of active set and final x;
!>               iprint>100  print details of every iteration including x and g;
!>               When iprint > 0, the file iterate.dat will be created to summarize the iteration.
!> \param info   info = 0       for normal return,
!>                    = nonzero for abnormal return when the matrix K is ill-conditioned.
!> \author       NEOS, November 1994. (Latest revision June 1996.)
!>               Optimization Technology Center.
!>               Argonne National Laboratory and Northwestern University.
!>               Written by
!>                           Ciyou Zhu
!>               in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
! **************************************************************************************************
      SUBROUTINE subsm ( n, m, nsub, ind, lower_bound, upper_bound, nbd, x, d, xp, ws, wy,  &
                         theta, xx, gg,                                 &
                         col, head, iword, wv, wn, iprint, info )
      INTEGER, INTENT(in)                                :: n, m, nsub, ind(nsub)
      REAL(KIND=dp), INTENT(in)                          :: lower_bound(n), upper_bound(n)
      INTEGER, INTENT(in)                                :: nbd(n)
      REAL(KIND=dp), INTENT(inout)                       :: x(n), d(n)
      REAL(KIND=dp)                                      :: xp(n)
      REAL(KIND=dp), INTENT(in)                          :: ws(n, m), wy(n, m), theta, xx(n), gg(n)
      INTEGER, INTENT(in)                                :: col, head
      INTEGER, INTENT(out)                               :: iword
      REAL(KIND=dp)                                      :: wv(2*m)
      REAL(KIND=dp), INTENT(in)                          :: wn(2*m, 2*m)
      INTEGER                                            :: iprint
      INTEGER, INTENT(out)                               :: info

      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: col2, i, ibd, j, js, jy, k, m2, pointr
      REAL(KIND=dp)                                      :: alpha, dd_p, dk, temp1, temp2, xk

!     References:
!
!       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
!       memory algorithm for bound constrained optimization'',
!       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
!
!
!
!                           *  *  *
!

      IF (nsub <= 0) RETURN
      IF (iprint >= 99) WRITE (*,1001)

!     Compute wv = W'Zd.

      pointr = head
      DO i = 1, col
         temp1 = zero
         temp2 = zero
         DO j = 1, nsub
            k = ind(j)
            temp1 = temp1 + wy(k,pointr)*d(j)
            temp2 = temp2 + ws(k,pointr)*d(j)
         END DO
         wv(i) = temp1
         wv(col + i) = theta*temp2
         pointr = MOD(pointr,m) + 1
      END DO

!     Compute wv:=K^(-1)wv.

      m2 = 2*m
      col2 = 2*col
      CALL dtrsl(wn,m2,col2,wv,11,info)
      IF (info /= 0) RETURN
      DO i = 1, col
         wv(i) = -wv(i)
      END DO
      CALL dtrsl(wn,m2,col2,wv,01,info)
      IF (info /= 0) RETURN

!     Compute d = (1/theta)d + (1/theta**2)Z'W wv.

      pointr = head
      DO jy = 1, col
         js = col + jy
         DO i = 1, nsub
            k = ind(i)
            d(i) = d(i) + wy(k,pointr)*wv(jy)/theta                     &
     &                  + ws(k,pointr)*wv(js)
         END DO
         pointr = MOD(pointr,m) + 1
      END DO

      CALL dscal( nsub, one/theta, d, 1 )
!
!-----------------------------------------------------------------
!     Let us try the projection, d is the Newton direction

      iword = 0

      CALL dcopy ( n, x, 1, xp, 1 )
!
      DO i=1, nsub
         k  = ind(i)
         dk = d(i)
         xk = x(k)
         IF ( nbd(k) /= 0 ) THEN
!
                                             ! lower bounds only
            IF ( nbd(k).eq.1 ) THEN
               x(k) = MAX( lower_bound(k), xk + dk )
               IF ( x(k).eq.lower_bound(k) ) iword = 1
            ELSE
!
                                             ! upper and lower bounds
               IF ( nbd(k).eq.2 ) THEN
                  xk   = MAX( lower_bound(k), xk + dk )
                  x(k) = MIN( upper_bound(k), xk )
                  IF ( x(k).eq.lower_bound(k) .OR. x(k).eq.upper_bound(k) ) iword = 1
               ELSE
!
                                             ! upper bounds only
                  IF ( nbd(k).eq.3 ) THEN
                     x(k) = MIN( upper_bound(k), xk + dk )
                     IF ( x(k).eq.upper_bound(k) ) iword = 1
                  END IF
               END IF
            END IF
!
                                             ! free variables
         ELSE
            x(k) = xk + dk
         END IF
      END DO
!
      IF ( .NOT.(iword.eq.0) ) THEN
!
!     check sign of the directional derivative
!
      dd_p = zero
      DO i=1, n
         dd_p  = dd_p + (x(i) - xx(i))*gg(i)
      END DO
      IF ( dd_p .gt.zero ) THEN
         CALL dcopy( n, xp, 1, x, 1 )
         IF(iprint >0) WRITE(*,*) ' Positive dir derivative in projection '
         IF(iprint >0) WRITE(*,*) ' Using the backtracking step '
         alpha = one
         temp1 = alpha
         ibd   = 0
         DO i = 1, nsub
            k = ind(i)
            dk = d(i)
            IF (nbd(k) /= 0) THEN
               IF (dk < zero .AND. nbd(k) <= 2) THEN
                  temp2 = lower_bound(k) - x(k)
                  IF (temp2 >= zero) THEN
                     temp1 = zero
                  ELSE IF (dk*alpha < temp2) THEN
                     temp1 = temp2/dk
                  END IF
               ELSE IF (dk > zero .AND. nbd(k) >= 2) THEN
                  temp2 = upper_bound(k) - x(k)
                  IF (temp2 <= zero) THEN
                     temp1 = zero
                  ELSE IF (dk*alpha > temp2) THEN
                     temp1 = temp2/dk
                  END IF
               END IF
               IF (temp1 < alpha) THEN
                  alpha = temp1
                  ibd = i
               END IF
            END IF
         END DO

         IF (alpha < one) THEN
            dk = d(ibd)
            k = ind(ibd)
            IF (dk > zero) THEN
               x(k) = upper_bound(k)
               d(ibd) = zero
            ELSE IF (dk < zero) THEN
               x(k) = lower_bound(k)
               d(ibd) = zero
            END IF
         END IF
         DO i = 1, nsub
            k    = ind(i)
            x(k) = x(k) + alpha*d(i)
         END DO
      END IF
      END IF

      IF (iprint >= 99) WRITE (*,1004)

 1001 FORMAT (/,'----------------SUBSM entered-----------------',/)
 1004 FORMAT (/,'----------------exit SUBSM --------------------',/)

      RETURN

      END SUBROUTINE subsm

! **************************************************************************************************
!> \brief         This subroutine finds a step that satisfies a sufficient
!>                decrease condition and a curvature condition.
!>
!>                Each call of the subroutine updates an interval with
!>                endpoints stx and sty. The interval is initially chosen
!>                so that it contains a minimizer of the modified function
!>
!>                      psi(stp) = f(stp) - f(0) - ftol*stp*f'(0).
!>
!>                If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
!>                interval is chosen so that it contains a minimizer of f.
!>
!>                The algorithm is designed to find a step that satisfies
!>                the sufficient decrease condition
!>
!>                f(stp) <= f(0) + ftol*stp*f'(0),
!>
!>                and the curvature condition
!>
!>                abs(f'(stp)) <= gtol*abs(f'(0)).
!>
!>                If ftol is less than gtol and if, for example, the function
!>                is bounded below, then there is always a step which satisfies
!>                both conditions.
!>
!>                If no step can be found that satisfies both conditions, then
!>                the algorithm stops with a warning. In this case stp only
!>                satisfies the sufficient decrease condition.
!>
!>                A typical invocation of dcsrch has the following outline:
!>
!>                task = 'START'
!>                DO WHILE (.TRUE.)
!>                   call dcsrch( ... )
!>                   if (task .eq. 'FG') then
!>                      Evaluate the function and the gradient at stp
!>                   else
!>                      exit
!>                   end if
!>                END DO
!> \param f       On initial entry f is the value of the function at 0.
!>                On subsequent entries f is the value of the
!>                function at stp.
!>                On exit f is the value of the function at stp.
!> \param g       On initial entry g is the derivative of the function at 0.
!>                On subsequent entries g is the derivative of the
!>                function at stp.
!>                On exit g is the derivative of the function at stp.
!> \param stp     On entry stp is the current estimate of a satisfactory
!>                step. On initial entry, a positive initial estimate
!>                must be provided.
!>                On exit stp is the current estimate of a satisfactory step
!>                if task = 'FG'. If task = 'CONV' then stp satisfies
!>                the sufficient decrease and curvature condition.
!> \param ftol    ftol specifies a nonnegative tolerance for the
!>                sufficient decrease condition.
!> \param gtol    gtol specifies a nonnegative tolerance for the
!>                curvature condition.
!> \param xtol    xtol specifies a nonnegative relative tolerance
!>                for an acceptable step. The subroutine exits with a
!>                warning if the relative difference between sty and stx
!>                is less than xtol.
!> \param stpmin  stpmin is a nonnegative lower bound for the step.
!> \param stpmax  stpmax is a nonnegative upper bound for the step.
!> \param task    task is a character variable of length at least 60.
!>                On initial entry task must be set to 'START'.
!>                On exit task indicates the required action:
!>
!>                If task(1:2) = 'FG' then evaluate the function and
!>                derivative at stp and call dcsrch again.
!>
!>                If task(1:4) = 'CONV' then the search is successful.
!>
!>                If task(1:4) = 'WARN' then the subroutine is not able
!>                to satisfy the convergence conditions. The exit value of
!>                stp contains the best point found during the search.
!>
!>                If task(1:5) = 'ERROR' then there is an error in the
!>                input arguments.
!>
!>                On exit with convergence, a warning or an error, the
!>                variable task contains additional information.
!> \param isave   is work array
!> \param dsave   is a work array
! **************************************************************************************************
      SUBROUTINE dcsrch(f,g,stp,ftol,gtol,xtol,stpmin,stpmax,           &
                        task,isave,dsave)
      REAL(KIND=dp)                                      :: f, g
      REAL(KIND=dp), INTENT(inout)                       :: stp
      REAL(KIND=dp)                                      :: ftol, gtol, xtol, stpmin, stpmax
      CHARACTER(LEN=*)                                   :: task
      INTEGER                                            :: isave(2)
      REAL(KIND=dp)                                      :: dsave(13)

      REAL(KIND=dp), PARAMETER                           :: p5 = 0.5_dp, p66 = 0.66_dp, &
                                                            xtrapl = 1.1_dp, xtrapu = 4.0_dp, &
                                                            zero = 0.0_dp

      INTEGER                                            :: stage
      LOGICAL                                            :: brackt
      REAL(KIND=dp)                                      :: finit, fm, ftest, fx, fxm, fy, fym, &
                                                            ginit, gm, gtest, gx, gxm, gy, gym, &
                                                            stmax, stmin, stx, sty, width, width1

!
!     NOTE: The user must no alter work arrays between calls.
!
!
!     MINPACK-1 Project. June 1983.
!     Argonne National Laboratory.
!     Jorge J. More' and David J. Thuente.
!
!     MINPACK-2 Project. October 1993.
!     Argonne National Laboratory and University of Minnesota.
!     Brett M. Averick, Richard G. Carter, and Jorge J. More'.
!
!     **********
!     Initialization block.

      IF (task(1:5) == 'START') THEN

!        Check the input arguments for errors.

         IF (stp < stpmin) task = 'ERROR: STP < STPMIN'
         IF (stp > stpmax) task = 'ERROR: STP > STPMAX'
         IF (g >= zero) task = 'ERROR: INITIAL G >= ZERO'
         IF (ftol < zero) task = 'ERROR: FTOL < ZERO'
         IF (gtol < zero) task = 'ERROR: GTOL < ZERO'
         IF (xtol < zero) task = 'ERROR: XTOL < ZERO'
         IF (stpmin < zero) task = 'ERROR: STPMIN < ZERO'
         IF (stpmax < stpmin) task = 'ERROR: STPMAX < STPMIN'

!        Exit if there are errors on input.

         IF (task(1:5) == 'ERROR') RETURN

!        Initialize local variables.

         brackt = .FALSE.
         stage = 1
         finit = f
         ginit = g
         gtest = ftol*ginit
         width = stpmax - stpmin
         width1 = width/p5

!        The variables stx, fx, gx contain the values of the step,
!        function, and derivative at the best step.
!        The variables sty, fy, gy contain the value of the step,
!        function, and derivative at sty.
!        The variables stp, f, g contain the values of the step,
!        function, and derivative at stp.

         stx = zero
         fx = finit
         gx = ginit
         sty = zero
         fy = finit
         gy = ginit
         stmin = zero
         stmax = stp + xtrapu*stp
         task = 'FG'

      ELSE

!        Restore local variables.

         IF (isave(1) == 1) THEN
            brackt = .TRUE.
         ELSE
            brackt = .FALSE.
         END IF
         stage = isave(2)
         ginit = dsave(1)
         gtest = dsave(2)
         gx = dsave(3)
         gy = dsave(4)
         finit = dsave(5)
         fx = dsave(6)
         fy = dsave(7)
         stx = dsave(8)
         sty = dsave(9)
         stmin = dsave(10)
         stmax = dsave(11)
         width = dsave(12)
         width1 = dsave(13)

!        If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
!        algorithm enters the second stage.

         ftest = finit + stp*gtest
         IF (stage == 1 .AND. f <= ftest .AND. g >= zero)            &
            stage = 2

!        Test for warnings.

         IF (brackt .AND. (stp <= stmin .OR. stp >= stmax))            &
            task = 'WARNING: ROUNDING ERRORS PREVENT PROGRESS'
         IF (brackt .AND. stmax - stmin <= xtol*stmax)                   &
            task = 'WARNING: XTOL TEST SATISFIED'
         IF (stp == stpmax .AND. f <= ftest .AND. g <= gtest)        &
            task = 'WARNING: STP = STPMAX'
         IF (stp == stpmin .AND. (f > ftest .OR. g >= gtest))       &
            task = 'WARNING: STP = STPMIN'

!        Test for convergence.

         IF (f <= ftest .AND. ABS(g) <= gtol*(-ginit))                 &
            task = 'CONVERGENCE'

!        Test for termination.



         IF (.NOT.(task(1:4) == 'WARN' .OR. task(1:4) == 'CONV')) THEN

!        A modified function is used to predict the step during the
!        first stage if a lower function value has been obtained but
!        the decrease is not sufficient.

           IF (stage == 1 .AND. f <= fx .AND. f > ftest) THEN

!             Define the modified function and derivative values.

              fm = f - stp*gtest
              fxm = fx - stx*gtest
              fym = fy - sty*gtest
              gm = g - gtest
              gxm = gx - gtest
              gym = gy - gtest

!             Call dcstep to update stx, sty, and to compute the new step.

              CALL dcstep(stx,fxm,gxm,sty,fym,gym,stp,fm,gm,                 &
                       brackt,stmin,stmax)

!             Reset the function and derivative values for f.

              fx = fxm + stx*gtest
              fy = fym + sty*gtest
              gx = gxm + gtest
              gy = gym + gtest

           ELSE

!             Call dcstep to update stx, sty, and to compute the new step.

              CALL dcstep(stx,fx,gx,sty,fy,gy,stp,f,g,                        &
                   brackt,stmin,stmax)

           END IF

!          Decide if a bisection step is needed.

           IF (brackt) THEN
              IF (ABS(sty-stx) >= p66*width1) stp = stx + p5*(sty - stx)
              width1 = width
              width = ABS(sty-stx)
           END IF

!          Set the minimum and maximum steps allowed for stp.

           IF (brackt) THEN
              stmin = MIN(stx,sty)
              stmax = MAX(stx,sty)
           ELSE
              stmin = stp + xtrapl*(stp - stx)
              stmax = stp + xtrapu*(stp - stx)
           END IF

!          Force the step to be within the bounds stpmax and stpmin.

           stp = MAX(stp,stpmin)
           stp = MIN(stp,stpmax)

!          If further progress is not possible, let stp be the best
!          point obtained during the search.

           IF (brackt .AND. (stp <= stmin .OR. stp >= stmax)             &
             .OR. (brackt .AND. stmax-stmin <= xtol*stmax)) stp = stx

!          Obtain another function and derivative.

           task = 'FG'

        ENDIF
     ENDIF

!     Save local variables.

      IF (brackt) THEN
         isave(1) = 1
      ELSE
         isave(1) = 0
      END IF
      isave(2) = stage
      dsave(1) =  ginit
      dsave(2) =  gtest
      dsave(3) =  gx
      dsave(4) =  gy
      dsave(5) =  finit
      dsave(6) =  fx
      dsave(7) =  fy
      dsave(8) =  stx
      dsave(9) =  sty
      dsave(10) = stmin
      dsave(11) = stmax
      dsave(12) = width
      dsave(13) = width1

      RETURN
      END SUBROUTINE dcsrch

! **************************************************************************************************
!> \brief          This subroutine computes a safeguarded step for a search
!>                 procedure and updates an interval that contains a step that
!>                 satisfies a sufficient decrease and a curvature condition.
!>
!>                 The parameter stx contains the step with the least function
!>                 value. If brackt is set to .true. then a minimizer has
!>                 been bracketed in an interval with endpoints stx and sty.
!>                 The parameter stp contains the current step.
!>                 The subroutine assumes that if brackt is set to .true. then
!>
!>                    min(stx,sty) < stp < max(stx,sty),
!>
!>                 and that the derivative at stx is negative in the direction
!>                 of the step.
!> \param stx      On entry stx is the best step obtained so far and is an
!>                 endpoint of the interval that contains the minimizer.
!>                 On exit stx is the updated best step.
!> \param fx       fx is the function at stx.
!> \param dx       On entry dx is the derivative of the function at
!>                 stx. The derivative must be negative in the direction of
!>                 the step, that is, dx and stp - stx must have opposite
!>                 signs.
!>                 On exit dx is the derivative of the function at stx.
!> \param sty      On entry sty is the second endpoint of the interval that
!>                 contains the minimizer.
!>                 On exit sty is the updated endpoint of the interval that
!>                 contains the minimizer.
!> \param fy       fy is the function at sty.
!> \param dy       On entry dy is the derivative of the function at sty.
!>                 On exit dy is the derivative of the function at the exit sty.
!> \param stp      On entry stp is the current step. If brackt is set to .true.
!>                 then on input stp must be between stx and sty.
!>                 On exit stp is a new trial step.
!> \param fp       fp is the function at stp
!> \param dp_loc   dp_loc is the the derivative of the function at stp.
!> \param brackt   On entry brackt specifies if a minimizer has been bracketed.
!>                 Initially brackt must be set to .false.
!>                 On exit brackt specifies if a minimizer has been bracketed.
!>                 When a minimizer is bracketed brackt is set to .true.
!> \param stpmin   stpmin is a lower bound for the step.
!> \param stpmax   stpmax is an upper bound for the step.
! **************************************************************************************************
      SUBROUTINE dcstep(stx,fx,dx,sty,fy,dy,stp,fp,dp_loc,brackt,           &
                        stpmin,stpmax)
      REAL(KIND=dp), INTENT(inout)                       :: stx, fx, dx, sty, fy, dy, stp
      REAL(KIND=dp), INTENT(in)                          :: fp, dp_loc
      LOGICAL, INTENT(inout)                             :: brackt
      REAL(KIND=dp), INTENT(in)                          :: stpmin, stpmax

      REAL(KIND=dp), PARAMETER                           :: p66 = 0.66_dp, three = 3.0_dp, &
                                                            two = 2.0_dp, zero = 0.0_dp

      REAL(KIND=dp)                                      :: gamma, p, q, r, s, sgnd, stpc, stpf, &
                                                            stpq, theta

!
!     MINPACK-1 Project. June 1983
!     Argonne National Laboratory.
!     Jorge J. More' and David J. Thuente.
!
!     MINPACK-2 Project. October 1993.
!     Argonne National Laboratory and University of Minnesota.
!     Brett M. Averick and Jorge J. More'.
!
!     **********

      sgnd = dp_loc*SIGN(1.0_dp,dx)

!     First case: A higher function value. The minimum is bracketed.
!     If the cubic step is closer to stx than the quadratic step, the
!     cubic step is taken, otherwise the average of the cubic and
!     quadratic steps is taken.

      IF (fp > fx) THEN
         theta = three*(fx - fp)/(stp - stx) + dx + dp_loc
         s = MAX(ABS(theta),ABS(dx),ABS(dp_loc))
         gamma = s*SQRT((theta/s)**2 - (dx/s)*(dp_loc/s))
         IF (stp < stx) gamma = -gamma
         p = (gamma - dx) + theta
         q = ((gamma - dx) + gamma) + dp_loc
         r = p/q
         stpc = stx + r*(stp - stx)
         stpq = stx + ((dx/((fx - fp)/(stp - stx) + dx))/two)*          &
     &                                                       (stp - stx)
         IF (ABS(stpc-stx) < ABS(stpq-stx)) THEN
            stpf = stpc
         ELSE
            stpf = stpc + (stpq - stpc)/two
         END IF
         brackt = .TRUE.

!     Second case: A lower function value and derivatives of opposite
!     sign. The minimum is bracketed. If the cubic step is farther from
!     stp than the secant step, the cubic step is taken, otherwise the
!     secant step is taken.

      ELSE IF (sgnd < zero) THEN
         theta = three*(fx - fp)/(stp - stx) + dx + dp_loc
         s = MAX(ABS(theta),ABS(dx),ABS(dp_loc))
         gamma = s*SQRT((theta/s)**2 - (dx/s)*(dp_loc/s))
         IF (stp > stx) gamma = -gamma
         p = (gamma - dp_loc) + theta
         q = ((gamma - dp_loc) + gamma) + dx
         r = p/q
         stpc = stp + r*(stx - stp)
         stpq = stp + (dp_loc/(dp_loc - dx))*(stx - stp)
         IF (ABS(stpc-stp) > ABS(stpq-stp)) THEN
            stpf = stpc
         ELSE
            stpf = stpq
         END IF
         brackt = .TRUE.

!     Third case: A lower function value, derivatives of the same sign,
!     and the magnitude of the derivative decreases.

      ELSE IF (ABS(dp_loc) < ABS(dx)) THEN

!        The cubic step is computed only if the cubic tends to infinity
!        in the direction of the step or if the minimum of the cubic
!        is beyond stp. Otherwise the cubic step is defined to be the
!        secant step.

         theta = three*(fx - fp)/(stp - stx) + dx + dp_loc
         s = MAX(ABS(theta),ABS(dx),ABS(dp_loc))

!        The case gamma = 0 only arises if the cubic does not tend
!        to infinity in the direction of the step.

         gamma = s*SQRT(MAX(zero,(theta/s)**2-(dx/s)*(dp_loc/s)))
         IF (stp > stx) gamma = -gamma
         p = (gamma - dp_loc) + theta
         q = (gamma + (dx - dp_loc)) + gamma
         r = p/q
         IF (r < zero .AND. gamma /= zero) THEN
            stpc = stp + r*(stx - stp)
         ELSE IF (stp > stx) THEN
            stpc = stpmax
         ELSE
            stpc = stpmin
         END IF
         stpq = stp + (dp_loc/(dp_loc - dx))*(stx - stp)

         IF (brackt) THEN

!           A minimizer has been bracketed. If the cubic step is
!           closer to stp than the secant step, the cubic step is
!           taken, otherwise the secant step is taken.

            IF (ABS(stpc-stp) < ABS(stpq-stp)) THEN
               stpf = stpc
            ELSE
               stpf = stpq
            END IF
            IF (stp > stx) THEN
               stpf = MIN(stp+p66*(sty-stp),stpf)
            ELSE
               stpf = MAX(stp+p66*(sty-stp),stpf)
            END IF
         ELSE

!           A minimizer has not been bracketed. If the cubic step is
!           farther from stp than the secant step, the cubic step is
!           taken, otherwise the secant step is taken.

            IF (ABS(stpc-stp) > ABS(stpq-stp)) THEN
               stpf = stpc
            ELSE
               stpf = stpq
            END IF
            stpf = MIN(stpmax,stpf)
            stpf = MAX(stpmin,stpf)
         END IF

!     Fourth case: A lower function value, derivatives of the same sign,
!     and the magnitude of the derivative does not decrease. If the
!     minimum is not bracketed, the step is either stpmin or stpmax,
!     otherwise the cubic step is taken.

      ELSE
         IF (brackt) THEN
            theta = three*(fp - fy)/(sty - stp) + dy + dp_loc
            s = MAX(ABS(theta),ABS(dy),ABS(dp_loc))
            gamma = s*SQRT((theta/s)**2 - (dy/s)*(dp_loc/s))
            IF (stp > sty) gamma = -gamma
            p = (gamma - dp_loc) + theta
            q = ((gamma - dp_loc) + gamma) + dy
            r = p/q
            stpc = stp + r*(sty - stp)
            stpf = stpc
         ELSE IF (stp > stx) THEN
            stpf = stpmax
         ELSE
            stpf = stpmin
         END IF
      END IF

!     Update the interval which contains a minimizer.

      IF (fp > fx) THEN
         sty = stp
         fy = fp
         dy = dp_loc
      ELSE
         IF (sgnd < zero) THEN
            sty = stx
            fy = fx
            dy = dx
         END IF
         stx = stp
         fx = fp
         dx = dp_loc
      END IF

!     Compute the new step.

      stp = stpf

      RETURN
      END SUBROUTINE dcstep

!MK LINPACK

! **************************************************************************************************
!> \brief         factors a double precision symmetric positive definite
!>                matrix.
!>
!>                dpofa is usually called by dpoco, but it can be called
!>                directly with a saving in time if  rcond  is not needed.
!>                (time for dpoco) = (1 + 18/n)*(time for dpofa) .
!> \param a       the symmetric matrix to be factored.  only the
!>                diagonal and upper triangle are used.
!>                on return
!>                an upper triangular matrix  r  so that  a = trans(r)*r
!>                where  trans(r)  is the transpose.
!>                the strict lower triangle is unaltered.
!>                if  info .ne. 0 , the factorization is not complete.
!> \param lda     the leading dimension of the array  a .
!> \param n       the order of the matrix  a .
!> \param info    = 0  for normal return.
!>                = k  signals an error condition.  the leading minor
!>                     of order  k  is not positive definite.
! **************************************************************************************************
      SUBROUTINE dpofa(a,lda,n,info)
      INTEGER, INTENT(in)                                :: lda
      REAL(KIND=dp)                                      :: a(lda,*)
      INTEGER, INTENT(in)                                :: n
      INTEGER                                            :: info

      INTEGER                                            :: j, jm1, k
      REAL(KIND=dp)                                      :: ddot, s, t

!
!     linpack.  this version dated 08/14/78 .
!     cleve moler, university of new mexico, argonne national lab.
!
!     begin block with ...exits to 40
!
!

         DO j = 1, n
            info = j
            s = 0.0_dp
            jm1 = j - 1
            IF (.NOT.(jm1 < 1)) THEN
               DO k = 1, jm1
                  t = a(k,j) - ddot(k-1,a(1,k),1,a(1,j),1)
                  t = t/a(k,k)
                  a(k,j) = t
                  s = s + t*t
               END DO
            END IF
            s = a(j,j) - s
!     ......exit
            IF (s <= 0.0_dp) EXIT
            a(j,j) = SQRT(s)
            info=0
         END DO
      RETURN
      END SUBROUTINE dpofa

! **************************************************************************************************
!> \brief           dtrsl solves systems of the form
!>
!>                  t * x = b
!>                  or
!>                  trans(t) * x = b
!>
!>                  where t is a triangular matrix of order n. here trans(t)
!>                  denotes the transpose of the matrix t.
!> \param t         t contains the matrix of the system. the zero
!>                  elements of the matrix are not referenced, and
!>                  the corresponding elements of the array can be
!>                  used to store other information.
!> \param ldt       ldt is the leading dimension of the array t.
!> \param n         n is the order of the system.
!> \param b         contains the right hand side of the system.
!>                  on return
!>                  b contains the solution, if info .eq. 0.
!>                  otherwise b is unaltered.
!> \param job       job specifies what kind of system is to be solved.
!>                   if job is
!>                       00   solve t*x=b, t lower triangular,
!>                       01   solve t*x=b, t upper triangular,
!>                       10   solve trans(t)*x=b, t lower triangular,
!>                       11   solve trans(t)*x=b, t upper triangular.
!> \param info      on return
!>                  info contains zero if the system is nonsingular.
!>                  otherwise info contains the index of
!>                  the first zero diagonal element of t.
! **************************************************************************************************
      SUBROUTINE dtrsl(t,ldt,n,b,job,info)
      INTEGER, INTENT(in)                                :: ldt
      REAL(KIND=dp), INTENT(in)                          :: t(ldt,*)
      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(inout)                       :: b(*)
      INTEGER, INTENT(in)                                :: job
      INTEGER, INTENT(out)                               :: info

      CHARACTER(len=*), PARAMETER :: routineN = 'dtrsl', routineP = moduleN//':'//routineN

      INTEGER                                            :: CASE, j, jj
      REAL(KIND=dp)                                      :: ddot, temp

!     linpack. this version dated 08/14/78 .
!     g. w. stewart, university of maryland, argonne national lab.
!
!     begin block permitting ...exits to 150
!
!        check for zero diagonal elements.
!

         DO info = 1, n
!     ......exit
            IF (t(info,info) == 0.0_dp) RETURN
         END DO
         info = 0
!
!        determine the task and go to it.
!
         CASE = 1
         IF (MOD(job,10) /= 0) CASE = 2
         IF (MOD(job,100)/10 /= 0) CASE = CASE + 2

         SELECT CASE(CASE)
         CASE(1)
!
!        solve t*x=b for t lower triangular
!
            b(1) = b(1)/t(1,1)
            IF (n>1) THEN
               DO j = 2, n
                  temp = -b(j-1)
                  CALL daxpy(n-j+1,temp,t(j,j-1),1,b(j),1)
                  b(j) = b(j)/t(j,j)
               END DO
            END IF
         CASE(2)
!
!        solve t*x=b for t upper triangular.
!
            b(n) = b(n)/t(n,n)
            IF (n>1) THEN
               DO jj = 2, n
                  j = n - jj + 1
                  temp = -b(j+1)
                  CALL daxpy(j,temp,t(1,j+1),1,b(1),1)
                  b(j) = b(j)/t(j,j)
               END DO
            END IF
         CASE(3)
!
!        solve trans(t)*x=b for t lower triangular.
!
            b(n) = b(n)/t(n,n)
            IF (n>1) THEN
               DO jj = 2, n
                  j = n - jj + 1
                  b(j) = b(j) - ddot(jj-1,t(j+1,j),1,b(j+1),1)
                  b(j) = b(j)/t(j,j)
               END DO
            END IF
         CASE(4)
!
!        solve trans(t)*x=b for t upper triangular.
!
            b(1) = b(1)/t(1,1)
            IF (.NOT.(n < 2)) THEN
               DO j = 2, n
                  b(j) = b(j) - ddot(j-1,t(1,j),1,b(1),1)
                  b(j) = b(j)/t(j,j)
               END DO
            END IF
         CASE DEFAULT
            CALL cp__b("motion/cp_lbfgs.F",3793,"unexpected case")
         END SELECT

      RETURN
      END SUBROUTINE dtrsl

!MK Timer

! **************************************************************************************************    
!> \brief This routine computes cpu time in double precision; it makes use o
!>        the intrinsic f90 cpu_time therefore a conversion type is
!>        needed.
!> \param ttime ...
! **************************************************************************************************
      SUBROUTINE timer(ttime)
      REAL(KIND=dp)                                      :: ttime

!
!     REAL temp
!
!           J.L Morales  Departamento de Matematicas,
!                        Instituto Tecnologico Autonomo de Mexico
!                        Mexico D.F.
!
!           J.L Nocedal  Department of Electrical Engineering and
!                        Computer Science.
!                        Northwestern University. Evanston, IL. USA
!
!                        January 21, 2011
!
!MK      temp = sngl(ttime)
!MK      CALL cpu_time(temp)
!MK      ttime = DBLE(temp)

         ttime = m_walltime()

      END SUBROUTINE timer

! **************************************************************************************************    
!> \brief  Saves the lcoal variables, long term this should be replaces by a lbfgs type
!> \param lsave    lsave is a working array
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 If lsave(1) = .true.  then  the initial X has been replaced by
!>                               its projection in the feasible set
!>                 If lsave(2) = .true.  then  the problem is constrained;
!>                 If lsave(3) = .true.  then  each variable has upper and lower bounds;
!> \param isave    isave is a working array
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 isave(22) = the total number of intervals explored in the
!>                         search of Cauchy points;
!>                 isave(26) = the total number of skipped BFGS updates before the current iteration;
!>                 isave(30) = the number of current iteration;
!>                 isave(31) = the total number of BFGS updates prior the current iteration;
!>                 isave(33) = the number of intervals explored in the search of
!>                             Cauchy point in the current iteration;
!>                 isave(34) = the total number of function and gradient evaluations;
!>                 isave(36) = the number of function value or gradient
!>                             evaluations in the current iteration;
!>                 if isave(37) = 0  then the subspace argmin is within the box;
!>                 if isave(37) = 1  then the subspace argmin is beyond the box;
!>                 isave(38) = the number of free variables in the current iteration;
!>                 isave(39) = the number of active constraints in the current iteration;
!>                 n + 1 - isave(40) = the number of variables leaving the set of
!>                                     active constraints in the current iteration;
!>                 isave(41) = the number of variables entering the set of active
!>                             constraints in the current iteration.
!> \param dsave    dsave is a working array of dimension 29.
!>                 On exit with 'task' = NEW_X, the following information is available:
!>                 dsave(1) = current 'theta' in the BFGS matrix;
!>                 dsave(2) = f(x) in the previous iteration;
!>                 dsave(3) = factr*epsmch;
!>                 dsave(4) = 2-norm of the line search direction vector;
!>                 dsave(5) = the machine precision epsmch generated by the code;
!>                 dsave(7) = the accumulated time spent on searching for Cauchy points;
!>                 dsave(8) = the accumulated time spent on subspace minimization;
!>                 dsave(9) = the accumulated time spent on line search;
!>                 dsave(11) = the slope of the line search function at the current point of line search;
!>                 dsave(12) = the maximum relative step length imposed in line search;
!>                 dsave(13) = the infinity norm of the projected gradient;
!>                 dsave(14) = the relative step length in the line search;
!>                 dsave(15) = the slope of the line search function at the starting point of the line search;
!>                 dsave(16) = the square of the 2-norm of the line search direction vector.
!> \param x_projected ...
!> \param constrained ...
!> \param boxed ...
!> \param updatd ...
!> \param nintol ...
!> \param itfile ...
!> \param iback ...
!> \param nskip ...
!> \param head ...
!> \param col ...
!> \param itail ...
!> \param iter ...
!> \param iupdat ...
!> \param nseg ...
!> \param nfgv ...
!> \param info ...
!> \param ifun ...
!> \param iword ...
!> \param nfree ...
!> \param nact ...
!> \param ileave ...
!> \param nenter ...
!> \param theta ...
!> \param fold ...
!> \param tol ...
!> \param dnorm ...
!> \param epsmch ...
!> \param cpu1 ...
!> \param cachyt ...
!> \param sbtime ...
!> \param lnscht ...
!> \param time1 ...
!> \param gd ...
!> \param step_max ...
!> \param g_inf_norm ...
!> \param stp ...
!> \param gdold ...
!> \param dtd ...
!> \author Samuel Andermatt (01.15)
! **************************************************************************************************

      SUBROUTINE save_local(lsave,isave,dsave,x_projected,constrained,boxed,updatd,nintol,itfile,iback,nskip,head,col,itail,&
                            iter,iupdat,nseg,nfgv,info,ifun,iword,nfree,nact,ileave,nenter,theta,fold,tol,dnorm,epsmch,cpu1,&
                            cachyt,sbtime,lnscht,time1,gd,step_max,g_inf_norm,stp,gdold,dtd)
      LOGICAL, INTENT(out)                               :: lsave(4)
      INTEGER, INTENT(out)                               :: isave(23)
      REAL(KIND=dp), INTENT(out)                         :: dsave(29)
      LOGICAL, INTENT(in)                                :: x_projected, constrained, boxed, updatd
      INTEGER, INTENT(in)                                :: nintol, itfile, iback, nskip, head, col, &
                                                            itail, iter, iupdat, nseg, nfgv, info, &
                                                            ifun, iword, nfree, nact, ileave, &
                                                            nenter
      REAL(KIND=dp), INTENT(in)                          :: theta, fold, tol, dnorm, epsmch, cpu1, &
                                                            cachyt, sbtime, lnscht, time1, gd, &
                                                            step_max, g_inf_norm, stp, gdold, dtd

      lsave(1)  = x_projected
      lsave(2)  = constrained
      lsave(3)  = boxed
      lsave(4)  = updatd

      isave(1)  = nintol
      isave(3)  = itfile
      isave(4)  = iback
      isave(5)  = nskip
      isave(6)  = head
      isave(7)  = col
      isave(8)  = itail
      isave(9)  = iter
      isave(10) = iupdat
      isave(12) = nseg
      isave(13) = nfgv
      isave(14) = info
      isave(15) = ifun
      isave(16) = iword
      isave(17) = nfree
      isave(18) = nact
      isave(19) = ileave
      isave(20) = nenter

      dsave(1)  = theta
      dsave(2)  = fold
      dsave(3)  = tol
      dsave(4)  = dnorm
      dsave(5)  = epsmch
      dsave(6)  = cpu1
      dsave(7)  = cachyt
      dsave(8)  = sbtime
      dsave(9)  = lnscht
      dsave(10) = time1
      dsave(11) = gd
      dsave(12) = step_max
      dsave(13) = g_inf_norm
      dsave(14) = stp
      dsave(15) = gdold
      dsave(16) = dtd

      END SUBROUTINE save_local

END MODULE cp_lbfgs
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that optimize a functional using the limited memory bfgs
!>      quasi-newton method.
!>      The process set up so that a master runs the real optimizer and the
!>      others help then to calculate the objective function.
!>      The arguments for the objective function are physicaly present in
!>      every processor (nedeed in the actual implementation of pao).
!>      In the future tha arguments themselves could be distributed.
!> \par History
!>      09.2003 globenv->para_env, retain/release, better parallel behaviour
!> \author Fawzi Mohamed
!>      @version 2.2002
! **************************************************************************************************
MODULE cp_lbfgs_optimizer_gopt
   USE cp_lbfgs,                        ONLY: setulb
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE force_env_types,                 ONLY: force_env_type
   USE gopt_f_methods,                  ONLY: gopt_f_io
   USE gopt_f_types,                    ONLY: gopt_f_release,&
                                              gopt_f_retain,&
                                              gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
   USE message_passing,                 ONLY: mp_bcast
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param error variable to control error logging, stopping,...
!>             see module cp_error_handling
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! *****************************************************************************
INTERFACE

  RECURSIVE SUBROUTINE cp_eval_at(gopt_env,x,f,gradient,master,&
                                  final_evaluation,para_env)

    USE cp_para_types,                   ONLY: cp_para_env_type
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE kinds,                           ONLY: dp

    TYPE(gopt_f_type), POINTER               :: gopt_env
    REAL(KIND=dp), DIMENSION(:), POINTER     :: x
    REAL(KIND=dp), INTENT(out), OPTIONAL     :: f
    REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
    INTEGER, INTENT(IN)                      :: master
    LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
    TYPE(cp_para_env_type), POINTER          :: para_env

  END SUBROUTINE cp_eval_at

END INTERFACE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_lbfgs_optimizer_gopt'
   INTEGER, PRIVATE, SAVE :: last_lbfgs_optimizer_id = 0

   ! types
   PUBLIC :: cp_lbfgs_opt_gopt_type

   ! core methods

   ! special methos

   ! underlying functions
   PUBLIC :: cp_opt_gopt_create, cp_opt_gopt_release, &
             cp_opt_gopt_next, &
             cp_opt_gopt_stop

   ! initalize the object
   INTERFACE cp_create
      MODULE PROCEDURE cp_opt_gopt_create
   END INTERFACE
   ! releases the given object
   INTERFACE cp_release
      MODULE PROCEDURE cp_opt_gopt_release
   END INTERFACE
   ! retains the given object
   INTERFACE cp_retain
      MODULE PROCEDURE cp_opt_gopt_retain
   END INTERFACE
   ! returns attributes about the object
   INTERFACE cp_get
      MODULE PROCEDURE cp_opt_gopt_get
   END INTERFACE
   ! goes to the next point, returns true if not converged or error
   INTERFACE cp_next
      MODULE PROCEDURE cp_opt_gopt_next
   END INTERFACE
   ! goes to the next point
   INTERFACE cp_step
      MODULE PROCEDURE cp_opt_gopt_step
   END INTERFACE
   ! stops the iteration
   INTERFACE cp_stop
      MODULE PROCEDURE cp_opt_gopt_stop
   END INTERFACE

! **************************************************************************************************
!> \brief info for the optimizer (see the description of this module)
!> \param task the actual task of the optimizer (in the master it is up to
!>        date, in case of error also the slaves one get updated.
!> \param csave internal character string used by the lbfgs optimizer,
!>        meaningful only in the master
!> \param lsave logical array used by the lbfgs optimizer, updated only
!>        in the master
!>        On exit with task = 'NEW_X', the following information is
!>        available:
!>           lsave(1) = .true.  the initial x did not satisfy the bounds;
!>           lsave(2) = .true.  the problem contains bounds;
!>           lsave(3) = .true.  each variable has upper and lower bounds.
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param id_nr identification number (unique)
!> \param m the dimension of the subspace used to approximate the second
!>        derivative
!> \param print_every every how many iterations output should be written.
!>        if 0 only at end, if print_every<0 never
!> \param master the pid of the master processor
!> \param max_f_per_iter the maximum number of function evaluations per
!>        iteration
!> \param status 0: just initialized, 1: f g calculation,
!>        2: begin new iteration, 3: ended iteration,
!>        4: normal (converged) exit, 5: abnormal (error) exit,
!>        6: daellocated
!> \param n_iter the actual iteration number
!> \param kind_of_bound an array with 0 (no bound), 1 (lower bound),
!>        2 (both bounds), 3 (upper bound), to describe the bounds
!>        of every variable
!> \param i_work_array an integer workarray of dimension 3*n, present only
!>        in the master
!> \param isave is an INTEGER working array of dimension 44.
!>        On exit with task = 'NEW_X', it contains information that
!>        the user may want to access:
!> \param isave (30) = the current iteration number;
!> \param isave (34) = the total number of function and gradient
!>           evaluations;
!> \param isave (36) = the number of function value or gradient
!>           evaluations in the current iteration;
!> \param isave (38) = the number of free variables in the current
!>           iteration;
!> \param isave (39) = the number of active constraints at the current
!>           iteration;
!> \param f the actual best value of the object function
!> \param wanted_relative_f_delta the wanted relative error on f
!>        (to be multiplied by epsilon), 0.0 -> no check
!> \param wanted_projected_gradient the wanted error on the projected
!>        gradient (hessian times the gradient), 0.0 -> no check
!> \param last_f the value of f in the last iteration
!> \param projected_gradient the value of the sup norm of the projected
!>        gradient
!> \param x the actual evaluation point (best one if converged or stopped)
!> \param lower_bound the lower bounds
!> \param upper_bound the upper bounds
!> \param gradient the actual gradient
!> \param dsave info date for lbfgs (master only)
!> \param work_array a work array for lbfgs (master only)
!> \param para_env the parallel environment for this optimizer
!> \param obj_funct the objective function to be optimized
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
! **************************************************************************************************
   TYPE cp_lbfgs_opt_gopt_type
      CHARACTER(len=60) :: task
      CHARACTER(len=60) :: csave
      LOGICAL :: lsave(4)
      INTEGER :: m, print_every, master, max_f_per_iter, status, n_iter
      INTEGER :: ref_count, id_nr
      INTEGER, DIMENSION(:), POINTER :: kind_of_bound, i_work_array, isave
      REAL(kind=dp) :: f, wanted_relative_f_delta, wanted_projected_gradient, &
                       last_f, projected_gradient, eold, emin, trust_radius
      REAL(kind=dp), DIMENSION(:), POINTER :: x, lower_bound, upper_bound, &
                                              gradient, dsave, work_array
      TYPE(cp_para_env_type), POINTER :: para_env
      TYPE(gopt_f_type), POINTER :: obj_funct
   END TYPE cp_lbfgs_opt_gopt_type

CONTAINS

! **************************************************************************************************
!> \brief initializes the optimizer
!> \param optimizer ...
!> \param para_env ...
!> \param obj_funct ...
!> \param x0 ...
!> \param m ...
!> \param print_every ...
!> \param wanted_relative_f_delta ...
!> \param wanted_projected_gradient ...
!> \param lower_bound ...
!> \param upper_bound ...
!> \param kind_of_bound ...
!> \param master ...
!> \param max_f_per_iter ...
!> \param trust_radius ...
!> \par History
!>      02.2002 created [fawzi]
!>      09.2003 refactored (retain/release,para_env) [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      redirects the lbfgs output the the default unit
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_create(optimizer, para_env, obj_funct, x0, m, print_every, &
                                 wanted_relative_f_delta, wanted_projected_gradient, lower_bound, upper_bound, &
                                 kind_of_bound, master, max_f_per_iter, trust_radius)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gopt_f_type), POINTER                         :: obj_funct
      REAL(kind=dp), DIMENSION(:), INTENT(in)            :: x0
      INTEGER, INTENT(in), OPTIONAL                      :: m, print_every
      REAL(kind=dp), INTENT(in), OPTIONAL                :: wanted_relative_f_delta, &
                                                            wanted_projected_gradient
      REAL(kind=dp), DIMENSION(SIZE(x0)), INTENT(in), &
         OPTIONAL                                        :: lower_bound, upper_bound
      INTEGER, DIMENSION(SIZE(x0)), INTENT(in), OPTIONAL :: kind_of_bound
      INTEGER, INTENT(in), OPTIONAL                      :: master, max_f_per_iter
      REAL(kind=dp), INTENT(in), OPTIONAL                :: trust_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, lenwa, n

      CALL timeset(routineN, handle)

      ALLOCATE (optimizer)
      NULLIFY (optimizer%kind_of_bound, &
               optimizer%i_work_array, &
               optimizer%isave, &
               optimizer%x, &
               optimizer%lower_bound, &
               optimizer%upper_bound, &
               optimizer%gradient, &
               optimizer%dsave, &
               optimizer%work_array, &
               optimizer%para_env, &
               optimizer%obj_funct)
      optimizer%ref_count = 0
      last_lbfgs_optimizer_id = last_lbfgs_optimizer_id+1
      optimizer%id_nr = last_lbfgs_optimizer_id
      n = SIZE(x0)
      optimizer%m = 4
      IF (PRESENT(m)) optimizer%m = m
      optimizer%master = para_env%source
      optimizer%para_env => para_env
      CALL cp_para_env_retain(para_env)
      optimizer%obj_funct => obj_funct
      CALL gopt_f_retain(obj_funct)
      optimizer%max_f_per_iter = 20
      IF (PRESENT(max_f_per_iter)) optimizer%max_f_per_iter = max_f_per_iter
      optimizer%print_every = -1
      optimizer%n_iter = 0
      optimizer%f = -1.0_dp
      optimizer%last_f = -1.0_dp
      optimizer%projected_gradient = -1.0_dp
      IF (PRESENT(print_every)) optimizer%print_every = print_every
      IF (PRESENT(master)) optimizer%master = master
      IF (optimizer%master == optimizer%para_env%mepos) THEN
         !MK This has to be adapted for a new L-BFGS version possibly
         lenwa = 2*optimizer%m*n+5*n+11*optimizer%m*optimizer%m+8*optimizer%m
         ALLOCATE (optimizer%kind_of_bound(n), optimizer%i_work_array(3*n), &
                   optimizer%isave(44))
         ALLOCATE (optimizer%x(n), optimizer%lower_bound(n), &
                   optimizer%upper_bound(n), optimizer%gradient(n), &
                   optimizer%dsave(29), optimizer%work_array(lenwa))
         optimizer%x = x0
         optimizer%task = 'START'
         optimizer%wanted_relative_f_delta = wanted_relative_f_delta
         optimizer%wanted_projected_gradient = wanted_projected_gradient
         optimizer%kind_of_bound = 0
         IF (PRESENT(kind_of_bound)) optimizer%kind_of_bound = kind_of_bound
         IF (PRESENT(lower_bound)) optimizer%lower_bound = lower_bound
         IF (PRESENT(upper_bound)) optimizer%upper_bound = upper_bound
         IF (PRESENT(trust_radius)) optimizer%trust_radius = trust_radius

         CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                     optimizer%lower_bound, optimizer%upper_bound, &
                     optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                     optimizer%wanted_relative_f_delta, &
                     optimizer%wanted_projected_gradient, optimizer%work_array, &
                     optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                     optimizer%csave, optimizer%lsave, optimizer%isave, &
                     optimizer%dsave, optimizer%trust_radius)
      ELSE
         NULLIFY ( &
            optimizer%kind_of_bound, optimizer%i_work_array, optimizer%isave, &
            optimizer%lower_bound, optimizer%upper_bound, optimizer%gradient, &
            optimizer%dsave, optimizer%work_array)
         ALLOCATE (optimizer%x(n))
         optimizer%x(:) = 0.0_dp
         ALLOCATE (optimizer%gradient(n))
         optimizer%gradient(:) = 0.0_dp
      END IF
      CALL mp_bcast(optimizer%x, optimizer%master, optimizer%para_env%group)
      optimizer%status = 0
      optimizer%ref_count = 1

      CALL timestop(handle)

   END SUBROUTINE cp_opt_gopt_create

! **************************************************************************************************
!> \brief retains the given optimizer (see doc/ReferenceCounting.html)
!> \param optimizer the optimizer to retain
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_retain(optimizer)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",307)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",308)
      optimizer%ref_count = optimizer%ref_count+1
   END SUBROUTINE cp_opt_gopt_retain

! **************************************************************************************************
!> \brief releases the optimizer (see doc/ReferenceCounting.html)
!> \param optimizer the object that should be released
!> \par History
!>      02.2002 created [fawzi]
!>      09.2003 dealloc_ref->release [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_release(optimizer)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(optimizer)) THEN
         IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",331)
         optimizer%ref_count = optimizer%ref_count-1
         IF (optimizer%ref_count == 0) THEN
            optimizer%status = 6
            IF (ASSOCIATED(optimizer%kind_of_bound)) THEN
               DEALLOCATE (optimizer%kind_of_bound)
            END IF
            IF (ASSOCIATED(optimizer%i_work_array)) THEN
               DEALLOCATE (optimizer%i_work_array)
            END IF
            IF (ASSOCIATED(optimizer%isave)) THEN
               DEALLOCATE (optimizer%isave)
            END IF
            IF (ASSOCIATED(optimizer%x)) THEN
               DEALLOCATE (optimizer%x)
            END IF
            IF (ASSOCIATED(optimizer%lower_bound)) THEN
               DEALLOCATE (optimizer%lower_bound)
            END IF
            IF (ASSOCIATED(optimizer%upper_bound)) THEN
               DEALLOCATE (optimizer%upper_bound)
            END IF
            IF (ASSOCIATED(optimizer%gradient)) THEN
               DEALLOCATE (optimizer%gradient)
            END IF
            IF (ASSOCIATED(optimizer%dsave)) THEN
               DEALLOCATE (optimizer%dsave)
            END IF
            IF (ASSOCIATED(optimizer%work_array)) THEN
               DEALLOCATE (optimizer%work_array)
            END IF
            CALL cp_para_env_release(optimizer%para_env)
            CALL gopt_f_release(optimizer%obj_funct)
            NULLIFY (optimizer%obj_funct)
            DEALLOCATE (optimizer)
         END IF
      END IF
      NULLIFY (optimizer)
      CALL timestop(handle)
   END SUBROUTINE cp_opt_gopt_release

! **************************************************************************************************
!> \brief takes different valuse from the optimizer
!> \param optimizer ...
!> \param para_env ...
!> \param obj_funct ...
!> \param m ...
!> \param print_every ...
!> \param id_nr ...
!> \param wanted_relative_f_delta ...
!> \param wanted_projected_gradient ...
!> \param x ...
!> \param lower_bound ...
!> \param upper_bound ...
!> \param kind_of_bound ...
!> \param master ...
!> \param actual_projected_gradient ...
!> \param n_var ...
!> \param n_iter ...
!> \param status ...
!> \param max_f_per_iter ...
!> \param at_end ...
!> \param is_master ...
!> \param last_f ...
!> \param f ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_get(optimizer, para_env, &
                              obj_funct, m, print_every, id_nr, &
                              wanted_relative_f_delta, wanted_projected_gradient, &
                              x, lower_bound, upper_bound, kind_of_bound, master, &
                              actual_projected_gradient, &
                              n_var, n_iter, status, max_f_per_iter, at_end, &
                              is_master, last_f, f)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(gopt_f_type), OPTIONAL, POINTER               :: obj_funct
      INTEGER, INTENT(out), OPTIONAL                     :: m, print_every, id_nr
      REAL(kind=dp), INTENT(out), OPTIONAL               :: wanted_relative_f_delta, &
                                                            wanted_projected_gradient
      REAL(kind=dp), DIMENSION(:), OPTIONAL, POINTER     :: x, lower_bound, upper_bound
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: kind_of_bound
      INTEGER, INTENT(out), OPTIONAL                     :: master
      REAL(kind=dp), INTENT(out), OPTIONAL               :: actual_projected_gradient
      INTEGER, INTENT(out), OPTIONAL                     :: n_var, n_iter, status, max_f_per_iter
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end, is_master
      REAL(kind=dp), INTENT(out), OPTIONAL               :: last_f, f

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_get', &
         routineP = moduleN//':'//routineN

!  call timeset(routineN,handle)

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",427)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",428)

      IF (PRESENT(is_master)) is_master = optimizer%master == optimizer%para_env%mepos
      IF (PRESENT(master)) master = optimizer%master
      IF (PRESENT(status)) status = optimizer%status
      IF (PRESENT(para_env)) para_env => optimizer%para_env
      IF (PRESENT(id_nr)) id_nr = optimizer%id_nr
      IF (PRESENT(obj_funct)) obj_funct = optimizer%obj_funct
      IF (PRESENT(m)) m = optimizer%m
      IF (PRESENT(max_f_per_iter)) max_f_per_iter = optimizer%max_f_per_iter
      IF (PRESENT(wanted_projected_gradient)) &
         wanted_projected_gradient = optimizer%wanted_projected_gradient
      IF (PRESENT(wanted_relative_f_delta)) &
         wanted_relative_f_delta = optimizer%wanted_relative_f_delta
      IF (PRESENT(print_every)) print_every = optimizer%print_every
      IF (PRESENT(x)) x => optimizer%x
      IF (PRESENT(n_var)) n_var = SIZE(x)
      IF (PRESENT(lower_bound)) lower_bound => optimizer%lower_bound
      IF (PRESENT(upper_bound)) upper_bound => optimizer%upper_bound
      IF (PRESENT(kind_of_bound)) kind_of_bound => optimizer%kind_of_bound
      IF (PRESENT(n_iter)) n_iter = optimizer%n_iter
      IF (PRESENT(last_f)) last_f = optimizer%last_f
      IF (PRESENT(f)) f = optimizer%f
      IF (PRESENT(at_end)) at_end = optimizer%status > 3
      IF (PRESENT(actual_projected_gradient)) &
         actual_projected_gradient = optimizer%projected_gradient
      IF (optimizer%master == optimizer%para_env%mepos) THEN
         IF (optimizer%isave(30) > 1 .AND. (optimizer%task(1:5) == "NEW_X" .OR. &
                                            optimizer%task(1:4) == "STOP" .AND. optimizer%task(7:9) == "CPU")) THEN
            ! nr iterations >1 .and. dsave contains the wanted data
            IF (PRESENT(last_f)) last_f = optimizer%dsave(2)
            IF (PRESENT(actual_projected_gradient)) &
               actual_projected_gradient = optimizer%dsave(13)
         ELSE
            IF(.NOT.(.NOT. PRESENT(last_f)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",462)
            IF(.NOT.(.NOT. PRESENT(actual_projected_gradient)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",463)
         END IF
      ELSE
         IF (PRESENT(lower_bound) .OR. PRESENT(upper_bound) .OR. PRESENT(kind_of_bound)) &
            CALL cp__w("motion/cp_lbfgs_optimizer_gopt.F",467,"asked undefined types")
      END IF

      !  call timestop(handle)
   END SUBROUTINE cp_opt_gopt_get

! **************************************************************************************************
!> \brief does one optimization step
!> \param optimizer ...
!> \param n_iter ...
!> \param f ...
!> \param last_f ...
!> \param projected_gradient ...
!> \param converged ...
!> \param geo_section ...
!> \param force_env ...
!> \param gopt_param ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
!> \note
!>      use directly mainlb in place of setulb ??
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_opt_gopt_step(optimizer, n_iter, f, last_f, &
                                         projected_gradient, converged, geo_section, force_env, &
                                         gopt_param)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      INTEGER, INTENT(out), OPTIONAL                     :: n_iter
      REAL(kind=dp), INTENT(out), OPTIONAL               :: f, last_f, projected_gradient
      LOGICAL, INTENT(out), OPTIONAL                     :: converged
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_step', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=5)                                   :: wildcard
      INTEGER                                            :: dataunit, handle, its
      LOGICAL                                            :: conv, is_master, justEntred
      REAL(KIND=dp)                                      :: t_diff, t_now, t_old
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xold
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger, xold)
      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)
      justEntred = .TRUE.
      is_master = optimizer%master == optimizer%para_env%mepos
      IF (PRESENT(converged)) converged = optimizer%status == 4
      ALLOCATE (xold(SIZE(optimizer%x)))
      xold = optimizer%x
      t_old = m_walltime()

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",522)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",523)

      IF (optimizer%status >= 4) THEN
         CALL cp__w("motion/cp_lbfgs_optimizer_gopt.F",526,"status>=4, trying to restart")
         optimizer%status = 0
         IF (is_master) THEN
            optimizer%task = 'START'
            CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                        optimizer%lower_bound, optimizer%upper_bound, &
                        optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                        optimizer%wanted_relative_f_delta, &
                        optimizer%wanted_projected_gradient, optimizer%work_array, &
                        optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                        optimizer%csave, optimizer%lsave, optimizer%isave, &
                        optimizer%dsave, optimizer%trust_radius)
         END IF
      END IF

      DO
         ifMaster: IF (is_master) THEN
            IF (optimizer%task(1:7) == 'RESTART') THEN
               ! restart the optimizer
               optimizer%status = 0
               optimizer%task = 'START'
               CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                           optimizer%lower_bound, optimizer%upper_bound, &
                           optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                           optimizer%wanted_relative_f_delta, &
                           optimizer%wanted_projected_gradient, optimizer%work_array, &
                           optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                           optimizer%csave, optimizer%lsave, optimizer%isave, &
                           optimizer%dsave, optimizer%trust_radius)
            END IF
            IF (optimizer%task(1:2) == 'FG') THEN
               IF (optimizer%isave(36) > optimizer%max_f_per_iter) THEN
                  optimizer%task = 'STOP: CPU, hit max f eval in iter'
                  optimizer%status = 5 ! anormal exit
                  CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                              optimizer%lower_bound, optimizer%upper_bound, &
                              optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                              optimizer%wanted_relative_f_delta, &
                              optimizer%wanted_projected_gradient, optimizer%work_array, &
                              optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                              optimizer%csave, optimizer%lsave, optimizer%isave, &
                              optimizer%dsave, optimizer%trust_radius)
               ELSE
                  optimizer%status = 1
               END IF
            ELSE IF (optimizer%task(1:5) == 'NEW_X') THEN
               IF (justEntred) THEN
                  optimizer%status = 2
                  CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                              optimizer%lower_bound, optimizer%upper_bound, &
                              optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                              optimizer%wanted_relative_f_delta, &
                              optimizer%wanted_projected_gradient, optimizer%work_array, &
                              optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                              optimizer%csave, optimizer%lsave, optimizer%isave, &
                              optimizer%dsave, optimizer%trust_radius)
               ELSE
                  optimizer%status = 3
               END IF
            ELSE IF (optimizer%task(1:4) == 'CONV') THEN
               optimizer%status = 4
            ELSE IF (optimizer%task(1:4) == 'STOP') THEN
               optimizer%status = 5
               CALL cp__w("motion/cp_lbfgs_optimizer_gopt.F",589,"task became stop in an unknown way")
            ELSE IF (optimizer%task(1:5) == 'ERROR') THEN
               optimizer%status = 5
            ELSE
               CALL cp__w("motion/cp_lbfgs_optimizer_gopt.F",593,"unknown task '"//optimizer%task//"'")
            END IF
         END IF ifMaster
         CALL mp_bcast(optimizer%status, optimizer%master, optimizer%para_env%group)
         ! Dump info
         IF (optimizer%status == 3) THEN
            its = 0
            IF (is_master) THEN
               ! Iteration level is taken into account in the optimizer external loop
               its = optimizer%isave(30)
            END IF
         END IF
         !
         SELECT CASE (optimizer%status)
         CASE (1)
            !op=1 evaluate f and g
            CALL cp_eval_at(optimizer%obj_funct, x=optimizer%x, &
                            f=optimizer%f, &
                            gradient=optimizer%gradient, &
                            final_evaluation=.FALSE., &
                            master=optimizer%master, para_env=optimizer%para_env)
            ! do not use keywords?
            IF (is_master) THEN
               CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                           optimizer%lower_bound, optimizer%upper_bound, &
                           optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                           optimizer%wanted_relative_f_delta, &
                           optimizer%wanted_projected_gradient, optimizer%work_array, &
                           optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                           optimizer%csave, optimizer%lsave, optimizer%isave, &
                           optimizer%dsave, optimizer%trust_radius)
            END IF
            CALL mp_bcast(optimizer%x, optimizer%master, optimizer%para_env%group)
         CASE (2)
            !op=2 begin new iter
            CALL mp_bcast(optimizer%x, optimizer%master, optimizer%para_env%group)
            t_old = m_walltime()
         CASE (3)
            !op=3 ended iter
            wildcard = "LBFGS"
            dataunit = cp_print_key_unit_nr(logger, geo_section, &
                                            "PRINT%PROGRAM_RUN_INFO", extension=".geoLog")
            IF (is_master) its = optimizer%isave(30)
            CALL mp_bcast(its, optimizer%master, optimizer%para_env%group)

            ! Some IO and Convergence check
            t_now = m_walltime()
            t_diff = t_now-t_old
            t_old = t_now
            CALL gopt_f_io(optimizer%obj_funct, force_env, force_env%root_section, &
                           its, optimizer%f, dataunit, optimizer%eold, optimizer%emin, wildcard, gopt_param, &
                           SIZE(optimizer%x), optimizer%x-xold, optimizer%gradient, conv, used_time=t_diff)
            CALL mp_bcast(conv, optimizer%master, optimizer%para_env%group)
            CALL cp_print_key_finished_output(dataunit, logger, geo_section, &
                                              "PRINT%PROGRAM_RUN_INFO")
            optimizer%eold = optimizer%f
            optimizer%emin = MIN(optimizer%emin, optimizer%eold)
            xold = optimizer%x
            IF (PRESENT(converged)) converged = conv
            EXIT
         CASE (4)
            !op=4 (convergence - normal exit)
            ! Specific L-BFGS convergence criteria.. overrides the convergence criteria on
            ! stepsize and gradients
            dataunit = cp_print_key_unit_nr(logger, geo_section, &
                                            "PRINT%PROGRAM_RUN_INFO", extension=".geoLog")
            IF (dataunit > 0) THEN
               WRITE (dataunit, '(T2,A)') ""
               WRITE (dataunit, '(T2,A)') "***********************************************"
               WRITE (dataunit, '(T2,A)') "* Specific L-BFGS convergence criteria         "
               WRITE (dataunit, '(T2,A)') "* WANTED_PROJ_GRADIENT and WANTED_REL_F_ERROR  "
               WRITE (dataunit, '(T2,A)') "* satisfied .... run CONVERGED!                "
               WRITE (dataunit, '(T2,A)') "***********************************************"
               WRITE (dataunit, '(T2,A)') ""
            END IF
            CALL cp_print_key_finished_output(dataunit, logger, geo_section, &
                                              "PRINT%PROGRAM_RUN_INFO")
            IF (PRESENT(converged)) converged = .TRUE.
            EXIT
         CASE (5)
            ! op=5 abnormal exit ()
            CALL mp_bcast(optimizer%task, optimizer%master, &
                          optimizer%para_env%group)
         CASE (6)
            ! deallocated
            CALL cp__b("motion/cp_lbfgs_optimizer_gopt.F",678,"step on a deallocated opt structure ")
         CASE default
            CALL cp_abort(cp__l("motion/cp_lbfgs_optimizer_gopt.F",680), &
                          "unknown status "//cp_to_string(optimizer%status))
            optimizer%status = 5
            EXIT
         END SELECT
         IF (optimizer%status == 1 .AND. justEntred) THEN
            optimizer%eold = optimizer%f
            optimizer%emin = optimizer%eold
         END IF
         justEntred = .FALSE.
      END DO
      CALL mp_bcast(optimizer%x, optimizer%master, &
                    optimizer%para_env%group)
      CALL cp_opt_gopt_bcast_res(optimizer, &
                                 n_iter=optimizer%n_iter, &
                                 f=optimizer%f, last_f=optimizer%last_f, &
                                 projected_gradient=optimizer%projected_gradient)

      DEALLOCATE (xold)
      IF (PRESENT(f)) f = optimizer%f
      IF (PRESENT(last_f)) last_f = optimizer%last_f
      IF (PRESENT(projected_gradient)) &
         projected_gradient = optimizer%projected_gradient
      IF (PRESENT(n_iter)) n_iter = optimizer%n_iter
      CALL timestop(handle)

   END SUBROUTINE cp_opt_gopt_step

! **************************************************************************************************
!> \brief returns the results (and broadcasts them)
!> \param optimizer the optimizer object the info is taken from
!> \param n_iter the number of iterations
!> \param f the actual value of the objective function (f)
!> \param last_f the last value of f
!> \param projected_gradient the infinity norm of the projected gradient
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
!> \note
!>      private routine
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_bcast_res(optimizer, n_iter, f, last_f, &
                                    projected_gradient)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      INTEGER, INTENT(out), OPTIONAL                     :: n_iter
      REAL(kind=dp), INTENT(inout), OPTIONAL             :: f, last_f, projected_gradient

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_bcast_res', &
         routineP = moduleN//':'//routineN

      REAL(kind=dp), DIMENSION(4)                        :: results

!  call timeset(routineN,handle)

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",735)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",736)

      IF (optimizer%master == optimizer%para_env%mepos) THEN
         results = (/REAL(optimizer%isave(30), kind=dp), &
                     optimizer%f, optimizer%dsave(2), optimizer%dsave(13)/)
      END IF
      CALL mp_bcast(results, optimizer%master, &
                    optimizer%para_env%group)
      IF (PRESENT(n_iter)) n_iter = NINT(results(1))
      IF (PRESENT(f)) f = results(2)
      IF (PRESENT(last_f)) last_f = results(3)
      IF (PRESENT(projected_gradient)) &
         projected_gradient = results(4)

      !  call timestop(handle)
   END SUBROUTINE cp_opt_gopt_bcast_res

! **************************************************************************************************
!> \brief goes to the next optimal point (after an optimizer iteration)
!>      returns true if converged
!> \param optimizer the optimizer that goes to the next point
!> \param n_iter ...
!> \param f ...
!> \param last_f ...
!> \param projected_gradient ...
!> \param converged ...
!> \param geo_section ...
!> \param force_env ...
!> \param gopt_param ...
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
!> \note
!>      if you deactivate convergence control it returns never false
! **************************************************************************************************
   RECURSIVE FUNCTION cp_opt_gopt_next(optimizer, n_iter, f, last_f, &
                                       projected_gradient, converged, geo_section, force_env, &
                                       gopt_param) RESULT(res)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      INTEGER, INTENT(out), OPTIONAL                     :: n_iter
      REAL(kind=dp), INTENT(out), OPTIONAL               :: f, last_f, projected_gradient
      LOGICAL, INTENT(out)                               :: converged
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_next', &
         routineP = moduleN//':'//routineN

!call timeset(routineN,handle)

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",790)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",791)
      CALL cp_opt_gopt_step(optimizer, n_iter=n_iter, f=f, &
                            last_f=last_f, projected_gradient=projected_gradient, &
                            converged=converged, geo_section=geo_section, &
                            force_env=force_env, gopt_param=gopt_param)
      res = (optimizer%status < 40) .AND. .NOT. converged
      !call timestop(handle)
   END FUNCTION cp_opt_gopt_next

! **************************************************************************************************
!> \brief stops the optimization
!> \param optimizer ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!>      @version 2.2002
!> \note
!>      necessary???
! **************************************************************************************************
   SUBROUTINE cp_opt_gopt_stop(optimizer)
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_opt_gopt_stop', &
         routineP = moduleN//':'//routineN

!  call timeset(routineN,handle)

      IF(.NOT.(ASSOCIATED(optimizer)))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",818)
      IF(.NOT.(optimizer%ref_count > 0))CALL cp__a("motion/cp_lbfgs_optimizer_gopt.F",819)

      optimizer%task = 'STOPPED on user request'
      optimizer%status = 4 ! normal exit
      IF (optimizer%master == optimizer%para_env%mepos) THEN
         CALL setulb(SIZE(optimizer%x), optimizer%m, optimizer%x, &
                     optimizer%lower_bound, optimizer%upper_bound, &
                     optimizer%kind_of_bound, optimizer%f, optimizer%gradient, &
                     optimizer%wanted_relative_f_delta, &
                     optimizer%wanted_projected_gradient, optimizer%work_array, &
                     optimizer%i_work_array, optimizer%task, optimizer%print_every, &
                     optimizer%csave, optimizer%lsave, optimizer%isave, &
                     optimizer%dsave, optimizer%trust_radius)
      END IF

      ! call timestop(handle)
   END SUBROUTINE cp_opt_gopt_stop

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,type1_retain,type1_release,type1_eval_at)]
! ARGS:
!  INTERFACE = "#include "gopt_f77_methods.h""
!  USE = "USE gopt_f_types, ONLY: gopt_f_type, gopt_f_release,gopt_f_retain"
!  nametype1 = "gopt"
!  type1 = "type(gopt_f_type)"
!  type1_eval_at = "cp_eval_at"
!  type1_release = "gopt_f_release"
!  type1_retain = "gopt_f_retain"

END MODULE cp_lbfgs_optimizer_gopt
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Main driver for L-BFGS optimizer
!> \par History
!>      none
! **************************************************************************************************
MODULE cp_lbfgs_geo
   USE cell_types,                      ONLY: cell_type
   USE cp_external_control,             ONLY: external_control
   USE cp_lbfgs_optimizer_gopt,         ONLY: cp_lbfgs_opt_gopt_type,&
                                              cp_opt_gopt_create,&
                                              cp_opt_gopt_next,&
                                              cp_opt_gopt_release,&
                                              cp_opt_gopt_stop
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_methods,                  ONLY: gopt_f_ii,&
                                              gopt_f_io_finalize,&
                                              print_geo_opt_header,&
                                              print_geo_opt_nc
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_type
   USE input_constants,                 ONLY: default_ts_method_id
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_lbfgs_geo'

   PUBLIC :: geoopt_lbfgs

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \param gopt_param ...
!> \param globenv ...
!> \param geo_section ...
!> \param gopt_env ...
!> \param x0 ...
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   RECURSIVE SUBROUTINE geoopt_lbfgs(force_env, gopt_param, globenv, geo_section, gopt_env, &
                                     x0)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: geo_section
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0

      CHARACTER(len=*), PARAMETER :: routineN = 'geoopt_lbfgs', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iter_nr, its, output_unit
      LOGICAL                                            :: converged, should_stop
      REAL(KIND=dp)                                      :: trust_radius
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_lbfgs_opt_gopt_type), POINTER              :: optimizer
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section

      CALL timeset(routineN, handle)

      NULLIFY (optimizer, para_env)
      logger => cp_get_default_logger()
      root_section => force_env%root_section
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/cp_lbfgs_geo.F",87)
      IF(.NOT.(ASSOCIATED(gopt_param)))CALL cp__a("motion/cp_lbfgs_geo.F",88)
      IF(.NOT.(gopt_param%ref_count > 0))CALL cp__a("motion/cp_lbfgs_geo.F",89)

      CALL force_env_get(force_env, para_env=para_env, cell=cell)

      ! Geometry optimization starts now
      output_unit = cp_print_key_unit_nr(logger, geo_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".geoLog")
      CALL print_geo_opt_header(gopt_env, output_unit, "L-BFGS")

      ! Stop if not implemented
      IF (gopt_env%type_id == default_ts_method_id) &
         CALL cp__b("motion/cp_lbfgs_geo.F",100,"BFGS method not yet working with DIMER")

      CALL section_vals_val_get(geo_section, "LBFGS%TRUST_RADIUS", r_val=trust_radius)
      CALL cp_opt_gopt_create(optimizer, para_env=para_env, obj_funct=gopt_env, &
                              x0=x0, wanted_relative_f_delta=gopt_param%wanted_rel_f_error, &
                              wanted_projected_gradient=gopt_param%wanted_proj_gradient, m=gopt_param%max_h_rank, &
                              max_f_per_iter=gopt_param%max_f_per_iter, trust_radius=trust_radius)
      CALL cp_iterate(logger%iter_info, increment=0, iter_nr_out=iter_nr)
      converged = .FALSE.

      DO its = iter_nr+1, gopt_param%max_iter
         CALL cp_iterate(logger%iter_info, last=(its == gopt_param%max_iter))
         CALL section_vals_val_set(geo_section, "STEP_START_VAL", i_val=its)
         CALL gopt_f_ii(its, output_unit)

         ! Real optimization step..
         IF (.NOT. cp_opt_gopt_next(optimizer, geo_section=geo_section, &
                                    force_env=force_env, gopt_param=gopt_param, converged=converged)) EXIT

         ! Check for an external exit command
         CALL external_control(should_stop, "GEO", globenv=globenv)
         IF (should_stop) THEN
            CALL cp_opt_gopt_stop(optimizer)
            EXIT
         END IF
         IF (its == gopt_param%max_iter) EXIT
      END DO

      IF ((its == gopt_param%max_iter) .AND. (.NOT. converged)) THEN
         CALL print_geo_opt_nc(gopt_env, output_unit)
      END IF

      ! Write final output information, if converged
      CALL cp_iterate(logger%iter_info, last=.TRUE., increment=0)
      CALL gopt_f_io_finalize(gopt_env, force_env, optimizer%x, converged, its, root_section, &
                              optimizer%para_env, optimizer%master, output_unit)

      CALL cp_opt_gopt_release(optimizer)
      CALL cp_print_key_finished_output(output_unit, logger, geo_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE geoopt_lbfgs

END MODULE cp_lbfgs_geo
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief performs geometry optimization
!> \par History
!>      none
! **************************************************************************************************
MODULE geo_opt

   USE bfgs_optimizer,                  ONLY: geoopt_bfgs
   USE cg_optimizer,                    ONLY: geoopt_cg
   USE cp_lbfgs_geo,                    ONLY: geoopt_lbfgs
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_rm_iter_level
   USE force_env_types,                 ONLY: force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_methods,                  ONLY: gopt_f_create_x0
   USE gopt_f_types,                    ONLY: gopt_f_create,&
                                              gopt_f_release,&
                                              gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_read,&
                                              gopt_param_release,&
                                              gopt_param_type
   USE input_constants,                 ONLY: default_bfgs_method_id,&
                                              default_cg_method_id,&
                                              default_lbfgs_method_id
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'geo_opt'

   PUBLIC :: cp_geo_opt, cp_rot_opt

CONTAINS

! **************************************************************************************************
!> \brief Main driver to perform geometry optimization
!> \param force_env ...
!> \param globenv ...
!> \param eval_opt_geo ...
!> \param rm_restart_info ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_geo_opt(force_env, globenv, eval_opt_geo, rm_restart_info)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      LOGICAL, INTENT(IN), OPTIONAL                      :: eval_opt_geo, rm_restart_info

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_geo_opt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, step_start_val
      LOGICAL                                            :: my_rm_restart_info
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(section_vals_type), POINTER                   :: force_env_section, geo_section, &
                                                            root_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/geo_opt.F",75)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/geo_opt.F",76)
      NULLIFY (gopt_param, force_env_section, gopt_env, x0)
      root_section => force_env%root_section
      force_env_section => force_env%force_env_section
      geo_section => section_vals_get_subs_vals(root_section, "MOTION%GEO_OPT")

      CALL gopt_param_read(gopt_param, geo_section)
      CALL gopt_f_create(gopt_env, gopt_param, force_env=force_env, globenv=globenv, &
                         geo_opt_section=geo_section, eval_opt_geo=eval_opt_geo)
      CALL gopt_f_create_x0(gopt_env, x0)

      CALL section_vals_val_get(geo_section, "STEP_START_VAL", i_val=step_start_val)
      CALL cp_add_iter_level(logger%iter_info, "GEO_OPT")
      CALL cp_iterate(logger%iter_info, iter_nr=step_start_val)
      CALL cp_geo_opt_low(force_env, globenv, gopt_param, gopt_env, &
                          force_env_section, geo_section, x0)
      CALL cp_rm_iter_level(logger%iter_info, "GEO_OPT")

      ! Reset counter for next iteration, unless rm_restart_info==.FALSE.
      my_rm_restart_info = .TRUE.
      IF (PRESENT(rm_restart_info)) my_rm_restart_info = rm_restart_info
      IF (my_rm_restart_info) &
         CALL section_vals_val_set(geo_section, "STEP_START_VAL", i_val=0)

      DEALLOCATE (x0)
      CALL gopt_f_release(gopt_env)
      CALL gopt_param_release(gopt_param)
      CALL timestop(handle)

   END SUBROUTINE cp_geo_opt

! **************************************************************************************************
!> \brief Main driver to perform rotation optimization for Dimer
!> \param gopt_env ...
!> \param x0 ...
!> \param gopt_param ...
!> \param geo_section ...
! **************************************************************************************************
   SUBROUTINE cp_rot_opt(gopt_env, x0, gopt_param, geo_section)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(section_vals_type), POINTER                   :: geo_section

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_rot_opt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, step_start_val
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CALL timeset(routineN, handle)
      NULLIFY (force_env_section)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(gopt_env)))CALL cp__a("motion/geo_opt.F",129)
      IF(.NOT.(ASSOCIATED(gopt_env%force_env)))CALL cp__a("motion/geo_opt.F",130)
      IF(.NOT.(ASSOCIATED(gopt_env%globenv)))CALL cp__a("motion/geo_opt.F",131)
      force_env_section => gopt_env%force_env%force_env_section

      CALL section_vals_val_get(geo_section, "STEP_START_VAL", i_val=step_start_val)
      CALL cp_add_iter_level(logger%iter_info, "ROT_OPT")
      CALL cp_iterate(logger%iter_info, iter_nr=step_start_val)
      CALL cp_geo_opt_low(gopt_env%force_env, gopt_env%globenv, gopt_param, gopt_env, &
                          force_env_section, geo_section, x0)
      CALL cp_rm_iter_level(logger%iter_info, "ROT_OPT")

      ! Reset counter for next iteration
      CALL section_vals_val_set(geo_section, "STEP_START_VAL", i_val=0)
      CALL timestop(handle)

   END SUBROUTINE cp_rot_opt

! **************************************************************************************************
!> \brief call to low level geometry optimizers
!> \param force_env ...
!> \param globenv ...
!> \param gopt_param ...
!> \param gopt_env ...
!> \param force_env_section ...
!> \param geo_section ...
!> \param x0 ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_geo_opt_low(force_env, globenv, gopt_param, gopt_env, force_env_section, &
                                       geo_section, x0)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, geo_section
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_geo_opt_low', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/geo_opt.F",168)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/geo_opt.F",169)
      IF(.NOT.(ASSOCIATED(gopt_param)))CALL cp__a("motion/geo_opt.F",170)
      IF(.NOT.(ASSOCIATED(gopt_env)))CALL cp__a("motion/geo_opt.F",171)
      IF(.NOT.(ASSOCIATED(x0)))CALL cp__a("motion/geo_opt.F",172)
      IF(.NOT.(ASSOCIATED(force_env_section)))CALL cp__a("motion/geo_opt.F",173)
      IF(.NOT.(ASSOCIATED(geo_section)))CALL cp__a("motion/geo_opt.F",174)

      SELECT CASE (gopt_param%method_id)
      CASE (default_bfgs_method_id)
         CALL geoopt_bfgs(force_env, gopt_param, globenv, &
                          geo_section, gopt_env, x0)
      CASE (default_lbfgs_method_id)
         CALL geoopt_lbfgs(force_env, gopt_param, globenv, &
                           geo_section, gopt_env, x0)
      CASE (default_cg_method_id)
         CALL geoopt_cg(force_env, gopt_param, globenv, &
                        geo_section, gopt_env, x0)
      CASE DEFAULT
         CALL cp__b("motion/geo_opt.F",187,"")
      END SELECT

   END SUBROUTINE cp_geo_opt_low

END MODULE geo_opt
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE optimize_input
   USE cell_types,                      ONLY: parse_cell_line
   USE cp2k_info,                       ONLY: write_restart_header
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_read_line
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE environment,                     ONLY: cp2k_get_walltime
   USE f77_interface,                   ONLY: calc_force,&
                                              create_force_env,&
                                              destroy_force_env,&
                                              set_cell
   USE input_constants,                 ONLY: opt_force_matching
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set,&
                                              section_vals_write
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_comm_free,&
                                              mp_comm_split,&
                                              mp_comm_split_direct,&
                                              mp_environ,&
                                              mp_sum
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type
   USE physcon,                         ONLY: bohr
   USE powell,                          ONLY: opt_state_type,&
                                              powell_optimize
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE
  PRIVATE

  PUBLIC ::  run_optimize_input

  TYPE fm_env_type
     CHARACTER(LEN=default_path_length) :: optimize_file_name

     CHARACTER(LEN=default_path_length) :: ref_traj_file_name
     CHARACTER(LEN=default_path_length) :: ref_force_file_name
     CHARACTER(LEN=default_path_length) :: ref_cell_file_name

     INTEGER :: group_size

     REAL(KIND=dp) :: energy_weight
     REAL(KIND=dp) :: shift_mm
     REAL(KIND=dp) :: shift_qm
     LOGICAL       :: shift_average
     INTEGER :: frame_start,frame_stop,frame_stride,frame_count
  END TYPE

  TYPE variable_type
     CHARACTER(LEN=default_string_length) :: label
     REAL(KIND=dp)                        :: value
     LOGICAL                              :: fixed
  END TYPE

  TYPE oi_env_type
     INTEGER :: method
     INTEGER :: seed
     CHARACTER(LEN=default_path_length) :: project_name
     TYPE(fm_env_type) :: fm_env
     TYPE(variable_type), DIMENSION(:), ALLOCATABLE :: variables
     REAL(KIND=dp) :: rhobeg,rhoend
     INTEGER       :: maxfun
     INTEGER       :: iter_start_val
     REAL(KIND=dp) :: randomize_variables
     REAL(KIND=dp) :: start_time,target_time
  END TYPE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'optimize_input'

CONTAINS

! **************************************************************************************************
!> \brief main entry point for methods aimed at optimizing parameters in a CP2K input file
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \author Joost VandeVondele
! **************************************************************************************************
  SUBROUTINE run_optimize_input(input_declaration, root_section, para_env)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'run_optimize_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_var
      REAL(KIND=dp)                                      :: random_number, seed(3,2)
      TYPE(oi_env_type)                                  :: oi_env
      TYPE(rng_stream_type), POINTER                     :: rng_stream

    CALL timeset(routineN,handle)

    oi_env%start_time=m_walltime()

    CALL parse_input(oi_env,root_section)

    ! if we have been asked to randomize the variables, we do this.
    IF (oi_env%randomize_variables.NE.0.0_dp) THEN
       NULLIFY(rng_stream)
       seed=REAL(oi_env%seed,KIND=dp)
       CALL create_rng_stream(rng_stream,"run_optimize_input",distribution_type=UNIFORM,seed=seed)
       DO i_var=1,SIZE(oi_env%variables,1)
          IF (.NOT.oi_env%variables(i_var)%fixed) THEN
             ! change with a random percentage the variable
             random_number   = next_random_number(rng_stream)
             oi_env%variables(i_var)%value=oi_env%variables(i_var)%value* &
                            (1.0_dp+(2*random_number-1.0_dp)*oi_env%randomize_variables/100.0_dp)
          ENDIF
       ENDDO
       CALL delete_rng_stream(rng_stream)
    ENDIF

    ! proceed to actual methods
    SELECT CASE(oi_env%method)
    CASE(opt_force_matching)
      CALL force_matching(oi_env,input_declaration,root_section,para_env)
    CASE DEFAULT
      CALL cp__b("optimize_input.F",148,"")
    END SELECT

    CALL timestop(handle)

  END SUBROUTINE run_optimize_input

! **************************************************************************************************
!> \brief optimizes parameters by force/energy matching results against reference values
!> \param oi_env ...
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \author Joost VandeVondele
! **************************************************************************************************
  SUBROUTINE force_matching(oi_env,input_declaration,root_section,para_env)
      TYPE(oi_env_type)                                  :: oi_env
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_matching', routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: input_path, output_path
      CHARACTER(len=default_string_length), &
         ALLOCATABLE, DIMENSION(:, :)                    :: initial_variables
      INTEGER :: color, energies_unit, handle, history_unit, i_atom, i_el, i_frame, i_free_var, &
         i_var, ierr, mepos_master, mepos_slave, mpi_comm_master, mpi_comm_slave, &
         mpi_comm_slave_primus, n_atom, n_el, n_frames, n_free_var, n_groups, n_var, new_env_id, &
         num_pe_master, num_pe_slave, output_unit, restart_unit, state
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: free_var_index
      INTEGER, DIMENSION(:), POINTER                     :: group_distribution
      LOGICAL                                            :: should_stop
      REAL(KIND=dp)                                      :: e1, e2, e3, e4, e_pot, energy_weight, &
                                                            re, rf, shift_mm, shift_qm, t1, t2, &
                                                            t3, t4, t5
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: force, free_vars, pos
      REAL(KIND=dp), DIMENSION(:), POINTER               :: energy_traj, energy_traj_read, energy_var
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: cell_traj, cell_traj_read, force_traj, &
                                                            force_traj_read, force_var, pos_traj, &
                                                            pos_traj_read
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(opt_state_type)                               :: ostate
      TYPE(section_vals_type), POINTER                   :: oi_section, variable_section

    CALL timeset(routineN,handle)

    logger=>cp_get_default_logger()
    CALL cp_add_iter_level(logger%iter_info,"POWELL_OPT")
    output_unit = cp_logger_get_default_io_unit(logger)

    IF (output_unit>0) THEN
       WRITE(output_unit,'(T2,A)') 'FORCE_MATCHING| good morning....'
    ENDIF

    ! do IO of ref traj / frc / cell
    NULLIFY(cell_traj_read,force_traj_read,pos_traj_read,energy_traj_read)
    CALL read_reference_data(oi_env,para_env,force_traj_read,pos_traj_read,energy_traj_read,cell_traj_read)
    n_atom=SIZE(pos_traj_read,2)

    ! adjust read data with respect to start/stop/stride
    IF (oi_env%fm_env%frame_stop<0) oi_env%fm_env%frame_stop=SIZE(pos_traj_read,3)

    IF (oi_env%fm_env%frame_count>0) THEN
       oi_env%fm_env%frame_stride=(oi_env%fm_env%frame_stop-oi_env%fm_env%frame_start+1+ &
                                   oi_env%fm_env%frame_count-1)/(oi_env%fm_env%frame_count)
    ENDIF
    n_frames=(oi_env%fm_env%frame_stop-oi_env%fm_env%frame_start+oi_env%fm_env%frame_stride)/oi_env%fm_env%frame_stride

    ALLOCATE(force_traj(3,n_atom,n_frames),pos_traj(3,n_atom,n_frames),energy_traj(n_frames))
    IF (ASSOCIATED(cell_traj_read)) ALLOCATE(cell_traj(3,3,n_frames))

    n_frames=0
    DO i_frame=oi_env%fm_env%frame_start,oi_env%fm_env%frame_stop,oi_env%fm_env%frame_stride
       n_frames=n_frames+1
       force_traj(:,:,n_frames)=force_traj_read(:,:,i_frame)
       pos_traj(:,:,n_frames)=pos_traj_read(:,:,i_frame)
       energy_traj(n_frames)=energy_traj_read(i_frame)
       IF (ASSOCIATED(cell_traj)) cell_traj(:,:,n_frames)=cell_traj_read(:,:,i_frame)
    ENDDO
    DEALLOCATE(force_traj_read,pos_traj_read,energy_traj_read)
    IF (ASSOCIATED(cell_traj_read)) DEALLOCATE(cell_traj_read)

    n_el=3*n_atom
    ALLOCATE(pos(n_el), force(n_el))
    ALLOCATE(energy_var(n_frames),force_var(3,n_atom,n_frames))


    ! split the para_env in a set of sub_para_envs that will do the force_env communications
    mpi_comm_master=para_env%group
    num_pe_master  =para_env%num_pe
    mepos_master   =para_env%mepos
    ALLOCATE(group_distribution(0:num_pe_master-1))
    IF (oi_env%fm_env%group_size>para_env%num_pe) oi_env%fm_env%group_size=para_env%num_pe

    CALL mp_comm_split(mpi_comm_master,mpi_comm_slave,n_groups,group_distribution,subgroup_min_size=oi_env%fm_env%group_size)
    CALL mp_environ(num_pe_slave,mepos_slave,mpi_comm_slave)
    color=0
    IF (mepos_slave==0) color=1
    CALL mp_comm_split_direct(mpi_comm_master,mpi_comm_slave_primus,color)

    ! assign initial variables
    n_var=SIZE(oi_env%variables,1)
    ALLOCATE(initial_variables(2,n_var))
    n_free_var=0
    DO i_var=1,n_var
       initial_variables(1,i_var)=oi_env%variables(i_var)%label
       WRITE(initial_variables(2,i_var),*) oi_env%variables(i_var)%value
       IF (.NOT.oi_env%variables(i_var)%fixed) n_free_var=n_free_var+1
    ENDDO
    ALLOCATE(free_vars(n_free_var),free_var_index(n_free_var))
    i_free_var=0
    DO i_var=1,n_var
       IF (.NOT.oi_env%variables(i_var)%fixed) THEN
          i_free_var=i_free_var+1
          free_var_index(i_free_var)=i_var
          free_vars(i_free_var)=oi_env%variables(free_var_index(i_free_var))%value
       ENDIF
    ENDDO

    ! create input and output file names.
    input_path=oi_env%fm_env%optimize_file_name
    WRITE(output_path,'(A,I0,A)') TRIM(oi_env%project_name)//"-worker-",group_distribution(mepos_master),".out"

    ! initialize the powell optimizer
    energy_weight=oi_env%fm_env%energy_weight
    shift_mm=oi_env%fm_env%shift_mm
    shift_qm=oi_env%fm_env%shift_qm

    IF (para_env%mepos==para_env%source) THEN
       ostate%nf = 0
       ostate%nvar = n_free_var
       ostate%rhoend = oi_env%rhoend
       ostate%rhobeg = oi_env%rhobeg
       ostate%maxfun = oi_env%maxfun
       ostate%iprint = 1
       ostate%unit = output_unit
       ostate%state = 0
    ENDIF

    IF (output_unit>0) THEN
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| number of atoms per frame ',n_atom
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| number of frames ',n_frames
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| number of parallel groups ',n_groups
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| number of variables ',n_var
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| number of free variables ',n_free_var
       WRITE(output_unit,'(T2,A,A)')          'FORCE_MATCHING| optimize file name ',TRIM(input_path)
       WRITE(output_unit,'(T2,A,T60,F20.12)') 'FORCE_MATCHING| accuracy',ostate%rhoend
       WRITE(output_unit,'(T2,A,T60,F20.12)') 'FORCE_MATCHING| step size',ostate%rhobeg
       WRITE(output_unit,'(T2,A,T60,I20)')    'FORCE_MATCHING| max function evaluation',ostate%maxfun
       WRITE(output_unit,'(T2,A,T60,L20)')    'FORCE_MATCHING| shift average',oi_env%fm_env%shift_average
       WRITE(output_unit,'(T2,A)')            'FORCE_MATCHING| initial values:'
       DO i_var=1,n_var
          WRITE(output_unit,'(T2,A,1X,E28.16)') TRIM(oi_env%variables(i_var)%label),oi_env%variables(i_var)%value
       ENDDO
       WRITE(output_unit,'(T2,A)')            'FORCE_MATCHING| switching to POWELL optimization of the free parameters'
       WRITE(output_unit,'()')
       WRITE(output_unit,'(T2,A20,A20,A11,A11)')   'iteration number','function value','time','time Force'
       CALL m_flush(output_unit)
    ENDIF


    t1 = m_walltime()

    DO


      ! globalize the state
      IF (para_env%mepos==para_env%source) state=ostate%state
      CALL mp_bcast(state,para_env%source,para_env%group)

      ! if required get the energy of this set of params
      IF (state == 2) THEN

         CALL cp_iterate(logger%iter_info,last=.FALSE.)

         ! create a new force env, updating the free vars as needed
         DO i_free_var=1,n_free_var
            WRITE(initial_variables(2,free_var_index(i_free_var)),*) free_vars(i_free_var)
            oi_env%variables(free_var_index(i_free_var))%value=free_vars(i_free_var)
         ENDDO

         ierr=0
         CALL create_force_env(new_env_id=new_env_id, input_declaration=input_declaration,&
                               input_path=input_path, output_path=output_path, &
                               mpi_comm=mpi_comm_slave,initial_variables=initial_variables,ierr=ierr)

         ! set to zero initialy, for easier mp_summing
         energy_var=0.0_dp
         force_var=0.0_dp

         ! compute energies and forces for all frames, doing the work on a slave sub group based on round robin
         t5 = 0.0_dp
         DO i_frame=group_distribution(mepos_master)+1,n_frames,n_groups

            ! set new cell if needed
            IF (ASSOCIATED(cell_traj)) THEN
               CALL set_cell(env_id=new_env_id, new_cell=cell_traj(:,:,i_frame), ierr=ierr)
            ENDIF

            ! copy pos from ref
            i_el=0
            DO i_atom=1,n_atom
               pos(i_el+1)=pos_traj(1,i_atom,i_frame)
               pos(i_el+2)=pos_traj(2,i_atom,i_frame)
               pos(i_el+3)=pos_traj(3,i_atom,i_frame)
               i_el=i_el+3
            ENDDO

            ! evaluate energy/force with new pos
            t3 = m_walltime()
            CALL calc_force(env_id=new_env_id,pos=pos,n_el_pos=n_el,e_pot=e_pot,force=force,n_el_force=n_el,ierr=ierr)
            t4 = m_walltime()
            t5 = t5 + t4-t3

            ! copy force and energy in place
            energy_var(i_frame)=e_pot
            i_el=0
            DO i_atom=1,n_atom
               force_var(1,i_atom,i_frame)=force(i_el+1)
               force_var(2,i_atom,i_frame)=force(i_el+2)
               force_var(3,i_atom,i_frame)=force(i_el+3)
               i_el=i_el+3
            ENDDO

         ENDDO

         ! clean up force env, get ready for the next round
         CALL destroy_force_env(env_id=new_env_id,ierr=ierr)

         ! get data everywhere on the master group, we could reduce the amount of data by reducing to partial RMSD first
         ! furthermore, we should only do this operation among the masters of the slave group
         IF (mepos_slave==0) THEN
             CALL mp_sum(energy_var,mpi_comm_slave_primus)
             CALL mp_sum(force_var,mpi_comm_slave_primus)
         ENDIF

         ! now evaluate the target function to be minimized (only valid on mepos_slave==0)
         IF (para_env%mepos==para_env%source) THEN
             rf = SQRT(SUM((force_var-force_traj)**2)/(REAL(n_frames,KIND=dp) * REAL(n_atom,KIND=dp)))
             IF (oi_env%fm_env%shift_average) THEN
                shift_mm=SUM(energy_var)/n_frames
                shift_qm=SUM(energy_traj)/n_frames
             ENDIF
             re = SQRT(SUM(((energy_var-shift_mm)-(energy_traj-shift_qm))**2)/n_frames)
             ostate%f= energy_weight * re + rf
             t2 = m_walltime()
             WRITE(output_unit,'(T2,I20,F20.12,F11.3,F11.3)') oi_env%iter_start_val + ostate%nf, ostate%f, t2-t1, t5
             CALL m_flush(output_unit)
             t1 = m_walltime()
         ENDIF

         ! the history file with the trajectory of the parameters
         history_unit=cp_print_key_unit_nr(logger,root_section,"OPTIMIZE_INPUT%HISTORY", &
                                           extension=".dat")
         IF (history_unit>0) THEN
            WRITE (UNIT=history_unit,FMT="(I20,F20.12,1000F20.12)") oi_env%iter_start_val + ostate%nf, ostate%f, free_vars
         END IF
         CALL cp_print_key_finished_output(history_unit,logger,root_section,"OPTIMIZE_INPUT%HISTORY")

         ! the energy profile for all frames
         energies_unit=cp_print_key_unit_nr(logger,root_section,"OPTIMIZE_INPUT%FORCE_MATCHING%COMPARE_ENERGIES", &
                                           file_position="REWIND",extension=".dat")
         IF (energies_unit>0) THEN
            WRITE (UNIT=energies_unit,FMT="(A20,A20,A20,A20)") "#frame","ref","fit","diff"
            DO i_frame=1,n_frames
               e1=energy_traj(i_frame)-shift_qm
               e2=energy_var(i_frame)-shift_mm
               WRITE (UNIT=energies_unit,FMT="(I20,F20.12,F20.12,F20.12)") i_frame,e1,e2,e1-e2
            ENDDO
         END IF
         CALL cp_print_key_finished_output(energies_unit,logger,root_section,"OPTIMIZE_INPUT%FORCE_MATCHING%COMPARE_ENERGIES")

         ! the force profile for all frames
         energies_unit=cp_print_key_unit_nr(logger,root_section,"OPTIMIZE_INPUT%FORCE_MATCHING%COMPARE_FORCES", &
                                           file_position="REWIND",extension=".dat")
         IF (energies_unit>0) THEN
            WRITE (UNIT=energies_unit,FMT="(A20,A20,A20,A20)") "#frame","normalized diff","diff","ref","ref sum"
            DO i_frame=1,n_frames
               e1=SQRT(SUM((force_var(:,:,i_frame)-force_traj(:,:,i_frame))**2))
               e2=SQRT(SUM((force_traj(:,:,i_frame))**2))
               e3=SQRT(SUM(SUM(force_traj(:,:,i_frame),DIM=2)**2))
               e4=SQRT(SUM(SUM(force_var(:,:,i_frame),DIM=2)**2))
               WRITE (UNIT=energies_unit,FMT="(I20,F20.12,F20.12,F20.12,2F20.12)") i_frame,e1/e2,e1,e2,e3,e4
            ENDDO
         END IF
         CALL cp_print_key_finished_output(energies_unit,logger,root_section,"OPTIMIZE_INPUT%FORCE_MATCHING%COMPARE_FORCES")


         ! a restart file with the current values of the parameters
         restart_unit = cp_print_key_unit_nr(logger,root_section,"OPTIMIZE_INPUT%RESTART", extension=".restart",&
                                     file_position="REWIND", do_backup=.TRUE.)
         IF (restart_unit>0) THEN
            oi_section => section_vals_get_subs_vals(root_section,"OPTIMIZE_INPUT")
            CALL section_vals_val_set(oi_section,"ITER_START_VAL",i_val=oi_env%iter_start_val + ostate%nf)
            variable_section  => section_vals_get_subs_vals(oi_section,"VARIABLE")
            DO i_free_var=1,n_free_var
               CALL section_vals_val_set(variable_section,"VALUE",i_rep_section=free_var_index(i_free_var), &
                                         r_val=free_vars(i_free_var))
            ENDDO
            CALL write_restart_header(restart_unit)
            CALL section_vals_write(root_section,unit_nr=restart_unit,hide_root=.TRUE.)
         ENDIF
         CALL cp_print_key_finished_output(restart_unit,logger,root_section,"OPTIMIZE_INPUT%RESTART")

      ENDIF

      IF ( state == -1 ) EXIT

      CALL external_control(should_stop,"OPTIMIZE_INPUT",target_time=oi_env%target_time,start_time=oi_env%start_time)

      IF (should_stop) EXIT

      ! do a powell step if needed
      IF (para_env%mepos==para_env%source) THEN
         CALL powell_optimize (ostate%nvar, free_vars , ostate)
      ENDIF
      CALL mp_bcast(free_vars,para_env%source,para_env%group)

    ENDDO

    ! finally, get the best set of variables
    IF (para_env%mepos==para_env%source) THEN
        ostate%state = 8
       CALL powell_optimize (ostate%nvar, free_vars , ostate)
    ENDIF
    CALL mp_bcast(free_vars,para_env%source,para_env%group)
    DO i_free_var=1,n_free_var
       WRITE(initial_variables(2,free_var_index(i_free_var)),*) free_vars(i_free_var)
       oi_env%variables(free_var_index(i_free_var))%value=free_vars(i_free_var)
    ENDDO
    IF (para_env%mepos==para_env%source) THEN
       WRITE(output_unit,'(T2,A)')         ''
       WRITE(output_unit,'(T2,A,T60,F20.12)')         'FORCE_MATCHING| optimal function value found so far:',ostate%fopt
       WRITE(output_unit,'(T2,A)')         'FORCE_MATCHING| optimal variables found so far:'
       DO i_var=1,n_var
          WRITE(output_unit,'(T2,A,1X,E28.16)') TRIM(oi_env%variables(i_var)%label),oi_env%variables(i_var)%value
       ENDDO
    ENDIF

    CALL cp_rm_iter_level(logger%iter_info,"POWELL_OPT")

    ! deallocate for cleanup
    IF (ASSOCIATED(cell_traj)) DEALLOCATE(cell_traj)
    DEALLOCATE(pos,force,force_traj, pos_traj, force_var)
    DEALLOCATE(group_distribution,energy_traj, energy_var)
    CALL mp_comm_free(mpi_comm_slave)
    CALL mp_comm_free(mpi_comm_slave_primus)

    CALL timestop(handle)

  END SUBROUTINE force_matching

! **************************************************************************************************
!> \brief reads the reference data for force matching results, the format of the files needs to be the CP2K xyz trajectory format
!> \param oi_env ...
!> \param para_env ...
!> \param force_traj forces 
!> \param pos_traj position 
!> \param energy_traj energies, as extracted from the forces file
!> \param cell_traj cell parameters, as extracted from a CP2K cell file 
!> \author Joost VandeVondele
! **************************************************************************************************
  SUBROUTINE read_reference_data(oi_env,para_env,force_traj,pos_traj,energy_traj,cell_traj)
      TYPE(oi_env_type)                                  :: oi_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: force_traj, pos_traj
      REAL(KIND=dp), DIMENSION(:), POINTER               :: energy_traj
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: cell_traj

      CHARACTER(len=*), PARAMETER :: routineN = 'read_reference_data', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: filename
      CHARACTER(len=default_string_length)               :: AA
      INTEGER                                            :: cell_itimes, handle, i, iframe, &
                                                            n_frames, n_frames_current, nread, &
                                                            trj_itimes
      LOGICAL                                            :: at_end, test_ok
      REAL(KIND=dp)                                      :: cell_time, trj_epot, trj_time, vec(3), &
                                                            vol
      TYPE(cp_parser_type), POINTER                      :: local_parser

    CALL timeset(routineN,handle)

    ! do IO of ref traj / frc / cell

    ! trajectory
    n_frames=0
    n_frames_current=0
    NULLIFY(local_parser,pos_traj,energy_traj,force_traj)
    filename=oi_env%fm_env%ref_traj_file_name
    IF(filename.EQ."")&
       CALL cp__b("optimize_input.F",541,"The reference trajectory file name is empty")
    CALL parser_create(local_parser,filename,para_env=para_env)
    DO
       CALL parser_read_line(local_parser,1,at_end=at_end)
       IF (at_end) EXIT
       READ(local_parser%input_line,FMT="(I8)") nread
       n_frames=n_frames+1

       IF (n_frames>n_frames_current) THEN
           n_frames_current=5*(n_frames_current+10)/3
           CALL reallocate(pos_traj,1,3,1,nread,1,n_frames_current)
       ENDIF

       ! title line
       CALL parser_read_line(local_parser,1)

       ! actual coordinates
       DO i = 1,nread
          CALL parser_read_line(local_parser,1)
          READ(local_parser%input_line(1:LEN_TRIM(local_parser%input_line)),*) AA,vec
          pos_traj(:,i,n_frames)=vec*bohr
       END DO

    ENDDO
    CALL parser_release(local_parser)

    n_frames_current=n_frames
    CALL reallocate(energy_traj,1,n_frames_current)
    CALL reallocate(force_traj,1,3,1,nread,1,n_frames_current)
    CALL reallocate(pos_traj,1,3,1,nread,1,n_frames_current)

    ! now force reference trajectory
    filename=oi_env%fm_env%ref_force_file_name
    IF(filename.EQ."")&
       CALL cp__b("optimize_input.F",575,"The reference force file name is empty")
    CALL parser_create(local_parser,filename,para_env=para_env)
    DO iframe=1,n_frames
       CALL parser_read_line(local_parser,1)
       READ(local_parser%input_line,FMT="(I8)") nread

       ! title line
       test_ok = .FALSE.
       CALL parser_read_line(local_parser,1)
       READ(local_parser%input_line,FMT="(T6,I8,T23,F12.3,T41,F20.10)",ERR=999) trj_itimes, trj_time, trj_epot
       test_ok = .TRUE.
999    CONTINUE
       IF (.NOT. test_ok) THEN
           CALL cp__b("optimize_input.F",588,"Could not parse the title line of the trajectory file")
       END IF
       energy_traj(iframe)=trj_epot

       ! actual forces, in a.u.
       DO i = 1,nread
          CALL parser_read_line(local_parser,1)
          READ(local_parser%input_line(1:LEN_TRIM(local_parser%input_line)),*) AA,vec
          force_traj(:,i,iframe)=vec
       END DO
    ENDDO
    CALL parser_release(local_parser)


    ! and cell, which is optional
    NULLIFY(cell_traj)
    filename=oi_env%fm_env%ref_cell_file_name
    IF (filename.NE."") THEN
       CALL parser_create(local_parser,filename,para_env=para_env)
       ALLOCATE(cell_traj(3,3,n_frames))
       DO iframe=1,n_frames
          CALL parser_read_line(local_parser,1)
          CALL parse_cell_line(local_parser%input_line, cell_itimes, cell_time, cell_traj(:,:,iframe), vol)
       ENDDO
       CALL parser_release(local_parser)
    ENDIF

    CALL timestop(handle)

  END SUBROUTINE read_reference_data

! **************************************************************************************************
!> \brief parses the input section, and stores in the optimize input environment
!> \param oi_env optimize input environment
!> \param root_section ...
!> \author Joost VandeVondele
! **************************************************************************************************
  SUBROUTINE parse_input(oi_env,root_section)
      TYPE(oi_env_type)                                  :: oi_env
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ivar, n_var
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: fm_section, oi_section, variable_section

    CALL timeset(routineN,handle)

    CALL section_vals_val_get(root_section,"GLOBAL%PROJECT",c_val=oi_env%project_name)
    CALL section_vals_val_get(root_section,"GLOBAL%SEED",i_val=oi_env%seed)
    CALL cp2k_get_walltime(section=root_section, keyword_name="GLOBAL%WALLTIME", &
                             walltime=oi_env%target_time)

    oi_section => section_vals_get_subs_vals(root_section,"OPTIMIZE_INPUT")
    variable_section  => section_vals_get_subs_vals(oi_section,"VARIABLE")

    CALL section_vals_val_get(oi_section,"ACCURACY", r_val=oi_env%rhoend)
    CALL section_vals_val_get(oi_section,"STEP_SIZE", r_val=oi_env%rhobeg)
    CALL section_vals_val_get(oi_section,"MAX_FUN", i_val=oi_env%maxfun)
    CALL section_vals_val_get(oi_section,"ITER_START_VAL", i_val=oi_env%iter_start_val)
    CALL section_vals_val_get(oi_section,"RANDOMIZE_VARIABLES", r_val=oi_env%randomize_variables)

    CALL section_vals_get(variable_section,explicit=explicit, n_repetition=n_var)
    IF (explicit) THEN
       ALLOCATE(oi_env%variables(1:n_var))
       DO ivar=1,n_var
          CALL section_vals_val_get(variable_section,"VALUE",i_rep_section=ivar, &
                r_val=oi_env%variables(ivar)%value)
          CALL section_vals_val_get(variable_section,"FIXED",i_rep_section=ivar, &
                l_val=oi_env%variables(ivar)%fixed)
          CALL section_vals_val_get(variable_section,"LABEL",i_rep_section=ivar, &
                c_val=oi_env%variables(ivar)%label)
       ENDDO
    ENDIF

    CALL section_vals_val_get(oi_section,"METHOD",i_val=oi_env%method)
    SELECT CASE(oi_env%method)
    CASE(opt_force_matching)
       fm_section => section_vals_get_subs_vals(oi_section,"FORCE_MATCHING")
       CALL section_vals_val_get(fm_section,"REF_TRAJ_FILE_NAME",c_val=oi_env%fm_env%ref_traj_file_name)
       CALL section_vals_val_get(fm_section,"REF_FORCE_FILE_NAME",c_val=oi_env%fm_env%ref_force_file_name)
       CALL section_vals_val_get(fm_section,"REF_CELL_FILE_NAME",c_val=oi_env%fm_env%ref_cell_file_name)
       CALL section_vals_val_get(fm_section,"OPTIMIZE_FILE_NAME",c_val=oi_env%fm_env%optimize_file_name)
       CALL section_vals_val_get(fm_section,"FRAME_START",i_val=oi_env%fm_env%frame_start)
       CALL section_vals_val_get(fm_section,"FRAME_STOP",i_val=oi_env%fm_env%frame_stop)
       CALL section_vals_val_get(fm_section,"FRAME_STRIDE",i_val=oi_env%fm_env%frame_stride)
       CALL section_vals_val_get(fm_section,"FRAME_COUNT",i_val=oi_env%fm_env%frame_count)

       CALL section_vals_val_get(fm_section,"GROUP_SIZE",i_val=oi_env%fm_env%group_size)

       CALL section_vals_val_get(fm_section,"ENERGY_WEIGHT",r_val=oi_env%fm_env%energy_weight)
       CALL section_vals_val_get(fm_section,"SHIFT_MM",r_val=oi_env%fm_env%shift_mm)
       CALL section_vals_val_get(fm_section,"SHIFT_QM",r_val=oi_env%fm_env%shift_qm)
       CALL section_vals_val_get(fm_section,"SHIFT_AVERAGE",l_val=oi_env%fm_env%shift_average)
    CASE DEFAULT
      CALL cp__b("optimize_input.F",684,"")
    END SELECT

    CALL timestop(handle)

  END SUBROUTINE parse_input

END MODULE optimize_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Public path integral routines that can be called from other modules
!> \author Lukasz Walewski
!> \date   2009-07-24
!> \note   Avoiding circular dependencies: please design new members of this
!>         module in such a way that they use pint_types module only.
! **************************************************************************************************
MODULE pint_public

   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE pint_types,                      ONLY: pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_public'

   PUBLIC :: pint_com_pos
   PUBLIC :: pint_levy_walk
   PUBLIC :: pint_calc_centroid

CONTAINS

! ***************************************************************************
!> \brief  Return the center of mass of the PI system
!> \param pint_env ...
!> \retval com_r ...
!> \date   2009-07-24
!> \par    History
!>           2009-11-30 fixed serious bug in pint_env%x indexing [lwalewski]
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION pint_com_pos(pint_env) RESULT(com_r)

      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(3)                        :: com_r

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_com_pos', routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, ic
      REAL(kind=dp)                                      :: tmass

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_public.F",53)

      tmass = 0.0_dp
      com_r(:) = 0.0_dp
      DO ia = 1, pint_env%ndim/3
         DO ib = 1, pint_env%p
            DO ic = 1, 3
               com_r(ic) = com_r(ic)+ &
                           pint_env%x(ib, (ia-1)*3+ic)*pint_env%mass((ia-1)*3+ic)
               tmass = tmass+pint_env%mass((ia-1)*3+ic)
            END DO
         END DO
      END DO
      ! pint_env%mass is REAL, DIMENSION(NDIM) which means that each atom
      ! has it's mass defined three times - here we hope that all three
      ! values are equal
      tmass = tmass/3.0_dp
      com_r(:) = com_r(:)/tmass
      RETURN
   END FUNCTION pint_com_pos

! ***************************************************************************
!> \brief  Return the center of geometry of the PI system
!> \param pint_env ...
!> \retval cntrd_r ...
!> \date   2009-11-30
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION pint_cog_pos(pint_env) RESULT(cntrd_r)

      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(3)                        :: cntrd_r

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_cog_pos', routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, ic, natoms

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_public.F",90)

      cntrd_r(:) = 0.0_dp
      natoms = pint_env%ndim/3
      DO ia = 1, natoms
         DO ib = 1, pint_env%p
            DO ic = 1, 3
               cntrd_r(ic) = cntrd_r(ic)+pint_env%x(ib, (ia-1)*3+ic)
            END DO
         END DO
      END DO
      cntrd_r(:) = cntrd_r(:)/REAL(pint_env%p, dp)/REAL(natoms, dp)
      RETURN
   END FUNCTION pint_cog_pos

! ***************************************************************************
!> \brief  Given the number of beads n and the variance t returns the
!>         positions of the beads for a non-interacting particle.
!> \param n ...
!> \param t ...
!> \param rng_gaussian ...
!> \param x ...
!> \param nout ...
!> \date   2010-12-13
!> \author Lukasz Walewski
!> \note  This routine implements Levy argorithm (see e.g. Rev. Mod. Phys.
!>         67 (1995) 279, eq. 5.35) and requires that n is a power of 2. The
!>         resulting bead positions are centered around (0,0,0).
! **************************************************************************************************
   SUBROUTINE pint_free_part_bead_x(n, t, rng_gaussian, x, nout)
!
!TODO this routine gives wrong spread of the particles, please fix before usage.
!
      INTEGER, INTENT(IN)                                :: n
      REAL(kind=dp), INTENT(IN)                          :: t
      TYPE(rng_stream_type), POINTER                     :: rng_gaussian
      REAL(kind=dp), DIMENSION(:), POINTER               :: x
      INTEGER, INTENT(OUT)                               :: nout

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_free_part_bead_x', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dl, i1, i2, ib, ic, il, ip, j, nlevels, &
                                                            np
      REAL(kind=dp)                                      :: rtmp, tcheck, vrnc, xc
      REAL(kind=dp), DIMENSION(3)                        :: cntrd_r

      nout = 0

      IF (n < 1) THEN
         RETURN
      END IF

      ! if number of beads is not a power of 2 return
      nlevels = NINT(LOG(REAL(n, KIND=dp))/LOG(2.0_dp))
      rtmp = 2**nlevels
      tcheck = ABS(REAL(n, KIND=dp)-rtmp)
      IF (tcheck > 100.0_dp*EPSILON(0.0_dp)) THEN
         RETURN
      END IF

      ! generate at least first point at (0,0,0)
      DO ic = 1, 3
         x(ic) = 0.0_dp
      END DO
      nout = 1

      ! loop over Levy levels
      vrnc = 2.0_dp*t
      DO il = 0, nlevels-1

         np = 2**il ! number of points to be generated at this level
         dl = n/(2*np) ! interval betw points (in index numbers)
         vrnc = vrnc/2.0_dp; ! variance at this level (=t at level 0)

         ! loop over points added in this level
         DO ip = 0, np-1

            j = (2*ip+1)*dl ! index of currently generated point

            ! indices of two points betw which to generate a new point
            i1 = 2*dl*ip
            i2 = 2*dl*(ip+1)
            IF (i2 .EQ. n) THEN
               i2 = 0
            END IF

            ! generate new point and save it under j
            DO ic = 1, 3
               xc = (x(3*i1+ic)+x(3*i2+ic))/2.0
               xc = xc+next_random_number(rng_stream=rng_gaussian, &
                                          variance=vrnc)
               x(3*j+ic) = xc
            END DO
            nout = nout+1

         END DO
      END DO

      ! translate the centroid to the origin
      cntrd_r(:) = 0.0_dp
      DO ib = 1, n
         DO ic = 1, 3
            cntrd_r(ic) = cntrd_r(ic)+x((ib-1)*3+ic)
         END DO
      END DO
      cntrd_r(:) = cntrd_r(:)/REAL(n, dp)
      DO ib = 1, n
         DO ic = 1, 3
            x((ib-1)*3+ic) = x((ib-1)*3+ic)-cntrd_r(ic)
         END DO
      END DO

      RETURN
   END SUBROUTINE pint_free_part_bead_x

! ***************************************************************************
!> \brief  Perform a Brownian walk of length n around x0 with the variance v.
!> \param x0 ...
!> \param n ...
!> \param v ...
!> \param x ...
!> \param rng_gaussian ...
!> \date   2011-01-06
!> \author Lukasz Walewski
!> \note  This routine implements Levy argorithm (Phys. Rev. 143 (1966) 58)
! **************************************************************************************************
   SUBROUTINE pint_levy_walk(x0, n, v, x, rng_gaussian)

      REAL(kind=dp), DIMENSION(3), INTENT(IN)            :: x0
      INTEGER, INTENT(IN)                                :: n
      REAL(kind=dp), INTENT(IN)                          :: v
      REAL(kind=dp), DIMENSION(:), POINTER               :: x
      TYPE(rng_stream_type), POINTER                     :: rng_gaussian

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_levy_walk', routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, ic
      REAL(kind=dp)                                      :: r, tau_i, tau_i1
      REAL(kind=dp), DIMENSION(3)                        :: cntrd_r

      x(1) = x0(1)
      x(2) = x0(2)
      x(3) = x0(3)
      DO ib = 1, n-1
         DO ic = 1, 3
            r = next_random_number(rng_stream=rng_gaussian, &
                                   variance=1.0_dp)
            tau_i = (REAL(ib, dp)-1.0_dp)/REAL(n, dp)
            tau_i1 = (REAL(ib+1, dp)-1.0_dp)/REAL(n, dp)
            x(ib*3+ic) = (x((ib-1)*3+ic)*(1.0_dp-tau_i1)+ &
                          x(ic)*(tau_i1-tau_i))/ &
                         (1.0_dp-tau_i)+ &
                         r*v*SQRT( &
                         (tau_i1-tau_i)* &
                         (1.0_dp-tau_i1)/ &
                         (1.0_dp-tau_i) &
                         )
         END DO
      END DO

      ! translate the centroid to the origin
      cntrd_r(:) = 0.0_dp
      DO ib = 1, n
         DO ic = 1, 3
            cntrd_r(ic) = cntrd_r(ic)+x((ib-1)*3+ic)
         END DO
      END DO
      cntrd_r(:) = cntrd_r(:)/REAL(n, dp)
      DO ib = 1, n
         DO ic = 1, 3
            x((ib-1)*3+ic) = x((ib-1)*3+ic)-cntrd_r(ic)
         END DO
      END DO

      RETURN
   END SUBROUTINE pint_levy_walk

   ! ***************************************************************************
   !> \brief  Calculate the centroid
   !> \param  pint_env path integral environment
   !> \date   2013-02-10
   !> \author lwalewski
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_calc_centroid(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_centroid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib
      REAL(KIND=dp)                                      :: invp

      invp = 1.0_dp/pint_env%p
      pint_env%centroid(:) = 0.0_dp
      DO ia = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            pint_env%centroid(ia) = pint_env%centroid(ia)+pint_env%x(ib, ia)
         END DO
         pint_env%centroid(ia) = pint_env%centroid(ia)*invp
      END DO

   END SUBROUTINE pint_calc_centroid

END MODULE pint_public
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  I/O subroutines for pint_env
!> \author Lukasz Walewski
!> \date   2009-06-04
! **************************************************************************************************
MODULE pint_io

   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get
   USE input_constants,                 ONLY: dump_atomic,&
                                              dump_dcd,&
                                              dump_dcd_aligned_cell,&
                                              dump_xmol
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE pint_public,                     ONLY: pint_com_pos
   USE pint_transformations,            ONLY: pint_u2x
   USE pint_types,                      ONLY: e_conserved_id,&
                                              e_kin_thermo_id,&
                                              e_kin_virial_id,&
                                              e_potential_id,&
                                              pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_io'

   PUBLIC :: pint_write_line
   PUBLIC :: pint_write_centroids
   PUBLIC :: pint_write_trajectory
   PUBLIC :: pint_write_com
   PUBLIC :: pint_write_ener
   PUBLIC :: pint_write_step_info
   PUBLIC :: pint_write_rgyr

CONTAINS

! ***************************************************************************
!> \brief  Writes out a line of text to the default output unit.
!> \param line ...
!> \date   2009-07-10
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_write_line(line)

      CHARACTER(len=*), INTENT(IN)                       :: line

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_line', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_label
      INTEGER                                            :: unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      my_label = "PINT|"

      IF (logger%para_env%ionode) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger)
         WRITE (unit_nr, '(T2,A)') TRIM(my_label)//" "//TRIM(line)
      END IF

      RETURN
   END SUBROUTINE pint_write_line

! ***************************************************************************
!> \brief Write out the trajectory of the centroid (positions and velocities)
!> \param pint_env ...
!> \par History
!>      various bug fixes - hforbert
!>      2010-11-25 rewritten, added support for velocity printing,
!>                 calc of the stddev of the beads turned off [lwalewski]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_write_centroids(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_centroids', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: n_ids = 2, pos_id = 1, vel_id = 2

      CHARACTER(len=default_string_length)               :: ext, form, my_middle_name, unit_str
      CHARACTER(len=default_string_length), DIMENSION(2) :: content_id, middle_name, sect_path, title
      INTEGER                                            :: handle, handle1, iat, ib, id, idim, &
                                                            idir, ierr, outformat, should_output, &
                                                            unit_nr
      LOGICAL                                            :: new_file
      REAL(kind=dp)                                      :: nb, ss, unit_conv, vv
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle1)

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",127)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_io.F",128)

      sect_path(pos_id) = "MOTION%PINT%PRINT%CENTROID_POS"
      sect_path(vel_id) = "MOTION%PINT%PRINT%CENTROID_VEL"
      middle_name(pos_id) = "centroid-pos"
      middle_name(vel_id) = "centroid-vel"
      content_id(pos_id) = "POS"
      content_id(vel_id) = "VEL"
      WRITE (UNIT=title(pos_id), FMT="(A,I8,A,F20.10)") &
         " i =", pint_env%iter, &
         ", E =", SUM(pint_env%e_pot_bead)*pint_env%propagator%physpotscale
      WRITE (UNIT=title(vel_id), FMT="(A,I8,A,F20.10,A,F20.10)") &
         " i =", pint_env%iter, &
         ", E_trm =", pint_env%energy(e_kin_thermo_id), &
         ", E_vir =", pint_env%energy(e_kin_virial_id)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      CALL pint_u2x(pint_env, ux=pint_env%uv, x=pint_env%v)

      ! iterate over the properties that we know how to print
      ! (currently positions and velocities)
      DO id = 1, n_ids

         print_key => section_vals_get_subs_vals(pint_env%input, &
                                                 TRIM(sect_path(id)))

         should_output = cp_print_key_should_output( &
                         iteration_info=logger%iter_info, &
                         basis_section=print_key)
         IF (.NOT. BTEST(should_output, cp_p_file)) CONTINUE

         ! get units of measure for output (if available)
         CALL section_vals_val_get(print_key, "UNIT", &
                                   c_val=unit_str)
         unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

         ! get the format for output
         CALL section_vals_val_get(print_key, "FORMAT", i_val=outformat)

         SELECT CASE (outformat)
         CASE (dump_dcd, dump_dcd_aligned_cell)
            form = "UNFORMATTED"
            ext = ".dcd"
         CASE (dump_atomic)
            form = "FORMATTED"
            ext = ""
         CASE (dump_xmol)
            form = "FORMATTED"
            ext = ".xyz"
         CASE default
            CALL cp__b("motion/pint_io.F",180,"")
         END SELECT

         NULLIFY (f_env, cell, subsys)
         CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id, &
                                 f_env=f_env, handle=handle)
         CALL force_env_get(force_env=f_env%force_env, &
                            cell=cell, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)

         ! calculate and copy the requested property
         ! to the particles structure
         nb = REAL(pint_env%p, dp)
         idim = 0
         DO iat = 1, pint_env%ndim/3
            DO idir = 1, 3
               idim = idim+1
               ss = 0.0_dp
               vv = 0.0_dp
!          ss2=0.0_dp
               DO ib = 1, pint_env%p
                  ss = ss+pint_env%x(ib, idim)
                  vv = vv+pint_env%v(ib, idim)
!            ss2=ss2+pint_env%x(ib,idim)**2
               END DO
               particles%els(iat)%r(idir) = ss/nb
               particles%els(iat)%v(idir) = vv/nb
!          particles%els(iat)%v(idir)=SQRT(ss2/nb-(ss/nb)**2)
            END DO
         END DO

         ! set up the output unit number and file name
         ! for the current property
         my_middle_name = TRIM(middle_name(id))
         unit_nr = cp_print_key_unit_nr(logger=logger, &
                                        basis_section=print_key, print_key_path="", &
                                        extension=TRIM(ext), middle_name=TRIM(my_middle_name), &
                                        local=.FALSE., file_form=form, is_new_file=new_file)

         ! don't write the 0-th frame if the file already exists
         IF (.NOT. new_file .AND. (pint_env%iter .LE. pint_env%first_step)) THEN
            CALL cp_print_key_finished_output(unit_nr, logger, &
                                              print_key)
            CONTINUE
         END IF

         ! actually perform the i/o - on the ionode only
         IF (unit_nr > 0) THEN

            CALL write_particle_coordinates( &
               particles%els, &
               iunit=unit_nr, &
               output_format=outformat, &
               content=content_id(id), &
               title=title(id), &
               cell=cell, &
               unit_conv=unit_conv)

            CALL cp_print_key_finished_output(unit_nr, logger, &
                                              print_key, "", local=.FALSE.)

         END IF

         CALL f_env_rm_defaults(f_env, ierr, handle)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/pint_io.F",244)

      END DO

      CALL timestop(handle1)
      RETURN
   END SUBROUTINE pint_write_centroids

! ***************************************************************************
!> \brief  Write out the trajectory of the beads (positions and velocities)
!> \param pint_env ...
!> \par    History
!>         2010-11-25 added support for velocity printing [lwalewski]
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE pint_write_trajectory(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_trajectory', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: n_ids = 2, pos_id = 1, vel_id = 2

      CHARACTER(len=default_string_length)               :: ext, form, ib_str, my_middle_name, &
                                                            title, unit_str
      CHARACTER(len=default_string_length), DIMENSION(2) :: content_id, middle_name, sect_path
      INTEGER                                            :: handle, handle1, iat, ib, id, idim, &
                                                            idir, ierr, imag_stride, outformat, &
                                                            should_output, unit_nr
      LOGICAL                                            :: new_file
      REAL(kind=dp)                                      :: unit_conv
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle1)

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",283)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_io.F",284)

      sect_path(pos_id) = "MOTION%PRINT%TRAJECTORY"
      sect_path(vel_id) = "MOTION%PRINT%VELOCITIES"
      middle_name(pos_id) = "pos-"
      middle_name(vel_id) = "vel-"
      content_id(pos_id) = "POS"
      content_id(vel_id) = "VEL"

      NULLIFY (logger)
      logger => cp_get_default_logger()

      CALL pint_u2x(pint_env, ux=pint_env%uv, x=pint_env%v)

      ! iterate over the properties that we know how to print
      ! (currently positions and velocities)
      DO id = 1, n_ids

         print_key => section_vals_get_subs_vals(pint_env%input, &
                                                 TRIM(sect_path(id)))

         should_output = cp_print_key_should_output( &
                         iteration_info=logger%iter_info, &
                         basis_section=print_key)
         IF (.NOT. BTEST(should_output, cp_p_file)) CONTINUE

         ! get units of measure for output (if available)
         CALL section_vals_val_get(print_key, "UNIT", &
                                   c_val=unit_str)
         unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

         ! get the format for output
         CALL section_vals_val_get(print_key, "FORMAT", i_val=outformat)

         SELECT CASE (outformat)
         CASE (dump_dcd, dump_dcd_aligned_cell)
            form = "UNFORMATTED"
            ext = ".dcd"
         CASE (dump_atomic)
            form = "FORMATTED"
            ext = ""
         CASE (dump_xmol)
            form = "FORMATTED"
            ext = ".xyz"
         CASE default
            CALL cp__b("motion/pint_io.F",329,"")
         END SELECT

         NULLIFY (f_env, cell, subsys)
         CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id, &
                                 f_env=f_env, handle=handle)
         CALL force_env_get(force_env=f_env%force_env, &
                            cell=cell, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)

         !Get print stride for bead trajectories
         CALL section_vals_val_get(pint_env%input, &
                                   "MOTION%PINT%PRINT%IMAGINARY_TIME_STRIDE", &
                                   i_val=imag_stride)

         ! iterate over beads
         DO ib = 1, pint_env%p, imag_stride

            ! copy the requested property of the current bead
            ! to the particles structure
            idim = 0
            DO iat = 1, pint_env%ndim/3
               DO idir = 1, 3
                  idim = idim+1
                  particles%els(iat)%r(idir) = pint_env%x(ib, idim)
                  particles%els(iat)%v(idir) = pint_env%v(ib, idim)
               END DO
            END DO

            ! set up the output unit number and file name
            ! for the current property and bead
            ib_str = ""
            WRITE (ib_str, *) ib
            my_middle_name = TRIM(middle_name(id))//TRIM(ADJUSTL(ib_str))
            unit_nr = cp_print_key_unit_nr(logger=logger, &
                                           basis_section=print_key, print_key_path="", &
                                           extension=TRIM(ext), middle_name=TRIM(my_middle_name), &
                                           local=.FALSE., file_form=form, is_new_file=new_file)

            ! don't write the 0-th frame if the file already exists
            IF (.NOT. new_file .AND. (pint_env%iter .LE. pint_env%first_step)) THEN
               CALL cp_print_key_finished_output(unit_nr, logger, &
                                                 print_key)
               CONTINUE
            END IF

            ! actually perform the i/o - on the ionode only
            IF (unit_nr > 0) THEN

               IF (outformat == dump_xmol) THEN
                  WRITE (UNIT=title, FMT="(A,I8,A,F20.10)") &
                     " i =", pint_env%iter, &
                     ", E =", pint_env%e_pot_bead(ib)
               END IF

               CALL write_particle_coordinates( &
                  particles%els, &
                  iunit=unit_nr, &
                  output_format=outformat, &
                  content=content_id(id), &
                  title=title, &
                  cell=cell, &
                  unit_conv=unit_conv)

               CALL cp_print_key_finished_output(unit_nr, logger, &
                                                 print_key, "", local=.FALSE.)

            END IF

         END DO

         CALL f_env_rm_defaults(f_env, ierr, handle)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/pint_io.F",401)

      END DO

      CALL timestop(handle1)
      RETURN
   END SUBROUTINE pint_write_trajectory

! ***************************************************************************
!> \brief  Write center of mass (COM) position according to PINT%PRINT%COM
!> \param pint_env ...
!> \date   2010-02-17
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_write_com(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_com', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: stmp1, stmp2
      INTEGER                                            :: ic, unit_nr
      LOGICAL                                            :: new_file, should_output
      REAL(kind=dp), DIMENSION(3)                        :: com_r
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",428)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! decide whether to write anything or not
      NULLIFY (print_key)
      print_key => section_vals_get_subs_vals(pint_env%input, &
                                              "MOTION%PINT%PRINT%COM")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)
      IF (.NOT. should_output) THEN
         RETURN
      END IF

      com_r = pint_com_pos(pint_env)
      DO ic = 1, 3
         com_r(ic) = cp_unit_from_cp2k(com_r(ic), "angstrom")
      END DO

      unit_nr = cp_print_key_unit_nr(logger, print_key, is_new_file=new_file, &
                                     middle_name="com-pos", extension=".xyz")

      ! don't write the 0-th frame if the file already exists
      IF (.NOT. new_file .AND. (pint_env%iter .LE. pint_env%first_step)) THEN
         CALL cp_print_key_finished_output(unit_nr, logger, &
                                           print_key)
         RETURN
      END IF

      ! actually perform the i/o - on the ionode only
      IF (unit_nr > 0) THEN

         WRITE (unit_nr, '(I2)') 1
         WRITE (stmp1, *) pint_env%iter
         WRITE (stmp2, '(F20.10)') pint_env%energy(e_conserved_id)
         WRITE (unit_nr, '(4A)') " Iteration = ", TRIM(ADJUSTL(stmp1)), &
            ", E_conserved = ", TRIM(ADJUSTL(stmp2))
         WRITE (unit_nr, '(A2,3(1X,F20.10))') "X ", (com_r(ic), ic=1, 3)

         CALL m_flush(unit_nr)

      END IF

      CALL cp_print_key_finished_output(unit_nr, logger, print_key)

      RETURN
   END SUBROUTINE pint_write_com

! ***************************************************************************
!> \brief  Writes out the energies according to PINT%PRINT%ENERGY
!> \param  pint_env path integral environment
!> \par    History
!>           various bug fixes [hforbert]
!>           2009-11-16 energy components calc moved out of here [lwalewski]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_write_ener(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_ener', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ndof, unit_nr
      LOGICAL                                            :: file_is_new
      REAL(kind=dp)                                      :: t, temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",498)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_io.F",499)

      NULLIFY (print_key, logger)
      print_key => section_vals_get_subs_vals(pint_env%input, &
                                              "MOTION%PINT%PRINT%ENERGY")
      logger => cp_get_default_logger()
      IF (BTEST(cp_print_key_should_output(iteration_info=logger%iter_info, &
                                           basis_section=print_key), cp_p_file)) THEN

         unit_nr = cp_print_key_unit_nr(logger, print_key, middle_name="energy", &
                                        extension=".dat", is_new_file=file_is_new)

         ! don't write the 0-th frame if the file already exists
         IF (.NOT. file_is_new .AND. (pint_env%iter .LE. pint_env%first_step)) THEN
            CALL cp_print_key_finished_output(unit_nr, logger, &
                                              print_key)
            RETURN
         END IF

         ! cp_print_key_unit_nr returns -1 on nodes other than logger%para_env%ionode
         IF (unit_nr > 0) THEN

            ! please keep the format explaination up to date
            ! keep the constant of motion the true constant of motion !
            IF (file_is_new) THEN
               WRITE (unit_nr, "(A8,1X,A12,1X,5(A20,1X),A12)") &
                  "# StepNr", &
                  "   Time [fs]", &
                  "      Kinetic [a.u.]", &
                  "    VirialKin [a.u.]", &
                  "     Temperature [K]", &
                  "    Potential [a.u.]", &
                  "      ConsQty [a.u.]", &
                  "     CPU [s]"
            END IF

            t = cp_unit_from_cp2k(pint_env%t, "fs")

            ndof = pint_env%p
            IF (pint_env%first_propagated_mode .EQ. 2) THEN
               ndof = ndof-1
            END IF
            temp = cp_unit_from_cp2k(2.0_dp*pint_env%e_kin_beads/ &
                                     REAL(ndof, dp)/REAL(pint_env%ndim, dp), &
                                     "K")*pint_env%propagator%temp_sim2phys

            WRITE (unit_nr, "(I8,1X,F12.3,1X,5(F20.9,1X),F12.1)") &
               pint_env%iter, &
               t, &
               pint_env%energy(e_kin_thermo_id), &
               pint_env%energy(e_kin_virial_id), &
               temp, &
               pint_env%energy(e_potential_id), &
               pint_env%energy(e_conserved_id), &
               pint_env%time_per_step
            CALL m_flush(unit_nr)

         END IF

         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
      END IF

      RETURN
   END SUBROUTINE pint_write_ener

! ***************************************************************************
!> \brief  Write step info to the output file.
!> \param pint_env ...
!> \date   2009-11-16
!> \par History
!>      2010-01-27 getting default unit nr now only on ionode [lwalewski]
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_write_step_info(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_step_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msgstr, stmp, time_unit
      INTEGER                                            :: unit_nr
      REAL(kind=dp)                                      :: time_used
      TYPE(cp_logger_type), POINTER                      :: logger

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",583)

      unit_nr = 0
      NULLIFY (logger)
      logger => cp_get_default_logger()

      time_used = pint_env%time_per_step
      time_unit = "sec"
      IF (time_used .GE. 60.0_dp) THEN
         time_used = time_used/60.0_dp
         time_unit = "min"
      END IF
      IF (time_used .GE. 60.0_dp) THEN
         time_used = time_used/60.0_dp
         time_unit = "hours"
      END IF
      msgstr = "MD step"
      stmp = ""
      WRITE (stmp, *) pint_env%iter
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))//" of"
      stmp = ""
      WRITE (stmp, *) pint_env%last_step
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))//" in"
      stmp = ""
      WRITE (stmp, '(F20.1)') time_used
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(time_unit))//"."

      IF (logger%para_env%ionode) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger)
         WRITE (unit_nr, '(T2,A)') "PINT| "//TRIM(ADJUSTL(msgstr))
      END IF

      ! print out the total energy - for regtest evaluation
      stmp = ""
      WRITE (stmp, *) pint_env%energy(e_conserved_id)
      msgstr = "Total energy = "//TRIM(ADJUSTL(stmp))
      IF (logger%para_env%ionode) THEN
         WRITE (unit_nr, '(T2,A)') "PINT| "//TRIM(ADJUSTL(msgstr))
      END IF

      RETURN
   END SUBROUTINE pint_write_step_info

! ***************************************************************************
!> \brief  Write radii of gyration according to PINT%PRINT%CENTROID_GYR
!> \param pint_env ...
!> \date   2011-01-07
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_write_rgyr(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_write_rgyr', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: unit_str
      INTEGER                                            :: ia, ib, ic, idim, unit_nr
      LOGICAL                                            :: new_file, should_output
      REAL(kind=dp)                                      :: nb, ss, unit_conv
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_io.F",647)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! decide whether to write anything or not
      NULLIFY (print_key)
      print_key => section_vals_get_subs_vals(pint_env%input, &
                                              "MOTION%PINT%PRINT%CENTROID_GYR")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)
      IF (.NOT. should_output) THEN
         RETURN
      END IF

      ! get the units conversion factor
      CALL section_vals_val_get(print_key, "UNIT", c_val=unit_str)
      unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

      ! calculate the centroid positions
      nb = REAL(pint_env%p, dp)
      idim = 0
      DO ia = 1, pint_env%ndim/3
         DO ic = 1, 3
            idim = idim+1
            ss = 0.0_dp
            DO ib = 1, pint_env%p
               ss = ss+pint_env%x(ib, idim)
            END DO
            pint_env%rtmp_ndim(idim) = ss/nb
         END DO
      END DO

      ! calculate the radii of gyration
      idim = 0
      DO ia = 1, pint_env%ndim/3
         ss = 0.0_dp
         DO ic = 1, 3
            idim = idim+1
            DO ib = 1, pint_env%p
               ss = ss+(pint_env%x(ib, idim)-pint_env%rtmp_ndim(idim))**2
            END DO
         END DO
         pint_env%rtmp_natom(ia) = SQRT(ss/nb)*unit_conv
      END DO

      unit_nr = cp_print_key_unit_nr(logger, print_key, is_new_file=new_file, &
                                     middle_name="centroid-gyr", extension=".dat")

      ! don't write the 0-th frame if the file already exists
      IF (.NOT. new_file .AND. (pint_env%iter .LE. pint_env%first_step)) THEN
         CALL cp_print_key_finished_output(unit_nr, logger, &
                                           print_key)
         RETURN
      END IF

      ! actually perform the i/o - on the ionode only
      IF (unit_nr > 0) THEN

         DO ia = 1, pint_env%ndim/3
            WRITE (unit_nr, '(F20.10,1X)', ADVANCE='NO') pint_env%rtmp_natom(ia)
         END DO
         WRITE (unit_nr, '(A)') ""

         CALL m_flush(unit_nr)

      END IF

      CALL cp_print_key_finished_output(unit_nr, logger, print_key)

      RETURN
   END SUBROUTINE pint_write_rgyr

END MODULE pint_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods to apply a simple Lagevin thermostat to PI runs.
!>         v_new = c1*vold + SQRT(kT/m)*c2*random
!> \author Felix Uhl
!> \par History
!>      10.2014 created [Felix Uhl]
! **************************************************************************************************
MODULE pint_pile
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              read_rng_stream,&
                                              rng_record_length
   USE pint_types,                      ONLY: normalmode_env_type,&
                                              pile_therm_type,&
                                              pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pint_pile_step, &
             pint_pile_init, &
             pint_pile_release, &
             pint_calc_pile_energy

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_pile'

CONTAINS

   ! ***************************************************************************
   !> \brief initializes the data for a pile run
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param pile_therm ...
!> \param pint_env ...
!> \param normalmode_env ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE pint_pile_init(pile_therm, pint_env, normalmode_env, section)
      TYPE(pile_therm_type), POINTER                     :: pile_therm
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(normalmode_env_type), POINTER                 :: normalmode_env
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_pile_init', routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: i, j, p
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: dti2, ex
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed
      TYPE(section_vals_type), POINTER                   :: rng_section

      pint_env%e_pile = 0.0_dp
      ALLOCATE (pile_therm)
      pile_therm%ref_count = 1
      pile_therm%thermostat_energy = 0.0_dp
      !Get input parameter
      CALL section_vals_val_get(section, "TAU", r_val=pile_therm%tau)
      CALL section_vals_val_get(section, "LAMBDA", r_val=pile_therm%lamb)
      pile_therm%tau = cp_unit_to_cp2k(pile_therm%tau, "fs")

      p = pint_env%p
      dti2 = 0.5_dp*pint_env%dt
      ALLOCATE (pile_therm%c1(p))
      ALLOCATE (pile_therm%c2(p))
      ALLOCATE (pile_therm%g_fric(p))
      ALLOCATE (pile_therm%massfact(p, pint_env%ndim))
      !Initialize everything
      ! If tau is negative or zero the thermostat does not act on the centroid
      ! (TRPMD)
      IF (pile_therm%tau <= 0.0_dp) THEN
         pile_therm%g_fric(1) = 0.0_dp
      ELSE
         pile_therm%g_fric(1) = 1.0_dp/pile_therm%tau
      END IF
      DO i = 2, p
         pile_therm%g_fric(i) = 2.0_dp*pile_therm%lamb*normalmode_env%lambda(i)
      END DO
      DO i = 1, p
         ex = -dti2*pile_therm%g_fric(i)
         pile_therm%c1(i) = EXP(ex)
         ex = pile_therm%c1(i)*pile_therm%c1(i)
         pile_therm%c2(i) = SQRT(1.0_dp-ex)
      END DO
      DO j = 1, pint_env%ndim
         DO i = 1, pint_env%p
            pile_therm%massfact(i, j) = SQRT(pint_env%kT/pint_env%mass_fict(i, j))
         END DO
      END DO

      !prepare Random number generator
      NULLIFY (rng_section)
      NULLIFY (pile_therm%gaussian_rng_stream)
      rng_section => section_vals_get_subs_vals(section, &
                                                subsection_name="RNG_INIT")
      CALL section_vals_get(rng_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(rng_section, "_DEFAULT_KEYWORD_", &
                                   i_rep_val=1, c_val=rng_record)
         CALL read_rng_stream(rng_stream=pile_therm%gaussian_rng_stream, &
                              rng_record=rng_record)
      ELSE
         initial_seed(:, :) = REAL(pint_env%thermostat_rng_seed, dp)
         CALL create_rng_stream(rng_stream=pile_therm%gaussian_rng_stream, &
                                name="pile_rng_gaussian", distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE., &
                                seed=initial_seed)
      END IF

   END SUBROUTINE pint_pile_init

! **************************************************************************************************
!> \brief ...
!> \param vold ...
!> \param vnew ...
!> \param p ...
!> \param ndim ...
!> \param first_mode ...
!> \param masses ...
!> \param pile_therm ...
! **************************************************************************************************
   SUBROUTINE pint_pile_step(vold, vnew, p, ndim, first_mode, masses, pile_therm)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vold, vnew
      INTEGER, INTENT(IN)                                :: p, ndim, first_mode
      REAL(kind=dp), DIMENSION(:, :), INTENT(IN)         :: masses
      TYPE(pile_therm_type), POINTER                     :: pile_therm

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_pile_step', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ibead, idim
      REAL(KIND=dp)                                      :: delta_ekin

      CALL timeset(routineN, handle)
      delta_ekin = 0.0_dp
      DO idim = 1, ndim
         DO ibead = first_mode, p
            vnew(ibead, idim) = pile_therm%c1(ibead)*vold(ibead, idim)+ &
                                pile_therm%massfact(ibead, idim)*pile_therm%c2(ibead)* &
                                next_random_number(pile_therm%gaussian_rng_stream)
            delta_ekin = delta_ekin+masses(ibead, idim)*( &
                         vnew(ibead, idim)*vnew(ibead, idim)- &
                         vold(ibead, idim)*vold(ibead, idim))
         END DO
      END DO
      pile_therm%thermostat_energy = pile_therm%thermostat_energy-0.5_dp*delta_ekin

      CALL timestop(handle)
   END SUBROUTINE pint_pile_step

   ! ***************************************************************************
   !> \brief releases the pile environment
   !> \param pile_therm pile data to be released
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param pile_therm ...
! **************************************************************************************************
   SUBROUTINE pint_pile_release(pile_therm)

      TYPE(pile_therm_type), POINTER                     :: pile_therm

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_pile_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pile_therm)) THEN
         pile_therm%ref_count = pile_therm%ref_count-1
         IF (pile_therm%ref_count == 0) THEN
            DEALLOCATE (pile_therm%c1)
            DEALLOCATE (pile_therm%c2)
            DEALLOCATE (pile_therm%g_fric)
            DEALLOCATE (pile_therm%massfact)
            CALL delete_rng_stream(pile_therm%gaussian_rng_stream)
            DEALLOCATE (pile_therm)
         END IF
      END IF
      NULLIFY (pile_therm)

      RETURN
   END SUBROUTINE pint_pile_release

   ! ***************************************************************************
   !> \brief returns the pile kinetic energy contribution
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_calc_pile_energy(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_pile_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pint_env%pile_therm)) THEN
         pint_env%e_pile = pint_env%pile_therm%thermostat_energy
      END IF

      RETURN

   END SUBROUTINE pint_calc_pile_energy
END MODULE
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!> \author
! **************************************************************************************************
MODULE gle_system_dynamics

   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: map_info_type
   USE gle_system_types,                ONLY: gle_thermo_create,&
                                              gle_type
   USE input_constants,                 ONLY: &
        do_thermo_communication, do_thermo_no_communication, isokin_ensemble, langevin_ensemble, &
        npe_f_ensemble, npe_i_ensemble, nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, &
        npt_f_ensemble, npt_i_ensemble, nve_ensemble, nvt_ensemble, reftraj_ensemble
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              read_rng_stream,&
                                              rng_record_length,&
                                              rng_stream_type
   USE particle_types,                  ONLY: particle_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_mapping,              ONLY: thermostat_mapping_region
   USE thermostat_types,                ONLY: thermostat_info_type
   USE thermostat_utils,                ONLY: ke_region_particles,&
                                              momentum_region_particles,&
                                              vel_rescale_particles
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: gle_particles, &
             initialize_gle_part, &
             gle_cholesky_stab, &
             gle_matrix_exp, &
             restart_gle

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gle_system_dynamics'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gle ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param group ...
!> \param shell_adiabatic ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \date
!> \par History
! **************************************************************************************************
   SUBROUTINE gle_particles(gle, molecule_kind_set, molecule_set, particle_set, local_molecules, &
                            group, shell_adiabatic, shell_particle_set, core_particle_set, vel, shell_vel, core_vel)

      TYPE(gle_type), POINTER                            :: gle
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      INTEGER, INTENT(IN)                                :: group
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell_adiabatic
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'gle_particles', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iadd, ideg, imap, ndim, num
      LOGICAL                                            :: my_shell_adiabatic, present_vel
      REAL(dp)                                           :: alpha, beta, rr
      REAL(dp), DIMENSION(:, :), POINTER                 :: a_mat, e_tmp, h_tmp, s_tmp
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CALL timeset(routineN, handle)
      my_shell_adiabatic = .FALSE.
      IF (PRESENT(shell_adiabatic)) my_shell_adiabatic = shell_adiabatic
      NULLIFY (rng_stream)
      present_vel = PRESENT(vel)
      ndim = gle%ndim
      ALLOCATE (s_tmp(ndim, gle%loc_num_gle))
      s_tmp = 0.0_dp
      ALLOCATE (e_tmp(ndim, gle%loc_num_gle))
      ALLOCATE (h_tmp(ndim, gle%loc_num_gle))

      map_info => gle%map_info
      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel)
      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%map_index(ideg)
         gle%nvt(ideg)%kin_energy = map_info%s_kin(imap)
      END DO

      CALL momentum_region_particles(map_info, particle_set, molecule_kind_set, &
                                     local_molecules, molecule_set, group, vel)

      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%map_index(ideg)
         IF (gle%nvt(ideg)%nkt == 0.0_dp) CYCLE
         gle%nvt(ideg)%s(1) = map_info%s_kin(imap)
         s_tmp(1, imap) = map_info%s_kin(imap)
         rng_stream => gle%nvt(ideg)%gaussian_rng_stream
         rr = next_random_number(rng_stream)
         e_tmp(1, imap) = rr
         DO iadd = 2, ndim
            s_tmp(iadd, imap) = gle%nvt(ideg)%s(iadd)
            rr = next_random_number(rng_stream)
            e_tmp(iadd, imap) = rr
         END DO
      END DO
      num = gle%loc_num_gle
      a_mat => gle%gle_s
      alpha = 1.0_dp
      beta = 0.0_dp
      CALL DGEMM('N', 'N', ndim, num, ndim, alpha, a_mat(1, 1), ndim, e_tmp(1, 1), ndim, beta, h_tmp(1, 1), ndim)
!
      a_mat => gle%gle_t
      beta = 1.0_dp
      CALL dgemm("N", "N", ndim, num, ndim, alpha, a_mat(1, 1), ndim, s_tmp(1, 1), ndim, beta, h_tmp(1, 1), ndim)

      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%map_index(ideg)
         IF (gle%nvt(ideg)%nkt == 0.0_dp) CYCLE

         map_info%v_scale(imap) = h_tmp(1, imap)/s_tmp(1, imap)
         DO iadd = 2, ndim
            gle%nvt(ideg)%s(iadd) = h_tmp(iadd, ideg)
         END DO
      END DO

      CALL vel_rescale_particles(map_info, molecule_kind_set, molecule_set, particle_set, &
                                 local_molecules, my_shell_adiabatic, shell_particle_set, core_particle_set, &
                                 vel, shell_vel, core_vel)

      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel)
      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%map_index(ideg)
         gle%nvt(ideg)%thermostat_energy = gle%nvt(ideg)%thermostat_energy+ &
                                           0.5_dp*(gle%nvt(ideg)%kin_energy-map_info%s_kin(imap))
      END DO

      DEALLOCATE (e_tmp, s_tmp, h_tmp)

      CALL timestop(handle)
   END SUBROUTINE gle_particles

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param gle ...
!> \param gle_section ...
!> \param gci ...
!> \param save_mem ...
!> \author
! **************************************************************************************************
   SUBROUTINE initialize_gle_part(thermostat_info, simpar, local_molecules, &
                                  molecule, molecule_kind_set, para_env, gle, gle_section, &
                                  gci, save_mem)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gle_type), POINTER                            :: gle
      TYPE(section_vals_type), POINTER                   :: gle_section
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN)                                :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_gle_part', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: restart
      REAL(dp)                                           :: Mtmp(gle%ndim, gle%ndim)

      restart = .FALSE.

      CALL gle_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                   molecule, molecule_kind_set, gle, para_env, gci)

      IF (gle%ndim /= 0) THEN
         CALL init_gle_variables(gle)
      END IF
      CALL restart_gle(gle, gle_section, save_mem, restart)

      ! here we should have read a_mat and c_mat; whe can therefore compute S and T
      ! deterministic part of the propagator
      CALL gle_matrix_exp((-simpar%dt*0.5_dp)*gle%a_mat, gle%ndim, 15, 15, gle%gle_t)
      ! stochastic part
      Mtmp = gle%c_mat-MATMUL(gle%gle_t, MATMUL(gle%c_mat, TRANSPOSE(gle%gle_t)))
      CALL gle_cholesky_stab(Mtmp, gle%gle_s, gle%ndim)

   END SUBROUTINE initialize_gle_part

! **************************************************************************************************
!> \brief ...
!> \param M ...
!> \param n ...
!> \param j ...
!> \param k ...
!> \param EM ...
!> \author Michele
!> routine to compute matrix exponential via scale & square
! **************************************************************************************************
   SUBROUTINE gle_matrix_exp(M, n, j, k, EM)

      INTEGER, INTENT(in)                                :: n
      REAL(dp), INTENT(in)                               :: M(n, n)
      INTEGER, INTENT(in)                                :: j, k
      REAL(dp), INTENT(out)                              :: EM(n, n)

      INTEGER                                            :: i, p
      REAL(dp)                                           :: SM(n, n), tc(j+1)

      tc(1) = 1._dp
      DO i = 1, j
         tc(i+1) = tc(i)/REAL(i, KIND=dp)
      ENDDO

      !scale
      SM = M*(1._dp/2._dp**k)
      EM = 0._dp
      DO i = 1, n
         EM(i, i) = tc(j+1)
      ENDDO

      !taylor exp of scaled matrix
      DO p = j, 1, -1
         EM = MATMUL(SM, EM)
         DO i = 1, n
            EM(i, i) = EM(i, i)+tc(p)
         ENDDO
      ENDDO

      !square
      DO p = 1, k
         EM = MATMUL(EM, EM)
      ENDDO
   END SUBROUTINE gle_matrix_exp

! **************************************************************************************************
!> \brief ...
!> \param SST ...
!> \param S ...
!> \param n ...
!> \author Michele
!>  "stabilized" cholesky to deal with small & negative diagonal elements,
!>  in practice a LDL^T decomposition is computed, and negative els. are zeroed
!>  \par History
!>      05.11.2015: Bug fix: In rare cases D(j) is negative due to numerics [Felix Uhl]
! **************************************************************************************************
   SUBROUTINE gle_cholesky_stab(SST, S, n)
      INTEGER, INTENT(in)                                :: n
      REAL(dp), INTENT(out)                              :: S(n, n)
      REAL(dp), INTENT(in)                               :: SST(n, n)

      INTEGER                                            :: i, j, k
      REAL(dp)                                           :: D(n), L(n, n)

      D = 0._dp
      L = 0._dp
      S = 0._dp
      DO i = 1, n
         L(i, i) = 1.0_dp
         D(i) = SST(i, i)
         DO j = 1, i-1
            L(i, j) = SST(i, j); 
            DO k = 1, j-1
               L(i, j) = L(i, j)-L(i, k)*L(j, k)*D(k)
            ENDDO
            IF (ABS(D(j)) > EPSILON(1.0_dp)) L(i, j) = L(i, j)/D(j)
         ENDDO
         DO k = 1, i-1
            D(i) = D(i)-L(i, k)*L(i, k)*D(k)
         END DO
      ENDDO
      DO i = 1, n
         DO j = 1, i
            IF ((ABS(D(j)) > EPSILON(1.0_dp)) .AND. (D(j) > 0.0_dp)) THEN
               S(i, j) = S(i, j)+L(i, j)*SQRT(D(j))
            END IF
         ENDDO
      ENDDO

   END SUBROUTINE gle_cholesky_stab

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param gle ...
!> \param para_env ...
!> \param gci ...
!> \author
! **************************************************************************************************
   SUBROUTINE gle_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                      molecule_set, molecule_kind_set, gle, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(gle_type), POINTER                            :: gle
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gle_to_particle_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, j, mal_size, natoms_local, &
                                                            nkind, number, region, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      LOGICAL                                            :: do_shell
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      do_shell = .FALSE.
      NULLIFY (massive_atom_list, deg_of_freedom)
      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/gle_system_dynamics.F",356,"Unknown ensemble!")
      CASE (nve_ensemble, isokin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/gle_system_dynamics.F",359,"Never reach this point!")
      CASE (nvt_ensemble, npt_i_ensemble, npt_f_ensemble)

         map_info => gle%map_info
         nkind = SIZE(molecule_kind_set)
         sum_of_thermostats = thermostat_info%sum_of_thermostats
         map_info%dis_type = thermostat_info%dis_type
         number = thermostat_info%number_of_thermostats
         region = gle%region

         CALL thermostat_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                        molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                        simpar, number, region, gci, do_shell, thermostat_info%map_loc_thermo_gen, &
                                        sum_of_thermostats)

         ! This is the local number of available thermostats
         gle%loc_num_gle = number
         gle%glob_num_gle = sum_of_thermostats
         mal_size = SIZE(massive_atom_list)
         IF(.NOT.(mal_size /= 0))CALL cp__a("motion/thermostat/gle_system_dynamics.F",378)
         CALL gle_thermo_create(gle, mal_size)
         gle%mal(1:mal_size) = massive_atom_list(1:mal_size)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! If thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/gle_system_dynamics.F",399,"Zero degrees of freedom. Nothing to thermalize!")
            END IF
            gle%nvt(1)%nkt = simpar%temp_ext*fac
            gle%nvt(1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, gle%loc_num_gle
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               gle%nvt(i)%nkt = simpar%temp_ext*fac
               gle%nvt(i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF
         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)
      END SELECT

   END SUBROUTINE gle_to_particle_mapping

! **************************************************************************************************
!> \brief ...
!> \param gle ...
!> \param gle_section ...
!> \param save_mem ...
!> \param restart ...
! **************************************************************************************************
   SUBROUTINE restart_gle(gle, gle_section, save_mem, restart)

      TYPE(gle_type), POINTER                            :: gle
      TYPE(section_vals_type), POINTER                   :: gle_section
      LOGICAL, INTENT(IN)                                :: save_mem
      LOGICAL, INTENT(OUT)                               :: restart

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_gle', routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: glob_num, i, ind, j, loc_num, n_rep
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buffer
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(section_vals_type), POINTER                   :: work_section

      NULLIFY (buffer, work_section)

      explicit = .FALSE.
      restart = .FALSE.

      IF (ASSOCIATED(gle_section)) THEN
         work_section => section_vals_get_subs_vals(gle_section, "S")
         CALL section_vals_get(work_section, explicit=explicit)
         restart = explicit
      END IF

      IF (restart) THEN
         map_info => gle%map_info
         CALL section_vals_val_get(gle_section, "S%_DEFAULT_KEYWORD_", r_vals=buffer)
         DO i = 1, SIZE(gle%nvt, 1)
            ind = map_info%index(i)
            ind = (ind-1)*(gle%ndim)
            DO j = 1, SIZE(gle%nvt(i)%s, 1)
               ind = ind+1
               gle%nvt(i)%s(j) = buffer(ind)
            END DO
         END DO

         IF (save_mem) THEN
            NULLIFY (work_section)
            work_section => section_vals_get_subs_vals(gle_section, "S")
            CALL section_vals_remove_values(work_section)
         END IF

         ! Possibly restart the initial thermostat energy
         work_section => section_vals_get_subs_vals(section_vals=gle_section, &
                                                    subsection_name="THERMOSTAT_ENERGY")
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                      n_rep_val=n_rep)
            IF (n_rep == gle%glob_num_gle) THEN
               DO i = 1, gle%loc_num_gle
                  ind = map_info%index(i)
                  CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                            i_rep_val=ind, r_val=gle%nvt(i)%thermostat_energy)
               END DO
            ELSE
               CALL cp_abort(cp__l("motion/thermostat/gle_system_dynamics.F",483), &
                             "Number of thermostat energies not equal to the number of "// &
                             "total thermostats!")
            END IF
         END IF

         ! Possibly restart the random number generators for the different thermostats
         work_section => section_vals_get_subs_vals(section_vals=gle_section, &
                                                    subsection_name="RNG_INIT")

         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                      n_rep_val=n_rep)

            glob_num = gle%glob_num_gle
            loc_num = gle%loc_num_gle
            IF (n_rep == glob_num) THEN
               DO i = 1, loc_num
                  ind = map_info%index(i)
                  CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                            i_rep_val=ind, c_val=rng_record)
                  CALL read_rng_stream(rng_stream=gle%nvt(i)%gaussian_rng_stream, &
                                       rng_record=rng_record)
               END DO
            ELSE
               CALL cp_abort(cp__l("motion/thermostat/gle_system_dynamics.F",509), &
                             "Number pf restartable stream not equal to the number of "// &
                             "total thermostats!")
            END IF
         END IF
      END IF

   END SUBROUTINE restart_gle

! **************************************************************************************************
!> \brief ...
!> \param gle ...
! **************************************************************************************************
   SUBROUTINE init_gle_variables(gle)

      TYPE(gle_type), POINTER                            :: gle

      CHARACTER(len=*), PARAMETER :: routineN = 'init_gle_variables', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(dp)                                           :: rr(gle%ndim), cc(gle%ndim, gle%ndim)
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CALL gle_cholesky_stab(gle%c_mat, cc, gle%ndim)
      DO i = 1, gle%loc_num_gle
         rng_stream => gle%nvt(i)%gaussian_rng_stream
         DO j = 1, gle%ndim
            ! here s should be properly initialized, when it is not read from restart
            rr(j) = next_random_number(rng_stream)
         END DO
         gle%nvt(i)%s = MATMUL(cc, rr)
      END DO

   END SUBROUTINE init_gle_variables

END MODULE gle_system_dynamics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods to apply the piglet thermostat to PI runs.
!> \author Felix Uhl
!> \par History
!>      10.2014 created [Felix Uhl]
! **************************************************************************************************
MODULE pint_piglet
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE gle_system_dynamics,             ONLY: gle_cholesky_stab,&
                                              gle_matrix_exp
   USE input_constants,                 ONLY: matrix_init_cholesky,&
                                              matrix_init_diagonal,&
                                              propagator_rpmd
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              read_rng_stream,&
                                              rng_record_length
   USE pint_io,                         ONLY: pint_write_line
   USE pint_types,                      ONLY: piglet_therm_type,&
                                              pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pint_piglet_create, &
             pint_piglet_init, &
             pint_piglet_step, &
             pint_piglet_release, &
             pint_calc_piglet_energy

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_piglet'

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                           !
!       _._ _..._ .-',     _.._   _         !
!      '-. `     '  /-._.-'    ',/ )        !
!         )         \            '.         !
!        / _    _    |             \        !
!       |  o    o    /              |       !
!       \   .-.                     ;       !
!        '-('' ).-'       ,'       ;        !
!           '-;           |      .'         !
!              \           \    /           !
!              | 7  .__  _.-\   \           !
!              | |  |  ``/  /`  /           !
!             /,_|  |   /,_/   /            !
!                /,_/      '`-'             !
!                                           !
!    (    (             (                   !
!    )\ ) )\ )  (       )\ )        *   )   !
!   (()/((()/(  )\ )   (()/(  (   ` )  /(   !
!    /(_))/(_))(()/(    /(_)) )\   ( )(_))  !
!   (_)) (_))   /(_))_ (_))  ((_) (_(_())   !
!   | _ \|_ _| (_)) __|| |   | __||_   _|   !
!   |  _/ | |    | (_ || |__ | _|   | |     !
!   |_|  |___|    \___||____||___|  |_|     !
!                                           !
!        Make Quantum Mechanics Hot         !
!                                           !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

CONTAINS

   ! ***************************************************************************
   !> \brief creates the data structure for a piglet thermostating in PI runs
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param piglet_therm ...
!> \param pint_env ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE pint_piglet_create(piglet_therm, pint_env, section)
      TYPE(piglet_therm_type), POINTER                   :: piglet_therm
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_piglet_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ndim, p

      NULLIFY (piglet_therm)
      ALLOCATE (piglet_therm)
      piglet_therm%ref_count = 1

      CALL section_vals_val_get(section, "NEXTRA_DOF", i_val=piglet_therm%nsp1)
      !add real degree of freedom to ns to reach nsp1
      piglet_therm%nsp1 = piglet_therm%nsp1+1
      p = pint_env%p
      piglet_therm%p = pint_env%p
      ndim = pint_env%ndim
      piglet_therm%ndim = pint_env%ndim
      NULLIFY (piglet_therm%a_mat)
      ALLOCATE (piglet_therm%a_mat(piglet_therm%nsp1, piglet_therm%nsp1, p))
      piglet_therm%a_mat(:, :, :) = 0.0_dp
      NULLIFY (piglet_therm%c_mat)
      ALLOCATE (piglet_therm%c_mat(piglet_therm%nsp1, piglet_therm%nsp1, p))
      piglet_therm%c_mat(:, :, :) = 0.0_dp
      NULLIFY (piglet_therm%gle_t)
      ALLOCATE (piglet_therm%gle_t(piglet_therm%nsp1, piglet_therm%nsp1, p))
      piglet_therm%gle_t(:, :, :) = 0.0_dp
      NULLIFY (piglet_therm%gle_s)
      ALLOCATE (piglet_therm%gle_s(piglet_therm%nsp1, piglet_therm%nsp1, p))
      piglet_therm%gle_s(:, :, :) = 0.0_dp
      NULLIFY (piglet_therm%smalls)
      ALLOCATE (piglet_therm%smalls(piglet_therm%nsp1, ndim*p))
      piglet_therm%smalls(:, :) = 0.0_dp
      NULLIFY (piglet_therm%temp1)
      ALLOCATE (piglet_therm%temp1(piglet_therm%nsp1, ndim*p))
      piglet_therm%temp1(:, :) = 0.0_dp
      NULLIFY (piglet_therm%temp2)
      ALLOCATE (piglet_therm%temp2(piglet_therm%nsp1))
      piglet_therm%temp2(:) = 0.0_dp
      NULLIFY (piglet_therm%sqrtmass)
      ALLOCATE (piglet_therm%sqrtmass(piglet_therm%p, piglet_therm%ndim))
      piglet_therm%sqrtmass(:, :) = 0.0_dp

      RETURN

   END SUBROUTINE pint_piglet_create

   ! ***************************************************************************
   !> \brief initializes the data for a piglet run
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param piglet_therm ...
!> \param pint_env ...
!> \param section ...
!> \param dt ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE pint_piglet_init(piglet_therm, pint_env, section, dt, para_env)
      TYPE(piglet_therm_type), POINTER                   :: piglet_therm
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(section_vals_type), POINTER                   :: section
      REAL(KIND=dp), INTENT(IN)                          :: dt
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_piglet_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=20)                                  :: default_format, read_unit
      CHARACTER(len=default_path_length)                 :: matrices_file_name
      CHARACTER(len=default_string_length)               :: msg, temp_input
      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: cbrac, i, ibead, idim, imode, &
                                                            input_unit, isp, j, matrix_init, ns, &
                                                            obrac, p, read_err
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed
      REAL(KIND=dp), DIMENSION(:), POINTER               :: smallstmp
      REAL(KIND=dp), &
         DIMENSION(piglet_therm%nsp1, piglet_therm%nsp1) :: Mtmp, tmpmatrix
      TYPE(section_vals_type), POINTER                   :: rng_section, smalls_section

      p = piglet_therm%p
      pint_env%e_piglet = 0.0_dp
      pint_env%piglet_therm%thermostat_energy = 0.0_dp
      CALL section_vals_val_get(section, "SMATRIX_INIT", i_val=matrix_init)
      !Read the matices

      IF (pint_env%propagator%prop_kind /= propagator_rpmd) THEN
         CALL cp__b("motion/pint_piglet.F",188,"PIGLET is designed to work with the RPMD propagator")
      END IF
      ! Select algorithm for S-matrix initialization
      IF (matrix_init == matrix_init_cholesky) THEN
         IF (para_env%ionode) THEN
            CALL pint_write_line("PIGLET| Initalizing S-matrices using cholesky decomposition.")
         END IF
      ELSE IF (matrix_init == matrix_init_diagonal) THEN
         IF (para_env%ionode) THEN
            CALL pint_write_line("PIGLET| Initalizing S-matrices using full diagonalization.")
         END IF
      ELSE
         CALL cp__w("motion/pint_piglet.F",200,"No PIGLET init algorithm selected. Selecting cholesky decomposition")
         matrix_init = matrix_init_cholesky
      END IF

      IF (para_env%ionode) THEN
         ! Read input matrices
         WRITE (default_format, '(A,I10,A)') "(A", default_string_length, ")"
         CALL section_vals_val_get(section, "MATRICES_FILE_NAME", &
                                   c_val=matrices_file_name)
         CALL pint_write_line("PIGLET| Reading PIGLET matrices from file: ")
         CALL pint_write_line("PIGLET|    "//TRIM(matrices_file_name))
         CALL open_file(file_name=TRIM(matrices_file_name), &
                        file_action="READ", file_status="OLD", unit_number=input_unit)
         read_err = 0
         msg = ""
         DO WHILE (read_err == 0)
            READ (input_unit, default_format, iostat=read_err) temp_input
            IF (read_err /= 0) EXIT
            !Parse comment section
            IF (INDEX(temp_input, "#") /= 0) THEN
               IF (INDEX(temp_input, "T=") /= 0) THEN
                  CALL check_temperature(line=temp_input, &
                                         propagator=pint_env%propagator%prop_kind, &
                                         targettemp=pint_env%kT*pint_env%propagator%temp_sim2phys)
               ELSE IF (INDEX(temp_input, "A MATRIX") /= 0) THEN
                  obrac = INDEX(temp_input, "(")+1
                  cbrac = INDEX(temp_input, ")")-1
                  read_unit = temp_input(obrac:cbrac)
                  DO imode = 1, p
                     READ (input_unit, default_format) temp_input
                     DO i = 1, piglet_therm%nsp1
                        READ (input_unit, *, iostat=read_err) &
                           (piglet_therm%a_mat(i, j, imode), j=1, piglet_therm%nsp1)
                        IF (read_err /= 0) THEN
                           WRITE (UNIT=msg, FMT=*) "Invalid PIGLET A-matrix Nr.", i-1
                           CALL cp__b("motion/pint_piglet.F",235,msg)
                           EXIT
                        END IF
                     END DO
                  END DO
                  !convert to cp2k units
                  IF (read_err == 0) THEN
                     CALL a_mat_to_cp2k(piglet_therm%a_mat, p, &
                                        piglet_therm%nsp1, read_unit, msg)
                  END IF
               ELSE IF (INDEX(temp_input, "C MATRIX") /= 0) THEN
                  obrac = INDEX(temp_input, "(")+1
                  cbrac = INDEX(temp_input, ")")-1
                  read_unit = temp_input(obrac:cbrac)
                  DO imode = 1, p
                     READ (input_unit, default_format) temp_input
                     DO i = 1, piglet_therm%nsp1
                        READ (input_unit, *, iostat=read_err) &
                           (piglet_therm%c_mat(i, j, imode), j=1, piglet_therm%nsp1)
                        IF (read_err /= 0) THEN
                           WRITE (UNIT=msg, FMT=*) "Invalid PIGLET C-matrix Nr.", i-1
                           CALL cp__b("motion/pint_piglet.F",256,msg)
                           EXIT
                        END IF
                     END DO
                  END DO
                  IF (read_err == 0) THEN
                     CALL c_mat_to_cp2k(piglet_therm%c_mat, p, &
                                        piglet_therm%nsp1, read_unit, msg)
                  END IF
               END IF
            END IF
         END DO
         CALL close_file(unit_number=input_unit)
      END IF
      ! communicate A and C matrix to other nodes
      CALL mp_bcast(piglet_therm%a_mat, &
                    para_env%source, &
                    para_env%group)
      CALL mp_bcast(piglet_therm%c_mat, &
                    para_env%source, &
                    para_env%group)

      !prepare Random number generator
      NULLIFY (rng_section)
      NULLIFY (piglet_therm%gaussian_rng_stream)
      rng_section => section_vals_get_subs_vals(section, &
                                                subsection_name="RNG_INIT")
      CALL section_vals_get(rng_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(rng_section, "_DEFAULT_KEYWORD_", &
                                   i_rep_val=1, c_val=rng_record)
         CALL read_rng_stream(rng_stream=piglet_therm%gaussian_rng_stream, &
                              rng_record=rng_record)
      ELSE
         initial_seed(:, :) = REAL(pint_env%thermostat_rng_seed, dp)
         CALL create_rng_stream(rng_stream=piglet_therm%gaussian_rng_stream, &
                                name="piglet_rng_gaussian", distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE., &
                                seed=initial_seed)
      END IF

      !Compute the T and S matrices on every mpi process
      DO i = 1, p
         ! T = EXP(-A_mat*dt) = EXP(-A_mat*dt/2)
         ! Values for j and k = 15 are way to high, but to be sure.
         ! (its only executed once anyway)
         CALL gle_matrix_exp(M=(-0.5_dp*dt)*piglet_therm%a_mat(:, :, i), & !dt scaled A matrix
                             n=piglet_therm%nsp1, & !size of matrix
                             j=15, & !truncation for taylor expansion
                             k=15, & !scaling parameter for faster convergence of taylor expansion
                             EM=piglet_therm%gle_t(:, :, i)) !output T matrices
         ! S*TRANSPOSE(S) = C-T*C*TRANSPOSE(T)
         ! T*C:
         CALL DGEMM('N', & ! T-matrix is not transposed
                    'N', & ! C-matrix is not transposed
                    piglet_therm%nsp1, & ! number of rows of T-matrix
                    piglet_therm%nsp1, & ! number of columns of C-matrix
                    piglet_therm%nsp1, & ! number of columns of T-matrix and number of rows of C-matrix
                    1.0D0, & ! scaling factor alpha
                    piglet_therm%gle_t(1, 1, i), & ! T-matrix
                    piglet_therm%nsp1, & ! leading dimension of T-matrix as declared
                    piglet_therm%c_mat(1, 1, i), & ! C-matrix
                    piglet_therm%nsp1, & ! leading dimension of C-matrix as declared
                    0.0D0, & ! scaling of tmpmatrix as additive
                    tmpmatrix, & ! result matrix: tmpmatrix
                    piglet_therm%nsp1) ! leading dimension of tmpmatrix
         ! T*C*TRANSPOSE(T):
         CALL DGEMM('N', & ! tmpmatrix is not transposed
                    'T', & ! T-matrix is transposed
                    piglet_therm%nsp1, & ! number of rows of tmpmatrix
                    piglet_therm%nsp1, & ! number of columns of T-matrix
                    piglet_therm%nsp1, & ! number of columns of tmpmatrix and number of rows of T-matrix
                    1.0D0, & ! scaling factor alpha
                    tmpmatrix, & ! tmpmatrix
                    piglet_therm%nsp1, & ! leading dimension of tmpmatrix as declared
                    piglet_therm%gle_t(1, 1, i), & ! T-matrix
                    piglet_therm%nsp1, & ! leading dimension of T-matrix as declared
                    0.0D0, & ! scaling of Mtmp as additive
                    Mtmp, & ! result matrix: Mtmp
                    piglet_therm%nsp1) ! leading dimension of Mtmp
         ! C - T*C*TRANSPOSE(T):
         Mtmp(:, :) = piglet_therm%c_mat(:, :, i)-Mtmp(:, :)

         IF (matrix_init == matrix_init_cholesky) THEN
            ! Get S by cholesky decomposition of Mtmp
            CALL gle_cholesky_stab(Mtmp, & ! Matrix to decompose
                                   piglet_therm%gle_s(:, :, i), & ! result
                                   piglet_therm%nsp1) ! Size of the matrix
         ELSE IF (matrix_init == matrix_init_diagonal) THEN
            ! Get S by full diagonalization of MTmp matrix
            CALL sqrt_pos_def_mat(piglet_therm%nsp1, & ! Size of the matrix
                                  Mtmp, & ! matrix to decompose
                                  piglet_therm%gle_s(:, :, i)) ! result
         END IF

      END DO

      ! Initialize extra degrees of freedom for Markovian Dynamics
      ! as a cholesky decomposition of C-matrix multiplied by a random number vector
      ! Or from restart
      piglet_therm%smalls = 0.0_dp
      NULLIFY (smalls_section)
      smalls_section => section_vals_get_subs_vals(section, subsection_name="EXTRA_DOF")
      CALL section_vals_get(smalls_section, explicit=explicit)
      IF (explicit) THEN
         NULLIFY (smallstmp)
         CALL section_vals_val_get(smalls_section, "_DEFAULT_KEYWORD_", &
                                   n_rep_val=ns)
         CALL section_vals_val_get(smalls_section, "_DEFAULT_KEYWORD_", &
                                   r_vals=smallstmp)
         i = 1
         DO isp = 2, piglet_therm%nsp1
            DO ibead = 1, piglet_therm%p*piglet_therm%ndim
               piglet_therm%smalls(isp, ibead) = smallstmp(i)
               i = i+1
            END DO
         END DO
      ELSE
         DO ibead = 1, piglet_therm%p
            IF (matrix_init == matrix_init_cholesky) THEN
               CALL gle_cholesky_stab(piglet_therm%c_mat(:, :, ibead), & ! Matrix to decompose
                                      Mtmp, & ! Result
                                      piglet_therm%nsp1) ! Size of Matrix
            ELSE IF (matrix_init == matrix_init_diagonal) THEN
               ! Get S by full diagonalization of c_mat matrix
               CALL sqrt_pos_def_mat(piglet_therm%nsp1, & ! Size of the matrix
                                     piglet_therm%c_mat(:, :, ibead), & ! matrix to decompose
                                     Mtmp) ! result
            END IF
            DO idim = 1, piglet_therm%ndim
               i = (ibead-1)*piglet_therm%ndim+idim
               ! Fill a vector with random numbers
               DO j = 1, piglet_therm%nsp1
                  piglet_therm%temp2(j) = next_random_number(piglet_therm%gaussian_rng_stream)
               END DO
               !piglet_therm%smalls(:,i) = MATMUL(Mtmp, piglet_therm%temp2)
               CALL DGEMV('N', & ! Matrix Mtmp is not transposed
                          piglet_therm%nsp1, & ! Number of rows of matrix
                          piglet_therm%nsp1, & ! Number of columns of matrix
                          1.0_dp, & ! scaling of matrix
                          Mtmp(1, 1), & ! Matrix Mtmp
                          piglet_therm%nsp1, & ! leading dimension of Mtmp
                          piglet_therm%temp2, & ! Vector
                          1, & ! increment of vector
                          0.0_dp, & ! scaling of smalls as additive
                          piglet_therm%smalls(1, i), & ! result vector
                          1) ! increment of result vector
            END DO
         END DO
      END IF

      !Fill the array for the sqrt of the masses
      DO idim = 1, piglet_therm%ndim
         DO ibead = 1, piglet_therm%p
            piglet_therm%sqrtmass(ibead, idim) = SQRT(pint_env%mass_fict(ibead, idim))
         END DO
      END DO

      RETURN

   END SUBROUTINE pint_piglet_init

! **************************************************************************************************
!> \brief ...
!> \param vold ...
!> \param vnew ...
!> \param first_mode ...
!> \param masses ...
!> \param piglet_therm ...
! **************************************************************************************************
   SUBROUTINE pint_piglet_step(vold, vnew, first_mode, masses, piglet_therm)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vold, vnew
      INTEGER, INTENT(IN)                                :: first_mode
      REAL(kind=dp), DIMENSION(:, :), INTENT(IN)         :: masses
      TYPE(piglet_therm_type), POINTER                   :: piglet_therm

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_piglet_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ibead, idim, j, ndim, ndimp, &
                                                            nsp1, p
      REAL(KIND=dp)                                      :: alpha, beta, delta_ekin

      CALL timeset(routineN, handle)
      nsp1 = piglet_therm%nsp1
      ndim = piglet_therm%ndim
      p = piglet_therm%p
      ndimp = ndim*p
      ! perform the following operation for all 3N*P
      ! smalls = gle_t*smalls + gle_s*rand_mat
      ! Copy mass scaled momenta to temp1 matrix
      ! p/sqrt(m) we use velocities so v*sqrt(m)
      DO idim = 1, ndim
         DO ibead = first_mode, p
            i = (idim-1)*p+ibead
            piglet_therm%temp1(1, i) = vold(ibead, idim)*piglet_therm%sqrtmass(ibead, idim)
         END DO
      END DO
      ! copy the extra degrees of freedom to the temp1 matrix
      DO j = 1, ndimp
         DO i = 2, nsp1
            piglet_therm%temp1(i, j) = piglet_therm%smalls(i, j)
         END DO
      END DO

      alpha = 1.0_dp
      beta = 0.0_dp
      ! Start the big multiplication loop
      DO idim = 1, ndim
         ! TODO could be accelerated by one big matrix matrix multiplication.
         DO ibead = first_mode, p
            !fill temp2 with gaussian random noise
            DO j = 1, nsp1
               piglet_therm%temp2(j) = next_random_number(piglet_therm%gaussian_rng_stream)
            END DO
            beta = 0.0_dp

            i = (idim-1)*p+ibead
            !smalls(:,i) = alpha * S *temp2 + 0 * smalls
            CALL DGEMM("N", & ! S-matrix should not be transposed
                       "N", & ! tmp2 matrix shoud not be transposed
                       nsp1, & ! Number of rows of S-Matrix
                       1, & ! Number of columns of temp2 vector
                       nsp1, & ! Number of Columns of S-Matrix
                       alpha, & ! Scaling of S-Matrix
                       piglet_therm%gle_s(1, 1, ibead), & ! First element of S-matrix
                       nsp1, & ! Leading dimension of S-matrix
                       piglet_therm%temp2(1), & ! First element of temp2 vector
                       nsp1, & ! Leading dimension of temp2
                       beta, & ! scaling factor of added smalls vector
                       piglet_therm%smalls(1, i), & ! First element of result vector
                       nsp1) ! Leading dimension of result vector
            ! Now add the product of T-matrix * old smalls vectors
            ! smalls (:,i) = alpha * T * temp1 + 1.0 * smalls
            beta = 1.0_dp
            CALL DGEMM("N", & ! T-matrix should not be transposed
                       "N", & ! temp1 matrix shoud not be transposed
                       nsp1, & ! Number of rows of T-Matrix
                       1, & ! Number of columns of temp1 vector
                       nsp1, & ! Number of Columns of T-Matrix
                       alpha, & ! Scaling of T-Matrix
                       piglet_therm%gle_t(1, 1, ibead), & ! First element of T-matrix
                       nsp1, & ! Leading dimension of T-matrix
                       piglet_therm%temp1(1, i), & ! First element of temp1 vector
                       nsp1, & ! Leading dimension of temp1
                       beta, & ! scaling factor of added smalls vector
                       piglet_therm%smalls(1, i), & ! First element of result vector
                       nsp1) ! Leading dimension of result vector
         END DO
      END DO

      ! Copy the mass scales momenta to the outgoing velocities
      delta_ekin = 0.0_dp
      DO idim = 1, ndim
         DO ibead = first_mode, p
            i = (idim-1)*p+ibead
            vnew(ibead, idim) = piglet_therm%smalls(1, i)/piglet_therm%sqrtmass(ibead, idim)
            delta_ekin = delta_ekin+masses(ibead, idim)*( &
                         vnew(ibead, idim)*vnew(ibead, idim)- &
                         vold(ibead, idim)*vold(ibead, idim))
         END DO
      END DO
      ! the piglet is such a strong thermostat, that it messes up the "exact" integration. The thermostats energy will rise lineary, because "it will suck up its own mess" (quote from Michele Ceriotti)
      piglet_therm%thermostat_energy = piglet_therm%thermostat_energy-0.5_dp*delta_ekin

      CALL timestop(handle)

      RETURN

   END SUBROUTINE pint_piglet_step

   ! ***************************************************************************
   !> \brief releases the piglet environment
   !> \param piglet_therm piglet data to be released
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param piglet_therm ...
! **************************************************************************************************
   SUBROUTINE pint_piglet_release(piglet_therm)

      TYPE(piglet_therm_type), POINTER                   :: piglet_therm

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_piglet_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(piglet_therm)) THEN
         piglet_therm%ref_count = piglet_therm%ref_count-1
         IF (piglet_therm%ref_count == 0) THEN
            DEALLOCATE (piglet_therm%a_mat)
            DEALLOCATE (piglet_therm%c_mat)
            DEALLOCATE (piglet_therm%gle_t)
            DEALLOCATE (piglet_therm%gle_s)
            CALL delete_rng_stream(piglet_therm%gaussian_rng_stream)
            DEALLOCATE (piglet_therm%smalls)
            DEALLOCATE (piglet_therm%temp1)
            DEALLOCATE (piglet_therm%temp2)
            DEALLOCATE (piglet_therm%sqrtmass)
            DEALLOCATE (piglet_therm)
         END IF
      END IF
      NULLIFY (piglet_therm)

      RETURN
   END SUBROUTINE pint_piglet_release

   ! ***************************************************************************
   !> \brief adjust the unit of A MAT for piglet
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param a_mat ...
!> \param p ...
!> \param nsp1 ...
!> \param myunit ...
!> \param msg ...
! **************************************************************************************************
   SUBROUTINE a_mat_to_cp2k(a_mat, p, nsp1, myunit, msg)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: a_mat
      INTEGER, INTENT(IN)                                :: p, nsp1
      CHARACTER(LEN=20), INTENT(IN)                      :: myunit
      CHARACTER(default_string_length), INTENT(OUT)      :: msg

      CHARACTER(len=*), PARAMETER :: routineN = 'a_mat_to_cp2k', routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: isunit
      INTEGER                                            :: i, imode, j

      msg = ""
      SELECT CASE (TRIM (myunit))
      CASE ("femtoseconds^-1")
         isunit = "fs^-1"
      CASE ("picoseconds^-1")
         isunit = "ps^-1"
      CASE ("seconds^-1")
         isunit = "s^-1"
      CASE ("atomic time units^-1")
         RETURN
      CASE DEFAULT
         msg = "Unknown unit of A-Matrices for PIGLET. Assuming a.u."
         CALL cp__w("motion/pint_piglet.F",599,msg)
         RETURN
      END SELECT

      DO imode = 1, p
         DO j = 1, nsp1
            DO i = 1, nsp1
               a_mat(i, j, imode) = cp_unit_to_cp2k(a_mat(i, j, imode), TRIM(isunit))
            END DO
         END DO
      END DO

      RETURN

   END SUBROUTINE
! ***************************************************************************
!> \brief adjust the unit of C MAT for piglet
!> \param c_mat ...
!> \param p ...
!> \param nsp1 ...
!> \param myunit ...
!> \param msg ...
!> \author Felix Uhl
! **************************************************************************************************
   SUBROUTINE c_mat_to_cp2k(c_mat, p, nsp1, myunit, msg)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: c_mat
      INTEGER, INTENT(IN)                                :: p, nsp1
      CHARACTER(LEN=20), INTENT(IN)                      :: myunit
      CHARACTER(default_string_length), INTENT(OUT)      :: msg

      CHARACTER(len=*), PARAMETER :: routineN = 'c_mat_to_cp2k', routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: isunit
      INTEGER                                            :: i, imode, j

      msg = ""
      SELECT CASE (TRIM (myunit))
      CASE ("eV")
         isunit = "eV"
      CASE ("K")
         isunit = "K_e"
      CASE ("atomic energy units ")
         RETURN
      CASE DEFAULT
         msg = "Unknown unit of C-Matrices for PIGLET. Assuming a.u."
         CALL cp__w("motion/pint_piglet.F",645,msg)
         RETURN
      END SELECT

      DO imode = 1, p
         DO j = 1, nsp1
            DO i = 1, nsp1
               c_mat(i, j, imode) = cp_unit_to_cp2k(c_mat(i, j, imode), TRIM(isunit))
            END DO
         END DO
      END DO

      RETURN

   END SUBROUTINE c_mat_to_cp2k

! ***************************************************************************
!> \brief checks if the matrices are suited for the target temperature
!> \param line ...
!> \param propagator ...
!> \param targettemp ...
!> \author Felix Uhl
! **************************************************************************************************
   SUBROUTINE check_temperature(line, propagator, targettemp)

      CHARACTER(len=*), INTENT(IN)                       :: line
      INTEGER, INTENT(IN)                                :: propagator
      REAL(KIND=dp), INTENT(IN)                          :: targettemp

      CHARACTER(len=*), PARAMETER :: routineN = 'check_temperature', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msg
      INTEGER                                            :: posnumber
      REAL(KIND=dp)                                      :: convttemp, deviation, matrixtemp, ttemp

      deviation = 100.0d0
      posnumber = INDEX(line, "T=")+2
      IF (propagator == propagator_rpmd) ttemp = targettemp
      !Get the matrix temperature
      READ (line(posnumber:), *) matrixtemp
      msg = ""
      IF (INDEX(line, "K") /= 0) THEN
         convttemp = cp_unit_from_cp2k(ttemp, "K")
         IF (ABS(convttemp-matrixtemp) > convttemp/deviation) THEN
            WRITE (UNIT=msg, FMT=*) "PIGLET Simulation temperature (", &
               convttemp, "K) /= matrix temperature (", matrixtemp, "K)"
            CALL cp__w("motion/pint_piglet.F",692,msg)
         END IF
      ELSE IF (INDEX(line, "eV") /= 0) THEN
         convttemp = cp_unit_from_cp2k(ttemp, "K")/11604.505_dp
         IF (ABS(convttemp-matrixtemp) > convttemp/deviation) THEN
            WRITE (UNIT=msg, FMT=*) "PIGLET Simulation temperature (", &
               convttemp, "K) /= matrix temperature (", matrixtemp, "K)"
            CALL cp__w("motion/pint_piglet.F",699,msg)
         END IF
      ELSE IF (INDEX(line, "atomic energy units") /= 0) THEN
         convttemp = ttemp
         IF (ABS(convttemp-matrixtemp) > convttemp/deviation) THEN
            WRITE (UNIT=msg, FMT=*) "PIGLET Simulation temperature (", &
               convttemp, "K) /= matrix temperature (", matrixtemp, "K)"
            CALL cp__w("motion/pint_piglet.F",706,msg)
         END IF
      ELSE
         WRITE (UNIT=msg, FMT=*) "Unknown PIGLET matrix temperature. Assuming a.u."
         CALL cp__w("motion/pint_piglet.F",710,msg)
         convttemp = ttemp
         IF (ABS(convttemp-matrixtemp) > convttemp/deviation) THEN
            WRITE (UNIT=msg, FMT=*) "PIGLET Simulation temperature (", &
               convttemp, "K) /= matrix temperature (", matrixtemp, "K)"
            CALL cp__w("motion/pint_piglet.F",715,msg)
         END IF
      END IF

      RETURN

   END SUBROUTINE check_temperature

   ! ***************************************************************************
   !> \brief returns the piglet kinetic energy contribution
   !> \author Felix Uhl
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_calc_piglet_energy(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_piglet_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pint_env%piglet_therm)) THEN
         pint_env%e_piglet = pint_env%piglet_therm%thermostat_energy
      ELSE
         pint_env%e_piglet = 0.0d0
      END IF

      RETURN

   END SUBROUTINE pint_calc_piglet_energy

   ! ***************************************************************************
   !> \brief calculates S from S*TRANSPOSED(S) by diagonalizing
   !>        if S*TRANSPOSED(S) is a positive definite matrix
   !> \param n order of input matrix
   !> \param SST matrix to be decomposed
   !> \param S result matrix
   !> \author Felix Uhl
   !> \param LOG: 11-08-2015: Replaced DSYEVR by DSYEV (Felix Uhl)
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param SST ...
!> \param S ...
! **************************************************************************************************
   SUBROUTINE sqrt_pos_def_mat(n, SST, S)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(n, n), INTENT(IN)         :: SST
      REAL(KIND=dp), DIMENSION(n, n), INTENT(OUT)        :: S

      CHARACTER(len=*), PARAMETER :: routineN = 'sqrt_pos_def_mat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, lwork
      REAL(KIND=dp)                                      :: tmplwork
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work
      REAL(KIND=dp), DIMENSION(n)                        :: eigval
      REAL(KIND=dp), DIMENSION(n, n)                     :: A, tmpmatrix

! order of input matrix
! matrix to be decomposed
! result matrix
! Variables for eigenvalue/vector computation
! store matrix here to pass to lapack routine DSYEVR
! Array to contain the eigenvalues
! size of temporary real work array
! temporary real work array
! information about success
! counter

      eigval(:) = 0.0_dp
      A(:, :) = 0.0_dp
      A(:, :) = SST(:, :)

      !first call to figure out how big work array needs to be
      CALL DSYEV('V', & ! Compute eigenvalues and eigenvectors
                 'U', & ! Store upper triagonal matrix
                 n, & ! order of matrix A to calculate the eigenvalues/vectors from
                 A, & ! Matrix to calculate the eigenvalues/vectors from
                 n, & ! leading order of matrix A
                 eigval, & ! Array to contain the eigenvalues
                 tmplwork, & ! temporary real work array
                 -1, & ! size of temporary real work array
                 info) ! information about success

      lwork = INT(tmplwork+0.5)
      ALLOCATE (work(lwork))
      work(:) = 0.0_dp

      CALL DSYEV('V', & ! Compute eigenvalues and eigenvectors
                 'U', & ! Store upper triagonal matrix
                 n, & ! order of matrix A to calculate the eigenvalues/vectors from
                 A, & ! Matrix to calculate the eigenvalues/vectors from
                 n, & ! leading order of matrix A
                 eigval, & ! Array to contain the eigenvalues
                 work, & ! temporary real work array
                 lwork, & ! size of temporary real work array
                 info) ! information about success
      DEALLOCATE (work)
      ! A-matrix now contains the eigenvectors

      S(:, :) = 0.0_dp
      DO i = 1, n
         ! In case that numerics made some eigenvalues negative
         IF (eigval(i) > 0.0_dp) THEN
            S(i, i) = SQRT(eigval(i))
         END IF
      END DO

      tmpmatrix(:, :) = 0.0_dp
      ! Transform matrix back
      !tmpmatrix = A*S
      CALL DGEMM('N', & ! A-matrix is not transposed
                 'N', & ! S-matrix is not transposed
                 n, & ! number of rows of A-matrix
                 n, & ! number of columns of S-matrix
                 n, & ! number of columns of A-matrix and number of rows of S-matrix
                 1.0D0, & ! scaling factor of A-matrix
                 A, & ! A-matrix
                 n, & ! leading dimension of A-matrix as declared
                 S, & ! S-matrix
                 n, & ! leading dimension of S-matrix as declared
                 0.0D0, & ! scaling of tmpmatrix as additive
                 tmpmatrix, & ! result matrix: tmpmatrix
                 n) ! leading dimension of tmpmatrix
      !S = tmpmatrix*TRANSPOSE(A) = A*S*TRANSPOSE(A)
      CALL DGEMM('N', & ! tmpmatrix not transposed
                 'T', & ! A-matrix is transposed
                 n, & ! number of rows of tmpmatrix
                 n, & ! number of columns of A-matrix
                 n, & ! number of columns of tmpmatrix and rows of A-matrix
                 1.0D0, & ! scaling factor of tmpmatrix
                 tmpmatrix, & ! tmpmatrix
                 n, & ! leading dimension of tmpmatrix as declared
                 A, & ! A-matrix
                 n, & ! leading dimension of A-matrix as declared
                 0.0D0, & ! scaling of S-matrix as additive
                 S, & ! result matrix: S-matrix
                 n) ! leading dimension of S-matrix as declared

      RETURN

   END SUBROUTINE sqrt_pos_def_mat

END MODULE pint_piglet
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Independent helium subroutines shared with other modules
!> \author Lukasz Walewski
!> \date   2009-07-14
!> \note   Avoiding circular deps: do not USE any other helium_* modules here.
! **************************************************************************************************
MODULE helium_common

   USE helium_types,                    ONLY: helium_solvent_type,&
                                              int_arr_ptr,&
                                              rho_atom_number,&
                                              rho_moment_of_inertia,&
                                              rho_num,&
                                              rho_projected_area,&
                                              rho_winding_cycle,&
                                              rho_winding_number
   USE input_constants,                 ONLY: denominator_inertia,&
                                              denominator_natoms,&
                                              denominator_rperp2,&
                                              helium_cell_shape_cube,&
                                              helium_cell_shape_octahedron
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE parallel_rng_types,              ONLY: next_random_number
   USE physcon,                         ONLY: angstrom,&
                                              bohr
   USE pint_public,                     ONLY: pint_calc_centroid,&
                                              pint_com_pos
   USE pint_types,                      ONLY: pint_env_type
   USE splines_methods,                 ONLY: spline_value
   USE splines_types,                   ONLY: spline_data_p_type,&
                                              spline_data_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_common'

   PUBLIC :: helium_pbc
   PUBLIC :: helium_boxmean_3d
   PUBLIC :: helium_calc_rdf
   PUBLIC :: helium_calc_rho
   PUBLIC :: helium_calc_plength
   PUBLIC :: helium_rotate
   PUBLIC :: helium_eval_expansion
   PUBLIC :: helium_update_transition_matrix
   PUBLIC :: helium_spline
   PUBLIC :: helium_cycle_number
   PUBLIC :: helium_path_length
   PUBLIC :: helium_is_winding
   PUBLIC :: helium_cycle_of
   PUBLIC :: helium_total_winding_number
   PUBLIC :: helium_total_projected_area
   PUBLIC :: helium_link_projected_area
   PUBLIC :: helium_cycle_projected_area
   PUBLIC :: helium_cycle_projected_area_pbc
   PUBLIC :: helium_total_moment_of_inertia
   PUBLIC :: helium_total_moment_of_inertia_linkwise
   PUBLIC :: helium_com
   PUBLIC :: helium_total_projected_area_linkwise
   PUBLIC :: helium_link_vector
   PUBLIC :: helium_calc_cycles
   PUBLIC :: helium_update_coord_system
   PUBLIC :: helium_set_rdf_coord_system

CONTAINS

! ***************************************************************************
!> \brief  General PBC routine for helium.
!> \param helium ...
!> \param r ...
!> \param enforce ...
!> \date   2009-09-30
!> \author Lukasz Walewski
!> \note  Check wheather PBC should be applied, if yes call low level
!>         routine according to the unit cell shape.
! **************************************************************************************************
   SUBROUTINE helium_pbc(helium, r, enforce)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: r
      LOGICAL, OPTIONAL                                  :: enforce

      IF (.NOT. helium%periodic) THEN
         IF (.NOT. PRESENT(enforce)) THEN
            RETURN
         END IF
      END IF

      SELECT CASE (helium%cell_shape)
      CASE (helium_cell_shape_cube)
         CALL helium_pbc_cube(helium, r)
      CASE (helium_cell_shape_octahedron)
         CALL helium_pbc_trocta_opt(helium, r)
      CASE DEFAULT
         CALL cp__b("motion/helium_common.F",105,"we should never get here")
      END SELECT

      RETURN
   END SUBROUTINE helium_pbc

! ***************************************************************************
!> \brief  Wrap r back to the helium box (if helium box is periodic)
!> \param    helium - helium environment for which to wrap
!> \param    r      - 3D vector to be wraped back to periodic box
!> \par History
!>      2009-10-02 renamed, originally was helium_box
!>      2009-10-02 redesigned so it is now called as a subroutine [lwalewski]
!>      2009-10-02 redesigned so it now gets/returns a 3D vector [lwalewski]
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_pbc_cube(helium, r)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: r

      REAL(KIND=dp)                                      :: s

! x coordinate

      s = helium%cell_size_inv*r(1)
      IF (s > 0.5_dp) THEN
         s = s-REAL(INT(s+0.5_dp), dp)
      ELSEIF (s < -0.5_dp) THEN
         s = s-REAL(INT(s-0.5_dp), dp)
      END IF
      r(1) = s*helium%cell_size

      ! y coordinate
      s = helium%cell_size_inv*r(2)
      IF (s > 0.5_dp) THEN
         s = s-REAL(INT(s+0.5_dp), dp)
      ELSEIF (s < -0.5_dp) THEN
         s = s-REAL(INT(s-0.5_dp), dp)
      END IF
      r(2) = s*helium%cell_size

      ! z coordinate
      s = helium%cell_size_inv*r(3)
      IF (s > 0.5_dp) THEN
         s = s-REAL(INT(s+0.5_dp), dp)
      ELSEIF (s < -0.5_dp) THEN
         s = s-REAL(INT(s-0.5_dp), dp)
      END IF
      r(3) = s*helium%cell_size

      RETURN
   END SUBROUTINE helium_pbc_cube


! ***************************************************************************
!> \brief  Apply PBC within truncated octahedral unit cell.
!> \param helium ...
!> \param r ...
!> \date   2009-10-22
!> \author hforbert
!> \note   Version of the original Allen & Tildesley routine optimized for
!>         g95 and intel compilers on x86-64.
! **************************************************************************************************
   SUBROUTINE helium_pbc_trocta_opt(helium, r)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: r

      REAL(KIND=dp)                                      :: cell_size, cell_size_inv, corr, rx, ry, &
                                                            rz, sx, sy, sz

      cell_size = helium%cell_size
      cell_size_inv = helium%cell_size_inv

      rx = r(1)*cell_size_inv
      IF (rx > 0.5_dp) THEN
         rx = rx-REAL(INT(rx+0.5_dp), dp)
      ELSEIF (rx < -0.5_dp) THEN
         rx = rx-REAL(INT(rx-0.5_dp), dp)
      END IF

      ry = r(2)*cell_size_inv
      IF (ry > 0.5_dp) THEN
         ry = ry-REAL(INT(ry+0.5_dp), dp)
      ELSEIF (ry < -0.5_dp) THEN
         ry = ry-REAL(INT(ry-0.5_dp), dp)
      END IF

      rz = r(3)*cell_size_inv
      IF (rz > 0.5_dp) THEN
         rz = rz-REAL(INT(rz+0.5_dp), dp)
      ELSEIF (rz < -0.5_dp) THEN
         rz = rz-REAL(INT(rz-0.5_dp), dp)
      END IF

      corr = 0.0_dp
      IF (rx > 0.0_dp) THEN
         corr = corr+rx
         sx = 0.5_dp
      ELSE
         corr = corr-rx
         sx = -0.5_dp
      END IF
      IF (ry > 0.0_dp) THEN
         corr = corr+ry
         sy = 0.5_dp
      ELSE
         corr = corr-ry
         sy = -0.5_dp
      END IF
      IF (rz > 0.0_dp) THEN
         corr = corr+rz
         sz = 0.5_dp
      ELSE
         corr = corr-rz
         sz = -0.5_dp
      END IF
      IF (corr > 0.75_dp) THEN
         rx = rx-sx
         ry = ry-sy
         rz = rz-sz
      END IF

      r(1) = rx*cell_size
      r(2) = ry*cell_size
      r(3) = rz*cell_size

      RETURN
   END SUBROUTINE helium_pbc_trocta_opt

! ***************************************************************************
!> \brief  Calculate the point equidistant from two other points a and b
!>         within the helium box - 3D version
!> \param    helium - helium environment for which
!> \param a vectors for which to find the mean within the He box
!> \param b vectors for which to find the mean within the He box
!> \param c ...
!> \par History
!>      2009-10-02 renamed, originally was helium_boxmean [lwalewski]
!>      2009-10-02 redesigned so it is now called as a subroutine [lwalewski]
!>      2009-10-02 redesigned so it now gets/returns a 3D vectors [lwalewski]
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_boxmean_3d(helium, a, b, c)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a, b
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: c

      c(:) = b(:)-a(:)
      CALL helium_pbc(helium, c)
      c(:) = a(:)+0.5_dp*c(:)
      CALL helium_pbc(helium, c)
      RETURN
   END SUBROUTINE helium_boxmean_3d

! ***************************************************************************
!> \brief  Given the permutation state assign cycle lengths to all He atoms.
!> \param helium ...
!> \date   2011-07-06
!> \author Lukasz Walewski
!> \note   The helium%atom_plength array is filled with cycle lengths,
!>         each atom gets the length of the permutation cycle it belongs to.
! **************************************************************************************************
   SUBROUTINE helium_calc_atom_cycle_length(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_atom_cycle_length', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str
      INTEGER                                            :: clen, curr_idx, handle, ia, start_idx
      INTEGER, DIMENSION(:), POINTER                     :: atoms_in_cycle
      LOGICAL                                            :: atoms_left, path_end_reached
      LOGICAL, DIMENSION(:), POINTER                     :: atom_was_used

      CALL timeset(routineN, handle)

      NULLIFY (atoms_in_cycle)
      atoms_in_cycle => helium%itmp_atoms_1d
      atoms_in_cycle(:) = 0

      NULLIFY (atom_was_used)
      atom_was_used => helium%ltmp_atoms_1d
      atom_was_used(:) = .FALSE.

      helium%atom_plength(:) = 0

      start_idx = 1
      DO
         clen = 0
         path_end_reached = .FALSE.
         curr_idx = start_idx
         DO ia = 1, helium%atoms
            clen = clen+1
            atoms_in_cycle(clen) = curr_idx
            atom_was_used(curr_idx) = .TRUE.

            ! follow to the next atom in the cycle
            curr_idx = helium%permutation(curr_idx)
            IF (curr_idx .EQ. start_idx) THEN
               path_end_reached = .TRUE.
               EXIT
            END IF
         END DO
         err_str = "Permutation path is not cyclic."
         IF (.NOT. path_end_reached) THEN
            CALL cp__b("motion/helium_common.F",359,err_str)
         END IF

         ! assign the cycle length to the collected atoms
         DO ia = 1, clen
            helium%atom_plength(atoms_in_cycle(ia)) = clen
         END DO

         ! go to the next "unused" atom
         atoms_left = .FALSE.
         DO ia = 1, helium%atoms
            IF (.NOT. atom_was_used(ia)) THEN
               start_idx = ia
               atoms_left = .TRUE.
               EXIT
            END IF
         END DO

         IF (.NOT. atoms_left) EXIT
      END DO

      atoms_in_cycle(:) = 0
      NULLIFY (atoms_in_cycle)
      atom_was_used(:) = .FALSE.
      NULLIFY (atom_was_used)

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_calc_atom_cycle_length

! ***************************************************************************
!> \brief  Decompose a permutation into cycles
!> \param permutation ...
!> \retval cycles ...
!> \date   2013-12-11
!> \author Lukasz Walewski
!> \note   Given a permutation return a pointer to an array of pointers,
!>         with each element pointing to a cycle of this permutation.
!> \note   This function allocates memory and returns a pointer to it,
!>         do not forget to deallocate once finished with the results.
! **************************************************************************************************
   FUNCTION helium_calc_cycles(permutation) RESULT(cycles)

      INTEGER, DIMENSION(:), POINTER                     :: permutation
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: cycles

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_cycles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: curat, ncycl, nsize, nused
      INTEGER, DIMENSION(:), POINTER                     :: cur_cycle, used_indices
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: my_cycles

      nsize = SIZE(permutation)

      ! most pesimistic scenario: no cycles longer than 1
      ALLOCATE (my_cycles(nsize))

      ! go over the permutation and identify cycles
      curat = 1
      nused = 0
      ncycl = 0
      DO WHILE (curat .LE. nsize)

         ! get the permutation cycle the current atom belongs to
         cur_cycle => helium_cycle_of(curat, permutation)

         ! include the current cycle in the pool of "used" indices
         nused = nused+SIZE(cur_cycle)
         CALL reallocate(used_indices, 1, nused)
         used_indices(nused-SIZE(cur_cycle)+1:nused) = cur_cycle(1:SIZE(cur_cycle))

         ! store the pointer to the current cycle
         ncycl = ncycl+1
         my_cycles(ncycl)%iap => cur_cycle

         ! free the local pointer
         NULLIFY (cur_cycle)

         ! try to increment the current atom index
         DO WHILE (ANY(used_indices .EQ. curat))
            curat = curat+1
         END DO

      END DO

      DEALLOCATE (used_indices)
      NULLIFY (used_indices)

      ! assign the result
      ALLOCATE (cycles(ncycl))
      cycles(1:ncycl) = my_cycles(1:ncycl)

      DEALLOCATE (my_cycles)
      NULLIFY (my_cycles)

      RETURN
   END FUNCTION helium_calc_cycles

! ***************************************************************************
!> \brief  Calculate helium density distribution functions and store them
!>         in helium%rho_inst
!> \param helium ...
!> \date   2011-06-14
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_calc_rho(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_rho', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msgstr
      INTEGER                                            :: aa, bx, by, bz, handle, ia, ib, ic, id, &
                                                            ii, ir, n_out_of_range, nbin
      INTEGER, DIMENSION(3)                              :: nw
      INTEGER, DIMENSION(:), POINTER                     :: cycl_len
      LOGICAL                                            :: ltmp1, ltmp2, ltmp3
      REAL(KIND=dp)                                      :: invd3r, invdr, invp, rtmp
      REAL(KIND=dp), DIMENSION(3)                        :: maxr_half, r, vlink, vtotal, wn
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: cycles

      CALL timeset(routineN, handle)

      maxr_half(:) = helium%rho_maxr/2.0_dp
      invdr = 1.0_dp/helium%rho_delr
      invd3r = invdr*invdr*invdr
      invp = 1.0_dp/REAL(helium%beads, dp)
      nbin = helium%rho_nbin
      ! assign the cycle lengths to all the atoms
      CALL helium_calc_atom_cycle_length(helium)
      NULLIFY (cycl_len)
      cycl_len => helium%atom_plength
      DO ir = 1, rho_num ! loop over densities ---

         IF (.NOT. helium%rho_property(ir)%is_calculated) THEN
            ! skip densities that are not requested by the user
            CYCLE
         END IF

         SELECT CASE (ir)

         CASE (rho_atom_number)
            ii = helium%rho_property(ir)%component_index(1)
            helium%rho_incr(ii, :, :) = invp

         CASE (rho_projected_area)
            vtotal(:) = helium_total_projected_area(helium)
            DO ia = 1, helium%atoms
               DO ib = 1, helium%beads
                  vlink(:) = helium_link_projected_area(helium, ia, ib)
                  DO ic = 1, 3
                     ii = helium%rho_property(ir)%component_index(ic)
                     helium%rho_incr(ii, ia, ib) = vtotal(ic)*vlink(ic)*angstrom*angstrom*angstrom*angstrom
                  END DO
               END DO
            END DO

!        CASE (rho_winding_number)
!          vtotal(:) = helium_total_winding_number(helium)
!          DO ia = 1, helium%atoms
!            DO ib = 1, helium%beads
!              vlink(:) = helium_link_winding_number(helium,ia,ib)
!              DO ic = 1, 3
!                ii = helium%rho_property(ir)%component_index(ic)
!                helium%rho_incr(ii,ia,ib) = vtotal(ic)*vlink(ic)*angstrom*angstrom
!              END DO
!            END DO
!          END DO

         CASE (rho_winding_number)
            vtotal(:) = helium_total_winding_number(helium)
            DO id = 1, 3
               ii = helium%rho_property(ir)%component_index(id)
               helium%rho_incr(ii, :, :) = 0.0_dp
            END DO
            NULLIFY (cycles)
            cycles => helium_calc_cycles(helium%permutation)
            DO ic = 1, SIZE(cycles)
               wn = helium_cycle_winding_number(helium, cycles(ic)%iap, helium%pos)
               DO ia = 1, SIZE(cycles(ic)%iap)
                  aa = cycles(ic)%iap(ia)
                  DO ib = 1, helium%beads
                     vlink(:) = helium_link_winding_number(helium, aa, ib)
                     DO id = 1, 3
                        IF (ABS(wn(id)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                           ii = helium%rho_property(ir)%component_index(id)
                           helium%rho_incr(ii, aa, ib) = vtotal(id)*vlink(id)*angstrom*angstrom
                        END IF
                     END DO
                  END DO
               END DO
            END DO
            DEALLOCATE (cycles)

         CASE (rho_winding_cycle)
            vtotal(:) = helium_total_winding_number(helium)
            DO id = 1, 3
               ii = helium%rho_property(ir)%component_index(id)
               helium%rho_incr(ii, :, :) = 0.0_dp
            END DO
            NULLIFY (cycles)
            cycles => helium_calc_cycles(helium%permutation)
            ! compute number of atoms in all winding cycles
            nw(:) = 0
            DO ic = 1, SIZE(cycles)
               wn = helium_cycle_winding_number(helium, cycles(ic)%iap, helium%pos)
               DO id = 1, 3
                  IF (ABS(wn(id)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                     nw(id) = nw(id)+SIZE(cycles(ic)%iap)
                  END IF
               END DO
            END DO
            ! assign contributions to all beads of winding cycles only
            DO ic = 1, SIZE(cycles)
               wn = helium_cycle_winding_number(helium, cycles(ic)%iap, helium%pos)
               DO id = 1, 3
                  IF (ABS(wn(id)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                     DO ia = 1, SIZE(cycles(ic)%iap)
                        aa = cycles(ic)%iap(ia)
                        DO ib = 1, helium%beads
                           IF (nw(id) .GT. 0) THEN ! this test should always get passed
                              ii = helium%rho_property(ir)%component_index(id)
                              rtmp = invp/REAL(nw(id), dp)
                              helium%rho_incr(ii, aa, ib) = rtmp*vtotal(id)*vtotal(id)*angstrom*angstrom
                           END IF
                        END DO
                     END DO
                  END IF
               END DO
            END DO
            DEALLOCATE (cycles)

         CASE (rho_moment_of_inertia)
            vtotal(:) = helium_total_moment_of_inertia(helium)
            DO ia = 1, helium%atoms
               DO ib = 1, helium%beads
                  vlink(:) = helium_link_moment_of_inertia(helium, ia, ib)
                  DO ic = 1, 3
                     ii = helium%rho_property(ir)%component_index(ic)
                     helium%rho_incr(ii, ia, ib) = vlink(ic)*angstrom*angstrom
                  END DO
               END DO
            END DO

         CASE DEFAULT
            ! do nothing

         END SELECT

      END DO ! loop over densities ---

      n_out_of_range = 0
      helium%rho_inst(:, :, :, :) = 0.0_dp
      DO ia = 1, helium%atoms
         ! bin the bead positions of the current atom using the increments set above
         DO ib = 1, helium%beads
            ! map the current bead position to the corresponding voxel
            r(:) = helium%pos(:, ia, ib)-helium%center(:)
            ! enforce PBC even if this is a non-periodic calc to avoid density leakage
            CALL helium_pbc(helium, r, enforce=.TRUE.)
            ! set up bin indices (translate by L/2 to avoid non-positive array indices)
            bx = INT((r(1)+maxr_half(1))*invdr)+1
            by = INT((r(2)+maxr_half(2))*invdr)+1
            bz = INT((r(3)+maxr_half(3))*invdr)+1
            ! check that the resulting bin numbers are within array bounds
            ltmp1 = (0 .LT. bx) .AND. (bx .LE. nbin)
            ltmp2 = (0 .LT. by) .AND. (by .LE. nbin)
            ltmp3 = (0 .LT. bz) .AND. (bz .LE. nbin)
            IF (ltmp1 .AND. ltmp2 .AND. ltmp3) THEN
               ! increment all the estimators (those that the current atom does not
               ! contribute to have increment incr(ic)==0)
               DO ic = 1, helium%rho_num_act
                  helium%rho_inst(ic, bx, by, bz) = helium%rho_inst(ic, bx, by, bz)+helium%rho_incr(ic, ia, ib)
               END DO
            ELSE
               n_out_of_range = n_out_of_range+1
            END IF
         END DO
      END DO
      ! scale by volume element
      helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)*invd3r

      ! stop if any bead fell out of the range
      ! since enforced PBC should have taken care of such leaks
      WRITE (msgstr, *) n_out_of_range
      msgstr = "Number of bead positions out of range: "//TRIM(ADJUSTL(msgstr))
      IF (n_out_of_range .GT. 0) THEN
         CALL cp__b("motion/helium_common.F",649,msgstr)
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_calc_rho


! ***************************************************************************
!> \brief  Calculate helium radial distribution functions wrt positions given
!>         by <centers> using the atomic weights given by <weights>.
!> \param helium ...
!> \param centers ...
!> \date   2009-07-22
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_calc_rdf(helium, centers)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(:), POINTER               :: centers

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_rdf', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msgstr
      INTEGER                                            :: bin, handle, ia, ib, ic, id, &
                                                            n_out_of_range, nbin
      REAL(KIND=dp)                                      :: const, invdr, invp, nideal, pi, ri, &
                                                            rlower, rupper, sdens
      REAL(KIND=dp), DIMENSION(3)                        :: r, r0
      REAL(KIND=dp), DIMENSION(:), POINTER               :: incr

      CALL timeset(routineN, handle)

      pi = 4.0_dp*ATAN(1.0_dp)
      invdr = 1.0_dp/helium%rdf_delr
      invp = 1.0_dp/helium%beads
      nbin = helium%rdf_nbin
      sdens = helium%wpref*invp*helium%atoms
      ALLOCATE (incr(helium%rdf_num))
      incr(:) = 0.0d0

      ! calculate the histogram of distances
      n_out_of_range = 0
      helium%rdf_inst(:, :, :) = 0.0_dp
      DO ic = 1, SIZE(centers)/3

         r0(1) = centers(3*(ic-1)+1)
         r0(2) = centers(3*(ic-1)+2)
         r0(3) = centers(3*(ic-1)+3)
         DO ia = 1, helium%atoms

            ! set the increments for this atom
            incr(1) = invp

            DO ib = 1, helium%beads
               r(:) = helium%pos(:, ia, ib)-r0(:)
               CALL helium_pbc(helium, r)
               ri = SQRT(r(1)*r(1)+r(2)*r(2)+r(3)*r(3))
               bin = INT(ri*invdr)+1
               IF ((0 .LT. bin) .AND. (bin .LE. nbin)) THEN
                  ! increment the RDF value for He atoms inside the r_6 sphere
                  DO id = 1, helium%rdf_num
                     helium%rdf_inst(id, bin, ic) = helium%rdf_inst(id, bin, ic)+incr(id)
                  END DO
               ELSE
                  !WRITE(msgstr,*) helium%center * angstrom
                  !msgstr = "center = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  !WRITE(msgstr,*) r0 * angstrom
                  !msgstr = "r0     = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  !WRITE(msgstr,*) helium%pos(:,ia,ib) * angstrom
                  !msgstr = "pos    = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  !WRITE(msgstr,*) ri * angstrom
                  !msgstr = "ri     = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  !WRITE(msgstr,*) ri * invdr
                  !msgstr = "ri/dr  = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  !WRITE(msgstr,*) helium%current_step
                  !msgstr = "step   = "//TRIM(ADJUSTL(msgstr))
                  !CALL cp_error_message(cp_warning_level, routineP, msgstr, error)
                  n_out_of_range = n_out_of_range+1
               END IF
            END DO
         END DO
      END DO

      IF (.NOT. helium%periodic) THEN
         IF (n_out_of_range .GT. 0) THEN
            WRITE (msgstr, *) n_out_of_range
            msgstr = "Number of bead positions out of range: "//TRIM(ADJUSTL(msgstr))
            CALL cp__b("motion/helium_common.F",806,msgstr)
         END IF
      END IF
      ! for periodic case we intentionally truncate RDFs to spherical volume
      ! so we skip atoms in the corners

      ! normalize the histogram to get g(r)
      ! note: helium%density refers to the number of atoms, not the beads
      const = 4.0_dp*pi*helium%density/3.0_dp
      DO bin = 1, helium%rdf_nbin
         rlower = REAL(bin-1, dp)*helium%rdf_delr
         rupper = rlower+helium%rdf_delr
         nideal = const*(rupper**3-rlower**3)
         DO id = 1, helium%rdf_num
            helium%rdf_inst(id, bin, :) = helium%rdf_inst(id, bin, :)/nideal
         END DO
      END DO

      DEALLOCATE (incr)
      NULLIFY (incr)

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_calc_rdf

! ***************************************************************************
!> \brief  Calculate probability distribution of the permutation lengths
!> \param helium ...
!> \date   2010-06-07
!> \author Lukasz Walewski
!> \note   Valid permutation path length is an integer (1, NATOMS), number
!>         of paths of a given length is calculated here and average over
!>         inner loop iterations and helium environments is done in
!>         helium_sample.
! **************************************************************************************************
   SUBROUTINE helium_calc_plength(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_plength', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k

      helium%plength_inst(:) = 0.0_dp
      DO i = 1, helium%atoms
         j = helium%permutation(i)
         k = 1
         DO
            IF (j == i) EXIT
            k = k+1
            j = helium%permutation(j)
         END DO
         helium%plength_inst(k) = helium%plength_inst(k)+1
      END DO
      helium%plength_inst(:) = helium%plength_inst(:)/helium%atoms

      RETURN
   END SUBROUTINE helium_calc_plength

! ***************************************************************************
!> \brief  Rotate helium particles in imaginary time by nslices
!> \param helium ...
!> \param nslices ...
!> \author hforbert
!> \note   Positions of helium beads in helium%pos array are reorganized such
!>         that the indices are cyclically translated in a permutation-aware
!>         manner. helium%relrot is given a new value that represents the new
!>         'angle' of the beads. This is done modulo helium%beads, so relrot
!>         should be always within 0 (no rotation) and helium%beads-1 (almost
!>         full rotation). [lwalewski]
! **************************************************************************************************
   SUBROUTINE helium_rotate(helium, nslices)
      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: nslices

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rotate', routineP = moduleN//':'//routineN

      INTEGER                                            :: b, i, j, k, n

      b = helium%beads
      n = helium%atoms
      i = MOD(nslices, b)
      IF (i < 0) i = i+b
      IF ((i >= b) .OR. (i < 1)) RETURN
      helium%relrot = MOD(helium%relrot+i, b)
      DO k = 1, i
         helium%work(:, :, k) = helium%pos(:, :, k)
      END DO
      DO k = i+1, b
         helium%pos(:, :, k-i) = helium%pos(:, :, k)
      END DO
      DO k = 1, i
         DO j = 1, n
            helium%pos(:, j, b-i+k) = helium%work(:, helium%permutation(j), k)
         END DO
      END DO
      RETURN
   END SUBROUTINE helium_rotate

! ***************************************************************************
!> \brief  Calculate the pair-product action or energy by evaluating the
!>         power series expansion according to Eq. 4.46 in Ceperley 1995.
!> \param helium ...
!> \param r ...
!> \param rp ...
!> \param tab ...
!> \param cut ...
!> \retval res ...
!> \author Harald Forbert
! **************************************************************************************************
   FUNCTION helium_eval_expansion(helium, r, rp, tab, cut) RESULT(res)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r, rp
      TYPE(spline_data_p_type), DIMENSION(:, :), POINTER :: tab
      INTEGER, INTENT(IN)                                :: cut
      REAL(KIND=dp)                                      :: res

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: ar, arp, q, s, v, z
      REAL(KIND=dp), DIMENSION(3)                        :: br, brp

      br(:) = r(:)
      brp(:) = rp(:)
      CALL helium_pbc(helium, br)
      CALL helium_pbc(helium, brp)

      ar = SQRT(br(1)**2+br(2)**2+br(3)**2)
      arp = SQRT(brp(1)**2+brp(2)**2+brp(3)**2)
      q = 0.5_dp*(ar+arp)

      IF (helium%periodic .AND. ((ar > 0.5_dp*helium%cell_size) &
                                 .OR. (arp > 0.5_dp*helium%cell_size))) THEN
         v = 0.0_dp
         IF (arp > 0.5_dp*helium%cell_size) THEN
            v = v+REAL(cut, dp)*helium_spline(tab(1, 1)%spline_data, 0.5_dp*helium%cell_size)
         ELSE
            v = v+helium_spline(tab(1, 1)%spline_data, arp)
         END IF
         IF (ar > 0.5_dp*helium%cell_size) THEN
            v = v+REAL(cut, dp)*helium_spline(tab(1, 1)%spline_data, 0.5_dp*helium%cell_size)
         ELSE
            v = v+helium_spline(tab(1, 1)%spline_data, ar)
         END IF
         res = 0.5_dp*v
      ELSE
         ! end-point action (first term):
         v = 0.5_dp*(helium_spline(tab(1, 1)%spline_data, ar)+helium_spline(tab(1, 1)%spline_data, arp))
         DO i = 1, 3
            br(i) = br(i)-brp(i)
         END DO
         CALL helium_pbc(helium, br)
         s = br(1)**2+br(2)**2+br(3)**2
         z = (ar-arp)**2
         arp = 1.0_dp
         ! j=0 terms
         DO i = 2, SIZE(tab, 1)
            arp = arp*s
            v = v+arp*helium_spline(tab(i, 1)%spline_data, q)
         END DO
         ar = 1.0_dp
         DO j = 2, SIZE(tab, 2)
            ar = ar*z
            arp = ar
            DO i = j, SIZE(tab, 1)
               v = v+arp*helium_spline(tab(i, j)%spline_data, q)
               arp = arp*s
            END DO
         END DO
         res = v
      END IF
      RETURN
   END FUNCTION helium_eval_expansion

! **************************************************************************************************
!> \brief ...
!> \param helium ...
!> \author Harald Forbert
! **************************************************************************************************
   SUBROUTINE helium_update_transition_matrix(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      INTEGER                                            :: b, c, i, j, k, m, n, nb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: lens, order
      INTEGER, DIMENSION(:), POINTER                     :: perm
      INTEGER, DIMENSION(:, :), POINTER                  :: nmatrix
      REAL(KIND=dp)                                      :: f, q, t, v
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: p
      REAL(KIND=dp), DIMENSION(3)                        :: r
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ipmatrix, pmatrix, tmatrix
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos

      nb = helium%atoms
      !TODO: check allocation status
      ALLOCATE (p(2*nb))
      ALLOCATE (order(nb))
      ALLOCATE (lens(2*nb))
      b = helium%beads-helium%bisection+1
      f = -0.5_dp/(helium%hb2m*helium%tau*helium%bisection)
      tmatrix => helium%tmatrix
      pmatrix => helium%pmatrix
      ipmatrix => helium%ipmatrix
      nmatrix => helium%nmatrix
      perm => helium%permutation
      pos => helium%pos
      DO i = 1, nb
         DO j = 1, nb
            v = 0.0_dp
            r(:) = pos(:, i, b)-pos(:, j, 1)
            CALL helium_pbc(helium, r)
            v = v+r(1)*r(1)+r(2)*r(2)+r(3)*r(3)
            pmatrix(i, j) = f*v
         END DO
         t = pmatrix(i, perm(i)) ! just some reference
         v = 0.0_dp
         DO j = 1, nb
            tmatrix(i, j) = EXP(pmatrix(i, j)-t)
            v = v+tmatrix(i, j)
         END DO
         ! normalize
         q = t+LOG(v)
         t = 1.0_dp/v
         DO j = 1, nb
            tmatrix(i, j) = tmatrix(i, j)*t
            ipmatrix(i, j) = 1.0_dp/tmatrix(i, j)
         END DO

         ! at this point we have:
         ! tmatrix(i,j) = exp(-f*(r_i^b - r_j^1)**2) normalized such
         !    that sum_j tmatrix(i,j) = 1.
         !    ( tmatrix(k1,k2) = t_{k1,k2} / h_{k1} of ceperly. )
         !    so tmatrix(i,j) is the probability to try to change a permutation
         !    with particle j (assuming particle i is already selected as well)
         ! ipmatrix(i,j) = 1.0/tmatrix(i,j)
         ! pmatrix(i,j) = log(tmatrix(i,j))  + some_offset(i)

         ! generate optimal search tree so we can select which particle j
         ! belongs to a given random_number as fast as possible.
         ! (the traditional approach would be to generate a table
         !  of cumulative probabilities and to search that table)
         ! so for example if we have:
         ! tmatrix(i,:) = ( 0.1 , 0.4 , 0.2 , 0.3 )
         ! traditionally we would build the running sum table:
         !  ( 0.1 , 0.5 , 0.7 , 1.0 ) and for a random number r
         ! would search this table for the lowest index larger than r
         ! (which would then be the particle index chosen by this random number)
         ! we build an optimal binary search tree instead, so here
         ! we would have:
         ! if ( r > 0.6 ) then take index 2,
         ! else if ( r > 0.3 ) then take index 4,
         ! else if ( r > 0.1 ) then take index 3 else index 1.
         ! the search tree is generated in tmatrix and nmatrix.
         ! tmatrix contains the decision values (0.6,0.3,0.1 in this case)
         ! and nmatrix contains the two branches (what to do if lower or higher)
         ! negative numbers in nmatrix mean take minus that index
         ! positive number means go down the tree to that next node, since we
         ! put the root of the tree at the end the arrays in the example would
         ! look like this:
         ! tmatrix(i,:) = ( 0.1 , 0.3 , 0.6 , arbitrary )
         ! namtrix(i,:) = ( -1 , -3 , 1 , -4 , 2 , -2 , arb. , arb. )
         !
         ! the way to generate this tree may not be the best, but the
         ! tree generation itself shouldn't be needed quite that often:
         !
         ! first sort values (with some variant of heap sort)

         DO j = 1, nb
            order(j) = j
            p(j) = tmatrix(i, j)
         END DO
         IF (nb > 1) THEN ! if nb = 1 it is already sorted.
            k = nb/2+1
            c = nb
            DO
               IF (k > 1) THEN
                  ! building up the heap:
                  k = k-1
                  n = order(k)
                  v = p(k)
               ELSE
                  ! removing the top of the heap
                  n = order(c)
                  v = p(c)
                  order(c) = order(1)
                  p(c) = p(1)
                  c = c-1
                  IF (c == 1) THEN
                     order(1) = n
                     p(1) = v
                     EXIT
                  END IF
               END IF
               m = k
               j = 2*k
               ! restoring heap order between k and c
               DO
                  IF (j > c) EXIT
                  IF (j < c) THEN
                     IF (p(j) < p(j+1)) j = j+1
                  END IF
                  IF (v >= p(j)) EXIT
                  order(m) = order(j)
                  p(m) = p(j)
                  m = j
                  j = 2*j
               END DO
               order(m) = n
               p(m) = v
            END DO
         END IF

         ! now:
         !    p(1:nb)     : tmatrix(i,1:nb) sorted in ascending order
         !    order(1:nb) : corresponding index: p(j) == tmatrix(i,order(j))
         !                                                       for all j

         ! merge sort with elements as we generate new interior search nodes
         ! by combining older elements/nodes

         ! first fill unused part of array with guard values:
         DO j = nb+1, 2*nb
            p(j) = 2.0_dp
         END DO

         ! j   - head of leaf queue
         ! c+1 - head of node queue in p (c in lens)
         ! m+1 - tail of node queue in p (m in lens)
         c = nb+1
         j = 1
         DO m = nb+1, 2*nb-1
            ! get next smallest element
            IF (p(j) < p(c+1)) THEN
               v = p(j)
               lens(j) = m
               j = j+1
            ELSE
               v = p(c+1)
               lens(c) = m
               c = c+1
            END IF
            ! get the second next smallest element
            IF (p(j) < p(c+1)) THEN
               p(m+1) = v+p(j)
               lens(j) = m
               j = j+1
            ELSE
               p(m+1) = v+p(c+1)
               lens(c) = m
               c = c+1
            END IF
         END DO

         ! lens(:) now has the tree with lens(j) pointing to its parent
         ! the root of the tree is at 2*nb-1
         ! calculate the depth of each node in the tree now: (root = 0)

         lens(2*nb-1) = 0
         DO m = 2*nb-2, 1, -1
            lens(m) = lens(lens(m))+1
         END DO

         ! lens(:) now has the depths of the nodes/leafs


         ! reset the nodes, for the canonical tree we just need the leaf info
         DO j = 1, nb
            lens(j+nb) = 0
            p(j+nb) = 0.0_dp
         END DO

         ! build the canonical tree (number of decisions on average are
         ! the same to the tree we build above, but it has better caching behavior

         ! c head of leafs
         ! m head of interior nodes
         c = 1
         m = nb+1
         DO k = 1, 2*nb-2
            j = nb+1+(k-1)/2
            IF (lens(c) > lens(m+1)) THEN
               nmatrix(i, k) = -order(c)
               lens(j+1) = lens(c)-1
               v = p(c)
               c = c+1
            ELSE
               nmatrix(i, k) = m-nb
               lens(j+1) = lens(m+1)-1
               v = p(m)
               m = m+1
            END IF
            p(j) = p(j)+v
            IF (MOD(k, 2) == 1) tmatrix(i, j-nb) = v
         END DO

         ! now:
         !    nmatrix(i,2*j+1) left child of node j
         !    nmatrix(i,2*j+2) right child of node j
         !       children:
         !           negative : leaf with particle index == abs(value)
         !           positive : child node index
         !    p(j) weight of leaf j
         !    p(nb+j) weight of node j
         !    tmatrix(i,j) weight of left child of node j

         ! fix offsets for decision tree:

         p(nb-1) = 0.0_dp
         DO m = nb-1, 1, -1
            ! if right child is a node, set its offset and
            ! change its decision value
            IF (nmatrix(i, 2*m) > 0) THEN
               p(nmatrix(i, 2*m)) = tmatrix(i, m)
               tmatrix(i, nmatrix(i, 2*m)) = tmatrix(i, nmatrix(i, 2*m))+tmatrix(i, m)
            END IF
            ! if left child is a node, set its offset and
            ! change its decision value
            IF (nmatrix(i, 2*m-1) > 0) THEN
               p(nmatrix(i, 2*m-1)) = p(m)
               tmatrix(i, nmatrix(i, 2*m-1)) = tmatrix(i, nmatrix(i, 2*m-1))+p(m)
            END IF
         END DO

         ! canonical optimal search tree done


      END DO

      ! initialize trial permutation with some identity permutation
      ! (should not be taken, but just in case it does we have something valid)

      helium%pweight = 0.0_dp
      t = next_random_number(helium%rng_stream_uniform)
      helium%ptable(1) = 1+INT(t*nb)
      helium%ptable(2) = -1

      ! recalculate inverse permutation table (just in case)
      DO i = 1, nb
         helium%iperm(perm(i)) = i
      END DO

      ! clean up:
      DEALLOCATE (lens)
      DEALLOCATE (order)
      DEALLOCATE (p)

      RETURN
   END SUBROUTINE helium_update_transition_matrix

! **************************************************************************************************
!> \brief ...
!> \param spl ...
!> \param xx ...
!> \retval res ...
!> \author Harald Forbert
! **************************************************************************************************
   FUNCTION helium_spline(spl, xx) RESULT(res)
      TYPE(spline_data_type), POINTER                    :: spl
      REAL(KIND=dp), INTENT(IN)                          :: xx
      REAL(KIND=dp)                                      :: res

      REAL(KIND=dp)                                      :: a, b

      IF (xx < spl%x1) THEN
         b = spl%invh*(xx-spl%x1)
         a = 1.0_dp-b
         res = a*spl%y(1)+b*(spl%y(2)-spl%y2(2)*spl%h26)
      ELSE IF (xx > spl%xn) THEN
         b = spl%invh*(xx-spl%xn)+1.0_dp
         a = 1.0_dp-b
         res = b*spl%y(spl%n)+a*(spl%y(spl%n-1)-spl%y2(spl%n-1)*spl%h26)
      ELSE
         res = spline_value(spl, xx)
      END IF
      RETURN
   END FUNCTION helium_spline

! ***************************************************************************
!> \brief  Return the distance <rij> between bead <ib> of atom <ia>
!>         and bead <jb> of atom <ja>.
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \param ja ...
!> \param jb ...
!> \retval rij ...
!> \date   2009-07-17
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_bead_rij(helium, ia, ib, ja, jb) RESULT(rij)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: ia, ib, ja, jb
      REAL(KIND=dp)                                      :: rij

      REAL(KIND=dp)                                      :: dx, dy, dz

      dx = helium%pos(1, ia, ib)-helium%pos(1, ja, jb)
      dy = helium%pos(2, ia, ib)-helium%pos(2, ja, jb)
      dz = helium%pos(3, ia, ib)-helium%pos(3, ja, jb)
      rij = SQRT(dx*dx+dy*dy+dz*dz)

      RETURN
   END FUNCTION helium_bead_rij

! ***************************************************************************
!> \brief  Given the atom number and permutation state return the cycle
!>         number the atom belongs to.
!> \param helium ...
!> \param atom_number ...
!> \param permutation ...
!> \retval cycle_number ...
!> \date   2009-07-21
!> \author Lukasz Walewski
!> \note   Cycles (or paths) are numbered from 1 to <num_cycles>, where
!>         <num_cycles> is in the range of (1, <helium%atoms>).
!>         if (num_cycles .EQ. 1) then all atoms belong to one cycle
!>         if (num_cycles .EQ. helium%atoms) then there are no cycles of
!>         length greater than 1 (i.e. no atoms are connected)
! **************************************************************************************************
   FUNCTION helium_cycle_number(helium, atom_number, permutation) RESULT(cycle_number)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: atom_number
      INTEGER, DIMENSION(:), POINTER                     :: permutation
      INTEGER                                            :: cycle_number

      INTEGER                                            :: atom_idx, cycle_idx, cycle_num, ia, ib, &
                                                            ic, num_cycles
      INTEGER, DIMENSION(:), POINTER                     :: cycle_index
      LOGICAL                                            :: found, new_cycle

      NULLIFY (cycle_index)
      cycle_index => helium%itmp_atoms_1d
      cycle_index(:) = 0

      num_cycles = 0
      found = .FALSE.
      cycle_num = -1
      DO ia = 1, helium%atoms
         ! this loop reaches its maximum iteration count when atom in question
         ! is the last one (i.e. when atom_number .EQ. helium%atoms)

         ! exit if we have found the cycle number for the atom in question
         IF (found) THEN
            EXIT
         END IF

         ! initialize current cycle index with the current atom
         cycle_idx = ia

         atom_idx = ia
         DO ib = 1, helium%atoms*helium%beads
            ! this loop reaches its maximum iteration count when all He atoms
            ! form one cycle (i.e. all beads belong to one path)

            ! proceed along the path
            atom_idx = permutation(atom_idx)

            IF (atom_idx .EQ. ia) THEN
               ! end of cycle detected (looped back to the first atom)

               ! check if this is a new cycle
               new_cycle = .TRUE.
               DO ic = 1, num_cycles
                  IF (cycle_index(ic) .EQ. cycle_idx) THEN
                     new_cycle = .FALSE.
                  END IF
               END DO

               IF (new_cycle) THEN
                  ! increase number of cycles and update the current cycle's index
                  num_cycles = num_cycles+1
                  cycle_index(num_cycles) = cycle_idx
               END IF

               ! if this was the atom in question
               IF (ia .EQ. atom_number) THEN
                  ! save the cycle index it belongs to
                  cycle_num = cycle_idx

                  ! exit the loop over atoms, we've found what we've been looking for
                  found = .TRUE.
               END IF

               ! exit the loop over beads, there are no more (new) beads in this cycle
               EXIT
            END IF

            ! set the cycle index to the lowest atom index in this cycle
            IF (atom_idx .LT. cycle_idx) THEN
               cycle_idx = atom_idx
            END IF

         END DO

      END DO

!TODO make it cp2k way
      IF (.NOT. found) THEN
         CALL cp__w("motion/helium_common.F",1497,"helium_cycle_number: we are going to return -1, problems ahead!")
      END IF

      ! at this point we know the cycle index for atom <atom_number>
      ! but it is expressed as the atom number of the first atom in that cycle

      ! renumber cycle indices, so that they form a range (1, <num_cycles>)
      ! (don't do it actually - just return the requested <cycle_number>)
      cycle_number = 0
      DO ic = 1, num_cycles
         IF (cycle_index(ic) .EQ. cycle_num) THEN
            cycle_number = ic
            EXIT
         END IF
      END DO

      NULLIFY (cycle_index)

      RETURN
   END FUNCTION helium_cycle_number

! ***************************************************************************
!> \brief  Given the atom number and permutation state return the length of
!>         the path this atom belongs to.
!> \param helium ...
!> \param atom_number ...
!> \param permutation ...
!> \retval path_length ...
!> \date   2009-10-07
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_path_length(helium, atom_number, permutation) RESULT(path_length)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: atom_number
      INTEGER, DIMENSION(:), POINTER                     :: permutation
      INTEGER                                            :: path_length

      INTEGER                                            :: atom_idx, ia
      LOGICAL                                            :: path_end_reached

      atom_idx = atom_number
      path_length = 0
      path_end_reached = .FALSE.
      DO ia = 1, helium%atoms
         path_length = path_length+1
         atom_idx = permutation(atom_idx)
         IF (atom_idx .EQ. atom_number) THEN
            path_end_reached = .TRUE.
            EXIT
         END IF
      END DO

      IF (.NOT. path_end_reached) THEN
         path_length = -1
      END IF

      RETURN
   END FUNCTION helium_path_length

! ***************************************************************************
!> \brief  Given an element of a permutation return the cycle it belongs to.
!> \param element ...
!> \param permutation ...
!> \retval CYCLE ...
!> \date   2013-12-10
!> \author Lukasz Walewski
!> \note   This function allocates memory and returns a pointer to it,
!>         do not forget to deallocate once finished with the results.
! **************************************************************************************************
   FUNCTION helium_cycle_of(element, permutation) RESULT(CYCLE)

      INTEGER, INTENT(IN)                                :: element
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: permutation
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE

      INTEGER                                            :: ia, icur, len, nsize
      CHARACTER(len=*), PARAMETER :: routineN = 'helium_cycle_of', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(:), POINTER                     :: my_cycle
      LOGICAL                                            :: cycle_end_reached

      nsize = SIZE(permutation)

      ! maximum possible cycle length is the number of atoms
      NULLIFY (my_cycle)
      ALLOCATE (my_cycle(nsize))

      ! traverse the permutation table
      len = 0
      icur = element
      cycle_end_reached = .FALSE.
      DO ia = 1, nsize
         len = len+1
         my_cycle(len) = icur
         icur = permutation(icur)
         IF (icur .EQ. element) THEN
            cycle_end_reached = .TRUE.
            EXIT
         END IF
      END DO

      IF (.NOT. cycle_end_reached) THEN
         ! return null
         NULLIFY (CYCLE)
      ELSE
         ! assign the result
         ALLOCATE (CYCLE(len))
         CYCLE(1:len) = my_cycle(1:len)
      END IF

      ! clean up
      DEALLOCATE (my_cycle)
      NULLIFY (my_cycle)

      RETURN
   END FUNCTION helium_cycle_of

! ***************************************************************************
!> \brief  Return total winding number
!> \param helium ...
!> \retval wnum ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_total_winding_number(helium) RESULT(wnum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_winding_number', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: rcur
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ri, rj

      wnum(:) = 0.0_dp
      DO ia = 1, helium%atoms
         ! sum of contributions from the rest of bead pairs
         DO ib = 1, helium%beads-1
            ri => helium%pos(:, ia, ib)
            rj => helium%pos(:, ia, ib+1)
            rcur(:) = ri(:)-rj(:)
            CALL helium_pbc(helium, rcur)
            wnum(:) = wnum(:)+rcur(:)
         END DO
         ! contribution from the last and the first bead
         ri => helium%pos(:, ia, helium%beads)
         rj => helium%pos(:, helium%permutation(ia), 1)
         rcur(:) = ri(:)-rj(:)
         CALL helium_pbc(helium, rcur)
         wnum(:) = wnum(:)+rcur(:)
      END DO

   END FUNCTION helium_total_winding_number

! ***************************************************************************
!> \brief  Return link winding number
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \retval wnum ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_link_winding_number(helium, ia, ib) RESULT(wnum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_link_winding_number', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ja1, ja2, jb1, jb2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r1, r2

      IF (ib .EQ. helium%beads) THEN
         ja1 = ia
         ja2 = helium%permutation(ia)
         jb1 = ib
         jb2 = 1
      ELSE
         ja1 = ia
         ja2 = ia
         jb1 = ib
         jb2 = ib+1
      END IF
      r1 => helium%pos(:, ja1, jb1)
      r2 => helium%pos(:, ja2, jb2)
      wnum(:) = r1(:)-r2(:)
      CALL helium_pbc(helium, wnum)

      RETURN
   END FUNCTION helium_link_winding_number

! ***************************************************************************
!> \brief  Return total winding number (sum over all links)
!> \param helium ...
!> \retval wnum ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_winding_number_linkwise(helium) RESULT(wnum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_winding_number_linkwise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib

      wnum(:) = 0.0_dp
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            wnum(:) = wnum(:)+helium_link_winding_number(helium, ia, ib)
         END DO
      END DO

      RETURN
   END FUNCTION helium_total_winding_number_linkwise

! ***************************************************************************
!> \brief  Return cycle winding number
!> \param helium ...
!> \param CYCLE ...
!> \param pos ...
!> \retval wnum ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_cycle_winding_number(helium, CYCLE, pos) RESULT(wnum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_cycle_winding_number', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ia, ib, nsize
      REAL(KIND=dp), DIMENSION(3)                        :: rcur
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ri, rj

      nsize = SIZE(CYCLE)

      ! traverse the path
      wnum(:) = 0.0_dp
      DO ia = 1, nsize
         ! contributions from all bead pairs of the current atom
         DO ib = 1, helium%beads-1
            ri => pos(:, CYCLE(ia), ib)
            rj => pos(:, CYCLE(ia), ib+1)
            rcur(:) = ri(:)-rj(:)
            CALL helium_pbc(helium, rcur)
            wnum(:) = wnum(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         i1 = CYCLE(ia)
         IF (ia .EQ. nsize) THEN
            i2 = CYCLE(1)
         ELSE
            i2 = CYCLE(ia+1)
         END IF
         ri => pos(:, i1, helium%beads)
         rj => pos(:, i2, 1)
         rcur(:) = ri(:)-rj(:)
         CALL helium_pbc(helium, rcur)
         wnum(:) = wnum(:)+rcur(:)
      END DO

      RETURN
   END FUNCTION helium_cycle_winding_number

! ***************************************************************************
!> \brief  Return total winding number (sum over all cycles)
!> \param helium ...
!> \retval wnum ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_winding_number_cyclewise(helium) RESULT(wnum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_winding_number_cyclewise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic
      REAL(KIND=dp), DIMENSION(3)                        :: wn
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: cycles

! decompose the current permutation state into permutation cycles

      NULLIFY (cycles)
      cycles => helium_calc_cycles(helium%permutation)

      wnum(:) = 0.0_dp
      DO ic = 1, SIZE(cycles)
         wn = helium_cycle_winding_number(helium, cycles(ic)%iap, helium%pos)
         wnum(:) = wnum(:)+wn(:)
      END DO

      DEALLOCATE (cycles)

      RETURN
   END FUNCTION helium_total_winding_number_cyclewise

! ***************************************************************************
!> \brief  Return total projected area
!> \param helium ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_total_projected_area(helium) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_projected_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r12, r2, rcur

      area(:) = 0.0_dp
      DO ia = 1, helium%atoms
         ! contributions from all links of the current atom
         DO ib = 1, helium%beads-1
            r1(:) = helium%pos(:, ia, ib)
            r2(:) = helium%pos(:, ia, ib+1)
            ! comment out for non-PBC version -->
            r12(:) = r2(:)-r1(:)
            CALL helium_pbc(helium, r1)
            CALL helium_pbc(helium, r12)
            r2(:) = r1(:)+r12(:)
            ! comment out for non-PBC version <--
            rcur(1) = r1(2)*r2(3)-r1(3)*r2(2)
            rcur(2) = r1(3)*r2(1)-r1(1)*r2(3)
            rcur(3) = r1(1)*r2(2)-r1(2)*r2(1)
            area(:) = area(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         r1(:) = helium%pos(:, ia, helium%beads)
         r2(:) = helium%pos(:, helium%permutation(ia), 1)
         ! comment out for non-PBC version -->
         r12(:) = r2(:)-r1(:)
         CALL helium_pbc(helium, r1)
         CALL helium_pbc(helium, r12)
         r2(:) = r1(:)+r12(:)
         ! comment out for non-PBC version <--
         rcur(1) = r1(2)*r2(3)-r1(3)*r2(2)
         rcur(2) = r1(3)*r2(1)-r1(1)*r2(3)
         rcur(3) = r1(1)*r2(2)-r1(2)*r2(1)
         area(:) = area(:)+rcur(:)
      END DO
      area(:) = 0.5_dp*area(:)

      RETURN
   END FUNCTION helium_total_projected_area

! ***************************************************************************
!> \brief  Return link projected area
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_link_projected_area(helium, ia, ib) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_link_projected_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ja1, ja2, jb1, jb2
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r12, r2

      IF (ib .EQ. helium%beads) THEN
         ja1 = ia
         ja2 = helium%permutation(ia)
         jb1 = ib
         jb2 = 1
      ELSE
         ja1 = ia
         ja2 = ia
         jb1 = ib
         jb2 = ib+1
      END IF
      r1(:) = helium%pos(:, ja1, jb1)
      r2(:) = helium%pos(:, ja2, jb2)
      ! comment out for non-PBC version -->
      r12(:) = r2(:)-r1(:)
      CALL helium_pbc(helium, r1)
      CALL helium_pbc(helium, r12)
      r2(:) = r1(:)+r12(:)
      ! comment out for non-PBC version <--
      area(1) = r1(2)*r2(3)-r1(3)*r2(2)
      area(2) = r1(3)*r2(1)-r1(1)*r2(3)
      area(3) = r1(1)*r2(2)-r1(2)*r2(1)
      area(:) = 0.5_dp*area(:)

      RETURN
   END FUNCTION helium_link_projected_area

! ***************************************************************************
!> \brief  Return total projected area (sum over all links)
!> \param helium ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_projected_area_linkwise(helium) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_projected_area_linkwise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib

      area(:) = 0.0_dp
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            area(:) = area(:)+helium_link_projected_area(helium, ia, ib)
         END DO
      END DO

   END FUNCTION helium_total_projected_area_linkwise

! ***************************************************************************
!> \brief  Return cycle projected area
!> \param helium ...
!> \param CYCLE ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_cycle_projected_area(helium, CYCLE) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_cycle_projected_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ia, ib, nsize
      REAL(KIND=dp), DIMENSION(3)                        :: rcur, rsum
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ri, rj

      nsize = SIZE(CYCLE)

      ! traverse the path
      rsum(:) = 0.0_dp
      DO ia = 1, nsize
         ! contributions from all bead pairs of the current atom
         DO ib = 1, helium%beads-1
            ri => helium%pos(:, CYCLE(ia), ib)
            rj => helium%pos(:, CYCLE(ia), ib+1)
            rcur(1) = ri(2)*rj(3)-ri(3)*rj(2)
            rcur(2) = ri(3)*rj(1)-ri(1)*rj(3)
            rcur(3) = ri(1)*rj(2)-ri(2)*rj(1)
            rsum(:) = rsum(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         i1 = CYCLE(ia)
         IF (ia .EQ. nsize) THEN
            i2 = CYCLE(1)
         ELSE
            i2 = CYCLE(ia+1)
         END IF
         ri => helium%pos(:, i1, helium%beads)
         rj => helium%pos(:, i2, 1)
         rcur(1) = ri(2)*rj(3)-ri(3)*rj(2)
         rcur(2) = ri(3)*rj(1)-ri(1)*rj(3)
         rcur(3) = ri(1)*rj(2)-ri(2)*rj(1)
         rsum(:) = rsum(:)+rcur(:)
      END DO
      area(:) = 0.5_dp*rsum(:)

      RETURN
   END FUNCTION helium_cycle_projected_area

! ***************************************************************************
!> \brief  Return cycle projected area (sum over all links)
!> \param helium ...
!> \param CYCLE ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_cycle_projected_area_pbc(helium, CYCLE) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_cycle_projected_area_pbc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ia, ib, nsize
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r12, r2, rcur

      nsize = SIZE(CYCLE)

      ! traverse the path
      area(:) = 0.0_dp
      DO ia = 1, nsize
         ! contributions from all bead pairs of the current atom
         DO ib = 1, helium%beads-1
            r1(:) = helium%pos(:, CYCLE(ia), ib)
            r2(:) = helium%pos(:, CYCLE(ia), ib+1)
            r12(:) = r2(:)-r1(:)
            CALL helium_pbc(helium, r1)
            CALL helium_pbc(helium, r12)
            r2(:) = r1(:)+r12(:)
            rcur(1) = r1(2)*r2(3)-r1(3)*r2(2)
            rcur(2) = r1(3)*r2(1)-r1(1)*r2(3)
            rcur(3) = r1(1)*r2(2)-r1(2)*r2(1)
            area(:) = area(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         i1 = CYCLE(ia)
         IF (ia .EQ. nsize) THEN
            i2 = CYCLE(1)
         ELSE
            i2 = CYCLE(ia+1)
         END IF
         r1(:) = helium%pos(:, i1, helium%beads)
         r2(:) = helium%pos(:, i2, 1)
         r12(:) = r2(:)-r1(:)
         CALL helium_pbc(helium, r1)
         CALL helium_pbc(helium, r12)
         r2(:) = r1(:)+r12(:)
         rcur(1) = r1(2)*r2(3)-r1(3)*r2(2)
         rcur(2) = r1(3)*r2(1)-r1(1)*r2(3)
         rcur(3) = r1(1)*r2(2)-r1(2)*r2(1)
         area(:) = area(:)+rcur(:)
      END DO
      area(:) = 0.5_dp*area(:)

      RETURN
   END FUNCTION helium_cycle_projected_area_pbc

! ***************************************************************************
!> \brief  Return total projected area (sum over all cycles)
!> \param helium ...
!> \retval area ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_projected_area_cyclewise(helium) RESULT(area)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_projected_area_cyclewise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic
      REAL(KIND=dp), DIMENSION(3)                        :: pa
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: cycles

! decompose the current permutation state into permutation cycles

      NULLIFY (cycles)
      cycles => helium_calc_cycles(helium%permutation)

      area(:) = 0.0_dp
      DO ic = 1, SIZE(cycles)
         pa = helium_cycle_projected_area(helium, cycles(ic)%iap)
         area(:) = area(:)+pa(:)
      END DO

      RETURN
   END FUNCTION helium_total_projected_area_cyclewise

! ***************************************************************************
!> \brief  Return total moment of inertia divided by m_He
!> \param helium ...
!> \retval moit ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_total_moment_of_inertia(helium) RESULT(moit)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: moit

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_moment_of_inertia', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: com, r1, r12, r2, rcur

      com(:) = helium%center(:)

      moit(:) = 0.0_dp
      DO ia = 1, helium%atoms
         ! contributions from all the links of the current atom
         DO ib = 1, helium%beads-1
            r1(:) = helium%pos(:, ia, ib)-com(:)
            r2(:) = helium%pos(:, ia, ib+1)-com(:)
            ! comment out for non-PBC version -->
            r12(:) = r2(:)-r1(:)
            CALL helium_pbc(helium, r1)
            CALL helium_pbc(helium, r12)
            r2(:) = r1(:)+r12(:)
            ! comment out for non-PBC version <--
            rcur(1) = r1(2)*r2(2)+r1(3)*r2(3)
            rcur(2) = r1(3)*r2(3)+r1(1)*r2(1)
            rcur(3) = r1(1)*r2(1)+r1(2)*r2(2)
            moit(:) = moit(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         r1(:) = helium%pos(:, ia, helium%beads)-com(:)
         r2(:) = helium%pos(:, helium%permutation(ia), 1)-com(:)
         ! comment out for non-PBC version -->
         r12(:) = r2(:)-r1(:)
         CALL helium_pbc(helium, r1)
         CALL helium_pbc(helium, r12)
         r2(:) = r1(:)+r12(:)
         ! comment out for non-PBC version <--
         rcur(1) = r1(2)*r2(2)+r1(3)*r2(3)
         rcur(2) = r1(3)*r2(3)+r1(1)*r2(1)
         rcur(3) = r1(1)*r2(1)+r1(2)*r2(2)
         moit(:) = moit(:)+rcur(:)
      END DO
      moit(:) = moit(:)/helium%beads

      RETURN
   END FUNCTION helium_total_moment_of_inertia

! ***************************************************************************
!> \brief  Return link moment of inertia divided by m_He
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \retval moit ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_link_moment_of_inertia(helium, ia, ib) RESULT(moit)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: moit

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_link_moment_of_inertia', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ja1, ja2, jb1, jb2
      REAL(KIND=dp), DIMENSION(3)                        :: com, r1, r12, r2

      com(:) = helium%center(:)

      IF (ib .EQ. helium%beads) THEN
         ja1 = ia
         ja2 = helium%permutation(ia)
         jb1 = ib
         jb2 = 1
      ELSE
         ja1 = ia
         ja2 = ia
         jb1 = ib
         jb2 = ib+1
      END IF
      r1(:) = helium%pos(:, ja1, jb1)-com(:)
      r2(:) = helium%pos(:, ja2, jb2)-com(:)
      ! comment out for non-PBC version -->
      r12(:) = r2(:)-r1(:)
      CALL helium_pbc(helium, r1)
      CALL helium_pbc(helium, r12)
      r2(:) = r1(:)+r12(:)
      ! comment out for non-PBC version <--
      moit(1) = r1(2)*r2(2)+r1(3)*r2(3)
      moit(2) = r1(3)*r2(3)+r1(1)*r2(1)
      moit(3) = r1(1)*r2(1)+r1(2)*r2(2)
      moit(:) = moit(:)/helium%beads

      RETURN
   END FUNCTION helium_link_moment_of_inertia

! ***************************************************************************
!> \brief  Return total moment of inertia (sum over all links)
!> \param helium ...
!> \retval moit ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_moment_of_inertia_linkwise(helium) RESULT(moit)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: moit

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_moment_of_inertia_linkwise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib

      moit(:) = 0.0_dp
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            moit(:) = moit(:)+helium_link_moment_of_inertia(helium, ia, ib)
         END DO
      END DO

   END FUNCTION helium_total_moment_of_inertia_linkwise

! ***************************************************************************
!> \brief  Return moment of inertia of a cycle divided by m_He
!> \param helium ...
!> \param CYCLE ...
!> \param pos ...
!> \retval moit ...
!> \date   2014-04-24
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_cycle_moment_of_inertia(helium, CYCLE, pos) RESULT(moit)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos
      REAL(KIND=dp), DIMENSION(3)                        :: moit

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_cycle_moment_of_inertia', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ia, ib, nsize
      REAL(KIND=dp), DIMENSION(3)                        :: com, rcur, ri, rj

      nsize = SIZE(CYCLE)

      ! traverse the path
      moit(:) = 0.0_dp
      com(:) = helium_com(helium)
      DO ia = 1, nsize
         ! contributions from all bead pairs of the current atom
         DO ib = 1, helium%beads-1
            ri = pos(:, CYCLE(ia), ib)-com(:)
            rj = pos(:, CYCLE(ia), ib+1)-com(:)
            rcur(1) = ri(2)*rj(2)+ri(3)*rj(3)
            rcur(2) = ri(3)*rj(3)+ri(1)*rj(1)
            rcur(3) = ri(1)*rj(1)+ri(2)*rj(2)
            moit(:) = moit(:)+rcur(:)
         END DO
         ! contribution from the last bead of the current atom
         ! and the first bead of the next atom
         i1 = CYCLE(ia)
         IF (ia .EQ. nsize) THEN
            i2 = CYCLE(1)
         ELSE
            i2 = CYCLE(ia+1)
         END IF
         ! rotation invariant bead index
         ri = pos(:, i1, helium%beads)-com(:)
         rj = pos(:, i2, 1)-com(:)
         rcur(1) = ri(2)*rj(2)+ri(3)*rj(3)
         rcur(2) = ri(3)*rj(3)+ri(1)*rj(1)
         rcur(3) = ri(1)*rj(1)+ri(2)*rj(2)
         moit(:) = moit(:)+rcur(:)
      END DO
      moit(:) = moit(:)/helium%beads

      RETURN
   END FUNCTION helium_cycle_moment_of_inertia

! ***************************************************************************
!> \brief  Return total moment of inertia (sum over all cycles)
!> \param helium ...
!> \retval moit ...
!> \date   2014-04-24
!> \author Lukasz Walewski
!> \note   mostly for sanity checks
! **************************************************************************************************
   FUNCTION helium_total_moment_of_inertia_cyclewise(helium) RESULT(moit)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: moit

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_total_moment_of_inertia_cyclewise', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic
      REAL(KIND=dp), DIMENSION(3)                        :: pa
      TYPE(int_arr_ptr), DIMENSION(:), POINTER           :: cycles

! decompose the current permutation state into permutation cycles

      NULLIFY (cycles)
      cycles => helium_calc_cycles(helium%permutation)

      moit(:) = 0.0_dp
      DO ic = 1, SIZE(cycles)
         pa = helium_cycle_moment_of_inertia(helium, cycles(ic)%iap, helium%pos)
         moit(:) = moit(:)+pa(:)
      END DO

      DEALLOCATE (cycles)

      RETURN
   END FUNCTION helium_total_moment_of_inertia_cyclewise

! ***************************************************************************
!> \brief  Set coordinate system, e.g. for RHO calculations
!> \param helium ...
!> \param pint_env ...
!> \date   2014-04-25
!> \author Lukasz Walewski
!> \note   Sets the origin (center of the coordinate system) wrt which
!>         spatial distribution functions are calculated.
!> \note   It can be extended later to set the axes of the coordinate system
!>         as well, e.g. for dynamic analysis with moving solute.
! **************************************************************************************************
   SUBROUTINE helium_update_coord_system(helium, pint_env)

      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_update_coord_system', &
         routineP = moduleN//':'//routineN

      IF (helium%solute_present) THEN
         helium%center(:) = pint_com_pos(pint_env)
      ELSE
         IF (helium%periodic) THEN
            helium%center(:) = (/0.0_dp, 0.0_dp, 0.0_dp/)
         ELSE
            helium%center(:) = helium_com(helium)
         END IF
      END IF

      RETURN
   END SUBROUTINE helium_update_coord_system

! ***************************************************************************
!> \brief  Set coordinate system for RDF calculations
!> \param helium ...
!> \param pint_env ...
!> \date   2014-04-25
!> \author Lukasz Walewski
!> \note   Sets the centers wrt which radial distribution functions are
!>         calculated.
! **************************************************************************************************
   SUBROUTINE helium_set_rdf_coord_system(helium, pint_env)

      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_set_rdf_coord_system', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (helium%solute_present) THEN
         CALL pint_calc_centroid(pint_env)
         i = 3*helium%solute_atoms
         helium%rdf_centers(1:i) = pint_env%centroid(:)
      ELSE
         helium%rdf_centers(:) = helium%center(:)
      END IF

      RETURN
   END SUBROUTINE helium_set_rdf_coord_system

! ***************************************************************************
!> \brief  Calculate center of mass
!> \param helium ...
!> \retval com ...
!> \date   2014-04-09
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_com(helium) RESULT(com)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: com

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_com', routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib

      com(:) = 0.0_dp
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            com(:) = com(:)+helium%pos(:, ia, ib)
         END DO
      END DO
      com(:) = com(:)/helium%atoms/helium%beads

   END FUNCTION helium_com

! ***************************************************************************
!> \brief  Return link vector, i.e. displacement vector of two consecutive
!>         beads along the path starting at bead ib of atom ia
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \retval lvec ...
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_link_vector(helium, ia, ib) RESULT(lvec)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: lvec

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_link_vector', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia1, ia2, ib1, ib2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r1, r2

      IF (ib .EQ. helium%beads) THEN
         ia1 = ia
         ia2 = helium%permutation(ia)
         ib1 = ib
         ib2 = 1
      ELSE
         ia1 = ia
         ia2 = ia
         ib1 = ib
         ib2 = ib+1
      END IF
      r1 => helium%pos(:, ia1, ib1)
      r2 => helium%pos(:, ia2, ib2)
      lvec(:) = r2(:)-r1(:)
      CALL helium_pbc(helium, lvec)

   END FUNCTION helium_link_vector

! **************************************************************************************************
!> \brief ...
!> \param helium ...
!> \param ia ...
!> \param ib ...
!> \retval rperp ...
! **************************************************************************************************
   FUNCTION helium_rperpendicular(helium, ia, ib) RESULT(rperp)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER                                            :: ia, ib
      REAL(KIND=dp), DIMENSION(3)                        :: rperp

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rperpendicular', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(:), POINTER               :: ri

      ri => helium%pos(:, ia, ib)
      rperp(1) = SQRT(ri(2)*ri(2)+ri(3)*ri(3))
      rperp(2) = SQRT(ri(3)*ri(3)+ri(1)*ri(1))
      rperp(3) = SQRT(ri(1)*ri(1)+ri(2)*ri(2))

      RETURN
   END FUNCTION helium_rperpendicular

! ***************************************************************************
!> \brief  Convert a winding number vector from real number representation
!>         (in internal units) to integer number representation (in cell
!>         vector units)
!> \param helium ...
!> \param wnum ...
!> \retval inum ...
!> \author Lukasz Walewski
! **************************************************************************************************
   FUNCTION helium_wnumber_to_integer(helium, wnum) RESULT(inum)

      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3)                        :: wnum
      INTEGER, DIMENSION(3)                              :: inum

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_wnumber_to_integer', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3)                        :: wcur
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: invcell

      invcell => helium%cell_m_inv
      CALL DGEMV('N', 3, 3, 1.0_dp, invcell, SIZE(invcell, 1), wnum, 1, 0.0_dp, wcur, 1)
      inum(:) = NINT(wcur(:))

      RETURN
   END FUNCTION helium_wnumber_to_integer

! ***************************************************************************
!> \brief  Given the atom index and permutation state returns .TRUE. if the
!>         atom belongs to a winding path, .FASLE. otherwise.
!> \param helium ...
!> \param atmidx ...
!> \param pos ...
!> \param permutation ...
!> \retval is_winding ...
!> \date   2010-09-21
!> \author Lukasz Walewski
!> \note   The path winds around the periodic box if any component of its
!>         widing number vector differs from 0.
! **************************************************************************************************
   FUNCTION helium_is_winding(helium, atmidx, pos, permutation) RESULT(is_winding)

      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: atmidx
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos
      INTEGER, DIMENSION(:), POINTER                     :: permutation
      LOGICAL                                            :: is_winding

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_is_winding', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic
      INTEGER, DIMENSION(3)                              :: inum
      INTEGER, DIMENSION(:), POINTER                     :: CYCLE
      REAL(KIND=dp), DIMENSION(3)                        :: wnum

      is_winding = .FALSE.
      NULLIFY (CYCLE)
      CYCLE => helium_cycle_of(atmidx, permutation)
      wnum(:) = bohr*helium_cycle_winding_number(helium, CYCLE, pos)
      inum(:) = helium_wnumber_to_integer(helium, wnum)
      DO ic = 1, 3
         IF (ABS(inum(ic)) .GT. 0) THEN
            is_winding = .TRUE.
            EXIT
         END IF
      END DO
      DEALLOCATE (CYCLE)

      RETURN
   END FUNCTION helium_is_winding

END MODULE helium_common
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  I/O subroutines for helium
!> \author Lukasz Walewski
!> \date   2009-06-08
! **************************************************************************************************
MODULE helium_io

   USE cell_types,                      ONLY: get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE helium_common,                   ONLY: helium_cycle_number,&
                                              helium_cycle_of,&
                                              helium_is_winding,&
                                              helium_path_length,&
                                              helium_pbc
   USE helium_types,                    ONLY: e_id_interact,&
                                              e_id_kinetic,&
                                              e_id_potential,&
                                              e_id_thermo,&
                                              e_id_total,&
                                              e_id_virial,&
                                              helium_solvent_type,&
                                              rho_num
   USE input_constants,                 ONLY: helium_cell_shape_cube,&
                                              helium_cell_shape_octahedron,&
                                              helium_sampling_ceperley,&
                                              helium_solute_intpot_mwater,&
                                              helium_solute_intpot_none,&
                                              perm_cycle,&
                                              perm_plain
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              int_8
   USE machine,                         ONLY: m_flush
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_gather,&
                                              mp_sum
   USE physcon,                         ONLY: angstrom
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_io'

   CHARACTER(len=*), DIMENSION(6), PARAMETER, PRIVATE :: m_dist_name = (/ &
                                                         "SINGLEV    ", &
                                                         "UNIFORM    ", &
                                                         "LINEAR     ", &
                                                         "QUADRATIC  ", &
                                                         "EXPONENTIAL", &
                                                         "GAUSSIAN   "/)

   PUBLIC :: helium_read_xyz
   PUBLIC :: helium_print_rdf
   PUBLIC :: helium_print_rho
   PUBLIC :: helium_write_line
   PUBLIC :: helium_write_setup
   PUBLIC :: helium_print_energy
   PUBLIC :: helium_print_vector
   PUBLIC :: helium_print_plength
   PUBLIC :: helium_print_coordinates
   PUBLIC :: helium_print_force
   PUBLIC :: helium_print_accepts
   PUBLIC :: helium_print_perm
   PUBLIC :: helium_write_cubefile

CONTAINS

! ***************************************************************************
!> \brief  Read XYZ coordinates from file
!> \param coords ...
!> \param file_name ...
!> \param para_env ...
!> \date   2009-06-03
!> \author Lukasz Walewski
!> \note   This is a parallel routine, all ranks get coords defined
! **************************************************************************************************
   SUBROUTINE helium_read_xyz(coords, file_name, para_env)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: coords
      CHARACTER(LEN=default_path_length)                 :: file_name
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_read_xyz', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: strtmp
      INTEGER                                            :: frame, handle, istat, j, natom
      LOGICAL                                            :: found, my_end
      TYPE(cp_parser_type), POINTER                      :: parser

      CALL timeset(routineN, handle)

      ! check if the file can be accessed
      INQUIRE (FILE=file_name, EXIST=found, IOSTAT=istat)
      IF (istat /= 0) THEN
         WRITE (UNIT=strtmp, FMT="(A,I0,A)") &
            "An error occurred inquiring the file <"// &
            TRIM(file_name)//">"
         CALL cp__b("motion/helium_io.F",125,TRIM(strtmp))
      END IF
      IF (.NOT. found) THEN
         CALL cp__b("motion/helium_io.F",128,"Coordinate file <"//TRIM(file_name)//"> not found.")
      END IF

      NULLIFY (parser)
      CALL parser_create( &
         parser, &
         file_name, &
         para_env=para_env, &
         parse_white_lines=.TRUE.)

      natom = 0
      frame = 0
      CALL parser_get_next_line(parser, 1)
      Frames: DO
         ! Atom number
         CALL parser_get_object(parser, natom)
         frame = frame+1
         IF (frame == 1) THEN
            ALLOCATE (coords(3*natom))
         ELSE
            strtmp = "Warning: more than one frame on file <"//TRIM(file_name)//">"
            CALL helium_write_line(strtmp)
            strtmp = "Warning: using the first frame only!"
            CALL helium_write_line(strtmp)
            EXIT Frames
         END IF
         ! Dummy line
         CALL parser_get_next_line(parser, 2)
         DO j = 1, natom
            ! Atom coordinates
            READ (parser%input_line, *) strtmp, &
               coords(3*(j-1)+1), &
               coords(3*(j-1)+2), &
               coords(3*(j-1)+3)
            coords(3*(j-1)+1) = cp_unit_to_cp2k(coords(3*(j-1)+1), "angstrom")
            coords(3*(j-1)+2) = cp_unit_to_cp2k(coords(3*(j-1)+2), "angstrom")
            coords(3*(j-1)+3) = cp_unit_to_cp2k(coords(3*(j-1)+3), "angstrom")
            ! If there's a white line or end of file exit.. otherwise go on
            CALL parser_get_next_line(parser, 1, at_end=my_end)
            my_end = my_end .OR. (LEN_TRIM(parser%input_line) == 0)
            IF (my_end) THEN
               IF (j /= natom) THEN
                  CALL cp__b("motion/helium_io.F",170,"Error in XYZ format.")
               END IF
               EXIT Frames
            END IF
         END DO
      END DO Frames
      CALL parser_release(parser)

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_read_xyz

! ***************************************************************************
!> \brief  Write helium parameters to the output unit
!> \param helium ...
!> \date   2009-06-03
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_write_setup(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_write_setup', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msg_str, my_label, stmp, stmp1, stmp2, &
                                                            unit_str
      INTEGER                                            :: i, itmp, j, unit_nr
      INTEGER(KIND=int_8)                                :: i8tmp
      LOGICAL                                            :: first
      REAL(KIND=dp)                                      :: rtmp, v1, v2, v3
      REAL(KIND=dp), DIMENSION(3)                        :: my_abc
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      my_label = "HELIUM| "

      IF (logger%para_env%ionode) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger)

         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Number of helium environments:     ", helium%num_env

         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Number of solvent atoms:           ", helium%atoms
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Number of solvent beads:           ", helium%beads
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Total number of solvent particles: ", helium%atoms*helium%beads

         unit_str = "angstrom^-3"
         rtmp = cp_unit_from_cp2k(helium%density, &
                                  unit_str)
         WRITE (unit_nr, '(T2,A,F12.6)') TRIM(my_label)//" Density   ["// &
            TRIM(unit_str)//"]:", rtmp

         unit_str = "angstrom"
         rtmp = cp_unit_from_cp2k(helium%cell_size, &
                                  unit_str)
         WRITE (unit_nr, '(T2,A,F12.6)') TRIM(my_label)//" Cell size ["// &
            TRIM(unit_str)//"]:   ", rtmp

         IF (helium%periodic) THEN
            SELECT CASE (helium%cell_shape)
            CASE (helium_cell_shape_cube)
               CALL helium_write_line("PBC cell shape: CUBE.")
            CASE (helium_cell_shape_octahedron)
               CALL helium_write_line("PBC cell shape: TRUNCATED OCTAHEDRON.")
            CASE DEFAULT
               CALL helium_write_line("*** Warning: unknown cell shape.")
            END SELECT
         ELSE
            CALL helium_write_line("PBC turned off.")
         END IF

         IF (helium%droplet_radius .LT. HUGE(1.0_dp)) THEN
            WRITE (stmp, *) helium%droplet_radius*angstrom
            CALL helium_write_line("Droplet radius: "//TRIM(ADJUSTL(stmp))//" [angstrom]")
         END IF

         ! first step gets incremented during first iteration
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " First MC step                      :", helium%first_step+1
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Last MC step                       :", helium%last_step
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Total number of MC steps           :", helium%num_steps
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Number of outer MC trials per step :", helium%iter_rot
         WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
            " Number of inner MC trials per step :", helium%iter_norot
         i8tmp = INT(helium%iter_rot, int_8)
         i8tmp = i8tmp*INT(helium%iter_norot, int_8)
         stmp = ""
         WRITE (stmp, *) i8tmp
         WRITE (unit_nr, '(T2,A)') TRIM(my_label)// &
            " Total number of MC trials per step : "//TRIM(ADJUSTL(stmp))
         i8tmp = INT(helium%num_steps, int_8)
         i8tmp = i8tmp*INT(helium%iter_rot, int_8)
         i8tmp = i8tmp*INT(helium%iter_norot, int_8)
         stmp = ""
         WRITE (stmp, *) i8tmp
         WRITE (unit_nr, '(T2,A)') TRIM(my_label)// &
            " Total number of MC trials          : "//TRIM(ADJUSTL(stmp))

         SELECT CASE (helium%sampling_method)

         CASE (helium_sampling_ceperley)

            ! permutation cycle length sampling
            stmp = ""
            CALL helium_write_line(stmp)
            WRITE (stmp, *) helium%maxcycle
            stmp2 = ""
            WRITE (stmp2, *) "Using maximum permutation cycle length: "// &
               TRIM(ADJUSTL(stmp))
            CALL helium_write_line(stmp2)
            stmp = ""
            WRITE (stmp, *) "Permutation cycle length distribution: "// &
               TRIM(ADJUSTL(m_dist_name(helium%m_dist_type)))
            CALL helium_write_line(stmp)
            stmp = ""
            stmp1 = ""
            WRITE (stmp1, *) helium%m_ratio
            stmp2 = ""
            WRITE (stmp2, *) helium%m_value
            WRITE (stmp, *) "Using ratio "//TRIM(ADJUSTL(stmp1))// &
               " for M = "//TRIM(ADJUSTL(stmp2))
            CALL helium_write_line(stmp)
            stmp = ""
            CALL helium_write_line(stmp)

         CASE DEFAULT
            WRITE (msg_str, *) helium%sampling_method
            msg_str = "Sampling method ("//TRIM(ADJUSTL(msg_str))//") not supported."
            CALL cp__b("motion/helium_io.F",308,msg_str)

         END SELECT

         ! solute related data
         IF (helium%solute_present) THEN
            WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
               " Number of solute atoms:            ", helium%solute_atoms
            WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
               " Number of solute beads:            ", helium%solute_beads
            WRITE (unit_nr, '(T2,A,1X,I0)') TRIM(my_label)// &
               " Total number of solute particles:  ", helium%solute_atoms* &
               helium%solute_beads
            DO i = 1, helium%enum
               WRITE (unit_nr, '(T2,A,1X,A2,A,I0,A,I0,A)', ADVANCE='NO') &
                  TRIM(my_label)//" Solute atom type: ", &
                  helium%ename(i), &
                  "(", &
                  helium%eid(i), &
                  "), count: ", &
                  SIZE(helium%solute_i(helium%eid(i))%iap), &
                  ", "
               first = .TRUE.
               DO j = 1, SIZE(helium%solute_i(helium%eid(i))%iap)
                  IF (first) THEN
                     WRITE (unit_nr, '(A)', ADVANCE='NO') "indices: "
                     first = .FALSE.
                  ELSE
                     WRITE (unit_nr, '(A)', ADVANCE='NO') ", "
                  END IF
                  WRITE (unit_nr, '(I0)', ADVANCE='NO') helium%solute_i(helium%eid(i))%iap(j)
               END DO
               WRITE (unit_nr, '(1X)')
            END DO

            stmp1 = ""
            SELECT CASE (helium%solute_interaction)
            CASE (helium_solute_intpot_none)
               WRITE (stmp1, *) "NONE"
            CASE (helium_solute_intpot_mwater)
               WRITE (stmp1, *) "MWATER"
            CASE DEFAULT
               WRITE (stmp1, *) "***UNKNOWN***"
            END SELECT
            WRITE (unit_nr, '(T2,A,A,A)') &
               TRIM(my_label), &
               " Solute interaction type: ", &
               TRIM(ADJUSTL(stmp1))

            CALL get_cell(helium%solute_cell, abc=my_abc)
            unit_str = "angstrom"
            v1 = cp_unit_from_cp2k(my_abc(1), unit_str)
            v2 = cp_unit_from_cp2k(my_abc(2), unit_str)
            v3 = cp_unit_from_cp2k(my_abc(3), unit_str)
            WRITE (unit_nr, '(T2,A,F12.6,1X,F12.6,1X,F12.6)') &
               TRIM(my_label)//" Solute cell size ["// &
               TRIM(unit_str)//"]:   ", v1, v2, v3
         ELSE
            WRITE (unit_nr, '(T2,A)') TRIM(my_label)//" Solute is NOT present"
         END IF
      END IF
      CALL helium_write_line("")

      ! RDF related settings
      IF (helium%rdf_present) THEN
         WRITE (stmp, '(I6)') helium%rdf_num_ctr
         CALL helium_write_line("RDF| number of centers: "//TRIM(stmp))
         rtmp = cp_unit_from_cp2k(helium%rdf_delr, "angstrom")
         WRITE (stmp, '(1X,F12.6)') rtmp
         CALL helium_write_line("RDF| delr [angstrom]  : "//TRIM(stmp))
         rtmp = cp_unit_from_cp2k(helium%rdf_maxr, "angstrom")
         WRITE (stmp, '(1X,F12.6)') rtmp
         CALL helium_write_line("RDF| maxr [angstrom]  : "//TRIM(stmp))
         itmp = helium%rdf_nbin
         WRITE (stmp, '(I6)') itmp
         CALL helium_write_line("RDF| nbin             : "//TRIM(stmp))
         CALL helium_write_line("")
      ELSE
         CALL helium_write_line("RDF| radial distributions will NOT be calculated.")
         CALL helium_write_line("")
      END IF

      ! RHO related settings
      IF (helium%rho_present) THEN
         CALL helium_write_line('RHO| The following densities will be calculated:')
         DO i = 1, rho_num
            IF (helium%rho_property(i)%is_calculated) THEN
               WRITE (stmp, '(A)') 'RHO|    '//TRIM(helium%rho_property(i)%name)
               CALL helium_write_line(TRIM(stmp))
            END IF
         END DO
         CALL helium_write_line('RHO|')
         rtmp = cp_unit_from_cp2k(helium%rho_delr, "angstrom")
         WRITE (stmp, '(1X,F12.6)') rtmp
         CALL helium_write_line("RHO| delr [angstrom]  : "//TRIM(stmp))
         rtmp = cp_unit_from_cp2k(helium%rho_maxr, "angstrom")
         WRITE (stmp, '(1X,F12.6)') rtmp
         CALL helium_write_line("RHO| maxr [angstrom]  : "//TRIM(stmp))
         itmp = helium%rho_nbin
         WRITE (stmp, '(I6)') itmp
         CALL helium_write_line("RHO| nbin             : "//TRIM(stmp))
         CALL helium_write_line("")
      ELSE
         CALL helium_write_line("RHO| Density distributions will NOT be calculated.")
         CALL helium_write_line("")
      END IF

      RETURN
   END SUBROUTINE helium_write_setup

! ***************************************************************************
!> \brief  Writes out a line of text to the default output unit
!> \param line ...
!> \date   2009-07-10
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_write_line(line)

      CHARACTER(len=*), INTENT(IN)                       :: line

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_write_line', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_label
      INTEGER                                            :: unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      my_label = "HELIUM|"

      IF (logger%para_env%ionode) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger)
         WRITE (unit_nr, '(T2,A)') TRIM(my_label)//" "//TRIM(line)
      END IF

      RETURN
   END SUBROUTINE helium_write_line

! ***************************************************************************
!> \brief  Print energies according to HELIUM%PRINT%ENERGY
!> \param helium ...
!> \date   2009-06-08
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_energy(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iteration, m, unit_nr
      LOGICAL                                            :: file_is_new, should_output
      REAL(KIND=dp)                                      :: naccptd
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_energy
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      my_energy => helium%energy_avrg

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%ENERGY")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (should_output) THEN

         iteration = logger%iter_info%iteration(2)
         naccptd = 0.0_dp
         DO m = 1, helium%maxcycle
            naccptd = naccptd+helium%num_accepted(helium%bisctlog2+2, m)
         END DO
         CALL mp_sum(naccptd, logger%para_env%group)
         naccptd = naccptd/REAL(logger%para_env%num_pe, dp)

         IF (logger%para_env%ionode) THEN

            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name="helium-energy", &
                      extension=".dat", &
                      is_new_file=file_is_new)

            IF (file_is_new) THEN
               WRITE (unit_nr, '(A9,1X,A12,6(1X,A20))') &
                  "#    Step", &
                  "     Naccptd", &
                  "               E_pot", &
                  "               E_kin", &
                  "            E_thermo", &
                  "            E_virial", &
                  "             E_inter", &
                  "               E_tot"
            END IF

            WRITE (unit_nr, "(I9,1X,F12.1,6(1X,F20.9))") &
               iteration, &
               naccptd, &
               my_energy(e_id_potential), &
               my_energy(e_id_kinetic), &
               my_energy(e_id_thermo), &
               my_energy(e_id_virial), &
               my_energy(e_id_interact), &
               my_energy(e_id_total)

            CALL m_flush(unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, print_key)

         END IF
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_energy

! ***************************************************************************
!> \brief  Print a 3D real vector accoring to printkey <pkey>
!> \param helium ...
!> \param pkey ...
!> \param DATA ...
!> \param uconv ...
!> \param col_label ...
!> \param cmmnt ...
!> \param fname ...
!> \param fpos ...
!> \param avg ...
!> \date   2014-09-09
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_vector(helium, pkey, DATA, uconv, col_label, cmmnt, fname, fpos, avg)

      TYPE(helium_solvent_type), POINTER                 :: helium
      CHARACTER(len=*)                                   :: pkey
      REAL(KIND=dp), DIMENSION(3)                        :: DATA
      REAL(KIND=dp)                                      :: uconv
      CHARACTER(len=*), DIMENSION(3)                     :: col_label
      CHARACTER(len=*)                                   :: cmmnt, fname
      CHARACTER(len=*), OPTIONAL                         :: fpos
      LOGICAL, OPTIONAL                                  :: avg

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_vector', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_fpos
      INTEGER                                            :: handle, i, irank, msglen, nranks, &
                                                            offset, unit_nr
      LOGICAL                                            :: is_new, my_avg, should_output
      REAL(KIND=dp), DIMENSION(3)                        :: avg_data
      REAL(KIND=dp), DIMENSION(:), POINTER               :: data_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: psection

      CALL timeset(routineN, handle)

      IF (PRESENT(avg)) THEN
         my_avg = avg
      ELSE
         my_avg = .FALSE.
      END IF

      IF (PRESENT(fpos)) THEN
         my_fpos = fpos
      ELSE
         my_fpos = "APPEND"
      END IF

      NULLIFY (logger, psection)
      logger => cp_get_default_logger()
      psection => section_vals_get_subs_vals(helium%input, pkey)
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=psection), cp_p_file)

      IF (.NOT. should_output) THEN
         CALL timestop(handle)
         RETURN
      END IF

      IF (my_avg) THEN
         ! average data over all processors
         avg_data(:) = DATA(:)
         CALL mp_sum(avg_data, logger%para_env%group)
         nranks = logger%para_env%num_pe
      ELSE
         ! gather data from all processors
         CALL mp_gather(DATA, &
                        helium%rtmp_3_np_1d, &
                        logger%para_env%source, &
                        logger%para_env%group)
      END IF

      unit_nr = cp_print_key_unit_nr( &
                logger, &
                psection, &
                middle_name=fname, &
                extension=".dat", &
                file_position=my_fpos, &
                is_new_file=is_new)

      IF (logger%para_env%ionode) THEN

         IF (is_new) THEN
            ! comment
            IF (cmmnt .NE. "") THEN
               WRITE (unit_nr, '(A)') "# "//cmmnt
            END IF
            ! column labels
            WRITE (unit_nr, '(A2,A18,1X,A20,1X,A20)') &
               "# ", &
               col_label(1), &
               col_label(2), &
               col_label(3)
         END IF

         IF (my_avg) THEN
            DO i = 1, 3
               WRITE (unit_nr, '(E27.20)', ADVANCE='NO') uconv*avg_data(i)
               IF (i .LT. 3) THEN
                  WRITE (unit_nr, '(1X)', ADVANCE='NO')
               END IF
            END DO
            WRITE (unit_nr, '(A)') ""
         ELSE
            ! iterate over processors/helium environments
            DO irank = 1, helium%num_env
               ! unpack data (actually point to the right fragment only)
               msglen = SIZE(DATA)
               offset = (irank-1)*msglen
               data_p => helium%rtmp_3_np_1d(offset+1:offset+msglen)
               ! write out the data
               DO i = 1, 3
                  WRITE (unit_nr, '(E27.20)', ADVANCE='NO') uconv*data_p(i)
                  IF (i .LT. 3) THEN
                     WRITE (unit_nr, '(1X)', ADVANCE='NO')
                  END IF
               END DO
               WRITE (unit_nr, '(A)') ""
            END DO
         END IF

         CALL m_flush(unit_nr)
         CALL cp_print_key_finished_output(unit_nr, logger, psection)

      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_vector

! ***************************************************************************
!> \brief  Print acceptance counts according to HELIUM%PRINT%ACCEPTS
!> \param helium ...
!> \date   2010-05-27
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_accepts(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_accepts', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, unit_nr
      LOGICAL                                            :: file_is_new, should_output
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%ACCEPTS")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (should_output) THEN

         IF (logger%para_env%ionode) THEN

            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name="helium-accepts", &
                      extension=".dat", &
                      is_new_file=file_is_new)

            IF (file_is_new) THEN
               WRITE (unit_nr, '(A8,1X,A15,1X,A20)', ADVANCE='NO') &
                  "# Length", &
                  "         Trials", &
                  "            Selected"
               DO j = 1, helium%bisctlog2
                  WRITE (unit_nr, '(A17,1X,I3)', ADVANCE='NO') "            Level", j
               END DO
               WRITE (unit_nr, '(A)') ""
            END IF

            DO i = 1, helium%maxcycle
               WRITE (unit_nr, '(I3)', ADVANCE='NO') i
               DO j = 1, helium%bisctlog2+2
                  WRITE (unit_nr, '(1X,F20.2)', ADVANCE='NO') helium%num_accepted(j, i)
               END DO
               WRITE (unit_nr, '(A)') ""
            END DO
            WRITE (unit_nr, '(A)') "&"

            CALL m_flush(unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, print_key)

         END IF
      END IF

      CALL timestop(handle)
      RETURN
   END SUBROUTINE helium_print_accepts

! ***************************************************************************
!> \brief  Print permutation state according to HELIUM%PRINT%PERM
!> \param helium ...
!> \date   2010-06-07
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_perm(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_perm', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: left_delim, right_delim
      CHARACTER(len=default_string_length)               :: msg_str, my_middle_name, stmp
      INTEGER                                            :: curat, handle, i, irank, lb, msglen, &
                                                            nused, offset, outformat, ub, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: my_cycle, my_perm, used_indices
      LOGICAL                                            :: first, first_cycle, should_output
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%PERM")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (.NOT. should_output) THEN
         CALL timestop(handle)
         RETURN
      END IF

      ! get the output type
      CALL section_vals_val_get(print_key, "FORMAT", i_val=outformat)

      IF (outformat .EQ. perm_cycle) THEN
         ! gather positions from all processors to logger%para_env%source
         helium%rtmp_3_atoms_beads_1d(:) = PACK(helium%pos(:, :, 1:helium%beads), .TRUE.)
         CALL mp_gather(helium%rtmp_3_atoms_beads_1d, &
                        helium%rtmp_3_atoms_beads_np_1d, &
                        logger%para_env%source, logger%para_env%group)
         ! set logical mask for unpacking coordinates gathered from other ranks
         helium%ltmp_3_atoms_beads_3d(:, :, :) = .TRUE.
      END IF

      ! gather permutation state from all processors to logger%para_env%source
      helium%itmp_atoms_np_1d(:) = 0
      CALL mp_gather(helium%permutation, &
                     helium%itmp_atoms_np_1d, &
                     logger%para_env%source, logger%para_env%group)

      IF (logger%para_env%ionode) THEN

         ! iterate over processors/helium environments
         DO irank = 1, helium%num_env

            ! generate one file per processor
            stmp = ""
            WRITE (stmp, *) irank
            my_middle_name = "helium-perm-"//TRIM(ADJUSTL(stmp))
            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name=TRIM(my_middle_name), &
                      extension=".dat")

            ! unpack permutation state (actually point to the right section only)
            lb = (irank-1)*helium%atoms+1
            ub = irank*helium%atoms
            my_perm => helium%itmp_atoms_np_1d(lb:ub)

            SELECT CASE (outformat)

            CASE (perm_cycle)
               ! write the permutation grouped into cycles

               ! unpack coordinates (necessary only for winding path delimiters)
               msglen = SIZE(helium%rtmp_3_atoms_beads_1d)
               offset = (irank-1)*msglen
               helium%work(:, :, 1:helium%beads) = &
                  UNPACK(helium%rtmp_3_atoms_beads_np_1d(offset+1:offset+msglen), &
                         MASK=helium%ltmp_3_atoms_beads_3d, FIELD=0.0_dp)

               curat = 1
               nused = 0
               first_cycle = .TRUE.
               DO WHILE (curat .LE. helium%atoms)

                  ! get the permutation cycle the current atom belongs to
                  my_cycle => helium_cycle_of(curat, my_perm)

                  ! include the current cycle in the pool of "used" indices
                  nused = nused+SIZE(my_cycle)
                  CALL reallocate(used_indices, 1, nused)
                  used_indices(nused-SIZE(my_cycle)+1:nused) = my_cycle(1:SIZE(my_cycle))

                  ! select delimiters accoring to the cycle's winding state
                  IF (helium_is_winding(helium, curat, helium%work, my_perm)) THEN
                     left_delim = "["
                     right_delim = "]"
                  ELSE
                     left_delim = "("
                     right_delim = ")"
                  END IF

                  ! cycle delimiter
                  IF (first_cycle) THEN
                     first_cycle = .FALSE.
                  ELSE
                     WRITE (unit_nr, '(1X)', ADVANCE='NO')
                  END IF

                  ! write out the current cycle
                  WRITE (unit_nr, '(A1)', ADVANCE='NO') left_delim
                  first = .TRUE.
                  DO i = 1, SIZE(my_cycle)
                     IF (first) THEN
                        first = .FALSE.
                     ELSE
                        WRITE (unit_nr, '(1X)', ADVANCE='NO')
                     END IF
                     WRITE (unit_nr, '(I0)', ADVANCE='NO') my_cycle(i)
                  END DO
                  WRITE (unit_nr, '(A1)', ADVANCE='NO') right_delim

                  ! clean up
                  DEALLOCATE (my_cycle)
                  NULLIFY (my_cycle)

                  ! try to increment the current atom index
                  DO WHILE (ANY(used_indices .EQ. curat))
                     curat = curat+1
                  END DO

               END DO
               WRITE (unit_nr, *)

               DEALLOCATE (used_indices)
               NULLIFY (used_indices)

            CASE (perm_plain)
               ! write the plain permutation

               first = .TRUE.
               DO i = 1, helium%atoms
                  IF (first) THEN
                     first = .FALSE.
                  ELSE
                     WRITE (unit_nr, '(1X)', ADVANCE='NO')
                  END IF
                  WRITE (unit_nr, '(I0)', ADVANCE='NO') my_perm(i)
               END DO
               WRITE (unit_nr, '(A)') ""

            CASE default

               msg_str = "Format for permutation output unknown."
               CALL cp__b("motion/helium_io.F",897,msg_str)

            END SELECT

            CALL m_flush(unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, print_key)

         END DO

      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_perm

! ***************************************************************************
!> \brief  Print coordinates according to HELIUM%PRINT%COORDINATES
!> \param helium ...
!> \date   2009-07-16
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_coordinates(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(3)                                       :: resName
      CHARACTER(len=default_string_length)               :: fmt_string, my_middle_name, stmp
      INTEGER                                            :: handle, ia, ib, ib1, ib2, ic, icycle, &
                                                            irank, msglen, offset, tmp1, tmp2, &
                                                            unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: my_perm
      LOGICAL                                            :: are_connected, is_winding, ltmp, &
                                                            should_output
      REAL(KIND=dp)                                      :: xtmp, ytmp, ztmp
      REAL(KIND=dp), DIMENSION(3)                        :: r0, r1, r2
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%COORDINATES")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (.NOT. should_output) THEN
         CALL timestop(handle)
         RETURN
      END IF

      ! prepare the coordinates for output (use unit cell centered around r0)
      r0(:) = helium%center(:)
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            r1(:) = helium%pos(:, ia, ib)-r0(:)
            r2(:) = helium%pos(:, ia, ib)-r0(:)
            CALL helium_pbc(helium, r2)
            ltmp = .FALSE.
            DO ic = 1, 3
               IF (ABS(r1(ic)-r2(ic)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                  ltmp = .TRUE.
                  CYCLE
               END IF
            END DO
            IF (ltmp) THEN
               helium%work(:, ia, ib) = r0(:)+r2(:)
            ELSE
               helium%work(:, ia, ib) = helium%pos(:, ia, ib)
            END IF
         END DO
      END DO

      ! gather positions from all processors to logger%para_env%source
      helium%rtmp_3_atoms_beads_1d(:) = PACK(helium%work(:, :, 1:helium%beads), .TRUE.)
      CALL mp_gather(helium%rtmp_3_atoms_beads_1d, &
                     helium%rtmp_3_atoms_beads_np_1d, &
                     logger%para_env%source, logger%para_env%group)

      ! gather permutation state from all processors to logger%para_env%source
      CALL mp_gather(helium%permutation, &
                     helium%itmp_atoms_np_1d, &
                     logger%para_env%source, logger%para_env%group)

      ! set logical mask for unpacking coordinates gathered from other ranks
      helium%ltmp_3_atoms_beads_3d(:, :, :) = .TRUE.

      IF (logger%para_env%ionode) THEN

         ! iterate over processors/helium environments
         DO irank = 1, helium%num_env

            ! generate one file per processor
            stmp = ""
            WRITE (stmp, *) irank
            my_middle_name = "helium-pos-"//TRIM(ADJUSTL(stmp))
            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name=TRIM(my_middle_name), &
                      extension=".pdb")

            ! write out the unit cell parameters
            fmt_string = "(A6,3F9.3,3F7.2,1X,A11,1X,I3)"
            xtmp = helium%cell_size
            xtmp = cp_unit_from_cp2k(xtmp, "angstrom")
            SELECT CASE (helium%cell_shape)
            CASE (helium_cell_shape_cube)
               stmp = "C          "
            CASE (helium_cell_shape_octahedron)
               stmp = "O          "
            CASE DEFAULT
               stmp = "UNKNOWN    "
            END SELECT
            WRITE (unit_nr, fmt_string) "CRYST1", &
               xtmp, xtmp, xtmp, &
               90.0_dp, 90.0_dp, 90.0_dp, &
               stmp, helium%beads

            ! unpack coordinates
            msglen = SIZE(helium%rtmp_3_atoms_beads_1d)
            offset = (irank-1)*msglen
            helium%work(:, :, 1:helium%beads) = &
               UNPACK(helium%rtmp_3_atoms_beads_np_1d(offset+1:offset+msglen), &
                      MASK=helium%ltmp_3_atoms_beads_3d, FIELD=0.0_dp)

            ! unpack permutation state (actually point to the right section only)
            msglen = SIZE(helium%permutation)
            offset = (irank-1)*msglen
            my_perm => helium%itmp_atoms_np_1d(offset+1:offset+msglen)

            ! write out coordinates
            fmt_string = &
               "(A6,I5,1X,A4,A1,A3,1X,A1,I4,A1,3X,3F8.3,2F6.2,10X,A2,A2)"
            DO ia = 1, helium%atoms
               icycle = helium_cycle_number(helium, ia, my_perm)
               is_winding = helium_is_winding(helium, ia, helium%work, my_perm)
               IF (is_winding) THEN
                  resName = "SPR"
               ELSE
                  resName = "NRM"
               END IF
               DO ib = 1, helium%beads
                  xtmp = helium%work(1, ia, ib)
                  xtmp = cp_unit_from_cp2k(xtmp, "angstrom")
                  ytmp = helium%work(2, ia, ib)
                  ytmp = cp_unit_from_cp2k(ytmp, "angstrom")
                  ztmp = helium%work(3, ia, ib)
                  ztmp = cp_unit_from_cp2k(ztmp, "angstrom")
                  WRITE (unit_nr, fmt_string) "ATOM  ", &
                     (ia-1)*helium%beads+ib, &
                     " He ", " ", resName, "X", &
                     icycle, &
                     " ", &
                     xtmp, ytmp, ztmp, &
                     1.0_dp, 0.0_dp, "HE", "  "
               END DO
            END DO

            ! write out the bead connectivity information
            DO ia = 1, helium%atoms

               ! write connectivity records for this atom only if the path
               ! it belongs to is longer than 1.
               IF (helium_path_length(helium, ia, my_perm) .LE. 1) THEN
                  CYCLE
               END IF

               DO ib = 1, helium%beads-1
                  ! check wheather the consecutive beads belong to the same box
                  r1(:) = helium%work(:, ia, ib)-helium%work(:, ia, ib+1)
                  r2(:) = r1(:)
                  CALL helium_pbc(helium, r2)
                  are_connected = .TRUE.
                  DO ic = 1, 3
                     IF (ABS(r1(ic)-r2(ic)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                        ! if the distance betw ib and ib+1 changes upon applying
                        ! PBC do not connect them
                        are_connected = .FALSE.
                        CYCLE
                     END IF
                  END DO
                  IF (are_connected) THEN
                     tmp1 = (ia-1)*helium%beads+ib
                     tmp2 = (ia-1)*helium%beads+ib+1
                     ! smaller value has to go first
                     IF (tmp1 .LT. tmp2) THEN
                        ib1 = tmp1
                        ib2 = tmp2
                     ELSE
                        ib1 = tmp2
                        ib2 = tmp1
                     END IF
                     WRITE (unit_nr, '(A6,2I5)') "CONECT", ib1, ib2
                  END IF
               END DO

               ! last bead of atom <ia> connects to the first bead
               ! of the next atom in the permutation cycle
               r1(:) = helium%work(:, ia, helium%beads)-helium%work(:, my_perm(ia), 1)
               r2(:) = r1(:)
               CALL helium_pbc(helium, r2)
               are_connected = .TRUE.
               DO ic = 1, 3
                  IF (ABS(r1(ic)-r2(ic)) .GT. 100.0_dp*EPSILON(0.0_dp)) THEN
                     ! if the distance betw ib and ib+1 changes upon applying
                     ! PBC do not connect them
                     are_connected = .FALSE.
                     CYCLE
                  END IF
               END DO
               IF (are_connected) THEN
                  tmp1 = ia*helium%beads
                  tmp2 = (my_perm(ia)-1)*helium%beads+1
                  IF (tmp1 .LT. tmp2) THEN
                     ib1 = tmp1
                     ib2 = tmp2
                  ELSE
                     ib1 = tmp2
                     ib2 = tmp1
                  END IF
                  WRITE (unit_nr, '(A6,2I5)') "CONECT", ib1, ib2
               END IF
            END DO
            WRITE (unit_nr, '(A)') "END"

            CALL m_flush(unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, print_key)

         END DO

      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_coordinates

! ***************************************************************************
!> \brief  Print radial distribution functions according to HELIUM%PRINT%RDF
!> \param helium ...
!> \date   2009-07-23
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_rdf(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_rdf', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: comment, stmp
      INTEGER                                            :: handle, ia, ic, id, itmp, iweight, &
                                                            nsteps, unit_nr
      LOGICAL                                            :: should_output
      REAL(KIND=dp)                                      :: inv_norm, rtmp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: message
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%RDF")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (should_output) THEN

         ! work on the temporary array so that accumulated data remains intact
         helium%rdf_inst(:, :, :) = helium%rdf_accu(:, :, :)

         ! average over helium environments
         NULLIFY (message)
         message => helium%rdf_inst(:, :, :)
         CALL mp_sum(message, logger%para_env%group)
         itmp = logger%para_env%num_pe
         inv_norm = 1.0_dp/REAL(itmp, dp)
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)*inv_norm

         ! average over steps performed so far in this run
         nsteps = helium%current_step-helium%first_step
         inv_norm = 1.0_dp/REAL(nsteps, dp)
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)*inv_norm

         iweight = helium%rdf_iweight
         ! average over the old and the current density (observe the weights!)
         helium%rdf_inst(:, :, :) = nsteps*helium%rdf_inst(:, :, :)+ &
                                    iweight*helium%rdf_rstr(:, :, :)
         helium%rdf_inst(:, :, :) = helium%rdf_inst(:, :, :)/REAL(nsteps+iweight, dp)

         IF (logger%para_env%ionode) THEN

            DO id = 1, helium%rdf_num ! loop over estimators ---

               ! overwrite RDF file each time it is written
               stmp = ""
               WRITE (stmp, *) id
               unit_nr = cp_print_key_unit_nr( &
                         logger, &
                         print_key, &
                         middle_name="helium-rdf-"//TRIM(ADJUSTL(stmp)), &
                         extension=".dat", &
                         file_position="REWIND", &
                         do_backup=.FALSE.)

               comment = "Radial distribution function"

               DO ic = 1, helium%rdf_nbin
                  rtmp = (REAL(ic, dp)-0.5_dp)*helium%rdf_delr
                  rtmp = cp_unit_from_cp2k(rtmp, "angstrom")
                  WRITE (unit_nr, '(F20.10)', ADVANCE='NO') rtmp
                  DO ia = 1, helium%rdf_num_ctr
                     WRITE (unit_nr, '(F20.10)', ADVANCE='NO') helium%rdf_inst(id, ic, ia)
                  END DO
                  WRITE (unit_nr, *)
               END DO

               CALL m_flush(unit_nr)
               CALL cp_print_key_finished_output(unit_nr, logger, print_key)

            END DO ! loop over estimators ---

         END IF
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_rdf

! ***************************************************************************
!> \brief  Print densities according to HELIUM%PRINT%RHO
!> \param helium ...
!> \date   2011-06-21
!> \par History
!>      08.2015 cleaned code from unneded arrays
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_rho(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_rho', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: comment, fname
      INTEGER                                            :: handle, ic, id, itmp, iweight, nsteps, &
                                                            unit_nr
      LOGICAL                                            :: should_output
      REAL(KIND=dp)                                      :: inv_norm, invproc
      REAL(KIND=dp), DIMENSION(3)                        :: center
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: cubdata
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%RHO")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (.NOT. should_output) THEN
         CALL timestop(handle)
         RETURN
      END IF

      ! work on temporary array so that the average remains intact
      helium%rho_inst(:, :, :, :) = helium%rho_accu(:, :, :, :)

      ! average over processors / helium environments
      CALL mp_sum(helium%rho_inst, logger%para_env%group)
      itmp = logger%para_env%num_pe
      invproc = 1.0_dp/REAL(itmp, dp)
      helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)*invproc

      ! average over steps performed so far in this run
      nsteps = helium%current_step-helium%first_step
      inv_norm = 1.0_dp/REAL(nsteps, dp)
      helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)*inv_norm

      iweight = helium%rho_iweight
      ! average over the old and the current density (observe the weights!)
      helium%rho_inst(:, :, :, :) = nsteps*helium%rho_inst(:, :, :, :)+ &
                                    iweight*helium%rho_rstr(:, :, :, :)
      helium%rho_inst(:, :, :, :) = helium%rho_inst(:, :, :, :)/REAL(nsteps+iweight, dp)

      ! set center of the cubefile
      IF (helium%solute_present) THEN
         ! should be set to solute's COM
         center(:) = helium%center(:)
      ELSE
         ! regardless of whether we are periodic or not use the origin, since
         ! pure cluster's COM might drift, but we want the cube fixed (note that
         ! the densities are correctly calculated wrt to COM in such case)
         center(:) = (/0.0_dp, 0.0_dp, 0.0_dp/)
      END IF

      DO id = 1, rho_num ! loop over densities ---

         IF (.NOT. helium%rho_property(id)%is_calculated) THEN
            ! skip densities that are not requested by the user
            CYCLE
         END IF

         DO ic = 1, helium%rho_property(id)%num_components ! loop over components

            WRITE (fname, '(A)') "helium-rho-"// &
               TRIM(ADJUSTL(helium%rho_property(id)%filename_suffix(ic)))
            IF (helium%rho_property(id)%component_name(ic) .EQ. "") THEN
               WRITE (comment, '(A)') TRIM(helium%rho_property(id)%name)
            ELSE
               WRITE (comment, '(A)') TRIM(helium%rho_property(id)%name)// &
                  ", "// &
                  TRIM(helium%rho_property(id)%component_name(ic))
            END IF
            cubdata => helium%rho_inst(helium%rho_property(id)%component_index(ic), :, :, :)

            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name=TRIM(ADJUSTL(fname)), &
                      extension=".cube", &
                      file_position="REWIND", &
                      do_backup=.FALSE.)

            IF (logger%para_env%ionode) THEN
               CALL helium_write_cubefile( &
                  unit_nr, &
                  comment, &
                  center-0.5_dp*(helium%rho_maxr-helium%rho_delr), &
                  helium%rho_delr, &
                  helium%rho_nbin, &
                  cubdata)
               CALL m_flush(unit_nr)
               CALL cp_print_key_finished_output(unit_nr, logger, print_key)
            END IF

         END DO ! loop over components

      END DO ! loop over densities ---

      CALL timestop(handle)

   END SUBROUTINE helium_print_rho

! ***************************************************************************
!> \brief Write volumetric data to an orthorhombic cubefile
!> \param   unit - unit number to which output will be sent
!> \param   comment - description of the data stored in the cubefile
!> \param   origin - position of the cubefile origin
!> \param   deltar - voxel side length
!> \param   ndim - number of voxels in each dimension
!> \param   DATA - array (ndim x ndim x ndim) with the data for output
!> \date 2013-11-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_write_cubefile(unit, comment, origin, deltar, ndim, DATA)

      INTEGER, INTENT(IN)                                :: unit
      CHARACTER(len=default_string_length), INTENT(IN)   :: comment
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: origin
      REAL(KIND=dp), INTENT(IN)                          :: deltar
      INTEGER, INTENT(IN)                                :: ndim
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN), &
         POINTER                                         :: DATA

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_write_cubefile', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ix, iy, iz, nw
      REAL(KIND=dp)                                      :: delr, inva3
      REAL(KIND=dp), DIMENSION(3)                        :: orig

      CALL timeset(routineN, handle)

      ! convert cubefile data to the proper units of measure
      delr = angstrom*deltar
      orig(:) = angstrom*origin(:)

      ! output cube file header
      WRITE (unit, '(A)') comment
      WRITE (unit, '(A)') "Volumetric data in cubefile format generated by CP2K"
      WRITE (unit, '(I5,3(1X,F12.8))') 0, orig(1), orig(2), orig(3)
      WRITE (unit, '(I5,3(1X,F12.8))') ndim, delr, 0.0_dp, 0.0_dp
      WRITE (unit, '(I5,3(1X,F12.8))') ndim, 0.0_dp, delr, 0.0_dp
      WRITE (unit, '(I5,3(1X,F12.8))') ndim, 0.0_dp, 0.0_dp, delr

      ! output cube file data
      nw = 0
      inva3 = 1.0_dp/(angstrom*angstrom*angstrom)
      DO ix = 1, ndim
         DO iy = 1, ndim
            DO iz = 1, ndim
               WRITE (unit, '(1X,E13.5)', ADVANCE='NO') inva3*DATA(ix, iy, iz)
               nw = nw+1
               IF (MOD(nw, 6) .EQ. 0) THEN
                  nw = 0
                  WRITE (unit, *)
               END IF
            END DO
         END DO
      END DO
      ! some compilers write over the first entry on a line loosing all previous
      ! values written on that line unless line terminator is written at the end
      ! so make sure that the last WRITE statement runs without ADVANCE='NO'
      ! (observed for ifort 14.0.1 and 14.0.2 but not for gfortran 4.8.2)
      IF (nw .NE. 0) THEN
         WRITE (unit, *)
      END IF

      CALL timestop(handle)

   END SUBROUTINE helium_write_cubefile

! ***************************************************************************
!> \brief  Print permutation length according to HELIUM%PRINT%PLENGTH
!> \param helium ...
!> \date   2010-06-07
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_plength(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_plength', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, unit_nr
      LOGICAL                                            :: is_new, should_output
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%PLENGTH")
      should_output = BTEST(cp_print_key_should_output( &
                            iteration_info=logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (should_output) THEN

         IF (logger%para_env%ionode) THEN

            unit_nr = cp_print_key_unit_nr( &
                      logger, &
                      print_key, &
                      middle_name="helium-plength", &
                      extension=".dat", &
                      is_new_file=is_new)

            DO i = 1, helium%atoms
               WRITE (unit_nr, '(F20.10)', ADVANCE='NO') helium%plength_avrg(i)
               IF (i .LT. helium%atoms) THEN
                  WRITE (unit_nr, '(1X)', ADVANCE='NO')
               END IF
            END DO
            WRITE (unit_nr, '(A)') ""

            CALL m_flush(unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, print_key)

         END IF
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_plength

! ***************************************************************************
!> \brief  Print helium force according to HELIUM%PRINT%FORCE
!> \param helium ...
!> \date   2010-01-27
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_print_force(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_print_force', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msgstr
      INTEGER                                            :: handle, ia, ib, ic, idim, unit_nr
      LOGICAL                                            :: should_output
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      NULLIFY (logger, print_key)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%FORCES")
      should_output = BTEST(cp_print_key_should_output( &
                            logger%iter_info, &
                            basis_section=print_key), cp_p_file)

      IF (.NOT. should_output) THEN
         CALL timestop(handle)
         RETURN
      END IF

      ! check if there is anything to be printed out
      IF (.NOT. helium%solute_present) THEN
         msgstr = "Warning: force printout requested but there is no solute!"
         CALL helium_write_line(msgstr)
         CALL timestop(handle)
         RETURN
      END IF

      IF (logger%para_env%ionode) THEN

         unit_nr = cp_print_key_unit_nr( &
                   logger, &
                   print_key, &
                   middle_name="helium-force", &
                   extension=".dat")

         ! print all force components in one line
         DO ib = 1, helium%solute_beads
            idim = 0
            DO ia = 1, helium%solute_atoms
               DO ic = 1, 3
                  idim = idim+1
                  WRITE (unit_nr, '(F20.10)', ADVANCE='NO') helium%force_avrg(ib, idim)
               END DO
            END DO
         END DO
         WRITE (unit_nr, *)

         CALL m_flush(unit_nr)
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)

      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_print_force


END MODULE helium_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods that handle helium-solvent and helium-helium interactions
!> \author Lukasz Walewski
!> \date   2009-06-10
! **************************************************************************************************
MODULE helium_interactions

   USE helium_common,                   ONLY: helium_eval_expansion,&
                                              helium_pbc
   USE helium_types,                    ONLY: &
        e_id_interact, e_id_kinetic, e_id_potential, e_id_thermo, e_id_total, e_id_virial, &
        helium_solvent_type, hid_chlorine, hid_hydrogen, hid_oxygen, int_arr_ptr
   USE input_constants,                 ONLY: helium_solute_intpot_mwater,&
                                              helium_solute_intpot_none
   USE kinds,                           ONLY: dp
   USE physcon,                         ONLY: angstrom,&
                                              kelvin
   USE pint_types,                      ONLY: pint_env_type
   USE splines_types,                   ONLY: spline_data_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_interactions'

   PUBLIC :: helium_calc_energy
   PUBLIC :: helium_solute_e_f
   PUBLIC :: helium_bead_solute_e_f
   PUBLIC :: helium_intpot_scan

CONTAINS

! ***************************************************************************
!> \brief  Calculate the helium energy (including helium-solute interaction)
!> \param    helium   - helium environment
!> \param    pint_env - path integral environment
!> \par History
!>         2009-06 moved I/O out from here [lwalewski]
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_calc_energy(helium, pint_env)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_calc_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: b, bead, c, i, j, n
      INTEGER, DIMENSION(:), POINTER                     :: perm
      LOGICAL                                            :: nperiodic
      REAL(KIND=dp)                                      :: a, cell_size, en, interac, kin, pot, &
                                                            rmax, rmin, vkin
      REAL(KIND=dp), DIMENSION(3)                        :: r, rp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos
      TYPE(spline_data_p_type), DIMENSION(:, :), POINTER :: eij

      pos => helium%pos
      perm => helium%permutation
      eij => helium%eij
      cell_size = 0.5_dp*helium%cell_size
      nperiodic = .NOT. helium%periodic
      n = helium%atoms
      b = helium%beads
      en = 0.0_dp
      pot = 0.0_dp
      rmin = 1.0e20_dp
      rmax = 0.0_dp
      DO i = 1, n-1
         DO j = i+1, n
            rp(:) = pos(:, i, 1)-pos(:, j, 1)
            CALL helium_pbc(helium, rp)
            DO bead = 2, b
               a = 0.0_dp
               DO c = 1, 3
                  r(c) = rp(c)
                  a = a+r(c)**2
                  rp(c) = pos(c, i, bead)-pos(c, j, bead)
               END DO
               CALL helium_pbc(helium, rp)
               en = en+helium_eval_expansion(helium, r, rp, eij, 0)
               a = SQRT(a)
               IF (a < rmin) rmin = a
               IF (a > rmax) rmax = a
               IF ((a < cell_size) .OR. nperiodic) THEN
                  pot = pot+helium_vij(a)
               END IF
            END DO
            a = 0.0_dp
            DO c = 1, 3
               r(c) = rp(c)
               a = a+r(c)**2
               rp(c) = pos(c, perm(i), 1)-pos(c, perm(j), 1)
            END DO
            CALL helium_pbc(helium, rp)
            en = en+helium_eval_expansion(helium, r, rp, eij, 0)
            a = SQRT(a)
            IF (a < rmin) rmin = a
            IF (a > rmax) rmax = a
            IF ((a < cell_size) .OR. nperiodic) THEN
               pot = pot+helium_vij(a)
            END IF
         END DO
      END DO
      pot = pot/b
      en = en/b

      ! helium-solute interaction energy (all beads of all particles)
      interac = 0.0_dp
      IF (helium%solute_present) THEN
         CALL helium_solute_e(pint_env, helium, interac)
      END IF
      interac = interac/b

!TODO:
      vkin = 0.0_dp
!   vkin = helium_virial_energy(helium)

      kin = 0.0_dp
      DO i = 1, n
         r(:) = pos(:, i, b)-pos(:, perm(i), 1)
         CALL helium_pbc(helium, r)
         kin = kin+r(1)*r(1)+r(2)*r(2)+r(3)*r(3)
         DO bead = 2, b
            r(:) = pos(:, i, bead-1)-pos(:, i, bead)
            CALL helium_pbc(helium, r)
            kin = kin+r(1)*r(1)+r(2)*r(2)+r(3)*r(3)
         END DO
      END DO
      kin = 1.5_dp*n/helium%tau-0.5*kin/(b*helium%tau**2*helium%hb2m)

! TODO: move printing somwhere else ?
!   print *,"POT = ",(pot/n+helium%e_corr)*kelvin,"K"
!   print *,"INTERAC = ",interac*kelvin,"K"
!   print *,"RMIN= ",rmin*angstrom,"A"
!   print *,"RMAX= ",rmax*angstrom,"A"
!   print *,"EVIRIAL not valid!"
!   print *,"ETHERMO= ",((en+kin)/n+helium%e_corr)*kelvin,"K"
!   print *,"ECORR= ",helium%e_corr*kelvin,"K"
!!   kin = helium_total_action(helium)
!!   print *,"ACTION= ",kin
!   print *,"WINDING#= ",helium_calc_winding(helium)

      helium%energy_inst(e_id_potential) = pot/n+helium%e_corr
      helium%energy_inst(e_id_kinetic) = (en-pot+kin)/n
      helium%energy_inst(e_id_interact) = interac
      helium%energy_inst(e_id_thermo) = (en+kin)/n+helium%e_corr
      helium%energy_inst(e_id_virial) = 0.0_dp !(en+vkin)/n+helium%e_corr
      helium%energy_inst(e_id_total) = (en+vkin)/n+helium%e_corr

      RETURN
   END SUBROUTINE helium_calc_energy

! ***************************************************************************
!> \brief Calculate general helium-solute interaction energy (and forces)
!>        between one helium bead and the corresponding solute time slice.
!> \param pint_env           path integral environment
!> \param helium ...
!> \param helium_part_index  helium particle index
!> \param helium_slice_index helium time slice index
!> \param helium_r_opt       explicit helium bead coordinates (optional)
!> \param energy             calculated energy
!> \param force              calculated force (if requested)
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_bead_solute_e_f(pint_env, helium, helium_part_index, &
                                     helium_slice_index, helium_r_opt, energy, force)

      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: helium_part_index, helium_slice_index
      REAL(KIND=dp), DIMENSION(3), INTENT(IN), OPTIONAL  :: helium_r_opt
      REAL(KIND=dp), INTENT(OUT)                         :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL, POINTER                               :: force

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_bead_solute_e_f', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: hbeads, hi, qi
      REAL(KIND=dp), DIMENSION(3)                        :: helium_r
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_force

      hbeads = helium%beads
      ! helium bead index that is invariant wrt the rotations
      hi = MOD(helium_slice_index-1+hbeads+helium%relrot, hbeads)+1
      ! solute bead index that belongs to hi helium index
      qi = ((hi-1)*pint_env%p)/hbeads+1

      ! coordinates of the helium bead
      IF (PRESENT(helium_r_opt)) THEN
         helium_r(:) = helium_r_opt(:)
      ELSE
         helium_r(:) = helium%pos(:, helium_part_index, helium_slice_index)
      END IF

      SELECT CASE (helium%solute_interaction)

      CASE (helium_solute_intpot_mwater)
         IF (PRESENT(force)) THEN
            force(:, :) = 0.0_dp
            my_force => force(qi, :)
            CALL helium_intpot_model_water( &
               pint_env%x(qi, :), &
               helium%solute_i, &
               helium, &
               helium_r, &
               energy, &
               my_force &
               )
         ELSE
            CALL helium_intpot_model_water( &
               pint_env%x(qi, :), &
               helium%solute_i, &
               helium, &
               helium_r, &
               energy &
               )
         END IF

      CASE (helium_solute_intpot_none)
         energy = 0.0_dp
         IF (PRESENT(force)) THEN
            force(:, :) = 0.0_dp
         END IF

      CASE DEFAULT

      END SELECT

      RETURN
   END SUBROUTINE helium_bead_solute_e_f

! ***************************************************************************
!> \brief Calculate total helium-solute interaction energy and forces.
!> \param   pint_env - path integral environment
!> \param helium ...
!> \param   energy   - calculated interaction energy
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_solute_e_f(pint_env, helium, energy)

      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), INTENT(OUT)                         :: energy

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_solute_e_f', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, jb, jc
      REAL(KIND=dp)                                      :: my_energy
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: force

      NULLIFY (force)
      force => helium%force_inst

      energy = 0.0_dp
      force(:, :) = 0.0_dp

      ! calculate the total interaction energy and gradients between the
      ! solute and the helium, sum over all beads of all He particles
      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            CALL helium_bead_solute_e_f(pint_env, helium, ia, ib, &
                                        energy=my_energy, force=helium%rtmp_p_ndim_2d)
            energy = energy+my_energy
            DO jb = 1, pint_env%p
               DO jc = 1, pint_env%ndim
                  force(jb, jc) = force(jb, jc)+helium%rtmp_p_ndim_2d(jb, jc)
               END DO
            END DO
         END DO
      END DO

      RETURN
   END SUBROUTINE helium_solute_e_f

! ***************************************************************************
!> \brief Calculate total helium-solute interaction energy.
!> \param   pint_env - path integral environment
!> \param helium ...
!> \param   energy   - calculated interaction energy
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_solute_e(pint_env, helium, energy)

      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), INTENT(OUT)                         :: energy

      INTEGER                                            :: ia, ib
      REAL(KIND=dp)                                      :: my_energy

      energy = 0.0_dp

      DO ia = 1, helium%atoms
         DO ib = 1, helium%beads
            CALL helium_bead_solute_e_f(pint_env, helium, ia, ib, energy=my_energy)
            energy = energy+my_energy
         END DO
      END DO

      RETURN
   END SUBROUTINE helium_solute_e

! ***************************************************************************
!> \brief Calculate l-th Legendre polynomial P_{l}(x) at point x
!> \param x ...
!> \param n ...
!> \retval Pl ...
! **************************************************************************************************
   FUNCTION Pl(x, n)
      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp)                                      :: Pl

      INTEGER                                            :: k
      REAL(KIND=dp)                                      :: pln(0:n)

      pln(0) = 1.0_dp
      pln(1) = x

      IF (n <= 1) THEN
         Pl = pln(n)
      ELSE
         DO k = 1, n-1
            pln(k+1) = ((2.0*k+1.0)*x*pln(k)-REAL(k, dp)*pln(k-1))/(REAL(k+1, dp))
         END DO
         Pl = pln(n)
      END IF
      RETURN
   END FUNCTION Pl

! ***************************************************************************
!> \brief  Scan the helium-solute interaction energy within the periodic cell
!> \param pint_env ...
!> \param helium ...
!> \date   2014-01-22
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_intpot_scan(pint_env, helium)

      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_intpot_scan', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, ix, iy, iz, nbin
      LOGICAL                                            :: wrapped
      REAL(KIND=dp)                                      :: delr, my_en, ox, oy, oz
      REAL(kind=dp), DIMENSION(3)                        :: pbc1, pbc2, pos

      CALL timeset(routineN, handle)

      nbin = helium%rho_nbin
      delr = helium%rho_delr
      ox = helium%center(1)-helium%rho_maxr/2.0_dp
      oy = helium%center(2)-helium%rho_maxr/2.0_dp
      oz = helium%center(3)-helium%rho_maxr/2.0_dp

      DO ix = 1, nbin
         DO iy = 1, nbin
            DO iz = 1, nbin

               ! put the probe in the center of the current voxel
               pos(:) = (/ox+(ix-0.5_dp)*delr, oy+(iy-0.5_dp)*delr, oz+(iz-0.5_dp)*delr/)

               ! calc interaction energy for the current probe position
               helium%pos(:, 1, 1) = pos(:)
               CALL helium_bead_solute_e_f(pint_env, helium, 1, 1, energy=my_en)

               ! check if the probe fits within the unit cell
               pbc1(:) = pos(:)-helium%center
               pbc2(:) = pbc1(:)
               CALL helium_pbc(helium, pbc2)
               wrapped = .FALSE.
               DO ic = 1, 3
                  IF (ABS(pbc1(ic)-pbc2(ic)) .GT. 10.0_dp*EPSILON(0.0_dp)) THEN
                     wrapped = .TRUE.
                  END IF
               END DO

               ! set the interaction energy value
               IF (wrapped) THEN
                  helium%rho_inst(1, ix, iy, iz) = 0.0_dp
               ELSE
                  helium%rho_inst(1, ix, iy, iz) = my_en
               END IF

            END DO
         END DO
      END DO

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_intpot_scan

! ***************************************************************************
!> \brief Calculate model helium-solute interaction energy and forces
!>        between one helium bead and the corresponding solute time
!>        slice asuming water solute.
!> \param solute_x  solute positions ARR(3*NATOMS)
!> \param solute_i  solute indices, mapping from instances of a given element
!>        to global atom indices
!> \param helium    only needed for helium_pbc call at the moment
!> \param helium_x  helium bead position ARR(3)
!> \param energy    calculated interaction energy
!> \param force ...
!> \author Felix Uhl
! **************************************************************************************************
   SUBROUTINE helium_intpot_model_water(solute_x, solute_i, helium, helium_x, energy, force)

      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: solute_x
      TYPE(int_arr_ptr), DIMENSION(:), INTENT(IN)        :: solute_i
      TYPE(helium_solvent_type), POINTER                 :: helium
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: helium_x
      REAL(KIND=dp), INTENT(OUT)                         :: energy
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT), &
         OPTIONAL, POINTER                               :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'helium_intpot_model_water', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ig, num_chlorine, num_hydrogen, &
                                                            num_oxygen
      REAL(KIND=dp)                                      :: d, d2, dd, ep, eps, s1, s2, sig
      REAL(KIND=dp), DIMENSION(3)                        :: dr, solute_r

      IF (ASSOCIATED(solute_i(hid_chlorine)%iap)) THEN
         num_chlorine = SIZE(solute_i(hid_chlorine)%iap)
      ELSE
         num_chlorine = 0
      END IF

      IF (ASSOCIATED(solute_i(hid_oxygen)%iap)) THEN
         num_oxygen = SIZE(solute_i(hid_oxygen)%iap)
      ELSE
         num_oxygen = 0
      END IF

      IF (ASSOCIATED(solute_i(hid_hydrogen)%iap)) THEN
         num_hydrogen = SIZE(solute_i(hid_hydrogen)%iap)
      ELSE
         num_hydrogen = 0
      END IF

      energy = 0.0_dp
      IF (PRESENT(force)) THEN
         force(:) = 0.0_dp
      END IF

      sig = 2.69_dp ! 1.4 Angstrom
      eps = 60.61e-6_dp ! 19 K
      s1 = 0.0_dp
      DO i = 1, num_hydrogen
         ig = solute_i(hid_hydrogen)%iap(i)-1 ! global hydrogen index (3 == H)
         solute_r(1) = solute_x(3*ig+1)
         solute_r(2) = solute_x(3*ig+2)
         solute_r(3) = solute_x(3*ig+3)
         dr(:) = solute_r(:)-helium_x(:)
         CALL helium_pbc(helium, dr)
         d2 = dr(1)*dr(1)+dr(2)*dr(2)+dr(3)*dr(3)
         d = SQRT(d2)
         dd = (sig/d)**6
         ep = 4.0_dp*eps*dd*(dd-1.0_dp)
         s1 = s1+ep
         s2 = 24.0_dp*eps*dd*(2.0_dp*dd-1.0_dp)/d2
         IF (PRESENT(force)) THEN
            force(3*ig+1) = force(3*ig+1)+s2*dr(1)
            force(3*ig+2) = force(3*ig+2)+s2*dr(2)
            force(3*ig+3) = force(3*ig+3)+s2*dr(3)
         END IF
      END DO ! i = 1, num_hydrogen
      energy = energy+s1

      sig = 5.01_dp ! 2.6 Angstrom
      eps = 104.5e-6_dp ! 33 K
      s1 = 0.0_dp
      DO i = 1, num_oxygen
         ig = solute_i(hid_oxygen)%iap(i)-1 ! global oxygen index (2 == O)
         solute_r(1) = solute_x(3*ig+1)
         solute_r(2) = solute_x(3*ig+2)
         solute_r(3) = solute_x(3*ig+3)
         dr(:) = solute_r(:)-helium_x(:)
         CALL helium_pbc(helium, dr)
         d2 = dr(1)*dr(1)+dr(2)*dr(2)+dr(3)*dr(3)
         d = SQRT(d2)
         dd = (sig/d)**6
         ep = 4.0_dp*eps*dd*(dd-1.0_dp)
         s1 = s1+ep
         s2 = 24.0_dp*eps*dd*(2.0_dp*dd-1.0_dp)/d2
         IF (PRESENT(force)) THEN
            force(3*ig+1) = force(3*ig+1)+s2*dr(1)
            force(3*ig+2) = force(3*ig+2)+s2*dr(2)
            force(3*ig+3) = force(3*ig+3)+s2*dr(3)
         END IF
      END DO ! i = 1, num_chlorine
      energy = energy+s1

      RETURN

   END SUBROUTINE helium_intpot_model_water

! ***************************************************************************
!> \brief Helium-helium pair interaction potential.
!> \param r ...
!> \retval vij ...
! **************************************************************************************************
   FUNCTION helium_vij(r) RESULT(vij)

      REAL(kind=dp), INTENT(IN)                          :: r
      REAL(kind=dp)                                      :: vij

      REAL(kind=dp)                                      :: f, x, x2

      x = angstrom*r/2.9673_dp
      IF (x < 1.241314_dp) THEN
         x2 = 1.241314_dp/x-1.0_dp
         f = EXP(-x2*x2)
      ELSE
         f = 1.0_dp
      END IF
      x2 = 1.0_dp/(x*x)
      vij = 10.8_dp/kelvin*(544850.4_dp*EXP(-13.353384_dp*x)-f* &
                            ((0.1781_dp*x2+0.4253785_dp)*x2+1.3732412_dp)*x2*x2*x2)
      RETURN
   END FUNCTION helium_vij


END MODULE helium_interactions
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods for sampling helium variables
!> \author Lukasz Walewski
!> \date   2009-06-10
! **************************************************************************************************
MODULE helium_sampling

   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_rm_iter_level
   USE global_types,                    ONLY: global_environment_type
   USE helium_common,                   ONLY: &
        helium_boxmean_3d, helium_calc_plength, helium_calc_rdf, helium_calc_rho, helium_com, &
        helium_eval_expansion, helium_pbc, helium_rotate, helium_set_rdf_coord_system, &
        helium_spline, helium_total_moment_of_inertia, helium_total_projected_area, &
        helium_total_winding_number, helium_update_transition_matrix
   USE helium_interactions,             ONLY: helium_bead_solute_e_f,&
                                              helium_calc_energy,&
                                              helium_solute_e_f
   USE helium_io,                       ONLY: &
        helium_print_accepts, helium_print_coordinates, helium_print_energy, helium_print_force, &
        helium_print_perm, helium_print_plength, helium_print_rdf, helium_print_rho, &
        helium_print_vector, helium_write_line
   USE helium_types,                    ONLY: e_id_total,&
                                              helium_solvent_type
   USE input_constants,                 ONLY: &
        helium_forces_average, helium_forces_last, helium_mdist_exponential, &
        helium_mdist_gaussian, helium_mdist_linear, helium_mdist_quadratic, helium_mdist_singlev, &
        helium_mdist_uniform, helium_sampling_ceperley
   USE input_cp2k_restarts,             ONLY: write_restart
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_walltime
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
   USE parallel_rng_types,              ONLY: next_random_number
   USE physcon,                         ONLY: angstrom
   USE pint_types,                      ONLY: pint_env_type
   USE splines_types,                   ONLY: spline_data_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_sampling'

   PUBLIC :: helium_do_run
   PUBLIC :: helium_sample
   PUBLIC :: helium_step

CONTAINS

! ***************************************************************************
!> \brief  Performs MC simulation for helium (only)
!> \param helium ...
!> \param globenv ...
!> \date   2009-07-14
!> \author Lukasz Walewski
!> \note   This routine gets called only when HELIUM_ONLY is set to .TRUE.,
!>         so do not put any property calculations here!
! **************************************************************************************************
   SUBROUTINE helium_do_run(helium, globenv)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_do_run', routineP = moduleN//':'//routineN

      INTEGER                                            :: step, tot_steps
      LOGICAL                                            :: should_stop
      TYPE(pint_env_type), POINTER                       :: pint_env

      NULLIFY (pint_env)

      NULLIFY (helium%logger)
      helium%logger => cp_get_default_logger()

      ! create iteration level
      ! The iteration level 'MD' can be misleading since this is a pure MC
      ! simulation, however the helium code is meant to be run mainly from PINT.
      ! We keep it like that for compatibility with PINT+HELIUM simulations.
      CALL cp_add_iter_level(helium%logger%iter_info, "MD")
      CALL cp_iterate(helium%logger%iter_info, iter_nr=helium%first_step)
      tot_steps = helium%first_step

      ! set the properties accumulated over the whole MC process to 0
      helium%proarea%accu(:) = 0.0_dp
      helium%prarea2%accu(:) = 0.0_dp
      helium%wnmber2%accu(:) = 0.0_dp
      helium%mominer%accu(:) = 0.0_dp
      IF (helium%rho_present) THEN
         helium%rho_accu(:, :, :, :) = 0.0_dp
      END IF
      IF (helium%rdf_present) THEN
         helium%rdf_accu(:, :, :) = 0.0_dp
      END IF

      DO step = 1, helium%num_steps

         tot_steps = tot_steps+1
         CALL cp_iterate(helium%logger%iter_info, &
                         last=(step == helium%num_steps), &
                         iter_nr=tot_steps)
         !level_name="MD")
         helium%current_step = tot_steps

         CALL helium_step(helium, pint_env)

         ! call write_restart here to avoid interference with PINT write_restart
         CALL write_restart(root_section=helium%input, helium_env=helium)

         ! exit from the main loop if soft exit has been requested
         CALL external_control(should_stop, "MD", globenv=globenv)
         IF (should_stop) EXIT

      END DO

      ! remove iteration level
      CALL cp_rm_iter_level(helium%logger%iter_info, "MD")

      RETURN
   END SUBROUTINE helium_do_run

! ***************************************************************************
!> \brief  Sample the helium phase space
!> \param helium ...
!> \param pint_env ...
!> \date   2009-10-27
!> \author Lukasz Walewski
!> \note   Samples helium variable space according to multilevel Metropolis
!>         MC scheme, calculates the forces exerted by helium solvent on the
!>         solute and stores them in helium%force_avrg array. The forces are
!>         averaged over outer MC loop.
!> \note   The implicit assumption is that we simulate solute _with_ solvent
!>         most of the time, so for performance reasons I do not check that
!>         everywhere I should. This leads to some redundancy in the case of
!>         helium-only calculations.
! **************************************************************************************************
   SUBROUTINE helium_sample(helium, pint_env)

      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_sample', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msg_str
      INTEGER                                            :: irot, iweight, nslices, nsteps, &
                                                            sel_mp_source
      REAL(KIND=dp)                                      :: inv_num_pe, inv_xn, rnd, rtmp, rweight
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (helium%solute_present) helium%force_avrg(:, :) = 0.0_dp
      helium%energy_avrg(:) = 0.0_dp
      helium%plength_avrg(:) = 0.0_dp
      helium%num_accepted(:, :) = 0.0_dp

      ! helium parallelization: each processor gets different RN stream and
      ! runs independent helium simulation, the properties and forces are
      ! averaged over parallel helium environments once per step.

      inv_xn = 0.0_dp
      SELECT CASE (helium%sampling_method)
      CASE (helium_sampling_ceperley)
         ! helium sampling (outer MC loop)
         DO irot = 1, helium%iter_rot

            ! rotate helium beads in imaginary time at random, store current
            ! 'rotation state' in helium%relrot wich is within (0, helium%beads-1)
            ! (this is needed to sample different fragments of the permutation
            ! paths in try_permutations)
            rnd = next_random_number(helium%rng_stream_uniform)
            nslices = INT(rnd*helium%beads)
            CALL helium_rotate(helium, nslices)

            CALL helium_try_permutations(helium, pint_env)

            ! calculate & accumulate instantaneous properties for averaging
            IF (helium%solute_present) THEN
               IF (helium%get_helium_forces == helium_forces_average) THEN
                  CALL helium_solute_e_f(pint_env, helium, rtmp)
                  helium%force_avrg(:, :) = helium%force_avrg(:, :)+helium%force_inst(:, :)
               END IF
            END IF
            CALL helium_calc_energy(helium, pint_env)
            helium%energy_avrg(:) = helium%energy_avrg(:)+helium%energy_inst(:)
            CALL helium_calc_plength(helium)
            helium%plength_avrg(:) = helium%plength_avrg(:)+helium%plength_inst(:)

            ! instantaneous force output according to HELIUM%PRINT%FORCES_INST
            ! Warning: file I/O here may cost A LOT of cpu time!
            ! switched off here to save cpu
            !CALL helium_print_force_inst( helium, error )

         END DO ! outer MC loop

         ! If only the last forces are taken, calculate them for the last outer MC loop configuration
         IF ((helium%solute_present) .AND. (helium%get_helium_forces == helium_forces_last)) THEN
            CALL helium_solute_e_f(pint_env, helium, rtmp)
            helium%force_avrg(:, :) = helium%force_inst(:, :)
         END IF

         ! restore the original alignment of beads in imaginary time
         ! (this is useful to make a single bead's position easy to follow
         ! in the trajectory, otherwise it's index would change every step)
         CALL helium_rotate(helium, -helium%relrot)

         inv_xn = 1.0_dp/REAL(helium%iter_rot, dp)

      CASE DEFAULT
         WRITE (msg_str, *) helium%sampling_method
         msg_str = "Sampling method ("//TRIM(ADJUSTL(msg_str))//") not supported."
         CALL cp__b("motion/helium_sampling.F",224,msg_str)

      END SELECT

      ! actually average the properties over the outer MC loop
      IF ((helium%solute_present) .AND. (helium%get_helium_forces == helium_forces_average)) THEN
         helium%force_avrg(:, :) = helium%force_avrg(:, :)*inv_xn
      END IF
      helium%energy_avrg(:) = helium%energy_avrg(:)*inv_xn
      helium%plength_avrg(:) = helium%plength_avrg(:)*inv_xn

      ! instantaneous properties
      helium%proarea%inst(:) = helium_total_projected_area(helium)
      helium%prarea2%inst(:) = helium%proarea%inst(:)*helium%proarea%inst(:)
      helium%wnumber%inst(:) = helium_total_winding_number(helium)
      helium%wnmber2%inst(:) = helium%wnumber%inst(:)*helium%wnumber%inst(:)
      helium%mominer%inst(:) = helium_total_moment_of_inertia(helium)

      ! properties accumulated over the whole MC process
      helium%proarea%accu(:) = helium%proarea%accu(:)+helium%proarea%inst(:)
      helium%prarea2%accu(:) = helium%prarea2%accu(:)+helium%prarea2%inst(:)
      helium%wnmber2%accu(:) = helium%wnmber2%accu(:)+helium%wnmber2%inst(:)
      helium%mominer%accu(:) = helium%mominer%accu(:)+helium%mominer%inst(:)
      IF (helium%rho_present) THEN
         CALL helium_calc_rho(helium)
         helium%rho_accu(:, :, :, :) = helium%rho_accu(:, :, :, :)+helium%rho_inst(:, :, :, :)
      END IF
      IF (helium%rdf_present) THEN
         CALL helium_set_rdf_coord_system(helium, pint_env)
         CALL helium_calc_rdf(helium, helium%rdf_centers)
         helium%rdf_accu(:, :, :) = helium%rdf_accu(:, :, :)+helium%rdf_inst(:, :, :)
      END IF

      ! running averages (restart-aware)
      nsteps = helium%current_step-helium%first_step
      iweight = helium%averages_iweight
      rweight = REAL(iweight, dp)
      rtmp = 1.0_dp/(REAL(MAX(1, nsteps+iweight), dp))
      helium%proarea%ravr(:) = (helium%proarea%accu(:)+rweight*helium%proarea%rstr(:))*rtmp
      helium%prarea2%ravr(:) = (helium%prarea2%accu(:)+rweight*helium%prarea2%rstr(:))*rtmp
      helium%wnmber2%ravr(:) = (helium%wnmber2%accu(:)+rweight*helium%wnmber2%rstr(:))*rtmp
      helium%mominer%ravr(:) = (helium%mominer%accu(:)+rweight*helium%mominer%rstr(:))*rtmp

      ! average over helium environments sitting at different processors
!TODO the following involves message passing, maybe move it to i/o routines?
      inv_num_pe = 1.0_dp/REAL(logger%para_env%num_pe, dp)
      IF (helium%solute_present) THEN
         IF (helium%get_helium_forces == helium_forces_average) THEN
            CALL mp_sum(helium%force_avrg, logger%para_env%group)
            helium%force_avrg(:, :) = helium%force_avrg(:, :)*inv_num_pe
         ELSE IF (helium%get_helium_forces == helium_forces_last) THEN
            ! create random number from 0 to num pe on IO node
            IF (logger%para_env%ionode) THEN
               sel_mp_source = INT(next_random_number(helium%rng_stream_uniform) &
                                   *REAL(logger%para_env%num_pe, dp))
            END IF
            ! broadcast this number
            CALL mp_bcast(sel_mp_source, logger%para_env%source, &
                          logger%para_env%group)
            ! broadcast forces from this environment to all others
            CALL mp_bcast(helium%force_avrg, sel_mp_source, logger%para_env%group)
         END IF
      END IF

      CALL mp_sum(helium%energy_avrg, logger%para_env%group)
      CALL mp_sum(helium%plength_avrg, logger%para_env%group)
      CALL mp_sum(helium%num_accepted, logger%para_env%group)
      helium%energy_avrg(:) = helium%energy_avrg(:)*inv_num_pe
      helium%plength_avrg(:) = helium%plength_avrg(:)*inv_num_pe
      helium%num_accepted(:, :) = helium%num_accepted(:, :)*inv_num_pe

      RETURN
   END SUBROUTINE helium_sample

! ***************************************************************************
!> \brief  Perform MC step for helium
!> \param helium ...
!> \param pint_env ...
!> \date   2009-06-12
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_step(helium, pint_env)

      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_step', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msgstr, stmp, time_unit
      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: time_start, time_stop, time_used

      CALL timeset(routineN, handle)
      time_start = m_walltime()

      ! perform the actual phase space sampling
      CALL helium_sample(helium, pint_env)

      ! print properties
      CALL helium_print_energy(helium)
      WRITE (stmp, *) helium%apref
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%PROJECTED_AREA", &
                               helium%proarea%inst, &
                               angstrom*angstrom, &
                               (/"A_x", "A_y", "A_z"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-pa")
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%PROJECTED_AREA_2_AVG", &
                               helium%prarea2%ravr, &
                               angstrom*angstrom*angstrom*angstrom, &
                               (/"<A_x^2>", "<A_y^2>", "<A_z^2>"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-pa-avg", &
                               "REWIND", &
                               .TRUE.)
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%MOMENT_OF_INERTIA", &
                               helium%mominer%inst, &
                               angstrom*angstrom, &
                               (/"I_x/m", "I_y/m", "I_z/m"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-mi")
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%MOMENT_OF_INERTIA_AVG", &
                               helium%mominer%ravr, &
                               angstrom*angstrom, &
                               (/"<I_x/m>", "<I_y/m>", "<I_z/m>"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-mi-avg", &
                               "REWIND", &
                               .TRUE.)
      WRITE (stmp, *) helium%wpref
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%WINDING_NUMBER", &
                               helium%wnumber%inst, &
                               angstrom, &
                               (/"W_x", "W_y", "W_z"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-wn")
      CALL helium_print_vector(helium, &
                               "MOTION%PINT%HELIUM%PRINT%WINDING_NUMBER_2_AVG", &
                               helium%wnmber2%ravr, &
                               angstrom*angstrom, &
                               (/"<W_x^2>", "<W_y^2>", "<W_z^2>"/), &
                               "prefactor = "//TRIM(ADJUSTL(stmp))//" [Angstrom^-2]", &
                               "helium-wn-avg", &
                               "REWIND", &
                               .TRUE.)
      CALL helium_print_plength(helium)
      IF (helium%rdf_present) CALL helium_print_rdf(helium)
      IF (helium%rho_present) CALL helium_print_rho(helium)
      CALL helium_print_perm(helium)
      CALL helium_print_accepts(helium)
      CALL helium_print_coordinates(helium)
      CALL helium_print_force(helium)

      time_stop = m_walltime()
      time_used = time_stop-time_start
      time_unit = "sec"
      IF (time_used .GE. 60.0_dp) THEN
         time_used = time_used/60.0_dp
         time_unit = "min"
         IF (time_used .GE. 60.0_dp) THEN
            time_used = time_used/60.0_dp
            time_unit = "hours"
         END IF
      END IF
      msgstr = "MC step"
      stmp = ""
      WRITE (stmp, *) helium%current_step
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))//" of"
      stmp = ""
      WRITE (stmp, *) helium%last_step
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))//" in"
      stmp = ""
      WRITE (stmp, '(F20.1)') time_used
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(stmp))
      msgstr = TRIM(ADJUSTL(msgstr))//" "//TRIM(ADJUSTL(time_unit))//"."
      CALL helium_write_line(TRIM(msgstr))

      ! print out the total energy - for regtest evaluation
      stmp = ""
      WRITE (stmp, *) helium%energy_avrg(e_id_total)
      msgstr = "Total energy = "//TRIM(ADJUSTL(stmp))
      CALL helium_write_line(TRIM(msgstr))

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_step

! ***************************************************************************
!> \brief  ...
!> \param helium ...
!> \param  pint_env  - path integral environment
!> \par    History
!>        2010-06-17 added different distributions for m-sampling [lwalewski]
!>        2010-06-17 ratio for m-value added (m-sampling related) [lwalewski]
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_try_permutations(helium, pint_env)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_try_permutations', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: cyclen, ni
      LOGICAL                                            :: accepted, selected
      REAL(KIND=dp)                                      :: r, rnd, x, y, z

      IF (helium%maxcycle > 1) CALL helium_update_transition_matrix(helium)

      ! consecutive calls to helium_slice_metro_cyclic require that
      ! ALL(helium%pos.EQ.helium%work) - see a comment there,
      ! besides there is a magic regarding helium%permutation
      ! you have been warned
      helium%work(:, :, :) = helium%pos(:, :, :)

      ! the inner MC loop (without rotation in imaginary time)
      DO ni = 1, helium%iter_norot

         ! set the probability threshold for m_value: 1/(1+(m-1)/helium%m_ratio)
         r = 1.0_dp/(1.0_dp+(helium%maxcycle-1)/helium%m_ratio)

         ! draw permutation length for this trial from the distribution of choice
         !
         SELECT CASE (helium%m_dist_type)

         CASE (helium_mdist_singlev)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = 1
            ELSE
               cyclen = helium%m_value
            END IF

         CASE (helium_mdist_uniform)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = helium%m_value
            ELSE
               DO
                  x = next_random_number(helium%rng_stream_uniform)
                  cyclen = INT(helium%maxcycle*x)+1
                  IF (cyclen .NE. helium%m_value) EXIT
               END DO
            END IF

         CASE (helium_mdist_linear)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = helium%m_value
            ELSE
               DO
                  x = next_random_number(helium%rng_stream_uniform)
                  y = SQRT(2.0_dp*x)
                  cyclen = INT(helium%maxcycle*y/SQRT(2.0_dp))+1
                  IF (cyclen .NE. helium%m_value) EXIT
               END DO
            END IF

         CASE (helium_mdist_quadratic)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = helium%m_value
            ELSE
               DO
                  x = next_random_number(helium%rng_stream_uniform)
                  y = (3.0_dp*x)**(1.0_dp/3.0_dp)
                  z = 3.0_dp**(1.0_dp/3.0_dp)
                  cyclen = INT(helium%maxcycle*y/z)+1
                  IF (cyclen .NE. helium%m_value) EXIT
               END DO
            END IF

         CASE (helium_mdist_exponential)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = helium%m_value
            ELSE
               DO
                  DO
                     x = next_random_number(helium%rng_stream_uniform)
                     IF (x .GE. 0.01_dp) EXIT
                  END DO
                  z = -LOG(0.01_dp)
                  y = LOG(x)/z+1.0_dp; 
                  cyclen = INT(helium%maxcycle*y)+1
                  IF (cyclen .NE. helium%m_value) EXIT
               END DO
            END IF

         CASE (helium_mdist_gaussian)
            x = next_random_number(helium%rng_stream_uniform)
            IF (x .LT. r) THEN
               cyclen = 1
            ELSE
               DO
                  x = next_random_number(helium%rng_stream_gaussian)
                  cyclen = INT(x*0.75_dp+helium%m_value-0.5_dp)+1
                  IF (cyclen .NE. 1) EXIT
               END DO
            END IF

         CASE DEFAULT
            WRITE (stmp, *) helium%m_dist_type
            err_str = "M distribution type unknown ("//TRIM(ADJUSTL(stmp))//")"
            CALL cp__b("motion/helium_sampling.F",535,err_str)
            cyclen = 1

         END SELECT

         IF (cyclen < 1) cyclen = 1
         IF (cyclen > helium%maxcycle) cyclen = helium%maxcycle
         helium%num_accepted(1, cyclen) = helium%num_accepted(1, cyclen)+1

         ! check, if permutation of this length can be constructed
         IF (cyclen == 1) THEN
            rnd = next_random_number(helium%rng_stream_uniform)
            helium%ptable(1) = 1+INT(rnd*helium%atoms)
            helium%ptable(2) = -1
            helium%pweight = 0.0_dp
            selected = .TRUE.
         ELSE
            selected = helium_select_permutation(helium, cyclen)
         END IF

         IF (selected) THEN
            ! permutation was successfully selected - actually sample this permutation
            accepted = helium_slice_metro_cyclic(helium, pint_env, cyclen)
         ELSE
            accepted = .FALSE.
         END IF

!if (any(helium%pos .ne. helium%work)) then
!  print *, ""
!  print *, "pos and work are different!"
!  print *, ""
!end if

         IF (accepted) THEN
            ! positions changed
            IF (helium%solute_present) THEN
               ! use COM of the solute, but it did not move here so do nothing to save cpu
            ELSE
               IF (helium%periodic) THEN
                  ! use center of the cell, but it did not move here so do nothing to save cpu
               ELSE
                  ! update center of mass
                  helium%center(:) = helium_com(helium)
               END IF
            END IF
         END IF

      END DO

      RETURN
   END SUBROUTINE helium_try_permutations

! ***************************************************************************
!> \brief  Sample path variables for permutation of length <cyclen>
!> \param helium ...
!> \param pint_env ...
!> \param cyclen ...
!> \retval res ...
!> \author hforbert
! **************************************************************************************************
   FUNCTION helium_slice_metro_cyclic(helium, pint_env, cyclen) RESULT(res)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env
      INTEGER, INTENT(IN)                                :: cyclen
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_slice_metro_cyclic', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: c, ifix, j, k, l, level, pk1, pk2, stride
      INTEGER, DIMENSION(:), POINTER                     :: p, perm
      LOGICAL                                            :: nperiodic, should_reject
      REAL(KIND=dp)                                      :: cell_size, ds, dtk, e1, e2, pds, &
                                                            prev_ds, r, rtmp, sigma, x
      REAL(KIND=dp), DIMENSION(3)                        :: bis, rm1, rm2, tmp1, tmp2
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pos, work
      TYPE(spline_data_p_type), DIMENSION(:, :), POINTER :: uij

! trial permutation implicit in p
! since we (momentarily) only use cyclic permutations:
! cyclen = 1 : no permutation, sample p[0] anew
! cyclen = 2 : p[0] -> p[1] -> p[0]
! cyclen = 3 : p[0] -> p[1] -> p[2] -> p[0]
! cyclen = 4 : p[0] -> p[1] -> p[2] -> p[3] -> p[0]

      p => helium%ptable
      prev_ds = helium%pweight

      helium%num_accepted(2, cyclen) = helium%num_accepted(2, cyclen)+1
      level = 1
      res = .FALSE.

      ! nothing to be done
      IF (helium%bisection == 0) RETURN

      ! On entry helium_slice_metro_cyclic ASSUMES that work is a copy of pos
      ! and constructs the trial move there. If the move is accepted, the
      ! changed parts are copied to pos, but if it fails, only the CHANGED parts
      ! of work are overwritten by the corresponding parts of pos. So on exit work
      ! will AGAIN be a copy of pos (either way accept/reject). This is done here
      ! so we do not have to copy around the whole pos array in the caller, and
      ! the caller does not need to know which parts of work might have changed.
      pos => helium%pos
      work => helium%work
      perm => helium%permutation
      uij => helium%uij
      cell_size = (0.5_dp*helium%cell_size)**2
      nperiodic = .NOT. helium%periodic

      pds = prev_ds
      ifix = helium%beads-helium%bisection+1

      ! sanity checks
      !
      IF (ifix < 1) THEN
         WRITE (stmp, *) ifix
         err_str = "ifix<1 test failed (ifix="//TRIM(ADJUSTL(stmp))//")"
         CALL cp__b("motion/helium_sampling.F",653,err_str)
      END IF
      !
      j = 1
      k = helium%bisection
      DO
         IF (k < 2) EXIT
         j = j*2
         k = k/2
      END DO
      !
      IF (j /= helium%bisection) THEN
         WRITE (stmp, *) helium%bisection
         err_str = "helium%bisection not a power of 2 (helium%bisection="//TRIM(ADJUSTL(stmp))//")"
         CALL cp__b("motion/helium_sampling.F",667,err_str)
      END IF
      !
      IF (helium%bisection < 2) THEN
         WRITE (stmp, *) helium%bisection
         err_str = "helium%bisection less than 2 (helium%bisection="//TRIM(ADJUSTL(stmp))//")"
         CALL cp__b("motion/helium_sampling.F",673,err_str)
      END IF

      stride = helium%bisection
      DO
         IF (stride <= 2) EXIT
         ! calc new trial positions
         ! trial action: 0.5*stride*endpointapprox
         sigma = SQRT(0.5_dp*helium%hb2m*(stride/2)*helium%tau)
         dtk = 0.0_dp
         ds = 0.0_dp

         j = ifix+stride/2
         DO
            IF (j > helium%beads-stride/2) EXIT
            pk1 = j-stride/2
            pk2 = j+stride/2
            ! calculate log(T(s)):
            DO k = 1, cyclen
               CALL helium_boxmean_3d(helium, pos(:, p(k), pk1), pos(:, p(k), pk2), bis)
               tmp1(:) = bis(:)-pos(:, p(k), j)
               CALL helium_pbc(helium, tmp1)
               tmp1(:) = tmp1(:)/sigma
               dtk = dtk-0.5_dp*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
            END DO
            ! calculate log(T(sprime)) and sprime itself
            DO k = 1, cyclen
               CALL helium_boxmean_3d(helium, work(:, p(k), pk1), work(:, p(k), pk2), tmp1)
               DO c = 1, 3
                  x = next_random_number(rng_stream=helium%rng_stream_gaussian, variance=1.0_dp)
                  x = sigma*x
                  tmp1(c) = tmp1(c)+x
                  tmp2(c) = x
               END DO
               CALL helium_pbc(helium, tmp1)
               CALL helium_pbc(helium, tmp2)
               work(:, p(k), j) = tmp1(:)
               tmp2(:) = tmp2(:)/sigma
               dtk = dtk+0.5_dp*(tmp2(1)*tmp2(1)+tmp2(2)*tmp2(2)+tmp2(3)*tmp2(3))
            END DO
            j = j+stride
         END DO

         j = helium%beads-stride/2+1
         pk1 = j-stride/2
         DO k = 1, cyclen
            CALL helium_boxmean_3d(helium, pos(:, p(k), pk1), pos(:, perm(p(k)), 1), bis)
            tmp1(:) = bis(:)-pos(:, p(k), j)
            CALL helium_pbc(helium, tmp1)
            tmp1(:) = tmp1(:)/sigma
            dtk = dtk-0.5_dp*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
         END DO
         DO k = 1, cyclen
            CALL helium_boxmean_3d(helium, work(:, p(k), pk1), work(:, perm(p(1+MOD(k, cyclen))), 1), tmp1)
            DO c = 1, 3
               x = next_random_number(rng_stream=helium%rng_stream_gaussian, variance=1.0_dp)
               x = sigma*x
               tmp1(c) = tmp1(c)+x
               tmp2(c) = x
            END DO
            CALL helium_pbc(helium, tmp1)
            CALL helium_pbc(helium, tmp2)
            work(:, p(k), j) = tmp1(:)
            tmp2(:) = tmp2(:)/sigma
            dtk = dtk+0.5_dp*(tmp2(1)*tmp2(1)+tmp2(2)*tmp2(2)+tmp2(3)*tmp2(3))
         END DO
         ! ok we got the new positions
         ! calculate action_k(s)-action_k(sprime)
         x = 1.0_dp/(helium%tau*helium%hb2m*stride)
         j = ifix
         DO
            IF (j > helium%beads-stride/2) EXIT
            pk1 = j+stride/2
            DO k = 1, cyclen
               tmp1(:) = pos(:, p(k), j)-pos(:, p(k), pk1)
               CALL helium_pbc(helium, tmp1)
               ds = ds+x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
               tmp1(:) = work(:, p(k), j)-work(:, p(k), pk1)
               CALL helium_pbc(helium, tmp1)
               ds = ds-x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
               ! interaction change
               IF (helium%solute_present) THEN
                  CALL helium_bead_solute_e_f(pint_env, helium, p(k), pk1, energy=e1)
                  CALL helium_bead_solute_e_f(pint_env, helium, p(k), pk1, work(:, p(k), pk1), e2)
                  ds = ds+(stride/2)*(e1-e2)*helium%tau
               END IF
               DO l = 1, helium%atoms
                  IF (l /= p(k)) THEN
                     tmp1(:) = pos(:, p(k), pk1)-pos(:, l, pk1)
                     CALL helium_pbc(helium, tmp1)
                     r = tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3)
                     IF ((r < cell_size) .OR. nperiodic) THEN
                        r = SQRT(r)
                        ds = ds+REAL(stride/2, dp)*helium_spline(uij(1, 1)%spline_data, r)
                     END IF
                     tmp1(:) = work(:, p(k), pk1)-work(:, l, pk1)
                     CALL helium_pbc(helium, tmp1)
                     r = tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3)
                     IF ((r < cell_size) .OR. nperiodic) THEN
                        r = SQRT(r)
                        ds = ds-REAL(stride/2, dp)*helium_spline(uij(1, 1)%spline_data, r)
                     END IF
                  END IF
               END DO
               ! counted p[k], p[m] twice. subtract those again
               IF (k < cyclen) THEN
                  DO l = k+1, cyclen
                     tmp1(:) = pos(:, p(k), pk1)-pos(:, p(l), pk1)
                     CALL helium_pbc(helium, tmp1)
                     r = tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3)
                     IF ((r < cell_size) .OR. nperiodic) THEN
                        r = SQRT(r)
                        ds = ds-REAL(stride/2, dp)*helium_spline(uij(1, 1)%spline_data, r)
                     END IF
                     tmp1(:) = work(:, p(k), pk1)-work(:, p(l), pk1)
                     CALL helium_pbc(helium, tmp1)
                     r = tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3)
                     IF ((r < cell_size) .OR. nperiodic) THEN
                        r = SQRT(r)
                        ds = ds+REAL(stride/2, dp)*helium_spline(uij(1, 1)%spline_data, r)
                     END IF
                  END DO
               END IF
            END DO
            j = j+stride/2
         END DO
         ! last link
         pk1 = helium%beads-stride/2+1
         DO k = 1, cyclen
            tmp1(:) = pos(:, p(k), pk1)-pos(:, perm(p(k)), 1)
            CALL helium_pbc(helium, tmp1)
            ds = ds+x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
            tmp1(:) = work(:, p(k), pk1)-work(:, perm(p(1+MOD(k, cyclen))), 1)
            CALL helium_pbc(helium, tmp1)
            ds = ds-x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
         END DO
         ! ok now accept or reject:
         rtmp = next_random_number(helium%rng_stream_uniform)
!      IF ((dtk+ds-pds < 0.0_dp).AND.(EXP(dtk+ds-pds)<rtmp)) THEN
         IF (dtk+ds-pds < 0.0_dp) THEN
            IF (EXP(dtk+ds-pds) < rtmp) THEN
               DO c = ifix, helium%beads
                  DO k = 1, cyclen
                     work(:, p(k), c) = pos(:, p(k), c)
                  END DO
               END DO
               RETURN
            END IF
         END IF
         ! accepted. go on to the next level
         helium%num_accepted(level+2, cyclen) = helium%num_accepted(level+2, cyclen)+1
         level = level+1
         pds = ds
         stride = stride/2
      END DO
      ! we are on the lowest level now
      ! calc new trial positions
      ! trial action: endpointapprox for T, full action otherwise
      sigma = SQRT(0.5_dp*helium%hb2m*helium%tau)
      dtk = 0.0_dp
      ds = 0.0_dp
      DO j = ifix+1, helium%beads-1, 2
         pk1 = j-1
         pk2 = j+1
         ! calculate log(T(s)):
         DO k = 1, cyclen
            CALL helium_boxmean_3d(helium, pos(:, p(k), pk1), pos(:, p(k), pk2), bis)
            tmp1(:) = bis(:)-pos(:, p(k), j)
            CALL helium_pbc(helium, tmp1)
            tmp1(:) = tmp1(:)/sigma
            dtk = dtk-0.5_dp*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
         END DO
         ! calculate log(T(sprime)) and sprime itself
         DO k = 1, cyclen
            CALL helium_boxmean_3d(helium, work(:, p(k), pk1), work(:, p(k), pk2), tmp1)
            DO c = 1, 3
               x = next_random_number(rng_stream=helium%rng_stream_gaussian, variance=1.0_dp)
               x = sigma*x
               tmp1(c) = tmp1(c)+x
               tmp2(c) = x
            END DO
            CALL helium_pbc(helium, tmp1)
            CALL helium_pbc(helium, tmp2)
            work(:, p(k), j) = tmp1(:)
            tmp2(:) = tmp2(:)/sigma
            dtk = dtk+0.5_dp*(tmp2(1)*tmp2(1)+tmp2(2)*tmp2(2)+tmp2(3)*tmp2(3))
         END DO
      END DO
      j = helium%beads
      pk1 = j-1
      DO k = 1, cyclen
         CALL helium_boxmean_3d(helium, pos(:, p(k), pk1), pos(:, perm(p(k)), 1), bis)
         tmp1(:) = bis(:)-pos(:, p(k), j)
         CALL helium_pbc(helium, tmp1)
         tmp1(:) = tmp1(:)/sigma
         dtk = dtk-0.5_dp*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
      END DO
      DO k = 1, cyclen
         CALL helium_boxmean_3d(helium, work(:, p(k), pk1), work(:, perm(p(1+MOD(k, cyclen))), 1), tmp1)
         DO c = 1, 3
            x = next_random_number(rng_stream=helium%rng_stream_gaussian, variance=1.0_dp)
            x = sigma*x
            tmp1(c) = tmp1(c)+x
            tmp2(c) = x
         END DO
         CALL helium_pbc(helium, tmp1)
         CALL helium_pbc(helium, tmp2)
         work(:, p(k), j) = tmp1(:)
         tmp2 = tmp2/sigma
         dtk = dtk+0.5_dp*(tmp2(1)*tmp2(1)+tmp2(2)*tmp2(2)+tmp2(3)*tmp2(3))
      END DO
      ! ok we got the new positions.
      ! calculate action_k(s)-action_k(sprime)
      ! interaction change
!TODO interaction ONLY here? or some simplified 12-6 in the upper part?
      IF (helium%solute_present) THEN
         DO j = ifix, helium%beads
            DO k = 1, cyclen
               CALL helium_bead_solute_e_f(pint_env, helium, p(k), j, energy=e1)
               CALL helium_bead_solute_e_f(pint_env, helium, p(k), j, work(:, p(k), j), e2)
               ds = ds+(e1-e2)*helium%tau
            END DO
         END DO
      END IF
      x = 1.0_dp/(helium%tau*helium%hb2m*stride)
      DO j = ifix, helium%beads-1
         pk1 = j+1
         DO k = 1, cyclen
            tmp1(:) = pos(:, p(k), j)-pos(:, p(k), pk1)
            CALL helium_pbc(helium, tmp1)
            ds = ds+x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
            tmp1(:) = work(:, p(k), j)-work(:, p(k), pk1)
            CALL helium_pbc(helium, tmp1)
            ds = ds-x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
            DO l = 1, helium%atoms
               IF (l /= p(k)) THEN
                  rm1(:) = pos(:, p(k), j)-pos(:, l, j)
                  rm2(:) = pos(:, p(k), pk1)-pos(:, l, pk1)
                  ds = ds+helium_eval_expansion(helium, rm1, rm2, uij, 1)
                  rm1(:) = work(:, p(k), j)-work(:, l, j)
                  rm2(:) = work(:, p(k), pk1)-work(:, l, pk1)
                  ds = ds-helium_eval_expansion(helium, rm1, rm2, uij, 1)
               END IF
            END DO
            ! counted p[k], p[m] twice. subtract those again
            IF (k < cyclen) THEN
               DO l = k+1, cyclen
                  rm1(:) = pos(:, p(k), j)-pos(:, p(l), j)
                  rm2(:) = pos(:, p(k), pk1)-pos(:, p(l), pk1)
                  ds = ds-helium_eval_expansion(helium, rm1, rm2, uij, 1)
                  rm1(:) = work(:, p(k), j)-work(:, p(l), j)
                  rm2(:) = work(:, p(k), pk1)-work(:, p(l), pk1)
                  ds = ds+helium_eval_expansion(helium, rm1, rm2, uij, 1)
               END DO
            END IF
         END DO
      END DO
      j = helium%beads
      pk1 = 1
      DO k = 1, cyclen
         tmp1(:) = pos(:, p(k), j)-pos(:, perm(p(k)), 1)
         CALL helium_pbc(helium, tmp1)
         ds = ds+x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
         tmp1(:) = work(:, p(k), j)-work(:, perm(p(1+MOD(k, cyclen))), 1)
         CALL helium_pbc(helium, tmp1)
         ds = ds-x*(tmp1(1)*tmp1(1)+tmp1(2)*tmp1(2)+tmp1(3)*tmp1(3))
         DO l = 1, helium%atoms
            IF (l /= p(k)) THEN
               rm1(:) = pos(:, p(k), j)-pos(:, l, j)
               rm2(:) = pos(:, perm(p(k)), 1)-pos(:, perm(l), 1)
               ds = ds+helium_eval_expansion(helium, rm1, rm2, uij, 1)
            END IF
         END DO
         ! counted p[k], p[m] twice. subtract those again
         IF (k < cyclen) THEN
            DO l = k+1, cyclen
               rm1(:) = pos(:, p(k), j)-pos(:, p(l), j)
               rm2(:) = pos(:, perm(p(k)), pk1)-pos(:, perm(p(l)), pk1)
               ds = ds-helium_eval_expansion(helium, rm1, rm2, uij, 1)
            END DO
         END IF
      END DO
      IF (cyclen > 1) THEN
         !k,c,l
         c = perm(p(1))
         DO k = 1, cyclen-1
            perm(p(k)) = perm(p(k+1))
         END DO
         perm(p(cyclen)) = c
      END IF
      DO k = 1, cyclen
         DO l = 1, helium%atoms
            IF (l /= p(k)) THEN
               rm1(:) = work(:, p(k), j)-work(:, l, j)
               rm2(:) = work(:, perm(p(k)), 1)-work(:, perm(l), 1)
               ds = ds-helium_eval_expansion(helium, rm1, rm2, uij, 1)
            END IF
         END DO
         ! counted p[k], p[m] twice. subtract those again
         IF (k < cyclen) THEN
            DO l = k+1, cyclen
               rm1(:) = work(:, p(k), j)-work(:, p(l), j)
               rm2(:) = work(:, perm(p(k)), 1)-work(:, perm(p(l)), 1)
               ds = ds+helium_eval_expansion(helium, rm1, rm2, uij, 1)
            END DO
         END IF
      END DO
      ! ok now accept or reject:
      rtmp = next_random_number(helium%rng_stream_uniform)
!   IF ((dtk+ds-pds<0.0_dp).AND.(EXP(dtk+ds-pds)<rtmp)) THEN
      IF (dtk+ds-pds < 0.0_dp) THEN
         IF (EXP(dtk+ds-pds) < rtmp) THEN
            DO c = ifix, helium%beads
               DO k = 1, cyclen
                  work(:, p(k), c) = pos(:, p(k), c)
               END DO
            END DO
            IF (cyclen > 1) THEN
               c = perm(p(cyclen))
               DO k = cyclen-1, 1, -1
                  perm(p(k+1)) = perm(p(k))
               END DO
               perm(p(1)) = c
            END IF
            RETURN
         END IF
      END IF
      ! accepted.

      ! rejection of the whole move if any of the new positions is farther away
      ! from the current center of gravity than HELIUM%DROPLET_RADIUS [lwalewski]
      IF (helium%droplet_radius .LT. HUGE(1.0_dp)) THEN
         ! cycle through new positions
         should_reject = .FALSE.
         outer: DO c = ifix, helium%beads
            DO k = 1, cyclen
               bis(:) = work(:, p(k), c)-helium%center(:)
               rtmp = SQRT(bis(1)*bis(1)+bis(2)*bis(2)+bis(3)*bis(3))
               IF (rtmp .GE. helium%droplet_radius) THEN
                  ! found - this one does not fit within R from the center
                  should_reject = .TRUE.
                  EXIT outer
               END IF
            END DO
         END DO outer
         IF (should_reject) THEN
            ! restore work and perm, then return
            DO c = ifix, helium%beads
               DO k = 1, cyclen
                  work(:, p(k), c) = pos(:, p(k), c)
               END DO
            END DO
            IF (cyclen > 1) THEN
               c = perm(p(cyclen))
               DO k = cyclen-1, 1, -1
                  perm(p(k+1)) = perm(p(k))
               END DO
               perm(p(1)) = c
            END IF
            RETURN
         END IF
         ! Note: after CoM position gets updated according to the new positions
         ! some beads may end up outside of the DROPLET_RADIUS range from the new
         ! CoM, they should not be very far off, though.
      END IF
      ! accepted.

      ! copy trial over to the real thing
      DO c = ifix, helium%beads
         DO k = 1, cyclen
            pos(:, p(k), c) = work(:, p(k), c)
         END DO
      END DO
      DO k = 1, cyclen
         helium%iperm(perm(p(k))) = p(k)
      END DO
      helium%num_accepted(level+2, cyclen) = helium%num_accepted(level+2, cyclen)+1
      res = .TRUE.

      RETURN
   END FUNCTION helium_slice_metro_cyclic

! ***************************************************************************
!> \brief  ...
!> \param helium ...
!> \param len ...
!> \retval res ...
!> \author hforbert
! **************************************************************************************************
   FUNCTION helium_select_permutation(helium, len) RESULT(res)
      TYPE(helium_solvent_type), POINTER                 :: helium
      INTEGER, INTENT(IN)                                :: len
      LOGICAL                                            :: res

      INTEGER                                            :: i, j, k, n
      INTEGER, DIMENSION(:), POINTER                     :: iperm, p, perm
      INTEGER, DIMENSION(:, :), POINTER                  :: nmatrix
      REAL(KIND=dp)                                      :: rnd, s1, s2, t
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ipmatrix, pmatrix, tmatrix

      s1 = 0.0_dp
      s2 = 0.0_dp
      res = .FALSE.
      n = helium%atoms
      tmatrix => helium%tmatrix
      pmatrix => helium%pmatrix
      ipmatrix => helium%ipmatrix
      perm => helium%permutation
      iperm => helium%iperm
      p => helium%ptable
      nmatrix => helium%nmatrix

      p(len+1) = -1
      rnd = next_random_number(helium%rng_stream_uniform)
      p(1) = INT(n*rnd)+1
      DO k = 1, len-1
         t = next_random_number(helium%rng_stream_uniform)
         ! find the corresponding path to connect to
         ! using the precalculated optimal decision tree:
         i = n-1
         DO
            IF (tmatrix(p(k), i) > t) THEN
               i = nmatrix(p(k), 2*i-1)
            ELSE
               i = nmatrix(p(k), 2*i)
            END IF
            IF (i < 0) EXIT
         END DO
         i = -i
         ! which particle was it previously connected to?
         p(k+1) = iperm(i)
         ! is it unique? quit if it was already part of the permutation
         DO j = 1, k
            IF (p(j) == p(k+1)) RETURN
         END DO
         ! acummulate the needed values for the final
         ! accept/reject step:
         s1 = s1+ipmatrix(p(k), i)
         s2 = s2+ipmatrix(p(k), perm(p(k)))
      END DO
      ! close the permutation loop:
      s1 = s1+ipmatrix(p(len), perm(p(1)))
      s2 = s2+ipmatrix(p(len), perm(p(len)))
      ! final accept/reject:
      rnd = next_random_number(helium%rng_stream_uniform)
      t = s1*rnd
      IF (t > s2) RETURN
      ! ok, we have accepted the permutation
      ! calculate the action bias for the subsequent resampling
      ! of the paths:
      s1 = pmatrix(p(len), perm(p(1)))-pmatrix(p(len), perm(p(len)))
      DO k = 1, len-1
         s1 = s1+pmatrix(p(k), perm(p(k+1)))-pmatrix(p(k), perm(p(k)))
      END DO
      helium%pweight = s1
      res = .TRUE.
      RETURN
   END FUNCTION helium_select_permutation

END MODULE helium_sampling
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods dealing with helium_solvent_type
!> \author Lukasz Walewski
!> \date   2009-06-10
! **************************************************************************************************
MODULE helium_methods

   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE bibliography,                    ONLY: Walewski2014,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_printkey_is_on
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get
   USE helium_common,                   ONLY: helium_com,&
                                              helium_pbc
   USE helium_io,                       ONLY: helium_read_xyz,&
                                              helium_write_line,&
                                              helium_write_setup
   USE helium_sampling,                 ONLY: helium_sample
   USE helium_types,                    ONLY: &
        he_mass, helium_destroy_int_arr_ptr, helium_solvent_type, hid_carbon, hid_chlorine, &
        hid_hydrogen, hid_num, hid_oxygen, rho_atom_number, rho_moment_of_inertia, rho_num, &
        rho_projected_area, rho_winding_cycle, rho_winding_number
   USE input_constants,                 ONLY: helium_cell_shape_cube,&
                                              helium_cell_shape_octahedron,&
                                              helium_solute_intpot_none
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              max_line_length
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_bcast
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type,&
                                              set_rng_stream
   USE particle_list_types,             ONLY: particle_list_type
   USE physcon,                         ONLY: a_mass,&
                                              angstrom,&
                                              boltzmann,&
                                              h_bar,&
                                              kelvin,&
                                              massunit
   USE pint_public,                     ONLY: pint_com_pos
   USE pint_types,                      ONLY: pint_env_type
   USE splines_methods,                 ONLY: init_spline,&
                                              init_splinexy
   USE splines_types,                   ONLY: spline_data_create,&
                                              spline_data_release,&
                                              spline_data_retain
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'helium_methods'
   INTEGER, SAVE, PRIVATE :: last_helium_id = 0

   PUBLIC :: helium_create
   PUBLIC :: helium_init
   PUBLIC :: helium_release

CONTAINS

! ***************************************************************************
!> \brief  Data-structure that holds all needed information about
!>         (superfluid) helium solvent
!> \param helium ...
!> \param input ...
!> \param solute ...
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_create(helium, input, solute)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(pint_env_type), OPTIONAL, POINTER             :: solute

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_create', routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: msg_str, potential_file_name
      CHARACTER(len=default_string_length)               :: stmp
      INTEGER                                            :: handle, i, input_unit, isize, itmp, j, &
                                                            nlines, ntab
      LOGICAL                                            :: expl_cell, expl_dens, expl_nats, &
                                                            explicit, ltmp
      REAL(KIND=dp)                                      :: cgeof, dx, mHe, rtmp, T, tcheck, x1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pot_transfer
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: helium_section

      CALL timeset(routineN, handle)

      CALL cite_reference(Walewski2014)
      NULLIFY (helium_section)
      helium_section => section_vals_get_subs_vals(input, &
                                                   "MOTION%PINT%HELIUM")
      CALL section_vals_get(helium_section, explicit=explicit)
      ALLOCATE (helium)
      NULLIFY (helium%input, &
               helium%ptable, helium%permutation, &
               helium%iperm, &
               helium%itmp_atoms_1d, &
               helium%ltmp_atoms_1d, &
               helium%itmp_atoms_np_1d, &
               helium%pos, helium%work, &
               helium%force_avrg, &
               helium%force_inst, &
               helium%rtmp_3_np_1d, &
               helium%rtmp_p_ndim_1d, &
               helium%rtmp_p_ndim_np_1d, &
               helium%rtmp_3_atoms_beads_1d, &
               helium%rtmp_3_atoms_beads_np_1d, &
               helium%rtmp_p_ndim_2d, &
               helium%ltmp_3_atoms_beads_3d, &
               helium%tmatrix, helium%pmatrix, &
               helium%nmatrix, helium%ipmatrix, &
               helium%uij, helium%eij, &
               helium%rdf_inst, &
               helium%plength_avrg, &
               helium%plength_inst, &
               helium%atom_plength, &
               helium%ename, &
               helium%eid &
               )

      helium%ref_count = 1
      last_helium_id = last_helium_id+1
      helium%id_nr = last_helium_id
      helium%input => input
      helium%accepts = 0
      helium%relrot = 0

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! check if solute is present in our simulation
      helium%solute_present = .FALSE.
      helium%solute_atoms = 0
      helium%solute_beads = 0
      CALL section_vals_val_get( &
         helium_section, &
         "HELIUM_ONLY", &
         l_val=ltmp)
      IF (.NOT. ltmp) THEN
         IF (PRESENT(solute)) THEN
            IF (ASSOCIATED(solute)) THEN
               helium%solute_present = .TRUE.
               helium%solute_atoms = solute%ndim/3
               helium%solute_beads = solute%p
            END IF
         END IF
      END IF

      ! get number of environments in the restart file (if present)
      CALL section_vals_val_get(helium_section, "NUM_ENV", &
                                explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(helium_section, "NUM_ENV", &
                                   i_val=itmp)
         helium%num_env_restart = itmp
      ELSE
         helium%num_env_restart = -1
      END IF

      ! set current number of environments
      helium%num_env = logger%para_env%num_pe
      CALL section_vals_val_set(helium%input, &
                                "MOTION%PINT%HELIUM%NUM_ENV", &
                                i_val=helium%num_env)

      ! if the restart file contains more environments than the runtime
      ! the additional environments (possibly with valuable permutation state)
      ! will be lost, we won't proceed in this case by default, the user might
      ! enforce this explicitly though
      IF (helium%num_env_restart .GT. helium%num_env) THEN
         stmp = ""
         WRITE (stmp, *) helium%num_env
         msg_str = "Number of He environments in the runtime ("// &
                   TRIM(ADJUSTL(stmp))//") smaller than in the restart ("
         stmp = ""
         WRITE (stmp, *) helium%num_env_restart
         msg_str = TRIM(ADJUSTL(msg_str))//TRIM(ADJUSTL(stmp))//")!"
         CALL helium_write_line(msg_str)
         CALL section_vals_val_get(helium_section, "DROP_UNUSED_ENVS", &
                                   l_val=ltmp)
         IF (ltmp) THEN
            msg_str = "DROP_UNUSED_ENVS set - proceeding anyways."
            CALL cp__w("motion/helium_methods.F",214,msg_str)
         ELSE
            CALL cp__b("motion/helium_methods.F",216,msg_str)
         END IF
      END IF

      CALL section_vals_val_get(helium_section, "NBEADS", &
                                i_val=helium%beads)
      CALL section_vals_val_get(helium_section, "INOROT", &
                                i_val=helium%iter_norot)
      CALL section_vals_val_get(helium_section, "IROT", &
                                i_val=helium%iter_rot)

      ! get number of steps and current step number from PINT
      CALL section_vals_val_get(input, "MOTION%PINT%ITERATION", &
                                i_val=itmp)
      helium%first_step = itmp
      CALL section_vals_val_get(input, "MOTION%PINT%MAX_STEP", &
                                explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(input, "MOTION%PINT%MAX_STEP", &
                                   i_val=itmp)
         helium%last_step = itmp
         helium%num_steps = helium%last_step-helium%first_step
      ELSE
         CALL section_vals_val_get(input, "MOTION%PINT%NUM_STEPS", &
                                   i_val=itmp)
         helium%num_steps = itmp
         helium%last_step = helium%first_step+helium%num_steps
      END IF

      ! boundary conditions
      CALL section_vals_val_get(helium_section, "PERIODIC", &
                                l_val=helium%periodic)
      CALL section_vals_val_get(helium_section, "CELL_SHAPE", &
                                i_val=helium%cell_shape)

      CALL section_vals_val_get(helium_section, "DROPLET_RADIUS", &
                                r_val=helium%droplet_radius)

      ! Set density Rho, number of atoms N and volume V ( Rho = N / V ).
      ! Allow only 2 out of 3 values to be defined at the same time, calculate
      ! the third.
      ! Note, that DENSITY and NATOMS keywords have default values, while
      ! CELL_SIZE does not. Thus if CELL_SIZE is given explicitly then one and
      ! only one of the two remaining options must be give explicitly as well.
      ! If CELL_SIZE is not given explicitly then all four combinations of the
      ! two other options are valid.
      CALL section_vals_val_get(helium_section, "DENSITY", &
                                explicit=expl_dens, r_val=helium%density)
      CALL section_vals_val_get(helium_section, "NATOMS", &
                                explicit=expl_nats, i_val=helium%atoms)
      CALL section_vals_val_get(helium_section, "CELL_SIZE", &
                                explicit=expl_cell)
      cgeof = 1.0_dp
      IF (helium%periodic) THEN
         IF (helium%cell_shape .EQ. helium_cell_shape_octahedron) cgeof = 2.0_dp
      END IF
      rtmp = (cgeof*helium%atoms/helium%density)**(1.0_dp/3.0_dp)
      IF (.NOT. expl_cell) THEN
         helium%cell_size = rtmp
      ELSE
         CALL section_vals_val_get(helium_section, "CELL_SIZE", &
                                   r_val=helium%cell_size)
         ! only more work if not all three values are consistent:
         IF (ABS(helium%cell_size-rtmp) .GT. 100.0_dp*EPSILON(0.0_dp)* &
             (ABS(helium%cell_size)+rtmp)) THEN
            IF (expl_dens .AND. expl_nats) THEN
               msg_str = "DENSITY, NATOMS and CELL_SIZE options "// &
                         "contradict each other"
               CALL cp__w("motion/helium_methods.F",284,msg_str)
            END IF
            !ok we have enough freedom to resolve the conflict:
            IF (.NOT. expl_dens) THEN
               helium%density = cgeof*helium%atoms/helium%cell_size**3.0_dp
               IF (.NOT. expl_nats) THEN
                  msg_str = "CELL_SIZE defined but neither "// &
                            "NATOMS nor DENSITY given, using default NATOMS."
                  CALL cp__w("motion/helium_methods.F",292,msg_str)
               END IF
            ELSE ! ( expl_dens .AND. .NOT. expl_nats )
               ! calculate the nearest number of atoms for given conditions
               helium%atoms = NINT(helium%density* &
                                   helium%cell_size**3.0_dp/cgeof)
               ! adjust cell size to maintain correct density
               ! (should be a small correction)
               rtmp = (cgeof*helium%atoms/helium%density &
                       )**(1.0_dp/3.0_dp)
               IF (ABS(helium%cell_size-rtmp) .GT. 100.0_dp*EPSILON(0.0_dp) &
                   *(ABS(helium%cell_size)+rtmp)) THEN
                  msg_str = "Adjusting actual cell size "// &
                            "to maintain correct density."
                  CALL cp__w("motion/helium_methods.F",306,msg_str)
                  helium%cell_size = rtmp
               END IF
            END IF
         END IF
      END IF
      helium%cell_size_inv = 1.0_dp/helium%cell_size
      ! From now on helium%density, helium%atoms and helium%cell_size are
      ! correctly defined.

      ! set the M matrix for winding number calculations
      SELECT CASE (helium%cell_shape)

      CASE (helium_cell_shape_octahedron)
         helium%cell_m(1, 1) = helium%cell_size; 
         helium%cell_m(2, 1) = 0.0_dp; 
         helium%cell_m(3, 1) = 0.0_dp; 
         helium%cell_m(1, 2) = 0.0_dp; 
         helium%cell_m(2, 2) = helium%cell_size; 
         helium%cell_m(3, 2) = 0.0_dp; 
         helium%cell_m(1, 3) = helium%cell_size/2.0_dp; 
         helium%cell_m(2, 3) = helium%cell_size/2.0_dp; 
         helium%cell_m(3, 3) = helium%cell_size/2.0_dp; 
         helium%cell_m_inv(1, 1) = 1.0_dp/helium%cell_size; 
         helium%cell_m_inv(2, 1) = 0.0_dp; 
         helium%cell_m_inv(3, 1) = 0.0_dp; 
         helium%cell_m_inv(1, 2) = 0.0_dp; 
         helium%cell_m_inv(2, 2) = 1.0_dp/helium%cell_size; 
         helium%cell_m_inv(3, 2) = 0.0_dp; 
         helium%cell_m_inv(1, 3) = -1.0_dp/helium%cell_size; 
         helium%cell_m_inv(2, 3) = -1.0_dp/helium%cell_size; 
         helium%cell_m_inv(3, 3) = 2.0_dp/helium%cell_size; 
      CASE (helium_cell_shape_cube)

         helium%cell_m(1, 1) = helium%cell_size; 
         helium%cell_m(2, 1) = 0.0_dp; 
         helium%cell_m(3, 1) = 0.0_dp; 
         helium%cell_m(1, 2) = 0.0_dp; 
         helium%cell_m(2, 2) = helium%cell_size; 
         helium%cell_m(3, 2) = 0.0_dp; 
         helium%cell_m(1, 3) = 0.0_dp; 
         helium%cell_m(2, 3) = 0.0_dp; 
         helium%cell_m(3, 3) = helium%cell_size; 
         helium%cell_m_inv(1, 1) = 1.0_dp/helium%cell_size; 
         helium%cell_m_inv(2, 1) = 0.0_dp; 
         helium%cell_m_inv(3, 1) = 0.0_dp; 
         helium%cell_m_inv(1, 2) = 0.0_dp; 
         helium%cell_m_inv(2, 2) = 1.0_dp/helium%cell_size; 
         helium%cell_m_inv(3, 2) = 0.0_dp; 
         helium%cell_m_inv(1, 3) = 0.0_dp; 
         helium%cell_m_inv(2, 3) = 0.0_dp; 
         helium%cell_m_inv(3, 3) = 1.0_dp/helium%cell_size; 
      CASE DEFAULT
         helium%cell_m(:, :) = 0.0_dp
         helium%cell_m_inv(:, :) = 0.0_dp

      END SELECT

      ! check value of maxcycle
      CALL section_vals_val_get(helium_section, "MAX_PERM_CYCLE", &
                                i_val=helium%maxcycle)
      i = helium%maxcycle
      i = helium%atoms-helium%maxcycle

      ! set m-distribution parameters
      CALL section_vals_val_get(helium_section, "M-SAMPLING%DISTRIBUTION-TYPE", &
                                i_val=i)
      helium%m_dist_type = i
      CALL section_vals_val_get(helium_section, "M-SAMPLING%M-VALUE", &
                                i_val=i)
      helium%m_value = i
      CALL section_vals_val_get(helium_section, "M-SAMPLING%M-RATIO", &
                                r_val=rtmp)
      helium%m_ratio = rtmp

      CALL section_vals_val_get(helium_section, "BISECTION", &
                                i_val=helium%bisection)
      ! precheck bisection value (not all invalids are filtered out here yet)
      i = helium%bisection
      i = helium%beads-helium%bisection
      !
      itmp = helium%bisection
      rtmp = 2.0_dp**(ANINT(LOG(REAL(itmp, dp))/LOG(2.0_dp)))
      tcheck = ABS(REAL(itmp, KIND=dp)-rtmp)
      IF (tcheck > 100.0_dp*EPSILON(0.0_dp)) THEN
         msg_str = "BISECTION should be integer power of 2."
         CALL cp__b("motion/helium_methods.F",392,msg_str)
      END IF
      helium%bisctlog2 = NINT(LOG(REAL(itmp, dp))/LOG(2.0_dp))

      CALL section_vals_val_get(helium_section, "SAMPLING_METHOD", &
                                i_val=helium%sampling_method)

      ! hard coded He4 directly (mass so i get my original hb2m value)
      ! He4 mass defined as a constant in helium_types.F now [lwalewski]
      helium%hb2m = 1.0_dp/(he_mass*massunit)

      helium%pweight = 0.0_dp

      IF (logger%para_env%ionode) THEN
         CALL section_vals_val_get(helium_section, "POTENTIAL_FILE_NAME", &
                                   c_val=potential_file_name)
         CALL open_file(file_name=TRIM(potential_file_name), &
                        file_action="READ", file_status="OLD", unit_number=input_unit)
         READ (input_unit, *) nlines, helium%pdx, helium%tau, &
            x1, dx
         helium%tau = kelvin/helium%tau
         x1 = x1/angstrom
         dx = dx/angstrom
      END IF
      CALL mp_bcast(nlines, logger%para_env%source, &
                    logger%para_env%group)
      CALL mp_bcast(helium%pdx, logger%para_env%source, &
                    logger%para_env%group)
      CALL mp_bcast(helium%tau, logger%para_env%source, &
                    logger%para_env%group)
      CALL mp_bcast(x1, logger%para_env%source, &
                    logger%para_env%group)
      CALL mp_bcast(dx, logger%para_env%source, &
                    logger%para_env%group)

      ! boltzmann : Boltzmann constant [J/K]
      ! h_bar     : Planck constant [J*s]
      ! J = kg*m^2/s^2
      ! 4He mass in [kg]
      mHe = he_mass*a_mass
      ! physical temperature [K]
      T = kelvin/helium%tau/helium%beads
      ! prefactors for calculating superfluid fractions [Angstrom^-2]
      helium%wpref = (((1e-20/h_bar)*mHe)/h_bar)*boltzmann*T
      helium%apref = (((4e-20/h_bar)*mHe)/h_bar)*boltzmann*T

      isize = helium%pdx+1
      ALLOCATE (helium%uij(isize, isize))
      ALLOCATE (helium%eij(isize, isize))
      DO i = 1, isize
         DO j = 1, i
            CALL spline_data_create(helium%uij(i, j)%spline_data)
            CALL init_splinexy(helium%uij(i, j)%spline_data, nlines)
            helium%uij(i, j)%spline_data%x1 = x1
            CALL spline_data_create(helium%eij(i, j)%spline_data)
            CALL init_splinexy(helium%eij(i, j)%spline_data, nlines)
            helium%eij(i, j)%spline_data%x1 = x1
         END DO
      END DO
      DO i = 1, isize-1
         DO j = i+1, isize
            helium%uij(i, j) = helium%uij(j, i)
            CALL spline_data_retain(helium%uij(i, j)%spline_data)
            helium%eij(i, j) = helium%eij(j, i)
            CALL spline_data_retain(helium%eij(i, j)%spline_data)
         END DO
      END DO

      isize = (helium%pdx+1)*(helium%pdx+2)
      ALLOCATE (pot_transfer(nlines, isize))
      IF (logger%para_env%ionode) THEN
         DO i = 1, nlines
            READ (input_unit, *) pot_transfer(i, :)
         END DO
         CALL close_file(unit_number=input_unit)
      END IF
      CALL mp_bcast(pot_transfer, logger%para_env%source, &
                    logger%para_env%group)
      isize = helium%pdx+1
      ntab = 1
      DO i = 1, isize
         DO j = 1, i
            helium%uij(i, j)%spline_data%y(:) = pot_transfer(:, ntab)* &
                                                angstrom**(2*i-2)
            CALL init_spline(helium%uij(i, j)%spline_data, dx=dx)
            ntab = ntab+1
         END DO
      END DO
      DO i = 1, isize
         DO j = 1, i
            helium%eij(i, j)%spline_data%y(:) = pot_transfer(:, ntab)* &
                                                angstrom**(2*i-2)/kelvin
            CALL init_spline(helium%eij(i, j)%spline_data, dx=dx)
            ntab = ntab+1
         END DO
      END DO
      DEALLOCATE (pot_transfer)

      ! ALLOCATE helium-related arrays
      i = helium%atoms
      j = helium%beads
      ALLOCATE (helium%pos(3, i, j+1))
      helium%pos = 0.0_dp
      ALLOCATE (helium%work(3, i, j+1))
      ALLOCATE (helium%ptable(helium%maxcycle+1))
      ALLOCATE (helium%permutation(i))
      ALLOCATE (helium%iperm(i))
      ALLOCATE (helium%tmatrix(i, i))
      ALLOCATE (helium%nmatrix(i, 2*i))
      ALLOCATE (helium%pmatrix(i, i))
      ALLOCATE (helium%ipmatrix(i, i))
      itmp = helium%bisctlog2+2
      ALLOCATE (helium%num_accepted(itmp, helium%maxcycle))
      ALLOCATE (helium%plength_avrg(helium%atoms))
      ALLOCATE (helium%plength_inst(helium%atoms))
      ALLOCATE (helium%atom_plength(helium%atoms))

      ! check whether rdfs should be calculated and printed
      helium%rdf_present = helium_property_active(helium, "RDF")
      IF (helium%rdf_present) THEN
         ! allocate & initialize rdf related data structures
         CALL helium_rdf_init(helium)
      END IF

      ! check whether densities should be calculated and printed
      helium%rho_present = helium_property_active(helium, "RHO")
      IF (helium%rho_present) THEN
         ! allocate & initialize density related data structures
         NULLIFY (helium%rho_property)
         CALL helium_rho_init(helium)
      END IF

      ! restore averages calculated in previous runs
      CALL helium_averages_restore(helium)

      ! RNG state create & init
      CALL helium_rng_init(helium)

      ! fill in the solute-related data structures
      helium%e_corr = 0.0_dp
      IF (helium%solute_present) THEN
         helium%bead_ratio = helium%beads/helium%solute_beads

         ! check if bead numbers are commensurate:
         i = helium%bead_ratio*helium%solute_beads-helium%beads
!TODO Adjust helium bead number if not comm. and if coords not given expl.

         ! check if tau, temperature and bead number are consistent:
         tcheck = ABS((helium%tau*helium%beads-solute%beta)/solute%beta)
         IF (tcheck > 1.0e-14_dp) THEN
            msg_str = "Tau, temperature and bead number are inconsistent."
            CALL cp__b("motion/helium_methods.F",543,msg_str)
         END IF

         CALL helium_set_solute_indices(helium, solute)
         CALL helium_set_solute_cell(helium, solute)

         ! set the interaction potential type
         CALL section_vals_val_get(helium_section, "SOLUTE_INTERACTION", &
                                   i_val=helium%solute_interaction)
         IF (helium%solute_interaction .EQ. helium_solute_intpot_none) THEN
            WRITE (msg_str, '(A,I0,A)') &
               "Solute found but no helium-solute interaction selected "// &
               "(see SOLUTE_INTERACTION keyword)"
            CALL cp__b("motion/helium_methods.F",556,msg_str)
         END IF

         ! ALLOCATE solute-related arrays
         ALLOCATE (helium%force_avrg(helium%solute_beads, &
                                     helium%solute_atoms*3))
         ALLOCATE (helium%force_inst(helium%solute_beads, &
                                     helium%solute_atoms*3))

         ALLOCATE (helium%rtmp_p_ndim_1d(solute%p*solute%ndim))
         ALLOCATE (helium%rtmp_p_ndim_np_1d(solute%p*solute%ndim*helium%num_env))
         ALLOCATE (helium%rtmp_p_ndim_2d(solute%p, solute%ndim))

      ELSE
         helium%bead_ratio = 0
         IF (helium%periodic) THEN
            ! this assumes a specific potential (and its ugly):
            x1 = angstrom*0.5_dp*helium%cell_size
            ! 10.8 is in Kelvin, x1 needs to be in Angstrom,
            ! since 2.9673 is in Angstrom
            helium%e_corr = (twopi*helium%density/angstrom**3*10.8_dp* &
                             (544850.4_dp*EXP(-13.353384_dp*x1/2.9673_dp)*(2.9673_dp/13.353384_dp)**3* &
                              (2.0_dp+2.0_dp*13.353384_dp*x1/2.9673_dp+(13.353384_dp*x1/2.9673_dp)**2)- &
                              (((0.1781_dp/7.0_dp*(2.9673_dp/x1)**2+0.4253785_dp/5.0_dp)*(2.9673_dp/x1)**2+ &
                                1.3732412_dp/3.0_dp)*(2.9673_dp/x1)**3)*2.9673_dp**3))/kelvin
         END IF
      END IF

      ! ALLOCATE temporary arrays
      ALLOCATE (helium%itmp_atoms_1d(helium%atoms))
      ALLOCATE (helium%ltmp_atoms_1d(helium%atoms))
      ALLOCATE (helium%itmp_atoms_np_1d(helium%atoms*helium%num_env))
      ALLOCATE (helium%rtmp_3_np_1d(3*helium%num_env))
      ALLOCATE (helium%rtmp_3_atoms_beads_1d(3*helium%atoms* &
                                             helium%beads))
      ALLOCATE (helium%rtmp_3_atoms_beads_np_1d(3*helium%atoms* &
                                                helium%beads*helium%num_env))
      ALLOCATE (helium%ltmp_3_atoms_beads_3d(3, helium%atoms, &
                                             helium%beads))

      CALL helium_write_setup(helium)

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_create

! ***************************************************************************
!> \brief  Releases helium_solvent_type
!> \param helium ...
!> \author hforbert
! **************************************************************************************************
   SUBROUTINE helium_release(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(helium)) THEN
         helium%ref_count = helium%ref_count-1
         IF (helium%ref_count < 1) THEN

            ! DEALLOCATE temporary arrays
            DEALLOCATE ( &
               helium%ltmp_3_atoms_beads_3d, &
               helium%rtmp_3_atoms_beads_np_1d, &
               helium%rtmp_3_atoms_beads_1d, &
               helium%rtmp_3_np_1d, &
               helium%itmp_atoms_np_1d, &
               helium%ltmp_atoms_1d, &
               helium%itmp_atoms_1d)

            NULLIFY ( &
               helium%ltmp_3_atoms_beads_3d, &
               helium%rtmp_3_atoms_beads_np_1d, &
               helium%rtmp_3_atoms_beads_1d, &
               helium%rtmp_3_np_1d, &
               helium%itmp_atoms_np_1d, &
               helium%ltmp_atoms_1d, &
               helium%itmp_atoms_1d &
               )

            IF (helium%solute_present) THEN
               ! DEALLOCATE solute-related arrays
               DEALLOCATE ( &
                  helium%rtmp_p_ndim_2d, &
                  helium%rtmp_p_ndim_np_1d, &
                  helium%rtmp_p_ndim_1d, &
                  helium%force_inst, &
                  helium%force_avrg)
               NULLIFY ( &
                  helium%rtmp_p_ndim_2d, &
                  helium%rtmp_p_ndim_np_1d, &
                  helium%rtmp_p_ndim_1d, &
                  helium%force_inst, &
                  helium%force_avrg)
            END IF

            IF (helium%rho_present) THEN
               DEALLOCATE ( &
                  helium%rho_rstr, &
                  helium%rho_accu, &
                  helium%rho_inst, &
                  helium%rho_incr)
               NULLIFY ( &
                  helium%rho_rstr, &
                  helium%rho_accu, &
                  helium%rho_inst, &
                  helium%rho_incr)
               ! DEALLOCATE everything
               DEALLOCATE (helium%rho_property(rho_atom_number)%filename_suffix)
               DEALLOCATE (helium%rho_property(rho_atom_number)%component_name)
               DEALLOCATE (helium%rho_property(rho_atom_number)%component_index)
               NULLIFY (helium%rho_property(rho_atom_number)%filename_suffix)
               NULLIFY (helium%rho_property(rho_atom_number)%component_name)
               NULLIFY (helium%rho_property(rho_atom_number)%component_index)
               DEALLOCATE (helium%rho_property(rho_winding_number)%filename_suffix)
               DEALLOCATE (helium%rho_property(rho_winding_number)%component_name)
               DEALLOCATE (helium%rho_property(rho_winding_number)%component_index)
               NULLIFY (helium%rho_property(rho_winding_number)%filename_suffix)
               NULLIFY (helium%rho_property(rho_winding_number)%component_name)
               NULLIFY (helium%rho_property(rho_winding_number)%component_index)
               DEALLOCATE (helium%rho_property(rho_winding_cycle)%filename_suffix)
               DEALLOCATE (helium%rho_property(rho_winding_cycle)%component_name)
               DEALLOCATE (helium%rho_property(rho_winding_cycle)%component_index)
               NULLIFY (helium%rho_property(rho_winding_cycle)%filename_suffix)
               NULLIFY (helium%rho_property(rho_winding_cycle)%component_name)
               NULLIFY (helium%rho_property(rho_winding_cycle)%component_index)
               DEALLOCATE (helium%rho_property(rho_projected_area)%filename_suffix)
               DEALLOCATE (helium%rho_property(rho_projected_area)%component_name)
               DEALLOCATE (helium%rho_property(rho_projected_area)%component_index)
               NULLIFY (helium%rho_property(rho_projected_area)%filename_suffix)
               NULLIFY (helium%rho_property(rho_projected_area)%component_name)
               NULLIFY (helium%rho_property(rho_projected_area)%component_index)
               DEALLOCATE (helium%rho_property(rho_moment_of_inertia)%filename_suffix)
               DEALLOCATE (helium%rho_property(rho_moment_of_inertia)%component_name)
               DEALLOCATE (helium%rho_property(rho_moment_of_inertia)%component_index)
               NULLIFY (helium%rho_property(rho_moment_of_inertia)%filename_suffix)
               NULLIFY (helium%rho_property(rho_moment_of_inertia)%component_name)
               NULLIFY (helium%rho_property(rho_moment_of_inertia)%component_index)
               DEALLOCATE (helium%rho_property)
               NULLIFY (helium%rho_property)
            END IF

            CALL helium_rdf_release(helium)

            ! DEALLOCATE helium-related arrays
            DEALLOCATE ( &
               helium%atom_plength, &
               helium%plength_inst, &
               helium%plength_avrg, &
               helium%num_accepted, &
               helium%ipmatrix, &
               helium%pmatrix, &
               helium%nmatrix, &
               helium%tmatrix, &
               helium%iperm, &
               helium%permutation, &
               helium%ptable, &
               helium%work, &
               helium%pos)
            NULLIFY ( &
               helium%atom_plength, &
               helium%plength_inst, &
               helium%plength_avrg, &
               helium%num_accepted, &
               helium%ipmatrix, &
               helium%pmatrix, &
               helium%nmatrix, &
               helium%tmatrix, &
               helium%iperm, &
               helium%permutation, &
               helium%ptable, &
               helium%work, &
               helium%pos &
               )

            DO i = 1, SIZE(helium%eij, 1)
               DO j = 1, SIZE(helium%eij, 1)
                  CALL spline_data_release(helium%eij(i, j)%spline_data)
                  CALL spline_data_release(helium%uij(i, j)%spline_data)
                  !TODO: shouldn't that be done in spline_data_release??
                  NULLIFY (helium%eij(i, j)%spline_data, &
                           helium%uij(i, j)%spline_data)
               END DO
            END DO

            DEALLOCATE (helium%eij)
            NULLIFY (helium%eij)

            DEALLOCATE (helium%uij)
            NULLIFY (helium%uij)

            CALL delete_rng_stream(helium%rng_stream_uniform)
            CALL delete_rng_stream(helium%rng_stream_gaussian)

            ! deallocate solute-related arrays
            IF (helium%solute_present) THEN
               CALL helium_destroy_int_arr_ptr(helium%solute_i)
               DEALLOCATE (helium%solute_element)
               NULLIFY (helium%solute_element)
            END IF

            ! Deallocate everything from the helium_set_solute_indices
            IF (ASSOCIATED(helium%ename)) THEN
               DEALLOCATE (helium%ename)
               NULLIFY (helium%ename)
            END IF

            IF (ASSOCIATED(helium%eid)) THEN
               DEALLOCATE (helium%eid)
               NULLIFY (helium%eid)
            END IF

            DEALLOCATE (helium)

         END IF
      END IF
      RETURN
   END SUBROUTINE helium_release

! ***************************************************************************
!> \brief  Initialize helium data structures.
!> \param helium ...
!> \param pint_env ...
!> \par    History
!>         removed refereces to pint_env_type data structure [lwalewski]
!>         2009-11-10 init/restore coords, perm, RNG and forces [lwalewski]
!> \author hforbert
!> \note   Initializes helium coordinates either as random positions or from
!>         HELIUM%COORD section if it's present in the input file.
!>         Initializes helium permutation state as identity permutation or
!>         from HELIUM%PERM section if it's present in the input file.
! **************************************************************************************************
   SUBROUTINE helium_init(helium, pint_env)

      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      LOGICAL                                            :: coords_presampled, explicit, presample
      TYPE(section_vals_type), POINTER                   :: helium_section, sec

      CALL timeset(routineN, handle)

      NULLIFY (helium_section)
      helium_section => section_vals_get_subs_vals(helium%input, &
                                                   "MOTION%PINT%HELIUM")

      ! restore RNG state
      NULLIFY (sec)
      sec => section_vals_get_subs_vals(helium_section, "RNG_STATE")
      CALL section_vals_get(sec, explicit=explicit)
      IF (explicit) THEN
         CALL helium_rng_restore(helium)
      ELSE
         CALL helium_write_line("RNG state initialized as new.")
      END IF

      ! init/restore permutation state
      NULLIFY (sec)
      sec => section_vals_get_subs_vals(helium_section, "PERM")
      CALL section_vals_get(sec, explicit=explicit)
      IF (explicit) THEN
         CALL helium_perm_restore(helium)
      ELSE
         CALL helium_perm_init(helium)
         CALL helium_write_line("Permutation state initialized as identity.")
      END IF

      ! Specify if forces should be obtained as AVG or LAST
      CALL section_vals_val_get(helium_section, "GET_FORCES", &
                                i_val=helium%get_helium_forces)

      ! init/restore coordinates
      NULLIFY (sec)
      sec => section_vals_get_subs_vals(helium_section, "COORD")
      CALL section_vals_get(sec, explicit=explicit)
      IF (explicit) THEN
         CALL helium_coord_restore(helium)
      ELSE
         CALL helium_coord_init(helium)
      END IF
      DO i = 1, helium%atoms
         helium%pos(:, i, helium%beads+1) = helium%pos(:, helium%permutation(i), 1)
      END DO
      helium%work = helium%pos

      ! init center of mass
      IF (helium%solute_present) THEN
         helium%center(:) = pint_com_pos(pint_env)
      ELSE
         IF (helium%periodic) THEN
            helium%center(:) = (/0.0_dp, 0.0_dp, 0.0_dp/)
         ELSE
            helium%center(:) = helium_com(helium)
         END IF
      END IF

      ! Optional helium coordinate presampling:
      CALL section_vals_val_get(helium_section, "PRESAMPLE", &
                                l_val=presample)
      coords_presampled = .FALSE.
      IF (presample) THEN
         helium%current_step = 0
         CALL helium_sample(helium, pint_env)
         IF (helium%solute_present) helium%force_avrg(:, :) = 0.0_dp
         helium%energy_avrg(:) = 0.0_dp
         helium%plength_avrg(:) = 0.0_dp
         helium%num_accepted(:, :) = 0.0_dp
         ! Reset properties accumulated over presample:
         helium%proarea%accu(:) = 0.0_dp
         helium%prarea2%accu(:) = 0.0_dp
         helium%wnmber2%accu(:) = 0.0_dp
         helium%mominer%accu(:) = 0.0_dp
         IF (helium%rho_present) THEN
            helium%rho_accu(:, :, :, :) = 0.0_dp
         END IF
         IF (helium%rdf_present) THEN
            helium%rdf_accu(:, :, :) = 0.0_dp
         END IF
         coords_presampled = .TRUE.
         CALL helium_write_line("Bead coordinates pre-sampled.")
      END IF

      IF (helium%solute_present) THEN
         ! restore helium forces
         NULLIFY (sec)
         sec => section_vals_get_subs_vals(helium_section, "FORCE")
         CALL section_vals_get(sec, explicit=explicit)
         IF (explicit) THEN
            IF (.NOT. coords_presampled) THEN
               CALL helium_force_restore(helium)
            END IF
         ELSE
            IF (.NOT. coords_presampled) THEN
               CALL helium_force_init(helium)
               CALL helium_write_line("Forces on the solute initialized as zero.")
            END IF
         END IF
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE helium_init

! ***************************************************************************
! Data transfer functions.
!
! These functions manipulate and transfer data between the runtime
! environment and the input structure.
! ***************************************************************************

! ***************************************************************************
!> \brief  Initialize helium coordinates with random positions.
!> \param helium ...
!> \date   2009-11-09
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_coord_init(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_coord_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, ic
      REAL(KIND=dp)                                      :: r1, r2

      DO ia = 1, helium%atoms
         DO ic = 1, 3
            r1 = next_random_number(helium%rng_stream_uniform)
            r1 = r1*helium%cell_size
            DO ib = 1, helium%beads
!TODO use thermal gaussian instead
               r2 = next_random_number(helium%rng_stream_uniform)
               helium%pos(ic, ia, ib) = r1+0.1_dp*r2
            END DO
         END DO
         DO ib = 1, helium%beads
            CALL helium_pbc(helium, helium%pos(:, ia, ib))
         END DO
      END DO

      ! store positions at time slice nbeads+1 (rperez):
      DO ia = 1, helium%atoms
         helium%pos(:, ia, helium%beads+1) = helium%pos(:, ia, 1)
      END DO
      ! initialize work array (rperez):
      helium%work = helium%pos

      RETURN
   END SUBROUTINE helium_coord_init

! ***************************************************************************
!> \brief  Restore coordinates from the input structure.
!> \param helium ...
!> \date   2009-11-09
!> \par    History
!>         2010-07-22 accomodate additional cpus in the runtime wrt the
!>                    restart [lwalewski]
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_coord_restore(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_coord_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: actlen, msglen, num_env_restart, offset
      LOGICAL, DIMENSION(:, :, :), POINTER               :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: f
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! assign the pointer to the memory location of the input structure, where
      ! the coordinates are stored
      NULLIFY (message)
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%COORD%_DEFAULT_KEYWORD_", &
                                r_vals=message)

      ! check that the number of values in the input match the current runtime
      actlen = SIZE(message)
      num_env_restart = actlen/helium%atoms/helium%beads/3

      ! distribute coordinates over processors (no message passing)
      msglen = helium%atoms*helium%beads*3
      offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
      NULLIFY (m, f)
      ALLOCATE (m(3, helium%atoms, helium%beads))
      ALLOCATE (f(3, helium%atoms, helium%beads))
      m(:, :, :) = .TRUE.
      f(:, :, :) = 0.0_dp
      helium%pos(:, :, 1:helium%beads) = UNPACK(message(offset+1:offset+msglen), MASK=m, FIELD=f)
      DEALLOCATE (f, m)

      IF (num_env_restart .NE. logger%para_env%num_pe) THEN
         err_str = "Reading bead coordinates from the input file."
         CALL helium_write_line(err_str)
         err_str = "Number of environments in the restart...: '"
         stmp = ""
         WRITE (stmp, *) num_env_restart
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         err_str = "Number of current run time environments.: '"
         stmp = ""
         WRITE (stmp, *) logger%para_env%num_pe
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         IF (num_env_restart .LT. logger%para_env%num_pe) THEN
            err_str = "Replicated bead coordinates from the restarted environments."
            CALL helium_write_line(err_str)
         END IF
         IF (num_env_restart .GT. logger%para_env%num_pe) THEN
            err_str = "Dropped bead coordinates from some restarted environments."
            CALL helium_write_line(err_str)
         END IF
         err_str = "Done."
         CALL helium_write_line(err_str)
      ELSE
         CALL helium_write_line("Bead coordinates read from the input file.")
      END IF

      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_coord_restore

! ***************************************************************************
!> \brief  Initialize forces exerted on the solute
!> \param helium ...
!> \date   2009-11-10
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_force_init(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_force_init', &
         routineP = moduleN//':'//routineN

      IF (helium%solute_present) THEN
         helium%force_avrg(:, :) = 0.0_dp
         helium%force_inst(:, :) = 0.0_dp
      END IF

      RETURN
   END SUBROUTINE helium_force_init

! ***************************************************************************
!> \brief  Restore forces from the input structure to the runtime environment.
!> \param helium ...
!> \date   2009-11-10
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_force_restore(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_force_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: actlen, msglen
      LOGICAL, DIMENSION(:, :), POINTER                  :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: f

! assign the pointer to the memory location of the input structure, where
! the forces are stored

      NULLIFY (message)
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%FORCE%_DEFAULT_KEYWORD_", &
                                r_vals=message)

      ! check if the destination array has correct size
      msglen = helium%solute_atoms*helium%solute_beads*3
      actlen = SIZE(helium%force_avrg)
      err_str = "Invalid size of helium%force_avrg array: actual '"
      stmp = ""
      WRITE (stmp, *) actlen
      err_str = TRIM(ADJUSTL(err_str))// &
                TRIM(ADJUSTL(stmp))//"' but expected '"
      stmp = ""
      WRITE (stmp, *) msglen
      IF (actlen /= msglen) THEN
         err_str = TRIM(ADJUSTL(err_str))// &
                   TRIM(ADJUSTL(stmp))//"'."
         CALL cp__b("motion/helium_methods.F",1092,err_str)
      END IF

      ! restore forces on all processors (no message passing)
      NULLIFY (m, f)
      ALLOCATE (m(helium%solute_beads, helium%solute_atoms*3))
      ALLOCATE (f(helium%solute_beads, helium%solute_atoms*3))
      m(:, :) = .TRUE.
      f(:, :) = 0.0_dp
      helium%force_avrg(:, :) = UNPACK(message(1:msglen), MASK=m, FIELD=f)
      helium%force_inst(:, :) = 0.0_dp
      DEALLOCATE (f, m)

      CALL helium_write_line("Forces on the solute read from the input file.")

      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_force_restore

! ***************************************************************************
!> \brief  Initialize the permutation state.
!> \param helium ...
!> \date   2009-11-05
!> \author Lukasz Walewski
!> \note   Assign the identity permutation at each processor. Inverse
!>         permutation array gets assigned as well.
! **************************************************************************************************
   SUBROUTINE helium_perm_init(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_perm_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia

      DO ia = 1, helium%atoms
         helium%permutation(ia) = ia
         helium%iperm(ia) = ia
      END DO

      RETURN
   END SUBROUTINE helium_perm_init

! ***************************************************************************
!> \brief  Restore permutation state from the input structre.
!> \param helium ...
!> \date   2009-11-05
!> \par    History
!>         2010-07-22 accomodate additional cpus in the runtime wrt the
!>                    restart [lwalewski]
!> \author Lukasz Walewski
!> \note   Transfer permutation state from the input tree to the runtime
!>         data structures on each processor. Inverse permutation array is
!>         recalculated according to the restored permutation state.
! **************************************************************************************************
   SUBROUTINE helium_perm_restore(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_perm_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: actlen, ia, ic, msglen, num_env_restart, &
                                                            offset
      INTEGER, DIMENSION(:), POINTER                     :: message
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! assign the pointer to the memory location of the input structure, where
      ! the permutation state is stored
      NULLIFY (message)
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%PERM%_DEFAULT_KEYWORD_", &
                                i_vals=message)

      ! check the number of environments presumably stored in the restart
      actlen = SIZE(message)
      num_env_restart = actlen/helium%atoms
!TODO maybe add some sanity checks here:
! is num_env_restart integer ?
! is num_env_restart == helium%num_env_restart ?

      ! distribute permutation state over processors (no message passing)
      msglen = helium%atoms
      offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
      helium%permutation(:) = message(offset+1:offset+msglen)

      IF (num_env_restart .NE. logger%para_env%num_pe) THEN
         err_str = "Reading permutation state from the input file."
         CALL helium_write_line(err_str)
         err_str = "Number of environments in the restart...: '"
         stmp = ""
         WRITE (stmp, *) num_env_restart
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         err_str = "Number of current run time environments.: '"
         stmp = ""
         WRITE (stmp, *) logger%para_env%num_pe
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         IF (num_env_restart .LT. logger%para_env%num_pe) THEN
            err_str = "Replicated permutation state from the restarted environments."
            CALL helium_write_line(err_str)
         END IF
         IF (num_env_restart .GT. logger%para_env%num_pe) THEN
            err_str = "Dropped permutation state from some restarted environments."
            CALL helium_write_line(err_str)
         END IF
         err_str = "Done."
         CALL helium_write_line(err_str)
      ELSE
         CALL helium_write_line("Permutation state read from the input file.")
      END IF

      ! recalculate the inverse permutation array
      helium%iperm(:) = 0
      ic = 0
      DO ia = 1, msglen
         IF ((helium%permutation(ia) > 0) .AND. (helium%permutation(ia) <= msglen)) THEN
            helium%iperm(helium%permutation(ia)) = ia
            ic = ic+1
         END IF
      END DO
      err_str = "Invalid HELIUM%PERM state: some numbers not within (1,"
      stmp = ""
      WRITE (stmp, *) msglen
      IF (ic /= msglen) THEN
         err_str = TRIM(ADJUSTL(err_str))// &
                   TRIM(ADJUSTL(stmp))//")."
         CALL cp__b("motion/helium_methods.F",1224,err_str)
      END IF

      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_perm_restore

! ***************************************************************************
!> \brief  Restore averages from the input structure
!> \param helium ...
!> \date   2014-06-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_averages_restore(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_averages_restore', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: msglen, num_env_restart, offset
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! restore projected area
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA", &
                                   r_vals=message)
         num_env_restart = SIZE(message)/3 ! apparent number of environments
         msglen = 3
         offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
         helium%proarea%rstr(:) = message(offset+1:offset+msglen)
      ELSE
         helium%proarea%rstr(:) = 0.0_dp
      END IF

      ! restore projected area squared
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA_2", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%PROJECTED_AREA_2", &
                                   r_vals=message)
         num_env_restart = SIZE(message)/3 ! apparent number of environments
         msglen = 3
         offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
         helium%prarea2%rstr(:) = message(offset+1:offset+msglen)
      ELSE
         helium%prarea2%rstr(:) = 0.0_dp
      END IF

      ! restore winding number squared
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%WINDING_NUMBER_2", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%WINDING_NUMBER_2", &
                                   r_vals=message)
         num_env_restart = SIZE(message)/3 ! apparent number of environments
         msglen = 3
         offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
         helium%wnmber2%rstr(:) = message(offset+1:offset+msglen)
      ELSE
         helium%wnmber2%rstr(:) = 0.0_dp
      END IF

      ! restore moment of inertia
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%MOMENT_OF_INERTIA", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%MOMENT_OF_INERTIA", &
                                   r_vals=message)
         num_env_restart = SIZE(message)/3 ! apparent number of environments
         msglen = 3
         offset = msglen*MOD(logger%para_env%mepos, num_env_restart)
         helium%mominer%rstr(:) = message(offset+1:offset+msglen)
      ELSE
         helium%mominer%rstr(:) = 0.0_dp
      END IF

      IF (helium%rdf_present) THEN
         CALL helium_rdf_restore(helium)
      END IF

      IF (helium%rho_present) THEN
         ! restore densities
         CALL helium_rho_restore(helium)
      END IF

      ! get the weighting factor
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%AVERAGES%IWEIGHT", &
         i_val=helium%averages_iweight)

      ! set the flag indicating whether the averages have been restarted
      CALL section_vals_val_get( &
         helium%input, &
         "EXT_RESTART%RESTART_HELIUM_AVERAGES", &
         l_val=helium%averages_restarted)

      RETURN
   END SUBROUTINE helium_averages_restore

! ***************************************************************************
!> \brief  Create RNG streams and initialize their state.
!> \param helium ...
!> \date   2009-11-04
!> \author Lukasz Walewski
!> \note   TODO: This function shouldn't create (allocate) objects! Only
!>         initialization, i.e. setting the seed values etc, should be done
!>         here, allocation should be moved to helium_create
! **************************************************************************************************
   SUBROUTINE helium_rng_init(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rng_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: rank
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(rng_stream_type), POINTER                     :: next_rngs, prev_rngs

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! Create two RNG streams on each processor and initialize them so, that
      ! each processor gets unique RN sequences.
      NULLIFY (prev_rngs, next_rngs)
      NULLIFY (helium%rng_stream_uniform, helium%rng_stream_gaussian)

      ! Create two RNG strems at each processor by seeding one from the
      ! other. Then, on rank 0: save the pointers to both structures, while
      ! on all other ranks: delete the first structure and keep the second
      ! one to seed next RNG stream to be created.
      CALL create_rng_stream(prev_rngs, &
                             name="helium_rns_uniform", &
                             distribution_type=UNIFORM, &
                             extended_precision=.TRUE.)
      IF (logger%para_env%mepos .EQ. 0) THEN
         helium%rng_stream_uniform => prev_rngs
      END IF
      CALL create_rng_stream(next_rngs, &
                             name="helium_rns_gaussian", &
                             last_rng_stream=prev_rngs, &
                             distribution_type=GAUSSIAN, &
                             extended_precision=.TRUE.)
      IF (logger%para_env%mepos .EQ. 0) THEN
         helium%rng_stream_gaussian => next_rngs
         NULLIFY (prev_rngs)
      ELSE
         CALL delete_rng_stream(prev_rngs)
         prev_rngs => next_rngs
      END IF
      NULLIFY (next_rngs)

      ! At all ranks higher than 0 keep creating new RNG streams one from
      ! the other (so they are all different) and on each rank cut this
      ! process at different stage.
      DO rank = 1, logger%para_env%mepos
         CALL create_rng_stream(next_rngs, &
                                name="helium_rns_uniform", &
                                last_rng_stream=prev_rngs, &
                                distribution_type=UNIFORM, &
                                extended_precision=.TRUE.)
         IF (logger%para_env%mepos .EQ. rank) THEN
            helium%rng_stream_uniform => next_rngs
         END IF
         CALL delete_rng_stream(prev_rngs)
         prev_rngs => next_rngs
         NULLIFY (next_rngs)
         CALL create_rng_stream(next_rngs, &
                                name="helium_rns_gaussian", &
                                last_rng_stream=prev_rngs, &
                                distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE.)
         IF (logger%para_env%mepos .EQ. rank) THEN
            helium%rng_stream_gaussian => next_rngs
            NULLIFY (prev_rngs)
         ELSE
            CALL delete_rng_stream(prev_rngs)
            prev_rngs => next_rngs
         END IF
         NULLIFY (next_rngs)
      END DO

      RETURN
   END SUBROUTINE helium_rng_init

! ***************************************************************************
!> \brief  Restore RNG state from the input structure.
!> \param helium ...
!> \date   2009-11-04
!> \par    History
!>         2010-07-22 Create new rng streams if more cpus available in the
!>         runtime than in the restart [lwalewski]
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rng_restore(helium)
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rng_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: actlen, msglen, num_env_restart, offset, &
                                                            rank
      LOGICAL                                            :: lbf
      LOGICAL, DIMENSION(3, 2)                           :: m
      REAL(KIND=dp)                                      :: bf, bu
      REAL(KIND=dp), DIMENSION(3, 2)                     :: bg, cg, f, ig
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(rng_stream_type), POINTER                     :: next_rngs, prev_rngs

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! assign the pointer to the memory location of the input structure
      ! where the RNG state is stored
      NULLIFY (message)
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%RNG_STATE%_DEFAULT_KEYWORD_", &
                                r_vals=message)

      ! check the number of environments presumably stored in the restart
      actlen = SIZE(message)
      num_env_restart = actlen/40

      IF (logger%para_env%mepos .LT. num_env_restart) THEN

         ! unpack the buffer at each processor, set RNG state (no message passing)
         msglen = 40
         offset = msglen*logger%para_env%mepos
         m(:, :) = .TRUE.
         f(:, :) = 0.0_dp
         bg(:, :) = UNPACK(message(offset+1:offset+6), MASK=m, FIELD=f)
         cg(:, :) = UNPACK(message(offset+7:offset+12), MASK=m, FIELD=f)
         ig(:, :) = UNPACK(message(offset+13:offset+18), MASK=m, FIELD=f)
         bf = message(offset+19)
         bu = message(offset+20)
         IF (bf .GT. 0) THEN
            lbf = .TRUE.
         ELSE
            lbf = .FALSE.
         END IF
         CALL set_rng_stream(helium%rng_stream_uniform, bg=bg, cg=cg, ig=ig, &
                             buffer=bu, buffer_filled=lbf)
         bg(:, :) = UNPACK(message(offset+21:offset+26), MASK=m, FIELD=f)
         cg(:, :) = UNPACK(message(offset+27:offset+32), MASK=m, FIELD=f)
         ig(:, :) = UNPACK(message(offset+33:offset+38), MASK=m, FIELD=f)
         bf = message(offset+39)
         bu = message(offset+40)
         IF (bf .GT. 0) THEN
            lbf = .TRUE.
         ELSE
            lbf = .FALSE.
         END IF
         CALL set_rng_stream(helium%rng_stream_gaussian, bg=bg, cg=cg, ig=ig, &
                             buffer=bu, buffer_filled=lbf)

      ELSE
         ! On processors that did not receive rng state from the restart file
         ! delete rng streams created in helium_rng_init and create them
         ! anew, as they have been initialized with default initial state. Here
         ! the sequence of rng streams starts from the last stream from the
         ! restart file, each stream is initialized from the previously created
         ! one.

         CALL delete_rng_stream(helium%rng_stream_uniform)
         CALL delete_rng_stream(helium%rng_stream_gaussian)
         NULLIFY (prev_rngs, next_rngs)
         NULLIFY (helium%rng_stream_uniform, helium%rng_stream_gaussian)

         ! take the last uniform rng stream from the restart as a starting point
         msglen = 40
         offset = msglen*(num_env_restart-1)
         m(:, :) = .TRUE.
         f(:, :) = 0.0_dp
         bg(:, :) = UNPACK(message(offset+1:offset+6), MASK=m, FIELD=f)
         cg(:, :) = UNPACK(message(offset+7:offset+12), MASK=m, FIELD=f)
         ig(:, :) = UNPACK(message(offset+13:offset+18), MASK=m, FIELD=f)
         bf = message(offset+19)
         bu = message(offset+20)
         IF (bf .GT. 0) THEN
            lbf = .TRUE.
         ELSE
            lbf = .FALSE.
         END IF
         ! create new uniform rng from scratch
         CALL create_rng_stream(prev_rngs, &
                                name="helium_rns_uniform", &
                                distribution_type=UNIFORM, &
                                extended_precision=.TRUE.)
         ! set it to the last uniform rng stream from the restart
         CALL set_rng_stream(prev_rngs, bg=bg, cg=cg, ig=ig, &
                             buffer=bu, buffer_filled=lbf)
         ! use this on the first non-restarted environment as the new rng
         IF (logger%para_env%mepos .EQ. num_env_restart) THEN
            helium%rng_stream_uniform => prev_rngs
         END IF
         ! unpack the last gaussian rng stream from the restart
         bg(:, :) = UNPACK(message(offset+21:offset+26), MASK=m, FIELD=f)
         cg(:, :) = UNPACK(message(offset+27:offset+32), MASK=m, FIELD=f)
         ig(:, :) = UNPACK(message(offset+33:offset+38), MASK=m, FIELD=f)
         bf = message(offset+39)
         bu = message(offset+40)
         IF (bf .GT. 0) THEN
            lbf = .TRUE.
         ELSE
            lbf = .FALSE.
         END IF
         ! create new gaussian rng stream from scratch
         CALL create_rng_stream(next_rngs, &
                                name="helium_rns_gaussian", &
                                last_rng_stream=prev_rngs, &
                                distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE.)
         ! set it to the last gaussian rng stream from the restart
         CALL set_rng_stream(next_rngs, bg=bg, cg=cg, ig=ig, &
                             buffer=bu, buffer_filled=lbf)
         ! use this on the first non-restarted environment as the new gaussian rng
         IF (logger%para_env%mepos .EQ. num_env_restart) THEN
            helium%rng_stream_gaussian => next_rngs
            NULLIFY (prev_rngs)
         ELSE
            CALL delete_rng_stream(prev_rngs)
            prev_rngs => next_rngs
         END IF
         NULLIFY (next_rngs)

         ! At all ranks higher than num_env_restart keep creating new RNG streams
         ! one from the other (so they are all different) and on each rank cut
         ! this process at different stage.
         DO rank = num_env_restart+1, logger%para_env%mepos
            CALL create_rng_stream(next_rngs, &
                                   name="helium_rns_uniform", &
                                   last_rng_stream=prev_rngs, &
                                   distribution_type=UNIFORM, &
                                   extended_precision=.TRUE.)
            IF (logger%para_env%mepos .EQ. rank) THEN
               helium%rng_stream_uniform => next_rngs
            END IF
            CALL delete_rng_stream(prev_rngs)
            prev_rngs => next_rngs
            NULLIFY (next_rngs)
            CALL create_rng_stream(next_rngs, &
                                   name="helium_rns_gaussian", &
                                   last_rng_stream=prev_rngs, &
                                   distribution_type=GAUSSIAN, &
                                   extended_precision=.TRUE.)
            IF (logger%para_env%mepos .EQ. rank) THEN
               helium%rng_stream_gaussian => next_rngs
               NULLIFY (prev_rngs)
            ELSE
               CALL delete_rng_stream(prev_rngs)
               prev_rngs => next_rngs
            END IF
            NULLIFY (next_rngs)
         END DO

      END IF

      ! say what has been done
      IF (num_env_restart .NE. logger%para_env%num_pe) THEN
         err_str = "Reading RNG state from the input file."
         CALL helium_write_line(err_str)
         err_str = "Number of environments in the restart...: '"
         stmp = ""
         WRITE (stmp, *) num_env_restart
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         err_str = "Number of current run time environments.: '"
         stmp = ""
         WRITE (stmp, *) logger%para_env%num_pe
         err_str = TRIM(ADJUSTL(err_str))//TRIM(ADJUSTL(stmp))//"'."
         CALL helium_write_line(err_str)
         IF (num_env_restart .LT. logger%para_env%num_pe) THEN
            err_str = "Created some new RNGs from the restarted environments."
            CALL helium_write_line(err_str)
         END IF
         IF (num_env_restart .GT. logger%para_env%num_pe) THEN
            err_str = "Dropped RNG state from some restarted environments."
            CALL helium_write_line(err_str)
         END IF
         err_str = "Done."
         CALL helium_write_line(err_str)
      ELSE
         CALL helium_write_line("RNG state read from the input file.")
      END IF

      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_rng_restore

! ***************************************************************************
!> \brief  Create the RDF related data structures and initialize
!> \param helium ...
!> \date   2014-02-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rdf_init(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rdf_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: file_name
      CHARACTER(len=default_string_length)               :: err_str, stmp
      INTEGER                                            :: ifirst, ii, ij, ik
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: coords
      TYPE(cp_logger_type), POINTER                      :: logger

      IF (helium%solute_present) THEN
         ! get number of centers from solute
         helium%rdf_num_ctr = helium%solute_atoms
      ELSE
         ! use center of the unit cell or COM of the He droplet
         helium%rdf_num_ctr = 1
      END IF
      ifirst = helium%rdf_num_ctr

      ! read auxiliary centers from file
      NULLIFY (logger)
      logger => cp_get_default_logger()
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RDF%CENTERS_FILE_NAME", &
         c_val=file_name, &
         explicit=explicit)
      NULLIFY (coords)
      IF (explicit) THEN
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%RDF%CENTERS_FILE_NAME", &
            c_val=file_name)
         CALL helium_read_xyz(coords, file_name, logger%para_env)
      END IF

      ! increment number of centers if necessary
      IF (ASSOCIATED(coords)) THEN
         helium%rdf_num_ctr = helium%rdf_num_ctr+SIZE(coords)/3
      END IF

      ! set the flag for RDF and either proceed or return
      IF (helium%rdf_num_ctr > 0) THEN
         helium%rdf_present = .TRUE.
      ELSE
         helium%rdf_present = .FALSE.
         RETURN
      END IF

      ! allocate & store auxiliary centers (align to the end of array)
      ALLOCATE (helium%rdf_centers(3*helium%rdf_num_ctr))
      helium%rdf_centers(:) = 0.0_dp
      IF (ASSOCIATED(coords)) THEN
         ii = 3*ifirst+1
         ij = 3*helium%rdf_num_ctr
         helium%rdf_centers(ii:ij) = coords(:)
         DEALLOCATE (coords)
         NULLIFY (coords)
      END IF

      helium%rdf_num = 4 ! see also helium%rho_num_def

      ! set the maximum RDF range
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RDF%MAXR", &
         explicit=explicit)
      IF (explicit) THEN
         ! use the value explicitly set in the input
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%RDF%MAXR", &
            r_val=helium%rdf_maxr)
      ELSE
         ! use the default value
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%DROPLET_RADIUS", &
            explicit=explicit)
         IF (explicit) THEN
            ! use the droplet radius
            helium%rdf_maxr = helium%droplet_radius
         ELSE
            ! use cell_size and cell_shape
            ! (they are set regardless of us being periodic or not)
            SELECT CASE (helium%cell_shape)
            CASE (helium_cell_shape_cube)
               helium%rdf_maxr = helium%cell_size/2.0_dp
            CASE (helium_cell_shape_octahedron)
               helium%rdf_maxr = helium%cell_size*SQRT(3.0_dp)/4.0_dp
            CASE DEFAULT
               helium%rdf_maxr = 0.0_dp
               WRITE (stmp, *) helium%cell_shape
               err_str = "cell shape unknown ("//TRIM(ADJUSTL(stmp))//")"
               CALL cp__b("motion/helium_methods.F",1741,err_str)
            END SELECT
         END IF
      END IF

      ! get number of bins and set bin spacing
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RDF%NBIN", &
         i_val=helium%rdf_nbin)
      helium%rdf_delr = helium%rdf_maxr/REAL(helium%rdf_nbin, dp)

      ! get the weighting factor
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%AVERAGES%IWEIGHT", &
         i_val=helium%rdf_iweight)

      ! allocate and initialize memory for RDF storage
      ii = helium%rdf_num
      ij = helium%rdf_nbin
      ik = helium%rdf_num_ctr
      ALLOCATE (helium%rdf_inst(ii, ij, ik))
      ALLOCATE (helium%rdf_accu(ii, ij, ik))
      ALLOCATE (helium%rdf_rstr(ii, ij, ik))
      helium%rdf_inst(:, :, :) = 0.0_dp
      helium%rdf_accu(:, :, :) = 0.0_dp
      helium%rdf_rstr(:, :, :) = 0.0_dp

      RETURN
   END SUBROUTINE helium_rdf_init

! ***************************************************************************
!> \brief  Restore the RDFs from the input structure
!> \param helium ...
!> \date   2011-06-22
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rdf_restore(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rdf_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: stmp1, stmp2
      CHARACTER(len=max_line_length)                     :: err_str
      INTEGER                                            :: ii, ij, ik, itmp, msglen
      LOGICAL                                            :: explicit, ltmp
      LOGICAL, DIMENSION(:, :, :), POINTER               :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: f

      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%RDF", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%RDF", &
                                   r_vals=message)
         msglen = SIZE(message)
         itmp = SIZE(helium%rdf_rstr)
         ltmp = (msglen .EQ. itmp)
         IF (.NOT. ltmp) THEN
            stmp1 = ""
            WRITE (stmp1, *) msglen
            stmp2 = ""
            WRITE (stmp2, *) itmp
            err_str = "Size of the RDF array in the input ("// &
                      TRIM(ADJUSTL(stmp1))// &
                      ") .NE. that in the runtime ("// &
                      TRIM(ADJUSTL(stmp2))//")."
            CALL cp__b("motion/helium_methods.F",1814,err_str)
         END IF
      ELSE
         RETURN
      END IF

      ii = helium%rdf_num
      ij = helium%rdf_nbin
      ik = helium%rdf_num_ctr
      NULLIFY (m, f)
      ALLOCATE (m(ii, ij, ik))
      ALLOCATE (f(ii, ij, ik))
      m(:, :, :) = .TRUE.
      f(:, :, :) = 0.0_dp

      helium%rdf_rstr(:, :, :) = UNPACK(message(1:msglen), MASK=m, FIELD=f)
      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%IWEIGHT", &
                                i_val=helium%rdf_iweight)

      DEALLOCATE (f, m)
      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_rdf_restore

! ***************************************************************************
!> \brief  Release/deallocate RDF related data structures
!> \param helium ...
!> \date   2014-02-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rdf_release(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rdf_release', &
         routineP = moduleN//':'//routineN

      IF (helium%rdf_present) THEN

         DEALLOCATE ( &
            helium%rdf_centers, &
            helium%rdf_rstr, &
            helium%rdf_accu, &
            helium%rdf_inst)

         NULLIFY ( &
            helium%rdf_centers, &
            helium%rdf_rstr, &
            helium%rdf_accu, &
            helium%rdf_inst)

      END IF

      RETURN
   END SUBROUTINE helium_rdf_release

! ***************************************************************************
!> \brief  Check whether property <prop> is activated in the input structure
!> \param helium ...
!> \param prop ...
!> \retval is_active ...
!> \date   2014-06-26
!> \author Lukasz Walewski
!> \note   The property is controlled by two items in the input structure,
!>         the printkey and the control section. Two settings result in
!>         the property being considered active:
!>         1. printkey is on at the given print level
!>         2. control section is explicit and on
!>         If the property is considered active it should be calculated
!>         and printed through out the run.
! **************************************************************************************************
   FUNCTION helium_property_active(helium, prop) RESULT(is_active)

      TYPE(helium_solvent_type), POINTER                 :: helium
      CHARACTER(len=*)                                   :: prop
      LOGICAL                                            :: is_active

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_property_active', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: input_path
      INTEGER                                            :: print_level
      LOGICAL                                            :: explicit, is_on
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key, section

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! if the printkey is on at this runlevel consider prop to be active
      NULLIFY (print_key)
      input_path = "MOTION%PINT%HELIUM%PRINT%"//TRIM(ADJUSTL(prop))
      print_key => section_vals_get_subs_vals( &
                   helium%input, &
                   input_path)
      is_on = cp_printkey_is_on( &
              iteration_info=logger%iter_info, &
              print_key=print_key)
      IF (is_on) THEN
         is_active = .TRUE.
         RETURN
      END IF

      ! if the control section is explicit and on consider prop to be active
      ! and activate the printkey
      is_active = .FALSE.
      NULLIFY (section)
      input_path = "MOTION%PINT%HELIUM%"//TRIM(ADJUSTL(prop))
      section => section_vals_get_subs_vals( &
                 helium%input, &
                 input_path)
      CALL section_vals_get(section, explicit=explicit)
      IF (explicit) THEN
         ! control section explicitly present, continue checking
         CALL section_vals_val_get( &
            section, &
            "_SECTION_PARAMETERS_", &
            l_val=is_on)
         IF (is_on) THEN
            ! control section is explicit and on, activate the property
            is_active = .TRUE.
            ! activate the corresponding print_level as well
            print_level = logger%iter_info%print_level
            CALL section_vals_val_set( &
               print_key, &
               "_SECTION_PARAMETERS_", &
               i_val=print_level)
         END IF
      END IF

      RETURN
   END FUNCTION helium_property_active

! ***************************************************************************
!> \brief  Create the density related data structures and initialize
!> \param helium ...
!> \date   2014-02-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rho_property_init(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rho_property_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nc

      ALLOCATE (helium%rho_property(rho_num))

      helium%rho_property(rho_atom_number)%name = 'Atom number density'
      nc = 1
      helium%rho_property(rho_atom_number)%num_components = nc
      ALLOCATE (helium%rho_property(rho_atom_number)%filename_suffix(nc))
      ALLOCATE (helium%rho_property(rho_atom_number)%component_name(nc))
      ALLOCATE (helium%rho_property(rho_atom_number)%component_index(nc))
      helium%rho_property(rho_atom_number)%filename_suffix(1) = 'an'
      helium%rho_property(rho_atom_number)%component_name(1) = ''
      helium%rho_property(rho_atom_number)%component_index(:) = 0

      helium%rho_property(rho_projected_area)%name = 'Projected area squared density, A*A(r)'
      nc = 3
      helium%rho_property(rho_projected_area)%num_components = nc
      ALLOCATE (helium%rho_property(rho_projected_area)%filename_suffix(nc))
      ALLOCATE (helium%rho_property(rho_projected_area)%component_name(nc))
      ALLOCATE (helium%rho_property(rho_projected_area)%component_index(nc))
      helium%rho_property(rho_projected_area)%filename_suffix(1) = 'pa_x'
      helium%rho_property(rho_projected_area)%filename_suffix(2) = 'pa_y'
      helium%rho_property(rho_projected_area)%filename_suffix(3) = 'pa_z'
      helium%rho_property(rho_projected_area)%component_name(1) = 'component x'
      helium%rho_property(rho_projected_area)%component_name(2) = 'component y'
      helium%rho_property(rho_projected_area)%component_name(3) = 'component z'
      helium%rho_property(rho_projected_area)%component_index(:) = 0

      helium%rho_property(rho_winding_number)%name = 'Winding number squared density, W*W(r)'
      nc = 3
      helium%rho_property(rho_winding_number)%num_components = nc
      ALLOCATE (helium%rho_property(rho_winding_number)%filename_suffix(nc))
      ALLOCATE (helium%rho_property(rho_winding_number)%component_name(nc))
      ALLOCATE (helium%rho_property(rho_winding_number)%component_index(nc))
      helium%rho_property(rho_winding_number)%filename_suffix(1) = 'wn_x'
      helium%rho_property(rho_winding_number)%filename_suffix(2) = 'wn_y'
      helium%rho_property(rho_winding_number)%filename_suffix(3) = 'wn_z'
      helium%rho_property(rho_winding_number)%component_name(1) = 'component x'
      helium%rho_property(rho_winding_number)%component_name(2) = 'component y'
      helium%rho_property(rho_winding_number)%component_name(3) = 'component z'
      helium%rho_property(rho_winding_number)%component_index(:) = 0

      helium%rho_property(rho_winding_cycle)%name = 'Winding number squared density, W^2(r)'
      nc = 3
      helium%rho_property(rho_winding_cycle)%num_components = nc
      ALLOCATE (helium%rho_property(rho_winding_cycle)%filename_suffix(nc))
      ALLOCATE (helium%rho_property(rho_winding_cycle)%component_name(nc))
      ALLOCATE (helium%rho_property(rho_winding_cycle)%component_index(nc))
      helium%rho_property(rho_winding_cycle)%filename_suffix(1) = 'wc_x'
      helium%rho_property(rho_winding_cycle)%filename_suffix(2) = 'wc_y'
      helium%rho_property(rho_winding_cycle)%filename_suffix(3) = 'wc_z'
      helium%rho_property(rho_winding_cycle)%component_name(1) = 'component x'
      helium%rho_property(rho_winding_cycle)%component_name(2) = 'component y'
      helium%rho_property(rho_winding_cycle)%component_name(3) = 'component z'
      helium%rho_property(rho_winding_cycle)%component_index(:) = 0

      helium%rho_property(rho_moment_of_inertia)%name = 'Moment of inertia'
      nc = 3
      helium%rho_property(rho_moment_of_inertia)%num_components = nc
      ALLOCATE (helium%rho_property(rho_moment_of_inertia)%filename_suffix(nc))
      ALLOCATE (helium%rho_property(rho_moment_of_inertia)%component_name(nc))
      ALLOCATE (helium%rho_property(rho_moment_of_inertia)%component_index(nc))
      helium%rho_property(rho_moment_of_inertia)%filename_suffix(1) = 'mi_x'
      helium%rho_property(rho_moment_of_inertia)%filename_suffix(2) = 'mi_y'
      helium%rho_property(rho_moment_of_inertia)%filename_suffix(3) = 'mi_z'
      helium%rho_property(rho_moment_of_inertia)%component_name(1) = 'component x'
      helium%rho_property(rho_moment_of_inertia)%component_name(2) = 'component y'
      helium%rho_property(rho_moment_of_inertia)%component_name(3) = 'component z'
      helium%rho_property(rho_moment_of_inertia)%component_index(:) = 0

      helium%rho_property(:)%is_calculated = .FALSE.

      RETURN
   END SUBROUTINE helium_rho_property_init

! ***************************************************************************
!> \brief  Create the density related data structures and initialize
!> \param helium ...
!> \date   2014-02-25
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rho_init(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rho_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, itmp, jtmp
      LOGICAL                                            :: explicit, ltmp

      CALL helium_rho_property_init(helium)

      helium%rho_num_act = 0

      ! check for atom number density
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%ATOM_NUMBER", &
         l_val=ltmp)
      IF (ltmp) THEN
         helium%rho_property(rho_atom_number)%is_calculated = .TRUE.
         helium%rho_num_act = helium%rho_num_act+1
         helium%rho_property(rho_atom_number)%component_index(1) = helium%rho_num_act
      END IF

      ! check for projected area density
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%PROJECTED_AREA_2", &
         l_val=ltmp)
      IF (ltmp) THEN
         helium%rho_property(rho_projected_area)%is_calculated = .TRUE.
         DO ii = 1, helium%rho_property(rho_projected_area)%num_components
            helium%rho_num_act = helium%rho_num_act+1
            helium%rho_property(rho_projected_area)%component_index(ii) = helium%rho_num_act
         END DO
      END IF

      ! check for winding number density
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%WINDING_NUMBER_2", &
         l_val=ltmp)
      IF (ltmp) THEN
         helium%rho_property(rho_winding_number)%is_calculated = .TRUE.
         DO ii = 1, helium%rho_property(rho_winding_number)%num_components
            helium%rho_num_act = helium%rho_num_act+1
            helium%rho_property(rho_winding_number)%component_index(ii) = helium%rho_num_act
         END DO
      END IF

      ! check for winding cycle density
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%WINDING_CYCLE_2", &
         l_val=ltmp)
      IF (ltmp) THEN
         helium%rho_property(rho_winding_cycle)%is_calculated = .TRUE.
         DO ii = 1, helium%rho_property(rho_winding_cycle)%num_components
            helium%rho_num_act = helium%rho_num_act+1
            helium%rho_property(rho_winding_cycle)%component_index(ii) = helium%rho_num_act
         END DO
      END IF

      ! check for moment of inertia density
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%MOMENT_OF_INERTIA", &
         l_val=ltmp)
      IF (ltmp) THEN
         helium%rho_property(rho_moment_of_inertia)%is_calculated = .TRUE.
         DO ii = 1, helium%rho_property(rho_moment_of_inertia)%num_components
            helium%rho_num_act = helium%rho_num_act+1
            helium%rho_property(rho_moment_of_inertia)%component_index(ii) = helium%rho_num_act
         END DO
      END IF

      ! set the cube dimensions, etc (common to all estimators)
      helium%rho_maxr = helium%cell_size
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%NBIN", &
         i_val=helium%rho_nbin)
      helium%rho_delr = helium%rho_maxr/REAL(helium%rho_nbin, dp)

      ! check for optional estimators based on winding paths
      helium%rho_num_min_len_wdg = 0
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_WDG", &
         explicit=explicit)
      IF (explicit) THEN
         NULLIFY (helium%rho_min_len_wdg_vals)
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_WDG", &
            i_vals=helium%rho_min_len_wdg_vals)
         itmp = SIZE(helium%rho_min_len_wdg_vals)
         IF (itmp .GT. 0) THEN
            helium%rho_num_min_len_wdg = itmp
            helium%rho_num_act = helium%rho_num_act+itmp
         END IF
      END IF

      ! check for optional estimators based on non-winding paths
      helium%rho_num_min_len_non = 0
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_NON", &
         explicit=explicit)
      IF (explicit) THEN
         NULLIFY (helium%rho_min_len_non_vals)
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_NON", &
            i_vals=helium%rho_min_len_non_vals)
         itmp = SIZE(helium%rho_min_len_non_vals)
         IF (itmp .GT. 0) THEN
            helium%rho_num_min_len_non = itmp
            helium%rho_num_act = helium%rho_num_act+itmp
         END IF
      END IF

      ! check for optional estimators based on all paths
      helium%rho_num_min_len_all = 0
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_ALL", &
         explicit=explicit)
      IF (explicit) THEN
         NULLIFY (helium%rho_min_len_all_vals)
         CALL section_vals_val_get( &
            helium%input, &
            "MOTION%PINT%HELIUM%RHO%MIN_CYCLE_LENGTHS_ALL", &
            i_vals=helium%rho_min_len_all_vals)
         itmp = SIZE(helium%rho_min_len_all_vals)
         IF (itmp .GT. 0) THEN
            helium%rho_num_min_len_all = itmp
            helium%rho_num_act = helium%rho_num_act+itmp
         END IF
      END IF

      ! get the weighting factor
      CALL section_vals_val_get( &
         helium%input, &
         "MOTION%PINT%HELIUM%AVERAGES%IWEIGHT", &
         i_val=helium%rho_iweight)

      ! allocate and initialize memory for density storage
      itmp = helium%rho_nbin
      jtmp = helium%rho_num_act
      ALLOCATE (helium%rho_inst(jtmp, itmp, itmp, itmp))
      ALLOCATE (helium%rho_accu(jtmp, itmp, itmp, itmp))
      ALLOCATE (helium%rho_rstr(jtmp, itmp, itmp, itmp))
      ALLOCATE (helium%rho_incr(jtmp, helium%atoms, helium%beads))

      helium%rho_incr(:, :, :) = 0.0_dp
      helium%rho_inst(:, :, :, :) = 0.0_dp
      helium%rho_accu(:, :, :, :) = 0.0_dp
      helium%rho_rstr(:, :, :, :) = 0.0_dp

      RETURN
   END SUBROUTINE helium_rho_init

! ***************************************************************************
!> \brief  Restore the densities from the input structure.
!> \param helium ...
!> \date   2011-06-22
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE helium_rho_restore(helium)

      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_rho_restore', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: stmp1, stmp2
      CHARACTER(len=max_line_length)                     :: err_str
      INTEGER                                            :: itmp, msglen
      LOGICAL                                            :: explicit, ltmp
      LOGICAL, DIMENSION(:, :, :, :), POINTER            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: message
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: f

      CALL section_vals_val_get(helium%input, &
                                "MOTION%PINT%HELIUM%AVERAGES%RHO", &
                                explicit=explicit)
      IF (explicit) THEN
         NULLIFY (message)
         CALL section_vals_val_get(helium%input, &
                                   "MOTION%PINT%HELIUM%AVERAGES%RHO", &
                                   r_vals=message)
         msglen = SIZE(message)
         itmp = SIZE(helium%rho_rstr)
         ltmp = (msglen .EQ. itmp)
         IF (.NOT. ltmp) THEN
            stmp1 = ""
            WRITE (stmp1, *) msglen
            stmp2 = ""
            WRITE (stmp2, *) itmp
            err_str = "Size of the S density array in the input ("// &
                      TRIM(ADJUSTL(stmp1))// &
                      ") .NE. that in the runtime ("// &
                      TRIM(ADJUSTL(stmp2))//")."
            CALL cp__b("motion/helium_methods.F",2248,err_str)
         END IF
      ELSE
         RETURN
      END IF

      itmp = helium%rho_nbin
      NULLIFY (m, f)
      ALLOCATE (m(helium%rho_num_act, itmp, itmp, itmp))
      ALLOCATE (f(helium%rho_num_act, itmp, itmp, itmp))
      m(:, :, :, :) = .TRUE.
      f(:, :, :, :) = 0.0_dp

      helium%rho_rstr(:, :, :, :) = UNPACK(message(1:msglen), MASK=m, FIELD=f)

      DEALLOCATE (f, m)
      NULLIFY (message)

      RETURN
   END SUBROUTINE helium_rho_restore

! ***************************************************************************
!> \brief Count atoms of different types and store their global indices.
!> \param helium ...
!> \param pint_env ...
!> \author Lukasz Walewski
!> \note  Arrays ALLOCATEd here are (should be) DEALLOCATEd in
!>        helium_release.
! **************************************************************************************************
   SUBROUTINE helium_set_solute_indices(helium, pint_env)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_set_solute_indices', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: cur_ename
      CHARACTER(len=default_path_length)                 :: msg_str
      INTEGER                                            :: cur_eid, i, j, mnum, natoms
      INTEGER, DIMENSION(hid_num)                        :: el_counts
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: mass
      TYPE(cp_subsys_type), POINTER                      :: my_subsys
      TYPE(f_env_type), POINTER                          :: my_f_env
      TYPE(particle_list_type), POINTER                  :: my_particles

! set up my_particles structure

      NULLIFY (my_f_env, my_subsys, my_particles)
      CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id, &
                              f_env=my_f_env)
      CALL force_env_get(force_env=my_f_env%force_env, subsys=my_subsys)
      CALL cp_subsys_get(my_subsys, particles=my_particles)
      CALL f_env_rm_defaults(my_f_env)

      natoms = helium%solute_atoms
      NULLIFY (helium%solute_element)
      ALLOCATE (helium%solute_element(natoms))

      NULLIFY (helium%ename, helium%eid)
      ALLOCATE (helium%ename(hid_num))
      ALLOCATE (helium%eid(hid_num))

      NULLIFY (helium%solute_i)
      ALLOCATE (helium%solute_i(hid_num))
      DO i = 1, hid_num
!TODO important to nullify here, otherwise helium_destroy_int_arr_ptr
!     in helium_release will fail
!     safer solution: change int_arr_ptr%iap from POINTER to ALLOCATABLE
         NULLIFY (helium%solute_i(i)%iap)
      END DO

      ! find out how many different atomic types are there
      helium%enum = 0
      el_counts(:) = 0
      DO i = 1, natoms
         CALL get_atomic_kind(my_particles%els(i)%atomic_kind, mass=mass)
         mnum = NINT(cp_unit_from_cp2k(mass, "amu"))
         SELECT CASE (mnum)
         CASE (35)
            cur_ename = "CL"
            cur_eid = hid_chlorine
         CASE (16)
            cur_ename = "O "
            cur_eid = hid_oxygen
         CASE (1)
            cur_ename = "H "
            cur_eid = hid_hydrogen
         CASE (12)
            cur_ename = "C "
            cur_eid = hid_carbon
         CASE DEFAULT
            cur_ename = "XX"
            cur_eid = -1
            WRITE (msg_str, '(A,I0,A)') &
               "Atomic mass number '", &
               mnum, &
               "' not supported by the HELIUM code."
            CALL cp__b("motion/helium_methods.F",2346,msg_str)
         END SELECT
         helium%solute_element(i) = cur_ename
         el_counts(cur_eid) = el_counts(cur_eid)+1
         ! check if this element symbol is already present in element table
         found = .FALSE.
         DO j = 1, helium%enum
            IF (helium%ename(j) == cur_ename) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
         IF (.NOT. found) THEN
            ! increase current number of different elements
            helium%enum = helium%enum+1
            helium%eid(helium%enum) = cur_eid
            helium%ename(helium%enum) = cur_ename
         END IF
      END DO

      DO i = 1, helium%enum
         ALLOCATE (helium%solute_i(helium%eid(i))%iap(el_counts(helium%eid(i))))
      END DO

      ! store atomic indices of different atomic kinds
      el_counts(:) = 0
      DO i = 1, natoms
         SELECT CASE (helium%solute_element (i))
         CASE ("CL")
            el_counts(hid_chlorine) = el_counts(hid_chlorine)+1
            helium%solute_i(hid_chlorine)%iap(el_counts(hid_chlorine)) = i
         CASE ("O ")
            el_counts(hid_oxygen) = el_counts(hid_oxygen)+1
            helium%solute_i(hid_oxygen)%iap(el_counts(hid_oxygen)) = i
         CASE ("H ")
            el_counts(hid_hydrogen) = el_counts(hid_hydrogen)+1
            helium%solute_i(hid_hydrogen)%iap(el_counts(hid_hydrogen)) = i
         CASE ("C ")
            el_counts(hid_carbon) = el_counts(hid_carbon)+1
            helium%solute_i(hid_carbon)%iap(el_counts(hid_carbon)) = i
         CASE DEFAULT
            WRITE (msg_str, '(A)') &
               "Should never get here, check what happened!"
            CALL cp__b("motion/helium_methods.F",2389,msg_str)
         END SELECT
      END DO

      RETURN
   END SUBROUTINE helium_set_solute_indices

! ***************************************************************************
!> \brief Sets helium%solute_cell based on the solute's force_env.
!> \param helium ...
!> \param pint_env ...
!> \author Lukasz Walewski
!> \note  The simulation cell for the solvated molecule is taken from force_env
!>        which should assure that we get proper cell dimensions regardless of
!>        the method used for the solute (QS, FIST). Helium solvent needs the
!>        solute's cell dimensions to calculate the solute-solvent distances
!>        correctly.
!> \note  At the moment only orthorhombic cells are supported.
! **************************************************************************************************
   SUBROUTINE helium_set_solute_cell(helium, pint_env)
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'helium_set_solute_cell', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_orthorhombic
      TYPE(cell_type), POINTER                           :: my_cell
      TYPE(f_env_type), POINTER                          :: my_f_env

! get the cell structure from pint_env

      NULLIFY (my_f_env, my_cell)
      CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id, &
                              f_env=my_f_env)
      CALL force_env_get(force_env=my_f_env%force_env, cell=my_cell)
      CALL f_env_rm_defaults(my_f_env)

      CALL get_cell(my_cell, orthorhombic=my_orthorhombic)
      IF (.NOT. my_orthorhombic) THEN
         CALL cp__b("motion/helium_methods.F",2429,"Helium solvent not implemented for non-orthorhombic cells.")
      ELSE
         helium%solute_cell => my_cell
      END IF

      RETURN
   END SUBROUTINE helium_set_solute_cell

END MODULE helium_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods to apply GLE to PI runs.
!> \author michelec
!> \par History
!>      06.2010 created [michelec]
!> \note   trying to keep duplication at a minimum....
! **************************************************************************************************

MODULE pint_gle
   USE gle_system_dynamics,             ONLY: gle_cholesky_stab
   USE gle_system_types,                ONLY: gle_type
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE pint_types,                      ONLY: pint_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pint_gle_step, pint_gle_init, pint_calc_gle_energy

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_gle'

CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_calc_gle_energy(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      INTEGER                                            :: i

      pint_env%e_gle = 0._dp
      IF (ASSOCIATED(pint_env%gle)) THEN
         DO i = 1, pint_env%gle%loc_num_gle
            pint_env%e_gle = pint_env%e_gle+pint_env%gle%nvt(i)%thermostat_energy
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_gle_init(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      INTEGER                                            :: i, ib, idim, imap, j
      REAL(dp) :: mf, rr(pint_env%gle%ndim), cc(pint_env%gle%ndim, pint_env%gle%ndim)
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CALL gle_cholesky_stab(pint_env%gle%c_mat, cc, pint_env%gle%ndim)
      DO i = 1, pint_env%gle%loc_num_gle
         imap = pint_env%gle%map_info%index(i)
         ib = 1+(imap-1)/pint_env%ndim
         idim = 1+MOD(imap-1, pint_env%ndim)
         mf = 1.0_dp/SQRT(pint_env%mass_fict(ib, idim))
         rng_stream => pint_env%gle%nvt(i)%gaussian_rng_stream
         DO j = 1, pint_env%gle%ndim
            rr(j) = next_random_number(rng_stream)*mf
         END DO
         pint_env%gle%nvt(i)%s = MATMUL(cc, rr)
      END DO

   END SUBROUTINE pint_gle_init

! **************************************************************************************************
!> \brief ...
!> \param pint_env ...
! **************************************************************************************************
   SUBROUTINE pint_gle_step(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_gle_step', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iadd, ib, ideg, idim, imap, &
                                                            ndim, num
      REAL(dp)                                           :: alpha, beta, mf, rr
      REAL(dp), DIMENSION(:, :), POINTER                 :: a_mat, e_tmp, h_tmp, s_tmp
      TYPE(gle_type), POINTER                            :: gle
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CALL timeset(routineN, handle)

      gle => pint_env%gle
      NULLIFY (rng_stream)
      ndim = gle%ndim

      ALLOCATE (s_tmp(ndim, gle%loc_num_gle))
      s_tmp = 0.0_dp
      ALLOCATE (e_tmp(ndim, gle%loc_num_gle))
      ALLOCATE (h_tmp(ndim, gle%loc_num_gle))

      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%index(ideg)
         ib = 1+(imap-1)/pint_env%ndim
         idim = 1+MOD(imap-1, pint_env%ndim)

         gle%nvt(ideg)%s(1) = pint_env%uv_t(ib, idim)
         gle%nvt(ideg)%thermostat_energy = gle%nvt(ideg)%thermostat_energy &
                                           +0.5_dp*pint_env%mass_fict(ib, idim)*gle%nvt(ideg)%s(1)**2
         s_tmp(1, imap) = gle%nvt(ideg)%s(1)
         rng_stream => gle%nvt(ideg)%gaussian_rng_stream
         rr = next_random_number(rng_stream)
         mf = 1.0_dp/SQRT(pint_env%mass_fict(ib, idim))
         e_tmp(1, imap) = rr*mf
         DO iadd = 2, ndim
            s_tmp(iadd, imap) = gle%nvt(ideg)%s(iadd)
            rr = next_random_number(rng_stream)
            e_tmp(iadd, imap) = rr*mf
         END DO
      END DO
      num = gle%loc_num_gle
      a_mat => gle%gle_s
      alpha = 1.0_dp
      beta = 0.0_dp

      CALL DGEMM('N', 'N', ndim, num, ndim, alpha, a_mat(1, 1), ndim, e_tmp(1, 1), ndim, beta, h_tmp(1, 1), ndim)

      a_mat => gle%gle_t
      beta = 1.0_dp
      CALL DGEMM("N", "N", ndim, num, ndim, alpha, a_mat(1, 1), ndim, s_tmp(1, 1), ndim, beta, h_tmp(1, 1), ndim)

      DO ideg = 1, gle%loc_num_gle
         imap = gle%map_info%index(ideg)

         DO iadd = 1, ndim
            gle%nvt(ideg)%s(iadd) = h_tmp(iadd, imap)
         END DO
         ib = 1+(imap-1)/pint_env%ndim
         idim = 1+MOD(imap-1, pint_env%ndim)
         pint_env%uv_t(ib, idim) = gle%nvt(ideg)%s(1)
         gle%nvt(ideg)%thermostat_energy = gle%nvt(ideg)%thermostat_energy &
                                           -0.5_dp*pint_env%mass_fict(ib, idim)*gle%nvt(ideg)%s(1)**2
      END DO
      pint_env%e_kin_t = 0.0_dp
      DEALLOCATE (e_tmp, s_tmp, h_tmp)
      CALL timestop(handle)
   END SUBROUTINE
END MODULE
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods to performs a path integral run
!> \author fawzi
!> \par History
!>      02.2005 created [fawzi]
!>           11.2006 modified so it might actually work [hforbert]
!>           10.2015 added RPMD propagator
!>           10.2015 added exact harmonic integrator [Felix Uhl]
!> \note   quick & dirty rewrite of my python program
! **************************************************************************************************
MODULE pint_methods

   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE bibliography,                    ONLY: Ceriotti2010,&
                                              Ceriotti2012,&
                                              cite_reference
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get
   USE gle_system_dynamics,             ONLY: gle_cholesky_stab,&
                                              gle_matrix_exp,&
                                              restart_gle
   USE gle_system_types,                ONLY: gle_dealloc,&
                                              gle_init,&
                                              gle_thermo_create
   USE global_types,                    ONLY: global_environment_type
   USE helium_interactions,             ONLY: helium_intpot_scan
   USE helium_io,                       ONLY: helium_write_cubefile
   USE helium_methods,                  ONLY: helium_create,&
                                              helium_init,&
                                              helium_release
   USE helium_sampling,                 ONLY: helium_do_run,&
                                              helium_step
   USE helium_types,                    ONLY: helium_solvent_type
   USE input_constants,                 ONLY: integrate_exact,&
                                              integrate_numeric,&
                                              propagator_rpmd,&
                                              transformation_normal,&
                                              transformation_stage
   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_section_types,             ONLY: &
        section_type, section_vals_get, section_vals_get_subs_vals, section_vals_release, &
        section_vals_retain, section_vals_type, section_vals_val_get, section_vals_val_set, &
        section_vals_val_unset
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathconstants,                   ONLY: twopi
   USE mathlib,                         ONLY: gcd
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              next_rng_seed,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE pint_gle,                        ONLY: pint_calc_gle_energy,&
                                              pint_gle_init,&
                                              pint_gle_step
   USE pint_io,                         ONLY: pint_write_centroids,&
                                              pint_write_com,&
                                              pint_write_ener,&
                                              pint_write_line,&
                                              pint_write_rgyr,&
                                              pint_write_step_info,&
                                              pint_write_trajectory
   USE pint_normalmode,                 ONLY: normalmode_calc_uf_h,&
                                              normalmode_env_create,&
                                              normalmode_init_masses,&
                                              normalmode_release
   USE pint_piglet,                     ONLY: pint_calc_piglet_energy,&
                                              pint_piglet_create,&
                                              pint_piglet_init,&
                                              pint_piglet_release,&
                                              pint_piglet_step
   USE pint_pile,                       ONLY: pint_calc_pile_energy,&
                                              pint_pile_init,&
                                              pint_pile_release,&
                                              pint_pile_step
   USE pint_public,                     ONLY: pint_com_pos,&
                                              pint_levy_walk
   USE pint_staging,                    ONLY: staging_calc_uf_h,&
                                              staging_env_create,&
                                              staging_init_masses,&
                                              staging_release
   USE pint_transformations,            ONLY: pint_f2uf,&
                                              pint_u2x,&
                                              pint_x2u
   USE pint_types,                      ONLY: &
        e_conserved_id, e_kin_thermo_id, e_kin_virial_id, e_potential_id, pint_env_type, &
        thermostat_gle, thermostat_none, thermostat_nose, thermostat_piglet, thermostat_pile
   USE replica_methods,                 ONLY: rep_env_calc_e_f,&
                                              rep_env_create
   USE replica_types,                   ONLY: rep_env_release,&
                                              replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pint_methods'
   INTEGER, SAVE, PRIVATE :: last_pint_id = 0

   PUBLIC :: do_pint_run

CONTAINS

! ***************************************************************************
!> \brief  Create a path integral environment
!> \param pint_env ...
!> \param input ...
!> \param input_declaration ...
!> \param para_env ...
!> \par    History
!>           Fixed some bugs [hforbert]
!>           Added normal mode transformation [hforbert]
!>           10.2015 Added RPMD propagator and harmonic integrator [Felix Uhl]
!> \author fawzi
!> \note   Might return an unassociated pointer in parallel on the processors
!>         that are not needed.
! **************************************************************************************************
   SUBROUTINE pint_create(pint_env, input, input_declaration, para_env)

      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_create', routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: output_file_name, project_name
      CHARACTER(len=default_string_length)               :: msg
      INTEGER                                            :: handle, iat, ibead, idim, idir, ierr, &
                                                            itmp, nrep, prep, stat
      LOGICAL                                            :: explicit, ltmp, wrong_input
      REAL(kind=dp)                                      :: dt, mass, omega
      REAL(kind=dp), DIMENSION(3, 2)                     :: seed
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(section_vals_type), POINTER                   :: gle_section, nose_section, &
                                                            piglet_section, pile_section, &
                                                            pint_section, transform_section

      CALL timeset(routineN, handle)

      NULLIFY (f_env, subsys, particles, nose_section, gle_section)

      IF(.NOT.(.NOT. ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",174)
      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("motion/pint_methods.F",175)
      IF(.NOT.(input%ref_count > 0))CALL cp__a("motion/pint_methods.F",176)
      NULLIFY (rep_env)
      pint_section => section_vals_get_subs_vals(input, "MOTION%PINT")
      CALL section_vals_val_get(pint_section, "p", i_val=nrep)
      CALL section_vals_val_get(pint_section, "proc_per_replica", &
                                i_val=prep)
      ! Maybe let the user have his/her way as long as prep is
      ! within the bounds of number of CPUs??
      IF ((prep < 1) .OR. (prep > para_env%num_pe) .OR. &
          (MOD(prep*nrep, para_env%num_pe) /= 0)) THEN
         prep = para_env%num_pe/gcd(para_env%num_pe, nrep)
         IF (para_env%ionode) THEN
            WRITE (UNIT=msg, FMT=*) "PINT WARNING: Adjusting number of processors per replica to ", prep
            CALL cp__w("motion/pint_methods.F",189,msg)
         END IF
      END IF

      ! replica_env modifies the global input structure - which is wrong - one
      ! of the side effects is the inifite adding of the -r-N string to the
      ! project name and the output file name, which corrupts restart files.
      ! For now: save the project name and output file name and restore them
      ! after the rep_env_create has executed - the initialization of the
      ! replicas will run correctly anyways.
      ! TODO: modify rep_env so that it behaves better
      CALL section_vals_val_get(input, "GLOBAL%PROJECT_NAME", c_val=project_name)
      CALL section_vals_val_get(input, "GLOBAL%OUTPUT_FILE_NAME", c_val=output_file_name)
      CALL rep_env_create(rep_env, para_env=para_env, input=input, &
                          input_declaration=input_declaration, nrep=nrep, prep=prep, row_force=.TRUE.)
      CALL section_vals_val_set(input, "GLOBAL%PROJECT_NAME", c_val=TRIM(project_name))
      IF (LEN_TRIM(output_file_name) .GT. 0) THEN
         CALL section_vals_val_set(input, "GLOBAL%OUTPUT_FILE_NAME", c_val=TRIM(output_file_name))
      ELSE
         CALL section_vals_val_unset(input, "GLOBAL%OUTPUT_FILE_NAME")
      END IF
      IF (.NOT. ASSOCIATED(rep_env)) RETURN

      ALLOCATE (pint_env)
      last_pint_id = last_pint_id+1
      pint_env%id_nr = last_pint_id
      pint_env%ref_count = 1
      NULLIFY (pint_env%replicas, pint_env%input, pint_env%staging_env, &
               pint_env%normalmode_env, pint_env%propagator)
      pint_env%p = nrep
      pint_env%replicas => rep_env
      pint_env%ndim = rep_env%ndim
      pint_env%input => input

      CALL section_vals_retain(pint_env%input)

      ! get first step, last step, number of steps, etc
      CALL section_vals_val_get(input, "MOTION%PINT%ITERATION", &
                                i_val=itmp)
      pint_env%first_step = itmp
      CALL section_vals_val_get(input, "MOTION%PINT%MAX_STEP", &
                                explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(input, "MOTION%PINT%MAX_STEP", &
                                   i_val=itmp)
         pint_env%last_step = itmp
         pint_env%num_steps = pint_env%last_step-pint_env%first_step
      ELSE
         CALL section_vals_val_get(input, "MOTION%PINT%NUM_STEPS", &
                                   i_val=itmp)
         pint_env%num_steps = itmp
         pint_env%last_step = pint_env%first_step+pint_env%num_steps
      END IF

      CALL section_vals_val_get(pint_section, "DT", &
                                r_val=pint_env%dt)
      pint_env%t = pint_env%first_step*pint_env%dt

      CALL section_vals_val_get(pint_section, "nrespa", i_val=pint_env%nrespa)
      CALL section_vals_val_get(pint_section, "Temp", r_val=pint_env%kT)
      CALL section_vals_val_get(pint_section, "T_TOL", &
                                r_val=pint_env%t_tol)

      ALLOCATE (pint_env%propagator)
      CALL section_vals_val_get(pint_section, "propagator", &
                                i_val=pint_env%propagator%prop_kind)
      !Initialize simulation temperature depending on the propagator
      !As well as the scaling factor for the physical potential
      IF (pint_env%propagator%prop_kind == propagator_rpmd) THEN
         pint_env%propagator%temp_phys2sim = REAL(pint_env%p, dp)
         pint_env%propagator%physpotscale = 1.0_dp
      ELSE
         pint_env%propagator%temp_phys2sim = 1.0_dp
         pint_env%propagator%physpotscale = 1.0_dp/REAL(pint_env%p, dp)
      END IF
      pint_env%propagator%temp_sim2phys = 1.0_dp/pint_env%propagator%temp_phys2sim
      pint_env%kT = pint_env%kT*pint_env%propagator%temp_phys2sim

      CALL section_vals_val_get(pint_section, "transformation", &
                                i_val=pint_env%transform)

      NULLIFY (pint_env%tx, pint_env%tv, pint_env%tv_t, pint_env%tv_old, pint_env%tv_new, pint_env%tf)

      pint_env%nnos = 0
      pint_env%pimd_thermostat = thermostat_none
      nose_section => section_vals_get_subs_vals(input, "MOTION%PINT%NOSE")
      CALL section_vals_get(nose_section, explicit=explicit)
      IF (explicit) THEN
         IF (pint_env%propagator%prop_kind == propagator_rpmd) THEN
            CALL cp__b("motion/pint_methods.F",278,"RPMD Propagator with Nose-thermostat not implemented!")
         END IF
         CALL section_vals_val_get(nose_section, "nnos", i_val=pint_env%nnos)
         IF (pint_env%nnos > 0) THEN
            pint_env%pimd_thermostat = thermostat_nose
            ALLOCATE ( &
               pint_env%tx(pint_env%nnos, pint_env%p, pint_env%ndim), &
               pint_env%tv(pint_env%nnos, pint_env%p, pint_env%ndim), &
               pint_env%tv_t(pint_env%nnos, pint_env%p, pint_env%ndim), &
               pint_env%tv_old(pint_env%nnos, pint_env%p, pint_env%ndim), &
               pint_env%tv_new(pint_env%nnos, pint_env%p, pint_env%ndim), &
               pint_env%tf(pint_env%nnos, pint_env%p, pint_env%ndim))
            pint_env%tx = 0._dp
            pint_env%tv = 0._dp
            pint_env%tv_t = 0._dp
            pint_env%tv_old = 0._dp
            pint_env%tv_new = 0._dp
            pint_env%tf = 0._dp
         END IF
      END IF

      pint_env%beta = 1._dp/(pint_env%kT*pint_env%propagator%temp_sim2phys)
!TODO
! v_tol not in current input structure
! should also probably be part of nose_section
!       CALL section_vals_val_get(transform_section,"v_tol_nose",r_val=pint_env%v_tol)
!MK ... but we have to initialise v_tol
      pint_env%v_tol = 0.0_dp ! to be fixed

      NULLIFY (pint_env%randomG)

      seed(:, :) = next_rng_seed()
      CALL create_rng_stream(pint_env%randomG, &
                             name="pint_randomG", &
                             distribution_type=GAUSSIAN, &
                             extended_precision=.TRUE., &
                             seed=seed)

      ALLOCATE (pint_env%e_pot_bead(pint_env%p))
      pint_env%e_pot_bead = 0._dp
      pint_env%e_pot_h = 0._dp
      pint_env%e_kin_beads = 0._dp
      pint_env%e_pot_t = 0._dp
      pint_env%e_gle = 0._dp
      pint_env%e_pile = 0._dp
      pint_env%e_piglet = 0._dp
      pint_env%e_kin_t = 0._dp
      pint_env%energy(:) = 0.0_dp

!TODO: rearrange to use standard nose hoover chain functions/data types

      ALLOCATE ( &
         pint_env%x(pint_env%p, pint_env%ndim), &
         pint_env%v(pint_env%p, pint_env%ndim), &
         pint_env%f(pint_env%p, pint_env%ndim), &
         pint_env%external_f(pint_env%p, pint_env%ndim), &
         pint_env%ux(pint_env%p, pint_env%ndim), &
         pint_env%ux_t(pint_env%p, pint_env%ndim), &
         pint_env%uv(pint_env%p, pint_env%ndim), &
         pint_env%uv_t(pint_env%p, pint_env%ndim), &
         pint_env%uv_new(pint_env%p, pint_env%ndim), &
         pint_env%uf(pint_env%p, pint_env%ndim), &
         pint_env%uf_h(pint_env%p, pint_env%ndim), &
         pint_env%centroid(pint_env%ndim), &
         pint_env%rtmp_ndim(pint_env%ndim), &
         pint_env%rtmp_natom(pint_env%ndim/3), &
         STAT=stat)
      IF(.NOT.(stat == 0))CALL cp__a("motion/pint_methods.F",345)
      pint_env%x = 0._dp
      pint_env%v = 0._dp
      pint_env%f = 0._dp
      pint_env%external_f = 0._dp
      pint_env%ux = 0._dp
      pint_env%ux_t = 0._dp
      pint_env%uv = 0._dp
      pint_env%uv_t = 0._dp
      pint_env%uv_new = 0._dp
      pint_env%uf = 0._dp
      pint_env%uf_h = 0._dp
      pint_env%centroid(:) = 0.0_dp
      pint_env%rtmp_ndim = 0._dp
      pint_env%rtmp_natom = 0._dp
      pint_env%time_per_step = 0.0_dp

      IF (pint_env%transform == transformation_stage) THEN
         transform_section => section_vals_get_subs_vals(input, &
                                                         "MOTION%PINT%STAGING")
         CALL staging_env_create(pint_env%staging_env, transform_section, &
                                 p=pint_env%p, kT=pint_env%kT)
      ELSE
         transform_section => section_vals_get_subs_vals(input, &
                                                         "MOTION%PINT%NORMALMODE")
         CALL normalmode_env_create(pint_env%normalmode_env, &
                                    transform_section, p=pint_env%p, kT=pint_env%kT, propagator=pint_env%propagator%prop_kind)
         wrong_input = pint_env%nrespa*twopi/(SQRT(pint_env%p/pint_env%normalmode_env%modefactor)*pint_env%kT)/pint_env%dt > 10
         IF(.NOT.(wrong_input))CALL cp__a("motion/pint_methods.F",373)
      END IF
      ALLOCATE (pint_env%mass(pint_env%ndim))
      CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id, &
                              f_env=f_env)
      CALL force_env_get(force_env=f_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, particles=particles)

!TODO length of pint_env%mass is redundant
      idim = 0
      DO iat = 1, pint_env%ndim/3
         CALL get_atomic_kind(particles%els(iat)%atomic_kind, mass=mass)
         DO idir = 1, 3
            idim = idim+1
            pint_env%mass(idim) = mass
         END DO
      END DO
      CALL f_env_rm_defaults(f_env, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/pint_methods.F",391)

      ALLOCATE (pint_env%Q(pint_env%p), &
                pint_env%mass_beads(pint_env%p, pint_env%ndim), &
                pint_env%mass_fict(pint_env%p, pint_env%ndim))
      IF (pint_env%transform == transformation_stage) THEN
         CALL staging_init_masses(pint_env%staging_env, mass=pint_env%mass, &
                                  mass_beads=pint_env%mass_beads, mass_fict=pint_env%mass_fict, &
                                  Q=pint_env%Q)
      ELSE
         CALL normalmode_init_masses(pint_env%normalmode_env, &
                                     mass=pint_env%mass, mass_beads=pint_env%mass_beads, &
                                     mass_fict=pint_env%mass_fict, Q=pint_env%Q)
      END IF

      NULLIFY (pint_env%gle)
      gle_section => section_vals_get_subs_vals(input, "MOTION%PINT%GLE")
      CALL section_vals_get(gle_section, explicit=explicit)
      IF (explicit) THEN
         ALLOCATE (pint_env%gle)
         CALL gle_init(pint_env%gle, dt=pint_env%dt/pint_env%nrespa, temp=pint_env%kT, &
                       section=gle_section)
         IF (pint_env%pimd_thermostat == thermostat_none .AND. pint_env%gle%ndim .GT. 0) THEN
            pint_env%pimd_thermostat = thermostat_gle

            ! initialize a GLE with ALL degrees of freedom on node 0,
            ! as it seems to me that here everything but force eval is replicated
            pint_env%gle%loc_num_gle = pint_env%p*pint_env%ndim
            pint_env%gle%glob_num_gle = pint_env%gle%loc_num_gle
            ALLOCATE (pint_env%gle%map_info%index(pint_env%gle%loc_num_gle))
            IF(.NOT.(stat == 0))CALL cp__a("motion/pint_methods.F",421)
            DO itmp = 1, pint_env%gle%loc_num_gle
               pint_env%gle%map_info%index(itmp) = itmp
            ENDDO
            CALL gle_thermo_create(pint_env%gle, pint_env%gle%loc_num_gle)

            ! here we should have read a_mat and c_mat;
            !we can therefore compute the matrices needed for the propagator
            ! deterministic part of the propagator
            CALL gle_matrix_exp((-pint_env%dt/pint_env%nrespa*0.5_dp)*pint_env%gle%a_mat, &
                                pint_env%gle%ndim, 15, 15, pint_env%gle%gle_t)
            ! stochastic part
            CALL gle_cholesky_stab(pint_env%gle%c_mat-MATMUL(pint_env%gle%gle_t, &
                                                             MATMUL(pint_env%gle%c_mat, TRANSPOSE(pint_env%gle%gle_t))), &
                                   pint_env%gle%gle_s, pint_env%gle%ndim)
            ! and initialize the additional momenta
            CALL pint_gle_init(pint_env)
         END IF
      END IF

      !Setup pile thermostat
      NULLIFY (pint_env%pile_therm)
      pile_section => section_vals_get_subs_vals(input, "MOTION%PINT%PILE")
      CALL section_vals_get(pile_section, explicit=explicit)
      IF (explicit) THEN
         CALL cite_reference(Ceriotti2010)
         CALL section_vals_val_get(pint_env%input, &
                                   "MOTION%PINT%INIT%THERMOSTAT_SEED", &
                                   i_val=pint_env%thermostat_rng_seed)
         IF (pint_env%pimd_thermostat == thermostat_none) THEN
            pint_env%pimd_thermostat = thermostat_pile
            CALL pint_pile_init(pile_therm=pint_env%pile_therm, &
                                pint_env=pint_env, &
                                normalmode_env=pint_env%normalmode_env, &
                                section=pile_section)
         ELSE
            CALL cp__b("motion/pint_methods.F",457,"PILE thermostat can't be used with another thermostat.")
         END IF
      END IF

      !Setup PIGLET thermostat
      NULLIFY (pint_env%piglet_therm)
      piglet_section => section_vals_get_subs_vals(input, "MOTION%PINT%PIGLET")
      CALL section_vals_get(piglet_section, explicit=explicit)
      IF (explicit) THEN
         CALL cite_reference(Ceriotti2012)
         CALL section_vals_val_get(pint_env%input, &
                                   "MOTION%PINT%INIT%THERMOSTAT_SEED", &
                                   i_val=pint_env%thermostat_rng_seed)
         IF (pint_env%pimd_thermostat == thermostat_none) THEN
            pint_env%pimd_thermostat = thermostat_piglet
            CALL pint_piglet_create(pint_env%piglet_therm, &
                                    pint_env, &
                                    piglet_section)
            CALL pint_piglet_init(pint_env%piglet_therm, &
                                  pint_env, &
                                  piglet_section, &
                                  dt=pint_env%dt, para_env=para_env)
         ELSE
            CALL cp__b("motion/pint_methods.F",480,"PILE thermostat can't be used with another thermostat.")
         END IF
      END IF

      !Initialize integrator scheme
      CALL section_vals_val_get(pint_section, "HARM_INT", i_val=pint_env%harm_integrator)
      IF (pint_env%harm_integrator == integrate_exact) THEN
         IF (pint_env%pimd_thermostat == thermostat_nose) THEN
            WRITE (UNIT=msg, FMT=*) "PINT WARNING| Nose Thermostat only avaliable in "// &
               "the numeric harmonic integrator. Switching to numeric harmonic integrator."
            CALL cp__w("motion/pint_methods.F",490,msg)
            pint_env%harm_integrator = integrate_numeric
         END IF
         IF (pint_env%pimd_thermostat == thermostat_gle) THEN
            WRITE (UNIT=msg, FMT=*) "PINT WARNING| GLE Thermostat only avaliable in "// &
               "the numeric harmonic integrator. Switching to numeric harmonic integrator."
            CALL cp__w("motion/pint_methods.F",496,msg)
            pint_env%harm_integrator = integrate_numeric
         END IF
      ELSE IF (pint_env%harm_integrator == integrate_numeric) THEN
         IF (pint_env%pimd_thermostat == thermostat_pile) THEN
            WRITE (UNIT=msg, FMT=*) "PINT WARNING| PILE Thermostat only avaliable in "// &
               "the exact harmonic integrator. Switching to exact harmonic integrator."
            CALL cp__w("motion/pint_methods.F",503,msg)
            pint_env%harm_integrator = integrate_exact
         END IF
         IF (pint_env%pimd_thermostat == thermostat_piglet) THEN
            WRITE (UNIT=msg, FMT=*) "PINT WARNING| PIGLET Thermostat only avaliable in "// &
               "the exact harmonic integrator. Switching to exact harmonic integrator."
            CALL cp__w("motion/pint_methods.F",509,msg)
            pint_env%harm_integrator = integrate_exact
         END IF
      END IF

      IF (pint_env%harm_integrator == integrate_exact) THEN
         IF (pint_env%nrespa /= 1) THEN
            pint_env%nrespa = 1
            WRITE (UNIT=msg, FMT=*) "PINT WARNING| Adjusting NRESPA to 1 for exact harmonic integration."
            CALL cp__w("motion/pint_methods.F",518,msg)
         END IF
         NULLIFY (pint_env%wsinex)
         ALLOCATE (pint_env%wsinex(pint_env%p))
         NULLIFY (pint_env%iwsinex)
         ALLOCATE (pint_env%iwsinex(pint_env%p))
         NULLIFY (pint_env%cosex)
         ALLOCATE (pint_env%cosex(pint_env%p))
         dt = pint_env%dt/REAL(pint_env%nrespa, KIND=dp)
         !Centroid mode shoud not be propagated
         pint_env%wsinex(1) = 0.0_dp
         pint_env%iwsinex(1) = dt
         pint_env%cosex(1) = 1.0_dp
         DO ibead = 2, pint_env%p
            omega = SQRT(pint_env%normalmode_env%lambda(ibead))
            pint_env%wsinex(ibead) = SIN(omega*dt)*omega
            pint_env%iwsinex(ibead) = SIN(omega*dt)/omega
            pint_env%cosex(ibead) = COS(omega*dt)
         END DO
      END IF

      CALL section_vals_val_get(pint_section, "FIX_CENTROID_POS", &
                                l_val=ltmp)
      IF (ltmp .AND. (pint_env%transform .EQ. transformation_normal)) THEN
         pint_env%first_propagated_mode = 2
      ELSE
         pint_env%first_propagated_mode = 1
      END IF

      CALL timestop(handle)

      RETURN
   END SUBROUTINE pint_create

! ***************************************************************************
!> \brief Retain a path integral environment
!> \param pint_env the pint_env to retain
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pint_retain(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",562)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",563)
      pint_env%ref_count = pint_env%ref_count+1
      RETURN
   END SUBROUTINE pint_retain

! ***************************************************************************
!> \brief Release a path integral environment
!> \param pint_env the pint_env to release
!> \par History
!>      Added normal mode transformation [hforbert]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pint_release(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pint_env)) THEN
         IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",581)
         pint_env%ref_count = pint_env%ref_count-1
         IF (pint_env%ref_count == 0) THEN
            CALL rep_env_release(pint_env%replicas)
            CALL section_vals_release(pint_env%input)
            IF (ASSOCIATED(pint_env%staging_env)) THEN
               CALL staging_release(pint_env%staging_env)
            END IF
            IF (ASSOCIATED(pint_env%normalmode_env)) THEN
               CALL normalmode_release(pint_env%normalmode_env)
            END IF
            CALL delete_rng_stream(pint_env%randomG)

            DEALLOCATE (pint_env%mass)
            DEALLOCATE (pint_env%e_pot_bead)

            DEALLOCATE (pint_env%x)
            DEALLOCATE (pint_env%v)
            DEALLOCATE (pint_env%f)
            DEALLOCATE (pint_env%external_f)
            DEALLOCATE (pint_env%mass_beads)
            DEALLOCATE (pint_env%mass_fict)
            DEALLOCATE (pint_env%ux)
            DEALLOCATE (pint_env%ux_t)
            DEALLOCATE (pint_env%uv)
            DEALLOCATE (pint_env%uv_t)
            DEALLOCATE (pint_env%uv_new)
            DEALLOCATE (pint_env%uf)
            DEALLOCATE (pint_env%uf_h)
            DEALLOCATE (pint_env%centroid)
            DEALLOCATE (pint_env%rtmp_ndim)
            DEALLOCATE (pint_env%rtmp_natom)
            DEALLOCATE (pint_env%propagator)

            IF (pint_env%harm_integrator == integrate_exact) THEN
               DEALLOCATE (pint_env%wsinex)
               DEALLOCATE (pint_env%iwsinex)
               DEALLOCATE (pint_env%cosex)
            END IF

            SELECT CASE (pint_env%pimd_thermostat)
            CASE (thermostat_nose)
               DEALLOCATE (pint_env%tx)
               DEALLOCATE (pint_env%tv)
               DEALLOCATE (pint_env%tv_t)
               DEALLOCATE (pint_env%tv_old)
               DEALLOCATE (pint_env%tv_new)
               DEALLOCATE (pint_env%tf)
            CASE (thermostat_gle)
               CALL gle_dealloc(pint_env%gle)
            CASE (thermostat_pile)
               CALL pint_pile_release(pint_env%pile_therm)
            CASE (thermostat_piglet)
               CALL pint_piglet_release(pint_env%piglet_therm)
            END SELECT

            DEALLOCATE (pint_env%Q)

            DEALLOCATE (pint_env)

         END IF
      END IF

      NULLIFY (pint_env)

      RETURN
   END SUBROUTINE pint_release

! ***************************************************************************
!> \brief Tests the path integral methods
!> \param para_env parallel environment
!> \param input the input to test
!> \param input_declaration ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_test(para_env, input, input_declaration)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ib, idim, unit_nr
      REAL(kind=dp)                                      :: c, e_h, err
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: x1
      TYPE(pint_env_type), POINTER                       :: pint_env

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("motion/pint_methods.F",668)
      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("motion/pint_methods.F",669)
      IF(.NOT.(para_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",670)
      IF(.NOT.(input%ref_count > 0))CALL cp__a("motion/pint_methods.F",671)
      NULLIFY (pint_env)
      unit_nr = cp_logger_get_default_io_unit()
      CALL pint_create(pint_env, input, input_declaration, para_env)
      IF (ASSOCIATED(pint_env)) THEN
         ALLOCATE (x1(pint_env%ndim, pint_env%p))
         x1(:, :) = pint_env%x
         CALL pint_x2u(pint_env)
         pint_env%x = 0._dp
         CALL pint_u2x(pint_env)
         err = 0._dp
         DO i = 1, pint_env%ndim
            err = MAX(err, ABS(x1(1, i)-pint_env%x(1, i)))
         END DO
         IF (unit_nr > 0) WRITE (unit_nr, *) "diff_r1="//cp_to_string(err)

         CALL pint_calc_uf_h(pint_env, e_h=e_h)
         c = -pint_env%staging_env%w_p**2
         pint_env%f = 0._dp
         DO idim = 1, pint_env%ndim
            DO ib = 1, pint_env%p
               pint_env%f(ib, idim) = pint_env%f(ib, idim)+ &
                                      c*(2._dp*pint_env%x(ib, idim) &
                                         -pint_env%x(MODULO(ib-2, pint_env%p)+1, idim) &
                                         -pint_env%x(MODULO(ib, pint_env%p)+1, idim))
            END DO
         END DO
         CALL pint_f2uf(pint_env)
         err = 0._dp
         DO idim = 1, pint_env%ndim
            DO ib = 1, pint_env%p
               err = MAX(err, ABS(pint_env%uf(ib, idim)-pint_env%uf_h(ib, idim)))
            END DO
         END DO
         IF (unit_nr > 0) WRITE (unit_nr, *) "diff_f_h="//cp_to_string(err)
      END IF
      RETURN
   END SUBROUTINE pint_test

! ***************************************************************************
!> \brief  Perform a path integral simulation
!> \param  para_env parallel environment
!> \param  input the input to test
!> \param input_declaration ...
!> \param globenv ...
!> \par    History
!>         2003-11 created [fawzi]
!>         2009-12-14 globenv parameter added to handle soft exit
!>           requests [lwalewski]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE do_pint_run(para_env, input, input_declaration, globenv)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'do_pint_run', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: helium_only_mid = 1, &
                                                            int_pot_scan_mid = 4, &
                                                            solute_only_mid = 2, &
                                                            solute_with_helium_mid = 3

      CHARACTER(len=default_string_length)               :: stmp
      INTEGER                                            :: handle, mode
      LOGICAL                                            :: explicit, helium_only, int_pot_scan, &
                                                            solvent_present
      TYPE(helium_solvent_type), POINTER                 :: helium
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(section_vals_type), POINTER                   :: helium_section

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("motion/pint_methods.F",744)
      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("motion/pint_methods.F",745)
      IF(.NOT.(para_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",746)
      IF(.NOT.(input%ref_count > 0))CALL cp__a("motion/pint_methods.F",747)

      ! check if helium solvent is present
      NULLIFY (helium_section)
      helium_section => section_vals_get_subs_vals(input, &
                                                   "MOTION%PINT%HELIUM")
      CALL section_vals_get(helium_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(helium_section, "_SECTION_PARAMETERS_", &
                                   l_val=solvent_present)
      ELSE
         solvent_present = .FALSE.
      END IF

      ! check if there is anything but helium
      IF (solvent_present) THEN
         CALL section_vals_val_get(helium_section, "HELIUM_ONLY", &
                                   l_val=helium_only)
      ELSE
         helium_only = .FALSE.
      END IF

      ! check wheather to perform solute-helium interaction pot scan
      IF (solvent_present) THEN
         CALL section_vals_val_get(helium_section, "INTERACTION_POT_SCAN", &
                                   l_val=int_pot_scan)
      ELSE
         int_pot_scan = .FALSE.
      END IF

      ! input consistency check
      IF (helium_only .AND. int_pot_scan) THEN
         stmp = "Options HELIUM_ONLY and INTERACTION_POT_SCAN are exclusive"
         CALL cp__b("motion/pint_methods.F",780,stmp)
      END IF

      ! select mode of operation
      mode = 0
      IF (solvent_present) THEN
         IF (helium_only) THEN
            mode = helium_only_mid
         ELSE
            IF (int_pot_scan) THEN
               mode = int_pot_scan_mid
            ELSE
               mode = solute_with_helium_mid
            END IF
         END IF
      ELSE
         mode = solute_only_mid
      END IF

      NULLIFY (pint_env, helium)

      ! perform the simulation according to the chosen mode
      SELECT CASE (mode)

      CASE (helium_only_mid)
         CALL helium_create(helium, input)
         CALL helium_init(helium, pint_env)
         CALL helium_do_run(helium, globenv)
         CALL helium_release(helium)

      CASE (solute_only_mid)
         CALL pint_create(pint_env, input, input_declaration, para_env)
         CALL pint_init(pint_env)
         CALL pint_do_run(pint_env, globenv)
         CALL pint_release(pint_env)

      CASE (int_pot_scan_mid)
         CALL pint_create(pint_env, input, input_declaration, para_env)
! TODO only initialization of positions is necessary, but rep_env_calc_e_f called
! from within pint_init_f does something to the replica environments which can not be
! avoided (has something to do with f_env_add_defaults) so leaving for now..
         CALL pint_init(pint_env)
         CALL helium_create(helium, input, solute=pint_env)
         CALL pint_run_scan(pint_env, helium)
         CALL helium_release(helium)
         CALL pint_release(pint_env)

      CASE (solute_with_helium_mid)
         CALL pint_create(pint_env, input, input_declaration, para_env)
         ! init pint wihtout helium forces (they are not yet initialized)
         CALL pint_init(pint_env)
         ! init helium with solute's positions (they are already initialized)
         CALL helium_create(helium, input, solute=pint_env)
         CALL helium_init(helium, pint_env)
         ! reinit pint forces with helium forces (they are now initialized)
         CALL pint_init_f(pint_env, helium_solvent=helium)

         CALL pint_do_run(pint_env, globenv, helium=helium)
         CALL helium_release(helium)
         CALL pint_release(pint_env)

      CASE DEFAULT
         CALL cp__b("motion/pint_methods.F",842,"Unknown mode ("//TRIM(ADJUSTL(cp_to_string(mode)))//")")
      END SELECT

      CALL timestop(handle)

      RETURN
   END SUBROUTINE do_pint_run

! ***************************************************************************
!> \brief  Reads the restart, initializes the beads, etc.
!> \param pint_env ...
!> \par    History
!>           11.2003 created [fawzi]
!>           actually ASSIGN input pointer [hforbert]
!>           2010-12-16 turned into a wrapper routine [lwalewski]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pint_init(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_init', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",865)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",866)

      CALL pint_init_x(pint_env)
      CALL pint_init_v(pint_env)
      CALL pint_init_t(pint_env)
      CALL pint_init_f(pint_env)

      RETURN
   END SUBROUTINE pint_init

! ***************************************************************************
!> \brief  Assign initial postions to the beads.
!> \param pint_env ...
!> \date   2010-12-15
!> \author Lukasz Walewski
!> \note  Initialization is done in the following way:
!>           1. assign all beads with the same classical positions from
!>              FORCE_EVAL (hot start)
!>           2. spread the beads around classical positions as if they were
!>              free particles (if requested)
!>           3. replace positions generated in steps 1-2 with the explicit
!>              ones if they are explicitly given in the input structure
!>           4. apply Gaussian noise to the positions generated so far (if
!>              requested)
! **************************************************************************************************
   SUBROUTINE pint_init_x(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_init_x', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msg, tmp
      INTEGER                                            :: ia, ib, ic, idim, input_seed, n_rep_val
      LOGICAL                                            :: done_init, done_levy, done_rand, &
                                                            explicit, levycorr, ltmp
      REAL(kind=dp)                                      :: tcorr, var
      REAL(kind=dp), DIMENSION(3)                        :: x0
      REAL(kind=dp), DIMENSION(3, 2)                     :: seed
      REAL(kind=dp), DIMENSION(:), POINTER               :: bx, r_vals
      TYPE(rng_stream_type), POINTER                     :: rng_gaussian
      TYPE(section_vals_type), POINTER                   :: input_section

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",908)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",909)

      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            pint_env%x(ib, idim) = pint_env%replicas%r(idim, ib)
         END DO
      END DO

      done_levy = .FALSE.
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%LEVY_POS_SAMPLE", &
                                l_val=ltmp)
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%LEVY_TEMP_FACTOR", &
                                r_val=tcorr)
      IF (ltmp) THEN

         NULLIFY (bx)
         ALLOCATE (bx(3*pint_env%p))
         NULLIFY (rng_gaussian)
         CALL section_vals_val_get(pint_env%input, &
                                   "MOTION%PINT%INIT%LEVY_SEED", i_val=input_seed)
         seed(:, :) = REAL(input_seed, KIND=dp)
!      seed(:,:) = next_rng_seed()
         CALL create_rng_stream(rng_gaussian, &
                                name="tmp_rng_gaussian", &
                                distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE., &
                                seed=seed)

         CALL section_vals_val_get(pint_env%input, &
                                   "MOTION%PINT%INIT%LEVY_CORRELATED", &
                                   l_val=levycorr)

         IF (levycorr) THEN

            ! correlated Levy walk - the same path for all atoms
            x0 = (/0.0_dp, 0.0_dp, 0.0_dp/)
            CALL pint_levy_walk(x0, pint_env%p, 1.0_dp, bx, rng_gaussian)
            idim = 0
            DO ia = 1, pint_env%ndim/3
               var = SQRT(1.0_dp/(pint_env%kT*tcorr*pint_env%mass(3*ia)))
               DO ic = 1, 3
                  idim = idim+1
                  DO ib = 1, pint_env%p
                     pint_env%x(ib, idim) = pint_env%x(ib, idim)+bx(3*(ib-1)+ic)*var
                  END DO
               END DO
            END DO

         ELSE

            ! uncorrelated bead initialization - distinct Levy walk for each atom
            idim = 0
            DO ia = 1, pint_env%ndim/3
               x0(1) = pint_env%x(1, 3*(ia-1)+1)
               x0(2) = pint_env%x(1, 3*(ia-1)+2)
               x0(3) = pint_env%x(1, 3*(ia-1)+3)
               var = SQRT(1.0_dp/(pint_env%kT*tcorr*pint_env%mass(3*ia)))
               CALL pint_levy_walk(x0, pint_env%p, var, bx, rng_gaussian)
               DO ic = 1, 3
                  idim = idim+1
                  DO ib = 1, pint_env%p
                     pint_env%x(ib, idim) = pint_env%x(ib, idim)+bx(3*(ib-1)+ic)
                  END DO
               END DO
            END DO

         END IF

         CALL delete_rng_stream(rng_gaussian)
         DEALLOCATE (bx)
         done_levy = .TRUE.
      END IF

      done_init = .FALSE.
      NULLIFY (input_section)
      input_section => section_vals_get_subs_vals(pint_env%input, &
                                                  "MOTION%PINT%BEADS%COORD")
      CALL section_vals_get(input_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep_val)
         IF (n_rep_val > 0) THEN
            IF(.NOT.(n_rep_val == 1))CALL cp__a("motion/pint_methods.F",993)
            CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                      r_vals=r_vals)
            IF (SIZE(r_vals) /= pint_env%p*pint_env%ndim) &
               CALL cp__b("motion/pint_methods.F",997,"Invalid size of MOTION%PINT%BEADS%COORD")
            ic = 0
            DO idim = 1, pint_env%ndim
               DO ib = 1, pint_env%p
                  ic = ic+1
                  pint_env%x(ib, idim) = r_vals(ic)
               END DO
            END DO
            done_init = .TRUE.
         END IF
      END IF

      done_rand = .FALSE.
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%RANDOMIZE_POS", &
                                l_val=ltmp)
      IF (ltmp) THEN
         DO idim = 1, pint_env%ndim
            DO ib = 1, pint_env%p
               pint_env%x(ib, idim) = pint_env%x(ib, idim)+ &
                                      next_random_number(rng_stream=pint_env%randomG, &
                                                         variance=pint_env%beta/ &
                                                         SQRT(12.0_dp*pint_env%mass(idim)))
            END DO
         END DO
         done_rand = .TRUE.
      END IF

      WRITE (tmp, '(A)') "Bead positions initialization:"
      IF (done_init) THEN
         WRITE (msg, '(A,A)') TRIM(tmp), " input structure"
      ELSE IF (done_levy) THEN
         WRITE (msg, '(A,A)') TRIM(tmp), " Levy random walk"
      ELSE
         WRITE (msg, '(A,A)') TRIM(tmp), " hot start"
      END IF
      CALL pint_write_line(msg)

      IF (done_levy) THEN
         WRITE (msg, '(A,F6.3)') "Levy walk at effective temperature: ", tcorr
      END IF

      IF (done_rand) THEN
         WRITE (msg, '(A)') "Added gaussian noise to the positions of the beads."
         CALL pint_write_line(msg)
      END IF

      RETURN
   END SUBROUTINE pint_init_x

! ***************************************************************************
!> \brief  Initialize velocities
!> \param  pint_env the pint env in which you should initialize the
!>         velocity
!> \par    History
!>         2010-12-16 gathered all velocity-init code here [lwalewski]
!>         2011-04-05 added centroid velocity initialization [lwalewski]
!>         2011-12-19 removed optional parameter kT, target temperature is
!>                    now determined from the input directly [lwalewski]
!> \author fawzi
!> \note   Initialization is done according to the following protocol:
!>         1. set all the velocities to FORCE_EVAL%SUBSYS%VELOCITY if present
!>         2. scale the velocities according to the actual temperature
!>            (has no effect if vels not present in 1.)
!>         3. draw vels for the remaining dof from MB distribution
!>            (all or non-centroid modes only depending on 1.)
!>         4. add random noise to the centroid vels if CENTROID_SPEED == T
!>         5. set the vels for all dof to 0.0 if VELOCITY_QUENCH == T
!>         6. set the vels according to the explicit values from the input
!>            if present
! **************************************************************************************************
   SUBROUTINE pint_init_v(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_init_v', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: msg, stmp, stmp1, stmp2, unit_str
      INTEGER                                            :: first_mode, ia, ib, ic, idim, itmp, &
                                                            n_rep_val
      LOGICAL                                            :: done_init, done_quench, done_scale, &
                                                            done_sped, explicit, ltmp, vels_present
      REAL(kind=dp)                                      :: actual_t, ek, rtmp, target_t, unit_conv
      REAL(kind=dp), DIMENSION(:), POINTER               :: r_vals
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: input_section

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",1083)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",1084)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! read the velocities from the input file if they are given explicitly
      vels_present = .FALSE.
      NULLIFY (input_section)
      input_section => section_vals_get_subs_vals(pint_env%input, &
                                                  "FORCE_EVAL%SUBSYS%VELOCITY")
      CALL section_vals_get(input_section, explicit=explicit)
      IF (explicit) THEN

         CALL section_vals_val_get(input_section, "PINT_UNIT", &
                                   c_val=unit_str)
         unit_conv = cp_unit_to_cp2k(1.0_dp, TRIM(unit_str))

         ! assign all the beads with the same velocities from FORCE_EVAL%SUBSYS%VELOCITY
         NULLIFY (r_vals)
         CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep_val)
         stmp = ""
         WRITE (stmp, *) n_rep_val
         msg = "Invalid number of atoms in FORCE_EVAL%SUBSYS%VELOCITY ("// &
               TRIM(ADJUSTL(stmp))//")."
         IF (3*n_rep_val /= pint_env%ndim) &
            CALL cp__b("motion/pint_methods.F",1110,msg)
         DO ia = 1, pint_env%ndim/3
            CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                      i_rep_val=ia, r_vals=r_vals)
            itmp = SIZE(r_vals)
            stmp = ""
            WRITE (stmp, *) itmp
            msg = "Number of coordinates != 3 in FORCE_EVAL%SUBSYS%VELOCITY ("// &
                  TRIM(ADJUSTL(stmp))//")."
            IF (itmp /= 3) &
               CALL cp__b("motion/pint_methods.F",1120,msg)
            DO ib = 1, pint_env%p
               DO ic = 1, 3
                  idim = 3*(ia-1)+ic
                  pint_env%v(ib, idim) = r_vals(ic)*unit_conv
               END DO
            END DO
         END DO

         vels_present = .TRUE.
      END IF

      ! set the actual temperature...
      IF (vels_present) THEN
         ! ...from the initial velocities
         ek = 0.0_dp
         DO ia = 1, pint_env%ndim/3
            rtmp = 0.0_dp
            DO ic = 1, 3
               idim = 3*(ia-1)+ic
               rtmp = rtmp+pint_env%v(1, idim)*pint_env%v(1, idim)
            END DO
            ek = ek+0.5_dp*pint_env%mass(idim)*rtmp
         END DO
         actual_t = 2.0_dp*ek/pint_env%ndim
      ELSE
         ! ...using the temperature value from the input
         actual_t = pint_env%kT
      END IF

      ! set the target temperature
      target_t = pint_env%kT
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%VELOCITY_SCALE", &
                                l_val=done_scale)
      IF (vels_present) THEN
         IF (done_scale) THEN
            ! rescale the velocities to match the target temperature
            rtmp = SQRT(target_t/actual_t)
            DO ia = 1, pint_env%ndim/3
               DO ib = 1, pint_env%p
                  DO ic = 1, 3
                     idim = 3*(ia-1)+ic
                     pint_env%v(ib, idim) = rtmp*pint_env%v(ib, idim)
                  END DO
               END DO
            END DO
         ELSE
            target_t = actual_t
         END IF
      END IF

      ! draw velocities from the M-B distribution...
      IF (vels_present) THEN
         ! ...for non-centroid modes only
         CALL pint_x2u(pint_env, ux=pint_env%uv, x=pint_env%v)
         first_mode = 2
      ELSE
         ! ...for all the modes
         first_mode = 1
      END IF
      DO idim = 1, SIZE(pint_env%uv, 2)
         DO ib = first_mode, SIZE(pint_env%uv, 1)
            pint_env%uv(ib, idim) = &
               next_random_number(rng_stream=pint_env%randomG, &
                                  variance=target_t/pint_env%mass_fict(ib, idim))
         END DO
      END DO

      ! add random component to the centroid velocity if requsted
      done_sped = .FALSE.
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%CENTROID_SPEED", &
                                l_val=ltmp)
      IF (ltmp) THEN
         CALL pint_u2x(pint_env, ux=pint_env%uv, x=pint_env%v)
         DO idim = 1, pint_env%ndim
            rtmp = next_random_number(rng_stream=pint_env%randomG, &
                                      variance=pint_env%mass(idim)*pint_env%kT) &
                   /pint_env%mass(idim)
            DO ib = 1, pint_env%p
               pint_env%v(ib, idim) = pint_env%v(ib, idim)+rtmp
            END DO
         END DO
         CALL pint_x2u(pint_env, ux=pint_env%uv, x=pint_env%v)
         done_sped = .TRUE.
      END IF

      ! quench (set to zero) velocities for all the modes if requested
      ! (disregard all the initialization done so far)
      done_quench = .FALSE.
      CALL section_vals_val_get(pint_env%input, &
                                "MOTION%PINT%INIT%VELOCITY_QUENCH", &
                                l_val=ltmp)
      IF (ltmp) THEN
         DO idim = 1, pint_env%ndim
            DO ib = 1, pint_env%p
               pint_env%v(ib, idim) = 0.0_dp
            END DO
         END DO
         CALL pint_x2u(pint_env, ux=pint_env%uv, x=pint_env%v)
         done_quench = .TRUE.
      END IF

      ! set the velocities to the values from the input if they are explicit
      ! (disregard all the initialization done so far)
      done_init = .FALSE.
      NULLIFY (input_section)
      input_section => section_vals_get_subs_vals(pint_env%input, &
                                                  "MOTION%PINT%BEADS%VELOCITY")
      CALL section_vals_get(input_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep_val)
         IF (n_rep_val > 0) THEN
            IF(.NOT.(n_rep_val == 1))CALL cp__a("motion/pint_methods.F",1235)
            CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                      r_vals=r_vals)
            IF (SIZE(r_vals) /= pint_env%p*pint_env%ndim) &
               CALL cp__b("motion/pint_methods.F",1239,"Invalid size of MOTION%PINT%BEAD%VELOCITY")
            itmp = 0
            DO idim = 1, pint_env%ndim
               DO ib = 1, pint_env%p
                  itmp = itmp+1
                  pint_env%v(ib, idim) = r_vals(itmp)
               END DO
            END DO
            CALL pint_x2u(pint_env, ux=pint_env%uv, x=pint_env%v)
            done_init = .TRUE.
         END IF
      END IF

      unit_conv = cp_unit_from_cp2k(1.0_dp, "K")
      WRITE (stmp1, '(F10.2)') target_t*pint_env%propagator%temp_sim2phys*unit_conv
      msg = "Bead velocities initialization:"
      IF (done_init) THEN
         msg = TRIM(msg)//" input structure"
      ELSE IF (done_quench) THEN
         msg = TRIM(msg)//" quenching (set to 0.0)"
      ELSE
         IF (vels_present) THEN
            msg = TRIM(ADJUSTL(msg))//" centroid +"
         END IF
         msg = TRIM(ADJUSTL(msg))//" Maxwell-Boltzmann at "//TRIM(ADJUSTL(stmp1))//" K."
      END IF
      CALL pint_write_line(msg)

      IF (done_init .AND. done_quench) THEN
         msg = "WARNING: exclusive options requested (velocity restart and quenching)"
         CALL cp__w("motion/pint_methods.F",1269,msg)
         msg = "WARNING: velocity restart took precedence"
         CALL cp__w("motion/pint_methods.F",1271,msg)
      END IF

      IF ((.NOT. done_init) .AND. (.NOT. done_quench)) THEN
         IF (vels_present .AND. done_scale) THEN
            WRITE (stmp1, '(F10.2)') actual_t*unit_conv
            WRITE (stmp2, '(F10.2)') target_t*unit_conv
            msg = "Scaled initial velocities from "//TRIM(ADJUSTL(stmp1))// &
                  " to "//TRIM(ADJUSTL(stmp2))//" K as requested."
            CALL cp__w("motion/pint_methods.F",1280,msg)
         END IF
         IF (done_sped) THEN
            msg = "Added random component to the initial centroid velocities."
            CALL cp__w("motion/pint_methods.F",1284,msg)
         END IF
      END IF

      RETURN
   END SUBROUTINE pint_init_v

! ***************************************************************************
!> \brief  Assign initial postions and velocities to the thermostats.
!> \param pint_env ...
!> \param kT ...
!> \date   2010-12-15
!> \author Lukasz Walewski
!> \note   Extracted from pint_init
! **************************************************************************************************
   SUBROUTINE pint_init_t(pint_env, kT)

      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), INTENT(in), OPTIONAL                :: kT

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_init_t', routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim, ii, inos, n_rep_val
      LOGICAL                                            :: explicit, gle_restart
      REAL(kind=dp)                                      :: mykt
      REAL(kind=dp), DIMENSION(:), POINTER               :: r_vals
      TYPE(section_vals_type), POINTER                   :: input_section

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",1312)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",1313)

      IF (pint_env%pimd_thermostat == thermostat_nose) THEN

         mykt = pint_env%kT
         IF (PRESENT(kT)) mykt = kT
         DO idim = 1, SIZE(pint_env%tv, 3)
            DO ib = 1, SIZE(pint_env%tv, 2)
               DO inos = 1, SIZE(pint_env%tv, 1)
                  pint_env%tv(inos, ib, idim) = &
                     next_random_number(rng_stream=pint_env%randomG, &
                                        variance=mykt/pint_env%Q(ib))
               END DO
            END DO
         END DO

         NULLIFY (input_section)
         input_section => section_vals_get_subs_vals(pint_env%input, &
                                                     "MOTION%PINT%NOSE%COORD")
         CALL section_vals_get(input_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=n_rep_val)
            IF (n_rep_val > 0) THEN
               IF(.NOT.(n_rep_val == 1))CALL cp__a("motion/pint_methods.F",1337)
               CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                         r_vals=r_vals)
               IF (SIZE(r_vals) /= pint_env%p*pint_env%ndim*pint_env%nnos) &
                  CALL cp__b("motion/pint_methods.F",1341,"Invalid size of MOTION%PINT%NOSE%COORD")
               ii = 0
               DO idim = 1, pint_env%ndim
                  DO ib = 1, pint_env%p
                     DO inos = 1, pint_env%nnos
                        ii = ii+1
                        pint_env%tx(inos, ib, idim) = r_vals(ii)
                     END DO
                  END DO
               END DO
            END IF
         END IF

         NULLIFY (input_section)
         input_section => section_vals_get_subs_vals(pint_env%input, &
                                                     "MOTION%PINT%NOSE%VELOCITY")
         CALL section_vals_get(input_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=n_rep_val)
            IF (n_rep_val > 0) THEN
               IF(.NOT.(n_rep_val == 1))CALL cp__a("motion/pint_methods.F",1362)
               CALL section_vals_val_get(input_section, "_DEFAULT_KEYWORD_", &
                                         r_vals=r_vals)
               IF (SIZE(r_vals) /= pint_env%p*pint_env%ndim*pint_env%nnos) &
                  CALL cp__b("motion/pint_methods.F",1366,"Invalid size of MOTION%PINT%NOSE%VELOCITY")
               ii = 0
               DO idim = 1, pint_env%ndim
                  DO ib = 1, pint_env%p
                     DO inos = 1, pint_env%nnos
                        ii = ii+1
                        pint_env%tv(inos, ib, idim) = r_vals(ii)
                     END DO
                  END DO
               END DO
            END IF
         END IF

      ELSEIF (pint_env%pimd_thermostat == thermostat_gle) THEN
         NULLIFY (input_section)
         input_section => section_vals_get_subs_vals(pint_env%input, &
                                                     "MOTION%PINT%GLE")
         CALL section_vals_get(input_section, explicit=explicit)
         IF (explicit) THEN
            CALL restart_gle(pint_env%gle, input_section, save_mem=.FALSE., &
                             restart=gle_restart)
         END IF
      END IF

      RETURN
   END SUBROUTINE pint_init_t

! ***************************************************************************
!> \brief  Prepares the forces, etc. to perform an MD step
!> \param pint_env ...
!> \param helium_solvent ...
!> \par    History
!>           Added nh_energy calculation [hforbert]
!>           Bug fixes for no thermostats [hforbert]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_init_f(pint_env, helium_solvent)
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium_solvent

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_init_f', routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim, inos
      REAL(kind=dp)                                      :: e_h

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",1411)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",1412)
      CALL pint_x2u(pint_env)
      CALL pint_calc_uf_h(pint_env=pint_env, e_h=e_h)
      CALL pint_calc_f(pint_env)

      ! add helium forces to the solute's internal ones
      ! Assume that helium has been already initialized and helium_solvent
      ! contains proper forces in force_avrg array.
      IF (PRESENT(helium_solvent)) THEN
         IF (ASSOCIATED(helium_solvent)) THEN
            pint_env%f(:, :) = pint_env%f(:, :)+helium_solvent%force_avrg(:, :)
         END IF
      END IF
      CALL pint_f2uf(pint_env)

      ! set the centroid forces to 0 if FIX_CENTROID_POS
      IF (pint_env%first_propagated_mode .EQ. 2) THEN
         pint_env%uf(1, :) = 0.0_dp
      END IF

      CALL pint_calc_e_kin_beads_u(pint_env)
      CALL pint_calc_e_vir(pint_env)
      DO idim = 1, SIZE(pint_env%uf_h, 2)
         DO ib = pint_env%first_propagated_mode, SIZE(pint_env%uf_h, 1)
            pint_env%uf(ib, idim) = REAL(pint_env%nrespa, dp)*pint_env%uf(ib, idim)
         END DO
      END DO

      IF (pint_env%nnos > 0) THEN
         DO idim = 1, SIZE(pint_env%uf_h, 2)
            DO ib = 1, SIZE(pint_env%uf_h, 1)
               pint_env%tf(1, ib, idim) = (pint_env%mass_fict(ib, idim)* &
                                           pint_env%uv(ib, idim)**2-pint_env%kT)/pint_env%Q(ib)
            END DO
         END DO

         DO idim = 1, pint_env%ndim
            DO ib = 1, pint_env%p
               DO inos = 1, pint_env%nnos-1
                  pint_env%tf(inos+1, ib, idim) = pint_env%tv(inos, ib, idim)**2- &
                                                  pint_env%kT/pint_env%Q(ib)
               END DO
               DO inos = 1, pint_env%nnos-1
                  pint_env%tf(inos, ib, idim) = pint_env%tf(inos, ib, idim) &
                                                -pint_env%tv(inos, ib, idim)*pint_env%tv(inos+1, ib, idim)
               END DO
            END DO
         END DO
         CALL pint_calc_nh_energy(pint_env)
      END IF
      RETURN
   END SUBROUTINE pint_init_f

! ***************************************************************************
!> \brief  Perform the PIMD simulation (main MD loop)
!> \param pint_env ...
!> \param globenv ...
!> \param helium ...
!> \par    History
!>         2003-11 created [fawzi]
!>         renamed from pint_run to pint_do_run because of conflicting name
!>           of pint_run in input_constants [hforbert]
!>         2009-12-14 globenv parameter added to handle soft exit
!>           requests [lwalewski]
!> \author Fawzi Mohamed
!> \note   Everything should be read for an md step.
! **************************************************************************************************
   SUBROUTINE pint_do_run(pint_env, globenv, helium)
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_do_run', routineP = moduleN//':'//routineN

      INTEGER                                            :: step
      LOGICAL                                            :: should_stop
      REAL(kind=dp)                                      :: scal

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",1490)

      NULLIFY (pint_env%logger)
      pint_env%logger => cp_get_default_logger()

      ! create iteration level and initialize iteration info
      CALL cp_add_iter_level(pint_env%logger%iter_info, "MD")
      CALL cp_iterate(pint_env%logger%iter_info, iter_nr=pint_env%first_step)
      pint_env%iter = pint_env%first_step

      IF (PRESENT(helium)) THEN
         IF (ASSOCIATED(helium)) THEN
            ! set the properties accumulated over the whole MC process to 0
            helium%proarea%accu(:) = 0.0_dp
            helium%prarea2%accu(:) = 0.0_dp
            helium%wnmber2%accu(:) = 0.0_dp
            helium%mominer%accu(:) = 0.0_dp
            IF (helium%rho_present) THEN
               helium%rho_accu(:, :, :, :) = 0.0_dp
            END IF
            IF (helium%rdf_present) THEN
               helium%rdf_accu(:, :, :) = 0.0_dp
            END IF
         END IF
      END IF

      ! write the properties at 0-th step
      CALL pint_calc_energy(pint_env)
      CALL pint_write_ener(pint_env)
      CALL pint_write_centroids(pint_env)
      CALL pint_write_trajectory(pint_env)
      CALL pint_write_com(pint_env)
      CALL pint_write_rgyr(pint_env)

      ! main PIMD loop
      DO step = 1, pint_env%num_steps

         pint_env%iter = pint_env%iter+1
         CALL cp_iterate(pint_env%logger%iter_info, &
                         last=(step == pint_env%num_steps), &
                         iter_nr=pint_env%iter)
         pint_env%t = pint_env%t+pint_env%dt

         IF (pint_env%t_tol > 0.0_dp) THEN
            IF (ABS(2._dp*pint_env%e_kin_beads/(pint_env%p*pint_env%ndim) &
                    -pint_env%kT) > pint_env%t_tol) THEN
               scal = SQRT(pint_env%kT*(pint_env%p*pint_env%ndim)/(2.0_dp*pint_env%e_kin_beads))
               pint_env%uv = scal*pint_env%uv
               CALL pint_init_f(pint_env, helium_solvent=helium)
            END IF
         END IF
         CALL pint_step(pint_env, helium_solvent=helium)

         CALL pint_write_ener(pint_env)
         CALL pint_write_centroids(pint_env)
         CALL pint_write_trajectory(pint_env)
         CALL pint_write_com(pint_env)
         CALL pint_write_rgyr(pint_env)

         CALL write_restart(root_section=pint_env%input, &
                            pint_env=pint_env, helium_env=helium)

         ! exit from the main loop if soft exit has been requested
         CALL external_control(should_stop, "MD", globenv=globenv)
         IF (should_stop) EXIT

      END DO

      ! remove iteration level
      CALL cp_rm_iter_level(pint_env%logger%iter_info, "MD")

      RETURN
   END SUBROUTINE pint_do_run

! ***************************************************************************
!> \brief  Performs a scan of the helium-solute interaction energy
!> \param pint_env ...
!> \param helium ...
!> \date   2013-11-26
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_run_scan(pint_env, helium)
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), POINTER                 :: helium

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_run_scan', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: comment
      INTEGER                                            :: unit_nr
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: DATA
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (pint_env%logger, print_key)
      pint_env%logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(helium%input, &
                                              "MOTION%PINT%HELIUM%PRINT%RHO")

      ! perform the actual scan wrt the COM of the solute
      helium%rho_inst(1, :, :, :) = 0.0_dp
      CALL helium_intpot_scan(pint_env, helium)

      ! output the interaction potential into a cubefile
      IF (pint_env%logger%para_env%ionode) THEN

         unit_nr = cp_print_key_unit_nr( &
                   pint_env%logger, &
                   print_key, &
                   middle_name="helium-pot", &
                   extension=".cube", &
                   file_position="REWIND", &
                   do_backup=.FALSE.)

         comment = "Solute - helium interaction potential"
         NULLIFY (DATA)
         DATA => helium%rho_inst(1, :, :, :)
         CALL helium_write_cubefile( &
            unit_nr, &
            comment, &
            helium%center-0.5_dp*(helium%rho_maxr-helium%rho_delr), &
            helium%rho_delr, &
            helium%rho_nbin, &
            DATA)

         CALL m_flush(unit_nr)
         CALL cp_print_key_finished_output(unit_nr, pint_env%logger, print_key)

      END IF

      ! output solute positions
      CALL pint_write_centroids(pint_env)
      CALL pint_write_trajectory(pint_env)

      RETURN
   END SUBROUTINE pint_run_scan

! ***************************************************************************
!> \brief  Does an MD step (and nrespa harmonic evaluations)
!> \param pint_env ...
!> \param helium_solvent ...
!> \par    History
!>           various bug fixes [hforbert]
!>           10.2015 Added RPMD propagator and harmonic integrator [Felix Uhl]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_step(pint_env, helium_solvent)
      TYPE(pint_env_type), POINTER                       :: pint_env
      TYPE(helium_solvent_type), OPTIONAL, POINTER       :: helium_solvent

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_step', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ib, idim, inos, iresp
      REAL(kind=dp)                                      :: dti, dti2, dti22, e_h, rn, tdti, &
                                                            time_start, time_stop, tol
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: tmp

      CALL timeset(routineN, handle)
      time_start = m_walltime()

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",1648)

      rn = REAL(pint_env%nrespa, dp)
      dti = pint_env%dt/rn
      dti2 = dti/2._dp
      tdti = 2.*dti
      dti22 = dti**2/2._dp

      SELECT CASE (pint_env%harm_integrator)
      CASE (integrate_numeric)
         DO iresp = 1, pint_env%nrespa
            ! integrate bead positions, first_propagated_mode = { 1, 2 }
            ! Nose needs an extra step
            IF (pint_env%pimd_thermostat == thermostat_nose) THEN
               DO i = pint_env%first_propagated_mode, pint_env%p
                  pint_env%ux(i, :) = pint_env%ux(i, :)- &
                                      dti22*pint_env%uv(i, :)*pint_env%tv(1, i, :)
               END DO
               pint_env%tx = pint_env%tx+dti*pint_env%tv+dti22*pint_env%tf
            END IF
            !Integrate position in harmonic springs (uf_h) and physical potential
            !(uf)
            DO i = pint_env%first_propagated_mode, pint_env%p
               pint_env%ux_t(i, :) = pint_env%ux(i, :)+ &
                                     dti*pint_env%uv(i, :)+ &
                                     dti22*(pint_env%uf_h(i, :)+ &
                                            pint_env%uf(i, :))
            END DO

            ! apply thermostats to velocities
            SELECT CASE (pint_env%pimd_thermostat)
            CASE (thermostat_nose)
               pint_env%uv_t = pint_env%uv-dti2* &
                               pint_env%uv*pint_env%tv(1, :, :)
               tmp => pint_env%tv_t
               pint_env%tv_t => pint_env%tv
               pint_env%tv => tmp
               pint_env%tv = pint_env%tv_old+tdti*pint_env%tf
               pint_env%tv_old = pint_env%tv_t
               pint_env%tv_t = pint_env%tv_t+dti2*pint_env%tf
            CASE DEFAULT
               pint_env%uv_t = pint_env%uv
            END SELECT

            !Integrate hamonic velocities and physical velocities
            pint_env%uv_t = pint_env%uv_t+dti2*(pint_env%uf_h+pint_env%uf)
            ! physical forces are only applied in first respa step.
            pint_env%uf = 0.0_dp
            ! calc harmonic forces at new pos
            pint_env%ux = pint_env%ux_t
            CALL pint_calc_uf_h(pint_env=pint_env, e_h=e_h)
            pint_env%uv_t = pint_env%uv_t+dti2*(pint_env%uf_h+pint_env%uf)

            ! For last respa step include integration of physical and helium
            ! forces
            IF (iresp == pint_env%nrespa) THEN
               CALL pint_u2x(pint_env)
               CALL pint_calc_f(pint_env)
               ! perform helium step and add helium forces
               IF (PRESENT(helium_solvent)) THEN
                  IF (ASSOCIATED(helium_solvent)) THEN
                     helium_solvent%current_step = pint_env%iter
                     helium_solvent%origin = pint_com_pos(pint_env)
                     CALL helium_step(helium_solvent, pint_env)
                     pint_env%f(:, :) = pint_env%f(:, :)+helium_solvent%force_avrg(:, :)
                  END IF
               END IF

               CALL pint_f2uf(pint_env)
               ! set the centroid forces to 0 if FIX_CENTROID_POS
               IF (pint_env%first_propagated_mode .EQ. 2) THEN
                  pint_env%uf(1, :) = 0.0_dp
               END IF
               !Scale physical forces and integrate velocities with physical
               !forces
               pint_env%uf = pint_env%uf*rn
               pint_env%uv_t = pint_env%uv_t+dti2*pint_env%uf
            END IF

            ! Apply second half of thermostats
            SELECT CASE (pint_env%pimd_thermostat)
            CASE (thermostat_nose)
               DO i = 1, 6
                  tol = 0._dp
                  pint_env%uv_new = pint_env%uv_t/(1.+dti2*pint_env%tv(1, :, :))
                  DO idim = 1, pint_env%ndim
                     DO ib = 1, pint_env%p
                        pint_env%tf(1, ib, idim) = (pint_env%mass_fict(ib, idim)* &
                                                    pint_env%uv_new(ib, idim)**2-pint_env%kT)/ &
                                                   pint_env%Q(ib)
                     END DO
                  END DO

                  DO idim = 1, pint_env%ndim
                     DO ib = 1, pint_env%p
                        DO inos = 1, pint_env%nnos-1
                           pint_env%tv_new(inos, ib, idim) = &
                              (pint_env%tv_t(inos, ib, idim)+dti2*pint_env%tf(inos, ib, idim))/ &
                              (1._dp+dti2*pint_env%tv(inos+1, ib, idim))
                           pint_env%tf(inos+1, ib, idim) = &
                              (pint_env%tv_new(inos, ib, idim)**2- &
                               pint_env%kT/pint_env%Q(ib))
                           tol = MAX(tol, ABS(pint_env%tv(inos, ib, idim) &
                                              -pint_env%tv_new(inos, ib, idim)))
                        END DO
                        pint_env%tv_new(pint_env%nnos, ib, idim) = &
                           pint_env%tv_t(pint_env%nnos, ib, idim)+ &
                           dti2*pint_env%tf(pint_env%nnos, ib, idim)
                        tol = MAX(tol, ABS(pint_env%tv(pint_env%nnos, ib, idim) &
                                           -pint_env%tv_new(pint_env%nnos, ib, idim)))
                        tol = MAX(tol, ABS(pint_env%uv(ib, idim) &
                                           -pint_env%uv_new(ib, idim)))
                     END DO
                  END DO

                  pint_env%uv = pint_env%uv_new
                  pint_env%tv = pint_env%tv_new
                  IF (tol <= pint_env%v_tol) EXIT
               END DO
               DO inos = 1, pint_env%nnos-1
                  pint_env%tf(inos, :, :) = pint_env%tf(inos, :, :) &
                                            -pint_env%tv(inos, :, :)*pint_env%tv(inos+1, :, :)
               END DO
            CASE (thermostat_gle)
               CALL pint_gle_step(pint_env)
               pint_env%uv = pint_env%uv_t
            CASE DEFAULT
               pint_env%uv = pint_env%uv_t
            END SELECT
         END DO
      CASE (integrate_exact)
         ! The Liouvillian splitting is as follows:
         ! 1. Thermostat
         ! 2. 0.5*physical integration
         ! 3. Exact harmonic integration
         ! 4. 0.5*physical integration
         ! 5. Thermostat

         ! 1. Apply thermostats
         SELECT CASE (pint_env%pimd_thermostat)
         CASE (thermostat_pile)
            CALL pint_pile_step(vold=pint_env%uv, &
                                vnew=pint_env%uv_t, &
                                p=pint_env%p, &
                                ndim=pint_env%ndim, &
                                first_mode=pint_env%first_propagated_mode, &
                                masses=pint_env%mass_fict, &
                                pile_therm=pint_env%pile_therm)
         CASE (thermostat_piglet)
            CALL pint_piglet_step(vold=pint_env%uv, &
                                  vnew=pint_env%uv_t, &
                                  first_mode=pint_env%first_propagated_mode, &
                                  masses=pint_env%mass_fict, &
                                  piglet_therm=pint_env%piglet_therm)
         CASE DEFAULT
            pint_env%uv_t = pint_env%uv
         END SELECT

         ! 2. 1/2*Physical integration
         pint_env%uv_t = pint_env%uv_t+dti2*pint_env%uf

         ! 3. Exact harmonic integration
         IF (pint_env%first_propagated_mode == 1) THEN
            ! The centroid is integrated via standard velocity-verlet
            pint_env%ux_t(1, :) = pint_env%ux(1, :)+ &
                                  dti*pint_env%uv_t(1, :)+ &
                                  dti22*pint_env%uf_h(1, :)
            pint_env%uv_t(1, :) = pint_env%uv_t(1, :)+ &
                                  dti2*pint_env%uf_h(1, :)

         ELSE
            ! set velocities to zero for fixed centroids
            pint_env%ux_t(1, :) = pint_env%ux(1, :)
            pint_env%uv_t(1, :) = 0.0_dp
         END IF

         DO i = 2, pint_env%p
            pint_env%ux_t(i, :) = pint_env%cosex(i)*pint_env%ux(i, :) &
                                  +pint_env%iwsinex(i)*pint_env%uv_t(i, :)
            pint_env%uv_t(i, :) = pint_env%cosex(i)*pint_env%uv_t(i, :) &
                                  -pint_env%wsinex(i)*pint_env%ux(i, :)
         END DO
         pint_env%ux = pint_env%ux_t

         ! 4. 1/2*Physical integration
         pint_env%uf = 0.0_dp
         CALL pint_u2x(pint_env)
         CALL pint_calc_f(pint_env)
         ! perform helium step and add helium forces
         IF (PRESENT(helium_solvent)) THEN
            IF (ASSOCIATED(helium_solvent)) THEN
               helium_solvent%current_step = pint_env%iter
               helium_solvent%origin = pint_com_pos(pint_env)
               CALL helium_step(helium_solvent, pint_env)
               pint_env%f(:, :) = pint_env%f(:, :)+helium_solvent%force_avrg(:, :)
            END IF
         END IF
         CALL pint_f2uf(pint_env)
         ! set the centroid forces to 0 if FIX_CENTROID_POS
         IF (pint_env%first_propagated_mode .EQ. 2) THEN
            pint_env%uf(1, :) = 0.0_dp
         END IF
         pint_env%uv_t = pint_env%uv_t+dti2*pint_env%uf

         ! 5. Apply thermostats
         SELECT CASE (pint_env%pimd_thermostat)
         CASE (thermostat_pile)
            CALL pint_pile_step(vold=pint_env%uv_t, &
                                vnew=pint_env%uv, &
                                p=pint_env%p, &
                                ndim=pint_env%ndim, &
                                first_mode=pint_env%first_propagated_mode, &
                                masses=pint_env%mass_fict, &
                                pile_therm=pint_env%pile_therm)
         CASE (thermostat_piglet)
            CALL pint_piglet_step(vold=pint_env%uv_t, &
                                  vnew=pint_env%uv, &
                                  first_mode=pint_env%first_propagated_mode, &
                                  masses=pint_env%mass_fict, &
                                  piglet_therm=pint_env%piglet_therm)
         CASE DEFAULT
            pint_env%uv = pint_env%uv_t
         END SELECT

      END SELECT

      ! calculate the energy components
      CALL pint_calc_energy(pint_env)

      ! check that the number of MD steps matches
      ! the number of force evaluations done so far
!TODO make this check valid if we start from ITERATION != 0
!     CALL f_env_add_defaults(f_env_id=pint_env%replicas%f_env_id,&
!          f_env=f_env,new_error=new_error)
!     NULLIFY(logger)
!     logger => cp_error_get_logger(new_error)
!     IF(logger%iter_info%iteration(2)/=pint_env%iter+1)&
!        CALL cp__b("motion/pint_methods.F",1885,"md & force_eval lost sychro")
!     CALL f_env_rm_defaults(f_env,new_error,ierr)

      time_stop = m_walltime()
      pint_env%time_per_step = time_stop-time_start
      CALL pint_write_step_info(pint_env)
      CALL timestop(handle)

      RETURN
   END SUBROUTINE pint_step

! ***************************************************************************
!> \brief  Calculate the energy components (private wrapper function)
!> \param pint_env ...
!> \date   2011-01-07
!> \author Lukasz Walewski
! **************************************************************************************************
   SUBROUTINE pint_calc_energy(pint_env)

      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_energy', &
         routineP = moduleN//':'//routineN

      CALL pint_calc_e_kin_beads_u(pint_env)
      CALL pint_calc_e_vir(pint_env)

      SELECT CASE (pint_env%pimd_thermostat)
      CASE (thermostat_nose)
         CALL pint_calc_nh_energy(pint_env)
      CASE (thermostat_gle)
         CALL pint_calc_gle_energy(pint_env)
      CASE (thermostat_pile)
         CALL pint_calc_pile_energy(pint_env)
      CASE (thermostat_piglet)
         CALL pint_calc_piglet_energy(pint_env)
      END SELECT

      pint_env%energy(e_kin_thermo_id) = &
         0.5_dp*REAL(pint_env%p, dp)*REAL(pint_env%ndim, dp)*pint_env%kT- &
         pint_env%e_pot_h

      pint_env%energy(e_potential_id) = SUM(pint_env%e_pot_bead)

      pint_env%energy(e_conserved_id) = &
         pint_env%energy(e_potential_id)*pint_env%propagator%physpotscale+ &
         pint_env%e_pot_h+ &
         pint_env%e_kin_beads+ &
         pint_env%e_pot_t+ &
         pint_env%e_kin_t+pint_env%e_gle+pint_env%e_pile+pint_env%e_piglet

      pint_env%energy(e_potential_id) = &
         pint_env%energy(e_potential_id)/REAL(pint_env%p, dp)

      RETURN
   END SUBROUTINE pint_calc_energy

! ***************************************************************************
!> \brief  Calculate the harmonic force in the u basis
!> \param  pint_env the path integral environment in which the harmonic
!>         forces should be calculated
!> \param e_h ...
!> \par    History
!>           Added normal mode transformation [hforbert]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_calc_uf_h(pint_env, e_h)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(KIND=dp), INTENT(OUT)                         :: e_h

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_uf_h', routineP = moduleN//':'//routineN

      IF (pint_env%transform == transformation_stage) THEN
         CALL staging_calc_uf_h(pint_env%staging_env, &
                                pint_env%mass_beads, &
                                pint_env%ux, &
                                pint_env%uf_h, &
                                pint_env%e_pot_h)
      ELSE
         CALL normalmode_calc_uf_h(pint_env%normalmode_env, &
                                   pint_env%mass_beads, &
                                   pint_env%ux, &
                                   pint_env%uf_h, &
                                   pint_env%e_pot_h)
      END IF
      e_h = pint_env%e_pot_h
      pint_env%uf_h = pint_env%uf_h/pint_env%mass_fict
      RETURN
   END SUBROUTINE pint_calc_uf_h

! ***************************************************************************
!> \brief calculates the force (and energy) in each bead, returns the sum
!>      of the potential energy
!> \param pint_env path integral environment on which you want to calculate
!>        the forces
!> \param x positions at which you want to evaluate the forces
!> \param f the forces
!> \param e potential energy on each bead
!> \par    History
!>           2009-06-15 moved helium calls out from here [lwalewski]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_calc_f(pint_env, x, f, e)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(in), &
         OPTIONAL, TARGET                                :: x
      REAL(kind=dp), DIMENSION(:, :), INTENT(out), &
         OPTIONAL, TARGET                                :: f
      REAL(kind=dp), DIMENSION(:), INTENT(out), &
         OPTIONAL, TARGET                                :: e

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_f', routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim
      REAL(kind=dp), DIMENSION(:), POINTER               :: my_e
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: my_f, my_x

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",2002)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",2003)
      my_x => pint_env%x
      IF (PRESENT(x)) my_x => x
      my_f => pint_env%f
      IF (PRESENT(f)) my_f => f
      my_e => pint_env%e_pot_bead
      IF (PRESENT(e)) my_e => e
      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            pint_env%replicas%r(idim, ib) = my_x(ib, idim)
         END DO
      END DO
      CALL rep_env_calc_e_f(pint_env%replicas, calc_f=.TRUE.)
      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            !ljw: is that fine ? - idim <-> ib
            my_f(ib, idim) = pint_env%replicas%f(idim, ib)
         END DO
      END DO
      my_e = pint_env%replicas%f(SIZE(pint_env%replicas%f, 1), :)

      RETURN
   END SUBROUTINE pint_calc_f

! ***************************************************************************
!> \brief  Calculate the kinetic energy of the beads (in the u variables)
!> \param pint_env ...
!> \param uv ...
!> \param e_k ...
!> \par    History
!>         Bug fix to give my_uv a default location if not given in call [hforbert]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_calc_e_kin_beads_u(pint_env, uv, e_k)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), DIMENSION(:, :), INTENT(in), &
         OPTIONAL, TARGET                                :: uv
      REAL(kind=dp), INTENT(out), OPTIONAL               :: e_k

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_e_kin_beads_u', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim
      REAL(kind=dp)                                      :: res
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: my_uv

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",2049)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",2050)
      res = -1.0_dp
      my_uv => pint_env%uv
      IF (PRESENT(uv)) my_uv => uv
      res = 0._dp
      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            res = res+pint_env%mass_fict(ib, idim)*my_uv(ib, idim)**2
         END DO
      END DO
      res = res*0.5
      IF (.NOT. PRESENT(uv)) pint_env%e_kin_beads = res
      IF (PRESENT(e_k)) e_k = res
      RETURN
   END SUBROUTINE pint_calc_e_kin_beads_u

! ***************************************************************************
!> \brief  Calculate the virial estimator of the real (quantum) kinetic energy
!> \param pint_env ...
!> \param e_vir ...
!> \author hforbert
!> \note   This subroutine modifies pint_env%energy(e_kin_virial_id) global
!>         variable [lwalewski]
! **************************************************************************************************
   SUBROUTINE pint_calc_e_vir(pint_env, e_vir)
      TYPE(pint_env_type), POINTER                       :: pint_env
      REAL(kind=dp), INTENT(out), OPTIONAL               :: e_vir

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_e_vir', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim
      REAL(kind=dp)                                      :: res, xcentroid

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",2084)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",2085)
      res = -1.0_dp
      res = 0._dp
      DO idim = 1, pint_env%ndim
         ! calculate the centroid
         xcentroid = 0._dp
         DO ib = 1, pint_env%p
            xcentroid = xcentroid+pint_env%x(ib, idim)
         END DO
         xcentroid = xcentroid/REAL(pint_env%p, dp)
         DO ib = 1, pint_env%p
            res = res+(pint_env%x(ib, idim)-xcentroid)*pint_env%f(ib, idim)
         END DO
      END DO
      res = 0.5_dp*(REAL(pint_env%ndim, dp)* &
                    (pint_env%kT*pint_env%propagator%temp_sim2phys)-res/REAL(pint_env%p, dp))
      pint_env%energy(e_kin_virial_id) = res
      IF (PRESENT(e_vir)) e_vir = res
      RETURN
   END SUBROUTINE pint_calc_e_vir

! ***************************************************************************
!> \brief calculates the energy (potential and kinetic) of the Nose-Hoover
!>      chain thermostats
!> \param pint_env the path integral environment
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pint_calc_nh_energy(pint_env)
      TYPE(pint_env_type), POINTER                       :: pint_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pint_calc_nh_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ib, idim, inos
      REAL(kind=dp)                                      :: ekin, epot

      IF(.NOT.(ASSOCIATED(pint_env)))CALL cp__a("motion/pint_methods.F",2121)
      IF(.NOT.(pint_env%ref_count > 0))CALL cp__a("motion/pint_methods.F",2122)
      ekin = 0._dp
      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            DO inos = 1, pint_env%nnos
               ekin = ekin+pint_env%Q(ib)*pint_env%tv(inos, ib, idim)**2
            END DO
         END DO
      END DO
      pint_env%e_kin_t = 0.5_dp*ekin
      epot = 0._dp
      DO idim = 1, pint_env%ndim
         DO ib = 1, pint_env%p
            DO inos = 1, pint_env%nnos
               epot = epot+pint_env%tx(inos, ib, idim)
            END DO
         END DO
      END DO
      pint_env%e_pot_t = pint_env%kT*epot
      RETURN
   END SUBROUTINE pint_calc_nh_energy

END MODULE pint_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE eri_mme_test

   USE cell_types,                      ONLY: cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              init_cell
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE eri_mme_error_control,           ONLY: eri_mme_set_params_custom
   USE eri_mme_integrate,               ONLY: eri_mme_2c_integrate_low
   USE eri_mme_types,                   ONLY: create_eri_mme_section,&
                                              dir_lattice_space,&
                                              eri_mme_init_read_input,&
                                              eri_mme_param,&
                                              eri_mme_release,&
                                              rec_lattice_space
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: real_t
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eri_mme_test'

   PUBLIC :: create_eri_mme_test_section, eri_mme_perf_acc_test

CONTAINS

! **************************************************************************************************
!> \brief Create input section for unit testing
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_eri_mme_test_section(section)
      TYPE(section_type), INTENT(INOUT), POINTER         :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eri_mme_test_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("eri_mme_test.F",68)
      CALL section_create(section, "ERI_MME_TEST", &
                          description="Parameters for testing the ERI_MME method for electron repulsion integrals. "// &
                          "Testing w.r.t. performance and accuracy. ", &
                          n_keywords=5, n_subsections=1)

      CALL create_eri_mme_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation the ERI_MME test. ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ABC", &
                          description="Specify the lengths of the cell vectors A, B, and C. ", &
                          usage="ABC 10.000 10.000 10.000", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NAB_MIN", &
                          description="Minimum number of atomic distances to consider. ", &
                          default_i_val=8)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NREP", &
                          description="Number of repeated calculation of each integral. ", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_ACCURACY", &
                          description="Whether integrals should be compared to reference values, "// &
                          "created on the fly by exact method (G-space sum on grid without "// &
                          "minimax approximation). Note: only feasible for not too many "// &
                          "integrals and maximum exponent around 10.0. ", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LMAX", &
                          description="Maximum total angular momentum quantum number. ", &
                          default_i_val=6)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ZET_MIN", &
                          description="Minimum exponent. ", &
                          default_r_val=0.001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ZET_MAX", &
                          description="Maximum exponent. ", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NZET", &
                          description="Number of exponents (logarithmic partition between ZET_MIN and ZET_MAX). ", &
                          default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_eri_mme_test_section

! **************************************************************************************************
!> \brief Unit test for performance and accuracy (orthorhombic cell only)
!> \param para_env ...
!> \param iw ...
!> \param eri_mme_test_section ...
! **************************************************************************************************
   SUBROUTINE eri_mme_perf_acc_test(para_env, iw, eri_mme_test_section)
      TYPE(cp_para_env_type), INTENT(IN), POINTER        :: para_env
      INTEGER, INTENT(IN)                                :: iw
      TYPE(section_vals_type), INTENT(INOUT), POINTER    :: eri_mme_test_section

      CHARACTER(len=*), PARAMETER :: routineN = 'eri_mme_perf_acc_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: count_ab, i, iab, irep, izet, jab, kab, &
                                                            l, l_max, nab, nab_min, nab_xyz, nrep, &
                                                            nzet
      LOGICAL                                            :: acc_check, test_accuracy
      REAL(KIND=dp)                                      :: acc_G, acc_R, G_res, t0, t1, t2, &
                                                            zet_fac, zetmax, zetmin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rab, time_G, time_R
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: I_diff_G, I_diff_R, I_ref, I_test_G, &
                                                            I_test_R
      REAL(KIND=dp), DIMENSION(3, 3)                     :: ht
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cell_par
      TYPE(cell_type), POINTER                           :: box
      TYPE(eri_mme_param)                                :: param
      TYPE(section_vals_type), POINTER                   :: eri_mme_section

      NULLIFY (box, eri_mme_section, cell_par)

      eri_mme_section => section_vals_get_subs_vals(eri_mme_test_section, "ERI_MME")
      CALL eri_mme_init_read_input(eri_mme_section, param)
      CALL section_vals_val_get(eri_mme_test_section, "ABC", r_vals=cell_par)
      CALL section_vals_val_get(eri_mme_test_section, "NAB_MIN", i_val=nab_min)
      CALL section_vals_val_get(eri_mme_test_section, "NREP", i_val=nrep)
      CALL section_vals_val_get(eri_mme_test_section, "CHECK_ACCURACY", l_val=test_accuracy)
      CALL section_vals_val_get(eri_mme_test_section, "LMAX", i_val=l_max)
      CALL section_vals_val_get(eri_mme_test_section, "ZET_MIN", r_val=zetmin)
      CALL section_vals_val_get(eri_mme_test_section, "ZET_MAX", r_val=zetmax)
      CALL section_vals_val_get(eri_mme_test_section, "NZET", i_val=nzet)
      IF (nzet .LE. 0) &
         CALL cp__b("eri_mme_test.F",183,"Number of exponents NZET must be greater than 0.")

      CALL init_orbital_pointers(l_max)

      ! Create ranges of zet to be tested
      ALLOCATE (zet(nzet))

      zet(1) = zetmin
      IF (nzet .GT. 1) THEN
         zet_fac = (zetmax/zetmin)**(1.0_dp/(nzet-1))
         DO i = 1, nzet-1
            zet(i+1) = zet(i)*zet_fac
         ENDDO
      ENDIF

      ! Create range of rab (atomic distances) to be tested
      nab_xyz = CEILING(REAL(nab_min, KIND=dp)**(1.0_dp/3.0_dp)-1.0E-06)
      nab = nab_xyz**3

      ALLOCATE (rab(3, nab))
      count_ab = 0
      DO iab = 1, nab_xyz
         DO jab = 1, nab_xyz
            DO kab = 1, nab_xyz
               count_ab = count_ab+1
               rab(:, count_ab) = [iab*ABS(cell_par(1)), jab*ABS(cell_par(2)), kab*ABS(cell_par(3))]/nab_xyz
            ENDDO
         ENDDO
      ENDDO

      ! initialize cell
      CALL cell_create(box)
      box%hmat = 0.0_dp
      box%hmat(1, 1) = cell_par(1)
      box%hmat(2, 2) = cell_par(2)
      box%hmat(3, 3) = cell_par(3)
      CALL init_cell(box)

      ! initialize MME method
      CALL eri_mme_set_params_custom(param, box, MINVAL(zet), [MAXVAL(zet)], [l_max], l_max, para_env)
      IF (iw > 0) CALL m_flush(iw)

      ! Calculate reference values (Exact expression in G space converged to high precision)
      IF (test_accuracy) THEN
         ht = twopi*TRANSPOSE(param%h_inv)
         G_res = 0.5_dp*param%G_min

         ALLOCATE (I_ref(ncoset(l_max), ncoset(l_max), nab, nzet))
         I_ref(:, :, :, :) = 0.0_dp

         DO izet = 1, nzet
            DO iab = 1, nab
               CALL eri_mme_2c_integrate_low(param, 0, l_max, 0, l_max, zet(izet), zet(izet), rab(:, iab), &
                                             I_ref(:, :, iab, izet), 0, 0, normalize=.TRUE., exact_method=.TRUE.)

            ENDDO
         ENDDO
      ENDIF

      ! test performance and accuracy of MME method
      ALLOCATE (I_test_G(ncoset(l_max), ncoset(l_max), nab, nzet))
      ALLOCATE (I_test_R(ncoset(l_max), ncoset(l_max), nab, nzet))
      ALLOCATE (I_diff_G(ncoset(l_max), ncoset(l_max), nab, nzet))
      ALLOCATE (I_diff_R(ncoset(l_max), ncoset(l_max), nab, nzet))

      ALLOCATE (time_G(0:l_max, nzet))
      ALLOCATE (time_R(0:l_max, nzet))
      DO l = 0, l_max
         DO izet = 1, nzet
            CALL CPU_TIME(t0)
            DO irep = 1, nrep
               DO iab = 1, nab
                  CALL eri_mme_2c_integrate_low(param, 0, l, 0, l, zet(izet), zet(izet), rab(:, iab), &
                                                I_test_G(:, :, iab, izet), 0, 0, normalize=.TRUE., &
                                                lattice_space=rec_lattice_space)
               ENDDO
            ENDDO
            CALL CPU_TIME(t1)
            DO irep = 1, nrep
               DO iab = 1, nab
                  CALL eri_mme_2c_integrate_low(param, 0, l, 0, l, zet(izet), zet(izet), rab(:, iab), &
                                                I_test_R(:, :, iab, izet), 0, 0, normalize=.TRUE., &
                                                lattice_space=dir_lattice_space)
               ENDDO
            ENDDO
            CALL CPU_TIME(t2)
            time_G(l, izet) = t1-t0
            time_R(l, izet) = t2-t1
         ENDDO
      ENDDO

      CALL mp_sum(time_G, para_env%group)
      CALL mp_sum(time_R, para_env%group)

      IF (test_accuracy) THEN
         I_diff_G(:, :, :, :) = ABS(I_test_G-I_ref)
         I_diff_R(:, :, :, :) = ABS(I_test_R-I_ref)
      ENDIF

      IF (iw > 0) THEN
         WRITE (iw, '(T2, A, T61, I20)') "ERI_MME| Number of atomic distances:", nab
         WRITE (iw, '(/T2, A)') "ERI_MME| Test results for time and accuracy in G and R space"
         WRITE (iw, '(T11, A)') "l, zet, cpu time G, cpu time R, accuracy G, accuracy R"

         DO l = 0, l_max
            DO izet = 1, nzet
               IF (test_accuracy) THEN
                  acc_G = MAXVAL(I_diff_G(ncoset(l-1)+1:ncoset(l), ncoset(l-1)+1:ncoset(l), :, izet))
                  acc_R = MAXVAL(I_diff_R(ncoset(l-1)+1:ncoset(l), ncoset(l-1)+1:ncoset(l), :, izet))
               ELSE
                  acc_G = 0.0_dp
                  acc_R = 0.0_dp
               ENDIF

               WRITE (iw, '(T11, I1, 1X, ES9.2, 1X, ES9.2, 1X, ES9.2, 1X, ES9.2, 1X, ES9.2)') &
                  l, zet(izet), time_G(l, izet), time_R(l, izet), acc_G, acc_R
            ENDDO
         ENDDO

         IF (test_accuracy) THEN
            WRITE (iw, '(/T2, A, 32X, ES9.2, 1X, ES9.2, 1X)') "ERI_MME| Maximum error G, R:", &
               MAXVAL(I_diff_G), MAXVAL(I_diff_R)

            acc_check = param%error_tot .GE. MAX(MAXVAL(I_diff_G), MAXVAL(I_diff_R))
            IF (.NOT. acc_check) &
               CALL cp__b("eri_mme_test.F",308,"Actual error greater than upper bound estimate.")

         ENDIF
      ENDIF

      CALL cell_release(box)
      CALL eri_mme_release(param)

   END SUBROUTINE eri_mme_perf_acc_test

END MODULE eri_mme_test
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Test of Clebsch-Gordon Coefficients
!> \par History
!>      none
!> \author JGH (28.02.2002)
! **************************************************************************************************
MODULE cg_test

   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE lebedev,                         ONLY: deallocate_lebedev_grids,&
                                              get_number_of_lebedev_grid,&
                                              init_lebedev_grids,&
                                              lebedev_grid
   USE machine,                         ONLY: m_walltime
   USE mathconstants,                   ONLY: pi
   USE spherical_harmonics,             ONLY: clebsch_gordon,&
                                              clebsch_gordon_deallocate,&
                                              clebsch_gordon_init,&
                                              y_lm
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cg_test'
   PUBLIC :: clebsch_gordon_test

CONTAINS

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE clebsch_gordon_test()

      CHARACTER(len=*), PARAMETER :: routineN = 'clebsch_gordon_test', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: l = 7

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: a1, a2, a3
      INTEGER                                            :: il, iw, l1, l2, ll, lp, m1, m2, mm, mp, &
                                                            na
      REAL(KIND=dp)                                      :: ca, cga(10), cn, rga(10, 21), tend, &
                                                            tstart
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: b1, b2, b3, wa

      iw = cp_logger_get_default_io_unit()

      IF (iw > 0) THEN

         WRITE (iw, '(/,A,/)') " Test of Clebsch-Gordon Coefficients"
         WRITE (iw, '(T40,A,T77,I4)') " Maximum l value tested:", l

         na = 500
         CALL init_lebedev_grids
         ll = get_number_of_lebedev_grid(n=na)
         na = lebedev_grid(ll)%n
         ALLOCATE (wa(na))
         ALLOCATE (a1(na), a2(na), a3(na))
         ALLOCATE (b1(na), b2(na), b3(na))

         wa(1:na) = 4.0_dp*pi*lebedev_grid(ll)%w(1:na)

         tstart = m_walltime()
         CALL clebsch_gordon_init(l)
         tend = m_walltime()
         tend = tend-tstart
         WRITE (iw, '(T30,A,T71,F10.3)') " Time for Clebsch-Gordon Table [s] ", tend
         lp = (l**4+6*l**3+15*l**2+18*l+8)/8
         lp = 2*lp*(l+1)
         WRITE (iw, '(T30,A,T71,I10)') "      Size of Clebsch-Gordon Table ", lp
         WRITE (iw, '(/,A)') " Start Test for Complex Spherical Harmonics "

         DO l1 = 0, l
            DO m1 = -l1, l1
               CALL y_lm(lebedev_grid(ll)%r, a1, l1, m1)
               DO l2 = 0, l
                  DO m2 = -l2, l2
                     CALL y_lm(lebedev_grid(ll)%r, a2, l2, m2)
                     CALL clebsch_gordon(l1, m1, l2, m2, cga)
                     DO lp = MOD(l1+l2, 2), l1+l2, 2
                        mp = m1+m2
                        IF (lp < ABS(mp)) CYCLE
                        CALL y_lm(lebedev_grid(ll)%r, a3, lp, mp)
                        cn = REAL(SUM(a1*a2*CONJG(a3)*wa), KIND=dp)
                        il = lp/2+1
                        ca = cga(il)
                        IF (ABS(ca-cn) > 1.e-10_dp) THEN
                           WRITE (*, '(A,3I5,A,F20.12)') " l ", l1, l2, lp, " A ", ca
                           WRITE (*, '(A,3I5,A,F20.12)') " m ", m1, m2, mp, " N ", cn
                           WRITE (*, *)
                        END IF
                     END DO
                  END DO
               END DO
            END DO
            WRITE (iw, '(A,i2,A)') " Test for l = ", l1, " done"
         END DO

         WRITE (iw, '(/,A)') " Start Test for Real Spherical Harmonics "
         DO l1 = 0, l
            DO m1 = -l1, l1
               CALL y_lm(lebedev_grid(ll)%r, b1, l1, m1)
               DO l2 = 0, l
                  DO m2 = -l2, l2
                     CALL y_lm(lebedev_grid(ll)%r, b2, l2, m2)
                     CALL clebsch_gordon(l1, m1, l2, m2, rga)
                     mp = m1+m2
                     mm = m1-m2
                     IF (m1*m2 < 0 .OR. (m1*m2 == 0 .AND. (m1 < 0 .OR. m2 < 0))) THEN
                        mp = -ABS(mp)
                        mm = -ABS(mm)
                     ELSE
                        mp = ABS(mp)
                        mm = ABS(mm)
                     END IF
                     DO lp = MOD(l1+l2, 2), l1+l2, 2
                        IF (ABS(mp) <= lp) THEN
                           CALL y_lm(lebedev_grid(ll)%r, b3, lp, mp)
                           cn = SUM(b1*b2*b3*wa)
                           il = lp/2+1
                           ca = rga(il, 1)
                           IF (ABS(ca-cn) > 1.e-10_dp) THEN
                              WRITE (*, '(A,3I5,A,F20.12)') " l ", l1, l2, lp, " A ", ca
                              WRITE (*, '(A,3I5,A,F20.12)') " m ", m1, m2, mp, " N ", cn
                              WRITE (*, *)
                           END IF
                        END IF
                        IF (mp /= mm .AND. ABS(mm) <= lp) THEN
                           CALL y_lm(lebedev_grid(ll)%r, b3, lp, mm)
                           cn = SUM(b1*b2*b3*wa)
                           il = lp/2+1
                           ca = rga(il, 2)
                           IF (ABS(ca-cn) > 1.e-10_dp) THEN
                              WRITE (*, '(A,3I5,A,F20.12)') " l ", l1, l2, lp, " A ", ca
                              WRITE (*, '(A,3I5,A,F20.12)') " m ", m1, m2, mm, " N ", cn
                              WRITE (*, *)
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            END DO
            WRITE (iw, '(A,i2,A)') " Test for l = ", l1, " done"
         END DO

         DEALLOCATE (wa)
         DEALLOCATE (a1, a2, a3)
         DEALLOCATE (b1, b2, b3)

         CALL deallocate_lebedev_grids()
         CALL clebsch_gordon_deallocate()

      END IF

   END SUBROUTINE clebsch_gordon_test

END MODULE cg_test

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Test of Electron Repulsion Routines (ERI)
!> \par History
!>      none
!> \author JGH (01.07.2009)
! **************************************************************************************************
MODULE ai_coulomb_test

   USE ai_coulomb,                      ONLY: coulomb2
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
   USE orbital_pointers,                ONLY: deallocate_orbital_pointers,&
                                              init_orbital_pointers,&
                                              nco,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_coulomb_test'

   REAL(KIND=dp), PARAMETER :: threshold = 1.0E-6_dp

   PRIVATE

   PUBLIC :: eri_test
! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE eri_test(iw)

      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eri_test', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: lmax = 6

      CHARACTER(LEN=11), DIMENSION(0:lmax)               :: i2g
      CHARACTER(LEN=5), DIMENSION(0:lmax)                :: i2c
      CHARACTER(LEN=7), DIMENSION(0:lmax)                :: i2e
      CHARACTER(LEN=9), DIMENSION(0:lmax)                :: i2f
      INTEGER                                            :: i, ii, l, la_max, la_min, lc_max, &
                                                            lc_min, ll, n, npgfa, npgfb, npgfc, &
                                                            npgfd
      REAL(KIND=dp)                                      :: perf, rac2, t, tend, tstart
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: f
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vac
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: v
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rb, rc, rd
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rpgf, zeta, zetb, zetc, zetd

      IF (iw > 0) WRITE (iw, '(/,A)') " Test of Electron Repulsion Integrals (ERI) "

      CALL init_orbital_pointers(lmax)

      i2c(0) = "(s|s)"
      i2c(1) = "(p|p)"
      i2c(2) = "(d|d)"
      i2c(3) = "(f|f)"
      i2c(4) = "(g|g)"
      i2c(5) = "(h|h)"
      i2c(6) = "(i|i)"

      i2g(0) = "[(ss)|(ss)]"
      i2g(1) = "[(pp)|(pp)]"
      i2g(2) = "[(dd)|(dd)]"
      i2g(3) = "[(ff)|(ff)]"
      i2g(4) = "[(gg)|(gg)]"
      i2g(5) = "[(hh)|(hh)]"
      i2g(6) = "[(ii)|(ii)]"

      i2f(0) = "[ss|(ss)]"
      i2f(1) = "[pp|(pp)]"
      i2f(2) = "[dd|(dd)]"
      i2f(3) = "[ff|(ff)]"
      i2f(4) = "[gg|(gg)]"
      i2f(5) = "[hh|(hh)]"
      i2f(6) = "[ii|(ii)]"

      i2e(0) = "(ss|ss)"
      i2e(1) = "(pp|pp)"
      i2e(2) = "(dd|dd)"
      i2e(3) = "(ff|ff)"
      i2e(4) = "(gg|gg)"
      i2e(5) = "(hh|hh)"
      i2e(6) = "(ii|ii)"

      npgfa = 4
      npgfb = 2
      npgfc = 4
      npgfd = 1
      n = MAX(npgfa, npgfb, npgfc, npgfd)

      ALLOCATE (zeta(npgfa), zetb(npgfb), zetc(npgfc), zetd(npgfd), rpgf(n))

      zeta(1:npgfa) = 0.5_dp
      zetb(1:npgfb) = 0.4_dp
      zetc(1:npgfc) = 0.3_dp
      zetd(1:npgfd) = 0.2_dp

      ra = (/0.0_dp, 0.0_dp, 0.0_dp/)
      rb = (/1.0_dp, 0.0_dp, 0.0_dp/)
      rc = (/0.0_dp, 0.3_dp, 0.3_dp/)
      rd = (/0.7_dp, 0.2_dp, 0.1_dp/)

      rac2 = SUM((ra-rc)**2)
      rpgf = 1.e10_dp

      ! Performance test of coulomb2 routine
      IF (iw > 0) THEN

         WRITE (iw, '(//,A,/)') " Test of 2-Electron-2-Center Integrals (coulomb2) "
         DO l = 0, lmax
            la_max = l
            la_min = l
            lc_max = l
            lc_min = l
            ll = ncoset(l)
            ALLOCATE (f(0:2*l+2), v(npgfa*ll, npgfc*ll, 2*l+1), vac(npgfa*ll, npgfc*ll))
            vac = 0._dp
            ii = MAX(100/(l+1)**2, 1)
            tstart = m_walltime()
            DO i = 1, ii
               CALL coulomb2(la_max, npgfa, zeta, rpgf, la_min, lc_max, npgfc, zetc, rpgf, lc_min, rc, rac2, vac, v, f)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            perf = REAL(ii*nco(l)**2)*1.e-6_dp*REAL(npgfa*npgfc, dp)/t
            WRITE (iw, '(A,T40,A,T66,F15.3)') " Performance [Mintegrals/s] ", i2c(l), perf
            DEALLOCATE (f, v, vac)
         END DO

      END IF

      DEALLOCATE (zeta, zetb, zetc, zetd, rpgf)

      CALL deallocate_orbital_pointers()

   END SUBROUTINE eri_test

! **************************************************************************************************

END MODULE ai_coulomb_test

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Performance tests for basic tasks like matrix multiplies, copy, fft.
!> \par History
!>      30-Nov-2000 (JGH) added input
!>      02-Jan-2001 (JGH) Parallel FFT
!>      28-Feb-2002 (JGH) Clebsch-Gordon Coefficients
!>      06-Jun-2003 (JGH) Real space grid test
!>      Eigensolver test (29.08.05,MK)
!> \author JGH  6-NOV-2000
! **************************************************************************************************
MODULE library_tests

   USE ai_coulomb_test,                 ONLY: eri_test
   USE cell_types,                      ONLY: cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              init_cell
   USE cg_test,                         ONLY: clebsch_gordon_test
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: dbcsr_run_tests
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_basic_linalg,              ONLY: cp_fm_gemm
   USE cp_fm_diag,                      ONLY: cp_fm_syevd,&
                                              cp_fm_syevx
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_init,          ONLY: init_input_type
   USE eri_mme_test,                    ONLY: eri_mme_perf_acc_test
   USE fft_tools,                       ONLY: BWFFT,&
                                              FFT_RADIX_CLOSEST,&
                                              FWFFT,&
                                              fft3d,&
                                              fft_radix_operations,&
                                              finalize_fft,&
                                              init_fft
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: do_diag_syevd,&
                                              do_diag_syevx,&
                                              do_mat_random,&
                                              do_mat_read,&
                                              do_pwgrid_ns_fullspace,&
                                              do_pwgrid_ns_halfspace,&
                                              do_pwgrid_spherical
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_max,&
                                              mp_sum,&
                                              mp_sync,&
                                              mpi_perf_test
   USE minimax,                         ONLY: minimax_validate
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              UNIFORM,&
                                              check_rng,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type,&
                                              write_rng_stream
   USE pw_grid_types,                   ONLY: FULLSPACE,&
                                              HALFSPACE,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_release,&
                                              pw_grid_setup
   USE pw_methods,                      ONLY: pw_transfer,&
                                              pw_zero
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              COMPLEXDATA3D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_type, realspace_grid_input_type, realspace_grid_type, rs2pw, &
        rs_grid_create, rs_grid_create_descriptor, rs_grid_print, rs_grid_release, &
        rs_grid_release_descriptor, rs_grid_zero, rs_pw_transfer
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: lib_test

   INTEGER                  :: runtest(100)
   REAL(KIND=dp)           :: max_memory
   REAL(KIND=dp), PARAMETER :: threshold = 1.0E-8_dp
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'library_tests'

CONTAINS

! **************************************************************************************************
!> \brief Master routine for tests
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \par History
!>      none
!> \author JGH  6-NOV-2000
! **************************************************************************************************
   SUBROUTINE lib_test(root_section, para_env, globenv)

      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lib_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: explicit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER :: cp_dbcsr_test_section, cp_fm_gemm_test_section, &
         eigensolver_section, eri_mme_test_section, pw_transfer_section, rs_pw_transfer_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, root_section, "TEST%PROGRAM_RUN_INFO", extension=".log")

      IF (iw > 0) THEN
         WRITE (iw, '(T2,79("*"))')
         WRITE (iw, '(A,T31,A,T80,A)') ' *', ' PERFORMANCE TESTS ', '*'
         WRITE (iw, '(T2,79("*"))')
      END IF
      !
      CALL test_input(root_section, para_env)
      !
      IF (runtest(1) /= 0) CALL copy_test(para_env, iw)
      !
      IF (runtest(2) /= 0) CALL matmul_test(para_env, test_matmul=.TRUE., test_dgemm=.FALSE., iw=iw)
      IF (runtest(5) /= 0) CALL matmul_test(para_env, test_matmul=.FALSE., test_dgemm=.TRUE., iw=iw)
      !
      IF (runtest(3) /= 0) CALL fft_test(para_env, iw, globenv%fftw_plan_type, &
                                         globenv%fftw_wisdom_file_name)
      !
      IF (runtest(4) /= 0) CALL eri_test(iw)
      !
      IF (runtest(6) /= 0) CALL clebsch_gordon_test()
      !
      ! runtest 7 has been deleted and can be recycled
      !
      IF (runtest(8) /= 0) CALL mpi_perf_test(para_env%group, runtest(8), iw)
      !
      IF (runtest(9) /= 0) CALL rng_test(para_env, iw)
      !
      IF (runtest(10) /= 0) CALL minimax_validate(runtest(10), iw)
      !
      rs_pw_transfer_section => section_vals_get_subs_vals(root_section, "TEST%RS_PW_TRANSFER")
      CALL section_vals_get(rs_pw_transfer_section, explicit=explicit)
      IF (explicit) THEN
         CALL rs_pw_transfer_test(para_env, iw, globenv, rs_pw_transfer_section)
      ENDIF

      pw_transfer_section => section_vals_get_subs_vals(root_section, "TEST%PW_TRANSFER")
      CALL section_vals_get(pw_transfer_section, explicit=explicit)
      IF (explicit) THEN
         CALL pw_fft_test(para_env, iw, globenv, pw_transfer_section)
      ENDIF

      cp_fm_gemm_test_section => section_vals_get_subs_vals(root_section, "TEST%CP_FM_GEMM")
      CALL section_vals_get(cp_fm_gemm_test_section, explicit=explicit)
      IF (explicit) THEN
         CALL cp_fm_gemm_test(para_env, iw, cp_fm_gemm_test_section)
      ENDIF

      eigensolver_section => section_vals_get_subs_vals(root_section, "TEST%EIGENSOLVER")
      CALL section_vals_get(eigensolver_section, explicit=explicit)
      IF (explicit) THEN
         CALL eigensolver_test(para_env, iw, eigensolver_section)
      ENDIF

      eri_mme_test_section => section_vals_get_subs_vals(root_section, "TEST%ERI_MME_TEST")
      CALL section_vals_get(eri_mme_test_section, explicit=explicit)
      IF (explicit) THEN
         CALL eri_mme_perf_acc_test(para_env, iw, eri_mme_test_section)
      ENDIF

      ! DBCSR tests
      ! matrix_multiplication
      cp_dbcsr_test_section => section_vals_get_subs_vals(root_section, &
                                                          "TEST%CP_DBCSR")
      CALL section_vals_get(cp_dbcsr_test_section, explicit=explicit)
      IF (explicit) THEN
         CALL cp_dbcsr_tests(para_env, iw, cp_dbcsr_test_section)
      ENDIF

      CALL cp_print_key_finished_output(iw, logger, root_section, "TEST%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE lib_test

! **************************************************************************************************
!> \brief Reads input section &TEST ... &END
!> \param root_section ...
!> \param para_env ...
!> \author JGH 30-NOV-2000
!> \note
!> I---------------------------------------------------------------------------I
!> I SECTION: &TEST ... &END                                                   I
!> I                                                                           I
!> I    MEMORY   max_memory                                                    I
!> I    COPY     n                                                             I
!> I    MATMUL   n                                                             I
!> I    FFT      n                                                             I
!> I    ERI      n                                                             I
!> I    PW_FFT   n                                                             I
!> I    Clebsch-Gordon n                                                       I
!> I    RS_GRIDS n                                                             I
!> I    MPI      n                                                             I
!> I    RNG      n             -> Parallel random number generator             I
!> I---------------------------------------------------------------------------I
! **************************************************************************************************
   SUBROUTINE test_input(root_section, para_env)
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'test_input', routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: test_section

!
!..defaults
! using this style is not recommended, introduce sections instead (see e.g. cp_fm_gemm)

      runtest = 0
      test_section => section_vals_get_subs_vals(root_section, "TEST")
      CALL section_vals_val_get(test_section, "MEMORY", r_val=max_memory)
      CALL section_vals_val_get(test_section, 'COPY', i_val=runtest(1))
      CALL section_vals_val_get(test_section, 'MATMUL', i_val=runtest(2))
      CALL section_vals_val_get(test_section, 'DGEMM', i_val=runtest(5))
      CALL section_vals_val_get(test_section, 'FFT', i_val=runtest(3))
      CALL section_vals_val_get(test_section, 'ERI', i_val=runtest(4))
      CALL section_vals_val_get(test_section, 'CLEBSCH_GORDON', i_val=runtest(6))
      CALL section_vals_val_get(test_section, 'MPI', i_val=runtest(8))
      CALL section_vals_val_get(test_section, 'RNG', i_val=runtest(9))
      CALL section_vals_val_get(test_section, 'MINIMAX', i_val=runtest(10))

      CALL mp_sync(para_env%group)
   END SUBROUTINE test_input

! **************************************************************************************************
!> \brief Tests the performance to copy two vectors.
!> \param para_env ...
!> \param iw ...
!> \par History
!>      none
!> \author JGH  6-NOV-2000
!> \note
!>      The results of these tests allow to determine the size of the cache
!>      of the CPU. This can be used to optimize the performance of the
!>      FFTSG library.
! **************************************************************************************************
   SUBROUTINE copy_test(para_env, iw)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw

      INTEGER                                            :: i, ierr, j, len, ntim, siz
      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_test', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: perf, t, tend, tstart
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ca, cb

! test for copy --> Cache size

      siz = ABS(runtest(1))
      IF (para_env%ionode) WRITE (iw, '(//,A,/)') " Test of copy ( F95 ) "
      DO i = 6, 24
         len = 2**i
         IF (8.0_dp*REAL(len, KIND=dp) > max_memory*0.5_dp) EXIT
         ALLOCATE (ca(len), STAT=ierr)
         IF (ierr /= 0) EXIT
         ALLOCATE (cb(len), STAT=ierr)
         IF (ierr /= 0) EXIT

         CALL RANDOM_NUMBER(ca)
         ntim = NINT(1.e7_dp/REAL(len, KIND=dp))
         ntim = MAX(ntim, 1)
         ntim = MIN(ntim, siz*10000)

         tstart = m_walltime()
         DO j = 1, ntim
            cb(:) = ca(:)
            ca(1) = REAL(j, KIND=dp)
         END DO
         tend = m_walltime()
         t = tend-tstart+threshold
         IF (t > 0.0_dp) THEN
            perf = REAL(ntim, KIND=dp)*REAL(len, KIND=dp)*1.e-6_dp/t
         ELSE
            perf = 0.0_dp
         END IF

         IF (para_env%ionode) THEN
            WRITE (iw, '(A,i2,i10,A,T59,F14.4,A)') " Copy test:   Size = 2^", i, &
               len/1024, " Kwords", perf, " Mcopy/s"
         END IF

         DEALLOCATE (ca)
         DEALLOCATE (cb)
      END DO
      CALL mp_sync(para_env%group)
   END SUBROUTINE copy_test

! **************************************************************************************************
!> \brief Tests the performance of different kinds of matrix matrix multiply
!>      kernels for the BLAS and F95 intrinsic matmul.
!> \param para_env ...
!> \param test_matmul ...
!> \param test_dgemm ...
!> \param iw ...
!> \par History
!>      none
!> \author JGH  6-NOV-2000
! **************************************************************************************************
   SUBROUTINE matmul_test(para_env, test_matmul, test_dgemm, iw)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL                                            :: test_matmul, test_dgemm
      INTEGER                                            :: iw

      INTEGER                                            :: i, ierr, j, len, ntim, siz
      CHARACTER(LEN=*), PARAMETER :: routineN = 'matmul_test', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: perf, t, tend, tstart, xdum
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: ma, mb, mc

! test for matrix multpies

      IF (test_matmul) THEN
         siz = ABS(runtest(2))
         IF (para_env%ionode) WRITE (iw, '(//,A,/)') " Test of matmul ( F95 ) "
         DO i = 5, siz, 2
            len = 2**i+1
            IF (8.0_dp*REAL(len*len, KIND=dp) > max_memory*0.3_dp) EXIT
            ALLOCATE (ma(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            ALLOCATE (mb(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            ALLOCATE (mc(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            mc = 0.0_dp

            CALL RANDOM_NUMBER(xdum)
            ma = xdum
            CALL RANDOM_NUMBER(xdum)
            mb = xdum
            ntim = NINT(1.e8_dp/(2.0_dp*REAL(len, KIND=dp)**3))
            ntim = MAX(ntim, 1)
            ntim = MIN(ntim, siz*200)
            tstart = m_walltime()
            DO j = 1, ntim
               mc(:, :) = MATMUL(ma, mb)
               ma(1, 1) = REAL(j, KIND=dp)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a * b         Size = ", len, perf, " Mflop/s"
            END IF
            tstart = m_walltime()
            DO j = 1, ntim
               mc(:, :) = mc+MATMUL(ma, mb)
               ma(1, 1) = REAL(j, KIND=dp)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: a = a * b         Size = ", len, perf, " Mflop/s"
            END IF

            tstart = m_walltime()
            DO j = 1, ntim
               mc(:, :) = mc+MATMUL(ma, TRANSPOSE(mb))
               ma(1, 1) = REAL(j, KIND=dp)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a * b(T)      Size = ", len, perf, " Mflop/s"
            END IF

            tstart = m_walltime()
            DO j = 1, ntim
               mc(:, :) = mc+MATMUL(TRANSPOSE(ma), mb)
               ma(1, 1) = REAL(j, KIND=dp)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a(T) * b      Size = ", len, perf, " Mflop/s"
            END IF

            DEALLOCATE (ma)
            DEALLOCATE (mb)
            DEALLOCATE (mc)
         END DO
      ENDIF

      ! test for matrix multpies
      IF (test_dgemm) THEN
         siz = ABS(runtest(5))
         IF (para_env%ionode) WRITE (iw, '(//,A,/)') " Test of matmul ( BLAS ) "
         DO i = 5, siz, 2
            len = 2**i+1
            IF (8.0_dp*REAL(len*len, KIND=dp) > max_memory*0.3_dp) EXIT
            ALLOCATE (ma(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            ALLOCATE (mb(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            ALLOCATE (mc(len, len), STAT=ierr)
            IF (ierr /= 0) EXIT
            mc = 0.0_dp

            CALL RANDOM_NUMBER(xdum)
            ma = xdum
            CALL RANDOM_NUMBER(xdum)
            mb = xdum
            ntim = NINT(1.e8_dp/(2.0_dp*REAL(len, KIND=dp)**3))
            ntim = MAX(ntim, 1)
            ntim = MIN(ntim, 1000)

            tstart = m_walltime()
            DO j = 1, ntim
               CALL dgemm("N", "N", len, len, len, 1.0_dp, ma, len, mb, len, 1.0_dp, mc, len)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a * b         Size = ", len, perf, " Mflop/s"
            END IF

            tstart = m_walltime()
            DO j = 1, ntim
               CALL dgemm("N", "N", len, len, len, 1.0_dp, ma, len, mb, len, 1.0_dp, mc, len)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: a = a * b         Size = ", len, perf, " Mflop/s"
            END IF

            tstart = m_walltime()
            DO j = 1, ntim
               CALL dgemm("N", "T", len, len, len, 1.0_dp, ma, len, mb, len, 1.0_dp, mc, len)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a * b(T)      Size = ", len, perf, " Mflop/s"
            END IF

            tstart = m_walltime()
            DO j = 1, ntim
               CALL dgemm("T", "N", len, len, len, 1.0_dp, ma, len, mb, len, 1.0_dp, mc, len)
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(ntim, KIND=dp)*2.0_dp*REAL(len, KIND=dp)**3*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            IF (para_env%ionode) THEN
               WRITE (iw, '(A,i6,T59,F14.4,A)') &
                  " Matrix multiply test: c = a(T) * b      Size = ", len, perf, " Mflop/s"
            END IF

            DEALLOCATE (ma)
            DEALLOCATE (mb)
            DEALLOCATE (mc)
         END DO
      ENDIF

      CALL mp_sync(para_env%group)

   END SUBROUTINE matmul_test

! **************************************************************************************************
!> \brief Tests the performance of all available FFT libraries for 3D FFTs
!> \param para_env ...
!> \param iw ...
!> \param fftw_plan_type ...
!> \param wisdom_file where FFTW3 should look to save/load wisdom
!> \par History
!>      none
!> \author JGH  6-NOV-2000
! **************************************************************************************************
   SUBROUTINE fft_test(para_env, iw, fftw_plan_type, wisdom_file)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw, fftw_plan_type
      CHARACTER(LEN=*), INTENT(IN)                       :: wisdom_file

      INTEGER                                            :: iall, ierr, it, j, len, n(3), ntim, &
                                                            radix_in, radix_out, siz, stat
      INTEGER, PARAMETER                                 :: ndate(3) = (/12, 48, 96/)
      CHARACTER(LEN=*), PARAMETER :: routineN = 'fft_test', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(4, 4, 4)               :: zz
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: ca, cb, cc
      CHARACTER(LEN=7)                                   :: method
      REAL(KIND=dp)                                      :: flops, perf, scale, t, tdiff, tend, &
                                                            tstart
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: ra

! test for 3d FFT

      IF (para_env%ionode) WRITE (iw, '(//,A,/)') " Test of 3D-FFT "
      siz = ABS(runtest(3))

      DO iall = 1, 100
         SELECT CASE (iall)
         CASE DEFAULT
            EXIT
         CASE (1)
            CALL init_fft("FFTSG", alltoall=.FALSE., fftsg_sizes=.TRUE., wisdom_file=wisdom_file, &
                          pool_limit=10, plan_style=fftw_plan_type)
            method = "FFTSG  "
         CASE (2)
            CYCLE
         CASE (3)
            CALL init_fft("FFTW3", alltoall=.FALSE., fftsg_sizes=.TRUE., wisdom_file=wisdom_file, &
                          pool_limit=10, plan_style=fftw_plan_type)
            method = "FFTW3  "
         END SELECT
         n = 4
         zz = 0.0_dp
         CALL fft3d(1, n, zz, status=stat)
         IF (stat == 0) THEN
            DO it = 1, 3
               radix_in = ndate(it)
               CALL fft_radix_operations(radix_in, radix_out, FFT_RADIX_CLOSEST)
               len = radix_out
               n = len
               IF (16.0_dp*REAL(len*len*len, KIND=dp) > max_memory*0.5_dp) EXIT
               ALLOCATE (ra(len, len, len), STAT=ierr)
               IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 613, &
                                               "ra", dp_size*len*len*len)
               ALLOCATE (ca(len, len, len), STAT=ierr)
               IF (ierr /= 0) THEN
                  CALL stop_memory(routineN, moduleN, 617, &
                                   "ca", 2*dp_size*len*len*len)
               ELSE
                  CALL RANDOM_NUMBER(ra)
                  ca(:, :, :) = ra
                  CALL RANDOM_NUMBER(ra)
                  ca(:, :, :) = ca+CMPLX(0.0_dp, 1.0_dp, KIND=dp)*ra
                  flops = REAL(len**3, KIND=dp)*15.0_dp*LOG(REAL(len, KIND=dp))
                  ntim = NINT(siz*1.e7_dp/flops)
                  ntim = MAX(ntim, 1)
                  ntim = MIN(ntim, 200)
                  scale = 1.0_dp/REAL(len**3, KIND=dp)
                  tstart = m_walltime()
                  DO j = 1, ntim
                     CALL fft3d(FWFFT, n, ca)
                     CALL fft3d(BWFFT, n, ca, SCALE=scale)
                  END DO
                  tend = m_walltime()
                  t = tend-tstart+threshold
                  IF (t > 0.0_dp) THEN
                     perf = REAL(ntim, KIND=dp)*2.0_dp*flops*1.e-6_dp/t
                  ELSE
                     perf = 0.0_dp
                  END IF

                  IF (para_env%ionode) THEN
                     WRITE (iw, '(T2,A,A,i6,T59,F14.4,A)') &
                        ADJUSTR(method), " test (in-place)    Size = ", len, perf, " Mflop/s"
                  END IF
                  DEALLOCATE (ca)
                  DEALLOCATE (ra)
               END IF
            END DO
            IF (para_env%ionode) WRITE (iw, *)
            ! test if input data is preserved
            len = 24
            n = len
            ALLOCATE (ra(len, len, len), STAT=ierr)
            IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 655, &
                                            "ra", dp_size*len*len*len)
            ALLOCATE (ca(len, len, len), STAT=ierr)
            IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 658, &
                                            "ca", 2*dp_size*len*len*len)
            ALLOCATE (cb(len, len, len), STAT=ierr)
            IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 661, &
                                            "cb", 2*dp_size*len*len*len)
            ALLOCATE (cc(len, len, len), STAT=ierr)
            IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 664, &
                                            "cc", 2*dp_size*len*len*len)
            CALL RANDOM_NUMBER(ra)
            ca(:, :, :) = ra
            CALL RANDOM_NUMBER(ra)
            ca(:, :, :) = ca+CMPLX(0.0_dp, 1.0_dp, KIND=dp)*ra
            cc(:, :, :) = ca
            CALL fft3d(FWFFT, n, ca, cb)
            tdiff = MAXVAL(ABS(ca-cc))
            IF (tdiff > 1.0E-12_dp) THEN
               IF (para_env%ionode) &
                  WRITE (iw, '(T2,A,A,A)') ADJUSTR(method), "         FWFFT ", &
                  "             Input array is changed in out-of-place FFT !"
            ELSE
               IF (para_env%ionode) &
                  WRITE (iw, '(T2,A,A,A)') ADJUSTR(method), "         FWFFT ", &
                  "         Input array is not changed in out-of-place FFT !"
            END IF
            ca(:, :, :) = cc
            CALL fft3d(BWFFT, n, ca, cb)
            tdiff = MAXVAL(ABS(ca-cc))
            IF (tdiff > 1.0E-12_dp) THEN
               IF (para_env%ionode) &
                  WRITE (iw, '(T2,A,A,A)') ADJUSTR(method), "         BWFFT ", &
                  "             Input array is changed in out-of-place FFT !"
            ELSE
               IF (para_env%ionode) &
                  WRITE (iw, '(T2,A,A,A)') ADJUSTR(method), "         BWFFT ", &
                  "         Input array is not changed in out-of-place FFT !"
            END IF
            IF (para_env%ionode) WRITE (iw, *)

            DEALLOCATE (ra)
            DEALLOCATE (ca)
            DEALLOCATE (cb)
            DEALLOCATE (cc)
         END IF
         CALL finalize_fft(para_env, wisdom_file=wisdom_file)
      END DO

   END SUBROUTINE fft_test

! **************************************************************************************************
!> \brief   test rs_pw_transfer performance
!> \param para_env ...
!> \param iw ...
!> \param globenv ...
!> \param rs_pw_transfer_section ...
!> \author  Joost VandeVondele
!>      9.2008 Randomise rs grid [Iain Bethune]
!>      (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE rs_pw_transfer_test(para_env, iw, globenv, rs_pw_transfer_section)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: rs_pw_transfer_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_pw_transfer_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dir, halo_size, handle, i_loop, n_loop, &
                                                            ns_max
      INTEGER, DIMENSION(3)                              :: no, np
      INTEGER, DIMENSION(:), POINTER                     :: i_vals
      LOGICAL                                            :: do_rs2pw
      REAL(KIND=dp)                                      :: tend, tstart
      TYPE(cell_type), POINTER                           :: box
      TYPE(pw_grid_type), POINTER                        :: grid
      TYPE(pw_p_type)                                    :: ca
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_input_type)                    :: input_settings
      TYPE(realspace_grid_type), POINTER                 :: rs_grid
      TYPE(section_vals_type), POINTER                   :: rs_grid_section

      CALL timeset(routineN, handle)

      !..set fft lib
      CALL init_fft(globenv%default_fft_library, alltoall=.FALSE., fftsg_sizes=.TRUE., &
                    pool_limit=globenv%fft_pool_scratch_limit, &
                    wisdom_file=globenv%fftw_wisdom_file_name, &
                    plan_style=globenv%fftw_plan_type)

      ! .. set cell (should otherwise be irrelevant)
      NULLIFY (box)
      CALL cell_create(box)
      box%hmat = RESHAPE((/20.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, 20.0_dp, 0.0_dp, &
                           0.0_dp, 0.0_dp, 20.0_dp/), (/3, 3/))
      CALL init_cell(box)

      ! .. grid type and pw_grid
      NULLIFY (grid)
      CALL section_vals_val_get(rs_pw_transfer_section, "GRID", i_vals=i_vals)
      np = i_vals
      CALL pw_grid_create(grid, para_env%group)
      CALL pw_grid_setup(box%hmat, grid, grid_span=FULLSPACE, npts=np, fft_usage=.TRUE., iounit=iw)
      no = grid%npts

      NULLIFY (ca%pw)
      CALL pw_create(ca%pw, grid, REALDATA3D)
      ca%pw%in_space = REALSPACE
      CALL pw_zero(ca%pw)

      ! .. rs input setting type
      CALL section_vals_val_get(rs_pw_transfer_section, "HALO_SIZE", i_val=halo_size)
      rs_grid_section => section_vals_get_subs_vals(rs_pw_transfer_section, "RS_GRID")
      ns_max = 2*halo_size+1
      CALL init_input_type(input_settings, ns_max, rs_grid_section, 1, (/-1, -1, -1/))

      ! .. rs type
      NULLIFY (rs_grid)
      NULLIFY (rs_desc)
      CALL rs_grid_create_descriptor(rs_desc, pw_grid=grid, input_settings=input_settings)
      CALL rs_grid_create(rs_grid, rs_desc)
      CALL rs_grid_print(rs_grid, iw)
      CALL rs_grid_zero(rs_grid)

      ! Put random values on the grid, so summation check will pick up errors
      CALL RANDOM_NUMBER(rs_grid%r)

      CALL section_vals_val_get(rs_pw_transfer_section, "N_loop", i_val=N_loop)
      CALL section_vals_val_get(rs_pw_transfer_section, "RS2PW", l_val=do_rs2pw)
      IF (do_rs2pw) THEN
         dir = rs2pw
      ELSE
         dir = pw2rs
      ENDIF

      ! go for the real loops, sync to get max timings
      IF (para_env%ionode) THEN
         WRITE (iw, '(T2,A)') ""
         WRITE (iw, '(T2,A)') "Timing rs_pw_transfer routine"
         WRITE (iw, '(T2,A)') ""
         WRITE (iw, '(T2,A)') "iteration      time[s]"
      ENDIF
      DO i_loop = 1, N_loop
         CALL mp_sync(para_env%group)
         tstart = m_walltime()
         CALL rs_pw_transfer(rs_grid, ca%pw, dir)
         CALL mp_sync(para_env%group)
         tend = m_walltime()
         IF (para_env%ionode) THEN
            WRITE (iw, '(T2,I9,1X,F12.6)') i_loop, tend-tstart
         ENDIF
      ENDDO

      !cleanup
      CALL rs_grid_release(rs_grid)
      CALL rs_grid_release_descriptor(rs_desc)
      CALL pw_release(ca%pw)
      CALL pw_grid_release(grid)
      CALL cell_release(box)
      CALL finalize_fft(para_env, wisdom_file=globenv%fftw_wisdom_file_name)

      CALL timestop(handle)

   END SUBROUTINE rs_pw_transfer_test

! **************************************************************************************************
!> \brief Tests the performance of PW calls to FFT routines
!> \param para_env ...
!> \param iw ...
!> \param globenv ...
!> \param pw_transfer_section ...
!> \par History
!>      JGH  6-Feb-2001 : Test and performance code
!>      Made input sensitive [Joost VandeVondele]
!> \author JGH  1-JAN-2001
! **************************************************************************************************
   SUBROUTINE pw_fft_test(para_env, iw, globenv, pw_transfer_section)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: pw_transfer_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pw_fft_test', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: toler = 1.e-11_dp

      INTEGER                                            :: blocked_id, grid_span, i_layout, i_rep, &
                                                            ig, ip, itmp, n_loop, n_rep, nn, p, q
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: layouts
      INTEGER, DIMENSION(2)                              :: distribution_layout
      INTEGER, DIMENSION(3)                              :: no, np
      INTEGER, DIMENSION(:), POINTER                     :: i_vals
      LOGICAL                                            :: debug, is_fullspace, odd, &
                                                            pw_grid_layout_all, spherical
      REAL(KIND=dp)                                      :: em, et, flops, gsq, perf, t, t_max, &
                                                            t_min, tend, tstart
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: t_end, t_start
      TYPE(cell_type), POINTER                           :: box
      TYPE(pw_grid_type), POINTER                        :: grid
      TYPE(pw_p_type)                                    :: ca, cb, cc

!..set fft lib

      CALL init_fft(globenv%default_fft_library, alltoall=.FALSE., fftsg_sizes=.TRUE., &
                    pool_limit=globenv%fft_pool_scratch_limit, &
                    wisdom_file=globenv%fftw_wisdom_file_name, &
                    plan_style=globenv%fftw_plan_type)

      !..the unit cell (should not really matter, the number of grid points do)
      NULLIFY (box, grid)
      CALL cell_create(box)
      box%hmat = RESHAPE((/10.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, 8.0_dp, 0.0_dp, &
                           0.0_dp, 0.0_dp, 7.0_dp/), (/3, 3/))
      CALL init_cell(box)

      CALL section_vals_get(pw_transfer_section, n_repetition=n_rep)
      DO i_rep = 1, n_rep

         ! how often should we do the transfer
         CALL section_vals_val_get(pw_transfer_section, "N_loop", i_rep_section=i_rep, i_val=N_loop)
         ALLOCATE (t_start(N_loop))
         ALLOCATE (t_end(N_loop))

         ! setup of the grids
         CALL section_vals_val_get(pw_transfer_section, "GRID", i_rep_section=i_rep, i_vals=i_vals)
         np = i_vals

         CALL section_vals_val_get(pw_transfer_section, "PW_GRID_BLOCKED", i_rep_section=i_rep, i_val=blocked_id)
         CALL section_vals_val_get(pw_transfer_section, "DEBUG", i_rep_section=i_rep, l_val=debug)

         CALL section_vals_val_get(pw_transfer_section, "PW_GRID_LAYOUT_ALL", i_rep_section=i_rep, &
                                   l_val=pw_grid_layout_all)

         ! prepare to loop over all or a specific layout
         IF (pw_grid_layout_all) THEN
            ! count layouts that fit
            itmp = 0
            ! start from 2, (/1,para_env%num_pe/) is not supported
            DO p = 2, para_env%num_pe
               q = para_env%num_pe/p
               IF (p*q == para_env%num_pe) THEN
                  itmp = itmp+1
               ENDIF
            ENDDO
            ! build list
            ALLOCATE (layouts(2, itmp))
            itmp = 0
            DO p = 2, para_env%num_pe
               q = para_env%num_pe/p
               IF (p*q == para_env%num_pe) THEN
                  itmp = itmp+1
                  layouts(:, itmp) = (/p, q/)
               ENDIF
            ENDDO
         ELSE
            CALL section_vals_val_get(pw_transfer_section, "PW_GRID_LAYOUT", i_rep_section=i_rep, i_vals=i_vals)
            ALLOCATE (layouts(2, 1))
            layouts(:, 1) = i_vals
         ENDIF

         DO i_layout = 1, SIZE(layouts, 2)

            distribution_layout = layouts(:, i_layout)

            CALL pw_grid_create(grid, para_env%group)

            CALL section_vals_val_get(pw_transfer_section, "PW_GRID", i_rep_section=i_rep, i_val=itmp)

            ! from cp_control_utils
            SELECT CASE (itmp)
            CASE (do_pwgrid_spherical)
               spherical = .TRUE.
               is_fullspace = .FALSE.
            CASE (do_pwgrid_ns_fullspace)
               spherical = .FALSE.
               is_fullspace = .TRUE.
            CASE (do_pwgrid_ns_halfspace)
               spherical = .FALSE.
               is_fullspace = .FALSE.
            END SELECT

            ! from pw_env_methods
            IF (spherical) THEN
               grid_span = HALFSPACE
               spherical = .TRUE.
               odd = .TRUE.
            ELSE IF (is_fullspace) THEN
               grid_span = FULLSPACE
               spherical = .FALSE.
               odd = .FALSE.
            ELSE
               grid_span = HALFSPACE
               spherical = .FALSE.
               odd = .TRUE.
            END IF

            ! actual setup
            CALL pw_grid_setup(box%hmat, grid, grid_span=grid_span, odd=odd, spherical=spherical, &
                               blocked=blocked_id, npts=np, fft_usage=.TRUE., &
                               rs_dims=distribution_layout, iounit=iw)

            IF (iw > 0) CALL m_flush(iw)

            ! note that the number of grid points might be different from what the user requested (fft-able needed)
            no = grid%npts

            NULLIFY (ca%pw)
            NULLIFY (cb%pw)
            NULLIFY (cc%pw)

            CALL pw_create(ca%pw, grid, COMPLEXDATA1D)
            CALL pw_create(cb%pw, grid, COMPLEXDATA3D)
            CALL pw_create(cc%pw, grid, COMPLEXDATA1D)

            ! initialize data
            CALL pw_zero(ca%pw)
            CALL pw_zero(cb%pw)
            CALL pw_zero(cc%pw)
            ca%pw%in_space = RECIPROCALSPACE
            nn = SIZE(ca%pw%cc)
            DO ig = 1, nn
               gsq = grid%gsq(ig)
               ca%pw%cc(ig) = EXP(-gsq)
            END DO

            flops = PRODUCT(no)*30.0_dp*LOG(REAL(MAXVAL(no), KIND=dp))
            tstart = m_walltime()
            DO ip = 1, n_loop
               CALL mp_sync(para_env%group)
               t_start(ip) = m_walltime()
               CALL pw_transfer(ca%pw, cb%pw, debug)
               CALL pw_transfer(cb%pw, cc%pw, debug)
               CALL mp_sync(para_env%group)
               t_end(ip) = m_walltime()
            END DO
            tend = m_walltime()
            t = tend-tstart+threshold
            IF (t > 0.0_dp) THEN
               perf = REAL(n_loop, KIND=dp)*2.0_dp*flops*1.e-6_dp/t
            ELSE
               perf = 0.0_dp
            END IF

            em = MAXVAL(ABS(ca%pw%cc(:)-cc%pw%cc(:)))
            CALL mp_max(em, para_env%group)
            et = SUM(ABS(ca%pw%cc(:)-cc%pw%cc(:)))
            CALL mp_sum(et, para_env%group)
            t_min = MINVAL(t_end-t_start)
            t_max = MAXVAL(t_end-t_start)

            IF (para_env%ionode) THEN
               WRITE (iw, *)
               WRITE (iw, '(A,T67,E14.6)') " Parallel FFT Tests: Maximal Error ", em
               WRITE (iw, '(A,T67,E14.6)') " Parallel FFT Tests: Total Error ", et
               WRITE (iw, '(A,T67,F14.0)') &
                  " Parallel FFT Tests: Performance [Mflops] ", perf
               WRITE (iw, '(A,T67,F14.6)') " Best time : ", t_min
               WRITE (iw, '(A,T67,F14.6)') " Worst time: ", t_max
               IF (iw > 0) CALL m_flush(iw)
            END IF

            ! need debugging ???
            IF (em > toler .OR. et > toler) THEN
               CALL cp__w("library_tests.F",1021,"The FFT results are not accurate ... starting debug pw_transfer")
               CALL pw_transfer(ca%pw, cb%pw, .TRUE.)
               CALL pw_transfer(cb%pw, cc%pw, .TRUE.)
            ENDIF

            ! done with these grids
            CALL pw_release(ca%pw)
            CALL pw_release(cb%pw)
            CALL pw_release(cc%pw)
            CALL pw_grid_release(grid)

         END DO

         ! local arrays
         DEALLOCATE (layouts)
         DEALLOCATE (t_start)
         DEALLOCATE (t_end)

      ENDDO

      ! cleanup
      CALL cell_release(box)
      CALL finalize_fft(para_env, wisdom_file=globenv%fftw_wisdom_file_name)

   END SUBROUTINE pw_fft_test

! **************************************************************************************************
!> \brief Test the parallel (pseudo)random number generator (RNG).
!> \param para_env ...
!> \param output_unit ...
!> \par History
!>      JGH  6-Feb-2001 : Test and performance code
!> \author JGH  1-JAN-2001
! **************************************************************************************************
   SUBROUTINE rng_test(para_env, output_unit)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rng_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      LOGICAL                                            :: ionode
      REAL(KIND=dp)                                      :: t, tend, tmax, tmin, tstart, tsum, tsum2
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      ionode = para_env%ionode
      n = runtest(9)
      NULLIFY (rng_stream)

      ! Check correctness

      CALL check_rng(output_unit, ionode)

      ! Check performance

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,/,T2,A,I10,A)") &
            "Check distributions using", n, " random numbers:"
      END IF

      ! Test uniform distribution [0,1]

      CALL create_rng_stream(rng_stream=rng_stream, &
                             name="Test uniform distribution [0,1]", &
                             distribution_type=UNIFORM, &
                             extended_precision=.TRUE.)

      IF (ionode) THEN
         CALL write_rng_stream(rng_stream, output_unit, write_all=.TRUE.)
      END IF

      tmax = -HUGE(0.0_dp)
      tmin = +HUGE(0.0_dp)
      tsum = 0.0_dp
      tsum2 = 0.0_dp

      tstart = m_walltime()
      DO i = 1, n
         t = next_random_number(rng_stream)
         tsum = tsum+t
         tsum2 = tsum2+t*t
         IF (t > tmax) tmax = t
         IF (t < tmin) tmin = t
      END DO
      tend = m_walltime()

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,(T4,A,F12.6))") &
            "Minimum: ", tmin, &
            "Maximum: ", tmax, &
            "Average: ", tsum/REAL(n, KIND=dp), &
            "Variance:", tsum2/REAL(n, KIND=dp), &
            "Time [s]:", tend-tstart
      END IF

      CALL delete_rng_stream(rng_stream)

      ! Test normal Gaussian distribution

      CALL create_rng_stream(rng_stream=rng_stream, &
                             name="Test normal Gaussian distribution", &
                             distribution_type=GAUSSIAN, &
                             extended_precision=.TRUE.)

      tmax = -HUGE(0.0_dp)
      tmin = +HUGE(0.0_dp)
      tsum = 0.0_dp
      tsum2 = 0.0_dp

      tstart = m_walltime()
      DO i = 1, n
         t = next_random_number(rng_stream)
         tsum = tsum+t
         tsum2 = tsum2+t*t
         IF (t > tmax) tmax = t
         IF (t < tmin) tmin = t
      END DO
      tend = m_walltime()

      IF (ionode) THEN
         CALL write_rng_stream(rng_stream, output_unit)
         WRITE (UNIT=output_unit, FMT="(/,(T4,A,F12.6))") &
            "Minimum: ", tmin, &
            "Maximum: ", tmax, &
            "Average: ", tsum/REAL(n, KIND=dp), &
            "Variance:", tsum2/REAL(n, KIND=dp), &
            "Time [s]:", tend-tstart
      END IF

      CALL delete_rng_stream(rng_stream)

   END SUBROUTINE rng_test

! **************************************************************************************************
!> \brief Tests the eigensolver library routines
!> \param para_env ...
!> \param iw ...
!> \param eigensolver_section ...
!> \par History
!>      JGH  6-Feb-2001 : Test and performance code
!> \author JGH  1-JAN-2001
! **************************************************************************************************
   SUBROUTINE eigensolver_test(para_env, iw, eigensolver_section)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw
      TYPE(section_vals_type), POINTER                   :: eigensolver_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eigensolver_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: diag_method, group, i, i_loop, i_rep, &
                                                            init_method, j, n, n_loop, n_rep, &
                                                            neig, source, unit_number
      REAL(KIND=dp)                                      :: t1, t2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: buffer
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct
      TYPE(cp_fm_type), POINTER                          :: eigenvectors, matrix, work
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      group = para_env%group
      source = para_env%source

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,/,T2,A,/)") "EIGENSOLVER TEST"
      END IF

      ! create blacs env corresponding to para_env
      NULLIFY (blacs_env)
      CALL cp_blacs_env_create(blacs_env=blacs_env, &
                               para_env=para_env)

      ! loop over all tests
      CALL section_vals_get(eigensolver_section, n_repetition=n_rep)
      DO i_rep = 1, n_rep

         ! parse section
         CALL section_vals_val_get(eigensolver_section, "N", i_rep_section=i_rep, i_val=n)
         CALL section_vals_val_get(eigensolver_section, "EIGENVALUES", i_rep_section=i_rep, i_val=neig)
         CALL section_vals_val_get(eigensolver_section, "DIAG_METHOD", i_rep_section=i_rep, i_val=diag_method)
         CALL section_vals_val_get(eigensolver_section, "INIT_METHOD", i_rep_section=i_rep, i_val=init_method)
         CALL section_vals_val_get(eigensolver_section, "N_loop", i_rep_section=i_rep, i_val=n_loop)

         ! proper number of eigs
         IF (neig < 0) neig = n
         neig = MIN(neig, n)

         ! report
         IF (iw > 0) THEN
            WRITE (iw, *) "Matrix size", n
            WRITE (iw, *) "Number of eigenvalues", neig
            WRITE (iw, *) "Timing loops", n_loop
            SELECT CASE (diag_method)
            CASE (do_diag_syevd)
               WRITE (iw, *) "Diag using syevd"
            CASE (do_diag_syevx)
               WRITE (iw, *) "Diag using syevx"
            CASE DEFAULT
               ! stop
            END SELECT

            SELECT CASE (init_method)
            CASE (do_mat_random)
               WRITE (iw, *) "using random matrix"
            CASE (do_mat_read)
               WRITE (iw, *) "reading from file"
            CASE DEFAULT
               ! stop
            END SELECT
         ENDIF

         ! create matrix struct type
         NULLIFY (fmstruct)
         CALL cp_fm_struct_create(fmstruct=fmstruct, &
                                  para_env=para_env, &
                                  context=blacs_env, &
                                  nrow_global=n, &
                                  ncol_global=n)

         ! create all needed matrices, and buffers for the eigenvalues
         NULLIFY (matrix)
         CALL cp_fm_create(matrix=matrix, &
                           matrix_struct=fmstruct, &
                           name="MATRIX")
         CALL cp_fm_set_all(matrix, 0.0_dp)

         NULLIFY (eigenvectors)
         CALL cp_fm_create(matrix=eigenvectors, &
                           matrix_struct=fmstruct, &
                           name="EIGENVECTORS")
         CALL cp_fm_set_all(eigenvectors, 0.0_dp)

         NULLIFY (work)
         CALL cp_fm_create(matrix=work, &
                           matrix_struct=fmstruct, &
                           name="WORK")
         CALL cp_fm_set_all(matrix, 0.0_dp)

         ALLOCATE (eigenvalues(n))
         eigenvalues = 0.0_dp
         ALLOCATE (buffer(1, n))

         ! generate initial matrix, either by reading a file, or using random numbers
         IF (para_env%mepos == para_env%source) THEN
            SELECT CASE (init_method)
            CASE (do_mat_random)
               NULLIFY (rng_stream)
               CALL create_rng_stream(rng_stream=rng_stream, &
                                      name="rng_stream", &
                                      distribution_type=UNIFORM, &
                                      extended_precision=.TRUE.)
            CASE (do_mat_read)
               CALL open_file(file_name="/mnt/sdcard/cp2k/work/MATRIX", &
                              file_action="READ", &
                              file_form="FORMATTED", &
                              file_status="OLD", &
                              unit_number=unit_number)
            END SELECT
         END IF

         DO i = 1, n
            IF (para_env%mepos == para_env%source) THEN
               SELECT CASE (init_method)
               CASE (do_mat_random)
                  DO j = i, n
                     buffer(1, j) = next_random_number(rng_stream)-0.5_dp
                  END DO
                  !MK activate/modify for a diagonal dominant symmetric matrix:
                  !MK buffer(1,i) = 10.0_dp*buffer(1,i)
               CASE (do_mat_read)
                  READ (UNIT=unit_number, FMT=*) buffer(1, 1:n)
               END SELECT
            END IF
            CALL mp_bcast(buffer, source, group)
            SELECT CASE (init_method)
            CASE (do_mat_random)
               CALL cp_fm_set_submatrix(fm=matrix, &
                                        new_values=buffer, &
                                        start_row=i, &
                                        start_col=i, &
                                        n_rows=1, &
                                        n_cols=n-i+1, &
                                        alpha=1.0_dp, &
                                        beta=0.0_dp, &
                                        transpose=.FALSE.)
               CALL cp_fm_set_submatrix(fm=matrix, &
                                        new_values=buffer, &
                                        start_row=i, &
                                        start_col=i, &
                                        n_rows=n-i+1, &
                                        n_cols=1, &
                                        alpha=1.0_dp, &
                                        beta=0.0_dp, &
                                        transpose=.TRUE.)
            CASE (do_mat_read)
               CALL cp_fm_set_submatrix(fm=matrix, &
                                        new_values=buffer, &
                                        start_row=i, &
                                        start_col=1, &
                                        n_rows=1, &
                                        n_cols=n, &
                                        alpha=1.0_dp, &
                                        beta=0.0_dp, &
                                        transpose=.FALSE.)
            END SELECT
         END DO

         DEALLOCATE (buffer)

         IF (para_env%mepos == para_env%source) THEN
            SELECT CASE (init_method)
            CASE (do_mat_random)
               CALL delete_rng_stream(rng_stream=rng_stream)
            CASE (do_mat_read)
               CALL close_file(unit_number=unit_number)
            END SELECT
         END IF

         DO i_loop = 1, n_loop
            eigenvalues = 0.0_dp
            CALL cp_fm_set_all(eigenvectors, 0.0_dp)
            CALL cp_fm_to_fm(source=matrix, &
                             destination=work)

            ! DONE, now testing
            t1 = m_walltime()
            SELECT CASE (diag_method)
            CASE (do_diag_syevd)
               CALL cp_fm_syevd(matrix=work, &
                                eigenvectors=eigenvectors, &
                                eigenvalues=eigenvalues)
            CASE (do_diag_syevx)
               CALL cp_fm_syevx(matrix=work, &
                                eigenvectors=eigenvectors, &
                                eigenvalues=eigenvalues, &
                                neig=neig, &
                                work_syevx=1.0_dp)
            END SELECT
            t2 = m_walltime()
            IF (iw > 0) WRITE (iw, *) "Timing for loop ", i_loop, " : ", t2-t1
         ENDDO

         IF (iw > 0) THEN
            WRITE (iw, *) "Eigenvalues: "
            WRITE (UNIT=iw, FMT="(T3,5F14.6)") eigenvalues(1:neig)
            WRITE (UNIT=iw, FMT="(T3,A4,F16.6)") "Sum:", SUM(eigenvalues(1:neig))
            WRITE (iw, *) ""
         END IF

         ! Clean up
         DEALLOCATE (eigenvalues)
         CALL cp_fm_release(matrix=work)
         CALL cp_fm_release(matrix=eigenvectors)
         CALL cp_fm_release(matrix=matrix)
         CALL cp_fm_struct_release(fmstruct=fmstruct)

      ENDDO

      CALL cp_blacs_env_release(blacs_env=blacs_env)

   END SUBROUTINE eigensolver_test

! **************************************************************************************************
!> \brief Tests the parallel matrix multiply
!> \param para_env ...
!> \param iw ...
!> \param cp_fm_gemm_test_section ...
! **************************************************************************************************
   SUBROUTINE cp_fm_gemm_test(para_env, iw, cp_fm_gemm_test_section)

      TYPE(cp_para_env_type), POINTER          :: para_env
      INTEGER                                  :: iw
      TYPE(section_vals_type), POINTER         :: cp_fm_gemm_test_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_gemm_test', &
                                     routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                         :: transa, transb
      INTEGER :: i_loop, i_rep, k, m, n, N_loop, n_rep, ncol_block, &
                 ncol_block_actual, ncol_global, nrow_block, nrow_block_actual, &
                 nrow_global
      INTEGER, DIMENSION(:), POINTER           :: grid_2d
      LOGICAL                                  :: force_blocksize, row_major, &
                                                  transa_p, transb_p
      REAL(KIND=dp)                            :: t1, t2, t3, t4
      TYPE(cp_blacs_env_type), POINTER         :: blacs_env
      TYPE(cp_fm_struct_type), POINTER         :: fmstruct_a, fmstruct_b, &
                                                  fmstruct_c
      TYPE(cp_fm_type), POINTER                :: matrix_a, matrix_b, matrix_c

      CALL section_vals_get(cp_fm_gemm_test_section, n_repetition=n_rep)
      DO i_rep = 1, n_rep

         ! how often should we do the multiply
         CALL section_vals_val_get(cp_fm_gemm_test_section, "N_loop", i_rep_section=i_rep, i_val=N_loop)

         ! matrices def.
         CALL section_vals_val_get(cp_fm_gemm_test_section, "K", i_rep_section=i_rep, i_val=k)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "N", i_rep_section=i_rep, i_val=n)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "M", i_rep_section=i_rep, i_val=m)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "transa", i_rep_section=i_rep, l_val=transa_p)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "transb", i_rep_section=i_rep, l_val=transb_p)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "nrow_block", i_rep_section=i_rep, i_val=nrow_block)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "ncol_block", i_rep_section=i_rep, i_val=ncol_block)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "ROW_MAJOR", i_rep_section=i_rep, l_val=row_major)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "GRID_2D", i_rep_section=i_rep, i_vals=grid_2d)
         CALL section_vals_val_get(cp_fm_gemm_test_section, "FORCE_BLOCKSIZE", i_rep_section=i_rep, l_val=force_blocksize)
         transa = "N"
         transb = "N"
         IF (transa_p) transa = "T"
         IF (transb_p) transb = "T"

         IF (iw > 0) THEN
            WRITE (iw, '(T2,A)') "----------- TESTING PARALLEL MATRIX MULTIPLY -------------"
            WRITE (iw, '(T2,A)', ADVANCE="NO") "C = "
            IF (transa_p) THEN
               WRITE (iw, '(A)', ADVANCE="NO") "TRANSPOSE(A) x"
            ELSE
               WRITE (iw, '(A)', ADVANCE="NO") "A x "
            ENDIF
            IF (transb_p) THEN
               WRITE (iw, '(A)') "TRANSPOSE(B) "
            ELSE
               WRITE (iw, '(A)') "B "
            ENDIF
            WRITE (iw, '(T2,A,T50,I5,A,I5)') 'requested block size', nrow_block, ' by ', ncol_block
            WRITE (iw, '(T2,A,T50,I5)') 'number of repetitions of cp_fm_gemm ', n_loop
            WRITE (iw, '(T2,A,T50,L5)') 'Row Major', row_major
            WRITE (iw, '(T2,A,T50,2I7)') 'GRID_2D ', grid_2d
            WRITE (iw, '(T2,A,T50,L5)') 'Force blocksize ', force_blocksize
            ! check the return value of pilaenv, too small values limit the performance (assuming pdgemm is the vanilla variant)
         ENDIF

         NULLIFY (blacs_env)
         CALL cp_blacs_env_create(blacs_env=blacs_env, &
                                  para_env=para_env, &
                                  row_major=row_major, &
                                  grid_2d=grid_2d)

         NULLIFY (fmstruct_a)
         IF (transa_p) THEN
            nrow_global = m; ncol_global = k
         ELSE
            nrow_global = k; ncol_global = m
         ENDIF
         CALL cp_fm_struct_create(fmstruct=fmstruct_a, para_env=para_env, context=blacs_env, &
                                  nrow_global=nrow_global, ncol_global=ncol_global, &
                                  nrow_block=nrow_block, ncol_block=ncol_block, force_block=force_blocksize)
         CALL cp_fm_struct_get(fmstruct_a, nrow_block=nrow_block_actual, ncol_block=ncol_block_actual)
         IF (iw > 0) WRITE (iw, '(T2,A,I9,A,I9,A,I5,A,I5)') 'matrix A ', nrow_global, " by ", ncol_global, &
            ' using blocks of ', nrow_block_actual, ' by ', ncol_block_actual

         IF (transb_p) THEN
            nrow_global = n; ncol_global = m
         ELSE
            nrow_global = m; ncol_global = n
         ENDIF
         NULLIFY (fmstruct_b)
         CALL cp_fm_struct_create(fmstruct=fmstruct_b, para_env=para_env, context=blacs_env, &
                                  nrow_global=nrow_global, ncol_global=ncol_global, &
                                  nrow_block=nrow_block, ncol_block=ncol_block, force_block=force_blocksize)
         CALL cp_fm_struct_get(fmstruct_b, nrow_block=nrow_block_actual, ncol_block=ncol_block_actual)
         IF (iw > 0) WRITE (iw, '(T2,A,I9,A,I9,A,I5,A,I5)') 'matrix B ', nrow_global, " by ", ncol_global, &
            ' using blocks of ', nrow_block_actual, ' by ', ncol_block_actual

         NULLIFY (fmstruct_c)
         nrow_global = k
         ncol_global = n
         CALL cp_fm_struct_create(fmstruct=fmstruct_c, para_env=para_env, context=blacs_env, &
                                  nrow_global=nrow_global, ncol_global=ncol_global, &
                                  nrow_block=nrow_block, ncol_block=ncol_block, force_block=force_blocksize)
         CALL cp_fm_struct_get(fmstruct_c, nrow_block=nrow_block_actual, ncol_block=ncol_block_actual)
         IF (iw > 0) WRITE (iw, '(T2,A,I9,A,I9,A,I5,A,I5)') 'matrix C ', nrow_global, " by ", ncol_global, &
            ' using blocks of ', nrow_block_actual, ' by ', ncol_block_actual

         NULLIFY (matrix_a)
         CALL cp_fm_create(matrix=matrix_a, matrix_struct=fmstruct_a, name="MATRIX A")
         NULLIFY (matrix_b)
         CALL cp_fm_create(matrix=matrix_b, matrix_struct=fmstruct_b, name="MATRIX B")
         NULLIFY (matrix_c)
         CALL cp_fm_create(matrix=matrix_c, matrix_struct=fmstruct_c, name="MATRIX C")

         CALL RANDOM_NUMBER(matrix_a%local_data)
         CALL RANDOM_NUMBER(matrix_b%local_data)
         CALL RANDOM_NUMBER(matrix_c%local_data)

         IF (iw > 0) CALL m_flush(iw)

         t1 = m_walltime()
         DO i_loop = 1, N_loop
            t3 = m_walltime()
            CALL cp_gemm(transa, transb, k, n, m, 1.0_dp, matrix_a, matrix_b, 0.0_dp, matrix_c)
            t4 = m_walltime()
            IF (iw > 0) THEN
               WRITE (iw, '(T2,A,T50,F12.6)') "cp_fm_gemm timing: ", (t4-t3)
               CALL m_flush(iw)
            ENDIF
         ENDDO
         t2 = m_walltime()

         IF (iw > 0) THEN
            WRITE (iw, '(T2,A,T50,F12.6)') "average cp_fm_gemm timing: ", (t2-t1)/N_loop
            IF (t2 > t1) THEN
               WRITE (iw, '(T2,A,T50,F12.6)') "cp_fm_gemm Gflops per MPI task: ", &
                  2*REAL(m, kind=dp)*REAL(n, kind=dp)*REAL(k, kind=dp)*N_loop/MAX(0.001_dp, t2-t1)/1.0E9_dp/para_env%num_pe
            ENDIF
         ENDIF

         CALL cp_fm_release(matrix=matrix_a)
         CALL cp_fm_release(matrix=matrix_b)
         CALL cp_fm_release(matrix=matrix_c)
         CALL cp_fm_struct_release(fmstruct=fmstruct_a)
         CALL cp_fm_struct_release(fmstruct=fmstruct_b)
         CALL cp_fm_struct_release(fmstruct=fmstruct_c)
         CALL cp_blacs_env_release(blacs_env=blacs_env)

      ENDDO

   END SUBROUTINE cp_fm_gemm_test

! **************************************************************************************************
!> \brief Tests the DBCSR interface.
!> \param para_env ...
!> \param iw ...
!> \param input_section ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_tests(para_env, iw, input_section)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iw
      TYPE(section_vals_type), POINTER                   :: input_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_tests', routineP = moduleN//':'//routineN

      CHARACTER, DIMENSION(3)                            :: types
      INTEGER                                            :: data_type, i_rep, k, m, n, N_loop, &
                                                            n_rep, test_type
      INTEGER, DIMENSION(:), POINTER                     :: bs_k, bs_m, bs_n, nproc
      LOGICAL                                            :: always_checksum, retain_sparsity, &
                                                            transa_p, transb_p
      REAL(KIND=dp)                                      :: alpha, beta, filter_eps, s_a, s_b, s_c

!   ---------------------------------------------------------------------------

      NULLIFY (bs_m, bs_n, bs_k)
      CALL section_vals_get(input_section, n_repetition=n_rep)
      DO i_rep = 1, n_rep
         ! how often should we do the multiply
         CALL section_vals_val_get(input_section, "N_loop", i_rep_section=i_rep, i_val=N_loop)

         ! matrices def.
         CALL section_vals_val_get(input_section, "TEST_TYPE", i_rep_section=i_rep, i_val=test_type)
         CALL section_vals_val_get(input_section, "DATA_TYPE", i_rep_section=i_rep, i_val=data_type)
         CALL section_vals_val_get(input_section, "K", i_rep_section=i_rep, i_val=k)
         CALL section_vals_val_get(input_section, "N", i_rep_section=i_rep, i_val=n)
         CALL section_vals_val_get(input_section, "M", i_rep_section=i_rep, i_val=m)
         CALL section_vals_val_get(input_section, "transa", i_rep_section=i_rep, l_val=transa_p)
         CALL section_vals_val_get(input_section, "transb", i_rep_section=i_rep, l_val=transb_p)
         CALL section_vals_val_get(input_section, "bs_m", i_rep_section=i_rep, &
                                   i_vals=bs_m)
         CALL section_vals_val_get(input_section, "bs_n", i_rep_section=i_rep, &
                                   i_vals=bs_n)
         CALL section_vals_val_get(input_section, "bs_k", i_rep_section=i_rep, &
                                   i_vals=bs_k)
         CALL section_vals_val_get(input_section, "keepsparse", i_rep_section=i_rep, l_val=retain_sparsity)
         CALL section_vals_val_get(input_section, "asparsity", i_rep_section=i_rep, r_val=s_a)
         CALL section_vals_val_get(input_section, "bsparsity", i_rep_section=i_rep, r_val=s_b)
         CALL section_vals_val_get(input_section, "csparsity", i_rep_section=i_rep, r_val=s_c)
         CALL section_vals_val_get(input_section, "alpha", i_rep_section=i_rep, r_val=alpha)
         CALL section_vals_val_get(input_section, "beta", i_rep_section=i_rep, r_val=beta)
         CALL section_vals_val_get(input_section, "nproc", i_rep_section=i_rep, &
                                   i_vals=nproc)
         CALL section_vals_val_get(input_section, "atype", i_rep_section=i_rep, &
                                   c_val=types(1))
         CALL section_vals_val_get(input_section, "btype", i_rep_section=i_rep, &
                                   c_val=types(2))
         CALL section_vals_val_get(input_section, "ctype", i_rep_section=i_rep, &
                                   c_val=types(3))
         CALL section_vals_val_get(input_section, "filter_eps", &
                                   i_rep_section=i_rep, r_val=filter_eps)
         CALL section_vals_val_get(input_section, "ALWAYS_CHECKSUM", i_rep_section=i_rep, l_val=always_checksum)

         CALL dbcsr_run_tests(para_env%group, iw, nproc, &
                              (/m, n, k/), &
                              (/transa_p, transb_p/), &
                              bs_m, bs_n, bs_k, &
                              (/s_a, s_b, s_c/), &
                              alpha, beta, &
                              data_type=data_type, &
                              test_type=test_type, &
                              n_loops=n_loop, eps=filter_eps, retain_sparsity=retain_sparsity, &
                              always_checksum=always_checksum)
      END DO
   END SUBROUTINE cp_dbcsr_tests

END MODULE library_tests
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for optimize_input
!> \par History
!>      09.2010 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE input_optimize_input
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              low_print_level
   USE input_constants,                 ONLY: opt_force_matching
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_optimize_input'
   PUBLIC :: create_optimize_input_section

CONTAINS

! **************************************************************************************************
!> \brief creates the optimize_input section
!> \param section ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_optimize_input_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_optimize_input_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: sub_section, subsubsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_input.F",52)
      CALL section_create(section, name="OPTIMIZE_INPUT", &
                          description="describes an input optimization job, in which parameters in input files get optimized.", &
                          repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="METHOD", &
                          description="What kind of input optimization to perform.", &
                          usage="METHOD FORCE_MATCHING", &
                          enum_c_vals=s2a("FORCE_MATCHING"), &
                          enum_desc=s2a("Perform a force matching minimization."), &
                          enum_i_vals=(/opt_force_matching/), &
                          default_i_val=opt_force_matching)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ACCURACY", &
                          description="Final accuracy requested in optimization (RHOEND)", &
                          usage="ACCURACY 0.00001", &
                          default_r_val=1.e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="Initial step size for search algorithm (RHOBEG)", &
                          usage="STEP_SIZE 0.005", &
                          default_r_val=0.05_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FUN", &
                          description="Maximum number of function evaluations", &
                          usage="MAX_FUN 1000", &
                          default_i_val=5000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ITER_START_VAL", &
                          description="Used for restarting, starting value of the iteration", &
                          usage="ITER_START_VAL 0", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANDOMIZE_VARIABLES", &
                          description="Percentage randomization of the free variables applied initially", &
                          usage="RANDOMIZE_VARIABLES 20", &
                          default_r_val=0.00_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !
      ! variables section
      !

      NULLIFY (sub_section)
      CALL section_create(sub_section, name="VARIABLE", &
                          description="Defines initial values for variables and their labels", &
                          n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="VALUE", &
                          description="Initial value of the variable", &
                          usage="VALUE 0.0", &
                          type_of_var=real_t, unit_str="internal_cp2k")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIXED", &
                          description="Is this variable fixed or should it be optimized.", &
                          usage="FIXED", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LABEL", &
                          description="The label used in the input file, i.e. ${LABEL} will be replaced by the VALUE specified.", &
                          usage="LABEL PRM01", &
                          type_of_var=char_t)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, sub_section)
      CALL section_release(sub_section)

      !
      ! force matching sub sectiong
      !

      NULLIFY (sub_section)
      CALL section_create(sub_section, name="FORCE_MATCHING", &
                          description="Specify the force matching input.", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="OPTIMIZE_FILE_NAME", &
                          description="the filename of the input file which contains the parameters to be optimized", &
                          usage="OPTIMIZE_FILE_NAME my_input.inp", &
                          default_lc_val="")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REF_TRAJ_FILE_NAME", &
                          description="the filename of the reference coordinates.", &
                          usage="REF_TRAJ_FILE_NAME pos.xyz", &
                          default_lc_val="")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REF_FORCE_FILE_NAME", &
                          description="the filename of the reference forces, should also contain the energy", &
                          usage="REF_FORCE_FILE_NAME frc.xyz", &
                          default_lc_val="")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REF_CELL_FILE_NAME", &
                          description="the filename of the reference cell", &
                          usage="REF_CELL_FILE_NAME project.cell", &
                          default_lc_val="")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GROUP_SIZE", &
                          description="Gives the preferred size of a working group, "// &
                          "groups will always be equal or larger than this size."// &
                          "Usually this should take the number of cores per socket into account for good performance.", &
                          usage="group_size 2", default_i_val=6)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FRAME_START", &
                          description="starting frame to be used from the reference trajectory", &
                          usage="FRAME_START 1", default_i_val=1)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FRAME_STOP", &
                          description="final frame to be used from the reference trajectory (all=-1)", &
                          usage="FRAME_STOP -1", default_i_val=-1)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FRAME_STRIDE", &
                          description="stride when using the reference trajectory", &
                          usage="FRAME_STRIDE 1", default_i_val=1)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FRAME_COUNT", &
                          description="Use at most FRAME_COUNT frames from the reference trajectory, "// &
                          "adjusting the stride to have them as fas apart as possible (all=-1).", &
                          usage="FRAME_COUNT 100", default_i_val=-1)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY_WEIGHT", &
                          description="Relative weight of the energy RMSD vs the force RMSD", &
                          usage="ENERGY_WEIGHT 0.1", default_r_val=0.1_dp)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT_AVERAGE", &
                          description="Shift averages of the energies before computing energy RMSD.", &
                          usage="SHIFT_AVERAGE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT_QM", &
                          description="Shift of the reference energies applied before computing energy RMSD.", &
                          usage="SHIFT_QM -17.0", default_r_val=0.0_dp)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT_MM", &
                          description="Shift of the fit energies applied before computing energy RMSD.", &
                          usage="SHIFT_MM 0.0", default_r_val=0.0_dp)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsubsection)
      CALL cp_print_key_section_create(subsubsection, "COMPARE_ENERGIES", &
                                       description="A comparison of energies between fit and reference", &
                                       print_level=low_print_level, filename="compare_energies", common_iter_levels=1)
      CALL section_add_subsection(sub_section, subsubsection)
      CALL section_release(subsubsection)

      NULLIFY (subsubsection)
      CALL cp_print_key_section_create(subsubsection, "COMPARE_FORCES", &
                                       description="A comparison of forces between fit and reference", &
                                       print_level=low_print_level, filename="compare_forces", common_iter_levels=1)
      CALL section_add_subsection(sub_section, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, sub_section)
      CALL section_release(sub_section)

      NULLIFY (subsubsection)
      CALL cp_print_key_section_create(subsubsection, "HISTORY", &
                                       description="writes a history of the function value and parameters", &
                                       print_level=low_print_level, filename="history", common_iter_levels=1)
      CALL section_add_subsection(section, subsubsection)
      CALL section_release(subsubsection)

      CALL cp_print_key_section_create(subsubsection, "RESTART", &
                                       description="writes an input file that can be used to restart ", &
                                       print_level=low_print_level, filename="optimize", common_iter_levels=1)
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsubsection)
      CALL section_release(subsubsection)

   END SUBROUTINE create_optimize_input_section

END MODULE input_optimize_input

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for optimize_basis
!> \par History
!>      03.2012 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************
MODULE input_optimize_basis

   USE input_constants,                 ONLY: do_opt_all,&
                                              do_opt_coeff,&
                                              do_opt_exps,&
                                              do_opt_none
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_optimize_basis'
   PUBLIC :: create_optimize_basis_section

CONTAINS

! **************************************************************************************************
!> \brief creates the optimize_basis section
!> \param section ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE create_optimize_basis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_optimize_basis_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",55)
      CALL section_create(section, name="OPTIMIZE_BASIS", &
                          description="describes a basis optimization job, in which an ADMM like approach is used to"// &
                          " find the best exponents and/or coefficients to match a given training set.", &
                          repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="BASIS_TEMPLATE_FILE", &
                          description="Name of the basis set file, containing the structure of the new basis set", &
                          usage="BASIS_TEMPLATE_FILE <FILENAME>", &
                          type_of_var=char_t, repeats=.FALSE., &
                          default_c_val="BASIS_SET", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_WORK_FILE", &
                          description="Name of the basis set file which is created to be read as initial guess", &
                          usage="BASIS_WORK_FILE <FILENAME>", &
                          type_of_var=char_t, repeats=.FALSE., &
                          default_c_val="BASIS_WORK_FILE", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_OUTPUT_FILE", &
                          description="Name of the basis set file containing the optimized basis", &
                          usage="BASIS_OUTPUT_FILE <FILENAME>", &
                          type_of_var=char_t, repeats=.FALSE., &
                          default_c_val="BASIS_OUTPUT_FILE", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WRITE_FREQUENCY", &
                          description="Frequency at which the intermediate results should be written", &
                          usage="WRITE_FREQUENCY 1000", &
                          default_i_val=5000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_CONDITION_NUMBER", &
                          description="Determines whether condition number should be part of optimization or not", &
                          usage="USE_CONDITION_NUMBER", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="BASIS_COMBINATIONS", &
         description="If multiple atomic kinds are fitted at the same time, this keyword "// &
         "allows to specify which basis sets should be used together in optimization (underived set ID=0). "// &
         "If skipped all combinations are used. The order is taken as the kinds and sets are specified in the input", &
         repeats=.TRUE., &
         usage="BASIS_COMBINATIONS SET_ID(KIND1) SET_ID(KIND2) ... ", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="RESIDUUM_WEIGHT", &
         description="This keyword allows to give different weight factors to the "// &
         "residuum of the different basis combinations. "// &
         "The first entry corresponds to the original basis sets. Every further value is assigned to the combinations "// &
         "in the order given for BASIS_COMBINATIONS.", &
         repeats=.TRUE., &
         usage="RESIDUUM_WEIGHT REAL ", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="CONDITION_WEIGHT", &
         description="This keyword allows to give different weight factors to the "// &
         "condition number of different basis combinations (LOG(cond) is used). "// &
         "The first entry corresponds to the original basis sets. Every further value is assigned to the combinations "// &
         "in the order given for BASIS_COMBINATIONS.", &
         repeats=.TRUE., &
         usage="CONTITION_WEIGHT REAL ", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GROUP_PARTITION", &
                          description="Allows the specification of the group mpi group sizes in parallel "// &
                          "runs. If less Groups than tasks are speciefied, consecutive calculations "// &
                          " Will be assigned to one group (derived basis sets and then training sets)"// &
                          " If keyword is skipped, equal group sizes will be generated trying to fit all calculations.", &
                          repeats=.TRUE., &
                          usage="GROUP_PARTITION INT INT ... ", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_fit_kinds_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_training_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_powell_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_optimize_basis_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_fit_kinds_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_fit_kinds_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",170)
      CALL section_create(section, name="FIT_KIND", &
                          description="specicifies the atomic kinds to be fitted and the basis"// &
                          " sets associated with the kind.", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The name of the kind described in this section.", &
                          usage="H", default_c_val="DEFAULT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_SET", &
                          description="The name of the basis set for the kind. Has to be specified in BASIS_TEMPLATE_FILE.", &
                          usage="H", default_c_val="DEFAULT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INITIAL_DEGREES_OF_FREEDOM", &
                          description="Specifies the initial degrees of freedom in the basis optimization."// &
                          "This can be used to make further specifications easier", &
                          usage="INITIAL_DEGREES_OF_FREEDOM ALL", &
                          enum_c_vals=s2a("ALL", "NONE", "COEFFICIENTS", "EXPONENTS"), &
                          enum_desc=s2a("Set all parameters in the basis to be variable.", &
                                        "Set all parameters in the basis to be fixed.", &
                                        "Set all coefficients in the basis set to be variable.", &
                                        "Set all exponents in the basis to be variable."), &
                          enum_i_vals=(/do_opt_all, do_opt_none, do_opt_coeff, do_opt_exps/), &
                          default_i_val=do_opt_coeff)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SWITCH_COEFF_STATE", &
                          description="Allows to switch the state of a given coefficient from current state "// &
                          "(varibale/fixed)) to the opposite state. The three integers indicate "// &
                          "the set number, the angular momentum i'th contraction and i'th coefficient", repeats=.TRUE., &
                          usage="SWITCH_COEFF_STATE SET L CONTRACTION IPGF", type_of_var=integer_t, n_var=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SWITCH_CONTRACTION_STATE", &
                          description="Allows to switch the state of a given contraction from current state "// &
                          "(varibale/fixed)) to the opposite state. The three integers indicate "// &
                          "the set number, the angular momentum and i'th contraction ", repeats=.TRUE., &
                          usage="SWITCH_CONTRACTION_STATE SET L CONTRACTION ", type_of_var=integer_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SWITCH_EXP_STATE", &
                          description="Allows to switch the state of a given exponent from current state "// &
                          "(varibale/fixed)) to the opposite state. The two integers indicate "// &
                          "the set number and i'th exponent", repeats=.TRUE., &
                          usage="SWITCH_EXP_STATE SET IEXP", type_of_var=integer_t, n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SWITCH_SET_STATE", &
                          description="Allows to switch the states of in a set from current state "// &
                          "(varibale/fixed)) to the opposite state. The two integers indicate "// &
                          "the affected part (0=ALL,1=EXPS,2=COEFF) and i'th set", repeats=.TRUE., &
                          usage="SWITCH_SET_STATE SET IEXP", type_of_var=integer_t, n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_constrain_exp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_derived_sets_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_fit_kinds_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_derived_sets_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_derived_sets_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",257)
      CALL section_create(section, name="DERIVED_BASIS_SETS", &
                          description="This section can be used to create subsets of a basis "// &
                          " which will be fitted at the same time. This is especially useful if connected"// &
                          " bsis sets e.g. TZVP, DZVP, SZV should be fitted. ", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="REFERENCE_SET", &
                          description="Specifies the reference basis ID which is used as template to create the new set. "// &
                          "The original basis has ID 0. All follwing sets are counted in order as specified in the Input."// &
                          " The decriptors always assume the structure of the input basis set.", &
                          repeats=.FALSE., usage="REFERNCE_SET INTEGER", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REMOVE_CONTRACTION", &
                          description="Can be used to remove a contraction from the reference basis set. "// &
                          "The contraction is speciefied by set number, angular momentum and number of contraction."// &
                          " The decriptors always assume the structure of the input basis set.", &
                          repeats=.TRUE., usage="REMOVE_CONTRACTION SET L ICONTRACTION", type_of_var=integer_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REMOVE_SET", &
                          description="Can be used to remove a set from the reference basis set. ", &
                          repeats=.TRUE., usage="REMOVE_SET SET", type_of_var=integer_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_derived_sets_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_constrain_exp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_constrain_exp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",301)
      CALL section_create(section, name="CONSTRAIN_EXPONENTS", &
                          description="specicifies contraints for the exponents to be fitted."// &
                          " Only a single constraint can be applied to an exponent", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="USE_EXP", &
                          description="Defines the exponent to be constraint. The two integers indicate "// &
                          "the set number and i'th exponent. The value -1 can be used to mark all sets/expoenents in a set.", &
                          repeats=.FALSE., usage="USE_EXP SET IEXP", type_of_var=integer_t, n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BOUNDARIES", &
                          description="Defines the boundaries to which the optimization is restricted."// &
                          " First value is the lower bound, second value is the upper bound.", &
                          repeats=.FALSE., usage="BOUNDARIES LOWER UPPER", type_of_var=real_t, n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_VAR_FRACTION", &
                          description="Defines the maximum fractionr by which the exponent is allowed to vary."// &
                          " e.g. 0.5 allows the exp to vary by 0.5*exp in both directions.", &
                          repeats=.FALSE., usage="MAX_VAR_FRACTION REAL", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_constrain_exp_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_training_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_training_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",343)
      CALL section_create(section, name="TRAINING_FILES", &
                          description="specicifies the location in which the files necessary for"// &
                          " fitting procedure are located. Each Training set needs a reptition of this section.", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="DIRECTORY", &
                          description="the directory in which the files are placed", &
                          usage="DIRECTORY /my/path", &
                          default_lc_val=".")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INPUT_FILE_NAME", &
                          description="the filename of the input file used to run the original calcuation", &
                          usage="INPUT_FILE_NAME my_input.inp", &
                          default_lc_val="input.inp")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_training_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_powell_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_powell_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_optimize_basis.F",378)
      CALL section_create(section, name="OPTIMIZATION", &
                          description="sets the parameters for optimizition, output frequency and restarts", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="ACCURACY", &
                          description="Final accuracy requested in optimization (RHOEND)", &
                          usage="ACCURACY 0.00001", &
                          default_r_val=1.e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="Initial step size for search algorithm (RHOBEG)", &
                          usage="STEP_SIZE 0.005", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FUN", &
                          description="Maximum number of function evaluations", &
                          usage="MAX_FUN 1000", &
                          default_i_val=5000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_powell_section

END MODULE input_optimize_basis
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Declares the input for global optimization
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_input
   USE bibliography,                    ONLY: Goedecker2004
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              low_print_level
   USE input_constants,                 ONLY: glbopt_do_mincrawl,&
                                              glbopt_do_minhop
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_input'

   PUBLIC :: glbopt_declare_input

CONTAINS

! **************************************************************************************************
!> \brief Declares the SWARM%GLOBAL_OPT input section
!> \param swarm_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_declare_input(swarm_section)
      TYPE(section_type), POINTER                        :: swarm_section

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: glbopt_section, printkey

      NULLIFY (glbopt_section, keyword, printkey)

      CALL section_create(glbopt_section, name="GLOBAL_OPT", &
                          description="Section to control global geometry optimizations.", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="METHOD", &
                          description="Methods to use for optimization.", &
                          default_i_val=glbopt_do_minhop, &
                          enum_c_vals=s2a("MINIMA_HOPPING", "MINIMA_CRAWLING"), &
                          enum_desc=s2a("Runs Minima-Hopping algorithm.", &
                                        "Runs Minima-Crawling algorithm."), &
                          enum_i_vals=(/glbopt_do_minhop, glbopt_do_mincrawl/))
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="E_TARGET", &
                          description="Target Energy, the optimization will quit once a lower potential energy is reached.", &
                          default_r_val=-1*HUGE(1.0_dp), type_of_var=real_t, unit_str="hartree")
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MD_BUMPS_MAX", &
                          description="Number of bumps in potential energy after which MD runs ends.", &
                          type_of_var=integer_t, default_i_val=3)
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BUMP_STEPS_UPWARDS", &
                          description="Number of MD steps with potential energy increases required for a bump.", &
                          type_of_var=integer_t, default_i_val=2)
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BUMP_STEPS_DOWNWARDS", &
                          description="Number of MD steps with potential energy decreases required for a bump.", &
                          type_of_var=integer_t, default_i_val=2)
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FRAGMENTATION_THRESHOLD", &
                          description="Threshold for atom distance used for detecting fragmentation of clusters.", &
                          default_r_val=2.0_dp, unit_str="angstrom", type_of_var=real_t)
      CALL section_add_keyword(glbopt_section, keyword)
      CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="MD_ADAPTIVE_TIMESTEP",&
      !     description="Make MD timesteps longer for lower temperatures.", &
      !     default_r_val=0.0_dp, type_of_var=real_t)
      !CALL section_add_keyword(glbopt_section, keyword)
      !CALL keyword_release(keyword)

      CALL cp_print_key_section_create( &
         printkey, "PROGRESS_TRAJECTORY", &
         description="Printkey to controll the writting of the progress trajectory. "// &
         "This trajectory contains the minima, which are lower in energy than the by then lowerest.", &
         print_level=low_print_level, common_iter_levels=1, &
         filename="", unit_str="angstrom")
      CALL section_add_subsection(glbopt_section, printkey)
      CALL section_release(printkey)

      CALL history_declare_input(glbopt_section)
      CALL minhop_declare_input(glbopt_section)
      CALL mincrawl_declare_input(glbopt_section)

      CALL section_add_subsection(swarm_section, glbopt_section)
      CALL section_release(glbopt_section)
   END SUBROUTINE glbopt_declare_input

! **************************************************************************************************
!> \brief Declares the SWARM%GLOBAL_OPT%HISTORY input section
!> \param glbopt_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE history_declare_input(glbopt_section)
      TYPE(section_type), POINTER                        :: glbopt_section

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: history_section

      NULLIFY (history_section, keyword)

      CALL section_create(history_section, name="HISTORY", &
                          description="Section controling the history of visited minima and "// &
                          "how minima are recognized at a later point.", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="ENERGY_PRECISION", &
                          description="If the difference of two energies is below this threshold "// &
                          "they are considert equal.", &
                          default_r_val=1.0e-5_dp, type_of_var=real_t)
      CALL section_add_keyword(history_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FINGERPRINT_PRECISION", &
                          description="If the euclidean distance of two fingerprints is below "// &
                          "this threshold, they are considert equal.", &
                          default_r_val=1.0e-2_dp, type_of_var=real_t)
      CALL section_add_keyword(history_section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(glbopt_section, history_section)
      CALL section_release(history_section)
   END SUBROUTINE history_declare_input

! **************************************************************************************************
!> \brief Declares the SWARM%GLOBAL_OPT%MINIMA_HOPPING input section
!> \param glbopt_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE minhop_declare_input(glbopt_section)
      TYPE(section_type), POINTER                        :: glbopt_section

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: minhop_section

      NULLIFY (minhop_section, keyword)

      CALL section_create(minhop_section, name="MINIMA_HOPPING", &
                          description="Section controlling the Minima Hopping method.", &
                          citations=(/Goedecker2004/), &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="BETA_1", &
                          description="Factor used to increase temperature when escape failed, "// &
                          "should be larger than 1.", &
                          default_r_val=1.1_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BETA_2", &
                          description="Factor used to increase temperature when escape found "// &
                          "known minima, should be larger than 1.", &
                          default_r_val=1.1_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BETA_3", &
                          description="Factor used to decrease temperature when escape succeeded"// &
                          ", should be smaller than 1.", &
                          default_r_val=1.0/1.1_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA_1", &
                          description="Factor used to decrease acceptance energy, when minima was accepted"// &
                          ", should be smaller than 1.", &
                          default_r_val=0.98_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA_2", &
                          description="Factor used to increase acceptance energy, when minima was rejected"// &
                          ", should be larger than 1.", &
                          default_r_val=1.0/0.98_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="E_ACCEPT_INIT", &
                          description="Initial value of acceptance energy", &
                          default_r_val=0.005_dp, type_of_var=real_t, unit_str="hartree")
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE_INIT", &
                          description="Initially temperature in Kelvin", &
                          default_r_val=100.0_dp, type_of_var=real_t)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHARE_HISTORY", &
                          description="If set all worker will use a single share history of visited minima.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(minhop_section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(glbopt_section, minhop_section)
      CALL section_release(minhop_section)
   END SUBROUTINE minhop_declare_input

! **************************************************************************************************
!> \brief Declares the SWARM%GLOBAL_OPT%MINIMA_CRAWLING input section
!> \param glbopt_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mincrawl_declare_input(glbopt_section)
      TYPE(section_type), POINTER                        :: glbopt_section

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: mincrawl_section, printkey

      NULLIFY (mincrawl_section, keyword, printkey)

      CALL section_create(mincrawl_section, name="MINIMA_CRAWLING", &
                          description="Section controls Minima Crawling run.", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="TEMPSTEP_BASE", &
                          description="Base used to calculate temperature steps base**n", &
                          default_r_val=1.1_dp, type_of_var=real_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPSTEP_MAX", &
                          description="Maximum number of temperature steps.", &
                          default_i_val=100, type_of_var=integer_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPDIST_UPDATE_WIDTH", &
                          description="Width of gaussian used to update temperature distribution.", &
                          default_r_val=2.0_dp, type_of_var=real_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPDIST_UPDATE_HEIGHT", &
                          description="Height of gaussian used to update temperature distribution.", &
                          default_r_val=0.1_dp, type_of_var=real_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE_INIT", &
                          description="Initial temperature in Kelvin", &
                          default_r_val=100.0_dp, type_of_var=real_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPDIST_INIT_WIDTH", &
                          description="Initial width of temperature distribution.", &
                          default_r_val=5.0_dp, type_of_var=real_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WORKER_PER_MINIMA", &
                          description="Maximum number of active workers per Minima.", &
                          default_i_val=3, type_of_var=integer_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ESCAPE_HISTORY_LENGTH", &
                          description="Number of escapes averaged for scoring of minima.", &
                          default_i_val=10, type_of_var=integer_t)
      CALL section_add_keyword(mincrawl_section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(printkey, "MINIMA_TRAJECTORY", &
                                       description="Printkey to controll the writting of the minima trajectory. "// &
                                       "This trajectory contains all encountered local minima.", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="", unit_str="angstrom")
      CALL section_add_subsection(mincrawl_section, printkey)
      CALL section_release(printkey)

      CALL section_add_subsection(glbopt_section, mincrawl_section)
      CALL section_release(mincrawl_section)
   END SUBROUTINE mincrawl_declare_input

END MODULE glbopt_input

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Declares the input for swarm framework
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm_input
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              low_print_level
   USE glbopt_input,                    ONLY: glbopt_declare_input
   USE input_constants,                 ONLY: swarm_do_glbopt
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm_input'

   PUBLIC :: create_swarm_section

CONTAINS

! **************************************************************************************************
!> \brief Declares the SWARM input section
!> \param swarm_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE create_swarm_section(swarm_section)
      TYPE(section_type), POINTER                        :: swarm_section

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_section, printkey

      NULLIFY (swarm_section, print_section, printkey, keyword)

      CALL section_create(swarm_section, name="SWARM", &
                          description="Section to control swarm runs. "// &
                          "The swarm framework provides a common ground for master/worker algorithms.", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="BEHAVIOR", &
                          description="Which behaviour should control the swarm.", &
                          usage="BEHAVIOR <STRING>", &
                          default_i_val=swarm_do_glbopt, &
                          enum_c_vals=s2a("GLOBAL_OPT"), &
                          enum_desc=s2a("Runs global geometry optimisation"), &
                          enum_i_vals=(/swarm_do_glbopt/))
      CALL section_add_keyword(swarm_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUMBER_OF_WORKERS", &
                          description="Number of workers used for swarm. "// &
                          "Of the total number of processors one is used for the master, "// &
                          "the remaining processors should be divisible by the number of workers.", &
                          type_of_var=integer_t)
      CALL section_add_keyword(swarm_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REPLAY_COMMUNICATION_LOG", &
                          description="Filename of communication log of previous run. Use this to restart a swarm.", &
                          repeats=.FALSE., &
                          usage="REPLAY_COMMUNICATION_LOG <CHARACTER>", default_lc_val="swarm_translog_replay.xyz")
      CALL section_add_keyword(swarm_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="The maximum number iterations the master should perform", &
                          type_of_var=integer_t, default_i_val=HUGE(1))
      CALL section_add_keyword(swarm_section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(print_section, name="PRINT", &
                          description="Controls the printing properties during a global optimization run", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL cp_print_key_section_create( &
         printkey, "WORKER_RUN_INFO", &
         description="Controls the printing of the worker's basic information during the global optimization", &
         print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(print_section, printkey)
      CALL section_release(printkey)

      CALL cp_print_key_section_create( &
         printkey, "MASTER_RUN_INFO", &
         description="Controls the printing of the masters's basic information during the global optimization", &
         print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(print_section, printkey)
      CALL section_release(printkey)

      CALL cp_print_key_section_create(printkey, "COMMUNICATION_LOG", &
                                       description="Log all the communication between workers and master. Needed for restart.", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="", unit_str="angstrom")
      CALL section_add_subsection(print_section, printkey)
      CALL section_release(printkey)

      CALL section_add_subsection(swarm_section, print_section)
      CALL section_release(print_section)

      CALL glbopt_declare_input(swarm_section)

   END SUBROUTINE create_swarm_section

END MODULE swarm_input

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      10.2005 split input_cp2k into smaller modules [fawzi]
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_thermostats
   USE bibliography,                    ONLY: Bussi2007,&
                                              Ceriotti2009,&
                                              Ceriotti2009b,&
                                              Jones2011,&
                                              Nose1984a,&
                                              Nose1984b
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        do_constr_atomic, do_constr_molec, do_constr_none, do_region_defined, do_region_global, &
        do_region_massive, do_region_molecule, do_region_none, do_thermo_al, do_thermo_csvr, &
        do_thermo_gle, do_thermo_nose, do_thermo_same_as_part
   USE input_cp2k_subsys,               ONLY: create_rng_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_thermostats'

   PUBLIC :: create_thermostat_section, &
             create_thermo_fast_section, &
             create_thermo_slow_section, &
             create_coord_section, &
             create_region_section, &
             create_velocity_section, &
             create_mass_section, &
             create_gle_section

!***
CONTAINS
! **************************************************************************************************
!> \brief Specifies parameter for thermostat for constant temperature ensembles
!> \param section will contain the coeff section
!> \param coupled_thermostat ...
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
   SUBROUTINE create_thermo_slow_section(section, coupled_thermostat)
      TYPE(section_type), POINTER                        :: section
      LOGICAL, INTENT(IN), OPTIONAL                      :: coupled_thermostat

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermo_slow_section', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_coupled_thermostat
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: nose_section, region_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",76)
      my_coupled_thermostat = .FALSE.
      IF (PRESENT(coupled_thermostat)) my_coupled_thermostat = coupled_thermostat
      NULLIFY (nose_section, region_section)

      CALL section_create(section, name="THERMOSTAT_SLOW", &
                          description="Specify thermostat type and parameters controlling the thermostat.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      IF (.NOT. my_coupled_thermostat) THEN
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_nose, &
                             enum_c_vals=s2a("NOSE"), &
                             enum_i_vals=(/do_thermo_nose/), &
                             enum_desc=s2a("Uses only the Nose-Hoover thermostat."))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="REGION", &
                             description="Determines the defined region for slow thermostat", &
                             usage="REGION (GLOBAL||MOLECULE||MASSIVE||DEFINED||NONE)", &
                             enum_c_vals=s2a("GLOBAL", "MOLECULE", "MASSIVE", "DEFINED", "NONE"), &
                             enum_i_vals=(/do_region_global, do_region_molecule, &
                                           do_region_massive, do_region_defined, do_region_none/), &
                             default_i_val=do_region_global)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL create_region_section(region_section, "slow thermostat")
         CALL section_add_subsection(section, region_section)
         CALL section_release(region_section)
      ELSE
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_same_as_part, &
                             enum_c_vals=s2a("SAME_AS_PARTICLE", "NOSE", "CSVR"), &
                             enum_i_vals=(/do_thermo_same_as_part, do_thermo_nose, do_thermo_csvr/), &
                             enum_desc=s2a("Use the same kind of thermostat used for particles.", &
                                           "Uses the Nose-Hoover thermostat.", &
                                           "Uses the canonical sampling through velocity rescaling."))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)
      END IF

      CALL create_nose_section(nose_section)
      CALL section_add_subsection(section, nose_section)
      CALL section_release(nose_section)

      ! Print Section
!       CALL create_print_section(subsection)
!       CALL section_add_subsection(section, subsection)
!       CALL section_release(subsection)

   END SUBROUTINE create_thermo_slow_section

! **************************************************************************************************
!> \brief Specifies parameter for thermostat for constant temperature ensembles
!> \param section will contain the coeff section
!> \param coupled_thermostat ...
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
   SUBROUTINE create_thermo_fast_section(section, coupled_thermostat)
      TYPE(section_type), POINTER                        :: section
      LOGICAL, INTENT(IN), OPTIONAL                      :: coupled_thermostat

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermo_fast_section', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_coupled_thermostat
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: nose_section, region_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",152)
      my_coupled_thermostat = .FALSE.
      IF (PRESENT(coupled_thermostat)) my_coupled_thermostat = coupled_thermostat
      NULLIFY (nose_section, region_section)

      CALL section_create(section, name="THERMOSTAT_FAST", &
                          description="Specify thermostat type and parameters controlling the thermostat.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      IF (.NOT. my_coupled_thermostat) THEN
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_nose, &
                             enum_c_vals=s2a("NOSE"), &
                             enum_i_vals=(/do_thermo_nose/), &
                             enum_desc=s2a("Uses only the Nose-Hoover thermostat."))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="REGION", &
                             description="Determines the defined region for fast thermostat", &
                             usage="REGION (GLOBAL||MOLECULE||MASSIVE||DEFINED||NONE)", &
                             enum_c_vals=s2a("GLOBAL", "MOLECULE", "MASSIVE", "DEFINED", "NONE"), &
                             enum_i_vals=(/do_region_global, do_region_molecule, &
                                           do_region_massive, do_region_defined, do_region_none/), &
                             default_i_val=do_region_global)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL create_region_section(region_section, "fast thermostat")
         CALL section_add_subsection(section, region_section)
         CALL section_release(region_section)
      ELSE
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_same_as_part, &
                             enum_c_vals=s2a("SAME_AS_PARTICLE", "NOSE", "CSVR"), &
                             enum_i_vals=(/do_thermo_same_as_part, do_thermo_nose, do_thermo_csvr/), &
                             enum_desc=s2a("Use the same kind of thermostat used for particles.", &
                                           "Uses the Nose-Hoover thermostat.", &
                                           "Uses the canonical sampling through velocity rescaling."))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)
      END IF

      CALL create_nose_section(nose_section)
      CALL section_add_subsection(section, nose_section)
      CALL section_release(nose_section)

      ! Print Section
!       CALL create_print_section(subsection)
!       CALL section_add_subsection(section, subsection)
!       CALL section_release(subsection)

   END SUBROUTINE create_thermo_fast_section

! **************************************************************************************************
!> \brief Specifies parameter for thermostat for constant temperature ensembles
!> \param section will contain the coeff section
!> \param coupled_thermostat ...
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
   SUBROUTINE create_thermostat_section(section, coupled_thermostat)
      TYPE(section_type), POINTER                        :: section
      LOGICAL, INTENT(IN), OPTIONAL                      :: coupled_thermostat

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermostat_section', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_coupled_thermostat
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: al_section, csvr_section, gle_section, &
                                                            nose_section, region_section, &
                                                            subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",230)
      my_coupled_thermostat = .FALSE.
      IF (PRESENT(coupled_thermostat)) my_coupled_thermostat = coupled_thermostat
      NULLIFY (csvr_section, gle_section, al_section, nose_section, subsection, region_section)

      CALL section_create(section, name="THERMOSTAT", &
                          description="Specify thermostat type and parameters controlling the thermostat.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      IF (.NOT. my_coupled_thermostat) THEN
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_nose, &
                             enum_c_vals=s2a("NOSE", "CSVR", "GLE", "AD_LANGEVIN"), &
                             enum_i_vals=(/do_thermo_nose, &
                                           do_thermo_csvr, do_thermo_gle, do_thermo_al/), &
                             enum_desc=s2a("Uses the Nose-Hoover thermostat.", &
                                           "Uses the canonical sampling through velocity rescaling.", &
                                           "Uses GLE thermostat", &
                                           "Uses adaptive-Langevin thermostat"))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="REGION", &
                             description="Determines the region each thermostat is attached to.", &
                             usage="REGION (GLOBAL||MOLECULE||MASSIVE||DEFINED||NONE)", &
                             enum_c_vals=s2a("GLOBAL", "MOLECULE", "MASSIVE", "DEFINED", "NONE"), &
                             enum_i_vals=(/do_region_global, do_region_molecule, &
                                           do_region_massive, do_region_defined, do_region_none/), &
                             default_i_val=do_region_global)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL create_region_section(region_section, "thermostat")
         CALL section_add_subsection(section, region_section)
         CALL section_release(region_section)
      ELSE
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify the thermostat used for the constant temperature ensembles.", &
                             usage="thermostat NOSE", &
                             default_i_val=do_thermo_same_as_part, &
                             enum_c_vals=s2a("SAME_AS_PARTICLE", "NOSE", "CSVR"), &
                             enum_i_vals=(/do_thermo_same_as_part, do_thermo_nose, do_thermo_csvr/), &
                             enum_desc=s2a("Use the same kind of thermostat used for particles.", &
                                           "Uses the Nose-Hoover thermostat.", &
                                           "Uses the canonical sampling through velocity rescaling."))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)
      END IF

      CALL create_nose_section(nose_section)
      CALL section_add_subsection(section, nose_section)
      CALL section_release(nose_section)

      CALL create_csvr_section(csvr_section)
      CALL section_add_subsection(section, csvr_section)
      CALL section_release(csvr_section)

      CALL create_gle_section(gle_section)
      CALL section_add_subsection(section, gle_section)
      CALL section_release(gle_section)

      CALL create_al_section(al_section)
      CALL section_add_subsection(section, al_section)
      CALL section_release(al_section)

      ! Print Section
      CALL create_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_thermostat_section

! **************************************************************************************************
!> \brief Creates print section for thermostat section
!> \param section ...
!> \author teo [tlaino] - University of Zurich - 02.2008
! **************************************************************************************************
   SUBROUTINE create_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",318)
      NULLIFY (print_key)
      CALL section_create(section, name="PRINT", &
                          description="Collects all print_keys for thermostat", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "THERMOSTAT_INFO", &
                                       description="Controls output information of the corresponding thermostat.", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "TEMPERATURE", &
                                       description="Controls the output of the temperatures of the regions corresponding to "// &
                                       "the present thermostat", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the output of kinetic energy, and potential energy "// &
                                       " of the defined thermostat.", print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_print_section

! **************************************************************************************************
!> \brief Creates a section to arbitrary define a region to thermostat
!> \param section will contain the coeff section
!> \param label ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_region_section(section, label)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'create_region_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",362)

      CALL section_create(section, name="DEFINE_REGION", &
                          description="This section provides the possibility to define arbitrary region "// &
                          " for the "//TRIM(label)//".", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of atoms to thermostat.", &
                          usage="LIST {integer} {integer} .. {integer}", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecules to thermostat", &
                          usage="MOLNAME WAT MEOH", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MM_SUBSYS", &
                          variants=(/"PROTEIN"/), &
                          description="In a QM/MM run all  MM atoms are specified as a whole ensemble to be thermostated", &
                          usage="MM_SUBSYS (NONE|ATOMIC|MOLECULAR)", &
                          enum_c_vals=s2a("NONE", "ATOMIC", "MOLECULAR"), &
                          enum_i_vals=(/do_constr_none, do_constr_atomic, do_constr_molec/), &
                          enum_desc=s2a("Thermostat nothing", &
                                        "Only the MM atoms itself", &
                                        "The full molecule/residue that contains a MM atom"), &
                          default_i_val=do_constr_none, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QM_SUBSYS", &
                          description="In a QM/MM run all QM atoms are specified as a whole ensemble to be thermostated", &
                          usage="QM_SUBSYS (NONE|ATOMIC|MOLECULAR)", &
                          enum_c_vals=s2a("NONE", "ATOMIC", "MOLECULAR"), &
                          enum_desc=s2a("Thermostat nothing", &
                                        "Only the QM atoms itself", &
                                        "The full molecule/residue that contains a QM atom"), &
                          enum_i_vals=(/do_constr_none, do_constr_atomic, do_constr_molec/), &
                          default_i_val=do_constr_none, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_region_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the ewald section
!> \author gloria
! **************************************************************************************************
   SUBROUTINE create_nose_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_nose_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",426)
      CALL section_create(section, name="nose", &
                          description="paramameters of the Nose Hoover thermostat chain", &
                          citations=(/Nose1984a, Nose1984b/))

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="length", &
                          description="length of the Nose-Hoover chain", usage="length integer", &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Yoshida", &
                          description="order of the yoshida integretor used for the thermostat", &
                          usage="Yoshida integer", &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="timecon", &
                          description="timeconstant of the thermostat chain", &
                          usage="timecon <REAL>", &
                          default_r_val=cp_unit_to_cp2k(1000.0_dp, "fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="mts", &
                          variants=s2a("multiple_time_steps", "mult_t_steps"), &
                          description="number of multiple timesteps to be used for the NoseHoover chain", &
                          usage="mts integer", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_coord_section(subsection, "NOSE HOOVER")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_velocity_section(subsection, "NOSE HOOVER")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mass_section(subsection, "NOSE HOOVER")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_force_section(subsection, "NOSE HOOVER")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_nose_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param
!> \author
! **************************************************************************************************
   SUBROUTINE create_gle_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gle_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",494)
      CALL section_create(section, name="GLE", &
                          description="paramameters of the gle thermostat. This section can be generated "// &
                          " from https://epfl-cosmo.github.io/gle4md/index.html?page=matrix ", &
                          citations=(/Ceriotti2009, Ceriotti2009b/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="NDIM", &
                          description="Size of the gle matrix", usage="NDIM 6", &
                          default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A_SCALE", &
                          description="scaling factor for matrix A (for generic matrix A, depends "// &
                          "on the characteristic frequency of the system).", usage="A_SCALE 0.5", &
                          default_r_val=cp_unit_to_cp2k(1.0_dp, "ps^-1"), unit_str="ps^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A_LIST", &
                          description="A matrix The defaults give optimal sampling for most "// &
                          "cristalline and liquid compounds. Generated with the parameters set kv_4-4.a"// &
                          "centered on w_0=40 cm^-1.", usage="A_LIST real real real", &
                          type_of_var=real_t, unit_str="internal_cp2k", &
                          n_var=-1, repeats=.TRUE.)
!             default_r_vals=(/ &
!    1.859575861256e+2_dp,  2.726385349840e-1_dp,  1.152610045461e+1_dp, -3.641457826260e+1_dp,  2.317337581602e+2_dp, &
!   -2.780952471206e-1_dp,  8.595159180871e-5_dp,  7.218904801765e-1_dp, -1.984453934386e-1_dp,  4.240925758342e-1_dp, &
!   -1.482580813121e+1_dp, -7.218904801765e-1_dp,  1.359090212128e+0_dp,  5.149889628035e+0_dp, -9.994926845099e+0_dp, &
!   -1.037218912688e+1_dp,  1.984453934386e-1_dp, -5.149889628035e+0_dp,  2.666191089117e+1_dp,  1.150771549531e+1_dp, &
!    2.180134636042e+2_dp, -4.240925758342e-1_dp,  9.994926845099e+0_dp, -1.150771549531e+1_dp,  3.095839456559e+2_dp /), &
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C_LIST", &
                          description="C matrix", usage="C_LIST real real real", &
                          unit_str="K_e", &
                          type_of_var=real_t, n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_thermo_energy_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_rng_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_gles_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_gle_section

! **************************************************************************************************
!> \brief Creates the gles section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_gles_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gles_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",564)
      CALL section_create(section, name="s", &
                          description="The s variable for GLE used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify s variable for GLE thermostat ", repeats=.FALSE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_gles_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the ewald section
!> \author teo [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
   SUBROUTINE create_csvr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_csvr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",592)
      CALL section_create(section, name="csvr", &
                          description="Parameters of the canonical sampling through velocity rescaling thermostat.", &
                          citations=(/Bussi2007/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="timecon", &
                          description="Time constant of the CSVR thermostat. A small time "// &
                          "constant will result in strong thermostatting (useful for "// &
                          "initial equilibrations) and a large time constant would be adequate "// &
                          "to get weak thermostatting in production runs.", &
                          usage="timecon <REAL>", &
                          default_r_val=cp_unit_to_cp2k(1000.0_dp, "fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_thermo_energy_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_rng_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_csvr_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the adaptive langevin section
!> \author Noam [bernstei]
! **************************************************************************************************
   SUBROUTINE create_al_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_al_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",634)
      CALL section_create(section, name="ad_langevin", &
                          description="Parameters of the adaptive-Langevin thermostat."// &
                          " Known to work with NVT ensemble, but not tested with"// &
                          " other ensembles.  Also tested with FIXED_ATOMS constraints, but"// &
                          " may not work with other constraints (restraints should be OK, but"// &
                          " haven't been well tested)", &
                          citations=(/Jones2011/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="timecon_nh", &
                          description="Time constant of the Nose-Hoover part of the AD_LANGEVIN thermostat. A small time "// &
                          "constant will result in strong thermostatting (useful for "// &
                          "initial equilibrations) and a large time constant would be adequate "// &
                          "to get weak thermostatting in production runs.", &
                          usage="timecon_nh <REAL>", &
                          default_r_val=cp_unit_to_cp2k(1000.0_dp, "fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="timecon_langevin", &
                          description="Time constant of the Langevin part of the AD_LANGEVIN thermostat. A small time "// &
                          "constant will result in strong thermostatting (useful for "// &
                          "initial equilibrations) and a large time constant would be adequate "// &
                          "to get weak thermostatting in production runs.", &
                          usage="timecon_langevin <REAL>", &
                          default_r_val=cp_unit_to_cp2k(1000.0_dp, "fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_thermo_chi_mass_section(subsection, "CHI")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_thermo_chi_mass_section(subsection, "MASS")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_al_section

! **************************************************************************************************
!> \brief Creates the thermostat chi restarting section
!> \param section the section to create
!> \param sec_name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_thermo_chi_mass_section(section, sec_name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*)                                   :: sec_name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermo_chi_mass_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",692)
      CALL section_create(section, name=TRIM(sec_name), &
                          description="Information to initialize the Ad-Langevin thermostat DOF "//TRIM(sec_name), &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify an initial thermostat DOF "//TRIM(sec_name)// &
                          " for Ad-Langevin thermostat.", repeats=.TRUE., &
                          unit_str="fs^-1", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_thermo_chi_mass_section

! **************************************************************************************************
!> \brief Creates the thermostat energy restarting section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_thermo_energy_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermo_energy_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",720)
      CALL section_create(section, name="THERMOSTAT_ENERGY", &
                          description="Information to initialize the CSVR thermostat energy.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify an initial thermostat energy for CSVR thermostat.", &
                          repeats=.TRUE., unit_str="internal_cp2k", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_thermo_energy_section

! **************************************************************************************************
!> \brief Creates the mass section
!> \param section the section to create
!> \param name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_force_section(section, name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_force_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",749)
      CALL section_create(section, name="force", &
                          description="The forces for "//TRIM(name)//" used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify masses of the system", repeats=.FALSE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_force_section

! **************************************************************************************************
!> \brief Creates the mass section
!> \param section the section to create
!> \param name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_mass_section(section, name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mass_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",778)
      CALL section_create(section, name="mass", &
                          description="The masses for "//TRIM(name)//" used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify masses of the system", repeats=.FALSE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mass_section

! **************************************************************************************************
!> \brief Creates the velocity section
!> \param section the section to create
!> \param name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_velocity_section(section, name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_velocity_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",807)
      CALL section_create(section, name="velocity", &
                          description="The velocities for "//TRIM(name)//" used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify velocities of the system", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_velocity_section

! **************************************************************************************************
!> \brief Creates the coord section
!> \param section the section to create
!> \param name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_coord_section(section, name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_coord_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_thermostats.F",836)
      CALL section_create(section, name="coord", &
                          description="The positions for "//TRIM(name)//" used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify positions of the system", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_coord_section

END MODULE input_cp2k_thermostats
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief tree nodes creation, searching, deallocation, references etc.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_stati
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_stati'

   ! IO
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_default_trajectory_file_name = "tmc_trajectory.dat"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_default_restart_out_file_name = "tmc_restart.dat"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_default_restart_in_file_name = "tmc_restart.last"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_energy_worker_out_file_name = "tmc_E_worker.out"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_NMC_worker_out_file_name = "tmc_NMC_worker.out"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_master_out_file_name = "tmc_master.out"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_ana_out_file_name = "tmc_ana.out"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_default_dot_file_name = "tmc_tree.dot"
   CHARACTER(LEN=*), PARAMETER, &
      PUBLIC                   :: tmc_default_unspecified_name = "xxx_unspecified_xxx"

   ! TASK TYPES
   INTEGER, PARAMETER, PUBLIC :: task_type_MC = 1
   INTEGER, PARAMETER, PUBLIC :: task_type_ideal_gas = 2
   INTEGER, PARAMETER, PUBLIC :: task_type_pauling = 3
   INTEGER, PARAMETER, PUBLIC :: task_type_gaussian_adaptation = 4

   !-- communication status --
   !message tags
   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_OK = 0
   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_WAIT_FOR_NEW_TASK = -42

   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_WORKER_INIT = 666

   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_CALCULATING = 1000
   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_FAILED = 998
   INTEGER, PARAMETER, PUBLIC :: TMC_STATUS_STOP_RECEIPT = 999

   INTEGER, PARAMETER, PUBLIC :: TMC_MESSAGE_INT = 1001
   INTEGER, PARAMETER, PUBLIC :: TMC_MASSAGE_REAL = 1002

   INTEGER, PARAMETER, PUBLIC :: TMC_CANCELING_MESSAGE = 1003
   INTEGER, PARAMETER, PUBLIC :: TMC_CANCELING_RECEIPT = 1004

!  INTEGER, PARAMETER, PUBLIC :: TMC_STAT_REQUEST_REJECTED        = 1005

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_APPROX_ENERGY_REQUEST = 1007
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_APPROX_ENERGY_RESULT = 1008

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_START_CONF_REQUEST = 1009
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_START_CONF_RESULT = 1010

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_ENERGY_REQUEST = 1011
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_ENERGY_RESULT = 1012

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_NMC_REQUEST = 1020
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_NMC_RESULT = 1021
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_NMC_BROADCAST = 1022

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_MD_REQUEST = 1030
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_MD_RESULT = 1031
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_MD_BROADCAST = 1032

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_SCF_STEP_ENER_RECEIVE = 2011

   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_INIT_ANALYSIS = 3000
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_ANALYSIS_REQUEST = 3001
   INTEGER, PARAMETER, PUBLIC :: TMC_STAT_ANALYSIS_RESULT = 3002

!  INTEGER, PARAMETER, PUBLIC :: TMC_STAT_SYNC_RND_SEED           = 1040

END MODULE tmc_stati

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      06.2013 split input_cp2k into smaller modules
!> \author Mandes
! **************************************************************************************************
MODULE input_cp2k_tmc
   USE bibliography,                    ONLY: Schonherr2014
   USE input_cp2k_mm,                   ONLY: create_CHARGE_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE tmc_stati,                       ONLY: tmc_default_dot_file_name,&
                                              tmc_default_unspecified_name
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_tmc'

   PUBLIC :: create_TMC_section

CONTAINS

! **************************************************************************************************
!> \brief creates the TreeMonteCarlo subsection
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("tmc/input_cp2k_tmc.F",54)
      CALL section_create( &
         section, name="TMC", &
         description="A parallelized MC algorithm, presampling the configurations, espacially the Markov chain elements", &
         citations=(/Schonherr2014/), &
         n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      ! size of a group calculating the exact energy
      CALL keyword_create(keyword=keyword, &
                          name="GROUP_ENERGY_SIZE", &
                          description="amount of CPUs per group for energy calculation", &
                          usage="GROUP_ENERGY_SIZE {INTEGER}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! size of a group calculating the exact energy
      CALL keyword_create(keyword=keyword, &
                          name="GROUP_ENERGY_NR", &
                          description="amount of groups for exact energy calculation", &
                          usage="GROUP_ENERGY_NR {INTEGER}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! size of a group calculating the exact energy
      CALL keyword_create(keyword=keyword, &
                          name="GROUP_CC_SIZE", &
                          description="amount of of CPUs per group for configurational change", &
                          usage="GROUP_CC_SIZE {INTEGER}", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! size of a group calculating the exact energy
      CALL keyword_create(keyword=keyword, &
                          name="GROUP_ANLYSIS_NR", &
                          description="amount of groups (cores) for analysing the configurations", &
                          usage="GROUP_ANALYSIS_NR {INTEGER}", &
                          default_i_val=1, lone_keyword_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="NUM_MC_ELEM", &
                          description="the minimum Markov Chain elements, to be sampled", &
                          usage="NUM_MC_ELEM {INTEGER}", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! the start value for the random number generator
      CALL keyword_create(keyword=keyword, &
                          name="RND_DETERMINISTIC", &
                          description="the initialisation number for the random number generator", &
                          usage="RND_INIT {INTEGER}", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="TASK_TYPE", &
                          description="Select specialized types. Selectable:"// &
                          "IDEAL_GAS (constant configuration energy E=0.0), ", &
                          usage="TASK_TYPE {OPTION}", &
                          default_c_val="", lone_keyword_c_val=tmc_default_unspecified_name)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="NR_TEMPERATURE", &
                          description="the number of different temperature for parallel tempering", &
                          usage="NR_TEMP {INTEGER}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="TEMPERATURE", &
                          description="one temperatue OR for parallel tempering: "// &
                          "Tmin Tmax or each temperature T1 T2 T3 ..."// &
                          " If every single temperature is specified, "// &
                          "do NOT use keyword NR_TEMPERATURE", &
                          usage="TEMPERATURE {REAL} |OR| TEMPERATURE {REAL} {REAL} ...", &
                          default_r_vals=(/330.0_dp/), &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="NUM_MV_ELEM_IN_CELL", &
                          description="the number of elements (atoms or molecules) "// &
                          "moves in cell or sub box."// &
                          "if 0 all elements are moved once in a MC move", &
                          usage="NUM_MV_ELEM_IN_CELL {INTEGER}", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

!       CALL keyword_create(keyword=keyword,&
!            name="NR_NMC_STEPS",&
!            description="the number of Nested Mont Carlo moves with in one MC move "//&
!                        "should be huge enough to reach euilibrium state", &
!            usage="NR_NMC_STEPS {INTEGER}",&
!            default_i_val=-1)
!       CALL section_add_keyword(section,keyword)
!       CALL keyword_release(keyword)

      ! the moves MOVE_TYPE on exact potential
      CALL create_TMC_move_type_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! the moves MOVE_TYPE on approx potential
      CALL create_TMC_NMC_move_type_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, &
                          name="SUB_BOX", &
                          description="specifies the size ot the sub box. "// &
                          "Standard moves only within subbox of random position, "// &
                          "to compensate the potential difference of the approximate potential.", &
                          usage="SUB_BOX {REAL} {REAL} {REAL} OR SUB_BOX {REAL} for cubic", &
                          default_r_vals=(/-1.0_dp/), &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRESSURE", &
                          description="enables NPT calculation with specified constant pressure [bar]", &
                          usage="PRESSURE {REAL}", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="VOLUME_ISOTROPIC", &
                          description="volume move is equal in each direction", &
                          usage="VOLUME_ISOTROPIC {LOGICAL}", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="MOVE_CENTER_OF_MASS", &
                          description="Moves the center of mass of defined molecules (in volume moves)", &
                          usage="MOVE_CENTER_OF_MASS {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="ESIMATE_ACC_PROB", &
                          description="set the estimation of the acceptance probability using run time information of the energy", &
                          usage="ESIMATE_ACC_PROB {LOGICAL}", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="SPECULATIVE_CANCELING", &
                          description="enables or disables the speculative canceling. If we have further knowledge of "// &
                          "acceptance probabilities using parent acceptance or the estimated energy.", &
                          usage="SPECULATIVE_CANCELING {LOGICAL}", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="USE_SCF_ENERGY_INFO", &
                          description="enables or disables the usage of SCF energy information for "// &
                          " estimating the acceptance probability. ", &
                          usage="USE_SCF_ENERGY_INFO {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="RESULT_LIST_IN_MEMORY", &
                          description="enables the storing of the whole Markov Chain", &
                          usage="RESULT_LIST_IN_MEMORY {LOGICAL}", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="INFO_OUT_STEP_SIZE", &
                          description="the number the amount of calculated configurations between to output printings.", &
                          usage="INFO_OUT_STEP_SIZE {INTEGER}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="RESTART_IN", &
                          description="if existing use the last restart file", &
                          usage="RESTART or RSTART {FILENAME}", &
                          default_c_val="", lone_keyword_c_val=tmc_default_unspecified_name)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="RESTART_OUT", &
                          description="Defines the frequency to write restart files. "// &
                          "If no frequency is specified (lone keyword) "// &
                          "the restart file is written at the end (only). "// &
                          "If the value is 0, no restart file is written at all. "// &
                          "The frequency specifies is related "// &
                          "to the calculated Markov chain elements", &
                          usage="RESTART or RESTART {INTEGER}", &
                          default_i_val=-1, lone_keyword_i_val=-9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY_FILE_NAME", &
                          description="input file name for the exact potential energy calculation.", &
                          usage="ENERGY_FILE_NAME {filename}", &
                          default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

!       CALL keyword_create(keyword, name="NMC_FILE_NAME",&
!            description="input file name for the approximate potential for Nested Monte Carlo.",&
!            usage="NMC_FILE_NAME {filename}",&
!            default_c_val="",lone_keyword_c_val=tmc_default_unspecified_name)
!       CALL section_add_keyword(section,keyword)
!       CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_ONLY_ACC", &
                          description="printing only accepted elements of the Markov Chain.", &
                          usage="PRINT_ONLY_ACC {LOGICAL}", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_COORDS", &
                          description="printing coordinates of the Markov Chain elements", &
                          usage="PRINT_COORDS {LOGICAL}", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_FORCES", &
                          description="printing forces of the Markov Chain elements", &
                          usage="PRINT_FORCES {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_DIPOLE", &
                          description="enables the calculation and printing the exact cell dipoles"// &
                          " (only for QS methods)", &
                          usage="PRINT_DIPOLE {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_CELL", &
                          description="printing the cell vectors of the Markov Chain elements", &
                          usage="PRINT_CELL {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_ENERGIES", &
                          description="printing the different calculated energies (approximated, scf and exact)", &
                          usage="PRINT_ENERGIES {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DOT_TREE", &
                          description="file name for GrapgViz dot file", &
                          usage="DOT_TREE {filename}", &
                          default_c_val="", lone_keyword_c_val=tmc_default_dot_file_name)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALL_CONF_FILE_NAME", &
                          description="file name for printing every single calculated configuration (e.g. for fitting).", &
                          usage="ALL_CONF_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="PRINT_TEST_OUTPUT", &
                          description="printing different values for regtest comparison", &
                          usage="PRINT_TEST_OUTPUT {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_TMC_ana_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_TMC_ana_files(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_TMC_section

! **************************************************************************************************
!> \brief creates the TMC section to select the move types performed within the
!>        NMC (on approximate potential)
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_NMC_move_type_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_NMC_move_type_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("tmc/input_cp2k_tmc.F",384)
      CALL section_create(section, name="NMC_MOVES", &
                          description="This section specifies the TMC move types, "// &
                          "which are performed within the nested Monte Carlo (NMC). "// &
                          "thus using the approximate potential.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword=keyword, &
                          name="NR_NMC_STEPS", &
                          description="the number of Nested Mont Carlo moves with in one MC move "// &
                          "should be huge enough to reach euilibrium state", &
                          usage="NR_NMC_STEPS {INTEGER}", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NMC_FILE_NAME", &
                          description="input file name for the approximate potential for Nested Monte Carlo.", &
                          usage="NMC_FILE_NAME {filename}", &
                          default_c_val="", lone_keyword_c_val=tmc_default_unspecified_name)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROB", &
                          description="Defines the probability of the NMC move "// &
                          "(considering the ration between the selected other moves)"// &
                          "the probabilities of the move types in the NMC section "// &
                          "defines only the weight within the NMC steps", &
                          usage="PROB {real}", type_of_var=real_t, &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INIT_ACC_PROB", &
                          description="Defines the initial probability of accepting the move. ", &
                          usage="INIT_ACC_PROB {real}", type_of_var=real_t, &
                          default_r_val=0.5_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! the moves types
      CALL create_TMC_move_type_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_TMC_NMC_move_type_section

! **************************************************************************************************
!> \brief creates the TMC section to select the move types
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_move_type_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_move_type_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("tmc/input_cp2k_tmc.F",445)
      CALL section_create(section, name="MOVE_TYPE", &
                          description="This section specifies the TMC move type, "// &
                          "and its properties. "// &
                          "Selectable types are: "// &
                          "ATOM_TRANS atom translation, "// &
                          "MOL_TRANS molecule translation, "// &
                          "MOL_ROT molecule rotation, "// &
                          "PROT_REORDER proton reordering, "// &
                          "PT_SWAP Parallel Tempering swap, "// &
                          "VOL_MOVE volume change, "// &
                          "ATOM_SWAP swaps two atoms of different type.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The name of the move type described in this section.", &
                          usage="ATOM_TRANS", default_c_val="DEFAULT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIZE", &
                          description="Defines the size of the move:"// &
                          "ATOM_TRANS [A], "// &
                          "MOL_TRANS [A], "// &
                          "MOL_ROT [degree], "// &
                          "PROT_REORDER [], "// &
                          "VOL_MOVE [A], "// &
                          "ATOM_SWAP", &
                          usage="SIZE {real}", type_of_var=real_t, &
                          default_r_val=-1.0_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROB", &
                          description="Defines the probability of the move "// &
                          "(considering the ration between the selected moves)", &
                          usage="PROB {real}", type_of_var=real_t, &
                          default_r_val=1.0_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INIT_ACC_PROB", &
                          description="Defines the initial probability of accepting the move. ", &
                          usage="INIT_ACC_PROB {real}", type_of_var=real_t, &
                          default_r_val=0.23_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the move. "// &
                          "Up to now only used for the atom swap.", &
                          usage="ATOMS {KIND1} {KIND2} ... ", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_TMC_move_type_section

! **************************************************************************************************
!> \brief creates the TreeMonteCarlo subsection
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_ana_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_ana_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("tmc/input_cp2k_tmc.F",516)
      CALL section_create(section, name="TMC_ANALYSIS", &
                          description="Analysing the Markov Chain elments with the specified methods", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      CALL create_TMC_ana_kinds(section=section)

   END SUBROUTINE create_TMC_ana_section

! **************************************************************************************************
!> \brief creates the TreeMonteCarlo subsection
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_ana_files(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_ana_files', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("tmc/input_cp2k_tmc.F",540)
      CALL section_create(section, name="TMC_ANALYSIS_FILES", &
                          description="Analysing the Markov Chain elments with the specified methods, "// &
                          "reading form default or specified files", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      CALL create_TMC_ana_kinds(section=section)

      CALL keyword_create(keyword=keyword, &
                          name="NR_TEMPERATURE", &
                          description="the number of different temperature for parallel tempering", &
                          usage="NR_TEMP {INTEGER}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="TEMPERATURE", &
                          description="one temperatue OR for parallel tempering: "// &
                          "Tmin Tmax or each temperature T1 T2 T3 ..."// &
                          " If every single temperature is specified, "// &
                          "do NOT use keyword NR_TEMPERATURE", &
                          usage="TEMPERATURE {REAL} |OR| TEMPERATURE {REAL} {REAL} ...", &
                          default_r_vals=(/330.0_dp/), &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DIRECTORIES", &
                          description="Analysing multiple directories, "// &
                          "created by standard parallel MC (e.g. using TMC farming ", &
                          usage="DIRECTORIES {DIR1/} {DIR2/} ...", &
                          default_c_vals=(/"./"/), &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FORCE_ENV_FILE", &
                          description="input file name for force env, "// &
                          "to get initial configuration with dimensions and cell", &
                          usage="FORCE_ENV_FILE {filename}", &
                          default_c_val="", lone_keyword_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POSITION_FILE", &
                          description="file name for analysing the position file", &
                          usage="POSITION_FILE {filename}", &
                          default_c_val="", lone_keyword_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_FILE", &
                          description="file name for analysing the cell file", &
                          usage="CELL_FILE {filename}", &
                          default_c_val="", lone_keyword_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIPOLE_FILE", &
                          description="file name for analysing the dipole file", &
                          usage="DIPOLE_FILE {filename}", &
                          default_c_val="", lone_keyword_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="START_ELEM", &
                          description="start analysis at element with number #", &
                          usage="START_ELEM {INTEGER}", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="END_ELEM", &
                          description="end analysis at element with number #", &
                          usage="END_ELEM {INTEGER}", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_TMC_ana_files

! **************************************************************************************************
!> \brief creates the TreeMonteCarlo subsection
!> \param section the section to be created
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE create_TMC_ana_kinds(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TMC_ana_kinds', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword=keyword, &
                          name="RESTART", &
                          description="Enables/disables the reading and writing of "// &
                          " analysis restart files ", &
                          usage="RESTART {LOGICAL}", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PREFIX_ANA_FILES", &
                          description="specifies a prefix for all analysis files.", &
                          usage="ANA_FILES_PREFIX {prefix}", &
                          default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="DENSITY", &
                          description="Mass density in the simulation cell, or if specified in sub cubes", &
                          usage="DENSITY or DENSITY {INTEGER} {INTEGER} {INTEGER}", &
                          default_i_vals=(/1/), &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="G_R", &
                          description="Radial Distribution Function for each pair of atoms"// &
                          "using the amount of specified bins within MAX(cell_lenght)/2", &
                          usage="G_R or G_R {INTEGER}", &
                          default_i_val=-1, lone_keyword_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="CLASSICAL_DIPOLE_MOMENTS", &
                          description="calculates the classical dipole Moment. "// &
                          "Following flag specifies if they should be written. "// &
                          "Class. Dip. Mom. are also used to unfold the exact dipole moment. ", &
                          usage="CLASSICAL_DIPOLE_MOMENTS or CLASSICAL_DIPOLE_MOMENTS {LOGICAL}", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! for calculating the classical dipole moment we need charges
      CALL create_CHARGE_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword=keyword, &
                          name="DIPOLE_ANALYSIS", &
                          description="Enables analysis of dipole moments, espacially dielectric constant. "// &
                          "An additional type can be specified, e.g. analyzing ice structures."// &
                          "using SYM_XYZ also dipoles (-x,y,z) .. .. (-x,-y,z).... (-x,-y-z) "// &
                          "are regarded, only use it if this configurations have "// &
                          "all the same energy. ", &
                          usage="DIPOLE_ANALYSIS or DIPOLE_ANALYSIS {type}", &
                          default_c_val="", lone_keyword_c_val=tmc_default_unspecified_name)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="DEVIATION", &
                          description="Calculates the deviation of the position from the last configuration", &
                          usage="DEVIATION {LOGICAL}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_TMC_ana_kinds

END MODULE input_cp2k_tmc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - taken out of input_cp2k_motion
!> \author Ole Schuett
! **************************************************************************************************

MODULE input_cp2k_neb
   USE bibliography,                    ONLY: Elber1987,&
                                              Jonsson1998,&
                                              Jonsson2000_1,&
                                              Jonsson2000_2,&
                                              Wales2004
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        band_diis_opt, band_md_opt, do_b_neb, do_ci_neb, do_d_neb, do_eb, do_it_neb, &
        do_rep_blocked, do_rep_interleaved, do_sm, pot_neb_fe, pot_neb_full, pot_neb_me
   USE input_cp2k_thermostats,          ONLY: create_coord_section,&
                                              create_velocity_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_neb'

   PUBLIC :: create_band_section

CONTAINS

! **************************************************************************************************
!> \brief creates the section for a BAND run
!> \param section will contain the pint section
!> \author Teodoro Laino 09.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE create_band_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_band_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsubsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_neb.F",66)
      CALL section_create(section, name="band", &
                          description="The section that controls a BAND run", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Elber1987, Jonsson1998, Jonsson2000_1, Jonsson2000_2, Wales2004/))
      NULLIFY (keyword, print_key, subsection, subsubsection)

      CALL keyword_create(keyword, name="NPROC_REP", &
                          description="Specify the number of processors to be used per replica "// &
                          "environment (for parallel runs)", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROC_DIST_TYPE", &
                          description="Specify the topology of the mapping of processors into replicas.", &
                          usage="PROC_DIST_TYPE (INTERLEAVED|BLOCKED)", &
                          enum_c_vals=s2a("INTERLEAVED", &
                                          "BLOCKED"), &
                          enum_desc=s2a("Interleaved distribution", &
                                        "Blocked distribution"), &
                          enum_i_vals=(/do_rep_interleaved, do_rep_blocked/), &
                          default_i_val=do_rep_blocked)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BAND_TYPE", &
                          description="Specifies the type of BAND calculation", &
                          usage="BAND_TYPE (B-NEB|IT-NEB|CI-NEB|D-NEB|SM|EB)", &
                          default_i_val=do_it_neb, &
                          enum_c_vals=s2a("B-NEB", &
                                          "IT-NEB", &
                                          "CI-NEB", &
                                          "D-NEB", &
                                          "SM", &
                                          "EB"), &
                          enum_desc=s2a("Bisection nudged elastic band", &
                                        "Improved tangent nudged elastic band", &
                                        "Climbing image nudged elastic band", &
                                        "Doubly nudged elastic band", &
                                        "String Method", &
                                        "Elastic band (Hamiltonian formulation)"), &
                          enum_i_vals=(/do_b_neb, do_it_neb, do_ci_neb, do_d_neb, do_sm, do_eb/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUMBER_OF_REPLICA", &
                          description="Specify the number of Replica to use in the BAND", &
                          default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_COLVARS", &
                          description="Uses a version of the band scheme projected in a subspace of colvars.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POT_TYPE", &
                          description="Specifies the type of potential used in the BAND calculation", &
                          usage="POT_TYPE (FULL|FE|ME)", &
                          default_i_val=pot_neb_full, &
                          enum_c_vals=s2a("FULL", &
                                          "FE", &
                                          "ME"), &
                          enum_desc=s2a("Full potential (no projections in a subspace of colvars)", &
                                        "Free energy (requires a projections in a subspace of colvars)", &
                                        "Minimum energy (requires a projections in a subspace of colvars)"), &
                          enum_i_vals=(/pot_neb_full, pot_neb_fe, pot_neb_me/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROTATE_FRAMES", &
                          description="Compute at each BAND step the RMSD and rotate the frames in order"// &
                          " to minimize it.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALIGN_FRAMES", &
                          description="Enables the alignment of the frames at the beginning of a BAND calculation. "// &
                          "This keyword does not affect the rotation of the replicas during a BAND calculation.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K_SPRING", &
                          variants=(/"K"/), &
                          description="Specify the value of the spring constant", &
                          default_r_val=0.02_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Convergence_control
      CALL section_create(subsection, name="CONVERGENCE_CONTROL", &
                          description="Setup parameters to control the convergence criteria for BAND", &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="MAX_DR", &
                          description="Tolerance on the maximum value of the displacement on the BAND.", &
                          usage="MAX_DR {real}", &
                          default_r_val=0.0002_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FORCE", &
                          description="Tolerance on the maximum value of Forces on the BAND.", &
                          usage="MAX_FORCE {real}", &
                          default_r_val=0.00045_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMS_DR", &
                          description="Tolerance on RMS displacements on the BAND.", &
                          usage="RMS_DR {real}", &
                          default_r_val=0.0001_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMS_FORCE", &
                          description="Tolerance on RMS Forces on the BAND.", &
                          usage="RMS_FORCE {real}", &
                          default_r_val=0.00030_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection, subsubsection)
      ! CI-NEB section
      CALL section_create(subsection, name="CI_NEB", &
                          description="Controls parameters for CI-NEB type calculation only.", &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="NSTEPS_IT", &
                          description="Specify the number of steps of IT-NEB to perform before "// &
                          "switching on the CI algorithm", &
                          default_i_val=5)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! String Method section
      CALL section_create(subsection, name="STRING_METHOD", &
                          description="Controls parameters for String Method type calculation only.", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="SPLINE_ORDER", &
                          description="Specify the oder of the spline used in the String Method.", &
                          default_i_val=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SMOOTHING", &
                          description="Smoothing parameter for the reparametrization of the frames.", &
                          default_r_val=0.2_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Optimization section
      CALL section_create(subsection, name="optimize_band", &
                          description="Specify the optimization method for the band", &
                          repeats=.TRUE.)
      CALL create_opt_band_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! replica section: to specify coordinates and velocities (possibly) of the
      ! different replica used in the BAND
      CALL section_create(subsection, name="replica", &
                          description="Specify coordinates and velocities (possibly) of the replica", &
                          repeats=.TRUE.)
      ! Colvar
      CALL keyword_create(keyword, name="COLLECTIVE", &
                          description="Specifies the value of the collective variables used in the projected"// &
                          " BAND method. The order of the values is the order of the COLLECTIVE section in the"// &
                          " constraints/restraints section", &
                          usage="COLLECTIVE {real} .. {real}", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      ! Coordinates read through an external file
      CALL keyword_create(keyword, name="COORD_FILE_NAME", &
                          description="Name of the xyz file with coordinates (alternative to &COORD section)", &
                          usage="COORD_FILE_NAME <CHAR>", &
                          default_lc_val="")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      ! Coordinates and velocities
      CALL create_coord_section(subsubsection, "BAND")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL create_velocity_section(subsubsection, "BAND")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Print key section
      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the BAND run", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")

      CALL keyword_create(keyword, name="INITIAL_CONFIGURATION_INFO", &
                          description="Print information for the setup of the initial configuration.", &
                          usage="INITIAL_CONFIGURATION_INFO <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "convergence_info", &
                                       description="Controls the printing of the convergence criteria during a BAND run", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "replica_info", &
                                       description="Controls the printing of each replica info during a BAND run", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the printing of the ENER file in a BAND run", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "BANNER", &
                                       description="Controls the printing of the BAND banner", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_band_section

! **************************************************************************************************
!> \brief creates the optimization section for a BAND run
!> \param section will contain the pint section
!> \author Teodoro Laino 02.2007 [tlaino]
! **************************************************************************************************
   SUBROUTINE create_opt_band_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_opt_band_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsubsection

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_neb.F",322)
      NULLIFY (keyword, print_key, subsection, subsubsection)

      CALL keyword_create(keyword, name="OPT_TYPE", &
                          description="Specifies the type optimizer used for the band", &
                          usage="OPT_TYPE (MD|DIIS)", &
                          default_i_val=band_diis_opt, &
                          enum_c_vals=s2a("MD", &
                                          "DIIS"), &
                          enum_desc=s2a("Molecular dynamics-based optimizer", &
                                        "Coupled steepest descent / direct inversion in the iterative subspace"), &
                          enum_i_vals=(/band_md_opt, band_diis_opt/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OPTIMIZE_END_POINTS", &
                          description="Performs also an optimization of the end points of the band.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! MD optimization section
      CALL section_create(subsection, name="MD", &
                          description="Activate the MD based optimization procedure for BAND", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="MAX_STEPS", &
                          description="Specify the maximum number of MD steps", &
                          default_i_val=100)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="timestep", &
         description="The length of an integration step", &
         usage="timestep 1.0", &
         default_r_val=cp_unit_to_cp2k(value=0.5_dp, &
                                       unit_str="fs"), &
         unit_str="fs")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Specify the initial temperature", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, &
                                                        unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      ! Temp_control
      CALL section_create(subsubsection, name="TEMP_CONTROL", &
                          description="Setup parameters to control the temperature during a BAND MD run.", &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Specify the target temperature", &
                          type_of_var=real_t, unit_str="K")
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMP_TOL", &
                          description="Specify the tolerance on the temperature for rescaling", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, &
                                                        unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMP_TOL_STEPS", &
                          description="Specify the number of steps to apply a temperature control", &
                          default_i_val=0)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      ! Vel_control
      CALL section_create(subsubsection, name="VEL_CONTROL", &
                          description="Setup parameters to control the velocity during a BAND MD run.", &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="ANNEALING", &
                          description="Specify the annealing coefficient", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PROJ_VELOCITY_VERLET", &
                          description="Uses a Projected Velocity Verlet instead of a normal Velocity Verlet."// &
                          " Every time the cosine between velocities and forces is < 0 velocities are"// &
                          " zeroed.", &
                          usage="PROJ_VELOCITY_VERLET <LOGICAL>", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SD_LIKE", &
                          description="Zeros velocity at each MD step emulating a steepest descent like"// &
                          "(SD_LIKE) approach", &
                          usage="SD_LIKE <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      ! End of MD
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! DIIS optimization section
      CALL section_create(subsection, name="DIIS", &
                          description="Activate the DIIS based optimization procedure for BAND", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="MAX_SD_STEPS", &
                          description="Specify the maximum number of SD steps to perform"// &
                          " before switching on DIIS (the minimum number will always be equal to N_DIIS).", &
                          default_i_val=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_STEPS", &
                          description="Specify the maximum number of optimization steps", &
                          default_i_val=100)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_DIIS", &
                          variants=(/"NDIIS"/), &
                          description="Number of history vectors to be used with DIIS", &
                          usage="N_DIIS 4", &
                          default_i_val=7)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEPSIZE", &
                          description="Initial stepsize used for the line search, sometimes this parameter"// &
                          "can be reduced to stablize DIIS", &
                          usage="STEPSIZE <REAL>", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_STEPSIZE", &
                          description="Maximum stepsize used for the line search, sometimes this parameter"// &
                          "can be reduced to stablize the LS for particularly difficult initial geometries", &
                          usage="MAX_STEPSIZE <REAL>", &
                          default_r_val=2.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NP_LS", &
                          description="Number of points used in the line search SD.", &
                          usage="NP_LS <INTEGER>", &
                          default_i_val=2)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NO_LS", &
                          description="Does not perform LS during SD. Useful in combination with a proper STEPSIZE"// &
                          " for particularly out of equilibrium starting geometries.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_DIIS", &
                          description="Performes a series of checks on the DIIS solution in order to accept the DIIS step."// &
                          " If set to .FALSE. the only check performed is that the angle between the DIIS solution and the"// &
                          " reference vector is less than Pi/2. Can be useful if many DIIS steps are rejected.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "diis_info", &
                                       description="Controls the printing of DIIS info during a BAND run", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_opt_band_section

END MODULE input_cp2k_neb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      10.2005 split input_cp2k into smaller modules [fawzi]
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_constraints
   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              high_print_level
   USE input_constants,                 ONLY: do_constr_atomic,&
                                              do_constr_molec,&
                                              do_constr_none
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_constraints'

   PUBLIC :: create_constraint_section

!***
CONTAINS

! **************************************************************************************************
!> \brief Create the constraint section. This section is useful to impose
!>      constraints
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_constraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_constraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",67)
      CALL section_create(section, name="constraint", &
                          description="Section specifying information regarding how to impose constraints"// &
                          " on the system.", &
                          n_keywords=0, n_subsections=2, repeats=.FALSE.)

      NULLIFY (subsection, keyword, print_key)
      CALL keyword_create(keyword, name="SHAKE_TOLERANCE", &
                          variants=s2a("SHAKE_TOL", "SHAKE"), &
                          description="Set the tolerance for the shake/rattle constraint algorithm.", &
                          usage="SHAKE_TOLERANCE <REAL>", &
                          default_r_val=1.0E-6_dp, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROLL_TOLERANCE", &
                          variants=s2a("ROLL_TOL", "ROLL"), &
                          description="Set the tolerance for the roll constraint algorithm.", &
                          usage="ROLL_TOLERANCE <REAL>", &
                          default_r_val=1.0E-10_dp, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONSTRAINT_INIT", &
                          description="Apply constraints to the initial position and velocities."// &
                          " Default is to apply constraints only after the first MD step.", &
                          usage="CONSTRAINT_INIT <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_hbonds_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_g3x3_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_g4x6_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_vsite_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_collective_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_fixed_atom_section(subsection)
      CALL restraint_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_f_a_rest_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_clv_rest_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(print_key, "constraint_info", &
                                       description="Prints information about iterative constraints solutions", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "lagrange_multipliers", &
         description="Prints out the lagrange multipliers of the specified constraints during an MD.", &
         print_level=high_print_level, filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_constraint_section

! **************************************************************************************************
!> \brief Create the restart section for colvar restraints
!>      This section will be only used for restraint restarts.
!>      Constraints are handled automatically
!> \param section the section to create
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE create_clv_rest_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_clv_rest_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",166)
      NULLIFY (keyword)
      CALL section_create(section, name="COLVAR_RESTART", &
                          description="Specify restart position only for COLVAR restraints.", &
                          n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The restarting values for COLVAR restraints."// &
                          " The order is an internal order. So if you decide to modify these values by hand"// &
                          " first think what you're doing!", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_clv_rest_section

! **************************************************************************************************
!> \brief Create the restart section for fixed atoms restraints
!>      This section will be only used for restraint restarts.
!>      Constraints are handled automatically
!> \param section the section to create
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE create_f_a_rest_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_f_a_rest_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",197)
      NULLIFY (keyword)
      CALL section_create(section, name="FIX_ATOM_RESTART", &
                          description="Specify restart position only for FIXED_ATOMS restraints.", &
                          n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The restarting position of fixed atoms for restraints."// &
                          " The order is an internal order. So if you decide to modify these values by hand"// &
                          " first think what you're doing!", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_f_a_rest_section

! **************************************************************************************************
!> \brief Create the restraint info section in the constraint section
!> \param section the section to create
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE restraint_info_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'restraint_info_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",227)
      NULLIFY (subsection, keyword)
      CALL section_create(subsection, name="RESTRAINT", &
                          description="Activate and specify information on restraint instead of constraint", &
                          n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="K", &
                          description="Specifies the force constant for the harmonic restraint. The functional "// &
                          "form for the restraint is: K*(X-TARGET)^2.", &
                          usage="K {real}", &
                          type_of_var=real_t, default_r_val=0.0_dp, unit_str="internal_cp2k")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE restraint_info_section

! **************************************************************************************************
!> \brief Create the constraint section for collective constraints
!> \param section the section to create
!> \author Joost VandeVondele [01.2006]
! **************************************************************************************************
   SUBROUTINE create_collective_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_collective_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",259)
      CALL section_create(section, name="COLLECTIVE", &
                          description="Used to constraint collective (general) degrees of freedom, "// &
                          "writing langrangian multipliers to file.", &
                          n_subsections=0, repeats=.TRUE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="COLVAR", &
                          description="Specifies the index (in input file order) of the type of colvar to constrain.", &
                          usage="COLVAR {int}", &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLECULE", &
                          description="Specifies the index of the molecule kind (in input file order)"// &
                          "on which the constraint will be applied."// &
                          " MOLECULE and MOLNAME keyword exclude themself mutually.", &
                          usage="MOLECULE {integer}", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecule on which the constraint will be applied.", &
                          usage="MOLNAME {character}", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERMOLECULAR", &
                          description="Specify if the constraint/restraint is intermolecular.", &
                          usage="INTERMOLECULAR <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="Specifies the target value of the constrained collective"// &
                          " variable (units depend on the colvar).", &
                          usage="TARGET {real}", &
                          type_of_var=real_t, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_GROWTH", &
                          description="Specifies the growth speed of the target value of the constrained collective"// &
                          " variable.", &
                          usage="TARGET_GROWTH {real}", &
                          default_r_val=0.0_dp, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_LIMIT", &
                          description="Specifies the limit of the growth of the target value of the constrained collective"// &
                          " variable. By default no limit at the colvar growth is set.", &
                          usage="TARGET_LIMIT {real}", type_of_var=real_t, &
                          unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not apply the constraint to the QM region within a QM/MM calculation", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not apply the constraint to the MM region within a QM/MM calculation", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_collective_section

! **************************************************************************************************
!> \brief Create the constraint section that fixes atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_fixed_atom_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_fixed_atom_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",348)
      CALL section_create(section, name="fixed_atoms", &
                          description="This section is used to constraint the overall atomic position (X,Y,Z). In case "// &
                          "a restraint is specified the value of the TARGET is considered to be the value of the "// &
                          "coordinates at the beginning of the run or alternatively the corresponding value in the section: "// &
                          "FIX_ATOM_RESTART.", n_keywords=3, n_subsections=0, repeats=.TRUE.)
      NULLIFY (keyword)

      ! Section Parameter
      CALL keyword_create(keyword, name="COMPONENTS_TO_FIX", &
                          description="Specify which components (X,Y,Z or combinations) of the atoms specified in the section "// &
                          "will be constrained/restrained.", &
                          usage="COMPONENTS_TO_FIX (x|y|z|xy|xz|yz|xyz)", &
                          default_i_val=use_perd_xyz, &
                          enum_c_vals=s2a("x", "y", "z", "xy", "xz", "yz", "xyz"), &
                          enum_i_vals=(/use_perd_x, use_perd_y, use_perd_z, &
                                        use_perd_xy, use_perd_xz, use_perd_yz, &
                                        use_perd_xyz/), &
                          enum_desc=s2a("Fix only X component", &
                                        "Fix only Y component", &
                                        "Fix only Z component", &
                                        "Fix X-Y components", &
                                        "Fix X-Z components", &
                                        "Fix Y-Z components", &
                                        "Fix the full components of the atomic position."), &
                          repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Integer
      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of atoms to freeze.", &
                          usage="LIST {integer} {integer} .. {integer}", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecule to fix", &
                          usage="MOLNAME WAT MEOH", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MM_SUBSYS", &
         variants=(/"PROTEIN"/), &
         description="In a QM/MM run all  MM atoms are fixed according to the argument.", &
         usage="MM_SUBSYS (NONE|ATOMIC|MOLECULAR)", &
         enum_c_vals=s2a("NONE", "ATOMIC", "MOLECULAR"), &
         enum_i_vals=(/do_constr_none, do_constr_atomic, do_constr_molec/), &
         enum_desc=s2a("fix nothing", &
                       "only the MM atoms itself", &
                       "the full molecule/residue that contains a MM atom (i.e. some QM atoms might be fixed as well)"), &
         default_i_val=do_constr_none, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="QM_SUBSYS", &
         description="In a QM/MM run all QM atoms are fixed according to the argument.", &
         usage="QM_SUBSYS (NONE|ATOMIC|MOLECULAR)", &
         enum_c_vals=s2a("NONE", "ATOMIC", "MOLECULAR"), &
         enum_desc=s2a("fix nothing", &
                       "only the QM atoms itself", &
                       "the full molecule/residue that contains a QM atom (i.e. some MM atoms might be fixed as well)"), &
         enum_i_vals=(/do_constr_none, do_constr_atomic, do_constr_molec/), &
         default_i_val=do_constr_none, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not apply the constraint to the QM region within a QM/MM calculation."// &
                          " This keyword is active only together with MOLNAME", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not apply the constraint to the MM region within a QM/MM calculation."// &
                          " This keyword is active only together with MOLNAME", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_fixed_atom_section

! **************************************************************************************************
!> \brief Create the constraint section specialized on g3x3 constraints
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_g3x3_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_g3x3_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",451)
      CALL section_create(section, name="g3x3", &
                          description="This section is used to set 3x3 (3 atoms and 3 distances) constraints.", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      ! Integer
      CALL keyword_create(keyword, name="MOLECULE", &
                          variants=(/"MOL"/), &
                          description="Specifies the molecule kind number on which constraint will be applied."// &
                          " MOLECULE and MOLNAME keyword exclude themself mutually.", &
                          usage="MOL {integer}", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecule on which the constraint will be applied.", &
                          usage="MOLNAME {character}", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERMOLECULAR", &
                          description="Specify if the constraint/restraint is intermolecular.", &
                          usage="INTERMOLECULAR <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Atoms' index on which apply the constraint", usage="ATOMS 1 3 6", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Real
      CALL keyword_create(keyword, name="DISTANCES", &
                          description="The constrained distances' values.", &
                          usage="DISTANCES {real} {real} {real}", type_of_var=real_t, &
                          unit_str="internal_cp2k", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logical
      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not apply the constraint to the QM region within a QM/MM calculation", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not apply the constraint to the MM region within a QM/MM calculation", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_g3x3_section

! **************************************************************************************************
!> \brief Create the constraint section specialized on H BONDS constraints
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_hbonds_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hbonds_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",525)
      CALL section_create(section, name="HBONDS", &
                          description="This section is used to set bonds constraints involving Hydrogen atoms", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      ! Character
      CALL keyword_create(keyword, name="ATOM_TYPE", &
                          description="Defines the atoms' type forming a bond with an hydrogen. If not specified "// &
                          " the default bond value of the first molecule is used as constraint target", &
                          usage="ATOMS <CHARACTER>", &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLECULE", &
                          description="Specifies the indexes of the molecule kind (in input file order)"// &
                          "on which the constraint will be applied."// &
                          " MOLECULE and MOLNAME keyword exclude themself mutually.", &
                          usage="MOLECULE {integer} .. {integer} ", n_var=-1, &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the names of the molecule on which the constraint will be applied.", &
                          usage="MOLNAME {character} .. {character} ", n_var=-1, &
                          type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not shake HBONDS in the QM region within a QM/MM calculation", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not shake HBONDS in the MM region within a QM/MM calculation", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Real
      CALL keyword_create(keyword, name="TARGETS", &
                          description="The constrained distances' values  for the types defines in ATOM_TYPE.", &
                          usage="TARGETS {real} {real} {real}", type_of_var=real_t, n_var=-1, &
                          unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_hbonds_section

! **************************************************************************************************
!> \brief Create the constraint section specialized on g4x6 constraints
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_g4x6_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_g4x6_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",594)
      CALL section_create(section, name="g4x6", &
                          description="This section is used to set 4x6 (4 atoms and 6 distances) constraints.", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      ! Integer
      CALL keyword_create(keyword, name="MOLECULE", &
                          variants=(/"MOL"/), &
                          description="Specifies the molecule number on which constraint will be applied."// &
                          " MOLECULE and MOLNAME keyword exclude themself mutually.", &
                          usage="MOL {integer}", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecule on which the constraint will be applied.", &
                          usage="MOLNAME {character}", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERMOLECULAR", &
                          description="Specify if the constraint/restraint is intermolecular.", &
                          usage="INTERMOLECULAR <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Atoms' index on which apply the constraint", usage="ATOMS 1 3 6 4", &
                          n_var=4, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Real
      CALL keyword_create(keyword, name="DISTANCES", &
                          description="The constrained distances' values.", &
                          usage="DISTANCES {real} {real} {real} {real} {real} {real}", &
                          type_of_var=real_t, n_var=6, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logical
      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not apply the constraint to the QM region within a QM/MM calculation", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not apply the constraint to the MM region within a QM/MM calculation", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_g4x6_section

! **************************************************************************************************
!> \brief Create the constraint section specialized on vsite constraints
!> \param section the section to create
!> \author marcel baer
! **************************************************************************************************
   SUBROUTINE create_vsite_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_vsite_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_constraints.F",668)
      CALL section_create(section, name="virtual_site", &
                          description="This section is used to set a virtual interaction-site constraint.", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      ! Integer
      CALL keyword_create(keyword, name="MOLECULE", &
                          variants=(/"MOL"/), &
                          description="Specifies the molecule number on which constraint will be applied."// &
                          " MOLECULE and MOLNAME keyword exclude themself mutually.", &
                          usage="MOL {integer}", n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLNAME", &
                          variants=(/"SEGNAME"/), &
                          description="Specifies the name of the molecule on which the constraint will be applied.", &
                          usage="MOLNAME {character}", n_var=1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERMOLECULAR", &
                          description="Specify if the constraint/restraint is intermolecular.", &
                          usage="INTERMOLECULAR <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Atoms' index on which apply the constraint (v i j k), first is virtual site", &
                          usage="ATOMS 1 2 3 4", &
                          n_var=4, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Real
      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="The constrained paramters' values to construct virtual site."// &
                          "r_v=a*r_ij+b*r_kj", &
                          usage="PARAMETERS {real} {real}", &
                          type_of_var=real_t, n_var=2, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logical
      CALL keyword_create(keyword, name="EXCLUDE_QM", &
                          description="Does not apply the constraint to the QM region within a QM/MM calculation", &
                          usage="EXCLUDE_QM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_MM", &
                          description="Does not apply the constraint to the MM region within a QM/MM calculation", &
                          usage="EXCLUDE_MM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_vsite_section
END MODULE input_cp2k_constraints
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      10.2005 split input_cp2k into smaller modules [fawzi]
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_barostats
   USE barostat_types,                  ONLY: do_clv_x,&
                                              do_clv_xy,&
                                              do_clv_xyz,&
                                              do_clv_xz,&
                                              do_clv_y,&
                                              do_clv_yz,&
                                              do_clv_z
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              high_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_cp2k_thermostats,          ONLY: create_mass_section,&
                                              create_thermostat_section,&
                                              create_velocity_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_barostats'

   PUBLIC :: create_barostat_section

!***
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param section will contain the coeff section
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_barostat_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_barostat_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection, thermo_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/thermostat/input_cp2k_barostats.F",63)
      CALL section_create(section, name="barostat", &
                          description="Parameters of barostat.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection, thermo_section)
      CALL keyword_create(keyword, name="PRESSURE", &
                          description="Initial pressure", &
                          usage="PRESSURE real", &
                          default_r_val=0._dp, unit_str='bar')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TIMECON", &
                          description="Barostat time constant", &
                          usage="TIMECON real", &
                          default_r_val=cp_unit_to_cp2k(1000.0_dp, "fs"), &
                          unit_str='fs')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Barostat initial temperature. If not set, the ensemble temperature is used instead.", &
                          usage="TEMPERATURE real", type_of_var=real_t, &
                          unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMP_TOL", &
                          description="Maximum oscillation of the Barostat temperature imposed by rescaling.", &
                          usage="TEMP_TOL real", default_r_val=0._dp, &
                          unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VIRIAL", &
                          description="For NPT_F only: allows the screening of one or more components of the virial in order"// &
                          " to relax the cell only along specific cartesian axis", &
                          usage="VIRIAL (XYZ | X | Y | Z | XY| XZ | YZ)", &
                          enum_c_vals=s2a("XYZ", "X", "Y", "Z", "XY", "XZ", "YZ"), &
                          enum_i_vals=(/do_clv_xyz, do_clv_x, do_clv_y, do_clv_z, do_clv_xy, do_clv_xz, do_clv_yz/), &
                          default_i_val=do_clv_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_velocity_section(subsection, "BAROSTAT")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mass_section(subsection, "BAROSTAT")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_thermostat_section(thermo_section, coupled_thermostat=.TRUE.)
      CALL section_add_subsection(section, thermo_section)
      CALL section_release(thermo_section)

      CALL create_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_barostat_section

! **************************************************************************************************
!> \brief Creates print section for barostat section
!> \param section ...
!> \author teo [tlaino] - University of Zurich - 02.2008
! **************************************************************************************************
   SUBROUTINE create_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/thermostat/input_cp2k_barostats.F",139)
      NULLIFY (print_key)
      CALL section_create(section, name="PRINT", &
                          description="Collects all print_keys for barostat", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the output of kinetic energy, and potential energy "// &
                                       " of the defined barostat.", print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_print_section

END MODULE input_cp2k_barostats
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - taken out of input_cp2k_motion
!> \author Ole Schuett
! **************************************************************************************************

MODULE input_cp2k_md
   USE bibliography,                    ONLY: &
        Evans1983, Guidon2008, Kantorovich2008, Kantorovich2008a, Kolafa2004, Kuhne2007, &
        Minary2003, Ricci2003, Tuckerman1992, VandeVondele2002
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble, reftraj_ensemble
   USE input_cp2k_barostats,            ONLY: create_barostat_section
   USE input_cp2k_thermostats,          ONLY: create_region_section,&
                                              create_thermo_fast_section,&
                                              create_thermo_slow_section,&
                                              create_thermostat_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_md'

   PUBLIC :: create_md_section

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param section will contain the md section
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_md_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_md_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",71)
      CALL section_create(section, name="MD", &
                          description="This section defines the whole set of parameters needed perform an MD run.", &
                          n_keywords=13, n_subsections=6, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="ensemble", &
                          description="The ensemble/integrator that you want to use for MD propagation", &
                          usage="ensemble nve", &
                          default_i_val=nve_ensemble, &
                          enum_c_vals=s2a("NVE", "NVT", "NPT_I", "NPT_F", "MSST", "MSST_DAMPED", &
                                          "HYDROSTATICSHOCK", "ISOKIN", "REFTRAJ", "LANGEVIN", "NPE_F", &
                                          "NPE_I", "NVT_ADIABATIC"), &
                          enum_desc=s2a("constant energy (microcanonical)", &
                                        "constant temperature and volume (canonical)", &
                                        "constant temperature and pressure using an isotropic cell", &
                                        "constant temperature and pressure using a flexible cell", &
                                        "simulate steady shock (uniaxial)", &
                                        "simulate steady shock (uniaxial) with extra viscosity", &
                                        "simulate steady shock with hydrostatic pressure", &
                                        "constant kinetic energy", &
                                        "reading frames from a file called reftraj.xyz (e.g. for property calculation)", &
                                        "langevin dynamics (constant temperature)", &
                                        "constant pressure ensemble (no thermostat)", &
                                        "constant pressure ensemble using an isotropic cell (no thermostat)", &
                                        "adiabatic dynamics in constant temperature and volume ensemble (CAFES)"), &
                          citations=(/Evans1983, VandeVondele2002, Minary2003, Kolafa2004/), &
                          enum_i_vals=(/nve_ensemble, nvt_ensemble, npt_i_ensemble, npt_f_ensemble, &
                                        nph_uniaxial_ensemble, nph_uniaxial_damped_ensemble, nph_ensemble, isokin_ensemble, &
                                        reftraj_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
                                        nvt_adiabatic_ensemble/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="steps", &
                          description="The number of MD steps to perform", &
                          usage="steps 100", default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="timestep", &
                          description="The length of an integration step (in case RESPA the large TIMESTEP)", &
                          usage="timestep 1.0", default_r_val=cp_unit_to_cp2k(value=0.5_dp, unit_str="fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="step_start_val", &
                          description="The starting step value for the MD", usage="step_start_val <integer>", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="time_start_val", &
                          description="The starting timer value for the MD", &
                          usage="time_start_val <real>", default_r_val=cp_unit_to_cp2k(value=0.0_dp, unit_str="fs"), &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="econs_start_val", &
                          description="The starting  value of the conserved quantity", &
                          usage="econs_start_val <real>", default_r_val=0.0_dp, &
                          unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="temperature", &
                          description="The temperature in K used to initialize "// &
                          "the velocities with init and pos restart, and in the NPT/NVT simulations", &
                          usage="temperature 325.0", default_r_val=cp_unit_to_cp2k(value=300.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_tol", &
                          variants=s2a("temp_to", "temperature_tolerance"), &
                          description="The maximum accepted deviation of the (global) temperature"// &
                          "from the desired target temperature before a rescaling of the velocites "// &
                          "is performed. If it is 0 no rescaling is performed. NOTE: This keyword is "// &
                          "obsolescent; Using a CSVR thermostat with a short timeconstant is "// &
                          "recommended as a better alternative.", &
                          usage="temp_tol 0.0", default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_kind", &
                          description="Compute the temperature per each kind separately", &
                          usage="temp_kind LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="scale_temp_kind", &
                          description="When necessary rescale the temperature per each kind separately", &
                          usage="scale_temp_kind LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="comvel_tol", &
                          description="The maximum accepted velocity of the center of mass. "// &
                          "With Shell-Model, comvel may drift if MD%THERMOSTAT%REGION /= GLOBAL ", &
                          usage="comvel_tol 0.1", type_of_var=real_t, n_var=1, unit_str="bohr*au_t^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="angvel_tol", &
                          description="The maximum accepted angular velocity. This option is ignored "// &
                          "when the system is periodic. Removes the components of the velocities that"// &
                          "project on the external rotational degrees of freedom.", &
                          usage="angvel_tol 0.1", type_of_var=real_t, n_var=1, unit_str="bohr*au_t^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="angvel_zero", &
                          description="Set the initial angular velocity to zero. This option is ignored "// &
                          "when the system is periodic or when initial velocities are defined. Technically, "// &
                          "the part of the random initial velocities that projects on the external "// &
                          "rotational degrees of freedom is subtracted.", &
                          usage="angvel_zero LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANNEALING", &
                          description="Specifies the rescaling factor for annealing velocities. "// &
                          "Automatically enables the annealing procedure. This scheme works only for ensembles "// &
                          "that do not have thermostats on particles.", &
                          usage="annealing <REAL>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANNEALING_CELL", &
                          description="Specifies the rescaling factor for annealing velocities of the CELL "// &
                          "Automatically enables the annealing procedure for the CELL. This scheme works only "// &
                          "for ensambles that do not have thermostat on CELLS velocities.", &
                          usage="ANNEALING_CELL <REAL>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPLACEMENT_TOL", &
                          description="This keyword sets a maximum atomic displacement "// &
                          " in each Cartesian direction."// &
                          "The maximum velocity is evaluated and if it is too large to remain"// &
                          "within the assigned limit, the time step is rescaled accordingly,"// &
                          "and the first half step of the velocity verlet is repeated.", &
                          usage="DISPLACEMENT_TOL <REAL>", default_r_val=100.0_dp, &
                          unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_langevin_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_msst_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_barostat_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_thermostat_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_respa_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_shell_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_adiabatic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_softening_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_reftraj_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_avgs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_thermal_region_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_md_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cascade_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_md_section

! **************************************************************************************************
!> \brief Defines LANGEVIN section
!> \param section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_langevin_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_langevin_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",290)
      CALL section_create(section, name="Langevin", &
                          description="Controls the set of parameters to run a Langevin MD. "// &
                          "The integrator used follows that given in the article by Ricci et al. "// &
                          "The user can define regions in the system where the atoms inside "// &
                          "undergoes Langevin MD, while those outside the regions undergoes "// &
                          "NVE Born Oppenheimer MD. To define the regions, the user should "// &
                          "use THERMAL_REGION subsection of MOTION%MD. The theory for "// &
                          "Langevin MD involving sub-regions can be found in articles by "// &
                          "Kantorovitch et al. All the references can be found in the links below.", &
                          citations=(/Ricci2003, Kantorovich2008, Kantorovich2008a/), &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="gamma", &
                          description="Gamma parameter for the Langevin dynamics (LD)", &
                          usage="gamma 0.001", &
                          default_r_val=0.0_dp, unit_str='fs^-1')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Noisy_Gamma", &
                          variants=(/"NoisyGamma"/), &
                          description="Imaginary Langevin Friction term for LD with noisy forces.", &
                          citations=(/Kuhne2007/), &
                          usage="Noisy_Gamma 4.0E-5", default_r_val=0.0_dp, unit_str='fs^-1')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Shadow_Gamma", &
                          variants=(/"ShadowGamma"/), &
                          description="Shadow Langevin Friction term for LD with noisy forces in order to adjust Noisy_Gamma.", &
                          citations=(/Kuhne2007/), &
                          usage="Shadow_Gamma 0.001", default_r_val=0.0_dp, unit_str='fs^-1')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_langevin_section

! **************************************************************************************************
!> \brief Defines print section for MD
!> \param section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_md_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_md_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",342)
      CALL section_create(section, name="print", &
                          description="Controls the printing properties during an MD run", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (print_key, keyword)

      CALL keyword_create(keyword, name="FORCE_LAST", &
                          description="Print the output and restart file if walltime is reached or "// &
                          "if an external EXIT command is given. It still requires the keyword LAST "// &
                          "to be present for the specific print key (in case the last step should not "// &
                          "match with the print_key iteration number).", &
                          usage="FORCE_LAST LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the output the ener file", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SHELL_ENERGY", &
                                       description="Controls the output of the shell-energy file (only if shell-model)", &
                                       print_level=medium_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "TEMP_KIND", &
                                       description="Controls the output of the temperature"// &
                                       " computed separately for each kind", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "TEMP_SHELL_KIND", &
                                       description="Controls the output of the temperature of the"// &
                                       " shell-core motion computed separately for each kind", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CENTER_OF_MASS", &
                                       description="Controls the printing of COM velocity during an MD", &
                                       print_level=medium_print_level, common_iter_levels=1, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COEFFICIENTS", &
                                       description="Controls the printing of coefficients during an MD run.", &
                                       print_level=medium_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ROTATIONAL_INFO", &
                                       description="Controls the printing basic info during the calculation of the "// &
                                       "translational/rotational degrees of freedom.", print_level=low_print_level, &
                                       add_last=add_last_numeric, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="COORDINATES", &
                          description="Prints atomic coordinates in the standard orientation. "// &
                          "Coordinates are not affected during the calculation.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of basic and summary information during the"// &
                                       " Molecular Dynamics", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_md_print_section

! **************************************************************************************************
!> \brief Defines parameters for RESPA integration scheme
!> \param section will contain the coeff section
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_respa_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_respa_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",436)

      CALL section_create(section, name="RESPA", &
                          description="Multiple timestep integration based on RESPA (implemented for NVE only)."// &
                          " RESPA exploits multiple force_eval."// &
                          " In this case the order of the force_eval maps "// &
                          " the order of the respa shells from the slowest to the fastest force evaluation."// &
                          " If force_evals share the same subsys, it's enough then to specify the "// &
                          " subsys in the force_eval corresponding at the first index in the multiple_force_eval list."// &
                          " Can be used to speedup classical and ab initio MD simulations.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Tuckerman1992, Guidon2008/))

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="FREQUENCY", &
                          description="The number of reference MD steps between two RESPA corrections.", &
                          usage="FREQUENCY <INTEGER>", default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_respa_section

! **************************************************************************************************
!> \brief Defines parameters for REFTRAJ analysis
!> \param section will contain the coeff section
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_reftraj_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_reftraj_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",472)

      CALL section_create(section, name="REFTRAJ", &
                          description="Loads an external trajectory file and performs analysis on the"// &
                          " loaded snapshots.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, print_key, subsection)

      CALL keyword_create(keyword, name="TRAJ_FILE_NAME", &
                          description="Specify the filename where the trajectory is stored.", repeats=.FALSE., &
                          usage="TRAJ_FILE_NAME <CHARACTER>", default_lc_val="reftraj.xyz")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_FILE_NAME", &
                          description="Specify the filename where the cell is stored "// &
                          "(for trajectories generated within variable cell ensembles).", repeats=.FALSE., &
                          usage="CELL_FILE_NAME <CHARACTER>", default_lc_val="reftraj.cell")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="VARIABLE_VOLUME", &
         description="Enables the possibility to read a CELL file with information on the CELL size during the MD.", &
         repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIRST_SNAPSHOT", &
                          description="Index of the snapshot stored in the trajectory file "// &
                          "from which to start a REFTRAJ run", &
                          repeats=.FALSE., usage="FIRST_SNAPSHOT <INTEGER>", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAST_SNAPSHOT", &
                          description="Index of the last snapshot stored in the trajectory file that is read along a REFTRAJ run", &
                          repeats=.FALSE., usage="LAST_SNAPSHOT", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRIDE", &
                          description=" Stride in number of snapshot for the  reftraj analysis", &
                          repeats=.FALSE., usage="STRIDE", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eval_energy_forces", &
                          description="Evaluate energy and forces for each retrieved snapshot during a REFTRAJ run", &
                          repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_msd_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="print", &
                          description="The section that controls the output of a reftraj run", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "msd_kind", &
                                       description="Controls the output of msd per kind", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "msd_molecule", &
                                       description="Controls the output of msd per molecule kind", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "displaced_atom", &
                                       description="Controls the output of index and dislacement of "// &
                                       "atoms that moved away from the initial position of more than a"// &
                                       "given distance (see msd%disp_tol)", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_reftraj_section

! **************************************************************************************************
!> \brief Defines parameters for MSD calculation along a REFTRAJ analysis
!> \param section will contain the coeff section
!> \author MI
! **************************************************************************************************
   SUBROUTINE create_msd_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_msd_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",577)

      CALL section_create(section, name="MSD", &
                          description="Loads an external trajectory file and performs analysis on the"// &
                          " loaded snapshots.", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of core-level spectroscopy simulations", &
                          usage="&MSD T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REF0_FILENAME", &
                          description="Specify the filename where the initial reference configuration is stored.", &
                          repeats=.FALSE., usage="REF0_FILENAME <CHARACTER>", default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MSD_PER_KIND", &
                          description="Set up the calculation of the MSD for each atomic kind", &
                          usage="MSD_PER_KIND <LOGICAL>", repeats=.FALSE., &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MSD_PER_MOLKIND", &
                          description="Set up the calculation of the MSD for each molecule kind."// &
                          "The position of the center of mass of the molecule is considered.", &
                          usage="MSD_PER_MOLKIND <LOGICAL>", repeats=.FALSE., &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MSD_PER_REGION", &
                          description="Set up the calculation of the MSD for each defined region.", &
                          usage="MSD_PER_REGION <LOGICAL>", repeats=.FALSE., &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_region_section(subsection, "MSD calculation")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="DISPLACED_ATOM", &
                          description="Identify the atoms that moved from their initial"// &
                          "position of a distance larger than a given tolerance (see msd%displacement_tol).", &
                          usage="DISPLACED_ATOM <LOGICAL>", repeats=.FALSE., &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="displacement_tol", &
                          description="Lower limit to define displaced atoms", &
                          usage="DISPLACEMENT_TOL real", &
                          default_r_val=0._dp, n_var=1, unit_str='bohr')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_msd_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the coeff section
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_msst_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_msst_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",656)

      CALL section_create(section, name="msst", &
                          description="Parameters for Multi-Scale Shock Technique (MSST) "// &
                          "which simulate the effect of a steady planar shock on a unit cell. "// &
                          "Reed et. al. Physical Review Letters 90, 235503 (2003).", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="PRESSURE", &
                          description="Initial pressure", &
                          usage="PRESSURE real", &
                          default_r_val=0._dp, n_var=1, unit_str='bar')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY", &
                          description="Initial energy", &
                          usage="ENERGY real", &
                          default_r_val=0._dp, n_var=1, unit_str='hartree')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VOLUME", &
                          description="Initial volume", &
                          usage="VOLUME real", &
                          default_r_val=0._dp, n_var=1, unit_str='angstrom^3')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CMASS", &
                          description="Effective cell mass", &
                          usage="CMASS real", &
                          default_r_val=0._dp, n_var=1, unit_str='au_m')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VSHOCK", variants=(/"V_SHOCK"/), &
                          description="Velocity shock", &
                          usage="VSHOCK real", &
                          default_r_val=0._dp, n_var=1, unit_str='m/s')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GAMMA", &
                          description="Damping coefficient for cell volume", &
                          usage="GAMMA real", &
                          unit_str='fs^-1', &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_msst_section

! **************************************************************************************************
!> \brief section will contain some parameters for the shells dynamics
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_shell_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shell_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: thermo_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",723)

      CALL section_create(section, name="shell", &
                          description="Parameters of shell model in adiabatic dynamics.", &
                          n_keywords=4, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, thermo_section)

      CALL keyword_create(keyword=keyword, name="temperature", &
                          description="Temperature in K used to control "// &
                          "the internal velocities of the core-shell motion ", &
                          usage="temperature 5.0", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_tol", &
                          description="Maximum accepted temperature deviation"// &
                          " from the expected value, for the internal core-shell motion."// &
                          "If 0, no rescaling is performed", &
                          usage="temp_tol 0.0", default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="nose_particle", &
                          description="If nvt or npt, the core and shell velocities are controlled "// &
                          "by the same thermostat used for the particle. This might favour heat exchange "// &
                          "and additional rescaling of the internal core-shell velocity is needed (TEMP_TOL)", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPLACEMENT_SHELL_TOL", &
                          description="This keyword sets a maximum variation of the shell "// &
                          "core distance in each Cartesian direction."// &
                          "The maximum internal core-shell velocity is evaluated and"// &
                          " if it is too large to remain"// &
                          "within the assigned limit, the time step is rescaled accordingly,"// &
                          "and the first half step of the velocity verlet is repeated.", &
                          usage="DISPLACEMENT_SHELL_TOL <REAL>", default_r_val=100.0_dp, &
                          unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_thermostat_section(thermo_section)
      CALL section_add_subsection(section, thermo_section)
      CALL section_release(thermo_section)

   END SUBROUTINE create_shell_section

! **************************************************************************************************
!> \brief section will contain some parameters for the adiabatic dynamics
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_adiabatic_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_adiabatic_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: thermo_fast_section, thermo_slow_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",787)

      CALL section_create(section, name="ADIABATIC_DYNAMICS", &
                          description="Parameters used in canonical adiabatic free energy sampling (CAFES).", &
                          n_keywords=5, n_subsections=2, repeats=.FALSE., &
                          citations=(/VandeVondele2002/))

      NULLIFY (keyword, thermo_fast_section, thermo_slow_section)

      CALL keyword_create(keyword=keyword, name="temp_fast", &
                          description="Temperature in K used to control "// &
                          "the fast degrees of freedom ", &
                          usage="temp_fast 5.0", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="temp_slow", &
                          description="Temperature in K used to control "// &
                          "the slow degrees of freedom ", &
                          usage="temp_slow 5.0", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_tol_fast", &
                          description="Maximum accepted temperature deviation"// &
                          " from the expected value, for the fast motion."// &
                          "If 0, no rescaling is performed", &
                          usage="temp_tol 0.0", default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_tol_slow", &
                          description="Maximum accepted temperature deviation"// &
                          " from the expected value, for the slow motion."// &
                          "If 0, no rescaling is performed", &
                          usage="temp_tol 0.0", default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="n_resp_fast", &
                          description="number of respa steps for fast degrees of freedom", &
                          repeats=.FALSE., default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_thermo_fast_section(thermo_fast_section)
      CALL section_add_subsection(section, thermo_fast_section)
      CALL section_release(thermo_fast_section)

      CALL create_thermo_slow_section(thermo_slow_section)
      CALL section_add_subsection(section, thermo_slow_section)
      CALL section_release(thermo_slow_section)

   END SUBROUTINE create_adiabatic_section

! **************************************************************************************************
!> \brief section will contain parameters for the velocity softening
!> \param section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE create_softening_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_softening_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="VELOCITY_SOFTENING", &
                          description="A method to initialize the velocities along low-curvature " &
                          //"directions in order to favors MD trajectories to cross rapidly over " &
                          //"small energy barriers into neighboring basins. " &
                          //"In each iteration the forces are calculated at a point y, which " &
                          //"is slightly displaced from the current positions x in the direction " &
                          //"of the original velocities v. The velocities are then updated with " &
                          //"the force component F_t, which is perpendicular to N. " &
                          //"N = v / |v|;  y = x + delta * N;  F_t = F(y) - &lang; F(y) | N &rang; * N; " &
                          //"v' = v + alpha * F_t")

      NULLIFY (keyword)
      CALL keyword_create(keyword=keyword, name="STEPS", &
                          description="Number of softening iterations performed. " &
                          //"Typical values are around 40 steps.", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="DELTA", &
                          description="Displacement used to obtain y.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="ALPHA", &
                          description="Mixing factor used for updating velocities.", &
                          default_r_val=0.15_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_softening_section

! **************************************************************************************************
!> \brief input section used to define regions with different temperature
!>        initialization and control
!> \param section ...
!> \par History
!>   - Added input for langevin regions in thermal regions section
!>     (2014/02/04, LT)
! **************************************************************************************************
   SUBROUTINE create_thermal_region_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermal_region_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, region_section, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",909)

      CALL section_create(section, name="thermal_region", &
                          description="Define regions where different initialization and control "// &
                          "of the temperature is used. When MOTION%MD%ENSEMBLE is set to LANGEVIN, "// &
                          "this section controls if the atoms defined inside and outside the "// &
                          "thermal regions should undergo Langevin MD or NVE Born-Oppenheimer MD. "// &
                          "The theory behind Langevin MD using different regions can be found in "// &
                          "articles by Kantorovitch et al. listed below.", &
                          citations=(/Kantorovich2008, Kantorovich2008a/), &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (region_section)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword=keyword, name="force_rescaling", &
                          description="Control the rescaling ot the velocities in all the regions, "// &
                          "according to the temperature assigned to each reagion, when "// &
                          "RESTART_VELOCITY in EXT_RESTART is active.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="do_langevin_default", &
                          description="If ENSEMBLE is set to LANGEVIN, controls whether the "// &
                          "atoms NOT defined in the thermal regions to undergo langevin MD "// &
                          "or not. If not, then the atoms will undergo NVE Born-Oppenheimer MD.", &
                          usage="do_langevin_default .FALSE.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(region_section, name="DEFINE_REGION", &
                          description="This section provides the possibility to define arbitrary region ", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of atoms belonging to the region.", &
                          usage="LIST {integer} {integer} .. {integer}", &
                          repeats=.TRUE., n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(region_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="temperature", &
                          description="The temperature in K used to initialize the velocities "// &
                          "of the atoms in this region ", &
                          usage="temperature 5.0", &
                          default_r_val=cp_unit_to_cp2k(value=0.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(region_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="temp_tol", &
                          description="Maximum accepted temperature deviation from the expected "// &
                          "value for this region. If temp_tol=0 no rescaling is performed", &
                          usage="temp_tol 0.0", &
                          default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(region_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="do_langevin", &
                          description="When ENSEMBLE is set to LANGEVIN, Controls whether "// &
                          "the atoms in the thermal region should undergo Langevin MD. If "// &
                          "not, then they will undergo NVE Born-Oppenheimer MD.", &
                          usage="do_langevin .TRUE.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(region_section, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, region_section)
      CALL section_release(region_section)

      NULLIFY (print_key)
      CALL section_create(subsection, name="PRINT", &
                          description="Collects all print_keys for thermal_regions", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "TEMPERATURE", &
                                       description="Controls output of temperature per region.", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "LANGEVIN_REGIONS", &
                                       description="Controls output of information on which atoms "// &
                                       "underwent Langevin MD and which atoms did not.", &
                                       print_level=high_print_level, &
                                       filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_thermal_region_section

! **************************************************************************************************
!> \brief   Defines the parameters for the setup of a cascade simulation
!> \param section ...
!> \date    03.02.2012
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE create_cascade_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_cascade_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword)
      NULLIFY (subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",1026)

      CALL section_create(section, name="CASCADE", &
                          description="Defines the parameters for the setup of a cascade simulation.", &
                          n_keywords=1, &
                          n_subsections=1, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the CASCADE section.", &
                          usage="&CASCADE on", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY", &
                          description="Total energy transferred to the system during the cascade event.", &
                          usage="ENERGY 20.0", &
                          default_r_val=0.0_dp, &
                          unit_str="keV")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="ATOM_LIST", &
                          description="Defines a list of atoms for which the initial velocities are modified", &
                          n_keywords=1, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Defines the list of atoms for which the velocities are modified. "// &
                          "Each record consists of the atomic index, the velocity vector, and "// &
                          "a weight to define which fraction of the total energy is assigned "// &
                          "to the current atom:"// &
                          "<p><tt><big>Atomic_index v<sub>x</sub> v<sub>y</sub> v<sub>x</sub> "// &
                          "Weight</big></tt></p>", &
                          usage="{{Integer} {Real} {Real} {Real} {Real}}", &
                          repeats=.TRUE., &
                          type_of_var=lchar_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_cascade_section

! **************************************************************************************************
!> \brief Defines AVERAGES section
!> \param section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_avgs_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_avgs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",1089)
      CALL section_create(section, name="Averages", &
                          description="Controls the calculation of the averages during an MD run.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword, print_key, subsection)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Controls the calculations of the averages.", &
                          usage="&AVERAGES T", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ACQUISITION_START_TIME", &
                          description="Setup up the simulation time when the acquisition process to compute "// &
                          " averages is started.", &
                          usage="ACQUISITION_START_TIME <REAL>", &
                          default_r_val=0.0_dp, unit_str='fs')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVERAGE_COLVAR", &
                          description="Switch for computing the averages of COLVARs.", &
                          usage="AVERAGE_COLVAR <LOGICAL>", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "PRINT_AVERAGES", &
                                       description="Controls the output the averaged quantities", &
                                       print_level=debug_print_level+1, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_avgs_restart_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_avgs_section

! **************************************************************************************************
!> \brief Defines the AVERAGES RESTART section
!> \param section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_avgs_restart_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_avgs_restart_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_md.F",1141)
      CALL section_create(section, name="RESTART_AVERAGES", &
                          description="Stores information for restarting averages.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ITIMES_START", &
                          description="TIME STEP starting the evaluation of averages", &
                          usage="ITIMES_START <INTEGER>", type_of_var=integer_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVECPU", &
                          description="CPU average", usage="AVECPU <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEHUGONIOT", &
                          description="HUGONIOT average", usage="AVEHUGONIOT <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVETEMP_BARO", &
                          description="BAROSTAT TEMPERATURE average", usage="AVETEMP_BARO <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEPOT", &
                          description="POTENTIAL ENERGY average", usage="AVEPOT <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEKIN", &
                          description="KINETIC ENERGY average", usage="AVEKIN <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVETEMP", &
                          description="TEMPERATURE average", usage="AVETEMP <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEKIN_QM", &
                          description="QM KINETIC ENERGY average in QMMM runs", usage="AVEKIN_QM <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVETEMP_QM", &
                          description="QM TEMPERATURE average in QMMM runs", usage="AVETEMP_QM <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEVOL", &
                          description="VOLUME average", usage="AVEVOL <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVECELL_A", &
                          description="CELL VECTOR A average", usage="AVECELL_A <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVECELL_B", &
                          description="CELL VECTOR B average", usage="AVECELL_B <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVECELL_C", &
                          description="CELL VECTOR C average", usage="AVECELL_C <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEALPHA", &
                          description="ALPHA cell angle average", usage="AVEALPHA <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEBETA", &
                          description="BETA cell angle average", usage="AVEBETA <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVEGAMMA", &
                          description="GAMMA cell angle average", usage="AVEGAMMA <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_ECONS", &
                          description="CONSTANT ENERGY average", usage="AVE_ECONS <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PRESS", &
                          description="PRESSURE average", usage="AVE_PRESS <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PXX", &
                          description="P_{XX} average", usage="AVE_PXX <REAL>", &
                          type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_VIR", &
                          description="PV VIRIAL average", usage="AVE_PV_VIR <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_TOT", &
                          description="PV TOTAL average", usage="AVE_PV_TOT <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_KIN", &
                          description="PV KINETIC average", usage="AVE_PV_KIN <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_CNSTR", &
                          description="PV CONSTRAINTS average", usage="AVE_PV_CNSTR <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_XC", &
                          description="PV XC average", usage="AVE_PV_XC <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_PV_FOCK_4C", &
                          description="PV XC average", usage="AVE_PV_FOCK_4C <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=9)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_COLVARS", &
                          description="COLVARS averages", usage="AVE_COLVARS <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVE_MMATRIX", &
                          description="METRIC TENSOR averages", usage="AVE_MMATRIX <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_avgs_restart_section

END MODULE input_cp2k_md
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      10.2005 split input_cp2k into smaller modules [fawzi]
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_motion
   USE bibliography,                    ONLY: Byrd1995,&
                                              Ceriotti2010,&
                                              Ceriotti2012,&
                                              Ceriotti2014,&
                                              Henkelman1999,&
                                              Kapil2016
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        default_bfgs_method_id, default_cell_direct_id, default_cell_geo_opt_id, &
        default_cell_md_id, default_cg_method_id, default_dimer_method_id, &
        default_lbfgs_method_id, default_minimization_method_id, default_ts_method_id, &
        do_mc_gemc_npt, do_mc_gemc_nvt, do_mc_traditional, do_mc_virial, fmt_id_pdb, fmt_id_xyz, &
        gaussian, helium_cell_shape_cube, helium_cell_shape_octahedron, helium_forces_average, &
        helium_forces_last, helium_mdist_exponential, helium_mdist_gaussian, helium_mdist_linear, &
        helium_mdist_quadratic, helium_mdist_singlev, helium_mdist_uniform, &
        helium_sampling_ceperley, helium_solute_intpot_mwater, helium_solute_intpot_none, &
        integrate_exact, integrate_numeric, ls_2pnt, ls_3pnt, ls_fit, ls_gold, ls_none, &
        matrix_init_cholesky, matrix_init_diagonal, numerical, perm_cycle, perm_plain, &
        propagator_pimd, propagator_rpmd, transformation_normal, transformation_stage
   USE input_cp2k_constraints,          ONLY: create_constraint_section
   USE input_cp2k_free_energy,          ONLY: create_fe_section
   USE input_cp2k_md,                   ONLY: create_md_section
   USE input_cp2k_motion_print,         ONLY: add_format_keyword,&
                                              create_motion_print_section
   USE input_cp2k_neb,                  ONLY: create_band_section
   USE input_cp2k_subsys,               ONLY: create_rng_section
   USE input_cp2k_thermostats,          ONLY: create_coord_section,&
                                              create_gle_section,&
                                              create_velocity_section
   USE input_cp2k_tmc,                  ONLY: create_TMC_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              logical_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_motion'

   PUBLIC :: create_motion_section

CONTAINS

! **************************************************************************************************
!> \brief creates the motion section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_motion_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_motion_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",86)
      CALL section_create(section, name="motion", &
                          description="This section defines a set of tool connected with the motion of the nuclei.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (subsection)

      CALL create_geoopt_section(subsection, label="GEO_OPT", &
                                 description="This section sets the environment of the geometry optimizer.", &
                                 just_optimizers=.FALSE.)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cell_opt_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_shellcore_opt_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_md_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_driver_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_fe_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_constraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_fp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_TMC_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_pint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_band_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_motion_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_motion_section

! **************************************************************************************************
!> \brief creates the Monte Carlo section
!> \param section the section to be created
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_mc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",163)
      CALL section_create(section, name="mc", &
                          description="This section sets parameters to set up a MonteCarlo calculation.", &
                          n_keywords=10, n_subsections=2, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="NSTEP", &
                          description="Specifies the number of MC cycles.", &
                          usage="NSTEP {integer}", &
                          default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IPRINT", &
                          description="Prints coordinate/cell/etc information every IPRINT steps.", &
                          usage="IPRINT {integer}", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NMOVES", &
                          description="Specifies the number of classical moves between energy evaluations. ", &
                          usage="NMOVES {integer}", &
                          default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NSWAPMOVES", &
                          description="How many insertions to try per swap move.", &
                          usage="NSWAPMOVES {integer}", &
                          default_i_val=16)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LBIAS", &
                          description="Dictates if we presample moves with a different potential.", &
                          usage="LBIAS {logical}", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LSTOP", &
                          description="Makes nstep in terms of steps, instead of cycles.", &
                          usage="LSTOP {logical}", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LDISCRETE", &
                          description="Changes the volume of the box in discrete steps, one side at a time.", &
                          usage="LDISCRETE {logical}", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART", &
                          description="Read initial configuration from restart file.", &
                          usage="RESTART {logical}", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="NVIRIAL", &
         description="Use this many random orientations to compute the second virial coefficient (ENSEMBLE=VIRIAL)", &
         usage="NVIRIAL {integer}", &
         default_i_val=1000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENSEMBLE", &
                          description="Specify the type of simulation", &
                          usage="PROGRAM (TRADITIONAL|GEMC_NVT|GEMC_NPT|VIRIAL)", &
                          enum_c_vals=s2a("TRADITIONAL", "GEMC_NVT", "GEMC_NPT", "VIRIAL"), &
                          enum_i_vals=(/do_mc_traditional, do_mc_gemc_nvt, do_mc_gemc_npt, do_mc_virial/), &
                          default_i_val=do_mc_traditional)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_FILE_NAME", &
                          description="Name of the restart file for MC information.", &
                          usage="RESTART_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOVES_FILE_NAME", &
                          description="The file to print the move statistics to.", &
                          usage="MOVES_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOLECULES_FILE_NAME", &
                          description="The file to print the number of molecules to.", &
                          usage="MOLECULES_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORDINATE_FILE_NAME", &
                          description="The file to print the current coordinates to.", &
                          usage="COORDINATE_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY_FILE_NAME", &
                          description="The file to print current energies to.", &
                          usage="ENERGY_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DATA_FILE_NAME", &
                          description="The file to print current configurational info to.", &
                          usage="DATA_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_FILE_NAME", &
                          description="The file to print current cell length info to.", &
                          usage="CELL_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DISP_FILE_NAME", &
                          description="The file to print current maximum displacement info to.", &
                          usage="MAX_DISP_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BOX2_FILE_NAME", &
                          description="For GEMC, the name of the input file for the other box.", &
                          usage="BOX2_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRESSURE", &
                          description="The pressure for NpT simulations, in bar.", &
                          usage="PRESSURE {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="The temperature of the simulation, in Kelvin.", &
                          usage="TEMPERATURE {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="VIRIAL_TEMPS", &
         description="The temperatures you wish to compute the virial coefficient for.  Only used if ensemble=VIRIAL.", &
         usage="VIRIAL_TEMPS {real} {real} ... ", &
         n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISCRETE_STEP", &
                          description="The size of the discrete volume move step, in angstroms.", &
                          usage="DISCRETE_STEP {real}", &
                          default_r_val=1.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ETA", &
                          description="The free energy bias (in Kelvin) for swapping a molecule of each type into this box.", &
                          usage="ETA {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANDOMTOSKIP", &
                          description="Number of random numbers from the acceptance/rejection stream to skip", &
                          usage="RANDOMTOSKIP {integer}", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_avbmc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_move_prob_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_update_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_max_disp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_mc_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the AVBMC parameters for MC
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_avbmc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_avbmc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",380)

      CALL section_create(section, name="avbmc", &
                          description="Parameters for Aggregation Volume Bias Monte Carlo (AVBMC) "// &
                          "which explores cluster formation and destruction. "// &
                          "Chen and Siepmann, J. Phys. Chem. B 105, 11275-11282 (2001).", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="PBIAS", &
         description="The probability of swapping to an inner region in an AVBMC swap move for each molecule type.", &
         usage="PBIAS {real} {real} ... ", &
         n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVBMC_ATOM", &
                          description="The target atom for an AVBMC swap move for each molecule type.", &
                          usage="AVBMC_ATOM {integer} {integer} ... ", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVBMC_RMIN", &
                          description="The inner radius for an AVBMC swap move, in angstroms for every molecule type.", &
                          usage="AVBMC_RMIN {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AVBMC_RMAX", &
                          description="The outer radius for an AVBMC swap move, in angstroms, for every molecule type.", &
                          usage="AVBMC_RMAX {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_avbmc_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the probabilities for attempting each move
!>        type in Monte Carlo
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_move_prob_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_move_prob_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",436)

      CALL section_create(section, name="move_probabilities", &
                          description="Parameters for fraction of moves performed for each move type.", &
                          n_keywords=5, n_subsections=2, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="PMHMC", &
                          description="The probability of attempting a hybrid MC move.", &
                          usage="PMHMC {real}", &
                          type_of_var=real_t, default_r_val=0.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMTRANS", &
                          description="The probability of attempting a molecule translation.", &
                          usage="PMTRANS {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMAVBMC", &
                          description="The probability of attempting an AVBMC swap move.", &
                          usage="PMAVBMC {real}", &
                          default_r_val=0.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMTRAION", &
                          description="The probability of attempting a conformational change.", &
                          usage="PMTRAION {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMSWAP", &
                          description="The probability of attempting a swap move.", &
                          usage="PMSWAP {real}", &
                          type_of_var=real_t, default_r_val=0.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMVOLUME", &
                          description="The probability of attempting a volume move.", &
                          usage="PMVOLUME {real}", &
                          type_of_var=real_t, default_r_val=0.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_mol_prob_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_box_prob_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_move_prob_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the probabilities for attempting various moves
!>        on the various molecule types present in the system
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_mol_prob_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mol_prob_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",510)

      CALL section_create(section, name="mol_probabilities", &
                          description="Probabilities of attempting various moves types on "// &
                          "the various molecular types present in the simulation.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="PMAVBMC_MOL", &
                          description="The probability of attempting an AVBMC swap move on each molecule type.", &
                          usage="PMAVBMC_MOL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMSWAP_MOL", &
                          description="The probability of attempting a molecule swap of a given molecule type.", &
                          usage="PMSWAP_MOL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMROT_MOL", &
                          description="The probability of attempting a molecule rotation of a given molecule type.", &
                          usage="PMROT_MOL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMTRAION_MOL", &
                          description="The probability of attempting a conformational change of a given molecule type.", &
                          usage="PMTRAION_MOL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMTRANS_MOL", &
                          description="The probability of attempting a molecule translation of a given molecule type.", &
                          usage="PMTRANS_MOL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mol_prob_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the probabilities for attempting various moves
!>        on the box where the variable is present
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_box_prob_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_box_prob_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",570)

      CALL section_create(section, name="BOX_PROBABILITIES", &
                          description="Probabilities of attempting various moves types on "// &
                          "the box.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="PMHMC_BOX", &
                          description="The probability of attempting a HMC move on this box.", &
                          usage="PMHMC_BOX {real}", &
                          type_of_var=real_t, default_r_val=1.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PMVOL_BOX", &
                          description="The probability of attempting a volume move on this box (GEMC_NpT).", &
                          usage="PMVOL_BOX {real}", &
                          type_of_var=real_t, default_r_val=1.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_box_prob_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the frequency for updating maximum
!>        displacements for various moves
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_update_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_update_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",609)

      CALL section_create(section, name="MOVE_UPDATES", &
                          description="Frequency for updating move maximum displacements.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="IUPVOLUME", &
                          description="Every iupvolume steps update maximum volume displacement.", &
                          usage="IUPVOLUME {integer}", &
                          default_i_val=10000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IUPTRANS", &
                          description="Every iuptrans steps update maximum translation/rotation/"// &
                          "configurational changes.", &
                          usage="IUPTRANS {integer}", &
                          default_i_val=10000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_update_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the maximum displacements for various moves
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_max_disp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_max_disp_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",647)

      CALL section_create(section, name="max_displacements", &
                          description="The maximum displacements for all attempted moves.", &
                          n_keywords=1, n_subsections=2, repeats=.FALSE.)

      NULLIFY (subsection)

      CALL create_mol_disp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_box_disp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_max_disp_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the maximum displacements for all moves which
!>        require a value for each molecule type
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_mol_disp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mol_disp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",679)

      CALL section_create(section, name="mol_displacements", &
                          description="Maximum displacements for every move type that requires "// &
                          "a value for each molecular type in the simulation.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="RMBOND", &
                          description="Maximum bond length displacement, in angstroms, for each molecule type.", &
                          usage="RMBOND {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMANGLE", &
                          description="Maximum bond angle displacement, in degrees, for each molecule type.", &
                          usage="RMANGLE {real} {real} ...", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMDIHEDRAL", &
                          description="Maximum dihedral angle distplacement, in degrees, for each molecule type.", &
                          usage="RMDIHEDRAL {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMROT", &
                          description="Maximum rotational displacement, in degrees, for each molecule type.", &
                          usage="RMROT {real} {real} ... ", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMTRANS", &
                          description="Maximum translational displacement, in angstroms, for each molecule type.", &
                          usage="RMTRANS {real} {real} ...", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mol_disp_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the maximum displacements for any move that
!>        is done on each simulation box
!> \author matt
! **************************************************************************************************
   SUBROUTINE create_box_disp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_box_disp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",739)

      CALL section_create(section, name="BOX_DISPLACEMENTS", &
                          description="Maximum displacements for any move that is performed on each"// &
                          " simulation box.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="RMVOLUME", &
                          description="Maximum volume displacement, in angstrom**3.", &
                          usage="RMVOLUME {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_box_disp_section

! **************************************************************************************************
!> \brief creates the geometry optimization section
!> \param section the section to be created
!> \param label ...
!> \param description ...
!> \param just_optimizers ...
!> \author teo
! **************************************************************************************************
   RECURSIVE SUBROUTINE create_geoopt_section(section, label, description, just_optimizers)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: label, description
      LOGICAL, INTENT(IN)                                :: just_optimizers

      CHARACTER(len=*), PARAMETER :: routineN = 'create_geoopt_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",776)
      CALL section_create(section, name=label, description=description, &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword)
      IF (.NOT. just_optimizers) THEN
         CALL keyword_create(keyword, name="TYPE", &
                             description="Specify which kind of geometry optimization to perform", &
                             usage="TYPE (MINIMIZATION|TRANSITION_STATE)", &
                             enum_c_vals=s2a("MINIMIZATION", "TRANSITION_STATE"), &
                             enum_desc=s2a("Performs a geometry minimization.", &
                                           "Performs a transition state optimization."), &
                             enum_i_vals=(/default_minimization_method_id, default_ts_method_id/), &
                             default_i_val=default_minimization_method_id)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)
      END IF

      CALL keyword_create( &
         keyword, name="OPTIMIZER", &
         variants=(/"MINIMIZER"/), &
         citations=(/Byrd1995/), &
         description="Specify which method to use to perform a geometry optimization.", &
         usage="OPTIMIZER {BFGS|LBFGS|CG}", &
         enum_c_vals=s2a("BFGS", "LBFGS", "CG"), &
         enum_desc=s2a("Most efficient minimizer, but only for 'small' systems, "// &
                       "as it relies on diagonalization of a full Hessian matrix", &
                       "Limit memory variant of the above, can also be applied to large systems, not as well fine-tuned", &
                       "conjugate gradients, robust minimizer (depending on the line search) also OK for large systems"), &
         enum_i_vals=(/default_bfgs_method_id, default_lbfgs_method_id, default_cg_method_id/), &
         default_i_val=default_bfgs_method_id)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Specifies the maximum number of geometry optimization steps. "// &
                          "One step might imply several force evaluations for the CG and LBFGS optimizers.", &
                          usage="MAX_ITER {integer}", &
                          default_i_val=200)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DR", &
                          description="Convergence criterium for the maximum geometry change "// &
                          "between the current and the last optimizer iteration.", &
                          usage="MAX_DR {real}", &
                          default_r_val=0.0030_dp, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FORCE", &
                          description="Convergence criterium for the maximum force component of the current configuration.", &
                          usage="MAX_FORCE {real}", &
                          default_r_val=0.00045_dp, unit_str="hartree/bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMS_DR", &
                          description="Convergence criterium for the root mean square (RMS) geometry"// &
                          " change between the current and the last optimizer iteration.", &
                          usage="RMS_DR {real}", unit_str="bohr", &
                          default_r_val=0.0015_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMS_FORCE", &
                          description="Convergence criterium for the root mean square (RMS) force of the current configuration.", &
                          usage="RMS_FORCE {real}", unit_str="hartree/bohr", &
                          default_r_val=0.00030_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="step_start_val", &
                          description="The starting step value for the "//TRIM(label)//" module.", &
                          usage="step_start_val <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_lbfgs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cg_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_bfgs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      IF (.NOT. just_optimizers) THEN
         ! Transition states section
         CALL create_ts_section(subsection)
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)

         ! Create the PRINT subsection
         NULLIFY (subsection)
         CALL section_create(subsection, name="PRINT", &
                             description="Controls the printing properties during a geometry optimization run", &
                             n_keywords=0, n_subsections=1, repeats=.TRUE.)
         NULLIFY (print_key)
         CALL cp_print_key_section_create( &
            print_key, "program_run_info", &
            description="Controls the printing of basic information during the Geometry Optimization", &
            print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
         CALL section_add_subsection(subsection, print_key)
         CALL section_release(print_key)
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)
      END IF

   END SUBROUTINE create_geoopt_section

! **************************************************************************************************
!> \brief creates the section for the shell-core optimization
!> \param section the section to be created
!> \author Caino
! **************************************************************************************************
   SUBROUTINE create_shellcore_opt_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shellcore_opt_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key, subsection

      CALL create_geoopt_section( &
         section, label="SHELL_OPT", &
         description="This section sets the environment for the optimization of the shell-core distances"// &
         " that might turn to be necessary along a MD run using a shell-model potential. "// &
         " The optimization procedure is activated when at least one of the shell-core "// &
         "pairs becomes too elongated,  i.e. when the assumption of point dipole is not longer valid.", &
         just_optimizers=.TRUE.)

      NULLIFY (print_key, subsection)

      ! Create the PRINT subsection
      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="Controls the printing properties during a shell-core optimization procedure", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic information during the Optimization", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_shellcore_opt_section

! **************************************************************************************************
!> \brief creates the section for the cell optimization
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE create_cell_opt_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cell_opt_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      CALL create_geoopt_section(section, label="CELL_OPT", &
                                 description="This section sets the environment for the optimization of the simulation cell."// &
                                 " Two possible schemes are available: (1) Zero temperature optimization; "// &
                                 " (2) Finite temperature optimization. ", just_optimizers=.TRUE.)

      NULLIFY (keyword, print_key, subsection)
      CALL keyword_create( &
         keyword, name="TYPE", &
         description="Specify which kind of method to use for the optimization of the simulation cell", &
         usage="TYPE (GEO_OPT|MD|DIRECT_CELL_OPT)", &
         enum_c_vals=s2a("GEO_OPT", "MD", "DIRECT_CELL_OPT"), &
         enum_desc=s2a( &
         "Performs a geometry optimization (the GEO_OPT section must be defined) between cell optimization steps. "// &
         " The stress tensor is computed at the optimized geometry.", &
         "Performs a molecular dynamics run (the MD section needs must defined) for computing the stress tensor "// &
         " used for the cell optimization.", &
         "Performs a geometry and cell optimization at the same time."// &
         " The stress tensor is computed at every step"), &
         enum_i_vals=(/default_cell_geo_opt_id, default_cell_md_id, default_cell_direct_id/), &
         default_i_val=default_cell_direct_id)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EXTERNAL_PRESSURE", &
         description="Specifies the external pressure (1 value or the full 9 components of the pressure tensor) "// &
         "applied during the cell optimization.", &
         usage="EXTERNAL_PRESSURE {REAL} .. {REAL}", unit_str="bar", &
         default_r_vals=(/cp_unit_to_cp2k(100.0_dp, "bar"), 0.0_dp, 0.0_dp, &
                          0.0_dp, cp_unit_to_cp2k(100.0_dp, "bar"), 0.0_dp, &
                          0.0_dp, 0.0_dp, cp_unit_to_cp2k(100.0_dp, "bar")/), n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="KEEP_ANGLES", &
         description="Keep angles between the cell vectors constant, but allow the lenghts of the"// &
         " cell vectors to change independently."// &
         " Albeit general, this is most useful for triclinic cells, to enforce higher symmetry, see KEEP_SYMMETRY.", &
         usage="KEEP_ANGLES TRUE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KEEP_SYMMETRY", &
                          description="Keep the requested initial cell symmetry (e.g. during a cell optimisation). "// &
                          "The initial symmetry must be specified in the &CELL section.", &
                          usage="KEEP_SYMMETRY yes", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRESSURE_TOLERANCE", &
                          description="Specifies the Pressure tolerance (compared to the external pressure) to achieve "// &
                          "during the cell optimization.", &
                          usage="PRESSURE_TOLERANCE {REAL}", unit_str="bar", &
                          default_r_val=cp_unit_to_cp2k(100.0_dp, "bar"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Create the PRINT subsection
      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="Controls the printing properties during a geometry optimization run", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic information during the Geometry Optimization", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL cp_print_key_section_create( &
         print_key, "cell", &
         description="Controls the printing of the cell eveytime a calculation using a new cell is started.", &
         print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__", &
         unit_str="angstrom")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_cell_opt_section

! **************************************************************************************************
!> \brief creates the section for tuning transition states search
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! **************************************************************************************************
   SUBROUTINE create_ts_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ts_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsection2, &
                                                            subsection3

! Create the Transition State subsection

      NULLIFY (section, keyword, subsection, subsection2)
      CALL section_create(section, name="TRANSITION_STATE", &
                          description="Specifies parameters to perform a transition state search", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="METHOD", &
                          description="Specify which kind of method to use for locating transition states", &
                          citations=(/Henkelman1999/), &
                          usage="TYPE (DIMER)", &
                          enum_c_vals=s2a("DIMER"), &
                          enum_desc=s2a("Uses the dimer method to optimize transition states."), &
                          enum_i_vals=(/default_dimer_method_id/), &
                          default_i_val=default_dimer_method_id)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="DIMER", &
                          description="Specifies parameters for Dimer Method", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="DR", &
                          description="This keyword sets the value for the DR parameter.", &
                          usage="DR {real}", unit_str='angstrom', &
                          default_r_val=cp_unit_to_cp2k(0.01_dp, "angstrom"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERPOLATE_GRADIENT", &
                          description="This keyword controls the interpolation of the gradient whenever possible"// &
                          " during the optimization of the Dimer. The use of this keywords saves 1 evaluation "// &
                          " of energy/forces.", usage="INTERPOLATE_GRADIENT {logical}", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANGLE_TOLERANCE", &
                          description="This keyword sets the value of the tolerance angle for the line search "// &
                          " performed to optimize the orientation of the dimer.", &
                          usage="ANGLE_TOL {real}", unit_str='rad', &
                          default_r_val=cp_unit_to_cp2k(5.0_dp, "deg"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL create_geoopt_section( &
         subsection2, label="ROT_OPT", &
         description="This section sets the environment for the optimization of the rotation of the Dimer.", &
         just_optimizers=.TRUE.)
      NULLIFY (subsection3)
      CALL section_create(subsection3, name="PRINT", &
                          description="Controls the printing properties during the dimer rotation optimization run", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)
      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of basic information during the Geometry Optimization", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection3, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ROTATIONAL_INFO", &
                                       description="Controls the printing basic info during the cleaning of the "// &
                                       "rotational degrees of freedom.", print_level=low_print_level, &
                                       add_last=add_last_numeric, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="COORDINATES", &
                          description="Prints atomic coordinates after rotation", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection3, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(subsection2, subsection3)
      CALL section_release(subsection3)
      CALL section_add_subsection(subsection, subsection2)
      CALL section_release(subsection2)

      CALL section_create(subsection2, name="DIMER_VECTOR", &
                          description="Specifies the initial dimer vector (used frequently to restart DIMER calculations)."// &
                          " If not provided the starting orientation of the dimer is chosen randomly.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify on each line the components of the dimer vector.", repeats=.TRUE., &
                          usage="{Real} {Real} {Real}", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection2, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsection2)
      CALL section_release(subsection2)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_ts_section

! **************************************************************************************************
!> \brief creates the BFGS section
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! **************************************************************************************************
   SUBROUTINE create_bfgs_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_bfgs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

! create the BFGS subsection

      NULLIFY (section, keyword, print_key)
      CALL section_create(section, name="BFGS", &
                          description="Provides parameters to tune the BFGS optimization", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="TRUST_RADIUS", &
                          description="Trust radius used in BFGS. Previously set to 0.1. "// &
                          "Large values can lead to instabilities", &
                          usage="TRUST_RADIUS {real}", unit_str='angstrom', &
                          default_r_val=cp_unit_to_cp2k(0.25_dp, "angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_MODEL_HESSIAN", &
                          description="Uses a model Hessian as initial guess instead of a unit matrix."// &
                          " Should lead in general to improved convergence might be switched off for exotic cases", &
                          usage="USE_MODEL_HESSIAN", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_RAT_FUN_OPT", &
                          description="Includes a rational function optimization to determine the step."// &
                          " Previously default but did not improve convergence in many cases", &
                          usage="USE_RAT_FUN_OPT", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_HESSIAN", &
                          description="Controls the reading of the initial Hessian from file.", &
                          usage="RESTART_HESSIAN", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_FILE_NAME", &
                          description="Specifies the name of the file used to read the initial Hessian.", &
                          usage="RESTART_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the printing of Hessian Restart file", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="BFGS", &
                                       common_iter_levels=2)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_bfgs_section

! **************************************************************************************************
!> \brief creates the CG section
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! **************************************************************************************************
   SUBROUTINE create_cg_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cg_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection, subsubsection

! create the CG subsection

      NULLIFY (section, subsection, subsubsection, keyword)
      CALL section_create(section, name="CG", &
                          description="Provides parameters to tune the conjugate gradient optimization", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="MAX_STEEP_STEPS", &
                          description="Maximum number of steepest descent steps before starting the"// &
                          " conjugate gradients optimization.", &
                          usage="MAX_STEEP_STEPS {integer}", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_LIMIT", &
                          description="Cosine of the angle between two consecutive searching directions."// &
                          " If the angle during a CG optimization is less than the one corresponding to "// &
                          " to the RESTART_LIMIT the CG is reset and one step of steepest descent is "// &
                          " performed.", &
                          usage="RESTART_LIMIT {real}", &
                          default_r_val=0.9_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FLETCHER_REEVES", &
                          description="Uses FLETCHER-REEVES instead of POLAK-RIBIERE when using Conjugate Gradients", &
                          usage="FLETCHER-REEVES", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Line Search section
      CALL section_create(subsection, name="LINE_SEARCH", &
                          description="Provides parameters to tune the line search during the conjugate gradient optimization", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="TYPE", &
                          description="1D line search algorithm to be used with the CG optimizer,"// &
                          " in increasing order of robustness and cost. ", &
                          usage="TYPE GOLD", &
                          default_i_val=ls_gold, &
                          enum_c_vals=s2a("NONE", "2PNT", "3PNT", "GOLD", "FIT"), &
                          enum_desc=s2a("take fixed lenght steps", &
                                        "extrapolate based on 2 points", &
                                        "extrapolate based on on 3 points", &
                                        "perform 1D golden section search of the minimum (very expensive)", &
                                        "perform 1D fit of a parabola on several evaluation of energy "// &
                                        "(very expensive and more robust vs numerical noise)"), &
                          enum_i_vals=(/ls_none, ls_2pnt, ls_3pnt, ls_gold, ls_fit/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      ! 2PNT
      NULLIFY (subsubsection)
      CALL section_create(subsubsection, name="2PNT", &
                          description="Provides parameters to tune the line search for the two point based line search.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="MAX_ALLOWED_STEP", &
                          description="Max allowed value for the line search step.", &
                          usage="MAX_ALLOWED_STEP {real}", unit_str="internal_cp2k", &
                          default_r_val=0.25_dp)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="LINMIN_GRAD_ONLY", &
         description="Use only the gradient, not the energy for line minimizations (e.g. in conjugate gradients).", &
         usage="LINMIN_GRAD_ONLY T", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      ! GOLD
      NULLIFY (subsubsection)
      CALL section_create(subsubsection, name="GOLD", &
                          description="Provides parameters to tune the line search for the gold search.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="INITIAL_STEP", &
                          description="Initial step size used, e.g. for bracketing or minimizers. "// &
                          "Might need to be reduced for systems with close contacts", &
                          usage="INITIAL_STEP {real}", unit_str="internal_cp2k", &
                          default_r_val=0.2_dp)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BRACK_LIMIT", &
                          description="Limit in 1D bracketing during line search in Conjugate Gradients Optimization.", &
                          usage="BRACK_LIMIT {real}", unit_str="internal_cp2k", &
                          default_r_val=100.0_dp)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BRENT_TOL", &
                          description="Tolerance requested during Brent line search in Conjugate Gradients Optimization.", &
                          usage="BRENT_TOL {real}", unit_str="internal_cp2k", &
                          default_r_val=0.01_dp)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BRENT_MAX_ITER", &
                          description="Maximum number of iterations in brent algorithm "// &
                          "(used for the line search in Conjugated Gradients Optimization)", &
                          usage="BRENT_MAX_ITER {integer}", &
                          default_i_val=100)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_cg_section

! **************************************************************************************************
!> \brief creates the LBFGS section
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! **************************************************************************************************
   SUBROUTINE create_lbfgs_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_lbfgs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

! create the LBFGS section

      NULLIFY (section, keyword)
      CALL section_create(section, name="LBFGS", &
                          description="Provides parameters to tune the limited memory BFGS (LBFGS) optimization", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE., &
                          citations=(/Byrd1995/))

      CALL keyword_create(keyword, name="MAX_H_RANK", &
                          description="Maximum rank (and consequently size) of the "// &
                          "approximate Hessian matrix used by the LBFGS optimizer. "// &
                          "Larger values (e.g. 30) will accelerate the convergence behaviour "// &
                          "at the cost of a larger memory consumption.", &
                          usage="MAX_H_RANK {integer}", &
                          default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_F_PER_ITER", &
                          description="Maximum number of force evaluations per iteration"// &
                          "(used for the line search)", &
                          usage="MAX_F_PER_ITER {integer}", &
                          default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WANTED_PROJ_GRADIENT", &
                          description="Convergence criterium (overrides the general ones):"// &
                          "Requested norm threshold of the gradient multiplied "// &
                          "by the approximate Hessian.", &
                          usage="WANTED_PROJ_GRADIENT {real}", unit_str="internal_cp2k", &
                          default_r_val=1.0E-16_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WANTED_REL_F_ERROR", &
                          description="Convergence criterium (overrides the general ones):"// &
                          "Requested relative error on the objective function"// &
                          "of the optimizer (the energy)", &
                          usage="WANTED_REL_F_ERROR {real}", unit_str="internal_cp2k", &
                          default_r_val=1.0E-16_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="TRUST_RADIUS", &
         description="Trust radius used in LBFGS. Not completely in dept tested. Negativ values means no trust radius is used.", &
         usage="TRUST_RADIUS {real}", unit_str='angstrom', &
         default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_lbfgs_section

! **************************************************************************************************
!> \brief creates the flexible_partitioning section
!> \param section the section to be created
!> \author Joost VandeVondele [04.2006]
! **************************************************************************************************
   SUBROUTINE create_fp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_fp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",1416)
      CALL section_create(section, name="FLEXIBLE_PARTITIONING", &
                          description="This section sets up flexible_partitioning", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="CENTRAL_ATOM", &
                          description="Specifies the central atom.", &
                          usage="CENTRAL_ATOM {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INNER_ATOMS", &
                          description="Specifies the list of atoms that should remain close to the central atom.", &
                          usage="INNER_ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OUTER_ATOMS", &
                          description="Specifies the list of atoms that should remain far from the central atom.", &
                          usage="OUTER_ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INNER_RADIUS", &
                          description="radius of the inner wall", &
                          usage="INNER_RADIUS {real} ", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OUTER_RADIUS", &
                          description="radius of the outer wall", &
                          usage="OUTER_RADIUS {real} ", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="Sets the force constant of the repulsive harmonic potential", &
                          usage="STRENGTH 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BIAS", &
                          description="If a bias potential counter-acting the weight term should be applied (recommended).", &
                          usage="BIAS F", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Sets the temperature parameter that is used in the baising potential."// &
                          "It is recommended to use the actual simulation temperature", &
                          usage="TEMPERATURE 300", default_r_val=300.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTH_WIDTH", &
                          description="Sets the width of the smooth counting function.", &
                          usage="SMOOTH_WIDTH 0.2", default_r_val=0.02_dp, unit_str='angstrom')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "WEIGHTS", &
                                       description="Controls the printing of FP info during flexible partitioning simulations.", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="FLEXIBLE_PARTIONING")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CONTROL", &
                                       description="Controls the printing of FP info at startup", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_fp_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the driver section
!> \author mceriotti
! **************************************************************************************************
   SUBROUTINE create_driver_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_driver_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",1512)
      CALL section_create(section, name="DRIVER", &
                          description="This section defines the parameters needed to run in i-PI driver mode.", &
                          citations=(/Ceriotti2014, Kapil2016/), &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="unix", &
                          description="Use a UNIX socket rather than an INET socket.", &
                          usage="unix LOGICAL", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="port", &
                          description="Port number for the i-PI server.", &
                          usage="port <INTEGER>", &
                          default_i_val=12345)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="host", &
                          description="Host name for the i-PI server.", &
                          usage="host <HOSTNAME>", &
                          default_c_val="localhost")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SLEEP_TIME", &
                          description="Sleeping time while waiting for for driver commands [s].", &
                          usage="SLEEP_TIME 0.1", &
                          default_r_val=0.01_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_driver_section

! **************************************************************************************************
!> \brief creates the section for a path integral run
!> \param section will contain the pint section
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_pint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsubsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",1563)
      CALL section_create(section, name="PINT", &
                          description="The section that controls a path integral run", &
                          n_keywords=13, n_subsections=8, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="p", &
                          description="Specify number beads to use", repeats=.FALSE., &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="proc_per_replica", &
                          description="Specify number of processors to use for each replica", &
                          repeats=.FALSE., default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="num_steps", &
                          description="Number of steps (if MAX_STEP is not explicitly given"// &
                          " the program will perform this number of steps)", repeats=.FALSE., &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="MAX_STEP", &
                          description="Maximum step number (the program will stop if"// &
                          " ITERATION >= MAX_STEP even if NUM_STEPS has not been reached)", &
                          repeats=.FALSE., default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="iteration", &
                          description="Specify the iteration number from which it should be "// &
                          "counted", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="Temp", &
                          description="The temperature you want to simulate", &
                          default_r_val=cp_unit_to_cp2k(300._dp, "K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="T_tol", variants=(/"temp_to"/), &
                          description="threshold for the oscillations of the temperature "// &
                          "excedeed which the temperature is rescaled. 0 means no rescaling.", &
                          default_r_val=0._dp, unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="dt", &
                          description="timestep (might be subdivised in nrespa subtimesteps", &
                          repeats=.FALSE., &
                          default_r_val=cp_unit_to_cp2k(1.0_dp, "fs"), &
                          usage="dt 1.0", unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="HARM_INT", &
                          description="integrator scheme for integrating the harmonic bead springs.", &
                          usage="HARM_INT (NUMERIC|EXACT)", &
                          default_i_val=integrate_numeric, &
                          enum_c_vals=s2a("NUMERIC", "EXACT"), &
                          enum_i_vals=(/integrate_numeric, integrate_exact/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="nrespa", &
                          description="number of respa steps for the bead for each md step", &
                          repeats=.FALSE., default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="transformation", &
                          description="Specifies the coordinate transformation to use", &
                          usage="TRANSFORMATION (NORMAL|STAGE)", &
                          default_i_val=transformation_normal, &
                          enum_c_vals=s2a("NORMAL", "STAGE"), &
                          enum_i_vals=(/transformation_normal, transformation_stage/))

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="propagator", &
                          description="Specifies the real time propagator to use", &
                          usage="PROPAGATOR (PIMD|RPMD)", &
                          default_i_val=propagator_pimd, &
                          enum_c_vals=s2a("PIMD", "RPMD"), &
                          enum_i_vals=(/propagator_pimd, propagator_rpmd/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="FIX_CENTROID_POS", &
                          description="Propagate all DOF but the centroid - "// &
                          "useful for equilibration of the non-centroid modes "// &
                          "(activated only if TRANSFORMATION==NORMAL)", &
                          repeats=.FALSE., default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection, subsubsection)
      CALL section_create(subsection, name="NORMALMODE", &
                          description="Controls the normal mode transformation", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="Q_CENTROID", &
                          description="Value of the thermostat mass of centroid degree of freedom", &
                          repeats=.FALSE., default_r_val=-1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="Q_BEAD", &
                          description="Value of the thermostat mass of non-centroid degrees of freedom", &
                          repeats=.FALSE., default_r_val=-1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="MODEFACTOR", &
                          description="mass scale factor for non-centroid degrees of freedom", &
                          repeats=.FALSE., default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="staging", &
                          description="The section that controls the staging transformation", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="j", &
                          description="Value of the j parameter for the staging transformation", &
                          repeats=.FALSE., default_i_val=2)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="Q_END", &
                          description="Value of the nose-hoover mass for the endbead (Q_end)", &
                          repeats=.FALSE., default_i_val=2)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BEADS", &
                          description="Sets positions and velocities of the beads", &
                          n_keywords=0, n_subsections=2, &
                          repeats=.FALSE.)
      CALL create_coord_section(subsubsection, "BEADS")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL create_velocity_section(subsubsection, "BEADS")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="NOSE", &
                          description="Controls the Nose-Hoover thermostats", &
                          n_keywords=1, n_subsections=2, &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="nnos", &
                          description="length of nose-hoover chain. 0 means no thermostat", &
                          repeats=.FALSE., default_i_val=2)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL create_coord_section(subsubsection, "NOSE")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL create_velocity_section(subsubsection, "NOSE")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_gle_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PILE", &
                          description="Controls the PI Langevin Equation thermostat."// &
                          " Needs the exact harmonic integrator", &
                          citations=(/Ceriotti2010/), &
                          n_keywords=2, n_subsections=1, &
                          repeats=.FALSE.)
      CALL create_rng_section(subsubsection)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL keyword_create(keyword, name="TAU", &
                          description="Time constant for centroid motion. "// &
                          "If zero or negative the centroid is not thermostated.", &
                          usage="TAU {real}", type_of_var=real_t, &
                          n_var=1, default_r_val=10000.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Scaling of friction to mode coupling", &
                          usage="LAMBDA {real}", type_of_var=real_t, &
                          unit_str="fs", n_var=1, default_r_val=0.01_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PIGLET", &
                          description="Controls the PI Generalized Langevin Equation thermostat."// &
                          " Needs the exact harmonic integrator", &
                          citations=(/Ceriotti2012/), &
                          n_keywords=3, n_subsections=2, &
                          repeats=.FALSE.)
      CALL create_rng_section(subsubsection)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL section_create(subsubsection, name="EXTRA_DOF", &
                          description="Additional degrees of freedom to ensure Markovian Dynamics.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Restart values for additional degrees of freedom" &
                          //" (only for restarts, do not set explicitly)", &
                          repeats=.FALSE., &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)
      CALL keyword_create(keyword, name="NEXTRA_DOF", &
                          description="Number of extra degrees of freedom to ensure markovian dynamics", &
                          repeats=.FALSE., default_i_val=8)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="MATRICES_FILE_NAME", &
                          description="Filename containig the raw matrices from "// &
                          "https://epfl-cosmo.github.io/gle4md/index.html?page=matrix", &
                          repeats=.FALSE., default_lc_val="PIGLET.MAT")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SMATRIX_INIT", &
                          description="Select algorithm to initialize piglet S-matrices", &
                          usage="SMATRIX_INIT (CHOLESKY|DIAGONAL)", &
                          default_i_val=matrix_init_cholesky, &
                          enum_c_vals=s2a("CHOLESKY", "DIAGONAL"), &
                          enum_i_vals=(/matrix_init_cholesky, matrix_init_diagonal/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="INIT", &
                          description="Controls the initialization if the beads are not present", &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, name="LEVY_POS_SAMPLE", &
                          description="Sample bead positions assuming free particle "// &
                          "behavior (performs a Levy random walk of length P around "// &
                          "the classical position of each atom at the physical "// &
                          "temperature defined in PINT%TEMP)", &
                          repeats=.FALSE., default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="LEVY_CORRELATED", &
                          description="Use the same Levy path for all atoms, though "// &
                          "with mass-dependent variances (might help at very low T)", &
                          repeats=.FALSE., default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="LEVY_TEMP_FACTOR", &
                          description="Multiplicative correction factor for the "// &
                          "temperature at which the Levy walk is performed "// &
                          "(correction is due to the interactions that modify "// &
                          "the spread of a free particle)", &
                          repeats=.FALSE., default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="LEVY_SEED", &
                          description="Initial seed for the (pseudo)random number "// &
                          "generator that controls Levy walk for bead positions.", &
                          usage="LEVY_SEED <INTEGER>", default_i_val=1234, &
                          repeats=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="THERMOSTAT_SEED", &
                          description="Initial seed for the (pseudo)random number "// &
                          "generator that controls the PILE and PIGLET thermostats.", &
                          usage="THERMOSTAT_SEED <INTEGER>", default_i_val=12345, &
                          repeats=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RANDOMIZE_POS", &
                          description="add gaussian noise to the positions of the beads", &
                          repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CENTROID_SPEED", &
                          description="adds random velocity component to the centroid modes "// &
                          "(useful to correct for the averaging out of the speed of various beads)", &
                          repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VELOCITY_QUENCH", &
                          description="set the initial velocities to zero", &
                          repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="VELOCITY_SCALE", &
                          description="scale initial velocities to the temperature given in MOTION%PINT%TEMP", &
                          repeats=.FALSE., default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_helium_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PRINT", &
                          description="Controls the path integral-specific output", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the output of the path integral energies", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CENTROID_POS", &
                                       description="Controls the output of the centroid's position", &
                                       unit_str="angstrom", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL add_format_keyword(keyword, print_key, pos=.TRUE., &
                              description="Output file format for the positions of centroid")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CENTROID_VEL", &
                                       description="Controls the output of the centroid's velocity", &
                                       unit_str="bohr*au_t^-1", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Output file format for the velocity of centroid")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CENTROID_GYR", &
                                       description="Controls the output of the centroid's radii of gyration", &
                                       unit_str="angstrom", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COM", &
                                       description="Controls the output of the center of mass", &
                                       print_level=high_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL keyword_create(keyword, name="IMAGINARY_TIME_STRIDE", &
                          description="Prints only every nth bead trajectory", &
                          repeats=.FALSE., default_i_val=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_pint_section

   ! ***************************************************************************
   !> \brief  Create the input section for superfluid helium solvent.
   !> \author Lukasz Walewski
   ! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_helium_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_helium_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k_motion.F",1940)

      CALL section_create(section, name="HELIUM", &
                          description="The section that controls optional helium solvent"// &
                          " environment (highly experimental, not for general use yet)", &
                          n_keywords=28, n_subsections=9, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Whether or not to actually use this section", &
                          usage="silent", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HELIUM_ONLY", &
                          description="Simulate helium solvent only, "// &
                          "disregard solute entirely", &
                          repeats=.FALSE., default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERACTION_POT_SCAN", &
                          description="Scan solute-helium interaction potential, "// &
                          "cubefile parameters set in subsection RHO", &
                          repeats=.FALSE., default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_ENV", &
                          description="Number of independent helium environments"// &
                          " (only for restarts, do not set explicitly)", &
                          repeats=.FALSE., default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POTENTIAL_FILE_NAME", &
                          description="Name of the Helium interaction potential file", &
                          repeats=.FALSE., default_lc_val="HELIUM.POT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GET_FORCES", &
                          description="Get average MC forces or last MC forces to propagate MD", &
                          usage="GET_FORCES (AVERAGE|LAST)", &
                          default_i_val=helium_forces_average, &
                          enum_c_vals=s2a("AVERAGE", "LAST"), &
                          enum_i_vals=(/helium_forces_average, helium_forces_last/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SOLUTE_INTERACTION", &
                          description="Interaction potential between helium and the solute", &
                          usage="SOLUTE_INTERACTION (NONE | MWATER)", &
                          default_i_val=helium_solute_intpot_none, &
                          enum_c_vals=s2a("NONE", "MWATER"), &
                          enum_i_vals=(/ &
                          helium_solute_intpot_none, &
                          helium_solute_intpot_mwater/), &
                          enum_desc=s2a( &
                          "No interaction with solute", &
                          "Test interaction with wrong Water"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NATOMS", &
                          description="Number of helium atoms", &
                          repeats=.FALSE., default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NBEADS", &
                          description="Number of helium path integral beads", &
                          repeats=.FALSE., default_i_val=25)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_INNER", &
                          variants=s2a("INOROT"), &
                          description="Number of MC iterations at the same time slice(s) "// &
                          "(number of inner MC loop iterations)", &
                          repeats=.FALSE., default_i_val=6600)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_OUTER", &
                          variants=s2a("IROT"), &
                          description="how often to reselect the time slice(s) to work on "// &
                          "(number of outer MC loop iterations)", &
                          repeats=.FALSE., default_i_val=300)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BISECTION", &
                          description="how many time slices to change at once (+1). "// &
                          "Must be a power of 2 currently", &
                          repeats=.FALSE., default_i_val=8)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_PERM_CYCLE", &
                          description="how large cyclic permutations to try", &
                          repeats=.FALSE., default_i_val=6)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SAMPLING_METHOD", &
                          description="Choose between Ceperley or other not implemented algorithms", &
                          usage="SAMPLING_METHOD (CEPERLEY)", &
                          default_i_val=helium_sampling_ceperley, &
                          enum_c_vals=s2a("CEPERLEY"), &
                          enum_i_vals=(/helium_sampling_ceperley/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL section_create(subsection, name="M-SAMPLING", &
                          description="Permutation cycle length sampling settings", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="DISTRIBUTION-TYPE", &
                          description="Distribution from which the cycle length m is sampled", &
                          usage="DISTRIBUTION-TYPE (SINGLEV|UNIFORM|LINEAR|QUADRATIC|EXPONENTIAL|GAUSSIAN)", &
                          default_i_val=helium_mdist_uniform, &
                          enum_c_vals=s2a( &
                          "SINGLEV", &
                          "UNIFORM", &
                          "LINEAR", &
                          "QUADRATIC", &
                          "EXPONENTIAL", &
                          "GAUSSIAN"), &
                          enum_i_vals=(/ &
                          helium_mdist_singlev, &
                          helium_mdist_uniform, &
                          helium_mdist_linear, &
                          helium_mdist_quadratic, &
                          helium_mdist_exponential, &
                          helium_mdist_gaussian/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="M-VALUE", &
                          description="Value of m treated in a special way "// &
                          "(specific behavior depends on the distribution type chosen)", &
                          repeats=.FALSE., &
                          default_i_val=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="M-RATIO", &
                          description="Probability ratio betw M-VALUE and other cycle lengths", &
                          repeats=.FALSE., &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="PERIODIC", &
                          description="Use periodic boundary conditions for helium", &
                          repeats=.FALSE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_SIZE", &
                          description="PBC unit cell size (NOTE 1: density, number of atoms"// &
                          " and volume are interdependent - give only two of them; "// &
                          "NOTE 2: for small cell sizes specify NATOMS instead)", &
                          repeats=.FALSE., type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_SHAPE", &
                          description="PBC unit cell shape for helium", &
                          usage="CELL_SHAPE (CUBE|OCTAHEDRON)", &
                          default_i_val=helium_cell_shape_cube, &
                          enum_c_vals=s2a("CUBE", "OCTAHEDRON"), &
                          enum_i_vals=(/helium_cell_shape_cube, helium_cell_shape_octahedron/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DROPLET_RADIUS", &
                          description="Reject a move if any of the new positions does not lie within"// &
                          "  this range from the center of gravity", &
                          repeats=.FALSE., type_of_var=real_t, default_r_val=HUGE(1.0_dp), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DENSITY", &
                          description="trial density of helium for determining the helium "// &
                          "box size", &
                          repeats=.FALSE., &
                          default_r_val=cp_unit_to_cp2k(0.02186_dp, "angstrom^-3"), &
                          unit_str="angstrom^-3")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRESAMPLE", &
                          description="Presample He coordinates before first PIMD step", &
                          repeats=.FALSE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DROP_UNUSED_ENVS", &
                          description="Drop He environments if N_restart > N_runtime "// &
                          "(Warning: this will cause data loss in the restart file!)", &
                          repeats=.FALSE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="RDF", &
                          description="Radial distribution settings", &
                          n_keywords=4, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Whether or not to actually calculate this property", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAXR", &
                          description="Maximum RDF range, defaults to unit cell size", &
                          repeats=.FALSE., type_of_var=real_t, &
                          unit_str="angstrom")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NBIN", &
                          description="Number of bins", &
                          repeats=.FALSE., &
                          default_i_val=250)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CENTERS_FILE_NAME", &
                          description="Calculate RDFs wrt centers read from a file", &
                          repeats=.FALSE., default_lc_val="centers.xyz")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL section_create(subsection, name="RHO", &
                          description="Spatial distribution settings", &
                          n_keywords=10, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Whether or not to actually calculate densities "// &
                          "(requires significant amount of memory, depending on the value of NBIN)", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="NBIN", &
                          description="Number of grid points in each direction for density binning", &
                          repeats=.FALSE., &
                          default_i_val=100)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="MIN_CYCLE_LENGTHS_WDG", &
                          description="Density of winding paths "// &
                          "not shorter than the given length", &
                          repeats=.FALSE., usage="<INT> <INT> .. <INT>", &
                          type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="MIN_CYCLE_LENGTHS_NON", &
                          description="Density of non-winding paths "// &
                          "not shorter than the given length", &
                          repeats=.FALSE., usage="<INT> <INT> .. <INT>", &
                          type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="MIN_CYCLE_LENGTHS_ALL", &
                          description="Density of all paths "// &
                          "not shorter than the given length", &
                          repeats=.FALSE., usage="<INT> <INT> .. <INT>", &
                          type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="ATOM_NUMBER", &
                          description="Atom number density", &
                          repeats=.FALSE., &
                          type_of_var=logical_t, &
                          default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="PROJECTED_AREA_2", &
                          description="Projected area squared density, A*A(r)", &
                          repeats=.FALSE., &
                          type_of_var=logical_t, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="WINDING_NUMBER_2", &
                          description="Winding number squared density, W*W(r)", &
                          repeats=.FALSE., &
                          type_of_var=logical_t, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="WINDING_CYCLE_2", &
                          description="Winding number squared density, W^2(r)", &
                          repeats=.FALSE., &
                          type_of_var=logical_t, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      !
      CALL keyword_create(keyword, name="MOMENT_OF_INERTIA", &
                          description="Moment of inertia density", &
                          repeats=.FALSE., &
                          type_of_var=logical_t, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
      ! end of subsection RHO

      CALL create_coord_section(subsection, "HELIUM")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PERM", &
                          description="Permutation state used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify particle index permutation for every "// &
                          "helium atom", repeats=.TRUE., usage="<INT> <INT> .. <INT>", &
                          type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="AVERAGES", &
                          description="Average properties (used for restarts)", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="PROJECTED_AREA", &
                          description="Projected area vector for all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PROJECTED_AREA_2", &
                          description="Projected area vector squared for all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="WINDING_NUMBER_2", &
                          description="Winding number vector squared for all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="MOMENT_OF_INERTIA", &
                          description="Moment of inertia vector for all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RDF", &
                          description="Radial distributions averaged over all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RHO", &
                          description="Spatial distributions averaged over all environments", &
                          repeats=.TRUE., usage="<REAL> <REAL> .. <REAL>", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="IWEIGHT", &
                          description="Weight for the restarted quantities "// &
                          "(number of MC steps used to calculate the accumulated averages)", &
                          repeats=.FALSE., &
                          default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="FORCE", &
                          description="Forces exerted by the helium on the solute system"// &
                          " (used for restarts)", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Number of real values should be 3 * "// &
                          "<num_solute_atoms> * <num_solute_beads>", repeats=.TRUE., &
                          usage="<REAL> <REAL> .. <REAL>", type_of_var=real_t, &
                          n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="RNG_STATE", &
                          description="Random number generator state for all processors", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Three real arrays of DIMENSION(3,2) times two RNG "// &
                          "streams - 36 real values per processor", &
                          repeats=.TRUE., usage="automatically filled, do not edit by hand", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PRINT", &
                          description="The section that controls the output of the helium code", &
                          n_keywords=15, n_subsections=0, repeats=.FALSE.)

      ! *************************************************************************
      !> Printkeys for properites output
      ! *************************************************************************
      NULLIFY (print_key)

      ! Properties printed at SILENT print level
      !

      ! Properties printed at LOW print level
      !
      CALL cp_print_key_section_create(print_key, "ENERGY", &
                                       description="Controls the output of helium energies"// &
                                       " (averaged over MC step)", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "PROJECTED_AREA_2_AVG", &
                                       description="Controls the output of the average projected area squared vector", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "WINDING_NUMBER_2_AVG", &
                                       description="Controls the output of the average winding number vector squared", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "MOMENT_OF_INERTIA_AVG", &
                                       description="Controls the output of the average moment of inertia vector", &
                                       print_level=low_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      ! Properties printed at MEDIUM print level
      !
      CALL cp_print_key_section_create(print_key, "RDF", &
                                       description="Controls the output of helium radial distribution functions", &
                                       print_level=medium_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RHO", &
                                       description="Controls the output of the helium density "// &
                                       "(Gaussian cube file format)", &
                                       each_iter_names=s2a("MD"), each_iter_values=(/100/), &
                                       print_level=medium_print_level, common_iter_levels=1, &
                                       add_last=add_last_numeric)
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum number of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "PROJECTED_AREA", &
                                       description="Controls the output of the projected area vector", &
                                       print_level=medium_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "WINDING_NUMBER", &
                                       description="Controls the output of the winding number vector", &
                                       print_level=medium_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "MOMENT_OF_INERTIA", &
                                       description="Controls the output of the moment of inertia vector", &
                                       print_level=medium_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "PLENGTH", &
                                       description="Controls the output of the helium permutation length", &
                                       print_level=medium_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      ! Properties printed at HIGH print level
      !
      CALL cp_print_key_section_create(print_key, "COORDINATES", &
                                       description="Controls the output of helium coordinates", &
                                       print_level=high_print_level, common_iter_levels=1)
      CALL keyword_create(keyword, name="FORMAT", &
                          description="Output file format for the coordinates", &
                          usage="FORMAT (PDB|XYZ)", &
                          default_i_val=fmt_id_pdb, &
                          enum_c_vals=s2a("PDB", "XYZ"), &
                          enum_i_vals=(/fmt_id_pdb, fmt_id_xyz/), &
                          enum_desc=s2a("Bead coordinates and connectivity is written in PDB format", &
                                        "Only bead coordinates are written in XYZ format"))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "PERM", &
                                       description="Controls the output of the helium permutation state", &
                                       print_level=high_print_level, common_iter_levels=1)
      CALL keyword_create(keyword, name="FORMAT", &
                          description="Output format for the permutation", &
                          usage="FORMAT (CYCLE|PLAIN)", &
                          default_i_val=perm_cycle, &
                          enum_c_vals=s2a("CYCLE", "PLAIN"), &
                          enum_i_vals=(/perm_cycle, perm_plain/), &
                          enum_desc=s2a( &
                          "Cycle notation with winding cycles enclosed"// &
                          " in '[...]' and non-winding ones enclosed in '(...)'", &
                          "Plain permutation output, i.e. P(1) ... P(N)"))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FORCES", &
                                       description="Controls the output of the helium forces on the solute", &
                                       print_level=high_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      ! Properties printed at DEBUG print level
      !
      CALL cp_print_key_section_create(print_key, "ACCEPTS", &
                                       description="Controls the output of the helium acceptance data", &
                                       print_level=debug_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      !
      CALL cp_print_key_section_create(print_key, "FORCES_INST", &
                                       description="Controls the output of the instantaneous helium forces on the solute", &
                                       print_level=debug_print_level, common_iter_levels=1)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      RETURN
   END SUBROUTINE create_helium_section

END MODULE input_cp2k_motion
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for the ATOM module
!> \author jgh
! **************************************************************************************************
MODULE input_cp2k_atom
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE input_constants,                 ONLY: &
        atom_basis_run, atom_energy_run, atom_no_run, atom_pseudo_run, barrier_conf, &
        contracted_gto, do_analytic, do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, &
        do_dkh4_atom, do_dkh5_atom, do_gapw_gcs, do_gapw_gct, do_gapw_log, do_nonrel_atom, &
        do_numeric, do_rhf_atom, do_rks_atom, do_rohf_atom, do_sczoramp_atom, do_semi_analytic, &
        do_uhf_atom, do_uks_atom, do_zoramp_atom, ecp_pseudo, gaussian, geometrical_gto, &
        gth_pseudo, no_conf, no_pseudo, numerical, poly_conf, slater, upf_pseudo
   USE input_cp2k_xc,                   ONLY: create_xc_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              logical_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_atom'

   PUBLIC :: create_atom_section

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Creates the input section for the atom code
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_atom_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",67)
      CALL section_create(section, name="ATOM", &
                          description="Section handling input for atomic calculations.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMIC_NUMBER", &
                          description="Specify the atomic number", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ELEMENT", &
                          description="Specify the element to be calculated", &
                          usage="ELEMENT char", n_var=1, type_of_var=char_t, &
                          default_c_val="H")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RUN_TYPE", &
                          description="Type of run that you want to perform "// &
                          "[ENERGY,BASIS_OPTIMIZATION,PSEUDOPOTENTIAL_OPTIMIZATION,,...] ", &
                          usage="RUN_TYPE (NONE|ENERGY|BASIS_OPTIMIZATION|PSEUDOPOTENTIAL_OPTIMIZATION)", &
                          default_i_val=atom_energy_run, &
                          enum_c_vals=s2a("NONE", "ENERGY", "BASIS_OPTIMIZATION", "PSEUDOPOTENTIAL_OPTIMIZATION"), &
                          enum_i_vals=(/atom_no_run, atom_energy_run, atom_basis_run, atom_pseudo_run/), &
                          enum_desc=s2a("Perform no run", &
                                        "Perform energy optimization", &
                                        "Perform basis optimization", &
                                        "Perform pseudopotential optimization"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COULOMB_INTEGRALS", &
                          description="Method to calculate Coulomb integrals", &
                          usage="COULOMB_INTEGRALS (ANALYTIC|SEMI_ANALYTIC|NUMERIC)", &
                          default_i_val=do_numeric, &
                          enum_c_vals=(/"ANALYTIC                 ", &
                                        "SEMI_ANALYTIC            ", &
                                        "NUMERIC                  "/), &
                          enum_i_vals=(/do_analytic, do_semi_analytic, do_numeric/), &
                          enum_desc=s2a("Use analytical method", &
                                        "Use semi-analytical method", &
                                        "Use numerical method"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCHANGE_INTEGRALS", &
                          description="Method to calculate Exchange integrals", &
                          usage="EXCHANGE_INTEGRALS (ANALYTIC|SEMI_ANALYTIC|NUMERIC)", &
                          default_i_val=do_numeric, &
                          enum_c_vals=(/"ANALYTIC                 ", &
                                        "SEMI_ANALYTIC            ", &
                                        "NUMERIC                  "/), &
                          enum_i_vals=(/do_analytic, do_semi_analytic, do_numeric/), &
                          enum_desc=s2a("Use analytical method", &
                                        "Use semi-analytical method", &
                                        "Use numerical method"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORE", &
                          description="Specifies the core electrons for a pseudopotential", &
                          usage="CORE 1s2 ... or CORE [Ne] or CORE none for 0 electron cores", repeats=.FALSE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ELECTRON_CONFIGURATION", &
                          description="Specifies the electron configuration. "// &
                          "Optional the multiplicity (m) and a core state [XX] can be declared", &
                          usage="ELECTRON_CONFIGURATION (1) [Ne] 3s2 ... ", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ANGULAR_MOMENTUM", &
                          description="Specifies the largest angular momentum calculated [0-3]", &
                          usage="MAX_ANGULAR_MOMENTUM 3", repeats=.FALSE., &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CALCULATE_STATES", &
                          description="Specifies the number of states calculated per l value", &
                          usage="CALCULATE_STATES  5 5 5 3 ", repeats=.FALSE., &
                          default_i_val=0, n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_atom_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_atom_aebasis_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_atom_ppbasis_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_atom_method_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_optimization_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_powell_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_atom_section

! **************************************************************************************************
!> \brief Create the print atom section
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_atom_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",201)
      CALL section_create(section, name="print", &
                          description="Section of possible print options specific of the ATOM code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key, keyword)

      CALL cp_print_key_section_create(print_key, "PROGRAM_BANNER", &
                                       description="Controls the printing of the banner of the ATOM program", &
                                       print_level=silent_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "METHOD_INFO", &
                                       description="Controls the printing of method information", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "BASIS_SET", &
                                       description="Controls the printing of the basis sets", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "POTENTIAL", &
                                       description="Controls the printing of the potentials", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "FIT_DENSITY", &
         description="Fit the total electronic density to a linear combination of Gaussian functions", &
         print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="NUM_GTO", &
                          description="Number of Gaussian type functions for density fit", &
                          usage="NUM_GTO integer ", type_of_var=integer_t, &
                          default_i_val=40)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FIT_KGPOT", &
                                       description="Fit an approximation to the non-additive kinetic energy potential used in KG", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="NUM_GAUSSIAN", &
                          description="Number of Gaussian terms for the fit", &
                          usage="NUM_GAUSSIAN integer ", type_of_var=integer_t, &
                          default_i_val=1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="NUM_POLYNOM", &
                          description="Number of terms in the polynomial expansion", &
                          usage="NUM_POLYNOM integer ", type_of_var=integer_t, &
                          default_i_val=4)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESPONSE_BASIS", &
                                       description="Calculate a response basis set contraction scheme", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="DELTA_CHARGE", &
                          description="Variation of charge used in finite difference calculation", &
                          usage="DELTA_CHARGE real ", type_of_var=real_t, &
                          default_r_val=0.05_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="DERIVATIVES", &
                          description="Number of wavefunction derivatives to calculate", &
                          usage="DERIVATIVES integer ", type_of_var=integer_t, &
                          default_i_val=2)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SCF_INFO", &
                                       description="Controls the printing of SCF information", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ORBITALS", &
                                       description="Controls the printing of the optimized orbitals information", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ANALYZE_BASIS", &
                                       description="Calculates some basis set analysis data", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="OVERLAP_CONDITION_NUMBER", &
                          description="Condition number of the basis set overlap matrix calculated for a cubic crystal", &
                          usage="OVERLAP_CONDITION_NUMBER <logical>", type_of_var=logical_t, default_l_val=.FALSE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="COMPLETENESS", &
                          description="Calculate a completeness estimate for the basis set.", &
                          usage="COMPLETENESS <logical>", type_of_var=logical_t, default_l_val=.FALSE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FIT_PSEUDO", &
                                       description="Controls the printing of FIT PSEUDO task", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FIT_BASIS", &
                                       description="Controls the printing of FIT BASIS task", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "UPF_FILE", &
                                       description="Write GTH pseudopotential in UPF format", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ENERGIES_MINUS_KINETIC", &
                                       description="Print out the total energy and orbital energies without "// &
                                       "the kinetic energy component. Useful for atomic calculations used"// &
                                       "during SCPTB parametrization", &
                                       print_level=debug_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)
   END SUBROUTINE create_atom_print_section

! **************************************************************************************************
!> \brief Create the all-electron basis section
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_atom_aebasis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_aebasis_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",348)
      CALL section_create(section, name="AE_BASIS", &
                          description="Section of basis set information for all-electron calculations.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL atom_basis_section(section)

   END SUBROUTINE create_atom_aebasis_section

! **************************************************************************************************
!> \brief Create the pseudopotential basis section
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_atom_ppbasis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_ppbasis_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",368)
      CALL section_create(section, name="PP_BASIS", &
                          description="Section of basis set information for pseudopotential calculations.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL atom_basis_section(section)

   END SUBROUTINE create_atom_ppbasis_section

! **************************************************************************************************
!> \brief Keywords in the atom basis section
!> \param section the section to fill
!> \author jgh
! **************************************************************************************************
   SUBROUTINE atom_basis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_basis_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",391)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="BASIS_TYPE", &
                          description="Basis set type", &
                          usage="BASIS_TYPE (GAUSSIAN|GEOMETRICAL_GTO|CONTRACTED_GTO|SLATER|NUMERICAL)", &
                          default_i_val=gaussian, &
                          enum_c_vals=(/"GAUSSIAN                 ", &
                                        "GEOMETRICAL_GTO          ", &
                                        "CONTRACTED_GTO           ", &
                                        "SLATER                   ", &
                                        "NUMERICAL                "/), &
                          enum_i_vals=(/gaussian, geometrical_gto, contracted_gto, slater, numerical/), &
                          enum_desc=s2a("Gaussian type orbitals", &
                                        "Geometrical Gaussian type orbitals", &
                                        "Contracted Gaussian type orbitals", &
                                        "Slater-type orbitals", &
                                        "Numerical basis type"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_GTO", &
                          description="Number of Gaussian type functions for s, p, d, ...", &
                          usage="NUM_GTO 5 5 5 ", n_var=-1, type_of_var=integer_t, &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_SLATER", &
                          description="Number of Slater type functions for s, p, d, ...", &
                          usage="NUM_SLATER 5 5 5 ", n_var=-1, type_of_var=integer_t, &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="START_INDEX", &
                          description="Starting index for Geometrical Basis sets", &
                          usage="START_INDEX 0 2 5 4 ", n_var=-1, type_of_var=integer_t, &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="S_EXPONENTS", &
                          description="Exponents for s functions", &
                          usage="S_EXPONENTS 1.0 2.0 ... ", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="P_EXPONENTS", &
                          description="Exponents for p functions", &
                          usage="P_EXPONENTS 1.0 2.0 ... ", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="D_EXPONENTS", &
                          description="Exponents for d functions", &
                          usage="D_EXPONENTS 1.0 2.0 ... ", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="F_EXPONENTS", &
                          description="Exponents for f functions", &
                          usage="F_EXPONENTS 1.0 2.0 ... ", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="S_QUANTUM_NUMBERS", &
                          description="Main quantum numbers for s functions", &
                          usage="S_QUANTUM_NUMBERS 1 2 ... ", n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="P_QUANTUM_NUMBERS", &
                          description="Main quantum numbers for p functions", &
                          usage="P_QUANTUM_NUMBERS 2 3 ... ", n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="D_QUANTUM_NUMBERS", &
                          description="Main quantum numbers for d functions", &
                          usage="D_QUANTUM_NUMBERS 3 4 ... ", n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="F_QUANTUM_NUMBERS", &
                          description="Main quantum numbers for f functions", &
                          usage="F_QUANTUM_NUMBERS 4 5 ... ", n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GEOMETRICAL_FACTOR", &
                          description="Geometrical basis: factor C in a*C^k", &
                          usage="GEOMETRICAL_FACTOR real", &
                          default_r_val=2.6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GEO_START_VALUE", &
                          description="Geometrical basis: starting value a in a*C^k", &
                          usage="GEO_START_VALUE real", &
                          default_r_val=0.016_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_SET_FILE_NAME", &
                          description="Name of the basis set file, may include a path", &
                          usage="BASIS_SET_FILE_NAME <FILENAME>", &
                          default_lc_val="BASIS_SET")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_SET", &
                          variants=s2a("ORBITAL_BASIS_SET", "ORB_BASIS"), &
                          description="The contracted Gaussian basis set", &
                          usage="BASIS_SET DZVP", default_c_val=" ", &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="QUADRATURE", &
                          description="Algorithm to construct the atomic radial grids", &
                          usage="QUADRATURE (GC_SIMPLE|GC_TRANSFORMED|GC_LOG)", &
                          enum_c_vals=s2a("GC_SIMPLE", "GC_TRANSFORMED", "GC_LOG"), &
                          enum_i_vals=(/do_gapw_gcs, do_gapw_gct, do_gapw_log/), &
                          enum_desc=s2a("Gauss-Chebyshev quadrature", &
                                        "Transformed Gauss-Chebyshev quadrature", &
                                        "Logarithmic transformed Gauss-Chebyshev quadrature"), &
                          default_i_val=do_gapw_log)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GRID_POINTS", &
                          description="Number of radial grid points", &
                          usage="GRID_POINTS integer", &
                          default_i_val=400)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_EIGENVALUE", &
                          description="Cutoff of overlap matrix eigenvalues included into basis", &
                          usage="EPS_EIGENVALUE real", &
                          default_r_val=1.e-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_basis_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE atom_basis_section

! **************************************************************************************************
!> \brief Create the method section for Atom calculations
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_atom_method_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_method_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection, keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",552)
      CALL section_create(section, name="METHOD", &
                          description="Section of information on method to use.", &
                          n_keywords=0, n_subsections=2, repeats=.TRUE.)

      CALL keyword_create(keyword, name="METHOD_TYPE", &
                          description="Type of electronic structure method to be used", &
                          usage="METHOD_TYPE (KOHN-SHAM|RKS|UKS|HARTREE-FOCK|RHF|UHF|ROHF)", &
                          default_i_val=do_rks_atom, &
                          enum_c_vals=(/"KOHN-SHAM                ", &
                                        "RKS                      ", &
                                        "UKS                      ", &
                                        "HARTREE-FOCK             ", &
                                        "RHF                      ", &
                                        "UHF                      ", &
                                        "ROHF                     "/), &
                          enum_i_vals=(/do_rks_atom, do_rks_atom, do_uks_atom, do_rhf_atom, &
                                        do_rhf_atom, do_uhf_atom, do_rohf_atom/), &
                          enum_desc=s2a("Kohn-Sham electronic structure method", &
                                        "Restricted Kohn-Sham electronic structure method", &
                                        "Unrestricted Kohn-Sham electronic structure method", &
                                        "Hartree-Fock electronic structure method", &
                                        "Restricted Hartree-Fock electronic structure method", &
                                        "Unrestricted Hartree-Fock electronic structure method", &
                                        "Restricted open-shell Hartree-Fock electronic structure method"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RELATIVISTIC", &
                          description="Type of scalar relativistic method to be used", &
                          usage="RELATIVISTIC (OFF|ZORA(MP)|scZORA(MP)|DKH(0)|DKH(1)|DKH(2)|DKH(3)|DKH(4)|(DKH(5))", &
                          default_i_val=do_nonrel_atom, &
                          enum_c_vals=(/"OFF                      ", &
                                        "ZORA(MP)                 ", &
                                        "scZORA(MP)               ", &
                                        "DKH(0)                   ", &
                                        "DKH(1)                   ", &
                                        "DKH(2)                   ", &
                                        "DKH(3)                   ", &
                                        "DKH(4)                   ", &
                                        "DKH(5)                   "/), &
                          enum_i_vals=(/do_nonrel_atom, do_zoramp_atom, do_sczoramp_atom, do_dkh0_atom, &
                                        do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, do_dkh5_atom/), &
                          enum_desc=s2a("Use no scalar relativistic method", &
                                        "Use ZORA method with atomic model potential", &
                                        "Use scaled ZORA method with atomic model potential", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 0", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 1", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 2", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 3", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 4", &
                                        "Use Douglas-Kroll-Hess Hamiltonian of order 5"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_xc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

! ZMP creating zubsection for the zmp calculations
      CALL create_zmp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_external_vxc(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_atom_method_section

! **************************************************************************************************
!> \brief Create the ZMP subsection for Atom calculations
!>
!> \param section ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE create_zmp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_zmp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection, keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",638)
      CALL section_create(section, name="ZMP", &
                          description="Section used to specify ZMP Potentials.", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="FILE_DENSITY", &
                          description="Specifies the filename containing the target density ", &
                          usage="FILE_DENSITY <FILENAME>", &
                          type_of_var=char_t, default_c_val="RHO_O.dat", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GRID_TOL", &
                          description="Tolerance in the equivalence of read-grid in ZMP method", &
                          usage="GRID_TOL <REAL>", default_r_val=1.E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Parameter used for the constraint in ZMP method", &
                          usage="LAMBDA <REAL>", default_r_val=10.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DM", &
                          description="read external density from density matrix", &
                          usage="DM <LOGICAL>", type_of_var=logical_t, default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_zmp_restart_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_zmp_section

! **************************************************************************************************
!> \brief Create the ZMP restart subsection for Atom calculations
!>
!> \param section ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE create_zmp_restart_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_zmp_restart_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",689)
      CALL section_create(section, name="RESTART", &
                          description="Section used to specify the restart option in the ZMP"// &
                          "procedure, and the file that must be read.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="FILE_RESTART", &
                          description="Specifies the filename containing the restart file density ", &
                          usage="FILE_RESTART <FILENAME>", &
                          type_of_var=char_t, default_c_val="RESTART.wfn", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_zmp_restart_section

! **************************************************************************************************
!> \brief Subroutine to create the external v_xc potential
!>
!> \param section ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE create_external_vxc(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_external_vxc', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",719)
      CALL section_create(section, name="EXTERNAL_VXC", &
                          description="Section used to specify exernal VXC Potentials.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="FILE_VXC", &
                          description="Specifies the filename containing the external vxc ", &
                          usage="FILE_VXC <FILENAME>", &
                          type_of_var=char_t, default_c_val="VXC.dat", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GRID_TOL", &
                          description="Tolerance in the equivalence of read-grid in ZMP method", &
                          usage="GRID_TOL <REAL>", default_r_val=1.E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_external_vxc

! **************************************************************************************************
!> \brief Create the optimization section for Atom calculations
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_optimization_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_optimization_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",753)
      CALL section_create(section, name="OPTIMIZATION", &
                          description="Section of information on optimization thresholds and algorithms.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of iterations for optimization", &
                          usage="MAX_ITER  50", default_i_val=200)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SCF", &
                          description="Convergence criterion for SCF", &
                          usage="EPS_SCF 1.e-10", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DAMPING", &
                          description="Damping parameter for extrapolation method", &
                          usage="DAMPING 0.4", default_r_val=0.4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_DIIS", &
                          description="Starting DIIS method at convergence to EPS_DIIS", &
                          usage="EPS_DIIS  0.01", default_r_val=10000._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_DIIS", &
                          description="Maximum number of DIIS vectors", &
                          usage="N_DIIS  6", default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_optimization_section

! **************************************************************************************************
!> \brief Create the potential section for Atom calculations
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_atom.F",805)
      CALL section_create(section, name="POTENTIAL", &
                          description="Section of information on potential.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="CONFINEMENT_TYPE", &
                          description="Define functional form of confinement potential.", &
                          usage="CONFINEMENT_TYPE (NONE|POLYNOM|BARRIER)", &
                          default_i_val=poly_conf, &
                          enum_c_vals=(/"NONE                     ", &
                                        "POLYNOM                  ", &
                                        "BARRIER                  "/), &
                          enum_i_vals=(/no_conf, poly_conf, barrier_conf/), &
                          enum_desc=s2a("Do not use confinement potential", &
                                        "Use polynomial confinement potential: a*R^b", &
                                        "Use a smooth barrier potential: a*F[r-ron)/w]"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONFINEMENT", &
                          description="Definition of parameters for confinement potential", &
                          usage="CONFINEMENT prefactor range exponent (POLYNOM)"// &
                          "CONFINEMENT prefactor range r_onset (BARRIER)", &
                          default_r_vals=(/0._dp, 0._dp, 0._dp/), &
                          repeats=.FALSE., n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PSEUDO_TYPE", &
                          description="Pseudopotential type", &
                          usage="PSEUDO_TYPE (NONE|GTH|UPF|ECP)", &
                          default_i_val=no_pseudo, &
                          enum_c_vals=(/"NONE                     ", &
                                        "GTH                      ", &
                                        "UPF                      ", &
                                        "ECP                      "/), &
                          enum_i_vals=(/no_pseudo, gth_pseudo, upf_pseudo, ecp_pseudo/), &
                          enum_desc=s2a("Do not use pseudopotentials", &
                                        "Use Goedecker-Teter-Hutter pseudopotentials", &
                                        "Use UPF norm-conserving pseudopotentials", &
                                        "Use ECP semi-local pseudopotentials"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POTENTIAL_FILE_NAME", &
                          description="Name of the pseudo potential file, may include a path", &
                          usage="POTENTIAL_FILE_NAME <FILENAME>", &
                          default_lc_val="POTENTIAL")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POTENTIAL_NAME", &
                          variants=(/"POT_NAME"/), &
                          description="The name of the pseudopotential for the defined kind.", &
                          usage="POTENTIAL_NAME <PSEUDO-POTENTIAL-NAME>", default_c_val=" ", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_gthpotential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_ecp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_potential_section

! **************************************************************************************************
!> \brief Creates the &GTH_POTENTIAL section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_gthpotential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gthpotential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="GTH_POTENTIAL", &
                          description="Section used to specify Potentials.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="CP2K Pseudo Potential Standard Format (GTH, ALL or KG)", &
                          repeats=.TRUE., type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_gthpotential_section

! **************************************************************************************************
!> \brief Creates the &ECP section
!> \param section the section to create
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_ecp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ecp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="ECP", &
                          description="Section used to specify ECP's.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Effective Core Potentials definition", &
                          repeats=.TRUE., type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_ecp_section

! **************************************************************************************************
!> \brief Creates the &BASIS section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_basis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_basis_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="basis", &
                          description="Section used to specify a general basis set for QM calculations.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="CP2K Basis Set Standard Format", repeats=.TRUE., &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_basis_section

! **************************************************************************************************
!> \brief Creates the &POWELL section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_powell_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_powell_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="powell", &
                          description="Section defines basic parameters for Powell optimization", &
                          n_keywords=4, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ACCURACY", &
                          description="Final accuracy requested in optimization (RHOEND)", &
                          usage="ACCURACY 0.00001", &
                          default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="Initial step size for search algorithm (RHOBEG)", &
                          usage="STEP_SIZE 0.005", &
                          default_r_val=0.005_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FUN", &
                          description="Maximum number of function evaluations", &
                          usage="MAX_FUN 1000", &
                          default_i_val=5000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_INIT", &
                          description="Maximum number of re-initialization of Powell method", &
                          usage="MAX_INIT 5", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_POT_VIRTUAL", &
                          description="Weight for virtual states in pseudopotential optimization", &
                          usage="WEIGHT_POT_VIRTUAL 1.0", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_POT_SEMICORE", &
                          description="Weight for semi core states in pseudopotential optimization", &
                          usage="WEIGHT_POT_SEMICORE 1.0", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_POT_VALENCE", &
                          description="Weight for valence states in pseudopotential optimization", &
                          usage="WEIGHT_POT_VALENCE 1.0", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_POT_NODE", &
                          description="Weight for node mismatch in pseudopotential optimization", &
                          usage="WEIGHT_POT_NODE 1.0", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_DELTA_ENERGY", &
                          description="Weight for energy differences in pseudopotential optimization", &
                          usage="WEIGHT_DELTA_ENERGY 1.0", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_ELECTRON_CONFIGURATION", &
                          description="Weight for different electronic states in optimization", &
                          usage="WEIGHT_ELECTRON_CONFIGURATION 1.0 0.1 ...", &
                          n_var=-1, type_of_var=real_t, default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_METHOD", &
                          description="Weight for different methods in optimization", &
                          usage="WEIGHT_METHOD 1.0 0.1 ...", &
                          n_var=-1, type_of_var=real_t, default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_POT_VIRTUAL", &
                          description="Target accuracy for virtual state eigenvalues in pseudopotential optimization", &
                          usage="TARGET_POT_VIRTUAL 0.0001", &
                          default_r_val=1.0e-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_POT_VALENCE", &
                          description="Target accuracy for valence state eigenvalues in pseudopotential optimization", &
                          usage="TARGET_POT_VALENCE 0.0001", &
                          default_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_POT_SEMICORE", &
                          description="Target accuracy for semicore state eigenvalues in pseudopotential optimization", &
                          usage="TARGET_POT_SEMICORE 0.01", &
                          default_r_val=1.0e-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET_DELTA_ENERGY", &
                          description="Target accuracy for energy differences in pseudopotential optimization", &
                          usage="TARGET_DELTA_ENERGY 0.01", &
                          default_r_val=1.0e-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_PSIR0", &
                          description="Weight for the wavefunctions at r=0 (only occupied states)", &
                          usage="WEIGHT_PSIR0 0.01", &
                          default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCOV_MULTIPLICATION", &
                          description="Multiply Rcov integration limit for charge conservation", &
                          usage="RCOV_MULTIPLICATION  1.10", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_powell_section

! **************************************************************************************************

END MODULE input_cp2k_atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for the VIBRATIONAL_ANALYSIS module
!> \par History
!>      01.2008 [tlaino] Teodoro Laino - University of Zurich
!>                       Creating an own module for vibrational analysis
!> \author [tlaino]
! **************************************************************************************************
MODULE input_cp2k_vib
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: do_rep_blocked,&
                                              do_rep_interleaved,&
                                              ms_guess_atomic,&
                                              ms_guess_bfgs,&
                                              ms_guess_molden,&
                                              ms_guess_restart,&
                                              ms_guess_restart_vec
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_vib'

   PUBLIC :: create_vib_section
CONTAINS

! **************************************************************************************************
!> \brief Creates the exteranal restart section
!> \param section the section to create
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_vib_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_vib_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_vib.F",65)
      CALL section_create( &
         section, name="VIBRATIONAL_ANALYSIS", &
         description="Section to setup parameters to perform a Normal Modes, vibrational, or phonon analysis. "// &
         "Vibrations are computed using finite differences, "// &
         "which implies a very tight (e.g. 1E-8) threshold is needed for EPS_SCF to get accurate low frequencies. "// &
         "The analysis assumes a stationary state (minimum or TS),"// &
         " i.e. tight geometry optimization (MAX_FORCE) is needed as well.", &
         n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="DX", &
                          description="Specify the increment to be used to construct the HESSIAN with "// &
                          "finite difference method", &
                          default_r_val=1.0E-2_dp, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NPROC_REP", &
                          description="Specify the number of processors to be used per replica "// &
                          "environment (for parallel runs). "// &
                          "In case of mode selective calculations more than one replica will start"// &
                          " a block Davidson algorithm to track more than only one frequency", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROC_DIST_TYPE", &
                          description="Specify the topology of the mapping of processors into replicas.", &
                          usage="PROC_DIST_TYPE (INTERLEAVED|BLOCKED)", &
                          enum_c_vals=s2a("INTERLEAVED", &
                                          "BLOCKED"), &
                          enum_desc=s2a("Interleaved distribution", &
                                        "Blocked distribution"), &
                          enum_i_vals=(/do_rep_interleaved, do_rep_blocked/), &
                          default_i_val=do_rep_blocked)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FULLY_PERIODIC", &
                          description="Avoids to clean rotations from the Hessian matrix.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTENSITIES", &
                          description="Calculation of the IR-Intensities. Calculation of dipoles has to be specified explicitly ", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="THERMOCHEMISTRY", &
                          description="Calculation of the thermochemical data. Valid for molecules in the gas phase. ", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TC_TEMPERATURE", &
                          description="Temperature for the calculation of the thermochemical data ", &
                          usage="tc_temperature 325.0", default_r_val=cp_unit_to_cp2k(value=273.150_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TC_PRESSURE", &
                          description="Pressure for the calculation of the thermochemical data  ", &
                          default_r_val=cp_unit_to_cp2k(value=101325.0_dp, unit_str="Pa"), unit_str="Pa")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_mode_selective_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_vib_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_vib_section

! **************************************************************************************************
!> \brief Create the print section for VIB
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - 10.2008
! **************************************************************************************************
   SUBROUTINE create_print_vib_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_vib_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_vib.F",158)
      CALL section_create(section, name="PRINT", &
                          description="Section controlling the print information during a vibrational "// &
                          "analysis.", n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, print_key)

      CALL cp_print_key_section_create(print_key, "BANNER", &
                                       description="Controls the printing of the vibrational analysis banner", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing basic info about the vibrational method", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "MOLDEN_VIB", &
                                       description="Controls the printing for visualization in molden format", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="VIBRATIONS")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ROTATIONAL_INFO", &
                                       description="Controls the printing basic info during the cleaning of the "// &
                                       "rotational degrees of freedom.", &
                                       print_level=debug_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      ! Print_key keywords
      CALL keyword_create(keyword, name="COORDINATES", &
                          description="Prints atomic coordinates after rotation", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_print_vib_section

! **************************************************************************************************
!> \brief Create the input section for MODE selective
!> \param section the section to create
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE create_mode_selective_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mode_selective_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, subsection, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_vib.F",213)
      CALL section_create(section, name="MODE_SELECTIVE", &
                          description="All parameters needed for to run a mode selective vibrational analysis", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="FREQUENCY", &
                          description="value close to the expected value of the frequency for to look for. "// &
                          "If the block Davidson algorithm is applied, the nrep closest frequencies are tracked. ", &
                          usage="FREQUENCY {REAL}", default_r_val=-1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANGE", &
                          description="Track modes in a given range of frequencies. "// &
                          "No warranty that the set of frequencies is complete.", &
                          usage="RANGE {REAL} {REAL}", &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the list of atoms which should be displaced for the Initial guess", &
                          usage="ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_MAX_VAL", &
                          description="Convergence criterium for the davidson algorithm. Specifies the maximal value in the "// &
                          "residuum vectors ", &
                          usage="EPS_MAX_VAL {REAL}", default_r_val=5.0E-7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_NORM", &
         description="Convergence criterium for the davidson algorithm. Specifies the maximal value of the norm "// &
         "of the residuum vectors ", &
         usage="EPS_NORM {REAL}", default_r_val=2.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="INITIAL_GUESS", &
         description="The type of initial guess for the normal modes", &
         usage="INITIAL_GUESS BFGS_HESS", &
         default_i_val=ms_guess_atomic, &
         enum_c_vals=s2a("BFGS_HESS", "ATOMIC", "RESTART", "RESTART_VEC", "MOLDEN_RESTART"), &
         enum_desc=s2a("get the first displacement vector out of the BFGS approximate Hessian", &
                       "use random displacements for a set of atoms specified", &
                       "use data from MS_RESTART as initial guess", &
                       "use a vector from MS_RESTART, useful if you want to increase accurcy by changing functionals or basis", &
                       "use the .mol file of a former run, to restart a vector"// &
                       "(similar to Restart_vec, but a different file FORMAT is used)"), &
         enum_i_vals=(/ms_guess_bfgs, ms_guess_atomic, ms_guess_restart, ms_guess_restart_vec, ms_guess_molden/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_FILE_NAME", &
                          description="Specifies the name of the file used to create the restarted vectors", &
                          usage="RESTART_FILE_NAME {filename}", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_involved_atoms_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PRINT", &
                          description="Controls the printing mode selective vibrational analysis", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL cp_print_key_section_create(print_key, "MS_RESTART", &
                                       description="Controls the printing of the Mode Selective Restart file.", &
                                       print_level=silent_print_level, common_iter_levels=1, &
                                       add_last=add_last_numeric, filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_mode_selective_section

! **************************************************************************************************
!> \brief Create the input section for Ivolved_atoms keyword in mode selective
!> \param section the section to create
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE create_involved_atoms_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_involved_atoms_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("motion/input_cp2k_vib.F",312)
      CALL section_create( &
         section, name="INVOLVED_ATOMS", &
         description="All parameters needed for the tracking of modes dominated by the motion of selected atoms ", &
         n_keywords=2, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create( &
         keyword, name="RANGE", &
         description=" Specifies the range of wavenumbers in which the modes related to the ATOMS have to be tracked. "// &
         " If not specified frequencies >400cm-1 will be used to avoid tracking of translational or rotational modes", &
         usage="RANGE {REAL} {REAL}", &
         n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="INVOLVED_ATOMS", &
         description="Specifies the list of atoms on which the tracked eigenvector should have the highest value "// &
         "similar to looking for the vibration of a set of atoms", &
         usage="INVOLVED_ATOMS {integer} {integer} .. {integer}", &
         n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_involved_atoms_section

END MODULE input_cp2k_vib
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the input structure for cp2k
!> \par History
!>      06.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k
   USE cp_dbcsr_interface,              ONLY: dbcsr_test_arnoldi,&
                                              dbcsr_test_binary_io,&
                                              dbcsr_test_mm,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE eri_mme_test,                    ONLY: create_eri_mme_test_section
   USE input_constants,                 ONLY: &
        do_diag_syevd, do_diag_syevx, do_mat_random, do_mat_read, do_pwgrid_ns_fullspace, &
        do_pwgrid_ns_halfspace, do_pwgrid_spherical, ehrenfest, numerical
   USE input_cp2k_atom,                 ONLY: create_atom_section
   USE input_cp2k_force_eval,           ONLY: create_force_eval_section
   USE input_cp2k_global,               ONLY: create_global_section
   USE input_cp2k_motion,               ONLY: create_motion_section
   USE input_cp2k_rsgrid,               ONLY: create_rsgrid_section
   USE input_cp2k_vib,                  ONLY: create_vib_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_optimize_basis,            ONLY: create_optimize_basis_section
   USE input_optimize_input,            ONLY: create_optimize_input_section
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              logical_t
   USE kinds,                           ONLY: dp
   USE pw_grids,                        ONLY: do_pw_grid_blocked_false,&
                                              do_pw_grid_blocked_free,&
                                              do_pw_grid_blocked_true
   USE string_utilities,                ONLY: newline,&
                                              s2a
   USE swarm_input,                     ONLY: create_swarm_section
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k'

   PUBLIC :: create_cp2k_root_section

CONTAINS

! **************************************************************************************************
!> \brief creates the input structure of the file used by cp2k
!> \param root_section the input structure to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_cp2k_root_section(root_section)
      TYPE(section_type), POINTER                        :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cp2k_root_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_type), POINTER                        :: section

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(root_section)))CALL cp__a("start/input_cp2k.F",84)
      CALL section_create(root_section, name="__ROOT__", &
                          description="input file of cp2k", n_keywords=0, n_subsections=10, &
                          repeats=.FALSE.)
      NULLIFY (section)

      CALL create_global_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_test_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_debug_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_motion_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_multi_force_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_force_eval_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_farming_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_optimize_input_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_optimize_basis_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_swarm_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_ext_restart_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_vib_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)

      CALL create_atom_section(section)
      CALL section_add_subsection(root_section, section)
      CALL section_release(section)
      CALL timestop(handle)

   END SUBROUTINE create_cp2k_root_section

! **************************************************************************************************
!> \brief section with the tests of the libraries or external code that cp2k uses
!> \param section the section to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_test_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_test_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      CALL section_create(section, name="TEST", &
                          description="Tests to perform on the supported libraries.", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)
      CALL keyword_create(keyword, name="MEMORY", &
                          description="Set the maximum amount of memory allocated for a given test (in bytes)", &
                          usage="MEMORY <REAL>", default_r_val=256.e6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COPY", &
                          description="Tests the performance to copy two vectors."// &
                          "The results of these tests allow to determine the size of the cache "// &
                          "of the CPU. This can be used to optimize the performance of the"// &
                          "FFTSG library. Tests are repeated the given number of times.", &
                          usage="copy 10", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MATMUL", &
                          description="Tests the performance of different kinds of matrix matrix "// &
                          "multiply kernels for the F95 INTRINSIC matmul. Matrices up to 2**N+1 will be tested. ", &
                          usage="matmul 10", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DGEMM", &
                          description="Tests the performance of different kinds of matrix matrix "// &
                          "multiply kernels for the BLAS INTRINSIC DGEMM. Matrices up to 2**N+1 will be tested. ", &
                          usage="DGEMM 10", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FFT", &
                          description="Tests the performance of all available FFT libraries for "// &
                          "3D FFTs Tests are repeated the given number of times.", &
                          usage="fft 10", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERI", &
                          description="Tests the performance and correctness of ERI libraries ", &
                          usage="eri 1", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CLEBSCH_GORDON", variants=(/"CLEBSCH"/), &
                          description="Tests the Clebsch-Gordon Coefficients. "// &
                          "Tests are repeated the given number of times.", &
                          usage="clebsch_gordon 10", default_i_val=0)

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MPI", &
                          description="Tests mpi, quickly adapted benchmark code,"// &
                          "will ONLY work on an even number of CPUs. comm is the relevant, "// &
                          "initialized communicator. This test will produce messages "// &
                          "of the size 8*10**requested_size, where requested_size is the value "// &
                          "given to this keyword", &
                          usage="mpi 6", default_i_val=0)

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANDOM_NUMBER_GENERATOR", variants=(/"rng"/), &
                          description=" Tests the parallel random number generator (RNG)", &
                          usage="rng 1000000", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MINIMAX", &
                          description="Tests validity of minimax coefficients for approximating 1/x "// &
                          "as a sum of exponentials. "// &
                          "Checks numerical error against tabulated error, testing "// &
                          "the given number of different Rc values.", &
                          usage="MINIMAX 1000", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create( &
         print_key, "GRID_INFORMATION", &
         description="Controls the printing of information regarding the PW and RS grid structures"// &
         " (ONLY for TEST run).", &
         print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="controls the printing of tests output", &
                                       print_level=silent_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      NULLIFY (subsection)
      CALL create_rs_pw_transfer_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eigensolver_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_pw_transfer_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cp_fm_gemm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cp_dbcsr_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eri_mme_test_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_test_section

! **************************************************************************************************
!> \brief section to setup debugging parameter
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_debug_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_debug_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      CALL section_create(section, name="DEBUG", &
                          description="Section to setup parameters for debug runs.", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="DEBUG_FORCES", &
                          description="Activates the debugging of the atomic forces", &
                          usage="DEBUG_FORCES {LOGICAL}", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DEBUG_STRESS_TENSOR", &
                          description="Activates the debugging of the stress tensor", &
                          usage="DEBUG_STRESS_TENSOR {LOGICAL}", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Increment for the calculation of the numerical derivatives", &
                          usage="DX {REAL}", default_r_val=0.001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_NO_ERROR_CHECK", &
                          description="The mismatch between the numerical and the "// &
                          "analytical value is not checked for analytical "// &
                          "values smaller than this threshold value", &
                          usage="EPS_NO_ERROR_CHECK {REAL}", default_r_val=1.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STOP_ON_MISMATCH", &
                          description="Stop the debug run when a mismatch between the numerical and "// &
                          "the analytical value is detected", &
                          usage="STOP_ON_MISMATCH {LOGICAL}", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of the DEBUG specific output", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_debug_section

! **************************************************************************************************
!> \brief creates the multiple force_eval section
!> \param section the section to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_multi_force_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multi_force_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",358)
      CALL section_create(section, name="MULTIPLE_FORCE_EVALS", &
                          description="Describes how to handle multiple force_evals.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="FORCE_EVAL_ORDER", &
                          description='Specify the orders of the different force_eval. When using a MIXED force_eval'// &
                          " this does not need to be specified in this list, because it that takes into account only the real"// &
                          " energy contributions", &
                          usage="FORCE_EVAL_ORDER <INTEGER> .. <INTEGER>", type_of_var=integer_t, n_var=-1, &
                          default_i_vals=(/1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIPLE_SUBSYS", &
                          description="Specify if force_eval have different subsys. In case they share the same subsys,"// &
                          " it needs to be specified only in the MIXED force_eval (if using MIXED) or"// &
                          " in the force_eval corresponding to first force_eval of the previous order (when not using MIXED).", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_multi_force_section

! **************************************************************************************************
!> \brief Creates the exteranal restart section
!> \param section the section to create
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_ext_restart_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ext_restart_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",396)
      CALL section_create(section, name="EXT_RESTART", &
                          description="Section for external restart, specifies an external "// &
                          "input file where to take positions,...", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="RESTART_FILE_NAME", variants=(/"EXTERNAL_FILE"/), &
                          description="Specifies the name of restart file (or any other input file)"// &
                          " to be read. Only fields relevant to a restart will be used"// &
                          " (unless switched off with the keywords in this section)", &
                          default_lc_val=" ")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BINARY_RESTART_FILE_NAME", &
                          variants=(/"BINARY_RESTART_FILE"/), &
                          description="Specifies the name of an additional restart file "// &
                          "from which selected input sections are read in binary format "// &
                          "(see SPLIT_RESTART_FILE).", &
                          default_lc_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_DEFAULT", &
                          description="This keyword controls the default value for all possible "// &
                          " restartable keywords, unless explicitly defined. For example setting"// &
                          " this keyword to .FALSE. does not restart any quantity. If, at the "// &
                          " same time, one keyword is set to .TRUE. only that quantity will be"// &
                          " restarted.", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_COUNTERS", &
                          description="Restarts the counters in MD schemes", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_POS", &
                          description="Takes the positions from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_VEL", &
                          description="Takes the velocities from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_RANDOMG", &
                          description="Restarts the random number generator from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_SHELL_POS", &
                          description="Takes the positions of the shells from the external file (only if shell-model)", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_CORE_POS", &
                          description="Takes the positions of the cores from the external file (only if shell-model)", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_OPTIMIZE_INPUT_VARIABLES", &
                          description="Restart with the optimize input variables", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_SHELL_VELOCITY", &
                          description="Takes the velocities of the shells from the external file (only if shell-model)", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_CORE_VELOCITY", &
                          description="Takes the velocities of the shells from the external file (only if shell-model)", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_BAROSTAT", &
                          description="Restarts the barostat from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_BAROSTAT_THERMOSTAT", &
                          description="Restarts the barostat thermostat from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_SHELL_THERMOSTAT", &
                          description="Restarts the shell thermostat from the external file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_THERMOSTAT", &
                          description="Restarts the nose thermostats of the particles "// &
                          "from the EXTERNAL file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_CELL", &
                          description="Restarts the cell (and cell_ref) "// &
                          "from the EXTERNAL file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_METADYNAMICS", &
                          description="Restarts hills from a previous metadynamics run "// &
                          "from the EXTERNAL file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_WALKERS", &
                          description="Restarts walkers informations from a previous metadynamics run "// &
                          "from the EXTERNAL file", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_BAND", &
                          description="Restarts positions and velocities of the Band.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_QMMM", &
                          description="Restarts the following specific QMMM info: translation vectors.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_CONSTRAINT", &
                          description="Restarts constraint section. It's necessary when doing restraint"// &
                          " calculation to have a perfect energy conservation. For constraints only it's"// &
                          " use is optional.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_BSSE", &
                          description="Restarts information for BSSE calculations.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_DIMER", &
                          description="Restarts information for DIMER geometry optimizations.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_AVERAGES", &
                          description="Restarts information for AVERAGES.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_RTP", &
                          description="Restarts information for REAL TIME PROPAGATION and EHRENFEST DYNAMICS.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CUSTOM_PATH", &
                          description="Restarts the given path from the EXTERNAL file. Allows a major flexibility for restarts.", &
                          type_of_var=char_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! PIMD
      CALL keyword_create(keyword, name="RESTART_PINT_POS", &
                          description="Restart bead positions from PINT%BEADS%COORD.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_PINT_VEL", &
                          description="Restart bead velocities from PINT%BEADS%VELOCITY.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_PINT_NOSE", &
                          description="Restart Nose thermostat for beads from PINT%NOSE.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_PINT_GLE", &
                          description="Restart GLE thermostat for beads from PINT%GLE.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! PIMC
      CALL keyword_create(keyword, name="RESTART_HELIUM_POS", &
                          description="Restart helium positions from PINT%HELIUM%COORD.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_HELIUM_PERMUTATION", &
                          description="Restart helium permutation state from PINT%HELIUM%PERM.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_HELIUM_FORCE", &
                          description="Restart helium forces exerted on the solute from PINT%HELIUM%FORCE.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_HELIUM_RNG", &
                          description="Restarts helium random number generators from PINT%HELIUM%RNG_STATE.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_HELIUM_DENSITIES", &
                          description="Restarts helium density distributions from PINT%HELIUM%RHO.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="RESTART_HELIUM_AVERAGES", &
                          description="Restarts average properties from PINT%HELIUM%AVERAGES.", &
                          type_of_var=logical_t, lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ext_restart_section

! **************************************************************************************************
!> \brief creates the farming section
!> \param section the section to create
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_farming_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_farming_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, sub_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",628)
      CALL section_create(section, name="farming", &
                          description="Describes a farming job, in which multiple inputs are executed."//newline// &
                          "The RUN_TYPE in the global section has to be set to NONE for FARMING."//newline// &
                          "The different groups are executed in parallel. The jobs inside the same groups in series.", &
                          repeats=.FALSE.)
      NULLIFY (keyword, print_key)

      CALL keyword_create( &
         keyword, name="MASTER_SLAVE", &
         description="If a master-slave setup should be employed, in which one process is used to distribute the tasks. "// &
         "This is most useful to load-balance if not all jobs have the same length, "// &
         "and a lot of CPUs/groups are availabe.", &
         usage="MASTER_SLAVE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGROUPS", variants=(/"NGROUP"/), &
                          description="Gives the preferred number of working groups.", &
                          usage="ngroups 4", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GROUP_SIZE", &
                          description="Gives the preferred size of a working group, "// &
                          "groups will always be equal or larger than this size.", &
                          usage="group_size 2", default_i_val=8)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRIDE", &
                          description="Stride to be used when building working groups from the parent MPI comm."// &
                          "Can be used to layout slave groups over nodes in advanced ways (1 rank per node / 2 groups per node).", &
                          usage="STRIDE 2", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GROUP_PARTITION", &
                          description="gives the exact number of processors for each group.", &
                          usage="group_partition  2 2 4 2 4 ", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_JOBS_PER_GROUP", &
                          variants=(/"MAX_JOBS"/), &
                          description="maximum number of jobs executed per group", &
                          usage="max_step 4", default_i_val=HUGE(0)/32768)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="CYCLE", &
         description="If farming should process all jobs in a cyclic way, stopping only if MAX_JOBS_PER_GROUP is exceeded.", &
         usage="CYCLE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="WAIT_TIME", &
         description="Time to wait [s] for a new task if no task is currently available, make this zero if no clock is available", &
         usage="WAIT_TIME 0.1", default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (sub_section)
      CALL section_create(sub_section, name="JOB", &
                          description="description of the jobs to be executed", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="DIRECTORY", &
                          description="the directory in which the job should be executed", &
                          usage="DIRECTORY /my/path", &
                          default_lc_val=".")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INPUT_FILE_NAME", &
                          description="the filename of the input file", &
                          usage="INPUT_FILE_NAME my_input.inp", &
                          default_lc_val="input.inp")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="OUTPUT_FILE_NAME", &
         description="the filename of the output file, if not specified will use the project name in the &GLOBAL section.", &
         usage="OUTPUT_FILE_NAME my_input.inp", &
         default_lc_val="")
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="JOB_ID", &
                          description="An ID used to indentify a job in DEPENDENCIES. "// &
                          "JOB_IDs do not need to be unique, dependencies will be on all jobs with a given ID. "// &
                          "If no JOB_ID is given, the index of the &JOB section in the input file will be used. ", &
                          usage="JOB_ID 13", type_of_var=integer_t)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="DEPENDENCIES", &
         description="specifies a list of JOB_IDs on which the current job depends. "// &
         "The current job will not be executed before all the dependencies have finished. "// &
         "The keyword requires a MASTER_SLAVE farming run. "// &
         "Beyond the default case, some special cases might arise: "// &
         "1) circular dependencies will lead to a deadlock. "// &
         "2) This keyword is not compatible with CYCLE. "// &
         "3) MAX_JOBS_PER_GROUP is ignored (though only a total of MAX_JOBS_PER_GROUP*NGROUPS jobs will be executed) "// &
         "4) dependencies on jobs that will not be executed (due to RESTART or MAX_JOBS_PER_GROUP) are ignored. "// &
         "Additionally, note that, on some file systems, "// &
         " output (restart) files might not be immediately available on all compute nodes,"// &
         "potentially resulting in unexpected failures.", &
         usage="DEPENDENCIES 13 1 7", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(sub_section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, sub_section)
      CALL section_release(sub_section)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="controls the printing of FARMING specific output", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL keyword_create(keyword, name="DO_RESTART", &
                          description="Restart a farming job (and should pick up where the previous left off)", &
                          usage="DO_RESTART", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_FILE_NAME", &
                          description="Name of the restart file to use for restarting a FARMING run. If not "// &
                          "specified the name is determined from PROJECT name.", &
                          usage="RESTART_FILE_NAME <FILENAME>", type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="controls the printing of the restart for FARMING.", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="FARMING")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_farming_section

! **************************************************************************************************
!> \brief   creates the rs_pw_transfer section for use in the test section
!> \param section ...
!> \date    2008-03-09
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_rs_pw_transfer_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_rs_pw_transfer_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",788)
      CALL section_create(section, name="RS_PW_TRANSFER", &
                          description="Describes how to benchmark the rs_pw_transfer routines.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="GRID", &
                          description="Specify the number of grid points (not all grid points are allowed)", &
                          usage="GRID_DIMENSIONS 128 128 128", type_of_var=integer_t, n_var=3, &
                          default_i_vals=(/128, 128, 128/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HALO_SIZE", &
                          description="number of grid points of the halo", &
                          usage="HALO_SIZE 17", default_i_val=17)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_LOOP", &
                          description="Number of rs_pw_transfers being timed", &
                          usage="N_LOOP 100", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RS2PW", &
                          description="should the direction be rs2pw (pw2rs otherwise)", &
                          usage="rs2pw TRUE", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_rsgrid_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_rs_pw_transfer_section

! **************************************************************************************************
!> \brief   creates the rs_pw_transfer section for use in the test section
!> \param section ...
!> \date    2008-03-09
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_pw_transfer_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pw_transfer_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",840)
      CALL section_create(section, name="PW_TRANSFER", &
                          description="Benchmark and test the pw_transfer routines.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="GRID", &
                          description="Specify the number of grid points (not all grid points are allowed)", &
                          usage="GRID_DIMENSIONS 128 128 128", type_of_var=integer_t, n_var=3, &
                          default_i_vals=(/128, 128, 128/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_LOOP", &
                          description="Number of pw_transfers (backward&forward) being timed", &
                          usage="N_LOOP 100", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID", &
                          description="What kind of PW_GRID should be employed", &
                          usage="PW_GRID NS-FULLSPACE", &
                          enum_c_vals=s2a("SPHERICAL", "NS-FULLSPACE", "NS-HALFSPACE"), &
                          enum_desc=s2a("- not tested", " tested", " - not tested"), &
                          enum_i_vals=(/do_pwgrid_spherical, do_pwgrid_ns_fullspace, do_pwgrid_ns_halfspace/), &
                          default_i_val=do_pwgrid_ns_fullspace)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID_LAYOUT_ALL", &
                          description="loop overal all PW_GRID_LAYOUTs that are compatible with a given number of CPUs ", &
                          usage="PW_GRID_LAYOUT_ALL", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DEBUG", &
                          description="Do the FFT in debug mode in all cases", &
                          usage="DEBUG", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID_LAYOUT", &
                          description="Expert use only, leave the default..."// &
                          "Can be used to set the distribution for ray-distributed FFT.", &
                          usage="PW_GRID_LAYOUT", &
                          repeats=.FALSE., n_var=2, &
                          default_i_vals=(/-1, -1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID_BLOCKED", &
                          description="Expert use only, leave the default..."// &
                          "Can be used to set the distribution in g-space for the pw grids and their FFT.", &
                          usage="PW_GRID_BLOCKED FREE", &
                          enum_c_vals=s2a("FREE", "TRUE", "FALSE"), &
                          enum_desc=s2a("CP2K will select the optimal value", "blocked", "not blocked"), &
                          enum_i_vals=(/do_pw_grid_blocked_free, do_pw_grid_blocked_true, do_pw_grid_blocked_false/), &
                          default_i_val=do_pw_grid_blocked_false)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_pw_transfer_section

! **************************************************************************************************
!> \brief   creates the cp_fm_gemm section for use in the test section
!> \param section ...
!> \date    2009-06-15
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_cp_fm_gemm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cp_fm_gemm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",917)
      CALL section_create(section, name="CP_FM_GEMM", &
                          description="Benchmark and test the cp_fm_gemm routines by multiplying C=A*B  ", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="N_LOOP", &
                          description="Number of cp_fm_gemm operations being timed (useful for small matrices).", &
                          usage="N_LOOP 10", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Dimension 1 of C", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="M", &
                          description="Inner dimension M   ", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="N", &
                          description="Dimension 2 of C", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NROW_BLOCK", &
                          description="block_size for rows", &
                          usage="nrow_block 64", default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NCOL_BLOCK", &
                          description="block_size for cols", &
                          usage="nrow_block 64", default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROW_MAJOR", &
                          description="Use a row major blacs grid", &
                          usage="ROW_MAJOR .FALSE.", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FORCE_BLOCKSIZE", &
                          description="Forces the blocksize, even if this implies that a few processes might have no data", &
                          usage="FORCE_BLOCKSIZE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GRID_2D", &
                          description="Explicitly set the blacs 2D processor layout."// &
                          " If the product differs from the number of MPI ranks,"// &
                          " it is ignored and a default nearly square layout is used.", n_var=2, &
                          usage="GRID_2D 64 16 ", default_i_vals=(/1, 1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRANSA", &
                          description="Transpose matrix A", &
                          usage="TRANSA", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRANSB", &
                          description="Transpose matrix B", &
                          usage="TRANSB", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_cp_fm_gemm_section

! **************************************************************************************************
!> \brief   creates the eigensolver section for use in the test section
!> \param section ...
!> \date    2010-03-10
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_eigensolver_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eigensolver_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",1005)
      CALL section_create(section, name="EIGENSOLVER", &
                          description="Benchmark and test the eigensolver routines.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="N", &
                          description="Dimension of the square matrix", &
                          usage="N 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_LOOP", &
                          description="Number of operations being timed (useful for small matrices).", &
                          usage="N_LOOP 10", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIAG_METHOD", &
                          description="Diagonalization strategy", &
                          usage="DIAG_METHOD syevd", &
                          enum_c_vals=s2a("syevd", "syevx"), &
                          enum_desc=s2a("(sca)lapacks syevd", "(sca)lapacks syevx"), &
                          enum_i_vals=(/do_diag_syevd, do_diag_syevx/), &
                          default_i_val=do_diag_syevd)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EIGENVALUES", &
                          description="number of eigenvalues to be computed (all=<0) ", &
                          usage="EIGENVALUES 13", default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INIT_METHOD", &
                          description="Initialization approach", &
                          usage="INIT_METHOD RANDOM", &
                          enum_c_vals=s2a("random", "read"), &
                          enum_desc=s2a("use a random initial matrix", "read a matrix from file MATRIX"), &
                          enum_i_vals=(/do_mat_random, do_mat_read/), &
                          default_i_val=do_mat_random)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_eigensolver_section

! **************************************************************************************************
!> \brief   creates the cp_dbcsr section for use in the test section
!> \param section ...
!> \date    2010-02-08
!> \author  Urban Borstnik
! **************************************************************************************************
   SUBROUTINE create_cp_dbcsr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cp_dbcsr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("start/input_cp2k.F",1065)
      CALL section_create(section, name="CP_DBCSR", &
                          description="Benchmark and test the cp_dbcsr routines", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="N_LOOP", &
                          description="Number of operations being timed (useful for small matrices).", &
                          usage="N_LOOP 10", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DATA_TYPE", &
                          description="Data type of the matrices", &
                          usage="DATA_TYPE real_4", &
                          default_i_val=dbcsr_type_real_8, &
                          enum_c_vals=s2a("real_4", "real_8", "complex_4", "complex_8"), &
                          enum_i_vals=(/dbcsr_type_real_4, dbcsr_type_real_8, &
                                        dbcsr_type_complex_4, dbcsr_type_complex_8/), &
                          enum_desc=s2a( &
                          "Real (Single Precision)", &
                          "Real (Double Precision)", &
                          "Complex (Single Precision)", &
                          "Complex (Double Precision)"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEST_TYPE", &
                          description="Which part of DBCSR is tested", &
                          usage="TEST_TYPE MM", &
                          default_i_val=dbcsr_test_mm, &
                          enum_c_vals=s2a("MM", "ARNOLDI", "Binary_IO"), &
                          enum_i_vals=(/dbcsr_test_mm, dbcsr_test_arnoldi, dbcsr_test_binary_io/), &
                          enum_desc=s2a( &
                          "Run matrix multiplications", &
                          "Run the Arnoldi eigenvalue routines (acts as check for the matrix vector part as well)", &
                          "Run binary IO checks"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="M", &
                          description="Dimension 1 of C", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="N", &
                          description="Dimension 2 of C", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="K", &
                          description="Inner dimension M   ", &
                          usage="A 1024", default_i_val=256)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRANSA", &
                          description="Transpose matrix A", &
                          usage="TRANSA", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRANSB", &
                          description="Transpose matrix B", &
                          usage="TRANSB", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BS_M", &
                          description="Row block sizes of C", n_var=-1, &
                          usage="BS_M 1 13 2 5", default_i_vals=(/1, 13, 2, 15/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BS_N", &
                          description="Column block sizes of C", n_var=-1, &
                          usage="BS_N 1 13 2 5", default_i_vals=(/1, 13, 2, 15/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BS_K", &
                          description="Block sizes of inner dimension", n_var=-1, &
                          usage="BS_K 1 13 2 5", default_i_vals=(/1, 13, 2, 15/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATYPE", &
                          description="Matrix A type", &
                          usage="ATYPE N", default_c_val='N')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="BTYPE", &
                          description="Matrix B type", &
                          usage="BTYPE N", default_c_val='N')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CTYPE", &
                          description="Matrix C type", &
                          usage="CTYPE N", default_c_val='N')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NPROC", &
                          description="Number of processors to test", n_var=-1, &
                          usage="NPROC 128 16 1", default_i_vals=(/0/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KEEPSPARSE", &
                          description="Keep product sparse", &
                          usage="KEEPSPARSE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ASPARSITY", &
                          description="Sparsity of A matrix", &
                          usage="ASPARSITY 70", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BSPARSITY", &
                          description="Sparsity of B matrix", &
                          usage="ASPARSITY 80", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CSPARSITY", &
                          description="Sparsity of C matrix", &
                          usage="ASPARSITY 90", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA", &
                          description="Multiplication factor", &
                          usage="ALPHA 2.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BETA", &
                          description="Product premultiplication factor", &
                          usage="BETA 1.0", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FILTER_EPS", &
                          description="Threshold for on-the-fly and final filtering.", &
                          usage="FILTER_EPS 1.0", default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALWAYS_CHECKSUM", &
                          description="perform a checksum after each multiplication", &
                          usage="ALWAYS_CHECKSUM", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_cp_dbcsr_section

END MODULE input_cp2k
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE farming_methods
   USE cp_files,                        ONLY: get_unit_number
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE farming_types,                   ONLY: farming_env_type,&
                                              init_job_type,&
                                              job_finished,&
                                              job_pending,&
                                              job_running
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE message_passing,                 ONLY: mp_bcast
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: farming_parse_input, get_next_job

   ! must be negative in order to avoid confusion with job numbers
   INTEGER, PARAMETER, PUBLIC    :: do_nothing = -1, &
                                    do_wait = -2, &
                                    do_deadlock = -3

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'farming_methods'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param farming_env ...
!> \param start ...
!> \param END ...
!> \param current ...
!> \param todo ...
! **************************************************************************************************
   SUBROUTINE get_next_job(farming_env, start, END, current, todo)
      TYPE(farming_env_type), POINTER                    :: farming_env
      INTEGER, INTENT(IN)                                :: start, END
      INTEGER, INTENT(INOUT)                             :: current
      INTEGER, INTENT(OUT)                               :: todo

      INTEGER                                            :: icheck, idep, itry, ndep
      LOGICAL                                            :: dep_ok

      IF (farming_env%cycle) THEN
         IF (current < start) THEN
            current = start
         ELSE
            current = current+1
         ENDIF
         IF (current > END) THEN
            todo = do_nothing
         ELSE
            todo = MODULO(current-1, farming_env%njobs)+1
         ENDIF
      ELSE
         ! find a pending job
         itry = start
         todo = do_nothing
         DO itry = start, END
            IF (farming_env%job(itry)%status == job_pending) THEN

               ! see if all dependencies are OK
               ndep = SIZE(farming_env%job(itry)%dependencies)
               dep_ok = .TRUE.
               dep: DO idep = 1, ndep
                  DO icheck = start, END
                     IF (farming_env%job(icheck)%status .NE. job_finished) THEN
                        IF (farming_env%job(icheck)%id == farming_env%job(itry)%dependencies(idep)) THEN
                           dep_ok = .FALSE.
                           EXIT dep
                        ENDIF
                     ENDIF
                  ENDDO
               ENDDO dep

               ! if there are pending jobs, the slave can not be told to stop
               ! at least wait if there are unresolved dependencies
               IF (dep_OK) THEN
                  todo = itry
                  EXIT
               ELSE
                  todo = do_wait
               ENDIF
            ENDIF
         ENDDO
         ! If we have to wait, but there are no running jobs we are deadlocked
         ! which we signal
         IF (todo == do_wait) THEN
            dep_OK = .FALSE.
            DO itry = start, END
               IF (farming_env%job(itry)%status .EQ. job_running) dep_OK = .TRUE.
            ENDDO
            IF (.NOT. dep_OK) todo = do_deadlock
         ENDIF
      ENDIF
   END SUBROUTINE get_next_job

! **************************************************************************************************
!> \brief ...
!> \param farming_env ...
!> \param root_section ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE farming_parse_input(farming_env, root_section, para_env)
      TYPE(farming_env_type), POINTER                    :: farming_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'farming_parse_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: text
      INTEGER                                            :: i, iunit, n_rep_val, num_slaves, &
                                                            output_unit, stat
      INTEGER, DIMENSION(:), POINTER                     :: dependencies, i_vals
      LOGICAL                                            :: explicit, has_dep
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: farming_section, jobs_section, print_key

      NULLIFY (farming_section, jobs_section, print_key, logger, dependencies, i_vals)
      logger => cp_get_default_logger()
      farming_env%group_size_wish_set = .FALSE.
      farming_env%ngroup_wish_set = .FALSE.
      farming_section => section_vals_get_subs_vals(root_section, "FARMING")

      IF (ASSOCIATED(farming_env%group_partition)) THEN
         DEALLOCATE (farming_env%group_partition)
      END IF

      ! The following input order is used
      ! 1) GROUP_PARTITION
      ! 2) NGROUP
      ! 3) GROUP_SIZE (default 8)
      CALL section_vals_val_get(farming_section, "GROUP_PARTITION", &
                                n_rep_val=n_rep_val)
      IF (n_rep_val > 0) THEN
         CALL section_vals_val_get(farming_section, "GROUP_PARTITION", &
                                   i_vals=i_vals)
         ALLOCATE (farming_env%group_partition(0:SIZE(i_vals)-1))
         farming_env%group_partition(:) = i_vals
         farming_env%ngroup_wish_set = .TRUE.
         farming_env%ngroup_wish = SIZE(i_vals)
      ELSE
         CALL section_vals_val_get(farming_section, "NGROUP", &
                                   n_rep_val=n_rep_val)
         IF (n_rep_val > 0) THEN
            CALL section_vals_val_get(farming_section, "NGROUP", &
                                      i_val=farming_env%ngroup_wish)
            farming_env%ngroup_wish_set = .TRUE.
         ELSE
            CALL section_vals_val_get(farming_section, "GROUP_SIZE", &
                                      i_val=farming_env%group_size_wish)
            farming_env%group_size_wish_set = .TRUE.
         END IF
      END IF
      CALL section_vals_val_get(farming_section, "STRIDE", &
                                i_val=farming_env%stride)

      CALL section_vals_val_get(farming_section, "RESTART_FILE_NAME", &
                                explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(farming_section, "RESTART_FILE_NAME", &
                                   c_val=farming_env%restart_file_name)
      ELSE
         print_key => section_vals_get_subs_vals(farming_section, "RESTART")
         farming_env%restart_file_name = cp_print_key_generate_filename(logger, print_key, extension=".restart", &
                                                                        my_local=.FALSE.)
      END IF

      CALL section_vals_val_get(farming_section, "DO_RESTART", &
                                l_val=farming_env%restart)
      CALL section_vals_val_get(farming_section, "MAX_JOBS_PER_GROUP", &
                                i_val=farming_env%max_steps)
      CALL section_vals_val_get(farming_section, "CYCLE", &
                                l_val=farming_env%cycle)
      CALL section_vals_val_get(farming_section, "WAIT_TIME", &
                                r_val=farming_env%wait_time)

      CALL section_vals_val_get(farming_section, "MASTER_SLAVE", &
                                l_val=farming_env%master_slave)

      jobs_section => section_vals_get_subs_vals(farming_section, "JOB")
      CALL section_vals_get(jobs_section, n_repetition=farming_env%njobs)

      ALLOCATE (farming_env%Job(farming_env%njobs))
      CALL init_job_type(farming_env%job)

      has_dep = .FALSE.
      DO i = 1, farming_env%njobs
         CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                   keyword_name="DIRECTORY", c_val=farming_env%Job(i)%cwd)
         CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                   keyword_name="INPUT_FILE_NAME", c_val=farming_env%Job(i)%input)
         CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                   keyword_name="OUTPUT_FILE_NAME", c_val=farming_env%Job(i)%output)

         ! if job id is not specified the job id is the index
         CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                   keyword_name="JOB_ID", n_rep_val=n_rep_val)
         IF (n_rep_val == 0) THEN
            farming_env%Job(i)%id = i
         ELSE
            CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                      keyword_name="JOB_ID", i_val=farming_env%Job(i)%id)
         ENDIF

         ! get dependencies
         CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                   keyword_name="DEPENDENCIES", n_rep_val=n_rep_val)
         IF (n_rep_val == 0) THEN
            ALLOCATE (farming_env%Job(i)%dependencies(0))
         ELSE
            CALL section_vals_val_get(jobs_section, i_rep_section=i, &
                                      keyword_name="DEPENDENCIES", i_vals=dependencies)
            ALLOCATE (farming_env%Job(i)%dependencies(SIZE(dependencies, 1)))
            farming_env%Job(i)%dependencies = dependencies
            IF (SIZE(dependencies, 1) .NE. 0) has_dep = .TRUE.
         ENDIF
      END DO

      IF (has_dep) THEN
         IF(.NOT.(farming_env%master_slave))CALL cp__a("farming_methods.F",235)
         IF(.NOT.(.NOT. farming_env%cycle))CALL cp__a("farming_methods.F",236)
      ENDIF

      output_unit = cp_print_key_unit_nr(logger, farming_section, "PROGRAM_RUN_INFO", &
                                         extension=".log")

      ! master slave not supported
      IF (para_env%num_pe == 1) THEN
         farming_env%master_slave = .FALSE.
         WRITE (output_unit, FMT="(T2,A)") "FARMING| Master-slave setup not support for serial runs"
      ENDIF
      IF (farming_env%master_slave) THEN
         num_slaves = para_env%num_pe-1
      ELSE
         num_slaves = para_env%num_pe
      ENDIF

      IF (output_unit > 0) THEN
         WRITE (output_unit, FMT="(T2,A,T71,I10)") "FARMING| Number of jobs found", farming_env%njobs
         IF (farming_env%ngroup_wish_set) THEN
            WRITE (output_unit, FMT="(T2,A,T71,I10)") "FARMING| Ngroup wish:", farming_env%ngroup_wish
            IF (ASSOCIATED(farming_env%group_partition)) THEN
               WRITE (output_unit, FMT="(T2,A)", ADVANCE="NO") "FARMING| User partition:"
               DO i = 0, SIZE(farming_env%group_partition)-1
                  IF (MODULO(i, 4) == 0) WRITE (output_unit, *)
                  WRITE (output_unit, FMT='(I4)', ADVANCE="NO") farming_env%group_partition(i)
               END DO
               WRITE (output_unit, *)
               IF (SUM(farming_env%group_partition) .NE. num_slaves) THEN
                  WRITE (output_unit, FMT="(T2,A,T61,I10,T71,I10)") &
                     "FARMING| WARNING : group partition CPUs not equal to the available number (ignoring master) ", &
                     num_slaves, SUM(farming_env%group_partition)
                  WRITE (output_unit, FMT="(T2,A)") "FARMING|          partition data ignored" ! any better idea ??
                  DEALLOCATE (farming_env%group_partition)
               ENDIF
            ENDIF
         ENDIF
         IF (farming_env%group_size_wish_set) THEN
            WRITE (output_unit, FMT="(T2,A,T71,I10)") "FARMING| Group size wish:", &
               farming_env%group_size_wish
         ENDIF
         WRITE (output_unit, FMT="(T2,A,T71,I10)") "FARMING| Max steps      :", farming_env%max_steps
         IF (farming_env%cycle) THEN
            text = "YES"
         ELSE
            text = " NO"
         ENDIF
         WRITE (output_unit, FMT="(T2,A,T78,A3)") "FARMING| Cyclic jobs execution:", text
         IF (farming_env%restart) THEN
            text = "YES"
         ELSE
            text = " NO"
         ENDIF
         WRITE (output_unit, FMT="(T2,A,T78,A3)") "FARMING| Restarting farm:", text
         farming_env%restart_n = 1
         IF (farming_env%restart) THEN
            iunit = get_unit_number()
            OPEN (UNIT=iunit, FILE=farming_env%restart_file_name, IOSTAT=stat)
            IF (stat == 0) THEN
               READ (UNIT=iunit, FMT=*, IOSTAT=stat) farming_env%restart_n
               IF (stat /= 0) THEN
                  WRITE (output_unit, "(T2,A)") &
                     "FARMING| ---- WARNING ---- failed to read from ("//TRIM(farming_env%restart_file_name)//") starting at 1"
               ELSE
                  WRITE (output_unit, "(T2,A)") &
                     "FARMING| restarting from ("//TRIM(farming_env%restart_file_name)//")"
                  WRITE (output_unit, "(T2,A,T71,I10)") &
                     "FARMING| restarting at ", farming_env%restart_n
               ENDIF
            ELSE
               WRITE (output_unit, "(T2,A)") &
                  "FARMING| ---- WARNING ---- failed to open ("//TRIM(farming_env%restart_file_name)//"), starting at 1"
            ENDIF
            CLOSE (iunit, IOSTAT=stat)
         ENDIF

         CALL cp_print_key_finished_output(output_unit, logger, farming_section, &
                                           "PROGRAM_RUN_INFO")
      ENDIF
      CALL mp_bcast(farming_env%restart_n, para_env%source, para_env%group)

   END SUBROUTINE

END MODULE farming_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Debug energy and derivatives w.r.t. finite differences
!> \note
!>      Use INTERPOLATION USE_GUESS, in order to perform force and energy
!>      calculations with the same density. This is not compulsory when iterating
!>      to selfconsistency, but essential in the non-selfconsistent case [08.2005,TdK].
!> \par History
!>      12.2004 created [tlaino]
!>      08.2005 consistent_energies option added, to allow FD calculations
!>              with the correct energies in the non-selfconsistent case, but
!>              keep in mind, that the QS energies and forces are then NOT
!>              consistent to each other [TdK].
!>      08.2005 In case the Harris functional is used, consistent_energies is
!>              et to .FALSE., otherwise the QS energies are spuriously used [TdK].
!>      01.2015 Remove Harris functional option
!>      - Revised (20.11.2013,MK)
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp2k_debug
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force,&
                                              force_env_calc_num_pressure
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              use_qs_force
   USE input_constants,                 ONLY: do_stress_analytical,&
                                              do_stress_diagonal_anal,&
                                              do_stress_diagonal_numer,&
                                              do_stress_numerical
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE particle_methods,                ONLY: write_fist_particle_coordinates,&
                                              write_qs_particle_coordinates
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE virial_types,                    ONLY: cp_virial,&
                                              virial_create,&
                                              virial_release,&
                                              virial_set,&
                                              virial_type,&
                                              zero_virial
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp2k_debug'
   REAL(KIND=dp), PRIVATE, PARAMETER :: maxerr = 5.0_dp

   PUBLIC :: cp2k_debug_energy_and_forces

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
! **************************************************************************************************
   SUBROUTINE cp2k_debug_energy_and_forces(force_env)

      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_debug_energy_and_forces', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=3*default_string_length)             :: message
      INTEGER                                            :: i, ip, iw, j, k, np, stress_tensor
      LOGICAL                                            :: check_failed, debug_forces, &
                                                            debug_stress_tensor, skip, &
                                                            stop_on_mismatch
      REAL(KIND=dp)                                      :: dx, eps_no_error_check, std_value
      REAL(KIND=dp), DIMENSION(2)                        :: numer_energy
      REAL(KIND=dp), DIMENSION(3)                        :: err, my_maxerr
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: analyt_forces, numer_forces
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: root_section, subsys_section
      TYPE(virial_type), POINTER                         :: virial, virial_analytical, &
                                                            virial_numerical

      NULLIFY (analyt_forces, numer_forces, subsys, particles)

      root_section => force_env%root_section

      CALL force_env_get(force_env, para_env=para_env, subsys=subsys, cell=cell)
      subsys_section => section_vals_get_subs_vals(force_env%force_env_section, &
                                                   "SUBSYS")

      CALL section_vals_val_get(root_section, "DEBUG%DEBUG_STRESS_TENSOR", &
                                l_val=debug_stress_tensor)
      CALL section_vals_val_get(root_section, "DEBUG%DEBUG_FORCES", &
                                l_val=debug_forces)
      CALL section_vals_val_get(root_section, "DEBUG%DX", &
                                r_val=dx)
      dx = ABS(dx)
      CALL section_vals_val_get(root_section, "DEBUG%EPS_NO_ERROR_CHECK", &
                                r_val=eps_no_error_check)
      eps_no_error_check = MAX(eps_no_error_check, EPSILON(0.0_dp))
      CALL section_vals_val_get(root_section, "DEBUG%STOP_ON_MISMATCH", &
                                l_val=stop_on_mismatch)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, root_section, "DEBUG%PROGRAM_RUN_INFO", &
                                extension=".log")
      IF (debug_stress_tensor) THEN
         ! To debug stress tensor the stress tensor calculation must be
         ! first enabled..
         CALL section_vals_val_get(force_env%force_env_section, "STRESS_TENSOR", &
                                   i_val=stress_tensor)
         skip = .FALSE.
         SELECT CASE (stress_tensor)
         CASE (do_stress_analytical, do_stress_diagonal_anal)
            ! OK
         CASE (do_stress_numerical, do_stress_diagonal_numer)
            ! Nothing to check
            CALL cp_warn(cp__l("cp2k_debug.F",134), "Numerical stress tensor was requested in "// &
                         "the FORCE_EVAL section. Nothing to debug!")
            skip = .TRUE.
         CASE DEFAULT
            CALL cp_warn(cp__l("cp2k_debug.F",138), "Stress tensor calculation was not enabled in "// &
                         "the FORCE_EVAL section. Nothing to debug!")
            skip = .TRUE.
         END SELECT

         IF (.NOT. skip) THEN

            ! Compute the analytical stress tensor
            CALL cp_subsys_get(subsys, virial=virial)
            CALL virial_set(virial, pv_numer=.FALSE.)
            CALL force_env_calc_energy_force(force_env, &
                                             calc_force=.TRUE., &
                                             calc_stress_tensor=.TRUE.)

            ! Retrieve the analytical virial
            CALL virial_create(virial_analytical)
            CALL zero_virial(virial_analytical)
            CALL cp_virial(virial, virial_analytical)

            ! Debug stress tensor (numerical vs analytical)
            CALL virial_set(virial, pv_numer=.TRUE.)
            CALL force_env_calc_num_pressure(force_env, dx=dx)

            ! Retrieve the numerical virial
            CALL cp_subsys_get(subsys, virial=virial)
            CALL virial_create(virial_numerical)
            CALL zero_virial(virial_numerical)
            CALL cp_virial(virial, virial_numerical)

            ! Print results
            IF (iw > 0) THEN
               WRITE (UNIT=iw, FMT="((T2,A))") &
                  "DEBUG| Numerical pv_virial"
               WRITE (UNIT=iw, FMT="((T2,A,3F16.10))") &
                  ("DEBUG|", virial_numerical%pv_virial(i, 1:3), i=1, 3)
               WRITE (UNIT=iw, FMT="(/,(T2,A))") &
                  "DEBUG| Analytical pv_virial"
               WRITE (UNIT=iw, FMT="((T2,A,3F16.10))") &
                  ("DEBUG|", virial_analytical%pv_virial(i, 1:3), i=1, 3)
               WRITE (UNIT=iw, FMT="(/,(T2,A))") &
                  "DEBUG| Difference pv_virial"
               WRITE (UNIT=iw, FMT="((T2,A,3F16.10))") &
                  ("DEBUG|", virial_numerical%pv_virial(i, 1:3)-virial_analytical%pv_virial(i, 1:3), i=1, 3)
               WRITE (UNIT=iw, FMT="(T2,A,T40,F16.10,/)") &
                  "DEBUG| Sum of differences:", &
                  SUM(ABS(virial_numerical%pv_virial(:, :)-virial_analytical%pv_virial(:, :)))
            END IF

            ! Check and abort on failure
            check_failed = .FALSE.
            DO i = 1, 3
               err(:) = 0.0_dp
               DO k = 1, 3
                  IF (ABS(virial_analytical%pv_virial(i, k)) >= eps_no_error_check) THEN
                     err(k) = 100.0_dp*(virial_numerical%pv_virial(i, k)-virial_analytical%pv_virial(i, k))/ &
                              virial_analytical%pv_virial(i, k)
                  END IF
               END DO
               IF (ANY(ABS(err(1:3)) > maxerr)) check_failed = .TRUE.
            END DO
            CALL virial_release(virial_analytical)
            CALL virial_release(virial_numerical)
            IF (check_failed) THEN
               message = "A mismatch between the analytical and the numerical "// &
                         "stress tensor has been detected. Check the implementation "// &
                         "of the stress tensor"
               IF (stop_on_mismatch) THEN
                  CALL cp__b("cp2k_debug.F",205,TRIM(message))
               ELSE
                  CALL cp__w("cp2k_debug.F",207,TRIM(message))
               END IF
            END IF
         END IF
      END IF

      IF (debug_forces) THEN
         ! Debug forces (numerical vs analytical)
         particles => subsys%particles%els
         SELECT CASE (force_env%in_use)
         CASE (use_qs_force)
            CALL get_qs_env(force_env%qs_env, qs_kind_set=qs_kind_set)
            CALL write_qs_particle_coordinates(particles, qs_kind_set, subsys_section, "DEBUG")
         CASE DEFAULT
            CALL write_fist_particle_coordinates(particles, subsys_section, charges=Null())
         END SELECT
         ! First evaluate energy and forces
         CALL force_env_calc_energy_force(force_env, &
                                          calc_force=.TRUE., &
                                          calc_stress_tensor=.FALSE.)
         ! Copy forces in array and start the numerical calculation
         IF (ASSOCIATED(analyt_forces)) DEALLOCATE (analyt_forces)
         np = subsys%particles%n_els
         ALLOCATE (analyt_forces(np, 3))
         DO ip = 1, np
            analyt_forces(ip, 1:3) = particles(ip)%f(1:3)
         END DO
         ! Loop on atoms and coordinates
         IF (ASSOCIATED(numer_forces)) DEALLOCATE (numer_forces)
         ALLOCATE (numer_forces(subsys%particles%n_els, 3))
         Atom: DO ip = 1, np
            Coord: DO k = 1, 3
               numer_energy = 0.0_dp
               std_value = particles(ip)%r(k)
               DO j = 1, 2
                  particles(ip)%r(k) = std_value-(-1.0_dp)**j*dx
                  SELECT CASE (force_env%in_use)
                  CASE (use_qs_force)
                     CALL get_qs_env(force_env%qs_env, qs_kind_set=qs_kind_set)
                     CALL write_qs_particle_coordinates(particles, qs_kind_set, subsys_section, "DEBUG")
                  CASE DEFAULT
                     CALL write_fist_particle_coordinates(particles, subsys_section, charges=Null())
                  END SELECT
                  ! Compute energy
                  CALL force_env_calc_energy_force(force_env, &
                                                   calc_force=.FALSE., &
                                                   calc_stress_tensor=.FALSE., &
                                                   consistent_energies=.TRUE.)
                  CALL force_env_get(force_env, potential_energy=numer_energy(j))
               END DO
               particles(ip)%r(k) = std_value
               numer_forces(ip, k) = -0.5_dp*(numer_energy(1)-numer_energy(2))/dx
               IF (iw > 0) THEN
                  WRITE (UNIT=iw, FMT="(T2,A,T17,A,F7.4,A,T34,A,F7.4,A,T52,A,T68,A)") &
                     "DEBUG| Atom", "E("//ACHAR(119+k)//" +", dx, ")", &
                     "E("//ACHAR(119+k)//" -", dx, ")", &
                     "f(numerical)", "f(analytical)"
                  WRITE (UNIT=iw, FMT="(T2,A,I5,4(1X,F16.8))") &
                     "DEBUG|", ip, numer_energy(1:2), numer_forces(ip, k), analyt_forces(ip, k)
               END IF
            END DO Coord
            ! Check analytical forces using the numerical forces as reference
            my_maxerr = maxerr
            err(1:3) = 0.0_dp
            DO k = 1, 3
               ! Calculate percentage but ignore very small force values
               IF (ABS(analyt_forces(ip, k)) >= eps_no_error_check) THEN
                  err(k) = 100.0_dp*(numer_forces(ip, k)-analyt_forces(ip, k))/analyt_forces(ip, k)
               END IF
               ! Increase error tolerance for small force values
               IF (ABS(analyt_forces(ip, k)) <= 0.0001_dp) my_maxerr(k) = 5.0_dp*my_maxerr(k)
            END DO
            IF (iw > 0) THEN
               WRITE (UNIT=iw, FMT="(/,T2,A)") &
                  "DEBUG| Atom  Coordinate   f(numerical)   f(analytical)   Difference   Error [%]"
               DO k = 1, 3
                  WRITE (UNIT=iw, FMT="(T2,A,I5,T19,A1,T26,F14.8,T42,F14.8,T57,F12.8,T71,F10.2)") &
                     "DEBUG|", ip, ACHAR(119+k), numer_forces(ip, k), analyt_forces(ip, k), &
                     analyt_forces(ip, k)-numer_forces(ip, k), err(k)
               END DO
            END IF
            IF (ANY(ABS(err(1:3)) > my_maxerr(1:3))) THEN
               message = "A mismatch between analytical and numerical forces "// &
                         "has been detected. Check the implementation of the "// &
                         "analytical force calculation"
               IF (stop_on_mismatch) THEN
                  CALL cp__b("cp2k_debug.F",293,message)
               ELSE
                  CALL cp__w("cp2k_debug.F",295,message)
               END IF
            END IF
         END DO Atom
         ! Print summary
         IF (iw > 0) THEN
            WRITE (UNIT=iw, FMT="(/,(T2,A))") &
               "DEBUG|======================== BEGIN OF SUMMARY ===============================", &
               "DEBUG| Atom  Coordinate   f(numerical)   f(analytical)   Difference   Error [%]"
            DO ip = 1, np
               err(1:3) = 0.0_dp
               DO k = 1, 3
                  IF (ABS(analyt_forces(ip, k)) >= eps_no_error_check) THEN
                     err(k) = 100_dp*(numer_forces(ip, k)-analyt_forces(ip, k))/analyt_forces(ip, k)
                     WRITE (UNIT=iw, FMT="(T2,A,I5,T19,A1,T26,F14.8,T42,F14.8,T57,F12.8,T71,F10.2)") &
                        "DEBUG|", ip, ACHAR(119+k), numer_forces(ip, k), analyt_forces(ip, k), &
                        analyt_forces(ip, k)-numer_forces(ip, k), err(k)
                  ELSE
                     WRITE (UNIT=iw, FMT="(T2,A,I5,T19,A1,T26,F14.8,T42,F14.8,T57,F12.8)") &
                        "DEBUG|", ip, ACHAR(119+k), numer_forces(ip, k), analyt_forces(ip, k), &
                        analyt_forces(ip, k)-numer_forces(ip, k)
                  END IF
               END DO
            END DO
            WRITE (UNIT=iw, FMT="(T2,A)") &
               "DEBUG|======================== END OF SUMMARY ================================="
         END IF
         ! Release work storage
         IF (ASSOCIATED(analyt_forces)) DEALLOCATE (analyt_forces)
         IF (ASSOCIATED(numer_forces)) DEALLOCATE (numer_forces)
      END IF
      CALL cp_print_key_finished_output(iw, logger, root_section, &
                                        "DEBUG%PROGRAM_RUN_INFO")

   END SUBROUTINE cp2k_debug_energy_and_forces

END MODULE cp2k_debug
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Driver mode - To communicate with i-PI Python wrapper
!> \par History
!>      none
!> \author Michele Ceriotti 03.2012
! **************************************************************************************************
MODULE ipi_driver
   USE ISO_C_BINDING,                   ONLY: C_CHAR,&
                                              C_DOUBLE,&
                                              C_INT,&
                                              C_LOC,&
                                              C_NULL_CHAR,&
                                              C_PTR
   USE bibliography,                    ONLY: Ceriotti2014,&
                                              Kapil2016,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_create,&
                                              cell_type,&
                                              init_cell
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              int_4
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_irecv,&
                                              mp_send,&
                                              mp_sync,&
                                              mp_testany
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ipi_driver'

   PUBLIC :: run_driver


CONTAINS


! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \param globenv ...
!> \par History
!>       12.2013 included in repository
!> \author Ceriotti
! **************************************************************************************************

   SUBROUTINE run_driver(force_env, globenv)
      TYPE(force_env_type), POINTER            :: force_env
      TYPE(global_environment_type), POINTER   :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'run_driver', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle
      CALL timeset(routineN, handle)
      CALL cp__b("ipi_driver.F",315,"CP2K was compiled with the __NO_IPI_DRIVER option!")
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(globenv))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(force_env))==-1) EXIT ;  END DO ; ENDIF

      CALL timestop(handle)

   END SUBROUTINE run_driver
END MODULE ipi_driver
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief History of minima, calulates, stores and compares fingerprints of minima.
!>        Used by Minima Hopping and Minima Crawling.
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_history
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   TYPE history_fingerprint_type
      PRIVATE
      REAL(KIND=dp)                            :: Epot = 0.0
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: goedecker
   END TYPE history_fingerprint_type

   TYPE history_entry_type
      TYPE(history_fingerprint_type), POINTER :: p => Null()
      INTEGER                                 :: id = -1
   END TYPE history_entry_type

   TYPE history_type
      PRIVATE
      TYPE(history_entry_type), DIMENSION(:), POINTER :: entries => Null()
      INTEGER                              :: length = 0
      INTEGER                              :: iw = -1
      REAL(KIND=dp)                        :: E_precision = 0.0
      REAL(KIND=dp)                        :: FP_precision = 0.0
   END TYPE history_type

   PUBLIC :: history_type, history_fingerprint_type
   PUBLIC :: history_init, history_finalize
   PUBLIC :: history_add, history_lookup
   PUBLIC :: history_fingerprint
   PUBLIC :: history_fingerprint_match

   LOGICAL, PARAMETER                     :: debug = .FALSE.
   INTEGER, PARAMETER                     :: history_grow_unit = 1000
CONTAINS

! **************************************************************************************************
!> \brief Initializes a history.
!> \param history ...
!> \param history_section ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE history_init(history, history_section, iw)
      TYPE(history_type), INTENT(INOUT)                  :: history
      TYPE(section_vals_type), POINTER                   :: history_section
      INTEGER                                            :: iw

      ALLOCATE (history%entries(history_grow_unit))
      history%iw = iw
      CALL section_vals_val_get(history_section, "ENERGY_PRECISION", &
                                r_val=history%E_precision)
      CALL section_vals_val_get(history_section, "FINGERPRINT_PRECISION", &
                                r_val=history%FP_precision)

      IF (iw > 0) THEN
         WRITE (iw, '(A,T66,E15.3)') &
            " GLBOPT| History energy precision", history%E_precision
         WRITE (iw, '(A,T66,E15.3)') &
            " GLBOPT| History fingerprint precision", history%FP_precision
      END IF
   END SUBROUTINE history_init

! **************************************************************************************************
!> \brief Calculates a fingerprint for a given configuration.
!> \param Epot ...
!> \param pos ...
!> \retval fp ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION history_fingerprint(Epot, pos) RESULT(fp)
      REAL(KIND=dp), INTENT(IN)                          :: Epot
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pos
      TYPE(history_fingerprint_type)                     :: fp

      INTEGER                                            :: handle
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp

      CALL timeset("glbopt_history_fingerprint", handle)

      NULLIFY (tmp)
      fp%Epot = Epot
      CALL goedecker_fingerprint(pos, tmp)

      !copy pointer to allocatable
      ALLOCATE (fp%goedecker(SIZE(tmp)))
      fp%goedecker(:) = tmp
      DEALLOCATE (tmp)

      CALL timestop(handle)
   END FUNCTION history_fingerprint

! **************************************************************************************************
!> \brief Helper routine for history_fingerprint.
!>        Calulates a fingerprint based on inter-atomic distances.
!> \param pos ...
!> \param res ...
!> \author Stefan Goedecker
! **************************************************************************************************
   SUBROUTINE goedecker_fingerprint(pos, res)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pos
      REAL(KIND=dp), DIMENSION(:), POINTER               :: res

      INTEGER                                            :: i, info, j, N
      REAL(KIND=dp)                                      :: d2, t
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: matrix, work
      REAL(KIND=dp), DIMENSION(3)                        :: d

      IF (ASSOCIATED(res)) CALL cp__b("swarm/glbopt_history.F",122,"goedecker_fingerprint: res already allocated")
      N = SIZE(pos)/3 ! number of atoms

      ALLOCATE (matrix(N, N), work(N, N))
      DO i = 1, N
         matrix(i, i) = 1.0
         DO j = i+1, N
            d = pos(3*i-2:3*i)-pos(3*j-2:3*j)
            d2 = SUM(d**2)
            t = EXP(-0.5*d2)
            matrix(i, j) = t
            matrix(j, i) = t
         ENDDO
      ENDDO
      !TODO: call dsyv through cp2k wrappers
      !TODO: do we need to store lower triangle of matrix?
      ALLOCATE (res(N))
      CALL DSYEV('N', 'U', N, matrix, N, res, work, N**2, info)
      IF (info /= 0) CALL cp__b("swarm/glbopt_history.F",140,"goedecker_fingerprint: DSYEV failed")
   END SUBROUTINE goedecker_fingerprint

! **************************************************************************************************
!> \brief Checks if two given fingerprints match.
!> \param history ...
!> \param fp1 ...
!> \param fp2 ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION history_fingerprint_match(history, fp1, fp2) RESULT(res)
      TYPE(history_type), INTENT(IN)                     :: history
      TYPE(history_fingerprint_type), INTENT(IN)         :: fp1, fp2
      LOGICAL                                            :: res

      res = (ABS(fp1%Epot-fp2%Epot) < history%E_precision) .AND. &
            (fingerprint_distance(fp1, fp2) < history%fp_precision)

   END FUNCTION history_fingerprint_match

! **************************************************************************************************
!> \brief Helper routine for history_fingerprint_match
!>        Calulates the distance between two given fingerprints.
!> \param fp1 ...
!> \param fp2 ...
!> \retval res ...
!> \author Stefan Goedecker
! **************************************************************************************************
   PURE FUNCTION fingerprint_distance(fp1, fp2) RESULT(res)
      TYPE(history_fingerprint_type), INTENT(IN)         :: fp1, fp2
      REAL(KIND=dp)                                      :: res

      res = SQRT(SUM((fp1%goedecker-fp2%goedecker)**2)/SIZE(fp1%goedecker))
   END FUNCTION fingerprint_distance

! **************************************************************************************************
!> \brief Addes a new fingerprints to the history.
!>        Optionally, an abitrary id can be stored alongside the fingerprint.
!> \param history ...
!> \param fingerprint ...
!> \param id ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE history_add(history, fingerprint, id)
      TYPE(history_type), INTENT(INOUT)                  :: history
      TYPE(history_fingerprint_type), INTENT(IN)         :: fingerprint
      INTEGER, INTENT(IN), OPTIONAL                      :: id

      INTEGER                                            :: handle, i, k, n
      TYPE(history_entry_type), DIMENSION(:), POINTER    :: tmp

      CALL timeset("glbopt_history_add", handle)

      n = SIZE(history%entries)
      IF (n == history%length) THEN
         ! grow history%entries array
         tmp => history%entries
         ALLOCATE (history%entries(n+history_grow_unit))
         history%entries(1:n) = tmp(:)
         DEALLOCATE (tmp)
         n = n+history_grow_unit
      ENDIF

      k = interpolation_search(history, fingerprint%Epot)

      !history%entries(k+1:) = history%entries(k:n-1)
      !Workaround for an XLF bug - pointer array copy does
      !not work correctly
      DO i = n, k+1, -1
         history%entries(i) = history%entries(i-1)
      END DO

      ALLOCATE (history%entries(k)%p)
      history%entries(k)%p = fingerprint
      IF (PRESENT(id)) &
         history%entries(k)%id = id
      history%length = history%length+1

      IF (debug) THEN
         ! check history for correct order
         DO k = 1, history%length
            !WRITE(*,*) "history: ", k, "Epot",history%entries(k)%p%Epot
            IF (k > 1) THEN
               IF (history%entries(k-1)%p%Epot > history%entries(k)%p%Epot) &
                  CALL cp__b("swarm/glbopt_history.F",225,"history_add: history in wrong order")
            END IF
         END DO
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE history_add

! **************************************************************************************************
!> \brief Checks if a given fingerprints is contained in the history.
!> \param history ...
!> \param fingerprint ...
!> \param found ...
!> \param id ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE history_lookup(history, fingerprint, found, id)
      TYPE(history_type), INTENT(IN)                     :: history
      TYPE(history_fingerprint_type), INTENT(IN)         :: fingerprint
      LOGICAL, INTENT(OUT)                               :: found
      INTEGER, INTENT(OUT), OPTIONAL                     :: id

      INTEGER                                            :: found_i, handle, i, k, k_max, k_min
      REAL(KIND=dp)                                      :: best_match, dist, Epot

      CALL timeset("glbopt_history_lookup", handle)

      found = .FALSE.
      IF (PRESENT(id)) id = -1
      best_match = HUGE(1.0_dp)

      IF (history%length > 0) THEN
         Epot = fingerprint%Epot
         k = interpolation_search(history, fingerprint%Epot)

         DO k_min = k-1, 1, -1
            IF (history%entries(k_min)%p%Epot < Epot-history%E_precision) EXIT
         ENDDO

         DO k_max = k, history%length
            IF (history%entries(k_max)%p%Epot > Epot+history%E_precision) EXIT
         ENDDO

         k_min = MAX(k_min+1, 1)
         k_max = MIN(k_max-1, history%length)

         IF (debug) found_i = -1

         DO i = k_min, k_max
            dist = fingerprint_distance(fingerprint, history%entries(i)%p)
            !WRITE(*,*) "entry ", i, " dist: ",dist
            IF (dist < history%fp_precision .AND. dist < best_match) THEN
               best_match = dist
               found = .TRUE.
               IF (PRESENT(id)) id = history%entries(i)%id
               IF (debug) found_i = i
            END IF
         ENDDO

         IF (debug) CALL verify_history_lookup(history, fingerprint, found_i)
      END IF

      CALL timestop(handle)

   END SUBROUTINE history_lookup

! **************************************************************************************************
!> \brief Helper routine for history_lookup
!> \param history ...
!> \param Efind ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION interpolation_search(history, Efind) RESULT(res)
      TYPE(history_type), INTENT(IN)                     :: history
      REAL(KIND=dp), INTENT(IN)                          :: Efind
      INTEGER                                            :: res

      INTEGER                                            :: high, low, mid
      REAL(KIND=dp)                                      :: slope

      low = 1
      high = history%length

      DO WHILE (low < high)
         !linear interpolation
         slope = REAL(high-low)/(history%entries(high)%p%Epot-history%entries(low)%p%Epot)
         mid = low+INT(slope*(Efind-history%entries(low)%p%Epot))
         mid = MIN(MAX(mid, low), high)

         IF (history%entries(mid)%p%Epot < Efind) THEN
            low = mid+1
         ELSE
            high = mid-1
         END IF
      END DO

      IF (0 < low .AND. low <= history%length) THEN
         IF (Efind > history%entries(low)%p%Epot) low = low+1
      END IF

      res = low
   END FUNCTION interpolation_search

! **************************************************************************************************
!> \brief Debugging routine, performs a slow (but robust) linear search.
!> \param history ...
!> \param fingerprint ...
!> \param found_i_ref ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE verify_history_lookup(history, fingerprint, found_i_ref)
      TYPE(history_type), INTENT(IN)                     :: history
      TYPE(history_fingerprint_type), INTENT(IN)         :: fingerprint
      INTEGER, INTENT(IN)                                :: found_i_ref

      INTEGER                                            :: found_i, i
      REAL(KIND=dp)                                      :: best_fp_match, Epot_dist, fp_dist

      found_i = -1
      best_fp_match = HUGE(1.0_dp)

      DO i = 1, history%length
         Epot_dist = ABS(fingerprint%Epot-history%entries(i)%p%Epot)
         IF (Epot_dist > history%E_precision) CYCLE
         fp_dist = fingerprint_distance(fingerprint, history%entries(i)%p)
         !WRITE(*,*) "entry ", i, " dist: ",dist
         IF (fp_dist < history%fp_precision .AND. fp_dist < best_fp_match) THEN
            best_fp_match = fp_dist
            found_i = i
         END IF
      ENDDO

      IF (found_i /= found_i_ref) THEN
         WRITE (*, *) found_i, found_i_ref
         CALL cp__b("swarm/glbopt_history.F",360,"verify_history_lookup failed")
      ENDIF

   END SUBROUTINE verify_history_lookup

! **************************************************************************************************
!> \brief Finalizes a history.
!> \param history ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE history_finalize(history)
      TYPE(history_type)                                 :: history

      INTEGER                                            :: i

      DO i = 1, history%length
         IF (ASSOCIATED(history%entries(i)%p)) &
            DEALLOCATE (history%entries(i)%p)
      END DO

      DEALLOCATE (history%entries)

   END SUBROUTINE history_finalize

END MODULE glbopt_history
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Swarm-message, a convenient data-container for with build-in serialization.
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm_message

   USE cp_parser_methods,               ONLY: parser_get_next_line
   USE cp_parser_types,                 ONLY: cp_parser_type
   USE kinds,                           ONLY: default_string_length,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_environ,&
                                              mp_recv,&
                                              mp_send
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm_message'

   TYPE swarm_message_type
      PRIVATE
      TYPE(message_entry_type), POINTER :: root => Null()
   END TYPE swarm_message_type

   INTEGER, PARAMETER  :: key_length = 20

   TYPE message_entry_type
      CHARACTER(LEN=key_length)                        :: key
      TYPE(message_entry_type), POINTER     :: next => Null()
      CHARACTER(LEN=default_string_length), POINTER     :: value_str => Null()
      INTEGER(KIND=int_4), POINTER     :: value_i4 => Null()
      INTEGER(KIND=int_8), POINTER     :: value_i8 => Null()
      REAL(KIND=real_4), POINTER     :: value_r4 => Null()
      REAL(KIND=real_8), POINTER     :: value_r8 => Null()
      INTEGER(KIND=int_4), DIMENSION(:), POINTER     :: value_1d_i4 => Null()
      INTEGER(KIND=int_8), DIMENSION(:), POINTER     :: value_1d_i8 => Null()
      REAL(KIND=real_4), DIMENSION(:), POINTER     :: value_1d_r4 => Null()
      REAL(KIND=real_8), DIMENSION(:), POINTER     :: value_1d_r8 => Null()
   END TYPE message_entry_type

! **************************************************************************************************
!> \brief Adds an entry from a swarm-message.
!> \author Ole Schuett
! **************************************************************************************************
   INTERFACE swarm_message_add
      MODULE PROCEDURE swarm_message_add_str
      MODULE PROCEDURE swarm_message_add_i4, swarm_message_add_i8
      MODULE PROCEDURE swarm_message_add_r4, swarm_message_add_r8
      MODULE PROCEDURE swarm_message_add_1d_i4, swarm_message_add_1d_i8
      MODULE PROCEDURE swarm_message_add_1d_r4, swarm_message_add_1d_r8
   END INTERFACE swarm_message_add

! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \author Ole Schuett
! **************************************************************************************************
   INTERFACE swarm_message_get
      MODULE PROCEDURE swarm_message_get_str
      MODULE PROCEDURE swarm_message_get_i4, swarm_message_get_i8
      MODULE PROCEDURE swarm_message_get_r4, swarm_message_get_r8
      MODULE PROCEDURE swarm_message_get_1d_i4, swarm_message_get_1d_i8
      MODULE PROCEDURE swarm_message_get_1d_r4, swarm_message_get_1d_r8
   END INTERFACE swarm_message_get

   PUBLIC :: swarm_message_type, swarm_message_add, swarm_message_get
   PUBLIC :: swarm_message_mpi_send, swarm_message_mpi_recv, swarm_message_mpi_bcast
   PUBLIC :: swarm_message_file_write, swarm_message_file_read
   PUBLIC :: swarm_message_haskey, swarm_message_equal
   PUBLIC :: swarm_message_free

CONTAINS

! **************************************************************************************************
!> \brief Returns the number of entries contained in a swarm-message.
!> \param msg ...
!> \retval l ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION swarm_message_length(msg) RESULT(l)
      TYPE(swarm_message_type), INTENT(IN)               :: msg
      INTEGER                                            :: l

      TYPE(message_entry_type), POINTER                  :: curr_entry

      l = 0
      curr_entry => msg%root
      DO WHILE (ASSOCIATED(curr_entry))
         l = l+1
         curr_entry => curr_entry%next
      END DO
   END FUNCTION swarm_message_length

! **************************************************************************************************
!> \brief Checks if a swarm-message contains an entry with the given key.
!> \param msg ...
!> \param key ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION swarm_message_haskey(msg, key) RESULT(res)
      TYPE(swarm_message_type), INTENT(IN)               :: msg
      CHARACTER(LEN=*), INTENT(IN)                       :: key
      LOGICAL                                            :: res

      TYPE(message_entry_type), POINTER                  :: curr_entry

      res = .FALSE.
      curr_entry => msg%root
      DO WHILE (ASSOCIATED(curr_entry))
         IF (TRIM(curr_entry%key) == TRIM(key)) THEN
            res = .TRUE.
            EXIT
         END IF
         curr_entry => curr_entry%next
      END DO
   END FUNCTION swarm_message_haskey

! **************************************************************************************************
!> \brief Deallocates all entries contained in a swarm-message.
!> \param msg ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_free(msg)
      TYPE(swarm_message_type), INTENT(INOUT)            :: msg

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_free', &
         routineP = moduleN//':'//routineN

      TYPE(message_entry_type), POINTER                  :: ENTRY, old_entry

      ENTRY => msg%root
      DO WHILE (ASSOCIATED(ENTRY))
         IF (ASSOCIATED(entry%value_str)) DEALLOCATE (entry%value_str)
         IF (ASSOCIATED(entry%value_i4)) DEALLOCATE (entry%value_i4)
         IF (ASSOCIATED(entry%value_i8)) DEALLOCATE (entry%value_i8)
         IF (ASSOCIATED(entry%value_r4)) DEALLOCATE (entry%value_r4)
         IF (ASSOCIATED(entry%value_r8)) DEALLOCATE (entry%value_r8)
         IF (ASSOCIATED(entry%value_1d_i4)) DEALLOCATE (entry%value_1d_i4)
         IF (ASSOCIATED(entry%value_1d_i8)) DEALLOCATE (entry%value_1d_i8)
         IF (ASSOCIATED(entry%value_1d_r4)) DEALLOCATE (entry%value_1d_r4)
         IF (ASSOCIATED(entry%value_1d_r8)) DEALLOCATE (entry%value_1d_r8)
         old_entry => ENTRY
         ENTRY => entry%next
         DEALLOCATE (old_entry)
      END DO

      NULLIFY (msg%root)

      IF(.NOT.(swarm_message_length(msg) == 0))CALL cp__a("swarm/swarm_message.F",159)
   END SUBROUTINE swarm_message_free

! **************************************************************************************************
!> \brief Checks if two swarm-messages are equal
!> \param msg1 ...
!> \param msg2 ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION swarm_message_equal(msg1, msg2) RESULT(res)
      TYPE(swarm_message_type), INTENT(IN)               :: msg1, msg2
      LOGICAL                                            :: res

      res = swarm_message_equal_oneway(msg1, msg2) .AND. &
            swarm_message_equal_oneway(msg2, msg1)

   END FUNCTION swarm_message_equal

! **************************************************************************************************
!> \brief Sends a swarm message via MPI.
!> \param msg ...
!> \param group ...
!> \param dest ...
!> \param tag ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_mpi_send(msg, group, dest, tag)
      TYPE(swarm_message_type), INTENT(IN)               :: msg
      INTEGER, INTENT(IN)                                :: group, dest, tag

      TYPE(message_entry_type), POINTER                  :: curr_entry

      CALL mp_send(swarm_message_length(msg), dest, tag, group)
      curr_entry => msg%root
      DO WHILE (ASSOCIATED(curr_entry))
         CALL swarm_message_entry_mpi_send(curr_entry, group, dest, tag)
         curr_entry => curr_entry%next
      END DO
   END SUBROUTINE swarm_message_mpi_send

! **************************************************************************************************
!> \brief Receives a swarm message via MPI.
!> \param msg ...
!> \param group ...
!> \param src ...
!> \param tag ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_mpi_recv(msg, group, src, tag)
      TYPE(swarm_message_type), INTENT(INOUT)            :: msg
      INTEGER, INTENT(IN)                                :: group
      INTEGER, INTENT(INOUT)                             :: src, tag

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_mpi_recv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, length
      TYPE(message_entry_type), POINTER                  :: new_entry

      IF (ASSOCIATED(msg%root)) CALL cp__b("swarm/swarm_message.F",219,"message not empty")
      CALL mp_recv(length, src, tag, group)
      DO i = 1, length
         ALLOCATE (new_entry)
         CALL swarm_message_entry_mpi_recv(new_entry, group, src, tag)
         new_entry%next => msg%root
         msg%root => new_entry
      END DO

   END SUBROUTINE swarm_message_mpi_recv

! **************************************************************************************************
!> \brief Broadcasts a swarm message via MPI.
!> \param msg ...
!> \param src ...
!> \param group ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_mpi_bcast(msg, src, group)
      TYPE(swarm_message_type), INTENT(INOUT)            :: msg
      INTEGER, INTENT(IN)                                :: src, group

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_mpi_bcast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, length, mepos, num_pe
      TYPE(message_entry_type), POINTER                  :: curr_entry

      CALL mp_environ(num_pe, mepos, group)

      IF (mepos /= src .AND. ASSOCIATED(msg%root)) CALL cp__b("swarm/swarm_message.F",249,"message not empty")
      length = swarm_message_length(msg)
      CALL mp_bcast(length, src, group)

      IF (mepos == src) curr_entry => msg%root

      DO i = 1, length
         IF (mepos /= src) ALLOCATE (curr_entry)

         CALL swarm_message_entry_mpi_bcast(curr_entry, src, group, mepos)

         IF (mepos == src) THEN
            curr_entry => curr_entry%next
         ELSE
            curr_entry%next => msg%root
            msg%root => curr_entry
         END IF
      END DO

   END SUBROUTINE swarm_message_mpi_bcast

! **************************************************************************************************
!> \brief Write a swarm-message to a given file / unit.
!> \param msg ...
!> \param unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_file_write(msg, unit)
      TYPE(swarm_message_type), INTENT(IN)               :: msg
      INTEGER, INTENT(IN)                                :: unit

      INTEGER                                            :: handle
      TYPE(message_entry_type), POINTER                  :: curr_entry

      IF (unit <= 0) RETURN

      CALL timeset("swarm_message_file_write", handle)
      WRITE (unit, "(A)") "BEGIN SWARM_MESSAGE"
      WRITE (unit, "(A,I10)") "msg_length: ", swarm_message_length(msg)

      curr_entry => msg%root
      DO WHILE (ASSOCIATED(curr_entry))
         CALL swarm_message_entry_file_write(curr_entry, unit)
         curr_entry => curr_entry%next
      END DO

      WRITE (unit, "(A)") "END SWARM_MESSAGE"
      WRITE (unit, "()")
      CALL timestop(handle)
   END SUBROUTINE swarm_message_file_write

! **************************************************************************************************
!> \brief Reads a swarm-message from a given file / unit.
!> \param msg ...
!> \param parser ...
!> \param at_end ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_file_read(msg, parser, at_end)
      TYPE(swarm_message_type), INTENT(OUT)              :: msg
      TYPE(cp_parser_type), POINTER                      :: parser
      LOGICAL, INTENT(INOUT)                             :: at_end

      INTEGER                                            :: handle

      CALL timeset("swarm_message_file_read", handle)
      CALL swarm_message_file_read_low(msg, parser, at_end)
      CALL timestop(handle)
   END SUBROUTINE swarm_message_file_read

! **************************************************************************************************
!> \brief Helper routine, does the actual work of swarm_message_file_read().
!> \param msg ...
!> \param parser ...
!> \param at_end ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_file_read_low(msg, parser, at_end)
      TYPE(swarm_message_type), INTENT(OUT)              :: msg
      TYPE(cp_parser_type), POINTER                      :: parser
      LOGICAL, INTENT(INOUT)                             :: at_end

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_file_read_low', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: label
      INTEGER                                            :: i, length
      TYPE(message_entry_type), POINTER                  :: new_entry

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN
      IF(.NOT.(TRIM(parser%input_line(1:20)) == "BEGIN SWARM_MESSAGE"))CALL cp__a("swarm/swarm_message.F",341)

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN
      READ (parser%input_line(1:40), *) label, length
      IF(.NOT.(TRIM(label) == "msg_length:"))CALL cp__a("swarm/swarm_message.F",347)

      DO i = 1, length
         ALLOCATE (new_entry)
         CALL swarm_message_entry_file_read(new_entry, parser, at_end)
         new_entry%next => msg%root
         msg%root => new_entry
      END DO

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN
      IF(.NOT.(TRIM(parser%input_line(1:20)) == "END SWARM_MESSAGE"))CALL cp__a("swarm/swarm_message.F",359)

   END SUBROUTINE swarm_message_file_read_low

! **************************************************************************************************
!> \brief Helper routine for swarm_message_equal
!> \param msg1 ...
!> \param msg2 ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION swarm_message_equal_oneway(msg1, msg2) RESULT(res)
      TYPE(swarm_message_type), INTENT(IN)               :: msg1, msg2
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_equal_oneway', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: found
      TYPE(message_entry_type), POINTER                  :: entry1, entry2

      res = .FALSE.

      !loop over entries of msg1
      entry1 => msg1%root
      DO WHILE (ASSOCIATED(entry1))

         ! finding matching entry in msg2
         entry2 => msg2%root
         found = .FALSE.
         DO WHILE (ASSOCIATED(entry2))
            IF (TRIM(entry2%key) == TRIM(entry1%key)) THEN
               found = .TRUE.
               EXIT
            END IF
            entry2 => entry2%next
         END DO
         IF (.NOT. found) RETURN

         !compare the two entries
         IF (ASSOCIATED(entry1%value_str)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_str)) RETURN
            IF (TRIM(entry1%value_str) /= TRIM(entry2%value_str)) RETURN

         ELSE IF (ASSOCIATED(entry1%value_i4)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_i4)) RETURN
            IF (entry1%value_i4 /= entry2%value_i4) RETURN

         ELSE IF (ASSOCIATED(entry1%value_i8)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_i8)) RETURN
            IF (entry1%value_i8 /= entry2%value_i8) RETURN

         ELSE IF (ASSOCIATED(entry1%value_r4)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_r4)) RETURN
            IF (ABS(entry1%value_r4-entry2%value_r4) > 1e-5) RETURN

         ELSE IF (ASSOCIATED(entry1%value_r8)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_r8)) RETURN
            IF (ABS(entry1%value_r8-entry2%value_r8) > 1e-10) RETURN

         ELSE IF (ASSOCIATED(entry1%value_1d_i4)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_1d_i4)) RETURN
            IF (ANY(entry1%value_1d_i4 /= entry2%value_1d_i4)) RETURN

         ELSE IF (ASSOCIATED(entry1%value_1d_i8)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_1d_i8)) RETURN
            IF (ANY(entry1%value_1d_i8 /= entry2%value_1d_i8)) RETURN

         ELSE IF (ASSOCIATED(entry1%value_1d_r4)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_1d_r4)) RETURN
            IF (ANY(ABS(entry1%value_1d_r4-entry2%value_1d_r4) > 1e-5)) RETURN

         ELSE IF (ASSOCIATED(entry1%value_1d_r8)) THEN
            IF (.NOT. ASSOCIATED(entry2%value_1d_r8)) RETURN
            IF (ANY(ABS(entry1%value_1d_r8-entry2%value_1d_r8) > 1e-10)) RETURN
         ELSE
            CALL cp__b("swarm/swarm_message.F",435,"no value ASSOCIATED")
         END IF

         entry1 => entry1%next
      END DO

      ! if we reach this point no differences were found
      res = .TRUE.
   END FUNCTION swarm_message_equal_oneway

! **************************************************************************************************
!> \brief Helper routine for swarm_message_mpi_send.
!> \param ENTRY ...
!> \param group ...
!> \param dest ...
!> \param tag ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_entry_mpi_send(ENTRY, group, dest, tag)
      TYPE(message_entry_type), INTENT(IN)               :: ENTRY
      INTEGER, INTENT(IN)                                :: group, dest, tag

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_entry_mpi_send', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(default_string_length)          :: value_str_arr
      INTEGER, DIMENSION(key_length)                     :: key_arr

      key_arr = str2iarr(entry%key)
      CALL mp_send(key_arr, dest, tag, group)

      IF (ASSOCIATED(entry%value_i4)) THEN
         CALL mp_send(1, dest, tag, group)
         CALL mp_send(entry%value_i4, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_i8)) THEN
         CALL mp_send(2, dest, tag, group)
         CALL mp_send(entry%value_i8, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_r4)) THEN
         CALL mp_send(3, dest, tag, group)
         CALL mp_send(entry%value_r4, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_r8)) THEN
         CALL mp_send(4, dest, tag, group)
         CALL mp_send(entry%value_r8, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_1d_i4)) THEN
         CALL mp_send(5, dest, tag, group)
         CALL mp_send(SIZE(entry%value_1d_i4), dest, tag, group)
         CALL mp_send(entry%value_1d_i4, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_1d_i8)) THEN
         CALL mp_send(6, dest, tag, group)
         CALL mp_send(SIZE(entry%value_1d_i8), dest, tag, group)
         CALL mp_send(entry%value_1d_i8, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_1d_r4)) THEN
         CALL mp_send(7, dest, tag, group)
         CALL mp_send(SIZE(entry%value_1d_r4), dest, tag, group)
         CALL mp_send(entry%value_1d_r4, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_1d_r8)) THEN
         CALL mp_send(8, dest, tag, group)
         CALL mp_send(SIZE(entry%value_1d_r8), dest, tag, group)
         CALL mp_send(entry%value_1d_r8, dest, tag, group)

      ELSE IF (ASSOCIATED(entry%value_str)) THEN
         CALL mp_send(9, dest, tag, group)
         value_str_arr = str2iarr(entry%value_str)
         CALL mp_send(value_str_arr, dest, tag, group)
      ELSE
         CALL cp__b("swarm/swarm_message.F",507,"no value ASSOCIATED")
      END IF
   END SUBROUTINE swarm_message_entry_mpi_send

! **************************************************************************************************
!> \brief Helper routine for swarm_message_mpi_recv.
!> \param ENTRY ...
!> \param group ...
!> \param src ...
!> \param tag ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_entry_mpi_recv(ENTRY, group, src, tag)
      TYPE(message_entry_type), INTENT(INOUT)            :: ENTRY
      INTEGER, INTENT(IN)                                :: group
      INTEGER, INTENT(INOUT)                             :: src, tag

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_entry_mpi_recv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: datatype, s
      INTEGER, DIMENSION(default_string_length)          :: value_str_arr
      INTEGER, DIMENSION(key_length)                     :: key_arr

      CALL mp_recv(key_arr, src, tag, group)
      entry%key = iarr2str(key_arr)

      CALL mp_recv(datatype, src, tag, group)

      SELECT CASE (datatype)
      CASE (1)
         ALLOCATE (entry%value_i4)
         CALL mp_recv(entry%value_i4, src, tag, group)
      CASE (2)
         ALLOCATE (entry%value_i8)
         CALL mp_recv(entry%value_i8, src, tag, group)
      CASE (3)
         ALLOCATE (entry%value_r4)
         CALL mp_recv(entry%value_r4, src, tag, group)
      CASE (4)
         ALLOCATE (entry%value_r8)
         CALL mp_recv(entry%value_r8, src, tag, group)

      CASE (5)
         CALL mp_recv(s, src, tag, group)
         ALLOCATE (entry%value_1d_i4(s))
         CALL mp_recv(entry%value_1d_i4, src, tag, group)
      CASE (6)
         CALL mp_recv(s, src, tag, group)
         ALLOCATE (entry%value_1d_i8(s))
         CALL mp_recv(entry%value_1d_i8, src, tag, group)
      CASE (7)
         CALL mp_recv(s, src, tag, group)
         ALLOCATE (entry%value_1d_r4(s))
         CALL mp_recv(entry%value_1d_r4, src, tag, group)
      CASE (8)
         CALL mp_recv(s, src, tag, group)
         ALLOCATE (entry%value_1d_r8(s))
         CALL mp_recv(entry%value_1d_r8, src, tag, group)
      CASE (9)
         ALLOCATE (entry%value_str)
         CALL mp_recv(value_str_arr, src, tag, group)
         entry%value_str = iarr2str(value_str_arr)
      CASE DEFAULT
         CALL cp__b("swarm/swarm_message.F",571,"unknown datatype")
      END SELECT
   END SUBROUTINE swarm_message_entry_mpi_recv

! **************************************************************************************************
!> \brief Helper routine for swarm_message_mpi_bcast.
!> \param ENTRY ...
!> \param src ...
!> \param group ...
!> \param mepos ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_entry_mpi_bcast(ENTRY, src, group, mepos)
      TYPE(message_entry_type), INTENT(INOUT)            :: ENTRY
      INTEGER, INTENT(IN)                                :: src, group, mepos

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_entry_mpi_bcast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: datasize, datatype
      INTEGER, DIMENSION(default_string_length)          :: value_str_arr
      INTEGER, DIMENSION(key_length)                     :: key_arr

      IF (src == mepos) key_arr = str2iarr(entry%key)
      CALL mp_bcast(key_arr, src, group)
      IF (src /= mepos) entry%key = iarr2str(key_arr)

      IF (src == mepos) THEN
         datasize = 1
         IF (ASSOCIATED(entry%value_i4)) THEN
            datatype = 1
         ELSE IF (ASSOCIATED(entry%value_i8)) THEN
            datatype = 2
         ELSE IF (ASSOCIATED(entry%value_r4)) THEN
            datatype = 3
         ELSE IF (ASSOCIATED(entry%value_r8)) THEN
            datatype = 4
         ELSE IF (ASSOCIATED(entry%value_1d_i4)) THEN
            datatype = 5
            datasize = SIZE(entry%value_1d_i4)
         ELSE IF (ASSOCIATED(entry%value_1d_i8)) THEN
            datatype = 6
            datasize = SIZE(entry%value_1d_i8)
         ELSE IF (ASSOCIATED(entry%value_1d_r4)) THEN
            datatype = 7
            datasize = SIZE(entry%value_1d_r4)
         ELSE IF (ASSOCIATED(entry%value_1d_r8)) THEN
            datatype = 8
            datasize = SIZE(entry%value_1d_r8)
         ELSE IF (ASSOCIATED(entry%value_str)) THEN
            datatype = 9
         ELSE
            CALL cp__b("swarm/swarm_message.F",623,"no value ASSOCIATED")
         END IF
      END IF
      CALL mp_bcast(datatype, src, group)
      CALL mp_bcast(datasize, src, group)

      SELECT CASE (datatype)
      CASE (1)
         IF (src /= mepos) ALLOCATE (entry%value_i4)
         CALL mp_bcast(entry%value_i4, src, group)
      CASE (2)
         IF (src /= mepos) ALLOCATE (entry%value_i8)
         CALL mp_bcast(entry%value_i8, src, group)
      CASE (3)
         IF (src /= mepos) ALLOCATE (entry%value_r4)
         CALL mp_bcast(entry%value_r4, src, group)
      CASE (4)
         IF (src /= mepos) ALLOCATE (entry%value_r8)
         CALL mp_bcast(entry%value_r8, src, group)
      CASE (5)
         IF (src /= mepos) ALLOCATE (entry%value_1d_i4(datasize))
         CALL mp_bcast(entry%value_1d_i4, src, group)
      CASE (6)
         IF (src /= mepos) ALLOCATE (entry%value_1d_i8(datasize))
         CALL mp_bcast(entry%value_1d_i8, src, group)
      CASE (7)
         IF (src /= mepos) ALLOCATE (entry%value_1d_r4(datasize))
         CALL mp_bcast(entry%value_1d_r4, src, group)
      CASE (8)
         IF (src /= mepos) ALLOCATE (entry%value_1d_r8(datasize))
         CALL mp_bcast(entry%value_1d_r8, src, group)
      CASE (9)
         IF (src == mepos) value_str_arr = str2iarr(entry%value_str)
         CALL mp_bcast(value_str_arr, src, group)
         IF (src /= mepos) THEN
            ALLOCATE (entry%value_str)
            entry%value_str = iarr2str(value_str_arr)
         END IF
      CASE DEFAULT
         CALL cp__b("swarm/swarm_message.F",662,"unknown datatype")
      END SELECT

   END SUBROUTINE swarm_message_entry_mpi_bcast

! **************************************************************************************************
!> \brief Helper routine for swarm_message_file_write.
!> \param ENTRY ...
!> \param unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_entry_file_write(ENTRY, unit)
      TYPE(message_entry_type), INTENT(IN)               :: ENTRY
      INTEGER, INTENT(IN)                                :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_entry_file_write', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      WRITE (unit, "(A,A)") "key: ", entry%key
      IF (ASSOCIATED(entry%value_i4)) THEN
         WRITE (unit, "(A)") "datatype: i4"
         WRITE (unit, "(A,I10)") "value: ", entry%value_i4

      ELSE IF (ASSOCIATED(entry%value_i8)) THEN
         WRITE (unit, "(A)") "datatype: i8"
         WRITE (unit, "(A,I20)") "value: ", entry%value_i8

      ELSE IF (ASSOCIATED(entry%value_r4)) THEN
         WRITE (unit, "(A)") "datatype: r4"
         WRITE (unit, "(A,E30.20)") "value: ", entry%value_r4

      ELSE IF (ASSOCIATED(entry%value_r8)) THEN
         WRITE (unit, "(A)") "datatype: r8"
         WRITE (unit, "(A,E30.20)") "value: ", entry%value_r8

      ELSE IF (ASSOCIATED(entry%value_str)) THEN
         WRITE (unit, "(A)") "datatype: str"
         WRITE (unit, "(A,A)") "value: ", entry%value_str

      ELSE IF (ASSOCIATED(entry%value_1d_i4)) THEN
         WRITE (unit, "(A)") "datatype: 1d_i4"
         WRITE (unit, "(A,I10)") "size: ", SIZE(entry%value_1d_i4)
         DO i = 1, SIZE(entry%value_1d_i4)
            WRITE (unit, *) entry%value_1d_i4(i)
         END DO

      ELSE IF (ASSOCIATED(entry%value_1d_i8)) THEN
         WRITE (unit, "(A)") "datatype: 1d_i8"
         WRITE (unit, "(A,I20)") "size: ", SIZE(entry%value_1d_i8)
         DO i = 1, SIZE(entry%value_1d_i8)
            WRITE (unit, *) entry%value_1d_i8(i)
         END DO

      ELSE IF (ASSOCIATED(entry%value_1d_r4)) THEN
         WRITE (unit, "(A)") "datatype: 1d_r4"
         WRITE (unit, "(A,I8)") "size: ", SIZE(entry%value_1d_r4)
         DO i = 1, SIZE(entry%value_1d_r4)
            WRITE (unit, "(1X,E30.20)") entry%value_1d_r4(i)
         END DO

      ELSE IF (ASSOCIATED(entry%value_1d_r8)) THEN
         WRITE (unit, "(A)") "datatype: 1d_r8"
         WRITE (unit, "(A,I8)") "size: ", SIZE(entry%value_1d_r8)
         DO i = 1, SIZE(entry%value_1d_r8)
            WRITE (unit, "(1X,E30.20)") entry%value_1d_r8(i)
         END DO

      ELSE
         CALL cp__b("swarm/swarm_message.F",732,"no value ASSOCIATED")
      END IF
   END SUBROUTINE swarm_message_entry_file_write

! **************************************************************************************************
!> \brief Helper routine for swarm_message_file_read.
!> \param ENTRY ...
!> \param parser ...
!> \param at_end ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_message_entry_file_read(ENTRY, parser, at_end)
      TYPE(message_entry_type), INTENT(INOUT)            :: ENTRY
      TYPE(cp_parser_type), POINTER                      :: parser
      LOGICAL, INTENT(INOUT)                             :: at_end

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_message_entry_file_read', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=15)                                  :: datatype, label
      INTEGER                                            :: arr_size, i
      LOGICAL                                            :: is_scalar

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN
      READ (parser%input_line(1:key_length+10), *) label, entry%key
      IF(.NOT.(TRIM(label) == "key:"))CALL cp__a("swarm/swarm_message.F",759)

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN
      READ (parser%input_line(1:30), *) label, datatype
      IF(.NOT.(TRIM(label) == "datatype:"))CALL cp__a("swarm/swarm_message.F",765)

      CALL parser_get_next_line(parser, 1, at_end)
      at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
      IF (at_end) RETURN

      is_scalar = .TRUE.
      SELECT CASE (TRIM (datatype))
      CASE ("i4")
         ALLOCATE (entry%value_i4)
         READ (parser%input_line(1:40), *) label, entry%value_i4
      CASE ("i8")
         ALLOCATE (entry%value_i8)
         READ (parser%input_line(1:40), *) label, entry%value_i8
      CASE ("r4")
         ALLOCATE (entry%value_r4)
         READ (parser%input_line(1:40), *) label, entry%value_r4
      CASE ("r8")
         ALLOCATE (entry%value_r8)
         READ (parser%input_line(1:40), *) label, entry%value_r8
      CASE ("str")
         ALLOCATE (entry%value_str)
         READ (parser%input_line(1:40), *) label, entry%value_str
      CASE DEFAULT
         is_scalar = .FALSE.
      END SELECT

      IF (is_scalar) THEN
         IF(.NOT.(TRIM(label) == "value:"))CALL cp__a("swarm/swarm_message.F",793)
         RETURN
      END IF

      ! musst be an array-datatype
      READ (parser%input_line(1:30), *) label, arr_size
      IF(.NOT.(TRIM(label) == "size:"))CALL cp__a("swarm/swarm_message.F",799)

      SELECT CASE (TRIM (datatype))
      CASE ("1d_i4")
         ALLOCATE (entry%value_1d_i4(arr_size))
      CASE ("1d_i8")
         ALLOCATE (entry%value_1d_i8(arr_size))
      CASE ("1d_r4")
         ALLOCATE (entry%value_1d_r4(arr_size))
      CASE ("1d_r8")
         ALLOCATE (entry%value_1d_r8(arr_size))
      CASE DEFAULT
         CALL cp__b("swarm/swarm_message.F",811,"unknown datatype")
      END SELECT

      DO i = 1, arr_size
         CALL parser_get_next_line(parser, 1, at_end)
         at_end = at_end .OR. LEN_TRIM(parser%input_line(1:10)) == 0
         IF (at_end) RETURN

         !Numbers were written with at most 31 characters.
         SELECT CASE (TRIM (datatype))
         CASE ("1d_i4")
            READ (parser%input_line(1:31), *) entry%value_1d_i4(i)
         CASE ("1d_i8")
            READ (parser%input_line(1:31), *) entry%value_1d_i8(i)
         CASE ("1d_r4")
            READ (parser%input_line(1:31), *) entry%value_1d_r4(i)
         CASE ("1d_r8")
            READ (parser%input_line(1:31), *) entry%value_1d_r8(i)
         CASE DEFAULT
            CALL cp__b("swarm/swarm_message.F",830,"swarm_message_entry_file_read: unknown datatype")
         END SELECT
      END DO

   END SUBROUTINE swarm_message_entry_file_read

! **************************************************************************************************
!> \brief Helper routine, converts a string into an integer-array
!> \param str ...
!> \retval arr ...
!> \author Ole Schuett
! **************************************************************************************************
   PURE FUNCTION str2iarr(str) RESULT(arr)
      CHARACTER(LEN=*), INTENT(IN)                       :: str
      INTEGER, DIMENSION(LEN(str))                       :: arr

      INTEGER                                            :: i

      DO i = 1, LEN(str)
         arr(i) = ICHAR(str(i:i))
      ENDDO
   END FUNCTION str2iarr

! **************************************************************************************************
!> \brief Helper routine, converts an integer-array into a string
!> \param arr ...
!> \retval str ...
!> \author Ole Schuett
! **************************************************************************************************
   PURE FUNCTION iarr2str(arr) RESULT(str)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: arr
      CHARACTER(LEN=SIZE(arr))                           :: str

      INTEGER                                            :: i

      DO i = 1, SIZE(arr)
         str(i:i) = CHAR(arr(i))
      ENDDO
   END FUNCTION iarr2str

! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_str(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   CHARACTER(LEN=*), INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_str: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_str)

   new_entry%value_str = value

   !WRITE (*,*) "swarm_message_add_str: key=",key, " value=",new_entry%value_str

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_str


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_str(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   CHARACTER(LEN=default_string_length), INTENT(OUT)                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_str: key=",key

   

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_str)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_str: value not associated key: "//TRIM(key))
         
         value = curr_entry%value_str
         !WRITE (*,*) "swarm_message_get_str: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_str

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_i4(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   INTEGER(KIND=int_4), INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_i4: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_i4)

   new_entry%value_i4 = value

   !WRITE (*,*) "swarm_message_add_i4: key=",key, " value=",new_entry%value_i4

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_i4


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_i4(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   INTEGER(KIND=int_4), INTENT(OUT)                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_i4: key=",key

   

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_i4)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_i4: value not associated key: "//TRIM(key))
         
         value = curr_entry%value_i4
         !WRITE (*,*) "swarm_message_get_i4: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_i4

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_i8(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   INTEGER(KIND=int_8), INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_i8: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_i8)

   new_entry%value_i8 = value

   !WRITE (*,*) "swarm_message_add_i8: key=",key, " value=",new_entry%value_i8

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_i8


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_i8(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   INTEGER(KIND=int_8), INTENT(OUT)                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_i8: key=",key

   

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_i8)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_i8: value not associated key: "//TRIM(key))
         
         value = curr_entry%value_i8
         !WRITE (*,*) "swarm_message_get_i8: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_i8

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_r4(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   REAL(KIND=real_4), INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_r4: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_r4)

   new_entry%value_r4 = value

   !WRITE (*,*) "swarm_message_add_r4: key=",key, " value=",new_entry%value_r4

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_r4


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_r4(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   REAL(KIND=real_4), INTENT(OUT)                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_r4: key=",key

   

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_r4)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_r4: value not associated key: "//TRIM(key))
         
         value = curr_entry%value_r4
         !WRITE (*,*) "swarm_message_get_r4: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_r4

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_r8(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   REAL(KIND=real_8), INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_r8: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_r8)

   new_entry%value_r8 = value

   !WRITE (*,*) "swarm_message_add_r8: key=",key, " value=",new_entry%value_r8

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_r8


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_r8(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   REAL(KIND=real_8), INTENT(OUT)                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_r8: key=",key

   

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_r8)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_r8: value not associated key: "//TRIM(key))
         
         value = curr_entry%value_r8
         !WRITE (*,*) "swarm_message_get_r8: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_r8

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_1d_i4(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   INTEGER(KIND=int_4), DIMENSION(:) , INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_1d_i4: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_1d_i4(SIZE(value)))

   new_entry%value_1d_i4 = value

   !WRITE (*,*) "swarm_message_add_1d_i4: key=",key, " value=",new_entry%value_1d_i4

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_1d_i4


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_1d_i4(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   INTEGER(KIND=int_4), DIMENSION(:), POINTER                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_1d_i4: key=",key

   IF(ASSOCIATED(value)) CALL cp__b("swarm/swarm_message.F",52,"swarm_message_get_1d_i4: value already associated")

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_1d_i4)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_1d_i4: value not associated key: "//TRIM(key))
         ALLOCATE(value(SIZE(curr_entry%value_1d_i4)))
         value = curr_entry%value_1d_i4
         !WRITE (*,*) "swarm_message_get_1d_i4: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_1d_i4

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_1d_i8(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   INTEGER(KIND=int_8), DIMENSION(:) , INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_1d_i8: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_1d_i8(SIZE(value)))

   new_entry%value_1d_i8 = value

   !WRITE (*,*) "swarm_message_add_1d_i8: key=",key, " value=",new_entry%value_1d_i8

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_1d_i8


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_1d_i8(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   INTEGER(KIND=int_8), DIMENSION(:), POINTER                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_1d_i8: key=",key

   IF(ASSOCIATED(value)) CALL cp__b("swarm/swarm_message.F",52,"swarm_message_get_1d_i8: value already associated")

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_1d_i8)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_1d_i8: value not associated key: "//TRIM(key))
         ALLOCATE(value(SIZE(curr_entry%value_1d_i8)))
         value = curr_entry%value_1d_i8
         !WRITE (*,*) "swarm_message_get_1d_i8: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_1d_i8

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_1d_r4(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   REAL(KIND=real_4), DIMENSION(:) , INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_1d_r4: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_1d_r4(SIZE(value)))

   new_entry%value_1d_r4 = value

   !WRITE (*,*) "swarm_message_add_1d_r4: key=",key, " value=",new_entry%value_1d_r4

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_1d_r4


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_1d_r4(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   REAL(KIND=real_4), DIMENSION(:), POINTER                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_1d_r4: key=",key

   IF(ASSOCIATED(value)) CALL cp__b("swarm/swarm_message.F",52,"swarm_message_get_1d_r4: value already associated")

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_1d_r4)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_1d_r4: value not associated key: "//TRIM(key))
         ALLOCATE(value(SIZE(curr_entry%value_1d_r4)))
         value = curr_entry%value_1d_r4
         !WRITE (*,*) "swarm_message_get_1d_r4: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_1d_r4

!EOF
! **************************************************************************************************
!> \brief Addes an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_add_1d_r8(msg, key, value)
   TYPE(swarm_message_type), INTENT(INOUT)   :: msg
   CHARACTER(LEN=*), INTENT(IN)              :: key
   REAL(KIND=real_8), DIMENSION(:) , INTENT(IN)                        :: value

   TYPE(message_entry_type), POINTER :: new_entry

   IF(swarm_message_haskey(msg, key)) &
      CALL cp__b("swarm/swarm_message.F",16,"swarm_message_add_1d_r8: key already exists: "//TRIM(key))

   ALLOCATE(new_entry)
   new_entry%key = key

   ALLOCATE(new_entry%value_1d_r8(SIZE(value)))

   new_entry%value_1d_r8 = value

   !WRITE (*,*) "swarm_message_add_1d_r8: key=",key, " value=",new_entry%value_1d_r8

   IF(.NOT. ASSOCIATED(msg%root)) THEN
      msg%root => new_entry
   ELSE
      new_entry%next => msg%root
      msg%root => new_entry
   ENDIF

 END SUBROUTINE swarm_message_add_1d_r8


! **************************************************************************************************
!> \brief Returns an entry from a swarm-message.
!> \param msg ...
!> \param key ...
!> \param value ...
!> \author Ole Schuett
! **************************************************************************************************
 SUBROUTINE swarm_message_get_1d_r8(msg, key, value)
   TYPE(swarm_message_type), INTENT(IN)  :: msg
   CHARACTER(LEN=*), INTENT(IN)          :: key
   REAL(KIND=real_8), DIMENSION(:), POINTER                             :: value

   TYPE(message_entry_type), POINTER :: curr_entry
   !WRITE (*,*) "swarm_message_get_1d_r8: key=",key

   IF(ASSOCIATED(value)) CALL cp__b("swarm/swarm_message.F",52,"swarm_message_get_1d_r8: value already associated")

   curr_entry => msg%root
   DO WHILE(ASSOCIATED(curr_entry))
      IF(TRIM(curr_entry%key) == TRIM(key)) THEN
         IF(.NOT. ASSOCIATED(curr_entry%value_1d_r8)) &
            CALL cp__b("swarm/swarm_message.F",58,"swarm_message_get_1d_r8: value not associated key: "//TRIM(key))
         ALLOCATE(value(SIZE(curr_entry%value_1d_r8)))
         value = curr_entry%value_1d_r8
         !WRITE (*,*) "swarm_message_get_1d_r8: value=",value
         RETURN
      ENDIF
      curr_entry => curr_entry%next
   END DO
   CALL cp__b("swarm/swarm_message.F",66,"swarm_message_get: key not found: "//TRIM(key))
 END SUBROUTINE swarm_message_get_1d_r8

!EOF

END MODULE swarm_message

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the Minima Hopping global optimization scheme
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_minhop
   USE bibliography,                    ONLY: Goedecker2004,&
                                              cite_reference
   USE glbopt_history,                  ONLY: history_add,&
                                              history_finalize,&
                                              history_fingerprint,&
                                              history_fingerprint_match,&
                                              history_fingerprint_type,&
                                              history_init,&
                                              history_lookup,&
                                              history_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE physcon,                         ONLY: kelvin
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_get,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_minhop'

   PUBLIC :: minhop_type
   PUBLIC :: minhop_init, minhop_finalize
   PUBLIC :: minhop_steer

   TYPE worker_state_type
      REAL(KIND=dp)                                       :: Eaccept = -1.0
      REAL(KIND=dp)                                       :: temp = -1.0
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: pos
      REAL(KIND=dp)                                       :: Epot = -1.0
      TYPE(history_fingerprint_type)                      :: fp
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: pos_hop
      REAL(KIND=dp)                                       :: Epot_hop = HUGE(1.0)
      TYPE(history_fingerprint_type)                      :: fp_hop
      INTEGER                                             :: minima_id = -1
      INTEGER                                             :: iframe = 1
   END TYPE worker_state_type

   TYPE minima_state_type
      REAL(KIND=dp)                                       :: Eaccept = -1.0
      REAL(KIND=dp)                                       :: temp = -1.0
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: pos
      REAL(KIND=dp)                                       :: Epot = -1.0
      TYPE(history_fingerprint_type)                      :: fp
      LOGICAL                                             :: disabled = .FALSE.
      INTEGER                                             :: n_active = 0
      INTEGER                                             :: n_sampled = 0
   END TYPE minima_state_type

   TYPE minhop_type
      PRIVATE
      TYPE(history_type), DIMENSION(:), ALLOCATABLE        :: history
      TYPE(worker_state_type), DIMENSION(:), ALLOCATABLE  :: worker_state
      TYPE(minima_state_type), DIMENSION(:), ALLOCATABLE  :: minima_state
      INTEGER                                             :: n_minima = 0
      REAL(KIND=dp)                                       :: beta1 = 0
      REAL(KIND=dp)                                       :: beta2 = 0
      REAL(KIND=dp)                                       :: beta3 = 0
      REAL(KIND=dp)                                       :: Eaccept0 = 0
      REAL(KIND=dp)                                       :: temp_init = 0
      REAL(KIND=dp)                                       :: temp_max = 0
      REAL(KIND=dp)                                       :: temp_min = 0
      REAL(KIND=dp)                                       :: alpha1 = 0
      REAL(KIND=dp)                                       :: alpha2 = 0
      INTEGER                                             :: n_accepted = 0
      INTEGER                                             :: n_rejected = 0
      INTEGER                                             :: iw = 0
      INTEGER                                             :: n_workers = 0
      LOGICAL                                             :: share_history = .FALSE.
   END TYPE minhop_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes master for Minima Hopping
!> \param this ...
!> \param glbopt_section ...
!> \param n_workers ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE minhop_init(this, glbopt_section, n_workers, iw)
      TYPE(minhop_type)                                  :: this
      TYPE(section_vals_type), POINTER                   :: glbopt_section
      INTEGER, INTENT(IN)                                :: n_workers, iw

      INTEGER                                            :: i, n_histories
      REAL(kind=dp)                                      :: temp_in_kelvin
      TYPE(section_vals_type), POINTER                   :: history_section, minhop_section

      CALL cite_reference(Goedecker2004)

      ! read input
      minhop_section => section_vals_get_subs_vals(glbopt_section, "MINIMA_HOPPING")
      CALL section_vals_val_get(minhop_section, "BETA_1", r_val=this%beta1)
      CALL section_vals_val_get(minhop_section, "BETA_2", r_val=this%beta2)
      CALL section_vals_val_get(minhop_section, "BETA_3", r_val=this%beta3)
      CALL section_vals_val_get(minhop_section, "ALPHA_1", r_val=this%alpha1)
      CALL section_vals_val_get(minhop_section, "ALPHA_2", r_val=this%alpha2)
      CALL section_vals_val_get(minhop_section, "E_ACCEPT_INIT", r_val=this%Eaccept0)
      CALL section_vals_val_get(minhop_section, "TEMPERATURE_INIT", r_val=temp_in_kelvin)
      this%temp_init = temp_in_kelvin/kelvin
      CALL section_vals_val_get(minhop_section, "SHARE_HISTORY", l_val=this%share_history)

      ! allocate history / histories
      history_section => section_vals_get_subs_vals(glbopt_section, "HISTORY")
      n_histories = n_workers
      IF (this%share_history) n_histories = 1
      ALLOCATE (this%history(n_histories))

      !only the first history shall write to iw
      CALL history_init(this%history(1), history_section, iw=iw)
      DO i = 2, n_histories
         CALL history_init(this%history(i), history_section, iw=-1)
      END DO

      ALLOCATE (this%worker_state(n_workers))
      this%n_workers = n_workers
      this%iw = iw

      IF (this%iw > 0) THEN
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| beta_1", this%beta1
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| beta_2", this%beta2
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| beta_3", this%beta3
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| alpha_1", this%alpha1
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| alpha_2", this%alpha2
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| Initial acceptance energy [Hartree]", this%Eaccept0
         WRITE (this%iw, '(A,T71,F10.3)') " MINHOP| Initial temperature [Kelvin]", this%temp_init*kelvin
         WRITE (this%iw, '(A,T71,L10)') " MINHOP| All workers share a single history", this%share_history
      ENDIF
   END SUBROUTINE minhop_init

! **************************************************************************************************
!> \brief Central steering routine of Minima Hopping
!> \param this ...
!> \param report ...
!> \param cmd ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE minhop_steer(this, report, cmd)
      TYPE(minhop_type)                                  :: this
      TYPE(swarm_message_type)                           :: report, cmd

      CHARACTER(len=default_string_length)               :: status
      INTEGER                                            :: hid, iframe, wid
      LOGICAL                                            :: minima_known
      REAL(KIND=dp)                                      :: report_Epot
      REAL(KIND=dp), DIMENSION(:), POINTER               :: report_positions
      TYPE(history_fingerprint_type)                     :: report_fp

      NULLIFY (report_positions)
      CALL swarm_message_get(report, "worker_id", wid)
      CALL swarm_message_get(report, "status", status)

      IF (TRIM(status) == "initial_hello") THEN
         this%worker_state(wid)%temp = this%temp_init
         this%worker_state(wid)%Eaccept = this%Eaccept0
         CALL swarm_message_add(cmd, "command", "md_and_gopt")
         CALL swarm_message_add(cmd, "iframe", 1)
         CALL swarm_message_add(cmd, "temperature", this%worker_state(wid)%temp)
         IF (this%iw > 0) WRITE (this%iw, '(1X,A,1X,I10,1X,A,7X,F10.3)') &
            "MINHOP| Sending worker", wid, &
            "initial temperature [Kelvin]", this%worker_state(wid)%temp*kelvin
         RETURN
      ENDIF

      hid = wid ! history_id = worker_id unless ....
      IF (this%share_history) hid = 1 !...there is ONE shared history.

      CALL swarm_message_get(report, "Epot", report_Epot)
      CALL swarm_message_get(report, "positions", report_positions)

      report_fp = history_fingerprint(report_Epot, report_positions)

      IF (.NOT. ALLOCATED(this%worker_state(wid)%pos)) THEN
         !init (first real report)
         this%worker_state(wid)%Epot = report_Epot
         ALLOCATE (this%worker_state(wid)%pos(SIZE(report_positions)))
         this%worker_state(wid)%pos(:) = report_positions
         this%worker_state(wid)%fp = report_fp
      END IF

      IF (history_fingerprint_match(this%history(hid), this%worker_state(wid)%fp, report_fp)) THEN
         ! not escaped
         IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| Not escaped"
         this%worker_state(wid)%temp = this%worker_state(wid)%temp*this%beta1 !increasing temperature
      ELSE
         ! escaped
         CALL history_lookup(this%history(hid), report_fp, minima_known)
         IF (minima_known) THEN
            IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| Escaped, old minima"
            this%worker_state(wid)%temp = this%worker_state(wid)%temp*this%beta2 !increasing temperature
         ELSE
            IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| Escaped, new minima"
            this%worker_state(wid)%temp = this%worker_state(wid)%temp*this%beta3 !decreasing temperature
            CALL history_add(this%history(hid), report_fp)
         ENDIF

         IF (report_Epot < this%worker_state(wid)%Epot_hop) THEN
            ! new locally lowest
            IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| New locally lowest"
            this%worker_state(wid)%Epot_hop = report_Epot
            IF (.NOT. ALLOCATED(this%worker_state(wid)%pos_hop)) &
               ALLOCATE (this%worker_state(wid)%pos_hop(SIZE(report_positions)))
            this%worker_state(wid)%pos_hop(:) = report_positions
            this%worker_state(wid)%fp_hop = report_fp
         ENDIF

         IF (this%worker_state(wid)%Epot_hop-this%worker_state(wid)%Epot < this%worker_state(wid)%Eaccept) THEN
            ! accept
            IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| Accept"
            this%worker_state(wid)%Epot = this%worker_state(wid)%Epot_hop
            this%worker_state(wid)%pos(:) = this%worker_state(wid)%pos_hop
            this%worker_state(wid)%fp = this%worker_state(wid)%fp_hop
            this%worker_state(wid)%Epot_hop = HUGE(1.0)

            this%worker_state(wid)%Eaccept = this%worker_state(wid)%Eaccept*this%alpha1 !decreasing Eaccept
            this%n_accepted = this%n_accepted+1
         ELSE
            ! not accept
            IF (this%iw > 0) WRITE (this%iw, '(A)') " MINHOP| Reject"
            this%worker_state(wid)%Eaccept = this%worker_state(wid)%Eaccept*this%alpha2 !increasing Eaccept
            this%n_rejected = this%n_rejected+1
         ENDIF
      END IF

      IF (this%iw > 0) THEN
         WRITE (this%iw, '(A,15X,E20.10)') &
            " MINHOP| Worker's acceptance Energy [Hartree]", this%worker_state(wid)%Eaccept
         WRITE (this%iw, '(A,22X,F20.3)') &
            " MINHOP| Worker's temperature [Kelvin]", this%worker_state(wid)%temp*kelvin
      END IF

      CALL swarm_message_get(report, "iframe", iframe)
      CALL swarm_message_add(cmd, "iframe", iframe)
      CALL swarm_message_add(cmd, "command", "md_and_gopt")
      CALL swarm_message_add(cmd, "positions", this%worker_state(wid)%pos)
      CALL swarm_message_add(cmd, "temperature", this%worker_state(wid)%temp)

      IF (this%iw > 0) THEN
         WRITE (this%iw, '(A,30X,I10)') &
            " MINHOP| Total number of accepted minima", this%n_accepted
         WRITE (this%iw, '(A,30X,I10)') &
            " MINHOP| Total number of rejected minima", this%n_rejected
      ENDIF

      DEALLOCATE (report_positions)
   END SUBROUTINE minhop_steer

! **************************************************************************************************
!> \brief Finalizes master for Minima Hopping
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE minhop_finalize(this)
      TYPE(minhop_type)                                  :: this

      INTEGER                                            :: i

      DO i = 1, SIZE(this%history)
         CALL history_finalize(this%history(i))
      END DO
   END SUBROUTINE minhop_finalize

END MODULE glbopt_minhop

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the Minima Crawling global optimization scheme
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_mincrawl
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE glbopt_history,                  ONLY: history_add,&
                                              history_finalize,&
                                              history_fingerprint,&
                                              history_fingerprint_type,&
                                              history_init,&
                                              history_lookup,&
                                              history_type
   USE input_constants,                 ONLY: dump_xmol
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE parallel_rng_types,              ONLY: create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: kelvin
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_get,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_mincrawl'

   PUBLIC :: mincrawl_type
   PUBLIC :: mincrawl_init, mincrawl_finalize
   PUBLIC :: mincrawl_steer

   TYPE minima_type
      INTEGER                                             :: id = -1
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: pos
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: escape_hist
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: tempdist
      REAL(KIND=dp)                                       :: Epot = -1.0
      TYPE(history_fingerprint_type)                      :: fp
      LOGICAL                                             :: disabled = .FALSE.
      INTEGER                                             :: n_active = 0
      INTEGER                                             :: n_sampled = 0
   END TYPE minima_type

   TYPE minima_p_type
      TYPE(minima_type), POINTER                          :: p => Null()
   END TYPE minima_p_type

   TYPE worker_state_type
      TYPE(minima_type), POINTER                          :: start_minima => Null()
      INTEGER                                             :: tempstep = 0
      INTEGER                                             :: iframe = 1
   END TYPE worker_state_type

   TYPE mincrawl_type
      PRIVATE
      TYPE(history_type)                                  :: history
      TYPE(worker_state_type), DIMENSION(:), ALLOCATABLE  :: workers
      TYPE(minima_p_type), DIMENSION(:), ALLOCATABLE      :: minimas
      REAL(KIND=dp)                                       :: tempstep_base = 0
      INTEGER                                             :: tempstep_max = 0
      REAL(KIND=dp)                                       :: tempdist_init_width = 0
      REAL(KIND=dp)                                       :: tempdist_update_width = 0
      REAL(KIND=dp)                                       :: tempdist_update_height = 0
      INTEGER                                             :: esc_hist_len = 0
      INTEGER                                             :: tempstep_init = 0
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE            :: tempdist_init
      INTEGER                                             :: n_minima = 0
      INTEGER                                             :: n_workers = 0
      INTEGER                                             :: worker_per_min = 0
      INTEGER                                             :: iw = 0
      INTEGER                                             :: minima_traj_unit = 0
      TYPE(section_vals_type), POINTER                    :: mincrawl_section => Null()
      TYPE(rng_stream_type), POINTER                       :: rng_stream => Null()
      TYPE(particle_type), DIMENSION(:), POINTER          :: particle_set => Null()
   END TYPE mincrawl_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes master for Minima Crawling
!> \param this ...
!> \param glbopt_section ...
!> \param n_workers ...
!> \param iw ...
!> \param particle_set ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mincrawl_init(this, glbopt_section, n_workers, iw, particle_set)
      TYPE(mincrawl_type)                                :: this
      TYPE(section_vals_type), POINTER                   :: glbopt_section
      INTEGER, INTENT(IN)                                :: n_workers, iw
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      INTEGER                                            :: i
      REAL(kind=dp)                                      :: temp_in_kelvin
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: history_section

      NULLIFY (logger, history_section)

      ! read input
      this%mincrawl_section => section_vals_get_subs_vals(glbopt_section, "MINIMA_CRAWLING")
      CALL section_vals_val_get(this%mincrawl_section, "TEMPSTEP_BASE", r_val=this%tempstep_base)
      CALL section_vals_val_get(this%mincrawl_section, "TEMPSTEP_MAX", i_val=this%tempstep_max)
      CALL section_vals_val_get(this%mincrawl_section, "TEMPDIST_INIT_WIDTH", r_val=this%tempdist_init_width)
      CALL section_vals_val_get(this%mincrawl_section, "TEMPDIST_UPDATE_WIDTH", r_val=this%tempdist_update_width)
      CALL section_vals_val_get(this%mincrawl_section, "TEMPDIST_UPDATE_HEIGHT", r_val=this%tempdist_update_height)
      CALL section_vals_val_get(this%mincrawl_section, "TEMPERATURE_INIT", r_val=temp_in_kelvin)
      this%tempstep_init = temp2tempstep(this, temp_in_kelvin/kelvin)
      CALL section_vals_val_get(this%mincrawl_section, "WORKER_PER_MINIMA", i_val=this%worker_per_min)
      CALL section_vals_val_get(this%mincrawl_section, "ESCAPE_HISTORY_LENGTH", i_val=this%esc_hist_len)

      !init minima trajectory
      logger => cp_get_default_logger()
      this%minima_traj_unit = cp_print_key_unit_nr(logger, &
                                                   this%mincrawl_section, "MINIMA_TRAJECTORY", &
                                                   middle_name="minima", extension=".xyz", &
                                                   file_action="WRITE", file_position="REWIND")

      !init history
      history_section => section_vals_get_subs_vals(glbopt_section, "HISTORY")
      CALL history_init(this%history, history_section, iw=iw)

      !allocate data structures
      ALLOCATE (this%minimas(1000)) !will be grown if needed

      ALLOCATE (this%workers(n_workers))
      this%n_workers = n_workers
      this%iw = iw
      this%particle_set => particle_set

      ! call fermi-like stepfunction for initial temp-dist
      ALLOCATE (this%tempdist_init(this%tempstep_max))
      this%tempdist_init = 0.0
      DO i = 1, this%tempstep_max
         this%tempdist_init(i) = 1.0/(1.0+EXP((this%tempstep_init-i)/this%tempdist_init_width))
      ENDDO

      CALL create_rng_stream(this%rng_stream, name="mincrawl")
   END SUBROUTINE mincrawl_init

! **************************************************************************************************
!> \brief Central steering routine of Minima Crawling
!> \param this ...
!> \param report ...
!> \param cmd ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mincrawl_steer(this, report, cmd)
      TYPE(mincrawl_type)                                :: this
      TYPE(swarm_message_type)                           :: report, cmd

      CHARACTER(len=default_string_length)               :: status
      INTEGER                                            :: wid
      TYPE(minima_type), POINTER                         :: best_minima

      CALL swarm_message_get(report, "status", status)
      CALL swarm_message_get(report, "worker_id", wid)

      IF (TRIM(status) == "initial_hello") THEN
         this%workers(wid)%tempstep = this%tempstep_init
         CALL swarm_message_add(cmd, "command", "md_and_gopt")
         CALL swarm_message_add(cmd, "iframe", 1)
         CALL swarm_message_add(cmd, "temperature", tempstep2temp(this, this%workers(wid)%tempstep))
         RETURN
      ENDIF

      IF (TRIM(status) == "ok") &
         CALL mincrawl_register_minima(this, report)

      IF (.FALSE.) CALL print_tempdist(best_minima)

      best_minima => choose_promising_minima(this)

      IF (.NOT. ASSOCIATED(best_minima)) THEN ! no suitable minima found
         CALL swarm_message_add(cmd, "command", "wait")
         !WRITE(this%iw,*) " MINCRAWL| Waiting until new minima become available"
         RETURN
      ENDIF

      best_minima%n_active = best_minima%n_active+1
      best_minima%n_sampled = best_minima%n_sampled+1
      this%workers(wid)%start_minima => best_minima
      this%workers(wid)%tempstep = choose_tempstep(this, best_minima)

      CALL swarm_message_add(cmd, "command", "md_and_gopt")
      CALL swarm_message_add(cmd, "iframe", this%workers(wid)%iframe)
      CALL swarm_message_add(cmd, "temperature", tempstep2temp(this, this%workers(wid)%tempstep))
      CALL swarm_message_add(cmd, "positions", best_minima%pos)

      IF (this%iw > 0) THEN
         WRITE (this%iw, '(1X,A,T71,I10)') &
            "MINCRAWL| Total number of found minima", this%n_minima
         WRITE (this%iw, '(1X,A,T71,I10)') &
            "MINCRAWL| Sampling minima with id", best_minima%id
         WRITE (this%iw, '(1X,A,I10,A,A,T71,F10.3)') &
            "MINCRAWL| Temperature  (step ", this%workers(wid)%tempstep, " ) ", &
            "[Kelvin]", kelvin*tempstep2temp(this, this%workers(wid)%tempstep)
      ENDIF

   END SUBROUTINE mincrawl_steer

! **************************************************************************************************
!> \brief Helper routine for mincrawl_steer, choses minimum based on its score.
!> \param this ...
!> \retval minima ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION choose_promising_minima(this) RESULT(minima)
      TYPE(mincrawl_type)                                :: this
      TYPE(minima_type), POINTER                         :: minima

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: score, score_best

      score_best = HUGE(1.0)
      NULLIFY (minima)

      DO i = 1, this%n_minima
         IF (this%minimas(i)%p%disabled) CYCLE
         IF (this%minimas(i)%p%n_active > this%worker_per_min) CYCLE
         score = minima_score(this%minimas(i)%p)
!       WRITE (*,*) "Minima: ", i, " active: ",this%minimas(i)%active, " E_expect: ", E_expect
         IF (score < score_best) THEN
            score_best = score
            minima => this%minimas(i)%p
         ENDIF
      END DO
   END FUNCTION choose_promising_minima

! **************************************************************************************************
!> \brief Helper routine for choose_promising_minima, calculates a minimum's score
!> \param minima ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION minima_score(minima) RESULT(res)
      TYPE(minima_type), POINTER                         :: minima
      REAL(KIND=dp)                                      :: res

      res = SUM(minima%escape_hist)/SIZE(minima%escape_hist)
   END FUNCTION minima_score

! **************************************************************************************************
!> \brief Helper routine for mincrawl_steer, samples from a temp-dist.
!> \param this ...
!> \param minima ...
!> \retval step ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION choose_tempstep(this, minima) RESULT(step)
      TYPE(mincrawl_type)                                :: this
      TYPE(minima_type), POINTER                         :: minima
      INTEGER                                            :: step

      REAL(KIND=dp)                                      :: a, r

      DO
         r = next_random_number(this%rng_stream)
         step = INT(r*SIZE(minima%tempdist))+1
         a = 1.0-2.0*ABS(minima%tempdist(step)-0.5)
         r = next_random_number(this%rng_stream)
         IF (r < a) EXIT
      END DO

   END FUNCTION choose_tempstep

! **************************************************************************************************
!> \brief Debugging routine, prints a minimum's temp-distribution.
!> \param minima ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE print_tempdist(minima)
      TYPE(minima_type), POINTER                         :: minima

      INTEGER                                            :: i

!WRITE (*,*) "tempdist: ", SUM(minima%tempdist, DIM=1)

      DO i = 1, SIZE(minima%tempdist)
         WRITE (*, *) "tempstep: ", i, minima%tempdist(i)
      ENDDO
   END SUBROUTINE print_tempdist

! **************************************************************************************************
!> \brief Helper routine, convertes a  discrete temp-step to a temperature.
!> \param this ...
!> \param step ...
!> \retval temp_in_au ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION tempstep2temp(this, step) RESULT(temp_in_au)
      TYPE(mincrawl_type)                                :: this
      INTEGER                                            :: step
      REAL(KIND=dp)                                      :: temp_in_au

      temp_in_au = (this%tempstep_base**step)/kelvin
   END FUNCTION tempstep2temp

! **************************************************************************************************
!> \brief Helper routine, convertes a temperature to a discrete temp-step.
!> \param this ...
!> \param temp_in_au ...
!> \retval step ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION temp2tempstep(this, temp_in_au) RESULT(step)
      TYPE(mincrawl_type)                                :: this
      REAL(KIND=dp)                                      :: temp_in_au
      INTEGER                                            :: step

      CHARACTER(len=*), PARAMETER :: routineN = 'temp2tempstep', routineP = moduleN//':'//routineN

      step = INT(LOG(temp_in_au*kelvin)/LOG(this%tempstep_base))
      !WRITE(*,*) "temp: ", temp_in_au*kelvin, this%tempstep_base
      !WRITE(*,*) "step: ", step
      IF (step > this%tempstep_max) CALL cp__b("swarm/glbopt_mincrawl.F",335,"temp2tempstep: step > tempstep_max")
   END FUNCTION temp2tempstep

! **************************************************************************************************
!> \brief Helper routine for mincrawl_steer
!>        Incorporates information of new report into history.
!> \param this ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mincrawl_register_minima(this, report)
      TYPE(mincrawl_type)                                :: this
      TYPE(swarm_message_type)                           :: report

      CHARACTER(len=*), PARAMETER :: routineN = 'mincrawl_register_minima', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: new_mid, tempstep, wid
      LOGICAL                                            :: minima_known
      REAL(KIND=dp)                                      :: report_Epot
      REAL(KIND=dp), DIMENSION(:), POINTER               :: report_positions
      TYPE(history_fingerprint_type)                     :: report_fp
      TYPE(minima_p_type), ALLOCATABLE, DIMENSION(:)     :: minimas_tmp
      TYPE(minima_type), POINTER                         :: new_minima, start_minima

      NULLIFY (start_minima, new_minima, report_positions)

      CALL swarm_message_get(report, "worker_id", wid)
      CALL swarm_message_get(report, "Epot", report_Epot)
      CALL swarm_message_get(report, "positions", report_positions)
      CALL swarm_message_get(report, "iframe", this%workers(wid)%iframe)

      start_minima => this%workers(wid)%start_minima
      tempstep = this%workers(wid)%tempstep

      report_fp = history_fingerprint(report_Epot, report_positions)
      CALL history_lookup(this%history, report_fp, minima_known)

      IF (ASSOCIATED(start_minima)) THEN
         start_minima%n_active = start_minima%n_active-1
         IF (start_minima%n_active < 0) CALL cp__b("swarm/glbopt_mincrawl.F",375,"negative n_active")

         ! update tempdist and escape_hist
         IF (minima_known) THEN
            CALL update_tempdist(this, start_minima%tempdist, tempstep, -1)
         ELSE
            CALL update_tempdist(this, start_minima%tempdist, tempstep, +1)
            start_minima%escape_hist(:) = EOSHIFT(start_minima%escape_hist, 1)
            start_minima%escape_hist(1) = report_Epot
         ENDIF

      ENDIF

      IF (.NOT. minima_known) THEN
         this%n_minima = this%n_minima+1
         IF (this%n_minima > SIZE(this%minimas)) THEN
            ALLOCATE (minimas_tmp(SIZE(this%minimas)))
            minimas_tmp(:) = this%minimas
            DEALLOCATE (this%minimas)
            ALLOCATE (this%minimas(SIZE(minimas_tmp)+1000))
            this%minimas(:SIZE(minimas_tmp)) = minimas_tmp
            DEALLOCATE (minimas_tmp)
         ENDIF

         new_mid = this%n_minima
         ALLOCATE (this%minimas(new_mid)%p)
         new_minima => this%minimas(new_mid)%p
         new_minima%id = new_mid
         ALLOCATE (new_minima%escape_hist(this%esc_hist_len))
         ALLOCATE (new_minima%tempdist(this%tempstep_max))

         new_minima%escape_hist(:) = report_Epot !init with Epot

         IF (ASSOCIATED(start_minima)) THEN
            new_minima%tempdist(:) = start_minima%tempdist ! inherit tempdist
         ELSE
            new_minima%tempdist(:) = this%tempdist_init
         ENDIF

         new_minima%Epot = report_Epot
         new_minima%fp = report_fp
         ALLOCATE (new_minima%pos(SIZE(report_positions)))
         new_minima%pos(:) = report_positions

         IF (ASSOCIATED(start_minima)) THEN
            IF (report_Epot < start_minima%Epot) THEN
               start_minima%disabled = .TRUE.
               IF (this%iw > 0) WRITE (this%iw, '(1X,A,T71,I10)') &
                  "MINCRAWL| Disabling minimum with id", start_minima%id
            ENDIF
         ENDIF

         IF (this%iw > 0) WRITE (this%iw, '(1X,A,T71,I10)') &
            "MINCRAWL| Adding new minima with id", new_mid

         CALL history_add(this%history, report_fp, id=new_mid)
         CALL write_minima_traj(this, wid, new_mid, report_Epot, report_positions)
      ENDIF
      DEALLOCATE (report_positions)
   END SUBROUTINE mincrawl_register_minima

! **************************************************************************************************
!> \brief Helper routine for mincrawl_register_minima.
!>        Adds or substracts small Gaussian from a minimum's temp-distribution.
!> \param this ...
!> \param tempdist ...
!> \param center ...
!> \param direction ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE update_tempdist(this, tempdist, center, direction)
      TYPE(mincrawl_type)                                :: this
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: tempdist
      INTEGER                                            :: center, direction

      INTEGER                                            :: i

      DO i = 1, SIZE(tempdist)
         tempdist(i) = tempdist(i)+this%tempdist_update_height &
                       *REAL(direction, KIND=dp)*EXP(-((center-i)/this%tempdist_update_width)**2)
         tempdist(i) = MAX(0.0_dp, MIN(1.0_dp, tempdist(i)))
      ENDDO
   END SUBROUTINE update_tempdist

! **************************************************************************************************
!> \brief Helper routine for mincrawl_register_minima, write trajectory.
!> \param this ...
!> \param worker_id ...
!> \param minimum_id ...
!> \param Epot ...
!> \param positions ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE write_minima_traj(this, worker_id, minimum_id, Epot, positions)
      TYPE(mincrawl_type), INTENT(INOUT)                 :: this
      INTEGER, INTENT(IN)                                :: worker_id, minimum_id
      REAL(KIND=dp), INTENT(IN)                          :: Epot
      REAL(KIND=dp), DIMENSION(:), POINTER               :: positions

      CHARACTER(len=default_string_length)               :: title, unit_str
      REAL(KIND=dp)                                      :: unit_conv

      IF (this%minima_traj_unit <= 0) RETURN

      WRITE (title, '(A,I8,A,I5,A,F20.10)') 'minimum_id = ', minimum_id, &
         ' worker_id = ', worker_id, ' Epot = ', Epot

      !get the conversion factor for the length unit
      CALL section_vals_val_get(this%mincrawl_section, "MINIMA_TRAJECTORY%UNIT", &
                                c_val=unit_str)
      unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

      CALL write_particle_coordinates(this%particle_set, &
                                      iunit=this%minima_traj_unit, &
                                      output_format=dump_xmol, &
                                      content="POS", &
                                      title=TRIM(title), &
                                      array=positions, &
                                      unit_conv=unit_conv)
   END SUBROUTINE write_minima_traj

! **************************************************************************************************
!> \brief Finalizes master for Minima Crawling
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mincrawl_finalize(this)
      TYPE(mincrawl_type)                                :: this

      INTEGER                                            :: i
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)

      DO i = 1, this%n_minima
         !WRITE (*,*) "Minima: ", i, " n_sampled: ",this%minimas(i)%n_sampled
         DEALLOCATE (this%minimas(i)%p)
      ENDDO

      logger => cp_get_default_logger()
      CALL cp_print_key_finished_output(this%minima_traj_unit, logger, &
                                        this%mincrawl_section, "MINIMA_TRAJECTORY")

      CALL history_finalize(this%history)
      CALL delete_rng_stream(this%rng_stream)
   END SUBROUTINE mincrawl_finalize

END MODULE glbopt_mincrawl

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Master's routines for global optimization
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_master
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              deallocate_atomic_kind_set
   USE colvar_types,                    ONLY: colvar_p_release,&
                                              colvar_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE exclusion_types,                 ONLY: exclusion_release,&
                                              exclusion_type
   USE glbopt_mincrawl,                 ONLY: mincrawl_finalize,&
                                              mincrawl_init,&
                                              mincrawl_steer,&
                                              mincrawl_type
   USE glbopt_minhop,                   ONLY: minhop_finalize,&
                                              minhop_init,&
                                              minhop_steer,&
                                              minhop_type
   USE input_constants,                 ONLY: dump_xmol,&
                                              glbopt_do_mincrawl,&
                                              glbopt_do_minhop
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE molecule_kind_types,             ONLY: deallocate_molecule_kind_set,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: deallocate_global_constraint,&
                                              deallocate_molecule_set,&
                                              global_constraint_type,&
                                              molecule_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE particle_types,                  ONLY: deallocate_particle_set,&
                                              particle_type
   USE swarm_message,                   ONLY: swarm_message_get,&
                                              swarm_message_type
   USE topology,                        ONLY: topology_control
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_master'

   PUBLIC :: glbopt_master_type
   PUBLIC :: glbopt_master_init, glbopt_master_finalize
   PUBLIC :: glbopt_master_steer

   TYPE glbopt_master_type
      PRIVATE
      REAL(KIND=dp)                                       :: E_lowest = HUGE(1.0_dp)
      REAL(KIND=dp)                                       :: E_target = TINY(1.0_dp)
      INTEGER                                             :: iw = 0
      INTEGER                                             :: progress_traj_unit = 0
      INTEGER(int_8)                                      :: total_md_steps = 0
      INTEGER(int_8)                                      :: total_gopt_steps = 0
      INTEGER(int_8)                                      :: count_reports = 0
      INTEGER                                             :: method
      TYPE(minhop_type), POINTER                           :: minhop
      TYPE(mincrawl_type), POINTER                        :: mincrawl
      INTEGER                                             :: i_iteration = 0
      TYPE(atomic_kind_type), DIMENSION(:), POINTER       :: atomic_kind_set => Null()
      TYPE(particle_type), DIMENSION(:), POINTER          :: particle_set => Null()
      TYPE(section_vals_type), POINTER                    :: glbopt_section => Null()
   END TYPE glbopt_master_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes the master of the global optimizer
!> \param this ...
!> \param para_env ...
!> \param root_section ...
!> \param n_walkers ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_master_init(this, para_env, root_section, n_walkers, iw)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: n_walkers, iw

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_master_init', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)

      this%iw = iw

      this%glbopt_section => section_vals_get_subs_vals(root_section, "SWARM%GLOBAL_OPT")
      CALL section_vals_retain(this%glbopt_section)

      CALL section_vals_val_get(this%glbopt_section, "E_TARGET", r_val=this%E_target)
      CALL section_vals_val_get(this%glbopt_section, "METHOD", i_val=this%method)

      CALL glbopt_read_particle_set(this, para_env, root_section)

      logger => cp_get_default_logger()
      this%progress_traj_unit = cp_print_key_unit_nr(logger, &
                                                     this%glbopt_section, "PROGRESS_TRAJECTORY", &
                                                     middle_name="progress", extension=".xyz", &
                                                     file_action="WRITE", file_position="REWIND")

      SELECT CASE (this%method)
      CASE (glbopt_do_minhop)
         ALLOCATE (this%minhop)
         CALL minhop_init(this%minhop, this%glbopt_section, n_walkers, iw)
      CASE (glbopt_do_mincrawl)
         ALLOCATE (this%mincrawl)
         CALL mincrawl_init(this%mincrawl, this%glbopt_section, n_walkers, iw, this%particle_set)
      CASE DEFAULT
         CALL cp__b("swarm/glbopt_master.F",131,"Unknown glbopt_method")
      END SELECT
   END SUBROUTINE glbopt_master_init

! **************************************************************************************************
!> \brief Helper-routine for glbopt_master_init, reads part of SUBSYS-section
!> \param this ...
!> \param para_env ...
!> \param root_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_read_particle_set(this, para_env, root_section)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(exclusion_type), DIMENSION(:), POINTER        :: exclusions
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section

      NULLIFY (atomic_kind_set, particle_set, molecule_kind_set, molecule_set)
      NULLIFY (colvar_p, gci, exclusions, force_env_section, subsys_section)

      force_env_section => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
      subsys_section => section_vals_get_subs_vals(root_section, "FORCE_EVAL%SUBSYS")

      CALL topology_control(atomic_kind_set, &
                            particle_set, &
                            molecule_kind_set, &
                            molecule_set, &
                            colvar_p, &
                            gci, &
                            root_section=root_section, &
                            para_env=para_env, &
                            force_env_section=force_env_section, &
                            subsys_section=subsys_section, &
                            use_motion_section=.FALSE., &
                            exclusions=exclusions)

      !This is the only thing we need to write trajectories.
      this%atomic_kind_set => atomic_kind_set
      this%particle_set => particle_set
      CALL exclusion_release(exclusions)
      CALL deallocate_molecule_set(molecule_set)
      CALL deallocate_molecule_kind_set(molecule_kind_set)
      CALL deallocate_global_constraint(gci)
      CALL colvar_p_release(colvar_p)

   END SUBROUTINE glbopt_read_particle_set

! **************************************************************************************************
!> \brief Central steering routine of global optimizer
!> \param this ...
!> \param report ...
!> \param cmd ...
!> \param should_stop ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_master_steer(this, report, cmd, should_stop)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this
      TYPE(swarm_message_type)                           :: report, cmd
      LOGICAL, INTENT(INOUT)                             :: should_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_master_steer', &
         routineP = moduleN//':'//routineN

      CALL progress_report(this, report)

      SELECT CASE (this%method)
      CASE (glbopt_do_minhop)
         CALL minhop_steer(this%minhop, report, cmd)
      CASE (glbopt_do_mincrawl)
         CALL mincrawl_steer(this%mincrawl, report, cmd)
      CASE DEFAULT
         CALL cp__b("swarm/glbopt_master.F",210,"Unknown glbopt_method")
      END SELECT

      IF (this%E_lowest < this%E_target) THEN
         IF (this%iw > 0) WRITE (this%iw, "(A,I8,A)") &
            " GLBOPT| Reached E_pot < E_target after ", this%i_iteration, " iterations. Quitting."
         should_stop = .TRUE.
      END IF
   END SUBROUTINE glbopt_master_steer

! **************************************************************************************************
!> \brief Helper routine for glbopt_master_steer(), updates stats, etc.
!> \param this ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE progress_report(this, report)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this
      TYPE(swarm_message_type)                           :: report

      CHARACTER(len=default_string_length)               :: status
      INTEGER                                            :: gopt_steps, md_steps, report_worker_id
      REAL(KIND=dp)                                      :: report_Epot

      this%i_iteration = this%i_iteration+1

      CALL swarm_message_get(report, "worker_id", report_worker_id)
      CALL swarm_message_get(report, "status", status)

      IF (TRIM(status) == "ok") THEN
         CALL swarm_message_get(report, "Epot", report_Epot)
         CALL swarm_message_get(report, "md_steps", md_steps)
         CALL swarm_message_get(report, "gopt_steps", gopt_steps)
         this%total_md_steps = this%total_md_steps+md_steps
         this%total_gopt_steps = this%total_gopt_steps+gopt_steps
         this%count_reports = this%count_reports+1

         IF (report_Epot < this%E_lowest) THEN
            this%E_lowest = report_Epot
            CALL write_progress_traj(this, report)
         ENDIF

         IF (this%iw > 0) THEN
            WRITE (this%iw, '(A,46X,I8)') &
               " GLBOPT| Reporting worker ", report_worker_id
            WRITE (this%iw, '(A,20X,E15.8)') &
               " GLBOPT| Reported potential energy [Hartree] ", report_Epot
            WRITE (this%iw, '(A,13X,E15.8)') &
               " GLBOPT| Lowest reported potential energy [Hartree] ", this%E_lowest
            WRITE (this%iw, '(A,T71,F10.1)') &
               " GLBOPT| Average number of MD steps", REAL(this%total_md_steps, KIND=dp)/this%count_reports
            WRITE (this%iw, '(A,T71,F10.1)') &
               " GLBOPT| Average number of Geo-Opt steps", REAL(this%total_gopt_steps, KIND=dp)/this%count_reports
         END IF
      END IF
   END SUBROUTINE progress_report

! **************************************************************************************************
!> \brief Helper routine for progress_report(), write frames to trajectory.
!> \param this ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE write_progress_traj(this, report)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this
      TYPE(swarm_message_type), INTENT(IN)               :: report

      CHARACTER(len=default_string_length)               :: title, unit_str
      INTEGER                                            :: report_worker_id
      REAL(KIND=dp)                                      :: report_Epot, unit_conv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: report_positions

      NULLIFY (report_positions)

      IF (this%progress_traj_unit <= 0) RETURN

      CALL swarm_message_get(report, "worker_id", report_worker_id)
      CALL swarm_message_get(report, "positions", report_positions)
      CALL swarm_message_get(report, "Epot", report_Epot)

      WRITE (title, '(A,I8,A,I5,A,F20.10)') 'i = ', this%i_iteration, &
         ' worker_id = ', report_worker_id, ' Epot = ', report_Epot

      !get the conversion factor for the length unit
      CALL section_vals_val_get(this%glbopt_section, "PROGRESS_TRAJECTORY%UNIT", &
                                c_val=unit_str)
      unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      CALL write_particle_coordinates(this%particle_set, &
                                      iunit=this%progress_traj_unit, &
                                      output_format=dump_xmol, &
                                      content="POS", &
                                      title=TRIM(title), &
                                      array=report_positions, &
                                      unit_conv=unit_conv)
      DEALLOCATE (report_positions)
   END SUBROUTINE write_progress_traj

! **************************************************************************************************
!> \brief Finalized the master of the global optimizer
!> \param this ...
!> \author Ole
! **************************************************************************************************
   SUBROUTINE glbopt_master_finalize(this)
      TYPE(glbopt_master_type), INTENT(INOUT)            :: this

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_master_finalize', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)

      SELECT CASE (this%method)
      CASE (glbopt_do_minhop)
         CALL minhop_finalize(this%minhop)
         DEALLOCATE (this%minhop)
      CASE (glbopt_do_mincrawl)
         CALL mincrawl_finalize(this%mincrawl)
         DEALLOCATE (this%mincrawl)
      CASE DEFAULT
         CALL cp__b("swarm/glbopt_master.F",330,"Unknown glbopt_method")
      END SELECT

      logger => cp_get_default_logger()
      CALL cp_print_key_finished_output(this%progress_traj_unit, logger, &
                                        this%glbopt_section, "PROGRESS_TRAJECTORY")

      CALL section_vals_release(this%glbopt_section)
      CALL deallocate_particle_set(this%particle_set)
      CALL deallocate_atomic_kind_set(this%atomic_kind_set)

   END SUBROUTINE glbopt_master_finalize

END MODULE glbopt_master

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Master's routines for the swarm-framework
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm_master
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE glbopt_master,                   ONLY: glbopt_master_finalize,&
                                              glbopt_master_init,&
                                              glbopt_master_steer,&
                                              glbopt_master_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: swarm_do_glbopt
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_equal,&
                                              swarm_message_file_read,&
                                              swarm_message_file_write,&
                                              swarm_message_free,&
                                              swarm_message_get,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm_master'

   PUBLIC :: swarm_master_type
   PUBLIC :: swarm_master_init, swarm_master_finalize
   PUBLIC :: swarm_master_steer

   TYPE swarm_message_p_type
      TYPE(swarm_message_type), POINTER                   :: p => Null()
   END TYPE swarm_message_p_type

   TYPE swarm_master_type
      PRIVATE
      INTEGER                                             :: behavior = -1
      TYPE(glbopt_master_type), POINTER                   :: glbopt => Null()
      !possibly more behaviors ...
      INTEGER                                             :: iw = 0
      INTEGER                                             :: i_iteration = 0
      INTEGER                                             :: max_iter = 0
      LOGICAL                                             :: should_stop = .FALSE.
      INTEGER                                             :: n_workers = -1
      INTEGER                                             :: comlog_unit
      TYPE(section_vals_type), POINTER                    :: swarm_section => Null()
      TYPE(cp_para_env_type), POINTER                     :: para_env => Null()
      TYPE(swarm_message_p_type), DIMENSION(:), POINTER   :: queued_commands => Null()
      TYPE(global_environment_type), POINTER              :: globenv => Null()
      LOGICAL                                             :: ignore_last_iteration = .FALSE.
      INTEGER                                             :: n_waiting = 0
   END TYPE swarm_master_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes the swarm master
!> \param master ...
!> \param para_env ...
!> \param globenv ...
!> \param root_section ...
!> \param n_workers ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_master_init(master, para_env, globenv, root_section, n_workers)
      TYPE(swarm_master_type)                            :: master
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: n_workers

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_master_init', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      master%swarm_section => section_vals_get_subs_vals(root_section, "SWARM")

      logger => cp_get_default_logger()
      master%n_workers = n_workers
      master%para_env => para_env
      master%globenv => globenv
      ALLOCATE (master%queued_commands(master%n_workers))
      master%iw = cp_print_key_unit_nr(logger, master%swarm_section, &
                                       "PRINT%MASTER_RUN_INFO", extension=".masterLog")

      CALL section_vals_val_get(master%swarm_section, "BEHAVIOR", i_val=master%behavior)

      ! uses logger%iter_info%project_name to construct filename
      master%comlog_unit = cp_print_key_unit_nr(logger, master%swarm_section, "PRINT%COMMUNICATION_LOG", &
                                                !middle_name="comlog", extension=".xyz", &
                                                extension=".comlog", &
                                                file_action="WRITE", file_position="REWIND")

      CALL section_vals_val_get(master%swarm_section, "MAX_ITER", i_val=master%max_iter)

      SELECT CASE (master%behavior)
      CASE (swarm_do_glbopt)
         ALLOCATE (master%glbopt)
         CALL glbopt_master_init(master%glbopt, para_env, root_section, n_workers, master%iw)
      CASE DEFAULT
         CALL cp__b("swarm/swarm_master.F",120,"got unknown behavior")
      END SELECT

      CALL replay_comlog(master)
   END SUBROUTINE swarm_master_init

! **************************************************************************************************
!> \brief Helper routine for swarm_master_init, restarts a calculation
!> \param master ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE replay_comlog(master)
      TYPE(swarm_master_type)                            :: master

      CHARACTER(len=*), PARAMETER :: routineN = 'replay_comlog', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: command_log
      INTEGER                                            :: handle, i, worker_id
      LOGICAL                                            :: at_end, explicit
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(swarm_message_type)                           :: cmd_log, report_log
      TYPE(swarm_message_type), &
         DIMENSION(master%n_workers)                     :: last_commands
      TYPE(swarm_message_type), POINTER                  :: cmd_now

      NULLIFY (parser)

      ! Initialize parser for trajectory
      CALL section_vals_val_get(master%swarm_section, "REPLAY_COMMUNICATION_LOG", &
                                c_val=filename, explicit=explicit)

      IF (.NOT. explicit) RETURN
      IF (master%iw > 0) WRITE (master%iw, '(A,A)') &
         " SWARM| Starting replay of communication-log: ", TRIM(filename)

      CALL timeset("swarm_master_replay_comlog", handle)
      CALL parser_create(parser, filename, para_env=master%para_env)

      at_end = .FALSE.
      DO
         CALL swarm_message_file_read(report_log, parser, at_end)
         IF (at_end) EXIT

         CALL swarm_message_file_read(cmd_log, parser, at_end)
         IF (at_end) EXIT

         ALLOCATE (cmd_now)
         CALL swarm_master_steer(master, report_log, cmd_now)

         !TODO: maybe we should just exit the loop instead of stopping?
         CALL swarm_message_get(cmd_log, "command", command_log)
         IF (TRIM(command_log) /= "shutdown") THEN
            IF (.NOT. commands_equal(cmd_now, cmd_log, master%iw)) CALL cp__b("swarm/swarm_master.F",173,"wrong behaviour")
         END IF

         CALL swarm_message_free(cmd_log)
         CALL swarm_message_free(report_log)
         CALL swarm_message_get(cmd_now, "worker_id", worker_id)
         CALL swarm_message_free(last_commands(worker_id))
         last_commands(worker_id) = cmd_now
         DEALLOCATE (cmd_now)
      END DO

      CALL swarm_message_free(report_log) !don't worry about double-frees
      CALL swarm_message_free(cmd_log)

      IF (master%iw > 0) WRITE (master%iw, '(A,A)') &
         " SWARM| Reached end of communication log. Queueing last commands."

      DO i = 1, master%n_workers
         ALLOCATE (master%queued_commands(i)%p)
         master%queued_commands(i)%p = last_commands(i)
      END DO

      CALL parser_release(parser)
      CALL timestop(handle)
   END SUBROUTINE replay_comlog

! **************************************************************************************************
!> \brief Helper routine for replay_comlog, compares two commands
!> \param cmd1 ...
!> \param cmd2 ...
!> \param iw ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION commands_equal(cmd1, cmd2, iw) RESULT(res)
      TYPE(swarm_message_type)                           :: cmd1, cmd2
      INTEGER                                            :: iw
      LOGICAL                                            :: res

      res = swarm_message_equal(cmd1, cmd2)
      IF (.NOT. res .AND. iw > 0) THEN
         WRITE (iw, *) "Command 1:"
         CALL swarm_message_file_write(cmd1, iw)
         WRITE (iw, *) "Command 2:"
         CALL swarm_message_file_write(cmd2, iw)
      END IF
   END FUNCTION commands_equal

! **************************************************************************************************
!> \brief Central steering routine of the swarm master
!> \param master ...
!> \param report ...
!> \param cmd ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_master_steer(master, report, cmd)
      TYPE(swarm_master_type), INTENT(INOUT)             :: master
      TYPE(swarm_message_type), INTENT(IN)               :: report
      TYPE(swarm_message_type), INTENT(OUT)              :: cmd

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_master_steer', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: command, status
      INTEGER                                            :: handle, worker_id
      LOGICAL                                            :: should_stop

      should_stop = .FALSE.

      CALL timeset("swarm_master_steer", handle)

      ! First check if there are queued commands for this worker
      CALL swarm_message_get(report, "worker_id", worker_id)

      IF (ASSOCIATED(master%queued_commands(worker_id)%p)) THEN
         cmd = master%queued_commands(worker_id)%p
         DEALLOCATE (master%queued_commands(worker_id)%p)
         IF (master%iw > 0) WRITE (master%iw, '(A,A,A,I9,1X,A)') ' SWARM| ', &
            REPEAT("*", 9), " Sending out queued command to worker: ", &
            worker_id, REPEAT("*", 9)
         CALL timestop(handle)
         RETURN
      END IF

      IF (.NOT. master%ignore_last_iteration) THEN
         ! There are no queued commands. Do the normal processing.
         master%i_iteration = master%i_iteration+1

         IF (master%iw > 0) WRITE (master%iw, '(A,A,1X,I8,A,A)') ' SWARM| ', REPEAT("*", 15), &
            master%i_iteration, ' Master / Worker Communication  ', REPEAT("*", 15)
      ENDIF

      IF (master%i_iteration >= master%max_iter .AND. .NOT. master%should_stop) THEN
         IF (master%iw > 0) WRITE (master%iw, '(A)') " SWARM| Reached MAX_ITER. Quitting."
         master%should_stop = .TRUE.
      ENDIF

      IF (.NOT. master%should_stop) THEN
         CALL external_control(master%should_stop, "SWARM", master%globenv)
         IF (master%should_stop .AND. master%iw > 0) &
            WRITE (master%iw, *) " SWARM| Received stop from external_control. Quitting."
      END IF

      !IF(unit > 0) &

      IF (master%should_stop) THEN
         CALL swarm_message_add(cmd, "command", "shutdown")
         IF (master%iw > 0) WRITE (master%iw, '(1X,A,T71,I10)') &
            "SWARM| Sending shutdown command to worker", worker_id
      ELSE
         SELECT CASE (master%behavior)
         CASE (swarm_do_glbopt)
            CALL glbopt_master_steer(master%glbopt, report, cmd, should_stop)
         CASE DEFAULT
            CALL cp__b("swarm/swarm_master.F",287,"got unknown behavior")
         END SELECT

         IF (should_stop) THEN
            CALL swarm_message_free(cmd)
            CALL swarm_message_add(cmd, "command", "shutdown") !overwrite command
            IF (master%iw > 0) WRITE (master%iw, '(1X,A,T71,I10)') &
               "SWARM| Sending shutdown command to worker", worker_id
            master%should_stop = .TRUE.
         ENDIF
      ENDIF

      CALL swarm_message_add(cmd, "worker_id", worker_id)

      ! Don't pollute comlog with "continue waiting"-commands.
      CALL swarm_message_get(report, "status", status)
      CALL swarm_message_get(cmd, "command", command)
      IF (TRIM(status) == "wait_done") master%n_waiting = master%n_waiting-1
      IF (TRIM(command) == "wait") master%n_waiting = master%n_waiting+1
      IF (master%n_waiting < 0) CALL cp__b("swarm/swarm_master.F",306,"master%n_waiting < 0")
      IF (TRIM(status) /= "wait_done" .OR. TRIM(command) /= "wait") THEN
         CALL swarm_message_file_write(report, master%comlog_unit)
         CALL swarm_message_file_write(cmd, master%comlog_unit)
         IF (master%n_waiting > 0 .AND. master%iw > 0) WRITE (master%iw, '(1X,A,T71,I10)') &
            "SWARM| Number of waiting workers:", master%n_waiting
         master%ignore_last_iteration = .FALSE.
      ELSE
         master%ignore_last_iteration = .TRUE.
      END IF
      CALL timestop(handle)
   END SUBROUTINE swarm_master_steer

! **************************************************************************************************
!> \brief Finalizes the swarm master
!> \param master ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_master_finalize(master)
      TYPE(swarm_master_type)                            :: master

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_master_finalize', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      IF (master%iw > 0) THEN
         WRITE (master%iw, "(1X,A,T71,I10)") "SWARM| Total number of iterations ", master%i_iteration
         WRITE (master%iw, "(A)") " SWARM| Shutting down the master."
      ENDIF

      SELECT CASE (master%behavior)
      CASE (swarm_do_glbopt)
         CALL glbopt_master_finalize(master%glbopt)
         DEALLOCATE (master%glbopt)
      CASE DEFAULT
         CALL cp__b("swarm/swarm_master.F",342,"got unknown behavior")
      END SELECT

      DEALLOCATE (master%queued_commands)

      logger => cp_get_default_logger()
      CALL cp_print_key_finished_output(master%iw, logger, &
                                        master%swarm_section, "PRINT%MASTER_RUN_INFO")
      CALL cp_print_key_finished_output(master%comlog_unit, logger, &
                                        master%swarm_section, "PRINT%COMMUNICATION_LOG")

      !CALL rm_timer_env() !pops the top-most timer
   END SUBROUTINE swarm_master_finalize

END MODULE swarm_master

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A common interface for passing a callback into the md_run loop.
!> \par History
!> \author Ole
! **************************************************************************************************
MODULE mdctrl_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   TYPE glbopt_mdctrl_data_type
      INTEGER                                    :: md_bump_counter
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE   :: epot_history
      INTEGER                                    :: output_unit
      INTEGER                                    :: itimes
      INTEGER                                    :: bump_steps_upwards
      INTEGER                                    :: bump_steps_downwards
      INTEGER                                    :: md_bumps_max
   END TYPE glbopt_mdctrl_data_type

   TYPE mdctrl_type
      TYPE(glbopt_mdctrl_data_type), POINTER                 :: glbopt => Null()
      !... and possible more in the future
   END TYPE mdctrl_type

   PUBLIC :: mdctrl_type, glbopt_mdctrl_data_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mdctrl_types'

END MODULE mdctrl_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods to perform on the fly statistical analysis of data
!>      -) Schiferl and Wallace, J. Chem. Phys. 83 (10) 1985
!> \author Teodoro Laino (01.2007) [tlaino]
!> \par History
!>         - Teodoro Laino (10.2008) [tlaino] - University of Zurich
!>           module made publicly available
! **************************************************************************************************
MODULE statistical_methods
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE erf_fn,                          ONLY: erf
   USE global_types,                    ONLY: global_environment_type
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: next_random_number
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'statistical_methods'
   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.
   INTEGER, PARAMETER, PUBLIC           :: min_sample_size = 20
   PUBLIC :: sw_test, &
             k_test, &
             vn_test

CONTAINS

! **************************************************************************************************
!> \brief Shapiro - Wilk's test or W-statistic to test normality of a distribution
!>      R94  APPL. STATIST. (1995) VOL.44, NO.4
!>      Calculates the Shapiro-Wilk W test and its significance level
!> \param ix ...
!> \param n ...
!> \param w ...
!> \param pw ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE sw_test(ix, n, w, pw)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ix
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(OUT)                         :: w, pw

      CHARACTER(len=*), PARAMETER :: routineN = 'sw_test', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: c1(6) = (/0.000000_dp, 0.221157_dp, -0.147981_dp, -2.071190_dp, &
         4.434685_dp, -2.706056_dp/), c2(6) = (/0.000000_dp, 0.042981_dp, -0.293762_dp, &
         -1.752461_dp, 5.682633_dp, -3.582633_dp/), &
         c3(4) = (/0.544000_dp, -0.399780_dp, 0.025054_dp, -0.6714E-3_dp/), &
         c4(4) = (/1.3822_dp, -0.77857_dp, 0.062767_dp, -0.0020322_dp/), &
         c5(4) = (/-1.5861_dp, -0.31082_dp, -0.083751_dp, 0.0038915_dp/), &
         c6(3) = (/-0.4803_dp, -0.082676_dp, 0.0030302_dp/), g(2) = (/-2.273_dp, 0.459_dp/), &
         one = 1.0_dp, pi6 = 1.909859_dp, qtr = 0.25_dp, small = EPSILON(0.0_dp), &
         sqrth = 0.70711_dp
      REAL(KIND=dp), PARAMETER :: stqr = 1.047198_dp, th = 0.375_dp, two = 2.0_dp, zero = 0.0_dp

      INTEGER                                            :: i, i1, j, n2, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: itmp
      LOGICAL                                            :: failure
      REAL(KIND=dp)                                      :: a1, a2, an, an25, asa, fac, gamma, m, &
                                                            range, rsn, s, sa, sax, ssa, ssassx, &
                                                            ssumm2, ssx, summ2, sx, w1, xi, xsx, &
                                                            xx, y
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: a, x

      failure = .FALSE.
      output_unit = cp_logger_get_default_io_unit()
      ! Check for  N < 3
      IF (n < 3 .OR. n > 5000) THEN
         IF (output_unit > 0) WRITE (output_unit, '(A)') &
            "Shapiro Wilk test: number of points less than 3 or greated than 5000."
         IF (output_unit > 0) WRITE (output_unit, '(A)') &
            "Not able to perform the test!"
      END IF
      ! Sort input array of data in ascending order
      IF (MOD(n, 2) == 0) THEN
         n2 = n/2
      ELSE
         n2 = (n-1)/2
      END IF
      ALLOCATE (x(n))
      ALLOCATE (itmp(n))
      ALLOCATE (a(n2))
      x(:) = ix
      CALL sort(x, n, itmp)
      ! Check for zero range
      range = x(n)-x(1)
      IF (range < small) failure = .TRUE.
      IF (failure .AND. (output_unit > 0)) THEN
         WRITE (output_unit, '(A)') "Shapiro Wilk test: two data points are numerically identical."
         WRITE (output_unit, '(A)') "Not able to perform the test!"
      END IF
      pw = one
      w = one
      an = n
      ! Calculates coefficients for the test
      IF (n == 3) THEN
         a(1) = sqrth
      ELSE
         an25 = an+qtr
         summ2 = zero
         DO i = 1, n2
            CALL ppnd7((i-th)/an25, a(i))
            summ2 = summ2+a(i)**2
         END DO
         summ2 = summ2*two
         ssumm2 = SQRT(summ2)
         rsn = one/SQRT(an)
         a1 = poly(c1, 6, rsn)-a(1)/ssumm2
         ! Normalize coefficients
         IF (n > 5) THEN
            i1 = 3
            a2 = -a(2)/ssumm2+poly(c2, 6, rsn)
            fac = SQRT((summ2-two*a(1)**2-two*a(2)**2)/(one-two*a1**2-two*a2**2))
            a(1) = a1
            a(2) = a2
         ELSE
            i1 = 2
            fac = SQRT((summ2-two*a(1)**2)/(one-two*a1**2))
            a(1) = a1
         END IF
         DO i = i1, n2
            a(i) = -a(i)/fac
         END DO
      END IF
      ! scaled X
      xx = x(1)/range
      sx = xx
      sa = -a(1)
      j = n-1
      DO i = 2, n
         xi = x(i)/range
         sx = sx+xi
         IF (i /= j) sa = sa+SIGN(1, i-j)*a(MIN(i, j))
         xx = xi
         j = j-1
      END DO
      ! Calculate W statistic as squared correlation
      ! between data and coefficients
      sa = sa/n
      sx = sx/n
      ssa = zero
      ssx = zero
      sax = zero
      j = n
      DO i = 1, n
         IF (i /= j) THEN
            asa = SIGN(1, i-j)*a(MIN(i, j))-sa
         ELSE
            asa = -sa
         END IF
         xsx = x(i)/range-sx
         ssa = ssa+asa*asa
         ssx = ssx+xsx*xsx
         sax = sax+asa*xsx
         j = j-1
      END DO
      ! W1 equals (1-W) calculated to avoid excessive rounding error
      ! for W very near 1 (a potential problem in very large samples)
      ssassx = SQRT(ssa*ssx)
      w1 = (ssassx-sax)*(ssassx+sax)/(ssa*ssx)
      w = one-w1
      ! Calculate significance level for W (exact for N=3)
      IF (n == 3) THEN
         pw = pi6*(ASIN(SQRT(w))-stqr)
      ELSE
         y = LOG(w1)
         xx = LOG(an)
         m = zero
         s = one
         IF (n <= 11) THEN
            gamma = poly(g, 2, an)
            IF (y >= gamma) THEN
               pw = small
            ELSE
               y = -LOG(gamma-y)
               m = poly(c3, 4, an)
               s = EXP(poly(c4, 4, an))
               pw = alnorm((y-m)/s, .TRUE.)
            END IF
         ELSE
            m = poly(c5, 4, xx)
            s = EXP(poly(c6, 3, xx))
            pw = alnorm((y-m)/s, .TRUE.)
         END IF
      END IF
      DEALLOCATE (x)
      DEALLOCATE (itmp)
      DEALLOCATE (a)

   END SUBROUTINE sw_test

! **************************************************************************************************
!> \brief Produces the normal deviate Z corresponding to a given lower tail area of P
!>      Z is accurate to about 1 part in 10**7.
!>      AS241  APPL. STATIST. (1988) VOL. 37, NO. 3, 477- 484.
!> \param p ...
!> \param normal_dev ...
!> \par History
!>      Original version by Alain J. Miller - 1996
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE ppnd7(p, normal_dev)
      REAL(KIND=dp), INTENT(IN)                          :: p
      REAL(KIND=dp), INTENT(OUT)                         :: normal_dev

      REAL(KIND=dp), PARAMETER :: a0 = 3.3871327179E+00_dp, a1 = 5.0434271938E+01_dp, &
         a2 = 1.5929113202E+02_dp, a3 = 5.9109374720E+01_dp, b1 = 1.7895169469E+01_dp, &
         b2 = 7.8757757664E+01_dp, b3 = 6.7187563600E+01_dp, c0 = 1.4234372777E+00_dp, &
         c1 = 2.7568153900E+00_dp, c2 = 1.3067284816E+00_dp, c3 = 1.7023821103E-01_dp, &
         const1 = 0.180625_dp, const2 = 1.6_dp, d1 = 7.3700164250E-01_dp, &
         d2 = 1.2021132975E-01_dp, e0 = 6.6579051150E+00_dp, e1 = 3.0812263860E+00_dp, &
         e2 = 4.2868294337E-01_dp, e3 = 1.7337203997E-02_dp, f1 = 2.4197894225E-01_dp, &
         f2 = 1.2258202635E-02_dp, half = 0.5_dp, one = 1.0_dp
      REAL(KIND=dp), PARAMETER :: split1 = 0.425_dp, split2 = 5.0_dp, zero = 0.0_dp

      REAL(KIND=dp)                                      :: q, r

      q = p-half
      IF (ABS(q) <= split1) THEN
         r = const1-q*q
         normal_dev = q*(((a3*r+a2)*r+a1)*r+a0)/ &
                      (((b3*r+b2)*r+b1)*r+one)
         RETURN
      ELSE
         IF (q < zero) THEN
            r = p
         ELSE
            r = one-p
         END IF
         IF (r <= zero) THEN
            normal_dev = zero
            RETURN
         END IF
         r = SQRT(-LOG(r))
         IF (r <= split2) THEN
            r = r-const2
            normal_dev = (((c3*r+c2)*r+c1)*r+c0)/((d2*r+d1)*r+one)
         ELSE
            r = r-split2
            normal_dev = (((e3*r+e2)*r+e1)*r+e0)/((f2*r+f1)*r+one)
         END IF
         IF (q < zero) normal_dev = -normal_dev
         RETURN
      END IF
   END SUBROUTINE ppnd7

! **************************************************************************************************
!> \brief Evaluates the tail area of the standardised normal curve
!>      from x to infinity if upper is .true. or
!>      from minus infinity to x if upper is .false.
!>      AS66 Applied Statistics (1973) vol.22, no.3
!> \param x ...
!> \param upper ...
!> \retval fn_val ...
!> \par History
!>      Original version by Alain J. Miller - 1996
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   FUNCTION alnorm(x, upper) RESULT(fn_val)
      REAL(KIND=dp), INTENT(IN)                          :: x
      LOGICAL, INTENT(IN)                                :: upper
      REAL(KIND=dp)                                      :: fn_val

      REAL(KIND=dp), PARAMETER :: a1 = 5.75885480458_dp, a2 = 2.62433121679_dp, &
         a3 = 5.92885724438_dp, b1 = -29.8213557807_dp, b2 = 48.6959930692_dp, c1 = -3.8052E-8_dp, &
         c2 = 3.98064794E-4_dp, c3 = -0.151679116635_dp, c4 = 4.8385912808_dp, &
         c5 = 0.742380924027_dp, c6 = 3.99019417011_dp, con = 1.28_dp, d1 = 1.00000615302_dp, &
         d2 = 1.98615381364_dp, d3 = 5.29330324926_dp, d4 = -15.1508972451_dp, &
         d5 = 30.789933034_dp, half = 0.5_dp, ltone = 7.0_dp, one = 1.0_dp, p = 0.398942280444_dp, &
         q = 0.39990348504_dp, r = 0.398942280385_dp, utzero = 18.66_dp, zero = 0.0_dp

      LOGICAL                                            :: up
      REAL(KIND=dp)                                      :: y, z

      up = upper
      z = x
      IF (z < zero) THEN
         up = .NOT. up
         z = -z
      ENDIF
      IF (.NOT. (z <= ltone .OR. up .AND. z <= utzero)) THEN
         fn_val = zero
         IF (.NOT. up) fn_val = one-fn_val
         RETURN
      ENDIF
      y = half*z*z
      IF (z <= con) THEN
         fn_val = r*EXP(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
      ELSE
         fn_val = half-z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
      ENDIF
      IF (.NOT. up) fn_val = one-fn_val

   END FUNCTION alnorm

! **************************************************************************************************
!> \brief Calculates the algebraic polynomial of order nored-1 with
!>      array of coefficients c.  Zero order coefficient is c(1)
!>      AS 181.2   Appl. Statist.  (1982) Vol. 31, No. 2
!> \param c ...
!> \param nord ...
!> \param x ...
!> \retval fn_val ...
!> \par History
!>      Original version by Alain J. Miller - 1996
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   FUNCTION poly(c, nord, x) RESULT(fn_val)

      REAL(KIND=dp), INTENT(IN)                          :: c(:)
      INTEGER, INTENT(IN)                                :: nord
      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: fn_val

      INTEGER                                            :: i, j, n2
      REAL(KIND=dp)                                      :: p

      fn_val = c(1)
      IF (nord == 1) RETURN
      p = x*c(nord)
      IF (nord == 2) THEN
         fn_val = fn_val+p
         RETURN
      ENDIF
      n2 = nord-2
      j = n2+1
      DO i = 1, n2
         p = (p+c(j))*x
         j = j-1
      END DO
      fn_val = fn_val+p
   END FUNCTION poly

! **************************************************************************************************
!> \brief Kandall's test for correlation
!> \param xdata ...
!> \param istart ...
!> \param n ...
!> \param tau ...
!> \param z ...
!> \param prob ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
!> \note
!>      tau:  Kendall's Tau
!>      z:    number of std devs from 0 of tau
!>      prob: tau's probability
! **************************************************************************************************
   SUBROUTINE k_test(xdata, istart, n, tau, z, prob)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdata
      INTEGER, INTENT(IN)                                :: istart, n
      REAL(KIND=dp)                                      :: tau, z, prob

      CHARACTER(len=*), PARAMETER :: routineN = 'k_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: is, j, k, nt
      REAL(KIND=dp)                                      :: a1, var

      nt = n-istart+1
      IF (nt .GE. min_sample_size) THEN
         is = 0
         DO j = istart, n-1
            DO k = j+1, n
               a1 = xdata(j)-xdata(k)
               IF (a1 .GT. 0.0_dp) THEN
                  is = is+1
               ELSE IF (a1 .LT. 0.0_dp) THEN
                  is = is-1
               END IF
            END DO
         END DO
         tau = REAL(is, KIND=dp)
         var = REAL(nt, KIND=dp)*REAL(nt-1, KIND=dp)*REAL(2*nt+5, KIND=dp)/18.0_dp
         z = tau/SQRT(var)
         prob = erf(ABS(z)/SQRT(2.0_dp))
      ELSE
         tau = 0.0_dp
         z = 0.0_dp
         prob = 1.0_dp
      END IF
   END SUBROUTINE k_test

! **************************************************************************************************
!> \brief Von Neumann test for serial correlation
!> \param xdata ...
!> \param n ...
!> \param r ...
!> \param u ...
!> \param prob ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE vn_test(xdata, n, r, u, prob)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdata
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp)                                      :: r, u, prob

      CHARACTER(len=*), PARAMETER :: routineN = 'vn_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: q, s, var, x

      IF (n .GE. min_sample_size) THEN
         x = 0.0_dp
         q = 0.0_dp
         s = 0.0_dp
         DO i = 1, n-1
            x = x+xdata(i)
            q = q+(xdata(i+1)-xdata(i))**2
         END DO
         x = x+xdata(n)
         x = x/REAL(n, KIND=dp)
         DO i = 1, n
            s = s+(xdata(i)-x)**2
         END DO
         s = s/REAL(n-1, KIND=dp)
         q = q/REAL(2*(n-1), KIND=dp)
         r = q/s
         var = SQRT(1.0_dp/REAL(n+1, KIND=dp)*(1.0_dp+1.0_dp/REAL(n-1, KIND=dp)))
         u = (r-1.0_dp)/var
         prob = erf(ABS(u)/SQRT(2.0_dp))
      ELSE
         r = 0.0_dp
         u = 0.0_dp
         prob = 1.0_dp
      END IF

   END SUBROUTINE vn_test

! **************************************************************************************************
!> \brief Performs tests on statistical methods
!>      Debug use only
!> \param xdata ...
!> \param globenv ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE tests(xdata, globenv)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdata
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'tests', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      REAL(KINd=dp)                                      :: prob, pw, r, tau, u, w, z
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ydata

      IF (debug_this_module) THEN
         n = 50 ! original sample size
         NULLIFY (xdata)
         ALLOCATE (xdata(n))
         DO i = 1, 10
            xdata(i) = 5.0_dp-REAL(i, KIND=dp)/2.0_dp+0.1* &
                       next_random_number(globenv%gaussian_rng_stream)
            WRITE (3, *) xdata(i)
         END DO
         DO i = 11, n
            xdata(i) = 0.1*next_random_number(globenv%gaussian_rng_stream)
         END DO

         ! Test for trend
         DO i = 1, n
            CALL k_test(xdata, i, n, tau, z, prob)
            IF (prob <= 0.2_dp) EXIT
         END DO
         WRITE (*, *) "Mann-Kendall test", i

         ! Test for normality distribution and for serial correlation
         DO i = 1, n
            ALLOCATE (ydata(n-i+1))
            ydata = xdata(i:n)
            CALL sw_test(ydata, n-i+1, w, pw)
            CALL vn_test(ydata, n-i+1, r, u, prob)
            WRITE (*, *) "Shapiro Wilks test", i, w, pw
            WRITE (*, *) "Von Neu", i, r, u, prob
            DEALLOCATE (ydata)
         END DO

         DEALLOCATE (xdata)
      END IF
   END SUBROUTINE tests

END MODULE statistical_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods to perform free energy and free energy derivatives calculations
!> \author Teodoro Laino (01.2007) [tlaino]
! **************************************************************************************************
MODULE free_energy_methods
   USE colvar_methods,                  ONLY: colvar_eval_glob_f
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE fparser,                         ONLY: evalf,&
                                              evalfd,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE free_energy_types,               ONLY: free_energy_type,&
                                              ui_var_type
   USE input_constants,                 ONLY: do_fe_ac,&
                                              do_fe_ui
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathlib,                         ONLY: diamat_all
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE memory_utilities,                ONLY: reallocate
   USE simpar_types,                    ONLY: simpar_type
   USE statistical_methods,             ONLY: k_test,&
                                              min_sample_size,&
                                              sw_test,&
                                              vn_test
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'free_energy_methods'
   PUBLIC :: free_energy_evaluate

CONTAINS

! **************************************************************************************************
!> \brief Main driver for free energy calculations
!>      In this routine we handle specifically biased MD.
!> \param md_env ...
!> \param converged ...
!> \param fe_section ...
!> \par History
!>      Teodoro Laino (01.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE free_energy_evaluate(md_env, converged, fe_section)
      TYPE(md_environment_type), POINTER                 :: md_env
      LOGICAL, INTENT(OUT)                               :: converged
      TYPE(section_vals_type), POINTER                   :: fe_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'free_energy_evaluate', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: my_par
      INTEGER                                            :: handle, ic, icolvar, nforce_eval, &
                                                            output_unit, stat_sign_points
      INTEGER, POINTER                                   :: istep
      REAL(KIND=dp)                                      :: beta, dx, lerr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_val
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(free_energy_type), POINTER                    :: fe_env
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(ui_var_type), POINTER                         :: cv

      NULLIFY (force_env, istep, subsys, cv, simpar)
      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)
      converged = .FALSE.
      CALL get_md_env(md_env, force_env=force_env, fe_env=fe_env, simpar=simpar, &
                      itimes=istep)
      ! Metadynamics is also a free energy calculation but is handled in a different
      ! module.
      IF (.NOT. ASSOCIATED(force_env%meta_env) .AND. ASSOCIATED(fe_env)) THEN
         SELECT CASE (fe_env%type)
         CASE (do_fe_ui)
            ! Umbrella Integration..
            CALL force_env_get(force_env, subsys=subsys)
            fe_env%nr_points = fe_env%nr_points+1
            output_unit = cp_logger_get_default_io_unit(logger)
            DO ic = 1, fe_env%ncolvar
               cv => fe_env%uivar(ic)
               icolvar = cv%icolvar
               CALL colvar_eval_glob_f(icolvar, force_env)
               CALL reallocate(cv%ss, 1, fe_env%nr_points)
               cv%ss(fe_env%nr_points) = subsys%colvar_p(icolvar)%colvar%ss
               IF (output_unit > 0) THEN
                  WRITE (output_unit, *) "COLVAR::", cv%ss(fe_env%nr_points)
               END IF
            END DO
            stat_sign_points = fe_env%nr_points-fe_env%nr_rejected
            IF (output_unit > 0) THEN
               WRITE (output_unit, *) fe_env%nr_points, stat_sign_points
            END IF
            ! Start statistical analysis when enough CG data points have been collected
            IF ((fe_env%conv_par%cg_width*fe_env%conv_par%cg_points <= stat_sign_points) .AND. &
                (MOD(stat_sign_points, fe_env%conv_par%cg_width) == 0)) THEN
               output_unit = cp_print_key_unit_nr(logger, fe_section, "FREE_ENERGY_INFO", &
                                                  extension=".FreeEnergyLog", log_filename=.FALSE.)
               CALL print_fe_prolog(output_unit)
               ! Trend test..  recomputes the number of statistically significant points..
               CALL ui_check_trend(fe_env, fe_env%conv_par%test_k, stat_sign_points, output_unit)
               stat_sign_points = fe_env%nr_points-fe_env%nr_rejected
               ! Normality and serial correlation tests..
               IF (fe_env%conv_par%cg_width*fe_env%conv_par%cg_points <= stat_sign_points .AND. &
                   fe_env%conv_par%test_k) THEN
                  ! Statistical tests
                  CALL ui_check_convergence(fe_env, converged, stat_sign_points, output_unit)
               END IF
               CALL print_fe_epilog(output_unit)
               CALL cp_print_key_finished_output(output_unit, logger, fe_section, "FREE_ENERGY_INFO")
            END IF
         CASE (do_fe_ac)
            CALL initf(2)
            ! Alchemical Changes
            IF (.NOT. ASSOCIATED(force_env%mixed_env)) &
               CALL cp_abort(cp__l("motion/free_energy_methods.F",139), &
                             'ASSERTION (cond) failed at line '//cp_to_string(140)// &
                             ' Free Energy calculations require the definition of a mixed env!')
            my_par => force_env%mixed_env%par
            my_val => force_env%mixed_env%val
            dx = force_env%mixed_env%dx
            lerr = force_env%mixed_env%lerr
            coupling_function = force_env%mixed_env%coupling_function
            beta = 1/simpar%temp_ext
            CALL parsef(1, TRIM(coupling_function), my_par)
            nforce_eval = SIZE(force_env%sub_force_env)
            CALL dump_ac_info(my_val, my_par, dx, lerr, fe_section, nforce_eval, &
                              fe_env%covmx, istep, beta)
            CALL finalizef()
         CASE DEFAULT
            ! Do Nothing
         END SELECT
      END IF
      CALL timestop(handle)

   END SUBROUTINE free_energy_evaluate

! **************************************************************************************************
!> \brief Print prolog of free energy output section
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE print_fe_prolog(output_unit)
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_fe_prolog', &
         routineP = moduleN//':'//routineN

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,79("*"))')
         WRITE (output_unit, '(T30,"FREE ENERGY CALCULATION",/)')
      END IF
   END SUBROUTINE print_fe_prolog

! **************************************************************************************************
!> \brief Print epilog of free energy output section
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE print_fe_epilog(output_unit)
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_fe_epilog', &
         routineP = moduleN//':'//routineN

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,79("*"),/)')
      END IF
   END SUBROUTINE print_fe_epilog

! **************************************************************************************************
!> \brief Test for trend in coarse grained data set
!> \param fe_env ...
!> \param trend_free ...
!> \param nr_points ...
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (01.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE ui_check_trend(fe_env, trend_free, nr_points, output_unit)
      TYPE(free_energy_type), POINTER                    :: fe_env
      LOGICAL, INTENT(OUT)                               :: trend_free
      INTEGER, INTENT(IN)                                :: nr_points, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ui_check_trend', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, j, k, my_reject, ncolvar, &
                                                            ng_points, rejected_points
      LOGICAL                                            :: test_avg, test_std
      REAL(KIND=dp)                                      :: prob, tau, z
      REAL(KIND=dp), DIMENSION(:), POINTER               :: wrk

      CALL timeset(routineN, handle)
      trend_free = .FALSE.
      test_avg = .TRUE.
      test_std = .TRUE.
      ncolvar = fe_env%ncolvar
      ! Number of coarse grained points
      IF (output_unit > 0) THEN
         WRITE (output_unit, *) nr_points, fe_env%conv_par%cg_width
      END IF
      ng_points = nr_points/fe_env%conv_par%cg_width
      my_reject = 0
      ! Allocate storage
      CALL create_tmp_data(fe_env, wrk, ng_points, ncolvar)
      ! Compute the Coarse Grained data set using a reverse cumulative strategy
      CALL create_csg_data(fe_env, ng_points, output_unit)
      ! Test on coarse grained average
      DO j = 1, ncolvar
         ii = 1
         DO i = ng_points, 1, -1
            wrk(ii) = fe_env%cg_data(i)%avg(j)
            ii = ii+1
         END DO
         DO i = my_reject+1, ng_points
            IF ((ng_points-my_reject) .LT. min_sample_size) THEN
               my_reject = MAX(0, my_reject-1)
               test_avg = .FALSE.
               EXIT
            END IF
            CALL k_test(wrk, my_reject+1, ng_points, tau, z, prob)
            PRINT*, prob, fe_env%conv_par%k_conf_lm
            IF (prob < fe_env%conv_par%k_conf_lm) EXIT
            my_reject = my_reject+1
         END DO
         my_reject = MIN(ng_points, my_reject)
      END DO
      rejected_points = my_reject*fe_env%conv_par%cg_width
      ! Print some info
      IF (output_unit > 0) THEN
         WRITE (output_unit, *) "Kendall trend test (Average)", test_avg, &
            "number of points rejected:", rejected_points+fe_env%nr_rejected
         WRITE (output_unit, *) "Reject Nr.", my_reject, " coarse grained points testing average"
      END IF
      ! Test on coarse grained covariance matrix
      DO j = 1, ncolvar
         DO k = j, ncolvar
            ii = 1
            DO i = ng_points, 1, -1
               wrk(ii) = fe_env%cg_data(i)%var(j, k)
               ii = ii+1
            END DO
            DO i = my_reject+1, ng_points
               IF ((ng_points-my_reject) .LT. min_sample_size) THEN
                  my_reject = MAX(0, my_reject-1)
                  test_std = .FALSE.
                  EXIT
               END IF
               CALL k_test(wrk, my_reject+1, ng_points, tau, z, prob)
               PRINT*, prob, fe_env%conv_par%k_conf_lm
               IF (prob < fe_env%conv_par%k_conf_lm) EXIT
               my_reject = my_reject+1
            END DO
            my_reject = MIN(ng_points, my_reject)
         END DO
      END DO
      rejected_points = my_reject*fe_env%conv_par%cg_width
      fe_env%nr_rejected = fe_env%nr_rejected+rejected_points
      trend_free = test_avg .AND. test_std
      ! Print some info
      IF (output_unit > 0) THEN
         WRITE (output_unit, *) "Kendall trend test (Std. Dev.)", test_std, &
            "number of points rejected:", fe_env%nr_rejected
         WRITE (output_unit, *) "Reject Nr.", my_reject, " coarse grained points testing standard dev."
         WRITE (output_unit, *) "Kendall test passed:", trend_free
      END IF
      ! Release storage
      CALL destroy_tmp_data(fe_env, wrk, ng_points)
      CALL timestop(handle)
   END SUBROUTINE ui_check_trend

! **************************************************************************************************
!> \brief Creates temporary data structures
!> \param fe_env ...
!> \param wrk ...
!> \param ng_points ...
!> \param ncolvar ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE create_tmp_data(fe_env, wrk, ng_points, ncolvar)
      TYPE(free_energy_type), POINTER                    :: fe_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: wrk
      INTEGER, INTENT(IN)                                :: ng_points, ncolvar

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_tmp_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      ALLOCATE (fe_env%cg_data(ng_points))
      DO i = 1, ng_points
         ALLOCATE (fe_env%cg_data(i)%avg(ncolvar))
         ALLOCATE (fe_env%cg_data(i)%var(ncolvar, ncolvar))
      END DO
      IF (PRESENT(wrk)) THEN
         ALLOCATE (wrk(ng_points))
      END IF
   END SUBROUTINE create_tmp_data

! **************************************************************************************************
!> \brief Destroys temporary data structures
!> \param fe_env ...
!> \param wrk ...
!> \param ng_points ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE destroy_tmp_data(fe_env, wrk, ng_points)
      TYPE(free_energy_type), POINTER                    :: fe_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: wrk
      INTEGER, INTENT(IN)                                :: ng_points

      CHARACTER(LEN=*), PARAMETER :: routineN = 'destroy_tmp_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, ng_points
         DEALLOCATE (fe_env%cg_data(i)%avg)
         DEALLOCATE (fe_env%cg_data(i)%var)
      END DO
      DEALLOCATE (fe_env%cg_data)
      IF (PRESENT(wrk)) THEN
         DEALLOCATE (wrk)
      END IF
   END SUBROUTINE destroy_tmp_data

! **************************************************************************************************
!> \brief Fills in temporary arrays with coarse grained data
!> \param fe_env ...
!> \param ng_points ...
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE create_csg_data(fe_env, ng_points, output_unit)
      TYPE(free_energy_type), POINTER                    :: fe_env
      INTEGER, INTENT(IN)                                :: ng_points, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_csg_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iend, istart

      DO i = 1, ng_points
         istart = fe_env%nr_points-(i)*fe_env%conv_par%cg_width+1
         iend = fe_env%nr_points-(i-1)*fe_env%conv_par%cg_width
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) istart, iend
         END IF
         CALL eval_cov_matrix(fe_env, cg_index=i, istart=istart, iend=iend, output_unit=output_unit)
      END DO

   END SUBROUTINE create_csg_data

! **************************************************************************************************
!> \brief Checks Normality of the distribution and Serial Correlation of
!>      coarse grained data
!> \param fe_env ...
!> \param test_passed ...
!> \param nr_points ...
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE ui_check_norm_sc(fe_env, test_passed, nr_points, output_unit)
      TYPE(free_energy_type), POINTER                    :: fe_env
      LOGICAL, INTENT(OUT)                               :: test_passed
      INTEGER, INTENT(IN)                                :: nr_points, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ui_check_norm_sc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ng_points

      CALL timeset(routineN, handle)
      test_passed = .FALSE.
      DO WHILE (fe_env%conv_par%cg_width < fe_env%conv_par%max_cg_width)
         ng_points = nr_points/fe_env%conv_par%cg_width
         PRINT*, ng_points
         IF (ng_points < min_sample_size) EXIT
         CALL ui_check_norm_sc_low(fe_env, nr_points, output_unit)
         test_passed = fe_env%conv_par%test_vn .AND. fe_env%conv_par%test_sw
         IF (test_passed) EXIT
         fe_env%conv_par%cg_width = fe_env%conv_par%cg_width+1
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "New coarse grained width:", fe_env%conv_par%cg_width
         END IF
      END DO
      IF (fe_env%conv_par%cg_width == fe_env%conv_par%max_cg_width .AND. (.NOT. (test_passed))) THEN
         CALL ui_check_norm_sc_low(fe_env, nr_points, output_unit)
         test_passed = fe_env%conv_par%test_vn .AND. fe_env%conv_par%test_sw
      END IF
      CALL timestop(handle)
   END SUBROUTINE ui_check_norm_sc

! **************************************************************************************************
!> \brief Checks Normality of the distribution and Serial Correlation of
!>      coarse grained data - Low Level routine
!> \param fe_env ...
!> \param nr_points ...
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE ui_check_norm_sc_low(fe_env, nr_points, output_unit)
      TYPE(free_energy_type), POINTER                    :: fe_env
      INTEGER, INTENT(IN)                                :: nr_points, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ui_check_norm_sc_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, ncolvar, ng_points
      LOGICAL                                            :: avg_test_passed, sdv_test_passed
      REAL(KIND=dp)                                      :: prob, pw, r, u, w
      REAL(KIND=dp), DIMENSION(:), POINTER               :: wrk

      CALL timeset(routineN, handle)
      ncolvar = fe_env%ncolvar
      ! Compute the Coarse Grained data set using a reverse cumulative strategy
      fe_env%conv_par%test_sw = .FALSE.
      fe_env%conv_par%test_vn = .FALSE.
      ! Number of coarse grained points
      avg_test_passed = .TRUE.
      sdv_test_passed = .TRUE.
      ng_points = nr_points/fe_env%conv_par%cg_width
      CALL create_tmp_data(fe_env, wrk, ng_points, ncolvar)
      CALL create_csg_data(fe_env, ng_points, output_unit)
      ! Testing Averages
      DO j = 1, ncolvar
         DO i = 1, ng_points
            wrk(i) = fe_env%cg_data(i)%avg(j)
         END DO
         ! Test of Shapiro - Wilks for normality
         !                 - Average
         CALL sw_test(wrk, ng_points, w, pw)
         PRINT*, 1.0_dp-pw, fe_env%conv_par%sw_conf_lm
         avg_test_passed = (1.0_dp-pw) <= fe_env%conv_par%sw_conf_lm
         fe_env%conv_par%test_sw = avg_test_passed
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "Shapiro-Wilks normality test (Avg)", avg_test_passed
         END IF
         ! Test of von Neumann for serial correlation
         !                 - Average
         CALL vn_test(wrk, ng_points, r, u, prob)
         PRINT*, prob, fe_env%conv_par%vn_conf_lm
         avg_test_passed = prob <= fe_env%conv_par%vn_conf_lm
         fe_env%conv_par%test_vn = avg_test_passed
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "von Neumann serial correlation test (Avg)", avg_test_passed
         END IF
      END DO
      ! If tests on average are ok let's proceed with Standard Deviation
      IF (fe_env%conv_par%test_vn .AND. fe_env%conv_par%test_sw) THEN
         ! Testing Standard Deviations
         DO j = 1, ncolvar
            DO k = j, ncolvar
               DO i = 1, ng_points
                  wrk(i) = fe_env%cg_data(i)%var(j, k)
               END DO
               ! Test of Shapiro - Wilks for normality
               !                 - Standard Deviation
               CALL sw_test(wrk, ng_points, w, pw)
               PRINT*, 1.0_dp-pw, fe_env%conv_par%sw_conf_lm
               sdv_test_passed = (1.0_dp-pw) <= fe_env%conv_par%sw_conf_lm
               fe_env%conv_par%test_sw = fe_env%conv_par%test_sw .AND. sdv_test_passed
               IF (output_unit > 0) THEN
                  WRITE (output_unit, *) "Shapiro-Wilks normality test (Std. Dev.)", sdv_test_passed
               END IF
               ! Test of von Neumann for serial correlation
               !                 - Standard Deviation
               CALL vn_test(wrk, ng_points, r, u, prob)
               PRINT*, prob, fe_env%conv_par%vn_conf_lm
               sdv_test_passed = prob <= fe_env%conv_par%vn_conf_lm
               fe_env%conv_par%test_vn = fe_env%conv_par%test_vn .AND. sdv_test_passed
               IF (output_unit > 0) THEN
                  WRITE (output_unit, *) "von Neumann serial correlation test (Std. Dev.)", sdv_test_passed
               END IF
            END DO
         END DO
         CALL destroy_tmp_data(fe_env, wrk, ng_points)
      ELSE
         CALL destroy_tmp_data(fe_env, wrk, ng_points)
      END IF
      CALL timestop(handle)
   END SUBROUTINE ui_check_norm_sc_low

! **************************************************************************************************
!> \brief Convergence criteria (Error on average and covariance matrix)
!>      for free energy method
!> \param fe_env ...
!> \param converged ...
!> \param nr_points ...
!> \param output_unit which unit to print to
!> \par History
!>      Teodoro Laino (01.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE ui_check_convergence(fe_env, converged, nr_points, output_unit)
      TYPE(free_energy_type), POINTER                    :: fe_env
      LOGICAL, INTENT(OUT)                               :: converged
      INTEGER, INTENT(IN)                                :: nr_points, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ui_check_convergence', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ic, ncolvar, ng_points
      LOGICAL                                            :: test_passed
      REAL(KIND=dp)                                      :: max_error_avg, max_error_std
      REAL(KIND=dp), DIMENSION(:), POINTER               :: avg_std, avgmx
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cov_std, covmx

      CALL timeset(routineN, handle)
      converged = .FALSE.
      ncolvar = fe_env%ncolvar
      NULLIFY (avgmx, avg_std, covmx, cov_std)
      CALL ui_check_norm_sc(fe_env, test_passed, nr_points, output_unit)
      IF (test_passed) THEN
         ng_points = nr_points/fe_env%conv_par%cg_width
         ! We can finally compute the error on average and covariance matrix
         ! and check if we converged..
         CALL create_tmp_data(fe_env, ng_points=ng_points, ncolvar=ncolvar)
         CALL create_csg_data(fe_env, ng_points, output_unit)
         ALLOCATE (covmx(ncolvar, ncolvar))
         ALLOCATE (avgmx(ncolvar))
         ALLOCATE (cov_std(ncolvar*(ncolvar+1)/2, ncolvar*(ncolvar+1)/2))
         ALLOCATE (avg_std(ncolvar))
         covmx = 0.0_dp
         avgmx = 0.0_dp
         DO i = 1, ng_points
            covmx = covmx+fe_env%cg_data(i)%var
            avgmx = avgmx+fe_env%cg_data(i)%avg
         END DO
         covmx = covmx/REAL(ng_points, KIND=dp)
         avgmx = avgmx/REAL(ng_points, KIND=dp)

         ! Compute errors on average and standard deviation
         CALL compute_avg_std_errors(fe_env, ncolvar, avgmx, covmx, avg_std, cov_std)
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "pippo", avgmx, covmx
            WRITE (output_unit, *) "pippo", avg_std, cov_std
         END IF
         ! Convergence of the averages
         max_error_avg = SQRT(MAXVAL(ABS(avg_std))/REAL(ng_points, KIND=dp))/MINVAL(avgmx)
         max_error_std = SQRT(MAXVAL(ABS(cov_std))/REAL(ng_points, KIND=dp))/MINVAL(covmx)
         IF (max_error_avg <= fe_env%conv_par%eps_conv .AND. &
             max_error_std <= fe_env%conv_par%eps_conv) converged = .TRUE.

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(/,T2,"CG SAMPLING LENGTH = ",I7,20X,"REQUESTED ACCURACY  = ",E12.6)') ng_points, &
               fe_env%conv_par%eps_conv
            WRITE (output_unit, '(T50,"PRESENT ACCURACY AVG= ",E12.6)') max_error_avg
            WRITE (output_unit, '(T50,"PRESENT ACCURACY STD= ",E12.6)') max_error_std
            WRITE (output_unit, '(T50,"CONVERGED FE-DER = ",L12)') converged

            WRITE (output_unit, '(/,T33, "COVARIANCE MATRIX")')
            WRITE (output_unit, '(T8,'//cp_to_string(ncolvar)//'(3X,I7,6X))') (ic, ic=1, ncolvar)
            DO ic = 1, ncolvar
               WRITE (output_unit, '(T2,I6,'//cp_to_string(ncolvar)//'(3X,E12.6,1X))') ic, covmx(ic, :)
            END DO
            WRITE (output_unit, '(T33, "ERROR OF COVARIANCE MATRIX")')
            WRITE (output_unit, '(T8,'//cp_to_string(ncolvar)//'(3X,I7,6X))') (ic, ic=1, ncolvar)
            DO ic = 1, ncolvar
               WRITE (output_unit, '(T2,I6,'//cp_to_string(ncolvar)//'(3X,E12.6,1X))') ic, cov_std(ic, :)
            END DO

            WRITE (output_unit, '(/,T2,"COLVAR Nr.",18X,13X,"AVERAGE",13X,"STANDARD DEVIATION")')
            WRITE (output_unit, '(T2,"CV",I8,21X,7X,E12.6,14X,E12.6)') &
               (ic, avgmx(ic), SQRT(ABS(avg_std(ic))), ic=1, ncolvar)
         END IF
         CALL destroy_tmp_data(fe_env, ng_points=ng_points)
         DEALLOCATE (covmx)
         DEALLOCATE (avgmx)
         DEALLOCATE (cov_std)
         DEALLOCATE (avg_std)
      END IF
      CALL timestop(handle)
   END SUBROUTINE ui_check_convergence

! **************************************************************************************************
!> \brief Computes the errors on averages and standard deviations for a
!>      correlation-independent coarse grained data set
!> \param fe_env ...
!> \param ncolvar ...
!> \param avgmx ...
!> \param covmx ...
!> \param avg_std ...
!> \param cov_std ...
!> \par History
!>      Teodoro Laino (02.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE compute_avg_std_errors(fe_env, ncolvar, avgmx, covmx, avg_std, cov_std)
      TYPE(free_energy_type), POINTER                    :: fe_env
      INTEGER, INTENT(IN)                                :: ncolvar
      REAL(KIND=dp), DIMENSION(:), POINTER               :: avgmx
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: covmx
      REAL(KIND=dp), DIMENSION(:), POINTER               :: avg_std
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cov_std

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_avg_std_errors', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind, j, k, nvar
      REAL(KIND=dp)                                      :: fac
      REAL(KIND=dp), DIMENSION(:), POINTER               :: awrk, eig, tmp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: wrk

! Averages

      nvar = ncolvar
      ALLOCATE (wrk(nvar, nvar))
      ALLOCATE (eig(nvar))
      fac = REAL(SIZE(fe_env%cg_data), KIND=dp)
      wrk = 0.0_dp
      eig = 0.0_dp
      DO k = 1, SIZE(fe_env%cg_data)
         DO j = 1, nvar
            DO i = j, nvar
               wrk(i, j) = wrk(i, j)+fe_env%cg_data(k)%avg(i)*fe_env%cg_data(k)%avg(j)
            END DO
         END DO
      END DO
      DO j = 1, nvar
         DO i = j, nvar
            wrk(i, j) = wrk(i, j)-avgmx(i)*avgmx(j)*fac
            wrk(j, i) = wrk(i, j)
         END DO
      END DO
      wrk = wrk/(fac-1.0_dp)
      ! Diagonalize the covariance matrix and check for the maximum error
      CALL diamat_all(wrk, eig)
      DO i = 1, nvar
         avg_std(i) = eig(i)
      END DO
      DEALLOCATE (wrk)
      DEALLOCATE (eig)
      ! Standard Deviations
      nvar = ncolvar*(ncolvar+1)/2
      ALLOCATE (wrk(nvar, nvar))
      ALLOCATE (eig(nvar))
      ALLOCATE (awrk(nvar))
      ALLOCATE (tmp(nvar))
      wrk = 0.0_dp
      eig = 0.0_dp
      ind = 0
      DO i = 1, ncolvar
         DO j = i, ncolvar
            ind = ind+1
            awrk(ind) = covmx(i, j)
         END DO
      END DO
      DO k = 1, SIZE(fe_env%cg_data)
         ind = 0
         DO i = 1, ncolvar
            DO j = i, ncolvar
               ind = ind+1
               tmp(ind) = fe_env%cg_data(k)%var(i, j)
            END DO
         END DO
         DO i = 1, nvar
            DO j = i, nvar
               wrk(i, j) = wrk(i, j)+tmp(i)*tmp(j)-awrk(i)*awrk(j)
            END DO
         END DO
      END DO
      DO i = 1, nvar
         DO j = i, nvar
            wrk(i, j) = wrk(i, j)-fac*awrk(i)*awrk(j)
            wrk(j, i) = wrk(i, j)
         END DO
      END DO
      wrk = wrk/(fac-1.0_dp)
      ! Diagonalize the covariance matrix and check for the maximum error
      CALL diamat_all(wrk, eig)
      ind = 0
      DO i = 1, ncolvar
         DO j = i, ncolvar
            ind = ind+1
            cov_std(i, j) = eig(ind)
            cov_std(j, i) = cov_std(i, j)
         END DO
      END DO
      DEALLOCATE (wrk)
      DEALLOCATE (eig)
      DEALLOCATE (awrk)
      DEALLOCATE (tmp)

   END SUBROUTINE compute_avg_std_errors

! **************************************************************************************************
!> \brief Computes the covariance matrix
!> \param fe_env ...
!> \param cg_index ...
!> \param istart ...
!> \param iend ...
!> \param output_unit which unit to print to
!> \param covmx ...
!> \param avgs ...
!> \par History
!>      Teodoro Laino (01.2007) [tlaino]
! **************************************************************************************************
   SUBROUTINE eval_cov_matrix(fe_env, cg_index, istart, iend, output_unit, covmx, avgs)
      TYPE(free_energy_type), POINTER                    :: fe_env
      INTEGER, INTENT(IN)                                :: cg_index, istart, iend, output_unit
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: covmx
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: avgs

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_cov_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, jc, jstep, ncolvar, nlength
      REAL(KIND=dp)                                      :: tmp_ic, tmp_jc
      TYPE(ui_var_type), POINTER                         :: cv

      CALL timeset(routineN, handle)
      ncolvar = fe_env%ncolvar
      nlength = iend-istart+1
      fe_env%cg_data(cg_index)%avg = 0.0_dp
      fe_env%cg_data(cg_index)%var = 0.0_dp
      IF (nlength > 1) THEN
         ! Update the info on averages and variances
         DO jstep = istart, iend
            DO ic = 1, ncolvar
               cv => fe_env%uivar(ic)
               tmp_ic = cv%ss(jstep)
               fe_env%cg_data(cg_index)%avg(ic) = fe_env%cg_data(cg_index)%avg(ic)+tmp_ic
            END DO
            DO ic = 1, ncolvar
               cv => fe_env%uivar(ic)
               tmp_ic = cv%ss(jstep)
               DO jc = 1, ic
                  cv => fe_env%uivar(jc)
                  tmp_jc = cv%ss(jstep)
                  fe_env%cg_data(cg_index)%var(jc, ic) = fe_env%cg_data(cg_index)%var(jc, ic)+tmp_ic*tmp_jc
               END DO
            END DO
         END DO
         ! Normalized the variances and the averages
         ! Unbiased estimator
         fe_env%cg_data(cg_index)%var = fe_env%cg_data(cg_index)%var/REAL(nlength-1, KIND=dp)
         fe_env%cg_data(cg_index)%avg = fe_env%cg_data(cg_index)%avg/REAL(nlength, KIND=dp)
         ! Compute the covariance matrix
         DO ic = 1, ncolvar
            tmp_ic = fe_env%cg_data(cg_index)%avg(ic)
            DO jc = 1, ic
               tmp_jc = fe_env%cg_data(cg_index)%avg(jc)*REAL(nlength, KIND=dp)/REAL(nlength-1, KIND=dp)
               fe_env%cg_data(cg_index)%var(jc, ic) = fe_env%cg_data(cg_index)%var(jc, ic)-tmp_ic*tmp_jc
               fe_env%cg_data(cg_index)%var(ic, jc) = fe_env%cg_data(cg_index)%var(jc, ic)
            END DO
         END DO
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "eval_cov_matrix", istart, iend, fe_env%cg_data(cg_index)%avg, fe_env%cg_data(cg_index)%var
         END IF
         IF (PRESENT(covmx)) covmx = fe_env%cg_data(cg_index)%var
         IF (PRESENT(avgs)) avgs = fe_env%cg_data(cg_index)%avg
      END IF
      CALL timestop(handle)
   END SUBROUTINE eval_cov_matrix

! **************************************************************************************************
!> \brief Dumps information when performing an alchemical change run
!> \param my_val ...
!> \param my_par ...
!> \param dx ...
!> \param lerr ...
!> \param fe_section ...
!> \param nforce_eval ...
!> \param cum_res ...
!> \param istep ...
!> \param beta ...
!> \author Teodoro Laino - University of Zurich [tlaino] - 05.2007
! **************************************************************************************************
   SUBROUTINE dump_ac_info(my_val, my_par, dx, lerr, fe_section, nforce_eval, cum_res, &
                           istep, beta)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_val
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: my_par
      REAL(KIND=dp), INTENT(IN)                          :: dx, lerr
      TYPE(section_vals_type), POINTER                   :: fe_section
      INTEGER, INTENT(IN)                                :: nforce_eval
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cum_res
      INTEGER, POINTER                                   :: istep
      REAL(KIND=dp), INTENT(IN)                          :: beta

      CHARACTER(len=*), PARAMETER :: routineN = 'dump_ac_info', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length)               :: def_error, par, this_error
      INTEGER                                            :: i, iforce_eval, ipar, isize, iw, j, &
                                                            NEquilStep
      REAL(KIND=dp)                                      :: avg_BP, avg_DET, avg_DUE, d_ene_w, dedf, &
                                                            ene_w, err, Err_DET, Err_DUE, std_DET, &
                                                            std_DUE, tmp, tmp2, wfac
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: alch_section

      logger => cp_get_default_logger()
      alch_section => section_vals_get_subs_vals(fe_section, "ALCHEMICAL_CHANGE")
      CALL section_vals_val_get(alch_section, "PARAMETER", c_val=par)
      DO i = 1, SIZE(my_par)
         IF (my_par(i) == par) EXIT
      END DO
      IF(.NOT.(i <= SIZE(my_par)))CALL cp__a("motion/free_energy_methods.F",828)
      ipar = i
      dedf = evalfd(1, ipar, my_val, dx, err)
      IF (ABS(err) > lerr) THEN
         WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
         WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
         CALL compress(this_error, .TRUE.)
         CALL compress(def_error, .TRUE.)
         CALL cp_warn(cp__l("motion/free_energy_methods.F",836), &
                      'ASSERTION (cond) failed at line '//cp_to_string(837)// &
                      ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                      TRIM(def_error)//' .')
      END IF

      ! We must print now the energy of the biased system, the weigthing energy
      ! and the derivative w.r.t.the coupling parameter of the biased energy
      ! Retrieve the expression of the weighting function:
      CALL section_vals_val_get(alch_section, "WEIGHTING_FUNCTION", c_val=coupling_function)
      CALL compress(coupling_function, full=.TRUE.)
      CALL parsef(2, TRIM(coupling_function), my_par)
      ene_w = evalf(2, my_val)
      d_ene_w = evalfd(2, ipar, my_val, dx, err)
      IF (ABS(err) > lerr) THEN
         WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
         WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
         CALL compress(this_error, .TRUE.)
         CALL compress(def_error, .TRUE.)
         CALL cp_warn(cp__l("motion/free_energy_methods.F",855), &
                      'ASSERTION (cond) failed at line '//cp_to_string(856)// &
                      ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                      TRIM(def_error)//' .')
      END IF
      CALL section_vals_val_get(alch_section, "NEQUIL_STEPS", i_val=NEquilStep)
      ! Store results
      IF (istep > NEquilStep) THEN
         isize = SIZE(cum_res, 2)+1
         CALL reallocate(cum_res, 1, 3, 1, isize)
         cum_res(1, isize) = dedf
         cum_res(2, isize) = dedf-d_ene_w
         cum_res(3, isize) = ene_w
         ! Compute derivative of biased and total energy
         ! Total Free Energy
         avg_DET = SUM(cum_res(1, 1:isize))/REAL(isize, KIND=dp)
         std_DET = SUM(cum_res(1, 1:isize)**2)/REAL(isize, KIND=dp)
         ! Unbiased Free Energy
         avg_BP = SUM(cum_res(3, 1:isize))/REAL(isize, KIND=dp)
         wfac = 0.0_dp
         DO j = 1, isize
            wfac = wfac+EXP(beta*(cum_res(3, j)-avg_BP))
         END DO
         avg_DUE = 0.0_dp
         std_DUE = 0.0_dp
         DO j = 1, isize
            tmp = cum_res(2, j)
            tmp2 = EXP(beta*(cum_res(3, j)-avg_BP))/wfac
            avg_DUE = avg_DUE+tmp*tmp2
            std_DUE = std_DUE+tmp**2*tmp2
         END DO
         IF (isize > 1) THEN
            Err_DUE = SQRT(std_DUE-avg_DUE**2)/SQRT(REAL(isize-1, KIND=dp))
            Err_DET = SQRT(std_DET-avg_DET**2)/SQRT(REAL(isize-1, KIND=dp))
         END IF
         ! Print info
         iw = cp_print_key_unit_nr(logger, fe_section, "FREE_ENERGY_INFO", &
                                   extension=".free_energy")
         IF (iw > 0) THEN
            WRITE (iw, '(T2,79("-"),T37," oOo ")')
            DO iforce_eval = 1, nforce_eval
               WRITE (iw, '(T2,"ALCHEMICAL CHANGE| FORCE_EVAL Nr.",I5,T48,"ENERGY (Hartree)= ",F15.9)') &
                  iforce_eval, my_val(iforce_eval)
            END DO
            WRITE (iw, '(T2,"ALCHEMICAL CHANGE| DERIVATIVE OF TOTAL ENERGY  [ PARAMETER (",A,") ]",T66,F15.9)') &
               TRIM(par), dedf
            WRITE (iw, '(T2,"ALCHEMICAL CHANGE| DERIVATIVE OF BIASED ENERGY [ PARAMETER (",A,") ]",T66,F15.9)') &
               TRIM(par), dedf-d_ene_w
            WRITE (iw, '(T2,"ALCHEMICAL CHANGE| BIASING UMBRELLA POTENTIAL  ",T66,F15.9)') &
               ene_w

            IF (isize > 1) THEN
               WRITE (iw, '(/,T2,"ALCHEMICAL CHANGE| DERIVATIVE TOTAL FREE ENERGY  ",T50,F15.9,1X,"+/-",1X,F11.9)') &
                  avg_DET, Err_DET
               WRITE (iw, '(T2,"ALCHEMICAL CHANGE| DERIVATIVE UNBIASED FREE ENERGY  ",T50,F15.9,1X,"+/-",1X,F11.9)') &
                  avg_DUE, Err_DUE
            ELSE
               WRITE (iw, '(/,T2,"ALCHEMICAL CHANGE| DERIVATIVE TOTAL FREE ENERGY  ",T50,F15.9,1X,"+/-",1X,T76,A)') &
                  avg_DET, "UNDEF"
               WRITE (iw, '(T2,"ALCHEMICAL CHANGE| DERIVATIVE UNBIASED FREE ENERGY  ",T50,F15.9,1X,"+/-",1X,T76,A)') &
                  avg_DUE, "UNDEF"
            END IF
            WRITE (iw, '(T2,79("-"))')
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, fe_section, "FREE_ENERGY_INFO")

   END SUBROUTINE dump_ac_info

END MODULE free_energy_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Barostat utils
!> \author teo [tlaino] - University of Zurich - 02.2008
! **************************************************************************************************
MODULE barostat_utils
   USE barostat_types,                  ONLY: barostat_type
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE extended_system_types,           ONLY: npt_info_type
   USE input_constants,                 ONLY: npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE physcon,                         ONLY: angstrom,&
                                              femtoseconds,&
                                              kelvin
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC ::  get_baro_energies, print_barostat_status

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'barostat_utils'

CONTAINS
! **************************************************************************************************
!> \brief Calculates kinetic energy and potential of barostat
!> \param cell ...
!> \param simpar ...
!> \param npt ...
!> \param baro_kin ...
!> \param baro_pot ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_baro_energies(cell, simpar, npt, baro_kin, baro_pot)

      TYPE(cell_type), POINTER                           :: cell
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      TYPE(npt_info_type), DIMENSION(:, :), INTENT(IN)   :: npt
      REAL(KIND=dp), INTENT(OUT)                         :: baro_kin, baro_pot

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_baro_energies', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(dp)                                           :: iv0, v0, v_shock

      IF (simpar%ensemble == npt_i_ensemble .OR. simpar%ensemble == npe_i_ensemble) THEN
         baro_pot = simpar%p_ext*cell%deth
         baro_kin = 0.5_dp*npt(1, 1)%v**2*npt(1, 1)%mass
      ELSE IF (simpar%ensemble == npt_f_ensemble .OR. simpar%ensemble == npe_f_ensemble) THEN
         baro_pot = simpar%p_ext*cell%deth
         baro_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, 3
               baro_kin = baro_kin+0.5_dp*npt(i, j)%v**2*npt(i, j)%mass
            END DO
         END DO
      ELSEIF (simpar%ensemble == nph_uniaxial_ensemble .OR. simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         v0 = simpar%v0
         iv0 = 1._dp/v0
         v_shock = simpar%v_shock

         ! Valid only for orthorhombic cell
         baro_pot = -0.5_dp*v_shock*v_shock*(1._dp-cell%deth*iv0)**2-simpar%p0*(v0-cell%deth)
         ! Valid only for orthorhombic cell
         baro_kin = 0.5_dp*npt(1, 1)%v*npt(1, 1)%v*npt(1, 1)%mass
      END IF

   END SUBROUTINE get_baro_energies

! **************************************************************************************************
!> \brief Prints status of barostat during an MD run
!> \param barostat ...
!> \param simpar ...
!> \param my_pos ...
!> \param my_act ...
!> \param cell ...
!> \param itimes ...
!> \param time ...
!> \author Teodoro Laino [tlaino] - 02.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE print_barostat_status(barostat, simpar, my_pos, my_act, cell, itimes, time)
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      CHARACTER(LEN=default_string_length)               :: my_pos, my_act
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_barostat_status', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: baro, nfree
      LOGICAL                                            :: new_file
      REAL(KIND=dp)                                      :: baro_kin, baro_pot, temp
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (ASSOCIATED(barostat)) THEN
         baro = cp_print_key_unit_nr(logger, barostat%section, "PRINT%ENERGY", &
                                     extension=".bener", file_position=my_pos, file_action=my_act, is_new_file=new_file)
         CALL get_baro_energies(cell, simpar, barostat%npt, baro_kin, baro_pot)
         nfree = SIZE(barostat%npt, 1)*SIZE(barostat%npt, 2)
         temp = 2.0_dp*baro_kin/REAL(nfree, dp)*kelvin
         IF (baro > 0) THEN
            IF (new_file) THEN
               WRITE (baro, '("#",3X,A,10X,A,8X,3(5X,A,5X),3X,A)') "Step Nr.", "Time[fs]", "Kin.[a.u.]", &
                  "Temp.[K]", "Pot.[a.u.]", "Vol[Ang.^3]"
            END IF
            WRITE (UNIT=baro, FMT="(I10, F20.3,4F20.10)") itimes, time*femtoseconds, &
               baro_kin, temp, baro_pot, cell%deth*angstrom*angstrom*angstrom
            CALL m_flush(baro)
         END IF
         CALL cp_print_key_finished_output(baro, logger, barostat%section, "PRINT%ENERGY")
      END IF

   END SUBROUTINE print_barostat_status

END MODULE barostat_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief computes the conserved quantities for a given md ensemble
!>      and also kinetic energies, thermo/barostat stuff
!> \author gtb, 05.02.2003
! **************************************************************************************************
MODULE md_conserved_quantities
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE barostat_utils,                  ONLY: get_baro_energies
   USE cell_types,                      ONLY: cell_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_constants,                 ONLY: &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble, reftraj_ensemble
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: zero
   USE md_ener_types,                   ONLY: md_ener_type,&
                                              zero_md_ener
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type,&
                                              set_md_env
   USE message_passing,                 ONLY: mp_sum
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: kelvin
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE qmmm_types_low,                  ONLY: force_mixing_label_QM_dynamics
   USE qmmmx_types,                     ONLY: qmmmx_env_type
   USE shell_potential_types,           ONLY: shell_kind_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: thermostat_type
   USE thermostat_utils,                ONLY: get_thermostat_energies
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: compute_conserved_quantity, calc_nfree_qm
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_conserved_quantities'

CONTAINS

! **************************************************************************************************
!> \brief calculates conserved quantity.
!> \param md_env ...
!> \param md_ener ...
!> \param tkind ...
!> \param tshell ...
!> \param natom ...
!> \par Input Arguments
!>     md_env is the md_environment
!>     epot is the total potential energy
!> \par Output Arguments
!>     cons is the conserved quantity
!> \par Output Optional Arguments
!>     cons_rel : relative cons. quantity (to the first md step)
!>     ekin : kinetic energy of particles
!>     temp : temperature
!>     temp_qm : temperature of the QM system in a QM/MM calculation
!> \par History
!>      none
!> \author gloria
! **************************************************************************************************
   SUBROUTINE compute_conserved_quantity(md_env, md_ener, tkind, tshell, &
                                         natom)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      LOGICAL, INTENT(IN)                                :: tkind, tshell
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_conserved_quantity', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nfree_qm, nkind
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: init
      REAL(KIND=dp), POINTER                             :: constant
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(simpar_type), POINTER                         :: simpar

      NULLIFY (itimes, para_env, simpar)

      CALL zero_md_ener(md_ener, tkind, tshell)

      CALL get_md_env(md_env=md_env, &
                      constant=constant, &
                      itimes=itimes, &
                      init=init, &
                      simpar=simpar, &
                      para_env=para_env)

      CALL get_part_ke(md_env, md_ener, tkind, tshell, para_env%group)

      IF (md_ener%nfree /= 0) THEN
         md_ener%temp_part = 2.0_dp*md_ener%ekin/REAL(simpar%nfree, KIND=dp)
         md_ener%temp_part = md_ener%temp_part*kelvin
      END IF

      nfree_qm = calc_nfree_qm(md_env, md_ener)
      IF (nfree_qm > 0) THEN
         md_ener%temp_qm = 2.0_dp*md_ener%ekin_qm/REAL(nfree_qm, KIND=dp)
         md_ener%temp_qm = md_ener%temp_qm*kelvin
      END IF

      IF (md_ener%nfree_shell > 0) THEN
         md_ener%temp_shell = 2.0_dp*md_ener%ekin_shell/REAL(md_ener%nfree_shell, KIND=dp)
         md_ener%temp_shell = md_ener%temp_shell*kelvin
      END IF

      IF (tkind) THEN
         nkind = SIZE(md_ener%temp_kind)
         DO ikind = 1, nkind
            md_ener%temp_kind(ikind) = 2.0_dp* &
                                       md_ener%ekin_kind(ikind)/REAL(md_ener%nfree_kind(ikind), KIND=dp)
            md_ener%temp_kind(ikind) = md_ener%temp_kind(ikind)*kelvin
         END DO
         IF (tshell) THEN
            DO ikind = 1, nkind
               md_ener%temp_shell_kind(ikind) = 2.0_dp* &
                                                md_ener%ekin_shell_kind(ikind)/REAL(md_ener%nfree_shell_kind(ikind), KIND=dp)
               md_ener%temp_shell_kind(ikind) = md_ener%temp_shell_kind(ikind)*kelvin
            END DO
         END IF
      END IF

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/md_conserved_quantities.F",144,'Unknown ensemble')
      CASE (isokin_ensemble)
         md_ener%constant = md_ener%ekin
      CASE (reftraj_ensemble) ! no constant of motion available
         md_ener%constant = md_ener%epot
      CASE (nve_ensemble)
         CALL get_econs_nve(md_env, md_ener, para_env)
      CASE (nvt_ensemble)
         CALL get_econs_nvt(md_env, md_ener, para_env)
      CASE (npt_i_ensemble, npt_f_ensemble)
         CALL get_econs_npt(md_env, md_ener, para_env)
         md_ener%temp_baro = md_ener%temp_baro*kelvin
      CASE (nph_uniaxial_ensemble)
         CALL get_econs_nph_uniaxial(md_env, md_ener)
         md_ener%temp_baro = md_ener%temp_baro*kelvin
      CASE (nph_uniaxial_damped_ensemble)
         CALL get_econs_nph_uniaxial(md_env, md_ener)
         md_ener%temp_baro = md_ener%temp_baro*kelvin
      CASE (langevin_ensemble)
         md_ener%constant = md_ener%ekin+md_ener%epot
      CASE (npe_f_ensemble, npe_i_ensemble)
         CALL get_econs_npe(md_env, md_ener, para_env)
         md_ener%temp_baro = md_ener%temp_baro*kelvin
      CASE (nvt_adiabatic_ensemble)
         CALL get_econs_nvt_adiabatic(md_env, md_ener, para_env)
      END SELECT

      IF (init) THEN
         ! If the value was not read from input let's set it at the begin of the MD
         IF (constant == 0.0_dp) THEN
            constant = md_ener%constant
            CALL set_md_env(md_env=md_env, constant=constant)
         END IF
      ELSE
         CALL get_md_env(md_env=md_env, constant=constant)
         md_ener%delta_cons = (md_ener%constant-constant)/REAL(natom, KIND=dp)*kelvin
      END IF

   END SUBROUTINE compute_conserved_quantity

! **************************************************************************************************
!> \brief Calculates the number of QM degress of freedom
!> \param md_env ...
!> \param md_ener ...
!> \retval nfree_qm ...
! **************************************************************************************************
   FUNCTION calc_nfree_qm(md_env, md_ener) RESULT(nfree_qm)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      INTEGER                                            :: nfree_qm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_nfree_qm', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      INTEGER, POINTER                                   :: cur_indices(:), cur_labels(:)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      TYPE(section_vals_type), POINTER                   :: force_env_section

      NULLIFY (qmmm_env, qmmmx_env, subsys, particles, force_env, force_env_section)
      nfree_qm = 0

      CALL get_md_env(md_env, force_env=force_env)
      CALL force_env_get(force_env, &
                         subsys=subsys, &
                         qmmm_env=qmmm_env, &
                         qmmmx_env=qmmmx_env, &
                         force_env_section=force_env_section)

      IF (ASSOCIATED(qmmm_env)) THEN ! conventional QM/MM
         CALL cp_subsys_get(subsys, particles=particles)
         ! The degrees of freedom for the quantum part of the system
         ! are set to 3*Number of QM atoms and to simpar%nfree in case all the MM
         ! system is treated at QM level (not really QM/MM, just for consistency).
         ! The degree of freedom will not be correct if 1-3 atoms are treated only
         ! MM. In this case we should take care of rotations
         nfree_qm = 3*SIZE(qmmm_env%qm%qm_atom_index)
         IF (nfree_qm == 3*(particles%n_els)) nfree_qm = md_ener%nfree
      END IF

      IF (ASSOCIATED(qmmmx_env)) THEN ! doing force mixing
         CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%INDICES", i_vals=cur_indices)
         CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%LABELS", i_vals=cur_labels)
         nfree_qm = 0
         DO ip = 1, SIZE(cur_indices)
            IF (cur_labels(ip) >= force_mixing_label_QM_dynamics) THEN ! this is a QM atom
               nfree_qm = nfree_qm+3
            END IF
         END DO
      END IF

      IF(.NOT.(.NOT. (ASSOCIATED(qmmm_env) .AND. ASSOCIATED(qmmmx_env))))CALL cp__a("motion/md_conserved_quantities.F",238)
   END FUNCTION calc_nfree_qm

! **************************************************************************************************
!> \brief calculates conserved quantity for nvt ensemble
!> \param md_env ...
!> \param md_ener ...
!> \param para_env ...
!> \par History
!>      none
!> \author gloria
! **************************************************************************************************
   SUBROUTINE get_econs_nve(md_env, md_ener, para_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_nve', routineP = moduleN//':'//routineN

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(thermostat_type), POINTER                     :: thermostat_coeff, thermostat_shell

      NULLIFY (force_env, thermostat_coeff, thermostat_shell)

      CALL get_md_env(md_env, force_env=force_env, thermostat_coeff=thermostat_coeff, &
                      thermostat_shell=thermostat_shell)
      md_ener%constant = md_ener%ekin+md_ener%epot+md_ener%ekin_shell

      CALL get_thermostat_energies(thermostat_shell, md_ener%thermostat_shell_pot, &
                                   md_ener%thermostat_shell_kin, para_env)
      md_ener%constant = md_ener%constant+md_ener%thermostat_shell_kin+md_ener%thermostat_shell_pot

   END SUBROUTINE get_econs_nve

! **************************************************************************************************
!> \brief calculates conserved quantity for nvt ensemble
!> \param md_env ...
!> \param md_ener ...
!> \param para_env ...
!> \par History
!>      none
!> \author gloria
! **************************************************************************************************
   SUBROUTINE get_econs_nvt_adiabatic(md_env, md_ener, para_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_nvt_adiabatic', &
         routineP = moduleN//':'//routineN

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(thermostat_type), POINTER                     :: thermostat_fast, thermostat_slow

      NULLIFY (force_env, thermostat_fast, thermostat_slow)
      CALL get_md_env(md_env, force_env=force_env, thermostat_fast=thermostat_fast, &
                      thermostat_slow=thermostat_slow)
      CALL get_thermostat_energies(thermostat_fast, md_ener%thermostat_fast_pot, &
                                   md_ener%thermostat_fast_kin, para_env)
      md_ener%constant = md_ener%ekin+md_ener%epot+ &
                         md_ener%thermostat_fast_kin+md_ener%thermostat_fast_pot
      CALL get_thermostat_energies(thermostat_slow, md_ener%thermostat_slow_pot, &
                                   md_ener%thermostat_slow_kin, para_env)
      md_ener%constant = md_ener%constant+ &
                         md_ener%thermostat_slow_kin+md_ener%thermostat_slow_pot

   END SUBROUTINE get_econs_nvt_adiabatic

! **************************************************************************************************
!> \brief calculates conserved quantity for nvt ensemble
!> \param md_env ...
!> \param md_ener ...
!> \param para_env ...
!> \par History
!>      none
!> \author gloria
! **************************************************************************************************
   SUBROUTINE get_econs_nvt(md_env, md_ener, para_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_nvt', routineP = moduleN//':'//routineN

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(thermostat_type), POINTER                     :: thermostat_coeff, thermostat_part, &
                                                            thermostat_shell

      NULLIFY (force_env, thermostat_part, thermostat_coeff, thermostat_shell)
      CALL get_md_env(md_env, force_env=force_env, thermostat_part=thermostat_part, &
                      thermostat_coeff=thermostat_coeff, thermostat_shell=thermostat_shell)
      CALL get_thermostat_energies(thermostat_part, md_ener%thermostat_part_pot, &
                                   md_ener%thermostat_part_kin, para_env)
      md_ener%constant = md_ener%ekin+md_ener%epot+md_ener%ekin_shell+ &
                         md_ener%thermostat_part_kin+md_ener%thermostat_part_pot

      CALL get_thermostat_energies(thermostat_shell, md_ener%thermostat_shell_pot, &
                                   md_ener%thermostat_shell_kin, para_env)
      md_ener%constant = md_ener%constant+md_ener%thermostat_shell_kin+md_ener%thermostat_shell_pot

   END SUBROUTINE get_econs_nvt

! **************************************************************************************************
!> \brief calculates conserved quantity for npe ensemble
!> \param md_env ...
!> \param md_ener ...
!> \param para_env ...
!> \par History
!>      none
!> \author  marcella (02-2008)
! **************************************************************************************************
   SUBROUTINE get_econs_npe(md_env, md_ener, para_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_npe', routineP = moduleN//':'//routineN

      INTEGER                                            :: nfree
      TYPE(cell_type), POINTER                           :: box
      TYPE(npt_info_type), POINTER                       :: npt(:, :)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_baro, thermostat_shell

      NULLIFY (thermostat_baro, thermostat_shell, npt)
      CALL get_md_env(md_env, thermostat_baro=thermostat_baro, &
                      simpar=simpar, npt=npt, cell=box, &
                      thermostat_shell=thermostat_shell)
      CALL get_baro_energies(box, simpar, npt, md_ener%baro_kin, &
                             md_ener%baro_pot)
      nfree = SIZE(npt, 1)*SIZE(npt, 2)
      md_ener%temp_baro = 2.0_dp*md_ener%baro_kin/nfree

      md_ener%constant = md_ener%ekin+md_ener%epot+md_ener%ekin_shell &
                         +md_ener%baro_kin+md_ener%baro_pot

      CALL get_thermostat_energies(thermostat_shell, md_ener%thermostat_shell_pot, &
                                   md_ener%thermostat_shell_kin, para_env)
      md_ener%constant = md_ener%constant+md_ener%thermostat_shell_kin+ &
                         md_ener%thermostat_shell_pot

   END SUBROUTINE get_econs_npe

! **************************************************************************************************
!> \brief calculates conserved quantity for npt ensemble
!> \param md_env ...
!> \param md_ener ...
!> \param para_env ...
!> \par History
!>      none
!> \author gloria
! **************************************************************************************************
   SUBROUTINE get_econs_npt(md_env, md_ener, para_env)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_npt', routineP = moduleN//':'//routineN

      INTEGER                                            :: nfree
      TYPE(cell_type), POINTER                           :: box
      TYPE(npt_info_type), POINTER                       :: npt(:, :)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_baro, thermostat_part, &
                                                            thermostat_shell

      NULLIFY (thermostat_baro, thermostat_part, thermostat_shell, npt, simpar, box)
      CALL get_md_env(md_env, thermostat_part=thermostat_part, thermostat_baro=thermostat_baro, &
                      simpar=simpar, npt=npt, cell=box, thermostat_shell=thermostat_shell)
      CALL get_thermostat_energies(thermostat_part, md_ener%thermostat_part_pot, &
                                   md_ener%thermostat_part_kin, para_env)
      CALL get_thermostat_energies(thermostat_baro, md_ener%thermostat_baro_pot, &
                                   md_ener%thermostat_baro_kin, para_env)
      CALL get_baro_energies(box, simpar, npt, md_ener%baro_kin, md_ener%baro_pot)
      nfree = SIZE(npt, 1)*SIZE(npt, 2)
      md_ener%temp_baro = 2.0_dp*md_ener%baro_kin/nfree

      md_ener%constant = md_ener%ekin+md_ener%epot+md_ener%ekin_shell &
                         +md_ener%thermostat_part_kin+md_ener%thermostat_part_pot &
                         +md_ener%thermostat_baro_kin+md_ener%thermostat_baro_pot &
                         +md_ener%baro_kin+md_ener%baro_pot

      CALL get_thermostat_energies(thermostat_shell, md_ener%thermostat_shell_pot, &
                                   md_ener%thermostat_shell_kin, para_env)
      md_ener%constant = md_ener%constant+md_ener%thermostat_shell_kin+md_ener%thermostat_shell_pot

   END SUBROUTINE get_econs_npt

! **************************************************************************************************
!> \brief calculates conserved quantity for nph_uniaxial
!> \param md_env ...
!> \param md_ener ...
!> \par History
!>      none
!> \author cjm
! **************************************************************************************************
   SUBROUTINE get_econs_nph_uniaxial(md_env, md_ener)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), INTENT(inout)                  :: md_ener

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_econs_nph_uniaxial', &
         routineP = moduleN//':'//routineN

      TYPE(cell_type), POINTER                           :: box
      TYPE(npt_info_type), POINTER                       :: npt(:, :)
      TYPE(simpar_type), POINTER                         :: simpar

      CALL get_md_env(md_env, simpar=simpar, npt=npt, cell=box)

      CALL get_baro_energies(box, simpar, npt, md_ener%baro_kin, md_ener%baro_pot)
      md_ener%temp_baro = 2.0_dp*md_ener%baro_kin
      md_ener%constant = md_ener%ekin+md_ener%epot+md_ener%baro_kin+md_ener%baro_pot
   END SUBROUTINE get_econs_nph_uniaxial

! **************************************************************************************************
!> \brief Calculates kinetic energy of particles
!> \param md_env ...
!> \param md_ener ...
!> \param tkind ...
!> \param tshell ...
!> \param group ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_part_ke(md_env, md_ener, tkind, tshell, group)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      LOGICAL, INTENT(IN)                                :: tkind, tshell
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_part_ke', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local, shell_index
      INTEGER, POINTER                                   :: cur_indices(:), cur_labels(:)
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: ekin_c, ekin_com, ekin_s, mass
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (force_env, qmmm_env, qmmmx_env, subsys, force_env_section)
      CALL get_md_env(md_env, force_env=force_env)
      CALL force_env_get(force_env, &
                         subsys=subsys, &
                         qmmm_env=qmmm_env, &
                         qmmmx_env=qmmmx_env, &
                         force_env_section=force_env_section)

      CALL cp_subsys_get(subsys=subsys, &
                         atomic_kinds=atomic_kinds, &
                         local_particles=local_particles, &
                         particles=particles, shell_particles=shell_particles, &
                         core_particles=core_particles)

      nparticle_kind = atomic_kinds%n_els
      atomic_kind_set => atomic_kinds%els

      ekin_s = zero
      ekin_c = zero
      ekin_com = zero
      IF (tkind) THEN
         md_ener%nfree_kind = 0
         IF (tshell) THEN
            md_ener%nfree_shell_kind = 0
         END IF
      END IF

      particle_set => particles%els
      IF (tshell) THEN
         shell_particle_set => shell_particles%els
         core_particle_set => core_particles%els
         DO iparticle_kind = 1, nparticle_kind
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, &
                                 shell_active=is_shell, shell=shell)
            nparticle_local = local_particles%n_el(iparticle_kind)
            IF (is_shell) THEN
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                  shell_index = particle_set(iparticle)%shell_index
                  !ekin
                  ekin_com = 0.5_dp*mass* &
                             (particle_set(iparticle)%v(1)*particle_set(iparticle)%v(1) &
                              +particle_set(iparticle)%v(2)*particle_set(iparticle)%v(2) &
                              +particle_set(iparticle)%v(3)*particle_set(iparticle)%v(3))
                  !vcom
                  md_ener%vcom(1) = md_ener%vcom(1)+particle_set(iparticle)%v(1)*mass
                  md_ener%vcom(2) = md_ener%vcom(2)+particle_set(iparticle)%v(2)*mass
                  md_ener%vcom(3) = md_ener%vcom(3)+particle_set(iparticle)%v(3)*mass
                  md_ener%total_mass = md_ener%total_mass+mass

                  md_ener%ekin = md_ener%ekin+ekin_com
                  ekin_c = 0.5_dp*shell%mass_core* &
                           (core_particle_set(shell_index)%v(1)*core_particle_set(shell_index)%v(1) &
                            +core_particle_set(shell_index)%v(2)*core_particle_set(shell_index)%v(2) &
                            +core_particle_set(shell_index)%v(3)*core_particle_set(shell_index)%v(3))
                  ekin_s = 0.5_dp*shell%mass_shell* &
                           (shell_particle_set(shell_index)%v(1)*shell_particle_set(shell_index)%v(1) &
                            +shell_particle_set(shell_index)%v(2)*shell_particle_set(shell_index)%v(2) &
                            +shell_particle_set(shell_index)%v(3)*shell_particle_set(shell_index)%v(3))
                  md_ener%ekin_shell = md_ener%ekin_shell+ekin_c+ekin_s-ekin_com

                  IF (tkind) THEN
                     md_ener%ekin_kind(iparticle_kind) = md_ener%ekin_kind(iparticle_kind)+ekin_com
                     md_ener%nfree_kind(iparticle_kind) = md_ener%nfree_kind(iparticle_kind)+3
                     md_ener%ekin_shell_kind(iparticle_kind) = md_ener%ekin_shell_kind(iparticle_kind)+ &
                                                               ekin_c+ekin_s-ekin_com
                     md_ener%nfree_shell_kind(iparticle_kind) = md_ener%nfree_shell_kind(iparticle_kind)+3
                  END IF

               END DO ! iparticle_local
            ELSE
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                  ekin_com = 0.5_dp*mass* &
                             (particle_set(iparticle)%v(1)*particle_set(iparticle)%v(1) &
                              +particle_set(iparticle)%v(2)*particle_set(iparticle)%v(2) &
                              +particle_set(iparticle)%v(3)*particle_set(iparticle)%v(3))
                  !vcom
                  md_ener%vcom(1) = md_ener%vcom(1)+particle_set(iparticle)%v(1)*mass
                  md_ener%vcom(2) = md_ener%vcom(2)+particle_set(iparticle)%v(2)*mass
                  md_ener%vcom(3) = md_ener%vcom(3)+particle_set(iparticle)%v(3)*mass
                  md_ener%total_mass = md_ener%total_mass+mass

                  md_ener%ekin = md_ener%ekin+ekin_com
                  IF (tkind) THEN
                     md_ener%ekin_kind(iparticle_kind) = md_ener%ekin_kind(iparticle_kind)+ekin_com
                     md_ener%nfree_kind(iparticle_kind) = md_ener%nfree_kind(iparticle_kind)+3
                  END IF
               END DO ! iparticle_local
            END IF
         END DO ! iparticle_kind
         IF (tkind) THEN
            CALL mp_sum(md_ener%ekin_kind, group)
            CALL mp_sum(md_ener%nfree_kind, group)
            CALL mp_sum(md_ener%ekin_shell_kind, group)
            CALL mp_sum(md_ener%nfree_shell_kind, group)
         END IF
         ! sum all contributions to energy over calculated parts on all processors
         CALL mp_sum(md_ener%ekin_shell, group)
      ELSE
         DO iparticle_kind = 1, nparticle_kind
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               ! ekin
               ekin_com = 0.5_dp*mass* &
                          (particle_set(iparticle)%v(1)*particle_set(iparticle)%v(1) &
                           +particle_set(iparticle)%v(2)*particle_set(iparticle)%v(2) &
                           +particle_set(iparticle)%v(3)*particle_set(iparticle)%v(3))

               !vcom
               md_ener%vcom(1) = md_ener%vcom(1)+particle_set(iparticle)%v(1)*mass
               md_ener%vcom(2) = md_ener%vcom(2)+particle_set(iparticle)%v(2)*mass
               md_ener%vcom(3) = md_ener%vcom(3)+particle_set(iparticle)%v(3)*mass
               md_ener%total_mass = md_ener%total_mass+mass

               md_ener%ekin = md_ener%ekin+ekin_com
               IF (tkind) THEN
                  md_ener%ekin_kind(iparticle_kind) = md_ener%ekin_kind(iparticle_kind)+ekin_com
                  md_ener%nfree_kind(iparticle_kind) = md_ener%nfree_kind(iparticle_kind)+3
               END IF
            END DO
         END DO ! iparticle_kind
         IF (tkind) THEN
            CALL mp_sum(md_ener%ekin_kind, group)
            CALL mp_sum(md_ener%nfree_kind, group)
         END IF
      END IF

      ! sum all contributions to energy over calculated parts on all processors
      CALL mp_sum(md_ener%ekin, group)
      CALL mp_sum(md_ener%vcom, group)
      CALL mp_sum(md_ener%total_mass, group)
      md_ener%vcom = md_ener%vcom/md_ener%total_mass
      !
      ! Compute the QM/MM kinetic energy

      IF (ASSOCIATED(qmmm_env)) THEN ! conventional QM/MM
         DO i = 1, SIZE(qmmm_env%qm%qm_atom_index)
            iparticle = qmmm_env%qm%qm_atom_index(i)
            mass = particle_set(iparticle)%atomic_kind%mass
            md_ener%ekin_qm = md_ener%ekin_qm+0.5_dp*mass* &
                              (particle_set(iparticle)%v(1)*particle_set(iparticle)%v(1) &
                               +particle_set(iparticle)%v(2)*particle_set(iparticle)%v(2) &
                               +particle_set(iparticle)%v(3)*particle_set(iparticle)%v(3))
         END DO
      END IF

      IF (ASSOCIATED(qmmmx_env)) THEN ! doing force mixing
         CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%INDICES", i_vals=cur_indices)
         CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%LABELS", i_vals=cur_labels)
         DO i = 1, SIZE(cur_indices)
            IF (cur_labels(i) >= force_mixing_label_QM_dynamics) THEN ! this is a QM atom
               iparticle = cur_indices(i)
               mass = particle_set(iparticle)%atomic_kind%mass
               md_ener%ekin_qm = md_ener%ekin_qm+0.5_dp*mass* &
                                 (particle_set(iparticle)%v(1)*particle_set(iparticle)%v(1) &
                                  +particle_set(iparticle)%v(2)*particle_set(iparticle)%v(2) &
                                  +particle_set(iparticle)%v(3)*particle_set(iparticle)%v(3))
            END IF
         END DO
      END IF

      IF (ASSOCIATED(qmmm_env) .AND. ASSOCIATED(qmmmx_env)) &
         CALL cp__b("motion/md_conserved_quantities.F",659,"get_part_ke: qmmm bug")
   END SUBROUTINE get_part_ke

! **************************************************************************************************

END MODULE md_conserved_quantities
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Setup of regions with different temperature
!> \par History
!>   - Added support for Langevin regions (2014/01/08, LT)
!>   - Added print subroutine for langevin regions (2014/02/04, LT)
!>   - Changed print_thermal_regions to print_thermal_regions_temperature
!>     (2014/02/04, LT)
!> \author MI
! **************************************************************************************************
MODULE thermal_region_utils

   USE bibliography,                    ONLY: Kantorovich2008,&
                                              Kantorovich2008a,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_constants,                 ONLY: langevin_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble,&
                                              nvt_ensemble
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_list_types,             ONLY: particle_list_type
   USE physcon,                         ONLY: kelvin
   USE simpar_types,                    ONLY: simpar_type
   USE thermal_region_types,            ONLY: allocate_thermal_regions,&
                                              release_thermal_regions,&
                                              thermal_region_type,&
                                              thermal_regions_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: create_thermal_regions, &
             print_thermal_regions_temperature, &
             print_thermal_regions_langevin

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermal_region_utils'

CONTAINS

! **************************************************************************************************
!> \brief create thermal_regions
!> \param thermal_regions ...
!> \param md_section ...
!> \param simpar ...
!> \param force_env ...
!> \par History
!>   - Added support for Langevin regions (2014/01/08, LT)
!> \author
! **************************************************************************************************
   SUBROUTINE create_thermal_regions(thermal_regions, md_section, simpar, force_env)
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermal_regions', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, il, ipart, ireg, nlist, nregions
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      LOGICAL                                            :: apply_thermostat, do_langevin, &
                                                            do_langevin_default, explicit
      REAL(KIND=dp)                                      :: temp, temp_tol
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: region_sections, thermal_region_section
      TYPE(thermal_region_type), POINTER                 :: t_region

      NULLIFY (region_sections, t_region, thermal_region_section, particles, subsys, tmplist)
      CALL allocate_thermal_regions(thermal_regions)
      thermal_region_section => section_vals_get_subs_vals(md_section, "THERMAL_REGION")
      CALL section_vals_get(thermal_region_section, explicit=explicit)
      IF (explicit) THEN
         apply_thermostat = (simpar%ensemble == nvt_ensemble) .OR. &
                            (simpar%ensemble == npt_f_ensemble) .OR. &
                            (simpar%ensemble == npt_i_ensemble)
         IF (apply_thermostat) THEN
            CALL cp_warn(cp__l("motion/thermal_region_utils.F",100), &
                         "With the chosen ensemble the temperature is "// &
                         "controlled by thermostats. The definition of different thermal "// &
                         "regions might result inconsistent with the presence of thermostats.")
         END IF
         IF (simpar%temp_tol > 0.0_dp) THEN
            CALL cp_warn(cp__l("motion/thermal_region_utils.F",106), &
                         "Control of the global temperature by rescaling of the velocity"// &
                         "is not consistent with the presence of different thermal regions."// &
                         "The temperature of different regions is rescaled separatedly.")
         END IF
         CALL section_vals_val_get(thermal_region_section, "FORCE_RESCALING", &
                                   l_val=thermal_regions%force_rescaling)
         region_sections => section_vals_get_subs_vals(thermal_region_section, &
                                                       "DEFINE_REGION")
         CALL section_vals_get(region_sections, n_repetition=nregions)
         IF (nregions > 0) THEN
            thermal_regions%nregions = nregions
            thermal_regions%section => thermal_region_section
            ALLOCATE (thermal_regions%thermal_region(nregions))
            CALL force_env_get(force_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            IF (simpar%ensemble == langevin_ensemble) THEN
               CALL cite_reference(Kantorovich2008)
               CALL cite_reference(Kantorovich2008a)
               CALL section_vals_val_get(thermal_region_section, "DO_LANGEVIN_DEFAULT", &
                                         l_val=do_langevin_default)
               ALLOCATE (thermal_regions%do_langevin(particles%n_els))
               thermal_regions%do_langevin = do_langevin_default
            END IF
            DO ireg = 1, nregions
               NULLIFY (t_region)
               t_region => thermal_regions%thermal_region(ireg)
               t_region%region_index = ireg
               CALL section_vals_val_get(region_sections, "LIST", &
                                         i_rep_section=ireg, n_rep_val=nlist)
               NULLIFY (t_region%part_index)
               t_region%npart = 0
               IF (simpar%ensemble == langevin_ensemble) THEN
                  CALL section_vals_val_get(region_sections, "DO_LANGEVIN", &
                                            i_rep_section=ireg, l_val=do_langevin)
               END IF
               DO il = 1, nlist
                  CALL section_vals_val_get(region_sections, "LIST", i_rep_section=ireg, &
                                            i_rep_val=il, i_vals=tmplist)
                  CALL reallocate(t_region%part_index, 1, t_region%npart+SIZE(tmplist))
                  DO i = 1, SIZE(tmplist)
                     ipart = tmplist(i)
                     IF(.NOT.(((ipart > 0) .AND. (ipart <= particles%n_els))))CALL cp__a("motion/thermal_region_utils.F",148)
                     t_region%npart = t_region%npart+1
                     t_region%part_index(t_region%npart) = ipart
                     particles%els(ipart)%t_region_index = ireg
                     IF (simpar%ensemble == langevin_ensemble) THEN
                        thermal_regions%do_langevin(ipart) = do_langevin
                     END IF
                  END DO
               END DO
               CALL section_vals_val_get(region_sections, "TEMPERATURE", i_rep_section=ireg, &
                                         r_val=temp)
               t_region%temp_expected = temp
               CALL section_vals_val_get(region_sections, "TEMP_TOL", i_rep_section=ireg, &
                                         r_val=temp_tol)
               t_region%temp_tol = temp_tol
            END DO
            simpar%do_thermal_region = .TRUE.
         ELSE
            CALL release_thermal_regions(thermal_regions)
            simpar%do_thermal_region = .FALSE.
         END IF
      ELSE
         CALL release_thermal_regions(thermal_regions)
         simpar%do_thermal_region = .FALSE.
      END IF

   END SUBROUTINE create_thermal_regions

! **************************************************************************************************
!> \brief print_thermal_regions_temperature
!> \param thermal_regions : thermal regions type contains information
!>                          about the regions
!> \param itimes          : iteration number of the time step
!> \param time            : simulation time of the time step
!> \param pos             : file position
!> \param act             : file action
!> \par History
!>   - added doxygen header and changed subroutine name from
!>     print_thermal_regions to print_thermal_regions_temperature
!>     (2014/02/04, LT)
!> \author
! **************************************************************************************************
   SUBROUTINE print_thermal_regions_temperature(thermal_regions, itimes, time, pos, act)
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time
      CHARACTER(LEN=default_string_length)               :: pos, act

      CHARACTER(len=*), PARAMETER :: routineN = 'print_thermal_regions_temperature', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: fmd
      INTEGER                                            :: ireg, nregions, unit
      LOGICAL                                            :: new_file
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (ASSOCIATED(thermal_regions)) THEN
         print_key => section_vals_get_subs_vals(thermal_regions%section, "PRINT%TEMPERATURE")
         IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
            unit = cp_print_key_unit_nr(logger, thermal_regions%section, "PRINT%TEMPERATURE", &
                                        extension=".tregion", file_position=pos, &
                                        file_action=act, is_new_file=new_file)
            IF (unit > 0) THEN
               IF (new_file) THEN
                  WRITE (unit, '(A)') "# Temperature per Region"
                  WRITE (unit, '("#",3X,A,2X,A,13X,A)') "Step Nr.", "Time[fs]", "Temp.[K] ...."
               END IF
               nregions = thermal_regions%nregions
               ALLOCATE (temp(0:nregions))
               temp = 0.0_dp
               temp(0) = thermal_regions%temp_reg0
               DO ireg = 1, nregions
                  temp(ireg) = thermal_regions%thermal_region(ireg)%temperature
               END DO
               fmd = "(I10,F20.3,"//TRIM(ADJUSTL(cp_to_string(nregions+1)))//"F20.6)"
               fmd = TRIM(fmd)
               WRITE (UNIT=unit, FMT=fmd) itimes, time, temp(0:nregions)
               DEALLOCATE (temp)
            END IF
            CALL cp_print_key_finished_output(unit, logger, thermal_regions%section, "PRINT%TEMPERATURE")
         END IF
      END IF
   END SUBROUTINE print_thermal_regions_temperature

! **************************************************************************************************
!> \brief print out information regarding to langevin regions defined in
!>        thermal_regions section
!> \param thermal_regions : thermal regions type containing the relevant
!>                          langevin regions information
!> \param simpar          : wrapper for simulation parameters
!> \param pos             : file position
!> \param act             : file action
!> \par History
!>   - created (2014/02/02, LT)
!> \author Lianheng Tong [LT] (tonglianheng@gmail.com)
! **************************************************************************************************
   SUBROUTINE print_thermal_regions_langevin(thermal_regions, simpar, pos, act)
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(simpar_type), POINTER                         :: simpar
      CHARACTER(LEN=default_string_length)               :: pos, act

      CHARACTER(len=*), PARAMETER :: routineN = 'print_thermal_regions_langevin', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipart, ipart_reg, ireg, natoms, &
                                                            print_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: region_id
      LOGICAL                                            :: new_file
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: temperature
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (ASSOCIATED(thermal_regions)) THEN
         IF (ASSOCIATED(thermal_regions%do_langevin)) THEN
            print_key => section_vals_get_subs_vals(thermal_regions%section, &
                                                    "PRINT%LANGEVIN_REGIONS")
            IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                      cp_p_file)) THEN
               print_unit = cp_print_key_unit_nr(logger, thermal_regions%section, &
                                                 "PRINT%LANGEVIN_REGIONS", &
                                                 extension=".lgv_regions", &
                                                 file_position=pos, file_action=act, &
                                                 is_new_file=new_file)
               IF (print_unit > 0) THEN
                  IF (new_file) THEN
                     WRITE (print_unit, '(A)') "# Atoms Undergoing Langevin MD"
                     WRITE (print_unit, '(A,3X,A,3X,A,3X,A,3X,A)') &
                        "#", "Atom_ID", "Region_ID", "Langevin(L)/NVE(N)", "Expected_Temperature[K]"
                  END IF
                  natoms = SIZE(thermal_regions%do_langevin)
                  ALLOCATE (temperature(natoms))
                  ALLOCATE (region_id(natoms))
                  temperature(:) = simpar%temp_ext
                  region_id(:) = 0
                  DO ireg = 1, thermal_regions%nregions
                     DO ipart_reg = 1, thermal_regions%thermal_region(ireg)%npart
                        ipart = thermal_regions%thermal_region(ireg)%part_index(ipart_reg)
                        temperature(ipart) = thermal_regions%thermal_region(ireg)%temp_expected
                        region_id(ipart) = thermal_regions%thermal_region(ireg)%region_index
                     END DO
                  END DO
                  DO ipart = 1, natoms
                     WRITE (print_unit, '(1X,I10,3X)', advance='no') ipart
                     WRITE (print_unit, '(I10,3X)', advance='no') region_id(ipart)
                     IF (thermal_regions%do_langevin(ipart)) THEN
                        WRITE (print_unit, '(A,3X)', advance='no') "L"
                        WRITE (print_unit, '(F20.3)') temperature(ipart)*kelvin
                     ELSE
                        WRITE (print_unit, '(A,3X)', advance='no') "N"
                        WRITE (print_unit, '(18X,A)') "--"
                     END IF
                  END DO
                  DEALLOCATE (region_id)
                  DEALLOCATE (temperature)
               END IF
               CALL cp_print_key_finished_output(print_unit, logger, thermal_regions%section, &
                                                 "PRINT%LANGEVIN_REGIONS")
            END IF
         END IF
      END IF
   END SUBROUTINE print_thermal_regions_langevin

END MODULE thermal_region_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief prints all energy info per timestep to the screen or to
!>      user defined output files
!> \author Joost VandeVondele (copy from md_fist_energies)
!>
!> \par History
!>   - New MD data are appended to the old data (15.09.2003,MK)
! **************************************************************************************************
MODULE md_energies
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE averages_types,                  ONLY: average_quantities_type,&
                                              compute_averages
   USE barostat_types,                  ONLY: barostat_type
   USE barostat_utils,                  ONLY: print_barostat_status
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              use_mixed_force
   USE input_constants,                 ONLY: npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble,&
                                              reftraj_ensemble
   USE input_cp2k_md,                   ONLY: create_md_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE md_conserved_quantities,         ONLY: calc_nfree_qm,&
                                              compute_conserved_quantity
   USE md_ener_types,                   ONLY: md_ener_type,&
                                              zero_md_ener
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type,&
                                              set_md_env
   USE motion_utils,                    ONLY: write_simulation_cell,&
                                              write_stress_tensor,&
                                              write_trajectory
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: write_structure_data
   USE physcon,                         ONLY: angstrom,&
                                              femtoseconds,&
                                              kelvin
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE reftraj_types,                   ONLY: reftraj_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermal_region_types,            ONLY: thermal_regions_type
   USE thermal_region_utils,            ONLY: print_thermal_regions_temperature
   USE thermostat_types,                ONLY: thermostats_type
   USE thermostat_utils,                ONLY: print_thermostats_status
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_energies'

   PUBLIC :: initialize_md_ener, &
             md_energy, &
             md_ener_reftraj, &
             md_write_output

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param md_ener ...
!> \param force_env ...
!> \param simpar ...
!> \par History
!>   - 10-2007 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE initialize_md_ener(md_ener, force_env, simpar)

      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(simpar_type), POINTER                         :: simpar

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_md_ener', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nkind
      LOGICAL                                            :: shell_adiabatic
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles, shell_particles

      NULLIFY (subsys)
      NULLIFY (atomic_kinds, atomic_kind_set, particles, shell_particles)

      IF(.NOT.(ASSOCIATED(md_ener)))CALL cp__a("motion/md_energies.F",124)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/md_energies.F",125)

      CALL force_env_get(force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, particles=particles, &
                         shell_particles=shell_particles)
      atomic_kind_set => atomic_kinds%els
      nkind = SIZE(atomic_kind_set)
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               shell_adiabatic=shell_adiabatic)

      md_ener%nfree = simpar%nfree
      md_ener%nfree_shell = -HUGE(0)

      IF (shell_adiabatic) THEN
         md_ener%nfree_shell = 3*(shell_particles%n_els)
      END IF

      IF (simpar%temperature_per_kind) THEN
         ALLOCATE (md_ener%temp_kind(nkind))
         ALLOCATE (md_ener%ekin_kind(nkind))
         ALLOCATE (md_ener%nfree_kind(nkind))
         md_ener%nfree_kind = 0

         IF (shell_adiabatic) THEN
            ALLOCATE (md_ener%temp_shell_kind(nkind))
            ALLOCATE (md_ener%ekin_shell_kind(nkind))
            ALLOCATE (md_ener%nfree_shell_kind(nkind))
            md_ener%nfree_shell_kind = 0
         END IF

      END IF
      CALL zero_md_ener(md_ener, tkind=simpar%temperature_per_kind, &
                        tshell=shell_adiabatic)
      md_ener%epot = 0.0_dp

   END SUBROUTINE initialize_md_ener

! **************************************************************************************************
!> \brief ...
!> \param md_env ...
!> \param md_ener ...
!> \par History
!>   - 10-2007 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE md_energy(md_env, md_ener)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener

      CHARACTER(len=*), PARAMETER :: routineN = 'md_energy', routineP = moduleN//':'//routineN

      INTEGER                                            :: natom
      LOGICAL                                            :: shell_adiabatic, tkind, tshell
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(simpar_type), POINTER                         :: simpar

      NULLIFY (atomic_kinds, atomic_kind_set, force_env, &
               particles, subsys, simpar)
      CALL get_md_env(md_env=md_env, force_env=force_env, &
                      simpar=simpar)

      CALL force_env_get(force_env, &
                         potential_energy=md_ener%epot, subsys=subsys)

      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds)
      atomic_kind_set => atomic_kinds%els
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               shell_adiabatic=shell_adiabatic)

      tkind = simpar%temperature_per_kind
      tshell = shell_adiabatic

      CALL cp_subsys_get(subsys, particles=particles)
      natom = particles%n_els

      CALL compute_conserved_quantity(md_env, md_ener, tkind=tkind, &
                                      tshell=tshell, natom=natom)

   END SUBROUTINE md_energy

! **************************************************************************************************
!> \brief ...
!> \param md_env ...
!> \param md_ener ...
!> \par History
!>   - 10.2007 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE md_ener_reftraj(md_env, md_ener)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener

      CHARACTER(len=*), PARAMETER :: routineN = 'md_ener_reftraj', &
         routineP = moduleN//':'//routineN

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(reftraj_type), POINTER                        :: reftraj

      CALL zero_md_ener(md_ener, tkind=.FALSE., tshell=.FALSE.)
      CALL get_md_env(md_env=md_env, force_env=force_env, reftraj=reftraj)

      IF (reftraj%info%eval_ef) THEN
         CALL force_env_get(force_env, potential_energy=md_ener%epot)
      ELSE
         md_ener%epot = reftraj%epot
         md_ener%delta_epot = (reftraj%epot-reftraj%epot0)/REAL(reftraj%natom, kind=dp)*kelvin
      END IF

   END SUBROUTINE md_ener_reftraj

! **************************************************************************************************
!> \brief This routine computes the conserved quantity, temperature
!>      and things like that and prints them out
!> \param md_env ...
!> \par History
!>   - New MD data are appended to the old data (15.09.2003,MK)
!>   - 02.2008 - Teodoro Laino [tlaino] - University of Zurich
!>               Cleaning code and collecting the many commons routines..
!> \author CJM
! **************************************************************************************************
   SUBROUTINE md_write_output(md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'md_write_output', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: fmd, my_act, my_pos
      INTEGER                                            :: ene, ener_mix, handle, i, nat, nkind, &
                                                            shene, tempkind, trsl
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: init, is_mixed, new_file, qmmm, &
                                                            shell_adiabatic, shell_present
      REAL(dp)                                           :: abc(3), cell_angle(3), dt, econs, &
                                                            pv_scalar, pv_xx, pv_xx_nc
      REAL(KIND=dp)                                      :: harm_shell, hugoniot
      REAL(KIND=dp), POINTER                             :: time, used_time
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(average_quantities_type), POINTER             :: averages
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(section_vals_type), POINTER                   :: motion_section, print_key, root_section
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (logger)
      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)

      ! Zeroing
      hugoniot = 0.0_dp
      econs = 0.0_dp
      shell_adiabatic = .FALSE.
      shell_present = .FALSE.
      NULLIFY (motion_section, atomic_kinds, atomic_kind_set, cell, subsys, &
               force_env, md_ener, qmmm_env, reftraj, core_particles, particles, &
               shell_particles, print_key, root_section, simpar, virial, &
               thermostats, thermal_regions)

      CALL get_md_env(md_env=md_env, itimes=itimes, t=time, used_time=used_time, &
                      simpar=simpar, force_env=force_env, init=init, md_ener=md_ener, &
                      reftraj=reftraj, thermostats=thermostats, barostat=barostat, &
                      para_env=para_env, averages=averages, thermal_regions=thermal_regions)

      root_section => force_env%root_section
      motion_section => section_vals_get_subs_vals(root_section, "MOTION")

      CALL force_env_get(force_env, cell=cell, subsys=subsys, qmmm_env=qmmm_env)

      qmmm = calc_nfree_qm(md_env, md_ener) > 0
      is_mixed = (force_env%in_use == use_mixed_force)

      CALL cp_subsys_get(subsys, particles=particles, virial=virial)
      nat = particles%n_els
      dt = simpar%dt*simpar%dt_fact

      ! Computing the scalar pressure
      IF (virial%pv_availability) THEN
         pv_scalar = 0._dp
         DO i = 1, 3
            pv_scalar = pv_scalar+virial%pv_total(i, i)
         END DO
         pv_scalar = pv_scalar/3._dp/cell%deth
         pv_scalar = cp_unit_from_cp2k(pv_scalar, "bar")
         pv_xx_nc = virial%pv_total(1, 1)/cell%deth
         pv_xx = cp_unit_from_cp2k(virial%pv_total(1, 1)/cell%deth, "bar")
      ENDIF

      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds)
      atomic_kind_set => atomic_kinds%els
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               shell_present=shell_present, &
                               shell_adiabatic=shell_adiabatic)

      CALL get_cell(cell, abc=abc, alpha=cell_angle(3), beta=cell_angle(2), gamma=cell_angle(1))

      ! Determine POS and ACT for I/O
      my_pos = "APPEND"
      my_act = "WRITE"
      IF (init .AND. (itimes == 0)) THEN
         my_pos = "REWIND"
         my_act = "WRITE"
      END IF

      ! In case of REFTRAJ ensemble the POS is determined differently..
      ! according the REFTRAJ counter
      IF (simpar%ensemble == reftraj_ensemble) THEN
         IF ((reftraj%isnap == reftraj%info%first_snapshot)) THEN
            my_pos = "REWIND"
         END IF
      ENDIF

      ! Performing protocol relevant to the first step of an MD run
      IF (init) THEN
         ! Computing the Hugoniot for NPH calculations
         IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
             simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
            IF (simpar%e0 == 0._dp) simpar%e0 = md_ener%epot+md_ener%ekin
            hugoniot = md_ener%epot+md_ener%ekin-simpar%e0-0.5_dp*(pv_xx_nc+simpar%p0)* &
                       (simpar%v0-cell%deth)
         ENDIF

         IF (simpar%ensemble == reftraj_ensemble) reftraj%init = init
      ELSE
         ! Performing protocol for anything beyond the first step of MD
         IF (simpar%ensemble == nph_uniaxial_ensemble .OR. simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
            hugoniot = md_ener%epot+md_ener%ekin-simpar%e0-0.5_dp*(pv_xx_nc+simpar%p0)* &
                       (simpar%v0-cell%deth)
         END IF

         IF (simpar%ensemble == reftraj_ensemble) THEN
            time = reftraj%time
            econs = md_ener%delta_epot
            itimes = reftraj%itimes
         ELSE
            econs = md_ener%delta_cons
         END IF

         ! Compute average quantities
         CALL compute_averages(averages, force_env, md_ener, cell, virial, pv_scalar, &
                               pv_xx, used_time, hugoniot, abc, cell_angle, nat, itimes, time, my_pos, &
                               my_act)
      END IF

      ! Print md information
      CALL md_write_info_low(simpar, md_ener, qmmm, virial, reftraj, cell, abc, &
                             cell_angle, itimes, dt, time, used_time, averages, econs, pv_scalar, pv_xx, &
                             hugoniot, nat, init, logger, motion_section, my_pos, my_act)

      ! Real Ouput driven by the PRINT sections
      IF ((.NOT. init) .OR. (itimes == 0) .OR. simpar%ensemble == reftraj_ensemble) THEN
         ! Print Energy
         ene = cp_print_key_unit_nr(logger, motion_section, "MD%PRINT%ENERGY", &
                                    extension=".ener", file_position=my_pos, file_action=my_act, is_new_file=new_file)
         IF (ene > 0) THEN
            IF (new_file) THEN
               ! Please change also the corresponding format explaination below
               ! keep the constant of motion the true constant of motion !
               WRITE (ene, '("#",5X,A,10X,A,8X,A,10X,A,12X,A,2(8X,A))') "Step Nr.", "Time[fs]", "Kin.[a.u.]", "Temp[K]", &
                  "Pot.[a.u.]", "Cons Qty[a.u.]", "UsedTime[s]"
            END IF
            WRITE (ene, "(I10,F20.6,F20.9,F20.9,F20.9,F20.9,F20.9)") &
               itimes, time*femtoseconds, md_ener%ekin, md_ener%temp_part, md_ener%epot, md_ener%constant, used_time
            CALL m_flush(ene)
         END IF
         CALL cp_print_key_finished_output(ene, logger, motion_section, "MD%PRINT%ENERGY")

         ! Possibly Print MIXED Energy
         IF (is_mixed) THEN
            ener_mix = cp_print_key_unit_nr(logger, motion_section, "PRINT%MIXED_ENERGIES", &
                                            extension=".ener", file_position=my_pos, file_action=my_act, &
                                            middle_name="mix")
            IF (ener_mix > 0) THEN
               WRITE (ener_mix, "(I8,F12.3,F20.9,"//cp_to_string(SIZE(force_env%mixed_env%energies))//"F20.9,F20.9)") &
                  itimes, time*femtoseconds, md_ener%epot, force_env%mixed_env%energies, md_ener%constant
               CALL m_flush(ener_mix)
            END IF
            CALL cp_print_key_finished_output(ener_mix, logger, motion_section, "PRINT%MIXED_ENERGIES")
         ENDIF

         ! Print QMMM translation vector if requested
         IF (qmmm) THEN
            trsl = cp_print_key_unit_nr(logger, motion_section, "PRINT%TRANSLATION_VECTOR", &
                                        extension=".translation", middle_name="qmmm")
            IF (trsl > 0) THEN
               WRITE (trsl, '(I10,3F15.10)') itimes, qmmm_env%qm%transl_v
            END IF
            CALL cp_print_key_finished_output(trsl, logger, motion_section, &
                                              "PRINT%TRANSLATION_VECTOR")
         END IF

         ! Write Structure data
         CALL write_structure_data(particles%els, cell, motion_section)

         ! Print Coordinates
         CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                               pos=my_pos, act=my_act, extended_xmol_title=.TRUE.)

         ! Print Velocities
         CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                               "VELOCITIES", my_pos, my_act, middle_name="vel", extended_xmol_title=.TRUE.)

         ! Print Force
         CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                               "FORCES", my_pos, my_act, middle_name="frc", extended_xmol_title=.TRUE.)

         ! Print Force-Mixing labels
         CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                               "FORCE_MIXING_LABELS", my_pos, my_act, middle_name="fmlabels", extended_xmol_title=.TRUE.)

         ! Print Simulation Cell
         CALL write_simulation_cell(cell, motion_section, itimes, time*femtoseconds, my_pos, my_act)

         ! Print Thermostats status
         CALL print_thermostats_status(thermostats, para_env, my_pos, my_act, itimes, time)

         ! Print Barostat status
         CALL print_barostat_status(barostat, simpar, my_pos, my_act, cell, itimes, time)

         ! Print Stress Tensor
         CALL write_stress_tensor(virial, cell, motion_section, itimes, time*femtoseconds, &
                                  my_pos, my_act)

         ! Temperature per Kinds
         IF (simpar%temperature_per_kind) THEN
            tempkind = cp_print_key_unit_nr(logger, motion_section, "MD%PRINT%TEMP_KIND", &
                                            extension=".temp", file_position=my_pos, file_action=my_act)
            IF (tempkind > 0) THEN
               nkind = SIZE(md_ener%temp_kind)
               fmd = "(I10,F20.3,"//TRIM(ADJUSTL(cp_to_string(nkind)))//"F20.9)"
               fmd = TRIM(fmd)
               WRITE (tempkind, fmd) itimes, time*femtoseconds, md_ener%temp_kind(1:nkind)
               CALL m_flush(tempkind)
            END IF
            CALL cp_print_key_finished_output(tempkind, logger, motion_section, "MD%PRINT%TEMP_KIND")
         ELSE
            print_key => section_vals_get_subs_vals(motion_section, "MD%PRINT%TEMP_KIND")
            IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) &
               CALL cp_warn(cp__l("motion/md_energies.F",480), &
                            "The print_key MD%PRINT%TEMP_KIND has been activated but the "// &
                            "calculation of the temperature per kind has not been requested. "// &
                            "Please switch on the keyword MD%TEMP_KIND.")
         END IF
         !Thermal Region
         CALL print_thermal_regions_temperature(thermal_regions, itimes, time*femtoseconds, my_pos, my_act)

         ! Core/Shell Model
         IF (shell_present) THEN
            CALL force_env_get(force_env, harmonic_shell=harm_shell)
            CALL cp_subsys_get(subsys, shell_particles=shell_particles, core_particles=core_particles)

            ! Print Shell Energy
            shene = cp_print_key_unit_nr(logger, motion_section, "MD%PRINT%SHELL_ENERGY", &
                                         extension=".shener", file_position=my_pos, file_action=my_act, &
                                         file_form="FORMATTED", is_new_file=new_file)
            IF (shene > 0) THEN
               IF (new_file) THEN
                  WRITE (shene, '("#",3X,A,3X,A,3X,3(5X,A,5X))') "Step Nr.", "Time[fs]", "Kin.[a.u.]", &
                     "Temp.[K]", "Pot.[a.u.]"
               END IF

               WRITE (shene, "(I8,F12.3,F20.9,F20.9,F20.9,F20.9 )") &
                  itimes, time*femtoseconds, md_ener%ekin_shell, md_ener%temp_shell, harm_shell
               CALL m_flush(shene)
            END IF
            CALL cp_print_key_finished_output(shene, logger, motion_section, "MD%PRINT%SHELL_ENERGY")

            ! Print Shell Coordinates
            CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                  "SHELL_TRAJECTORY", my_pos, my_act, "shpos", shell_particles, extended_xmol_title=.TRUE.)

            IF (shell_adiabatic) THEN
               ! Print Shell Velocities
               CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                     "SHELL_VELOCITIES", my_pos, my_act, "shvel", shell_particles, extended_xmol_title=.TRUE.)

               ! Print Shell Forces
               CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                     "SHELL_FORCES", my_pos, my_act, "shfrc", shell_particles, extended_xmol_title=.TRUE.)

               ! Print Core Coordinates
               CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                     "CORE_TRAJECTORY", my_pos, my_act, "copos", core_particles, extended_xmol_title=.TRUE.)

               ! Print Core Velocities
               CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                     "CORE_VELOCITIES", my_pos, my_act, "covel", core_particles, extended_xmol_title=.TRUE.)

               ! Print Core Forces
               CALL write_trajectory(force_env, root_section, itimes, time*femtoseconds, dt*femtoseconds, md_ener%epot, &
                                     "CORE_FORCES", my_pos, my_act, "cofrc", core_particles, extended_xmol_title=.TRUE.)

               ! Temperature per Kinds
               IF (simpar%temperature_per_kind) THEN
                  tempkind = cp_print_key_unit_nr(logger, motion_section, "MD%PRINT%TEMP_SHELL_KIND", &
                                                  extension=".shtemp", file_position=my_pos, file_action=my_act)
                  IF (tempkind > 0) THEN
                     nkind = SIZE(md_ener%temp_shell_kind)
                     fmd = "(I10,F20.3,"//TRIM(ADJUSTL(cp_to_string(nkind)))//"F20.9)"
                     fmd = TRIM(fmd)
                     WRITE (tempkind, fmd) itimes, time*femtoseconds, md_ener%temp_shell_kind(1:nkind)
                     CALL m_flush(tempkind)
                  END IF
                  CALL cp_print_key_finished_output(tempkind, logger, motion_section, &
                                                    "MD%PRINT%TEMP_SHELL_KIND")
               ELSE
                  print_key => section_vals_get_subs_vals(motion_section, "MD%PRINT%TEMP_SHELL_KIND")
                  IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) &
                     CALL cp_warn(cp__l("motion/md_energies.F",550), &
                                  "The print_key MD%PRINT%TEMP_SHELL_KIND has been activated but the "// &
                                  "calculation of the temperature per kind has not been requested. "// &
                                  "Please switch on the keyword MD%TEMP_KIND.")
               END IF
            END IF
         END IF
      END IF
      init = .FALSE.
      CALL set_md_env(md_env, init=init)
      CALL timestop(handle)
   END SUBROUTINE md_write_output

! **************************************************************************************************
!> \brief This routine prints all basic information during MD steps
!> \param simpar ...
!> \param md_ener ...
!> \param qmmm ...
!> \param virial ...
!> \param reftraj ...
!> \param cell ...
!> \param abc ...
!> \param cell_angle ...
!> \param itimes ...
!> \param dt ...
!> \param time ...
!> \param used_time ...
!> \param averages ...
!> \param econs ...
!> \param pv_scalar ...
!> \param pv_xx ...
!> \param hugoniot ...
!> \param nat ...
!> \param init ...
!> \param logger ...
!> \param motion_section ...
!> \param my_pos ...
!> \param my_act ...
!> \par History
!>   - 10.2008 - Teodoro Laino [tlaino] - University of Zurich
!>               Refactoring: split into an independent routine.
!>               All output on screen must be included in this function!
!> \author CJM
! **************************************************************************************************
   SUBROUTINE md_write_info_low(simpar, md_ener, qmmm, virial, reftraj, cell, &
                                abc, cell_angle, itimes, dt, time, used_time, averages, econs, pv_scalar, &
                                pv_xx, hugoniot, nat, init, logger, motion_section, my_pos, my_act)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(md_ener_type), POINTER                        :: md_ener
      LOGICAL, INTENT(IN)                                :: qmmm
      TYPE(virial_type), POINTER                         :: virial
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: abc, cell_angle
      INTEGER, POINTER                                   :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: dt
      REAL(KIND=dp), POINTER                             :: time, used_time
      TYPE(average_quantities_type), POINTER             :: averages
      REAL(KIND=dp), INTENT(IN)                          :: econs, pv_scalar, pv_xx, hugoniot
      INTEGER, INTENT(IN)                                :: nat
      LOGICAL, INTENT(IN)                                :: init
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: motion_section
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: my_pos, my_act

      CHARACTER(len=*), PARAMETER :: routineN = 'md_write_info_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section

      NULLIFY (enum, keyword, section)
      ! Print to the screen info about MD
      iw = cp_print_key_unit_nr(logger, motion_section, "MD%PRINT%PROGRAM_RUN_INFO", &
                                extension=".mdLog", file_position=my_pos, file_action=my_act)

      ! Performing protocol relevant to the first step of an MD run
      IF (iw > 0) THEN
         CALL create_md_section(section)
         keyword => section_get_keyword(section, "ENSEMBLE")
         CALL keyword_get(keyword, enum=enum)

         IF (init) THEN
            ! Write initial values of quantities of interest
            WRITE (iw, *)
            WRITE (iw, '( A )') ' MD_ENERGIES| Initialization proceeding'
            WRITE (iw, *)
            WRITE (iw, '( )')
            WRITE (iw, '( A,A )') ' ******************************** ', &
               'GO CP2K GO! **********************************'
            WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL POTENTIAL ENERGY', &
               '[hartree]', '= ', md_ener%epot
            IF (simpar%ensemble /= reftraj_ensemble) THEN
               IF (.NOT. qmmm) THEN
                  ! NO QM/MM info
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL KINETIC ENERGY', &
                     '[hartree]', '= ', md_ener%ekin
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,F20.3) )') ' INITIAL TEMPERATURE', &
                     '[K]', '= ', md_ener%temp_part
               ELSE
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' TOTAL INITIAL KINETIC ENERGY', &
                     '[hartree]', '= ', md_ener%ekin
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' QM INITIAL KINETIC ENERGY', &
                     '[hartree]', '= ', md_ener%ekin_qm
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,F20.3) )') ' TOTAL INITIAL TEMPERATURE', &
                     '[K]', '= ', md_ener%temp_part
                  WRITE (iw, '( A,A,T40,A,T60,1(1X,F20.3) )') ' QM INITIAL TEMPERATURE', &
                     '[K]', '= ', md_ener%temp_qm
               END IF
            END IF
            IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
                simpar%ensemble == nph_uniaxial_damped_ensemble .OR. &
                simpar%ensemble == npt_i_ensemble .OR. &
                simpar%ensemble == npt_f_ensemble .OR. &
                simpar%ensemble == npe_i_ensemble .OR. &
                simpar%ensemble == npe_f_ensemble) &
               WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL BAROSTAT TEMP', &
               '[K]', '= ', md_ener%temp_baro
            IF (virial%pv_availability) &
               WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL PRESSURE', &
               '[bar]', '= ', pv_scalar
            IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
                simpar%ensemble == nph_uniaxial_damped_ensemble) &
               WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL HUGONIOT CONSTRAINT', &
               '[K]', '= ', hugoniot
            IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
                simpar%ensemble == nph_uniaxial_damped_ensemble) &
               WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL E0', &
               '[hartree]', '= ', simpar%e0
            WRITE (iw, '( A,A,T40,A,T60,1(1X,E20.12) )') ' INITIAL VOLUME', &
               '[bohr^3]', '= ', cell%deth
            WRITE (iw, '( A,A,T29,A,T33,3(1X,E15.7) )') ' INITIAL CELL LNTHS', &
               '[bohr]', '= ', abc(1), abc(2), abc(3)
            WRITE (iw, '( A,A,T29,A,T33,3(1X,E15.7) )') ' INITIAL CELL ANGLS', &
               '[deg]', '= ', cell_angle(3), cell_angle(2), cell_angle(1)
            WRITE (iw, '( A,A )') ' ******************************** ', &
               'GO CP2K GO! **********************************'
         ELSE
            ! Write seuquential values of quantities of interest
            WRITE (iw, '(/,T2,A)') REPEAT('*', 79)
            WRITE (iw, '(T2,A,T61,A20)') &
               'ENSEMBLE TYPE                = ', ADJUSTR(TRIM(enum_i2c(enum, simpar%ensemble)))
            WRITE (iw, '(T2,A,T71,I10)') &
               'STEP NUMBER                  = ', itimes
            IF (simpar%variable_dt) THEN
               WRITE (iw, '(T2,A,T60,1X,F20.6)') &
                  'TIME STEP [fs]               = ', dt*femtoseconds
            END IF
            WRITE (iw, '(T2,A,T60,1X,F20.6)') &
               'TIME [fs]                    = ', time*femtoseconds
            WRITE (iw, '(T2,A,T60,1X,E20.12)') &
               'CONSERVED QUANTITY [hartree] = ', md_ener%constant
            WRITE (iw, '(/,T47,A,T73,A)') 'INSTANTANEOUS', 'AVERAGES'
            WRITE (iw, '(T2,A,T39,2(1X,F20.2))') &
               'CPU TIME [s]                 = ', used_time, averages%avecpu
            WRITE (iw, '(T2,A,T39,2(1X,E20.12))') &
               'ENERGY DRIFT PER ATOM [K]    = ', econs, averages%econs
            WRITE (iw, '(T2,A,T39,2(1X,E20.12))') &
               'POTENTIAL ENERGY[hartree]    = ', md_ener%epot, averages%avepot
            IF (simpar%ensemble /= reftraj_ensemble) THEN
               IF (.NOT. qmmm) THEN
                  ! No QM/MM info
                  WRITE (iw, '(T2,A,T39,2(1X,E20.12))') &
                     'KINETIC ENERGY [hartree]     = ', md_ener%ekin, averages%avekin
                  WRITE (iw, '(T2,A,T39,2(1X,F20.3))') &
                     'TEMPERATURE [K]              = ', md_ener%temp_part, averages%avetemp
               ELSE
                  WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' TOTAL KINETIC ENERGY', &
                     '[hartree]', '= ', md_ener%ekin, averages%avekin
                  WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' QM KINETIC ENERGY', &
                     '[hartree]', '= ', md_ener%ekin_qm, averages%avekin_qm
                  WRITE (iw, '( A,A,T31,A,T39,2(1X,F20.3) )') ' TOTAL TEMPERATURE', &
                     '[K]', '= ', md_ener%temp_part, averages%avetemp
                  WRITE (iw, '( A,A,T31,A,T39,2(1X,F20.3) )') ' QM TEMPERATURE', &
                     '[K]', '= ', md_ener%temp_qm, averages%avetemp_qm
               END IF
            END IF
            IF (virial%pv_availability) THEN
               WRITE (iw, '(T2,A,T39,2(1X,E20.12))') &
                  'PRESSURE [bar]               = ', pv_scalar, averages%avepress
            END IF
            IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
                simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
               WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' P_XX', &
                  '[bar]', '= ', pv_xx, averages%avepxx
               WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' HUGONIOT', &
                  '[K]', '= ', hugoniot/3._dp/nat*kelvin, &
                  averages%avehugoniot/3._dp/nat*kelvin
            END IF
            IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
                simpar%ensemble == nph_uniaxial_damped_ensemble .OR. &
                simpar%ensemble == npt_i_ensemble .OR. &
                simpar%ensemble == npt_f_ensemble .OR. &
                simpar%ensemble == npe_i_ensemble .OR. &
                simpar%ensemble == npe_f_ensemble) THEN
               WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' BAROSTAT TEMP', &
                  '[K]', '= ', md_ener%temp_baro, averages%avetemp_baro
               WRITE (iw, '( A,A,T31,A,T39,2(1X,E20.12) )') ' VOLUME', &
                  '[bohr^3]', '= ', cell%deth, averages%avevol
               WRITE (iw, '( A,A,T31,A,T33,3(1X,E15.7) )') ' CELL LNTHS', &
                  '[bohr]', '= ', abc(1), abc(2), abc(3)
               WRITE (iw, '( A,A,T31,A,T33,3(1X,E15.7) )') ' AVE. CELL LNTHS', &
                  '[bohr]', '= ', averages%aveca, averages%avecb, &
                  averages%avecc
            END IF
            IF (simpar%ensemble == npt_f_ensemble .OR. &
                simpar%ensemble == npe_f_ensemble) THEN
               WRITE (iw, '( A,A,T31,A,T33,3(1X,E15.7) )') ' CELL ANGLS', &
                  '[deg]', '= ', cell_angle(3), cell_angle(2), cell_angle(1)
               WRITE (iw, '( A,A,T31,A,T33,3(1X,E15.7) )') ' AVE. CELL ANGLS', &
                  '[deg]', '= ', averages%aveal, averages%avebe, averages%avega
            END IF
            IF (simpar%ensemble == reftraj_ensemble) THEN
               IF (reftraj%info%msd) THEN
                  IF (reftraj%msd%msd_kind) &
                     WRITE (iw, '(/,A,T50,3f10.5)') ' COM displacement (dx,dy,dz) [angstrom]:  ', &
                     reftraj%msd%drcom(1:3)*angstrom
               END IF
            END IF
            WRITE (iw, '(T2,A,/)') REPEAT('*', 79)
         END IF
      END IF
      CALL section_release(section)
      CALL cp_print_key_finished_output(iw, logger, motion_section, &
                                        "MD%PRINT%PROGRAM_RUN_INFO")
   END SUBROUTINE md_write_info_low

END MODULE md_energies
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides integrator utility routines for the integrators
!> \par History
!>      Teodoro Laino [tlaino] 02.2008 - Splitting from integrator and creation
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
! **************************************************************************************************
MODULE integrator_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE barostat_types,                  ONLY: do_clv_x,&
                                              do_clv_xy,&
                                              do_clv_xyz,&
                                              do_clv_xz,&
                                              do_clv_y,&
                                              do_clv_yz,&
                                              do_clv_z
   USE cell_types,                      ONLY: cell_type
   USE constraint,                      ONLY: rattle_roll_control
   USE constraint_util,                 ONLY: pv_constraint
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: debug_isotropic_limit,&
                                              debug_uniaxial_limit,&
                                              npt_info_type
   USE input_constants,                 ONLY: &
        isokin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_uniaxial_damped_ensemble, &
        nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, nve_ensemble, nvt_ensemble
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: matvec_3x3,&
                                              transpose_3d
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type,&
                                              update_particle_set
   USE shell_potential_types,           ONLY: shell_kind_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: set_thermostats,&
                                              thermostats_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'integrator_utils'

! **************************************************************************************************
   TYPE old_variables_type
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: v
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: r
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: eps
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: veps
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: h
   END TYPE old_variables_type

! **************************************************************************************************
   TYPE tmp_variables_type
      INTEGER, POINTER :: itimes
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: pos
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: vel
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: shell_pos
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: shell_vel
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: core_pos
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: core_vel
      REAL(KIND=dp) :: max_vel, max_vel_sc
      REAL(KIND=dp) :: max_dvel, max_dvel_sc
      REAL(KIND=dp) :: max_dr, max_dsc
      REAL(KIND=dp) :: arg_r(3), arg_v(3), u(3, 3), e_val(3), s, ds
      REAL(KIND=dp), DIMENSION(3) :: poly_r, &
                                     poly_v, scale_r, scale_v
   END TYPE tmp_variables_type

   INTERFACE set
      MODULE PROCEDURE set_particle_set, set_vel
   END INTERFACE

   INTERFACE update_pv
      MODULE PROCEDURE update_pv_particle_set, update_pv_velocity
   END INTERFACE

   INTERFACE damp_v
      MODULE PROCEDURE damp_v_particle_set, damp_v_velocity
   END INTERFACE

   PUBLIC :: old_variables_type, tmp_variables_type, allocate_old, deallocate_old, &
             allocate_tmp, update_dealloc_tmp, damp_v, set, update_pv, get_s_ds, &
             rattle_roll_setup, damp_veps, update_veps, vv_first, vv_second, &
             variable_timestep

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param old ...
!> \param particle_set ...
!> \param npt ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE allocate_old(old, particle_set, npt)
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(npt_info_type), POINTER                       :: npt(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_old', routineP = moduleN//':'//routineN

      INTEGER                                            :: idim, jdim, natoms

      natoms = SIZE(particle_set)
      idim = SIZE(npt, 1)
      jdim = SIZE(npt, 2)
      IF(.NOT.(.NOT. ASSOCIATED(old)))CALL cp__a("motion/integrator_utils.F",126)
      ALLOCATE (old)

      ALLOCATE (old%v(natoms, 3))
      old%v = 0.0_dp
      ALLOCATE (old%r(natoms, 3))
      old%r = 0.0_dp
      ALLOCATE (old%eps(idim, jdim))
      old%eps = 0.0_dp
      ALLOCATE (old%veps(idim, jdim))
      old%veps = 0.0_dp
      ALLOCATE (old%h(3, 3))
      old%h = 0.0_dp

   END SUBROUTINE allocate_old

! **************************************************************************************************
!> \brief ...
!> \param old ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE deallocate_old(old)
      TYPE(old_variables_type), POINTER                  :: old

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_old', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(old)) THEN
         IF (ASSOCIATED(old%v)) THEN
            DEALLOCATE (old%v)
         END IF
         IF (ASSOCIATED(old%r)) THEN
            DEALLOCATE (old%r)
         END IF
         IF (ASSOCIATED(old%eps)) THEN
            DEALLOCATE (old%eps)
         END IF
         IF (ASSOCIATED(old%veps)) THEN
            DEALLOCATE (old%veps)
         END IF
         IF (ASSOCIATED(old%h)) THEN
            DEALLOCATE (old%h)
         END IF
         DEALLOCATE (old)
      END IF

   END SUBROUTINE deallocate_old

! **************************************************************************************************
!> \brief allocate temporary variables to store positions and velocities
!>        used by the velocity-verlet integrator
!> \param md_env ...
!> \param tmp ...
!> \param nparticle ...
!> \param nshell ...
!> \param shell_adiabatic ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE allocate_tmp(md_env, tmp, nparticle, nshell, shell_adiabatic)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(tmp_variables_type), POINTER                  :: tmp
      INTEGER, INTENT(IN)                                :: nparticle, nshell
      LOGICAL, INTENT(IN)                                :: shell_adiabatic

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_tmp', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(tmp)))CALL cp__a("motion/integrator_utils.F",195)
      ALLOCATE (tmp)

      ! Nullify Components
      NULLIFY (tmp%pos)
      NULLIFY (tmp%vel)
      NULLIFY (tmp%shell_pos)
      NULLIFY (tmp%shell_vel)
      NULLIFY (tmp%core_pos)
      NULLIFY (tmp%core_vel)
      NULLIFY (tmp%itimes)

      !     *** Allocate work storage for positions and velocities ***
      ALLOCATE (tmp%pos(3, nparticle))

      ALLOCATE (tmp%vel(3, nparticle))

      tmp%pos(:, :) = 0.0_dp
      tmp%vel(:, :) = 0.0_dp

      IF (shell_adiabatic) THEN
         !     *** Allocate work storage for positions and velocities ***
         ALLOCATE (tmp%shell_pos(3, nshell))

         ALLOCATE (tmp%core_pos(3, nshell))

         ALLOCATE (tmp%shell_vel(3, nshell))

         ALLOCATE (tmp%core_vel(3, nshell))

         tmp%shell_pos(:, :) = 0.0_dp
         tmp%shell_vel(:, :) = 0.0_dp
         tmp%core_pos(:, :) = 0.0_dp
         tmp%core_vel(:, :) = 0.0_dp
      END IF

      tmp%arg_r = 0.0_dp
      tmp%arg_v = 0.0_dp
      tmp%u = 0.0_dp
      tmp%e_val = 0.0_dp
      tmp%max_vel = 0.0_dp
      tmp%max_vel_sc = 0.0_dp
      tmp%max_dvel = 0.0_dp
      tmp%max_dvel_sc = 0.0_dp
      tmp%scale_r = 1.0_dp
      tmp%scale_v = 1.0_dp
      tmp%poly_r = 1.0_dp
      tmp%poly_v = 1.0_dp

      CALL get_md_env(md_env=md_env, itimes=tmp%itimes)

   END SUBROUTINE allocate_tmp

! **************************************************************************************************
!> \brief update positions and deallocate temporary variable
!> \param tmp ...
!> \param particle_set ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param para_env ...
!> \param shell_adiabatic ...
!> \param pos ...
!> \param vel ...
!> \param should_deall_vel ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE update_dealloc_tmp(tmp, particle_set, shell_particle_set, &
                                 core_particle_set, para_env, shell_adiabatic, pos, vel, &
                                 should_deall_vel)

      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, shell_particle_set, &
                                                            core_particle_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      LOGICAL, INTENT(IN), OPTIONAL                      :: pos, vel, should_deall_vel

      CHARACTER(len=*), PARAMETER :: routineN = 'update_dealloc_tmp', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_deall, my_pos, my_vel

      IF(.NOT.(ASSOCIATED(tmp)))CALL cp__a("motion/integrator_utils.F",279)
      my_pos = .FALSE.
      my_vel = .FALSE.
      my_deall = .TRUE.
      IF (PRESENT(pos)) my_pos = pos
      IF (PRESENT(vel)) my_vel = vel
      IF (PRESENT(should_deall_vel)) my_deall = should_deall_vel

      !      *** Broadcast the new particle positions ***
      IF (my_pos) THEN
         CALL update_particle_set(particle_set, para_env%group, pos=tmp%pos)
         DEALLOCATE (tmp%pos)

         IF (shell_adiabatic) THEN
            CALL update_particle_set(shell_particle_set, para_env%group, pos=tmp%shell_pos)
            CALL update_particle_set(core_particle_set, para_env%group, pos=tmp%core_pos)
            DEALLOCATE (tmp%shell_pos)
            DEALLOCATE (tmp%core_pos)
         END IF
      END IF

      !     *** Broadcast the new particle velocities ***
      IF (my_vel) THEN
         CALL update_particle_set(particle_set, para_env%group, vel=tmp%vel)
         IF (shell_adiabatic) THEN
            CALL update_particle_set(shell_particle_set, para_env%group, vel=tmp%shell_vel)
            CALL update_particle_set(core_particle_set, para_env%group, vel=tmp%core_vel)
         END IF
         IF (my_deall) THEN
            DEALLOCATE (tmp%vel)
            IF (shell_adiabatic) THEN
               DEALLOCATE (tmp%shell_vel)
               DEALLOCATE (tmp%core_vel)
            END IF
            IF(.NOT.(.NOT. ASSOCIATED(tmp%pos)))CALL cp__a("motion/integrator_utils.F",313)
            IF(.NOT.(.NOT. ASSOCIATED(tmp%shell_pos)))CALL cp__a("motion/integrator_utils.F",314)
            IF(.NOT.(.NOT. ASSOCIATED(tmp%core_pos)))CALL cp__a("motion/integrator_utils.F",315)
            DEALLOCATE (tmp)
         END IF
      END IF

   END SUBROUTINE update_dealloc_tmp

! **************************************************************************************************
!> \brief ...
!> \param tmp ...
!> \param nparticle_kind ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param dt ...
!> \param para_env ...
!> \param tmpv ...
!> \par History
!>      - Created [2004-07]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE get_s_ds(tmp, nparticle_kind, atomic_kind_set, local_particles, particle_set, &
                       dt, para_env, tmpv)
      TYPE(tmp_variables_type), POINTER                  :: tmp
      INTEGER                                            :: nparticle_kind
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp)                                      :: dt
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: tmpv

      CHARACTER(len=*), PARAMETER :: routineN = 'get_s_ds', routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_local
      LOGICAL                                            :: my_tmpv
      REAL(KIND=dp)                                      :: a, b, K, mass, rb
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      my_tmpv = .FALSE.
      IF (PRESENT(tmpv)) my_tmpv = tmpv

      K = 0.0_dp
      a = 0.0_dp
      b = 0.0_dp
      DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         IF (mass /= 0.0_dp) THEN
            nparticle_local = local_particles%n_el(iparticle_kind)
            IF (my_tmpv) THEN
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                  K = K+0.5_dp*mass*DOT_PRODUCT(tmp%vel(:, iparticle), tmp%vel(:, iparticle))
                  a = a+DOT_PRODUCT(tmp%vel(:, iparticle), particle_set(iparticle)%f(:))
                  b = b+(1.0_dp/mass)*DOT_PRODUCT(particle_set(iparticle)%f(:), particle_set(iparticle)%f(:))
               END DO
            ELSE
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                  K = K+0.5_dp*mass*DOT_PRODUCT(particle_set(iparticle)%v(:), particle_set(iparticle)%v(:))
                  a = a+DOT_PRODUCT(particle_set(iparticle)%v(:), particle_set(iparticle)%f(:))
                  b = b+(1.0_dp/mass)*DOT_PRODUCT(particle_set(iparticle)%f(:), particle_set(iparticle)%f(:))
               END DO
            END IF
         END IF
      END DO
      CALL mp_sum(K, para_env%group)
      CALL mp_sum(a, para_env%group)
      CALL mp_sum(b, para_env%group)
      a = a/(2.0_dp*K)
      b = b/(2.0_dp*K)
      rb = SQRT(b)
      tmp%s = (a/b)*(COSH(dt*rb/2.0_dp)-1)+SINH(dt*rb/2.0_dp)/rb
      tmp%ds = (a/b)*(SINH(dt*rb/2.0_dp)*rb)+COSH(dt*rb/2.0_dp)

   END SUBROUTINE get_s_ds

! **************************************************************************************************
!> \brief ...
!> \param old ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param cell ...
!> \param npt ...
!> \param char ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE set_particle_set(old, atomic_kind_set, particle_set, local_particles, cell, npt, char)
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(cell_type), POINTER                           :: cell
      TYPE(npt_info_type), DIMENSION(:, :), POINTER      :: npt
      CHARACTER(LEN=*), INTENT(IN)                       :: char

      CHARACTER(len=*), PARAMETER :: routineN = 'set_particle_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idim, iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local

      nparticle_kind = SIZE(atomic_kind_set)
      SELECT CASE (char)
      CASE ('F') ! forward assigning the old
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               DO idim = 1, 3
                  old%v(iparticle, idim) = particle_set(iparticle)%v(idim)
                  old%r(iparticle, idim) = particle_set(iparticle)%r(idim)
               END DO
            END DO
         END DO
         old%eps(:, :) = npt(:, :)%eps
         old%veps(:, :) = npt(:, :)%v
         old%h(:, :) = cell%hmat(:, :)
      CASE ('B') ! back assigning the original variables
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               DO idim = 1, 3
                  particle_set(iparticle)%v(idim) = old%v(iparticle, idim)
                  particle_set(iparticle)%r(idim) = old%r(iparticle, idim)
               END DO
            END DO
         END DO
         npt(:, :)%eps = old%eps(:, :)
         npt(:, :)%v = old%veps(:, :)
         cell%hmat(:, :) = old%h(:, :)
      END SELECT

   END SUBROUTINE set_particle_set

! **************************************************************************************************
!> \brief ...
!> \param old ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param vel ...
!> \param local_particles ...
!> \param cell ...
!> \param npt ...
!> \param char ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE set_vel(old, atomic_kind_set, particle_set, vel, local_particles, cell, npt, char)
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(cell_type), POINTER                           :: cell
      TYPE(npt_info_type), DIMENSION(:, :), POINTER      :: npt
      CHARACTER(LEN=*), INTENT(IN)                       :: char

      CHARACTER(len=*), PARAMETER :: routineN = 'set_vel', routineP = moduleN//':'//routineN

      INTEGER                                            :: idim, iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local

      nparticle_kind = SIZE(atomic_kind_set)
      SELECT CASE (char)
      CASE ('F') ! forward assigning the old
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               DO idim = 1, 3
                  old%v(iparticle, idim) = vel(idim, iparticle)
                  old%r(iparticle, idim) = particle_set(iparticle)%r(idim)
               END DO
            END DO
         END DO
         old%eps(:, :) = npt(:, :)%eps
         old%veps(:, :) = npt(:, :)%v
         old%h(:, :) = cell%hmat(:, :)
      CASE ('B') ! back assigning the original variables
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               DO idim = 1, 3
                  vel(idim, iparticle) = old%v(iparticle, idim)
                  particle_set(iparticle)%r(idim) = old%r(iparticle, idim)
               END DO
            END DO
         END DO
         npt(:, :)%eps = old%eps(:, :)
         npt(:, :)%v = old%veps(:, :)
         cell%hmat(:, :) = old%h(:, :)
      END SELECT

   END SUBROUTINE set_vel

! **************************************************************************************************
!> \brief overloaded routine provides damping for particles via nph_uniaxial_damped dynamics
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param gamma1 ...
!> \param npt ...
!> \param dt ...
!> \param group ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE damp_v_particle_set(molecule_kind_set, molecule_set, &
                                  particle_set, local_molecules, gamma1, npt, dt, group)

      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      REAL(KIND=dp), INTENT(IN)                          :: gamma1
      TYPE(npt_info_type), INTENT(IN)                    :: npt
      REAL(KIND=dp), INTENT(IN)                          :: dt
      INTEGER, INTENT(IN)                                :: group

      INTEGER                                            :: first_atom, ikind, imol, imol_local, &
                                                            ipart, last_atom, nmol_local
      REAL(KIND=dp)                                      :: alpha, ikin, kin, mass, scale
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule

      kin = 0.0_dp
      DO ikind = 1, SIZE(molecule_kind_set)
         ! Compute the total kinetic energy
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, &
                              last_atom=last_atom)
            DO ipart = first_atom, last_atom
               atomic_kind => particle_set(ipart)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
               kin = kin+mass*particle_set(ipart)%v(1)* &
                     particle_set(ipart)%v(1)
               kin = kin+mass*particle_set(ipart)%v(2)* &
                     particle_set(ipart)%v(2)
               kin = kin+mass*particle_set(ipart)%v(3)* &
                     particle_set(ipart)%v(3)
            END DO
         END DO
      END DO
      !
      CALL mp_sum(kin, group)
      !
      ikin = 1.0_dp/kin
      scale = 1.0_dp
      alpha = 2.0_dp*npt%mass*npt%v*npt%v*gamma1*ikin
      scale = scale*SQRT(1.0_dp+alpha*0.5_dp*dt)
      ! Scale
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, &
                              last_atom=last_atom)
            DO ipart = first_atom, last_atom
               particle_set(ipart)%v(1) = particle_set(ipart)%v(1)*scale
               particle_set(ipart)%v(2) = particle_set(ipart)%v(2)*scale
               particle_set(ipart)%v(3) = particle_set(ipart)%v(3)*scale
            END DO
         END DO
      END DO

   END SUBROUTINE damp_v_particle_set

! **************************************************************************************************
!> \brief overloaded provides damping for particles via nph_uniaxial_damped dynamics
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param vel ...
!> \param gamma1 ...
!> \param npt ...
!> \param dt ...
!> \param group ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE damp_v_velocity(molecule_kind_set, molecule_set, &
                              particle_set, local_molecules, vel, gamma1, npt, dt, group)

      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: gamma1
      TYPE(npt_info_type), INTENT(IN)                    :: npt
      REAL(KIND=dp), INTENT(IN)                          :: dt
      INTEGER, INTENT(IN)                                :: group

      INTEGER                                            :: first_atom, ikind, imol, imol_local, &
                                                            ipart, last_atom, nmol_local
      REAL(KIND=dp)                                      :: alpha, ikin, kin, mass, scale
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule

      kin = 0.0_dp
      ! Compute the total kinetic energy
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, &
                              last_atom=last_atom)
            DO ipart = first_atom, last_atom
               atomic_kind => particle_set(ipart)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
               kin = kin+mass*vel(1, ipart)*vel(1, ipart)
               kin = kin+mass*vel(2, ipart)*vel(2, ipart)
               kin = kin+mass*vel(3, ipart)*vel(3, ipart)
            END DO
         END DO
      END DO
      !
      CALL mp_sum(kin, group)
      !
      ikin = 1.0_dp/kin
      scale = 1.0_dp
      alpha = 2.0_dp*npt%mass*npt%v*npt%v*gamma1*ikin
      scale = scale*SQRT(1.0_dp+alpha*0.5_dp*dt)
      ! Scale
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, &
                              last_atom=last_atom)
            DO ipart = first_atom, last_atom
               vel(1, ipart) = vel(1, ipart)*scale
               vel(2, ipart) = vel(2, ipart)*scale
               vel(3, ipart) = vel(3, ipart)*scale
            END DO
         END DO
      END DO

   END SUBROUTINE damp_v_velocity

! **************************************************************************************************
!> \brief provides damping for barostat via nph_uniaxial_damped dynamics
!> \param npt ...
!> \param gamma1 ...
!> \param dt ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE damp_veps(npt, gamma1, dt)

      TYPE(npt_info_type), INTENT(INOUT)                 :: npt
      REAL(KIND=dp), INTENT(IN)                          :: gamma1, dt

      REAL(KIND=dp)                                      :: scale

      scale = 1.0_dp
      scale = scale*EXP(-gamma1*0.25_dp*dt)
      ! Scale
      npt%v = npt%v*scale

   END SUBROUTINE damp_veps

! **************************************************************************************************
!> \brief update veps using multiplier obtained from SHAKE
!> \param old ...
!> \param gci ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param local_molecules ...
!> \param vel ...
!> \param dt ...
!> \param cell ...
!> \param npt ...
!> \param simpar ...
!> \param virial ...
!> \param vector_v ...
!> \param roll_tol ...
!> \param iroll ...
!> \param infree ...
!> \param first ...
!> \param para_env ...
!> \param u ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE rattle_roll_setup(old, gci, atomic_kind_set, particle_set, local_particles, &
                                molecule_kind_set, molecule_set, local_molecules, vel, dt, cell, npt, simpar, virial, &
                                vector_v, roll_tol, iroll, infree, first, para_env, u)

      TYPE(old_variables_type), POINTER                  :: old
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: dt
      TYPE(cell_type), POINTER                           :: cell
      TYPE(npt_info_type), DIMENSION(:, :), POINTER      :: npt
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      TYPE(virial_type), POINTER                         :: virial
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vector_v
      REAL(KIND=dp), INTENT(OUT)                         :: roll_tol
      INTEGER, INTENT(INOUT)                             :: iroll
      REAL(KIND=dp), INTENT(IN)                          :: infree
      LOGICAL, INTENT(INOUT)                             :: first
      TYPE(cp_para_env_type), INTENT(IN)                 :: para_env
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: u(:, :)

      REAL(KIND=dp)                                      :: kin
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_kin
      TYPE(npt_info_type), DIMENSION(3, 3)               :: npt_loc

      IF (first) THEN
         CALL update_pv(gci, simpar, atomic_kind_set, vel, particle_set, &
                        local_molecules, molecule_set, molecule_kind_set, &
                        local_particles, kin, pv_kin, virial, para_env%group)
         CALL update_veps(cell, npt, simpar, pv_kin, kin, virial, infree)

      ENDIF
      first = .FALSE.

      ! assigning local variable
      SELECT CASE (simpar%ensemble)
      CASE (npt_i_ensemble)
         npt_loc(:, :)%v = 0.0_dp
         npt_loc(:, :)%mass = 0.0_dp
         npt_loc(1, 1)%v = npt(1, 1)%v
         npt_loc(2, 2)%v = npt(1, 1)%v
         npt_loc(3, 3)%v = npt(1, 1)%v
         npt_loc(1, 1)%mass = npt(1, 1)%mass
         npt_loc(2, 2)%mass = npt(1, 1)%mass
         npt_loc(3, 3)%mass = npt(1, 1)%mass
      CASE (npt_f_ensemble)
         npt_loc = npt
      END SELECT

      ! resetting
      CALL set(old, atomic_kind_set, particle_set, vel, local_particles, cell, npt, 'B')
      CALL rattle_roll_control(gci, local_molecules, molecule_set, molecule_kind_set, &
                               particle_set, vel, dt, simpar, vector_v, npt_loc%v, roll_tol, iroll, &
                               para_env, u, cell, local_particles)

   END SUBROUTINE rattle_roll_setup

! **************************************************************************************************
!> \brief Overloaded routine to compute veps given the particles
!>      structure or a local copy of the velocity array
!> \param gci ...
!> \param simpar ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param local_particles ...
!> \param kin ...
!> \param pv_kin ...
!> \param virial ...
!> \param int_group ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE update_pv_particle_set(gci, simpar, atomic_kind_set, particle_set, &
                                     local_molecules, molecule_set, molecule_kind_set, &
                                     local_particles, kin, pv_kin, virial, int_group)
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(OUT)                         :: kin
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: pv_kin
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      INTEGER, INTENT(IN)                                :: int_group

      INTEGER                                            :: i, iparticle, iparticle_kind, &
                                                            iparticle_local, j, nparticle_local
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      pv_kin = 0.0_dp
      kin = 0.0_dp
      DO iparticle_kind = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            DO i = 1, 3
               DO j = 1, 3
                  pv_kin(i, j) = pv_kin(i, j)+ &
                                 mass*particle_set(iparticle)%v(i)* &
                                 particle_set(iparticle)%v(j)
               END DO
               kin = kin+mass*particle_set(iparticle)%v(i)* &
                     particle_set(iparticle)%v(i)
            END DO
         END DO
      END DO

      CALL mp_sum(pv_kin, int_group)
      CALL mp_sum(kin, int_group)

      ! updating the constraint virial
      IF (simpar%constraint) CALL pv_constraint(gci, local_molecules, &
                                                molecule_set, &
                                                molecule_kind_set, &
                                                particle_set, virial, &
                                                int_group)

   END SUBROUTINE update_pv_particle_set

! **************************************************************************************************
!> \brief Overloaded routine to compute kinetic virials given the particles
!>      structure or a local copy of the velocity array
!> \param gci ...
!> \param simpar ...
!> \param atomic_kind_set ...
!> \param vel ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param local_particles ...
!> \param kin ...
!> \param pv_kin ...
!> \param virial ...
!> \param int_group ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE update_pv_velocity(gci, simpar, atomic_kind_set, vel, particle_set, &
                                 local_molecules, molecule_set, molecule_kind_set, &
                                 local_particles, kin, pv_kin, virial, int_group)
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(OUT)                         :: kin
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pv_kin
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      INTEGER, INTENT(IN)                                :: int_group

      INTEGER                                            :: i, iparticle, iparticle_kind, &
                                                            iparticle_local, j, nparticle_local
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      pv_kin = 0.0_dp
      kin = 0._dp
      DO iparticle_kind = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            DO i = 1, 3
               DO j = 1, 3
                  pv_kin(i, j) = pv_kin(i, j)+ &
                                 mass*vel(i, iparticle)*vel(j, iparticle)
               END DO
               kin = kin+mass*vel(i, iparticle)*vel(i, iparticle)
            END DO
         END DO
      END DO

      CALL mp_sum(pv_kin, int_group)
      CALL mp_sum(kin, int_group)

      ! updating the constraint virial
      IF (simpar%constraint) CALL pv_constraint(gci, local_molecules, &
                                                molecule_set, &
                                                molecule_kind_set, &
                                                particle_set, virial, &
                                                int_group)

   END SUBROUTINE update_pv_velocity

! **************************************************************************************************
!> \brief Routine to compute veps
!> \param box ...
!> \param npt ...
!> \param simpar ...
!> \param pv_kin ...
!> \param kin ...
!> \param virial ...
!> \param infree ...
!> \param virial_components ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE update_veps(box, npt, simpar, pv_kin, kin, virial, infree, virial_components)

      TYPE(cell_type), INTENT(IN)                        :: box
      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: pv_kin
      REAL(KIND=dp), INTENT(IN)                          :: kin
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      REAL(KIND=dp), INTENT(IN)                          :: infree
      INTEGER, INTENT(IN), OPTIONAL                      :: virial_components

      INTEGER                                            :: ii
      REAL(KIND=dp)                                      :: fdotr, trace, v, v0, v0i, vi
      REAL(KIND=dp), DIMENSION(3, 3)                     :: unit

! initialize unit

      unit = 0.0_dp
      unit(1, 1) = 1.0_dp
      unit(2, 2) = 1.0_dp
      unit(3, 3) = 1.0_dp

      IF (simpar%ensemble == npt_i_ensemble .OR. &
          simpar%ensemble == npe_i_ensemble) THEN
         ! get force on barostat
         fdotr = 0.0_dp
         DO ii = 1, 3
            fdotr = fdotr+virial%pv_virial(ii, ii)+ &
                    virial%pv_constraint(ii, ii)
         END DO

         npt(:, :)%f = (1.0_dp+(3.0_dp*infree))*kin+fdotr- &
                       3.0_dp*simpar%p_ext*box%deth
      ELSEIF (simpar%ensemble == npt_f_ensemble .OR. &
              simpar%ensemble == npe_f_ensemble) THEN
         npt(:, :)%f = virial%pv_virial(:, :)+ &
                       pv_kin(:, :)+virial%pv_constraint(:, :)- &
                       unit(:, :)*simpar%p_ext*box%deth+ &
                       infree*kin*unit(:, :)
         IF (debug_isotropic_limit) THEN
            trace = npt(1, 1)%f+npt(2, 2)%f+npt(3, 3)%f
            trace = trace/3.0_dp
            npt(:, :)%f = trace*unit(:, :)
         END IF
      ELSEIF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
              simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         v = box%deth
         vi = 1._dp/v
         v0 = simpar%v0
         v0i = 1._dp/v0
         ! orthorhombic box ONLY
         ! Chooses only the compressive solution
         IF (v < v0) THEN
            npt(1, 1)%f = virial%pv_virial(1, 1)+ &
                          pv_kin(1, 1)+virial%pv_constraint(1, 1)- &
                          simpar%p0*v-simpar%v_shock*simpar%v_shock* &
                          v*v0i*(1._dp-v*v0i)+infree*kin
         ELSE
            npt(1, 1)%f = virial%pv_virial(1, 1)+ &
                          pv_kin(1, 1)+virial%pv_constraint(1, 1)- &
                          simpar%p0*v+infree*kin
         ENDIF
         IF (debug_uniaxial_limit) THEN
            ! orthorhombic box ONLY
            npt(1, 1)%f = virial%pv_virial(1, 1)+ &
                          pv_kin(1, 1)+virial%pv_constraint(1, 1)- &
                          simpar%p0*box%deth+infree*kin
         END IF
      ENDIF

      ! update barostat velocities
      npt(:, :)%v = npt(:, :)%v+ &
                    0.5_dp*simpar%dt*npt(:, :)%f/npt(:, :)%mass

      ! Screen the dynamics of the barostat according user request
      IF (PRESENT(virial_components)) THEN
         SELECT CASE (virial_components)
         CASE (do_clv_xyz)
            ! Do nothing..
         CASE (do_clv_x)
            npt(1, 2)%v = 0.0_dp
            npt(1, 3)%v = 0.0_dp
            npt(1, 2)%f = 0.0_dp
            npt(1, 3)%f = 0.0_dp
            npt(2, 1)%v = 0.0_dp
            npt(2, 2)%v = 0.0_dp
            npt(2, 3)%v = 0.0_dp
            npt(2, 1)%f = 0.0_dp
            npt(2, 2)%f = 0.0_dp
            npt(2, 3)%f = 0.0_dp
            npt(3, 1)%v = 0.0_dp
            npt(3, 2)%v = 0.0_dp
            npt(3, 3)%v = 0.0_dp
            npt(3, 1)%f = 0.0_dp
            npt(3, 2)%f = 0.0_dp
            npt(3, 3)%f = 0.0_dp
         CASE (do_clv_y)
            npt(1, 1)%v = 0.0_dp
            npt(1, 2)%v = 0.0_dp
            npt(1, 3)%v = 0.0_dp
            npt(1, 1)%f = 0.0_dp
            npt(1, 2)%f = 0.0_dp
            npt(1, 3)%f = 0.0_dp
            npt(2, 1)%v = 0.0_dp
            npt(2, 3)%v = 0.0_dp
            npt(2, 1)%f = 0.0_dp
            npt(2, 3)%f = 0.0_dp
            npt(3, 1)%v = 0.0_dp
            npt(3, 2)%v = 0.0_dp
            npt(3, 3)%v = 0.0_dp
            npt(3, 1)%f = 0.0_dp
            npt(3, 2)%f = 0.0_dp
            npt(3, 3)%f = 0.0_dp
         CASE (do_clv_z)
            npt(1, 1)%v = 0.0_dp
            npt(1, 2)%v = 0.0_dp
            npt(1, 3)%v = 0.0_dp
            npt(1, 1)%f = 0.0_dp
            npt(1, 2)%f = 0.0_dp
            npt(1, 3)%f = 0.0_dp
            npt(2, 1)%v = 0.0_dp
            npt(2, 2)%v = 0.0_dp
            npt(2, 3)%v = 0.0_dp
            npt(2, 1)%f = 0.0_dp
            npt(2, 2)%f = 0.0_dp
            npt(2, 3)%f = 0.0_dp
            npt(3, 1)%v = 0.0_dp
            npt(3, 2)%v = 0.0_dp
            npt(3, 1)%f = 0.0_dp
            npt(3, 2)%f = 0.0_dp
         CASE (do_clv_xy)
            npt(1, 3)%v = 0.0_dp
            npt(1, 3)%f = 0.0_dp
            npt(2, 3)%v = 0.0_dp
            npt(2, 3)%f = 0.0_dp
            npt(3, 1)%v = 0.0_dp
            npt(3, 2)%v = 0.0_dp
            npt(3, 3)%v = 0.0_dp
            npt(3, 1)%f = 0.0_dp
            npt(3, 2)%f = 0.0_dp
            npt(3, 3)%f = 0.0_dp
         CASE (do_clv_xz)
            npt(1, 2)%v = 0.0_dp
            npt(1, 2)%f = 0.0_dp
            npt(2, 1)%v = 0.0_dp
            npt(2, 2)%v = 0.0_dp
            npt(2, 3)%v = 0.0_dp
            npt(2, 1)%f = 0.0_dp
            npt(2, 2)%f = 0.0_dp
            npt(2, 3)%f = 0.0_dp
            npt(3, 2)%v = 0.0_dp
            npt(3, 2)%f = 0.0_dp
         CASE (do_clv_yz)
            npt(1, 1)%v = 0.0_dp
            npt(1, 2)%v = 0.0_dp
            npt(1, 3)%v = 0.0_dp
            npt(1, 1)%f = 0.0_dp
            npt(1, 2)%f = 0.0_dp
            npt(1, 3)%f = 0.0_dp
            npt(2, 1)%v = 0.0_dp
            npt(2, 1)%f = 0.0_dp
            npt(3, 1)%v = 0.0_dp
            npt(3, 1)%f = 0.0_dp
         END SELECT
      END IF

   END SUBROUTINE update_veps

! **************************************************************************************************
!> \brief First half of the velocity-verlet algorithm : update velocity by half
!>        step and positions by full step
!> \param tmp ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param nparticle_kind ...
!> \param shell_adiabatic ...
!> \param dt ...
!> \param u ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, &
                       shell_adiabatic, dt, u)

      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, core_particle_set, &
                                                            shell_particle_set
      INTEGER, INTENT(IN)                                :: nparticle_kind
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: u

      CHARACTER(len=*), PARAMETER :: routineN = 'vv_first', routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_local, &
                                                            shell_index
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: dm, dmc, dms, dsc(3), dvsc(3), &
                                                            fac_massc, fac_masss, mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (atomic_kind, shell)
      tmp%max_vel = 0.0_dp
      tmp%max_vel_sc = 0.0_dp
      tmp%max_dvel = 0.0_dp
      tmp%max_dvel_sc = 0.0_dp
      tmp%max_dr = 0.0_dp
      tmp%max_dsc = 0.0_dp
      dsc = 0.0_dp
      dvsc = 0.0_dp

      !     *** Velocity Verlet (first part) ***
      DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, shell_active=is_shell)
         IF (mass /= 0.0_dp) THEN
            dm = 0.5_dp*dt/mass
            IF (is_shell .AND. shell_adiabatic) THEN
               CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
               dms = 0.5_dp*dt/shell%mass_shell
               dmc = 0.5_dp*dt/shell%mass_core
               fac_masss = shell%mass_shell/mass
               fac_massc = shell%mass_core/mass
               nparticle_local = local_particles%n_el(iparticle_kind)
               IF (PRESENT(u)) THEN
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     shell_index = particle_set(iparticle)%shell_index
                     ! Transform positions and velocities and forces of the shell
                     CALL transform_first(shell_particle_set, tmp%shell_pos, tmp%shell_vel, &
                                          shell_index, u, dms, dt, tmp%poly_v, tmp%poly_r, tmp%scale_v, tmp%scale_r)

                     ! Transform positions and velocities and forces of the core
                     CALL transform_first(core_particle_set, tmp%core_pos, tmp%core_vel, &
                                          shell_index, u, dmc, dt, tmp%poly_v, tmp%poly_r, tmp%scale_v, tmp%scale_r)

                     ! Derive velocities and positions of the COM
                     tmp%vel(:, iparticle) = fac_masss*tmp%shell_vel(:, shell_index)+ &
                                             fac_massc*tmp%core_vel(:, shell_index)
                     tmp%pos(:, iparticle) = fac_masss*tmp%shell_pos(:, shell_index)+ &
                                             fac_massc*tmp%core_pos(:, shell_index)

                     tmp%max_vel = MAX(tmp%max_vel, ABS(tmp%vel(1, iparticle)), &
                                       ABS(tmp%vel(2, iparticle)), ABS(tmp%vel(3, iparticle)))
                     tmp%max_vel_sc = MAX(tmp%max_vel_sc, &
                                          ABS(tmp%shell_vel(1, shell_index)-tmp%core_vel(1, shell_index)), &
                                          ABS(tmp%shell_vel(2, shell_index)-tmp%core_vel(2, shell_index)), &
                                          ABS(tmp%shell_vel(3, shell_index)-tmp%core_vel(3, shell_index)))
                     tmp%max_dr = MAX(tmp%max_dr, &
                                      ABS(particle_set(iparticle)%r(1)-tmp%pos(1, iparticle)), &
                                      ABS(particle_set(iparticle)%r(2)-tmp%pos(2, iparticle)), &
                                      ABS(particle_set(iparticle)%r(3)-tmp%pos(3, iparticle)))
                     tmp%max_dvel = MAX(tmp%max_dvel, &
                                        ABS(particle_set(iparticle)%v(1)-tmp%vel(1, iparticle)), &
                                        ABS(particle_set(iparticle)%v(2)-tmp%vel(2, iparticle)), &
                                        ABS(particle_set(iparticle)%v(3)-tmp%vel(3, iparticle)))

                     dsc(:) = tmp%shell_pos(:, shell_index)-tmp%core_pos(:, shell_index)- &
                              shell_particle_set(shell_index)%r(:)+core_particle_set(shell_index)%r(:)
                     tmp%max_dsc = MAX(tmp%max_dsc, ABS(dsc(1)), ABS(dsc(2)), ABS(dsc(3)))
                     dvsc(:) = tmp%shell_vel(:, shell_index)-tmp%core_vel(:, shell_index)- &
                               shell_particle_set(shell_index)%v(:)+core_particle_set(shell_index)%v(:)
                     tmp%max_dvel_sc = MAX(tmp%max_dvel_sc, ABS(dvsc(1)), ABS(dvsc(2)), ABS(dvsc(3)))
                  END DO ! iparticle_local
               ELSE
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     shell_index = particle_set(iparticle)%shell_index
                     tmp%shell_vel(:, shell_index) = &
                        shell_particle_set(shell_index)%v(:)*tmp%scale_v(:)*tmp%scale_v(:)+ &
                        tmp%scale_v(:)*tmp%poly_v(:)*dms*shell_particle_set(shell_index)%f(:)
                     tmp%shell_pos(:, shell_index) = &
                        shell_particle_set(shell_index)%r(:)*tmp%scale_r(:)*tmp%scale_r(:)+ &
                        tmp%scale_r(:)*tmp%poly_r(:)*dt*tmp%shell_vel(:, shell_index)
                     tmp%core_vel(:, shell_index) = &
                        core_particle_set(shell_index)%v(:)*tmp%scale_v(:)*tmp%scale_v(:)+ &
                        tmp%scale_v(:)*tmp%poly_v(:)*dmc*core_particle_set(shell_index)%f(:)
                     tmp%core_pos(:, shell_index) = &
                        core_particle_set(shell_index)%r(:)*tmp%scale_r(:)*tmp%scale_r(:)+ &
                        tmp%scale_r(:)*tmp%poly_r(:)*dt*tmp%core_vel(:, shell_index)

                     tmp%vel(:, iparticle) = fac_masss*tmp%shell_vel(:, shell_index)+ &
                                             fac_massc*tmp%core_vel(:, shell_index)
                     tmp%pos(:, iparticle) = fac_masss*tmp%shell_pos(:, shell_index)+ &
                                             fac_massc*tmp%core_pos(:, shell_index)
                     tmp%max_vel = MAX(tmp%max_vel, &
                                       ABS(tmp%vel(1, iparticle)), ABS(tmp%vel(2, iparticle)), ABS(tmp%vel(3, iparticle)))
                     tmp%max_vel_sc = MAX(tmp%max_vel_sc, &
                                          ABS(tmp%shell_vel(1, shell_index)-tmp%core_vel(1, shell_index)), &
                                          ABS(tmp%shell_vel(2, shell_index)-tmp%core_vel(2, shell_index)), &
                                          ABS(tmp%shell_vel(3, shell_index)-tmp%core_vel(3, shell_index)))
                     tmp%max_dr = MAX(tmp%max_dr, &
                                      ABS(particle_set(iparticle)%r(1)-tmp%pos(1, iparticle)), &
                                      ABS(particle_set(iparticle)%r(2)-tmp%pos(2, iparticle)), &
                                      ABS(particle_set(iparticle)%r(3)-tmp%pos(3, iparticle)))
                     tmp%max_dvel = MAX(tmp%max_dvel, &
                                        ABS(particle_set(iparticle)%v(1)-tmp%vel(1, iparticle)), &
                                        ABS(particle_set(iparticle)%v(2)-tmp%vel(2, iparticle)), &
                                        ABS(particle_set(iparticle)%v(3)-tmp%vel(3, iparticle)))
                     dsc(:) = tmp%shell_pos(:, shell_index)-tmp%core_pos(:, shell_index)- &
                              shell_particle_set(shell_index)%r(:)+core_particle_set(shell_index)%r(:)
                     tmp%max_dsc = MAX(tmp%max_dsc, ABS(dsc(1)), ABS(dsc(2)), ABS(dsc(3)))
                     dvsc(:) = tmp%shell_vel(:, shell_index)-tmp%core_vel(:, shell_index)- &
                               shell_particle_set(shell_index)%v(:)+core_particle_set(shell_index)%v(:)
                     tmp%max_dvel_sc = MAX(tmp%max_dvel_sc, ABS(dvsc(1)), ABS(dvsc(2)), ABS(dvsc(3)))
                  END DO ! iparticle_local
               END IF
            ELSE
               nparticle_local = local_particles%n_el(iparticle_kind)
               IF (PRESENT(u)) THEN
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     ! Transform positions and velocities and forces
                     CALL transform_first(particle_set, tmp%pos, tmp%vel, &
                                          iparticle, u, dm, dt, tmp%poly_v, tmp%poly_r, tmp%scale_v, tmp%scale_r)

                     tmp%max_vel = MAX(tmp%max_vel, &
                                       ABS(tmp%vel(1, iparticle)), ABS(tmp%vel(2, iparticle)), ABS(tmp%vel(3, iparticle)))
                     tmp%max_dr = MAX(tmp%max_dr, ABS(particle_set(iparticle)%r(1)-tmp%pos(1, iparticle)), &
                                      ABS(particle_set(iparticle)%r(2)-tmp%pos(2, iparticle)), &
                                      ABS(particle_set(iparticle)%r(3)-tmp%pos(3, iparticle)))
                     tmp%max_dvel = MAX(tmp%max_dvel, &
                                        ABS(particle_set(iparticle)%v(1)-tmp%vel(1, iparticle)), &
                                        ABS(particle_set(iparticle)%v(2)-tmp%vel(2, iparticle)), &
                                        ABS(particle_set(iparticle)%v(3)-tmp%vel(3, iparticle)))
                  END DO ! iparticle_local
               ELSE
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     tmp%vel(1, iparticle) = &
                        particle_set(iparticle)%v(1)*tmp%scale_v(1)*tmp%scale_v(1)+ &
                        tmp%scale_v(1)*tmp%poly_v(1)*dm*particle_set(iparticle)%f(1)
                     tmp%vel(2, iparticle) = &
                        particle_set(iparticle)%v(2)*tmp%scale_v(2)*tmp%scale_v(2)+ &
                        tmp%scale_v(2)*tmp%poly_v(2)*dm*particle_set(iparticle)%f(2)
                     tmp%vel(3, iparticle) = &
                        particle_set(iparticle)%v(3)*tmp%scale_v(3)*tmp%scale_v(3)+ &
                        tmp%scale_v(3)*tmp%poly_v(3)*dm*particle_set(iparticle)%f(3)
                     tmp%pos(1, iparticle) = &
                        particle_set(iparticle)%r(1)*tmp%scale_r(1)*tmp%scale_r(1)+ &
                        tmp%scale_r(1)*tmp%poly_r(1)*dt*tmp%vel(1, iparticle)
                     tmp%pos(2, iparticle) = &
                        particle_set(iparticle)%r(2)*tmp%scale_r(2)*tmp%scale_r(2)+ &
                        tmp%scale_r(2)*tmp%poly_r(2)*dt*tmp%vel(2, iparticle)
                     tmp%pos(3, iparticle) = &
                        particle_set(iparticle)%r(3)*tmp%scale_r(3)*tmp%scale_r(3)+ &
                        tmp%scale_r(3)*tmp%poly_r(3)*dt*tmp%vel(3, iparticle)

                     tmp%max_vel = MAX(tmp%max_vel, &
                                       ABS(tmp%vel(1, iparticle)), ABS(tmp%vel(2, iparticle)), ABS(tmp%vel(3, iparticle)))
                     tmp%max_dr = MAX(tmp%max_dr, &
                                      ABS(particle_set(iparticle)%r(1)-tmp%pos(1, iparticle)), &
                                      ABS(particle_set(iparticle)%r(2)-tmp%pos(2, iparticle)), &
                                      ABS(particle_set(iparticle)%r(3)-tmp%pos(3, iparticle)))
                     tmp%max_dvel = MAX(tmp%max_dvel, &
                                        ABS(particle_set(iparticle)%v(1)-tmp%vel(1, iparticle)), &
                                        ABS(particle_set(iparticle)%v(2)-tmp%vel(2, iparticle)), &
                                        ABS(particle_set(iparticle)%v(3)-tmp%vel(3, iparticle)))
                  END DO
               END IF
            END IF
         END IF
      END DO
   END SUBROUTINE vv_first

! **************************************************************************************************
!> \brief Second half of the velocity-verlet algorithm : update velocity by half
!>        step  using the new forces
!> \param tmp ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param nparticle_kind ...
!> \param shell_adiabatic ...
!> \param dt ...
!> \param u ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
                        core_particle_set, shell_particle_set, nparticle_kind, &
                        shell_adiabatic, dt, u)

      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, core_particle_set, &
                                                            shell_particle_set
      INTEGER, INTENT(IN)                                :: nparticle_kind
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: u

      CHARACTER(len=*), PARAMETER :: routineN = 'vv_second', routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_local, &
                                                            shell_index
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: dm, dmc, dms, fac_massc, fac_masss, mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (atomic_kind, shell)

      !     *** Velocity Verlet (second part) ***
      DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, &
                              shell_active=is_shell)
         IF (mass /= 0.0_dp) THEN
            dm = 0.5_dp*dt/mass
            IF (is_shell .AND. shell_adiabatic) THEN
               CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
               dms = 0.5_dp*dt/shell%mass_shell
               dmc = 0.5_dp*dt/shell%mass_core
               fac_masss = shell%mass_shell/mass
               fac_massc = shell%mass_core/mass
               nparticle_local = local_particles%n_el(iparticle_kind)
               IF (PRESENT(u)) THEN
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     shell_index = particle_set(iparticle)%shell_index
                     ! Transform velocities and forces shell
                     CALL transform_second(shell_particle_set, tmp%shell_vel, shell_index, &
                                           u, dms, tmp%poly_v, tmp%scale_v)

                     ! Transform velocities and forces core
                     CALL transform_second(core_particle_set, tmp%core_vel, shell_index, &
                                           u, dmc, tmp%poly_v, tmp%scale_v)

                     ! Derive velocties of the COM
                     tmp%vel(1, iparticle) = fac_masss*tmp%shell_vel(1, shell_index)+ &
                                             fac_massc*tmp%core_vel(1, shell_index)
                     tmp%vel(2, iparticle) = fac_masss*tmp%shell_vel(2, shell_index)+ &
                                             fac_massc*tmp%core_vel(2, shell_index)
                     tmp%vel(3, iparticle) = fac_masss*tmp%shell_vel(3, shell_index)+ &
                                             fac_massc*tmp%core_vel(3, shell_index)
                  END DO ! iparticle_local
               ELSE
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     shell_index = particle_set(iparticle)%shell_index
                     tmp%shell_vel(1, shell_index) = &
                        tmp%shell_vel(1, shell_index)*tmp%scale_v(1)*tmp%scale_v(1)+ &
                        tmp%scale_v(1)*tmp%poly_v(1)*dms*shell_particle_set(shell_index)%f(1)
                     tmp%shell_vel(2, shell_index) = &
                        tmp%shell_vel(2, shell_index)*tmp%scale_v(2)*tmp%scale_v(2)+ &
                        tmp%scale_v(2)*tmp%poly_v(2)*dms*shell_particle_set(shell_index)%f(2)
                     tmp%shell_vel(3, shell_index) = &
                        tmp%shell_vel(3, shell_index)*tmp%scale_v(3)*tmp%scale_v(3)+ &
                        tmp%scale_v(3)*tmp%poly_v(3)*dms*shell_particle_set(shell_index)%f(3)
                     tmp%core_vel(1, shell_index) = &
                        tmp%core_vel(1, shell_index)*tmp%scale_v(1)*tmp%scale_v(1)+ &
                        tmp%scale_v(1)*tmp%poly_v(1)*dmc*core_particle_set(shell_index)%f(1)
                     tmp%core_vel(2, shell_index) = &
                        tmp%core_vel(2, shell_index)*tmp%scale_v(2)*tmp%scale_v(2)+ &
                        tmp%scale_v(2)*tmp%poly_v(2)*dmc*core_particle_set(shell_index)%f(2)
                     tmp%core_vel(3, shell_index) = &
                        tmp%core_vel(3, shell_index)*tmp%scale_v(3)*tmp%scale_v(3)+ &
                        tmp%scale_v(3)*tmp%poly_v(3)*dmc*core_particle_set(shell_index)%f(3)

                     tmp%vel(1, iparticle) = fac_masss*tmp%shell_vel(1, shell_index)+ &
                                             fac_massc*tmp%core_vel(1, shell_index)
                     tmp%vel(2, iparticle) = fac_masss*tmp%shell_vel(2, shell_index)+ &
                                             fac_massc*tmp%core_vel(2, shell_index)
                     tmp%vel(3, iparticle) = fac_masss*tmp%shell_vel(3, shell_index)+ &
                                             fac_massc*tmp%core_vel(3, shell_index)
                  END DO ! iparticle_local
               END IF
            ELSE
               nparticle_local = local_particles%n_el(iparticle_kind)
               IF (PRESENT(u)) THEN
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     CALL transform_second(particle_set, tmp%vel, iparticle, &
                                           u, dm, tmp%poly_v, tmp%scale_v)
                  END DO ! iparticle_local
               ELSE
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     tmp%vel(1, iparticle) = &
                        tmp%vel(1, iparticle)*tmp%scale_v(1)*tmp%scale_v(1)+ &
                        tmp%scale_v(1)*tmp%poly_v(1)*dm*particle_set(iparticle)%f(1)
                     tmp%vel(2, iparticle) = &
                        tmp%vel(2, iparticle)*tmp%scale_v(2)*tmp%scale_v(2)+ &
                        tmp%scale_v(2)*tmp%poly_v(2)*dm*particle_set(iparticle)%f(2)
                     tmp%vel(3, iparticle) = &
                        tmp%vel(3, iparticle)*tmp%scale_v(3)*tmp%scale_v(3)+ &
                        tmp%scale_v(3)*tmp%poly_v(3)*dm*particle_set(iparticle)%f(3)
                  END DO
               END IF
            END IF
         END IF
      END DO

   END SUBROUTINE vv_second

! **************************************************************************************************
!> \brief Transform position and velocities for the first half of the
!>        Velocity-Verlet integration in the npt_f ensemble
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param index ...
!> \param u ...
!> \param dm ...
!> \param dt ...
!> \param poly_v ...
!> \param poly_r ...
!> \param scale_v ...
!> \param scale_r ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE transform_first(particle_set, pos, vel, index, u, dm, dt, poly_v, &
                              poly_r, scale_v, scale_r)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pos, vel
      INTEGER, INTENT(IN)                                :: index
      REAL(KIND=dp), INTENT(IN)                          :: u(3, 3), dm, dt, poly_v(3), poly_r(3), &
                                                            scale_v(3), scale_r(3)

      CHARACTER(len=*), PARAMETER :: routineN = 'transform_first', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3)                        :: uf, ur, uv

      CALL matvec_3x3(ur, TRANSPOSE_3D(u), particle_set(index)%r(:))
      CALL matvec_3x3(uv, TRANSPOSE_3D(u), particle_set(index)%v(:))
      CALL matvec_3x3(uf, TRANSPOSE_3D(u), particle_set(index)%f(:))
      !
      uv(1) = uv(1)*scale_v(1)*scale_v(1)+uf(1)*scale_v(1)*poly_v(1)*dm
      uv(2) = uv(2)*scale_v(2)*scale_v(2)+uf(2)*scale_v(2)*poly_v(2)*dm
      uv(3) = uv(3)*scale_v(3)*scale_v(3)+uf(3)*scale_v(3)*poly_v(3)*dm

      ur(1) = ur(1)*scale_r(1)*scale_r(1)+uv(1)*scale_r(1)*poly_r(1)*dt
      ur(2) = ur(2)*scale_r(2)*scale_r(2)+uv(2)*scale_r(2)*poly_r(2)*dt
      ur(3) = ur(3)*scale_r(3)*scale_r(3)+uv(3)*scale_r(3)*poly_r(3)*dt
      !
      CALL MATVEC_3x3(pos(:, index), u, ur)
      CALL MATVEC_3x3(vel(:, index), u, uv)

   END SUBROUTINE transform_first

! **************************************************************************************************
!> \brief Transform position and velocities for the second half of the
!>        Velocity-Verlet integration in the npt_f ensemble
!> \param particle_set ...
!> \param vel ...
!> \param index ...
!> \param u ...
!> \param dm ...
!> \param poly_v ...
!> \param scale_v ...
!> \par History
!>      none
!> \author MI (February 2008)
! **************************************************************************************************
   SUBROUTINE transform_second(particle_set, vel, index, u, dm, poly_v, scale_v)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vel
      INTEGER, INTENT(IN)                                :: index
      REAL(KIND=dp), INTENT(IN)                          :: u(3, 3), dm, poly_v(3), scale_v(3)

      CHARACTER(len=*), PARAMETER :: routineN = 'transform_second', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3)                        :: uf, uv

      CALL matvec_3x3(uv, TRANSPOSE_3D(u), vel(:, index))
      CALL matvec_3x3(uf, TRANSPOSE_3D(u), particle_set(index)%f(:))
      !
      uv(1) = uv(1)*scale_v(1)*scale_v(1)+uf(1)*scale_v(1)*poly_v(1)*dm
      uv(2) = uv(2)*scale_v(2)*scale_v(2)+uf(2)*scale_v(2)*poly_v(2)*dm
      uv(3) = uv(3)*scale_v(3)*scale_v(3)+uf(3)*scale_v(3)*poly_v(3)*dm

      CALL MATVEC_3x3(vel(:, index), u, uv)

   END SUBROUTINE transform_second

! **************************************************************************************************
!> \brief  Compute the timestep rescaling factor
!> \param md_env ...
!> \param tmp ...
!> \param dt ...
!> \param simpar ...
!> \param para_env ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param nparticle_kind ...
!> \param shell_adiabatic ...
!> \param npt ...
!> \par History
!>      none
!> \author MI (October 2008)
! **************************************************************************************************
   SUBROUTINE variable_timestep(md_env, tmp, dt, simpar, para_env, atomic_kind_set, local_particles, &
                                particle_set, core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, npt)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(tmp_variables_type), POINTER                  :: tmp
      REAL(KIND=dp), INTENT(INOUT)                       :: dt
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, core_particle_set, &
                                                            shell_particle_set
      INTEGER, INTENT(IN)                                :: nparticle_kind
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      TYPE(npt_info_type), OPTIONAL, POINTER             :: npt(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'variable_timestep', &
         routineP = moduleN//':'//routineN

      INTEGER, SAVE                                      :: itime = 0
      REAL(KIND=dp)                                      :: dt_fact, dt_fact_f, dt_fact_old, &
                                                            dt_fact_sc, dt_fact_v, dt_old
      REAL(KIND=dp), POINTER                             :: time
      TYPE(thermostats_type), POINTER                    :: thermostats

      dt_fact = 1.0_dp
      dt_fact_sc = 1.0_dp
      dt_fact_f = 1.0_dp
      dt_fact_v = 1.0_dp
      dt_old = dt
      dt_fact_old = simpar%dt_fact
      simpar%dt_fact = 1.0_dp
      NULLIFY (thermostats)

      itime = itime+1
      CALL mp_max(tmp%max_dr, para_env%group)
      IF (tmp%max_dr > simpar%dr_tol) THEN
         CALL mp_max(tmp%max_dvel, para_env%group)
         dt_fact = SQRT(simpar%dr_tol*dt/tmp%max_dvel)/dt
      END IF

      IF (shell_adiabatic) THEN
         CALL mp_max(tmp%max_dsc, para_env%group)
         IF (tmp%max_dsc > simpar%dsc_tol) THEN
            CALL mp_max(tmp%max_dvel_sc, para_env%group)
            dt_fact_sc = SQRT(simpar%dsc_tol*dt/tmp%max_dvel_sc)/dt
         END IF
      END IF

      dt_fact_f = MIN(dt_fact, dt_fact_sc, 1.0_dp)
      IF (dt_fact_f < 1.0_dp) THEN
         IF (dt_fact_f < 0.1_dp) dt_fact_f = 0.1_dp
         ! repeat the first vv half-step with the rescaled time step
         dt = dt*dt_fact_f
         simpar%dt_fact = dt_fact_f
         CALL rescaled_vv_first(tmp, dt, simpar, atomic_kind_set, local_particles, &
                                particle_set, core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, npt)
      END IF

      dt_fact = 1.0_dp
      dt_fact_sc = 1.0_dp
      CALL mp_max(tmp%max_dr, para_env%group)
      IF (tmp%max_dr > simpar%dr_tol) THEN
         CALL mp_max(tmp%max_vel, para_env%group)
         dt_fact = simpar%dr_tol/tmp%max_vel/dt
      END IF

      IF (shell_adiabatic) THEN
         CALL mp_max(tmp%max_dsc, para_env%group)
         IF (tmp%max_dsc > simpar%dsc_tol) THEN
            CALL mp_max(tmp%max_vel_sc, para_env%group)
            dt_fact_sc = simpar%dsc_tol/tmp%max_vel_sc/dt
         END IF
      END IF
      dt_fact_v = MIN(dt_fact, dt_fact_sc, 1.0_dp)

      IF (dt_fact_v < 1.0_dp) THEN
         IF (dt_fact_v < 0.1_dp) dt_fact_v = 0.1_dp
         ! repeat the first vv half-step with the rescaled time step
         dt = dt*dt_fact_v
         simpar%dt_fact = dt_fact_f*dt_fact_v
         CALL rescaled_vv_first(tmp, dt, simpar, atomic_kind_set, local_particles, &
                                particle_set, core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, npt)
      END IF

      simpar%dt_fact = dt_fact_f*dt_fact_v
      IF (simpar%dt_fact < 1.0_dp) THEN
         CALL get_md_env(md_env, t=time, thermostats=thermostats)
         time = time-dt_old+dt_old*simpar%dt_fact
         IF (ASSOCIATED(thermostats)) THEN
            CALL set_thermostats(thermostats, dt_fact=simpar%dt_fact)
         END IF
      END IF

   END SUBROUTINE variable_timestep

! **************************************************************************************************
!> \brief Repeat the first step of velocity verlet witht the rescaled time step
!> \param tmp ...
!> \param dt ...
!> \param simpar ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param nparticle_kind ...
!> \param shell_adiabatic ...
!> \param npt ...
!> \par History
!>      none
!> \author MI (October 2008)
!>     I soliti ignoti
! **************************************************************************************************
   SUBROUTINE rescaled_vv_first(tmp, dt, simpar, atomic_kind_set, local_particles, &
                                particle_set, core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, npt)

      TYPE(tmp_variables_type), POINTER                  :: tmp
      REAL(KIND=dp), INTENT(IN)                          :: dt
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, core_particle_set, &
                                                            shell_particle_set
      INTEGER, INTENT(IN)                                :: nparticle_kind
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      TYPE(npt_info_type), OPTIONAL, POINTER             :: npt(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'rescaled_vv_first', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: e2 = 1.0_dp/6.0_dp, e4 = e2/20.0_dp, &
                                                            e6 = e4/42.0_dp, e8 = e6/72.0_dp

      REAL(KIND=dp)                                      :: arg_r(3), arg_v(3), e_val(3), infree, &
                                                            trvg, u(3, 3)

      infree = 1.0_dp/REAL(simpar%nfree, dp)
      arg_r = tmp%arg_r
      arg_v = tmp%arg_v
      u = tmp%u
      e_val = tmp%e_val

      SELECT CASE (simpar%ensemble)
      CASE (nve_ensemble, nvt_ensemble)
         CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)
      CASE (isokin_ensemble)
         tmp%scale_v(1:3) = SQRT(1.0_dp/tmp%ds)
         tmp%poly_v(1:3) = 2.0_dp*tmp%s/SQRT(tmp%ds)/dt
         CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, &
                       shell_adiabatic, dt)

      CASE (npt_i_ensemble, npe_i_ensemble)
         arg_r = arg_r*simpar%dt_fact*simpar%dt_fact
         tmp%poly_r(1:3) = 1.0_dp+e2*arg_r(1)+e4*arg_r(1)*arg_r(1)+e6*arg_r(1)**3+e8*arg_r(1)**4
         arg_v = arg_v*simpar%dt_fact*simpar%dt_fact
         tmp%poly_v(1:3) = 1.0_dp+e2*arg_v(1)+e4*arg_v(1)*arg_v(1)+e6*arg_v(1)**3+e8*arg_v(1)**4
         tmp%scale_r(1:3) = EXP(0.5_dp*dt*npt(1, 1)%v)
         tmp%scale_v(1:3) = EXP(-0.25_dp*dt*npt(1, 1)%v* &
                                (1.0_dp+3.0_dp*infree))
         CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, &
                       shell_adiabatic, dt)

      CASE (npt_f_ensemble, npe_f_ensemble)
         trvg = npt(1, 1)%v+npt(2, 2)%v+npt(3, 3)%v
         arg_r(:) = arg_r(:)*simpar%dt_fact*simpar%dt_fact
         tmp%poly_r = 1._dp+e2*arg_r+e4*arg_r*arg_r+e6*arg_r**3+e8*arg_r**4
         tmp%scale_r(:) = EXP(0.5_dp*dt*e_val(:))
         arg_v(:) = arg_v(:)*simpar%dt_fact*simpar%dt_fact
         tmp%poly_v = 1.0_dp+e2*arg_v+e4*arg_v*arg_v+e6*arg_v**3+e8*arg_v**4
         tmp%scale_v(:) = EXP(-0.25_dp*dt*( &
                              e_val(:)+trvg*infree))

         CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, &
                       shell_adiabatic, dt, u)

      CASE (nph_uniaxial_ensemble, nph_uniaxial_damped_ensemble)
         arg_r = arg_r*simpar%dt_fact*simpar%dt_fact
         tmp%poly_r(1) = 1._dp+e2*arg_r(1)+e4*arg_r(1)*arg_r(1)+e6*arg_r(1)**3+e8*arg_r(1)**4
         arg_v(2) = arg_v(2)*simpar%dt_fact*simpar%dt_fact
         arg_v(1) = arg_v(1)*simpar%dt_fact*simpar%dt_fact
         tmp%poly_v(1) = 1._dp+e2*arg_v(1)+e4*arg_v(1)*arg_v(1)+e6*arg_v(1)**3+e8*arg_v(1)**4
         tmp%poly_v(2) = 1._dp+e2*arg_v(2)+e4*arg_v(2)*arg_v(2)+e6*arg_v(2)**3+e8*arg_v(2)**4
         tmp%poly_v(3) = 1._dp+e2*arg_v(2)+e4*arg_v(2)*arg_v(2)+e6*arg_v(2)**3+e8*arg_v(2)**4
         tmp%scale_r(1) = EXP(0.5_dp*dt*npt(1, 1)%v)
         tmp%scale_v(1) = EXP(-0.25_dp*dt*npt(1, 1)%v* &
                              (1._dp+infree))
         tmp%scale_v(2) = EXP(-0.25_dp*dt*npt(1, 1)%v*infree)
         tmp%scale_v(3) = EXP(-0.25_dp*dt*npt(1, 1)%v*infree)
         CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
                       core_particle_set, shell_particle_set, nparticle_kind, &
                       shell_adiabatic, dt)

      END SELECT

   END SUBROUTINE rescaled_vv_first

END MODULE integrator_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief
!> \author
! **************************************************************************************************
MODULE shell_opt

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cg_optimizer,                    ONLY: geoopt_cg
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_types,                    ONLY: gopt_f_create,&
                                              gopt_f_release,&
                                              gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_read,&
                                              gopt_param_release,&
                                              gopt_param_type
   USE input_constants,                 ONLY: default_shellcore_method_id
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE integrator_utils,                ONLY: tmp_variables_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE shell_potential_types,           ONLY: shell_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: optimize_shell_core
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'shell_opt'

CONTAINS

! **************************************************************************************************
!> \brief Optimize shell-core positions along an MD run
!> \param force_env ...
!> \param particle_set ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param globenv ...
!> \param tmp ...
!> \param check ...
!> \author
! **************************************************************************************************

   SUBROUTINE optimize_shell_core(force_env, particle_set, shell_particle_set, core_particle_set, globenv, tmp, check)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, shell_particle_set, &
                                                            core_particle_set
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(tmp_variables_type), OPTIONAL, POINTER        :: tmp
      LOGICAL, INTENT(IN), OPTIONAL                      :: check

      CHARACTER(len=*), PARAMETER :: routineN = 'optimize_shell_core', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iat, nshell
      LOGICAL                                            :: do_update, explicit, my_check, optimize
      REAL(dp), DIMENSION(:), POINTER                    :: dvec_sc, dvec_sc_0
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(section_vals_type), POINTER                   :: force_env_section, geo_section, &
                                                            root_section

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/shell_opt.F",90)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/shell_opt.F",91)

      NULLIFY (gopt_param, force_env_section, gopt_env, dvec_sc, dvec_sc_0, root_section, geo_section)
      root_section => force_env%root_section
      force_env_section => force_env%force_env_section
      geo_section => section_vals_get_subs_vals(root_section, "MOTION%SHELL_OPT")

      CALL section_vals_get(geo_section, explicit=explicit)
      IF (.NOT. explicit) RETURN

      CALL timeset(routineN, handle)

      optimize = .FALSE.
      my_check = .FALSE.
      IF (PRESENT(check)) my_check = check
      IF (my_check) THEN
         NULLIFY (subsys, para_env, atomic_kinds, local_particles)
         CALL force_env_get(force_env=force_env, subsys=subsys, para_env=para_env)
         CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds, local_particles=local_particles)
         CALL check_shell_core_distance(atomic_kinds, local_particles, particle_set, shell_particle_set, &
                                        core_particle_set, para_env, optimize)

         IF (.NOT. optimize) THEN
            CALL timestop(handle)
            RETURN
         END IF
      END IF

      nshell = SIZE(shell_particle_set)
      ALLOCATE (dvec_sc(3*nshell))
      ALLOCATE (dvec_sc_0(3*nshell))
      DO i = 1, nshell
         dvec_sc(1+3*(i-1)) = core_particle_set(i)%r(1)-shell_particle_set(i)%r(1)
         dvec_sc(2+3*(i-1)) = core_particle_set(i)%r(2)-shell_particle_set(i)%r(2)
         dvec_sc(3+3*(i-1)) = core_particle_set(i)%r(3)-shell_particle_set(i)%r(3)
      END DO
      dvec_sc_0 = dvec_sc

      CALL gopt_param_read(gopt_param, geo_section, type_id=default_shellcore_method_id)
      CALL gopt_f_create(gopt_env, gopt_param, force_env=force_env, globenv=globenv, &
                         geo_opt_section=geo_section)

      CALL cp_add_iter_level(logger%iter_info, "SHELL_OPT")
      gopt_env%eval_opt_geo = .FALSE.
      CALL geoopt_cg(force_env, gopt_param, globenv, &
                     geo_section, gopt_env, dvec_sc, do_update=do_update)
      IF (.NOT. do_update) THEN
         DO i = 1, nshell
            shell_particle_set(i)%r(1) = -dvec_sc_0(1+3*(i-1))+core_particle_set(i)%r(1)
            shell_particle_set(i)%r(2) = -dvec_sc_0(2+3*(i-1))+core_particle_set(i)%r(2)
            shell_particle_set(i)%r(3) = -dvec_sc_0(3+3*(i-1))+core_particle_set(i)%r(3)
         END DO
      END IF
      CALL cp_rm_iter_level(logger%iter_info, "SHELL_OPT")

      CALL gopt_f_release(gopt_env)
      CALL gopt_param_release(gopt_param)
      DEALLOCATE (dvec_sc)
      DEALLOCATE (dvec_sc_0)

      IF (PRESENT(tmp)) THEN
         DO i = 1, nshell
            iat = shell_particle_set(i)%atom_index
            tmp%shell_vel(1:3, i) = tmp%vel(1:3, iat)
            tmp%core_vel(1:3, i) = tmp%vel(1:3, iat)
         END DO
      ELSE
         DO i = 1, nshell
            iat = shell_particle_set(i)%atom_index
            shell_particle_set(i)%v(1:3) = particle_set(iat)%v(1:3)
            core_particle_set(i)%v(1:3) = particle_set(iat)%v(1:3)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE optimize_shell_core

! **************************************************************************************************
!> \brief Check shell_core_distance
!> \param atomic_kinds ...
!> \param local_particles ...
!> \param particle_set ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param para_env ...
!> \param optimize ...
!> \par History
!>      none
!> \author MI (October 2008)
!>     I soliti ignoti
! **************************************************************************************************
   SUBROUTINE check_shell_core_distance(atomic_kinds, local_particles, particle_set, &
                                        shell_particle_set, core_particle_set, para_env, optimize)

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, shell_particle_set, &
                                                            core_particle_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(INOUT)                             :: optimize

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_shell_core_distance', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, iparticle, iparticle_local, &
                                                            itest, nkind, nparticle_local, &
                                                            shell_index
      LOGICAL                                            :: is_shell
      REAL(dp)                                           :: dsc, rc(3), rs(3)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      nkind = atomic_kinds%n_els
      itest = 0
      DO ikind = 1, nkind
         NULLIFY (atomic_kind)
         atomic_kind => atomic_kinds%els(ikind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, shell_active=is_shell, shell=shell)
         IF (is_shell) THEN
            IF (shell%max_dist > 0.0_dp) THEN
               nparticle_local = local_particles%n_el(ikind)
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(ikind)%array(iparticle_local)
                  shell_index = particle_set(iparticle)%shell_index

                  rc(:) = core_particle_set(shell_index)%r(:)
                  rs(:) = shell_particle_set(shell_index)%r(:)
                  dsc = SQRT((rc(1)-rs(1))**2+(rc(2)-rs(2))**2+(rc(3)-rs(3))**2)
                  IF (dsc > shell%max_dist) THEN
                     itest = 1
                  END IF
               END DO
            END IF
         END IF
      END DO

      CALL mp_sum(itest, para_env%group)
      IF (itest > 0) optimize = .TRUE.

   END SUBROUTINE check_shell_core_distance
END MODULE shell_opt
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Collection of utilities for setting-up and handle velocities in MD
!>         runs
!> \author CJM
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
!>         reorganization of the original routines/modules
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
! **************************************************************************************************
MODULE md_vel_utils
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: &
        cell_type, use_perd_none, use_perd_x, use_perd_xy, use_perd_xyz, use_perd_xz, use_perd_y, &
        use_perd_yz, use_perd_z
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE force_env_utils,                 ONLY: force_env_rattle,&
                                              force_env_shake
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble,&
                                              reftraj_ensemble
   USE input_cp2k_binary_restarts,      ONLY: read_binary_velocities
   USE input_restart_force_eval,        ONLY: update_subsys
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_list_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: diamat_all
   USE md_ener_types,                   ONLY: md_ener_type
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              get_molecule_kind,&
                                              get_molecule_kind_set,&
                                              molecule_kind_type
   USE parallel_rng_types,              ONLY: next_random_number
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: kelvin
   USE shell_opt,                       ONLY: optimize_shell_core
   USE shell_potential_types,           ONLY: shell_kind_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermal_region_types,            ONLY: thermal_region_type,&
                                              thermal_regions_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_vel_utils'


  PUBLIC :: temperature_control,&
            comvel_control,&
            angvel_control,&
            setup_velocities

CONTAINS

! **************************************************************************************************
!> \brief compute center of mass position
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param is_fixed ...
!> \param rcom ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE compute_rcom(part,is_fixed,rcom)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, DIMENSION(:), INTENT(IN)                  :: is_fixed
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: rcom

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_rcom', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: denom, mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    rcom(:)  = 0.0_dp
    denom = 0.0_dp
    DO i = 1, SIZE(part)
       atomic_kind => part(i)%atomic_kind
       CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
       SELECT CASE(is_fixed(i))
       CASE(use_perd_x,use_perd_y,use_perd_z,use_perd_xy,use_perd_xz,use_perd_yz,use_perd_none)
          rcom(1) = rcom(1) + part(i)%r(1) * mass
          rcom(2) = rcom(2) + part(i)%r(2) * mass
          rcom(3) = rcom(3) + part(i)%r(3) * mass
          denom = denom + mass
       END SELECT
    END DO
    rcom = rcom/denom

  END SUBROUTINE compute_rcom

! **************************************************************************************************
!> \brief compute center of mass velocity
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param is_fixed ...
!> \param vcom ...
!> \param ecom ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE compute_vcom(part,is_fixed,vcom,ecom)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, DIMENSION(:), INTENT(IN)                  :: is_fixed
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: vcom
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: ecom

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_vcom', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: denom, mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    vcom  = 0.0_dp
    denom = 0.0_dp
    DO i = 1, SIZE(part)
       atomic_kind => part(i)%atomic_kind
       CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
       IF (mass.NE.0.0) THEN
          SELECT CASE(is_fixed(i))
          CASE(use_perd_x,use_perd_y,use_perd_z,use_perd_xy,use_perd_xz,use_perd_yz,use_perd_none)
             vcom(1) = vcom(1) + part(i)%v(1) * mass
             vcom(2) = vcom(2) + part(i)%v(2) * mass
             vcom(3) = vcom(3) + part(i)%v(3) * mass
             denom = denom + mass
          END SELECT
       END IF
    END DO
    vcom = vcom/denom
    IF (PRESENT(ecom)) THEN
       ecom = 0.5_dp*denom*SUM(vcom*vcom)
    END IF

  END SUBROUTINE compute_vcom

! **************************************************************************************************
!> \brief Copy atom velocities into core and shell velocities
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param shell_part ...
!> \param core_part ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE clone_core_shell_vel(part,shell_part,core_part)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part, shell_part, core_part

      CHARACTER(len=*), PARAMETER :: routineN = 'clone_core_shell_vel', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: is_shell
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    DO i = 1, SIZE(part)
       atomic_kind => part(i)%atomic_kind
       CALL get_atomic_kind(atomic_kind=atomic_kind, shell_active=is_shell)
       IF (is_shell) THEN
          shell_part( part(i)%shell_index )%v(:) =  part(i)%v(:)
          core_part( part(i)%shell_index )%v(:)  =  part(i)%v(:)
       END IF
    END DO

  END SUBROUTINE clone_core_shell_vel

! **************************************************************************************************
!> \brief Compute the kinetic energy. Does not subtract the center of mass kinetic
!>      energy.
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param ireg ...
!> \retval ekin ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  FUNCTION compute_ekin(part,ireg) RESULT(ekin)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, INTENT(IN), OPTIONAL                      :: ireg
      REAL(KIND=dp)                                      :: ekin

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_ekin', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    NULLIFY(atomic_kind)
    ekin = 0.0_dp
    IF(PRESENT(ireg)) THEN
      DO i = 1, SIZE(part)
       IF(part(i)%t_region_index==ireg) THEN
         atomic_kind => part(i)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         ekin = ekin + 0.5_dp * mass * SUM(part(i)%v(:) * part(i)%v(:))
       END IF
      END DO
    ELSE
      DO i = 1, SIZE(part)
         atomic_kind => part(i)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         ekin = ekin + 0.5_dp * mass * SUM(part(i)%v(:) * part(i)%v(:))
      END DO
    END IF

  END FUNCTION compute_ekin

! **************************************************************************************************
!> \brief Rescale the velocity to mimic the given external kinetic temperature.
!>      Optionally also rescale vcom.
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param simpar ...
!> \param ekin ...
!> \param vcom ...
!> \param ireg ...
!> \param nfree ...
!> \param temp ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE rescale_vel(part,simpar,ekin,vcom,ireg,nfree,temp)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(simpar_type), POINTER                         :: simpar
      REAL(KIND=dp), INTENT(INOUT)                       :: ekin
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT), &
         OPTIONAL                                        :: vcom
      INTEGER, INTENT(IN), OPTIONAL                      :: ireg, nfree
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: temp

      CHARACTER(len=*), PARAMETER :: routineN = 'rescale_vel', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_ireg, my_nfree
      REAL(KIND=dp)                                      :: factor, my_temp

    IF(PRESENT(ireg).AND.PRESENT(nfree).AND.PRESENT(temp)) THEN
      my_ireg = ireg
      my_nfree = nfree
      my_temp = temp
    ELSEIF(PRESENT(nfree)) THEN
      my_ireg = 0
      my_nfree = nfree
      my_temp = simpar%temp_ext
    ELSE
      my_ireg = 0
      my_nfree = simpar%nfree
      my_temp = simpar%temp_ext
    END IF
    IF (my_nfree/=0) THEN
       factor = my_temp / ( 2.0_dp * ekin ) * REAL(my_nfree,KIND=dp)
    ELSE
       factor = 0.0_dp
    ENDIF
    ! Note:
    ! this rescaling is still wrong, it should take the masses into account
    ! rescaling is generally not correct, so needs fixing
    ekin = ekin * factor
    factor = SQRT(factor)
    IF(PRESENT(ireg)) THEN
      DO i = 1, SIZE(part)
        IF(part(i)%t_region_index == my_ireg) part(i)%v(:) = factor*part(i)%v(:)
      END DO
    ELSE
      DO i = 1, SIZE(part)
        part(i)%v(:) = factor*part(i)%v(:)
      END DO
      IF (PRESENT(vcom)) THEN
         vcom = factor*vcom
      END IF
    END IF

  END SUBROUTINE rescale_vel

! **************************************************************************************************
!> \brief Rescale the velocity of separated regions independently
!> \param part ...
!> \param md_env ...
!> \param simpar ...
!> \par History
!>      2008-11
!> \author  MI
! **************************************************************************************************
  SUBROUTINE rescale_vel_region(part,md_env,simpar)

      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(simpar_type), POINTER                         :: simpar

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rescale_vel_region', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ireg, nfree, nfree0, nfree_done
      REAL(KIND=dp)                                      :: ekin, temp
      TYPE(thermal_region_type), POINTER                 :: t_region
      TYPE(thermal_regions_type), POINTER                :: thermal_regions

    NULLIFY(thermal_regions, t_region)

    CALL get_md_env(md_env,thermal_regions=thermal_regions)
    nfree_done = 0
    DO ireg = 1,thermal_regions%nregions
      NULLIFY(t_region)
      t_region => thermal_regions%thermal_region(ireg)
      nfree = t_region%npart*3
      ekin = compute_ekin(part,ireg)
      temp =  t_region%temp_expected
      CALL rescale_vel(part,simpar,ekin,ireg=ireg,nfree=nfree,temp=temp)
      nfree_done = nfree_done+nfree
      ekin = compute_ekin(part,ireg)
      temp = 2.0_dp* ekin/REAL(nfree,dp)*kelvin
      t_region%temperature = temp
    END DO
    nfree0 = simpar%nfree - nfree_done
    IF(nfree0>0) THEN
      ekin = compute_ekin(part,0)
      CALL rescale_vel(part,simpar,ekin,ireg=0,nfree=nfree0,temp=simpar%temp_ext)
      ekin = compute_ekin(part,0)
      temp = 2.0_dp* ekin/REAL(nfree0,dp)*kelvin
      thermal_regions%temp_reg0 = temp
    END IF
  END SUBROUTINE  rescale_vel_region

! **************************************************************************************************
!> \brief subtract center of mass velocity
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param is_fixed ...
!> \param vcom ...
!> \par History
!>      2007-11-6: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE subtract_vcom(part,is_fixed,vcom)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, DIMENSION(:), INTENT(IN)                  :: is_fixed
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: vcom

      CHARACTER(len=*), PARAMETER :: routineN = 'subtract_vcom', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

    DO i = 1, SIZE(part)
       SELECT CASE(is_fixed(i))
       CASE(use_perd_x)
          part(i)%v(2) = part(i)%v(2) - vcom(2)
          part(i)%v(3) = part(i)%v(3) - vcom(3)
       CASE(use_perd_y)
          part(i)%v(1) = part(i)%v(1) - vcom(1)
          part(i)%v(3) = part(i)%v(3) - vcom(3)
       CASE(use_perd_z)
          part(i)%v(1) = part(i)%v(1) - vcom(1)
          part(i)%v(2) = part(i)%v(2) - vcom(2)
       CASE(use_perd_xy)
          part(i)%v(3) = part(i)%v(3) - vcom(3)
       CASE(use_perd_xz)
          part(i)%v(2) = part(i)%v(2) - vcom(2)
       CASE(use_perd_yz)
          part(i)%v(1) = part(i)%v(1) - vcom(1)
       CASE(use_perd_none)
          part(i)%v(:) = part(i)%v(:) - vcom(:)
       END SELECT
    END DO
  END SUBROUTINE subtract_vcom

! **************************************************************************************************
!> \brief compute the angular velocity
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param is_fixed ...
!> \param rcom ...
!> \param vang ...
!> \par History
!>      2007-11-9: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE compute_vang(part,is_fixed,rcom,vang)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, DIMENSION(:), INTENT(IN)                  :: is_fixed
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rcom
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: vang

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_vang', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: mass, proj
      REAL(KIND=dp), DIMENSION(3)                        :: evals, mang, r
      REAL(KIND=dp), DIMENSION(3, 3)                     :: iner
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    NULLIFY(atomic_kind)
    mang(:)   = 0.0_dp
    iner(:,:) = 0.0_dp
    DO i=1,SIZE(part)
       ! compute angular momentum and inertia tensor
       SELECT CASE(is_fixed(i))
       CASE(use_perd_x,use_perd_y,use_perd_z,use_perd_xy,use_perd_xz,use_perd_yz,use_perd_none)
          r(:) = part(i)%r(:) - rcom(:)
          atomic_kind => part(i)%atomic_kind
          CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
          mang(1) = mang(1) + mass*(r(2)*part(i)%v(3) - r(3)*part(i)%v(2))
          mang(2) = mang(2) + mass*(r(3)*part(i)%v(1) - r(1)*part(i)%v(3))
          mang(3) = mang(3) + mass*(r(1)*part(i)%v(2) - r(2)*part(i)%v(1))

          iner(1,1) = iner(1,1) + mass*(r(2)*r(2) + r(3)*r(3))
          iner(2,2) = iner(2,2) + mass*(r(3)*r(3) + r(1)*r(1))
          iner(3,3) = iner(3,3) + mass*(r(1)*r(1) + r(2)*r(2))

          iner(1,2) = iner(1,2) - mass*r(1)*r(2)
          iner(2,3) = iner(2,3) - mass*r(2)*r(3)
          iner(3,1) = iner(3,1) - mass*r(3)*r(1)
       END SELECT
    END DO
    iner(2,1) = iner(1,2)
    iner(3,2) = iner(2,3)
    iner(1,3) = iner(3,1)

    ! Take the safest route, i.e. diagonalize the inertia tensor and solve
    ! the angular velocity only with the non-zero eigenvalues. A plain inversion
    ! would fail for linear molecules.
    CALL diamat_all(iner, evals)

    vang(:) = 0.0_dp
    DO i=1,3
       IF (evals(i) > 0.0_dp) THEN
         proj = SUM(iner(:,i)*mang)/evals(i)
         vang(1) = vang(1) + proj*iner(1,i)
         vang(2) = vang(2) + proj*iner(2,i)
         vang(3) = vang(3) + proj*iner(3,i)
       END IF
    END DO

  END SUBROUTINE compute_vang

! **************************************************************************************************
!> \brief subtract the angular velocity
!>      *** is only used by initialize_velocities below ***
!> \param part ...
!> \param is_fixed ...
!> \param rcom ...
!> \param vang ...
!> \par History
!>      2007-11-9: created
!> \author Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE subtract_vang(part,is_fixed,rcom,vang)
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      INTEGER, DIMENSION(:), INTENT(IN)                  :: is_fixed
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rcom, vang

      CHARACTER(len=*), PARAMETER :: routineN = 'subtract_vang', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(3)                        :: r

    DO i=1,SIZE(part)
       r(:) = part(i)%r(:) - rcom(:)
       SELECT CASE(is_fixed(i))
       CASE(use_perd_x)
          part(i)%v(2) = part(i)%v(2) - (vang(3)*r(1) - vang(1)*r(3))
          part(i)%v(3) = part(i)%v(3) - (vang(1)*r(2) - vang(2)*r(1))
       CASE(use_perd_y)
          part(i)%v(1) = part(i)%v(1) - (vang(2)*r(3) - vang(3)*r(2))
          part(i)%v(3) = part(i)%v(3) - (vang(1)*r(2) - vang(2)*r(1))
       CASE(use_perd_z)
          part(i)%v(1) = part(i)%v(1) - (vang(2)*r(3) - vang(3)*r(2))
          part(i)%v(2) = part(i)%v(2) - (vang(3)*r(1) - vang(1)*r(3))
       CASE(use_perd_xy)
          part(i)%v(3) = part(i)%v(3) - (vang(1)*r(2) - vang(2)*r(1))
       CASE(use_perd_xz)
          part(i)%v(2) = part(i)%v(2) - (vang(3)*r(1) - vang(1)*r(3))
       CASE(use_perd_yz)
          part(i)%v(1) = part(i)%v(1) - (vang(2)*r(3) - vang(3)*r(2))
       CASE(use_perd_none)
          part(i)%v(1) = part(i)%v(1) - (vang(2)*r(3) - vang(3)*r(2))
          part(i)%v(2) = part(i)%v(2) - (vang(3)*r(1) - vang(1)*r(3))
          part(i)%v(3) = part(i)%v(3) - (vang(1)*r(2) - vang(2)*r(1))
       END SELECT
    END DO

  END SUBROUTINE subtract_vang

! **************************************************************************************************
!> \brief Initializes the velocities to the Maxwellian distribution
!> \param simpar ...
!> \param part ...
!> \param force_env ...
!> \param globenv ...
!> \param md_env ...
!> \param molecule_kinds ...
!> \param label ...
!> \param print_section ...
!> \param subsys_section ...
!> \param shell_present ...
!> \param shell_part ...
!> \param core_part ...
!> \param force_rescaling ...
!> \param para_env ...
!> \par History
!>      - is_fixed removed from particle_type
!>      - 2007-11-07: Cleanup (TV)
!>      - 2007-11-09: Added angvel_zero feature
!> \author CJM,MK,Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE initialize_velocities(simpar, part, force_env, globenv, md_env, &
       molecule_kinds, label, print_section, subsys_section, shell_present,&
       shell_part, core_part, force_rescaling, para_env)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      CHARACTER(LEN=*), INTENT(IN)                       :: label
      TYPE(section_vals_type), POINTER                   :: print_section, subsys_section
      LOGICAL, INTENT(IN)                                :: shell_present
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_part, core_part
      LOGICAL, INTENT(IN)                                :: force_rescaling
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_velocities', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_format
      INTEGER                                            :: handle, i, ifixd, imolecule_kind, iw, &
                                                            natoms, num_x
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: is_fixed
      LOGICAL                                            :: success
      REAL(KIND=dp)                                      :: ecom, ekin, mass, mass_tot, temp, tmp_r1
      REAL(KIND=dp), DIMENSION(3)                        :: rcom, vang, vcom
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

    CALL timeset(routineN,handle)

    ! Initializing parameters
    natoms = SIZE(part)
    NULLIFY (atomic_kind, fixd_list, logger, molecule_kind)
    NULLIFY (molecule_kind_set)

    ! Logging
    logger => cp_get_default_logger()
    iw=cp_print_key_unit_nr(logger,print_section,"PROGRAM_RUN_INFO",extension=".log")
    IF (iw>0) THEN
       num_x = (79-LEN_TRIM(ADJUSTL(label))-2)/2
       WRITE(my_format,'(A,I0,A,I0,A)')'(1X,',num_x,'("*"),1X,A,1X,',num_x,'("*"))'
       WRITE ( iw, TRIM(my_format))TRIM(ADJUSTL(label))
    END IF

    ! Build a list of all fixed atoms (if any)
    ALLOCATE (is_fixed(natoms))

    is_fixed = use_perd_none
    molecule_kind_set => molecule_kinds%els
    DO imolecule_kind=1,molecule_kinds%n_els
       molecule_kind => molecule_kind_set(imolecule_kind)
       CALL get_molecule_kind(molecule_kind=molecule_kind,fixd_list=fixd_list)
       IF (ASSOCIATED(fixd_list)) THEN
          DO ifixd=1,SIZE(fixd_list)
             IF (.NOT.fixd_list(ifixd)%restraint%active) is_fixed(fixd_list(ifixd)%fixd) = fixd_list(ifixd)%itype
          END DO
       END IF
    END DO

    ! Compute the total mass when needed
    IF ( simpar%ensemble == nph_uniaxial_ensemble .OR.&
         simpar%ensemble == nph_uniaxial_damped_ensemble ) THEN
       mass_tot = 0.0_dp
       DO i = 1, natoms
          atomic_kind => part(i)%atomic_kind
          CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
          mass_tot = mass_tot + mass
       END DO
       simpar%v_shock = simpar%v_shock * SQRT ( mass_tot )
    END IF


    CALL read_input_velocities(simpar, part, force_env, md_env, subsys_section,&
             shell_present, shell_part, core_part, force_rescaling, para_env, is_fixed, success)
    IF(.NOT.success) &
       CALL generate_velocities(simpar, part, force_env, globenv, md_env, shell_present,&
                shell_part, core_part, is_fixed, iw)


    IF (iw>0) THEN
       ! Recompute vcom, ecom and ekin for IO
       CALL compute_vcom(part,is_fixed,vcom,ecom)
       ekin = compute_ekin(part) - ecom
       IF (simpar%nfree == 0) THEN
          IF(.NOT.(ekin==0.0_dp))CALL cp__a("motion/md_vel_utils.F",634)
          temp = 0.0_dp
       ELSE
          temp = 2.0_dp * ekin / REAL ( simpar%nfree,KIND=dp)
       END IF
       tmp_r1 = cp_unit_from_cp2k(temp,"K")
       WRITE (iw, '( A, T61, F18.2, A2 )' ) ' Initial Temperature ', tmp_r1, " K"
       WRITE (iw, '( A, T21, F20.12 , F20.12 , F20.12 )' ) ' COM velocity:', vcom ( 1 ), vcom ( 2 ), vcom ( 3 )

       ! compute and log rcom and vang if not periodic
       CALL force_env_get(force_env, cell=cell)
       IF (SUM(cell%perd(1:3)) == 0) THEN
          CALL compute_rcom(part,is_fixed,rcom)
          CALL compute_vang(part,is_fixed,rcom,vang)
          WRITE (iw, '( A, T21, F20.12 , F20.12 , F20.12 )' ) ' COM position:', rcom ( 1 ), rcom ( 2 ), rcom ( 3 )
          WRITE (iw, '( A, T21, F20.12 , F20.12 , F20.12 )' ) ' Angular velocity:', vang ( 1 ), vang ( 2 ), vang ( 3 )
       END IF
       WRITE ( iw, '( 1X,79("*"),/)' )
    END IF

    DEALLOCATE (is_fixed)
    CALL cp_print_key_finished_output(iw,logger,print_section,"PROGRAM_RUN_INFO")
    CALL timestop(handle)

  END SUBROUTINE initialize_velocities


! **************************************************************************************************
!> \brief Read velocities from binary restart file if available
!> \param simpar ...
!> \param part ...
!> \param force_env ...
!> \param md_env ...
!> \param subsys_section ...
!> \param shell_present ...
!> \param shell_part ...
!> \param core_part ...
!> \param force_rescaling ...
!> \param para_env ...
!> \param is_fixed ...
!> \param success ...
!> \author CJM,MK,Toon Verstraelen <Toon.Verstraelen@gmail.com>
! **************************************************************************************************
  SUBROUTINE read_input_velocities(simpar, part, force_env, md_env, subsys_section,&
       shell_present, shell_part, core_part, force_rescaling, para_env, is_fixed, success)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: subsys_section
      LOGICAL, INTENT(IN)                                :: shell_present
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_part, core_part
      LOGICAL, INTENT(IN)                                :: force_rescaling
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: is_fixed
      LOGICAL, INTENT(OUT)                               :: success

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_input_velocities', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, natoms, nshell, shell_index
      LOGICAL :: atomvel_explicit, atomvel_read, corevel_explicit, corevel_read, is_ok, &
         rescale_regions, shellvel_explicit, shellvel_read
      REAL(KIND=dp)                                      :: ecom, ekin, fac_massc, fac_masss, mass
      REAL(KIND=dp), DIMENSION(3)                        :: vc, vcom, vs
      REAL(KIND=dp), DIMENSION(:), POINTER               :: vel
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_sll_val_type), POINTER                     :: atom_list, core_list, shell_list
      TYPE(section_vals_type), POINTER                   :: atomvel_section, corevel_section, &
                                                            shellvel_section
      TYPE(shell_kind_type), POINTER                     :: shell
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(val_type), POINTER                            :: val

! Initializing parameters

    success = .FALSE.
    natoms = SIZE(part)
    atomvel_read = .FALSE.
    corevel_read = .FALSE.
    shellvel_read = .FALSE.
    NULLIFY (vel, atomic_kind, atom_list, core_list, shell_list)
    NULLIFY (atomvel_section,shellvel_section, corevel_section)
    NULLIFY (shell, thermal_regions, val)


    ! Core-Shell Model
    nshell = 0
    IF (shell_present) THEN
       IF(.NOT.(ASSOCIATED(core_part)))CALL cp__a("motion/md_vel_utils.F",723)
       IF(.NOT.(ASSOCIATED(shell_part)))CALL cp__a("motion/md_vel_utils.F",724)
       nshell = SIZE(shell_part)
    END IF

    atomvel_section  => section_vals_get_subs_vals(subsys_section,"VELOCITY")
    shellvel_section => section_vals_get_subs_vals(subsys_section,"SHELL_VELOCITY")
    corevel_section => section_vals_get_subs_vals(subsys_section,"CORE_VELOCITY")

    ! Read or initialize the particle velocities
    CALL section_vals_get(atomvel_section,explicit=atomvel_explicit)
    CALL section_vals_get(shellvel_section,explicit=shellvel_explicit)
    CALL section_vals_get(corevel_section,explicit=corevel_explicit)
    IF(.NOT.(shellvel_explicit.EQV.corevel_explicit))CALL cp__a("motion/md_vel_utils.F",736)

    CALL read_binary_velocities("",part,force_env%root_section,para_env,&
                                subsys_section,atomvel_read)
    CALL read_binary_velocities("SHELL",shell_part,force_env%root_section,para_env,&
                                subsys_section,shellvel_read)
    CALL read_binary_velocities("CORE",core_part,force_env%root_section,para_env,&
                                subsys_section,corevel_read)

    IF (.NOT.(atomvel_explicit.OR.atomvel_read)) RETURN
    success = .TRUE.

     IF (.NOT.atomvel_read) THEN
        ! Read the atom velocities if explicitly given in the input file
        CALL section_vals_list_get(atomvel_section,"_DEFAULT_KEYWORD_",list=atom_list)
        DO i = 1, natoms
           is_ok = cp_sll_val_next(atom_list,val)
           CALL val_get(val,r_vals=vel)
           part(i)%v = vel
        END DO
     END IF
     DO i = 1, natoms
       SELECT CASE(is_fixed(i))
       CASE (use_perd_x)
          part(i)%v(1) = 0.0_dp
       CASE (use_perd_y)
          part(i)%v(2) = 0.0_dp
       CASE (use_perd_z)
          part(i)%v(3) = 0.0_dp
       CASE (use_perd_xy)
          part(i)%v(1) = 0.0_dp
          part(i)%v(2) = 0.0_dp
       CASE (use_perd_xz)
          part(i)%v(1) = 0.0_dp
          part(i)%v(3) = 0.0_dp
       CASE (use_perd_yz)
          part(i)%v(2) = 0.0_dp
          part(i)%v(3) = 0.0_dp
       CASE (use_perd_xyz)
          part(i)%v    = 0.0_dp
       END SELECT
     END DO
     IF (shell_present) THEN
        IF (shellvel_explicit) THEN
           ! If the atoms positions are given (?) and core and shell velocities are
           ! present in the input, read the latter.
           CALL section_vals_list_get(shellvel_section,"_DEFAULT_KEYWORD_",list=shell_list)
           CALL section_vals_list_get(corevel_section,"_DEFAULT_KEYWORD_",list=core_list)
           DO i=1,nshell
              is_ok = cp_sll_val_next(shell_list,val)
              CALL val_get(val,r_vals=vel)
              shell_part(i)%v = vel
              is_ok = cp_sll_val_next(core_list,val)
              CALL val_get(val,r_vals=vel)
              core_part(i)%v = vel
           END DO
        ELSE
           IF (.NOT.(shellvel_read.AND.corevel_read)) THEN
              ! Otherwise, just copy atom velocties into shell and core velocities.
              CALL clone_core_shell_vel(part,shell_part,core_part)
           END IF
        END IF
     END IF

     ! compute vcom, ecom and ekin
     CALL compute_vcom(part,is_fixed,vcom,ecom)
     ekin = compute_ekin(part) - ecom

     IF(simpar%do_thermal_region) THEN
       CALL get_md_env (md_env, thermal_regions=thermal_regions)
       IF(ASSOCIATED(thermal_regions)) THEN
         rescale_regions = thermal_regions%force_rescaling
       END IF
     ELSE
       rescale_regions = .FALSE.
     END IF
     IF (simpar%nfree /= 0 .AND. (force_rescaling .OR. rescale_regions)) THEN
        IF(simpar%do_thermal_region) THEN
          CALL rescale_vel_region(part,md_env,simpar)
        ELSE
          CALL rescale_vel(part,simpar,ekin,vcom=vcom)
        END IF

        ! After rescaling, the core and shell velocities must also adapt.
        DO i = 1, natoms
          shell_index = part(i)%shell_index
          IF(shell_present .AND. shell_index/=0) THEN
              atomic_kind => part(i)%atomic_kind
              CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, shell=shell)
              fac_masss = shell%mass_shell/mass
              fac_massc = shell%mass_core/mass
              vs = shell_part(shell_index)%v
              vc = core_part(shell_index)%v

              shell_part(shell_index)%v(1) = part(i)%v(1) + fac_massc*(vs(1)-vc(1))
              shell_part(shell_index)%v(2) = part(i)%v(2) + fac_massc*(vs(2)-vc(2))
              shell_part(shell_index)%v(3) = part(i)%v(3) + fac_massc*(vs(3)-vc(3))
              core_part(shell_index)%v(1)  = part(i)%v(1) + fac_masss*(vc(1)-vs(1))
              core_part(shell_index)%v(2)  = part(i)%v(2) + fac_masss*(vc(2)-vs(2))
              core_part(shell_index)%v(3)  = part(i)%v(3) + fac_masss*(vc(3)-vs(3))
          END IF
        END DO
     END IF
  END SUBROUTINE read_input_velocities


! **************************************************************************************************
!> \brief Initializing velocities deterministically on all processors, if not given in input
!> \param simpar ...
!> \param part ...
!> \param force_env ...
!> \param globenv ...
!> \param md_env ...
!> \param shell_present ...
!> \param shell_part ...
!> \param core_part ...
!> \param is_fixed ...
!> \param iw ...
!> \author CJM,MK,Toon Verstraelen <Toon.Verstraelen@gmail.com>, Ole Schuett
! **************************************************************************************************
  SUBROUTINE generate_velocities(simpar, part, force_env, globenv, md_env,&
                          shell_present, shell_part, core_part, is_fixed, iw)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(md_environment_type), POINTER                 :: md_env
      LOGICAL, INTENT(IN)                                :: shell_present
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_part, core_part
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: is_fixed
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'generate_velocities', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, natoms
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

    NULLIFY (atomic_kind)
    natoms = SIZE(part)

     DO i = 1, natoms
        atomic_kind => part(i)%atomic_kind
        CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
        part(i)%v(1) = 0.0_dp
        part(i)%v(2) = 0.0_dp
        part(i)%v(3) = 0.0_dp
        IF (mass.NE.0.0) THEN
           SELECT CASE(is_fixed(i))
           CASE (use_perd_x)
              part(i)%v(2) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
              part(i)%v(3) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_y)
              part(i)%v(1) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
              part(i)%v(3) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_z)
              part(i)%v(1) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
              part(i)%v(2) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_xy)
              part(i)%v(3) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_xz)
              part(i)%v(2) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_yz)
              part(i)%v(1) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           CASE (use_perd_none)
              part(i)%v(1) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
              part(i)%v(2) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
              part(i)%v(3) = next_random_number(globenv%gaussian_rng_stream) / SQRT(mass)
           END SELECT
        END IF
     END DO

     CALL normalize_velocities(simpar, part, force_env, md_env, is_fixed)
     CALL soften_velocities(simpar, part, force_env, md_env, is_fixed, iw)

     ! Initialize the core and the shell velocity. Atom velocities are just
     ! copied so that the initial relative core-shell velocity is zero.
     IF (shell_present) THEN
       CALL optimize_shell_core(force_env,part,shell_part, core_part, globenv)
     ENDIF
  END SUBROUTINE generate_velocities

! **************************************************************************************************
!> \brief Direct velocities along a low-curvature direction in order to 
!>        favors MD trajectories to cross rapidly over small energy barriers
!>        into neighboring basins.
!> \param simpar ...
!> \param part ...
!> \param force_env ...
!> \param md_env ...
!> \param is_fixed ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
  SUBROUTINE soften_velocities(simpar, part, force_env, md_env, is_fixed, iw)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_environment_type), POINTER                 :: md_env
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: is_fixed
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'soften_velocities', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k
      REAL(KIND=dp), DIMENSION(SIZE(part), 3)            :: F, F_t, N, x0

    IF(simpar%soften_nsteps <= 0) RETURN !nothing todo

    IF(ANY(is_fixed/=use_perd_none))&
       CALL cp__b("motion/md_vel_utils.F",948,"Velocitiy softening with constraints is not supported.")

    !backup positions
    DO i = 1, SIZE(part)
          x0(i,:) = part(i)%r
    END DO

    DO k=1, simpar%soften_nsteps

       !use normalized velocities as displace direction
       DO i = 1, SIZE(part)
          N(i,:) = part(i)%v
       END DO
       N = N / SQRT(SUM(N**2))

       ! displace system temporarly to calculate forces
       DO i = 1, SIZE(part)
          part(i)%r = part(i)%r + simpar%soften_delta * N(i,:)
       END DO
       CALL force_env_calc_energy_force(force_env)

       ! calculate velocity update direction F_t
       DO i = 1, SIZE(part)
          F(i,:) = part(i)%f
       END DO
       F_t = F - N * SUM(N*F)

       ! restore positions and update velocities
       DO i = 1, SIZE(part)
          part(i)%r = x0(i,:)
          part(i)%v = part(i)%v + simpar%soften_alpha * F_t(i,:)
       END DO

       CALL normalize_velocities(simpar, part, force_env, md_env, is_fixed)
    END DO

    IF(iw>0) THEN
      WRITE (iw, "(A,T71, I10)") " Velocities softening Steps: ",simpar%soften_nsteps
      WRITE (iw, "(A,T71, E10.3)") " Velocities softening NORM(F_t): ",SQRT(SUM(F_t**2))
    END IF
  END SUBROUTINE soften_velocities


! **************************************************************************************************
!> \brief Scale velocities according to temperature and remove rigid body motion.
!> \param simpar ...
!> \param part ...
!> \param force_env ...
!> \param md_env ...
!> \param is_fixed ...
!> \author CJM,MK,Toon Verstraelen <Toon.Verstraelen@gmail.com>, Ole Schuett
! **************************************************************************************************
  SUBROUTINE normalize_velocities(simpar, part, force_env, md_env, is_fixed)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_environment_type), POINTER                 :: md_env
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: is_fixed

      CHARACTER(LEN=*), PARAMETER :: routineN = 'normalize_velocities', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: ekin
      REAL(KIND=dp), DIMENSION(3)                        :: rcom, vang, vcom
      TYPE(cell_type), POINTER                           :: cell

     NULLIFY(cell)

     ! Subtract the vcom
     CALL compute_vcom(part,is_fixed,vcom)
     CALL subtract_vcom(part,is_fixed,vcom)
     ! If requested and the system is not periodic, subtract the angular velocity
     CALL force_env_get(force_env, cell=cell)
     IF (SUM(cell%perd(1:3)) == 0 .AND. simpar%angvel_zero) THEN
        CALL compute_rcom(part,is_fixed,rcom)
        CALL compute_vang(part,is_fixed,rcom,vang)
        CALL subtract_vang(part,is_fixed,rcom,vang)
     END IF
     ! Rescale the velocities
     IF(simpar%do_thermal_region) THEN
       CALL rescale_vel_region(part,md_env,simpar)
     ELSE
       ekin = compute_ekin(part)
       CALL rescale_vel(part,simpar,ekin)
     END IF
  END SUBROUTINE normalize_velocities


! **************************************************************************************************
!> \brief Computes Ekin, VCOM and Temp for particles
!> \param subsys ...
!> \param md_ener ...
!> \param vsubtract ...
!> \par History
!>     Teodoro Laino - University of Zurich - 09.2007 [tlaino]
! **************************************************************************************************
  SUBROUTINE reset_vcom(subsys, md_ener, vsubtract)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(md_ener_type), POINTER                        :: md_ener
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: vsubtract

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reset_vcom', routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, handle, iatom, ikind, natom, &
                                                            shell_index
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: ekin_old, imass_c, imass_s, mass, v2
      REAL(KIND=dp), DIMENSION(3)                        :: tmp, v
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(shell_kind_type), POINTER                     :: shell

    NULLIFY(particles, atomic_kind, atomic_kinds, atom_list, shell)
    CALL timeset(routineN,handle)

    CALL cp_subsys_get(subsys,&
                       atomic_kinds=atomic_kinds,&
                       particles=particles,&
                       shell_particles=shell_particles,&
                       core_particles=core_particles)

    ekin_old = md_ener%ekin
    ! Possibly subtract a quantity from all velocities
       DO ikind=1,atomic_kinds%n_els
          atomic_kind => atomic_kinds%els(ikind)
          CALL get_atomic_kind(atomic_kind=atomic_kind, atom_list=atom_list,&
             natom=natom, mass=mass, shell_active=is_shell, shell=shell)
          IF (is_shell) THEN
             tmp = 0.5_dp*vsubtract*mass
             imass_s = 1.0_dp/shell%mass_shell
             imass_c = 1.0_dp/shell%mass_core
             DO iatom = 1, natom
                atom = atom_list(iatom)
                shell_index = particles%els(atom)%shell_index
                shell_particles%els(shell_index)%v = shell_particles%els(shell_index)%v - tmp*imass_s
                core_particles%els(shell_index)%v  = core_particles%els(shell_index)%v  - tmp*imass_c
                particles%els(atom)%v = particles%els(atom)%v - vsubtract
             END DO
          ELSE
             DO iatom = 1, natom
                atom = atom_list(iatom)
                particles%els(atom)%v = particles%els(atom)%v - vsubtract
             END DO
          END IF
       END DO
    ! Compute Kinetic Energy and COM Velocity
    md_ener%vcom       = 0.0_dp
    md_ener%total_mass = 0.0_dp
    md_ener%ekin       = 0.0_dp
    DO ikind=1,atomic_kinds%n_els
       atomic_kind => atomic_kinds%els(ikind)
       CALL get_atomic_kind(atomic_kind=atomic_kind, atom_list=atom_list, mass=mass, natom=natom)
          v2   = 0.0_dp
          v    = 0.0_dp
          DO iatom = 1, natom
             atom = atom_list(iatom)
             v2 = v2 + SUM(particles%els(atom)%v**2)
             v(1) = v(1) + particles%els(atom)%v(1)
             v(2) = v(2) + particles%els(atom)%v(2)
             v(3) = v(3) + particles%els(atom)%v(3)
          END DO
          md_ener%ekin  = md_ener%ekin + 0.5_dp*mass*v2
          md_ener%vcom(1)    = md_ener%vcom(1) + mass*v(1)
          md_ener%vcom(2)    = md_ener%vcom(2) + mass*v(2)
          md_ener%vcom(3)    = md_ener%vcom(3) + mass*v(3)
          md_ener%total_mass = md_ener%total_mass + REAL(natom,KIND=dp)*mass
    END DO
    md_ener%vcom = md_ener%vcom / md_ener%total_mass
    md_ener%constant = md_ener%constant - ekin_old + md_ener%ekin
    IF (md_ener%nfree /=0) THEN
         md_ener%temp_part = 2.0_dp*md_ener%ekin/REAL(md_ener%nfree,KIND=dp)*kelvin
    END IF
    CALL timestop(handle)

  END SUBROUTINE reset_vcom

! **************************************************************************************************
!> \brief Scale velocities to get the correct temperature
!> \param subsys ...
!> \param md_ener ...
!> \param temp_expected ...
!> \param temp_tol ...
!> \param iw ...
!> \par History
!>     Teodoro Laino - University of Zurich - 09.2007 [tlaino]
! **************************************************************************************************
  SUBROUTINE scale_velocity(subsys, md_ener, temp_expected, temp_tol, iw)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(md_ener_type), POINTER                        :: md_ener
      REAL(KIND=dp), INTENT(IN)                          :: temp_expected, temp_tol
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scale_velocity', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: ekin_old, scale, temp_old

    IF (ABS(temp_expected - md_ener%temp_part/kelvin) > temp_tol) THEN
       scale = 0.0_dp
       IF (md_ener%temp_part>0.0_dp) scale = SQRT((temp_expected/md_ener%temp_part)*kelvin)
       ekin_old = md_ener%ekin
       temp_old = md_ener%temp_part
       md_ener%ekin       = 0.0_dp
       md_ener%temp_part  = 0.0_dp
       md_ener%vcom       = 0.0_dp
       md_ener%total_mass = 0.0_dp

       CALL scale_velocity_low(subsys,scale,ireg=0,ekin=md_ener%ekin,vcom=md_ener%vcom)
       IF(md_ener%nfree /=0) THEN
           md_ener%temp_part = 2.0_dp*md_ener%ekin/REAL(md_ener%nfree,KIND=dp)*kelvin
       END IF
       md_ener%constant = md_ener%constant - ekin_old + md_ener%ekin

       IF (iw>0) THEN
          WRITE (UNIT=iw,FMT="(/,T2,A,F10.2,A,F10.2,A)")"Temperature scaled to requested temperature:",&
             temp_old," K  ->",md_ener%temp_part," K"
       END IF
    END IF
  END SUBROUTINE scale_velocity

! **************************************************************************************************
!> \brief Scale velocities of set of regions
!> \param md_env ...
!> \param subsys ...
!> \param md_ener ...
!> \param simpar ...
!> \param iw ...
!> \par author MI
! **************************************************************************************************
  SUBROUTINE scale_velocity_region(md_env, subsys, md_ener, simpar, iw)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scale_velocity_region', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ireg, nfree, nfree_done, nregions
      REAL(KIND=dp)                                      :: ekin, ekin_old, ekin_total_new, fscale, &
                                                            vcom(3), vcom_total(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: temp_new, temp_old
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: part
      TYPE(thermal_region_type), POINTER                 :: t_region
      TYPE(thermal_regions_type), POINTER                :: thermal_regions

     NULLIFY( particles, part, thermal_regions, t_region)
     CALL cp_subsys_get(subsys, particles=particles)
     part => particles%els
     CALL get_md_env(md_env, thermal_regions=thermal_regions)

     nregions = thermal_regions%nregions
     nfree_done = 0
     ekin_total_new = 0.0_dp
     ekin_old = md_ener%ekin
     vcom_total = 0.0_dp
     ALLOCATE(temp_new(0:nregions),temp_old(0:nregions))
     temp_new = 0.0_dp
     temp_old = 0.0_dp
     !loop regions
     DO ireg = 1,nregions
       NULLIFY(t_region)
       t_region => thermal_regions%thermal_region(ireg)
       nfree = 3*t_region%npart
       ekin = compute_ekin(part,ireg)
       IF(nfree > 0) t_region%temperature = 2.0_dp*ekin/REAL(nfree,KIND=dp)*kelvin
       temp_old(ireg) =  t_region%temperature
       IF(t_region%temp_tol > 0.0_dp .AND. &
         ABS(t_region%temp_expected - t_region%temperature/kelvin) > t_region%temp_tol) THEN
           fscale =  SQRT((t_region%temp_expected/t_region%temperature)*kelvin)
           CALL scale_velocity_low(subsys,fscale,ireg,ekin,vcom)
           t_region%temperature = 2.0_dp*ekin/REAL(nfree,KIND=dp)*kelvin
           temp_new(ireg) = t_region%temperature
       END IF
       nfree_done = nfree_done + nfree
       ekin_total_new = ekin_total_new + ekin
     END DO
     nfree = simpar%nfree - nfree_done
     ekin = compute_ekin(part,ireg=0)
     IF(nfree>0) thermal_regions%temp_reg0 = 2.0_dp*ekin/REAL(nfree,KIND=dp)*kelvin
     temp_old(0) =  thermal_regions%temp_reg0
     IF(simpar%temp_tol > 0.0_dp .AND. nfree>0) THEN
        IF (ABS(simpar%temp_ext - thermal_regions%temp_reg0/kelvin) > simpar%temp_tol) THEN
           fscale =  SQRT((simpar%temp_ext/thermal_regions%temp_reg0)*kelvin)
           CALL scale_velocity_low(subsys,fscale,0,ekin,vcom)
           thermal_regions%temp_reg0 = 2.0_dp*ekin/REAL(nfree,KIND=dp)*kelvin
           temp_new(0) = thermal_regions%temp_reg0
        END IF
     END IF
     ekin_total_new = ekin_total_new + ekin

     md_ener%ekin = ekin_total_new
     IF(md_ener%nfree /=0) THEN
         md_ener%temp_part = 2.0_dp*md_ener%ekin/REAL(md_ener%nfree,KIND=dp)*kelvin
     END IF
     md_ener%constant = md_ener%constant - ekin_old + md_ener%ekin

     DO ireg = 0,nregions
       IF (iw>0 .AND. temp_new(ireg)>0.0_dp) THEN
          WRITE (UNIT=iw,FMT="(/,T2,A,I5, A,F10.2,A,F10.2,A)")"Temperature region ", ireg, &
             " rescaled from:",temp_old(ireg)," K  to ",temp_new(ireg)," K"
       END IF
     END DO
     DEALLOCATE(temp_new,temp_old)

  END SUBROUTINE scale_velocity_region

! **************************************************************************************************
!> \brief Scale velocities  for a specific region
!> \param subsys ...
!> \param fscale ...
!> \param ireg ...
!> \param ekin ...
!> \param vcom ...
!> \par author MI
! **************************************************************************************************
  SUBROUTINE scale_velocity_low(subsys,fscale,ireg,ekin,vcom)

      TYPE(cp_subsys_type), POINTER                      :: subsys
      REAL(KIND=dp), INTENT(IN)                          :: fscale
      INTEGER, INTENT(IN)                                :: ireg
      REAL(KIND=dp), INTENT(OUT)                         :: ekin, vcom(3)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scale_velocity_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, iatom, ikind, my_ireg, natom, &
                                                            shell_index
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: imass, mass, tmass, v2
      REAL(KIND=dp), DIMENSION(3)                        :: tmp, v, vc, vs
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(shell_kind_type), POINTER                     :: shell

    NULLIFY(atomic_kinds,particles,shell_particles,core_particles,shell,atom_list)

    my_ireg = ireg
    ekin = 0.0_dp
    tmass = 0.0_dp
    vcom = 0.0_dp

    CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, particles=particles,&
          shell_particles=shell_particles, core_particles=core_particles)

    DO ikind=1,atomic_kinds%n_els
       atomic_kind => atomic_kinds%els(ikind)
       CALL get_atomic_kind(atomic_kind=atomic_kind, atom_list=atom_list, mass=mass,&
            natom=natom, shell_active=is_shell, shell=shell)
       IF(is_shell) THEN
          imass = 1.0_dp / mass
          v2 = 0.0_dp
          v  = 0.0_dp
          DO iatom=1,natom
             atom = atom_list(iatom)
             !check region
             IF( particles%els(atom)%t_region_index/=my_ireg) CYCLE

             particles%els(atom)%v(:) = fscale*particles%els(atom)%v
             shell_index = particles%els(atom)%shell_index
             vs = shell_particles%els(shell_index)%v
             vc = core_particles %els(shell_index)%v
             tmp(1) = imass*(vs(1)-vc(1))
             tmp(2) = imass*(vs(2)-vc(2))
             tmp(3) = imass*(vs(3)-vc(3))

             shell_particles%els(shell_index)%v(1) = particles%els(atom)%v(1) + tmp(1)*shell%mass_core
             shell_particles%els(shell_index)%v(2) = particles%els(atom)%v(2) + tmp(2)*shell%mass_core
             shell_particles%els(shell_index)%v(3) = particles%els(atom)%v(3) + tmp(3)*shell%mass_core

             core_particles%els(shell_index)%v(1) =  particles%els(atom)%v(1) - tmp(1)*shell%mass_shell
             core_particles%els(shell_index)%v(2) =  particles%els(atom)%v(2) - tmp(2)*shell%mass_shell
             core_particles%els(shell_index)%v(3) =  particles%els(atom)%v(3) - tmp(3)*shell%mass_shell

             ! kinetic energy and velocity of COM
             v2 = v2 + SUM(particles%els(atom)%v**2)
             v(1) = v(1) + particles%els(atom)%v(1)
             v(2) = v(2) + particles%els(atom)%v(2)
             v(3) = v(3) + particles%els(atom)%v(3)
             tmass = tmass + mass
          END DO
       ELSE
          v2 = 0.0_dp
          v = 0.0_dp
          DO iatom=1,natom
             atom = atom_list(iatom)
             !check region
             IF( particles%els(atom)%t_region_index/=my_ireg) CYCLE

             particles%els(atom)%v(:) = fscale*particles%els(atom)%v
            ! kinetic energy and velocity of COM
             v2 = v2 + SUM(particles%els(atom)%v**2)
             v(1) = v(1) + particles%els(atom)%v(1)
             v(2) = v(2) + particles%els(atom)%v(2)
             v(3) = v(3) + particles%els(atom)%v(3)
             tmass = tmass + mass
          END DO
       END IF
       ekin  = ekin + 0.5_dp*mass*v2
       vcom(1)    = vcom(1) + mass*v(1)
       vcom(2)    = vcom(2) + mass*v(2)
       vcom(3)    = vcom(3) + mass*v(3)

     END DO
     vcom = vcom / tmass

  END SUBROUTINE scale_velocity_low

! **************************************************************************************************
!> \brief Scale internal motion of CORE-SHELL model to the correct temperature
!> \param subsys ...
!> \param md_ener ...
!> \param temp_expected ...
!> \param temp_tol ...
!> \param iw ...
!> \par History
!>     Teodoro Laino - University of Zurich - 09.2007 [tlaino]
! **************************************************************************************************
  SUBROUTINE scale_velocity_internal(subsys, md_ener,  temp_expected, temp_tol, iw)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(md_ener_type), POINTER                        :: md_ener
      REAL(KIND=dp), INTENT(IN)                          :: temp_expected, temp_tol
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scale_velocity_internal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, iatom, ikind, natom, shell_index
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: ekin_shell_old, fac_mass, mass, scale, &
                                                            temp_shell_old, v2
      REAL(KIND=dp), DIMENSION(3)                        :: tmp, v, vc, vs
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(shell_kind_type), POINTER                     :: shell

    NULLIFY(atom_list,atomic_kinds,atomic_kind,core_particles,particles,shell_particles,shell)
    IF (ABS(temp_expected - md_ener%temp_shell/kelvin) > temp_tol) THEN
       scale = 0.0_dp
       IF (md_ener%temp_shell>EPSILON(0.0_dp)) scale = SQRT((temp_expected/md_ener%temp_shell)*kelvin)
       ekin_shell_old = md_ener%ekin_shell
       temp_shell_old = md_ener%temp_shell
       md_ener%ekin_shell = 0.0_dp
       md_ener%temp_shell = 0.0_dp

       CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, particles=particles, shell_particles=shell_particles,&
          core_particles=core_particles)

       DO ikind=1,atomic_kinds%n_els
          atomic_kind => atomic_kinds%els(ikind)
          CALL get_atomic_kind(atomic_kind=atomic_kind, atom_list=atom_list, mass=mass, natom=natom,&
             shell_active=is_shell, shell=shell)
          IF(is_shell) THEN
             fac_mass = 1.0_dp/mass
             v2 = 0.0_dp
             DO iatom= 1, natom
                atom = atom_list(iatom)
                shell_index = particles%els(atom)%shell_index
                vs  = shell_particles%els(shell_index)%v
                vc  = core_particles%els(shell_index)%v
                v   = particles%els(atom)%v
                tmp(1) = fac_mass*(vc(1)-vs(1))
                tmp(2) = fac_mass*(vc(2)-vs(2))
                tmp(3) = fac_mass*(vc(3)-vs(3))

                shell_particles%els(shell_index)%v(1) = v(1) - shell%mass_core*scale*tmp(1)
                shell_particles%els(shell_index)%v(2) = v(2) - shell%mass_core*scale*tmp(2)
                shell_particles%els(shell_index)%v(3) = v(3) - shell%mass_core*scale*tmp(3)

                core_particles%els(shell_index)%v(1) =  v(1) + shell%mass_shell*scale*tmp(1)
                core_particles%els(shell_index)%v(2) =  v(2) + shell%mass_shell*scale*tmp(2)
                core_particles%els(shell_index)%v(3) =  v(3) + shell%mass_shell*scale*tmp(3)

                vs  = shell_particles%els(shell_index)%v
                vc  = core_particles%els(shell_index)%v
                tmp(1) = vc(1) - vs(1)
                tmp(2) = vc(2) - vs(2)
                tmp(3) = vc(3) - vs(3)
                v2 = v2 + SUM(tmp**2)
             END DO
             md_ener%ekin_shell = md_ener%ekin_shell + 0.5_dp*shell%mass_core*shell%mass_shell*fac_mass*v2
          END IF
       END DO
       IF(md_ener%nfree_shell>0)THEN
           md_ener%temp_shell = 2.0_dp*md_ener%ekin_shell/REAL(md_ener%nfree_shell,KIND=dp)*kelvin
       END IF
       md_ener%constant = md_ener%constant - ekin_shell_old + md_ener%ekin_shell

       IF (iw>0) THEN
          WRITE (UNIT=iw,FMT="(/,T2,A,F10.2,A,F10.2,A)")&
               "Temperature  shell internal motion scaled to requested temperature:",&
               temp_shell_old," K  ->",md_ener%temp_shell," K"
       END IF
    ENDIF
  END SUBROUTINE scale_velocity_internal

! **************************************************************************************************
!> \brief Scale barostat velocities to get the desired temperature
!> \param md_env ...
!> \param md_ener ...
!> \param temp_expected ...
!> \param temp_tol ...
!> \param iw ...
!> \par History
!>     MI 02.2008
! **************************************************************************************************
  SUBROUTINE scale_velocity_baro(md_env, md_ener, temp_expected, temp_tol, iw)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      REAL(KIND=dp), INTENT(IN)                          :: temp_expected, temp_tol
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scale_velocity_baro', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nfree
      REAL(KIND=dp)                                      :: ekin_old, scale, temp_old
      TYPE(npt_info_type), POINTER                       :: npt( :, : )
      TYPE(simpar_type), POINTER                         :: simpar

    NULLIFY( npt, simpar)
    CALL get_md_env ( md_env, simpar = simpar, npt = npt)
    IF (ABS(temp_expected - md_ener%temp_baro/kelvin) > temp_tol) THEN
       scale = 0.0_dp
       IF (md_ener%temp_baro>0.0_dp) scale = SQRT((temp_expected/md_ener%temp_baro)*kelvin)
       ekin_old = md_ener%baro_kin
       temp_old = md_ener%temp_baro
       md_ener%baro_kin       = 0.0_dp
       md_ener%temp_baro  = 0.0_dp
      IF ( simpar%ensemble==npt_i_ensemble .OR. simpar%ensemble==npe_i_ensemble) THEN
         npt ( 1, 1 )%v = npt ( 1, 1 )%v*scale
         md_ener%baro_kin = 0.5_dp * npt ( 1, 1 )%v**2 * npt ( 1, 1 )%mass
       ELSE IF (simpar%ensemble==npt_f_ensemble .OR. simpar%ensemble==npe_f_ensemble) THEN
         md_ener%baro_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, 3
             npt(i,j)%v = npt(i,j)%v*scale
             md_ener%baro_kin = md_ener%baro_kin + 0.5_dp * npt(i,j)%v**2  * npt ( i, j )%mass
            END DO
         END DO
      END IF

      nfree = SIZE ( npt, 1 ) * SIZE ( npt, 2 )
      md_ener%temp_baro = 2.0_dp * md_ener%baro_kin / REAL(nfree,dp)*kelvin
      IF (iw>0) THEN
          WRITE (UNIT=iw,FMT="(/,T2,A,F10.2,A,F10.2,A)")&
               "Temperature  of barostat motion scaled to requested temperature:",&
               temp_old," K  ->",md_ener%temp_baro," K"
      END IF
   END IF
  END SUBROUTINE scale_velocity_baro

! **************************************************************************************************
!> \brief Perform all temperature manipulations during a QS MD run.
!> \param simpar ...
!> \param md_env ...
!> \param md_ener ...
!> \param force_env ...
!> \param logger ...
!> \par History
!>     Creation (15.09.2003,MK)
!>     adapted to force_env (05.10.2003,fawzi)
!>     Cleaned (09.2007) Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
  SUBROUTINE temperature_control(simpar, md_env, md_ener,force_env,logger)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(LEN=*), PARAMETER :: routineN = 'temperature_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys

    CALL timeset(routineN,handle)
    NULLIFY(subsys, para_env)
    IF(.NOT.(ASSOCIATED(simpar)))CALL cp__a("motion/md_vel_utils.F",1540)
    IF(.NOT.(ASSOCIATED(md_ener)))CALL cp__a("motion/md_vel_utils.F",1541)
    IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/md_vel_utils.F",1542)
    CALL force_env_get(force_env,subsys=subsys,para_env=para_env)
    iw = cp_print_key_unit_nr(logger,force_env%root_section,"MOTION%MD%PRINT%PROGRAM_RUN_INFO",&
         extension=".mdLog")

    ! Control the particle motion
    IF(simpar%do_thermal_region) THEN
       CALL scale_velocity_region(md_env, subsys, md_ener, simpar, iw)
    ELSE
      IF (simpar%temp_tol > 0.0_dp ) THEN
         CALL scale_velocity(subsys, md_ener, simpar%temp_ext, simpar%temp_tol, iw)
      END IF
    END IF
    ! Control the internal core-shell motion
    IF(simpar%temp_sh_tol > 0.0_dp) THEN
       CALL scale_velocity_internal(subsys, md_ener, simpar%temp_sh_ext, simpar%temp_sh_tol, iw)
    END IF
    ! Control cell motion
    SELECT CASE (simpar%ensemble)
    CASE( nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, &
            npt_f_ensemble, npt_i_ensemble, npe_f_ensemble, npe_i_ensemble)
       IF(simpar%temp_baro_tol > 0.0_dp) THEN
          CALL scale_velocity_baro(md_env, md_ener, simpar%temp_baro_ext, simpar%temp_baro_tol, iw)
       END IF
    END SELECT

    CALL cp_print_key_finished_output(iw,logger,force_env%root_section,&
         "MOTION%MD%PRINT%PROGRAM_RUN_INFO")
    CALL timestop(handle)
  END SUBROUTINE temperature_control

! **************************************************************************************************
!> \brief Set to 0 the velocity of the COM along MD runs, if required.
!> \param md_ener ...
!> \param force_env ...
!> \param md_section ...
!> \param logger ...
!> \par History
!>      Creation (29.04.2007,MI)
!>      Cleaned (09.2007) Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
  SUBROUTINE comvel_control(md_ener,force_env, md_section, logger)

      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(LEN=*), PARAMETER :: routineN = 'comvel_control', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: comvel_tol, temp_old, vel_com
      REAL(KIND=dp), DIMENSION(3)                        :: vcom_old
      TYPE(cp_subsys_type), POINTER                      :: subsys

    CALL timeset(routineN,handle)
    NULLIFY(subsys)
    IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/md_vel_utils.F",1600)
    CALL force_env_get(force_env,subsys=subsys)

    ! Print COMVEL and COM Position
    iw = cp_print_key_unit_nr(logger,md_section,"PRINT%CENTER_OF_MASS",extension=".mdLog")
    IF (iw>0) THEN
       WRITE (UNIT=iw,FMT="(/,T2,A,(T58,A3,F20.10))")&
            "Centre of mass motion (COM):","x =",md_ener%vcom(1),"y =",md_ener%vcom(2),"z =",md_ener%vcom(3)
    END IF
    CALL cp_print_key_finished_output(iw,logger,md_section,"PRINT%CENTER_OF_MASS")

    ! If requested rescale COMVEL
    CALL section_vals_val_get(md_section,"COMVEL_TOL",explicit=explicit)
    IF ( explicit ) THEN
       CALL section_vals_val_get(md_section,"COMVEL_TOL",r_val=comvel_tol)
       iw = cp_print_key_unit_nr(logger,md_section,"PRINT%PROGRAM_RUN_INFO",&
            extension=".mdLog")
       vel_com = SQRT(md_ener%vcom(1)**2+md_ener%vcom(2)**2+md_ener%vcom(3)**2)

       ! Subtract the velocity of the COM, if requested
       IF (vel_com > comvel_tol) THEN
          temp_old = md_ener%temp_part/kelvin
          vcom_old = md_ener%vcom
          CALL reset_vcom( subsys, md_ener, vsubtract=vcom_old)
          CALL scale_velocity(subsys, md_ener, temp_old, 0.0_dp, iw)
          IF (iw>0) THEN
             WRITE (UNIT=iw,FMT="(T2,'MD| ',A,3F16.10,A)") &
                  "Old VCOM = ",vcom_old(1:3)," a.u.",&
                  "New VCOM = ",md_ener%vcom(1:3)," a.u"
          END IF
       END IF
       CALL cp_print_key_finished_output(iw,logger,md_section,&
            "PRINT%PROGRAM_RUN_INFO")
    END IF

    CALL timestop(handle)
  END SUBROUTINE comvel_control

! **************************************************************************************************
!> \brief Set to 0 the angular velocity along MD runs, if required.
!> \param md_ener ...
!> \param force_env ...
!> \param md_section ...
!> \param logger ...
!> \par History
!>      Creation (10.2009) Teodoro Laino [tlaino]
! **************************************************************************************************
  SUBROUTINE angvel_control(md_ener, force_env, md_section, logger)

      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(LEN=*), PARAMETER :: routineN = 'angvel_control', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ifixd, imolecule_kind, iw, natoms
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: is_fixed
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: angvel_tol, rcom(3), temp_old, vang(3), &
                                                            vang_new(3)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles

    CALL timeset(routineN,handle)
    ! If requested rescale ANGVEL
    CALL section_vals_val_get(md_section,"ANGVEL_TOL",explicit=explicit)
    IF ( explicit ) THEN
       NULLIFY(subsys, cell)
       IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/md_vel_utils.F",1674)
       CALL force_env_get(force_env,subsys=subsys,cell=cell)

       IF (SUM(cell%perd(1:3)) == 0) THEN
          CALL section_vals_val_get(md_section,"ANGVEL_TOL",r_val=angvel_tol)
          iw = cp_print_key_unit_nr(logger,md_section,"PRINT%PROGRAM_RUN_INFO",&
               extension=".mdLog")

          CALL cp_subsys_get(subsys,molecule_kinds_new=molecule_kinds,&
               particles=particles)

          natoms = SIZE(particles%els)
          ! Build a list of all fixed atoms (if any)
          ALLOCATE (is_fixed(natoms))

          is_fixed = use_perd_none
          molecule_kind_set => molecule_kinds%els
          DO imolecule_kind=1,molecule_kinds%n_els
             molecule_kind => molecule_kind_set(imolecule_kind)
             CALL get_molecule_kind(molecule_kind=molecule_kind,fixd_list=fixd_list)
             IF (ASSOCIATED(fixd_list)) THEN
                DO ifixd=1,SIZE(fixd_list)
                   IF (.NOT.fixd_list(ifixd)%restraint%active) &
                        is_fixed(fixd_list(ifixd)%fixd) = fixd_list(ifixd)%itype
                END DO
             END IF
          END DO

          ! If requested and the system is not periodic, subtract the angular velocity
          CALL compute_rcom(particles%els,is_fixed,rcom)
          CALL compute_vang(particles%els,is_fixed,rcom,vang)
          IF (SQRT(DOT_PRODUCT(vang,vang))>angvel_tol) THEN
             CALL subtract_vang(particles%els,is_fixed,rcom,vang)

             ! Rescale velocities after removal
             temp_old = md_ener%temp_part/kelvin
             CALL scale_velocity(subsys, md_ener, temp_old, 0.0_dp, iw)
             CALL compute_vang(particles%els,is_fixed,rcom,vang_new)
             IF (iw>0) THEN
                WRITE (UNIT=iw,FMT="(T2,'MD| ',A,3F16.10,A)") &
                     "Old VANG = ",vang(1:3)," a.u.",&
                     "New VANG = ",vang_new(1:3)," a.u"
             END IF
          END IF

          DEALLOCATE (is_fixed)

          CALL cp_print_key_finished_output(iw,logger,md_section,&
               "PRINT%PROGRAM_RUN_INFO")
       END IF
    END IF

    CALL timestop(handle)
  END SUBROUTINE angvel_control

! **************************************************************************************************
!> \brief Initialize Velocities for MD runs
!> \param force_env ...
!> \param simpar ...
!> \param globenv ...
!> \param md_env ...
!> \param md_section ...
!> \param constraint_section ...
!> \param write_binary_restart_file ...
!> \par History
!>     Teodoro Laino - University of Zurich - 09.2007 [tlaino]
! **************************************************************************************************
  SUBROUTINE setup_velocities(force_env, simpar, globenv, md_env, md_section, &
                              constraint_section, write_binary_restart_file)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: md_section, constraint_section
      LOGICAL, INTENT(IN)                                :: write_binary_restart_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_velocities', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nconstraint, nconstraint_fixd
      LOGICAL                                            :: apply_cns0, shell_adiabatic, &
                                                            shell_present
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(section_vals_type), POINTER                   :: force_env_section, print_section, &
                                                            subsys_section

    CALL timeset(routineN,handle)

    NULLIFY (atomic_kinds,cell,para_env,subsys,molecule_kinds,core_particles,particles)
    NULLIFY (shell_particles,core_particle_set,particle_set,shell_particle_set)
    NULLIFY (force_env_section,print_section,subsys_section)

    print_section => section_vals_get_subs_vals(md_section,"PRINT")
    apply_cns0 = .FALSE.
    IF (simpar%constraint) THEN
       CALL section_vals_val_get(constraint_section,"CONSTRAINT_INIT",l_val=apply_cns0)
    END IF
    ! Always initialize velocities and possibly restart them
    CALL force_env_get(force_env, subsys=subsys, cell=cell, para_env=para_env,&
         force_env_section=force_env_section)
    subsys_section => section_vals_get_subs_vals(force_env_section,"SUBSYS")

    CALL cp_subsys_get(subsys,&
                       atomic_kinds=atomic_kinds,&
                       core_particles=core_particles,&
                       molecule_kinds_new=molecule_kinds,&
                       particles=particles,&
                       shell_particles=shell_particles)

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kinds%els,&
                             shell_present=shell_present,&
                             shell_adiabatic=shell_adiabatic)

    NULLIFY (core_particle_set)
    NULLIFY (particle_set)
    NULLIFY (shell_particle_set)
    particle_set => particles%els

    IF (shell_present.AND.shell_adiabatic) THEN
       ! Constraints are not yet implemented for core-shell models generally
       CALL get_molecule_kind_set(molecule_kind_set=molecule_kinds%els,&
                                  nconstraint=nconstraint,&
                                  nconstraint_fixd=nconstraint_fixd)
       IF(nconstraint - nconstraint_fixd /= 0)&
          CALL cp__b("motion/md_vel_utils.F",1807,"Only the fixed atom constraint is implemented for core-shell models")
!MK    CPPostcondition(.NOT.simpar%constraint,cp_failure_level,routineP,failure)
       IF(.NOT.(ASSOCIATED(shell_particles)))CALL cp__a("motion/md_vel_utils.F",1809)
       IF(.NOT.(ASSOCIATED(core_particles)))CALL cp__a("motion/md_vel_utils.F",1810)
       shell_particle_set => shell_particles%els
       core_particle_set  =>  core_particles%els
    END IF

    CALL initialize_velocities(simpar,particle_set, molecule_kinds=molecule_kinds,&
         force_env=force_env, globenv=globenv, md_env=md_env,&
         label="Velocities initialization",print_section=print_section, &
         subsys_section=subsys_section, shell_present=(shell_present.AND.shell_adiabatic), &
         shell_part=shell_particle_set, core_part=core_particle_set, force_rescaling=.FALSE., &
         para_env=para_env)

    ! Apply constraints if required and rescale velocities..
    IF (simpar%ensemble /= reftraj_ensemble) THEN
       IF (apply_cns0) THEN
          CALL force_env_calc_energy_force ( force_env, calc_force=.TRUE.)
          CALL force_env_shake(force_env,shake_tol=simpar%shake_tol,&
               log_unit=simpar%info_constraint,lagrange_mult=simpar%lagrange_multipliers,&
               dump_lm=simpar%dump_lm,compold=.TRUE.)
          CALL force_env_rattle(force_env,shake_tol=simpar%shake_tol,&
               log_unit=simpar%info_constraint,lagrange_mult=simpar%lagrange_multipliers,&
               dump_lm=simpar%dump_lm,reset=.TRUE.)
          IF (simpar%do_respa)THEN
             CALL force_env_calc_energy_force (force_env%sub_force_env(1)%force_env,&
                  calc_force=.TRUE.)
             CALL force_env_shake(force_env%sub_force_env(1)%force_env,&
                  shake_tol=simpar%shake_tol,log_unit=simpar%info_constraint,&
                  lagrange_mult=simpar%lagrange_multipliers,dump_lm=simpar%dump_lm,compold=.TRUE.)
             CALL force_env_rattle(force_env%sub_force_env(1)%force_env,&
                  shake_tol=simpar%shake_tol,log_unit=simpar%info_constraint,&
                  lagrange_mult=simpar%lagrange_multipliers,dump_lm=simpar%dump_lm,reset=.TRUE.)
          END IF
          ! Reinitialize velocities rescaling properly after rattle
          subsys_section => section_vals_get_subs_vals(force_env_section,"SUBSYS")
          CALL update_subsys(subsys_section,force_env,.FALSE.,write_binary_restart_file)

          CALL initialize_velocities(simpar,particle_set, molecule_kinds=molecule_kinds,&
               force_env=force_env, globenv=globenv, md_env=md_env,&
               label="Re-Initializing velocities after applying constraints",print_section=print_section, &
               subsys_section=subsys_section, shell_present=(shell_present.AND.shell_adiabatic), &
               shell_part=shell_particle_set, core_part=core_particle_set, force_rescaling=.TRUE., &
               para_env=para_env)
       END IF
    END IF

    ! Perform setup for a cascade run
    CALL initialize_cascade(simpar,particle_set,molecule_kinds,md_section)

    CALL timestop(handle)

  END SUBROUTINE setup_velocities

! **************************************************************************************************
!> \brief   Perform the initialization for a cascade run
!> \param simpar ...
!> \param particle_set ...
!> \param molecule_kinds ...
!> \param md_section ...
!> \date    05.02.2012
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
  SUBROUTINE initialize_cascade(simpar,particle_set,molecule_kinds,md_section)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(section_vals_type), POINTER                   :: md_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_cascade', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2*default_string_length)             :: line
      INTEGER                                            :: handle, iatom, ifixd, imolecule_kind, &
                                                            iparticle, iw, natom, nparticle
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_index, is_fixed
      LOGICAL                                            :: init_cascade, is_ok, no_read_error
      REAL(KIND=dp)                                      :: ecom, ekin, energy, norm, temp, &
                                                            temperature
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: matom, weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vatom
      REAL(KIND=dp), DIMENSION(3)                        :: vcom
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_sll_val_type), POINTER                     :: atom_list
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(section_vals_type), POINTER                   :: atom_list_section, cascade_section, &
                                                            print_section
      TYPE(val_type), POINTER                            :: val

    CALL timeset(routineN,handle)


    NULLIFY (atom_list)
    NULLIFY (atom_list_section)
    NULLIFY (atomic_kind)
    NULLIFY (cascade_section)
    NULLIFY (fixd_list)
    NULLIFY (molecule_kind)
    NULLIFY (molecule_kind_set)
    NULLIFY (logger)
    NULLIFY (val)

    logger => cp_get_default_logger()
    print_section => section_vals_get_subs_vals(md_section,"PRINT")
    iw = cp_print_key_unit_nr(logger,print_section,"PROGRAM_RUN_INFO",extension=".log")

    cascade_section => section_vals_get_subs_vals(md_section,"CASCADE")
    CALL section_vals_val_get(cascade_section,"_SECTION_PARAMETERS_",l_val=init_cascade)

    nparticle = SIZE(particle_set)

    IF (init_cascade) THEN

       CALL section_vals_val_get(cascade_section,"ENERGY",r_val=energy)
       IF(energy < 0.0_dp)&
          CALL cp__b("motion/md_vel_utils.F",1928,"Error occurred reading &CASCADE section: Negative energy found")

       IF (iw > 0) THEN
          ekin = cp_unit_from_cp2k(energy,"keV")
          WRITE (UNIT=iw,FMT="(T2,A,T61,F20.6)")&
           "CASCADE| Energy [keV]",ekin
          WRITE (UNIT=iw,FMT="(T2,A)")&
           "CASCADE|"
       END IF

       ! Read the atomic velocities given in the input file
       atom_list_section => section_vals_get_subs_vals(cascade_section,"ATOM_LIST")
       CALL section_vals_val_get(atom_list_section,"_DEFAULT_KEYWORD_",n_rep_val=natom)
       CALL section_vals_list_get(atom_list_section,"_DEFAULT_KEYWORD_",list=atom_list)
       IF(natom <= 0)&
          CALL cp__b("motion/md_vel_utils.F",1943,"Error occurred reading &CASCADE section: No atom list found")

       IF (iw > 0) THEN
          WRITE (UNIT=iw,FMT="(T2,A,T11,A,3(11X,A),9X,A)")&
           "CASCADE| ","Atom index","v(x)","v(y)","v(z)","weight"
       END IF

       ALLOCATE (atom_index(natom))
       ALLOCATE (matom(natom))
       ALLOCATE (vatom(3,natom))
       ALLOCATE (weight(natom))

       DO iatom=1,natom
          is_ok = cp_sll_val_next(atom_list,val)
          CALL val_get(val,c_val=line)
          ! Read atomic index, velocity vector, and weight
          no_read_error = .FALSE.
          READ (UNIT=line,FMT=*,ERR=999) atom_index(iatom),vatom(1:3,iatom),weight(iatom)
          no_read_error = .TRUE.
999       IF(.NOT.no_read_error)&
             CALL cp__b("motion/md_vel_utils.F",1963,"Error occurred reading &CASCADE section. Last line read <"//TRIM(line)//">")
          IF((atom_index(iatom) <= 0).OR.((atom_index(iatom) > nparticle)))&
             CALL cp__b("motion/md_vel_utils.F",1965,"Error occurred reading &CASCADE section: Invalid atom index found")
          IF(weight(iatom) < 0.0_dp)&
             CALL cp__b("motion/md_vel_utils.F",1967,"Error occurred reading &CASCADE section: Negative weight found")
          IF (iw > 0) THEN
             WRITE (UNIT=iw,FMT="(T2,A,I10,4(1X,F14.6))")&
              "CASCADE| ",atom_index(iatom),vatom(1:3,iatom),weight(iatom)
          END IF
       END DO

       ! Normalise velocities and weights
       norm = 0.0_dp
       DO iatom=1,natom
          iparticle = atom_index(iatom)
          IF(particle_set(iparticle)%shell_index /= 0)&
             CALL cp__w("motion/md_vel_utils.F",1979,"Warning: The primary knock-on atom is a core-shell atom")
          atomic_kind => particle_set(iparticle)%atomic_kind
          CALL get_atomic_kind(atomic_kind=atomic_kind,mass=matom(iatom))
          norm = norm + matom(iatom)*weight(iatom)
       END DO
       weight(:) = matom(:)*weight(:)*energy/norm
       DO iatom=1,natom
          norm = SQRT(DOT_PRODUCT(vatom(1:3,iatom),vatom(1:3,iatom)))
          vatom(1:3,iatom) = vatom(1:3,iatom)/norm
       END DO

       IF (iw > 0) THEN
          WRITE (UNIT=iw,FMT="(T2,A)")&
           "CASCADE|",&
           "CASCADE| Normalised velocities and additional kinetic energy [keV]",&
           "CASCADE|"
          WRITE (UNIT=iw,FMT="(T2,A,T11,A,3(11X,A),9X,A)")&
           "CASCADE| ","Atom index","v(x)","v(y)","v(z)","E(kin)"
          DO iatom=1,natom
             ekin = cp_unit_from_cp2k(weight(iatom),"keV")
             WRITE (UNIT=iw,FMT="(T2,A,I10,4(1X,F14.6))")&
              "CASCADE| ",atom_index(iatom),vatom(1:3,iatom),ekin
          END DO
       END IF

       ! Apply velocity modifications
       DO iatom=1,natom
          iparticle = atom_index(iatom)
          particle_set(iparticle)%v(:) = particle_set(iparticle)%v(:) +&
                                         SQRT(2.0_dp*weight(iatom)/matom(iatom))*vatom(1:3,iatom)
       END DO

       DEALLOCATE (atom_index)
       DEALLOCATE (matom)
       DEALLOCATE (vatom)
       DEALLOCATE (weight)

       IF (iw > 0) THEN
          ! Build a list of all fixed atoms (if any)
          ALLOCATE (is_fixed(nparticle))
          is_fixed = use_perd_none
          molecule_kind_set => molecule_kinds%els
          DO imolecule_kind=1,molecule_kinds%n_els
             molecule_kind => molecule_kind_set(imolecule_kind)
             CALL get_molecule_kind(molecule_kind=molecule_kind,fixd_list=fixd_list)
             IF (ASSOCIATED(fixd_list)) THEN
                DO ifixd=1,SIZE(fixd_list)
                   IF (.NOT.fixd_list(ifixd)%restraint%active) is_fixed(fixd_list(ifixd)%fixd) = fixd_list(ifixd)%itype
                END DO
             END IF
          END DO
          ! Compute vcom, ecom and ekin for printout
          CALL compute_vcom(particle_set,is_fixed,vcom,ecom)
          ekin = compute_ekin(particle_set) - ecom
          IF (simpar%nfree == 0) THEN
             IF(.NOT.(ekin == 0.0_dp))CALL cp__a("motion/md_vel_utils.F",2034)
             temp = 0.0_dp
          ELSE
             temp = 2.0_dp*ekin/REAL(simpar%nfree,KIND=dp)
          END IF
          temperature = cp_unit_from_cp2k(temp,"K")
          WRITE (UNIT=iw,FMT="(T2,A)")&
           "CASCADE|"
          WRITE (UNIT=iw,FMT="(T2,A,T61,F18.2,A2)")&
           "CASCADE| Temperature after cascade initialization",temperature," K"
          WRITE (UNIT=iw,FMT="(T2,A,T30,3(1X,ES16.8),/)")&
           "CASCADE| COM velocity: ",vcom(1:3)
!MK          ! compute and log rcom and vang if not periodic
!MK          CALL force_env_get(force_env,cell=cell)
!MK          IF (SUM(cell%perd(1:3)) == 0) THEN
!MK             CALL compute_rcom(particle_set,is_fixed,rcom)
!MK             CALL compute_vang(particle_set,is_fixed,rcom,vang)
!MK             WRITE (iw, '( A, T21, F20.12 , F20.12 , F20.12 )' ) ' COM position:',rcom(1:3)
!MK             WRITE (iw, '( A, T21, F20.12 , F20.12 , F20.12 )' ) ' Angular velocity:',vang(1:3)
!MK          END IF
          DEALLOCATE (is_fixed)
       END IF

    END IF

    CALL cp_print_key_finished_output(iw,logger,print_section,"PROGRAM_RUN_INFO")

    CALL timestop(handle)

  END SUBROUTINE initialize_cascade

END MODULE md_vel_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handling of the Wiener process currently employed in turn of the
!>      Langevin dynamics.
!> \par History
!>      none
!> \author Matthias Krack (05.07.2005)
! **************************************************************************************************
MODULE wiener_process

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type,&
                                              need_per_atom_wiener_process
   USE metadynamics_types,              ONLY: meta_env_type
   USE parallel_rng_types,              ONLY: GAUSSIAN,&
                                              create_rng_stream,&
                                              next_rng_seed,&
                                              read_rng_stream,&
                                              rng_record_length
   USE particle_list_types,             ONLY: particle_list_type
   USE simpar_types,                    ONLY: simpar_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters in this module
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'wiener_process'

   ! Public subroutines
   PUBLIC :: create_wiener_process, create_wiener_process_cv

CONTAINS

! **************************************************************************************************
!> \brief Create a Wiener process for Langevin dynamics and initialize an
!>      independent random number generator for each atom in all force
!>      environment and all the subsystems/fragments therein.
!> \param md_env ...
!> \par History
!>      Creation (06.07.2005,MK)
! **************************************************************************************************
   SUBROUTINE create_wiener_process(md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_wiener_process', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=40)                                  :: name
      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle, &
                                                            nparticle_kind, nparticle_local
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: seed
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section, &
                                                            work_section
      TYPE(simpar_type), POINTER                         :: simpar

      NULLIFY (work_section, force_env)
      IF(.NOT.(ASSOCIATED(md_env)))CALL cp__a("motion/wiener_process.F",85)

      CALL get_md_env(md_env=md_env, force_env=force_env, para_env=para_env, &
                      simpar=simpar)

      ![NB] shouldn't the calling process know if it's needed
      IF (need_per_atom_wiener_process(md_env)) THEN

         ! Load initial seed (not needed for a restart)

         initial_seed = next_rng_seed()

         CALL force_env_get(force_env, force_env_section=force_env_section, &
                            subsys=subsys)

         subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")

         CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                            particles=particles)

         nparticle_kind = atomic_kinds%n_els
         nparticle = particles%n_els

         ! Allocate the (local) data structures for the Wiener process

         ALLOCATE (local_particles%local_particle_set(nparticle_kind))

         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            ALLOCATE (local_particles%local_particle_set(iparticle_kind)%rng(nparticle_local))
            DO iparticle_local = 1, nparticle_local
               NULLIFY (local_particles%local_particle_set(iparticle_kind)%rng(iparticle_local)%stream)
            END DO
         END DO

         ! Each process generates all seeds. The seed generation should be
         ! quite fast and in this way a broadcast is avoided.

         ALLOCATE (seed(3, 2, nparticle))

         seed(:, :, 1) = initial_seed
         DO iparticle = 2, nparticle
            seed(:, :, iparticle) = next_rng_seed(seed(:, :, iparticle-1))
         END DO

         ! Update initial seed

         initial_seed = next_rng_seed(seed(:, :, nparticle))

         ! Create a random number stream (Wiener process) for each particle

         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               WRITE (UNIT=name, FMT="(A,I8)") "Wiener process for particle", iparticle
               CALL compress(name)
               CALL create_rng_stream(rng_stream=local_particles%local_particle_set(iparticle_kind)% &
                                      rng(iparticle_local)%stream, name=name, distribution_type=GAUSSIAN, &
                                      extended_precision=.TRUE., seed=seed(:, :, iparticle))
            END DO
         END DO

         DEALLOCATE (seed)

         ! Possibly restart Wiener process
         NULLIFY (work_section)
         work_section => section_vals_get_subs_vals(section_vals=subsys_section, &
                                                    subsection_name="RNG_INIT")
         CALL init_local_particle_set(distribution_1d=local_particles, &
                                      nparticle_kind=nparticle_kind, &
                                      work_section=work_section)
      END IF

   END SUBROUTINE create_wiener_process

! **************************************************************************************************
!> \brief Helper routine for create_wiener_process.
!> \param distribution_1d ...
!> \param nparticle_kind ...
!> \param work_section ...
!> \par History
!>      01.2014 moved from distribution_1d_types (Ole Schuett)
! **************************************************************************************************
   SUBROUTINE init_local_particle_set(distribution_1d, nparticle_kind, &
                                      work_section)

      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      INTEGER, INTENT(in)                                :: nparticle_kind
      TYPE(section_vals_type), POINTER                   :: work_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_local_particle_set', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_local
      LOGICAL                                            :: explicit

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(distribution_1d)))CALL cp__a("motion/wiener_process.F",186)

      IF (ASSOCIATED(work_section)) THEN
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            DO iparticle_kind = 1, nparticle_kind
               nparticle_local = distribution_1d%n_el(iparticle_kind)
               DO iparticle_local = 1, nparticle_local
                  iparticle = distribution_1d%list(iparticle_kind)%array(iparticle_local)
                  IF (iparticle == distribution_1d%list(iparticle_kind)%array(iparticle_local)) THEN
                     CALL section_vals_val_get(section_vals=work_section, &
                                               keyword_name="_DEFAULT_KEYWORD_", &
                                               i_rep_val=iparticle, &
                                               c_val=rng_record)
                     CALL read_rng_stream(rng_stream=distribution_1d% &
                                          local_particle_set(iparticle_kind)% &
                                          rng(iparticle_local)%stream, &
                                          rng_record=rng_record)
                  END IF
               END DO
            END DO
         END IF
      END IF

   END SUBROUTINE init_local_particle_set

! **************************************************************************************************
!> \brief Create a Wiener process for Langevin dynamics used for
!>        metadynamics and initialize an
!>        independent random number generator for each COLVAR.
!> \param meta_env ...
!> \date   01.2009
!> \author Fabio Sterpone
!>
! **************************************************************************************************
   SUBROUTINE create_wiener_process_cv(meta_env)

      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_wiener_process_cv', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=40)                                  :: name
      INTEGER                                            :: i_c
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: seed
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed

      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      initial_seed = next_rng_seed()

      DO i_c = 1, meta_env%n_colvar
         NULLIFY (meta_env%rng(i_c)%stream)
      END DO

      ! Each process generates all seeds. The seed generation should be
      ! quite fast and in this way a broadcast is avoided.

      ALLOCATE (seed(3, 2, meta_env%n_colvar))

      seed(:, :, 1) = initial_seed
      DO i_c = 2, meta_env%n_colvar
         seed(:, :, i_c) = next_rng_seed(seed(:, :, i_c-1))
      END DO

      ! Update initial seed
      initial_seed = next_rng_seed(seed(:, :, meta_env%n_colvar))

      ! Create a random number stream (Wiener process) for each particle
      DO i_c = 1, meta_env%n_colvar
         WRITE (UNIT=name, FMT="(A,I8)") "Wiener process for COLVAR", i_c
         CALL compress(name)
         CALL create_rng_stream(rng_stream=meta_env%rng(i_c)%stream, name=name, distribution_type=GAUSSIAN, &
                                extended_precision=.TRUE., seed=seed(:, :, i_c))
      END DO
      DEALLOCATE (seed)

   END SUBROUTINE create_wiener_process_cv

END MODULE wiener_process
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Methods for storing MD parameters type
!> \author CJM
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
!>         reorganization of the original routines/modules
! **************************************************************************************************
MODULE simpar_methods
   USE bibliography,                    ONLY: Evans1983,&
                                              Kuhne2007,&
                                              Minary2003,&
                                              Ricci2003,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_constants,                 ONLY: &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nvt_ensemble, reftraj_ensemble
   USE input_cp2k_md,                   ONLY: create_md_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'simpar_methods'
   PUBLIC :: read_md_section

CONTAINS

! **************************************************************************************************
!> \brief Reads the MD section and setup the simulation parameters type
!> \param simpar ...
!> \param motion_section ...
!> \param md_section ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE read_md_section(simpar, motion_section, md_section)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(section_vals_type), POINTER                   :: motion_section, md_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_md_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: iprint, iw
      REAL(kind=dp)                                      :: tmp_r1, tmp_r2, tmp_r3
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger, print_key, enum, keyword, section)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, md_section, "PRINT%PROGRAM_RUN_INFO", extension=".log")

      CALL read_md_low(simpar, motion_section, md_section)
      IF (iw > 0) WRITE (iw, *)

      ! Begin setup Langevin dynamics
      IF (simpar%ensemble == langevin_ensemble) THEN
         CALL cite_reference(Ricci2003)
         IF (simpar%noisy_gamma > 0.0_dp) CALL cite_reference(Kuhne2007)
         ! Normalization factor using a normal Gaussian random number distribution
         simpar%var_w = 2.0_dp*simpar%temp_ext*simpar%dt*(simpar%gamma+simpar%noisy_gamma)
         IF (iw > 0) THEN
            tmp_r1 = cp_unit_from_cp2k(simpar%gamma, "fs^-1")
            tmp_r2 = cp_unit_from_cp2k(simpar%noisy_gamma, "fs^-1")
            tmp_r3 = cp_unit_from_cp2k(simpar%shadow_gamma, "fs^-1")
            WRITE (UNIT=iw, FMT="(T2,A,T71,ES10.3)") &
               "LD| Gamma [1/fs] ", tmp_r1, &
               "LD| Noisy Gamma [1/fs]", tmp_r2, &
               "LD| Shadow Gamma [1/fs]", tmp_r3, &
               "LD| Variance [a.u.]", simpar%var_w
         END IF
      END IF

      ! Create section for output enumeration infos
      CALL create_md_section(section)
      keyword => section_get_keyword(section, "ENSEMBLE")
      CALL keyword_get(keyword, enum=enum)

      !..write some information to output
      IF (iw > 0) THEN
         WRITE (iw, '( A )') ' MD| Molecular Dynamics Protocol '
         WRITE (iw, '( A,T61,A20)') ' MD| Ensemble Type ', &
            ADJUSTR(TRIM(enum_i2c(enum, simpar%ensemble)))
         WRITE (iw, '( A,T71,I10 )') ' MD| Number of Time Steps ', &
            simpar%nsteps
         IF (simpar%variable_dt) THEN
            WRITE (iw, '( A )') ' MD| Variable Time Step is activated '
            tmp_r1 = cp_unit_from_cp2k(simpar%dt, "fs")
            WRITE (iw, '( A,A2,A,T71,F10.2 )') ' MD| Max. Time Step [', 'fs', '] ', tmp_r1
            tmp_r1 = cp_unit_from_cp2k(simpar%dr_tol, "angstrom")
            WRITE (iw, '( A, T71, F10.4 )') ' MD| Max. atomic displacement permitted [A] ', tmp_r1
         ELSE
            tmp_r1 = cp_unit_from_cp2k(simpar%dt, "fs")
            WRITE (iw, '( A,A2,A,T71,F10.2 )') ' MD| Time Step [', 'fs', '] ', tmp_r1
         END IF
         tmp_r1 = cp_unit_from_cp2k(simpar%temp_ext, "K")
         WRITE (iw, '( A,T71,F10.2 )') ' MD| Temperature [K] ', tmp_r1
         tmp_r1 = cp_unit_from_cp2k(simpar%temp_tol, "K")
         WRITE (iw, '( A,T71,F10.2 )') ' MD| Temperature tolerance [K] ', tmp_r1

         IF (simpar%annealing) &
            WRITE (iw, '( A,T71,F10.2 )') ' MD| Annealing ion factor      ', &
            simpar%f_annealing
         IF ((simpar%ensemble == npe_f_ensemble .OR. simpar%ensemble == npe_i_ensemble) &
             .AND. simpar%annealing_cell) &
            WRITE (iw, '( A,T71,F10.2 )') ' MD| Annealing cell factor      ', &
            simpar%f_annealing_cell
         IF (simpar%ensemble == npt_i_ensemble .OR. &
             simpar%ensemble == npt_f_ensemble .OR. &
             simpar%ensemble == npe_i_ensemble .OR. &
             simpar%ensemble == npe_f_ensemble) THEN
            tmp_r1 = cp_unit_from_cp2k(simpar%p_ext, "bar")
            WRITE (iw, '( A,A3,A, T71, F10.2 )') &
               ' MD| Pressure [', 'Bar', '] ', tmp_r1
            tmp_r1 = cp_unit_from_cp2k(simpar%tau_cell, "fs")
            WRITE (iw, '( A,A4,A, T71, F10.2 )') &
               ' MD| Barostat time constant [', 'fs', '] ', tmp_r1
         END IF
         IF (simpar%ensemble == isokin_ensemble) THEN
            CALL cite_reference(Evans1983)
            CALL cite_reference(Minary2003)
            WRITE (iw, '( A )') ' MD| Simulation in the isokinetic ensemble'
         END IF
         IF (simpar%constraint) THEN
            WRITE (iw, '( A )') ' MD| Constraints activated '
            WRITE (iw, '( A,T71,G10.4 )') ' MD| Tolerance for shake ', &
               simpar%shake_tol
         END IF

         print_key => section_vals_get_subs_vals(motion_section, "MD%PRINT%PROGRAM_RUN_INFO")
         CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
         WRITE (iw, '( A,T63,i10,A )') ' MD| Print MD information every', iprint, ' step(s)'
         WRITE (iw, '( A,T20,A,T71,A10 )') ' MD| File type', 'Print frequency[steps]', 'File names'

         print_key => section_vals_get_subs_vals(motion_section, "PRINT%TRAJECTORY")
         CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   extension=".xyz", middle_name="pos", my_local=.FALSE.)
         WRITE (iw, '( A,T20,i10,T31,A50 )') ' MD| Coordinates', iprint, &
            ADJUSTR(TRIM(filename))

         IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
             simpar%ensemble == nph_uniaxial_damped_ensemble .OR. &
             simpar%ensemble == npt_i_ensemble .OR. &
             simpar%ensemble == npt_f_ensemble .OR. &
             simpar%ensemble == npe_i_ensemble .OR. &
             simpar%ensemble == npe_f_ensemble) THEN

            print_key => section_vals_get_subs_vals(motion_section, "PRINT%CELL")
            CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension=".cell", my_local=.FALSE.)
            WRITE (iw, '( A,T20,i10,T31,A50 )') ' MD| Simulation Cell', iprint, &
               ADJUSTR(TRIM(filename))
         END IF

         print_key => section_vals_get_subs_vals(motion_section, "PRINT%VELOCITIES")
         CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   extension=".xyz", middle_name="vel", my_local=.FALSE.)
         WRITE (iw, '( A,T20,i10,T31,A50 )') ' MD| Velocities', iprint, &
            ADJUSTR(TRIM(filename))

         print_key => section_vals_get_subs_vals(motion_section, "MD%PRINT%ENERGY")
         CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   extension=".ener", my_local=.FALSE.)
         WRITE (iw, '( A,T20,i10,T31,A50 )') ' MD| Energies', iprint, &
            ADJUSTR(TRIM(filename))

         print_key => section_vals_get_subs_vals(motion_section, "PRINT%RESTART")
         CALL section_vals_val_get(print_key, "EACH%MD", i_val=iprint)
         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   extension=".restart", my_local=.FALSE.)
         WRITE (iw, '( A,T20,i10,T31,A50 )') ' MD| Dump', iprint, &
            ADJUSTR(TRIM(filename))

         WRITE (iw, *)
         IF (simpar%ensemble == nph_uniaxial_ensemble .OR. &
             simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
            WRITE (iw, '( A )') ' SHOCK| Uniaxial Shock Parameters: '
            tmp_r1 = cp_unit_from_cp2k(simpar%v_shock, "m*s^-1")
            WRITE (iw, '( A,A4,A,T71,F10.4 )') &
               ' SHOCK| Shock Velocity [', 'm/s', '] ', tmp_r1
            tmp_r1 = cp_unit_from_cp2k(simpar%gamma_nph, "fs^-1")
            WRITE (iw, '( A,A4,A,T71,F10.4 )') &
               ' SHOCK| Damping Coefficient [', '1/fs', '] ', tmp_r1
            tmp_r1 = cp_unit_from_cp2k(simpar%p0, "bar")
            WRITE (iw, '( A,A3,A, T71, F10.2 )') &
               ' SHOCK| Pressure [', 'Bar', '] ', tmp_r1
            WRITE (iw, '( A,A4,A, T71, E10.4 )') &
               ' SHOCK| Barostat Mass [', 'au', '] ', simpar%cmass
         END IF
         ! Print warning for temp_tol
         IF (simpar%temp_tol > 0.0_dp) THEN
            CALL cp_warn(cp__l("motion/simpar_methods.F",223), &
                         "A temperature tolerance (TEMP_TOL) is used during the MD. "// &
                         "Due to the velocity rescaling algorithm jumps may appear in the conserved quantity.")
         END IF
         ! Print warning for annealing
         IF (simpar%annealing) THEN
            IF (simpar%ensemble == nvt_ensemble .OR. &
                simpar%ensemble == npt_i_ensemble .OR. &
                simpar%ensemble == npt_f_ensemble) THEN
               CALL cp_abort(cp__l("motion/simpar_methods.F",232), &
                             "Annealing of the ions has been required "// &
                             "even if the thermostat is active (nvt or npt_i or npt_f) "// &
                             "These two methods to control the temperature act one against the other.")
            END IF
         END IF
         ! Print warning for variable time step
         IF (simpar%variable_dt) THEN
            IF (simpar%ensemble == langevin_ensemble .OR. &
                simpar%ensemble == reftraj_ensemble .OR. &
                simpar%do_respa) THEN
               CALL cp_warn( &
                  cp__l("motion/simpar_methods.F",244), &
                  "The variable timestep  has been required, however "// &
                  "this option is not available either with the Langevin ensemble or with the multiple timestep schme. "// &
                  "The run will proceed with constant timestep, as read from input.")
            END IF
         END IF
      END IF
      CALL section_release(section)
      CALL cp_print_key_finished_output(iw, logger, md_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE read_md_section

! **************************************************************************************************
!> \brief Low Level: Parses the MD input section
!> \param simpar ...
!> \param motion_section ...
!> \param md_section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_md_low(simpar, motion_section, md_section)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(section_vals_type), POINTER                   :: motion_section, md_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_md_low', routineP = moduleN//':'//routineN

      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: tmp_section

      NULLIFY (tmp_section)
      CALL section_vals_val_get(md_section, "ENSEMBLE", i_val=simpar%ensemble)
      CALL section_vals_val_get(md_section, "STEPS", i_val=simpar%nsteps)
      CALL section_vals_val_get(md_section, "TEMPERATURE", r_val=simpar%temp_ext)
      CALL section_vals_val_get(md_section, "TEMP_TOL", r_val=simpar%temp_tol)
      CALL section_vals_val_get(md_section, "ANGVEL_ZERO", l_val=simpar%angvel_zero)
      CALL section_vals_val_get(md_section, "TEMP_KIND", l_val=simpar%temperature_per_kind)
      CALL section_vals_val_get(md_section, "SCALE_TEMP_KIND", l_val=simpar%scale_temperature_per_kind)
      CALL section_vals_val_get(md_section, "ANNEALING", r_val=simpar%f_annealing, explicit=simpar%annealing)
      CALL section_vals_val_get(md_section, "ANNEALING_CELL", r_val=simpar%f_annealing_cell, &
                                explicit=simpar%annealing_cell)
      CALL section_vals_val_get(md_section, "DISPLACEMENT_TOL", r_val=simpar%dr_tol, &
                                explicit=simpar%variable_dt)
      CALL section_vals_val_get(md_section, "TIMESTEP", r_val=simpar%dt)
      ! Initialize dt_fact to 1.0
      simpar%dt_fact = 1.0_dp

      IF (simpar%ensemble == langevin_ensemble) THEN
         CALL section_vals_val_get(md_section, "LANGEVIN%GAMMA", r_val=simpar%gamma)
         CALL section_vals_val_get(md_section, "LANGEVIN%NOISY_GAMMA", r_val=simpar%noisy_gamma)
         CALL section_vals_val_get(md_section, "LANGEVIN%SHADOW_GAMMA", r_val=simpar%shadow_gamma)
      END IF

      tmp_section => section_vals_get_subs_vals(motion_section, "CONSTRAINT")
      CALL section_vals_get(tmp_section, explicit=simpar%constraint)
      IF (simpar%constraint) THEN
         CALL section_vals_val_get(tmp_section, "SHAKE_TOLERANCE", r_val=simpar%shake_tol)
         IF (simpar%shake_tol <= EPSILON(0.0_dp)*1000.0_dp) &
            CALL cp_warn(cp__l("motion/simpar_methods.F",301), &
                         "Shake tolerance lower than 1000*EPSILON, where EPSILON is the machine precision. "// &
                         "This may lead to numerical problems. Setting up shake_tol to 1000*EPSILON!")
         simpar%shake_tol = MAX(EPSILON(0.0_dp)*1000.0_dp, simpar%shake_tol)

         CALL section_vals_val_get(tmp_section, "ROLL_TOLERANCE", r_val=simpar%roll_tol)
         IF (simpar%roll_tol <= EPSILON(0.0_dp)*1000.0_dp) &
            CALL cp_warn(cp__l("motion/simpar_methods.F",308), &
                         "Roll tolerance lower than 1000*EPSILON, where EPSILON is the machine precision. "// &
                         "This may lead to numerical problems. Setting up roll_tol to 1000*EPSILON!")
         simpar%roll_tol = MAX(EPSILON(0.0_dp)*1000.0_dp, simpar%roll_tol)
      END IF

      IF (simpar%ensemble == nph_uniaxial_ensemble .OR. simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         tmp_section => section_vals_get_subs_vals(md_section, "MSST")
         CALL section_vals_val_get(tmp_section, "PRESSURE", r_val=simpar%p0)
         CALL section_vals_val_get(tmp_section, "ENERGY", r_val=simpar%e0)
         CALL section_vals_val_get(tmp_section, "VOLUME", r_val=simpar%v0)
         CALL section_vals_val_get(tmp_section, "GAMMA", r_val=simpar%gamma_nph)
         IF (simpar%gamma_nph /= 0.0_dp) simpar%ensemble = nph_uniaxial_damped_ensemble
         CALL section_vals_val_get(tmp_section, "CMASS", r_val=simpar%cmass)
         CALL section_vals_val_get(tmp_section, "VSHOCK", r_val=simpar%v_shock)
      END IF

      SELECT CASE (simpar%ensemble)
      CASE (nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, &
            npt_f_ensemble, npt_i_ensemble, npe_f_ensemble, npe_i_ensemble)
         tmp_section => section_vals_get_subs_vals(md_section, "BAROSTAT")
         CALL section_vals_val_get(tmp_section, "PRESSURE", r_val=simpar%p_ext)
         CALL section_vals_val_get(tmp_section, "TIMECON", r_val=simpar%tau_cell)
      END SELECT

      ! RESPA
      tmp_section => section_vals_get_subs_vals(md_section, "RESPA")
      CALL section_vals_get(tmp_section, explicit=simpar%do_respa)
      CALL section_vals_val_get(tmp_section, "FREQUENCY", i_val=simpar%n_time_steps)
      simpar%multi_time_switch = simpar%do_respa

      ! CORE-SHELL MODEL
      tmp_section => section_vals_get_subs_vals(md_section, "SHELL")
      CALL section_vals_val_get(tmp_section, "TEMPERATURE", r_val=simpar%temp_sh_ext)
      CALL section_vals_val_get(tmp_section, "TEMP_TOL", r_val=simpar%temp_sh_tol)

      CALL section_vals_val_get(tmp_section, "DISPLACEMENT_SHELL_TOL", r_val=simpar%dsc_tol, &
                                explicit=explicit)
      simpar%variable_dt = simpar%variable_dt .OR. explicit
      ! ADIABATIC DYNAMICS
      tmp_section => section_vals_get_subs_vals(md_section, "ADIABATIC_DYNAMICS")
      CALL section_vals_val_get(tmp_section, "TEMP_FAST", r_val=simpar%temp_fast)
      CALL section_vals_val_get(tmp_section, "TEMP_SLOW", r_val=simpar%temp_slow)
      CALL section_vals_val_get(tmp_section, "TEMP_TOL_FAST", r_val=simpar%temp_tol_fast)
      CALL section_vals_val_get(tmp_section, "TEMP_TOL_SLOW", r_val=simpar%temp_tol_slow)
      CALL section_vals_val_get(tmp_section, "N_RESP_FAST", i_val=simpar%n_resp_fast)

      ! VELOCITY SOFTENING
      tmp_section => section_vals_get_subs_vals(md_section, "VELOCITY_SOFTENING")
      CALL section_vals_val_get(tmp_section, "STEPS", i_val=simpar%soften_nsteps)
      CALL section_vals_val_get(tmp_section, "ALPHA", r_val=simpar%soften_alpha)
      CALL section_vals_val_get(tmp_section, "DELTA", r_val=simpar%soften_delta)
   END SUBROUTINE read_md_low

END MODULE simpar_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines needed for EMD
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE rt_propagation_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              rtp_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_binary_read, cp_dbcsr_checksum, cp_dbcsr_copy, &
        cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_desymmetrize, cp_dbcsr_distribution, &
        cp_dbcsr_filter, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: use_restart_wfn,&
                                              use_rt_restart
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE mathconstants,                   ONLY: zero
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_dftb_matrices,                ONLY: build_dftb_overlap
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_mo_io,                        ONLY: read_mo_set,&
                                              read_rt_mos_from_restart
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_set,&
                                              qs_rho_type
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: get_restart_wfn, &
             calc_S_derivs, &
             calc_update_rho, &
             calc_update_rho_sparse, &
             calculate_P_imaginary

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation_utils'

CONTAINS

! **************************************************************************************************
!> \brief Calculates dS/dR respectily the velocity weighted derivatves
!>        only needed for ehrenfest MD.
!>
!> \param qs_env the qs environment
!> \par History
!>      02.2009 created [Manuel Guidon]
!>      02.2014 switched to dbcsr matrices [Samuel Andermatt]
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE calc_S_derivs(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_S_derivs', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: col_atom, handle, i, j, m, maxder, n, &
                                                            nder, row_atom
      INTEGER, DIMENSION(6, 2)                           :: c_map_mat
      LOGICAL                                            :: return_s_derivatives
      REAL(dp), DIMENSION(:), POINTER                    :: block_values
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: C_mat, S_der, s_derivs
      TYPE(cp_dbcsr_type), POINTER                       :: B_mat, tmp_mat, tmp_mat2
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(rt_prop_type), POINTER                        :: rtp

      CALL timeset(routineN, handle)

      return_s_derivatives = .TRUE.

      NULLIFY (particle_set)
      NULLIFY (rtp)
      NULLIFY (s_derivs)
      NULLIFY (dft_control)
      NULLIFY (ks_env)

      CALL get_qs_env(qs_env=qs_env, &
                      rtp=rtp, &
                      particle_set=particle_set, &
                      sab_orb=sab_orb, &
                      dft_control=dft_control, &
                      ks_env=ks_env)

      CALL get_rtp(rtp=rtp, B_mat=B_mat, C_mat=C_mat, S_der=S_der)

      nder = 2
      maxder = ncoset(nder)

      NULLIFY (tmp_mat)
      ALLOCATE (tmp_mat)
      CALL cp_dbcsr_init(tmp_mat)
      CALL cp_dbcsr_create(tmp_mat, template=S_der(1)%matrix, matrix_type="N")

      IF (rtp%iter < 2) THEN
         ! calculate the overlap derivative matrices
         IF (dft_control%qs_control%dftb) THEN
            CALL build_dftb_overlap(qs_env, nder, s_derivs)
         ELSE
            CALL build_overlap_matrix(ks_env, nderivative=nder, matrix_s=s_derivs, &
                                      basis_type_a="ORB", basis_type_b="ORB", sab_nl=sab_orb, qs_input=qs_env%input)
         END IF

         NULLIFY (tmp_mat2)
         ALLOCATE (tmp_mat2)
         CALL cp_dbcsr_init(tmp_mat2)
         CALL cp_dbcsr_create(tmp_mat2, template=S_der(1)%matrix, matrix_type="S")
         DO m = 1, 9
            CALL cp_dbcsr_copy(tmp_mat2, s_derivs(m+1)%matrix)
            CALL cp_dbcsr_desymmetrize(tmp_mat2, S_der(m)%matrix)
            CALL cp_dbcsr_scale(S_der(m)%matrix, -one)
            CALL cp_dbcsr_filter(S_der(m)%matrix, rtp%filter_eps)
            !The diagonal should be zero
            CALL cp_dbcsr_iterator_start(iter, S_der(m)%matrix)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
               IF (row_atom == col_atom) block_values = 0.0_dp
            END DO
            CALL cp_dbcsr_iterator_stop(iter)
         END DO
         CALL cp_dbcsr_deallocate_matrix_set(s_derivs)
         CALL cp_dbcsr_deallocate_matrix(tmp_mat2)
      END IF

      !calculate scalar product v(Rb)*<alpha|d/dRb beta> (B_mat), and store the first derivatives

      CALL cp_dbcsr_set(B_mat, zero)
      DO m = 1, 3
         CALL cp_dbcsr_copy(tmp_mat, S_der(m)%matrix)
         CALL cp_dbcsr_iterator_start(iter, tmp_mat)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            IF (row_atom == col_atom) block_values = 0.0_dp
            block_values = block_values*particle_set(col_atom)%v(m)
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
         CALL cp_dbcsr_add(B_mat, tmp_mat, one, one)
      END DO
      CALL cp_dbcsr_filter(B_mat, rtp%filter_eps)
      !calculate C matrix: v(Rb)*<d/dRa alpha| d/dRb beta>

      c_map_mat = 0
      n = 0
      DO j = 1, 3
         DO m = j, 3
            n = n+1
            c_map_mat(n, 1) = j
            IF (m == j) CYCLE
            c_map_mat(n, 2) = m
         END DO
      END DO

      DO i = 1, 3
         CALL cp_dbcsr_set(C_mat(i)%matrix, zero)
      END DO
      DO m = 1, 6
         CALL cp_dbcsr_copy(tmp_mat, S_der(m+3)%matrix)
         DO j = 1, 2
            IF (c_map_mat(m, j) == 0) CYCLE
            CALL cp_dbcsr_add(C_mat(c_map_mat(m, j))%matrix, tmp_mat, one, one)
         END DO
      END DO

      DO m = 1, 3
         CALL cp_dbcsr_iterator_start(iter, C_mat(m)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            block_values = block_values*particle_set(row_atom)%v(m)
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
         CALL cp_dbcsr_filter(C_mat(m)%matrix, rtp%filter_eps)
      END DO

      CALL cp_dbcsr_deallocate_matrix(tmp_mat)
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief reads the restart file. At the moment only SCF (means only real)
!> \param qs_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE get_restart_wfn(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_restart_wfn', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, project_name
      INTEGER                                            :: i, id_nr, im, ispin, ncol, nspin, re, &
                                                            unit_nr
      REAL(KIND=dp)                                      :: alpha, cs_pos
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_rmpv, rho_new, rho_old
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_old
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(section_vals_type), POINTER                   :: dft_section, input

      NULLIFY (atomic_kind_set, qs_kind_set, mo_array, particle_set, rho_struct, para_env)

      CALL get_qs_env(qs_env, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, &
                      mos=mo_array, &
                      input=input, &
                      rtp=rtp, &
                      dft_control=dft_control, &
                      rho=rho_struct, &
                      para_env=para_env)
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      id_nr = 0
      nspin = SIZE(mo_array)
      CALL qs_rho_get(rho_struct, rho_ao=p_rmpv)
      dft_section => section_vals_get_subs_vals(input, "DFT")
      SELECT CASE (dft_control%rtp_control%initial_wfn)
      CASE (use_restart_wfn)
         CALL read_mo_set(mo_array, atomic_kind_set, qs_kind_set, particle_set, para_env, &
                          id_nr=id_nr, multiplicity=dft_control%multiplicity, dft_section=dft_section)
         DO ispin = 1, nspin
            CALL calculate_density_matrix(mo_array(ispin)%mo_set, p_rmpv(ispin)%matrix)
         ENDDO
         IF (rtp%linear_scaling) THEN
            CALL get_rtp(rtp=rtp, rho_old=rho_old, rho_new=rho_new)
            DO ispin = 1, nspin
               re = 2*ispin-1
               im = 2*ispin
               CALL cp_fm_get_info(mo_array(ispin)%mo_set%mo_coeff, ncol_global=ncol)
               alpha = 1.0_dp
               IF (SIZE(mo_array) == 1) alpha = 2*alpha
               CALL cp_dbcsr_plus_fm_fm_t( &
                  sparse_matrix=rho_old(re)%matrix, &
                  matrix_v=mo_array(ispin)%mo_set%mo_coeff, matrix_g=mo_array(ispin)%mo_set%mo_coeff, ncol=ncol, &
                  keep_sparsity=.FALSE., alpha=alpha)
            END DO
            DO i = 1, nspin
               CALL cp_dbcsr_copy(rho_new(i)%matrix, rho_old(i)%matrix)
            ENDDO
            CALL calc_update_rho_sparse(qs_env)
         ELSE
            CALL get_rtp(rtp=rtp, mos_old=mos_old)
            DO i = 1, SIZE(qs_env%mos)
               CALL cp_fm_to_fm(mo_array(i)%mo_set%mo_coeff, mos_old(2*i-1)%matrix)
               CALL cp_fm_set_all(mos_old(2*i)%matrix, zero, zero)
            END DO
         ENDIF
      CASE (use_rt_restart)
         IF (rtp%linear_scaling) THEN
            CALL get_rtp(rtp=rtp, rho_old=rho_old, rho_new=rho_new)
            project_name = logger%iter_info%project_name
            DO ispin = 1, nspin
               re = 2*ispin-1
               im = 2*ispin
               WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_RE", ispin, "_RESTART.dm"
               CALL cp_dbcsr_binary_read(file_name, distribution=cp_dbcsr_distribution(rho_old(re)%matrix), &
                                         matrix_new=rho_old(re)%matrix)
               cs_pos = cp_dbcsr_checksum(rho_old(re)%matrix, pos=.TRUE.)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, '(T2,A,E20.8)') "Read restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
               ENDIF
               WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_IM", ispin, "_RESTART.dm"
               CALL cp_dbcsr_binary_read(file_name, distribution=cp_dbcsr_distribution(rho_old(im)%matrix), &
                                         matrix_new=rho_old(im)%matrix)
               cs_pos = cp_dbcsr_checksum(rho_old(im)%matrix, pos=.TRUE.)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, '(T2,A,E20.8)') "Read restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
               ENDIF
            ENDDO
            DO i = 1, SIZE(rho_new)
               CALL cp_dbcsr_copy(rho_new(i)%matrix, rho_old(i)%matrix)
            ENDDO
            CALL calc_update_rho_sparse(qs_env)
         ELSE
            CALL get_rtp(rtp=rtp, mos_old=mos_old, mos_new=mos_new)
            CALL read_rt_mos_from_restart(mo_array, mos_old, atomic_kind_set, qs_kind_set, particle_set, para_env, &
                                          id_nr, dft_control%multiplicity, dft_section)
            DO ispin = 1, nspin
               CALL calculate_density_matrix(mo_array(ispin)%mo_set, &
                                             p_rmpv(ispin)%matrix)
            ENDDO
         ENDIF
      END SELECT

   END SUBROUTINE get_restart_wfn

! **************************************************************************************************
!> \brief calculates the density from the complex MOs and passes the density to
!>        qs_env.
!> \param qs_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE calc_update_rho(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_update_rho', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, ncol, re
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao, rho_ao_im
      TYPE(cp_fm_p_type)                                 :: mos_occ
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rt_prop_type), POINTER                        :: rtp

      CALL timeset(routineN, handle)

      NULLIFY (rho, ks_env, mos_new, rtp)
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      rho=rho, &
                      rtp=rtp, &
                      mos=mos)
      CALL get_rtp(rtp=rtp, mos_new=mos_new)
      CALL qs_rho_get(rho_struct=rho, rho_ao=rho_ao)
      DO i = 1, SIZE(mos_new)/2
         re = 2*i-1; im = 2*i
         alpha = 3*one-REAL(SIZE(mos_new)/2, dp)
         CALL cp_dbcsr_set(rho_ao(i)%matrix, zero)
         CALL cp_fm_get_info(mos_new(re)%matrix, ncol_global=ncol)
         CALL cp_fm_create(mos_occ%matrix, &
                           matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                           name="mos_occ")
         CALL cp_fm_to_fm(mos_new(re)%matrix, mos_occ%matrix)
         CALL cp_fm_column_scale(mos_occ%matrix, mos(i)%mo_set%occupation_numbers/alpha)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=rho_ao(i)%matrix, &
                                    matrix_v=mos_occ%matrix, &
                                    ncol=ncol, &
                                    alpha=alpha)
         ! It is actually complex conjugate but i*i=-1 therfore it must be added
         CALL cp_fm_to_fm(mos_new(im)%matrix, mos_occ%matrix)
         CALL cp_fm_column_scale(mos_occ%matrix, mos(i)%mo_set%occupation_numbers/alpha)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=rho_ao(i)%matrix, &
                                    matrix_v=mos_occ%matrix, &
                                    ncol=ncol, &
                                    alpha=alpha)
         CALL cp_fm_release(mos_occ%matrix)
      END DO
      CALL qs_rho_update_rho(rho, qs_env)

      IF (rtp%do_hfx) THEN
         CALL qs_rho_get(rho_struct=rho, rho_ao_im=rho_ao_im)
         CALL calculate_P_imaginary(qs_env, rtp, rho_ao_im, keep_sparsity=.TRUE.)
         CALL qs_rho_set(rho, rho_ao_im=rho_ao_im)
      END IF

      CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)

      CALL timestop(handle)

   END SUBROUTINE calc_update_rho

! **************************************************************************************************
!> \brief Copies the density matrix back into the qs_env%rho%rho_ao
!> \param qs_env ...
!> \author Samuel Andermatt (3.14)
! **************************************************************************************************

   SUBROUTINE calc_update_rho_sparse(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_update_rho_sparse', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, im, ispin, re
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao, rho_ao_im, rho_new
      TYPE(cp_dbcsr_type), POINTER                       :: S_minus_half, tmp, tmp2
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      NULLIFY (rho, ks_env, rtp, tmp, tmp2, dft_control)
      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      rho=rho, &
                      rtp=rtp, &
                      dft_control=dft_control)
      rtp_control => dft_control%rtp_control
      CALL get_rtp(rtp=rtp, rho_new=rho_new)
      CALL qs_rho_get(rho_struct=rho, rho_ao=rho_ao)
      IF (rtp%do_hfx) CALL qs_rho_get(rho_struct=rho, rho_ao_im=rho_ao_im)
      IF (rtp_control%orthonormal) THEN
         CALL get_rtp(rtp=rtp, S_minus_half=S_minus_half)
         ALLOCATE (tmp)
         CALL cp_dbcsr_init(tmp)
         CALL cp_dbcsr_create(tmp, template=rho_new(1)%matrix)
         ALLOCATE (tmp2)
         CALL cp_dbcsr_init(tmp2)
         CALL cp_dbcsr_create(tmp2, template=rho_new(1)%matrix)
         DO ispin = 1, SIZE(rho_ao)
            re = 2*ispin-1
            CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, rho_new(re)%matrix, zero, tmp, filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp, S_minus_half, zero, tmp2, filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_set(rho_ao(ispin)%matrix, zero)
            CALL cp_dbcsr_copy_into_existing(rho_ao(ispin)%matrix, tmp2)
         END DO
         IF (rtp%do_hfx) THEN
            DO ispin = 1, SIZE(rho_ao_im)
               im = 2*ispin
               CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, rho_new(im)%matrix, zero, tmp, filter_eps=rtp%filter_eps)
               CALL cp_dbcsr_multiply("N", "N", one, tmp, S_minus_half, zero, tmp2, filter_eps=rtp%filter_eps)
               CALL cp_dbcsr_set(rho_ao_im(ispin)%matrix, zero)
               CALL cp_dbcsr_copy_into_existing(rho_ao_im(ispin)%matrix, tmp2)
            END DO
         ENDIF
         CALL cp_dbcsr_deallocate_matrix(tmp)
         CALL cp_dbcsr_deallocate_matrix(tmp2)
      ELSE
         DO ispin = 1, SIZE(rho_ao)
            CALL cp_dbcsr_set(rho_ao(ispin)%matrix, zero)
            CALL cp_dbcsr_copy_into_existing(rho_ao(ispin)%matrix, rho_new(ispin*2-1)%matrix)
            IF (rtp%do_hfx) CALL cp_dbcsr_copy_into_existing(rho_ao_im(ispin)%matrix, rho_new(ispin*2)%matrix)
         END DO
      ENDIF

      CALL qs_rho_update_rho(rho, qs_env)
      CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)

      CALL timestop(handle)

   END SUBROUTINE calc_update_rho_sparse

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rtp ...
!> \param matrix_p_im ...
!> \param keep_sparsity ...
! **************************************************************************************************
   SUBROUTINE calculate_P_imaginary(qs_env, rtp, matrix_p_im, keep_sparsity)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p_im
      LOGICAL, OPTIONAL                                  :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_P_imaginary', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, im, ncol, re
      LOGICAL                                            :: my_keep_sparsity
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_occ
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CALL get_rtp(rtp=rtp, mos_new=mos_new)

      my_keep_sparsity = .FALSE.
      IF (PRESENT(keep_sparsity)) my_keep_sparsity = keep_sparsity
      CALL get_qs_env(qs_env, mos=mos)
      ALLOCATE (mos_occ(SIZE(mos)*2))

      DO i = 1, SIZE(mos_new)/2
         re = 2*i-1; im = 2*i
         alpha = 3.0_dp-REAL(SIZE(matrix_p_im), dp)
         CALL cp_fm_create(mos_occ(re)%matrix, &
                           matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                           name="mos_occ")
         CALL cp_fm_create(mos_occ(im)%matrix, &
                           matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                           name="mos_occ")
         CALL cp_dbcsr_set(matrix_p_im(i)%matrix, 0.0_dp)
         CALL cp_fm_get_info(mos_new(re)%matrix, ncol_global=ncol)
         CALL cp_fm_to_fm(mos_new(re)%matrix, mos_occ(re)%matrix)
         CALL cp_fm_column_scale(mos_occ(re)%matrix, mos(i)%mo_set%occupation_numbers/alpha)
         CALL cp_fm_to_fm(mos_new(im)%matrix, mos_occ(im)%matrix)
         CALL cp_fm_column_scale(mos_occ(im)%matrix, mos(i)%mo_set%occupation_numbers/alpha)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=matrix_p_im(i)%matrix, &
                                    matrix_v=mos_occ(im)%matrix, &
                                    matrix_g=mos_occ(re)%matrix, &
                                    ncol=ncol, &
                                    keep_sparsity=my_keep_sparsity, &
                                    alpha=alpha)
         ! It is actually complex conjugate not only transposed
         alpha = -alpha
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=matrix_p_im(i)%matrix, &
                                    matrix_v=mos_occ(re)%matrix, &
                                    matrix_g=mos_occ(im)%matrix, &
                                    ncol=ncol, &
                                    keep_sparsity=my_keep_sparsity, &
                                    alpha=alpha)
      END DO
      CALL cp_fm_vect_dealloc(mos_occ)

   END SUBROUTINE calculate_P_imaginary

END MODULE rt_propagation_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routine for the real time propagation output.
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE rt_propagation_output
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_binary_write, cp_dbcsr_checksum, &
        cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_desymmetrize, cp_dbcsr_distribution, &
        cp_dbcsr_filter, cp_dbcsr_get_occupation, cp_dbcsr_init, cp_dbcsr_init_p, &
        cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, &
        cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, cp_dbcsr_p_type, cp_dbcsr_set, &
        cp_dbcsr_type, dbcsr_distribution_mp, dbcsr_mp_group
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_double,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_iter_string,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE input_constants,                 ONLY: ehrenfest,&
                                              real_time_propagation
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE message_passing,                 ONLY: mp_max
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind_set,&
                                              qs_kind_type
   USE qs_linres_current,               ONLY: calculate_jrho_resp
   USE qs_linres_types,                 ONLY: current_env_type
   USE qs_mo_io,                        ONLY: write_rt_mos_to_restart
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_post_dftb,                ONLY: scf_post_calculation_dftb
   USE qs_scf_post_gpw,                 ONLY: qs_scf_post_moments,&
                                              write_available_results,&
                                              write_mo_free_results
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
   USE rt_propagation_utils,            ONLY: calculate_P_imaginary
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation_output'

   PUBLIC :: rt_prop_output, &
             rt_convergence, &
             rt_convergence_density, &
             report_density_occupation

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param run_type ...
!> \param delta_iter ...
!> \param used_time ...
! **************************************************************************************************
   SUBROUTINE rt_prop_output(qs_env, run_type, delta_iter, used_time)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(in)                                :: run_type
      REAL(dp), INTENT(in), OPTIONAL                     :: delta_iter, used_time

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_prop_output', routineP = moduleN//':'//routineN

      INTEGER                                            :: n_electrons, nspin, output_unit, spin
      REAL(dp)                                           :: orthonormality, tot_rho_r
      REAL(KIND=dp), DIMENSION(:), POINTER               :: qs_tot_rho_r
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, P_im, rho_new
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(section_vals_type), POINTER                   :: dft_section, input, rtp_section

      NULLIFY (logger, dft_control)

      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env, &
                      rtp=rtp, &
                      matrix_s=matrix_s, &
                      input=input, &
                      rho=rho, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control)

      rtp_section => section_vals_get_subs_vals(input, "DFT%REAL_TIME_PROPAGATION")

      CALL get_qs_kind_set(qs_kind_set, nelectron=n_electrons)
      n_electrons = n_electrons-dft_control%charge

      CALL qs_rho_get(rho_struct=rho, tot_rho_r=qs_tot_rho_r)

      tot_rho_r = accurate_sum(qs_tot_rho_r)

      output_unit = cp_print_key_unit_nr(logger, rtp_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")

      IF (output_unit > 0) THEN
         WRITE (output_unit, FMT="(/,(T3,A,T40,I5))") &
            "Information at iteration step:", rtp%iter
         WRITE (UNIT=output_unit, FMT="((T3,A,T41,2F20.10))") &
            "Total electronic density (r-space): ", &
            tot_rho_r, &
            tot_rho_r+ &
            REAL(n_electrons, dp)
         WRITE (UNIT=output_unit, FMT="((T3,A,T59,F22.14))") &
            "Total energy:", rtp%energy_new
         IF (run_type == ehrenfest) &
            WRITE (UNIT=output_unit, FMT="((T3,A,T61,F20.14))") &
            "Energy difference to previous iteration step:", rtp%energy_new-rtp%energy_old
         IF (run_type == real_time_propagation) &
            WRITE (UNIT=output_unit, FMT="((T3,A,T61,F20.14))") &
            "Energy difference to initial state:", rtp%energy_new-rtp%energy_old
         IF (PRESENT(delta_iter)) &
            WRITE (UNIT=output_unit, FMT="((T3,A,T61,E20.6))") &
            "Convergence:", delta_iter
         IF (rtp%converged) THEN
            IF (run_type == real_time_propagation) &
               WRITE (UNIT=output_unit, FMT="((T3,A,T61,F12.2))") &
               "Time needed for propagation:", used_time
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,3X,F16.14))") &
               "CONVERGENCE REACHED", rtp%energy_new-rtp%energy_old
         END IF
      END IF

      IF (rtp%converged) THEN
         IF (.NOT. rtp%linear_scaling) THEN
            CALL get_rtp(rtp=rtp, mos_new=mos_new)
            CALL rt_calculate_orthonormality(orthonormality, &
                                             mos_new, matrix_s(1)%matrix)
            IF (output_unit > 0) &
               WRITE (output_unit, FMT="(/,(T3,A,T60,F20.10))") &
               "Max deviation from orthonormalization:", orthonormality
         ENDIF
      END IF

      IF (output_unit > 0) &
         CALL m_flush(output_unit)
      CALL cp_print_key_finished_output(output_unit, logger, rtp_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      IF (rtp%converged) THEN
         CALL make_moment(qs_env)
         dft_section => section_vals_get_subs_vals(input, "DFT")
         IF (rtp%linear_scaling) THEN
            CALL get_rtp(rtp=rtp, rho_new=rho_new)
            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 dft_section, "REAL_TIME_PROPAGATION%PRINT%RESTART"), cp_p_file)) THEN
               CALL write_rt_p_to_restart(rho_new, .FALSE.)
            ENDIF
            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 dft_section, "REAL_TIME_PROPAGATION%PRINT%RESTART_HISTORY"), cp_p_file)) THEN
               CALL write_rt_p_to_restart(rho_new, .TRUE.)
            ENDIF
            IF (.NOT. dft_control%qs_control%dftb) THEN
               !Not sure if these things could also work with dftb or not
               CALL write_mo_free_results(qs_env)
               IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                    dft_section, "REAL_TIME_PROPAGATION%PRINT%CURRENT"), cp_p_file)) THEN
                  DO spin = 1, SIZE(rho_new)/2
                     CALL rt_current(qs_env, rho_new(2*spin)%matrix, dft_section, spin)
                  END DO
               ENDIF
            ENDIF
         ELSE
            CALL get_rtp(rtp=rtp, mos_new=mos_new)
            IF (.NOT. dft_control%qs_control%dftb) THEN
               CALL write_available_results(qs_env=qs_env)
               IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                    dft_section, "REAL_TIME_PROPAGATION%PRINT%CURRENT"), cp_p_file)) THEN
                  NULLIFY (P_im)
                  nspin = SIZE(mos_new)/2
                  CALL cp_dbcsr_allocate_matrix_set(P_im, nspin)
                  DO spin = 1, nspin
                     CALL cp_dbcsr_init_p(P_im(spin)%matrix)
                     CALL cp_dbcsr_create(P_im(spin)%matrix, template=matrix_s(1)%matrix, matrix_type="N")
                  END DO
                  CALL calculate_P_imaginary(qs_env, rtp, P_im)
                  DO spin = 1, nspin
                     CALL rt_current(qs_env, P_im(spin)%matrix, dft_section, spin)
                  END DO
                  CALL cp_dbcsr_deallocate_matrix_set(P_im)
               ENDIF
            ENDIF
            CALL write_rt_mos_to_restart(qs_env%mos, mos_new, particle_set, &
                                         dft_section, qs_kind_set)
         ENDIF
      ENDIF

      rtp%energy_old = rtp%energy_new

      IF (.NOT. rtp%converged .AND. rtp%iter >= dft_control%rtp_control%max_iter) &
         CALL cp_abort(cp__l("emd/rt_propagation_output.F",251), "EMD did not converge, either increase MAX_ITER "// &
                       "or use a smaller TIMESTEP")

   END SUBROUTINE rt_prop_output

! **************************************************************************************************
!> \brief computes the effective orthonormality of a set of mos given an s-matrix
!>        orthonormality is the max deviation from unity of the C^T S C
!> \param orthonormality ...
!> \param mos_new ...
!> \param matrix_s ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************
   SUBROUTINE rt_calculate_orthonormality(orthonormality, mos_new, matrix_s)
      REAL(KIND=dp), INTENT(out)                         :: orthonormality
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_calculate_orthonormality', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, im, ispin, j, k, n, &
                                                            ncol_local, nrow_local, nspin, re
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: alpha, max_alpha, max_beta
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: overlap_re, svec_im, svec_re

      NULLIFY (tmp_fm_struct, svec_im, svec_re, overlap_re)

      CALL timeset(routineN, handle)

      nspin = SIZE(mos_new)/2
      max_alpha = 0.0_dp
      max_beta = 0.0_dp
      DO ispin = 1, nspin
         re = ispin*2-1
         im = ispin*2
         ! get S*C
         CALL cp_fm_create(svec_re, mos_new(im)%matrix%matrix_struct)
         CALL cp_fm_create(svec_im, mos_new(im)%matrix%matrix_struct)
         CALL cp_fm_get_info(mos_new(im)%matrix, &
                             nrow_global=n, ncol_global=k)
         CALL cp_dbcsr_sm_fm_multiply(matrix_s, mos_new(re)%matrix, &
                                      svec_re, k)
         CALL cp_dbcsr_sm_fm_multiply(matrix_s, mos_new(im)%matrix, &
                                      svec_im, k)

         ! get C^T (S*C)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=k, ncol_global=k, &
                                  para_env=mos_new(re)%matrix%matrix_struct%para_env, &
                                  context=mos_new(re)%matrix%matrix_struct%context)
         CALL cp_fm_create(overlap_re, tmp_fm_struct)

         CALL cp_fm_struct_release(tmp_fm_struct)

         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, mos_new(re)%matrix, &
                      svec_re, 0.0_dp, overlap_re)
         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, mos_new(im)%matrix, &
                      svec_im, 1.0_dp, overlap_re)

         CALL cp_fm_release(svec_re)
         CALL cp_fm_release(svec_im)

         CALL cp_fm_get_info(overlap_re, nrow_local=nrow_local, ncol_local=ncol_local, &
                             row_indices=row_indices, col_indices=col_indices)
         DO i = 1, nrow_local
            DO j = 1, ncol_local
               alpha = overlap_re%local_data(i, j)
               IF (row_indices(i) .EQ. col_indices(j)) alpha = alpha-1.0_dp
               max_alpha = MAX(max_alpha, ABS(alpha))
            ENDDO
         ENDDO
         CALL cp_fm_release(overlap_re)
      ENDDO
      CALL mp_max(max_alpha, mos_new(1)%matrix%matrix_struct%para_env%group)
      CALL mp_max(max_beta, mos_new(1)%matrix%matrix_struct%para_env%group)
      orthonormality = max_alpha

      CALL timestop(handle)

   END SUBROUTINE rt_calculate_orthonormality

! **************************************************************************************************
!> \brief computs the convergence criterium for RTP and EMD
!> \param rtp ...
!> \param matrix_s Overlap matrix without the derivatives
!> \param delta_mos ...
!> \param delta_eps ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE rt_convergence(rtp, matrix_s, delta_mos, delta_eps)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: delta_mos
      REAL(dp), INTENT(out)                              :: delta_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_convergence', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, icol, im, ispin, j, lcol, &
                                                            lrow, nao, newdim, nmo, nspin, re
      LOGICAL                                            :: double_col, double_row
      REAL(KIND=dp)                                      :: alpha, max_alpha
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(cp_fm_struct_type), POINTER                   :: newstruct, newstruct1, tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: work, work1, work2

      NULLIFY (tmp_fm_struct)

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, mos_new=mos_new)

      nspin = SIZE(delta_mos)/2
      max_alpha = 0.0_dp

      DO i = 1, SIZE(mos_new)
         CALL cp_fm_scale_and_add(-one, delta_mos(i)%matrix, one, mos_new(i)%matrix)
      END DO

      DO ispin = 1, nspin
         re = ispin*2-1
         im = ispin*2

         double_col = .TRUE.
         double_row = .FALSE.
         CALL cp_fm_struct_double(newstruct, &
                                  delta_mos(re)%matrix%matrix_struct, &
                                  delta_mos(re)%matrix%matrix_struct%context, &
                                  double_col, &
                                  double_row)

         CALL cp_fm_create(work, matrix_struct=newstruct)
         CALL cp_fm_create(work1, matrix_struct=newstruct)

         CALL cp_fm_get_info(delta_mos(re)%matrix, ncol_local=lcol, ncol_global=nmo, &
                             nrow_global=nao)
         CALL cp_fm_get_info(work, ncol_global=newdim)

         CALL cp_fm_set_all(work, zero, zero)

         DO icol = 1, lcol
            work%local_data(:, icol) = delta_mos(re)%matrix%local_data(:, icol)
            work%local_data(:, icol+lcol) = delta_mos(im)%matrix%local_data(:, icol)
         END DO

         CALL cp_dbcsr_sm_fm_multiply(matrix_s, work, work1, ncol=newdim)

         CALL cp_fm_release(work)

         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmo, ncol_global=nmo, &
                                  para_env=delta_mos(re)%matrix%matrix_struct%para_env, &
                                  context=delta_mos(re)%matrix%matrix_struct%context)
         CALL cp_fm_struct_double(newstruct1, &
                                  tmp_fm_struct, &
                                  delta_mos(re)%matrix%matrix_struct%context, &
                                  double_col, &
                                  double_row)

         CALL cp_fm_create(work, matrix_struct=newstruct1)
         CALL cp_fm_create(work2, matrix_struct=newstruct1)

         CALL cp_gemm("T", "N", nmo, newdim, nao, one, delta_mos(re)%matrix, &
                      work1, zero, work)

         CALL cp_gemm("T", "N", nmo, newdim, nao, one, delta_mos(im)%matrix, &
                      work1, zero, work2)

         CALL cp_fm_get_info(work, nrow_local=lrow)
         DO i = 1, lrow
            DO j = 1, lcol
               alpha = SQRT((work%local_data(i, j)+work2%local_data(i, j+lcol))**2+ &
                            (work%local_data(i, j+lcol)-work2%local_data(i, j))**2)
               max_alpha = MAX(max_alpha, ABS(alpha))
            ENDDO
         ENDDO

         CALL cp_fm_release(work)
         CALL cp_fm_release(work1)
         CALL cp_fm_release(work2)
         CALL cp_fm_struct_release(tmp_fm_struct)
         CALL cp_fm_struct_release(newstruct)
         CALL cp_fm_struct_release(newstruct1)

      ENDDO

      CALL mp_max(max_alpha, delta_mos(1)%matrix%matrix_struct%para_env%group)
      delta_eps = SQRT(max_alpha)

      CALL timestop(handle)

   END SUBROUTINE rt_convergence

! **************************************************************************************************
!> \brief computs the convergence criterium for RTP and EMD based on the density matrix
!> \param rtp ...
!> \param delta_P ...
!> \param delta_eps ...
!> \author Samuel Andermatt (02.14)
! **************************************************************************************************

   SUBROUTINE rt_convergence_density(rtp, delta_P, delta_eps)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: delta_P
      REAL(dp), INTENT(out)                              :: delta_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_convergence_density', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: col_atom, handle, i, ispin, row_atom
      REAL(dp)                                           :: alpha, max_alpha
      REAL(dp), DIMENSION(:), POINTER                    :: block_values
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_new
      TYPE(cp_dbcsr_type), POINTER                       :: tmp

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, rho_new=rho_new)

      DO i = 1, SIZE(rho_new)
         CALL cp_dbcsr_add(delta_P(i)%matrix, rho_new(i)%matrix, one, -one)
      ENDDO
      !get the maximum value of delta_P
      DO i = 1, SIZE(delta_P)
         !square all entries of both matrices
         CALL cp_dbcsr_iterator_start(iter, delta_P(i)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            block_values = block_values*block_values
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
      END DO
      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=delta_P(1)%matrix, matrix_type="N")
      DO ispin = 1, SIZE(delta_P)/2
         CALL cp_dbcsr_desymmetrize(delta_P(2*ispin-1)%matrix, tmp)
         CALL cp_dbcsr_add(delta_P(2*ispin)%matrix, tmp, one, one)
      END DO
      !the absolute values are now in the even entries of delta_P
      max_alpha = zero
      DO ispin = 1, SIZE(delta_P)/2
         CALL cp_dbcsr_iterator_start(iter, delta_P(2*ispin)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            alpha = MAXVAL(block_values)
            IF (alpha > max_alpha) max_alpha = alpha
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
      END DO
      CALL mp_max(max_alpha, dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(delta_P(1)%matrix))))
      delta_eps = SQRT(max_alpha)
      CALL cp_dbcsr_deallocate_matrix(tmp)
      CALL timestop(handle)

   END SUBROUTINE rt_convergence_density

! **************************************************************************************************
!> \brief interface to qs_moments. Does only work for nonperiodic dipole
!> \param qs_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE make_moment(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'make_moment', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: dft_section, input

      CALL timeset(routineN, handle)

      NULLIFY (input, dft_control)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL get_qs_env(qs_env, input=input, dft_control=dft_control)
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      IF (dft_control%qs_control%dftb) THEN
         CALL scf_post_calculation_dftb(dft_section, qs_env=qs_env)
      ELSE
         CALL qs_scf_post_moments(input, logger, qs_env, output_unit)
      END IF
      CALL timestop(handle)

   END SUBROUTINE make_moment

! **************************************************************************************************
!> \brief Reports the sparsity pattern of the complex density matrix
!> \param filter_eps ...
!> \param rho ...
!> \author Samuel Andermatt (09.14)
! **************************************************************************************************

   SUBROUTINE report_density_occupation(filter_eps, rho)

      REAL(KIND=dp)                                      :: filter_eps
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho

      CHARACTER(len=*), PARAMETER :: routineN = 'report_density_occupation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, im, ispin, re, unit_nr
      REAL(KIND=dp)                                      :: eps, occ
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      unit_nr = cp_logger_get_default_io_unit(logger)
      NULLIFY (tmp)
      CALL cp_dbcsr_allocate_matrix_set(tmp, SIZE(rho))
      DO i = 1, SIZE(rho)
         CALL cp_dbcsr_init_p(tmp(i)%matrix)
         CALL cp_dbcsr_create(tmp(i)%matrix, template=rho(i)%matrix)
         CALL cp_dbcsr_copy(tmp(i)%matrix, rho(i)%matrix)
      ENDDO
      DO ispin = 1, SIZE(rho)/2
         re = 2*ispin-1
         im = 2*ispin
         eps = MAX(filter_eps, 10E-12_dp)
         DO WHILE (eps < 1.1_dp)
            CALL cp_dbcsr_filter(tmp(re)%matrix, eps)
            occ = cp_dbcsr_get_occupation(tmp(re)%matrix)
            IF (unit_nr > 0) WRITE (unit_nr, FMT="((T3,A,I1,A,F15.12,A,T61,F20.10))") "Occupation of rho spin ", &
               ispin, " eps ", eps, " real: ", occ
            eps = eps*10
         ENDDO
         eps = MAX(filter_eps, 10E-12_dp)
         DO WHILE (eps < 1.1_dp)
            CALL cp_dbcsr_filter(tmp(im)%matrix, eps)
            occ = cp_dbcsr_get_occupation(tmp(im)%matrix)
            IF (unit_nr > 0) WRITE (unit_nr, FMT="((T3,A,I1,A,F15.12,A,T61,F20.10))") "Occupation of rho spin ", &
               ispin, " eps ", eps, " imag: ", occ
            eps = eps*10
         ENDDO
      ENDDO
      CALL cp_dbcsr_deallocate_matrix_set(tmp)
      CALL timestop(handle)

   END SUBROUTINE report_density_occupation

! **************************************************************************************************
!> \brief Writes the density matrix and the atomic positions to a restart file
!> \param rho_new ...
!> \param history ...
!> \author Samuel Andermatt (09.14)
! **************************************************************************************************

   SUBROUTINE write_rt_p_to_restart(rho_new, history)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_new
      LOGICAL                                            :: history

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_rt_p_to_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, project_name
      INTEGER                                            :: handle, im, ispin, re, unit_nr
      REAL(KIND=dp)                                      :: cs_pos
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      project_name = logger%iter_info%project_name
      DO ispin = 1, SIZE(rho_new)/2
         re = 2*ispin-1
         im = 2*ispin
         IF (history) THEN
            WRITE (file_name, '(A,I0,A)') &
               TRIM(project_name)//"_LS_DM_SPIN_RE", ispin, "_"//TRIM(cp_iter_string(logger%iter_info))//"_RESTART.dm"
         ELSE
            WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_RE", ispin, "_RESTART.dm"
         ENDIF
         cs_pos = cp_dbcsr_checksum(rho_new(re)%matrix, pos=.TRUE.)
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T2,A,E20.8)') "Writing restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
         ENDIF
         CALL cp_dbcsr_binary_write(rho_new(re)%matrix, file_name)
         IF (history) THEN
            WRITE (file_name, '(A,I0,A)') &
               TRIM(project_name)//"_LS_DM_SPIN_IM", ispin, "_"//TRIM(cp_iter_string(logger%iter_info))//"_RESTART.dm"
         ELSE
            WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_IM", ispin, "_RESTART.dm"
         ENDIF
         cs_pos = cp_dbcsr_checksum(rho_new(im)%matrix, pos=.TRUE.)
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T2,A,E20.8)') "Writing restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
         ENDIF
         CALL cp_dbcsr_binary_write(rho_new(im)%matrix, file_name)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE write_rt_p_to_restart

! **************************************************************************************************
!> \brief Collocation of the current and printing of it in a cube file
!> \param qs_env ...
!> \param P_im ...
!> \param dft_section ...
!> \param spin ...
!> \author Samuel Andermatt (06.15)
! **************************************************************************************************
   SUBROUTINE rt_current(qs_env, P_im, dft_section, spin)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_type), POINTER                       :: P_im
      TYPE(section_vals_type), POINTER                   :: dft_section
      INTEGER                                            :: spin

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_current', routineP = moduleN//':'//routineN

      CHARACTER(len=1)                                   :: char_spin
      CHARACTER(len=14)                                  :: ext
      CHARACTER(len=2)                                   :: sdir
      INTEGER                                            :: dir, handle, print_unit
      INTEGER, DIMENSION(:), POINTER                     :: stride(:)
      TYPE(cp_dbcsr_type), POINTER                       :: zero
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(current_env_type)                             :: current_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_p_type), POINTER                           :: gs, rs
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env=qs_env, subsys=subsys, pw_env=pw_env, rho=rho)
      CALL qs_subsys_get(subsys, particles=particles)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, rs_grids=rs_rho, rs_descs=rs_descs)
      CALL qs_rho_get(rho, rho_r=rho_r, rho_g=rho_g)

      NULLIFY (zero, rs, gs)
      ALLOCATE (rs, gs)
      NULLIFY (rs%pw, gs%pw)
      ALLOCATE (zero)
      CALL cp_dbcsr_init(zero)
      CALL cp_dbcsr_create(zero, template=P_im)
      CALL cp_dbcsr_copy(zero, P_im)
      CALL cp_dbcsr_set(zero, 0.0_dp)
      current_env%gauge = -1
      current_env%gauge_init = .FALSE.
      CALL pw_pool_create_pw(auxbas_pw_pool, rs%pw, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, gs%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      NULLIFY (stride)
      ALLOCATE (stride(3))

      DO dir = 1, 3

         CALL pw_zero(rs%pw)
         CALL pw_zero(gs%pw)

         CALL calculate_jrho_resp(zero, P_im, zero, zero, dir, dir, rs, gs, qs_env, current_env, retain_rsgrid=.TRUE.)

         stride = section_get_ivals(dft_section, "REAL_TIME_PROPAGATION%PRINT%CURRENT%STRIDE")

         IF (dir == 1) THEN
            sdir = "-x"
         ELSEIF (dir == 2) THEN
            sdir = "-y"
         ELSE
            sdir = "-z"
         ENDIF
         WRITE (char_spin, "(I1)") spin

         ext = "-SPIN-"//char_spin//sdir//".cube"

         print_unit = cp_print_key_unit_nr(logger, dft_section, "REAL_TIME_PROPAGATION%PRINT%CURRENT", &
                                           extension=ext, file_status="REPLACE", file_action="WRITE", &
                                           log_filename=.FALSE.)

         CALL cp_pw_to_cube(rs%pw, print_unit, "EMD current", particles=particles, stride=stride)

         CALL cp_print_key_finished_output(print_unit, logger, dft_section, "REAL_TIME_PROPAGATION%PRINT%CURRENT")

      END DO

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rs%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, gs%pw)

      DEALLOCATE (rs)
      DEALLOCATE (gs)

      CALL cp_dbcsr_deallocate_matrix(zero)

      DEALLOCATE (stride)

      CALL timestop(handle)

   END SUBROUTINE rt_current

END MODULE rt_propagation_output
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      CJM 20-Feb-2001: Now npt_ifo is allocated to zero when not used
!>      CJM 11-apr-2001: adding routines to thermostat ao_type
!>      CJM 02-Aug-2003: renamed
!> \author CJM
! **************************************************************************************************
MODULE extended_system_dynamics

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: lnhc_parameters_type,&
                                              map_info_type,&
                                              npt_info_type
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE shell_potential_types,           ONLY: shell_kind_type
   USE thermostat_utils,                ONLY: ke_region_baro,&
                                              ke_region_particles,&
                                              ke_region_shells,&
                                              vel_rescale_baro,&
                                              vel_rescale_particles,&
                                              vel_rescale_shells
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   LOGICAL, PARAMETER :: debug_this_module = .FALSE.
   PUBLIC :: shell_scale_comv, &
             lnhc_particles, &
             lnhc_barostat, &
             lnhc_shells

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'extended_system_dynamics'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param nhc ...
!> \param npt ...
!> \param group ...
!> \date 13-DEC-2000
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE lnhc_barostat(nhc, npt, group)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(len=*), PARAMETER :: routineN = 'lnhc_barostat', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      map_info => nhc%map_info

      ! Compute the kinetic energy of the barostat
      CALL ke_region_baro(map_info, npt, group)

      ! Calculate forces on the Nose-Hoover Thermostat and apply chains
      CALL do_nhc(nhc, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_baro(map_info, npt)

      CALL timestop(handle)
   END SUBROUTINE lnhc_barostat

! **************************************************************************************************
!> \brief ...
!> \param nhc ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param group ...
!> \param shell_adiabatic ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE lnhc_particles(nhc, molecule_kind_set, molecule_set, &
                             particle_set, local_molecules, group, shell_adiabatic, &
                             shell_particle_set, core_particle_set, vel, shell_vel, core_vel)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      INTEGER, INTENT(IN)                                :: group
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell_adiabatic
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'lnhc_particles', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_shell_adiabatic
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      my_shell_adiabatic = .FALSE.
      IF (PRESENT(shell_adiabatic)) my_shell_adiabatic = shell_adiabatic
      map_info => nhc%map_info

      ! Compute the kinetic energy for the region to thermostat
      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel)

      ! Calculate forces on the Nose-Hoover Thermostat and apply chains
      CALL do_nhc(nhc, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_particles(map_info, molecule_kind_set, molecule_set, particle_set, &
                                 local_molecules, my_shell_adiabatic, shell_particle_set, core_particle_set, &
                                 vel, shell_vel, core_vel)

      CALL timestop(handle)
   END SUBROUTINE lnhc_particles

! **************************************************************************************************
!> \brief ...
!> \param nhc ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param group ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE lnhc_shells(nhc, atomic_kind_set, particle_set, local_particles, &
                          group, shell_particle_set, core_particle_set, vel, shell_vel, core_vel)

      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      INTEGER, INTENT(IN)                                :: group
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'lnhc_shells', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      map_info => nhc%map_info

      ! Compute the kinetic energy of the region to thermostat
      CALL ke_region_shells(map_info, particle_set, atomic_kind_set, local_particles, &
                            group, core_particle_set, shell_particle_set, core_vel, shell_vel)

      ! Calculate forces on the Nose-Hoover Thermostat and apply chains
      CALL do_nhc(nhc, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_shells(map_info, atomic_kind_set, particle_set, local_particles, &
                              shell_particle_set, core_particle_set, shell_vel, core_vel, vel)

      CALL timestop(handle)
   END SUBROUTINE lnhc_shells

! **************************************************************************************************
!> \brief ...
!> \param nhc ...
!> \param map_info ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE do_nhc(nhc, map_info)
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      TYPE(map_info_type), POINTER                       :: map_info

      CHARACTER(len=*), PARAMETER :: routineN = 'do_nhc', routineP = moduleN//':'//routineN

      INTEGER                                            :: imap, n

! Force on the first bead in every thermostat chain

      DO n = 1, nhc%loc_num_nhc
         imap = nhc%map_info%map_index(n)
         IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
         nhc%nvt(1, n)%f = (map_info%s_kin(imap)-nhc%nvt(1, n)%nkt)/nhc%nvt(1, n)%mass
      END DO

      ! Perform multiple time stepping using Yoshida
      CALL multiple_step_yoshida(nhc)

   END SUBROUTINE do_nhc

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param com_vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shell_scale_comv(atomic_kind_set, local_particles, particle_set, &
                               com_vel, shell_vel, core_vel)

      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(IN)                          :: com_vel(:, :)
      REAL(KIND=dp), INTENT(INOUT)                       :: shell_vel(:, :), core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_scale_comv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local, shell_index
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: fac_massc, fac_masss, mass, vc(3), vs(3)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      nparticle_kind = SIZE(atomic_kind_set)

      DO iparticle_kind = 1, nparticle_kind
         atomic_kind => atomic_kind_set(iparticle_kind)
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass, &
                              shell_active=is_shell, shell=shell)
         IF (is_shell) THEN
            fac_masss = shell%mass_shell/mass
            fac_massc = shell%mass_core/mass
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               shell_index = particle_set(iparticle)%shell_index
               vs(1:3) = shell_vel(1:3, shell_index)
               vc(1:3) = core_vel(1:3, shell_index)
               shell_vel(1, shell_index) = com_vel(1, iparticle)+fac_massc*(vs(1)-vc(1))
               shell_vel(2, shell_index) = com_vel(2, iparticle)+fac_massc*(vs(2)-vc(2))
               shell_vel(3, shell_index) = com_vel(3, iparticle)+fac_massc*(vs(3)-vc(3))
               core_vel(1, shell_index) = com_vel(1, iparticle)+fac_masss*(vc(1)-vs(1))
               core_vel(2, shell_index) = com_vel(2, iparticle)+fac_masss*(vc(2)-vs(2))
               core_vel(3, shell_index) = com_vel(3, iparticle)+fac_masss*(vc(3)-vs(3))
            END DO
         END IF ! is_shell
      END DO ! iparticle_kind
   END SUBROUTINE shell_scale_comv

! **************************************************************************************************
!> \brief ...
!> \param nhc ...
!> \date 14-NOV-2000
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE multiple_step_yoshida(nhc)

      TYPE(lnhc_parameters_type), POINTER                :: nhc

      CHARACTER(len=*), PARAMETER :: routineN = 'multiple_step_yoshida', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imap, inc, inhc, iyosh, n, nx1, nx2
      REAL(KIND=dp)                                      :: scale
      TYPE(map_info_type), POINTER                       :: map_info

      nx1 = SIZE(nhc%nvt, 1)
      nx2 = SIZE(nhc%nvt, 2)
      map_info => nhc%map_info
      ! perform multiple time stepping using Yoshida
      NCLOOP: DO inc = 1, nhc%nc
         YOSH: DO iyosh = 1, nhc%nyosh

            ! update velocity on the last thermostat in the chain    ! O1
            nhc%nvt(nhc%nhc_len, :)%v = nhc%nvt(nhc%nhc_len, :)%v+ &
                                        nhc%nvt(nhc%nhc_len, :)%f*0.25_dp*nhc%dt_yosh(iyosh)*nhc%dt_fact

            ! update velocity of other thermostats on chain (from nhc_len-1 to 1)  ! O2
            DO n = 1, nhc%loc_num_nhc
               IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
               DO inhc = nhc%nhc_len-1, 1, -1
                  scale = EXP(-0.125_dp*nhc%nvt(inhc+1, n)%v*nhc%dt_yosh(iyosh)*nhc%dt_fact)
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v*scale ! scale
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v+ &
                                       nhc%nvt(inhc, n)%f*0.25_dp*nhc%dt_yosh(iyosh)*nhc%dt_fact ! shift
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v*scale ! scale
               END DO
            END DO

            ! the core of the operator ----- START------
            ! update nhc positions
            nhc%nvt(:, :)%eta = nhc%nvt(:, :)%eta+ &
                                0.5_dp*nhc%nvt(:, :)%v*nhc%dt_yosh(iyosh)*nhc%dt_fact

            ! now accumulate the scale factor for particle velocities
            DO n = 1, nhc%loc_num_nhc
               imap = nhc%map_info%map_index(n)
               IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
               map_info%v_scale(imap) = map_info%v_scale(imap)*EXP(-0.5_dp*nhc%dt_yosh(iyosh)*nhc%dt_fact*nhc%nvt(1, n)%v)
            END DO
            ! the core of the operator ------ END ------

            ! update the force on first thermostat again (since particle velocities changed)
            DO n = 1, nhc%loc_num_nhc
               imap = nhc%map_info%map_index(n)
               IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
               nhc%nvt(1, n)%f = (map_info%s_kin(imap)*map_info%v_scale(imap)* &
                                  map_info%v_scale(imap)-nhc%nvt(1, n)%nkt)/nhc%nvt(1, n)%mass
            END DO

            ! update velocity of other thermostats on chain (from 1 to nhc_len-1)  ! O2
            DO inhc = 1, nhc%nhc_len-1
               DO n = 1, nhc%loc_num_nhc
                  IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
                  scale = EXP(-0.125_dp*nhc%nvt(inhc+1, n)%v*nhc%dt_yosh(iyosh)*nhc%dt_fact)
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v*scale ! scale
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v+ &
                                       nhc%nvt(inhc, n)%f*0.25_dp*nhc%dt_yosh(iyosh)*nhc%dt_fact ! shift
                  nhc%nvt(inhc, n)%v = nhc%nvt(inhc, n)%v*scale ! scale
               END DO

               ! updating the forces on all the thermostats
               DO n = 1, nhc%loc_num_nhc
                  IF (nhc%nvt(1, n)%nkt == 0.0_dp) CYCLE
                  nhc%nvt(inhc+1, n)%f = (nhc%nvt(inhc, n)%mass*nhc%nvt(inhc, n)%v &
                                          *nhc%nvt(inhc, n)%v-nhc%nvt(inhc+1, n)%nkt)/nhc%nvt(inhc+1, n)%mass
               END DO
            END DO
            ! update velocity on last thermostat                             ! O1
            nhc%nvt(nhc%nhc_len, :)%v = nhc%nvt(nhc%nhc_len, :)%v+ &
                                        nhc%nvt(nhc%nhc_len, :)%f*0.25_dp*nhc%dt_yosh(iyosh)*nhc%dt_fact
         END DO YOSH
      END DO NCLOOP
   END SUBROUTINE multiple_step_yoshida

END MODULE extended_system_dynamics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for calculating a complex matrix exponential with dbcsr matrices.
!>        Based on the code in matrix_exp.F from Florian Schiffmann
!> \author Samuel Andermatt (02.14)
! **************************************************************************************************

MODULE ls_matrix_exp

   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix, cp_dbcsr_filter, cp_dbcsr_frobenius_norm, cp_dbcsr_init, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_transposed, &
        cp_dbcsr_type, dbcsr_type_complex_8
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ls_matrix_exp'

   PUBLIC :: taylor_only_imaginary_dbcsr, &
             taylor_full_complex_dbcsr, &
             cp_complex_dbcsr_gemm_3, &
             bch_expansion_imaginary_propagator, &
             bch_expansion_complex_propagator

CONTAINS

! **************************************************************************************************
!> \brief Convenience function. Computes the matrix multiplications needed
!>        for the multiplication of complex sparse matrices.
!>        C = beta * C + alpha * ( A  ** transa ) * ( B ** transb )
!> \param transa : 'N' -> normal   'T' -> transpose
!>      alpha,beta :: can be 0.0_dp and 1.0_dp
!> \param transb ...
!> \param alpha ...
!> \param A_re m x k matrix ( ! for transa = 'N'), real part
!> \param A_im m x k matrix ( ! for transa = 'N'), imaginary part
!> \param B_re k x n matrix ( ! for transb = 'N'), real part
!> \param B_im k x n matrix ( ! for transb = 'N'), imaginary part
!> \param beta ...
!> \param C_re m x n matrix, real part
!> \param C_im m x n matrix, imaginary part
!> \param filter_eps ...
!> \author Samuel Andermatt
!> \note
!>      C should have no overlap with A, B
!>      This subroutine uses three real matrix multiplications instead of two complex
!>      This reduces the amount of flops and memory bandwidth by 25%, but for memory bandwidth
!>      true complex algebra is still superior (one third less bandwidth needed)
!>      limited cases matrix multiplications
! **************************************************************************************************

   SUBROUTINE cp_complex_dbcsr_gemm_3(transa, transb, alpha, A_re, A_im, &
                                      B_re, B_im, beta, C_re, C_im, filter_eps)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: A_re, A_im, B_re, B_im
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: C_re, C_im
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: filter_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_complex_dbcsr_gemm_3', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      CHARACTER(LEN=1)                                   :: transa2, transb2
      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: alpha2, alpha3, alpha4
      TYPE(cp_dbcsr_type), POINTER                       :: a_plus_b, ac, bd, c_plus_d

      CALL timeset(routineN, handle)
      !A complex matrix matrix multiplication can be done with only three multiplications
      !(a+ib)*(c+id)=ac-bd+i((a+b)*(c+d) - ac - bd)
      !A_re=a, A_im=b, B_re=c, B_im=d

      alpha2 = -alpha
      alpha3 = alpha
      alpha4 = alpha

      IF (transa == "C") THEN
         alpha2 = -alpha2
         alpha3 = -alpha3
         transa2 = "T"
      ELSE
         transa2 = transa
      END IF
      IF (transb == "C") THEN
         alpha2 = -alpha2
         alpha4 = -alpha4
         transb2 = "T"
      ELSE
         transb2 = transb
      END IF

      !create the work matrices
      NULLIFY (ac)
      ALLOCATE (ac)
      CALL cp_dbcsr_init(ac)
      CALL cp_dbcsr_create(ac, template=A_re, matrix_type="N")
      NULLIFY (bd)
      ALLOCATE (bd)
      CALL cp_dbcsr_init(bd)
      CALL cp_dbcsr_create(bd, template=A_re, matrix_type="N")
      NULLIFY (a_plus_b)
      ALLOCATE (a_plus_b)
      CALL cp_dbcsr_init(a_plus_b)
      CALL cp_dbcsr_create(a_plus_b, template=A_re, matrix_type="N")
      NULLIFY (c_plus_d)
      ALLOCATE (c_plus_d)
      CALL cp_dbcsr_init(c_plus_d)
      CALL cp_dbcsr_create(c_plus_d, template=A_re, matrix_type="N")

      !Do the neccesarry multiplications
      CALL cp_dbcsr_multiply(transa2, transb2, alpha, A_re, B_re, zero, ac, filter_eps=filter_eps)
      CALL cp_dbcsr_multiply(transa2, transb2, alpha2, A_im, B_im, zero, bd, filter_eps=filter_eps)

      CALL cp_dbcsr_add(a_plus_b, A_re, zero, alpha)
      CALL cp_dbcsr_add(a_plus_b, A_im, one, alpha3)
      CALL cp_dbcsr_add(c_plus_d, B_re, zero, alpha)
      CALL cp_dbcsr_add(c_plus_d, B_im, one, alpha4)

      !this can already be written into C_im
      !now both matrixes have been scaled which means we currently multiplied by alpha squared
      CALL cp_dbcsr_multiply(transa2, transb2, one/alpha, a_plus_b, c_plus_d, beta, C_im, filter_eps=filter_eps)

      !now add up all the terms into the result
      CALL cp_dbcsr_add(C_re, ac, beta, one)
      !the minus sign was already taken care of at the definition of alpha2
      CALL cp_dbcsr_add(C_re, bd, one, one)
      CALL cp_dbcsr_filter(C_re, filter_eps)

      CALL cp_dbcsr_add(C_im, ac, one, -one)
      !the minus sign was already taken care of at the definition of alpha2
      CALL cp_dbcsr_add(C_im, bd, one, one)
      CALL cp_dbcsr_filter(C_im, filter_eps)

      !Deallocate the work matrices
      CALL cp_dbcsr_deallocate_matrix(ac)
      CALL cp_dbcsr_deallocate_matrix(bd)
      CALL cp_dbcsr_deallocate_matrix(a_plus_b)
      CALL cp_dbcsr_deallocate_matrix(c_plus_d)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief specialized subroutine for purely imaginary matrix exponentials
!> \param exp_H ...
!> \param im_matrix ...
!> \param nsquare ...
!> \param ntaylor ...
!> \param filter_eps ...
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************

   SUBROUTINE taylor_only_imaginary_dbcsr(exp_H, im_matrix, nsquare, ntaylor, filter_eps)

      TYPE(cp_dbcsr_p_type), DIMENSION(2)                :: exp_H
      TYPE(cp_dbcsr_type), POINTER                       :: im_matrix
      INTEGER, INTENT(in)                                :: nsquare, ntaylor
      REAL(KIND=dp), INTENT(in)                          :: filter_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'taylor_only_imaginary_dbcsr', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, nloop
      REAL(KIND=dp)                                      :: square_fac, Tfac, tmp
      TYPE(cp_dbcsr_type), POINTER                       :: T1, T2, Tres_im, Tres_re

      CALL timeset(routineN, handle)

      !The divider that comes from the scaling and squaring procedure
      square_fac = 1.0_dp/(2.0_dp**REAL(nsquare, dp))

      !Allocate work matrices
      NULLIFY (T1)
      ALLOCATE (T1)
      CALL cp_dbcsr_init(T1)
      CALL cp_dbcsr_create(T1, template=im_matrix, matrix_type="N")
      NULLIFY (T2)
      ALLOCATE (T2)
      CALL cp_dbcsr_init(T2)
      CALL cp_dbcsr_create(T2, template=im_matrix, matrix_type="N")
      NULLIFY (Tres_re)
      ALLOCATE (Tres_re)
      CALL cp_dbcsr_init(Tres_re)
      CALL cp_dbcsr_create(Tres_re, template=im_matrix, matrix_type="N")
      NULLIFY (Tres_im)
      ALLOCATE (Tres_im)
      CALL cp_dbcsr_init(Tres_im)
      CALL cp_dbcsr_create(Tres_im, template=im_matrix, matrix_type="N")

      !Create the unit matrices
      CALL cp_dbcsr_set(T1, zero)
      CALL cp_dbcsr_add_on_diag(T1, one)
      CALL cp_dbcsr_set(Tres_re, zero)
      CALL cp_dbcsr_add_on_diag(Tres_re, one)
      CALL cp_dbcsr_set(Tres_im, zero)

      nloop = CEILING(REAL(ntaylor, dp)/2.0_dp)
      !the inverse of the prefactor in the taylor series
      tmp = 1.0_dp
      DO i = 1, nloop
         CALL cp_dbcsr_scale(T1, 1.0_dp/(REAL(i, dp)*2.0_dp-1.0_dp))
         CALL cp_dbcsr_filter(T1, filter_eps)
         CALL cp_dbcsr_multiply("N", "N", square_fac, im_matrix, T1, zero, &
                                T2, filter_eps=filter_eps)
         Tfac = one
         IF (MOD(i, 2) == 0) Tfac = -Tfac
         CALL cp_dbcsr_add(Tres_im, T2, one, Tfac)
         CALL cp_dbcsr_scale(T2, 1.0_dp/(REAL(i, dp)*2.0_dp))
         CALL cp_dbcsr_filter(T2, filter_eps)
         CALL cp_dbcsr_multiply("N", "N", square_fac, im_matrix, T2, zero, &
                                T1, filter_eps=filter_eps)
         Tfac = one
         IF (MOD(i, 2) == 1) Tfac = -Tfac
         CALL cp_dbcsr_add(Tres_re, T1, one, Tfac)
      END DO

      !Square the matrices, due to the scaling and squaring procedure
      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_complex_dbcsr_gemm_3("N", "N", one, Tres_re, Tres_im, &
                                         Tres_re, Tres_im, zero, exp_H(1)%matrix, exp_H(2)%matrix, &
                                         filter_eps=filter_eps)
            CALL cp_dbcsr_copy(Tres_re, exp_H(1)%matrix)
            CALL cp_dbcsr_copy(Tres_im, exp_H(2)%matrix)
         END DO
      ELSE
         CALL cp_dbcsr_copy(exp_H(1)%matrix, Tres_re)
         CALL cp_dbcsr_copy(exp_H(2)%matrix, Tres_im)
      END IF
      CALL cp_dbcsr_deallocate_matrix(T1)
      CALL cp_dbcsr_deallocate_matrix(T2)
      CALL cp_dbcsr_deallocate_matrix(Tres_re)
      CALL cp_dbcsr_deallocate_matrix(Tres_im)

      CALL timestop(handle)

   END SUBROUTINE taylor_only_imaginary_dbcsr

! **************************************************************************************************
!> \brief subroutine for general complex matrix exponentials
!>        on input a separate cp_fm_type for real and complex part
!>        on output a size 2 cp_fm_p_type, first element is the real part of
!>        the exponential second the imaginary
!> \param exp_H ...
!> \param re_part ...
!> \param im_part ...
!> \param nsquare ...
!> \param ntaylor ...
!> \param filter_eps ...
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************
   SUBROUTINE taylor_full_complex_dbcsr(exp_H, re_part, im_part, nsquare, ntaylor, filter_eps)
      TYPE(cp_dbcsr_p_type), DIMENSION(2)                :: exp_H
      TYPE(cp_dbcsr_type), POINTER                       :: re_part, im_part
      INTEGER, INTENT(in)                                :: nsquare, ntaylor
      REAL(KIND=dp), INTENT(in)                          :: filter_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'taylor_full_complex_dbcsr', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp)                                   :: square_fac
      INTEGER                                            :: handle, i
      TYPE(cp_dbcsr_type), POINTER                       :: T1, T2, T3, Tres

      CALL timeset(routineN, handle)

      !The divider that comes from the scaling and squaring procedure
      square_fac = CMPLX(1.0_dp/(2.0_dp**REAL(nsquare, dp)), 0.0_dp, KIND=dp)

      !Allocate work matrices
      NULLIFY (T1)
      ALLOCATE (T1)
      CALL cp_dbcsr_init(T1)
      CALL cp_dbcsr_create(T1, template=re_part, matrix_type="N", &
                           data_type=dbcsr_type_complex_8)
      NULLIFY (T2)
      ALLOCATE (T2)
      CALL cp_dbcsr_init(T2)
      CALL cp_dbcsr_create(T2, template=re_part, matrix_type="N", &
                           data_type=dbcsr_type_complex_8)
      NULLIFY (T3)
      ALLOCATE (T3)
      CALL cp_dbcsr_init(T3)
      CALL cp_dbcsr_create(T3, template=re_part, matrix_type="N", &
                           data_type=dbcsr_type_complex_8)
      NULLIFY (Tres)
      ALLOCATE (Tres)
      CALL cp_dbcsr_init(Tres)
      CALL cp_dbcsr_create(Tres, template=re_part, matrix_type="N", &
                           data_type=dbcsr_type_complex_8)

      !Fuse the input matrices to a single complex matrix
      CALL cp_dbcsr_copy(T3, re_part)
      CALL cp_dbcsr_copy(Tres, im_part) !will later on be set back to zero
      CALL cp_dbcsr_scale(Tres, CMPLX(0.0_dp, 1.0_dp, KIND=dp))
      CALL cp_dbcsr_add(T3, Tres, one, one)

      !Create the unit matrices
      CALL cp_dbcsr_set(T1, zero)
      CALL cp_dbcsr_add_on_diag(T1, 1.0_dp)
      CALL cp_dbcsr_set(Tres, zero)
      CALL cp_dbcsr_add_on_diag(Tres, 1.0_dp)

      DO i = 1, ntaylor
         CALL cp_dbcsr_scale(T1, one/CMPLX(i*1.0_dp, 0.0_dp, KIND=dp))
         CALL cp_dbcsr_filter(T1, filter_eps)
         CALL cp_dbcsr_multiply("N", "N", square_fac, T1, T3, &
                                zero, T2, filter_eps=filter_eps)
         CALL cp_dbcsr_add(Tres, T2, one, one)
         CALL cp_dbcsr_copy(T1, T2)
      END DO

      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_dbcsr_multiply("N", "N", one, Tres, Tres, zero, &
                                   T2, filter_eps=filter_eps)
            CALL cp_dbcsr_copy(Tres, T2)
         END DO
      END IF

      CALL cp_dbcsr_copy(exp_H(1)%matrix, Tres, keep_imaginary=.FALSE.)
      CALL cp_dbcsr_scale(Tres, CMPLX(0.0_dp, -1.0_dp, KIND=dp))
      CALL cp_dbcsr_copy(exp_H(2)%matrix, Tres, keep_imaginary=.FALSE.)

      CALL cp_dbcsr_deallocate_matrix(T1)
      CALL cp_dbcsr_deallocate_matrix(T2)
      CALL cp_dbcsr_deallocate_matrix(T3)
      CALL cp_dbcsr_deallocate_matrix(Tres)

      CALL timestop(handle)

   END SUBROUTINE taylor_full_complex_dbcsr

! **************************************************************************************************
!> \brief  The Baker-Campbell-Hausdorff expansion for a purely imaginary exponent (e.g. rtp)
!>         Works for a non unitary propagator, because the density matrix is hermitian
!> \param propagator The exponent of the matrix exponential
!> \param density_re Real part of the density matrix
!> \param density_im Imaginary part of the density matrix
!> \param filter_eps The filtering threshold for all matrices
!> \param filter_eps_small ...
!> \param eps_exp The accuracy of the exponential
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************

   SUBROUTINE bch_expansion_imaginary_propagator(propagator, density_re, density_im, filter_eps, filter_eps_small, eps_exp)
      TYPE(cp_dbcsr_type), POINTER                       :: propagator, density_re, density_im
      REAL(KIND=dp), INTENT(in)                          :: filter_eps, filter_eps_small, eps_exp

      CHARACTER(len=*), PARAMETER :: routineN = 'bch_expansion_imaginary_propagator', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, unit_nr
      LOGICAL                                            :: convergence
      REAL(KIND=dp)                                      :: alpha, max_alpha, prefac
      TYPE(cp_dbcsr_type), POINTER                       :: comm, comm2, tmp, tmp2

      CALL timeset(routineN, handle)

      unit_nr = cp_logger_get_default_io_unit()

      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=propagator)
      NULLIFY (tmp2)
      ALLOCATE (tmp2)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=propagator)
      NULLIFY (comm)
      ALLOCATE (comm)
      CALL cp_dbcsr_init(comm)
      CALL cp_dbcsr_create(comm, template=propagator)
      NULLIFY (comm2)
      ALLOCATE (comm2)
      CALL cp_dbcsr_init(comm2)
      CALL cp_dbcsr_create(comm2, template=propagator)

      CALL cp_dbcsr_copy(tmp, density_re)
      CALL cp_dbcsr_copy(tmp2, density_im)

      convergence = .FALSE.
      DO i = 1, 20
         prefac = one/i
         CALL cp_dbcsr_multiply("N", "N", -prefac, propagator, tmp2, zero, comm, &
                                filter_eps=filter_eps_small)
         CALL cp_dbcsr_multiply("N", "N", prefac, propagator, tmp, zero, comm2, &
                                filter_eps=filter_eps_small)
         CALL cp_dbcsr_transposed(tmp, comm)
         CALL cp_dbcsr_transposed(tmp2, comm2)
         CALL cp_dbcsr_add(comm, tmp, one, one)
         CALL cp_dbcsr_add(comm2, tmp2, one, -one)
         CALL cp_dbcsr_add(density_re, comm, one, one)
         CALL cp_dbcsr_add(density_im, comm2, one, one)
         CALL cp_dbcsr_copy(tmp, comm)
         CALL cp_dbcsr_copy(tmp2, comm2)
         !check for convergence
         max_alpha = zero
         alpha = cp_dbcsr_frobenius_norm(comm)
         IF (alpha > max_alpha) max_alpha = alpha
         alpha = cp_dbcsr_frobenius_norm(comm2)
         IF (alpha > max_alpha) max_alpha = alpha
         IF (max_alpha < eps_exp) convergence = .TRUE.
         IF (convergence) THEN
            IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="((T3,A,I2,A))") &
               "BCH converged after ", i, " steps"
            EXIT
         ENDIF
      ENDDO

      CALL cp_dbcsr_filter(density_re, filter_eps)
      CALL cp_dbcsr_filter(density_im, filter_eps)

      IF (.NOT. convergence) &
         CALL cp__w("ls_matrix_exp.F",432,"BCH method did not converge")

      CALL cp_dbcsr_deallocate_matrix(tmp)
      CALL cp_dbcsr_deallocate_matrix(tmp2)
      CALL cp_dbcsr_deallocate_matrix(comm)
      CALL cp_dbcsr_deallocate_matrix(comm2)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief  The Baker-Campbell-Hausdorff expansion for a complex exponent (e.g. rtp)
!>         Works for a non unitary propagator, because the density matrix is hermitian
!> \param propagator_re Real part of the exponent
!> \param propagator_im Imaginary part of the exponent
!> \param density_re Real part of the density matrix
!> \param density_im Imaginary part of the density matrix
!> \param filter_eps The filtering threshold for all matrices
!> \param filter_eps_small ...
!> \param eps_exp The accuracy of the exponential
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************

   SUBROUTINE bch_expansion_complex_propagator(propagator_re, propagator_im, density_re, density_im, filter_eps, &
                                               filter_eps_small, eps_exp)
      TYPE(cp_dbcsr_type), POINTER                       :: propagator_re, propagator_im, &
                                                            density_re, density_im
      REAL(KIND=dp), INTENT(in)                          :: filter_eps, filter_eps_small, eps_exp

      CHARACTER(len=*), PARAMETER :: routineN = 'bch_expansion_complex_propagator', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, unit_nr
      LOGICAL                                            :: convergence
      REAL(KIND=dp)                                      :: alpha, max_alpha, prefac
      TYPE(cp_dbcsr_type), POINTER                       :: comm, comm2, tmp, tmp2

      CALL timeset(routineN, handle)

      unit_nr = cp_logger_get_default_io_unit()

      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=propagator_re)
      NULLIFY (tmp2)
      ALLOCATE (tmp2)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=propagator_re)
      NULLIFY (comm)
      ALLOCATE (comm)
      CALL cp_dbcsr_init(comm)
      CALL cp_dbcsr_create(comm, template=propagator_re)
      NULLIFY (comm2)
      ALLOCATE (comm2)
      CALL cp_dbcsr_init(comm2)
      CALL cp_dbcsr_create(comm2, template=propagator_re)

      CALL cp_dbcsr_copy(tmp, density_re)
      CALL cp_dbcsr_copy(tmp2, density_im)

      convergence = .FALSE.

      DO i = 1, 20
         prefac = one/i
         CALL cp_complex_dbcsr_gemm_3("N", "N", prefac, propagator_re, propagator_im, &
                                      tmp, tmp2, zero, comm, comm2, filter_eps=filter_eps_small)
         CALL cp_dbcsr_transposed(tmp, comm)
         CALL cp_dbcsr_transposed(tmp2, comm2)
         CALL cp_dbcsr_add(comm, tmp, one, one)
         CALL cp_dbcsr_add(comm2, tmp2, one, -one)
         CALL cp_dbcsr_add(density_re, comm, one, one)
         CALL cp_dbcsr_add(density_im, comm2, one, one)
         CALL cp_dbcsr_copy(tmp, comm)
         CALL cp_dbcsr_copy(tmp2, comm2)
         !check for convergence
         max_alpha = zero
         alpha = cp_dbcsr_frobenius_norm(comm)
         IF (alpha > max_alpha) max_alpha = alpha
         alpha = cp_dbcsr_frobenius_norm(comm2)
         IF (alpha > max_alpha) max_alpha = alpha
         IF (max_alpha < eps_exp) convergence = .TRUE.
         IF (convergence) THEN
            IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="((T3,A,I2,A))") &
               "BCH converged after ", i, " steps"
            EXIT
         ENDIF
      ENDDO

      CALL cp_dbcsr_filter(density_re, filter_eps)
      CALL cp_dbcsr_filter(density_im, filter_eps)

      IF (.NOT. convergence) &
         CALL cp__w("ls_matrix_exp.F",527,"BCH method did not converge ")

      CALL cp_dbcsr_deallocate_matrix(tmp)
      CALL cp_dbcsr_deallocate_matrix(tmp2)
      CALL cp_dbcsr_deallocate_matrix(comm)
      CALL cp_dbcsr_deallocate_matrix(comm2)

      CALL timestop(handle)

   END SUBROUTINE

END MODULE ls_matrix_exp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for calculating a complex matrix exponential.
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE rt_make_propagators

   USE cp_control_types,                ONLY: rtp_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_create,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_scale,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_to_fm
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE input_constants,                 ONLY: do_etrs,&
                                              do_pade,&
                                              do_taylor
   USE kinds,                           ONLY: dp
   USE ls_matrix_exp,                   ONLY: bch_expansion_complex_propagator,&
                                              bch_expansion_imaginary_propagator,&
                                              cp_complex_dbcsr_gemm_3,&
                                              taylor_full_complex_dbcsr,&
                                              taylor_only_imaginary_dbcsr
   USE matrix_exp,                      ONLY: arnoldi,&
                                              exp_pade_full_complex,&
                                              exp_pade_only_imaginary,&
                                              taylor_full_complex,&
                                              taylor_only_imaginary
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_make_propagators'

   PUBLIC :: propagate_exp, &
             propagate_arnoldi, &
             compute_exponential, &
             compute_exponential_sparse, &
             propagate_exp_density, &
             propagate_bch

CONTAINS
! **************************************************************************************************
!> \brief performs propagations if explicit matrix exponentials are used
!>        ETRS:  exp(i*H(t+dt)*dt/2)*exp(i*H(t)*dt/2)*MOS
!>        EM:    exp[-idt/2H(t+dt/2)*MOS
!> \param rtp ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE propagate_exp(rtp, rtp_control)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'propagate_exp', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, nmo, re
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_new, exp_H_old, propagator_matrix
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_next, mos_old

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, propagator_matrix=propagator_matrix, mos_old=mos_old, mos_new=mos_new, &
                   mos_next=mos_next, exp_H_new=exp_H_new, exp_H_old=exp_H_old)

      ! Only compute exponential if a new propagator matrix is available
      CALL compute_exponential(exp_H_new, propagator_matrix, rtp_control, rtp)

      DO i = 1, SIZE(mos_new)/2
         re = 2*i-1
         im = 2*i

         CALL cp_fm_get_info(mos_new(re)%matrix, ncol_global=nmo)
         !Save some work by computing the first half of the propagation only once in case of ETRS
         !For EM this matrix has to be the initial matrix, thus a copy is enough
         IF (rtp%iter == 1) THEN
            IF (rtp_control%propagator == do_etrs) THEN
               CALL cp_dbcsr_sm_fm_multiply(exp_H_old(re)%matrix, mos_old(re)%matrix, &
                                            mos_next(re)%matrix, nmo, alpha=one, beta=zero)
               CALL cp_dbcsr_sm_fm_multiply(exp_H_old(im)%matrix, mos_old(im)%matrix, &
                                            mos_next(re)%matrix, nmo, alpha=-one, beta=one)
               CALL cp_dbcsr_sm_fm_multiply(exp_H_old(re)%matrix, mos_old(im)%matrix, &
                                            mos_next(im)%matrix, nmo, alpha=one, beta=zero)
               CALL cp_dbcsr_sm_fm_multiply(exp_H_old(im)%matrix, mos_old(re)%matrix, &
                                            mos_next(im)%matrix, nmo, alpha=one, beta=one)
            ELSE
               CALL cp_fm_to_fm(mos_old(re)%matrix, mos_next(re)%matrix)
               CALL cp_fm_to_fm(mos_old(im)%matrix, mos_next(im)%matrix)
            END IF
         END IF
         CALL cp_dbcsr_sm_fm_multiply(exp_H_new(re)%matrix, mos_next(re)%matrix, &
                                      mos_new(re)%matrix, nmo, alpha=one, beta=zero)
         CALL cp_dbcsr_sm_fm_multiply(exp_H_new(im)%matrix, mos_next(im)%matrix, &
                                      mos_new(re)%matrix, nmo, alpha=-one, beta=one)
         CALL cp_dbcsr_sm_fm_multiply(exp_H_new(re)%matrix, mos_next(im)%matrix, &
                                      mos_new(im)%matrix, nmo, alpha=one, beta=zero)
         CALL cp_dbcsr_sm_fm_multiply(exp_H_new(im)%matrix, mos_next(re)%matrix, &
                                      mos_new(im)%matrix, nmo, alpha=one, beta=one)
      END DO

      CALL timestop(handle)

   END SUBROUTINE propagate_exp

! **************************************************************************************************
!> \brief Propagation of the density matrix instead of the atomic orbitals
!>        via a matrix exponential
!> \param rtp ...
!> \param rtp_control ...
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************

   SUBROUTINE propagate_exp_density(rtp, rtp_control)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'propagate_exp_density', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, re
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_new, exp_H_old, propagator_matrix, &
                                                            rho_new, rho_next, rho_old
      TYPE(cp_dbcsr_type), POINTER                       :: tmp_im, tmp_re

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, propagator_matrix=propagator_matrix, exp_H_new=exp_H_new, &
                   exp_H_old=exp_H_old, rho_old=rho_old, rho_new=rho_new, rho_next=rho_next)

      CALL compute_exponential_sparse(exp_H_new, propagator_matrix, rtp_control, rtp)

      !I could store these matrices in the type
      NULLIFY (tmp_re)
      ALLOCATE (tmp_re)
      CALL cp_dbcsr_init(tmp_re)
      CALL cp_dbcsr_create(tmp_re, template=propagator_matrix(1)%matrix, matrix_type="N")
      NULLIFY (tmp_im)
      ALLOCATE (tmp_im)
      CALL cp_dbcsr_init(tmp_im)
      CALL cp_dbcsr_create(tmp_im, template=propagator_matrix(1)%matrix, matrix_type="N")

      DO i = 1, SIZE(exp_H_new)/2
         re = 2*i-1
         im = 2*i
         !Save some work by computing the first half of the propagation only once in case of ETRS
         !For EM this matrix has to be the initial matrix, thus a copy is enough
         IF (rtp%iter == 1) THEN
            IF (rtp_control%propagator == do_etrs) THEN
               CALL cp_complex_dbcsr_gemm_3("N", "N", one, exp_H_old(re)%matrix, exp_H_old(im)%matrix, &
                                            rho_old(re)%matrix, rho_old(im)%matrix, zero, tmp_re, tmp_im, filter_eps=rtp%filter_eps)
               CALL cp_complex_dbcsr_gemm_3("N", "C", one, tmp_re, tmp_im, exp_H_old(re)%matrix, exp_H_old(im)%matrix, &
                                            zero, rho_next(re)%matrix, rho_next(im)%matrix, filter_eps=rtp%filter_eps)
            ELSE
               CALL cp_dbcsr_copy(rho_next(re)%matrix, rho_old(re)%matrix)
               CALL cp_dbcsr_copy(rho_next(im)%matrix, rho_old(im)%matrix)
            ENDIF
         END IF
         CALL cp_complex_dbcsr_gemm_3("N", "N", one, exp_H_new(re)%matrix, exp_H_new(im)%matrix, &
                                      rho_next(re)%matrix, rho_next(im)%matrix, zero, tmp_re, tmp_im, filter_eps=rtp%filter_eps)
         CALL cp_complex_dbcsr_gemm_3("N", "C", one, tmp_re, tmp_im, exp_H_new(re)%matrix, exp_H_new(im)%matrix, &
                                      zero, rho_new(re)%matrix, rho_new(im)%matrix, filter_eps=rtp%filter_eps)
      END DO

      CALL cp_dbcsr_deallocate_matrix(tmp_re)
      CALL cp_dbcsr_deallocate_matrix(tmp_im)

      CALL timestop(handle)

   END SUBROUTINE propagate_exp_density

! **************************************************************************************************
!> \brief computes U_prop*MOs using arnoldi subspace algorithm
!> \param rtp ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE propagate_arnoldi(rtp, rtp_control)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'propagate_arnoldi', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, im, ispin, nspin, re
      REAL(dp)                                           :: eps_arnoldi, t
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: propagator_matrix
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_next, mos_old, &
                                                            propagator_matrix_fm

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, dt=t, mos_new=mos_new, mos_old=mos_old, &
                   mos_next=mos_next, propagator_matrix=propagator_matrix)

      nspin = SIZE(mos_new)/2
      eps_arnoldi = rtp_control%eps_exp
      ! except for the first step the further propagated mos_next
      ! must be copied on mos_old so that we have the half propagated mos
      ! ready on mos_old and only need to perform the second half propagatioon
      IF (rtp_control%propagator == do_etrs .AND. rtp%iter == 1) THEN
         DO i = 1, SIZE(mos_new)
            CALL cp_fm_to_fm(mos_next(i)%matrix, mos_old(i)%matrix)
         END DO
      END IF

      NULLIFY (propagator_matrix_fm)
      ALLOCATE (propagator_matrix_fm(SIZE(propagator_matrix)))
      DO i = 1, SIZE(propagator_matrix)
         CALL cp_fm_create(propagator_matrix_fm(i)%matrix, &
                           matrix_struct=rtp%ao_ao_fmstruct, &
                           name="prop_fm")
         CALL copy_dbcsr_to_fm(propagator_matrix(i)%matrix, propagator_matrix_fm(i)%matrix)
      END DO

      DO ispin = 1, nspin
         re = ispin*2-1
         im = ispin*2
         IF (rtp_control%fixed_ions .AND. .NOT. rtp%do_hfx) THEN
            CALL arnoldi(mos_old(re:im), mos_new(re:im), &
                         eps_arnoldi, Him=propagator_matrix_fm(im)%matrix, &
                         mos_next=mos_next(re:im), narn_old=rtp%narn_old)
         ELSE
            CALL arnoldi(mos_old(re:im), mos_new(re:im), &
                         eps_arnoldi, Hre=propagator_matrix_fm(re)%matrix, &
                         Him=propagator_matrix_fm(im)%matrix, mos_next=mos_next(re:im), &
                         narn_old=rtp%narn_old)
         END IF
      END DO

!    DO i=1,SIZE(propagator_matrix)
!         CALL copy_fm_to_dbcsr(propagator_matrix_fm(i)%matrix,propagator_matrix(i)%matrix)
!    END DO
      CALL cp_fm_vect_dealloc(propagator_matrix_fm)

      CALL timestop(handle)

   END SUBROUTINE propagate_arnoldi

! **************************************************************************************************
!> \brief  Propagation using the Baker-Campbell-Hausdorff expansion,
!>         currently only works for rtp
!> \param rtp ...
!> \param rtp_control ...
!> \author Samuel Andermatt (02.2014)
! **************************************************************************************************

   SUBROUTINE propagate_bch(rtp, rtp_control)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'propagate_bch', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, im, ispin, re
      REAL(dp)                                           :: dt
      REAL(KIND=dp)                                      :: prefac
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_old, propagator_matrix, rho_new, &
                                                            rho_next, rho_old

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, propagator_matrix=propagator_matrix, rho_old=rho_old, rho_new=rho_new, &
                   rho_next=rho_next)

      DO ispin = 1, SIZE(propagator_matrix)/2
         re = 2*ispin-1
         im = 2*ispin

         IF (rtp%iter == 1) THEN
            ! For EM I have to copy rho_old onto rho_next and for ETRS,
            ! this is the first term of the series of commutators that result in rho_next
            CALL cp_dbcsr_copy(rho_next(re)%matrix, rho_old(re)%matrix)
            CALL cp_dbcsr_copy(rho_next(im)%matrix, rho_old(im)%matrix)
            IF (rtp_control%propagator == do_etrs) THEN
               !since we never calculated the matrix exponential the old matrix exponential stores the unscalled propagator
               CALL get_rtp(rtp=rtp, exp_H_old=exp_H_old, dt=dt)
               prefac = -0.5_dp*dt
               CALL cp_dbcsr_scale(exp_H_old(im)%matrix, prefac)
               IF (.NOT. rtp%do_hfx .AND. rtp_control%fixed_ions) THEN
                  CALL bch_expansion_imaginary_propagator( &
                     exp_H_old(im)%matrix, rho_next(re)%matrix, rho_next(im)%matrix, &
                     rtp%filter_eps, rtp%filter_eps_small, rtp_control%eps_exp)
               ELSE
                  CALL cp_dbcsr_scale(exp_H_old(re)%matrix, prefac)
                  CALL bch_expansion_complex_propagator( &
                     exp_H_old(re)%matrix, exp_H_old(im)%matrix, rho_next(re)%matrix, rho_next(im)%matrix, &
                     rtp%filter_eps, rtp%filter_eps_small, rtp_control%eps_exp)
               ENDIF
            END IF
         END IF
         CALL cp_dbcsr_copy(rho_new(re)%matrix, rho_next(re)%matrix)
         CALL cp_dbcsr_copy(rho_new(im)%matrix, rho_next(im)%matrix)
         IF (.NOT. rtp%do_hfx .AND. rtp_control%fixed_ions) THEN
            CALL bch_expansion_imaginary_propagator( &
               propagator_matrix(im)%matrix, rho_new(re)%matrix, rho_new(im)%matrix, &
               rtp%filter_eps, rtp%filter_eps_small, rtp_control%eps_exp)
         ELSE
            CALL bch_expansion_complex_propagator( &
               propagator_matrix(re)%matrix, propagator_matrix(im)%matrix, rho_new(re)%matrix, rho_new(im)%matrix, &
               rtp%filter_eps, rtp%filter_eps_small, rtp_control%eps_exp)
         ENDIF

      END DO

      CALL timestop(handle)

   END SUBROUTINE propagate_bch

! **************************************************************************************************
!> \brief decides which type of exponential has to be computed
!> \param propagator ...
!> \param propagator_matrix ...
!> \param rtp_control ...
!> \param rtp ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE compute_exponential(propagator, propagator_matrix, rtp_control, rtp)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: propagator, propagator_matrix
      TYPE(rtp_control_type), POINTER                    :: rtp_control
      TYPE(rt_prop_type), POINTER                        :: rtp

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_exponential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, im, ispin, re
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: propagator_fm, propagator_matrix_fm

      NULLIFY (propagator_fm)
      ALLOCATE (propagator_fm(SIZE(propagator)))
      NULLIFY (propagator_matrix_fm)
      ALLOCATE (propagator_matrix_fm(SIZE(propagator_matrix)))
      DO i = 1, SIZE(propagator)
         CALL cp_fm_create(propagator_fm(i)%matrix, &
                           matrix_struct=rtp%ao_ao_fmstruct, &
                           name="prop_fm")
         CALL copy_dbcsr_to_fm(propagator(i)%matrix, propagator_fm(i)%matrix)
         CALL cp_fm_create(propagator_matrix_fm(i)%matrix, &
                           matrix_struct=rtp%ao_ao_fmstruct, &
                           name="prop_mat_fm")
         CALL copy_dbcsr_to_fm(propagator_matrix(i)%matrix, propagator_matrix_fm(i)%matrix)
      END DO

      DO ispin = 1, SIZE(propagator)/2
         re = 2*ispin-1
         im = 2*ispin

         SELECT CASE (rtp_control%mat_exp)

         CASE (do_taylor)
            IF (rtp_control%fixed_ions .AND. .NOT. rtp%do_hfx) THEN
               CALL taylor_only_imaginary(propagator_fm(re:im), propagator_matrix_fm(im)%matrix, &
                                          rtp%orders(1, ispin), rtp%orders(2, ispin))
            ELSE
               CALL taylor_full_complex(propagator_fm(re:im), propagator_matrix_fm(re)%matrix, propagator_matrix_fm(im)%matrix, &
                                        rtp%orders(1, ispin), rtp%orders(2, ispin))
            END IF
         CASE (do_pade)
            IF (rtp_control%fixed_ions .AND. .NOT. rtp%do_hfx) THEN
               CALL exp_pade_only_imaginary(propagator_fm(re:im), propagator_matrix_fm(im)%matrix, &
                                            rtp%orders(1, ispin), rtp%orders(2, ispin))
            ELSE
               CALL exp_pade_full_complex(propagator_fm(re:im), propagator_matrix_fm(re)%matrix, propagator_matrix_fm(im)%matrix, &
                                          rtp%orders(1, ispin), rtp%orders(2, ispin))
            END IF
         END SELECT
      END DO

      DO i = 1, SIZE(propagator)
         CALL copy_fm_to_dbcsr(propagator_fm(i)%matrix, propagator(i)%matrix)
         CALL copy_fm_to_dbcsr(propagator_matrix_fm(i)%matrix, propagator_matrix(i)%matrix)
      END DO
      CALL cp_fm_vect_dealloc(propagator_fm)
      CALL cp_fm_vect_dealloc(propagator_matrix_fm)

   END SUBROUTINE compute_exponential

! **************************************************************************************************
!> \brief Sparse versions of the matrix exponentials
!> \param propagator ...
!> \param propagator_matrix ...
!> \param rtp_control ...
!> \param rtp ...
!> \author Samuel Andermatt (02.14)
! **************************************************************************************************

   SUBROUTINE compute_exponential_sparse(propagator, propagator_matrix, rtp_control, rtp)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: propagator, propagator_matrix
      TYPE(rtp_control_type), POINTER                    :: rtp_control
      TYPE(rt_prop_type), POINTER                        :: rtp

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_exponential_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, im, ispin, re

      CALL timeset(routineN, handle)

      DO ispin = 1, SIZE(propagator)/2
         re = 2*ispin-1
         im = 2*ispin
         IF (rtp_control%fixed_ions .AND. .NOT. rtp%do_hfx) THEN
            CALL taylor_only_imaginary_dbcsr(propagator(re:im), propagator_matrix(im)%matrix, &
                                             rtp%orders(1, ispin), rtp%orders(2, ispin), rtp%filter_eps)
         ELSE
            CALL taylor_full_complex_dbcsr(propagator(re:im), propagator_matrix(re)%matrix, propagator_matrix(im)%matrix, &
                                           rtp%orders(1, ispin), rtp%orders(2, ispin), rtp%filter_eps)
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE compute_exponential_sparse

END MODULE rt_make_propagators
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for propagating the orbitals
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************
MODULE rt_propagation_methods
   USE bibliography,                    ONLY: Kolafa2004,&
                                              cite_reference
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_cholesky_decompose,&
                                              cp_cfm_gemm,&
                                              cp_cfm_triangular_multiply
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_release,&
                                              cp_cfm_type
   USE cp_control_types,                ONLY: rtp_control_type
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_invert
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix, cp_dbcsr_deallocate_matrix_set, cp_dbcsr_desymmetrize, &
        cp_dbcsr_filter, cp_dbcsr_frobenius_norm, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_init_p, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, &
        cp_dbcsr_transposed, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_double,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE input_constants,                 ONLY: do_arnoldi,&
                                              do_bch,&
                                              do_em,&
                                              do_pade,&
                                              do_taylor
   USE iterate_matrix,                  ONLY: matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE ls_matrix_exp,                   ONLY: cp_complex_dbcsr_gemm_3
   USE mathlib,                         ONLY: binomial
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_energy_utils,                 ONLY: qs_energies_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE rt_make_propagators,             ONLY: propagate_arnoldi,&
                                              propagate_bch,&
                                              propagate_exp,&
                                              propagate_exp_density
   USE rt_propagation_output,           ONLY: report_density_occupation,&
                                              rt_convergence,&
                                              rt_convergence_density
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
   USE rt_propagation_utils,            ONLY: calc_S_derivs,&
                                              calc_update_rho,&
                                              calc_update_rho_sparse
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation_methods'

   PUBLIC :: propagation_step, &
             s_matrices_create, &
             calc_sinvH, &
             put_data_to_history

CONTAINS

! **************************************************************************************************
!> \brief performes a single propagation step a(t+Dt)=U(t+Dt,t)*a(0)
!>        and calculates the new exponential
!> \param qs_env ...
!> \param rtp ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE propagation_step(qs_env, rtp, rtp_control)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'propagation_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aspc_order, handle, i, im, re, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: delta_P, H_last_iter, ks_mix, ks_mix_im, &
                                                            matrix_ks, matrix_ks_im, matrix_s, &
                                                            rho_new
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: delta_mos, mos_new
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      NULLIFY (delta_P, rho_new, delta_mos, mos_new)
      NULLIFY (ks_mix, ks_mix_im)
      ! get everything needed and set some values
      CALL get_qs_env(qs_env, matrix_s=matrix_s)
      IF (rtp%iter == 1) THEN
         CALL qs_energies_init(qs_env, .FALSE.)
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         IF (.NOT. rtp_control%fixed_ions) THEN
            CALL s_matrices_create(matrix_s, rtp)
         END IF
         rtp%delta_iter = 100.0_dp
         rtp%mixing_factor = 1.0_dp
         rtp%mixing = .FALSE.
         aspc_order = rtp_control%aspc_order
         CALL aspc_extrapolate(rtp, matrix_s, aspc_order)
         IF (rtp%linear_scaling) THEN
            CALL calc_update_rho_sparse(qs_env)
         ELSE
            CALL calc_update_rho(qs_env)
         ENDIF
         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE.)
      END IF
      IF (.NOT. rtp_control%fixed_ions) THEN
         CALL calc_S_derivs(qs_env)
      END IF
      rtp%converged = .FALSE.

      IF (rtp%linear_scaling) THEN
         ! keep temporary copy of the starting density matrix to check for convergence
         CALL get_rtp(rtp=rtp, rho_new=rho_new)
         NULLIFY (delta_P)
         CALL cp_dbcsr_allocate_matrix_set(delta_P, SIZE(rho_new))
         DO i = 1, SIZE(rho_new)
            CALL cp_dbcsr_init_p(delta_P(i)%matrix)
            CALL cp_dbcsr_create(delta_P(i)%matrix, template=rho_new(i)%matrix)
            CALL cp_dbcsr_copy(delta_P(i)%matrix, rho_new(i)%matrix)
         END DO
      ELSE
         ! keep temporary copy of the starting mos to check for convergence
         CALL get_rtp(rtp=rtp, mos_new=mos_new)
         ALLOCATE (delta_mos(SIZE(mos_new)))
         DO i = 1, SIZE(mos_new)
            CALL cp_fm_create(delta_mos(i)%matrix, &
                              matrix_struct=mos_new(i)%matrix%matrix_struct, &
                              name="delta_mos"//TRIM(ADJUSTL(cp_to_string(i))))
            CALL cp_fm_to_fm(mos_new(i)%matrix, delta_mos(i)%matrix)
         END DO
      ENDIF

      CALL get_qs_env(qs_env, &
                      matrix_ks=matrix_ks, &
                      matrix_ks_im=matrix_ks_im)

      CALL get_rtp(rtp=rtp, H_last_iter=H_last_iter)
      IF (rtp%mixing) THEN
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(t3,a,2f16.8)') "Mixing the Hamiltonians to improve robustness, mixing factor: ", rtp%mixing_factor
         ENDIF
         CALL cp_dbcsr_allocate_matrix_set(ks_mix, SIZE(matrix_ks))
         CALL cp_dbcsr_allocate_matrix_set(ks_mix_im, SIZE(matrix_ks))
         DO i = 1, SIZE(matrix_ks)
            CALL cp_dbcsr_init_p(ks_mix(i)%matrix)
            CALL cp_dbcsr_create(ks_mix(i)%matrix, template=matrix_ks(1)%matrix)
            CALL cp_dbcsr_init_p(ks_mix_im(i)%matrix)
            CALL cp_dbcsr_create(ks_mix_im(i)%matrix, template=matrix_ks(1)%matrix)
         ENDDO
         DO i = 1, SIZE(matrix_ks)
            re = 2*i-1
            im = 2*i
            CALL cp_dbcsr_add(ks_mix(i)%matrix, matrix_ks(i)%matrix, 0.0_dp, rtp%mixing_factor)
            CALL cp_dbcsr_add(ks_mix(i)%matrix, H_last_iter(re)%matrix, 1.0_dp, 1.0_dp-rtp%mixing_factor)
            IF (rtp%do_hfx) THEN
               CALL cp_dbcsr_add(ks_mix_im(i)%matrix, matrix_ks_im(i)%matrix, 0.0_dp, rtp%mixing_factor)
               CALL cp_dbcsr_add(ks_mix_im(i)%matrix, H_last_iter(im)%matrix, 1.0_dp, 1.0_dp-rtp%mixing_factor)
            ENDIF
         ENDDO
         CALL calc_SinvH(rtp, ks_mix, ks_mix_im, rtp_control)
         DO i = 1, SIZE(matrix_ks)
            re = 2*i-1
            im = 2*i
            CALL cp_dbcsr_copy(H_last_iter(re)%matrix, ks_mix(i)%matrix)
            IF (rtp%do_hfx) THEN
               CALL cp_dbcsr_copy(H_last_iter(im)%matrix, ks_mix_im(i)%matrix)
            ENDIF
         ENDDO
         CALL cp_dbcsr_deallocate_matrix_set(ks_mix)
         CALL cp_dbcsr_deallocate_matrix_set(ks_mix_im)
      ELSE
         CALL calc_SinvH(rtp, matrix_ks, matrix_ks_im, rtp_control)
         DO i = 1, SIZE(matrix_ks)
            re = 2*i-1
            im = 2*i
            CALL cp_dbcsr_copy(H_last_iter(re)%matrix, matrix_ks(i)%matrix)
            IF (rtp%do_hfx) THEN
               CALL cp_dbcsr_copy(H_last_iter(im)%matrix, matrix_ks_im(i)%matrix)
            ENDIF
         ENDDO
      ENDIF

      CALL compute_propagator_matrix(rtp, rtp_control%propagator)

      SELECT CASE (rtp_control%mat_exp)
      CASE (do_pade, do_taylor)
         IF (rtp%linear_scaling) THEN
            CALL propagate_exp_density(rtp, rtp_control)
            CALL calc_update_rho_sparse(qs_env)
         ELSE
            CALL propagate_exp(rtp, rtp_control)
            CALL calc_update_rho(qs_env)
         END IF
      CASE (do_arnoldi)
         CALL propagate_arnoldi(rtp, rtp_control)
         CALL calc_update_rho(qs_env)
      CASE (do_bch)
         CALL propagate_bch(rtp, rtp_control)
         CALL calc_update_rho_sparse(qs_env)
      END SELECT
      CALL step_finalize(qs_env, rtp_control, delta_mos, delta_P)
      IF (rtp%linear_scaling) THEN
         CALL cp_dbcsr_deallocate_matrix_set(delta_P)
      ELSE
         CALL cp_fm_vect_dealloc(delta_mos)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE propagation_step

! **************************************************************************************************
!> \brief Performes all the stuff to finish the step:
!>        convergence checks
!>        copying stuff into right place for the next step
!>        updating the history for extrapolation
!> \param qs_env ...
!> \param rtp_control ...
!> \param delta_mos ...
!> \param delta_P ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE step_finalize(qs_env, rtp_control, delta_mos, delta_P)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rtp_control_type), POINTER                    :: rtp_control
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: delta_mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: delta_P

      CHARACTER(len=*), PARAMETER :: routineN = 'step_finalize', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ihist
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_new, exp_H_old, matrix_ks, &
                                                            matrix_ks_im, rho_new, rho_old, s_mat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_old
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(rt_prop_type), POINTER                        :: rtp

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, rtp=rtp, matrix_s=s_mat, matrix_ks=matrix_ks, matrix_ks_im=matrix_ks_im, energy=energy)
      CALL get_rtp(rtp=rtp, exp_H_old=exp_H_old, exp_H_new=exp_H_new)

      IF (rtp_control%sc_check_start .LT. rtp%iter) THEN
         rtp%delta_iter_old = rtp%delta_iter
         IF (rtp%linear_scaling) THEN
            CALL rt_convergence_density(rtp, delta_P, rtp%delta_iter)
         ELSE
            CALL rt_convergence(rtp, s_mat(1)%matrix, delta_mos, rtp%delta_iter)
         END IF
         rtp%converged = (rtp%delta_iter .LT. rtp_control%eps_ener)
         !Apply mixing if scf loop is not converging

         !It would be better to redo the current step with mixixng,
         !but currently the decision is made to use mixing from the next step on
         IF (rtp_control%sc_check_start .LT. rtp%iter+1) THEN
            IF (rtp%delta_iter/rtp%delta_iter_old > 0.9) THEN
               rtp%mixing_factor = MAX(rtp%mixing_factor/2.0_dp, 0.125_dp)
               rtp%mixing = .TRUE.
            ENDIF
         ENDIF
      END IF

      IF (rtp%converged) THEN
         IF (rtp%linear_scaling) THEN
            CALL get_rtp(rtp=rtp, rho_old=rho_old, rho_new=rho_new)
            IF (rtp_control%orthonormal) THEN
               CALL purify_mcweeny_complex_orth(rho_new, rtp%filter_eps, rtp%filter_eps_small, &
                                                rtp_control%mcweeny_max_iter, rtp_control%mcweeny_eps)
            ELSE
               CALL purify_mcweeny_complex_nonorth(rho_new, s_mat, rtp%filter_eps, rtp%filter_eps_small, &
                                                   rtp_control%mcweeny_max_iter, rtp_control%mcweeny_eps)
            ENDIF
            IF (rtp_control%mcweeny_max_iter > 0) CALL calc_update_rho_sparse(qs_env)
            CALL report_density_occupation(rtp%filter_eps, rho_new)
            DO i = 1, SIZE(rho_new)
               CALL cp_dbcsr_copy(rho_old(i)%matrix, rho_new(i)%matrix)
            END DO
         ELSE
            CALL get_rtp(rtp=rtp, mos_old=mos_old, mos_new=mos_new)
            DO i = 1, SIZE(mos_new)
               CALL cp_fm_to_fm(mos_new(i)%matrix, mos_old(i)%matrix)
            END DO
         ENDIF
         IF (rtp_control%propagator == do_em) CALL calc_SinvH(rtp, matrix_ks, matrix_ks_im, rtp_control)
         DO i = 1, SIZE(exp_H_new)
            CALL cp_dbcsr_copy(exp_H_old(i)%matrix, exp_H_new(i)%matrix)
         END DO
         ihist = MOD(rtp%istep, rtp_control%aspc_order)+1
         IF (rtp_control%fixed_ions) THEN
            CALL put_data_to_history(rtp, rho=rho_new, mos=mos_new, ihist=ihist)
         ELSE
            CALL put_data_to_history(rtp, rho=rho_new, mos=mos_new, s_mat=s_mat, ihist=ihist)
         END IF
      END IF

      rtp%energy_new = energy%total

      CALL timestop(handle)

   END SUBROUTINE step_finalize

! **************************************************************************************************
!> \brief computes the propagator matrix for EM/ETRS, RTP/EMD
!> \param rtp ...
!> \param propagator ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE compute_propagator_matrix(rtp, propagator)
      TYPE(rt_prop_type), POINTER                        :: rtp
      INTEGER                                            :: propagator

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_propagator_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(Kind=dp)                                      :: dt, prefac
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_new, exp_H_old, propagator_matrix

      CALL timeset(routineN, handle)
      CALL get_rtp(rtp=rtp, exp_H_new=exp_H_new, exp_H_old=exp_H_old, &
                   propagator_matrix=propagator_matrix, dt=dt)

      prefac = -0.5_dp*dt

      DO i = 1, SIZE(exp_H_new)
         CALL cp_dbcsr_add(propagator_matrix(i)%matrix, exp_H_new(i)%matrix, 0.0_dp, prefac)
         IF (propagator == do_em) &
            CALL cp_dbcsr_add(propagator_matrix(i)%matrix, exp_H_old(i)%matrix, 1.0_dp, prefac)
      END DO

      CALL timestop(handle)

   END SUBROUTINE compute_propagator_matrix

! **************************************************************************************************
!> \brief computes t*S_inv*H, if needed t*Sinv*B
!> \param rtp ...
!> \param matrix_ks ...
!> \param matrix_ks_im ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE calc_SinvH(rtp, matrix_ks, matrix_ks_im, rtp_control)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_ks_im
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_SinvH', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, im, ispin, re
      REAL(dp)                                           :: t
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H, SinvB, SinvH
      TYPE(cp_dbcsr_type)                                :: matrix_ks_nosym, tmp
      TYPE(cp_dbcsr_type), POINTER                       :: B_mat, S_inv, S_minus_half

      CALL timeset(routineN, handle)
      CALL get_rtp(rtp=rtp, S_inv=S_inv, S_minus_half=S_minus_half, exp_H_new=exp_H, dt=t)
      CALL cp_dbcsr_init(matrix_ks_nosym)
      CALL cp_dbcsr_create(matrix_ks_nosym, template=matrix_ks(1)%matrix, matrix_type="N")
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=matrix_ks(1)%matrix, matrix_type="N")
      DO ispin = 1, SIZE(matrix_ks)
         re = ispin*2-1
         im = ispin*2
         CALL cp_dbcsr_desymmetrize(matrix_ks(ispin)%matrix, matrix_ks_nosym)
         IF (rtp_control%orthonormal) THEN
            CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, matrix_ks_nosym, zero, tmp, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp, S_minus_half, zero, exp_H(im)%matrix, &
                                   filter_eps=rtp%filter_eps)
         ELSE
            CALL cp_dbcsr_multiply("N", "N", one, S_inv, matrix_ks_nosym, zero, exp_H(im)%matrix, &
                                   filter_eps=rtp%filter_eps)
         ENDIF
         IF (.NOT. rtp_control%fixed_ions) THEN
            CALL get_rtp(rtp=rtp, SinvH=SinvH)
            CALL cp_dbcsr_copy(SinvH(ispin)%matrix, exp_H(im)%matrix)
         END IF
      END DO
      IF (.NOT. rtp_control%fixed_ions .OR. rtp%do_hfx) THEN
         CALL get_rtp(rtp=rtp, B_mat=B_mat, SinvB=SinvB)
         IF (rtp%do_hfx) THEN
            DO ispin = 1, SIZE(matrix_ks)
               re = ispin*2-1
               im = ispin*2
               CALL cp_dbcsr_set(matrix_ks_nosym, 0.0_dp)
               CALL cp_dbcsr_desymmetrize(matrix_ks_im(ispin)%matrix, matrix_ks_nosym)

               ! take care of the EMD case and add the velocity scaled S_derivative
               IF (.NOT. rtp_control%fixed_ions) &
                  CALL cp_dbcsr_add(matrix_ks_nosym, B_mat, 1.0_dp, -1.0_dp)

               IF (rtp_control%orthonormal) THEN
                  CALL cp_dbcsr_multiply("N", "N", -one, S_minus_half, matrix_ks_nosym, zero, tmp, &
                                         filter_eps=rtp%filter_eps)
                  CALL cp_dbcsr_multiply("N", "N", one, tmp, S_minus_half, zero, exp_H(re)%matrix, &
                                         filter_eps=rtp%filter_eps)
               ELSE
                  CALL cp_dbcsr_multiply("N", "N", -one, S_inv, matrix_ks_nosym, zero, exp_H(re)%matrix, &
                                         filter_eps=rtp%filter_eps)
               ENDIF

               IF (.NOT. rtp_control%fixed_ions) &
                  CALL cp_dbcsr_copy(SinvB(ispin)%matrix, exp_H(re)%matrix)
            END DO
         ELSE
            ! in case of pure EMD its only needed once as B is the same for both spins
            IF (rtp_control%orthonormal) THEN
               CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, B_mat, zero, tmp, &
                                      filter_eps=rtp%filter_eps)
               CALL cp_dbcsr_multiply("N", "N", one, tmp, S_minus_half, zero, exp_H(1)%matrix, &
                                      filter_eps=rtp%filter_eps)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", one, S_inv, B_mat, zero, exp_H(1)%matrix, filter_eps=rtp%filter_eps)
            ENDIF

            CALL cp_dbcsr_copy(SinvB(1)%matrix, exp_H(1)%matrix)

            IF (SIZE(matrix_ks) == 2) CALL cp_dbcsr_copy(exp_H(3)%matrix, exp_H(1)%matrix)
            IF (SIZE(matrix_ks) == 2) CALL cp_dbcsr_copy(SinvB(2)%matrix, SinvB(1)%matrix)
         END IF
      ELSE
         !set real part to zero
         DO ispin = 1, SIZE(exp_H)/2
            re = ispin*2-1
            im = ispin*2
            CALL cp_dbcsr_set(exp_H(re)%matrix, zero)
         ENDDO
      END IF
      CALL cp_dbcsr_release(matrix_ks_nosym)
      CALL cp_dbcsr_release(tmp)
      CALL timestop(handle)
   END SUBROUTINE calc_SinvH

! **************************************************************************************************
!> \brief calculates the needed overlap-like matrices
!>        depending on the way the exponential is calculated, only S^-1 is needed
!> \param s_mat ...
!> \param rtp ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE s_matrices_create(s_mat, rtp)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: s_mat
      TYPE(rt_prop_type), POINTER                        :: rtp

      CHARACTER(len=*), PARAMETER :: routineN = 's_matrices_create', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type), POINTER                       :: S_half, S_inv, S_minus_half

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, S_inv=S_inv)

      IF (rtp%linear_scaling) THEN
         CALL get_rtp(rtp=rtp, S_half=S_half, S_minus_half=S_minus_half)
         CALL matrix_sqrt_Newton_Schulz(S_half, S_minus_half, s_mat(1)%matrix, rtp%filter_eps, &
                                        rtp%newton_schulz_order, rtp%lanzcos_threshold, rtp%lanzcos_max_iter)
         CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, S_minus_half, zero, S_inv, &
                                filter_eps=rtp%filter_eps)
      ELSE
         CALL cp_dbcsr_copy(S_inv, s_mat(1)%matrix)
         CALL cp_dbcsr_cholesky_decompose(S_inv, para_env=rtp%ao_ao_fmstruct%para_env, &
                                          blacs_env=rtp%ao_ao_fmstruct%context)
         CALL cp_dbcsr_cholesky_invert(S_inv, para_env=rtp%ao_ao_fmstruct%para_env, &
                                       blacs_env=rtp%ao_ao_fmstruct%context, upper_to_full=.TRUE.)
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE s_matrices_create

! **************************************************************************************************
!> \brief Calculates the frobenius norm of a complex matrix represented by two real matrices
!> \param frob_norm ...
!> \param mat_re ...
!> \param mat_im ...
!> \author Samuel Andermatt (04.14)
! **************************************************************************************************

   SUBROUTINE complex_frobenius_norm(frob_norm, mat_re, mat_im)

      REAL(KIND=dp), INTENT(out)                         :: frob_norm
      TYPE(cp_dbcsr_type), POINTER                       :: mat_re, mat_im

      CHARACTER(len=*), PARAMETER :: routineN = 'complex_frobenius_norm', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: col_atom, handle, row_atom
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:), POINTER                    :: block_values, block_values2
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type), POINTER                       :: tmp

      CALL timeset(routineN, handle)

      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=mat_re)
      !make sure the tmp has the same sparsity pattern as the real and the complex part combined
      CALL cp_dbcsr_add(tmp, mat_re, zero, one)
      CALL cp_dbcsr_add(tmp, mat_im, zero, one)
      CALL cp_dbcsr_set(tmp, zero)
      !calculate the hadamard product
      CALL cp_dbcsr_iterator_start(iter, tmp)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
         CALL cp_dbcsr_get_block_p(mat_re, row_atom, col_atom, block_values2, found=found)
         IF (found) THEN
            block_values = block_values2*block_values2
         ENDIF
         CALL cp_dbcsr_get_block_p(mat_im, row_atom, col_atom, block_values2, found=found)
         IF (found) THEN
            block_values = block_values+block_values2*block_values2
         ENDIF
         block_values = SQRT(block_values)
      END DO
      CALL cp_dbcsr_iterator_stop(iter)
      frob_norm = cp_dbcsr_frobenius_norm(tmp)

      CALL cp_dbcsr_deallocate_matrix(tmp)

      CALL timestop(handle)

   END SUBROUTINE complex_frobenius_norm

! **************************************************************************************************
!> \brief Does McWeeny for complex matrices in the non-orthogonal basis
!> \param P ...
!> \param s_mat ...
!> \param eps ...
!> \param eps_small ...
!> \param max_iter ...
!> \param threshold ...
!> \author Samuel Andermatt (04.14)
! **************************************************************************************************

   SUBROUTINE purify_mcweeny_complex_nonorth(P, s_mat, eps, eps_small, max_iter, threshold)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: P, s_mat
      REAL(KIND=dp), INTENT(in)                          :: eps, eps_small
      INTEGER, INTENT(in)                                :: max_iter
      REAL(KIND=dp), INTENT(in)                          :: threshold

      CHARACTER(len=*), PARAMETER :: routineN = 'purify_mcweeny_complex_nonorth', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, imax, ispin, re, unit_nr
      REAL(KIND=dp)                                      :: frob_norm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: PS, PSP, tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      NULLIFY (tmp, PS, PSP)
      CALL cp_dbcsr_allocate_matrix_set(tmp, SIZE(P))
      CALL cp_dbcsr_allocate_matrix_set(PSP, SIZE(P))
      CALL cp_dbcsr_allocate_matrix_set(PS, SIZE(P))
      DO i = 1, SIZE(P)
         CALL cp_dbcsr_init_p(PS(i)%matrix)
         CALL cp_dbcsr_create(PS(i)%matrix, template=P(1)%matrix)
         CALL cp_dbcsr_init_p(PSP(i)%matrix)
         CALL cp_dbcsr_create(PSP(i)%matrix, template=P(1)%matrix)
         CALL cp_dbcsr_init_p(tmp(i)%matrix)
         CALL cp_dbcsr_create(tmp(i)%matrix, template=P(1)%matrix)
      ENDDO
      IF (SIZE(P) == 2) THEN
         CALL cp_dbcsr_scale(P(1)%matrix, one/2)
         CALL cp_dbcsr_scale(P(2)%matrix, one/2)
      ENDIF
      DO ispin = 1, SIZE(P)/2
         re = 2*ispin-1
         im = 2*ispin
         imax = MAX(max_iter, 1) !if max_iter is 0 then only the deviation from idempotency needs to be calculated
         DO i = 1, imax
            CALL cp_dbcsr_multiply("N", "N", one, P(re)%matrix, s_mat(1)%matrix, &
                                   zero, PS(re)%matrix, filter_eps=eps_small)
            CALL cp_dbcsr_multiply("N", "N", one, P(im)%matrix, s_mat(1)%matrix, &
                                   zero, PS(im)%matrix, filter_eps=eps_small)
            CALL cp_complex_dbcsr_gemm_3("N", "N", one, PS(re)%matrix, PS(im)%matrix, &
                                         P(re)%matrix, P(im)%matrix, zero, PSP(re)%matrix, PSP(im)%matrix, &
                                         filter_eps=eps_small)
            CALL cp_dbcsr_copy(tmp(re)%matrix, PSP(re)%matrix)
            CALL cp_dbcsr_copy(tmp(im)%matrix, PSP(im)%matrix)
            CALL cp_dbcsr_add(tmp(re)%matrix, P(re)%matrix, 1.0_dp, -1.0_dp)
            CALL cp_dbcsr_add(tmp(im)%matrix, P(im)%matrix, 1.0_dp, -1.0_dp)
            CALL complex_frobenius_norm(frob_norm, tmp(re)%matrix, tmp(im)%matrix)
            IF (unit_nr .GT. 0) WRITE (unit_nr, '(t3,a,2f16.8)') "Deviation from idempotency: ", frob_norm
            IF (frob_norm .GT. threshold .AND. max_iter > 0) THEN
               CALL cp_dbcsr_copy(P(re)%matrix, PSP(re)%matrix)
               CALL cp_dbcsr_copy(P(im)%matrix, PSP(im)%matrix)
               CALL cp_complex_dbcsr_gemm_3("N", "N", -2.0_dp, PS(re)%matrix, PS(im)%matrix, &
                                            PSP(re)%matrix, PSP(im)%matrix, 3.0_dp, P(re)%matrix, P(im)%matrix, &
                                            filter_eps=eps_small)
               CALL cp_dbcsr_filter(P(re)%matrix, eps)
               CALL cp_dbcsr_filter(P(im)%matrix, eps)
               !make sure P is exactly hermitian
               CALL cp_dbcsr_transposed(tmp(re)%matrix, P(re)%matrix)
               CALL cp_dbcsr_add(P(re)%matrix, tmp(re)%matrix, one/2, one/2)
               CALL cp_dbcsr_transposed(tmp(im)%matrix, P(im)%matrix)
               CALL cp_dbcsr_add(P(im)%matrix, tmp(im)%matrix, one/2, -one/2)
            ELSE
               EXIT
            END IF
         END DO
         !make sure P is hermitian
         CALL cp_dbcsr_transposed(tmp(re)%matrix, P(re)%matrix)
         CALL cp_dbcsr_add(P(re)%matrix, tmp(re)%matrix, one/2, one/2)
         CALL cp_dbcsr_transposed(tmp(im)%matrix, P(im)%matrix)
         CALL cp_dbcsr_add(P(im)%matrix, tmp(im)%matrix, one/2, -one/2)
      END DO
      IF (SIZE(P) == 2) THEN
         CALL cp_dbcsr_scale(P(1)%matrix, one*2)
         CALL cp_dbcsr_scale(P(2)%matrix, one*2)
      ENDIF
      CALL cp_dbcsr_deallocate_matrix_set(tmp)
      CALL cp_dbcsr_deallocate_matrix_set(PS)
      CALL cp_dbcsr_deallocate_matrix_set(PSP)

      CALL timestop(handle)

   END SUBROUTINE purify_mcweeny_complex_nonorth

! **************************************************************************************************
!> \brief Does McWeeny for complex matrices in the orthonormal basis
!> \param P ...
!> \param eps ...
!> \param eps_small ...
!> \param max_iter ...
!> \param threshold ...
!> \author Samuel Andermatt (04.14)
! **************************************************************************************************

   SUBROUTINE purify_mcweeny_complex_orth(P, eps, eps_small, max_iter, threshold)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: P
      REAL(KIND=dp), INTENT(in)                          :: eps, eps_small
      INTEGER, INTENT(in)                                :: max_iter
      REAL(KIND=dp), INTENT(in)                          :: threshold

      CHARACTER(len=*), PARAMETER :: routineN = 'purify_mcweeny_complex_orth', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, imax, ispin, re, unit_nr
      REAL(KIND=dp)                                      :: frob_norm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: PP, tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      NULLIFY (tmp, PP)
      CALL cp_dbcsr_allocate_matrix_set(tmp, SIZE(P))
      CALL cp_dbcsr_allocate_matrix_set(PP, SIZE(P))
      DO i = 1, SIZE(P)
         CALL cp_dbcsr_init_p(PP(i)%matrix)
         CALL cp_dbcsr_create(PP(i)%matrix, template=P(1)%matrix)
         CALL cp_dbcsr_init_p(tmp(i)%matrix)
         CALL cp_dbcsr_create(tmp(i)%matrix, template=P(1)%matrix)
      ENDDO
      IF (SIZE(P) == 2) THEN
         CALL cp_dbcsr_scale(P(1)%matrix, one/2)
         CALL cp_dbcsr_scale(P(2)%matrix, one/2)
      ENDIF
      DO ispin = 1, SIZE(P)/2
         re = 2*ispin-1
         im = 2*ispin
         imax = MAX(max_iter, 1) !if max_iter is 0 then only the deviation from idempotency needs to be calculated
         DO i = 1, imax
            CALL cp_complex_dbcsr_gemm_3("N", "N", one, P(re)%matrix, P(im)%matrix, &
                                         P(re)%matrix, P(im)%matrix, zero, PP(re)%matrix, PP(im)%matrix, &
                                         filter_eps=eps_small)
            CALL cp_dbcsr_copy(tmp(re)%matrix, PP(re)%matrix)
            CALL cp_dbcsr_copy(tmp(im)%matrix, PP(im)%matrix)
            CALL cp_dbcsr_add(tmp(re)%matrix, P(re)%matrix, 1.0_dp, -1.0_dp)
            CALL cp_dbcsr_add(tmp(im)%matrix, P(im)%matrix, 1.0_dp, -1.0_dp)
            CALL complex_frobenius_norm(frob_norm, tmp(re)%matrix, tmp(im)%matrix)
            IF (unit_nr .GT. 0) WRITE (unit_nr, '(t3,a,2f16.8)') "Deviation from idempotency: ", frob_norm
            IF (frob_norm .GT. threshold .AND. max_iter > 0) THEN
               CALL cp_dbcsr_copy(tmp(re)%matrix, P(re)%matrix)
               CALL cp_dbcsr_copy(tmp(im)%matrix, P(im)%matrix)
               CALL cp_dbcsr_copy(P(re)%matrix, PP(re)%matrix)
               CALL cp_dbcsr_copy(P(im)%matrix, PP(im)%matrix)
               CALL cp_complex_dbcsr_gemm_3("N", "N", -2.0_dp, tmp(re)%matrix, tmp(im)%matrix, &
                                            PP(re)%matrix, PP(im)%matrix, 3.0_dp, P(re)%matrix, P(im)%matrix, &
                                            filter_eps=eps_small)
               CALL cp_dbcsr_filter(P(re)%matrix, eps)
               CALL cp_dbcsr_filter(P(im)%matrix, eps)
               !make sure P is exactly hermitian
               CALL cp_dbcsr_transposed(tmp(re)%matrix, P(re)%matrix)
               CALL cp_dbcsr_add(P(re)%matrix, tmp(re)%matrix, one/2, one/2)
               CALL cp_dbcsr_transposed(tmp(im)%matrix, P(im)%matrix)
               CALL cp_dbcsr_add(P(im)%matrix, tmp(im)%matrix, one/2, -one/2)
            ELSE
               EXIT
            END IF
         END DO
      END DO
      IF (SIZE(P) == 2) THEN
         CALL cp_dbcsr_scale(P(1)%matrix, one*2)
         CALL cp_dbcsr_scale(P(2)%matrix, one*2)
      ENDIF
      CALL cp_dbcsr_deallocate_matrix_set(tmp)
      CALL cp_dbcsr_deallocate_matrix_set(PP)

      CALL timestop(handle)

   END SUBROUTINE purify_mcweeny_complex_orth

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
!> \param matrix_s ...
!> \param aspc_order ...
! **************************************************************************************************
   SUBROUTINE aspc_extrapolate(rtp, matrix_s, aspc_order)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      INTEGER, INTENT(in)                                :: aspc_order

      CHARACTER(len=*), PARAMETER :: routineN = 'aspc_extrapolate', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: cone = (1.0_dp, 0.0_dp), &
                                                            czero = (0.0_dp, 0.0_dp)
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, iaspc, icol_local, ihist, &
                                                            imat, k, kdbl, n, naspc, ncol_local, &
                                                            nmat
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_cfm_type), POINTER                         :: cfm_tmp, cfm_tmp1, csc
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_new, s_hist
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_hist
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: mo_hist
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct, matrix_struct_new
      TYPE(cp_fm_type), POINTER                          :: fm_tmp, fm_tmp1, fm_tmp2

      NULLIFY (rho_hist)
      CALL timeset(routineN, handle)
      CALL cite_reference(Kolafa2004)

      IF (rtp%linear_scaling) THEN
         CALL get_rtp(rtp=rtp, rho_new=rho_new)
      ELSE
         CALL get_rtp(rtp=rtp, mos_new=mos_new)
      ENDIF

      naspc = MIN(rtp%istep, aspc_order)
      IF (rtp%linear_scaling) THEN
         nmat = SIZE(rho_new)
         rho_hist => rtp%history%rho_history
         DO imat = 1, nmat
            DO iaspc = 1, naspc
               alpha = (-1.0_dp)**(iaspc+1)*REAL(iaspc, KIND=dp)* &
                       binomial(2*naspc, naspc-iaspc)/binomial(2*naspc-2, naspc-1)
               ihist = MOD(rtp%istep-iaspc, aspc_order)+1
               IF (iaspc == 1) THEN
                  CALL cp_dbcsr_add(rho_new(imat)%matrix, rho_hist(imat, ihist)%matrix, zero, alpha)
               ELSE
                  CALL cp_dbcsr_add(rho_new(imat)%matrix, rho_hist(imat, ihist)%matrix, one, alpha)
               END IF
            END DO
         END DO
      ELSE
         mo_hist => rtp%history%mo_history
         nmat = SIZE(mos_new)
         DO imat = 1, nmat
            DO iaspc = 1, naspc
               alpha = (-1.0_dp)**(iaspc+1)*REAL(iaspc, KIND=dp)* &
                       binomial(2*naspc, naspc-iaspc)/binomial(2*naspc-2, naspc-1)
               ihist = MOD(rtp%istep-iaspc, aspc_order)+1
               IF (iaspc == 1) THEN
                  CALL cp_fm_scale_and_add(zero, mos_new(imat)%matrix, alpha, mo_hist(imat, ihist)%matrix)
               ELSE
                  CALL cp_fm_scale_and_add(one, mos_new(imat)%matrix, alpha, mo_hist(imat, ihist)%matrix)
               END IF
            END DO
         END DO

         mo_hist => rtp%history%mo_history
         s_hist => rtp%history%s_history
         DO i = 1, SIZE(mos_new)/2
            NULLIFY (matrix_struct, matrix_struct_new, csc, fm_tmp, fm_tmp1, fm_tmp2, cfm_tmp, cfm_tmp1)

            CALL cp_fm_struct_double(matrix_struct, &
                                     mos_new(2*i)%matrix%matrix_struct, &
                                     mos_new(2*i)%matrix%matrix_struct%context, &
                                     .TRUE., .FALSE.)

            CALL cp_fm_create(fm_tmp, matrix_struct)
            CALL cp_fm_create(fm_tmp1, matrix_struct)
            CALL cp_fm_create(fm_tmp2, mos_new(2*i)%matrix%matrix_struct)
            CALL cp_cfm_create(cfm_tmp, mos_new(2*i)%matrix%matrix_struct)
            CALL cp_cfm_create(cfm_tmp1, mos_new(2*i)%matrix%matrix_struct)

            CALL cp_fm_get_info(fm_tmp, &
                                ncol_global=kdbl)

            CALL cp_fm_get_info(mos_new(2*i)%matrix, &
                                nrow_global=n, &
                                ncol_global=k, &
                                ncol_local=ncol_local)

            CALL cp_fm_struct_create(matrix_struct_new, &
                                     template_fmstruct=matrix_struct, &
                                     nrow_global=k, &
                                     ncol_global=k)
            CALL cp_cfm_create(csc, matrix_struct_new)

            CALL cp_fm_struct_release(matrix_struct_new)
            CALL cp_fm_struct_release(matrix_struct)

            ! first the most recent

! reorthogonalize vectors

            DO icol_local = 1, ncol_local
               fm_tmp%local_data(:, icol_local) = mos_new(2*i-1)%matrix%local_data(:, icol_local)
               fm_tmp%local_data(:, icol_local+ncol_local) = mos_new(2*i)%matrix%local_data(:, icol_local)
            END DO

            CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, fm_tmp, fm_tmp1, kdbl)

            DO icol_local = 1, ncol_local
               cfm_tmp%local_data(:, icol_local) = CMPLX(fm_tmp1%local_data(:, icol_local), &
                                                         fm_tmp1%local_data(:, icol_local+ncol_local), dp)
               cfm_tmp1%local_data(:, icol_local) = CMPLX(mos_new(2*i-1)%matrix%local_data(:, icol_local), &
                                                          mos_new(2*i)%matrix%local_data(:, icol_local), dp)
            END DO
            CALL cp_cfm_gemm('C', 'N', k, k, n, cone, cfm_tmp1, cfm_tmp, czero, csc)
            CALL cp_cfm_cholesky_decompose(csc)
            CALL cp_cfm_triangular_multiply(csc, cfm_tmp1, n_cols=k, side='R', invert_tr=.TRUE.)
            DO icol_local = 1, ncol_local
               mos_new(2*i-1)%matrix%local_data(:, icol_local) = REAL(cfm_tmp1%local_data(:, icol_local), dp)
               mos_new(2*i)%matrix%local_data(:, icol_local) = AIMAG(cfm_tmp1%local_data(:, icol_local))
            END DO

! deallocate work matrices
            CALL cp_cfm_release(csc)
            CALL cp_fm_release(fm_tmp)
            CALL cp_fm_release(fm_tmp)
            CALL cp_fm_release(fm_tmp1)
            CALL cp_fm_release(fm_tmp2)
            CALL cp_cfm_release(cfm_tmp)
            CALL cp_cfm_release(cfm_tmp1)
         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE aspc_extrapolate

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
!> \param mos ...
!> \param rho ...
!> \param s_mat ...
!> \param ihist ...
! **************************************************************************************************
   SUBROUTINE put_data_to_history(rtp, mos, rho, s_mat, ihist)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: s_mat
      INTEGER                                            :: ihist

      CHARACTER(len=*), PARAMETER :: routineN = 'put_data_to_history', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (rtp%linear_scaling) THEN
         DO i = 1, SIZE(rho)
            CALL cp_dbcsr_copy(rtp%history%rho_history(i, ihist)%matrix, rho(i)%matrix)
         END DO
      ELSE
         DO i = 1, SIZE(mos)
            CALL cp_fm_to_fm(mos(i)%matrix, rtp%history%mo_history(i, ihist)%matrix)
         END DO
         IF (PRESENT(s_mat)) THEN
            IF (ASSOCIATED(rtp%history%s_history(ihist)%matrix)) THEN ! the sparsity might be different
               ! (future struct:check)
               CALL cp_dbcsr_deallocate_matrix(rtp%history%s_history(ihist)%matrix)
            END IF
            ALLOCATE (rtp%history%s_history(ihist)%matrix)
            CALL cp_dbcsr_init(rtp%history%s_history(ihist)%matrix)
            CALL cp_dbcsr_copy(rtp%history%s_history(ihist)%matrix, s_mat(1)%matrix)
         END IF
      END IF

   END SUBROUTINE put_data_to_history

END MODULE rt_propagation_methods
