!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE farming_types

   USE kinds,                           ONLY: default_path_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: farming_env_type, deallocate_farming_env, init_farming_env, init_job_type

   INTEGER, PUBLIC, PARAMETER :: job_pending = 1, job_running = 2, job_finished = 3

! **************************************************************************************************
   TYPE job_type
      CHARACTER(LEN=default_path_length) :: cwd ! the directory to go to
      CHARACTER(LEN=default_path_length) :: input ! the input file to use
      CHARACTER(LEN=default_path_length) :: output ! the output file to use
      INTEGER                            :: ID ! the ID of this job
      INTEGER, POINTER, DIMENSION(:)     :: dependencies ! the dependencies of this job
      INTEGER                            :: status ! pending,running,finished
   END TYPE job_type

! **************************************************************************************************
   TYPE farming_env_type
      INTEGER :: group_size_wish
      LOGICAL :: group_size_wish_set
      INTEGER :: ngroup_wish
      LOGICAL :: ngroup_wish_set
      LOGICAL :: restart
      LOGICAL :: CYCLE
      LOGICAL :: master_slave
      INTEGER, DIMENSION(:), POINTER                              :: group_partition ! user preference for partitioning the cpus
      CHARACTER(LEN=default_path_length)                          :: restart_file_name ! restart file for farming
      CHARACTER(LEN=default_path_length)                          :: cwd ! directory we started from
      INTEGER                                                     :: Njobs ! how many jobs to run
      INTEGER                                                     :: restart_n ! where to start
      INTEGER                                                     :: max_steps ! max number of steps,
      ! results in max_steps*Ngroup jobs being run
      INTEGER                                                     :: stride ! for creating slave groups.
      TYPE(job_type), DIMENSION(:), POINTER                       :: job ! a list of jobs
      REAL(KIND=dp) :: wait_time
   END TYPE farming_env_type

CONTAINS

! **************************************************************************************************
!> \brief help poor compilers do their job
!>       i.e. provide a default initialization
!> \param farming_env an associated farming env pointer
!> \par History
!>      03.2004 created [Joost VandeVondele ]
! **************************************************************************************************
   SUBROUTINE init_farming_env(farming_env)
      TYPE(farming_env_type), POINTER                    :: farming_env

      IF (ASSOCIATED(farming_env)) THEN
         farming_env%group_size_wish = 0
         farming_env%group_size_wish_set = .FALSE.
         farming_env%ngroup_wish = 0
         farming_env%ngroup_wish_set = .FALSE.
         farming_env%restart = .FALSE.
         farming_env%restart_n = 1
         farming_env%cycle = .FALSE.
         farming_env%master_slave = .FALSE.
         NULLIFY (farming_env%group_partition)
         farming_env%cwd = "."
         farming_env%Njobs = 0
         ! so that maxsteps*ngroup is not overflowing
         farming_env%max_steps = HUGE(0)/32768
         NULLIFY (farming_env%Job)
      ENDIF
   END SUBROUTINE

! **************************************************************************************************
!> \brief provide a default initialization
!> \param job ...
!> \par History
!>      09.2007 created [Joost VandeVondele ]
! **************************************************************************************************
   ELEMENTAL SUBROUTINE init_job_type(job)
      TYPE(job_type), INTENT(OUT)                        :: job

      job%cwd = ""
      job%input = ""
      job%output = ""
      job%ID = -1
      job%status = job_pending
      NULLIFY (job%dependencies)

   END SUBROUTINE init_job_type

! **************************************************************************************************
!> \brief deallocates all memory associated with this job
!> \param job ...
!> \par History
!>      09.2007 created [Joost VandeVondele ]
! **************************************************************************************************
   SUBROUTINE deallocate_job_type(job)
      TYPE(job_type)                                     :: job

      IF (ASSOCIATED(job%dependencies)) DEALLOCATE (job%dependencies)

   END SUBROUTINE deallocate_job_type

! **************************************************************************************************
!> \brief deallocates all associated fields of the farming_env type
!>      and the type itself
!> \param farming_env ...
!> \par History
!>      03.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE deallocate_farming_env(farming_env)
      TYPE(farming_env_type), POINTER                    :: farming_env

      INTEGER                                            :: I

      IF (ASSOCIATED(farming_env)) THEN
         IF (ASSOCIATED(farming_env%job)) THEN
            DO I = 1, SIZE(farming_env%job, 1)
               CALL deallocate_job_type(farming_env%job(I))
            ENDDO
            DEALLOCATE (farming_env%job)
         ENDIF
         IF (ASSOCIATED(farming_env%group_partition)) DEALLOCATE (farming_env%group_partition)
         DEALLOCATE (farming_env) ! and the type itself
      ENDIF
   END SUBROUTINE deallocate_farming_env
END MODULE farming_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The environment for the empirical interatomic potential methods.
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
MODULE eip_environment_types
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_create,&
                                              atomic_kind_list_release,&
                                              atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_retain,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_create,&
                                              mol_kind_new_list_release,&
                                              mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_create,&
                                              mol_new_list_release,&
                                              mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_list_types,             ONLY: particle_list_create,&
                                              particle_list_release,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eip_environment_types'
   INTEGER, PRIVATE, SAVE :: last_eip_id = 0

   ! *** Public data types ***
   PUBLIC :: eip_environment_type

   ! *** Public subroutines ***
   PUBLIC :: eip_env_retain, &
             eip_env_release, &
             eip_env_set, &
             eip_env_get, &
             eip_env_create

! **************************************************************************************************
!> \brief The empirical interatomic potential environment
!> \param id_nr unique identifier
!> \param ref_count reference counter (see doc/ReferenceCounting.html)
!> \param eip_model Specifies which EIP model is in use.
!> \param eip_kinetic_energy The EIP kinetic energy
!> \param eip_potential_energy The EIP potential energy
!> \param eip_energy The total eip energy
!> \param eip_energy_var Variance of the energy/atom
!> \param eip_forces The final eip forces [eV/A]
!> \param coord_avg The average coordination number
!> \param coord_var The variance of the coordination number
!> \param count Counts how often the empirical interatomic potential function
!>               is called. Don't ask why this is a real!
!> \param subsystem the particles, molecules,... of this environment
!> \param eip_input Pointer to the EIP input section
!> \param force_env_input Pointer to the force_env input section
!> \param cell The simulation cell
!> \param cell_ref The reference simulation cell
!> \param use_ref_cell Logical which indicates if reference
!>                      simulation cell is used
!> \param virial Dummy virial pointer
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   TYPE eip_environment_type
      INTEGER                                          :: id_nr, &
                                                          ref_count, &
                                                          eip_model
      REAL(KIND=dp)                                    :: eip_energy, &
                                                          eip_kinetic_energy, &
                                                          eip_potential_energy, &
                                                          eip_energy_var
      REAL(KIND=dp), DIMENSION(:, :), POINTER           :: eip_forces
      REAL(KIND=dp)                                    :: coord_avg, &
                                                          coord_var, &
                                                          count
      TYPE(cp_subsys_type), POINTER                    :: subsys
      TYPE(section_vals_type), POINTER                 :: eip_input, &
                                                          force_env_input
      TYPE(cell_type), POINTER                         :: cell_ref
      LOGICAL                                          :: use_ref_cell
   END TYPE eip_environment_type

! **************************************************************************************************
!> \brief Type which allows the use of arrays of pointers
!> \param eip_env_ptr Pointer to the actual eip environment
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   TYPE eip_environment_p_type
      TYPE(eip_environment_type), POINTER :: eip_env_ptr
   END TYPE eip_environment_p_type

CONTAINS

! **************************************************************************************************
!> \brief Retains a eip environment (see doc/ReferenceCounting.html)
!> \param eip_env The eip environment to retain
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_env_retain(eip_env)

      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_retain', routineP = moduleN//':'//routineN

!   ------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_environment_types.F",133)
      IF(.NOT.(eip_env%ref_count > 0))CALL cp__a("eip_environment_types.F",134)
      eip_env%ref_count = eip_env%ref_count+1
   END SUBROUTINE eip_env_retain

! **************************************************************************************************
!> \brief Releases the given eip environment (see doc/ReferenceCounting.html)
!> \param eip_env The eip environment to release
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_env_release(eip_env)

      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_release', &
         routineP = moduleN//':'//routineN

!   ------------------------------------------------------------------------

      IF (ASSOCIATED(eip_env)) THEN
         IF(.NOT.(eip_env%ref_count > 0))CALL cp__a("eip_environment_types.F",155)
         eip_env%ref_count = eip_env%ref_count-1
         IF (eip_env%ref_count < 1) THEN
            IF (ASSOCIATED(eip_env%eip_forces)) THEN
               DEALLOCATE (eip_env%eip_forces)
            END IF
            IF (ASSOCIATED(eip_env%subsys)) THEN
               CALL cp_subsys_release(eip_env%subsys)
            END IF
            IF (ASSOCIATED(eip_env%subsys)) THEN
               CALL cp_subsys_release(eip_env%subsys)
            END IF
            !IF (ASSOCIATED(eip_env%eip_input)) THEN
            !  CALL section_vals_release(eip_env%eip_input)
            !END IF
            !IF (ASSOCIATED(eip_env%force_env_input)) THEN
            !  CALL section_vals_release(eip_env%force_env_input)
            !END IF
            IF (ASSOCIATED(eip_env%cell_ref)) THEN
               CALL cell_release(eip_env%cell_ref)
            END IF
            DEALLOCATE (eip_env)
         END IF
      END IF
      NULLIFY (eip_env)
   END SUBROUTINE eip_env_release

! **************************************************************************************************
!> \brief Returns various attributes of the eip environment
!> \param eip_env The enquired eip environment
!> \param id_nr unique identifier
!> \param eip_model Specifies which EIP model is in use.
!> \param eip_energy The total eip energy
!> \param eip_energy_var Variance of the energy/atom
!> \param eip_forces The final eip forces [eV/A]
!> \param coord_avg The average coordination number
!> \param coord_var The variance of the coordination number
!> \param count Counts how often the empirical interatomic potential function
!>               is called. Don't ask why this is a real!
!> \param subsys the particles, molecules,... of this environment
!> \param atomic_kind_set The set of all atomic kinds involved
!> \param particle_set The set of all particles
!> \param local_particles All particles on this particular node
!> \param molecule_kind_set The set of all different molecule kinds involved
!> \param molecule_set The set of all molecules
!> \param local_molecules All molecules on this particular node
!> \param eip_input the pointer to the EIP input section
!> \param force_env_input Pointer to the force_env input section
!> \param cell The simulation cell
!> \param cell_ref The reference simulation cell
!> \param use_ref_cell Logical which indicates if reference
!>                      simulation cell is used
!> \param eip_kinetic_energy The EIP kinetic energy
!> \param eip_potential_energy The EIP potential energy
!> \param virial Dummy virial pointer
!>
!>      For possible missing arguments see the attributes of
!>      eip_environment_type
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_env_get(eip_env, id_nr, eip_model, eip_energy, eip_energy_var, &
                          eip_forces, coord_avg, coord_var, count, subsys, &
                          atomic_kind_set, particle_set, local_particles, &
                          molecule_kind_set, molecule_set, local_molecules, &
                          eip_input, force_env_input, cell, cell_ref, &
                          use_ref_cell, eip_kinetic_energy, eip_potential_energy, &
                          virial)

      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(OUT), OPTIONAL                     :: id_nr, eip_model
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: eip_energy, eip_energy_var
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: eip_forces
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: coord_avg, coord_var, count
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules
      TYPE(section_vals_type), OPTIONAL, POINTER         :: eip_input, force_env_input
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      LOGICAL, INTENT(OUT), OPTIONAL                     :: use_ref_cell
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: eip_kinetic_energy, eip_potential_energy
      TYPE(virial_type), OPTIONAL, POINTER               :: virial

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_get', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles

!   ------------------------------------------------------------------------

      NULLIFY (atomic_kinds, particles, molecules_new, molecule_kinds_new)

      IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_environment_types.F",258)
      IF(.NOT.(eip_env%ref_count > 0))CALL cp__a("eip_environment_types.F",259)

      IF (PRESENT(id_nr)) id_nr = eip_env%id_nr
      IF (PRESENT(eip_model)) eip_model = eip_env%eip_model
      IF (PRESENT(eip_kinetic_energy)) THEN
         eip_kinetic_energy = eip_env%eip_kinetic_energy
      END IF
      IF (PRESENT(eip_potential_energy)) THEN
         eip_potential_energy = eip_env%eip_potential_energy
      END IF
      IF (PRESENT(eip_energy)) eip_energy = eip_env%eip_energy
      IF (PRESENT(eip_energy_var)) eip_energy_var = eip_env%eip_energy_var
      IF (PRESENT(eip_forces)) eip_forces = eip_env%eip_forces
      IF (PRESENT(coord_avg)) coord_avg = eip_env%coord_avg
      IF (PRESENT(coord_var)) coord_var = eip_env%coord_var
      IF (PRESENT(count)) count = eip_env%count
      IF (PRESENT(subsys)) subsys => eip_env%subsys
      CALL cp_subsys_get(eip_env%subsys, &
                         atomic_kinds=atomic_kinds, &
                         particles=particles, &
                         molecule_kinds_new=molecule_kinds_new, &
                         molecules_new=molecules_new, &
                         local_molecules_new=local_molecules, &
                         local_particles=local_particles, &
                         virial=virial, &
                         cell=cell)
      IF (PRESENT(atomic_kind_set)) atomic_kind_set => atomic_kinds%els
      IF (PRESENT(particle_set)) particle_set => particles%els
      IF (PRESENT(molecule_kind_set)) molecule_kind_set => molecule_kinds_new%els
      IF (PRESENT(molecule_set)) molecule_set => molecules_new%els

      IF (PRESENT(eip_input)) eip_input => eip_env%eip_input
      IF (PRESENT(force_env_input)) force_env_input => eip_env%force_env_input
      IF (PRESENT(cell_ref)) cell_ref => eip_env%cell_ref
      IF (PRESENT(use_ref_cell)) use_ref_cell = eip_env%use_ref_cell

   END SUBROUTINE eip_env_get

! **************************************************************************************************
!> \brief Sets various attributes of the eip environment
!> \param eip_env The enquired eip environment
!> \param id_nr unique identifier
!> \param eip_model Specifies which EIP model is in use
!> \param eip_energy The total eip energy
!> \param eip_energy_var Variance of the energy/atom
!> \param eip_forces The final eip forces [eV/A]
!> \param coord_avg The average coordination number
!> \param coord_var The variance of the coordination number
!> \param count Counts how often the empirical interatomic potential function
!>               is called. Don't ask why this is a real!
!> \param subsys the particles, molecules,... of this environment
!> \param atomic_kind_set The set of all atomic kinds involved
!> \param particle_set The set of all particles
!> \param local_particles All particles on this particular node
!> \param molecule_kind_set The set of all different molecule kinds involved
!> \param molecule_set The set of all molecules
!> \param local_molecules All molecules on this particular node
!> \param eip_input the pointer to the EIP input section
!> \param force_env_input Pointer to the force_env input section
!> \param cell_ref The reference simulation cell
!> \param use_ref_cell Logical which indicates if reference
!>                      simulation cell is used
!> \param eip_kinetic_energy The EIP kinetic energy
!> \param eip_potential_energy The EIP potential energy
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
!> \note
!>   For possible missing arguments see the attributes of eip_environment_type
! **************************************************************************************************
   SUBROUTINE eip_env_set(eip_env, id_nr, eip_model, eip_energy, eip_energy_var, &
                          eip_forces, coord_avg, coord_var, count, subsys, &
                          atomic_kind_set, particle_set, local_particles, &
                          molecule_kind_set, molecule_set, local_molecules, &
                          eip_input, force_env_input, cell_ref, &
                          use_ref_cell, eip_kinetic_energy, eip_potential_energy)

      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN), OPTIONAL                      :: id_nr, eip_model
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: eip_energy, eip_energy_var
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: eip_forces
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: coord_avg, coord_var, count
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules
      TYPE(section_vals_type), OPTIONAL, POINTER         :: eip_input, force_env_input
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_ref
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_ref_cell
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: eip_kinetic_energy, eip_potential_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_set', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles

!   ------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_environment_types.F",366)
      IF(.NOT.(eip_env%ref_count > 0))CALL cp__a("eip_environment_types.F",367)

      IF (PRESENT(id_nr)) eip_env%id_nr = id_nr
      IF (PRESENT(eip_model)) eip_env%eip_model = eip_model
      IF (PRESENT(eip_kinetic_energy)) THEN
         eip_env%eip_kinetic_energy = eip_kinetic_energy
      END IF
      IF (PRESENT(eip_potential_energy)) THEN
         eip_env%eip_potential_energy = eip_potential_energy
      END IF
      IF (PRESENT(eip_energy)) eip_env%eip_energy = eip_energy
      IF (PRESENT(eip_energy_var)) eip_env%eip_energy_var = eip_energy_var
      IF (PRESENT(eip_forces)) eip_env%eip_forces = eip_forces
      IF (PRESENT(coord_avg)) eip_env%coord_avg = coord_avg
      IF (PRESENT(coord_var)) eip_env%coord_var = coord_var
      IF (PRESENT(count)) eip_env%count = count
      IF (PRESENT(subsys)) THEN
         CALL cp_subsys_retain(subsys)
         CALL cp_subsys_release(eip_env%subsys)
         eip_env%subsys => subsys
      END IF
      IF (PRESENT(atomic_kind_set)) THEN
         CALL atomic_kind_list_create(atomic_kinds, &
                                      els_ptr=atomic_kind_set)
         CALL cp_subsys_set(eip_env%subsys, &
                            atomic_kinds=atomic_kinds)
         CALL atomic_kind_list_release(atomic_kinds)
      END IF
      IF (PRESENT(particle_set)) THEN
         CALL particle_list_create(particles, &
                                   els_ptr=particle_set)
         CALL cp_subsys_set(eip_env%subsys, &
                            particles=particles)
         CALL particle_list_release(particles)
      END IF
      IF (PRESENT(molecule_kind_set)) THEN
         CALL mol_kind_new_list_create(molecule_kinds_new, &
                                       els_ptr=molecule_kind_set)
         CALL cp_subsys_set(eip_env%subsys, &
                            molecule_kinds_new=molecule_kinds_new)
         CALL mol_kind_new_list_release(molecule_kinds_new)
      END IF
      IF (PRESENT(molecule_set)) THEN
         CALL mol_new_list_create(molecules_new, &
                                  els_ptr=molecule_set)
         CALL cp_subsys_set(eip_env%subsys, &
                            molecules_new=molecules_new)
         CALL mol_new_list_release(molecules_new)
      END IF
      IF (PRESENT(local_particles)) THEN
         CALL cp_subsys_set(eip_env%subsys, &
                            local_particles=local_particles)
      END IF
      IF (PRESENT(local_molecules)) THEN
         CALL cp_subsys_set(eip_env%subsys, &
                            local_molecules_new=local_molecules)
      END IF

      IF (PRESENT(eip_input)) eip_env%eip_input => eip_input
      IF (PRESENT(force_env_input)) THEN
         eip_env%force_env_input => force_env_input
      END IF
      IF (PRESENT(cell_ref)) THEN
         CALL cell_retain(cell_ref)
         CALL cell_release(eip_env%cell_ref)
         eip_env%cell_ref => cell_ref
      END IF
      IF (PRESENT(use_ref_cell)) eip_env%use_ref_cell = use_ref_cell
   END SUBROUTINE eip_env_set

! **************************************************************************************************
!> \brief Reinitializes the eip environment
!> \param eip_env The eip environment to be reinitialized
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_env_clear(eip_env)

      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_clear', routineP = moduleN//':'//routineN

!   ------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_environment_types.F",452)

      eip_env%eip_model = 0
      eip_env%eip_kinetic_energy = 0.0_dp
      eip_env%eip_potential_energy = 0.0_dp
      eip_env%eip_energy = 0.0_dp
      eip_env%eip_energy_var = 0.0_dp
      eip_env%coord_avg = 0.0_dp
      eip_env%coord_var = 0.0_dp
      eip_env%count = 0.0_dp
      IF (ASSOCIATED(eip_env%eip_forces)) THEN
         eip_env%eip_forces(:, :) = 0.0_dp
      END IF
      IF (ASSOCIATED(eip_env%subsys)) THEN
         CALL cp_subsys_release(eip_env%subsys)
      END IF
      IF (ASSOCIATED(eip_env%eip_input)) THEN
         CALL section_vals_release(eip_env%eip_input)
      END IF
      IF (ASSOCIATED(eip_env%force_env_input)) THEN
         CALL section_vals_release(eip_env%force_env_input)
      END IF
      IF (ASSOCIATED(eip_env%cell_ref)) THEN
         CALL cell_release(eip_env%cell_ref)
      END IF
   END SUBROUTINE eip_env_clear

! **************************************************************************************************
!> \brief Creates the eip environment
!> \param eip_env The eip environment to be created
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_env_create(eip_env)

      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (eip_env)

      NULLIFY (eip_env%eip_forces)
      NULLIFY (eip_env%subsys)
      NULLIFY (eip_env%eip_input)
      NULLIFY (eip_env%force_env_input)
      NULLIFY (eip_env%cell_ref)

      eip_env%ref_count = 1
      last_eip_id = last_eip_id+1
      eip_env%id_nr = last_eip_id
      eip_env%use_ref_cell = .FALSE.
      CALL eip_env_clear(eip_env)
   END SUBROUTINE eip_env_create

END MODULE eip_environment_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (11.08.2002) exchange and correlation energy now in exc
!> \author MK (13.06.2002)
! **************************************************************************************************
MODULE fist_energy_types

   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_energy_types'

   PRIVATE

! **************************************************************************************************
   TYPE fist_energy_type
      REAL(kind=dp) :: kin, pot, e_gspace, e_self, e_neut, e_bonded, e_induction
      REAL(kind=dp) :: kin_shell, harm_shell
   END TYPE fist_energy_type

! *** Public data types ***

   PUBLIC :: fist_energy_type

! *** Public subroutines ***

   PUBLIC :: allocate_fist_energy, &
             deallocate_fist_energy

CONTAINS

! **************************************************************************************************
!> \brief   Allocate and/or initialise a Fist energy data structure.
!> \param fist_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_fist_energy(fist_energy)
      TYPE(fist_energy_type), POINTER                    :: fist_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_fist_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      IF (.NOT. ASSOCIATED(fist_energy)) THEN
         ALLOCATE (fist_energy, STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 57, &
                                          "fist_energy", 8*dp_size)
      END IF

      CALL init_fist_energy(fist_energy)

   END SUBROUTINE allocate_fist_energy

! **************************************************************************************************
!> \brief   Deallocate a Fist energy data structure.
!> \param fist_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_fist_energy(fist_energy)
      TYPE(fist_energy_type), POINTER                    :: fist_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_fist_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fist_energy)) THEN
         DEALLOCATE (fist_energy)
      ELSE
         CALL cp_abort(cp__l("fist_energy_types.F",81), &
                       "The fist_energy pointer is not associated "// &
                       "and cannot be deallocated.")
      END IF

   END SUBROUTINE deallocate_fist_energy

! **************************************************************************************************
!> \brief   Initialise a Fist energy data structure.
!> \param fist_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_fist_energy(fist_energy)
      TYPE(fist_energy_type), POINTER                    :: fist_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'init_fist_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fist_energy)) THEN
         fist_energy%kin = 0.0_dp
         fist_energy%pot = 0.0_dp
         fist_energy%e_gspace = 0.0_dp
         fist_energy%e_self = 0.0_dp
         fist_energy%e_neut = 0.0_dp
         fist_energy%e_bonded = 0.0_dp
         fist_energy%e_induction = 0.0_dp
         fist_energy%kin_shell = 0.0_dp
         fist_energy%harm_shell = 0.0_dp
      ELSE
         CALL cp_abort(cp__l("fist_energy_types.F",112), &
                       "The fist_energy pointer is not associated "// &
                       "and cannot be initialised.")
      END IF

   END SUBROUTINE init_fist_energy

END MODULE fist_energy_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>    gt sept-23-02 added atomic_kind_set to replica_environment_type
!>    to allow use of kind_based neighbor list
!>    CJM rewrite
!> \author CJM SEPT-01-02
! **************************************************************************************************
MODULE fist_environment_types
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_create,&
                                              atomic_kind_list_release,&
                                              atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_retain,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_release,&
                                              ewald_env_retain,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_release,&
                                              ewald_pw_retain,&
                                              ewald_pw_type
   USE exclusion_types,                 ONLY: exclusion_release,&
                                              exclusion_type
   USE fist_energy_types,               ONLY: deallocate_fist_energy,&
                                              fist_energy_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_release,&
                                              fist_nonbond_env_retain,&
                                              fist_nonbond_env_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_create,&
                                              mol_kind_new_list_release,&
                                              mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_create,&
                                              mol_new_list_release,&
                                              mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE multipole_types,                 ONLY: multipole_type
   USE particle_list_types,             ONLY: particle_list_create,&
                                              particle_list_release,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_create,&
                                              qmmm_env_mm_release,&
                                              qmmm_env_mm_retain,&
                                              qmmm_env_mm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   TYPE fist_environment_type
      PRIVATE
      INTEGER                                      :: id_nr, ref_count
      LOGICAL                                      :: qmmm
      LOGICAL                                      :: shell_model, shell_model_ad
      TYPE(qmmm_env_mm_type), POINTER             :: qmmm_env
      TYPE(cell_type), POINTER                    :: cell_ref
      TYPE(ewald_environment_type), POINTER       :: ewald_env
      TYPE(ewald_pw_type), POINTER       :: ewald_pw
      TYPE(fist_energy_type), POINTER             :: thermo
      TYPE(cp_para_env_type), POINTER             :: para_env
      TYPE(cp_subsys_type), POINTER               :: subsys
      TYPE(fist_nonbond_env_type), POINTER        :: fist_nonbond_env
      TYPE(section_vals_type), POINTER             :: input
      TYPE(exclusion_type), DIMENSION(:), POINTER  :: exclusions
   END TYPE fist_environment_type

! **************************************************************************************************
!> \brief to build arrays of pointers
!> \param fist_env the pointer to the fist_env
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   TYPE fist_environment_p_type
      TYPE(fist_environment_type), POINTER :: fist_env
   END TYPE fist_environment_p_type

! *** Public data types ***
   PUBLIC :: fist_environment_type

! *** Public subroutines ***
   PUBLIC :: fist_env_get, &
             fist_env_set, &
             fist_env_create, &
             fist_env_release, &
             fist_env_retain

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_environment_types'
   INTEGER, PRIVATE, SAVE :: last_fist_env_id_nr = 0

!***

CONTAINS

! **************************************************************************************************
!> \brief Purpose: Get the FIST environment.
!> \param fist_env the pointer to the fist_env
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param ewald_pw ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param cell ...
!> \param cell_ref ...
!> \param ewald_env ...
!> \param fist_nonbond_env ...
!> \param thermo ...
!> \param para_env ...
!> \param subsys ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param input ...
!> \param shell_model ...
!> \param shell_model_ad ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param multipoles ...
!> \param results ...
!> \param exclusions ...
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   SUBROUTINE fist_env_get(fist_env, atomic_kind_set, particle_set, ewald_pw, &
                           local_particles, local_molecules, molecule_kind_set, molecule_set, cell, &
                           cell_ref, ewald_env, fist_nonbond_env, thermo, para_env, subsys, qmmm, &
                           qmmm_env, input, shell_model, shell_model_ad, shell_particle_set, &
                           core_particle_set, multipoles, results, exclusions)

      TYPE(fist_environment_type), INTENT(IN)            :: fist_env
      TYPE(atomic_kind_type), OPTIONAL, POINTER          :: atomic_kind_set(:)
      TYPE(particle_type), OPTIONAL, POINTER             :: particle_set(:)
      TYPE(ewald_pw_type), OPTIONAL, POINTER             :: ewald_pw
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind_set(:)
      TYPE(molecule_type), OPTIONAL, POINTER             :: molecule_set(:)
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      TYPE(ewald_environment_type), OPTIONAL, POINTER    :: ewald_env
      TYPE(fist_nonbond_env_type), OPTIONAL, POINTER     :: fist_nonbond_env
      TYPE(fist_energy_type), OPTIONAL, POINTER          :: thermo
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      LOGICAL, OPTIONAL                                  :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      LOGICAL, OPTIONAL                                  :: shell_model, shell_model_ad
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      TYPE(multipole_type), OPTIONAL, POINTER            :: multipoles
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exclusions

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_env_get', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(multipole_type), POINTER                      :: fist_multipoles
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles

      NULLIFY (atomic_kinds, particles, molecules_new, molecule_kinds_new, fist_multipoles)

      IF (PRESENT(input)) input => fist_env%input
      IF (PRESENT(qmmm)) qmmm = fist_env%qmmm
      IF (PRESENT(qmmm_env)) qmmm_env => fist_env%qmmm_env
      IF (PRESENT(cell_ref)) cell_ref => fist_env%cell_ref
      IF (PRESENT(ewald_env)) ewald_env => fist_env%ewald_env
      IF (PRESENT(thermo)) thermo => fist_env%thermo
      IF (PRESENT(exclusions)) exclusions => fist_env%exclusions
      IF (PRESENT(para_env)) para_env => fist_env%para_env
      IF (PRESENT(ewald_pw)) ewald_pw => fist_env%ewald_pw
      IF (PRESENT(fist_nonbond_env)) fist_nonbond_env => fist_env%fist_nonbond_env
      IF (PRESENT(shell_model)) shell_model = fist_env%shell_model
      IF (PRESENT(shell_model_ad)) shell_model_ad = fist_env%shell_model_ad
      IF (PRESENT(subsys)) subsys => fist_env%subsys

      IF (ASSOCIATED(fist_env%subsys)) &
         CALL cp_subsys_get(fist_env%subsys, &
                            atomic_kinds=atomic_kinds, &
                            local_molecules_new=local_molecules, &
                            local_particles=local_particles, &
                            particles=particles, &
                            molecule_kinds_new=molecule_kinds_new, &
                            molecules_new=molecules_new, &
                            shell_particles=shell_particles, &
                            core_particles=core_particles, &
                            multipoles=fist_multipoles, &
                            results=results, &
                            cell=cell)
      IF (PRESENT(atomic_kind_set)) atomic_kind_set => atomic_kinds%els
      IF (PRESENT(particle_set)) particle_set => particles%els
      IF (PRESENT(molecule_kind_set)) molecule_kind_set => molecule_kinds_new%els
      IF (PRESENT(molecule_set)) molecule_set => molecules_new%els
      IF (PRESENT(shell_particle_set)) shell_particle_set => shell_particles%els
      IF (PRESENT(core_particle_set)) core_particle_set => core_particles%els
      IF (PRESENT(multipoles)) multipoles => fist_multipoles
   END SUBROUTINE fist_env_get

! **************************************************************************************************
!> \brief Initialise the FIST environment.
!> \param fist_env the pointer to the fist_env
!> \param para_env ...
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   SUBROUTINE init_fist_env(fist_env, para_env)

      TYPE(fist_environment_type), INTENT(OUT)           :: fist_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      NULLIFY (fist_env%input)
      NULLIFY (fist_env%qmmm_env)
      NULLIFY (fist_env%cell_ref)
      NULLIFY (fist_env%ewald_env)
      NULLIFY (fist_env%ewald_pw)
      NULLIFY (fist_env%thermo)
      NULLIFY (fist_env%fist_nonbond_env)
      NULLIFY (fist_env%subsys)
      NULLIFY (fist_env%exclusions)
      fist_env%qmmm = .FALSE.
      fist_env%shell_model = .FALSE.
      fist_env%shell_model_ad = .FALSE.
      CALL qmmm_env_mm_create(fist_env%qmmm_env)
      NULLIFY (fist_env%subsys)
      CALL cp_para_env_retain(para_env)
      fist_env%para_env => para_env
      fist_env%ref_count = 1
      last_fist_env_id_nr = last_fist_env_id_nr+1
      fist_env%id_nr = last_fist_env_id_nr

   END SUBROUTINE init_fist_env

! **************************************************************************************************
!> \brief Set the FIST environment.
!> \param fist_env the pointer to the fist_env
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param ewald_pw ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param cell_ref ...
!> \param ewald_env ...
!> \param fist_nonbond_env ...
!> \param thermo ...
!> \param subsys ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param input ...
!> \param shell_model ...
!> \param shell_model_ad ...
!> \param exclusions ...
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   SUBROUTINE fist_env_set(fist_env, atomic_kind_set, particle_set, ewald_pw, &
                           local_particles, local_molecules, molecule_kind_set, &
                           molecule_set, cell_ref, ewald_env, &
                           fist_nonbond_env, thermo, subsys, qmmm, qmmm_env, &
                           input, shell_model, shell_model_ad, exclusions)

      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(atomic_kind_type), OPTIONAL, POINTER          :: atomic_kind_set(:)
      TYPE(particle_type), OPTIONAL, POINTER             :: particle_set(:)
      TYPE(ewald_pw_type), OPTIONAL, POINTER             :: ewald_pw
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind_set(:)
      TYPE(molecule_type), OPTIONAL, POINTER             :: molecule_set(:)
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_ref
      TYPE(ewald_environment_type), OPTIONAL, POINTER    :: ewald_env
      TYPE(fist_nonbond_env_type), OPTIONAL, POINTER     :: fist_nonbond_env
      TYPE(fist_energy_type), OPTIONAL, POINTER          :: thermo
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      LOGICAL, OPTIONAL                                  :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      LOGICAL, OPTIONAL                                  :: shell_model, shell_model_ad
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exclusions

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_env_set', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles

      IF(.NOT.(fist_env%ref_count > 0))CALL cp__a("fist_environment_types.F",319)

      IF (PRESENT(qmmm)) fist_env%qmmm = qmmm
      IF (PRESENT(qmmm_env)) THEN
         CALL qmmm_env_mm_retain(qmmm_env)
         CALL qmmm_env_mm_release(fist_env%qmmm_env)
         fist_env%qmmm_env => qmmm_env
      END IF
      IF (PRESENT(ewald_env)) THEN
         CALL ewald_env_retain(ewald_env)
         CALL ewald_env_release(fist_env%ewald_env)
         fist_env%ewald_env => ewald_env
      ENDIF
      IF (PRESENT(ewald_pw)) THEN
         CALL ewald_pw_retain(ewald_pw)
         CALL ewald_pw_release(fist_env%ewald_pw)
         fist_env%ewald_pw => ewald_pw
      ENDIF
      IF (PRESENT(cell_ref)) THEN
         CALL cell_retain(cell_ref)
         CALL cell_release(fist_env%cell_ref)
         fist_env%cell_ref => cell_ref
      END IF
      IF (PRESENT(fist_nonbond_env)) THEN
         CALL fist_nonbond_env_retain(fist_nonbond_env)
         CALL fist_nonbond_env_release(fist_env%fist_nonbond_env)
         fist_env%fist_nonbond_env => fist_nonbond_env
      ENDIF
      IF (PRESENT(input)) THEN
         CALL section_vals_retain(input)
         CALL section_vals_release(fist_env%input)
         fist_env%input => input
      END IF
      IF (PRESENT(thermo)) fist_env%thermo => thermo
      IF (PRESENT(subsys)) THEN
         CALL cp_subsys_retain(subsys)
         CALL cp_subsys_release(fist_env%subsys)
         fist_env%subsys => subsys
      END IF
      IF (PRESENT(atomic_kind_set)) THEN
         CALL atomic_kind_list_create(atomic_kinds, &
                                      els_ptr=atomic_kind_set)
         CALL cp_subsys_set(fist_env%subsys, &
                            atomic_kinds=atomic_kinds)
         CALL atomic_kind_list_release(atomic_kinds)
      END IF
      IF (PRESENT(particle_set)) THEN
         CALL particle_list_create(particles, &
                                   els_ptr=particle_set)
         CALL cp_subsys_set(fist_env%subsys, &
                            particles=particles)
         CALL particle_list_release(particles)
      END IF
      IF (PRESENT(local_particles)) THEN
         CALL cp_subsys_set(fist_env%subsys, &
                            local_particles=local_particles)
      END IF
      IF (PRESENT(local_molecules)) THEN
         CALL cp_subsys_set(fist_env%subsys, &
                            local_molecules_new=local_molecules)
      END IF
      IF (PRESENT(molecule_kind_set)) THEN
         CALL mol_kind_new_list_create(molecule_kinds_new, &
                                       els_ptr=molecule_kind_set)
         CALL cp_subsys_set(fist_env%subsys, &
                            molecule_kinds_new=molecule_kinds_new)
         CALL mol_kind_new_list_release(molecule_kinds_new)
      END IF
      IF (PRESENT(molecule_set)) THEN
         CALL mol_new_list_create(molecules_new, &
                                  els_ptr=molecule_set)
         CALL cp_subsys_set(fist_env%subsys, &
                            molecules_new=molecules_new)
         CALL mol_new_list_release(molecules_new)
      END IF
      IF (PRESENT(exclusions)) fist_env%exclusions => exclusions
      IF (PRESENT(shell_model)) THEN
         fist_env%shell_model = shell_model
      END IF
      IF (PRESENT(shell_model_ad)) THEN
         fist_env%shell_model_ad = shell_model_ad
      END IF

   END SUBROUTINE fist_env_set

! **************************************************************************************************
!> \brief allocates and intitializes a fist_env
!> \param fist_env the object to create
!> \param para_env the parallel environement for the qs_env
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_env_create(fist_env, para_env)
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_env_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (fist_env)
      CALL init_fist_env(fist_env, para_env=para_env)
   END SUBROUTINE fist_env_create

! **************************************************************************************************
!> \brief retains the given fist_env (see doc/ReferenceCounting.html)
!> \param fist_env the object to retain
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_env_retain(fist_env)
      TYPE(fist_environment_type), POINTER               :: fist_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fist_env)))CALL cp__a("fist_environment_types.F",436)
      IF(.NOT.(fist_env%ref_count > 0))CALL cp__a("fist_environment_types.F",437)
      fist_env%ref_count = fist_env%ref_count+1
   END SUBROUTINE fist_env_retain

! **************************************************************************************************
!> \brief releases the given fist_env (see doc/ReferenceCounting.html)
!> \param fist_env the object to release
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_env_release(fist_env)
      TYPE(fist_environment_type), POINTER               :: fist_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fist_env)) THEN
         IF(.NOT.(fist_env%ref_count > 0))CALL cp__a("fist_environment_types.F",455)
         fist_env%ref_count = fist_env%ref_count-1
         IF (fist_env%ref_count < 1) THEN
            CALL qmmm_env_mm_release(fist_env%qmmm_env)
            CALL cell_release(fist_env%cell_ref)
            CALL ewald_pw_release(fist_env%ewald_pw)
            CALL ewald_env_release(fist_env%ewald_env)
            CALL cp_para_env_release(fist_env%para_env)
            CALL deallocate_fist_energy(fist_env%thermo)

            CALL fist_nonbond_env_release(fist_env%fist_nonbond_env)
            CALL cp_subsys_release(fist_env%subsys)
            CALL section_vals_release(fist_env%input)
            CALL exclusion_release(fist_env%exclusions)

            DEALLOCATE (fist_env)
         END IF
      END IF
      NULLIFY (fist_env)
   END SUBROUTINE fist_env_release

END MODULE fist_environment_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Basic container type for QM/MM.
!> \author Ole Schuett
! **************************************************************************************************
MODULE qmmm_types
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE fist_energy_types,               ONLY: fist_energy_type
   USE fist_environment_types,          ONLY: fist_env_get,&
                                              fist_env_release,&
                                              fist_environment_type
   USE kinds,                           ONLY: dp
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_release,&
                                              qmmm_env_qm_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_env_release,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_types'

   PUBLIC :: qmmm_env_type, qmmm_env_retain, qmmm_env_release, qmmm_env_get

   TYPE qmmm_env_type
      INTEGER                                                 :: ref_count = 1
      TYPE(qs_environment_type), POINTER                      :: qs_env => Null()
      TYPE(fist_environment_type), POINTER                    :: fist_env => Null()
      TYPE(qmmm_env_qm_type), POINTER                         :: qm => Null()
   END TYPE qmmm_env_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \param subsys ...
!> \param potential_energy ...
!> \param kinetic_energy ...
! **************************************************************************************************
   SUBROUTINE qmmm_env_get(qmmm_env, subsys, potential_energy, kinetic_energy)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: potential_energy, kinetic_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_get', routineP = moduleN//':'//routineN

      TYPE(fist_energy_type), POINTER                    :: thermo
      TYPE(qs_energy_type), POINTER                      :: qs_energy

      NULLIFY (qs_energy, thermo)

      IF(.NOT.(ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_types.F",60)
      IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types.F",61)

      IF (PRESENT(kinetic_energy)) THEN
         CALL fist_env_get(qmmm_env%fist_env, thermo=thermo)
         kinetic_energy = thermo%kin
      END IF
      IF (PRESENT(subsys)) THEN
         CALL fist_env_get(qmmm_env%fist_env, subsys=subsys)
      ENDIF
      IF (PRESENT(potential_energy)) THEN
         ! get the underlying energies from primary subsys.  This is the only subsys
         ! for conventional QM/MM, and force-mixing knows to put relevant energy there.
         CALL fist_env_get(qmmm_env%fist_env, thermo=thermo)
         CALL get_qs_env(qmmm_env%qs_env, energy=qs_energy)
         potential_energy = thermo%pot+qs_energy%total
      ENDIF
   END SUBROUTINE qmmm_env_get

! **************************************************************************************************
!> \brief releases the given qmmm_env (see doc/ReferenceCounting.html)
!> \param qmmm_env the object to release
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qmmm_env_release(qmmm_env)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qmmm_env)) THEN
         IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types.F",91)
         qmmm_env%ref_count = qmmm_env%ref_count-1
         IF (qmmm_env%ref_count == 0) THEN
            CALL qs_env_release(qmmm_env%qs_env)
            CALL qmmm_env_qm_release(qmmm_env%qm)
            CALL fist_env_release(qmmm_env%fist_env)
            DEALLOCATE (qmmm_env)
         END IF
      END IF
      NULLIFY (qmmm_env)
   END SUBROUTINE qmmm_env_release

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qmmm_env_retain(qmmm_env)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_types.F",114)
      IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types.F",115)
      qmmm_env%ref_count = qmmm_env%ref_count+1
   END SUBROUTINE qmmm_env_retain

END MODULE qmmm_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Basic container type for QM/MM with force mixing.
!> \author Ole Schuett
! **************************************************************************************************
MODULE qmmmx_types
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE kinds,                           ONLY: dp
   USE qmmm_types,                      ONLY: qmmm_env_get,&
                                              qmmm_env_release,&
                                              qmmm_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmmx_types'

   PUBLIC :: qmmmx_env_type, qmmmx_env_get, qmmmx_env_retain, qmmmx_env_release

   TYPE qmmmx_env_type
      INTEGER                                                 :: ref_count = 1
      TYPE(qmmm_env_type), POINTER                            :: core => Null()
      TYPE(qmmm_env_type), POINTER                            :: ext => Null()
   END TYPE qmmmx_env_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qmmmx_env ...
!> \param subsys ...
!> \param potential_energy ...
!> \param kinetic_energy ...
! **************************************************************************************************
   SUBROUTINE qmmmx_env_get(qmmmx_env, subsys, potential_energy, kinetic_energy)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: potential_energy, kinetic_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_env_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qmmmx_env)))CALL cp__a("qmmmx_types.F",47)
      IF(.NOT.(qmmmx_env%ref_count > 0))CALL cp__a("qmmmx_types.F",48)

      ! get the underlying energies from primary subsys.  This is the only subsys
      ! for conventional QM/MM, and force-mixing knows to put relevant energy there.
      CALL qmmm_env_get(qmmmx_env%ext, &
                        kinetic_energy=kinetic_energy, &
                        potential_energy=potential_energy, &
                        subsys=subsys)

   END SUBROUTINE qmmmx_env_get

! **************************************************************************************************
!> \brief ...
!> \param qmmmx_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qmmmx_env_retain(qmmmx_env)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qmmmx_env)))CALL cp__a("qmmmx_types.F",70)
      IF(.NOT.(qmmmx_env%ref_count > 0))CALL cp__a("qmmmx_types.F",71)
      qmmmx_env%ref_count = qmmmx_env%ref_count+1
   END SUBROUTINE qmmmx_env_retain

! **************************************************************************************************
!> \brief releases the given qmmmx_env (see doc/ReferenceCounting.html)
!> \param qmmmx_env the object to release
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qmmmx_env_release(qmmmx_env)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qmmmx_env)) THEN
         IF(.NOT.(qmmmx_env%ref_count > 0))CALL cp__a("qmmmx_types.F",87)
         qmmmx_env%ref_count = qmmmx_env%ref_count-1
         IF (qmmmx_env%ref_count == 0) THEN
            CALL qmmm_env_release(qmmmx_env%core)
            CALL qmmm_env_release(qmmmx_env%ext)
            DEALLOCATE (qmmmx_env)
         END IF
      END IF
      NULLIFY (qmmmx_env)
   END SUBROUTINE qmmmx_env_release

END MODULE qmmmx_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author fschiff
!> \date   11.06
! **************************************************************************************************
MODULE mixed_energy_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mixed_energy_types'
   PRIVATE

! **************************************************************************************************
   TYPE mixed_energy_type
      REAL(kind=dp) :: pot
      REAL(kind=dp) :: kin
   END TYPE mixed_energy_type

! **************************************************************************************************
   TYPE mixed_force_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER         :: forces
   END TYPE mixed_force_type

! Public data types
   PUBLIC :: mixed_energy_type, &
             mixed_force_type

! Public subroutines
   PUBLIC :: allocate_mixed_energy, &
             deallocate_mixed_energy

CONTAINS

! **************************************************************************************************
!> \brief   Allocate and/or initialise a mixed energy data structure.
!> \param mixed_energy ...
!> \date    11.06
!> \author  fschiff
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_mixed_energy(mixed_energy)
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_mixed_energy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(mixed_energy)) THEN
         ALLOCATE (mixed_energy)
      END IF
      CALL init_mixed_energy(mixed_energy)
   END SUBROUTINE allocate_mixed_energy

! **************************************************************************************************
!> \brief   Deallocate a mixed energy data structure.
!> \param mixed_energy ...
!> \date    11.06
!> \author  fschiff
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_mixed_energy(mixed_energy)
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_mixed_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mixed_energy)) THEN
         DEALLOCATE (mixed_energy)
      END IF
   END SUBROUTINE deallocate_mixed_energy

! **************************************************************************************************
!> \brief ...
!> \param mixed_energy ...
! **************************************************************************************************
   SUBROUTINE init_mixed_energy(mixed_energy)
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'init_mixed_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mixed_energy)) THEN
         mixed_energy%pot = 0.0_dp
      ELSE
         CALL cp_abort(cp__l("mixed_energy_types.F",90), &
                       "The mixed_energy pointer is not associated "// &
                       "and cannot be initialised")
      END IF
   END SUBROUTINE init_mixed_energy

END MODULE mixed_energy_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author fschiff SEPT-11-06
! **************************************************************************************************
MODULE mixed_environment_types
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_create,&
                                              atomic_kind_list_release,&
                                              atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE cp_log_handling,                 ONLY: cp_logger_p_type,&
                                              cp_logger_release
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_p_type,&
                                              cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_retain,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mixed_energy_types,              ONLY: deallocate_mixed_energy,&
                                              mixed_energy_type
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_create,&
                                              mol_kind_new_list_release,&
                                              mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_create,&
                                              mol_new_list_release,&
                                              mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_list_types,             ONLY: particle_list_create,&
                                              particle_list_release,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \param mixed_env the pointer to the mixed_env
!> \par History
!>      11/06
!> \author fschiff
! **************************************************************************************************
   TYPE mixed_environment_type
      INTEGER :: id_nr, ref_count
      TYPE(cell_type), POINTER                      :: cell_ref
      TYPE(mixed_energy_type), POINTER              :: mixed_energy
      TYPE(cp_para_env_type), POINTER               :: para_env
      TYPE(cp_subsys_type), POINTER                 :: subsys
      TYPE(section_vals_type), POINTER                 :: input
      REAL(KIND=dp), DIMENSION(:), POINTER             :: energies
      ! Parallelization of multiple force_eval
      INTEGER                                          :: new_group, ngroups
      INTEGER, DIMENSION(:), POINTER                   :: group_distribution
      TYPE(cp_para_env_p_type), DIMENSION(:), POINTER  :: sub_para_env
      TYPE(cp_logger_p_type), DIMENSION(:), POINTER   :: sub_logger
      REAL(KIND=dp), POINTER, DIMENSION(:)             :: val
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                       :: par
      REAL(KIND=dp)                                    :: dx, lerr
      CHARACTER(default_path_length)                   :: coupling_function
   END TYPE mixed_environment_type

! **************************************************************************************************
!> \brief to build arrays of pointers
!> \param mixed_env the pointer to the mixed_env
!> \par History
!>      11/06
!> \author fschiff
! **************************************************************************************************
   TYPE mixed_environment_p_type
      TYPE(mixed_environment_type), POINTER :: mixed_env
   END TYPE mixed_environment_p_type

! *** Public data types ***

   PUBLIC :: mixed_environment_type

! *** Public subroutines ***

   PUBLIC :: get_mixed_env, &
             set_mixed_env, &
             mixed_env_create, &
             mixed_env_release, &
             mixed_env_retain

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mixed_environment_types'
   INTEGER, PRIVATE, SAVE :: last_mixed_env_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief Get the MIXED environment.
!> \param mixed_env the pointer to the mixed_env
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param cell ...
!> \param cell_ref ...
!> \param mixed_energy ...
!> \param para_env ...
!> \param sub_para_env ...
!> \param subsys ...
!> \param input ...
!> \param results ...
! **************************************************************************************************
   SUBROUTINE get_mixed_env(mixed_env, atomic_kind_set, particle_set, &
                            local_particles, local_molecules, molecule_kind_set, &
                            molecule_set, cell, cell_ref, &
                            mixed_energy, para_env, sub_para_env, subsys, &
                            input, results)

      TYPE(mixed_environment_type), INTENT(IN)           :: mixed_env
      TYPE(atomic_kind_type), OPTIONAL, POINTER          :: atomic_kind_set(:)
      TYPE(particle_type), OPTIONAL, POINTER             :: particle_set(:)
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind_set(:)
      TYPE(molecule_type), OPTIONAL, POINTER             :: molecule_set(:)
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      TYPE(mixed_energy_type), OPTIONAL, POINTER         :: mixed_energy
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_para_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: sub_para_env
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'get_mixed_env', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles

      NULLIFY (atomic_kinds, particles, molecules_new, molecule_kinds_new)
      IF(.NOT.(ASSOCIATED(mixed_env%subsys)))CALL cp__a("mixed_environment_types.F",156)

      IF (PRESENT(input)) input => mixed_env%input
      IF (PRESENT(cell_ref)) cell_ref => mixed_env%cell_ref
      IF (PRESENT(mixed_energy)) mixed_energy => mixed_env%mixed_energy
      IF (PRESENT(para_env)) para_env => mixed_env%para_env
      IF (PRESENT(sub_para_env)) sub_para_env => mixed_env%sub_para_env
      IF (PRESENT(subsys)) subsys => mixed_env%subsys
      CALL cp_subsys_get(mixed_env%subsys, &
                         atomic_kinds=atomic_kinds, &
                         local_molecules_new=local_molecules, &
                         local_particles=local_particles, &
                         particles=particles, &
                         molecule_kinds_new=molecule_kinds_new, &
                         molecules_new=molecules_new, &
                         results=results, &
                         cell=cell)
      IF (PRESENT(atomic_kind_set)) atomic_kind_set => atomic_kinds%els
      IF (PRESENT(particle_set)) particle_set => particles%els
      IF (PRESENT(molecule_kind_set)) molecule_kind_set => molecule_kinds_new%els
      IF (PRESENT(molecule_set)) molecule_set => molecules_new%els

   END SUBROUTINE get_mixed_env

! **************************************************************************************************
!> \brief Initialise the MIXED environment.
!> \param mixed_env the pointer to the mixed_env
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE init_mixed_env(mixed_env, para_env)

      TYPE(mixed_environment_type), INTENT(OUT)          :: mixed_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      NULLIFY (mixed_env%input)
      NULLIFY (mixed_env%cell_ref)
      NULLIFY (mixed_env%mixed_energy)
      NULLIFY (mixed_env%subsys)
      NULLIFY (mixed_env%sub_para_env)
      NULLIFY (mixed_env%sub_logger)
      NULLIFY (mixed_env%energies)
      NULLIFY (mixed_env%par)
      NULLIFY (mixed_env%val)
      NULLIFY (mixed_env%subsys)
      CALL cp_para_env_retain(para_env)
      mixed_env%para_env => para_env
      mixed_env%ref_count = 1
      last_mixed_env_id_nr = last_mixed_env_id_nr+1
      mixed_env%id_nr = last_mixed_env_id_nr

   END SUBROUTINE init_mixed_env

! **************************************************************************************************
!> \brief Set the MIXED environment.
!> \param mixed_env the pointer to the mixed_env
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param cell_ref ...
!> \param mixed_energy ...
!> \param subsys ...
!> \param input ...
!> \param sub_para_env ...
! **************************************************************************************************
   SUBROUTINE set_mixed_env(mixed_env, atomic_kind_set, particle_set, &
                            local_particles, local_molecules, molecule_kind_set, &
                            molecule_set, cell_ref, mixed_energy, subsys, &
                            input, sub_para_env)

      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(atomic_kind_type), OPTIONAL, POINTER          :: atomic_kind_set(:)
      TYPE(particle_type), OPTIONAL, POINTER             :: particle_set(:)
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind_set(:)
      TYPE(molecule_type), OPTIONAL, POINTER             :: molecule_set(:)
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_ref
      TYPE(mixed_energy_type), OPTIONAL, POINTER         :: mixed_energy
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      TYPE(cp_para_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: sub_para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'set_mixed_env', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles

      IF(.NOT.(mixed_env%ref_count > 0))CALL cp__a("mixed_environment_types.F",248)

      IF (PRESENT(cell_ref)) THEN
         CALL cell_retain(cell_ref)
         CALL cell_release(mixed_env%cell_ref)
         mixed_env%cell_ref => cell_ref
      END IF
      IF (PRESENT(input)) THEN
         CALL section_vals_retain(input)
         CALL section_vals_release(mixed_env%input)
         mixed_env%input => input
      END IF
      IF (PRESENT(mixed_energy)) mixed_env%mixed_energy => mixed_energy
      IF (PRESENT(subsys)) THEN
         CALL cp_subsys_retain(subsys)
         CALL cp_subsys_release(mixed_env%subsys)
         mixed_env%subsys => subsys
      END IF
      IF (PRESENT(sub_para_env)) THEN
         mixed_env%sub_para_env => sub_para_env
      END IF
      IF (PRESENT(atomic_kind_set)) THEN
         CALL atomic_kind_list_create(atomic_kinds, &
                                      els_ptr=atomic_kind_set)
         CALL cp_subsys_set(mixed_env%subsys, &
                            atomic_kinds=atomic_kinds)
         CALL atomic_kind_list_release(atomic_kinds)
      END IF
      IF (PRESENT(particle_set)) THEN
         CALL particle_list_create(particles, &
                                   els_ptr=particle_set)
         CALL cp_subsys_set(mixed_env%subsys, &
                            particles=particles)
         CALL particle_list_release(particles)
      END IF
      IF (PRESENT(local_particles)) THEN
         CALL cp_subsys_set(mixed_env%subsys, &
                            local_particles=local_particles)
      END IF
      IF (PRESENT(local_molecules)) THEN
         CALL cp_subsys_set(mixed_env%subsys, &
                            local_molecules_new=local_molecules)
      END IF
      IF (PRESENT(molecule_kind_set)) THEN
         CALL mol_kind_new_list_create(molecule_kinds_new, &
                                       els_ptr=molecule_kind_set)
         CALL cp_subsys_set(mixed_env%subsys, &
                            molecule_kinds_new=molecule_kinds_new)
         CALL mol_kind_new_list_release(molecule_kinds_new)
      END IF
      IF (PRESENT(molecule_set)) THEN
         CALL mol_new_list_create(molecules_new, &
                                  els_ptr=molecule_set)
         CALL cp_subsys_set(mixed_env%subsys, &
                            molecules_new=molecules_new)
         CALL mol_new_list_release(molecules_new)
      END IF

   END SUBROUTINE set_mixed_env

! **************************************************************************************************
!> \brief allocates and intitializes a mixed_env
!> \param mixed_env the object to create
!> \param para_env the parallel environement for the qs_env
!> \author fschiff 11.06
! **************************************************************************************************
   SUBROUTINE mixed_env_create(mixed_env, para_env)
      TYPE(mixed_environment_type), POINTER              :: mixed_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_env_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (mixed_env)
      CALL init_mixed_env(mixed_env, para_env=para_env)
   END SUBROUTINE mixed_env_create

! **************************************************************************************************
!> \brief retains the given mixed_env (see doc/ReferenceCounting.html)
!> \param mixed_env the object to retain
!> \author fschiff 11.06
! **************************************************************************************************
   SUBROUTINE mixed_env_retain(mixed_env)
      TYPE(mixed_environment_type), POINTER              :: mixed_env

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(mixed_env)))CALL cp__a("mixed_environment_types.F",336)
      IF(.NOT.(mixed_env%ref_count > 0))CALL cp__a("mixed_environment_types.F",337)
      mixed_env%ref_count = mixed_env%ref_count+1
   END SUBROUTINE mixed_env_retain

! **************************************************************************************************
!> \brief releases the given mixed_env (see doc/ReferenceCounting.html)
!> \param mixed_env the object to release
!> \author fschiff 11.06
! **************************************************************************************************
   SUBROUTINE mixed_env_release(mixed_env)
      TYPE(mixed_environment_type), POINTER              :: mixed_env

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ngroups

      IF (ASSOCIATED(mixed_env)) THEN
         IF(.NOT.(mixed_env%ref_count > 0))CALL cp__a("mixed_environment_types.F",355)
         mixed_env%ref_count = mixed_env%ref_count-1
         IF (mixed_env%ref_count < 1) THEN
            ngroups = SIZE(mixed_env%sub_para_env)
            DO i = 1, ngroups
               IF (ASSOCIATED(mixed_env%sub_para_env(i)%para_env)) THEN
                  CALL cp_logger_release(mixed_env%sub_logger(i)%p)
                  CALL cp_para_env_release(mixed_env%sub_para_env(i)%para_env)
               END IF
            END DO
            DEALLOCATE (mixed_env%sub_para_env)
            DEALLOCATE (mixed_env%sub_logger)
            DEALLOCATE (mixed_env%energies)
            IF (ASSOCIATED(mixed_env%par)) THEN
               DEALLOCATE (mixed_env%par)
            END IF
            IF (ASSOCIATED(mixed_env%val)) THEN
               DEALLOCATE (mixed_env%val)
            END IF
            CALL cell_release(mixed_env%cell_ref)
            CALL cp_para_env_release(mixed_env%para_env)
            CALL deallocate_mixed_energy(mixed_env%mixed_energy)
            CALL cp_subsys_release(mixed_env%subsys)
            CALL section_vals_release(mixed_env%input)
            IF (ASSOCIATED(mixed_env%group_distribution)) THEN
               DEALLOCATE (mixed_env%group_distribution)
            END IF
            DEALLOCATE (mixed_env)
         END IF
      END IF
      NULLIFY (mixed_env)

   END SUBROUTINE mixed_env_release

END MODULE mixed_environment_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief defines types for metadynamics calculation
!> \par History
!>      01.2005 created [fawzi and ale]
! **************************************************************************************************
MODULE metadynamics_types
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE parallel_rng_types,              ONLY: delete_rng_stream,&
                                              rng_stream_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER, SAVE, PRIVATE               :: last_meta_env_id = 0
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'metadynamics_types'

   PUBLIC :: metadyn_create, &
             hills_env_type, &
             meta_env_type, &
             meta_env_retain, &
             set_meta_env, &
             meta_env_release, &
             metavar_type, &
             multiple_walkers_type

! **************************************************************************************************
!> \brief defines types for HILLS
! **************************************************************************************************
   TYPE hills_env_type
      LOGICAL                                   :: restart, slow_growth, wtcontrol
      !RG Adaptive hills
      REAL(KIND=dp)                             :: ww, min_disp, energy
      INTEGER                                   :: n_hills, nt_hills, min_nt_hills
      INTEGER                                   :: old_hill_number, old_hill_step
      !RG Adaptive hills
      REAL(KIND=dp), DIMENSION(:, :), POINTER    :: ss_history
      REAL(KIND=dp), DIMENSION(:, :), POINTER    :: delta_s_history
      REAL(KIND=dp), DIMENSION(:), POINTER    :: ww_history
      REAL(KIND=dp), DIMENSION(:), POINTER    :: invdt_history
      !Hills tail damping
      REAL(KIND=dp)                             :: tail_cutoff
      INTEGER                                   :: p_exp, q_exp
   END TYPE hills_env_type

! **************************************************************************************************
!> \brief defines types for WALLS
! **************************************************************************************************
   TYPE wall_type
      INTEGER                                   :: id_type, id_direction
      REAL(KIND=dp)                             :: pos, pos0
      REAL(KIND=dp)                             :: k_quadratic, k_quartic, ww_gauss, sigma_gauss
   END TYPE wall_type

! **************************************************************************************************
!> \brief defines types for COLVAR used in the metadynamics
! **************************************************************************************************
   TYPE metavar_type
      INTEGER                                   :: icolvar
      LOGICAL                                   :: do_wall, periodic
      REAL(KIND=dp)                             :: mass, lambda, vvp
      REAL(KIND=dp)                             :: gamma
      REAL(KIND=dp)                             :: epot_s, delta_s, epot_walls
      REAL(KIND=dp)                             :: ss, ss0, ff_s, ff_hills, ff_walls
      TYPE(wall_type), DIMENSION(:), POINTER    :: walls
   END TYPE metavar_type

! **************************************************************************************************
!> \brief defines types for multiple walkers run
! **************************************************************************************************
   TYPE multiple_walkers_type
      INTEGER                                   :: n_hills_local
      INTEGER                                   :: walker_id
      INTEGER                                   :: walkers_tot_nr
      INTEGER                                   :: walkers_freq_comm
      INTEGER, DIMENSION(:), POINTER            :: walkers_status
      CHARACTER(LEN=default_path_length), &
         DIMENSION(:), POINTER                :: walkers_file_name
   END TYPE multiple_walkers_type

! **************************************************************************************************
!> \brief defines meta_env type
! **************************************************************************************************
   TYPE meta_env_type
      INTEGER                                   :: ref_count, id_nr
      LOGICAL                                   :: do_hills, do_multiple_walkers
      LOGICAL                                   :: extended_lagrange
      LOGICAL                                   :: well_tempered
      LOGICAL                                   :: langevin
      LOGICAL                                   :: use_plumed
      CHARACTER(LEN=default_path_length)        :: plumed_input_file
      INTEGER                                   :: n_colvar
      REAL(KIND=dp)                             :: ekin_s, epot_s, dt, &
                                                   avg_temp, epot_walls
      LOGICAL                                   :: tempcontrol, restart
      REAL(KIND=dp)                             :: temp_wanted, toll_temp
      REAL(KIND=dp)                             :: delta_t, invdt, &
                                                   wtgamma, wttemperature
      INTEGER                                   :: n_steps
      ! time pointer should never be allocated itself.. that's a quite bad
      ! choice derived from the md_env.. So whenever the meta_env type is
      ! used the pointer time should be pointer to the high level time counter
      REAL(KIND=dp), POINTER                    :: time
      TYPE(hills_env_type), POINTER             :: hills_env
      TYPE(metavar_type), POINTER, DIMENSION(:) :: metavar
      TYPE(multiple_walkers_type), POINTER      :: multiple_walkers
      TYPE(cp_para_env_type), POINTER           :: para_env
      TYPE(section_vals_type), POINTER          :: metadyn_section
      TYPE(rng_stream_p_type), DIMENSION(:), &
         POINTER :: rng
      INTEGER                                   :: TAMCSteps
      REAL(kind=dp)                             :: zdt
   END TYPE meta_env_type

CONTAINS

! **************************************************************************************************
!> \brief allocates a metadynamic environment (performs only minimal
!>      initialization)
!> \param meta_env the meta env_ that will be allocated
!> \param n_colvar number of collectiva variables
!> \param dt ...
!> \param para_env ...
!> \param metadyn_section ...
!> \par History
!>      04.2004 created
!>      02.2006 Reorganized the structure of the restart for Metadynamics (teo)
!>              cleaned the metadynamic type
!> \author - alessandro laio and fawzi mohamed
!>         - Teodoro Laino [tlaino] - University of Zurich. 11.2007
!>         - Teodoro Laino [tlaino] - University of Zurich. 10.2008
!>           Major rewriting and addition of multiple walkers
! **************************************************************************************************
   SUBROUTINE metadyn_create(meta_env, n_colvar, dt, para_env, metadyn_section)
      TYPE(meta_env_type), POINTER                       :: meta_env
      INTEGER, INTENT(in)                                :: n_colvar
      REAL(dp), INTENT(in)                               :: dt
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: metadyn_section

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: do_langevin

      IF(.NOT.(.NOT. ASSOCIATED(meta_env)))CALL cp__a("metadynamics_types.F",158)
      ALLOCATE (meta_env)
      NULLIFY (meta_env%multiple_walkers, &
               meta_env%metadyn_section, &
               meta_env%time, &
               meta_env%rng, &
               meta_env%hills_env)

      meta_env%use_plumed = .FALSE.
      meta_env%plumed_input_file = ""
      meta_env%metadyn_section => metadyn_section
      meta_env%ref_count = 1
      meta_env%restart = .TRUE.
      last_meta_env_id = last_meta_env_id+1
      meta_env%id_nr = last_meta_env_id
      meta_env%n_colvar = n_colvar
      meta_env%para_env => para_env
      CALL cp_para_env_retain(para_env)

      meta_env%ekin_s = 0.0_dp
      meta_env%epot_s = 0.0_dp
      meta_env%epot_walls = 0.0_dp
      meta_env%n_steps = 0
      meta_env%dt = dt
      meta_env%tempcontrol = .FALSE.

      ! Hills_env
      ALLOCATE (meta_env%hills_env)
      ALLOCATE (meta_env%hills_env%ss_history(n_colvar, 0))
      ALLOCATE (meta_env%hills_env%delta_s_history(n_colvar, 0))
      ALLOCATE (meta_env%hills_env%ww_history(0))
      ALLOCATE (meta_env%hills_env%invdt_history(0))
      meta_env%hills_env%n_hills = 0
      meta_env%hills_env%energy = 0.0_dp
      meta_env%hills_env%restart = .TRUE.

      ! Colvar
      ALLOCATE (meta_env%metavar(n_colvar))
      DO i = 1, n_colvar
         NULLIFY (meta_env%metavar(i)%walls)
         meta_env%metavar(i)%mass = -HUGE(0.0_dp)
         meta_env%metavar(i)%lambda = -HUGE(0.0_dp)
         meta_env%metavar(i)%gamma = 0.0_dp
         meta_env%metavar(i)%ss = 0.0_dp
         meta_env%metavar(i)%ss0 = 0.0_dp
         meta_env%metavar(i)%ff_s = 0.0_dp
         meta_env%metavar(i)%vvp = 0.0_dp
         meta_env%metavar(i)%epot_s = 0.0_dp
         meta_env%metavar(i)%epot_walls = 0.0_dp
         meta_env%metavar(i)%delta_s = 0.0_dp
         meta_env%metavar(i)%ff_hills = 0.0_dp
         meta_env%metavar(i)%ff_walls = 0.0_dp
         meta_env%metavar(i)%do_wall = .FALSE.
         meta_env%metavar(i)%periodic = .FALSE.
         meta_env%metavar(i)%icolvar = 0
      END DO

      ! Multiple Walkers
      CALL section_vals_val_get(metadyn_section, "MULTIPLE_WALKERS%_SECTION_PARAMETERS_", &
                                l_val=meta_env%do_multiple_walkers)
      IF (meta_env%do_multiple_walkers) THEN
         ALLOCATE (meta_env%multiple_walkers)

         ! Walkers status and Walkers file name
         NULLIFY (meta_env%multiple_walkers%walkers_status, &
                  meta_env%multiple_walkers%walkers_file_name)
         meta_env%multiple_walkers%n_hills_local = 0
      END IF

      CALL section_vals_val_get(metadyn_section, "LANGEVIN", l_val=do_langevin)
      IF (do_langevin) THEN
         ALLOCATE (meta_env%rng(meta_env%n_colvar))
         DO i = 1, meta_env%n_colvar
            NULLIFY (meta_env%rng(meta_env%n_colvar)%stream)
         END DO
      ENDIF
   END SUBROUTINE metadyn_create

! **************************************************************************************************
!> \brief sets the meta_env
!> \param meta_env ...
!> \param time ...
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   SUBROUTINE set_meta_env(meta_env, time)
      TYPE(meta_env_type), POINTER                       :: meta_env
      REAL(KIND=dp), OPTIONAL, POINTER                   :: time

      CHARACTER(len=*), PARAMETER :: routineN = 'set_meta_env', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(meta_env)) THEN
         IF (PRESENT(time)) THEN
            NULLIFY (meta_env%time)
            meta_env%time => time
         END IF
      END IF
   END SUBROUTINE set_meta_env

! **************************************************************************************************
!> \brief retains the meta_env
!> \param meta_env ...
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   SUBROUTINE meta_env_retain(meta_env)
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(len=*), PARAMETER :: routineN = 'meta_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(meta_env)))CALL cp__a("metadynamics_types.F",267)
      IF(.NOT.(meta_env%ref_count > 0))CALL cp__a("metadynamics_types.F",268)
      meta_env%ref_count = meta_env%ref_count+1
   END SUBROUTINE meta_env_retain

! **************************************************************************************************
!> \brief releases the meta_env
!> \param meta_env ...
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   SUBROUTINE meta_env_release(meta_env)
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(len=*), PARAMETER :: routineN = 'meta_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(meta_env)) THEN
         IF(.NOT.(meta_env%ref_count > 0))CALL cp__a("metadynamics_types.F",286)
         meta_env%ref_count = meta_env%ref_count-1
         IF (meta_env%ref_count == 0) THEN
            CALL cp_para_env_release(meta_env%para_env)
            IF (ASSOCIATED(meta_env%metavar)) THEN
               DO i = 1, SIZE(meta_env%metavar)
                  IF (ASSOCIATED(meta_env%metavar(i)%walls)) THEN
                     DEALLOCATE (meta_env%metavar(i)%walls)
                  END IF
               END DO
               DEALLOCATE (meta_env%metavar)
            END IF

            ! Hills env
            CALL hills_env_release(meta_env%hills_env)
            ! Walkers type
            IF (ASSOCIATED(meta_env%multiple_walkers)) THEN
               IF (ASSOCIATED(meta_env%multiple_walkers%walkers_status)) THEN
                  DEALLOCATE (meta_env%multiple_walkers%walkers_status)
               END IF
               IF (ASSOCIATED(meta_env%multiple_walkers%walkers_file_name)) THEN
                  DEALLOCATE (meta_env%multiple_walkers%walkers_file_name)
               END IF
               DEALLOCATE (meta_env%multiple_walkers)
            END IF

            ! Langevin on COLVARS
            IF (meta_env%langevin) THEN
               DO i = 1, SIZE(meta_env%rng)
                  IF (ASSOCIATED(meta_env%rng(i)%stream)) THEN
                     CALL delete_rng_stream(meta_env%rng(i)%stream)
                  END IF
               END DO
               DEALLOCATE (meta_env%rng)
            ENDIF
            NULLIFY (meta_env%time)
            NULLIFY (meta_env%metadyn_section)
            DEALLOCATE (meta_env)
         END IF
      END IF
   END SUBROUTINE meta_env_release

! **************************************************************************************************
!> \brief releases the hills_env
!> \param hills_env ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE hills_env_release(hills_env)
      TYPE(hills_env_type), POINTER                      :: hills_env

      CHARACTER(len=*), PARAMETER :: routineN = 'hills_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(hills_env)) THEN
         IF (ASSOCIATED(hills_env%ss_history)) THEN
            DEALLOCATE (hills_env%ss_history)
         END IF
         IF (ASSOCIATED(hills_env%delta_s_history)) THEN
            DEALLOCATE (hills_env%delta_s_history)
         END IF
         IF (ASSOCIATED(hills_env%ww_history)) THEN
            DEALLOCATE (hills_env%ww_history)
         END IF
         IF (ASSOCIATED(hills_env%invdt_history)) THEN
            DEALLOCATE (hills_env%invdt_history)
         END IF
         DEALLOCATE (hills_env)
      END IF
   END SUBROUTINE hills_env_release

END MODULE metadynamics_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types used in the flexible partitioning scheme
!> \par History
!>      04.2006 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE fp_types
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: fp_type
   PUBLIC :: fp_env_create, fp_env_release, fp_env_retain, fp_env_read, fp_env_write

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fp_types'

! **************************************************************************************************
   TYPE fp_type

      ! input related objects
      LOGICAL                        :: use_fp
      INTEGER                        :: ref_count

      INTEGER                        :: central_atom
      INTEGER, DIMENSION(:), POINTER :: inner_atoms, outer_atoms
      REAL(KIND=dp)                  :: inner_radius, outer_radius
      REAL(KIND=dp)                  :: strength, smooth_width
      LOGICAL                        :: bias
      REAL(KIND=dp)                  :: temperature
      TYPE(section_vals_type), POINTER   :: print_section

      ! computed during runs
      INTEGER                        :: i1, i2, o1, o2
      REAL(KIND=dp)                  :: ri1, ri2, ro1, ro2
      REAL(KIND=dp)                  :: weight, comb_weight, bias_weight
      REAL(KIND=dp)                  :: energy, bias_energy, restraint_energy
   END TYPE fp_type

CONTAINS

! **************************************************************************************************
!> \brief create retain release the flexible partitioning environment
!> \param fp_env ...
!> \par History
!>      04.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE fp_env_create(fp_env)
      TYPE(fp_type), POINTER                             :: fp_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(fp_env)))CALL cp__a("fp_types.F",73)
      ALLOCATE (fp_env)
      fp_env%ref_count = 1

      fp_env%use_fp = .FALSE.
      NULLIFY (fp_env%inner_atoms)
      NULLIFY (fp_env%outer_atoms)
      NULLIFY (fp_env%print_section)
      CALL timestop(handle)

   END SUBROUTINE fp_env_create

! **************************************************************************************************
!> \brief ...
!> \param fp_env ...
! **************************************************************************************************
   SUBROUTINE fp_env_release(fp_env)
      TYPE(fp_type), POINTER                             :: fp_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(fp_env)) THEN
         IF(.NOT.(fp_env%ref_count > 0))CALL cp__a("fp_types.F",99)
         fp_env%ref_count = fp_env%ref_count-1
         IF (fp_env%ref_count == 0) THEN
            IF (ASSOCIATED(fp_env%inner_atoms)) DEALLOCATE (fp_env%inner_atoms)
            IF (ASSOCIATED(fp_env%outer_atoms)) DEALLOCATE (fp_env%outer_atoms)
            IF (ASSOCIATED(fp_env%print_section)) CALL section_vals_release(fp_env%print_section)
            fp_env%use_fp = .FALSE.
            DEALLOCATE (fp_env)
         ENDIF
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE fp_env_release

! **************************************************************************************************
!> \brief ...
!> \param fp_env ...
! **************************************************************************************************
   SUBROUTINE fp_env_retain(fp_env)
      TYPE(fp_type), POINTER                             :: fp_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fp_env)))CALL cp__a("fp_types.F",122)
      fp_env%ref_count = fp_env%ref_count+1

   END SUBROUTINE fp_env_retain

! **************************************************************************************************
!> \brief reads the corresponding input section and stores it in the fp_env
!> \param fp_env ...
!> \param fp_section ...
!> \par History
!>      04.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE fp_env_read(fp_env, fp_section)
      TYPE(fp_type), POINTER                             :: fp_env
      TYPE(section_vals_type), POINTER                   :: fp_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_env_read', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(:), POINTER                     :: tmplist

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(fp_env)))CALL cp__a("fp_types.F",144)
      CALL section_vals_get(fp_section, explicit=fp_env%use_fp)
      IF (fp_env%use_fp) THEN
         CALL section_vals_val_get(fp_section, "CENTRAL_ATOM", i_val=fp_env%central_atom)

         CALL section_vals_val_get(fp_section, "INNER_ATOMS", i_vals=tmplist)
         ALLOCATE (fp_env%inner_atoms(SIZE(tmplist, 1)))
         fp_env%inner_atoms = tmplist

         CALL section_vals_val_get(fp_section, "OUTER_ATOMS", i_vals=tmplist)
         ALLOCATE (fp_env%outer_atoms(SIZE(tmplist, 1)))
         fp_env%outer_atoms = tmplist

         CALL section_vals_val_get(fp_section, "INNER_RADIUS", r_val=fp_env%inner_radius)
         CALL section_vals_val_get(fp_section, "OUTER_RADIUS", r_val=fp_env%outer_radius)
         CALL section_vals_val_get(fp_section, "STRENGTH", r_val=fp_env%strength)
         CALL section_vals_val_get(fp_section, "SMOOTH_WIDTH", r_val=fp_env%smooth_width)
         CALL section_vals_val_get(fp_section, "BIAS", l_val=fp_env%bias)
         CALL section_vals_val_get(fp_section, "TEMPERATURE", r_val=fp_env%temperature)

         fp_env%print_section => section_vals_get_subs_vals(fp_section, "WEIGHTS")
         CALL section_vals_retain(fp_env%print_section)
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE fp_env_read

! **************************************************************************************************
!> \brief writes information concerning the fp_env to the output
!> \param fp_env ...
!> \param fp_section ...
!> \par History
!>      04.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE fp_env_write(fp_env, fp_section)
      TYPE(fp_type), POINTER                             :: fp_env
      TYPE(section_vals_type), POINTER                   :: fp_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_env_write', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: PRINT

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(fp_env)))CALL cp__a("fp_types.F",190)

      IF (fp_env%use_fp) THEN
         PRINT => section_vals_get_subs_vals(fp_section, "CONTROL")
         output_unit = cp_print_key_unit_nr(logger, PRINT, "", extension=".Log")
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T79,A)") &
               "FP| Flexible partitioning is ", "ON"
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
               "FP| Central atom ", fp_env%central_atom
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
               "FP| number of inner atoms", SIZE(fp_env%inner_atoms, 1)
            WRITE (UNIT=output_unit, FMT="(1(T2,8I8))") fp_env%inner_atoms
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
               "FP| number of outer atoms", SIZE(fp_env%outer_atoms, 1)
            WRITE (UNIT=output_unit, FMT="(1(T2,8I8))") fp_env%outer_atoms
            WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.10)") &
               "FP| inner radius [a.u.] ", fp_env%inner_radius
            WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.10)") &
               "FP| outer radius [a.u.] ", fp_env%outer_radius
            WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.10)") &
               "FP| reflecting restraint strength ", fp_env%strength
            IF (fp_env%bias) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,T79,A)") &
                  "FP| Flexible partitioning bias is ", "ON"
               WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.10)") &
                  "FP| bias temperature [kT a.u.]", fp_env%temperature
               WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.10)") &
                  "FP| smooth width [a.u.] ", fp_env%smooth_width
            ELSE
               WRITE (UNIT=output_unit, FMT="(T2,A,T78,A)") &
                  "FP| Flexible partitioning bias is", "OFF"
            ENDIF
         ENDIF
         CALL cp_print_key_finished_output(output_unit, logger, PRINT, "")
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE fp_env_write

END MODULE fp_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface for the force calculations
!> \par History
!>      cjm, FEB-20-2001: pass variable box_ref
!>      cjm, SEPT-12-2002: major reorganization
!>      fawzi, APR-12-2003: introduced force_env
!>      cjm, FEB-27-2006: no more box_change
!>      MK, Nov. 2010: new interfaces added and others were updated
!> \author CJM & JGH
! **************************************************************************************************
MODULE force_env_types
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_logger_type,&
                                              cp_rm_default_logger
   USE cp_para_env,                     ONLY: cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type,&
                                              pack_subsys_particles
   USE eip_environment_types,           ONLY: eip_env_get,&
                                              eip_env_release,&
                                              eip_environment_type
   USE fist_energy_types,               ONLY: fist_energy_type
   USE fist_environment_types,          ONLY: fist_env_get,&
                                              fist_env_release,&
                                              fist_environment_type
   USE fp_types,                        ONLY: fp_env_release,&
                                              fp_env_retain,&
                                              fp_type
   USE global_types,                    ONLY: global_environment_type,&
                                              globenv_release
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE metadynamics_types,              ONLY: meta_env_release,&
                                              meta_env_retain,&
                                              meta_env_type
   USE mixed_energy_types,              ONLY: mixed_energy_type
   USE mixed_environment_types,         ONLY: get_mixed_env,&
                                              mixed_env_release,&
                                              mixed_environment_type
   USE qmmm_types,                      ONLY: qmmm_env_get,&
                                              qmmm_env_release,&
                                              qmmm_env_type
   USE qmmmx_types,                     ONLY: qmmmx_env_get,&
                                              qmmmx_env_release,&
                                              qmmmx_env_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_env_release,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_env_types'

   INTEGER, PARAMETER, PUBLIC :: use_fist_force = 501, &
                                 use_qs_force = 502, &
                                 use_qmmm = 503, &
                                 use_qmmmx = 504, &
                                 use_eip_force = 505, &
                                 use_mixed_force = 506

   CHARACTER(LEN=10), DIMENSION(501:506), PARAMETER, PUBLIC :: &
      use_prog_name = (/ &
      "FIST  ", &
      "QS    ", &
      "QMMM  ", &
      "QMMMX ", &
      "EIP   ", &
      "MIXED "/)

   PUBLIC :: force_env_type, &
             force_env_p_type

   PUBLIC :: force_env_retain, &
             force_env_release, &
             force_env_get, &
             force_env_get_natom, &
             force_env_get_nparticle, &
             force_env_get_frc, &
             force_env_get_pos, &
             force_env_get_vel, &
             force_env_set, &
             multiple_fe_list

! **************************************************************************************************
!> \brief wrapper to abstract the force evaluation of the various methods
!> \param id_nr identification number
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param in_use which method is in use
!> \param fist_env the fist environment (allocated only if fist is in use)
!> \param qs_env qs_env (activated only if quickstep is in use)
!> \param globenv the globenv to have the input that generated this force_env
!> \param para_env the parallel environment that contains all the parallel
!>        environment of the fragments
!> \param meta_env the metadynamics environment, allocated if there is
!>        metadynamics
!> \param fp_env the flexible partitioning environment
!>      read-only attributes (get them *only* through force_env_get):
!> \param subsys the fragments that build up the actual system.
!> \param cell the cell of the actual system
!> \note
!>      as always direct manipulation of these attributes can have very
!>      bad effects. In this case it can be quite bad and the variables
!>      might not be up to date. You are warned, use only the get method...
!> \par History
!>      04.2003 created [fawzi]
!>      07.2003 tried to adapt to multiple mpi groups
!> \author fawzi
! **************************************************************************************************
   TYPE force_env_type
      INTEGER :: id_nr, ref_count, in_use, method_name_id
      REAL(KIND=dp)                                    :: additional_potential
      TYPE(fist_environment_type), POINTER             :: fist_env
      TYPE(meta_env_type), POINTER                     :: meta_env
      TYPE(fp_type), POINTER                           :: fp_env
      TYPE(qs_environment_type), POINTER               :: qs_env
      TYPE(eip_environment_type), POINTER              :: eip_env
      TYPE(global_environment_type), POINTER           :: globenv
      TYPE(cp_para_env_type), POINTER                  :: para_env
      TYPE(force_env_p_type), DIMENSION(:), POINTER    :: sub_force_env
      TYPE(qmmm_env_type), POINTER                     :: qmmm_env
      TYPE(qmmmx_env_type), POINTER                    :: qmmmx_env
      TYPE(mixed_environment_type), POINTER            :: mixed_env
      TYPE(section_vals_type), POINTER                  :: force_env_section
      TYPE(section_vals_type), POINTER                  :: root_section
   END TYPE force_env_type

! **************************************************************************************************
!> \brief allows for the creation of an array of force_env
!> \param force_env a force environment (see above)
!> \note
!>      added by MJM for MC swap moves
!> \author MJM
! **************************************************************************************************
   TYPE force_env_p_type
      TYPE(force_env_type), POINTER :: force_env
   END TYPE force_env_p_type

CONTAINS

! **************************************************************************************************
!> \brief retains the given force env
!> \param force_env the force environment to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE force_env_retain(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",170)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",171)
      force_env%ref_count = force_env%ref_count+1
   END SUBROUTINE force_env_retain

! **************************************************************************************************
!> \brief releases the given force env
!> \param force_env the force environment to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   RECURSIVE SUBROUTINE force_env_release(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_group
      TYPE(cp_logger_type), POINTER                      :: my_logger

      IF (ASSOCIATED(force_env)) THEN
         IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",194)
         force_env%ref_count = force_env%ref_count-1
         IF (force_env%ref_count == 0) THEN
            ! Deallocate SUB_FORCE_ENV
            IF (ASSOCIATED(force_env%sub_force_env)) THEN
               DO i = 1, SIZE(force_env%sub_force_env)
                  IF (.NOT. ASSOCIATED(force_env%sub_force_env(i)%force_env)) CYCLE
                  ! Use the proper logger to deallocate..
                  IF (force_env%in_use == use_mixed_force) THEN
                     my_group = force_env%mixed_env%group_distribution(force_env%para_env%mepos)
                     my_logger => force_env%mixed_env%sub_logger(my_group+1)%p
                     CALL cp_add_default_logger(my_logger)
                  END IF
                  CALL force_env_release(force_env%sub_force_env(i)%force_env)
                  IF (force_env%in_use == use_mixed_force) &
                     CALL cp_rm_default_logger()
               END DO
               DEALLOCATE (force_env%sub_force_env)
            END IF

            SELECT CASE (force_env%in_use)
            CASE (use_fist_force)
               CALL fist_env_release(force_env%fist_env)
            CASE (use_qs_force)
               CALL qs_env_release(force_env%qs_env)
            CASE (use_eip_force)
               CALL eip_env_release(force_env%eip_env)
            CASE (use_mixed_force)
               CALL mixed_env_release(force_env%mixed_env)
            END SELECT
            CALL globenv_release(force_env%globenv)
            CALL cp_para_env_release(force_env%para_env)
            ! Not deallocated
            IF(.NOT.(.NOT. ASSOCIATED(force_env%fist_env)))CALL cp__a("force_env_types.F",227)
            IF(.NOT.(.NOT. ASSOCIATED(force_env%qs_env)))CALL cp__a("force_env_types.F",228)
            IF(.NOT.(.NOT. ASSOCIATED(force_env%eip_env)))CALL cp__a("force_env_types.F",229)
            IF(.NOT.(.NOT. ASSOCIATED(force_env%mixed_env)))CALL cp__a("force_env_types.F",230)
            CALL meta_env_release(force_env%meta_env)
            CALL fp_env_release(force_env%fp_env)
            CALL qmmm_env_release(force_env%qmmm_env)
            CALL qmmmx_env_release(force_env%qmmmx_env)
            CALL section_vals_release(force_env%force_env_section)
            CALL section_vals_release(force_env%root_section)
            DEALLOCATE (force_env)
         END IF
      END IF
      NULLIFY (force_env)
   END SUBROUTINE force_env_release

! **************************************************************************************************
!> \brief returns various attributes about the force environment
!> \param force_env the force environment you what informations about
!> \param in_use ...
!> \param fist_env ...
!> \param qs_env ...
!> \param meta_env ...
!> \param fp_env ...
!> \param subsys ...
!> \param para_env ...
!> \param potential_energy ...
!> \param additional_potential ...
!> \param kinetic_energy ...
!> \param harmonic_shell ...
!> \param kinetic_shell ...
!> \param cell ...
!> \param sub_force_env ...
!> \param qmmm_env ...
!> \param qmmmx_env ...
!> \param eip_env ...
!> \param globenv ...
!> \param input ...
!> \param force_env_section ...
!> \param method_name_id ...
!> \param root_section ...
!> \param mixed_env ...
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE force_env_get(force_env, in_use, fist_env, qs_env, &
                                      meta_env, fp_env, subsys, para_env, potential_energy, additional_potential, &
                                      kinetic_energy, harmonic_shell, kinetic_shell, cell, sub_force_env, &
                                      qmmm_env, qmmmx_env, eip_env, globenv, input, force_env_section, &
                                      method_name_id, root_section, mixed_env)
      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER, INTENT(out), OPTIONAL                     :: in_use
      TYPE(fist_environment_type), OPTIONAL, POINTER     :: fist_env
      TYPE(qs_environment_type), OPTIONAL, POINTER       :: qs_env
      TYPE(meta_env_type), OPTIONAL, POINTER             :: meta_env
      TYPE(fp_type), OPTIONAL, POINTER                   :: fp_env
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: potential_energy, additional_potential, &
                                                            kinetic_energy, harmonic_shell, &
                                                            kinetic_shell
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      TYPE(force_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: sub_force_env
      TYPE(qmmm_env_type), OPTIONAL, POINTER             :: qmmm_env
      TYPE(qmmmx_env_type), OPTIONAL, POINTER            :: qmmmx_env
      TYPE(eip_environment_type), OPTIONAL, POINTER      :: eip_env
      TYPE(global_environment_type), OPTIONAL, POINTER   :: globenv
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input, force_env_section
      INTEGER, INTENT(out), OPTIONAL                     :: method_name_id
      TYPE(section_vals_type), OPTIONAL, POINTER         :: root_section
      TYPE(mixed_environment_type), OPTIONAL, POINTER    :: mixed_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_get', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: eip_kinetic_energy, eip_potential_energy
      TYPE(cp_subsys_type), POINTER                      :: subsys_tmp
      TYPE(fist_energy_type), POINTER                    :: thermo
      TYPE(mixed_energy_type), POINTER                   :: mixed_energy
      TYPE(qs_energy_type), POINTER                      :: qs_energy

      NULLIFY (subsys_tmp)

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",311)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",312)

      SELECT CASE (force_env%in_use)
      CASE (use_qs_force)
         IF(.NOT.(ASSOCIATED(force_env%qs_env)))CALL cp__a("force_env_types.F",316)
         IF(.NOT.(.NOT. PRESENT(fist_env)))CALL cp__a("force_env_types.F",317)
         IF(.NOT.(.NOT. PRESENT(eip_env)))CALL cp__a("force_env_types.F",318)
         CALL get_qs_env(force_env%qs_env, &
                         energy=qs_energy, &
                         input=input, &
                         cp_subsys=subsys)
         IF (PRESENT(potential_energy)) potential_energy = qs_energy%total
         IF(.NOT.(.NOT. PRESENT(kinetic_energy)))CALL cp__a("force_env_types.F",324)
      CASE (use_fist_force)
         IF(.NOT.(ASSOCIATED(force_env%fist_env)))CALL cp__a("force_env_types.F",326)
!          CPPreconditionNoFail(.NOT.PRESENT(qs_env),cp_warning_level,routineP)
!          CPPreconditionNoFail(.NOT.PRESENT(eip_env),cp_warning_level,routineP)
         IF(.NOT.(.NOT. PRESENT(input)))CALL cp__a("force_env_types.F",329)
         CALL fist_env_get(force_env%fist_env, &
                           thermo=thermo, &
                           subsys=subsys)
         IF (PRESENT(potential_energy)) potential_energy = thermo%pot
         IF (PRESENT(kinetic_energy)) kinetic_energy = thermo%kin
         IF (PRESENT(kinetic_shell)) kinetic_shell = thermo%kin_shell
         IF (PRESENT(harmonic_shell)) harmonic_shell = thermo%harm_shell
      CASE (use_eip_force)
         IF(.NOT.(ASSOCIATED(force_env%eip_env)))CALL cp__a("force_env_types.F",338)
         IF(.NOT.(.NOT. PRESENT(qs_env)))CALL cp__a("force_env_types.F",339)
         IF(.NOT.(.NOT. PRESENT(fist_env)))CALL cp__a("force_env_types.F",340)
         CALL eip_env_get(force_env%eip_env, &
                          eip_potential_energy=eip_potential_energy, &
                          eip_kinetic_energy=eip_kinetic_energy, &
                          subsys=subsys)
         IF (PRESENT(potential_energy)) THEN
            potential_energy = eip_potential_energy
         END IF
         IF (PRESENT(kinetic_energy)) kinetic_energy = eip_kinetic_energy
         IF(.NOT.(.NOT. PRESENT(kinetic_energy)))CALL cp__a("force_env_types.F",349)
      CASE (use_qmmm)
         CALL qmmm_env_get(force_env%qmmm_env, &
                           subsys=subsys, &
                           potential_energy=potential_energy, &
                           kinetic_energy=kinetic_energy)
      CASE (use_qmmmx)
         CALL qmmmx_env_get(force_env%qmmmx_env, &
                            subsys=subsys, &
                            potential_energy=potential_energy, &
                            kinetic_energy=kinetic_energy)
      CASE (use_mixed_force)
         IF(.NOT.(ASSOCIATED(force_env%mixed_env)))CALL cp__a("force_env_types.F",361)
!          CPPreconditionNoFail(.NOT.PRESENT(qs_env),cp_warning_level,routineP)
!          CPPreconditionNoFail(.NOT.PRESENT(eip_env),cp_warning_level,routineP)
         IF(.NOT.(.NOT. PRESENT(input)))CALL cp__a("force_env_types.F",364)
         CALL get_mixed_env(force_env%mixed_env, &
                            mixed_energy=mixed_energy, &
                            subsys=subsys)
         IF (PRESENT(potential_energy)) potential_energy = mixed_energy%pot
         IF (PRESENT(kinetic_energy)) kinetic_energy = mixed_energy%kin
      CASE DEFAULT
         CALL cp__b("force_env_types.F",371,"unknown in_use flag value ")
      END SELECT

      IF (PRESENT(force_env_section)) force_env_section => force_env%force_env_section
      IF (PRESENT(in_use)) in_use = force_env%in_use
      IF (PRESENT(method_name_id)) method_name_id = force_env%method_name_id
      IF (PRESENT(fist_env)) THEN
         fist_env => force_env%fist_env
      END IF
      IF (PRESENT(qs_env)) THEN
         qs_env => force_env%qs_env
      END IF
      IF (PRESENT(eip_env)) THEN
         eip_env => force_env%eip_env
      END IF
      IF (PRESENT(para_env)) para_env => force_env%para_env
      ! adjust the total energy for the metadynamics
      IF (ASSOCIATED(force_env%meta_env)) THEN
         IF (PRESENT(potential_energy)) THEN
            potential_energy = potential_energy+ &
                               force_env%meta_env%epot_s+ &
                               force_env%meta_env%epot_walls+ &
                               force_env%meta_env%hills_env%energy
         END IF
         IF (PRESENT(kinetic_energy)) THEN
            kinetic_energy = kinetic_energy+force_env%meta_env%ekin_s
         END IF
      END IF
      ! adjust the total energy for the flexible partitioning
      IF (ASSOCIATED(force_env%fp_env) .AND. PRESENT(potential_energy)) THEN
         IF (force_env%fp_env%use_fp) THEN
            potential_energy = potential_energy+force_env%fp_env%energy
         ENDIF
      ENDIF
      IF (PRESENT(potential_energy)) THEN
         potential_energy = potential_energy+force_env%additional_potential
      END IF
      IF (PRESENT(additional_potential)) THEN
         additional_potential = force_env%additional_potential
      END IF
      IF (PRESENT(cell)) THEN
         CALL force_env_get(force_env, subsys=subsys_tmp)
         CALL cp_subsys_get(subsys_tmp, cell=cell)
      END IF
      IF (PRESENT(fp_env)) fp_env => force_env%fp_env
      IF (PRESENT(meta_env)) meta_env => force_env%meta_env
      IF (PRESENT(sub_force_env)) sub_force_env => force_env%sub_force_env
      IF (PRESENT(qmmm_env)) qmmm_env => force_env%qmmm_env
      IF (PRESENT(qmmmx_env)) qmmmx_env => force_env%qmmmx_env
      IF (PRESENT(mixed_env)) mixed_env => force_env%mixed_env
      IF (PRESENT(globenv)) globenv => force_env%globenv
      IF (PRESENT(root_section)) root_section => force_env%root_section

   END SUBROUTINE force_env_get

! **************************************************************************************************
!> \brief returns the number of atoms
!> \param force_env the force_env you what information about
!> \retval n_atom the number of atoms
!> \date   22.11.2010 updated (MK)
!> \author fawzi
! **************************************************************************************************
   FUNCTION force_env_get_natom(force_env) RESULT(n_atom)

      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER                                            :: n_atom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_env_get_natom', &
         routineP = moduleN//':'//routineN

      TYPE(cp_subsys_type), POINTER                      :: subsys

      n_atom = 0
      NULLIFY (subsys)
      CALL force_env_get(force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, natom=n_atom)

   END FUNCTION force_env_get_natom

! **************************************************************************************************
!> \brief returns the number of particles in a force environment
!> \param force_env the force_env you what information about
!> \retval n_particle the number of particles
!> \date   22.11.2010 (MK)
!> \author Matthias Krack
! **************************************************************************************************
   FUNCTION force_env_get_nparticle(force_env) RESULT(n_particle)

      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER                                            :: n_particle

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_env_get_nparticle', &
         routineP = moduleN//':'//routineN

      TYPE(cp_subsys_type), POINTER                      :: subsys

      n_particle = 0
      NULLIFY (subsys)
      CALL force_env_get(force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, nparticle=n_particle)

   END FUNCTION force_env_get_nparticle

! **************************************************************************************************
!> \brief returns the particle forces in a dimension(*) array
!> \param force_env the force_env you want to get the forces
!> \param frc the array of the forces
!> \param n ...
!> \date   22.11.2010 Creation
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE force_env_get_frc(force_env, frc, n)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(*), INTENT(OUT)           :: frc
      INTEGER, INTENT(IN)                                :: n

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_env_get_frc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",495)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",496)
      CALL force_env_get(force_env, subsys=subsys)
      CALL pack_subsys_particles(subsys=subsys, f=frc(1:n))
      CALL timestop(handle)

   END SUBROUTINE force_env_get_frc

! **************************************************************************************************
!> \brief returns the particle positions in a dimension(*) array
!> \param force_env the force_env you want to get the positions
!> \param pos the array of the positions
!> \param n ...
!> \date   22.11.2010 updated (MK)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE force_env_get_pos(force_env, pos, n)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(kind=dp), DIMENSION(*), INTENT(OUT)           :: pos
      INTEGER, INTENT(IN)                                :: n

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_env_get_pos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",524)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",525)
      CALL force_env_get(force_env, subsys=subsys)
      CALL pack_subsys_particles(subsys=subsys, r=pos(1:n))
      CALL timestop(handle)

   END SUBROUTINE force_env_get_pos

! **************************************************************************************************
!> \brief returns the particle velocities in a dimension(*) array
!> \param force_env the force_env you want to get the velocities
!> \param vel the array of the velocities
!> \param n ...
!> \date   22.11.2010 Creation (MK)
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE force_env_get_vel(force_env, vel, n)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(*), INTENT(OUT)           :: vel
      INTEGER, INTENT(IN)                                :: n

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_env_get_vel', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",553)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",554)
      CALL force_env_get(force_env, subsys=subsys)
      CALL pack_subsys_particles(subsys=subsys, v=vel(1:n))
      CALL timestop(handle)

   END SUBROUTINE force_env_get_vel

! **************************************************************************************************
!> \brief changes some attributes of the force_env
!> \param force_env the force environment where the cell should be changed
!> \param meta_env the new meta environment
!> \param fp_env ...
!> \param force_env_section ...
!> \param method_name_id ...
!> \param additional_potential ...
!> \par History
!>      09.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE force_env_set(force_env, meta_env, fp_env, force_env_section, &
                            method_name_id, additional_potential)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(meta_env_type), OPTIONAL, POINTER             :: meta_env
      TYPE(fp_type), OPTIONAL, POINTER                   :: fp_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: force_env_section
      INTEGER, OPTIONAL                                  :: method_name_id
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: additional_potential

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_set', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_types.F",585)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_types.F",586)
      IF (PRESENT(meta_env)) THEN
         IF (ASSOCIATED(meta_env)) THEN
            CALL meta_env_retain(meta_env)
         END IF
         CALL meta_env_release(force_env%meta_env)
         force_env%meta_env => meta_env
      END IF
      IF (PRESENT(fp_env)) THEN
         CALL fp_env_retain(fp_env)
         CALL fp_env_release(force_env%fp_env)
         force_env%fp_env => fp_env
      END IF
      IF (PRESENT(force_env_section)) THEN
         IF (ASSOCIATED(force_env_section)) THEN
            CALL section_vals_retain(force_env_section)
            CALL section_vals_release(force_env%force_env_section)
            force_env%force_env_section => force_env_section
         END IF
      END IF
      IF (PRESENT(additional_potential)) THEN
         force_env%additional_potential = additional_potential
      END IF
      IF (PRESENT(method_name_id)) THEN
         force_env%method_name_id = method_name_id
      END IF

   END SUBROUTINE force_env_set

! **************************************************************************************************
!> \brief returns the order of the multiple force_env
!> \param force_env_sections ...
!> \param root_section ...
!> \param i_force_eval ...
!> \param nforce_eval ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE multiple_fe_list(force_env_sections, root_section, i_force_eval, nforce_eval)

      TYPE(section_vals_type), POINTER                   :: force_env_sections, root_section
      INTEGER, DIMENSION(:), POINTER                     :: i_force_eval
      INTEGER                                            :: nforce_eval

      CHARACTER(len=*), PARAMETER :: routineN = 'multiple_fe_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iforce_eval, main_force_eval
      INTEGER, DIMENSION(:), POINTER                     :: my_i_force_eval

! Let's treat the case of Multiple force_eval

      CALL section_vals_get(force_env_sections, n_repetition=nforce_eval)
      CALL section_vals_val_get(root_section, "MULTIPLE_FORCE_EVALS%FORCE_EVAL_ORDER", &
                                i_vals=my_i_force_eval)
      ALLOCATE (i_force_eval(nforce_eval))
      IF (nforce_eval > 0) THEN
         IF (nforce_eval == SIZE(my_i_force_eval)) THEN
            i_force_eval = my_i_force_eval
         ELSE
            ! The difference in the amount of defined force_env MUST be one..
            IF(.NOT.(nforce_eval-SIZE(my_i_force_eval) == 1))CALL cp__a("force_env_types.F",646)
            DO iforce_eval = 1, nforce_eval
               IF (ANY(my_i_force_eval == iforce_eval)) CYCLE
               main_force_eval = iforce_eval
               EXIT
            END DO
            i_force_eval(1) = main_force_eval
            i_force_eval(2:nforce_eval) = my_i_force_eval
         END IF
      END IF

   END SUBROUTINE multiple_fe_list

END MODULE force_env_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for evaluating the residual part (1/r^3) of Integrals for
!>        semi-empiric methods
!> \author Teodoro Laino (11.2008) [tlaino]
! **************************************************************************************************
MODULE semi_empirical_int3_utils

   USE input_constants,                 ONLY: do_method_pchg
   USE kinds,                           ONLY: dp
   USE semi_empirical_int_arrays,       ONLY: clm_d,&
                                              indexb
   USE semi_empirical_types,            ONLY: semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int3_utils'

   PUBLIC ::   ijkl_low_3, charg_int_3, dcharg_int_3, coeff_int_3

CONTAINS

! **************************************************************************************************
!> \brief Low level general driver for computing residual part of semi-empirical
!>        integrals <ij|kl> and their derivatives
!>        The residual part is the leading 1/r^3 term
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param itype ...
!> \param eval ...
!> \retval res ...
!> \date 11.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION ijkl_low_3(sepi, sepj, ij, kl, li, lj, lk, ll, ic, itype, eval) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic, itype
      REAL(KIND=dp)                                      :: eval, res

      CHARACTER(len=*), PARAMETER :: routineN = 'ijkl_low_3', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2, lij, lkl
      REAL(KIND=dp)                                      :: add, ccc, chrg, pij, pkl, sum

      sum = 0.0_dp
      l1 = ABS(li-lj)
      lij = indexb(li+1, lj+1)
      l2 = ABS(lk-ll)
      lkl = indexb(lk+1, ll+1)

      ! Standard value of the integral
      IF (l1 == 0) THEN
         IF (lij == 1) THEN
            pij = sepi%ko(1)
            IF (ic == 1) THEN
               pij = sepi%ko(9)
            END IF
         ELSE IF (lij == 3) THEN
            pij = sepi%ko(7)
         ELSE IF (lij == 6) THEN
            pij = sepi%ko(8)
         END IF
      END IF
      !
      IF (l2 == 0) THEN
         IF (lkl == 1) THEN
            pkl = sepj%ko(1)
            IF (ic == 2) THEN
               pkl = sepj%ko(9)
            END IF
         ELSE IF (lkl == 3) THEN
            pkl = sepj%ko(7)
         ELSE IF (lkl == 6) THEN
            pkl = sepj%ko(8)
         END IF
      END IF
      IF (l1 == 0 .AND. l2 == 0) THEN
         IF (itype == do_method_pchg) THEN
            add = 0.0_dp
         ELSE
            add = (pij+pkl)**2
         END IF
         ccc = clm_d(ij, l1, 0)*clm_d(kl, l2, 0)
         IF (ABS(ccc) > EPSILON(0.0_dp)) THEN
            chrg = eval(l1, l2, add)
            sum = chrg
         END IF
      END IF
      res = sum
   END FUNCTION ijkl_low_3

! **************************************************************************************************
!> \brief Evaluates the residual Interaction function between two point-charges
!>        The term evaluated is the 1/r^3 (for short range interactions)
!>        r    -  Distance r12
!>        l1   -  Quantum numbers for multipole of configuration 1
!>        l2   -  Quantum numbers for multipole of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1 ...
!> \param l2 ...
!> \param add ...
!> \retval charg ...
!> \date 11.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION charg_int_3(r, l1, l2, add) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1, l2
      REAL(KIND=dp), INTENT(in)                          :: add
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'charg_int_3', routineP = moduleN//':'//routineN

! Computing only residual Integral Values

      charg = 0.0_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = -add/(2.0_dp*r**3)
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int3_utils.F",139,"")
   END FUNCTION charg_int_3

! **************************************************************************************************
!> \brief Evaluates the coefficient for the residual Interaction function
!>        between two point-charges
!>        l1   -  Quantum numbers for multipole of configuration 1
!>        l2   -  Quantum numbers for multipole of configuration 2
!>        add  -  additive term
!>
!> \param l1 ...
!> \param l2 ...
!> \param add ...
!> \retval coeff ...
!> \date 11.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION coeff_int_3(l1, l2, add) RESULT(coeff)
      INTEGER, INTENT(in)                                :: l1, l2
      REAL(KIND=dp), INTENT(in)                          :: add
      REAL(KIND=dp)                                      :: coeff

      CHARACTER(len=*), PARAMETER :: routineN = 'coeff_int_3', routineP = moduleN//':'//routineN

! Computing only residual Integral Values

      coeff = 0.0_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         coeff = -add/2.0_dp
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int3_utils.F",172,"")
   END FUNCTION coeff_int_3

! **************************************************************************************************
!> \brief Derivatives of residual interaction function between two point-charges
!>
!>        r    -  Distance r12
!>        l1   -  Quantum numbers for multipole of configuration 1
!>        l2   -  Quantum numbers for multipole of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1 ...
!> \param l2 ...
!> \param add ...
!> \retval charg ...
!> \date 11.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION dcharg_int_3(r, l1, l2, add) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1, l2
      REAL(KIND=dp), INTENT(in)                          :: add
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'dcharg_int_3', routineP = moduleN//':'//routineN

! Computing only residual Integral Derivatives

      charg = 0.0_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = 3.0_dp*add/(2.0_dp*r**4)
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int3_utils.F",208,"")
   END FUNCTION dcharg_int_3

END MODULE semi_empirical_int3_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods for handling the 1/R^3 residual integral part
!> \author Teodoro Laino (12.2008) [tlaino]
! **************************************************************************************************
MODULE semi_empirical_expns3_methods
   USE cp_control_types,                ONLY: semi_empirical_control_type
   USE input_constants,                 ONLY: do_method_undef
   USE kinds,                           ONLY: dp
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE semi_empirical_expns3_types,     ONLY: semi_empirical_expns3_create
   USE semi_empirical_int3_utils,       ONLY: coeff_int_3,&
                                              ijkl_low_3
   USE semi_empirical_int_arrays,       ONLY: indexa,&
                                              l_index
   USE semi_empirical_types,            ONLY: semi_empirical_type
   USE semi_empirical_utils,            ONLY: get_se_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_expns3_methods'

   PUBLIC :: semi_empirical_expns3_setup

CONTAINS
! **************************************************************************************************
!> \brief Setup the quantity necessary to handle the slowly convergent
!>        residual integral term 1/R^3
!>
!> \param qs_kind_set ...
!> \param se_control ...
!> \param method_id ...
!> \date 12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE semi_empirical_expns3_setup(qs_kind_set, se_control, method_id)
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      INTEGER, INTENT(IN)                                :: method_id

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_expns3_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, itype, j, nkinds
      LOGICAL                                            :: check
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj

      IF (se_control%do_ewald_r3) THEN
         NULLIFY (sepi, sepj)
         nkinds = SIZE(qs_kind_set)
         DO i = 1, nkinds
            CALL get_qs_kind(qs_kind_set(i), se_parameter=sepi)
            check = .NOT. ASSOCIATED(sepi%expns3_int)
            IF(.NOT.(check))CALL cp__a("semi_empirical_expns3_methods.F",61)
            ALLOCATE (sepi%expns3_int(nkinds))
            DO j = 1, nkinds
               NULLIFY (sepi%expns3_int(j)%expns3)
               CALL semi_empirical_expns3_create(sepi%expns3_int(j)%expns3)
            END DO
         END DO

         itype = get_se_type(method_id)
         DO i = 1, nkinds
            CALL get_qs_kind(qs_kind_set(i), se_parameter=sepi)
            DO j = 1, nkinds
               CALL get_qs_kind(qs_kind_set(j), se_parameter=sepj)
               CALL setup_c3_coeff(sepi, sepj, i, j, itype)
            END DO
         END DO
      END IF
   END SUBROUTINE semi_empirical_expns3_setup

! **************************************************************************************************
!> \brief For any given semi-empirical pair i,j evaluates the coefficient of
!>        the integral residual part ( 1/r^3 term )
!>        The integral expression, unfortunately, does not allow any kind of
!>        separability. It is, therefore, mandatory to compute this coefficient
!>        as a pair term, instead of as an atomic quantity.
!>
!> \param sepi ...
!> \param sepj ...
!> \param ikind ...
!> \param jkind ...
!> \param itype ...
!> \date 12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE setup_c3_coeff(sepi, sepj, ikind, jkind, itype)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ikind, jkind, itype

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_c3_coeff', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, kl, kr, li, lk
      REAL(KIND=dp)                                      :: core_core, e1b(9), e2a(9), r, zi, zj

! Set the distance to 0 (the coefficient is anyway independent of the atomic
! position)

      r = 0.0_dp
      ! Nuclei-Nuclei     contribution
      ij = indexa(1, 1)
      zi = -sepi%zeff
      zj = -sepj%zeff
      core_core = ijkl_low_3(sepi, sepj, ij, ij, 0, 0, 0, 0, -1, itype, coeff_int_3)*zi*zj

      ! Electron(i)-Nuclei(j)   contribution
      kl = indexa(1, 1)
      e1b(1) = ijkl_low_3(sepi, sepj, kl, ij, 0, 0, 0, 0, 2, itype, coeff_int_3)*zj
      IF (sepi%natorb > 1) THEN
         kl = indexa(2, 2)
         e1b(2) = ijkl_low_3(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, itype, coeff_int_3)*zj
         kl = indexa(3, 3)
         e1b(3) = ijkl_low_3(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, itype, coeff_int_3)*zj
         kl = indexa(4, 4)
         e1b(4) = ijkl_low_3(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, itype, coeff_int_3)*zj
         ! Consistency check
         IF(.NOT.(e1b(2) == e1b(3)))CALL cp__a("semi_empirical_expns3_methods.F",125)
         IF(.NOT.(e1b(3) == e1b(4)))CALL cp__a("semi_empirical_expns3_methods.F",126)
         IF (sepi%dorb) THEN
            kl = indexa(5, 5)
            e1b(5) = ijkl_low_3(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, itype, coeff_int_3)*zj
            kl = indexa(6, 6)
            e1b(6) = ijkl_low_3(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, itype, coeff_int_3)*zj
            kl = indexa(7, 7)
            e1b(7) = ijkl_low_3(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, itype, coeff_int_3)*zj
            kl = indexa(8, 8)
            e1b(8) = ijkl_low_3(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, itype, coeff_int_3)*zj
            kl = indexa(9, 9)
            e1b(9) = ijkl_low_3(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, itype, coeff_int_3)*zj
            ! Consistency check
            IF(.NOT.(e1b(5) == e1b(6)))CALL cp__a("semi_empirical_expns3_methods.F",139)
            IF(.NOT.(e1b(6) == e1b(7)))CALL cp__a("semi_empirical_expns3_methods.F",140)
            IF(.NOT.(e1b(7) == e1b(8)))CALL cp__a("semi_empirical_expns3_methods.F",141)
            IF(.NOT.(e1b(8) == e1b(9)))CALL cp__a("semi_empirical_expns3_methods.F",142)
         END IF
      END IF

      ! Electron(j)-Nuclei(i)   contribution
      kl = indexa(1, 1)
      e2a(1) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 0, 0, 1, itype, coeff_int_3)*zi
      IF (sepj%natorb > 1) THEN
         kl = indexa(2, 2)
         e2a(2) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, itype, coeff_int_3)*zi
         kl = indexa(3, 3)
         e2a(3) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, itype, coeff_int_3)*zi
         kl = indexa(4, 4)
         e2a(4) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, itype, coeff_int_3)*zi
         ! Consistency check
         IF(.NOT.(e2a(2) == e2a(3)))CALL cp__a("semi_empirical_expns3_methods.F",157)
         IF(.NOT.(e2a(3) == e2a(4)))CALL cp__a("semi_empirical_expns3_methods.F",158)
         IF (sepj%dorb) THEN
            kl = indexa(5, 5)
            e2a(5) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, itype, coeff_int_3)*zi
            kl = indexa(6, 6)
            e2a(6) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, itype, coeff_int_3)*zi
            kl = indexa(7, 7)
            e2a(7) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, itype, coeff_int_3)*zi
            kl = indexa(8, 8)
            e2a(8) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, itype, coeff_int_3)*zi
            kl = indexa(9, 9)
            e2a(9) = ijkl_low_3(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, itype, coeff_int_3)*zi
            ! Consistency check
            IF(.NOT.(e2a(5) == e2a(6)))CALL cp__a("semi_empirical_expns3_methods.F",171)
            IF(.NOT.(e2a(6) == e2a(7)))CALL cp__a("semi_empirical_expns3_methods.F",172)
            IF(.NOT.(e2a(7) == e2a(8)))CALL cp__a("semi_empirical_expns3_methods.F",173)
            IF(.NOT.(e2a(8) == e2a(9)))CALL cp__a("semi_empirical_expns3_methods.F",174)
         END IF
      END IF

      ! Copy info into the semi-empirical type (i)
      sepi%expns3_int(jkind)%expns3%core_core = core_core
      sepi%expns3_int(jkind)%expns3%e1b(1:sepi%natorb) = e1b(1:sepi%natorb)
      sepi%expns3_int(jkind)%expns3%e2a(1:sepj%natorb) = e2a(1:sepj%natorb)
      ! Copy info into the semi-empirical type (j)
      sepj%expns3_int(ikind)%expns3%core_core = core_core
      sepj%expns3_int(ikind)%expns3%e1b(1:sepj%natorb) = e2a(1:sepj%natorb)
      sepj%expns3_int(ikind)%expns3%e2a(1:sepi%natorb) = e1b(1:sepi%natorb)

      ! Electron-Electron contribution - sepi/sepj
      kr = 0
      DO i = 1, sepi%natorb
         li = l_index(i)
         ij = indexa(i, i)
         DO j = 1, sepj%natorb
            lk = l_index(j)
            kl = indexa(j, j)
            kr = kr+1
            sepi%expns3_int(jkind)%expns3%w(kr) = &
               ijkl_low_3(sepi, sepj, ij, kl, li, li, lk, lk, 0, do_method_undef, coeff_int_3)
         END DO
      END DO

      ! Electron-Electron contribution - sepj/sepi
      kr = 0
      DO i = 1, sepj%natorb
         li = l_index(i)
         ij = indexa(i, i)
         DO j = 1, sepi%natorb
            lk = l_index(j)
            kl = indexa(j, j)
            kr = kr+1
            sepj%expns3_int(ikind)%expns3%w(kr) = &
               ijkl_low_3(sepj, sepi, ij, kl, li, li, lk, lk, 0, do_method_undef, coeff_int_3)
         END DO
      END DO

   END SUBROUTINE setup_c3_coeff

END MODULE semi_empirical_expns3_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author teo
! **************************************************************************************************
MODULE qmmm_ff_fist

   USE kinds,                           ONLY: default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_ff_fist'

   PUBLIC :: qmmm_ff_precond_only_qm

CONTAINS

! **************************************************************************************************
!> \brief This function handles the atom names and modifies
!>      the "_QM_" prefix, in order to find the parameters
!>      and return .TRUE. if all input labels belong to QM atoms
!> \param id1 ...
!> \param id2 ...
!> \param id3 ...
!> \param id4 ...
!> \param is_link ...
!> \retval only_qm ...
!> \par History
!>      11.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   FUNCTION qmmm_ff_precond_only_qm(id1, id2, id3, id4, is_link) RESULT(only_qm)
      CHARACTER(LEN=default_string_length), &
         INTENT(INOUT)                                   :: id1
      CHARACTER(LEN=default_string_length), &
         INTENT(INOUT), OPTIONAL                         :: id2, id3, id4
      LOGICAL, INTENT(OUT), OPTIONAL                     :: is_link
      LOGICAL                                            :: only_qm

      CHARACTER(LEN=default_string_length)               :: tmp
      INTEGER                                            :: j, my_index
      LOGICAL                                            :: my_link

      only_qm = .FALSE.
      my_link = .FALSE.
      DO WHILE (INDEX(id1, "_QM_") /= 0)
         my_link = qmmm_ff_precond_only_link(id1) .OR. my_link
         my_index = INDEX(id1, "_QM_")+LEN_TRIM("_QM_")
         only_qm = .TRUE.
         tmp = TRIM(id1(my_index:))
         clean_string_1: DO j = 1, default_string_length
            id1(j:j) = " "
         END DO clean_string_1
         id1 = TRIM(tmp)
      END DO

      IF (PRESENT(id2)) THEN
         IF (INDEX(id2, "_QM_") == 0) only_qm = .FALSE.
         DO WHILE (INDEX(id2, "_QM_") /= 0)
            my_link = qmmm_ff_precond_only_link(id2) .OR. my_link
            my_index = INDEX(id2, "_QM_")+LEN_TRIM("_QM_")
            tmp = TRIM(id2(my_index:))
            clean_string_2: DO j = 1, default_string_length
               id2(j:j) = " "
            END DO clean_string_2
            id2 = TRIM(tmp)
         END DO
      END IF

      IF (PRESENT(id3)) THEN
         IF (INDEX(id3, "_QM_") == 0) only_qm = .FALSE.
         DO WHILE (INDEX(id3, "_QM_") /= 0)
            my_link = qmmm_ff_precond_only_link(id3) .OR. my_link
            my_index = INDEX(id3, "_QM_")+LEN_TRIM("_QM_")
            tmp = TRIM(id3(my_index:))
            clean_string_3: DO j = 1, default_string_length
               id3(j:j) = " "
            END DO clean_string_3
            id3 = TRIM(tmp)
         END DO
      END IF

      IF (PRESENT(id4)) THEN
         IF (INDEX(id4, "_QM_") == 0) only_qm = .FALSE.
         DO WHILE (INDEX(id4, "_QM_") /= 0)
            my_link = qmmm_ff_precond_only_link(id4) .OR. my_link
            my_index = INDEX(id4, "_QM_")+LEN_TRIM("_QM_")
            tmp = TRIM(id4(my_index:))
            clean_string_4: DO j = 1, default_string_length
               id4(j:j) = " "
            END DO clean_string_4
            id4 = TRIM(tmp)
         END DO
      END IF

      IF (PRESENT(is_link)) is_link = my_link

   END FUNCTION qmmm_ff_precond_only_qm

! **************************************************************************************************
!> \brief ...
!> \param id1 ...
!> \retval is_link ...
! **************************************************************************************************
   FUNCTION qmmm_ff_precond_only_link(id1) RESULT(is_link)
      CHARACTER(LEN=default_string_length), &
         INTENT(INOUT)                                   :: id1
      LOGICAL                                            :: is_link

      CHARACTER(LEN=default_string_length)               :: tmp
      INTEGER                                            :: j, my_index

      is_link = .FALSE.
      DO WHILE (INDEX(id1, "_LNK") /= 0)
         my_index = INDEX(id1, "_LNK")+1
         my_index = INDEX(id1(my_index:), "_QM_")+my_index-1
         is_link = .TRUE.
         tmp = TRIM(id1(my_index:))
         clean_string_1: DO j = 1, default_string_length
            id1(j:j) = " "
         END DO clean_string_1
         id1 = TRIM(tmp)
      END DO

   END FUNCTION qmmm_ff_precond_only_link

END MODULE qmmm_ff_fist
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define methods related to particle_type
!> \par History
!>            10.2014 Move routines out of particle_types.F [Ole Schuett]
!> \author Ole Schuett
! **************************************************************************************************
MODULE particle_methods
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              get_cell,&
                                              pbc,&
                                              real_to_scaled,&
                                              set_cell_param
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE external_potential_types,        ONLY: fist_potential_type,&
                                              get_potential
   USE input_constants,                 ONLY: dump_atomic,&
                                              dump_dcd,&
                                              dump_dcd_aligned_cell,&
                                              dump_pdb,&
                                              dump_xmol
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              dp_size,&
                                              sp,&
                                              sp_size
   USE mathconstants,                   ONLY: degree
   USE mathlib,                         ONLY: angle,&
                                              dihedral_angle
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: get_particle_pos_or_vel,&
                                              particle_type
   USE physcon,                         ONLY: massunit
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE shell_potential_types,           ONLY: get_shell,&
                                              shell_kind_type
   USE string_utilities,                ONLY: uppercase
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: sort,&
                                              sort_unique
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Public subroutines

   PUBLIC :: write_fist_particle_coordinates, &
             write_qs_particle_coordinates, &
             write_particle_distances, &
             write_particle_coordinates, &
             write_structure_data, &
             get_particle_set, &
             write_particle_matrix

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'particle_methods'

CONTAINS

! **************************************************************************************************
!> \brief   Get the components of a particle set.
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param first_sgf ...
!> \param last_sgf ...
!> \param nsgf ...
!> \param nmao ...
!> \param basis ...
!> \date    14.01.2002
!> \par History
!>      - particle type cleaned (13.10.2003,MK)
!>      - refactoring and add basis set option (17.08.2010,jhu)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_particle_set(particle_set, qs_kind_set, first_sgf, last_sgf, nsgf, &
                               nmao, basis)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, DIMENSION(:), INTENT(INOUT), OPTIONAL     :: first_sgf, last_sgf, nsgf, nmao
      TYPE(gto_basis_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'get_particle_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, iparticle, isgf, nparticle, ns

      IF(.NOT.(ASSOCIATED(particle_set)))CALL cp__a("particle_methods.F",110)

      nparticle = SIZE(particle_set)
      IF (PRESENT(first_sgf)) THEN
         IF(.NOT.(SIZE(first_sgf) >= nparticle))CALL cp__a("particle_methods.F",114)
      END IF
      IF (PRESENT(last_sgf)) THEN
         IF(.NOT.(SIZE(last_sgf) >= nparticle))CALL cp__a("particle_methods.F",117)
      END IF
      IF (PRESENT(nsgf)) THEN
         IF(.NOT.(SIZE(nsgf) >= nparticle))CALL cp__a("particle_methods.F",120)
      END IF
      IF (PRESENT(nmao)) THEN
         IF(.NOT.(SIZE(nmao) >= nparticle))CALL cp__a("particle_methods.F",123)
      END IF

      IF (PRESENT(first_sgf) .OR. PRESENT(last_sgf) .OR. PRESENT(nsgf)) THEN
         isgf = 0
         DO iparticle = 1, nparticle
            CALL get_atomic_kind(particle_set(iparticle)%atomic_kind, kind_number=ikind)
            IF (PRESENT(basis)) THEN
               CALL get_gto_basis_set(gto_basis_set=basis(ikind)%gto_basis_set, nsgf=ns)
            ELSE
               CALL get_qs_kind(qs_kind_set(ikind), nsgf=ns)
            END IF
            IF (PRESENT(nsgf)) nsgf(iparticle) = ns
            IF (PRESENT(first_sgf)) first_sgf(iparticle) = isgf+1
            isgf = isgf+ns
            IF (PRESENT(last_sgf)) last_sgf(iparticle) = isgf
         END DO
      END IF

      IF (PRESENT(first_sgf)) THEN
         IF (SIZE(first_sgf) > nparticle) first_sgf(nparticle+1) = isgf+1
      END IF

      IF (PRESENT(nmao)) THEN
         DO iparticle = 1, nparticle
            CALL get_atomic_kind(particle_set(iparticle)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), mao=ns)
            nmao(iparticle) = ns
         END DO
      END IF

   END SUBROUTINE get_particle_set

! **************************************************************************************************
!> \brief   Should be able to write a few formats e.g. xmol, and some binary
!>          format (dcd) some format can be used for x,v,f
!>
!>          FORMAT   CONTENT                                    UNITS x, v, f
!>          XMOL     POS, VEL, FORCE, POS_VEL, POS_VEL_FORCE    Angstrom, a.u., a.u.
!>
!> \param particle_set ...
!> \param iunit ...
!> \param output_format ...
!> \param content ...
!> \param title ...
!> \param cell ...
!> \param array ...
!> \param unit_conv ...
!> \param charge_occup ...
!> \param charge_beta ...
!> \param charge_extended ...
!> \date    14.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_particle_coordinates(particle_set, iunit, output_format, &
                                         content, title, cell, array, unit_conv, &
                                         charge_occup, charge_beta, &
                                         charge_extended)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER                                            :: iunit, output_format
      CHARACTER(LEN=*)                                   :: content, title
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: array
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: unit_conv
      LOGICAL, INTENT(IN), OPTIONAL                      :: charge_occup, charge_beta, &
                                                            charge_extended

      CHARACTER(len=*), PARAMETER :: routineN = 'write_particle_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=120)                                 :: line
      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=4)                                   :: name
      CHARACTER(LEN=default_string_length)               :: atm_name, my_format
      INTEGER                                            :: handle, iatom, natom, stat
      LOGICAL                                            :: dummy, my_charge_beta, &
                                                            my_charge_extended, my_charge_occup
      REAL(KIND=dp)                                      :: angle_alpha, angle_beta, angle_gamma, &
                                                            factor, qeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: arr
      REAL(KIND=dp), DIMENSION(3)                        :: abc, angles, f, r, v
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h
      REAL(KIND=sp), ALLOCATABLE, DIMENSION(:)           :: x4, y4, z4
      TYPE(cell_type), POINTER                           :: cell_dcd
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(shell_kind_type), POINTER                     :: shell

      CALL timeset(routineN, handle)

      natom = SIZE(particle_set)
      IF (PRESENT(array)) THEN
         SELECT CASE (TRIM (content))
         CASE ("POS_VEL", "POS_VEL_FORCE")
            CALL cp__b("particle_methods.F",218,"Illegal usage")
         END SELECT
      END IF
      factor = 1.0_dp
      IF (PRESENT(unit_conv)) THEN
         factor = unit_conv
      END IF
      SELECT CASE (output_format)
      CASE (dump_xmol)
         WRITE (iunit, "(I8)") natom
         WRITE (iunit, "(A)") TRIM(title)
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol)
            IF (LEN_TRIM(element_symbol) == 0) THEN
               CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                    name=atm_name)
               dummy = qmmm_ff_precond_only_qm(id1=atm_name)
               my_format = "(A4,"
               name = TRIM(atm_name)
            ELSE
               my_format = "(T2,A2,"
               name = TRIM(element_symbol)
            END IF
            SELECT CASE (TRIM (content))
            CASE ("POS")
               IF (PRESENT(array)) THEN
                  r(1:3) = get_particle_pos_or_vel(iatom, particle_set, array)
               ELSE
                  r(:) = particle_set(iatom)%r(:)
               END IF
               WRITE (iunit, TRIM(my_format)//"1X,3F20.10)") TRIM(name), r(1:3)*factor
            CASE ("VEL")
               IF (PRESENT(array)) THEN
                  v(1:3) = get_particle_pos_or_vel(iatom, particle_set, array)
               ELSE
                  v(:) = particle_set(iatom)%v(:)
               END IF
               WRITE (iunit, TRIM(my_format)//"1X,3F20.10)") TRIM(name), v(1:3)*factor
            CASE ("FORCE")
               IF (PRESENT(array)) THEN
                  f(:) = array((iatom-1)*3+1:(iatom-1)*3+3)
               ELSE
                  f(:) = particle_set(iatom)%f(:)
               END IF
               WRITE (iunit, TRIM(my_format)//"1X,3F20.10)") TRIM(name), f(1:3)*factor
            CASE ("FORCE_MIXING_LABELS")
               IF (PRESENT(array)) THEN
                  f(:) = array((iatom-1)*3+1:(iatom-1)*3+3)
               ELSE
                  f(:) = particle_set(iatom)%f(:)
               END IF
               WRITE (iunit, TRIM(my_format)//"1X,3F20.10)") TRIM(name), f(1:3)*factor
            END SELECT
         END DO
      CASE (dump_atomic)
         DO iatom = 1, natom
            SELECT CASE (TRIM (content))
            CASE ("POS")
               IF (PRESENT(array)) THEN
                  r(1:3) = get_particle_pos_or_vel(iatom, particle_set, array)
               ELSE
                  r(:) = particle_set(iatom)%r(:)
               END IF
               WRITE (iunit, "(3F20.10)") r(1:3)*factor
            CASE ("VEL")
               IF (PRESENT(array)) THEN
                  v(1:3) = get_particle_pos_or_vel(iatom, particle_set, array)
               ELSE
                  v(:) = particle_set(iatom)%v(:)
               END IF
               WRITE (iunit, "(3F20.10)") v(1:3)*factor
            CASE ("FORCE")
               IF (PRESENT(array)) THEN
                  f(:) = array((iatom-1)*3+1:(iatom-1)*3+3)
               ELSE
                  f(:) = particle_set(iatom)%f(:)
               END IF
               WRITE (iunit, "(3F20.10)") f(1:3)*factor
            CASE ("FORCE_MIXING_LABELS")
               IF (PRESENT(array)) THEN
                  f(:) = array((iatom-1)*3+1:(iatom-1)*3+3)
               ELSE
                  f(:) = particle_set(iatom)%f(:)
               END IF
               WRITE (iunit, "(3F20.10)") f(1:3)*factor
            END SELECT
         END DO
      CASE (dump_dcd, dump_dcd_aligned_cell)
         IF (.NOT. (PRESENT(cell))) THEN
            CALL cp__b("particle_methods.F",308,"Cell is not present! Report this bug!")
         END IF
         CALL get_cell(cell, alpha=angle_alpha, beta=angle_beta, gamma=angle_gamma, &
                       abc=abc)
         IF (.NOT. cell%orthorhombic .AND. (output_format == dump_dcd_aligned_cell)) THEN
            ! In the case of a non-orthorhombic cell adopt a common convention
            ! for the orientation of the cell with respect to the Cartesian axes:
            ! Cell vector a is aligned with the x axis and the cell vector b lies
            ! in the xy plane.
            NULLIFY (cell_dcd)
            CALL cell_create(cell_dcd)
            CALL cell_clone(cell, cell_dcd)
            angles(1) = angle_alpha/degree
            angles(2) = angle_beta/degree
            angles(3) = angle_gamma/degree
            CALL set_cell_param(cell_dcd, abc, angles, &
                                do_init_cell=.TRUE.)
            h(1:3, 1:3) = MATMUL(cell_dcd%hmat(1:3, 1:3), cell%h_inv(1:3, 1:3))
            CALL cell_release(cell_dcd)
         END IF
         ALLOCATE (arr(3, natom), STAT=stat)
         IF (stat /= 0) CALL stop_memory(routineN, moduleN, 329, &
                                         "arr", 3*natom*dp_size)
         IF (PRESENT(array)) THEN
            arr(1:3, 1:natom) = RESHAPE(array, (/3, natom/))
         ELSE
            SELECT CASE (TRIM (content))
            CASE ("POS")
               DO iatom = 1, natom
                  arr(1:3, iatom) = particle_set(iatom)%r(1:3)
               END DO
            CASE ("VEL")
               DO iatom = 1, natom
                  arr(1:3, iatom) = particle_set(iatom)%v(1:3)
               END DO
            CASE ("FORCE")
               DO iatom = 1, natom
                  arr(1:3, iatom) = particle_set(iatom)%f(1:3)
               END DO
            CASE DEFAULT
               CALL cp__b("particle_methods.F",348,"Illegal DCD dump type")
            END SELECT
         END IF
         ALLOCATE (x4(natom), STAT=stat)
         IF (stat /= 0) CALL stop_memory(routineN, moduleN, 352, &
                                         "x4", natom*sp_size)
         ALLOCATE (y4(natom), STAT=stat)
         IF (stat /= 0) CALL stop_memory(routineN, moduleN, 355, &
                                         "y4", natom*sp_size)
         ALLOCATE (z4(natom), STAT=stat)
         IF (stat /= 0) CALL stop_memory(routineN, moduleN, 358, &
                                         "z4", natom*sp_size)
         IF (.NOT. cell%orthorhombic .AND. (output_format == dump_dcd_aligned_cell)) THEN
            x4(1:natom) = REAL(MATMUL(h(1, 1:3), arr(1:3, 1:natom)), KIND=sp)
            y4(1:natom) = REAL(MATMUL(h(2, 1:3), arr(1:3, 1:natom)), KIND=sp)
            z4(1:natom) = REAL(MATMUL(h(3, 1:3), arr(1:3, 1:natom)), KIND=sp)
         ELSE
            x4(1:natom) = REAL(arr(1, 1:natom), KIND=sp)
            y4(1:natom) = REAL(arr(2, 1:natom), KIND=sp)
            z4(1:natom) = REAL(arr(3, 1:natom), KIND=sp)
         END IF
         WRITE (iunit) abc(1)*factor, angle_gamma, abc(2)*factor, &
            angle_beta, angle_alpha, abc(3)*factor
         WRITE (iunit) x4*REAL(factor, KIND=sp)
         WRITE (iunit) y4*REAL(factor, KIND=sp)
         WRITE (iunit) z4*REAL(factor, KIND=sp)
         ! Release work storage
         DEALLOCATE (arr)
         DEALLOCATE (x4)
         DEALLOCATE (y4)
         DEALLOCATE (z4)
      CASE (dump_pdb)
         my_charge_occup = .FALSE.
         IF (PRESENT(charge_occup)) my_charge_occup = charge_occup
         my_charge_beta = .FALSE.
         IF (PRESENT(charge_beta)) my_charge_beta = charge_beta
         my_charge_extended = .FALSE.
         IF (PRESENT(charge_extended)) my_charge_extended = charge_extended
         IF (LEN_TRIM(title) > 0) THEN
            WRITE (UNIT=iunit, FMT="(A6,T11,A)") &
               "REMARK", TRIM(title)
         END IF
         CALL get_cell(cell, alpha=angle_alpha, beta=angle_beta, gamma=angle_gamma, abc=abc)
         ! COLUMNS       DATA TYPE      CONTENTS
         ! --------------------------------------------------
         !  1 -  6       Record name    "CRYST1"
         !  7 - 15       Real(9.3)      a (Angstroms)
         ! 16 - 24       Real(9.3)      b (Angstroms)
         ! 25 - 33       Real(9.3)      c (Angstroms)
         ! 34 - 40       Real(7.2)      alpha (degrees)
         ! 41 - 47       Real(7.2)      beta (degrees)
         ! 48 - 54       Real(7.2)      gamma (degrees)
         ! 56 - 66       LString        Space group
         ! 67 - 70       Integer        Z value
         WRITE (UNIT=iunit, FMT="(A6,3F9.3,3F7.2)") &
            "CRYST1", abc(1:3)*factor, angle_alpha, angle_beta, angle_gamma
         WRITE (UNIT=line(1:6), FMT="(A6)") "ATOM  "
         DO iatom = 1, natom
            line = ""
            ! COLUMNS        DATA TYPE       CONTENTS
            !  1 -  6        Record name     "ATOM  "
            !  7 - 11        Integer         Atom serial number
            ! 13 - 16        Atom            Atom name
            ! 17             Character       Alternate location indicator
            ! 18 - 20        Residue name    Residue name
            ! 22             Character       Chain identifier
            ! 23 - 26        Integer         Residue sequence number
            ! 27             AChar           Code for insertion of residues
            ! 31 - 38        Real(8.3)       Orthogonal coordinates for X in Angstrom
            ! 39 - 46        Real(8.3)       Orthogonal coordinates for Y in Angstrom
            ! 47 - 54        Real(8.3)       Orthogonal coordinates for Z in Angstrom
            ! 55 - 60        Real(6.2)       Occupancy
            ! 61 - 66        Real(6.2)       Temperature factor (Default = 0.0)
            ! 73 - 76        LString(4)      Segment identifier, left-justified
            ! 77 - 78        LString(2)      Element symbol, right-justified
            ! 79 - 80        LString(2)      Charge on the atom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol, name=atm_name, &
                                 fist_potential=fist_potential, shell=shell)
            IF (LEN_TRIM(element_symbol) == 0) THEN
               dummy = qmmm_ff_precond_only_qm(id1=atm_name)
            END IF
            name = TRIM(atm_name)
            IF (ASSOCIATED(fist_potential)) THEN
               CALL get_potential(potential=fist_potential, qeff=qeff)
            ELSE
               qeff = 0.0_dp
            END IF
            IF (ASSOCIATED(shell)) CALL get_shell(shell=shell, charge=qeff)
            WRITE (UNIT=line(1:6), FMT="(A6)") "ATOM  "
            WRITE (UNIT=line(7:11), FMT="(I5)") MODULO(iatom, 100000)
            WRITE (UNIT=line(13:16), FMT="(A4)") ADJUSTL(name)
            ! WRITE (UNIT=line(18:20),FMT="(A3)") TRIM(resname)
            ! WRITE (UNIT=line(23:26),FMT="(I4)") MODULO(idres,10000)
            SELECT CASE (TRIM (content))
            CASE ("POS")
               IF (PRESENT(array)) THEN
                  r(1:3) = get_particle_pos_or_vel(iatom, particle_set, array)
               ELSE
                  r(:) = particle_set(iatom)%r(:)
               END IF
               WRITE (UNIT=line(31:54), FMT="(3F8.3)") r(1:3)*factor
            CASE DEFAULT
               CALL cp__b("particle_methods.F",451,"PDB dump only for trajectory available")
            END SELECT
            IF (my_charge_occup) THEN
               WRITE (UNIT=line(55:60), FMT="(F6.2)") qeff
            ELSE
               WRITE (UNIT=line(55:60), FMT="(F6.2)") 0.0_dp
            END IF
            IF (my_charge_beta) THEN
               WRITE (UNIT=line(61:66), FMT="(F6.2)") qeff
            ELSE
               WRITE (UNIT=line(61:66), FMT="(F6.2)") 0.0_dp
            END IF
            ! WRITE (UNIT=line(73:76),FMT="(A4)") ADJUSTL(TRIM(molname))
            WRITE (UNIT=line(77:78), FMT="(A2)") ADJUSTR(TRIM(element_symbol))
            IF (my_charge_extended) THEN
               WRITE (UNIT=line(81:), FMT="(SP,F0.8)") qeff
            END IF
            WRITE (UNIT=iunit, FMT="(A)") TRIM(line)
         END DO
         WRITE (UNIT=iunit, FMT="(A)") "END"
      CASE DEFAULT
         CALL cp__b("particle_methods.F",472,"Illegal dump type")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE write_particle_coordinates

! **************************************************************************************************
!> \brief   Write the atomic coordinates to the output unit.
!> \param particle_set ...
!> \param subsys_section ...
!> \param charges ...
!> \date    05.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_fist_particle_coordinates(particle_set, subsys_section, &
                                              charges)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: subsys_section
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'write_fist_particle_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: iatom, ikind, iw, natom
      REAL(KIND=dp)                                      :: conv, mass, qcore, qeff, qshell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(shell_kind_type), POINTER                     :: shell_kind

      NULLIFY (logger)
      NULLIFY (shell_kind)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, &
                                "PRINT%ATOMIC_COORDINATES", extension=".coordLog")

      CALL section_vals_val_get(subsys_section, "PRINT%ATOMIC_COORDINATES%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      CALL uppercase(unit_str)
      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,/,T2,A)") &
            "MODULE FIST:  ATOMIC COORDINATES IN "//TRIM(unit_str)
         WRITE (UNIT=iw, &
                FMT="(/,T3,A,7X,2(A1,11X),A1,8X,A8,5X,A6,/)") &
            "Atom  Kind  ATM_TYP", "X", "Y", "Z", "  q(eff)", "  Mass"
         natom = SIZE(particle_set)
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 kind_number=ikind, &
                                 name=name, &
                                 mass=mass, &
                                 qeff=qeff, &
                                 shell=shell_kind)
            IF (ASSOCIATED(charges)) qeff = charges(iatom)
            IF (ASSOCIATED(shell_kind)) THEN
               CALL get_shell(shell=shell_kind, &
                              charge_core=qcore, &
                              charge_shell=qshell)
               qeff = qcore+qshell
            END IF
            WRITE (UNIT=iw, &
                   FMT="(T2,I5,1X,I4,3X,A4,3X,3F12.6,4X,F6.2,2X,F11.4)") &
               iatom, ikind, name, &
               particle_set(iatom)%r(1:3)*conv, qeff, mass/massunit
         END DO
         WRITE (iw, '(/)')
      END IF

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%ATOMIC_COORDINATES")

   END SUBROUTINE write_fist_particle_coordinates

! **************************************************************************************************
!> \brief   Write the atomic coordinates to the output unit.
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \param label ...
!> \date    05.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_qs_particle_coordinates(particle_set, qs_kind_set, subsys_section, label)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'write_qs_particle_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_string_length)               :: unit_str
      INTEGER                                            :: handle, iatom, ikind, iw, natom, z
      REAL(KIND=dp)                                      :: conv, mass, zeff
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, &
                                "PRINT%ATOMIC_COORDINATES", extension=".coordLog")

      CALL section_vals_val_get(subsys_section, "PRINT%ATOMIC_COORDINATES%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,/,T2,A)") &
            "MODULE "//TRIM(label)//":  ATOMIC COORDINATES IN "//TRIM(unit_str)
         WRITE (UNIT=iw, &
                FMT="(/,T3,A,7X,2(A1,11X),A1,8X,A8,5X,A6,/)") &
            "Atom  Kind  Element", "X", "Y", "Z", "  Z(eff)", "  Mass"

         natom = SIZE(particle_set)
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 kind_number=ikind, &
                                 element_symbol=element_symbol, &
                                 mass=mass, &
                                 z=z)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)
            WRITE (UNIT=iw, &
                   FMT="(T2,I7,1X,I5,1X,A2,1X,I3,3F12.6,4X,F6.2,2X,F11.4)") &
               iatom, ikind, element_symbol, z, &
               particle_set(iatom)%r(1:3)*conv, zeff, mass/massunit
         END DO
         WRITE (iw, '(/)')
      END IF

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%ATOMIC_COORDINATES")

      CALL timestop(handle)

   END SUBROUTINE write_qs_particle_coordinates

! **************************************************************************************************
!> \brief   Write the matrix of the particle distances to the output unit.
!> \param particle_set ...
!> \param cell ...
!> \param subsys_section ...
!> \date    06.10.2000
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_particle_distances(particle_set, cell, subsys_section)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_particle_distances', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: unit_str
      INTEGER                                            :: handle, iatom, iw, jatom, natom
      INTEGER, DIMENSION(3)                              :: periodic
      REAL(KIND=dp)                                      :: conv, dab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: distance_matrix
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, &
                                "PRINT%INTERATOMIC_DISTANCES", extension=".distLog")

      CALL section_vals_val_get(subsys_section, "PRINT%INTERATOMIC_DISTANCES%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (iw > 0) THEN
         CALL get_cell(cell=cell, periodic=periodic)
         natom = SIZE(particle_set)
         ALLOCATE (distance_matrix(natom, natom))
         distance_matrix(:, :) = 0.0_dp
         DO iatom = 1, natom
            DO jatom = iatom+1, natom
               rab(:) = pbc(particle_set(iatom)%r(:), &
                            particle_set(jatom)%r(:), cell)
               dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))
               distance_matrix(iatom, jatom) = dab*conv
               distance_matrix(jatom, iatom) = distance_matrix(iatom, jatom)
            END DO
         END DO

         ! Print the distance matrix
         WRITE (UNIT=iw, FMT="(/,/,T2,A)") &
            "INTERATOMIC DISTANCES IN "//TRIM(unit_str)

         CALL write_particle_matrix(distance_matrix, particle_set, iw)
      END IF

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%INTERATOMIC_DISTANCES")

      CALL timestop(handle)

   END SUBROUTINE write_particle_distances

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param particle_set ...
!> \param iw ...
!> \param el_per_part ...
!> \param Ilist ...
! **************************************************************************************************
   SUBROUTINE write_particle_matrix(matrix, particle_set, iw, el_per_part, Ilist)
      REAL(KIND=dp), DIMENSION(:, :)                     :: matrix
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: iw
      INTEGER, INTENT(IN), OPTIONAL                      :: el_per_part
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Ilist

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: from, i, iatom, icol, jatom, katom, &
                                                            my_el_per_part, natom, to
      INTEGER, DIMENSION(:), POINTER                     :: my_list

      my_el_per_part = 1
      IF (PRESENT(el_per_part)) my_el_per_part = el_per_part
      IF (PRESENT(Ilist)) THEN
         natom = SIZE(Ilist)
      ELSE
         natom = SIZE(particle_set)
      END IF
      ALLOCATE (my_list(natom))
      IF (PRESENT(Ilist)) THEN
         my_list = Ilist
      ELSE
         DO i = 1, natom
            my_list(i) = i
         END DO
      END IF
      natom = natom*my_el_per_part
      DO jatom = 1, natom, 5
         from = jatom
         to = MIN(from+4, natom)
         WRITE (UNIT=iw, FMT="(/,T2,11X,5(4X,I5,4X))") &
            (icol, icol=from, to)
         DO iatom = 1, natom
            katom = iatom/my_el_per_part
            IF (MOD(iatom, my_el_per_part) /= 0) katom = katom+1
            CALL get_atomic_kind(atomic_kind=particle_set(my_list(katom))%atomic_kind, &
                                 element_symbol=element_symbol)
            WRITE (UNIT=iw, FMT="(T2,I5,2X,A2,2X,5(1X,F12.6))") &
               iatom, element_symbol, &
               (matrix(iatom, icol), icol=from, to)
         END DO
      END DO
      DEALLOCATE (my_list)
   END SUBROUTINE write_particle_matrix

! **************************************************************************************************
!> \brief   Write structure data requested by a separate structure data input
!>          section to the output unit.
!>          input_section can be either motion_section or subsys_section.
!>
!> \param particle_set ...
!> \param cell ...
!> \param input_section ...
!> \date    11.03.04
!> \par History
!>          Recovered (23.03.06,MK)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_structure_data(particle_set, cell, input_section)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: input_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_structure_data', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string, unit_str
      INTEGER                                            :: handle, i, i_rep, iw, n, n_rep, n_vals, &
                                                            natom, new_size, old_size, wrk2(2), &
                                                            wrk3(3), wrk4(4)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: work
      INTEGER, DIMENSION(:), POINTER                     :: atomic_indices, index_list
      LOGICAL                                            :: unique
      REAL(KIND=dp)                                      :: conv, dab
      REAL(KIND=dp), DIMENSION(3)                        :: r, rab, rbc, rcd, s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: section

      CALL timeset(routineN, handle)
      NULLIFY (atomic_indices)
      NULLIFY (index_list)
      NULLIFY (logger)
      NULLIFY (section)
      string = ""

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger=logger, &
                                basis_section=input_section, &
                                print_key_path="PRINT%STRUCTURE_DATA", &
                                extension=".coordLog")

      CALL section_vals_val_get(input_section, "PRINT%STRUCTURE_DATA%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      CALL uppercase(unit_str)
      IF (iw > 0) THEN
         natom = SIZE(particle_set)
         section => section_vals_get_subs_vals(section_vals=input_section, &
                                               subsection_name="PRINT%STRUCTURE_DATA")

         WRITE (UNIT=iw, FMT="(/,T2,A)") "REQUESTED STRUCTURE DATA"
         ! Print the requested atomic position vectors
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="POSITION", &
                                   n_rep_val=n_rep)
         IF (n_rep > 0) THEN
            WRITE (UNIT=iw, FMT="(/,T3,A,/)") &
               "Position vectors r(i) of the atoms i in "//TRIM(unit_str)
            old_size = 0
            DO i_rep = 1, n_rep
               CALL section_vals_val_get(section_vals=section, &
                                         keyword_name="POSITION", &
                                         i_rep_val=i_rep, &
                                         i_vals=atomic_indices)
               n_vals = SIZE(atomic_indices)
               new_size = old_size+n_vals
               CALL reallocate(index_list, 1, new_size)
               index_list(old_size+1:new_size) = atomic_indices(1:n_vals)
               old_size = new_size
            END DO
            ALLOCATE (work(new_size))
            CALL sort(index_list, new_size, work)
            DEALLOCATE (work)
            DO i = 1, new_size
               WRITE (UNIT=string, FMT="(A,I0,A)") "(", index_list(i), ")"
               IF ((index_list(i) < 1) .OR. (index_list(i) > natom)) THEN
                  WRITE (UNIT=iw, FMT="(T3,A)") &
                     "Invalid atomic index "//TRIM(string)//" specified. Print request is ignored."
                  CYCLE
               END IF
               IF (i > 1) THEN
                  ! Skip redundant indices
                  IF (index_list(i) == index_list(i-1)) CYCLE
               END IF
               WRITE (UNIT=iw, FMT="(T3,A,T20,A,3F13.6)") &
                  "r"//TRIM(string), "=", pbc(particle_set(index_list(i))%r(1:3), cell)*conv
            END DO
            DEALLOCATE (index_list)
         END IF

         ! Print the requested atomic position vectors in scaled coordinates
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="POSITION_SCALED", &
                                   n_rep_val=n_rep)
         IF (n_rep > 0) THEN
            WRITE (UNIT=iw, FMT="(/,T3,A,/)") &
               "Position vectors s(i) of the atoms i in scaled coordinates"
            old_size = 0
            DO i_rep = 1, n_rep
               CALL section_vals_val_get(section_vals=section, &
                                         keyword_name="POSITION_SCALED", &
                                         i_rep_val=i_rep, &
                                         i_vals=atomic_indices)
               n_vals = SIZE(atomic_indices)
               new_size = old_size+n_vals
               CALL reallocate(index_list, 1, new_size)
               index_list(old_size+1:new_size) = atomic_indices(1:n_vals)
               old_size = new_size
            END DO
            ALLOCATE (work(new_size))
            CALL sort(index_list, new_size, work)
            DEALLOCATE (work)
            DO i = 1, new_size
               WRITE (UNIT=string, FMT="(A,I0,A)") "(", index_list(i), ")"
               IF ((index_list(i) < 1) .OR. (index_list(i) > natom)) THEN
                  WRITE (UNIT=iw, FMT="(T3,A)") &
                     "Invalid atomic index "//TRIM(string)//" specified. Print request is ignored."
                  CYCLE
               END IF
               IF (i > 1) THEN
                  ! Skip redundant indices
                  IF (index_list(i) == index_list(i-1)) CYCLE
               END IF
               r(1:3) = pbc(particle_set(index_list(i))%r(1:3), cell)
               CALL real_to_scaled(s, r, cell)
               WRITE (UNIT=iw, FMT="(T3,A,T20,A,3F13.6)") &
                  "s"//TRIM(string), "=", s(1:3)
            END DO
            DEALLOCATE (index_list)
         END IF

         ! Print the requested distances
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="DISTANCE", &
                                   n_rep_val=n)
         IF (n > 0) THEN
            WRITE (UNIT=iw, FMT="(/,T3,A,/)") &
               "Distance vector r(i,j) between the atom i and j in "// &
               TRIM(unit_str)
            DO i = 1, n
               CALL section_vals_val_get(section_vals=section, &
                                         keyword_name="DISTANCE", &
                                         i_rep_val=i, &
                                         i_vals=atomic_indices)
               string = ""
               WRITE (UNIT=string, FMT="(A,2(I0,A))") &
                  "(", atomic_indices(1), ",", atomic_indices(2), ")"
               wrk2 = atomic_indices
               CALL sort_unique(wrk2, unique)
               IF (((wrk2(1) >= 1) .AND. (wrk2(SIZE(wrk2)) <= natom)) .AND. unique) THEN
                  rab(:) = pbc(particle_set(atomic_indices(1))%r(:), &
                               particle_set(atomic_indices(2))%r(:), cell)
                  dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))
                  WRITE (UNIT=iw, FMT="(T3,A,T20,A,3F13.6,3X,A,F13.6)") &
                     "r"//TRIM(string), "=", rab(:)*conv, &
                     "|r| =", dab*conv
               ELSE
                  WRITE (UNIT=iw, FMT="(T3,A)") &
                     "Invalid atomic indices "//TRIM(string)//" specified. Print request is ignored."
               END IF
            END DO
         END IF

         ! Print the requested angles
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="ANGLE", &
                                   n_rep_val=n)
         IF (n > 0) THEN
            WRITE (UNIT=iw, FMT="(/,T3,A,/)") &
               "Angle a(i,j,k) between the atomic distance vectors r(j,i) and "// &
               "r(j,k) in DEGREE"
            DO i = 1, n
               CALL section_vals_val_get(section_vals=section, &
                                         keyword_name="ANGLE", &
                                         i_rep_val=i, &
                                         i_vals=atomic_indices)
               string = ""
               WRITE (UNIT=string, FMT="(A,3(I0,A))") &
                  "(", atomic_indices(1), ",", atomic_indices(2), ",", atomic_indices(3), ")"
               wrk3 = atomic_indices
               CALL sort_unique(wrk3, unique)
               IF (((wrk3(1) >= 1) .AND. (wrk3(SIZE(wrk3)) <= natom)) .AND. unique) THEN
                  rab(:) = pbc(particle_set(atomic_indices(1))%r(:), &
                               particle_set(atomic_indices(2))%r(:), cell)
                  rbc(:) = pbc(particle_set(atomic_indices(2))%r(:), &
                               particle_set(atomic_indices(3))%r(:), cell)
                  WRITE (UNIT=iw, FMT="(T3,A,T26,A,F9.3)") &
                     "a"//TRIM(string), "=", angle(-rab, rbc)*degree
               ELSE
                  WRITE (UNIT=iw, FMT="(T3,A)") &
                     "Invalid atomic indices "//TRIM(string)//" specified. Print request is ignored."
               END IF
            END DO
         END IF

         ! Print the requested dihedral angles
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="DIHEDRAL_ANGLE", &
                                   n_rep_val=n)
         IF (n > 0) THEN
            WRITE (UNIT=iw, FMT="(/,T3,A,/)") &
               "Dihedral angle d(i,j,k,l) between the planes (i,j,k) and (j,k,l) "// &
               "in DEGREE"
            DO i = 1, n
               CALL section_vals_val_get(section_vals=section, &
                                         keyword_name="DIHEDRAL_ANGLE", &
                                         i_rep_val=i, &
                                         i_vals=atomic_indices)
               string = ""
               WRITE (UNIT=string, FMT="(A,4(I0,A))") &
                  "(", atomic_indices(1), ",", atomic_indices(2), ",", &
                  atomic_indices(3), ",", atomic_indices(4), ")"
               wrk4 = atomic_indices
               CALL sort_unique(wrk4, unique)
               IF (((wrk4(1) >= 1) .AND. (wrk4(SIZE(wrk4)) <= natom)) .AND. unique) THEN
                  rab(:) = pbc(particle_set(atomic_indices(1))%r(:), &
                               particle_set(atomic_indices(2))%r(:), cell)
                  rbc(:) = pbc(particle_set(atomic_indices(2))%r(:), &
                               particle_set(atomic_indices(3))%r(:), cell)
                  rcd(:) = pbc(particle_set(atomic_indices(3))%r(:), &
                               particle_set(atomic_indices(4))%r(:), cell)
                  WRITE (UNIT=iw, FMT="(T3,A,T26,A,F9.3)") &
                     "d"//TRIM(string), "=", dihedral_angle(rab, rbc, rcd)*degree
               ELSE
                  WRITE (UNIT=iw, FMT="(T3,A)") &
                     "Invalid atomic indices "//TRIM(string)//" specified. Print request is ignored."
               END IF
            END DO
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, input_section, &
                                        "PRINT%STRUCTURE_DATA")

      CALL timestop(handle)

   END SUBROUTINE write_structure_data

END MODULE particle_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Distribution methods for atoms, particles, or molecules
!> \par History
!>      - 1d-distribution of molecules and particles (Sep. 2003, MK)
!>      - 2d-distribution for Quickstep updated with molecules (Oct. 2003, MK)
!> \author MK (22.08.2003)
! **************************************************************************************************
MODULE distribution_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE cp_array_utils_i,                ONLY: cp_1d_i_p_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_distribution_get_num_images,&
                                              heap_fill,&
                                              heap_get_first,&
                                              heap_new,&
                                              heap_release,&
                                              heap_reset_first,&
                                              heap_t
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE distribution_1d_types,           ONLY: distribution_1d_create,&
                                              distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_create,&
                                              distribution_2d_type,&
                                              distribution_2d_write
   USE input_constants,                 ONLY: model_block_count,&
                                              model_block_lmax
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              int_size
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: gcd,&
                                              lcm
   USE message_passing,                 ONLY: mp_sum,&
                                              mp_sync
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              get_molecule_kind_set,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (in this module) ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'distribution_methods'

! *** Public subroutines ***

   PUBLIC :: distribute_molecules_1d, &
             distribute_molecules_2d

CONTAINS

! **************************************************************************************************
!> \brief Distribute molecules and particles
!> \param atomic_kind_set particle (atomic) kind information
!> \param particle_set particle information
!> \param local_particles distribution of particles created by this routine
!> \param molecule_kind_set molecule kind information
!> \param molecule_set molecule information
!> \param local_molecules distribution of molecules created by this routine
!> \param force_env_section ...
!> \param prev_molecule_kind_set previous molecule kind information, used with
!>        prev_local_molecules
!> \param prev_local_molecules previous distribution of molecules, new one will
!>        be identical if all the prev_* arguments are present and associated
!> \par History
!>      none
!> \author MK (Jun. 2003)
! **************************************************************************************************
   SUBROUTINE distribute_molecules_1d(atomic_kind_set, particle_set, &
                                      local_particles, &
                                      molecule_kind_set, molecule_set, &
                                      local_molecules, force_env_section, &
                                      prev_molecule_kind_set, &
                                      prev_local_molecules)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(section_vals_type), POINTER                   :: force_env_section
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: prev_molecule_kind_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: prev_local_molecules

      CHARACTER(len=*), PARAMETER :: routineN = 'distribute_molecules_1d', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, bin, bin_price, group, handle, iatom, imolecule, imolecule_kind, &
         imolecule_local, imolecule_prev_kind, iparticle_kind, ipe, istat, iw, kind_a, molecule_a, &
         mype, n, natom, nbins, nload, nmolecule, nmolecule_kind, nparticle_kind, npe, nsgf, &
         output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nmolecule_local, nparticle_local, work, &
                                                            workload_count, workload_fill
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, has_prev_subsys_info, heap_error, &
                                                            is_local
      TYPE(cp_1d_i_p_type), ALLOCATABLE, DIMENSION(:)    :: local_molecule
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(heap_t)                                       :: bin_heap_count, bin_heap_fill
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

! integer n_local_molecules
! integer, allocatable :: all_local_molecules(:)

      CALL timeset(routineN, handle)

      has_prev_subsys_info = .FALSE.
      IF (PRESENT(prev_local_molecules) .AND. &
          PRESENT(prev_molecule_kind_set)) THEN
         IF (ASSOCIATED(prev_local_molecules) .AND. &
             ASSOCIATED(prev_molecule_kind_set)) THEN
            has_prev_subsys_info = .TRUE.
         ENDIF
      ENDIF

      logger => cp_get_default_logger()

      group = logger%para_env%group
      mype = logger%para_env%mepos+1
      npe = logger%para_env%num_pe

      ALLOCATE (workload_count(npe), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 163, &
                                       "workload", npe*int_size)
      workload_count(:) = 0

      ALLOCATE (workload_fill(npe), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 168, &
                                       "workload_old", npe*int_size)
      workload_fill(:) = 0

      nmolecule_kind = SIZE(molecule_kind_set)

      ALLOCATE (nmolecule_local(nmolecule_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 175, &
                                       "nmolecule_local", nmolecule_kind*int_size)
      nmolecule_local(:) = 0

      ALLOCATE (local_molecule(nmolecule_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 180, &
                                       "local_molecule", nmolecule_kind*int_size)

      nparticle_kind = SIZE(atomic_kind_set)

      ALLOCATE (nparticle_local(nparticle_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 186, &
                                       "nparticle_local", nparticle_kind*int_size)
      nparticle_local(:) = 0

      nbins = npe

      CALL heap_new(bin_heap_count, nbins)
      CALL heap_fill(bin_heap_count, &
                     (/(bin, bin=1, nbins)/), workload_count, heap_error)
      IF (heap_error) &
         CALL cp__b("distribution_methods.F",196,"Error initially filling the heap.")

      CALL heap_new(bin_heap_fill, nbins)
      CALL heap_fill(bin_heap_fill, &
                     (/(bin, bin=1, nbins)/), workload_fill, heap_error)
      IF (heap_error) &
         CALL cp__b("distribution_methods.F",202,"Error initially filling the heap.")

      DO imolecule_kind = 1, nmolecule_kind

         molecule_kind => molecule_kind_set(imolecule_kind)

         NULLIFY (molecule_list)

!     *** Get the number of molecules and the number of ***
!     *** atoms in each molecule of that molecular kind ***

         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nsgf=nsgf)

!     *** Consider the number of atoms or basis ***
!     *** functions which depends on the method ***

         nload = MAX(natom, nsgf)
         nmolecule = SIZE(molecule_list)

!     *** Get the number of local molecules of the current molecule kind ***

         DO imolecule = 1, nmolecule
            IF (has_prev_subsys_info) THEN
               DO imolecule_prev_kind = 1, SIZE(prev_molecule_kind_set)
                  IF (ANY(prev_local_molecules%list(imolecule_prev_kind)%array( &
                          1:prev_local_molecules%n_el(imolecule_prev_kind)) == molecule_list(imolecule))) THEN
                     ! molecule used to be local
                     nmolecule_local(imolecule_kind) = nmolecule_local(imolecule_kind)+1
                  ENDIF
               END DO
            ELSE
               CALL heap_get_first(bin_heap_count, bin, bin_price, found, heap_error)
               IF (heap_error) &
                  CALL cp__b("distribution_methods.F",238,"Error getting topmost heap element.")
               IF (.NOT. found) &
                  CALL cp__b("distribution_methods.F",240,"No topmost heap element found.")

               ipe = bin
               IF (bin_price /= workload_count(ipe)) &
                  CALL cp__b("distribution_methods.F",244,"inconsistent heap")

               workload_count(ipe) = workload_count(ipe)+nload
               IF (ipe == mype) THEN
                  nmolecule_local(imolecule_kind) = nmolecule_local(imolecule_kind)+1
               END IF

               bin_price = workload_count(ipe)
               CALL heap_reset_first(bin_heap_count, bin_price, heap_error)
               IF (heap_error) &
                  CALL cp__w("distribution_methods.F",254,"Error setting price of top heap element.")
            END IF
         END DO

!     *** Distribute the molecules ***
         n = nmolecule_local(imolecule_kind)

         IF (n > 0) THEN
            ALLOCATE (local_molecule(imolecule_kind)%array(n), STAT=istat)
            IF (istat /= 0) THEN
               CALL stop_memory(routineN, moduleN, 264, &
                                "local_molecule(imolecule_kind)%array", &
                                n*int_size)
            END IF
         ELSE
            NULLIFY (local_molecule(imolecule_kind)%array)
         END IF

         imolecule_local = 0
         DO imolecule = 1, nmolecule
            is_local = .FALSE.
            IF (has_prev_subsys_info) THEN
               DO imolecule_prev_kind = 1, SIZE(prev_molecule_kind_set)
                  IF (ANY(prev_local_molecules%list(imolecule_prev_kind)%array( &
                          1:prev_local_molecules%n_el(imolecule_prev_kind)) == molecule_list(imolecule))) THEN
                     is_local = .TRUE.
                  END IF
               END DO
            ELSE
               CALL heap_get_first(bin_heap_fill, bin, bin_price, found, heap_error)
               IF (heap_error) &
                  CALL cp__b("distribution_methods.F",285,"Error getting topmost heap element.")
               IF (.NOT. found) &
                  CALL cp__b("distribution_methods.F",287,"No topmost heap element found.")

               ipe = bin
               IF (bin_price /= workload_fill(ipe)) &
                  CALL cp__b("distribution_methods.F",291,"inconsistent heap")

               workload_fill(ipe) = workload_fill(ipe)+nload
               is_local = (ipe == mype)
            ENDIF
            IF (is_local) THEN
               imolecule_local = imolecule_local+1
               molecule_a = molecule_list(imolecule)
               local_molecule(imolecule_kind)%array(imolecule_local) = molecule_a
               DO iatom = 1, natom
                  atom_a = molecule_set(molecule_a)%first_atom+iatom-1

                  CALL get_atomic_kind(atomic_kind=particle_set(atom_a)%atomic_kind, &
                                       kind_number=kind_a)
                  nparticle_local(kind_a) = nparticle_local(kind_a)+1
               END DO
            END IF
            IF (.NOT. has_prev_subsys_info) THEN
               bin_price = workload_fill(ipe)
               CALL heap_reset_first(bin_heap_fill, bin_price, heap_error)
               IF (heap_error) &
                  CALL cp__w("distribution_methods.F",312,"Error setting price of top heap element.")
            END IF
         END DO

      END DO

      IF (ANY(workload_fill .NE. workload_count)) &
         CALL cp__b("distribution_methods.F",319,"Inconsistent heaps encountered")

      CALL heap_release(bin_heap_count)
      CALL heap_release(bin_heap_fill)

!   *** Create the local molecule structure ***

      CALL distribution_1d_create(local_molecules, &
                                  n_el=nmolecule_local, &
                                  para_env=logger%para_env)

!   *** Create the local particle structure ***

      CALL distribution_1d_create(local_particles, &
                                  n_el=nparticle_local, &
                                  para_env=logger%para_env)

!   *** Store the generated local molecule and particle distributions ***

      nparticle_local(:) = 0

      DO imolecule_kind = 1, nmolecule_kind

         IF (nmolecule_local(imolecule_kind) == 0) CYCLE

         local_molecules%list(imolecule_kind)%array(:) = &
            local_molecule(imolecule_kind)%array(:)

         molecule_kind => molecule_kind_set(imolecule_kind)

         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                natom=natom)

         DO imolecule = 1, nmolecule_local(imolecule_kind)
            molecule_a = local_molecule(imolecule_kind)%array(imolecule)
            DO iatom = 1, natom
               atom_a = molecule_set(molecule_a)%first_atom+iatom-1
               CALL get_atomic_kind(atomic_kind=particle_set(atom_a)%atomic_kind, &
                                    kind_number=kind_a)
               nparticle_local(kind_a) = nparticle_local(kind_a)+1
               local_particles%list(kind_a)%array(nparticle_local(kind_a)) = atom_a
            END DO
         END DO

      END DO

!   *** Print distribution, if requested ***

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           force_env_section, "PRINT%DISTRIBUTION1D"), cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, force_env_section, "PRINT%DISTRIBUTION1D", &
                                            extension=".Log")

         iw = output_unit
         IF (output_unit < 0) iw = cp_logger_get_default_unit_nr(logger, LOCAL=.TRUE.)

!     *** Print molecule distribution ***

         ALLOCATE (work(npe), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 379, &
                                          "work", npe*int_size)
         work(:) = 0

         work(mype) = SUM(nmolecule_local)
         CALL mp_sum(work, group)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/, T2, A, T51, A, /, (T52, I6, T73, I8))") &
               "DISTRIBUTION OF THE MOLECULES", &
               "Process    Number of molecules", &
               (ipe-1, work(ipe), ipe=1, npe)
            WRITE (UNIT=output_unit, FMT="(T55, A3, T73, I8)") &
               "Sum", SUM(work)
            CALL m_flush(output_unit)
         END IF

         CALL mp_sync(group)

         DO ipe = 1, npe
            IF (ipe == mype) THEN
               WRITE (UNIT=iw, FMT="(/, T3, A)") &
                  "Process   Kind   Local molecules (global indices)"
               DO imolecule_kind = 1, nmolecule_kind
                  IF (imolecule_kind == 1) THEN
                     WRITE (UNIT=iw, FMT="(T4, I6, 2X, I5, (T21, 10I6))") &
                        ipe-1, imolecule_kind, &
                        (local_molecules%list(imolecule_kind)%array(imolecule), &
                         imolecule=1, nmolecule_local(imolecule_kind))
                  ELSE
                     WRITE (UNIT=iw, FMT="(T12, I5, (T21, 10I6))") &
                        imolecule_kind, &
                        (local_molecules%list(imolecule_kind)%array(imolecule), &
                         imolecule=1, nmolecule_local(imolecule_kind))
                  END IF
               END DO
            END IF
            CALL m_flush(iw)
            CALL mp_sync(group)
         END DO

!     *** Print particle distribution ***

         work(:) = 0

         work(mype) = SUM(nparticle_local)
         CALL mp_sum(work, group)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/, T2, A, T51, A, /, (T52, I6, T73, I8))") &
               "DISTRIBUTION OF THE PARTICLES", &
               "Process    Number of particles", &
               (ipe-1, work(ipe), ipe=1, npe)
            WRITE (UNIT=output_unit, FMT="(T55, A3, T73, I8)") &
               "Sum", SUM(work)
            CALL m_flush(output_unit)
         END IF

         CALL mp_sync(group)

         DO ipe = 1, npe
            IF (ipe == mype) THEN
               WRITE (UNIT=iw, FMT="(/, T3, A)") &
                  "Process   Kind   Local particles (global indices)"
               DO iparticle_kind = 1, nparticle_kind
                  IF (iparticle_kind == 1) THEN
                     WRITE (UNIT=iw, FMT="(T4, I6, 2X, I5, (T20, 10I6))") &
                        ipe-1, iparticle_kind, &
                        (local_particles%list(iparticle_kind)%array(iatom), &
                         iatom=1, nparticle_local(iparticle_kind))
                  ELSE
                     WRITE (UNIT=iw, FMT="(T12, I5, (T20, 10I6))") &
                        iparticle_kind, &
                        (local_particles%list(iparticle_kind)%array(iatom), &
                         iatom=1, nparticle_local(iparticle_kind))
                  END IF
               END DO
            END IF
            CALL m_flush(iw)
            CALL mp_sync(group)
         END DO
         DEALLOCATE (work)

         CALL cp_print_key_finished_output(output_unit, logger, force_env_section, &
                                           "PRINT%DISTRIBUTION1D")
      END IF
!   *** Release work storage ***

      DEALLOCATE (workload_count)

      DEALLOCATE (workload_fill)

      DEALLOCATE (nmolecule_local)

      DEALLOCATE (nparticle_local)

      DO imolecule_kind = 1, nmolecule_kind
         IF (ASSOCIATED(local_molecule(imolecule_kind)%array)) THEN
            DEALLOCATE (local_molecule(imolecule_kind)%array)
         END IF
      END DO
      DEALLOCATE (local_molecule)

      CALL timestop(handle)

   END SUBROUTINE distribute_molecules_1d

! **************************************************************************************************
!> \brief Distributes the particle pairs creating a 2d distribution optimally
!>      suited for quickstep
!> \param cell ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param distribution_2d the distribution that will be created by this
!>                         method
!> \param blacs_env the parallel environement at the basis of the
!>                   distribution
!> \param force_env_section ...
!> \par History
!>      - local_rows & cols blocksize optimizations (Aug. 2003, MK)
!>      - cleanup of distribution_2d (Sep. 2003, fawzi)
!>      - update for molecules (Oct. 2003, MK)
!> \author fawzi (Feb. 2003)
!> \note
!>      Intermediate generation of a 2d distribution of the molecules, but
!>      only the corresponding particle (atomic) distribution is currently
!>      used. The 2d distribution of the molecules is deleted, but may easily
!>      be recovered (MK).
! **************************************************************************************************
   SUBROUTINE distribute_molecules_2d(cell, atomic_kind_set, particle_set, &
                                      qs_kind_set, molecule_kind_set, molecule_set, &
                                      distribution_2d, blacs_env, force_env_section)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'distribute_molecules_2d', &
         routineP = moduleN//':'//routineN

      INTEGER :: cluster_price, cost_model, group, handle, iatom, iatom_mol, iatom_one, ikind, &
         imol, imolecule, imolecule_kind, iparticle_kind, ipcol, iprow, istat, iw, kind_a, mypcol, &
         myprow, n, natom, natom_mol, nclusters, nmolecule, nmolecule_kind, nparticle_kind, npcol, &
         nprow, nsgf, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cluster_list, cluster_prices, &
                                                            nparticle_local_col, &
                                                            nparticle_local_row, work
      INTEGER, DIMENSION(:), POINTER                     :: lmax_basis, molecule_list
      INTEGER, DIMENSION(:, :), POINTER                  :: cluster_col_distribution, &
                                                            cluster_row_distribution, &
                                                            col_distribution, row_distribution
      LOGICAL :: basic_cluster_optimization, basic_optimization, basic_spatial_optimization, &
         molecular_distribution, skip_optimization
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coords, pbc_scaled_coords
      REAL(KIND=dp), DIMENSION(3)                        :: center
      TYPE(cp_1d_i_p_type), DIMENSION(:), POINTER        :: local_particle_col, local_particle_row
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(section_vals_type), POINTER                   :: distribution_section

!...

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      distribution_section => section_vals_get_subs_vals(force_env_section, "DFT%QS%DISTRIBUTION")

      CALL section_vals_val_get(distribution_section, "2D_MOLECULAR_DISTRIBUTION", l_val=molecular_distribution)
      CALL section_vals_val_get(distribution_section, "SKIP_OPTIMIZATION", l_val=skip_optimization)
      CALL section_vals_val_get(distribution_section, "BASIC_OPTIMIZATION", l_val=basic_optimization)
      CALL section_vals_val_get(distribution_section, "BASIC_SPATIAL_OPTIMIZATION", l_val=basic_spatial_optimization)
      CALL section_vals_val_get(distribution_section, "BASIC_CLUSTER_OPTIMIZATION", l_val=basic_cluster_optimization)

      CALL section_vals_val_get(distribution_section, "COST_MODEL", i_val=cost_model)
      !

      group = blacs_env%para_env%group
      myprow = blacs_env%mepos(1)+1
      mypcol = blacs_env%mepos(2)+1
      nprow = blacs_env%num_pe(1)
      npcol = blacs_env%num_pe(2)

      nmolecule_kind = SIZE(molecule_kind_set)
      CALL get_molecule_kind_set(molecule_kind_set, nmolecule=nmolecule)

      nparticle_kind = SIZE(atomic_kind_set)
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, natom=natom)

      !
      ! we need to generate two representations of the distribution, one as a straight array with global particles
      ! one ordered wrt to kinds and only listing the local particles
      !
      ALLOCATE (row_distribution(natom, 2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 584, &
                                       "row_distribution", natom*int_size)

      ALLOCATE (col_distribution(natom, 2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 588, &
                                       "col_distribution", natom*int_size)
      ! Initialize the distributions to -1, as the second dimension only gets set with cluster optimization
      ! but the information is needed by dbcsr
      row_distribution = -1; col_distribution = -1

      ALLOCATE (local_particle_col(nparticle_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 595, &
                                       "local_particle_col", &
                                       nparticle_kind*int_size)

      ALLOCATE (local_particle_row(nparticle_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 600, &
                                       "local_particle_col", &
                                       nparticle_kind*int_size)

      ALLOCATE (nparticle_local_row(nparticle_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 605, &
                                       "nparticle_local_row", &
                                       nparticle_kind*int_size)

      ALLOCATE (nparticle_local_col(nparticle_kind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 610, &
                                       "nparticle_local_col", &
                                       nparticle_kind*int_size)

      IF (basic_optimization .OR. basic_spatial_optimization .OR. basic_cluster_optimization) THEN

         IF (molecular_distribution) THEN
            nclusters = nmolecule
         ELSE
            nclusters = natom
         ENDIF

         ALLOCATE (cluster_list(nclusters), stat=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 623, &
                                          "cluster_list", nclusters*int_size)
         ALLOCATE (cluster_prices(nclusters), stat=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 626, &
                                          "cluster_prices", nclusters*int_size)
         ALLOCATE (cluster_row_distribution(nclusters, 2), stat=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 629, &
                                          "cluster_prices", nclusters*int_size)
         ALLOCATE (cluster_col_distribution(nclusters, 2), stat=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 632, &
                                          "cluster_prices", nclusters*int_size)
         cluster_row_distribution = -1; cluster_col_distribution = -1

         ! Fill in the clusters and their prices
         CALL section_vals_val_get(distribution_section, "COST_MODEL", i_val=cost_model)
         IF (.NOT. molecular_distribution) THEN
            DO iatom = 1, natom
               IF (iatom .GT. nclusters) &
                  CALL cp__b("distribution_methods.F",641,"Bounds error")
               CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
               cluster_list(iatom) = iatom
               SELECT CASE (cost_model)
               CASE (model_block_count)
                  CALL get_qs_kind(qs_kind_set(ikind), nsgf=nsgf)
                  cluster_price = nsgf
               CASE (model_block_lmax)
                  CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
                  CALL get_gto_basis_set(orb_basis_set, lmax=lmax_basis)
                  cluster_price = MAXVAL(lmax_basis)
               CASE default
                  CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
                  CALL get_gto_basis_set(orb_basis_set, lmax=lmax_basis)
                  cluster_price = 8+(MAXVAL(lmax_basis)**2)
               END SELECT
               cluster_prices(iatom) = cluster_price
            ENDDO
         ELSE
            imol = 0
            DO imolecule_kind = 1, nmolecule_kind
               molecule_kind => molecule_kind_set(imolecule_kind)
               CALL get_molecule_kind(molecule_kind=molecule_kind, molecule_list=molecule_list, natom=natom_mol)
               DO imolecule = 1, SIZE(molecule_list)
                  imol = imol+1
                  cluster_list(imol) = imol
                  cluster_price = 0
                  DO iatom_mol = 1, natom_mol
                     iatom = molecule_set(molecule_list(imolecule))%first_atom+iatom_mol-1
                     CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
                     SELECT CASE (cost_model)
                     CASE (model_block_count)
                        CALL get_qs_kind(qs_kind_set(ikind), nsgf=nsgf)
                        cluster_price = cluster_price+nsgf
                     CASE (model_block_lmax)
                        CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
                        CALL get_gto_basis_set(orb_basis_set, lmax=lmax_basis)
                        cluster_price = cluster_price+MAXVAL(lmax_basis)
                     CASE default
                        CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
                        CALL get_gto_basis_set(orb_basis_set, lmax=lmax_basis)
                        cluster_price = cluster_price+8+(MAXVAL(lmax_basis)**2)
                     END SELECT
                  ENDDO
                  cluster_prices(imol) = cluster_price
               ENDDO
            ENDDO
         ENDIF

         ! And distribute
         IF (basic_optimization) THEN
            CALL make_basic_distribution(cluster_list, cluster_prices, &
                                         nprow, cluster_row_distribution(:, 1), npcol, cluster_col_distribution(:, 1))
         ELSE
            IF (basic_cluster_optimization) THEN
               IF (molecular_distribution) &
                  CALL cp__b("distribution_methods.F",697,"clustering and molecular blocking NYI")
               ALLOCATE (pbc_scaled_coords(3, natom), coords(3, natom))
               DO iatom = 1, natom
                  CALL real_to_scaled(pbc_scaled_coords(:, iatom), pbc(particle_set(iatom)%r(:), cell), cell)
                  coords(:, iatom) = pbc(particle_set(iatom)%r(:), cell)
               ENDDO
               CALL make_cluster_distribution(coords, pbc_scaled_coords, cell, cluster_prices, &
                                              nprow, cluster_row_distribution, npcol, cluster_col_distribution)
            ELSE ! basic_spatial_optimization
               ALLOCATE (pbc_scaled_coords(3, nclusters))
               IF (.NOT. molecular_distribution) THEN
                  ! just scaled coords
                  DO iatom = 1, natom
                     CALL real_to_scaled(pbc_scaled_coords(:, iatom), pbc(particle_set(iatom)%r(:), cell), cell)
                  ENDDO
               ELSE
                  ! use scaled coords of geometric center, folding when appropriate
                  imol = 0
                  DO imolecule_kind = 1, nmolecule_kind
                     molecule_kind => molecule_kind_set(imolecule_kind)
                     CALL get_molecule_kind(molecule_kind=molecule_kind, molecule_list=molecule_list, natom=natom_mol)
                     DO imolecule = 1, SIZE(molecule_list)
                        imol = imol+1
                        iatom_one = molecule_set(molecule_list(imolecule))%first_atom
                        center = 0.0_dp
                        DO iatom_mol = 1, natom_mol
                           iatom = molecule_set(molecule_list(imolecule))%first_atom+iatom_mol-1
                           center = center+ &
                                    pbc(particle_set(iatom)%r(:)-particle_set(iatom_one)%r(:), cell)+particle_set(iatom_one)%r(:)
                        ENDDO
                        center = center/natom_mol
                        CALL real_to_scaled(pbc_scaled_coords(:, imol), pbc(center, cell), cell)
                     ENDDO
                  ENDDO
               ENDIF

               CALL make_basic_spatial_distribution(pbc_scaled_coords, cluster_prices, &
                                                    nprow, cluster_row_distribution(:, 1), npcol, cluster_col_distribution(:, 1))

               DEALLOCATE (pbc_scaled_coords)
            END IF
         ENDIF

         ! And assign back
         IF (.NOT. molecular_distribution) THEN
            row_distribution = cluster_row_distribution
            col_distribution = cluster_col_distribution
         ELSE
            imol = 0
            DO imolecule_kind = 1, nmolecule_kind
               molecule_kind => molecule_kind_set(imolecule_kind)
               CALL get_molecule_kind(molecule_kind=molecule_kind, molecule_list=molecule_list, natom=natom_mol)
               DO imolecule = 1, SIZE(molecule_list)
                  imol = imol+1
                  DO iatom_mol = 1, natom_mol
                     iatom = molecule_set(molecule_list(imolecule))%first_atom+iatom_mol-1
                     row_distribution(iatom, :) = cluster_row_distribution(imol, :)
                     col_distribution(iatom, :) = cluster_col_distribution(imol, :)
                  ENDDO
               ENDDO
            ENDDO
         ENDIF

         ! cleanup
         DEALLOCATE (cluster_list)
         DEALLOCATE (cluster_prices)
         DEALLOCATE (cluster_row_distribution)
         DEALLOCATE (cluster_col_distribution)

      ELSE
         ! expects nothing else
         CALL cp__b("distribution_methods.F",768,"")
      ENDIF

      ! prepare the lists of local particles

      ! count local particles of a given kind
      nparticle_local_col = 0
      nparticle_local_row = 0
      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, kind_number=kind_a)
         IF (row_distribution(iatom, 1) == myprow) nparticle_local_row(kind_a) = nparticle_local_row(kind_a)+1
         IF (col_distribution(iatom, 1) == mypcol) nparticle_local_col(kind_a) = nparticle_local_col(kind_a)+1
      ENDDO

      ! allocate space
      DO iparticle_kind = 1, nparticle_kind
         n = nparticle_local_row(iparticle_kind)
         ALLOCATE (local_particle_row(iparticle_kind)%array(n), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 786, &
                                          "local_particle_row(iparticle_kind)%array", n*int_size)

         n = nparticle_local_col(iparticle_kind)
         ALLOCATE (local_particle_col(iparticle_kind)%array(n), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 791, &
                                          "local_particle_col(iparticle_kind)%array", n*int_size)
      ENDDO

      ! store
      nparticle_local_col = 0
      nparticle_local_row = 0
      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, kind_number=kind_a)
         IF (row_distribution(iatom, 1) == myprow) THEN
            nparticle_local_row(kind_a) = nparticle_local_row(kind_a)+1
            local_particle_row(kind_a)%array(nparticle_local_row(kind_a)) = iatom
         ENDIF
         IF (col_distribution(iatom, 1) == mypcol) THEN
            nparticle_local_col(kind_a) = nparticle_local_col(kind_a)+1
            local_particle_col(kind_a)%array(nparticle_local_col(kind_a)) = iatom
         ENDIF
      ENDDO

!   *** Generate the 2d distribution structure  but take care of the zero offsets required
      row_distribution(:, 1) = row_distribution(:, 1)-1
      col_distribution(:, 1) = col_distribution(:, 1)-1
      CALL distribution_2d_create(distribution_2d, &
                                  row_distribution_ptr=row_distribution, &
                                  col_distribution_ptr=col_distribution, &
                                  local_rows_ptr=local_particle_row, &
                                  local_cols_ptr=local_particle_col, &
                                  blacs_env=blacs_env)

      NULLIFY (local_particle_row)
      NULLIFY (local_particle_col)
      NULLIFY (row_distribution)
      NULLIFY (col_distribution)

!   *** Print distribution, if requested ***
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           force_env_section, "PRINT%DISTRIBUTION"), cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, force_env_section, "PRINT%DISTRIBUTION", &
                                            extension=".Log")

!     *** Print row distribution ***

         ALLOCATE (work(nprow), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 836, &
                             "work", nprow*int_size)
         END IF
         work(:) = 0

         IF (mypcol == 1) work(myprow) = SUM(distribution_2d%n_local_rows)

         CALL mp_sum(work, group)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/, T2, A, /, T15, A, /, (T16, I10, T41, I10, T71, I10))") &
               "DISTRIBUTION OF THE PARTICLES (ROWS)", &
               "Process row      Number of particles         Number of matrix rows", &
               (iprow-1, work(iprow), -1, iprow=1, nprow)
            WRITE (UNIT=output_unit, FMT="(T23, A3, T41, I10, T71, I10)") &
               "Sum", SUM(work), -1
            CALL m_flush(output_unit)
         END IF

         DEALLOCATE (work)

!     *** Print column distribution ***

         ALLOCATE (work(npcol), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 862, &
                             "work", npcol*int_size)
         END IF
         work(:) = 0

         IF (myprow == 1) work(mypcol) = SUM(distribution_2d%n_local_cols)

         CALL mp_sum(work, group)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/, T2, A, /, T15, A, /, (T16, I10, T41, I10, T71, I10))") &
               "DISTRIBUTION OF THE PARTICLES (COLUMNS)", &
               "Process col      Number of particles      Number of matrix columns", &
               (ipcol-1, work(ipcol), -1, ipcol=1, npcol)
            WRITE (UNIT=output_unit, FMT="(T23, A3, T41, I10, T71, I10)") &
               "Sum", SUM(work), -1
            CALL m_flush(output_unit)
         END IF

         DEALLOCATE (work)

         CALL cp_print_key_finished_output(output_unit, logger, force_env_section, &
                                           "PRINT%DISTRIBUTION")
      END IF

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           force_env_section, "PRINT%DISTRIBUTION2D"), cp_p_file)) THEN

         iw = cp_logger_get_default_unit_nr(logger, LOCAL=.TRUE.)
         CALL distribution_2d_write(distribution_2d, &
                                    unit_nr=iw, &
                                    local=.TRUE., &
                                    long_description=.TRUE.)

      ENDIF

!   *** Release work storage ***

      DEALLOCATE (nparticle_local_row)

      DEALLOCATE (nparticle_local_col)

      CALL timestop(handle)

   END SUBROUTINE distribute_molecules_2d

! **************************************************************************************************
!> \brief Creates a basic distribution
!> \param cluster_list ...
!> \param cluster_prices ...
!> \param nprows ...
!> \param row_distribution ...
!> \param npcols ...
!> \param col_distribution ...
!> \par History
!> - Created 2010-08-06 UB
! **************************************************************************************************
   SUBROUTINE make_basic_distribution(cluster_list, cluster_prices, &
                                      nprows, row_distribution, npcols, col_distribution)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: cluster_list, cluster_prices
      INTEGER, INTENT(IN)                                :: nprows
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: row_distribution
      INTEGER, INTENT(IN)                                :: npcols
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: col_distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'make_basic_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, bin_price, cluster, cluster_index, &
                                                            cluster_price, nbins, nclusters, pcol, &
                                                            pgrid_gcd, prow, timing_handle
      LOGICAL                                            :: found, heap_error
      TYPE(heap_t)                                       :: bin_heap

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, timing_handle)
      nbins = lcm(nprows, npcols)
      pgrid_gcd = gcd(nprows, npcols)
      CALL sort(cluster_prices, SIZE(cluster_list), cluster_list)
      CALL heap_new(bin_heap, nbins)
      CALL heap_fill(bin_heap, &
                     (/(bin, bin=0, nbins-1)/), (/(0, bin=1, nbins)/), heap_error)
      IF (heap_error) &
         CALL cp__b("distribution_methods.F",947,"Error initially filling the heap.")
      !
      nclusters = SIZE(cluster_list)
      ! Put the most expensive cluster in the bin with the smallest
      ! price and repeat.
      DO cluster_index = nclusters, 1, -1
         cluster = cluster_list(cluster_index)
         CALL heap_get_first(bin_heap, bin, bin_price, found, heap_error)
         IF (heap_error) &
            CALL cp__b("distribution_methods.F",956,"Error getting topmost heap element.")
         IF (.NOT. found) &
            CALL cp__b("distribution_methods.F",958,"No topmost heap element found.")
         !
         prow = INT(bin*pgrid_gcd/npcols)
         IF (prow .GE. nprows) &
            CALL cp__b("distribution_methods.F",962,"Invalid process row.")
         pcol = INT(bin*pgrid_gcd/nprows)
         IF (pcol .GE. npcols) &
            CALL cp__b("distribution_methods.F",965,"Invalid process column.")
         row_distribution(cluster) = prow+1
         col_distribution(cluster) = pcol+1
         !
         cluster_price = cluster_prices(cluster_index)
         bin_price = bin_price+cluster_price
         CALL heap_reset_first(bin_heap, bin_price, heap_error)
         IF (heap_error) &
            CALL cp__w("distribution_methods.F",973,"Error setting price of top heap element.")
      ENDDO
      CALL heap_release(bin_heap)
      CALL timestop(timing_handle)
   END SUBROUTINE make_basic_distribution

! **************************************************************************************************
!> \brief Creates a basic spatial distribution
!>        that tries to make the corresponding blocks as homogeneous as possible
!> \param pbc_scaled_coords ...
!> \param costs ...
!> \param nprows ...
!> \param row_distribution ...
!> \param npcols ...
!> \param col_distribution ...
!> \par History
!> - Created 2010-11-11 Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE make_basic_spatial_distribution(pbc_scaled_coords, costs, &
                                              nprows, row_distribution, npcols, col_distribution)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: pbc_scaled_coords
      INTEGER, DIMENSION(:), INTENT(IN)                  :: costs
      INTEGER, INTENT(IN)                                :: nprows
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: row_distribution
      INTEGER, INTENT(IN)                                :: npcols
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: col_distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'make_basic_spatial_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, natoms, nbins, pgrid_gcd
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bin_costs, distribution

      CALL timeset(routineN, handle)

      natoms = SIZE(costs)
      nbins = lcm(nprows, npcols)
      pgrid_gcd = gcd(nprows, npcols)
      ALLOCATE (bin_costs(nbins), distribution(natoms))
      bin_costs = 0

      CALL spatial_recurse(pbc_scaled_coords, costs, (/(iatom, iatom=1, natoms)/), bin_costs, distribution, 0)

      ! WRITE(*, *) "Final bin costs: ", bin_costs

      ! final row_distribution / col_distribution
      DO iatom = 1, natoms
         row_distribution(iatom) = (distribution(iatom)-1)*pgrid_gcd/npcols+1
         col_distribution(iatom) = (distribution(iatom)-1)*pgrid_gcd/nprows+1
      ENDDO

      DEALLOCATE (bin_costs, distribution)

      CALL timestop(handle)

   END SUBROUTINE make_basic_spatial_distribution

! **************************************************************************************************
!> \brief ...
!> \param pbc_scaled_coords ...
!> \param costs ...
!> \param indices ...
!> \param bin_costs ...
!> \param distribution ...
!> \param level ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE spatial_recurse(pbc_scaled_coords, costs, indices, bin_costs, distribution, level)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: pbc_scaled_coords
      INTEGER, DIMENSION(:), INTENT(IN)                  :: costs, indices
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: bin_costs, distribution
      INTEGER, INTENT(IN)                                :: level

      INTEGER                                            :: iatom, ibin, natoms, nbins, nhalf
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_costs_sorted, atom_permutation, &
                                                            bin_costs_sorted, permutation
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: coord

      natoms = SIZE(costs)
      nbins = SIZE(bin_costs)
      nhalf = (natoms+1)/2

      IF (natoms <= nbins) THEN
         ! assign the most expensive atom to the least costly bin
         ALLOCATE (bin_costs_sorted(nbins), permutation(nbins))
         bin_costs_sorted(:) = bin_costs
         CALL sort(bin_costs_sorted, nbins, permutation)
         ALLOCATE (atom_costs_sorted(natoms), atom_permutation(natoms))
         atom_costs_sorted(:) = costs
         CALL sort(atom_costs_sorted, natoms, atom_permutation)
         ibin = 0
         ! WRITE(*, *) "Dealing with a new bunch of atoms "
         DO iatom = natoms, 1, -1
            ibin = ibin+1
            ! WRITE(*, *) "atom", indices(atom_permutation(iatom)), "cost", atom_costs_sorted(iatom), &
            !            "bin", permutation(ibin), "its cost", bin_costs(permutation(ibin))
            ! WRITE(100, '(A, I0, 3F12.6)') "A", permutation(ibin), pbc_scaled_coords(:, atom_permutation(iatom))
            bin_costs(permutation(ibin)) = bin_costs(permutation(ibin))+atom_costs_sorted(iatom)
            distribution(indices(atom_permutation(iatom))) = permutation(ibin)
         ENDDO
         DEALLOCATE (bin_costs_sorted, permutation, atom_costs_sorted, atom_permutation)
      ELSE
         ! divide atoms in two subsets, sorting according to their coordinates, alternatively x, y, z
         ! recursively do this for both subsets
         ALLOCATE (coord(natoms), permutation(natoms))
         coord(:) = pbc_scaled_coords(MOD(level, 3)+1, :)
         CALL sort(coord, natoms, permutation)
         CALL spatial_recurse(pbc_scaled_coords(:, permutation(1:nhalf)), costs(permutation(1:nhalf)), &
                              indices(permutation(1:nhalf)), bin_costs, distribution, level+1)
         CALL spatial_recurse(pbc_scaled_coords(:, permutation(nhalf+1:)), costs(permutation(nhalf+1:)), &
                              indices(permutation(nhalf+1:)), bin_costs, distribution, level+1)
         DEALLOCATE (coord, permutation)
      ENDIF

   END SUBROUTINE spatial_recurse

! **************************************************************************************************
!> \brief creates a distribution placing close by atoms into clusters and
!>        putting them on the smae processors. Load balancing is
!>        performed by balancing sum of the cluster costs per processor
!> \param coords coordinates of the system
!> \param scaled_coords scaled coordinates
!> \param cell the cell_type
!> \param costs costs per atomic block
!> \param nprows number of precessors per row on the 2d grid
!> \param row_distribution the resulting distribution over proc_rows of atomic blocks
!> \param npcols number of precessors per col on the 2d grid
!> \param col_distribution the resulting distribution over proc_cols of atomic blocks
! **************************************************************************************************
   SUBROUTINE make_cluster_distribution(coords, scaled_coords, cell, costs, &
                                        nprows, row_distribution, npcols, col_distribution)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: coords, scaled_coords
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(:), INTENT(IN)                  :: costs
      INTEGER, INTENT(IN)                                :: nprows
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: row_distribution
      INTEGER, INTENT(IN)                                :: npcols
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: col_distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'make_cluster_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, icluster, level, natom, &
                                                            output_unit
      INTEGER(KIND=int_8)                                :: ncluster
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_to_cluster, cluster_cost, &
                                                            cluster_count, cluster_to_col, &
                                                            cluster_to_row, piv_cost, proc_cost, &
                                                            sorted_cost
      REAL(KIND=dp)                                      :: fold(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cluster_center, cluster_high, cluster_low

      CALL timeset(routineN, handle)

      output_unit = cp_logger_get_default_io_unit()

      natom = SIZE(costs)
      ncluster = cp_distribution_get_num_images(SUM(costs), natom, nprows, npcols)
      ALLOCATE (atom_to_cluster(natom))
      ALLOCATE (cluster_cost(ncluster))
      ALLOCATE (cluster_to_row(ncluster))
      ALLOCATE (cluster_to_col(ncluster))
      ALLOCATE (sorted_cost(ncluster))
      ALLOCATE (piv_cost(ncluster))
      cluster_cost(:) = 0

      icluster = 0
      CALL cluster_recurse(coords, scaled_coords, cell, costs, atom_to_cluster, ncluster, icluster, cluster_cost)

      sorted_cost(:) = cluster_cost(:)
      CALL sort(sorted_cost, INT(ncluster), piv_cost)

      ALLOCATE (proc_cost(nprows))
      proc_cost = 0; level = 1
      CALL assign_clusters(cluster_cost, piv_cost, proc_cost, cluster_to_row, nprows)

      DEALLOCATE (proc_cost); ALLOCATE (proc_cost(npcols))
      proc_cost = 0; level = 1
      CALL assign_clusters(cluster_cost, piv_cost, proc_cost, cluster_to_col, npcols)

      DO i = 1, natom
         row_distribution(i, 1) = cluster_to_row(atom_to_cluster(i))
         row_distribution(i, 2) = atom_to_cluster(i)
         col_distribution(i, 1) = cluster_to_col(atom_to_cluster(i))
         col_distribution(i, 2) = atom_to_cluster(i)
      END DO

      ! generate some statistics on clusters
      ALLOCATE (cluster_center(3, ncluster))
      ALLOCATE (cluster_low(3, ncluster))
      ALLOCATE (cluster_high(3, ncluster))
      ALLOCATE (cluster_count(ncluster))
      cluster_count = 0
      DO i = 1, natom
         cluster_count(atom_to_cluster(i)) = cluster_count(atom_to_cluster(i))+1
         cluster_center(:, atom_to_cluster(i)) = coords(:, i)
      ENDDO
      cluster_low = HUGE(0.0_dp)
      cluster_high = -HUGE(0.0_dp)
      DO i = 1, natom
         fold = pbc(coords(:, i)-cluster_center(:, atom_to_cluster(i)), cell)+cluster_center(:, atom_to_cluster(i))
         cluster_low(:, atom_to_cluster(i)) = MIN(cluster_low(:, atom_to_cluster(i)), fold(:))
         cluster_high(:, atom_to_cluster(i)) = MAX(cluster_high(:, atom_to_cluster(i)), fold(:))
      ENDDO
      IF (output_unit > 0) THEN
         WRITE (output_unit, *)
         WRITE (output_unit, '(T2,A)') "Cluster distribution information"
         WRITE (output_unit, '(T2,A,T48,I8)') "Number of atoms", natom
         WRITE (output_unit, '(T2,A,T48,I8)') "Number of clusters", ncluster
         WRITE (output_unit, '(T2,A,T48,I8)') "Largest cluster in atoms", MAXVAL(cluster_count)
         WRITE (output_unit, '(T2,A,T48,I8)') "Smallest cluster in atoms", MINVAL(cluster_count)
         WRITE (output_unit, '(T2,A,T48,F8.3,I8)') "Largest cartesian extend [a.u.]/cluster x=", &
            MAXVAL(cluster_high(1, :)-cluster_low(1, :)), MAXLOC(cluster_high(1, :)-cluster_low(1, :))
         WRITE (output_unit, '(T2,A,T48,F8.3,I8)') "Largest cartesian extend [a.u.]/cluster y=", &
            MAXVAL(cluster_high(2, :)-cluster_low(2, :)), MAXLOC(cluster_high(2, :)-cluster_low(2, :))
         WRITE (output_unit, '(T2,A,T48,F8.3,I8)') "Largest cartesian extend [a.u.]/cluster z=", &
            MAXVAL(cluster_high(3, :)-cluster_low(3, :)), MAXLOC(cluster_high(3, :)-cluster_low(3, :))
      ENDIF

      DEALLOCATE (atom_to_cluster, cluster_cost, cluster_to_row, cluster_to_col, sorted_cost, piv_cost, proc_cost)
      CALL timestop(handle)

   END SUBROUTINE make_cluster_distribution

! **************************************************************************************************
!> \brief assigns the clusters to processors, tryimg to balance the cost on the nodes
!> \param cluster_cost vector with the cost of each cluster
!> \param piv_cost pivoting vector sorting the cluster_cost
!> \param proc_cost cost per processor, on input 0 everywhere
!> \param cluster_assign assgnment of clusters on proc
!> \param nproc number of processor over which clusters are distributed
! **************************************************************************************************
   SUBROUTINE assign_clusters(cluster_cost, piv_cost, proc_cost, cluster_assign, nproc)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cluster_cost, piv_cost, proc_cost, &
                                                            cluster_assign
      INTEGER                                            :: nproc

      CHARACTER(len=*), PARAMETER :: routineN = 'assign_clusters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ilevel, offset, &
                                                            piv_pcost(nproc), sort_proc_cost(nproc)

      CALL timeset(routineN, handle)

      DO ilevel = 1, SIZE(cluster_cost)/nproc
         sort_proc_cost(:) = proc_cost(:)
         CALL sort(sort_proc_cost, nproc, piv_pcost)

         offset = (SIZE(cluster_cost)/nproc-ilevel+1)*nproc+1
         DO i = 1, nproc
            cluster_assign(piv_cost(offset-i)) = piv_pcost(i)
            proc_cost(piv_pcost(i)) = proc_cost(piv_pcost(i))+cluster_cost(piv_cost(offset-i))
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE assign_clusters

! **************************************************************************************************
!> \brief recursive routine to cluster atoms.
!>        Low level uses a modified KMEANS algorithm
!>        recursion is used to reduce cost.
!>        each level will subdivide a cluster into smaller clusters
!>        If only a single split is necessary atoms are assigned to the current cluster
!> \param coord coordinates of the system
!> \param scaled_coord scaled coordinates
!> \param cell the cell_type
!> \param costs costs per atomic block
!> \param cluster_inds the atom_to cluster mapping
!> \param ncluster number of clusters still to be created on a given recursion level
!> \param icluster the index of the current cluster to be created
!> \param fin_cluster_cost total cost of the final clusters
! **************************************************************************************************
   RECURSIVE SUBROUTINE cluster_recurse(coord, scaled_coord, cell, costs, cluster_inds, ncluster, icluster, fin_cluster_cost)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: coord, scaled_coord
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(:), INTENT(IN)                  :: costs
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: cluster_inds
      INTEGER(KIND=int_8), INTENT(INOUT)                 :: ncluster
      INTEGER, INTENT(INOUT)                             :: icluster
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: fin_cluster_cost

      CHARACTER(len=*), PARAMETER :: routineN = 'cluster_recurse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ibeg, iend, maxv(1), min_seed, &
                                                            natoms, nleft, nsplits, seed, tot_cost
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: ncluster_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cluster_cost, inds_tmp, nat_cluster, piv
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: balance, balance_new, conv

      natoms = SIZE(coord, 2)
      ! This is a bit of an arbitrary choice, simply a try to avoid too many clusters on large systems and too few for balancing on
      ! small systems or subclusters
      IF (natoms .LE. 1) THEN
         nsplits = 1
      ELSE
         nsplits = MIN(INT(MIN(INT(MAX(6, INT(60.00/LOG(REAL(natoms)))), KIND=int_8), ncluster)), natoms)
      ENDIF
      IF (nsplits == 1) THEN
         icluster = icluster+1
         cluster_inds = icluster
         fin_cluster_cost(icluster) = SUM(costs)
      ELSE
         ALLOCATE (cluster_cost(nsplits), ncluster_new(nsplits), inds_tmp(natoms), piv(natoms), nat_cluster(nsplits))
         ! initialise some values
         cluster_cost = 0; seed = 300; found = .TRUE.; min_seed = seed
         CALL kmeans(nsplits, coord, scaled_coord, cell, cluster_inds, nat_cluster, seed, conv)
         balance = MAXVAL(REAL(nat_cluster))/MINVAL(REAL(nat_cluster))

         ! If the system is small enough try to do better in terms of balancing number of atoms per cluster
         ! by changing the seed for the initial guess
         IF (natoms .LT. 1000 .AND. balance .GT. 1.1) THEN
            found = .FALSE.
            DO i = 1, 5
               IF (balance .GT. 1.1) THEN
                  CALL kmeans(nsplits, coord, scaled_coord, cell, cluster_inds, nat_cluster, seed+i*40, conv)
                  balance_new = MAXVAL(REAL(nat_cluster))/MINVAL(REAL(nat_cluster))
                  IF (balance_new .LT. balance) THEN
                     balance = balance_new
                     min_seed = seed+i*40; 
                  END IF
               ELSE
                  found = .TRUE.
                  EXIT
               END IF
            END DO
         END IF
         !If we do not match the convergence than recompute at least the best assignment
         IF (.NOT. found) CALL kmeans(nsplits, coord, scaled_coord, cell, cluster_inds, nat_cluster, min_seed, conv)

         ! compute the cost of each cluster to decide how many splits have to be performed on the next lower level
         DO i = 1, natoms
            cluster_cost(cluster_inds(i)) = cluster_cost(cluster_inds(i))+costs(i)
         END DO
         tot_cost = SUM(cluster_cost)
         ! compute new splitting, can be done more elegant
         ncluster_new(:) = ncluster*cluster_cost(:)/tot_cost
         nleft = INT(ncluster-SUM(ncluster_new))
         ! As we won't have empty clusters, we can not have 0 as new size, so we correct for this at first
         DO i = 1, nsplits
            IF (ncluster_new(i) == 0) THEN
               ncluster_new(i) = 1
               nleft = nleft-1
            END IF
         END DO
         ! now comes the next part that the number of clusters will not match anymore, so try to correct in a meaningful way without
         ! introducing 0 sized blocks again
         IF (nleft .NE. 0) THEN
            DO i = 1, ABS(nleft)
               IF (nleft < 0) THEN
                  maxv = MINLOC(cluster_cost/ncluster_new)
                  IF (ncluster_new(maxv(1)) .NE. 1) THEN
                     ncluster_new(maxv) = ncluster_new(maxv)-1
                  ELSE
                     maxv = MAXLOC(ncluster_new)
                     ncluster_new(maxv) = ncluster_new(maxv)-1
                  END IF
               ELSE
                  maxv = MAXLOC(cluster_cost/ncluster_new)
                  ncluster_new(maxv) = ncluster_new(maxv)+1
               END IF
            END DO
         END IF

         !Now get the permutations to sort the atoms in the nsplits clusters for the next level of iteration
         inds_tmp(:) = cluster_inds(:)
         CALL sort(inds_tmp, natoms, piv)

         ibeg = 1; iend = 0
         DO i = 1, nsplits
            IF (nat_cluster(i) == 0) CYCLE
            iend = iend+nat_cluster(i)
            CALL cluster_recurse(coord(:, piv(ibeg:iend)), scaled_coord(:, piv(ibeg:iend)), cell, costs(piv(ibeg:iend)), &
                                 inds_tmp(ibeg:iend), ncluster_new(i), icluster, fin_cluster_cost)
            ibeg = ibeg+nat_cluster(i)
         END DO
         ! copy the sorted cluster IDs on the old layout, inds_tmp gets set at the lowest level of recursion
         cluster_inds(piv(:)) = inds_tmp
         DEALLOCATE (cluster_cost, ncluster_new, inds_tmp, piv, nat_cluster)

      END IF

   END SUBROUTINE cluster_recurse

! **************************************************************************************************
!> \brief A modified version of the kmeans algorithm.
!>        The assignment has a penalty function in case clusters become
!>        larger than average. Like this more even sized clusters are created
!>        trading it for locality
!> \param ncent number of centers to be created
!> \param coord coordinates
!> \param scaled_coord scaled coord
!> \param cell the cell_type
!> \param cluster atom to cluster assignment
!> \param nat_cl atoms per cluster
!> \param seed seed for the RNG. Algorithm might need multiple tries to deliver best results
!> \param tot_var the total variance of the clusters around the centers
! **************************************************************************************************
   SUBROUTINE kmeans(ncent, coord, scaled_coord, cell, cluster, nat_cl, seed, tot_var)
      INTEGER                                            :: ncent
      REAL(KIND=dp), DIMENSION(:, :)                     :: coord, scaled_coord
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(:)                              :: cluster, nat_cl
      INTEGER                                            :: seed
      REAL(KIND=dp)                                      :: tot_var

      CHARACTER(len=*), PARAMETER :: routineN = 'kmeans', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ind, itn, j, nat, oldc
      LOGICAL                                            :: changed
      REAL(KIND=dp) :: average(3, ncent, 2), cent_coord(3, ncent), devi, deviat(ncent), dist, &
         dvec(3), old_var, rn, scaled_cent(3, ncent), var_cl(ncent)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dmat
      REAL(KIND=dp), DIMENSION(3, 2)                     :: initial_seed
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CALL timeset(routineN, handle)

      initial_seed = REAL(seed, dp); nat = SIZE(coord, 2)
      NULLIFY (rng_stream)
      ALLOCATE (dmat(ncent, nat))

      CALL create_rng_stream(rng_stream=rng_stream, &
                             name="kmeans uniform distribution [0,1]", &
                             distribution_type=UNIFORM, seed=initial_seed)

! try to find a clever initial guess with centers being somewhat distributed
      rn = next_random_number(rng_stream)
      ind = CEILING(rn*nat)
      cent_coord(:, 1) = coord(:, ind)
      DO i = 2, ncent
         DO
            rn = next_random_number(rng_stream)
            ind = CEILING(rn*nat)
            cent_coord(:, i) = coord(:, ind)
            devi = HUGE(1.0_dp)
            DO j = 1, i-1
               dvec = pbc(cent_coord(:, j), cent_coord(:, i), cell)
               dist = SQRT(DOT_PRODUCT(dvec, dvec))
               IF (dist .LT. devi) devi = dist
            END DO
            rn = next_random_number(rng_stream)
            IF (rn .LT. devi**2/169.0) EXIT
         END DO
      END DO

! Now start the KMEANS but penalise it in case it starts packing too many atoms into a single set
! Unfoirtunatelz as this is dependent on what happend before it cant be parallel
      cluster = 0; old_var = HUGE(1.0_dp)
      DO itn = 1, 1000
         changed = .FALSE.; var_cl = 0.0_dp; tot_var = 0.0_dp; nat_cl = 0; deviat = 0.0_dp
!      !$OMP PARALLEL DO PRIVATE(i,j,dvec)
         DO i = 1, nat
            DO j = 1, ncent
               dvec = pbc(cent_coord(:, j), coord(:, i), cell)
               dmat(j, i) = DOT_PRODUCT(dvec, dvec)
            END DO
         END DO
         DO i = 1, nat
            devi = HUGE(1.0_dp); oldc = cluster(i)
            DO j = 1, ncent
               dist = dmat(j, i)+MAX(nat_cl(j)**2/nat*ncent, nat/ncent)
               IF (dist .LT. devi) THEN
                  devi = dist; cluster(i) = j
               END IF
            END DO
            deviat(cluster(i)) = deviat(cluster(i))+SQRT(devi)
            nat_cl(cluster(i)) = nat_cl(cluster(i))+1
            tot_var = tot_var+devi
            IF (oldc .NE. cluster(i)) changed = .TRUE.
         END DO
         ! get the update of the centers done, add a new one in case one center lost all its atoms
         ! the algorithm would survive, but its nice to really create what you demand
         IF (tot_var .GE. old_var) EXIT
         IF (changed) THEN
            ! Here misery of computing the center of geometry of the clusters in PBC.
            ! The mapping on the unit circle allows to circumvent all problems
            average = 0.0_dp
            DO i = 1, SIZE(coord, 2)
               average(:, cluster(i), 1) = average(:, cluster(i), 1)+COS(scaled_coord(:, i)*2.0_dp*pi)
               average(:, cluster(i), 2) = average(:, cluster(i), 2)+SIN(scaled_coord(:, i)*2.0_dp*pi)
            END DO

            DO i = 1, ncent
               IF (nat_cl(i) == 0) THEN
                  rn = next_random_number(rng_stream)
                  scaled_cent(:, i) = scaled_coord(:, CEILING(rn*nat))
               ELSE
                  average(:, i, 1) = average(:, i, 1)/REAL(nat_cl(i), dp)
                  average(:, i, 2) = average(:, i, 2)/REAL(nat_cl(i), dp)
                  scaled_cent(:, i) = (ATAN2(-average(:, i, 2), -average(:, i, 1))+pi)/(2.0_dp*pi)
                  CALL scaled_to_real(cent_coord(:, i), scaled_cent(:, i), cell)
               END IF
            END DO
         ELSE
            EXIT
         END IF
      END DO

      CALL delete_rng_stream(rng_stream)

      CALL timestop(handle)

   END SUBROUTINE kmeans

END MODULE distribution_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH [04042007] code refactoring
! **************************************************************************************************
MODULE virial_methods

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC:: virial_evaluate, virial_pair_force, virial_update

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'virial_methods'

CONTAINS
! **************************************************************************************************
!> \brief Updates the virial given the virial and subsys
!> \param virial ...
!> \param subsys ...
!> \param para_env ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - 03.2008 - Zurich University
! **************************************************************************************************
   SUBROUTINE virial_update(virial, subsys, para_env)
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'virial_update', routineP = moduleN//':'//routineN

      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_list_type), POINTER                  :: particles

      CALL cp_subsys_get(subsys, local_particles=local_particles, atomic_kinds=atomic_kinds, &
                         particles=particles)

      CALL virial_evaluate(atomic_kinds%els, particles%els, local_particles, &
                           virial, para_env%group)

   END SUBROUTINE virial_update

! **************************************************************************************************
!> \brief Computes the kinetic part of the pressure tensor and updates
!>      the full VIRIAL (PV)
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param virial ...
!> \param igroup ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE virial_evaluate(atomic_kind_set, particle_set, local_particles, &
                              virial, igroup)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      INTEGER, INTENT(IN)                                :: igroup

      CHARACTER(LEN=*), PARAMETER :: routineN = 'virial_evaluate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iparticle, iparticle_kind, &
                                                            iparticle_local, j, nparticle_kind, &
                                                            nparticle_local
      REAL(KIND=dp)                                      :: mass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      IF (virial%pv_availability) THEN
         CALL timeset(routineN, handle)
         NULLIFY (atomic_kind)
         nparticle_kind = SIZE(atomic_kind_set)
         virial%pv_kinetic = 0.0_dp
         DO i = 1, 3
            DO j = 1, i
               DO iparticle_kind = 1, nparticle_kind
                  atomic_kind => atomic_kind_set(iparticle_kind)
                  CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
                  nparticle_local = local_particles%n_el(iparticle_kind)
                  DO iparticle_local = 1, nparticle_local
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                     virial%pv_kinetic(i, j) = virial%pv_kinetic(i, j)+ &
                                               mass*particle_set(iparticle)%v(i)*particle_set(iparticle)%v(j)
                  END DO
               END DO
               virial%pv_kinetic(j, i) = virial%pv_kinetic(i, j)
            END DO
         END DO

         CALL mp_sum(virial%pv_kinetic, igroup)

         ! total virial
         virial%pv_total = virial%pv_virial+virial%pv_kinetic+virial%pv_constraint

         CALL timestop(handle)
      ENDIF

   END SUBROUTINE virial_evaluate

! **************************************************************************************************
!> \brief Computes the contribution to the stress tensor from two-body
!>      pair-wise forces
!> \param pv_virial ...
!> \param f0 ...
!> \param force ...
!> \param rab ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   PURE SUBROUTINE virial_pair_force(pv_virial, f0, force, rab)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: pv_virial
      REAL(KIND=dp), INTENT(IN)                          :: f0
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: force, rab

      CHARACTER(LEN=*), PARAMETER :: routineN = 'virial_pair_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      DO i = 1, 3
         DO j = 1, 3
            pv_virial(i, j) = pv_virial(i, j)+f0*force(i)*rab(j)
         END DO
      END DO

   END SUBROUTINE virial_pair_force

END MODULE virial_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of dispersion using pair potentials
!> \author JGH
! **************************************************************************************************
MODULE qs_dispersion_pairpot
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE bibliography,                    ONLY: cite_reference,&
                                              grimme2006,&
                                              grimme2010,&
                                              grimme2011
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc,&
                                              plane_distance
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_constants,                 ONLY: vdw_pairpot_dftd2,&
                                              vdw_pairpot_dftd3,&
                                              vdw_pairpot_dftd3bj,&
                                              xc_vdw_fun_none,&
                                              xc_vdw_fun_pairpot
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_max,&
                                              mp_shift,&
                                              mp_sum
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: bohr,&
                                              kcalmol,&
                                              kjmol
   USE qs_dispersion_types,             ONLY: dftd2_pp,&
                                              dftd3_pp,&
                                              qs_atom_dispersion_type,&
                                              qs_dispersion_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type,&
                                              set_qs_kind
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dispersion_pairpot'

   PUBLIC :: qs_dispersion_pairpot_init, calculate_dispersion_pairpot, &
             qs_scaling_init, qs_scaling_dftd3, qs_scaling_dftd3bj

   TYPE dcnum_type
      INTEGER                                :: neighbors
      INTEGER, DIMENSION(:), POINTER         :: nlist
      REAL(KIND=dp), DIMENSION(:), POINTER   :: dvals
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: rik
   END TYPE dcnum_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param dispersion_env ...
!> \param pp_section ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE qs_dispersion_pairpot_init(atomic_kind_set, qs_kind_set, dispersion_env, pp_section, para_env)
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: pp_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dispersion_pairpot_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: symbol
      CHARACTER(LEN=default_string_length)               :: aname, filename
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: elem, handle, i, ikind, j, max_elem, &
                                                            maxc, n_rep, nkind, nl, vdw_pp_type, &
                                                            vdw_type
      LOGICAL                                            :: at_end, found, is_available
      REAL(KIND=dp)                                      :: dum
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(qs_atom_dispersion_type), POINTER             :: disp

      CALL timeset(routineN, handle)

      vdw_type = dispersion_env%type
      SELECT CASE (vdw_type)
      CASE DEFAULT
         ! do nothing
      CASE (xc_vdw_fun_pairpot)
         ! setup information on pair potentials
         vdw_pp_type = dispersion_env%type
         SELECT CASE (dispersion_env%pp_type)
         CASE DEFAULT
            ! do nothing
         CASE (vdw_pairpot_dftd3, vdw_pairpot_dftd3bj)
            !DFT-D3 Method initial setup
            CALL cite_reference(Grimme2010)
            CALL cite_reference(Grimme2011)
            max_elem = 94
            maxc = 5
            dispersion_env%max_elem = max_elem
            dispersion_env%maxc = maxc
            ALLOCATE (dispersion_env%maxci(max_elem))
            ALLOCATE (dispersion_env%c6ab(max_elem, max_elem, maxc, maxc, 3))
            ALLOCATE (dispersion_env%r0ab(max_elem, max_elem))
            ALLOCATE (dispersion_env%rcov(max_elem))
            ALLOCATE (dispersion_env%r2r4(max_elem))
            ALLOCATE (dispersion_env%cn(max_elem))

            ! get filename of parameter file
            filename = dispersion_env%parameter_file_name
            CALL dftd3_c6_param(dispersion_env%c6ab, dispersion_env%maxci, filename, para_env)
            CALL setr0ab(dispersion_env%r0ab, dispersion_env%rcov, dispersion_env%r2r4)
            ! the default coordination numbers
            CALL setcn(dispersion_env%cn)
            ! scale r4/r2 values of the atoms by sqrt(Z)
            ! sqrt is also globally close to optimum
            ! together with the factor 1/2 this yield reasonable
            ! c8 for he, ne and ar. for larger Z, C8 becomes too large
            ! which effectively mimics higher R^n terms neglected due
            ! to stability reasons
            DO i = 1, max_elem
               dum = 0.5_dp*dispersion_env%r2r4(i)*REAL(i, dp)**0.5_dp
               ! store it as sqrt because the geom. av. is taken
               dispersion_env%r2r4(i) = SQRT(dum)
            ENDDO
            ! parameters
            dispersion_env%k1 = 16.0_dp
            dispersion_env%k2 = 4._dp/3._dp
            ! reasonable choices are between 3 and 5
            ! this gives smoth curves with maxima around the integer values
            ! k3=3 give for CN=0 a slightly smaller value than computed
            ! for the free atom. This also yields to larger CN for atoms
            ! in larger molecules but with the same chem. environment
            ! which is physically not right
            ! values >5 might lead to bumps in the potential
            dispersion_env%k3 = -4._dp
            dispersion_env%rcov = dispersion_env%k2*dispersion_env%rcov*bohr
            ! alpha default parameter
            dispersion_env%alp = 14._dp
         END SELECT
      END SELECT

      nkind = SIZE(atomic_kind_set)
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), element_symbol=symbol, z=elem)
         ALLOCATE (disp)
         SELECT CASE (vdw_type)
         CASE DEFAULT
            CALL cp__b("qs_dispersion_pairpot.F",192,"")
         CASE (xc_vdw_fun_none)
            disp%defined = .FALSE.
         CASE (xc_vdw_fun_pairpot)
            ! setup information on pair potentials
            SELECT CASE (dispersion_env%pp_type)
            CASE DEFAULT
               CALL cp__b("qs_dispersion_pairpot.F",199,"")
            CASE (vdw_pairpot_dftd2)
               CALL cite_reference(Grimme2006)
               disp%type = dftd2_pp
               ! get filename of parameter file
               filename = dispersion_env%parameter_file_name
               ! check for local parameters
               found = .FALSE.
               IF (PRESENT(pp_section)) THEN
                  CALL section_vals_val_get(pp_section, "ATOMPARM", n_rep_val=n_rep)
                  DO i = 1, n_rep
                     CALL section_vals_val_get(pp_section, "ATOMPARM", i_rep_val=i, &
                                               c_vals=tmpstringlist)
                     IF (TRIM(tmpstringlist(1)) == TRIM(symbol)) THEN
                        ! we assume the parameters are in atomic units!
                        READ (tmpstringlist(2), *) disp%c6
                        READ (tmpstringlist(3), *) disp%vdw_radii
                        found = .TRUE.
                        EXIT
                     END IF
                  END DO
               END IF
               IF (.NOT. found) THEN
                  ! check for internal parameters
                  CALL dftd2_param(elem, disp%c6, disp%vdw_radii, found)
               END IF
               IF (.NOT. found) THEN
                  ! check on file
                  INQUIRE (FILE=filename, EXIST=is_available)
                  IF (is_available) THEN
                     NULLIFY (parser)
                     CALL parser_create(parser, filename, para_env=para_env)
                     DO
                        at_end = .FALSE.
                        CALL parser_get_next_line(parser, 1, at_end)
                        IF (at_end) EXIT
                        CALL parser_get_object(parser, aname)
                        IF (TRIM(aname) == TRIM(symbol)) THEN
                           CALL parser_get_object(parser, disp%c6)
                           ! we have to change the units J*nm^6*mol^-1 -> Hartree*Bohr^6
                           disp%c6 = disp%c6*1000._dp*bohr**6/kjmol
                           CALL parser_get_object(parser, disp%vdw_radii)
                           disp%vdw_radii = disp%vdw_radii*bohr
                           found = .TRUE.
                           EXIT
                        END IF
                     END DO
                     CALL parser_release(parser)
                  END IF
               END IF
               IF (found) THEN
                  disp%defined = .TRUE.
               ELSE
                  disp%defined = .FALSE.
               END IF
            CASE (vdw_pairpot_dftd3, vdw_pairpot_dftd3bj)
               disp%type = dftd3_pp
               IF (elem <= 94) THEN
                  disp%defined = .TRUE.
               ELSE
                  disp%defined = .FALSE.
               END IF

               ! Check for coordination numbers
               IF (PRESENT(pp_section)) THEN
                  CALL section_vals_val_get(pp_section, "KIND_COORDINATION_NUMBERS", n_rep_val=n_rep)
                  IF (n_rep > 0) THEN
                     ALLOCATE (dispersion_env%cnkind(n_rep))
                     DO i = 1, n_rep
                        CALL section_vals_val_get(pp_section, "KIND_COORDINATION_NUMBERS", i_rep_val=i, &
                                                  c_vals=tmpstringlist)
                        READ (tmpstringlist(1), *) dispersion_env%cnkind(i)%cnum
                        READ (tmpstringlist(2), *) dispersion_env%cnkind(i)%kind
                     END DO
                  END IF
                  CALL section_vals_val_get(pp_section, "ATOM_COORDINATION_NUMBERS", n_rep_val=n_rep)
                  IF (n_rep > 0) THEN
                     ALLOCATE (dispersion_env%cnlist(n_rep))
                     DO i = 1, n_rep
                        CALL section_vals_val_get(pp_section, "ATOM_COORDINATION_NUMBERS", i_rep_val=i, &
                                                  c_vals=tmpstringlist)
                        nl = SIZE(tmpstringlist)
                        ALLOCATE (dispersion_env%cnlist(i)%atom(nl-1))
                        dispersion_env%cnlist(i)%natom = nl-1
                        READ (tmpstringlist(1), *) dispersion_env%cnlist(i)%cnum
                        DO j = 1, nl-1
                           READ (tmpstringlist(j+1), *) dispersion_env%cnlist(i)%atom(j)
                        END DO
                     END DO
                  END IF
               END IF

            END SELECT
         END SELECT
         ! Check if the parameter is defined
         IF (.NOT. disp%defined) &
            CALL cp_abort(cp__l("qs_dispersion_pairpot.F",295), &
                          "Dispersion parameters for element ("//TRIM(symbol)//") are not defined! "// &
                          "Please provide a valid set of parameters through the input section or "// &
                          "through an external file! ")
         CALL set_qs_kind(qs_kind_set(ikind), dispersion=disp)
      END DO

      CALL timestop(handle)

   END SUBROUTINE qs_dispersion_pairpot_init

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param scaling ...
!> \param vdw_section ...
! **************************************************************************************************
   SUBROUTINE qs_scaling_init(scaling, vdw_section)
      REAL(KIND=dp), INTENT(inout)                       :: scaling
      TYPE(section_vals_type), POINTER                   :: vdw_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scaling_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: functional

      CALL section_vals_val_get(vdw_section, "PAIR_POTENTIAL%REFERENCE_FUNCTIONAL", c_val=functional)

      SELECT CASE (TRIM (functional))
      CASE DEFAULT
         ! unknown functional
         CALL cp__b("qs_dispersion_pairpot.F",327,"No DFT-D2 s6 value available for this functional:"//TRIM(functional))
      CASE ("BLYP")
         scaling = 1.20_dp
      CASE ("B3LYP")
         scaling = 1.05_dp
      CASE ("TPSS")
         scaling = 1.00_dp
      CASE ("PBE")
         scaling = 0.75_dp
      CASE ("PBE0")
         scaling = 0.6_dp
      CASE ("B2PLYP")
         scaling = 0.55_dp
      CASE ("BP86")
         scaling = 1.05_dp
      CASE ("B97")
         scaling = 1.25_dp
      END SELECT

   END SUBROUTINE qs_scaling_init

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param s6 ...
!> \param sr6 ...
!> \param s8 ...
!> \param vdw_section ...
! **************************************************************************************************
   SUBROUTINE qs_scaling_dftd3(s6, sr6, s8, vdw_section)
      REAL(KIND=dp), INTENT(inout)                       :: s6, sr6, s8
      TYPE(section_vals_type), POINTER                   :: vdw_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scaling_dftd3', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: functional

      CALL section_vals_val_get(vdw_section, "PAIR_POTENTIAL%REFERENCE_FUNCTIONAL", c_val=functional)

      ! values for different functionals from:
      ! http://www.thch.uni-bonn.de/tc/downloads/DFT-D3/functionals.html
      SELECT CASE (TRIM (functional))
      CASE DEFAULT
         ! unknown functional
         CALL cp__b("qs_dispersion_pairpot.F",373,"No DFT-D3 values available for this functional:"//TRIM(functional))
      CASE ("B1B95")
         s6 = 1.000_dp
         sr6 = 1.613_dp
         s8 = 1.868_dp
      CASE ("B2GPPLYP")
         ! L. Goerigk and S. Grimme
         ! J. Chem. Theory Comput. 2011, 7, 291-309; doi:10.1021/ct100466k
         s6 = 0.56_dp
         sr6 = 1.586_dp
         s8 = 0.760_dp
      CASE ("B2PLYP")
         ! L. Goerigk and S. Grimme
         ! J. Chem. Theory Comput. 2011, 7, 291-309; doi:10.1021/ct100466k
         s6 = 0.64_dp
         sr6 = 1.427_dp
         s8 = 1.022_dp
      CASE ("DSD-BLYP")
         ! L. Goerigk and S. Grimme
         ! J. Chem. Theory Comput. 2011, 7, 291-309; doi:10.1021/ct100466k
         s6 = 0.50_dp
         sr6 = 1.569_dp
         s8 = 0.705_dp
      CASE ("B3LYP")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.261_dp
         s8 = 1.703_dp
      CASE ("B97-D")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 0.892_dp
         s8 = 0.909_dp
      CASE ("BHLYP")
         s6 = 1.000_dp
         sr6 = 1.370_dp
         s8 = 1.442_dp
      CASE ("BLYP")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.094_dp
         s8 = 1.682_dp
      CASE ("BP86")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.139_dp
         s8 = 1.683_dp
      CASE ("BPBE")
         s6 = 1.000_dp
         sr6 = 1.087_dp
         s8 = 2.033_dp
      CASE ("MPWLYP")
         s6 = 1.000_dp
         sr6 = 1.239_dp
         s8 = 1.098_dp
      CASE ("PBE")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.217_dp
         s8 = 0.722_dp
      CASE ("PBE0")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.287_dp
         s8 = 0.928_dp
      CASE ("PW6B95")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.532_dp
         s8 = 0.862_dp
      CASE ("PWB6K")
         s6 = 1.000_dp
         sr6 = 1.660_dp
         s8 = 0.550_dp
      CASE ("revPBE")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 0.923_dp
         s8 = 1.010_dp
      CASE ("TPSS")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.166_dp
         s8 = 1.105_dp
      CASE ("TPSS0")
         ! S. Grimme, J. Antony, S. Ehrlich, and H. Krieg
         ! J. Chem. Phys. 132,154104 (2010); doi:10.1063/1.3382344
         s6 = 1.000_dp
         sr6 = 1.252_dp
         s8 = 1.242_dp
      CASE ("TPSSh")
         s6 = 1.000_dp
         sr6 = 1.223_dp
         s8 = 1.219_dp
      CASE ("B3PW91")
         s6 = 1.000_dp
         sr6 = 1.176_dp
         s8 = 1.775_dp
      CASE ("BMK")
         s6 = 1.000_dp
         sr6 = 1.931_dp
         s8 = 2.168_dp
      CASE ("CAMB3LYP")
         s6 = 1.000_dp
         sr6 = 1.378_dp
         s8 = 1.217_dp
      CASE ("LCWPBE")
         s6 = 1.000_dp
         sr6 = 1.355_dp
         s8 = 1.279_dp
      CASE ("M052X")
         s6 = 1.000_dp
         sr6 = 1.417_dp
         s8 = 0.000_dp
      CASE ("M05")
         s6 = 1.000_dp
         sr6 = 1.373_dp
         s8 = 0.595_dp
      CASE ("M062X")
         s6 = 1.000_dp
         sr6 = 1.619_dp
         s8 = 0.000_dp
      CASE ("M06HF")
         s6 = 1.000_dp
         sr6 = 1.446_dp
         s8 = 0.000_dp
      CASE ("M06L")
         s6 = 1.000_dp
         sr6 = 1.581_dp
         s8 = 0.000_dp
      CASE ("M06N")
         s6 = 1.000_dp
         sr6 = 1.325_dp
         s8 = 0.000_dp
      CASE ("HCTH120")
         s6 = 1.000_dp
         sr6 = 1.221_dp
         s8 = 1.206_dp
      CASE ("PTPSS")
         s6 = 0.750_dp
         sr6 = 1.541_dp
         s8 = 0.879_dp
      CASE ("PWPB95")
         s6 = 0.820_dp
         sr6 = 1.557_dp
         s8 = 0.705_dp
      CASE ("OLYP")
         s6 = 1.000_dp
         sr6 = 0.806_dp
         s8 = 1.764_dp
      CASE ("OPBE")
         s6 = 1.000_dp
         sr6 = 0.837_dp
         s8 = 2.055_dp
      CASE ("OTPSS")
         s6 = 1.000_dp
         sr6 = 1.128_dp
         s8 = 1.494_dp
      CASE ("PBE38")
         s6 = 1.000_dp
         sr6 = 1.333_dp
         s8 = 0.998_dp
      CASE ("PBEsol")
         s6 = 1.000_dp
         sr6 = 1.345_dp
         s8 = 0.612_dp
      CASE ("revSSB")
         s6 = 1.000_dp
         sr6 = 1.221_dp
         s8 = 0.560_dp
      CASE ("SSB")
         s6 = 1.000_dp
         sr6 = 1.215_dp
         s8 = 0.663_dp
      CASE ("BOP")
         s6 = 1.000_dp
         sr6 = 0.929_dp
         s8 = 1.975_dp
      CASE ("MPW1B95")
         s6 = 1.000_dp
         sr6 = 1.605_dp
         s8 = 1.118_dp
      CASE ("MPWB1K")
         s6 = 1.000_dp
         sr6 = 1.671_dp
         s8 = 1.061_dp
      CASE ("revPBE0")
         s6 = 1.000_dp
         sr6 = 0.949_dp
         s8 = 0.792_dp
      CASE ("revPBE38")
         s6 = 1.000_dp
         sr6 = 1.021_dp
         s8 = 0.862_dp
      CASE ("rPW86PBE")
         s6 = 1.000_dp
         sr6 = 1.224_dp
         s8 = 0.901_dp
      END SELECT

   END SUBROUTINE qs_scaling_dftd3

! **************************************************************************************************
!> \brief ...
!> \param s6 ...
!> \param a1 ...
!> \param s8 ...
!> \param a2 ...
!> \param vdw_section ...
! **************************************************************************************************
   SUBROUTINE qs_scaling_dftd3bj(s6, a1, s8, a2, vdw_section)
      REAL(KIND=dp), INTENT(inout)                       :: s6, a1, s8, a2
      TYPE(section_vals_type), POINTER                   :: vdw_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scaling_dftd3bj', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: functional

      CALL section_vals_val_get(vdw_section, "PAIR_POTENTIAL%REFERENCE_FUNCTIONAL", c_val=functional)

      ! values for different functionals from:
      ! http://www.thch.uni-bonn.de/tc/downloads/DFT-D3/functionalsbj.html
      SELECT CASE (TRIM (functional))
      CASE DEFAULT
         ! unknown functional
         CALL cp__b("qs_dispersion_pairpot.F",608,"No DFT-D3(BJ) values available for this functional:"//TRIM(functional))
      CASE ("B1B95")
         s6 = 1.0000_dp
         a1 = 0.2092_dp
         s8 = 1.4507_dp
         a2 = 5.5545_dp
      CASE ("B2GPPLYP")
         s6 = 0.5600_dp
         a1 = 0.0000_dp
         s8 = 0.2597_dp
         a2 = 6.3332_dp
      CASE ("B3PW91")
         s6 = 1.0000_dp
         a1 = 0.4312_dp
         s8 = 2.8524_dp
         a2 = 4.4693_dp
      CASE ("BHLYP")
         s6 = 1.0000_dp
         a1 = 0.2793_dp
         s8 = 1.0354_dp
         a2 = 4.9615_dp
      CASE ("BMK")
         s6 = 1.0000_dp
         a1 = 0.1940_dp
         s8 = 2.0860_dp
         a2 = 5.9197_dp
      CASE ("BOP")
         s6 = 1.0000_dp
         a1 = 0.4870_dp
         s8 = 3.2950_dp
         a2 = 3.5043_dp
      CASE ("BPBE")
         s6 = 1.0000_dp
         a1 = 0.4567_dp
         s8 = 4.0728_dp
         a2 = 4.3908_dp
      CASE ("CAMB3LYP")
         s6 = 1.0000_dp
         a1 = 0.3708_dp
         s8 = 2.0674_dp
         a2 = 5.4743_dp
      CASE ("LCWPBE")
         s6 = 1.0000_dp
         a1 = 0.3919_dp
         s8 = 1.8541_dp
         a2 = 5.0897_dp
      CASE ("MPW1B95")
         s6 = 1.0000_dp
         a1 = 0.1955_dp
         s8 = 1.0508_dp
         a2 = 6.4177_dp
      CASE ("MPWB1K")
         s6 = 1.0000_dp
         a1 = 0.1474_dp
         s8 = 0.9499_dp
         a2 = 6.6223_dp
      CASE ("MPWLYP")
         s6 = 1.0000_dp
         a1 = 0.4831_dp
         s8 = 2.0077_dp
         a2 = 4.5323_dp
      CASE ("OLYP")
         s6 = 1.0000_dp
         a1 = 0.5299_dp
         s8 = 2.6205_dp
         a2 = 2.8065_dp
      CASE ("OPBE")
         s6 = 1.0000_dp
         a1 = 0.5512_dp
         s8 = 3.3816_dp
         a2 = 2.9444_dp
      CASE ("OTPSS")
         s6 = 1.0000_dp
         a1 = 0.4634_dp
         s8 = 2.7495_dp
         a2 = 4.3153_dp
      CASE ("PBE38")
         s6 = 1.0000_dp
         a1 = 0.3995_dp
         s8 = 1.4623_dp
         a2 = 5.1405_dp
      CASE ("PBEsol")
         s6 = 1.0000_dp
         a1 = 0.4466_dp
         s8 = 2.9491_dp
         a2 = 6.1742_dp
      CASE ("PTPSS")
         s6 = 0.7500_dp
         a1 = 0.0000_dp
         s8 = 0.2804_dp
         a2 = 6.5745_dp
      CASE ("PWB6K")
         s6 = 1.0000_dp
         a1 = 0.1805_dp
         s8 = 0.9383_dp
         a2 = 7.7627_dp
      CASE ("revSSB")
         s6 = 1.0000_dp
         a1 = 0.4720_dp
         s8 = 0.4389_dp
         a2 = 4.0986_dp
      CASE ("SSB")
         s6 = 1.0000_dp
         a1 = -0.0952_dp
         s8 = -0.1744_dp
         a2 = 5.2170_dp
      CASE ("TPSSh")
         s6 = 1.0000_dp
         a1 = 0.4529_dp
         s8 = 2.2382_dp
         a2 = 4.6550_dp
      CASE ("HCTH120")
         s6 = 1.0000_dp
         a1 = 0.3563_dp
         s8 = 1.0821_dp
         a2 = 4.3359_dp
      CASE ("B2PLYP")
         s6 = 0.6400_dp
         a1 = 0.3065_dp
         s8 = 0.9147_dp
         a2 = 5.0570_dp
      CASE ("B3LYP")
         s6 = 1.0000_dp
         a1 = 0.3981_dp
         s8 = 1.9889_dp
         a2 = 4.4211_dp
      CASE ("B97-D")
         s6 = 1.0000_dp
         a1 = 0.5545_dp
         s8 = 2.2609_dp
         a2 = 3.2297_dp
      CASE ("BLYP")
         s6 = 1.0000_dp
         a1 = 0.4298_dp
         s8 = 2.6996_dp
         a2 = 4.2359_dp
      CASE ("BP86")
         s6 = 1.0000_dp
         a1 = 0.3946_dp
         s8 = 3.2822_dp
         a2 = 4.8516_dp
      CASE ("DSD-BLYP")
         s6 = 0.5000_dp
         a1 = 0.0000_dp
         s8 = 0.2130_dp
         a2 = 6.0519_dp
      CASE ("PBE0")
         s6 = 1.0000_dp
         a1 = 0.4145_dp
         s8 = 1.2177_dp
         a2 = 4.8593_dp
      CASE ("PBE")
         s6 = 1.0000_dp
         a1 = 0.4289_dp
         s8 = 0.7875_dp
         a2 = 4.4407_dp
      CASE ("PW6B95")
         s6 = 1.0000_dp
         a1 = 0.2076_dp
         s8 = 0.7257_dp
         a2 = 6.3750_dp
      CASE ("PWPB95")
         s6 = 0.8200_dp
         a1 = 0.0000_dp
         s8 = 0.2904_dp
         a2 = 7.3141_dp
      CASE ("revPBE0")
         s6 = 1.0000_dp
         a1 = 0.4679_dp
         s8 = 1.7588_dp
         a2 = 3.7619_dp
      CASE ("revPBE38")
         s6 = 1.0000_dp
         a1 = 0.4309_dp
         s8 = 1.4760_dp
         a2 = 3.9446_dp
      CASE ("revPBE")
         s6 = 1.0000_dp
         a1 = 0.5238_dp
         s8 = 2.3550_dp
         a2 = 3.5016_dp
      CASE ("RPW86PBE")
         s6 = 1.0000_dp
         a1 = 0.4613_dp
         s8 = 1.3845_dp
         a2 = 4.5062_dp
      CASE ("TPSS0")
         s6 = 1.0000_dp
         a1 = 0.3768_dp
         s8 = 1.2576_dp
         a2 = 4.5865_dp
      CASE ("TPSS")
         s6 = 1.0000_dp
         a1 = 0.4535_dp
         s8 = 1.9435_dp
         a2 = 4.4752_dp
      END SELECT

   END SUBROUTINE qs_scaling_dftd3bj

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param dispersion_env ...
!> \param energy ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE calculate_dispersion_pairpot(qs_env, dispersion_env, energy, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      REAL(KIND=dp), INTENT(OUT)                         :: energy
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_dispersion_pairpot', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, atom_c, atom_d, handle, hashb, hashc, i, ia, iatom, icx, icy, &
         icz, idmp, ikind, ilist, jatom, jkind, katom, kkind, kstart, latom, lkind, max_elem, &
         maxc, na, natom, nb, nc, ni, nj, nkind, unit_nr, za, zb, zc
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, atomnumber, kind_of
      INTEGER, DIMENSION(3)                              :: cell_b, cell_c, ncell, periodic
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: atenergy, atstress, debug, debugall, &
                                                            debugx, ghost_a, ghost_b, ghost_c, &
                                                            is000, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: dodisp, ghost
      REAL(KIND=dp) :: a1, a2, alp6, alp8, ang, c6, c8, c9, cc6ab, cc6bc, cc6ca, cnab, cnum, dc6a, &
         dc6b, dc8a, dc8b, dcc6aba, dcc6abb, dcc6bcb, dcc6bcc, dcc6caa, dcc6cac, dcnab, dd, de6, &
         de8, de91, de921, de922, dea, devdw, dfdab6, dfdab8, dfdabc, dfdmp, dr, drk, e6, e6tot, &
         e8, e8tot, e9, e9tot, elrc6, elrc8, elrc9, eps_cn, er, evdw, f0ab, fac, fac0, fdab6, &
         fdab8, fdabc, fdmp, gnorm, nab, nabc, r0, r0ab, r2ab, r2abc, r2bc, r2ca, r6, r8, rabc, &
         rc2, rcc, rcova, rcovb, rcut, rs6, rs8, s1, s2, s3, s6, s8, s9, xp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: c6d2, cnkind, cnumbers, cnumfix, radd2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rcpbc
      REAL(KIND=dp), DIMENSION(3)                        :: fdij, fdik, ra, rab, rb, rb0, rbc, rc, &
                                                            rc0, rca, rij, rik, sab_max
      REAL(KIND=dp), DIMENSION(3, 3)                     :: dvirial
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atener
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: atstr
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dcnum_type), ALLOCATABLE, DIMENSION(:)        :: dcnum
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_cn, sab_vdw
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_atom_dispersion_type), POINTER             :: disp_a, disp_b, disp_c
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      energy = 0._dp

      IF (dispersion_env%type == xc_vdw_fun_pairpot) THEN

         NULLIFY (atomic_kind_set, qs_kind_set, sab_cn, sab_vdw)

         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                         cell=cell, virial=virial, para_env=para_env, atprop=atprop)

         debugx = dispersion_env%verbose
         debugall = debugx
         debug = debugx .AND. para_env%ionode
         nkind = SIZE(atomic_kind_set)

         NULLIFY (logger)
         logger => cp_get_default_logger()
         IF (ASSOCIATED(dispersion_env%dftd_section)) THEN
            unit_nr = cp_print_key_unit_nr(logger, dispersion_env%dftd_section, "PRINT_DFTD", &
                                           extension=".dftd")
         ELSE
            unit_nr = -1
         END IF

         ! atomic energy and stress arrays
         atenergy = atprop%energy
         IF (atenergy) THEN
            NULLIFY (particle_set)
            CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
            natom = SIZE(particle_set)
            CALL atprop_array_init(atprop%atevdw, natom)
            atener => atprop%atevdw
         END IF
         atstress = atprop%stress
         IF (atstress) THEN
            atstr => atprop%atstress
         END IF

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *)
            WRITE (unit_nr, *) " Pair potential vdW calculation"
            IF (dispersion_env%pp_type == vdw_pairpot_dftd2) THEN
               WRITE (unit_nr, *) " Dispersion potential type: DFT-D2"
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
               WRITE (unit_nr, *) " Dispersion potential type: DFT-D3"
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
               WRITE (unit_nr, *) " Dispersion potential type: DFT-D3(BJ)"
            END IF
         END IF

         NULLIFY (particle_set)
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
         natom = SIZE(particle_set)
         IF (calculate_forces .OR. debugall) THEN
            NULLIFY (force)
            CALL get_qs_env(qs_env=qs_env, force=force)
            ALLOCATE (atom_of_kind(natom), kind_of(natom))
            CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind, kind_of=kind_of)
            use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
            IF (use_virial .AND. debugall) THEN
               dvirial = virial%pv_virial
            END IF
         ELSE IF ((dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
                   dispersion_env%pp_type == vdw_pairpot_dftd3bj) .AND. dispersion_env%doabc) THEN
            ALLOCATE (atom_of_kind(natom), kind_of(natom))
            CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind, kind_of=kind_of)
         END IF

         ALLOCATE (dodisp(nkind), ghost(nkind), atomnumber(nkind), c6d2(nkind), radd2(nkind))
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), z=za)
            CALL get_qs_kind(qs_kind_set(ikind), dispersion=disp_a, ghost=ghost_a)
            dodisp(ikind) = disp_a%defined
            ghost(ikind) = ghost_a
            atomnumber(ikind) = za
            c6d2(ikind) = disp_a%c6
            radd2(ikind) = disp_a%vdw_radii
         END DO
         ALLOCATE (rcpbc(3, natom))
         DO iatom = 1, natom
            rcpbc(:, iatom) = pbc(particle_set(iatom)%r(:), cell)
         END DO

         rcut = 2._dp*dispersion_env%rc_disp
         rc2 = rcut*rcut
         IF (dispersion_env%pp_type == vdw_pairpot_dftd2) THEN
            s6 = dispersion_env%scaling
            dd = dispersion_env%exp_pre
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) " Scaling parameter (s6) ", s6
               WRITE (unit_nr, *) " Exponential prefactor  ", dd
            END IF
         ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
                  dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
            maxc = SIZE(dispersion_env%c6ab, 3)
            max_elem = SIZE(dispersion_env%c6ab, 1)
            alp6 = dispersion_env%alp
            alp8 = alp6+2._dp
            s6 = dispersion_env%s6
            s8 = dispersion_env%s8
            s9 = dispersion_env%s6
            rs6 = dispersion_env%sr6
            rs8 = 1._dp
            a1 = dispersion_env%a1
            a2 = dispersion_env%a2
            eps_cn = dispersion_env%eps_cn
            e6tot = 0._dp
            e8tot = 0._dp
            e9tot = 0._dp
            ! damping type
            idmp = 0
            IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
               idmp = 1
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
               idmp = 2
            END IF

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) " Scaling parameter (s6) ", s6
               WRITE (unit_nr, *) " Scaling parameter (s8) ", s8
               IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
                  WRITE (unit_nr, *) " Zero Damping parameter (sr6)", rs6
                  WRITE (unit_nr, *) " Zero Damping parameter (sr8)", rs8
               ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
                  WRITE (unit_nr, *) " BJ Damping parameter (a1) ", a1
                  WRITE (unit_nr, *) " BJ Damping parameter (a2) ", a2
               END IF
               WRITE (unit_nr, *) " Cutoff coordination numbers", eps_cn
               IF (dispersion_env%lrc) THEN
                  WRITE (unit_nr, *) " Apply a long range correction"
               END IF
            END IF
            ! Calculate coordination numbers
            NULLIFY (particle_set)
            CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
            natom = SIZE(particle_set)
            ALLOCATE (cnumbers(natom))
            cnumbers = 0._dp
            IF (calculate_forces .OR. debugall) THEN
               ALLOCATE (dcnum(natom))
               dcnum(:)%neighbors = 0
               DO iatom = 1, natom
                  ALLOCATE (dcnum(iatom)%nlist(10), dcnum(iatom)%dvals(10), dcnum(iatom)%rik(3, 10))
               END DO
            END IF
            sab_cn => dispersion_env%sab_cn
            CALL neighbor_list_iterator_create(nl_iterator, sab_cn)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)
               IF (ghost(ikind) .OR. ghost(jkind)) CYCLE
               rcc = SQRT(rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3))
               IF (rcc > 1.e-6_dp) THEN
                  za = atomnumber(ikind)
                  zb = atomnumber(jkind)
                  rcova = dispersion_env%rcov(za)
                  rcovb = dispersion_env%rcov(zb)
                  CALL cnparam_d3(rcc, rcova, rcovb, dispersion_env%k1, cnab, dcnab)
                  IF (cnab > eps_cn) THEN
                     cnumbers(iatom) = cnumbers(iatom)+cnab
                     IF (iatom /= jatom) cnumbers(jatom) = cnumbers(jatom)+cnab
                  END IF
                  IF (calculate_forces .OR. debugall .AND. cnab > eps_cn) THEN
                     dcnum(iatom)%neighbors = dcnum(iatom)%neighbors+1
                     ni = dcnum(iatom)%neighbors
                     IF (ni > SIZE(dcnum(iatom)%nlist)) THEN
                        CALL reallocate(dcnum(iatom)%nlist, 1, 2*ni)
                        CALL reallocate(dcnum(iatom)%dvals, 1, 2*ni)
                        CALL reallocate(dcnum(iatom)%rik, 1, 3, 1, 2*ni)
                     END IF
                     dcnum(iatom)%nlist(ni) = jatom
                     dcnum(iatom)%dvals(ni) = dcnab
                     dcnum(iatom)%rik(1:3, ni) = rij(1:3)
                     IF (iatom /= jatom) THEN
                        dcnum(jatom)%neighbors = dcnum(jatom)%neighbors+1
                        nj = dcnum(jatom)%neighbors
                        IF (nj > SIZE(dcnum(jatom)%dvals)) THEN
                           CALL reallocate(dcnum(jatom)%nlist, 1, 2*nj)
                           CALL reallocate(dcnum(jatom)%dvals, 1, 2*nj)
                           CALL reallocate(dcnum(jatom)%rik, 1, 3, 1, 2*nj)
                        END IF
                        dcnum(jatom)%nlist(nj) = iatom
                        dcnum(jatom)%dvals(nj) = dcnab
                        dcnum(jatom)%rik(1:3, nj) = -rij(1:3)
                     END IF
                  END IF
               END IF
            END DO
            CALL neighbor_list_iterator_release(nl_iterator)
            CALL mp_sum(cnumbers, para_env%group)
            ! for parallel runs we have to update dcnum on all processors
            IF (calculate_forces .OR. debugall) THEN
               CALL dcnum_distribute(dcnum, para_env)
               IF (unit_nr > 0 .AND. SIZE(dcnum) > 0) THEN
                  WRITE (unit_nr, *)
                  WRITE (unit_nr, *) "  ATOM       Coordination   Neighbors"
                  DO i = 1, natom
                     WRITE (unit_nr, "(I6,F20.10,I12)") i, cnumbers(i), dcnum(i)%neighbors
                  END DO
                  WRITE (unit_nr, *)
               END IF
            END IF
         END IF

         nab = 0._dp
         nabc = 0._dp
         IF (dispersion_env%doabc) THEN
            rcc = 2._dp*dispersion_env%rc_disp
            CALL get_cell(cell=cell, periodic=periodic)
            sab_max(1) = rcc/plane_distance(1, 0, 0, cell)
            sab_max(2) = rcc/plane_distance(0, 1, 0, cell)
            sab_max(3) = rcc/plane_distance(0, 0, 1, cell)
            ncell(:) = (INT(sab_max(:))+1)*periodic(:)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) " Calculate C9 Terms"
               WRITE (unit_nr, "(A,T20,I3,A,I3)") "  Search in cells ", -ncell(1), ":", ncell(1)
               WRITE (unit_nr, "(T20,I3,A,I3)")-ncell(2), ":", ncell(2)
               WRITE (unit_nr, "(T20,I3,A,I3)")-ncell(3), ":", ncell(3)
               WRITE (unit_nr, *)
            END IF
            IF (dispersion_env%c9cnst) THEN
               IF (unit_nr > 0) WRITE (unit_nr, *) " Use reference coordination numbers for C9 term"
               ALLOCATE (cnumfix(natom))
               cnumfix = 0._dp
               ! first use the default values
               DO iatom = 1, natom
                  ikind = kind_of(iatom)
                  CALL get_atomic_kind(atomic_kind_set(ikind), z=za)
                  cnumfix(iatom) = dispersion_env%cn(za)
               END DO
               ! now check for changes from default
               IF (ASSOCIATED(dispersion_env%cnkind)) THEN
                  DO i = 1, SIZE(dispersion_env%cnkind)
                     ikind = dispersion_env%cnkind(i)%kind
                     cnum = dispersion_env%cnkind(i)%cnum
                     IF(.NOT.(ikind <= nkind))CALL cp__a("qs_dispersion_pairpot.F",1099)
                     IF(.NOT.(ikind > 0))CALL cp__a("qs_dispersion_pairpot.F",1100)
                     CALL get_atomic_kind(atomic_kind_set(ikind), natom=na, atom_list=atom_list)
                     DO ia = 1, na
                        iatom = atom_list(ia)
                        cnumfix(iatom) = cnum
                     END DO
                  END DO
               END IF
               IF (ASSOCIATED(dispersion_env%cnlist)) THEN
                  DO i = 1, SIZE(dispersion_env%cnlist)
                     DO ilist = 1, dispersion_env%cnlist(i)%natom
                        iatom = dispersion_env%cnlist(i)%atom(ilist)
                        cnumfix(iatom) = dispersion_env%cnlist(i)%cnum
                     END DO
                  END DO
               END IF
               IF (unit_nr > 0) THEN
                  DO i = 1, natom
                     IF (ABS(cnumbers(i)-cnumfix(i)) > 0.5_dp) THEN
                        WRITE (unit_nr, "(A,T20,A,I6,T41,2F10.3)") "  Difference in CN ", "Atom:", &
                           i, cnumbers(i), cnumfix(i)
                     END IF
                  END DO
                  WRITE (unit_nr, *)
               END IF
            END IF
         END IF

         evdw = 0._dp

         sab_vdw => dispersion_env%sab_vdw

         nkind = SIZE(atomic_kind_set)

         CALL neighbor_list_iterator_create(nl_iterator, sab_vdw)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)
            IF (ghost(ikind) .OR. ghost(jkind)) CYCLE
            IF (.NOT. (dodisp(ikind) .AND. dodisp(jkind))) CYCLE
            za = atomnumber(ikind)
            zb = atomnumber(jkind)
            ! vdW potential
            dr = SQRT(SUM(rij(:)**2))
            IF (dr <= rcut) THEN
               nab = nab+1._dp
               fac = 1._dp
               IF (iatom == jatom) fac = 0.5_dp
               IF (disp_a%type == dftd2_pp .AND. dr > 0.001_dp) THEN
                  c6 = SQRT(c6d2(ikind)*c6d2(jkind))
                  rcc = radd2(ikind)+radd2(jkind)
                  er = EXP(-dd*(dr/rcc-1._dp))
                  fdmp = 1._dp/(1._dp+er)
                  xp = s6*c6/dr**6
                  evdw = evdw-xp*fdmp*fac
                  IF (calculate_forces) THEN
                     dfdmp = dd/rcc*er*fdmp*fdmp
                     devdw = -xp*(-6._dp*fdmp/dr+dfdmp)
                     fdij(:) = devdw*rij(:)/dr*fac
                     atom_a = atom_of_kind(iatom)
                     atom_b = atom_of_kind(jatom)
                     force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdij(:)
                     force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)+fdij(:)
                     IF (use_virial) THEN
                        CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rij)
                     END IF
                     IF (atstress) THEN
                        CALL virial_pair_force(atstr(:, :, iatom), -0.5_dp, fdij, rij)
                        CALL virial_pair_force(atstr(:, :, jatom), -0.5_dp, fdij, rij)
                     END IF
                  END IF
                  IF (atenergy) THEN
                     atener(iatom) = atener(iatom)-0.5_dp*xp*fdmp*fac
                     atener(jatom) = atener(jatom)-0.5_dp*xp*fdmp*fac
                  END IF
               ELSE IF (disp_a%type == dftd3_pp .AND. dr > 0.001_dp) THEN
                  ! C6 coefficient
                  CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, za, zb, &
                             cnumbers(iatom), cnumbers(jatom), dispersion_env%k3, c6, dc6a, dc6b)
                  c8 = 3.0d0*c6*dispersion_env%r2r4(za)*dispersion_env%r2r4(zb)
                  dc8a = 3.0d0*dc6a*dispersion_env%r2r4(za)*dispersion_env%r2r4(zb)
                  dc8b = 3.0d0*dc6b*dispersion_env%r2r4(za)*dispersion_env%r2r4(zb)
                  r6 = dr**6
                  r8 = r6*dr*dr
                  ! damping
                  IF (idmp == 1) THEN
                     ! zero
                     CALL damping_d3(dr, dispersion_env%r0ab(za, zb), rs6, alp6, rcut, fdab6, dfdab6)
                     CALL damping_d3(dr, dispersion_env%r0ab(za, zb), rs8, alp8, rcut, fdab8, dfdab8)
                     e6 = s6*fac*c6*fdab6/r6
                     e8 = s8*fac*c8*fdab8/r8
                  ELSE IF (idmp == 2) THEN
                     ! BJ
                     r0ab = SQRT(3.0d0*dispersion_env%r2r4(za)*dispersion_env%r2r4(zb))
                     f0ab = a1*r0ab+a2
                     fdab6 = 1.0_dp/(r6+f0ab**6)
                     fdab8 = 1.0_dp/(r8+f0ab**8)
                     e6 = s6*fac*c6*fdab6
                     e8 = s8*fac*c8*fdab8
                  ELSE
                     CALL cp__b("qs_dispersion_pairpot.F",1199,"Unknown DFT-D3 damping function:")
                  END IF
                  evdw = evdw-e6-e8
                  e6tot = e6tot-e6
                  e8tot = e8tot-e8
                  IF (atenergy) THEN
                     atener(iatom) = atener(iatom)-0.5_dp*(e6+e8)
                     atener(jatom) = atener(jatom)-0.5_dp*(e6+e8)
                  END IF
                  IF (calculate_forces) THEN
                     ! damping
                     IF (idmp == 1) THEN
                        ! zero
                        de6 = -s6*c6/r6*(dfdab6-6._dp*fdab6/dr)
                        de8 = -s8*c8/r8*(dfdab8-8._dp*fdab8/dr)
                     ELSE IF (idmp == 2) THEN
                        ! BJ
                        de6 = s6*c6*fdab6*fdab6*6.0_dp*dr**5
                        de8 = s8*c8*fdab8*fdab8*8.0_dp*dr**7
                     ELSE
                        CALL cp__b("qs_dispersion_pairpot.F",1219,"Unknown DFT-D3 damping function:")
                     END IF
                     fdij(:) = (de6+de8)*rij(:)/dr*fac
                     atom_a = atom_of_kind(iatom)
                     atom_b = atom_of_kind(jatom)
                     force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdij(:)
                     force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)+fdij(:)
                     IF (use_virial) THEN
                        CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rij)
                     END IF
                     IF (atstress) THEN
                        CALL virial_pair_force(atstr(:, :, iatom), -0.5_dp, fdij, rij)
                        CALL virial_pair_force(atstr(:, :, jatom), -0.5_dp, fdij, rij)
                     END IF
                     ! forces from the r-dependence of the coordination numbers
                     IF (idmp == 1) THEN
                        ! zero
                        dc6a = -s6*fac*dc6a*fdab6/r6
                        dc6b = -s6*fac*dc6b*fdab6/r6
                        dc8a = -s8*fac*dc8a*fdab8/r8
                        dc8b = -s8*fac*dc8b*fdab8/r8
                     ELSE IF (idmp == 2) THEN
                        ! BJ
                        dc6a = -s6*fac*dc6a*fdab6
                        dc6b = -s6*fac*dc6b*fdab6
                        dc8a = -s8*fac*dc8a*fdab8
                        dc8b = -s8*fac*dc8b*fdab8
                     ELSE
                        CALL cp__b("qs_dispersion_pairpot.F",1247,"Unknown DFT-D3 damping function:")
                     END IF
                     DO i = 1, dcnum(iatom)%neighbors
                        katom = dcnum(iatom)%nlist(i)
                        kkind = kind_of(katom)
                        rik = dcnum(iatom)%rik(:, i)
                        drk = SQRT(SUM(rik(:)**2))
                        fdik(:) = (dc6a+dc8a)*dcnum(iatom)%dvals(i)*rik(:)/drk
                        atom_c = atom_of_kind(katom)
                        force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdik(:)
                        force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)+fdik(:)
                        IF (use_virial) THEN
                           CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                        END IF
                        IF (atstress) THEN
                           CALL virial_pair_force(atstr(:, :, iatom), -0.5_dp, fdik, rik)
                           CALL virial_pair_force(atstr(:, :, katom), -0.5_dp, fdik, rik)
                        END IF
                     END DO
                     DO i = 1, dcnum(jatom)%neighbors
                        katom = dcnum(jatom)%nlist(i)
                        kkind = kind_of(katom)
                        rik = dcnum(jatom)%rik(:, i)
                        drk = SQRT(SUM(rik(:)**2))
                        fdik(:) = (dc6b+dc8b)*dcnum(jatom)%dvals(i)*rik(:)/drk
                        atom_c = atom_of_kind(katom)
                        force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)-fdik(:)
                        force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)+fdik(:)
                        IF (use_virial) THEN
                           CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                        END IF
                        IF (atstress) THEN
                           CALL virial_pair_force(atstr(:, :, jatom), -0.5_dp, fdik, rik)
                           CALL virial_pair_force(atstr(:, :, katom), -0.5_dp, fdik, rik)
                        END IF
                     END DO
                  END IF
                  IF (dispersion_env%doabc) THEN
                     CALL get_iterator_info(nl_iterator, cell=cell_b)
                     hashb = cellhash(cell_b, ncell)
                     is000 = (ALL(cell_b == 0))
                     rb0(:) = MATMUL(cell%hmat, cell_b)
                     ra(:) = pbc(particle_set(iatom)%r(:), cell)
                     rb(:) = pbc(particle_set(jatom)%r(:), cell)+rb0
                     r2ab = SUM((ra-rb)**2)
                     DO icx = -ncell(1), ncell(1)
                        DO icy = -ncell(2), ncell(2)
                           DO icz = -ncell(3), ncell(3)
                              cell_c(1) = icx
                              cell_c(2) = icy
                              cell_c(3) = icz
                              hashc = cellhash(cell_c, ncell)
                              IF (is000 .AND. (ALL(cell_c == 0))) THEN
                                 ! CASE 1: all atoms in (000), use only ordered triples
                                 kstart = MAX(jatom+1, iatom+1)
                                 fac0 = 1.0_dp
                              ELSE IF (is000) THEN
                                 ! CASE 2: AB in (000), C in other cell
                                 !         This case covers also all instances with BC in same
                                 !         cell not (000)
                                 kstart = 1
                                 fac0 = 1.0_dp
                              ELSE
                                 ! These are case 2 again, cycle
                                 IF (hashc == hashb) CYCLE
                                 IF (ALL(cell_c == 0)) CYCLE
                                 ! CASE 3: A in (000) and B and C in different cells
                                 kstart = 1
                                 fac0 = 1.0_dp/3.0_dp
                              END IF
                              rc0 = MATMUL(cell%hmat, cell_c)
                              DO katom = kstart, natom
                                 kkind = kind_of(katom)
                                 IF (ghost(kkind) .OR. .NOT. dodisp(kkind)) CYCLE
                                 rc(:) = rcpbc(:, katom)+rc0(:)
                                 r2bc = SUM((rb-rc)**2)
                                 IF (r2bc >= rc2) CYCLE
                                 r2ca = SUM((rc-ra)**2)
                                 IF (r2ca >= rc2) CYCLE
                                 r2abc = r2ab*r2bc*r2ca
                                 IF (r2abc <= 0.001_dp) CYCLE
                                 ! this is an empirical scaling
                                 IF (r2abc <= 0.01*rc2*rc2*rc2) THEN
                                    kkind = kind_of(katom)
                                    atom_c = atom_of_kind(katom)
                                    zc = atomnumber(kkind)
                                    ! avoid double counting!
                                    fac = 1._dp
                                    IF (iatom == jatom .OR. iatom == katom .OR. jatom == katom) fac = 0.5_dp
                                    IF (iatom == jatom .AND. iatom == katom) fac = 1._dp/3._dp
                                    fac = fac*fac0
                                    nabc = nabc+1._dp
                                    IF (dispersion_env%c9cnst) THEN
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, za, zb, &
                                                  cnumfix(iatom), cnumfix(jatom), dispersion_env%k3, cc6ab, dcc6aba, dcc6abb)
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zb, zc, &
                                                  cnumfix(jatom), cnumfix(katom), dispersion_env%k3, cc6bc, dcc6bcb, dcc6bcc)
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zc, za, &
                                                  cnumfix(katom), cnumfix(iatom), dispersion_env%k3, cc6ca, dcc6cac, dcc6caa)
                                    ELSE
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, za, zb, &
                                                  cnumbers(iatom), cnumbers(jatom), dispersion_env%k3, cc6ab, dcc6aba, dcc6abb)
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zb, zc, &
                                                  cnumbers(jatom), cnumbers(katom), dispersion_env%k3, cc6bc, dcc6bcb, dcc6bcc)
                                       CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zc, za, &
                                                  cnumbers(katom), cnumbers(iatom), dispersion_env%k3, cc6ca, dcc6cac, dcc6caa)
                                    END IF
                                    c9 = -SQRT(cc6ab*cc6bc*cc6ca)
                                    rabc = r2abc**(1._dp/6._dp)
                                    r0 = (dispersion_env%r0ab(za, zb)*dispersion_env%r0ab(zb, zc)* &
                                          dispersion_env%r0ab(zc, za))**(1._dp/3._dp)
                                    CALL damping_d3(rabc, r0, 4._dp/3._dp, alp6, rcut, fdabc, dfdabc)
                                    s1 = r2ab+r2bc-r2ca
                                    s2 = r2ab+r2ca-r2bc
                                    s3 = r2ca+r2bc-r2ab
                                    ang = 0.375_dp*s1*s2*s3/r2abc+1.0_dp

                                    e9 = s9*fac*fdabc*c9*ang/r2abc**1.50d0
                                    evdw = evdw-e9
                                    e9tot = e9tot-e9
                                    IF (atenergy) THEN
                                       atener(iatom) = atener(iatom)-e9/3._dp
                                       atener(jatom) = atener(jatom)-e9/3._dp
                                       atener(katom) = atener(katom)-e9/3._dp
                                    END IF

                                    IF (calculate_forces) THEN
                                       rab = rb-ra; rbc = rc-rb; rca = ra-rc
                                       de91 = s9*fac*dfdabc*c9*ang/r2abc**1.50d0
                                       de91 = -de91/3._dp*rabc+3._dp*e9
                                       dea = s9*fac*fdabc*c9/r2abc**2.50d0*0.75_dp
                                       fdij(:) = de91*rab(:)/r2ab
                                       fdij(:) = fdij(:)+dea*s1*s2*s3*rab(:)/r2ab
                                       fdij(:) = fdij(:)-dea*(s2*s3+s1*s3-s1*s2)*rab(:)
                                       force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdij(:)
                                       force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)+fdij(:)
                                       IF (use_virial) THEN
                                          CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rab)
                                       END IF
                                       IF (atstress) THEN
                                          CALL virial_pair_force(atstr(:, :, iatom), -0.5_dp, fdij, rab)
                                          CALL virial_pair_force(atstr(:, :, jatom), -0.5_dp, fdij, rab)
                                       END IF
                                       fdij(:) = de91*rbc(:)/r2bc
                                       fdij(:) = fdij(:)+dea*s1*s2*s3*rbc(:)/r2bc
                                       fdij(:) = fdij(:)-dea*(s2*s3-s1*s3+s1*s2)*rbc(:)
                                       force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)-fdij(:)
                                       force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)+fdij(:)
                                       IF (use_virial) THEN
                                          CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rbc)
                                       END IF
                                       IF (atstress) THEN
                                          CALL virial_pair_force(atstr(:, :, jatom), -0.5_dp, fdij, rbc)
                                          CALL virial_pair_force(atstr(:, :, katom), -0.5_dp, fdij, rbc)
                                       END IF
                                       fdij(:) = de91*rca(:)/r2ca
                                       fdij(:) = fdij(:)+dea*s1*s2*s3*rca(:)/r2ca
                                       fdij(:) = fdij(:)-dea*(-s2*s3+s1*s3+s1*s2)*rca(:)
                                       force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)-fdij(:)
                                       force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)+fdij(:)
                                       IF (use_virial) THEN
                                          CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rca)
                                       END IF
                                       IF (atstress) THEN
                                          CALL virial_pair_force(atstr(:, :, iatom), -0.5_dp, fdij, rca)
                                          CALL virial_pair_force(atstr(:, :, katom), -0.5_dp, fdij, rca)
                                       END IF
                                       IF (.NOT. dispersion_env%c9cnst) THEN
                                          ! forces from the r-dependence of the coordination numbers
                                          ! atomic stress not implemented
                                          de91 = 0.5_dp*e9/cc6ab
                                          de921 = de91*dcc6aba
                                          de922 = de91*dcc6abb
                                          DO i = 1, dcnum(iatom)%neighbors
                                             latom = dcnum(iatom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(iatom)%rik(1, i)
                                             rik(2) = dcnum(iatom)%rik(2, i)
                                             rik(3) = dcnum(iatom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de921*dcnum(iatom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO
                                          DO i = 1, dcnum(jatom)%neighbors
                                             latom = dcnum(jatom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(jatom)%rik(1, i)
                                             rik(2) = dcnum(jatom)%rik(2, i)
                                             rik(3) = dcnum(jatom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de922*dcnum(jatom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO

                                          de91 = 0.5_dp*e9/cc6bc
                                          de921 = de91*dcc6bcb
                                          de922 = de91*dcc6bcc
                                          DO i = 1, dcnum(jatom)%neighbors
                                             latom = dcnum(jatom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(jatom)%rik(1, i)
                                             rik(2) = dcnum(jatom)%rik(2, i)
                                             rik(3) = dcnum(jatom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de921*dcnum(jatom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(jkind)%dispersion(:, atom_b) = force(jkind)%dispersion(:, atom_b)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO
                                          DO i = 1, dcnum(katom)%neighbors
                                             latom = dcnum(katom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(katom)%rik(1, i)
                                             rik(2) = dcnum(katom)%rik(2, i)
                                             rik(3) = dcnum(katom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de922*dcnum(katom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO

                                          de91 = 0.5_dp*e9/cc6ca
                                          de921 = de91*dcc6cac
                                          de922 = de91*dcc6caa
                                          DO i = 1, dcnum(katom)%neighbors
                                             latom = dcnum(katom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(katom)%rik(1, i)
                                             rik(2) = dcnum(katom)%rik(2, i)
                                             rik(3) = dcnum(katom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de921*dcnum(katom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(kkind)%dispersion(:, atom_c) = force(kkind)%dispersion(:, atom_c)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO
                                          DO i = 1, dcnum(iatom)%neighbors
                                             latom = dcnum(iatom)%nlist(i)
                                             lkind = kind_of(latom)
                                             rik(1) = dcnum(iatom)%rik(1, i)
                                             rik(2) = dcnum(iatom)%rik(2, i)
                                             rik(3) = dcnum(iatom)%rik(3, i)
                                             drk = SQRT(rik(1)*rik(1)+rik(2)*rik(2)+rik(3)*rik(3))
                                             fdik(:) = -de922*dcnum(iatom)%dvals(i)*rik(:)/drk
                                             atom_d = atom_of_kind(latom)
                                             force(ikind)%dispersion(:, atom_a) = force(ikind)%dispersion(:, atom_a)-fdik(:)
                                             force(lkind)%dispersion(:, atom_d) = force(lkind)%dispersion(:, atom_d)+fdik(:)
                                             IF (use_virial) THEN
                                                CALL virial_pair_force(virial%pv_virial, -1._dp, fdik, rik)
                                             END IF
                                          END DO
                                       END IF

                                    END IF

                                 END IF
                              END DO
                           END DO
                        END DO
                     END DO

                  END IF
               END IF
            END IF
         END DO
         CALL neighbor_list_iterator_release(nl_iterator)

         elrc6 = 0._dp
         elrc8 = 0._dp
         elrc9 = 0._dp
         IF (dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
             dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
            ! Long range correction (atomic contributions not implemented)
            IF (dispersion_env%lrc) THEN
               ALLOCATE (cnkind(nkind))
               cnkind = 0._dp
               ! first use the default values
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), z=za)
                  cnkind(ikind) = dispersion_env%cn(za)
               END DO
               ! now check for changes from default
               IF (ASSOCIATED(dispersion_env%cnkind)) THEN
                  DO i = 1, SIZE(dispersion_env%cnkind)
                     ikind = dispersion_env%cnkind(i)%kind
                     cnkind(ikind) = dispersion_env%cnkind(i)%cnum
                  END DO
               END IF
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), natom=na, z=za)
                  CALL get_qs_kind(qs_kind_set(ikind), dispersion=disp_a, ghost=ghost_a)
                  IF (.NOT. disp_a%defined .OR. ghost_a) CYCLE
                  DO jkind = 1, nkind
                     CALL get_atomic_kind(atomic_kind_set(jkind), natom=nb, z=zb)
                     CALL get_qs_kind(qs_kind_set(jkind), dispersion=disp_b, ghost=ghost_b)
                     IF (.NOT. disp_b%defined .OR. ghost_b) CYCLE
                     CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, za, zb, &
                                cnkind(ikind), cnkind(jkind), dispersion_env%k3, cc6ab, dcc6aba, dcc6abb)
                     elrc6 = elrc6-s6*twopi*REAL(na*nb, KIND=dp)*cc6ab/(3._dp*rcut**3*cell%deth)
                     c8 = 3.0d0*cc6ab*dispersion_env%r2r4(za)*dispersion_env%r2r4(zb)
                     elrc8 = elrc8-s8*twopi*REAL(na*nb, KIND=dp)*c8/(5._dp*rcut**5*cell%deth)
                     IF (dispersion_env%doabc) THEN
                        DO kkind = 1, nkind
                           CALL get_atomic_kind(atomic_kind_set(kkind), natom=nc, z=zc)
                           CALL get_qs_kind(qs_kind_set(kkind), dispersion=disp_c, ghost=ghost_c)
                           IF (.NOT. disp_c%defined .OR. ghost_c) CYCLE
                           CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, za, zb, &
                                      cnkind(ikind), cnkind(jkind), dispersion_env%k3, cc6ab, dcc6aba, dcc6abb)
                           CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zc, za, &
                                      cnkind(kkind), cnkind(ikind), dispersion_env%k3, cc6ca, dcc6aba, dcc6abb)
                           CALL getc6(maxc, max_elem, dispersion_env%c6ab, dispersion_env%maxci, zb, zc, &
                                      cnkind(jkind), cnkind(kkind), dispersion_env%k3, cc6bc, dcc6aba, dcc6abb)
                           c9 = -SQRT(cc6ab*cc6bc*cc6ca)
                           elrc9 = elrc9-s9*64._dp*twopi*REAL(na*nb*nc, KIND=dp)*c9/(6._dp*rcut**3*cell%deth**2)
                        END DO
                     END IF
                  END DO
               END DO
               IF (use_virial) THEN
                  DO i = 1, 3
                     virial%pv_virial(i, i) = virial%pv_virial(i, i)+(elrc6+elrc8+2._dp*elrc9)
                  END DO
               END IF
               DEALLOCATE (cnkind)
            END IF
         END IF

         IF (dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
             dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
            DEALLOCATE (cnumbers)
            IF (dispersion_env%doabc .AND. dispersion_env%c9cnst) THEN
               DEALLOCATE (cnumfix)
            END IF
            IF (calculate_forces .OR. debugall) THEN
               DO iatom = 1, natom
                  DEALLOCATE (dcnum(iatom)%nlist, dcnum(iatom)%dvals, dcnum(iatom)%rik)
               END DO
               DEALLOCATE (dcnum)
            END IF
         END IF

         ! set dispersion energy
         CALL mp_sum(evdw, para_env%group)
         evdw = evdw+(elrc6+elrc8+elrc9)
         energy = evdw

         IF ((dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
              dispersion_env%pp_type == vdw_pairpot_dftd3bj) .AND. debugall) THEN
            CALL mp_sum(e6tot, para_env%group)
            CALL mp_sum(e8tot, para_env%group)
            CALL mp_sum(e9tot, para_env%group)
            CALL mp_sum(nab, para_env%group)
            CALL mp_sum(nabc, para_env%group)
            e6tot = e6tot+elrc6
            e8tot = e8tot+elrc8
            e9tot = e9tot+elrc9
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, "(A,F20.0)") "  E6 vdW terms              :", nab
               WRITE (unit_nr, *) " E6 vdW energy [au/kcal]   :", e6tot, e6tot*kcalmol
               WRITE (unit_nr, *) " E8 vdW energy [au/kcal]   :", e8tot, e8tot*kcalmol
               WRITE (unit_nr, *) " %E8 on total vdW energy   :", e8tot/evdw*100._dp
               WRITE (unit_nr, "(A,F20.0)") "  E9 vdW terms              :", nabc
               WRITE (unit_nr, *) " E9 vdW energy [au/kcal]   :", e9tot, e9tot*kcalmol
               WRITE (unit_nr, *) " %E9 on total vdW energy   :", e9tot/evdw*100._dp
               IF (dispersion_env%lrc) THEN
                  WRITE (unit_nr, *) " E LRC C6 [au/kcal]        :", elrc6, elrc6*kcalmol
                  WRITE (unit_nr, *) " E LRC C8 [au/kcal]        :", elrc8, elrc8*kcalmol
                  WRITE (unit_nr, *) " E LRC C9 [au/kcal]        :", elrc9, elrc9*kcalmol
               END IF
            END IF
         END IF
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *) " Total vdW energy [au]     :", evdw
            WRITE (unit_nr, *) " Total vdW energy [kcal]   :", evdw*kcalmol
            WRITE (unit_nr, *)
         END IF
         IF (calculate_forces .AND. debugall) THEN
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) " Dispersion Forces         "
               WRITE (unit_nr, *) " Atom   Kind                            Forces    "
            END IF
            gnorm = 0._dp
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               atom_a = atom_of_kind(iatom)
               fdij(1:3) = force(ikind)%dispersion(:, atom_a)
               CALL mp_sum(fdij, para_env%group)
               gnorm = gnorm+SUM(ABS(fdij))
               IF (unit_nr > 0) WRITE (unit_nr, "(i5,i7,3F20.14)") iatom, ikind, fdij
            END DO
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *)
               WRITE (unit_nr, *) "|G| = ", gnorm
               WRITE (unit_nr, *)
            END IF
            IF (use_virial) THEN
               dvirial = virial%pv_virial-dvirial
               CALL mp_sum(dvirial, para_env%group)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "Stress Tensor (dispersion)"
                  WRITE (unit_nr, "(3G20.12)") dvirial
                  WRITE (unit_nr, *) "  Tr(P)/3 :  ", (dvirial(1, 1)+dvirial(2, 2)+dvirial(3, 3))/3._dp
                  WRITE (unit_nr, *)
               END IF
            END IF
         END IF

         IF (calculate_forces .OR. debugall .OR. dispersion_env%doabc) THEN
            DEALLOCATE (atom_of_kind, kind_of)
         END IF

         DEALLOCATE (dodisp, ghost, atomnumber, rcpbc, radd2, c6d2)

         IF (ASSOCIATED(dispersion_env%dftd_section)) THEN
            CALL cp_print_key_finished_output(unit_nr, logger, dispersion_env%dftd_section, &
                                              "PRINT_DFTD")
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_dispersion_pairpot

! **************************************************************************************************
!> \brief ...
!> \param cell ...
!> \param ncell ...
!> \retval hash ...
! **************************************************************************************************
   FUNCTION cellhash(cell, ncell) RESULT(hash)
      INTEGER, DIMENSION(3), INTENT(IN)                  :: cell, ncell
      INTEGER                                            :: hash

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cellhash', routineP = moduleN//':'//routineN

      INTEGER                                            :: ix, iy, iz, nx, ny, nz

      IF(.NOT.(ALL(ABS(cell) <= ncell)))CALL cp__a("qs_dispersion_pairpot.F",1705)

      ix = cell(1)
      IF (ix /= 0) THEN
         ix = 2*ABS(ix)-(1+SIGN(1, ix))/2
      END IF
      iy = cell(2)
      IF (iy /= 0) THEN
         iy = 2*ABS(iy)-(1+SIGN(1, iy))/2
      END IF
      iz = cell(3)
      IF (iz /= 0) THEN
         iz = 2*ABS(iz)-(1+SIGN(1, iz))/2
      END IF

      nx = 2*ncell(1)+1
      ny = 2*ncell(2)+1
      nz = 2*ncell(3)+1

      hash = ix*ny*nz+iy*nz+iz+1

   END FUNCTION cellhash
! **************************************************************************************************
!> \brief ...
!> \param z ...
!> \param c6 ...
!> \param r ...
!> \param found ...
! **************************************************************************************************
   SUBROUTINE dftd2_param(z, c6, r, found)

      INTEGER, INTENT(in)                                :: z
      REAL(KIND=dp), INTENT(inout)                       :: c6, r
      LOGICAL, INTENT(inout)                             :: found

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dftd2_param', routineP = moduleN//':'//routineN
      REAL(KIND=dp), DIMENSION(54), PARAMETER :: c6val = (/0.14_dp, 0.08_dp, 1.61_dp, 1.61_dp, &
         3.13_dp, 1.75_dp, 1.23_dp, 0.70_dp, 0.75_dp, 0.63_dp, 5.71_dp, 5.71_dp, 10.79_dp, 9.23_dp,&
         7.84_dp, 5.57_dp, 5.07_dp, 4.61_dp, 10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, &
         10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, 10.80_dp, 16.99_dp, 17.10_dp, &
         16.37_dp, 12.64_dp, 12.47_dp, 12.01_dp, 24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, &
         24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, 24.67_dp, 37.32_dp, 38.71_dp, &
         38.44_dp, 31.74_dp, 31.50_dp, 29.99_dp/)
      REAL(KIND=dp), DIMENSION(54), PARAMETER :: rval = (/1.001_dp, 1.012_dp, 0.825_dp, 1.408_dp, &
         1.485_dp, 1.452_dp, 1.397_dp, 1.342_dp, 1.287_dp, 1.243_dp, 1.144_dp, 1.364_dp, 1.639_dp, &
         1.716_dp, 1.705_dp, 1.683_dp, 1.639_dp, 1.595_dp, 1.485_dp, 1.474_dp, 1.562_dp, 1.562_dp, &
         1.562_dp, 1.562_dp, 1.562_dp, 1.562_dp, 1.562_dp, 1.562_dp, 1.562_dp, 1.562_dp, 1.650_dp, &
         1.727_dp, 1.760_dp, 1.771_dp, 1.749_dp, 1.727_dp, 1.628_dp, 1.606_dp, 1.639_dp, 1.639_dp, &
         1.639_dp, 1.639_dp, 1.639_dp, 1.639_dp, 1.639_dp, 1.639_dp, 1.639_dp, 1.639_dp, 1.672_dp, &
         1.804_dp, 1.881_dp, 1.892_dp, 1.892_dp, 1.881_dp/)

!
! GRIMME DISPERSION PARAMETERS
! Stefan Grimme, Semiempirical GGA-Type Density Functional Constructed
!                with a Long-Range Dispersion Correction, J. Comp. Chem. 27: 1787-1799 (2006)
! doi:10.1002/jcc.20495
!
! Conversion factor [Jnm^6mol^-1] -> [a.u.] : 17.34527758021901
! Conversion factor [A] -> [a.u.] : 1.889726132885643
!
! C6 values in [Jnm^6/mol]
! vdW radii [A]

      IF (z > 0 .AND. z <= 54) THEN
         found = .TRUE.
         c6 = c6val(z)*1000._dp*bohr**6/kjmol
         r = rval(z)*bohr
      ELSE
         found = .FALSE.
      END IF

   END SUBROUTINE dftd2_param
! **************************************************************************************************
!> \brief ...
!> \param c6ab ...
!> \param maxci ...
!> \param filename ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE dftd3_c6_param(c6ab, maxci, filename, para_env)

      REAL(KIND=dp), DIMENSION(:, :, :, :, :)            :: c6ab
      INTEGER, DIMENSION(:)                              :: maxci
      CHARACTER(LEN=*)                                   :: filename
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dftd3_c6_param', routineP = moduleN//':'//routineN

      INTEGER                                            :: funit, iadr, iat, jadr, jat, kk, nl, &
                                                            nlines, nn
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pars

      IF (para_env%source == para_env%mepos) THEN
         ! Read the DFT-D3 C6AB parameters from file "filename"
         CALL open_file(file_name=filename, unit_number=funit, file_form="FORMATTED")
         READ (funit, *) nl, nlines
      END IF
      CALL mp_bcast(nl, para_env%source, para_env%group)
      CALL mp_bcast(nlines, para_env%source, para_env%group)
      ALLOCATE (pars(nl))
      IF (para_env%source == para_env%mepos) THEN
         READ (funit, *) pars(1:nl)
         CALL close_file(unit_number=funit)
      END IF
      CALL mp_bcast(pars, para_env%source, para_env%group)

      ! Store C6AB coefficients in an array
      c6ab = -1._dp
      maxci = 0
      kk = 1
      DO nn = 1, nlines
         iat = NINT(pars(kk+1))
         jat = NINT(pars(kk+2))
         CALL limit(iat, jat, iadr, jadr)
         maxci(iat) = MAX(maxci(iat), iadr)
         maxci(jat) = MAX(maxci(jat), jadr)
         c6ab(iat, jat, iadr, jadr, 1) = pars(kk)
         c6ab(iat, jat, iadr, jadr, 2) = pars(kk+3)
         c6ab(iat, jat, iadr, jadr, 3) = pars(kk+4)

         c6ab(jat, iat, jadr, iadr, 1) = pars(kk)
         c6ab(jat, iat, jadr, iadr, 2) = pars(kk+4)
         c6ab(jat, iat, jadr, iadr, 3) = pars(kk+3)
         kk = (nn*5)+1
      ENDDO

      DEALLOCATE (pars)

   END SUBROUTINE dftd3_c6_param

! **************************************************************************************************
!> \brief ...
!> \param iat ...
!> \param jat ...
!> \param iadr ...
!> \param jadr ...
! **************************************************************************************************
   SUBROUTINE limit(iat, jat, iadr, jadr)
      INTEGER                                            :: iat, jat, iadr, jadr

      INTEGER                                            :: i

      iadr = 1
      jadr = 1
      i = 100
      DO WHILE (iat .GT. 100)
         iat = iat-100
         iadr = iadr+1
      ENDDO

      i = 100
      DO WHILE (jat .GT. 100)
         jat = jat-100
         jadr = jadr+1
      ENDDO
   END SUBROUTINE limit

! **************************************************************************************************
!> \brief ...
!> \param rout ...
!> \param rcov ...
!> \param r2r4 ...
! **************************************************************************************************
   SUBROUTINE setr0ab(rout, rcov, r2r4)
      ! set cut-off radii
      REAL(KIND=dp), DIMENSION(:, :)                     :: rout
      REAL(KIND=dp), DIMENSION(:)                        :: rcov, r2r4

      INTEGER                                            :: i, j, k
      REAL(KIND=dp), DIMENSION(4465)                     :: r0ab(4465)

      r0ab(1:70) = (/ &
                   2.1823, 1.8547, 1.7347, 2.9086, 2.5732, 3.4956, 2.3550 &
                   , 2.5095, 2.9802, 3.0982, 2.5141, 2.3917, 2.9977, 2.9484 &
                   , 3.2160, 2.4492, 2.2527, 3.1933, 3.0214, 2.9531, 2.9103 &
                   , 2.3667, 2.1328, 2.8784, 2.7660, 2.7776, 2.7063, 2.6225 &
                   , 2.1768, 2.0625, 2.6395, 2.6648, 2.6482, 2.5697, 2.4846 &
                   , 2.4817, 2.0646, 1.9891, 2.5086, 2.6908, 2.6233, 2.4770 &
                   , 2.3885, 2.3511, 2.2996, 1.9892, 1.9251, 2.4190, 2.5473 &
                   , 2.4994, 2.4091, 2.3176, 2.2571, 2.1946, 2.1374, 2.9898 &
                   , 2.6397, 3.6031, 3.1219, 3.7620, 3.2485, 2.9357, 2.7093 &
                   , 2.5781, 2.4839, 3.7082, 2.5129, 2.7321, 3.1052, 3.2962 &
                   /)
      r0ab(71:140) = (/ &
                     3.1331, 3.2000, 2.9586, 3.0822, 2.8582, 2.7120, 3.2570 &
                     , 3.4839, 2.8766, 2.7427, 3.2776, 3.2363, 3.5929, 3.2826 &
                     , 3.0911, 2.9369, 2.9030, 2.7789, 3.3921, 3.3970, 4.0106 &
                     , 2.8884, 2.6605, 3.7513, 3.1613, 3.3605, 3.3325, 3.0991 &
                     , 2.9297, 2.8674, 2.7571, 3.8129, 3.3266, 3.7105, 3.7917 &
                     , 2.8304, 2.5538, 3.3932, 3.1193, 3.1866, 3.1245, 3.0465 &
                     , 2.8727, 2.7664, 2.6926, 3.4608, 3.2984, 3.5142, 3.5418 &
                     , 3.5017, 2.6190, 2.4797, 3.1331, 3.0540, 3.0651, 2.9879 &
                     , 2.9054, 2.8805, 2.7330, 2.6331, 3.2096, 3.5668, 3.3684 &
                     , 3.3686, 3.3180, 3.3107, 2.4757, 2.4019, 2.9789, 3.1468 &
                     /)
      r0ab(141:210) = (/ &
                      2.9768, 2.8848, 2.7952, 2.7457, 2.6881, 2.5728, 3.0574 &
                      , 3.3264, 3.3562, 3.2529, 3.1916, 3.1523, 3.1046, 2.3725 &
                      , 2.3289, 2.8760, 2.9804, 2.9093, 2.8040, 2.7071, 2.6386 &
                      , 2.5720, 2.5139, 2.9517, 3.1606, 3.2085, 3.1692, 3.0982 &
                      , 3.0352, 2.9730, 2.9148, 3.2147, 2.8315, 3.8724, 3.4621 &
                      , 3.8823, 3.3760, 3.0746, 2.8817, 2.7552, 2.6605, 3.9740 &
                      , 3.6192, 3.6569, 3.9586, 3.6188, 3.3917, 3.2479, 3.1434 &
                      , 4.2411, 2.7597, 3.0588, 3.3474, 3.6214, 3.4353, 3.4729 &
                      , 3.2487, 3.3200, 3.0914, 2.9403, 3.4972, 3.7993, 3.6773 &
                      , 3.8678, 3.5808, 3.8243, 3.5826, 3.4156, 3.8765, 4.1035 &
                      /)
      r0ab(211:280) = (/ &
                      2.7361, 2.9765, 3.2475, 3.5004, 3.4185, 3.4378, 3.2084 &
                      , 3.2787, 3.0604, 2.9187, 3.4037, 3.6759, 3.6586, 3.8327 &
                      , 3.5372, 3.7665, 3.5310, 3.3700, 3.7788, 3.9804, 3.8903 &
                      , 2.6832, 2.9060, 3.2613, 3.4359, 3.3538, 3.3860, 3.1550 &
                      , 3.2300, 3.0133, 2.8736, 3.4024, 3.6142, 3.5979, 3.5295 &
                      , 3.4834, 3.7140, 3.4782, 3.3170, 3.7434, 3.9623, 3.8181 &
                      , 3.7642, 2.6379, 2.8494, 3.1840, 3.4225, 3.2771, 3.3401 &
                      , 3.1072, 3.1885, 2.9714, 2.8319, 3.3315, 3.5979, 3.5256 &
                      , 3.4980, 3.4376, 3.6714, 3.4346, 3.2723, 3.6859, 3.8985 &
                      , 3.7918, 3.7372, 3.7211, 2.9230, 2.6223, 3.4161, 2.8999 &
                      /)
      r0ab(281:350) = (/ &
                      3.0557, 3.3308, 3.0555, 2.8508, 2.7385, 2.6640, 3.5263 &
                      , 3.0277, 3.2990, 3.7721, 3.5017, 3.2751, 3.1368, 3.0435 &
                      , 3.7873, 3.2858, 3.2140, 3.1727, 3.2178, 3.4414, 2.5490 &
                      , 2.7623, 3.0991, 3.3252, 3.1836, 3.2428, 3.0259, 3.1225 &
                      , 2.9032, 2.7621, 3.2490, 3.5110, 3.4429, 3.3845, 3.3574 &
                      , 3.6045, 3.3658, 3.2013, 3.6110, 3.8241, 3.7090, 3.6496 &
                      , 3.6333, 3.0896, 3.5462, 2.4926, 2.7136, 3.0693, 3.2699 &
                      , 3.1272, 3.1893, 2.9658, 3.0972, 2.8778, 2.7358, 3.2206 &
                      , 3.4566, 3.3896, 3.3257, 3.2946, 3.5693, 3.3312, 3.1670 &
                      , 3.5805, 3.7711, 3.6536, 3.5927, 3.5775, 3.0411, 3.4885 &
                      /)
      r0ab(351:420) = (/ &
                      3.4421, 2.4667, 2.6709, 3.0575, 3.2357, 3.0908, 3.1537 &
                      , 2.9235, 3.0669, 2.8476, 2.7054, 3.2064, 3.4519, 3.3593 &
                      , 3.2921, 3.2577, 3.2161, 3.2982, 3.1339, 3.5606, 3.7582 &
                      , 3.6432, 3.5833, 3.5691, 3.0161, 3.4812, 3.4339, 3.4327 &
                      , 2.4515, 2.6338, 3.0511, 3.2229, 3.0630, 3.1265, 2.8909 &
                      , 3.0253, 2.8184, 2.6764, 3.1968, 3.4114, 3.3492, 3.2691 &
                      , 3.2320, 3.1786, 3.2680, 3.1036, 3.5453, 3.7259, 3.6090 &
                      , 3.5473, 3.5327, 3.0018, 3.4413, 3.3907, 3.3593, 3.3462 &
                      , 2.4413, 2.6006, 3.0540, 3.1987, 3.0490, 3.1058, 2.8643 &
                      , 2.9948, 2.7908, 2.6491, 3.1950, 3.3922, 3.3316, 3.2585 &
                      /)
      r0ab(421:490) = (/ &
                      3.2136, 3.1516, 3.2364, 3.0752, 3.5368, 3.7117, 3.5941 &
                      , 3.5313, 3.5164, 2.9962, 3.4225, 3.3699, 3.3370, 3.3234 &
                      , 3.3008, 2.4318, 2.5729, 3.0416, 3.1639, 3.0196, 3.0843 &
                      , 2.8413, 2.7436, 2.7608, 2.6271, 3.1811, 3.3591, 3.3045 &
                      , 3.2349, 3.1942, 3.1291, 3.2111, 3.0534, 3.5189, 3.6809 &
                      , 3.5635, 3.5001, 3.4854, 2.9857, 3.3897, 3.3363, 3.3027 &
                      , 3.2890, 3.2655, 3.2309, 2.8502, 2.6934, 3.2467, 3.1921 &
                      , 3.5663, 3.2541, 3.0571, 2.9048, 2.8657, 2.7438, 3.3547 &
                      , 3.3510, 3.9837, 3.6871, 3.4862, 3.3389, 3.2413, 3.1708 &
                      , 3.6096, 3.6280, 3.6860, 3.5568, 3.4836, 3.2868, 3.3994 &
                      /)
      r0ab(491:560) = (/ &
                      3.3476, 3.3170, 3.2950, 3.2874, 3.2606, 3.9579, 2.9226 &
                      , 2.6838, 3.7867, 3.1732, 3.3872, 3.3643, 3.1267, 2.9541 &
                      , 2.8505, 2.7781, 3.8475, 3.3336, 3.7359, 3.8266, 3.5733 &
                      , 3.3959, 3.2775, 3.1915, 3.9878, 3.8816, 3.5810, 3.5364 &
                      , 3.5060, 3.8097, 3.3925, 3.3348, 3.3019, 3.2796, 3.2662 &
                      , 3.2464, 3.7136, 3.8619, 2.9140, 2.6271, 3.4771, 3.1774 &
                      , 3.2560, 3.1970, 3.1207, 2.9406, 2.8322, 2.7571, 3.5455 &
                      , 3.3514, 3.5837, 3.6177, 3.5816, 3.3902, 3.2604, 3.1652 &
                      , 3.7037, 3.6283, 3.5858, 3.5330, 3.4884, 3.5789, 3.4094 &
                      , 3.3473, 3.3118, 3.2876, 3.2707, 3.2521, 3.5570, 3.6496 &
                      /)
      r0ab(561:630) = (/ &
                      3.6625, 2.7300, 2.5870, 3.2471, 3.1487, 3.1667, 3.0914 &
                      , 3.0107, 2.9812, 2.8300, 2.7284, 3.3259, 3.3182, 3.4707 &
                      , 3.4748, 3.4279, 3.4182, 3.2547, 3.1353, 3.5116, 3.9432 &
                      , 3.8828, 3.8303, 3.7880, 3.3760, 3.7218, 3.3408, 3.3059 &
                      , 3.2698, 3.2446, 3.2229, 3.4422, 3.5023, 3.5009, 3.5268 &
                      , 2.6026, 2.5355, 3.1129, 3.2863, 3.1029, 3.0108, 2.9227 &
                      , 2.8694, 2.8109, 2.6929, 3.1958, 3.4670, 3.4018, 3.3805 &
                      , 3.3218, 3.2815, 3.2346, 3.0994, 3.3937, 3.7266, 3.6697 &
                      , 3.6164, 3.5730, 3.2522, 3.5051, 3.4686, 3.4355, 3.4084 &
                      , 3.3748, 3.3496, 3.3692, 3.4052, 3.3910, 3.3849, 3.3662 &
                      /)
      r0ab(631:700) = (/ &
                      2.5087, 2.4814, 3.0239, 3.1312, 3.0535, 2.9457, 2.8496 &
                      , 2.7780, 2.7828, 2.6532, 3.1063, 3.3143, 3.3549, 3.3120 &
                      , 3.2421, 3.1787, 3.1176, 3.0613, 3.3082, 3.5755, 3.5222 &
                      , 3.4678, 3.4231, 3.1684, 3.3528, 3.3162, 3.2827, 3.2527 &
                      , 3.2308, 3.2029, 3.3173, 3.3343, 3.3092, 3.2795, 3.2452 &
                      , 3.2096, 3.2893, 2.8991, 4.0388, 3.6100, 3.9388, 3.4475 &
                      , 3.1590, 2.9812, 2.8586, 2.7683, 4.1428, 3.7911, 3.8225 &
                      , 4.0372, 3.7059, 3.4935, 3.3529, 3.2492, 4.4352, 4.0826 &
                      , 3.9733, 3.9254, 3.8646, 3.9315, 3.7837, 3.7465, 3.7211 &
                      , 3.7012, 3.6893, 3.6676, 3.7736, 4.0660, 3.7926, 3.6158 &
                      /)
      r0ab(701:770) = (/ &
                      3.5017, 3.4166, 4.6176, 2.8786, 3.1658, 3.5823, 3.7689 &
                      , 3.5762, 3.5789, 3.3552, 3.4004, 3.1722, 3.0212, 3.7241 &
                      , 3.9604, 3.8500, 3.9844, 3.7035, 3.9161, 3.6751, 3.5075 &
                      , 4.1151, 4.2877, 4.1579, 4.1247, 4.0617, 3.4874, 3.9848 &
                      , 3.9280, 3.9079, 3.8751, 3.8604, 3.8277, 3.8002, 3.9981 &
                      , 3.7544, 4.0371, 3.8225, 3.6718, 4.3092, 4.4764, 2.8997 &
                      , 3.0953, 3.4524, 3.6107, 3.6062, 3.5783, 3.3463, 3.3855 &
                      , 3.1746, 3.0381, 3.6019, 3.7938, 3.8697, 3.9781, 3.6877 &
                      , 3.8736, 3.6451, 3.4890, 3.9858, 4.1179, 4.0430, 3.9563 &
                      , 3.9182, 3.4002, 3.8310, 3.7716, 3.7543, 3.7203, 3.7053 &
                      /)
      r0ab(771:840) = (/ &
                      3.6742, 3.8318, 3.7631, 3.7392, 3.9892, 3.7832, 3.6406 &
                      , 4.1701, 4.3016, 4.2196, 2.8535, 3.0167, 3.3978, 3.5363 &
                      , 3.5393, 3.5301, 3.2960, 3.3352, 3.1287, 2.9967, 3.6659 &
                      , 3.7239, 3.8070, 3.7165, 3.6368, 3.8162, 3.5885, 3.4336 &
                      , 3.9829, 4.0529, 3.9584, 3.9025, 3.8607, 3.3673, 3.7658 &
                      , 3.7035, 3.6866, 3.6504, 3.6339, 3.6024, 3.7708, 3.7283 &
                      , 3.6896, 3.9315, 3.7250, 3.5819, 4.1457, 4.2280, 4.1130 &
                      , 4.0597, 3.0905, 2.7998, 3.6448, 3.0739, 3.2996, 3.5262 &
                      , 3.2559, 3.0518, 2.9394, 2.8658, 3.7514, 3.2295, 3.5643 &
                      , 3.7808, 3.6931, 3.4723, 3.3357, 3.2429, 4.0280, 3.5589 &
                      /)
      r0ab(841:910) = (/ &
                      3.4636, 3.4994, 3.4309, 3.6177, 3.2946, 3.2376, 3.2050 &
                      , 3.1847, 3.1715, 3.1599, 3.5555, 3.8111, 3.7693, 3.5718 &
                      , 3.4498, 3.3662, 4.1608, 3.7417, 3.6536, 3.6154, 3.8596 &
                      , 3.0301, 2.7312, 3.5821, 3.0473, 3.2137, 3.4679, 3.1975 &
                      , 2.9969, 2.8847, 2.8110, 3.6931, 3.2076, 3.4943, 3.5956 &
                      , 3.6379, 3.4190, 3.2808, 3.1860, 3.9850, 3.5105, 3.4330 &
                      , 3.3797, 3.4155, 3.6033, 3.2737, 3.2145, 3.1807, 3.1596 &
                      , 3.1461, 3.1337, 3.4812, 3.6251, 3.7152, 3.5201, 3.3966 &
                      , 3.3107, 4.1128, 3.6899, 3.6082, 3.5604, 3.7834, 3.7543 &
                      , 2.9189, 2.6777, 3.4925, 2.9648, 3.1216, 3.2940, 3.0975 &
                      /)
      r0ab(911:980) = (/ &
                      2.9757, 2.8493, 2.7638, 3.6085, 3.1214, 3.4006, 3.4793 &
                      , 3.5147, 3.3806, 3.2356, 3.1335, 3.9144, 3.4183, 3.3369 &
                      , 3.2803, 3.2679, 3.4871, 3.1714, 3.1521, 3.1101, 3.0843 &
                      , 3.0670, 3.0539, 3.3890, 3.5086, 3.5895, 3.4783, 3.3484 &
                      , 3.2559, 4.0422, 3.5967, 3.5113, 3.4576, 3.6594, 3.6313 &
                      , 3.5690, 2.8578, 2.6334, 3.4673, 2.9245, 3.0732, 3.2435 &
                      , 3.0338, 2.9462, 2.8143, 2.7240, 3.5832, 3.0789, 3.3617 &
                      , 3.4246, 3.4505, 3.3443, 3.1964, 3.0913, 3.8921, 3.3713 &
                      , 3.2873, 3.2281, 3.2165, 3.4386, 3.1164, 3.1220, 3.0761 &
                      , 3.0480, 3.0295, 3.0155, 3.3495, 3.4543, 3.5260, 3.4413 &
                      /)
      r0ab(981:1050) = (/ &
                       3.3085, 3.2134, 4.0170, 3.5464, 3.4587, 3.4006, 3.6027 &
                       , 3.5730, 3.4945, 3.4623, 2.8240, 2.5960, 3.4635, 2.9032 &
                       , 3.0431, 3.2115, 2.9892, 2.9148, 2.7801, 2.6873, 3.5776 &
                       , 3.0568, 3.3433, 3.3949, 3.4132, 3.3116, 3.1616, 3.0548 &
                       , 3.8859, 3.3719, 3.2917, 3.2345, 3.2274, 3.4171, 3.1293 &
                       , 3.0567, 3.0565, 3.0274, 3.0087, 2.9939, 3.3293, 3.4249 &
                       , 3.4902, 3.4091, 3.2744, 3.1776, 4.0078, 3.5374, 3.4537 &
                       , 3.3956, 3.5747, 3.5430, 3.4522, 3.4160, 3.3975, 2.8004 &
                       , 2.5621, 3.4617, 2.9154, 3.0203, 3.1875, 2.9548, 2.8038 &
                       , 2.7472, 2.6530, 3.5736, 3.0584, 3.3304, 3.3748, 3.3871 &
                       /)
      r0ab(1051:1120) = (/ &
                        3.2028, 3.1296, 3.0214, 3.8796, 3.3337, 3.2492, 3.1883 &
                        , 3.1802, 3.4050, 3.0756, 3.0478, 3.0322, 3.0323, 3.0163 &
                        , 3.0019, 3.3145, 3.4050, 3.4656, 3.3021, 3.2433, 3.1453 &
                        , 3.9991, 3.5017, 3.4141, 3.3520, 3.5583, 3.5251, 3.4243 &
                        , 3.3851, 3.3662, 3.3525, 2.7846, 2.5324, 3.4652, 2.8759 &
                        , 3.0051, 3.1692, 2.9273, 2.7615, 2.7164, 2.6212, 3.5744 &
                        , 3.0275, 3.3249, 3.3627, 3.3686, 3.1669, 3.0584, 2.9915 &
                        , 3.8773, 3.3099, 3.2231, 3.1600, 3.1520, 3.4023, 3.0426 &
                        , 3.0099, 2.9920, 2.9809, 2.9800, 2.9646, 3.3068, 3.3930 &
                        , 3.4486, 3.2682, 3.1729, 3.1168, 3.9952, 3.4796, 3.3901 &
                        /)
      r0ab(1121:1190) = (/ &
                        3.3255, 3.5530, 3.5183, 3.4097, 3.3683, 3.3492, 3.3360 &
                        , 3.3308, 2.5424, 2.6601, 3.2555, 3.2807, 3.1384, 3.1737 &
                        , 2.9397, 2.8429, 2.8492, 2.7225, 3.3875, 3.4910, 3.4520 &
                        , 3.3608, 3.3036, 3.2345, 3.2999, 3.1487, 3.7409, 3.8392 &
                        , 3.7148, 3.6439, 3.6182, 3.1753, 3.5210, 3.4639, 3.4265 &
                        , 3.4075, 3.3828, 3.3474, 3.4071, 3.3754, 3.3646, 3.3308 &
                        , 3.4393, 3.2993, 3.8768, 3.9891, 3.8310, 3.7483, 3.3417 &
                        , 3.3019, 3.2250, 3.1832, 3.1578, 3.1564, 3.1224, 3.4620 &
                        , 2.9743, 2.8058, 3.4830, 3.3474, 3.6863, 3.3617, 3.1608 &
                        , 3.0069, 2.9640, 2.8427, 3.5885, 3.5219, 4.1314, 3.8120 &
                        /)
      r0ab(1191:1260) = (/ &
                        3.6015, 3.4502, 3.3498, 3.2777, 3.8635, 3.8232, 3.8486 &
                        , 3.7215, 3.6487, 3.4724, 3.5627, 3.5087, 3.4757, 3.4517 &
                        , 3.4423, 3.4139, 4.1028, 3.8388, 3.6745, 3.5562, 3.4806 &
                        , 3.4272, 4.0182, 3.9991, 4.0007, 3.9282, 3.7238, 3.6498 &
                        , 3.5605, 3.5211, 3.5009, 3.4859, 3.4785, 3.5621, 4.2623 &
                        , 3.0775, 2.8275, 4.0181, 3.3385, 3.5379, 3.5036, 3.2589 &
                        , 3.0804, 3.0094, 2.9003, 4.0869, 3.5088, 3.9105, 3.9833 &
                        , 3.7176, 3.5323, 3.4102, 3.3227, 4.2702, 4.0888, 3.7560 &
                        , 3.7687, 3.6681, 3.6405, 3.5569, 3.4990, 3.4659, 3.4433 &
                        , 3.4330, 3.4092, 3.8867, 4.0190, 3.7961, 3.6412, 3.5405 &
                        /)
      r0ab(1261:1330) = (/ &
                        3.4681, 4.3538, 4.2136, 3.9381, 3.8912, 3.9681, 3.7909 &
                        , 3.6774, 3.6262, 3.5999, 3.5823, 3.5727, 3.5419, 4.0245 &
                        , 4.1874, 3.0893, 2.7917, 3.7262, 3.3518, 3.4241, 3.5433 &
                        , 3.2773, 3.0890, 2.9775, 2.9010, 3.8048, 3.5362, 3.7746 &
                        , 3.7911, 3.7511, 3.5495, 3.4149, 3.3177, 4.0129, 3.8370 &
                        , 3.7739, 3.7125, 3.7152, 3.7701, 3.5813, 3.5187, 3.4835 &
                        , 3.4595, 3.4439, 3.4242, 3.7476, 3.8239, 3.8346, 3.6627 &
                        , 3.5479, 3.4639, 4.1026, 3.9733, 3.9292, 3.8667, 3.9513 &
                        , 3.8959, 3.7698, 3.7089, 3.6765, 3.6548, 3.6409, 3.5398 &
                        , 3.8759, 3.9804, 4.0150, 2.9091, 2.7638, 3.5066, 3.3377 &
                        /)
      r0ab(1331:1400) = (/ &
                        3.3481, 3.2633, 3.1810, 3.1428, 2.9872, 2.8837, 3.5929 &
                        , 3.5183, 3.6729, 3.6596, 3.6082, 3.5927, 3.4224, 3.2997 &
                        , 3.8190, 4.1865, 4.1114, 4.0540, 3.6325, 3.5697, 3.5561 &
                        , 3.5259, 3.4901, 3.4552, 3.4315, 3.4091, 3.6438, 3.6879 &
                        , 3.6832, 3.7043, 3.5557, 3.4466, 3.9203, 4.2919, 4.2196 &
                        , 4.1542, 3.7573, 3.7039, 3.6546, 3.6151, 3.5293, 3.4849 &
                        , 3.4552, 3.5192, 3.7673, 3.8359, 3.8525, 3.8901, 2.7806 &
                        , 2.7209, 3.3812, 3.4958, 3.2913, 3.1888, 3.0990, 3.0394 &
                        , 2.9789, 2.8582, 3.4716, 3.6883, 3.6105, 3.5704, 3.5059 &
                        , 3.4619, 3.4138, 3.2742, 3.7080, 3.9773, 3.9010, 3.8409 &
                        /)
      r0ab(1401:1470) = (/ &
                        3.7944, 3.4465, 3.7235, 3.6808, 3.6453, 3.6168, 3.5844 &
                        , 3.5576, 3.5772, 3.5959, 3.5768, 3.5678, 3.5486, 3.4228 &
                        , 3.8107, 4.0866, 4.0169, 3.9476, 3.6358, 3.5800, 3.5260 &
                        , 3.4838, 3.4501, 3.4204, 3.3553, 3.6487, 3.6973, 3.7398 &
                        , 3.7405, 3.7459, 3.7380, 2.6848, 2.6740, 3.2925, 3.3386 &
                        , 3.2473, 3.1284, 3.0301, 2.9531, 2.9602, 2.8272, 3.3830 &
                        , 3.5358, 3.5672, 3.5049, 3.4284, 3.3621, 3.3001, 3.2451 &
                        , 3.6209, 3.8299, 3.7543, 3.6920, 3.6436, 3.3598, 3.5701 &
                        , 3.5266, 3.4904, 3.4590, 3.4364, 3.4077, 3.5287, 3.5280 &
                        , 3.4969, 3.4650, 3.4304, 3.3963, 3.7229, 3.9402, 3.8753 &
                        /)
      r0ab(1471:1540) = (/ &
                        3.8035, 3.5499, 3.4913, 3.4319, 3.3873, 3.3520, 3.3209 &
                        , 3.2948, 3.5052, 3.6465, 3.6696, 3.6577, 3.6388, 3.6142 &
                        , 3.5889, 3.3968, 3.0122, 4.2241, 3.7887, 4.0049, 3.5384 &
                        , 3.2698, 3.1083, 2.9917, 2.9057, 4.3340, 3.9900, 4.6588 &
                        , 4.1278, 3.8125, 3.6189, 3.4851, 3.3859, 4.6531, 4.3134 &
                        , 4.2258, 4.1309, 4.0692, 4.0944, 3.9850, 3.9416, 3.9112 &
                        , 3.8873, 3.8736, 3.8473, 4.6027, 4.1538, 3.8994, 3.7419 &
                        , 3.6356, 3.5548, 4.8353, 4.5413, 4.3891, 4.3416, 4.3243 &
                        , 4.2753, 4.2053, 4.1790, 4.1685, 4.1585, 4.1536, 4.0579 &
                        , 4.1980, 4.4564, 4.2192, 4.0528, 3.9489, 3.8642, 5.0567 &
                        /)
      r0ab(1541:1610) = (/ &
                        3.0630, 3.3271, 4.0432, 4.0046, 4.1555, 3.7426, 3.5130 &
                        , 3.5174, 3.2884, 3.1378, 4.1894, 4.2321, 4.1725, 4.1833 &
                        , 3.8929, 4.0544, 3.8118, 3.6414, 4.6373, 4.6268, 4.4750 &
                        , 4.4134, 4.3458, 3.8582, 4.2583, 4.1898, 4.1562, 4.1191 &
                        , 4.1069, 4.0639, 4.1257, 4.1974, 3.9532, 4.1794, 3.9660 &
                        , 3.8130, 4.8160, 4.8272, 4.6294, 4.5840, 4.0770, 4.0088 &
                        , 3.9103, 3.8536, 3.8324, 3.7995, 3.7826, 4.2294, 4.3380 &
                        , 4.4352, 4.1933, 4.4580, 4.2554, 4.1072, 5.0454, 5.1814 &
                        , 3.0632, 3.2662, 3.6432, 3.8088, 3.7910, 3.7381, 3.5093 &
                        , 3.5155, 3.3047, 3.1681, 3.7871, 3.9924, 4.0637, 4.1382 &
                        /)
      r0ab(1611:1680) = (/ &
                        3.8591, 4.0164, 3.7878, 3.6316, 4.1741, 4.3166, 4.2395 &
                        , 4.1831, 4.1107, 3.5857, 4.0270, 3.9676, 3.9463, 3.9150 &
                        , 3.9021, 3.8708, 4.0240, 4.1551, 3.9108, 4.1337, 3.9289 &
                        , 3.7873, 4.3666, 4.5080, 4.4232, 4.3155, 3.8461, 3.8007 &
                        , 3.6991, 3.6447, 3.6308, 3.5959, 3.5749, 4.0359, 4.3124 &
                        , 4.3539, 4.1122, 4.3772, 4.1785, 4.0386, 4.7004, 4.8604 &
                        , 4.6261, 2.9455, 3.2470, 3.6108, 3.8522, 3.6625, 3.6598 &
                        , 3.4411, 3.4660, 3.2415, 3.0944, 3.7514, 4.0397, 3.9231 &
                        , 4.0561, 3.7860, 3.9845, 3.7454, 3.5802, 4.1366, 4.3581 &
                        , 4.2351, 4.2011, 4.1402, 3.5381, 4.0653, 4.0093, 3.9883 &
                        /)
      r0ab(1681:1750) = (/ &
                        3.9570, 3.9429, 3.9112, 3.8728, 4.0682, 3.8351, 4.1054 &
                        , 3.8928, 3.7445, 4.3415, 4.5497, 4.3833, 4.3122, 3.8051 &
                        , 3.7583, 3.6622, 3.6108, 3.5971, 3.5628, 3.5408, 4.0780 &
                        , 4.0727, 4.2836, 4.0553, 4.3647, 4.1622, 4.0178, 4.5802 &
                        , 4.9125, 4.5861, 4.6201, 2.9244, 3.2241, 3.5848, 3.8293 &
                        , 3.6395, 3.6400, 3.4204, 3.4499, 3.2253, 3.0779, 3.7257 &
                        , 4.0170, 3.9003, 4.0372, 3.7653, 3.9672, 3.7283, 3.5630 &
                        , 4.1092, 4.3347, 4.2117, 4.1793, 4.1179, 3.5139, 4.0426 &
                        , 3.9867, 3.9661, 3.9345, 3.9200, 3.8883, 3.8498, 4.0496 &
                        , 3.8145, 4.0881, 3.8756, 3.7271, 4.3128, 4.5242, 4.3578 &
                        /)
      r0ab(1751:1820) = (/ &
                        4.2870, 3.7796, 3.7318, 3.6364, 3.5854, 3.5726, 3.5378 &
                        , 3.5155, 4.0527, 4.0478, 4.2630, 4.0322, 4.3449, 4.1421 &
                        , 3.9975, 4.5499, 4.8825, 4.5601, 4.5950, 4.5702, 2.9046 &
                        , 3.2044, 3.5621, 3.8078, 3.6185, 3.6220, 3.4019, 3.4359 &
                        , 3.2110, 3.0635, 3.7037, 3.9958, 3.8792, 4.0194, 3.7460 &
                        , 3.9517, 3.7128, 3.5474, 4.0872, 4.3138, 4.1906, 4.1593 &
                        , 4.0973, 3.4919, 4.0216, 3.9657, 3.9454, 3.9134, 3.8986 &
                        , 3.8669, 3.8289, 4.0323, 3.7954, 4.0725, 3.8598, 3.7113 &
                        , 4.2896, 4.5021, 4.3325, 4.2645, 3.7571, 3.7083, 3.6136 &
                        , 3.5628, 3.5507, 3.5155, 3.4929, 4.0297, 4.0234, 4.2442 &
                        /)
      r0ab(1821:1890) = (/ &
                        4.0112, 4.3274, 4.1240, 3.9793, 4.5257, 4.8568, 4.5353 &
                        , 4.5733, 4.5485, 4.5271, 2.8878, 3.1890, 3.5412, 3.7908 &
                        , 3.5974, 3.6078, 3.3871, 3.4243, 3.1992, 3.0513, 3.6831 &
                        , 3.9784, 3.8579, 4.0049, 3.7304, 3.9392, 3.7002, 3.5347 &
                        , 4.0657, 4.2955, 4.1705, 4.1424, 4.0800, 3.4717, 4.0043 &
                        , 3.9485, 3.9286, 3.8965, 3.8815, 3.8500, 3.8073, 4.0180 &
                        , 3.7796, 4.0598, 3.8470, 3.6983, 4.2678, 4.4830, 4.3132 &
                        , 4.2444, 3.7370, 3.6876, 3.5935, 3.5428, 3.5314, 3.4958 &
                        , 3.4730, 4.0117, 4.0043, 4.2287, 3.9939, 4.3134, 4.1096 &
                        , 3.9646, 4.5032, 4.8356, 4.5156, 4.5544, 4.5297, 4.5083 &
                        /)
      r0ab(1891:1960) = (/ &
                        4.4896, 2.8709, 3.1737, 3.5199, 3.7734, 3.5802, 3.5934 &
                        , 3.3724, 3.4128, 3.1877, 3.0396, 3.6624, 3.9608, 3.8397 &
                        , 3.9893, 3.7145, 3.9266, 3.6877, 3.5222, 4.0448, 4.2771 &
                        , 4.1523, 4.1247, 4.0626, 3.4530, 3.9866, 3.9310, 3.9115 &
                        , 3.8792, 3.8641, 3.8326, 3.7892, 4.0025, 3.7636, 4.0471 &
                        , 3.8343, 3.6854, 4.2464, 4.4635, 4.2939, 4.2252, 3.7169 &
                        , 3.6675, 3.5739, 3.5235, 3.5126, 3.4768, 3.4537, 3.9932 &
                        , 3.9854, 4.2123, 3.9765, 4.2992, 4.0951, 3.9500, 4.4811 &
                        , 4.8135, 4.4959, 4.5351, 4.5105, 4.4891, 4.4705, 4.4515 &
                        , 2.8568, 3.1608, 3.5050, 3.7598, 3.5665, 3.5803, 3.3601 &
                        /)
      r0ab(1961:2030) = (/ &
                        3.4031, 3.1779, 3.0296, 3.6479, 3.9471, 3.8262, 3.9773 &
                        , 3.7015, 3.9162, 3.6771, 3.5115, 4.0306, 4.2634, 4.1385 &
                        , 4.1116, 4.0489, 3.4366, 3.9732, 3.9176, 3.8983, 3.8659 &
                        , 3.8507, 3.8191, 3.7757, 3.9907, 3.7506, 4.0365, 3.8235 &
                        , 3.6745, 4.2314, 4.4490, 4.2792, 4.2105, 3.7003, 3.6510 &
                        , 3.5578, 3.5075, 3.4971, 3.4609, 3.4377, 3.9788, 3.9712 &
                        , 4.1997, 3.9624, 4.2877, 4.0831, 3.9378, 4.4655, 4.7974 &
                        , 4.4813, 4.5209, 4.4964, 4.4750, 4.4565, 4.4375, 4.4234 &
                        , 2.6798, 3.0151, 3.2586, 3.5292, 3.5391, 3.4902, 3.2887 &
                        , 3.3322, 3.1228, 2.9888, 3.4012, 3.7145, 3.7830, 3.6665 &
                        /)
      r0ab(2031:2100) = (/ &
                        3.5898, 3.8077, 3.5810, 3.4265, 3.7726, 4.0307, 3.9763 &
                        , 3.8890, 3.8489, 3.2706, 3.7595, 3.6984, 3.6772, 3.6428 &
                        , 3.6243, 3.5951, 3.7497, 3.6775, 3.6364, 3.9203, 3.7157 &
                        , 3.5746, 3.9494, 4.2076, 4.1563, 4.0508, 3.5329, 3.4780 &
                        , 3.3731, 3.3126, 3.2846, 3.2426, 3.2135, 3.7491, 3.9006 &
                        , 3.8332, 3.8029, 4.1436, 3.9407, 3.7998, 4.1663, 4.5309 &
                        , 4.3481, 4.2911, 4.2671, 4.2415, 4.2230, 4.2047, 4.1908 &
                        , 4.1243, 2.5189, 2.9703, 3.3063, 3.6235, 3.4517, 3.3989 &
                        , 3.2107, 3.2434, 3.0094, 2.8580, 3.4253, 3.8157, 3.7258 &
                        , 3.6132, 3.5297, 3.7566, 3.5095, 3.3368, 3.7890, 4.1298 &
                        /)
      r0ab(2101:2170) = (/ &
                        4.0190, 3.9573, 3.9237, 3.2677, 3.8480, 3.8157, 3.7656 &
                        , 3.7317, 3.7126, 3.6814, 3.6793, 3.6218, 3.5788, 3.8763 &
                        , 3.6572, 3.5022, 3.9737, 4.3255, 4.1828, 4.1158, 3.5078 &
                        , 3.4595, 3.3600, 3.3088, 3.2575, 3.2164, 3.1856, 3.8522 &
                        , 3.8665, 3.8075, 3.7772, 4.1391, 3.9296, 3.7772, 4.2134 &
                        , 4.7308, 4.3787, 4.3894, 4.3649, 4.3441, 4.3257, 4.3073 &
                        , 4.2941, 4.1252, 4.2427, 3.0481, 2.9584, 3.6919, 3.5990 &
                        , 3.8881, 3.4209, 3.1606, 3.1938, 2.9975, 2.8646, 3.8138 &
                        , 3.7935, 3.7081, 3.9155, 3.5910, 3.4808, 3.4886, 3.3397 &
                        , 4.1336, 4.1122, 3.9888, 3.9543, 3.8917, 3.5894, 3.8131 &
                        /)
      r0ab(2171:2240) = (/ &
                        3.7635, 3.7419, 3.7071, 3.6880, 3.6574, 3.6546, 3.9375 &
                        , 3.6579, 3.5870, 3.6361, 3.5039, 4.3149, 4.2978, 4.1321 &
                        , 4.1298, 3.8164, 3.7680, 3.7154, 3.6858, 3.6709, 3.6666 &
                        , 3.6517, 3.8174, 3.8608, 4.1805, 3.9102, 3.8394, 3.8968 &
                        , 3.7673, 4.5274, 4.6682, 4.3344, 4.3639, 4.3384, 4.3162 &
                        , 4.2972, 4.2779, 4.2636, 4.0253, 4.1168, 4.1541, 2.8136 &
                        , 3.0951, 3.4635, 3.6875, 3.4987, 3.5183, 3.2937, 3.3580 &
                        , 3.1325, 2.9832, 3.6078, 3.8757, 3.7616, 3.9222, 3.6370 &
                        , 3.8647, 3.6256, 3.4595, 3.9874, 4.1938, 4.0679, 4.0430 &
                        , 3.9781, 3.3886, 3.9008, 3.8463, 3.8288, 3.7950, 3.7790 &
                        /)
      r0ab(2241:2310) = (/ &
                        3.7472, 3.7117, 3.9371, 3.6873, 3.9846, 3.7709, 3.6210 &
                        , 4.1812, 4.3750, 4.2044, 4.1340, 3.6459, 3.5929, 3.5036 &
                        , 3.4577, 3.4528, 3.4146, 3.3904, 3.9014, 3.9031, 4.1443 &
                        , 3.8961, 4.2295, 4.0227, 3.8763, 4.4086, 4.7097, 4.4064 &
                        , 4.4488, 4.4243, 4.4029, 4.3842, 4.3655, 4.3514, 4.1162 &
                        , 4.2205, 4.1953, 4.2794, 2.8032, 3.0805, 3.4519, 3.6700 &
                        , 3.4827, 3.5050, 3.2799, 3.3482, 3.1233, 2.9747, 3.5971 &
                        , 3.8586, 3.7461, 3.9100, 3.6228, 3.8535, 3.6147, 3.4490 &
                        , 3.9764, 4.1773, 4.0511, 4.0270, 3.9614, 3.3754, 3.8836 &
                        , 3.8291, 3.8121, 3.7780, 3.7619, 3.7300, 3.6965, 3.9253 &
                        /)
      r0ab(2311:2380) = (/ &
                        3.6734, 3.9733, 3.7597, 3.6099, 4.1683, 4.3572, 4.1862 &
                        , 4.1153, 3.6312, 3.5772, 3.4881, 3.4429, 3.4395, 3.4009 &
                        , 3.3766, 3.8827, 3.8868, 4.1316, 3.8807, 4.2164, 4.0092 &
                        , 3.8627, 4.3936, 4.6871, 4.3882, 4.4316, 4.4073, 4.3858 &
                        , 4.3672, 4.3485, 4.3344, 4.0984, 4.2036, 4.1791, 4.2622 &
                        , 4.2450, 2.7967, 3.0689, 3.4445, 3.6581, 3.4717, 3.4951 &
                        , 3.2694, 3.3397, 3.1147, 2.9661, 3.5898, 3.8468, 3.7358 &
                        , 3.9014, 3.6129, 3.8443, 3.6054, 3.4396, 3.9683, 4.1656 &
                        , 4.0394, 4.0158, 3.9498, 3.3677, 3.8718, 3.8164, 3.8005 &
                        , 3.7662, 3.7500, 3.7181, 3.6863, 3.9170, 3.6637, 3.9641 &
                        /)
      r0ab(2381:2450) = (/ &
                        3.7503, 3.6004, 4.1590, 4.3448, 4.1739, 4.1029, 3.6224 &
                        , 3.5677, 3.4785, 3.4314, 3.4313, 3.3923, 3.3680, 3.8698 &
                        , 3.8758, 4.1229, 3.8704, 4.2063, 3.9987, 3.8519, 4.3832 &
                        , 4.6728, 4.3759, 4.4195, 4.3952, 4.3737, 4.3551, 4.3364 &
                        , 4.3223, 4.0861, 4.1911, 4.1676, 4.2501, 4.2329, 4.2208 &
                        , 2.7897, 3.0636, 3.4344, 3.6480, 3.4626, 3.4892, 3.2626 &
                        , 3.3344, 3.1088, 2.9597, 3.5804, 3.8359, 3.7251, 3.8940 &
                        , 3.6047, 3.8375, 3.5990, 3.4329, 3.9597, 4.1542, 4.0278 &
                        , 4.0048, 3.9390, 3.3571, 3.8608, 3.8056, 3.7899, 3.7560 &
                        , 3.7400, 3.7081, 3.6758, 3.9095, 3.6552, 3.9572, 3.7436 &
                        /)
      r0ab(2451:2520) = (/ &
                        3.5933, 4.1508, 4.3337, 4.1624, 4.0916, 3.6126, 3.5582 &
                        , 3.4684, 3.4212, 3.4207, 3.3829, 3.3586, 3.8604, 3.8658 &
                        , 4.1156, 3.8620, 4.1994, 3.9917, 3.8446, 4.3750, 4.6617 &
                        , 4.3644, 4.4083, 4.3840, 4.3625, 4.3439, 4.3253, 4.3112 &
                        , 4.0745, 4.1807, 4.1578, 4.2390, 4.2218, 4.2097, 4.1986 &
                        , 2.8395, 3.0081, 3.3171, 3.4878, 3.5360, 3.5145, 3.2809 &
                        , 3.3307, 3.1260, 2.9940, 3.4741, 3.6675, 3.7832, 3.6787 &
                        , 3.6156, 3.8041, 3.5813, 3.4301, 3.8480, 3.9849, 3.9314 &
                        , 3.8405, 3.8029, 3.2962, 3.7104, 3.6515, 3.6378, 3.6020 &
                        , 3.5849, 3.5550, 3.7494, 3.6893, 3.6666, 3.9170, 3.7150 &
                        /)
      r0ab(2521:2590) = (/ &
                        3.5760, 4.0268, 4.1596, 4.1107, 3.9995, 3.5574, 3.5103 &
                        , 3.4163, 3.3655, 3.3677, 3.3243, 3.2975, 3.7071, 3.9047 &
                        , 3.8514, 3.8422, 3.8022, 3.9323, 3.7932, 4.2343, 4.4583 &
                        , 4.3115, 4.2457, 4.2213, 4.1945, 4.1756, 4.1569, 4.1424 &
                        , 4.0620, 4.0494, 3.9953, 4.0694, 4.0516, 4.0396, 4.0280 &
                        , 4.0130, 2.9007, 2.9674, 3.8174, 3.5856, 3.6486, 3.5339 &
                        , 3.2832, 3.3154, 3.1144, 2.9866, 3.9618, 3.8430, 3.9980 &
                        , 3.8134, 3.6652, 3.7985, 3.5756, 3.4207, 4.4061, 4.2817 &
                        , 4.1477, 4.0616, 3.9979, 3.6492, 3.8833, 3.8027, 3.7660 &
                        , 3.7183, 3.6954, 3.6525, 3.9669, 3.8371, 3.7325, 3.9160 &
                        /)
      r0ab(2591:2660) = (/ &
                        3.7156, 3.5714, 4.6036, 4.4620, 4.3092, 4.2122, 3.8478 &
                        , 3.7572, 3.6597, 3.5969, 3.5575, 3.5386, 3.5153, 3.7818 &
                        , 4.1335, 4.0153, 3.9177, 3.8603, 3.9365, 3.7906, 4.7936 &
                        , 4.7410, 4.5461, 4.5662, 4.5340, 4.5059, 4.4832, 4.4604 &
                        , 4.4429, 4.2346, 4.4204, 4.3119, 4.3450, 4.3193, 4.3035 &
                        , 4.2933, 4.1582, 4.2450, 2.8559, 2.9050, 3.8325, 3.5442 &
                        , 3.5077, 3.4905, 3.2396, 3.2720, 3.0726, 2.9467, 3.9644 &
                        , 3.8050, 3.8981, 3.7762, 3.6216, 3.7531, 3.5297, 3.3742 &
                        , 4.3814, 4.2818, 4.1026, 4.0294, 3.9640, 3.6208, 3.8464 &
                        , 3.7648, 3.7281, 3.6790, 3.6542, 3.6117, 3.8650, 3.8010 &
                        /)
      r0ab(2661:2730) = (/ &
                        3.6894, 3.8713, 3.6699, 3.5244, 4.5151, 4.4517, 4.2538 &
                        , 4.1483, 3.8641, 3.7244, 3.6243, 3.5589, 3.5172, 3.4973 &
                        , 3.4715, 3.7340, 4.0316, 3.9958, 3.8687, 3.8115, 3.8862 &
                        , 3.7379, 4.7091, 4.7156, 4.5199, 4.5542, 4.5230, 4.4959 &
                        , 4.4750, 4.4529, 4.4361, 4.1774, 4.3774, 4.2963, 4.3406 &
                        , 4.3159, 4.3006, 4.2910, 4.1008, 4.1568, 4.0980, 2.8110 &
                        , 2.8520, 3.7480, 3.5105, 3.4346, 3.3461, 3.1971, 3.2326 &
                        , 3.0329, 2.9070, 3.8823, 3.7928, 3.8264, 3.7006, 3.5797 &
                        , 3.7141, 3.4894, 3.3326, 4.3048, 4.2217, 4.0786, 3.9900 &
                        , 3.9357, 3.6331, 3.8333, 3.7317, 3.6957, 3.6460, 3.6197 &
                        /)
      r0ab(2731:2800) = (/ &
                        3.5779, 3.7909, 3.7257, 3.6476, 3.5729, 3.6304, 3.4834 &
                        , 4.4368, 4.3921, 4.2207, 4.1133, 3.8067, 3.7421, 3.6140 &
                        , 3.5491, 3.5077, 3.4887, 3.4623, 3.6956, 3.9568, 3.8976 &
                        , 3.8240, 3.7684, 3.8451, 3.6949, 4.6318, 4.6559, 4.4533 &
                        , 4.4956, 4.4641, 4.4366, 4.4155, 4.3936, 4.3764, 4.1302 &
                        , 4.3398, 4.2283, 4.2796, 4.2547, 4.2391, 4.2296, 4.0699 &
                        , 4.1083, 4.0319, 3.9855, 2.7676, 2.8078, 3.6725, 3.4804 &
                        , 3.3775, 3.2411, 3.1581, 3.1983, 2.9973, 2.8705, 3.8070 &
                        , 3.7392, 3.7668, 3.6263, 3.5402, 3.6807, 3.4545, 3.2962 &
                        , 4.2283, 4.1698, 4.0240, 3.9341, 3.8711, 3.5489, 3.7798 &
                        /)
      r0ab(2801:2870) = (/ &
                        3.7000, 3.6654, 3.6154, 3.5882, 3.5472, 3.7289, 3.6510 &
                        , 3.6078, 3.5355, 3.5963, 3.4480, 4.3587, 4.3390, 4.1635 &
                        , 4.0536, 3.7193, 3.6529, 3.5512, 3.4837, 3.4400, 3.4191 &
                        , 3.3891, 3.6622, 3.8934, 3.8235, 3.7823, 3.7292, 3.8106 &
                        , 3.6589, 4.5535, 4.6013, 4.3961, 4.4423, 4.4109, 4.3835 &
                        , 4.3625, 4.3407, 4.3237, 4.0863, 4.2835, 4.1675, 4.2272 &
                        , 4.2025, 4.1869, 4.1774, 4.0126, 4.0460, 3.9815, 3.9340 &
                        , 3.8955, 2.6912, 2.7604, 3.6037, 3.4194, 3.3094, 3.1710 &
                        , 3.0862, 3.1789, 2.9738, 2.8427, 3.7378, 3.6742, 3.6928 &
                        , 3.5512, 3.4614, 3.4087, 3.4201, 3.2607, 4.1527, 4.0977 &
                        /)
      r0ab(2871:2940) = (/ &
                        3.9523, 3.8628, 3.8002, 3.4759, 3.7102, 3.6466, 3.6106 &
                        , 3.5580, 3.5282, 3.4878, 3.6547, 3.5763, 3.5289, 3.5086 &
                        , 3.5593, 3.4099, 4.2788, 4.2624, 4.0873, 3.9770, 3.6407 &
                        , 3.5743, 3.5178, 3.4753, 3.3931, 3.3694, 3.3339, 3.6002 &
                        , 3.8164, 3.7478, 3.7028, 3.6952, 3.7669, 3.6137, 4.4698 &
                        , 4.5488, 4.3168, 4.3646, 4.3338, 4.3067, 4.2860, 4.2645 &
                        , 4.2478, 4.0067, 4.2349, 4.0958, 4.1543, 4.1302, 4.1141 &
                        , 4.1048, 3.9410, 3.9595, 3.8941, 3.8465, 3.8089, 3.7490 &
                        , 2.7895, 2.5849, 3.6484, 3.0162, 3.1267, 3.2125, 3.0043 &
                        , 2.9572, 2.8197, 2.7261, 3.7701, 3.2446, 3.5239, 3.4696 &
                        /)
      r0ab(2941:3010) = (/ &
                        3.4261, 3.3508, 3.1968, 3.0848, 4.1496, 3.6598, 3.5111 &
                        , 3.4199, 3.3809, 3.5382, 3.2572, 3.2100, 3.1917, 3.1519 &
                        , 3.1198, 3.1005, 3.5071, 3.5086, 3.5073, 3.4509, 3.3120 &
                        , 3.2082, 4.2611, 3.8117, 3.6988, 3.5646, 3.6925, 3.6295 &
                        , 3.5383, 3.4910, 3.4625, 3.4233, 3.4007, 3.2329, 3.6723 &
                        , 3.6845, 3.6876, 3.6197, 3.4799, 3.3737, 4.4341, 4.0525 &
                        , 3.9011, 3.8945, 3.8635, 3.8368, 3.8153, 3.7936, 3.7758 &
                        , 3.4944, 3.4873, 3.9040, 3.7110, 3.6922, 3.6799, 3.6724 &
                        , 3.5622, 3.6081, 3.5426, 3.4922, 3.4498, 3.3984, 3.4456 &
                        , 2.7522, 2.5524, 3.5742, 2.9508, 3.0751, 3.0158, 2.9644 &
                        /)
      r0ab(3011:3080) = (/ &
                        2.8338, 2.7891, 2.6933, 3.6926, 3.1814, 3.4528, 3.4186 &
                        , 3.3836, 3.2213, 3.1626, 3.0507, 4.0548, 3.5312, 3.4244 &
                        , 3.3409, 3.2810, 3.4782, 3.1905, 3.1494, 3.1221, 3.1128 &
                        , 3.0853, 3.0384, 3.4366, 3.4562, 3.4638, 3.3211, 3.2762 &
                        , 3.1730, 4.1632, 3.6825, 3.5822, 3.4870, 3.6325, 3.5740 &
                        , 3.4733, 3.4247, 3.3969, 3.3764, 3.3525, 3.1984, 3.5989 &
                        , 3.6299, 3.6433, 3.4937, 3.4417, 3.3365, 4.3304, 3.9242 &
                        , 3.7793, 3.7623, 3.7327, 3.7071, 3.6860, 3.6650, 3.6476 &
                        , 3.3849, 3.3534, 3.8216, 3.5870, 3.5695, 3.5584, 3.5508 &
                        , 3.4856, 3.5523, 3.4934, 3.4464, 3.4055, 3.3551, 3.3888 &
                        /)
      r0ab(3081:3150) = (/ &
                        3.3525, 2.7202, 2.5183, 3.4947, 2.8731, 3.0198, 3.1457 &
                        , 2.9276, 2.7826, 2.7574, 2.6606, 3.6090, 3.0581, 3.3747 &
                        , 3.3677, 3.3450, 3.1651, 3.1259, 3.0147, 3.9498, 3.3857 &
                        , 3.2917, 3.2154, 3.1604, 3.4174, 3.0735, 3.0342, 3.0096 &
                        , 3.0136, 2.9855, 2.9680, 3.3604, 3.4037, 3.4243, 3.2633 &
                        , 3.1810, 3.1351, 4.0557, 3.5368, 3.4526, 3.3699, 3.5707 &
                        , 3.5184, 3.4085, 3.3595, 3.3333, 3.3143, 3.3041, 3.1094 &
                        , 3.5193, 3.5745, 3.6025, 3.4338, 3.3448, 3.2952, 4.2158 &
                        , 3.7802, 3.6431, 3.6129, 3.5853, 3.5610, 3.5406, 3.5204 &
                        , 3.5036, 3.2679, 3.2162, 3.7068, 3.4483, 3.4323, 3.4221 &
                        /)
      r0ab(3151:3220) = (/ &
                        3.4138, 3.3652, 3.4576, 3.4053, 3.3618, 3.3224, 3.2711 &
                        , 3.3326, 3.2950, 3.2564, 2.5315, 2.6104, 3.2734, 3.2299 &
                        , 3.1090, 2.9942, 2.9159, 2.8324, 2.8350, 2.7216, 3.3994 &
                        , 3.4475, 3.4354, 3.3438, 3.2807, 3.2169, 3.2677, 3.1296 &
                        , 3.7493, 3.8075, 3.6846, 3.6104, 3.5577, 3.2052, 3.4803 &
                        , 3.4236, 3.3845, 3.3640, 3.3365, 3.3010, 3.3938, 3.3624 &
                        , 3.3440, 3.3132, 3.4035, 3.2754, 3.8701, 3.9523, 3.8018 &
                        , 3.7149, 3.3673, 3.3199, 3.2483, 3.2069, 3.1793, 3.1558 &
                        , 3.1395, 3.4097, 3.5410, 3.5228, 3.5116, 3.4921, 3.4781 &
                        , 3.4690, 4.0420, 4.1759, 4.0078, 4.0450, 4.0189, 3.9952 &
                        /)
      r0ab(3221:3290) = (/ &
                        3.9770, 3.9583, 3.9434, 3.7217, 3.8228, 3.7826, 3.8640 &
                        , 3.8446, 3.8314, 3.8225, 3.6817, 3.7068, 3.6555, 3.6159 &
                        , 3.5831, 3.5257, 3.2133, 3.1689, 3.1196, 3.3599, 2.9852 &
                        , 2.7881, 3.5284, 3.3493, 3.6958, 3.3642, 3.1568, 3.0055 &
                        , 2.9558, 2.8393, 3.6287, 3.5283, 4.1511, 3.8259, 3.6066 &
                        , 3.4527, 3.3480, 3.2713, 3.9037, 3.8361, 3.8579, 3.7311 &
                        , 3.6575, 3.5176, 3.5693, 3.5157, 3.4814, 3.4559, 3.4445 &
                        , 3.4160, 4.1231, 3.8543, 3.6816, 3.5602, 3.4798, 3.4208 &
                        , 4.0542, 4.0139, 4.0165, 3.9412, 3.7698, 3.6915, 3.6043 &
                        , 3.5639, 3.5416, 3.5247, 3.5153, 3.5654, 4.2862, 4.0437 &
                        /)
      r0ab(3291:3360) = (/ &
                        3.8871, 3.7741, 3.6985, 3.6413, 4.2345, 4.3663, 4.3257 &
                        , 4.0869, 4.0612, 4.0364, 4.0170, 3.9978, 3.9834, 3.9137 &
                        , 3.8825, 3.8758, 3.9143, 3.8976, 3.8864, 3.8768, 3.9190 &
                        , 4.1613, 4.0566, 3.9784, 3.9116, 3.8326, 3.7122, 3.6378 &
                        , 3.5576, 3.5457, 4.3127, 3.1160, 2.8482, 4.0739, 3.3599 &
                        , 3.5698, 3.5366, 3.2854, 3.1039, 2.9953, 2.9192, 4.1432 &
                        , 3.5320, 3.9478, 4.0231, 3.7509, 3.5604, 3.4340, 3.3426 &
                        , 4.3328, 3.8288, 3.7822, 3.7909, 3.6907, 3.6864, 3.5793 &
                        , 3.5221, 3.4883, 3.4649, 3.4514, 3.4301, 3.9256, 4.0596 &
                        , 3.8307, 3.6702, 3.5651, 3.4884, 4.4182, 4.2516, 3.9687 &
                        /)
      r0ab(3361:3430) = (/ &
                        3.9186, 3.9485, 3.8370, 3.7255, 3.6744, 3.6476, 3.6295 &
                        , 3.6193, 3.5659, 4.0663, 4.2309, 4.0183, 3.8680, 3.7672 &
                        , 3.6923, 4.5240, 4.4834, 4.1570, 4.3204, 4.2993, 4.2804 &
                        , 4.2647, 4.2481, 4.2354, 3.8626, 3.8448, 4.2267, 4.1799 &
                        , 4.1670, 3.8738, 3.8643, 3.8796, 4.0575, 4.0354, 3.9365 &
                        , 3.8611, 3.7847, 3.7388, 3.6826, 3.6251, 3.5492, 4.0889 &
                        , 4.2764, 3.1416, 2.8325, 3.7735, 3.3787, 3.4632, 3.5923 &
                        , 3.3214, 3.1285, 3.0147, 2.9366, 3.8527, 3.5602, 3.8131 &
                        , 3.8349, 3.7995, 3.5919, 3.4539, 3.3540, 4.0654, 3.8603 &
                        , 3.7972, 3.7358, 3.7392, 3.8157, 3.6055, 3.5438, 3.5089 &
                        /)
      r0ab(3431:3500) = (/ &
                        3.4853, 3.4698, 3.4508, 3.7882, 3.8682, 3.8837, 3.7055 &
                        , 3.5870, 3.5000, 4.1573, 4.0005, 3.9568, 3.8936, 3.9990 &
                        , 3.9433, 3.8172, 3.7566, 3.7246, 3.7033, 3.6900, 3.5697 &
                        , 3.9183, 4.0262, 4.0659, 3.8969, 3.7809, 3.6949, 4.2765 &
                        , 4.2312, 4.1401, 4.0815, 4.0580, 4.0369, 4.0194, 4.0017 &
                        , 3.9874, 3.8312, 3.8120, 3.9454, 3.9210, 3.9055, 3.8951 &
                        , 3.8866, 3.8689, 3.9603, 3.9109, 3.9122, 3.8233, 3.7438 &
                        , 3.7436, 3.6981, 3.6555, 3.5452, 3.9327, 4.0658, 4.1175 &
                        , 2.9664, 2.8209, 3.5547, 3.3796, 3.3985, 3.3164, 3.2364 &
                        , 3.1956, 3.0370, 2.9313, 3.6425, 3.5565, 3.7209, 3.7108 &
                        /)
      r0ab(3501:3570) = (/ &
                        3.6639, 3.6484, 3.4745, 3.3492, 3.8755, 4.2457, 3.7758 &
                        , 3.7161, 3.6693, 3.6155, 3.5941, 3.5643, 3.5292, 3.4950 &
                        , 3.4720, 3.4503, 3.6936, 3.7392, 3.7388, 3.7602, 3.6078 &
                        , 3.4960, 3.9800, 4.3518, 4.2802, 3.8580, 3.8056, 3.7527 &
                        , 3.7019, 3.6615, 3.5768, 3.5330, 3.5038, 3.5639, 3.8192 &
                        , 3.8883, 3.9092, 3.9478, 3.7995, 3.6896, 4.1165, 4.5232 &
                        , 4.4357, 4.4226, 4.4031, 4.3860, 4.3721, 4.3580, 4.3466 &
                        , 4.2036, 4.2037, 3.8867, 4.2895, 4.2766, 4.2662, 4.2598 &
                        , 3.8408, 3.9169, 3.8681, 3.8250, 3.7855, 3.7501, 3.6753 &
                        , 3.5499, 3.4872, 3.5401, 3.8288, 3.9217, 3.9538, 4.0054 &
                        /)
      r0ab(3571:3640) = (/ &
                        2.8388, 2.7890, 3.4329, 3.5593, 3.3488, 3.2486, 3.1615 &
                        , 3.1000, 3.0394, 2.9165, 3.5267, 3.7479, 3.6650, 3.6263 &
                        , 3.5658, 3.5224, 3.4762, 3.3342, 3.7738, 4.0333, 3.9568 &
                        , 3.8975, 3.8521, 3.4929, 3.7830, 3.7409, 3.7062, 3.6786 &
                        , 3.6471, 3.6208, 3.6337, 3.6519, 3.6363, 3.6278, 3.6110 &
                        , 3.4825, 3.8795, 4.1448, 4.0736, 4.0045, 3.6843, 3.6291 &
                        , 3.5741, 3.5312, 3.4974, 3.4472, 3.4034, 3.7131, 3.7557 &
                        , 3.7966, 3.8005, 3.8068, 3.8015, 3.6747, 4.0222, 4.3207 &
                        , 4.2347, 4.2191, 4.1990, 4.1811, 4.1666, 4.1521, 4.1401 &
                        , 3.9970, 3.9943, 3.9592, 4.0800, 4.0664, 4.0559, 4.0488 &
                        /)
      r0ab(3641:3710) = (/ &
                        3.9882, 4.0035, 3.9539, 3.9138, 3.8798, 3.8355, 3.5359 &
                        , 3.4954, 3.3962, 3.5339, 3.7595, 3.8250, 3.8408, 3.8600 &
                        , 3.8644, 2.7412, 2.7489, 3.3374, 3.3950, 3.3076, 3.1910 &
                        , 3.0961, 3.0175, 3.0280, 2.8929, 3.4328, 3.5883, 3.6227 &
                        , 3.5616, 3.4894, 3.4241, 3.3641, 3.3120, 3.6815, 3.8789 &
                        , 3.8031, 3.7413, 3.6939, 3.4010, 3.6225, 3.5797, 3.5443 &
                        , 3.5139, 3.4923, 3.4642, 3.5860, 3.5849, 3.5570, 3.5257 &
                        , 3.4936, 3.4628, 3.7874, 3.9916, 3.9249, 3.8530, 3.5932 &
                        , 3.5355, 3.4757, 3.4306, 3.3953, 3.3646, 3.3390, 3.5637 &
                        , 3.7053, 3.7266, 3.7177, 3.6996, 3.6775, 3.6558, 3.9331 &
                        /)
      r0ab(3711:3780) = (/ &
                        4.1655, 4.0879, 4.0681, 4.0479, 4.0299, 4.0152, 4.0006 &
                        , 3.9883, 3.8500, 3.8359, 3.8249, 3.9269, 3.9133, 3.9025 &
                        , 3.8948, 3.8422, 3.8509, 3.7990, 3.7570, 3.7219, 3.6762 &
                        , 3.4260, 3.3866, 3.3425, 3.5294, 3.7022, 3.7497, 3.7542 &
                        , 3.7494, 3.7370, 3.7216, 3.4155, 3.0522, 4.2541, 3.8218 &
                        , 4.0438, 3.5875, 3.3286, 3.1682, 3.0566, 2.9746, 4.3627 &
                        , 4.0249, 4.6947, 4.1718, 3.8639, 3.6735, 3.5435, 3.4479 &
                        , 4.6806, 4.3485, 4.2668, 4.1690, 4.1061, 4.1245, 4.0206 &
                        , 3.9765, 3.9458, 3.9217, 3.9075, 3.8813, 3.9947, 4.1989 &
                        , 3.9507, 3.7960, 3.6925, 3.6150, 4.8535, 4.5642, 4.4134 &
                        /)
      r0ab(3781:3850) = (/ &
                        4.3688, 4.3396, 4.2879, 4.2166, 4.1888, 4.1768, 4.1660 &
                        , 4.1608, 4.0745, 4.2289, 4.4863, 4.2513, 4.0897, 3.9876 &
                        , 3.9061, 5.0690, 5.0446, 4.6186, 4.6078, 4.5780, 4.5538 &
                        , 4.5319, 4.5101, 4.4945, 4.1912, 4.2315, 4.5534, 4.4373 &
                        , 4.4224, 4.4120, 4.4040, 4.2634, 4.7770, 4.6890, 4.6107 &
                        , 4.5331, 4.4496, 4.4082, 4.3095, 4.2023, 4.0501, 4.2595 &
                        , 4.5497, 4.3056, 4.1506, 4.0574, 3.9725, 5.0796, 3.0548 &
                        , 3.3206, 3.8132, 3.9720, 3.7675, 3.7351, 3.5167, 3.5274 &
                        , 3.3085, 3.1653, 3.9500, 4.1730, 4.0613, 4.1493, 3.8823 &
                        , 4.0537, 3.8200, 3.6582, 4.3422, 4.5111, 4.3795, 4.3362 &
                        /)
      r0ab(3851:3920) = (/ &
                        4.2751, 3.7103, 4.1973, 4.1385, 4.1129, 4.0800, 4.0647 &
                        , 4.0308, 4.0096, 4.1619, 3.9360, 4.1766, 3.9705, 3.8262 &
                        , 4.5348, 4.7025, 4.5268, 4.5076, 3.9562, 3.9065, 3.8119 &
                        , 3.7605, 3.7447, 3.7119, 3.6916, 4.1950, 4.2110, 4.3843 &
                        , 4.1631, 4.4427, 4.2463, 4.1054, 4.7693, 5.0649, 4.7365 &
                        , 4.7761, 4.7498, 4.7272, 4.7076, 4.6877, 4.6730, 4.4274 &
                        , 4.5473, 4.5169, 4.5975, 4.5793, 4.5667, 4.5559, 4.3804 &
                        , 4.6920, 4.6731, 4.6142, 4.5600, 4.4801, 4.0149, 3.8856 &
                        , 3.7407, 4.1545, 4.2253, 4.4229, 4.1923, 4.5022, 4.3059 &
                        , 4.1591, 4.7883, 4.9294, 3.3850, 3.4208, 3.7004, 3.8800 &
                        /)
      r0ab(3921:3990) = (/ &
                        3.9886, 3.9040, 3.6719, 3.6547, 3.4625, 3.3370, 3.8394 &
                        , 4.0335, 4.2373, 4.3023, 4.0306, 4.1408, 3.9297, 3.7857 &
                        , 4.1907, 4.3230, 4.2664, 4.2173, 4.1482, 3.6823, 4.0711 &
                        , 4.0180, 4.0017, 3.9747, 3.9634, 3.9383, 4.1993, 4.3205 &
                        , 4.0821, 4.2547, 4.0659, 3.9359, 4.3952, 4.5176, 4.3888 &
                        , 4.3607, 3.9583, 3.9280, 3.8390, 3.7971, 3.7955, 3.7674 &
                        , 3.7521, 4.1062, 4.3633, 4.2991, 4.2767, 4.4857, 4.3039 &
                        , 4.1762, 4.6197, 4.8654, 4.6633, 4.5878, 4.5640, 4.5422 &
                        , 4.5231, 4.5042, 4.4901, 4.3282, 4.3978, 4.3483, 4.4202 &
                        , 4.4039, 4.3926, 4.3807, 4.2649, 4.6135, 4.5605, 4.5232 &
                        /)
      r0ab(3991:4060) = (/ &
                        4.4676, 4.3948, 4.0989, 3.9864, 3.8596, 4.0942, 4.2720 &
                        , 4.3270, 4.3022, 4.5410, 4.3576, 4.2235, 4.6545, 4.7447 &
                        , 4.7043, 3.0942, 3.2075, 3.5152, 3.6659, 3.8289, 3.7459 &
                        , 3.5156, 3.5197, 3.3290, 3.2069, 3.6702, 3.8448, 4.0340 &
                        , 3.9509, 3.8585, 3.9894, 3.7787, 3.6365, 4.1425, 4.1618 &
                        , 4.0940, 4.0466, 3.9941, 3.5426, 3.8952, 3.8327, 3.8126 &
                        , 3.7796, 3.7635, 3.7356, 4.0047, 3.9655, 3.9116, 4.1010 &
                        , 3.9102, 3.7800, 4.2964, 4.3330, 4.2622, 4.2254, 3.8195 &
                        , 3.7560, 3.6513, 3.5941, 3.5810, 3.5420, 3.5178, 3.8861 &
                        , 4.1459, 4.1147, 4.0772, 4.3120, 4.1207, 3.9900, 4.4733 &
                        /)
      r0ab(4061:4130) = (/ &
                        4.6157, 4.4580, 4.4194, 4.3954, 4.3739, 4.3531, 4.3343 &
                        , 4.3196, 4.2140, 4.2339, 4.1738, 4.2458, 4.2278, 4.2158 &
                        , 4.2039, 4.1658, 4.3595, 4.2857, 4.2444, 4.1855, 4.1122 &
                        , 3.7839, 3.6879, 3.5816, 3.8633, 4.1585, 4.1402, 4.1036 &
                        , 4.3694, 4.1735, 4.0368, 4.5095, 4.5538, 4.5240, 4.4252 &
                        , 3.0187, 3.1918, 3.5127, 3.6875, 3.7404, 3.6943, 3.4702 &
                        , 3.4888, 3.2914, 3.1643, 3.6669, 3.8724, 3.9940, 4.0816 &
                        , 3.8054, 3.9661, 3.7492, 3.6024, 4.0428, 4.1951, 4.1466 &
                        , 4.0515, 4.0075, 3.5020, 3.9158, 3.8546, 3.8342, 3.8008 &
                        , 3.7845, 3.7549, 3.9602, 3.8872, 3.8564, 4.0793, 3.8835 &
                        /)
      r0ab(4131:4200) = (/ &
                        3.7495, 4.2213, 4.3704, 4.3300, 4.2121, 3.7643, 3.7130 &
                        , 3.6144, 3.5599, 3.5474, 3.5093, 3.4853, 3.9075, 4.1115 &
                        , 4.0473, 4.0318, 4.2999, 4.1050, 3.9710, 4.4320, 4.6706 &
                        , 4.5273, 4.4581, 4.4332, 4.4064, 4.3873, 4.3684, 4.3537 &
                        , 4.2728, 4.2549, 4.2032, 4.2794, 4.2613, 4.2491, 4.2375 &
                        , 4.2322, 4.3665, 4.3061, 4.2714, 4.2155, 4.1416, 3.7660 &
                        , 3.6628, 3.5476, 3.8790, 4.1233, 4.0738, 4.0575, 4.3575 &
                        , 4.1586, 4.0183, 4.4593, 4.5927, 4.4865, 4.3813, 4.4594 &
                        , 2.9875, 3.1674, 3.4971, 3.6715, 3.7114, 3.6692, 3.4446 &
                        , 3.4676, 3.2685, 3.1405, 3.6546, 3.8579, 3.9637, 4.0581 &
                        /)
      r0ab(4201:4270) = (/ &
                        3.7796, 3.9463, 3.7275, 3.5792, 4.0295, 4.1824, 4.1247 &
                        , 4.0357, 3.9926, 3.4827, 3.9007, 3.8392, 3.8191, 3.7851 &
                        , 3.7687, 3.7387, 3.9290, 3.8606, 3.8306, 4.0601, 3.8625 &
                        , 3.7269, 4.2062, 4.3566, 4.3022, 4.1929, 3.7401, 3.6888 &
                        , 3.5900, 3.5350, 3.5226, 3.4838, 3.4594, 3.8888, 4.0813 &
                        , 4.0209, 4.0059, 4.2810, 4.0843, 3.9486, 4.4162, 4.6542 &
                        , 4.5005, 4.4444, 4.4196, 4.3933, 4.3741, 4.3552, 4.3406 &
                        , 4.2484, 4.2413, 4.1907, 4.2656, 4.2474, 4.2352, 4.2236 &
                        , 4.2068, 4.3410, 4.2817, 4.2479, 4.1921, 4.1182, 3.7346 &
                        , 3.6314, 3.5168, 3.8582, 4.0927, 4.0469, 4.0313, 4.3391 &
                        /)
      r0ab(4271:4340) = (/ &
                        4.1381, 3.9962, 4.4429, 4.5787, 4.4731, 4.3588, 4.4270 &
                        , 4.3957, 2.9659, 3.1442, 3.4795, 3.6503, 3.6814, 3.6476 &
                        , 3.4222, 3.4491, 3.2494, 3.1209, 3.6324, 3.8375, 3.9397 &
                        , 3.8311, 3.7581, 3.9274, 3.7085, 3.5598, 4.0080, 4.1641 &
                        , 4.1057, 4.0158, 3.9726, 3.4667, 3.8802, 3.8188, 3.7989 &
                        , 3.7644, 3.7474, 3.7173, 3.9049, 3.8424, 3.8095, 4.0412 &
                        , 3.8436, 3.7077, 4.1837, 4.3366, 4.2816, 4.1686, 3.7293 &
                        , 3.6709, 3.5700, 3.5153, 3.5039, 3.4684, 3.4437, 3.8663 &
                        , 4.0575, 4.0020, 3.9842, 4.2612, 4.0643, 3.9285, 4.3928 &
                        , 4.6308, 4.4799, 4.4244, 4.3996, 4.3737, 4.3547, 4.3358 &
                        /)
      r0ab(4341:4410) = (/ &
                        4.3212, 4.2275, 4.2216, 4.1676, 4.2465, 4.2283, 4.2161 &
                        , 4.2045, 4.1841, 4.3135, 4.2562, 4.2226, 4.1667, 4.0932 &
                        , 3.7134, 3.6109, 3.4962, 3.8352, 4.0688, 4.0281, 4.0099 &
                        , 4.3199, 4.1188, 3.9768, 4.4192, 4.5577, 4.4516, 4.3365 &
                        , 4.4058, 4.3745, 4.3539, 2.8763, 3.1294, 3.5598, 3.7465 &
                        , 3.5659, 3.5816, 3.3599, 3.4024, 3.1877, 3.0484, 3.7009 &
                        , 3.9451, 3.8465, 3.9873, 3.7079, 3.9083, 3.6756, 3.5150 &
                        , 4.0829, 4.2780, 4.1511, 4.1260, 4.0571, 3.4865, 3.9744 &
                        , 3.9150, 3.8930, 3.8578, 3.8402, 3.8073, 3.7977, 4.0036 &
                        , 3.7604, 4.0288, 3.8210, 3.6757, 4.2646, 4.4558, 4.2862 &
                        /)
      r0ab(4411:4465) = (/ &
                        4.2122, 3.7088, 3.6729, 3.5800, 3.5276, 3.5165, 3.4783 &
                        , 3.4539, 3.9553, 3.9818, 4.2040, 3.9604, 4.2718, 4.0689 &
                        , 3.9253, 4.4869, 4.7792, 4.4918, 4.5342, 4.5090, 4.4868 &
                        , 4.4680, 4.4486, 4.4341, 4.2023, 4.3122, 4.2710, 4.3587 &
                        , 4.3407, 4.3281, 4.3174, 4.1499, 4.3940, 4.3895, 4.3260 &
                        , 4.2725, 4.1961, 3.7361, 3.6193, 3.4916, 3.9115, 3.9914 &
                        , 3.9809, 3.9866, 4.3329, 4.1276, 3.9782, 4.5097, 4.6769 &
                        , 4.5158, 4.3291, 4.3609, 4.3462, 4.3265, 4.4341 &
                        /)

      k = 0
      DO i = 1, SIZE(rout, 1)
         DO j = 1, i
            k = k+1
            rout(i, j) = r0ab(k)*bohr
            rout(j, i) = r0ab(k)*bohr
         ENDDO
      ENDDO

      ! covalent radii (taken from Pyykko and Atsumi, Chem. Eur. J. 15, 2009, 188-197)
      ! values for metals decreased by 10 %
      rcov(1:94) = (/ &
                   0.32, 0.46, 1.20, 0.94, 0.77, 0.75, 0.71, 0.63, 0.64, 0.67 &
                   , 1.40, 1.25, 1.13, 1.04, 1.10, 1.02, 0.99, 0.96, 1.76, 1.54 &
                   , 1.33, 1.22, 1.21, 1.10, 1.07, 1.04, 1.00, 0.99, 1.01, 1.09 &
                   , 1.12, 1.09, 1.15, 1.10, 1.14, 1.17, 1.89, 1.67, 1.47, 1.39 &
                   , 1.32, 1.24, 1.15, 1.13, 1.13, 1.08, 1.15, 1.23, 1.28, 1.26 &
                   , 1.26, 1.23, 1.32, 1.31, 2.09, 1.76, 1.62, 1.47, 1.58, 1.57 &
                   , 1.56, 1.55, 1.51, 1.52, 1.51, 1.50, 1.49, 1.49, 1.48, 1.53 &
                   , 1.46, 1.37, 1.31, 1.23, 1.18, 1.16, 1.11, 1.12, 1.13, 1.32 &
                   , 1.30, 1.30, 1.36, 1.31, 1.38, 1.42, 2.01, 1.81, 1.67, 1.58 &
                   , 1.52, 1.53, 1.54, 1.55 &
                   /)

      ! PBE0/def2-QZVP atomic values
      r2r4(1:94) = (/ &
                   8.0589, 3.4698, 29.0974, 14.8517, 11.8799, 7.8715, 5.5588, &
                   4.7566, 3.8025, 3.1036, 26.1552, 17.2304, 17.7210, 12.7442, &
                   9.5361, 8.1652, 6.7463, 5.6004, 29.2012, 22.3934, 19.0598, &
                   16.8590, 15.4023, 12.5589, 13.4788, 12.2309, 11.2809, 10.5569, &
                   10.1428, 9.4907, 13.4606, 10.8544, 8.9386, 8.1350, 7.1251, &
                   6.1971, 30.0162, 24.4103, 20.3537, 17.4780, 13.5528, 11.8451, &
                   11.0355, 10.1997, 9.5414, 9.0061, 8.6417, 8.9975, 14.0834, &
                   11.8333, 10.0179, 9.3844, 8.4110, 7.5152, 32.7622, 27.5708, &
                   23.1671, 21.6003, 20.9615, 20.4562, 20.1010, 19.7475, 19.4828, &
                   15.6013, 19.2362, 17.4717, 17.8321, 17.4237, 17.1954, 17.1631, &
                   14.5716, 15.8758, 13.8989, 12.4834, 11.4421, 10.2671, 8.3549, &
                   7.8496, 7.3278, 7.4820, 13.5124, 11.6554, 10.0959, 9.7340, &
                   8.8584, 8.0125, 29.8135, 26.3157, 19.1885, 15.8542, 16.1305, &
                   15.6161, 15.1226, 16.1576 &
                   /)

   END SUBROUTINE setr0ab

! **************************************************************************************************
!> \brief ...
!> \param cnout ...
! **************************************************************************************************
   SUBROUTINE setcn(cnout)
      ! default coordination numbers
      REAL(KIND=dp), DIMENSION(:)                        :: cnout

      INTEGER                                            :: n
      REAL(KIND=dp), DIMENSION(104)                      :: cn

      cn(1:104) = (/ &
                  1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 3.0, 2.0, 1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0 &
                  , 2.0, 1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 3.0, 2.0, 3.0, 2.0, 2.0, 2.0, 2.0 &
                  , 3.0, 4.0, 3.0, 4.0, 1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 4.0, 3.0 &
                  , 2.0, 1.0, 2.0, 3.0, 2.0, 3.0, 4.0, 1.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0 &
                  , 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 5.0, 6.0, 7.0 &
                  , 4.0, 4.0, 4.0, 3.0, 2.0, 1.0, 2.0, 3.0, 4.0, 1.0, 0.0, 1.0, 2.0, 3.0, 4.0 &
                  , 5.0, 6.0, 5.0, 4.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 2.0, 3.0, 3.0 &
                  /)

      cnout = 0._dp
      n = MIN(SIZE(cn), SIZE(cnout))
      cnout(1:n) = cn(1:n)

   END SUBROUTINE setcn

! **************************************************************************************************
!> \brief ...
!> \param rab ...
!> \param rcova ...
!> \param rcovb ...
!> \param k1 ...
!> \param cnab ...
!> \param dcnab ...
! **************************************************************************************************
   SUBROUTINE cnparam_d3(rab, rcova, rcovb, k1, cnab, dcnab)

      REAL(KIND=dp), INTENT(IN)                          :: rab, rcova, rcovb, k1
      REAL(KIND=dp), INTENT(OUT)                         :: cnab, dcnab

      REAL(KIND=dp)                                      :: dfz, ee, fz, rco, rr

! covalent distance in Bohr

      rco = rcova+rcovb
      rr = rco/rab
      ! counting function exponential has a better long-range behavior
      ! than MHGs inverse damping
      ! factor k2 already included into rcov
      ee = EXP(-k1*(rr-1.0_dp))
      ! force the function to zero using a second step function
      fz = 0.5_dp*(1.0_dp-TANH(rab-2.0_dp*rco))
      dfz = 0.5_dp*((TANH(rab-2.0_dp*rco))**2-1.0_dp)
      cnab = 1.0_dp/(1.0_dp+ee)*fz
      dcnab = -cnab*cnab*k1*rr/rab*ee+1.0_dp/(1.0_dp+ee)*dfz

   END SUBROUTINE cnparam_d3

! **************************************************************************************************
!> \brief ...
!> \param rab ...
!> \param rcutab ...
!> \param srn ...
!> \param alpn ...
!> \param rcut ...
!> \param fdab ...
!> \param dfdab ...
! **************************************************************************************************
   SUBROUTINE damping_d3(rab, rcutab, srn, alpn, rcut, fdab, dfdab)

      REAL(KIND=dp), INTENT(IN)                          :: rab, rcutab, srn, alpn, rcut
      REAL(KIND=dp), INTENT(OUT)                         :: fdab, dfdab

      REAL(KIND=dp)                                      :: a, b, c, d, dd, dfab, dfcc, dz, fab, &
                                                            fcc, rl, rr, ru, z, zz

      rl = rcut-1._dp
      ru = rcut
      IF (rab >= ru) THEN
         fcc = 0._dp
         dfcc = 0._dp
      ELSEIF (rab <= rl) THEN
         fcc = 1._dp
         dfcc = 0._dp
      ELSE
         z = rab*rab-rl*rl
         dz = 2._dp*rab
         zz = z*z*z
         d = ru*ru-rl*rl
         dd = d*d*d
         a = -10._dp/dd
         b = 15._dp/(dd*d)
         c = -6._dp/(dd*d*d)
         fcc = 1._dp+zz*(a+b*z+c*z*z)
         dfcc = zz*dz/z*(3._dp*a+4._dp*b*z+5._dp*c*z*z)
      END IF

      rr = 6._dp*(rab/(srn*rcutab))**(-alpn)
      fab = 1._dp/(1._dp+rr)
      dfab = fab*fab*rr*alpn/rab
      fdab = fab*fcc
      dfdab = dfab*fcc+fab*dfcc

   END SUBROUTINE damping_d3

! **************************************************************************************************
!> \brief ...
!> \param maxc ...
!> \param max_elem ...
!> \param c6ab ...
!> \param mxc ...
!> \param iat ...
!> \param jat ...
!> \param nci ...
!> \param ncj ...
!> \param k3 ...
!> \param c6 ...
!> \param dc6a ...
!> \param dc6b ...
! **************************************************************************************************
   SUBROUTINE getc6(maxc, max_elem, c6ab, mxc, iat, jat, nci, ncj, k3, c6, dc6a, dc6b)

      INTEGER, INTENT(IN)                                :: maxc, max_elem
      REAL(KIND=dp), INTENT(IN) :: c6ab(max_elem, max_elem, maxc, maxc, 3)
      INTEGER, INTENT(IN)                                :: mxc(max_elem), iat, jat
      REAL(KIND=dp), INTENT(IN)                          :: nci, ncj, k3
      REAL(KIND=dp), INTENT(OUT)                         :: c6, dc6a, dc6b

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: c6mem, cn1, cn2, csum, dtmpa, dtmpb, &
                                                            dwa, dwb, dza, dzb, r, rsave, rsum, &
                                                            tmp1

! the exponential is sensitive to numerics
! when nci or ncj is much larger than cn1/cn2

      c6mem = -1.0e+99_dp
      rsave = 1.0e+99_dp
      rsum = 0.0_dp
      csum = 0.0_dp
      dza = 0.0_dp
      dzb = 0.0_dp
      dwa = 0.0_dp
      dwb = 0.0_dp
      c6 = 0.0_dp
      DO i = 1, mxc(iat)
         DO j = 1, mxc(jat)
            c6 = c6ab(iat, jat, i, j, 1)
            IF (c6 > 0.0_dp) THEN
               cn1 = c6ab(iat, jat, i, j, 2)
               cn2 = c6ab(iat, jat, i, j, 3)
               ! distance
               r = (cn1-nci)**2+(cn2-ncj)**2
               IF (r < rsave) THEN
                  rsave = r
                  c6mem = c6
               ENDIF
               tmp1 = EXP(k3*r)
               dtmpa = -2.0_dp*k3*(cn1-nci)*tmp1
               dtmpb = -2.0_dp*k3*(cn2-ncj)*tmp1
               rsum = rsum+tmp1
               csum = csum+tmp1*c6
               dza = dza+dtmpa*c6
               dwa = dwa+dtmpa
               dzb = dzb+dtmpb*c6
               dwb = dwb+dtmpb
            ENDIF
         ENDDO
      ENDDO

      IF (c6 == 0.0_dp) c6mem = 0.0_dp

      IF (rsum > 1.0e-66_dp) THEN
         c6 = csum/rsum
         dc6a = (dza-c6*dwa)/rsum
         dc6b = (dzb-c6*dwb)/rsum
      ELSE
         c6 = c6mem
         dc6a = 0._dp
         dc6b = 0._dp
      ENDIF

   END SUBROUTINE getc6

! **************************************************************************************************
!> \brief ...
!> \param dcnum ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE dcnum_distribute(dcnum, para_env)

      TYPE(dcnum_type), DIMENSION(:)                     :: dcnum
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dcnum_distribute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, i, ia, ipe, jn, n, natom, ntmax, &
                                                            ntot
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: list, nloc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dvals
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rik

! we only have to do something if this is a parallel run

      IF (para_env%num_pe > 1) THEN
         group = para_env%group
         natom = SIZE(dcnum)
         !pack my dcnum data
         ALLOCATE (nloc(natom))
         DO ia = 1, natom
            nloc(ia) = dcnum(ia)%neighbors
         END DO
         ntot = SUM(nloc)
         ntmax = ntot
         CALL mp_max(ntmax, group)
         ALLOCATE (list(ntmax), dvals(ntmax), rik(3, ntmax))
         list = 0
         dvals = 0._dp
         rik = 0._dp
         i = 0
         DO ia = 1, natom
            DO jn = 1, nloc(ia)
               i = i+1
               list(i) = dcnum(ia)%nlist(jn)
               dvals(i) = dcnum(ia)%dvals(jn)
               rik(1:3, i) = dcnum(ia)%rik(1:3, jn)
            END DO
         END DO
         DO ipe = 1, para_env%num_pe-1
            !send/receive packed data
            CALL mp_shift(nloc, group)
            !unpack received data
            CALL mp_shift(list, group)
            CALL mp_shift(dvals, group)
            CALL mp_shift(rik, group)
            !add data to local dcnum
            i = 0
            DO ia = 1, natom
               n = dcnum(ia)%neighbors+nloc(ia)
               IF (n > SIZE(dcnum(ia)%nlist)) THEN
                  CALL reallocate(dcnum(ia)%nlist, 1, 2*n)
                  CALL reallocate(dcnum(ia)%dvals, 1, 2*n)
                  CALL reallocate(dcnum(ia)%rik, 1, 3, 1, 2*n)
               END IF
               DO jn = 1, nloc(ia)
                  i = i+1
                  n = dcnum(ia)%neighbors+jn
                  dcnum(ia)%nlist(n) = list(i)
                  dcnum(ia)%dvals(n) = dvals(i)
                  dcnum(ia)%rik(1:3, n) = rik(1:3, i)
               END DO
               dcnum(ia)%neighbors = dcnum(ia)%neighbors+nloc(ia)
            END DO
         END DO
         DEALLOCATE (nloc)
         DEALLOCATE (list, dvals, rik)
      END IF

   END SUBROUTINE dcnum_distribute

! **************************************************************************************************

END MODULE qs_dispersion_pairpot
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set disperson types for DFT calculations
!> \author JGH (04.2014)
! **************************************************************************************************
MODULE qs_dispersion_utils

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: vdw_nl_DRSLL,&
                                              vdw_nl_LMKLL,&
                                              vdw_nl_RVV10,&
                                              vdw_pairpot_dftd2,&
                                              vdw_pairpot_dftd3,&
                                              vdw_pairpot_dftd3bj,&
                                              xc_vdw_fun_nonloc,&
                                              xc_vdw_fun_pairpot
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE physcon,                         ONLY: bohr,&
                                              kjmol
   USE qs_dispersion_pairpot,           ONLY: qs_scaling_dftd3,&
                                              qs_scaling_dftd3bj,&
                                              qs_scaling_init
   USE qs_dispersion_types,             ONLY: qs_atom_dispersion_type,&
                                              qs_dispersion_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dispersion_utils'

   PUBLIC :: qs_dispersion_env_set, qs_write_dispersion

! **************************************************************************************************
CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param dispersion_env ...
!> \param xc_section ...
! **************************************************************************************************
   SUBROUTINE qs_dispersion_env_set(dispersion_env, xc_section)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(section_vals_type), POINTER                   :: xc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dispersion_env_set', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: exfun, explicit
      REAL(dp), POINTER                                  :: params(:), scal(:)
      TYPE(section_vals_type), POINTER                   :: nl_section, pp_section, vdw_section, &
                                                            xc_fun_section

      IF(.NOT.(ASSOCIATED(dispersion_env)))CALL cp__a("qs_dispersion_utils.F",70)

      ! set general defaults
      dispersion_env%doabc = .FALSE.
      dispersion_env%c9cnst = .FALSE.
      dispersion_env%lrc = .FALSE.
      dispersion_env%verbose = .FALSE.
      NULLIFY (dispersion_env%c6ab, dispersion_env%maxci, dispersion_env%r0ab, dispersion_env%rcov, &
               dispersion_env%r2r4, dispersion_env%cn, dispersion_env%cnkind, dispersion_env%cnlist)
      NULLIFY (dispersion_env%q_mesh, dispersion_env%kernel, dispersion_env%d2phi_dk2, &
               dispersion_env%d2y_dx2)
      NULLIFY (dispersion_env%sab_vdw, dispersion_env%sab_cn)
      NULLIFY (dispersion_env%dftd_section)
      NULLIFY (vdw_section, xc_fun_section)
      vdw_section => section_vals_get_subs_vals(xc_section, "vdw_potential")
      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      CALL section_vals_val_get(vdw_section, "POTENTIAL_TYPE", i_val=dispersion_env%type)
      IF (dispersion_env%type == xc_vdw_fun_pairpot) THEN
         NULLIFY (pp_section)
         pp_section => section_vals_get_subs_vals(vdw_section, "PAIR_POTENTIAL")
         CALL section_vals_val_get(pp_section, "VERBOSE_OUTPUT", l_val=dispersion_env%verbose)
         CALL section_vals_val_get(pp_section, "TYPE", i_val=dispersion_env%pp_type)
         IF (dispersion_env%pp_type == vdw_pairpot_dftd2) THEN
            ! functional parameters for Grimme D2 type
            CALL section_vals_val_get(pp_section, "EXP_PRE", r_val=dispersion_env%exp_pre)
            CALL section_vals_val_get(pp_section, "SCALING", explicit=explicit)
            IF (.NOT. explicit) THEN
               CALL section_vals_val_get(pp_section, "REFERENCE_FUNCTIONAL", explicit=exfun)
               IF(.NOT.(exfun))CALL cp__a("qs_dispersion_utils.F",98)
               CALL qs_scaling_init(dispersion_env%scaling, vdw_section)
            ELSE
               CALL section_vals_val_get(pp_section, "SCALING", r_val=dispersion_env%scaling)
            END IF
         ELSE
            dispersion_env%exp_pre = 0._dp
            dispersion_env%scaling = 0._dp
         END IF
         IF (dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
             dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
            ! functional parameters for Grimme DFT-D3 type
            CALL section_vals_val_get(pp_section, "EPS_CN", r_val=dispersion_env%eps_cn)
            CALL section_vals_val_get(pp_section, "CALCULATE_C9_TERM", l_val=dispersion_env%doabc)
            CALL section_vals_val_get(pp_section, "REFERENCE_C9_TERM", l_val=dispersion_env%c9cnst)
            CALL section_vals_val_get(pp_section, "LONG_RANGE_CORRECTION", l_val=dispersion_env%lrc)
            IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
               CALL section_vals_val_get(pp_section, "D3_SCALING", explicit=explicit)
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
               CALL section_vals_val_get(pp_section, "D3BJ_SCALING", explicit=explicit)
            END IF
            IF (.NOT. explicit) THEN
               CALL section_vals_val_get(pp_section, "REFERENCE_FUNCTIONAL", explicit=exfun)
               IF(.NOT.(exfun))CALL cp__a("qs_dispersion_utils.F",121)
               IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
                  CALL qs_scaling_dftd3(dispersion_env%s6, dispersion_env%sr6, dispersion_env%s8, vdw_section)
               ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
                  CALL qs_scaling_dftd3bj(dispersion_env%s6, dispersion_env%a1, dispersion_env%s8, &
                                          dispersion_env%a2, vdw_section)
               END IF
            ELSE
               IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
                  ! zero damping
                  CALL section_vals_val_get(pp_section, "D3_SCALING", r_vals=scal)
                  dispersion_env%s6 = scal(1)
                  dispersion_env%sr6 = scal(2)
                  dispersion_env%s8 = scal(3)
                  dispersion_env%a1 = 0.0_dp
                  dispersion_env%a2 = 0.0_dp
               ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
                  ! BJ damping
                  CALL section_vals_val_get(pp_section, "D3BJ_SCALING", r_vals=scal)
                  dispersion_env%s6 = scal(1)
                  dispersion_env%a1 = scal(2)
                  dispersion_env%s8 = scal(3)
                  dispersion_env%a2 = scal(4)
                  dispersion_env%sr6 = 0.0_dp
               END IF
            END IF
         ELSE
            dispersion_env%s6 = 0._dp
            dispersion_env%sr6 = 0._dp
            dispersion_env%s8 = 0._dp
            dispersion_env%a1 = 0._dp
            dispersion_env%a2 = 0._dp
            dispersion_env%eps_cn = 0._dp
         END IF
         CALL section_vals_val_get(pp_section, "R_CUTOFF", r_val=dispersion_env%rc_disp)
         CALL section_vals_val_get(pp_section, "PARAMETER_FILE_NAME", &
                                   c_val=dispersion_env%parameter_file_name)
         ! set DFTD section for output handling
         dispersion_env%dftd_section => pp_section
      ELSE IF (dispersion_env%type == xc_vdw_fun_nonloc) THEN
         NULLIFY (nl_section)
         nl_section => section_vals_get_subs_vals(vdw_section, "NON_LOCAL")
         CALL section_vals_val_get(nl_section, "VERBOSE_OUTPUT", l_val=dispersion_env%verbose)
         CALL section_vals_val_get(nl_section, "KERNEL_FILE_NAME", &
                                   c_val=dispersion_env%kernel_file_name)
         CALL section_vals_val_get(nl_section, "TYPE", i_val=dispersion_env%nl_type)
         CALL section_vals_val_get(nl_section, "CUTOFF", r_val=dispersion_env%pw_cutoff)
         dispersion_env%pw_cutoff = 0.5_dp*dispersion_env%pw_cutoff
         CALL section_vals_val_get(nl_section, "PARAMETERS", r_vals=params)
         dispersion_env%b_value = params(1)
         dispersion_env%c_value = params(2)
      END IF
   END SUBROUTINE qs_dispersion_env_set

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param dispersion_env ...
!> \param ounit ...
! **************************************************************************************************
   SUBROUTINE qs_write_dispersion(qs_env, dispersion_env, ounit)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      INTEGER, INTENT(in), OPTIONAL                      :: ounit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_write_dispersion', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: symbol
      INTEGER                                            :: i, ikind, nkind, output_unit
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_atom_dispersion_type), POINTER             :: disp
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: dft_section

      IF (PRESENT(ounit)) THEN
         output_unit = ounit
      ELSE
         NULLIFY (logger)
         logger => cp_get_default_logger()

         dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
         output_unit = cp_print_key_unit_nr(logger, dft_section, &
                                            "PRINT%DFT_CONTROL_PARAMETERS", extension=".Log")
      END IF

      IF (output_unit > 0) THEN
         ! vdW type specific output
         IF (dispersion_env%type == xc_vdw_fun_pairpot) THEN
            WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T67,'Pair Potential')")
            ! Pair potentials
            IF (dispersion_env%pp_type == vdw_pairpot_dftd2) THEN
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T35,'Potential Form: S. Grimme, JCC 27: 1787 (2006)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T35,'Cutoff Radius [Bohr]:',T73,F8.2)") dispersion_env%rc_disp
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T35,'Scaling Factor:',T73,F8.4)") dispersion_env%scaling
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T35,'Exp Prefactor for Damping:',T73,F8.1)") dispersion_env%exp_pre
               CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set)
               nkind = SIZE(atomic_kind_set)
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), element_symbol=symbol)
                  CALL get_qs_kind(qs_kind_set(ikind), dispersion=disp)
                  IF (disp%defined) THEN
                     WRITE (output_unit, fmt="(' vdW PARAMETER| ',T18,'Atom=',A2,"// &
                            "T28,'C6[J*nm^6*mol^-1]=',F8.4,T63,'r(vdW)[A]=',F8.4)") &
                        symbol, disp%c6/(1000._dp*bohr**6/kjmol), disp%vdw_radii/bohr
                  ELSE
                     WRITE (output_unit, fmt="(' vdW PARAMETER| ',T20,'Atom=',A2,T70,'not defined')")
                  END IF
               END DO
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3) THEN
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'DFT-D3 (Version 3.1)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Potential Form: S. Grimme et al, JCP 132: 154104 (2010)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Zero Damping')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Cutoff Radius [Bohr]:',T73,F8.2)") dispersion_env%rc_disp
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'s6 Scaling Factor:',T73,F8.4)") dispersion_env%s6
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'sr6 Scaling Factor:',T73,F8.4)") dispersion_env%sr6
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'s8 Scaling Factor:',T73,F8.4)") dispersion_env%s8
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Cutoff for CN calculation:',T69,E12.4)") dispersion_env%eps_cn
            ELSE IF (dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'DFT-D3 (Version 3.1)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Potential Form: S. Grimme et al, JCP 132: 154104 (2010)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'BJ Damping: S. Grimme et al, JCC 32: 1456 (2011)')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Cutoff Radius [Bohr]:',T73,F8.2)") dispersion_env%rc_disp
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'s6 Scaling Factor:',T73,F8.4)") dispersion_env%s6
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'a1 Damping Factor:',T73,F8.4)") dispersion_env%a1
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'s8 Scaling Factor:',T73,F8.4)") dispersion_env%s8
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'a2 Damping Factor:',T73,F8.4)") dispersion_env%a2
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T26,'Cutoff for CN calculation:',T69,E12.4)") dispersion_env%eps_cn
            END IF
         ELSE IF (dispersion_env%type == xc_vdw_fun_nonloc) THEN
            WRITE (output_unit, fmt="(' vdW POTENTIAL| ',T61,'Non-local Functional')")
            WRITE (output_unit, &
                   fmt="(' vdW POTENTIAL| ','Implementation: G. Roman-Perez, J. Soler, PRL 103: 096102 (2009)')")
            WRITE (output_unit, &
                   fmt="(' vdW POTENTIAL| ',T38,' T. Thonhauser et al, PRB 76: 125112 (2007)')")
            WRITE (output_unit, &
                   fmt="(' vdW POTENTIAL| ',T22,' R. Sabatini et al, J.Phys:Condens Matter 24: 424209 (2012)')")
            WRITE (output_unit, &
                   fmt="(' vdW POTENTIAL| ',T16,' Based on QE implementation by Brian Kolb, Timo Thonhauser (2009)')")
            SELECT CASE (dispersion_env%nl_type)
            CASE DEFAULT
               ! unknown functional
               CALL cp__b("qs_dispersion_utils.F",266,"")
            CASE (vdw_nl_DRSLL)
               WRITE (output_unit, &
                      fmt="(' vdW POTENTIAL| ','DRSLL Functional:           M. Dion et al, PRL 92: 246401 (2004)')")
            CASE (vdw_nl_LMKLL)
               WRITE (output_unit, &
                      fmt="(' vdW POTENTIAL| ','LMKLL Functional:            K. Lee et al, PRB 82: 081101 (2010)')")
            CASE (vdw_nl_RVV10)
               WRITE (output_unit, &
                      fmt="(' vdW POTENTIAL| ','RVV10 Functional:    R. Sabatini et al, PRB 87: 041108(R) (2013)')")
            END SELECT
            IF (dispersion_env%verbose) THEN
               WRITE (output_unit, &
                      fmt="(' vdW POTENTIAL| ','         Carrying out vdW-DF run using the following parameters:')")
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ','Nqs =',I8,'        Nr_points =',I8,'       r_max =',F10.3)") &
                  dispersion_env%nqs, dispersion_env%nr_points, dispersion_env%r_max
               WRITE (output_unit, fmt="(' vdW POTENTIAL| ','q_mesh =')")
               WRITE (output_unit, fmt="(8X,4F18.8)") (dispersion_env%q_mesh(i), i=1, dispersion_env%nqs)
               WRITE (output_unit, &
                      fmt="(' vdW POTENTIAL| ','Density cutoff for comvolution:',T68,F10.1,' Ry')") &
                  2.0_dp*dispersion_env%pw_cutoff
            END IF
         END IF
      END IF
      IF (.NOT. PRESENT(ounit)) THEN
         CALL cp_print_key_finished_output(output_unit, logger, dft_section, &
                                           "PRINT%DFT_CONTROL_PARAMETERS")
      END IF

   END SUBROUTINE qs_write_dispersion

! **************************************************************************************************

END MODULE qs_dispersion_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set of procedures to modify the density matrix, in order to collocate
!>      on the real space grid specific functions of the primitives
!> \author MI 06.2006
! **************************************************************************************************
MODULE qs_modify_pab_block

   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: coset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_modify_pab_block'

! *** Public subroutines ***

   PUBLIC :: prepare_dadb, prepare_adb_m_dab, prepare_dab_p_adb, prepare_ardb_m_darb, prepare_arb
   PUBLIC :: prepare_diadib, prepare_dijadijb, prepare_diiadiib
   PUBLIC :: FUNC_AB, FUNC_ARB, FUNC_DADB, FUNC_ADBmDAB, FUNC_ARDBmDARB, FUNC_DABpADB
   PUBLIC :: FUNC_DER, FUNC_DXDY, FUNC_DYDZ, FUNC_DZDX, FUNC_DX, FUNC_DY, FUNC_DZ
   PUBLIC :: FUNC_DXDX, FUNC_DYDY, FUNC_DZDZ

   ! Flags for the type of function to be collocated
   INTEGER, PARAMETER :: FUNC_AB = 401, FUNC_DADB = 402
   INTEGER, PARAMETER :: FUNC_ADBmDAB = 403, FUNC_ARDBmDARB = 404
   INTEGER, PARAMETER :: FUNC_DABpADB = 405, FUNC_ARB = 406
   INTEGER, DIMENSION(0:9), PARAMETER :: FUNC_DER = (/401, 501, 502, 503, 601, 602, 603, 604, 605, 606/)
   INTEGER, PARAMETER :: FUNC_DXDY = 601, FUNC_DYDZ = 602, FUNC_DZDX = 603
   INTEGER, PARAMETER :: FUNC_DXDX = 604, FUNC_DYDY = 605, FUNC_DZDZ = 606
   INTEGER, PARAMETER :: FUNC_DX = 501, FUNC_DY = 502, FUNC_DZ = 503

CONTAINS

!

! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with 0.5 * (nabla pgf_a) . (nabla pgf_b)
!>      (ddx pgf_a ) (ddx pgf_b) = (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x})*(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x})
!> \param pab_local ...
!> \param pab ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_dadb(pab_local, pab, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pab
      INTEGER, INTENT(IN)                                :: lxa, lya, lza, lxb, lyb, lzb, o1, o2
      REAL(dp), INTENT(IN)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco, jco_l

! this element of pab results in 12 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)
      ! x  (all safe if lxa = 0, as the spurious added terms have zero prefactor)

      ico_l = coset(MAX(lxa-1, 0), lya, lza)
      jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxa*lxb*pab(o1+ico, o2+jco)
      ico_l = coset(MAX(lxa-1, 0), lya, lza)
      jco_l = coset((lxb+1), lyb, lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lxa*zetb*pab(o1+ico, o2+jco)
      ico_l = coset((lxa+1), lya, lza)
      jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lxb*pab(o1+ico, o2+jco)
      ico_l = coset((lxa+1), lya, lza)
      jco_l = coset((lxb+1), lyb, lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)

      ! y

      ico_l = coset(lxa, MAX(lya-1, 0), lza)
      jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lya*lyb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, MAX(lya-1, 0), lza)
      jco_l = coset(lxb, (lyb+1), lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lya*zetb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, (lya+1), lza)
      jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lyb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, (lya+1), lza)
      jco_l = coset(lxb, (lyb+1), lzb)
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)

      ! z

      ico_l = coset(lxa, lya, MAX(lza-1, 0))
      jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lza*lzb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, lya, MAX(lza-1, 0))
      jco_l = coset(lxb, lyb, (lzb+1))
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lza*zetb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, lya, (lza+1))
      jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lzb*pab(o1+ico, o2+jco)
      ico_l = coset(lxa, lya, (lza+1))
      jco_l = coset(lxb, lyb, (lzb+1))
      pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)

   END SUBROUTINE prepare_dadb

!
! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with  (ddi pgf_a) . (ddi pgf_b)
!>      (ddx pgf_a ) (ddx pgf_b) = (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x})*(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x})
!> \param pab_local ...
!> \param pab ...
!> \param ider ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_diadib(pab_local, pab, ider, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(inout)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(in)              :: pab
      INTEGER, INTENT(in)                                :: ider, lxa, lya, lza, lxb, lyb, lzb, o1, &
                                                            o2
      REAL(dp), INTENT(in)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco, jco_l

! this element of pab results in 12 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)
      IF (ider == 1) THEN
         ! x  (all safe if lxa = 0, as the spurious added terms have zero prefactor)
         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxa*lxb*pab(o1+ico, o2+jco)
         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lxa*zetb*pab(o1+ico, o2+jco)
         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lxb*pab(o1+ico, o2+jco)
         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)

      ELSEIF (ider == 2) THEN
         ! y

         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lya*lyb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lya*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lyb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)
      ELSEIF (ider == 3) THEN
         ! z
         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lza*lzb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*lza*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*lzb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zeta*zetb*pab(o1+ico, o2+jco)
      END IF

   END SUBROUTINE prepare_diadib
!
! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with  (ddidj pgf_a) . (ddidj pgf_b)
!>      (ddxdy pgf_a ) (ddxdy pgf_b) =
!> \param pab_local ...
!> \param pab ...
!> \param ider1 ...
!> \param ider2 ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_dijadijb(pab_local, pab, ider1, ider2, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(inout)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(in)              :: pab
      INTEGER, INTENT(in)                                :: ider1, ider2, lxa, lya, lza, lxb, lyb, &
                                                            lzb, o1, o2
      REAL(dp), INTENT(in)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco
      REAL(dp)                                           :: func_a

! this element of pab results in 12 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF ((ider1 == 1 .AND. ider2 == 2) .OR. (ider1 == 2 .AND. ider2 == 1)) THEN ! xy
         ico_l = coset(MAX(lxa-1, 0), MAX(lya-1, 0), lza)
         func_a = lxa*lya*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
         ico_l = coset(lxa+1, MAX(lya-1, 0), lza)
         func_a = -2.0_dp*zeta*lya*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
         ico_l = coset(MAX(lxa-1, 0), lya+1, lza)
         func_a = -2.0_dp*zeta*lxa*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
         ico_l = coset(lxa+1, lya+1, lza)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
      ELSEIF ((ider1 == 2 .AND. ider2 == 3) .OR. (ider1 == 3 .AND. ider2 == 2)) THEN ! yz
         ico_l = coset(lxa, MAX(lya-1, 0), MAX(lza-1, 0))
         func_a = lya*lza*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
         ico_l = coset(lxa, lya+1, MAX(lza-1, 0))
         func_a = -2.0_dp*zeta*lza*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
         ico_l = coset(lxa, MAX(lya-1, 0), lza+1)
         func_a = -2.0_dp*zeta*lya*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
         ico_l = coset(lxa, lya+1, lza+1)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
      ELSEIF ((ider1 == 3 .AND. ider2 == 1) .OR. (ider1 == 1 .AND. ider2 == 3)) THEN ! zx
         ico_l = coset(MAX(lxa-1, 0), lya, MAX(lza-1, 0))
         func_a = lza*lxa*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
         ico_l = coset(MAX(lxa-1, 0), lya, lza+1)
         func_a = -2.0_dp*zeta*lxa*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
         ico_l = coset(lxa+1, lya, MAX(lza-1, 0))
         func_a = -2.0_dp*zeta*lza*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
         ico_l = coset(lxa+1, lya, lza+1)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_dijdij(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
      END IF

   END SUBROUTINE prepare_dijadijb

! **************************************************************************************************
!> \brief ...
!> \param pab_local ...
!> \param func_a ...
!> \param ico_l ...
!> \param lx ...
!> \param ly ...
!> \param lz ...
!> \param zet ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE oneterm_dijdij(pab_local, func_a, ico_l, lx, ly, lz, zet, idir)

      REAL(dp), DIMENSION(:, :), INTENT(inout)           :: pab_local
      REAL(dp), INTENT(in)                               :: func_a
      INTEGER, INTENT(in)                                :: ico_l, lx, ly, lz
      REAL(dp), INTENT(in)                               :: zet
      INTEGER, INTENT(in)                                :: idir

      INTEGER                                            :: jco_l, l1, l2

      IF (idir == 1) THEN
         l1 = lx
         l2 = ly
         jco_l = coset(MAX(lx-1, 0), MAX(ly-1, 0), lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*l2*func_a
         jco_l = coset(lx+1, MAX(ly-1, 0), lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l2*func_a
         jco_l = coset(MAX(lx-1, 0), ly+1, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l1*func_a
         jco_l = coset(lx+1, ly+1, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a
      ELSEIF (idir == 2) THEN
         l1 = ly
         l2 = lz
         jco_l = coset(lx, MAX(ly-1, 0), MAX(lz-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*l2*func_a
         jco_l = coset(lx, ly+1, MAX(lz-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l2*func_a
         jco_l = coset(lx, MAX(ly-1, 0), lz+1)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l1*func_a
         jco_l = coset(lx, ly+1, lz+1)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a
      ELSEIF (idir == 3) THEN
         l1 = lz
         l2 = lx
         jco_l = coset(MAX(lx-1, 0), ly, MAX(lz-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*l2*func_a
         jco_l = coset(MAX(lx-1, 0), ly, lz+1)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l2*func_a
         jco_l = coset(lx+1, ly, MAX(lz-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*l1*func_a
         jco_l = coset(lx+1, ly, lz+1)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a

      END IF
   END SUBROUTINE oneterm_dijdij

!
! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with  (ddidj pgf_a) . (ddidj pgf_b)
!>      (ddxdx pgf_a ) (ddxdx pgf_b) =
!> \param pab_local ...
!> \param pab ...
!> \param ider ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_diiadiib(pab_local, pab, ider, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(inout)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(in)              :: pab
      INTEGER, INTENT(in)                                :: ider, lxa, lya, lza, lxb, lyb, lzb, o1, &
                                                            o2
      REAL(dp), INTENT(in)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco
      REAL(dp)                                           :: func_a

! this element of pab results in  9 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF (ider == 1) THEN ! x
         ico_l = coset(MAX(lxa-2, 0), lya, lza)
         func_a = lxa*(lxa-1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
         ico_l = coset(lxa, lya, lza)
         func_a = -2.0_dp*zeta*(2*lxa+1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
         ico_l = coset(lxa+2, lya, lza)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=1)
      ELSEIF (ider == 2) THEN ! y
         ico_l = coset(lxa, MAX(lya-2, 0), lza)
         func_a = lya*(lya-1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
         ico_l = coset(lxa, lya, lza)
         func_a = -2.0_dp*zeta*(2*lya+1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
         ico_l = coset(lxa, lya+2, lza)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=2)
      ELSEIF (ider == 3) THEN ! z
         ico_l = coset(lxa, lya, MAX(lza-2, 0))
         func_a = lza*(lza-1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
         ico_l = coset(lxa, lya, lza)
         func_a = -2.0_dp*zeta*(2*lza+1)*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
         ico_l = coset(lxa, lya, lza+2)
         func_a = 4.0_dp*zeta*zeta*pab(o1+ico, o2+jco)
         CALL oneterm_diidii(pab_local, func_a, ico_l, lxb, lyb, lzb, zetb, idir=3)
      END IF

   END SUBROUTINE prepare_diiadiib

! **************************************************************************************************
!> \brief ...
!> \param pab_local ...
!> \param func_a ...
!> \param ico_l ...
!> \param lx ...
!> \param ly ...
!> \param lz ...
!> \param zet ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE oneterm_diidii(pab_local, func_a, ico_l, lx, ly, lz, zet, idir)
      REAL(dp), DIMENSION(:, :), INTENT(inout)           :: pab_local
      REAL(dp), INTENT(in)                               :: func_a
      INTEGER, INTENT(in)                                :: ico_l, lx, ly, lz
      REAL(dp), INTENT(in)                               :: zet
      INTEGER, INTENT(in)                                :: idir

      INTEGER                                            :: jco_l, l1

      IF (idir == 1) THEN
         l1 = lx
         jco_l = coset(MAX(lx-2, 0), ly, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*(l1-1)*func_a
         jco_l = coset(lx, ly, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*(2*l1+1)*func_a
         jco_l = coset(lx+2, ly, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a
      ELSEIF (idir == 2) THEN
         l1 = ly
         jco_l = coset(lx, MAX(ly-2, 0), lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*(l1-1)*func_a
         jco_l = coset(lx, ly, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*(2*l1+1)*func_a
         jco_l = coset(lx, ly+2, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a
      ELSEIF (idir == 3) THEN
         l1 = lz
         jco_l = coset(lx, ly, MAX(lz-2, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+l1*(l1-1)*func_a
         jco_l = coset(lx, ly, lz)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zet*(2*l1+1)*func_a
         jco_l = coset(lx, ly, lz+2)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+4.0_dp*zet*zet*func_a
      END IF

   END SUBROUTINE oneterm_diidii

! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with  pgf_a (nabla_{idir} pgf_b) - (nabla_{idir} pgf_a) pgf_b
!>      ( pgf_a ) (ddx pgf_b) - (ddx pgf_a)( pgf_b ) =
!>              pgf_a *(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x}) - (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x}) pgf_b
!> \param pab_local ...
!> \param pab ...
!> \param idir ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_adb_m_dab(pab_local, pab, idir, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pab
      INTEGER, INTENT(IN)                                :: idir, lxa, lya, lza, lxb, lyb, lzb, o1, &
                                                            o2
      REAL(dp), INTENT(IN)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco, jco_l

! this element of pab results in 4 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF (idir == 1) THEN ! x
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lxa*pab(o1+ico, o2+jco)
         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)
      ELSEIF (idir == 2) THEN ! y
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lyb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lya*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)
      ELSE ! z
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lzb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lza*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)
      END IF

   END SUBROUTINE prepare_adb_m_dab

!
! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      is equivalent to mapping pab with  pgf_a (nabla_{idir} pgf_b) + (nabla_{idir} pgf_a) pgf_b
!>      ( pgf_a ) (ddx pgf_b) + (ddx pgf_a)( pgf_b ) =
!>              pgf_a *(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x}) + (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x}) pgf_b
!> \param pab_local ...
!> \param pab ...
!> \param idir ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_dab_p_adb(pab_local, pab, idir, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pab
      INTEGER, INTENT(IN)                                :: idir, lxa, lya, lza, lxb, lyb, lzb, o1, &
                                                            o2
      REAL(dp), INTENT(IN)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco, jco_l

! this element of pab results in 4 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF (idir == 1) THEN ! x
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(MAX(lxb-1, 0), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxa*pab(o1+ico, o2+jco)
         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*pab(o1+ico, o2+jco)
      ELSEIF (idir == 2) THEN ! y
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, MAX(lyb-1, 0), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lyb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lya*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*pab(o1+ico, o2+jco)
      ELSE ! z
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lzb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lza*pab(o1+ico, o2+jco)
         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zeta*pab(o1+ico, o2+jco)
      END IF

   END SUBROUTINE prepare_dab_p_adb

!

! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>      pgf_a (r-Rb)_{ir} (nabla_{idir} pgf_b) - (nabla_{idir} pgf_a) (r-Rb)_{ir}  pgf_b
!>       ( pgf_a )(r-Rb)_{ir} (ddx pgf_b) - (ddx pgf_a) (r-Rb)_{ir} ( pgf_b ) =
!>                     pgf_a *(lbx pgf_{b-1x+1ir} - 2*zetb*pgf_{b+1x+1ir}) -
!>                              (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x}) pgf_{b+1ir}
!> \param pab_local ...
!> \param pab ...
!> \param idir ...
!> \param ir ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
!> \param zeta ...
!> \param zetb ...
! **************************************************************************************************
   SUBROUTINE prepare_ardb_m_darb(pab_local, pab, idir, ir, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pab
      INTEGER, INTENT(IN)                                :: idir, ir, lxa, lya, lza, lxb, lyb, lzb, &
                                                            o1, o2
      REAL(dp), INTENT(IN)                               :: zeta, zetb

      INTEGER                                            :: ico, ico_l, jco, jco_l

! this element of pab results in 4 elements of pab_local

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF (idir == 1 .AND. ir == 1) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+2), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lxa*pab(o1+ico, o2+jco)

         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 1 .AND. ir == 2) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(MAX(lxb-1, 0), (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lxa*pab(o1+ico, o2+jco)

         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 1 .AND. ir == 3) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(MAX(lxb-1, 0), lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lxb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(MAX(lxa-1, 0), lya, lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lxa*pab(o1+ico, o2+jco)

         ico_l = coset((lxa+1), lya, lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 2 .AND. ir == 1) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), MAX(lyb-1, 0), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lyb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lya*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 2 .AND. ir == 2) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lyb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+2), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lya*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 2 .AND. ir == 3) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, MAX(lyb-1, 0), (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lyb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, MAX(lya-1, 0), lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lya*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, (lya+1), lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 3 .AND. ir == 1) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lzb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lza*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 3 .AND. ir == 2) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), MAX(lzb-1, 0))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lzb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lza*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      ELSEIF (idir == 3 .AND. ir == 3) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+lzb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, (lzb+2))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-2.0_dp*zetb*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, MAX(lza-1, 0))
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)-lza*pab(o1+ico, o2+jco)

         ico_l = coset(lxa, lya, (lza+1))
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+2.0_dp*zeta*pab(o1+ico, o2+jco)

      END IF

   END SUBROUTINE prepare_ardb_m_darb

! **************************************************************************************************
!> \brief create a new pab_local so that mapping pab_local with pgf_a pgf_b
!>    pgf_a (r-Rb)_{ir} pgf_b = pgf_a * pgf_b{+1ir}
!> \param pab_local ...
!> \param pab ...
!> \param ir ...
!> \param lxa ...
!> \param lya ...
!> \param lza ...
!> \param lxb ...
!> \param lyb ...
!> \param lzb ...
!> \param o1 ...
!> \param o2 ...
! **************************************************************************************************
   SUBROUTINE prepare_arb(pab_local, pab, ir, lxa, lya, lza, lxb, lyb, lzb, o1, o2)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: pab_local
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pab
      INTEGER, INTENT(IN)                                :: ir, lxa, lya, lza, lxb, lyb, lzb, o1, o2

      INTEGER                                            :: ico, ico_l, jco, jco_l

      ico = coset(lxa, lya, lza)
      jco = coset(lxb, lyb, lzb)

      IF (ir == 1) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset((lxb+1), lyb, lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+pab(o1+ico, o2+jco)

      ELSEIF (ir == 2) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, (lyb+1), lzb)
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+pab(o1+ico, o2+jco)

      ELSEIF (ir == 3) THEN

         ico_l = coset(lxa, lya, lza)
         jco_l = coset(lxb, lyb, (lzb+1))
         pab_local(ico_l, jco_l) = pab_local(ico_l, jco_l)+pab(o1+ico, o2+jco)

      END IF

   END SUBROUTINE prepare_arb

END MODULE qs_modify_pab_block

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief utils to manipulate splines on the regular grid of a pw
!> \par History
!>      01.2014 move routines related to input_section_types to seperate file.
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_spline_utils
   USE input_constants,                 ONLY: spline3_nopbc_interp,&
                                              spline3_pbc_interp
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: &
        add_coarse2fine, add_fine2coarse, find_coeffs, pw_spline_do_precond, &
        pw_spline_precond_create, pw_spline_precond_release, pw_spline_precond_set_kind, &
        pw_spline_precond_type, spl3_1d_transf_border1, spl3_1d_transf_coeffs, spl3_nopbc, &
        spl3_nopbct, spl3_pbc
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_spline_utils'

   PUBLIC ::  pw_prolongate_s3, pw_restrict_s3

CONTAINS

! **************************************************************************************************
!> \brief restricts the function from a fine grid to a coarse one
!> \param pw_fine_in the fine grid
!> \param pw_coarse_out the coarse grid
!> \param coarse_pool ...
!> \param param_section ...
!> \author fawzi
!> \note
!>      extremely slow (but correct) version
! **************************************************************************************************
   SUBROUTINE pw_restrict_s3(pw_fine_in, pw_coarse_out, coarse_pool, param_section)
      TYPE(pw_type), POINTER                             :: pw_fine_in, pw_coarse_out
      TYPE(pw_pool_type), POINTER                        :: coarse_pool
      TYPE(section_vals_type), POINTER                   :: param_section

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_restrict_s3', routineP = moduleN//':'//routineN

      INTEGER                                            :: aint_precond, handle, interp_kind, &
                                                            max_iter, precond_kind
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, SAVE                                      :: ifile = 0
      LOGICAL                                            :: pbc, safe_computation, success
      REAL(kind=dp)                                      :: eps_r, eps_x
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(pw_type), POINTER                             :: coeffs, values

      ifile = ifile+1
      CALL timeset(routineN, handle)
      CALL section_vals_val_get(param_section, "safe_computation", &
                                l_val=safe_computation)
      CALL section_vals_val_get(param_section, "aint_precond", &
                                i_val=aint_precond)
      CALL section_vals_val_get(param_section, "precond", &
                                i_val=precond_kind)
      CALL section_vals_val_get(param_section, "max_iter", &
                                i_val=max_iter)
      CALL section_vals_val_get(param_section, "eps_r", &
                                r_val=eps_r)
      CALL section_vals_val_get(param_section, "eps_x", &
                                r_val=eps_x)
      CALL section_vals_val_get(param_section, "kind", &
                                i_val=interp_kind)

      pbc = (interp_kind == spline3_pbc_interp)
      IF(.NOT.(pbc .OR. interp_kind == spline3_nopbc_interp))CALL cp__a("cp_spline_utils.F",86)
      bo = pw_coarse_out%pw_grid%bounds_local
      NULLIFY (values, coeffs)
      CALL pw_pool_create_pw(coarse_pool, values, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_zero(values)

!FM       nullify(tst_pw)
!FM       CALL pw_pool_create_pw(coarse_pool,tst_pw, use_data=REALDATA3D,&
!FM            in_space=REALSPACE)
!FM       call pw_copy(values,tst_pw)
!FM       call add_fine2coarse(fine_values_pw=pw_fine_in,&
!FM            coarse_coeffs_pw=tst_pw,&
!FM            weights_1d=spl3_1d_transf_coeffs/2._dp, w_border0=0.5_dp,&
!FM            w_border1=spl3_1d_transf_border1/2._dp,pbc=pbc,&
!FM            safe_computation=.false.)

      CALL add_fine2coarse(fine_values_pw=pw_fine_in, &
                           coarse_coeffs_pw=values, &
                           weights_1d=spl3_1d_transf_coeffs/2._dp, w_border0=0.5_dp, &
                           w_border1=spl3_1d_transf_border1/2._dp, pbc=pbc, &
                           safe_computation=safe_computation)

!FM       CALL pw_compare_debug(tst_pw,values,max_diff)
!FM       WRITE(cp_logger_get_default_unit_nr(logger,.TRUE.),*)"f2cmax_diff=",max_diff
!FM       CALL pw_pool_give_back_pw(coarse_pool,tst_pw)

      CALL pw_pool_create_pw(coarse_pool, coeffs, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      NULLIFY (precond)
      CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                    pool=coarse_pool, pbc=pbc, transpose=.TRUE.)
      CALL pw_spline_do_precond(precond, values, coeffs)
      CALL pw_spline_precond_set_kind(precond, precond_kind)
      IF (pbc) THEN
         success = find_coeffs(values=values, coeffs=coeffs, &
                               linOp=spl3_pbc, preconditioner=precond, pool=coarse_pool, &
                               eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
      ELSE
         success = find_coeffs(values=values, coeffs=coeffs, &
                               linOp=spl3_nopbct, preconditioner=precond, pool=coarse_pool, &
                               eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
      END IF
      CALL pw_spline_precond_release(precond)

      CALL pw_zero(pw_coarse_out)
      CALL pw_axpy(coeffs, pw_coarse_out)

      CALL pw_pool_give_back_pw(coarse_pool, values)
      CALL pw_pool_give_back_pw(coarse_pool, coeffs)
      CALL timestop(handle)
   END SUBROUTINE pw_restrict_s3

! **************************************************************************************************
!> \brief prolongates a function from a coarse grid into a fine one
!> \param pw_coarse_in the coarse grid
!> \param pw_fine_out the fine grid
!> \param coarse_pool ...
!> \param param_section ...
!> \author fawzi
!> \note
!>      extremely slow (but correct) version
! **************************************************************************************************
   SUBROUTINE pw_prolongate_s3(pw_coarse_in, pw_fine_out, coarse_pool, &
                               param_section)
      TYPE(pw_type), POINTER                             :: pw_coarse_in, pw_fine_out
      TYPE(pw_pool_type), POINTER                        :: coarse_pool
      TYPE(section_vals_type), POINTER                   :: param_section

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_prolongate_s3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aint_precond, handle, interp_kind, &
                                                            max_iter, precond_kind
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, SAVE                                      :: ifile = 0
      LOGICAL                                            :: pbc, safe_computation, success
      REAL(kind=dp)                                      :: eps_r, eps_x
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(pw_type), POINTER                             :: coeffs

      ifile = ifile+1
      CALL timeset(routineN, handle)
      NULLIFY (coeffs)
      CALL pw_pool_create_pw(coarse_pool, coeffs, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      bo = pw_coarse_in%pw_grid%bounds_local
      CALL section_vals_val_get(param_section, "safe_computation", &
                                l_val=safe_computation)
      CALL section_vals_val_get(param_section, "aint_precond", &
                                i_val=aint_precond)
      CALL section_vals_val_get(param_section, "precond", &
                                i_val=precond_kind)
      CALL section_vals_val_get(param_section, "max_iter", &
                                i_val=max_iter)
      CALL section_vals_val_get(param_section, "eps_r", &
                                r_val=eps_r)
      CALL section_vals_val_get(param_section, "eps_x", &
                                r_val=eps_x)
      CALL section_vals_val_get(param_section, "kind", &
                                i_val=interp_kind)

      pbc = (interp_kind == spline3_pbc_interp)
      IF(.NOT.(pbc .OR. interp_kind == spline3_nopbc_interp))CALL cp__a("cp_spline_utils.F",189)
      NULLIFY (precond)
      CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                    pool=coarse_pool, pbc=pbc, transpose=.FALSE.)
      CALL pw_spline_do_precond(precond, pw_coarse_in, coeffs)
      CALL pw_spline_precond_set_kind(precond, precond_kind)
      IF (pbc) THEN
         success = find_coeffs(values=pw_coarse_in, coeffs=coeffs, &
                               linOp=spl3_pbc, preconditioner=precond, pool=coarse_pool, &
                               eps_r=eps_r, eps_x=eps_x, &
                               max_iter=max_iter)
      ELSE
         success = find_coeffs(values=pw_coarse_in, coeffs=coeffs, &
                               linOp=spl3_nopbc, preconditioner=precond, pool=coarse_pool, &
                               eps_r=eps_r, eps_x=eps_x, &
                               max_iter=max_iter)
      END IF
      IF(.NOT.(success))CALL cp__a("cp_spline_utils.F",206)
      CALL pw_spline_precond_release(precond)

!FM       nullify(tst_pw)
!FM       call pw_create(tst_pw, pw_fine_out%pw_grid, use_data=REALDATA3D,&
!FM            in_space=REALSPACE)
!FM       call pw_copy(pw_fine_out,tst_pw)
!FM       CALL add_coarse2fine(coarse_coeffs_pw=coeffs,&
!FM            fine_values_pw=tst_pw,&
!FM            weights_1d=spl3_1d_transf_coeffs,&
!FM            w_border0=1._dp,&
!FM            w_border1=spl3_1d_transf_border1,&
!FM            pbc=pbc,safe_computation=.false.,&
!FM

      CALL add_coarse2fine(coarse_coeffs_pw=coeffs, &
                           fine_values_pw=pw_fine_out, &
                           weights_1d=spl3_1d_transf_coeffs, &
                           w_border0=1._dp, &
                           w_border1=spl3_1d_transf_border1, &
                           pbc=pbc, safe_computation=safe_computation)

!FM       CALL pw_compare_debug(tst_pw,pw_fine_out,max_diff)
!FM       WRITE(cp_logger_get_default_unit_nr(logger,.TRUE.),*)"c2fmax_diff=",max_diff
!FM       CALL pw_release(tst_pw)

      CALL pw_pool_give_back_pw(coarse_pool, coeffs)

      CALL timestop(handle)
   END SUBROUTINE pw_prolongate_s3

END MODULE cp_spline_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Transfers densities from PW to RS grids and potentials from PW to RS
!> \par History
!>      - Copied from qs_coolocate_Density and qs_integrate_potenntial
!> \author JGH (04.2014)
! **************************************************************************************************
MODULE rs_pw_interface
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_spline_utils,                 ONLY: pw_prolongate_s3,&
                                              pw_restrict_s3
   USE gaussian_gridlevels,             ONLY: gridlevel_info_type
   USE input_constants,                 ONLY: pw_interp,&
                                              spline3_pbc_interp
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_p_type,&
                                              pw_pools_create_pws,&
                                              pw_pools_give_back_pws
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE realspace_grid_types,            ONLY: pw2rs,&
                                              realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              rs2pw,&
                                              rs_grid_release,&
                                              rs_pw_transfer
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rs_pw_interface'
! *** Public subroutines ***

   PUBLIC :: density_rs2pw, &
             density_rs2pw_basic, &
             potential_pw2rs

CONTAINS

! **************************************************************************************************
!> \brief given partial densities on the realspace multigrids,
!>      computes the full density on the plane wave grids, both in real and
!>      gspace
!> \param pw_env ...
!> \param rs_rho ...
!> \param rho ...
!> \param rho_gspace ...
!> \note
!>      should contain all communication in the collocation of the density
!>      in the case of replicated grids
! **************************************************************************************************
   SUBROUTINE density_rs2pw(pw_env, rs_rho, rho, rho_gspace)

      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho, rho_gspace

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_rs2pw', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, igrid_level, interp_kind
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: mgrid_gspace, mgrid_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs

      CALL timeset(routineN, handle)
      NULLIFY (gridlevel_info, mgrid_gspace, mgrid_rspace, rs_descs, pw_pools)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("rs_pw_interface.F",85)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, pw_pools=pw_pools)

      gridlevel_info => pw_env%gridlevel_info

      CALL section_vals_val_get(pw_env%interp_section, "KIND", i_val=interp_kind)

      CALL pw_pools_create_pws(pw_pools, mgrid_rspace, &
                               use_data=REALDATA3D, &
                               in_space=REALSPACE)

      CALL pw_pools_create_pws(pw_pools, mgrid_gspace, &
                               use_data=COMPLEXDATA1D, &
                               in_space=RECIPROCALSPACE)

      IF (gridlevel_info%ngrid_levels == 1) THEN
         CALL rs_pw_transfer(rs_rho(1)%rs_grid, rho%pw, rs2pw)
         CALL rs_grid_release(rs_rho(1)%rs_grid)
         CALL pw_transfer(rho%pw, rho_gspace%pw)
         IF (rho%pw%pw_grid%spherical) THEN ! rho_gspace = rho
            CALL pw_transfer(rho_gspace%pw, rho%pw)
         ENDIF
      ELSE
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            CALL rs_pw_transfer(rs_rho(igrid_level)%rs_grid, &
                                mgrid_rspace(igrid_level)%pw, rs2pw)
            CALL rs_grid_release(rs_rho(igrid_level)%rs_grid)
         ENDDO

         ! we want both rho and rho_gspace, the latter for Hartree and co-workers.
         SELECT CASE (interp_kind)
         CASE (pw_interp)
            CALL pw_zero(rho_gspace%pw)
            DO igrid_level = 1, gridlevel_info%ngrid_levels
               CALL pw_transfer(mgrid_rspace(igrid_level)%pw, &
                                mgrid_gspace(igrid_level)%pw)
               CALL pw_axpy(mgrid_gspace(igrid_level)%pw, rho_gspace%pw)
            END DO
            CALL pw_transfer(rho_gspace%pw, rho%pw)
         CASE (spline3_pbc_interp)
            DO igrid_level = gridlevel_info%ngrid_levels, 2, -1
               CALL pw_prolongate_s3(mgrid_rspace(igrid_level)%pw, &
                                     mgrid_rspace(igrid_level-1)%pw, pw_pools(igrid_level)%pool, &
                                     pw_env%interp_section)
            END DO
            CALL pw_copy(mgrid_rspace(1)%pw, rho%pw)
            CALL pw_transfer(rho%pw, rho_gspace%pw)
         CASE default
            CALL cp_abort(cp__l("rs_pw_interface.F",133), &
                          "interpolator "// &
                          cp_to_string(interp_kind))
         END SELECT
      END IF

      ! *** give back the pw multi-grids
      CALL pw_pools_give_back_pws(pw_pools, mgrid_gspace)
      CALL pw_pools_give_back_pws(pw_pools, mgrid_rspace)
      CALL timestop(handle)

   END SUBROUTINE density_rs2pw

! **************************************************************************************************
!> \brief given partial densities on the realspace multigrids,
!>      computes the full density on the plane wave grids
!> \param pw_env ...
!> \param rs_rho ...
!> \param rho ...
!> \param rho_gspace ...
!> \note
!>      should contain the all communication in the collocation of the density
!>      in the case of replicated grids
! **************************************************************************************************
   SUBROUTINE density_rs2pw_basic(pw_env, rs_rho, rho, rho_gspace)

      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho, rho_gspace

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_rs2pw_basic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, igrid_level, interp_kind
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: mgrid_gspace, mgrid_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs

      CALL timeset(routineN, handle)
      NULLIFY (gridlevel_info, mgrid_gspace, mgrid_rspace, rs_descs, pw_pools)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("rs_pw_interface.F",175)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, pw_pools=pw_pools)

      gridlevel_info => pw_env%gridlevel_info

      CALL section_vals_val_get(pw_env%interp_section, "KIND", i_val=interp_kind)

      CALL pw_pools_create_pws(pw_pools, mgrid_rspace, &
                               use_data=REALDATA3D, &
                               in_space=REALSPACE)

      CALL pw_pools_create_pws(pw_pools, mgrid_gspace, &
                               use_data=COMPLEXDATA1D, &
                               in_space=RECIPROCALSPACE)

      IF (gridlevel_info%ngrid_levels == 1) THEN
         CALL rs_pw_transfer(rs_rho(1)%rs_grid, rho%pw, rs2pw)
         CALL pw_transfer(rho%pw, rho_gspace%pw)
      ELSE
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            CALL rs_pw_transfer(rs_rho(igrid_level)%rs_grid, &
                                mgrid_rspace(igrid_level)%pw, rs2pw)
         ENDDO

         ! we want both rho and rho_gspace, the latter for Hartree and co-workers.
         SELECT CASE (interp_kind)
         CASE (pw_interp)
            DO igrid_level = 1, gridlevel_info%ngrid_levels
               CALL pw_transfer(mgrid_rspace(igrid_level)%pw, &
                                mgrid_gspace(igrid_level)%pw)
               IF (igrid_level /= 1) THEN
                  CALL pw_axpy(mgrid_gspace(igrid_level)%pw, mgrid_gspace(1)%pw)
               END IF
            END DO
            CALL pw_transfer(mgrid_gspace(1)%pw, rho%pw)
            CALL pw_transfer(mgrid_rspace(1)%pw, rho_gspace%pw)
         CASE (spline3_pbc_interp)
            DO igrid_level = gridlevel_info%ngrid_levels, 2, -1
               CALL pw_prolongate_s3(mgrid_rspace(igrid_level)%pw, &
                                     mgrid_rspace(igrid_level-1)%pw, pw_pools(igrid_level)%pool, &
                                     pw_env%interp_section)
            END DO
            CALL pw_copy(mgrid_rspace(1)%pw, rho%pw)
            CALL pw_transfer(rho%pw, rho_gspace%pw)
         CASE default
            CALL cp_abort(cp__l("rs_pw_interface.F",220), &
                          "interpolator "// &
                          cp_to_string(interp_kind))
         END SELECT
      END IF

      ! *** give back the pw multi-grids
      CALL pw_pools_give_back_pws(pw_pools, mgrid_gspace)
      CALL pw_pools_give_back_pws(pw_pools, mgrid_rspace)
      CALL timestop(handle)

   END SUBROUTINE density_rs2pw_basic

! **************************************************************************************************
!> \brief transfers a potential from a pw_grid to a vector of
!>      realspace multigrids
!> \param rs_v OUTPUT: the potential on the realspace multigrids
!> \param v_rspace INPUT : the potential on a planewave grid in Rspace
!> \param pw_env ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      extracted from integrate_v_rspace
!>      should contain all parallel communication of integrate_v_rspace in the
!>      case of replicated grids.
! **************************************************************************************************
   SUBROUTINE potential_pw2rs(rs_v, v_rspace, pw_env)

      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_v
      TYPE(pw_p_type), INTENT(IN)                        :: v_rspace
      TYPE(pw_env_type), POINTER                         :: pw_env

      CHARACTER(len=*), PARAMETER :: routineN = 'potential_pw2rs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: auxbas_grid, handle, igrid_level, &
                                                            interp_kind
      REAL(KIND=dp)                                      :: scale
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: mgrid_gspace, mgrid_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools

      CALL timeset(routineN, handle)

      ! *** set up of the potential on the multigrids
      CALL pw_env_get(pw_env, pw_pools=pw_pools, gridlevel_info=gridlevel_info, &
                      auxbas_grid=auxbas_grid)

      CALL pw_pools_create_pws(pw_pools, mgrid_rspace, &
                               use_data=REALDATA3D, &
                               in_space=REALSPACE)

      ! use either realspace or fft techniques to get the potential on the rs multigrids
      CALL section_vals_val_get(pw_env%interp_section, "KIND", i_val=interp_kind)
      SELECT CASE (interp_kind)
      CASE (pw_interp)
         CALL pw_pools_create_pws(pw_pools, mgrid_gspace, &
                                  use_data=COMPLEXDATA1D, &
                                  in_space=RECIPROCALSPACE)
         CALL pw_transfer(v_rspace%pw, mgrid_gspace(auxbas_grid)%pw)
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            IF (igrid_level /= auxbas_grid) THEN
               CALL pw_copy(mgrid_gspace(auxbas_grid)%pw, mgrid_gspace(igrid_level)%pw)
               CALL pw_transfer(mgrid_gspace(igrid_level)%pw, mgrid_rspace(igrid_level)%pw)
            ELSE
               IF (mgrid_gspace(auxbas_grid)%pw%pw_grid%spherical) THEN
                  CALL pw_transfer(mgrid_gspace(auxbas_grid)%pw, mgrid_rspace(auxbas_grid)%pw)
               ELSE ! fft forward + backward should be identical
                  CALL pw_copy(v_rspace%pw, mgrid_rspace(auxbas_grid)%pw)
               ENDIF
            ENDIF
            ! *** Multiply by the grid volume element ratio ***
            IF (igrid_level /= auxbas_grid) THEN
               scale = mgrid_rspace(igrid_level)%pw%pw_grid%dvol/ &
                       mgrid_rspace(auxbas_grid)%pw%pw_grid%dvol
               mgrid_rspace(igrid_level)%pw%cr3d = &
                  scale*mgrid_rspace(igrid_level)%pw%cr3d
            END IF
         END DO
         CALL pw_pools_give_back_pws(pw_pools, mgrid_gspace)
      CASE (spline3_pbc_interp)
         CALL pw_copy(v_rspace%pw, mgrid_rspace(1)%pw)
         DO igrid_level = 1, gridlevel_info%ngrid_levels-1
            CALL pw_zero(mgrid_rspace(igrid_level+1)%pw)
            CALL pw_restrict_s3(mgrid_rspace(igrid_level)%pw, &
                                mgrid_rspace(igrid_level+1)%pw, pw_pools(igrid_level+1)%pool, &
                                pw_env%interp_section)
            ! *** Multiply by the grid volume element ratio
            mgrid_rspace(igrid_level+1)%pw%cr3d = &
               mgrid_rspace(igrid_level+1)%pw%cr3d*8._dp
         END DO
      CASE default
         CALL cp_abort(cp__l("rs_pw_interface.F",312), &
                       "interpolation not supported "// &
                       cp_to_string(interp_kind))
      END SELECT

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_pw_transfer(rs_v(igrid_level)%rs_grid, &
                             mgrid_rspace(igrid_level)%pw, pw2rs)
      ENDDO
      ! *** give back the pw multi-grids
      CALL pw_pools_give_back_pws(pw_pools, mgrid_rspace)

      CALL timestop(handle)

   END SUBROUTINE potential_pw2rs

END MODULE rs_pw_interface
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief generate the tasks lists used by collocate and integrate routines
!> \par History
!>      01.2008 [Joost VandeVondele] refactered out of qs_collocate / qs_integrate
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE task_list_methods
   USE ao_util,                         ONLY: exp_radius_very_extended
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        convert_sizes_to_offsets, cp_dbcsr_col_block_sizes, cp_dbcsr_finalize, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_info, cp_dbcsr_p_type, cp_dbcsr_put_block, &
        cp_dbcsr_row_block_sizes, cp_dbcsr_type, cp_dbcsr_work_create
   USE cube_utils,                      ONLY: compute_cube_center,&
                                              cube_info_type,&
                                              return_cube,&
                                              return_cube_nonortho
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel,&
                                              gridlevel_info_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_alltoall,&
                                              mp_bcast,&
                                              mp_sum,&
                                              mp_sum_partial
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_desc_type,&
                                              realspace_grid_p_type,&
                                              rs_grid_create,&
                                              rs_grid_locate_rank,&
                                              rs_grid_release,&
                                              rs_grid_reorder_ranks
   USE task_list_types,                 ONLY: task_list_type
   USE util,                            ONLY: sort

!$ USE OMP_LIB, ONLY: omp_destroy_lock, omp_get_num_threads, omp_init_lock,&
!$                    omp_lock_kind, omp_set_lock, omp_unset_lock
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'task_list_methods'

   PUBLIC :: generate_qs_task_list, &
             task_list_inner_loop
   PUBLIC :: distribute_tasks, &
             int2pair, &
             rs_distribute_matrix

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ks_env ...
!> \param task_list ...
!> \param reorder_rs_grid_ranks Flag that indicates if this routine should
!>        or should not overwrite the rs descriptor (see comment below)
!> \param skip_load_balance_distributed ...
!> \param soft_valid ...
!> \param basis_type ...
!> \param pw_env_external ...
!> \param sab_orb_external ...
!> \par History
!>      01.2008 factored out of calculate_rho_elec [Joost VandeVondele]
!>      04.2010 divides tasks into grid levels and atom pairs for integrate/collocate [Iain Bethune]
!>              (c) The Numerical Algorithms Group (NAG) Ltd, 2010 on behalf of the HECToR project
!>      06.2015 adjusted to be used with multiple images (k-points) [JGH]
!> \note  If this routine is called several times with different task lists,
!>        the default behaviour is to re-optimize the grid ranks and overwrite
!>        the rs descriptor and grids. reorder_rs_grid_ranks = .FALSE. prevents the code
!>        of performing a new optimization by leaving the rank order in
!>        its current state.
! **************************************************************************************************
   SUBROUTINE generate_qs_task_list(ks_env, task_list, &
                                    reorder_rs_grid_ranks, skip_load_balance_distributed, &
                                    soft_valid, basis_type, pw_env_external, sab_orb_external)

      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(task_list_type), POINTER                      :: task_list
      LOGICAL, INTENT(IN)                                :: reorder_rs_grid_ranks, &
                                                            skip_load_balance_distributed
      LOGICAL, INTENT(IN), OPTIONAL                      :: soft_valid
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: basis_type
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env_external
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sab_orb_external

      CHARACTER(LEN=*), PARAMETER :: routineN = 'generate_qs_task_list', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: max_tasks = 2000

      CHARACTER(LEN=default_string_length)               :: my_basis_type
      INTEGER :: cindex, curr_tasks, handle, i, iatom, iatom_old, igrid_level, igrid_level_old, &
         ikind, ilevel, img, img_old, inode, ipair, ipgf, iset, itask, jatom, jatom_old, jkind, &
         jpgf, jset, maxpgf, maxset, natoms, nimages, nkind, nseta, nsetb
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: blocks
      INTEGER, DIMENSION(3)                              :: cellind
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: dokp, my_soft
      REAL(KIND=dp)                                      :: kind_radius_a, kind_radius_b
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, zeta, zetb
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b, orb_basis_set
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_grids

      CALL timeset(routineN, handle)

      ! by default, the full density is calculated
      my_soft = .FALSE.
      IF (PRESENT(soft_valid)) my_soft = soft_valid

      CALL get_ks_env(ks_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      particle_set=particle_set, &
                      dft_control=dft_control)

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      CALL get_ks_env(ks_env, sab_orb=sab_orb)
      IF (PRESENT(sab_orb_external)) sab_orb => sab_orb_external

      CALL get_ks_env(ks_env, pw_env=pw_env)
      IF (PRESENT(pw_env_external)) pw_env => pw_env_external
      CALL pw_env_get(pw_env, rs_descs=rs_descs, rs_grids=rs_grids)

      ! *** assign from pw_env
      gridlevel_info => pw_env%gridlevel_info
      cube_info => pw_env%cube_info

      ! find maximum numbers
      nkind = SIZE(qs_kind_set)
      natoms = SIZE(particle_set)
      maxset = 0
      maxpgf = 0
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, &
                          softb=my_soft, &
                          basis_set=orb_basis_set, basis_type=my_basis_type)

         IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, npgf=npgfa, nset=nseta)

         maxset = MAX(nseta, maxset)
         maxpgf = MAX(MAXVAL(npgfa), maxpgf)
      END DO

      ! kpoint related
      nimages = dft_control%nimages
      IF (nimages > 1) THEN
         dokp = .TRUE.
         NULLIFY (kpoints)
         CALL get_ks_env(ks_env=ks_env, kpoints=kpoints)
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      ELSE
         dokp = .FALSE.
         NULLIFY (cell_to_index)
      END IF

      ! free the atom_pair lists if allocated
      IF (ASSOCIATED(task_list%atom_pair_send)) DEALLOCATE (task_list%atom_pair_send)
      IF (ASSOCIATED(task_list%atom_pair_recv)) DEALLOCATE (task_list%atom_pair_recv)

      ! construct a list of all tasks
      IF (.NOT. ASSOCIATED(task_list%tasks)) THEN
         CALL reallocate(task_list%tasks, 1, 6, 1, max_tasks)
         CALL reallocate(task_list%dist_ab, 1, 3, 1, max_tasks)
      ENDIF
      task_list%ntasks = 0
      curr_tasks = SIZE(task_list%tasks, 2)

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, softb=my_soft, basis_set=basis_set_a, &
                          basis_type=my_basis_type)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab, cell=cellind)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra(:) = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         kind_radius_a = basis_set_a%kind_radius
         zeta => basis_set_a%zet
         ! basis jkind
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         kind_radius_b = basis_set_b%kind_radius
         zetb => basis_set_b%zet

         IF (dokp) THEN
            cindex = cell_to_index(cellind(1), cellind(2), cellind(3))
         ELSE
            cindex = 1
         END IF

         CALL task_list_inner_loop(task_list%tasks, task_list%dist_ab, task_list%ntasks, curr_tasks, &
                                   rs_descs, dft_control, cube_info, gridlevel_info, cindex, &
                                   iatom, jatom, rpgfa, rpgfb, zeta, zetb, kind_radius_b, set_radius_a, set_radius_b, ra, rab, &
                                   la_max, la_min, lb_max, lb_min, npgfa, npgfb, maxpgf, maxset, natoms, nimages, nseta, nsetb)

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (basis_set_list)

      ! redistribute the task list so that all tasks map on the local rs grids
      CALL distribute_tasks( &
         rs_descs, task_list%ntasks, natoms, maxset, maxpgf, nimages, &
         task_list%tasks, rval=task_list%dist_ab, atom_pair_send=task_list%atom_pair_send, &
         atom_pair_recv=task_list%atom_pair_recv, symmetric=.TRUE., &
         reorder_rs_grid_ranks=reorder_rs_grid_ranks, skip_load_balance_distributed=skip_load_balance_distributed)

      ! If the rank order has changed, reallocate any of the distributed rs_grids

      IF (reorder_rs_grid_ranks) THEN
         DO i = 1, gridlevel_info%ngrid_levels
            IF (rs_descs(i)%rs_desc%distributed) THEN
               CALL rs_grid_release(rs_grids(i)%rs_grid)
               NULLIFY (rs_grids(i)%rs_grid)
               CALL rs_grid_create(rs_grids(i)%rs_grid, rs_descs(i)%rs_desc)
            END IF
         END DO
      END IF

      ! Now we have the final list of tasks, setup the task_list with the
      ! data needed for the loops in integrate_v/calculate_rho

      IF (ASSOCIATED(task_list%taskstart)) THEN
         DEALLOCATE (task_list%taskstart)
      END IF
      IF (ASSOCIATED(task_list%taskstop)) THEN
         DEALLOCATE (task_list%taskstop)
      END IF
      IF (ASSOCIATED(task_list%npairs)) THEN
         DEALLOCATE (task_list%npairs)
      END IF

      ! First, count the number of unique atom pairs per grid level

      ALLOCATE (task_list%npairs(SIZE(rs_descs)))

      iatom_old = -1; jatom_old = -1; igrid_level_old = -1; img_old = -1
      ipair = 0
      task_list%npairs = 0

      DO i = 1, task_list%ntasks
         CALL int2pair(task_list%tasks(3, i), igrid_level, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natoms, maxset, maxpgf)
         IF (igrid_level .NE. igrid_level_old) THEN
            IF (igrid_level_old .NE. -1) THEN
               task_list%npairs(igrid_level_old) = ipair
            END IF
            ipair = 1
            igrid_level_old = igrid_level
            iatom_old = iatom
            jatom_old = jatom
            img_old = img
         ELSE IF (iatom .NE. iatom_old .OR. jatom .NE. jatom_old .OR. img .NE. img_old) THEN
            ipair = ipair+1
            iatom_old = iatom
            jatom_old = jatom
            img_old = img
         END IF
      END DO
      ! Take care of the last iteration
      IF (task_list%ntasks /= 0) THEN
         task_list%npairs(igrid_level) = ipair
      END IF

      ! Second, for each atom pair, find the indices in the task list
      ! of the first and last task

      ! Array sized for worst case
      ALLOCATE (task_list%taskstart(MAXVAL(task_list%npairs), SIZE(rs_descs)))
      ALLOCATE (task_list%taskstop(MAXVAL(task_list%npairs), SIZE(rs_descs)))

      iatom_old = -1; jatom_old = -1; igrid_level_old = -1; img_old = -1
      ipair = 0
      task_list%taskstart = 0
      task_list%taskstop = 0

      DO i = 1, task_list%ntasks
         CALL int2pair(task_list%tasks(3, i), igrid_level, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natoms, maxset, maxpgf)
         IF (igrid_level .NE. igrid_level_old) THEN
            IF (igrid_level_old .NE. -1) THEN
               task_list%taskstop(ipair, igrid_level_old) = i-1
            END IF
            ipair = 1
            task_list%taskstart(ipair, igrid_level) = i
            igrid_level_old = igrid_level
            iatom_old = iatom
            jatom_old = jatom
            img_old = img
         ELSE IF (iatom .NE. iatom_old .OR. jatom .NE. jatom_old .OR. img .NE. img_old) THEN
            ipair = ipair+1
            task_list%taskstart(ipair, igrid_level) = i
            task_list%taskstop(ipair-1, igrid_level) = i-1
            iatom_old = iatom
            jatom_old = jatom
            img_old = img
         END IF
      END DO
      ! Take care of the last iteration
      IF (task_list%ntasks /= 0) THEN
         task_list%taskstop(ipair, igrid_level) = task_list%ntasks
      END IF

      ! Debug task destribution
      IF (debug_this_module) THEN
         tasks => task_list%tasks
         WRITE (6, *)
         WRITE (6, *) "Total number of tasks              ", task_list%ntasks
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            WRITE (6, *) "Total number of pairs(grid_level)  ", igrid_level, task_list%npairs(igrid_level)
         END DO
         WRITE (6, *)

         DO igrid_level = 1, gridlevel_info%ngrid_levels

            ALLOCATE (blocks(natoms, natoms, nimages))
            blocks = -1
            DO ipair = 1, task_list%npairs(igrid_level)
               itask = task_list%taskstart(ipair, igrid_level)
               CALL int2pair(tasks(3, itask), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                             nimages, natoms, maxset, maxpgf)
               IF (blocks(iatom, jatom, img) == -1 .AND. blocks(jatom, iatom, img) == -1) THEN
                  blocks(iatom, jatom, img) = 1
                  blocks(jatom, iatom, img) = 1
               ELSE
                  WRITE (6, *) "TASK LIST CONFLICT IN PAIR       ", ipair
                  WRITE (6, *) "Reuse of iatom, jatom, image     ", iatom, jatom, img
               END IF

               iatom_old = iatom
               jatom_old = jatom
               img_old = img
               DO itask = task_list%taskstart(ipair, igrid_level), task_list%taskstop(ipair, igrid_level)

                  CALL int2pair(tasks(3, itask), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                                nimages, natoms, maxset, maxpgf)
                  IF (iatom /= iatom_old .OR. jatom /= jatom_old .OR. img /= img_old) THEN
                     WRITE (6, *) "TASK LIST CONFLICT IN TASK       ", itask
                     WRITE (6, *) "Inconsistent iatom, jatom, image ", iatom, jatom, img
                     WRITE (6, *) "Should be    iatom, jatom, image ", iatom_old, jatom_old, img_old
                  END IF

               END DO
            END DO
            DEALLOCATE (blocks)

         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE generate_qs_task_list

! **************************************************************************************************
!> \brief ...
!> \param tasks ...
!> \param dist_ab ...
!> \param ntasks ...
!> \param curr_tasks ...
!> \param rs_descs ...
!> \param dft_control ...
!> \param cube_info ...
!> \param gridlevel_info ...
!> \param cindex ...
!> \param iatom ...
!> \param jatom ...
!> \param rpgfa ...
!> \param rpgfb ...
!> \param zeta ...
!> \param zetb ...
!> \param kind_radius_b ...
!> \param set_radius_a ...
!> \param set_radius_b ...
!> \param ra ...
!> \param rab ...
!> \param la_max ...
!> \param la_min ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfa ...
!> \param npgfb ...
!> \param maxpgf ...
!> \param maxset ...
!> \param natoms ...
!> \param nimages ...
!> \param nseta ...
!> \param nsetb ...
!> \par History
!>      Joost VandeVondele: 10.2008 refactored
! **************************************************************************************************
   SUBROUTINE task_list_inner_loop(tasks, dist_ab, ntasks, curr_tasks, rs_descs, dft_control, &
                                   cube_info, gridlevel_info, cindex, &
                                   iatom, jatom, rpgfa, rpgfb, zeta, zetb, kind_radius_b, set_radius_a, set_radius_b, ra, rab, &
                                   la_max, la_min, lb_max, lb_min, npgfa, npgfb, maxpgf, maxset, natoms, nimages, nseta, nsetb)

      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dist_ab
      INTEGER                                            :: ntasks, curr_tasks
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      INTEGER                                            :: cindex, iatom, jatom
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, zeta, zetb
      REAL(KIND=dp)                                      :: kind_radius_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb
      INTEGER                                            :: maxpgf, maxset, natoms, nimages, nseta, &
                                                            nsetb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'task_list_inner_loop', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cube_center(3), igrid_level, ipgf, iset, &
                                                            jpgf, jset, lb_cube(3), ub_cube(3)
      REAL(KIND=dp)                                      :: dab, rab2, radius, zetp

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      dab = SQRT(rab2)

      loop_iset: DO iset = 1, nseta

         IF (set_radius_a(iset)+kind_radius_b < dab) CYCLE

         loop_jset: DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            loop_ipgf: DO ipgf = 1, npgfa(iset)

               IF (rpgfa(ipgf, iset)+set_radius_b(jset) < dab) CYCLE

               loop_jpgf: DO jpgf = 1, npgfb(jset)

                  IF (rpgfa(ipgf, iset)+rpgfb(jpgf, jset) < dab) CYCLE

                  zetp = zeta(ipgf, iset)+zetb(jpgf, jset)
                  igrid_level = gaussian_gridlevel(gridlevel_info, zetp)

                  CALL compute_pgf_properties(cube_center, lb_cube, ub_cube, radius, &
                                              rs_descs(igrid_level)%rs_desc, cube_info(igrid_level), &
                                              la_max(iset), zeta(ipgf, iset), la_min(iset), &
                                              lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                              ra, rab, rab2, dft_control%qs_control%eps_rho_rspace)

                  CALL pgf_to_tasks(tasks, dist_ab, ntasks, curr_tasks, &
                                    rab, cindex, iatom, jatom, iset, jset, ipgf, jpgf, nimages, natoms, maxset, maxpgf, &
                                    la_max(iset), lb_max(jset), rs_descs(igrid_level)%rs_desc, &
                                    igrid_level, gridlevel_info%ngrid_levels, cube_center, &
                                    lb_cube, ub_cube)

               END DO loop_jpgf

            END DO loop_ipgf

         END DO loop_jset

      END DO loop_iset

   END SUBROUTINE task_list_inner_loop

! **************************************************************************************************
!> \brief combines the calculation of several basic properties of a given pgf:
!>  its center, the bounding cube, the radius, the cost,
!>  tries to predict the time needed for processing this task
!>      in this way an improved load balance might be obtained
!> \param cube_center ...
!> \param lb_cube ...
!> \param ub_cube ...
!> \param radius ...
!> \param rs_desc ...
!> \param cube_info ...
!> \param la_max ...
!> \param zeta ...
!> \param la_min ...
!> \param lb_max ...
!> \param zetb ...
!> \param lb_min ...
!> \param ra ...
!> \param rab ...
!> \param rab2 ...
!> \param eps ...
!> \par History
!>      10.2008 refactored [Joost VandeVondele]
!> \note
!>      -) this requires the radius to be computed in the same way as
!>      collocate_pgf_product_rspace, we should factor that part into a subroutine
!>      -) we're assuming that integrate_pgf and collocate_pgf are the same cost for load balancing
!>         this is more or less true for map_consistent
!>      -) in principle, the computed radius could be recycled in integrate_pgf/collocate_pgf if it is certainly
!>         the same, this could lead to a small speedup
!>      -) the cost function is a fit through the median cost of mapping a pgf with a given l and a given radius (in grid points)
!>         fitting the measured data on an opteron/g95 using the expression
!>         a*(l+b)(r+c)**3+d which is based on the innerloop of the collocating routines
! **************************************************************************************************
   SUBROUTINE compute_pgf_properties(cube_center, lb_cube, ub_cube, radius, &
                                     rs_desc, cube_info, la_max, zeta, la_min, lb_max, zetb, lb_min, ra, rab, rab2, eps)

      INTEGER, DIMENSION(3), INTENT(OUT)                 :: cube_center, lb_cube, ub_cube
      REAL(KIND=dp), INTENT(OUT)                         :: radius
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(cube_info_type), INTENT(IN)                   :: cube_info
      INTEGER, INTENT(IN)                                :: la_max
      REAL(KIND=dp), INTENT(IN)                          :: zeta
      INTEGER, INTENT(IN)                                :: la_min, lb_max
      REAL(KIND=dp), INTENT(IN)                          :: zetb
      INTEGER, INTENT(IN)                                :: lb_min
      REAL(KIND=dp), INTENT(IN)                          :: ra(3), rab(3), rab2, eps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_pgf_properties', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: extent(3)
      INTEGER, DIMENSION(:), POINTER                     :: sphere_bounds
      REAL(KIND=dp)                                      :: cutoff, f, prefactor, rb(3), zetp
      REAL(KIND=dp), DIMENSION(3)                        :: rp

! the radius for this task

      zetp = zeta+zetb
      rp(:) = ra(:)+zetb/zetp*rab(:)
      rb(:) = ra(:)+rab(:)
      cutoff = 1.0_dp
      f = zetb/zetp
      prefactor = EXP(-zeta*f*rab2)
      radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, ra=ra, rb=rb, rp=rp, &
                                        zetp=zetp, eps=eps, prefactor=prefactor, cutoff=cutoff)

      CALL compute_cube_center(cube_center, rs_desc, zeta, zetb, ra, rab)
      ! compute cube_center, the center of the gaussian product to map (folded to within the unit cell)
      cube_center(:) = MODULO(cube_center(:), rs_desc%npts(:))
      cube_center(:) = cube_center(:)+rs_desc%lb(:)

      IF (rs_desc%orthorhombic) THEN
         CALL return_cube(cube_info, radius, lb_cube, ub_cube, sphere_bounds)
      ELSE
         CALL return_cube_nonortho(cube_info, radius, lb_cube, ub_cube, rp)
         !? unclear if extent is computed correctly.
         extent(:) = ub_cube(:)-lb_cube(:)
         lb_cube(:) = -extent(:)/2-1
         ub_cube(:) = extent(:)/2
      ENDIF

   END SUBROUTINE compute_pgf_properties
! **************************************************************************************************
!> \brief predicts the cost of a task in kcycles for a given task
!>        the model is based on a fit of actual data, and might need updating
!>        as collocate_pgf_product_rspace changes (or CPUs/compilers change)
!>        maybe some dynamic approach, improving the cost model on the fly could
!>        work as well
!>        the cost model does not yet take into account the fraction of space
!>        that is mapped locally for a given cube and rs_grid (generalised tasks)
!> \param lb_cube ...
!> \param ub_cube ...
!> \param fraction ...
!> \param lmax ...
!> \param is_ortho ...
!> \retval cost_model ...
! **************************************************************************************************
   INTEGER FUNCTION cost_model(lb_cube, ub_cube, fraction, lmax, is_ortho)
      INTEGER, DIMENSION(3), INTENT(IN)       :: lb_cube, ub_cube
      REAL(KIND=dp), INTENT(IN)               :: fraction

      INTEGER :: cmax, lmax
      LOGICAL :: is_ortho
      REAL(KIND=dp) :: v1, v2, v3, v4, v5

      cmax = MAXVAL(((ub_cube-lb_cube)+1)/2)

      IF (is_ortho) THEN
         v1 = 1.504760E+00_dp
         v2 = 3.126770E+00_dp
         v3 = 5.074106E+00_dp
         v4 = 1.091568E+00_dp
         v5 = 1.070187E+00_dp
      ELSE
         v1 = 7.831105E+00_dp
         v2 = 2.675174E+00_dp
         v3 = 7.546553E+00_dp
         v4 = 6.122446E-01_dp
         v5 = 3.886382E+00_dp
      ENDIF
      cost_model = CEILING(((lmax+v1)*(cmax+v2)**3*v3*fraction+v4+v5*lmax**7)/1000.0_dp)

   END FUNCTION cost_model
! **************************************************************************************************
!> \brief pgf_to_tasks converts a given pgf to one or more tasks, in particular
!>        this determines by which CPUs a given pgf gets collocated
!>        the format of the task array is as follows
!>        tasks(1,i) := destination
!>        tasks(2,i) := source
!>        tasks(3,i) := compressed type (iatom, jatom, ....)
!>        tasks(4,i) := type (0: replicated, 1: distributed local, 2: distributed generalised)
!>        tasks(5,i) := cost
!>        tasks(6,i) := alternate destination code (0 if none availabe)
!>
!> \param tasks ...
!> \param dist_ab ...
!> \param ntasks ...
!> \param curr_tasks ...
!> \param rab ...
!> \param cindex ...
!> \param iatom ...
!> \param jatom ...
!> \param iset ...
!> \param jset ...
!> \param ipgf ...
!> \param jpgf ...
!> \param nimages ...
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \param la_max ...
!> \param lb_max ...
!> \param rs_desc ...
!> \param igrid_level ...
!> \param n_levels ...
!> \param cube_center ...
!> \param lb_cube ...
!> \param ub_cube ...
!> \par History
!>      10.2008 Refactored based on earlier routines by MattW [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE pgf_to_tasks(tasks, dist_ab, ntasks, curr_tasks, &
                           rab, cindex, iatom, jatom, iset, jset, ipgf, jpgf, nimages, natoms, maxset, maxpgf, &
                           la_max, lb_max, rs_desc, igrid_level, n_levels, cube_center, lb_cube, ub_cube)

      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dist_ab
      INTEGER, INTENT(INOUT)                             :: ntasks, curr_tasks
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      INTEGER, INTENT(IN)                                :: cindex, iatom, jatom, iset, jset, ipgf, &
                                                            jpgf, nimages, natoms, maxset, maxpgf, &
                                                            la_max, lb_max
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      INTEGER, INTENT(IN)                                :: igrid_level, n_levels
      INTEGER, DIMENSION(3), INTENT(IN)                  :: cube_center, lb_cube, ub_cube

      INTEGER, PARAMETER                                 :: add_tasks = 1000
      REAL(kind=dp), PARAMETER                           :: mult_tasks = 2.0_dp

      INTEGER                                            :: added_tasks, cost, j, lmax
      LOGICAL                                            :: is_ortho
      REAL(KIND=dp)                                      :: tfraction

      ntasks = ntasks+1
      IF (ntasks > curr_tasks) THEN
         curr_tasks = INT((curr_tasks+add_tasks)*mult_tasks)
         CALL reallocate(tasks, 1, 6, 1, curr_tasks)
      END IF

      IF (rs_desc%distributed) THEN

         ! finds the node(s) that need to process this task
         ! on exit tasks(4,:) is 1 for distributed tasks and 2 for generalised tasks
         CALL rs_find_node(rs_desc, igrid_level, n_levels, cube_center, &
                           ntasks=ntasks, tasks=tasks, lb_cube=lb_cube, ub_cube=ub_cube, added_tasks=added_tasks)

      ELSE
         tasks(1, ntasks) = encode_rank(rs_desc%my_pos, igrid_level, n_levels)
         tasks(4, ntasks) = 0
         tasks(6, ntasks) = 0
         added_tasks = 1
      ENDIF

      IF (SIZE(dist_ab, 2) .NE. SIZE(tasks, 2)) THEN
         CALL reallocate(dist_ab, 1, 3, 1, SIZE(tasks, 2))
      ENDIF

      lmax = la_max+lb_max
      is_ortho = (tasks(4, ntasks) == 0 .OR. tasks(4, ntasks) == 1) .AND. rs_desc%orthorhombic
      ! we assume the load is shared equally between processes dealing with a generalised Gaussian.
      ! this could be refined in the future
      tfraction = 1.0_dp/added_tasks

      cost = cost_model(lb_cube, ub_cube, tfraction, lmax, is_ortho)

      DO j = 1, added_tasks

         tasks(2, ntasks-added_tasks+j) = encode_rank(rs_desc%my_pos, igrid_level, n_levels)
         tasks(5, ntasks-added_tasks+j) = cost

         !encode the atom pairs and basis info as a single long integer
         CALL pair2int(tasks(3, ntasks-added_tasks+j), igrid_level, cindex, &
                       iatom, jatom, iset, jset, ipgf, jpgf, nimages, natoms, maxset, maxpgf)

         dist_ab(1, ntasks-added_tasks+j) = rab(1)
         dist_ab(2, ntasks-added_tasks+j) = rab(2)
         dist_ab(3, ntasks-added_tasks+j) = rab(3)

      ENDDO

   END SUBROUTINE pgf_to_tasks

! **************************************************************************************************
!> \brief converts a pgf index pair (ipgf,iset,iatom),(jpgf,jset,jatom)
!>      to a unique integer.
!>      a list of integers can be sorted, and will result in a list of pgf pairs
!>      for which all atom pairs are grouped, and for each atom pair all set pairs are grouped
!>      and for each set pair, all pgfs are grouped
!> \param res ...
!> \param ilevel ...
!> \param image ...
!> \param iatom ...
!> \param jatom ...
!> \param iset ...
!> \param jset ...
!> \param ipgf ...
!> \param jpgf ...
!> \param nimages ...
!> \param natom ...
!> \param maxset ...
!> \param maxpgf ...
!> \par History
!>      11.2007 created [Joost]
!> \note
!>      will hopefully not overflow any more
! **************************************************************************************************
   SUBROUTINE pair2int(res, ilevel, image, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natom, maxset, maxpgf)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: res
      INTEGER, INTENT(IN)                                :: ilevel, image, iatom, jatom, iset, jset, &
                                                            ipgf, jpgf, nimages, natom, maxset, &
                                                            maxpgf

      INTEGER(KIND=int_8)                                :: maxpgf8, maxset8, natom8, nimages8

      natom8 = natom; maxset8 = maxset; maxpgf8 = maxpgf; nimages8 = nimages
      !
      ! this encoding yields the right order of the tasks for collocation after the sort
      ! in distribute_tasks. E.g. for a atom pair, all sets and pgfs are computed in one go.
      ! The exception is the gridlevel. Tasks are first ordered wrt to grid_level. This implies
      ! that a given density matrix block will be decontracted several times, but cache effects on the
      ! grid make up for this.
      !
      res = ilevel*nimages8*natom8**2*maxset8**2*maxpgf8**2+ &
            (image-1)*natom8**2*maxset8**2*maxpgf8**2+ &
            ((iatom-1)*natom8+jatom-1)*maxset8**2*maxpgf8**2+ &
            ((iset-1)*maxset8+jset-1)*maxpgf8**2+ &
            (ipgf-1)*maxpgf8+jpgf-1

   END SUBROUTINE pair2int

! **************************************************************************************************
!> \brief ...
!> \param res ...
!> \param ilevel ...
!> \param image ...
!> \param iatom ...
!> \param jatom ...
!> \param iset ...
!> \param jset ...
!> \param ipgf ...
!> \param jpgf ...
!> \param nimages ...
!> \param natom ...
!> \param maxset ...
!> \param maxpgf ...
! **************************************************************************************************
   SUBROUTINE int2pair(res, ilevel, image, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natom, maxset, maxpgf)
      INTEGER(KIND=int_8), INTENT(IN)                    :: res
      INTEGER, INTENT(OUT)                               :: ilevel, image, iatom, jatom, iset, jset, &
                                                            ipgf, jpgf
      INTEGER, INTENT(IN)                                :: nimages, natom, maxset, maxpgf

      INTEGER(KIND=int_8)                                :: iatom8, ijatom, ijpgf, ijset, img, &
                                                            ipgf8, iset8, jatom8, jpgf8, jset8, &
                                                            maxpgf8, maxset8, natom8, nimages8, tmp

      natom8 = natom; maxset8 = maxset; maxpgf8 = maxpgf; nimages8 = nimages

      ilevel = INT(res/(nimages8*natom8**2*maxset8**2*maxpgf8**2))
      tmp = MOD(res, nimages8*natom8**2*maxset8**2*maxpgf8**2)
      img = tmp/(natom8**2*maxset8**2*maxpgf8**2)+1
      tmp = MOD(res, natom8**2*maxset8**2*maxpgf8**2)
      ijatom = tmp/(maxpgf8**2*maxset8**2)
      iatom8 = ijatom/natom8+1
      jatom8 = MOD(ijatom, natom8)+1
      tmp = MOD(tmp, maxpgf8**2*maxset8**2)
      ijset = tmp/maxpgf8**2
      iset8 = ijset/maxset8+1
      jset8 = MOD(ijset, maxset8)+1
      ijpgf = MOD(tmp, maxpgf8**2)
      ipgf8 = ijpgf/maxpgf8+1
      jpgf8 = MOD(ijpgf, maxpgf8)+1

      image = INT(img)
      iatom = INT(iatom8); jatom = INT(jatom8); iset = INT(iset8); jset = INT(jset8)
      ipgf = INT(ipgf8); jpgf = INT(jpgf8)

   END SUBROUTINE int2pair

! **************************************************************************************************
!> \brief performs load balancing of the tasks on the distributed grids
!> \param tasks ...
!> \param ntasks ...
!> \param rs_descs ...
!> \param grid_level ...
!> \param nimages ...
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \par History
!>      created 2008-10-03 [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE load_balance_distributed(tasks, ntasks, rs_descs, grid_level, &
                                       nimages, natoms, maxset, maxpgf)

      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER                                            :: ntasks
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: grid_level, nimages, natoms, maxset, &
                                                            maxpgf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'load_balance_distributed', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(:, :, :), POINTER               :: list

      CALL timeset(routineN, handle)

      NULLIFY (list)
      ! here we create for each cpu (0:ncpu-1) a list of possible destinations.
      ! if a destination would not be in this list, it is a bug
      CALL create_destination_list(list, rs_descs, grid_level)

      ! now, walk over the tasks, filling in the loads of each destination
      CALL compute_load_list(list, rs_descs, grid_level, tasks, ntasks, nimages, natoms, maxset, maxpgf, &
                             create_list=.TRUE.)

      ! optimize loads & fluxes
      CALL optimize_load_list(list, rs_descs(1)%rs_desc%group, rs_descs(1)%rs_desc%my_pos)

      ! now, walk over the tasks, using the list to set the destinations
      CALL compute_load_list(list, rs_descs, grid_level, tasks, ntasks, nimages, natoms, maxset, maxpgf, &
                             create_list=.FALSE.)

      DEALLOCATE (list)

      CALL timestop(handle)

   END SUBROUTINE load_balance_distributed

! **************************************************************************************************
!> \brief this serial routine adjusts the fluxes in the global list
!>
!> \param list_global ...
!> \par History
!>      created 2008-10-06 [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE balance_global_list(list_global)
      INTEGER, DIMENSION(:, :, 0:)                       :: list_global

      CHARACTER(LEN=*), PARAMETER :: routineN = 'balance_global_list', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: Max_Iter = 100
      REAL(KIND=dp), PARAMETER                           :: Tolerance_factor = 0.005_dp

      INTEGER                                            :: dest, handle, icpu, idest, iflux, &
                                                            ilocal, k, maxdest, Ncpu, Nflux
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: flux_connections
      LOGICAL                                            :: solution_optimal
      REAL(KIND=dp)                                      :: average, load_shift, max_load_shift, &
                                                            tolerance
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: load, optimized_flux, optimized_load
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: flux_limits

      CALL timeset(routineN, handle)

      Ncpu = SIZE(list_global, 3)
      maxdest = SIZE(list_global, 2)
      ALLOCATE (load(0:Ncpu-1))
      load = 0.0_dp
      ALLOCATE (optimized_load(0:Ncpu-1))

      ! figure out the number of fluxes
      ! we assume that the global_list is symmetric
      Nflux = 0
      DO icpu = 0, ncpu-1
         DO idest = 1, maxdest
            dest = list_global(1, idest, icpu)
            IF (dest < ncpu .AND. dest > icpu) Nflux = Nflux+1
         ENDDO
      ENDDO
      ALLOCATE (optimized_flux(Nflux))
      ALLOCATE (flux_limits(2, Nflux))
      ALLOCATE (flux_connections(2, Nflux))

      ! reorder data
      flux_limits = 0
      Nflux = 0
      DO icpu = 0, ncpu-1
         load(icpu) = SUM(list_global(2, :, icpu))
         DO idest = 1, maxdest
            dest = list_global(1, idest, icpu)
            IF (dest < ncpu) THEN
               IF (dest .NE. icpu) THEN
                  IF (dest > icpu) THEN
                     Nflux = Nflux+1
                     flux_limits(2, Nflux) = list_global(2, idest, icpu)
                     flux_connections(1, Nflux) = icpu
                     flux_connections(2, Nflux) = dest
                  ELSE
                     DO iflux = 1, Nflux
                        IF (flux_connections(1, iflux) == dest .AND. flux_connections(2, iflux) == icpu) THEN
                           flux_limits(1, iflux) = -list_global(2, idest, icpu)
                           EXIT
                        ENDIF
                     ENDDO
                  ENDIF
               ENDIF
            ENDIF
         ENDDO
      ENDDO

      solution_optimal = .FALSE.
      optimized_flux = 0.0_dp

      ! an iterative solver, if iterated till convergence the maximum load is minimal
      ! we terminate before things are fully converged, since this does show up in the timings
      ! once the largest shift becomes less than a small fraction of the average load, we're done
      ! we're perfectly happy if the load balance is within 1 percent or so
      ! the maximum load normally converges even faster
      average = SUM(load)/SIZE(load)
      tolerance = Tolerance_factor*average

      optimized_load(:) = load
      DO k = 1, Max_iter
         max_load_shift = 0.0_dp
         DO iflux = 1, Nflux
            load_shift = (optimized_load(flux_connections(1, iflux))-optimized_load(flux_connections(2, iflux)))/2
            load_shift = MAX(flux_limits(1, iflux)-optimized_flux(iflux), load_shift)
            load_shift = MIN(flux_limits(2, iflux)-optimized_flux(iflux), load_shift)
            max_load_shift = MAX(ABS(load_shift), max_load_shift)
            optimized_load(flux_connections(1, iflux)) = optimized_load(flux_connections(1, iflux))-load_shift
            optimized_load(flux_connections(2, iflux)) = optimized_load(flux_connections(2, iflux))+load_shift
            optimized_flux(iflux) = optimized_flux(iflux)+load_shift
         ENDDO
         IF (max_load_shift < tolerance) THEN
            solution_optimal = .TRUE.
            EXIT
         ENDIF
      ENDDO

      ! now adjust the load list to reflect the optimized fluxes
      ! reorder data
      Nflux = 0
      DO icpu = 0, ncpu-1
         DO idest = 1, maxdest
            IF (list_global(1, idest, icpu) == icpu) ilocal = idest
         ENDDO
         DO idest = 1, maxdest
            dest = list_global(1, idest, icpu)
            IF (dest < ncpu) THEN
               IF (dest .NE. icpu) THEN
                  IF (dest > icpu) THEN
                     Nflux = Nflux+1
                     IF (optimized_flux(Nflux) > 0) THEN
                        list_global(2, ilocal, icpu) = list_global(2, ilocal, icpu)+ &
                                                       list_global(2, idest, icpu)-NINT(optimized_flux(Nflux))
                        list_global(2, idest, icpu) = NINT(optimized_flux(Nflux))
                     ELSE
                        list_global(2, ilocal, icpu) = list_global(2, ilocal, icpu)+ &
                                                       list_global(2, idest, icpu)
                        list_global(2, idest, icpu) = 0
                     ENDIF
                  ELSE
                     DO iflux = 1, Nflux
                        IF (flux_connections(1, iflux) == dest .AND. flux_connections(2, iflux) == icpu) THEN
                           IF (optimized_flux(iflux) > 0) THEN
                              list_global(2, ilocal, icpu) = list_global(2, ilocal, icpu)+ &
                                                             list_global(2, idest, icpu)
                              list_global(2, idest, icpu) = 0
                           ELSE
                              list_global(2, ilocal, icpu) = list_global(2, ilocal, icpu)+ &
                                                             list_global(2, idest, icpu)+NINT(optimized_flux(iflux))
                              list_global(2, idest, icpu) = -NINT(optimized_flux(iflux))
                           ENDIF
                           EXIT
                        ENDIF
                     ENDDO
                  ENDIF
               ENDIF
            ENDIF
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE balance_global_list

! **************************************************************************************************
!> \brief this routine gets back optimized loads for all destinations
!>
!> \param list ...
!> \param group ...
!> \param my_pos ...
!> \par History
!>      created 2008-10-06 [Joost VandeVondele]
!>      Modified 2016-01   [EPCC] Reduce memory requirements on P processes
!>                                from O(P^2) to O(P)
! **************************************************************************************************
   SUBROUTINE optimize_load_list(list, group, my_pos)
      INTEGER, DIMENSION(:, :, 0:)                       :: list
      INTEGER, INTENT(IN)                                :: group, my_pos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'optimize_load_list', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: rank_of_root = 0

      INTEGER                                            :: handle, icpu, idest, maxdest, ncpu
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: load_all
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: load_partial
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: list_global

      CALL timeset(routineN, handle)

      ncpu = SIZE(list, 3)
      maxdest = SIZE(list, 2)

      !find total workload ...
      ALLOCATE (load_all(maxdest*ncpu))
      load_all(:) = RESHAPE(list(2, :, :), (/maxdest*ncpu/))
      CALL mp_sum(load_all(:), rank_of_root, group)

      ! ... and optimise the work per process
      ALLOCATE (list_global(2, maxdest, ncpu))
      IF (rank_of_root .EQ. my_pos) THEN
         list_global(1, :, :) = list(1, :, :)
         list_global(2, :, :) = RESHAPE(load_all, (/maxdest, ncpu/))
         CALL balance_global_list(list_global)
      ENDIF
      CALL mp_bcast(list_global, rank_of_root, group)

      !figure out how much can be sent to other processes
      ALLOCATE (load_partial(maxdest, ncpu))
      CALL mp_sum_partial(list(2, :, :), load_partial(:, :), group)

      DO icpu = 1, ncpu
         DO idest = 1, maxdest

            !need to deduct 1 because `list' was passed in to this routine as being indexed from zero
            IF (load_partial(idest, icpu) > list_global(2, idest, icpu)) THEN
               IF (load_partial(idest, icpu)-list(2, idest, icpu-1) < list_global(2, idest, icpu)) THEN
                  list(2, idest, icpu-1) = list_global(2, idest, icpu) &
                                           -(load_partial(idest, icpu)-list(2, idest, icpu-1))
               ELSE
                  list(2, idest, icpu-1) = 0
               ENDIF
            ENDIF

         ENDDO
      ENDDO

      !clean up before leaving
      DEALLOCATE (load_all)
      DEALLOCATE (list_global)
      DEALLOCATE (load_partial)

      CALL timestop(handle)
   END SUBROUTINE optimize_load_list

! **************************************************************************************************
!> \brief fill the load list with values derived from the tasks array
!>        from the alternate locations, we select the alternate location that
!>        can be used without increasing the number of matrix blocks needed to
!>        distribute.
!>        Replicated tasks are not yet considered
!>
!> \param list ...
!> \param rs_descs ...
!> \param grid_level ...
!> \param tasks ...
!> \param ntasks ...
!> \param nimages ...
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \param create_list ...
!> \par History
!>      created 2008-10-06 [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE compute_load_list(list, rs_descs, grid_level, tasks, &
                                ntasks, nimages, natoms, maxset, maxpgf, create_list)
      INTEGER, DIMENSION(:, :, 0:)                       :: list
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: grid_level
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER                                            :: ntasks, nimages, natoms, maxset, maxpgf
      LOGICAL                                            :: create_list

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_load_list', &
         routineP = moduleN//':'//routineN

      INTEGER :: cost, dest, handle, i, iatom, ilevel, img, img_old, iopt, ipgf, iset, itask, &
         itask_start, itask_stop, jatom, jpgf, jset, li, maxdest, ncpu, ndest_pair, nopt, nshort, &
         rank
      INTEGER(KIND=int_8)                                :: bit_pattern, ipair, ipair_old, natom8
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: loads
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: all_dests, index
      INTEGER, DIMENSION(6)                              :: options

      CALL timeset(routineN, handle)

      ALLOCATE (loads(0:rs_descs(grid_level)%rs_desc%group_size-1))
      CALL get_current_loads(loads, rs_descs, grid_level, ntasks, nimages, natoms, maxset, maxpgf, &
                             tasks, use_reordered_ranks=.FALSE.)

      maxdest = SIZE(list, 2)
      ncpu = SIZE(list, 3)
      natom8 = natoms

      ! first find the tasks that deal with the same atom pair
      itask_stop = 0
      ipair_old = HUGE(ipair_old)
      img_old = -1
      ALLOCATE (all_dests(0))
      ALLOCATE (INDEX(0))

      DO

         ! first find the range of tasks that deal with the same atom pair
         itask_start = itask_stop+1
         itask_stop = itask_start
         IF (itask_stop > ntasks) EXIT
         CALL int2pair(tasks(3, itask_stop), ilevel, img_old, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natoms, maxset, maxpgf)
         ipair_old = (iatom-1)*natom8+(jatom-1)
         DO
            IF (itask_stop+1 > ntasks) EXIT
            CALL int2pair(tasks(3, itask_stop+1), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natoms, maxset, maxpgf)
            ipair = (iatom-1)*natom8+(jatom-1)
            IF (ipair == ipair_old .AND. img == img_old) THEN
               itask_stop = itask_stop+1
            ELSE
               EXIT
            ENDIF
         ENDDO
         ipair = ipair_old
         nshort = itask_stop-itask_start+1

         ! find the unique list of destinations on this grid level only
         DEALLOCATE (all_dests)
         ALLOCATE (all_dests(nshort))
         DEALLOCATE (index)
         ALLOCATE (INDEX(nshort))
         DO i = 1, nshort
            CALL int2pair(tasks(3, itask_start+i-1), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natoms, maxset, maxpgf)
            IF (ilevel .EQ. grid_level) THEN
               all_dests(i) = decode_rank(tasks(1, itask_start+i-1), SIZE(rs_descs))
            ELSE
               all_dests(i) = HUGE(all_dests(i))
            END IF
         ENDDO
         CALL sort(all_dests, nshort, index)
         ndest_pair = 1
         DO i = 2, nshort
            IF ((all_dests(ndest_pair) .NE. all_dests(i)) .AND. (all_dests(i) .NE. HUGE(all_dests(i)))) THEN
               ndest_pair = ndest_pair+1
               all_dests(ndest_pair) = all_dests(i)
            ENDIF
         ENDDO

         DO itask = itask_start, itask_stop

            dest = decode_rank(tasks(1, itask), SIZE(rs_descs)) ! notice that dest can be changed
            CALL int2pair(tasks(3, itask), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natoms, maxset, maxpgf)
            ! Only proceed with tasks which are on this grid level
            IF (ilevel .NE. grid_level) CYCLE
            ipair = (iatom-1)*natom8+(jatom-1)
            cost = INT(tasks(5, itask))

            SELECT CASE (tasks (4, itask))
            CASE (1)
               bit_pattern = tasks(6, itask)
               nopt = 0
               IF (BTEST(bit_pattern, 0)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/-1, 0, 0/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (BTEST(bit_pattern, 1)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/+1, 0, 0/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (BTEST(bit_pattern, 2)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/0, -1, 0/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (BTEST(bit_pattern, 3)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/0, +1, 0/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (BTEST(bit_pattern, 4)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/0, 0, -1/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (BTEST(bit_pattern, 5)) THEN
                  rank = rs_grid_locate_rank(rs_descs(ilevel)%rs_desc, dest, (/0, 0, +1/))
                  IF (ANY(all_dests(1:ndest_pair) .EQ. rank)) THEN
                     nopt = nopt+1
                     options(nopt) = rank
                  ENDIF
               ENDIF
               IF (nopt > 0) THEN
                  ! set it to the rank with the lowest load
                  rank = options(1)
                  DO iopt = 2, nopt
                     IF (loads(rank) > loads(options(iopt))) rank = options(iopt)
                  ENDDO
               ELSE
                  rank = dest
               ENDIF
               li = list_index(list, rank, dest)
               IF (create_list) THEN
                  list(2, li, dest) = list(2, li, dest)+cost
               ELSE
                  IF (list(1, li, dest) == dest) THEN
                     tasks(1, itask) = encode_rank(dest, ilevel, SIZE(rs_descs))
                  ELSE
                     IF (list(2, li, dest) >= cost) THEN
                        list(2, li, dest) = list(2, li, dest)-cost
                        tasks(1, itask) = encode_rank(list(1, li, dest), ilevel, SIZE(rs_descs))
                     ELSE
                        tasks(1, itask) = encode_rank(dest, ilevel, SIZE(rs_descs))
                     ENDIF
                  ENDIF
               ENDIF
            CASE (2) ! generalised
               li = list_index(list, dest, dest)
               IF (create_list) THEN
                  list(2, li, dest) = list(2, li, dest)+cost
               ELSE
                  IF (list(1, li, dest) == dest) THEN
                     tasks(1, itask) = encode_rank(dest, ilevel, SIZE(rs_descs))
                  ELSE
                     IF (list(2, li, dest) >= cost) THEN
                        list(2, li, dest) = list(2, li, dest)-cost
                        tasks(1, itask) = encode_rank(list(1, li, dest), ilevel, SIZE(rs_descs))
                     ELSE
                        tasks(1, itask) = encode_rank(dest, ilevel, SIZE(rs_descs))
                     ENDIF
                  ENDIF
               ENDIF
            CASE DEFAULT
               CALL cp__b("task_list_methods.F",1354,"")
            END SELECT

         ENDDO

      ENDDO

      CALL timestop(handle)

   END SUBROUTINE compute_load_list
! **************************************************************************************************
!> \brief small helper function to return the proper index in the list array
!>
!> \param list ...
!> \param rank ...
!> \param dest ...
!> \retval list_index ...
!> \par History
!>      created 2008-10-06 [Joost VandeVondele]
! **************************************************************************************************
   INTEGER FUNCTION list_index(list, rank, dest)
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)  :: list
      INTEGER, INTENT(IN)                      :: rank, dest
      list_index = 1
      DO
         IF (list(1, list_index, dest) == rank) EXIT
         list_index = list_index+1
      ENDDO
   END FUNCTION list_index
! **************************************************************************************************
!> \brief create a list with possible destinations (i.e. the central cpu and neighbors) for each cpu
!>        note that we allocate it with an additional field to store the load of this destination
!>
!> \param list ...
!> \param rs_descs ...
!> \param grid_level ...
!> \par History
!>      created 2008-10-06 [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE create_destination_list(list, rs_descs, grid_level)
      INTEGER, DIMENSION(:, :, :), POINTER               :: list
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER, INTENT(IN)                                :: grid_level

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_destination_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, icpu, j, maxcount, ncpu, &
                                                            ultimate_max
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index, sublist

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(list)))CALL cp__a("task_list_methods.F",1408)
      ncpu = rs_descs(grid_level)%rs_desc%group_size
      ultimate_max = 7

      ALLOCATE (list(2, ultimate_max, 0:ncpu-1))

      ALLOCATE (INDEX(ultimate_max))
      ALLOCATE (sublist(ultimate_max))
      sublist = HUGE(sublist)

      maxcount = 1
      DO icpu = 0, ncpu-1
         sublist(1) = icpu
         sublist(2) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/-1, 0, 0/))
         sublist(3) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/+1, 0, 0/))
         sublist(4) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/0, -1, 0/))
         sublist(5) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/0, +1, 0/))
         sublist(6) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/0, 0, -1/))
         sublist(7) = rs_grid_locate_rank(rs_descs(grid_level)%rs_desc, icpu, (/0, 0, +1/))
         ! only retain unique values of the destination
         CALL sort(sublist, ultimate_max, index)
         j = 1
         DO i = 2, 7
            IF (sublist(i) .NE. sublist(j)) THEN
               j = j+1
               sublist(j) = sublist(i)
            ENDIF
         ENDDO
         maxcount = MAX(maxcount, j)
         sublist(j+1:ultimate_max) = HUGE(sublist)
         list(1, :, icpu) = sublist
         list(2, :, icpu) = 0
      ENDDO

      CALL reallocate(list, 1, 2, 1, maxcount, 0, ncpu-1)

      CALL timestop(handle)

   END SUBROUTINE create_destination_list

! **************************************************************************************************
!> \brief given a task list, compute the load of each process everywhere
!>        giving this function the ability to loop over a (sub)set of rs_grids,
!>        and do all the communication in one shot, would speed it up
!> \param loads ...
!> \param rs_descs ...
!> \param grid_level ...
!> \param ntasks ...
!> \param nimages ...
!> \param natom ...
!> \param maxset ...
!> \param maxpgf ...
!> \param tasks ...
!> \param use_reordered_ranks ...
!> \par History
!>      none
!> \author MattW 21/11/2007
! **************************************************************************************************
   SUBROUTINE get_current_loads(loads, rs_descs, grid_level, ntasks, nimages, natom, maxset, &
                                maxpgf, tasks, use_reordered_ranks)
      INTEGER(KIND=int_8), DIMENSION(:)                  :: loads
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: grid_level, ntasks, nimages, natom, &
                                                            maxset, maxpgf
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      LOGICAL, INTENT(IN)                                :: use_reordered_ranks

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_current_loads', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ilevel, img, ipgf, &
                                                            iset, jatom, jpgf, jset
      INTEGER(KIND=int_8)                                :: total_cost_local
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: recv_buf_i, send_buf_i
      TYPE(realspace_grid_desc_type), POINTER            :: desc

      CALL timeset(routineN, handle)

      desc => rs_descs(grid_level)%rs_desc

      ! allocate local arrays
      ALLOCATE (send_buf_i(desc%group_size))
      ALLOCATE (recv_buf_i(desc%group_size))

      ! communication step 1 : compute the total local cost of the tasks
      !                        each proc needs to know the amount of work he will receive

      ! send buffer now contains for each target the cost of the tasks it will receive
      send_buf_i = 0
      DO i = 1, ntasks
         CALL int2pair(tasks(3, i), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, nimages, natom, maxset, maxpgf)
         IF (ilevel .NE. grid_level) CYCLE
         IF (use_reordered_ranks) THEN
            send_buf_i(rs_descs(ilevel)%rs_desc%virtual2real(decode_rank(tasks(1, i), SIZE(rs_descs)))+1) = &
               send_buf_i(rs_descs(ilevel)%rs_desc%virtual2real(decode_rank(tasks(1, i), SIZE(rs_descs)))+1) &
               +tasks(5, i)
         ELSE
            send_buf_i(decode_rank(tasks(1, i), SIZE(rs_descs))+1) = &
               send_buf_i(decode_rank(tasks(1, i), SIZE(rs_descs))+1) &
               +tasks(5, i)
         END IF
      END DO
      CALL mp_alltoall(send_buf_i, recv_buf_i, 1, desc%group)

      ! communication step 2 : compute the global cost of the tasks
      total_cost_local = SUM(recv_buf_i)

      ! after this step, the recv buffer contains the local cost for each CPU
      CALL mp_allgather(total_cost_local, loads, desc%group)

      CALL timestop(handle)

   END SUBROUTINE get_current_loads
! **************************************************************************************************
!> \brief performs load balancing shifting tasks on the replicated grids
!>        this modifies the destination of some of the tasks on replicated
!>        grids, and in this way balances the load
!> \param rs_descs ...
!> \param ntasks ...
!> \param tasks ...
!> \param nimages ...
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \par History
!>      none
!> \author MattW 21/11/2007
! **************************************************************************************************
   SUBROUTINE load_balance_replicated(rs_descs, ntasks, tasks, nimages, natoms, maxset, maxpgf)

      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: ntasks
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, INTENT(IN)                                :: nimages, natoms, maxset, maxpgf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'load_balance_replicated', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ilevel, img, ipgf, &
                                                            iset, j, jatom, jpgf, jset, &
                                                            no_overloaded, no_underloaded, &
                                                            proc_receiving
      INTEGER(KIND=int_8)                                :: average_cost, cost_task_rep, count, &
                                                            offset, total_cost_global
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: load_imbalance, loads, recv_buf_i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index
      TYPE(realspace_grid_desc_type), POINTER            :: desc

      CALL timeset(routineN, handle)

      desc => rs_descs(1)%rs_desc

      ! allocate local arrays
      ALLOCATE (recv_buf_i(desc%group_size))
      ALLOCATE (loads(desc%group_size))

      recv_buf_i = 0
      DO i = 1, SIZE(rs_descs)
         CALL get_current_loads(loads, rs_descs, i, ntasks, nimages, natoms, maxset, maxpgf, &
                                tasks, use_reordered_ranks=.TRUE.)
         recv_buf_i(:) = recv_buf_i+loads
      END DO

      total_cost_global = SUM(recv_buf_i)
      average_cost = total_cost_global/desc%group_size

      !
      ! compute how to redistribute the replicated tasks so that the average cost is reached
      !

      ! load imbalance measures the load of a given CPU relative
      ! to the optimal load distribution (load=average)
      ALLOCATE (load_imbalance(desc%group_size))
      ALLOCATE (INDEX(desc%group_size))

      load_imbalance(:) = recv_buf_i-average_cost
      no_overloaded = 0
      no_underloaded = 0

      DO i = 1, desc%group_size
         IF (load_imbalance(i) .GT. 0) no_overloaded = no_overloaded+1
         IF (load_imbalance(i) .LT. 0) no_underloaded = no_underloaded+1
      ENDDO

      ! sort the recv_buffer on number of tasks, gives us index which provides a
      ! mapping between processor ranks and how overloaded the processor
      CALL sort(recv_buf_i, SIZE(recv_buf_i), index)

      ! find out the number of replicated tasks each proc has
      ! but only those tasks which have not yet been assigned
      cost_task_rep = 0
      DO i = 1, ntasks
         IF (tasks(4, i) .EQ. 0 &
             .AND. decode_rank(tasks(1, i), SIZE(rs_descs)) == decode_rank(tasks(2, i), SIZE(rs_descs))) THEN
            cost_task_rep = cost_task_rep+tasks(5, i)
         ENDIF
      ENDDO

      ! now, correct the load imbalance for the overloaded CPUs
      ! they will send away not more than the total load of replicated tasks
      CALL mp_allgather(cost_task_rep, recv_buf_i, desc%group)

      DO i = 1, desc%group_size
         ! At the moment we can only offload replicated tasks
         IF (load_imbalance(i) .GT. 0) &
            load_imbalance(i) = MIN(load_imbalance(i), recv_buf_i(i))
      ENDDO

      ! simplest algorithm I can think of of is that the processor with the most
      ! excess tasks fills up the process needing most, then moves on to next most.
      ! At the moment if we've got less replicated tasks than we're overloaded then
      ! task balancing will be incomplete

      ! only need to do anything if I've excess tasks
      IF (load_imbalance(desc%my_pos+1) .GT. 0) THEN

         count = 0 ! weighted amount of tasks offloaded
         offset = 0 ! no of underloaded processes already filled by other more overloaded procs

         ! calculate offset
         DO i = desc%group_size, desc%group_size-no_overloaded+1, -1
            IF (INDEX(i) .EQ. desc%my_pos+1) THEN
               EXIT
            ELSE
               offset = offset+load_imbalance(INDEX(i))
            ENDIF
         ENDDO

         ! find my starting processor to send to
         proc_receiving = HUGE(proc_receiving)
         DO i = 1, no_underloaded
            offset = offset+load_imbalance(INDEX(i))
            IF (offset .LE. 0) THEN
               proc_receiving = i
               EXIT
            ENDIF
         ENDDO

         ! offset now contains minus the number of tasks proc_receiving requires
         ! we fill this up by adjusting the destination of tasks on the replicated grid,
         ! then move to next most underloaded proc
         DO j = 1, ntasks
            IF (tasks(4, j) .EQ. 0 &
                .AND. decode_rank(tasks(1, j), SIZE(rs_descs)) == decode_rank(tasks(2, j), SIZE(rs_descs))) THEN
               ! just avoid sending to non existing procs due to integer truncation
               ! in the computation of the average
               IF (proc_receiving .GT. no_underloaded) EXIT
               ! set new destination
               CALL int2pair(tasks(3, j), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, nimages, natoms, maxset, maxpgf)
               tasks(1, j) = encode_rank(INDEX(proc_receiving)-1, ilevel, SIZE(rs_descs))
               offset = offset+tasks(5, j)
               count = count+tasks(5, j)
               IF (count .GE. load_imbalance(desc%my_pos+1)) EXIT
               IF (offset .GT. 0) THEN
                  proc_receiving = proc_receiving+1
                  ! just avoid sending to non existing procs due to integer truncation
                  ! in the computation of the average
                  IF (proc_receiving .GT. no_underloaded) EXIT
                  offset = load_imbalance(INDEX(proc_receiving))
               ENDIF
            ENDIF
         ENDDO
      ENDIF

      DEALLOCATE (index)
      DEALLOCATE (load_imbalance)

      CALL timestop(handle)

   END SUBROUTINE load_balance_replicated

! **************************************************************************************************
!> \brief given an input task list, redistribute so that all tasks can be processed locally,
!>        i.e. dest equals rank
!> \param rs_descs ...
!> \param ntasks ...
!> \param tasks ...
!> \param rval ...
!> \param ntasks_recv ...
!> \param tasks_recv ...
!> \param rval_recv ...
!> \par History
!>      none
!> \author MattW 21/11/2007
! **************************************************************************************************
   SUBROUTINE create_local_tasks(rs_descs, ntasks, tasks, rval, &
                                 ntasks_recv, tasks_recv, rval_recv)

      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: ntasks
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rval
      INTEGER                                            :: ntasks_recv
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_recv
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rval_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_local_tasks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, l, rank, task_dim
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: recv_buf_i, send_buf_i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: recv_disps, recv_sizes, send_disps, &
                                                            send_sizes
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf_r, send_buf_r
      TYPE(realspace_grid_desc_type), POINTER            :: desc

      CALL timeset(routineN, handle)

      desc => rs_descs(1)%rs_desc

      ! allocate local arrays
      ALLOCATE (send_sizes(desc%group_size))
      ALLOCATE (recv_sizes(desc%group_size))
      ALLOCATE (send_disps(desc%group_size))
      ALLOCATE (recv_disps(desc%group_size))
      ALLOCATE (send_buf_i(desc%group_size))
      ALLOCATE (recv_buf_i(desc%group_size))

      ! fill send buffer, now counting how many tasks will be send (stored in an int8 array for convenience only).
      send_buf_i = 0
      DO i = 1, ntasks
         rank = rs_descs(decode_level(tasks(1, i), SIZE(rs_descs)))%rs_desc%virtual2real(decode_rank(tasks(1, i), SIZE(rs_descs)))
         send_buf_i(rank+1) = send_buf_i(rank+1)+1
      END DO

      CALL mp_alltoall(send_buf_i, recv_buf_i, 1, desc%group)

      ! pack the tasks, and send them around

      task_dim = SIZE(tasks, 1)

      send_sizes = 0
      send_disps = 0
      recv_sizes = 0
      recv_disps = 0

      send_sizes(1) = INT(send_buf_i(1)*task_dim)
      recv_sizes(1) = INT(recv_buf_i(1)*task_dim)
      DO i = 2, desc%group_size
         send_sizes(i) = INT(send_buf_i(i)*task_dim)
         recv_sizes(i) = INT(recv_buf_i(i)*task_dim)
         send_disps(i) = send_disps(i-1)+send_sizes(i-1)
         recv_disps(i) = recv_disps(i-1)+recv_sizes(i-1)
      ENDDO

      ! deallocate old send/recv buffers
      DEALLOCATE (send_buf_i)
      DEALLOCATE (recv_buf_i)

      ! allocate them with new sizes
      ALLOCATE (send_buf_i(SUM(send_sizes)))
      ALLOCATE (recv_buf_i(SUM(recv_sizes)))

      ! do packing
      send_buf_i = 0
      send_sizes = 0
      DO j = 1, ntasks
         i = rs_descs(decode_level(tasks(1, j), SIZE(rs_descs)))%rs_desc%virtual2real(decode_rank(tasks(1, j), SIZE(rs_descs)))+1
         DO k = 1, task_dim
            send_buf_i(send_disps(i)+send_sizes(i)+k) = tasks(k, j)
         ENDDO
         send_sizes(i) = send_sizes(i)+task_dim
      ENDDO

      ! do communication
      CALL mp_alltoall(send_buf_i, send_sizes, send_disps, recv_buf_i, recv_sizes, recv_disps, desc%group)

      DEALLOCATE (send_buf_i)

      ntasks_recv = SUM(recv_sizes)/task_dim
      ALLOCATE (tasks_recv(task_dim, ntasks_recv))

      ! do unpacking
      l = 0
      DO i = 1, desc%group_size
         DO j = 0, recv_sizes(i)/task_dim-1
            l = l+1
            DO k = 1, task_dim
               tasks_recv(k, l) = recv_buf_i(recv_disps(i)+j*task_dim+k)
            ENDDO
         ENDDO
      ENDDO

      DEALLOCATE (recv_buf_i)

      ! send rvals (to be removed :-)

      ! take care of the new task_dim in the send_sizes
      send_sizes(:) = (send_sizes(:)/task_dim)*SIZE(rval, 1)
      recv_sizes(:) = (recv_sizes(:)/task_dim)*SIZE(rval, 1)
      send_disps(:) = (send_disps(:)/task_dim)*SIZE(rval, 1)
      recv_disps(:) = (recv_disps(:)/task_dim)*SIZE(rval, 1)
      task_dim = SIZE(rval, 1)

      ALLOCATE (send_buf_r(SUM(send_sizes)))
      ALLOCATE (recv_buf_r(SUM(recv_sizes)))

      !do packing
      send_sizes = 0
      DO j = 1, ntasks
         i = rs_descs(decode_level(tasks(1, j), SIZE(rs_descs)))%rs_desc%virtual2real(decode_rank(tasks(1, j), SIZE(rs_descs)))+1
         DO k = 1, task_dim
            send_buf_r(send_disps(i)+send_sizes(i)+k) = rval(k, j)
         ENDDO
         send_sizes(i) = send_sizes(i)+task_dim
      ENDDO

      ! do communication
      CALL mp_alltoall(send_buf_r, send_sizes, send_disps, &
                       recv_buf_r, recv_sizes, recv_disps, desc%group)

      DEALLOCATE (send_buf_r)
      ALLOCATE (rval_recv(task_dim, SUM(recv_sizes)/task_dim))

      ! do unpacking
      l = 0
      DO i = 1, desc%group_size
         DO j = 0, recv_sizes(i)/task_dim-1
            l = l+1
            DO k = 1, task_dim
               rval_recv(k, l) = recv_buf_r(recv_disps(i)+j*task_dim+k)
            ENDDO
         ENDDO
      ENDDO

      DEALLOCATE (recv_buf_r)
      DEALLOCATE (send_sizes)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (recv_disps)

      CALL timestop(handle)

   END SUBROUTINE create_local_tasks

! **************************************************************************************************
!> \brief Assembles tasks to be performed on local grid
!> \param rs_descs the grids
!> \param ntasks Number of tasks for local processing
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \param nimages ...
!> \param tasks the task set generated on this processor
!> \param rval ...
!> \param atom_pair_send ...
!> \param atom_pair_recv ...
!> \param symmetric ...
!> \param reorder_rs_grid_ranks ...
!> \param skip_load_balance_distributed ...
!> \par History
!>      none
!> \author MattW 21/11/2007
! **************************************************************************************************
   SUBROUTINE distribute_tasks(rs_descs, ntasks, natoms, maxset, maxpgf, nimages, &
                               tasks, rval, atom_pair_send, atom_pair_recv, &
                               symmetric, reorder_rs_grid_ranks, skip_load_balance_distributed)

      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      INTEGER                                            :: ntasks, natoms, maxset, maxpgf, nimages
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rval
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: atom_pair_send, atom_pair_recv
      LOGICAL, INTENT(IN)                                :: symmetric, reorder_rs_grid_ranks, &
                                                            skip_load_balance_distributed

      CHARACTER(LEN=*), PARAMETER :: routineN = 'distribute_tasks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, igrid_level, irank, k, &
                                                            ntasks_recv
      INTEGER(KIND=int_8)                                :: load_gap, max_load, replicated_load
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: taskid, total_loads, total_loads_tmp, &
                                                            trial_loads
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: loads, tasks_recv
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index, real2virtual, total_index
      LOGICAL                                            :: distributed_grids, fixed_first_grid
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rval_recv
      TYPE(realspace_grid_desc_type), POINTER            :: desc

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tasks)))CALL cp__a("task_list_methods.F",1894)

      ! *** figure out if we have distributed grids
      distributed_grids = .FALSE.
      DO igrid_level = 1, SIZE(rs_descs)
         IF (rs_descs(igrid_level)%rs_desc%distributed) THEN
            distributed_grids = .TRUE.
         ENDIF
      END DO
      desc => rs_descs(1)%rs_desc

      IF (distributed_grids) THEN

         ALLOCATE (loads(0:desc%group_size-1, SIZE(rs_descs)))
         ALLOCATE (total_loads(0:desc%group_size-1))

         total_loads = 0

         ! First round of balancing on the distributed grids
         ! we just balance each of the distributed grids independently
         DO igrid_level = 1, SIZE(rs_descs)
            IF (rs_descs(igrid_level)%rs_desc%distributed) THEN

               IF (.NOT. skip_load_balance_distributed) &
                  CALL load_balance_distributed(tasks, ntasks, rs_descs, igrid_level, nimages, natoms, maxset, maxpgf)

               CALL get_current_loads(loads(:, igrid_level), rs_descs, igrid_level, ntasks, nimages, natoms, maxset, maxpgf, &
                                      tasks, use_reordered_ranks=.FALSE.)

               total_loads(:) = total_loads+loads(:, igrid_level)

            END IF
         END DO

         ! calculate the total load of replicated tasks, so we can decide if it is worth
         ! reordering the distributed grid levels

         replicated_load = 0
         DO igrid_level = 1, SIZE(rs_descs)
            IF (.NOT. rs_descs(igrid_level)%rs_desc%distributed) THEN
               CALL get_current_loads(loads(:, igrid_level), rs_descs, igrid_level, ntasks, nimages, natoms, maxset, maxpgf, &
                                      tasks, use_reordered_ranks=.FALSE.)
               replicated_load = replicated_load+SUM(loads(:, igrid_level))
            END IF
         END DO

         !IF (desc%my_pos==0) THEN
         ! WRITE(*,*) "Total replicated load is ",replicated_load
         !END IF

         ! Now we adjust the rank ordering based on the current loads
         ! we leave the first distributed level and all the replicated levels in the default order
         IF (reorder_rs_grid_ranks) THEN
            fixed_first_grid = .FALSE.
            DO igrid_level = 1, SIZE(rs_descs)
               IF (rs_descs(igrid_level)%rs_desc%distributed) THEN
                  IF (fixed_first_grid .EQV. .FALSE.) THEN
                     total_loads(:) = loads(:, igrid_level)
                     fixed_first_grid = .TRUE.
                  ELSE
                     ALLOCATE (trial_loads(0:desc%group_size-1))

                     trial_loads(:) = total_loads+loads(:, igrid_level)
                     max_load = MAXVAL(trial_loads)
                     load_gap = 0
                     DO irank = 0, desc%group_size-1
                        load_gap = load_gap+max_load-trial_loads(irank)
                     END DO

                     ! If there is not enough replicated load to load balance well enough
                     ! then we will reorder this grid level
                     IF (load_gap > replicated_load*1.05_dp) THEN

                        ALLOCATE (INDEX(0:desc%group_size-1))
                        ALLOCATE (total_index(0:desc%group_size-1))
                        ALLOCATE (total_loads_tmp(0:desc%group_size-1))
                        ALLOCATE (real2virtual(0:desc%group_size-1))

                        total_loads_tmp(:) = total_loads
                        CALL sort(total_loads_tmp, desc%group_size, total_index)
                        CALL sort(loads(:, igrid_level), desc%group_size, index)

                        ! Reorder so that the rank with smallest load on this grid level is paired with
                        ! the highest load in total
                        DO irank = 0, desc%group_size-1
                           total_loads(total_index(irank)-1) = total_loads(total_index(irank)-1)+ &
                                                               loads(desc%group_size-irank-1, igrid_level)
                           real2virtual(total_index(irank)-1) = INDEX(desc%group_size-irank-1)-1
                        END DO

                        CALL rs_grid_reorder_ranks(rs_descs(igrid_level)%rs_desc, real2virtual)

                        DEALLOCATE (index)
                        DEALLOCATE (total_index)
                        DEALLOCATE (total_loads_tmp)
                        DEALLOCATE (real2virtual)
                     ELSE
                        total_loads(:) = trial_loads
                     ENDIF

                     DEALLOCATE (trial_loads)

                  ENDIF
               ENDIF
            END DO
         END IF

         ! Now we use the replicated tasks to balance out the rest of the load
         CALL load_balance_replicated(rs_descs, ntasks, tasks, nimages, natoms, maxset, maxpgf)

         !total_loads = 0
         !DO igrid_level=1,SIZE(rs_descs)
         !  CALL get_current_loads(loads(:,igrid_level), rs_descs, igrid_level, ntasks, nimages, natoms, maxset,maxpgf, &
         !                         tasks, use_reordered_ranks=.TRUE.)
         !  total_loads = total_loads + loads(:, igrid_level)
         !END DO

         !IF (desc%my_pos==0) THEN
         !  WRITE(*,*) ""
         !  WRITE(*,*) "At the end of the load balancing procedure"
         !  WRITE(*,*) "Maximum  load:",MAXVAL(total_loads)
         !  WRITE(*,*) "Average  load:",SUM(total_loads)/SIZE(total_loads)
         !  WRITE(*,*) "Minimum  load:",MINVAL(total_loads)
         !ENDIF

         ! given a list of tasks, this will do the needed reshuffle so that all tasks will be local
         CALL create_local_tasks(rs_descs, ntasks, tasks, rval, ntasks_recv, tasks_recv, rval_recv)

         !
         ! tasks list are complete, we can compute the list of atomic blocks (atom pairs)
         ! we will be sending. These lists are needed for redistribute_matrix.
         !
         ALLOCATE (atom_pair_send(ntasks))
         CALL get_atom_pair(atom_pair_send, tasks, send=.TRUE., &
                            symmetric=symmetric, natoms=natoms, maxset=maxset, maxpgf=maxpgf, &
                            nimages=nimages, rs_descs=rs_descs)

         ! natom_send=SIZE(atom_pair_send)
         ! CALL mp_sum(natom_send,desc%group)
         ! IF (desc%my_pos==0) THEN
         !     WRITE(*,*) ""
         !     WRITE(*,*) "Total number of atomic blocks to be send:",natom_send
         ! ENDIF

         ALLOCATE (atom_pair_recv(ntasks_recv))
         CALL get_atom_pair(atom_pair_recv, tasks_recv, send=.FALSE., &
                            symmetric=symmetric, natoms=natoms, maxset=maxset, maxpgf=maxpgf, &
                            nimages=nimages, rs_descs=rs_descs)

         ! cleanup, at this point we  don't need the original tasks & rvals anymore
         DEALLOCATE (tasks)
         DEALLOCATE (rval)
         DEALLOCATE (loads)
         DEALLOCATE (total_loads)

      ELSE

         tasks_recv => tasks
         ntasks_recv = ntasks
         rval_recv => rval

      ENDIF

      !
      ! here we sort the task list we will process locally.
      ! the sort is determined by pair2int
      !
      rval => rval_recv

      ALLOCATE (taskid(ntasks_recv))
      ALLOCATE (INDEX(ntasks_recv))

      taskid(:) = tasks_recv(3, 1:ntasks_recv)
      CALL sort(taskid, SIZE(taskid), index)

      DO k = 1, SIZE(tasks_recv, 1)
         tasks_recv(k, 1:ntasks_recv) = tasks_recv(k, index)
      ENDDO

      DO k = 1, SIZE(rval, 1)
         rval(k, 1:ntasks_recv) = rval(k, index)
      ENDDO

      DEALLOCATE (taskid)
      DEALLOCATE (index)

      !
      ! final lists are ready
      !

      tasks => tasks_recv
      ntasks = ntasks_recv

      CALL timestop(handle)

   END SUBROUTINE distribute_tasks

! **************************************************************************************************
!> \brief ...
!> \param atom_pair ...
!> \param my_tasks ...
!> \param send ...
!> \param symmetric ...
!> \param natoms ...
!> \param maxset ...
!> \param maxpgf ...
!> \param nimages ...
!> \param rs_descs ...
! **************************************************************************************************
   SUBROUTINE get_atom_pair(atom_pair, my_tasks, send, symmetric, natoms, maxset, &
                            maxpgf, nimages, rs_descs)

      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: atom_pair
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: my_tasks
      LOGICAL                                            :: send, symmetric
      INTEGER                                            :: natoms, maxset, maxpgf, nimages
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_atom_pair', routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, i, iatom, ilevel, img, ipgf, &
                                                            iset, j, jatom, jpgf, jset
      INTEGER(KIND=int_8)                                :: natom8, nim8
      INTEGER, DIMENSION(:), POINTER                     :: index

! calculate list of atom pairs
! fill pair list taking into acount symmetry

      natom8 = natoms
      nim8 = nimages

      atom_pair = 0
      DO i = 1, SIZE(atom_pair)
         CALL int2pair(my_tasks(3, i), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, nimages, natoms, maxset, maxpgf)
         IF (symmetric) THEN
            IF (iatom .LE. jatom) THEN
               arow = iatom
               acol = jatom
            ELSE
               arow = jatom
               acol = iatom
            ENDIF
         ELSE
            arow = iatom
            acol = jatom
         ENDIF
         IF (send) THEN
            ! If sending, we need to use the 'real rank' as the pair has to be sent to the process which
            ! actually has the correct part of the rs_grid to do the mapping
            atom_pair(i) = rs_descs(ilevel)%rs_desc%virtual2real(decode_rank(my_tasks(1, i), SIZE(rs_descs))) &
                           *natom8*natom8*nim8+(arow-1)*natom8*nim8+(acol-1)*nim8+(img-1)
         ELSE
            ! If we are recieving, then no conversion is needed as the rank is that of the process with the
            ! required matrix block, and the ordering of the rs grid is irrelevant
            atom_pair(i) = decode_rank(my_tasks(2, i), SIZE(rs_descs)) &
                           *natom8*natom8*nim8+(arow-1)*natom8*nim8+(acol-1)*nim8+(img-1)
         ENDIF
      ENDDO

      ALLOCATE (INDEX(SIZE(atom_pair)))

      ! find unique atom pairs that I'm sending/receiving
      CALL sort(atom_pair, SIZE(atom_pair), index)

      DEALLOCATE (index)

      IF (SIZE(atom_pair) > 1) THEN
         j = 1
         ! first atom pair must be allowed
         DO i = 2, SIZE(atom_pair)
            IF (atom_pair(i) .GT. atom_pair(i-1)) THEN
               j = j+1
               atom_pair(j) = atom_pair(i)
            ENDIF
         ENDDO
         ! reallocate the atom pair list
         CALL reallocate(atom_pair, 1, j)
      ENDIF

   END SUBROUTINE get_atom_pair

! **************************************************************************************************
!> \brief redistributes the matrix so that it can be used in realspace operations
!>        i.e. according to the task lists for collocate and integrate.
!>        This routine can become a bottleneck in large calculations.
!> \param rs_descs ...
!> \param pmats ...
!> \param atom_pair_send ...
!> \param atom_pair_recv ...
!> \param natoms ...
!> \param nimages ...
!> \param scatter ...
!> \param hmats ...
! **************************************************************************************************
   SUBROUTINE rs_distribute_matrix(rs_descs, pmats, atom_pair_send, atom_pair_recv, &
                                   natoms, nimages, scatter, hmats)

      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmats
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: atom_pair_send, atom_pair_recv
      INTEGER                                            :: natoms, nimages
      LOGICAL                                            :: scatter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: hmats

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_distribute_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, handle, i, img, j, k, l, me, &
                                                            nblkcols_total, nblkrows_total, ncol, &
                                                            nrow, nthread, nthread_left
      INTEGER(KIND=int_8)                                :: natom8, nim8, pair
      INTEGER, ALLOCATABLE, DIMENSION(:) :: first_col, first_row, last_col, last_row, recv_disps, &
         recv_pair_count, recv_pair_disps, recv_sizes, send_disps, send_pair_count, &
         send_pair_disps, send_sizes
      LOGICAL                                            :: found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf_r, send_buf_r
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: h_block, p_block
      TYPE(cp_dbcsr_type), POINTER                       :: hmat, pmat
      TYPE(realspace_grid_desc_type), POINTER            :: desc

!$    INTEGER(kind=omp_lock_kind), ALLOCATABLE, DIMENSION(:) :: locks

      CALL timeset(routineN, handle)

      IF (.NOT. scatter) THEN
         IF(.NOT.(PRESENT(hmats)))CALL cp__a("task_list_methods.F",2222)
      END IF

      desc => rs_descs(1)%rs_desc
      me = desc%my_pos+1

      ! allocate local arrays
      ALLOCATE (send_sizes(desc%group_size))
      ALLOCATE (recv_sizes(desc%group_size))
      ALLOCATE (send_disps(desc%group_size))
      ALLOCATE (recv_disps(desc%group_size))
      ALLOCATE (send_pair_count(desc%group_size))
      ALLOCATE (recv_pair_count(desc%group_size))
      ALLOCATE (send_pair_disps(desc%group_size))
      ALLOCATE (recv_pair_disps(desc%group_size))

      pmat => pmats(1)%matrix
      CALL cp_dbcsr_get_info(pmat, &
                             nblkrows_total=nblkrows_total, &
                             nblkcols_total=nblkcols_total)
      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total), &
                first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(cp_dbcsr_row_block_sizes(pmat), first_row, last_row)
      CALL convert_sizes_to_offsets(cp_dbcsr_col_block_sizes(pmat), first_col, last_col)
      ! set up send buffer sizes
      natom8 = natoms
      nim8 = nimages

      send_sizes = 0
      send_pair_count = 0
      DO i = 1, SIZE(atom_pair_send)

         ! proc we're sending this block to
         k = INT(atom_pair_send(i)/(nim8*natom8**2))+1

         pair = MOD(atom_pair_send(i), nim8*natom8**2)

         arow = INT(pair/(nim8*natom8))+1
         acol = INT(MOD(pair, nim8*natom8)/nim8)+1

         nrow = last_row(arow)-first_row(arow)+1
         ncol = last_col(acol)-first_col(acol)+1

         send_sizes(k) = send_sizes(k)+nrow*ncol
         send_pair_count(k) = send_pair_count(k)+1

      ENDDO

      send_disps = 0
      send_pair_disps = 0
      DO i = 2, desc%group_size
         send_disps(i) = send_disps(i-1)+send_sizes(i-1)
         send_pair_disps(i) = send_pair_disps(i-1)+send_pair_count(i-1)
      ENDDO

      ALLOCATE (send_buf_r(SUM(send_sizes)))

      ! set up recv buffer

      recv_sizes = 0
      recv_pair_count = 0
      DO i = 1, SIZE(atom_pair_recv)

         ! proc we're receiving this data from
         k = INT(atom_pair_recv(i)/(nim8*natom8**2))+1

         pair = MOD(atom_pair_recv(i), nim8*natom8**2)

         arow = INT(pair/(nim8*natom8))+1
         acol = INT(MOD(pair, nim8*natom8)/nim8)+1

         nrow = last_row(arow)-first_row(arow)+1
         ncol = last_col(acol)-first_col(acol)+1

         recv_sizes(k) = recv_sizes(k)+nrow*ncol
         recv_pair_count(k) = recv_pair_count(k)+1

      ENDDO

      recv_disps = 0
      recv_pair_disps = 0
      DO i = 2, desc%group_size
         recv_disps(i) = recv_disps(i-1)+recv_sizes(i-1)
         recv_pair_disps(i) = recv_pair_disps(i-1)+recv_pair_count(i-1)
      ENDDO
      ALLOCATE (recv_buf_r(SUM(recv_sizes)))

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          SHARED(desc,send_pair_count,send_pair_disps,natom8,nim8,nimages),&
!$OMP          SHARED(last_row,first_row,last_col,first_col),&
!$OMP          SHARED(pmats,send_buf_r,send_disps,send_sizes),&
!$OMP          SHARED(atom_pair_send,me,hmats,nblkrows_total),&
!$OMP          SHARED(atom_pair_recv,recv_buf_r,scatter,recv_pair_disps), &
!$OMP          SHARED(recv_sizes,recv_disps,recv_pair_count,locks), &
!$OMP          PRIVATE(i,img,pair,arow,acol,nrow,ncol,p_block,found,j,k,l),&
!$OMP          PRIVATE(nthread,h_block,nthread_left,hmat,pmat)

      nthread = 1
!$    nthread = omp_get_num_threads()
      nthread_left = 1
!$    nthread_left = MAX(1, nthread-1)

      ! do packing
!$OMP DO schedule(guided)
      DO l = 1, desc%group_size
         IF (l .EQ. me) CYCLE
         send_sizes(l) = 0
         DO i = 1, send_pair_count(l)
            pair = MOD(atom_pair_send(send_pair_disps(l)+i), nim8*natom8**2)

            arow = INT(pair/(nim8*natom8))+1
            acol = INT(MOD(pair, nim8*natom8)/nim8)+1
            img = INT(MOD(pair, nim8))+1

            nrow = last_row(arow)-first_row(arow)+1
            ncol = last_col(acol)-first_col(acol)+1

            pmat => pmats(img)%matrix
            CALL cp_dbcsr_get_block_p(matrix=pmat, row=arow, col=acol, &
                                      block=p_block, found=found)
            IF(.NOT.(found))CALL cp__a("task_list_methods.F",2342)

            DO k = 1, ncol
               DO j = 1, nrow
                  send_buf_r(send_disps(l)+send_sizes(l)+j+(k-1)*nrow) = p_block(j, k)
               ENDDO
            ENDDO
            send_sizes(l) = send_sizes(l)+nrow*ncol
         ENDDO
      ENDDO
!$OMP END DO

      IF (.NOT. scatter) THEN
         ! We need locks to protect concurrent summation into H
!$OMP SINGLE
!$       ALLOCATE (locks(nthread*10))
!$OMP END SINGLE

!$OMP DO
!$       do i = 1, nthread*10
!$          call omp_init_lock(locks(i))
!$       end do
!$OMP END DO
      END IF

!$OMP MASTER
      ! do communication
      CALL mp_alltoall(send_buf_r, send_sizes, send_disps, &
                       recv_buf_r, recv_sizes, recv_disps, desc%group)
!$OMP END MASTER

      ! If this is a scatter, then no need to copy local blocks,
      ! If not, we sum them directly into H (bypassing the alltoall)
      IF (.NOT. scatter) THEN

         ! Distribute work over remaining threads assuming one is still in the alltoall
!$OMP DO schedule(dynamic,MAX(1,send_pair_count(me)/nthread_left))
         DO i = 1, send_pair_count(me)
            pair = MOD(atom_pair_send(send_pair_disps(me)+i), natom8**2)

            arow = INT(pair/(nim8*natom8))+1
            acol = INT(MOD(pair, nim8*natom8)/nim8)+1
            img = INT(MOD(pair, nim8))+1

            nrow = last_row(arow)-first_row(arow)+1
            ncol = last_col(acol)-first_col(acol)+1

            hmat => hmats(img)%matrix
            pmat => pmats(img)%matrix
            CALL cp_dbcsr_get_block_p(matrix=hmat, row=arow, col=acol, &
                                      BLOCK=h_block, found=found)
            IF(.NOT.(found))CALL cp__a("task_list_methods.F",2393)
            CALL cp_dbcsr_get_block_p(matrix=pmat, row=arow, col=acol, &
                                      BLOCK=p_block, found=found)
            IF(.NOT.(found))CALL cp__a("task_list_methods.F",2396)

!$          call omp_set_lock(locks((arow-1)*nthread*10/nblkrows_total+1))
            DO k = 1, ncol
               DO j = 1, nrow
                  h_block(j, k) = h_block(j, k)+p_block(j, k)
               ENDDO
            ENDDO
!$          call omp_unset_lock(locks((arow-1)*nthread*10/nblkrows_total+1))
         ENDDO
!$OMP END DO
      ELSE
         ! We will insert new blocks into P, so create mutable work matrices
         DO img = 1, nimages
            pmat => pmats(img)%matrix
            CALL cp_dbcsr_work_create(pmat, work_mutable=.TRUE., &
                                      nblks_guess=SIZE(atom_pair_recv)/nthread, sizedata_guess=SIZE(recv_buf_r)/nthread, &
                                      n=nthread)
         END DO
      END IF

! wait for comm and setup to finish
!$OMP BARRIER

      !do unpacking
!$OMP DO schedule(guided)
      DO l = 1, desc%group_size
         IF (l .EQ. me) CYCLE
         recv_sizes(l) = 0
         DO i = 1, recv_pair_count(l)
            pair = MOD(atom_pair_recv(recv_pair_disps(l)+i), natom8**2)

            arow = INT(pair/(nim8*natom8))+1
            acol = INT(MOD(pair, nim8*natom8)/nim8)+1
            img = INT(MOD(pair, nim8))+1

            nrow = last_row(arow)-first_row(arow)+1
            ncol = last_col(acol)-first_col(acol)+1

            pmat => pmats(img)%matrix
            NULLIFY (p_block)
            CALL cp_dbcsr_get_block_p(matrix=pmat, row=arow, col=acol, &
                                      BLOCK=p_block, found=found)

            IF (PRESENT(hmats)) THEN
               hmat => hmats(img)%matrix
               CALL cp_dbcsr_get_block_p(matrix=hmat, row=arow, col=acol, &
                                         BLOCK=h_block, found=found)
               IF(.NOT.(found))CALL cp__a("task_list_methods.F",2444)
            ENDIF

            IF (scatter .AND. .NOT. ASSOCIATED(p_block)) THEN
               CALL cp_dbcsr_put_block(pmat, arow, acol, &
                                       block=recv_buf_r(recv_disps(l)+recv_sizes(l)+1:recv_disps(l)+recv_sizes(l)+nrow*ncol))
            END IF
            IF (.NOT. scatter) THEN
!$             call omp_set_lock(locks((arow-1)*nthread*10/nblkrows_total+1))
               DO k = 1, ncol
                  DO j = 1, nrow
                     h_block(j, k) = h_block(j, k)+recv_buf_r(recv_disps(l)+recv_sizes(l)+j+(k-1)*nrow)
                  ENDDO
               ENDDO
!$             call omp_unset_lock(locks((arow-1)*nthread*10/nblkrows_total+1))
            ENDIF
            recv_sizes(l) = recv_sizes(l)+nrow*ncol
         ENDDO
      ENDDO
!$OMP END DO

!$    IF (.not. scatter) THEN
!$OMP DO
!$       do i = 1, nthread*10
!$          call omp_destroy_lock(locks(i))
!$       end do
!$OMP END DO
!$    END IF

!$OMP SINGLE
!$    IF (.not. scatter) THEN
!$       DEALLOCATE (locks)
!$    END IF
!$OMP END SINGLE NOWAIT

      IF (scatter) THEN
         ! Blocks were added to P
         DO img = 1, nimages
            pmat => pmats(img)%matrix
            CALL cp_dbcsr_finalize(pmat)
         END DO
      END IF
!$OMP END PARALLEL

      DEALLOCATE (send_buf_r)
      DEALLOCATE (recv_buf_r)

      DEALLOCATE (send_sizes)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (recv_disps)
      DEALLOCATE (send_pair_count)
      DEALLOCATE (recv_pair_count)
      DEALLOCATE (send_pair_disps)
      DEALLOCATE (recv_pair_disps)

      DEALLOCATE (first_row, last_row, first_col, last_col)

      CALL timestop(handle)

   END SUBROUTINE rs_distribute_matrix

! **************************************************************************************************
!> \brief determines the rank of the processor who's real rs grid contains point
!> \param rs_desc ...
!> \param igrid_level ...
!> \param n_levels ...
!> \param cube_center ...
!> \param ntasks ...
!> \param tasks ...
!> \param lb_cube ...
!> \param ub_cube ...
!> \param added_tasks ...
!> \par History
!>      11.2007 created [MattW]
!>      10.2008 rewritten [Joost VandeVondele]
!> \author MattW
! **************************************************************************************************
   SUBROUTINE rs_find_node(rs_desc, igrid_level, n_levels, cube_center, ntasks, tasks, &
                           lb_cube, ub_cube, added_tasks)

      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      INTEGER, INTENT(IN)                                :: igrid_level, n_levels
      INTEGER, DIMENSION(3), INTENT(IN)                  :: cube_center
      INTEGER, INTENT(INOUT)                             :: ntasks
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, DIMENSION(3), INTENT(IN)                  :: lb_cube, ub_cube
      INTEGER, INTENT(OUT)                               :: added_tasks

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_find_node', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: add_tasks = 1000
      REAL(kind=dp), PARAMETER                           :: mult_tasks = 2.0_dp

      INTEGER :: bit_index, coord(3), curr_tasks, dest, i, icoord(3), idest, itask, ix, iy, iz, &
         lb_coord(3), lb_domain(3), lbc(3), ub_coord(3), ub_domain(3), ubc(3)
      INTEGER(KIND=int_8)                                :: bit_pattern
      LOGICAL                                            :: dir_periodic(3)

      coord(1) = rs_desc%x2coord(cube_center(1))
      coord(2) = rs_desc%y2coord(cube_center(2))
      coord(3) = rs_desc%z2coord(cube_center(3))
      dest = rs_desc%coord2rank(coord(1), coord(2), coord(3))

      ! the real cube coordinates
      lbc = lb_cube+cube_center
      ubc = ub_cube+cube_center

      IF (ALL((rs_desc%lb_global(:, dest)-rs_desc%border) .LE. lbc) .AND. &
          ALL((rs_desc%ub_global(:, dest)+rs_desc%border) .GE. ubc)) THEN
         !standard distributed collocation/integration
         tasks(1, ntasks) = encode_rank(dest, igrid_level, n_levels)
         tasks(4, ntasks) = 1
         tasks(6, ntasks) = 0
         added_tasks = 1

         ! here we figure out if there is an alternate location for this task
         ! i.e. even though the cube_center is not in the real local domain,
         ! it might fully fit in the halo of the neighbor
         ! if its radius is smaller than the maximum radius
         ! the list of possible neighbors is stored here as a bit pattern
         ! to reconstruct what the rank of the neigbor is,
         ! we can use (note this requires the correct rs_grid)
         !  IF (BTEST(bit_pattern,0)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/-1,0,0/))
         !  IF (BTEST(bit_pattern,1)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/+1,0,0/))
         !  IF (BTEST(bit_pattern,2)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/0,-1,0/))
         !  IF (BTEST(bit_pattern,3)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/0,+1,0/))
         !  IF (BTEST(bit_pattern,4)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/0,0,-1/))
         !  IF (BTEST(bit_pattern,5)) rank=rs_grid_locate_rank(rs_desc,tasks(1,ntasks),(/0,0,+1/))
         bit_index = 0
         bit_pattern = 0
         DO i = 1, 3
            IF (rs_desc%perd(i) == 1) THEN
               bit_pattern = IBCLR(bit_pattern, bit_index)
               bit_index = bit_index+1
               bit_pattern = IBCLR(bit_pattern, bit_index)
               bit_index = bit_index+1
            ELSE
               ! fits the left neighbor ?
               IF (ubc(i) <= rs_desc%lb_global(i, dest)-1+rs_desc%border) THEN
                  bit_pattern = IBSET(bit_pattern, bit_index)
                  bit_index = bit_index+1
               ELSE
                  bit_pattern = IBCLR(bit_pattern, bit_index)
                  bit_index = bit_index+1
               ENDIF
               ! fits the right neighbor ?
               IF (lbc(i) >= rs_desc%ub_global(i, dest)+1-rs_desc%border) THEN
                  bit_pattern = IBSET(bit_pattern, bit_index)
                  bit_index = bit_index+1
               ELSE
                  bit_pattern = IBCLR(bit_pattern, bit_index)
                  bit_index = bit_index+1
               ENDIF
            ENDIF
         ENDDO
         tasks(6, ntasks) = bit_pattern

      ELSE
         ! generalised collocation/integration needed
         ! first we figure out how many neighbors we have to add to include the lbc/ubc
         ! in the available domains (inclusive of halo points)
         ! first we 'ignore' periodic boundary conditions
         ! i.e. ub_coord-lb_coord+1 might be larger than group_dim
         lb_coord = coord
         ub_coord = coord
         lb_domain = rs_desc%lb_global(:, dest)-rs_desc%border
         ub_domain = rs_desc%ub_global(:, dest)+rs_desc%border
         DO i = 1, 3
            ! only if the grid is not periodic in this direction we need to take care of adding neighbors
            IF (rs_desc%perd(i) == 0) THEN
               ! if the domain lower bound is greater than the lbc we need to add the size of the neighbor domain
               DO
                  IF (lb_domain(i) > lbc(i)) THEN
                     lb_coord(i) = lb_coord(i)-1
                     icoord = MODULO(lb_coord, rs_desc%group_dim)
                     idest = rs_desc%coord2rank(icoord(1), icoord(2), icoord(3))
                     lb_domain(i) = lb_domain(i)-(rs_desc%ub_global(i, idest)-rs_desc%lb_global(i, idest)+1)
                  ELSE
                     EXIT
                  ENDIF
               ENDDO
               ! same for the upper bound
               DO
                  IF (ub_domain(i) < ubc(i)) THEN
                     ub_coord(i) = ub_coord(i)+1
                     icoord = MODULO(ub_coord, rs_desc%group_dim)
                     idest = rs_desc%coord2rank(icoord(1), icoord(2), icoord(3))
                     ub_domain(i) = ub_domain(i)+(rs_desc%ub_global(i, idest)-rs_desc%lb_global(i, idest)+1)
                  ELSE
                     EXIT
                  ENDIF
               ENDDO
            ENDIF
         ENDDO

         ! some care is needed for the periodic boundaries
         DO i = 1, 3
            IF (ub_domain(i)-lb_domain(i)+1 >= rs_desc%npts(i)) THEN
               dir_periodic(i) = .TRUE.
               lb_coord(i) = 0
               ub_coord(i) = rs_desc%group_dim(i)-1
            ELSE
               dir_periodic(i) = .FALSE.
            ENDIF
         ENDDO

         added_tasks = PRODUCT(ub_coord-lb_coord+1)
         itask = ntasks
         ntasks = ntasks+added_tasks-1
         IF (ntasks > SIZE(tasks, 2)) THEN
            curr_tasks = INT((SIZE(tasks, 2)+add_tasks)*mult_tasks)
            CALL reallocate(tasks, 1, 6, 1, curr_tasks)
         END IF
         DO iz = lb_coord(3), ub_coord(3)
         DO iy = lb_coord(2), ub_coord(2)
         DO ix = lb_coord(1), ub_coord(1)
            icoord = MODULO((/ix, iy, iz/), rs_desc%group_dim)
            idest = rs_desc%coord2rank(icoord(1), icoord(2), icoord(3))
            tasks(1, itask) = encode_rank(idest, igrid_level, n_levels)
            tasks(4, itask) = 2
            tasks(6, itask) = 0
            ! encode the domain size for this task
            ! if the bit is set, we need to add the border in that direction
            IF (ix == lb_coord(1) .AND. .NOT. dir_periodic(1)) tasks(6, itask) = IBSET(tasks(6, itask), 0)
            IF (ix == ub_coord(1) .AND. .NOT. dir_periodic(1)) tasks(6, itask) = IBSET(tasks(6, itask), 1)
            IF (iy == lb_coord(2) .AND. .NOT. dir_periodic(2)) tasks(6, itask) = IBSET(tasks(6, itask), 2)
            IF (iy == ub_coord(2) .AND. .NOT. dir_periodic(2)) tasks(6, itask) = IBSET(tasks(6, itask), 3)
            IF (iz == lb_coord(3) .AND. .NOT. dir_periodic(3)) tasks(6, itask) = IBSET(tasks(6, itask), 4)
            IF (iz == ub_coord(3) .AND. .NOT. dir_periodic(3)) tasks(6, itask) = IBSET(tasks(6, itask), 5)
            itask = itask+1
         ENDDO
         ENDDO
         ENDDO
      ENDIF

   END SUBROUTINE rs_find_node

! **************************************************************************************************
!> \brief utility functions for encoding the grid level with a rank, allowing
!>        different grid levels to maintain a different rank ordering without
!>        losing information.  These encoded_ints are stored in the tasks(1:2,:) array
!> \param rank ...
!> \param grid_level ...
!> \param n_levels ...
!> \retval encoded_int ...
!> \par History
!>      4.2009 created [Iain Bethune]
!>        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
! **************************************************************************************************
   FUNCTION encode_rank(rank, grid_level, n_levels) RESULT(encoded_int)

      INTEGER, INTENT(IN)                                :: rank, grid_level, n_levels
      INTEGER(KIND=int_8)                                :: encoded_int

! ordered so can still sort by rank

      encoded_int = rank*n_levels+grid_level-1

   END FUNCTION

! **************************************************************************************************
!> \brief ...
!> \param encoded_int ...
!> \param n_levels ...
!> \retval rank ...
! **************************************************************************************************
   FUNCTION decode_rank(encoded_int, n_levels) RESULT(rank)

      INTEGER(KIND=int_8), INTENT(IN)                    :: encoded_int
      INTEGER, INTENT(IN)                                :: n_levels
      INTEGER                                            :: rank

      rank = INT(encoded_int/n_levels)

   END FUNCTION

! **************************************************************************************************
!> \brief ...
!> \param encoded_int ...
!> \param n_levels ...
!> \retval grid_level ...
! **************************************************************************************************
   FUNCTION decode_level(encoded_int, n_levels) RESULT(grid_level)

      INTEGER(KIND=int_8), INTENT(IN)                    :: encoded_int
      INTEGER, INTENT(IN)                                :: n_levels
      INTEGER                                            :: grid_level

      INTEGER(KIND=int_8)                                :: n_levels8

      n_levels8 = n_levels

      grid_level = INT(MODULO(encoded_int, n_levels8))+1

   END FUNCTION decode_level

END MODULE task_list_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!>   \brief
!>     Routines to efficently handle dense polynomial in 3 variables up to
!>     a given degree.
!>     Multiplication, partial evalution, affine transform (change of reference
!>     system), differentiation are efficiently implemented.
!>     some functions accept or return several polynomial together,
!>     these have to have all the same size, and are stored one after the other
!>     in an unique 1d array. This gives them an easy handling and even seem to
!>     be faster than the transposed layout.
!>   \note
!>     not all routines have been fully optimized.
!>     original available also with a BSD style license
!>   \author Fawzi Mohamed
! **************************************************************************************************
MODULE d3_poly

   USE kinds,                           ONLY: dp

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: poly_size1, poly_size2, poly_size3, &
             grad_size3, &
             poly_affine_t3, &
             poly_p_eval2b, &
             poly_p_eval3b, poly_cp2k2d3, &
             poly_padd_uneval3b, poly_padd_uneval2b, &
             poly_affine_t3t, poly_d32cp2k, init_d3_poly_module


   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'd3_poly'
! maximum grad for cached values
   INTEGER, PUBLIC, PARAMETER :: max_grad2 = 5
   INTEGER, PUBLIC, PARAMETER :: max_grad3 = 3
   INTEGER, PUBLIC, PARAMETER :: cached_dim1 = max_grad2+1
   INTEGER, PUBLIC, PARAMETER :: cached_dim2 = (max_grad2+1)*(max_grad2+2)/2
   INTEGER, PUBLIC, PARAMETER :: cached_dim3 = (max_grad3+1)*(max_grad3+2)*(max_grad3+3)/6

! cached index -> monomial exponents
   LOGICAL, SAVE                           :: module_initialized = .FALSE.
   INTEGER, SAVE, DIMENSION(2, cached_dim2) :: a_mono_exp2
   INTEGER, SAVE, DIMENSION(3, cached_dim3) :: a_mono_exp3
   INTEGER, SAVE, DIMENSION(cached_dim3)   :: a_reduce_idx3
   INTEGER, SAVE, DIMENSION(3, cached_dim3) :: a_deriv_idx3
   INTEGER, SAVE, DIMENSION(cached_dim2, cached_dim2) :: a_mono_mult2
   INTEGER, SAVE, DIMENSION(cached_dim3, cached_dim3) :: a_mono_mult3
   INTEGER, SAVE, DIMENSION(4, cached_dim3) :: a_mono_mult3a

CONTAINS

! **************************************************************************************************
!> \brief initialization of the cache, is called by functions in this module
!> that use cached values
! **************************************************************************************************
   SUBROUTINE init_d3_poly_module()
      INTEGER                                            :: grad, i, ii, ij, j, nthreads, subG
      INTEGER, DIMENSION(2)                              :: monoRes2
      INTEGER, DIMENSION(3)                              :: monoRes3

      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()
      IF (nthreads /= 1) CALL cp__b("d3_poly.F",78,"init_d3_poly_module must not be called within OMP PARALLEL")

      IF (module_initialized) RETURN

      ii = 1
      DO grad = 0, max_grad2
         DO i = grad, 0, -1
            a_mono_exp2(1, ii) = i
            a_mono_exp2(2, ii) = grad-i
            ii = ii+1
         END DO
      END DO
      ii = 1
      DO grad = 0, max_grad3
         DO i = grad, 0, -1
            DO j = grad-i, 0, -1
               a_mono_exp3(1, ii) = i
               a_mono_exp3(2, ii) = j
               a_mono_exp3(3, ii) = grad-i-j
               ii = ii+1
            END DO
         END DO
      END DO
      DO ii = 1, cached_dim3
         subG = a_mono_exp3(2, ii)+a_mono_exp3(3, ii)
         a_reduce_idx3(ii) = subG*(subG+1)/2+a_mono_exp3(3, ii)+1
      END DO
      DO ii = 1, cached_dim3
         IF (a_mono_exp3(1, ii) > 0) THEN
            a_deriv_idx3(1, ii) = mono_index3(a_mono_exp3(1, ii)-1, a_mono_exp3(2, ii), a_mono_exp3(3, ii))
         ELSE
            a_deriv_idx3(1, ii) = 0
         END IF
         IF (a_mono_exp3(2, ii) > 0) THEN
            a_deriv_idx3(2, ii) = mono_index3(a_mono_exp3(1, ii), a_mono_exp3(2, ii)-1, a_mono_exp3(3, ii))
         ELSE
            a_deriv_idx3(2, ii) = 0
         END IF
         IF (a_mono_exp3(3, ii) > 0) THEN
            a_deriv_idx3(3, ii) = mono_index3(a_mono_exp3(1, ii), a_mono_exp3(2, ii), a_mono_exp3(3, ii)-1)
         ELSE
            a_deriv_idx3(3, ii) = 0
         END IF
      END DO
      DO ii = 1, cached_dim2
         DO ij = ii, cached_dim2
            monoRes2 = a_mono_exp2(:, ii)+a_mono_exp2(:, ij)
            a_mono_mult2(ii, ij) = mono_index2(monoRes2(1), monoRes2(2))+1
            a_mono_mult2(ij, ii) = a_mono_mult2(ii, ij)
         END DO
      END DO
      DO ii = 1, cached_dim3
         DO ij = ii, cached_dim3
            monoRes3 = a_mono_exp3(:, ii)+a_mono_exp3(:, ij)
            a_mono_mult3(ii, ij) = mono_index3(monoRes3(1), monoRes3(2), monoRes3(3))+1
            a_mono_mult3(ij, ii) = a_mono_mult3(ii, ij)
         END DO
      END DO
      ii = 1
      DO i = 1, cached_dim3
         DO j = 1, 4
            a_mono_mult3a(j, i) = a_mono_mult3(j, i)
            ii = ii+1
         END DO
      END DO

      module_initialized = .TRUE.
   END SUBROUTINE

! **************************************************************************************************
!> \brief size of a polynomial in x up to the given degree
!> \param maxgrad ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION poly_size1(maxgrad) RESULT(res)
      INTEGER, INTENT(in)                                :: maxgrad
      INTEGER                                            :: res

      res = maxgrad+1
   END FUNCTION

! **************************************************************************************************
!> \brief size of a polynomial in x,y up to the given degree
!> \param maxgrad ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION poly_size2(maxgrad) RESULT(res)
      INTEGER, INTENT(in)                                :: maxgrad
      INTEGER                                            :: res

      res = (maxgrad+1)*(maxgrad+2)/2
   END FUNCTION

! **************************************************************************************************
!> \brief size of a polynomial in x,y,z up to the given degree
!> \param maxgrad ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION poly_size3(maxgrad) RESULT(res)
      INTEGER, INTENT(in)                                :: maxgrad
      INTEGER                                            :: res

      res = (maxgrad+1)*(maxgrad+2)*(maxgrad+3)/6
   END FUNCTION

! **************************************************************************************************
!> \brief max grad for a polynom of the given size
!> \param n ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION grad_size1(n) RESULT(res)
      INTEGER, INTENT(in)                                :: n
      INTEGER                                            :: res

      res = n-1
   END FUNCTION

! **************************************************************************************************
!> \brief max grad for a polynom of the given size
!> \param n ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION grad_size2(n) RESULT(res)
      INTEGER, INTENT(in)                                :: n
      INTEGER                                            :: res

      res = INT(FLOOR(0.5_dp*(SQRT(1.0_dp+8.0_dp*REAL(n, dp))-1.0_dp)-2.e-6_dp))
   END FUNCTION

! **************************************************************************************************
!> \brief max grad for a polynom of the given size
!> \param n ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION grad_size3(n) RESULT(res)
      INTEGER, INTENT(in)                                :: n
      INTEGER                                            :: res

      INTEGER                                            :: nn
      REAL(dp)                                           :: g1

      IF (n < 1) THEN
         res = -1
      ELSE
         nn = n*6
         g1 = (108.0_dp*nn+12.0_dp*SQRT(81.0_dp*nn*nn-12.0_dp))**(1.0_dp/3.0_dp)
         res = FLOOR(g1/6.0_dp+2.0_dp/g1-1.0_dp-2.e-6_dp)
      END IF
   END FUNCTION

! **************************************************************************************************
!> \brief 0-based index of monomial of the given degree
!> \param i ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_index1(i) RESULT(res)
      INTEGER, INTENT(in)                                :: i
      INTEGER                                            :: res

      res = i
   END FUNCTION

! **************************************************************************************************
!> \brief 0-based index of monomial of the given degree
!> \param i ...
!> \param j ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_index2(i, j) RESULT(res)
      INTEGER, INTENT(in)                                :: i, j
      INTEGER                                            :: res

      INTEGER                                            :: grad

      grad = i+j
      res = grad*(grad+1)/2+j
   END FUNCTION

! **************************************************************************************************
!> \brief 0-based index of monomial of the given degree
!> \param i ...
!> \param j ...
!> \param k ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_index3(i, j, k) RESULT(res)
      INTEGER, INTENT(in)                                :: i, j, k
      INTEGER                                            :: res

      INTEGER                                            :: grad, sgrad

      sgrad = j+k
      grad = i+sgrad
      res = grad*(grad+1)*(grad+2)/6+(sgrad)*(sgrad+1)/2+k
   END FUNCTION

! **************************************************************************************************
!> \brief exponents of the monomial at the given 0-based index
!> \param ii ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_exp1(ii) RESULT(res)
      INTEGER, INTENT(in)                                :: ii
      INTEGER                                            :: res

      res = ii
   END FUNCTION

! **************************************************************************************************
!> \brief exponents of the monomial at the given 0-based index
!> \param ii ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_exp2(ii) RESULT(res)
      INTEGER, INTENT(in)                                :: ii
      INTEGER, DIMENSION(2)                              :: res

      INTEGER                                            :: grad

      grad = INT(FLOOR(0.5_dp*(SQRT(9.0_dp+8.0_dp*ii)-1.0_dp)-2.e-6_dp))
      res(2) = ii-(grad)*(grad+1)/2
      res(1) = grad-res(2)
   END FUNCTION

! **************************************************************************************************
!> \brief exponents of the monomial at the given 0-based index
!> \param n ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_exp3(n) RESULT(res)
      INTEGER, INTENT(in)                                :: n
      INTEGER, DIMENSION(3)                              :: res

      INTEGER                                            :: grad, grad1, ii, nn
      REAL(dp)                                           :: g1

      nn = (n+1)*6
      g1 = (108.0_dp*nn+12.0_dp*SQRT(81.0_dp*nn*nn-12.0_dp))**(1.0_dp/3.0_dp)
      grad1 = INT(FLOOR(g1/6.0_dp+2.0_dp/g1-1.0_dp-2.e-6_dp))
      ii = n-grad1*(grad1+1)*(grad1+2)/6
      grad = INT(FLOOR(0.5_dp*(SQRT(9.0_dp+8.0_dp*ii)-1.0_dp)-1.e-6_dp))
      res(3) = ii-grad*(grad+1)/2
      res(2) = grad-res(3)
      res(1) = grad1-grad
   END FUNCTION

! **************************************************************************************************
!> \brief the index of the result of the multiplication of the two monomials
!> \param ii ...
!> \param ij ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_mult1(ii, ij) RESULT(res)
      INTEGER, INTENT(in)                                :: ii, ij
      INTEGER                                            :: res

      res = ii+ij
   END FUNCTION

! **************************************************************************************************
!> \brief the index of the result of the multiplication of the two monomials
!> \param ii ...
!> \param ij ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_mult2(ii, ij) RESULT(res)
      INTEGER, INTENT(in)                                :: ii, ij
      INTEGER                                            :: res

      INTEGER, DIMENSION(2)                              :: monoRes

      monoRes = mono_exp2(ii)+mono_exp2(ij)
      res = mono_index2(monoRes(1), monoRes(2))
   END FUNCTION

! **************************************************************************************************
!> \brief the index of the result of the multiplication of the two monomials
!> \param ii ...
!> \param ij ...
!> \retval res ...
! **************************************************************************************************
   PURE FUNCTION mono_mult3(ii, ij) RESULT(res)
      INTEGER, INTENT(in)                                :: ii, ij
      INTEGER                                            :: res

      INTEGER, DIMENSION(3)                              :: monoRes

      monoRes = mono_exp3(ii)+mono_exp3(ij)
      res = mono_index3(monoRes(1), monoRes(2), monoRes(3))
   END FUNCTION

! **************************************************************************************************
!> \brief multiplies the polynomials p1 with p2 using pRes to store the result
!> \param p1 ...
!> \param p2 ...
!> \param pRes ...
!> \param np1 ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_mult1(p1, p2, pRes, np1, sumUp)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p1, p2
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: np1
      LOGICAL, INTENT(in), OPTIONAL                      :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_mult1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ipoly, iPos, j, myNp1, newGrad, &
                                                            newSize, resPos, resShift_0, size_p1, &
                                                            size_p2
      LOGICAL                                            :: mySumUp

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",390,"module d3_poly not initialized")
      mySumUp = .FALSE.
      myNp1 = 1
      IF (PRESENT(np1)) myNp1 = np1
      IF (PRESENT(sumUp)) mySumUp = sumUp
      size_p1 = SIZE(p1)/myNp1
      size_p2 = SIZE(p2)
      newGrad = grad_size1(size_p1)+grad_size1(size_p2)
      newSize = SIZE(pRes)/myNp1
      IF(.NOT.(newSize >= poly_size1(newGrad)))CALL cp__a("d3_poly.F",399)
      IF (.NOT. mySumUp) pRes = 0
      iPos = 1
      resShift_0 = 0
      DO ipoly = 0, myNp1-1
         DO i = 1, size_p1
            resPos = resShift_0+i
            DO j = 1, size_p2
               pRes(resPos) = pRes(resPos)+p1(iPos)*p2(j)
               resPos = resPos+1
            END DO
            iPos = iPos+1
         END DO
         resShift_0 = resShift_0+newSize
      END DO
   END SUBROUTINE

! **************************************************************************************************
!> \brief multiplies p1 with p2 using pRes to store the result
!> \param p1 ...
!> \param p2 ...
!> \param pRes ...
!> \param np1 ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_mult2(p1, p2, pRes, np1, sumUp)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p1, p2
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: np1
      LOGICAL, INTENT(in), OPTIONAL                      :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_mult2', routineP = moduleN//':'//routineN

      INTEGER :: g1, g1g2, g2, grad1, grad2, i, ipoly, iShift, j, msize_p1, myNp1, newGrad, &
         newSize, shift1, shift2, shiftRes, shiftRes_0, size_p1, size_p2, subG2
      LOGICAL                                            :: mySumUp

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",436,"module d3_poly not initialized")
      mySumUp = .FALSE.
      IF (PRESENT(sumUp)) mySumUp = sumUp
      myNp1 = 1
      IF (PRESENT(np1)) myNp1 = np1
      size_p1 = SIZE(p1)/myNp1
      size_p2 = SIZE(p2)
      grad1 = grad_size2(size_p1)
      grad2 = grad_size2(size_p2)
      newGrad = grad1+grad2
      newSize = SIZE(pRes)/myNp1
      IF(.NOT.(newSize >= poly_size2(newGrad)))CALL cp__a("d3_poly.F",447)
      IF (.NOT. mySumUp) pRes = 0
      iShift = 0
      shiftRes = 0
      DO ipoly = 1, myNp1
         DO i = 1, MIN(size_p1, cached_dim2)
            DO j = 1, MIN(size_p2, cached_dim2)
               pRes(shiftRes+a_mono_mult2(j, i)) = pRes(shiftRes+a_mono_mult2(j, i)) &
                                                   +p1(iShift+i)*p2(j)
            END DO
         END DO
         iShift = iShift+size_p1
         shiftRes = shiftRes+newSize
      END DO
      IF (grad1 > max_grad2 .OR. grad2 > max_grad2) THEN
         msize_p1 = size_p1
         shiftRes_0 = 0
         DO ipoly = 0, myNp1-1
            shift1 = ipoly*size_p1
            DO g1 = 0, grad1
               ! shift1=g1*(g1+1)/2
               IF (g1 > max_grad2) THEN
                  subG2 = 0
                  shift2 = 0
                  g1g2 = shiftRes_0-1
               ELSE
                  subG2 = max_grad2+1
                  shift2 = cached_dim2
                  g1g2 = shiftRes_0+g1*subG2-1
               END IF
               DO g2 = subG2, grad2
                  ! shift2=g2*(g2+1)/2
                  shiftRes = shift1+shift2+g1g2 ! shiftRes=(g1+g2)*(g1+g2+1)/2-1+ipoly*newSize
                  DO i = 1, MIN(g1+1, msize_p1-shift1)
                     DO j = 1, MIN(g2+1, size_p2-shift2)
                        pRes(shiftRes+i+j) = pRes(shiftRes+i+j)+p1(shift1+i)*p2(shift2+j)
                     END DO
                  END DO
                  shift2 = shift2+g2+1 !
                  g1g2 = g1g2+g1 !
               END DO
               shift1 = shift1+g1+1 !
            END DO
            shiftRes_0 = shiftRes_0+newSize-size_p1
            msize_p1 = msize_p1+size_p1
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief multiplies p1 with p2 using pRes to store the result
!> \param p1 ...
!> \param p2 ...
!> \param pRes ...
!> \param np1 ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_mult3(p1, p2, pRes, np1, sumUp)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p1, p2
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: np1
      LOGICAL, INTENT(in), OPTIONAL                      :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_mult3', routineP = moduleN//':'//routineN

      INTEGER                                            :: grad1, grad2, myNp1, newGrad, newSize, &
                                                            size_p1, size_p2
      LOGICAL                                            :: mySumUp

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",516,"module d3_poly not initialized")
      mySumUp = .FALSE.
      IF (PRESENT(sumUp)) mySumUp = sumUp
      myNp1 = 1
      IF (PRESENT(np1)) myNp1 = np1
      size_p1 = SIZE(p1)/myNp1
      size_p2 = SIZE(p2)
      grad1 = grad_size3(size_p1)
      grad2 = grad_size3(size_p2)
      newGrad = grad1+grad2
      newSize = SIZE(pRes)/myNp1
      IF(.NOT.(newSize >= poly_size3(newGrad)))CALL cp__a("d3_poly.F",527)
      CALL poly_mult3b(p1, SIZE(p1), grad1, p2, SIZE(p2), grad2, pRes, SIZE(pRes), myNp1, mySumUp)
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine of poly_mult3 without checks
!> \param p1 ...
!> \param size_p1 ...
!> \param grad1 ...
!> \param p2 ...
!> \param size_p2 ...
!> \param grad2 ...
!> \param pRes ...
!> \param size_pRes ...
!> \param np1 ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_mult3b(p1, size_p1, grad1, p2, size_p2, grad2, pRes, size_pRes, np1, sumUp)
      INTEGER, INTENT(in)                                :: size_p1
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: p1
      INTEGER, INTENT(in)                                :: grad1, size_p2
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: p2
      INTEGER, INTENT(in)                                :: grad2, size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: pRes
      INTEGER, INTENT(in)                                :: np1
      LOGICAL, INTENT(in)                                :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_mult3b', routineP = moduleN//':'//routineN

      INTEGER :: g1, g2, i, i1, i2, ipoly, iShift, j, j1, j2, msize_p1, my_size_p1, newSize, &
         shift1, shift1I, shift1J, shift2, shift2I, shift2J, shiftRes, shiftRes_0, shiftResI, &
         shiftResI_0, shiftResJ, subG2, subGrad

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",563,"module d3_poly not initialized")
      my_size_p1 = size_p1/np1
      newSize = size_pRes/np1

      IF (.NOT. sumUp) pRes(1:size_pRes) = 0.0_dp
      iShift = 0
      shiftRes = 0
      DO ipoly = 1, np1
         DO i = 1, MIN(my_size_p1, cached_dim3)
            DO j = 1, MIN(size_p2, cached_dim3)
               pRes(shiftRes+a_mono_mult3(j, i)) = pRes(shiftRes+a_mono_mult3(j, i)) &
                                                   +p1(iShift+i)*p2(j)
            END DO
         END DO
         iShift = iShift+my_size_p1
         shiftRes = shiftRes+newSize
      END DO
      IF (grad1 > max_grad3 .OR. grad2 > max_grad3) THEN
         ! one could remove multiplications even more agressively...
         msize_p1 = my_size_p1
         DO ipoly = 0, np1-1
            shift1 = 1+ipoly*my_size_p1
            shiftRes_0 = 1+ipoly*newSize
            DO g1 = 0, grad1
               IF (g1 > max_grad3) THEN
                  subG2 = 0
                  shift2 = 1
               ELSE
                  subG2 = max_grad3+1
                  shift2 = subG2*(subG2+1)*(subG2+2)/6+1
               END IF
               DO g2 = subG2, grad2
                  shiftRes = (g1+g2)*(g1+g2+1)*(g1+g2+2)/6+shiftRes_0
                  shift1I = shift1
                  shiftResI_0 = shiftRes
                  DO i1 = g1, 0, -1
                     IF (shift1I > msize_p1) EXIT
                     shift2I = shift2
                     shiftResI = shiftResI_0
                     subGrad = g1-i1
                     DO i2 = g2, 0, -1
                        !subGrad=g1+g2-i1-i2
                        !shiftResI=shiftRes+(subGrad)*(subGrad+1)/2
                        !shift2I=shift2+(g2-i2)*(g2-i2+1)/2
                        IF (shift2I > size_p2) EXIT
                        DO j1 = g1-i1, 0, -1
                           shift1J = shift1I+g1-i1-j1
                           IF (shift1J > msize_p1) EXIT
                           DO j2 = g2-i2, 0, -1
                              shift2J = shift2I+g2-i2-j2
                              IF (shift2J > size_p2) EXIT
                              shiftResJ = shiftResI+(subGrad-j1-j2)
                              ! shift1J=mono_index3(i1,j1,g1-i1-j1)+ipoly*my_size_p1+1
                              ! shift2J=mono_index3(i2,j2,g2-i2-j2)+1
                              ! shiftResJ=mono_index3(i1+i2,j1+j2,g1+g2-i1-i2-j1-j2)+ipoly*newSize+1
                              pRes(shiftResJ) = pRes(shiftResJ)+p1(shift1J)*p2(shift2J)
                           END DO
                        END DO
                        subGrad = subGrad+1
                        shift2I = shift2I+(g2-i2+1)
                        shiftResI = shiftResI+subGrad
                     END DO
                     shift1I = shift1I+(g1-i1+1)
                     shiftResI_0 = shiftResI_0+(g1-i1+1)
                  END DO
                  shift2 = shift2+(g2+1)*(g2+2)/2
               END DO
               shift1 = shift1+(g1+1)*(g1+2)/2
            END DO
            msize_p1 = msize_p1+my_size_p1
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine that multiplies with a polynomial of grad 1
!> \param p1 ...
!> \param size_p1 ...
!> \param grad1 ...
!> \param p2 ...
!> \param pRes ...
!> \param size_pRes ...
!> \param np1 ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_mult3ab(p1, size_p1, grad1, p2, pRes, size_pRes, np1, sumUp)
      INTEGER, INTENT(in)                                :: size_p1
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: p1
      INTEGER, INTENT(in)                                :: grad1
      REAL(dp), DIMENSION( *), INTENT(in)    :: p2
      INTEGER, INTENT(in)                                :: size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: pRes
      INTEGER, INTENT(in)                                :: np1
      LOGICAL, INTENT(in)                                :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_mult3ab', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: grad2 = 1

      INTEGER :: g1, g2, i, i1, i2, ipoly, iShift, j1, j2, msize_p1, my_size_p1, newSize, shift1, &
         shift1I, shift1J, shift2, shift2I, shift2J, shiftRes, shiftRes_0, shiftResI, shiftResI_0, &
         shiftResJ, subG2, subGrad

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",667,"module d3_poly not initialized")
      my_size_p1 = size_p1/np1
      newSize = size_pRes/np1

      IF (.NOT. sumUp) pRes(1:size_pRes) = 0.0_dp
      iShift = 0
      shiftRes = 0
      DO ipoly = 1, np1
         DO i = 1, MIN(my_size_p1, cached_dim3)
            pRes(shiftRes+a_mono_mult3a(1, i)) = pRes(shiftRes+a_mono_mult3a(1, i)) &
                                                 +p1(iShift+i)*p2(1)
            pRes(shiftRes+a_mono_mult3a(2, i)) = pRes(shiftRes+a_mono_mult3a(2, i)) &
                                                 +p1(iShift+i)*p2(2)
            pRes(shiftRes+a_mono_mult3a(3, i)) = pRes(shiftRes+a_mono_mult3a(3, i)) &
                                                 +p1(iShift+i)*p2(3)
            pRes(shiftRes+a_mono_mult3a(4, i)) = pRes(shiftRes+a_mono_mult3a(4, i)) &
                                                 +p1(iShift+i)*p2(4)
         END DO
         iShift = iShift+my_size_p1
         shiftRes = shiftRes+newSize
      END DO
      IF (grad1 > max_grad3 .OR. grad2 > max_grad3) THEN
         ! one could remove multiplications even more agressively...
         msize_p1 = my_size_p1
         DO ipoly = 0, np1-1
            shift1 = 1+ipoly*my_size_p1+(max_grad3+1)*(max_grad3+2)*(max_grad3+3)/6
            shiftRes_0 = 1+ipoly*newSize
            DO g1 = max_grad3+1, grad1
               subG2 = 0
               shift2 = 1
               DO g2 = subG2, grad2
                  shiftRes = (g1+g2)*(g1+g2+1)*(g1+g2+2)/6+shiftRes_0
                  shift1I = shift1
                  shiftResI_0 = shiftRes
                  DO i1 = g1, 0, -1
                     IF (shift1I > msize_p1) EXIT
                     shift2I = shift2
                     shiftResI = shiftResI_0
                     subGrad = g1-i1
                     DO i2 = g2, 0, -1
                        !subGrad=g1+g2-i1-i2
                        !shiftResI=shiftRes+(subGrad)*(subGrad+1)/2
                        !shift2I=shift2+(g2-i2)*(g2-i2+1)/2
                        DO j1 = g1-i1, 0, -1
                           shift1J = shift1I+g1-i1-j1
                           IF (shift1J > msize_p1) EXIT
                           DO j2 = g2-i2, 0, -1
                              shift2J = shift2I+g2-i2-j2
                              shiftResJ = shiftResI+(subGrad-j1-j2)
                              ! shift1J=mono_index3(i1,j1,g1-i1-j1)+ipoly*my_size_p1+1
                              ! shift2J=mono_index3(i2,j2,g2-i2-j2)+1
                              ! shiftResJ=mono_index3(i1+i2,j1+j2,g1+g2-i1-i2-j1-j2)+ipoly*newSize+1
                              pRes(shiftResJ) = pRes(shiftResJ)+p1(shift1J)*p2(shift2J)
                           END DO
                        END DO
                        subGrad = subGrad+1
                        shift2I = shift2I+(g2-i2+1)
                        shiftResI = shiftResI+subGrad
                     END DO
                     shift1I = shift1I+(g1-i1+1)
                     shiftResI_0 = shiftResI_0+(g1-i1+1)
                  END DO
                  shift2 = shift2+(g2+1)*(g2+2)/2
               END DO
               shift1 = shift1+(g1+1)*(g1+2)/2
            END DO
            msize_p1 = msize_p1+my_size_p1
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief writes out a 1d polynomial in a human readable form
!> \param p ...
!> \param out_f ...
! **************************************************************************************************
   SUBROUTINE poly_write1(p, out_f)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      INTEGER, INTENT(in)                                :: out_f

      INTEGER                                            :: i
      LOGICAL                                            :: did_write

      did_write = .FALSE.
      DO i = 1, SIZE(p)
         IF (p(i) /= 0) THEN
            IF (p(i) >= 0) WRITE (out_f, '("+")', advance='NO')
            WRITE (out_f, '(G20.10)', advance='NO') p(i)
            IF (i /= 1) WRITE (out_f, '("*x^",I3)', advance='NO') i-1
            did_write = .TRUE.
         END IF
      END DO
      IF (.NOT. did_write) WRITE (out_f, '("0.0")', advance='NO')
   END SUBROUTINE

! **************************************************************************************************
!> \brief writes out a 2d polynomial in a human readable form
!> \param p ...
!> \param out_f ...
! **************************************************************************************************
   SUBROUTINE poly_write2(p, out_f)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      INTEGER, INTENT(in)                                :: out_f

      INTEGER                                            :: i
      INTEGER, DIMENSION(2)                              :: mono_e
      LOGICAL                                            :: did_write

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",775,"module d3_poly not initialized")
      did_write = .FALSE.
      DO i = 1, SIZE(p)
         mono_e = mono_exp2(i-1)
         IF (p(i) /= 0) THEN
            IF (p(i) >= 0) WRITE (out_f, '("+")', advance='NO')
            WRITE (out_f, '(G20.10)', advance='NO') p(i)
            IF (mono_e(1) /= 0) WRITE (out_f, '("*x^",I3)', advance='NO') mono_e(1)
            IF (mono_e(2) /= 0) WRITE (out_f, '("*y^",I3)', advance='NO') mono_e(2)
            did_write = .TRUE.
         END IF
      END DO
      IF (.NOT. did_write) WRITE (out_f, '("0.0")', advance='NO')
   END SUBROUTINE

! **************************************************************************************************
!> \brief writes out the polynomial in a human readable form
!> \param p ...
!> \param out_f ...
! **************************************************************************************************
   SUBROUTINE poly_write3(p, out_f)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      INTEGER, INTENT(in)                                :: out_f

      INTEGER                                            :: i
      INTEGER, DIMENSION(3)                              :: mono_e
      LOGICAL                                            :: did_write

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",803,"module d3_poly not initialized")
      did_write = .FALSE.
      DO i = 1, SIZE(p)
         mono_e = mono_exp3(i-1)
         IF (p(i) /= 0) THEN
            IF (p(i) >= 0) WRITE (out_f, '("+")', advance='NO')
            WRITE (out_f, '(G20.10)', advance='NO') p(i)
            IF (mono_e(1) /= 0) WRITE (out_f, '("*x^",I3)', advance='NO') mono_e(1)
            IF (mono_e(2) /= 0) WRITE (out_f, '("*y^",I3)', advance='NO') mono_e(2)
            IF (mono_e(3) /= 0) WRITE (out_f, '("*z^",I3)', advance='NO') mono_e(3)
            did_write = .TRUE.
         END IF
      END DO
      IF (.NOT. did_write) WRITE (out_f, '("0.0")', advance='NO')
   END SUBROUTINE

! **************************************************************************************************
!> \brief random poly with coeffiecents that are easy to print exactly,
!>        of the given maximum size (for testing purposes)
!> \param p ...
!> \param maxSize ...
!> \param minSize ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION poly_random(p, maxSize, minSize) RESULT(res)
      REAL(dp), DIMENSION(:), INTENT(out)                :: p
      INTEGER, INTENT(in)                                :: maxSize
      INTEGER, INTENT(in), OPTIONAL                      :: minSize
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_random', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, myMinSize, pSize
      REAL(dp)                                           :: g

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",838,"module d3_poly not initialized")
      myMinSize = 1
      IF (PRESENT(minSize)) myMinSize = minSize
      CALL RANDOM_NUMBER(g)
      pSize = MIN(maxSize, myMinSize+INT((maxSize-myMinSize+1)*g))
      IF(.NOT.(SIZE(p) >= pSize))CALL cp__a("d3_poly.F",843)
      CALL RANDOM_NUMBER(p)
      DO i = 1, pSize
         p(i) = REAL(INT(p(i)*200.0_dp-100.0_dp), dp)/100.0_dp
      END DO
      DO i = pSize+1, SIZE(p)
         p(i) = 0.0_dp
      END DO
      res = pSize
   END FUNCTION

! **************************************************************************************************
!> \brief returns in the polynomials pRes the transpose of the
!> affine transformation x -> m*x+b of p
!> \param p ...
!> \param m ...
!> \param b ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_affine_t3t(p, m, b, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), DIMENSION(3, 3), INTENT(in)              :: m
      REAL(dp), DIMENSION(3), INTENT(in)                 :: b
      REAL(dp), DIMENSION(:), INTENT(out)                :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_affine_t3t', &
         routineP = moduleN//':'//routineN

      INTEGER :: grad, i, igrad, ii, ii1, ipoly, j, k, minResSize, monoDim1, monoDim2, monoDimAtt, &
         monoDimAtt2, monoFullDim1, monoFullDim2, monoSize1, monoSize2, my_npoly, pcoeff, pIdx, &
         pShift, rescoeff, resShift, rest_size_p, size_p, size_res, start_idx1
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: monoG1, monoG2
      REAL(dp), DIMENSION(4, 3)                          :: basepoly

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",879,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      basepoly(1, :) = b
      DO j = 1, 3
         DO i = 1, 3
            basepoly(j+1, i) = m(i, j)
         END DO
      END DO
      size_p = SIZE(pRes)/my_npoly
      size_res = SIZE(p)/my_npoly
      grad = grad_size3(size_res)
      minResSize = poly_size3(grad)
      IF(.NOT.(size_res == minResSize))CALL cp__a("d3_poly.F",892)
      IF(.NOT.(size_p >= minResSize))CALL cp__a("d3_poly.F",893)
      pRes = 0
      IF (size_p == 0) RETURN
      ii1 = 1
      ii = 1
      DO ipoly = 1, my_npoly
         pRes(ii1) = p(ii)
         ii = ii+size_res
         ii1 = ii1+size_p
      END DO
      IF (size_p == 1) RETURN

      ALLOCATE (monoG1((grad+1)*(grad+2)/2*minResSize), &
                monoG2((grad+1)*(grad+2)/2*minResSize))
      !monoG1=0
      !monoG2=0
      ii1 = 1
      DO j = 1, 3
         DO i = 1, 4
            monoG1(ii1) = basepoly(i, j)
            ii1 = ii1+1
         END DO
      END DO
      ii1 = 2
      igrad = 1
      monoDim1 = 4
      monoSize1 = 3
      monoFullDim1 = monoDim1*monoSize1
      rest_size_p = size_p-1
      DO
         k = MIN(rest_size_p, monoSize1)
         !call dgemm('T','N',monoDim1,my_npoly,k,&
         !    1.0_dp,monoG1,monoDim1,p(ii1:),size_p,1.0_dp,pRes,size_res)
         resShift = 0
         pShift = ii1
         DO ipoly = 1, my_npoly
            pIdx = pShift
            ii = 1
            DO pcoeff = 1, k
               DO rescoeff = 1, monoDim1
                  pRes(pIdx) = pRes(pIdx)+p(resShift+rescoeff)*monoG1(ii)
                  ii = ii+1
               END DO
               pIdx = pIdx+1
            END DO
            resShift = resShift+size_res
            pShift = pShift+size_p
         END DO

         rest_size_p = rest_size_p-k
         ii1 = ii1+k
         IF (rest_size_p <= 0) EXIT

         monoSize2 = igrad+2+monoSize1
         monoDim2 = monoDim1+monoSize2
         monoFullDim2 = monoSize2*monoDim2
         monoDimAtt = monoSize1*monoDim2
         CALL poly_mult3ab( monoG1(1), monoFullDim1, igrad, &
                            basepoly(1, 1), &
                            monoG2(1), monoDimAtt, monoSize1, .FALSE.)
         monoDimAtt2 = monoFullDim2-monoDim2
         start_idx1 = (monoSize1-igrad-1)*monoDim1
         CALL poly_mult3ab( monoG1(start_idx1+1), &
                           monoFullDim1-start_idx1, igrad,  basepoly(1, 2), &
                            monoG2(monoDimAtt+1), &
                           monoDimAtt2-monoDimAtt, igrad+1, .FALSE.)
         CALL poly_mult3ab( monoG1(monoFullDim1-monoDim1+1), &
                           monoDim1, igrad,  basepoly(1, 3), &
                            monoG2(monoDimAtt2+1), &
                           monoFullDim2-monoDimAtt2, 1, .FALSE.)
         igrad = igrad+1

         ! even grads

         k = MIN(rest_size_p, monoSize2)
         !call dgemm('T','N',monoDim2,my_npoly,k,&
         !    1.0_dp,monoG2,monoDim2,p(ii1:),size_p,1.0_dp,pRes,size_res)
         resShift = 0
         pShift = ii1
         DO ipoly = 1, my_npoly
            pIdx = pShift
            ii = 1
            DO pcoeff = 1, k
               DO rescoeff = 1, monoDim2
                  pRes(pIdx) = pRes(pIdx)+p(resShift+rescoeff)*monoG2(ii)
                  ii = ii+1
               END DO
               pIdx = pIdx+1
            END DO
            resShift = resShift+size_res
            pShift = pShift+size_p
         END DO

         rest_size_p = rest_size_p-k
         ii1 = ii1+k
         IF (rest_size_p <= 0) EXIT

         monoSize1 = igrad+2+monoSize2
         monoDim1 = monoDim2+monoSize1
         monoFullDim1 = monoSize1*monoDim1
         monoDimAtt = monoSize2*monoDim1
         CALL poly_mult3ab( monoG2(1), monoFullDim2, igrad, &
                            basepoly(1, 1),  monoG1(1), &
                           monoDimAtt, monoSize2, .FALSE.)
         monoDimAtt2 = monoFullDim1-monoDim1
         start_idx1 = (monoSize2-igrad-1)*monoDim2
         CALL poly_mult3ab( monoG2(start_idx1+1), &
                           monoFullDim2-start_idx1, igrad,  basepoly(1, 2), &
                            monoG1(monoDimAtt+1), monoDimAtt2-monoDimAtt, &
                           igrad+1, .FALSE.)
         CALL poly_mult3ab( monoG2(monoFullDim2-monoDim2+1), &
                           monoDim2, igrad,  basepoly(1, 3), &
                            monoG1(monoDimAtt2+1), &
                           monoFullDim1-monoDimAtt2, 1, .FALSE.)
         igrad = igrad+1

         ! ! alterantive to unrolling
         ! monoG1=monoG2
         ! monoSize1=monoSize2
         ! monoDim1=monoDim2
         ! monoFullDim1=monoFullDim2
      END DO
      DEALLOCATE (monoG1, monoG2)
   END SUBROUTINE

! **************************************************************************************************
!> \brief returns in the polynomials pRes the affine transformation x -> m*x+b of p
!> \param p ...
!> \param m ...
!> \param b ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_affine_t3(p, m, b, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), DIMENSION(3, 3), INTENT(in)              :: m
      REAL(dp), DIMENSION(3), INTENT(in)                 :: b
      REAL(dp), DIMENSION(:), INTENT(out)                :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_affine_t3', routineP = moduleN//':'//routineN

      INTEGER :: grad, i, igrad, ii, ii1, ipoly, j, k, minResSize, monoDim1, monoDim2, monoDimAtt, &
         monoDimAtt2, monoFullDim1, monoFullDim2, monoSize1, monoSize2, my_npoly, pcoeff, pIdx, &
         pShift, rescoeff, resShift, rest_size_p, size_p, size_res, start_idx1
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: monoG1, monoG2
      REAL(dp), DIMENSION(4, 3)                          :: basepoly

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1041,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      basepoly(1, :) = b
      DO j = 1, 3
         DO i = 1, 3
            basepoly(j+1, i) = m(i, j)
         END DO
      END DO
      size_p = SIZE(p)/my_npoly
      grad = grad_size3(size_p)
      size_res = SIZE(pRes)/my_npoly
      minResSize = poly_size3(grad)
      IF(.NOT.(size_res >= minResSize))CALL cp__a("d3_poly.F",1054)
      pRes = 0
      IF (size_p == 0) RETURN
      ii1 = 1
      ii = 1
      DO ipoly = 1, my_npoly
         pRes(ii) = p(ii1)
         ii = ii+size_res
         ii1 = ii1+size_p
      END DO
      IF (size_p == 1) RETURN

      ALLOCATE (monoG1((grad+1)*(grad+2)/2*minResSize), &
                monoG2((grad+1)*(grad+2)/2*minResSize))
      monoG1 = 0
      monoG2 = 0
      ii1 = 1
      DO j = 1, 3
         DO i = 1, 4
            monoG1(ii1) = basepoly(i, j)
            ii1 = ii1+1
         END DO
      END DO
      ii1 = 2
      igrad = 1
      monoDim1 = 4
      monoSize1 = 3
      monoFullDim1 = monoDim1*monoSize1
      rest_size_p = size_p-1
      DO
         k = MIN(rest_size_p, monoSize1)
         !call dgemm('T','N',monoDim1,my_npoly,k,&
         !    1.0_dp,monoG1,monoDim1,p(ii1:),size_p,1.0_dp,pRes,size_res)
         resShift = 0
         pShift = ii1
         DO ipoly = 1, my_npoly
            pIdx = pShift
            ii = 1
            DO pcoeff = 1, k
               DO rescoeff = 1, monoDim1
                  pRes(resShift+rescoeff) = pRes(resShift+rescoeff)+p(pIdx)*monoG1(ii)
                  ii = ii+1
               END DO
               pIdx = pIdx+1
            END DO
            resShift = resShift+size_res
            pShift = pShift+size_p
         END DO

         rest_size_p = rest_size_p-k
         ii1 = ii1+k
         IF (rest_size_p <= 0) EXIT

         monoSize2 = igrad+2+monoSize1
         monoDim2 = monoDim1+monoSize2
         monoFullDim2 = monoSize2*monoDim2
         monoDimAtt = monoSize1*monoDim2
         CALL poly_mult3ab( monoG1(1), monoFullDim1, igrad, &
                            basepoly(1, 1), &
                            monoG2(1), monoDimAtt, monoSize1, .FALSE.)
         monoDimAtt2 = monoFullDim2-monoDim2
         start_idx1 = (monoSize1-igrad-1)*monoDim1
         CALL poly_mult3ab( monoG1(start_idx1+1), &
                           monoFullDim1-start_idx1, igrad,  basepoly(1, 2), &
                            monoG2(monoDimAtt+1), &
                           monoDimAtt2-monoDimAtt, igrad+1, .FALSE.)
         CALL poly_mult3ab( monoG1(monoFullDim1-monoDim1+1), &
                           monoDim1, igrad,  basepoly(1, 3), &
                            monoG2(monoDimAtt2+1), &
                           monoFullDim2-monoDimAtt2, 1, .FALSE.)
         igrad = igrad+1

         ! even grads

         k = MIN(rest_size_p, monoSize2)
         !call dgemm('T','N',monoDim2,my_npoly,k,&
         !    1.0_dp,monoG2,monoDim2,p(ii1:),size_p,1.0_dp,pRes,size_res)
         resShift = 0
         pShift = ii1
         DO ipoly = 1, my_npoly
            pIdx = pShift
            ii = 1
            DO pcoeff = 1, k
               DO rescoeff = 1, monoDim2
                  pRes(resShift+rescoeff) = pRes(resShift+rescoeff)+p(pIdx)*monoG2(ii)
                  ii = ii+1
               END DO
               pIdx = pIdx+1
            END DO
            resShift = resShift+size_res
            pShift = pShift+size_p
         END DO

         rest_size_p = rest_size_p-k
         ii1 = ii1+k
         IF (rest_size_p <= 0) EXIT

         monoSize1 = igrad+2+monoSize2
         monoDim1 = monoDim2+monoSize1
         monoFullDim1 = monoSize1*monoDim1
         monoDimAtt = monoSize2*monoDim1
         CALL poly_mult3ab( monoG2(1), monoFullDim2, igrad, &
                            basepoly(1, 1), &
                            monoG1(1), monoDimAtt, monoSize2, .FALSE.)
         monoDimAtt2 = monoFullDim1-monoDim1
         start_idx1 = (monoSize2-igrad-1)*monoDim2
         CALL poly_mult3ab( monoG2(start_idx1+1), &
                           monoFullDim2-start_idx1, igrad, &
                            basepoly(1, 2), &
                            monoG1(monoDimAtt+1), monoDimAtt2-monoDimAtt, &
                           igrad+1, .FALSE.)
         CALL poly_mult3ab( monoG2(monoFullDim2-monoDim2+1), &
                           monoDim2, igrad,  basepoly(1, 3), &
                            monoG1(monoDimAtt2+1), &
                           monoFullDim1-monoDimAtt2, 1, .FALSE.)
         igrad = igrad+1

         ! ! alterantive to unrolling
         ! monoG1=monoG2
         ! monoSize1=monoSize2
         ! monoDim1=monoDim2
         ! monoFullDim1=monoFullDim2
      END DO
      DEALLOCATE (monoG1, monoG2)
   END SUBROUTINE

! **************************************************************************************************
!> \brief evaluates the 3d polymial at x (the result is a polynomial in two variables)
!> \param p ...
!> \param x ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_p_eval3(p, x, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), INTENT(in)                               :: x
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_p_eval3', routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, my_npoly, newSize, size_p
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1198,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      grad = grad_size3(size_p)
      newSize = SIZE(pRes)/my_npoly
      IF(.NOT.(newSize >= poly_size2(grad)))CALL cp__a("d3_poly.F",1204)
      pRes = 0.0
      ALLOCATE (xi(grad+1))
      CALL poly_p_eval3b(p, SIZE(p), x, pRes, SIZE(pRes), my_npoly, grad, xi)
      DEALLOCATE (xi)
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine of poly_p_eval3 without checks
!> \param p ...
!> \param size_p ...
!> \param x ...
!> \param pRes ...
!> \param size_pRes ...
!> \param npoly ...
!> \param grad ...
!> \param xi ...
! **************************************************************************************************
   SUBROUTINE poly_p_eval3b(p, size_p, x, pRes, size_pRes, npoly, grad, xi)
      INTEGER, INTENT(in)                                :: size_p
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: p
      REAL(dp), INTENT(in)                               :: x
      INTEGER, INTENT(in)                                :: size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: pRes
      INTEGER, INTENT(in)                                :: npoly, grad
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: xi

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_p_eval3b', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, igrad, ii, ii0, inSize, ipoly, j, &
                                                            msize_p, newSize, pShift, shiftRes, &
                                                            shiftRes_0, subG

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1240,"module d3_poly not initialized")
      inSize = size_p/npoly
      newSize = size_pRes/npoly
      pRes(1:size_pRes) = 0.0
      xi(1) = 1.0
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      shiftRes = 0
      pShift = 0
      DO ipoly = 1, npoly
         DO ii = 1, MIN(inSize, cached_dim3)
            pRes(shiftRes+a_reduce_idx3(ii)) = pRes(shiftRes+a_reduce_idx3(ii))+p(pShift+ii)*xi(a_mono_exp3(1, ii)+1)
         END DO
         shiftRes = shiftRes+newSize
         pShift = pShift+inSize
      END DO
      IF (grad > max_grad3) THEN
         ii0 = (max_grad3+1)*(max_grad3+2)*(max_grad3+3)/6+1
         shiftRes_0 = 1
         msize_p = inSize
         DO ipoly = 1, npoly
            ii = ii0
            grad_do: DO igrad = max_grad3+1, grad
               !ii=igrad*(igrad+1)*(igrad+2)/6+1
               shiftRes = shiftRes_0
               subG = 0
               DO i = igrad, 0, -1
                  !subG=igrad-i
                  !shiftRes=subG*(subG+3)/2+1
                  DO j = subG, 0, -1
                     IF (msize_p < ii) EXIT grad_do
                     pRes(shiftRes-j) = pRes(shiftRes-j)+p(ii)*xi(i+1)
                     ii = ii+1
                  END DO
                  shiftRes = shiftRes+subG+2
                  subG = subG+1
               END DO
            END DO grad_do
            ii0 = ii0+inSize
            shiftRes_0 = shiftRes_0+newSize
            msize_p = msize_p+inSize
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief unevaluates a 2d polymial to a 3d polynomial at x
!>  p(a,b,c)=p(a,b,c)+sum(pRes(b,c)*(x*a)^i,i), this is *not* the inverse of poly_p_eval3
!>  adds to p
!> \param p ...
!> \param x ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_padd_uneval3(p, x, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(inout)              :: p
      REAL(dp), INTENT(in)                               :: x
      REAL(dp), DIMENSION(:), INTENT(in)                 :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_padd_uneval3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, my_npoly, newSize, size_p
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1307,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      newSize = SIZE(pRes)/my_npoly
      grad = grad_size2(newSize)
      IF(.NOT.(size_p >= poly_size3(grad)))CALL cp__a("d3_poly.F",1313)
      IF(.NOT.(newSize == poly_size2(grad)))CALL cp__a("d3_poly.F",1314)
      ALLOCATE (xi(grad+1))
      CALL poly_padd_uneval3b(p, SIZE(p), x, pRes, SIZE(pRes), my_npoly, grad, xi)
      DEALLOCATE (xi)
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine of poly_padd_uneval3 without checks
!> \param p ...
!> \param size_p ...
!> \param x ...
!> \param pRes ...
!> \param size_pRes ...
!> \param npoly ...
!> \param grad ...
!> \param xi ...
!> \note loop should be structured differently (more contiguous pRes access)
! **************************************************************************************************
   SUBROUTINE poly_padd_uneval3b(p, size_p, x, pRes, size_pRes, npoly, grad, xi)
      INTEGER, INTENT(in)                                :: size_p
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: p
      REAL(dp), INTENT(in)                               :: x
      INTEGER, INTENT(in)                                :: size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: pRes
      INTEGER, INTENT(in)                                :: npoly, grad
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: xi

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_padd_uneval3b', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, igrad, ii, ii0, inSize, ipoly, j, &
                                                            msize_p, newSize, pShift, shiftRes, &
                                                            shiftRes_0, subG, upSize

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1351,"module d3_poly not initialized")
      inSize = size_p/npoly
      newSize = size_pRes/npoly
      upSize = (grad+1)*(grad+2)*(grad+3)/6
      xi(1) = 1.0
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      shiftRes = 0
      pShift = 0
      DO ipoly = 1, npoly
         DO ii = 1, MIN(upSize, cached_dim3)
            p(pShift+ii) = p(pShift+ii)+pRes(shiftRes+a_reduce_idx3(ii))*xi(a_mono_exp3(1, ii)+1)
         END DO
         shiftRes = shiftRes+newSize
         pShift = pShift+inSize
      END DO
      IF (grad > max_grad3) THEN
         ii0 = (max_grad3+1)*(max_grad3+2)*(max_grad3+3)/6+1
         shiftRes_0 = 1
         msize_p = upSize
         DO ipoly = 1, npoly
            ii = ii0
            grad_do: DO igrad = max_grad3+1, grad
               !ii=igrad*(igrad+1)*(igrad+2)/6+1
               shiftRes = shiftRes_0
               subG = 0
               DO i = igrad, 0, -1
                  !subG=igrad-i
                  !shiftRes=subG*(subG+3)/2+1
                  DO j = subG, 0, -1
                     IF (msize_p < ii) EXIT grad_do
                     p(ii) = p(ii)+pRes(shiftRes-j)*xi(i+1)
                     ii = ii+1
                  END DO
                  shiftRes = shiftRes+subG+2
                  subG = subG+1
               END DO
            END DO grad_do
            ii0 = ii0+inSize
            shiftRes_0 = shiftRes_0+newSize
            msize_p = msize_p+inSize
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief evaluates the 2d polynomial at x (the result is a polynomial in one variable)
!> \param p ...
!> \param x ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_p_eval2(p, x, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), INTENT(in)                               :: x
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_p_eval2', routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, my_npoly, newSize, size_p
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1415,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      grad = grad_size2(size_p)
      newSize = SIZE(pRes)/my_npoly
      pRes = 0.0_dp
      IF(.NOT.(newSize >= poly_size1(grad)))CALL cp__a("d3_poly.F",1422)
      ALLOCATE (xi(grad+1))
      CALL poly_p_eval2b(p, SIZE(p), x, pRes, SIZE(pRes), my_npoly, grad, xi)
      DEALLOCATE (xi)
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine of poly_p_eval2 without checks
!> \param p ...
!> \param size_p ...
!> \param x ...
!> \param pRes ...
!> \param size_pRes ...
!> \param npoly ...
!> \param grad ...
!> \param xi ...
! **************************************************************************************************
   SUBROUTINE poly_p_eval2b(p, size_p, x, pRes, size_pRes, npoly, grad, xi)
      INTEGER, INTENT(in)                                :: size_p
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: p
      REAL(dp), INTENT(in)                               :: x
      INTEGER, INTENT(in)                                :: size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: pRes
      INTEGER, INTENT(in)                                :: npoly
      INTEGER                                            :: grad
      REAL(dp), DIMENSION( *)           :: xi

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_p_eval2b', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, igrad, ii, ii0, ij, inSize, ipoly, &
                                                            msize_p, newSize, pShift, shiftRes

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1456,"module d3_poly not initialized")
      inSize = size_p/npoly
      newSize = size_pRes/npoly
      pRes(1:size_pRes) = 0.0_dp
      !CPPreconditionNoFail(newSize>grad,cp_failure_level,routineP)
      xi(1) = 1.0_dp
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      shiftRes = 1
      pShift = 0
      DO ipoly = 1, npoly
         DO ii = 1, MIN(inSize, cached_dim2)
            pRes(shiftRes+a_mono_exp2(2, ii)) = pRes(shiftRes+a_mono_exp2(2, ii))+p(pShift+ii)*xi(a_mono_exp2(1, ii)+1)
         END DO
         shiftRes = shiftRes+newSize
         pShift = pShift+inSize
      END DO
      IF (grad > max_grad2) THEN
         ii0 = (max_grad2+1)*(max_grad2+2)/2+1
         shiftRes = 1
         msize_p = inSize
         DO ipoly = 1, npoly
            ii = ii0
            grad_do2: DO igrad = max_grad2+1, grad
               !ii=igrad*(igrad+1)/2+1
               ij = shiftRes
               DO i = igrad, 0, -1
                  IF (msize_p < ii) EXIT grad_do2
                  ! ij=igrad-i
                  pRes(ij) = pRes(ij)+p(ii)*xi(i+1)
                  ii = ii+1
                  ij = ij+1
               END DO
            END DO grad_do2
            msize_p = msize_p+inSize
            shiftRes = shiftRes+newSize
            ii0 = ii0+inSize
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief unevaluates a 1d polynomial to 2d at x
!>  p(a,b)=sum(pRes(b)*(x*a)^i,i), this is *not* the inverse of poly_p_eval2
!>  overwrites p
!> \param p ...
!> \param x ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_padd_uneval2(p, x, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(inout)              :: p
      REAL(dp), INTENT(in)                               :: x
      REAL(dp), DIMENSION(:), INTENT(in)                 :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_padd_uneval2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, my_npoly, newSize, size_p
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1519,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      newSize = SIZE(pRes)/my_npoly
      grad = grad_size1(newSize)
      IF(.NOT.(size_p >= poly_size2(grad)))CALL cp__a("d3_poly.F",1525)
      IF(.NOT.(newSize == poly_size1(grad)))CALL cp__a("d3_poly.F",1526)
      ALLOCATE (xi(grad+1))
      CALL poly_padd_uneval2b(p, SIZE(p), x, pRes, SIZE(pRes), my_npoly, grad, xi)
      DEALLOCATE (xi)
   END SUBROUTINE

! **************************************************************************************************
!> \brief low level routine of poly_p_uneval2 without checks
!> \param p ...
!> \param size_p ...
!> \param x ...
!> \param pRes ...
!> \param size_pRes ...
!> \param npoly ...
!> \param grad ...
!> \param xi ...
! **************************************************************************************************
   SUBROUTINE poly_padd_uneval2b(p, size_p, x, pRes, size_pRes, npoly, grad, xi)
      INTEGER, INTENT(in)                                :: size_p
      REAL(dp), DIMENSION( *), &
         INTENT(inout)                                   :: p
      REAL(dp), INTENT(in)                               :: x
      INTEGER, INTENT(in)                                :: size_pRes
      REAL(dp), DIMENSION( *), &
         INTENT(in)                                      :: pRes
      INTEGER, INTENT(in)                                :: npoly
      INTEGER                                            :: grad
      REAL(dp), DIMENSION( *)           :: xi

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_padd_uneval2b', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, igrad, ii, ii0, ij, inSize, ipoly, &
                                                            msize_p, newSize, pShift, shiftRes, &
                                                            upSize

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1562,"module d3_poly not initialized")
      inSize = size_p/npoly
      upSize = (grad+1)*(grad+2)/2
      newSize = size_pRes/npoly
      !CPPreconditionNoFail(newSize>grad,cp_failure_level,routineP)
      xi(1) = 1.0_dp
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      shiftRes = 1
      pShift = 0
      DO ipoly = 1, npoly
         DO ii = 1, MIN(upSize, cached_dim2)
            p(pShift+ii) = p(pShift+ii)+pRes(shiftRes+a_mono_exp2(2, ii))*xi(a_mono_exp2(1, ii)+1)
         END DO
         shiftRes = shiftRes+newSize
         pShift = pShift+inSize
      END DO
      IF (grad > max_grad2) THEN
         ii0 = (max_grad2+1)*(max_grad2+2)/2+1
         shiftRes = 1
         msize_p = upSize
         DO ipoly = 1, npoly
            ii = ii0
            grad_do2: DO igrad = max_grad2+1, grad
               !ii=igrad*(igrad+1)/2+1
               ij = shiftRes
               DO i = igrad, 0, -1
                  IF (msize_p < ii) EXIT grad_do2
                  ! ij=igrad-i
                  p(ii) = p(ii)+pRes(ij)*xi(i+1)
                  ii = ii+1
                  ij = ij+1
               END DO
            END DO grad_do2
            msize_p = msize_p+inSize
            shiftRes = shiftRes+newSize
            ii0 = ii0+inSize
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief evaluates the 1d polynomial at the given place, results are stored contiguosly
!> \param p ...
!> \param x ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_eval1(p, x, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), INTENT(in)                               :: x
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_eval1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ipoly, my_npoly, pShift, size_p
      REAL(dp)                                           :: vv, xx

      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      IF(.NOT.(SIZE(pRes) >= my_npoly))CALL cp__a("d3_poly.F",1625)
      pShift = 0
      DO ipoly = 1, my_npoly
         xx = 1.0_dp
         vv = 0.0_dp
         DO i = 1, size_p
            vv = vv+p(pShift+i)*xx
            xx = xx*x
         END DO
         pRes(ipoly) = vv
         pShift = pShift+size_p
      END DO
   END SUBROUTINE

! **************************************************************************************************
!> \brief evaluates the 2d polynomial at the given place, results are stored contiguosly
!> \param p ...
!> \param x ...
!> \param y ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_eval2(p, x, y, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), INTENT(in)                               :: x, y
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_eval2', routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, i, igrad, ii, ipoly, j, msize_p, &
                                                            my_npoly, pShift, size_p
      REAL(dp)                                           :: v
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi, yi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1660,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      grad = grad_size2(size_p)
      IF(.NOT.(SIZE(pRes) >= my_npoly))CALL cp__a("d3_poly.F",1665)
      ALLOCATE (xi(grad+1), yi(grad+1))
      xi(1) = 1.0_dp
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      yi(1) = 1.0_dp
      DO i = 1, grad
         yi(i+1) = yi(i)*y
      END DO
      pShift = 0
      DO ipoly = 1, my_npoly
         v = 0.0_dp
         DO ii = 1, MIN(size_p, cached_dim2)
            v = v+p(pShift+ii)*xi(a_mono_exp2(1, ii)+1)*yi(a_mono_exp2(2, ii)+1)
         END DO
         pRes(ipoly) = v
         pShift = pShift+size_p
      END DO
      IF (grad > max_grad2) THEN
         pShift = (max_grad2+1)*(max_grad2+2)/2+1
         msize_p = size_p
         DO ipoly = 1, my_npoly
            ii = pShift
            v = 0.0_dp
            grad_do4: DO igrad = max_grad2+1, grad
               ! ii=igrad*(igrad+1)*(igrad+2)/6+1
               j = 1
               DO i = igrad, 0, -1
                  IF (msize_p < ii) EXIT grad_do4
                  v = v+p(ii)*xi(i+1)*yi(j)
                  j = j+1
                  ii = ii+1
               END DO
            END DO grad_do4
            pRes(ipoly) = pRes(ipoly)+v
            pShift = pShift+size_p
            msize_p = msize_p+size_p
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief evaluates the 3d polynomial at the given place, results are stored contiguosly
!> \param p ...
!> \param x ...
!> \param y ...
!> \param z ...
!> \param pRes ...
!> \param npoly ...
! **************************************************************************************************
   SUBROUTINE poly_eval3(p, x, y, z, pRes, npoly)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), INTENT(in)                               :: x, y, z
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_eval3', routineP = moduleN//':'//routineN

      INTEGER                                            :: grad, i, igrad, ii, ipoly, j, k, &
                                                            msize_p, my_npoly, pShift, size_p
      REAL(dp)                                           :: v
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xi, yi, zi

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1729,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      size_p = SIZE(p)/my_npoly
      grad = grad_size3(size_p)
      IF(.NOT.(SIZE(pRes) >= my_npoly))CALL cp__a("d3_poly.F",1734)
      ALLOCATE (xi(grad+1), yi(grad+1), zi(grad+1))
      xi(1) = 1.0_dp
      DO i = 1, grad
         xi(i+1) = xi(i)*x
      END DO
      yi(1) = 1.0_dp
      DO i = 1, grad
         yi(i+1) = yi(i)*y
      END DO
      zi(1) = 1.0_dp
      DO i = 1, grad
         zi(i+1) = zi(i)*z
      END DO
      pShift = 0
      DO ipoly = 1, my_npoly
         v = 0.0_dp
         DO ii = 1, MIN(size_p, cached_dim3)
            v = v+p(pShift+ii)*xi(a_mono_exp3(1, ii)+1)*yi(a_mono_exp3(2, ii)+1) &
                *zi(a_mono_exp3(3, ii)+1)
         END DO
         pRes(ipoly) = v
         pShift = pShift+size_p
      END DO
      IF (grad > max_grad3) THEN
         pShift = (max_grad3+1)*(max_grad3+2)*(max_grad3+3)/6+1
         msize_p = size_p
         DO ipoly = 1, my_npoly
            ii = pShift
            v = 0.0_dp
            grad_do3: DO igrad = max_grad3+1, grad
               ! ii=igrad*(igrad+1)*(igrad+2)/6+1
               DO i = igrad, 0, -1
                  k = 1
                  DO j = igrad-i, 0, -1
                     ii = (ipoly-1)*size_p+mono_index3(i, j, igrad-i-j)+1
                     IF (msize_p < ii) EXIT grad_do3
                     v = v+p(ii)*xi(i+1)*yi(j+1)*zi(k)
                     k = k+1
                     ii = ii+1
                  END DO
               END DO
            END DO grad_do3
            pRes(ipoly) = pRes(ipoly)+v
            pShift = pShift+size_p
            msize_p = msize_p+size_p
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief returns an array with all dp/dx, the all dp/dy, and finally all dp/dz
!> \param p ...
!> \param pRes ...
!> \param npoly ...
!> \param sumUp ...
! **************************************************************************************************
   SUBROUTINE poly_derive3(p, pRes, npoly, sumUp)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: p
      REAL(dp), DIMENSION(:), INTENT(inout)              :: pRes
      INTEGER, INTENT(in), OPTIONAL                      :: npoly
      LOGICAL, INTENT(in), OPTIONAL                      :: sumUp

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_derive3', routineP = moduleN//':'//routineN

      INTEGER :: grad, i, igrad, ii, ii2, ipoly, j, k, msize_p, my_npoly, newSize, pShift, size_p, &
         xDerivShift, yDerivShift, yShift, zDerivShift, zShift
      LOGICAL                                            :: my_sumUp

      IF (.NOT. module_initialized) CALL cp__b("d3_poly.F",1803,"module d3_poly not initialized")
      my_npoly = 1
      IF (PRESENT(npoly)) my_npoly = npoly
      my_sumUp = .FALSE.
      IF (PRESENT(sumUp)) my_sumUp = sumUp
      size_p = SIZE(p)/my_npoly
      newSize = SIZE(pRes)/(3*my_npoly)
      grad = grad_size3(size_p)
      IF(.NOT.(newSize >= poly_size3(grad)))CALL cp__a("d3_poly.F",1811)
      IF (.NOT. my_sumUp) pRes = 0
      xDerivShift = 1
      yDerivShift = my_npoly*newSize+1
      zDerivShift = 2*yDerivShift-1
      pShift = 0
      DO ipoly = 1, my_npoly
         ! split derivs to have less streams to follow (3 vs 5)?
         DO ii = 1, MIN(cached_dim3, size_p)
            pRes(xDerivShift+a_deriv_idx3(1, ii)) = pRes(xDerivShift+a_deriv_idx3(1, ii)) &
                                                    +p(pShift+ii)*a_mono_exp3(1, ii)
            pRes(yDerivShift+a_deriv_idx3(2, ii)) = pRes(yDerivShift+a_deriv_idx3(2, ii)) &
                                                    +p(pShift+ii)*a_mono_exp3(2, ii)
            pRes(zDerivShift+a_deriv_idx3(3, ii)) = pRes(zDerivShift+a_deriv_idx3(3, ii)) &
                                                    +p(pShift+ii)*a_mono_exp3(3, ii)
         END DO
         xDerivShift = xDerivShift+newSize
         yDerivShift = yDerivShift+newSize
         zDerivShift = zDerivShift+newSize
         pShift = pShift+size_p
      END DO
      IF (grad > max_grad3) THEN
         xDerivShift = 0
         yDerivShift = my_npoly*newSize
         zDerivShift = 2*yDerivShift-1
         msize_p = size_p
         xDerivShift = max_grad3*(max_grad3+1)*(max_grad3+2)/6+1
         pShift = xDerivShift+(max_grad3+1)*(max_grad3+2)/2
         DO ipoly = 1, my_npoly
            ii = pShift
            ii2 = xDerivShift
            grad_do5: DO igrad = max_grad3+1, grad
               yShift = yDerivShift
               zShift = zDerivShift
               DO i = igrad, 0, -1
                  k = 0
                  DO j = igrad-i, 0, -1
                     IF (ii > msize_p) EXIT grad_do5
                     ! remove ifs?
                     IF (i > 0) pRes(ii2) = pRes(ii2)+p(ii)*i
                     IF (j > 0) pRes(yShift+ii2) = pRes(yShift+ii2)+p(ii)*j
                     IF (k > 0) pRes(zShift+ii2) = pRes(zShift+ii2)+k*p(ii)
                     ii = ii+1
                     ii2 = ii2+1
                     k = k+1
                  END DO
                  yShift = yShift-1
                  zShift = zShift-1
               END DO
               ii2 = ii2-igrad-1
            END DO grad_do5
            pShift = pShift+size_p
            xDerivShift = xDerivShift+newSize
            msize_p = msize_p+size_p
         END DO
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief subroutine that converts from the cp2k poly format to the d3 poly format
!> \param poly_cp2k ...
!> \param grad ...
!> \param poly_d3 ...
! **************************************************************************************************
   SUBROUTINE poly_cp2k2d3(poly_cp2k, grad, poly_d3)
      REAL(dp), DIMENSION(:), INTENT(in)                 :: poly_cp2k
      INTEGER, INTENT(in)                                :: grad
      REAL(dp), DIMENSION(:), INTENT(out)                :: poly_d3

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_cp2k2d3', routineP = moduleN//':'//routineN

      INTEGER                                            :: cp_ii, i, j, k, mgrad, mgrad2, sgrad, &
                                                            sgrad2, sgrad2k, sgrad3, sgrad3k, &
                                                            shifti, shiftj, shiftk, size_p

      size_p = (grad+1)*(grad+2)*(grad+3)/6
      IF(.NOT.(SIZE(poly_cp2k) >= size_p))CALL cp__a("d3_poly.F",1887)
      IF(.NOT.(SIZE(poly_d3) >= size_p))CALL cp__a("d3_poly.F",1888)
      cp_ii = 0
      sgrad2k = 0
      sgrad3k = 0
      DO k = 0, grad
         shiftk = k+1
         sgrad2k = sgrad2k+k
         sgrad3k = sgrad3k+sgrad2k
         sgrad2 = sgrad2k
         sgrad3 = sgrad3k
         DO j = 0, grad-k
            sgrad = j+k
            mgrad2 = sgrad2
            shiftj = mgrad2+shiftk
            mgrad = sgrad
            shifti = shiftj+sgrad3
            DO i = 0, grad-j-k
               cp_ii = cp_ii+1
               poly_d3(shifti) = poly_cp2k(cp_ii)
               mgrad = mgrad+1
               mgrad2 = mgrad2+mgrad
               shifti = shifti+mgrad2
            END DO
            sgrad2 = sgrad2+sgrad+1
            sgrad3 = sgrad3+sgrad2
         END DO
      END DO
      IF (SIZE(poly_d3) >= size_p) THEN
         poly_d3(size_p+1:) = 0.0_dp
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief subroutine that converts from the d3 poly format to the cp2k poly format
!> \param poly_cp2k ...
!> \param grad ...
!> \param poly_d3 ...
! **************************************************************************************************
   SUBROUTINE poly_d32cp2k(poly_cp2k, grad, poly_d3)
      REAL(dp), DIMENSION(:), INTENT(out)                :: poly_cp2k
      INTEGER, INTENT(in)                                :: grad
      REAL(dp), DIMENSION(:), INTENT(in)                 :: poly_d3

      CHARACTER(len=*), PARAMETER :: routineN = 'poly_d32cp2k', routineP = moduleN//':'//routineN

      INTEGER                                            :: cp_ii, i, j, k, mgrad, mgrad2, sgrad, &
                                                            sgrad2, sgrad2k, sgrad3, sgrad3k, &
                                                            shifti, shiftj, shiftk, size_p

      size_p = (grad+1)*(grad+2)*(grad+3)/6
      IF(.NOT.(SIZE(poly_cp2k) >= size_p))CALL cp__a("d3_poly.F",1938)
      IF(.NOT.(SIZE(poly_d3) >= size_p))CALL cp__a("d3_poly.F",1939)
      cp_ii = 0
      sgrad2k = 0
      sgrad3k = 0
      DO k = 0, grad
         shiftk = k+1
         sgrad2k = sgrad2k+k
         sgrad3k = sgrad3k+sgrad2k
         sgrad2 = sgrad2k
         sgrad3 = sgrad3k
         DO j = 0, grad-k
            sgrad = j+k
            mgrad2 = sgrad2
            shiftj = mgrad2+shiftk
            mgrad = sgrad
            shifti = shiftj+sgrad3
            DO i = 0, grad-j-k
               cp_ii = cp_ii+1
               poly_cp2k(cp_ii) = poly_d3(shifti)
               mgrad = mgrad+1
               mgrad2 = mgrad2+mgrad
               shifti = shifti+mgrad2
            END DO
            sgrad2 = sgrad2+sgrad+1
            sgrad3 = sgrad3+sgrad2
         END DO
      END DO
      IF (SIZE(poly_d3) >= size_p) THEN
         poly_cp2k(size_p+1:) = 0.0_dp
      END IF
   END SUBROUTINE

END MODULE d3_poly
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!>   \brief
!>     Routines to efficently collocate and integrate gaussians on a grid
!>     These use most of Joost's tricks and a couple more...
!>     result is *speed* and genericity
!>   \author Fawzi Mohamed, 2007
!>   \notes original available with BSD style license
! **************************************************************************************************
MODULE gauss_colloc
   USE d3_poly,                         ONLY: &
        grad_size3, poly_affine_t3, poly_affine_t3t, poly_p_eval2b, poly_p_eval3b, &
        poly_padd_uneval2b, poly_padd_uneval3b, poly_size1, poly_size2, poly_size3
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE lgrid_types,                     ONLY: lgrid_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: collocGauss, &
             integrateGaussFull


   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gauss_colloc'

   REAL(dp), PARAMETER :: small = TINY(1.0_dp)

   ! keep prettify happy (does not see the include)
   INTEGER(KIND=int_8), PARAMETER           :: unused_import_of_int_8 = 1

CONTAINS

! **************************************************************************************************
!> \brief collocate a periodically repeated gaussian on a non orthormbic grid
!>
!> this routine has been tested and works well with cells with
!> det(h)/sqrt(tr(dot(h^T,h)))>0.2 (2 angles bigger than 24 deg or one angle
!> bigger than 11 deg).
!> Because of its numerics it might fail badly (infinity or NaN) with
!> with more deformed cells. Avoiding this would be bossible only using
!> IEEE numerics controls, which would also make everything slower and
!> less supported.
!> Still the actual numeric has been carefully tuned, and in normal cases
!> and most anormal it should work.
!> With det(h)/sqrt(tr(dot(h^T,h)))>0.2 I could not find any failure.
!>
!> \param h cell matrix
!> \param h_inv inverse of the cell matrix
!> \param grid the grid
!> \param poly polynomial (d3_poly format)
!> \param alphai exponential coeff
!> \param posi position of the gaussian
!> \param max_r2 maximum radius of collocation squared
!> \param periodic array of 0 or 1 that says which dimensions have pbc (1=pbc)
!> \param gdim dimension of the grid (grid might be a subset)
!> \param local_bounds local bounds of the grid piece that is kept locally
!>   (i.e. of grid) the global grid is assumed to atart at 0,0,0
!> \param local_shift start indexes of the local slice (i.e. of grid)
!> \param poly_shift position of posi in the polynomial reference system.
!>  Set it to posi to use the global reference system.
!> \param scale a global scale factor
!> \param lgrid ...
! **************************************************************************************************
   SUBROUTINE collocGauss(h, h_inv, grid, poly, alphai, posi, max_r2, &
                          periodic, gdim, local_bounds, local_shift, poly_shift, scale, lgrid)
      REAL(dp), DIMENSION(0:2, 0:2), INTENT(in)          :: h, h_inv
      REAL(dp), DIMENSION(0:, 0:, 0:), INTENT(inout)     :: grid
      REAL(dp), DIMENSION(:), INTENT(inout)              :: poly
      REAL(dp), INTENT(in)                               :: alphai
      REAL(dp), DIMENSION(0:2), INTENT(in)               :: posi
      REAL(dp), INTENT(in)                               :: max_r2
      INTEGER, DIMENSION(0:2), INTENT(in)                :: periodic
      INTEGER, DIMENSION(0:2), INTENT(in), OPTIONAL      :: gdim
      INTEGER, DIMENSION(2, 0:2), INTENT(in), OPTIONAL   :: local_bounds
      INTEGER, DIMENSION(0:2), INTENT(in), OPTIONAL      :: local_shift
      REAL(dp), DIMENSION(0:2), INTENT(in), OPTIONAL     :: poly_shift
      REAL(dp), INTENT(in), OPTIONAL                     :: scale
      TYPE(lgrid_type), INTENT(inout), OPTIONAL          :: lgrid

      CHARACTER(len=*), PARAMETER :: routineN = 'collocGauss', routineP = moduleN//':'//routineN

      CALL colloc_int_body(h, h_inv, grid, poly, alphai, posi, max_r2, &
                           periodic, gdim, local_bounds, local_shift, &
                           poly_shift, scale, lgrid, integrate=.FALSE.)

   END SUBROUTINE

! **************************************************************************************************
!> \brief integrates a gaussian times any polynomial up to a give order.
!>
!> Most things are the same as for collocGauss (see its comments).
!> Returns the integrals of all the monomials in d3 format into poly
!> \param h ...
!> \param h_inv ...
!> \param grid ...
!> \param poly ...
!> \param alphai ...
!> \param posi ...
!> \param max_r2 ...
!> \param periodic ...
!> \param gdim ...
!> \param local_bounds ...
!> \param local_shift ...
!> \param poly_shift ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE integrateGaussFull(h, h_inv, grid, poly, alphai, posi, max_r2, &
                                 periodic, gdim, local_bounds, local_shift, poly_shift, scale)
      REAL(dp), DIMENSION(0:2, 0:2), INTENT(in)          :: h, h_inv
      REAL(dp), DIMENSION(0:, 0:, 0:), INTENT(inout)     :: grid
      REAL(dp), DIMENSION(:), INTENT(inout)              :: poly
      REAL(dp), INTENT(in)                               :: alphai
      REAL(dp), DIMENSION(0:2), INTENT(in)               :: posi
      REAL(dp), INTENT(in)                               :: max_r2
      INTEGER, DIMENSION(0:2), INTENT(in)                :: periodic
      INTEGER, DIMENSION(0:2), INTENT(in), OPTIONAL      :: gdim
      INTEGER, DIMENSION(2, 0:2), INTENT(in), OPTIONAL   :: local_bounds
      INTEGER, DIMENSION(0:2), INTENT(in), OPTIONAL      :: local_shift
      REAL(dp), DIMENSION(0:2), INTENT(in), OPTIONAL     :: poly_shift
      REAL(dp), INTENT(in), OPTIONAL                     :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'integrateGaussFull', &
         routineP = moduleN//':'//routineN

      CALL colloc_int_body(h, h_inv, grid, poly, alphai, posi, max_r2, &
                           periodic, gdim, local_bounds, local_shift, &
                           poly_shift, scale, integrate=.TRUE.)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Common code for collocGauss() and integrateGaussFull()
!> \param h ...
!> \param h_inv ...
!> \param grid ...
!> \param poly ...
!> \param alphai ...
!> \param posi ...
!> \param max_r2 ...
!> \param periodic ...
!> \param gdim ...
!> \param local_bounds ...
!> \param local_shift ...
!> \param poly_shift ...
!> \param scale ...
!> \param lgrid ...
!> \param integrate ...
! **************************************************************************************************
   SUBROUTINE colloc_int_body(h, h_inv, grid, poly, alphai, posi, max_r2, &
                              periodic, gdim, local_bounds, local_shift, poly_shift, scale, lgrid, integrate)
      REAL(dp), DIMENSION(0:2, 0:2), &
         INTENT(in)                             :: h, h_inv
      REAL(dp), DIMENSION(0:, 0:, 0:), &
         INTENT(inout)                          :: grid
      REAL(dp), DIMENSION(:), INTENT(inout)    :: poly
      REAL(dp), INTENT(in)                     :: alphai
      REAL(dp), DIMENSION(0:2), INTENT(in)     :: posi
      REAL(dp), INTENT(in)                     :: max_r2
      INTEGER, DIMENSION(0:2), INTENT(in)      :: periodic
      INTEGER, DIMENSION(0:2), INTENT(in), &
         OPTIONAL                               :: gdim
      INTEGER, DIMENSION(2, 0:2), INTENT(in), &
         OPTIONAL                               :: local_bounds
      INTEGER, DIMENSION(0:2), INTENT(in), &
         OPTIONAL                               :: local_shift
      REAL(dp), DIMENSION(0:2), INTENT(in), &
         OPTIONAL                               :: poly_shift
      REAL(dp), INTENT(in), OPTIONAL           :: scale
      TYPE(lgrid_type), INTENT(inout), &
         OPTIONAL                               :: lgrid
      LOGICAL                                  :: integrate

      CHARACTER(len=*), PARAMETER :: routineN = 'colloc_int_body', &
                                     routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(0:2), PARAMETER       :: permut = (/2, 1, 0/)
      INTEGER :: grad, i, i0, ii, iiShift, iiShift2, iistart, iistart2, ij, &
                 ijShift, iJump, ik, ikShift, ikShift2, ikstart, ikstart2, iend, iend2, &
                 imax, imax1, imin, imin1, istart, istart2, j, jend, jJump, jmax, jmax1, jmin, &
                 jmin1, jstart, k, kend, kend2, kgrad, kJump, kmax, kmax1, kmin, kmin1, &
                 kstart, kstart2, max_j, size_jk, size_k, size_ijk, ig, ithread, nthread
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :) :: k_bounds
      INTEGER, DIMENSION(0:2)                  :: cellShift, l_shift, l_ub, &
                                                  ndim, period, shiftPos, ldim2
      INTEGER, DIMENSION(2, 0:2)               :: l_bounds
      LOGICAL                                  :: has_overlap
      REAL(dp) :: cci0, cci1, cci2, ccj0, ccj0_i0, ccj0_i1, ccj0_i2, ccj1, &
                  ccj1_i0, ccj1_i1, ccj2, cck0, cck0_0, cck0_0_p, cck0_i, cck0_i2, &
                  cck0_ij, cck0_j, cck0_j2, cck0_j_p, cck1, cck1_0, cck1_0_p, cck1_i, &
                  cck1_j, cck2, delta_i, delta_j, delta_k, g_scale, i_coeffn_i, icoeff0, &
                  ii_coeff0, ii_coeff2, ii_coeff2_jump, ii_coeffn, ii_coeffn_jump, &
                  ij_coeff0, ij_coeff0_jump, ij_coeff1, ik_coeff0, ik_coeff0_jump, &
                  ik_coeff1, j_coeffn_i, j_coeffn_j, jcoeff0, jj_coeff0, jj_coeff2, &
                  jj_coeffn, jk_coeff0, jk_coeff1, k_coeffn_i, k_coeffn_j, k_coeffn_k, &
                  kcoeff0, kk_coeff0, kk_coeff2, kk_coeffn, m(0:2, 0:2), maxr2, p_kk, &
                  r_0
      REAL(dp) :: res_0, res_i, res_j, res_k, scaled_h(0:2, 0:2), sqDi, sqDj, &
                  sqDk
      REAL(dp), ALLOCATABLE, DIMENSION(:)      :: poly_ijk, poly_jk, xi
      REAL(dp), DIMENSION(0:2)                 :: l, normD, p_shift, resPos, &
                                                  resPosReal, riPos, rpos, wrPos

      REAL(dp) :: det, gval
      REAL(dp), ALLOCATABLE, DIMENSION(:)      :: k_vals
      INTEGER, PARAMETER :: npoly = 1
      REAL(dp), ALLOCATABLE, DIMENSION(:)      :: poly_k
      REAL(dp) :: p_v


      ig = 0
      ithread = 0
!$    ithread = omp_get_thread_num()

      nthread = 1
!$    nthread = omp_get_num_threads()

      IF (integrate) THEN
         poly = 0.0_dp
      ELSE
         IF (ALL(poly == 0.0_dp)) RETURN
      ENDIF

      IF (PRESENT(poly_shift)) THEN
         p_shift = poly_shift
      ELSE
         p_shift = 0.0_dp
      END IF

      ldim2(permut(0)) = SIZE(grid, 1)
      ldim2(permut(1)) = SIZE(grid, 2)
      ldim2(permut(2)) = SIZE(grid, 3)

      IF (PRESENT(gdim)) THEN
         DO i = 0, 2
            ndim(permut(i)) = gdim(i)
         END DO
      ELSE
         ndim = ldim2
      END IF
      g_scale = 1.0_dp
      IF (PRESENT(scale)) g_scale = scale

      IF (integrate) THEN
         det = (h(0, 0)*(h(1, 1)*h(2, 2)-h(1, 2)*h(2, 1)) &
                -h(1, 0)*(h(0, 1)*h(2, 2)-h(0, 2)*h(2, 1)) &
                +h(2, 0)*(h(0, 1)*h(1, 2)-h(0, 2)*h(1, 1)))
         g_scale = g_scale*ABS(det)/REAL(INT(ndim(0), KIND=int_8)*INT(ndim(1), KIND=int_8)*INT(ndim(2), KIND=int_8), dp)
      ENDIF

      IF (PRESENT(local_bounds)) THEN
         DO i = 0, 2
            l_bounds(:, permut(i)) = local_bounds(:, i)
         END DO
      ELSE
         l_bounds(1, :) = 0
         l_bounds(2, :) = ndim-1
      END IF
      IF (PRESENT(local_shift)) THEN
         DO i = 0, 2
            l_shift(permut(i)) = local_shift(i)
         END DO
      ELSE
         l_shift = 0 ! l_bounds(1,:)
      END IF
      l_ub = l_bounds(2, :)-l_bounds(1, :)+l_shift
      DO i = 0, 2
         IF(.NOT.(l_ub(i) < ldim2(i)))CALL cp__a("gauss_colloc.F",281)
      END DO

      DO i = 0, 2
         period(permut(i)) = periodic(i)
      END DO
      IF(.NOT.(ALL(l_bounds(2, :) < ndim .OR. period(:) == 1)))CALL cp__a("gauss_colloc.F",287)
      IF(.NOT.(ALL(l_bounds(1, :) >= 0 .OR. period(:) == 1)))CALL cp__a("gauss_colloc.F",288)
      IF(.NOT.(ALL(l_bounds(2, :)-l_bounds(1, :) < ndim)))CALL cp__a("gauss_colloc.F",289)
      rPos = 0.0_dp
      DO j = 0, 2
         DO i = 0, 2
            rPos(permut(i)) = rPos(permut(i))+h_inv(i, j)*posi(j)
         END DO
      END DO
      cellShift = FLOOR(rPos)*period
      wrPos = rPos-REAL(cellShift, dp)
      riPos = wrPos*ndim
      shiftPos = FLOOR(riPos+0.5_dp)
      resPos = riPos-shiftPos
      normD = 1.0_dp/REAL(ndim, dp)
      scaled_h = 0.0_dp
      DO j = 0, 2
         DO i = 0, 2
            scaled_h(i, permut(j)) = h(i, j)*normD(permut(j))
         END DO
      END DO
      resPosReal = 0.0_dp
      DO j = 0, 2
         DO i = 0, 2
            resPosReal(i) = resPosReal(i)+h(i, j)*(wrPos(permut(j))-normD(permut(j))*REAL(shiftPos(permut(j)), dp))
         END DO
      END DO

      !maxr2=0.0_dp
      !DO j=0,2
      !    DO i=0,2
      !        ! guarantee at least the nearest points (this increases the sphere, increase just the box?)
      !        maxr2=maxr2+(scaled_h(i,j))**2
      !    END DO
      !END DO
      maxr2 = max_r2 !MAX(max_r2,maxr2)

      ! build up quadratic form (ellipsoid)
      m = 0.0_dp
      DO j = 0, 2
         DO i = 0, 2
            DO k = 0, 2
               m(i, j) = m(i, j)+scaled_h(k, i)*scaled_h(k, j)
            END DO
         END DO
      END DO

      l = 0.0_dp
      DO j = 0, 2
         DO i = 0, 2
            l(j) = l(j)-2.0*resPos(i)*m(i, j)
         END DO
      END DO

      r_0 = 0.0_dp
      DO i = 0, 2
         r_0 = r_0-0.5*resPos(i)*l(i)
      END DO

      ! calc i boundaries
      cci2 = (m(2, 2)*m(0, 0)*m(1, 1)-m(2, 2)*m(0, 1)**2-m(1, 1)*m(0, 2)**2 &
              +2.0_dp*m(0, 2)*m(0, 1)*m(1, 2)-m(0, 0)*m(1, 2)**2)/(m(2, 2)*m(1, 1)-m(1, 2)**2)
      cci1 = -(-m(2, 2)*l(0)*m(1, 1)+m(2, 2)*m(0, 1)*l(1)+l(2)*m(0, 2)*m(1, 1) &
               +l(0)*m(1, 2)**2-l(2)*m(0, 1)*m(1, 2)-m(1, 2)*l(1)*m(0, 2))/(m(2, 2)*m(1, 1)-m(1, 2)**2)
      cci0 = -((-4.0_dp*m(2, 2)*r_0*m(1, 1)+m(2, 2)*l(1)**2+l(2)**2*m(1, 1) &
                -2.0_dp*l(1)*m(1, 2)*l(2)+4.0_dp*r_0*m(1, 2)**2) &
               /(m(2, 2)*m(1, 1)-m(1, 2)**2))/4.0_dp-maxr2
      delta_i = cci1*cci1-4.0_dp*cci2*cci0
      IF (delta_i <= 0) RETURN
      sqDi = SQRT(delta_i)
      imin = CEILING((-cci1-sqDi)/(2.0_dp*cci2))
      imax = FLOOR((-cci1+sqDi)/(2.0_dp*cci2))
      !! box early return

      IF (period(0) == 1) THEN
         has_overlap = imax-imin+1 > ndim(0) .OR. (l_bounds(1, 0) == 0 .AND. l_bounds(2, 0) == ndim(0)-1)
         IF (.NOT. has_overlap) THEN
            imin1 = MODULO(imin+shiftPos(0), ndim(0))
            imax1 = imin1+imax-imin+1
            IF (imin1 < l_bounds(1, 0)) THEN
               has_overlap = imax1 >= l_bounds(1, 0)
            ELSE
               has_overlap = imin1 <= l_bounds(2, 0) .OR. (imax1 >= ndim(0) .AND. l_bounds(1, 0) <= imax1+ndim(0))
            END IF
            IF (.NOT. has_overlap) RETURN
         END IF
      ELSE
         IF (imax+shiftPos(0) < l_bounds(1, 0) .OR. imin+shiftPos(0) > l_bounds(2, 0)) RETURN
      END IF

      ! j box bounds
      has_overlap = l_bounds(1, 1) == 0 .AND. l_bounds(2, 1) == ndim(1)-1
      IF (.NOT. has_overlap) THEN
         ccj2 = (m(0, 0)*m(2, 2)*m(1, 1)-m(0, 0)*m(1, 2)**2-m(0, 1)**2*m(2, 2) &
                 +2.0_dp*m(0, 1)*m(0, 2)*m(1, 2)-m(1, 1)*m(0, 2)**2) &
                /(m(0, 0)*m(2, 2)-m(0, 2)**2)
         ccj1 = -(-m(0, 0)*l(1)*m(2, 2)+m(0, 0)*l(2)*m(1, 2)+l(0)*m(0, 1)*m(2, 2) &
                  -m(0, 2)*l(2)*m(0, 1)-l(0)*m(0, 2)*m(1, 2)+l(1)*m(0, 2)**2) &
                /(m(0, 0)*m(2, 2)-m(0, 2)**2)
         ccj0 = (4.0_dp*m(0, 0)*m(2, 2)*r_0-m(0, 0)*l(2)**2-m(2, 2)*l(0)**2 &
                 +2.0_dp*m(0, 2)*l(2)*l(0)-4.0_dp*r_0*m(0, 2)**2) &
                /(m(0, 0)*m(2, 2)-m(0, 2)**2)/4.0_dp-maxr2
         delta_j = ccj1*ccj1-4.0_dp*ccj2*ccj0
         IF (delta_j <= 0) RETURN
         sqDj = SQRT(delta_j)
         jmin = CEILING((-ccj1-sqDj)/(2.0_dp*ccj2))
         jmax = FLOOR((-ccj1+sqDj)/(2.0_dp*ccj2))
         IF (period(1) == 1) THEN
            IF (jmax-jmin+1 < ndim(1)) THEN
               jmin1 = MODULO(jmin+shiftPos(1), ndim(1))
               jmax1 = jmin1+jmax-jmin+1
               IF (jmin1 < l_bounds(1, 1)) THEN
                  has_overlap = jmax1 >= l_bounds(1, 1)
               ELSE
                  has_overlap = jmin1 <= l_bounds(2, 1) .OR. (jmax1 >= ndim(1) .AND. (l_bounds(1, 1) <= jmax1-ndim(1)))
               END IF
               IF (.NOT. has_overlap) RETURN
            END IF
         ELSE
            IF (jmax+shiftPos(1) < l_bounds(1, 1) .OR. jmin+shiftPos(1) > l_bounds(2, 1)) RETURN
         END IF
      END IF

      ! k box bounds
      has_overlap = l_bounds(1, 2) == 0 .AND. l_bounds(2, 2) == ndim(2)-1
      IF (.NOT. has_overlap) THEN
         cck2 = (m(0, 0)*m(2, 2)*m(1, 1)-m(0, 0)*m(1, 2)**2-m(0, 1)**2*m(2, 2) &
                 +2.0_dp*m(0, 1)*m(0, 2)*m(1, 2)-m(1, 1)*m(0, 2)**2)/(m(0, 0)*m(1, 1)-m(0, 1)**2)
         cck1 = (m(0, 0)*l(2)*m(1, 1)-m(0, 0)*m(1, 2)*l(1)-m(0, 2)*l(0)*m(1, 1) &
                 +l(0)*m(0, 1)*m(1, 2)-l(2)*m(0, 1)**2+m(0, 1)*l(1)*m(0, 2))/(m(0, 0)*m(1, 1)-m(0, 1)**2)
         cck0 = (4.0_dp*m(0, 0)*m(1, 1)*r_0-m(0, 0)*l(1)**2-m(1, 1)*l(0)**2 &
                 +2.0_dp*m(0, 1)*l(1)*l(0)-4.0_dp*r_0*m(0, 1)**2) &
                /(m(0, 0)*m(1, 1)-m(0, 1)**2)/4.0_dp-maxr2
         delta_k = cck1*cck1-4.0_dp*cck2*cck0
         IF (delta_k <= 0) RETURN
         sqDk = SQRT(delta_k)
         kmin = CEILING((-cck1-sqDk)/(2.0_dp*cck2))
         kmax = FLOOR((-cck1+sqDk)/(2.0_dp*cck2))

         IF (period(2) == 1) THEN
            IF (kmax-kmin+1 < ndim(2)) THEN
               kmin1 = MODULO(kmin+shiftPos(2), ndim(2))
               kmax1 = kmin1+kmax-kmin+1
               IF (kmin1 < l_bounds(1, 2)) THEN
                  has_overlap = kmax1 >= l_bounds(1, 2)
               ELSE
                  has_overlap = kmin1 <= l_bounds(2, 2) .OR. &
                                (kmax1 >= ndim(2) .AND. (l_bounds(1, 2) <= MODULO(kmax1, ndim(2))))
               END IF
               IF (.NOT. has_overlap) RETURN
            END IF
         ELSE
            IF (kmax+shiftPos(2) < l_bounds(1, 2) .OR. kmin+shiftPos(2) > l_bounds(2, 2)) RETURN
         END IF
      END IF

      ! k bounds (cache a la cube_info, or inversely integrate in the collocate loop?)
      ccj2 = (m(2, 2)*m(1, 1)-m(1, 2)**2)/m(2, 2)
      ccj1_i1 = (2*m(2, 2)*m(0, 1)-2*m(0, 2)*m(1, 2))/m(2, 2)
      ccj1_i0 = (-l(2)*m(1, 2)+m(2, 2)*l(1))/m(2, 2)
      ccj0_i2 = (m(2, 2)*m(0, 0)-m(0, 2)**2)/m(2, 2)
      ccj0_i1 = (m(2, 2)*l(0)-m(0, 2)*l(2))/m(2, 2)
      ccj0_i0 = (m(2, 2)*r_0-0.25*l(2)**2)/m(2, 2)-maxr2
      cck2 = m(2, 2)
      cck1_i = 2*m(0, 2)
      cck1_j = 2*m(1, 2)
      cck1_0 = l(2)
      cck0_i2 = m(0, 0)
      cck0_ij = 2*m(0, 1)
      cck0_i = l(0)
      cck0_j2 = m(1, 1)
      cck0_j = l(1)
      cck0_0 = r_0-maxr2

      ! find maximum number of j
      max_j = 0
      DO i0 = 0, 1
         i = (imin+imax)/2+i0
         ccj1 = ccj1_i1*i+ccj1_i0
         ccj0 = (ccj0_i2*i+ccj0_i1)*i+ccj0_i0
         delta_j = ccj1*ccj1-4*ccj2*ccj0
         IF (delta_j >= 0) THEN
            sqDj = SQRT(delta_j)
            max_j = MAX(max_j, FLOOR((-ccj1+sqDj)/(2.0_dp*ccj2)) &
                        -CEILING((-ccj1-sqDj)/(2.0_dp*ccj2))+1)
         END IF
      END DO
      max_j = max_j+1 ! just to be sure...
      IF (period(1) == 0) max_j = MIN(max_j, l_bounds(2, 1)-l_bounds(1, 1)+1)

      IF (period(0) == 0) THEN
         imin = MAX(l_bounds(1, 0)-shiftPos(0), imin)
         imax = MIN(l_bounds(2, 0)-shiftPos(0), imax)
      END IF

      ! k bounds (cache a la cube_info?)
      has_overlap = .FALSE.
      ALLOCATE (k_bounds(0:1, 0:max_j-1, 0:MAX(0, imax-imin+1)))
      ! k_bounds=0
      istart = imin
      iiShift = shiftPos(0)-l_bounds(2, 0)+istart
      IF (iiShift > 0) iiShift = iiShift+ndim(0)-1
      iiShift = (iiShift/ndim(0))*ndim(0)-shiftPos(0)
      !iiShift=CEILING(REAL(shiftPos(0)+istart-l_bounds(2,0))/REAL(ndim(0)))*ndim(0)-shiftPos(0))
      istart = MAX(iiShift+l_bounds(1, 0), istart)
      iend = MIN(iiShift+l_bounds(2, 0), imax)
      iJump = ndim(0)-l_bounds(2, 0)+l_bounds(1, 0)-1
      jJump = ndim(1)-l_bounds(2, 1)+l_bounds(1, 1)-1
      DO
         DO i = istart, iend
            ccj1 = ccj1_i1*i+ccj1_i0
            ccj0 = (ccj0_i2*i+ccj0_i1)*i+ccj0_i0
            delta_j = ccj1*ccj1-4*ccj2*ccj0
            IF (delta_j < 0) CONTINUE
            sqDj = SQRT(delta_j)
            jmin = CEILING((-ccj1-sqDj)/(2.0_dp*ccj2))
            jmax = FLOOR((-ccj1+sqDj)/(2.0_dp*ccj2))
            cck0_0_p = cck0_0+(cck0_i2*i+cck0_i)*i
            cck0_j_p = cck0_j+cck0_ij*i
            cck1_0_p = cck1_0+cck1_i*i
            IF (period(1) == 0) THEN
               jmin = MAX(l_bounds(1, 1)-shiftPos(1), jmin)
               jmax = MIN(l_bounds(2, 1)-shiftPos(1), jmax)
            END IF
            jstart = jmin
            ijShift = shiftPos(1)+jstart-l_bounds(2, 1)
            IF (ijShift > 0) ijShift = ijShift+ndim(1)-1
            ijShift = (ijShift/ndim(1))*ndim(1)-shiftPos(1)
            ! ijShift=CEILING(REAL(shiftPos(1)+jstart-l_bounds(2,1))/REAL(ndim(1)))*ndim(1)-shiftPos(1)
            jstart = MAX(ijShift+l_bounds(1, 1), jstart)
            jend = MIN(ijShift+l_bounds(2, 1), jmax)
            DO
               DO j = jstart, jend
                  cck1 = cck1_0_p+cck1_j*j
                  cck0 = cck0_0_p+(cck0_j_p+cck0_j2*j)*j

                  delta_k = cck1*cck1-4*cck2*cck0
                  IF (delta_k < 0) THEN
                     k_bounds(0, j-jmin, i-imin) = 0 ! CEILING((-cck1)/(2.0_dp*cck2))
                     k_bounds(1, j-jmin, i-imin) = -1 ! k_bounds(0,j-jmin,i-imin)-1
                  ELSE
                     sqDk = SQRT(delta_k)
                     kmin = CEILING((-cck1-sqDk)/(2.0_dp*cck2))
                     kmax = FLOOR((-cck1+sqDk)/(2.0_dp*cck2))

                     ! ! reduce kmax,kmin
                     ! ! this should be done later if k_bounds are shared by threads with different slices
                     ! ikShift=FLOOR(REAL(shiftPos(2)+kmax-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
                     ! kmax=MIN(kmax,ikShift+l_bounds(2,2))
                     ! ikShift2=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kmin)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
                     ! kmin=MAX(kmin,ikShift2+l_bounds(1,2))

                     k_bounds(0, j-jmin, i-imin) = kmin
                     k_bounds(1, j-jmin, i-imin) = kmax
                     IF (kmax >= kmin) has_overlap = .TRUE.
                  END IF
               END DO
               jstart = jend+jJump+1
               IF (jstart > jmax) EXIT
               jend = MIN(jend+ndim(1), jmax)
            END DO
         END DO
         istart = iend+iJump+1
         IF (istart > imax) EXIT
         iend = MIN(iend+ndim(0), imax)
      END DO
      IF (period(2) == 0) THEN
         k_bounds(0, :, :) = MAX(l_bounds(1, 2)-shiftPos(2), k_bounds(0, :, :))
         k_bounds(1, :, :) = MIN(l_bounds(2, 2)-shiftPos(2), k_bounds(1, :, :))
      END IF

      IF (.NOT. has_overlap) RETURN

      ! poly x,y,z -> i,j,k
      grad = grad_size3(SIZE(poly)/npoly)
      size_jk = poly_size2(grad)*npoly
      size_k = poly_size1(grad)*npoly
      size_ijk = poly_size3(grad)*npoly
      ALLOCATE (poly_ijk(size_ijk), &
                poly_jk(size_jk), &
                xi(grad+1))

      IF (integrate) THEN
         ALLOCATE (k_vals(0:grad))
      ELSE
         ALLOCATE (poly_k(0:size_k-1))
      ENDIF

      IF (integrate) THEN
         IF(.NOT.(SIZE(poly) == poly_size3(grad)))CALL cp__a("gauss_colloc.F",576)
         poly_ijk = 0.0_dp
      ELSE
         CALL poly_affine_t3(poly, scaled_h, -resPosReal+p_shift, poly_ijk, &
                             npoly=npoly)
      ENDIF

      ij_coeff0 = EXP(-2.0_dp*alphai*m(0, 1))
      ik_coeff0 = EXP(-2.0_dp*alphai*m(0, 2))
      ii_coeff0 = EXP(-alphai*m(0, 0))
      jk_coeff0 = EXP(-2.0_dp*alphai*m(1, 2))
      jj_coeff0 = EXP(-alphai*m(1, 1))
      kk_coeff0 = EXP(-alphai*m(2, 2))
      jk_coeff1 = 1.0_dp/jk_coeff0
      ij_coeff1 = 1.0_dp/ij_coeff0
      ik_coeff1 = 1.0_dp/ik_coeff0
      ii_coeff2 = ii_coeff0*ii_coeff0
      jj_coeff2 = jj_coeff0*jj_coeff0
      kk_coeff2 = kk_coeff0*kk_coeff0
      icoeff0 = EXP(-alphai*l(0))
      jcoeff0 = EXP(-alphai*l(1))
      kcoeff0 = EXP(-alphai*l(2))
      res_0 = EXP(-alphai*r_0)*g_scale

      i_coeffn_i = icoeff0
      j_coeffn_i = jcoeff0
      k_coeffn_i = kcoeff0
      ii_coeffn = i_coeffn_i*ii_coeff0
      res_i = res_0

      iJump = ndim(0)-l_bounds(2, 0)+l_bounds(1, 0)-1
      istart = MAX(0, imin)
      iiShift = shiftPos(0)-l_bounds(2, 0)+istart
      IF (iiShift > 0) iiShift = iiShift+ndim(0)-1
      iiShift = (iiShift/ndim(0))*ndim(0)-shiftPos(0)
      !iiShift=CEILING(REAL(shiftPos(0)+istart-l_bounds(2,0))/REAL(ndim(0)))*ndim(0)-shiftPos(0)
      istart = MAX(iiShift+l_bounds(1, 0), istart)
      iistart = istart-iiShift-l_bounds(1, 0)+l_shift(0)
      istart2 = MIN(-1, imax)
      iiShift2 = shiftPos(0)+istart2-l_bounds(1, 0)
      IF (iiShift2 < 0) iiShift2 = iiShift2-ndim(0)+1
      iiShift2 = (iiShift2/ndim(0))*ndim(0)-shiftPos(0)
      !iiShift2=FLOOR(REAL(shiftPos(0)+istart2-l_bounds(1,0))/REAL(ndim(0)))*ndim(0)-shiftPos(0)
      istart2 = MIN(iiShift2+l_bounds(2, 0), istart2)
      iistart2 = istart2-iiShift2-l_bounds(1, 0)+l_shift(0)

      IF (iJump /= 0 .AND. (iistart+imax-istart >= ndim(0)+l_shift(0) .OR. &
                            iistart2+imin-istart2 <= l_ub(0)-ndim(0))) THEN
         ! will wrap
         ij_coeff0_jump = ij_coeff0**(iJump)
         ik_coeff0_jump = ik_coeff0**(iJump)
         ii_coeff2_jump = ii_coeff2**(iJump)
         ii_coeffn_jump = ii_coeff0**((iJump)*(iJump-1))
      ELSE
         ij_coeff0_jump = 1.0_dp
         ik_coeff0_jump = 1.0_dp
         ii_coeff2_jump = 1.0_dp
         ii_coeffn_jump = 1.0_dp
      END IF

      iend = MIN(iiShift+l_bounds(2, 0), imax)
      ii = iistart 
      IF (i > 0) THEN
         ii_coeffn = i_coeffn_i*ii_coeff0**(2*istart+1)
         j_coeffn_i = jcoeff0*ij_coeff0**istart
         k_coeffn_i = kcoeff0*ik_coeff0**istart
         res_i = res_0*(ii_coeff0**istart*i_coeffn_i)**istart
      END IF
      DO
         DO i = istart, iend
            ! perform j loop
            IF (ABS(res_i) > small) THEN
               CALL j_loop
            END IF
            j_coeffn_i = j_coeffn_i*ij_coeff0
            k_coeffn_i = k_coeffn_i*ik_coeff0
            res_i = res_i*ii_coeffn
            ii_coeffn = ii_coeffn*ii_coeff2
            ii = ii+ 1
         END DO
         istart = iend+iJump+1
         IF (istart > imax) EXIT
         iend = MIN(iend+ndim(0), imax)
         ii = l_shift(0) 
         j_coeffn_i = j_coeffn_i*ij_coeff0_jump
         k_coeffn_i = k_coeffn_i*ik_coeff0_jump
         res_i = res_i*ii_coeffn**(iJump)*ii_coeffn_jump
         ii_coeffn = ii_coeffn*ii_coeff2_jump
      END DO

      ! neg i side
      i_coeffn_i = 1.0_dp/icoeff0
      j_coeffn_i = jcoeff0
      k_coeffn_i = kcoeff0
      res_i = res_0
      ii_coeffn = i_coeffn_i*ii_coeff0

      iend2 = MAX(iiShift2+l_bounds(1, 0), imin)
      ii = iistart2 
      IF (istart2 < -1) THEN
         ii_coeffn = i_coeffn_i*ii_coeff0**(-(2*istart2+1))
         j_coeffn_i = jcoeff0*ij_coeff0**(istart2+1)
         k_coeffn_i = kcoeff0*ik_coeff0**(istart2+1)
         res_i = res_0*(ii_coeff0**(-istart2-1)*i_coeffn_i)**(-istart2-1)
      END IF
      DO
         DO i = istart2, iend2, -1
            j_coeffn_i = j_coeffn_i*ij_coeff1
            k_coeffn_i = k_coeffn_i*ik_coeff1
            res_i = res_i*ii_coeffn
            ii_coeffn = ii_coeffn*ii_coeff2

            ! perform j loop
            IF (ABS(res_i) > small) THEN
               CALL j_loop
            END IF
            ii = ii- 1
         END DO
         istart2 = iend2-iJump-1
         IF (istart2 < imin) EXIT
         iend2 = MAX(iend2-ndim(0), imin)
         ii = l_ub(0) 
         j_coeffn_i = j_coeffn_i/ij_coeff0_jump
         k_coeffn_i = k_coeffn_i/ik_coeff0_jump
         res_i = res_i*ii_coeffn**iJump*ii_coeffn_jump
         ii_coeffn = ii_coeffn*ii_coeff2_jump
      END DO

      ! the final cleanup
      IF (integrate) THEN
         CALL poly_affine_t3t(poly_ijk, scaled_h, -resPosReal+p_shift, poly, &
                              npoly=npoly)
      END IF

      ! rely on compiler to deallocate ALLOCATABLEs

   CONTAINS

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE j_loop()
         ! calculate j bounds
         ccj1 = ccj1_i1*i+ccj1_i0
         ccj0 = (ccj0_i2*i+ccj0_i1)*i+ccj0_i0
         delta_j = ccj1*ccj1-4*ccj2*ccj0
         IF (delta_j < 0) THEN
            RETURN
         END IF
         sqDj = SQRT(delta_j)
         jmin = CEILING((-ccj1-sqDj)/(2.0_dp*ccj2))
         jmax = FLOOR((-ccj1+sqDj)/(2.0_dp*ccj2))

         IF (integrate) THEN
            poly_jk = 0.0_dp
         ELSE
            CALL poly_p_eval3b( poly_ijk(1), size_ijk, REAL(i, dp), &
                                poly_jk(1), size_jk, &
                               npoly=npoly, grad=grad, xi= xi(1))
         ENDIF

         IF (period(1) == 0) THEN
            jmin = MAX(l_bounds(1, 1)-shiftPos(1), jmin)
            jmax = MIN(l_bounds(2, 1)-shiftPos(1), jmax)
         END IF

         ! pos j side
         j_coeffn_j = j_coeffn_i
         k_coeffn_j = k_coeffn_i
         jj_coeffn = j_coeffn_j*jj_coeff0
         res_j = res_i

         jJump = ndim(1)-l_bounds(2, 1)+l_bounds(1, 1)
         jstart = MAX(0, jmin)
         ijShift = shiftPos(1)+jstart-l_bounds(2, 1)
         IF (ijShift > 0) ijShift = ijShift+ndim(1)-1
         ijShift = (ijShift/ndim(1))*ndim(1)-shiftPos(1)
         !ijShift=CEILING(REAL(shiftPos(1)+jstart-l_bounds(2,1))/REAL(ndim(1)))*ndim(1)-shiftPos(1)
         jstart = MAX(ijShift+l_bounds(1, 1), jstart)
         jend = MIN(ijShift+l_bounds(2, 1), jmax)
         ij = (jstart-ijShift-l_bounds(1, 1)+l_shift(1)) 

         IF (jstart > 0) THEN
            k_coeffn_j = k_coeffn_i*jk_coeff0**jstart
            jj_coeffn = j_coeffn_j*jj_coeff0**(2*jstart+1)
            res_j = res_i*(jj_coeff0**jstart*j_coeffn_j)**jstart
         END IF
         DO
            DO j = jstart, jend
               kmin = k_bounds(0, j-jmin, i-imin)
               kmax = k_bounds(1, j-jmin, i-imin)
               ! do k loop
               IF (res_j /= 0.0_dp .AND. k_coeffn_j /= 0.0_dp .AND. kmin <= kmax &
                   .AND. ABS(res_j) > small) THEN
                  CALL k_loop
               END IF
               k_coeffn_j = k_coeffn_j*jk_coeff0
               res_j = res_j*jj_coeffn
               jj_coeffn = jj_coeffn*jj_coeff2
               ij = ij+ 1
            END DO
            jstart = jend+jJump
            IF (jstart > jmax) EXIT
            ij = l_shift(1) 
            jend = MIN(jend+ndim(1), jmax)
            IF (jJump /= 1) THEN ! remove if?
               k_coeffn_j = k_coeffn_i*jk_coeff0**jstart
               jj_coeffn = j_coeffn_j*jj_coeff0**(2*jstart+1)
               res_j = res_i*(jj_coeff0**jstart*j_coeffn_j)**jstart
            END IF
         END DO

         ! neg j side
         j_coeffn_j = 1.0_dp/j_coeffn_i
         k_coeffn_j = k_coeffn_i
         jj_coeffn = j_coeffn_j*jj_coeff0
         res_j = res_i

         jstart = MIN(-1, jmax)
         ijShift = shiftPos(1)+jstart-l_bounds(1, 1)
         IF (ijShift < 0) ijShift = ijShift-ndim(1)+1
         ijShift = (ijShift/ndim(1))*ndim(1)-shiftPos(1)
         !ijShift=FLOOR(REAL(shiftPos(1)+jstart-l_bounds(1,1))/REAL(ndim(1)))*ndim(1)-shiftPos(1))
         jstart = MIN(ijShift+l_bounds(2, 1), jstart)
         jend = MAX(ijShift+l_bounds(1, 1), jmin)
         ij = (jstart-ijShift-l_bounds(1, 1)+l_shift(1)) 
         IF (jstart < -1) THEN
            k_coeffn_j = k_coeffn_i*jk_coeff0**(jstart+1)
            jj_coeffn = j_coeffn_j*jj_coeff0**(-(2*jstart+1))
            res_j = res_i*(jj_coeff0**(-jstart-1)*j_coeffn_j)**(-jstart-1)
         END IF
         DO
            DO j = jstart, jend, -1
               k_coeffn_j = k_coeffn_j*jk_coeff1
               res_j = res_j*jj_coeffn
               jj_coeffn = jj_coeffn*jj_coeff2

               kmin = k_bounds(0, j-jmin, i-imin)
               kmax = k_bounds(1, j-jmin, i-imin)
               ! perform k loop
               IF (res_j /= 0.0_dp .AND. k_coeffn_j /= 0.0_dp .AND. kmin <= kmax &
                   .AND. ABS(res_j) > small) THEN
                  CALL k_loop
               END IF
               ij = ij- 1
            END DO
            jstart = jend-jJump
            IF (jstart < jmin) EXIT
            jend = MAX(jend-ndim(1), jmin)
            ij = l_ub(1) 
            IF (jJump /= 1) THEN ! remove if?
               k_coeffn_j = k_coeffn_i*jk_coeff0**(jstart+1)
               jj_coeffn = j_coeffn_j*jj_coeff0**(-(2*jstart+1))
               res_j = res_i*(jj_coeff0**(-jstart-1)*j_coeffn_j)**(-jstart-1)
            END IF
         END DO

         IF (integrate) THEN
            CALL poly_padd_uneval3b( poly_ijk(1), size_ijk, REAL(i, dp), &
                                     poly_jk(1), size_jk, &
                                    npoly=npoly, grad=grad, xi= xi(1))
            !CALL poly_padd_uneval3(poly_ijk,REAL(i,dp),poly_jk,npoly=npoly)
         ENDIF
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE k_loop()
         IF (integrate) THEN
            SELECT CASE (grad)
            CASE (1)
               CALL kloop1_int
            CASE (2)
               CALL kloop2_int
            CASE (3)
               CALL kloop3_int
            CASE (4)
               CALL kloop4_int
            CASE (5)
               CALL kloop5_int
            CASE (6)
               CALL kloop6_int
            CASE (7)
               CALL kloop7_int
            CASE (8)
               CALL kloop8_int
            CASE default
               CALL kloopdefault_int
            END SELECT
         ELSE
            SELECT CASE (grad)
            CASE (1)
               CALL kloop1_col
            CASE (2)
               CALL kloop2_col
            CASE (3)
               CALL kloop3_col
            CASE (4)
               CALL kloop4_col
            CASE (5)
               CALL kloop5_col
            CASE (6)
               CALL kloop6_col
            CASE (7)
               CALL kloop7_col
            CASE (8)
               CALL kloop8_col
            CASE default
               CALL kloopdefault_col
            END SELECT
         ENDIF
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop1_int()
         INTEGER, PARAMETER :: grad_val = 1
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop2_int()
         INTEGER, PARAMETER :: grad_val = 2
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop3_int()
         INTEGER, PARAMETER :: grad_val = 3
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop4_int()
         INTEGER, PARAMETER :: grad_val = 4
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop5_int()
         INTEGER, PARAMETER :: grad_val = 5
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop6_int()
         INTEGER, PARAMETER :: grad_val = 6
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop7_int()
         INTEGER, PARAMETER :: grad_val = 7
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop8_int()
         INTEGER, PARAMETER :: grad_val = 8
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(8)=k_vals(8)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(8)=k_vals(8)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(8)=k_vals(8)+p_kk

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(1)=k_vals(1)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(2)=k_vals(2)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(3)=k_vals(3)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(4)=k_vals(4)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(5)=k_vals(5)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(6)=k_vals(6)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(7)=k_vals(7)+p_kk
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(8)=k_vals(8)+p_kk
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloopdefault_int()
      INTEGER                                            :: grad_val

         grad_val = grad
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    k_vals=0.0_dp
    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                DO kgrad=1,grad_val
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(kgrad)=k_vals(kgrad)+p_kk
                END DO

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                DO kgrad=1,grad_val
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(kgrad)=k_vals(kgrad)+p_kk
                END DO
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                DO kgrad=1,grad_val
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(kgrad)=k_vals(kgrad)+p_kk
                END DO

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                gval=grid(ik,ij,ii)*res_k
                k_vals(0)=k_vals(0)+gval
                p_kk=gval
                DO kgrad=1,grad_val
                    p_kk=p_kk*REAL(k,dp)
                    k_vals(kgrad)=k_vals(kgrad)+p_kk
                END DO
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
    CALL poly_padd_uneval2b(poly_jk(1),size_jk,REAL(j,dp),k_vals(0),&
        size_k,npoly=npoly,grad=grad_val,xi=xi(1))
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop1_col()
      INTEGER, PARAMETER                                 :: grad_val = 1

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop2_col()
      INTEGER, PARAMETER                                 :: grad_val = 2

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop3_col()
      INTEGER, PARAMETER                                 :: grad_val = 3

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop4_col()
      INTEGER, PARAMETER                                 :: grad_val = 4

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop5_col()
      INTEGER, PARAMETER                                 :: grad_val = 5

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop6_col()
      INTEGER, PARAMETER                                 :: grad_val = 6

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop7_col()
      INTEGER, PARAMETER                                 :: grad_val = 7

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloop8_col()
      INTEGER, PARAMETER                                 :: grad_val = 8

! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(8)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(8)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                    p_v=p_v+poly_k(8)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                    p_v=p_v+poly_k(1)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(2)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(3)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(4)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(5)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(6)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(7)*p_kk
                    p_kk=p_kk*k
                    p_v=p_v+poly_k(8)*p_kk
                    p_kk=p_kk*k
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE kloopdefault_col()
      INTEGER                                            :: grad_val

         grad_val = grad
! to generate the unrolled loops use the following python scripts
!import os,re
!
!doStart=re.compile(r" *do +kgrad *= *(?P<start>[0-9]+) *, *grad_val *$",re.IGNORECASE)
!doEnd=re.compile(r" *end do",re.IGNORECASE)
!for i in xrange(1,9):
!    fIn=file('colloc_int_kloop.f90')
!    fout=file('colloc_int_kloop%d.f90'%(i),'w')
!    while 1:
!        line=fIn.readline()
!        if not line:
!            break
!        m=doStart.match(line)
!        if m:
!            body=""
!            while 1:
!                line=fIn.readline()
!                if not line:
!                    raise Exception('missing end do')
!                if doEnd.match(line):
!                    print 'unrolling',repr(body)
!                    for j in xrange(int(m.group('start')),i+1):
!                        fout.write(body.replace('kgrad',str(j)))
!                    break
!                body+=line
!        else:
!            fout.write(line)
!    fout.close()
!
    CALL poly_p_eval2b(poly_jk(1),size_jk,REAL(j,dp),&
         poly_k(0),size_k,npoly=npoly,grad=grad_val,xi=xi(1))
    ! starting point
    kJump=ndim(2)-l_bounds(2,2)+l_bounds(1,2)
    kstart=MAX(0,kmin)
    ikShift=shiftPos(2)-l_bounds(2,2)+kstart
    IF (ikShift>0) ikShift=ikShift+ndim(2)-1
    ikShift=(ikShift/ndim(2))*ndim(2)-shiftPos(2)
    ! ikShift=CEILING(REAL(shiftPos(2)-l_bounds(2,2)+kstart)/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart=MAX(ikShift+l_bounds(1,2),kstart)
    kend=MIN(ikShift+l_bounds(2,2),kmax)
    ikstart=kstart-ikShift-l_bounds(1,2)+l_shift(2)
    kstart2=MIN(-1,kmax)
    ikShift2=shiftPos(2)+kstart2-l_bounds(1,2)
    IF (ikShift2<0) ikShift2=ikShift2-ndim(2)+1
    ikShift2=(ikShift2/ndim(2))*ndim(2)-shiftPos(2)
    !ikShift2=FLOOR(REAL(shiftPos(2)+kstart2-l_bounds(1,2))/REAL(ndim(2)))*ndim(2)-shiftPos(2)
    kstart2=MIN(ikShift2+l_bounds(2,2),kstart2)
    kend2=MAX(ikShift2+l_bounds(1,2),kmin)
    ikstart2=kstart2-ikShift2-l_bounds(1,2)+l_shift(2)

    IF (kJump/=1 .AND. (ikstart+kmax-kstart>=ndim(2)+l_shift(2) .OR.&
        ikstart2+kmin-kstart2<=l_ub(2)-ndim(2))) THEN
        ! will wrap
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                DO kgrad=1,grad_val
                    p_v=p_v+poly_k(kgrad)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                END DO
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                DO kgrad=1,grad_val
                    p_v=p_v+poly_k(kgrad)*p_kk
                    p_kk=p_kk*k
                END DO
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END DO
    ELSE
        ! no jump
        ! pos k side
        k_coeffn_k=k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart
        ik=ikstart
        IF (k>0) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(2*kstart+1)
            res_k=res_j*(kk_coeff0**kstart*k_coeffn_k)**kstart
        END IF
        DO
            DO k=kstart,kend
                p_v=poly_k(0)
                p_kk=REAL(k,dp)
                DO kgrad=1,grad_val
                    p_v=p_v+poly_k(kgrad)*p_kk
                    p_kk=p_kk*REAL(k,dp)
                END DO
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF

                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                ik=ik+1
            END DO
            kstart=kend+kJump
            IF (kstart>kmax) EXIT
            kend=MIN(kend+ndim(2),kmax)
            ik=l_shift(2)
        END DO

        ! neg k side
        k_coeffn_k=1.0_dp/k_coeffn_j
        kk_coeffn=k_coeffn_k*kk_coeff0
        res_k=res_j
        k=kstart2
        ik=ikstart2
        IF (k<-1) THEN
            kk_coeffn=k_coeffn_k*kk_coeff0**(-(2*kstart2+1))
            res_k=res_j*(kk_coeff0**(-kstart2-1)*k_coeffn_k)**(-kstart2-1)
        END IF
        DO
            DO k=kstart2,kend2,-1
                res_k=res_k*kk_coeffn
                kk_coeffn=kk_coeffn*kk_coeff2
                p_v=poly_k(0)
                p_kk=k
                DO kgrad=1,grad_val
                    p_v=p_v+poly_k(kgrad)*p_kk
                    p_kk=p_kk*k
                END DO
                IF ( PRESENT ( lgrid ) ) THEN
                  ig = lgrid%ldim * ithread + ii * (l_bounds(2,2)-l_bounds(1,2)+1) * (l_bounds(2,1)-l_bounds(1,1)+1) + &
                       ij * (l_bounds(2,2)-l_bounds(1,2)+1) + ik + 1
                  lgrid%r(ig)=lgrid%r(ig) + p_v*res_k
                ELSE
                  grid(ik,ij,ii) = grid(ik,ij,ii) + p_v*res_k
                END IF
                ik=ik-1
            END DO
            kstart2=kend2-kJump
            IF (kstart2<kmin) EXIT
            kend2=MAX(kend2-ndim(2),kmin)
            ik=l_ub(2)
        END DO
    END IF
      END SUBROUTINE

   END SUBROUTINE colloc_int_body

END MODULE gauss_colloc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the plane wave density by collocating the primitive Gaussian
!>      functions (pgf).
!> \par History
!>      - rewrote collocate for increased accuracy and speed
!>      - introduced the PGI hack for increased speed with that compiler
!>        (22.02.02)
!>      - Added Multiple Grid feature
!>      - new way to get over the grid (01.03.02)
!>      - removed timing calls since they were getting expensive
!>      - Updated with the new QS data structures (09.04.02,MK)
!>      - introduction of the real space grid type ( prelim. version JVdV 05.02)
!>      - parallel FFT (JGH 22.05.02)
!>      - multigrid arrays independent from density (JGH 30.08.02)
!>      - old density stored in g space (JGH 30.08.02)
!>      - distributed real space code (JGH 17.07.03)
!>      - refactoring and new loop ordering (JGH 23.11.03)
!>      - OpenMP parallelization (JGH 03.12.03)
!>      - Modified to compute tau (Joost 12.03)
!>      - removed the incremental density rebuild (Joost 01.04)
!>      - introduced realspace multigridding (Joost 02.04)
!>      - introduced map_consistent (Joost 02.04)
!>      - Addition of the subroutine calculate_atomic_charge_density (TdK, 08.05)
!>      - rewrite of the collocate/integrate kernels (Joost VandeVondele, 03.07)
!> \author Matthias Krack (03.04.2001)
!>      1) Joost VandeVondele (01.2002)
!>      Thomas D. Kuehne (04.08.2005)
! **************************************************************************************************
MODULE qs_collocate_density
   USE ao_util,                         ONLY: exp_radius_very_extended
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_fm_types,                     ONLY: cp_fm_get_element,&
                                              cp_fm_get_info,&
                                              cp_fm_type
   USE cube_utils,                      ONLY: compute_cube_center,&
                                              cube_info_type,&
                                              return_cube,&
                                              return_cube_nonortho
   USE d3_poly,                         ONLY: poly_cp2k2d3
   USE external_potential_types,        ONLY: get_potential,&
                                              gth_potential_type
   USE gauss_colloc,                    ONLY: collocGauss
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel,&
                                              gridlevel_info_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_8
   USE lgrid_types,                     ONLY: lgrid_allocate_grid,&
                                              lgrid_type
   USE lri_environment_types,           ONLY: lri_kind_type
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              pi,&
                                              twopi
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_integrate_function,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type,&
                                              pw_pools_create_pws,&
                                              pw_pools_give_back_pws
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_modify_pab_block,             ONLY: &
        FUNC_AB, FUNC_ADBmDAB, FUNC_ARB, FUNC_ARDBmDARB, FUNC_DABpADB, FUNC_DADB, FUNC_DER, &
        FUNC_DX, FUNC_DXDX, FUNC_DXDY, FUNC_DY, FUNC_DYDY, FUNC_DYDZ, FUNC_DZ, FUNC_DZDX, &
        FUNC_DZDZ, prepare_adb_m_dab, prepare_arb, prepare_ardb_m_darb, prepare_dab_p_adb, &
        prepare_dadb, prepare_diadib, prepare_diiadiib, prepare_dijadijb
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              realspace_grid_type,&
                                              rs2pw,&
                                              rs_grid_release,&
                                              rs_grid_retain,&
                                              rs_grid_zero,&
                                              rs_pw_transfer
   USE rs_pw_interface,                 ONLY: density_rs2pw,&
                                              density_rs2pw_basic
   USE scptb_types,                     ONLY: get_scptb_parameter,&
                                              scp_vector_type,&
                                              scptb_parameter_type
   USE task_list_methods,               ONLY: int2pair,&
                                              rs_distribute_matrix
   USE task_list_types,                 ONLY: task_list_type
   USE termination,                     ONLY: stop_memory

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_collocate_density'
! *** Public subroutines ***

   PUBLIC :: calculate_ppl_grid, &
             calculate_rho_core, &
             calculate_lri_rho_elec, &
             calculate_rho_single_gaussian, &
             calculate_rho_metal, &
             calculate_rho_resp_single, &
             calculate_rho_resp_all, &
             calculate_rho_elec, &
             calculate_drho_elec, &
             calculate_scp_charge, &
             calculate_wavefunction, &
             collocate_pgf_product_gspace, &
             collocate_pgf_product_rspace, &
             calculate_rho_nlcc

   INTEGER :: debug_count = 0

CONTAINS

! **************************************************************************************************
!> \brief computes the SCP charge distribution on the grid
!> \param scp ...
!> \param qs_env ...
!> \param scpv ...
! **************************************************************************************************
   SUBROUTINE calculate_scp_charge(scp, qs_env, scpv)

      TYPE(pw_p_type), INTENT(INOUT)                     :: scp
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(scp_vector_type), POINTER                     :: scpv

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_scp_charge', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, i, iatom, ii, ikind, &
                                                            ithread, j, jj, l, lmax_global, &
                                                            lmaxscp, natom, ni, nj, npme, nthread
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: defined
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace, norm, pp
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type)                               :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_rho
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, pab, qs_kind_set, particle_set, atomic_kind_set, &
               atom_list, pw_env, rs_rho, auxbas_pw_pool, cores)

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control, &
                      cell=cell, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      cube_info = pw_env%cube_info(1)
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      !Find the maximum la_max over all of the atomic_kind_set
      !We are using FUNC_AB so no need to cater for different ga_gb_functions
      lmax_global = 0
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
         CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp)
         lmax_global = MAX(lmaxscp, lmax_global)
      END DO

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
         CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=alpha)

         IF (.NOT. defined) CYCLE

         ni = ncoset(lmaxscp)
         ALLOCATE (pab(ni, 1))

         nthread = 1
         ithread = 0

         ALLOCATE (cores(natom))
         cores = 0
         npme = 0

         DO iatom = 1, natom
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         END DO

         DO nj = 1, npme

            iatom = cores(nj)
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            subpatch_pattern = 0
            pab = 0._dp
            DO l = 0, lmaxscp
               pp = (2._dp*l+3._dp)/2._dp
               norm = 2._dp**(l+2)/SQRT(pi)/dfac(2*l+1)
               norm = SQRT(0.25_dp*dfac(2*l+1)/pi)*norm*alpha**pp
               DO jj = 1, nco(l)
                  j = ncoset(l-1)+jj
                  DO ii = 1, nso(l)
                     i = nsoset(l-1)+ii
                     pab(j, 1) = pab(j, 1)+scpv%vector(ikind)%vmat(i, iatom)* &
                                 orbtramat(l)%s2c(ii, jj)*norm
                  END DO
               END DO
            END DO
            CALL collocate_pgf_product_rspace(lmaxscp, alpha, 0, 0, 0.0_dp, 0, ra, &
                                              (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                              cell, cube_info, eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                              ithread=ithread, use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                              lmax_global=lmax_global)

         END DO

         DEALLOCATE (pab, cores)

      END DO

      CALL rs_pw_transfer(rs_rho, scp%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL timestop(handle)

   END SUBROUTINE calculate_scp_charge
! **************************************************************************************************
!> \brief computes the density of the non-linear core correction on the grid
!> \param rho_nlcc ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE calculate_rho_nlcc(rho_nlcc, qs_env)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_nlcc
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_nlcc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, iexp_nlcc, ikind, &
                                                            ithread, j, lmax_global, n, natom, nc, &
                                                            nexp_nlcc, ni, npme, nthread
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores, mylmax, nct_nlcc
      LOGICAL                                            :: nlcc
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_nlcc
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval_nlcc, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type)                               :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, pab, particle_set, atomic_kind_set, qs_kind_set, &
               atom_list, pw_env, rs_rho, auxbas_pw_pool, cores, mylmax)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      cube_info = pw_env%cube_info(1)
      ! be careful in parallel nsmax is choosen with multigrid in mind!
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      !Find the maximum la_max over all of the atomic_kind_set
      !We are using FUNC_AB so no need to cater for different ga_gb_functions
      lmax_global = 0
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, nct_nlcc=mylmax, nlcc_present=nlcc)
         IF (.NOT. nlcc) CYCLE
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO
      !lmax_global set according to ni = 2*nc-2 below
      lmax_global = 2*lmax_global-2

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)

         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, nlcc_present=nlcc, nexp_nlcc=nexp_nlcc, &
                            alpha_nlcc=alpha_nlcc, nct_nlcc=nct_nlcc, cval_nlcc=cval_nlcc)

         IF (.NOT. nlcc) CYCLE

         DO iexp_nlcc = 1, nexp_nlcc

            alpha = alpha_nlcc(iexp_nlcc)
            nc = nct_nlcc(iexp_nlcc)

            ni = ncoset(2*nc-2)
            ALLOCATE (pab(ni, 1))
            pab = 0._dp

            nthread = 1
            ithread = 0

            CALL reallocate(cores, 1, natom)
            npme = 0
            cores = 0

            ! prepare core function
            DO j = 1, nc
               SELECT CASE (j)
               CASE (1)
                  pab(1, 1) = cval_nlcc(1, iexp_nlcc)
               CASE (2)
                  n = coset(2, 0, 0)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
                  n = coset(0, 2, 0)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
                  n = coset(0, 0, 2)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
               CASE (3)
                  n = coset(4, 0, 0)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 4, 0)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 0, 4)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(2, 2, 0)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(2, 0, 2)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 2, 2)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
               CASE (4)
                  n = coset(6, 0, 0)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 6, 0)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 0, 6)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(4, 2, 0)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(4, 0, 2)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 4, 0)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 0, 4)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 4, 2)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 2, 4)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 2, 2)
                  pab(n, 1) = 6._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
               CASE DEFAULT
                  CALL cp__b("qs_collocate_density.F",438,"")
               END SELECT
            END DO
            IF (dft_control%nspins == 2) pab = pab*0.5_dp

            DO iatom = 1, natom
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
                  ! replicated realspace grid, split the atoms up between procs
                  IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
                     npme = npme+1
                     cores(npme) = iatom
                  ENDIF
               ELSE
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            END DO

            DO j = 1, npme

               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               subpatch_pattern = 0
               ni = 2*nc-2
               CALL collocate_pgf_product_rspace(ni, 1/(2*alpha**2), 0, 0, 0.0_dp, 0, ra, &
                                                 (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                                 cell, cube_info, eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                                 ithread=ithread, use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                                 lmax_global=lmax_global)

            END DO

            DEALLOCATE (pab)

         END DO

      END DO

      IF (ASSOCIATED(cores)) THEN
         DEALLOCATE (cores)
      END IF

      CALL rs_pw_transfer(rs_rho, rho_nlcc%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_nlcc

! **************************************************************************************************
!> \brief computes the local pseudopotential (without erf term) on the grid
!> \param vppl ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE calculate_ppl_grid(vppl, qs_env)

      TYPE(pw_p_type), INTENT(INOUT)                     :: vppl
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_ppl_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ikind, ithread, &
                                                            j, lmax_global, lppl, n, natom, ni, &
                                                            npme, nthread
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cexp_ppl
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type)                               :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, pab, atomic_kind_set, qs_kind_set, particle_set, &
               atom_list, pw_env, rs_rho, auxbas_pw_pool, cores)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      cube_info = pw_env%cube_info(1)
      ! be careful in parallel nsmax is choosen with multigrid in mind!
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      !Find maximum la_max over all of the atomic kind set where la_max=ni & ni=2*lppl-2
      !Thus need to find the max of lppl over atomic_kind_set and use to define lmax_global
      !We are using FUNC_AB so no need to cater for different ga_gb_functions
      lmax_global = 0
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, nexp_ppl=lppl)
         lmax_global = MAX(lppl, lmax_global)
      END DO
      lmax_global = 2*lmax_global-2

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)

         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, alpha_ppl=alpha, nexp_ppl=lppl, cexp_ppl=cexp_ppl)

         IF (lppl <= 0) CYCLE

         ni = ncoset(2*lppl-2)
         ALLOCATE (pab(ni, 1))
         pab = 0._dp

         nthread = 1
         ithread = 0

         CALL reallocate(cores, 1, natom)
         npme = 0
         cores = 0

         ! prepare core function
         DO j = 1, lppl
            SELECT CASE (j)
            CASE (1)
               pab(1, 1) = cexp_ppl(1)
            CASE (2)
               n = coset(2, 0, 0)
               pab(n, 1) = cexp_ppl(2)
               n = coset(0, 2, 0)
               pab(n, 1) = cexp_ppl(2)
               n = coset(0, 0, 2)
               pab(n, 1) = cexp_ppl(2)
            CASE (3)
               n = coset(4, 0, 0)
               pab(n, 1) = cexp_ppl(3)
               n = coset(0, 4, 0)
               pab(n, 1) = cexp_ppl(3)
               n = coset(0, 0, 4)
               pab(n, 1) = cexp_ppl(3)
               n = coset(2, 2, 0)
               pab(n, 1) = 2._dp*cexp_ppl(3)
               n = coset(2, 0, 2)
               pab(n, 1) = 2._dp*cexp_ppl(3)
               n = coset(0, 2, 2)
               pab(n, 1) = 2._dp*cexp_ppl(3)
            CASE (4)
               n = coset(6, 0, 0)
               pab(n, 1) = cexp_ppl(4)
               n = coset(0, 6, 0)
               pab(n, 1) = cexp_ppl(4)
               n = coset(0, 0, 6)
               pab(n, 1) = cexp_ppl(4)
               n = coset(4, 2, 0)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(4, 0, 2)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 4, 0)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 0, 4)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(0, 4, 2)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(0, 2, 4)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 2, 2)
               pab(n, 1) = 6._dp*cexp_ppl(4)
            CASE DEFAULT
               CALL cp__b("qs_collocate_density.F",623,"")
            END SELECT
         END DO

         DO iatom = 1, natom
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         END DO

         IF (npme .GT. 0) THEN
            DO j = 1, npme

               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               subpatch_pattern = 0
               ni = 2*lppl-2
               CALL collocate_pgf_product_rspace(ni, alpha, 0, 0, 0.0_dp, 0, ra, &
                                                 (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                                 cell, cube_info, eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                                 ithread=ithread, use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                                 lmax_global=lmax_global)

            END DO
         ENDIF

         DEALLOCATE (pab)

      END DO

      IF (ASSOCIATED(cores)) THEN
         DEALLOCATE (cores)
      END IF

      CALL rs_pw_transfer(rs_rho, vppl%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL timestop(handle)

   END SUBROUTINE calculate_ppl_grid

! **************************************************************************************************
!> \brief Collocates the fitted lri density on a grid.
!> \param lri_rho_g ...
!> \param lri_rho_r ...
!> \param qs_env ...
!> \param lri_coef ...
!> \param total_rho ...
!> \par History
!>      04.2013
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calculate_lri_rho_elec(lri_rho_g, lri_rho_r, qs_env, &
                                     lri_coef, total_rho)

      TYPE(pw_p_type), INTENT(INOUT)                     :: lri_rho_g, lri_rho_r
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: lri_coef
      REAL(KIND=dp), INTENT(OUT)                         :: total_rho

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_lri_rho_elec', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, dir, group_size, handle, iatom, ierr, igrid_level, ikind, ipgf, iset, &
         lmax_global, m1, maxco, my_pos, na1, natom, ncoa, nseta, offset, sgfa
      INTEGER, DIMENSION(3)                              :: lb, location, tp, ub
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, la_max, la_min, mylmax, &
                                                            npgfa, nsgfa
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: map_consistent, map_it_here
      REAL(KIND=dp)                                      :: eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:), POINTER               :: aci
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab, sphi_a, work, zeta
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: lri_basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: mgrid_gspace, mgrid_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      NULLIFY (aci, atomic_kind_set, qs_kind_set, atom_list, cell, cube_info, &
               dft_control, first_sgfa, gridlevel_info, la_max, la_min, lri_basis_set, &
               mgrid_gspace, mgrid_rspace, npgfa, nsgfa, pab, &
               particle_set, pw_env, pw_pools, rs_grid, rs_rho, sphi_a, &
               work, zeta, mylmax)

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      cell=cell, particle_set=particle_set, &
                      pw_env=pw_env, &
                      dft_control=dft_control)

      cube_info => pw_env%cube_info
      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      map_consistent = dft_control%qs_control%map_consistent
      gridlevel_info => pw_env%gridlevel_info

      ! *** set up the pw multi-grids *** !
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_collocate_density.F",741)
      CALL pw_env_get(pw_env=pw_env, rs_grids=rs_rho, pw_pools=pw_pools)

      CALL pw_pools_create_pws(pw_pools, mgrid_rspace, &
                               use_data=REALDATA3D, &
                               in_space=REALSPACE)

      CALL pw_pools_create_pws(pw_pools, mgrid_gspace, &
                               use_data=COMPLEXDATA1D, &
                               in_space=RECIPROCALSPACE)

      ! *** set up the rs multi-grids *** !
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_grid_retain(rs_rho(igrid_level)%rs_grid)
         CALL rs_grid_zero(rs_rho(igrid_level)%rs_grid)
      END DO

      !take maxco from the LRI basis set!
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, basis_type="LRI")

      ALLOCATE (pab(maxco, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 763, &
                                      "pab", maxco*dp_size)
      offset = 0
      my_pos = mgrid_rspace(1)%pw%pw_grid%para%my_pos
      group_size = mgrid_rspace(1)%pw%pw_grid%para%group_size

      !Find lmax_global across atomic_kind_set
      lmax_global = 0
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
         CALL get_gto_basis_set(gto_basis_set=lri_basis_set, lmax=mylmax)
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO

      DO ikind = 1, SIZE(atomic_kind_set)

         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")

         !Take the lri basis set here!
         CALL get_gto_basis_set(gto_basis_set=lri_basis_set, lmax=la_max, &
                                lmin=la_min, zet=zeta, nset=nseta, npgf=npgfa, &
                                sphi=sphi_a, first_sgf=first_sgfa, nsgf_set=nsgfa)

         DO iatom = 1, natom
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            aci => lri_coef(ikind)%acoef(iatom, :)

            DO iset = 1, nseta
               sgfa = first_sgfa(1, iset)
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               m1 = sgfa+nsgfa(iset)-1
               CALL reallocate(work, 1, nsgfa(iset), 1, 1)
               work(1:nsgfa(iset), 1) = aci(sgfa:m1)
               pab = 0._dp

               CALL dgemm("N", "N", ncoa, 1, nsgfa(iset), 1.0_dp, lri_basis_set%sphi(1, sgfa), &
                          SIZE(lri_basis_set%sphi, 1), work(1, 1), SIZE(work, 1), 0.0_dp, pab(1, 1), &
                          SIZE(pab, 1))

               DO ipgf = 1, npgfa(iset)
                  na1 = (ipgf-1)*ncoset(la_max(iset))
                  igrid_level = gaussian_gridlevel(gridlevel_info, zeta(ipgf, iset))
                  rs_grid => rs_rho(igrid_level)%rs_grid

                  map_it_here = .FALSE.

                  IF (.NOT. ALL(rs_grid%desc%perd == 1)) THEN
                     DO dir = 1, 3
                        ! bounds of local grid (i.e. removing the 'wings'), if periodic
                        tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_grid%desc%npts(dir))
                        tp(dir) = MODULO(tp(dir), rs_grid%desc%npts(dir))
                        IF (rs_grid%desc%perd(dir) .NE. 1) THEN
                           lb(dir) = rs_grid%lb_local(dir)+rs_grid%desc%border
                           ub(dir) = rs_grid%ub_local(dir)-rs_grid%desc%border
                        ELSE
                           lb(dir) = rs_grid%lb_local(dir)
                           ub(dir) = rs_grid%ub_local(dir)
                        ENDIF
                        ! distributed grid, only map if it is local to the grid
                        location(dir) = tp(dir)+rs_grid%desc%lb(dir)
                     ENDDO
                     IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                         lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                         lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                        map_it_here = .TRUE.
                     ENDIF
                  ELSE
                     ! not distributed, just a round-robin distribution over the full set of CPUs
                     IF (MODULO(offset, group_size) == my_pos) map_it_here = .TRUE.
                  ENDIF

                  IF (map_it_here) THEN
                     CALL collocate_pgf_product_rspace(la_max=la_max(iset), &
                                                       zeta=zeta(ipgf, iset), &
                                                       la_min=la_min(iset), &
                                                       lb_max=0, zetb=0.0_dp, lb_min=0, &
                                                       ra=ra, rab=(/0.0_dp, 0.0_dp, 0.0_dp/), &
                                                       rab2=0.0_dp, scale=1._dp, &
                                                       pab=pab, o1=na1, o2=0, &
                                                       rsgrid=rs_grid, cell=cell, &
                                                       cube_info=cube_info(igrid_level), &
                                                       eps_rho_rspace=eps_rho_rspace, &
                                                       ga_gb_function=FUNC_AB, &
                                                       map_consistent=map_consistent, &
                                                       lmax_global=lmax_global)
                  ENDIF
               END DO

               offset = offset+nsgfa(iset)

            END DO

            DEALLOCATE (work)

         END DO
      END DO

      DEALLOCATE (pab)

      CALL pw_zero(lri_rho_g%pw)
      CALL pw_zero(lri_rho_r%pw)

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL pw_zero(mgrid_rspace(igrid_level)%pw)
         CALL rs_pw_transfer(rs=rs_rho(igrid_level)%rs_grid, &
                             pw=mgrid_rspace(igrid_level)%pw, dir=rs2pw)
         CALL rs_grid_release(rs_rho(igrid_level)%rs_grid)
      END DO

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL pw_zero(mgrid_gspace(igrid_level)%pw)
         CALL pw_transfer(mgrid_rspace(igrid_level)%pw, &
                          mgrid_gspace(igrid_level)%pw)
         CALL pw_axpy(mgrid_gspace(igrid_level)%pw, lri_rho_g%pw)
      END DO
      CALL pw_transfer(lri_rho_g%pw, lri_rho_r%pw)
      total_rho = pw_integrate_function(lri_rho_r%pw, isign=-1)

      ! *** give back the multi-grids *** !
      CALL pw_pools_give_back_pws(pw_pools, mgrid_gspace)
      CALL pw_pools_give_back_pws(pw_pools, mgrid_rspace)

      CALL timestop(handle)

   END SUBROUTINE calculate_lri_rho_elec

! **************************************************************************************************
!> \brief computes the density of the core charges on the grid
!> \param rho_core ...
!> \param total_rho ...
!> \param qs_env ...
!> \param only_nopaw ...
! **************************************************************************************************
   SUBROUTINE calculate_rho_core(rho_core, total_rho, qs_env, only_nopaw)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_core
      REAL(KIND=dp), INTENT(OUT)                         :: total_rho
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: only_nopaw

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_core', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ierr, ikind, &
                                                            ithread, j, natom, npme, nthread
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: my_only_nopaw, paw_atom
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type)                               :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhoc_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, pab, atomic_kind_set, qs_kind_set, particle_set, &
               atom_list, pw_env, rs_rho, auxbas_pw_pool, cores)
      ALLOCATE (pab(1, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 931, &
                                      "pab", dp_size)

      my_only_nopaw = .FALSE.
      IF (PRESENT(only_nopaw)) my_only_nopaw = only_nopaw

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      cube_info = pw_env%cube_info(1)
      ! be careful in parallel nsmax is choosen with multigrid in mind!
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom, &
                          alpha_core_charge=alpha, ccore_charge=pab(1, 1))

         IF (my_only_nopaw .AND. paw_atom) CYCLE
         IF (alpha == 0.0_dp .OR. pab(1, 1) == 0.0_dp) CYCLE

         nthread = 1
         ithread = 0

         CALL reallocate(cores, 1, natom)
         npme = 0
         cores = 0

         DO iatom = 1, natom
            IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         END DO

         IF (npme .GT. 0) THEN
            DO j = 1, npme

               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               subpatch_pattern = 0
               !lmax == 0 so set lmax_global to 0
               CALL collocate_pgf_product_rspace(0, alpha, 0, 0, 0.0_dp, 0, ra, &
                                                 (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, -1.0_dp, pab, 0, 0, rs_rho, &
                                                 cell, cube_info, eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                                 ithread=ithread, use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                                 lmax_global=0)

            END DO
         ENDIF

      END DO

      IF (ASSOCIATED(cores)) THEN
         DEALLOCATE (cores)
      END IF
      DEALLOCATE (pab)

      CALL pw_pool_create_pw(auxbas_pw_pool, rhoc_r%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL rs_pw_transfer(rs_rho, rhoc_r%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      total_rho = pw_integrate_function(rhoc_r%pw, isign=-1)

      CALL pw_transfer(rhoc_r%pw, rho_core%pw)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhoc_r%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_core

! **************************************************************************************************
!> \brief collocate a single Gaussian on the grid
!> \param rho_gb charge density generated by a single gaussian
!> \param qs_env qs environment
!> \param iatom_in atom index
!> \par History
!>        12.2011 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calculate_rho_single_gaussian(rho_gb, qs_env, iatom_in)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_gb
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom_in

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_single_gaussian', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ierr, npme
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      REAL(KIND=dp)                                      :: eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhoc_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, pab, pw_env, rs_rho, auxbas_pw_pool)

      ALLOCATE (pab(1, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 1055, &
                                      "pab", dp_size)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      pab(1, 1) = 1.0_dp
      iatom = iatom_in

      npme = 0

      IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
         IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
            npme = npme+1
         ENDIF
      ELSE
         npme = npme+1
      ENDIF

      IF (npme .GT. 0) THEN
         atom_a = qs_env%qmmm_env_qm%image_charge_pot%image_mm_list(iatom)
         ra(:) = pbc(qs_env%qmmm_env_qm%image_charge_pot%particles_all(atom_a)%r, cell)
         subpatch_pattern = 0
         !lmax == 0 so set lmax_global to 0
         CALL collocate_pgf_product_rspace(0, qs_env%qmmm_env_qm%image_charge_pot%eta, &
                                           0, 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                           cell, pw_env%cube_info(1), eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                           use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                           lmax_global=0)
      ENDIF

      DEALLOCATE (pab)

      CALL pw_pool_create_pw(auxbas_pw_pool, rhoc_r%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL rs_pw_transfer(rs_rho, rhoc_r%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL pw_transfer(rhoc_r%pw, rho_gb%pw)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhoc_r%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_single_gaussian

! **************************************************************************************************
!> \brief computes the image charge density on the grid (including coeffcients)
!> \param rho_metal image charge density
!> \param coeff expansion coefficients of the image charge density, i.e.
!>        rho_metal=sum_a c_a*g_a
!> \param total_rho_metal total induced image charge density
!> \param qs_env qs environment
!> \par History
!>        01.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calculate_rho_metal(rho_metal, coeff, total_rho_metal, qs_env)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_metal
      REAL(KIND=dp), DIMENSION(:), POINTER               :: coeff
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: total_rho_metal
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_metal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ierr, j, natom, &
                                                            npme
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: cores
      REAL(KIND=dp)                                      :: eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhoc_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, pab, pw_env, rs_rho, auxbas_pw_pool, cores)

      ALLOCATE (pab(1, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 1149, &
                                      "pab", dp_size)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      pab(1, 1) = 1.0_dp

      natom = SIZE(qs_env%qmmm_env_qm%image_charge_pot%image_mm_list)

      CALL reallocate(cores, 1, natom)
      npme = 0
      cores = 0

      DO iatom = 1, natom
         IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
            IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         ELSE
            npme = npme+1
            cores(npme) = iatom
         ENDIF
      ENDDO

      IF (npme .GT. 0) THEN
         DO j = 1, npme
            iatom = cores(j)
            atom_a = qs_env%qmmm_env_qm%image_charge_pot%image_mm_list(iatom)
            ra(:) = pbc(qs_env%qmmm_env_qm%image_charge_pot%particles_all(atom_a)%r, cell)
            subpatch_pattern = 0
            !lmax == 0 so set lmax_global to 0
            CALL collocate_pgf_product_rspace( &
               0, qs_env%qmmm_env_qm%image_charge_pot%eta, &
               0, 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, coeff(iatom), pab, 0, 0, rs_rho, &
               cell, pw_env%cube_info(1), eps_rho_rspace, ga_gb_function=FUNC_AB, &
               use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, lmax_global=0)
         ENDDO
      ENDIF

      DEALLOCATE (pab, cores)

      CALL pw_pool_create_pw(auxbas_pw_pool, rhoc_r%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL rs_pw_transfer(rs_rho, rhoc_r%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      IF (PRESENT(total_rho_metal)) &
         !minus sign: account for the fact that rho_metal has opposite sign
         total_rho_metal = pw_integrate_function(rhoc_r%pw, isign=-1)

      CALL pw_transfer(rhoc_r%pw, rho_metal%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhoc_r%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_metal

! **************************************************************************************************
!> \brief collocate a single Gaussian on the grid for periodic RESP fitting
!> \param rho_gb charge density generated by a single gaussian
!> \param qs_env qs environment
!> \param eta width of single Gaussian
!> \param iatom_in atom index
!> \par History
!>        06.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calculate_rho_resp_single(rho_gb, qs_env, eta, iatom_in)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_gb
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(IN)                          :: eta
      INTEGER, INTENT(IN)                                :: iatom_in

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_resp_single', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ierr, npme
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      REAL(KIND=dp)                                      :: eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhoc_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, pab, pw_env, rs_rho, auxbas_pw_pool, &
               particle_set)

      ALLOCATE (pab(1, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 1254, &
                                      "pab", dp_size)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      pab(1, 1) = 1.0_dp
      iatom = iatom_in

      npme = 0

      IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
         IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
            npme = npme+1
         ENDIF
      ELSE
         npme = npme+1
      ENDIF

      IF (npme .GT. 0) THEN
         ra(:) = pbc(particle_set(iatom)%r, cell)
         subpatch_pattern = 0
         ! lmax == 0 so set lmax_global to 0
         CALL collocate_pgf_product_rspace(0, eta, 0, 0, 0.0_dp, 0, ra, &
                                           (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                           cell, pw_env%cube_info(1), eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                           use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                           lmax_global=0)
      ENDIF

      DEALLOCATE (pab)

      CALL pw_pool_create_pw(auxbas_pw_pool, rhoc_r%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL rs_pw_transfer(rs_rho, rhoc_r%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL pw_transfer(rhoc_r%pw, rho_gb%pw)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhoc_r%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_resp_single

! **************************************************************************************************
!> \brief computes the RESP charge density on a grid based on the RESP charges
!> \param rho_resp RESP charge density
!> \param coeff RESP charges, take care of normalization factor
!>        (eta/pi)**1.5 later
!> \param natom number of atoms
!> \param eta width of single Gaussian
!> \param qs_env qs environment
!> \par History
!>        01.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calculate_rho_resp_all(rho_resp, coeff, natom, eta, qs_env)

      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_resp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: coeff
      INTEGER, INTENT(IN)                                :: natom
      REAL(KIND=dp), INTENT(IN)                          :: eta
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_resp_all', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ierr, j, npme
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: cores
      REAL(KIND=dp)                                      :: eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhoc_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)

      NULLIFY (cell, cores, dft_control, pab, pw_env, rs_rho, auxbas_pw_pool, &
               particle_set)

      ALLOCATE (pab(1, 1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 1351, &
                                      "pab", dp_size)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      pab(1, 1) = 1.0_dp

      CALL reallocate(cores, 1, natom)
      npme = 0
      cores = 0

      DO iatom = 1, natom
         IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
            IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         ELSE
            npme = npme+1
            cores(npme) = iatom
         ENDIF
      ENDDO

      IF (npme .GT. 0) THEN
         DO j = 1, npme
            iatom = cores(j)
            ra(:) = pbc(particle_set(iatom)%r, cell)
            subpatch_pattern = 0
            ! la_max==0 so set lmax_global to 0
            CALL collocate_pgf_product_rspace( &
               0, eta, &
               0, 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, coeff(iatom), pab, 0, 0, rs_rho, &
               cell, pw_env%cube_info(1), eps_rho_rspace, ga_gb_function=FUNC_AB, &
               use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, lmax_global=0)
         ENDDO
      ENDIF

      DEALLOCATE (pab, cores)

      CALL pw_pool_create_pw(auxbas_pw_pool, rhoc_r%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL rs_pw_transfer(rs_rho, rhoc_r%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL pw_transfer(rhoc_r%pw, rho_resp%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhoc_r%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_resp_all

! **************************************************************************************************
!> \brief computes the density corresponding to a given density matrix on the grid
!> \param matrix_p ...
!> \param matrix_p_kp ...
!> \param rho ...
!> \param rho_gspace ...
!> \param total_rho ...
!> \param ks_env ...
!> \param soft_valid ...
!> \param compute_tau ...
!> \param compute_grad ...
!> \param basis_type ...
!> \param der_type ...
!> \param idir ...
!> \param task_list_external ...
!> \param pw_env_external ...
!> \par History
!>      IAB (15-Feb-2010): Added OpenMP parallelisation to task loop
!>                         (c) The Numerical Algorithms Group (NAG) Ltd, 2010 on behalf of the HECToR project
!> \note
!>      both rho and rho_gspace contain the new rho
!>      (in real and g-space respectively)
! **************************************************************************************************
   SUBROUTINE calculate_rho_elec(matrix_p, matrix_p_kp, rho, rho_gspace, total_rho, &
                                 ks_env, soft_valid, compute_tau, compute_grad, &
                                 basis_type, der_type, idir, task_list_external, pw_env_external)

      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: matrix_p_kp
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho, rho_gspace
      REAL(KIND=dp), INTENT(OUT)                         :: total_rho
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: soft_valid, compute_tau, compute_grad
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: basis_type
      INTEGER, INTENT(IN), OPTIONAL                      :: der_type, idir
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list_external
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env_external

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_elec', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_basis_type
      INTEGER :: bcol, brow, ga_gb_function, handle, iatom, iatom_old, igrid_level, &
         igrid_level_dummy, ikind, ikind_old, img, img_old, ipair, ipgf, iset, iset_old, itask, &
         ithread, jatom, jatom_old, jkind, jkind_old, jpgf, jset, jset_old, lb, lbr, lbw, &
         lmax_global, maxco, maxpgf, maxset, maxsgf, maxsgf_set, my_der_type, my_idir, n, na1, &
         na2, natoms, nb1, nb2, nblock, ncoa, ncob, nimages, nr, nrlevel, nseta, nsetb, ntasks, &
         nthread, nxy, nz, nzsize, sgfa, sgfb, ub
      INTEGER(kind=int_8), DIMENSION(:), POINTER         :: atom_pair_recv, atom_pair_send
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, mylmax, &
                                                            npgfa, npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL :: atom_pair_changed, distributed_rs_grids, do_kp, found, map_consistent, &
         my_compute_grad, my_compute_tau, my_soft, use_subpatch
      REAL(KIND=dp)                                      :: eps_rho_rspace, rab2, scale, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rab_inv, rb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dist_ab, p_block, pab, sphi_a, sphi_b, &
                                                            work, zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pabt, workt
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: deltap
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(lgrid_type), POINTER                          :: lgrid
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho
      TYPE(task_list_type), POINTER                      :: task_list, task_list_soft

      CALL timeset(routineN, handle)

      IF(.NOT.(PRESENT(matrix_p) .OR. PRESENT(matrix_p_kp)))CALL cp__a("qs_collocate_density.F",1491)
      do_kp = PRESENT(matrix_p_kp)

      NULLIFY (qs_kind, cell, dft_control, orb_basis_set, deltap, &
               qs_kind_set, particle_set, rs_rho, pw_env, rs_descs, &
               dist_ab, la_max, la_min, &
               lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb, p_block, &
               sphi_a, sphi_b, zeta, zetb, first_sgfa, first_sgfb, tasks, pabt, &
               workt, lgrid, mylmax)

      debug_count = debug_count+1

      ! by default, the full density is calculated
      my_soft = .FALSE.
      IF (PRESENT(soft_valid)) my_soft = soft_valid

      ! by default, do not compute the kinetic energy density (tau)
      ! if compute_tau, all grids referening to rho are actually tau
      IF (PRESENT(compute_tau)) THEN
         my_compute_tau = compute_tau
      ELSE
         my_compute_tau = .FALSE.
      ENDIF

      IF (PRESENT(compute_grad)) THEN
         my_compute_grad = compute_grad
      ELSE
         my_compute_grad = .FALSE.
      ENDIF

      my_der_type = 0
      my_idir = 0
      IF (PRESENT(der_type)) THEN
         my_der_type = der_type
         ga_gb_function = FUNC_DER(my_der_type)
      ELSE IF (my_compute_tau) THEN
         ga_gb_function = FUNC_DADB
      ELSE IF (my_compute_grad) THEN
         ga_gb_function = FUNC_DABpADB
      ELSE
         ga_gb_function = FUNC_AB
      ENDIF
      IF (PRESENT(idir)) my_idir = idir

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      CALL get_ks_env(ks_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env)

      SELECT CASE (my_basis_type)
      CASE ("ORB")
         CALL get_ks_env(ks_env, &
                         task_list=task_list, &
                         task_list_soft=task_list_soft)
      CASE ("AUX_FIT")
         CALL get_ks_env(ks_env, &
                         task_list_aux_fit=task_list, &
                         task_list_soft=task_list_soft)
      END SELECT

      nimages = dft_control%nimages
      IF(.NOT.(nimages == 1 .OR. do_kp))CALL cp__a("qs_collocate_density.F",1560)

      IF (PRESENT(pw_env_external)) pw_env => pw_env_external
      IF (PRESENT(task_list_external)) task_list => task_list_external

      ! *** assign from pw_env
      gridlevel_info => pw_env%gridlevel_info
      cube_info => pw_env%cube_info

      !   *** Allocate work storage ***
      nthread = 1
!$    nthread = omp_get_max_threads()
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxsgf=maxsgf, &
                           maxsgf_set=maxsgf_set, &
                           basis_type=my_basis_type)
      CALL reallocate(pabt, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(workt, 1, maxco, 1, maxsgf_set, 0, nthread-1)

      ! find maximum numbers
      natoms = SIZE(particle_set)
      maxset = 0
      maxpgf = 0
      lmax_global = 0

      DO ikind = 1, SIZE(qs_kind_set)
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, &
                          softb=my_soft, &
                          basis_set=orb_basis_set, basis_type=my_basis_type)

         IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                npgf=npgfa, nset=nseta, lmax=mylmax)

         maxset = MAX(nseta, maxset)
         maxpgf = MAX(MAXVAL(npgfa), maxpgf)
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO

      ! Optionally increment lmax_global depending on which ga_gb_function
      ! will be used for collocation
      SELECT CASE (ga_gb_function)
      CASE (FUNC_DADB)
         lmax_global = lmax_global+1
      CASE (FUNC_ADBmDAB)
         lmax_global = lmax_global+1
      CASE (FUNC_DABpADB)
         lmax_global = lmax_global+1
      CASE (FUNC_DX, FUNC_DY, FUNC_DZ)
         lmax_global = lmax_global+1
      CASE (FUNC_DXDY, FUNC_DYDZ, FUNC_DZDX)
         lmax_global = lmax_global+2
      CASE (FUNC_DXDX, FUNC_DYDY, FUNC_DZDZ)
         lmax_global = lmax_global+2
      CASE (FUNC_ARDBmDARB)
         lmax_global = lmax_global+2
      CASE (FUNC_ARB)
         lmax_global = lmax_global+1
      CASE (FUNC_AB)
         lmax_global = lmax_global
      CASE DEFAULT
         CALL cp__b("qs_collocate_density.F",1624,"")
      END SELECT

      ! get the task lists
      IF (my_soft) task_list => task_list_soft
      IF(.NOT.(ASSOCIATED(task_list)))CALL cp__a("qs_collocate_density.F",1629)
      tasks => task_list%tasks
      dist_ab => task_list%dist_ab
      atom_pair_send => task_list%atom_pair_send
      atom_pair_recv => task_list%atom_pair_recv
      ntasks = task_list%ntasks

      ! *** set up the pw multi-grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_collocate_density.F",1637)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, rs_grids=rs_rho, lgrid=lgrid)
      ! *** set up the rs multi-grids
      IF(.NOT.(ASSOCIATED(rs_rho)))CALL cp__a("qs_collocate_density.F",1640)
      IF(.NOT.(ASSOCIATED(rs_descs)))CALL cp__a("qs_collocate_density.F",1641)
      IF(.NOT.(ASSOCIATED(lgrid)))CALL cp__a("qs_collocate_density.F",1642)
      distributed_rs_grids = .FALSE.

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_grid_retain(rs_rho(igrid_level)%rs_grid)
         CALL rs_grid_zero(rs_rho(igrid_level)%rs_grid)
         IF (rs_rho(igrid_level)%rs_grid%desc%distributed) THEN
            distributed_rs_grids = .TRUE.
         ENDIF
      END DO

      IF (nthread > 1) THEN
         IF (.NOT. ASSOCIATED(lgrid%r)) THEN
            CALL lgrid_allocate_grid(lgrid, nthread)
         ENDIF
      END IF

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      map_consistent = dft_control%qs_control%map_consistent

      !   *** Initialize working density matrix ***
      ! distributed rs grids require a matrix that will be changed
      ! whereas this is not the case for replicated grids
      NULLIFY (deltap)
      CALL cp_dbcsr_allocate_matrix_set(deltap, nimages)
      IF (distributed_rs_grids) THEN
         DO img = 1, nimages
            ALLOCATE (deltap(img)%matrix)
            CALL cp_dbcsr_init(deltap(img)%matrix)
         END DO
         ! this matrix has no strict sparsity pattern in parallel
         ! deltap%sparsity_id=-1
         IF (do_kp) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_copy(deltap(img)%matrix, matrix_p_kp(img)%matrix, &
                                  name="DeltaP")
            END DO
         ELSE
            CALL cp_dbcsr_copy(deltap(1)%matrix, matrix_p, name="DeltaP")
         END IF
      ELSE
         IF (do_kp) THEN
            DO img = 1, nimages
               deltap(img)%matrix => matrix_p_kp(img)%matrix
            END DO
         ELSE
            deltap(1)%matrix => matrix_p
         END IF
      ENDIF

      ! distribute the matrix
      IF (distributed_rs_grids) THEN
         CALL rs_distribute_matrix(rs_descs, deltap, atom_pair_send, atom_pair_recv, &
                                   natoms, nimages, scatter=.TRUE.)
      ENDIF

      ! map all tasks on the grids

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          SHARED(ntasks,tasks,nimages,natoms,maxset,maxpgf,particle_set,pabt,workt), &
!$OMP          SHARED(my_basis_type,my_soft,deltap,maxco,dist_ab,ncoset,nthread), &
!$OMP          SHARED(cell,cube_info,eps_rho_rspace,ga_gb_function, my_idir,map_consistent), &
!$OMP          SHARED(rs_rho,lgrid,gridlevel_info,task_list,qs_kind_set,lmax_global), &
!$OMP          PRIVATE(igrid_level,iatom,jatom,iset,jset,ipgf,jpgf,ikind,jkind,pab,work), &
!$OMP          PRIVATE(img,img_old,iatom_old,jatom_old,iset_old,jset_old,ikind_old,jkind_old), &
!$OMP          PRIVATE(brow,bcol,qs_kind,orb_basis_set,first_sgfa,la_max,la_min), &
!$OMP          PRIVATE(npgfa,nseta,nsgfa,sphi_a,zeta,first_sgfb,lb_max,lb_min,npgfb), &
!$OMP          PRIVATE(nsetb,nsgfb,sphi_b,zetb,p_block,found), &
!$OMP          PRIVATE(atom_pair_changed,ncoa,sgfa,ncob,sgfb,rab,rab2,ra,rb,zetp), &
!$OMP          PRIVATE(na1,na2,nb1,nb2,scale,use_subpatch,rab_inv,ithread,lb,ub,n), &
!$OMP          PRIVATE(itask,nz,nxy,nzsize,nrlevel,nblock,lbw,lbr,nr,igrid_level_dummy)

      ithread = 0
!$    ithread = omp_get_thread_num()
      pab => pabt(:, :, ithread)
      work => workt(:, :, ithread)

      iatom_old = -1; jatom_old = -1; iset_old = -1; jset_old = -1
      ikind_old = -1; jkind_old = -1; img_old = -1

      ! Loop over each gridlevel first, then loop and load balance over atom pairs
      ! We only need a single lgrid, which we sum back onto the rsgrid after each
      ! grid level is completed
      loop_gridlevels: DO igrid_level = 1, gridlevel_info%ngrid_levels

         ! Only zero the region of the lgrid required for this grid level
         IF (nthread > 1) THEN
            lb = ithread*lgrid%ldim+1
            ub = ithread*lgrid%ldim+rs_rho(igrid_level)%rs_grid%ngpts_local
            lgrid%r(lb:ub) = 0._dp
         END IF
!$OMP DO schedule(dynamic, MAX(1,task_list%npairs(igrid_level)/(nthread*50)))
         loop_pairs: DO ipair = 1, task_list%npairs(igrid_level)
         loop_tasks: DO itask = task_list%taskstart(ipair, igrid_level), task_list%taskstop(ipair, igrid_level)
            !decode the atom pair and basis info (igrid_level_dummy equals do loop variable by construction).
            CALL int2pair(tasks(3, itask), igrid_level_dummy, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natoms, maxset, maxpgf)
            ikind = particle_set(iatom)%atomic_kind%kind_number
            jkind = particle_set(jatom)%atomic_kind%kind_number

            IF (iatom .NE. iatom_old .OR. jatom .NE. jatom_old .OR. img .NE. img_old) THEN

               IF (iatom .NE. iatom_old) ra(:) = pbc(particle_set(iatom)%r, cell)

               IF (iatom <= jatom) THEN
                  brow = iatom
                  bcol = jatom
               ELSE
                  brow = jatom
                  bcol = iatom
               END IF

               IF (ikind .NE. ikind_old) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), softb=my_soft, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfa, &
                                         lmax=la_max, &
                                         lmin=la_min, &
                                         npgf=npgfa, &
                                         nset=nseta, &
                                         nsgf_set=nsgfa, &
                                         sphi=sphi_a, &
                                         zet=zeta)
               ENDIF

               IF (jkind .NE. jkind_old) THEN
                  CALL get_qs_kind(qs_kind_set(jkind), softb=my_soft, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfb, &
                                         lmax=lb_max, &
                                         lmin=lb_min, &
                                         npgf=npgfb, &
                                         nset=nsetb, &
                                         nsgf_set=nsgfb, &
                                         sphi=sphi_b, &
                                         zet=zetb)
               ENDIF

               CALL cp_dbcsr_get_block_p(matrix=deltap(img)%matrix, &
                                         row=brow, col=bcol, BLOCK=p_block, found=found)
               IF(.NOT.(found))CALL cp__a("qs_collocate_density.F",1784)

               iatom_old = iatom
               jatom_old = jatom
               ikind_old = ikind
               jkind_old = jkind
               img_old = img
               atom_pair_changed = .TRUE.

            ELSE

               atom_pair_changed = .FALSE.

            ENDIF

            IF (atom_pair_changed .OR. iset_old .NE. iset .OR. jset_old .NE. jset) THEN
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (iatom <= jatom) THEN
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             p_block(sgfa, sgfb), SIZE(p_block, 1), &
                             0.0_dp, work(1, 1), maxco)
                  CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                             1.0_dp, work(1, 1), maxco, &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, pab(1, 1), maxco)
               ELSE
                  CALL dgemm("N", "N", ncob, nsgfa(iset), nsgfb(jset), &
                             1.0_dp, sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             p_block(sgfb, sgfa), SIZE(p_block, 1), &
                             0.0_dp, work(1, 1), maxco)
                  CALL dgemm("N", "T", ncob, ncoa, nsgfa(iset), &
                             1.0_dp, work(1, 1), maxco, &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             0.0_dp, pab(1, 1), maxco)
               END IF

               iset_old = iset
               jset_old = jset
            ENDIF

            rab(:) = dist_ab(:, itask)
            rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
            rb(:) = ra(:)+rab(:)
            zetp = zeta(ipgf, iset)+zetb(jpgf, jset)

            na1 = (ipgf-1)*ncoset(la_max(iset))+1
            na2 = ipgf*ncoset(la_max(iset))
            nb1 = (jpgf-1)*ncoset(lb_max(jset))+1
            nb2 = jpgf*ncoset(lb_max(jset))

            ! takes the density matrix symmetry in account, i.e. off-diagonal blocks need to be mapped 'twice'
            IF (iatom == jatom) THEN
               scale = 1.0_dp
            ELSE
               scale = 2.0_dp
            END IF

            ! check whether we need to use fawzi's generalised collocation scheme
            IF (rs_rho(igrid_level)%rs_grid%desc%distributed) THEN
               !tasks(4,:) is 0 for replicated, 1 for distributed 2 for exceptional distributed tasks
               IF (tasks(4, itask) .EQ. 2) THEN
                  use_subpatch = .TRUE.
               ELSE
                  use_subpatch = .FALSE.
               ENDIF
            ELSE
               use_subpatch = .FALSE.
            ENDIF

            IF (nthread > 1) THEN
               IF (iatom <= jatom) THEN
                  CALL collocate_pgf_product_rspace( &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     ra, rab, rab2, scale, pab, na1-1, nb1-1, &
                     rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                     eps_rho_rspace, &
                     ga_gb_function=ga_gb_function, &
                     idir=my_idir, &
                     lgrid=lgrid, ithread=ithread, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, &
                     subpatch_pattern=tasks(6, itask), lmax_global=lmax_global)
               ELSE
                  rab_inv = -rab
                  CALL collocate_pgf_product_rspace( &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     rb, rab_inv, rab2, scale, pab, nb1-1, na1-1, &
                     rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                     eps_rho_rspace, &
                     ga_gb_function=ga_gb_function, &
                     idir=my_idir, &
                     lgrid=lgrid, ithread=ithread, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, &
                     subpatch_pattern=tasks(6, itask), lmax_global=lmax_global)
               END IF
            ELSE
               IF (iatom <= jatom) THEN
                  CALL collocate_pgf_product_rspace( &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     ra, rab, rab2, scale, pab, na1-1, nb1-1, &
                     rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                     eps_rho_rspace, &
                     ga_gb_function=ga_gb_function, &
                     idir=my_idir, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, &
                     subpatch_pattern=tasks(6, itask), lmax_global=lmax_global)
               ELSE
                  rab_inv = -rab
                  CALL collocate_pgf_product_rspace( &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     rb, rab_inv, rab2, scale, pab, nb1-1, na1-1, &
                     rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                     eps_rho_rspace, &
                     ga_gb_function=ga_gb_function, &
                     idir=my_idir, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, &
                     subpatch_pattern=tasks(6, itask), lmax_global=lmax_global)
               END IF
            END IF
         END DO loop_tasks
         END DO loop_pairs
!$OMP END DO

         ! Now sum the thread-local grids back into the rs_grid
         ! (in parallel, each thread writes to a section of the rs_grid at a time)
         IF (nthread > 1) THEN
            nz = (1+rs_rho(igrid_level)%rs_grid%ub_local(3) &
                  -rs_rho(igrid_level)%rs_grid%lb_local(3))
            nxy = (1+rs_rho(igrid_level)%rs_grid%ub_local(1) &
                   -rs_rho(igrid_level)%rs_grid%lb_local(1)) &
                  *(1+rs_rho(igrid_level)%rs_grid%ub_local(2) &
                    -rs_rho(igrid_level)%rs_grid%lb_local(2))

            ! Work out the number of tree levels, and start the reduction

            nrlevel = CEILING(LOG10(1.0_dp*nthread)/LOG10(2.0_dp))

            DO nr = 1, nrlevel
               nblock = 2**nr
               nzsize = MIN(nblock, nthread-(ithread/nblock)*nblock)
               itask = MODULO(ithread, nblock)

               lb = (nz*itask)/nzsize
               ub = (nz*(itask+1))/nzsize-1
               n = (ithread/nblock)*nblock
               lbw = 1+n*lgrid%ldim+nxy*lb

               n = n+nblock/2
               IF (n < nthread) THEN
                  lbr = 1+n*lgrid%ldim+nxy*lb

                  CALL daxpy(nxy*(1+ub-lb), 1.0_dp, lgrid%r(lbr), 1, lgrid%r(lbw), 1)
               END IF
!$OMP BARRIER
            END DO

            ! Copy final result from first local grid to rs_grid
            lb = (nz*ithread)/nthread
            ub = (nz*(ithread+1))/nthread-1
            nzsize = 1+(ub-lb)

            lb = lb+rs_rho(igrid_level)%rs_grid%lb_local(3)
            ub = ub+rs_rho(igrid_level)%rs_grid%lb_local(3)
            lbr = 1+nxy*(nz*ithread/nthread)

            CALL daxpy(nxy*nzsize, 1.0_dp, lgrid%r(lbr), 1, &
                       rs_rho(igrid_level)%rs_grid%r(:, :, lb:ub), 1)
!$OMP BARRIER
         END IF
      END DO loop_gridlevels
!$OMP END PARALLEL

      !   *** Release work storage ***

      IF (distributed_rs_grids) THEN
         CALL cp_dbcsr_deallocate_matrix_set(deltap)
      ELSE
         DO img = 1, nimages
            NULLIFY (deltap(img)%matrix)
         END DO
         DEALLOCATE (deltap)
      ENDIF

      DEALLOCATE (pabt, workt)

      CALL density_rs2pw(pw_env, rs_rho, rho, rho_gspace)

      total_rho = pw_integrate_function(rho%pw, isign=-1)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_elec

! **************************************************************************************************
!> \brief computes the gradient of the density corresponding to a given
!>        density matrix on the grid
!> \param matrix_p ...
!> \param matrix_p_kp ...
!> \param drho ...
!> \param drho_gspace ...
!> \param qs_env ...
!> \param soft_valid ...
!> \param basis_type ...
!> \note  this is an alternative to calculate the gradient through FFTs
! **************************************************************************************************
   SUBROUTINE calculate_drho_elec(matrix_p, matrix_p_kp, drho, drho_gspace, qs_env, &
                                  soft_valid, basis_type)

      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: matrix_p_kp
      TYPE(pw_p_type), DIMENSION(:), INTENT(INOUT)       :: drho, drho_gspace
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: soft_valid
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: basis_type

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_drho_elec', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_basis_type
      INTEGER :: bcol, brow, handle, i, iatom, iatom_old, idir, igrid_level, ikind, ikind_old, &
         img, img_old, ipgf, iset, iset_old, itask, ithread, jatom, jatom_old, jkind, jkind_old, &
         jpgf, jset, jset_old, lmax_global, maxco, maxpgf, maxset, maxsgf, maxsgf_set, na1, na2, &
         natoms, nb1, nb2, ncoa, ncob, nimages, nseta, nsetb, ntasks, nthread, sgfa, sgfb
      INTEGER(kind=int_8), DIMENSION(:), POINTER         :: atom_pair_recv, atom_pair_send
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, mylmax, &
                                                            npgfa, npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: atom_pair_changed, distributed_rs_grids, &
                                                            do_kp, found, map_consistent, my_soft, &
                                                            use_subpatch
      REAL(KIND=dp)                                      :: eps_rho_rspace, rab2, scale, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rab_inv, rb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dist_ab, p_block, pab, sphi_a, sphi_b, &
                                                            work, zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pabt, workt
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: deltap
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho
      TYPE(task_list_type), POINTER                      :: task_list, task_list_soft

      CALL timeset(routineN, handle)

      IF(.NOT.(PRESENT(matrix_p) .OR. PRESENT(matrix_p_kp)))CALL cp__a("qs_collocate_density.F",2048)
      do_kp = PRESENT(matrix_p_kp)

      NULLIFY (qs_kind, cell, dft_control, orb_basis_set, deltap, &
               qs_kind_set, sab_orb, particle_set, rs_rho, pw_env, rs_descs, &
               dist_ab, la_max, la_min, &
               lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb, p_block, &
               sphi_a, sphi_b, zeta, zetb, first_sgfa, first_sgfb, tasks, pabt, &
               workt, mylmax)

      debug_count = debug_count+1

      ! by default, the full density is calculated
      my_soft = .FALSE.
      IF (PRESENT(soft_valid)) my_soft = soft_valid

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      sab_orb=sab_orb, &
                      pw_env=pw_env)

      SELECT CASE (my_basis_type)
      CASE ("ORB")
         CALL get_qs_env(qs_env=qs_env, &
                         task_list=task_list, &
                         task_list_soft=task_list_soft)
      CASE ("AUX_FIT")
         CALL get_qs_env(qs_env=qs_env, &
                         task_list_aux_fit=task_list, &
                         task_list_soft=task_list_soft)
      END SELECT

      ! *** assign from pw_env
      gridlevel_info => pw_env%gridlevel_info
      cube_info => pw_env%cube_info

      !   *** Allocate work storage ***
      nthread = 1
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxsgf=maxsgf, &
                           maxsgf_set=maxsgf_set, &
                           basis_type=my_basis_type)
      CALL reallocate(pabt, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(workt, 1, maxco, 1, maxsgf_set, 0, nthread-1)

      ! find maximum numbers
      nimages = dft_control%nimages
      IF(.NOT.(nimages == 1 .OR. do_kp))CALL cp__a("qs_collocate_density.F",2105)

      natoms = SIZE(particle_set)
      maxset = 0
      maxpgf = 0
      lmax_global = 0

      DO ikind = 1, SIZE(qs_kind_set)
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, &
                          softb=my_soft, &
                          basis_set=orb_basis_set, basis_type=my_basis_type)

         IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                npgf=npgfa, nset=nseta, lmax=mylmax)

         maxset = MAX(nseta, maxset)
         maxpgf = MAX(MAXVAL(npgfa), maxpgf)
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO

      !update lmax_global since ga_gb_function=FUNC_DABpADB
      lmax_global = lmax_global+1

      ! get the task lists
      IF (my_soft) task_list => task_list_soft
      IF(.NOT.(ASSOCIATED(task_list)))CALL cp__a("qs_collocate_density.F",2133)
      tasks => task_list%tasks
      dist_ab => task_list%dist_ab
      atom_pair_send => task_list%atom_pair_send
      atom_pair_recv => task_list%atom_pair_recv
      ntasks = task_list%ntasks

      ! *** set up the rs multi-grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_collocate_density.F",2141)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, rs_grids=rs_rho)
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_grid_retain(rs_rho(igrid_level)%rs_grid)
         distributed_rs_grids = rs_rho(igrid_level)%rs_grid%desc%distributed
      END DO

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      map_consistent = dft_control%qs_control%map_consistent

      !   *** Initialize working density matrix ***
      ! distributed rs grids require a matrix that will be changed
      ! whereas this is not the case for replicated grids
      ALLOCATE (deltap(nimages))
      IF (distributed_rs_grids) THEN
         DO img = 1, nimages
            CALL cp_dbcsr_init(deltap(img)%matrix)
         END DO
         ! this matrix has no strict sparsity pattern in parallel
         ! deltap%sparsity_id=-1
         IF (do_kp) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_copy(deltap(img)%matrix, matrix_p_kp(img)%matrix, &
                                  name="DeltaP")
            END DO
         ELSE
            CALL cp_dbcsr_copy(deltap(1)%matrix, matrix_p, name="DeltaP")
         END IF
      ELSE
         IF (do_kp) THEN
            DO img = 1, nimages
               deltap(img)%matrix => matrix_p_kp(img)%matrix
            END DO
         ELSE
            deltap(1)%matrix => matrix_p
         END IF
      ENDIF

      ! distribute the matrix
      IF (distributed_rs_grids) THEN
         CALL rs_distribute_matrix(rs_descs, deltap, atom_pair_send, atom_pair_recv, &
                                   natoms, nimages, scatter=.TRUE.)
      ENDIF

      ! map all tasks on the grids

      ithread = 0
      pab => pabt(:, :, ithread)
      work => workt(:, :, ithread)

      loop_xyz: DO idir = 1, 3

         DO igrid_level = 1, gridlevel_info%ngrid_levels
            CALL rs_grid_zero(rs_rho(igrid_level)%rs_grid)
         END DO

         iatom_old = -1; jatom_old = -1; iset_old = -1; jset_old = -1
         ikind_old = -1; jkind_old = -1; img_old = -1
         loop_tasks: DO itask = 1, ntasks

            !decode the atom pair and basis info
            CALL int2pair(tasks(3, itask), igrid_level, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natoms, maxset, maxpgf)

            ikind = particle_set(iatom)%atomic_kind%kind_number
            jkind = particle_set(jatom)%atomic_kind%kind_number

            IF (iatom .NE. iatom_old .OR. jatom .NE. jatom_old .OR. img .NE. img_old) THEN

               IF (iatom .NE. iatom_old) ra(:) = pbc(particle_set(iatom)%r, cell)

               IF (iatom <= jatom) THEN
                  brow = iatom
                  bcol = jatom
               ELSE
                  brow = jatom
                  bcol = iatom
               END IF

               IF (ikind .NE. ikind_old) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), softb=my_soft, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfa, &
                                         lmax=la_max, &
                                         lmin=la_min, &
                                         npgf=npgfa, &
                                         nset=nseta, &
                                         nsgf_set=nsgfa, &
                                         sphi=sphi_a, &
                                         zet=zeta)
               ENDIF

               IF (jkind .NE. jkind_old) THEN
                  CALL get_qs_kind(qs_kind_set(jkind), softb=my_soft, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfb, &
                                         lmax=lb_max, &
                                         lmin=lb_min, &
                                         npgf=npgfb, &
                                         nset=nsetb, &
                                         nsgf_set=nsgfb, &
                                         sphi=sphi_b, &
                                         zet=zetb)
               ENDIF

               CALL cp_dbcsr_get_block_p(matrix=deltap(img)%matrix, &
                                         row=brow, col=bcol, BLOCK=p_block, found=found)
               IF(.NOT.(found))CALL cp__a("qs_collocate_density.F",2250)

               iatom_old = iatom
               jatom_old = jatom
               ikind_old = ikind
               jkind_old = jkind
               img_old = img
               atom_pair_changed = .TRUE.

            ELSE

               atom_pair_changed = .FALSE.

            ENDIF

            IF (atom_pair_changed .OR. iset_old .NE. iset .OR. jset_old .NE. jset) THEN

               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (iatom <= jatom) THEN
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             p_block(sgfa, sgfb), SIZE(p_block, 1), &
                             0.0_dp, work(1, 1), maxco)
                  CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                             1.0_dp, work(1, 1), maxco, &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, pab(1, 1), maxco)
               ELSE
                  CALL dgemm("N", "N", ncob, nsgfa(iset), nsgfb(jset), &
                             1.0_dp, sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             p_block(sgfb, sgfa), SIZE(p_block, 1), &
                             0.0_dp, work(1, 1), maxco)
                  CALL dgemm("N", "T", ncob, ncoa, nsgfa(iset), &
                             1.0_dp, work(1, 1), maxco, &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             0.0_dp, pab(1, 1), maxco)
               END IF

               iset_old = iset
               jset_old = jset

            ENDIF

            rab(:) = dist_ab(:, itask)
            rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
            rb(:) = ra(:)+rab(:)
            zetp = zeta(ipgf, iset)+zetb(jpgf, jset)

            na1 = (ipgf-1)*ncoset(la_max(iset))+1
            na2 = ipgf*ncoset(la_max(iset))
            nb1 = (jpgf-1)*ncoset(lb_max(jset))+1
            nb2 = jpgf*ncoset(lb_max(jset))

            ! takes the density matrix symmetry in account, i.e. off-diagonal blocks need to be mapped 'twice'
            IF (iatom == jatom .AND. img == 1) THEN
               scale = 1.0_dp
            ELSE
               scale = 2.0_dp
            END IF

            ! check whether we need to use fawzi's generalised collocation scheme
            IF (rs_rho(igrid_level)%rs_grid%desc%distributed) THEN
               !tasks(4,:) is 0 for replicated, 1 for distributed 2 for exceptional distributed tasks
               IF (tasks(4, itask) .EQ. 2) THEN
                  use_subpatch = .TRUE.
               ELSE
                  use_subpatch = .FALSE.
               ENDIF
            ELSE
               use_subpatch = .FALSE.
            ENDIF
            IF (iatom <= jatom) THEN
               CALL collocate_pgf_product_rspace( &
                  la_max(iset), zeta(ipgf, iset), la_min(iset), &
                  lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                  ra, rab, rab2, scale, pab, na1-1, nb1-1, &
                  rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                  eps_rho_rspace, &
                  ga_gb_function=FUNC_DABpADB, &
                  idir=idir, &
                  map_consistent=map_consistent, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask), &
                  lmax_global=lmax_global)
            ELSE
               rab_inv = -rab
               CALL collocate_pgf_product_rspace( &
                  lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                  la_max(iset), zeta(ipgf, iset), la_min(iset), &
                  rb, rab_inv, rab2, scale, pab, nb1-1, na1-1, &
                  rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                  eps_rho_rspace, &
                  ga_gb_function=FUNC_DABpADB, &
                  idir=idir, &
                  map_consistent=map_consistent, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask), &
                  lmax_global=lmax_global)
            END IF

         END DO loop_tasks

         CALL density_rs2pw_basic(pw_env, rs_rho, drho(idir), drho_gspace(idir))

      END DO loop_xyz

      !   *** Release work storage ***
      IF (ASSOCIATED(rs_rho)) THEN
         DO i = 1, SIZE(rs_rho)
            CALL rs_grid_release(rs_rho(i)%rs_grid)
         END DO
      END IF

      IF (distributed_rs_grids) THEN
         CALL cp_dbcsr_deallocate_matrix_set(deltap)
      ELSE
         DO img = 1, nimages
            NULLIFY (deltap(img)%matrix)
         END DO
         DEALLOCATE (deltap)
      ENDIF

      DEALLOCATE (pabt, workt)

      CALL timestop(handle)

   END SUBROUTINE calculate_drho_elec

! **************************************************************************************************
!> \brief maps a given wavefunction on the grid
!> \param mo_vectors ...
!> \param ivector ...
!> \param rho ...
!> \param rho_gspace ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param dft_control ...
!> \param particle_set ...
!> \param pw_env ...
!> \param basis_type ...
!> \param external_vector ...
!> \par History
!>      08.2002 created [Joost VandeVondele]
!>      03.2006 made independent of qs_env [Joost VandeVondele]
!> \note
!>      modified calculate_rho_elec, should write the wavefunction represented by
!>      it's presumably dominated by the FFT and the rs->pw and back routines
!>
!>    BUGS ???
!>      does it take correctly the periodic images of the basis functions into account
!>      i.e. is only correct if the basis functions are localised enough to be just in 1 cell ?
! **************************************************************************************************
   SUBROUTINE calculate_wavefunction(mo_vectors, ivector, rho, rho_gspace, &
                                     atomic_kind_set, qs_kind_set, cell, dft_control, particle_set, &
                                     pw_env, basis_type, external_vector)

      TYPE(cp_fm_type), POINTER                          :: mo_vectors
      INTEGER                                            :: ivector
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho, rho_gspace
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: basis_type
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: external_vector

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_wavefunction', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_basis_type
      INTEGER :: dir, group, group_size, handle, i, iatom, igrid_level, ikind, ipgf, iset, &
         lmax_global, maxco, maxsgf_set, my_pos, na1, na2, nao, natom, ncoa, nseta, offset, sgfa, &
         stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: where_is_the_point
      INTEGER, DIMENSION(3)                              :: lb, location, tp, ub
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, mylmax, npgfa, nsgfa
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: local, map_it_here
      REAL(KIND=dp)                                      :: dab, eps_rho_rspace, rab2, scale, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvector
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab, sphi_a, work, zeta
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: mgrid_gspace, mgrid_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(REALSPACE_GRID_P_TYPE), DIMENSION(:), POINTER :: rs_rho

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      CALL timeset(routineN, handle)

      NULLIFY (eigenvector, orb_basis_set, &
               pab, work, la_max, la_min, &
               npgfa, nsgfa, &
               sphi_a, zeta, first_sgfa, &
               rs_rho, pw_pools, mgrid_rspace, mgrid_gspace, mylmax)

      IF (PRESENT(external_vector)) THEN
         nao = SIZE(external_vector)
         ALLOCATE (eigenvector(nao), STAT=stat)
         IF (stat .NE. 0) CALL stop_memory(routineN, moduleN, 2459, "eigenvector", dp_size*nao)
         eigenvector = external_vector
      ELSE
         CALL cp_fm_get_info(matrix=mo_vectors, nrow_global=nao)
         ALLOCATE (eigenvector(nao), STAT=stat)
         IF (stat .NE. 0) CALL stop_memory(routineN, moduleN, 2464, "eigenvector", dp_size*nao)
         DO i = 1, nao
            CALL cp_fm_get_element(mo_vectors, i, ivector, eigenvector(i), local)
         ENDDO
      ENDIF

      ! *** set up the pw multi-grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_collocate_density.F",2471)
      CALL pw_env_get(pw_env, rs_grids=rs_rho, pw_pools=pw_pools, &
                      cube_info=cube_info, gridlevel_info=gridlevel_info)

      CALL pw_pools_create_pws(pw_pools, mgrid_gspace, &
                               use_data=COMPLEXDATA1D, &
                               in_space=RECIPROCALSPACE)
      CALL pw_pools_create_pws(pw_pools, mgrid_rspace, &
                               use_data=REALDATA3D, &
                               in_space=REALSPACE)

      ! *** set up rs multi-grids
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_grid_retain(rs_rho(igrid_level)%rs_grid)
         CALL rs_grid_zero(rs_rho(igrid_level)%rs_grid)
      END DO

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
!   *** Allocate work storage ***
      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, &
                           maxco=maxco, &
                           maxsgf_set=maxsgf_set, &
                           basis_type=my_basis_type)

      ALLOCATE (pab(maxco, 1), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 2497, &
                                      "pab", maxco*dp_size)
      ALLOCATE (work(maxco, 1), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 2500, &
                                      "work", maxco*dp_size)

      offset = 0
      group = mgrid_rspace(1)%pw%pw_grid%para%group
      my_pos = mgrid_rspace(1)%pw%pw_grid%para%my_pos
      group_size = mgrid_rspace(1)%pw%pw_grid%para%group_size
      ALLOCATE (where_is_the_point(0:group_size-1))

      !loop over natom to find maximum value of la_max
      lmax_global = 0
      DO iatom = 1, natom
         ikind = particle_set(iatom)%atomic_kind%kind_number
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, basis_type=my_basis_type)
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, lmax=mylmax)
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO

      DO iatom = 1, natom
         ikind = particle_set(iatom)%atomic_kind%kind_number
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, basis_type=my_basis_type)
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                first_sgf=first_sgfa, &
                                lmax=la_max, &
                                lmin=la_min, &
                                npgf=npgfa, &
                                nset=nseta, &
                                nsgf_set=nsgfa, &
                                sphi=sphi_a, &
                                zet=zeta)
         ra(:) = pbc(particle_set(iatom)%r, cell)
         rab(:) = 0.0_dp
         rab2 = 0.0_dp
         dab = 0.0_dp

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO i = 1, nsgfa(iset)
               work(i, 1) = eigenvector(offset+i)
            ENDDO

            CALL dgemm("N", "N", ncoa, 1, nsgfa(iset), &
                       1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                       work(1, 1), SIZE(work, 1), &
                       0.0_dp, pab(1, 1), SIZE(pab, 1))

            DO ipgf = 1, npgfa(iset)

               na1 = (ipgf-1)*ncoset(la_max(iset))+1
               na2 = ipgf*ncoset(la_max(iset))

               scale = 1.0_dp
               zetp = zeta(ipgf, iset)
               igrid_level = gaussian_gridlevel(gridlevel_info, zetp)

               map_it_here = .FALSE.

               IF (.NOT. ALL(rs_rho(igrid_level)%rs_grid%desc%perd == 1)) THEN
                  DO dir = 1, 3
                     ! bounds of local grid (i.e. removing the 'wings'), if periodic
                     tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_rho(igrid_level)%rs_grid%desc%npts(dir))
                     tp(dir) = MODULO(tp(dir), rs_rho(igrid_level)%rs_grid%desc%npts(dir))
                     IF (rs_rho(igrid_level)%rs_grid%desc%perd(dir) .NE. 1) THEN
                        lb(dir) = rs_rho(igrid_level)%rs_grid%lb_local(dir)+rs_rho(igrid_level)%rs_grid%desc%border
                        ub(dir) = rs_rho(igrid_level)%rs_grid%ub_local(dir)-rs_rho(igrid_level)%rs_grid%desc%border
                     ELSE
                        lb(dir) = rs_rho(igrid_level)%rs_grid%lb_local(dir)
                        ub(dir) = rs_rho(igrid_level)%rs_grid%ub_local(dir)
                     ENDIF
                     ! distributed grid, only map if it is local to the grid
                     location(dir) = tp(dir)+rs_rho(igrid_level)%rs_grid%desc%lb(dir)
                  ENDDO
                  IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                      lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                      lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                     map_it_here = .TRUE.
                  ENDIF
               ELSE
                  ! not distributed, just a round-robin distribution over the full set of CPUs
                  IF (MODULO(offset, group_size) == my_pos) map_it_here = .TRUE.
               ENDIF

               IF (map_it_here) CALL collocate_pgf_product_rspace( &
                  la_max(iset), zeta(ipgf, iset), la_min(iset), &
                  0, 0.0_dp, 0, &
                  ra, rab, rab2, scale, pab, na1-1, 0, &
                  rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                  eps_rho_rspace, map_consistent=.TRUE., ga_gb_function=FUNC_AB, &
                  lmax_global=lmax_global)

            END DO

            offset = offset+nsgfa(iset)

         END DO

      END DO

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL rs_pw_transfer(rs_rho(igrid_level)%rs_grid, &
                             mgrid_rspace(igrid_level)%pw, rs2pw)
         CALL rs_grid_release(rs_rho(igrid_level)%rs_grid)
      ENDDO

      CALL pw_zero(rho_gspace%pw)
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         CALL pw_transfer(mgrid_rspace(igrid_level)%pw, &
                          mgrid_gspace(igrid_level)%pw)
         CALL pw_axpy(mgrid_gspace(igrid_level)%pw, rho_gspace%pw)
      END DO

      CALL pw_transfer(rho_gspace%pw, rho%pw)

      ! Release work storage
      DEALLOCATE (eigenvector)

      DEALLOCATE (pab)

      DEALLOCATE (work)

      ! give back the pw multi-grids
      CALL pw_pools_give_back_pws(pw_pools, mgrid_gspace)
      CALL pw_pools_give_back_pws(pw_pools, mgrid_rspace)

      CALL timestop(handle)

   END SUBROUTINE calculate_wavefunction

! **************************************************************************************************
!> \brief low level collocation of primitive gaussian functions
!> \param la_max ...
!> \param zeta ...
!> \param la_min ...
!> \param lb_max ...
!> \param zetb ...
!> \param lb_min ...
!> \param ra ...
!> \param rab ...
!> \param rab2 ...
!> \param scale ...
!> \param pab ...
!> \param o1 ...
!> \param o2 ...
!> \param rsgrid ...
!> \param cell ...
!> \param cube_info ...
!> \param eps_rho_rspace ...
!> \param ga_gb_function ...
!> \param lgrid ...
!> \param ithread ...
!> \param map_consistent ...
!> \param collocate_rho0 ...
!> \param rpgf0_s ...
!> \param idir ...
!> \param ir ...
!> \param rsgauge ...
!> \param rsbuf ...
!> \param use_subpatch ...
!> \param subpatch_pattern ...
!> \param lmax_global Maximum possible value of lmax used to dimension arrays
! **************************************************************************************************
   SUBROUTINE collocate_pgf_product_rspace(la_max, zeta, la_min, &
                                           lb_max, zetb, lb_min, &
                                           ra, rab, rab2, scale, pab, o1, o2, &
                                           rsgrid, cell, cube_info, &
                                           eps_rho_rspace, ga_gb_function, &
                                           lgrid, ithread, &
                                           map_consistent, &
                                           collocate_rho0, &
                                           rpgf0_s, idir, ir, rsgauge, rsbuf, &
                                           use_subpatch, subpatch_pattern, &
                                           lmax_global)

      INTEGER, INTENT(IN)                      :: la_max
      REAL(KIND=dp), INTENT(IN)                :: zeta
      INTEGER, INTENT(IN)                      :: la_min, lb_max
      REAL(KIND=dp), INTENT(IN)                :: zetb
      INTEGER, INTENT(IN)                      :: lb_min
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: ra, rab
      REAL(KIND=dp), INTENT(IN)                :: rab2, scale
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: pab
      INTEGER, INTENT(IN)                      :: o1, o2
      TYPE(realspace_grid_type), POINTER       :: rsgrid
      TYPE(cell_type), POINTER                 :: cell
      TYPE(cube_info_type), INTENT(IN)         :: cube_info
      REAL(KIND=dp), INTENT(IN)                :: eps_rho_rspace
      INTEGER, INTENT(IN)                      :: ga_gb_function
      TYPE(lgrid_type), OPTIONAL               :: lgrid
      INTEGER, INTENT(IN), OPTIONAL            :: ithread
      LOGICAL, INTENT(IN), OPTIONAL            :: map_consistent, collocate_rho0
      REAL(dp), INTENT(IN), OPTIONAL           :: rpgf0_s
      INTEGER, INTENT(IN), OPTIONAL            :: idir, ir
      TYPE(realspace_grid_type), POINTER, OPTIONAL :: rsgauge, rsbuf
      LOGICAL, OPTIONAL                        :: use_subpatch
      INTEGER(KIND=int_8), OPTIONAL, INTENT(IN):: subpatch_pattern
      INTEGER, INTENT(IN)                      :: lmax_global

      CHARACTER(len=*), PARAMETER :: routineN = 'collocate_pgf_product_rspace', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: cmax, gridbounds(2, 3), i, ico, icoef, ider1, ider2, ig, ithread_l, &
                 jco, k, l, la_max_local, la_min_local, lb_max_local, lb_min_local, &
                 length, lxa, lxb, lxy, lxyz, lya, lyb, &
                 lza, lzb, o1_local, o2_local, offset, start
      INTEGER, DIMENSION(3)                    :: cubecenter, lb_cube, ng, &
                                                  ub_cube
      INTEGER, DIMENSION(:), POINTER           :: sphere_bounds
      LOGICAL                                  :: my_collocate_rho0, &
                                                  my_map_consistent, subpatch_collocate
      REAL(KIND=dp) :: a, b, binomial_k_lxa, binomial_l_lxb, cutoff, f, pg, &
                       prefactor, radius, rpg, zetp
      REAL(KIND=dp), DIMENSION(3)              :: dr, rap, rb, rbp, roffset, rp
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: pab_local
      REAL(KIND=dp), DIMENSION(:, :, :), &
         POINTER                                :: grid

      INTEGER :: lxp, lyp, lzp, lp, lxpm, iaxis
      INTEGER, ALLOCATABLE, DIMENSION(:, :) :: map
      REAL(kind=dp) :: p_ele
      REAL(kind=dp), DIMENSION(0:lmax_global*2, 0:lmax_global, 0:lmax_global, 3) :: alpha
      REAL(kind=dp), DIMENSION(((lmax_global*2+1)*(lmax_global*2+2)*(lmax_global*2+3))/6) :: coef_xyz
      REAL(kind=dp), DIMENSION(((lmax_global*2+1)*(lmax_global*2+2))/2) :: coef_xyt
      REAL(kind=dp), DIMENSION(0:lmax_global*2) :: coef_xtt

      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :, :) :: pol_z
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :, :) :: pol_y
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :) :: pol_x
      REAL(KIND=dp) :: t_exp_1, t_exp_2, t_exp_min_1, t_exp_min_2, t_exp_plus_1, t_exp_plus_2

      REAL(kind=dp), POINTER, DIMENSION(:, :, :) :: grid_tmp, gauge

      subpatch_collocate = .FALSE.

      IF (PRESENT(use_subpatch)) THEN
         IF (use_subpatch) THEN
            subpatch_collocate = .TRUE.
            IF(.NOT.(PRESENT(subpatch_pattern)))CALL cp__a("qs_collocate_density.F",2739)
         ENDIF
      ENDIF

      IF (PRESENT(ithread)) THEN
         ithread_l = ithread
      ELSE
         ithread_l = 0
      ENDIF

      ! use identical radii for integrate and collocate ?
      IF (PRESENT(map_consistent)) THEN
         my_map_consistent = map_consistent
      ELSE
         my_map_consistent = .FALSE.
      ENDIF

      IF (PRESENT(collocate_rho0) .AND. PRESENT(rpgf0_s)) THEN
         my_collocate_rho0 = collocate_rho0
      ELSE
         my_collocate_rho0 = .FALSE.
      END IF

      zetp = zeta+zetb
      f = zetb/zetp
      rap(:) = f*rab(:)
      rbp(:) = rap(:)-rab(:)
      rp(:) = ra(:)+rap(:)
      rb(:) = ra(:)+rab(:)

      IF (my_map_consistent) THEN
         cutoff = 1.0_dp
         prefactor = EXP(-zeta*f*rab2)
         radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, ra=ra, rb=rb, rp=rp, &
                                           zetp=zetp, eps=eps_rho_rspace, &
                                           prefactor=prefactor, cutoff=cutoff)
         prefactor = scale*EXP(-zeta*f*rab2)
      ELSE IF (my_collocate_rho0) THEN
         cutoff = 0.0_dp
         prefactor = 1.0_dp
         radius = rpgf0_s
      ELSE
         cutoff = 0.0_dp
         prefactor = scale*EXP(-zeta*f*rab2)
         radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, pab, o1, o2, ra, rb, rp, &
                                           zetp, eps_rho_rspace, prefactor, cutoff)
      ENDIF

      IF (radius .EQ. 0.0_dp) THEN
         RETURN
      END IF

      ! it's a choice to compute lX_min/max, pab here,
      ! this way we get the same radius as we use for the corresponding density
      SELECT CASE (ga_gb_function)
      CASE (FUNC_DADB)
         la_max_local = la_max+1
         la_min_local = MAX(la_min-1, 0)
         lb_max_local = lb_max+1
         lb_min_local = MAX(lb_min-1, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with 0.5 * (nabla pgf_a) . (nabla pgf_b)
         ! (ddx pgf_a ) (ddx pgf_b) = (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x})*(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x})
         ! cleaner would possibly be to touch pzyx directly (avoiding the following allocate)
         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb

                  ! this element of pab results in 12 elements of pab_local
                  CALL prepare_dadb(pab_local, pab, lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)

               ENDDO
               ENDDO
            ENDDO
            ENDDO
         ENDDO
         ENDDO
         o1_local = 0
         o2_local = 0
         pab_local = pab_local*0.5_dp
      CASE (FUNC_ADBmDAB)
         IF(.NOT.(PRESENT(idir)))CALL cp__a("qs_collocate_density.F",2825)
         la_max_local = la_max+1
         la_min_local = MAX(la_min-1, 0)
         lb_max_local = lb_max+1
         lb_min_local = MAX(lb_min-1, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         !    pgf_a (nabla_{idir} pgf_b) - (nabla_{idir} pgf_a) pgf_b
         ! ( pgf_a ) (ddx pgf_b) - (ddx pgf_a)( pgf_b ) =
         !          pgf_a *(lbx pgf_{b-1x} - 2*zetb*pgf_{b+1x}) -
         !                   (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x}) pgf_b

         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 4 elements of pab_local
                  CALL prepare_adb_m_dab(pab_local, pab, idir, &
                                         lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_DABpADB)
         IF(.NOT.(PRESENT(idir)))CALL cp__a("qs_collocate_density.F",2857)
         la_max_local = la_max+1
         la_min_local = MAX(la_min-1, 0)
         lb_max_local = lb_max+1
         lb_min_local = MAX(lb_min-1, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         !    pgf_a (nabla_{idir} pgf_b) + (nabla_{idir} pgf_a) pgf_b
         ! ( pgf_a ) (ddx pgf_b) + (ddx pgf_a)( pgf_b ) =
         !          pgf_a *(lbx pgf_{b-1x} + 2*zetb*pgf_{b+1x}) +
         !                   (lax pgf_{a-1x} + 2*zeta*pgf_{a+1x}) pgf_b

         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 4 elements of pab_local
                  CALL prepare_dab_p_adb(pab_local, pab, idir, &
                                         lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_DX, FUNC_DY, FUNC_DZ)
         ider1 = ga_gb_function-500
         la_max_local = la_max+1
         la_min_local = MAX(la_min-1, 0)
         lb_max_local = lb_max+1
         lb_min_local = MAX(lb_min-1, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         !   d_{ider1} pgf_a d_{ider1} pgf_b
         ! dx pgf_a dx pgf_b =
         !        (lax pgf_{a-1x})*(lbx pgf_{b-1x}) - 2*zetb*lax*pgf_{a-1x}*pgf_{b+1x} -
         !         lbx pgf_{b-1x}*2*zeta*pgf_{a+1x}+ 4*zeta*zetab*pgf_{a+1x}pgf_{b+1x}

         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 4 elements of pab_local
                  CALL prepare_dIadIb(pab_local, pab, ider1, &
                                      lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_DXDY, FUNC_DYDZ, FUNC_DZDX)
         ider1 = ga_gb_function-600
         ider2 = ga_gb_function-600+1
         IF (ider2 > 3) ider2 = ider1-2
         la_max_local = la_max+2
         la_min_local = MAX(la_min-2, 0)
         lb_max_local = lb_max+2
         lb_min_local = MAX(lb_min-2, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         !   d_{ider1} pgf_a d_{ider1} pgf_b
         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 16 elements of pab_local
                  CALL prepare_dijadijb(pab_local, pab, ider1, ider2, &
                                        lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_DXDX, FUNC_DYDY, FUNC_DZDZ)
         ider1 = ga_gb_function-603
         la_max_local = la_max+2
         la_min_local = MAX(la_min-2, 0)
         lb_max_local = lb_max+2
         lb_min_local = MAX(lb_min-2, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         !   dd_{ider1} pgf_a dd_{ider1} pgf_b

         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 9 elements of pab_local
                  CALL prepare_diiadiib(pab_local, pab, ider1, &
                                        lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_ARDBmDARB)
         IF(.NOT.(PRESENT(idir)))CALL cp__a("qs_collocate_density.F",2980)
         IF(.NOT.(PRESENT(ir)))CALL cp__a("qs_collocate_density.F",2981)
         la_max_local = la_max+1
         la_min_local = MAX(la_min-1, 0)
         lb_max_local = lb_max+2
         lb_min_local = MAX(lb_min-1, 0)
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         ! pgf_a (r-Rb)_{ir} (nabla_{idir} pgf_b) - (nabla_{idir} pgf_a) (r-Rb)_{ir}  pgf_b
         ! ( pgf_a )(r-Rb)_{ir} (ddx pgf_b) - (ddx pgf_a) (r-Rb)_{ir} ( pgf_b ) =
         !                        pgf_a *(lbx pgf_{b-1x+1ir} - 2*zetb*pgf_{b+1x+1ir}) -
         !                       (lax pgf_{a-1x} - 2*zeta*pgf_{a+1x}) pgf_{b+1ir}

         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb

                  ! this element of pab results in 4 elements of pab_local
                  CALL prepare_ardb_m_darb(pab_local, pab, idir, ir, &
                                           lxa, lya, lza, lxb, lyb, lzb, o1, o2, zeta, zetb)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_ARB)
         IF(.NOT.(PRESENT(ir)))CALL cp__a("qs_collocate_density.F",3014)
         la_max_local = la_max
         la_min_local = la_min
         lb_max_local = lb_max+1
         lb_min_local = lb_min
         ! create a new pab_local so that mapping pab_local with pgf_a pgf_b
         ! is equivalent to mapping pab with
         ! pgf_a (r-Rb)_{ir} pgf_b = pgf_a * pgf_{b+1ir}
         ALLOCATE (pab_local(ncoset(la_max_local), ncoset(lb_max_local)))
         pab_local = 0.0_dp
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  ! this element of pab results in 4 elements of pab_local
                  CALL prepare_arb(pab_local, pab, ir, lxa, lya, lza, lxb, lyb, lzb, o1, o2)
               END DO
               END DO
            END DO
            END DO
         END DO
         END DO
         o1_local = 0
         o2_local = 0
      CASE (FUNC_AB)
         la_max_local = la_max
         la_min_local = la_min
         lb_max_local = lb_max
         lb_min_local = lb_min
         pab_local => pab
         o1_local = o1
         o2_local = o2
      CASE DEFAULT
         IF(.NOT.(.FALSE.))CALL cp__a("qs_collocate_density.F",3049)
      END SELECT

      ng(:) = rsgrid%desc%npts(:)
      grid => rsgrid%r(:, :, :)
      gridbounds(1, 1) = LBOUND(GRID, 1)
      gridbounds(2, 1) = UBOUND(GRID, 1)
      gridbounds(1, 2) = LBOUND(GRID, 2)
      gridbounds(2, 2) = UBOUND(GRID, 2)
      gridbounds(1, 3) = LBOUND(GRID, 3)
      gridbounds(2, 3) = UBOUND(GRID, 3)

      IF (PRESENT(ir) .AND. PRESENT(rsgauge) .AND. PRESENT(rsbuf)) THEN
         grid => rsbuf%r(:, :, :)
         grid_tmp => rsgrid%r(:, :, :)
         gauge => rsgauge%r(:, :, :)
      ENDIF

!   *** initialise the coefficient matrix, we transform the sum
!
!   sum_{lxa,lya,lza,lxb,lyb,lzb} P_{lxa,lya,lza,lxb,lyb,lzb} *
!           (x-a_x)**lxa (y-a_y)**lya (z-a_z)**lza (x-b_x)**lxb (y-a_y)**lya (z-a_z)**lza
!
!   into
!
!   sum_{lxp,lyp,lzp} P_{lxp,lyp,lzp} (x-p_x)**lxp (y-p_y)**lyp (z-p_z)**lzp
!
!   where p is center of the product gaussian, and lp = la_max + lb_max
!   (current implementation is l**7)
!
      lp = la_max_local+lb_max_local
!
!   compute polynomial expansion coefs -> (x-a)**lxa (x-b)**lxb -> sum_{ls} alpha(ls,lxa,lxb,1)*(x-p)**ls
!
!   *** make the alpha matrix ***
      alpha(:, :, :, :) = 0.0_dp
      DO iaxis = 1, 3
      DO lxa = 0, la_max_local
      DO lxb = 0, lb_max_local
         binomial_k_lxa = 1.0_dp
         a = 1.0_dp
         DO k = 0, lxa
            binomial_l_lxb = 1.0_dp
            b = 1.0_dp
            DO l = 0, lxb
               alpha(lxa-l+lxb-k, lxa, lxb, iaxis) = alpha(lxa-l+lxb-k, lxa, lxb, iaxis)+ &
                                                     binomial_k_lxa*binomial_l_lxb*a*b
               binomial_l_lxb = binomial_l_lxb*REAL(lxb-l, dp)/REAL(l+1, dp)
               b = b*(rp(iaxis)-(ra(iaxis)+rab(iaxis)))
            ENDDO
            binomial_k_lxa = binomial_k_lxa*REAL(lxa-k, dp)/REAL(k+1, dp)
            a = a*(-ra(iaxis)+rp(iaxis))
         ENDDO
      ENDDO
      ENDDO
      ENDDO

!
!   compute P_{lxp,lyp,lzp} given P_{lxa,lya,lza,lxb,lyb,lzb} and alpha(ls,lxa,lxb,1)
!   use a three step procedure
!   we don't store zeros, so counting is done using lxyz,lxy in order to have contiguous memory access in collocate_fast.F
!
      lxyz = 0
      DO lzp = 0, lp
      DO lyp = 0, lp-lzp
      DO lxp = 0, lp-lzp-lyp
         lxyz = lxyz+1
         coef_xyz(lxyz) = 0.0_dp
      ENDDO
      ENDDO
      ENDDO
      DO lzb = 0, lb_max_local
      DO lza = 0, la_max_local
         lxy = 0
         DO lyp = 0, lp-lza-lzb
            DO lxp = 0, lp-lza-lzb-lyp
               lxy = lxy+1
               coef_xyt(lxy) = 0.0_dp
            ENDDO
            lxy = lxy+lza+lzb
         ENDDO
         DO lyb = 0, lb_max_local-lzb
         DO lya = 0, la_max_local-lza
            lxpm = (lb_max_local-lzb-lyb)+(la_max_local-lza-lya)
            coef_xtt(0:lxpm) = 0.0_dp
            DO lxb = MAX(lb_min_local-lzb-lyb, 0), lb_max_local-lzb-lyb
            DO lxa = MAX(la_min_local-lza-lya, 0), la_max_local-lza-lya
               ico = coset(lxa, lya, lza)
               jco = coset(lxb, lyb, lzb)
               p_ele = prefactor*pab_local(o1_local+ico, o2_local+jco)
               DO lxp = 0, lxa+lxb
                  coef_xtt(lxp) = coef_xtt(lxp)+p_ele*alpha(lxp, lxa, lxb, 1)
               ENDDO
            ENDDO
            ENDDO
            lxy = 0
            DO lyp = 0, lya+lyb
               DO lxp = 0, lp-lza-lzb-lya-lyb
                  lxy = lxy+1
                  coef_xyt(lxy) = coef_xyt(lxy)+alpha(lyp, lya, lyb, 2)*coef_xtt(lxp)
               ENDDO
               lxy = lxy+lza+lzb+lya+lyb-lyp
            ENDDO
         ENDDO
         ENDDO
         lxyz = 0
         DO lzp = 0, lza+lzb
            lxy = 0
            DO lyp = 0, lp-lza-lzb
               DO lxp = 0, lp-lza-lzb-lyp
                  lxy = lxy+1; lxyz = lxyz+1
                  coef_xyz(lxyz) = coef_xyz(lxyz)+alpha(lzp, lza, lzb, 3)*coef_xyt(lxy)
               ENDDO
               lxy = lxy+lza+lzb; lxyz = lxyz+lza+lzb-lzp
            ENDDO
            DO lyp = lp-lza-lzb+1, lp-lzp
               DO lxp = 0, lp-lyp-lzp
                  lxyz = lxyz+1
               ENDDO
            ENDDO
         ENDDO
      ENDDO
      ENDDO

      IF (subpatch_collocate) THEN
         CALL collocate_general_subpatch()
      ELSE
         IF (rsgrid%desc%orthorhombic) THEN
            CALL collocate_ortho()
            ! CALL collocate_general()
         ELSE
            CALL collocate_general_wings()
            !CALL collocate_general_opt()
         END IF
      END IF

      IF (ga_gb_function /= FUNC_AB) THEN
         DEALLOCATE (pab_local)
      ENDIF

   CONTAINS

      !
      ! this treats efficiently the orthogonal case
      !
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_ortho()

!   *** properties of the grid ***

         ! notice we're in the ortho case
         dr(1) = rsgrid%desc%dh(1, 1)
         dr(2) = rsgrid%desc%dh(2, 2)
         dr(3) = rsgrid%desc%dh(3, 3)

!   *** get the sub grid properties for the given radius ***
         CALL return_cube(cube_info, radius, lb_cube, ub_cube, sphere_bounds)
         cmax = MAXVAL(ub_cube)

!   *** position of the gaussian product
!
!   this is the actual definition of the position on the grid
!   i.e. a point rp(:) gets here grid coordinates
!   MODULO(rp(:)/dr(:),ng(:))+1
!   hence (0.0,0.0,0.0) in real space is rsgrid%lb on the rsgrid ((1,1,1) on grid)
!

         ALLOCATE (map(-cmax:cmax, 3))
         CALL compute_cube_center(cubecenter, rsgrid%desc, zeta, zetb, ra, rab)
         roffset(:) = rp(:)-REAL(cubecenter(:), dp)*dr(:)
!   *** a mapping so that the ig corresponds to the right grid point
         DO i = 1, 3
            IF (rsgrid%desc%perd(i) == 1) THEN
               start = lb_cube(i)
               DO
                  offset = MODULO(cubecenter(i)+start, ng(i))+1-start
                  length = MIN(ub_cube(i), ng(i)-offset)-start
                  DO ig = start, start+length
                     map(ig, i) = ig+offset
                  END DO
                  IF (start+length .GE. ub_cube(i)) EXIT
                  start = start+length+1
               END DO
            ELSE
               ! this takes partial grid + border regions into account
               offset = MODULO(cubecenter(i)+lb_cube(i)+rsgrid%desc%lb(i)-rsgrid%lb_local(i), ng(i))+1-lb_cube(i)
               ! check for out of bounds
               IF (ub_cube(i)+offset > UBOUND(grid, i) .OR. lb_cube(i)+offset < LBOUND(grid, i)) THEN
                  IF(.NOT.(.FALSE.))CALL cp__a("qs_collocate_density.F",3239)
               ENDIF
               DO ig = lb_cube(i), ub_cube(i)
                  map(ig, i) = ig+offset
               END DO
            END IF
         ENDDO
         ALLOCATE (pol_z(1:2, 0:lp, -cmax:0))
         ALLOCATE (pol_y(1:2, 0:lp, -cmax:0))
         ALLOCATE (pol_x(0:lp, -cmax:cmax))

         IF (PRESENT(ir) .AND. PRESENT(rsgauge)) CALL collocate_ortho_set_to_0()

!
!   compute the values of all (x-xp)**lp*exp(..)
!
! still requires the old trick:
!  new trick to avoid to many exps (reuse the result from the previous gridpoint):
!  exp( -a*(x+d)**2)=exp(-a*x**2)*(-2*a*x*d)*exp(-a*d**2)
!  exp(-2*a*(x+d)*d)=exp(-2*a*x*d)*exp(-2*a*d**2)

      iaxis=3
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(iaxis),-1
        rpg = REAL(ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_z(1,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_z(2,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO

      iaxis=2
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(iaxis),-1
        rpg = REAL(ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_y(1,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_y(2,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO

      iaxis=1
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(1),-1

        rpg = REAL(ig,dp)*dr(1) - roffset(1)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        !pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_x(icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(1) - roffset(1)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_x(icoef,1-ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO



         IF (PRESENT(lgrid)) THEN
            ig = lgrid%ldim*ithread_l+1
  SELECT CASE(lp)
 CASE(0)
 CALL collocate_core_0(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(1)
 CALL collocate_core_1(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(2)
 CALL collocate_core_2(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(3)
 CALL collocate_core_3(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(4)
 CALL collocate_core_4(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(5)
 CALL collocate_core_5(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(6)
 CALL collocate_core_6(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(7)
 CALL collocate_core_7(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(8)
 CALL collocate_core_8(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(9)
 CALL collocate_core_9(lgrid%r(ig),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
  CASE DEFAULT
 CALL collocate_core_default(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax),pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),lp,cmax,gridbounds(1,1))
  END SELECT
         ELSE
  SELECT CASE(lp)
 CASE(0)
 CALL collocate_core_0(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(1)
 CALL collocate_core_1(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(2)
 CALL collocate_core_2(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(3)
 CALL collocate_core_3(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(4)
 CALL collocate_core_4(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(5)
 CALL collocate_core_5(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(6)
 CALL collocate_core_6(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(7)
 CALL collocate_core_7(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(8)
 CALL collocate_core_8(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
 CASE(9)
 CALL collocate_core_9(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax), pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),cmax,gridbounds(1,1))
  CASE DEFAULT
 CALL collocate_core_default(grid(1,1,1),coef_xyz(1),pol_x(0,-cmax),pol_y(1,0,-cmax),pol_z(1,0,-cmax), &
                        map(-cmax,1),sphere_bounds(1),lp,cmax,gridbounds(1,1))
  END SELECT
         END IF

         IF (PRESENT(ir) .AND. PRESENT(rsgauge)) CALL collocate_gauge_ortho()

         ! deallocation needed to pass around a pgi bug..
         DEALLOCATE (pol_z)
         DEALLOCATE (pol_y)
         DEALLOCATE (pol_x)
         DEALLOCATE (map)

      END SUBROUTINE collocate_ortho

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_gauge_ortho()
      INTEGER                                            :: i, igmax, igmin, j, j2, jg, jg2, jgmin, &
                                                            k, k2, kg, kg2, kgmin, sci
      REAL(KIND=dp)                                      :: point(3, 4), res(4), x, y, y2, z, z2

! notice we're in the ortho case

         dr(1) = rsgrid%desc%dh(1, 1)
         dr(2) = rsgrid%desc%dh(2, 2)
         dr(3) = rsgrid%desc%dh(3, 3)
         !
         sci = 1
         kgmin = sphere_bounds(sci)
         sci = sci+1
         DO kg = kgmin, 0
            kg2 = 1-kg
            k = map(kg, 3)
            k2 = map(kg2, 3)
            jgmin = sphere_bounds(sci)
            sci = sci+1
            z = (REAL(kg, dp)+REAL(cubecenter(3), dp))*dr(3)
            z2 = (REAL(kg2, dp)+REAL(cubecenter(3), dp))*dr(3)
            DO jg = jgmin, 0
               jg2 = 1-jg
               j = map(jg, 2)
               j2 = map(jg2, 2)
               igmin = sphere_bounds(sci)
               sci = sci+1
               igmax = 1-igmin
               y = (REAL(jg, dp)+REAL(cubecenter(2), dp))*dr(2)
               y2 = (REAL(jg2, dp)+REAL(cubecenter(2), dp))*dr(2)
               DO ig = igmin, igmax
                  i = map(ig, 1)
                  x = (REAL(ig, dp)+REAL(cubecenter(1), dp))*dr(1)
                  point(1, 1) = x; point(2, 1) = y; point(3, 1) = z
                  point(1, 2) = x; point(2, 2) = y2; point(3, 2) = z
                  point(1, 3) = x; point(2, 3) = y; point(3, 3) = z2
                  point(1, 4) = x; point(2, 4) = y2; point(3, 4) = z2
                  !
                  res(1) = (point(ir, 1)-rb(ir))-gauge(i, j, k)
                  res(2) = (point(ir, 2)-rb(ir))-gauge(i, j2, k)
                  res(3) = (point(ir, 3)-rb(ir))-gauge(i, j, k2)
                  res(4) = (point(ir, 4)-rb(ir))-gauge(i, j2, k2)
                  !
                  grid_tmp(i, j, k) = grid_tmp(i, j, k)+grid(i, j, k)*res(1)
                  grid_tmp(i, j2, k) = grid_tmp(i, j2, k)+grid(i, j2, k)*res(2)
                  grid_tmp(i, j, k2) = grid_tmp(i, j, k2)+grid(i, j, k2)*res(3)
                  grid_tmp(i, j2, k2) = grid_tmp(i, j2, k2)+grid(i, j2, k2)*res(4)
               ENDDO
            ENDDO
         ENDDO
      END SUBROUTINE collocate_gauge_ortho

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_ortho_set_to_0()
      INTEGER                                            :: i, igmax, igmin, j, j2, jg, jg2, jgmin, &
                                                            k, k2, kg, kg2, kgmin, sci

!

         dr(1) = rsgrid%desc%dh(1, 1)
         dr(2) = rsgrid%desc%dh(2, 2)
         dr(3) = rsgrid%desc%dh(3, 3)
         !
         sci = 1
         kgmin = sphere_bounds(sci)
         sci = sci+1
         DO kg = kgmin, 0
            kg2 = 1-kg
            k = map(kg, 3)
            k2 = map(kg2, 3)
            jgmin = sphere_bounds(sci)
            sci = sci+1
            DO jg = jgmin, 0
               jg2 = 1-jg
               j = map(jg, 2)
               j2 = map(jg2, 2)
               igmin = sphere_bounds(sci)
               sci = sci+1
               igmax = 1-igmin
               DO ig = igmin, igmax
                  i = map(ig, 1)
                  grid(i, j, k) = 0.0_dp
                  grid(i, j2, k) = 0.0_dp
                  grid(i, j, k2) = 0.0_dp
                  grid(i, j2, k2) = 0.0_dp
               ENDDO
            ENDDO
         ENDDO
      END SUBROUTINE collocate_ortho_set_to_0

!
!   this is a general 'optimized' routine to do the collocation
!
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_general_opt()

      INTEGER :: i, i_index, il, ilx, ily, ilz, index_max(3), index_min(3), ismax, ismin, j, &
         j_index, jl, jlx, jly, jlz, k, k_index, kl, klx, kly, klz, lpx, lpy, lpz, lx, ly, lz, &
         offset(3)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: grid_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: coef_map
      REAL(KIND=dp)                                      :: a, b, c, d, di, dip, dj, djp, dk, dkp, &
                                                            exp0i, exp1i, exp2i, gp(3), &
                                                            hmatgrid(3, 3), pointj(3), pointk(3), &
                                                            res, v(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: coef_ijk
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: hmatgridp

!
! transform P_{lxp,lyp,lzp} into a P_{lip,ljp,lkp} such that
! sum_{lxp,lyp,lzp} P_{lxp,lyp,lzp} (x-x_p)**lxp (y-y_p)**lyp (z-z_p)**lzp =
! sum_{lip,ljp,lkp} P_{lip,ljp,lkp} (i-i_p)**lip (j-j_p)**ljp (k-k_p)**lkp
!

         ALLOCATE (coef_ijk(((lp+1)*(lp+2)*(lp+3))/6))

         ! aux mapping array to simplify life
         ALLOCATE (coef_map(0:lp, 0:lp, 0:lp))
         coef_map = HUGE(coef_map)
         lxyz = 0
         DO lzp = 0, lp
         DO lyp = 0, lp-lzp
         DO lxp = 0, lp-lzp-lyp
            lxyz = lxyz+1
            coef_ijk(lxyz) = 0.0_dp
            coef_map(lxp, lyp, lzp) = lxyz
         ENDDO
         ENDDO
         ENDDO

         ! cell hmat in grid points
         hmatgrid = rsgrid%desc%dh

         ! center in grid coords
         gp = MATMUL(rsgrid%desc%dh_inv, rp)
         cubecenter(:) = FLOOR(gp)

         ! transform using multinomials
         ALLOCATE (hmatgridp(3, 3, 0:lp))
         hmatgridp(:, :, 0) = 1.0_dp
         DO k = 1, lp
            hmatgridp(:, :, k) = hmatgridp(:, :, k-1)*hmatgrid(:, :)
         ENDDO

         lpx = lp
         DO klx = 0, lpx
         DO jlx = 0, lpx-klx
         DO ilx = 0, lpx-klx-jlx
            lx = ilx+jlx+klx
            lpy = lp-lx
            DO kly = 0, lpy
            DO jly = 0, lpy-kly
            DO ily = 0, lpy-kly-jly
               ly = ily+jly+kly
               lpz = lp-lx-ly
               DO klz = 0, lpz
               DO jlz = 0, lpz-klz
               DO ilz = 0, lpz-klz-jlz
                  lz = ilz+jlz+klz

                  il = ilx+ily+ilz
                  jl = jlx+jly+jlz
                  kl = klx+kly+klz
                  coef_ijk(coef_map(il, jl, kl)) = &
                     coef_ijk(coef_map(il, jl, kl))+coef_xyz(coef_map(lx, ly, lz))* &
                     hmatgridp(1, 1, ilx)*hmatgridp(1, 2, jlx)*hmatgridp(1, 3, klx)* &
                     hmatgridp(2, 1, ily)*hmatgridp(2, 2, jly)*hmatgridp(2, 3, kly)* &
                     hmatgridp(3, 1, ilz)*hmatgridp(3, 2, jlz)*hmatgridp(3, 3, klz)* &
                     fac(lx)*fac(ly)*fac(lz)/ &
                     (fac(ilx)*fac(ily)*fac(ilz)*fac(jlx)*fac(jly)*fac(jlz)*fac(klx)*fac(kly)*fac(klz))
               ENDDO
               ENDDO
               ENDDO
            ENDDO
            ENDDO
            ENDDO
         ENDDO
         ENDDO
         ENDDO

         CALL return_cube_nonortho(cube_info, radius, index_min, index_max, rp)

         offset(:) = MODULO(index_min(:)+rsgrid%desc%lb(:)-rsgrid%lb_local(:), ng(:))+1

         ALLOCATE (grid_map(index_min(1):index_max(1)))
         DO i = index_min(1), index_max(1)
            grid_map(i) = MODULO(i, ng(1))+1
            IF (rsgrid%desc%perd(1) == 1) THEN
               grid_map(i) = MODULO(i, ng(1))+1
            ELSE
               grid_map(i) = i-index_min(1)+offset(1)
            ENDIF
         ENDDO

         ! go over the grid, but cycle if the point is not within the radius
         DO k = index_min(3), index_max(3)
            dk = k-gp(3)
            pointk = hmatgrid(:, 3)*dk

            IF (rsgrid%desc%perd(3) == 1) THEN
               k_index = MODULO(k, ng(3))+1
            ELSE
               k_index = k-index_min(3)+offset(3)
            ENDIF

            coef_xyt = 0.0_dp
            lxyz = 0
            dkp = 1.0_dp
            DO kl = 0, lp
               lxy = 0
               DO jl = 0, lp-kl
                  DO il = 0, lp-kl-jl
                     lxyz = lxyz+1; lxy = lxy+1
                     coef_xyt(lxy) = coef_xyt(lxy)+coef_ijk(lxyz)*dkp
                  ENDDO
                  lxy = lxy+kl
               ENDDO
               dkp = dkp*dk
            ENDDO

            DO j = index_min(2), index_max(2)
               dj = j-gp(2)
               pointj = pointk+hmatgrid(:, 2)*dj
               IF (rsgrid%desc%perd(2) == 1) THEN
                  j_index = MODULO(j, ng(2))+1
               ELSE
                  j_index = j-index_min(2)+offset(2)
               ENDIF

               coef_xtt = 0.0_dp
               lxy = 0
               djp = 1.0_dp
               DO jl = 0, lp
                  DO il = 0, lp-jl
                     lxy = lxy+1
                     coef_xtt(il) = coef_xtt(il)+coef_xyt(lxy)*djp
                  ENDDO
                  djp = djp*dj
               ENDDO

               ! find bounds for the inner loop
               ! based on a quadratic equation in i
               ! a*i**2+b*i+c=radius**2
               v = pointj-gp(1)*hmatgrid(:, 1)
               a = DOT_PRODUCT(hmatgrid(:, 1), hmatgrid(:, 1))
               b = 2*DOT_PRODUCT(v, hmatgrid(:, 1))
               c = DOT_PRODUCT(v, v)
               d = b*b-4*a*(c-radius**2)

               IF (d < 0) THEN
                  CYCLE
               ELSE
                  d = SQRT(d)
                  ismin = CEILING((-b-d)/(2*a))
                  ismax = FLOOR((-b+d)/(2*a))
               ENDIF
               ! prepare for computing -zetp*rsq
               a = -zetp*a
               b = -zetp*b
               c = -zetp*c
               i = ismin-1

               ! the recursion relation might have to be done
               ! from the center of the gaussian (in both directions)
               ! instead as the current implementation from an edge
               exp2i = EXP((a*i+b)*i+c)
               exp1i = EXP(2*a*i+a+b)
               exp0i = EXP(2*a)

               DO i = ismin, ismax
                  di = i-gp(1)

                  ! polynomial terms
                  res = 0.0_dp
                  dip = 1.0_dp
                  DO il = 0, lp
                     res = res+coef_xtt(il)*dip
                     dip = dip*di
                  ENDDO

                  ! the exponential recursion
                  exp2i = exp2i*exp1i
                  exp1i = exp1i*exp0i
                  res = res*exp2i

                  i_index = grid_map(i)
                  IF (PRESENT(lgrid)) THEN
                     ig = lgrid%ldim*ithread_l+(k_index-1)*ng(2)*ng(1)+(j_index-1)*ng(1)+(i_index-1)+1
                     lgrid%r(ig) = lgrid%r(ig)+res
                  ELSE
                     grid(i_index, j_index, k_index) = grid(i_index, j_index, k_index)+res
                  ENDIF
               ENDDO
            ENDDO
         ENDDO
         !t2=nanotime_ia32()
         !write(*,*) t2-t1
         ! deallocation needed to pass around a pgi bug..
         DEALLOCATE (coef_ijk)
         DEALLOCATE (coef_map)
         DEALLOCATE (hmatgridp)
         DEALLOCATE (grid_map)

      END SUBROUTINE collocate_general_opt

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_general_subpatch()
      INTEGER, DIMENSION(2, 3)                           :: local_b
      INTEGER, DIMENSION(3)                              :: local_s, periodic
      REAL(dp), DIMENSION((lp+1)*(lp+2)*(lp+3)/6)        :: poly_d3

         periodic = 1 ! cell%perd
         CALL poly_cp2k2d3(coef_xyz, lp, poly_d3)
         local_b(1, :) = rsgrid%lb_real-rsgrid%desc%lb
         local_b(2, :) = rsgrid%ub_real-rsgrid%desc%lb
         local_s = rsgrid%lb_real-rsgrid%lb_local
         IF (BTEST(subpatch_pattern, 0)) local_b(1, 1) = local_b(1, 1)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 1)) local_b(2, 1) = local_b(2, 1)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 2)) local_b(1, 2) = local_b(1, 2)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 3)) local_b(2, 2) = local_b(2, 2)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 4)) local_b(1, 3) = local_b(1, 3)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 5)) local_b(2, 3) = local_b(2, 3)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 0)) local_s(1) = local_s(1)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 2)) local_s(2) = local_s(2)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 4)) local_s(3) = local_s(3)-rsgrid%desc%border
         IF (PRESENT(lgrid)) THEN
            CALL collocGauss(h=cell%hmat, h_inv=cell%h_inv, &
                             grid=grid, poly=poly_d3, alphai=zetp, posi=rp, max_r2=radius*radius, &
                             periodic=periodic, gdim=ng, local_bounds=local_b, local_shift=local_s, &
                             lgrid=lgrid)
         ELSE
            CALL collocGauss(h=cell%hmat, h_inv=cell%h_inv, &
                             grid=grid, poly=poly_d3, alphai=zetp, posi=rp, max_r2=radius*radius, &
                             periodic=periodic, gdim=ng, local_bounds=local_b, local_shift=local_s)
         END IF
         ! defaults: local_shift=(/0,0,0/),poly_shift=(/0.0_dp,0.0_dp,0.0_dp/),scale=1.0_dp,

      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_general_wings()
      INTEGER, DIMENSION(2, 3)                           :: local_b
      INTEGER, DIMENSION(3)                              :: periodic
      REAL(dp), DIMENSION((lp+1)*(lp+2)*(lp+3)/6)        :: poly_d3
      REAL(dp), DIMENSION(3)                             :: local_shift, rShifted

         periodic = 1 ! cell%perd
         CALL poly_cp2k2d3(coef_xyz, lp, poly_d3)
         local_b(1, :) = 0
         local_b(2, :) = MIN(rsgrid%desc%npts-1, rsgrid%ub_local-rsgrid%lb_local)
         local_shift = REAL(rsgrid%desc%lb-rsgrid%lb_local, dp)/REAL(rsgrid%desc%npts, dp)
         rShifted(1) = rp(1)+cell%hmat(1, 1)*local_shift(1) &
                       +cell%hmat(1, 2)*local_shift(2) &
                       +cell%hmat(1, 3)*local_shift(3)
         rShifted(2) = rp(2)+cell%hmat(2, 1)*local_shift(1) &
                       +cell%hmat(2, 2)*local_shift(2) &
                       +cell%hmat(2, 3)*local_shift(3)
         rShifted(3) = rp(3)+cell%hmat(3, 1)*local_shift(1) &
                       +cell%hmat(3, 2)*local_shift(2) &
                       +cell%hmat(3, 3)*local_shift(3)
         IF (PRESENT(lgrid)) THEN
            CALL collocGauss(h=cell%hmat, h_inv=cell%h_inv, &
                             grid=grid, poly=poly_d3, alphai=zetp, posi=rShifted, max_r2=radius*radius, &
                             periodic=periodic, gdim=ng, local_bounds=local_b, &
                             lgrid=lgrid)
         ELSE
            CALL collocGauss(h=cell%hmat, h_inv=cell%h_inv, &
                             grid=grid, poly=poly_d3, alphai=zetp, posi=rShifted, max_r2=radius*radius, &
                             periodic=periodic, gdim=ng, local_bounds=local_b)
         END IF
         ! defaults: local_shift=(/0,0,0/),poly_shift=(/0.0_dp,0.0_dp,0.0_dp/),scale=1.0_dp,

      END SUBROUTINE

!
!   this is a general 'reference' routine to do the collocation
!
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE collocate_general()

      INTEGER                                            :: i, index_max(3), index_min(3), &
                                                            ipoint(3), j, k
      REAL(KIND=dp)                                      :: point(3)

! still hard-wired (see MODULO)

         IF(.NOT.(ALL(rsgrid%desc%perd == 1)))CALL cp__a("qs_collocate_density.F",3671)

         CALL return_cube_nonortho(cube_info, radius, index_min, index_max, rp)

         ! go over the grid, but cycle if the point is not within the radius
         DO k = index_min(3), index_max(3)
         DO j = index_min(2), index_max(2)
         DO i = index_min(1), index_max(1)
            ! point in real space
            point = MATMUL(cell%hmat, REAL((/i, j, k/), KIND=dp)/ng)
            ! skip if outside of the sphere
            IF (SUM((point-rp)**2) > radius**2) CYCLE
            ! point on the grid (including pbc)
            ipoint = MODULO((/i, j, k/), ng)+1
            ! add to grid
            IF (PRESENT(lgrid)) THEN
               ig = lgrid%ldim*ithread_l+ipoint(3)*ng(2)*ng(1)+ipoint(2)*ng(1)+ipoint(1)+1
               lgrid%r(ig) = lgrid%r(ig)+primitive_value(point)
            ELSE
               grid(ipoint(1), ipoint(2), ipoint(3)) = grid(ipoint(1), ipoint(2), ipoint(3))+primitive_value(point)
            ENDIF
         ENDDO
         ENDDO
         ENDDO

      END SUBROUTINE collocate_general

! **************************************************************************************************
!> \brief ...
!> \param point ...
!> \retval res ...
! **************************************************************************************************
      FUNCTION primitive_value(point) RESULT(res)
      REAL(KIND=dp)                                      :: point(3), res

      REAL(KIND=dp)                                      :: dra(3), drap(3), drb(3), drbp(3), myexp

         res = 0.0_dp
         myexp = EXP(-zetp*SUM((point-rp)**2))*prefactor
         dra = point-ra
         drb = point-rb
         drap(1) = 1.0_dp
         DO lxa = 0, la_max_local
            drbp(1) = 1.0_dp
            DO lxb = 0, lb_max_local
               drap(2) = 1.0_dp
               DO lya = 0, la_max_local-lxa
                  drbp(2) = 1.0_dp
                  DO lyb = 0, lb_max_local-lxb
                     drap(3) = 1.0_dp
                     DO lza = 1, MAX(la_min_local-lxa-lya, 0)
                        drap(3) = drap(3)*dra(3)
                     ENDDO
                     DO lza = MAX(la_min_local-lxa-lya, 0), la_max_local-lxa-lya
                        drbp(3) = 1.0_dp
                        DO lzb = 1, MAX(lb_min_local-lxb-lyb, 0)
                           drbp(3) = drbp(3)*drb(3)
                        ENDDO
                        DO lzb = MAX(lb_min_local-lxb-lyb, 0), lb_max_local-lxb-lyb
                           ico = coset(lxa, lya, lza)
                           jco = coset(lxb, lyb, lzb)
                           res = res+pab_local(ico+o1_local, jco+o2_local)*myexp*PRODUCT(drap)*PRODUCT(drbp)
                           drbp(3) = drbp(3)*drb(3)
                        ENDDO
                        drap(3) = drap(3)*dra(3)
                     ENDDO
                     drbp(2) = drbp(2)*drb(2)
                  ENDDO
                  drap(2) = drap(2)*dra(2)
               ENDDO
               drbp(1) = drbp(1)*drb(1)
            ENDDO
            drap(1) = drap(1)*dra(1)
         ENDDO

      END FUNCTION primitive_value

   END SUBROUTINE collocate_pgf_product_rspace

! **************************************************************************************************
!> \brief low level collcation of primitive gaussian functions in g-space
!> \param la_max ...
!> \param zeta ...
!> \param la_min ...
!> \param lb_max ...
!> \param zetb ...
!> \param lb_min ...
!> \param ra ...
!> \param rab ...
!> \param rab2 ...
!> \param scale ...
!> \param pab ...
!> \param na ...
!> \param nb ...
!> \param eps_rho_gspace ...
!> \param gsq_max ...
!> \param pw ...
! **************************************************************************************************
   SUBROUTINE collocate_pgf_product_gspace(la_max, zeta, la_min, &
                                           lb_max, zetb, lb_min, &
                                           ra, rab, rab2, scale, pab, na, nb, &
                                           eps_rho_gspace, gsq_max, pw)

      ! NOTE: this routine is much slower than collocate_pgf_product_rspace

      INTEGER, INTENT(IN)                                :: la_max
      REAL(dp), INTENT(IN)                               :: zeta
      INTEGER, INTENT(IN)                                :: la_min, lb_max
      REAL(dp), INTENT(IN)                               :: zetb
      INTEGER, INTENT(IN)                                :: lb_min
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: ra, rab
      REAL(dp), INTENT(IN)                               :: rab2, scale
      REAL(dp), DIMENSION(:, :), POINTER                 :: pab
      INTEGER, INTENT(IN)                                :: na, nb
      REAL(dp), INTENT(IN)                               :: eps_rho_gspace, gsq_max
      TYPE(pw_type), POINTER                             :: pw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'collocate_pgf_product_gspace', &
         routineP = moduleN//':'//routineN

      COMPLEX(dp), DIMENSION(3)                          :: phasefactor
      COMPLEX(dp), DIMENSION(:), POINTER                 :: rag, rbg
      COMPLEX(dp), DIMENSION(:, :, :, :), POINTER        :: cubeaxis
      INTEGER                                            :: ax, ay, az, bx, by, bz, handle, i, ico, &
                                                            ig, ig2, jco, jg, kg, la, lb, &
                                                            lb_cube_min, lb_grid, ub_cube_max, &
                                                            ub_grid
      INTEGER, DIMENSION(3)                              :: lb_cube, ub_cube
      REAL(dp)                                           :: f, fa, fb, pij, prefactor, rzetp, &
                                                            twozetp, zetp
      REAL(dp), DIMENSION(3)                             :: dg, expfactor, fap, fbp, rap, rbp, rp
      REAL(dp), DIMENSION(:), POINTER                    :: g

      CALL timeset(routineN, handle)

      dg(:) = twopi/(pw%pw_grid%npts(:)*pw%pw_grid%dr(:))

      zetp = zeta+zetb
      rzetp = 1.0_dp/zetp
      f = zetb*rzetp
      rap(:) = f*rab(:)
      rbp(:) = rap(:)-rab(:)
      rp(:) = ra(:)+rap(:)
      twozetp = 2.0_dp*zetp
      fap(:) = twozetp*rap(:)
      fbp(:) = twozetp*rbp(:)

      prefactor = scale*SQRT((pi*rzetp)**3)*EXP(-zeta*f*rab2)
      phasefactor(:) = EXP(CMPLX(0.0_dp, -rp(:)*dg(:), KIND=dp))
      expfactor(:) = EXP(-0.25*rzetp*dg(:)*dg(:))

      lb_cube(:) = pw%pw_grid%bounds(1, :)
      ub_cube(:) = pw%pw_grid%bounds(2, :)

      lb_cube_min = MINVAL(lb_cube(:))
      ub_cube_max = MAXVAL(ub_cube(:))

      NULLIFY (cubeaxis, g, rag, rbg)

      CALL reallocate(cubeaxis, lb_cube_min, ub_cube_max, 1, 3, 0, la_max, 0, lb_max)
      CALL reallocate(g, lb_cube_min, ub_cube_max)
      CALL reallocate(rag, lb_cube_min, ub_cube_max)
      CALL reallocate(rbg, lb_cube_min, ub_cube_max)

      lb_grid = LBOUND(pw%cc, 1)
      ub_grid = UBOUND(pw%cc, 1)

      DO i = 1, 3

         DO ig = lb_cube(i), ub_cube(i)
            ig2 = ig*ig
            cubeaxis(ig, i, 0, 0) = expfactor(i)**ig2*phasefactor(i)**ig
         END DO

         IF (la_max > 0) THEN
            DO ig = lb_cube(i), ub_cube(i)
               g(ig) = REAL(ig, dp)*dg(i)
               rag(ig) = CMPLX(fap(i), -g(ig), KIND=dp)
               cubeaxis(ig, i, 1, 0) = rag(ig)*cubeaxis(ig, i, 0, 0)
            END DO
            DO la = 2, la_max
               fa = REAL(la-1, dp)*twozetp
               DO ig = lb_cube(i), ub_cube(i)
                  cubeaxis(ig, i, la, 0) = rag(ig)*cubeaxis(ig, i, la-1, 0)+ &
                                           fa*cubeaxis(ig, i, la-2, 0)
               END DO
            END DO
            IF (lb_max > 0) THEN
               fa = twozetp
               DO ig = lb_cube(i), ub_cube(i)
                  rbg(ig) = CMPLX(fbp(i), -g(ig), KIND=dp)
                  cubeaxis(ig, i, 0, 1) = rbg(ig)*cubeaxis(ig, i, 0, 0)
                  cubeaxis(ig, i, 1, 1) = rbg(ig)*cubeaxis(ig, i, 1, 0)+ &
                                          fa*cubeaxis(ig, i, 0, 0)
               END DO
               DO lb = 2, lb_max
                  fb = REAL(lb-1, dp)*twozetp
                  DO ig = lb_cube(i), ub_cube(i)
                     cubeaxis(ig, i, 0, lb) = rbg(ig)*cubeaxis(ig, i, 0, lb-1)+ &
                                              fb*cubeaxis(ig, i, 0, lb-2)
                     cubeaxis(ig, i, 1, lb) = rbg(ig)*cubeaxis(ig, i, 1, lb-1)+ &
                                              fb*cubeaxis(ig, i, 1, lb-2)+ &
                                              fa*cubeaxis(ig, i, 0, lb-1)
                  END DO
               END DO
               DO la = 2, la_max
                  fa = REAL(la, dp)*twozetp
                  DO ig = lb_cube(i), ub_cube(i)
                     cubeaxis(ig, i, la, 1) = rbg(ig)*cubeaxis(ig, i, la, 0)+ &
                                              fa*cubeaxis(ig, i, la-1, 0)
                  END DO
                  DO lb = 2, lb_max
                     fb = REAL(lb-1, dp)*twozetp
                     DO ig = lb_cube(i), ub_cube(i)
                        cubeaxis(ig, i, la, lb) = rbg(ig)*cubeaxis(ig, i, la, lb-1)+ &
                                                  fb*cubeaxis(ig, i, la, lb-2)+ &
                                                  fa*cubeaxis(ig, i, la-1, lb-1)
                     END DO
                  END DO
               END DO
            END IF
         ELSE
            IF (lb_max > 0) THEN
               DO ig = lb_cube(i), ub_cube(i)
                  g(ig) = REAL(ig, dp)*dg(i)
                  rbg(ig) = CMPLX(fbp(i), -g(ig), KIND=dp)
                  cubeaxis(ig, i, 0, 1) = rbg(ig)*cubeaxis(ig, i, 0, 0)
               END DO
               DO lb = 2, lb_max
                  fb = REAL(lb-1, dp)*twozetp
                  DO ig = lb_cube(i), ub_cube(i)
                     cubeaxis(ig, i, 0, lb) = rbg(ig)*cubeaxis(ig, i, 0, lb-1)+ &
                                              fb*cubeaxis(ig, i, 0, lb-2)
                  END DO
               END DO
            END IF
         END IF

      END DO

      DO la = 0, la_max
         DO lb = 0, lb_max
            IF (la+lb == 0) CYCLE
            fa = (1.0_dp/twozetp)**(la+lb)
            DO i = 1, 3
               DO ig = lb_cube(i), ub_cube(i)
                  cubeaxis(ig, i, la, lb) = fa*cubeaxis(ig, i, la, lb)
               END DO
            END DO
         END DO
      END DO

      ! Add the current primitive Gaussian function product to grid

      DO ico = ncoset(la_min-1)+1, ncoset(la_max)

         ax = indco(1, ico)
         ay = indco(2, ico)
         az = indco(3, ico)

         DO jco = ncoset(lb_min-1)+1, ncoset(lb_max)

            pij = prefactor*pab(na+ico, nb+jco)

            IF (ABS(pij) < eps_rho_gspace) CYCLE

            bx = indco(1, jco)
            by = indco(2, jco)
            bz = indco(3, jco)

            DO i = lb_grid, ub_grid
               IF (pw%pw_grid%gsq(i) > gsq_max) CYCLE
               ig = pw%pw_grid%g_hat(1, i)
               jg = pw%pw_grid%g_hat(2, i)
               kg = pw%pw_grid%g_hat(3, i)
               pw%cc(i) = pw%cc(i)+pij*cubeaxis(ig, 1, ax, bx)* &
                          cubeaxis(jg, 2, ay, by)* &
                          cubeaxis(kg, 3, az, bz)
            END DO

         END DO

      END DO

      DEALLOCATE (cubeaxis)
      DEALLOCATE (g)
      DEALLOCATE (rag)
      DEALLOCATE (rbg)

      CALL timestop(handle)

   END SUBROUTINE collocate_pgf_product_gspace

END MODULE qs_collocate_density
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the overlap integrals over solid harmonic Gaussian
!>        (SHG) functions.
!> \par Literature (partly)
!>      T.J. Giese and D. M. York, J. Chem. Phys, 128, 064104 (2008)
!>      T. Helgaker, P Joergensen, J. Olsen, Molecular Electronic-Structure
!>                                           Theory, Wiley
!> \par History
!>      created [04.2015]
!> \author Dorothea Golze
! **************************************************************************************************
MODULE ai_shg_overlap
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              pi
   USE orbital_pointers,                ONLY: indso,&
                                              indso_inv,&
                                              nsoset_pm
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_shg_overlap'

! *** Public subroutines ***
   PUBLIC :: get_dW_matrix, get_W_matrix, get_real_scaled_solid_harmonic, &
             overlap_shg_ab, dev_overlap_shg_ab, overlap_shg_aba, dev_overlap_shg_aba, &
             overlap_shg_abb, dev_overlap_shg_abb, s_overlap_ab, s_overlap_abx

CONTAINS

! **************************************************************************************************
!> \brief calculates the uncontracted, not normalized [s|s] overlap
!> \param la_max maximal l quantum number on a
!> \param npgfa number of primitive Gaussian on a
!> \param zeta set of exponents on a
!> \param lb_max maximal l quantum number on b
!> \param npgfb number of primitive Gaussian on a
!> \param zetb set of exponents on a
!> \param rab distance vector between a and b
!> \param s uncontracted overlap of s functions
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE s_overlap_ab(la_max, npgfa, zeta, lb_max, npgfb, zetb, rab, s, calculate_forces)

      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: lb_max, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: s
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 's_overlap_ab', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ids, ipgfa, jpgfb, ndev
      REAL(KIND=dp)                                      :: a, b, rab2, xhi, zet

      CALL timeset(routineN, handle)

      ! Distance of the centers a and b
      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      ndev = 0
      IF (calculate_forces) ndev = 1
      ! Loops over all pairs of primitive Gaussian-type functions
      DO ipgfa = 1, npgfa
         DO jpgfb = 1, npgfb

            ! Distance Screening   !maybe later

            ! Calculate some prefactors
            a = zeta(ipgfa)
            b = zetb(jpgfb)
            zet = a+b
            xhi = a*b/zet

            ! [s|s] integral
            s(ipgfa, jpgfb, 1) = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            DO ids = 2, la_max+lb_max+ndev+1
               s(ipgfa, jpgfb, ids) = -xhi*s(ipgfa, jpgfb, ids-1)
            ENDDO

         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE s_overlap_ab

! **************************************************************************************************
!> \brief calculates [s|ra^n|s] integrals for [aba] and the [s|rb^n|s]
!>        integrals for [abb]
!>        abx ... x stands either for a and b respectively
!> \param la_max maximal l quantum number on a, orbital basis
!> \param npgfa number of primitive Gaussian on a, orbital basis
!> \param zeta set of exponents on a, orbital basis
!> \param lb_max maximal l quantum number on b, orbital basis
!> \param npgfb number of primitive Gaussian on a, orbital basis
!> \param zetb set of exponents on b, orbital basis
!> \param lx_max maximal l quantum number of aux basis on x (a or b)
!> \param npgfx number of primitive Gaussian on x, aux basis
!> \param zetx set of exponents on x, aux basis
!> \param rab distance vector between a and b
!> \param s uncontracted [s|r^n|s] integrals
!> \param calculate_forces ...
!> \param calc_aba flag if [aba] or [abb] is calculated
! **************************************************************************************************
   SUBROUTINE s_overlap_abx(la_max, npgfa, zeta, lb_max, npgfb, zetb, lx_max, npgfx, zetx, &
                            rab, s, calculate_forces, calc_aba)

      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: lb_max, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb
      INTEGER, INTENT(IN)                                :: lx_max, npgfx
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetx
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: s
      LOGICAL, INTENT(IN)                                :: calculate_forces, calc_aba

      CHARACTER(len=*), PARAMETER :: routineN = 's_overlap_abx', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ids, il, ipgfa, jpgfb, kpgfx, &
                                                            laa_max, lbb_max, lmax, n, ndev, nds, &
                                                            nl
      REAL(KIND=dp)                                      :: a, b, exp_rab2, k, pfac, prefac, rab2, &
                                                            sqrt_pi3, sqrt_zet, sr_int, temp, x, &
                                                            xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dsr_int, dtemp

      CALL timeset(routineN, handle)

      ! Distance of the centers a and b
      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      ndev = 0
      IF (calculate_forces) ndev = 1

      IF (calc_aba) THEN
         laa_max = la_max+lx_max
         nl = INT(laa_max/2)
         IF (la_max == 0 .OR. lx_max == 0) nl = 0
         lmax = laa_max+lb_max
      ELSE
         lbb_max = lb_max+lx_max
         nl = INT(lbb_max/2)
         IF (lb_max == 0 .OR. lx_max == 0) nl = 0
         lmax = la_max+lbb_max
      ENDIF

      ALLOCATE (dtemp(nl+1), dsr_int(nl+1))

      sqrt_pi3 = SQRT(pi**3)

      ! Loops over all pairs of primitive Gaussian-type functions
      DO ipgfa = 1, npgfa
         DO jpgfb = 1, npgfb
            DO kpgfx = 1, npgfx

               ! Calculate some prefactors
               IF (calc_aba) THEN
                  a = zeta(ipgfa)+zetx(kpgfx)
                  b = zetb(jpgfb)
                  x = b
               ELSE
                  a = zeta(ipgfa)
                  b = zetb(jpgfb)+zetx(kpgfx)
                  x = a
               ENDIF

               zet = a+b
               xhi = a*b/zet
               exp_rab2 = EXP(-xhi*rab2)

               pfac = x**2/zet
               sqrt_zet = SQRT(zet)

               DO il = 0, nl
                  nds = lmax-2*il+ndev+1
                  SELECT CASE (il)
                  CASE (0)
                     ! [s|s] integral
                     s(ipgfa, jpgfb, kpgfx, il, 1) = (pi/zet)**(1.5_dp)*exp_rab2
                     DO ids = 2, nds
                        n = ids-1
                        s(ipgfa, jpgfb, kpgfx, il, ids) = (-xhi)**n*s(ipgfa, jpgfb, kpgfx, il, 1)
                     ENDDO
                  CASE (1)
                     ![s|r^2|s] integral
                     sr_int = sqrt_pi3/sqrt_zet**5*(3.0_dp+2.0_dp*pfac*rab2)/2.0_dp
                     s(ipgfa, jpgfb, kpgfx, il, 1) = exp_rab2*sr_int
                     k = sqrt_pi3*x**2/sqrt_zet**7
                     DO ids = 2, nds
                        n = ids-1
                        s(ipgfa, jpgfb, kpgfx, il, ids) = (-xhi)**n*exp_rab2*sr_int &
                                                          +n*(-xhi)**(n-1)*k*exp_rab2
                     ENDDO
                  CASE (2)
                     ![s|r^4|s] integral
                     prefac = sqrt_pi3/4.0_dp/sqrt_zet**7
                     temp = 15.0_dp+20.0_dp*pfac*rab2+4.0_dp*(pfac*rab2)**2
                     sr_int = prefac*temp
                     s(ipgfa, jpgfb, kpgfx, il, 1) = exp_rab2*sr_int
                     !** derivatives
                     k = sqrt_pi3*x**4/sqrt_zet**11
                     dsr_int(1) = prefac*(20.0_dp*pfac+8.0_dp*pfac**2*rab2)
                     DO ids = 2, nds
                        n = ids-1
                        dtemp(1) = (-xhi)**n*exp_rab2*sr_int
                        dtemp(2) = n*(-xhi)**(n-1)*exp_rab2*dsr_int(1)
                        dtemp(3) = (n**2-n)*(-xhi)**(n-2)*k*exp_rab2
                        s(ipgfa, jpgfb, kpgfx, il, ids) = dtemp(1)+dtemp(2)+dtemp(3)
                     ENDDO
                  CASE (3)
                     ![s|r^6|s] integral
                     prefac = sqrt_pi3/8.0_dp/sqrt_zet**9
                     temp = 105.0_dp+210.0_dp*pfac*rab2
                     temp = temp+84.0_dp*(pfac*rab2)**2+8.0_dp*(pfac*rab2)**3
                     sr_int = prefac*temp
                     s(ipgfa, jpgfb, kpgfx, il, 1) = exp_rab2*sr_int
                     !** derivatives
                     k = sqrt_pi3*x**6/sqrt_zet**15
                     dsr_int(1) = prefac*(210.0_dp*pfac+168.0_dp*pfac**2*rab2 &
                                          +24.0_dp*pfac**3*rab2**2)
                     dsr_int(2) = prefac*(168.0_dp*pfac**2+48.0_dp*pfac**3*rab2)
                     DO ids = 2, nds
                        n = ids-1
                        dtemp(1) = (-xhi)**n*exp_rab2*sr_int
                        dtemp(2) = REAL(n, dp)*(-xhi)**(n-1)*exp_rab2*dsr_int(1)
                        dtemp(3) = REAL(n**2-n, dp)/2.0_dp*(-xhi)**(n-2) &
                                   *exp_rab2*dsr_int(2)
                        dtemp(4) = REAL(n*(n-1)*(n-2), dp)*(-xhi)**(n-3)*k*exp_rab2
                        s(ipgfa, jpgfb, kpgfx, il, ids) = dtemp(1)+dtemp(2) &
                                                          +dtemp(3)+dtemp(4)
                     ENDDO
                  CASE (4)
                     ![s|r^8|s] integral
                     prefac = sqrt_pi3/16.0_dp/sqrt_zet**11
                     temp = 945.0_dp+2520.0_dp*pfac*rab2+1512.0_dp*(pfac*rab2)**2
                     temp = temp+288.0_dp*(pfac*rab2)**3+16.0_dp*(pfac*rab2)**4
                     sr_int = prefac*temp
                     s(ipgfa, jpgfb, kpgfx, il, 1) = exp_rab2*sr_int
                     !** derivatives
                     k = sqrt_pi3*x**8/sqrt_zet**19
                     dsr_int(1) = 2520.0_dp*pfac+3024.0_dp*pfac**2*rab2
                     dsr_int(1) = dsr_int(1)+864.0_dp*pfac**3*rab2**2 &
                                  +64.0_dp*pfac**4*rab2**3
                     dsr_int(1) = prefac*dsr_int(1)
                     dsr_int(2) = 3024.0_dp*pfac**2+1728.0_dp*pfac**3*rab2
                     dsr_int(2) = dsr_int(2)+192.0_dp*pfac**4*rab2**2
                     dsr_int(2) = prefac*dsr_int(2)
                     dsr_int(3) = 1728.0_dp*pfac**3+384.0_dp*pfac**4*rab2
                     dsr_int(3) = prefac*dsr_int(3)
                     DO ids = 2, nds
                        n = ids-1
                        dtemp(1) = (-xhi)**n*exp_rab2*sr_int
                        dtemp(2) = REAL(n, dp)*(-xhi)**(n-1)*exp_rab2*dsr_int(1)
                        dtemp(3) = REAL(n**2-n, dp)/2.0_dp*(-xhi)**(n-2) &
                                   *exp_rab2*dsr_int(2)
                        dtemp(4) = REAL(n*(n-1)*(n-2), dp)/6.0_dp*(-xhi)**(n-3) &
                                   *exp_rab2*dsr_int(3)
                        dtemp(5) = REAL(n*(n-1)*(n-2)*(n-3), dp)*(-xhi)**(n-4) &
                                   *k*exp_rab2
                        s(ipgfa, jpgfb, kpgfx, il, ids) = dtemp(1)+dtemp(2)+dtemp(3) &
                                                          +dtemp(4)+dtemp(5)
                     ENDDO
                  CASE (5)
                     ![s|r^10|s] integral
                     prefac = sqrt_pi3/32.0_dp/sqrt_zet**13
                     temp = 10395.0_dp+34650.0_dp*pfac*rab2
                     temp = temp+27720.0_dp*(pfac*rab2)**2+7920.0_dp*(pfac*rab2)**3
                     temp = temp+880.0_dp*(pfac*rab2)**4+32.0_dp*(pfac*rab2)**5
                     sr_int = prefac*temp
                     s(ipgfa, jpgfb, kpgfx, il, 1) = exp_rab2*sr_int
                     !** derivatives
                     k = sqrt_pi3*x**10/sqrt_zet**23
                     dsr_int(1) = 34650.0_dp*pfac+55440.0_dp*pfac**2*rab2
                     dsr_int(1) = dsr_int(1)+23760.0_dp*pfac**3*rab2**2
                     dsr_int(1) = dsr_int(1)+3520.0_dp*pfac**4*rab2**3
                     dsr_int(1) = dsr_int(1)+160.0_dp*pfac**5*rab2**4
                     dsr_int(1) = prefac*dsr_int(1)
                     dsr_int(2) = 55440.0_dp*pfac**2+47520.0_dp*pfac**3*rab2
                     dsr_int(2) = dsr_int(2)+10560.0_dp*pfac**4*rab2**2
                     dsr_int(2) = dsr_int(2)+640.0_dp*pfac**5*rab2**3
                     dsr_int(2) = prefac*dsr_int(2)
                     dsr_int(3) = 47520.0_dp*pfac**3+21120.0_dp*pfac**4*rab2
                     dsr_int(3) = dsr_int(3)+1920.0_dp*pfac**5*rab2**2
                     dsr_int(3) = prefac*dsr_int(3)
                     dsr_int(4) = 21120.0_dp*pfac**4+3840.0_dp*pfac**5*rab2
                     dsr_int(4) = prefac*dsr_int(4)
                     DO ids = 2, nds
                        n = ids-1
                        dtemp(1) = (-xhi)**n*exp_rab2*sr_int
                        dtemp(2) = REAL(n, dp)*(-xhi)**(n-1)*exp_rab2*dsr_int(1)
                        dtemp(3) = REAL(n**2-n, dp)/2.0_dp*(-xhi)**(n-2) &
                                   *exp_rab2*dsr_int(2)
                        dtemp(4) = REAL(n*(n-1)*(n-2), dp)/6.0_dp*(-xhi)**(n-3) &
                                   *exp_rab2*dsr_int(3)
                        dtemp(5) = REAL(n*(n-1)*(n-2)*(n-3), dp)/24.0_dp*(-xhi)**(n-4) &
                                   *exp_rab2*dsr_int(4)
                        dtemp(6) = REAL(n*(n-1)*(n-2)*(n-3)*(n-4), dp)*(-xhi)**(n-5) &
                                   *k*exp_rab2
                        s(ipgfa, jpgfb, kpgfx, il, ids) = dtemp(1)+dtemp(2)+dtemp(3) &
                                                          +dtemp(4)+dtemp(5)+dtemp(6)
                     ENDDO
                  CASE DEFAULT
                     CALL cp_abort(cp__l("aobasis/ai_shg_overlap.F",315), &
                                   "SHG integrals not implemented when sum of l quantum"// &
                                   " number of orbital and ri basis larger than 11")
                  END SELECT

               ENDDO

            END DO
         END DO
      END DO

      DEALLOCATE (dtemp, dsr_int)

      CALL timestop(handle)

   END SUBROUTINE s_overlap_abx

! **************************************************************************************************
!> \brief computes the real scaled solid harmonics Rlm up to a given l
!> \param Rlm_c cosine part of real scaled soldi harmonics
!> \param Rlm_s sine part of real scaled soldi harmonics
!> \param l maximal l quantum up to where Rlm is calculated
!> \param r distance vector between a and b
!> \param r2 square of distance vector
! **************************************************************************************************
   SUBROUTINE get_real_scaled_solid_harmonic(Rlm_c, Rlm_s, l, r, r2)

      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(0:l, -2*l:2*l), &
         INTENT(OUT)                                     :: Rlm_s, Rlm_c
      REAL(KIND=dp), DIMENSION(3)                        :: r
      REAL(KIND=dp)                                      :: r2

      CHARACTER(len=*), PARAMETER :: routineN = 'get_real_scaled_solid_harmonic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, li, mi, prefac
      REAL(KIND=dp)                                      :: Rc, Rc_00, Rlm, Rmlm, Rplm, Rs, Rs_00, &
                                                            temp_c

      CALL timeset(routineN, handle)

      Rc_00 = 1.0_dp
      Rs_00 = 0.0_dp

      Rlm_c(0, 0) = Rc_00
      Rlm_s(0, 0) = Rs_00

      ! generate elements Rmm
      ! start
      IF (l > 0) THEN
         Rc = -0.5_dp*r(1)*Rc_00
         Rs = -0.5_dp*r(2)*Rc_00
         Rlm_c(1, 1) = Rc
         Rlm_s(1, 1) = Rs
         Rlm_c(1, -1) = -Rc
         Rlm_s(1, -1) = Rs
      ENDIF
      DO li = 2, l
         temp_c = (-r(1)*Rc+r(2)*Rs)/(REAL(2*(li-1)+2, dp))
         Rs = (-r(2)*Rc-r(1)*Rs)/(REAL(2*(li-1)+2, dp))
         Rc = temp_c
         Rlm_c(li, li) = Rc
         Rlm_s(li, li) = Rs
         IF (MODULO(li, 2) /= 0) THEN
            Rlm_c(li, -li) = -Rc
            Rlm_s(li, -li) = Rs
         ELSE
            Rlm_c(li, -li) = Rc
            Rlm_s(li, -li) = -Rs
         ENDIF
      ENDDO

      DO mi = 0, l-1
         Rmlm = Rlm_c(mi, mi)
         Rlm = r(3)*Rlm_c(mi, mi)
         Rlm_c(mi+1, mi) = Rlm
         IF (MODULO(mi, 2) /= 0) THEN
            Rlm_c(mi+1, -mi) = -Rlm
         ELSE
            Rlm_c(mi+1, -mi) = Rlm
         ENDIF
         DO li = mi+2, l
            prefac = (li+mi)*(li-mi)
            Rplm = (REAL(2*li-1, dp)*r(3)*Rlm-r2*Rmlm)/REAL(prefac, dp)
            Rmlm = Rlm
            Rlm = Rplm
            Rlm_c(li, mi) = Rlm
            IF (MODULO(mi, 2) /= 0) THEN
               Rlm_c(li, -mi) = -Rlm
            ELSE
               Rlm_c(li, -mi) = Rlm
            ENDIF
         ENDDO
      ENDDO
      DO mi = 1, l-1
         Rmlm = Rlm_s(mi, mi)
         Rlm = r(3)*Rlm_s(mi, mi)
         Rlm_s(mi+1, mi) = Rlm
         IF (MODULO(mi, 2) /= 0) THEN
            Rlm_s(mi+1, -mi) = Rlm
         ELSE
            Rlm_s(mi+1, -mi) = -Rlm
         ENDIF
         DO li = mi+2, l
            prefac = (li+mi)*(li-mi)
            Rplm = (REAL(2*li-1, dp)*r(3)*Rlm-r2*Rmlm)/REAL(prefac, dp)
            Rmlm = Rlm
            Rlm = Rplm
            Rlm_s(li, mi) = Rlm
            IF (MODULO(mi, 2) /= 0) THEN
               Rlm_s(li, -mi) = Rlm
            ELSE
               Rlm_s(li, -mi) = -Rlm
            ENDIF
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE get_real_scaled_solid_harmonic

! **************************************************************************************************
!> \brief calculates the angular dependent-part of the SHG integrals,
!>        transformation matrix W, see literature above
!> \param lamax array of maximal l quantum number on a;
!>        lamax(lb) with lb= 0..lbmax
!> \param lbmax maximal l quantum number on b
!> \param lmax maximal l quantum number
!> \param Rc cosine part of real scaled solid harmonics
!> \param Rs sine part of real scaled solid harmonics
!> \param Waux_mat stores the angular-dependent part of the SHG integrals
!>        last dimension is (1:4): cc(1), cs(2), sc(3), ss(4)
! **************************************************************************************************
   SUBROUTINE get_W_matrix(lamax, lbmax, lmax, Rc, Rs, Waux_mat)

      INTEGER, DIMENSION(:), POINTER                     :: lamax
      INTEGER, INTENT(IN)                                :: lbmax, lmax
      REAL(KIND=dp), DIMENSION(0:lmax, -2*lmax:2*lmax), &
         INTENT(IN)                                      :: Rc, Rs
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat

      CHARACTER(len=*), PARAMETER :: routineN = 'get_W_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, j, k, la, labmin, laj, lb, lbj, &
                                                            ma, ma_m, ma_p, mb, mb_m, mb_p, nla, &
                                                            nlb
      REAL(KIND=dp) :: A_jk, A_lama, A_lbmb, Alm_fac, delta_k, prefac, Rca_m, Rca_p, Rcb_m, Rcb_p, &
         Rsa_m, Rsa_p, Rsb_m, Rsb_p, sign_fac, Wa(4), Wb(4), Wmat(4)

      CALL timeset(routineN, handle)

      Wa(:) = 0.0_dp
      Wb(:) = 0.0_dp
      Wmat(:) = 0.0_dp

      DO lb = 0, lbmax
         nlb = nsoset_pm(lb-1)
         DO la = 0, lamax(lb)
            nla = nsoset_pm(la-1)
            labmin = MIN(la, lb)
            DO mb = 0, lb
               A_lbmb = SQRT(fac(lb+mb)*fac(lb-mb))
               IF (MODULO(mb, 2) /= 0) A_lbmb = -A_lbmb
               IF (mb /= 0) A_lbmb = A_lbmb*SQRT(2.0_dp)
               IF (MODULO(lb, 2) /= 0) A_lbmb = -A_lbmb
               DO ma = 0, la
                  A_lama = SQRT(fac(la+ma)*fac(la-ma))
                  IF (MODULO(ma, 2) /= 0) A_lama = -A_lama
                  IF (ma /= 0) A_lama = A_lama*SQRT(2.0_dp)
                  Alm_fac = A_lama*A_lbmb
                  DO j = 0, labmin
                     laj = la-j
                     lbj = lb-j
                     prefac = Alm_fac*REAL(2**(la+lb-j), dp)*dfac(2*j-1)
                     delta_k = 0.5_dp
                     Wmat = 0.0_dp
                     DO k = 0, j
                        ma_m = ma-k
                        ma_p = ma+k
                        IF (laj < ABS(ma_m) .AND. laj < ABS(ma_p)) CYCLE
                        mb_m = mb-k
                        mb_p = mb+k
                        IF (lbj < ABS(mb_m) .AND. lbj < ABS(mb_p)) CYCLE
                        IF (k /= 0) delta_k = 1.0_dp
                        A_jk = fac(j+k)*fac(j-k)
                        IF (k /= 0) A_jk = 2.0_dp*A_jk
                        IF (MODULO(k, 2) /= 0) THEN
                           sign_fac = -1.0_dp
                        ELSE
                           sign_fac = 1.0_dp
                        ENDIF
                        Rca_m = Rc(laj, ma_m)
                        Rsa_m = Rs(laj, ma_m)
                        Rca_p = Rc(laj, ma_p)
                        Rsa_p = Rs(laj, ma_p)
                        Rcb_m = Rc(lbj, mb_m)
                        Rsb_m = Rs(lbj, mb_m)
                        Rcb_p = Rc(lbj, mb_p)
                        Rsb_p = Rs(lbj, mb_p)
                        Wa(1) = delta_k*(Rca_m+sign_fac*Rca_p)
                        Wb(1) = delta_k*(Rcb_m+sign_fac*Rcb_p)
                        Wa(2) = -Rsa_m+sign_fac*Rsa_p
                        Wb(2) = -Rsb_m+sign_fac*Rsb_p
                        Wmat(1) = Wmat(1)+prefac/A_jk*(Wa(1)*Wb(1)+Wa(2)*Wb(2))
                        IF (mb > 0) THEN
                           Wb(3) = delta_k*(Rsb_m+sign_fac*Rsb_p)
                           Wb(4) = Rcb_m-sign_fac*Rcb_p
                           Wmat(2) = Wmat(2)+prefac/A_jk*(Wa(1)*Wb(3)+Wa(2)*Wb(4))
                        ENDIF
                        IF (ma > 0) THEN
                           Wa(3) = delta_k*(Rsa_m+sign_fac*Rsa_p)
                           Wa(4) = Rca_m-sign_fac*Rca_p
                           Wmat(3) = Wmat(3)+prefac/A_jk*(Wa(3)*Wb(1)+Wa(4)*Wb(2))
                        ENDIF
                        IF (ma > 0 .AND. mb > 0) THEN
                           Wmat(4) = Wmat(4)+prefac/A_jk*(Wa(3)*Wb(3)+Wa(4)*Wb(4))
                        ENDIF
                     ENDDO
                     Waux_mat(nla+ma+1, nlb+mb+1, j+1, 1) = Wmat(1)
                     IF (mb > 0) Waux_mat(nla+ma+1, nlb+mb+1, j+1, 2) = Wmat(2)
                     IF (ma > 0) Waux_mat(nla+ma+1, nlb+mb+1, j+1, 3) = Wmat(3)
                     IF (ma > 0 .AND. mb > 0) Waux_mat(nla+ma+1, nlb+mb+1, j+1, 4) = Wmat(4)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE get_W_matrix

! **************************************************************************************************
!> \brief calculates derivatives of transformation matrix W,
!> \param lamax array of maximal l quantum number on a;
!>        lamax(lb) with lb= 0..lbmax
!> \param lbmax maximal l quantum number on b
!> \param Waux_mat stores the angular-dependent part of the SHG integrals
!> \param dWaux_mat stores the derivatives of the angular-dependent part of
!>        the SHG integrals
!>        last dimension is (1:4): cc(1), cs(2), sc(3), ss(4)
! **************************************************************************************************
   SUBROUTINE get_dW_matrix(lamax, lbmax, Waux_mat, dWaux_mat)

      INTEGER, DIMENSION(:), POINTER                     :: lamax
      INTEGER, INTENT(IN)                                :: lbmax
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat

      CHARACTER(len=*), PARAMETER :: routineN = 'get_dW_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ia, iam, ib, ibm, j, jmax, la, &
                                                            labm, labmin, lamb, lb, ma, mb, nla, &
                                                            nlam, nlb, nlbm
      REAL(KIND=dp)                                      :: bma, bma_m, bma_p, bmb, bmb_m, bmb_p, &
                                                            dAa, dAa_m, dAa_p, dAb, dAb_m, dAb_p
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Wam, Wamm, Wamp, Wbm, Wbmm, Wbmp

      CALL timeset(routineN, handle)

      jmax = MIN(MAXVAL(lamax), lbmax)
      ALLOCATE (Wam(0:jmax, 4), Wamm(0:jmax, 4), Wamp(0:jmax, 4))
      ALLOCATE (Wbm(0:jmax, 4), Wbmm(0:jmax, 4), Wbmp(0:jmax, 4))

      DO lb = 0, lbmax
         nlb = nsoset_pm(lb-1)
         nlbm = 0
         IF (lb > 0) nlbm = nsoset_pm(lb-2)
         DO la = 0, lamax(lb)
            nla = nsoset_pm(la-1)
            nlam = 0
            IF (la > 0) nlam = nsoset_pm(la-2)
            labmin = MIN(la, lb)
            lamb = MIN(la-1, lb)
            labm = MIN(la, lb-1)
            DO mb = 0, lb
               bmb = 1.0_dp
               bmb_m = 1.0_dp
               bmb_p = 1.0_dp
               IF (mb /= 0) bmb = SQRT(2.0_dp)
               IF (mb-1 /= 0) bmb_m = SQRT(2.0_dp)
               IF (mb+1 /= 0) bmb_p = SQRT(2.0_dp)
               dAb_p = -bmb/bmb_p*SQRT(REAL((lb-mb)*(lb-mb-1), dp))
               dAb_m = -bmb/bmb_m*SQRT(REAL((lb+mb)*(lb+mb-1), dp))
               dAb = 2.0_dp*SQRT(REAL((lb+mb)*(lb-mb), dp))
               !*** for m=0 W_l-1,-1 can't be read from Waux_mat, but we use
               !*** W_l-1,-1 = -W_l-1,1 [cc(1), cs(2)] or W_l-1,-1 = W_l-1,1 [[sc(3), ss(4)]
               IF (mb == 0) dAb_p = 2.0_dp*dAb_p
               ib = nlb+mb+1
               ibm = nlbm+mb+1
               DO ma = 0, la
                  bma = 1.0_dp
                  bma_m = 1.0_dp
                  bma_p = 1.0_dp
                  IF (ma /= 0) bma = SQRT(2.0_dp)
                  IF (ma-1 /= 0) bma_m = SQRT(2.0_dp)
                  IF (ma+1 /= 0) bma_p = SQRT(2.0_dp)
                  dAa_p = -bma/bma_p*SQRT(REAL((la-ma)*(la-ma-1), dp))
                  dAa_m = -bma/bma_m*SQRT(REAL((la+ma)*(la+ma-1), dp))
                  dAa = 2.0_dp*SQRT(REAL((la+ma)*(la-ma), dp))
                  IF (ma == 0) dAa_p = 2.0_dp*dAa_p
                  ia = nla+ma+1
                  iam = nlam+ma+1
                  Wam(:, :) = 0.0_dp
                  Wamm(:, :) = 0.0_dp
                  Wamp(:, :) = 0.0_dp
                  !*** Wam: la-1, ma
                  IF (ma <= la-1) THEN
                     Wam(0:lamb, 1) = Waux_mat(iam, ib, 1:lamb+1, 1)
                     IF (mb > 0) Wam(0:lamb, 2) = Waux_mat(iam, ib, 1:lamb+1, 2)
                     IF (ma > 0) Wam(0:lamb, 3) = Waux_mat(iam, ib, 1:lamb+1, 3)
                     IF (ma > 0 .AND. mb > 0) Wam(0:lamb, 4) = Waux_mat(iam, ib, 1:lamb+1, 4)
                  ENDIF
                  !*** Wamm: la-1, ma-1
                  IF (ma-1 >= 0) THEN
                     Wamm(0:lamb, 1) = Waux_mat(iam-1, ib, 1:lamb+1, 1)
                     IF (mb > 0) Wamm(0:lamb, 2) = Waux_mat(iam-1, ib, 1:lamb+1, 2)
                     IF (ma-1 > 0) Wamm(0:lamb, 3) = Waux_mat(iam-1, ib, 1:lamb+1, 3)
                     IF (ma-1 > 0 .AND. mb > 0) Wamm(0:lamb, 4) = Waux_mat(iam-1, ib, 1:lamb+1, 4)
                  ENDIF
                  !*** Wamp: la-1, ma+1
                  IF (ma+1 <= la-1) THEN
                     Wamp(0:lamb, 1) = Waux_mat(iam+1, ib, 1:lamb+1, 1)
                     IF (mb > 0) Wamp(0:lamb, 2) = Waux_mat(iam+1, ib, 1:lamb+1, 2)
                     IF (ma+1 > 0) Wamp(0:lamb, 3) = Waux_mat(iam+1, ib, 1:lamb+1, 3)
                     IF (ma+1 > 0 .AND. mb > 0) Wamp(0:lamb, 4) = Waux_mat(iam+1, ib, 1:lamb+1, 4)
                  ENDIF
                  Wbm(:, :) = 0.0_dp
                  Wbmm(:, :) = 0.0_dp
                  Wbmp(:, :) = 0.0_dp
                  !*** Wbm: lb-1, mb
                  IF (mb <= lb-1) THEN
                     Wbm(0:labm, 1) = Waux_mat(ia, ibm, 1:labm+1, 1)
                     IF (mb > 0) Wbm(0:labm, 2) = Waux_mat(ia, ibm, 1:labm+1, 2)
                     IF (ma > 0) Wbm(0:labm, 3) = Waux_mat(ia, ibm, 1:labm+1, 3)
                     IF (ma > 0 .AND. mb > 0) Wbm(0:labm, 4) = Waux_mat(ia, ibm, 1:labm+1, 4)
                  ENDIF
                  !*** Wbmm: lb-1, mb-1
                  IF (mb-1 >= 0) THEN
                     Wbmm(0:labm, 1) = Waux_mat(ia, ibm-1, 1:labm+1, 1)
                     IF (mb-1 > 0) Wbmm(0:labm, 2) = Waux_mat(ia, ibm-1, 1:labm+1, 2)
                     IF (ma > 0) Wbmm(0:labm, 3) = Waux_mat(ia, ibm-1, 1:labm+1, 3)
                     IF (ma > 0 .AND. mb-1 > 0) Wbmm(0:labm, 4) = Waux_mat(ia, ibm-1, 1:labm+1, 4)
                  ENDIF
                  !*** Wbmp: lb-1, mb+1
                  IF (mb+1 <= lb-1) THEN
                     Wbmp(0:labm, 1) = Waux_mat(ia, ibm+1, 1:labm+1, 1)
                     IF (mb+1 > 0) Wbmp(0:labm, 2) = Waux_mat(ia, ibm+1, 1:labm+1, 2)
                     IF (ma > 0) Wbmp(0:labm, 3) = Waux_mat(ia, ibm+1, 1:labm+1, 3)
                     IF (ma > 0 .AND. mb+1 > 0) Wbmp(0:labm, 4) = Waux_mat(ia, ibm+1, 1:labm+1, 4)
                  ENDIF
                  DO j = 0, labmin
                     !*** x component
                     dWaux_mat(1, ia, ib, j+1, 1) = dAa_p*Wamp(j, 1)-dAa_m*Wamm(j, 1) &
                                                    -dAb_p*Wbmp(j, 1)+dAb_m*Wbmm(j, 1)
                     IF (mb > 0) THEN
                        dWaux_mat(1, ia, ib, j+1, 2) = dAa_p*Wamp(j, 2)-dAa_m*Wamm(j, 2) &
                                                       -dAb_p*Wbmp(j, 2)+dAb_m*Wbmm(j, 2)
                     ENDIF
                     IF (ma > 0) THEN
                        dWaux_mat(1, ia, ib, j+1, 3) = dAa_p*Wamp(j, 3)-dAa_m*Wamm(j, 3) &
                                                       -dAb_p*Wbmp(j, 3)+dAb_m*Wbmm(j, 3)
                     ENDIF
                     IF (ma > 0 .AND. mb > 0) THEN
                        dWaux_mat(1, ia, ib, j+1, 4) = dAa_p*Wamp(j, 4)-dAa_m*Wamm(j, 4) &
                                                       -dAb_p*Wbmp(j, 4)+dAb_m*Wbmm(j, 4)
                     ENDIF
                     !**** y component
                     dWaux_mat(2, ia, ib, j+1, 1) = dAa_p*Wamp(j, 3)+dAa_m*Wamm(j, 3) &
                                                    -dAb_p*Wbmp(j, 2)-dAb_m*Wbmm(j, 2)
                     IF (mb > 0) THEN
                        dWaux_mat(2, ia, ib, j+1, 2) = dAa_p*Wamp(j, 4)+dAa_m*Wamm(j, 4) &
                                                       +dAb_p*Wbmp(j, 1)+dAb_m*Wbmm(j, 1)
                     ENDIF
                     IF (ma > 0) THEN
                        dWaux_mat(2, ia, ib, j+1, 3) = -dAa_p*Wamp(j, 1)-dAa_m*Wamm(j, 1) &
                                                       -dAb_p*Wbmp(j, 4)-dAb_m*Wbmm(j, 4)
                     ENDIF
                     IF (ma > 0 .AND. mb > 0) THEN
                        dWaux_mat(2, ia, ib, j+1, 4) = -dAa_p*Wamp(j, 2)-dAa_m*Wamm(j, 2) &
                                                       +dAb_p*Wbmp(j, 3)+dAb_m*Wbmm(j, 3)
                     ENDIF
                     !**** z compnent
                     dWaux_mat(3, ia, ib, j+1, 1) = dAa*Wam(j, 1)-dAb*Wbm(j, 1)
                     IF (mb > 0) THEN
                        dWaux_mat(3, ia, ib, j+1, 2) = dAa*Wam(j, 2)-dAb*Wbm(j, 2)
                     ENDIF
                     IF (ma > 0) THEN
                        dWaux_mat(3, ia, ib, j+1, 3) = dAa*Wam(j, 3)-dAb*Wbm(j, 3)
                     ENDIF
                     IF (ma > 0 .AND. mb > 0) THEN
                        dWaux_mat(3, ia, ib, j+1, 4) = dAa*Wam(j, 4)-dAb*Wbm(j, 4)
                     ENDIF
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      DEALLOCATE (Wam, Wamm, Wamp)
      DEALLOCATE (Wbm, Wbmm, Wbmp)

      CALL timestop(handle)

   END SUBROUTINE get_dW_matrix

! **************************************************************************************************
!> \brief calculates [ab] SHG overlap integrals using precomputed angular-
!>        dependent part
!> \param la set of l quantum number on a
!> \param first_sgfa indexing
!> \param nshella number of shells for a
!> \param lb set of l quantum number on b
!> \param first_sgfb indexing
!> \param nshellb number of shells for b
!> \param swork_cont contracted and normalized [s|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param sab contracted overlap of spherical harmonic Gaussians
! **************************************************************************************************
   SUBROUTINE overlap_shg_ab(la, first_sgfa, nshella, lb, first_sgfb, nshellb, &
                             swork_cont, Waux_mat, sab)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: sab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_shg_ab', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ia, iaa, ishella, j, jb, jbb, &
                                                            jshellb, labmin, lai, lbj, mai, mbj, &
                                                            nla, nlb, sgfa, sgfb
      REAL(KIND=dp)                                      :: prefac

      CALL timeset(routineN, handle)

      DO ishella = 1, nshella
         lai = la(ishella)
         nla = nsoset_pm(lai-1)
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            labmin = MIN(lai, lbj)
            DO j = 0, labmin
               prefac = swork_cont(ishella, jshellb, lai+lbj-j+1)
               ! mai = 0 and mbj = 0
               iaa = nla+1
               jbb = nlb+1
               sab(ia, jb) = sab(ia, jb)+prefac*Waux_mat(iaa, jbb, j+1, 1)
               ! mai = 0 and mbj /= 0
               DO mbj = 1, lbj
                  jbb = nlb+mbj+1
                  sab(ia, jb+mbj) = sab(ia, jb+mbj)+prefac*Waux_mat(iaa, jbb, j+1, 1)
                  sab(ia, jb-mbj) = sab(ia, jb-mbj)+prefac*Waux_mat(iaa, jbb, j+1, 2)
               ENDDO
               ! mai /= 0 and mbj = 0
               jbb = nlb+1
               DO mai = 1, lai
                  iaa = nla+mai+1
                  sab(ia+mai, jb) = sab(ia+mai, jb)+prefac*Waux_mat(iaa, jbb, j+1, 1)
                  sab(ia-mai, jb) = sab(ia-mai, jb)+prefac*Waux_mat(iaa, jbb, j+1, 3)
               ENDDO
               ! mai /= 0 and mbj /= 0
               DO mai = 1, lai
                  iaa = nla+mai+1
                  DO mbj = 1, lbj
                     jbb = nlb+mbj+1
                     sab(ia+mai, jb+mbj) = sab(ia+mai, jb+mbj)+prefac*Waux_mat(iaa, jbb, j+1, 1)
                     sab(ia+mai, jb-mbj) = sab(ia+mai, jb-mbj)+prefac*Waux_mat(iaa, jbb, j+1, 2)
                     sab(ia-mai, jb+mbj) = sab(ia-mai, jb+mbj)+prefac*Waux_mat(iaa, jbb, j+1, 3)
                     sab(ia-mai, jb-mbj) = sab(ia-mai, jb-mbj)+prefac*Waux_mat(iaa, jbb, j+1, 4)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE overlap_shg_ab

! **************************************************************************************************
!> \brief calculates derivatives of [ab] SHG overlap integrals using precomputed
!>        angular-dependent part
!> \param la set of l quantum number on a
!> \param first_sgfa indexing
!> \param nshella number of shells for a
!> \param lb set of l quantum number on b
!> \param first_sgfb indexing
!> \param nshellb number of shells for b
!> \param rab distance vector Ra-Rb
!> \param swork_cont contracted and normalized [s|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param dWaux_mat derivatives of precomputed angular-dependent part
!> \param dsab derivative of contracted overlap of spherical harmonic Gaussians
! **************************************************************************************************
   SUBROUTINE dev_overlap_shg_ab(la, first_sgfa, nshella, lb, first_sgfb, nshellb, rab, &
                                 swork_cont, Waux_mat, dWaux_mat, dsab)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      REAL(KIND=dp), INTENT(IN)                          :: rab(3)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: dWaux_mat
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: dsab

      CHARACTER(len=*), PARAMETER :: routineN = 'dev_overlap_shg_ab', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ia, iaa, ishella, j, jb, jbb, &
                                                            jshellb, labmin, lai, lbj, mai, mbj, &
                                                            nla, nlb, sgfa, sgfb
      REAL(KIND=dp)                                      :: dprefac, prefac, rabx2(3)

      CALL timeset(routineN, handle)

      rabx2(:) = 2.0_dp*rab
      DO ishella = 1, nshella
         lai = la(ishella)
         nla = nsoset_pm(lai-1)
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            labmin = MIN(lai, lbj)
            DO j = 0, labmin
               prefac = swork_cont(ishella, jshellb, lai+lbj-j+1)
               dprefac = swork_cont(ishella, jshellb, lai+lbj-j+2) !j+1
               ! mai = 0 and mbj = 0
               iaa = nla+1
               jbb = nlb+1
               DO i = 1, 3
                  dsab(ia, jb, i) = dsab(ia, jb, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 1) &
                                    +prefac*dWaux_mat(i, iaa, jbb, j+1, 1)
               ENDDO
               ! mai = 0 and mbj /= 0
               DO mbj = 1, lbj
                  jbb = nlb+mbj+1
                  DO i = 1, 3
                     dsab(ia, jb+mbj, i) = dsab(ia, jb+mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 1) &
                                           +prefac*dWaux_mat(i, iaa, jbb, j+1, 1)
                     dsab(ia, jb-mbj, i) = dsab(ia, jb-mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 2) &
                                           +prefac*dWaux_mat(i, iaa, jbb, j+1, 2)
                  ENDDO
               ENDDO
               ! mai /= 0 and mbj = 0
               jbb = nlb+1
               DO mai = 1, lai
                  iaa = nla+mai+1
                  DO i = 1, 3
                     dsab(ia+mai, jb, i) = dsab(ia+mai, jb, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 1) &
                                           +prefac*dWaux_mat(i, iaa, jbb, j+1, 1)
                     dsab(ia-mai, jb, i) = dsab(ia-mai, jb, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 3) &
                                           +prefac*dWaux_mat(i, iaa, jbb, j+1, 3)
                  ENDDO
               ENDDO
               ! mai /= 0 and mbj /= 0
               DO mai = 1, lai
                  iaa = nla+mai+1
                  DO mbj = 1, lbj
                     jbb = nlb+mbj+1
                     DO i = 1, 3
                        dsab(ia+mai, jb+mbj, i) = dsab(ia+mai, jb+mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 1) &
                                                  +prefac*dWaux_mat(i, iaa, jbb, j+1, 1)
                        dsab(ia+mai, jb-mbj, i) = dsab(ia+mai, jb-mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 2) &
                                                  +prefac*dWaux_mat(i, iaa, jbb, j+1, 2)
                        dsab(ia-mai, jb+mbj, i) = dsab(ia-mai, jb+mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 3) &
                                                  +prefac*dWaux_mat(i, iaa, jbb, j+1, 3)
                        dsab(ia-mai, jb-mbj, i) = dsab(ia-mai, jb-mbj, i)+rabx2(i)*dprefac*Waux_mat(iaa, jbb, j+1, 4) &
                                                  +prefac*dWaux_mat(i, iaa, jbb, j+1, 4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE dev_overlap_shg_ab

! **************************************************************************************************
!> \brief calculates [aba] SHG overlap integrals using precomputed angular-
!>        dependent part
!> \param la set of l quantum number on a, orbital basis
!> \param first_sgfa indexing
!> \param nshella number of shells for a, orbital basis
!> \param lb set of l quantum number on b. orbital basis
!> \param first_sgfb indexing
!> \param nshellb number of shells for b, orbital basis
!> \param lca of l quantum number on a, aux basis
!> \param first_sgfca indexing
!> \param nshellca  number of shells for a, aux basis
!> \param cg_coeff Clebsch-Gordon coefficients
!> \param cg_none0_list list of none-zero Clebsch-Gordon coefficients
!> \param ncg_none0 number of non-zero Clebsch-Gordon coefficients
!> \param swork_cont contracted and normalized [s|ra^n|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param saba contracted overlap [aba] of spherical harmonic Gaussians
! **************************************************************************************************
   SUBROUTINE overlap_shg_aba(la, first_sgfa, nshella, lb, first_sgfb, nshellb, &
                              lca, first_sgfca, nshellca, cg_coeff, cg_none0_list, &
                              ncg_none0, swork_cont, Waux_mat, saba)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lca, first_sgfca
      INTEGER, INTENT(IN)                                :: nshellca
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: cg_coeff
      INTEGER, DIMENSION(:, :, :), INTENT(IN)            :: cg_none0_list
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: ncg_none0
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: saba

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_shg_aba', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, ia, il, ilist, ishella, isoa1, isoa2, isoaa, j, jb, jshellb, ka, kshella, &
         laa, labmin, lai, lak, lbj, maa, mai, mak, mbj, nla, nlb, sgfa, sgfb, sgfca
      REAL(KIND=dp)                                      :: prefac, stemp

      CALL timeset(routineN, handle)

      DO ishella = 1, nshella
         lai = la(ishella)
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            DO kshella = 1, nshellca
               lak = lca(kshella)
               sgfca = first_sgfca(kshella)
               ka = sgfca+lak
               DO mai = -lai, lai, 1
                  DO mbj = -lbj, lbj, 1
                     DO mak = -lak, lak, 1
                        isoa1 = indso_inv(lai, mai)
                        isoa2 = indso_inv(lak, mak)
                        DO ilist = 1, ncg_none0(isoa1, isoa2)
                           isoaa = cg_none0_list(isoa1, isoa2, ilist)
                           laa = indso(1, isoaa)
                           maa = indso(2, isoaa)
                           nla = nsoset_pm(laa-1)
                           labmin = MIN(laa, lbj)
                           il = INT((lai+lak-laa)/2)
                           stemp = 0.0_dp
                           DO j = 0, labmin
                              prefac = swork_cont(ishella, jshellb, kshella, il, laa+lbj-j+1)
                              IF (maa >= 0 .AND. mbj >= 0) stemp = stemp+prefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 1)
                              IF (maa >= 0 .AND. mbj < 0) stemp = stemp+prefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 2)
                              IF (maa < 0 .AND. mbj >= 0) stemp = stemp+prefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 3)
                              IF (maa < 0 .AND. mbj < 0) stemp = stemp+prefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 4)
                           ENDDO
                           saba(ia+mai, jb+mbj, ka+mak) = saba(ia+mai, jb+mbj, ka+mak)+cg_coeff(isoa1, isoa2, isoaa)*stemp
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE overlap_shg_aba

! **************************************************************************************************
!> \brief calculates derivatives of [aba] SHG overlap integrals using
!>        precomputed angular-dependent part
!> \param la set of l quantum number on a, orbital basis
!> \param first_sgfa indexing
!> \param nshella number of shells for a, orbital basis
!> \param lb set of l quantum number on b. orbital basis
!> \param first_sgfb indexing
!> \param nshellb number of shells for b, orbital basis
!> \param lca of l quantum number on a, aux basis
!> \param first_sgfca indexing
!> \param nshellca  number of shells for a, aux basis
!> \param cg_coeff Clebsch-Gordon coefficients
!> \param cg_none0_list list of none-zero Clebsch-Gordon coefficients
!> \param ncg_none0 number of non-zero Clebsch-Gordon coefficients
!> \param rab distance vector Ra-Rb
!> \param swork_cont contracted and normalized [s|ra^n|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param dWaux_mat derivatives of precomputed angular-dependent part
!> \param dsaba derivative of contracted overlap [aba] of spherical harmonic
!>              Gaussians
! **************************************************************************************************
   SUBROUTINE dev_overlap_shg_aba(la, first_sgfa, nshella, lb, first_sgfb, nshellb, &
                                  lca, first_sgfca, nshellca, cg_coeff, cg_none0_list, &
                                  ncg_none0, rab, swork_cont, Waux_mat, dWaux_mat, dsaba)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lca, first_sgfca
      INTEGER, INTENT(IN)                                :: nshellca
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: cg_coeff
      INTEGER, DIMENSION(:, :, :), INTENT(IN)            :: cg_none0_list
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: ncg_none0
      REAL(KIND=dp), INTENT(IN)                          :: rab(3)
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: dWaux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT)                                   :: dsaba

      CHARACTER(len=*), PARAMETER :: routineN = 'dev_overlap_shg_aba', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ia, il, ilist, ishella, isoa1, isoa2, isoaa, j, jb, jshellb, ka, &
         kshella, laa, labmin, lai, lak, lbj, maa, mai, mak, mbj, nla, nlb, sgfa, sgfb, sgfca
      REAL(KIND=dp)                                      :: dprefac, dtemp(3), prefac, rabx2(3)

      CALL timeset(routineN, handle)

      rabx2(:) = 2.0_dp*rab

      DO ishella = 1, nshella
         lai = la(ishella)
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            DO kshella = 1, nshellca
               lak = lca(kshella)
               sgfca = first_sgfca(kshella)
               ka = sgfca+lak
               DO mai = -lai, lai, 1
                  DO mbj = -lbj, lbj, 1
                     DO mak = -lak, lak, 1
                        isoa1 = indso_inv(lai, mai)
                        isoa2 = indso_inv(lak, mak)
                        DO ilist = 1, ncg_none0(isoa1, isoa2)
                           isoaa = cg_none0_list(isoa1, isoa2, ilist)
                           laa = indso(1, isoaa)
                           maa = indso(2, isoaa)
                           nla = nsoset_pm(laa-1)
                           labmin = MIN(laa, lbj)
                           il = INT((lai+lak-laa)/2)
                           dtemp = 0.0_dp
                           DO j = 0, labmin
                              prefac = swork_cont(ishella, jshellb, kshella, il, laa+lbj-j+1)
                              dprefac = swork_cont(ishella, jshellb, kshella, il, laa+lbj-j+2)
                              DO i = 1, 3
                                 IF (maa >= 0 .AND. mbj >= 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 1) &
                                               +prefac*dWaux_mat(i, nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 1)
                                 ELSEIF (maa >= 0 .AND. mbj < 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 2) &
                                               +prefac*dWaux_mat(i, nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 2)
                                 ELSEIF (maa < 0 .AND. mbj >= 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 3) &
                                               +prefac*dWaux_mat(i, nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 3)
                                 ELSEIF (maa < 0 .AND. mbj < 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 4) &
                                               +prefac*dWaux_mat(i, nla+ABS(maa)+1, nlb+ABS(mbj)+1, j+1, 4)
                                 ENDIF
                              ENDDO
                           ENDDO
                           DO i = 1, 3
                              dsaba(ia+mai, jb+mbj, ka+mak, i) = dsaba(ia+mai, jb+mbj, ka+mak, i) &
                                                                 +cg_coeff(isoa1, isoa2, isoaa)*dtemp(i)
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE dev_overlap_shg_aba

! **************************************************************************************************
!> \brief calculates [abb] SHG overlap integrals using precomputed angular-
!>        dependent part
!> \param la set of l quantum number on a, orbital basis
!> \param first_sgfa indexing
!> \param nshella number of shells for a, orbital basis
!> \param lb set of l quantum number on b. orbital basis
!> \param first_sgfb indexing
!> \param nshellb number of shells for b, orbital basis
!> \param lcb l quantum number on b, aux basis
!> \param first_sgfcb indexing
!> \param nshellcb number of shells for b, aux basis
!> \param cg_coeff Clebsch-Gordon coefficients
!> \param cg_none0_list list of none-zero Clebsch-Gordon coefficients
!> \param ncg_none0 number of non-zero Clebsch-Gordon coefficients
!> \param swork_cont contracted and normalized [s|rb^n|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param sabb contracted overlap [abb] of spherical harmonic Gaussians
! **************************************************************************************************
   SUBROUTINE overlap_shg_abb(la, first_sgfa, nshella, lb, first_sgfb, nshellb, &
                              lcb, first_sgfcb, nshellcb, cg_coeff, cg_none0_list, &
                              ncg_none0, swork_cont, Waux_mat, sabb)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lcb, first_sgfcb
      INTEGER, INTENT(IN)                                :: nshellcb
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: cg_coeff
      INTEGER, DIMENSION(:, :, :), INTENT(IN)            :: cg_none0_list
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: ncg_none0
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: sabb

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_shg_abb', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, ia, il, ilist, ishella, isob1, isob2, isobb, j, jb, jshellb, kb, kshellb, &
         labmin, lai, lbb, lbj, lbk, mai, mbb, mbj, mbk, nla, nlb, sgfa, sgfb, sgfcb
      REAL(KIND=dp)                                      :: prefac, stemp, tsign

      CALL timeset(routineN, handle)

      DO ishella = 1, nshella
         lai = la(ishella)
         nla = nsoset_pm(lai-1)+1
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            DO kshellb = 1, nshellcb
               lbk = lcb(kshellb)
               sgfcb = first_sgfcb(kshellb)
               kb = sgfcb+lbk
               DO mai = -lai, lai, 1
                  DO mbj = -lbj, lbj, 1
                     DO mbk = -lbk, lbk, 1
                        isob1 = indso_inv(lbj, mbj)
                        isob2 = indso_inv(lbk, mbk)
                        DO ilist = 1, ncg_none0(isob1, isob2)
                           isobb = cg_none0_list(isob1, isob2, ilist)
                           lbb = indso(1, isobb)
                           mbb = indso(2, isobb)
                           nlb = nsoset_pm(lbb-1)+1
                           tsign = 1.0_dp
                           IF (MODULO(lbb-lai, 2) /= 0) tsign = -1.0_dp
                           labmin = MIN(lai, lbb)
                           il = INT((lbj+lbk-lbb)/2)
                           stemp = 0.0_dp
                           DO j = 0, labmin
                              prefac = swork_cont(ishella, jshellb, kshellb, il, lai+lbb-j+1)
                              IF (mai >= 0 .AND. mbb >= 0) stemp = stemp+prefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 1)
                              ! take the transpose, since auxmat calculated for (lbb,lai)
                              IF (mai >= 0 .AND. mbb < 0) stemp = stemp+prefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 3)
                              IF (mai < 0 .AND. mbb >= 0) stemp = stemp+prefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 2)
                              IF (mai < 0 .AND. mbb < 0) stemp = stemp+prefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 4)
                           ENDDO
                           sabb(ia+mai, jb+mbj, kb+mbk) = sabb(ia+mai, jb+mbj, kb+mbk)+tsign*cg_coeff(isob1, isob2, isobb)*stemp
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE overlap_shg_abb

! **************************************************************************************************
!> \brief calculates derivatives of [abb] SHG overlap integrals using
!>        precomputed angular-dependent part
!> \param la set of l quantum number on a, orbital basis
!> \param first_sgfa indexing
!> \param nshella number of shells for a, orbital basis
!> \param lb set of l quantum number on b. orbital basis
!> \param first_sgfb indexing
!> \param nshellb number of shells for b, orbital basis
!> \param lcb l quantum number on b, aux basis
!> \param first_sgfcb indexing
!> \param nshellcb number of shells for b, aux basis
!> \param cg_coeff Clebsch-Gordon coefficients
!> \param cg_none0_list list of none-zero Clebsch-Gordon coefficients
!> \param ncg_none0 number of non-zero Clebsch-Gordon coefficients
!> \param rab distance vector Ra-Rb
!> \param swork_cont contracted and normalized [s|rb^n|s] integrals
!> \param Waux_mat precomputed angular-dependent part
!> \param dWaux_mat derivatives of precomputed angular-dependent part
!> \param dsabb derivative of contracted overlap [abb] of spherical harmonic
!>        Gaussians
! **************************************************************************************************
   SUBROUTINE dev_overlap_shg_abb(la, first_sgfa, nshella, lb, first_sgfb, nshellb, &
                                  lcb, first_sgfcb, nshellcb, cg_coeff, cg_none0_list, &
                                  ncg_none0, rab, swork_cont, Waux_mat, dWaux_mat, dsabb)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: la, first_sgfa
      INTEGER, INTENT(IN)                                :: nshella
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, first_sgfb
      INTEGER, INTENT(IN)                                :: nshellb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lcb, first_sgfcb
      INTEGER, INTENT(IN)                                :: nshellcb
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: cg_coeff
      INTEGER, DIMENSION(:, :, :), INTENT(IN)            :: cg_none0_list
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: ncg_none0
      REAL(KIND=dp), INTENT(IN)                          :: rab(3)
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork_cont
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: dWaux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT)                                   :: dsabb

      CHARACTER(len=*), PARAMETER :: routineN = 'dev_overlap_shg_abb', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ia, il, ilist, ishella, isob1, isob2, isobb, j, jb, jshellb, kb, &
         kshellb, labmin, lai, lbb, lbj, lbk, mai, mbb, mbj, mbk, nla, nlb, sgfa, sgfb, sgfcb
      REAL(KIND=dp)                                      :: dprefac, dtemp(3), prefac, rabx2(3), &
                                                            tsign

      CALL timeset(routineN, handle)

      rabx2(:) = 2.0_dp*rab

      DO ishella = 1, nshella
         lai = la(ishella)
         nla = nsoset_pm(lai-1)+1
         sgfa = first_sgfa(ishella)
         ia = sgfa+lai
         DO jshellb = 1, nshellb
            lbj = lb(jshellb)
            nlb = nsoset_pm(lbj-1)
            sgfb = first_sgfb(jshellb)
            jb = sgfb+lbj
            DO kshellb = 1, nshellcb
               lbk = lcb(kshellb)
               sgfcb = first_sgfcb(kshellb)
               kb = sgfcb+lbk
               DO mai = -lai, lai, 1
                  DO mbj = -lbj, lbj, 1
                     DO mbk = -lbk, lbk, 1
                        isob1 = indso_inv(lbj, mbj)
                        isob2 = indso_inv(lbk, mbk)
                        DO ilist = 1, ncg_none0(isob1, isob2)
                           isobb = cg_none0_list(isob1, isob2, ilist)
                           lbb = indso(1, isobb)
                           mbb = indso(2, isobb)
                           nlb = nsoset_pm(lbb-1)+1
                           tsign = 1.0_dp
                           IF (MODULO(lbb-lai, 2) /= 0) tsign = -1.0_dp
                           labmin = MIN(lai, lbb)
                           il = INT((lbj+lbk-lbb)/2)
                           dtemp = 0.0_dp
                           DO j = 0, labmin
                              prefac = swork_cont(ishella, jshellb, kshellb, il, lai+lbb-j+1)
                              dprefac = swork_cont(ishella, jshellb, kshellb, il, lai+lbb-j+2)
                              DO i = 1, 3
                                 IF (mai >= 0 .AND. mbb >= 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 1) &
                                               +prefac*dWaux_mat(i, nlb+ABS(mbb), nla+ABS(mai), j+1, 1)
                                 ENDIF
                                 ! take the transpose, since auxmat calculated for (lbb,lai)
                                 IF (mai >= 0 .AND. mbb < 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 3) &
                                               +prefac*dWaux_mat(i, nlb+ABS(mbb), nla+ABS(mai), j+1, 3)
                                 ENDIF
                                 IF (mai < 0 .AND. mbb >= 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 2) &
                                               +prefac*dWaux_mat(i, nlb+ABS(mbb), nla+ABS(mai), j+1, 2)
                                 ENDIF
                                 IF (mai < 0 .AND. mbb < 0) THEN
                                    dtemp(i) = dtemp(i)+rabx2(i)*dprefac*Waux_mat(nlb+ABS(mbb), nla+ABS(mai), j+1, 4) &
                                               +prefac*dWaux_mat(i, nlb+ABS(mbb), nla+ABS(mai), j+1, 4)
                                 ENDIF
                              ENDDO
                           ENDDO
                           DO i = 1, 3
                              dsabb(ia+mai, jb+mbj, kb+mbk, i) = dsabb(ia+mai, jb+mbj, kb+mbk, i) &
                                                                 +tsign*cg_coeff(isob1, isob2, isobb)*dtemp(i)
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE dev_overlap_shg_abb

END MODULE ai_shg_overlap

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates integral matrices for LRIGPW method using solid harmonic
!>        Gaussians (SHG)
!>        lri : local resolution of the identity
!> \par History
!>      created [06.2015]
!> \author Dorothea Golze
! **************************************************************************************************
MODULE lri_shg_integrals
   USE ai_shg_overlap,                  ONLY: &
        dev_overlap_shg_ab, dev_overlap_shg_aba, dev_overlap_shg_abb, get_W_matrix, get_dW_matrix, &
        get_real_scaled_solid_harmonic, overlap_shg_ab, overlap_shg_aba, overlap_shg_abb, &
        s_overlap_ab, s_overlap_abx
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE lri_environment_types,           ONLY: lri_environment_type
   USE mathconstants,                   ONLY: fac,&
                                              ifac,&
                                              pi
   USE orbital_pointers,                ONLY: nsoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_shg_integrals'

   PUBLIC :: precalc_angular_shg_part, lri_int_ab_shg, lri_int_aba_shg, &
             lri_int_abb_shg

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief precalculates the angular part of the SHG integrals for the matrices
!>        (fa,fb), (a,b), (a,b,fa) and (b,fb,a)
!> \param oba orbital basis on a
!> \param obb orbital basis on b
!> \param fba aux basis on a
!> \param fbb aux basis on b
!> \param rab distance vector between a and b
!> \param Waux_mat matrix storing angular-dependent part
!> \param dWaux_mat ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE precalc_angular_shg_part(oba, obb, fba, fbb, rab, Waux_mat, dWaux_mat, &
                                       calculate_forces)

      TYPE(gto_basis_set_type), POINTER                  :: oba, obb, fba, fbb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'precalc_angular_shg_part', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, isize, j, k, la_max, laa_max, &
                                                            lb_max, lbb_max, lca_max, lcb_max, &
                                                            li_max, lj_max, lmax, mdim(4), &
                                                            size_int(4, 2), temp
      INTEGER, DIMENSION(:), POINTER                     :: li_max_all
      REAL(KIND=dp)                                      :: rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Rc, Rs

      CALL timeset(routineN, handle)
      NULLIFY (li_max_all)

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)

      !*** 1 Waux_mat of size (li_max,lj_max) for elements
      !                    i        j
      !    [aab]    --> (laa_max, lb_max)
      !    [bba]    --> (lbb_max, la_max) --> use for [abb]
      !    [ab] ri  --> (lca_max, lcb_max)
      !    [ab] orb --> (la_max , lb_max)

      la_max = MAXVAL(oba%lmax)
      lb_max = MAXVAL(obb%lmax)
      lca_max = MAXVAL(fba%lmax)
      lcb_max = MAXVAL(fbb%lmax)

      laa_max = la_max+lca_max
      lbb_max = lb_max+lcb_max
      li_max = MAX(laa_max, lbb_max)
      lj_max = MAX(la_max, lb_max, lcb_max)
      lmax = li_max

      ALLOCATE (li_max_all(0:lj_max))
      ALLOCATE (Rc(0:lmax, -2*lmax:2*lmax), Rs(0:lmax, -2*lmax:2*lmax))
      Rc = 0._dp
      Rs = 0._dp
      mdim(1) = nsoset(li_max)+1
      mdim(2) = nsoset(lj_max)+1
      mdim(3) = li_max+lj_max+1
      mdim(4) = 4
      ALLOCATE (Waux_mat(mdim(1), mdim(2), mdim(3), mdim(4)))
      ALLOCATE (dWaux_mat(3, mdim(1), mdim(2), mdim(3), mdim(4)))
      !Waux_mat = 0._dp !.. takes time

      !*** Waux_mat (li_max,lj_max) contains elements not needed,
      !*** make indixing so that only required ones are computed
      !*** li_max_all(j) --> li_max dependent on j
      size_int(1, :) = (/laa_max, lb_max/)
      size_int(2, :) = (/lbb_max, la_max/)
      size_int(3, :) = (/lca_max, lcb_max/)
      size_int(4, :) = (/la_max, lb_max/)

      li_max_all(:) = 0
      DO isize = 1, 4
         i = size_int(isize, 1)
         j = size_int(isize, 2)
         k = li_max_all(j)
         IF (k < i) li_max_all(j) = i
      ENDDO
      temp = li_max_all(lj_max)
      DO j = lj_max, 0, -1
         IF (li_max_all(j) < temp) THEN
            li_max_all(j) = temp
         ELSE
            temp = li_max_all(j)
         ENDIF
      ENDDO

      !*** -rab, since Eq. in Ref. use Ra-Rb, not Rb-Ra
      CALL get_real_scaled_solid_harmonic(Rc, Rs, lmax, -rab, rab2)
      CALL get_W_matrix(li_max_all, lj_max, lmax, Rc, Rs, Waux_mat)
      IF (calculate_forces) THEN
         CALL get_dW_matrix(li_max_all, lj_max, Waux_mat, dWaux_mat)
      ENDIF

      DEALLOCATE (Rc, Rs, li_max_all)

      CALL timestop(handle)

   END SUBROUTINE precalc_angular_shg_part

! **************************************************************************************************
!> \brief calculate integrals (a,b,fa)
!> \param lri_env ...
!> \param abaint integral (a,b,fa)
!> \param dabdaint ...
!> \param rab distance vector between A and B
!> \param oba orbital basis at center A
!> \param obb orbital basis at center B
!> \param fba auxiliary basis set at center A
!> \param scona_shg contraction matrix for orb bas on A
!> \param sconb_shg contraction matrix for orb bas on B
!> \param sconca_shg contraction matrix for aux bas on A
!> \param Waux_mat angular-depedent part
!> \param dWaux_mat derivative of Waux_mat
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE lri_int_aba_shg(lri_env, abaint, dabdaint, rab, oba, obb, fba, scona_shg, &
                              sconb_shg, sconca_shg, Waux_mat, dWaux_mat, &
                              calculate_forces)

      TYPE(lri_environment_type), POINTER                :: lri_env
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: abaint
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: dabdaint
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      TYPE(gto_basis_set_type), POINTER                  :: oba, obb, fba
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scona_shg, sconb_shg, sconca_shg
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_aba_shg', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, ids, ids_start, iil, il, ipgfa, iset, ishella, jpgfb, jset, jshellb, &
         kpgfa, kset, kshella, la_max_set, laa, laa_max_set, lai, lak, lb_max_set, lbj, &
         lca_max_set, ndev, nds, nds_max, nl, nl_set, npgfa_set, npgfb_set, npgfca_set, nseta, &
         nsetb, nsetca, nshella_set, nshellb_set, nshellca_set
      INTEGER, DIMENSION(:), POINTER                     :: la_max, lb_max, lca_max, npgfa, npgfb, &
                                                            npgfca, nshella, nshellb, nshellca
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb, first_sgfca, la, &
                                                            lb, lca
      REAL(KIND=dp)                                      :: a, aif, bjf, const, consta, constca, &
                                                            dab, prefac, rab2, shg_fac(5)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b, set_radius_ca
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zeta, zetb, zetca
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork, swork_cont

      CALL timeset(routineN, handle)
      NULLIFY (la_max, lb_max, lca_max, npgfa, npgfb, npgfca)
      NULLIFY (first_sgfa, first_sgfb, first_sgfca, set_radius_a, set_radius_b, &
               set_radius_ca, zeta, zetb, zetca)

      ! basis ikind
      first_sgfa => oba%first_sgf
      la_max => oba%lmax
      la => oba%l
      npgfa => oba%npgf
      nshella => oba%nshell
      nseta = oba%nset
      set_radius_a => oba%set_radius
      zeta => oba%zet
      ! basis jkind
      first_sgfb => obb%first_sgf
      lb_max => obb%lmax
      lb => obb%l
      npgfb => obb%npgf
      nshellb => obb%nshell
      nsetb = obb%nset
      set_radius_b => obb%set_radius
      zetb => obb%zet

      ! basis RI A
      first_sgfca => fba%first_sgf
      lca_max => fba%lmax
      lca => fba%l
      npgfca => fba%npgf
      nshellca => fba%nshell
      nsetca = fba%nset
      set_radius_ca => fba%set_radius
      zetca => fba%zet

      dab = SQRT(SUM(rab**2))
      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)

      la_max_set = MAXVAL(la_max)
      lb_max_set = MAXVAL(lb_max)
      lca_max_set = MAXVAL(lca_max)
      npgfa_set = MAXVAL(npgfa)
      npgfb_set = MAXVAL(npgfb)
      npgfca_set = MAXVAL(npgfca)
      nshella_set = MAXVAL(nshella)
      nshellb_set = MAXVAL(nshellb)
      nshellca_set = MAXVAL(nshellca)
      !*** for forces: derivative+1 in auxiliary vector required
      ndev = 0
      IF (calculate_forces) ndev = 1

      laa_max_set = la_max_set+lca_max_set

      ! allocate some work storage....
      nds_max = laa_max_set+lb_max_set+ndev+1
      nl_set = INT((laa_max_set)/2)
      ALLOCATE (swork(npgfa_set, npgfb_set, npgfca_set, 0:nl_set, nds_max))
      ALLOCATE (swork_cont(nshella_set, nshellb_set, nshellca_set, 0:nl_set, nds_max))

      DO iset = 1, nseta

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            DO kset = 1, nsetca

               IF (set_radius_b(jset)+set_radius_ca(kset) < dab) CYCLE

               CALL s_overlap_abx(la_max(iset), npgfa(iset), zeta(:, iset), &
                                  lb_max(jset), npgfb(jset), zetb(:, jset), &
                                  lca_max(kset), npgfca(kset), zetca(:, kset), &
                                  rab, swork, calculate_forces, calc_aba=.TRUE.)

               swork_cont = 0.0_dp
               DO ishella = 1, nshella(iset)
                  lai = la(ishella, iset)
                  consta = SQRT(1.0_dp/REAL(2*lai+1, dp))
                  DO jshellb = 1, nshellb(jset)
                     lbj = lb(jshellb, jset)
                     DO kshella = 1, nshellca(kset)
                        lak = lca(kshella, kset)
                        constca = SQRT(1.0_dp/REAL(2*lak+1, dp))
                        nl = INT((lai+lak)/2)
                        IF (lai == 0 .OR. lak == 0) nl = 0
                        DO il = 0, nl
                           laa = lai+lak-2*il
                           nds = laa+lbj+1
                           ids_start = nds-MIN(laa, lbj)
                           const = consta*constca*2.0_dp*SQRT(pi*REAL(2*laa+1, dp))
                           DO iil = 1, il
                              shg_fac(iil) = fac(laa+iil-1)*ifac(laa)*REAL(laa, dp) &
                                             *fac(il)/fac(il-iil)/fac(iil)
                           ENDDO
                           DO ipgfa = 1, npgfa(iset)
                              DO jpgfb = 1, npgfb(jset)
                                 bjf = 1.0_dp/((2._dp*zetb(jpgfb, jset))**lbj)
                                 DO kpgfa = 1, npgfca(kset)
                                    a = zeta(ipgfa, iset)+zetca(kpgfa, kset)
                                    aif = 1.0_dp/((2._dp*a)**laa)
                                    prefac = const*aif*bjf &
                                             *scona_shg(ipgfa, ishella, iset) &
                                             *sconb_shg(jpgfb, jshellb, jset) &
                                             *sconca_shg(kpgfa, kshella, kset)
                                    DO ids = ids_start, nds+ndev
                                       IF (il == 0) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*swork(ipgfa, jpgfb, kpgfa, il, ids)
                                       ELSEIF (il == 1) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfa, il, ids) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfa, il-1, ids)/a)
                                       ELSEIF (il == 2) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfa, il, ids) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfa, il-1, ids)/a &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfa, il-2, ids)/(a**2))
                                       ELSEIF (il == 3) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfa, il, ids) &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfa, il-1, ids)/a &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfa, il-2, ids)/(a**2) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfa, il-3, ids)/(a**3))
                                       ELSEIF (il == 4) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfa, il, ids) &
                                                      +shg_fac(il-3)*swork(ipgfa, jpgfb, kpgfa, il-1, ids)/a &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfa, il-2, ids)/(a**2) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfa, il-3, ids)/(a**3) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfa, il-4, ids)/(a**4))
                                       ELSEIF (il == 5) THEN
                                          swork_cont(ishella, jshellb, kshella, il, ids) = &
                                             swork_cont(ishella, jshellb, kshella, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfa, il, ids) &
                                                      +shg_fac(il-4)*swork(ipgfa, jpgfb, kpgfa, il-1, ids)/a &
                                                      +shg_fac(il-3)*swork(ipgfa, jpgfb, kpgfa, il-2, ids)/a**2 &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfa, il-3, ids)/(a**3) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfa, il-4, ids)/(a**4) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfa, il-5, ids)/(a**5))
                                       ELSE
                                          CALL cp_abort(cp__l("lri_shg_integrals.F",338), &
                                                        "SHG integrals not implemented when l quantum number"// &
                                                        " of orbital and ri basis larger than 11")
                                       ENDIF
                                    ENDDO
                                 ENDDO
                              ENDDO
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
               CALL overlap_shg_aba(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                    lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                    lca(:, kset), first_sgfca(:, kset), nshellca(kset), &
                                    lri_env%cg_shg%cg_coeff, lri_env%cg_shg%cg_none0_list, &
                                    lri_env%cg_shg%ncg_none0, swork_cont, Waux_mat, abaint)
               !
               IF (calculate_forces) THEN
                  !*** -rab, since Eq. in Ref. use Ra-Rb, not Rb-Ra
                  CALL dev_overlap_shg_aba(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                           lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                           lca(:, kset), first_sgfca(:, kset), nshellca(kset), &
                                           lri_env%cg_shg%cg_coeff, lri_env%cg_shg%cg_none0_list, &
                                           lri_env%cg_shg%ncg_none0, -rab, swork_cont, &
                                           Waux_mat, dWaux_mat, dabdaint)
               ENDIF

            END DO
         END DO
      END DO

      DEALLOCATE (swork_cont)
      DEALLOCATE (swork)

      CALL timestop(handle)

   END SUBROUTINE lri_int_aba_shg

! **************************************************************************************************
!> \brief calculate integrals (a,b,fb)
!> \param lri_env ...
!> \param abbint integral (a,b,fb)
!> \param dabbint derivative of abbint
!> \param rab distance vector between A and B
!> \param oba orbital basis at center A
!> \param obb orbital basis at center B
!> \param fbb auxiliary basis set at center B
!> \param scona_shg contraction matrix for orb bas on A
!> \param sconb_shg contraction matrix for orb bas on B
!> \param sconcb_shg contraction matrix for aux bas on B
!> \param Waux_mat angular-dependent part
!> \param dWaux_mat derivative of Waux_mat
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE lri_int_abb_shg(lri_env, abbint, dabbint, rab, oba, obb, fbb, scona_shg, &
                              sconb_shg, sconcb_shg, Waux_mat, dWaux_mat, &
                              calculate_forces)

      TYPE(lri_environment_type), POINTER                :: lri_env
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: abbint
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: dabbint
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      TYPE(gto_basis_set_type), POINTER                  :: oba, obb, fbb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scona_shg, sconb_shg, sconcb_shg
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_abb_shg', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, ids, ids_start, iil, il, ipgfa, iset, ishella, jpgfb, jset, jshellb, &
         kpgfb, kset, kshellb, la_max_set, lai, lb_max_set, lbb, lbb_max_set, lbj, lbk, &
         lcb_max_set, ndev, nds, nds_max, nl, nl_set, npgfa_set, npgfb_set, npgfcb_set, nseta, &
         nsetb, nsetcb, nshella_set, nshellb_set, nshellcb_set
      INTEGER, DIMENSION(:), POINTER                     :: la_max, lb_max, lcb_max, npgfa, npgfb, &
                                                            npgfcb, nshella, nshellb, nshellcb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb, first_sgfcb, la, &
                                                            lb, lcb
      REAL(KIND=dp)                                      :: aif, b, bjf, const, constb, constcb, &
                                                            dab, prefac, rab2, shg_fac(5)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b, set_radius_cb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zeta, zetb, zetcb
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: swork, swork_cont

      CALL timeset(routineN, handle)
      NULLIFY (la_max, lb_max, lcb_max, npgfa, npgfb, npgfcb)
      NULLIFY (first_sgfa, first_sgfb, first_sgfcb, set_radius_a, set_radius_b, &
               set_radius_cb, zeta, zetb, zetcb)

      ! basis ikind
      first_sgfa => oba%first_sgf
      la_max => oba%lmax
      la => oba%l
      npgfa => oba%npgf
      nshella => oba%nshell
      nseta = oba%nset
      set_radius_a => oba%set_radius
      zeta => oba%zet
      ! basis jkind
      first_sgfb => obb%first_sgf
      lb_max => obb%lmax
      lb => obb%l
      npgfb => obb%npgf
      nshellb => obb%nshell
      nsetb = obb%nset
      set_radius_b => obb%set_radius
      zetb => obb%zet

      ! basis RI on B
      first_sgfcb => fbb%first_sgf
      lcb_max => fbb%lmax
      lcb => fbb%l
      npgfcb => fbb%npgf
      nshellcb => fbb%nshell
      nsetcb = fbb%nset
      set_radius_cb => fbb%set_radius
      zetcb => fbb%zet

      dab = SQRT(SUM(rab**2))
      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)

      la_max_set = MAXVAL(la_max)
      lb_max_set = MAXVAL(lb_max)
      lcb_max_set = MAXVAL(lcb_max)
      npgfa_set = MAXVAL(npgfa)
      npgfb_set = MAXVAL(npgfb)
      npgfcb_set = MAXVAL(npgfcb)
      nshella_set = MAXVAL(nshella)
      nshellb_set = MAXVAL(nshellb)
      nshellcb_set = MAXVAL(nshellcb)
      !*** for forces: derivative+1 in auxiliary vector required
      ndev = 0
      IF (calculate_forces) ndev = 1

      lbb_max_set = lb_max_set+lcb_max_set

      ! allocate some work storage....
      nds_max = la_max_set+lbb_max_set+ndev+1
      nl_set = INT((lbb_max_set)/2)
      ALLOCATE (swork(1:npgfa_set, 1:npgfb_set, 1:npgfcb_set, 0:nl_set, 1:nds_max))
      ALLOCATE (swork_cont(nshella_set, nshellb_set, nshellcb_set, 0:nl_set, nds_max))

      DO iset = 1, nseta

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            DO kset = 1, nsetcb

               IF (set_radius_a(iset)+set_radius_cb(kset) < dab) CYCLE

               CALL s_overlap_abx(la_max(iset), npgfa(iset), zeta(:, iset), &
                                  lb_max(jset), npgfb(jset), zetb(:, jset), &
                                  lcb_max(kset), npgfcb(kset), zetcb(:, kset), &
                                  rab, swork, calculate_forces, calc_aba=.FALSE.)

               swork_cont = 0.0_dp
               DO ishella = 1, nshella(iset)
                  lai = la(ishella, iset)
                  DO jshellb = 1, nshellb(jset)
                     lbj = lb(jshellb, jset)
                     constb = SQRT(1.0_dp/REAL(2*lbj+1, dp))
                     DO kshellb = 1, nshellcb(kset)
                        lbk = lcb(kshellb, kset)
                        constcb = SQRT(1.0_dp/REAL(2*lbk+1, dp))
                        nl = INT((lbj+lbk)/2)
                        IF (lbj == 0 .OR. lbk == 0) nl = 0
                        DO il = 0, nl
                           lbb = lbj+lbk-2*il
                           nds = lai+lbb+1
                           ids_start = nds-MIN(lai, lbb)
                           const = constb*constcb*2.0_dp*SQRT(pi*REAL(2*lbb+1, dp))
                           DO iil = 1, il
                              shg_fac(iil) = fac(lbb+iil-1)*ifac(lbb)*REAL(lbb, dp) &
                                             *fac(il)/fac(il-iil)/fac(iil)
                           ENDDO
                           DO ipgfa = 1, npgfa(iset)
                              aif = 1.0_dp/((2._dp*zeta(ipgfa, iset))**lai)
                              DO jpgfb = 1, npgfb(jset)
                                 DO kpgfb = 1, npgfcb(kset)
                                    b = zetb(jpgfb, jset)+zetcb(kpgfb, kset)
                                    bjf = 1.0_dp/((2._dp*b)**lbb)
                                    prefac = const*aif*bjf &
                                             *scona_shg(ipgfa, ishella, iset) &
                                             *sconb_shg(jpgfb, jshellb, jset) &
                                             *sconcb_shg(kpgfb, kshellb, kset)
                                    DO ids = ids_start, nds+ndev
                                       IF (il == 0) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*swork(ipgfa, jpgfb, kpgfb, il, ids)
                                       ELSEIF (il == 1) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfb, il, ids) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfb, il-1, ids)/b)
                                       ELSEIF (il == 2) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfb, il, ids) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfb, il-1, ids)/b &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfb, il-2, ids)/(b**2))
                                       ELSEIF (il == 3) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfb, il, ids) &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfb, il-1, ids)/b &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfb, il-2, ids)/(b**2) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfb, il-3, ids)/(b**3))
                                       ELSEIF (il == 4) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfb, il, ids) &
                                                      +shg_fac(il-3)*swork(ipgfa, jpgfb, kpgfb, il-1, ids)/b &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfb, il-2, ids)/(b**2) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfb, il-3, ids)/(b**3) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfb, il-4, ids)/(b**4))
                                       ELSEIF (il == 5) THEN
                                          swork_cont(ishella, jshellb, kshellb, il, ids) = &
                                             swork_cont(ishella, jshellb, kshellb, il, ids) &
                                             +prefac*(swork(ipgfa, jpgfb, kpgfb, il, ids) &
                                                      +shg_fac(il-4)*swork(ipgfa, jpgfb, kpgfb, il-1, ids)/b &
                                                      +shg_fac(il-3)*swork(ipgfa, jpgfb, kpgfb, il-2, ids)/b**2 &
                                                      +shg_fac(il-2)*swork(ipgfa, jpgfb, kpgfb, il-3, ids)/(b**3) &
                                                      +shg_fac(il-1)*swork(ipgfa, jpgfb, kpgfb, il-4, ids)/(b**4) &
                                                      +shg_fac(il)*swork(ipgfa, jpgfb, kpgfb, il-5, ids)/(b**5))
                                       ELSE
                                          CALL cp_abort(cp__l("lri_shg_integrals.F",568), &
                                                        "SHG integrals not implemented when l quantum number"// &
                                                        " of orbital and ri basis larger than 11")
                                       ENDIF
                                    ENDDO
                                 ENDDO
                              ENDDO
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
               CALL overlap_shg_abb(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                    lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                    lcb(:, kset), first_sgfcb(:, kset), nshellcb(kset), &
                                    lri_env%cg_shg%cg_coeff, lri_env%cg_shg%cg_none0_list, &
                                    lri_env%cg_shg%ncg_none0, swork_cont, Waux_mat, abbint)
               IF (calculate_forces) THEN
                  !*** -rab, since Eq. in Ref. use Ra-Rb, not Rb-Ra
                  CALL dev_overlap_shg_abb(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                           lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                           lcb(:, kset), first_sgfcb(:, kset), nshellcb(kset), &
                                           lri_env%cg_shg%cg_coeff, lri_env%cg_shg%cg_none0_list, &
                                           lri_env%cg_shg%ncg_none0, -rab, swork_cont, &
                                           Waux_mat, dWaux_mat, dabbint)
               ENDIF

            END DO
         END DO
      END DO

      DEALLOCATE (swork_cont)
      DEALLOCATE (swork)

      CALL timestop(handle)

   END SUBROUTINE lri_int_abb_shg

! **************************************************************************************************
!> \brief calculate overlap integrals (a,b)
!> \param sab integral (a,b)
!> \param dsab derivative of sab
!> \param rab distance vector
!> \param fba basis at center A
!> \param fbb basis at center B
!> \param scona_shg contraction matrix A
!> \param sconb_shg contraxtion matrix B
!> \param Waux_mat angular-dependent part
!> \param dWaux_mat derivative of Waux_mat
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE lri_int_ab_shg(sab, dsab, rab, fba, fbb, scona_shg, sconb_shg, Waux_mat, &
                             dWaux_mat, calculate_forces)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: sab
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dsab
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      TYPE(gto_basis_set_type), POINTER                  :: fba, fbb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scona_shg, sconb_shg
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_ab_shg', routineP = moduleN//':'//routineN

      INTEGER :: handle, ids, ids_start, ipgfa, iset, ishella, jpgfb, jset, jshellb, la_max_set, &
         lai, lb_max_set, lbj, ndev, nds, nds_max, npgfa_set, npgfb_set, nseta, nsetb, &
         nshella_set, nshellb_set
      INTEGER, DIMENSION(:), POINTER                     :: la_max, lb_max, npgfa, npgfb, nshella, &
                                                            nshellb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb, la, lb
      REAL(KIND=dp)                                      :: aif, bjf, dab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: swork, swork_cont

      CALL timeset(routineN, handle)
      NULLIFY (la_max, lb_max, npgfa, npgfb, first_sgfa, first_sgfb, set_radius_a, &
               set_radius_b, zeta, zetb)

      ! basis ikind
      first_sgfa => fba%first_sgf
      la_max => fba%lmax
      la => fba%l
      npgfa => fba%npgf
      nseta = fba%nset
      set_radius_a => fba%set_radius
      zeta => fba%zet
      nshella => fba%nshell
      ! basis jkind
      first_sgfb => fbb%first_sgf
      lb_max => fbb%lmax
      lb => fbb%l
      npgfb => fbb%npgf
      nsetb = fbb%nset
      set_radius_b => fbb%set_radius
      zetb => fbb%zet
      nshellb => fbb%nshell

      dab = SQRT(SUM(rab**2))

      la_max_set = MAXVAL(la_max)
      lb_max_set = MAXVAL(lb_max)

      ! allocate some work matrices
      npgfa_set = MAXVAL(npgfa)
      npgfb_set = MAXVAL(npgfb)
      nshella_set = MAXVAL(nshella)
      nshellb_set = MAXVAL(nshellb)
      ndev = 0
      IF (calculate_forces) ndev = 1
      nds_max = la_max_set+lb_max_set+ndev+1
      ALLOCATE (swork(npgfa_set, npgfb_set, nds_max))
      ALLOCATE (swork_cont(nshella_set, nshellb_set, nds_max))

      DO iset = 1, nseta

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            CALL s_overlap_ab(la_max(iset), npgfa(iset), zeta(:, iset), &
                              lb_max(jset), npgfb(jset), zetb(:, jset), &
                              rab, swork, calculate_forces)

            swork_cont = 0.0_dp
            DO ishella = 1, nshella(iset)
               lai = la(ishella, iset)
               DO jshellb = 1, nshellb(jset)
                  lbj = lb(jshellb, jset)
                  nds = lai+lbj+1
                  ids_start = nds-MIN(lai, lbj)
                  DO ipgfa = 1, npgfa(iset)
                     aif = 1.0_dp/((2._dp*zeta(ipgfa, iset))**lai)
                     DO jpgfb = 1, npgfb(jset)
                        bjf = 1.0_dp/((2._dp*zetb(jpgfb, jset))**lbj)
                        DO ids = ids_start, nds+ndev
                           swork_cont(ishella, jshellb, ids) = swork_cont(ishella, jshellb, ids) &
                                                               +scona_shg(ipgfa, ishella, iset) &
                                                               *sconb_shg(jpgfb, jshellb, jset) &
                                                               *aif*bjf*swork(ipgfa, jpgfb, ids)
                        ENDDO
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
            CALL overlap_shg_ab(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                swork_cont, Waux_mat, sab)
            IF (calculate_forces) THEN
               !*** -rab, since Eq. in Ref. use Ra-Rb, not Rb-Ra
               CALL dev_overlap_shg_ab(la(:, iset), first_sgfa(:, iset), nshella(iset), &
                                       lb(:, jset), first_sgfb(:, jset), nshellb(jset), &
                                       -rab, swork_cont, Waux_mat, dWaux_mat, dsab)
            ENDIF
         END DO
      END DO

      DEALLOCATE (swork, swork_cont)

      CALL timestop(handle)

   END SUBROUTINE lri_int_ab_shg

END MODULE lri_shg_integrals
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE ai_os_rr

   USE gamma,                           ONLY: fgamma=>fgamma_0
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: coset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_os_rr'
   PRIVATE

   ! *** Public subroutines ***

   PUBLIC :: os_rr_ovlp, os_rr_coul

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the basic Obara-Saika reccurence relation
!> \param rap ...
!> \param la_max ...
!> \param rbp ...
!> \param lb_max ...
!> \param zet ...
!> \param ldrr ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE os_rr_ovlp(rap, la_max, rbp, lb_max, zet, ldrr, rr)
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rap
      INTEGER, INTENT(IN)                                :: la_max
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rbp
      INTEGER, INTENT(IN)                                :: lb_max
      REAL(dp), INTENT(IN)                               :: zet
      INTEGER, INTENT(IN)                                :: ldrr
      REAL(dp), DIMENSION(0:ldrr-1, 0:ldrr-1, 3)         :: rr

      INTEGER                                            :: la, lam1, lap1, lb, lbm1, lbp1
      REAL(dp)                                           :: g

      g = 0.5_dp/zet
      rr(0, 0, 1) = 1.0_dp
      rr(0, 0, 2) = 1.0_dp
      rr(0, 0, 3) = 1.0_dp
      !
      ! recursion along la for lb=0
      !
      IF (la_max .GT. 0) THEN
         rr(1, 0, 1) = rap(1)
         rr(1, 0, 2) = rap(2)
         rr(1, 0, 3) = rap(3)
         !
         DO la = 1, la_max-1
            lap1 = la+1
            lam1 = la-1
            rr(lap1, 0, 1) = REAL(la, dp)*g*rr(lam1, 0, 1)+rap(1)*rr(la, 0, 1)
            rr(lap1, 0, 2) = REAL(la, dp)*g*rr(lam1, 0, 2)+rap(2)*rr(la, 0, 2)
            rr(lap1, 0, 3) = REAL(la, dp)*g*rr(lam1, 0, 3)+rap(3)*rr(la, 0, 3)
         ENDDO
      ENDIF
      !
      ! recursion along lb for all la
      !
      IF (lb_max .GT. 0) THEN
         rr(0, 1, 1) = rbp(1)
         rr(0, 1, 2) = rbp(2)
         rr(0, 1, 3) = rbp(3)
         !
         DO la = 1, la_max
            lam1 = la-1
            rr(la, 1, 1) = REAL(la, dp)*g*rr(lam1, 0, 1)+rbp(1)*rr(la, 0, 1)
            rr(la, 1, 2) = REAL(la, dp)*g*rr(lam1, 0, 2)+rbp(2)*rr(la, 0, 2)
            rr(la, 1, 3) = REAL(la, dp)*g*rr(lam1, 0, 3)+rbp(3)*rr(la, 0, 3)
         ENDDO
         !
         DO lb = 1, lb_max-1
            lbp1 = lb+1
            lbm1 = lb-1
            rr(0, lbp1, 1) = REAL(lb, dp)*g*rr(0, lbm1, 1)+rbp(1)*rr(0, lb, 1)
            rr(0, lbp1, 2) = REAL(lb, dp)*g*rr(0, lbm1, 2)+rbp(2)*rr(0, lb, 2)
            rr(0, lbp1, 3) = REAL(lb, dp)*g*rr(0, lbm1, 3)+rbp(3)*rr(0, lb, 3)
            DO la = 1, la_max
               lam1 = la-1
               rr(la, lbp1, 1) = g*(REAL(la, dp)*rr(lam1, lb, 1)+REAL(lb, dp)*rr(la, lbm1, 1))+rbp(1)*rr(la, lb, 1)
               rr(la, lbp1, 2) = g*(REAL(la, dp)*rr(lam1, lb, 2)+REAL(lb, dp)*rr(la, lbm1, 2))+rbp(2)*rr(la, lb, 2)
               rr(la, lbp1, 3) = g*(REAL(la, dp)*rr(lam1, lb, 3)+REAL(lb, dp)*rr(la, lbm1, 3))+rbp(3)*rr(la, lb, 3)
            ENDDO
         ENDDO
      ENDIF
      !
   END SUBROUTINE os_rr_ovlp

! **************************************************************************************************
!> \brief   Calculation of the Obara-Saika reccurence relation for 1/r_C
!> \param rap ...
!> \param la_max ...
!> \param rbp ...
!> \param lb_max ...
!> \param rcp ...
!> \param zet ...
!> \param ldrr1 ...
!> \param ldrr2 ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE os_rr_coul(rap, la_max, rbp, lb_max, rcp, zet, ldrr1, ldrr2, rr)
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rap
      INTEGER, INTENT(IN)                                :: la_max
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rbp
      INTEGER, INTENT(IN)                                :: lb_max
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rcp
      REAL(dp), INTENT(IN)                               :: zet
      INTEGER, INTENT(IN)                                :: ldrr1, ldrr2
      REAL(dp), DIMENSION(0:ldrr1-1, ldrr2, *), &
         INTENT(INOUT)                                   :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'os_rr_coul', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, coa1x, &
                                                            coa1y, coa1z, coa2x, coa2y, coa2z, &
                                                            cob, cob1x, cob1y, cob1z, cob2x, &
                                                            cob2y, cob2z, la, lb, m, mmax
      REAL(dp)                                           :: g, rcp2, t

      mmax = la_max+lb_max
      g = 0.5_dp/zet
      !
      ! rr(0:mmax) should be initialized before
      !
      rcp2 = rcp(1)**2+rcp(2)**2+rcp(3)**2
      t = zet*rcp2
      CALL fgamma(mmax, t, rr(0:mmax, 1, 1))
      !
      ! recursion in la with lb=0
      !
      DO la = 1, la_max
         DO ax = 0, la
         DO ay = 0, la-ax
            az = la-ax-ay
            coa = coset(ax, ay, az)
            coa1x = coset(MAX(ax-1, 0), ay, az)
            coa1y = coset(ax, MAX(ay-1, 0), az)
            coa1z = coset(ax, ay, MAX(az-1, 0))
            coa2x = coset(MAX(ax-2, 0), ay, az)
            coa2y = coset(ax, MAX(ay-2, 0), az)
            coa2z = coset(ax, ay, MAX(az-2, 0))
            IF (az .GT. 0) THEN
               DO m = 0, mmax-la
                  rr(m, coa, 1) = rap(3)*rr(m, coa1z, 1)-rcp(3)*rr(m+1, coa1z, 1)
               ENDDO
               IF (az .GT. 1) THEN
                  DO m = 0, mmax-la
                     rr(m, coa, 1) = rr(m, coa, 1)+g*REAL(az-1, dp)*(rr(m, coa2z, 1)-rr(m+1, coa2z, 1))
                  ENDDO
               ENDIF
            ELSEIF (ay .GT. 0) THEN
               DO m = 0, mmax-la
                  rr(m, coa, 1) = rap(2)*rr(m, coa1y, 1)-rcp(2)*rr(m+1, coa1y, 1)
               ENDDO
               IF (ay .GT. 1) THEN
                  DO m = 0, mmax-la
                     rr(m, coa, 1) = rr(m, coa, 1)+g*REAL(ay-1, dp)*(rr(m, coa2y, 1)-rr(m+1, coa2y, 1))
                  ENDDO
               ENDIF
            ELSEIF (ax .GT. 0) THEN
               DO m = 0, mmax-la
                  rr(m, coa, 1) = rap(1)*rr(m, coa1x, 1)-rcp(1)*rr(m+1, coa1x, 1)
               ENDDO
               IF (ax .GT. 1) THEN
                  DO m = 0, mmax-la
                     rr(m, coa, 1) = rr(m, coa, 1)+g*REAL(ax-1, dp)*(rr(m, coa2x, 1)-rr(m+1, coa2x, 1))
                  ENDDO
               ENDIF
            ELSE
               CALL cp__b("aobasis/ai_os_rr.F",184,"")
            ENDIF
         ENDDO
         ENDDO
      ENDDO
      !
      ! recursion in lb with all possible la
      !
      DO la = 0, la_max
         DO ax = 0, la
         DO ay = 0, la-ax
            az = la-ax-ay
            coa = coset(ax, ay, az)
            coa1x = coset(MAX(ax-1, 0), ay, az)
            coa1y = coset(ax, MAX(ay-1, 0), az)
            coa1z = coset(ax, ay, MAX(az-1, 0))
            coa2x = coset(MAX(ax-2, 0), ay, az)
            coa2y = coset(ax, MAX(ay-2, 0), az)
            coa2z = coset(ax, ay, MAX(az-2, 0))
            DO lb = 1, lb_max
               DO bx = 0, lb
               DO by = 0, lb-bx
                  bz = lb-bx-by
                  cob = coset(bx, by, bz)
                  cob1x = coset(MAX(bx-1, 0), by, bz)
                  cob1y = coset(bx, MAX(by-1, 0), bz)
                  cob1z = coset(bx, by, MAX(bz-1, 0))
                  cob2x = coset(MAX(bx-2, 0), by, bz)
                  cob2y = coset(bx, MAX(by-2, 0), bz)
                  cob2z = coset(bx, by, MAX(bz-2, 0))
                  IF (bz .GT. 0) THEN
                     DO m = 0, mmax-la-lb
                        rr(m, coa, cob) = rbp(3)*rr(m, coa, cob1z)-rcp(3)*rr(m+1, coa, cob1z)
                     ENDDO
                     IF (bz .GT. 1) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(bz-1, dp)*(rr(m, coa, cob2z)-rr(m+1, coa, cob2z))
                        ENDDO
                     ENDIF
                     IF (az .GT. 0) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(az, dp)*(rr(m, coa1z, cob1z)-rr(m+1, coa1z, cob1z))
                        ENDDO
                     ENDIF
                  ELSEIF (by .GT. 0) THEN
                     DO m = 0, mmax-la-lb
                        rr(m, coa, cob) = rbp(2)*rr(m, coa, cob1y)-rcp(2)*rr(m+1, coa, cob1y)
                     ENDDO
                     IF (by .GT. 1) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(by-1, dp)*(rr(m, coa, cob2y)-rr(m+1, coa, cob2y))
                        ENDDO
                     ENDIF
                     IF (ay .GT. 0) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(ay, dp)*(rr(m, coa1y, cob1y)-rr(m+1, coa1y, cob1y))
                        ENDDO
                     ENDIF
                  ELSEIF (bx .GT. 0) THEN
                     DO m = 0, mmax-la-lb
                        rr(m, coa, cob) = rbp(1)*rr(m, coa, cob1x)-rcp(1)*rr(m+1, coa, cob1x)
                     ENDDO
                     IF (bx .GT. 1) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(bx-1, dp)*(rr(m, coa, cob2x)-rr(m+1, coa, cob2x))
                        ENDDO
                     ENDIF
                     IF (ax .GT. 0) THEN
                        DO m = 0, mmax-la-lb
                           rr(m, coa, cob) = rr(m, coa, cob)+g*REAL(ax, dp)*(rr(m, coa1x, cob1x)-rr(m+1, coa1x, cob1x))
                        ENDDO
                     ENDIF
                  ELSE
                     CALL cp__b("aobasis/ai_os_rr.F",257,"")
                  ENDIF
               ENDDO
               ENDDO
            ENDDO
         ENDDO
         ENDDO
      ENDDO
      !
   END SUBROUTINE os_rr_coul

END MODULE ai_os_rr
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the overlap integrals over Cartesian Gaussian-type
!>      functions.
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      - Derivatives added (02.05.2002,MK)
!>      - New OS routine with simpler logic (11.07.2014, JGH)
!> \author Matthias Krack (08.10.1999)
! **************************************************************************************************
MODULE ai_overlap
   USE ai_os_rr,                        ONLY: os_rr_ovlp
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE orbital_pointers,                ONLY: coset,&
                                              nco,&
                                              ncoset,&
                                              nso
   USE orbital_transformation_matrices, ONLY: orbtramat
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap'

! *** Public subroutines ***
   PUBLIC :: overlap, overlap_ab, overlap_aab, overlap_ab_s, overlap_ab_sp

CONTAINS

! **************************************************************************************************
!> \brief   Purpose: Calculation of the two-center overlap integrals [a|b] over
!>          Cartesian Gaussian-type functions.
!> \param la_max_set Max L on center A
!> \param la_min_set Min L on center A
!> \param npgfa      Number of primitives on center A
!> \param rpgfa      Range of functions on A, used for screening
!> \param zeta       Exponents on center A
!> \param lb_max_set Max L on center B
!> \param lb_min_set Min L on center B
!> \param npgfb      Number of primitives on center B
!> \param rpgfb      Range of functions on B, used for screening
!> \param zetb       Exponents on center B
!> \param rab        Distance vector A-B
!> \param dab        Distance A-B
!> \param sab        Final Integrals, basic and derivatives
!> \param da_max_set Some additional derivative information
!> \param return_derivatives   Return integral derivatives
!> \param s          Work space
!> \param lds        Leading dimension of s
!> \param sdab       Return additional derivative integrals
!> \param pab        Density matrix block, used to calculate forces
!> \param force_a    Force vector [da/dR|b]
!> \date    19.09.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE overlap(la_max_set, la_min_set, npgfa, rpgfa, zeta, &
                      lb_max_set, lb_min_set, npgfb, rpgfb, zetb, &
                      rab, dab, sab, da_max_set, return_derivatives, s, lds, &
                      sdab, pab, force_a)
      INTEGER, INTENT(IN)                                :: la_max_set, la_min_set, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max_set, lb_min_set, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: dab
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: sab
      INTEGER, INTENT(IN)                                :: da_max_set
      LOGICAL, INTENT(IN)                                :: return_derivatives
      INTEGER, INTENT(IN)                                :: lds
      REAL(KIND=dp), DIMENSION(lds, lds, *), &
         INTENT(INOUT)                                   :: s
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: sdab
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: pab
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: force_a

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, cda, cdax, cday, cdaz, coa, coamx, coamy, coamz, coapx, &
         coapy, coapz, cob, cobm2x, cobm2y, cobm2z, cobmx, cobmy, cobmz, da, da_max, dax, day, &
         daz, i, ipgf, j, jk, jpgf, jstart, k, la, la_max, la_min, la_start, lb, lb_max, lb_min, &
         lb_start, na, nb, nda
      LOGICAL                                            :: calculate_force_a
      REAL(KIND=dp)                                      :: f0, f1, f2, f3, f4, fax, fay, faz, ftz, &
                                                            zetp
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

      IF (PRESENT(pab) .AND. PRESENT(force_a)) THEN
         calculate_force_a = .TRUE.
         force_a(:) = 0.0_dp
      ELSE
         calculate_force_a = .FALSE.
      END IF

      IF (PRESENT(sdab) .OR. calculate_force_a) THEN
         IF (da_max_set == 0) THEN
            da_max = 1
            la_max = la_max_set+1
            la_min = MAX(0, la_min_set-1)
         ELSE
            da_max = da_max_set
            la_max = la_max_set+da_max_set+1
            la_min = MAX(0, la_min_set-da_max_set-1)
         END IF
      ELSE
         da_max = da_max_set
         la_max = la_max_set+da_max_set
         la_min = MAX(0, la_min_set-da_max_set)
      END IF

      lb_max = lb_max_set
      lb_min = lb_min_set

!   *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nda = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

!       *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max_set)
                  DO i = na+1, na+ncoset(la_max_set)
                     sab(i, j) = 0.0_dp
                  END DO
               END DO
               IF (return_derivatives) THEN
                  DO k = 2, ncoset(da_max_set)
                     jstart = (k-1)*SIZE(sab, 1)
                     DO j = jstart+nb+1, jstart+nb+ncoset(lb_max_set)
                        DO i = na+1, na+ncoset(la_max_set)
                           sab(i, j) = 0.0_dp
                        END DO
                     END DO
                  END DO
               END IF
               nb = nb+ncoset(lb_max_set)
               CYCLE
            END IF

!       *** Calculate some prefactors ***

            zetp = 1.0_dp/(zeta(ipgf)+zetb(jpgf))

            f0 = SQRT((pi*zetp)**3)
            f1 = zetb(jpgf)*zetp
            f2 = 0.5_dp*zetp

!       *** Calculate the basic two-center overlap integral [s|s] ***

            s(1, 1, 1) = f0*EXP(-zeta(ipgf)*f1*dab*dab)

!       *** Recurrence steps: [s|s] -> [a|b] ***

            IF (la_max > 0) THEN

!         *** Vertical recurrence steps: [s|s] -> [a|s] ***

               rap(:) = f1*rab(:)

!         *** [p|s] = (Pi - Ai)*[s|s]  (i = x,y,z) ***

               s(2, 1, 1) = rap(1)*s(1, 1, 1) ! [px|s]
               s(3, 1, 1) = rap(2)*s(1, 1, 1) ! [py|s]
               s(4, 1, 1) = rap(3)*s(1, 1, 1) ! [pz|s]

               IF (la_max > 1) THEN

!           *** [d|s] ***

                  f3 = f2*s(1, 1, 1)

                  s(5, 1, 1) = rap(1)*s(2, 1, 1)+f3 ! [dx2|s]
                  s(6, 1, 1) = rap(1)*s(3, 1, 1) ! [dxy|s]
                  s(7, 1, 1) = rap(1)*s(4, 1, 1) ! [dxz|s]
                  s(8, 1, 1) = rap(2)*s(3, 1, 1)+f3 ! [dy2|s]
                  s(9, 1, 1) = rap(2)*s(4, 1, 1) ! [dyz|s]
                  s(10, 1, 1) = rap(3)*s(4, 1, 1)+f3 ! [dz2|s]

                  IF (la_max > 2) THEN

!             *** [f|s] ***

                     f3 = 2.0_dp*f2

                     s(11, 1, 1) = rap(1)*s(5, 1, 1)+f3*s(2, 1, 1) ! [fx3 |s]
                     s(12, 1, 1) = rap(1)*s(6, 1, 1)+f2*s(3, 1, 1) ! [fx2y|s]
                     s(13, 1, 1) = rap(1)*s(7, 1, 1)+f2*s(4, 1, 1) ! [fx2z|s]
                     s(14, 1, 1) = rap(1)*s(8, 1, 1) ! [fxy2|s]
                     s(15, 1, 1) = rap(1)*s(9, 1, 1) ! [fxyz|s]
                     s(16, 1, 1) = rap(1)*s(10, 1, 1) ! [fxz2|s]
                     s(17, 1, 1) = rap(2)*s(8, 1, 1)+f3*s(3, 1, 1) ! [fy3 |s]
                     s(18, 1, 1) = rap(2)*s(9, 1, 1)+f2*s(4, 1, 1) ! [fy2z|s]
                     s(19, 1, 1) = rap(2)*s(10, 1, 1) ! [fyz2|s]
                     s(20, 1, 1) = rap(3)*s(10, 1, 1)+f3*s(4, 1, 1) ! [fz3 |s]

                     IF (la_max > 3) THEN

!               *** [g|s] ***

                        f4 = 3.0_dp*f2

                        s(21, 1, 1) = rap(1)*s(11, 1, 1)+f4*s(5, 1, 1) ! [gx4  |s]
                        s(22, 1, 1) = rap(1)*s(12, 1, 1)+f3*s(6, 1, 1) ! [gx3y |s]
                        s(23, 1, 1) = rap(1)*s(13, 1, 1)+f3*s(7, 1, 1) ! [gx3z |s]
                        s(24, 1, 1) = rap(1)*s(14, 1, 1)+f2*s(8, 1, 1) ! [gx2y2|s]
                        s(25, 1, 1) = rap(1)*s(15, 1, 1)+f2*s(9, 1, 1) ! [gx2yz|s]
                        s(26, 1, 1) = rap(1)*s(16, 1, 1)+f2*s(10, 1, 1) ! [gx2z2|s]
                        s(27, 1, 1) = rap(1)*s(17, 1, 1) ! [gxy3 |s]
                        s(28, 1, 1) = rap(1)*s(18, 1, 1) ! [gxy2z|s]
                        s(29, 1, 1) = rap(1)*s(19, 1, 1) ! [gxyz2|s]
                        s(30, 1, 1) = rap(1)*s(20, 1, 1) ! [gxz3 |s]
                        s(31, 1, 1) = rap(2)*s(17, 1, 1)+f4*s(8, 1, 1) ! [gy4  |s]
                        s(32, 1, 1) = rap(2)*s(18, 1, 1)+f3*s(9, 1, 1) ! [gy3z |s]
                        s(33, 1, 1) = rap(2)*s(19, 1, 1)+f2*s(10, 1, 1) ! [gy2z2|s]
                        s(34, 1, 1) = rap(2)*s(20, 1, 1) ! [gyz3 |s]
                        s(35, 1, 1) = rap(3)*s(20, 1, 1)+f4*s(10, 1, 1) ! [gz4  |s]

!               *** [a|s] = (Pi - Ai)*[a-1i|s] + f2*Ni(a-1i)*[a-2i|s] ***

                        DO la = 5, la_max

!                 *** Increase the angular momentum component z of a ***

                           s(coset(0, 0, la), 1, 1) = &
                              rap(3)*s(coset(0, 0, la-1), 1, 1)+ &
                              f2*REAL(la-1, dp)*s(coset(0, 0, la-2), 1, 1)

!                 *** Increase the angular momentum component y of a ***

                           az = la-1
                           s(coset(0, 1, az), 1, 1) = rap(2)*s(coset(0, 0, az), 1, 1)
                           DO ay = 2, la
                              az = la-ay
                              s(coset(0, ay, az), 1, 1) = &
                                 rap(2)*s(coset(0, ay-1, az), 1, 1)+ &
                                 f2*REAL(ay-1, dp)*s(coset(0, ay-2, az), 1, 1)
                           END DO

!                 *** Increase the angular momentum component x of a ***

                           DO ay = 0, la-1
                              az = la-1-ay
                              s(coset(1, ay, az), 1, 1) = rap(1)*s(coset(0, ay, az), 1, 1)
                           END DO
                           DO ax = 2, la
                              f3 = f2*REAL(ax-1, dp)
                              DO ay = 0, la-ax
                                 az = la-ax-ay
                                 s(coset(ax, ay, az), 1, 1) = &
                                    rap(1)*s(coset(ax-1, ay, az), 1, 1)+ &
                                    f3*s(coset(ax-2, ay, az), 1, 1)
                              END DO
                           END DO

                        END DO

                     END IF

                  END IF

               END IF

!         *** Recurrence steps: [a|s] -> [a|b] ***

               IF (lb_max > 0) THEN

                  DO j = 2, ncoset(lb_max)
                     DO i = 1, ncoset(la_min)
                        s(i, j, 1) = 0.0_dp
                     END DO
                  END DO

!           *** Horizontal recurrence steps ***

                  rbp(:) = rap(:)-rab(:)

!           *** [a|p] = [a+1i|s] - (Bi - Ai)*[a|s] ***

                  IF (lb_max == 1) THEN
                     la_start = la_min
                  ELSE
                     la_start = MAX(0, la_min-1)
                  END IF

                  DO la = la_start, la_max-1
                     DO ax = 0, la
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           coa = coset(ax, ay, az)
                           coapx = coset(ax+1, ay, az)
                           coapy = coset(ax, ay+1, az)
                           coapz = coset(ax, ay, az+1)
                           s(coa, 2, 1) = s(coapx, 1, 1)-rab(1)*s(coa, 1, 1)
                           s(coa, 3, 1) = s(coapy, 1, 1)-rab(2)*s(coa, 1, 1)
                           s(coa, 4, 1) = s(coapz, 1, 1)-rab(3)*s(coa, 1, 1)
                        END DO
                     END DO
                  END DO

!           *** Vertical recurrence step ***

!           *** [a|p] = (Pi - Bi)*[a|s] + f2*Ni(a)*[a-1i|s] ***

                  DO ax = 0, la_max
                     fax = f2*REAL(ax, dp)
                     DO ay = 0, la_max-ax
                        fay = f2*REAL(ay, dp)
                        az = la_max-ax-ay
                        faz = f2*REAL(az, dp)
                        coa = coset(ax, ay, az)
                        coamx = coset(ax-1, ay, az)
                        coamy = coset(ax, ay-1, az)
                        coamz = coset(ax, ay, az-1)
                        s(coa, 2, 1) = rbp(1)*s(coa, 1, 1)+fax*s(coamx, 1, 1)
                        s(coa, 3, 1) = rbp(2)*s(coa, 1, 1)+fay*s(coamy, 1, 1)
                        s(coa, 4, 1) = rbp(3)*s(coa, 1, 1)+faz*s(coamz, 1, 1)
                     END DO
                  END DO

!           *** Recurrence steps: [a|p] -> [a|b] ***

                  DO lb = 2, lb_max

!             *** Horizontal recurrence steps ***

!             *** [a|b] = [a+1i|b-1i] - (Bi - Ai)*[a|b-1i] ***

                     IF (lb == lb_max) THEN
                        la_start = la_min
                     ELSE
                        la_start = MAX(0, la_min-1)
                     END IF

                     DO la = la_start, la_max-1
                        DO ax = 0, la
                           DO ay = 0, la-ax
                              az = la-ax-ay
                              coa = coset(ax, ay, az)
                              coapx = coset(ax+1, ay, az)
                              coapy = coset(ax, ay+1, az)
                              coapz = coset(ax, ay, az+1)

!                   *** Shift of angular momentum component z from a to b ***

                              cob = coset(0, 0, lb)
                              cobmz = coset(0, 0, lb-1)
                              s(coa, cob, 1) = s(coapz, cobmz, 1)-rab(3)*s(coa, cobmz, 1)

!                   *** Shift of angular momentum component y from a to b ***

                              DO by = 1, lb
                                 bz = lb-by
                                 cob = coset(0, by, bz)
                                 cobmy = coset(0, by-1, bz)
                                 s(coa, cob, 1) = s(coapy, cobmy, 1)-rab(2)*s(coa, cobmy, 1)
                              END DO

!                   *** Shift of angular momentum component x from a to b ***

                              DO bx = 1, lb
                                 DO by = 0, lb-bx
                                    bz = lb-bx-by
                                    cob = coset(bx, by, bz)
                                    cobmx = coset(bx-1, by, bz)
                                    s(coa, cob, 1) = s(coapx, cobmx, 1)-rab(1)*s(coa, cobmx, 1)
                                 END DO
                              END DO

                           END DO
                        END DO
                     END DO

!             *** Vertical recurrence step ***

!             *** [a|b] = (Pi - Bi)*[a|b-1i] + f2*Ni(a)*[a-1i|b-1i] + ***
!             ***         f2*Ni(b-1i)*[a|b-2i]                        ***

                     DO ax = 0, la_max
                        fax = f2*REAL(ax, dp)
                        DO ay = 0, la_max-ax
                           fay = f2*REAL(ay, dp)
                           az = la_max-ax-ay
                           faz = f2*REAL(az, dp)
                           coa = coset(ax, ay, az)
                           coamx = coset(ax-1, ay, az)
                           coamy = coset(ax, ay-1, az)
                           coamz = coset(ax, ay, az-1)

!                 *** Increase the angular momentum component z of b ***

                           f3 = f2*REAL(lb-1, dp)
                           cob = coset(0, 0, lb)
                           cobmz = coset(0, 0, lb-1)
                           cobm2z = coset(0, 0, lb-2)
                           s(coa, cob, 1) = rbp(3)*s(coa, cobmz, 1)+ &
                                            faz*s(coamz, cobmz, 1)+ &
                                            f3*s(coa, cobm2z, 1)

!                 *** Increase the angular momentum component y of b ***

                           bz = lb-1
                           cob = coset(0, 1, bz)
                           cobmy = coset(0, 0, bz)
                           s(coa, cob, 1) = rbp(2)*s(coa, cobmy, 1)+ &
                                            fay*s(coamy, cobmy, 1)
                           DO by = 2, lb
                              bz = lb-by
                              f3 = f2*REAL(by-1, dp)
                              cob = coset(0, by, bz)
                              cobmy = coset(0, by-1, bz)
                              cobm2y = coset(0, by-2, bz)
                              s(coa, cob, 1) = rbp(2)*s(coa, cobmy, 1)+ &
                                               fay*s(coamy, cobmy, 1)+ &
                                               f3*s(coa, cobm2y, 1)
                           END DO

!                 *** Increase the angular momentum component x of b ***

                           DO by = 0, lb-1
                              bz = lb-1-by
                              cob = coset(1, by, bz)
                              cobmx = coset(0, by, bz)
                              s(coa, cob, 1) = rbp(1)*s(coa, cobmx, 1)+ &
                                               fax*s(coamx, cobmx, 1)
                           END DO
                           DO bx = 2, lb
                              f3 = f2*REAL(bx-1, dp)
                              DO by = 0, lb-bx
                                 bz = lb-bx-by
                                 cob = coset(bx, by, bz)
                                 cobmx = coset(bx-1, by, bz)
                                 cobm2x = coset(bx-2, by, bz)
                                 s(coa, cob, 1) = rbp(1)*s(coa, cobmx, 1)+ &
                                                  fax*s(coamx, cobmx, 1)+ &
                                                  f3*s(coa, cobm2x, 1)
                              END DO
                           END DO

                        END DO
                     END DO

                  END DO

               END IF

            ELSE

               IF (lb_max > 0) THEN

!           *** Vertical recurrence steps: [s|s] -> [s|b] ***

                  rbp(:) = (f1-1.0_dp)*rab(:)

!           *** [s|p] = (Pi - Bi)*[s|s] ***

                  s(1, 2, 1) = rbp(1)*s(1, 1, 1) ! [s|px]
                  s(1, 3, 1) = rbp(2)*s(1, 1, 1) ! [s|py]
                  s(1, 4, 1) = rbp(3)*s(1, 1, 1) ! [s|pz]

                  IF (lb_max > 1) THEN

!             *** [s|d] ***

                     f3 = f2*s(1, 1, 1)

                     s(1, 5, 1) = rbp(1)*s(1, 2, 1)+f3 ! [s|dx2]
                     s(1, 6, 1) = rbp(1)*s(1, 3, 1) ! [s|dxy]
                     s(1, 7, 1) = rbp(1)*s(1, 4, 1) ! [s|dxz]
                     s(1, 8, 1) = rbp(2)*s(1, 3, 1)+f3 ! [s|dy2]
                     s(1, 9, 1) = rbp(2)*s(1, 4, 1) ! [s|dyz]
                     s(1, 10, 1) = rbp(3)*s(1, 4, 1)+f3 ! [s|dz2]

!             *** [s|b] = (Pi - Bi)*[s|b-1i] + f2*Ni(b-1i)*[s|b-2i] ***

                     DO lb = 3, lb_max

!               *** Increase the angular momentum component z of b ***

                        s(1, coset(0, 0, lb), 1) = &
                           rbp(3)*s(1, coset(0, 0, lb-1), 1)+ &
                           f2*REAL(lb-1, dp)*s(1, coset(0, 0, lb-2), 1)

!               *** Increase the angular momentum component y of b ***

                        bz = lb-1
                        s(1, coset(0, 1, bz), 1) = rbp(2)*s(1, coset(0, 0, bz), 1)
                        DO by = 2, lb
                           bz = lb-by
                           s(1, coset(0, by, bz), 1) = &
                              rbp(2)*s(1, coset(0, by-1, bz), 1)+ &
                              f2*REAL(by-1, dp)*s(1, coset(0, by-2, bz), 1)
                        END DO

!               *** Increase the angular momentum component x of b ***

                        DO by = 0, lb-1
                           bz = lb-1-by
                           s(1, coset(1, by, bz), 1) = rbp(1)*s(1, coset(0, by, bz), 1)
                        END DO
                        DO bx = 2, lb
                           f3 = f2*REAL(bx-1, dp)
                           DO by = 0, lb-bx
                              bz = lb-bx-by
                              s(1, coset(bx, by, bz), 1) = &
                                 rbp(1)*s(1, coset(bx-1, by, bz), 1)+ &
                                 f3*s(1, coset(bx-2, by, bz), 1)
                           END DO
                        END DO

                     END DO

                  END IF

               END IF

            END IF

!       *** Store the primitive overlap integrals ***

            DO j = 1, ncoset(lb_max_set)
               DO i = 1, ncoset(la_max_set)
                  sab(na+i, nb+j) = s(i, j, 1)
               END DO
            END DO

!       *** Calculate the requested derivatives with respect  ***
!       *** to the nuclear coordinates of the atomic center a ***

            IF (PRESENT(sdab) .OR. return_derivatives) THEN
               la_start = 0
               lb_start = 0
            ELSE
               la_start = la_min_set
               lb_start = lb_min_set
            END IF

            DO da = 0, da_max-1
               ftz = 2.0_dp*zeta(ipgf)
               DO dax = 0, da
                  DO day = 0, da-dax
                     daz = da-dax-day
                     cda = coset(dax, day, daz)
                     cdax = coset(dax+1, day, daz)
                     cday = coset(dax, day+1, daz)
                     cdaz = coset(dax, day, daz+1)

!             *** [da/dAi|b] = 2*zeta*[a+1i|b] - Ni(a)[a-1i|b] ***

                     DO la = la_start, la_max-da-1
                        DO ax = 0, la
                           fax = REAL(ax, dp)
                           DO ay = 0, la-ax
                              fay = REAL(ay, dp)
                              az = la-ax-ay
                              faz = REAL(az, dp)
                              coa = coset(ax, ay, az)
                              coamx = coset(ax-1, ay, az)
                              coamy = coset(ax, ay-1, az)
                              coamz = coset(ax, ay, az-1)
                              coapx = coset(ax+1, ay, az)
                              coapy = coset(ax, ay+1, az)
                              coapz = coset(ax, ay, az+1)
                              DO lb = lb_start, lb_max_set
                                 DO bx = 0, lb
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       cob = coset(bx, by, bz)
                                       s(coa, cob, cdax) = ftz*s(coapx, cob, cda)- &
                                                           fax*s(coamx, cob, cda)
                                       s(coa, cob, cday) = ftz*s(coapy, cob, cda)- &
                                                           fay*s(coamy, cob, cda)
                                       s(coa, cob, cdaz) = ftz*s(coapz, cob, cda)- &
                                                           faz*s(coamz, cob, cda)
                                    END DO
                                 END DO
                              END DO
                           END DO
                        END DO
                     END DO

                  END DO
               END DO
            END DO

!       *** Return all the calculated derivatives of the ***
!       *** primitive overlap integrals, if requested    ***

            IF (return_derivatives) THEN
               DO k = 2, ncoset(da_max_set)
                  jstart = (k-1)*SIZE(sab, 1)
                  DO j = 1, ncoset(lb_max_set)
                     jk = jstart+j
                     DO i = 1, ncoset(la_max_set)
                        sab(na+i, nb+jk) = s(i, j, k)
                     END DO
                  END DO
               END DO
            END IF

!       *** Calculate the force contribution for the atomic center a ***

            IF (calculate_force_a) THEN
               DO k = 1, 3
                  DO j = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
                     DO i = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                        force_a(k) = force_a(k)+pab(na+i, nb+j)*s(i, j, k+1)
                     END DO
                  END DO
               END DO
            END IF

!       *** Store the first derivatives of the primitive overlap integrals ***
!       *** which are used as auxiliary integrals for the calculation of   ***
!       *** the kinetic energy integrals if requested                      ***

            IF (PRESENT(sdab)) THEN
               sdab(nda+1, nb+1, 1) = s(1, 1, 1)
               DO k = 2, 4
                  DO j = 1, ncoset(lb_max_set)
                     DO i = 1, ncoset(la_max-1)
                        sdab(nda+i, nb+j, k) = s(i, j, k)
                     END DO
                  END DO
               END DO
            END IF

            nb = nb+ncoset(lb_max_set)

         END DO

         na = na+ncoset(la_max_set)
         nda = nda+ncoset(la_max-1)

      END DO

   END SUBROUTINE overlap

! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [a|b] over
!>          Cartesian Gaussian-type functions. First and second derivatives
!> \param la_max     Max L on center A
!> \param la_min     Min L on center A
!> \param npgfa      Number of primitives on center A
!> \param rpgfa      Range of functions on A, used for screening
!> \param zeta       Exponents on center A
!> \param lb_max     Max L on center B
!> \param lb_min     Min L on center B
!> \param npgfb      Number of primitives on center B
!> \param rpgfb      Range of functions on B, used for screening
!> \param zetb       Exponents on center B
!> \param rab        Distance vector A-B
!> \param sab        Final overlap integrals
!> \param dab        First derivative overlap integrals
!> \param ddab       Second derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_ab(la_max, la_min, npgfa, rpgfa, zeta, &
                         lb_max, lb_min, npgfb, rpgfb, zetb, &
                         rab, sab, dab, ddab)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: sab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: dab, ddab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_ab', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, cob, ia, &
                                                            ib, ipgf, jpgf, la, lb, ldrr, lma, &
                                                            lmb, ma, mb, na, nb, ofa, ofb
      REAL(KIND=dp)                                      :: a, ambm, ambp, apbm, apbp, b, dumx, &
                                                            dumy, dumz, f0, rab2, tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(sab)) THEN
         lma = la_max
         lmb = lb_max
      END IF
      IF (PRESENT(dab)) THEN
         lma = la_max+1
         lmb = lb_max
      END IF
      IF (PRESENT(ddab)) THEN
         lma = la_max+1
         lmb = lb_max+1
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa = ncoset(la_min-1)
      ofb = ncoset(lb_min-1)
      na = ncoset(la_max)-ofa
      nb = ncoset(lb_max)-ofb
      IF (PRESENT(sab)) THEN
         IF(.NOT.((SIZE(sab, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",728)
         IF(.NOT.((SIZE(sab, 2) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",729)
      END IF
      IF (PRESENT(dab)) THEN
         IF(.NOT.((SIZE(dab, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",732)
         IF(.NOT.((SIZE(dab, 2) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",733)
         IF(.NOT.((SIZE(dab, 3) >= 3)))CALL cp__a("aobasis/ai_overlap.F",734)
      END IF
      IF (PRESENT(ddab)) THEN
         IF(.NOT.((SIZE(ddab, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",737)
         IF(.NOT.((SIZE(ddab, 2) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",738)
         IF(.NOT.((SIZE(ddab, 3) >= 6)))CALL cp__a("aobasis/ai_overlap.F",739)
      END IF

      ! Loops over all pairs of primitive Gaussian-type functions
      ma = 0
      DO ipgf = 1, npgfa
         mb = 0
         DO jpgf = 1, npgfb
            ! Distance Screening
            IF (rpgfa(ipgf)+rpgfb(jpgf) < tab) THEN
               IF (PRESENT(sab)) sab(ma+1:ma+na, mb+1:mb+nb) = 0.0_dp
               IF (PRESENT(dab)) dab(ma+1:ma+na, mb+1:mb+nb, 1:3) = 0.0_dp
               IF (PRESENT(ddab)) ddab(ma+1:ma+na, mb+1:mb+nb, 1:6) = 0.0_dp
               mb = mb+nb
               CYCLE
            ENDIF

            ! Calculate some prefactors
            a = zeta(ipgf)
            b = zetb(jpgf)
            zet = a+b
            xhi = a*b/zet
            rap = b*rab/zet
            rbp = -a*rab/zet

            ! [s|s] integral
            f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! Calculate the reccurence relation
            CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)-ofb
               ib = mb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)-ofa
                  ia = ma+coa
                  ! integrals
                  IF (PRESENT(sab)) THEN
                     sab(ia, ib) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*rr(az, bz, 3)
                  END IF
                  ! first derivatives
                  IF (PRESENT(dab)) THEN
                     ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                     ! dx
                     dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                     IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                     dab(ia, ib, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                     ! dy
                     dumy = 2.0_dp*a*rr(ay+1, by, 2)
                     IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                     dab(ia, ib, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                     ! dz
                     dumz = 2.0_dp*a*rr(az+1, bz, 3)
                     IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                     dab(ia, ib, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                  END IF
                  ! 2nd derivatives
                  IF (PRESENT(ddab)) THEN
                     ! (dda|b) = -4*a*b*(a+1|b+1) + 2*a*N(b)*(a+1|b-1)
                     !           + 2*b*N(a)*(a-1|b+1) - N(a)*N(b)*(a-1|b-1)
                     ! dx dx
                     apbp = f0*rr(ax+1, bx+1, 1)*rr(ay, by, 2)*rr(az, bz, 3)
                     IF (bx > 0) THEN
                        apbm = f0*rr(ax+1, bx-1, 1)*rr(ay, by, 2)*rr(az, bz, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (ax > 0) THEN
                        ambp = f0*rr(ax-1, bx+1, 1)*rr(ay, by, 2)*rr(az, bz, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (ax > 0 .AND. bx > 0) THEN
                        ambm = f0*rr(ax-1, bx-1, 1)*rr(ay, by, 2)*rr(az, bz, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 1) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(bx, dp)*apbm &
                                       +2.0_dp*b*REAL(ax, dp)*ambp-REAL(ax, dp)*REAL(bx, dp)*ambm
                     ! dx dy
                     apbp = f0*rr(ax+1, bx, 1)*rr(ay, by+1, 2)*rr(az, bz, 3)
                     IF (by > 0) THEN
                        apbm = f0*rr(ax+1, bx, 1)*rr(ay, by-1, 2)*rr(az, bz, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (ax > 0) THEN
                        ambp = f0*rr(ax-1, bx, 1)*rr(ay, by+1, 2)*rr(az, bz, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (ax > 0 .AND. by > 0) THEN
                        ambm = f0*rr(ax-1, bx, 1)*rr(ay, by-1, 2)*rr(az, bz, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 2) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(by, dp)*apbm &
                                       +2.0_dp*b*REAL(ax, dp)*ambp-REAL(ax, dp)*REAL(by, dp)*ambm
                     ! dx dz
                     apbp = f0*rr(ax+1, bx, 1)*rr(ay, by, 2)*rr(az, bz+1, 3)
                     IF (bz > 0) THEN
                        apbm = f0*rr(ax+1, bx, 1)*rr(ay, by, 2)*rr(az, bz-1, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (ax > 0) THEN
                        ambp = f0*rr(ax-1, bx, 1)*rr(ay, by, 2)*rr(az, bz+1, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (ax > 0 .AND. bz > 0) THEN
                        ambm = f0*rr(ax-1, bx, 1)*rr(ay, by, 2)*rr(az, bz-1, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 3) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(bz, dp)*apbm &
                                       +2.0_dp*b*REAL(ax, dp)*ambp-REAL(ax, dp)*REAL(bz, dp)*ambm
                     ! dy dy
                     apbp = f0*rr(ax, bx, 1)*rr(ay+1, by+1, 2)*rr(az, bz, 3)
                     IF (by > 0) THEN
                        apbm = f0*rr(ax, bx, 1)*rr(ay+1, by-1, 2)*rr(az, bz, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (ay > 0) THEN
                        ambp = f0*rr(ax, bx, 1)*rr(ay-1, by+1, 2)*rr(az, bz, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (ay > 0 .AND. by > 0) THEN
                        ambm = f0*rr(ax, bx, 1)*rr(ay-1, by-1, 2)*rr(az, bz, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 4) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(by, dp)*apbm &
                                       +2.0_dp*b*REAL(ay, dp)*ambp-REAL(ay, dp)*REAL(by, dp)*ambm
                     ! dy dz
                     apbp = f0*rr(ax, bx, 1)*rr(ay+1, by, 2)*rr(az, bz+1, 3)
                     IF (bz > 0) THEN
                        apbm = f0*rr(ax, bx, 1)*rr(ay+1, by, 2)*rr(az, bz-1, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (ay > 0) THEN
                        ambp = f0*rr(ax, bx, 1)*rr(ay-1, by, 2)*rr(az, bz+1, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (ay > 0 .AND. bz > 0) THEN
                        ambm = f0*rr(ax, bx, 1)*rr(ay-1, by, 2)*rr(az, bz-1, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 5) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(bz, dp)*apbm &
                                       +2.0_dp*b*REAL(ay, dp)*ambp-REAL(ay, dp)*REAL(bz, dp)*ambm
                     ! dz dz
                     apbp = f0*rr(ax, bx, 1)*rr(ay, by, 2)*rr(az+1, bz+1, 3)
                     IF (bz > 0) THEN
                        apbm = f0*rr(ax, bx, 1)*rr(ay, by, 2)*rr(az+1, bz-1, 3)
                     ELSE
                        apbm = 0.0_dp
                     END IF
                     IF (az > 0) THEN
                        ambp = f0*rr(ax, bx, 1)*rr(ay, by, 2)*rr(az-1, bz+1, 3)
                     ELSE
                        ambp = 0.0_dp
                     END IF
                     IF (az > 0 .AND. bz > 0) THEN
                        ambm = f0*rr(ax, bx, 1)*rr(ay, by, 2)*rr(az-1, bz-1, 3)
                     ELSE
                        ambm = 0.0_dp
                     END IF
                     ddab(ia, ib, 6) = -4.0_dp*a*b*apbm+2.0_dp*a*REAL(bz, dp)*apbm &
                                       +2.0_dp*b*REAL(az, dp)*ambp-REAL(az, dp)*REAL(bz, dp)*ambm
                  END IF
                  !
               ENDDO
               ENDDO
               ENDDO !la
            ENDDO
            ENDDO
            ENDDO !lb

            mb = mb+nb
         END DO
         ma = ma+na
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_ab

! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [aa|b] over
!>          Cartesian Gaussian-type functions.
!> \param la1_max    Max L on center A (basis 1)
!> \param la1_min    Min L on center A (basis 1)
!> \param npgfa1     Number of primitives on center A (basis 1)
!> \param rpgfa1     Range of functions on A, used for screening (basis 1)
!> \param zeta1      Exponents on center A (basis 1)
!> \param la2_max    Max L on center A (basis 2)
!> \param la2_min    Min L on center A (basis 2)
!> \param npgfa2     Number of primitives on center A (basis 2)
!> \param rpgfa2     Range of functions on A, used for screening (basis 2)
!> \param zeta2      Exponents on center A (basis 2)
!> \param lb_max     Max L on center B
!> \param lb_min     Min L on center B
!> \param npgfb      Number of primitives on center B
!> \param rpgfb      Range of functions on B, used for screening
!> \param zetb       Exponents on center B
!> \param rab        Distance vector A-B
!> \param saab       Final overlap integrals
!> \param daab       First derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_aab(la1_max, la1_min, npgfa1, rpgfa1, zeta1, &
                          la2_max, la2_min, npgfa2, rpgfa2, zeta2, &
                          lb_max, lb_min, npgfb, rpgfb, zetb, &
                          rab, saab, daab)
      INTEGER, INTENT(IN)                                :: la1_max, la1_min, npgfa1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa1, zeta1
      INTEGER, INTENT(IN)                                :: la2_max, la2_min, npgfa2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa2, zeta2
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: saab
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: daab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_aab', routineP = moduleN//':'//routineN

      INTEGER :: ax, ax1, ax2, ay, ay1, ay2, az, az1, az2, bx, by, bz, coa1, coa2, cob, i1pgf, &
         i2pgf, ia1, ia2, ib, jpgf, la1, la2, lb, ldrr, lma, lmb, ma1, ma2, mb, na1, na2, nb, &
         ofa1, ofa2, ofb
      REAL(KIND=dp)                                      :: a, b, dumx, dumy, dumz, f0, rab2, rpgfa, &
                                                            tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(saab)) THEN
         lma = la1_max+la2_max
         lmb = lb_max
      END IF
      IF (PRESENT(daab)) THEN
         lma = la1_max+la2_max+1
         lmb = lb_max
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa1 = ncoset(la1_min-1)
      ofa2 = ncoset(la2_min-1)
      ofb = ncoset(lb_min-1)
      na1 = ncoset(la1_max)-ofa1
      na2 = ncoset(la2_max)-ofa2
      nb = ncoset(lb_max)-ofb
      IF (PRESENT(saab)) THEN
         IF(.NOT.((SIZE(saab, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1015)
         IF(.NOT.((SIZE(saab, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1016)
         IF(.NOT.((SIZE(saab, 3) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",1017)
      END IF
      IF (PRESENT(daab)) THEN
         IF(.NOT.((SIZE(daab, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1020)
         IF(.NOT.((SIZE(daab, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1021)
         IF(.NOT.((SIZE(daab, 3) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",1022)
         IF(.NOT.((SIZE(daab, 4) >= 3)))CALL cp__a("aobasis/ai_overlap.F",1023)
      END IF

      ! Loops over all primitive Gaussian-type functions
      ma1 = 0
      DO i1pgf = 1, npgfa1
         ma2 = 0
         DO i2pgf = 1, npgfa2
            rpgfa = MIN(rpgfa1(i1pgf), rpgfa2(i2pgf))
            mb = 0
            DO jpgf = 1, npgfb
               ! Distance Screening
               IF (rpgfa+rpgfb(jpgf) < tab) THEN
                  IF (PRESENT(saab)) saab(ma1+1:ma1+na1, ma2+1:ma2+na2, mb+1:mb+nb) = 0.0_dp
                  IF (PRESENT(daab)) daab(ma1+1:ma1+na1, ma2+1:ma2+na2, mb+1:mb+nb, 1:3) = 0.0_dp
                  mb = mb+nb
                  CYCLE
               ENDIF

               ! Calculate some prefactors
               a = zeta1(i1pgf)+zeta2(i2pgf)
               b = zetb(jpgf)
               zet = a+b
               xhi = a*b/zet
               rap = b*rab/zet
               rbp = -a*rab/zet

               ! [ss|s] integral
               f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

               ! Calculate the reccurence relation
               CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

               DO lb = lb_min, lb_max
               DO bx = 0, lb
               DO by = 0, lb-bx
                  bz = lb-bx-by
                  cob = coset(bx, by, bz)-ofb
                  ib = mb+cob
                  DO la2 = la2_min, la2_max
                  DO ax2 = 0, la2
                  DO ay2 = 0, la2-ax2
                     az2 = la2-ax2-ay2
                     coa2 = coset(ax2, ay2, az2)-ofa2
                     ia2 = ma2+coa2
                     DO la1 = la1_min, la1_max
                     DO ax1 = 0, la1
                     DO ay1 = 0, la1-ax1
                        az1 = la1-ax1-ay1
                        coa1 = coset(ax1, ay1, az1)-ofa1
                        ia1 = ma1+coa1
                        ! integrals
                        IF (PRESENT(saab)) THEN
                           saab(ia1, ia2, ib) = f0*rr(ax1+ax2, bx, 1)*rr(ay1+ay2, by, 2)*rr(az1+az2, bz, 3)
                        END IF
                        ! first derivatives
                        IF (PRESENT(daab)) THEN
                           ax = ax1+ax2
                           ay = ay1+ay2
                           az = az1+az2
                           ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                           ! dx
                           dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                           IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                           daab(ia1, ia2, ib, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                           ! dy
                           dumy = 2.0_dp*a*rr(ay+1, by, 2)
                           IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                           daab(ia1, ia2, ib, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                           ! dz
                           dumz = 2.0_dp*a*rr(az+1, bz, 3)
                           IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                           daab(ia1, ia2, ib, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                        END IF
                        !
                     ENDDO
                     ENDDO
                     ENDDO !la1
                  ENDDO
                  ENDDO
                  ENDDO !la2
               ENDDO
               ENDDO
               ENDDO !lb

               mb = mb+nb
            END DO
            ma2 = ma2+na2
         END DO
         ma1 = ma1+na1
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_aab

! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [a|bb] over
!>          Cartesian Gaussian-type functions.
!> \param la_max     Max L on center A
!> \param la_min     Min L on center A
!> \param npgfa      Number of primitives on center A
!> \param rpgfa      Range of functions on A, used for screening
!> \param zeta       Exponents on center A
!> \param lb1_max    Max L on center B (basis 1)
!> \param lb1_min    Min L on center B (basis 1)
!> \param npgfb1     Number of primitives on center B (basis 1)
!> \param rpgfb1     Range of functions on B, used for screening (basis 1)
!> \param zetb1      Exponents on center B (basis 1)
!> \param lb2_max    Max L on center B (basis 2)
!> \param lb2_min    Min L on center B (basis 2)
!> \param npgfb2     Number of primitives on center B (basis 2)
!> \param rpgfb2     Range of functions on B, used for screening (basis 2)
!> \param zetb2      Exponents on center B (basis 2)
!> \param rab        Distance vector A-B
!> \param sabb       Final overlap integrals
!> \param dabb       First derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_abb(la_max, la_min, npgfa, rpgfa, zeta, &
                          lb1_max, lb1_min, npgfb1, rpgfb1, zetb1, &
                          lb2_max, lb2_min, npgfb2, rpgfb2, zetb2, &
                          rab, sabb, dabb)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb1_max, lb1_min, npgfb1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb1, zetb1
      INTEGER, INTENT(IN)                                :: lb2_max, lb2_min, npgfb2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb2, zetb2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: sabb
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: dabb

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_abb', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, bx1, bx2, by, by1, by2, bz, bz1, bz2, coa, cob1, cob2, ia, ib1, &
         ib2, ipgf, j1pgf, j2pgf, la, lb1, lb2, ldrr, lma, lmb, ma, mb1, mb2, na, nb1, nb2, ofa, &
         ofb1, ofb2
      REAL(KIND=dp)                                      :: a, b, dumx, dumy, dumz, f0, rab2, rpgfb, &
                                                            tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(sabb)) THEN
         lma = la_max
         lmb = lb1_max+lb2_max
      END IF
      IF (PRESENT(dabb)) THEN
         lma = la_max+1
         lmb = lb1_max+lb2_max
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa = ncoset(la_min-1)
      ofb1 = ncoset(lb1_min-1)
      ofb2 = ncoset(lb2_min-1)
      na = ncoset(la_max)-ofa
      nb1 = ncoset(lb1_max)-ofb1
      nb2 = ncoset(lb2_max)-ofb2
      IF (PRESENT(sabb)) THEN
         IF(.NOT.((SIZE(sabb, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",1196)
         IF(.NOT.((SIZE(sabb, 2) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1197)
         IF(.NOT.((SIZE(sabb, 3) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1198)
      END IF
      IF (PRESENT(dabb)) THEN
         IF(.NOT.((SIZE(dabb, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",1201)
         IF(.NOT.((SIZE(dabb, 2) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1202)
         IF(.NOT.((SIZE(dabb, 3) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1203)
         IF(.NOT.((SIZE(dabb, 4) >= 3)))CALL cp__a("aobasis/ai_overlap.F",1204)
      END IF

      ! Loops over all pairs of primitive Gaussian-type functions
      ma = 0
      DO ipgf = 1, npgfa
         mb1 = 0
         DO j1pgf = 1, npgfb1
            mb2 = 0
            DO j2pgf = 1, npgfb2
               ! Distance Screening
               rpgfb = MIN(rpgfb1(j1pgf), rpgfb2(j2pgf))
               IF (rpgfa(ipgf)+rpgfb < tab) THEN
                  IF (PRESENT(sabb)) sabb(ma+1:ma+na, mb1+1:mb1+nb1, mb2+1:mb2+nb2) = 0.0_dp
                  IF (PRESENT(dabb)) dabb(ma+1:ma+na, mb1+1:mb1+nb1, mb2+1:mb2+nb2, 1:3) = 0.0_dp
                  mb2 = mb2+nb2
                  CYCLE
               ENDIF

               ! Calculate some prefactors
               a = zeta(ipgf)
               b = zetb1(j1pgf)+zetb2(j2pgf)
               zet = a+b
               xhi = a*b/zet
               rap = b*rab/zet
               rbp = -a*rab/zet

               ! [s|s] integral
               f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

               ! Calculate the reccurence relation
               CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

               DO lb2 = lb2_min, lb2_max
               DO bx2 = 0, lb2
               DO by2 = 0, lb2-bx2
                  bz2 = lb2-bx2-by2
                  cob2 = coset(bx2, by2, bz2)-ofb2
                  ib2 = mb2+cob2
                  DO lb1 = lb1_min, lb1_max
                  DO bx1 = 0, lb1
                  DO by1 = 0, lb1-bx1
                     bz1 = lb1-bx1-by1
                     cob1 = coset(bx1, by1, bz1)-ofb1
                     ib1 = mb1+cob1
                     DO la = la_min, la_max
                     DO ax = 0, la
                     DO ay = 0, la-ax
                        az = la-ax-ay
                        coa = coset(ax, ay, az)-ofa
                        ia = ma+coa
                        ! integrals
                        IF (PRESENT(sabb)) THEN
                           sabb(ia, ib1, ib2) = f0*rr(ax, bx1+bx2, 1)*rr(ay, by1+by2, 2)*rr(az, bz1+bz2, 3)
                        END IF
                        ! first derivatives
                        IF (PRESENT(dabb)) THEN
                           bx = bx1+bx2
                           by = by1+by2
                           bz = bz1+bz2
                           ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                           ! dx
                           dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                           IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                           dabb(ia, ib1, ib2, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                           ! dy
                           dumy = 2.0_dp*a*rr(ay+1, by, 2)
                           IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                           dabb(ia, ib1, ib2, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                           ! dz
                           dumz = 2.0_dp*a*rr(az+1, bz, 3)
                           IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                           dabb(ia, ib1, ib2, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                        END IF
                        !
                     ENDDO
                     ENDDO
                     ENDDO !la
                  ENDDO
                  ENDDO
                  ENDDO !lb1
               ENDDO
               ENDDO
               ENDDO !lb2

               mb2 = mb2+nb2
            END DO
            mb1 = mb1+nb1
         END DO
         ma = ma+na
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_abb

! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [aaa|b] over
!>          Cartesian Gaussian-type functions.
!> \param la1_max    Max L on center A (basis 1)
!> \param la1_min    Min L on center A (basis 1)
!> \param npgfa1     Number of primitives on center A (basis 1)
!> \param rpgfa1     Range of functions on A, used for screening (basis 1)
!> \param zeta1      Exponents on center A (basis 1)
!> \param la2_max    Max L on center A (basis 2)
!> \param la2_min    Min L on center A (basis 2)
!> \param npgfa2     Number of primitives on center A (basis 2)
!> \param rpgfa2     Range of functions on A, used for screening (basis 2)
!> \param zeta2      Exponents on center A (basis 2)
!> \param la3_max    Max L on center A (basis 3)
!> \param la3_min    Min L on center A (basis 3)
!> \param npgfa3     Number of primitives on center A (basis 3)
!> \param rpgfa3     Range of functions on A, used for screening (basis 3)
!> \param zeta3      Exponents on center A (basis 3)
!> \param lb_max     Max L on center B
!> \param lb_min     Min L on center B
!> \param npgfb      Number of primitives on center B
!> \param rpgfb      Range of functions on B, used for screening
!> \param zetb       Exponents on center B
!> \param rab        Distance vector A-B
!> \param saaab      Final overlap integrals
!> \param daaab      First derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_aaab(la1_max, la1_min, npgfa1, rpgfa1, zeta1, &
                           la2_max, la2_min, npgfa2, rpgfa2, zeta2, &
                           la3_max, la3_min, npgfa3, rpgfa3, zeta3, &
                           lb_max, lb_min, npgfb, rpgfb, zetb, &
                           rab, saaab, daaab)
      INTEGER, INTENT(IN)                                :: la1_max, la1_min, npgfa1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa1, zeta1
      INTEGER, INTENT(IN)                                :: la2_max, la2_min, npgfa2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa2, zeta2
      INTEGER, INTENT(IN)                                :: la3_max, la3_min, npgfa3
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa3, zeta3
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: saaab
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: daaab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_aaab', routineP = moduleN//':'//routineN

      INTEGER :: ax, ax1, ax2, ax3, ay, ay1, ay2, ay3, az, az1, az2, az3, bx, by, bz, coa1, coa2, &
         coa3, cob, i1pgf, i2pgf, i3pgf, ia1, ia2, ia3, ib, jpgf, la1, la2, la3, lb, ldrr, lma, &
         lmb, ma1, ma2, ma3, mb, na1, na2, na3, nb, ofa1, ofa2, ofa3, ofb
      REAL(KIND=dp)                                      :: a, b, dumx, dumy, dumz, f0, rab2, rpgfa, &
                                                            tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(saaab)) THEN
         lma = la1_max+la2_max+la3_max
         lmb = lb_max
      END IF
      IF (PRESENT(daaab)) THEN
         lma = la1_max+la2_max+la3_max+1
         lmb = lb_max
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa1 = ncoset(la1_min-1)
      ofa2 = ncoset(la2_min-1)
      ofa3 = ncoset(la3_min-1)
      ofb = ncoset(lb_min-1)
      na1 = ncoset(la1_max)-ofa1
      na2 = ncoset(la2_max)-ofa2
      na3 = ncoset(la3_max)-ofa3
      nb = ncoset(lb_max)-ofb
      IF (PRESENT(saaab)) THEN
         IF(.NOT.((SIZE(saaab, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1387)
         IF(.NOT.((SIZE(saaab, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1388)
         IF(.NOT.((SIZE(saaab, 3) >= na3*npgfa3)))CALL cp__a("aobasis/ai_overlap.F",1389)
         IF(.NOT.((SIZE(saaab, 4) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",1390)
      END IF
      IF (PRESENT(daaab)) THEN
         IF(.NOT.((SIZE(daaab, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1393)
         IF(.NOT.((SIZE(daaab, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1394)
         IF(.NOT.((SIZE(daaab, 3) >= na3*npgfa3)))CALL cp__a("aobasis/ai_overlap.F",1395)
         IF(.NOT.((SIZE(daaab, 4) >= nb*npgfb)))CALL cp__a("aobasis/ai_overlap.F",1396)
         IF(.NOT.((SIZE(daaab, 5) >= 3)))CALL cp__a("aobasis/ai_overlap.F",1397)
      END IF

      ! Loops over all primitive Gaussian-type functions
      ma1 = 0
      DO i1pgf = 1, npgfa1
         ma2 = 0
         DO i2pgf = 1, npgfa2
            ma3 = 0
            DO i3pgf = 1, npgfa3
               rpgfa = MIN(rpgfa1(i1pgf), rpgfa2(i2pgf), rpgfa3(i3pgf))
               mb = 0
               DO jpgf = 1, npgfb
                  ! Distance Screening
                  IF (rpgfa+rpgfb(jpgf) < tab) THEN
                     IF (PRESENT(saaab)) saaab(ma1+1:ma1+na1, ma2+1:ma2+na2, ma3+1:ma3+na3, mb+1:mb+nb) = 0.0_dp
                     IF (PRESENT(daaab)) daaab(ma1+1:ma1+na1, ma2+1:ma2+na2, ma3+1:ma3+na3, mb+1:mb+nb, 1:3) = 0.0_dp
                     mb = mb+nb
                     CYCLE
                  ENDIF

                  ! Calculate some prefactors
                  a = zeta1(i1pgf)+zeta2(i2pgf)+zeta3(i3pgf)
                  b = zetb(jpgf)
                  zet = a+b
                  xhi = a*b/zet
                  rap = b*rab/zet
                  rbp = -a*rab/zet

                  ! [sss|s] integral
                  f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

                  ! Calculate the reccurence relation
                  CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

                  DO lb = lb_min, lb_max
                  DO bx = 0, lb
                  DO by = 0, lb-bx
                     bz = lb-bx-by
                     cob = coset(bx, by, bz)-ofb
                     ib = mb+cob
                     DO la3 = la3_min, la3_max
                     DO ax3 = 0, la3
                     DO ay3 = 0, la3-ax3
                        az3 = la3-ax3-ay3
                        coa3 = coset(ax3, ay3, az3)-ofa3
                        ia3 = ma3+coa3
                        DO la2 = la2_min, la2_max
                        DO ax2 = 0, la2
                        DO ay2 = 0, la2-ax2
                           az2 = la2-ax2-ay2
                           coa2 = coset(ax2, ay2, az2)-ofa2
                           ia2 = ma2+coa2
                           DO la1 = la1_min, la1_max
                           DO ax1 = 0, la1
                           DO ay1 = 0, la1-ax1
                              az1 = la1-ax1-ay1
                              coa1 = coset(ax1, ay1, az1)-ofa1
                              ia1 = ma1+coa1
                              ! integrals
                              IF (PRESENT(saaab)) THEN
                                 saaab(ia1, ia2, ia3, ib) = f0*rr(ax1+ax2+ax3, bx, 1)* &
                                                            rr(ay1+ay2+ay3, by, 2)*rr(az1+az2+az3, bz, 3)
                              END IF
                              ! first derivatives
                              IF (PRESENT(daaab)) THEN
                                 ax = ax1+ax2+ax3
                                 ay = ay1+ay2+ay3
                                 az = az1+az2+az3
                                 ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                                 ! dx
                                 dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                                 IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                                 daaab(ia1, ia2, ia3, ib, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                                 ! dy
                                 dumy = 2.0_dp*a*rr(ay+1, by, 2)
                                 IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                                 daaab(ia1, ia2, ia3, ib, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                                 ! dz
                                 dumz = 2.0_dp*a*rr(az+1, bz, 3)
                                 IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                                 daaab(ia1, ia2, ia3, ib, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                              END IF
                              !
                           ENDDO
                           ENDDO
                           ENDDO !la1
                        ENDDO
                        ENDDO
                        ENDDO !la2
                     ENDDO
                     ENDDO
                     ENDDO !la3
                  ENDDO
                  ENDDO
                  ENDDO !lb

                  mb = mb+nb
               END DO
               ma3 = ma3+na3
            END DO
            ma2 = ma2+na2
         END DO
         ma1 = ma1+na1
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_aaab
! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [aa|bb] over
!>          Cartesian Gaussian-type functions.
!> \param la1_max    Max L on center A (basis 1)
!> \param la1_min    Min L on center A (basis 1)
!> \param npgfa1     Number of primitives on center A (basis 1)
!> \param rpgfa1     Range of functions on A, used for screening (basis 1)
!> \param zeta1      Exponents on center A (basis 1)
!> \param la2_max    Max L on center A (basis 2)
!> \param la2_min    Min L on center A (basis 2)
!> \param npgfa2     Number of primitives on center A (basis 2)
!> \param rpgfa2     Range of functions on A, used for screening (basis 2)
!> \param zeta2      Exponents on center A (basis 2)
!> \param lb1_max    Max L on center B (basis 3)
!> \param lb1_min    Min L on center B (basis 3)
!> \param npgfb1     Number of primitives on center B (basis 3)
!> \param rpgfb1     Range of functions on B, used for screening (basis 3)
!> \param zetb1      Exponents on center B (basis 3)
!> \param lb2_max    Max L on center B (basis 4)
!> \param lb2_min    Min L on center B (basis 4)
!> \param npgfb2     Number of primitives on center B (basis 4)
!> \param rpgfb2     Range of functions on B, used for screening (basis 4)
!> \param zetb2      Exponents on center B (basis 4)
!> \param rab        Distance vector A-B
!> \param saabb      Final overlap integrals
!> \param daabb      First derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_aabb(la1_max, la1_min, npgfa1, rpgfa1, zeta1, &
                           la2_max, la2_min, npgfa2, rpgfa2, zeta2, &
                           lb1_max, lb1_min, npgfb1, rpgfb1, zetb1, &
                           lb2_max, lb2_min, npgfb2, rpgfb2, zetb2, &
                           rab, saabb, daabb)
      INTEGER, INTENT(IN)                                :: la1_max, la1_min, npgfa1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa1, zeta1
      INTEGER, INTENT(IN)                                :: la2_max, la2_min, npgfa2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa2, zeta2
      INTEGER, INTENT(IN)                                :: lb1_max, lb1_min, npgfb1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb1, zetb1
      INTEGER, INTENT(IN)                                :: lb2_max, lb2_min, npgfb2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb2, zetb2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: saabb
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: daabb

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_aabb', routineP = moduleN//':'//routineN

      INTEGER :: ax, ax1, ax2, ay, ay1, ay2, az, az1, az2, bx, bx1, bx2, by, by1, by2, bz, bz1, &
         bz2, coa1, coa2, cob1, cob2, i1pgf, i2pgf, ia1, ia2, ib1, ib2, j1pgf, j2pgf, la1, la2, &
         lb1, lb2, ldrr, lma, lmb, ma1, ma2, mb1, mb2, na1, na2, nb1, nb2, ofa1, ofa2, ofb1, ofb2
      REAL(KIND=dp)                                      :: a, b, dumx, dumy, dumz, f0, rab2, rpgfa, &
                                                            rpgfb, tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(saabb)) THEN
         lma = la1_max+la2_max
         lmb = lb1_max+lb2_max
      END IF
      IF (PRESENT(daabb)) THEN
         lma = la1_max+la2_max+1
         lmb = lb1_max+lb2_max
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa1 = ncoset(la1_min-1)
      ofa2 = ncoset(la2_min-1)
      ofb1 = ncoset(lb1_min-1)
      ofb2 = ncoset(lb2_min-1)
      na1 = ncoset(la1_max)-ofa1
      na2 = ncoset(la2_max)-ofa2
      nb1 = ncoset(lb1_max)-ofb1
      nb2 = ncoset(lb2_max)-ofb2
      IF (PRESENT(saabb)) THEN
         IF(.NOT.((SIZE(saabb, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1593)
         IF(.NOT.((SIZE(saabb, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1594)
         IF(.NOT.((SIZE(saabb, 3) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1595)
         IF(.NOT.((SIZE(saabb, 4) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1596)
      END IF
      IF (PRESENT(daabb)) THEN
         IF(.NOT.((SIZE(daabb, 1) >= na1*npgfa1)))CALL cp__a("aobasis/ai_overlap.F",1599)
         IF(.NOT.((SIZE(daabb, 2) >= na2*npgfa2)))CALL cp__a("aobasis/ai_overlap.F",1600)
         IF(.NOT.((SIZE(daabb, 3) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1601)
         IF(.NOT.((SIZE(daabb, 4) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1602)
         IF(.NOT.((SIZE(daabb, 5) >= 3)))CALL cp__a("aobasis/ai_overlap.F",1603)
      END IF

      ! Loops over all primitive Gaussian-type functions
      ma1 = 0
      DO i1pgf = 1, npgfa1
         ma2 = 0
         DO i2pgf = 1, npgfa2
            mb1 = 0
            rpgfa = MIN(rpgfa1(i1pgf), rpgfa2(i2pgf))
            DO j1pgf = 1, npgfb1
               mb2 = 0
               DO j2pgf = 1, npgfb2
                  rpgfb = MIN(rpgfb1(j1pgf), rpgfb2(j2pgf))
                  ! Distance Screening
                  IF (rpgfa+rpgfb < tab) THEN
                     IF (PRESENT(saabb)) saabb(ma1+1:ma1+na1, ma2+1:ma2+na2, mb1+1:mb1+nb1, mb2+1:mb2+nb2) = 0.0_dp
                     IF (PRESENT(daabb)) daabb(ma1+1:ma1+na1, ma2+1:ma2+na2, mb1+1:mb1+nb1, mb2+1:mb2+nb2, 1:3) = 0.0_dp
                     mb2 = mb2+nb2
                     CYCLE
                  ENDIF

                  ! Calculate some prefactors
                  a = zeta1(i1pgf)+zeta2(i2pgf)
                  b = zetb1(j1pgf)+zetb2(j2pgf)
                  zet = a+b
                  xhi = a*b/zet
                  rap = b*rab/zet
                  rbp = -a*rab/zet

                  ! [ss|ss] integral
                  f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

                  ! Calculate the reccurence relation
                  CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

                  DO lb2 = lb2_min, lb2_max
                  DO bx2 = 0, lb2
                  DO by2 = 0, lb2-bx2
                     bz2 = lb2-bx2-by2
                     cob2 = coset(bx2, by2, bz2)-ofb2
                     ib2 = mb2+cob2
                     DO lb1 = lb1_min, lb1_max
                     DO bx1 = 0, lb1
                     DO by1 = 0, lb1-bx1
                        bz1 = lb1-bx1-by1
                        cob1 = coset(bx1, by1, bz1)-ofb1
                        ib1 = mb1+cob1
                        DO la2 = la2_min, la2_max
                        DO ax2 = 0, la2
                        DO ay2 = 0, la2-ax2
                           az2 = la2-ax2-ay2
                           coa2 = coset(ax2, ay2, az2)-ofa2
                           ia2 = ma2+coa2
                           DO la1 = la1_min, la1_max
                           DO ax1 = 0, la1
                           DO ay1 = 0, la1-ax1
                              az1 = la1-ax1-ay1
                              coa1 = coset(ax1, ay1, az1)-ofa1
                              ia1 = ma1+coa1
                              ! integrals
                              IF (PRESENT(saabb)) THEN
                                 saabb(ia1, ia2, ib1, ib2) = f0*rr(ax1+ax2, bx1+bx2, 1)* &
                                                             rr(ay1+ay2, by1+by2, 2)*rr(az1+az2, bz1+bz2, 3)
                              END IF
                              ! first derivatives
                              IF (PRESENT(daabb)) THEN
                                 ax = ax1+ax2
                                 ay = ay1+ay2
                                 az = az1+az2
                                 bx = bx1+bx2
                                 by = by1+by2
                                 bz = bz1+bz2
                                 ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                                 ! dx
                                 dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                                 IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                                 daabb(ia1, ia2, ib1, ib2, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                                 ! dy
                                 dumy = 2.0_dp*a*rr(ay+1, by, 2)
                                 IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                                 daabb(ia1, ia2, ib1, ib2, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                                 ! dz
                                 dumz = 2.0_dp*a*rr(az+1, bz, 3)
                                 IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                                 daabb(ia1, ia2, ib1, ib2, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                              END IF
                              !
                           ENDDO
                           ENDDO
                           ENDDO !la1
                        ENDDO
                        ENDDO
                        ENDDO !la2
                     ENDDO
                     ENDDO
                     ENDDO !lb1
                  ENDDO
                  ENDDO
                  ENDDO !lb2

                  mb2 = mb2+nb2
               END DO
               mb1 = mb1+nb1
            END DO
            ma2 = ma2+na2
         END DO
         ma1 = ma1+na1
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_aabb
! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [a|bbb] over
!>          Cartesian Gaussian-type functions.
!> \param la_max     Max L on center A
!> \param la_min     Min L on center A
!> \param npgfa      Number of primitives on center A
!> \param rpgfa      Range of functions on A, used for screening
!> \param zeta       Exponents on center A
!> \param lb1_max    Max L on center B (basis 1)
!> \param lb1_min    Min L on center B (basis 1)
!> \param npgfb1     Number of primitives on center B (basis 1)
!> \param rpgfb1     Range of functions on B, used for screening (basis 1)
!> \param zetb1      Exponents on center B (basis 1)
!> \param lb2_max    Max L on center B (basis 2)
!> \param lb2_min    Min L on center B (basis 2)
!> \param npgfb2     Number of primitives on center B (basis 2)
!> \param rpgfb2     Range of functions on B, used for screening (basis 2)
!> \param zetb2      Exponents on center B (basis 2)
!> \param lb3_max    Max L on center B (basis 3)
!> \param lb3_min    Min L on center B (basis 3)
!> \param npgfb3     Number of primitives on center B (basis 3)
!> \param rpgfb3     Range of functions on B, used for screening (basis 3)
!> \param zetb3      Exponents on center B (basis 3)
!> \param rab        Distance vector A-B
!> \param sabbb      Final overlap integrals
!> \param dabbb      First derivative overlap integrals
!> \date    01.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_abbb(la_max, la_min, npgfa, rpgfa, zeta, &
                           lb1_max, lb1_min, npgfb1, rpgfb1, zetb1, &
                           lb2_max, lb2_min, npgfb2, rpgfb2, zetb2, &
                           lb3_max, lb3_min, npgfb3, rpgfb3, zetb3, &
                           rab, sabbb, dabbb)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb1_max, lb1_min, npgfb1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb1, zetb1
      INTEGER, INTENT(IN)                                :: lb2_max, lb2_min, npgfb2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb2, zetb2
      INTEGER, INTENT(IN)                                :: lb3_max, lb3_min, npgfb3
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb3, zetb3
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: sabbb
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: dabbb

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_abbb', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, bx1, bx2, bx3, by, by1, by2, by3, bz, bz1, bz2, bz3, coa, cob1, &
         cob2, cob3, ia, ib1, ib2, ib3, ipgf, j1pgf, j2pgf, j3pgf, la, lb1, lb2, lb3, ldrr, lma, &
         lmb, ma, mb1, mb2, mb3, na, nb1, nb2, nb3, ofa, ofb1, ofb2, ofb3
      REAL(KIND=dp)                                      :: a, b, dumx, dumy, dumz, f0, rab2, rpgfb, &
                                                            tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(sabbb)) THEN
         lma = la_max
         lmb = lb1_max+lb2_max+lb3_max
      END IF
      IF (PRESENT(dabbb)) THEN
         lma = la_max+1
         lmb = lb1_max+lb2_max+lb3_max
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa = ncoset(la_min-1)
      ofb1 = ncoset(lb1_min-1)
      ofb2 = ncoset(lb2_min-1)
      ofb3 = ncoset(lb3_min-1)
      na = ncoset(la_max)-ofa
      nb1 = ncoset(lb1_max)-ofb1
      nb2 = ncoset(lb2_max)-ofb2
      nb3 = ncoset(lb3_max)-ofb3
      IF (PRESENT(sabbb)) THEN
         IF(.NOT.((SIZE(sabbb, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",1803)
         IF(.NOT.((SIZE(sabbb, 2) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1804)
         IF(.NOT.((SIZE(sabbb, 3) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1805)
         IF(.NOT.((SIZE(sabbb, 4) >= nb3*npgfb3)))CALL cp__a("aobasis/ai_overlap.F",1806)
      END IF
      IF (PRESENT(dabbb)) THEN
         IF(.NOT.((SIZE(dabbb, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_overlap.F",1809)
         IF(.NOT.((SIZE(dabbb, 2) >= nb1*npgfb1)))CALL cp__a("aobasis/ai_overlap.F",1810)
         IF(.NOT.((SIZE(dabbb, 3) >= nb2*npgfb2)))CALL cp__a("aobasis/ai_overlap.F",1811)
         IF(.NOT.((SIZE(dabbb, 4) >= nb3*npgfb3)))CALL cp__a("aobasis/ai_overlap.F",1812)
         IF(.NOT.((SIZE(dabbb, 5) >= 3)))CALL cp__a("aobasis/ai_overlap.F",1813)
      END IF

      ! Loops over all pairs of primitive Gaussian-type functions
      ma = 0
      DO ipgf = 1, npgfa
         mb1 = 0
         DO j1pgf = 1, npgfb1
            mb2 = 0
            DO j2pgf = 1, npgfb2
               mb3 = 0
               DO j3pgf = 1, npgfb3
                  ! Distance Screening
                  rpgfb = MIN(rpgfb1(j1pgf), rpgfb2(j2pgf), rpgfb3(j3pgf))
                  IF (rpgfa(ipgf)+rpgfb < tab) THEN
                     IF (PRESENT(sabbb)) sabbb(ma+1:ma+na, mb1+1:mb1+nb1, mb2+1:mb2+nb2, mb3+1:mb3+nb3) = 0.0_dp
                     IF (PRESENT(dabbb)) dabbb(ma+1:ma+na, mb1+1:mb1+nb1, mb2+1:mb2+nb2, mb3+1:mb3+nb3, 1:3) = 0.0_dp
                     mb3 = mb3+nb3
                     CYCLE
                  ENDIF

                  ! Calculate some prefactors
                  a = zeta(ipgf)
                  b = zetb1(j1pgf)+zetb2(j2pgf)+zetb3(j3pgf)
                  zet = a+b
                  xhi = a*b/zet
                  rap = b*rab/zet
                  rbp = -a*rab/zet

                  ! [s|s] integral
                  f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

                  ! Calculate the reccurence relation
                  CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

                  DO lb3 = lb3_min, lb3_max
                  DO bx3 = 0, lb3
                  DO by3 = 0, lb3-bx3
                     bz3 = lb3-bx3-by3
                     cob3 = coset(bx3, by3, bz3)-ofb3
                     ib3 = mb3+cob3
                     DO lb2 = lb2_min, lb2_max
                     DO bx2 = 0, lb2
                     DO by2 = 0, lb2-bx2
                        bz2 = lb2-bx2-by2
                        cob2 = coset(bx2, by2, bz2)-ofb2
                        ib2 = mb2+cob2
                        DO lb1 = lb1_min, lb1_max
                        DO bx1 = 0, lb1
                        DO by1 = 0, lb1-bx1
                           bz1 = lb1-bx1-by1
                           cob1 = coset(bx1, by1, bz1)-ofb1
                           ib1 = mb1+cob1
                           DO la = la_min, la_max
                           DO ax = 0, la
                           DO ay = 0, la-ax
                              az = la-ax-ay
                              coa = coset(ax, ay, az)-ofa
                              ia = ma+coa
                              ! integrals
                              IF (PRESENT(sabbb)) THEN
                                 sabbb(ia, ib1, ib2, ib3) = f0*rr(ax, bx1+bx2+bx3, 1)* &
                                                            rr(ay, by1+by2+by3, 2)*rr(az, bz1+bz2+bz3, 3)
                              END IF
                              ! first derivatives
                              IF (PRESENT(dabbb)) THEN
                                 bx = bx1+bx2+bx3
                                 by = by1+by2+by3
                                 bz = bz1+bz2+bz3
                                 ! (da|b) = 2*a*(a+1|b) - N(a)*(a-1|b)
                                 ! dx
                                 dumx = 2.0_dp*a*rr(ax+1, bx, 1)
                                 IF (ax > 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                                 dabbb(ia, ib1, ib2, ib3, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                                 ! dy
                                 dumy = 2.0_dp*a*rr(ay+1, by, 2)
                                 IF (ay > 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                                 dabbb(ia, ib1, ib2, ib3, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                                 ! dz
                                 dumz = 2.0_dp*a*rr(az+1, bz, 3)
                                 IF (az > 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                                 dabbb(ia, ib1, ib2, ib3, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                              END IF
                              !
                           ENDDO
                           ENDDO
                           ENDDO !la
                        ENDDO
                        ENDDO
                        ENDDO !lb1
                     ENDDO
                     ENDDO
                     ENDDO !lb2
                  ENDDO
                  ENDDO
                  ENDDO !lb3

                  mb3 = mb3+nb3
               END DO
               mb2 = mb2+nb2
            END DO
            mb1 = mb1+nb1
         END DO
         ma = ma+na
      END DO

      DEALLOCATE (rr)

   END SUBROUTINE overlap_abbb

! **************************************************************************************************
!> \brief   Calculation of the two-center overlap integrals [a|b] over
!>          Spherical Gaussian-type functions.
!> \param la         Max L on center A
!> \param zeta       Exponents on center A
!> \param lb         Max L on center B
!> \param zetb       Exponents on center B
!> \param rab        Distance vector A-B
!> \param sab        Final overlap integrals
!> \date    01.03.2016
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_ab_s(la, zeta, lb, zetb, rab, sab)
      INTEGER, INTENT(IN)                                :: la
      REAL(KIND=dp), INTENT(IN)                          :: zeta
      INTEGER, INTENT(IN)                                :: lb
      REAL(KIND=dp), INTENT(IN)                          :: zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: sab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_ab_s', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: huge4 = HUGE(1._dp)/4._dp

      INTEGER                                            :: nca, ncb, nsa, nsb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cab
      REAL(KIND=dp), DIMENSION(1)                        :: rpgf, za, zb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: c2sa, c2sb

      rpgf(1) = huge4
      za(1) = zeta
      zb(1) = zetb

      nca = nco(la)
      ncb = nco(lb)
      ALLOCATE (cab(nca, ncb))
      nsa = nso(la)
      nsb = nso(lb)

      CALL overlap_ab(la, la, 1, rpgf, za, lb, lb, 1, rpgf, zb, rab, cab)

      c2sa => orbtramat(la)%c2s
      c2sb => orbtramat(lb)%c2s
      sab(1:nsa, 1:nsb) = MATMUL(c2sa(1:nsa, 1:nca), &
                                 MATMUL(cab(1:nca, 1:ncb), TRANSPOSE(c2sb(1:nsb, 1:ncb))))

      DEALLOCATE (cab)

   END SUBROUTINE overlap_ab_s

! **************************************************************************************************
!> \brief   Calculation of the overlap integrals [a|b] over
!>          cubic periodic Spherical Gaussian-type functions.
!> \param la         Max L on center A
!> \param zeta       Exponents on center A
!> \param lb         Max L on center B
!> \param zetb       Exponents on center B
!> \param alat       Lattice constant
!> \param sab        Final overlap integrals
!> \date    01.03.2016
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE overlap_ab_sp(la, zeta, lb, zetb, alat, sab)
      INTEGER, INTENT(IN)                                :: la
      REAL(KIND=dp), INTENT(IN)                          :: zeta
      INTEGER, INTENT(IN)                                :: lb
      REAL(KIND=dp), INTENT(IN)                          :: zetb, alat
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: sab

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_ab_sp', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: zfg
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: fun, gun
      INTEGER                                            :: ax, ay, az, bx, by, bz, i, ia, ib, l, &
                                                            l1, l2, na, nb, nca, ncb, nmax, nsa, &
                                                            nsb
      REAL(KIND=dp)                                      :: oa, ob, ovol, zm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: fexp, gexp, gval
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cab
      REAL(KIND=dp), DIMENSION(0:3, 0:3)                 :: fgsum
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: c2sa, c2sb

      nca = nco(la)
      ncb = nco(lb)
      ALLOCATE (cab(nca, ncb))
      cab = 0.0_dp
      nsa = nso(la)
      nsb = nso(lb)

      zm = MIN(zeta, zetb)
      nmax = NINT(1.81_dp*alat*SQRT(zm)+1.0_dp)
      ALLOCATE (fun(-nmax:nmax, 0:la), gun(-nmax:nmax, 0:lb), &
                fexp(-nmax:nmax), gexp(-nmax:nmax), gval(-nmax:nmax))

      oa = 1._dp/zeta
      ob = 1._dp/zetb
      DO i = -nmax, nmax
         gval(i) = twopi/alat*REAL(i, KIND=dp)
         fexp(i) = SQRT(oa*pi)*EXP(-0.25_dp*oa*gval(i)**2)
         gexp(i) = SQRT(ob*pi)*EXP(-0.25_dp*ob*gval(i)**2)
      END DO
      DO l = 0, la
         IF (l == 0) THEN
            fun(:, l) = CMPLX(1.0_dp, 0.0_dp, KIND=dp)
         ELSEIF (l == 1) THEN
            fun(:, l) = CMPLX(0.0_dp, 0.5_dp*oa*gval(:), KIND=dp)
         ELSEIF (l == 2) THEN
            fun(:, l) = CMPLX(-(0.5_dp*oa*gval(:))**2, 0.0_dp, KIND=dp)
            fun(:, l) = fun(:, l)+CMPLX(0.5_dp*oa, 0.0_dp, KIND=dp)
         ELSEIF (l == 3) THEN
            fun(:, l) = CMPLX(0.0_dp, -(0.5_dp*oa*gval(:))**3, KIND=dp)
            fun(:, l) = fun(:, l)+CMPLX(0.0_dp, 0.75_dp*oa*oa*gval(:), KIND=dp)
         ELSE
            CALL cp__b("aobasis/ai_overlap.F",2035,"l value too high")
         END IF
      END DO
      DO l = 0, lb
         IF (l == 0) THEN
            gun(:, l) = CMPLX(1.0_dp, 0.0_dp, KIND=dp)
         ELSEIF (l == 1) THEN
            gun(:, l) = CMPLX(0.0_dp, 0.5_dp*ob*gval(:), KIND=dp)
         ELSEIF (l == 2) THEN
            gun(:, l) = CMPLX(-(0.5_dp*ob*gval(:))**2, 0.0_dp, KIND=dp)
            gun(:, l) = gun(:, l)+CMPLX(0.5_dp*ob, 0.0_dp, KIND=dp)
         ELSEIF (l == 3) THEN
            gun(:, l) = CMPLX(0.0_dp, -(0.5_dp*ob*gval(:))**3, KIND=dp)
            gun(:, l) = gun(:, l)+CMPLX(0.0_dp, 0.75_dp*ob*ob*gval(:), KIND=dp)
         ELSE
            CALL cp__b("aobasis/ai_overlap.F",2050,"l value too high")
         END IF
      END DO

      fgsum = 0.0_dp
      DO l1 = 0, la
         DO l2 = 0, lb
            zfg = SUM(CONJG(fun(:, l1))*fexp(:)*gun(:, l2)*gexp(:))
            fgsum(l1, l2) = REAL(zfg, KIND=dp)
         END DO
      END DO

      na = ncoset(la-1)
      nb = ncoset(lb-1)
      DO ax = 0, la
         DO ay = 0, la-ax
            az = la-ax-ay
            ia = coset(ax, ay, az)-na
            DO bx = 0, lb
               DO by = 0, lb-bx
                  bz = lb-bx-by
                  ib = coset(bx, by, bz)-nb
                  cab(ia, ib) = fgsum(ax, bx)*fgsum(ay, by)*fgsum(az, bz)
               END DO
            END DO
         END DO
      END DO

      c2sa => orbtramat(la)%c2s
      c2sb => orbtramat(lb)%c2s
      sab(1:nsa, 1:nsb) = MATMUL(c2sa(1:nsa, 1:nca), &
                                 MATMUL(cab(1:nca, 1:ncb), TRANSPOSE(c2sb(1:nsb, 1:ncb))))
      ovol = 1._dp/(alat**3)
      sab(1:nsa, 1:nsb) = ovol*sab(1:nsa, 1:nsb)

      DEALLOCATE (cab, fun, gun, fexp, gexp, gval)

   END SUBROUTINE overlap_ab_sp

END MODULE ai_overlap
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Two-center overlap integrals over Cartesian Gaussian-type functions
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \author Dorothea Golze
! **************************************************************************************************
MODULE ai_overlap_debug

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap_debug'

   INTEGER, PARAMETER            :: lmax = 5

   REAL(dp)                      :: xa, xb
   REAL(dp), DIMENSION(3)        :: A, B
   REAL(dp), DIMENSION(3)        :: P
   REAL(dp)                      :: xsi, zeta, ss

   PRIVATE
   PUBLIC :: init_os_overlap2, os_overlap2

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of overlap integrals over
!>          Cartesian Gaussian-type functions.
!> \param ya ...
!> \param yb ...
!> \param rA ...
!> \param rB ...
! **************************************************************************************************
   SUBROUTINE init_os_overlap2(ya, yb, rA, rB)
      REAL(dp)                                           :: ya, yb
      REAL(dp), DIMENSION(3)                             :: rA, rB

      xa = ya
      xb = yb
      A = rA
      B = rB

      xsi = xa+xb
      zeta = xa*xb/xsi

      P = (xa*A+xb*B)/xsi

      ss = (pi/xsi)**(3._dp/2._dp)*EXP(-zeta*SUM((A-B)**2))

   END SUBROUTINE init_os_overlap2

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param an ...
!> \param bn ...
!> \retval IAB ...
! **************************************************************************************************
   RECURSIVE FUNCTION os_overlap2(an, bn) RESULT(IAB)
      INTEGER, DIMENSION(3)                              :: an, bn
      REAL(dp)                                           :: IAB

      INTEGER, DIMENSION(3), PARAMETER                   :: i1 = (/1, 0, 0/), i2 = (/0, 1, 0/), &
                                                            i3 = (/0, 0, 1/)

      IAB = 0._dp
      IF (ANY(an < 0)) RETURN
      IF (ANY(bn < 0)) RETURN

      IF (SUM(an+bn) == 0) THEN
         IAB = ss
         RETURN
      END IF

      IF (bn(1) > 0) THEN
         IAB = os_overlap2(an+i1, bn-i1)+(A(1)-B(1))*os_overlap2(an, bn-i1)
      ELSEIF (bn(2) > 0) THEN
         IAB = os_overlap2(an+i2, bn-i2)+(A(2)-B(2))*os_overlap2(an, bn-i2)
      ELSEIF (bn(3) > 0) THEN
         IAB = os_overlap2(an+i3, bn-i3)+(A(3)-B(3))*os_overlap2(an, bn-i3)
      ELSE
         IF (an(1) > 0) THEN
            IAB = (P(1)-A(1))*os_overlap2(an-i1, bn)+ &
                  0.5_dp*(an(1)-1)/xsi*os_overlap2(an-i1-i1, bn)
         ELSEIF (an(2) > 0) THEN
            IAB = (P(2)-A(2))*os_overlap2(an-i2, bn)+ &
                  0.5_dp*(an(2)-1)/xsi*os_overlap2(an-i2-i2, bn)
         ELSEIF (an(3) > 0) THEN
            IAB = (P(3)-A(3))*os_overlap2(an-i3, bn)+ &
                  0.5_dp*(an(3)-1)/xsi*os_overlap2(an-i3-i3, bn)
         ELSE
            CALL cp__b("aobasis/ai_overlap_debug.F",103,"I(0000)")
         ENDIF
      ENDIF

   END FUNCTION os_overlap2

! **************************************************************************************************

END MODULE ai_overlap_debug
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Three-center integrals over Cartesian Gaussian-type functions
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \author Dorothea Golze
! **************************************************************************************************
MODULE ai_overlap3_debug

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap3_debug'

   INTEGER, PARAMETER            :: lmax = 5

   REAL(dp)                      :: xa, xb, xc
   REAL(dp), DIMENSION(3)        :: A, B, C
   REAL(dp), DIMENSION(3)        :: P, G
   REAL(dp)                      :: xsi, zeta, sss

   PRIVATE
   PUBLIC :: init_os_overlap3, os_overlap3

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of three-center integrals over
!>          Cartesian Gaussian-type functions.
!> \param ya ...
!> \param yb ...
!> \param yc ...
!> \param rA ...
!> \param rB ...
!> \param rC ...
! **************************************************************************************************
   SUBROUTINE init_os_overlap3(ya, yb, yc, rA, rB, rC)
      REAL(dp)                                           :: ya, yb, yc
      REAL(dp), DIMENSION(3)                             :: rA, rB, rC

      REAL(dp)                                           :: fpc, ss

      xa = ya
      xb = yb
      xc = yc
      A = rA
      B = rB
      C = rC

      xsi = xa+xb
      zeta = xa*xb/xsi

      P = (xa*A+xb*B)/xsi
      G = (xsi*P+xc*C)/(xsi+xc)

      ss = (pi/xsi)**(3._dp/2._dp)*EXP(-zeta*SUM((A-B)**2))

      fpc = EXP(-xsi*xc/(xsi+xc)*SUM((P-C)**2))
      sss = (xsi/(xsi+xc))**(3._dp/2._dp)*ss*fpc

   END SUBROUTINE init_os_overlap3

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param an ...
!> \param cn ...
!> \param bn ...
!> \retval IACB ...
! **************************************************************************************************
   RECURSIVE FUNCTION os_overlap3(an, cn, bn) RESULT(IACB)
      INTEGER, DIMENSION(3)                              :: an, cn, bn
      REAL(dp)                                           :: IACB

      INTEGER, DIMENSION(3), PARAMETER                   :: i1 = (/1, 0, 0/), i2 = (/0, 1, 0/), &
                                                            i3 = (/0, 0, 1/)

      IACB = 0._dp
      IF (ANY(an < 0)) RETURN
      IF (ANY(bn < 0)) RETURN
      IF (ANY(cn < 0)) RETURN

      IF (SUM(an+cn+bn) == 0) THEN
         IACB = sss
         RETURN
      END IF

      IF (bn(1) > 0) THEN
         IACB = os_overlap3(an, cn+i1, bn-i1)+(C(1)-B(1))*os_overlap3(an, cn, bn-i1)
      ELSEIF (bn(2) > 0) THEN
         IACB = os_overlap3(an, cn+i2, bn-i2)+(C(2)-B(2))*os_overlap3(an, cn, bn-i2)
      ELSEIF (bn(3) > 0) THEN
         IACB = os_overlap3(an, cn+i3, bn-i3)+(C(3)-B(3))*os_overlap3(an, cn, bn-i3)
      ELSE
         IF (cn(1) > 0) THEN
            IACB = os_overlap3(an+i1, cn-i1, bn)+(A(1)-C(1))*os_overlap3(an, cn-i1, bn)
         ELSEIF (cn(2) > 0) THEN
            IACB = os_overlap3(an+i2, cn-i2, bn)+(A(2)-C(2))*os_overlap3(an, cn-i2, bn)
         ELSEIF (cn(3) > 0) THEN
            IACB = os_overlap3(an+i3, cn-i3, bn)+(A(3)-C(3))*os_overlap3(an, cn-i3, bn)
         ELSE
            IF (an(1) > 0) THEN
               IACB = (G(1)-A(1))*os_overlap3(an-i1, cn, bn)+ &
                      0.5_dp*(an(1)-1)/(xsi+xc)*os_overlap3(an-i1-i1, cn, bn)
            ELSEIF (an(2) > 0) THEN
               IACB = (G(2)-A(2))*os_overlap3(an-i2, cn, bn)+ &
                      0.5_dp*(an(2)-1)/(xsi+xc)*os_overlap3(an-i2-i2, cn, bn)
            ELSEIF (an(3) > 0) THEN
               IACB = (G(3)-A(3))*os_overlap3(an-i3, cn, bn)+ &
                      0.5_dp*(an(3)-1)/(xsi+xc)*os_overlap3(an-i3-i3, cn, bn)
            ELSE
               CALL cp__b("aobasis/ai_overlap3_debug.F",122,"I(0000)")
            ENDIF
         ENDIF
      ENDIF

   END FUNCTION os_overlap3

! **************************************************************************************************

END MODULE ai_overlap3_debug
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the overlap integrals over Cartesian Gaussian-type
!>      functions.
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par Parameters
!>      - ax,ay,az  : Angular momentum index numbers of orbital a.
!>      - bx,by,bz  : Angular momentum index numbers of orbital b.
!>      - coset     : Cartesian orbital set pointer.
!>      - dab       : Distance between the atomic centers a and b.
!>      - l{a,b}    : Angular momentum quantum number of shell a or b.
!>      - l{a,b}_max: Maximum angular momentum quantum number of shell a or b.
!>      - l{a,b}_min: Minimum angular momentum quantum number of shell a or b.
!>      - rab       : Distance vector between the atomic centers a and b.
!>      - rpgf{a,b} : Radius of the primitive Gaussian-type function a or b.
!>      - sab       : Shell set of overlap integrals.
!>      - zet{a,b}  : Exponents of the Gaussian-type functions a or b.
!>      - zetp      : Reciprocal of the sum of the exponents of orbital a and b.
! **************************************************************************************************
MODULE ai_overlap_aabb

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap_aabb'

! *** Public subroutines ***
   PUBLIC :: overlap_aabb

CONTAINS

! **************************************************************************************************
!> \brief   Purpose: Calculation of the two-center overlap integrals [aa|bb]
!>          over Cartesian Gaussian-type functions.
!> \param la_max_set1 ...
!> \param la_min_set1 ...
!> \param npgfa1 ...
!> \param rpgfa1 ...
!> \param zeta1 ...
!> \param la_max_set2 ...
!> \param la_min_set2 ...
!> \param npgfa2 ...
!> \param rpgfa2 ...
!> \param zeta2 ...
!> \param lb_max_set1 ...
!> \param lb_min_set1 ...
!> \param npgfb1 ...
!> \param rpgfb1 ...
!> \param zetb1 ...
!> \param lb_max_set2 ...
!> \param lb_min_set2 ...
!> \param npgfb2 ...
!> \param rpgfb2 ...
!> \param zetb2 ...
!> \param asets_equal ...
!> \param bsets_equal ...
!> \param rab ...
!> \param dab ...
!> \param saabb ...
!> \param s ...
!> \param lds ...
!> \date    06.2014
!> \author  Dorothea Golze
! **************************************************************************************************
   SUBROUTINE overlap_aabb(la_max_set1, la_min_set1, npgfa1, rpgfa1, zeta1, &
                           la_max_set2, la_min_set2, npgfa2, rpgfa2, zeta2, &
                           lb_max_set1, lb_min_set1, npgfb1, rpgfb1, zetb1, &
                           lb_max_set2, lb_min_set2, npgfb2, rpgfb2, zetb2, &
                           asets_equal, bsets_equal, rab, dab, saabb, s, lds)

      INTEGER, INTENT(IN)                                :: la_max_set1, la_min_set1, npgfa1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa1, zeta1
      INTEGER, INTENT(IN)                                :: la_max_set2, la_min_set2, npgfa2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa2, zeta2
      INTEGER, INTENT(IN)                                :: lb_max_set1, lb_min_set1, npgfb1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb1, zetb1
      INTEGER, INTENT(IN)                                :: lb_max_set2, lb_min_set2, npgfb2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb2, zetb2
      LOGICAL, INTENT(IN)                                :: asets_equal, bsets_equal
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: dab
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT)                                   :: saabb
      INTEGER, INTENT(IN)                                :: lds
      REAL(KIND=dp), DIMENSION(lds, lds), INTENT(INOUT)  :: s

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap_aabb', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coa, coamx, coamy, coamz, coapx, coapy, coapz, cob, &
         cobm2x, cobm2y, cobm2z, cobmx, cobmy, cobmz, handle, i, ia, ib, ipgf, j, ja, jb, jpgf, &
         jpgf_start, kpgf, la, la_max, la_min, la_start, lb, lb_max, lb_min, lpgf, lpgf_start, &
         ncoa1, ncoa2, ncob1, ncob2
      INTEGER, DIMENSION(3)                              :: na, naa, nb, nbb, nia, nib, nja, njb
      REAL(KIND=dp)                                      :: f0, f1, f2, f3, f4, fax, fay, faz, zeta, &
                                                            zetb, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

      CALL timeset(routineN, handle)

!   *** Loop over all pairs of primitive Gaussian-type functions ***

      ncoa1 = 0
      ncoa2 = 0
      ncob1 = 0
      ncob2 = 0

      DO ipgf = 1, npgfa1

         ncoa2 = 0

         IF (asets_equal) THEN
            jpgf_start = ipgf
            DO i = 1, jpgf_start-1
               ncoa2 = ncoa2+ncoset(la_max_set2)
            ENDDO
         ELSE
            jpgf_start = 1
         ENDIF

         DO jpgf = jpgf_start, npgfa2

            ncob1 = 0
            zeta = zeta1(ipgf)+zeta2(jpgf)
            la_max = la_max_set1+la_max_set2
            la_min = la_min_set1+la_min_set2

            DO kpgf = 1, npgfb1

               ncob2 = 0

               IF (bsets_equal) THEN
                  lpgf_start = kpgf
                  DO i = 1, lpgf_start-1
                     ncob2 = ncob2+ncoset(lb_max_set2)
                  ENDDO
               ELSE
                  lpgf_start = 1
               ENDIF

               DO lpgf = lpgf_start, npgfb2

                  ! *** Screening ***
                  IF ((rpgfa1(ipgf)+rpgfb1(kpgf) < dab) .OR. &
                      (rpgfa2(jpgf)+rpgfb1(kpgf) < dab) .OR. &
                      (rpgfa1(ipgf)+rpgfb2(lpgf) < dab) .OR. &
                      (rpgfa2(jpgf)+rpgfb2(lpgf) < dab)) THEN
                     DO jb = ncoset(lb_min_set2-1)+1, ncoset(lb_max_set2)
                        DO ib = ncoset(lb_min_set1-1)+1, ncoset(lb_max_set1)
                           DO ja = ncoset(la_min_set2-1)+1, ncoset(la_max_set2)
                              DO ia = ncoset(la_min_set1-1)+1, ncoset(la_max_set1)
                                 saabb(ncoa1+ia, ncoa2+ja, ncob1+ib, ncob2+jb) = 0._dp
                                 IF (asets_equal) saabb(ncoa2+ja, ncoa1+ia, ncob1+ib, ncob2+jb) = 0._dp
                                 IF (bsets_equal) saabb(ncoa1+ia, ncoa2+ja, ncob2+jb, ncob1+ib) = 0._dp
                                 IF (asets_equal .AND. bsets_equal) THEN
                                    saabb(ncoa2+ja, ncoa1+ia, ncob1+ib, ncob2+jb) = 0._dp
                                    saabb(ncoa1+ia, ncoa2+ja, ncob2+jb, ncob1+ib) = 0._dp
                                    saabb(ncoa2+ja, ncoa1+ia, ncob2+jb, ncob1+ib) = 0._dp
                                 ENDIF
                              ENDDO
                           ENDDO
                        ENDDO
                     ENDDO
                     ncob2 = ncob2+ncoset(lb_max_set2)
                     CYCLE
                  END IF

                  zetb = zetb1(kpgf)+zetb2(lpgf)
                  lb_max = lb_max_set1+lb_max_set2
                  lb_min = lb_min_set1+lb_min_set2

!           *** Calculate some prefactors ***

                  zetp = 1.0_dp/(zeta+zetb)

                  f0 = SQRT((pi*zetp)**3)
                  f1 = zetb*zetp
                  f2 = 0.5_dp*zetp

!           *** Calculate the basic two-center overlap integral [s|s] ***

                  s(1, 1) = f0*EXP(-zeta*f1*dab*dab)

!           *** Recurrence steps: [s|s] -> [a|b] ***

                  IF (la_max > 0) THEN

!             *** Vertical recurrence steps: [s|s] -> [a|s] ***

                     rap(:) = f1*rab(:)

!             *** [p|s] = (Pi - Ai)*[s|s]  (i = x,y,z) ***

                     s(2, 1) = rap(1)*s(1, 1) ! [px|s]
                     s(3, 1) = rap(2)*s(1, 1) ! [py|s]
                     s(4, 1) = rap(3)*s(1, 1) ! [pz|s]

                     IF (la_max > 1) THEN

!               *** [d|s] ***

                        f3 = f2*s(1, 1)

                        s(5, 1) = rap(1)*s(2, 1)+f3 ! [dx2|s]
                        s(6, 1) = rap(1)*s(3, 1) ! [dxy|s]
                        s(7, 1) = rap(1)*s(4, 1) ! [dxz|s]
                        s(8, 1) = rap(2)*s(3, 1)+f3 ! [dy2|s]
                        s(9, 1) = rap(2)*s(4, 1) ! [dyz|s]
                        s(10, 1) = rap(3)*s(4, 1)+f3 ! [dz2|s]

                        IF (la_max > 2) THEN

!                 *** [f|s] ***

                           f3 = 2.0_dp*f2

                           s(11, 1) = rap(1)*s(5, 1)+f3*s(2, 1) ! [fx3 |s]
                           s(12, 1) = rap(1)*s(6, 1)+f2*s(3, 1) ! [fx2y|s]
                           s(13, 1) = rap(1)*s(7, 1)+f2*s(4, 1) ! [fx2z|s]
                           s(14, 1) = rap(1)*s(8, 1) ! [fxy2|s]
                           s(15, 1) = rap(1)*s(9, 1) ! [fxyz|s]
                           s(16, 1) = rap(1)*s(10, 1) ! [fxz2|s]
                           s(17, 1) = rap(2)*s(8, 1)+f3*s(3, 1) ! [fy3 |s]
                           s(18, 1) = rap(2)*s(9, 1)+f2*s(4, 1) ! [fy2z|s]
                           s(19, 1) = rap(2)*s(10, 1) ! [fyz2|s]
                           s(20, 1) = rap(3)*s(10, 1)+f3*s(4, 1) ! [fz3 |s]

                           IF (la_max > 3) THEN

!                   *** [g|s] ***

                              f4 = 3.0_dp*f2

                              s(21, 1) = rap(1)*s(11, 1)+f4*s(5, 1) ! [gx4  |s]
                              s(22, 1) = rap(1)*s(12, 1)+f3*s(6, 1) ! [gx3y |s]
                              s(23, 1) = rap(1)*s(13, 1)+f3*s(7, 1) ! [gx3z |s]
                              s(24, 1) = rap(1)*s(14, 1)+f2*s(8, 1) ! [gx2y2|s]
                              s(25, 1) = rap(1)*s(15, 1)+f2*s(9, 1) ! [gx2yz|s]
                              s(26, 1) = rap(1)*s(16, 1)+f2*s(10, 1) ! [gx2z2|s]
                              s(27, 1) = rap(1)*s(17, 1) ! [gxy3 |s]
                              s(28, 1) = rap(1)*s(18, 1) ! [gxy2z|s]
                              s(29, 1) = rap(1)*s(19, 1) ! [gxyz2|s]
                              s(30, 1) = rap(1)*s(20, 1) ! [gxz3 |s]
                              s(31, 1) = rap(2)*s(17, 1)+f4*s(8, 1) ! [gy4  |s]
                              s(32, 1) = rap(2)*s(18, 1)+f3*s(9, 1) ! [gy3z |s]
                              s(33, 1) = rap(2)*s(19, 1)+f2*s(10, 1) ! [gy2z2|s]
                              s(34, 1) = rap(2)*s(20, 1) ! [gyz3 |s]
                              s(35, 1) = rap(3)*s(20, 1)+f4*s(10, 1) ! [gz4  |s]

!                   *** [a|s] = (Pi - Ai)*[a-1i|s] + f2*Ni(a-1i)*[a-2i|s] ***

                              DO la = 5, la_max

!                     *** Increase the angular momentum component z of a ***

                                 s(coset(0, 0, la), 1) = &
                                    rap(3)*s(coset(0, 0, la-1), 1)+ &
                                    f2*REAL(la-1, dp)*s(coset(0, 0, la-2), 1)

!                     *** Increase the angular momentum component y of a ***

                                 az = la-1
                                 s(coset(0, 1, az), 1) = rap(2)*s(coset(0, 0, az), 1)
                                 DO ay = 2, la
                                    az = la-ay
                                    s(coset(0, ay, az), 1) = &
                                       rap(2)*s(coset(0, ay-1, az), 1)+ &
                                       f2*REAL(ay-1, dp)*s(coset(0, ay-2, az), 1)
                                 END DO

!                     *** Increase the angular momentum component x of a ***

                                 DO ay = 0, la-1
                                    az = la-1-ay
                                    s(coset(1, ay, az), 1) = rap(1)*s(coset(0, ay, az), 1)
                                 END DO
                                 DO ax = 2, la
                                    f3 = f2*REAL(ax-1, dp)
                                    DO ay = 0, la-ax
                                       az = la-ax-ay
                                       s(coset(ax, ay, az), 1) = &
                                          rap(1)*s(coset(ax-1, ay, az), 1)+ &
                                          f3*s(coset(ax-2, ay, az), 1)
                                    END DO
                                 END DO

                              END DO

                           END IF

                        END IF

                     END IF

!             *** Recurrence steps: [a|s] -> [a|b] ***

                     IF (lb_max > 0) THEN

                        DO j = 2, ncoset(lb_max)
                           DO i = 1, ncoset(la_min)
                              s(i, j) = 0.0_dp
                           END DO
                        END DO

!               *** Horizontal recurrence steps ***

                        rbp(:) = rap(:)-rab(:)

!               *** [a|p] = [a+1i|s] - (Bi - Ai)*[a|s] ***

                        IF (lb_max == 1) THEN
                           la_start = la_min
                        ELSE
                           la_start = MAX(0, la_min-1)
                        END IF

                        DO la = la_start, la_max-1
                           DO ax = 0, la
                              DO ay = 0, la-ax
                                 az = la-ax-ay
                                 coa = coset(ax, ay, az)
                                 coapx = coset(ax+1, ay, az)
                                 coapy = coset(ax, ay+1, az)
                                 coapz = coset(ax, ay, az+1)
                                 s(coa, 2) = s(coapx, 1)-rab(1)*s(coa, 1)
                                 s(coa, 3) = s(coapy, 1)-rab(2)*s(coa, 1)
                                 s(coa, 4) = s(coapz, 1)-rab(3)*s(coa, 1)
                              END DO
                           END DO
                        END DO

!               *** Vertical recurrence step ***

!               *** [a|p] = (Pi - Bi)*[a|s] + f2*Ni(a)*[a-1i|s] ***

                        DO ax = 0, la_max
                           fax = f2*REAL(ax, dp)
                           DO ay = 0, la_max-ax
                              fay = f2*REAL(ay, dp)
                              az = la_max-ax-ay
                              faz = f2*REAL(az, dp)
                              coa = coset(ax, ay, az)
                              coamx = coset(ax-1, ay, az)
                              coamy = coset(ax, ay-1, az)
                              coamz = coset(ax, ay, az-1)
                              s(coa, 2) = rbp(1)*s(coa, 1)+fax*s(coamx, 1)
                              s(coa, 3) = rbp(2)*s(coa, 1)+fay*s(coamy, 1)
                              s(coa, 4) = rbp(3)*s(coa, 1)+faz*s(coamz, 1)
                           END DO
                        END DO

!               *** Recurrence steps: [a|p] -> [a|b] ***

                        DO lb = 2, lb_max

!                 *** Horizontal recurrence steps ***

!                 *** [a|b] = [a+1i|b-1i] - (Bi - Ai)*[a|b-1i] ***

                           IF (lb == lb_max) THEN
                              la_start = la_min
                           ELSE
                              la_start = MAX(0, la_min-1)
                           END IF

                           DO la = la_start, la_max-1
                              DO ax = 0, la
                                 DO ay = 0, la-ax
                                    az = la-ax-ay
                                    coa = coset(ax, ay, az)
                                    coapx = coset(ax+1, ay, az)
                                    coapy = coset(ax, ay+1, az)
                                    coapz = coset(ax, ay, az+1)

!                       *** Shift of angular momentum component z from a to b ***

                                    cob = coset(0, 0, lb)
                                    cobmz = coset(0, 0, lb-1)
                                    s(coa, cob) = s(coapz, cobmz)-rab(3)*s(coa, cobmz)

!                       *** Shift of angular momentum component y from a to b ***

                                    DO by = 1, lb
                                       bz = lb-by
                                       cob = coset(0, by, bz)
                                       cobmy = coset(0, by-1, bz)
                                       s(coa, cob) = s(coapy, cobmy)-rab(2)*s(coa, cobmy)
                                    END DO

!                       *** Shift of angular momentum component x from a to b ***

                                    DO bx = 1, lb
                                       DO by = 0, lb-bx
                                          bz = lb-bx-by
                                          cob = coset(bx, by, bz)
                                          cobmx = coset(bx-1, by, bz)
                                          s(coa, cob) = s(coapx, cobmx)-rab(1)*s(coa, cobmx)
                                       END DO
                                    END DO

                                 END DO
                              END DO
                           END DO

!                 *** Vertical recurrence step ***

!                 *** [a|b] = (Pi - Bi)*[a|b-1i] + f2*Ni(a)*[a-1i|b-1i] + ***
!                 ***         f2*Ni(b-1i)*[a|b-2i]                        ***

                           DO ax = 0, la_max
                              fax = f2*REAL(ax, dp)
                              DO ay = 0, la_max-ax
                                 fay = f2*REAL(ay, dp)
                                 az = la_max-ax-ay
                                 faz = f2*REAL(az, dp)
                                 coa = coset(ax, ay, az)
                                 coamx = coset(ax-1, ay, az)
                                 coamy = coset(ax, ay-1, az)
                                 coamz = coset(ax, ay, az-1)

!                     *** Increase the angular momentum component z of b ***

                                 f3 = f2*REAL(lb-1, dp)
                                 cob = coset(0, 0, lb)
                                 cobmz = coset(0, 0, lb-1)
                                 cobm2z = coset(0, 0, lb-2)
                                 s(coa, cob) = rbp(3)*s(coa, cobmz)+ &
                                               faz*s(coamz, cobmz)+ &
                                               f3*s(coa, cobm2z)

!                     *** Increase the angular momentum component y of b ***

                                 bz = lb-1
                                 cob = coset(0, 1, bz)
                                 cobmy = coset(0, 0, bz)
                                 s(coa, cob) = rbp(2)*s(coa, cobmy)+ &
                                               fay*s(coamy, cobmy)
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    cob = coset(0, by, bz)
                                    cobmy = coset(0, by-1, bz)
                                    cobm2y = coset(0, by-2, bz)
                                    s(coa, cob) = rbp(2)*s(coa, cobmy)+ &
                                                  fay*s(coamy, cobmy)+ &
                                                  f3*s(coa, cobm2y)
                                 END DO

!                     *** Increase the angular momentum component x of b ***

                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    cob = coset(1, by, bz)
                                    cobmx = coset(0, by, bz)
                                    s(coa, cob) = rbp(1)*s(coa, cobmx)+ &
                                                  fax*s(coamx, cobmx)
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       cob = coset(bx, by, bz)
                                       cobmx = coset(bx-1, by, bz)
                                       cobm2x = coset(bx-2, by, bz)
                                       s(coa, cob) = rbp(1)*s(coa, cobmx)+ &
                                                     fax*s(coamx, cobmx)+ &
                                                     f3*s(coa, cobm2x)
                                    END DO
                                 END DO

                              END DO
                           END DO

                        END DO

                     END IF

                  ELSE

                     IF (lb_max > 0) THEN

!               *** Vertical recurrence steps: [s|s] -> [s|b] ***

                        rbp(:) = (f1-1.0_dp)*rab(:)

!               *** [s|p] = (Pi - Bi)*[s|s] ***

                        s(1, 2) = rbp(1)*s(1, 1) ! [s|px]
                        s(1, 3) = rbp(2)*s(1, 1) ! [s|py]
                        s(1, 4) = rbp(3)*s(1, 1) ! [s|pz]

                        IF (lb_max > 1) THEN

!                 *** [s|d] ***

                           f3 = f2*s(1, 1)

                           s(1, 5) = rbp(1)*s(1, 2)+f3 ! [s|dx2]
                           s(1, 6) = rbp(1)*s(1, 3) ! [s|dxy]
                           s(1, 7) = rbp(1)*s(1, 4) ! [s|dxz]
                           s(1, 8) = rbp(2)*s(1, 3)+f3 ! [s|dy2]
                           s(1, 9) = rbp(2)*s(1, 4) ! [s|dyz]
                           s(1, 10) = rbp(3)*s(1, 4)+f3 ! [s|dz2]

!                 *** [s|b] = (Pi - Bi)*[s|b-1i] + f2*Ni(b-1i)*[s|b-2i] ***

                           DO lb = 3, lb_max

!                   *** Increase the angular momentum component z of b ***

                              s(1, coset(0, 0, lb)) = &
                                 rbp(3)*s(1, coset(0, 0, lb-1))+ &
                                 f2*REAL(lb-1, dp)*s(1, coset(0, 0, lb-2))

!                   *** Increase the angular momentum component y of b ***

                              bz = lb-1
                              s(1, coset(0, 1, bz)) = rbp(2)*s(1, coset(0, 0, bz))
                              DO by = 2, lb
                                 bz = lb-by
                                 s(1, coset(0, by, bz)) = &
                                    rbp(2)*s(1, coset(0, by-1, bz))+ &
                                    f2*REAL(by-1, dp)*s(1, coset(0, by-2, bz))
                              END DO

!                   *** Increase the angular momentum component x of b ***

                              DO by = 0, lb-1
                                 bz = lb-1-by
                                 s(1, coset(1, by, bz)) = rbp(1)*s(1, coset(0, by, bz))
                              END DO
                              DO bx = 2, lb
                                 f3 = f2*REAL(bx-1, dp)
                                 DO by = 0, lb-bx
                                    bz = lb-bx-by
                                    s(1, coset(bx, by, bz)) = &
                                       rbp(1)*s(1, coset(bx-1, by, bz))+ &
                                       f3*s(1, coset(bx-2, by, bz))
                                 END DO
                              END DO

                           END DO

                        END IF

                     END IF

                  END IF

!           *** Store the primitive overlap integrals ***
                  DO jb = ncoset(lb_min_set2-1)+1, ncoset(lb_max_set2)
                     njb(1:3) = indco(1:3, jb)
                     DO ib = ncoset(lb_min_set1-1)+1, ncoset(lb_max_set1)
                        nib(1:3) = indco(1:3, ib)
                        nbb(1:3) = nib+njb
                        DO ja = ncoset(la_min_set2-1)+1, ncoset(la_max_set2)
                           nja(1:3) = indco(1:3, ja)
                           DO ia = ncoset(la_min_set1-1)+1, ncoset(la_max_set1)
                              nia(1:3) = indco(1:3, ia)
                              naa(1:3) = nia+nja
                              ! now loop over all elements of s
                              DO j = ncoset(lb_min-1)+1, ncoset(lb_max)
                                 nb(1:3) = indco(1:3, j)
                                 DO i = ncoset(la_min-1)+1, ncoset(la_max)
                                    na(1:3) = indco(1:3, i)
                                    IF (ALL(na == naa) .AND. ALL(nb == nbb)) THEN
                                       saabb(ncoa1+ia, ncoa2+ja, ncob1+ib, ncob2+jb) = s(i, j)
                                       IF (asets_equal) saabb(ncoa2+ja, ncoa1+ia, ncob1+ib, ncob2+jb) = s(i, j)
                                       IF (bsets_equal) saabb(ncoa1+ia, ncoa2+ja, ncob2+jb, ncob1+ib) = s(i, j)
                                       IF (asets_equal .AND. bsets_equal) THEN
                                          saabb(ncoa2+ja, ncoa1+ia, ncob1+ib, ncob2+jb) = s(i, j)
                                          saabb(ncoa1+ia, ncoa2+ja, ncob2+jb, ncob1+ib) = s(i, j)
                                          saabb(ncoa2+ja, ncoa1+ia, ncob2+jb, ncob1+ib) = s(i, j)
                                       ENDIF
                                    ENDIF
                                 END DO
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO

                  ncob2 = ncob2+ncoset(lb_max_set2)

               END DO

               ncob1 = ncob1+ncoset(lb_max_set1)

            END DO

            ncoa2 = ncoa2+ncoset(la_max_set2)

         END DO

         ncoa1 = ncoa1+ncoset(la_max_set1)

      END DO

      CALL timestop(handle)

   END SUBROUTINE overlap_aabb

END MODULE ai_overlap_aabb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
!!****** cp2k/ai_overlap3 [1.0] *
!!
!!   NAME
!!     ai_overlap3
!!
!!   FUNCTION
!!     Calculation of three-center overlap integrals over Cartesian
!!     Gaussian-type functions.
!!
!!   AUTHOR
!!     Matthias Krack (26.06.2001)
!!
!!   LITERATURE
!!     S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!!
!******************************************************************************

MODULE ai_overlap3

! **************************************************************************************************

! ax,ay,az   : Angular momentum index numbers of orbital a.
! bx,by,bz   : Angular momentum index numbers of orbital b.
! coset      : Cartesian orbital set pointer.
! dab        : Distance between the atomic centers a and b.
! dac        : Distance between the atomic centers a and c.
! dbc        : Distance between the atomic centers b and c.
! l{a,b,c}   : Angular momentum quantum number of shell a, b or c.
! l{a,b}_max : Maximum angular momentum quantum number of shell a, b or c.
! ncoset     : Number of Cartesian orbitals up to l.
! rab        : Distance vector between the atomic centers a and b.
! rac        : Distance vector between the atomic centers a and c.
! rbc        : Distance vector between the atomic centers b and c.
! rpgf{a,b,c}: Radius of the primitive Gaussian-type function a or b.
! zet{a,b,c} : Exponents of the Gaussian-type functions a or b.
! zetg       : Reciprocal of the sum of the exponents of orbital a, b and c.
! zetp       : Reciprocal of the sum of the exponents of orbital a and b.

! **************************************************************************************************

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap3'

! *** Public subroutines ***

   PUBLIC :: overlap3

!!***
! **************************************************************************************************

CONTAINS

! ***************************************************************************************************
!> \brief Calculation of three-center overlap integrals [a|b|c] over primitive
!>        Cartesian Gaussian functions
!> \param la_max_set ...
!> \param npgfa ...
!> \param zeta ...
!> \param rpgfa ...
!> \param la_min_set ...
!> \param lb_max_set ...
!> \param npgfb ...
!> \param zetb ...
!> \param rpgfb ...
!> \param lb_min_set ...
!> \param lc_max_set ...
!> \param npgfc ...
!> \param zetc ...
!> \param rpgfc ...
!> \param lc_min_set ...
!> \param rab ...
!> \param dab ...
!> \param rac ...
!> \param dac ...
!> \param rbc ...
!> \param dbc ...
!> \param sabc integrals [a|b|c]
!> \param sdabc derivative [da/dAi|b|c]
!> \param sabdc derivative [a|b|dc/dCi]
!> \par History
!>      05.2014 created (Dorothea Golze)
!> \author Dorothea Golze
!> \note  overlap3 essentially uses the setup of overlap3_old
! **************************************************************************************************

   SUBROUTINE overlap3(la_max_set, npgfa, zeta, rpgfa, la_min_set, &
                       lb_max_set, npgfb, zetb, rpgfb, lb_min_set, &
                       lc_max_set, npgfc, zetc, rpgfc, lc_min_set, &
                       rab, dab, rac, dac, rbc, dbc, sabc, sdabc, sabdc)

      INTEGER, INTENT(IN)                                :: la_max_set, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta, rpgfa
      INTEGER, INTENT(IN)                                :: la_min_set, lb_max_set, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb, rpgfb
      INTEGER, INTENT(IN)                                :: lb_min_set, lc_max_set, npgfc
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetc, rpgfc
      INTEGER, INTENT(IN)                                :: lc_min_set
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: dab
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: dac
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rbc
      REAL(KIND=dp), INTENT(IN)                          :: dbc
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: sabc
      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT), OPTIONAL                         :: sdabc, sabdc

      CHARACTER(len=*), PARAMETER :: routineN = 'overlap3', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coa, coax, coay, coaz, coc, cocx, cocy, cocz, cx, cy, cz, &
         handle, i, ipgf, j, jpgf, k, kpgf, l, la, la_max, la_min, la_start, lai, lb, lb_max, &
         lb_min, lbi, lc, lc_max, lc_min, lci, na, nb, nc, nda, ndc
      REAL(KIND=dp)                                      :: f0, f1, f2, f3, fcx, fcy, fcz, fx, fy, &
                                                            fz, rcp2, zetg, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: rag, rbg, rcg, rcp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: s
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: sda, sdc

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (s, sda, sdc)

      lai = 0
      lbi = 0
      lci = 0

      IF (PRESENT(sdabc)) lai = 1
      IF (PRESENT(sabdc)) lci = 1

      la_max = la_max_set+lai
      la_min = MAX(0, la_min_set-lai)
      lb_max = lb_max_set
      lb_min = lb_min_set
      lc_max = lc_max_set+lci
      lc_min = MAX(0, lc_min_set-lci)

      ALLOCATE (s(ncoset(la_max), ncoset(lb_max), ncoset(lc_max)))
      s = 0._dp
      IF (PRESENT(sdabc)) THEN
         ALLOCATE (sda(ncoset(la_max), ncoset(lb_max), ncoset(lc_max), 3))
         sda = 0._dp
      ENDIF
      IF (PRESENT(sabdc)) THEN
         ALLOCATE (sdc(ncoset(la_max), ncoset(lb_max), ncoset(lc_max), 3))
         sdc = 0._dp
      ENDIF

!   *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nda = 0
      DO ipgf = 1, npgfa

         nb = 0
         DO jpgf = 1, npgfb

            ! *** Screening ***
            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               nb = nb+ncoset(lb_max_set)
               CYCLE
            END IF

            nc = 0
            ndc = 0
            DO kpgf = 1, npgfc

               ! *** Screening ***
               IF ((rpgfb(jpgf)+rpgfc(kpgf) < dbc) .OR. &
                   (rpgfa(ipgf)+rpgfc(kpgf) < dac)) THEN
                  nc = nc+ncoset(lc_max_set)
                  ndc = ndc+ncoset(lc_max_set)
                  CYCLE
               END IF

               ! *** Calculate some prefactors ***
               zetg = 1.0_dp/(zeta(ipgf)+zetb(jpgf)+zetc(kpgf))
               zetp = 1.0_dp/(zeta(ipgf)+zetb(jpgf))
               f0 = (pi*zetg)**1.5_dp
               f1 = zetb(jpgf)*zetp
               f2 = 0.5_dp*zetg
               rcp(:) = f1*rab(:)-rac(:)
               rcp2 = rcp(1)*rcp(1)+rcp(2)*rcp(2)+rcp(3)*rcp(3)

               ! *** Calculate the basic three-center overlap integral [s|s|s] ***
               s(1, 1, 1) = f0*EXP(-(zeta(ipgf)*f1*dab*dab+zetc(kpgf)*zetg*rcp2/zetp))

!         *** Recurrence steps: [s|s|s] -> [a|s|s] ***

               IF (la_max > 0) THEN

!           *** Vertical recurrence steps: [s|s|s] -> [a|s|s] ***

                  rag(:) = zetg*(zetb(jpgf)*rab(:)+zetc(kpgf)*rac(:))

!           *** [p|s|s] = (Gi - Ai)*[s|s|s]  (i = x,y,z) ***

                  s(2, 1, 1) = rag(1)*s(1, 1, 1)
                  s(3, 1, 1) = rag(2)*s(1, 1, 1)
                  s(4, 1, 1) = rag(3)*s(1, 1, 1)

!           *** [a|s|s] = (Gi - Ai)*[a-1i|s|s] + f2*Ni(a-1i)*[a-2i|s|s] ***

                  DO la = 2, la_max

!             *** Increase the angular momentum component z of function a ***

                     s(coset(0, 0, la), 1, 1) = rag(3)*s(coset(0, 0, la-1), 1, 1)+ &
                                                f2*REAL(la-1, dp)*s(coset(0, 0, la-2), 1, 1)

!             *** Increase the angular momentum component y of function a ***

                     az = la-1
                     s(coset(0, 1, az), 1, 1) = rag(2)*s(coset(0, 0, az), 1, 1)

                     DO ay = 2, la
                        az = la-ay
                        s(coset(0, ay, az), 1, 1) = rag(2)*s(coset(0, ay-1, az), 1, 1)+ &
                                                    f2*REAL(ay-1, dp)*s(coset(0, ay-2, az), 1, 1)
                     END DO

!             *** Increase the angular momentum component x of function a ***

                     DO ay = 0, la-1
                        az = la-1-ay
                        s(coset(1, ay, az), 1, 1) = rag(1)*s(coset(0, ay, az), 1, 1)
                     END DO

                     DO ax = 2, la
                        f3 = f2*REAL(ax-1, dp)
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           s(coset(ax, ay, az), 1, 1) = rag(1)*s(coset(ax-1, ay, az), 1, 1)+ &
                                                        f3*s(coset(ax-2, ay, az), 1, 1)
                        END DO
                     END DO

                  END DO

!           *** Recurrence steps: [a|s|s] -> [a|s|b] ***

                  IF (lb_max > 0) THEN

!             *** Horizontal recurrence steps ***

                     rbg(:) = rag(:)-rab(:)

!             *** [a|s|p] = [a+1i|s|s] - (Bi - Ai)*[a|s|s] ***

                     IF (lb_max == 1) THEN
                        la_start = la_min
                     ELSE
                        la_start = MAX(0, la_min-1)
                     END IF

                     DO la = la_start, la_max-1
                        DO ax = 0, la
                           DO ay = 0, la-ax
                              az = la-ax-ay
                              coa = coset(ax, ay, az)
                              coax = coset(ax+1, ay, az)
                              coay = coset(ax, ay+1, az)
                              coaz = coset(ax, ay, az+1)
                              s(coset(ax, ay, az), 2, 1) = s(coax, 1, 1)-rab(1)*s(coa, 1, 1)
                              s(coset(ax, ay, az), 3, 1) = s(coay, 1, 1)-rab(2)*s(coa, 1, 1)
                              s(coset(ax, ay, az), 4, 1) = s(coaz, 1, 1)-rab(3)*s(coa, 1, 1)
                           END DO
                        END DO
                     END DO

!             *** Vertical recurrence step ***

!             *** [a|s|p] = (Gi - Bi)*[a|s|s] + f2*Ni(a)*[a-1i|s|s] ***

                     DO ax = 0, la_max
                        fx = f2*REAL(ax, dp)
                        DO ay = 0, la_max-ax
                           fy = f2*REAL(ay, dp)
                           az = la_max-ax-ay
                           fz = f2*REAL(az, dp)
                           coa = coset(ax, ay, az)
                           IF (ax == 0) THEN
                              s(coa, 2, 1) = rbg(1)*s(coa, 1, 1)
                           ELSE
                              s(coa, 2, 1) = rbg(1)*s(coa, 1, 1)+fx*s(coset(ax-1, ay, az), 1, 1)
                           END IF
                           IF (ay == 0) THEN
                              s(coa, 3, 1) = rbg(2)*s(coa, 1, 1)
                           ELSE
                              s(coa, 3, 1) = rbg(2)*s(coa, 1, 1)+fy*s(coset(ax, ay-1, az), 1, 1)
                           END IF
                           IF (az == 0) THEN
                              s(coa, 4, 1) = rbg(3)*s(coa, 1, 1)
                           ELSE
                              s(coa, 4, 1) = rbg(3)*s(coa, 1, 1)+fz*s(coset(ax, ay, az-1), 1, 1)
                           END IF
                        END DO
                     END DO

!             *** Recurrence steps: [a|s|p] -> [a|s|b] ***

                     DO lb = 2, lb_max

!               *** Horizontal recurrence steps ***

!               *** [a|s|b] = [a+1i|s|b-1i] - (Bi - Ai)*[a|s|b-1i] ***

                        IF (lb == lb_max) THEN
                           la_start = la_min
                        ELSE
                           la_start = MAX(0, la_min-1)
                        END IF

                        DO la = la_start, la_max-1
                           DO ax = 0, la
                              DO ay = 0, la-ax
                                 az = la-ax-ay

                                 coa = coset(ax, ay, az)
                                 coax = coset(ax+1, ay, az)
                                 coay = coset(ax, ay+1, az)
                                 coaz = coset(ax, ay, az+1)

!                     *** Shift of angular momentum component z from a to b ***

                                 s(coa, coset(0, 0, lb), 1) = &
                                    s(coaz, coset(0, 0, lb-1), 1)- &
                                    rab(3)*s(coa, coset(0, 0, lb-1), 1)

!                     *** Shift of angular momentum component y from a to b ***

                                 DO by = 1, lb
                                    bz = lb-by
                                    s(coa, coset(0, by, bz), 1) = &
                                       s(coay, coset(0, by-1, bz), 1)- &
                                       rab(2)*s(coa, coset(0, by-1, bz), 1)
                                 END DO

!                     *** Shift of angular momentum component x from a to b ***

                                 DO bx = 1, lb
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       s(coa, coset(bx, by, bz), 1) = &
                                          s(coax, coset(bx-1, by, bz), 1)- &
                                          rab(1)*s(coa, coset(bx-1, by, bz), 1)
                                    END DO
                                 END DO

                              END DO
                           END DO
                        END DO

!               *** Vertical recurrence step ***

!               *** [a|s|b] = (Gi - Bi)*[a|s|b-1i] +   ***
!               ***           f2*Ni(a)*[a-1i|s|b-1i] + ***
!               ***           f2*Ni(b-1i)*[a|s|b-2i]   ***

                        DO ax = 0, la_max
                           fx = f2*REAL(ax, dp)
                           DO ay = 0, la_max-ax
                              fy = f2*REAL(ay, dp)
                              az = la_max-ax-ay
                              fz = f2*REAL(az, dp)

                              coa = coset(ax, ay, az)

                              f3 = f2*REAL(lb-1, dp)

!                   *** Shift of angular momentum component z from a to b ***

                              IF (az == 0) THEN
                                 s(coa, coset(0, 0, lb), 1) = &
                                    rbg(3)*s(coa, coset(0, 0, lb-1), 1)+ &
                                    f3*s(coa, coset(0, 0, lb-2), 1)
                              ELSE
                                 coaz = coset(ax, ay, az-1)
                                 s(coa, coset(0, 0, lb), 1) = &
                                    rbg(3)*s(coa, coset(0, 0, lb-1), 1)+ &
                                    fz*s(coaz, coset(0, 0, lb-1), 1)+ &
                                    f3*s(coa, coset(0, 0, lb-2), 1)
                              END IF

!                   *** Shift of angular momentum component y from a to b ***

                              IF (ay == 0) THEN
                                 bz = lb-1
                                 s(coa, coset(0, 1, bz), 1) = &
                                    rbg(2)*s(coa, coset(0, 0, bz), 1)
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    s(coa, coset(0, by, bz), 1) = &
                                       rbg(2)*s(coa, coset(0, by-1, bz), 1)+ &
                                       f3*s(coa, coset(0, by-2, bz), 1)
                                 END DO
                              ELSE
                                 coay = coset(ax, ay-1, az)
                                 bz = lb-1
                                 s(coa, coset(0, 1, bz), 1) = &
                                    rbg(2)*s(coa, coset(0, 0, bz), 1)+ &
                                    fy*s(coay, coset(0, 0, bz), 1)
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    s(coa, coset(0, by, bz), 1) = &
                                       rbg(2)*s(coa, coset(0, by-1, bz), 1)+ &
                                       fy*s(coay, coset(0, by-1, bz), 1)+ &
                                       f3*s(coa, coset(0, by-2, bz), 1)
                                 END DO
                              END IF

!                   *** Shift of angular momentum component x from a to b ***

                              IF (ax == 0) THEN
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    s(coa, coset(1, by, bz), 1) = &
                                       rbg(1)*s(coa, coset(0, by, bz), 1)
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       s(coa, coset(bx, by, bz), 1) = &
                                          rbg(1)*s(coa, coset(bx-1, by, bz), 1)+ &
                                          f3*s(coa, coset(bx-2, by, bz), 1)
                                    END DO
                                 END DO
                              ELSE
                                 coax = coset(ax-1, ay, az)
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    s(coa, coset(1, by, bz), 1) = &
                                       rbg(1)*s(coa, coset(0, by, bz), 1)+ &
                                       fx*s(coax, coset(0, by, bz), 1)
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       s(coa, coset(bx, by, bz), 1) = &
                                          rbg(1)*s(coa, coset(bx-1, by, bz), 1)+ &
                                          fx*s(coax, coset(bx-1, by, bz), 1)+ &
                                          f3*s(coa, coset(bx-2, by, bz), 1)
                                    END DO
                                 END DO
                              END IF

                           END DO
                        END DO

                     END DO

                  END IF

               ELSE

                  IF (lb_max > 0) THEN

!             *** Vertical recurrence steps: [s|s|s] -> [s|s|b] ***

                     rbg(:) = -zetg*(zeta(ipgf)*rab(:)-zetc(kpgf)*rbc(:))

!             *** [s|s|p] = (Gi - Bi)*[s|s|s] ***

                     s(1, 2, 1) = rbg(1)*s(1, 1, 1)
                     s(1, 3, 1) = rbg(2)*s(1, 1, 1)
                     s(1, 4, 1) = rbg(3)*s(1, 1, 1)

!             *** [s|s|b] = (Gi - Bi)*[s|s|b-1i] + f2*Ni(b-1i)*[s|s|b-2i] ***

                     DO lb = 2, lb_max

!               *** Increase the angular momentum component z of function b ***

                        s(1, coset(0, 0, lb), 1) = rbg(3)*s(1, coset(0, 0, lb-1), 1)+ &
                                                   f2*REAL(lb-1, dp)*s(1, coset(0, 0, lb-2), 1)

!               *** Increase the angular momentum component y of function b ***

                        bz = lb-1
                        s(1, coset(0, 1, bz), 1) = rbg(2)*s(1, coset(0, 0, bz), 1)

                        DO by = 2, lb
                           bz = lb-by
                           s(1, coset(0, by, bz), 1) = &
                              rbg(2)*s(1, coset(0, by-1, bz), 1)+ &
                              f2*REAL(by-1, dp)*s(1, coset(0, by-2, bz), 1)
                        END DO

!               *** Increase the angular momentum component x of function b ***

                        DO by = 0, lb-1
                           bz = lb-1-by
                           s(1, coset(1, by, bz), 1) = rbg(1)*s(1, coset(0, by, bz), 1)
                        END DO

                        DO bx = 2, lb
                           f3 = f2*REAL(bx-1, dp)
                           DO by = 0, lb-bx
                              bz = lb-bx-by
                              s(1, coset(bx, by, bz), 1) = rbg(1)*s(1, coset(bx-1, by, bz), 1)+ &
                                                           f3*s(1, coset(bx-2, by, bz), 1)
                           END DO
                        END DO

                     END DO

                  END IF

               END IF

!         *** Recurrence steps: [a|s|b] -> [a|c|b] ***

               IF (lc_max > 0) THEN

!           *** Vertical recurrence steps: [s|s|s] -> [s|c|s] ***

                  rcg(:) = -zetg*(zeta(ipgf)*rac(:)+zetb(jpgf)*rbc(:))

!           *** [s|p|s] = (Gi - Ci)*[s|s|s]  (i = x,y,z) ***

                  s(1, 1, 2) = rcg(1)*s(1, 1, 1)
                  s(1, 1, 3) = rcg(2)*s(1, 1, 1)
                  s(1, 1, 4) = rcg(3)*s(1, 1, 1)

!           *** [s|c|s] = (Gi - Ci)*[s|c-1i|s] + f2*Ni(c-1i)*[s|c-2i|s] ***

                  DO lc = 2, lc_max

!             *** Increase the angular momentum component z of function c ***

                     s(1, 1, coset(0, 0, lc)) = rcg(3)*s(1, 1, coset(0, 0, lc-1))+ &
                                                f2*REAL(lc-1, dp)*s(1, 1, coset(0, 0, lc-2))

!             *** Increase the angular momentum component y of function c ***

                     cz = lc-1
                     s(1, 1, coset(0, 1, cz)) = rcg(2)*s(1, 1, coset(0, 0, cz))

                     DO cy = 2, lc
                        cz = lc-cy
                        s(1, 1, coset(0, cy, cz)) = rcg(2)*s(1, 1, coset(0, cy-1, cz))+ &
                                                    f2*REAL(cy-1, dp)*s(1, 1, coset(0, cy-2, cz))
                     END DO

!             *** Increase the angular momentum component x of function c ***

                     DO cy = 0, lc-1
                        cz = lc-1-cy
                        s(1, 1, coset(1, cy, cz)) = rcg(1)*s(1, 1, coset(0, cy, cz))
                     END DO

                     DO cx = 2, lc
                        f3 = f2*REAL(cx-1, dp)
                        DO cy = 0, lc-cx
                           cz = lc-cx-cy
                           s(1, 1, coset(cx, cy, cz)) = rcg(1)*s(1, 1, coset(cx-1, cy, cz))+ &
                                                        f3*s(1, 1, coset(cx-2, cy, cz))
                        END DO
                     END DO

                  END DO

!           *** Recurrence steps: [s|c|s] -> [a|c|b] ***

                  DO lc = 1, lc_max

                     DO cx = 0, lc
                        DO cy = 0, lc-cx
                           cz = lc-cx-cy

                           coc = coset(cx, cy, cz)
                           cocx = coset(MAX(0, cx-1), cy, cz)
                           cocy = coset(cx, MAX(0, cy-1), cz)
                           cocz = coset(cx, cy, MAX(0, cz-1))

                           fcx = f2*REAL(cx, dp)
                           fcy = f2*REAL(cy, dp)
                           fcz = f2*REAL(cz, dp)

!                 *** Recurrence steps: [s|c|s] -> [a|c|s] ***

                           IF (la_max > 0) THEN

!                   *** Vertical recurrence steps: [s|c|s] -> [a|c|s] ***

                              rag(:) = rcg(:)+rac(:)

!                   *** [p|c|s] = (Gi - Ai)*[s|c|s] + f2*Ni(c)*[s|c-1i|s] ***

                              s(2, 1, coc) = rag(1)*s(1, 1, coc)+fcx*s(1, 1, cocx)
                              s(3, 1, coc) = rag(2)*s(1, 1, coc)+fcy*s(1, 1, cocy)
                              s(4, 1, coc) = rag(3)*s(1, 1, coc)+fcz*s(1, 1, cocz)

!                   *** [a|c|s] = (Gi - Ai)*[a-1i|c|s] +   ***
!                   ***           f2*Ni(a-1i)*[a-2i|c|s] + ***
!                   ***           f2*Ni(c)*[a-1i|c-1i|s]   ***

                              DO la = 2, la_max

!                     *** Increase the angular momentum component z of a ***

                                 s(coset(0, 0, la), 1, coc) = &
                                    rag(3)*s(coset(0, 0, la-1), 1, coc)+ &
                                    f2*REAL(la-1, dp)*s(coset(0, 0, la-2), 1, coc)+ &
                                    fcz*s(coset(0, 0, la-1), 1, cocz)

!                     *** Increase the angular momentum component y of a ***

                                 az = la-1
                                 s(coset(0, 1, az), 1, coc) = &
                                    rag(2)*s(coset(0, 0, az), 1, coc)+ &
                                    fcy*s(coset(0, 0, az), 1, cocy)

                                 DO ay = 2, la
                                    az = la-ay
                                    s(coset(0, ay, az), 1, coc) = &
                                       rag(2)*s(coset(0, ay-1, az), 1, coc)+ &
                                       f2*REAL(ay-1, dp)*s(coset(0, ay-2, az), 1, coc)+ &
                                       fcy*s(coset(0, ay-1, az), 1, cocy)
                                 END DO

!                     *** Increase the angular momentum component x of a ***

                                 DO ay = 0, la-1
                                    az = la-1-ay
                                    s(coset(1, ay, az), 1, coc) = &
                                       rag(1)*s(coset(0, ay, az), 1, coc)+ &
                                       fcx*s(coset(0, ay, az), 1, cocx)
                                 END DO

                                 DO ax = 2, la
                                    f3 = f2*REAL(ax-1, dp)
                                    DO ay = 0, la-ax
                                       az = la-ax-ay
                                       s(coset(ax, ay, az), 1, coc) = &
                                          rag(1)*s(coset(ax-1, ay, az), 1, coc)+ &
                                          f3*s(coset(ax-2, ay, az), 1, coc)+ &
                                          fcx*s(coset(ax-1, ay, az), 1, cocx)
                                    END DO
                                 END DO

                              END DO

!                   *** Recurrence steps: [a|c|s] -> [a|c|b] ***

                              IF (lb_max > 0) THEN

!                     *** Horizontal recurrence steps ***

                                 rbg(:) = rag(:)-rab(:)

!                     *** [a|c|p] = [a+1i|c|s] - (Bi - Ai)*[a|c|s] ***

                                 IF (lb_max == 1) THEN
                                    la_start = la_min
                                 ELSE
                                    la_start = MAX(0, la_min-1)
                                 END IF

                                 DO la = la_start, la_max-1
                                    DO ax = 0, la
                                       DO ay = 0, la-ax
                                          az = la-ax-ay
                                          coa = coset(ax, ay, az)
                                          coax = coset(ax+1, ay, az)
                                          coay = coset(ax, ay+1, az)
                                          coaz = coset(ax, ay, az+1)
                                          s(coa, 2, coc) = s(coax, 1, coc)-rab(1)*s(coa, 1, coc)
                                          s(coa, 3, coc) = s(coay, 1, coc)-rab(2)*s(coa, 1, coc)
                                          s(coa, 4, coc) = s(coaz, 1, coc)-rab(3)*s(coa, 1, coc)
                                       END DO
                                    END DO
                                 END DO

!                     *** Vertical recurrence step ***

!                     *** [a|c|p] = (Gi - Bi)*[a|c|s] +   ***
!                                   f2*Ni(a)*[a-1i|c|s] + ***
!                                   f2*Ni(c)*[a|c-1i|s]   ***

                                 DO ax = 0, la_max
                                    fx = f2*REAL(ax, dp)
                                    DO ay = 0, la_max-ax
                                       fy = f2*REAL(ay, dp)
                                       az = la_max-ax-ay
                                       fz = f2*REAL(az, dp)
                                       coa = coset(ax, ay, az)
                                       IF (ax == 0) THEN
                                          s(coa, 2, coc) = rbg(1)*s(coa, 1, coc)+ &
                                                           fcx*s(coa, 1, cocx)
                                       ELSE
                                          s(coa, 2, coc) = rbg(1)*s(coa, 1, coc)+ &
                                                           fx*s(coset(ax-1, ay, az), 1, coc)+ &
                                                           fcx*s(coa, 1, cocx)
                                       END IF
                                       IF (ay == 0) THEN
                                          s(coa, 3, coc) = rbg(2)*s(coa, 1, coc)+ &
                                                           fcy*s(coa, 1, cocy)
                                       ELSE
                                          s(coa, 3, coc) = rbg(2)*s(coa, 1, coc)+ &
                                                           fy*s(coset(ax, ay-1, az), 1, coc)+ &
                                                           fcy*s(coa, 1, cocy)
                                       END IF
                                       IF (az == 0) THEN
                                          s(coa, 4, coc) = rbg(3)*s(coa, 1, coc)+ &
                                                           fcz*s(coa, 1, cocz)
                                       ELSE
                                          s(coa, 4, coc) = rbg(3)*s(coa, 1, coc)+ &
                                                           fz*s(coset(ax, ay, az-1), 1, coc)+ &
                                                           fcz*s(coa, 1, cocz)
                                       END IF
                                    END DO
                                 END DO

!                     *** Recurrence steps: [a|c|p] -> [a|c|b] ***

                                 DO lb = 2, lb_max

!                       *** Horizontal recurrence steps ***

!                       *** [a|c|b] = [a+1i|c|b-1i] - (Bi - Ai)*[a|c|b-1i] ***

                                    IF (lb == lb_max) THEN
                                       la_start = la_min
                                    ELSE
                                       la_start = MAX(0, la_min-1)
                                    END IF

                                    DO la = la_start, la_max-1
                                       DO ax = 0, la
                                          DO ay = 0, la-ax
                                             az = la-ax-ay

                                             coa = coset(ax, ay, az)
                                             coax = coset(ax+1, ay, az)
                                             coay = coset(ax, ay+1, az)
                                             coaz = coset(ax, ay, az+1)

!                             *** Shift of angular momentum ***
!                             *** component z from a to b   ***

                                             s(coa, coset(0, 0, lb), coc) = &
                                                s(coaz, coset(0, 0, lb-1), coc)- &
                                                rab(3)*s(coa, coset(0, 0, lb-1), coc)

!                             *** Shift of angular momentum ***
!                             *** component y from a to b   ***

                                             DO by = 1, lb
                                                bz = lb-by
                                                s(coa, coset(0, by, bz), coc) = &
                                                   s(coay, coset(0, by-1, bz), coc)- &
                                                   rab(2)*s(coa, coset(0, by-1, bz), coc)
                                             END DO

!                             *** Shift of angular momentum ***
!                             *** component x from a to b   ***

                                             DO bx = 1, lb
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   s(coa, coset(bx, by, bz), coc) = &
                                                      s(coax, coset(bx-1, by, bz), coc)- &
                                                      rab(1)*s(coa, coset(bx-1, by, bz), coc)
                                                END DO
                                             END DO

                                          END DO
                                       END DO
                                    END DO

!                       *** Vertical recurrence step ***

!                       *** [a|c|b] = (Gi - Bi)*[a|c|b-1i] +   ***
!                       ***           f2*Ni(a)*[a-1i|c|b-1i] + ***
!                       ***           f2*Ni(b-1i)*[a|c|b-2i] + ***
!                       ***           f2*Ni(c)*[a|c-1i|b-1i]   ***

                                    DO ax = 0, la_max
                                       fx = f2*REAL(ax, dp)
                                       DO ay = 0, la_max-ax
                                          fy = f2*REAL(ay, dp)
                                          az = la_max-ax-ay
                                          fz = f2*REAL(az, dp)

                                          coa = coset(ax, ay, az)
                                          coax = coset(MAX(0, ax-1), ay, az)
                                          coay = coset(ax, MAX(0, ay-1), az)
                                          coaz = coset(ax, ay, MAX(0, az-1))

                                          f3 = f2*REAL(lb-1, dp)

!                           *** Shift of angular momentum ***
!                           *** component z from a to b   ***

                                          IF (az == 0) THEN
                                             s(coa, coset(0, 0, lb), coc) = &
                                                rbg(3)*s(coa, coset(0, 0, lb-1), coc)+ &
                                                f3*s(coa, coset(0, 0, lb-2), coc)+ &
                                                fcz*s(coa, coset(0, 0, lb-1), cocz)
                                          ELSE
                                             s(coa, coset(0, 0, lb), coc) = &
                                                rbg(3)*s(coa, coset(0, 0, lb-1), coc)+ &
                                                fz*s(coaz, coset(0, 0, lb-1), coc)+ &
                                                f3*s(coa, coset(0, 0, lb-2), coc)+ &
                                                fcz*s(coa, coset(0, 0, lb-1), cocz)
                                          END IF

!                           *** Shift of angular momentum ***
!                           *** component y from a to b   ***

                                          IF (ay == 0) THEN
                                             bz = lb-1
                                             s(coa, coset(0, 1, bz), coc) = &
                                                rbg(2)*s(coa, coset(0, 0, bz), coc)+ &
                                                fcy*s(coa, coset(0, 0, bz), cocy)
                                             DO by = 2, lb
                                                bz = lb-by
                                                f3 = f2*REAL(by-1, dp)
                                                s(coa, coset(0, by, bz), coc) = &
                                                   rbg(2)*s(coa, coset(0, by-1, bz), coc)+ &
                                                   f3*s(coa, coset(0, by-2, bz), coc)+ &
                                                   fcy*s(coa, coset(0, by-1, bz), cocy)
                                             END DO
                                          ELSE
                                             bz = lb-1
                                             s(coa, coset(0, 1, bz), coc) = &
                                                rbg(2)*s(coa, coset(0, 0, bz), coc)+ &
                                                fy*s(coay, coset(0, 0, bz), coc)+ &
                                                fcy*s(coa, coset(0, 0, bz), cocy)
                                             DO by = 2, lb
                                                bz = lb-by
                                                f3 = f2*REAL(by-1, dp)
                                                s(coa, coset(0, by, bz), coc) = &
                                                   rbg(2)*s(coa, coset(0, by-1, bz), coc)+ &
                                                   fy*s(coay, coset(0, by-1, bz), coc)+ &
                                                   f3*s(coa, coset(0, by-2, bz), coc)+ &
                                                   fcy*s(coa, coset(0, by-1, bz), cocy)
                                             END DO
                                          END IF

!                           *** Shift of angular momentum ***
!                           *** component x from a to b   ***

                                          IF (ax == 0) THEN
                                             DO by = 0, lb-1
                                                bz = lb-1-by
                                                s(coa, coset(1, by, bz), coc) = &
                                                   rbg(1)*s(coa, coset(0, by, bz), coc)+ &
                                                   fcx*s(coa, coset(0, by, bz), cocx)
                                             END DO
                                             DO bx = 2, lb
                                                f3 = f2*REAL(bx-1, dp)
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   s(coa, coset(bx, by, bz), coc) = &
                                                      rbg(1)*s(coa, coset(bx-1, by, bz), coc)+ &
                                                      f3*s(coa, coset(bx-2, by, bz), coc)+ &
                                                      fcx*s(coa, coset(bx-1, by, bz), cocx)
                                                END DO
                                             END DO
                                          ELSE
                                             DO by = 0, lb-1
                                                bz = lb-1-by
                                                s(coa, coset(1, by, bz), coc) = &
                                                   rbg(1)*s(coa, coset(0, by, bz), coc)+ &
                                                   fx*s(coax, coset(0, by, bz), coc)+ &
                                                   fcx*s(coa, coset(0, by, bz), cocx)
                                             END DO
                                             DO bx = 2, lb
                                                f3 = f2*REAL(bx-1, dp)
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   s(coa, coset(bx, by, bz), coc) = &
                                                      rbg(1)*s(coa, coset(bx-1, by, bz), coc)+ &
                                                      fx*s(coax, coset(bx-1, by, bz), coc)+ &
                                                      f3*s(coa, coset(bx-2, by, bz), coc)+ &
                                                      fcx*s(coa, coset(bx-1, by, bz), cocx)
                                                END DO
                                             END DO
                                          END IF

                                       END DO
                                    END DO

                                 END DO

                              END IF

                           ELSE

                              IF (lb_max > 0) THEN

!                     *** Vertical recurrence steps: [s|c|s] -> [s|c|b] ***

                                 rbg(:) = rcg(:)+rbc(:)

!                     *** [s|c|p] = (Gi - Bi)*[s|c|s] + f2*Ni(c)*[s|c-1i|s] ***

                                 s(1, 2, coc) = rbg(1)*s(1, 1, coc)+fcx*s(1, 1, cocx)
                                 s(1, 3, coc) = rbg(2)*s(1, 1, coc)+fcy*s(1, 1, cocy)
                                 s(1, 4, coc) = rbg(3)*s(1, 1, coc)+fcz*s(1, 1, cocz)

!                     *** [s|c|b] = (Gi - Bi)*[s|c|b-1i] + ***
!                     ***           f2*Ni(b-1i)*[s|c|b-2i] ***
!                     ***           f2*Ni(c)*[s|c-1i|b-1i] ***

                                 DO lb = 2, lb_max

!                       *** Increase the angular momentum component z of b ***

                                    s(1, coset(0, 0, lb), coc) = &
                                       rbg(3)*s(1, coset(0, 0, lb-1), coc)+ &
                                       f2*REAL(lb-1, dp)*s(1, coset(0, 0, lb-2), coc)+ &
                                       fcz*s(1, coset(0, 0, lb-1), cocz)

!                       *** Increase the angular momentum component y of b ***

                                    bz = lb-1
                                    s(1, coset(0, 1, bz), coc) = &
                                       rbg(2)*s(1, coset(0, 0, bz), coc)+ &
                                       fcy*s(1, coset(0, 0, bz), cocy)

                                    DO by = 2, lb
                                       bz = lb-by
                                       s(1, coset(0, by, bz), coc) = &
                                          rbg(2)*s(1, coset(0, by-1, bz), coc)+ &
                                          f2*REAL(by-1, dp)*s(1, coset(0, by-2, bz), coc)+ &
                                          fcy*s(1, coset(0, by-1, bz), cocy)
                                    END DO

!                       *** Increase the angular momentum component x of b ***

                                    DO by = 0, lb-1
                                       bz = lb-1-by
                                       s(1, coset(1, by, bz), coc) = &
                                          rbg(1)*s(1, coset(0, by, bz), coc)+ &
                                          fcx*s(1, coset(0, by, bz), cocx)
                                    END DO

                                    DO bx = 2, lb
                                       f3 = f2*REAL(bx-1, dp)
                                       DO by = 0, lb-bx
                                          bz = lb-bx-by
                                          s(1, coset(bx, by, bz), coc) = &
                                             rbg(1)*s(1, coset(bx-1, by, bz), coc)+ &
                                             f3*s(1, coset(bx-2, by, bz), coc)+ &
                                             fcx*s(1, coset(bx-1, by, bz), cocx)
                                       END DO
                                    END DO

                                 END DO

                              END IF

                           END IF

                        END DO
                     END DO

                  END DO

               END IF

!         *** Store integrals

               DO k = ncoset(lc_min_set-1)+1, ncoset(lc_max_set)
                  DO j = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
                     DO i = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                        sabc(na+i, nb+j, nc+k) = s(i, j, k)
                     END DO
                  END DO
               END DO

!         *** Calculate the requested derivatives with respect to  ***
!         *** the nuclear coordinates of the atomic center a and c ***

               IF (PRESENT(sdabc) .OR. PRESENT(sabdc)) THEN
                  CALL derivatives_overlap3(la_max_set, la_min_set, lb_max_set, lb_min_set, &
                                            lc_max_set, lc_min_set, zeta(ipgf), zetc(kpgf), &
                                            s, sda, sdc)
               ENDIF

!         *** Store the first derivatives of the primitive overlap integrals ***

               IF (PRESENT(sdabc)) THEN
                  DO k = 1, 3
                     DO l = 1, ncoset(lc_max_set)
                        DO j = 1, ncoset(lb_max_set)
                           DO i = 1, ncoset(la_max_set)
                              sdabc(nda+i, nb+j, nc+l, k) = sda(i, j, l, k)
                           END DO
                        END DO
                     ENDDO
                  END DO
               END IF

               IF (PRESENT(sabdc)) THEN
                  DO k = 1, 3
                     DO l = 1, ncoset(lc_max_set)
                        DO j = 1, ncoset(lb_max_set)
                           DO i = 1, ncoset(la_max_set)
                              sabdc(na+i, nb+j, ndc+l, k) = sdc(i, j, l, k)
                           END DO
                        END DO
                     ENDDO
                  END DO
               END IF

               nc = nc+ncoset(lc_max_set)
               ndc = ndc+ncoset(lc_max_set)
            END DO

            nb = nb+ncoset(lb_max)
         END DO

         na = na+ncoset(la_max_set)
         nda = nda+ncoset(la_max_set)
      END DO

      DEALLOCATE (s)
      IF (PRESENT(sdabc)) THEN
         DEALLOCATE (sda)
      ENDIF
      IF (PRESENT(sabdc)) THEN
         DEALLOCATE (sdc)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE overlap3

! **************************************************************************************************
!> \brief Calculates the derivatives of the three-center overlap integral [a|b|c]
!>        with respect to the nuclear coordinates of the atomic center a and c
!> \param la_max_set ...
!> \param la_min_set ...
!> \param lb_max_set ...
!> \param lb_min_set ...
!> \param lc_max_set ...
!> \param lc_min_set ...
!> \param zeta ...
!> \param zetc ...
!> \param s integrals [a|b|c]
!> \param sda derivative [da/dAi|b|c]
!> \param sdc derivative [a|b|dc/dCi]
! **************************************************************************************************
   SUBROUTINE derivatives_overlap3(la_max_set, la_min_set, lb_max_set, lb_min_set, &
                                   lc_max_set, lc_min_set, zeta, zetc, s, sda, sdc)

      INTEGER, INTENT(IN)                                :: la_max_set, la_min_set, lb_max_set, &
                                                            lb_min_set, lc_max_set, lc_min_set
      REAL(KIND=dp), INTENT(IN)                          :: zeta, zetc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: s
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: sda, sdc

      CHARACTER(len=*), PARAMETER :: routineN = 'derivatives_overlap3', &
         routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coa, coamx, coamy, coamz, coapx, coapy, coapz, cob, coc, &
         cocmx, cocmy, cocmz, cocpx, cocpy, cocpz, cx, cy, cz, devx, devy, devz, handle, la, lb, lc
      REAL(KIND=dp)                                      :: fax, fay, faz, fcx, fcy, fcz, fexpa, &
                                                            fexpc

      CALL timeset(routineN, handle)

      fexpa = 2.0_dp*zeta
      fexpc = 2.0_dp*zetc

!   derviative with respec to x,y,z

      devx = 1
      devy = 2
      devz = 3

!   *** [da/dAi|b|c] = 2*zeta*[a+1i|b|c] - Ni(a)[a-1i|b|c] ***
!   *** [a|b|dc/dCi] = 2*zetc*[a|b|c+1i] - Ni(c)[a|b|c-1i] ***

      DO la = la_min_set, la_max_set
         DO ax = 0, la
            fax = REAL(ax, dp)
            DO ay = 0, la-ax
               fay = REAL(ay, dp)
               az = la-ax-ay
               faz = REAL(az, dp)
               coa = coset(ax, ay, az)
               coamx = coset(ax-1, ay, az)
               coamy = coset(ax, ay-1, az)
               coamz = coset(ax, ay, az-1)
               coapx = coset(ax+1, ay, az)
               coapy = coset(ax, ay+1, az)
               coapz = coset(ax, ay, az+1)
               DO lb = lb_min_set, lb_max_set
                  DO bx = 0, lb
                     DO by = 0, lb-bx
                        bz = lb-bx-by
                        cob = coset(bx, by, bz)
                        DO lc = lc_min_set, lc_max_set
                           DO cx = 0, lc
                              fcx = REAL(cx, dp)
                              DO cy = 0, lc-cx
                                 fcy = REAL(cy, dp)
                                 cz = lc-cx-cy
                                 fcz = REAL(cz, dp)
                                 coc = coset(cx, cy, cz)
                                 cocmx = coset(cx-1, cy, cz)
                                 cocmy = coset(cx, cy-1, cz)
                                 cocmz = coset(cx, cy, cz-1)
                                 cocpx = coset(cx+1, cy, cz)
                                 cocpy = coset(cx, cy+1, cz)
                                 cocpz = coset(cx, cy, cz+1)
                                 IF (ASSOCIATED(sda)) THEN
                                    sda(coa, cob, coc, devx) = fexpa*s(coapx, cob, coc)- &
                                                               fax*s(coamx, cob, coc)
                                    sda(coa, cob, coc, devy) = fexpa*s(coapy, cob, coc)- &
                                                               fay*s(coamy, cob, coc)
                                    sda(coa, cob, coc, devz) = fexpa*s(coapz, cob, coc)- &
                                                               faz*s(coamz, cob, coc)
                                 ENDIF
                                 IF (ASSOCIATED(sdc)) THEN
                                    sdc(coa, cob, coc, devx) = fexpc*s(coa, cob, cocpx)- &
                                                               fcx*s(coa, cob, cocmx)
                                    sdc(coa, cob, coc, devy) = fexpc*s(coa, cob, cocpy)- &
                                                               fcy*s(coa, cob, cocmy)
                                    sdc(coa, cob, coc, devz) = fexpc*s(coa, cob, cocpz)- &
                                                               fcz*s(coa, cob, cocmz)
                                 ENDIF
                              ENDDO
                           ENDDO
                        ENDDO
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE derivatives_overlap3

END MODULE ai_overlap3
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Debugs the  integral matrices for the LRIGPW method
!>        lri : local resolution of the identity
!> \par History
!>      created [07.2014]
!> \authors Dorothea Golze
! **************************************************************************************************
MODULE lri_debug_integrals

   USE ai_overlap,                      ONLY: overlap
   USE ai_overlap3,                     ONLY: overlap3
   USE ai_overlap3_debug,               ONLY: init_os_overlap3,&
                                              os_overlap3
   USE ai_overlap_aabb,                 ONLY: overlap_aabb
   USE ai_overlap_debug,                ONLY: init_os_overlap2,&
                                              os_overlap2
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_debug_integrals'

   PUBLIC :: overlap_ab_test, overlap_abc_test, overlap_aabb_test

! **************************************************************************************************

CONTAINS

! ***************************************************************************************************
!> \brief recursive test routines for integral (a,b) for only two exponents
! **************************************************************************************************
   SUBROUTINE overlap_ab_test_simple()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_ab_test_simple', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia1, iax, iay, iaz, ib1, ibx, iby, ibz, &
                                                            la_max, la_min, lb_max, lb_min, lds, &
                                                            ma, maxl, mb
      INTEGER, DIMENSION(3)                              :: na, nb
      REAL(KIND=dp)                                      :: dab, dmax, res1, xa, xb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: sab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: swork
      REAL(KIND=dp), DIMENSION(1)                        :: rpgfa, rpgfb, xa_work, xb_work
      REAL(KIND=dp), DIMENSION(3)                        :: A, B, rab

      xa = 0.783300000000 ! exponents
      xb = 1.239648746700

      A = (/0.329309000000, 0.28408240000, 0.28408240000/) !* bohr  !positions
      B = (/0.983983000000, 0.00453720000, 0.00432740000/) !* bohr

      la_min = 0
      lb_min = 0

      la_max = 3
      lb_max = 4

      !---------------------------------------
      ALLOCATE (sab(ncoset(la_max), ncoset(lb_max)))

      maxl = MAX(la_max, lb_max)
      lds = ncoset(maxl)
      ALLOCATE (swork(lds, lds, 1))
      sab = 0._dp
      rab(:) = B(:)-A(:)
      dab = SQRT(DOT_PRODUCT(rab, rab))
      xa_work(1) = xa
      xb_work(1) = xb
      rpgfa = 20._dp
      rpgfb = 20._dp
      CALL overlap(la_max_set=la_max, la_min_set=la_min, npgfa=1, rpgfa=rpgfa, zeta=xa_work, &
                   lb_max_set=lb_max, lb_min_set=lb_min, npgfb=1, rpgfb=rpgfb, zetb=xb_work, &
                   rab=rab, dab=dab, sab=sab, da_max_set=0, return_derivatives=.FALSE., s=swork, lds=lds)
      !---------------------------------------

      CALL init_os_overlap2(xa, xb, A, B)

      dmax = 0._dp
      DO ma = la_min, la_max
         DO mb = lb_min, lb_max
            DO iax = 0, ma
               DO iay = 0, ma-iax
                  iaz = ma-iax-iay
                  na(1) = iax; na(2) = iay; na(3) = iaz
                  ia1 = coset(iax, iay, iaz)
                  DO ibx = 0, mb
                     DO iby = 0, mb-ibx
                        ibz = mb-ibx-iby
                        nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                        ib1 = coset(ibx, iby, ibz)
                        res1 = os_overlap2(na, nb)
                        ! write(*,*) "la, lb,na, nb, res1", ma, mb, na, nb, res1
                        ! write(*,*) "sab ia1, ib1", ia1, ib1, sab(ia1,ib1)
                        dmax = MAX(dmax, ABS(res1-sab(ia1, ib1)))
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO

      WRITE (*, *) "dmax", dmax

      DEALLOCATE (sab, swork)

   END SUBROUTINE overlap_ab_test_simple

! ***************************************************************************************************
!> \brief recursive test routines for integral (a,b)
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param zetb ...
!> \param ra ...
!> \param rb ...
!> \param sab ...
!> \param dmax ...
! **************************************************************************************************
   SUBROUTINE overlap_ab_test(la_max, la_min, npgfa, zeta, lb_max, lb_min, npgfb, zetb, &
                              ra, rb, sab, dmax)

      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: sab
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_ab_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: coa, cob, ia1, iax, iay, iaz, ib1, ibx, &
                                                            iby, ibz, ipgf, jpgf, ma, mb
      INTEGER, DIMENSION(3)                              :: na, nb
      REAL(KIND=dp)                                      :: res1, res2, xa, xb
      REAL(KIND=dp), DIMENSION(3)                        :: A, B

      coa = 0
      DO ipgf = 1, npgfa
         cob = 0
         DO jpgf = 1, npgfb
            xa = zeta(ipgf) !exponents
            xb = zetb(jpgf)
            A = ra !positions
            B = rb
            CALL init_os_overlap2(xa, xb, A, B)
            DO ma = la_min, la_max
               DO mb = lb_min, lb_max
                  DO iax = 0, ma
                     DO iay = 0, ma-iax
                        iaz = ma-iax-iay
                        na(1) = iax; na(2) = iay; na(3) = iaz
                        ia1 = coset(iax, iay, iaz)
                        DO ibx = 0, mb
                           DO iby = 0, mb-ibx
                              ibz = mb-ibx-iby
                              nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                              ib1 = coset(ibx, iby, ibz)
                              res1 = os_overlap2(na, nb)
                              res2 = sab(coa+ia1, cob+ib1)
                              dmax = MAX(dmax, ABS(res1-res2))
                           END DO
                        END DO
                     END DO
                  END DO
               END DO
            END DO
            cob = cob+ncoset(lb_max)
         END DO
         coa = coa+ncoset(la_max)
      END DO
      !WRITE(*,*) "dmax overlap_ab_test", dmax

   END SUBROUTINE overlap_ab_test

! ***************************************************************************************************
!> \brief recursive test routines for integral (a,b,c) for only three exponents
! **************************************************************************************************
   SUBROUTINE overlap_abc_test_simple()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_abc_test_simple', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia1, iax, iay, iaz, ib1, ibx, iby, ibz, &
                                                            ic1, icx, icy, icz, la_max, la_min, &
                                                            lb_max, lb_min, lc_max, lc_min, ma, &
                                                            mb, mc
      INTEGER, DIMENSION(3)                              :: na, nb, nc
      REAL(KIND=dp)                                      :: dab, dac, dbc, dmax, res1, xa, xb, xc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: sabc
      REAL(KIND=dp), DIMENSION(1)                        :: rpgfa, rpgfb, rpgfc, xa_work, xb_work, &
                                                            xc_work
      REAL(KIND=dp), DIMENSION(3)                        :: A, B, C, rab, rac, rbc

      xa = 0.783300000000 ! exponents
      xb = 1.239648746700
      xc = 0.548370000000

      A = (/0.329309000000, 0.28408240000, 0.28408240000/) !* bohr  !positions
      B = (/0.983983000000, 0.00453720000, 0.00432740000/) !* bohr
      C = (/0.032380000000, 1.23470000000, 0.11137400000/) !* bohr

      la_min = 0
      lb_min = 0
      lc_min = 0

      la_max = 0
      lb_max = 0
      lc_max = 1

      !---------------------------------------
      rab(:) = B(:)-A(:)
      dab = SQRT(DOT_PRODUCT(rab, rab))
      rac(:) = C(:)-A(:)
      dac = SQRT(DOT_PRODUCT(rac, rac))
      rbc(:) = C(:)-B(:)
      dbc = SQRT(DOT_PRODUCT(rbc, rbc))
      ALLOCATE (sabc(ncoset(la_max), ncoset(lb_max), ncoset(lc_max)))
      xa_work(1) = xa
      xb_work(1) = xb
      xc_work(1) = xc
      rpgfa = 20._dp
      rpgfb = 20._dp
      rpgfc = 20._dp
      sabc = 0._dp

      CALL overlap3(la_max_set=la_max, npgfa=1, zeta=xa_work, rpgfa=rpgfa, la_min_set=la_min, &
                    lb_max_set=lb_max, npgfb=1, zetb=xb_work, rpgfb=rpgfb, lb_min_set=lb_min, &
                    lc_max_set=lc_max, npgfc=1, zetc=xc_work, rpgfc=rpgfc, lc_min_set=lc_min, &
                    rab=rab, dab=dab, rac=rac, dac=dac, rbc=rbc, dbc=dbc, sabc=sabc)
      !sabc(:,:,:)=0._dp
      !CALL overlap3_old(la_max=la_max,npgfa=1,zeta=xa_work,rpgfa=rpgfa,la_min=la_min,&
      !              lb_max=lb_max,npgfb=1,zetb=xb_work,rpgfb=rpgfb,lb_min=lb_min,&
      !              lc_max=lc_max,npgfc=1,zetc=xc,rpgfc=20._dp,lc_min=lc_min,&
      !              rab=rab,dab=dab,rac=rac,dac=dac,rbc=rbc,dbc=dbc,sabc=sabc)

      !---------------------------------------

      CALL init_os_overlap3(xa, xb, xc, A, B, C)

      dmax = 0._dp
      DO ma = la_min, la_max
         DO mc = lc_min, lc_max
            DO mb = lb_min, lb_max
               DO iax = 0, ma
                  DO iay = 0, ma-iax
                     iaz = ma-iax-iay
                     na(1) = iax; na(2) = iay; na(3) = iaz
                     ia1 = coset(iax, iay, iaz)
                     DO icx = 0, mc
                        DO icy = 0, mc-icx
                           icz = mc-icx-icy
                           nc(1) = icx; nc(2) = icy; nc(3) = icz
                           ic1 = coset(icx, icy, icz)
                           DO ibx = 0, mb
                              DO iby = 0, mb-ibx
                                 ibz = mb-ibx-iby
                                 nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                                 ib1 = coset(ibx, iby, ibz)
                                 res1 = os_overlap3(na, nc, nb)
                                 !write(*,*) "la, lc, lb,na,nc, nb, res1", ma, mc, mb, na, nc, nb, res1
                                 !write(*,*) "sabc ia1, ib1, ic1", ia1, ib1, ic1, sabc(ia1,ib1,ic1)
                                 dmax = MAX(dmax, ABS(res1-sabc(ia1, ib1, ic1)))
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO

      WRITE (*, *) "dmax", dmax

      DEALLOCATE (sabc)

   END SUBROUTINE overlap_abc_test_simple

! ***************************************************************************************************
!> \brief recursive test routines for integral (a,b,c)
!> \param la_max ...
!> \param npgfa ...
!> \param zeta ...
!> \param la_min ...
!> \param lb_max ...
!> \param npgfb ...
!> \param zetb ...
!> \param lb_min ...
!> \param lc_max ...
!> \param npgfc ...
!> \param zetc ...
!> \param lc_min ...
!> \param ra ...
!> \param rb ...
!> \param rc ...
!> \param sabc ...
!> \param dmax ...
! **************************************************************************************************
   SUBROUTINE overlap_abc_test(la_max, npgfa, zeta, la_min, &
                               lb_max, npgfb, zetb, lb_min, &
                               lc_max, npgfc, zetc, lc_min, &
                               ra, rb, rc, sabc, dmax)

      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: la_min, lb_max, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb
      INTEGER, INTENT(IN)                                :: lb_min, lc_max, npgfc
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetc
      INTEGER, INTENT(IN)                                :: lc_min
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rc
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: sabc
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_abc_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: coa, cob, coc, ia1, iax, iay, iaz, ib1, &
                                                            ibx, iby, ibz, ic1, icx, icy, icz, &
                                                            ipgf, jpgf, kpgf, ma, mb, mc
      INTEGER, DIMENSION(3)                              :: na, nb, nc
      REAL(KIND=dp)                                      :: res1, res2, xa, xb, xc
      REAL(KIND=dp), DIMENSION(3)                        :: A, B, C

      coa = 0
      DO ipgf = 1, npgfa
         cob = 0
         DO jpgf = 1, npgfb
            coc = 0
            DO kpgf = 1, npgfc

               xa = zeta(ipgf) ! exponents
               xb = zetb(jpgf)
               xc = zetc(kpgf)

               A = Ra !positions
               B = Rb
               C = Rc

               CALL init_os_overlap3(xa, xb, xc, A, B, C)

               DO ma = la_min, la_max
                  DO mc = lc_min, lc_max
                     DO mb = lb_min, lb_max
                        DO iax = 0, ma
                           DO iay = 0, ma-iax
                              iaz = ma-iax-iay
                              na(1) = iax; na(2) = iay; na(3) = iaz
                              ia1 = coset(iax, iay, iaz)
                              DO icx = 0, mc
                                 DO icy = 0, mc-icx
                                    icz = mc-icx-icy
                                    nc(1) = icx; nc(2) = icy; nc(3) = icz
                                    ic1 = coset(icx, icy, icz)
                                    DO ibx = 0, mb
                                       DO iby = 0, mb-ibx
                                          ibz = mb-ibx-iby
                                          nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                                          ib1 = coset(ibx, iby, ibz)
                                          res1 = os_overlap3(na, nc, nb)
                                          res2 = sabc(coa+ia1, cob+ib1, coc+ic1)
                                          dmax = MAX(dmax, ABS(res1-res2))
                                          !IF(dmax > 1.E-10) WRITE(*,*) "dmax in loop", dmax
                                       END DO
                                    END DO
                                 END DO
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO
               END DO
               coc = coc+ncoset(lc_max)
            END DO
            cob = cob+ncoset(lb_max)
         END DO
         coa = coa+ncoset(la_max)
      END DO
      !WRITE(*,*) "dmax abc", dmax

   END SUBROUTINE overlap_abc_test

! ***************************************************************************************************
!> \brief recursive test routines for integral (aa,bb) for only four exponents
! **************************************************************************************************
   SUBROUTINE overlap_aabb_test_simple()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_aabb_test_simple', &
         routineP = moduleN//':'//routineN

      INTEGER :: i, iax, iay, iaz, ibx, iby, ibz, j, k, l, la_max, la_max1, la_max2, la_min, &
         la_min1, la_min2, lb_max, lb_max1, lb_max2, lb_min, lb_min1, lb_min2, lds, ma, maxl, mb
      INTEGER, DIMENSION(3)                              :: na, naa, nb, nbb
      REAL(KIND=dp)                                      :: dab, dmax, res1, xa, xa1, xa2, xb, xb1, &
                                                            xb2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: swork
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: saabb
      REAL(KIND=dp), DIMENSION(1)                        :: rpgfa1, rpgfa2, rpgfb1, rpgfb2, &
                                                            xa_work1, xa_work2, xb_work1, xb_work2
      REAL(KIND=dp), DIMENSION(3)                        :: A, B, rab

      xa1 = 0.783300000000 ! exponents
      xb1 = 1.239648746700
      xa2 = 0.3400000000 ! exponents
      xb2 = 2.76

      A = (/0.329309000000, 0.28408240000, 0.28408240000/) !* bohr  !positions
      B = (/0.983983000000, 0.00453720000, 0.00432740000/) !* bohr

      la_min1 = 0
      la_min2 = 0
      lb_min1 = 3
      lb_min2 = 1

      la_max1 = 1
      la_max2 = 2
      lb_max1 = 3
      lb_max2 = 4

      !---------------------------------------
      ALLOCATE (saabb(ncoset(la_max1), ncoset(la_max2), ncoset(lb_max1), ncoset(lb_max2)))

      maxl = MAX(la_max1+la_max2, lb_max1+lb_max2)
      lds = ncoset(maxl)
      ALLOCATE (swork(lds, lds))
      saabb = 0._dp
      rab(:) = B(:)-A(:)
      dab = SQRT(DOT_PRODUCT(rab, rab))
      xa_work1(1) = xa1
      xa_work2(1) = xa2
      xb_work1(1) = xb1
      xb_work2(1) = xb2
      rpgfa1 = 20._dp
      rpgfa2 = 20._dp
      rpgfb1 = 20._dp
      rpgfb2 = 20._dp
      CALL overlap_aabb(la_max_set1=la_max1, la_min_set1=la_min1, npgfa1=1, rpgfa1=rpgfa1, zeta1=xa_work1, &
                        la_max_set2=la_max2, la_min_set2=la_min2, npgfa2=1, rpgfa2=rpgfa2, zeta2=xa_work2, &
                        lb_max_set1=lb_max1, lb_min_set1=lb_min1, npgfb1=1, rpgfb1=rpgfb1, zetb1=xb_work1, &
                        lb_max_set2=lb_max2, lb_min_set2=lb_min2, npgfb2=1, rpgfb2=rpgfb2, zetb2=xb_work2, &
                        asets_equal=.FALSE., bsets_equal=.FALSE., rab=rab, dab=dab, saabb=saabb, s=swork, lds=lds)
      !---------------------------------------

      xa = xa1+xa2
      xb = xb1+xb2
      la_min = la_min1+la_min2
      la_max = la_max1+la_max2
      lb_min = lb_min1+lb_min2
      lb_max = lb_max1+lb_max2

      CALL init_os_overlap2(xa, xb, A, B)

      dmax = 0._dp
      DO ma = la_min, la_max
         DO mb = lb_min, lb_max
            DO iax = 0, ma
               DO iay = 0, ma-iax
                  iaz = ma-iax-iay
                  na(1) = iax; na(2) = iay; na(3) = iaz
                  DO ibx = 0, mb
                     DO iby = 0, mb-ibx
                        ibz = mb-ibx-iby
                        nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                        res1 = os_overlap2(na, nb)
                        DO i = ncoset(la_min1-1)+1, ncoset(la_max1)
                           DO j = ncoset(la_min2-1)+1, ncoset(la_max2)
                              naa = indco(1:3, i)+indco(1:3, j)
                              DO k = ncoset(lb_min1-1)+1, ncoset(lb_max1)
                                 DO l = ncoset(lb_min2-1)+1, ncoset(lb_max2)
                                    nbb = indco(1:3, k)+indco(1:3, l)
                                    IF (ALL(na == naa) .AND. ALL(nb == nbb)) THEN
                                       dmax = MAX(dmax, ABS(res1-saabb(i, j, k, l)))
                                    END IF
                                 END DO
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO

      WRITE (*, *) "dmax", dmax

      DEALLOCATE (saabb, swork)

   END SUBROUTINE overlap_aabb_test_simple

! ***************************************************************************************************
!> \brief recursive test routines for integral (aa,bb)
!> \param la_max1 ...
!> \param la_min1 ...
!> \param npgfa1 ...
!> \param zeta1 ...
!> \param la_max2 ...
!> \param la_min2 ...
!> \param npgfa2 ...
!> \param zeta2 ...
!> \param lb_max1 ...
!> \param lb_min1 ...
!> \param npgfb1 ...
!> \param zetb1 ...
!> \param lb_max2 ...
!> \param lb_min2 ...
!> \param npgfb2 ...
!> \param zetb2 ...
!> \param ra ...
!> \param rb ...
!> \param saabb ...
!> \param dmax ...
! **************************************************************************************************
   SUBROUTINE overlap_aabb_test(la_max1, la_min1, npgfa1, zeta1, &
                                la_max2, la_min2, npgfa2, zeta2, &
                                lb_max1, lb_min1, npgfb1, zetb1, &
                                lb_max2, lb_min2, npgfb2, zetb2, &
                                ra, rb, saabb, dmax)

      INTEGER, INTENT(IN)                                :: la_max1, la_min1, npgfa1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta1
      INTEGER, INTENT(IN)                                :: la_max2, la_min2, npgfa2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta2
      INTEGER, INTENT(IN)                                :: lb_max1, lb_min1, npgfb1
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb1
      INTEGER, INTENT(IN)                                :: lb_max2, lb_min2, npgfb2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: saabb
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'overlap_aabb_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: coa1, coa2, cob1, cob2, i, iax, iay, &
                                                            iaz, ibx, iby, ibz, ipgf, j, jpgf, k, &
                                                            kpgf, l, la_max, la_min, lb_max, &
                                                            lb_min, lpgf, ma, mb
      INTEGER, DIMENSION(3)                              :: na, naa, nb, nbb
      REAL(KIND=dp)                                      :: res1, xa, xb
      REAL(KIND=dp), DIMENSION(3)                        :: A, B

      coa1 = 0
      DO ipgf = 1, npgfa1
         coa2 = 0
         DO jpgf = 1, npgfa2
            cob1 = 0
            DO kpgf = 1, npgfb1
               cob2 = 0
               DO lpgf = 1, npgfb2

                  xa = zeta1(ipgf)+zeta2(jpgf) ! exponents
                  xb = zetb1(kpgf)+zetb2(lpgf) ! exponents
                  la_max = la_max1+la_max2
                  lb_max = lb_max1+lb_max2
                  la_min = la_min1+la_min2
                  lb_min = lb_min1+lb_min2

                  A = ra !positions
                  B = rb

                  CALL init_os_overlap2(xa, xb, A, B)

                  DO ma = la_min, la_max
                     DO mb = lb_min, lb_max
                        DO iax = 0, ma
                           DO iay = 0, ma-iax
                              iaz = ma-iax-iay
                              na(1) = iax; na(2) = iay; na(3) = iaz
                              DO ibx = 0, mb
                                 DO iby = 0, mb-ibx
                                    ibz = mb-ibx-iby
                                    nb(1) = ibx; nb(2) = iby; nb(3) = ibz
                                    res1 = os_overlap2(na, nb)
                                    DO i = ncoset(la_min1-1)+1, ncoset(la_max1)
                                       DO j = ncoset(la_min2-1)+1, ncoset(la_max2)
                                          naa = indco(1:3, i)+indco(1:3, j)
                                          DO k = ncoset(lb_min1-1)+1, ncoset(lb_max1)
                                             DO l = ncoset(lb_min2-1)+1, ncoset(lb_max2)
                                                nbb = indco(1:3, k)+indco(1:3, l)
                                                IF (ALL(na == naa) .AND. ALL(nb == nbb)) THEN
                                                   dmax = MAX(dmax, ABS(res1-saabb(coa1+i, coa2+j, cob1+k, cob2+l)))
                                                ENDIF
                                             END DO
                                          END DO
                                       END DO
                                    END DO
                                 END DO
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO
                  cob2 = cob2+ncoset(lb_max2)
               END DO
               cob1 = cob1+ncoset(lb_max1)
            END DO
            coa2 = coa2+ncoset(la_max2)
         END DO
         coa1 = coa1+ncoset(la_max1)
      END DO

      !WRITE(*,*) "dmax aabb", dmax

   END SUBROUTINE overlap_aabb_test

END MODULE lri_debug_integrals
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates integral matrices for LRIGPW method using the Obara-Saika
!>        (OS) integral scheme
!>        lri : local resolution of the identity
!> \par History
!>      created [06.2015]
!> \author Dorothea Golze
! **************************************************************************************************
MODULE lri_os_integrals
   USE ai_overlap,                      ONLY: overlap
   USE ai_overlap3,                     ONLY: overlap3
   USE ai_overlap_aabb,                 ONLY: overlap_aabb
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE lri_debug_integrals,             ONLY: overlap_aabb_test,&
                                              overlap_ab_test,&
                                              overlap_abc_test
   USE orbital_pointers,                ONLY: ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_os_integrals'

   PUBLIC :: lri_int_ab, lri_int_aba, lri_int_abb, lri_int_aabb

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief calculate overlap integrals (a,b)
!> \param sab integral (a,b)
!> \param dsab derivative of sab with respect to A
!> \param ra ...
!> \param rb ...
!> \param rab ...
!> \param fba basis at center A
!> \param fbb basis at center B
!> \param calculate_forces ...
!> \param debug integrals are debugged by recursive routines if requested
!> \param dmax maximal deviation between integrals when debugging
! **************************************************************************************************
   SUBROUTINE lri_int_ab(sab, dsab, ra, rb, rab, fba, fbb, calculate_forces, debug, &
                         dmax)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: sab
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: dsab
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rab
      TYPE(gto_basis_set_type), POINTER                  :: fba, fbb
      LOGICAL, INTENT(IN)                                :: calculate_forces, debug
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_ab', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iset, jset, lds, m1, m2, &
                                                            maxco, maxcoa, maxcob, maxl, maxla, &
                                                            maxlb, ncoa, ncob, nseta, nsetb, sgfa, &
                                                            sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      REAL(KIND=dp)                                      :: dab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: sint
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: devab, swork
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb

      CALL timeset(routineN, handle)
      NULLIFY (la_max, la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb, &
               first_sgfa, first_sgfb, set_radius_a, set_radius_b, rpgfa, rpgfb, &
               sphi_a, sphi_b, zeta, zetb)

      ! basis ikind
      first_sgfa => fba%first_sgf
      la_max => fba%lmax
      la_min => fba%lmin
      npgfa => fba%npgf
      nseta = fba%nset
      nsgfa => fba%nsgf_set
      rpgfa => fba%pgf_radius
      set_radius_a => fba%set_radius
      sphi_a => fba%sphi
      zeta => fba%zet
      ! basis jkind
      first_sgfb => fbb%first_sgf
      lb_max => fbb%lmax
      lb_min => fbb%lmin
      npgfb => fbb%npgf
      nsetb = fbb%nset
      nsgfb => fbb%nsgf_set
      rpgfb => fbb%pgf_radius
      set_radius_b => fbb%set_radius
      sphi_b => fbb%sphi
      zetb => fbb%zet

      CALL get_gto_basis_set(fba, maxco=maxcoa, maxl=maxla)
      CALL get_gto_basis_set(fbb, maxco=maxcob, maxl=maxlb)
      maxco = MAX(maxcoa, maxcob)
      IF (calculate_forces) THEN
         maxl = MAX(maxla+1, maxlb)
      ELSE
         maxl = MAX(maxla, maxlb)
      ENDIF
      lds = ncoset(maxl)
      ALLOCATE (sint(maxco, maxco))
      IF (calculate_forces) THEN
         !derivatives will be stored in devab(:,:,2:4)
         ALLOCATE (swork(lds, lds, 4), devab(maxco, maxco, 4))
         devab = 0._dp
      ELSE
         ALLOCATE (swork(lds, lds, 1))
      ENDIF
      sint = 0._dp
      swork = 0._dp

      dab = SQRT(SUM(rab**2))

      DO iset = 1, nseta

         ncoa = npgfa(iset)*ncoset(la_max(iset))
         sgfa = first_sgfa(1, iset)

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            m1 = sgfa+nsgfa(iset)-1
            m2 = sgfb+nsgfb(jset)-1
            sint = 0._dp
            swork = 0._dp

            ! calculate integrals
            IF (calculate_forces) THEN
               devab = 0._dp
               CALL overlap(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                            lb_max(jset), lb_min(jset), npgfb(jset), rpgfb(:, jset), zetb(:, jset), &
                            rab, dab, sint, 0, .FALSE., swork, lds, sdab=devab)
               DO i = 1, 3
                  !NOTE: devab(:,:,2:4) contains all derivatives for lmin=0 to lmax=lmax
                  !      correct after contraction (multiply with zero for elements l < lmin)
                  CALL ab_contract(dsab(sgfa:m1, sgfb:m2, i), devab(:, :, i+1), sphi_a(:, sgfa:), &
                                   sphi_b(:, sgfb:), ncoa, ncob, nsgfa(iset), nsgfb(jset))
               ENDDO

            ELSE
               CALL overlap(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                            lb_max(jset), lb_min(jset), npgfb(jset), rpgfb(:, jset), zetb(:, jset), &
                            rab, dab, sint, 0, .FALSE., swork, lds)
            ENDIF
            ! debug if requested
            IF (debug) THEN
               CALL overlap_ab_test(la_max(iset), la_min(iset), npgfa(iset), zeta(:, iset), &
                                    lb_max(jset), lb_min(jset), npgfb(jset), zetb(:, jset), &
                                    ra, rb, sint, dmax)
            ENDIF

            CALL ab_contract(sab(sgfa:m1, sgfb:m2), sint, sphi_a(:, sgfa:), sphi_b(:, sgfb:), &
                             ncoa, ncob, nsgfa(iset), nsgfb(jset))
         END DO
      END DO

      IF (calculate_forces) THEN
         DEALLOCATE (devab)
      ENDIF

      DEALLOCATE (sint, swork)

      CALL timestop(handle)

   END SUBROUTINE lri_int_ab

! **************************************************************************************************
!> \brief calculate integrals (a,b,fa)
!> \param abaint integral (a,b,fa)
!> \param dabdaint derivative of abaint with respect to A
!> \param ra ...
!> \param rb ...
!> \param rab ...
!> \param oba orbital basis at center A
!> \param obb orbital basis at center B
!> \param fba auxiliary basis set at center A
!> \param calculate_forces ...
!> \param debug integrals are debugged by recursive routines if requested
!> \param dmax maximal deviation between integrals when debugging
! **************************************************************************************************
   SUBROUTINE lri_int_aba(abaint, dabdaint, ra, rb, rab, oba, obb, fba, &
                          calculate_forces, debug, dmax)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: abaint
      REAL(KIND=dp), DIMENSION(:, :, :, :), OPTIONAL, &
         POINTER                                         :: dabdaint
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rab
      TYPE(gto_basis_set_type), POINTER                  :: oba, obb, fba
      LOGICAL, INTENT(IN)                                :: calculate_forces, debug
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_aba', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iset, jset, kaset, m1, m2, &
                                                            m3, ncoa, ncob, ncoc, nseta, nsetb, &
                                                            nsetca, sgfa, sgfb, sgfc
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, lca_max, &
                                                            lca_min, npgfa, npgfb, npgfca, nsgfa, &
                                                            nsgfb, nsgfca
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb, first_sgfca
      REAL(KIND=dp)                                      :: dab, dac, dbc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: saba
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: sabda, sdaba, sdabda
      REAL(KIND=dp), DIMENSION(3)                        :: rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b, set_radius_ca
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, rpgfca, sphi_a, sphi_b, &
                                                            sphi_ca, zeta, zetb, zetca

      CALL timeset(routineN, handle)
      NULLIFY (la_max, la_min, lb_max, lb_min, lca_max, lca_min, npgfa, npgfb, &
               npgfca, nsgfa, nsgfb, nsgfca)
      NULLIFY (first_sgfa, first_sgfb, first_sgfca, set_radius_a, set_radius_b, &
               set_radius_ca, rpgfa, rpgfb, rpgfca, sphi_a, sphi_b, sphi_ca, &
               zeta, zetb, zetca)

      ! basis ikind
      first_sgfa => oba%first_sgf
      la_max => oba%lmax
      la_min => oba%lmin
      npgfa => oba%npgf
      nseta = oba%nset
      nsgfa => oba%nsgf_set
      rpgfa => oba%pgf_radius
      set_radius_a => oba%set_radius
      sphi_a => oba%sphi
      zeta => oba%zet
      ! basis jkind
      first_sgfb => obb%first_sgf
      lb_max => obb%lmax
      lb_min => obb%lmin
      npgfb => obb%npgf
      nsetb = obb%nset
      nsgfb => obb%nsgf_set
      rpgfb => obb%pgf_radius
      set_radius_b => obb%set_radius
      sphi_b => obb%sphi
      zetb => obb%zet

      ! basis RI A
      first_sgfca => fba%first_sgf
      lca_max => fba%lmax
      lca_min => fba%lmin
      npgfca => fba%npgf
      nsetca = fba%nset
      nsgfca => fba%nsgf_set
      rpgfca => fba%pgf_radius
      set_radius_ca => fba%set_radius
      sphi_ca => fba%sphi
      zetca => fba%zet

      dab = SQRT(SUM(rab**2))

      DO iset = 1, nseta

         ncoa = npgfa(iset)*ncoset(la_max(iset))
         sgfa = first_sgfa(1, iset)

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            m1 = sgfa+nsgfa(iset)-1
            m2 = sgfb+nsgfb(jset)-1

            ! calculate integrals abaint and derivative [d(a,b,a)/dA] dabdaint if requested
            rac = 0._dp
            dac = 0._dp
            rbc = -rab
            dbc = dab

            DO kaset = 1, nsetca

               IF (set_radius_b(jset)+set_radius_ca(kaset) < dab) CYCLE

               ncoc = npgfca(kaset)*ncoset(lca_max(kaset))
               sgfc = first_sgfca(1, kaset)
               m3 = sgfc+nsgfca(kaset)-1
               IF (ncoa*ncob*ncoc > 0) THEN
                  ALLOCATE (saba(ncoa, ncob, ncoc))
                  saba(:, :, :) = 0._dp
                  ! integrals
                  IF (calculate_forces) THEN
                     ALLOCATE (sdaba(ncoa, ncob, ncoc, 3))
                     ALLOCATE (sabda(ncoa, ncob, ncoc, 3))
                     ALLOCATE (sdabda(ncoa, ncob, ncoc, 3))
                     sdaba(:, :, :, :) = 0._dp
                     sabda(:, :, :, :) = 0._dp
                     sdabda(:, :, :, :) = 0._dp
                     CALL overlap3(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                   lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                   lca_max(kaset), npgfca(kaset), zetca(:, kaset), rpgfca(:, kaset), lca_min(kaset), &
                                   rab, dab, rac, dac, rbc, dbc, saba, sdaba, sabda)
                     !d(a,b,a)/dA = (da/dA,b,a) + (a,b,da/dA)
                     sdabda(:, :, :, :) = sdaba+sabda

                     DO i = 1, 3
                        CALL abc_contract(dabdaint(sgfa:m1, sgfb:m2, sgfc:m3, i), sdabda(:, :, :, i), &
                                          sphi_a(:, sgfa:), sphi_b(:, sgfb:), sphi_ca(:, sgfc:), &
                                          ncoa, ncob, ncoc, nsgfa(iset), nsgfb(jset), nsgfca(kaset))
                     ENDDO

                     DEALLOCATE (sdaba, sabda, sdabda)
                  ELSE
                     CALL overlap3(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                   lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                   lca_max(kaset), npgfca(kaset), zetca(:, kaset), rpgfca(:, kaset), lca_min(kaset), &
                                   rab, dab, rac, dac, rbc, dbc, saba)
                  ENDIF
                  ! debug if requested
                  IF (debug) THEN
                     CALL overlap_abc_test(la_max(iset), npgfa(iset), zeta(:, iset), la_min(iset), &
                                           lb_max(jset), npgfb(jset), zetb(:, jset), lb_min(jset), &
                                           lca_max(kaset), npgfca(kaset), zetca(:, kaset), lca_min(kaset), &
                                           ra, rb, ra, saba, dmax)
                  ENDIF
                  CALL abc_contract(abaint(sgfa:m1, sgfb:m2, sgfc:m3), saba, &
                                    sphi_a(:, sgfa:), sphi_b(:, sgfb:), sphi_ca(:, sgfc:), &
                                    ncoa, ncob, ncoc, nsgfa(iset), nsgfb(jset), nsgfca(kaset))
                  DEALLOCATE (saba)
               END IF
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE lri_int_aba

! **************************************************************************************************
!> \brief calculate integrals (a,b,fb)
!> \param abbint integral (a,b,fb)
!> \param dabbint derivative of abbint with respect to A
!> \param ra ...
!> \param rb ...
!> \param rab ...
!> \param oba orbital basis at center A
!> \param obb orbital basis at center B
!> \param fbb auxiliary basis set at center B
!> \param calculate_forces ...
!> \param debug integrals are debugged by recursive routines if requested
!> \param dmax maximal deviation between integrals when debugging
! **************************************************************************************************
   SUBROUTINE lri_int_abb(abbint, dabbint, ra, rb, rab, oba, obb, fbb, calculate_forces, &
                          debug, dmax)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: abbint
      REAL(KIND=dp), DIMENSION(:, :, :, :), OPTIONAL, &
         POINTER                                         :: dabbint
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rab
      TYPE(gto_basis_set_type), POINTER                  :: oba, obb, fbb
      LOGICAL, INTENT(IN)                                :: calculate_forces, debug
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_abb', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iset, jset, kbset, m1, m2, &
                                                            m3, ncoa, ncob, ncoc, nseta, nsetb, &
                                                            nsetcb, sgfa, sgfb, sgfc
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, lcb_max, &
                                                            lcb_min, npgfa, npgfb, npgfcb, nsgfa, &
                                                            nsgfb, nsgfcb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb, first_sgfcb
      REAL(KIND=dp)                                      :: dab, dac, dbc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: sabb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: sdabb
      REAL(KIND=dp), DIMENSION(3)                        :: rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b, set_radius_cb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, rpgfcb, sphi_a, sphi_b, &
                                                            sphi_cb, zeta, zetb, zetcb

      CALL timeset(routineN, handle)
      NULLIFY (la_max, la_min, lb_max, lb_min, lcb_max, lcb_min, npgfa, npgfb, &
               npgfcb, nsgfa, nsgfb, nsgfcb)
      NULLIFY (first_sgfa, first_sgfb, first_sgfcb, set_radius_a, set_radius_b, &
               set_radius_cb, rpgfa, rpgfb, rpgfcb, sphi_a, sphi_b, sphi_cb, &
               zeta, zetb, zetcb)

      ! basis ikind
      first_sgfa => oba%first_sgf
      la_max => oba%lmax
      la_min => oba%lmin
      npgfa => oba%npgf
      nseta = oba%nset
      nsgfa => oba%nsgf_set
      rpgfa => oba%pgf_radius
      set_radius_a => oba%set_radius
      sphi_a => oba%sphi
      zeta => oba%zet
      ! basis jkind
      first_sgfb => obb%first_sgf
      lb_max => obb%lmax
      lb_min => obb%lmin
      npgfb => obb%npgf
      nsetb = obb%nset
      nsgfb => obb%nsgf_set
      rpgfb => obb%pgf_radius
      set_radius_b => obb%set_radius
      sphi_b => obb%sphi
      zetb => obb%zet

      ! basis RI B
      first_sgfcb => fbb%first_sgf
      lcb_max => fbb%lmax
      lcb_min => fbb%lmin
      npgfcb => fbb%npgf
      nsetcb = fbb%nset
      nsgfcb => fbb%nsgf_set
      rpgfcb => fbb%pgf_radius
      set_radius_cb => fbb%set_radius
      sphi_cb => fbb%sphi
      zetcb => fbb%zet

      dab = SQRT(SUM(rab**2))

      DO iset = 1, nseta

         ncoa = npgfa(iset)*ncoset(la_max(iset))
         sgfa = first_sgfa(1, iset)

         DO jset = 1, nsetb

            IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            m1 = sgfa+nsgfa(iset)-1
            m2 = sgfb+nsgfb(jset)-1

            ! calculate integrals abbint and derivative [d(a,b,b)/dA] dabbint if requested
            rac = rab
            dac = dab
            rbc = 0._dp
            dbc = 0._dp

            DO kbset = 1, nsetcb

               IF (set_radius_a(iset)+set_radius_cb(kbset) < dab) CYCLE

               ncoc = npgfcb(kbset)*ncoset(lcb_max(kbset))
               sgfc = first_sgfcb(1, kbset)
               m3 = sgfc+nsgfcb(kbset)-1
               IF (ncoa*ncob*ncoc > 0) THEN
                  ALLOCATE (sabb(ncoa, ncob, ncoc))
                  sabb(:, :, :) = 0._dp
                  IF (calculate_forces) THEN
                     ALLOCATE (sdabb(ncoa, ncob, ncoc, 3))
                     sdabb(:, :, :, :) = 0._dp
                     CALL overlap3(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                   lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                   lcb_max(kbset), npgfcb(kbset), zetcb(:, kbset), rpgfcb(:, kbset), lcb_min(kbset), &
                                   rab, dab, rac, dac, rbc, dbc, sabb, sdabc=sdabb)
                     DO i = 1, 3
                        CALL abc_contract(dabbint(sgfa:m1, sgfb:m2, sgfc:m3, i), sdabb(:, :, :, i), &
                                          sphi_a(:, sgfa:), sphi_b(:, sgfb:), sphi_cb(:, sgfc:), &
                                          ncoa, ncob, ncoc, nsgfa(iset), nsgfb(jset), nsgfcb(kbset))
                     ENDDO
                     DEALLOCATE (sdabb)
                  ELSE
                     CALL overlap3(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                   lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                   lcb_max(kbset), npgfcb(kbset), zetcb(:, kbset), rpgfcb(:, kbset), lcb_min(kbset), &
                                   rab, dab, rac, dac, rbc, dbc, sabb)
                  ENDIF
                  ! debug if requested
                  IF (debug) THEN
                     CALL overlap_abc_test(la_max(iset), npgfa(iset), zeta(:, iset), la_min(iset), &
                                           lb_max(jset), npgfb(jset), zetb(:, jset), lb_min(jset), &
                                           lcb_max(kbset), npgfcb(kbset), zetcb(:, kbset), lcb_min(kbset), &
                                           ra, rb, rb, sabb, dmax)
                  ENDIF
                  CALL abc_contract(abbint(sgfa:m1, sgfb:m2, sgfc:m3), sabb, &
                                    sphi_a(:, sgfa:), sphi_b(:, sgfb:), sphi_cb(:, sgfc:), &
                                    ncoa, ncob, ncoc, nsgfa(iset), nsgfb(jset), nsgfcb(kbset))
                  DEALLOCATE (sabb)
               ENDIF
            END DO

         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE lri_int_abb

! **************************************************************************************************
!> \brief calculate overlap integrals (aa,bb)
!> \param saabb integral (aa,bb)
!> \param oba orbital basis at center A
!> \param obb orbital basis at center B
!> \param rab ...
!> \param ra ...
!> \param rb ...
!> \param debug integrals are debugged by recursive routines if requested
!> \param dmax maximal deviation between integrals when debugging
! **************************************************************************************************
   SUBROUTINE lri_int_aabb(saabb, oba, obb, rab, ra, rb, debug, dmax)

      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: saabb
      TYPE(gto_basis_set_type), POINTER                  :: oba, obb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab, ra, rb
      LOGICAL, INTENT(IN)                                :: debug
      REAL(KIND=dp), INTENT(INOUT)                       :: dmax

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_aabb', routineP = moduleN//':'//routineN

      INTEGER :: handle, iset, isgfa1, jset, jsgfa2, kset, ksgfb1, lds, lset, lsgfb2, m1, m2, m3, &
         m4, maxco, maxcoa, maxcob, maxl, maxla, maxlb, ncoa1, ncoa2, ncob1, ncob2, nseta, nsetb, &
         sgfa1, sgfa2, sgfb1, sgfb2
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: asets_equal, bsets_equal
      REAL(KIND=dp)                                      :: dab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: swork
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: sint
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb

      CALL timeset(routineN, handle)
      NULLIFY (la_max, la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb, &
               first_sgfa, first_sgfb, set_radius_a, set_radius_b, rpgfa, rpgfb, &
               sphi_a, sphi_b, zeta, zetb)

      ! basis ikind
      first_sgfa => oba%first_sgf
      la_max => oba%lmax
      la_min => oba%lmin
      npgfa => oba%npgf
      nseta = oba%nset
      nsgfa => oba%nsgf_set
      rpgfa => oba%pgf_radius
      set_radius_a => oba%set_radius
      sphi_a => oba%sphi
      zeta => oba%zet
      ! basis jkind
      first_sgfb => obb%first_sgf
      lb_max => obb%lmax
      lb_min => obb%lmin
      npgfb => obb%npgf
      nsetb = obb%nset
      nsgfb => obb%nsgf_set
      rpgfb => obb%pgf_radius
      set_radius_b => obb%set_radius
      sphi_b => obb%sphi
      zetb => obb%zet

      CALL get_gto_basis_set(oba, maxco=maxcoa, maxl=maxla)
      CALL get_gto_basis_set(obb, maxco=maxcob, maxl=maxlb)
      maxco = MAX(maxcoa, maxcob)
      maxla = 2*maxla
      maxlb = 2*maxlb
      maxl = MAX(maxla, maxlb)
      lds = ncoset(maxl)
      ALLOCATE (sint(maxco, maxco, maxco, maxco))
      ALLOCATE (swork(lds, lds))
      sint = 0._dp
      swork = 0._dp

      dab = SQRT(SUM(rab**2))

      DO iset = 1, nseta

         ncoa1 = npgfa(iset)*ncoset(la_max(iset))
         sgfa1 = first_sgfa(1, iset)
         m1 = sgfa1+nsgfa(iset)-1

         DO jset = iset, nseta

            ncoa2 = npgfa(jset)*ncoset(la_max(jset))
            sgfa2 = first_sgfa(1, jset)
            m2 = sgfa2+nsgfa(jset)-1

            DO kset = 1, nsetb

               ncob1 = npgfb(kset)*ncoset(lb_max(kset))
               sgfb1 = first_sgfb(1, kset)
               m3 = sgfb1+nsgfb(kset)-1

               DO lset = kset, nsetb

                  ncob2 = npgfb(lset)*ncoset(lb_max(lset))
                  sgfb2 = first_sgfb(1, lset)
                  m4 = sgfb2+nsgfb(lset)-1

                  ! check if sets are identical to spare some integral evaluation
                  asets_equal = .FALSE.
                  IF (iset == jset) asets_equal = .TRUE.
                  bsets_equal = .FALSE.
                  IF (kset == lset) bsets_equal = .TRUE.
                  ! calculate integrals
                  CALL overlap_aabb(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                                    la_max(jset), la_min(jset), npgfa(jset), rpgfa(:, jset), zeta(:, jset), &
                                    lb_max(kset), lb_min(kset), npgfb(kset), rpgfb(:, kset), zetb(:, kset), &
                                    lb_max(lset), lb_min(lset), npgfb(lset), rpgfb(:, lset), zetb(:, lset), &
                                    asets_equal, bsets_equal, rab, dab, sint, swork, lds)
                  ! debug if requested
                  IF (debug) THEN
                     CALL overlap_aabb_test(la_max(iset), la_min(iset), npgfa(iset), zeta(:, iset), &
                                            la_max(jset), la_min(jset), npgfa(jset), zeta(:, jset), &
                                            lb_max(kset), lb_min(kset), npgfb(kset), zetb(:, kset), &
                                            lb_max(lset), lb_min(lset), npgfb(lset), zetb(:, lset), &
                                            ra, rb, sint, dmax)
                  ENDIF

                  CALL abcd_contract(saabb(sgfa1:m1, sgfa2:m2, sgfb1:m3, sgfb2:m4), sint, sphi_a(:, sgfa1:), &
                                     sphi_a(:, sgfa2:), sphi_b(:, sgfb1:), sphi_b(:, sgfb2:), ncoa1, ncoa2, &
                                     ncob1, ncob2, nsgfa(iset), nsgfa(jset), nsgfb(kset), nsgfb(lset))

                  ! account for the fact that some integrals are alike
                  DO isgfa1 = sgfa1, m1
                     DO jsgfa2 = sgfa2, m2
                        DO ksgfb1 = sgfb1, m3
                           DO lsgfb2 = sgfb2, m4
                              saabb(jsgfa2, isgfa1, ksgfb1, lsgfb2) = saabb(isgfa1, jsgfa2, ksgfb1, lsgfb2)
                              saabb(isgfa1, jsgfa2, lsgfb2, ksgfb1) = saabb(isgfa1, jsgfa2, ksgfb1, lsgfb2)
                              saabb(jsgfa2, isgfa1, lsgfb2, ksgfb1) = saabb(isgfa1, jsgfa2, ksgfb1, lsgfb2)
                           END DO
                        END DO
                     END DO
                  END DO

               END DO
            END DO
         END DO
      END DO

      DEALLOCATE (sint, swork)

      CALL timestop(handle)

   END SUBROUTINE lri_int_aabb

! **************************************************************************************************
!> \brief contract overlap integrals (a,b) and transfer to spherical Gaussians
!> \param abint ...
!> \param sab ...
!> \param sphi_a ...
!> \param sphi_b ...
!> \param ncoa ...
!> \param ncob ...
!> \param nsgfa ...
!> \param nsgfb ...
! **************************************************************************************************
   SUBROUTINE ab_contract(abint, sab, sphi_a, sphi_b, ncoa, ncob, nsgfa, nsgfb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: abint
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: sab, sphi_a, sphi_b
      INTEGER, INTENT(IN)                                :: ncoa, ncob, nsgfa, nsgfb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ab_contract', routineP = moduleN//':'//routineN

      INTEGER                                            :: m1, m2, msphia, msphib, nn
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cpp

      msphia = SIZE(sphi_a, 1)
      msphib = SIZE(sphi_b, 1)

      m1 = SIZE(sab, 1)
      m2 = SIZE(sab, 2)

      nn = SIZE(abint, 1)

      ALLOCATE (cpp(nsgfa, m2))

      CALL dgemm("T", "N", nsgfa, m2, ncoa, 1._dp, sphi_a, msphia, sab, m1, 0.0_dp, cpp, nsgfa)
      CALL dgemm("N", "N", nsgfa, nsgfb, ncob, 1._dp, cpp, nsgfa, sphi_b, msphib, 0.0_dp, &
                 abint, nn)

      DEALLOCATE (cpp)

   END SUBROUTINE ab_contract

! **************************************************************************************************
!> \brief contract three-center overlap integrals (a,b,c) and transfer
!>        to spherical Gaussians
!> \param abcint ...
!> \param sabc ...
!> \param sphi_a ...
!> \param sphi_b ...
!> \param sphi_c ...
!> \param ncoa ...
!> \param ncob ...
!> \param ncoc ...
!> \param nsgfa ...
!> \param nsgfb ...
!> \param nsgfc ...
! **************************************************************************************************
   SUBROUTINE abc_contract(abcint, sabc, sphi_a, sphi_b, sphi_c, ncoa, ncob, ncoc, &
                           nsgfa, nsgfb, nsgfc)

      REAL(KIND=dp), DIMENSION(:, :, :)                  :: abcint, sabc
      REAL(KIND=dp), DIMENSION(:, :)                     :: sphi_a, sphi_b, sphi_c
      INTEGER, INTENT(IN)                                :: ncoa, ncob, ncoc, nsgfa, nsgfb, nsgfc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'abc_contract', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, isgfc, m1, m2, m3, msphia, &
                                                            msphib, msphic
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: temp_ccc, work_cpc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: cpc, cpp

      CALL timeset(routineN, handle)

      msphia = SIZE(sphi_a, 1)
      msphib = SIZE(sphi_b, 1)
      msphic = SIZE(sphi_c, 1)

      m1 = SIZE(sabc, 1)
      m2 = SIZE(sabc, 2)
      m3 = SIZE(sabc, 3)

      ALLOCATE (cpp(nsgfa, m2, m3), cpc(nsgfa, m2, nsgfc))
      cpp = 0._dp
      cpc = 0._dp
      ALLOCATE (work_cpc(nsgfa, m2), temp_ccc(nsgfa, nsgfb))
      work_cpc(:, :) = 0._dp
      temp_ccc(:, :) = 0._dp

      CALL dgemm("T", "N", nsgfa, m2*m3, ncoa, 1._dp, sphi_a, msphia, sabc, m1, 0.0_dp, cpp, nsgfa)
      CALL dgemm("N", "N", nsgfa*m2, nsgfc, ncoc, 1._dp, cpp, nsgfa*m2, sphi_c, msphic, 0.0_dp, &
                 cpc, nsgfa*m2)

      DO isgfc = 1, nsgfc
         work_cpc(:, :) = cpc(:, :, isgfc)
         CALL dgemm("N", "N", nsgfa, nsgfb, ncob, 1._dp, work_cpc, nsgfa, sphi_b, msphib, &
                    0.0_dp, temp_ccc, nsgfa)
         abcint(:, :, isgfc) = temp_ccc(:, :)
      END DO

      DEALLOCATE (cpp, cpc, work_cpc, temp_ccc)

      CALL timestop(handle)

   END SUBROUTINE abc_contract

! **************************************************************************************************
!> \brief contract four-center overlap integrals (a,b,c,d) and transfer
!>        to spherical Gaussians
!> \param abcdint ...
!> \param sabcd ...
!> \param sphi_a ...
!> \param sphi_b ...
!> \param sphi_c ...
!> \param sphi_d ...
!> \param ncoa ...
!> \param ncob ...
!> \param ncoc ...
!> \param ncod ...
!> \param nsgfa ...
!> \param nsgfb ...
!> \param nsgfc ...
!> \param nsgfd ...
! **************************************************************************************************
   SUBROUTINE abcd_contract(abcdint, sabcd, sphi_a, sphi_b, sphi_c, sphi_d, ncoa, ncob, &
                            ncoc, ncod, nsgfa, nsgfb, nsgfc, nsgfd)

      REAL(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(INOUT)                                   :: abcdint
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: sabcd
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: sphi_a, sphi_b, sphi_c, sphi_d
      INTEGER, INTENT(IN)                                :: ncoa, ncob, ncoc, ncod, nsgfa, nsgfb, &
                                                            nsgfc, nsgfd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'abcd_contract', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, isgfc, isgfd, m1, m2, m3, m4, &
                                                            msphia, msphib, msphic, msphid
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: temp_cccc, work_cpcc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: temp_cpcc, work_cppc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: cpcc, cppc, cppp

      CALL timeset(routineN, handle)

      msphia = SIZE(sphi_a, 1)
      msphib = SIZE(sphi_b, 1)
      msphic = SIZE(sphi_c, 1)
      msphid = SIZE(sphi_d, 1)

      m1 = SIZE(sabcd, 1)
      m2 = SIZE(sabcd, 2)
      m3 = SIZE(sabcd, 3)
      m4 = SIZE(sabcd, 4)

      ALLOCATE (cppp(nsgfa, m2, m3, m4), cppc(nsgfa, m2, m3, nsgfd), &
                cpcc(nsgfa, m2, nsgfc, nsgfd))

      ALLOCATE (work_cppc(nsgfa, m2, m3), temp_cpcc(nsgfa, m2, nsgfc))
      work_cppc = 0._dp
      temp_cpcc = 0._dp

      ALLOCATE (work_cpcc(nsgfa, m2), temp_cccc(nsgfa, nsgfb))
      work_cpcc = 0._dp
      temp_cccc = 0._dp

      CALL dgemm("T", "N", nsgfa, m2*m3*m4, ncoa, 1._dp, sphi_a, msphia, sabcd, m1, &
                 0.0_dp, cppp, nsgfa)
      CALL dgemm("N", "N", nsgfa*m2*m3, nsgfd, ncod, 1._dp, cppp, nsgfa*m2*m3, &
                 sphi_d, msphid, 0.0_dp, cppc, nsgfa*m2*m3)

      DO isgfd = 1, nsgfd
         work_cppc(:, :, :) = cppc(:, :, :, isgfd)
         CALL dgemm("N", "N", nsgfa*m2, nsgfc, ncoc, 1._dp, work_cppc, nsgfa*m2, &
                    sphi_c, msphic, 0.0_dp, temp_cpcc, nsgfa*m2)
         cpcc(:, :, :, isgfd) = temp_cpcc(:, :, :)
         DO isgfc = 1, nsgfc
            work_cpcc(:, :) = cpcc(:, :, isgfc, isgfd)
            CALL dgemm("N", "N", nsgfa, nsgfb, ncob, 1._dp, work_cpcc, nsgfa, sphi_b, &
                       msphib, 0.0_dp, temp_cccc, nsgfa)
            abcdint(:, :, isgfc, isgfd) = temp_cccc(:, :)
         END DO
      END DO

      DEALLOCATE (cpcc, cppc, cppp)
      DEALLOCATE (work_cpcc, work_cppc, temp_cpcc, temp_cccc)

      CALL timestop(handle)

   END SUBROUTINE abcd_contract

END MODULE lri_os_integrals
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates integral matrices for LRIGPW method
!>        lri : local resolution of the identity
!> \par History
!>      created JGH [08.2012]
!>      Dorothea Golze [02.2014] (1) extended, re-structured, cleaned
!>                               (2) heavily debugged
!> \authors JGH
!>          Dorothea Golze
! **************************************************************************************************
MODULE lri_environment_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_lri_inv,&
                                              do_lri_inv_auto,&
                                              do_lri_pseudoinv_diag,&
                                              do_lri_pseudoinv_svd
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE lri_environment_types,           ONLY: &
        allocate_lri_coefs, allocate_lri_ints, allocate_lri_ints_rho, allocate_lri_rhos, &
        deallocate_lri_ints, deallocate_lri_ints_rho, lri_density_create, lri_density_release, &
        lri_density_type, lri_environment_type, lri_int_rho_type, lri_int_type, lri_kind_type, &
        lri_list_type, lri_rhoab_type
   USE lri_os_integrals,                ONLY: lri_int_aabb,&
                                              lri_int_ab,&
                                              lri_int_aba,&
                                              lri_int_abb
   USE lri_shg_integrals,               ONLY: lri_int_ab_shg,&
                                              lri_int_aba_shg,&
                                              lri_int_abb_shg,&
                                              precalc_angular_shg_part
   USE mathlib,                         ONLY: get_pseudo_inverse_diag,&
                                              get_pseudo_inverse_svd,&
                                              invmat_symm
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_collocate_density,            ONLY: calculate_lri_rho_elec
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_environment_methods'

   PUBLIC :: build_lri_matrices, calculate_lri_densities, calculate_lri_integrals, &
             calculate_lri_overlap_aabb, calculate_avec

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief creates and initializes an lri_env
!> \param lri_env the lri_environment you want to create
!> \param qs_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE build_lri_matrices(lri_env, qs_env, calculate_forces)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_lri_matrices', &
         routineP = moduleN//':'//routineN

! calculate the integrals needed to do the local (2-center) expansion
! of the (pair) densities

      CALL calculate_lri_integrals(lri_env, qs_env, calculate_forces)

   END SUBROUTINE build_lri_matrices

! **************************************************************************************************
!> \brief calculates integrals needed for the LRI density fitting,
!>        integrals are calculated once, before the SCF starts
!> \param lri_env ...
!> \param qs_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE calculate_lri_integrals(lri_env, qs_env, calculate_forces)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_lri_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iac, iatom, ikind, ilist, jatom, &
                                                            jkind, jneighbor, nba, nbb, nfa, nfb, &
                                                            nkind, nlist, nn, nneighbor
      REAL(KIND=dp)                                      :: dab
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rb
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: fbasa, fbasb, obasa, obasb
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(lri_list_type), POINTER                       :: lri_ints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, fbasa, fbasb, lrii, lri_ints, nl_iterator, &
               obasa, obasb, particle_set, soo_list, virial)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         CALL get_qs_env(qs_env=qs_env, cell=cell, dft_control=dft_control, &
                         nkind=nkind, particle_set=particle_set, virial=virial)

         IF (ASSOCIATED(lri_env%lri_ints)) THEN
            CALL deallocate_lri_ints(lri_env%lri_ints)
         END IF

         ! allocate matrices storing the LRI integrals
         CALL allocate_lri_ints(lri_env, lri_env%lri_ints, nkind, &
                                calculate_forces, virial)
         lri_ints => lri_env%lri_ints

         CALL neighbor_list_iterator_create(nl_iterator, soo_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)

            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                   iatom=iatom, jatom=jatom, r=rab)

            iac = ikind+nkind*(jkind-1)
            dab = SQRT(SUM(rab*rab))

            obasa => lri_env%orb_basis(ikind)%gto_basis_set
            obasb => lri_env%orb_basis(jkind)%gto_basis_set
            fbasa => lri_env%ri_basis(ikind)%gto_basis_set
            fbasb => lri_env%ri_basis(jkind)%gto_basis_set

            IF (.NOT. ASSOCIATED(obasa)) CYCLE
            IF (.NOT. ASSOCIATED(obasb)) CYCLE

            lrii => lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

            nba = obasa%nsgf
            nbb = obasb%nsgf
            nfa = fbasa%nsgf
            nfb = fbasb%nsgf

            lrii%nba = nba
            lrii%nbb = nbb
            lrii%nfa = nfa
            lrii%nfb = nfb

            ra(:) = pbc(particle_set(iatom)%r, cell)
            rb(:) = pbc(particle_set(jatom)%r, cell)

            ! calculate integrals (fa,fb), (a,b), (a,b,fa) and (a,b,fb)
            IF (lri_env%use_shg_integrals) THEN
               CALL lri_int_shg(lri_env, lrii, ra, rb, rab, obasa, obasb, fbasa, fbasb, &
                                iatom, jatom, ikind, jkind)
            ELSE
               CALL lri_int_os(lri_env, lrii, ra, rb, rab, obasa, obasb, fbasa, fbasb, &
                               iatom, jatom, ikind)
            ENDIF

            ! construct and invert S matrix
            lrii%sinv(1:nfa, 1:nfa) = lri_env%bas_prop(ikind)%ri_ovlp(1:nfa, 1:nfa)
            IF (dab > lri_env%delta) THEN
               nn = nfa+nfb
               lrii%sinv(1:nfa, nfa+1:nn) = lrii%sab(1:nfa, 1:nfb)
               lrii%sinv(nfa+1:nn, 1:nfa) = TRANSPOSE(lrii%sab(1:nfa, 1:nfb))
               lrii%sinv(nfa+1:nn, nfa+1:nn) = lri_env%bas_prop(jkind)%ri_ovlp(1:nfb, 1:nfb)
            ENDIF
            CALL inverse_lri_overlap(lri_env, lrii%sinv)

            ! calculate Sinv*n and n*Sinv*n
            lrii%n(1:nfa) = lri_env%bas_prop(ikind)%int_fbas(1:nfa)
            IF (iatom == jatom .AND. dab < lri_env%delta) THEN
               lrii%sn(1:nfa) = MATMUL(lrii%sinv(1:nfa, 1:nfa), lrii%n(1:nfa))
               lrii%nsn = SUM(lrii%sn(1:nfa)*lrii%n(1:nfa))
            ELSE
               lrii%n(nfa+1:nn) = lri_env%bas_prop(jkind)%int_fbas(1:nfb)
               lrii%sn(1:nn) = MATMUL(lrii%sinv(1:nn, 1:nn), lrii%n(1:nn))
               lrii%nsn = SUM(lrii%sn(1:nn)*lrii%n(1:nn))
            ENDIF

            ! calculate derivative of fit coefficients, needed for update of KS matrix
            IF (.NOT. dft_control%qs_control%lri_optbas) THEN
               CALL lri_calculate_derivative_acoef(lri_env, lrii, iatom, jatom, nba, nbb, &
                                                   nfa, nfb, dab)
            ENDIF

         END DO

         CALL neighbor_list_iterator_release(nl_iterator)

         IF (lri_env%debug) THEN
            CALL output_debug_info(lri_env, qs_env, lri_ints, soo_list)
         ENDIF

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_lri_integrals

! **************************************************************************************************
!> \brief calculates overlap integrals (aabb) of the orbital basis set,
!>        reguired for LRI basis set optimization
!> \param lri_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE calculate_lri_overlap_aabb(lri_env, qs_env)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_lri_overlap_aabb', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iac, iatom, ikind, ilist, jatom, &
                                                            jkind, jneighbor, nba, nbb, nkind, &
                                                            nlist, nneighbor
      REAL(KIND=dp)                                      :: dab
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rb
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_type), POINTER                  :: obasa, obasb
      TYPE(lri_int_rho_type), POINTER                    :: lriir
      TYPE(lri_list_type), POINTER                       :: lri_ints_rho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      NULLIFY (cell, lriir, lri_ints_rho, nl_iterator, obasa, obasb, &
               particle_set, soo_list)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         CALL get_qs_env(qs_env=qs_env, nkind=nkind, particle_set=particle_set, &
                         cell=cell)

         IF (ASSOCIATED(lri_env%lri_ints_rho)) THEN
            CALL deallocate_lri_ints_rho(lri_env%lri_ints_rho)
         END IF

         CALL allocate_lri_ints_rho(lri_env, lri_env%lri_ints_rho, nkind)
         lri_ints_rho => lri_env%lri_ints_rho

         CALL neighbor_list_iterator_create(nl_iterator, soo_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)

            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                   iatom=iatom, jatom=jatom, r=rab)

            iac = ikind+nkind*(jkind-1)
            dab = SQRT(SUM(rab*rab))

            obasa => lri_env%orb_basis(ikind)%gto_basis_set
            obasb => lri_env%orb_basis(jkind)%gto_basis_set
            IF (.NOT. ASSOCIATED(obasa)) CYCLE
            IF (.NOT. ASSOCIATED(obasb)) CYCLE

            lriir => lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho(jneighbor)

            nba = obasa%nsgf
            nbb = obasb%nsgf
            ra(:) = pbc(particle_set(iatom)%r, cell)
            rb(:) = pbc(particle_set(jatom)%r, cell)

            ! calculate integrals (aa,bb)
            CALL lri_int_aabb(lriir%soaabb, obasa, obasb, rab, ra, rb, lri_env%debug, &
                              lriir%dmax_aabb)

         END DO

         CALL neighbor_list_iterator_release(nl_iterator)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calculate_lri_overlap_aabb

! **************************************************************************************************
!> \brief performs the fitting of the density and distributes the fitted
!>        density on the grid
!> \param lri_env the lri environment
!>        lri_density the environment for the fitting
!>        pmatrix density matrix
!>        lri_rho_struct where the fitted density is stored
!> \param lri_density ...
!> \param qs_env ...
!> \param pmatrix ...
!> \param lri_rho_struct ...
!> \param atomic_kind_set ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE calculate_lri_densities(lri_env, lri_density, qs_env, pmatrix, &
                                      lri_rho_struct, atomic_kind_set, para_env)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(qs_rho_type), POINTER                         :: lri_rho_struct
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_lri_densities', &
         routineP = moduleN//':'//routineN

      CALL calculate_avec(lri_env, lri_density, qs_env, pmatrix)

      CALL distribute_lri_density_on_the_grid(lri_env, lri_density, qs_env, &
                                              lri_rho_struct, atomic_kind_set, para_env)

   END SUBROUTINE calculate_lri_densities

! **************************************************************************************************
!> \brief performs the fitting of the density; solves the linear system of
!>        equations; yield the expansion coefficients avec
!> \param lri_env the lri environment
!>        lri_density the environment for the fitting
!>        pmatrix density matrix
!> \param lri_density ...
!> \param qs_env ...
!> \param pmatrix ...
! **************************************************************************************************
   SUBROUTINE calculate_avec(lri_env, lri_density, qs_env, pmatrix)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_avec', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iac, iatom, ikind, ilist, &
                                                            ispin, jatom, jkind, jneighbor, nba, &
                                                            nbb, nfa, nfb, nkind, nlist, nn, &
                                                            nneighbor, nspin
      LOGICAL                                            :: found, trans
      REAL(KIND=dp)                                      :: dab, rab(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: m
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pbij
      TYPE(cp_dbcsr_type), POINTER                       :: pmat
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(lri_list_type), POINTER                       :: lri_rho
      TYPE(lri_rhoab_type), POINTER                      :: lrho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list

      CALL timeset(routineN, handle)
      NULLIFY (lrii, lri_rho, nl_iterator, pbij, pmat, soo_list)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         nspin = SIZE(pmatrix)
         nkind = lri_env%lri_ints%nkind

         CALL lri_density_release(lri_density)
         CALL lri_density_create(lri_density)
         lri_density%nspin = nspin

         ! allocate structure lri_rhos and vectors tvec and avec
         CALL allocate_lri_rhos(lri_env, lri_density%lri_rhos, nspin, nkind)

         DO ispin = 1, nspin
            pmat => pmatrix(ispin)%matrix
            lri_rho => lri_density%lri_rhos(ispin)%lri_list

            CALL neighbor_list_iterator_create(nl_iterator, soo_list)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, &
                                      jatom=jatom, nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                      r=rab)

               iac = ikind+nkind*(jkind-1)
               dab = SQRT(SUM(rab*rab))

               IF (.NOT. ASSOCIATED(lri_env%lri_ints%lri_atom(iac)%lri_node)) CYCLE

               ! get the density matrix Pab
               NULLIFY (pbij)
               IF (iatom <= jatom) THEN
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=iatom, col=jatom, block=pbij, found=found)
                  trans = .FALSE.
               ELSE
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=jatom, col=iatom, block=pbij, found=found)
                  trans = .TRUE.
               END IF
               IF(.NOT.(found))CALL cp__a("lri_environment_methods.F",438)

               lrho => lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(jneighbor)
               lrii => lri_env%lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

               nba = lrii%nba
               nbb = lrii%nbb
               nfa = lrii%nfa
               nfb = lrii%nfb

               nn = nfa+nfb

               ! compute tvec = SUM_ab Pab *(a,b,x) and charge contraint
               IF (trans) THEN
                  lrho%charge = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%soo(1:nba, 1:nbb))
                  DO i = 1, nfa
                     lrho%tvec(i) = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%abaint(1:nba, 1:nbb, i))
                  END DO
                  IF (dab > lri_env%delta) THEN
                     DO i = 1, nfb
                        lrho%tvec(nfa+i) = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%abbint(1:nba, 1:nbb, i))
                     END DO
                  ENDIF
               ELSE
                  lrho%charge = SUM(pbij(1:nba, 1:nbb)*lrii%soo(1:nba, 1:nbb))
                  DO i = 1, nfa
                     lrho%tvec(i) = SUM(pbij(1:nba, 1:nbb)*lrii%abaint(1:nba, 1:nbb, i))
                  END DO
                  IF (dab > lri_env%delta) THEN
                     DO i = 1, nfb
                        lrho%tvec(nfa+i) = SUM(pbij(1:nba, 1:nbb)*lrii%abbint(1:nba, 1:nbb, i))
                     END DO
                  ENDIF
               END IF

               IF (iatom == jatom .AND. dab < lri_env%delta) THEN
                  lrho%nst = SUM(lrho%tvec(1:nfa)*lrii%sn(1:nfa))
               ELSE
                  lrho%nst = SUM(lrho%tvec(1:nn)*lrii%sn(1:nn))
               ENDIF
               lrho%lambda = (lrho%charge-lrho%nst)/lrii%nsn

               ! solve the linear system of equations
               ALLOCATE (m(nn))
               m = 0._dp
               IF (iatom == jatom .AND. dab < lri_env%delta) THEN
                  m(1:nfa) = lrho%tvec(1:nfa)+lrho%lambda*lrii%n(1:nfa)
                  CALL dgemv("N", nfa, nfa, 1.0_dp, lrii%sinv(1, 1), nfa, &
                             m(1), 1, 0.0_dp, lrho%avec, 1)
               ELSE
                  m(1:nn) = lrho%tvec(1:nn)+lrho%lambda*lrii%n(1:nn)
                  CALL dgemv("N", nn, nn, 1.0_dp, lrii%sinv(1, 1), nn, &
                             m(1), 1, 0.0_dp, lrho%avec, 1)
               ENDIF
               DEALLOCATE (m)

            END DO
            CALL neighbor_list_iterator_release(nl_iterator)

         END DO

         CALL set_qs_env(qs_env, lri_density=lri_density)

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_avec

! **************************************************************************************************
!> \brief sums up avec and  distributes the fitted density on the grid
!> \param lri_env the lri environment
!>        lri_density the environment for the fitting
!>        pmatrix density matrix
!>        lri_rho_struct where the fitted density is stored
!> \param lri_density ...
!> \param qs_env ...
!> \param lri_rho_struct ...
!> \param atomic_kind_set ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE distribute_lri_density_on_the_grid(lri_env, lri_density, qs_env, &
                                                 lri_rho_struct, atomic_kind_set, para_env)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: lri_rho_struct
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'distribute_lri_density_on_the_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, iac, iatom, &
                                                            ikind, ilist, ispin, jatom, jkind, &
                                                            jneighbor, nat, natom, nfa, nfb, &
                                                            nkind, nspin
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      REAL(KIND=dp)                                      :: dab, rab(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: aci, acj, tot_rho_r
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: lri_coef
      TYPE(lri_list_type), POINTER                       :: lri_rho
      TYPE(lri_rhoab_type), POINTER                      :: lrho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r

      CALL timeset(routineN, handle)
      NULLIFY (aci, acj, atomic_kind, atom_of_kind, lri_coef, lri_rho, &
               nl_iterator, soo_list, rho_r, rho_g, tot_rho_r)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         nspin = lri_density%nspin
         nkind = lri_env%lri_ints%nkind

         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, natom=nat)
         ALLOCATE (atom_of_kind(nat))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  atom_of_kind=atom_of_kind)

         ! allocate the arrays to hold RI expansion coefficients lri_coefs
         CALL allocate_lri_coefs(lri_env, lri_density, atomic_kind_set)
         DO ispin = 1, nspin

            lri_coef => lri_density%lri_coefs(ispin)%lri_kinds
            lri_rho => lri_density%lri_rhos(ispin)%lri_list

            ! sum up expansion coefficients
            CALL neighbor_list_iterator_create(nl_iterator, soo_list)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                      iatom=iatom, jatom=jatom, ilist=ilist, inode=jneighbor, r=rab)
               dab = SQRT(SUM(rab*rab))
               atom_a = atom_of_kind(iatom)
               atom_b = atom_of_kind(jatom)
               aci => lri_coef(ikind)%acoef(atom_a, :)
               acj => lri_coef(jkind)%acoef(atom_b, :)
               iac = ikind+nkind*(jkind-1)
               lrho => lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(jneighbor)
               nfa = lrho%nfa
               nfb = lrho%nfb
               IF (iatom == jatom .AND. dab < lri_env%delta) THEN
                  !self pair aa
                  aci(1:nfa) = aci(1:nfa)+lrho%avec(1:nfa)
               ELSE
                  IF (iatom == jatom) THEN
                     !periodic self pair aa'
                     aci(1:nfa) = aci(1:nfa)+lrho%avec(1:nfa)
                     acj(1:nfb) = acj(1:nfb)+lrho%avec(nfa+1:nfa+nfb)
                  ELSE
                     !pairs ab
                     aci(1:nfa) = aci(1:nfa)+2.0_dp*lrho%avec(1:nfa)
                     acj(1:nfb) = acj(1:nfb)+2.0_dp*lrho%avec(nfa+1:nfa+nfb)
                  ENDIF
               ENDIF
            END DO
            CALL neighbor_list_iterator_release(nl_iterator)

            ! replicate the acoef infomation
            DO ikind = 1, nkind
               atomic_kind => atomic_kind_set(ikind)
               CALL get_atomic_kind(atomic_kind=atomic_kind, natom=natom)
               DO iatom = 1, natom
                  aci => lri_coef(ikind)%acoef(iatom, :)
                  CALL mp_sum(aci, para_env%group)
               END DO
            END DO

         END DO

         !distribute fitted density on the grid
         CALL qs_rho_get(lri_rho_struct, rho_r=rho_r, rho_g=rho_g, &
                         tot_rho_r=tot_rho_r)
         DO ispin = 1, nspin
            CALL calculate_lri_rho_elec(rho_g(ispin), &
                                        rho_r(ispin), qs_env, &
                                        lri_density%lri_coefs(ispin)%lri_kinds, &
                                        tot_rho_r(ispin))
         ENDDO

         CALL set_qs_env(qs_env, lri_density=lri_density)

         DEALLOCATE (atom_of_kind)

      END IF

      CALL timestop(handle)

   END SUBROUTINE distribute_lri_density_on_the_grid

! **************************************************************************************************
!> \brief calculates the lri intergrals according to the Obara-Saika (OS)
!>        scheme
!> \param lri_env ...
!> \param lrii ...
!> \param ra position atom A
!> \param rb position atom B
!> \param rab distance vector
!> \param obasa orb basis on center A
!> \param obasb orb basis on center B
!> \param fbasa aux basis on center A
!> \param fbasb aux basis on center B
!> \param iatom index atom A
!> \param jatom index atom B
!> \param ikind kind atom A
! **************************************************************************************************
   SUBROUTINE lri_int_os(lri_env, lrii, ra, rb, rab, obasa, obasb, fbasa, fbasb, &
                         iatom, jatom, ikind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_int_type), POINTER                        :: lrii
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rab
      TYPE(gto_basis_set_type), POINTER                  :: obasa, obasb, fbasa, fbasb
      INTEGER, INTENT(IN)                                :: iatom, jatom, ikind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_os', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nba, nbb, nfa, nfb
      REAL(KIND=dp)                                      :: dab

      CALL timeset(routineN, handle)

      dab = SQRT(SUM(rab*rab))
      nba = obasa%nsgf
      nbb = obasb%nsgf
      nfa = fbasa%nsgf
      nfb = fbasb%nsgf

      !*** calculate overlap integrals; for iatom=jatom this is the self-overlap
      IF (iatom == jatom .AND. dab < lri_env%delta) THEN
         !*** (fa,fa)
         lrii%sab(1:nfa, 1:nfa) = lri_env%bas_prop(ikind)%ri_ovlp(1:nfa, 1:nfa)
         lrii%dsab = 0._dp
         !*** (a,a)
         lrii%soo(1:nba, 1:nba) = lri_env%bas_prop(ikind)%orb_ovlp(1:nba, 1:nba)
         lrii%dsoo = 0._dp
         !*** (a,a,fa)
         CALL lri_int_aba(lrii%abaint, ra=ra, rb=rb, rab=rab, oba=obasa, obb=obasb, &
                          fba=fbasa, calculate_forces=.FALSE., debug=lri_env%debug, &
                          dmax=lrii%dmax_aba)
         lrii%dabdaint = 0.0_dp
         lrii%dabbint = 0.0_dp
      ELSE
         !*** (fa,fb)
         CALL lri_int_ab(lrii%sab, lrii%dsab, ra, rb, rab, fbasa, fbasb, &
                         lrii%calc_force_pair, lri_env%debug, &
                         lrii%dmax_ab)
         !*** (a,b)
         CALL lri_int_ab(lrii%soo, lrii%dsoo, ra, rb, rab, obasa, obasb, &
                         lrii%calc_force_pair, lri_env%debug, &
                         lrii%dmax_oo)
         !*** (a,b,fa)
         CALL lri_int_aba(lrii%abaint, lrii%dabdaint, ra, rb, rab, obasa, obasb, fbasa, &
                          lrii%calc_force_pair, lri_env%debug, lrii%dmax_aba)
         !*** (a,b,fb)
         CALL lri_int_abb(lrii%abbint, lrii%dabbint, ra, rb, rab, obasa, obasb, fbasb, &
                          lrii%calc_force_pair, lri_env%debug, lrii%dmax_abb)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE lri_int_os

! **************************************************************************************************
!> \brief calcuates the lri integrals using solid harmonic Gaussians
!> \param lri_env ...
!> \param lrii ...
!> \param ra position of center A
!> \param rb position of center B
!> \param rab distance vector
!> \param obasa orb basis on A
!> \param obasb orb basis on B
!> \param fbasa aux basis on A
!> \param fbasb aux basis on B
!> \param iatom index atom A
!> \param jatom index atom B
!> \param ikind kind atom A
!> \param jkind kind atom B
! **************************************************************************************************
   SUBROUTINE lri_int_shg(lri_env, lrii, ra, rb, rab, obasa, obasb, fbasa, fbasb, &
                          iatom, jatom, ikind, jkind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_int_type), POINTER                        :: lrii
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb, rab
      TYPE(gto_basis_set_type), POINTER                  :: obasa, obasb, fbasa, fbasb
      INTEGER, INTENT(IN)                                :: iatom, jatom, ikind, jkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_int_shg', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nba, nbb, nfa, nfb
      REAL(KIND=dp)                                      :: dab
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scon_orba, scon_orbb, scon_ria, scon_rib
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: Waux_mat
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: dWaux_mat

      CALL timeset(routineN, handle)
      NULLIFY (scon_orba, scon_orbb, scon_ria, scon_rib, Waux_mat, dWaux_mat)

      dab = SQRT(SUM(rab*rab))
      nba = obasa%nsgf
      nbb = obasb%nsgf
      nfa = fbasa%nsgf
      nfb = fbasb%nsgf

      !*** calculate overlap integrals; for iatom=jatom this is the self-overlap
      IF (iatom == jatom .AND. dab < lri_env%delta) THEN
         !*** (fa,fa)
         lrii%sab(1:nfa, 1:nfa) = lri_env%bas_prop(ikind)%ri_ovlp(1:nfa, 1:nfa)
         lrii%dsab = 0._dp
         !*** (a,a)
         lrii%soo(1:nba, 1:nba) = lri_env%bas_prop(ikind)%orb_ovlp(1:nba, 1:nba)
         lrii%dsoo = 0._dp
         !*** (a,a,fa)
         CALL lri_int_aba(lrii%abaint, ra=ra, rb=rb, rab=rab, oba=obasa, obb=obasb, &
                          fba=fbasa, calculate_forces=.FALSE., debug=lri_env%debug, &
                          dmax=lrii%dmax_aba)
         lrii%dabdaint = 0.0_dp
         lrii%dabbint = 0.0_dp
      ELSE
         scon_orba => lri_env%bas_prop(ikind)%scon_orb
         scon_orbb => lri_env%bas_prop(jkind)%scon_orb
         scon_ria => lri_env%bas_prop(ikind)%scon_ri
         scon_rib => lri_env%bas_prop(jkind)%scon_ri
         CALL precalc_angular_shg_part(obasa, obasb, fbasa, fbasb, rab, Waux_mat, dWaux_mat, &
                                       lrii%calc_force_pair)
         !*** (fa,fb)
         CALL lri_int_ab_shg(lrii%sab, lrii%dsab, rab, fbasa, fbasb, scon_ria, scon_rib, &
                             Waux_mat, dWaux_mat, lrii%calc_force_pair)
         !*** (a,b)
         CALL lri_int_ab_shg(lrii%soo, lrii%dsoo, rab, obasa, obasb, scon_orba, scon_orbb, &
                             Waux_mat, dWaux_mat, lrii%calc_force_pair)
         !*** (a,b,fa)
         CALL lri_int_aba_shg(lri_env, lrii%abaint, lrii%dabdaint, rab, obasa, obasb, fbasa, &
                              scon_orba, scon_orbb, scon_ria, Waux_mat, dWaux_mat, &
                              lrii%calc_force_pair)
         !*** (a,b,fb)
         CALL lri_int_abb_shg(lri_env, lrii%abbint, lrii%dabbint, rab, obasa, obasb, fbasb, &
                              scon_orba, scon_orbb, scon_rib, Waux_mat, dWaux_mat, &
                              lrii%calc_force_pair)
         DEALLOCATE (Waux_mat, dWaux_mat)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE lri_int_shg

! **************************************************************************************************
!> \brief calculate derivate of fit coefficients acoef with respect to
!>        density matrix pmatrix
!>        R = (a,b)/nsn  - SUM_i (a,b,ai)*sn(i)/nsn
!>        Q = SUM_i sinv*(a,b,ai)
!>        derviate_aci = R + Q
!> \param lri_env ...
!> \param lrii ...
!> \param iatom ...
!> \param jatom ...
!> \param nba number of primary basis functions on a
!> \param nbb number of primary basis functions on b
!> \param nfa number of ri basis functions on a
!> \param nfb number of ri basis functions on b
!> \param dab distance between center a and b
! **************************************************************************************************
   SUBROUTINE lri_calculate_derivative_acoef(lri_env, lrii, iatom, jatom, nba, nbb, &
                                             nfa, nfb, dab)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_int_type), POINTER                        :: lrii
      INTEGER, INTENT(IN)                                :: iatom, jatom, nba, nbb, nfa, nfb
      REAL(KIND=dp), INTENT(IN)                          :: dab

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_calculate_derivative_acoef', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nn
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dlambdadp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: abxint_work, Q, R

      CALL timeset(routineN, handle)

      nn = nfa+nfb

      ALLOCATE (R(nba, nbb, nn))
      ALLOCATE (dlambdadp(nba, nbb))
      ALLOCATE (Q(nba, nbb, nn))
      ALLOCATE (abxint_work(nba, nbb, nn))

      R(:, :, :) = 0._dp
      dlambdadp = 0._dp
      Q(:, :, :) = 0._dp
      abxint_work(:, :, :) = 0._dp

      abxint_work(1:nba, 1:nbb, 1:nfa) = lrii%abaint(1:nba, 1:nbb, 1:nfa) !abaint
      abxint_work(1:nba, 1:nbb, nfa+1:nn) = lrii%abbint(1:nba, 1:nbb, 1:nfb) !abbint
      IF (iatom == jatom .AND. dab < lri_env%delta) THEN
         ! R = Sinv*abx_int
         CALL dgemm("N", "N", nba*nbb, nfa, nfa, 1.0_dp, abxint_work, nba*nbb, &
                    lrii%sinv, nfa, 0.0_dp, R, nba*nbb)
         ! dlambdadp = soo/nsn - sn*abx_int/nsn
         dlambdadp(1:nba, 1:nbb) = lrii%soo(1:nba, 1:nbb)
         CALL dgemm("N", "N", nba*nbb, 1, nfa, -1.0_dp/lrii%nsn, abxint_work, nba*nbb, &
                    lrii%sn, nfa, 1.0_dp/lrii%nsn, dlambdadp, nba*nbb)
         ! Q = dlambdadp * sn
         CALL dgemm("N", "T", nba*nbb, nfa, 1, 1.0_dp, dlambdadp, nba*nbb, &
                    lrii%sn, nfa, 0.0_dp, Q, nba*nbb)
         lrii%dacoef(1:nba, 1:nbb, 1:nfa) = R(1:nba, 1:nbb, 1:nfa)+Q(1:nba, 1:nbb, 1:nfa)
      ELSE
         ! R = Sinv*abx_int
         CALL dgemm("N", "N", nba*nbb, nn, nn, 1.0_dp, abxint_work, nba*nbb, &
                    lrii%sinv, nn, 0.0_dp, R, nba*nbb)
         ! dlambdadp = soo/nsn - sn*abx_int/nsn
         dlambdadp(1:nba, 1:nbb) = lrii%soo(1:nba, 1:nbb)
         CALL dgemm("N", "N", nba*nbb, 1, nn, -1.0_dp/lrii%nsn, abxint_work, nba*nbb, &
                    lrii%sn, nn, 1.0_dp/lrii%nsn, dlambdadp, nba*nbb)
         ! Q = dlambdadp * sn
         CALL dgemm("N", "T", nba*nbb, nn, 1, 1.0_dp, dlambdadp, nba*nbb, &
                    lrii%sn, nn, 0.0_dp, Q, nba*nbb)
         lrii%dacoef(1:nba, 1:nbb, 1:nn) = R(1:nba, 1:nbb, 1:nn)+Q(1:nba, 1:nbb, 1:nn)
      ENDIF

      DEALLOCATE (abxint_work)
      DEALLOCATE (R)
      DEALLOCATE (dlambdadp)
      DEALLOCATE (Q)

      CALL timestop(handle)

   END SUBROUTINE lri_calculate_derivative_acoef

! **************************************************************************************************
!> \brief get inverse or pseudoinverse of lri overlap matrix for aux basis set
!> \param lri_env lri environment
!> \param sinv on entry overlap matrix, on exit its inverse
! **************************************************************************************************
   SUBROUTINE inverse_lri_overlap(lri_env, sinv)

      TYPE(lri_environment_type)                         :: lri_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: sinv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'inverse_lri_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      REAL(KIND=dp)                                      :: anorm, rcond, rskip
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: s
      REAL(KIND=dp), EXTERNAL                            :: dlange

      CALL timeset(routineN, handle)

      NULLIFY (s)

      rskip = 1.E-8_dp ! parameter for pseudo inverse
      n = SIZE(sinv, 1)

      SELECT CASE (lri_env%lri_overlap_inv)
      CASE (do_lri_inv)
         CALL invmat_symm(sinv)
      CASE (do_lri_pseudoinv_svd)
         ALLOCATE (s(n, n))
         s(:, :) = sinv
         CALL get_pseudo_inverse_svd(s, sinv, rskip)
         DEALLOCATE (s)
      CASE (do_lri_pseudoinv_diag)
         ALLOCATE (s(n, n))
         s(:, :) = sinv
         CALL get_pseudo_inverse_diag(s, sinv, rskip)
         DEALLOCATE (s)
      CASE (do_lri_inv_auto)
         ! decide whether to calculate inverse or pseudoinverse
         ALLOCATE (s(n, n))
         s(:, :) = sinv
         ALLOCATE (work(3*n), iwork(n))
         ! norm of matrix
         anorm = dlange('1', n, n, sinv, n, work)
         ! Cholesky factorization (fails if matrix not positive definite)
         CALL dpotrf('U', n, sinv, n, info)
         IF (info == 0) THEN
            ! condition number
            CALL dpocon('U', n, sinv, n, anorm, rcond, work, iwork, info)
            IF (info /= 0) THEN
               CALL cp__b("lri_environment_methods.F",926,"DPOCON failed")
            ENDIF
            IF (LOG(1._dp/rcond) > lri_env%cond_max) THEN
               CALL get_pseudo_inverse_svd(s, sinv, rskip)
            ELSE
               CALL invmat_symm(sinv, "U")
            ENDIF
         ELSE
            CALL get_pseudo_inverse_svd(s, sinv, rskip)
         END IF
         DEALLOCATE (s, work, iwork)
      CASE DEFAULT
         CALL cp__b("lri_environment_methods.F",938,"No initialization for LRI overlap available?")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE inverse_lri_overlap

! **************************************************************************************************
!> \brief debug output
!> \param lri_env ...
!> \param qs_env ...
!> \param lri_ints ...
!> \param soo_list ...
! **************************************************************************************************
   SUBROUTINE output_debug_info(lri_env, qs_env, lri_ints, soo_list)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(lri_list_type), POINTER                       :: lri_ints
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list

      CHARACTER(LEN=*), PARAMETER :: routineN = 'output_debug_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iac, ikind, ilist, iunit, jkind, &
                                                            jneighbor, nkind
      REAL(KIND=dp)                                      :: dmax_aabb, dmax_ab, dmax_aba, dmax_abb, &
                                                            dmax_oo
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(lri_int_rho_type), POINTER                    :: lriir
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)
      NULLIFY (input, logger, lrii, lriir, nl_iterator, para_env)
      CALL get_qs_env(qs_env, dft_control=dft_control, input=input, nkind=nkind, &
                      para_env=para_env)
      dmax_ab = 0._dp
      dmax_oo = 0._dp
      dmax_aba = 0._dp
      dmax_abb = 0._dp
      dmax_aabb = 0._dp

      CALL neighbor_list_iterator_create(nl_iterator, soo_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)

         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                ilist=ilist, inode=jneighbor)

         iac = ikind+nkind*(jkind-1)
         lrii => lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

         dmax_ab = MAX(dmax_ab, lrii%dmax_ab)
         dmax_oo = MAX(dmax_oo, lrii%dmax_oo)
         dmax_aba = MAX(dmax_aba, lrii%dmax_aba)
         dmax_abb = MAX(dmax_abb, lrii%dmax_abb)

         IF (dft_control%qs_control%lri_optbas) THEN
            lriir => lri_env%lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho(jneighbor)
            dmax_aabb = MAX(dmax_aabb, lriir%dmax_aabb)
         ENDIF

      END DO

      CALL neighbor_list_iterator_release(nl_iterator)
      CALL mp_max(dmax_ab, para_env%group)
      CALL mp_max(dmax_oo, para_env%group)
      CALL mp_max(dmax_aba, para_env%group)
      CALL mp_max(dmax_abb, para_env%group)
      CALL mp_max(dmax_aabb, para_env%group)

      logger => cp_get_default_logger()
      iunit = cp_print_key_unit_nr(logger, input, "PRINT%PROGRAM_RUN_INFO", &
                                   extension=".lridebug")

      IF (iunit > 0) THEN
         WRITE (iunit, FMT="(/,T2,A)") "DEBUG INFO FOR LRI INTEGRALS"
         WRITE (iunit, FMT="(T2,A,T69,ES12.5)") "Maximal deviation of integrals "// &
            "[ai|bi]; fit basis", dmax_ab
         WRITE (iunit, FMT="(T2,A,T69,ES12.5)") "Maximal deviation of integrals "// &
            "[a|b]; orbital basis", dmax_oo
         WRITE (iunit, FMT="(T2,A,T69,ES12.5)") "Maximal deviation of integrals "// &
            "[a|b|ai]", dmax_aba
         WRITE (iunit, FMT="(T2,A,T69,ES12.5)") "Maximal deviation of integrals "// &
            "[a|b|bi]", dmax_abb
         IF (dft_control%qs_control%lri_optbas) THEN
            WRITE (iunit, FMT="(T2,A,T69,ES12.5,/)") "Maximal deviation of integrals "// &
               "[aa|bb]; orbital basis", &
               dmax_aabb
         ENDIF
      ENDIF

      CALL cp_print_key_finished_output(iunit, logger, input, &
                                        "PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)

   END SUBROUTINE output_debug_info

END MODULE lri_environment_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Some utility functions for the calculation of integrals
!> \par History
!>      JGH: initial version
!> \author JGH (10.07.2014)
! **************************************************************************************************
MODULE qs_integral_utils

   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE orbital_pointers,                ONLY: init_orbital_pointers
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_integral_utils'

! *** Interfaces ***

   INTERFACE get_memory_usage
      MODULE PROCEDURE get_memory_usage_a, get_memory_usage_ab, &
         get_memory_usage_abc, get_memory_usage_abcd
   END INTERFACE

! *** Public subroutines ***

   PUBLIC :: get_memory_usage, basis_set_list_setup

CONTAINS

! **************************************************************************************************
!> \brief Return the maximum memory usage in integral calculations
!> \param qs_kind_set The info for all atomic kinds
!> \param basis_type_a  Type of basis
!> \retval ldmem          Result
! **************************************************************************************************
   FUNCTION get_memory_usage_a(qs_kind_set, basis_type_a) RESULT(ldmem)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_type_a
      INTEGER                                            :: ldmem

      CHARACTER(len=*), PARAMETER :: routineN = 'get_memory_usage_a', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: maxc, maxl, maxs

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxc, maxlgto=maxl, maxsgf=maxs, &
                           basis_type=basis_type_a)
      ldmem = MAX(maxc, maxs)

      CALL init_orbital_pointers(maxl+2)

   END FUNCTION get_memory_usage_a

! **************************************************************************************************
!> \brief Return the maximum memory usage in integral calculations
!> \param qs_kind_set The info for all atomic kinds
!> \param basis_type_a  Type of basis
!> \param basis_type_b  Type of basis
!> \retval ldmem          Result
! **************************************************************************************************
   FUNCTION get_memory_usage_ab(qs_kind_set, basis_type_a, basis_type_b) RESULT(ldmem)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_type_a, basis_type_b
      INTEGER                                            :: ldmem

      CHARACTER(len=*), PARAMETER :: routineN = 'get_memory_usage_ab', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: lda, ldb

      lda = get_memory_usage_a(qs_kind_set, basis_type_a)
      ldb = get_memory_usage_a(qs_kind_set, basis_type_b)
      ldmem = MAX(lda, ldb)

   END FUNCTION get_memory_usage_ab

! **************************************************************************************************
!> \brief Return the maximum memory usage in integral calculations
!> \param qs_kind_set The info for all atomic kinds
!> \param basis_type_a  Type of basis
!> \param basis_type_b  Type of basis
!> \param basis_type_c  Type of basis
!> \retval ldmem          Result
! **************************************************************************************************
   FUNCTION get_memory_usage_abc(qs_kind_set, basis_type_a, &
                                 basis_type_b, basis_type_c) RESULT(ldmem)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_type_a, basis_type_b, basis_type_c
      INTEGER                                            :: ldmem

      CHARACTER(len=*), PARAMETER :: routineN = 'get_memory_usage_abc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: lda, ldb, ldc

      lda = get_memory_usage_a(qs_kind_set, basis_type_a)
      ldb = get_memory_usage_a(qs_kind_set, basis_type_b)
      ldc = get_memory_usage_a(qs_kind_set, basis_type_c)
      ldmem = MAX(lda, ldb, ldc)

   END FUNCTION get_memory_usage_abc

! **************************************************************************************************
!> \brief Return the maximum memory usage in integral calculations
!> \param qs_kind_set The info for all atomic kinds
!> \param basis_type_a  Type of basis
!> \param basis_type_b  Type of basis
!> \param basis_type_c  Type of basis
!> \param basis_type_d  Type of basis
!> \retval ldmem          Result
! **************************************************************************************************
   FUNCTION get_memory_usage_abcd(qs_kind_set, basis_type_a, &
                                  basis_type_b, basis_type_c, basis_type_d) RESULT(ldmem)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_type_a, basis_type_b, &
                                                            basis_type_c, basis_type_d
      INTEGER                                            :: ldmem

      CHARACTER(len=*), PARAMETER :: routineN = 'get_memory_usage_abcd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: lda, ldb, ldc, ldd

      lda = get_memory_usage_a(qs_kind_set, basis_type_a)
      ldb = get_memory_usage_a(qs_kind_set, basis_type_b)
      ldc = get_memory_usage_a(qs_kind_set, basis_type_c)
      ldd = get_memory_usage_a(qs_kind_set, basis_type_d)
      ldmem = MAX(lda, ldb, ldc, ldd)

   END FUNCTION get_memory_usage_abcd

! **************************************************************************************************

! **************************************************************************************************
!> \brief RSet up an easy accessible list of the basis sets for all kinds
!> \param basis_set_list    The basis set list
!> \param basis_type ...
!> \param qs_kind_set   Kind information, the basis is used
! **************************************************************************************************
   SUBROUTINE basis_set_list_setup(basis_set_list, basis_type, qs_kind_set)

      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      CHARACTER(len=*), INTENT(IN)                       :: basis_type
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_set_list_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

! set up basis sets

      DO ikind = 1, SIZE(qs_kind_set)
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set, &
                          basis_type=basis_type)
         NULLIFY (basis_set_list(ikind)%gto_basis_set)
         IF (ASSOCIATED(basis_set)) basis_set_list(ikind)%gto_basis_set => basis_set
      END DO

   END SUBROUTINE basis_set_list_setup

! **************************************************************************************************

END MODULE qs_integral_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Routines that link DBCSR and CP2K concepts together
!> \author Ole Schuett
!> \par History
!>         01.2014 created
! **************************************************************************************************
MODULE cp_dbcsr_cp2k_link
   USE ao_util,                         ONLY: exp_radius
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        accdrv_avoid_after_busy, accdrv_binning_binsize, accdrv_binning_nbins, &
        accdrv_do_inhomogenous, accdrv_min_flop_process, accdrv_min_flop_sort, &
        accdrv_posterior_buffers, accdrv_posterior_streams, accdrv_priority_buffers, &
        accdrv_priority_streams, cp_dbcsr_finalize, cp_dbcsr_get_block_p, &
        cp_dbcsr_get_matrix_type, cp_dbcsr_has_symmetry, cp_dbcsr_reserve_blocks, cp_dbcsr_set, &
        cp_dbcsr_type, dbcsr_get_conf_avg_elements_images, dbcsr_get_conf_comm_thread_load, &
        dbcsr_get_conf_mm_driver, dbcsr_get_conf_mm_stacksize, dbcsr_get_conf_mpi_mem, &
        dbcsr_get_conf_nstacks, dbcsr_get_conf_randmat_seed, dbcsr_get_conf_use_comm_thread, &
        dbcsr_get_conf_use_mpi_exp, dbcsr_get_conf_use_mpi_filtering, &
        dbcsr_set_conf_avg_elements_images, dbcsr_set_conf_comm_thread_load, &
        dbcsr_set_conf_max_ele_block, dbcsr_set_conf_mm_driver, dbcsr_set_conf_mm_stacksize, &
        dbcsr_set_conf_nstacks, dbcsr_set_conf_randmat_seed, dbcsr_set_conf_use_comm_thread, &
        dbcsr_set_conf_use_mpi_exp, dbcsr_set_conf_use_mpi_filtering, dbcsr_type_no_symmetry, &
        has_acc, has_mpi, mm_driver_blas, mm_driver_matmul, mm_driver_smm, mm_driver_xsmm, &
        mm_name_blas, mm_name_matmul, mm_name_smm, mm_name_xsmm, multrec_limit
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_logger_would_log,&
                                              cp_note_level
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              real_8
   USE orbital_pointers,                ONLY: nso
   USE qs_integral_utils,               ONLY: basis_set_list_setup
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              get_neighbor_list_set_p,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_dbcsr_cp2k_link'

   PUBLIC :: cp_dbcsr_config
   PUBLIC :: cp_dbcsr_print_config
   PUBLIC :: cp_dbcsr_alloc_block_from_nbl
   PUBLIC :: cp_dbcsr_to_csr_screening

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief Configures options for DBCSR
!> \param root_section ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_config(root_section)
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_config', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: avg_elements_images, comm_thread_load, &
                                                            error_handle, max_ele_block, &
                                                            mm_driver, mm_ss, randmat_seed
      INTEGER, DIMENSION(3)                              :: nstacks
      LOGICAL                                            :: use_comm_thread, use_mpi_exp, &
                                                            use_mpi_filtering
      TYPE(section_vals_type), POINTER                   :: dbcsr_section

      CALL timeset(routineN, error_handle)
      dbcsr_section => section_vals_get_subs_vals(root_section, &
                                                  "GLOBAL%DBCSR")
      CALL section_vals_val_get(dbcsr_section, &
                                "mm_stack_size", i_val=mm_ss)
      CALL section_vals_val_get(dbcsr_section, &
                                "mm_driver", i_val=mm_driver)
      CALL section_vals_val_get(dbcsr_section, &
                                "MAX_ELEMENTS_PER_BLOCK", i_val=max_ele_block)
      CALL section_vals_val_get(dbcsr_section, &
                                "avg_elements_images", i_val=avg_elements_images)
      CALL section_vals_val_get(dbcsr_section, &
                                "n_size_mnk_stacks", i_val=nstacks(1))
      nstacks(2:3) = nstacks(1)
      CALL section_vals_val_get(dbcsr_section, &
                                "use_mpi_filtering", l_val=use_mpi_filtering)
      CALL section_vals_val_get(dbcsr_section, &
                                "use_mpi_exp", l_val=use_mpi_exp)
      CALL section_vals_val_get(dbcsr_section, &
                                "use_comm_thread", l_val=use_comm_thread)
      CALL section_vals_val_get(dbcsr_section, &
                                "comm_thread_load", i_val=comm_thread_load)
      CALL section_vals_val_get(dbcsr_section, &
                                "multrec_limit", i_val=multrec_limit)
      CALL section_vals_val_get(dbcsr_section, &
                                "randmat_seed", i_val=randmat_seed)

      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%priority_streams", i_val=accdrv_priority_streams)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%priority_buffers", i_val=accdrv_priority_buffers)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%posterior_streams", i_val=accdrv_posterior_streams)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%posterior_buffers", i_val=accdrv_posterior_buffers)

      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%min_flop_process", i_val=accdrv_min_flop_process)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%min_flop_sort", i_val=accdrv_min_flop_sort)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%process_inhomogenous", l_val=accdrv_do_inhomogenous)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%avoid_after_busy", l_val=accdrv_avoid_after_busy)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%binning_nbins", i_val=accdrv_binning_nbins)
      CALL section_vals_val_get(dbcsr_section, &
                                "ACC%binning_binsize", i_val=accdrv_binning_binsize)

      CALL dbcsr_set_conf_mm_driver(mm_driver)
      CALL dbcsr_set_conf_max_ele_block(max_ele_block)
      IF (mm_ss > 0) &
         CALL dbcsr_set_conf_mm_stacksize(mm_ss)
      IF (avg_elements_images > 0) &
         CALL dbcsr_set_conf_avg_elements_images(avg_elements_images)
      IF (nstacks(1) > 0) &
         CALL dbcsr_set_conf_nstacks(nstacks)
      CALL dbcsr_set_conf_use_mpi_filtering(use_mpi_filtering)
      CALL dbcsr_set_conf_use_mpi_exp(use_mpi_exp)
      CALL dbcsr_set_conf_use_comm_thread(use_comm_thread)
      CALL dbcsr_set_conf_comm_thread_load(comm_thread_load)
      CALL dbcsr_set_conf_randmat_seed(randmat_seed)

      CALL timestop(error_handle)
   END SUBROUTINE cp_dbcsr_config

! **************************************************************************************************
!> \brief Prints configuration for DBCSR
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_print_config(unit_nr)
      INTEGER, INTENT(IN), OPTIONAL                      :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_print_config', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: mm_name
      INTEGER                                            :: unit_num
      INTEGER, DIMENSION(3)                              :: n_mnk_stacks
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      unit_num = cp_logger_get_default_io_unit(logger)
      IF (PRESENT(unit_nr)) unit_num = unit_nr

      IF (unit_num > 0 .AND. cp_logger_would_log(logger, cp_note_level)) THEN
         SELECT CASE (dbcsr_get_conf_mm_driver ())
         CASE (mm_driver_blas); mm_name = mm_name_blas
         CASE (mm_driver_matmul); mm_name = mm_name_matmul
         CASE (mm_driver_smm); mm_name = mm_name_smm
         CASE (mm_driver_xsmm); mm_name = mm_name_xsmm
         CASE DEFAULT; CALL cp__b("cp_dbcsr_cp2k_link.F",182,"Unkown MM driver")
         END SELECT
         WRITE (UNIT=unit_num, FMT='(1X,A,T41,A40)') &
            "DBCSR| Multiplication driver", ADJUSTR(mm_name(1:40))

         WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
            "DBCSR| Multrec recursion limit", multrec_limit
         WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
            "DBCSR| Multiplication stack size", dbcsr_get_conf_mm_stacksize()

         IF (dbcsr_get_conf_avg_elements_images() > 0) THEN
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Average elements for images", dbcsr_get_conf_avg_elements_images()
         ELSE
            WRITE (UNIT=unit_num, FMT='(1X,A,T72,A)') &
               "DBCSR| Maximum elements for images", "UNLIMITED"
         ENDIF
         WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
            "DBCSR| Randmat seed", dbcsr_get_conf_randmat_seed()

         CALL dbcsr_get_conf_nstacks(n_mnk_stacks)
         IF (ALL(n_mnk_stacks .EQ. n_mnk_stacks(1))) THEN
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Multiplication size stacks", n_mnk_stacks(1)
         ELSE
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Multiplication size m stacks", n_mnk_stacks(1)
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Multiplication size n stacks", n_mnk_stacks(2)
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Multiplication size l stacks", n_mnk_stacks(3)
         ENDIF

         IF (has_mpi) THEN
            WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
               "DBCSR| Use MPI memory allocation", dbcsr_get_conf_mpi_mem()
            WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
               "DBCSR| Use RMA algorithm", dbcsr_get_conf_use_mpi_exp()
            WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
               "DBCSR| Use Communication thread", dbcsr_get_conf_use_comm_thread()
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| Communication thread load", dbcsr_get_conf_comm_thread_load()
            IF (dbcsr_get_conf_use_mpi_exp()) THEN
               WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
                  "DBCSR| Use MPI filtering", dbcsr_get_conf_use_mpi_filtering()
            ENDIF
         ENDIF

         IF (has_acc) THEN
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Number of priority stack-buffers", accdrv_priority_buffers
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Number of posterior stack-buffers", accdrv_posterior_buffers
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Number of priority streams", accdrv_priority_streams
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Number of posterior streams", accdrv_posterior_streams
            WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
               "DBCSR| ACC: Avoid driver after busy ", accdrv_avoid_after_busy
            WRITE (UNIT=unit_num, FMT='(1X,A,T80,L1)') &
               "DBCSR| ACC: Process inhomogenous stacks", accdrv_do_inhomogenous
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Min. flop for processing", accdrv_min_flop_process
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Min. flop for sorting", accdrv_min_flop_sort
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Number of binning bins", accdrv_binning_nbins
            WRITE (UNIT=unit_num, FMT='(1X,A,T70,I11)') &
               "DBCSR| ACC: Size of binning bins", accdrv_binning_binsize
         END IF
      ENDIF

   END SUBROUTINE cp_dbcsr_print_config

! **************************************************************************************************
!> \brief allocate the blocks of a dbcsr based on the neighbor list
!> \param matrix        the matrix
!> \param sab_orb       the corresponding neighbor list
!> \param desymmetrize  Allocate all block of a non-symmetric matrix from a symmetric list
!> \par History
!>      11.2009 created vw
!>      01.2014 moved here from cp_dbcsr_operations (Ole Schuett)
!> \author vw
!> \note
! **************************************************************************************************

   SUBROUTINE cp_dbcsr_alloc_block_from_nbl(matrix, sab_orb, desymmetrize)

      TYPE(cp_dbcsr_type)                                :: matrix
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      LOGICAL, INTENT(IN), OPTIONAL                      :: desymmetrize

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_alloc_block_from_nbl', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: symmetry
      INTEGER                                            :: blk_cnt, handle, iatom, icol, inode, &
                                                            irow, jatom, last_jatom, nadd
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cols, rows, tmp
      LOGICAL                                            :: alloc_full, is_symmetric, new_atom_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)

      symmetry = cp_dbcsr_get_matrix_type(matrix)

      IF(.NOT.(ASSOCIATED(sab_orb)))CALL cp__a("cp_dbcsr_cp2k_link.F",290)

      CALL get_neighbor_list_set_p(neighbor_list_sets=sab_orb, symmetric=is_symmetric)
      alloc_full = .FALSE.
      IF (PRESENT(desymmetrize)) THEN
         IF (desymmetrize .AND. (symmetry == dbcsr_type_no_symmetry)) THEN
            IF (is_symmetric) alloc_full = .TRUE.
         END IF
      END IF

      CALL cp_dbcsr_finalize(matrix)
      ALLOCATE (rows(3), cols(3))
      blk_cnt = 0
      nadd = 1
      IF (alloc_full) nadd = 2

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom, inode=inode)
         IF (inode == 1) last_jatom = 0
         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
            CYCLE
         END IF
         IF (blk_cnt+nadd .GT. SIZE(rows)) THEN
            ALLOCATE (tmp(blk_cnt+nadd))
            tmp(1:blk_cnt) = rows(1:blk_cnt)
            DEALLOCATE (rows)
            ALLOCATE (rows((blk_cnt+nadd)*2))
            rows(1:blk_cnt) = tmp(1:blk_cnt)
            tmp(1:blk_cnt) = cols(1:blk_cnt)
            DEALLOCATE (cols)
            ALLOCATE (cols((blk_cnt+nadd)*2))
            cols(1:blk_cnt) = tmp(1:blk_cnt)
            DEALLOCATE (tmp)
         ENDIF
         IF (alloc_full) THEN
            blk_cnt = blk_cnt+1
            rows(blk_cnt) = iatom
            cols(blk_cnt) = jatom
            IF (iatom /= jatom) THEN
               blk_cnt = blk_cnt+1
               rows(blk_cnt) = jatom
               cols(blk_cnt) = iatom
            END IF
         ELSE
            blk_cnt = blk_cnt+1
            IF (symmetry == dbcsr_type_no_symmetry) THEN
               rows(blk_cnt) = iatom
               cols(blk_cnt) = jatom
            ELSE
               IF (iatom <= jatom) THEN
                  irow = iatom
                  icol = jatom
               ELSE
                  irow = jatom
                  icol = iatom
               END IF
               rows(blk_cnt) = irow
               cols(blk_cnt) = icol
            END IF
         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      !
      CALL cp_dbcsr_reserve_blocks(matrix, rows(1:blk_cnt), cols(1:blk_cnt))
      DEALLOCATE (rows)
      DEALLOCATE (cols)
      CALL cp_dbcsr_finalize(matrix)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_alloc_block_from_nbl

! **************************************************************************************************
!> \brief Apply distance screening to refine sparsity pattern of matrices in CSR
!>        format (using eps_pgf_orb). Currently this is used for the external
!>        library PEXSI.
!> \param ks_env ...
!> \param[in, out] csr_sparsity DBCSR matrix defining CSR sparsity pattern.
!>                              This matrix must be initialized and allocated
!>                              with exactly the same DBCSR sparsity pattern as
!>                              the DBCSR matrix that is used to create the CSR
!>                              matrix. It must have symmetric DBCSR format and
!>                              must not be filtered.
!> \par History
!>      02.2015 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_to_csr_screening(ks_env, csr_sparsity)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: csr_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_to_csr_screening', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, iatom, icol, ikind, ipgf, irow, iset, isgf, ishell, &
         jatom, jkind, jpgf, jset, jsgf, jshell, natom, nkind, nset_a, nset_b
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(:), POINTER                     :: npgf_a, npgf_b, nshell_a, nshell_b
      INTEGER, DIMENSION(:, :), POINTER                  :: l_a, l_b
      LOGICAL                                            :: do_symmetric, found
      REAL(KIND=dp)                                      :: dab, eps_pgf_orb, r_a, r_b
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, zet_a, zet_b
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc_a, gcc_b
      REAL(KIND=real_8), DIMENSION(:, :), POINTER        :: screen_blk
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list_a, basis_set_list_b
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: neighbour_list
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (screen_blk, atomic_kind_set, basis_set_list_a, &
               basis_set_list_b, basis_set_a, basis_set_b, nl_iterator, &
               qs_kind_set, dft_control)

      CALL timeset(routineN, handle)

      IF(.NOT.(cp_dbcsr_has_symmetry(csr_sparsity)))CALL cp__a("cp_dbcsr_cp2k_link.F",418)

      CALL get_ks_env(ks_env, &
                      sab_orb=neighbour_list, &
                      atomic_kind_set=atomic_kind_set, &
                      natom=natom, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control)

      eps_pgf_orb = dft_control%qs_control%eps_pgf_orb

      nkind = SIZE(qs_kind_set)
      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
      IF(.NOT.(SIZE(neighbour_list) > 0))CALL cp__a("cp_dbcsr_cp2k_link.F",432)
      CALL get_neighbor_list_set_p(neighbor_list_sets=neighbour_list, symmetric=do_symmetric)
      IF(.NOT.(do_symmetric))CALL cp__a("cp_dbcsr_cp2k_link.F",434)
      ALLOCATE (basis_set_list_a(nkind), basis_set_list_b(nkind))
      CALL basis_set_list_setup(basis_set_list_a, "ORB", qs_kind_set)
      CALL basis_set_list_setup(basis_set_list_b, "ORB", qs_kind_set)

      ! csr_sparsity can obtain values 0 (if zero element) or 1 (if non-zero element)
      CALL cp_dbcsr_set(csr_sparsity, 0.0)

      CALL neighbor_list_iterator_create(nl_iterator, neighbour_list)

      ! Iterate over interacting pairs of atoms corresponding to non-zero
      ! DBCSR blocks
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, &
                                ikind=ikind, jkind=jkind, &
                                iatom=iatom, jatom=jatom, &
                                r=rab)

         basis_set_a => basis_set_list_a(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list_b(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE

         atom_a = atom_of_kind(iatom)
         atom_b = atom_of_kind(jatom)

         nset_a = basis_set_a%nset
         nset_b = basis_set_b%nset
         npgf_a => basis_set_a%npgf
         npgf_b => basis_set_b%npgf
         nshell_a => basis_set_a%nshell
         nshell_b => basis_set_b%nshell

         l_a => basis_set_a%l
         l_b => basis_set_b%l
         gcc_a => basis_set_a%gcc
         gcc_b => basis_set_b%gcc
         zet_a => basis_set_a%zet
         zet_b => basis_set_b%zet

         rpgfa => basis_set_a%pgf_radius
         rpgfb => basis_set_b%pgf_radius

         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         END IF

         CALL cp_dbcsr_get_block_p(matrix=csr_sparsity, row=irow, col=icol, &
                                   block=screen_blk, found=found)

         IF(.NOT.(found))CALL cp__a("cp_dbcsr_cp2k_link.F",488)

         ! Distance between atoms a and b
         dab = SQRT(rab(1)**2+rab(2)**2+rab(3)**2)

         ! iterate over pairs of primitive GTOs i,j, get their radii r_i, r_j according
         ! to eps_pgf_orb. Define all matrix elements as non-zero to which a
         ! contribution from two Gaussians i,j exists with r_i + r_j >= dab.

         isgf = 0
         DO iset = 1, nset_a
            DO ishell = 1, nshell_a(iset)
               jsgf = 0
               DO jset = 1, nset_b
                  DO jshell = 1, nshell_b(jset)
                     gto_loop: DO ipgf = 1, npgf_a(iset)
                        DO jpgf = 1, npgf_b(jset)
                           IF (rpgfa(ipgf, iset)+rpgfb(jpgf, jset) .GE. dab) THEN
                              ! more selective screening with radius calculated for each primitive GTO
                              r_a = exp_radius(l_a(ishell, iset), &
                                               zet_a(ipgf, iset), &
                                               eps_pgf_orb, &
                                               gcc_a(ipgf, ishell, iset))
                              r_b = exp_radius(l_b(jshell, jset), &
                                               zet_b(jpgf, jset), &
                                               eps_pgf_orb, &
                                               gcc_b(jpgf, jshell, jset))
                              IF (r_a+r_b .GE. dab) THEN
                                 IF (irow .EQ. iatom) THEN
                                    screen_blk(isgf+1:isgf+nso(l_a(ishell, iset)), &
                                               jsgf+1:jsgf+nso(l_b(jshell, jset))) = 1.0_dp
                                 ELSE
                                    screen_blk(jsgf+1:jsgf+nso(l_b(jshell, jset)), &
                                               isgf+1:isgf+nso(l_a(ishell, iset))) = 1.0_dp
                                 ENDIF
                                 EXIT gto_loop
                              ENDIF
                           ENDIF
                        ENDDO
                     ENDDO gto_loop
                     jsgf = jsgf+nso(l_b(jshell, jset))
                  ENDDO
               ENDDO
               isgf = isgf+nso(l_a(ishell, iset))
            ENDDO
         ENDDO
      ENDDO

      CALL neighbor_list_iterator_release(nl_iterator)
      DEALLOCATE (basis_set_list_a, basis_set_list_b)

      CALL timestop(handle)
   END SUBROUTINE cp_dbcsr_to_csr_screening

END MODULE cp_dbcsr_cp2k_link
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the construction of the coefficients
!>      for the expansion  of the atomic
!>      densities rho1_hard and rho1_soft in terms of primitive spherical gaussians.
!> \par History
!>      05-2004 created
!> \author MI
! **************************************************************************************************
MODULE qs_oce_methods

   USE ai_overlap,                      ONLY: overlap
   USE ao_util,                         ONLY: exp_radius
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: dfac,&
                                              pi
   USE orbital_pointers,                ONLY: indco,&
                                              init_orbital_pointers,&
                                              nco,&
                                              ncoset,&
                                              nso
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE sap_kind_types,                  ONLY: clist_type,&
                                              sap_int_type,&
                                              sap_sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_oce_methods'

! *** Public subroutines ***

   PUBLIC :: build_oce_matrices, proj_blk, prj_scatter, prj_gather

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param oces ...
!> \param atom_ka ...
!> \param atom_kb ...
!> \param rab ...
!> \param nder ...
!> \param sgf_list ...
!> \param nsgf_cnt ...
!> \param sgf_soft_only ...
!> \param eps_fit ...
! **************************************************************************************************
   SUBROUTINE build_oce_block(oces, atom_ka, atom_kb, rab, nder, sgf_list, nsgf_cnt, sgf_soft_only, &
                              eps_fit)

      TYPE(block_p_type), DIMENSION(:), POINTER          :: oces
      TYPE(qs_kind_type), POINTER                        :: atom_ka, atom_kb
      REAL(dp), DIMENSION(3)                             :: rab
      INTEGER, INTENT(IN)                                :: nder
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sgf_list
      INTEGER, INTENT(OUT)                               :: nsgf_cnt
      LOGICAL, INTENT(OUT)                               :: sgf_soft_only
      REAL(dp), INTENT(IN)                               :: eps_fit

      CHARACTER(len=*), PARAMETER :: routineN = 'build_oce_block', &
         routineP = moduleN//':'//routineN

      INTEGER :: first_col, ic, ico, ider, ig1, igau, ip, ipgf, is, isgfb, isgfb_cnt, iso, isp, &
         jc, jset, lds, lm, lpoint, lprj, lsgfb, lsgfb_cnt, lshell, lx, ly, lz, m, m1, maxcob, &
         maxder, maxlb, maxlprj, maxnprja, maxsoa, msab, n, ncob, np_car, np_sph, nsatbas, nseta, &
         nsetb, nsoatot, ntotsgfb, sgf_hard_only
      INTEGER, DIMENSION(:), POINTER                     :: fp_cara, fp_spha, lb_max, lb_min, npgfb, &
                                                            nprjla, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfb
      LOGICAL                                            :: calculate_forces, paw_atom_a, paw_atom_b
      REAL(dp)                                           :: dab, hard_radius_a, hard_radius_b, &
                                                            radius, rcprja
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: c2s, ovs, spa_sb
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: s
      REAL(dp), DIMENSION(:), POINTER                    :: set_radius_b, zisominb
      REAL(dp), DIMENSION(:, :), POINTER                 :: cprj_s, csprj, rpgfb, rzetprja, spa_tmp, &
                                                            sphi_b, zetb, zetprja
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_a, orb_basis_b
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_a, paw_proj_b

      NULLIFY (orb_basis_a, paw_proj_a)
      CALL get_qs_kind(qs_kind=atom_ka, basis_set=orb_basis_a, &
                       paw_proj_set=paw_proj_a, paw_atom=paw_atom_a, &
                       hard_radius=hard_radius_a)

      NULLIFY (orb_basis_b, paw_proj_b)
      CALL get_qs_kind(qs_kind=atom_kb, basis_set=orb_basis_b, &
                       paw_proj_set=paw_proj_b, paw_atom=paw_atom_b, &
                       hard_radius=hard_radius_b)

      IF (.NOT. paw_atom_a) RETURN

      NULLIFY (cprj_s, nprjla, fp_cara, fp_spha, rzetprja, zetprja)
      CALL get_paw_proj_set(paw_proj_set=paw_proj_a, cprj_s=cprj_s, csprj=csprj, maxl=maxlprj, &
                            nprj=nprjla, ncgauprj=np_car, nsgauprj=np_sph, nsatbas=nsatbas, rcprj=rcprja, &
                            first_prj=fp_cara, first_prjs=fp_spha, &
                            rzetprj=rzetprja, zetprj=zetprja)

      NULLIFY (first_sgfb, lb_max, lb_min, npgfb, nsgfb, rpgfb, sphi_b, set_radius_b, zetb, zisominb)
      CALL get_gto_basis_set(gto_basis_set=orb_basis_b, nset=nsetb, nsgf=ntotsgfb, &
                             set_radius=set_radius_b, lmax=lb_max, lmin=lb_min, &
                             npgf=npgfb, nsgf_set=nsgfb, pgf_radius=rpgfb, &
                             sphi=sphi_b, zet=zetb, first_sgf=first_sgfb, &
                             maxco=maxcob, maxl=maxlb)

      CALL get_gto_basis_set(gto_basis_set=orb_basis_a, nset=nseta, maxso=maxsoa)

!  *** Add the block ab ***
      dab = SQRT(SUM(rab*rab))

      maxder = ncoset(nder)
      nsoatot = maxsoa*nseta
      maxnprja = SIZE(zetprja, 1)

      calculate_forces = .FALSE.
      IF (nder > 0) THEN
         calculate_forces = .TRUE.
      ENDIF

      lm = MAX(maxlb, maxlprj)
      lds = ncoset(lm+nder+1)
      msab = MAX(maxnprja*ncoset(maxlprj), maxcob)

      ALLOCATE (c2s(lds, lds))
      ALLOCATE (s(lds, lds, ncoset(nder+1)))
      ALLOCATE (spa_sb(np_car, ntotsgfb))
      ALLOCATE (spa_tmp(msab, msab*maxder))
      ALLOCATE (ovs(np_sph, maxcob*nsetb*maxder))

      m1 = 0
      nsgf_cnt = 0
      isgfb_cnt = 1
      sgf_hard_only = 0
      DO jset = 1, nsetb
         !
         ! Set the contribution list
         IF (hard_radius_a+set_radius_b(jset) >= dab) THEN
            isgfb = first_sgfb(1, jset)
            lsgfb = isgfb-1+nsgfb(jset)
            DO jc = isgfb, lsgfb
               nsgf_cnt = nsgf_cnt+1
               sgf_list(nsgf_cnt) = jc
            ENDDO

            ! check if this function is hard
            radius = exp_radius(lb_max(jset), MAXVAL(zetb(1:npgfb(jset), jset)), eps_fit, 1.0_dp)
            IF (radius .LE. hard_radius_b) sgf_hard_only = sgf_hard_only+1

            ! ***integral between proj of iatom and primitives of jatom
            ! *** Calculate the primitives overlap ***
            spa_tmp = 0.0_dp
            ovs = 0.0_dp
            s = 0.0_dp
            ncob = npgfb(jset)*ncoset(lb_max(jset))
            isgfb = first_sgfb(1, jset)
            lsgfb = isgfb-1+nsgfb(jset)

            lsgfb_cnt = isgfb_cnt-1+nsgfb(jset)

            DO lprj = 0, maxlprj
               CALL overlap(lprj, lprj, nprjla(lprj), &
                            rzetprja(:, lprj), zetprja(:, lprj), &
                            lb_max(jset), lb_min(jset), npgfb(jset), &
                            rpgfb(:, jset), zetb(:, jset), &
                            -rab, dab, spa_tmp, &
                            nder, .TRUE., s, lds)
               DO iso = 1, nso(lprj)
                  DO ico = 1, nco(lprj)
                     lx = indco(1, ico+ncoset(lprj-1))
                     ly = indco(2, ico+ncoset(lprj-1))
                     lz = indco(3, ico+ncoset(lprj-1))
                     c2s(iso, ico) = orbtramat(lprj)%c2s(iso, ico)/SQRT((4.0_dp*pi)/dfac(2*lprj+1)* &
                                                                        dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
                  ENDDO
               ENDDO
               DO ider = 1, maxder
                  is = (ider-1)*SIZE(spa_tmp, 1)
                  isp = (ider-1)*maxcob*nsetb
                  DO ipgf = 1, nprjla(lprj)
                     lpoint = ncoset(lprj-1)+1+(ipgf-1)*ncoset(lprj)
                     m = fp_spha(lprj)+(ipgf-1)*nso(lprj)
                     DO ip = 1, npgfb(jset)
                        ic = (ip-1)*ncoset(lb_max(jset))
                        igau = isp+ic+m1+ncoset(lb_min(jset)-1)+1
                        ig1 = is+ic+ncoset(lb_min(jset)-1)+1
                        n = ncoset(lb_max(jset))-ncoset(lb_min(jset)-1)
                        CALL DGEMM("N", "N", nso(lprj), n, nco(lprj), 1._dp, c2s, lds, spa_tmp(lpoint, ig1), msab, &
                                   0._dp, ovs(m, igau), np_sph)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            IF (paw_atom_b) THEN
               CALL get_paw_proj_set(paw_proj_set=paw_proj_b, zisomin=zisominb)
               DO ipgf = 1, npgfb(jset)
                  DO lshell = lb_min(jset), lb_max(jset)
                     IF (zetb(ipgf, jset) >= zisominb(lshell)) THEN
                        igau = n*(ipgf-1)+ncoset(lshell-1)
                        DO ider = 1, maxder
                           is = maxcob*(ider-1)
                           isp = (ider-1)*maxcob*nsetb
                           ovs(:, igau+1+isp+m1:igau+nco(lshell)+isp+m1) = 0.0_dp
                        ENDDO
                     ENDIF
                  ENDDO
               END DO
            ENDIF

            ! *** Contraction step (integrals and derivatives)
            DO ider = 1, maxder
               first_col = (ider-1)*maxcob*nsetb+1+m1
               CALL dgemm("N", "N", np_sph, nsgfb(jset), ncob, &
                          1.0_dp, ovs(1, first_col), SIZE(ovs, 1), &
                          sphi_b(1, isgfb), SIZE(sphi_b, 1), &
                          0.0_dp, spa_sb(1, isgfb), SIZE(spa_sb, 1))

               CALL dgemm("T", "N", nsatbas, nsgfb(jset), np_sph, &
                          1.0_dp, csprj(1, 1), SIZE(csprj, 1), &
                          spa_sb(1, isgfb), SIZE(spa_sb, 1), &
                          1.0_dp, oces(ider)%block(1, isgfb_cnt), SIZE(oces(ider)%block, 1))
            ENDDO
            isgfb_cnt = isgfb_cnt+nsgfb(jset)
         END IF ! radius
         m1 = m1+maxcob
      ENDDO !jset

      ! check if the screened functions are all soft
      sgf_soft_only = .FALSE.
      IF (sgf_hard_only .EQ. 0) sgf_soft_only = .TRUE.

      DEALLOCATE (c2s, s, spa_sb, spa_tmp, ovs)

   END SUBROUTINE build_oce_block

! **************************************************************************************************
!> \brief ...
!> \param oceh ...
!> \param oces ...
!> \param atom_ka ...
!> \param sgf_list ...
!> \param nsgf_cnt ...
! **************************************************************************************************
   SUBROUTINE build_oce_block_local(oceh, oces, atom_ka, sgf_list, nsgf_cnt)

      TYPE(block_p_type), DIMENSION(:), POINTER          :: oceh, oces
      TYPE(qs_kind_type), POINTER                        :: atom_ka
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sgf_list
      INTEGER, INTENT(OUT)                               :: nsgf_cnt

      CHARACTER(len=*), PARAMETER :: routineN = 'build_oce_block_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iset, isgfa, j, jc, lsgfa, maxlprj, &
                                                            maxsoa, n, nsatbas, nseta, nsgfa
      INTEGER, DIMENSION(:), POINTER                     :: n2oindex, nsgf_seta
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: paw_atom_a
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: prjloc_h, prjloc_s
      REAL(dp), DIMENSION(:, :), POINTER                 :: local_oce_h, local_oce_s
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_a
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_a

      NULLIFY (orb_basis_a, paw_proj_a)
      CALL get_qs_kind(qs_kind=atom_ka, basis_set=orb_basis_a, &
                       paw_proj_set=paw_proj_a, paw_atom=paw_atom_a)

      IF (.NOT. paw_atom_a) RETURN

      CALL get_paw_proj_set(paw_proj_set=paw_proj_a, maxl=maxlprj, nsatbas=nsatbas, &
                            n2oindex=n2oindex)

      CALL get_gto_basis_set(gto_basis_set=orb_basis_a, first_sgf=first_sgfa, &
                             nsgf=nsgfa, nsgf_set=nsgf_seta, nset=nseta, maxso=maxsoa)

      NULLIFY (local_oce_h, local_oce_s)
      CALL get_paw_proj_set(paw_proj_set=paw_proj_a, &
                            local_oce_sphi_h=local_oce_h, &
                            local_oce_sphi_s=local_oce_s)

      ALLOCATE (prjloc_h(nseta*maxsoa, nsgfa), prjloc_s(nseta*maxsoa, nsgfa))
      prjloc_h = 0._dp
      prjloc_s = 0._dp

      nsgf_cnt = 0
      DO iset = 1, nseta
         isgfa = first_sgfa(1, iset)
         lsgfa = isgfa-1+nsgf_seta(iset)
         DO jc = isgfa, lsgfa
            nsgf_cnt = nsgf_cnt+1
            sgf_list(nsgf_cnt) = jc
         ENDDO
         n = maxsoa*(iset-1)

         prjloc_h(n+1:n+maxsoa, isgfa:lsgfa) = local_oce_h(1:maxsoa, isgfa:lsgfa)
         prjloc_s(n+1:n+maxsoa, isgfa:lsgfa) = local_oce_s(1:maxsoa, isgfa:lsgfa)
      ENDDO

      DO i = 1, nsgfa
         DO j = 1, nsatbas
            jc = n2oindex(j)
            oceh(1)%block(j, i) = prjloc_h(jc, i)
            oces(1)%block(j, i) = prjloc_s(jc, i)
         END DO
      END DO

      DEALLOCATE (prjloc_h, prjloc_s)

   END SUBROUTINE build_oce_block_local

! **************************************************************************************************
!> \brief Set up the sparse matrix for the coefficients of one center expansions
!>      This routine uses the same logic as the nonlocal pseudopotential
!> \param intac TYPE that holds the integrals (a=basis; c=projector)
!> \param calculate_forces ...
!> \param nder ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param sap_oce ...
!> \param eps_fit ...
!> \par History
!>      02.2009 created
!> \author jgh
! **************************************************************************************************
   SUBROUTINE build_oce_matrices(intac, calculate_forces, nder, &
                                 qs_kind_set, particle_set, sap_oce, eps_fit)

      TYPE(sap_int_type), DIMENSION(:), POINTER          :: intac
      LOGICAL, INTENT(IN)                                :: calculate_forces
      INTEGER                                            :: nder
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sap_oce
      REAL(dp), INTENT(IN)                               :: eps_fit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_oce_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, i, iac, ikind, ilist, jkind, jneighbor, ldai, ldsab, &
         maxco, maxder, maxl, maxlgto, maxlprj, maxprj, maxsgf, maxsoa, maxsob, mlprj, natom, &
         ncoa_sum, nkind, nlist, nneighbor, nsatbas, nseta, nsetb, nsgf_cnt, nsgfa, nsobtot
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sgf_list
      INTEGER, DIMENSION(3)                              :: cell_b
      INTEGER, DIMENSION(:), POINTER                     :: fp_car, fp_sph, la_max, la_min, npgfa, &
                                                            nprjla, nsgf_seta
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: local, paw_atom_b, sgf_soft_only
      REAL(KIND=dp)                                      :: dab, rcprj
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: sab, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: ai_work
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rzetprj, sphi_a, zeta, zetb
      TYPE(block_p_type), DIMENSION(:), POINTER          :: oceh, oces
      TYPE(clist_type), POINTER                          :: clist
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_paw, orb_basis_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_b
      TYPE(qs_kind_type), POINTER                        :: at_a, at_b, qs_kind

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      IF (ASSOCIATED(sap_oce)) THEN

         nkind = SIZE(qs_kind_set)
         natom = SIZE(particle_set)

         maxder = ncoset(nder)

         CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                              maxco=maxco, &
                              maxlgto=maxlgto, &
                              maxlprj=maxlprj, &
                              maxco_proj=maxprj, &
                              maxsgf=maxsgf)

         maxl = MAX(maxlgto, maxlprj)
         CALL init_orbital_pointers(maxl+nder+1)

         ldsab = MAX(maxco, ncoset(maxlprj), maxsgf, maxprj)
         ldai = ncoset(maxl+nder+1)
         ALLOCATE (sab(ldsab, ldsab*maxder), work(ldsab, ldsab*maxder))
         sab = 0.0_dp
         ALLOCATE (ai_work(ldai, ldai, ncoset(nder+1)))
         ai_work = 0.0_dp

         ALLOCATE (oceh(maxder), oces(maxder))

         DO i = 1, nkind*nkind
            NULLIFY (intac(i)%alist, intac(i)%asort, intac(i)%aindex)
            intac(i)%nalist = 0
         END DO

         !calculate the overlap integrals <a|p>
         CALL neighbor_list_iterator_create(nl_iterator, sap_oce)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=atom_a, jatom=atom_b, &
                                   nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, cell=cell_b, r=rab)

            iac = ikind+nkind*(jkind-1)
            dab = SQRT(SUM(rab*rab))

            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, basis_set=orb_basis_set)

            IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE
            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                   first_sgf=first_sgfa, &
                                   lmax=la_max, &
                                   lmin=la_min, &
                                   nco_sum=ncoa_sum, &
                                   maxso=maxsoa, &
                                   npgf=npgfa, &
                                   nset=nseta, &
                                   nsgf=nsgfa, &
                                   nsgf_set=nsgf_seta, &
                                   pgf_radius=rpgfa, &
                                   set_radius=set_radius_a, &
                                   sphi=sphi_a, &
                                   zet=zeta)

            qs_kind => qs_kind_set(jkind)

            NULLIFY (paw_proj_b)
            CALL get_qs_kind(qs_kind=qs_kind, paw_proj_set=paw_proj_b, paw_atom=paw_atom_b)
            IF (.NOT. paw_atom_b) CYCLE

            CALL get_qs_kind(qs_kind=qs_kind, basis_set=orb_basis_paw)
            IF (.NOT. ASSOCIATED(orb_basis_paw)) CYCLE
            CALL get_gto_basis_set(gto_basis_set=orb_basis_paw, maxso=maxsob, nset=nsetb)

            CALL get_paw_proj_set(paw_proj_set=paw_proj_b, &
                                  maxl=mlprj, &
                                  nprj=nprjla, &
                                  nsatbas=nsatbas, &
                                  rcprj=rcprj, &
                                  first_prj=fp_car, &
                                  first_prjs=fp_sph, &
                                  rzetprj=rzetprj, &
                                  zetprj=zetb)

            nsobtot = nsatbas

            IF (.NOT. ASSOCIATED(intac(iac)%alist)) THEN
               intac(iac)%a_kind = ikind
               intac(iac)%p_kind = jkind
               intac(iac)%nalist = nlist
               ALLOCATE (intac(iac)%alist(nlist))
               DO i = 1, nlist
                  NULLIFY (intac(iac)%alist(i)%clist)
                  intac(iac)%alist(i)%aatom = 0
                  intac(iac)%alist(i)%nclist = 0
               END DO
            END IF
            IF (.NOT. ASSOCIATED(intac(iac)%alist(ilist)%clist)) THEN
               intac(iac)%alist(ilist)%aatom = atom_a
               intac(iac)%alist(ilist)%nclist = nneighbor
               ALLOCATE (intac(iac)%alist(ilist)%clist(nneighbor))
            END IF

            clist => intac(iac)%alist(ilist)%clist(jneighbor)
            clist%catom = atom_b
            clist%cell = cell_b
            clist%rac = rab
            clist%nsgf_cnt = 0
            clist%maxac = 0.0_dp
            clist%maxach = 0.0_dp
            NULLIFY (clist%acint, clist%achint, clist%sgf_list)

            ALLOCATE (sgf_list(nsgfa))

            at_a => qs_kind_set(jkind)
            at_b => qs_kind_set(ikind)

            local = (atom_a == atom_b .AND. ALL(cell_b == 0))

            IF (local) THEN
               DO i = 1, maxder
                  ALLOCATE (oceh(i)%block(nsobtot, nsgfa), oces(i)%block(nsobtot, nsgfa))
                  oceh(i)%block = 0._dp
                  oces(i)%block = 0._dp
               END DO
               CALL build_oce_block_local(oceh, oces, at_a, sgf_list, nsgf_cnt)
               clist%nsgf_cnt = nsgf_cnt
               clist%sgf_soft_only = .FALSE.
               IF (nsgf_cnt > 0) THEN
                  ALLOCATE (clist%acint(nsgf_cnt, nsobtot, maxder), clist%sgf_list(nsgf_cnt))
                  clist%acint(:, :, :) = 0._dp
                  clist%sgf_list(:) = HUGE(0)
                  IF(.NOT.(nsgf_cnt == nsgfa))CALL cp__a("qs_oce_methods.F",521)
                  ! *** Special case: A=B
                  ALLOCATE (clist%achint(nsgfa, nsobtot, maxder))
                  clist%achint = 0._dp
                  clist%acint(1:nsgfa, 1:nsobtot, 1) = TRANSPOSE(oces(1)%block(1:nsobtot, 1:nsgfa))
                  clist%achint(1:nsgfa, 1:nsobtot, 1) = TRANSPOSE(oceh(1)%block(1:nsobtot, 1:nsgfa))
                  clist%maxac = MAXVAL(ABS(clist%acint(:, :, 1)))
                  clist%maxach = 0._dp
                  clist%sgf_list(1:nsgf_cnt) = sgf_list(1:nsgf_cnt)
               ENDIF
               DO i = 1, maxder
                  DEALLOCATE (oceh(i)%block, oces(i)%block)
               END DO
            ELSE
               DO i = 1, maxder
                  ALLOCATE (oces(i)%block(nsobtot, nsgfa))
                  oces(i)%block = 0._dp
               END DO
               CALL build_oce_block(oces, at_a, at_b, rab, nder, sgf_list, nsgf_cnt, sgf_soft_only, eps_fit)
               clist%nsgf_cnt = nsgf_cnt
               clist%sgf_soft_only = sgf_soft_only
               IF (nsgf_cnt > 0) THEN
                  ALLOCATE (clist%acint(nsgf_cnt, nsobtot, maxder), clist%sgf_list(nsgf_cnt))
                  clist%acint(:, :, :) = 0._dp
                  clist%sgf_list(:) = HUGE(0)
                  DO i = 1, maxder
                     clist%acint(1:nsgf_cnt, 1:nsobtot, i) = TRANSPOSE(oces(i)%block(1:nsobtot, 1:nsgf_cnt))
                  END DO
                  clist%maxac = MAXVAL(ABS(clist%acint(:, :, 1)))
                  clist%maxach = 0._dp
                  clist%sgf_list(1:nsgf_cnt) = sgf_list(1:nsgf_cnt)
               ENDIF
               DO i = 1, maxder
                  DEALLOCATE (oces(i)%block)
               END DO
            END IF

            DEALLOCATE (sgf_list)

         END DO
         CALL neighbor_list_iterator_release(nl_iterator)

         DEALLOCATE (sab, work, ai_work)
         DEALLOCATE (oceh, oces)

         ! set up sort index
         CALL sap_sort(intac)

      END IF

      CALL timestop(handle)

   END SUBROUTINE build_oce_matrices

! **************************************************************************************************
!> \brief Project a matrix block onto the local atomic functions.
!>
!> \param h_a ...
!> \param s_a ...
!> \param na ...
!> \param h_b ...
!> \param s_b ...
!> \param nb ...
!> \param blk ...
!> \param ldb ...
!> \param proj_h ...
!> \param proj_s ...
!> \param nso ...
!> \param buf1 ...
!> \param buf2 ...
!> \param fac ...
!> \param distab ...
!> \par History
!>      02.2009 created
!> \author vw
! **************************************************************************************************
   SUBROUTINE proj_blk(h_a, s_a, na, h_b, s_b, nb, blk, ldb, proj_h, proj_s, nso, buf1, buf2, fac, distab)

      INTEGER, INTENT(IN)                                :: na
      REAL(dp), INTENT(IN)                               :: s_a(na, *), h_a(na, *)
      INTEGER, INTENT(IN)                                :: nb
      REAL(dp), INTENT(IN)                               :: s_b(nb, *), h_b(nb, *)
      INTEGER, INTENT(IN)                                :: ldb
      REAL(dp), INTENT(IN)                               :: blk(ldb, *)
      INTEGER, INTENT(IN)                                :: nso
      REAL(dp), INTENT(OUT)                              :: proj_s(nso, *), proj_h(nso, *)
      REAL(dp), INTENT(INOUT)                            :: buf1(*), buf2(*)
      REAL(dp), INTENT(IN)                               :: fac
      LOGICAL, INTENT(IN)                                :: distab

!

      IF (na .EQ. 0 .OR. nb .EQ. 0 .OR. nso .EQ. 0) RETURN
      !
      ! handle special cases
      IF (na .EQ. 1 .AND. nb .EQ. 1) THEN
         !
         ! hard
         CALL dger(nso, nso, fac*blk(1, 1), h_a(1, 1), 1, h_b(1, 1), 1, proj_h(1, 1), nso)
         !
         ! soft
         CALL dger(nso, nso, fac*blk(1, 1), s_a(1, 1), 1, s_b(1, 1), 1, proj_s(1, 1), nso)
      ELSE
         IF (distab) THEN
            !
            ! hard
            CALL dgemm('N', 'N', na, nso, nb, fac, blk(1, 1), ldb, h_b(1, 1), nb, 0.0_dp, buf1(1), na)
            CALL dgemm('T', 'N', nso, nso, na, 1.0_dp, h_a(1, 1), na, buf1(1), na, 0.0_dp, buf2(1), nso)
            CALL daxpy(nso*nso, 1.0_dp, buf2(1), 1, proj_h(1, 1), 1)
            !
            ! soft
            CALL daxpy(nso*nso, 1.0_dp, buf2(1), 1, proj_s(1, 1), 1)
         ELSE
            !
            ! hard
            CALL dgemm('N', 'N', na, nso, nb, fac, blk(1, 1), ldb, h_b(1, 1), nb, 0.0_dp, buf1(1), na)
            CALL dgemm('T', 'N', nso, nso, na, 1.0_dp, h_a(1, 1), na, buf1(1), na, 1.0_dp, proj_h(1, 1), nso)
            !
            ! soft
            CALL dgemm('N', 'N', na, nso, nb, fac, blk(1, 1), ldb, s_b(1, 1), nb, 0.0_dp, buf1(1), na)
            CALL dgemm('T', 'N', nso, nso, na, 1.0_dp, s_a(1, 1), na, buf1(1), na, 1.0_dp, proj_s(1, 1), nso)
         ENDIF
      ENDIF
      !
   END SUBROUTINE proj_blk

! **************************************************************************************************
!> \brief ...
!> \param ain matrix in old indexing
!> \param aout matrix in new compressed indexing
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE prj_gather(ain, aout, atom)

      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: ain
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: aout
      TYPE(qs_kind_type), INTENT(IN)                     :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'prj_gather', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ip, j, jp, nbas
      INTEGER, DIMENSION(:), POINTER                     :: n2oindex
      LOGICAL                                            :: paw_atom
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj

      NULLIFY (paw_proj)
      CALL get_qs_kind(qs_kind=atom, paw_proj_set=paw_proj, paw_atom=paw_atom)
      NULLIFY (n2oindex)
      CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nbas, n2oindex=n2oindex)

      DO i = 1, nbas
         ip = n2oindex(i)
         DO j = 1, nbas
            jp = n2oindex(j)
            aout(j, i) = ain(jp, ip)
         END DO
      END DO

   END SUBROUTINE prj_gather

! **************************************************************************************************
!> \brief ...
!> \param ain  matrix in new compressed indexing
!> \param aout matrix in old indexing (addup)
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE prj_scatter(ain, aout, atom)

      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: ain
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: aout
      TYPE(qs_kind_type), INTENT(IN)                     :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'prj_scatter', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ip, j, jp, nbas
      INTEGER, DIMENSION(:), POINTER                     :: n2oindex
      LOGICAL                                            :: paw_atom
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj

      NULLIFY (paw_proj)
      CALL get_qs_kind(qs_kind=atom, paw_proj_set=paw_proj, paw_atom=paw_atom)
      NULLIFY (n2oindex)
      CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nbas, n2oindex=n2oindex)

      DO i = 1, nbas
         ip = n2oindex(i)
         DO j = 1, nbas
            jp = n2oindex(j)
            aout(jp, ip) = aout(jp, ip)+ain(j, i)
         END DO
      END DO

   END SUBROUTINE prj_scatter

END MODULE qs_oce_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE qs_rho_atom_methods

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              gapw_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE lebedev,                         ONLY: deallocate_lebedev_grids,&
                                              get_number_of_lebedev_grid,&
                                              init_lebedev_grids,&
                                              lebedev_grid
   USE mathconstants,                   ONLY: fourpi,&
                                              pi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset,&
                                              soset
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_grid_atom,                    ONLY: create_grid_atom,&
                                              grid_atom_type
   USE qs_harmonics_atom,               ONLY: create_harmonics_atom,&
                                              get_maxl_CG,&
                                              get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_oce_methods,                  ONLY: proj_blk
   USE qs_oce_types,                    ONLY: oce_matrix_type
   USE qs_rho_atom_types,               ONLY: allocate_rho_atom_set,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE sap_kind_types,                  ONLY: alist_pre_align_blk,&
                                              alist_type,&
                                              get_alist
   USE spherical_harmonics,             ONLY: clebsch_gordon,&
                                              clebsch_gordon_deallocate,&
                                              clebsch_gordon_init
   USE util,                            ONLY: get_limit
   USE whittaker,                       ONLY: whittaker_c0a,&
                                              whittaker_ci

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho_atom_methods'

! *** Public subroutines ***

   PUBLIC :: allocate_rho_atom_internals, &
             calculate_rho_atom, &
             calculate_rho_atom_coeff, &
             init_rho_atom

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param para_env ...
!> \param rho_atom_set ...
!> \param qs_kind ...
!> \param atom_list ...
!> \param grid_atom ...
!> \param natom ...
!> \param nspins ...
!> \param tot_rho1_h ...
!> \param tot_rho1_s ...
! **************************************************************************************************
   SUBROUTINE calculate_rho_atom(para_env, rho_atom_set, qs_kind, atom_list, &
                                 grid_atom, natom, nspins, tot_rho1_h, tot_rho1_s)

      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      INTEGER, DIMENSION(:), INTENT(IN)                  :: atom_list
      TYPE(grid_atom_type), INTENT(IN)                   :: grid_atom
      INTEGER, INTENT(IN)                                :: natom, nspins
      REAL(dp), DIMENSION(:), INTENT(INOUT)              :: tot_rho1_h, tot_rho1_s

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_atom', &
         routineP = moduleN//':'//routineN

      INTEGER :: damax_iso_not0_local, dbmax_iso_not0_local, dmax_iso_not0, handle, i, i1, i2, &
         iat, iatom, icg, ipgf1, ipgf2, ir, iset1, iset2, iso, iso1, iso1_first, iso1_last, iso2, &
         iso2_first, iso2_last, isom1, isom2, istat, j, l, l1, l2, l_iso, l_sub, l_sum, lmax12, &
         lmax_expansion, lmin12, m1, m1s, m2, m2s, max_iso_not0, max_iso_not0_local, max_s_harm, &
         maxl, maxso, mepos, n1s, n2s, nr, nset, num_pe, size1, size2
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list, dacg_n_list, dbcg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list, dacg_list, dbcg_list
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, o2nindex
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: done_vgg
      REAL(dp)                                           :: c1, c2, rho_h, rho_s, root_zet12, zet12
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dd, erf_zet12, g1, g2, gg0, int1, int2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: CPCH_sphere, CPCS_sphere, dgg, gg, gg_lm1
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: vgg
      REAL(dp), DIMENSION(:, :), POINTER                 :: coeff, zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG, my_dCG
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: my_CG_dxyz
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj

      CALL timeset(routineN, handle)

      NULLIFY (orb_basis)
      NULLIFY (harmonics)
      NULLIFY (lmin, lmax, npgf, zet, my_CG, my_dCG, my_CG_dxyz, coeff)

      CALL get_qs_kind(qs_kind, basis_set=orb_basis, &
                       paw_proj_set=paw_proj, harmonics=harmonics)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                             maxl=maxl, npgf=npgf, nset=nset, zet=zet, &
                             maxso=maxso)

      CALL get_paw_proj_set(paw_proj_set=paw_proj, o2nindex=o2nindex)

      max_iso_not0 = harmonics%max_iso_not0
      dmax_iso_not0 = harmonics%dmax_iso_not0
      max_s_harm = harmonics%max_s_harm

      nr = grid_atom%nr
      lmax_expansion = indso(1, max_iso_not0)
!   Distribute the atoms of this kind
      num_pe = para_env%num_pe
      mepos = para_env%mepos
      bo = get_limit(natom, num_pe, mepos)

      my_CG => harmonics%my_CG
      my_dCG => harmonics%my_dCG
      my_CG_dxyz => harmonics%my_CG_dxyz

      ALLOCATE (CPCH_sphere(nsoset(maxl), nsoset(maxl)))
      ALLOCATE (CPCS_sphere(nsoset(maxl), nsoset(maxl)))
      ALLOCATE (g1(nr), g2(nr), gg0(nr), gg(nr, 0:2*maxl), dgg(nr, 0:2*maxl), gg_lm1(nr, 0:2*maxl))
      ALLOCATE (erf_zet12(nr), vgg(nr, 0:2*maxl, 0:indso(1, max_iso_not0)))
      ALLOCATE (done_vgg(0:2*maxl, 0:indso(1, max_iso_not0)))
      ALLOCATE (dd(nr), int1(nr), int2(nr))
      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm), &
                dacg_list(2, nsoset(maxl)**2, max_s_harm), dacg_n_list(max_s_harm), &
                dbcg_list(2, nsoset(maxl)**2, max_s_harm), dbcg_n_list(max_s_harm), &
                STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",176)

      DO iat = bo(1), bo(2)
         iatom = atom_list(iat)
         DO i = 1, nspins
            IF (.NOT. ASSOCIATED(rho_atom_set(iatom)%rho_rad_h(i)%r_coef)) THEN
               CALL allocate_rho_atom_rad(rho_atom_set, iatom, i, nr, max_iso_not0)
            ELSE
               CALL set2zero_rho_atom_rad(rho_atom_set, iatom, i)
            ENDIF
         END DO
      END DO

      m1s = 0
      DO iset1 = 1, nset
         m2s = 0
         DO iset2 = 1, nset

            CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, lmax_expansion, cg_list, cg_n_list, max_iso_not0_local)
            IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_rho_atom_methods.F",196)
            CALL get_none0_cg_list(my_CG_dxyz, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, lmax_expansion, dacg_list, dacg_n_list, damax_iso_not0_local)
            CALL get_none0_cg_list(my_dCG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, lmax_expansion, dbcg_list, dbcg_n_list, dbmax_iso_not0_local)
            n1s = nsoset(lmax(iset1))

            DO ipgf1 = 1, npgf(iset1)
               iso1_first = nsoset(lmin(iset1)-1)+1+n1s*(ipgf1-1)+m1s
               iso1_last = nsoset(lmax(iset1))+n1s*(ipgf1-1)+m1s
               size1 = iso1_last-iso1_first+1
               iso1_first = o2nindex(iso1_first)
               iso1_last = o2nindex(iso1_last)
               i1 = iso1_last-iso1_first+1
               IF(.NOT.(size1 == i1))CALL cp__a("qs_rho_atom_methods.F",210)
               i1 = nsoset(lmin(iset1)-1)+1

               g1(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))

               n2s = nsoset(lmax(iset2))
               DO ipgf2 = 1, npgf(iset2)
                  iso2_first = nsoset(lmin(iset2)-1)+1+n2s*(ipgf2-1)+m2s
                  iso2_last = nsoset(lmax(iset2))+n2s*(ipgf2-1)+m2s
                  size2 = iso2_last-iso2_first+1
                  iso2_first = o2nindex(iso2_first)
                  iso2_last = o2nindex(iso2_last)
                  i2 = iso2_last-iso2_first+1
                  IF(.NOT.(size2 == i2))CALL cp__a("qs_rho_atom_methods.F",223)
                  i2 = nsoset(lmin(iset2)-1)+1

                  g2(1:nr) = EXP(-zet(ipgf2, iset2)*grid_atom%rad2(1:nr))
                  lmin12 = lmin(iset1)+lmin(iset2)
                  lmax12 = lmax(iset1)+lmax(iset2)

                  zet12 = zet(ipgf1, iset1)+zet(ipgf2, iset2)
                  root_zet12 = SQRT(zet(ipgf1, iset1)+zet(ipgf2, iset2))
                  DO ir = 1, nr
                     erf_zet12(ir) = erf(root_zet12*grid_atom%rad(ir))
                  END DO

                  gg = 0.0_dp
                  dgg = 0.0_dp
                  gg_lm1 = 0.0_dp
                  vgg = 0.0_dp
                  done_vgg = .FALSE.
                  ! reduce the number of terms in the expansion local densities
                  IF (lmin12 .LE. lmax_expansion) THEN
                     IF (lmin12 == 0) THEN
                        gg(1:nr, lmin12) = g1(1:nr)*g2(1:nr)
                        gg_lm1(1:nr, lmin12) = 0.0_dp
                        gg0(1:nr) = gg(1:nr, lmin12)
                     ELSE
                        gg0(1:nr) = g1(1:nr)*g2(1:nr)
                        gg(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12)*g1(1:nr)*g2(1:nr)
                        gg_lm1(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12-1)*g1(1:nr)*g2(1:nr)
                     END IF

                     ! reduce the number of terms in the expansion local densities
                     IF (lmax12 .GT. lmax_expansion) lmax12 = lmax_expansion

                     DO l = lmin12+1, lmax12
                        gg(1:nr, l) = grid_atom%rad(1:nr)*gg(1:nr, l-1)
                        gg_lm1(1:nr, l) = gg(1:nr, l-1)
                        dgg(1:nr, l-1) = dgg(1:nr, l-1)-2.0_dp*(zet(ipgf1, iset1) &
                                                                +zet(ipgf2, iset2))*gg(1:nr, l)

                     END DO
                     dgg(1:nr, lmax12) = dgg(1:nr, lmax12)-2.0_dp*(zet(ipgf1, iset1) &
                                                                   +zet(ipgf2, iset2))*grid_atom%rad(1:nr)*gg(1:nr, lmax12)

                     c2 = SQRT(pi*pi*pi/(zet12*zet12*zet12))

                     DO iso = 1, max_iso_not0_local
                        l_iso = indso(1, iso)
                        c1 = fourpi/(2._dp*REAL(l_iso, dp)+1._dp)
                        DO icg = 1, cg_n_list(iso)
                           iso1 = cg_list(1, icg, iso)
                           iso2 = cg_list(2, icg, iso)

                           l = indso(1, iso1)+indso(1, iso2)
                           IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_rho_atom_methods.F",276)
                           IF (done_vgg(l, l_iso)) CYCLE
                           L_sum = l+l_iso
                           L_sub = l-l_iso

                           IF (l_sum == 0) THEN
                              vgg(1:nr, l, l_iso) = erf_zet12(1:nr)*grid_atom%oorad2l(1:nr, 1)*c2
                           ELSE
                              CALL whittaker_c0a(int1, grid_atom%rad, gg0, erf_zet12, zet12, l, l_iso, nr)
                              CALL whittaker_ci(int2, grid_atom%rad, gg0, zet12, L_sub, nr)

                              DO ir = 1, nr
                                 int2(ir) = grid_atom%rad2l(ir, l_iso)*int2(ir)
                                 vgg(ir, l, l_iso) = c1*(int1(ir)+int2(ir))
                              END DO
                           END IF
                           done_vgg(l, l_iso) = .TRUE.
                        END DO
                     END DO
                  END IF ! lmax_expansion

                  DO iat = bo(1), bo(2)
                     iatom = atom_list(iat)

                     DO i = 1, nspins
                        CPCH_sphere = 0.0_dp
                        CPCS_sphere = 0.0_dp

                        coeff => rho_atom_set(iatom)%cpc_h(i)%r_coef
                        CPCH_sphere(i1:i1+size1-1, i2:i2+size2-1) = coeff(iso1_first:iso1_last, iso2_first:iso2_last)

                        coeff => rho_atom_set(iatom)%cpc_s(i)%r_coef
                        CPCS_sphere(i1:i1+size1-1, i2:i2+size2-1) = coeff(iso1_first:iso1_last, iso2_first:iso2_last)

                        DO iso = 1, max_iso_not0_local
                           l_iso = indso(1, iso)
                           DO icg = 1, cg_n_list(iso)
                              iso1 = cg_list(1, icg, iso)
                              iso2 = cg_list(2, icg, iso)

                              l1 = indso(1, iso1)
                              l2 = indso(1, iso2)
                              m1 = indso(2, iso1)
                              m2 = indso(2, iso2)
                              isom1 = soset(l1, -m1)
                              isom2 = soset(l2, -m2)

                              dd(1:nr) = REAL(l1*l2, dp)/grid_atom%rad2(1:nr) &
                                         -2._dp*REAL(l1, dp)*zet(ipgf2, iset2) &
                                         -2._dp*REAL(l2, dp)*zet(ipgf1, iset1) &
                                         +4._dp*zet(ipgf1, iset1)*zet(ipgf2, iset2)*grid_atom%rad2(1:nr)
                              l = indso(1, iso1)+indso(1, iso2)
                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_rho_atom_methods.F",328)

                              rho_atom_set(iatom)%rho_rad_h(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%rho_rad_h(i)%r_coef(1:nr, iso)+ &
                                 gg(1:nr, l)*CPCH_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%rho_rad_s(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%rho_rad_s(i)%r_coef(1:nr, iso)+ &
                                 gg(1:nr, l)*CPCS_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%drho_rad_h(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%drho_rad_h(i)%r_coef(1:nr, iso)+ &
                                 dgg(1:nr, l)*CPCH_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%drho_rad_s(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%drho_rad_s(i)%r_coef(1:nr, iso)+ &
                                 dgg(1:nr, l)*CPCS_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%vrho_rad_h(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%vrho_rad_h(i)%r_coef(1:nr, iso)+ &
                                 vgg(1:nr, l, l_iso)*CPCH_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%vrho_rad_s(i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%vrho_rad_s(i)%r_coef(1:nr, iso)+ &
                                 vgg(1:nr, l, l_iso)*CPCS_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%trho_rad_h(1, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_h(1, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*gg(1:nr, l)*dd(1:nr)*CPCH_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%trho_rad_s(1, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_s(1, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*gg(1:nr, l)*dd(1:nr)*CPCS_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)

                              rho_atom_set(iatom)%trho_rad_h(3, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_h(3, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*REAL(m1*m2, dp)*gg(1:nr, l)*CPCH_sphere(iso1, iso2)* &
                                 my_CG(isom1, isom2, iso)/grid_atom%rad2(1:nr)

                              rho_atom_set(iatom)%trho_rad_s(3, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_s(3, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*REAL(m1*m2, dp)*gg(1:nr, l)*CPCS_sphere(iso1, iso2)* &
                                 my_CG(isom1, isom2, iso)/grid_atom%rad2(1:nr)

                           ENDDO ! icg

                        ENDDO ! iso

                        DO iso = 1, max_iso_not0 !damax_iso_not0_local
                           l_iso = indso(1, iso)
                           DO icg = 1, dacg_n_list(iso)
                              iso1 = dacg_list(1, icg, iso)
                              iso2 = dacg_list(2, icg, iso)
                              l = indso(1, iso1)+indso(1, iso2)
                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_rho_atom_methods.F",382)
                              DO j = 1, 3
                                 rho_atom_set(iatom)%rho_rad_h_d(j, i)%r_coef(1:nr, iso) = &
                                    rho_atom_set(iatom)%rho_rad_h_d(j, i)%r_coef(1:nr, iso)+ &
                                    gg_lm1(1:nr, l)*CPCH_sphere(iso1, iso2)*my_CG_dxyz(j, iso1, iso2, iso)

                                 rho_atom_set(iatom)%rho_rad_s_d(j, i)%r_coef(1:nr, iso) = &
                                    rho_atom_set(iatom)%rho_rad_s_d(j, i)%r_coef(1:nr, iso)+ &
                                    gg_lm1(1:nr, l)*CPCS_sphere(iso1, iso2)*my_CG_dxyz(j, iso1, iso2, iso)
                              END DO
                           END DO ! icg

                        ENDDO ! iso

                        !Expansion for tau functionals (polar angle term)
                        !We ignore the contributions iso > max_iso_not0 (should be dmax_iso_not0)

                        DO iso = 1, max_iso_not0 !dbmax_iso_not0_local
                           l_iso = indso(1, iso)
                           DO icg = 1, dbcg_n_list(iso)
                              iso1 = dbcg_list(1, icg, iso)
                              iso2 = dbcg_list(2, icg, iso)

                              l = indso(1, iso1)+indso(1, iso2)
                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_rho_atom_methods.F",406)
                              rho_atom_set(iatom)%trho_rad_h(2, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_h(2, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*gg(1:nr, l)*CPCH_sphere(iso1, iso2)* &
                                 my_dCG(iso1, iso2, iso)/grid_atom%rad2(1:nr)

                              rho_atom_set(iatom)%trho_rad_s(2, i)%r_coef(1:nr, iso) = &
                                 rho_atom_set(iatom)%trho_rad_s(2, i)%r_coef(1:nr, iso)+ &
                                 0.5_dp*gg(1:nr, l)*CPCS_sphere(iso1, iso2)* &
                                 my_dCG(iso1, iso2, iso)/grid_atom%rad2(1:nr)
                           END DO ! icg
                        ENDDO ! iso

                     ENDDO ! i
                  ENDDO ! iat

               ENDDO ! ipgf2
            ENDDO ! ipgf1
            m2s = m2s+maxso
         ENDDO ! iset2
         m1s = m1s+maxso
      ENDDO ! iset1

      DO iat = bo(1), bo(2)
         iatom = atom_list(iat)

         DO i = 1, nspins

            DO iso = 1, max_iso_not0
               rho_s = 0.0_dp
               rho_h = 0.0_dp
               DO ir = 1, nr
                  rho_h = rho_h+rho_atom_set(iatom)%rho_rad_h(i)%r_coef(ir, iso)*grid_atom%wr(ir)
                  rho_s = rho_s+rho_atom_set(iatom)%rho_rad_s(i)%r_coef(ir, iso)*grid_atom%wr(ir)
               END DO ! ir
               tot_rho1_h(i) = tot_rho1_h(i)+rho_h*harmonics%slm_int(iso)
               tot_rho1_s(i) = tot_rho1_s(i)+rho_s*harmonics%slm_int(iso)
            END DO ! iso

         END DO ! ispin

      END DO ! iat

      DEALLOCATE (CPCH_sphere, CPCS_sphere)
      DEALLOCATE (g1, g2, gg0, gg, gg_lm1, dgg, vgg, done_vgg, erf_zet12, int1, int2)
      DEALLOCATE (cg_list, cg_n_list, dacg_list, dacg_n_list, dbcg_list, dbcg_n_list)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_atom

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rho_ao ...
!> \param rho_atom_set ...
!> \par History
!>      Add OpenMP [Apr 2016, EPCC]
! **************************************************************************************************
   SUBROUTINE calculate_rho_atom_coeff(qs_env, rho_ao, rho_atom_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho_atom_coeff', &
         routineP = moduleN//':'//routineN

      INTEGER :: bo(2), handle, i, iac, iatom, ibc, icol, ikind, img, irow, ispin, istat, jatom, &
         jkind, kac, katom, kbc, kkind, max_gau, max_nsgf, maxsoc, mepos, n_cont_a, n_cont_b, &
         nat_kind, natom, nimages, nkind, nsatbas, nsetc, nsoctot, nspins, num_pe
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      INTEGER, DIMENSION(3)                              :: cell_b
      INTEGER, DIMENSION(:), POINTER                     :: a_list, list_a, list_b
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: dista, distab, distb, found, paw_atom
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: p_matrix
      REAL(KIND=dp)                                      :: eps_cpc, factor, pmax
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: CPC, PC1
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: C_coeff_hh_a, C_coeff_hh_b, &
                                                            C_coeff_ss_a, C_coeff_ss_b, r_coef_h, &
                                                            r_coef_s
      TYPE(alist_type), POINTER                          :: alist_ac, alist_bc
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b, orb_basis_set
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(oce_matrix_type), POINTER                     :: oce
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: p_block_spin
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)

      NULLIFY (dft_control)
      NULLIFY (qs_kind_set)
      NULLIFY (oce)
      NULLIFY (sab_orb)
      NULLIFY (rho_atom)
      NULLIFY (orb_basis_set)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      para_env=para_env, &
                      oce=oce, sab_orb=sab_orb, &
                      rho_atom_set=rho_atom)

      eps_cpc = dft_control%qs_control%gapw_control%eps_cpc

      IF (PRESENT(rho_atom_set)) rho_atom => rho_atom_set

      nspins = dft_control%nspins
      nimages = dft_control%nimages

      NULLIFY (cell_to_index)
      IF (nimages > 1) THEN
         CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      END IF

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, maxsgf=max_nsgf, maxgtops=max_gau)

      nkind = SIZE(atomic_kind_set)
!   Initialize to 0 the CPC coefficients and the local density arrays
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=a_list, natom=nat_kind)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom)

         IF (.NOT. paw_atom) CYCLE
         DO i = 1, nat_kind
            iatom = a_list(i)
            DO ispin = 1, nspins
               rho_atom(iatom)%cpc_h(ispin)%r_coef = 0.0_dp
               rho_atom(iatom)%cpc_s(ispin)%r_coef = 0.0_dp
            ENDDO ! ispin
         ENDDO ! i

         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(nat_kind, num_pe, mepos)
         DO i = bo(1), bo(2)
            iatom = a_list(i)
            DO ispin = 1, nspins
               rho_atom(iatom)%ga_Vlocal_gb_h(ispin)%r_coef = 0.0_dp
               rho_atom(iatom)%ga_Vlocal_gb_s(ispin)%r_coef = 0.0_dp
            ENDDO ! ispin
         ENDDO ! i
      END DO ! ikind

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO

      num_pe = 1
!$    num_pe = omp_get_max_threads()
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb, nthread=num_pe)

!$OMP PARALLEL DEFAULT( NONE )                            &
!$OMP           SHARED( max_nsgf, max_gau                 &
!$OMP                 , nl_iterator, basis_set_list       &
!$OMP                 , nimages, cell_to_index            &
!$OMP                 , nspins, rho_ao                    &
!$OMP                 , nkind, qs_kind_set                &
!$OMP                 , oce, eps_cpc                      &
!$OMP                 , rho_atom                          &
!$OMP                 )                                   &
!$OMP          PRIVATE( p_block_spin, ispin               &
!$OMP                 , PC1, CPC, p_matrix, istat         &
!$OMP                 , mepos                             &
!$OMP                 , ikind, jkind, iatom, jatom        &
!$OMP                 , cell_b, rab                       &
!$OMP                 , basis_set_a, basis_set_b          &
!$OMP                 , pmax, irow, icol, img             &
!$OMP                 , found                             &
!$OMP                 , kkind, orb_basis_set, paw_proj    &
!$OMP                 , paw_atom, nsetc, maxsoc, nsatbas  &
!$OMP                 , nsoctot, katom                    &
!$OMP                 , iac , alist_ac, kac, n_cont_a, list_a     &
!$OMP                 , ibc , alist_bc, kbc, n_cont_b, list_b     &
!$OMP                 , C_coeff_hh_a, C_coeff_ss_a, dista         &
!$OMP                 , C_coeff_hh_b, C_coeff_ss_b, distb         &
!$OMP                 , distab                                    &
!$OMP                 , factor, r_coef_h, r_coef_s                &
!$OMP                 )

      ALLOCATE (p_block_spin(nspins))
      DO ispin = 1, nspins
         NULLIFY (p_block_spin(ispin)%r_coef)
      END DO

      ALLOCATE (PC1(max_nsgf*max_gau), CPC(max_gau*max_gau), p_matrix(max_nsgf, max_nsgf), STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",615)

      mepos = 0
!$    mepos = omp_get_thread_num()
      DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)

         CALL get_iterator_info(nl_iterator, mepos=mepos, &
                                ikind=ikind, jkind=jkind, &
                                iatom=iatom, jatom=jatom, &
                                cell=cell_b, r=rab)

         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE

         pmax = 0._dp
         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         ENDIF

         IF (nimages > 1) THEN
            img = cell_to_index(cell_b(1), cell_b(2), cell_b(3))
            IF(.NOT.(img > 0))CALL cp__a("qs_rho_atom_methods.F",642)
         ELSE
            img = 1
         END IF

         DO ispin = 1, nspins
            NULLIFY (p_block_spin(ispin)%r_coef)
            CALL cp_dbcsr_get_block_p(matrix=rho_ao(ispin, img)%matrix, &
                                      row=irow, col=icol, BLOCK=p_block_spin(ispin)%r_coef, &
                                      found=found)
            pmax = pmax+MAXVAL(ABS(p_block_spin(ispin)%r_coef))
         ENDDO

         DO kkind = 1, nkind
            CALL get_qs_kind(qs_kind_set(kkind), &
                             basis_set=orb_basis_set, &
                             paw_proj_set=paw_proj, paw_atom=paw_atom)

            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, nset=nsetc, maxso=maxsoc)

            IF (.NOT. paw_atom) CYCLE

            CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nsatbas)
            nsoctot = nsatbas

            iac = ikind+nkind*(kkind-1)
            ibc = jkind+nkind*(kkind-1)
            IF (.NOT. ASSOCIATED(oce%intac(iac)%alist)) CYCLE
            IF (.NOT. ASSOCIATED(oce%intac(ibc)%alist)) CYCLE

            CALL get_alist(oce%intac(iac), alist_ac, iatom)
            CALL get_alist(oce%intac(ibc), alist_bc, jatom)
            IF (.NOT. ASSOCIATED(alist_ac)) CYCLE
            IF (.NOT. ASSOCIATED(alist_bc)) CYCLE

            DO kac = 1, alist_ac%nclist
               DO kbc = 1, alist_bc%nclist
                  IF (alist_ac%clist(kac)%catom /= alist_bc%clist(kbc)%catom) CYCLE
                  IF (ALL(cell_b+alist_bc%clist(kbc)%cell-alist_ac%clist(kac)%cell == 0)) THEN
                     IF (pmax*alist_bc%clist(kbc)%maxac*alist_ac%clist(kac)%maxac < eps_cpc) CYCLE

                     n_cont_a = alist_ac%clist(kac)%nsgf_cnt
                     n_cont_b = alist_bc%clist(kbc)%nsgf_cnt
                     IF (n_cont_a .EQ. 0 .OR. n_cont_b .EQ. 0) CYCLE

                     list_a => alist_ac%clist(kac)%sgf_list
                     list_b => alist_bc%clist(kbc)%sgf_list

                     katom = alist_ac%clist(kac)%catom

                     IF (iatom == katom .AND. ALL(alist_ac%clist(kac)%cell == 0)) THEN
                        C_coeff_hh_a => alist_ac%clist(kac)%achint(:, :, 1)
                        C_coeff_ss_a => alist_ac%clist(kac)%acint(:, :, 1)
                        dista = .FALSE.
                     ELSE
                        C_coeff_hh_a => alist_ac%clist(kac)%acint(:, :, 1)
                        C_coeff_ss_a => alist_ac%clist(kac)%acint(:, :, 1)
                        dista = .TRUE.
                     END IF
                     IF (jatom == katom .AND. ALL(alist_bc%clist(kbc)%cell == 0)) THEN
                        C_coeff_hh_b => alist_bc%clist(kbc)%achint(:, :, 1)
                        C_coeff_ss_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        distb = .FALSE.
                     ELSE
                        C_coeff_hh_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        C_coeff_ss_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        distb = .TRUE.
                     END IF

                     distab = dista .AND. distb

                     DO ispin = 1, nspins

                        IF (iatom <= jatom) THEN
                           CALL alist_pre_align_blk(p_block_spin(ispin)%r_coef, &
                                                    SIZE(p_block_spin(ispin)%r_coef, 1), p_matrix, SIZE(p_matrix, 1), &
                                                    list_a, n_cont_a, list_b, n_cont_b)
                        ELSE
                           CALL alist_pre_align_blk(p_block_spin(ispin)%r_coef, &
                                                    SIZE(p_block_spin(ispin)%r_coef, 1), p_matrix, SIZE(p_matrix, 1), &
                                                    list_b, n_cont_b, list_a, n_cont_a)
                        ENDIF

                        factor = 1.0_dp
                        IF (iatom == jatom) factor = 0.5_dp

                        r_coef_h => rho_atom(katom)%cpc_h(ispin)%r_coef
                        r_coef_s => rho_atom(katom)%cpc_s(ispin)%r_coef

!$OMP CRITICAL(qs_rho_atom_methods_proj_blk)
                        IF (iatom <= jatom) THEN
                           CALL proj_blk(C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                         C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                         p_matrix, max_nsgf, r_coef_h, r_coef_s, nsoctot, &
                                         PC1, CPC, factor, distab)
                        ELSE
                           CALL proj_blk(C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                         C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                         p_matrix, max_nsgf, r_coef_h, r_coef_s, nsoctot, &
                                         PC1, CPC, factor, distab)
                        ENDIF
!$OMP END CRITICAL(qs_rho_atom_methods_proj_blk)

                     END DO
                     EXIT !search loop over jatom-katom list
                  END IF
               END DO
            END DO
         END DO
      END DO

      DEALLOCATE (PC1, CPC, p_block_spin, p_matrix)
!$OMP END PARALLEL

      CALL neighbor_list_iterator_release(nl_iterator)

      ALLOCATE (kind_of(natom))
      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)

      DO iatom = 1, natom
         ikind = kind_of(iatom)

         DO ispin = 1, nspins
            IF (ASSOCIATED(rho_atom(iatom)%cpc_h(ispin)%r_coef)) THEN
               CALL mp_sum(rho_atom(iatom)%cpc_h(ispin)%r_coef, para_env%group)
               CALL mp_sum(rho_atom(iatom)%cpc_s(ispin)%r_coef, para_env%group)
               r_coef_h => rho_atom(iatom)%cpc_h(ispin)%r_coef
               r_coef_s => rho_atom(iatom)%cpc_s(ispin)%r_coef
               r_coef_h(:, :) = r_coef_h(:, :)+TRANSPOSE(r_coef_h(:, :))
               r_coef_s(:, :) = r_coef_s(:, :)+TRANSPOSE(r_coef_s(:, :))
            END IF
         ENDDO

      END DO

      DEALLOCATE (kind_of, basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE calculate_rho_atom_coeff

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param gapw_control ...
! **************************************************************************************************
   SUBROUTINE init_rho_atom(qs_env, gapw_control)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(gapw_control_type), POINTER                   :: gapw_control

      CHARACTER(len=*), PARAMETER :: routineN = 'init_rho_atom', routineP = moduleN//':'//routineN

      INTEGER :: handle, ikind, il, iso, iso1, iso2, l1, l1l2, l2, la, lc1, lc2, lcleb, ll, llmax, &
         lmax_sphere, lp, m1, m2, max_s_harm, max_s_set, maxl, maxlgto, maxs, mm, mp, na, nat, &
         natom, nr, nspins, quadrature
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: paw_atom
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: rga
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set

      CALL timeset(routineN, handle)

      NULLIFY (rho_atom_set)
      NULLIFY (qs_kind_set, atomic_kind_set)
      NULLIFY (dft_control)
      NULLIFY (orb_basis_set)
      NULLIFY (my_CG, grid_atom, harmonics, atom_list)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control)
      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, maxlgto=maxlgto)

      nspins = dft_control%nspins

      lmax_sphere = gapw_control%lmax_sphere

      llmax = MIN(lmax_sphere, 2*maxlgto)
      max_s_harm = nsoset(llmax)
      max_s_set = nsoset(maxlgto)

      lcleb = MAX(llmax, 2*maxlgto, 1)

!   *** allocate calculate the CG coefficients up to the maxl ***
      CALL clebsch_gordon_init(lcleb)
      CALL reallocate(my_CG, 1, max_s_set, 1, max_s_set, 1, max_s_harm)

      ALLOCATE (rga(lcleb, 2))
      DO lc1 = 0, maxlgto
         DO iso1 = nsoset(lc1-1)+1, nsoset(lc1)
            l1 = indso(1, iso1)
            m1 = indso(2, iso1)
            DO lc2 = 0, maxlgto
               DO iso2 = nsoset(lc2-1)+1, nsoset(lc2)
                  l2 = indso(1, iso2)
                  m2 = indso(2, iso2)
                  CALL clebsch_gordon(l1, m1, l2, m2, rga)
                  IF (l1+l2 > llmax) THEN
                     l1l2 = llmax
                  ELSE
                     l1l2 = l1+l2
                  END IF
                  mp = m1+m2
                  mm = m1-m2
                  IF (m1*m2 < 0 .OR. (m1*m2 == 0 .AND. (m1 < 0 .OR. m2 < 0))) THEN
                     mp = -ABS(mp)
                     mm = -ABS(mm)
                  ELSE
                     mp = ABS(mp)
                     mm = ABS(mm)
                  END IF
                  DO lp = MOD(l1+l2, 2), l1l2, 2
                     il = lp/2+1
                     IF (ABS(mp) <= lp) THEN
                     IF (mp >= 0) THEN
                        iso = nsoset(lp-1)+lp+1+mp
                     ELSE
                        iso = nsoset(lp-1)+lp+1-ABS(mp)
                     END IF
                     my_CG(iso1, iso2, iso) = rga(il, 1)
                     ENDIF
                     IF (mp /= mm .AND. ABS(mm) <= lp) THEN
                     IF (mm >= 0) THEN
                        iso = nsoset(lp-1)+lp+1+mm
                     ELSE
                        iso = nsoset(lp-1)+lp+1-ABS(mm)
                     END IF
                     my_CG(iso1, iso2, iso) = rga(il, 2)
                     ENDIF
                  END DO
               ENDDO ! iso2
            ENDDO ! lc2
         ENDDO ! iso1
      ENDDO ! lc1
      DEALLOCATE (rga)
      CALL clebsch_gordon_deallocate()

!   *** initialize the Lebedev grids ***
      CALL init_lebedev_grids()
      quadrature = gapw_control%quadrature

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          paw_atom=paw_atom, &
                          grid_atom=grid_atom, &
                          harmonics=harmonics, &
                          ngrid_rad=nr, ngrid_ang=na)

!     *** determine the Lebedev grid for this kind ***

         ll = get_number_of_lebedev_grid(n=na)
         na = lebedev_grid(ll)%n
         la = lebedev_grid(ll)%l
         grid_atom%ng_sphere = na
         grid_atom%nr = nr

         IF (llmax > la) THEN
            WRITE (*, '(/,72("*"))')
            WRITE (*, '(T2,A,T66,I4)') &
               "WARNING: the lebedev grid is built for angular momentum l up to ", la, &
               "         the max l of spherical harmonics is larger, l_max = ", llmax, &
               "         good integration is guaranteed only for l <= ", la
            WRITE (*, '(72("*"),/)')
         END IF

!     *** calculate the radial grid ***
         CALL create_grid_atom(grid_atom, nr, na, llmax, ll, quadrature)

!     *** calculate the spherical harmonics on the grid ***

         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, maxl=maxl)
         maxs = nsoset(maxl)
         CALL create_harmonics_atom(harmonics, &
                                    my_CG, na, llmax, maxs, max_s_harm, ll, grid_atom%wa, &
                                    grid_atom%pol, grid_atom%cos_pol, &
                                    grid_atom%cotan_pol)
         CALL get_maxl_CG(harmonics, orb_basis_set, llmax, max_s_harm)

      END DO

      CALL deallocate_lebedev_grids()
      DEALLOCATE (my_CG)

      CALL allocate_rho_atom_internals(qs_env, rho_atom_set)

      CALL set_qs_env(qs_env=qs_env, rho_atom_set=rho_atom_set)

      CALL timestop(handle)

   END SUBROUTINE init_rho_atom

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rho_atom_set ...
! **************************************************************************************************
   SUBROUTINE allocate_rho_atom_internals(qs_env, rho_atom_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho_atom_internals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2), handle, iat, iatom, ikind, ispin, &
                                                            istat, j, max_iso_not0, mepos, nat, &
                                                            natom, nr, nsatbas, nsotot, nspins, &
                                                            num_pe
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: paw_atom
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind_set, para_env, rho_atom_set, dft_control)

      CALL get_qs_env(qs_env, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      para_env=para_env, &
                      dft_control=dft_control)

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)

      nspins = dft_control%nspins

      CALL allocate_rho_atom_set(rho_atom_set, natom)

      DO ikind = 1, SIZE(atomic_kind_set)

         NULLIFY (atom_list, harmonics)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          paw_proj_set=paw_proj, &
                          paw_atom=paw_atom, &
                          harmonics=harmonics, &
                          ngrid_rad=nr)

         IF (paw_atom) THEN
            CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nsatbas, nsotot=nsotot)
         END IF

         max_iso_not0 = harmonics%max_iso_not0
         DO iat = 1, nat
            iatom = atom_list(iat)
            !       *** allocate the radial density for each LM,for each atom ***

            ALLOCATE (rho_atom_set(iatom)%rho_rad_h(nspins))
            ALLOCATE (rho_atom_set(iatom)%rho_rad_s(nspins))
            ALLOCATE (rho_atom_set(iatom)%vrho_rad_h(nspins))
            ALLOCATE (rho_atom_set(iatom)%vrho_rad_s(nspins))

            istat = 0
            DO ispin = 1, nspins

               NULLIFY (rho_atom_set(iatom)%rho_rad_h(ispin)%r_coef, &
                        rho_atom_set(iatom)%rho_rad_s(ispin)%r_coef, &
                        rho_atom_set(iatom)%vrho_rad_h(ispin)%r_coef, &
                        rho_atom_set(iatom)%vrho_rad_s(ispin)%r_coef)
            END DO

            ALLOCATE (rho_atom_set(iatom)%cpc_h(nspins), &
                      rho_atom_set(iatom)%cpc_s(nspins), &
                      rho_atom_set(iatom)%drho_rad_h(nspins), &
                      rho_atom_set(iatom)%drho_rad_s(nspins), &
                      rho_atom_set(iatom)%trho_rad_h(3, nspins), &
                      rho_atom_set(iatom)%trho_rad_s(3, nspins), &
                      rho_atom_set(iatom)%rho_rad_h_d(3, nspins), &
                      rho_atom_set(iatom)%rho_rad_s_d(3, nspins), &
                      STAT=istat)
            IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1027)

            IF (paw_atom) THEN
               DO ispin = 1, nspins
                  NULLIFY (rho_atom_set(iatom)%drho_rad_h(ispin)%r_coef, &
                           rho_atom_set(iatom)%drho_rad_s(ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_h(1, ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_h(2, ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_h(3, ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_s(1, ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_s(2, ispin)%r_coef, &
                           rho_atom_set(iatom)%trho_rad_s(3, ispin)%r_coef)
                  ALLOCATE (rho_atom_set(iatom)%cpc_h(ispin)%r_coef(1:nsatbas, 1:nsatbas), &
                            rho_atom_set(iatom)%cpc_s(ispin)%r_coef(1:nsatbas, 1:nsatbas), &
                            STAT=istat)
                  IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1042)

                  rho_atom_set(iatom)%cpc_h(ispin)%r_coef = 0.0_dp
                  rho_atom_set(iatom)%cpc_s(ispin)%r_coef = 0.0_dp
               END DO

            ELSE
               DO ispin = 1, nspins
                  NULLIFY (rho_atom_set(iatom)%cpc_h(ispin)%r_coef)
                  NULLIFY (rho_atom_set(iatom)%cpc_s(ispin)%r_coef)
                  NULLIFY (rho_atom_set(iatom)%drho_rad_h(ispin)%r_coef)
                  NULLIFY (rho_atom_set(iatom)%drho_rad_s(ispin)%r_coef)

                  DO j = 1, 3
                     NULLIFY (rho_atom_set(iatom)%trho_rad_h(j, ispin)%r_coef)
                     NULLIFY (rho_atom_set(iatom)%trho_rad_s(j, ispin)%r_coef)
                     NULLIFY (rho_atom_set(iatom)%rho_rad_h_d(j, ispin)%r_coef)
                     NULLIFY (rho_atom_set(iatom)%rho_rad_s_d(j, ispin)%r_coef)
                  END DO
               END DO
            ENDIF

         END DO ! iat

         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(nat, num_pe, mepos)
         DO iat = bo(1), bo(2)
            iatom = atom_list(iat)
            ALLOCATE (rho_atom_set(iatom)%ga_Vlocal_gb_h(nspins), &
                      rho_atom_set(iatom)%ga_Vlocal_gb_s(nspins), &
                      STAT=istat)
            IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1074)
            IF (paw_atom) THEN
               DO ispin = 1, nspins
                  NULLIFY (rho_atom_set(iatom)%ga_Vlocal_gb_h(ispin)%r_coef)
                  NULLIFY (rho_atom_set(iatom)%ga_Vlocal_gb_s(ispin)%r_coef)

                  CALL reallocate(rho_atom_set(iatom)%ga_Vlocal_gb_h(ispin)%r_coef, &
                                  1, nsotot, 1, nsotot)
                  CALL reallocate(rho_atom_set(iatom)%ga_Vlocal_gb_s(ispin)%r_coef, &
                                  1, nsotot, 1, nsotot)

                  rho_atom_set(iatom)%ga_Vlocal_gb_h(ispin)%r_coef = 0.0_dp
                  rho_atom_set(iatom)%ga_Vlocal_gb_s(ispin)%r_coef = 0.0_dp
               END DO
            ELSE
               DO ispin = 1, nspins
                  NULLIFY (rho_atom_set(iatom)%ga_Vlocal_gb_h(ispin)%r_coef)
                  NULLIFY (rho_atom_set(iatom)%ga_Vlocal_gb_s(ispin)%r_coef)
               END DO
            ENDIF

         END DO ! iat

      END DO

      CALL timestop(handle)

   END SUBROUTINE allocate_rho_atom_internals

! **************************************************************************************************
!> \brief ...
!> \param rho_atom_set ...
!> \param iatom ...
!> \param ispin ...
!> \param nr ...
!> \param max_iso_not0 ...
! **************************************************************************************************
   SUBROUTINE allocate_rho_atom_rad(rho_atom_set, iatom, ispin, nr, max_iso_not0)

      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      INTEGER, INTENT(IN)                                :: iatom, ispin, nr, max_iso_not0

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho_atom_rad', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, istat, j

      CALL timeset(routineN, handle)

      ALLOCATE (rho_atom_set(iatom)%rho_rad_h(ispin)%r_coef(1:nr, 1:max_iso_not0), &
                rho_atom_set(iatom)%rho_rad_s(ispin)%r_coef(1:nr, 1:max_iso_not0), &
                rho_atom_set(iatom)%vrho_rad_h(ispin)%r_coef(1:nr, 1:max_iso_not0), &
                rho_atom_set(iatom)%vrho_rad_s(ispin)%r_coef(1:nr, 1:max_iso_not0), &
                STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1128)

      rho_atom_set(iatom)%rho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%rho_rad_s(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%vrho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%vrho_rad_s(ispin)%r_coef = 0.0_dp

      ALLOCATE (rho_atom_set(iatom)%drho_rad_h(ispin)%r_coef(nr, max_iso_not0), &
                rho_atom_set(iatom)%drho_rad_s(ispin)%r_coef(nr, max_iso_not0), &
                STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1138)
      rho_atom_set(iatom)%drho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%drho_rad_s(ispin)%r_coef = 0.0_dp

      DO j = 1, 3
         ALLOCATE (rho_atom_set(iatom)%trho_rad_h(j, ispin)%r_coef(nr, max_iso_not0), &
                   rho_atom_set(iatom)%trho_rad_s(j, ispin)%r_coef(nr, max_iso_not0), &
                   STAT=istat)
         IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1146)
         rho_atom_set(iatom)%trho_rad_h(j, ispin)%r_coef = 0.0_dp
         rho_atom_set(iatom)%trho_rad_s(j, ispin)%r_coef = 0.0_dp
      END DO

      DO j = 1, 3
         NULLIFY (rho_atom_set(iatom)%rho_rad_h_d(j, ispin)%r_coef, &
                  rho_atom_set(iatom)%rho_rad_s_d(j, ispin)%r_coef)
         ALLOCATE (rho_atom_set(iatom)%rho_rad_h_d(j, ispin)%r_coef(nr, max_iso_not0), &
                   rho_atom_set(iatom)%rho_rad_s_d(j, ispin)%r_coef(nr, max_iso_not0), &
                   STAT=istat)
         IF(.NOT.(istat == 0))CALL cp__a("qs_rho_atom_methods.F",1157)
         rho_atom_set(iatom)%rho_rad_h_d(j, ispin)%r_coef = 0.0_dp
         rho_atom_set(iatom)%rho_rad_s_d(j, ispin)%r_coef = 0.0_dp
      END DO

      CALL timestop(handle)

   END SUBROUTINE allocate_rho_atom_rad

! **************************************************************************************************
!> \brief ...
!> \param rho_atom_set ...
!> \param iatom ...
!> \param ispin ...
! **************************************************************************************************
   SUBROUTINE set2zero_rho_atom_rad(rho_atom_set, iatom, ispin)

!    TYPE(rho_atom_type), POINTER             :: rho_atom
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      INTEGER, INTENT(IN)                                :: iatom, ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'set2zero_rho_atom_rad', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j

      rho_atom_set(iatom)%rho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%rho_rad_s(ispin)%r_coef = 0.0_dp

      rho_atom_set(iatom)%vrho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%vrho_rad_s(ispin)%r_coef = 0.0_dp

      rho_atom_set(iatom)%drho_rad_h(ispin)%r_coef = 0.0_dp
      rho_atom_set(iatom)%drho_rad_s(ispin)%r_coef = 0.0_dp

      DO j = 1, 3
         rho_atom_set(iatom)%trho_rad_h(j, ispin)%r_coef = 0.0_dp
         rho_atom_set(iatom)%trho_rad_s(j, ispin)%r_coef = 0.0_dp
         rho_atom_set(iatom)%rho_rad_h_d(j, ispin)%r_coef = 0.0_dp
         rho_atom_set(iatom)%rho_rad_s_d(j, ispin)%r_coef = 0.0_dp
      END DO

   END SUBROUTINE set2zero_rho_atom_rad

END MODULE qs_rho_atom_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods of the rho structure (defined in qs_rho_types)
!> \par History
!>      08.2002 created [fawzi]
!>      08.2014 kpoints [JGH]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_rho_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_distribution, cp_dbcsr_init, cp_dbcsr_p_type, &
        cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_symmetric
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE kpoint_types,                    ONLY: kpoint_type
   USE lri_environment_methods,         ONLY: calculate_lri_densities
   USE lri_environment_types,           ONLY: lri_density_type,&
                                              lri_environment_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qs_collocate_density,            ONLY: calculate_drho_elec,&
                                              calculate_rho_elec
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_local_rho_types,              ONLY: local_rho_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_atom_methods,             ONLY: calculate_rho_atom_coeff
   USE qs_rho_types,                    ONLY: qs_rho_clear,&
                                              qs_rho_get,&
                                              qs_rho_set,&
                                              qs_rho_type
   USE task_list_types,                 ONLY: task_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho_methods'

   PUBLIC :: qs_rho_update_rho, qs_rho_rebuild, duplicate_rho_type

CONTAINS

! **************************************************************************************************
!> \brief rebuilds rho (if necessary allocating and initializing it)
!> \param rho the rho type to rebuild (defaults to qs_env%rho)
!> \param qs_env the environment to which rho belongs
!> \param rebuild_ao if it is necessary to rebuild rho_ao. Defaults to true.
!> \param rebuild_grids if it in necessary to rebuild rho_r and rho_g.
!>        Defaults to false.
!> \param admm (use aux_fit basis)
!> \par History
!>      11.2002 created replacing qs_rho_create and qs_env_rebuild_rho[fawzi]
!> \author Fawzi Mohamed
!> \note
!>      needs updated  pw pools, s, s_mstruct and h in qs_env.
!>      The use of p to keep the structure of h (needed for the forces)
!>      is ugly and should be removed.
!>      Change so that it does not allocate a subcomponent if it is not
!>      associated and not requested?
! **************************************************************************************************
   SUBROUTINE qs_rho_rebuild(rho, qs_env, rebuild_ao, rebuild_grids, admm)
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in), OPTIONAL                      :: rebuild_ao, rebuild_grids, admm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_rho_rebuild', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: headline
      INTEGER                                            :: handle, i, ic, nimg, nspins
      LOGICAL                                            :: do_kpoints, my_admm, my_rebuild_ao, &
                                                            my_rebuild_grids
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s_kp, rho_ao_kp
      TYPE(cp_dbcsr_type), POINTER                       :: refmatrix, tmatrix
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: drho_g, drho_r, rho_g, rho_r, tau_g, &
                                                            tau_r
      TYPE(pw_p_type), POINTER                           :: rho_r_sccs
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, auxbas_pw_pool, matrix_s, matrix_s_kp, dft_control)
      NULLIFY (tot_rho_r, rho_ao_kp, rho_r, rho_g, drho_r, drho_g, tau_r, tau_g)
      NULLIFY (rho_r_sccs)
      NULLIFY (sab_orb)
      my_rebuild_ao = .TRUE.
      my_rebuild_grids = .TRUE.
      my_admm = .FALSE.
      IF (PRESENT(rebuild_ao)) my_rebuild_ao = rebuild_ao
      IF (PRESENT(rebuild_grids)) my_rebuild_grids = rebuild_grids
      IF (PRESENT(admm)) my_admm = admm

      CALL get_qs_env(qs_env, &
                      kpoints=kpoints, &
                      do_kpoints=do_kpoints, &
                      pw_env=pw_env, &
                      dft_control=dft_control)

      nimg = dft_control%nimages

      IF (my_admm) THEN
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_rho_methods.F",130)
         CALL get_qs_env(qs_env, matrix_s_aux_fit=matrix_s, sab_aux_fit=sab_orb)
         refmatrix => matrix_s(1)%matrix
      ELSE
         CALL get_qs_env(qs_env, matrix_s_kp=matrix_s_kp, sab_orb=sab_orb)
         refmatrix => matrix_s_kp(1, 1)%matrix
      END IF

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
      nspins = dft_control%nspins

      IF (.NOT. ASSOCIATED(rho)) CALL cp__b("qs_rho_methods.F",141,"rho not associated")
      CALL qs_rho_get(rho, &
                      tot_rho_r=tot_rho_r, &
                      rho_ao_kp=rho_ao_kp, &
                      rho_r=rho_r, &
                      rho_g=rho_g, &
                      drho_r=drho_r, &
                      drho_g=drho_g, &
                      tau_r=tau_r, &
                      tau_g=tau_g, &
                      rho_r_sccs=rho_r_sccs)

      IF (.NOT. ASSOCIATED(tot_rho_r)) THEN
         ALLOCATE (tot_rho_r(nspins))
         tot_rho_r = 0.0_dp
         CALL qs_rho_set(rho, tot_rho_r=tot_rho_r)
      END IF

      ! rho_ao
      IF (my_rebuild_ao .OR. (.NOT. ASSOCIATED(rho_ao_kp))) THEN
         IF (ASSOCIATED(rho_ao_kp)) &
            CALL cp_dbcsr_deallocate_matrix_set(rho_ao_kp)
         ! Create a new density matrix set
         CALL cp_dbcsr_allocate_matrix_set(rho_ao_kp, nspins, nimg)
         CALL qs_rho_set(rho, rho_ao_kp=rho_ao_kp)
         DO i = 1, nspins
            DO ic = 1, nimg
               IF (nspins > 1) THEN
                  IF (i == 1) THEN
                     headline = "DENSITY MATRIX FOR ALPHA SPIN"
                  ELSE
                     headline = "DENSITY MATRIX FOR BETA SPIN"
                  END IF
               ELSE
                  headline = "DENSITY MATRIX"
               END IF
               ALLOCATE (rho_ao_kp(i, ic)%matrix)
               tmatrix => rho_ao_kp(i, ic)%matrix
               CALL cp_dbcsr_init(tmatrix)
               CALL cp_dbcsr_create(matrix=tmatrix, name=TRIM(headline), &
                                    dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                    row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                    col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                    nze=0)
               CALL cp_dbcsr_alloc_block_from_nbl(tmatrix, sab_orb)
               CALL cp_dbcsr_set(tmatrix, 0.0_dp)
            END DO
         END DO
      END IF

      ! rho_r
      IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(rho_r)) THEN
         IF (ASSOCIATED(rho_r)) THEN
            DO i = 1, SIZE(rho_r)
               CALL pw_release(rho_r(i)%pw)
            END DO
            DEALLOCATE (rho_r)
         END IF
         ALLOCATE (rho_r(nspins))
         CALL qs_rho_set(rho, rho_r=rho_r)
         DO i = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r(i)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
         END DO
      END IF

      ! rho_g
      IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(rho_g)) THEN
         IF (ASSOCIATED(rho_g)) THEN
            DO i = 1, SIZE(rho_g)
               CALL pw_release(rho_g(i)%pw)
            END DO
            DEALLOCATE (rho_g)
         END IF
         ALLOCATE (rho_g(nspins))
         CALL qs_rho_set(rho, rho_g=rho_g)
         DO i = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_g(i)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         END DO
      END IF

      ! SCCS
      IF (dft_control%do_sccs) THEN
         IF (my_rebuild_grids .OR. (.NOT. ASSOCIATED(rho_r_sccs))) THEN
            IF (ASSOCIATED(rho_r_sccs)) THEN
               CALL pw_release(rho_r_sccs%pw)
               DEALLOCATE (rho_r_sccs)
            END IF
            ALLOCATE (rho_r_sccs)
            CALL qs_rho_set(rho, rho_r_sccs=rho_r_sccs)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r_sccs%pw, &
                                   use_data=REALDATA3D, &
                                   in_space=REALSPACE)
            CALL pw_zero(rho_r_sccs%pw)
         END IF
      END IF

      ! allocate drho_r and drho_g if xc_deriv_collocate
      IF (dft_control%drho_by_collocation) THEN
         ! drho_r
         IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(drho_r)) THEN
            IF (ASSOCIATED(drho_r)) THEN
               DO i = 1, SIZE(drho_r)
                  CALL pw_release(drho_r(i)%pw)
               END DO
               DEALLOCATE (drho_r)
            END IF
            ALLOCATE (drho_r(3*nspins))
            CALL qs_rho_set(rho, drho_r=drho_r)
            DO i = 1, 3*nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, drho_r(i)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
            END DO
         END IF
         ! drho_g
         IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(drho_g)) THEN
            IF (ASSOCIATED(drho_g)) THEN
               DO i = 1, SIZE(drho_g)
                  CALL pw_release(drho_g(i)%pw)
               END DO
               DEALLOCATE (drho_g)
            END IF
            ALLOCATE (drho_g(3*nspins))
            CALL qs_rho_set(rho, drho_g=drho_g)
            DO i = 1, 3*nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, drho_g(i)%pw, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            END DO
         END IF
      END IF

      ! allocate tau_r and tau_g if use_kinetic_energy_density
      IF (dft_control%use_kinetic_energy_density) THEN
         ! tau_r
         IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(tau_r)) THEN
            IF (ASSOCIATED(tau_r)) THEN
               DO i = 1, SIZE(tau_r)
                  CALL pw_release(tau_r(i)%pw)
               END DO
               DEALLOCATE (tau_r)
            END IF
            ALLOCATE (tau_r(nspins))
            CALL qs_rho_set(rho, tau_r=tau_r)
            DO i = 1, nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, tau_r(i)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
            END DO
         END IF

         ! tau_g
         IF (my_rebuild_grids .OR. .NOT. ASSOCIATED(tau_g)) THEN
            IF (ASSOCIATED(tau_g)) THEN
               DO i = 1, SIZE(tau_g)
                  CALL pw_release(tau_g(i)%pw)
               END DO
               DEALLOCATE (tau_g)
            END IF
            ALLOCATE (tau_g(nspins))
            CALL qs_rho_set(rho, tau_g=tau_g)
            DO i = 1, nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, tau_g(i)%pw, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            END DO
         END IF
      END IF ! use_kinetic_energy_density

      CALL timestop(handle)

   END SUBROUTINE qs_rho_rebuild

! **************************************************************************************************
!> \brief updates rho_r and rho_g to the rho%rho_ao.
!>      if use_kinetic_energy_density also computes tau_r and tau_g
!> \param rho_struct the rho structure that should be updated
!> \param qs_env the qs_env rho_struct refers to
!>        the integrated charge in r space
!> \param local_rho_set ...
!> \param task_list_external ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_rho_update_rho(rho_struct, qs_env, local_rho_set, task_list_external)
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(local_rho_type), OPTIONAL, POINTER            :: local_rho_set
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list_external

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_update_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, img, ispin, nimg, nspins
      LOGICAL                                            :: gapw, gapw_xc
      REAL(KIND=dp)                                      :: dum
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r, tot_rho_r_xc
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp, rho_xc_ao
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: drho_g, drho_r, drho_xc_g, rho_g, rho_r, &
                                                            rho_xc_g, rho_xc_r, tau_g, tau_r, &
                                                            tau_xc_g, tau_xc_r
      TYPE(pw_p_type), POINTER                           :: rho_r_sccs
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho_xc
      TYPE(task_list_type), POINTER                      :: task_list

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, rho_xc, ks_env, rho_ao, rho_r, rho_g, drho_r, drho_g, tau_r, tau_g)
      NULLIFY (rho_xc_ao, rho_xc_g, rho_xc_r, drho_xc_g, tau_xc_r, tau_xc_g, tot_rho_r, tot_rho_r_xc)
      NULLIFY (lri_env, lri_density, para_env, atomic_kind_set)

      IF(.NOT.(ASSOCIATED(rho_struct)))CALL cp__a("qs_rho_methods.F",358)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      task_list=task_list, &
                      lri_density=lri_density, &
                      lri_env=lri_env, &
                      atomic_kind_set=atomic_kind_set, &
                      para_env=para_env)

      CALL qs_rho_get(rho_struct, &
                      rho_r=rho_r, &
                      rho_g=rho_g, &
                      tot_rho_r=tot_rho_r, &
                      drho_r=drho_r, &
                      drho_g=drho_g, &
                      tau_r=tau_r, &
                      tau_g=tau_g, &
                      rho_r_sccs=rho_r_sccs)

      IF (PRESENT(task_list_external)) task_list => task_list_external

      nspins = dft_control%nspins
      nimg = dft_control%nimages
      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc

      IF (dft_control%qs_control%semi_empirical) THEN
         !
         CALL qs_rho_set(rho_struct, rho_r_valid=.FALSE., rho_g_valid=.FALSE.)
      ELSEIF (dft_control%qs_control%dftb) THEN
         !
         CALL qs_rho_set(rho_struct, rho_r_valid=.FALSE., rho_g_valid=.FALSE.)
      ELSEIF (dft_control%qs_control%scptb) THEN
         !
         CALL qs_rho_set(rho_struct, rho_r_valid=.FALSE., rho_g_valid=.FALSE.)
      ELSEIF (dft_control%qs_control%lrigpw) THEN
         CALL qs_rho_get(rho_struct, rho_ao=rho_ao)
         CALL calculate_lri_densities(lri_env, lri_density, qs_env, rho_ao, &
                                      lri_rho_struct=rho_struct, &
                                      atomic_kind_set=atomic_kind_set, &
                                      para_env=para_env)
         CALL qs_rho_set(rho_struct, rho_r_valid=.TRUE., rho_g_valid=.TRUE.)
      ELSE
         CALL qs_rho_get(rho_struct, rho_ao_kp=rho_ao_kp)
         DO ispin = 1, nspins
            rho_ao => rho_ao_kp(ispin, :)
            CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                    rho=rho_r(ispin), &
                                    rho_gspace=rho_g(ispin), &
                                    total_rho=tot_rho_r(ispin), &
                                    ks_env=ks_env, soft_valid=gapw, &
                                    task_list_external=task_list)
         END DO
         CALL qs_rho_set(rho_struct, rho_r_valid=.TRUE., rho_g_valid=.TRUE.)
      END IF

      ! if needed compute also the gradient of the density
      IF (dft_control%drho_by_collocation) THEN
         CALL qs_rho_get(rho_struct, rho_ao_kp=rho_ao_kp)
         DO ispin = 1, nspins
            IF(.NOT.(.NOT. PRESENT(task_list_external)))CALL cp__a("qs_rho_methods.F",420)
            rho_ao => rho_ao_kp(ispin, :)
            CALL calculate_drho_elec(matrix_p_kp=rho_ao, &
                                     drho=drho_r(3*(ispin-1)+1:3*ispin), &
                                     drho_gspace=drho_g(3*(ispin-1)+1:3*ispin), &
                                     qs_env=qs_env, soft_valid=gapw)
         END DO
         CALL qs_rho_set(rho_struct, drho_r_valid=.TRUE., drho_g_valid=.TRUE.)
      ENDIF

      ! if needed compute also the kinetic energy density
      IF (dft_control%use_kinetic_energy_density) THEN
         IF (dft_control%qs_control%lrigpw) THEN
            CALL cp__b("qs_rho_methods.F",433,"LRIGPW not implemented for meta-GGAs")
         ENDIF
         CALL qs_rho_get(rho_struct, rho_ao_kp=rho_ao_kp)
         DO ispin = 1, nspins
            IF(.NOT.(.NOT. PRESENT(task_list_external)))CALL cp__a("qs_rho_methods.F",437)
            rho_ao => rho_ao_kp(ispin, :)
            CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                    rho=tau_r(ispin), &
                                    rho_gspace=tau_g(ispin), &
                                    total_rho=dum, & ! presumably not meaningful
                                    ks_env=ks_env, soft_valid=gapw, &
                                    compute_tau=.TRUE.)
         END DO
         CALL qs_rho_set(rho_struct, tau_r_valid=.TRUE., tau_g_valid=.TRUE.)
      ENDIF

      ! GAPW o GAPW_XC require the calculation of hard and soft local densities
      IF (gapw) THEN
         IF(.NOT.(.NOT. PRESENT(task_list_external)))CALL cp__a("qs_rho_methods.F",451)
         CALL qs_rho_get(rho_struct, rho_ao_kp=rho_ao_kp)
         IF (PRESENT(local_rho_set)) THEN
            CALL calculate_rho_atom_coeff(qs_env, rho_ao_kp, local_rho_set%rho_atom_set)
         ELSE
            CALL calculate_rho_atom_coeff(qs_env, rho_ao_kp)
         ENDIF
      ENDIF
      IF (gapw_xc) THEN
         IF(.NOT.(.NOT. PRESENT(task_list_external)))CALL cp__a("qs_rho_methods.F",460)
         CALL get_qs_env(qs_env=qs_env, rho_xc=rho_xc)
         CALL qs_rho_get(rho_xc, &
                         rho_ao_kp=rho_xc_ao, &
                         rho_r=rho_xc_r, &
                         rho_g=rho_xc_g, &
                         tot_rho_r=tot_rho_r_xc, &
                         drho_g=drho_xc_g, &
                         tau_r=tau_xc_r, &
                         tau_g=tau_xc_g)
         CALL calculate_rho_atom_coeff(qs_env, rho_ao_kp)
         ! copy rho_ao into rho_xc_ao
         DO ispin = 1, nspins
            DO img = 1, nimg
               CALL cp_dbcsr_copy(rho_xc_ao(ispin, img)%matrix, rho_ao_kp(ispin, img)%matrix)
            END DO
         END DO
         DO ispin = 1, nspins
            rho_ao => rho_xc_ao(ispin, :)
            CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                    rho=rho_xc_r(ispin), &
                                    rho_gspace=rho_xc_g(ispin), &
                                    total_rho=tot_rho_r_xc(ispin), &
                                    ks_env=ks_env, soft_valid=gapw_xc)
         END DO
         CALL qs_rho_set(rho_xc, rho_r_valid=.TRUE., rho_g_valid=.TRUE.)
         ! if needed compute also the gradient of the density
         IF (dft_control%drho_by_collocation) THEN
            DO ispin = 1, nspins
               rho_ao => rho_xc_ao(ispin, :)
               CALL calculate_drho_elec(matrix_p_kp=rho_ao, &
                                        drho=rho_xc_r(3*(ispin-1)+1:3*ispin), &
                                        drho_gspace=drho_xc_g(3*(ispin-1)+1:3*ispin), &
                                        qs_env=qs_env, soft_valid=gapw_xc)
            END DO
            CALL qs_rho_set(rho_xc, drho_r_valid=.TRUE., drho_g_valid=.TRUE.)
         ENDIF
         ! if needed compute also the kinetic energy density
         IF (dft_control%use_kinetic_energy_density) THEN
            DO ispin = 1, nspins
               rho_ao => rho_xc_ao(ispin, :)
               CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                       rho=tau_xc_r(ispin), &
                                       rho_gspace=tau_xc_g(ispin), &
                                       total_rho=dum, & ! presumably not meaningful
                                       ks_env=ks_env, soft_valid=gapw_xc, &
                                       compute_tau=.TRUE.)
            END DO
            CALL qs_rho_set(rho_xc, tau_r_valid=.TRUE., tau_g_valid=.TRUE.)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_rho_update_rho

! **************************************************************************************************
!> \brief Duplicates a pointer physically
!> \param rho_input The rho structure to be duplicated
!> \param rho_output The duplicate rho structure
!> \param qs_env The QS environment from which the auxiliary PW basis-set
!>                pool is taken
!> \par History
!>      07.2005 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
!> \note
!>      Associated pointers are deallocated, nullified pointers are NOT accepted!
! **************************************************************************************************
   SUBROUTINE duplicate_rho_type(rho_input, rho_output, qs_env)

      TYPE(qs_rho_type), POINTER                         :: rho_input, rho_output
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'duplicate_rho_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, nspins, rebuild_each_in
      LOGICAL :: drho_g_valid_in, drho_r_valid_in, rho_g_valid_in, rho_r_valid_in, soft_valid_in, &
         tau_g_valid_in, tau_r_valid_in
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_g_in, tot_rho_g_out, &
                                                            tot_rho_r_in, tot_rho_r_out
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao_in, rho_ao_out
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER :: drho_g_in, drho_g_out, drho_r_in, drho_r_out, &
         rho_g_in, rho_g_out, rho_r_in, rho_r_out, tau_g_in, tau_g_out, tau_r_in, tau_r_out
      TYPE(pw_p_type), POINTER                           :: rho_r_sccs_in, rho_r_sccs_out
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, pw_env, auxbas_pw_pool)
      NULLIFY (rho_ao_in, rho_ao_out)
      NULLIFY (rho_r_in, rho_r_out, rho_g_in, rho_g_out, drho_r_in, drho_r_out)
      NULLIFY (drho_g_in, drho_g_out, tau_r_in, tau_r_out, tau_g_in, tau_g_out)
      NULLIFY (tot_rho_r_in, tot_rho_r_out, tot_rho_g_in, tot_rho_g_out)
      NULLIFY (rho_r_sccs_in, rho_r_sccs_out)

      IF(.NOT.(ASSOCIATED(rho_input)))CALL cp__a("qs_rho_methods.F",558)
      IF(.NOT.(ASSOCIATED(rho_output)))CALL cp__a("qs_rho_methods.F",559)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_rho_methods.F",560)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_rho_methods.F",561)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, dft_control=dft_control)
      CALL pw_env_get(pw_env=pw_env, auxbas_pw_pool=auxbas_pw_pool)
      nspins = dft_control%nspins

      CALL qs_rho_clear(rho_output)

      CALL qs_rho_get(rho_input, &
                      rho_ao=rho_ao_in, &
                      rho_r=rho_r_in, &
                      rho_g=rho_g_in, &
                      drho_r=drho_r_in, &
                      drho_g=drho_g_in, &
                      tau_r=tau_r_in, &
                      tau_g=tau_g_in, &
                      tot_rho_r=tot_rho_r_in, &
                      tot_rho_g=tot_rho_g_in, &
                      rho_g_valid=rho_g_valid_in, &
                      rho_r_valid=rho_r_valid_in, &
                      drho_g_valid=drho_g_valid_in, &
                      drho_r_valid=drho_r_valid_in, &
                      tau_r_valid=tau_r_valid_in, &
                      tau_g_valid=tau_g_valid_in, &
                      rho_r_sccs=rho_r_sccs_in, &
                      soft_valid=soft_valid_in, &
                      rebuild_each=rebuild_each_in)

      ! rho_ao
      IF (ASSOCIATED(rho_ao_in)) THEN
         CALL cp_dbcsr_allocate_matrix_set(rho_ao_out, nspins)
         CALL qs_rho_set(rho_output, rho_ao=rho_ao_out)
         DO i = 1, nspins
            ALLOCATE (rho_ao_out(i)%matrix)
            CALL cp_dbcsr_init(rho_ao_out(i)%matrix)
            CALL cp_dbcsr_copy(rho_ao_out(i)%matrix, rho_ao_in(i)%matrix, &
                               name="myDensityMatrix_for_Spin_"//TRIM(ADJUSTL(cp_to_string(i))))
            CALL cp_dbcsr_set(rho_ao_out(i)%matrix, 0.0_dp)
         END DO
      END IF

      ! rho_r
      IF (ASSOCIATED(rho_r_in)) THEN
         ALLOCATE (rho_r_out(nspins))
         CALL qs_rho_set(rho_output, rho_r=rho_r_out)
         DO i = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r_out(i)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            rho_r_out(i)%pw%cr3d(:, :, :) = rho_r_in(i)%pw%cr3d(:, :, :)
         END DO
      END IF

      ! rho_g
      IF (ASSOCIATED(rho_g_in)) THEN
         ALLOCATE (rho_g_out(nspins))
         CALL qs_rho_set(rho_output, rho_g=rho_g_out)
         DO i = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_g_out(i)%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            rho_g_out(i)%pw%cc(:) = rho_g_in(i)%pw%cc(:)
         END DO
      END IF

      ! SCCS
      IF (ASSOCIATED(rho_r_sccs_in)) THEN
         CALL qs_rho_set(rho_output, rho_r_sccs=rho_r_sccs_out)
         CALL pw_pool_create_pw(auxbas_pw_pool, rho_r_sccs_out%pw, &
                                in_space=REALSPACE, &
                                use_data=REALDATA3D)
         rho_r_sccs_out%pw%cr3d(:, :, :) = rho_r_sccs_in%pw%cr3d(:, :, :)
      END IF

      ! drho_r and drho_g are only needed if calculated by collocation
      IF (dft_control%drho_by_collocation) THEN
         ! drho_r
         IF (ASSOCIATED(drho_r_in)) THEN
            ALLOCATE (drho_r_out(3*nspins))
            CALL qs_rho_set(rho_output, drho_r=drho_r_out)
            DO i = 1, 3*nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, drho_r_out(i)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               drho_r_out(i)%pw%cr3d(:, :, :) = drho_r_in(i)%pw%cr3d(:, :, :)
            END DO
         END IF

         ! drho_g
         IF (ASSOCIATED(drho_g_in)) THEN
            ALLOCATE (drho_g_out(3*nspins))
            CALL qs_rho_set(rho_output, drho_g=drho_g_out)
            DO i = 1, 3*nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, drho_g_out(i)%pw, &
                                      use_data=COMPLEXDATA1D, &
                                      in_space=RECIPROCALSPACE)
               drho_g_out(i)%pw%cc(:) = drho_g_in(i)%pw%cc(:)
            END DO
         END IF
      END IF

      ! tau_r and tau_g are only needed in the case of Meta-GGA XC-functionals
      ! are used. Therefore they are only allocated if
      ! dft_control%use_kinetic_energy_density is true
      IF (dft_control%use_kinetic_energy_density) THEN
         ! tau_r
         IF (ASSOCIATED(tau_r_in)) THEN
            ALLOCATE (tau_r_out(nspins))
            CALL qs_rho_set(rho_output, tau_r=tau_r_out)
            DO i = 1, nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, tau_r_out(i)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               tau_r_out(i)%pw%cr3d(:, :, :) = tau_r_in(i)%pw%cr3d(:, :, :)
            END DO
         END IF

         ! tau_g
         IF (ASSOCIATED(tau_g_in)) THEN
            ALLOCATE (tau_g_out(nspins))
            CALL qs_rho_set(rho_output, tau_g=tau_g_out)
            DO i = 1, nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, tau_g_out(i)%pw, &
                                      use_data=COMPLEXDATA1D, &
                                      in_space=RECIPROCALSPACE)
               tau_g_out(i)%pw%cc(:) = tau_g_in(i)%pw%cc(:)
            END DO
         END IF
      END IF

      CALL qs_rho_set(rho_output, &
                      rho_g_valid=rho_g_valid_in, &
                      rho_r_valid=rho_r_valid_in, &
                      drho_g_valid=drho_g_valid_in, &
                      drho_r_valid=drho_r_valid_in, &
                      tau_r_valid=tau_r_valid_in, &
                      tau_g_valid=tau_g_valid_in, &
                      soft_valid=soft_valid_in, &
                      rebuild_each=rebuild_each_in)

      ! tot_rho_r
      IF (ASSOCIATED(tot_rho_r_in)) THEN
         ALLOCATE (tot_rho_r_out(nspins))
         CALL qs_rho_set(rho_output, tot_rho_r=tot_rho_r_out)
         DO i = 1, nspins
            tot_rho_r_out(i) = tot_rho_r_in(i)
         END DO
      END IF

      ! tot_rho_g
      IF (ASSOCIATED(tot_rho_g_in)) THEN
         ALLOCATE (tot_rho_g_out(nspins))
         CALL qs_rho_set(rho_output, tot_rho_g=tot_rho_g_out)
         DO i = 1, nspins
            tot_rho_g_out(i) = tot_rho_g_in(i)
         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE duplicate_rho_type

END MODULE qs_rho_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief simple routine to print charges for all atomic charge methods
!>      (currently mulliken, lowdin and ddapc)
!> \par History
!>      Joost VandeVondele [2006.03]
! **************************************************************************************************
MODULE atomic_charges
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atomic_charges'

   PUBLIC :: print_atomic_charges

CONTAINS

! **************************************************************************************************
!> \brief generates a unified output format for atomic charges
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param scr ...
!> \param title ...
!> \param electronic_charges (natom,nspin), the number of electrons of (so positive) per spin
!>                            if (nspin==1) it is the sum of alpha and beta electrons
!> \param atomic_charges truely the atomic charge (taking Z into account, atoms negative, no spin)
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \note
!>      charges are computed per spin in the LSD case
! **************************************************************************************************
   SUBROUTINE print_atomic_charges(particle_set, qs_kind_set, scr, title, electronic_charges, &
                                   atomic_charges)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER                                            :: scr
      CHARACTER(LEN=*)                                   :: title
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: electronic_charges
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: atomic_charges

      CHARACTER(len=*), PARAMETER :: routineN = 'print_atomic_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: handle, iatom, ikind, natom, nspin
      REAL(KIND=dp)                                      :: total_charge, zeff

      CALL timeset(routineN, handle)

      IF (PRESENT(electronic_charges)) THEN
         nspin = SIZE(electronic_charges, 2)
         natom = SIZE(electronic_charges, 1)
      ELSE
         natom = SIZE(atomic_charges, 1)
         nspin = 0
      ENDIF

      IF (scr > 0) THEN
         WRITE (scr, '(T2,A)') title
         SELECT CASE (nspin)
         CASE (0, 1)
            IF (title == "RESP charges:") THEN
               WRITE (scr, '(A)') "  Type |   Atom   |    Charge"
            ELSE
               WRITE (scr, '(A)') "  Atom     |    Charge"
            ENDIF
         CASE DEFAULT
            WRITE (scr, '(A)') "  Atom     |    Charge | Spin diff charge"
         END SELECT
         total_charge = 0.0_dp
         IF (SIZE(particle_set) /= natom) THEN
            CALL cp__b("atomic_charges.F",85,"Unexpected number of atoms/charges")
         END IF
         WRITE (scr, '(A)') ""
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)

            SELECT CASE (nspin)
            CASE (0)
               IF (title == "RESP charges:") THEN
                  WRITE (scr, '(T3,A4,2X,I6,A2,A2,F12.6)') "RESP", iatom, "  ", element_symbol, atomic_charges(iatom)
                  total_charge = total_charge+atomic_charges(iatom)
               ELSE
                  WRITE (scr, '(I6,A2,A2,F12.6)') iatom, "  ", element_symbol, atomic_charges(iatom)
                  total_charge = total_charge+atomic_charges(iatom)
               ENDIF
            CASE (1)
               WRITE (scr, '(I6,A2,A2,F12.6)') iatom, "  ", element_symbol, zeff-electronic_charges(iatom, 1)
               total_charge = total_charge+zeff-electronic_charges(iatom, 1)
            CASE DEFAULT
               WRITE (scr, '(I6,A2,A2,2F12.6)') iatom, "  ", element_symbol, &
                  zeff-(electronic_charges(iatom, 1)+electronic_charges(iatom, 2)), &
                  (electronic_charges(iatom, 1)-electronic_charges(iatom, 2))
               total_charge = total_charge+zeff-(electronic_charges(iatom, 1)+electronic_charges(iatom, 2))
            END SELECT
         ENDDO
         IF (title == "RESP charges:") THEN
            WRITE (scr, '(A,F10.6)') "  Total             ", total_charge
         ELSE
            WRITE (scr, '(A,F10.6)') "  Total     ", total_charge
         ENDIF
         WRITE (scr, '(A)') ""
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE print_atomic_charges

! **************************************************************************************************
!> \brief ...
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param scr ...
!> \param charge ...
!> \param dipole ...
!> \param quadrupole ...
! **************************************************************************************************
   SUBROUTINE print_multipoles(particle_set, qs_kind_set, scr, charge, dipole, quadrupole)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER                                            :: scr
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: charge
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: dipole
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL        :: quadrupole

      CHARACTER(len=*), PARAMETER :: routineN = 'print_multipoles', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: handle, i, iatom, ikind, natom
      REAL(KIND=dp)                                      :: zeff

      CALL timeset(routineN, handle)

      natom = 0
      IF (PRESENT(charge)) THEN
         natom = SIZE(charge)
      ENDIF

      IF (scr > 0) THEN

         WRITE (scr, '(T2,A)') 'multipoles:'

         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)

            WRITE (scr, '(a,i5)') ' iatom= ', iatom
            WRITE (scr, '(a,a2)') ' element_symbol= ', element_symbol
            WRITE (scr, '(a,f20.10)') ' zeff= ', zeff

            WRITE (scr, '(a, f20.10)') 'charge =     ', charge(iatom)
            WRITE (scr, '(a,3f20.10)') 'dipole =     ', dipole(:, iatom)
            WRITE (scr, '(a)') 'quadrupole = '
            DO i = 1, 3
               WRITE (scr, '(3f20.10)') quadrupole(i, :, iatom)
            ENDDO

         ENDDO
         WRITE (scr, '(A)') ""
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE print_multipoles

END MODULE atomic_charges
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Density Derived atomic point charges from a QM calculation
!>      (see J. Chem. Phys. Vol. 103 pp. 7422-7428)
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_ddapc_forces

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: ddapc_restraint_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE erf_fn,                          ONLY: erfc
   USE input_constants,                 ONLY: do_ddapc_constraint,&
                                              do_ddapc_restraint,&
                                              weight_type_mass,&
                                              weight_type_unit
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              pi,&
                                              rootpi,&
                                              twopi
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_spline_utils,                 ONLY: Eval_d_Interp_Spl3_pbc
   USE pw_types,                        ONLY: pw_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE spherical_harmonics,             ONLY: dlegendre,&
                                              legendre
!NB for reducing results of calculations that use dq, which is now spread over nodes
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_ddapc_forces'
   PUBLIC :: ewald_ddapc_force, &
             reset_ch_pulay, &
             evaluate_restraint_functional, &
             restraint_functional_force, &
             solvation_ddapc_force

CONTAINS

! **************************************************************************************************
!> \brief Evaluates the Ewald term E2 and E3 energy term for the decoupling/coupling
!>      of periodic images
!> \param qs_env ...
!> \param coeff ...
!> \param apply_qmmm_periodic ...
!> \param factor ...
!> \param multipole_section ...
!> \param cell ...
!> \param particle_set ...
!> \param radii ...
!> \param dq ...
!> \param charges ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE ewald_ddapc_force(qs_env, coeff, apply_qmmm_periodic, &
                                          factor, multipole_section, cell, particle_set, radii, dq, charges)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_type), POINTER                             :: coeff
      LOGICAL, INTENT(IN)                                :: apply_qmmm_periodic
      REAL(KIND=dp), INTENT(IN)                          :: factor
      TYPE(section_vals_type), POINTER                   :: multipole_section
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: dq
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_ddapc_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ip1, ip2, iparticle1, &
                                                            iparticle2, k1, k2, k3, n_rep, nmax1, &
                                                            nmax2, nmax3, r1, r2, r3, rmax1, &
                                                            rmax2, rmax3
      LOGICAL                                            :: analyt
      REAL(KIND=dp)                                      :: alpha, eps, fac, fac3, fs, galpha, gsq, &
                                                            gsqi, ij_fac, q1t, q2t, r, r2tmp, &
                                                            rcut, rcut2, t1, t2, tol, tol1
      REAL(KIND=dp), DIMENSION(3)                        :: drvec, fvec, gvec, ra, rvec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: d_el, M
      TYPE(cp_para_env_type), POINTER                    :: para_env

      NULLIFY (d_el, M, para_env)
      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env, para_env=para_env)
      IF(.NOT.(PRESENT(charges)))CALL cp__a("cp_ddapc_forces.F",106)
      IF(.NOT.(ASSOCIATED(radii)))CALL cp__a("cp_ddapc_forces.F",107)
      IF(.NOT.(cell%orthorhombic))CALL cp__a("cp_ddapc_forces.F",108)
      rcut = MIN(cell%hmat(1, 1), cell%hmat(2, 2), cell%hmat(3, 3))/2.0_dp
      CALL section_vals_val_get(multipole_section, "RCUT", n_rep_val=n_rep)
      IF (n_rep == 1) CALL section_vals_val_get(multipole_section, "RCUT", r_val=rcut)
      CALL section_vals_val_get(multipole_section, "EWALD_PRECISION", r_val=eps)
      CALL section_vals_val_get(multipole_section, "ANALYTICAL_GTERM", l_val=analyt)
      rcut2 = rcut**2
      !
      ! Setting-up parameters for Ewald summation
      !
      eps = MIN(ABS(eps), 0.5_dp)
      tol = SQRT(ABS(LOG(eps*rcut)))
      alpha = SQRT(ABS(LOG(eps*rcut*tol)))/rcut
      galpha = 1.0_dp/(4.0_dp*alpha*alpha)
      tol1 = SQRT(-LOG(eps*rcut*(2.0_dp*tol*alpha)**2))
      nmax1 = NINT(0.25_dp+cell%hmat(1, 1)*alpha*tol1/pi)
      nmax2 = NINT(0.25_dp+cell%hmat(2, 2)*alpha*tol1/pi)
      nmax3 = NINT(0.25_dp+cell%hmat(3, 3)*alpha*tol1/pi)

      rmax1 = CEILING(rcut/cell%hmat(1, 1))
      rmax2 = CEILING(rcut/cell%hmat(2, 2))
      rmax3 = CEILING(rcut/cell%hmat(3, 3))

      ALLOCATE (d_el(3, SIZE(particle_set)))
      d_el = 0.0_dp
      fac = 1.e0_dp/cell%deth
      fac3 = fac/8.0_dp
      fvec = twopi/(/cell%hmat(1, 1), cell%hmat(2, 2), cell%hmat(3, 3)/)
      !
      DO iparticle1 = 1, SIZE(particle_set)
         !NB parallelization
         IF (MOD(iparticle1, para_env%num_pe) /= para_env%mepos) CYCLE
         ip1 = (iparticle1-1)*SIZE(radii)
         q1t = SUM(charges(ip1+1:ip1+SIZE(radii)))
         DO iparticle2 = 1, iparticle1
            ij_fac = 1.0_dp
            IF (iparticle1 == iparticle2) ij_fac = 0.5_dp

            ip2 = (iparticle2-1)*SIZE(radii)
            q2t = SUM(charges(ip2+1:ip2+SIZE(radii)))
            !
            ! Real-Space Contribution
            !
            rvec = particle_set(iparticle1)%r-particle_set(iparticle2)%r
            IF (iparticle1 /= iparticle2) THEN
               ra = rvec
               r2tmp = DOT_PRODUCT(ra, ra)
               IF (r2tmp <= rcut2) THEN
                  r = SQRT(r2tmp)
                  t1 = erfc(alpha*r)/r
                  drvec = ra/r*q1t*q2t*factor
                  t2 = -2.0_dp*alpha*EXP(-alpha*alpha*r*r)/(r*rootpi)-t1/r
                  d_el(1:3, iparticle1) = d_el(1:3, iparticle1)-t2*drvec
                  d_el(1:3, iparticle2) = d_el(1:3, iparticle2)+t2*drvec
               END IF
            END IF
            DO r1 = -rmax1, rmax1
               DO r2 = -rmax2, rmax2
                  DO r3 = -rmax3, rmax3
                     IF ((r1 == 0) .AND. (r2 == 0) .AND. (r3 == 0)) CYCLE
                     ra(1) = rvec(1)+cell%hmat(1, 1)*r1
                     ra(2) = rvec(2)+cell%hmat(2, 2)*r2
                     ra(3) = rvec(3)+cell%hmat(3, 3)*r3
                     r2tmp = DOT_PRODUCT(ra, ra)
                     IF (r2tmp <= rcut2) THEN
                        r = SQRT(r2tmp)
                        t1 = erfc(alpha*r)/r
                        drvec = ra/r*q1t*q2t*factor*ij_fac
                        t2 = -2.0_dp*alpha*EXP(-alpha*alpha*r*r)/(r*rootpi)-t1/r
                        d_el(1, iparticle1) = d_el(1, iparticle1)-t2*drvec(1)
                        d_el(2, iparticle1) = d_el(2, iparticle1)-t2*drvec(2)
                        d_el(3, iparticle1) = d_el(3, iparticle1)-t2*drvec(3)
                        d_el(1, iparticle2) = d_el(1, iparticle2)+t2*drvec(1)
                        d_el(2, iparticle2) = d_el(2, iparticle2)+t2*drvec(2)
                        d_el(3, iparticle2) = d_el(3, iparticle2)+t2*drvec(3)
                     END IF
                  END DO
               END DO
            END DO
            !
            ! G-space Contribution
            !
            IF (analyt) THEN
               DO k1 = 0, nmax1
                  DO k2 = -nmax2, nmax2
                     DO k3 = -nmax3, nmax3
                        IF (k1 == 0 .AND. k2 == 0 .AND. k3 == 0) CYCLE
                        fs = 2.0_dp; IF (k1 == 0) fs = 1.0_dp
                        gvec = fvec*(/REAL(k1, KIND=dp), REAL(k2, KIND=dp), REAL(k3, KIND=dp)/)
                        gsq = DOT_PRODUCT(gvec, gvec)
                        gsqi = fs/gsq
                        t1 = fac*gsqi*EXP(-galpha*gsq)
                        t2 = -SIN(DOT_PRODUCT(gvec, rvec))*t1*q1t*q2t*factor*fourpi
                        d_el(1:3, iparticle1) = d_el(1:3, iparticle1)-t2*gvec
                        d_el(1:3, iparticle2) = d_el(1:3, iparticle2)+t2*gvec
                     END DO
                  END DO
               END DO
            ELSE
               gvec = Eval_d_Interp_Spl3_pbc(rvec, coeff)*q1t*q2t*factor*fourpi
               d_el(1:3, iparticle1) = d_el(1:3, iparticle1)-gvec
               d_el(1:3, iparticle2) = d_el(1:3, iparticle2)+gvec
            END IF
            IF (iparticle1 /= iparticle2) THEN
               ra = rvec
               r = SQRT(DOT_PRODUCT(ra, ra))
               t2 = -1.0_dp/(r*r)*factor
               drvec = ra/r*q1t*q2t
               d_el(1:3, iparticle1) = d_el(1:3, iparticle1)+t2*drvec
               d_el(1:3, iparticle2) = d_el(1:3, iparticle2)-t2*drvec
            END IF
         END DO ! iparticle2
      END DO ! iparticle1
      !NB parallelization
      CALL mp_sum(d_el, para_env%group)
      M => qs_env%cp_ddapc_env%Md
      IF (apply_qmmm_periodic) M => qs_env%cp_ddapc_env%Mr
      CALL cp_decpl_ddapc_forces(qs_env, M, charges, dq, d_el, particle_set)
      DEALLOCATE (d_el)
      CALL timestop(handle)
   END SUBROUTINE ewald_ddapc_force

! **************************************************************************************************
!> \brief Evaluation of the pulay forces due to the fitted charge density
!> \param qs_env ...
!> \param M ...
!> \param charges ...
!> \param dq ...
!> \param d_el ...
!> \param particle_set ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE cp_decpl_ddapc_forces(qs_env, M, charges, dq, d_el, particle_set)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: M
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dq
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: d_el
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_decpl_ddapc_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ikind, j, k, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: uv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: chf
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force

      NULLIFY (para_env)
      CALL timeset(routineN, handle)
      natom = SIZE(particle_set)
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      para_env=para_env, &
                      force=force)
      ALLOCATE (atom_of_kind(natom))
      ALLOCATE (kind_of(natom))
      ALLOCATE (chf(3, natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               atom_of_kind=atom_of_kind, &
                               kind_of=kind_of)

      ALLOCATE (uv(SIZE(M, 1)))
      uv(:) = MATMUL(M, charges)
      DO k = 1, natom
         DO j = 1, 3
            chf(j, k) = DOT_PRODUCT(uv, dq(:, k, j))
         END DO
      END DO
      !NB now that get_ddapc returns dq that's spread over nodes, must reduce chf here
      CALL mp_sum(chf, para_env%group)
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         i = atom_of_kind(iatom)
         force(ikind)%ch_pulay(1:3, i) = force(ikind)%ch_pulay(1:3, i)+chf(1:3, iatom)+d_el(1:3, iatom)
      END DO
      DEALLOCATE (atom_of_kind)
      DEALLOCATE (kind_of)
      DEALLOCATE (chf)
      DEALLOCATE (uv)
      CALL timestop(handle)
   END SUBROUTINE cp_decpl_ddapc_forces

! **************************************************************************************************
!> \brief Evaluation of the pulay forces due to the fitted charge density
!> \param qs_env ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE reset_ch_pulay(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'reset_ch_pulay', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ind
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force

      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env=qs_env, &
                      force=force)
      DO ind = 1, SIZE(force)
         force(ind)%ch_pulay = 0.0_dp
      END DO
      CALL timestop(handle)
   END SUBROUTINE reset_ch_pulay

! **************************************************************************************************
!> \brief computes energy and derivatives given a set of charges
!> \param ddapc_restraint_control ...
!> \param n_gauss ...
!> \param uv derivate of energy wrt the corresponding charge entry
!> \param charges current value of the charges (one number for each gaussian used)
!>
!> \param energy_res energy due to the restraint
!> \par History
!>      02.2006 [Joost VandeVondele]
!>               modified [Teo]
!> \note
!>       should be easy to adapt for other specialized cases
! **************************************************************************************************
   SUBROUTINE evaluate_restraint_functional(ddapc_restraint_control, n_gauss, uv, &
                                            charges, energy_res)
      TYPE(ddapc_restraint_type), INTENT(INOUT)          :: ddapc_restraint_control
      INTEGER, INTENT(in)                                :: n_gauss
      REAL(KIND=dp), DIMENSION(:)                        :: uv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_res

      CHARACTER(len=*), PARAMETER :: routineN = 'evaluate_restraint_functional', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, ind
      REAL(KIND=dp)                                      :: dE, order_p

! order parameter (i.e. the sum of the charges of the selected atoms)

      order_p = 0.0_dp
      DO I = 1, ddapc_restraint_control%natoms
         ind = (ddapc_restraint_control%atoms(I)-1)*n_gauss
         order_p = order_p+ddapc_restraint_control%coeff(I)*SUM(charges(ind+1:ind+n_gauss))
      ENDDO
      ddapc_restraint_control%ddapc_order_p = order_p

      SELECT CASE (ddapc_restraint_control%functional_form)
      CASE (do_ddapc_restraint)
         ! the restraint energy
         energy_res = ddapc_restraint_control%strength*(order_p-ddapc_restraint_control%target)**2.0_dp

         ! derivative of the energy
         dE = 2.0_dp*ddapc_restraint_control%strength*(order_p-ddapc_restraint_control%target)
         DO I = 1, ddapc_restraint_control%natoms
            ind = (ddapc_restraint_control%atoms(I)-1)*n_gauss
            uv(ind+1:ind+n_gauss) = dE*ddapc_restraint_control%coeff(I)
         ENDDO
      CASE (do_ddapc_constraint)
         energy_res = ddapc_restraint_control%strength*(order_p-ddapc_restraint_control%target)

         ! derivative of the energy
         DO I = 1, ddapc_restraint_control%natoms
            ind = (ddapc_restraint_control%atoms(I)-1)*n_gauss
            uv(ind+1:ind+n_gauss) = ddapc_restraint_control%strength*ddapc_restraint_control%coeff(I)
         ENDDO

      CASE DEFAULT
         CALL cp__b("cp_ddapc_forces.F",378,"")
      END SELECT

   END SUBROUTINE evaluate_restraint_functional

! **************************************************************************************************
!> \brief computes derivatives for DDAPC restraint
!> \param qs_env ...
!> \param ddapc_restraint_control ...
!> \param dq ...
!> \param charges ...
!> \param n_gauss ...
!> \param particle_set ...
!> \par History
!>      02.2006 [Joost VandeVondele]
!>              modified [Teo]
!> \note
!>       should be easy to adapt for other specialized cases
! **************************************************************************************************
   SUBROUTINE restraint_functional_force(qs_env, ddapc_restraint_control, dq, charges, &
                                         n_gauss, particle_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ddapc_restraint_type), INTENT(INOUT)          :: ddapc_restraint_control
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dq
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      INTEGER, INTENT(in)                                :: n_gauss
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'restraint_functional_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ikind, j, k, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      REAL(kind=dp)                                      :: dum
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: uv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: chf
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force

      NULLIFY (para_env)
      CALL timeset(routineN, handle)
      natom = SIZE(particle_set)
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      para_env=para_env, &
                      force=force)
      ALLOCATE (atom_of_kind(natom))
      ALLOCATE (kind_of(natom))
      ALLOCATE (chf(3, natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               atom_of_kind=atom_of_kind, &
                               kind_of=kind_of)

      ALLOCATE (uv(SIZE(dq, 1)))
      uv = 0.0_dp
      CALL evaluate_restraint_functional(ddapc_restraint_control, n_gauss, uv, &
                                         charges, dum)
      DO k = 1, natom
         DO j = 1, 3
            chf(j, k) = DOT_PRODUCT(uv, dq(:, k, j))
         END DO
      END DO
      !NB now that get_ddapc returns dq that's spread over nodes, must reduce chf here
      CALL mp_sum(chf, para_env%group)
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         i = atom_of_kind(iatom)
         force(ikind)%ch_pulay(1:3, i) = force(ikind)%ch_pulay(1:3, i)+chf(1:3, iatom)
      END DO
      DEALLOCATE (atom_of_kind)
      DEALLOCATE (kind_of)
      DEALLOCATE (chf)
      DEALLOCATE (uv)
      CALL timestop(handle)

   END SUBROUTINE restraint_functional_force

! **************************************************************************************************
!> \brief Evaluates the electrostatic potential due to a simple solvation model
!>      Spherical cavity in a dieletric medium
!> \param qs_env ...
!> \param solvation_section ...
!> \param particle_set ...
!> \param radii ...
!> \param dq ...
!> \param charges ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE solvation_ddapc_force(qs_env, solvation_section, particle_set, &
                                    radii, dq, charges)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: solvation_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: dq
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'solvation_ddapc_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ip1, ip2, iparticle1, iparticle2, l, &
                                                            lmax, n_rep1, n_rep2, weight
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: fixed_center
      REAL(KIND=dp) :: center(3), dcos1(3), dcos2(3), dpos1(3), dpos2(3), eps_in, eps_out, &
         factor1(3), factor2(3), lr, mass, mycos, pos1, pos1i, pos2, pos2i, ptcos, q1t, q2t, &
         r1(3), r1s, r2(3), r2s, Rs, rvec(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pos, R0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: d_el, LocP, M

      fixed_center = .FALSE.
      NULLIFY (d_el, M)
      eps_in = 1.0_dp
      CALL section_vals_val_get(solvation_section, "EPS_OUT", r_val=eps_out)
      CALL section_vals_val_get(solvation_section, "LMAX", i_val=lmax)
      CALL section_vals_val_get(solvation_section, "SPHERE%RADIUS", r_val=Rs)
      CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%XYZ", n_rep_val=n_rep1)
      IF (n_rep1 /= 0) THEN
         CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%XYZ", r_vals=R0)
         center = R0
      ELSE
         CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%ATOM_LIST", &
                                   n_rep_val=n_rep2)
         IF (n_rep2 /= 0) THEN
            CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%ATOM_LIST", i_vals=list)
            CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%WEIGHT_TYPE", i_val=weight)
            ALLOCATE (R0(SIZE(list)))
            SELECT CASE (weight)
            CASE (weight_type_unit)
               R0 = 0.0_dp
               DO i = 1, SIZE(list)
                  R0 = R0+particle_set(list(i))%r
               END DO
               R0 = R0/REAL(SIZE(list), KIND=dp)
            CASE (weight_type_mass)
               R0 = 0.0_dp
               mass = 0.0_dp
               DO i = 1, SIZE(list)
                  R0 = R0+particle_set(list(i))%r*particle_set(list(i))%atomic_kind%mass
                  mass = mass+particle_set(list(i))%atomic_kind%mass
               END DO
               R0 = R0/mass
            END SELECT
            center = R0
            DEALLOCATE (R0)
         END IF
      END IF
      IF(.NOT.(n_rep1 /= 0 .OR. n_rep2 /= 0))CALL cp__a("cp_ddapc_forces.F",530)
      ! Potential calculation
      ALLOCATE (LocP(0:lmax, SIZE(particle_set)))
      ALLOCATE (pos(SIZE(particle_set)))
      ALLOCATE (d_el(3, SIZE(particle_set)))
      d_el = 0.0_dp
      ! Determining the single atomic contribution to the dielectric dipole
      DO i = 1, SIZE(particle_set)
         rvec = particle_set(i)%r-center
         r2s = DOT_PRODUCT(rvec, rvec)
         r1s = SQRT(r2s)
         LocP(:, i) = 0.0_dp
         IF (r1s /= 0.0_dp) THEN
            DO l = 0, lmax
               LocP(l, i) = (r1s**l*REAL(l+1, KIND=dp)*(eps_in-eps_out))/ &
                            (Rs**(2*l+1)*eps_in*(REAL(l, KIND=dp)*eps_in+REAL(l+1, KIND=dp)*eps_out))
            END DO
         END IF
         pos(i) = r1s
      END DO
      ! Computes the full derivatives of the interaction energy
      DO iparticle1 = 1, SIZE(particle_set)
         ip1 = (iparticle1-1)*SIZE(radii)
         q1t = SUM(charges(ip1+1:ip1+SIZE(radii)))
         DO iparticle2 = 1, iparticle1
            ip2 = (iparticle2-1)*SIZE(radii)
            q2t = SUM(charges(ip2+1:ip2+SIZE(radii)))
            !
            r1 = particle_set(iparticle1)%r-center
            r2 = particle_set(iparticle2)%r-center
            pos1 = pos(iparticle1)
            pos2 = pos(iparticle2)
            factor1 = 0.0_dp
            factor2 = 0.0_dp
            IF (pos1*pos2 /= 0.0_dp) THEN
               pos1i = 1.0_dp/pos1
               pos2i = 1.0_dp/pos1
               dpos1 = pos1i*r1
               dpos2 = pos2i*r2
               ptcos = DOT_PRODUCT(r1, r2)
               mycos = ptcos/(pos1*pos2)
               IF (ABS(mycos) > 1.0_dp) mycos = SIGN(1.0_dp, mycos)
               dcos1 = (r2*(pos1*pos2)-pos2*dpos1*ptcos)/(pos1*pos2)**2
               dcos2 = (r1*(pos1*pos2)-pos1*dpos2*ptcos)/(pos1*pos2)**2

               DO l = 1, lmax
                  lr = REAL(l, KIND=dp)
                  factor1 = factor1+lr*LocP(l, iparticle2)*pos1**(l-1)*legendre(mycos, l, 0)*dpos1 &
                            +LocP(l, iparticle2)*pos1**l*dlegendre(mycos, l, 0)*dcos1
                  factor2 = factor2+lr*LocP(l, iparticle1)*pos2**(l-1)*legendre(mycos, l, 0)*dpos2 &
                            +LocP(l, iparticle1)*pos2**l*dlegendre(mycos, l, 0)*dcos2
               END DO
            END IF
            factor1 = factor1*q1t*q2t
            factor2 = factor2*q1t*q2t
            d_el(1:3, iparticle1) = d_el(1:3, iparticle1)+0.5_dp*factor1
            d_el(1:3, iparticle2) = d_el(1:3, iparticle2)+0.5_dp*factor2
         END DO
      END DO
      DEALLOCATE (pos)
      DEALLOCATE (LocP)
      M => qs_env%cp_ddapc_env%Ms
      CALL cp_decpl_ddapc_forces(qs_env, M, charges, dq, d_el, particle_set)
      DEALLOCATE (d_el)
   END SUBROUTINE solvation_ddapc_force

END MODULE cp_ddapc_forces
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Density Derived atomic point charges from a QM calculation
!>      (see Bloechl, J. Chem. Phys. Vol. 103 pp. 7422-7428)
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_ddapc_util

   USE atomic_charges,                  ONLY: print_atomic_charges
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: ddapc_restraint_type,&
                                              dft_control_type
   USE cp_ddapc_forces,                 ONLY: evaluate_restraint_functional
   USE cp_ddapc_methods,                ONLY: build_A_matrix,&
                                              build_b_vector,&
                                              build_der_A_matrix_rows,&
                                              build_der_b_vector,&
                                              cleanup_g_dot_rvec_sin_cos,&
                                              prep_g_dot_rvec_sin_cos
   USE cp_ddapc_types,                  ONLY: cp_ddapc_create,&
                                              cp_ddapc_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_full_density,&
                                              do_spin_density
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_transfer
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_ddapc_util'
   PUBLIC :: get_ddapc, &
             restraint_functional_potential, &
             modify_hartree_pot, &
             cp_ddapc_init

CONTAINS

! **************************************************************************************************
!> \brief Initialize the cp_ddapc_environment
!> \param qs_env ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE cp_ddapc_init(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw, iw2, n_rep_val, num_gauss
      LOGICAL                                            :: allocate_ddapc_env, unimplemented
      REAL(KIND=dp)                                      :: gcut, pfact, rcmin, Vol
      REAL(KIND=dp), DIMENSION(:), POINTER               :: inp_radii, radii
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), POINTER                           :: rho0_s_gs, rho_core
      TYPE(pw_pool_type), POINTER                        :: auxbas_pool
      TYPE(pw_type), POINTER                             :: rho_tot_g
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: density_fit_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      NULLIFY (dft_control, rho, rho_tot_g, rho_core, rho0_s_gs, pw_env, &
               radii, inp_radii, particle_set, qs_charges, para_env)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      allocate_ddapc_env = qs_env%cp_ddapc_ewald%do_solvation .OR. &
                           qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl .OR. &
                           qs_env%cp_ddapc_ewald%do_decoupling .OR. &
                           qs_env%cp_ddapc_ewald%do_restraint
      unimplemented = dft_control%qs_control%semi_empirical .OR. &
                      dft_control%qs_control%dftb
      IF (allocate_ddapc_env .AND. unimplemented) THEN
         CALL cp__b("cp_ddapc_util.F",117,"DDAP charges work only with GPW/GAPW code.")
      END IF
      allocate_ddapc_env = allocate_ddapc_env .OR. &
                           qs_env%cp_ddapc_ewald%do_property
      allocate_ddapc_env = allocate_ddapc_env .AND. (.NOT. unimplemented)
      IF (allocate_ddapc_env) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         dft_control=dft_control, &
                         rho=rho, &
                         rho_core=rho_core, &
                         rho0_s_gs=rho0_s_gs, &
                         pw_env=pw_env, &
                         qs_charges=qs_charges, &
                         particle_set=particle_set, &
                         cell=cell, &
                         super_cell=super_cell, &
                         para_env=para_env)
         density_fit_section => section_vals_get_subs_vals(qs_env%input, "DFT%DENSITY_FITTING")
         iw = cp_print_key_unit_nr(logger, density_fit_section, &
                                   "PROGRAM_RUN_INFO", ".FitCharge")
         IF (iw > 0) THEN
            WRITE (iw, '(/,A)') " Initializing the DDAPC Environment"
         END IF
         CALL pw_env_get(pw_env=pw_env, auxbas_pw_pool=auxbas_pool)
         CALL pw_pool_create_pw(auxbas_pool, rho_tot_g, in_space=RECIPROCALSPACE, &
                                use_data=COMPLEXDATA1D)
         Vol = rho_tot_g%pw_grid%vol
         !
         ! Get Input Parameters
         !
         CALL section_vals_val_get(density_fit_section, "RADII", n_rep_val=n_rep_val)
         IF (n_rep_val /= 0) THEN
            CALL section_vals_val_get(density_fit_section, "RADII", r_vals=inp_radii)
            num_gauss = SIZE(inp_radii)
            ALLOCATE (radii(num_gauss))
            radii = inp_radii
         ELSE
            CALL section_vals_val_get(density_fit_section, "NUM_GAUSS", i_val=num_gauss)
            CALL section_vals_val_get(density_fit_section, "MIN_RADIUS", r_val=rcmin)
            CALL section_vals_val_get(density_fit_section, "PFACTOR", r_val=pfact)
            ALLOCATE (radii(num_gauss))
            DO i = 1, num_gauss
               radii(i) = rcmin*pfact**(i-1)
            END DO
         END IF
         CALL section_vals_val_get(density_fit_section, "GCUT", r_val=gcut)
         ! Create DDAPC environment
         iw2 = cp_print_key_unit_nr(logger, density_fit_section, &
                                    "PROGRAM_RUN_INFO/CONDITION_NUMBER", ".FitCharge")
         ! Initialization of the cp_ddapc_env and of the cp_ddapc_ewald environment
         !NB pass qs_env%para_env for parallelization of ewald_ddapc_pot()
         CALL cp_ddapc_create(para_env, &
                              qs_env%cp_ddapc_env, &
                              qs_env%cp_ddapc_ewald, &
                              particle_set, &
                              radii, &
                              cell, &
                              super_cell, &
                              rho_tot_g, &
                              gcut, &
                              iw2, &
                              Vol, &
                              qs_env%input)
         CALL cp_print_key_finished_output(iw2, logger, density_fit_section, &
                                           "PROGRAM_RUN_INFO/CONDITION_NUMBER")
         DEALLOCATE (radii)
         CALL pw_pool_give_back_pw(auxbas_pool, rho_tot_g, &
                                   accept_non_compatible=.TRUE.)
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_ddapc_init

! **************************************************************************************************
!> \brief Computes the Density Derived Atomic Point Charges
!> \param qs_env ...
!> \param calc_force ...
!> \param density_fit_section ...
!> \param density_type ...
!> \param qout1 ...
!> \param qout2 ...
!> \param out_radii ...
!> \param dq_out ...
!> \param ext_rho_tot_g ...
!> \param Itype_of_density ...
!> \param iwc ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE get_ddapc(qs_env, calc_force, density_fit_section, &
                                  density_type, qout1, qout2, out_radii, dq_out, ext_rho_tot_g, &
                                  Itype_of_density, iwc)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in), OPTIONAL                      :: calc_force
      TYPE(section_vals_type), POINTER                   :: density_fit_section
      INTEGER, OPTIONAL                                  :: density_type
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: qout1, qout2, out_radii
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: dq_out
      TYPE(pw_type), OPTIONAL, POINTER                   :: ext_rho_tot_g
      CHARACTER(LEN=*), OPTIONAL                         :: Itype_of_density
      INTEGER, INTENT(IN), OPTIONAL                      :: iwc

      CHARACTER(len=*), PARAMETER :: routineN = 'get_ddapc', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: type_of_density
      INTEGER                                            :: handle, handle2, handle3, i, ii, &
                                                            iparticle, iparticle0, ispin, iw, j, &
                                                            myid, n_rep_val, ndim, nparticles, &
                                                            num_gauss, pmax, pmin
      LOGICAL                                            :: need_f
      REAL(KIND=dp)                                      :: c1, c3, ch_dens, gcut, pfact, rcmin, Vol
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: AmI_bv, AmI_cv, bv, cv, cvT_AmI, &
                                                            cvT_AmI_dAmj, dAmj_qv, qtot, qv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dbv, g_dot_rvec_cos, g_dot_rvec_sin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: dAm, dqv, tv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: inp_radii, radii
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_p_type), POINTER                           :: rho0_s_gs, rho_core
      TYPE(pw_pool_type), POINTER                        :: auxbas_pool
      TYPE(pw_type), POINTER                             :: rho_tot_g
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho

!NB variables for doing build_der_A_matrix_rows in blocks
!NB refactor math in inner loop - no need for dqv0
!!NB refactor math in inner loop - new temporaries

      EXTERNAL dgemv, dgemm

      CALL timeset(routineN, handle)
      need_f = .FALSE.
      IF (PRESENT(calc_force)) need_f = calc_force
      logger => cp_get_default_logger()
      NULLIFY (dft_control, rho, rho_tot_g, rho_core, rho0_s_gs, pw_env, rho_g, rho_r, &
               radii, inp_radii, particle_set, qs_kind_set, qs_charges, cp_ddapc_env)
      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      rho=rho, &
                      rho_core=rho_core, &
                      rho0_s_gs=rho0_s_gs, &
                      pw_env=pw_env, &
                      qs_charges=qs_charges, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      super_cell=super_cell)

      CALL qs_rho_get(rho, rho_r=rho_r, rho_g=rho_g)

      IF (PRESENT(iwc)) THEN
         iw = iwc
      ELSE
         iw = cp_print_key_unit_nr(logger, density_fit_section, &
                                   "PROGRAM_RUN_INFO", ".FitCharge")
      END IF
      CALL pw_env_get(pw_env=pw_env, &
                      auxbas_pw_pool=auxbas_pool)
      CALL pw_pool_create_pw(auxbas_pool, rho_tot_g, in_space=RECIPROCALSPACE, &
                             use_data=COMPLEXDATA1D)
      IF (PRESENT(ext_rho_tot_g)) THEN
         ! If provided use the input density in g-space
         CALL pw_transfer(ext_rho_tot_g, rho_tot_g)
         type_of_density = Itype_of_density
      ELSE
         IF (PRESENT(density_type)) THEN
            myid = density_type
         ELSE
            CALL section_vals_val_get(qs_env%input, &
                                      "PROPERTIES%FIT_CHARGE%TYPE_OF_DENSITY", i_val=myid)
         END IF
         SELECT CASE (myid)
         CASE (do_full_density)
            ! Otherwise build the total QS density (electron+nuclei) in G-space
            IF (dft_control%qs_control%gapw) THEN
               CALL pw_transfer(rho0_s_gs%pw, rho_tot_g)
            ELSE
               CALL pw_transfer(rho_core%pw, rho_tot_g)
            END IF
            DO ispin = 1, SIZE(rho_g)
               CALL pw_axpy(rho_g(ispin)%pw, rho_tot_g)
            END DO
            type_of_density = "FULL DENSITY"
         CASE (do_spin_density)
            CALL pw_copy(rho_g(1)%pw, rho_tot_g)
            CALL pw_axpy(rho_g(2)%pw, rho_tot_g, alpha=-1._dp)
            type_of_density = "SPIN DENSITY"
         END SELECT
      END IF
      Vol = rho_r(1)%pw%pw_grid%vol
      ch_dens = 0.0_dp
      ! should use pw_integrate
      IF (rho_tot_g%pw_grid%have_g0) ch_dens = REAL(rho_tot_g%cc(1), KIND=dp)
      CALL mp_sum(ch_dens, logger%para_env%group)
      !
      ! Get Input Parameters
      !
      CALL section_vals_val_get(density_fit_section, "RADII", n_rep_val=n_rep_val)
      IF (n_rep_val /= 0) THEN
         CALL section_vals_val_get(density_fit_section, "RADII", r_vals=inp_radii)
         num_gauss = SIZE(inp_radii)
         ALLOCATE (radii(num_gauss))
         radii = inp_radii
      ELSE
         CALL section_vals_val_get(density_fit_section, "NUM_GAUSS", i_val=num_gauss)
         CALL section_vals_val_get(density_fit_section, "MIN_RADIUS", r_val=rcmin)
         CALL section_vals_val_get(density_fit_section, "PFACTOR", r_val=pfact)
         ALLOCATE (radii(num_gauss))
         DO i = 1, num_gauss
            radii(i) = rcmin*pfact**(i-1)
         END DO
      END IF
      IF (PRESENT(out_radii)) THEN
         IF (ASSOCIATED(out_radii)) THEN
            DEALLOCATE (out_radii)
         END IF
         ALLOCATE (out_radii(SIZE(radii)))
         out_radii = radii
      END IF
      CALL section_vals_val_get(density_fit_section, "GCUT", r_val=gcut)
      cp_ddapc_env => qs_env%cp_ddapc_env
      !
      ! Start with the linear system
      !
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (bv(ndim))
      ALLOCATE (qv(ndim))
      ALLOCATE (qtot(SIZE(particle_set)))
      ALLOCATE (cv(ndim))
      CALL timeset(routineN//"-charges", handle2)
      bv(:) = 0.0_dp
      cv(:) = 1.0_dp/Vol
      CALL build_b_vector(bv, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                          particle_set, radii, rho_tot_g, gcut); bv(:) = bv(:)/Vol
      CALL mp_sum(bv, rho_tot_g%pw_grid%para%group)
      c1 = DOT_PRODUCT(cv, MATMUL(cp_ddapc_env%AmI, bv))-ch_dens
      c1 = c1/cp_ddapc_env%c0
      qv(:) = -MATMUL(cp_ddapc_env%AmI, (bv-c1*cv))
      j = 0
      qtot = 0.0_dp
      DO i = 1, ndim, num_gauss
         j = j+1
         DO ii = 1, num_gauss
            qtot(j) = qtot(j)+qv((i-1)+ii)
         END DO
      END DO
      IF (PRESENT(qout1)) THEN
         IF (ASSOCIATED(qout1)) THEN
            IF(.NOT.(SIZE(qout1) == SIZE(qv)))CALL cp__a("cp_ddapc_util.F",372)
         ELSE
            ALLOCATE (qout1(SIZE(qv)))
         END IF
         qout1 = qv
      END IF
      IF (PRESENT(qout2)) THEN
         IF (ASSOCIATED(qout2)) THEN
            IF(.NOT.(SIZE(qout2) == SIZE(qtot)))CALL cp__a("cp_ddapc_util.F",380)
         ELSE
            ALLOCATE (qout2(SIZE(qtot)))
         END IF
         qout2 = qtot
      END IF
      CALL print_atomic_charges(particle_set, qs_kind_set, iw, title=" DDAP "// &
                                TRIM(type_of_density)//" charges:", atomic_charges=qtot)
      CALL timestop(handle2)
      !
      ! If requested evaluate also the correction to derivatives due to Pulay Forces
      !
      IF (need_f) THEN
         CALL timeset(routineN//"-forces", handle3)
         IF (iw > 0) THEN
            WRITE (iw, '(T3,A)') " Evaluating DDAPC atomic derivatives .."
         END IF
         ALLOCATE (dAm(ndim, ndim, 3))
         ALLOCATE (dbv(ndim, 3))
         ALLOCATE (dqv(ndim, SIZE(particle_set), 3))
         !NB refactor math in inner loop - no more dqv0, but new temporaries instead
         ALLOCATE (cvT_AmI(ndim))
         ALLOCATE (cvT_AmI_dAmj(ndim))
         ALLOCATE (tv(ndim, SIZE(particle_set), 3))
         ALLOCATE (AmI_cv(ndim))
         cvT_AmI(:) = MATMUL(cv, cp_ddapc_env%AmI)
         AmI_cv(:) = MATMUL(cp_ddapc_env%AmI, cv)
         ALLOCATE (dAmj_qv(ndim))
         ALLOCATE (AmI_bv(ndim))
         AmI_bv(:) = MATMUL(cp_ddapc_env%AmI, bv)

         !NB call routine to precompute sin(g.r) and cos(g.r),
         ! so it doesn't have to be done for each r_i-r_j pair in build_der_A_matrix_rows()
         CALL prep_g_dot_rvec_sin_cos(rho_tot_g, particle_set, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
         !NB do build_der_A_matrix_rows in blocks, for more efficient use of DGEMM
         DO iparticle0 = 1, SIZE(particle_set), 100
            nparticles = MIN(100, SIZE(particle_set)-iparticle0+1)
            !NB each dAm is supposed to have one block of rows and one block of columns
            !NB for derivatives with respect to each atom.  build_der_A_matrix_rows()
            !NB just returns rows, since dAm is symmetric, and missing columns can be
            !NB reconstructed with a simple transpose, as below
            CALL build_der_A_matrix_rows(dAm, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                                         particle_set, radii, rho_tot_g, gcut, iparticle0, &
                                         nparticles, g_dot_rvec_sin, g_dot_rvec_cos)
            !NB no more reduction of dbv and dAm - instead we go through with each node's contribution
            !NB and reduce resulting charges/forces once, at the end.  Intermediate speedup can be
            !NB had by reducing dqv after the inner loop, and then other routines don't need to know
            !NB that contributions to dqv are distributed over the nodes.
            !NB also get rid of zeroing of dAm and division by Vol**2 - it's slow, and can be done
            !NB more quickly later, to a scalar or vector rather than a matrix
            DO iparticle = iparticle0, iparticle0+nparticles-1
            IF (debug_this_module) THEN
               CALL debug_der_A_matrix(dAm, particle_set, radii, rho_tot_g, &
                                       gcut, iparticle, Vol, qs_env)
               cp_ddapc_env => qs_env%cp_ddapc_env
            END IF
            dbv(:, :) = 0.0_dp
            CALL build_der_b_vector(dbv, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                                    particle_set, radii, rho_tot_g, gcut, iparticle)
            dbv(:, :) = dbv(:, :)/Vol
            IF (debug_this_module) THEN
               CALL debug_der_b_vector(dbv, particle_set, radii, rho_tot_g, &
                                       gcut, iparticle, Vol, qs_env)
               cp_ddapc_env => qs_env%cp_ddapc_env
            END IF
            DO j = 1, 3
               !NB dAmj is actually pretty sparse - one block of cols + one block of rows - use this here:
               pmin = (iparticle-1)*SIZE(radii)+1
               pmax = iparticle*SIZE(radii)
               !NB multiply by block of columns that aren't explicitly in dAm, but can be reconstructured
               !NB as transpose of relevant block of rows
               IF (pmin > 1) THEN
                  dAmj_qv(:pmin-1) = MATMUL(TRANSPOSE(dAm(pmin:pmax, :pmin-1, j)), qv(pmin:pmax))
                  cvT_AmI_dAmj(:pmin-1) = MATMUL(TRANSPOSE(dAm(pmin:pmax, :pmin-1, j)), cvT_AmI(pmin:pmax))
               ENDIF
               !NB multiply by block of rows that are explicitly in dAm
               dAmj_qv(pmin:pmax) = MATMUL(dAm(pmin:pmax, :, j), qv(:))
               cvT_AmI_dAmj(pmin:pmax) = MATMUL(dAm(pmin:pmax, :, j), cvT_AmI(:))
               !NB multiply by block of columns that aren't explicitly in dAm, but can be reconstructured
               !NB as transpose of relevant block of rows
               IF (pmax < SIZE(particle_set)*SIZE(radii)) THEN
                  dAmj_qv(pmax+1:) = MATMUL(TRANSPOSE(dAm(pmin:pmax, pmax+1:, j)), qv(pmin:pmax))
                  cvT_AmI_dAmj(pmax+1:) = MATMUL(TRANSPOSE(dAm(pmin:pmax, pmax+1:, j)), cvT_AmI(pmin:pmax))
               ENDIF
               dAmj_qv(:) = dAmj_qv(:)/(Vol*Vol)
               cvT_AmI_dAmj(:) = cvT_AmI_dAmj(:)/(Vol*Vol)
               c3 = DOT_PRODUCT(cvT_AmI_dAmj, AmI_bv)-DOT_PRODUCT(cvT_AmI, dbv(:, j))-c1*DOT_PRODUCT(cvT_AmI_dAmj, AmI_cv)
               tv(:, iparticle, j) = -(dAmj_qv(:)+dbv(:, j)+c3/cp_ddapc_env%c0*cv)
            END DO ! j
            !NB zero relevant parts of dAm here
            dAm((iparticle-1)*SIZE(radii)+1:iparticle*SIZE(radii), :, :) = 0.0_dp
            !! dAm(:,(iparticle-1)*SIZE(radii)+1:iparticle*SIZE(radii),:) = 0.0_dp
            END DO ! iparticle
         END DO ! iparticle0
         !NB final part of refactoring of math - one dgemm is faster than many dgemv
         CALL dgemm('N', 'N', SIZE(dqv, 1), SIZE(dqv, 2)*SIZE(dqv, 3), SIZE(cp_ddapc_env%AmI, 2), 1.0_dp, &
                    cp_ddapc_env%AmI, SIZE(cp_ddapc_env%AmI, 1), tv, SIZE(tv, 1), 0.0_dp, dqv, SIZE(dqv, 1))
         !NB deallocate g_dot_rvec_sin and g_dot_rvec_cos
         CALL cleanup_g_dot_rvec_sin_cos(g_dot_rvec_sin, g_dot_rvec_cos)
         !NB moved reduction out to where dqv is used to compute
         !NB  a force contribution (smaller array to reduce, just size(particle_set) x 3)
         !NB namely ewald_ddapc_force(), cp_decl_ddapc_forces(), restraint_functional_force()
         IF(.NOT.(PRESENT(dq_out)))CALL cp__a("cp_ddapc_util.F",483)
         IF (.NOT. ASSOCIATED(dq_out)) THEN
            ALLOCATE (dq_out(SIZE(dqv, 1), SIZE(dqv, 2), SIZE(dqv, 3)))
         ELSE
            IF(.NOT.(SIZE(dqv, 1) == SIZE(dq_out, 1)))CALL cp__a("cp_ddapc_util.F",487)
            IF(.NOT.(SIZE(dqv, 2) == SIZE(dq_out, 2)))CALL cp__a("cp_ddapc_util.F",488)
            IF(.NOT.(SIZE(dqv, 3) == SIZE(dq_out, 3)))CALL cp__a("cp_ddapc_util.F",489)
         END IF
         dq_out = dqv
         IF (debug_this_module) THEN
            CALL debug_charge(dqv, qs_env, density_fit_section, &
                              particle_set, radii, rho_tot_g, type_of_density)
            cp_ddapc_env => qs_env%cp_ddapc_env
         END IF
         DEALLOCATE (dqv)
         DEALLOCATE (dAm)
         DEALLOCATE (dbv)
         !NB deallocate new temporaries
         DEALLOCATE (cvT_AmI)
         DEALLOCATE (cvT_AmI_dAmj)
         DEALLOCATE (AmI_cv)
         DEALLOCATE (tv)
         DEALLOCATE (dAmj_qv)
         DEALLOCATE (AmI_bv)
         CALL timestop(handle3)
      END IF
      !
      ! End of charge fit
      !
      DEALLOCATE (radii)
      DEALLOCATE (bv)
      DEALLOCATE (cv)
      DEALLOCATE (qv)
      DEALLOCATE (qtot)
      IF (.NOT. PRESENT(iwc)) THEN
         CALL cp_print_key_finished_output(iw, logger, density_fit_section, &
                                           "PROGRAM_RUN_INFO")
      END IF
      CALL pw_pool_give_back_pw(auxbas_pool, rho_tot_g, &
                                accept_non_compatible=.TRUE.)
      CALL timestop(handle)
   END SUBROUTINE get_ddapc

! **************************************************************************************************
!> \brief modify hartree potential to handle restraints in DDAPC scheme
!> \param v_hartree_gspace ...
!> \param density_fit_section ...
!> \param particle_set ...
!> \param AmI ...
!> \param radii ...
!> \param charges ...
!> \param ddapc_restraint_control ...
!> \param energy_res ...
!> \par History
!>      02.2006  modified [Teo]
! **************************************************************************************************
   SUBROUTINE restraint_functional_potential(v_hartree_gspace, &
                                             density_fit_section, particle_set, AmI, radii, charges, &
                                             ddapc_restraint_control, energy_res)
      TYPE(pw_p_type)                                    :: v_hartree_gspace
      TYPE(section_vals_type), POINTER                   :: density_fit_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: AmI
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii, charges
      TYPE(ddapc_restraint_type), INTENT(INOUT)          :: ddapc_restraint_control
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_res

      CHARACTER(len=*), PARAMETER :: routineN = 'restraint_functional_potential', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: g_corr, phase
      INTEGER                                            :: handle, idim, ig, igauss, iparticle, &
                                                            n_gauss
      REAL(KIND=dp)                                      :: arg, fac, fac2, g2, gcut, gcut2, gfunc, &
                                                            gvec(3), rc, rc2, rvec(3), sfac, Vol, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cv, uv
      TYPE(pw_type), POINTER                             :: g_hartree

      CALL timeset(routineN, handle)
      NULLIFY (g_hartree)
      n_gauss = SIZE(radii)
      ALLOCATE (cv(n_gauss*SIZE(particle_set)))
      ALLOCATE (uv(n_gauss*SIZE(particle_set)))
      uv = 0.0_dp
      CALL evaluate_restraint_functional(ddapc_restraint_control, n_gauss, uv, &
                                         charges, energy_res)
      !
      CALL section_vals_val_get(density_fit_section, "GCUT", r_val=gcut)
      gcut2 = gcut*gcut
      g_hartree => v_hartree_gspace%pw
      Vol = g_hartree%pw_grid%vol
      cv = 1.0_dp/Vol
      sfac = -1.0_dp/Vol
      fac = DOT_PRODUCT(cv, MATMUL(AmI, cv))
      fac2 = DOT_PRODUCT(cv, MATMUL(AmI, uv))
      cv(:) = uv-cv*fac2/fac
      cv(:) = MATMUL(AmI, cv)
      IF (g_hartree%pw_grid%have_g0) g_hartree%cc(1) = g_hartree%cc(1)+sfac*fac2/fac
      DO ig = g_hartree%pw_grid%first_gne0, g_hartree%pw_grid%ngpts_cut_local
         g2 = g_hartree%pw_grid%gsq(ig)
         w = 4.0_dp*pi*(g2-gcut2)**2.0_dp/(g2*gcut2)
         IF (g2 > gcut2) EXIT
         gvec = g_hartree%pw_grid%g(:, ig)
         g_corr = 0.0_dp
         idim = 0
         DO iparticle = 1, SIZE(particle_set)
            DO igauss = 1, SIZE(radii)
               idim = idim+1
               rc = radii(igauss)
               rc2 = rc*rc
               rvec = particle_set(iparticle)%r
               arg = DOT_PRODUCT(gvec, rvec)
               phase = CMPLX(COS(arg), -SIN(arg), KIND=dp)
               gfunc = EXP(-g2*rc2/4.0_dp)
               g_corr = g_corr+gfunc*cv(idim)*phase
            END DO
         END DO
         g_corr = g_corr*w
         g_hartree%cc(ig) = g_hartree%cc(ig)+sfac*g_corr/Vol
      END DO
      CALL timestop(handle)
   END SUBROUTINE restraint_functional_potential

! **************************************************************************************************
!> \brief Modify the Hartree potential
!> \param v_hartree_gspace ...
!> \param density_fit_section ...
!> \param particle_set ...
!> \param M ...
!> \param AmI ...
!> \param radii ...
!> \param charges ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE modify_hartree_pot(v_hartree_gspace, density_fit_section, &
                                 particle_set, M, AmI, radii, charges)
      TYPE(pw_p_type)                                    :: v_hartree_gspace
      TYPE(section_vals_type), POINTER                   :: density_fit_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: M, AmI
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii, charges

      CHARACTER(len=*), PARAMETER :: routineN = 'modify_hartree_pot', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: g_corr, phase
      INTEGER                                            :: handle, idim, ig, igauss, iparticle
      REAL(kind=dp)                                      :: arg, fac, fac2, g2, gcut, gcut2, gfunc, &
                                                            gvec(3), rc, rc2, rvec(3), sfac, Vol, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cv, uv
      TYPE(pw_type), POINTER                             :: g_hartree

      CALL timeset(routineN, handle)
      NULLIFY (g_hartree)
      CALL section_vals_val_get(density_fit_section, "GCUT", r_val=gcut)
      gcut2 = gcut*gcut
      g_hartree => v_hartree_gspace%pw
      Vol = g_hartree%pw_grid%vol
      ALLOCATE (cv(SIZE(M, 1)))
      ALLOCATE (uv(SIZE(M, 1)))
      cv = 1.0_dp/Vol
      uv(:) = MATMUL(M, charges)
      sfac = -1.0_dp/Vol
      fac = DOT_PRODUCT(cv, MATMUL(AmI, cv))
      fac2 = DOT_PRODUCT(cv, MATMUL(AmI, uv))
      cv(:) = uv-cv*fac2/fac
      cv(:) = MATMUL(AmI, cv)
      IF (g_hartree%pw_grid%have_g0) g_hartree%cc(1) = g_hartree%cc(1)+sfac*fac2/fac
      DO ig = g_hartree%pw_grid%first_gne0, g_hartree%pw_grid%ngpts_cut_local
         g2 = g_hartree%pw_grid%gsq(ig)
         w = 4.0_dp*pi*(g2-gcut2)**2.0_dp/(g2*gcut2)
         IF (g2 > gcut2) EXIT
         gvec = g_hartree%pw_grid%g(:, ig)
         g_corr = 0.0_dp
         idim = 0
         DO iparticle = 1, SIZE(particle_set)
            DO igauss = 1, SIZE(radii)
               idim = idim+1
               rc = radii(igauss)
               rc2 = rc*rc
               rvec = particle_set(iparticle)%r
               arg = DOT_PRODUCT(gvec, rvec)
               phase = CMPLX(COS(arg), -SIN(arg), KIND=dp)
               gfunc = EXP(-g2*rc2/4.0_dp)
               g_corr = g_corr+gfunc*cv(idim)*phase
            END DO
         END DO
         g_corr = g_corr*w
         g_hartree%cc(ig) = g_hartree%cc(ig)+sfac*g_corr/Vol
      END DO
      CALL timestop(handle)
   END SUBROUTINE modify_hartree_pot

! **************************************************************************************************
!> \brief To Debug the derivative of the B vector for the solution of the
!>      linear system
!> \param dbv ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param iparticle ...
!> \param Vol ...
!> \param qs_env ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE debug_der_b_vector(dbv, particle_set, radii, &
                                 rho_tot_g, gcut, iparticle, Vol, qs_env)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: dbv
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      INTEGER, INTENT(in)                                :: iparticle
      REAL(KIND=dp), INTENT(IN)                          :: Vol
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_der_b_vector', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, kk, ndim
      REAL(KIND=dp)                                      :: dx, rvec(3), v0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: bv1, bv2, ddbv
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env

      NULLIFY (cp_ddapc_env)
      CALL timeset(routineN, handle)
      dx = 0.01_dp
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (bv1(ndim))
      ALLOCATE (bv2(ndim))
      ALLOCATE (ddbv(ndim))
      rvec = particle_set(iparticle)%r
      cp_ddapc_env => qs_env%cp_ddapc_env
      DO i = 1, 3
         bv1(:) = 0.0_dp
         bv2(:) = 0.0_dp
         particle_set(iparticle)%r(i) = rvec(i)+dx
         CALL build_b_vector(bv1, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                             particle_set, radii, rho_tot_g, gcut); bv1(:) = bv1(:)/Vol
         CALL mp_sum(bv1, rho_tot_g%pw_grid%para%group)
         particle_set(iparticle)%r(i) = rvec(i)-dx
         CALL build_b_vector(bv2, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                             particle_set, radii, rho_tot_g, gcut); bv2(:) = bv2(:)/Vol
         CALL mp_sum(bv2, rho_tot_g%pw_grid%para%group)
         ddbv(:) = (bv1(:)-bv2(:))/(2.0_dp*dx)
         DO kk = 1, SIZE(ddbv)
            IF (ddbv(kk) .GT. 1.0E-8_dp) THEN
               v0 = ABS(dbv(kk, i)-ddbv(kk))/ddbv(kk)*100.0_dp
               WRITE (*, *) "Error % on B ::", v0
               IF (v0 .GT. 0.1_dp) THEN
                  WRITE (*, '(A,2I5,2F15.9)') "ERROR IN DERIVATIVE OF B VECTOR, IPARTICLE, ICOORD:", iparticle, i, &
                     dbv(kk, i), ddbv(kk)
                  CALL cp__b("cp_ddapc_util.F",740,"")
               END IF
            END IF
         END DO
         particle_set(iparticle)%r = rvec
      END DO
      DEALLOCATE (bv1)
      DEALLOCATE (bv2)
      DEALLOCATE (ddbv)
      CALL timestop(handle)
   END SUBROUTINE debug_der_b_vector

! **************************************************************************************************
!> \brief To Debug the derivative of the A matrix for the solution of the
!>      linear system
!> \param dAm ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param iparticle ...
!> \param Vol ...
!> \param qs_env ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE debug_der_A_matrix(dAm, particle_set, radii, &
                                 rho_tot_g, gcut, iparticle, Vol, qs_env)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: dAm
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      INTEGER, INTENT(in)                                :: iparticle
      REAL(KIND=dp), INTENT(IN)                          :: Vol
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_der_A_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, kk, ll, ndim
      REAL(KIND=dp)                                      :: dx, rvec(3), v0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Am1, Am2, ddAm, g_dot_rvec_cos, &
                                                            g_dot_rvec_sin
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env

!NB new temporaries sin(g.r) and cos(g.r), as used in get_ddapc, to speed up build_der_A_matrix()

      NULLIFY (cp_ddapc_env)
      CALL timeset(routineN, handle)
      dx = 0.01_dp
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (Am1(ndim, ndim))
      ALLOCATE (Am2(ndim, ndim))
      ALLOCATE (ddAm(ndim, ndim))
      rvec = particle_set(iparticle)%r
      cp_ddapc_env => qs_env%cp_ddapc_env
      CALL prep_g_dot_rvec_sin_cos(rho_tot_g, particle_set, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
      DO i = 1, 3
         Am1 = 0.0_dp
         Am2 = 0.0_dp
         particle_set(iparticle)%r(i) = rvec(i)+dx
         CALL build_A_matrix(Am1, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                             particle_set, radii, rho_tot_g, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
         Am1(:, :) = Am1(:, :)/(Vol*Vol)
         CALL mp_sum(Am1, rho_tot_g%pw_grid%para%group)
         particle_set(iparticle)%r(i) = rvec(i)-dx
         CALL build_A_matrix(Am2, cp_ddapc_env%gfunc, cp_ddapc_env%w, &
                             particle_set, radii, rho_tot_g, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
         Am2(:, :) = Am2(:, :)/(Vol*Vol)
         CALL mp_sum(Am2, rho_tot_g%pw_grid%para%group)
         ddAm(:, :) = (Am1-Am2)/(2.0_dp*dx)
         DO kk = 1, SIZE(ddAm, 1)
            DO ll = 1, SIZE(ddAm, 2)
               IF (ddAm(kk, ll) .GT. 1.0E-8_dp) THEN
                  v0 = ABS(dAm(kk, ll, i)-ddAm(kk, ll))/ddAm(kk, ll)*100.0_dp
                  WRITE (*, *) "Error % on A ::", v0, Am1(kk, ll), Am2(kk, ll), iparticle, i, kk, ll
                  IF (v0 .GT. 0.1_dp) THEN
                     WRITE (*, '(A,4I5,2F15.9)') "ERROR IN DERIVATIVE OF A MATRIX, IPARTICLE, ICOORD:", iparticle, i, kk, ll, &
                        dAm(kk, ll, i), ddAm(kk, ll)
                     CALL cp__b("cp_ddapc_util.F",821,"")
                  END IF
               END IF
            END DO
         END DO
         particle_set(iparticle)%r = rvec
      END DO
      CALL cleanup_g_dot_rvec_sin_cos(g_dot_rvec_sin, g_dot_rvec_cos)
      DEALLOCATE (Am1)
      DEALLOCATE (Am2)
      DEALLOCATE (ddAm)
      CALL timestop(handle)
   END SUBROUTINE debug_der_A_matrix

! **************************************************************************************************
!> \brief To Debug the fitted charges
!> \param dqv ...
!> \param qs_env ...
!> \param density_fit_section ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param type_of_density ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE debug_charge(dqv, qs_env, density_fit_section, &
                           particle_set, radii, rho_tot_g, type_of_density)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: dqv
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: density_fit_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      CHARACTER(LEN=*)                                   :: type_of_density

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_charge', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iparticle, kk, ndim
      REAL(KIND=dp)                                      :: dx, rvec(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ddqv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: qtot1, qtot2

      CALL timeset(routineN, handle)
      WRITE (*, *) "DEBUG_CHARGE_ROUTINE"
      ndim = SIZE(particle_set)*SIZE(radii)
      NULLIFY (qtot1, qtot2)
      ALLOCATE (qtot1(ndim))
      ALLOCATE (qtot2(ndim))
      ALLOCATE (ddqv(ndim))
      !
      dx = 0.001_dp
      DO iparticle = 1, SIZE(particle_set)
         rvec = particle_set(iparticle)%r
         DO i = 1, 3
            particle_set(iparticle)%r(i) = rvec(i)+dx
            CALL get_ddapc(qs_env, .FALSE., density_fit_section, qout1=qtot1, &
                           ext_rho_tot_g=rho_tot_g, Itype_of_density=type_of_density)
            particle_set(iparticle)%r(i) = rvec(i)-dx
            CALL get_ddapc(qs_env, .FALSE., density_fit_section, qout1=qtot2, &
                           ext_rho_tot_g=rho_tot_g, Itype_of_density=type_of_density)
            ddqv(:) = (qtot1-qtot2)/(2.0_dp*dx)
            DO kk = 1, SIZE(qtot1)-1, SIZE(radii)
               IF (ANY(ddqv(kk:kk+2) .GT. 1.0E-8_dp)) THEN
                  WRITE (*, '(A,2F12.6,F12.2)') "Error :", SUM(dqv(kk:kk+2, iparticle, i)), SUM(ddqv(kk:kk+2)), &
                     ABS((SUM(ddqv(kk:kk+2))-SUM(dqv(kk:kk+2, iparticle, i)))/SUM(ddqv(kk:kk+2))*100.0_dp)
               END IF
            END DO
            particle_set(iparticle)%r = rvec
         END DO
      END DO
      !
      DEALLOCATE (qtot1)
      DEALLOCATE (qtot2)
      DEALLOCATE (ddqv)
      CALL timestop(handle)
   END SUBROUTINE debug_charge

END MODULE cp_ddapc_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for performing an outer scf loop
!> \par History
!>      Created [2006.03]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE qs_outer_scf
   USE cp_control_types,                ONLY: becke_restraint_type,&
                                              ddapc_restraint_type,&
                                              dft_control_type,&
                                              s2_restraint_type
   USE input_constants,                 ONLY: &
        do_ddapc_constraint, do_s2_constraint, outer_scf_becke_constraint, &
        outer_scf_ddapc_constraint, outer_scf_none, outer_scf_optimizer_bisect, &
        outer_scf_optimizer_diis, outer_scf_optimizer_none, outer_scf_optimizer_sd, &
        outer_scf_s2_constraint
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_outer_scf'

! *** Public subroutines ***

   PUBLIC :: outer_loop_gradient, outer_loop_optimize, outer_loop_update_qs_env, &
             outer_loop_variables_count, outer_loop_extrapolate

CONTAINS

! **************************************************************************************************
!> \brief returns the number of variables that is employed in the outer loop
!> \param scf_control ...
!> \retval res ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   FUNCTION outer_loop_variables_count(scf_control) RESULT(res)
      TYPE(scf_control_type), POINTER                    :: scf_control
      INTEGER                                            :: res

      SELECT CASE (scf_control%outer_scf%type)
      CASE (outer_scf_ddapc_constraint)
         res = 1
      CASE (outer_scf_s2_constraint)
         res = 1
      CASE (outer_scf_becke_constraint)
         res = 1
      CASE (outer_scf_none) ! just needed to communicate the gradient criterium
         res = 1
      CASE DEFAULT
         res = 0
      END SELECT

   END FUNCTION outer_loop_variables_count

! **************************************************************************************************
!> \brief computes the gradient wrt to the outer loop variables
!> \param qs_env ...
!> \param scf_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE outer_loop_gradient(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'outer_loop_gradient', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ihistory, n
      LOGICAL                                            :: is_constraint
      TYPE(becke_restraint_type), POINTER                :: becke_control
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(s2_restraint_type), POINTER                   :: s2_restraint_control
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, scf_control=scf_control, &
                      dft_control=dft_control, energy=energy)
      IF(.NOT.(scf_control%outer_scf%have_scf))CALL cp__a("qs_outer_scf.F",100)

      ihistory = scf_env%outer_scf%iter_count
      IF(.NOT.(ihistory <= SIZE(scf_env%outer_scf%energy, 1)))CALL cp__a("qs_outer_scf.F",103)

      scf_env%outer_scf%energy(ihistory) = energy%total

      SELECT CASE (scf_control%outer_scf%type)
      CASE (outer_scf_none)
         ! just pass the inner loop scf criterium to the outer loop one
         scf_env%outer_scf%variables(1, ihistory) = scf_env%iter_delta
         scf_env%outer_scf%gradient(1, ihistory) = scf_env%iter_delta
      CASE (outer_scf_ddapc_constraint)
         IF(.NOT.(dft_control%qs_control%ddapc_restraint))CALL cp__a("qs_outer_scf.F",113)
         DO n = 1, SIZE(dft_control%qs_control%ddapc_restraint_control)
            NULLIFY (ddapc_restraint_control)
            ddapc_restraint_control => dft_control%qs_control%ddapc_restraint_control(n)%ddapc_restraint_control
            is_constraint = (ddapc_restraint_control%functional_form == do_ddapc_constraint)
            IF (is_constraint) EXIT
         END DO
         IF(.NOT.(is_constraint))CALL cp__a("qs_outer_scf.F",120)

         scf_env%outer_scf%variables(:, ihistory) = ddapc_restraint_control%strength
         scf_env%outer_scf%gradient(:, ihistory) = ddapc_restraint_control%ddapc_order_p- &
                                                   ddapc_restraint_control%target
      CASE (outer_scf_s2_constraint)
         IF(.NOT.(dft_control%qs_control%s2_restraint))CALL cp__a("qs_outer_scf.F",126)
         s2_restraint_control => dft_control%qs_control%s2_restraint_control
         is_constraint = (s2_restraint_control%functional_form == do_s2_constraint)
         IF(.NOT.(is_constraint))CALL cp__a("qs_outer_scf.F",129)

         scf_env%outer_scf%variables(:, ihistory) = s2_restraint_control%strength
         scf_env%outer_scf%gradient(:, ihistory) = s2_restraint_control%s2_order_p- &
                                                   s2_restraint_control%target
      CASE (outer_scf_becke_constraint)
         IF(.NOT.(dft_control%qs_control%becke_restraint))CALL cp__a("qs_outer_scf.F",135)
         becke_control => dft_control%qs_control%becke_control
         scf_env%outer_scf%variables(:, ihistory) = becke_control%strength
         scf_env%outer_scf%gradient(:, ihistory) = becke_control%becke_order_p- &
                                                   becke_control%target
      CASE DEFAULT
         CALL cp__b("qs_outer_scf.F",141,"")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE outer_loop_gradient

! **************************************************************************************************
!> \brief optimizes the parameters of the outer_scf
!> \param scf_env ...
!> \param scf_control ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \note
!>       ought to be general, and independent of the actual kind of variables
! **************************************************************************************************
   SUBROUTINE outer_loop_optimize(scf_env, scf_control)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'outer_loop_optimize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ibuf, ihigh, ihistory, ilow, &
                                                            j, jbuf, nb, optimizer_type
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipivot
      REAL(KIND=dp)                                      :: interval, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ev
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: a, b

      CALL timeset(routineN, handle)

      ihistory = scf_env%outer_scf%iter_count
      optimizer_type = scf_control%outer_scf%optimizer

      DO WHILE (.TRUE.) ! if we need a different run type we'll restart here

         SELECT CASE (optimizer_type)
         CASE (outer_scf_optimizer_bisect) ! bisection on the gradient, needs to be 1D
            IF(.NOT.(SIZE(scf_env%outer_scf%gradient(:, 1)) == 1))CALL cp__a("qs_outer_scf.F",180)
            ! find the pair of points that bracket a zero of the gradient, with the smallest interval possible
            ilow = -1
            ihigh = -1
            interval = HUGE(interval)
            DO i = 1, ihistory
               DO j = i+1, ihistory
                  ! distrust often used points
                  IF (scf_env%outer_scf%count(i) .GT. scf_control%outer_scf%bisect_trust_count) CYCLE
                  IF (scf_env%outer_scf%count(j) .GT. scf_control%outer_scf%bisect_trust_count) CYCLE

                  ! if they bracket a zero use them
                  IF (scf_env%outer_scf%gradient(1, i)* &
                      scf_env%outer_scf%gradient(1, j) < 0.0_dp) THEN
                     tmp = ABS(scf_env%outer_scf%variables(1, i)-scf_env%outer_scf%variables(1, j))
                     IF (tmp < interval) THEN
                        ilow = i
                        ihigh = j
                        interval = tmp
                     ENDIF
                  ENDIF
               ENDDO
            ENDDO
            IF (ilow == -1) THEN ! we didn't bracket a minimum yet, try something else
               optimizer_type = outer_scf_optimizer_diis
               CYCLE
            ENDIF
            scf_env%outer_scf%count(ilow) = scf_env%outer_scf%count(ilow)+1
            scf_env%outer_scf%count(ihigh) = scf_env%outer_scf%count(ihigh)+1
            scf_env%outer_scf%variables(:, ihistory+1) = 0.5_dp*(scf_env%outer_scf%variables(:, ilow)+ &
                                                                 scf_env%outer_scf%variables(:, ihigh))
         CASE (outer_scf_optimizer_none)
            scf_env%outer_scf%variables(:, ihistory+1) = scf_env%outer_scf%variables(:, ihistory)
         CASE (outer_scf_optimizer_sd)
            ! Notice that we are just trying to find a stationary point
            ! e.g. the ddpac_constraint, one maximizes the function, so the stepsize might have
            ! to be negative
            scf_env%outer_scf%variables(:, ihistory+1) = scf_env%outer_scf%variables(:, ihistory)- &
                                                         scf_control%outer_scf%step_size*scf_env%outer_scf%gradient(:, ihistory)
         CASE (outer_scf_optimizer_diis)
            IF(.NOT.(scf_control%outer_scf%diis_buffer_length > 0))CALL cp__a("qs_outer_scf.F",220)
            ! set up DIIS matrix
            nb = MIN(ihistory, scf_control%outer_scf%diis_buffer_length)
            IF (nb < 2) THEN
               optimizer_type = outer_scf_optimizer_sd
               CYCLE
            ELSE
               ALLOCATE (b(nb+1, nb+1), a(nb+1, nb+1), ev(nb+1), ipivot(nb+1))
               DO I = 1, nb
                  DO J = I, nb
                     ibuf = ihistory-nb+i
                     jbuf = ihistory-nb+j
                     b(I, J) = DOT_PRODUCT(scf_env%outer_scf%gradient(:, ibuf), &
                                           scf_env%outer_scf%gradient(:, jbuf))
                     b(J, I) = b(I, J)
                  ENDDO
               ENDDO
               b(nb+1, :) = -1.0_dp
               b(:, nb+1) = -1.0_dp
               b(nb+1, nb+1) = 0.0_dp

               CALL diamat_all(b, ev)
               a(:, :) = b
               DO I = 1, nb+1
                  IF (ABS(ev(I)) .LT. 1.0E-12_dp) THEN
                     a(:, I) = 0.0_dp
                  ELSE
                     a(:, I) = a(:, I)/ev(I)
                  ENDIF
               END DO
               ev(:) = -MATMUL(a, b(nb+1, :))

               scf_env%outer_scf%variables(:, ihistory+1) = 0.0_dp
               DO i = 1, nb
                  ibuf = ihistory-nb+i
                  scf_env%outer_scf%variables(:, ihistory+1) = scf_env%outer_scf%variables(:, ihistory+1)+ &
                                                               ev(i)*scf_env%outer_scf%variables(:, ibuf)
               ENDDO
               DEALLOCATE (b, ev)
            ENDIF
         CASE DEFAULT
            CALL cp__b("qs_outer_scf.F",261,"")
         END SELECT
         EXIT
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE outer_loop_optimize

! **************************************************************************************************
!> \brief propagates the updated variables to wherever they need to be set in
!>       qs_env
!> \param qs_env ...
!> \param scf_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE outer_loop_update_qs_env(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'outer_loop_update_qs_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ihistory, n
      LOGICAL                                            :: is_constraint
      TYPE(becke_restraint_type), POINTER                :: becke_control
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(s2_restraint_type), POINTER                   :: s2_restraint_control
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, scf_control=scf_control, dft_control=dft_control)
      ihistory = scf_env%outer_scf%iter_count

      SELECT CASE (scf_control%outer_scf%type)
      CASE (outer_scf_none)
         ! do nothing
      CASE (outer_scf_ddapc_constraint)
         DO n = 1, SIZE(dft_control%qs_control%ddapc_restraint_control)
            NULLIFY (ddapc_restraint_control)
            ddapc_restraint_control => dft_control%qs_control%ddapc_restraint_control(n)%ddapc_restraint_control
            is_constraint = (ddapc_restraint_control%functional_form == do_ddapc_constraint)
            IF (is_constraint) EXIT
         END DO
         ddapc_restraint_control%strength = scf_env%outer_scf%variables(1, ihistory+1)
      CASE (outer_scf_s2_constraint)
         s2_restraint_control => dft_control%qs_control%s2_restraint_control
         s2_restraint_control%strength = scf_env%outer_scf%variables(1, ihistory+1)
      CASE (outer_scf_becke_constraint)
         becke_control => dft_control%qs_control%becke_control
         becke_control%strength = scf_env%outer_scf%variables(1, ihistory+1)
      CASE DEFAULT
         CALL cp__b("qs_outer_scf.F",316,"")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE outer_loop_update_qs_env

! **************************************************************************************************
!> \brief uses the outer_scf_history to extrapolate new values for the variables
!>       and updates their value in qs_env accordingly
!> \param qs_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \note
!>       it assumes that the current value of qs_env still needs to be added to the history
!>       simple multilinear extrapolation is employed
! **************************************************************************************************
   SUBROUTINE outer_loop_extrapolate(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'outer_loop_extrapolate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ihis, ivec, n, nhistory, &
                                                            nvariables, nvec, outer_scf_ihistory
      LOGICAL                                            :: is_constraint
      REAL(kind=dp)                                      :: alpha
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: extrapolation
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: outer_scf_history
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, outer_scf_history=outer_scf_history, &
                      outer_scf_ihistory=outer_scf_ihistory, &
                      scf_control=scf_control, dft_control=dft_control)

      nvariables = SIZE(outer_scf_history, 1)
      nhistory = SIZE(outer_scf_history, 2)
      ALLOCATE (extrapolation(nvariables))
      IF(.NOT.(nhistory > 0))CALL cp__a("qs_outer_scf.F",358)

      ! add the current version of qs_env to the history
      outer_scf_ihistory = outer_scf_ihistory+1
      ivec = 1+MODULO(outer_scf_ihistory-1, nhistory)
      SELECT CASE (scf_control%outer_scf%type)
      CASE (outer_scf_none)
         outer_scf_history(1, ivec) = 0.0_dp
      CASE (outer_scf_ddapc_constraint)
         DO n = 1, SIZE(dft_control%qs_control%ddapc_restraint_control)
            NULLIFY (ddapc_restraint_control)
            ddapc_restraint_control => dft_control%qs_control%ddapc_restraint_control(n)%ddapc_restraint_control
            is_constraint = (ddapc_restraint_control%functional_form == do_ddapc_constraint)
            IF (is_constraint) EXIT
         END DO
         outer_scf_history(1, ivec) = &
            ddapc_restraint_control%strength
      CASE (outer_scf_s2_constraint)
         outer_scf_history(1, ivec) = &
            dft_control%qs_control%s2_restraint_control%strength
      CASE (outer_scf_becke_constraint)
         outer_scf_history(1, ivec) = &
            dft_control%qs_control%becke_control%strength
      CASE DEFAULT
         CALL cp__b("qs_outer_scf.F",382,"")
      END SELECT
      CALL set_qs_env(qs_env, outer_scf_ihistory=outer_scf_ihistory)

      ! multilinear extrapolation
      nvec = MIN(nhistory, outer_scf_ihistory)
      alpha = nvec
      ivec = 1+MODULO(outer_scf_ihistory-1, nhistory)
      extrapolation(:) = alpha*outer_scf_history(:, ivec)
      DO ihis = 2, nvec
         alpha = -1.0_dp*alpha*REAL(nvec-ihis+1, dp)/REAL(ihis, dp)
         ivec = 1+MODULO(outer_scf_ihistory-ihis, nhistory)
         extrapolation(:) = extrapolation+alpha*outer_scf_history(:, ivec)
      ENDDO

      ! update qs_env to use this extrapolation
      SELECT CASE (scf_control%outer_scf%type)
      CASE (outer_scf_none)
         ! nothing
      CASE (outer_scf_ddapc_constraint)
         ddapc_restraint_control%strength = extrapolation(1)
      CASE (outer_scf_s2_constraint)
         dft_control%qs_control%s2_restraint_control%strength = extrapolation(1)
      CASE (outer_scf_becke_constraint)
         dft_control%qs_control%becke_control%strength = extrapolation(1)
      CASE DEFAULT
         CALL cp__b("qs_outer_scf.F",408,"")
      END SELECT

      DEALLOCATE (extrapolation)

      CALL timestop(handle)

   END SUBROUTINE outer_loop_extrapolate

END MODULE qs_outer_scf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief pw_methods
!> \author CJM
! **************************************************************************************************
MODULE ewald_pw_methods
   USE dg_rho0_types,                   ONLY: dg_rho0_get,&
                                              dg_rho0_init,&
                                              dg_rho0_set,&
                                              dg_rho0_type
   USE dg_types,                        ONLY: dg_get,&
                                              dg_set,&
                                              dg_type
   USE dgs,                             ONLY: dg_grid_change
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_env_set,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_set,&
                                              ewald_pw_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_change
   USE pw_poisson_methods,              ONLY: pw_poisson_set
! CJM, pw_poisson_rebuild
   USE pw_poisson_read_input,           ONLY: pw_poisson_read_parameters
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme,&
                                              pw_poisson_create,&
                                              pw_poisson_parameter_type,&
                                              pw_poisson_release,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_p_type,&
                                              pw_pool_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewald_pw_methods'

   PUBLIC :: ewald_pw_grid_update

CONTAINS

! **************************************************************************************************
!> \brief Rescales pw_grids for given box, if nessecary
!> \param ewald_pw ...
!> \param ewald_env ...
!> \param cell_hmat ...
!> \par History
!>      none
!> \author JGH (15-Mar-2001)
! **************************************************************************************************
   SUBROUTINE ewald_pw_grid_update(ewald_pw, ewald_env, cell_hmat)
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      REAL(KIND=dp), DIMENSION(3, 3)                     :: cell_hmat

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_grid_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, o_spline
      REAL(dp)                                           :: alpha
      REAL(KIND=dp), DIMENSION(3, 3)                     :: old_cell_hmat
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_poisson_parameter_type)                    :: poisson_params
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: pw_big_pool, pw_small_pool
      TYPE(section_vals_type), POINTER                   :: poisson_section

      NULLIFY (pw_big_pool)
      NULLIFY (pw_small_pool)
      NULLIFY (dg, poisson_env, poisson_section)

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, &
                         alpha=alpha, o_spline=o_spline, &
                         poisson_section=poisson_section, &
                         cell_hmat=old_cell_hmat)

      IF (ALL(cell_hmat == old_cell_hmat)) RETURN ! rebuild not needed

      CALL ewald_env_set(ewald_env, cell_hmat=cell_hmat)

      SELECT CASE (ewald_type)
      CASE (do_ewald_ewald)
         CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_big_pool, &
                           dg=dg, poisson_env=poisson_env)
         CALL pw_grid_change(cell_hmat, pw_big_pool%pw_grid)
         CALL ewald_pw_rho0_setup(ewald_env, pw_big_pool%pw_grid, dg)
         CALL pw_poisson_release(poisson_env)
         CALL ewald_pw_set(ewald_pw, pw_big_pool=pw_big_pool, dg=dg, &
                           poisson_env=poisson_env)
      CASE (do_ewald_pme)
         CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_big_pool, &
                           pw_small_pool=pw_small_pool, dg=dg, &
                           poisson_env=poisson_env)
         IF (.NOT. ASSOCIATED(poisson_env)) THEN
            CALL pw_poisson_create(poisson_env)
            CALL ewald_pw_set(ewald_pw, poisson_env=poisson_env)
         END IF
         CALL pw_grid_change(cell_hmat, pw_big_pool%pw_grid)
         CALL dg_grid_change(cell_hmat, pw_big_pool%pw_grid, pw_small_pool%pw_grid)
         CALL ewald_pw_rho0_setup(ewald_env, pw_small_pool%pw_grid, dg)
         CALL ewald_pw_set(ewald_pw, pw_big_pool=pw_big_pool, &
                           pw_small_pool=pw_small_pool, dg=dg, &
                           poisson_env=poisson_env)
      CASE (do_ewald_spme)
         CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_big_pool, &
                           poisson_env=poisson_env)
         IF (.NOT. ASSOCIATED(poisson_env)) THEN
            CALL pw_poisson_create(poisson_env)
         END IF
         CALL pw_grid_change(cell_hmat, pw_big_pool%pw_grid)
         CALL ewald_pw_set(ewald_pw, pw_big_pool=pw_big_pool, &
                           poisson_env=poisson_env)
      CASE (do_ewald_none)
      CASE default
         CALL cp__b("ewald_pw_methods.F",130,"")
      END SELECT
      IF (ASSOCIATED(poisson_env)) THEN
         ALLOCATE (pw_pools(1))
         pw_pools(1)%pool => pw_big_pool
         CALL pw_poisson_read_parameters(poisson_section, poisson_params)
         CALL pw_poisson_set(poisson_env, cell_hmat=cell_hmat, parameters=poisson_params, &
                             use_level=1, pw_pools=pw_pools)
         DEALLOCATE (pw_pools)
      END IF

   END SUBROUTINE ewald_pw_grid_update

! **************************************************************************************************
!> \brief Calculates the Fourier transform of the "Ewald function"
!> \param ewald_env ...
!> \param pw_grid ...
!> \param dg ...
!> \par History
!>      none
!> \author JGH (15-Mar-2001)
! **************************************************************************************************
   SUBROUTINE ewald_pw_rho0_setup(ewald_env, pw_grid, dg)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(dg_type), POINTER                             :: dg

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_rho0_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, grid_index, isos
      REAL(dp)                                           :: alpha
      REAL(dp), POINTER                                  :: gcc(:), zet(:)
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0

      CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type)
      CALL dg_get(dg, dg_rho0=dg_rho0)
      CALL dg_rho0_get(dg_rho0, gcc=gcc, zet=zet)

! This is the first ( and only ) double grid
      grid_index = 1

      IF (.NOT. ASSOCIATED(zet)) THEN
         ALLOCATE (zet(1), STAT=isos)
         IF (isos /= 0) CALL stop_memory(routineN, moduleN, 174, &
                                         "zet", dp_size)
      ENDIF

! No contracted Gaussians are used here
      NULLIFY (gcc)

      zet(1) = alpha
      CALL dg_rho0_set(dg_rho0, TYPE=ewald_type, zet=zet)

      CALL dg_rho0_init(dg_rho0, pw_grid)

      CALL dg_set(dg, dg_rho0=dg_rho0, grid_index=grid_index)

   END SUBROUTINE ewald_pw_rho0_setup

END MODULE ewald_pw_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods of pw_env that have dependence on qs_env
!> \par History
!>      10.2002 created [fawzi]
!>      JGH (22-Feb-03) PW grid options added
!>      04.2003 added rs grid pools [fawzi]
!>      02.2004 added commensurate grids
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE pw_env_methods
   USE ao_util,                         ONLY: exp_radius
   USE basis_set_container_types,       ONLY: translate_basis_type
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_init,          ONLY: init_input_type
   USE cube_utils,                      ONLY: destroy_cube_info,&
                                              init_cube_info,&
                                              return_cube_max_iradius
   USE d3_poly,                         ONLY: init_d3_poly_module
   USE dct,                             ONLY: neumannX,&
                                              neumannXY,&
                                              neumannXYZ,&
                                              neumannXZ,&
                                              neumannY,&
                                              neumannYZ,&
                                              neumannZ,&
                                              setup_dct_pw_grids
   USE dielectric_types,                ONLY: derivative_cd3,&
                                              derivative_cd5,&
                                              derivative_cd7,&
                                              rho_dependent
   USE gaussian_gridlevels,             ONLY: destroy_gaussian_gridlevel,&
                                              gaussian_gridlevel,&
                                              init_gaussian_gridlevel
   USE input_constants,                 ONLY: xc_vdw_fun_nonloc
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE lgrid_types,                     ONLY: lgrid_create,&
                                              lgrid_release
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE ps_wavelet_types,                ONLY: WAVELET0D,&
                                              WAVELET2D,&
                                              WAVELET3D
   USE pw_env_types,                    ONLY: pw_env_type
   USE pw_grid_info,                    ONLY: pw_grid_init_setup
   USE pw_grid_types,                   ONLY: FULLSPACE,&
                                              HALFSPACE,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: do_pw_grid_blocked_false,&
                                              pw_grid_change,&
                                              pw_grid_create,&
                                              pw_grid_release,&
                                              pw_grid_setup
   USE pw_poisson_methods,              ONLY: pw_poisson_set
   USE pw_poisson_read_input,           ONLY: pw_poisson_read_parameters
   USE pw_poisson_types,                ONLY: &
        pw_poisson_analytic, pw_poisson_create, pw_poisson_implicit, pw_poisson_mt, &
        pw_poisson_multipole, pw_poisson_none, pw_poisson_parameter_type, pw_poisson_periodic, &
        pw_poisson_wavelet
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_p_type,&
                                              pw_pool_release,&
                                              pw_pool_retain,&
                                              pw_pools_dealloc
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_rho0_types,                   ONLY: get_rho0_mpole,&
                                              rho0_mpole_type
   USE realspace_grid_types,            ONLY: &
        realspace_grid_desc_p_type, realspace_grid_desc_type, realspace_grid_input_type, &
        realspace_grid_p_type, realspace_grid_type, rs_grid_create, rs_grid_create_descriptor, &
        rs_grid_print, rs_grid_release, rs_grid_release_descriptor
   USE xc_input_constants,              ONLY: &
        xc_deriv_collocate, xc_deriv_nn10_smooth, xc_deriv_nn50_smooth, xc_deriv_pw, &
        xc_deriv_spline2, xc_deriv_spline2_smooth, xc_deriv_spline3, xc_deriv_spline3_smooth, &
        xc_rho_nn10, xc_rho_nn50, xc_rho_no_smooth, xc_rho_spline2_smooth, xc_rho_spline3_smooth
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_env_methods'

   PUBLIC :: pw_env_create, pw_env_rebuild
!***
CONTAINS

! **************************************************************************************************
!> \brief creates a pw_env, if qs_env is given calls pw_env_rebuild
!> \param pw_env the pw_env that gets created
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_env_create(pw_env)
      TYPE(pw_env_type), POINTER                         :: pw_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(pw_env)))CALL cp__a("pw_env_methods.F",130)
      ALLOCATE (pw_env)
      NULLIFY (pw_env%pw_pools, pw_env%gridlevel_info, pw_env%poisson_env, &
               pw_env%cube_info, pw_env%rs_descs, pw_env%rs_grids, &
               pw_env%xc_pw_pool, pw_env%vdw_pw_pool, pw_env%lgrid, &
               pw_env%interp_section)
      pw_env%auxbas_grid = -1
      pw_env%ref_count = 1

      CALL timestop(handle)

   END SUBROUTINE pw_env_create

! **************************************************************************************************
!> \brief rebuilds the pw_env data (necessary if cell or cutoffs change)
!> \param pw_env the environment to rebuild
!> \param qs_env the qs_env where to get the cell, cutoffs,...
!> \param external_para_env ...
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_env_rebuild(pw_env, qs_env, external_para_env)
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: external_para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_env_rebuild', routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: string
      INTEGER :: blocked_id, blocked_id_input, boundary_condition, grid_span, handle, i, &
         igrid_level, iounit, ncommensurate, ngrid_level, xc_deriv_method_id, xc_smooth_method_id
      INTEGER, DIMENSION(2)                              :: distribution_layout
      INTEGER, DIMENSION(3)                              :: higher_grid_layout
      LOGICAL                                            :: efg_present, linres_present, odd, &
                                                            set_vdw_pool, should_output, &
                                                            smooth_required, spherical, uf_grid, &
                                                            use_ref_cell
      REAL(KIND=dp)                                      :: cutilev, max_rpgf0_s, rel_cutoff, zet0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: radius
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cutoff
      TYPE(cell_type), POINTER                           :: cell, cell_ref, my_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_grid_type), POINTER                        :: dct_pw_grid, mt_super_ref_grid, &
                                                            old_pw_grid, pw_grid, super_ref_grid, &
                                                            vdw_grid, xc_super_ref_grid
      TYPE(pw_poisson_parameter_type)                    :: poisson_params
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_input_type)                    :: input_settings
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_grids
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(section_vals_type), POINTER                   :: efg_section, input, linres_section, &
                                                            poisson_section, print_section, &
                                                            rs_grid_section, xc_section

! a very small safety factor might be needed for roundoff issues
! e.g. radius being computed here as 12.998 (13) and 13.002 (14) during the collocation
! te latter can happen due to the lower precision in the computation of the radius in collocate
! parallel cost of rs_pw_transfer goes as safety_factor**3 so it is worthwhile keeping it tight
! Edit: Safety Factor was unused

      CALL timeset(routineN, handle)

!
!
! Part one, deallocate old data if needed
!
!
      NULLIFY (cutoff, cell, pw_grid, old_pw_grid, dft_control, qs_kind_set, &
               pw_pools, rho0_mpole, rs_descs, para_env, cell_ref, &
               mt_super_ref_grid, input, poisson_section, xc_super_ref_grid, &
               dct_pw_grid, vdw_grid, super_ref_grid, my_cell, rs_grids, dispersion_env)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      qs_kind_set=qs_kind_set, &
                      cell_ref=cell_ref, &
                      cell=cell, &
                      para_env=para_env, &
                      input=input, &
                      dispersion_env=dispersion_env)

      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("pw_env_methods.F",218)
      IF(.NOT.(pw_env%ref_count > 0))CALL cp__a("pw_env_methods.F",219)
      CALL pw_pool_release(pw_env%vdw_pw_pool)
      CALL pw_pool_release(pw_env%xc_pw_pool)
      CALL pw_pools_dealloc(pw_env%pw_pools)
      IF (ASSOCIATED(pw_env%rs_descs)) THEN
         DO i = 1, SIZE(pw_env%rs_descs)
            CALL rs_grid_release_descriptor(pw_env%rs_descs(i)%rs_desc)
         END DO
         DEALLOCATE (pw_env%rs_descs)
      END IF
      IF (ASSOCIATED(pw_env%rs_grids)) THEN
         DO i = 1, SIZE(pw_env%rs_grids)
            CALL rs_grid_release(pw_env%rs_grids(i)%rs_grid)
         END DO
         DEALLOCATE (pw_env%rs_grids)
      END IF
      CALL lgrid_release(pw_env%lgrid)
      IF (ASSOCIATED(pw_env%gridlevel_info)) THEN
         CALL destroy_gaussian_gridlevel(pw_env%gridlevel_info)
      ELSE
         ALLOCATE (pw_env%gridlevel_info)
      END IF
      IF (dft_control%qs_control%gapw) THEN
         CALL get_qs_env(qs_env=qs_env, rho0_mpole=rho0_mpole)
         IF(.NOT.(ASSOCIATED(rho0_mpole)))CALL cp__a("pw_env_methods.F",243)
         CALL get_rho0_mpole(rho0_mpole=rho0_mpole, &
                             zet0_h=zet0, max_rpgf0_s=max_rpgf0_s)
      END IF

      IF (ASSOCIATED(pw_env%cube_info)) THEN
         DO igrid_level = 1, SIZE(pw_env%cube_info)
            CALL destroy_cube_info(pw_env%cube_info(igrid_level))
         END DO
         DEALLOCATE (pw_env%cube_info)
      END IF
      NULLIFY (pw_env%pw_pools, pw_env%cube_info)

!
!
! Part two, setup the pw_grids
!
!

      IF (PRESENT(external_para_env)) THEN
         para_env => external_para_env
         IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("pw_env_methods.F",264)
      ENDIF
      ! interpolation section
      pw_env%interp_section => section_vals_get_subs_vals(input, "DFT%MGRID%INTERPOLATOR")

      CALL get_qs_env(qs_env, use_ref_cell=use_ref_cell)
      IF (use_ref_cell) THEN
         my_cell => cell_ref
      ELSE
         my_cell => cell
      END IF
      rel_cutoff = dft_control%qs_control%relative_cutoff
      cutoff => dft_control%qs_control%e_cutoff
      CALL section_vals_val_get(input, "DFT%XC%XC_GRID%USE_FINER_GRID", &
                                l_val=uf_grid)
      ngrid_level = SIZE(cutoff)

      ! init gridlevel_info XXXXXXXXX setup mapping to the effective cutoff ?
      !                     XXXXXXXXX the cutoff array here is more a 'wish-list'
      !                     XXXXXXXXX same holds for radius
      print_section => section_vals_get_subs_vals(input, &
                                                  "PRINT%GRID_INFORMATION")
      CALL init_gaussian_gridlevel(pw_env%gridlevel_info, &
                                   ngrid_levels=ngrid_level, cutoff=cutoff, rel_cutoff=rel_cutoff, &
                                   print_section=print_section)
      ! init pw_grids and pools
      ALLOCATE (pw_pools(ngrid_level))

      IF (dft_control%qs_control%commensurate_mgrids) THEN
         ncommensurate = ngrid_level
      ELSE
         ncommensurate = 0
      ENDIF
      !
      ! If Tuckerman is present let's perform the set-up of the super-reference-grid
      !
      cutilev = cutoff(1)
      IF (dft_control%qs_control%pw_grid_opt%spherical) THEN
         grid_span = HALFSPACE
         spherical = .TRUE.
      ELSE IF (dft_control%qs_control%pw_grid_opt%fullspace) THEN
         grid_span = FULLSPACE
         spherical = .FALSE.
      ELSE
         grid_span = HALFSPACE
         spherical = .FALSE.
      END IF

      CALL setup_super_ref_grid(super_ref_grid, mt_super_ref_grid, &
                                xc_super_ref_grid, cutilev, grid_span, spherical, my_cell, para_env, &
                                qs_env%input, ncommensurate, uf_grid=uf_grid, &
                                print_section=print_section)
      old_pw_grid => super_ref_grid
      !
      ! Setup of the multi-grid pw_grid and pw_pools
      !
      logger => cp_get_default_logger()
      iounit = cp_print_key_unit_nr(logger, print_section, '', &
                                    extension='.Log')

      IF (dft_control%qs_control%pw_grid_opt%spherical) THEN
         grid_span = HALFSPACE
         spherical = .TRUE.
         odd = .TRUE.
      ELSE IF (dft_control%qs_control%pw_grid_opt%fullspace) THEN
         grid_span = FULLSPACE
         spherical = .FALSE.
         odd = .FALSE.
      ELSE
         grid_span = HALFSPACE
         spherical = .FALSE.
         odd = .TRUE.
      END IF

      ! use input suggestion for blocked
      blocked_id_input = dft_control%qs_control%pw_grid_opt%blocked

      ! methods that require smoothing or nearest neighbor have to use a plane distributed setup
      ! find the xc properties (FIXME this could miss other xc sections that operate on the grid ...)
      CALL get_qs_env(qs_env=qs_env, input=input)
      xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      xc_deriv_method_id = section_get_ival(xc_section, "XC_GRID%XC_DERIV")
      xc_smooth_method_id = section_get_ival(xc_section, "XC_GRID%XC_SMOOTH_RHO")
      smooth_required = .FALSE.
      SELECT CASE (xc_deriv_method_id)
      CASE (xc_deriv_pw, xc_deriv_collocate, xc_deriv_spline3, xc_deriv_spline2)
         smooth_required = smooth_required .OR. .FALSE.
      CASE (xc_deriv_spline2_smooth, &
            xc_deriv_spline3_smooth, xc_deriv_nn10_smooth, xc_deriv_nn50_smooth)
         smooth_required = smooth_required .OR. .TRUE.
      CASE DEFAULT
         CALL cp__b("pw_env_methods.F",355,"")
      END SELECT
      SELECT CASE (xc_smooth_method_id)
      CASE (xc_rho_no_smooth)
         smooth_required = smooth_required .OR. .FALSE.
      CASE (xc_rho_spline2_smooth, xc_rho_spline3_smooth, xc_rho_nn10, xc_rho_nn50)
         smooth_required = smooth_required .OR. .TRUE.
      CASE DEFAULT
         CALL cp__b("pw_env_methods.F",363,"")
      END SELECT
      ! EPR, NMR, EFG can require splines. If the linres/EFG section is present we assume
      ! it could be on and splines might be used (not quite sure if this is due to their use of splines or something else)
      linres_section => section_vals_get_subs_vals(section_vals=input, &
                                                   subsection_name="PROPERTIES%LINRES")
      CALL section_vals_get(linres_section, explicit=linres_present)
      IF (linres_present) THEN
         smooth_required = smooth_required .OR. .TRUE.
      ENDIF

      efg_section => section_vals_get_subs_vals(section_vals=input, &
                                                subsection_name="DFT%PRINT%ELECTRIC_FIELD_GRADIENT")
      CALL section_vals_get(efg_section, explicit=efg_present)
      IF (efg_present) THEN
         smooth_required = smooth_required .OR. .TRUE.
      ENDIF

      DO igrid_level = 1, ngrid_level
         CALL pw_grid_create(pw_grid, para_env%group)

         cutilev = cutoff(igrid_level)

         ! the whole of QS seems to work fine with either blocked/non-blocked distribution in g-space
         ! the default choice should be made free
         blocked_id = blocked_id_input

         distribution_layout = dft_control%qs_control%pw_grid_opt%distribution_layout

         ! qmmm does not support a ray distribution
         ! FIXME ... check if a plane distributed lower grid is sufficient
         IF (qs_env%qmmm) THEN
            distribution_layout = (/para_env%num_pe, 1/)
         ENDIF

         ! If splines are required
         ! FIXME.... should only be true for the highest grid
         IF (smooth_required) THEN
            distribution_layout = (/para_env%num_pe, 1/)
         ENDIF

         IF (igrid_level == 1) THEN
            IF (ASSOCIATED(old_pw_grid)) THEN
               CALL pw_grid_setup(my_cell%hmat, pw_grid, grid_span=grid_span, &
                                  cutoff=cutilev, &
                                  spherical=spherical, odd=odd, fft_usage=.TRUE., &
                                  ncommensurate=ncommensurate, icommensurate=igrid_level, &
                                  blocked=do_pw_grid_blocked_false, &
                                  ref_grid=old_pw_grid, &
                                  rs_dims=distribution_layout, &
                                  iounit=iounit)
               old_pw_grid => pw_grid
            ELSE
               CALL pw_grid_setup(my_cell%hmat, pw_grid, grid_span=grid_span, &
                                  cutoff=cutilev, &
                                  spherical=spherical, odd=odd, fft_usage=.TRUE., &
                                  ncommensurate=ncommensurate, icommensurate=igrid_level, &
                                  blocked=blocked_id, &
                                  rs_dims=distribution_layout, &
                                  iounit=iounit)
               old_pw_grid => pw_grid
            END IF
         ELSE
            CALL pw_grid_setup(my_cell%hmat, pw_grid, grid_span=grid_span, &
                               cutoff=cutilev, &
                               spherical=spherical, odd=odd, fft_usage=.TRUE., &
                               ncommensurate=ncommensurate, icommensurate=igrid_level, &
                               blocked=do_pw_grid_blocked_false, &
                               ref_grid=old_pw_grid, &
                               rs_dims=distribution_layout, &
                               iounit=iounit)
         END IF

         ! init pw_pools
         NULLIFY (pw_pools(igrid_level)%pool)
         CALL pw_pool_create(pw_pools(igrid_level)%pool, pw_grid=pw_grid)

         CALL pw_grid_release(pw_grid)

      END DO

      pw_env%pw_pools => pw_pools

      CALL cp_print_key_finished_output(iounit, logger, print_section, '')

      ! init auxbas_grid
      DO i = 1, ngrid_level
         IF (cutoff(i) == dft_control%qs_control%cutoff) pw_env%auxbas_grid = i
      END DO

      ! init xc_pool
      IF (ASSOCIATED(xc_super_ref_grid)) THEN
         CALL pw_pool_create(pw_env%xc_pw_pool, &
                             pw_grid=xc_super_ref_grid)
         CALL pw_grid_release(xc_super_ref_grid)
      ELSE
         pw_env%xc_pw_pool => pw_pools(pw_env%auxbas_grid)%pool
         CALL pw_pool_retain(pw_env%xc_pw_pool)
      END IF

      ! init vdw_pool
      set_vdw_pool = .FALSE.
      IF (ASSOCIATED(dispersion_env)) THEN
         IF (dispersion_env%type == xc_vdw_fun_nonloc) THEN
            IF (dispersion_env%pw_cutoff > 0._dp) set_vdw_pool = .TRUE.
         END IF
      END IF
      IF (set_vdw_pool) THEN
         IF(.NOT.(ASSOCIATED(old_pw_grid)))CALL cp__a("pw_env_methods.F",471)
         CALL pw_grid_create(vdw_grid, para_env%group)
         CALL pw_grid_setup(my_cell%hmat, vdw_grid, grid_span=grid_span, &
                            cutoff=dispersion_env%pw_cutoff, &
                            spherical=spherical, odd=odd, fft_usage=.TRUE., &
                            ncommensurate=ncommensurate, icommensurate=0, &
                            blocked=do_pw_grid_blocked_false, &
                            ref_grid=old_pw_grid, &
                            rs_dims=distribution_layout, &
                            iounit=iounit)
         CALL pw_pool_create(pw_env%vdw_pw_pool, pw_grid=vdw_grid)
         CALL pw_grid_release(vdw_grid)
      ELSE
         pw_env%vdw_pw_pool => pw_pools(pw_env%auxbas_grid)%pool
         CALL pw_pool_retain(pw_env%vdw_pw_pool)
      END IF

      ! complete init of the poisson_env
      IF (.NOT. ASSOCIATED(pw_env%poisson_env)) THEN
         CALL pw_poisson_create(pw_env%poisson_env)
      END IF
      poisson_section => section_vals_get_subs_vals(input, "DFT%POISSON")

      CALL pw_poisson_read_parameters(poisson_section, poisson_params)
      CALL pw_poisson_set(pw_env%poisson_env, cell_hmat=my_cell%hmat, pw_pools=pw_env%pw_pools, &
                          parameters=poisson_params, mt_super_ref_pw_grid=mt_super_ref_grid, &
                          dct_pw_grid=dct_pw_grid, use_level=pw_env%auxbas_grid)
      CALL pw_grid_release(mt_super_ref_grid)
      CALL pw_grid_release(dct_pw_grid)
!
! If reference cell is present, then use pw_grid_change to keep bounds constant...
! do not re-init the Gaussian grid level (fix the gridlevel on which the pgf should go.
!
      IF (use_ref_cell) THEN
         DO igrid_level = 1, SIZE(pw_pools)
            CALL pw_grid_change(cell%hmat, pw_pools(igrid_level)%pool%pw_grid)
         ENDDO
         IF (set_vdw_pool) CALL pw_grid_change(cell%hmat, pw_env%vdw_pw_pool%pw_grid)
         CALL pw_poisson_read_parameters(poisson_section, poisson_params)
         CALL pw_poisson_set(pw_env%poisson_env, cell_hmat=cell%hmat, pw_pools=pw_env%pw_pools, &
                             parameters=poisson_params, mt_super_ref_pw_grid=mt_super_ref_grid, &
                             dct_pw_grid=dct_pw_grid, use_level=pw_env%auxbas_grid)
      END IF

      IF ((poisson_params%ps_implicit_params%boundary_condition .EQ. MIXED_PERIODIC_BC) .OR. &
          (poisson_params%ps_implicit_params%boundary_condition .EQ. MIXED_BC)) THEN
         pw_env%poisson_env%parameters%dbc_params%do_dbc_cube = &
            BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                             "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE"), cp_p_file)
      END IF
      ! setup dct_pw_grid (an extended pw_grid) for Discrete Cosine Transformation (DCT)
      IF ((poisson_params%ps_implicit_params%boundary_condition .EQ. NEUMANN_BC) .OR. &
          (poisson_params%ps_implicit_params%boundary_condition .EQ. MIXED_BC)) THEN
         CALL setup_dct_pw_grids(pw_env%poisson_env%pw_pools(pw_env%poisson_env%pw_level)%pool%pw_grid, &
                                 my_cell%hmat, poisson_params%ps_implicit_params%neumann_directions, &
                                 pw_env%poisson_env%dct_pw_grid)
      END IF
      ! setup real space grid for finite difference derivatives of dielectric constant function
      IF (poisson_params%has_dielectric .AND. &
          ((poisson_params%dielectric_params%derivative_method .EQ. derivative_cd3) .OR. &
           (poisson_params%dielectric_params%derivative_method .EQ. derivative_cd5) .OR. &
           (poisson_params%dielectric_params%derivative_method .EQ. derivative_cd7))) THEN

         SELECT CASE (poisson_params%ps_implicit_params%boundary_condition)
         CASE (NEUMANN_BC, MIXED_BC)
            CALL setup_diel_rs_grid(pw_env%poisson_env%diel_rs_grid, &
                                    poisson_params%dielectric_params%derivative_method, input, &
                                    pw_env%poisson_env%dct_pw_grid)
         CASE (PERIODIC_BC, MIXED_PERIODIC_BC)
            CALL setup_diel_rs_grid(pw_env%poisson_env%diel_rs_grid, &
                                    poisson_params%dielectric_params%derivative_method, input, &
                                    pw_env%poisson_env%pw_pools(pw_env%poisson_env%pw_level)%pool%pw_grid)
         END SELECT

      END IF

!
!
!    determine the maximum radii for mapped gaussians, needed to
!    set up distributed rs grids
!
!

      ALLOCATE (radius(ngrid_level))

      CALL compute_max_radius(radius, pw_env, qs_env)

!
!
!    set up the rs_grids and the cubes, requires 'radius' to be set up correctly
!
!
      ALLOCATE (rs_descs(ngrid_level))

      ALLOCATE (rs_grids(ngrid_level))

      ALLOCATE (pw_env%cube_info(ngrid_level))
      higher_grid_layout = (/-1, -1, -1/)

      DO igrid_level = 1, ngrid_level
         pw_grid => pw_pools(igrid_level)%pool%pw_grid

         CALL init_d3_poly_module() ! a fairly arbitrary but sufficient spot to do this
         CALL init_cube_info(pw_env%cube_info(igrid_level), &
                             pw_grid%dr(:), pw_grid%dh(:, :), pw_grid%dh_inv(:, :), pw_grid%orthorhombic, &
                             radius(igrid_level))

         rs_grid_section => section_vals_get_subs_vals(input, "DFT%MGRID%RS_GRID")

         CALL init_input_type(input_settings, nsmax=2*MAX(1, return_cube_max_iradius(pw_env%cube_info(igrid_level)))+1, &
                              rs_grid_section=rs_grid_section, ilevel=igrid_level, &
                              higher_grid_layout=higher_grid_layout)

         NULLIFY (rs_descs(igrid_level)%rs_desc)
         CALL rs_grid_create_descriptor(rs_descs(igrid_level)%rs_desc, pw_grid, input_settings)

         IF (rs_descs(igrid_level)%rs_desc%distributed) higher_grid_layout = rs_descs(igrid_level)%rs_desc%group_dim

         NULLIFY (rs_grids(igrid_level)%rs_grid)
         CALL rs_grid_create(rs_grids(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
      END DO
      pw_env%rs_descs => rs_descs
      pw_env%rs_grids => rs_grids

      DEALLOCATE (radius)

      ! Initialise the lgrids which may be used by OpenMP threads in QS routines
      ! (but don't yet allocate the lgrid, in case we don't need it)
      CALL lgrid_create(pw_env%lgrid, pw_env%rs_descs)

      ! Print grid information

      logger => cp_get_default_logger()
      iounit = cp_print_key_unit_nr(logger, print_section, '', &
                                    extension='.Log')
      IF (iounit > 0) THEN
         SELECT CASE (poisson_params%solver)
         CASE (pw_poisson_periodic)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "PERIODIC"
         CASE (pw_poisson_analytic)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "ANALYTIC"
         CASE (pw_poisson_mt)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", ADJUSTR("Martyna-Tuckerman (MT)")
            WRITE (UNIT=iounit, FMT="(T2,A,T71,F10.3,/,T2,A,T71,F10.1)") &
               "POISSON| MT| Alpha", poisson_params%mt_alpha, &
               "POISSON| MT| Relative cutoff", poisson_params%mt_rel_cutoff
         CASE (pw_poisson_multipole)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "MULTIPOLE (Bloechl)"
         CASE (pw_poisson_wavelet)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "WAVELET"
            WRITE (UNIT=iounit, FMT="(T2,A,T71,I10)") &
               "POISSON| Wavelet| Scaling function", poisson_params%wavelet_scf_type
            SELECT CASE (poisson_params%wavelet_method)
            CASE (WAVELET0D)
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Periodicity", "NONE"
            CASE (WAVELET2D)
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Periodicity", "XZ"
            CASE (WAVELET3D)
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Periodicity", "XYZ"
            CASE DEFAULT
               CALL cp__b("pw_env_methods.F",639,"Invalid periodicity for wavelet solver selected")
            END SELECT
         CASE (pw_poisson_implicit)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "IMPLICIT (GENERALIZED)"
            boundary_condition = poisson_params%ps_implicit_params%boundary_condition
            SELECT CASE (boundary_condition)
            CASE (PERIODIC_BC)
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Boundary Condition", "PERIODIC"
            CASE (NEUMANN_BC, MIXED_BC)
               IF (boundary_condition .EQ. NEUMANN_BC) THEN
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                     "POISSON| Boundary Condition", "NEUMANN"
               ELSE
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                     "POISSON| Boundary Condition", "MIXED"
               END IF
               SELECT CASE (poisson_params%ps_implicit_params%neumann_directions)
               CASE (neumannXYZ)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "X, Y, Z"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "NONE"
               CASE (neumannXY)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "X, Y"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "Z"
               CASE (neumannXZ)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "X, Z"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "Y"
               CASE (neumannYZ)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "Y, Z"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "X"
               CASE (neumannX)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "X"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "Y, Z"
               CASE (neumannY)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "Y"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "X, Z"
               CASE (neumannZ)
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| homogeneous Neumann directions", "Z"
                  WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") "POISSON| periodic directions", "X, Y"
               CASE DEFAULT
                  CALL cp__b("pw_env_methods.F",680,"Invalid combination of Neumann and periodic conditions.")
               END SELECT
            CASE (MIXED_PERIODIC_BC)
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Boundary Condition", "PERIODIC & DIRICHLET"
            CASE DEFAULT
               CALL cp__b("pw_env_methods.F",686,"Invalid boundary conditions for the implicit (generalized) poisson solver.")
            END SELECT
            WRITE (UNIT=iounit, FMT="(T2,A,T71,I10)") &
               "POISSON| Maximum number of iterations", poisson_params%ps_implicit_params%max_iter
            WRITE (UNIT=iounit, FMT="(T2,A,T51,ES30.2)") &
               "POISSON| Convergence threshold", poisson_params%ps_implicit_params%tol
            IF (poisson_params%dielectric_params%dielec_functiontype .EQ. rho_dependent) THEN
               WRITE (UNIT=iounit, FMT="(T2,A,T51,F30.2)") &
                  "POISSON| Dielectric Constant", poisson_params%dielectric_params%eps0
            ELSE
               WRITE (UNIT=iounit, FMT="(T2,A,T31,F9.2)", ADVANCE='NO') &
                  "POISSON| Dielectric Constants", poisson_params%dielectric_params%eps0
               DO i = 1, poisson_params%dielectric_params%n_aa_cuboidal
                  WRITE (UNIT=iounit, FMT="(F9.2)", ADVANCE='NO') &
                     poisson_params%dielectric_params%aa_cuboidal_eps(i)
               END DO
               DO i = 1, poisson_params%dielectric_params%n_xaa_annular
                  WRITE (UNIT=iounit, FMT="(F9.2)", ADVANCE='NO') &
                     poisson_params%dielectric_params%xaa_annular_eps(i)
               END DO
               WRITE (UNIT=iounit, FMT='(A1,/)')
            END IF
            WRITE (UNIT=iounit, FMT="(T2,A,T51,ES30.2)") &
               "POISSON| Relaxation parameter", poisson_params%ps_implicit_params%omega
         CASE (pw_poisson_none)
            WRITE (UNIT=iounit, FMT="(/,T2,A,T51,A30)") &
               "POISSON| Solver", "NONE"
         CASE default
            CALL cp__b("pw_env_methods.F",714,"Invalid Poisson solver selected")
         END SELECT
         IF ((poisson_params%solver .NE. pw_poisson_wavelet) .AND. &
             (poisson_params%solver .NE. pw_poisson_implicit)) THEN
            IF (SUM(poisson_params%periodic(1:3)) == 0) THEN
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Periodicity", "NONE"
            ELSE
               string = ""
               IF (poisson_params%periodic(1) == 1) string = TRIM(string)//"X"
               IF (poisson_params%periodic(2) == 1) string = TRIM(string)//"Y"
               IF (poisson_params%periodic(3) == 1) string = TRIM(string)//"Z"
               WRITE (UNIT=iounit, FMT="(T2,A,T51,A30)") &
                  "POISSON| Periodicity", string
            END IF
         END IF
      END IF
      should_output = BTEST(cp_print_key_should_output(logger%iter_info, &
                                                       print_section, ''), cp_p_file)
      IF (should_output) THEN
         DO igrid_level = 1, ngrid_level
            CALL rs_grid_print(rs_grids(igrid_level)%rs_grid, iounit)
         END DO
      END IF
      CALL cp_print_key_finished_output(iounit, logger, print_section, "")

      CALL timestop(handle)

   END SUBROUTINE pw_env_rebuild

! **************************************************************************************************
!> \brief computes the maximum radius
!> \param radius ...
!> \param pw_env ...
!> \param qs_env ...
!> \par History
!>      10.2010 refactored [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE compute_max_radius(radius, pw_env, qs_env)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: radius
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_max_radius', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: safety_factor = 1.015_dp

      CHARACTER(LEN=default_string_length)               :: basis_type
      INTEGER :: handle, ibasis_set_type, igrid_level, igrid_zet0_s, ikind, ipgf, iset, ishell, &
         jkind, jpgf, jset, jshell, la, lb, lgrid_level, ngrid_level, nkind, nseta, nsetb
      INTEGER, DIMENSION(:), POINTER                     :: npgfa, npgfb, nshella, nshellb
      INTEGER, DIMENSION(:, :), POINTER                  :: lshella, lshellb
      REAL(KIND=dp)                                      :: alpha, core_charge, max_rpgf0_s, &
                                                            maxradius, zet0, zetp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zeta, zetb
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole

! a very small safety factor might be needed for roundoff issues
! e.g. radius being computed here as 12.998 (13) and 13.002 (14) during the collocation
! te latter can happen due to the lower precision in the computation of the radius in collocate
! parallel cost of rs_pw_transfer goes as safety_factor**3 so it is worthwhile keeping it tight

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, qs_kind_set, rho0_mpole)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      dft_control=dft_control)

      IF (dft_control%qs_control%gapw) THEN
         CALL get_qs_env(qs_env=qs_env, rho0_mpole=rho0_mpole)
         IF(.NOT.(ASSOCIATED(rho0_mpole)))CALL cp__a("pw_env_methods.F",789)
         CALL get_rho0_mpole(rho0_mpole=rho0_mpole, zet0_h=zet0, max_rpgf0_s=max_rpgf0_s)
      END IF

      ngrid_level = SIZE(radius)
      nkind = SIZE(qs_kind_set)

      ! Find the grid level suitable for rho0_soft
      IF (dft_control%qs_control%gapw) THEN
         igrid_zet0_s = gaussian_gridlevel(pw_env%gridlevel_info, 2.0_dp*zet0)
         rho0_mpole%igrid_zet0_s = igrid_zet0_s
      ELSE
         igrid_zet0_s = 0
      END IF

      ! try to predict the maximum radius of the gaussians to be mapped on the grid
      ! up to now, it is not yet very good
      maxradius = 0.0_dp
      DO igrid_level = 1, ngrid_level

         maxradius = 0.0_dp

         ! Take into account the radius of the soft compensation charge rho0_soft1
         IF (dft_control%qs_control%gapw) THEN
            IF (igrid_zet0_s == igrid_level) maxradius = MAX(maxradius, max_rpgf0_s)
         END IF

         ! ibasis set type loops over the needed kinds of basis sets. Right now this is
         ! the orb_basis_set and the aux_fit_basis_set (and others!)
         ! Seemingly others (e.g. gapw soft basis set) are not needed?
         ! This is really ugly code and prone to fail at some point
         ! Urgent rethinking and refactoring needed (24.07.2015)
         DO ibasis_set_type = 1, 5
            basis_type = translate_basis_type(ibasis_set_type)
            DO ikind = 1, nkind
               qs_kind => qs_kind_set(ikind)

               IF (dft_control%qs_control%dftb) CYCLE

               ! this is to be sure that the core charge is mapped ok
               ! right now, the core is mapped on the auxiliary basis,
               ! this should, at a give point be changed
               ! so that also for the core a multigrid is used
               CALL get_qs_kind(qs_kind_set(ikind), &
                                alpha_core_charge=alpha, ccore_charge=core_charge)

               IF (alpha > 0.0_dp .AND. core_charge .NE. 0.0_dp) THEN
                  maxradius = MAX(maxradius, exp_radius(0, alpha, &
                                                        dft_control%qs_control%eps_rho_rspace, core_charge))
                  ! forces
                  maxradius = MAX(maxradius, exp_radius(1, alpha, &
                                                        dft_control%qs_control%eps_rho_rspace, core_charge))
               ENDIF

               !
               ! now for the basis sets
               !
               IF (basis_type == "LRI" .AND. dft_control%qs_control%lri_optbas) CYCLE
               CALL get_qs_kind(qs_kind=qs_kind, &
                                basis_set=orb_basis_set, basis_type=basis_type)

               IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE

               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      npgf=npgfa, nset=nseta, zet=zeta, l=lshella, nshell=nshella)

               DO jkind = 1, nkind
                  qs_kind => qs_kind_set(jkind)

                  CALL get_qs_kind(qs_kind=qs_kind, &
                                   basis_set=orb_basis_set, basis_type=basis_type)

                  IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         npgf=npgfb, nset=nsetb, zet=zetb, l=lshellb, nshell=nshellb)

                  DO iset = 1, nseta
                  DO ipgf = 1, npgfa(iset)
                  DO ishell = 1, nshella(iset)
                     ! for calculate wavefunction
                     zetp = zeta(ipgf, iset)
                     la = lshella(ishell, iset)

                     lgrid_level = gaussian_gridlevel(pw_env%gridlevel_info, zetp)

                     IF (lgrid_level .EQ. igrid_level) THEN
                        maxradius = MAX(maxradius, exp_radius(la, zetp, &
                                                              dft_control%qs_control%eps_rho_rspace, 1.0_dp))
                     ENDIF
                     DO jset = 1, nsetb
                     DO jpgf = 1, npgfb(jset)
                     DO jshell = 1, nshellb(jset)
                        zetp = zeta(ipgf, iset)+zetb(jpgf, jset)
                        lb = lshellb(jshell, jset)+la
                        lgrid_level = gaussian_gridlevel(pw_env%gridlevel_info, zetp)
                        IF (lgrid_level .EQ. igrid_level) THEN
                           ! density (scale is at most 2)
                           maxradius = MAX(maxradius, exp_radius(lb, zetp, &
                                                                 dft_control%qs_control%eps_rho_rspace, 2.0_dp))

                           ! should do tau (if tau will be used)

                           ! potential
                           maxradius = MAX(maxradius, exp_radius(lb, zetp, &
                                                                 dft_control%qs_control%eps_gvg_rspace, 2.0_dp))
                           ! forces
                           maxradius = MAX(maxradius, exp_radius(lb+1, zetp, &
                                                                 dft_control%qs_control%eps_gvg_rspace, 2.0_dp))

                           ! should do stresses (if stresses will be used)
                        ENDIF
                     ENDDO
                     ENDDO
                     ENDDO
                  ENDDO
                  ENDDO
                  ENDDO
               ENDDO ! jkind
            ENDDO ! ikind
         ENDDO ! ibasis_set_type

         ! this is a bit of hack, but takes into account numerics and rounding
         maxradius = maxradius*safety_factor
         radius(igrid_level) = maxradius
      END DO

      CALL timestop(handle)

   END SUBROUTINE compute_max_radius

! **************************************************************************************************
!> \brief Initialize the super-reference grid for Tuckerman or xc
!> \param super_ref_pw_grid ...
!> \param mt_super_ref_pw_grid ...
!> \param xc_super_ref_pw_grid ...
!> \param cutilev ...
!> \param grid_span ...
!> \param spherical ...
!> \param cell_ref ...
!> \param para_env ...
!> \param input ...
!> \param my_ncommensurate ...
!> \param uf_grid ...
!> \param print_section ...
!> \author 03-2005 Teodoro Laino [teo]
!> \note
!>      move somewere else?
! **************************************************************************************************
   SUBROUTINE setup_super_ref_grid(super_ref_pw_grid, mt_super_ref_pw_grid, &
                                   xc_super_ref_pw_grid, cutilev, grid_span, spherical, &
                                   cell_ref, para_env, input, my_ncommensurate, uf_grid, print_section)
      TYPE(pw_grid_type), POINTER                        :: super_ref_pw_grid, mt_super_ref_pw_grid, &
                                                            xc_super_ref_pw_grid
      REAL(KIND=dp), INTENT(IN)                          :: cutilev
      INTEGER, INTENT(IN)                                :: grid_span
      LOGICAL, INTENT(in)                                :: spherical
      TYPE(cell_type), POINTER                           :: cell_ref
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input
      INTEGER, INTENT(IN)                                :: my_ncommensurate
      LOGICAL, INTENT(in)                                :: uf_grid
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_super_ref_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iounit, my_val, nn(3), no(3)
      LOGICAL                                            :: mt_s_grid
      REAL(KIND=dp)                                      :: mt_rel_cutoff, my_cutilev
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: poisson_section

      NULLIFY (poisson_section)
      IF(.NOT.(.NOT. ASSOCIATED(mt_super_ref_pw_grid)))CALL cp__a("pw_env_methods.F",962)
      IF(.NOT.(.NOT. ASSOCIATED(xc_super_ref_pw_grid)))CALL cp__a("pw_env_methods.F",963)
      IF(.NOT.(.NOT. ASSOCIATED(super_ref_pw_grid)))CALL cp__a("pw_env_methods.F",964)
      poisson_section => section_vals_get_subs_vals(input, "DFT%POISSON")
      CALL section_vals_val_get(poisson_section, "POISSON_SOLVER", i_val=my_val)
      !
      ! Check if grids will be the same... In this case we don't use a super-reference grid
      !
      mt_s_grid = .FALSE.
      IF (my_val == pw_poisson_mt) THEN
         CALL section_vals_val_get(poisson_section, "MT%REL_CUTOFF", &
                                   r_val=mt_rel_cutoff)
         IF (mt_rel_cutoff > 1._dp) mt_s_grid = .TRUE.
      END IF

      logger => cp_get_default_logger()
      iounit = cp_print_key_unit_nr(logger, print_section, "", &
                                    extension=".Log")

      IF (uf_grid) THEN
         CALL pw_grid_create(xc_super_ref_pw_grid, para_env%group)
         CALL pw_grid_setup(cell_ref%hmat, xc_super_ref_pw_grid, grid_span=grid_span, &
                            cutoff=4._dp*cutilev, spherical=spherical, odd=.FALSE., fft_usage=.TRUE., &
                            ncommensurate=my_ncommensurate, icommensurate=1, &
                            blocked=do_pw_grid_blocked_false, rs_dims=(/para_env%num_pe, 1/), &
                            iounit=iounit)
         super_ref_pw_grid => xc_super_ref_pw_grid
      END IF
      IF (mt_s_grid) THEN
         CALL pw_grid_create(mt_super_ref_pw_grid, para_env%group)

         IF (ASSOCIATED(xc_super_ref_pw_grid)) THEN
            CALL cp__b("pw_env_methods.F",994,"special grid for mt and fine xc grid not compatible")
         ELSE
            my_cutilev = cutilev*mt_rel_cutoff

            no = pw_grid_init_setup(cell_ref%hmat, cutoff=cutilev, spherical=spherical, &
                                    odd=.FALSE., fft_usage=.TRUE., ncommensurate=0, icommensurate=1)
            nn = pw_grid_init_setup(cell_ref%hmat, cutoff=my_cutilev, spherical=spherical, &
                                    odd=.FALSE., fft_usage=.TRUE., ncommensurate=0, icommensurate=1)

            ! bug appears for nn==no, also in old versions
            IF(.NOT.(ALL(nn > no)))CALL cp__a("pw_env_methods.F",1004)
            CALL pw_grid_setup(cell_ref%hmat, mt_super_ref_pw_grid, &
                               cutoff=my_cutilev, spherical=spherical, fft_usage=.TRUE., &
                               blocked=do_pw_grid_blocked_false, rs_dims=(/para_env%num_pe, 1/), &
                               iounit=iounit)
            super_ref_pw_grid => mt_super_ref_pw_grid
         END IF
      END IF
      CALL cp_print_key_finished_output(iounit, logger, print_section, &
                                        "")
   END SUBROUTINE setup_super_ref_grid

! **************************************************************************************************
!> \brief   sets up a real-space grid for finite difference derivative of dielectric
!>          constant function
!> \param diel_rs_grid real space grid to be created
!> \param method preferred finite difference derivative method
!> \param input input file
!> \param pw_grid plane-wave grid
!> \par History
!>       12.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE setup_diel_rs_grid(diel_rs_grid, method, input, pw_grid)

      TYPE(realspace_grid_type), POINTER                 :: diel_rs_grid
      INTEGER, INTENT(IN)                                :: method
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_diel_rs_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: border_points, handle
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_input_type)                    :: input_settings
      TYPE(section_vals_type), POINTER                   :: rs_grid_section

      CALL timeset(routineN, handle)

      NULLIFY (rs_desc)
      rs_grid_section => section_vals_get_subs_vals(input, "DFT%MGRID%RS_GRID")
      SELECT CASE (method)
      CASE (derivative_cd3)
         border_points = 1
      CASE (derivative_cd5)
         border_points = 2
      CASE (derivative_cd7)
         border_points = 3
      END SELECT
      CALL init_input_type(input_settings, 2*border_points+1, rs_grid_section, &
                           1, (/-1, -1, -1/))
      CALL rs_grid_create_descriptor(rs_desc, pw_grid, input_settings, &
                                     border_points=border_points)
      CALL rs_grid_create(diel_rs_grid, rs_desc)
!   CALL rs_grid_print(diel_rs_grid,6)
      CALL rs_grid_release_descriptor(rs_desc)

      CALL timestop(handle)

   END SUBROUTINE setup_diel_rs_grid

END MODULE pw_env_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief qs_environement methods that use many other modules
!> \par History
!>      09.2002 created [fawzi]
!>      - local atom distribution (25.06.2003,MK)
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_environment_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: dbcsr_distribution_obj
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_dist2d_to_dist
   USE cp_ddapc_types,                  ONLY: cp_ddapc_release
   USE cp_ddapc_util,                   ONLY: cp_ddapc_init
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_2d_types,           ONLY: distribution_2d_release,&
                                              distribution_2d_type
   USE distribution_methods,            ONLY: distribute_molecules_2d
   USE ewald_environment_types,         ONLY: ewald_environment_type
   USE ewald_pw_methods,                ONLY: ewald_pw_grid_update
   USE ewald_pw_types,                  ONLY: ewald_pw_type
   USE external_potential_types,        ONLY: get_potential,&
                                              gth_potential_type
   USE input_constants,                 ONLY: do_ppl_analytic,&
                                              do_ppl_grid,&
                                              kg_tnadd_embed
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_methods,                  ONLY: pw_env_create,&
                                              pw_env_rebuild
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_release,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_transfer
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE qs_charges_types,                ONLY: qs_charges_create,&
                                              qs_charges_release,&
                                              qs_charges_type
   USE qs_collocate_density,            ONLY: calculate_ppl_grid,&
                                              calculate_rho_core,&
                                              calculate_rho_nlcc
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_did_change,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_matrix_pools,                 ONLY: mpools_rebuild_fm_pools
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_outer_scf,                    ONLY: outer_loop_variables_count
   USE qs_rho0_types,                   ONLY: rho0_mpole_type
   USE qs_rho_methods,                  ONLY: qs_rho_rebuild
   USE qs_rho_types,                    ONLY: qs_rho_type
   USE qs_scf_types,                    ONLY: scf_env_did_change
   USE scf_control_types,               ONLY: scf_control_type
   USE task_list_methods,               ONLY: generate_qs_task_list
   USE task_list_types,                 ONLY: allocate_task_list,&
                                              deallocate_task_list,&
                                              task_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_environment_methods'

   PUBLIC :: qs_env_update_s_mstruct, &
             qs_env_rebuild_pw_env, &
             qs_env_setup
!***
CONTAINS

! **************************************************************************************************
!> \brief initializes various components of the qs_env, that need only
!>      atomic_kind_set, cell, dft_control, scf_control, c(i)%nmo,
!>      c(i)%nao, and particle_set to be initialized.
!>      The previous components of qs_env must be valid.
!>      Initializes pools, charges and pw_env.
!> \param qs_env the qs_env to set up
!> \param unit_nr ...
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_setup(qs_env, unit_nr)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(in)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n_ao_aux_fit, nhistory, &
                                                            nvariables
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: outer_scf_history
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind_set, atomic_kind_set, dft_control, scf_control, qs_charges, para_env, &
               distribution_2d, molecule_kind_set, molecule_set, particle_set, cell, &
               ks_env, blacs_env)

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control, &
                      molecule_kind_set=molecule_kind_set, &
                      molecule_set=molecule_set, &
                      particle_set=particle_set, &
                      scf_control=scf_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env, &
                      cell=cell, &
                      ks_env=ks_env)

      IF(.NOT.(ASSOCIATED(qs_kind_set)))CALL cp__a("qs_environment_methods.F",151)
      IF(.NOT.(ASSOCIATED(atomic_kind_set)))CALL cp__a("qs_environment_methods.F",152)
      IF(.NOT.(ASSOCIATED(dft_control)))CALL cp__a("qs_environment_methods.F",153)
      IF(.NOT.(ASSOCIATED(scf_control)))CALL cp__a("qs_environment_methods.F",154)
      ! allocate qs_charges
      CALL qs_charges_create(qs_charges, nspins=dft_control%nspins)
      CALL set_qs_env(qs_env, qs_charges=qs_charges)
      CALL qs_charges_release(qs_charges)

      ! outer scf setup
      IF (scf_control%outer_scf%have_scf) THEN
         nvariables = outer_loop_variables_count(scf_control)
         nhistory = scf_control%outer_scf%extrapolation_order
         ALLOCATE (outer_scf_history(nvariables, nhistory))
         CALL set_qs_env(qs_env, outer_scf_history=outer_scf_history)
         CALL set_qs_env(qs_env, outer_scf_ihistory=0)
      ENDIF

      ! set up pw_env
      CALL qs_env_rebuild_pw_env(qs_env)

      ! rebuilds fm_pools

      ! XXXX should get rid of the mpools
      IF (ASSOCIATED(qs_env%mos)) THEN
         CALL mpools_rebuild_fm_pools(qs_env%mpools, mos=qs_env%mos, &
                                      blacs_env=blacs_env, para_env=para_env)
      ENDIF

      ! If we use auxiliary density matrix methods rebuild fm_pools
      IF (dft_control%do_admm) THEN
         CALL get_qs_kind_set(qs_kind_set, nsgf=n_ao_aux_fit, &
                              basis_type="AUX_FIT")
         CALL mpools_rebuild_fm_pools(qs_env%mpools_aux_fit, mos=qs_env%mos_aux_fit, &
                                      blacs_env=blacs_env, para_env=para_env)
      END IF

      ! create 2d distribution

      CALL distribute_molecules_2d(cell=cell, &
                                   atomic_kind_set=atomic_kind_set, &
                                   qs_kind_set=qs_kind_set, &
                                   particle_set=particle_set, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   distribution_2d=distribution_2d, &
                                   blacs_env=blacs_env, &
                                   force_env_section=qs_env%input)

      ! and use it to create the dbcsr_dist, which should be the sole user of distribution_2d by now.
      ALLOCATE (dbcsr_dist)
      CALL cp_dbcsr_dist2d_to_dist(distribution_2d, dbcsr_dist, unit_nr)
      CALL set_ks_env(ks_env, dbcsr_dist=dbcsr_dist)

      ! also keep distribution_2d in qs_env
      CALL set_ks_env(ks_env, distribution_2d=distribution_2d)
      CALL distribution_2d_release(distribution_2d)

      CALL timestop(handle)

   END SUBROUTINE qs_env_setup

! **************************************************************************************************
!> \brief updates the s_mstruct to reflect the new overlap structure,
!>      and also updates rho_core distribution.
!>      Should be called after the atoms have moved and the new overlap
!>      has been calculated.
!> \param qs_env the environment to update
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_update_s_mstruct(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_update_s_mstruct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_ppl
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), POINTER                           :: rho_core, rho_nlcc, rho_nlcc_g, vppl

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_environment_methods.F",236)

      NULLIFY (dft_control)
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control)

      ! *** updates rho core ***
      NULLIFY (rho_core)
      CALL get_qs_env(qs_env, rho_core=rho_core)
      IF (dft_control%qs_control%gapw) THEN
         qs_env%qs_charges%total_rho_core_rspace = qs_env%local_rho_set%rhoz_tot
         IF (dft_control%qs_control%gapw_control%nopaw_as_gpw) THEN
            IF(.NOT.(ASSOCIATED(rho_core)))CALL cp__a("qs_environment_methods.F",248)
            CALL calculate_rho_core(rho_core, &
                                    qs_env%qs_charges%total_rho_core_rspace, qs_env, only_nopaw=.TRUE.)
         ELSE
            IF (ASSOCIATED(rho_core)) THEN
               CALL pw_release(rho_core%pw)
               DEALLOCATE (rho_core)
            ENDIF
         ENDIF
         ! force analytic ppl calculation
         dft_control%qs_control%do_ppl_method = do_ppl_analytic
      ELSE IF (dft_control%qs_control%semi_empirical) THEN
         !??
      ELSE IF (dft_control%qs_control%dftb) THEN
         !??
      ELSE IF (dft_control%qs_control%scptb) THEN
         !??
      ELSE
         IF(.NOT.(ASSOCIATED(rho_core)))CALL cp__a("qs_environment_methods.F",266)
         CALL calculate_rho_core(rho_core, &
                                 qs_env%qs_charges%total_rho_core_rspace, qs_env)
      END IF

      ! calculate local pseudopotential on grid
      do_ppl = dft_control%qs_control%do_ppl_method == do_ppl_grid
      IF (do_ppl) THEN
         NULLIFY (vppl)
         CALL get_qs_env(qs_env, vppl=vppl)
         IF(.NOT.(ASSOCIATED(vppl)))CALL cp__a("qs_environment_methods.F",276)
         CALL calculate_ppl_grid(vppl, qs_env)
      END IF

      ! compute the rho_nlcc
      NULLIFY (rho_nlcc, rho_nlcc_g)
      CALL get_qs_env(qs_env, rho_nlcc=rho_nlcc, rho_nlcc_g=rho_nlcc_g)
      IF (ASSOCIATED(rho_nlcc)) THEN
         CALL calculate_rho_nlcc(rho_nlcc, qs_env)
         CALL pw_transfer(rho_nlcc%pw, rho_nlcc_g%pw)
      ENDIF

      ! allocates and creates the task_list
      CALL qs_create_task_list(qs_env)

      ! *** environment for ddapc ***
      IF (ASSOCIATED(qs_env%cp_ddapc_env)) THEN
         CALL cp_ddapc_release(qs_env%cp_ddapc_env)
      END IF
      CALL cp_ddapc_init(qs_env)

      ! *** tell ks_env ***
      CALL qs_ks_did_change(qs_env%ks_env, s_mstruct_changed=.TRUE.)

      !   *** Updates rho structure ***
      CALL qs_env_rebuild_rho(qs_env=qs_env)

      ! *** tell scf_env ***
      IF (ASSOCIATED(qs_env%scf_env)) THEN
         CALL scf_env_did_change(qs_env%scf_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_env_update_s_mstruct

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE qs_create_task_list(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_create_task_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, isub
      LOGICAL                                            :: skip_load_balance_distributed, soft_valid
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(task_list_type), POINTER                      :: task_list

      CALL timeset(routineN, handle)
      NULLIFY (ks_env, dft_control)
      CALL get_qs_env(qs_env, ks_env=ks_env, dft_control=dft_control)

      skip_load_balance_distributed = dft_control%qs_control%skip_load_balance_distributed
      IF (.NOT. (dft_control%qs_control%semi_empirical .OR. dft_control%qs_control%dftb)) THEN
         ! generate task lists (non-soft)
         IF (.NOT. dft_control%qs_control%gapw) THEN
            CALL get_ks_env(ks_env, task_list=task_list)
            IF (.NOT. ASSOCIATED(task_list)) THEN
               CALL allocate_task_list(task_list)
               CALL set_ks_env(ks_env, task_list=task_list)
            ENDIF
            CALL generate_qs_task_list(ks_env, task_list, &
                                       reorder_rs_grid_ranks=.TRUE., soft_valid=.FALSE., &
                                       skip_load_balance_distributed=skip_load_balance_distributed)
         ENDIF
         ! generate the soft task list
         IF (dft_control%qs_control%gapw .OR. dft_control%qs_control%gapw_xc) THEN
            CALL get_ks_env(ks_env, task_list_soft=task_list)
            IF (.NOT. ASSOCIATED(task_list)) THEN
               CALL allocate_task_list(task_list)
               CALL set_ks_env(ks_env, task_list_soft=task_list)
            ENDIF
            CALL generate_qs_task_list(ks_env, task_list, &
                                       reorder_rs_grid_ranks=.TRUE., soft_valid=.TRUE., &
                                       skip_load_balance_distributed=skip_load_balance_distributed)
         ENDIF
      ENDIF

      IF (dft_control%do_admm) THEN
         ! generate the aux_fit task list
         CALL get_ks_env(ks_env, task_list_aux_fit=task_list)
         IF (.NOT. ASSOCIATED(task_list)) THEN
            CALL allocate_task_list(task_list)
            CALL set_ks_env(ks_env, task_list_aux_fit=task_list)
         ENDIF
         CALL get_ks_env(ks_env, sab_aux_fit=sab_orb)
         CALL generate_qs_task_list(ks_env, task_list, &
                                    reorder_rs_grid_ranks=.FALSE., soft_valid=.FALSE., basis_type="AUX_FIT", &
                                    skip_load_balance_distributed=skip_load_balance_distributed, &
                                    sab_orb_external=sab_orb)
      END IF

      IF (dft_control%qs_control%do_kg) THEN
         soft_valid = (dft_control%qs_control%gapw .OR. dft_control%qs_control%gapw_xc)

         IF (qs_env%kg_env%tnadd_method == kg_tnadd_embed) THEN

            IF (ASSOCIATED(qs_env%kg_env%subset)) THEN
               DO isub = 1, qs_env%kg_env%nsubsets
                  IF (ASSOCIATED(qs_env%kg_env%subset(isub)%task_list)) &
                     CALL deallocate_task_list(qs_env%kg_env%subset(isub)%task_list)
               END DO
            ELSE
               ALLOCATE (qs_env%kg_env%subset(qs_env%kg_env%nsubsets))
            END IF

            DO isub = 1, qs_env%kg_env%nsubsets

               CALL allocate_task_list(qs_env%kg_env%subset(isub)%task_list)

               ! generate the subset task list from the neighborlist
               CALL generate_qs_task_list(ks_env, qs_env%kg_env%subset(isub)%task_list, &
                                          reorder_rs_grid_ranks=.FALSE., soft_valid=soft_valid, &
                                          skip_load_balance_distributed=skip_load_balance_distributed, &
                                          sab_orb_external=qs_env%kg_env%subset(isub)%sab_orb)

            END DO

         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_create_task_list

! **************************************************************************************************
!> \brief rebuilds the pw_env in the given qs_env, allocating it if necessary
!> \param qs_env the qs_env whose pw_env has to be rebuilt
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_rebuild_pw_env(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_rebuild_pw_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: nlcc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(pw_env_type), POINTER                         :: new_pw_env
      TYPE(pw_p_type), POINTER                           :: external_vxc, rho0_gs, rho0_rs, &
                                                            rho_core, rho_nlcc, rho_nlcc_g, vee, &
                                                            vppl
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: v_hartree_rspace
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole

      CALL timeset(routineN, handle)
      ! rebuild pw_env
      NULLIFY (dft_control, cell, ks_env, v_hartree_rspace, auxbas_pw_pool)
      NULLIFY (rho0_mpole, rho0_gs, rho0_rs)
      NULLIFY (ewald_env, ewald_pw, new_pw_env, external_vxc, rho_core, rho_nlcc, rho_nlcc_g, vee, vppl)

      CALL get_qs_env(qs_env, ks_env=ks_env, pw_env=new_pw_env)
      IF (.NOT. ASSOCIATED(new_pw_env)) THEN
         CALL pw_env_create(new_pw_env)
         CALL set_ks_env(ks_env, pw_env=new_pw_env)
         CALL pw_env_release(new_pw_env)
      ENDIF

      CALL get_qs_env(qs_env, pw_env=new_pw_env, dft_control=dft_control, &
                      cell=cell)

      IF (ANY(new_pw_env%cell_hmat /= cell%hmat)) THEN
         ! only rebuild if nessecary
         new_pw_env%cell_hmat = cell%hmat
         CALL pw_env_rebuild(new_pw_env, qs_env=qs_env)

         ! reallocate rho_core
         CALL get_qs_env(qs_env, pw_env=new_pw_env, rho_core=rho_core)
         IF(.NOT.(ASSOCIATED(new_pw_env)))CALL cp__a("qs_environment_methods.F",459)
         IF (dft_control%qs_control%gapw) THEN
            IF (ASSOCIATED(rho_core)) THEN
               CALL pw_release(rho_core%pw)
               DEALLOCATE (rho_core)
            END IF
            IF (dft_control%qs_control%gapw_control%nopaw_as_gpw) THEN
               ALLOCATE (rho_core)
               CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
               CALL pw_pool_create_pw(auxbas_pw_pool, rho_core%pw, &
                                      use_data=COMPLEXDATA1D)
               rho_core%pw%in_space = RECIPROCALSPACE
               CALL set_ks_env(ks_env, rho_core=rho_core)
            END IF
            CALL get_qs_env(qs_env=qs_env, rho0_mpole=rho0_mpole)
            IF (ASSOCIATED(rho0_mpole)) THEN
               rho0_rs => rho0_mpole%rho0_s_rs
               rho0_gs => rho0_mpole%rho0_s_gs
               IF (ASSOCIATED(rho0_rs)) THEN
                  CALL pw_release(rho0_rs%pw)
                  DEALLOCATE (rho0_rs)
               END IF
               ALLOCATE (rho0_rs)
               CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
               CALL pw_pool_create_pw(auxbas_pw_pool, rho0_rs%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               rho0_mpole%rho0_s_rs => rho0_rs

               IF (ASSOCIATED(rho0_gs)) THEN
                  CALL pw_release(rho0_gs%pw)
                  DEALLOCATE (rho0_gs)
               END IF
               ALLOCATE (rho0_gs)
               CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
               CALL pw_pool_create_pw(auxbas_pw_pool, rho0_gs%pw, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               rho0_mpole%rho0_s_gs => rho0_gs
            END IF
         ELSE IF (dft_control%qs_control%semi_empirical) THEN
            IF (dft_control%qs_control%se_control%do_ewald .OR. &
                dft_control%qs_control%se_control%do_ewald_gks) THEN
               ! rebuild Ewald environment
               CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
               CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)
            END IF
         ELSE IF (dft_control%qs_control%dftb) THEN
            IF (dft_control%qs_control%dftb_control%do_ewald) THEN
               ! rebuild Ewald environment
               CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
               CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)
            END IF
         ELSE IF (dft_control%qs_control%scptb) THEN
            IF (dft_control%qs_control%scptb_control%do_ewald) THEN
               ! rebuild Ewald environment
               CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
               CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)
            END IF
         ELSE
            IF (ASSOCIATED(rho_core)) THEN
               CALL pw_release(rho_core%pw)
               DEALLOCATE (rho_core)
            ENDIF
            ALLOCATE (rho_core)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_core%pw, &
                                   use_data=COMPLEXDATA1D)
            rho_core%pw%in_space = RECIPROCALSPACE
            CALL set_ks_env(ks_env, rho_core=rho_core)
         END IF

         ! reallocate vppl (realspace grid of local pseudopotential
         IF (dft_control%qs_control%do_ppl_method == do_ppl_grid) THEN
            NULLIFY (vppl)
            CALL get_qs_env(qs_env, pw_env=new_pw_env, vppl=vppl)
            IF (ASSOCIATED(vppl)) THEN
               CALL pw_release(vppl%pw)
               DEALLOCATE (vppl)
            ENDIF
            ALLOCATE (vppl)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, vppl%pw, use_data=REALDATA3D)
            vppl%pw%in_space = REALSPACE
            CALL set_ks_env(ks_env, vppl=vppl)
         END IF

         ! reallocate rho_nlcc
         CALL has_nlcc(nlcc, qs_env)
         IF (nlcc) THEN
            ! right now, not working with gapw/gapw_xc, needs implementation in the GAPW XC routines
            IF (dft_control%qs_control%gapw_xc .OR. dft_control%qs_control%gapw) THEN
               CALL cp__b("qs_environment_methods.F",549,"Non linear core correction for GAPW not implemented")
            ENDIF
            ! the realspace version
            NULLIFY (rho_nlcc)
            CALL get_qs_env(qs_env, pw_env=new_pw_env, rho_nlcc=rho_nlcc)
            IF (ASSOCIATED(rho_nlcc)) THEN
               CALL pw_release(rho_nlcc%pw)
               DEALLOCATE (rho_nlcc)
            ENDIF
            ALLOCATE (rho_nlcc)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_nlcc%pw, use_data=REALDATA3D)
            rho_nlcc%pw%in_space = REALSPACE
            CALL set_ks_env(ks_env, rho_nlcc=rho_nlcc)
            ! the g-space version
            NULLIFY (rho_nlcc_g)
            CALL get_qs_env(qs_env, pw_env=new_pw_env, rho_nlcc_g=rho_nlcc_g)
            IF (ASSOCIATED(rho_nlcc_g)) THEN
               CALL pw_release(rho_nlcc_g%pw)
               DEALLOCATE (rho_nlcc_g)
            ENDIF
            ALLOCATE (rho_nlcc_g)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_nlcc_g%pw, use_data=COMPLEXDATA1D)
            rho_nlcc_g%pw%in_space = RECIPROCALSPACE
            CALL set_ks_env(ks_env, rho_nlcc_g=rho_nlcc_g)
         END IF

         ! reallocate vee: external electrostatic potential
         IF (dft_control%apply_external_potential) THEN
            NULLIFY (vee)
            CALL get_qs_env(qs_env, pw_env=new_pw_env, vee=vee)
            IF (ASSOCIATED(vee)) THEN
               CALL pw_release(vee%pw)
               DEALLOCATE (vee)
            ENDIF
            ALLOCATE (vee)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, vee%pw, use_data=REALDATA3D)
            vee%pw%in_space = REALSPACE
            CALL set_ks_env(ks_env, vee=vee)
            dft_control%eval_external_potential = .TRUE.
         END IF

         ! ZMP Reallocate external_vxc: external vxc potential
         IF (dft_control%apply_external_vxc) THEN
            NULLIFY (external_vxc)
            CALL get_qs_env(qs_env, pw_env=new_pw_env, external_vxc=external_vxc)
            IF (ASSOCIATED(external_vxc)) THEN
               CALL pw_release(external_vxc%pw)
               DEALLOCATE (external_vxc)
            ENDIF
            ALLOCATE (external_vxc)
            CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, external_vxc%pw, use_data=REALDATA3D)
            external_vxc%pw%in_space = REALSPACE
            CALL set_qs_env(qs_env, external_vxc=external_vxc)
            dft_control%read_external_vxc = .TRUE.
         END IF

         CALL get_ks_env(ks_env, v_hartree_rspace=v_hartree_rspace)
         IF (ASSOCIATED(v_hartree_rspace)) &
            CALL pw_release(v_hartree_rspace)
         CALL get_qs_env(qs_env, pw_env=new_pw_env)
         CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_rspace, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         CALL set_ks_env(ks_env, v_hartree_rspace=v_hartree_rspace)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_env_rebuild_pw_env

! **************************************************************************************************
!> \brief finds if a given qs run needs to use nlcc
!> \param nlcc ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE has_nlcc(nlcc, qs_env)

      LOGICAL                                            :: nlcc
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'has_nlcc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind
      LOGICAL                                            :: nlcc_present
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      nlcc = .FALSE.

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set)
      DO ikind = 1, SIZE(qs_kind_set)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, nlcc_present=nlcc_present)
         nlcc = nlcc .OR. nlcc_present
      ENDDO

   END SUBROUTINE has_nlcc

! **************************************************************************************************
!> \brief rebuilds the rho structure, making sure that everything is allocated
!>      and has the right size
!> \param qs_env the environment in which rho should be rebuilt
!> \param rebuild_ao if it is necessary to rebuild rho_ao. Defaults to true.
!> \param rebuild_grids if it in necessary to rebuild rho_r and rho_g.
!>        Defaults to false.
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      needs updated  pw pools, s_mstruct and h.
!>      The use of p to keep the structure of h (needed for the forces)
!>      is ugly and should be removed.
!>      If necessary rho is created from scratch.
! **************************************************************************************************
   SUBROUTINE qs_env_rebuild_rho(qs_env, rebuild_ao, rebuild_grids)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in), OPTIONAL                      :: rebuild_ao, rebuild_grids

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_rebuild_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_admm, gapw_xc
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho, rho_aux_fit, rho_aux_fit_buffer, &
                                                            rho_external, rho_xc

      NULLIFY (rho)
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      rho=rho, &
                      rho_xc=rho_xc, &
                      rho_aux_fit=rho_aux_fit, &
                      rho_aux_fit_buffer=rho_aux_fit_buffer, &
                      rho_external=rho_external)

      gapw_xc = dft_control%qs_control%gapw_xc
      do_admm = dft_control%do_admm
      CALL qs_rho_rebuild(rho, qs_env=qs_env, &
                          rebuild_ao=rebuild_ao, rebuild_grids=rebuild_grids)

      IF (gapw_xc) THEN
         CALL qs_rho_rebuild(rho_xc, qs_env=qs_env, &
                             rebuild_ao=rebuild_ao, rebuild_grids=rebuild_grids)
      END IF
      IF (do_admm) THEN
         CALL qs_rho_rebuild(rho_aux_fit, qs_env=qs_env, &
                             rebuild_ao=rebuild_ao, rebuild_grids=rebuild_grids, &
                             admm=.TRUE.)
         CALL qs_rho_rebuild(rho_aux_fit_buffer, qs_env=qs_env, &
                             rebuild_ao=rebuild_ao, rebuild_grids=rebuild_grids, &
                             admm=.TRUE.)
      END IF
! ZMP rebuilding external density
      IF (dft_control%apply_external_density) THEN
         CALL qs_rho_rebuild(rho_external, qs_env=qs_env, &
                             rebuild_grids=rebuild_grids)
         dft_control%read_external_density = .TRUE.
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_env_rebuild_rho

END MODULE qs_environment_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of non local dispersion functionals
!>  Some routines adapted from:
!>  Copyright (C) 2001-2009 Quantum ESPRESSO group
!>  Copyright (C) 2009 Brian Kolb, Timo Thonhauser - Wake Forest University
!>  This file is distributed under the terms of the
!>  GNU General Public License. See the file `License'
!>  in the root directory of the present distribution,
!>  or http://www.gnu.org/copyleft/gpl.txt .
!> \author JGH
! **************************************************************************************************
MODULE qs_dispersion_nonloc

   USE bibliography,                    ONLY: Dion2004,&
                                              Romanperez2009,&
                                              Sabatini2013,&
                                              cite_reference
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: vdw_nl_DRSLL,&
                                              vdw_nl_LMKLL,&
                                              vdw_nl_RVV10,&
                                              xc_vdw_fun_nonloc
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
   USE pw_grid_types,                   ONLY: HALFSPACE,&
                                              pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_derive,&
                                              pw_transfer
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   REAL(KIND=dp), PARAMETER :: epsr = 1.e-12_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dispersion_nonloc'

   PUBLIC :: qs_dispersion_nonloc_init, calculate_dispersion_nonloc

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param dispersion_env ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE qs_dispersion_nonloc_init(dispersion_env, para_env)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dispersion_nonloc_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: filename
      INTEGER                                            :: funit, handle, nqs, nr_points, q1_i, &
                                                            q2_i, vdw_type

      CALL timeset(routineN, handle)

      SELECT CASE (dispersion_env%nl_type)
      CASE DEFAULT
         CALL cp__b("qs_dispersion_nonloc.F",87,"Unknown vdW-DF functional")
      CASE (vdw_nl_DRSLL, vdw_nl_LMKLL)
         CALL cite_reference(Dion2004)
      CASE (vdw_nl_RVV10)
         CALL cite_reference(Sabatini2013)
      END SELECT
      CALL cite_reference(RomanPerez2009)

      vdw_type = dispersion_env%type
      SELECT CASE (vdw_type)
      CASE DEFAULT
         ! do nothing
      CASE (xc_vdw_fun_nonloc)
         ! setup information on non local functionals
         filename = dispersion_env%kernel_file_name
         IF (para_env%source == para_env%mepos) THEN
            ! Read the kernel information from file "filename"
            CALL open_file(file_name=filename, unit_number=funit, file_form="FORMATTED")
            READ (funit, *) nqs, nr_points
            READ (funit, *) dispersion_env%r_max
         END IF
         CALL mp_bcast(nqs, para_env%source, para_env%group)
         CALL mp_bcast(nr_points, para_env%source, para_env%group)
         CALL mp_bcast(dispersion_env%r_max, para_env%source, para_env%group)
         ALLOCATE (dispersion_env%q_mesh(nqs), dispersion_env%kernel(0:nr_points, nqs, nqs), &
                   dispersion_env%d2phi_dk2(0:nr_points, nqs, nqs))
         dispersion_env%nqs = nqs
         dispersion_env%nr_points = nr_points
         IF (para_env%source == para_env%mepos) THEN
            !! Read in the values of the q points used to generate this kernel
            READ (funit, "(1p, 4e23.14)") dispersion_env%q_mesh
            !! For each pair of q values, read in the function phi_q1_q2(k).
            !! That is, the fourier transformed kernel function assuming q1 and q2
            !! for all the values of r used.
            DO q1_i = 1, nqs
               DO q2_i = 1, q1_i
                  READ (funit, "(1p, 4e23.14)") dispersion_env%kernel(0:nr_points, q1_i, q2_i)
                  dispersion_env%kernel(0:nr_points, q2_i, q1_i) = dispersion_env%kernel(0:nr_points, q1_i, q2_i)
               END DO
            END DO
            !! Again, for each pair of q values (q1 and q2), read in the value
            !! of the second derivative of the above mentiond Fourier transformed
            !! kernel function phi_alpha_beta(k).  These are used for spline
            !! interpolation of the Fourier transformed kernel.
            DO q1_i = 1, nqs
               DO q2_i = 1, q1_i
                  READ (funit, "(1p, 4e23.14)") dispersion_env%d2phi_dk2(0:nr_points, q1_i, q2_i)
                  dispersion_env%d2phi_dk2(0:nr_points, q2_i, q1_i) = dispersion_env%d2phi_dk2(0:nr_points, q1_i, q2_i)
               END DO
            END DO
            CALL close_file(unit_number=funit)
         END IF
         CALL mp_bcast(dispersion_env%q_mesh, para_env%source, para_env%group)
         CALL mp_bcast(dispersion_env%kernel, para_env%source, para_env%group)
         CALL mp_bcast(dispersion_env%d2phi_dk2, para_env%source, para_env%group)
         ! 2nd derivates for interpolation
         ALLOCATE (dispersion_env%d2y_dx2(nqs, nqs))
         CALL initialize_spline_interpolation(dispersion_env%q_mesh, dispersion_env%d2y_dx2)
         !
         dispersion_env%q_cut = dispersion_env%q_mesh(nqs)
         dispersion_env%q_min = dispersion_env%q_mesh(1)
         dispersion_env%dk = 2.0_dp*pi/dispersion_env%r_max

      END SELECT

      CALL timestop(handle)

   END SUBROUTINE qs_dispersion_nonloc_init

! **************************************************************************************************
!> \brief Calculates the non-local vdW functional using the method of Soler
!>        For spin polarized cases we use E(a,b) = E(a+b), i.e. total density
!> \param vxc_rho ...
!> \param rho_r ...
!> \param rho_g ...
!> \param edispersion ...
!> \param dispersion_env ...
!> \param energy_only ...
!> \param pw_pool ...
!> \param xc_pw_pool ...
!> \param para_env ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE calculate_dispersion_nonloc(vxc_rho, rho_r, rho_g, edispersion, &
                                          dispersion_env, energy_only, pw_pool, xc_pw_pool, para_env, virial)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_rho, rho_r, rho_g
      REAL(KIND=dp), INTENT(OUT)                         :: edispersion
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      LOGICAL, INTENT(IN)                                :: energy_only
      TYPE(pw_pool_type), POINTER                        :: pw_pool, xc_pw_pool
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(virial_type), OPTIONAL, POINTER               :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_dispersion_nonloc', &
         routineP = moduleN//':'//routineN
      INTEGER, DIMENSION(3, 3), PARAMETER :: nd = RESHAPE((/1, 0, 0, 0, 1, 0, 0, 0, 1/), (/3, 3/))

      INTEGER                                            :: handle, i, i_grid, idir, ispin, n1, n2, &
                                                            n3, np, nspin
      LOGICAL                                            :: use_virial
      REAL(KIND=dp)                                      :: b_value, beta, Ec_nl, sumnp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dq0_dgradrho, dq0_drho, hpot, potential, &
                                                            q0, rho
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: drho, thetas
      TYPE(pw_grid_type), POINTER                        :: grid
      TYPE(pw_p_type), ALLOCATABLE, DIMENSION(:)         :: thetas_g
      TYPE(pw_p_type), ALLOCATABLE, DIMENSION(:, :)      :: drho_r
      TYPE(pw_type), POINTER                             :: tmp_g, tmp_r, vxc_g, vxc_r

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(rho_r)))CALL cp__a("qs_dispersion_nonloc.F",198)
      IF(.NOT.(ASSOCIATED(rho_g)))CALL cp__a("qs_dispersion_nonloc.F",199)
      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("qs_dispersion_nonloc.F",200)

      IF (PRESENT(virial)) THEN
         use_virial = virial%pv_calculate .AND. (.NOT. virial%pv_numer)
      ELSE
         use_virial = .FALSE.
      ENDIF
      IF (use_virial) THEN
         IF(.NOT.(.NOT. energy_only))CALL cp__a("qs_dispersion_nonloc.F",208)
      END IF
      IF (.NOT. energy_only) THEN
         IF(.NOT.(ASSOCIATED(vxc_rho)))CALL cp__a("qs_dispersion_nonloc.F",211)
      END IF

      b_value = dispersion_env%b_value
      beta = 0.03125_dp*(3.0_dp/(b_value**2.0_dp))**0.75_dp

      ! tempory arrays for FFT
      CALL pw_pool_create_pw(pw_pool, tmp_g, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(pw_pool, tmp_r, use_data=REALDATA3D, in_space=REALSPACE)

      ! get density derivatives
      nspin = SIZE(rho_r)
      ALLOCATE (drho_r(3, nspin))
      DO ispin = 1, nspin
         DO idir = 1, 3
            NULLIFY (drho_r(idir, ispin)%pw)
            CALL pw_pool_create_pw(pw_pool, drho_r(idir, ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_transfer(rho_g(ispin)%pw, tmp_g)
            CALL pw_derive(tmp_g, nd(:, idir))
            CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
         END DO
      END DO

      np = SIZE(tmp_r%cr3d)
      ALLOCATE (rho(np), drho(np, 3))
      rho = 0.0_dp
      drho = 0.0_dp
      DO ispin = 1, nspin
         IF(.NOT.(rho_r(ispin)%pw%in_use == REALDATA3D))CALL cp__a("qs_dispersion_nonloc.F",240)
         CALL pw_transfer(rho_g(ispin)%pw, tmp_g)
         CALL pw_transfer(tmp_g, tmp_r)
         rho(:) = rho(:)+RESHAPE(tmp_r%cr3d, (/np/))
         drho(:, 1) = drho(:, 1)+RESHAPE(drho_r(1, ispin)%pw%cr3d, (/np/))
         drho(:, 2) = drho(:, 2)+RESHAPE(drho_r(2, ispin)%pw%cr3d, (/np/))
         drho(:, 3) = drho(:, 3)+RESHAPE(drho_r(3, ispin)%pw%cr3d, (/np/))
      END DO
      !! ---------------------------------------------------------------------------------
      !! Find the value of q0 for all assigned grid points.  q is defined in equations
      !! 11 and 12 of DION and q0 is the saturated version of q defined in equation
      !! 5 of SOLER.  This routine also returns the derivatives of the q0s with respect
      !! to the charge-density and the gradient of the charge-density.  These are needed
      !! for the potential calculated below.
      !! ---------------------------------------------------------------------------------

      IF (energy_only) THEN
         ALLOCATE (q0(np))
         SELECT CASE (dispersion_env%nl_type)
         CASE DEFAULT
            CALL cp__b("qs_dispersion_nonloc.F",260,"Unknown vdW-DF functional")
         CASE (vdw_nl_DRSLL, vdw_nl_LMKLL)
            CALL get_q0_on_grid_eo_vdw(rho, drho, q0, dispersion_env)
         CASE (vdw_nl_RVV10)
            CALL get_q0_on_grid_eo_rvv10(rho, drho, q0, dispersion_env)
         END SELECT
      ELSE
         ALLOCATE (q0(np), dq0_drho(np), dq0_dgradrho(np))
         SELECT CASE (dispersion_env%nl_type)
         CASE DEFAULT
            CALL cp__b("qs_dispersion_nonloc.F",270,"Unknown vdW-DF functional")
         CASE (vdw_nl_DRSLL, vdw_nl_LMKLL)
            CALL get_q0_on_grid_vdw(rho, drho, q0, dq0_drho, dq0_dgradrho, dispersion_env)
         CASE (vdw_nl_RVV10)
            CALL get_q0_on_grid_rvv10(rho, drho, q0, dq0_drho, dq0_dgradrho, dispersion_env)
         END SELECT
      END IF

      !! ---------------------------------------------------------------------------------------------
      !! Here we allocate and calculate the theta functions appearing in equations 8-12 of SOLER.
      !! They are defined as rho*P_i(q0(rho, gradient_rho)) for vdW and vdW2 or
      !! constant*rho^(3/4)*P_i(q0(rho, gradient_rho)) for rVV10 where P_i is a polynomial that
      !! interpolates a Kronecker delta function at the point q_i (taken from the q_mesh) and q0 is
      !! the saturated version of q.
      !! q is defined in equations 11 and 12 of DION and the saturation proceedure is defined in equation 5
      !! of soler.  This is the biggest memory consumer in the method since the thetas array is
      !! (total # of FFT points)*Nqs complex numbers.  In a parallel run, each processor will hold the
      !! values of all the theta functions on just the points assigned to it.
      !! --------------------------------------------------------------------------------------------------
      !! thetas are stored in reciprocal space as  theta_i(k) because this is the way they are used later
      !! for the convolution (equation 11 of SOLER).
      !! --------------------------------------------------------------------------------------------------
      ALLOCATE (thetas(np, dispersion_env%nqs))
      !! Interpolate the P_i polynomials defined in equation 3 in SOLER for the particular
      !! q0 values we have.
      CALL spline_interpolation(dispersion_env%q_mesh, dispersion_env%d2y_dx2, q0, thetas)

      !! Form the thetas where theta is defined as rho*p_i(q0) for vdW and vdW2 or
      !! constant*rho^(3/4)*p_i(q0) for rVV10
      !! ------------------------------------------------------------------------------------
      SELECT CASE (dispersion_env%nl_type)
      CASE DEFAULT
         CALL cp__b("qs_dispersion_nonloc.F",302,"Unknown vdW-DF functional")
      CASE (vdw_nl_DRSLL, vdw_nl_LMKLL)
         DO i = 1, dispersion_env%nqs
            thetas(:, i) = thetas(:, i)*rho(:)
         END DO
      CASE (vdw_nl_RVV10)
         DO i_grid = 1, np
            IF (rho(i_grid) > epsr) THEN
               DO i = 1, dispersion_env%nqs
                  thetas(i_grid, i) = thetas(i_grid, i)*(1.0_dp/(3.0_dp*SQRT(pi)* &
                                                                 (b_value**(3.0_dp/2.0_dp))))*(rho(i_grid)/pi)**(3.0_dp/4.0_dp)
               END DO
            ELSE
               thetas(i_grid, :) = 0.0_dp
            END IF
         END DO
      END SELECT
      !! ------------------------------------------------------------------------------------
      !! Get thetas in reciprocal space.
      n1 = SIZE(tmp_r%cr3d, 1)
      n2 = SIZE(tmp_r%cr3d, 2)
      n3 = SIZE(tmp_r%cr3d, 3)
      ALLOCATE (thetas_g(dispersion_env%nqs))
      DO i = 1, dispersion_env%nqs
         tmp_r%cr3d = RESHAPE(thetas(:, i), (/n1, n2, n3/))
         NULLIFY (thetas_g(i)%pw)
         CALL pw_pool_create_pw(pw_pool, thetas_g(i)%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         CALL pw_transfer(tmp_r, thetas_g(i)%pw)
      END DO
      grid => thetas_g(1)%pw%pw_grid
      !! ---------------------------------------------------------------------------------------------
      !! Carry out the integration in equation 8 of SOLER.  This also turns the thetas array into the
      !! precursor to the u_i(k) array which is inverse fourier transformed to get the u_i(r) functions
      !! of SOLER equation 11.  Add the energy we find to the output variable etxc.
      !! --------------------------------------------------------------------------------------------------
      sumnp = np
      CALL mp_sum(sumnp, para_env%group)
      IF (use_virial) THEN
         ! calculates kernel contribution to stress
         CALL vdW_energy(thetas_g, dispersion_env, Ec_nl, energy_only, virial)
         SELECT CASE (dispersion_env%nl_type)
         CASE (vdw_nl_RVV10)
            Ec_nl = 0.5_dp*Ec_nl+beta*SUM(rho(:))*grid%vol/sumnp
         END SELECT
         ! calculates energy contribution to stress
         ! potential contribution to stress is calculated together with other potentials (Hxc)
         DO idir = 1, 3
            virial%pv_xc(idir, idir) = virial%pv_xc(idir, idir)+Ec_nl
         END DO
      ELSE
         CALL vdW_energy(thetas_g, dispersion_env, Ec_nl, energy_only)
         SELECT CASE (dispersion_env%nl_type)
         CASE (vdw_nl_RVV10)
            Ec_nl = 0.5_dp*Ec_nl+beta*SUM(rho(:))*grid%vol/sumnp
         END SELECT
      END IF
      CALL mp_sum(Ec_nl, para_env%group)
      edispersion = Ec_nl

      IF (.NOT. energy_only) THEN
         !! ----------------------------------------------------------------------------
         !! Inverse Fourier transform the u_i(k) to get the u_i(r) of SOLER equation 11.
         !!-----------------------------------------------------------------------------
         DO i = 1, dispersion_env%nqs
            CALL pw_transfer(thetas_g(i)%pw, tmp_r)
            thetas(:, i) = RESHAPE(tmp_r%cr3d, (/np/))
         END DO
         !! -------------------------------------------------------------------------
         !! Here we allocate the array to hold the potential. This is calculated via
         !! equation 10 of SOLER, using the u_i(r) calculated from equations 11 and
         !! 12 of SOLER.  Each processor allocates the array to be the size of the
         !! full grid  because, as can be seen in SOLER equation 9, processors need
         !! to access grid points outside their allocated regions.
         !! -------------------------------------------------------------------------
         ALLOCATE (potential(np), hpot(np))
         IF (use_virial) THEN
            ! calculates gradient contribution to stress
            grid => tmp_g%pw_grid
            CALL get_potential(q0, dq0_drho, dq0_dgradrho, rho, thetas, potential, hpot, &
                               dispersion_env, drho, grid%dvol, virial)
         ELSE
            CALL get_potential(q0, dq0_drho, dq0_dgradrho, rho, thetas, potential, hpot, &
                               dispersion_env)
         END IF
         SELECT CASE (dispersion_env%nl_type)
         CASE (vdw_nl_RVV10)
            potential(:) = 0.5_dp*potential(:)+beta
            hpot(:) = 0.5_dp*hpot(:)
         END SELECT
         CALL pw_pool_create_pw(pw_pool, vxc_r, use_data=REALDATA3D, in_space=REALSPACE)
         vxc_r%cr3d = RESHAPE(potential, (/n1, n2, n3/))
         DO idir = 1, 3
            tmp_r%cr3d = 0.0_dp
            DO ispin = 1, nspin
               tmp_r%cr3d = tmp_r%cr3d+RESHAPE(hpot, (/n1, n2, n3/))*drho_r(idir, ispin)%pw%cr3d
            END DO
            CALL pw_transfer(tmp_r, tmp_g)
            CALL pw_derive(tmp_g, nd(:, idir))
            CALL pw_transfer(tmp_g, tmp_r)
            CALL pw_axpy(tmp_r, vxc_r, -1._dp)
         END DO
         CALL pw_transfer(vxc_r, tmp_g)
         CALL pw_pool_give_back_pw(pw_pool, vxc_r)
         CALL pw_pool_create_pw(xc_pw_pool, vxc_r, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_pool_create_pw(xc_pw_pool, vxc_g, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         CALL pw_transfer(tmp_g, vxc_g)
         CALL pw_transfer(vxc_g, vxc_r)
         DO ispin = 1, nspin
            CALL pw_axpy(vxc_r, vxc_rho(ispin)%pw, 1._dp)
         END DO
         CALL pw_pool_give_back_pw(xc_pw_pool, vxc_r)
         CALL pw_pool_give_back_pw(xc_pw_pool, vxc_g)
      END IF

      DEALLOCATE (thetas)

      DO i = 1, dispersion_env%nqs
         CALL pw_pool_give_back_pw(pw_pool, thetas_g(i)%pw)
      END DO
      DO ispin = 1, nspin
         DO idir = 1, 3
            CALL pw_pool_give_back_pw(pw_pool, drho_r(idir, ispin)%pw)
         END DO
      END DO
      CALL pw_pool_give_back_pw(pw_pool, tmp_r)
      CALL pw_pool_give_back_pw(pw_pool, tmp_g)

      DEALLOCATE (rho, drho, drho_r, thetas_g)
      IF (energy_only) THEN
         DEALLOCATE (q0)
      ELSE
         DEALLOCATE (q0, dq0_drho, dq0_dgradrho)
      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_dispersion_nonloc

! **************************************************************************************************
!> \brief This routine carries out the integration of equation 8 of SOLER.  It returns the non-local
!> exchange-correlation energy and the u_alpha(k) arrays used to find the u_alpha(r) arrays via
!> equations 11 and 12 in SOLER.
!> energy contribution to stress is added in qs_force
!> \param thetas_g ...
!> \param dispersion_env ...
!> \param vdW_xc_energy ...
!> \param energy_only ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE vdW_energy(thetas_g, dispersion_env, vdW_xc_energy, energy_only, virial)
      TYPE(pw_p_type), ALLOCATABLE, DIMENSION(:)         :: thetas_g
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      REAL(KIND=dp), INTENT(OUT)                         :: vdW_xc_energy
      LOGICAL, INTENT(IN)                                :: energy_only
      TYPE(virial_type), OPTIONAL, POINTER               :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'vdW_energy', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: uu
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: theta, theta_g, thetam
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: u_vdw(:, :)
      INTEGER                                            :: handle, ig, igs, iq, l, m, nqs, q1_i, &
                                                            q2_i
      LOGICAL                                            :: use_virial
      REAL(KIND=dp)                                      :: g, g2, g2_last, g_multiplier, gm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dkernel_of_dk, kernel_of_k
      TYPE(pw_grid_type), POINTER                        :: grid

      CALL timeset(routineN, handle)
      nqs = dispersion_env%nqs
      ALLOCATE (kernel_of_k(nqs, nqs), dkernel_of_dk(nqs, nqs))
      ALLOCATE (theta(nqs), thetam(nqs), theta_g(nqs))

      IF (PRESENT(virial)) THEN
         use_virial = .TRUE.
      ELSE
         use_virial = .FALSE.
      END IF

      vdW_xc_energy = 0._dp
      g2_last = HUGE(0._dp)
      grid => thetas_g(1)%pw%pw_grid

      IF (.NOT. energy_only) THEN
         ALLOCATE (u_vdW(grid%ngpts_cut_local, nqs))
         u_vdW(:, :) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      END IF

      IF (grid%grid_span == HALFSPACE) THEN
         g_multiplier = 2._dp
      ELSE
         g_multiplier = 1._dp
      END IF
      igs = 0
      DO ig = 1, grid%ngpts_cut_local
         g2 = grid%gsq(ig)
         IF (ABS(g2-g2_last) > 1.e-10) THEN
            g2_last = g2
            igs = igs+1
            g = SQRT(g2)
            CALL interpolate_kernel(g, kernel_of_k, dispersion_env)
            IF (use_virial) CALL interpolate_dkernel_dk(g, dkernel_of_dk, dispersion_env)
         END IF
         DO iq = 1, nqs
            theta(iq) = thetas_g(iq)%pw%cc(ig)
         END DO
         DO q2_i = 1, Nqs
            uu = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
            DO q1_i = 1, Nqs
               uu = uu+kernel_of_k(q2_i, q1_i)*theta(q1_i)
            END DO
            IF (ig < grid%first_gne0) THEN
               vdW_xc_energy = vdW_xc_energy+REAL(uu*CONJG(theta(q2_i)), KIND=dp)
            ELSE
               vdW_xc_energy = vdW_xc_energy+G_multiplier*REAL(uu*CONJG(theta(q2_i)), KIND=dp)
            END IF
            IF (.NOT. energy_only) u_vdW(ig, q2_i) = uu
            IF (use_virial .AND. ig >= grid%first_gne0) THEN
               DO q1_i = 1, Nqs
                  gm = 0.5_dp*G_multiplier*grid%vol*dkernel_of_dk(q1_i, q2_i)*REAL(theta(q1_i)*CONJG(theta(q2_i)), KIND=dp)
                  SELECT CASE (dispersion_env%nl_type)
                  CASE (vdw_nl_RVV10)
                     gm = 0.5_dp*gm
                  END SELECT
                  DO l = 1, 3
                     DO m = 1, l
                        virial%pv_xc(l, m) = virial%pv_xc(l, m)-gm*(grid%g(l, ig)*grid%g(m, ig))/g
                        virial%pv_xc(m, l) = virial%pv_xc(l, m)
                     END DO
                  END DO
               END DO
            END IF
         END DO
      END DO
      vdW_xc_energy = vdW_xc_energy*grid%vol*0.5_dp

      IF (.NOT. energy_only) THEN
         DO iq = 1, nqs
            DO ig = 1, grid%ngpts_cut_local
               thetas_g(iq)%pw%cc(ig) = u_vdW(ig, iq)
            END DO
         END DO
         DEALLOCATE (u_vdW)
      END IF
      DEALLOCATE (theta, thetam, theta_g)
      DEALLOCATE (kernel_of_k, dkernel_of_dk)

      CALL timestop(handle)

   END SUBROUTINE vdW_energy

! **************************************************************************************************
!! This routine finds the non-local correlation contribution to the potential
!! (i.e. the derivative of the non-local piece of the energy with respect to
!! density) given in SOLER equation 10.  The u_alpha(k) functions were found
!! while calculating the energy. They are passed in as the matrix u_vdW.
!! Most of the required derivatives were calculated in the "get_q0_on_grid"
!! routine, but the derivative of the interpolation polynomials, P_alpha(q),
!! (SOLER equation 3) with respect to q is interpolated here, along with the
!! polynomials themselves.
! **************************************************************************************************
!> \brief ...
!> \param q0 ...
!> \param dq0_drho ...
!> \param dq0_dgradrho ...
!> \param total_rho ...
!> \param u_vdW ...
!> \param potential ...
!> \param h_prefactor ...
!> \param dispersion_env ...
!> \param drho ...
!> \param dvol ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE get_potential(q0, dq0_drho, dq0_dgradrho, total_rho, u_vdW, potential, h_prefactor, &
                            dispersion_env, drho, dvol, virial)

      REAL(dp), DIMENSION(:), INTENT(in)                 :: q0, dq0_drho, dq0_dgradrho, total_rho
      REAL(dp), DIMENSION(:, :), INTENT(in)              :: u_vdW
      REAL(dp), DIMENSION(:), INTENT(inout)              :: potential, h_prefactor
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      REAL(dp), DIMENSION(:, :), OPTIONAL                :: drho
      REAL(dp), INTENT(IN), OPTIONAL                     :: dvol
      TYPE(virial_type), OPTIONAL, POINTER               :: virial

      CHARACTER(len=*), PARAMETER :: routineN = 'get_potential', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_grid, l, m, nqs, P_i, q, q_hi, &
                                                            q_low
      LOGICAL                                            :: use_virial
      REAL(dp)                                           :: a, b, b_value, c, const, d, dP_dq0, dq, &
                                                            e, f, P, prefactor
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: y
      REAL(dp), DIMENSION(:), POINTER                    :: q_mesh
      REAL(dp), DIMENSION(:, :), POINTER                 :: d2y_dx2

      CALL timeset(routineN, handle)

      IF (PRESENT(virial)) THEN
         use_virial = .TRUE.
         IF(.NOT.(PRESENT(drho)))CALL cp__a("qs_dispersion_nonloc.F",602)
         IF(.NOT.(PRESENT(dvol)))CALL cp__a("qs_dispersion_nonloc.F",603)
      ELSE
         use_virial = .FALSE.
      END IF

      b_value = dispersion_env%b_value
      const = 1.0_dp/(3.0_dp*b_value**(3.0_dp/2.0_dp)*pi**(5.0_dp/4.0_dp))
      potential = 0.0_dp
      h_prefactor = 0.0_dp

      d2y_dx2 => dispersion_env%d2y_dx2
      q_mesh => dispersion_env%q_mesh
      nqs = dispersion_env%nqs
      ALLOCATE (y(nqs))

      DO i_grid = 1, SIZE(u_vdw, 1)
         q_low = 1
         q_hi = Nqs
         ! Figure out which bin our value of q0 is in in the q_mesh
         DO WHILE ((q_hi-q_low) > 1)
            q = INT((q_hi+q_low)/2)
            IF (q_mesh(q) > q0(i_grid)) THEN
               q_hi = q
            ELSE
               q_low = q
            END IF
         END DO
         IF (q_hi == q_low) CALL cp__b("qs_dispersion_nonloc.F",630,"get_potential:  qhi == qlow")

         dq = q_mesh(q_hi)-q_mesh(q_low)

         a = (q_mesh(q_hi)-q0(i_grid))/dq
         b = (q0(i_grid)-q_mesh(q_low))/dq
         c = (a**3-a)*dq**2/6.0_dp
         d = (b**3-b)*dq**2/6.0_dp
         e = (3.0_dp*a**2-1.0_dp)*dq/6.0_dp
         f = (3.0_dp*b**2-1.0_dp)*dq/6.0_dp

         DO P_i = 1, Nqs
            y = 0.0_dp
            y(P_i) = 1.0_dp
            dP_dq0 = (y(q_hi)-y(q_low))/dq-e*d2y_dx2(P_i, q_low)+f*d2y_dx2(P_i, q_hi)
            P = a*y(q_low)+b*y(q_hi)+c*d2y_dx2(P_i, q_low)+d*d2y_dx2(P_i, q_hi)
            !! The first term in equation 13 of SOLER
            SELECT CASE (dispersion_env%nl_type)
            CASE DEFAULT
               CALL cp__b("qs_dispersion_nonloc.F",649,"Unknown vdW-DF functional")
            CASE (vdw_nl_DRSLL, vdw_nl_LMKLL)
               potential(i_grid) = potential(i_grid)+u_vdW(i_grid, P_i)*(P+dP_dq0*dq0_drho(i_grid))
               prefactor = u_vdW(i_grid, P_i)*dP_dq0*dq0_dgradrho(i_grid)
            CASE (vdw_nl_RVV10)
               IF (total_rho(i_grid) > epsr) THEN
                  potential(i_grid) = potential(i_grid)+u_vdW(i_grid, P_i)* &
                                      (const*(3.0_dp/4.0_dp)/(total_rho(i_grid)**(1.0_dp/4.0_dp))*P+ &
                                       const*total_rho(i_grid)**(3.0_dp/4.0_dp)*dP_dq0*dq0_drho(i_grid))
                  prefactor = u_vdW(i_grid, P_i)*const*total_rho(i_grid)**(3.0_dp/4.0_dp)* &
                              dP_dq0*dq0_dgradrho(i_grid)
               ELSE
                  prefactor = 0.0_dp
               ENDIF
            END SELECT
            IF (q0(i_grid) .NE. q_mesh(Nqs)) THEN
               h_prefactor(i_grid) = h_prefactor(i_grid)+prefactor
            END IF
            IF (use_virial .AND. ABS(prefactor) > 0.0_dp) THEN
               SELECT CASE (dispersion_env%nl_type)
               CASE DEFAULT
                  ! do nothing
               CASE (vdw_nl_RVV10)
                  prefactor = 0.5_dp*prefactor
               END SELECT
               prefactor = prefactor*dvol
               DO l = 1, 3
                  DO m = 1, l
                     virial%pv_xc(l, m) = virial%pv_xc(l, m)-prefactor*drho(i_grid, l)*drho(i_grid, m)
                     virial%pv_xc(m, l) = virial%pv_xc(l, m)
                  ENDDO
               ENDDO
            END IF
         END DO
      END DO
      DEALLOCATE (y)

      CALL timestop(handle)

   END SUBROUTINE get_potential

   !! This routine first calculates the q value defined in (DION equations 11 and 12), then
   !! saturates it according to (SOLER equation 5).
! **************************************************************************************************
!> \brief ...
!> \param total_rho ...
!> \param gradient_rho ...
!> \param q0 ...
!> \param dq0_drho ...
!> \param dq0_dgradrho ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE get_q0_on_grid_vdw(total_rho, gradient_rho, q0, dq0_drho, dq0_dgradrho, dispersion_env)
      !!
      !! more specifically it calculates the following
      !!
      !!     q0(ir) = q0 as defined above
      !!     dq0_drho(ir) = total_rho * d q0 /d rho
      !!     dq0_dgradrho = total_rho / |gradient_rho| * d q0 / d |gradient_rho|
      !!
      REAL(dp), INTENT(IN)                               :: total_rho(:), gradient_rho(:, :)
      REAL(dp), INTENT(OUT)                              :: q0(:), dq0_drho(:), dq0_dgradrho(:)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_q0_on_grid_vdw', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: m_cut = 12
      REAL(dp), PARAMETER :: LDA_A = 0.031091_dp, LDA_a1 = 0.2137_dp, LDA_b1 = 7.5957_dp, &
         LDA_b2 = 3.5876_dp, LDA_b3 = 1.6382_dp, LDA_b4 = 0.49294_dp

      INTEGER                                            :: i_grid, index
      REAL(dp)                                           :: dq0_dq, exponent, gradient_correction, &
                                                            kF, LDA_1, LDA_2, q, q_cut, q_min, &
                                                            r_s, sqrt_r_s, Z_ab

      q_cut = dispersion_env%q_cut
      q_min = dispersion_env%q_min
      SELECT CASE (dispersion_env%nl_type)
      CASE DEFAULT
         CALL cp__b("qs_dispersion_nonloc.F",728,"Unknown vdW-DF functional")
      CASE (vdw_nl_DRSLL)
         Z_ab = -0.8491_dp
      CASE (vdw_nl_LMKLL)
         Z_ab = -1.887_dp
      END SELECT

      ! initialize q0-related arrays ...
      q0(:) = q_cut
      dq0_drho(:) = 0.0_dp
      dq0_dgradrho(:) = 0.0_dp

      DO i_grid = 1, SIZE(total_rho)

         !! This prevents numerical problems.  If the charge density is negative (an
         !! unphysical situation), we simply treat it as very small.  In that case,
         !! q0 will be very large and will be saturated.  For a saturated q0 the derivative
         !! dq0_dq will be 0 so we set q0 = q_cut and dq0_drho = dq0_dgradrho = 0 and go on
         !! to the next point.
         !! ------------------------------------------------------------------------------------
         IF (total_rho(i_grid) < epsr) CYCLE
         !! ------------------------------------------------------------------------------------
         !! Calculate some intermediate values needed to find q
         !! ------------------------------------------------------------------------------------
         kF = (3.0_dp*pi*pi*total_rho(i_grid))**(1.0_dp/3.0_dp)
         r_s = (3.0_dp/(4.0_dp*pi*total_rho(i_grid)))**(1.0_dp/3.0_dp)
         sqrt_r_s = SQRT(r_s)

         gradient_correction = -Z_ab/(36.0_dp*kF*total_rho(i_grid)**2) &
                               *(gradient_rho(i_grid, 1)**2+gradient_rho(i_grid, 2)**2+gradient_rho(i_grid, 3)**2)

         LDA_1 = 8.0_dp*pi/3.0_dp*(LDA_A*(1.0_dp+LDA_a1*r_s))
         LDA_2 = 2.0_dp*LDA_A*(LDA_b1*sqrt_r_s+LDA_b2*r_s+LDA_b3*r_s*sqrt_r_s+LDA_b4*r_s*r_s)
         !! ---------------------------------------------------------------
         !! This is the q value defined in equations 11 and 12 of DION
         !! ---------------------------------------------------------------
         q = kF+LDA_1*LOG(1.0_dp+1.0_dp/LDA_2)+gradient_correction
         !! ---------------------------------------------------------------
         !! Here, we calculate q0 by saturating q according to equation 5 of SOLER.  Also, we find
         !! the derivative dq0_dq needed for the derivatives dq0_drho and dq0_dgradrh0 discussed below.
         !! ---------------------------------------------------------------------------------------
         exponent = 0.0_dp
         dq0_dq = 0.0_dp
         DO index = 1, m_cut
            exponent = exponent+((q/q_cut)**index)/index
            dq0_dq = dq0_dq+((q/q_cut)**(index-1))
         END DO
         q0(i_grid) = q_cut*(1.0_dp-EXP(-exponent))
         dq0_dq = dq0_dq*EXP(-exponent)
         !! ---------------------------------------------------------------------------------------
         !! This is to handle a case with q0 too small.  We simply set it to the smallest q value in
         !! out q_mesh.  Hopefully this doesn't get used often (ever)
         !! ---------------------------------------------------------------------------------------
         IF (q0(i_grid) < q_min) THEN
            q0(i_grid) = q_min
         END IF
         !! ---------------------------------------------------------------------------------------
         !! Here we find derivatives.  These are actually the density times the derivative of q0 with respect
         !! to rho and gradient_rho.  The density factor comes in since we are really differentiating
         !! theta = (rho)*P(q0) with respect to density (or its gradient) which will be
         !! dtheta_drho = P(q0) + dP_dq0 * [rho * dq0_dq * dq_drho]   and
         !! dtheta_dgradient_rho =  dP_dq0  * [rho * dq0_dq * dq_dgradient_rho]
         !! The parts in square brackets are what is calculated here.  The dP_dq0 term will be interpolated
         !! later.  There should actually be a factor of the magnitude of the gradient in the gradient_rho derivative
         !! but that cancels out when we differentiate the magnitude of the gradient with respect to a particular
         !! component.
         !! ------------------------------------------------------------------------------------------------

         dq0_drho(i_grid) = dq0_dq*(kF/3.0_dp-7.0_dp/3.0_dp*gradient_correction &
                                    -8.0_dp*pi/9.0_dp*LDA_A*LDA_a1*r_s*LOG(1.0_dp+1.0_dp/LDA_2) &
                                    +LDA_1/(LDA_2*(1.0_dp+LDA_2)) &
                                    *(2.0_dp*LDA_A*(LDA_b1/6.0_dp*sqrt_r_s+LDA_b2/3.0_dp*r_s+LDA_b3/2.0_dp*r_s*sqrt_r_s &
                                                    +2.0_dp*LDA_b4/3.0_dp*r_s**2)))

         dq0_dgradrho(i_grid) = total_rho(i_grid)*dq0_dq*2.0_dp*(-Z_ab)/(36.0_dp*kF*total_rho(i_grid)**2)

      END DO

   END SUBROUTINE get_q0_on_grid_vdw

! **************************************************************************************************
!> \brief ...
!> \param total_rho ...
!> \param gradient_rho ...
!> \param q0 ...
!> \param dq0_drho ...
!> \param dq0_dgradrho ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE get_q0_on_grid_rvv10(total_rho, gradient_rho, q0, dq0_drho, dq0_dgradrho, dispersion_env)
      !!
      !! more specifically it calculates the following
      !!
      !!     q0(ir) = q0 as defined above
      !!     dq0_drho(ir) = total_rho * d q0 /d rho
      !!     dq0_dgradrho = total_rho / |gradient_rho| * d q0 / d |gradient_rho|
      !!
      REAL(dp), INTENT(IN)                               :: total_rho(:), gradient_rho(:, :)
      REAL(dp), INTENT(OUT)                              :: q0(:), dq0_drho(:), dq0_dgradrho(:)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      INTEGER, PARAMETER                                 :: m_cut = 12

      INTEGER                                            :: i_grid, index
      REAL(dp)                                           :: b_value, C_value, dk_dn, dq0_dq, dw0_dn, &
                                                            exponent, gmod2, k, mod_grad, q, &
                                                            q_cut, q_min, w0, wg2, wp2

      q_cut = dispersion_env%q_cut
      q_min = dispersion_env%q_min
      b_value = dispersion_env%b_value
      C_value = dispersion_env%c_value

      ! initialize q0-related arrays ...
      q0(:) = q_cut
      dq0_drho(:) = 0.0_dp
      dq0_dgradrho(:) = 0.0_dp

      DO i_grid = 1, SIZE(total_rho)

         gmod2 = gradient_rho(i_grid, 1)**2+gradient_rho(i_grid, 2)**2+gradient_rho(i_grid, 3)**2

         !if (total_rho(i_grid) > epsr .and. gmod2 > epsr) cycle
         IF (total_rho(i_grid) > epsr) THEN

            !! Calculate some intermediate values needed to find q
            !! ------------------------------------------------------------------------------------
            mod_grad = SQRT(gmod2)

            wp2 = 16.0_dp*pi*total_rho(i_grid)
            wg2 = 4_dp*C_value*(mod_grad/total_rho(i_grid))**4

            k = b_value*3.0_dp*pi*((total_rho(i_grid)/(9.0_dp*pi))**(1.0_dp/6.0_dp))
            w0 = SQRT(wg2+wp2/3.0_dp)

            q = w0/k

            !! Here, we calculate q0 by saturating q according
            !! ---------------------------------------------------------------------------------------

            exponent = 0.0_dp
            dq0_dq = 0.0_dp

            DO index = 1, m_cut

               exponent = exponent+((q/q_cut)**index)/index
               dq0_dq = dq0_dq+((q/q_cut)**(index-1))

            END DO

            q0(i_grid) = q_cut*(1.0_dp-EXP(-exponent))
            dq0_dq = dq0_dq*EXP(-exponent)

            !! ---------------------------------------------------------------------------------------

            IF (q0(i_grid) < q_min) THEN
               q0(i_grid) = q_min
            END IF

            !!---------------------------------Final values---------------------------------

            dw0_dn = 1.0_dp/(2.0_dp*w0)*(16.0_dp/3.0_dp*pi-4.0_dp*wg2/total_rho(i_grid))
            dk_dn = k/(6.0_dp*total_rho(i_grid))

            dq0_drho(i_grid) = dq0_dq*1.0_dp/(k**2.0)*(dw0_dn*k-dk_dn*w0)
            dq0_dgradrho(i_grid) = dq0_dq*1.0_dp/(2.0_dp*k*w0)*4.0_dp*wg2/(mod_grad**2)

         ENDIF

      END DO

   END SUBROUTINE get_q0_on_grid_rvv10

! **************************************************************************************************
!> \brief ...
!> \param total_rho ...
!> \param gradient_rho ...
!> \param q0 ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE get_q0_on_grid_eo_vdw(total_rho, gradient_rho, q0, dispersion_env)

      REAL(dp), INTENT(IN)                               :: total_rho(:), gradient_rho(:, :)
      REAL(dp), INTENT(OUT)                              :: q0(:)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_q0_on_grid_eo_vdw', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: m_cut = 12
      REAL(dp), PARAMETER :: LDA_A = 0.031091_dp, LDA_a1 = 0.2137_dp, LDA_b1 = 7.5957_dp, &
         LDA_b2 = 3.5876_dp, LDA_b3 = 1.6382_dp, LDA_b4 = 0.49294_dp

      INTEGER                                            :: i_grid, index
      REAL(dp)                                           :: exponent, gradient_correction, kF, &
                                                            LDA_1, LDA_2, q, q_cut, q_min, r_s, &
                                                            sqrt_r_s, Z_ab

      q_cut = dispersion_env%q_cut
      q_min = dispersion_env%q_min
      SELECT CASE (dispersion_env%nl_type)
      CASE DEFAULT
         CALL cp__b("qs_dispersion_nonloc.F",929,"Unknown vdW-DF functional")
      CASE (vdw_nl_DRSLL)
         Z_ab = -0.8491_dp
      CASE (vdw_nl_LMKLL)
         Z_ab = -1.887_dp
      END SELECT

      ! initialize q0-related arrays ...
      q0(:) = q_cut
      DO i_grid = 1, SIZE(total_rho)
         !! This prevents numerical problems.  If the charge density is negative (an
         !! unphysical situation), we simply treat it as very small.  In that case,
         !! q0 will be very large and will be saturated.  For a saturated q0 the derivative
         !! dq0_dq will be 0 so we set q0 = q_cut and dq0_drho = dq0_dgradrho = 0 and go on
         !! to the next point.
         !! ------------------------------------------------------------------------------------
         IF (total_rho(i_grid) < epsr) CYCLE
         !! ------------------------------------------------------------------------------------
         !! Calculate some intermediate values needed to find q
         !! ------------------------------------------------------------------------------------
         kF = (3.0_dp*pi*pi*total_rho(i_grid))**(1.0_dp/3.0_dp)
         r_s = (3.0_dp/(4.0_dp*pi*total_rho(i_grid)))**(1.0_dp/3.0_dp)
         sqrt_r_s = SQRT(r_s)

         gradient_correction = -Z_ab/(36.0_dp*kF*total_rho(i_grid)**2) &
                               *(gradient_rho(i_grid, 1)**2+gradient_rho(i_grid, 2)**2+gradient_rho(i_grid, 3)**2)

         LDA_1 = 8.0_dp*pi/3.0_dp*(LDA_A*(1.0_dp+LDA_a1*r_s))
         LDA_2 = 2.0_dp*LDA_A*(LDA_b1*sqrt_r_s+LDA_b2*r_s+LDA_b3*r_s*sqrt_r_s+LDA_b4*r_s*r_s)
         !! ------------------------------------------------------------------------------------
         !! This is the q value defined in equations 11 and 12 of DION
         !! ---------------------------------------------------------------

         q = kF+LDA_1*LOG(1.0_dp+1.0_dp/LDA_2)+gradient_correction

         !! ---------------------------------------------------------------
         !! Here, we calculate q0 by saturating q according to equation 5 of SOLER.  Also, we find
         !! the derivative dq0_dq needed for the derivatives dq0_drho and dq0_dgradrh0 discussed below.
         !! ---------------------------------------------------------------------------------------
         exponent = 0.0_dp
         DO index = 1, m_cut
            exponent = exponent+((q/q_cut)**index)/index
         END DO
         q0(i_grid) = q_cut*(1.0_dp-EXP(-exponent))
         !! ---------------------------------------------------------------------------------------
         !! This is to handle a case with q0 too small.  We simply set it to the smallest q value in
         !! out q_mesh.  Hopefully this doesn't get used often (ever)
         !! ---------------------------------------------------------------------------------------
         IF (q0(i_grid) < q_min) THEN
            q0(i_grid) = q_min
         END IF
      END DO

   END SUBROUTINE get_q0_on_grid_eo_vdw

! **************************************************************************************************
!> \brief ...
!> \param total_rho ...
!> \param gradient_rho ...
!> \param q0 ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE get_q0_on_grid_eo_rvv10(total_rho, gradient_rho, q0, dispersion_env)

      REAL(dp), INTENT(IN)                               :: total_rho(:), gradient_rho(:, :)
      REAL(dp), INTENT(OUT)                              :: q0(:)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      INTEGER, PARAMETER                                 :: m_cut = 12

      INTEGER                                            :: i_grid, index
      REAL(dp)                                           :: b_value, C_value, exponent, gmod2, k, &
                                                            mod_grad, q, q_cut, q_min, w0, wg2, wp2

      q_cut = dispersion_env%q_cut
      q_min = dispersion_env%q_min
      b_value = dispersion_env%b_value
      C_value = dispersion_env%c_value

      ! initialize q0-related arrays ...
      q0(:) = q_cut

      DO i_grid = 1, SIZE(total_rho)

         gmod2 = gradient_rho(i_grid, 1)**2+gradient_rho(i_grid, 2)**2+gradient_rho(i_grid, 3)**2

         !if (total_rho(i_grid) > epsr .and. gmod2 > epsr) cycle
         IF (total_rho(i_grid) > epsr) THEN

            !! Calculate some intermediate values needed to find q
            !! ------------------------------------------------------------------------------------
            mod_grad = SQRT(gmod2)

            wp2 = 16.0_dp*pi*total_rho(i_grid)
            wg2 = 4_dp*C_value*(mod_grad/total_rho(i_grid))**4

            k = b_value*3.0_dp*pi*((total_rho(i_grid)/(9.0_dp*pi))**(1.0_dp/6.0_dp))
            w0 = SQRT(wg2+wp2/3.0_dp)

            q = w0/k

            !! Here, we calculate q0 by saturating q according
            !! ---------------------------------------------------------------------------------------

            exponent = 0.0_dp

            DO index = 1, m_cut

               exponent = exponent+((q/q_cut)**index)/index

            END DO

            q0(i_grid) = q_cut*(1.0_dp-EXP(-exponent))

            IF (q0(i_grid) < q_min) THEN
               q0(i_grid) = q_min
            END IF

         ENDIF

      END DO

   END SUBROUTINE get_q0_on_grid_eo_rvv10

   !! This routine is modeled after an algorithm from "Numerical Recipes in C" by Cambridge University
   !! press, page 97.  It was adapted for Fortran, of course and for the problem at hand, in that it finds
   !! the bin a particular x value is in and then loops over all the P_i functions so we only have to find
   !! the bin once.
! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param d2y_dx2 ...
!> \param evaluation_points ...
!> \param values ...
! **************************************************************************************************
   SUBROUTINE spline_interpolation(x, d2y_dx2, evaluation_points, values)

      REAL(dp), INTENT(in)                               :: x(:), d2y_dx2(:, :), evaluation_points(:)
      REAL(dp), INTENT(inout)                            :: values(:, :)

      INTEGER                                            :: i_grid, index, lower_bound, &
                                                            Ngrid_points, Nx, P_i, upper_bound
      REAL(dp)                                           :: a, b, c, d, dx
      REAL(dp), ALLOCATABLE                              :: y(:)

      Nx = SIZE(x)
      Ngrid_points = SIZE(evaluation_points)

      !! Allocate the temporary array
      ALLOCATE (y(Nx))

      DO i_grid = 1, Ngrid_points
         lower_bound = 1
         upper_bound = Nx
         DO WHILE ((upper_bound-lower_bound) > 1)
            index = (upper_bound+lower_bound)/2
            IF (evaluation_points(i_grid) > x(index)) THEN
               lower_bound = index
            ELSE
               upper_bound = index
            END IF
         END DO

         dx = x(upper_bound)-x(lower_bound)

         a = (x(upper_bound)-evaluation_points(i_grid))/dx
         b = (evaluation_points(i_grid)-x(lower_bound))/dx
         c = ((a**3-a)*dx**2)/6.0_dp
         d = ((b**3-b)*dx**2)/6.0_dp

         DO P_i = 1, Nx
            y = 0
            y(P_i) = 1
            values(i_grid, P_i) = a*y(lower_bound)+b*y(upper_bound) &
                                  +(c*d2y_dx2(P_i, lower_bound)+d*d2y_dx2(P_i, upper_bound))
         END DO
      END DO

      DEALLOCATE (y)

   END SUBROUTINE spline_interpolation

   !! This routine is modeled after an algorithm from "Numerical Recipes in C" by Cambridge
   !! University Press, pages 96-97.  It was adapted for Fortran and for the problem at hand.
! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param d2y_dx2 ...
! **************************************************************************************************
   SUBROUTINE initialize_spline_interpolation(x, d2y_dx2)

      REAL(dp), INTENT(in)                               :: x(:)
      REAL(dp), INTENT(inout)                            :: d2y_dx2(:, :)

      INTEGER                                            :: index, Nx, P_i
      REAL(dp)                                           :: temp1, temp2
      REAL(dp), ALLOCATABLE                              :: temp_array(:), y(:)

      Nx = SIZE(x)
      ALLOCATE (temp_array(Nx), y(Nx))

      DO P_i = 1, Nx
         !! In the Soler method, the polynomicals that are interpolated are Kronecker delta funcions
         !! at a particular q point.  So, we set all y values to 0 except the one corresponding to
         !! the particular function P_i.
         !! ----------------------------------------------------------------------------------------
         y = 0.0_dp
         y(P_i) = 1.0_dp
         !! ----------------------------------------------------------------------------------------

         d2y_dx2(P_i, 1) = 0.0_dp
         temp_array(1) = 0.0_dp
         DO index = 2, Nx-1
            temp1 = (x(index)-x(index-1))/(x(index+1)-x(index-1))
            temp2 = temp1*d2y_dx2(P_i, index-1)+2.0_dp
            d2y_dx2(P_i, index) = (temp1-1.0_dp)/temp2
            temp_array(index) = (y(index+1)-y(index))/(x(index+1)-x(index)) &
                                -(y(index)-y(index-1))/(x(index)-x(index-1))
            temp_array(index) = (6.0_dp*temp_array(index)/(x(index+1)-x(index-1)) &
                                 -temp1*temp_array(index-1))/temp2
         END DO
         d2y_dx2(P_i, Nx) = 0.0_dp
         DO index = Nx-1, 1, -1
            d2y_dx2(P_i, index) = d2y_dx2(P_i, index)*d2y_dx2(P_i, index+1)+temp_array(index)
         END DO
      END DO

      DEALLOCATE (temp_array, y)

   END SUBROUTINE initialize_spline_interpolation

   ! **************************************************************************************************
   !! This routine is modeled after an algorithm from "Numerical Recipes in C" by Cambridge
   !! University Press, page 97.  Adapted for Fortran and the problem at hand.  This function is used to
   !! find the Phi_alpha_beta needed for equations 8 and 11 of SOLER.
! **************************************************************************************************
!> \brief ...
!> \param k ...
!> \param kernel_of_k ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE interpolate_kernel(k, kernel_of_k, dispersion_env)

      REAL(dp), INTENT(in)                               :: k
      REAL(dp), INTENT(inout)                            :: kernel_of_k(:, :)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      CHARACTER(len=*), PARAMETER :: routineN = 'interpolate_kernel', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k_i, Nr_points, q1_i, q2_i
      REAL(dp)                                           :: A, B, C, D, dk, r_max
      REAL(dp), DIMENSION(:, :, :), POINTER              :: d2phi_dk2, kernel

      Nr_points = dispersion_env%nr_points
      r_max = dispersion_env%r_max
      dk = dispersion_env%dk
      kernel => dispersion_env%kernel
      d2phi_dk2 => dispersion_env%d2phi_dk2

      !! Check to make sure that the kernel table we have is capable of dealing with this
      !! value of k.  If k is larger than Nr_points*2*pi/r_max then we can't perform the
      !! interpolation.  In that case, a kernel file should be generated with a larger number
      !! of radial points.
      !! -------------------------------------------------------------------------------------
      IF(.NOT.(k < Nr_points*dk))CALL cp__a("qs_dispersion_nonloc.F",1194)
      !! -------------------------------------------------------------------------------------
      kernel_of_k = 0.0_dp
      !! This integer division figures out which bin k is in since the kernel
      !! is set on a uniform grid.
      k_i = INT(k/dk)

      !! Test to see if we are trying to interpolate a k that is one of the actual
      !! function points we have.  The value is just the value of the function in that
      !! case.
      !! ----------------------------------------------------------------------------------------
      IF (MOD(k, dk) == 0) THEN
         DO q1_i = 1, dispersion_env%Nqs
            DO q2_i = 1, q1_i
               kernel_of_k(q1_i, q2_i) = kernel(k_i, q1_i, q2_i)
               kernel_of_k(q2_i, q1_i) = kernel(k_i, q2_i, q1_i)
            END DO
         END DO
         RETURN
      END IF
      !! ----------------------------------------------------------------------------------------
      !! If we are not on a function point then we carry out the interpolation
      !! ----------------------------------------------------------------------------------------
      A = (dk*(k_i+1.0_dp)-k)/dk
      B = (k-dk*k_i)/dk
      C = (A**3-A)*dk**2/6.0_dp
      D = (B**3-B)*dk**2/6.0_dp
      DO q1_i = 1, dispersion_env%Nqs
         DO q2_i = 1, q1_i
            kernel_of_k(q1_i, q2_i) = A*kernel(k_i, q1_i, q2_i)+B*kernel(k_i+1, q1_i, q2_i) &
                                      +(C*d2phi_dk2(k_i, q1_i, q2_i)+D*d2phi_dk2(k_i+1, q1_i, q2_i))
            kernel_of_k(q2_i, q1_i) = kernel_of_k(q1_i, q2_i)
         END DO
      END DO

   END SUBROUTINE interpolate_kernel

! **************************************************************************************************
!> \brief ...
!> \param k ...
!> \param dkernel_of_dk ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE interpolate_dkernel_dk(k, dkernel_of_dk, dispersion_env)
      REAL(dp), INTENT(in)                               :: k
      REAL(dp), INTENT(inout)                            :: dkernel_of_dk(:, :)
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      CHARACTER(len=*), PARAMETER :: routineN = 'interpolate_dkernel_dk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k_i, Nr_points, q1_i, q2_i
      REAL(dp)                                           :: A, B, dAdk, dBdk, dCdk, dDdk, dk, r_max
      REAL(dp), DIMENSION(:, :, :), POINTER              :: d2phi_dk2, kernel

      Nr_points = dispersion_env%nr_points
      r_max = dispersion_env%r_max
      dk = dispersion_env%dk
      kernel => dispersion_env%kernel
      d2phi_dk2 => dispersion_env%d2phi_dk2

      IF(.NOT.(k < Nr_points*dk))CALL cp__a("qs_dispersion_nonloc.F",1255)

      dkernel_of_dk = 0.0_dp
      k_i = INT(k/dk)

      A = (dk*(k_i+1.0_dp)-k)/dk
      B = (k-dk*k_i)/dk
      dAdk = -1.0_dp/dk
      dBdk = 1.0_dp/dk
      dCdk = -((3*A**2-1.0_dp)/6.0_dp)*dk
      dDdk = ((3*B**2-1.0_dp)/6.0_dp)*dk
      DO q1_i = 1, dispersion_env%Nqs
         DO q2_i = 1, q1_i
            dkernel_of_dk(q1_i, q2_i) = dAdk*kernel(k_i, q1_i, q2_i)+dBdk*kernel(k_i+1, q1_i, q2_i) &
                                        +dCdk*d2phi_dk2(k_i, q1_i, q2_i)+dDdk*d2phi_dk2(k_i+1, q1_i, q2_i)
            dkernel_of_dk(q2_i, q1_i) = dkernel_of_dk(q1_i, q2_i)
         END DO
      END DO

   END SUBROUTINE interpolate_Dkernel_Dk

! **************************************************************************************************

END MODULE qs_dispersion_nonloc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief deal with the Fermi distribution, compute it, fix mu, get derivs
!> \author Joost VandeVondele
!> \date 09.2008
! **************************************************************************************************
MODULE fermi_utils

   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: Fermi, FermiFixed, FermiFixedDeriv, Fermikp, Fermikp2

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fermi_utils'

CONTAINS
! **************************************************************************************************
!> \brief   returns occupations according to Fermi-Dirac statistics
!>          for a given set of energies and fermi level.
!>          Note that singly occupied orbitals are assumed
!> \param   f occupations
!> \param   N total number of electrons (output)
!> \param kTS ...
!> \param   e eigenvalues
!> \param   mu Fermi level (input)
!> \param   T  electronic temperature
!> \param   maxocc maximum occupation of an orbital
!> \param   estate excited state in core level spectroscopy
!> \param   festate occupation of the excited state in core level spectroscopy
!> \date    09.2008
!> \par History
!>          - Made estate and festate optional (LT, 2014/02/26)
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE Fermi(f, N, kTS, e, mu, T, maxocc, estate, festate)

      REAL(KIND=dp), INTENT(out)                         :: f(:), N, kTS
      REAL(KIND=dp), INTENT(IN)                          :: e(:), mu, T, maxocc
      INTEGER, INTENT(IN), OPTIONAL                      :: estate
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: festate

      CHARACTER(len=*), PARAMETER :: routineN = 'Fermi', routineP = moduleN//':'//routineN

      INTEGER                                            :: I, Nstate
      REAL(KIND=dp)                                      :: arg, occupation, term1, term2, tmp, &
                                                            tmp2, tmp3, tmp4, tmplog

      Nstate = SIZE(e)
      kTS = 0.0_dp
      ! kTS is the entropic contribution to the energy i.e. -TS
      ! kTS= kT*[f ln f + (1-f) ln (1-f)]

      DO i = 1, Nstate
         IF (PRESENT(estate) .AND. PRESENT(festate)) THEN
            IF (i == estate) THEN
               occupation = festate
            ELSE
               occupation = maxocc
            END IF
         ELSE
            occupation = maxocc
         END IF
         ! have the result of exp go to zero instead of overflowing
         IF (e(i) > mu) THEN
            arg = -(e(i)-mu)/T
            ! tmp is smaller than 1
            tmp = EXP(arg)
            tmp4 = tmp+1.0_dp
            tmp2 = tmp/tmp4
            tmp3 = 1.0_dp/tmp4
            f(i) = occupation*tmp2
            ! log(1+eps), might need to be written more accurately
            tmplog = -LOG(tmp4)
            term1 = tmp2*(arg+tmplog)
            term2 = tmp3*tmplog
            kTS = kTS+T*occupation*(term1+term2)
         ELSE
            arg = (e(i)-mu)/T
            ! tmp is smaller than 1
            tmp = EXP(arg)
            tmp4 = tmp+1.0_dp
            tmp2 = 1.0_dp/tmp4
            tmp3 = tmp/tmp4
            f(i) = occupation*tmp2
            tmplog = -LOG(tmp4)
            term1 = tmp2*tmplog
            term2 = tmp3*(arg+tmplog)
            kTS = kTS+T*occupation*(term1+term2)
         END IF
      END DO

      N = accurate_sum(f)

   END SUBROUTINE Fermi

! **************************************************************************************************
!> \brief Fermi function for KP cases
!> \param f   Occupation numbers
!> \param nel Number of electrons (total)
!> \param kTS Entropic energy contribution
!> \param e   orbital (band) energies
!> \param mu  chemical potential
!> \param wk  kpoint weights
!> \param t   Temperature
!> \param maxocc Maximum occupation
! **************************************************************************************************
   SUBROUTINE Fermi2(f, nel, kTS, e, mu, wk, t, maxocc)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: f
      REAL(KIND=dp), INTENT(OUT)                         :: nel, kTS
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: e
      REAL(KIND=dp), INTENT(IN)                          :: mu
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wk
      REAL(KIND=dp), INTENT(IN)                          :: t, maxocc

      CHARACTER(len=*), PARAMETER :: routineN = 'Fermi2', routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, is, nkp, nmo
      REAL(KIND=dp)                                      :: arg, beta, term1, term2, tmp, tmp2, &
                                                            tmp3, tmp4, tmplog

      nmo = SIZE(e, 1)
      nkp = SIZE(e, 2)
      kTS = 0.0_dp
      ! kTS is the entropic contribution to the energy i.e. -TS
      ! kTS= kT*[f ln f + (1-f) ln (1-f)]
      IF (t > 1.0e-14_dp) THEN
         beta = 1.0_dp/t
         DO ik = 1, nkp
            DO is = 1, nmo
               IF (e(is, ik) > mu) THEN
                  arg = -(e(is, ik)-mu)*beta
                  tmp = EXP(arg)
                  tmp4 = tmp+1.0_dp
                  tmp2 = tmp/tmp4
                  tmp3 = 1.0_dp/tmp4
                  f(is, ik) = maxocc*tmp2
                  tmplog = -LOG(tmp4)
                  term1 = tmp2*(arg+tmplog)
                  term2 = tmp3*tmplog
                  kTS = kTS+t*maxocc*(term1+term2)*wk(ik)
               ELSE
                  arg = (e(is, ik)-mu)*beta
                  tmp = EXP(arg)
                  tmp4 = tmp+1.0_dp
                  tmp2 = 1.0_dp/tmp4
                  tmp3 = tmp/tmp4
                  f(is, ik) = maxocc*tmp2
                  tmplog = -LOG(tmp4)
                  term1 = tmp2*tmplog
                  term2 = tmp3*(arg+tmplog)
                  kTS = kTS+T*maxocc*(term1+term2)*wk(ik)
               END IF
            END DO
         END DO
      ELSE
         DO ik = 1, nkp
            DO is = 1, nmo
               IF (e(is, ik) <= mu) THEN
                  f(is, ik) = maxocc
               ELSE
                  f(is, ik) = 0.0_dp
               END IF
            END DO
         END DO
      END IF

      nel = 0.0_dp
      DO ik = 1, nkp
         nel = nel+accurate_sum(f(1:nmo, ik))*wk(ik)
      END DO

   END SUBROUTINE Fermi2

! **************************************************************************************************
!> \brief   returns occupations according to Fermi-Dirac statistics
!>          for a given set of energies and number of electrons.
!>          Note that singly occupied orbitals are assumed.
!>          could fail if the fermi level lies out of the range of eigenvalues
!>          (to be fixed)
!> \param   f occupations
!> \param   mu Fermi level (output)
!> \param kTS ...
!> \param   e eigenvalues
!> \param   N total number of electrons (input)
!> \param   T  electronic temperature
!> \param   maxocc maximum occupation of an orbital
!> \param   estate excited state in core level spectroscopy
!> \param   festate occupation of the excited state in core level spectroscopy
!> \date    09.2008
!> \par History
!>          - Made estate and festate optional (LT, 2014/02/26)
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE FermiFixed(f, mu, kTS, e, N, T, maxocc, estate, festate)
      REAL(KIND=dp), INTENT(OUT)                         :: f(:), mu, kTS
      REAL(KIND=dp), INTENT(IN)                          :: e(:), N, T, maxocc
      INTEGER, INTENT(IN), OPTIONAL                      :: estate
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: festate

      CHARACTER(len=*), PARAMETER :: routineN = 'FermiFixed', routineP = moduleN//':'//routineN

      INTEGER                                            :: iter, my_estate
      REAL(KIND=dp)                                      :: mu_max, mu_min, mu_now, my_festate, &
                                                            N_max, N_min, N_now

      IF (PRESENT(estate) .AND. PRESENT(festate)) THEN
         my_estate = estate
         my_festate = festate
      ELSE
         my_estate = NINT(maxocc)
         my_festate = my_estate
      END IF

! bisection search to find N
! first bracket

      mu_min = MINVAL(e)
      iter = 0
      DO
         iter = iter+1
         CALL Fermi(f, N_min, kTS, e, mu_min, T, maxocc, my_estate, my_festate)
         IF (N_min > N .OR. iter > 20) THEN
            mu_min = mu_min-T
         ELSE
            EXIT
         ENDIF
      ENDDO

      mu_max = MAXVAL(e)
      iter = 0
      DO
         iter = iter+1
         CALL Fermi(f, N_max, kTS, e, mu_max, T, maxocc, my_estate, my_festate)
         IF (N_max < N .OR. iter > 20) THEN
            mu_max = mu_max+T
         ELSE
            EXIT
         ENDIF
      ENDDO

      ! now bisect
      iter = 0
      DO WHILE (mu_max-mu_min > EPSILON(mu)*MAX(1.0_dp, ABS(mu_max), ABS(mu_min)))
         iter = iter+1
         mu_now = (mu_max+mu_min)/2.0_dp
         CALL Fermi(f, N_now, kTS, e, mu_now, T, maxocc, my_estate, my_festate)
         iter = iter+1
         IF (N_now <= N) THEN
            mu_min = mu_now
            N_min = N_now
         ELSE
            mu_max = mu_now
            N_max = N_now
         ENDIF
         IF (iter > 200) EXIT
      ENDDO

      mu = (mu_max+mu_min)/2.0_dp
      CALL Fermi(f, N_now, kTS, e, mu, T, maxocc, my_estate, my_festate)

   END SUBROUTINE FermiFixed

! **************************************************************************************************
!> \brief Bisection search to find mu for a given nel (kpoint case)
!> \param f   Occupation numbers
!> \param mu  chemical potential
!> \param kTS Entropic energy contribution
!> \param e   orbital (band) energies
!> \param nel Number of electrons (total)
!> \param wk  kpoint weights
!> \param t   Temperature
!> \param maxocc Maximum occupation
! **************************************************************************************************
   SUBROUTINE Fermikp(f, mu, kTS, e, nel, wk, t, maxocc)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: f
      REAL(KIND=dp), INTENT(OUT)                         :: mu, kTS
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: e
      REAL(KIND=dp), INTENT(IN)                          :: nel
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wk
      REAL(KIND=dp), INTENT(IN)                          :: t, maxocc

      CHARACTER(len=*), PARAMETER :: routineN = 'Fermikp', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: epsocc = 1.0e-12_dp

      INTEGER                                            :: iter
      REAL(KIND=dp)                                      :: de, mu_max, mu_min, mu_now, N_max, &
                                                            N_min, N_now

! bisection search to find mu for a given nel

      de = t*LOG((1.0_dp-epsocc)/epsocc)
      de = MAX(de, 0.5_dp)
      mu_min = MINVAL(e)-de
      mu_max = MAXVAL(e)+de
      iter = 0
      DO WHILE (mu_max-mu_min > EPSILON(mu)*MAX(1.0_dp, ABS(mu_max), ABS(mu_min)))
         mu_now = (mu_max+mu_min)/2.0_dp
         CALL Fermi2(f, N_now, kTS, e, mu_now, wk, T, maxocc)
         iter = iter+1
         IF (ABS(N_now-nel) < nel*epsocc) THEN
            mu_min = mu_now
            N_min = N_now
            mu_max = mu_now
            N_max = N_now
         ELSEIF (N_now <= nel) THEN
            mu_min = mu_now
            N_min = N_now
         ELSE
            mu_max = mu_now
            N_max = N_now
         ENDIF
         IF (iter > 400) EXIT
      ENDDO

      mu = (mu_max+mu_min)/2.0_dp
      CALL Fermi2(f, N_now, kTS, e, mu, wk, T, maxocc)

   END SUBROUTINE Fermikp

! **************************************************************************************************
!> \brief Bisection search to find mu for a given nel (kpoint case)
!> \param f   Occupation numbers
!> \param mu  chemical potential
!> \param kTS Entropic energy contribution
!> \param e   orbital (band) energies
!> \param nel Number of electrons (total)
!> \param wk  kpoint weights
!> \param t   Temperature
! **************************************************************************************************
   SUBROUTINE Fermikp2(f, mu, kTS, e, nel, wk, t)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: f
      REAL(KIND=dp), INTENT(OUT)                         :: mu, kTS
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: e
      REAL(KIND=dp), INTENT(IN)                          :: nel
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wk
      REAL(KIND=dp), INTENT(IN)                          :: t

      CHARACTER(len=*), PARAMETER :: routineN = 'Fermikp2', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: epsocc = 1.0e-12_dp

      INTEGER                                            :: iter
      REAL(KIND=dp)                                      :: de, kTSa, kTSb, mu_max, mu_min, mu_now, &
                                                            N_max, N_min, N_now, na, nb

! only do spin polarized case

      IF(.NOT.(SIZE(f, 3) == 2 .AND. SIZE(e, 3) == 2))CALL cp__a("fermi_utils.F",357)
      ! bisection search to find mu for a given nel
      de = t*LOG((1.0_dp-epsocc)/epsocc)
      de = MAX(de, 0.5_dp)
      mu_min = MINVAL(e)-de
      mu_max = MAXVAL(e)+de
      iter = 0
      DO WHILE (mu_max-mu_min > EPSILON(mu)*MAX(1.0_dp, ABS(mu_max), ABS(mu_min)))
         mu_now = (mu_max+mu_min)/2.0_dp
         CALL Fermi2(f(:, :, 1), na, kTSa, e(:, :, 1), mu_now, wk, T, 1.0_dp)
         CALL Fermi2(f(:, :, 2), nb, kTSb, e(:, :, 2), mu_now, wk, T, 1.0_dp)
         N_now = na+nb
         iter = iter+1
         IF (ABS(N_now-nel) < nel*epsocc) THEN
            mu_min = mu_now
            N_min = N_now
            mu_max = mu_now
            N_max = N_now
         ELSEIF (N_now <= nel) THEN
            mu_min = mu_now
            N_min = N_now
         ELSE
            mu_max = mu_now
            N_max = N_now
         ENDIF
         IF (iter > 400) EXIT
      ENDDO

      mu = (mu_max+mu_min)/2.0_dp
      CALL Fermi2(f(:, :, 1), na, kTSa, e(:, :, 1), mu, wk, T, 1.0_dp)
      CALL Fermi2(f(:, :, 2), na, kTSb, e(:, :, 2), mu, wk, T, 1.0_dp)
      kTS = kTSa+kTSb

   END SUBROUTINE Fermikp2

! **************************************************************************************************
!> \brief   returns f and dfde for a given set of energies and number of electrons
!>          it is a numerical derivative, trying to use a reasonable step length
!>          it ought to yield an accuracy of approximately EPSILON()^(2/3) (~10^-11)
!>          l ~ 10*T yields best accuracy
!>          Note that singly occupied orbitals are assumed.
!>          To be fixed: this could be parallellized for better efficiency
!> \param   dfde derivatives of the occupation numbers with respect to the eigenvalues
!>               the ith column is the derivative of f wrt to e_i
!> \param   f occupations
!> \param   mu Fermi level (input)
!> \param kTS ...
!> \param   e eigenvalues
!> \param   N total number of electrons (output)
!> \param   T  electronic temperature
!> \param maxocc ...
!> \param   l  typical lenght scale (~ 10 * T)
!> \param estate ...
!> \param festate ...
!> \date    09.2008
!> \par   History
!>          - Made estate and festate optional (LT, 2014/02/26)
!>          - Changed order of input, so l is before the two optional varaibles
!>            (LT, 2014/02/26)
!> \author  Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE FermiFixedDeriv(dfde, f, mu, kTS, e, N, T, maxocc, l, estate, festate)
      REAL(KIND=dp), INTENT(OUT)                         :: dfde(:, :), f(:), mu, kTS
      REAL(KIND=dp), INTENT(IN)                          :: e(:), N, T, maxocc, l
      INTEGER, INTENT(IN), OPTIONAL                      :: estate
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: festate

      CHARACTER(len=*), PARAMETER :: routineN = 'FermiFixedDeriv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, I, my_estate, Nstate
      REAL(KIND=dp)                                      :: h, mux, my_festate
      REAL(KIND=dp), ALLOCATABLE                         :: ex(:), fx(:)

      CALL timeset(routineN, handle)

      Nstate = SIZE(e)
      ALLOCATE (ex(Nstate), fx(Nstate))

      IF (PRESENT(estate) .AND. PRESENT(festate)) THEN
         my_estate = estate
         my_festate = festate
      ELSE
         my_estate = NINT(maxocc)
         my_festate = my_estate
      END IF

      DO I = 1, Nstate
         ! NR 5.7.8
         ! the problem here is that each f_i 'seems to have' a different lenght scale
         ! and it would be to expensive to compute each single df_i/de_i using a finite difference
         h = (EPSILON(h)**(1.0_dp/3.0_dp))*l
         ! get an exact machine representable number close to this h
         h = 2.0_dp**EXPONENT(h)
         ! this should write three times the same number
         ! write(*,*) h,(e(i)+h)-e(i),(e(i)-h)-e(i)
         ! and the symmetric finite difference
         ex(:) = e
         ex(i) = e(i)+h
         CALL FermiFixed(fx, mux, kTS, ex, N, T, maxocc, my_estate, my_festate)
         dfde(:, I) = fx
         ex(i) = e(i)-h
         CALL FermiFixed(fx, mux, kTS, ex, N, T, maxocc, my_estate, my_festate)
         dfde(:, I) = (dfde(:, I)-fx)/(2.0_dp*h)
      ENDDO
      DEALLOCATE (ex, fx)

      CALL FermiFixed(f, mu, kTS, e, N, T, maxocc, my_estate, my_festate)

      CALL timestop(handle)

   END SUBROUTINE FermiFixedDeriv

END MODULE fermi_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set occupation of molecular orbitals
!> \par History
!>      - set_mo_occupation subroutines moved from qs_mo_types (11.12.2014 MI)
!> \author  MI
! **************************************************************************************************

MODULE qs_mo_occupation

   USE cp_log_handling,                 ONLY: cp_to_string
   USE fermi_utils,                     ONLY: FermiFixed,&
                                              FermiFixedDeriv
   USE input_constants,                 ONLY: smear_energy_window,&
                                              smear_fermi_dirac,&
                                              smear_list
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type,&
                                              set_mo_set
   USE scf_control_types,               ONLY: smear_type
   USE util,                            ONLY: sort
   USE xas_env_types,                   ONLY: get_xas_env,&
                                              xas_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mo_occupation'

   PUBLIC :: set_mo_occupation

   INTERFACE set_mo_occupation
      MODULE PROCEDURE set_mo_occupation_1, set_mo_occupation_2
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief  Occupation for smeared spin polarized electronic structures
!>         with relaxed multiplicity
!>
!> \param mo_array ...
!> \param smear ...
!> \date    10.03.2011 (MI)
!> \author  MI
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_mo_occupation_3(mo_array, smear)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(smear_type), POINTER                          :: smear

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_mo_occupation_3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: all_nmo, handle, homo_a, homo_b, i, &
                                                            lfomo_a, lfomo_b, nmo_a, nmo_b, &
                                                            xas_estate
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: all_index
      LOGICAL                                            :: is_large
      REAL(KIND=dp)                                      :: all_nelec, kTS, mu, nelec_a, nelec_b, &
                                                            occ_estate
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: all_eigval, all_occ
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigval_a, eigval_b, occ_a, occ_b

      IF(.NOT.(ASSOCIATED(mo_array)))CALL cp__a("qs_mo_occupation.F",75)
      IF(.NOT.((SIZE(mo_array) == 2)))CALL cp__a("qs_mo_occupation.F",76)
      CALL timeset(routineN, handle)

      NULLIFY (eigval_a, eigval_b, occ_a, occ_b)
      CALL get_mo_set(mo_set=mo_array(1)%mo_set, nmo=nmo_a, eigenvalues=eigval_a, &
                      occupation_numbers=occ_a)
      CALL get_mo_set(mo_set=mo_array(2)%mo_set, nmo=nmo_b, eigenvalues=eigval_b, &
                      occupation_numbers=occ_b)
      all_nmo = nmo_a+nmo_b
      ALLOCATE (all_eigval(all_nmo))
      ALLOCATE (all_occ(all_nmo))
      ALLOCATE (all_index(all_nmo))

      all_eigval(1:nmo_a) = eigval_a(1:nmo_a)
      all_eigval(nmo_a+1:all_nmo) = eigval_b(1:nmo_b)

      CALL sort(all_eigval, all_nmo, all_index)

      xas_estate = -1
      occ_estate = 0.0_dp

      nelec_a = 0.0_dp
      nelec_b = 0.0_dp
      all_nelec = 0.0_dp
      nelec_a = accurate_sum(occ_a(:))
      nelec_b = accurate_sum(occ_b(:))
      all_nelec = nelec_a+nelec_b

      DO i = 1, all_nmo
         IF (all_index(i) <= nmo_a) THEN
            all_occ(i) = occ_a(all_index(i))
         ELSE
            all_occ(i) = occ_b(all_index(i)-nmo_a)
         END IF
      END DO

      CALL FermiFixed(all_occ, mu, kTS, all_eigval, all_nelec, &
                      smear%electronic_temperature, 1._dp, xas_estate, occ_estate)

      is_large = ABS(MAXVAL(all_occ)-1.0_dp) > smear%eps_fermi_dirac
      ! this is not a real problem, but the temperature might be a bit large
      IF (is_large) &
         CALL cp__w("qs_mo_occupation.F",118,"Fermi-Dirac smearing includes the first MO")

      is_large = ABS(MINVAL(all_occ)) > smear%eps_fermi_dirac
      IF (is_large) &
         CALL cp_warn(cp__l("qs_mo_occupation.F",122), &
                      "Fermi-Dirac smearing includes the last MO => "// &
                      "Add more MOs for proper smearing.")

      ! check that the total electron count is accurate
      is_large = (ABS(all_nelec-accurate_sum(all_occ(:))) > smear%eps_fermi_dirac*all_nelec)
      IF (is_large) &
         CALL cp__w("qs_mo_occupation.F",129,"Total number of electrons is not accurate")

      DO i = 1, all_nmo
         IF (all_index(i) <= nmo_a) THEN
            occ_a(all_index(i)) = all_occ(i)
            eigval_a(all_index(i)) = all_eigval(i)
         ELSE
            occ_b(all_index(i)-nmo_a) = all_occ(i)
            eigval_b(all_index(i)-nmo_a) = all_eigval(i)
         END IF
      END DO

      nelec_a = accurate_sum(occ_a(:))
      nelec_b = accurate_sum(occ_b(:))

      DO i = 1, nmo_a
         IF (occ_a(i) < 1.0_dp) THEN
            lfomo_a = i
            EXIT
         END IF
      END DO
      DO i = 1, nmo_b
         IF (occ_b(i) < 1.0_dp) THEN
            lfomo_b = i
            EXIT
         END IF
      END DO
      homo_a = lfomo_a-1
      DO i = nmo_a, lfomo_a, -1
         IF (occ_a(i) > smear%eps_fermi_dirac) THEN
            homo_a = i
            EXIT
         END IF
      END DO
      homo_b = lfomo_b-1
      DO i = nmo_b, lfomo_b, -1
         IF (occ_b(i) > smear%eps_fermi_dirac) THEN
            homo_b = i
            EXIT
         END IF
      END DO

      CALL set_mo_set(mo_set=mo_array(1)%mo_set, kTS=kTS/2.0_dp, mu=mu, n_el_f=nelec_a, &
                      lfomo=lfomo_a, homo=homo_a, uniform_occupation=.FALSE.)
      CALL set_mo_set(mo_set=mo_array(2)%mo_set, kTS=kTS/2.0_dp, mu=mu, n_el_f=nelec_b, &
                      lfomo=lfomo_b, homo=homo_b, uniform_occupation=.FALSE.)

      CALL timestop(handle)

   END SUBROUTINE set_mo_occupation_3

! **************************************************************************************************
!> \brief   Prepare an occupation of alpha and beta MOs following an Aufbau
!>          principle, i.e. allowing a change in multiplicity.
!> \param mo_array ...
!> \param smear ...
!> \param eval_deriv ...
!> \date    25.01.2010 (MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_mo_occupation_2(mo_array, smear, eval_deriv)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(smear_type), POINTER                          :: smear
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eval_deriv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_mo_occupation_2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, lumo_a, lumo_b, &
                                                            multiplicity_new, multiplicity_old, &
                                                            nelec
      REAL(KIND=dp)                                      :: nelec_f, threshold
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigval_a, eigval_b
      TYPE(mo_set_type), POINTER                         :: mo_set_a, mo_set_b

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(mo_array)))CALL cp__a("qs_mo_occupation.F",208)
      mo_set_a => mo_array(1)%mo_set
      IF(.NOT.(ASSOCIATED(mo_set_a)))CALL cp__a("qs_mo_occupation.F",210)
      ! Fall back for the case that we have only one MO set
      IF (SIZE(mo_array) == 1) THEN
         IF (PRESENT(eval_deriv)) THEN
            CALL set_mo_occupation_1(mo_set_a, smear=smear, eval_deriv=eval_deriv)
         ELSE
            CALL set_mo_occupation_1(mo_set_a, smear=smear)
         END IF
         CALL timestop(handle)
         RETURN
      END IF
      IF(.NOT.((SIZE(mo_array) == 2)))CALL cp__a("qs_mo_occupation.F",221)
      mo_set_b => mo_array(2)%mo_set

      IF (smear%do_smear) THEN
         IF (smear%fixed_mag_mom < 0.0_dp) THEN
            IF (smear%fixed_mag_mom /= -1.0_dp) THEN
               IF(.NOT.(.NOT. (PRESENT(eval_deriv))))CALL cp__a("qs_mo_occupation.F",227)
               CALL set_mo_occupation_3(mo_array, smear=smear)
               CALL timestop(handle)
               RETURN
            END IF
         ELSE
            nelec_f = mo_set_a%n_el_f+mo_set_b%n_el_f
            IF (ABS((mo_set_a%n_el_f-mo_set_b%n_el_f)-smear%fixed_mag_mom) > smear%eps_fermi_dirac*nelec_f) THEN
               mo_set_a%n_el_f = nelec_f/2.0_dp+smear%fixed_mag_mom/2.0_dp
               mo_set_b%n_el_f = nelec_f/2.0_dp-smear%fixed_mag_mom/2.0_dp
            END IF
            IF(.NOT.(.NOT. (PRESENT(eval_deriv))))CALL cp__a("qs_mo_occupation.F",238)
            CALL set_mo_occupation_1(mo_set_a, smear=smear)
            CALL set_mo_occupation_1(mo_set_b, smear=smear)
         END IF
      END IF

      IF (.NOT. ((mo_set_a%flexible_electron_count > 0.0_dp) .AND. &
                 (mo_set_b%flexible_electron_count > 0.0_dp))) THEN
         IF (PRESENT(eval_deriv)) THEN
            CALL set_mo_occupation_1(mo_set_a, smear=smear, eval_deriv=eval_deriv)
            CALL set_mo_occupation_1(mo_set_b, smear=smear, eval_deriv=eval_deriv)
         ELSE
            CALL set_mo_occupation_1(mo_set_a, smear=smear)
            CALL set_mo_occupation_1(mo_set_b, smear=smear)
         END IF
         CALL timestop(handle)
         RETURN
      END IF

      nelec = mo_set_a%nelectron+mo_set_b%nelectron

      multiplicity_old = mo_set_a%nelectron-mo_set_b%nelectron+1

      IF (mo_set_a%nelectron >= mo_set_a%nmo) &
         CALL cp_warn(cp__l("qs_mo_occupation.F",262), &
                      "All alpha MOs are occupied. Add more alpha MOs to "// &
                      "allow for a higher multiplicity")
      IF ((mo_set_b%nelectron >= mo_set_b%nmo) .AND. (mo_set_b%nelectron /= mo_set_a%nelectron)) &
         CALL cp_warn(cp__l("qs_mo_occupation.F",266), "All beta MOs are occupied. Add more beta MOs to "// &
                      "allow for a lower multiplicity")

      eigval_a => mo_set_a%eigenvalues
      eigval_b => mo_set_b%eigenvalues

      lumo_a = 1
      lumo_b = 1

      ! Apply Aufbau principle
      DO i = 1, nelec
         ! Threshold is needed to ensure a preference for alpha occupation in the case
         ! of degeneracy
         threshold = MAX(mo_set_a%flexible_electron_count, mo_set_b%flexible_electron_count)
         IF ((eigval_a(lumo_a)-threshold) < eigval_b(lumo_b)) THEN
            lumo_a = lumo_a+1
         ELSE
            lumo_b = lumo_b+1
         END IF
         IF (lumo_a > mo_set_a%nmo) THEN
            IF (i /= nelec) &
               CALL cp_warn(cp__l("qs_mo_occupation.F",287), &
                            "All alpha MOs are occupied. Add more alpha MOs to "// &
                            "allow for a higher multiplicity")
            IF (i < nelec) THEN
               lumo_a = lumo_a-1
               lumo_b = lumo_b+1
            END IF
         END IF
         IF (lumo_b > mo_set_b%nmo) THEN
            IF (lumo_b < lumo_a) &
               CALL cp_warn(cp__l("qs_mo_occupation.F",297), &
                            "All beta MOs are occupied. Add more beta MOs to "// &
                            "allow for a lower multiplicity")
            IF (i < nelec) THEN
               lumo_a = lumo_a+1
               lumo_b = lumo_b-1
            END IF
         END IF
      END DO

      mo_set_a%homo = lumo_a-1
      mo_set_b%homo = lumo_b-1

      IF (mo_set_b%homo > mo_set_a%homo) THEN
         CALL cp_warn(cp__l("qs_mo_occupation.F",311), &
                      "More beta ("// &
                      TRIM(ADJUSTL(cp_to_string(mo_set_b%homo)))// &
                      ") than alpha ("// &
                      TRIM(ADJUSTL(cp_to_string(mo_set_a%homo)))// &
                      ") MOs are occupied. Resorting to low spin state")
         mo_set_a%homo = nelec/2+MODULO(nelec, 2)
         mo_set_b%homo = nelec/2
      END IF

      mo_set_a%nelectron = mo_set_a%homo
      mo_set_b%nelectron = mo_set_b%homo
      multiplicity_new = mo_set_a%nelectron-mo_set_b%nelectron+1

      IF (multiplicity_new /= multiplicity_old) &
         CALL cp_warn(cp__l("qs_mo_occupation.F",326), &
                      "Multiplicity changed from "// &
                      TRIM(ADJUSTL(cp_to_string(multiplicity_old)))//" to "// &
                      TRIM(ADJUSTL(cp_to_string(multiplicity_new))))

      IF (PRESENT(eval_deriv)) THEN
         CALL set_mo_occupation_1(mo_set_a, smear=smear, eval_deriv=eval_deriv)
         CALL set_mo_occupation_1(mo_set_b, smear=smear, eval_deriv=eval_deriv)
      ELSE
         CALL set_mo_occupation_1(mo_set_a, smear=smear)
         CALL set_mo_occupation_1(mo_set_b, smear=smear)
      END IF

      CALL timestop(handle)

   END SUBROUTINE set_mo_occupation_2

! **************************************************************************************************
!> \brief   Smearing of the MO occupation with all kind of occupation numbers
!> \param   mo_set MO dataset structure
!> \param   smear optional smearing information
!> \param   eval_deriv on entry the derivative of the KS energy wrt to the occupation number
!>                     on exit  the derivative of the full free energy (i.e. KS and entropy) wrt to the eigenvalue
!> \param xas_env ...
!> \date    17.04.2002 (v1.0), 26.08.2008 (v1.1)
!> \author  Matthias Krack
!> \version 1.1
! **************************************************************************************************
   SUBROUTINE set_mo_occupation_1(mo_set, smear, eval_deriv, xas_env)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(smear_type), OPTIONAL, POINTER                :: smear
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eval_deriv
      TYPE(xas_environment_type), OPTIONAL, POINTER      :: xas_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_mo_occupation_1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_first, imo, nmo, nomo, &
                                                            xas_estate
      LOGICAL                                            :: equal_size, is_large
      REAL(KIND=dp)                                      :: e1, e2, edelta, edist, el_count, &
                                                            lengthscale, nelec, occ_estate, &
                                                            xas_nelectron
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dfde

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(mo_set)))CALL cp__a("qs_mo_occupation.F",374)
      IF(.NOT.(ASSOCIATED(mo_set%eigenvalues)))CALL cp__a("qs_mo_occupation.F",375)
      IF(.NOT.(ASSOCIATED(mo_set%occupation_numbers)))CALL cp__a("qs_mo_occupation.F",376)
      mo_set%occupation_numbers(:) = 0.0_dp

      ! Quick return, if no electrons are available
      IF (mo_set%nelectron == 0) THEN
         CALL timestop(handle)
         RETURN
      END IF

      xas_estate = -1
      occ_estate = 0.0_dp
      IF (PRESENT(xas_env)) THEN
         CALL get_xas_env(xas_env=xas_env, xas_nelectron=xas_nelectron, occ_estate=occ_estate, xas_estate=xas_estate)
         nomo = INT(xas_nelectron+1-occ_estate)
         IF (MOD(xas_nelectron+1-occ_estate, 1.0_dp) > EPSILON(0.0_dp)) nomo = nomo+1

         mo_set%occupation_numbers(1:nomo) = mo_set%maxocc
         IF (xas_estate > 0) mo_set%occupation_numbers(xas_estate) = occ_estate
         el_count = SUM(mo_set%occupation_numbers(1:nomo))
         IF (el_count > xas_nelectron) mo_set%occupation_numbers(nomo) = mo_set%occupation_numbers(nomo)-(el_count-xas_nelectron)
         el_count = SUM(mo_set%occupation_numbers(1:nomo))
         is_large = ABS(el_count-xas_nelectron) > xas_nelectron*EPSILON(el_count)
         IF(.NOT.(.NOT. is_large))CALL cp__a("qs_mo_occupation.F",398)
      ELSE
         IF (MODULO(mo_set%nelectron, INT(mo_set%maxocc)) == 0) THEN
            nomo = NINT(mo_set%nelectron/mo_set%maxocc)
            ! Initialize MO occupations
            mo_set%occupation_numbers(1:nomo) = mo_set%maxocc
         ELSE
            nomo = INT(mo_set%nelectron/mo_set%maxocc)+1
            ! Initialize MO occupations
            mo_set%occupation_numbers(1:nomo-1) = mo_set%maxocc
            mo_set%occupation_numbers(nomo) = mo_set%nelectron-(nomo-1)*mo_set%maxocc
         END IF
      END IF
      nmo = SIZE(mo_set%eigenvalues)

      IF(.NOT.(nmo >= nomo))CALL cp__a("qs_mo_occupation.F",413)
      IF(.NOT.((SIZE(mo_set%occupation_numbers) == nmo)))CALL cp__a("qs_mo_occupation.F",414)

      mo_set%homo = nomo
      mo_set%lfomo = nomo+1
      mo_set%mu = mo_set%eigenvalues(nomo)

      ! Check consistency of the array lengths
      IF (PRESENT(eval_deriv)) THEN
         equal_size = (SIZE(mo_set%occupation_numbers, 1) == SIZE(eval_deriv, 1))
         IF(.NOT.(equal_size))CALL cp__a("qs_mo_occupation.F",423)
      END IF

      ! Quick return, if no smearing information is supplied (TO BE FIXED, smear should become non-optional...)
      IF (.NOT. PRESENT(smear)) THEN
         ! there is no dependence of the energy on the eigenvalues
         mo_set%uniform_occupation = .TRUE.
         IF (PRESENT(eval_deriv)) THEN
            eval_deriv = 0.0_dp
         END IF
         CALL timestop(handle)
         RETURN
      END IF

      ! Check if proper eigenvalues are already available
      IF (smear%method /= smear_list) THEN
         IF ((ABS(mo_set%eigenvalues(1)) < 1.0E-12_dp) .AND. &
             (ABS(mo_set%eigenvalues(nmo)) < 1.0E-12_dp)) THEN
            CALL timestop(handle)
            RETURN
         END IF
      END IF

      ! Perform smearing
      IF (smear%do_smear) THEN
         IF (PRESENT(xas_env)) THEN
            i_first = xas_estate+1
            nelec = xas_nelectron
         ELSE
            i_first = 1
            IF (smear%fixed_mag_mom == -1.0_dp) THEN
               nelec = REAL(mo_set%nelectron, dp)
            ELSE
               nelec = mo_set%n_el_f
            END IF
         END IF
         SELECT CASE (smear%method)
         CASE (smear_fermi_dirac)
            IF (.NOT. PRESENT(eval_deriv)) THEN
               CALL FermiFixed(mo_set%occupation_numbers, mo_set%mu, mo_set%kTS, mo_set%eigenvalues, Nelec, &
                               smear%electronic_temperature, mo_set%maxocc, xas_estate, occ_estate)
            ELSE
               ! could be a relatively large matrix, but one could get rid of it by never storing it
               ! we only need dE/df * df/de, one could equally parallelize over entries, this could become expensive
               ALLOCATE (dfde(nmo, nmo))
               ! lengthscale could become a parameter, but this is pretty good
               lengthscale = 10*smear%electronic_temperature

               CALL FermiFixedDeriv(dfde, mo_set%occupation_numbers, mo_set%mu, mo_set%kTS, mo_set%eigenvalues, Nelec, &
                                    smear%electronic_temperature, mo_set%maxocc, lengthscale, xas_estate, occ_estate)

               ! deriv of E_{KS}-kT*S wrt to f_i
               eval_deriv = eval_deriv-mo_set%eigenvalues+mo_set%mu
               ! correspondingly the deriv of  E_{KS}-kT*S wrt to e_i
               eval_deriv = MATMUL(TRANSPOSE(dfde), eval_deriv)

               DEALLOCATE (dfde)
            END IF

            ! Find the lowest fractional occupied MO (LFOMO)
            DO imo = i_first, nmo
               IF (mo_set%occupation_numbers(imo) < mo_set%maxocc) THEN
                  mo_set%lfomo = imo
                  EXIT
               END IF
            END DO
            is_large = ABS(MAXVAL(mo_set%occupation_numbers)-mo_set%maxocc) > smear%eps_fermi_dirac
            ! this is not a real problem, but the temperature might be a bit large
            IF (is_large) &
               CALL cp__w("qs_mo_occupation.F",492,"Fermi-Dirac smearing includes the first MO")

            ! Find the highest (fractional) occupied MO which will be now the HOMO
            DO imo = nmo, mo_set%lfomo, -1
               IF (mo_set%occupation_numbers(imo) > smear%eps_fermi_dirac) THEN
                  mo_set%homo = imo
                  EXIT
               END IF
            END DO
            is_large = ABS(MINVAL(mo_set%occupation_numbers)) > smear%eps_fermi_dirac
            IF (is_large) &
               CALL cp_warn(cp__l("qs_mo_occupation.F",503), &
                            "Fermi-Dirac smearing includes the last MO => "// &
                            "Add more MOs for proper smearing.")

            ! check that the total electron count is accurate
            is_large = (ABS(nelec-accurate_sum(mo_set%occupation_numbers(:))) > smear%eps_fermi_dirac*nelec)
            IF (is_large) &
               CALL cp__w("qs_mo_occupation.F",510,"Total number of electrons is not accurate")

         CASE (smear_energy_window)
            ! not implemented
            IF(.NOT.(.NOT. PRESENT(eval_deriv)))CALL cp__a("qs_mo_occupation.F",514)

            ! Define the energy window for the eigenvalues
            e1 = mo_set%eigenvalues(mo_set%homo)-0.5_dp*smear%window_size
            IF (e1 <= mo_set%eigenvalues(1)) &
               CALL cp__w("qs_mo_occupation.F",519,"Energy window for smearing includes the first MO")

            e2 = mo_set%eigenvalues(mo_set%homo)+0.5_dp*smear%window_size
            IF (e2 >= mo_set%eigenvalues(nmo)) &
               CALL cp_warn(cp__l("qs_mo_occupation.F",523), &
                            "Energy window for smearing includes the last MO => "// &
                            "Add more MOs for proper smearing.")

            ! Find the lowest fractional occupied MO (LFOMO)
            DO imo = i_first, nomo
               IF (mo_set%eigenvalues(imo) > e1) THEN
                  mo_set%lfomo = imo
                  EXIT
               END IF
            END DO

            ! Find the highest fractional occupied (non-zero) MO which will be the HOMO
            DO imo = nmo, nomo, -1
               IF (mo_set%eigenvalues(imo) < e2) THEN
                  mo_set%homo = imo
                  EXIT
               END IF
            END DO

            ! Get the number of electrons to be smeared
            edist = 0.0_dp
            nelec = 0.0_dp

            DO imo = mo_set%lfomo, mo_set%homo
               nelec = nelec+mo_set%occupation_numbers(imo)
               edist = edist+ABS(e2-mo_set%eigenvalues(imo))
            END DO

            ! Smear electrons inside the energy window
            DO imo = mo_set%lfomo, mo_set%homo
               edelta = ABS(e2-mo_set%eigenvalues(imo))
               mo_set%occupation_numbers(imo) = MIN(mo_set%maxocc, nelec*edelta/edist)
               nelec = nelec-mo_set%occupation_numbers(imo)
               edist = edist-edelta
            END DO

         CASE (smear_list)
            equal_size = SIZE(mo_set%occupation_numbers, 1) == SIZE(smear%list, 1)
            IF(.NOT.(equal_size))CALL cp__a("qs_mo_occupation.F",562)
            mo_set%occupation_numbers = smear%list
            ! there is no dependence of the energy on the eigenvalues
            IF (PRESENT(eval_deriv)) THEN
               eval_deriv = 0.0_dp
            END IF
            ! most general case
            mo_set%lfomo = 1
            mo_set%homo = nmo
         END SELECT

         ! Check, if the smearing involves more than one MO
         IF (mo_set%lfomo == mo_set%homo) THEN
            mo_set%homo = nomo
            mo_set%lfomo = nomo+1
         ELSE
            mo_set%uniform_occupation = .FALSE.
         END IF

      END IF ! do smear

      ! zeros don't count as uniform
      mo_set%uniform_occupation = ALL(mo_set%occupation_numbers == mo_set%maxocc)

      CALL timestop(handle)

   END SUBROUTINE set_mo_occupation_1

END MODULE qs_mo_occupation
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basic linear algebra operations for complex full matrixes
!> \note
!>      - not all functionality implemented
!> \par History
!>      Nearly literal copy of Fawzi's routines
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE cp_cfm_basic_linalg
   USE cp_cfm_types,                    ONLY: cp_cfm_get_element,&
                                              cp_cfm_get_info,&
                                              cp_cfm_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_equivalent
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_cfm_basic_linalg'

   PUBLIC :: cp_cfm_add, cp_cfm_add_fm, cp_cfm_gemm, cp_cfm_lu_decompose, &
             cp_cfm_scale, cp_cfm_column_scale, cp_cfm_schur_product, &
             cp_cfm_solve, cp_cfm_cholesky_decompose, cp_cfm_triangular_multiply, &
             cp_cfm_triangular_invert, cp_cfm_axpy, cp_cfm_geadd, cp_cfm_transpose, &
             cp_cfm_cholesky_invert, cp_cfm_lu_invert, cp_cfm_norm, cp_cfm_latra

   REAL(KIND=dp), EXTERNAL :: zlange, pzlange, pzlatra

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param matrix_c ...
! **************************************************************************************************
   SUBROUTINE cp_cfm_schur_product(matrix_a, matrix_b, matrix_c)

      TYPE(cp_cfm_type), POINTER                         :: matrix_a, matrix_b, matrix_c

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_schur_product', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: a, b, c
      INTEGER                                            :: handle, icol_local, irow_local, mypcol, &
                                                            myprow, ncol_local, nrow_local

      CALL timeset(routineN, handle)

      myprow = matrix_a%matrix_struct%context%mepos(1)
      mypcol = matrix_a%matrix_struct%context%mepos(2)

      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data

      nrow_local = matrix_a%matrix_struct%nrow_locals(myprow)
      ncol_local = matrix_a%matrix_struct%ncol_locals(mypcol)

      DO icol_local = 1, ncol_local
         DO irow_local = 1, nrow_local
            c(irow_local, icol_local) = a(irow_local, icol_local)*b(irow_local, icol_local)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_schur_product

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param matrix_c ...
! **************************************************************************************************
   SUBROUTINE cp_cfm_schur_product_cc(matrix_a, matrix_b, matrix_c)

      TYPE(cp_cfm_type), POINTER                         :: matrix_a, matrix_b, matrix_c

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_schur_product_cc', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: a, b, c
      INTEGER                                            :: handle, icol_local, irow_local, mypcol, &
                                                            myprow, ncol_local, nrow_local

      CALL timeset(routineN, handle)

      myprow = matrix_a%matrix_struct%context%mepos(1)
      mypcol = matrix_a%matrix_struct%context%mepos(2)

      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data

      nrow_local = matrix_a%matrix_struct%nrow_locals(myprow)
      ncol_local = matrix_a%matrix_struct%ncol_locals(mypcol)

      DO icol_local = 1, ncol_local
         DO irow_local = 1, nrow_local
            c(irow_local, icol_local) = a(irow_local, icol_local)*CONJG(b(irow_local, icol_local))
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_schur_product_cc
! **************************************************************************************************
!> \brief   Scale and add two BLACS matrices (a <- alpha*a + beta*b).
!> \param alpha ...
!> \param matrix_a ...
!> \param beta ...
!> \param matrix_b ...
!> \date    11.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_add(alpha, matrix_a, beta, matrix_b)
      COMPLEX(KIND=dp), INTENT(IN)                       :: alpha
      TYPE(cp_cfm_type), POINTER                         :: matrix_a
      COMPLEX(KIND=dp), INTENT(in), OPTIONAL             :: beta
      TYPE(cp_cfm_type), OPTIONAL, POINTER               :: matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_add', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: my_beta
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: a, b
      INTEGER                                            :: handle, mypcol, myprow

      CALL timeset(routineN, handle)

      my_beta = CMPLX(0.0_dp, 0.0_dp, dp)
      IF (PRESENT(beta)) my_beta = beta
      NULLIFY (a, b)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_cfm_basic_linalg.F",147)
      IF(.NOT.(matrix_a%ref_count > 0))CALL cp__a("fm/cp_cfm_basic_linalg.F",148)
      ! to do: use dscal,dcopy,daxp
      myprow = matrix_a%matrix_struct%context%mepos(1)
      mypcol = matrix_a%matrix_struct%context%mepos(2)

      a => matrix_a%local_data

      IF (my_beta == 0.0_dp) THEN

         IF (alpha == CMPLX(0.0_dp, 0.0_dp, dp)) THEN
            a(:, :) = CMPLX(0.0_dp, 0.0_dp, dp)
         ELSE IF (alpha == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
            RETURN
         ELSE
            a(:, :) = alpha*a(:, :)
         END IF

      ELSE
         IF(.NOT.(PRESENT(matrix_b)))CALL cp__a("fm/cp_cfm_basic_linalg.F",166)
         IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_cfm_basic_linalg.F",167)
         IF(.NOT.(matrix_b%ref_count > 0))CALL cp__a("fm/cp_cfm_basic_linalg.F",168)
         IF (matrix_a%matrix_struct%context%group /= matrix_b%matrix_struct%context%group) &
            CALL cp__b("fm/cp_cfm_basic_linalg.F",170,"matrixes must be in the same blacs context")

         IF (cp_fm_struct_equivalent(matrix_a%matrix_struct, &
                                     matrix_b%matrix_struct)) THEN

            b => matrix_b%local_data

            IF (alpha == CMPLX(0.0_dp, 0.0_dp, dp)) THEN
               IF (my_beta == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
                  a(:, :) = b(:, :)
               ELSE
                  a(:, :) = my_beta*b(:, :)
               END IF
            ELSE IF (alpha == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
               IF (my_beta == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
                  a(:, :) = a(:, :)+b(:, :)
               ELSE
                  a(:, :) = a(:, :)+my_beta*b(:, :)
               END IF
            ELSE
               a(:, :) = alpha*a(:, :)+my_beta*b(:, :)
            END IF
         ELSE
            CALL cp__b("fm/cp_cfm_basic_linalg.F",196,"")
         END IF
      END IF
      CALL timestop(handle)

   END SUBROUTINE cp_cfm_add

! **************************************************************************************************
!> \brief interface to BLACS axpy:  matrix_b = matrix_b + alpha*matrix_a
!>        note that this is a level one routine, hence more efficient than
!>        cp_cfm_geadd
!> \param alpha  : complex scalar
!> \param matrix_a : input matrix_a
!> \param matrix_b : input matrix_b, upon out put the updated matrix_b
!> \author  Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_cfm_axpy(alpha, matrix_a, matrix_b)
      COMPLEX(KIND=dp), INTENT(IN) :: alpha
      TYPE(cp_cfm_type), POINTER :: matrix_a, matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_axpy', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrow_global, ncol_global, matrix_size, handle
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: aa, bb

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_cfm_basic_linalg.F",228)
      IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_cfm_basic_linalg.F",229)
      nrow_global = matrix_a%matrix_struct%nrow_global
      ncol_global = matrix_a%matrix_struct%ncol_global
      IF(.NOT.(nrow_global .EQ. matrix_b%matrix_struct%nrow_global))CALL cp__a("fm/cp_cfm_basic_linalg.F",232)
      IF(.NOT.(ncol_global .EQ. matrix_b%matrix_struct%ncol_global))CALL cp__a("fm/cp_cfm_basic_linalg.F",233)

      matrix_size = nrow_global*ncol_global
      aa => matrix_a%local_data
      bb => matrix_b%local_data

      CALL zaxpy(matrix_size, &
                 alpha, &
                 aa(1, 1), &
                 1, &
                 bb(1, 1), &
                 1)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_axpy

! **************************************************************************************************
!> \brief interface to BLACS geadd:
!>                matrix_b = beta*matrix_b + alpha*opt(matrix_a)
!>        where opt(matrix_a) can be either:
!>              'N':  matrix_a
!>              'T':  matrix_a^T
!>              'C':  matrix_a^H (Hermitian conjugate)
!>        note that this is a level three routine, use cp_cfm_axpy if that
!>        is sufficient for your needs
!> \param alpha  : complex scalar
!> \param trans  : 'N' normal, 'T' transposed, 'C' Hermitian conjugate
!> \param matrix_a : input matrix_a
!> \param beta   : complex scalar
!> \param matrix_b : input matrix_b, upon out put the updated matrix_b
!> \author  Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_cfm_geadd(alpha, trans, matrix_a, beta, matrix_b)
      COMPLEX(KIND=dp), INTENT(IN) :: alpha, beta
      CHARACTER, INTENT(IN) :: trans
      TYPE(cp_cfm_type), POINTER :: matrix_a, matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_geadd', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrow_global, ncol_global, handle
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: aa, bb
      INTEGER :: ii, jj

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_cfm_basic_linalg.F",299)
      IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_cfm_basic_linalg.F",300)
      nrow_global = matrix_a%matrix_struct%nrow_global
      ncol_global = matrix_a%matrix_struct%ncol_global
      IF(.NOT.(nrow_global .EQ. matrix_b%matrix_struct%nrow_global))CALL cp__a("fm/cp_cfm_basic_linalg.F",303)
      IF(.NOT.(ncol_global .EQ. matrix_b%matrix_struct%ncol_global))CALL cp__a("fm/cp_cfm_basic_linalg.F",304)

      aa => matrix_a%local_data
      bb => matrix_b%local_data

      ! zgeadd is not a standard BLAS function, although is implemented
      ! in some libraries like OpenBLAS, so not going to use it here
      SELECT CASE (trans)
      CASE ('T')
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               bb(ii, jj) = beta*bb(ii, jj)+alpha*aa(jj, ii)
            END DO
         END DO
      CASE ('C')
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               bb(ii, jj) = beta*bb(ii, jj)+alpha*CONJG(aa(jj, ii))
            END DO
         END DO
      CASE DEFAULT
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               bb(ii, jj) = beta*bb(ii, jj)+alpha*aa(ii, jj)
            END DO
         END DO
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_geadd

! **************************************************************************************************
!> \brief   Scale and add two BLACS matrices (a <- alpha*a + beta*b).
!>          where b is a real matrix (adapted from cp_cfm_add)
!> \param alpha ...
!> \param matrix_a ...
!> \param beta ...
!> \param matrix_b ...
!> \date    01.08.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_add_fm(alpha, matrix_a, beta, matrix_b)
      COMPLEX(KIND=dp), INTENT(IN)                       :: alpha
      TYPE(cp_cfm_type), POINTER                         :: matrix_a
      COMPLEX(KIND=dp), INTENT(IN)                       :: beta
      TYPE(cp_fm_type), POINTER                          :: matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_add_fm', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: a
      INTEGER                                            :: handle, mypcol, myprow
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: b

      CALL timeset(routineN, handle)

      NULLIFY (a, b)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_cfm_basic_linalg.F",379)
      myprow = matrix_a%matrix_struct%context%mepos(1)
      mypcol = matrix_a%matrix_struct%context%mepos(2)

      a => matrix_a%local_data

      IF (beta == CMPLX(0.0_dp, 0.0_dp, dp)) THEN

         IF (alpha == CMPLX(0.0_dp, 0.0_dp, dp)) THEN
            a(:, :) = CMPLX(0.0_dp, 0.0_dp, dp)
         ELSE IF (alpha == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
            RETURN
         ELSE
            a(:, :) = alpha*a(:, :)
         END IF

      ELSE
         IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_cfm_basic_linalg.F",396)
         IF (matrix_a%matrix_struct%context%group /= matrix_b%matrix_struct%context%group) &
            CALL cp__b("fm/cp_cfm_basic_linalg.F",398,"matrices must be in the same blacs context")

         IF (cp_fm_struct_equivalent(matrix_a%matrix_struct, &
                                     matrix_b%matrix_struct)) THEN

            b => matrix_b%local_data

            IF (alpha == CMPLX(0.0_dp, 0.0_dp, dp)) THEN
               IF (beta == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
                  a(:, :) = b(:, :)
               ELSE
                  a(:, :) = beta*b(:, :)
               END IF
            ELSE IF (alpha == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
               IF (beta == CMPLX(1.0_dp, 0.0_dp, dp)) THEN
                  a(:, :) = a(:, :)+b(:, :)
               ELSE
                  a(:, :) = a(:, :)+beta*b(:, :)
               END IF
            ELSE
               a(:, :) = alpha*a(:, :)+beta*b(:, :)
            END IF
         ELSE
            CALL cp__b("fm/cp_cfm_basic_linalg.F",424,"")
         END IF
      END IF
      CALL timestop(handle)

   END SUBROUTINE cp_cfm_add_fm

! **************************************************************************************************
!> \brief   Computes the LU decomposition of a given matrix
!>          the actual purpose right now is to compute the determinant of a given matrix
!>          which is most efficiently done this way, but, indeed, destroys the matrix
!>          SERIOUS WARNING (KNOWN BUG) : the sign of the determinant depends on ipivot
!>          one should be able to find out if ipivot is an even or an odd permutation...
!> \param matrix_a ...
!> \param almost_determinant ...
!> \date    11.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_lu_decompose(matrix_a, almost_determinant)
      TYPE(cp_cfm_type), POINTER               :: matrix_a
      COMPLEX(KIND=dp), INTENT(OUT)            :: almost_determinant

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_lu_decompose', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                         :: determinant
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a
      INTEGER                                  :: counter, handle, i, info, &
                                                  n, nrow_local
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: ipivot
      INTEGER, DIMENSION(:), POINTER           :: row_indices
      INTEGER                                  :: lda

      CALL timeset(routineN, handle)

      a => matrix_a%local_data
      n = matrix_a%matrix_struct%nrow_global
      ALLOCATE (ipivot(n))
      CALL cp_cfm_get_info(matrix_a, row_indices=row_indices, nrow_local=nrow_local)
      lda = SIZE(a, 1)
      CALL zgetrf(n, n, a(1, 1), lda, ipivot, info)
      counter = 0
      determinant = CMPLX(1.0_dp, 0.0_dp, dp)
      DO i = 1, n
         IF (ipivot(i) .NE. i) counter = counter+1
         determinant = determinant*a(i, i)
      ENDDO
      IF (MOD(counter, 2) == 1) determinant = -1.0_dp*determinant
      ! info is allowed to be zero
      ! this does just signal a zero diagonal element
      DEALLOCATE (ipivot)
      almost_determinant = determinant ! notice that the sign is random
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief   BLACS interface to the BLAS routine zgemm.
!> \param transa ...
!> \param transb ...
!> \param m ...
!> \param n ...
!> \param k ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param a_first_col ...
!> \param a_first_row ...
!> \param b_first_col ...
!> \param b_first_row ...
!> \param c_first_col ...
!> \param c_first_row ...
!> \date    07.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_gemm(transa, transb, m, n, k, alpha, matrix_a, matrix_b, beta, &
                          matrix_c, a_first_col, a_first_row, b_first_col, b_first_row, c_first_col, &
                          c_first_row)
      CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
      INTEGER, INTENT(IN)                      :: m, n, k
      COMPLEX(KIND=dp), INTENT(IN)             :: alpha
      TYPE(cp_cfm_type), POINTER               :: matrix_a, matrix_b
      COMPLEX(KIND=dp), INTENT(IN)             :: beta
      TYPE(cp_cfm_type), POINTER               :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL            :: a_first_col, a_first_row, &
                                                  b_first_col, b_first_row, &
                                                  c_first_col, c_first_row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_gemm', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a, b, c
      INTEGER                                  :: handle, i_a, i_b, i_c, j_a, j_b, j_c
      INTEGER                                  :: lda, ldb, ldc

      CALL timeset(routineN, handle)
      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data

      IF (PRESENT(a_first_row)) THEN
         i_a = a_first_row
      ELSE
         i_a = 1
      END IF
      IF (PRESENT(a_first_col)) THEN
         j_a = a_first_col
      ELSE
         j_a = 1
      END IF
      IF (PRESENT(b_first_row)) THEN
         i_b = b_first_row
      ELSE
         i_b = 1
      END IF
      IF (PRESENT(b_first_col)) THEN
         j_b = b_first_col
      ELSE
         j_b = 1
      END IF
      IF (PRESENT(c_first_row)) THEN
         i_c = c_first_row
      ELSE
         i_c = 1
      END IF
      IF (PRESENT(c_first_col)) THEN
         j_c = c_first_col
      ELSE
         j_c = 1
      END IF

      lda = SIZE(a, 1)
      ldb = SIZE(b, 1)
      ldc = SIZE(c, 1)

      CALL zgemm(transa, transb, m, n, k, alpha, a(i_a, j_a), &
                 lda, b(i_b, j_b), ldb, beta, c(i_c, j_c), ldc)
      CALL timestop(handle)

   END SUBROUTINE cp_cfm_gemm

! **************************************************************************************************
!> \brief scales column i of matrix a with scaling(i)
!> \param matrixa ...
!> \param scaling : an array used for scaling the columns, SIZE(scaling) determines the number of columns to be scaled
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_cfm_column_scale(matrixa, scaling)
      TYPE(cp_cfm_type), POINTER               :: matrixa
      COMPLEX(KIND=dp), DIMENSION(:), &
         INTENT(in)                             :: scaling

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_column_scale', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a
      INTEGER                                  :: k, mypcol, myprow, n, npcol, &
                                                  nprow, handle
      INTEGER                                  :: i

      CALL timeset(routineN, handle)

      myprow = matrixa%matrix_struct%context%mepos(1)
      mypcol = matrixa%matrix_struct%context%mepos(2)
      nprow = matrixa%matrix_struct%context%num_pe(1)
      npcol = matrixa%matrix_struct%context%num_pe(2)

      a => matrixa%local_data
      n = SIZE(a, 1)
      k = SIZE(scaling)

      DO i = 1, k
         CALL ZSCAL(n, scaling(i), a(1, i), 1)
      END DO

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_column_scale

! **************************************************************************************************
!> \brief scales a matrix
!>      matrix_a = alpha * matrix_b
!> \param alpha ...
!> \param matrix_a ...
!> \note
!>      use cp_fm_set_all to zero (avoids problems with nan)
! **************************************************************************************************
   SUBROUTINE cp_cfm_scale(alpha, matrix_a)
      COMPLEX(KIND=dp), INTENT(IN)                       :: alpha
      TYPE(cp_cfm_type), POINTER                         :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_scale', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: a
      INTEGER                                            :: handle, size_a

      CALL timeset(routineN, handle)

      NULLIFY (a)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_cfm_basic_linalg.F",686)
      IF(.NOT.(matrix_a%ref_count > 0))CALL cp__a("fm/cp_cfm_basic_linalg.F",687)

      a => matrix_a%local_data
      size_a = SIZE(a, 1)*SIZE(a, 2)

      CALL ZSCAL(size_a, alpha, a, 1)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_scale

! **************************************************************************************************
!> \brief computs the the solution to A*b=A_general using lu decomposition
!>        pay attention, both matrices are overwritten, a_general contais the result
!> \param matrix_a ...
!> \param general_a ...
!> \param determinant ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE cp_cfm_solve(matrix_a, general_a, determinant)
      TYPE(cp_cfm_type), POINTER               :: matrix_a, general_a
      COMPLEX(KIND=dp), OPTIONAL               :: determinant

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_solve', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a, a_general
      INTEGER                                  :: counter, handle, i, info, &
                                                  n, nrow_local
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: ipivot
      INTEGER, DIMENSION(:), POINTER           :: row_indices
      INTEGER                                  :: lda, ldb

      CALL timeset(routineN, handle)

      a => matrix_a%local_data
      a_general => general_a%local_data
      n = matrix_a%matrix_struct%nrow_global
      CALL cp_cfm_get_info(matrix_a, row_indices=row_indices, nrow_local=nrow_local)
      ALLOCATE (ipivot(n))

      lda = SIZE(a, 1)
      ldb = SIZE(a_general, 1)
      CALL zgetrf(n, n, a(1, 1), lda, ipivot, info)
      IF (PRESENT(determinant)) THEN
         counter = 0
         determinant = CMPLX(1.0_dp, 0.0_dp, dp)
         DO i = 1, n
            IF (ipivot(i) .NE. i) counter = counter+1
            determinant = determinant*a(i, i)
         ENDDO
         IF (MOD(counter, 2) == 1) determinant = -1.0_dp*determinant
      END IF
      CALL zgetrs("N", n, n, a(1, 1), lda, ipivot, a_general, ldb, info)
      ! info is allowed to be zero
      ! this does just signal a zero diagonal element
      DEALLOCATE (ipivot)
      CALL timestop(handle)

   END SUBROUTINE cp_cfm_solve

! **************************************************************************************************
!> \brief inverts a matrix using LU decomposition
!>        the input matrix will be overwritten
!> \param matrix   : input a general square non-singular matrix, outputs its inverse
!> \param info_out : optional, if present outputs the info from (p)zgetri
!> \author Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_cfm_lu_invert(matrix, info_out)
      TYPE(cp_cfm_type), POINTER               :: matrix
      INTEGER, INTENT(OUT), OPTIONAL           :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_lu_invert', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows_global, handle, info, lwork
      INTEGER, DIMENSION(:), ALLOCATABLE          :: ipivot
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER  :: mat
      COMPLEX(KIND=dp), DIMENSION(:), ALLOCATABLE :: work
      INTEGER                                  :: lda

      CALL timeset(routineN, handle)

      mat => matrix%local_data
      nrows_global = matrix%matrix_struct%nrow_global
      IF(.NOT.(nrows_global .EQ. matrix%matrix_struct%ncol_global))CALL cp__a("fm/cp_cfm_basic_linalg.F",810)
      ALLOCATE (ipivot(nrows_global))
      ! do LU decomposition
      lda = SIZE(mat, 1)
      CALL zgetrf(nrows_global, nrows_global, &
                  mat, lda, ipivot, info)
      IF (info /= 0) THEN
         CALL cp_abort(cp__l("fm/cp_cfm_basic_linalg.F",823), "LU decomposition has failed")
      END IF
      ! do inversion
      ALLOCATE (work(1))
      CALL zgetri(nrows_global, mat, lda, &
                  ipivot, work, -1, info)
      lwork = INT(work(1))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))
      CALL zgetri(nrows_global, mat, lda, &
                  ipivot, work, lwork, info)
      DEALLOCATE (work)
      DEALLOCATE (ipivot)

      IF (PRESENT(info_out)) THEN
         info = info_out
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_cfm_basic_linalg.F",856), "LU inversion has failed")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_lu_invert

! **************************************************************************************************
!> \brief used to replace a symmetric positive def. matrix M with its cholesky
!>      decomposition U: M = U^T * U, with U upper triangular
!> \param matrix : the matrix to replace with its cholesky decomposition
!> \param n : the number of row (and columns) of the matrix &
!>        (defaults to the min(size(matrix)))
!> \param info_out : if present, outputs info from (p)zpotrf
!> \par History
!>      05.2002 created [JVdV]
!>      12.2002 updated, added n optional parm [fawzi]
!> \author Joost
! **************************************************************************************************
   SUBROUTINE cp_cfm_cholesky_decompose(matrix, n, info_out)
      TYPE(cp_cfm_type), POINTER               :: matrix
      INTEGER, INTENT(in), OPTIONAL            :: n
      INTEGER, INTENT(OUT), OPTIONAL           :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_cholesky_decompose', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a
      INTEGER                                  :: handle, info, my_n

      CALL timeset(routineN, handle)

      my_n = MIN(matrix%matrix_struct%nrow_global, &
                 matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("fm/cp_cfm_basic_linalg.F",895)
         my_n = n
      END IF

      a => matrix%local_data

      CALL zpotrf('U', my_n, a(1, 1), SIZE(a, 1), info)

      IF (PRESENT(info_out)) THEN
         info = info_out
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_cfm_basic_linalg.F",912), &
                          "Cholesky decompose failed: matrix is not positive definite  or ill-conditioned")
      END IF

      IF(.NOT.(info == 0))CALL cp__a("fm/cp_cfm_basic_linalg.F",916)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_cholesky_decompose

! **************************************************************************************************
!> \brief used to replace the cholesky decomposition by the inverse.
!> \param matrix : the matrix to invert (must be an upper triangular matrix),
!>                 and is the output of the cholesky decomposition
!> \param n : size of the matrix to invert (defaults to the min(size(matrix)))
!> \param info_out : if present, outputs info of (p)zpotri
!> \par History
!>      05.2002 created Lianheng Tong, based on cp_fm_cholesky_invert
!> \author Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_cfm_cholesky_invert(matrix, n, info_out)
      TYPE(cp_cfm_type), POINTER                 :: matrix
      INTEGER, INTENT(in), OPTIONAL              :: n
      INTEGER, INTENT(OUT), OPTIONAL             :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_cholesky_invert', &
                                     routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER  :: aa
      INTEGER                                    :: info, handle
      INTEGER                                    :: my_n

      CALL timeset(routineN, handle)

      my_n = MIN(matrix%matrix_struct%nrow_global, &
                 matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("fm/cp_cfm_basic_linalg.F",951)
         my_n = n
      END IF

      aa => matrix%local_data

      CALL zpotri('U', my_n, aa(1, 1), SIZE(aa, 1), info)

      IF (PRESENT(info_out)) THEN
         info_out = info
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_cfm_basic_linalg.F",968), &
                          "Cholesky invert failed: the matrix is not positive definite or ill-conditioned.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_cholesky_invert

! **************************************************************************************************
!> \brief multiplies in place by a triangular matrix:
!>       matrix_b = alpha op(triangular_matrix) matrix_b
!>      or (if side='R')
!>       matrix_b = alpha matrix_b op(triangular_matrix)
!>      op(triangular_matrix) is:
!>       triangular_matrix (if transa="N" and invert_tr=.false.)
!>       triangular_matrix^T (if transa="T" and invert_tr=.false.)
!>       triangular_matrix^H (if transa="C" and invert_tr=.false.)
!>       triangular_matrix^(-1) (if transa="N" and invert_tr=.true.)
!>       triangular_matrix^(-T) (if transa="T" and invert_tr=.true.)
!>       triangular_matrix^(-H) (if transa="C" and invert_tr=.true.)
!> \param triangular_matrix the triangular matrix that multiplies the other
!> \param matrix_b the matrix that gets multiplied and stores the result
!> \param side on which side of matrix_b stays op(triangular_matrix)
!>        (defaults to 'L')
!> \param transa_tr ...
!> \param invert_tr if the triangular matrix should be inverted
!>        (defaults to false)
!> \param uplo_tr if triangular_matrix is stored in the upper ('U') or
!>        lower ('L') triangle (defaults to 'U')
!> \param unit_diag_tr if the diagonal elements of triangular_matrix should
!>        be assumed to be 1 (defaults to false)
!> \param n_rows the number of rows of the result (defaults to
!>        size(matrix_b,1))
!> \param n_cols the number of columns of the result (defaults to
!>        size(matrix_b,2))
!> \param alpha ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      needs an mpi env
! **************************************************************************************************
   SUBROUTINE cp_cfm_triangular_multiply(triangular_matrix, matrix_b, side, &
                                         transa_tr, invert_tr, uplo_tr, unit_diag_tr, n_rows, n_cols, &
                                         alpha)
      TYPE(cp_cfm_type), POINTER                         :: triangular_matrix, matrix_b
      CHARACTER, INTENT(in), OPTIONAL                    :: side, transa_tr
      LOGICAL, INTENT(in), OPTIONAL                      :: invert_tr
      CHARACTER, INTENT(in), OPTIONAL                    :: uplo_tr
      LOGICAL, INTENT(in), OPTIONAL                      :: unit_diag_tr
      INTEGER, INTENT(in), OPTIONAL                      :: n_rows, n_cols
      COMPLEX(KIND=dp), INTENT(in), OPTIONAL             :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_triangular_multiply', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: side_char, transa, unit_diag, uplo
      COMPLEX(KIND=dp)                                   :: al
      INTEGER                                            :: handle, m, n
      LOGICAL                                            :: invert

      CALL timeset(routineN, handle)
      side_char = 'L'
      unit_diag = 'N'
      uplo = 'U'
      transa = 'N'
      invert = .FALSE.
      al = CMPLX(1.0_dp, 0.0_dp, dp)
      CALL cp_cfm_get_info(matrix_b, nrow_global=m, ncol_global=n)
      IF (PRESENT(side)) side_char = side
      IF (PRESENT(invert_tr)) invert = invert_tr
      IF (PRESENT(uplo_tr)) uplo = uplo_tr
      IF (PRESENT(unit_diag_tr)) THEN
         IF (unit_diag_tr) THEN
            unit_diag = 'U'
         ELSE
            unit_diag = 'N'
         END IF
      END IF
      IF (PRESENT(transa_tr)) transa = transa_tr
      IF (PRESENT(alpha)) al = alpha
      IF (PRESENT(n_rows)) m = n_rows
      IF (PRESENT(n_cols)) n = n_cols

      IF (invert) THEN

         CALL ztrsm(side_char, uplo, transa, unit_diag, m, n, al, &
                    triangular_matrix%local_data(1, 1), &
                    SIZE(triangular_matrix%local_data, 1), &
                    matrix_b%local_data(1, 1), SIZE(matrix_b%local_data, 1))

      ELSE

         CALL ztrmm(side_char, uplo, transa, unit_diag, m, n, al, &
                    triangular_matrix%local_data(1, 1), &
                    SIZE(triangular_matrix%local_data, 1), &
                    matrix_b%local_data(1, 1), SIZE(matrix_b%local_data, 1))

      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_triangular_multiply

! **************************************************************************************************
!> \brief inverts a triangular matrix
!> \param matrix_a ...
!> \param uplo ...
!> \param info_out ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE cp_cfm_triangular_invert(matrix_a, uplo, info_out)
      TYPE(cp_cfm_type), POINTER               :: matrix_a
      CHARACTER, INTENT(IN), OPTIONAL          :: uplo
      INTEGER, INTENT(OUT), OPTIONAL           :: info_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_cfm_triangular_invert', &
                                     routineP = moduleN//':'//routineN

      CHARACTER                                :: unit_diag, my_uplo
      INTEGER                                  :: handle, info, ncol_global
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a

      CALL timeset(routineN, handle)

      unit_diag = 'N'
      my_uplo = 'U'
      IF (PRESENT(uplo)) my_uplo = uplo

      ncol_global = matrix_a%matrix_struct%ncol_global

      a => matrix_a%local_data

      CALL ztrtri(my_uplo, unit_diag, ncol_global, a(1, 1), ncol_global, info)

      IF (PRESENT(info_out)) THEN
         info = info_out
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_cfm_basic_linalg.F",1132), &
                          "triangular invert failed: matrix is not positive definite  or ill-conditioned")
      END IF

      CALL timestop(handle)
   END SUBROUTINE cp_cfm_triangular_invert

! **************************************************************************************************
!> \brief transposes a blacs distributed complex matrix
!> \param matrix  : input matrix
!> \param trans   : 'T' for transpose, 'C' for hermitian conjugate
!> \param matrixt : output matrix
!> \author Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_cfm_transpose(matrix, trans, matrixt)
      TYPE(cp_cfm_type), POINTER :: matrix, matrixt
      CHARACTER, INTENT(IN) :: trans

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_transpose', &
                                     routineP = moduleN//':'//routineN

      INTEGER nrow_global, ncol_global, handle
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: aa, cc
      INTEGER :: ii, jj

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_cfm_basic_linalg.F",1163)
      IF(.NOT.(ASSOCIATED(matrixt)))CALL cp__a("fm/cp_cfm_basic_linalg.F",1164)
      nrow_global = matrix%matrix_struct%nrow_global
      ncol_global = matrix%matrix_struct%ncol_global

      aa => matrix%local_data
      cc => matrixt%local_data

      SELECT CASE (trans)
      CASE ('T')
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               cc(ii, jj) = aa(jj, ii)
            END DO
         END DO
      CASE ('C')
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               cc(ii, jj) = CONJG(aa(jj, ii))
            END DO
         END DO
      CASE DEFAULT
         CALL cp__b("fm/cp_cfm_basic_linalg.F",1215,"trans only accepts 'T' or 'C'")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_transpose

! **************************************************************************************************
!> \brief norm of matrix using (p)zlange
!> \param matrix   : input a general matrix
!> \param mode     : 'M' max abs element value,
!>                   '1' or 'O' one norm, i.e. maximum column sum
!>                   'I' infinity norm, i.e. maximum row sum
!>                   'F' or 'E' Frobenius norm, i.e. sqrt of sum of all squares of elements
!> \retval res     : the norm according to mode
!> \author Lianheng Tong
! **************************************************************************************************
   FUNCTION cp_cfm_norm(matrix, mode) RESULT(res)
      TYPE(cp_cfm_type), POINTER :: matrix
      CHARACTER, INTENT(IN) :: mode
      REAL(KIND=dp) :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_norm', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows, ncols, handle, lwork, nrows_local, ncols_local
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: aa
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: work
      INTEGER :: lda

      CALL timeset(routineN, handle)

      CALL cp_cfm_get_info(matrix=matrix, &
                           nrow_global=nrows, &
                           ncol_global=ncols, &
                           nrow_local=nrows_local, &
                           ncol_local=ncols_local)
      aa => matrix%local_data

      SELECT CASE (mode)
      CASE ('M', 'm')
         lwork = 1
      CASE ('1', 'O', 'o')
         lwork = 1
      CASE ('I', 'i')
         lwork = nrows
      CASE ('F', 'f', 'E', 'e')
         lwork = 1
      CASE DEFAULT
         CALL cp__b("fm/cp_cfm_basic_linalg.F",1287,"mode input is not valid")
      END SELECT
      ALLOCATE (work(lwork))
      lda = SIZE(aa, 1)
      res = zlange(mode, nrows, ncols, aa, lda, work)
      DEALLOCATE (work)

      CALL timestop(handle)

   END FUNCTION cp_cfm_norm

! **************************************************************************************************
!> \brief trace of a matrix using pzlatra
!> \param matrix   : input a square matrix
!> \retval res     : the trace
!> \author Lianheng Tong
! **************************************************************************************************
   FUNCTION cp_cfm_latra(matrix) RESULT(res)
      TYPE(cp_cfm_type), POINTER :: matrix
      COMPLEX(KIND=dp) :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_latra', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows, ncols, handle
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: aa
      INTEGER :: ii

      CALL timeset(routineN, handle)

      nrows = matrix%matrix_struct%nrow_global
      ncols = matrix%matrix_struct%ncol_global
      IF(.NOT.(nrows .EQ. ncols))CALL cp__a("fm/cp_cfm_basic_linalg.F",1324)
      aa => matrix%local_data

      res = 0.0_dp
      DO ii = 1, nrows
         res = res+aa(ii, ii)
      END DO

      CALL timestop(handle)

   END FUNCTION cp_cfm_latra

END MODULE cp_cfm_basic_linalg
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief used for collecting diagonalization schemes available for cp_cfm_type
!> \note
!>      first version : only one routine right now
!> \author Joost VandeVondele (2003-09)
! **************************************************************************************************
MODULE cp_cfm_diag
  USE cp_cfm_basic_linalg,             ONLY: cp_cfm_cholesky_decompose,&
                                             cp_cfm_gemm,&
                                             cp_cfm_column_scale,&
                                             cp_cfm_scale,&
                                             cp_cfm_triangular_invert,&
                                             cp_cfm_triangular_multiply
  USE cp_cfm_types,                    ONLY: cp_cfm_get_info,&
                                             cp_cfm_set_element,&
                                             cp_cfm_to_cfm,&
                                             cp_cfm_type
  USE kinds,                           ONLY: dp

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_cfm_diag'

   PUBLIC :: cp_cfm_heevd, cp_cfm_geeig, cp_cfm_geeig_canon

CONTAINS

! **************************************************************************************************
!> \brief Perform a diagonalisation of a complex matrix
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \par History
!>      - (De)Allocation checks updated (15.02.2011,MK)
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_cfm_heevd(matrix, eigenvectors, eigenvalues)

      TYPE(cp_cfm_type), POINTER               :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT) :: eigenvalues

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_heevd', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER  :: work
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: m
      INTEGER                                  :: handle, info, liwork, &
                                                  lrwork, lwork, n
      INTEGER, DIMENSION(:), POINTER           :: iwork
      REAL(KIND=dp), DIMENSION(:), POINTER     :: rwork

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_cfm_diag.F",75)
      IF(.NOT.(ASSOCIATED(eigenvectors)))CALL cp__a("fm/cp_cfm_diag.F",76)

      n = matrix%matrix_struct%nrow_global
      m => matrix%local_data
      ALLOCATE (iwork(1), rwork(1), work(1))
      ! work space query
      lwork = -1
      lrwork = -1
      liwork = -1

      CALL ZHEEVD('V', 'U', n, m(1, 1), SIZE(m, 1), eigenvalues(1), &
                  work(1), lwork, rwork(1), lrwork, iwork(1), liwork, info)
      lwork = CEILING(REAL(work(1), KIND=dp))
      lrwork = CEILING(REAL(rwork(1), KIND=dp))
      liwork = iwork(1)

      DEALLOCATE (iwork, rwork, work)

      ALLOCATE (iwork(liwork))
      iwork(:) = 0
      ALLOCATE (rwork(lrwork))
      rwork(:) = 0.0_dp
      ALLOCATE (work(lwork))
      work(:) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)

      CALL ZHEEVD('V', 'U', n, m(1, 1), SIZE(m, 1), eigenvalues(1), &
                  work(1), lwork, rwork(1), lrwork, iwork(1), liwork, info)
      eigenvectors%local_data = matrix%local_data

      IF (info /= 0) &
         CALL cp__b("fm/cp_cfm_diag.F",134,"Diagonalisation of a complex matrix failed")
      DEALLOCATE (iwork, rwork, work)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_heevd

! **************************************************************************************************
!> \brief General Eigenvalue Problem  AX = BXE
!>        Single option version: Cholesky decomposition of B
!> \param amatrix ...
!> \param bmatrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param work ...
! **************************************************************************************************
   SUBROUTINE cp_cfm_geeig(amatrix, bmatrix, eigenvectors, eigenvalues, work)

      TYPE(cp_cfm_type), POINTER                         :: amatrix, bmatrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)                        :: eigenvalues
      TYPE(cp_cfm_type), POINTER                         :: work

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_geeig', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nao, nmo
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals

      CALL timeset(routineN, handle)

      CALL cp_cfm_get_info(amatrix, nrow_global=nao)
      ALLOCATE (evals(nao))
      ! Cholesky decompose S=U(T)U
      CALL cp_cfm_cholesky_decompose(bmatrix)
      ! Invert to get U^(-1)
      CALL cp_cfm_triangular_invert(bmatrix)
      ! Reduce to get U^(-T) * H * U^(-1)
      CALL cp_cfm_triangular_multiply(bmatrix, amatrix, side="R")
      CALL cp_cfm_triangular_multiply(bmatrix, amatrix, transa_tr="C")
      ! Diagonalize
      CALL cp_cfm_heevd(matrix=amatrix, eigenvectors=work, eigenvalues=evals)
      ! Restore vectors C = U^(-1) * C*
      CALL cp_cfm_triangular_multiply(bmatrix, work)
      nmo = SIZE(eigenvalues)
      CALL cp_cfm_to_cfm(work, eigenvectors, nmo)
      eigenvalues(1:nmo) = evals(1:nmo)

      DEALLOCATE (evals)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_geeig

! **************************************************************************************************
!> \brief General Eigenvalue Problem  AX = BXE
!>        Use canonical orthogonalization
!> \param amatrix ...
!> \param bmatrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param work ...
!> \param epseig ...
! **************************************************************************************************
   SUBROUTINE cp_cfm_geeig_canon(amatrix, bmatrix, eigenvectors, eigenvalues, work, epseig)

      TYPE(cp_cfm_type), POINTER                         :: amatrix, bmatrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)                        :: eigenvalues
      TYPE(cp_cfm_type), POINTER                         :: work
      REAL(KIND=dp), INTENT(IN)                          :: epseig

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_geeig_canon', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER :: cone = CMPLX(1.0_dp, 0.0_dp, KIND=dp), &
         czero = CMPLX(0.0_dp, 0.0_dp, KIND=dp)

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: cevals
      INTEGER                                            :: handle, i, icol, irow, nao, nc, ncol, &
                                                            nmo, nx
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals

      CALL timeset(routineN, handle)

      ! Test sizees
      CALL cp_cfm_get_info(amatrix, nrow_global=nao)
      nmo = SIZE(eigenvalues)
      ALLOCATE (evals(nao), cevals(nao))

      ! Diagonalize -S matrix, this way the NULL space is at the end of the spectrum
      CALL cp_cfm_scale(-cone, bmatrix)
      CALL cp_cfm_heevd(bmatrix, work, evals)
      evals(:) = -evals(:)
      nc = nao
      DO i = 1, nao
         IF (evals(i) < epseig) THEN
            nc = i-1
            EXIT
         END IF
      END DO
      IF(.NOT.(nc /= 0))CALL cp__a("fm/cp_cfm_diag.F",231)

      IF (nc /= nao) THEN
         IF (nc < nmo) THEN
            ! Copy NULL space definition to last vectors of eigenvectors (if needed)
            ncol = nmo-nc
            CALL cp_cfm_to_cfm(work, eigenvectors, ncol, nc+1, nc+1)
         END IF
         ! Set NULL space in eigenvector matrix of S to zero
         DO icol = nc+1, nao
            DO irow = 1, nao
               CALL cp_cfm_set_element(work, irow, icol, czero)
            END DO
         END DO
         ! Set small eigenvalues to a dummy save value
         evals(nc+1:nao) = 1.0_dp
      END IF
      ! calculate U*s**(-1/2)
      cevals(:) = CMPLX(1.0_dp/SQRT(evals(:)), 0.0_dp, KIND=dp)
      CALL cp_cfm_column_scale(work, cevals)
      ! Reduce to get U^(-C) * H * U^(-1)
      CALL cp_cfm_gemm("C", "N", nao, nao, nao, cone, work, amatrix, czero, bmatrix)
      CALL cp_cfm_gemm("N", "N", nao, nao, nao, cone, bmatrix, work, czero, amatrix)
      IF (nc /= nao) THEN
         ! set diagonal values to save large value
         DO icol = nc+1, nao
            CALL cp_cfm_set_element(amatrix, icol, icol, CMPLX(10000.0_dp, 0.0_dp, KIND=dp))
         END DO
      END IF
      ! Diagonalize
      CALL cp_cfm_heevd(amatrix, bmatrix, evals)
      eigenvalues(1:nmo) = evals(1:nmo)
      nx = MIN(nc, nmo)
      ! Restore vectors C = U^(-1) * C*
      CALL cp_cfm_gemm("N", "N", nao, nx, nc, cone, work, bmatrix, czero, eigenvectors)

      DEALLOCATE (evals)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_geeig_canon

END MODULE cp_cfm_diag
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief various cholesky decomposition related routines
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_fm_cholesky
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_cholesky'

   PUBLIC :: cp_fm_cholesky_decompose, cp_fm_cholesky_invert, &
             cp_fm_cholesky_reduce, cp_fm_cholesky_restore

!***
CONTAINS

! **************************************************************************************************
!> \brief used to replace a symmetric positive def. matrix M with its cholesky
!>      decomposition U: M = U^T * U, with U upper triangular
!> \param matrix the matrix to replace with its cholesky decomposition
!> \param n the number of row (and columns) of the matrix &
!>        (defaults to the min(size(matrix)))
!> \param info_out ...
!> \par History
!>      05.2002 created [JVdV]
!>      12.2002 updated, added n optional parm [fawzi]
!> \author Joost
! **************************************************************************************************
   SUBROUTINE cp_fm_cholesky_decompose(matrix, n, info_out)
      TYPE(cp_fm_type), POINTER                :: matrix
      INTEGER, INTENT(in), OPTIONAL            :: n
      INTEGER, INTENT(out), OPTIONAL           :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_cholesky_decompose', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, info, my_n
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp

      CALL timeset(routineN, handle)

      my_n = MIN(matrix%matrix_struct%nrow_global, &
                 matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("fm/cp_fm_cholesky.F",64)
         my_n = n
      END IF

      a => matrix%local_data
      a_sp => matrix%local_data_sp


      IF (matrix%use_sp) THEN
         CALL spotrf('U', my_n, a_sp(1, 1), SIZE(a_sp, 1), info)
      ELSE
         CALL dpotrf('U', my_n, a(1, 1), SIZE(a, 1), info)
      ENDIF


      IF (PRESENT(info_out)) THEN
         info_out = info
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_fm_cholesky.F",94), &
                          "Cholesky decompose failed: the matrix is not positive definite or ill-conditioned.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_cholesky_decompose

! **************************************************************************************************
!> \brief used to replace the cholesky decomposition by the inverse
!> \param matrix the matrix to invert (must be an upper triangular matrix)
!> \param n size of the matrix to invert (defaults to the min(size(matrix)))
!> \param info_out ...
!> \par History
!>      05.2002 created [JVdV]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_fm_cholesky_invert(matrix, n, info_out)
      TYPE(cp_fm_type), POINTER                 :: matrix
      INTEGER, INTENT(in), OPTIONAL             :: n
      INTEGER, INTENT(OUT), OPTIONAL            :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_cholesky_invert', &
                                     routineP = moduleN//':'//routineN
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp
      INTEGER                                   :: info, handle
      INTEGER                                   :: my_n

      CALL timeset(routineN, handle)

      my_n = MIN(matrix%matrix_struct%nrow_global, &
                 matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("fm/cp_fm_cholesky.F",131)
         my_n = n
      END IF

      a => matrix%local_data
      a_sp => matrix%local_data_sp


      IF (matrix%use_sp) THEN
         CALL spotri('U', my_n, a_sp(1, 1), SIZE(a_sp, 1), info)
      ELSE
         CALL dpotri('U', my_n, a(1, 1), SIZE(a, 1), info)
      ENDIF


      IF (PRESENT(info_out)) THEN
         info_out = info
      ELSE
         IF (info /= 0) &
            CALL cp__b("fm/cp_fm_cholesky.F",162,"Cholesky invert failed: the matrix is not positive definite or ill-conditioned.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_cholesky_invert

! **************************************************************************************************
!> \brief reduce a matrix pencil A,B to normal form
!>      B has to be cholesky decomposed with  cp_fm_cholesky_decompose
!>      before calling this routine
!>      A,B -> inv(U^T)*A*inv(U),1
!>      (AX=BX -> inv(U^T)*A*inv(U)*U*X=U*X hence evecs U*X)
!> \param matrix the symmetric matrix A
!> \param matrixb the cholesky decomposition of matrix B
!> \param itype ...
!> \par History
!>      05.2002 created [JVdV]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_fm_cholesky_reduce(matrix, matrixb, itype)
      TYPE(cp_fm_type), POINTER           :: matrix, matrixb
      INTEGER, OPTIONAL                   :: itype

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_cholesky_reduce', &
                                     routineP = moduleN//':'//routineN
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b
      INTEGER                                   :: info, handle
      INTEGER                                   :: n, my_itype

      CALL timeset(routineN, handle)

      n = matrix%matrix_struct%nrow_global

      my_itype = 1
      IF (PRESENT(itype)) my_itype = itype

      a => matrix%local_data
      b => matrixb%local_data


      CALL dsygst(my_itype, 'U', n, a(1, 1), n, b(1, 1), n, info)


      IF(.NOT.(info == 0))CALL cp__a("fm/cp_fm_cholesky.F",223)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_cholesky_reduce

!
! op can be "SOLVE" (out = U^-1 * in ) or "MULTIPLY"   (out = U * in )
! pos can be "LEFT" or "RIGHT" (U at the left or at the right)
!
! DEPRECATED, see cp_fm_basic_linalg:cp_fm_triangular_multiply
!
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param neig ...
!> \param matrixb ...
!> \param matrixout ...
!> \param op ...
!> \param pos ...
!> \param transa ...
! **************************************************************************************************
   SUBROUTINE cp_fm_cholesky_restore(matrix, neig, matrixb, matrixout, op, pos, transa)
      TYPE(cp_fm_type), POINTER          :: matrix, matrixb, matrixout
      INTEGER, INTENT(IN)                         :: neig
      CHARACTER(LEN=*), INTENT(IN)        :: op
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: pos
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: transa

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_cholesky_restore', &
                                     routineP = moduleN//':'//routineN
      REAL(KIND=dp), DIMENSION(:, :), POINTER         :: a, b, out
      REAL(KIND=sp), DIMENSION(:, :), POINTER         :: a_sp, b_sp, out_sp
      INTEGER                                   :: itype, handle
      INTEGER                                   :: n
      REAL(KIND=dp)                           :: alpha
      INTEGER                                   :: myprow, mypcol
      TYPE(cp_blacs_env_type), POINTER          :: context
      CHARACTER                                 :: chol_pos, chol_transa

      CALL timeset(routineN, handle)

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      n = matrix%matrix_struct%nrow_global
      itype = 1
      IF (op /= "SOLVE" .AND. op /= "MULTIPLY") &
         CALL cp__b("fm/cp_fm_cholesky.F",275,"wrong argument op")

      IF (PRESENT(pos)) THEN
         SELECT CASE (pos)
         CASE ("LEFT")
            chol_pos = 'L'
         CASE ("RIGHT")
            chol_pos = 'R'
         CASE DEFAULT
            CALL cp__b("fm/cp_fm_cholesky.F",284,"wrong argument pos")
         END SELECT
      ELSE
         chol_pos = 'L'
      ENDIF

      chol_transa = 'N'
      IF (PRESENT(transa)) chol_transa = transa

      IF ((matrix%use_sp .NEQV. matrixb%use_sp) .OR. (matrix%use_sp .NEQV. matrixout%use_sp)) &
         CALL cp__b("fm/cp_fm_cholesky.F",294,"not the same precision")

      ! notice b is the cholesky guy
      a => matrix%local_data
      b => matrixb%local_data
      out => matrixout%local_data
      a_sp => matrix%local_data_sp
      b_sp => matrixb%local_data_sp
      out_sp => matrixout%local_data_sp


      alpha = 1.0_dp
      IF (matrix%use_sp) THEN
         CALL scopy(neig*n, a_sp(1, 1), 1, out_sp(1, 1), 1)
      ELSE
         CALL dcopy(neig*n, a(1, 1), 1, out(1, 1), 1)
      ENDIF
      IF (op .EQ. "SOLVE") THEN
         IF (matrix%use_sp) THEN
            CALL strsm(chol_pos, 'U', chol_transa, 'N', n, neig, REAL(alpha, sp), b_sp(1, 1), SIZE(b_sp, 1), out_sp(1, 1), n)
         ELSE
            CALL dtrsm(chol_pos, 'U', chol_transa, 'N', n, neig, alpha, b(1, 1), SIZE(b, 1), out(1, 1), n)
         ENDIF
      ELSE
         IF (matrix%use_sp) THEN
            CALL strmm(chol_pos, 'U', chol_transa, 'N', n, neig, REAL(alpha, sp), b_sp(1, 1), n, out_sp(1, 1), n)
         ELSE
            CALL dtrmm(chol_pos, 'U', chol_transa, 'N', n, neig, alpha, b(1, 1), n, out(1, 1), n)
         ENDIF
      ENDIF


      CALL timestop(handle)

   END SUBROUTINE cp_fm_cholesky_restore

END MODULE cp_fm_cholesky
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief used for collecting some of the diagonalization shemes available for
!>      cp_fm_type. cp_fm_power also moved here as it is very related
!> \note
!>      first version : most routines imported
!> \par History
!>      - unused Jacobi routines removed, cosmetics (05.04.06,MK)
!> \author Joost VandeVondele (2003-08)
! **************************************************************************************************
MODULE cp_fm_diag
USE cp_log_handling, ONLY: cp_logger_type, cp_get_default_logger, cp_logger_get_default_unit_nr, cp_logger_get_unit_nr
  USE machine,                         ONLY: m_memory
  USE cp_blacs_calls,                  ONLY: cp_blacs_gridexit,&
                                             cp_blacs_gridinit
  USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                             cp_blacs_env_release
  USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
  USE cp_fm_basic_linalg,              ONLY: cp_fm_syrk,&
                                             cp_fm_gemm,&
                                             cp_fm_scale,&
                                             cp_fm_column_scale,&
                                             cp_fm_triangular_multiply,&
                                             cp_fm_triangular_invert,&
                                             cp_fm_upper_to_full
  USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                             cp_fm_struct_release,&
                                             cp_fm_struct_type,&
                                             cp_fm_struct_get
  USE cp_fm_types,                     ONLY: cp_fm_create,&
                                             cp_fm_release,&
                                             cp_fm_get_info,&
                                             cp_fm_set_element,&
                                             cp_fm_to_fm,&
                                             cp_fm_type
  USE cp_para_env,                     ONLY: cp_para_env_create,&
                                             cp_para_env_release
  USE cp_blacs_env,                    ONLY: cp_blacs_env_type
  USE cp_para_types,                   ONLY: cp_para_env_type
  USE kinds,                           ONLY: dp
  USE message_passing,                 ONLY: mp_bcast,&
                                             mp_comm_free,&
                                             mp_comm_split,&
                                             mp_comm_split_direct,&
                                             mp_sync


! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_diag'

   ! these saved variables are DIAGONALIZATION global
   INTEGER, SAVE :: diag_type = 0
   INTEGER, SAVE :: kernel_type = 0

   ! Public subroutines

   PUBLIC :: choose_eigv_solver, &
             cp_fm_block_jacobi, &
             cp_fm_power, &
             cp_fm_syevd, &
             cp_fm_syevx, &
             cp_fm_geeig, &
             cp_fm_geeig_canon, &
             diag_init

CONTAINS

! **************************************************************************************************
!> \brief Setup the diagonalization library to be used
!>         Check of availability not yet fully implemented
!>         It should change library to Scalapack if others are not available
!> \param diag_lib diag_library flag from GLOBAL section in input
!> \param switched ...
!> \param k_elpa ...
!> \author  MI 11.2013
! **************************************************************************************************
   SUBROUTINE diag_init(diag_lib, switched, k_elpa)
      CHARACTER(LEN=*), INTENT(IN)                       :: diag_lib
      LOGICAL, INTENT(INOUT)                             :: switched
      INTEGER, INTENT(IN)                                :: k_elpa

      CHARACTER(len=*), PARAMETER :: routineN = 'diag_init', routineP = moduleN//':'//routineN

      LOGICAL                                            :: try_library

! skalapack is the default and always linked

      IF (diag_lib .EQ. "SL") THEN
         try_library = .FALSE.
         diag_type = 1
      ELSE
         try_library = .TRUE.
      END IF

      IF (try_library) THEN
         IF (diag_lib .EQ. "ELPA") THEN

            ! ELPA library not linked, switch to SL
            diag_type = 1
            switched = .TRUE.
         ELSE IF (diag_lib .EQ. "SL2") THEN

            ! SL2 library not linked, switch to SL
            diag_type = 1
            switched = .TRUE.
         END IF
      END IF

      IF (diag_type == 3) THEN
         kernel_type = k_elpa
      END IF

      ! Check that one of the diagonalization type is set
      IF (diag_type < 1) THEN
         ! something wrong
         CALL cp__b("fm/cp_fm_diag.F",141,"Unknown DIAG type")
      END IF

   END SUBROUTINE diag_init

! **************************************************************************************************
!> \brief   Choose the Eigensolver depending on which library is available
!>           ELPA seems to be unstable for small systems
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param info ...
!> \par     info If present returns error code and prevents program stops.
!>               Works currently only for cp_fm_syevd with scalapack.
!>               Other solvers will end the program regardless of PRESENT(info).
! **************************************************************************************************
   SUBROUTINE choose_eigv_solver(matrix, eigenvectors, eigenvalues, info)

      TYPE(cp_fm_type), POINTER                          :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: eigenvalues
      INTEGER, INTENT(OUT), OPTIONAL                     :: info

      CHARACTER(len=*), PARAMETER :: routineN = 'choose_eigv_solver', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: myinfo, nmo

      myinfo = 0

      nmo = SIZE(eigenvalues, 1)

      !sample peak memory
      CALL m_memory()

      IF (diag_type == 3) THEN
         CALL cp_fm_elpa(matrix, eigenvectors, eigenvalues)
      ELSE IF (diag_type == 2) THEN
         CALL cp_fm_syevr(matrix, eigenvectors, eigenvalues, 1, nmo)
      ELSE IF (diag_type == 1) THEN
         CALL cp_fm_syevd(matrix, eigenvectors, eigenvalues, info=myinfo)
      END IF

      IF (PRESENT(info)) info = myinfo
   END SUBROUTINE choose_eigv_solver

! **************************************************************************************************
!> \brief   Computes all eigenvalues and vectors of a real symmetric matrix
!>          significantly faster than syevx, scales also much better.
!>          Needs workspace to allocate all the eigenvectors
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param info ...
!> \par     matrix is supposed to be in upper triangular form, and overwritten by this routine
!> \par     info If present returns error code and prevents program stops.
!>               Works currently only for scalapack.
!>               Other solvers will end the program regardless of PRESENT(info).
! **************************************************************************************************
   SUBROUTINE cp_fm_syevd(matrix, eigenvectors, eigenvalues, info)

      TYPE(cp_fm_type), POINTER                :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT) :: eigenvalues
      INTEGER, INTENT(OUT), OPTIONAL           :: info

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_syevd', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, myinfo, n, nmo
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:) :: eig
      INTEGER                                  :: liwork, lwork
      INTEGER, DIMENSION(:), POINTER           :: iwork
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: m
      REAL(KIND=dp), DIMENSION(:), POINTER     :: work

      CALL timeset(routineN, handle)

      myinfo = 0

      n = matrix%matrix_struct%nrow_global
      ALLOCATE (eig(n))


      !MK Retrieve the optimal work array sizes first
      myinfo = 0
      lwork = -1
      liwork = -1
      m => matrix%local_data
      eig(:) = 0.0_dp

      ALLOCATE (work(1))
      work(:) = 0.0_dp
      ALLOCATE (iwork(1))
      iwork(:) = 0

      CALL dsyevd('V', 'U', n, m(1, 1), n, eig(1), work(1), lwork, iwork(1), liwork, myinfo)

      IF (myinfo /= 0) THEN
         CALL cp__b("fm/cp_fm_diag.F",343,"ERROR in DSYEVD: Could not retrieve work array sizes")
      END IF

      ! Reallocate work arrays and perform diagonalisation
      lwork = INT(work(1))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))

      liwork = iwork(1)
      DEALLOCATE (iwork)
      ALLOCATE (iwork(liwork))
      iwork(:) = 0

      CALL dsyevd('V', 'U', n, m(1, 1), n, eig(1), work(1), lwork, iwork(1), liwork, myinfo)

      IF (myinfo /= 0) THEN
         CALL cp__b("fm/cp_fm_diag.F",359,"Matrix diagonalization failed")
      END IF

      CALL cp_fm_to_fm(matrix, eigenvectors)

      DEALLOCATE (iwork)
      DEALLOCATE (work)

      IF (PRESENT(info)) myinfo = 0

      nmo = SIZE(eigenvalues, 1)
      IF (nmo > n) THEN
         eigenvalues(1:n) = eig(1:n)
      ELSE
         eigenvalues(1:nmo) = eig(1:nmo)
      END IF

      DEALLOCATE (eig)
      CALL timestop(handle)

   END SUBROUTINE cp_fm_syevd

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eigenvectors ...
!> \param eig ...
!> \param info ...
! **************************************************************************************************
   SUBROUTINE cp_fm_syevd_base(matrix, eigenvectors, eig, info)

      TYPE(cp_fm_type), POINTER                :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)              :: eig
      INTEGER, INTENT(OUT)                     :: info

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_syevd_base', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(matrix))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(eigenvectors))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(eig))==-1) EXIT ;  END DO ; ENDIF
      info = -1
      CALL cp__b("fm/cp_fm_diag.F",471,"cp_fm_syevd_base called with out SCALAPACK compiled in")

      CALL timestop(handle)

   END SUBROUTINE cp_fm_syevd_base

! **************************************************************************************************
!> \brief   compute eigenvalues and optionally eigenvectors of a real symmetric matrix using scalapack.
!>          If eigenvectors are required this routine will replicate a full matrix on each CPU...
!>          if more than a handful of vectors are needed, use cp_fm_syevd instead
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param neig ...
!> \param work_syevx ...
!> \par     matrix is supposed to be in upper triangular form, and overwritten by this routine
!>          neig   is the number of vectors needed (default all)
!>          work_syevx evec calculation only, is the fraction of the working buffer allowed (1.0 use full buffer)
!>                     reducing this saves time, but might cause the routine to fail
! **************************************************************************************************
   SUBROUTINE cp_fm_syevx(matrix, eigenvectors, eigenvalues, neig, work_syevx)

      ! Diagonalise the symmetric n by n matrix using the LAPACK library.

      TYPE(cp_fm_type), POINTER                    :: matrix
      TYPE(cp_fm_type), POINTER, OPTIONAL          :: eigenvectors
      REAL(KIND=dp), OPTIONAL, INTENT(IN)        :: work_syevx
      INTEGER, INTENT(IN), OPTIONAL                :: neig
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)   :: eigenvalues

      CHARACTER(LEN=*), PARAMETER :: routineN = "cp_fm_syevx", &
                                     routineP = moduleN//":"//routineN

      REAL(KIND=dp), PARAMETER                     :: orfac = -1.0_dp, &
                                                      vl = 0.0_dp, &
                                                      vu = 0.0_dp

      TYPE(cp_blacs_env_type), POINTER             :: context
      TYPE(cp_logger_type), POINTER                :: logger
      CHARACTER(LEN=1)                             :: job_type
      REAL(KIND=dp)                                :: abstol, work_syevx_local
      INTEGER                                      :: handle, info, &
                                                      liwork, lwork, m, mypcol, &
                                                      myprow, n, nb, npcol, &
                                                      nprow, output_unit, &
                                                      neig_local
      LOGICAL                                      :: ionode, needs_evecs
      INTEGER, DIMENSION(:), ALLOCATABLE           :: ifail, iwork
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE     :: w, work
      REAL(KIND=dp), DIMENSION(:, :), POINTER       :: a, z

      REAL(KIND=dp), EXTERNAL                      :: dlamch

      INTEGER, EXTERNAL                            :: ilaenv

      ! by default all
      n = matrix%matrix_struct%nrow_global
      neig_local = n
      IF (PRESENT(neig)) neig_local = neig
      IF (neig_local == 0) RETURN

      CALL timeset(routineN, handle)

      needs_evecs = PRESENT(eigenvectors)

      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source
      n = matrix%matrix_struct%nrow_global

      ! by default allocate all needed space
      work_syevx_local = 1.0_dp
      IF (PRESENT(work_syevx)) work_syevx_local = work_syevx

      ! set scalapack job type
      IF (needs_evecs) THEN
         job_type = "V"
      ELSE
         job_type = "N"
      ENDIF

      ! target the most accurate calculation of the eigenvalues
      abstol = 2.0_dp*dlamch("S")

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      ALLOCATE (w(n))
      eigenvalues(:) = 0.0_dp

      a => matrix%local_data
      IF (needs_evecs) THEN
         z => eigenvectors%local_data
      ELSE
         ! z will not be referenced
         z => matrix%local_data
      ENDIF

      ! Get the optimal work storage size

      nb = MAX(ilaenv(1, "DSYTRD", "U", n, -1, -1, -1), &
               ilaenv(1, "DORMTR", "U", n, -1, -1, -1))

      lwork = MAX((nb+3)*n, 8*n)+n ! sun bug fix
      liwork = 5*n

      ALLOCATE (ifail(n))
      ifail = 0
      ALLOCATE (iwork(liwork))
      ALLOCATE (work(lwork))
      info = 0

      CALL dsyevx(job_type, "I", "U", n, a(1, 1), n, vl, vu, 1, neig_local, abstol, m, w, z(1, 1), n, work(1), lwork, &
                  iwork(1), ifail(1), info)

      ! Error handling

      IF (info /= 0) THEN
         output_unit = cp_logger_get_unit_nr(logger, local=.FALSE.)
         WRITE (unit=output_unit, FMT="(/,(T3,A,T12,1X,I10))") &
            "info    = ", info
         IF (info > 0) THEN
            WRITE (unit=output_unit, FMT="(/,T3,A,(T12,6(1X,I10)))") &
               "ifail   = ", ifail
         END IF
         CALL cp__b("fm/cp_fm_diag.F",698,"Error in dsyevx")
      END IF

      ! Release work storage

      DEALLOCATE (ifail)
      DEALLOCATE (iwork)
      DEALLOCATE (work)

      eigenvalues(1:neig_local) = w(1:neig_local)
      DEALLOCATE (w)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_syevx

! **************************************************************************************************
!> \brief  computes selected eigenvalues and, optionally, eigenvectors of
!>        a real symmetric matrix A distributed in 2D blockcyclic format by
!>       calling the recommended sequence of ScaLAPACK routines.
!>
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param ilow ...
!> \param iup ...
!> \par     matrix is supposed to be in upper triangular form, and overwritten by this routine
!>          subsets of eigenvalues/vectors can be selected by
!>          specifying a range of values or a range of indices for the desired eigenvalues.
! **************************************************************************************************
   SUBROUTINE cp_fm_syevr(matrix, eigenvectors, eigenvalues, ilow, iup)

      TYPE(cp_fm_type), POINTER                    :: matrix
      TYPE(cp_fm_type), POINTER, OPTIONAL          :: eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)   :: eigenvalues
      INTEGER, INTENT(IN), OPTIONAL                :: ilow, iup

      CHARACTER(LEN=*), PARAMETER :: routineN = "cp_fm_syevr", &
                                     routineP = moduleN//":"//routineN
      REAL(KIND=dp), PARAMETER  :: vl = 0.0_dp, &
                                   vu = 0.0_dp

      CHARACTER(LEN=1)                           :: job_type
      INTEGER                                    :: handle, ilow_local, &
                                                    info, iup_local, &
                                                    lwork, liwork, mypcol, &
                                                    myprow, n, neig
      INTEGER, DIMENSION(:), ALLOCATABLE         :: iwork
      LOGICAL                                    :: ionode, needs_evecs
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE   :: w, work
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: a, z

      TYPE(cp_blacs_env_type), POINTER           :: context
      TYPE(cp_logger_type), POINTER              :: logger

      REAL(KIND=dp), EXTERNAL :: dlamch

      INTEGER                             :: m, nb
      REAL(dp)                            :: abstol
      INTEGER, DIMENSION(:), ALLOCATABLE  :: ifail
      INTEGER, EXTERNAL                   :: ilaenv

      ! by default all
      n = matrix%matrix_struct%nrow_global
      neig = n
      iup_local = n
      ilow_local = 1
      IF (PRESENT(ilow) .AND. PRESENT(iup)) THEN
         neig = iup-ilow+1
         iup_local = iup
         ilow_local = ilow
      END IF
      IF (neig <= 0) RETURN

      CALL timeset(routineN, handle)

      needs_evecs = PRESENT(eigenvectors)

      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source
      n = matrix%matrix_struct%nrow_global

      ! set scalapack job type
      IF (needs_evecs) THEN
         job_type = "V"
      ELSE
         job_type = "N"
      ENDIF

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)

      ALLOCATE (w(n))

      eigenvalues(:) = 0.0_dp


      a => matrix%local_data
      IF (needs_evecs) THEN
         z => eigenvectors%local_data
      ELSE
         ! z will not be referenced
         z => matrix%local_data
      ENDIF

      ! Get the optimal work storage size

      nb = MAX(ilaenv(1, "DSYTRD", "U", n, -1, -1, -1), &
               ilaenv(1, "DORMTR", "U", n, -1, -1, -1))

      lwork = MAX((nb+3)*n, 8*n)+n ! sun bug fix
      liwork = 5*n

      ALLOCATE (ifail(n))
      ifail = 0

      ALLOCATE (iwork(liwork))
      ALLOCATE (work(lwork))

      ! target the most accurate calculation of the eigenvalues
      abstol = 2.0_dp*dlamch("S")

      info = 0
      CALL dsyevx(job_type, "I", "U", n, a(1, 1), n, vl, vu, ilow_local, iup_local, abstol, m, w, z(1, 1), n, work(1), lwork, &
                  iwork(1), ifail(1), info)

      ! Error handling
      IF(.NOT.(info == 0))CALL cp__a("fm/cp_fm_diag.F",889)

      ! Release work storage
      DEALLOCATE (iwork)
      DEALLOCATE (work)


      eigenvalues(ilow_local:iup_local) = w(ilow_local:iup_local)
      DEALLOCATE (w)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_syevr

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
! **************************************************************************************************
   SUBROUTINE cp_fm_elpa(matrix, eigenvectors, eigenvalues)

      TYPE(cp_fm_type), POINTER                :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)              :: eigenvalues

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_elpa', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(matrix))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(eigenvectors))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(eigenvalues))==-1) EXIT ;  END DO ; ENDIF

      CALL cp__b("fm/cp_fm_diag.F",990,"CP2K compiled without the ELPA library.")

      CALL timestop(handle)

   END SUBROUTINE cp_fm_elpa

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param work ...
!> \param exponent ...
!> \param threshold ...
!> \param n_dependent ...
!> \param verbose ...
! **************************************************************************************************
   SUBROUTINE cp_fm_power(matrix, work, exponent, threshold, n_dependent, verbose)

      ! Raise the real symmetric n by n matrix to the power given by
      ! the exponent. All eigenvectors with a corresponding eigenvalue lower
      ! than threshold are quenched. result in matrix

      ! - Creation (29.03.1999, Matthias Krack)
      ! - Parallelised using BLACS and ScaLAPACK (06.06.2001,MK)

      TYPE(cp_fm_type), POINTER                 :: matrix, work
      REAL(KIND=dp), INTENT(IN)               :: exponent, threshold
      INTEGER, INTENT(OUT)                      :: n_dependent
      LOGICAL, INTENT(IN), OPTIONAL             :: verbose

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_power', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                    :: handle, icol_global, &
                                                    mypcol, myprow, &
                                                    ncol_global, npcol, nprow, &
                                                    nrow_global
      LOGICAL                                    :: my_verbose
      REAL(KIND=dp)                              :: condition_number, f, p
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE   :: eigenvalues
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: eigenvectors
      TYPE(cp_blacs_env_type), POINTER           :: context


      CALL timeset(routineN, handle)

      my_verbose = .FALSE.
      IF (PRESENT(verbose)) my_verbose = verbose

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)
      n_dependent = 0
      p = 0.5_dp*exponent

      nrow_global = matrix%matrix_struct%nrow_global
      ncol_global = matrix%matrix_struct%ncol_global

      ALLOCATE (eigenvalues(ncol_global))

      eigenvalues(:) = 0.0_dp

      ! Compute the eigenvectors and eigenvalues

      CALL choose_eigv_solver(matrix, work, eigenvalues)


      eigenvectors => work%local_data

      ! Build matrix**exponent with eigenvector quenching

      DO icol_global = 1, ncol_global

         IF (eigenvalues(icol_global) < threshold) THEN

            n_dependent = n_dependent+1
            eigenvectors(1:nrow_global, icol_global) = 0.0_dp

         ELSE

            f = eigenvalues(icol_global)**p
            eigenvectors(1:nrow_global, icol_global) = &
               f*eigenvectors(1:nrow_global, icol_global)

         END IF

      END DO

      CALL cp_fm_syrk("U", "N", ncol_global, 1.0_dp, work, 1, 1, 0.0_dp, matrix)
      CALL cp_fm_upper_to_full(matrix, work)

      ! Print some warnings/notes

      IF (matrix%matrix_struct%para_env%mepos == 0 .AND. my_verbose) THEN
         condition_number = ABS(eigenvalues(ncol_global)/eigenvalues(1))
         WRITE (UNIT=cp_logger_get_default_unit_nr(), FMT="(/,(T2,A,ES15.6))") &
            "CP_FM_POWER: smallest eigenvalue:", eigenvalues(1), &
            "CP_FM_POWER: largest eigenvalue: ", eigenvalues(ncol_global), &
            "CP_FM_POWER: condition number:   ", condition_number
         IF (eigenvalues(1) <= 0.0_dp) THEN
            WRITE (UNIT=cp_logger_get_default_unit_nr(), FMT="(/,T2,A)") &
               "WARNING: matrix has a negative eigenvalue, tighten EPS_DEFAULT"
         END IF
         IF (condition_number > 1.0E12_dp) THEN
            WRITE (UNIT=cp_logger_get_default_unit_nr(), FMT="(/,T2,A)") &
               "WARNING: high condition number => possibly ill-conditioned matrix"
         END IF
      END IF

      DEALLOCATE (eigenvalues)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_power

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigval ...
!> \param thresh ...
!> \param start_sec_block ...
! **************************************************************************************************
   SUBROUTINE cp_fm_block_jacobi(matrix, eigenvectors, eigval, thresh, &
                                 start_sec_block)

      ! Calculates block diagonalizazion from full symmetric matrix
      ! It has its origin in cp_fm_syevx. This routine rotates only elements
      ! which are larger than a threshold thresh.
      ! start_sec_block is the start of the second block.
      ! IT DOES ONLY ONE SWEEP!

      ! - Creation (07.10.2002, Martin Fengler)
      ! - Cosmetics (05.04.06,MK)

      TYPE(cp_fm_type), POINTER                 :: eigenvectors, matrix
      REAL(KIND=dp), DIMENSION(:), INTENT(IN) :: eigval
      INTEGER, INTENT(IN)                       :: start_sec_block
      REAL(KIND=dp), INTENT(IN)               :: thresh

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_block_jacobi', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: handle
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, ev

      REAL(KIND=dp) :: tan_theta, tau, c, s
      INTEGER  :: q, p, N
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: c_ip


      ! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)


      N = matrix%matrix_struct%nrow_global ! Groesse der Matrix A, die bearbeitet werden soll

      ALLOCATE (c_ip(N)) ! Speicher fuer den lokalen Eigenwertvektor

      A => matrix%local_data ! Contains the Matrix to be worked on
      EV => eigenvectors%local_data ! Contains the eigenvectors up to blocksize: Rest ist Muell

      ! Start diagonalizing matrix

      tan_theta = 0.0_dp
      tau = 0.0_dp

      DO q = start_sec_block, N
         DO p = 1, (start_sec_block-1)

            IF (ABS(A(p, q)) > thresh) THEN

               tau = (eigval(q)-eigval(p))/(2.0_dp*A(p, q))

               tan_theta = SIGN(1.0_dp, tau)/(ABS(tau)+SQRT(1.0_dp+tau*tau))

               ! cos theta
               c = 1.0_dp/SQRT(1.0_dp+tan_theta*tan_theta)
               s = tan_theta*c

               ! Und jetzt noch die Eigenvektoren produzieren:
               ! Q * J
               !  Verstaendliche Version (bevor die BLAS-Aufrufe sie ersetzt haben)
               !  c_ip = c*EV(:,p) - s*EV(:,q)
               !  c_iq = s*EV(:,p) + c*EV(:,q)

               !  EV(:,p)=c_ip
               !  EV(:,q)=c_iq

               CALL dcopy(N, EV(1, p), 1, c_ip(1), 1)
               CALL dscal(N, c, EV(1, p), 1)
               CALL daxpy(N, -s, EV(1, q), 1, EV(1, p), 1)
               CALL dscal(N, c, EV(1, q), 1)
               CALL daxpy(N, s, c_ip(1), 1, EV(1, q), 1)

            END IF

         END DO
      END DO

      ! Release work storage

      DEALLOCATE (c_ip)


      CALL timestop(handle)

   END SUBROUTINE cp_fm_block_jacobi

! **************************************************************************************************
!> \brief General Eigenvalue Problem  AX = BXE
!>        Single option version: Cholesky decomposition of B
!> \param amatrix ...
!> \param bmatrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param work ...
! **************************************************************************************************
   SUBROUTINE cp_fm_geeig(amatrix, bmatrix, eigenvectors, eigenvalues, work)

      TYPE(cp_fm_type), POINTER                          :: amatrix, bmatrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)                        :: eigenvalues
      TYPE(cp_fm_type), POINTER                          :: work

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_geeig', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nao, nmo

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(amatrix, nrow_global=nao)
      nmo = SIZE(eigenvalues)
      ! Cholesky decompose S=U(T)U
      CALL cp_fm_cholesky_decompose(bmatrix)
      ! Invert to get U^(-1)
      CALL cp_fm_triangular_invert(bmatrix)
      ! Reduce to get U^(-T) * H * U^(-1)
      CALL cp_fm_triangular_multiply(bmatrix, amatrix, side="R")
      CALL cp_fm_triangular_multiply(bmatrix, amatrix, transpose_tr=.TRUE.)
      ! Diagonalize
      CALL choose_eigv_solver(matrix=amatrix, eigenvectors=work, &
                              eigenvalues=eigenvalues)
      ! Restore vectors C = U^(-1) * C*
      CALL cp_fm_triangular_multiply(bmatrix, work)
      CALL cp_fm_to_fm(work, eigenvectors, nmo)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_geeig

! **************************************************************************************************
!> \brief General Eigenvalue Problem  AX = BXE
!>        Use canonical diagonalization : U*s**(-1/2)
!> \param amatrix ...
!> \param bmatrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param work ...
!> \param epseig ...
! **************************************************************************************************
   SUBROUTINE cp_fm_geeig_canon(amatrix, bmatrix, eigenvectors, eigenvalues, work, epseig)

      TYPE(cp_fm_type), POINTER                          :: amatrix, bmatrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:)                        :: eigenvalues
      TYPE(cp_fm_type), POINTER                          :: work
      REAL(KIND=dp), INTENT(IN)                          :: epseig

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_geeig_canon', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, icol, irow, nao, nc, ncol, &
                                                            nmo, nx
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: seigval

      CALL timeset(routineN, handle)

      ! Test sizees
      CALL cp_fm_get_info(amatrix, nrow_global=nao)
      nmo = SIZE(eigenvalues)
      ALLOCATE (seigval(nao))

      ! Diagonalize -S matrix, this way the NULL space is at the end of the spectrum
      CALL cp_fm_scale(-1.0_dp, bmatrix)
      CALL choose_eigv_solver(matrix=bmatrix, eigenvectors=work, eigenvalues=seigval)
      seigval(:) = -seigval(:)
      nc = nao
      DO i = 1, nao
         IF (seigval(i) < epseig) THEN
            nc = i-1
            EXIT
         END IF
      END DO
      IF(.NOT.(nc /= 0))CALL cp__a("fm/cp_fm_diag.F",1471)

      IF (nc /= nao) THEN
         IF (nc < nmo) THEN
            ! Copy NULL space definition to last vectors of eigenvectors (if needed)
            ncol = nmo-nc
            CALL cp_fm_to_fm(work, eigenvectors, ncol, nc+1, nc+1)
         END IF
         ! Set NULL space in eigenvector matrix of S to zero
         DO icol = nc+1, nao
            DO irow = 1, nao
               CALL cp_fm_set_element(work, irow, icol, 0.0_dp)
            END DO
         END DO
         ! Set small eigenvalues to a dummy save value
         seigval(nc+1:nao) = 1.0_dp
      END IF
      ! calculate U*s**(-1/2)
      seigval(:) = 1.0_dp/SQRT(seigval(:))
      CALL cp_fm_column_scale(work, seigval)
      ! Reduce to get U^(-T) * H * U^(-1)
      CALL cp_fm_gemm("T", "N", nao, nao, nao, 1.0_dp, work, amatrix, 0.0_dp, bmatrix)
      CALL cp_fm_gemm("N", "N", nao, nao, nao, 1.0_dp, bmatrix, work, 0.0_dp, amatrix)
      IF (nc /= nao) THEN
         ! set diagonal values to save large value
         DO icol = nc+1, nao
            CALL cp_fm_set_element(amatrix, icol, icol, 10000.0_dp)
         END DO
      END IF
      ! Diagonalize
      CALL choose_eigv_solver(matrix=amatrix, eigenvectors=bmatrix, eigenvalues=eigenvalues)
      nx = MIN(nc, nmo)
      ! Restore vectors C = U^(-1) * C*
      CALL cp_fm_gemm("N", "N", nao, nx, nc, 1.0_dp, work, bmatrix, 0.0_dp, eigenvectors)

      DEALLOCATE (seigval)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_geeig_canon

END MODULE cp_fm_diag
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Interface to (sca)lapack for the Cholesky based procedures
!> \author  VW
!> \date    2009-11-09
!> \version 0.8
!>
!> <b>Modification history:</b>
!> - Created 2009-11-09
! **************************************************************************************************
MODULE cp_dbcsr_diag

   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_cfm_diag,                     ONLY: cp_cfm_heevd
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_release,&
                                              cp_cfm_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_info,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_cfm_to_dbcsr,&
                                              copy_dbcsr_to_cfm,&
                                              copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_syevx
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_dbcsr_diag'

   ! Public subroutines

   PUBLIC :: cp_dbcsr_syevd, &
             cp_dbcsr_syevx, &
             cp_dbcsr_heevd

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_syevd(matrix, eigenvectors, eigenvalues, para_env, blacs_env)

      ! Computes all eigenvalues and vectors of a real symmetric matrix
      ! should be quite a bit faster than syevx for that case
      ! especially in parallel with thightly clustered evals
      ! needs more workspace in the worst case, but much better distributed

      TYPE(cp_dbcsr_type)                                :: matrix, eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: eigenvalues
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_syevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nfullrows_total
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_eigenvectors, fm_matrix

      CALL timeset(routineN, handle)

      NULLIFY (fm_matrix, fm_eigenvectors, fm_struct)
      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nfullrows_total)

      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullrows_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix, fm_struct, name="fm_matrix")
      CALL cp_fm_create(fm_eigenvectors, fm_struct, name="fm_eigenvectors")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_fm(matrix, fm_matrix)

      CALL choose_eigv_solver(fm_matrix, fm_eigenvectors, eigenvalues)

      CALL copy_fm_to_dbcsr(fm_eigenvectors, eigenvectors)

      CALL cp_fm_release(fm_matrix)
      CALL cp_fm_release(fm_eigenvectors)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_syevd

! **************************************************************************************************
!> \brief   compute eigenvalues and optionally eigenvectors of a real symmetric matrix using scalapack.
!>          If eigenvectors are required this routine will replicate a full matrix on each CPU...
!>          if more than a handful of vectors are needed, use cp_dbcsr_syevd instead
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param neig ...
!> \param work_syevx ...
!> \param para_env ...
!> \param blacs_env ...
!> \par     matrix is supposed to be in upper triangular form, and overwritten by this routine
!>          neig   is the number of vectors needed (default all)
!>          work_syevx evec calculation only, is the fraction of the working buffer allowed (1.0 use full buffer)
!>                     reducing this saves time, but might cause the routine to fail
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_syevx(matrix, eigenvectors, eigenvalues, neig, work_syevx, &
                             para_env, blacs_env)

      ! Diagonalise the symmetric n by n matrix using the LAPACK library.

      TYPE(cp_dbcsr_type), POINTER                       :: matrix
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: eigenvalues
      INTEGER, INTENT(IN), OPTIONAL                      :: neig
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: work_syevx
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_syevx', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n, neig_local
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_eigenvectors, fm_matrix

      CALL timeset(routineN, handle)

      ! by default all
      CALL cp_dbcsr_get_info(matrix, nfullrows_total=n)
      neig_local = n
      IF (PRESENT(neig)) neig_local = neig
      IF (neig_local == 0) RETURN

      NULLIFY (fm_matrix, fm_eigenvectors, fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=n, &
                               ncol_global=n, para_env=para_env)
      CALL cp_fm_create(fm_matrix, fm_struct, name="fm_matrix")

      IF (PRESENT(eigenvectors)) THEN
         CALL cp_fm_create(fm_eigenvectors, fm_struct, name="fm_eigenvectors")
         CALL cp_fm_syevx(fm_matrix, fm_eigenvectors, eigenvalues, neig, work_syevx)
         CALL copy_fm_to_dbcsr(fm_eigenvectors, eigenvectors)
         CALL cp_fm_release(fm_eigenvectors)
      ELSE
         CALL cp_fm_syevx(fm_matrix, eigenvalues=eigenvalues, neig=neig, work_syevx=work_syevx)
      ENDIF

      CALL cp_fm_struct_release(fm_struct)
      CALL cp_fm_release(fm_matrix)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_syevx

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_heevd(matrix, eigenvectors, eigenvalues, para_env, blacs_env)

      TYPE(cp_dbcsr_type)                                :: matrix
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: eigenvectors
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: eigenvalues
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_heevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nfullrows_total
      TYPE(cp_cfm_type), POINTER                         :: fm_eigenvectors, fm_matrix
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct

      CALL timeset(routineN, handle)

      NULLIFY (fm_matrix, fm_eigenvectors, fm_struct)
      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nfullrows_total)

      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullrows_total, para_env=para_env)
      CALL cp_cfm_create(fm_matrix, fm_struct, name="fm_matrix")
      CALL cp_cfm_create(fm_eigenvectors, fm_struct, name="fm_eigenvectors")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_cfm(matrix, fm_matrix)

      CALL cp_cfm_heevd(fm_matrix, fm_eigenvectors, eigenvalues)

      CALL copy_cfm_to_dbcsr(fm_eigenvectors, eigenvectors)

      CALL cp_cfm_release(fm_matrix)
      CALL cp_cfm_release(fm_eigenvectors)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_heevd

END MODULE cp_dbcsr_diag
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basic linear algebra operations for full matrixes
!> \par History
!>      08.2002 splitted out of qs_blacs [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_gemm_interface
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_multiply,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm_bc,&
                                              copy_fm_to_dbcsr_bc
   USE cp_fm_basic_linalg,              ONLY: cp_fm_gemm
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_get_mm_type,&
                                              cp_fm_type
   USE input_constants,                 ONLY: do_dbcsr,&
                                              do_pdgemm
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_min
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_gemm_interface'

   PUBLIC :: cp_gemm

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param m ...
!> \param n ...
!> \param k ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param a_first_col ...
!> \param a_first_row ...
!> \param b_first_col ...
!> \param b_first_row ...
!> \param c_first_col ...
!> \param c_first_row ...
! **************************************************************************************************
   SUBROUTINE cp_gemm(transa, transb, m, n, k, alpha, matrix_a, matrix_b, beta, &
                      matrix_c, a_first_col, a_first_row, b_first_col, b_first_row, &
                      c_first_col, c_first_row)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      INTEGER, INTENT(IN)                                :: m, n, k
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(cp_fm_type), POINTER                          :: matrix_a, matrix_b
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(cp_fm_type), POINTER                          :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: a_first_col, a_first_row, b_first_col, &
                                                            b_first_row, c_first_col, c_first_row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_gemm', routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: my_trans
      INTEGER                                            :: handle, handle1, my_multi
      INTEGER, DIMENSION(:), POINTER                     :: a_col_loc, a_row_loc, b_col_loc, &
                                                            b_row_loc, c_col_loc, c_row_loc
      TYPE(cp_dbcsr_type)                                :: a_db, b_db, c_db

      CALL timeset(routineN, handle)

      my_multi = cp_fm_get_mm_type()
      ! catch the special case that matrices have different blocking
      ! SCALAPACK can deal with it but dbcsr doesn't like it
      CALL cp_fm_get_info(matrix_a, nrow_locals=a_row_loc, ncol_locals=a_col_loc)
      CALL cp_fm_get_info(matrix_b, nrow_locals=b_row_loc, ncol_locals=b_col_loc)
      CALL cp_fm_get_info(matrix_c, nrow_locals=c_row_loc, ncol_locals=c_col_loc)
      IF (PRESENT(a_first_row)) my_multi = do_pdgemm
      IF (PRESENT(a_first_col)) my_multi = do_pdgemm
      IF (PRESENT(b_first_row)) my_multi = do_pdgemm
      IF (PRESENT(b_first_col)) my_multi = do_pdgemm
      IF (PRESENT(c_first_row)) my_multi = do_pdgemm
      IF (PRESENT(c_first_col)) my_multi = do_pdgemm

      my_trans = transa; CALL uppercase(my_trans)
      IF (my_trans == 'T') THEN
         CALL cp_fm_get_info(matrix_a, nrow_locals=a_col_loc, ncol_locals=a_row_loc)
      END IF

      my_trans = transb; CALL uppercase(my_trans)
      IF (my_trans == 'T') THEN
         CALL cp_fm_get_info(matrix_b, nrow_locals=b_col_loc, ncol_locals=b_row_loc)
      END IF

      IF (my_multi .NE. do_pdgemm) THEN
         IF (SIZE(a_row_loc) == SIZE(c_row_loc)) THEN
            IF (ANY(a_row_loc-c_row_loc .NE. 0)) my_multi = do_pdgemm
         ELSE
            my_multi = do_pdgemm
         END IF
      END IF
      IF (my_multi .NE. do_pdgemm) THEN
         IF (SIZE(b_col_loc) == SIZE(c_col_loc)) THEN
            IF (ANY(b_col_loc-c_col_loc .NE. 0)) my_multi = do_pdgemm
         ELSE
            my_multi = do_pdgemm
         END IF
      END IF
      IF (my_multi .NE. do_pdgemm) THEN
         IF (SIZE(a_col_loc) == SIZE(b_row_loc)) THEN
            IF (ANY(a_col_loc-b_row_loc .NE. 0)) my_multi = do_pdgemm
         ELSE
            my_multi = do_pdgemm
         END IF
      END IF

      ! IMPORTANT do_pdgemm is lowest value. If one processor has it set make all do pdgemm
      IF (cp_fm_get_mm_type() .NE. do_pdgemm) CALL mp_min(my_multi, matrix_a%matrix_struct%para_env%group)

      SELECT CASE (my_multi)
      CASE (do_pdgemm)
         CALL timeset("cp_gemm_fm_gemm", handle1)
         CALL cp_fm_gemm(transa, transb, m, n, k, alpha, matrix_a, matrix_b, beta, matrix_c, &
                         a_first_col=a_first_col, &
                         a_first_row=a_first_row, &
                         b_first_col=b_first_col, &
                         b_first_row=b_first_row, &
                         c_first_col=c_first_col, &
                         c_first_row=c_first_row)
         CALL timestop(handle1)
      CASE (do_dbcsr)
         CALL timeset("cp_gemm_dbcsr_mm", handle1)
         CALL copy_fm_to_dbcsr_bc(matrix_a, a_db)
         CALL copy_fm_to_dbcsr_bc(matrix_b, b_db)
         CALL copy_fm_to_dbcsr_bc(matrix_c, c_db)

         CALL cp_dbcsr_multiply(transa, transb, alpha, a_db, b_db, beta, c_db, last_k=k)

         CALL copy_dbcsr_to_fm_bc(c_db, matrix_c)
         CALL cp_dbcsr_release(a_db)
         CALL cp_dbcsr_release(b_db)
         CALL cp_dbcsr_release(c_db)
         CALL timestop(handle1)
      END SELECT
      CALL timestop(handle)

   END SUBROUTINE cp_gemm

END MODULE cp_gemm_interface
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains methods used in the context of density fitting
!> \par History
!>      04.2008 created [Manuel Guidon]
!>      02.2013 moved from admm_methods
!> \author Manuel Guidon
! **************************************************************************************************
MODULE admm_utils
   USE admm_types,                      ONLY: admm_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix, cp_dbcsr_distribution, cp_dbcsr_get_data_size, &
        cp_dbcsr_get_data_type, cp_dbcsr_init, cp_dbcsr_row_block_sizes, cp_dbcsr_set, &
        cp_dbcsr_type, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE input_constants,                 ONLY: do_admm_purify_cauchy,&
                                              do_admm_purify_cauchy_subspace,&
                                              do_admm_purify_mo_diag,&
                                              do_admm_purify_mo_no_diag,&
                                              do_admm_purify_none
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: admm_correct_for_eigenvalues, &
             admm_uncorrect_for_eigenvalues

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'admm_utils'

!***

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ispin ...
!> \param admm_env ...
!> \param ks_matrix ...
! **************************************************************************************************
   SUBROUTINE admm_correct_for_eigenvalues(ispin, admm_env, ks_matrix)
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_dbcsr_type), POINTER                       :: ks_matrix

      INTEGER                                            :: nao_aux_fit, nao_orb
      TYPE(cp_dbcsr_type), POINTER                       :: work

      nao_aux_fit = admm_env%nao_aux_fit
      nao_orb = admm_env%nao_orb

      IF (.NOT. admm_env%block_dm) THEN
         SELECT CASE (admm_env%purification_method)
         CASE (do_admm_purify_cauchy_subspace)
            !* remove what has been added and add the correction
            NULLIFY (work)
            ALLOCATE (work)
            CALL cp_dbcsr_init(work)
            CALL cp_dbcsr_create(work, 'work', &
                                 cp_dbcsr_distribution(ks_matrix), dbcsr_type_symmetric, cp_dbcsr_row_block_sizes(ks_matrix), &
                                 cp_dbcsr_col_block_sizes(ks_matrix), &
                                 cp_dbcsr_get_data_size(ks_matrix), &
                                 cp_dbcsr_get_data_type(ks_matrix))

            CALL cp_dbcsr_copy(work, ks_matrix)
            CALL cp_dbcsr_set(work, 0.0_dp)
            CALL copy_fm_to_dbcsr(admm_env%ks_to_be_merged(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, -1.0_dp)

            ! ** calculate A^T*H_tilde*A
            CALL cp_gemm('N', 'N', nao_aux_fit, nao_orb, nao_aux_fit, &
                         1.0_dp, admm_env%K(ispin)%matrix, admm_env%A, 0.0_dp, &
                         admm_env%work_aux_orb)
            CALL cp_gemm('T', 'N', nao_orb, nao_orb, nao_aux_fit, &
                         1.0_dp, admm_env%A, admm_env%work_aux_orb, 0.0_dp, &
                         admm_env%H_corr(ispin)%matrix)

            CALL copy_fm_to_dbcsr(admm_env%H_corr(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, 1.0_dp)
            CALL cp_dbcsr_deallocate_matrix(work)

         CASE (do_admm_purify_mo_diag)
            !* remove what has been added and add the correction
            NULLIFY (work)
            ALLOCATE (work)
            CALL cp_dbcsr_init(work)
            CALL cp_dbcsr_create(work, 'work', &
                                 cp_dbcsr_distribution(ks_matrix), dbcsr_type_symmetric, cp_dbcsr_row_block_sizes(ks_matrix), &
                                 cp_dbcsr_col_block_sizes(ks_matrix), cp_dbcsr_get_data_size(ks_matrix), &
                                 cp_dbcsr_get_data_type(ks_matrix))

            CALL cp_dbcsr_copy(work, ks_matrix)
            CALL cp_dbcsr_set(work, 0.0_dp)
            CALL copy_fm_to_dbcsr(admm_env%ks_to_be_merged(ispin)%matrix, work, keep_sparsity=.TRUE.)

            ! ** calculate A^T*H_tilde*A
            CALL cp_gemm('N', 'N', nao_aux_fit, nao_orb, nao_aux_fit, &
                         1.0_dp, admm_env%K(ispin)%matrix, admm_env%A, 0.0_dp, &
                         admm_env%work_aux_orb)
            CALL cp_gemm('T', 'N', nao_orb, nao_orb, nao_aux_fit, &
                         1.0_dp, admm_env%A, admm_env%work_aux_orb, 0.0_dp, &
                         admm_env%H_corr(ispin)%matrix)

            CALL copy_fm_to_dbcsr(admm_env%H_corr(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, 1.0_dp)
            CALL cp_dbcsr_deallocate_matrix(work)

         CASE (do_admm_purify_mo_no_diag, do_admm_purify_none, do_admm_purify_cauchy)
            ! do nothing
         END SELECT
      ENDIF

   END SUBROUTINE admm_correct_for_eigenvalues

! **************************************************************************************************
!> \brief ...
!> \param ispin ...
!> \param admm_env ...
!> \param ks_matrix ...
! **************************************************************************************************
   SUBROUTINE admm_uncorrect_for_eigenvalues(ispin, admm_env, ks_matrix)
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_dbcsr_type), POINTER                       :: ks_matrix

      INTEGER                                            :: nao_aux_fit, nao_orb
      TYPE(cp_dbcsr_type), POINTER                       :: work

      nao_aux_fit = admm_env%nao_aux_fit
      nao_orb = admm_env%nao_orb

      IF (.NOT. admm_env%block_dm) THEN
         SELECT CASE (admm_env%purification_method)
         CASE (do_admm_purify_cauchy_subspace)
            !* remove what has been added and add the correction
            NULLIFY (work)
            ALLOCATE (work)
            CALL cp_dbcsr_init(work)
            CALL cp_dbcsr_create(work, 'work', &
                                 cp_dbcsr_distribution(ks_matrix), dbcsr_type_symmetric, cp_dbcsr_row_block_sizes(ks_matrix), &
                                 cp_dbcsr_col_block_sizes(ks_matrix), cp_dbcsr_get_data_size(ks_matrix), &
                                 cp_dbcsr_get_data_type(ks_matrix))

            CALL cp_dbcsr_copy(work, ks_matrix)
            CALL cp_dbcsr_set(work, 0.0_dp)
            CALL copy_fm_to_dbcsr(admm_env%H_corr(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, -1.0_dp)

            CALL copy_fm_to_dbcsr(admm_env%H_corr(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_set(work, 0.0_dp)
            CALL copy_fm_to_dbcsr(admm_env%ks_to_be_merged(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, 1.0_dp)
            CALL cp_dbcsr_deallocate_matrix(work)

         CASE (do_admm_purify_mo_diag)
            NULLIFY (work)
            ALLOCATE (work)
            CALL cp_dbcsr_init(work)
            CALL cp_dbcsr_create(work, 'work', &
                                 cp_dbcsr_distribution(ks_matrix), dbcsr_type_symmetric, cp_dbcsr_row_block_sizes(ks_matrix), &
                                 cp_dbcsr_col_block_sizes(ks_matrix), cp_dbcsr_get_data_size(ks_matrix), &
                                 cp_dbcsr_get_data_type(ks_matrix))

            CALL cp_dbcsr_copy(work, ks_matrix)
            CALL cp_dbcsr_set(work, 0.0_dp)

            CALL copy_fm_to_dbcsr(admm_env%H_corr(ispin)%matrix, work, keep_sparsity=.TRUE.)

            CALL cp_dbcsr_add(ks_matrix, work, 1.0_dp, -1.0_dp)
            CALL cp_dbcsr_deallocate_matrix(work)

         CASE (do_admm_purify_mo_no_diag, do_admm_purify_none, do_admm_purify_cauchy)
            ! do nothing
         END SELECT
      ENDIF
   END SUBROUTINE admm_uncorrect_for_eigenvalues

END MODULE admm_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief collects routines that perform operations directly related to MOs
!> \note
!>      first version : most routines imported
!> \author Joost VandeVondele (2003-08)
! **************************************************************************************************
MODULE qs_mo_methods
   USE admm_types,                      ONLY: admm_type
   USE admm_utils,                      ONLY: admm_correct_for_eigenvalues,&
                                              admm_uncorrect_for_eigenvalues
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_diag,                   ONLY: cp_dbcsr_syevd,&
                                              cp_dbcsr_syevx
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, cp_dbcsr_multiply, &
        cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_release_p, cp_dbcsr_scale_by_vector, &
        cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_m_by_n_from_template,&
                                              cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_syrk,&
                                              cp_fm_triangular_multiply
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_power,&
                                              cp_fm_syevx
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_max
   USE physcon,                         ONLY: evolt
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mo_methods'

   PUBLIC :: make_basis_simple, make_basis_cholesky, make_basis_sv, make_basis_sm, &
             make_basis_lowdin, calculate_density_matrix, calculate_subspace_eigenvalues, &
             calculate_orthonormality, calculate_magnitude, make_mo_eig

   INTERFACE calculate_density_matrix
      MODULE PROCEDURE calculate_dm_sparse
   END INTERFACE

   INTERFACE calculate_subspace_eigenvalues
      MODULE PROCEDURE subspace_eigenvalues_ks_fm
      MODULE PROCEDURE subspace_eigenvalues_ks_dbcsr
   END INTERFACE

   INTERFACE make_basis_sv
      MODULE PROCEDURE make_basis_sv_fm
      MODULE PROCEDURE make_basis_sv_dbcsr
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief returns an S-orthonormal basis v (v^T S v ==1)
!> \param vmatrix ...
!> \param ncol ...
!> \param matrix_s ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE make_basis_sm(vmatrix, ncol, matrix_s)
      TYPE(cp_fm_type), POINTER                          :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_sm', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: overlap_vv, svmatrix

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(matrix=vmatrix, nrow_global=n, ncol_global=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",105,"Wrong ncol value")

      CALL cp_fm_create(svmatrix, vmatrix%matrix_struct, "SV")
      CALL cp_dbcsr_sm_fm_multiply(matrix_s, vmatrix, svmatrix, ncol)

      NULLIFY (fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, ncol_global=ncol, &
                               para_env=vmatrix%matrix_struct%para_env, &
                               context=vmatrix%matrix_struct%context)
      CALL cp_fm_create(overlap_vv, fm_struct_tmp, "overlap_vv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_gemm('T', 'N', ncol, ncol, n, rone, vmatrix, svmatrix, rzero, overlap_vv)
      CALL cp_fm_cholesky_decompose(overlap_vv)
      CALL cp_fm_triangular_multiply(overlap_vv, vmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)

      CALL cp_fm_release(overlap_vv)
      CALL cp_fm_release(svmatrix)

      CALL timestop(handle)

   END SUBROUTINE make_basis_sm

! **************************************************************************************************
!> \brief returns an S-orthonormal basis v and the corresponding matrix S*v as well
!> \param vmatrix ...
!> \param ncol ...
!> \param svmatrix ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE make_basis_sv_fm(vmatrix, ncol, svmatrix)

      TYPE(cp_fm_type), POINTER                          :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol
      TYPE(cp_fm_type), POINTER                          :: svmatrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_sv_fm', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: overlap_vv

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)
      NULLIFY (fm_struct_tmp)

      CALL cp_fm_get_info(matrix=vmatrix, nrow_global=n, ncol_global=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",156,"Wrong ncol value")

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, ncol_global=ncol, &
                               para_env=vmatrix%matrix_struct%para_env, &
                               context=vmatrix%matrix_struct%context)
      CALL cp_fm_create(overlap_vv, fm_struct_tmp, "overlap_vv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_gemm('T', 'N', ncol, ncol, n, rone, vmatrix, svmatrix, rzero, overlap_vv)
      CALL cp_fm_cholesky_decompose(overlap_vv)
      CALL cp_fm_triangular_multiply(overlap_vv, vmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)
      CALL cp_fm_triangular_multiply(overlap_vv, svmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)

      CALL cp_fm_release(overlap_vv)

      CALL timestop(handle)

   END SUBROUTINE make_basis_sv_fm

! **************************************************************************************************
!> \brief ...
!> \param vmatrix ...
!> \param ncol ...
!> \param svmatrix ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE make_basis_sv_dbcsr(vmatrix, ncol, svmatrix, para_env, blacs_env)

      TYPE(cp_dbcsr_type)                                :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol
      TYPE(cp_dbcsr_type)                                :: svmatrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_sv_dbcsr', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_svmatrix, fm_vmatrix, overlap_vv

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)

      !CALL cp_fm_get_info(matrix=vmatrix,nrow_global=n,ncol_global=ncol_global)
      CALL cp_dbcsr_get_info(vmatrix, nfullrows_total=n, nfullcols_total=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",205,"Wrong ncol value")

      CALL cp_fm_struct_create(fm_struct_tmp, context=blacs_env, nrow_global=ncol, &
                               ncol_global=ncol, para_env=para_env)
      CALL cp_fm_create(overlap_vv, fm_struct_tmp, name="fm_overlap_vv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_fm_struct_create(fm_struct_tmp, context=blacs_env, nrow_global=n, &
                               ncol_global=ncol_global, para_env=para_env)
      CALL cp_fm_create(fm_vmatrix, fm_struct_tmp, name="fm_vmatrix")
      CALL cp_fm_create(fm_svmatrix, fm_struct_tmp, name="fm_svmatrix")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL copy_dbcsr_to_fm(vmatrix, fm_vmatrix)
      CALL copy_dbcsr_to_fm(svmatrix, fm_svmatrix)

      CALL cp_gemm('T', 'N', ncol, ncol, n, rone, fm_vmatrix, fm_svmatrix, rzero, overlap_vv)
      CALL cp_fm_cholesky_decompose(overlap_vv)
      CALL cp_fm_triangular_multiply(overlap_vv, fm_vmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)
      CALL cp_fm_triangular_multiply(overlap_vv, fm_svmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)

      CALL copy_fm_to_dbcsr(fm_vmatrix, vmatrix)
      CALL copy_fm_to_dbcsr(fm_svmatrix, svmatrix)

      CALL cp_fm_release(overlap_vv)
      CALL cp_fm_release(fm_vmatrix)
      CALL cp_fm_release(fm_svmatrix)

      CALL timestop(handle)

   END SUBROUTINE make_basis_sv_dbcsr

! **************************************************************************************************
!> \brief return a set of S orthonormal vectors (C^T S C == 1) where
!>      the cholesky decomposed form of S is passed as an argument
!> \param vmatrix ...
!> \param ncol ...
!> \param ortho cholesky decomposed S matrix
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \note
!>      if the cholesky decomposed S matrix is not available
!>      use make_basis_sm since this is much faster than computing the
!>      cholesky decomposition of S
! **************************************************************************************************
   SUBROUTINE make_basis_cholesky(vmatrix, ncol, ortho)

      TYPE(cp_fm_type), POINTER                          :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol
      TYPE(cp_fm_type), POINTER                          :: ortho

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_cholesky', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: overlap_vv

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)
      NULLIFY (fm_struct_tmp)

      CALL cp_fm_get_info(matrix=vmatrix, nrow_global=n, ncol_global=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",270,"Wrong ncol value")

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, ncol_global=ncol, &
                               para_env=vmatrix%matrix_struct%para_env, &
                               context=vmatrix%matrix_struct%context)
      CALL cp_fm_create(overlap_vv, fm_struct_tmp, "overlap_vv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_fm_triangular_multiply(ortho, vmatrix, n_cols=ncol)
      CALL cp_fm_syrk('U', 'T', n, rone, vmatrix, 1, 1, rzero, overlap_vv)
      CALL cp_fm_cholesky_decompose(overlap_vv)
      CALL cp_fm_triangular_multiply(overlap_vv, vmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)
      CALL cp_fm_triangular_multiply(ortho, vmatrix, n_cols=ncol, invert_tr=.TRUE.)

      CALL cp_fm_release(overlap_vv)

      CALL timestop(handle)

   END SUBROUTINE make_basis_cholesky

! **************************************************************************************************
!> \brief return a set of S orthonormal vectors (C^T S C == 1) where
!>      a Loedwin transformation is applied to keep the rotated vectors as close
!>      as possible to the original ones
!> \param vmatrix ...
!> \param ncol ...
!> \param matrix_s ...
!> \param
!> \par History
!>      05.2009 created [MI]
!> \note
! **************************************************************************************************
   SUBROUTINE make_basis_lowdin(vmatrix, ncol, matrix_s)

      TYPE(cp_fm_type), POINTER                          :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_lowdin', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global, ndep
      REAL(dp)                                           :: threshold
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: csc, sc, work

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)
      NULLIFY (fm_struct_tmp)
      threshold = 1.0E-7_dp
      CALL cp_fm_get_info(matrix=vmatrix, nrow_global=n, ncol_global=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",323,"Wrong ncol value")

      CALL cp_fm_create(sc, vmatrix%matrix_struct, "SC")
      CALL cp_dbcsr_sm_fm_multiply(matrix_s, vmatrix, sc, ncol)

      NULLIFY (fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, ncol_global=ncol, &
                               para_env=vmatrix%matrix_struct%para_env, &
                               context=vmatrix%matrix_struct%context)
      CALL cp_fm_create(csc, fm_struct_tmp, "csc")
      CALL cp_fm_create(work, fm_struct_tmp, "work")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_gemm('T', 'N', ncol, ncol, n, rone, vmatrix, sc, rzero, csc)
      CALL cp_fm_power(csc, work, -0.5_dp, threshold, ndep)
      CALL cp_gemm('N', 'N', n, ncol, ncol, rone, vmatrix, csc, rzero, sc)
      CALL cp_fm_to_fm(sc, vmatrix, ncol, 1, 1)

      CALL cp_fm_release(csc)
      CALL cp_fm_release(sc)
      CALL cp_fm_release(work)

      CALL timestop(handle)

   END SUBROUTINE make_basis_lowdin

! **************************************************************************************************
!> \brief given a set of vectors, return an orthogonal (C^T C == 1) set
!>      spanning the same space (notice, only for cases where S==1)
!> \param vmatrix ...
!> \param ncol ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE make_basis_simple(vmatrix, ncol)

      TYPE(cp_fm_type), POINTER                          :: vmatrix
      INTEGER, INTENT(IN)                                :: ncol

      CHARACTER(LEN=*), PARAMETER :: routineN = 'make_basis_simple', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, n, ncol_global
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: overlap_vv

      IF (ncol .EQ. 0) RETURN

      CALL timeset(routineN, handle)

      NULLIFY (fm_struct_tmp)

      CALL cp_fm_get_info(matrix=vmatrix, nrow_global=n, ncol_global=ncol_global)
      IF (ncol .GT. ncol_global) CALL cp__b("qs_mo_methods.F",377,"Wrong ncol value")

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, ncol_global=ncol, &
                               para_env=vmatrix%matrix_struct%para_env, &
                               context=vmatrix%matrix_struct%context)
      CALL cp_fm_create(overlap_vv, fm_struct_tmp, "overlap_vv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_gemm('T', 'N', ncol, ncol, n, rone, vmatrix, vmatrix, rzero, overlap_vv)
      CALL cp_fm_cholesky_decompose(overlap_vv)
      CALL cp_fm_triangular_multiply(overlap_vv, vmatrix, n_cols=ncol, side='R', invert_tr=.TRUE.)

      CALL cp_fm_release(overlap_vv)

      CALL timestop(handle)

   END SUBROUTINE make_basis_simple

! **************************************************************************************************
!> \brief   Calculate the density matrix
!> \param mo_set ...
!> \param density_matrix ...
!> \param use_dbcsr ...
!> \param retain_sparsity ...
!> \date    06.2002
!> \par History
!>       - Fractional occupied orbitals (MK)
!> \author  Joost VandeVondele
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE calculate_dm_sparse(mo_set, density_matrix, use_dbcsr, retain_sparsity)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: density_matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_dbcsr, retain_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_dm_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_retain_sparsity, my_use_dbcsr
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_dbcsr_type)                                :: dbcsr_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_tmp

      CALL timeset(routineN, handle)

      my_use_dbcsr = .FALSE.
      IF (PRESENT(use_dbcsr)) my_use_dbcsr = use_dbcsr
      my_retain_sparsity = .TRUE.
      IF (PRESENT(retain_sparsity)) my_retain_sparsity = retain_sparsity
      IF (my_use_dbcsr) THEN
         IF (.NOT. ASSOCIATED(mo_set%mo_coeff_b)) THEN
            CALL cp__b("qs_mo_methods.F",430,"mo_coeff_b NOT ASSOCIATED")
         END IF
      END IF

      CALL cp_dbcsr_set(density_matrix, 0.0_dp)

      IF (.NOT. mo_set%uniform_occupation) THEN ! not all orbitals 1..homo are equally occupied

         IF (my_use_dbcsr) THEN
            CALL cp_dbcsr_init(dbcsr_tmp)
            CALL cp_dbcsr_copy(dbcsr_tmp, mo_set%mo_coeff_b)
            CALL cp_dbcsr_scale_by_vector(dbcsr_tmp, mo_set%occupation_numbers(1:mo_set%homo), &
                                          side='right')
            CALL cp_dbcsr_multiply("N", "T", 1.0_dp, mo_set%mo_coeff_b, dbcsr_tmp, &
                                   1.0_dp, density_matrix, retain_sparsity=my_retain_sparsity, &
                                   last_k=mo_set%homo)
            CALL cp_dbcsr_release(dbcsr_tmp)
         ELSE
            NULLIFY (fm_tmp)
            CALL cp_fm_create(fm_tmp, mo_set%mo_coeff%matrix_struct)
            CALL cp_fm_to_fm(mo_set%mo_coeff, fm_tmp)
            CALL cp_fm_column_scale(fm_tmp, mo_set%occupation_numbers(1:mo_set%homo))
            alpha = 1.0_dp
            CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix, &
                                       matrix_v=mo_set%mo_coeff, &
                                       matrix_g=fm_tmp, &
                                       ncol=mo_set%homo, &
                                       alpha=alpha)
            CALL cp_fm_release(fm_tmp)
         ENDIF
      ELSE
         IF (my_use_dbcsr) THEN
            CALL cp_dbcsr_multiply("N", "T", mo_set%maxocc, mo_set%mo_coeff_b, mo_set%mo_coeff_b, &
                                   1.0_dp, density_matrix, retain_sparsity=my_retain_sparsity, &
                                   last_k=mo_set%homo)
         ELSE
            alpha = mo_set%maxocc
            CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix, &
                                       matrix_v=mo_set%mo_coeff, &
                                       ncol=mo_set%homo, &
                                       alpha=alpha)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calculate_dm_sparse

! **************************************************************************************************
!> \brief computes ritz values of a set of orbitals given a ks_matrix
!>      rotates the orbitals into eigenstates depending on do_rotation
!>      writes the evals to the screen depending on ionode/scr
!> \param orbitals S-orthonormal orbitals
!> \param ks_matrix Kohn-Sham matrix
!> \param evals_arg optional, filled with the evals
!> \param ionode /scr  : if present write to unit scr where ionode
!> \param scr ...
!> \param do_rotation optional rotate orbitals (default=.TRUE.)
!>        note that rotating the orbitals is slower
!> \param co_rotate an optional set of orbitals rotated by the same rotation matrix
!> \param co_rotate_dbcsr ...
!> \par History
!>      08.2004 documented and added do_rotation [Joost VandeVondele]
!>      09.2008 only compute eigenvalues if rotation is not needed
! **************************************************************************************************
   SUBROUTINE subspace_eigenvalues_ks_fm(orbitals, ks_matrix, evals_arg, ionode, scr, &
                                         do_rotation, co_rotate, co_rotate_dbcsr)

      TYPE(cp_fm_type), POINTER                          :: orbitals
      TYPE(cp_dbcsr_type), POINTER                       :: ks_matrix
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: evals_arg
      LOGICAL, INTENT(IN), OPTIONAL                      :: ionode
      INTEGER, INTENT(IN), OPTIONAL                      :: scr
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_rotation
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: co_rotate
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: co_rotate_dbcsr

      CHARACTER(len=*), PARAMETER :: routineN = 'subspace_eigenvalues_ks_fm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ncol_global, nrow_global
      LOGICAL                                            :: compute_evecs, do_rotation_local
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: e_vectors, h_block, weighted_vectors, &
                                                            weighted_vectors2

      CALL timeset(routineN, handle)

      do_rotation_local = .TRUE.
      IF (PRESENT(do_rotation)) do_rotation_local = do_rotation

      NULLIFY (weighted_vectors, weighted_vectors2, h_block, e_vectors, fm_struct_tmp)
      CALL cp_fm_get_info(matrix=orbitals, &
                          ncol_global=ncol_global, &
                          nrow_global=nrow_global)

      IF (do_rotation_local) THEN
         compute_evecs = .TRUE.
      ELSE
         ! this would be the logical choice if syevx computing only evals were faster than syevd computing evecs and evals.
         compute_evecs = .FALSE.
         ! this is not the case, so lets compute evecs always
         compute_evecs = .TRUE.
      ENDIF

      IF (ncol_global .GT. 0) THEN

         ALLOCATE (evals(ncol_global))

         CALL cp_fm_create(weighted_vectors, orbitals%matrix_struct, "weighted_vectors")
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol_global, ncol_global=ncol_global, &
                                  para_env=orbitals%matrix_struct%para_env, &
                                  context=orbitals%matrix_struct%context)
         CALL cp_fm_create(h_block, fm_struct_tmp, name="h block")
         IF (compute_evecs) THEN
            CALL cp_fm_create(e_vectors, fm_struct_tmp, name="e vectors")
         ENDIF
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! h subblock and diag
         CALL cp_dbcsr_sm_fm_multiply(ks_matrix, orbitals, weighted_vectors, ncol_global)

         CALL cp_gemm('T', 'N', ncol_global, ncol_global, nrow_global, 1.0_dp, &
                      orbitals, weighted_vectors, 0.0_dp, h_block)

         ! if eigenvectors are required, go for syevd, otherwise only compute eigenvalues
         IF (compute_evecs) THEN
            CALL choose_eigv_solver(h_block, e_vectors, evals)
         ELSE
            CALL cp_fm_syevx(h_block, eigenvalues=evals)
         ENDIF

         ! rotate the orbitals
         IF (do_rotation_local) THEN
            CALL cp_gemm('N', 'N', nrow_global, ncol_global, ncol_global, 1.0_dp, &
                         orbitals, e_vectors, 0.0_dp, weighted_vectors)
            CALL cp_fm_to_fm(weighted_vectors, orbitals)
            IF (PRESENT(co_rotate)) THEN
               IF (ASSOCIATED(co_rotate)) THEN
                  CALL cp_gemm('N', 'N', nrow_global, ncol_global, ncol_global, 1.0_dp, &
                               co_rotate, e_vectors, 0.0_dp, weighted_vectors)
                  CALL cp_fm_to_fm(weighted_vectors, co_rotate)
               ENDIF
            ENDIF
            IF (PRESENT(co_rotate_dbcsr)) THEN
               IF (ASSOCIATED(co_rotate_dbcsr)) THEN
                  CALL cp_fm_create(weighted_vectors2, orbitals%matrix_struct, "weighted_vectors")
                  CALL copy_dbcsr_to_fm(co_rotate_dbcsr, weighted_vectors2)
                  CALL cp_gemm('N', 'N', nrow_global, ncol_global, ncol_global, 1.0_dp, &
                               weighted_vectors2, e_vectors, 0.0_dp, weighted_vectors)
                  CALL copy_fm_to_dbcsr(weighted_vectors, co_rotate_dbcsr)
                  CALL cp_fm_release(weighted_vectors2)
               ENDIF
            ENDIF
         ENDIF

         ! give output
         IF (PRESENT(evals_arg)) THEN
            evals_arg(:) = evals(:)
         ENDIF

         IF (PRESENT(ionode) .OR. PRESENT(scr)) THEN
            IF (.NOT. PRESENT(ionode)) CALL cp__b("qs_mo_methods.F",593,"IONODE?")
            IF (.NOT. PRESENT(scr)) CALL cp__b("qs_mo_methods.F",594,"SCR?")
            IF (ionode) THEN
               DO i = 1, ncol_global, 4
                  j = MIN(3, ncol_global-i)
                  SELECT CASE (j)
                  CASE (3)
                     WRITE (scr, '(1X,4F16.8)') evals(i:i+j)
                  CASE (2)
                     WRITE (scr, '(1X,3F16.8)') evals(i:i+j)
                  CASE (1)
                     WRITE (scr, '(1X,2F16.8)') evals(i:i+j)
                  CASE (0)
                     WRITE (scr, '(1X,1F16.8)') evals(i:i+j)
                  END SELECT
               ENDDO
            ENDIF
         ENDIF

         CALL cp_fm_release(weighted_vectors)
         CALL cp_fm_release(h_block)
         IF (compute_evecs) THEN
            CALL cp_fm_release(e_vectors)
         ENDIF

         DEALLOCATE (evals)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE subspace_eigenvalues_ks_fm

! **************************************************************************************************
!> \brief ...
!> \param orbitals ...
!> \param ks_matrix ...
!> \param evals_arg ...
!> \param ionode ...
!> \param scr ...
!> \param do_rotation ...
!> \param co_rotate ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE subspace_eigenvalues_ks_dbcsr(orbitals, ks_matrix, evals_arg, ionode, scr, &
                                            do_rotation, co_rotate, para_env, blacs_env)

      TYPE(cp_dbcsr_type), POINTER                       :: orbitals, ks_matrix
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: evals_arg
      LOGICAL, INTENT(IN), OPTIONAL                      :: ionode
      INTEGER, INTENT(IN), OPTIONAL                      :: scr
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_rotation
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: co_rotate
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'subspace_eigenvalues_ks_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ncol_global, nrow_global
      LOGICAL                                            :: compute_evecs, do_rotation_local
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals
      TYPE(cp_dbcsr_type), POINTER                       :: e_vectors, h_block, weighted_vectors

      CALL timeset(routineN, handle)

      do_rotation_local = .TRUE.
      IF (PRESENT(do_rotation)) do_rotation_local = do_rotation

      NULLIFY (e_vectors, h_block, weighted_vectors)

      CALL cp_dbcsr_get_info(matrix=orbitals, &
                             nfullcols_total=ncol_global, &
                             nfullrows_total=nrow_global)

      IF (do_rotation_local) THEN
         compute_evecs = .TRUE.
      ELSE
         ! this would be the logical choice if syevx computing only evals were faster than syevd computing evecs and evals.
         compute_evecs = .FALSE.
         ! this is not the case, so lets compute evecs always
         compute_evecs = .TRUE.
      ENDIF

      IF (ncol_global .GT. 0) THEN

         ALLOCATE (evals(ncol_global))

         CALL cp_dbcsr_init_p(weighted_vectors)
         CALL cp_dbcsr_copy(weighted_vectors, orbitals, name="weighted_vectors")

         CALL cp_dbcsr_init_p(h_block)
         CALL cp_dbcsr_m_by_n_from_template(h_block, template=orbitals, m=ncol_global, n=ncol_global, &
                                            sym=dbcsr_type_no_symmetry)

!!!!!!!!!!!!!!XXXXXXXXXXXXXXXXXXX!!!!!!!!!!!!!
         !CALL cp_fm_create(h_block,fm_struct_tmp, name="h block")
         IF (compute_evecs) THEN
            CALL cp_dbcsr_init_p(e_vectors)
            CALL cp_dbcsr_m_by_n_from_template(e_vectors, template=orbitals, m=ncol_global, n=ncol_global, &
                                               sym=dbcsr_type_no_symmetry)
!           CALL cp_dbcsr_create(e_vectors, "e_vectors", dist, dbcsr_type_no_symmetry,&
!                row_blk_size, col_blk_size, 0, 0, dbcsr_type_real_default)
            !CALL cp_fm_create(e_vectors,fm_struct_tmp, name="e vectors")
         ENDIF
!        CALL cp_dbcsr_distribution_release (dist)
!        CALL array_release (row_blk_size);CALL array_release (col_blk_size)
         !CALL cp_fm_struct_release(fm_struct_tmp)

         ! h subblock and diag
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, ks_matrix, orbitals, &
                                0.0_dp, weighted_vectors)
         !CALL cp_dbcsr_sm_fm_multiply(ks_matrix,orbitals,weighted_vectors, ncol_global)

         CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, orbitals, weighted_vectors, 0.0_dp, h_block)
         !CALL cp_gemm('T','N',ncol_global,ncol_global,nrow_global,1.0_dp, &
         !                orbitals,weighted_vectors,0.0_dp,h_block)

         ! if eigenvectors are required, go for syevd, otherwise only compute eigenvalues
         IF (compute_evecs) THEN
            CALL cp_dbcsr_syevd(h_block, e_vectors, evals, para_env=para_env, blacs_env=blacs_env)
         ELSE
            CALL cp_dbcsr_syevx(h_block, eigenvalues=evals, para_env=para_env, blacs_env=blacs_env)
         ENDIF

         ! rotate the orbitals
         IF (do_rotation_local) THEN
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, orbitals, e_vectors, 0.0_dp, weighted_vectors)
            !CALL cp_gemm('N','N',nrow_global,ncol_global,ncol_global,1.0_dp, &
            !             orbitals,e_vectors,0.0_dp,weighted_vectors)
            CALL cp_dbcsr_copy(orbitals, weighted_vectors)
            !CALL cp_fm_to_fm(weighted_vectors,orbitals)
            IF (PRESENT(co_rotate)) THEN
               IF (ASSOCIATED(co_rotate)) THEN
                  CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, co_rotate, e_vectors, 0.0_dp, weighted_vectors)
                  !CALL cp_gemm('N','N',nrow_global,ncol_global,ncol_global,1.0_dp, &
                  !     co_rotate,e_vectors,0.0_dp,weighted_vectors)
                  CALL cp_dbcsr_copy(co_rotate, weighted_vectors)
                  !CALL cp_fm_to_fm(weighted_vectors,co_rotate)
               ENDIF
            ENDIF
         ENDIF

         ! give output
         IF (PRESENT(evals_arg)) THEN
            evals_arg(:) = evals(:)
         ENDIF

         IF (PRESENT(ionode) .OR. PRESENT(scr)) THEN
            IF (.NOT. PRESENT(ionode)) CALL cp__b("qs_mo_methods.F",743,"IONODE?")
            IF (.NOT. PRESENT(scr)) CALL cp__b("qs_mo_methods.F",744,"SCR?")
            IF (ionode) THEN
               DO i = 1, ncol_global, 4
                  j = MIN(3, ncol_global-i)
                  SELECT CASE (j)
                  CASE (3)
                     WRITE (scr, '(1X,4F16.8)') evals(i:i+j)
                  CASE (2)
                     WRITE (scr, '(1X,3F16.8)') evals(i:i+j)
                  CASE (1)
                     WRITE (scr, '(1X,2F16.8)') evals(i:i+j)
                  CASE (0)
                     WRITE (scr, '(1X,1F16.8)') evals(i:i+j)
                  END SELECT
               ENDDO
            ENDIF
         ENDIF

         CALL cp_dbcsr_release_p(weighted_vectors)
         CALL cp_dbcsr_release_p(h_block)
         !CALL cp_fm_release(weighted_vectors)
         !CALL cp_fm_release(h_block)
         IF (compute_evecs) THEN
            CALL cp_dbcsr_release_p(e_vectors)
            !CALL cp_fm_release(e_vectors)
         ENDIF

         DEALLOCATE (evals)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE subspace_eigenvalues_ks_dbcsr

! computes the effective orthonormality of a set of mos given an s-matrix
! orthonormality is the max deviation from unity of the C^T S C
! **************************************************************************************************
!> \brief ...
!> \param orthonormality ...
!> \param mo_array ...
!> \param matrix_s ...
! **************************************************************************************************
   SUBROUTINE calculate_orthonormality(orthonormality, mo_array, matrix_s)
      REAL(KIND=dp)                                      :: orthonormality
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_orthonormality', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, j, k, n, ncol_local, &
                                                            nrow_local, nspin
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: alpha, max_alpha
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: overlap, svec

      NULLIFY (tmp_fm_struct, svec, overlap)

      CALL timeset(routineN, handle)

      nspin = SIZE(mo_array)
      max_alpha = 0.0_dp

      DO ispin = 1, nspin
         IF (PRESENT(matrix_s)) THEN
            ! get S*C
            CALL cp_fm_create(svec, mo_array(ispin)%mo_set%mo_coeff%matrix_struct)
            CALL cp_fm_get_info(mo_array(ispin)%mo_set%mo_coeff, &
                                nrow_global=n, ncol_global=k)
            CALL cp_dbcsr_sm_fm_multiply(matrix_s, mo_array(ispin)%mo_set%mo_coeff, &
                                         svec, k)
            ! get C^T (S*C)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=k, ncol_global=k, &
                                     para_env=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%para_env, &
                                     context=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%context)
            CALL cp_fm_create(overlap, tmp_fm_struct)
            CALL cp_fm_struct_release(tmp_fm_struct)
            CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, mo_array(ispin)%mo_set%mo_coeff, &
                         svec, 0.0_dp, overlap)
            CALL cp_fm_release(svec)
         ELSE
            ! orthogonal basis C^T C
            CALL cp_fm_get_info(mo_array(ispin)%mo_set%mo_coeff, &
                                nrow_global=n, ncol_global=k)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=k, ncol_global=k, &
                                     para_env=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%para_env, &
                                     context=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%context)
            CALL cp_fm_create(overlap, tmp_fm_struct)
            CALL cp_fm_struct_release(tmp_fm_struct)
            CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, mo_array(ispin)%mo_set%mo_coeff, &
                         mo_array(ispin)%mo_set%mo_coeff, 0.0_dp, overlap)
         ENDIF
         CALL cp_fm_get_info(overlap, nrow_local=nrow_local, ncol_local=ncol_local, &
                             row_indices=row_indices, col_indices=col_indices)
         DO i = 1, nrow_local
            DO j = 1, ncol_local
               alpha = overlap%local_data(i, j)
               IF (row_indices(i) .EQ. col_indices(j)) alpha = alpha-1.0_dp
               max_alpha = MAX(max_alpha, ABS(alpha))
            ENDDO
         ENDDO
         CALL cp_fm_release(overlap)
      ENDDO
      CALL mp_max(max_alpha, mo_array(1)%mo_set%mo_coeff%matrix_struct%para_env%group)
      orthonormality = max_alpha
      ! write(*,*) "max deviation from orthonormalization ",orthonormality

      CALL timestop(handle)

   END SUBROUTINE calculate_orthonormality

! computes the minimum/maximum magnitudes of C^T C. This could be useful
! to detect problems in the case of nearly singular overlap matrices.
! in this case, we expect the ratio of min/max to be large
! this routine is only similar to mo_orthonormality if S==1
! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param mo_mag_min ...
!> \param mo_mag_max ...
! **************************************************************************************************
   SUBROUTINE calculate_magnitude(mo_array, mo_mag_min, mo_mag_max)
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      REAL(KIND=dp)                                      :: mo_mag_min, mo_mag_max

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_magnitude', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, k, n, nspin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: evecs, overlap

      NULLIFY (tmp_fm_struct, overlap)

      CALL timeset(routineN, handle)

      nspin = SIZE(mo_array)
      mo_mag_min = HUGE(0.0_dp)
      mo_mag_max = -HUGE(0.0_dp)
      DO ispin = 1, nspin
         CALL cp_fm_get_info(mo_array(ispin)%mo_set%mo_coeff, &
                             nrow_global=n, ncol_global=k)
         ALLOCATE (evals(k))
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=k, ncol_global=k, &
                                  para_env=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%para_env, &
                                  context=mo_array(ispin)%mo_set%mo_coeff%matrix_struct%context)
         CALL cp_fm_create(overlap, tmp_fm_struct)
         CALL cp_fm_create(evecs, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, mo_array(ispin)%mo_set%mo_coeff, &
                      mo_array(ispin)%mo_set%mo_coeff, 0.0_dp, overlap)
         CALL choose_eigv_solver(overlap, evecs, evals)
         mo_mag_min = MIN(MINVAL(evals), mo_mag_min)
         mo_mag_max = MAX(MAXVAL(evals), mo_mag_max)
         CALL cp_fm_release(overlap)
         CALL cp_fm_release(evecs)
         DEALLOCATE (evals)
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE calculate_magnitude

! **************************************************************************************************
!> \brief  Calculate KS eigenvalues starting from  OF MOS
!> \param mos ...
!> \param nspins ...
!> \param ks_rmpv ...
!> \param scf_control ...
!> \param mo_derivs ...
!> \param admm_env ...
!> \par History
!>         02.2013 moved from qs_scf_post_gpw
!>
! **************************************************************************************************
   SUBROUTINE make_mo_eig(mos, nspins, ks_rmpv, scf_control, mo_derivs, admm_env)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      INTEGER, INTENT(IN)                                :: nspins
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mo_derivs
      TYPE(admm_type), OPTIONAL, POINTER                 :: admm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'make_mo_eig', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, ispin, nmo, output_unit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_deriv
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff_deriv, mo_coeff, mo_eigenvalues)
      output_unit = cp_logger_get_default_io_unit()

      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                         eigenvalues=mo_eigenvalues, homo=homo, nmo=nmo)
         IF (output_unit > 0) WRITE (output_unit, *) " "
         IF (output_unit > 0) WRITE (output_unit, *) " Eigenvalues of the occupied subspace spin ", ispin
         !      IF (homo .NE. nmo) THEN
         !         IF (output_unit>0) WRITE(output_unit,*)" and ",nmo-homo," added MO eigenvalues"
         !      ENDIF
         IF (output_unit > 0) WRITE (output_unit, *) "---------------------------------------------"

         IF (scf_control%use_ot) THEN
            ! Also rotate the OT derivs, since they are needed for force calculations
            IF (ASSOCIATED(mo_derivs)) THEN
               mo_coeff_deriv => mo_derivs(ispin)%matrix
            ELSE
               mo_coeff_deriv => NULL()
            ENDIF

            ! ** If we do ADMM, we add have to modify the kohn-sham matrix
            IF (PRESENT(admm_env)) THEN
               CALL admm_correct_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
            END IF

            CALL calculate_subspace_eigenvalues(mo_coeff, ks_rmpv(ispin)%matrix, mo_eigenvalues, &
                                                scr=output_unit, ionode=output_unit > 0, do_rotation=.TRUE., &
                                                co_rotate_dbcsr=mo_coeff_deriv)

            ! ** If we do ADMM, we restore the original kohn-sham matrix
            IF (PRESENT(admm_env)) THEN
               CALL admm_uncorrect_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
            END IF
         ELSE
            IF (output_unit > 0) WRITE (output_unit, '(4(1X,1F16.8))') mo_eigenvalues(1:homo)
         END IF
         IF (.NOT. scf_control%diagonalization%mom) &
            CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, smear=scf_control%smear)
         IF (output_unit > 0) WRITE (output_unit, '(T2,A,F12.6)') &
            "Fermi Energy [eV] :", mos(ispin)%mo_set%mu*evolt
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE make_mo_eig

END MODULE qs_mo_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Storage of past states of the qs_env.
!>      Methods to interpolate (or actually normally extrapolate) the
!>      new guess for density and wavefunctions.
!> \note
!>      Most of the last snapshot should actually be in qs_env, but taking
!>      advantage of it would make the programming much convoluted
!> \par History
!>      02.2003 created [fawzi]
!>      11.2003 Joost VandeVondele : Implemented Nth order PS extrapolation
!>      02.2005 modified for KG_GPW [MI]
!> \author fawzi
! **************************************************************************************************
MODULE qs_wf_history_methods
   USE bibliography,                    ONLY: Kolafa2004,&
                                              VandeVondele2005a,&
                                              cite_reference
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale,&
                                              cp_fm_scale_and_add
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              cp_fm_pool_type,&
                                              fm_pools_create_fm_vect,&
                                              fm_pools_give_back_fm_vect
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              low_print_level
   USE input_constants,                 ONLY: &
        wfi_aspc_nr, wfi_frozen_method_nr, wfi_linear_p_method_nr, wfi_linear_ps_method_nr, &
        wfi_linear_wf_method_nr, wfi_ps_method_nr, wfi_use_guess_method_nr, &
        wfi_use_prev_p_method_nr, wfi_use_prev_rho_r_method_nr, wfi_use_prev_wf_method_nr
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: binomial
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_copy
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_did_change
   USE qs_matrix_pools,                 ONLY: mpools_get,&
                                              qs_matrix_pools_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix,&
                                              make_basis_cholesky,&
                                              make_basis_lowdin,&
                                              make_basis_simple,&
                                              make_basis_sm
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_set,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: ot_method_nr,&
                                              qs_scf_env_type
   USE qs_wf_history_types,             ONLY: qs_wf_history_type,&
                                              qs_wf_snapshot_type,&
                                              wfi_get_snapshot
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_wf_history_methods'
   INTEGER, SAVE, PRIVATE :: last_wfs_id = 0, last_wfi_id = 0

   PUBLIC :: wfi_create, wfi_update, wfi_create_for_kp, &
             wfi_extrapolate, wfi_get_method_label, &
             reorthogonalize_vectors

CONTAINS

! **************************************************************************************************
!> \brief allocates and initialize a wavefunction snapshot
!> \param snapshot the snapshot to create
!> \par History
!>      02.2003 created [fawzi]
!>      02.2005 added wf_mol [MI]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfs_create(snapshot)
      TYPE(qs_wf_snapshot_type), POINTER                 :: snapshot

      CHARACTER(len=*), PARAMETER :: routineN = 'wfs_create', routineP = moduleN//':'//routineN

      ALLOCATE (snapshot)
      last_wfs_id = last_wfs_id+1
      snapshot%id_nr = last_wfs_id
      NULLIFY (snapshot%wf, snapshot%rho_r, &
               snapshot%rho_g, snapshot%rho_ao, snapshot%rho_ao_kp, &
               snapshot%overlap, snapshot%rho_frozen)
      snapshot%dt = 1.0_dp
      snapshot%ref_count = 1
   END SUBROUTINE wfs_create

! **************************************************************************************************
!> \brief updates the given snapshot
!> \param snapshot the snapshot to be updated
!> \param wf_history the history
!> \param qs_env the qs_env that should be snapshotted
!> \param dt the time of the snapshot (wrt. to the previous snapshot)
!> \par History
!>      02.2003 created [fawzi]
!>      02.2005 added kg_fm_mol_set for KG_GPW [MI]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfs_update(snapshot, wf_history, qs_env, dt)
      TYPE(qs_wf_snapshot_type), POINTER                 :: snapshot
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: dt

      CHARACTER(len=*), PARAMETER :: routineN = 'wfs_update', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, img, ispin, nimg, nspins
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_pools
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, auxbas_pw_pool, ao_mo_pools, dft_control, mos, mo_coeff, &
               rho, rho_r, rho_g, rho_ao, matrix_s)
      CALL get_qs_env(qs_env, pw_env=pw_env, &
                      dft_control=dft_control, rho=rho)
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_pools)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_methods.F",170)
      IF(.NOT.(ASSOCIATED(dft_control)))CALL cp__a("qs_wf_history_methods.F",171)
      IF (.NOT. ASSOCIATED(snapshot)) THEN
         CALL wfs_create(snapshot)
      END IF
      IF(.NOT.(wf_history%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",175)
      IF(.NOT.(snapshot%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",176)

      nspins = dft_control%nspins
      snapshot%dt = 1.0_dp
      IF (PRESENT(dt)) snapshot%dt = dt
      IF (wf_history%store_wf) THEN
         CALL get_qs_env(qs_env, mos=mos)
         IF (.NOT. ASSOCIATED(snapshot%wf)) THEN
            CALL fm_pools_create_fm_vect(ao_mo_pools, snapshot%wf, &
                                         name="ws_snap"//TRIM(ADJUSTL(cp_to_string(snapshot%id_nr)))// &
                                         "ws")
            IF(.NOT.(nspins == SIZE(snapshot%wf)))CALL cp__a("qs_wf_history_methods.F",187)
         END IF
         DO ispin = 1, nspins
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_to_fm(mo_coeff, snapshot%wf(ispin)%matrix)
         END DO
      ELSE IF (ASSOCIATED(snapshot%wf)) THEN
         CALL fm_pools_give_back_fm_vect(ao_mo_pools, snapshot%wf)
      END IF

      IF (wf_history%store_rho_r) THEN
         CALL qs_rho_get(rho, rho_r=rho_r)
         IF(.NOT.(ASSOCIATED(rho_r)))CALL cp__a("qs_wf_history_methods.F",199)
         IF (.NOT. ASSOCIATED(snapshot%rho_r)) THEN
            ALLOCATE (snapshot%rho_r(nspins))
            DO ispin = 1, nspins
               NULLIFY (snapshot%rho_r(ispin)%pw)
               CALL pw_pool_create_pw(auxbas_pw_pool, snapshot%rho_r(ispin)%pw, &
                                      in_space=REALSPACE, use_data=REALDATA3D)
            END DO
         END IF
         DO ispin = 1, nspins
            CALL pw_copy(rho_r(ispin)%pw, snapshot%rho_r(ispin)%pw)
         END DO
      ELSE IF (ASSOCIATED(snapshot%rho_r)) THEN
         DO ispin = 1, SIZE(snapshot%rho_r)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, snapshot%rho_r(ispin)%pw)
         END DO
         DEALLOCATE (snapshot%rho_r)
      END IF

      IF (wf_history%store_rho_g) THEN
         CALL qs_rho_get(rho, rho_g=rho_g)
         IF(.NOT.(ASSOCIATED(rho_g)))CALL cp__a("qs_wf_history_methods.F",220)
         IF (.NOT. ASSOCIATED(snapshot%rho_g)) THEN
            ALLOCATE (snapshot%rho_g(nspins))
            DO ispin = 1, nspins
               NULLIFY (snapshot%rho_g(ispin)%pw)
               CALL pw_pool_create_pw(auxbas_pw_pool, snapshot%rho_g(ispin)%pw, &
                                      in_space=RECIPROCALSPACE, use_data=COMPLEXDATA1D)
            END DO
         END IF
         DO ispin = 1, nspins
            CALL pw_copy(rho_g(ispin)%pw, snapshot%rho_g(ispin)%pw)
         END DO
      ELSE IF (ASSOCIATED(snapshot%rho_g)) THEN
         DO ispin = 1, SIZE(snapshot%rho_g)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, snapshot%rho_g(ispin)%pw)
         END DO
         DEALLOCATE (snapshot%rho_g)
      END IF

      IF (ASSOCIATED(snapshot%rho_ao)) THEN ! the sparsity might be different
         ! (future struct:check)
         CALL cp_dbcsr_deallocate_matrix_set(snapshot%rho_ao)
      END IF
      IF (wf_history%store_rho_ao) THEN
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         IF(.NOT.(ASSOCIATED(rho_ao)))CALL cp__a("qs_wf_history_methods.F",245)

         CALL cp_dbcsr_allocate_matrix_set(snapshot%rho_ao, nspins)
         DO ispin = 1, nspins
            ALLOCATE (snapshot%rho_ao(ispin)%matrix)
            CALL cp_dbcsr_init(snapshot%rho_ao(ispin)%matrix)
            CALL cp_dbcsr_copy(snapshot%rho_ao(ispin)%matrix, rho_ao(ispin)%matrix)
         END DO
      END IF

      IF (ASSOCIATED(snapshot%rho_ao_kp)) THEN ! the sparsity might be different
         ! (future struct:check)
         CALL cp_dbcsr_deallocate_matrix_set(snapshot%rho_ao_kp)
      END IF
      IF (wf_history%store_rho_ao_kp) THEN
         CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
         IF(.NOT.(ASSOCIATED(rho_ao_kp)))CALL cp__a("qs_wf_history_methods.F",261)

         nimg = dft_control%nimages
         CALL cp_dbcsr_allocate_matrix_set(snapshot%rho_ao_kp, nspins, nimg)
         DO ispin = 1, nspins
            DO img = 1, nimg
               ALLOCATE (snapshot%rho_ao_kp(ispin, img)%matrix)
               CALL cp_dbcsr_init(snapshot%rho_ao_kp(ispin, img)%matrix)
               CALL cp_dbcsr_copy(snapshot%rho_ao_kp(ispin, img)%matrix, &
                                  rho_ao_kp(ispin, img)%matrix)
            END DO
         END DO
      END IF

      IF (ASSOCIATED(snapshot%overlap)) THEN ! the sparsity might be different
         ! (future struct:check)
         CALL cp_dbcsr_deallocate_matrix(snapshot%overlap)
      END IF
      IF (wf_history%store_overlap) THEN
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("qs_wf_history_methods.F",281)
         IF(.NOT.(ASSOCIATED(matrix_s(1)%matrix)))CALL cp__a("qs_wf_history_methods.F",282)
         ALLOCATE (snapshot%overlap)
         CALL cp_dbcsr_init(snapshot%overlap)
         CALL cp_dbcsr_copy(snapshot%overlap, matrix_s(1)%matrix)
      END IF

      IF (wf_history%store_frozen_density) THEN
         ! do nothing
         ! CALL deallocate_matrix_set(snapshot%rho_frozen%rho_ao)
      END IF

      CALL timestop(handle)

   END SUBROUTINE wfs_update

! **************************************************************************************************
!> \brief ...
!> \param wf_history ...
!> \param interpolation_method_nr the tag of the method used for
!>        the extrapolation of the intial density for the next md step
!>        (see qs_wf_history_types:wfi_*_method_nr)
!> \param extrapolation_order ...
!> \param has_unit_metric ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_create(wf_history, interpolation_method_nr, extrapolation_order, &
                         has_unit_metric)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      INTEGER, INTENT(in)                                :: interpolation_method_nr, &
                                                            extrapolation_order
      LOGICAL, INTENT(IN)                                :: has_unit_metric

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      ALLOCATE (wf_history)
      last_wfi_id = last_wfi_id+1
      wf_history%id_nr = last_wfi_id
      wf_history%ref_count = 1
      wf_history%memory_depth = 0
      wf_history%snapshot_count = 0
      wf_history%last_state_index = 1
      wf_history%store_wf = .FALSE.
      wf_history%store_rho_r = .FALSE.
      wf_history%store_rho_g = .FALSE.
      wf_history%store_rho_ao = .FALSE.
      wf_history%store_rho_ao_kp = .FALSE.
      wf_history%store_overlap = .FALSE.
      wf_history%store_frozen_density = .FALSE.
      NULLIFY (wf_history%past_states)

      wf_history%interpolation_method_nr = interpolation_method_nr

      SELECT CASE (wf_history%interpolation_method_nr)
      CASE (wfi_use_guess_method_nr)
         wf_history%memory_depth = 0
      CASE (wfi_use_prev_wf_method_nr)
         wf_history%memory_depth = 0
      CASE (wfi_use_prev_p_method_nr)
         wf_history%memory_depth = 1
         wf_history%store_rho_ao = .TRUE.
      CASE (wfi_use_prev_rho_r_method_nr)
         wf_history%memory_depth = 1
         wf_history%store_rho_ao = .TRUE.
      CASE (wfi_linear_wf_method_nr)
         wf_history%memory_depth = 2
         wf_history%store_wf = .TRUE.
      CASE (wfi_linear_p_method_nr)
         wf_history%memory_depth = 2
         wf_history%store_rho_ao = .TRUE.
      CASE (wfi_linear_ps_method_nr)
         wf_history%memory_depth = 2
         wf_history%store_wf = .TRUE.
         IF (.NOT. has_unit_metric) wf_history%store_overlap = .TRUE.
      CASE (wfi_ps_method_nr)
         CALL cite_reference(VandeVondele2005a)
         wf_history%memory_depth = extrapolation_order+1
         wf_history%store_wf = .TRUE.
         IF (.NOT. has_unit_metric) wf_history%store_overlap = .TRUE.
      CASE (wfi_frozen_method_nr)
         wf_history%memory_depth = 1
         wf_history%store_frozen_density = .TRUE.
      CASE (wfi_aspc_nr)
         wf_history%memory_depth = extrapolation_order+2
         wf_history%store_wf = .TRUE.
         IF (.NOT. has_unit_metric) wf_history%store_overlap = .TRUE.
      CASE default
         CALL cp_abort(cp__l("qs_wf_history_methods.F",374), &
                       "Unknown interpolation method: "// &
                       TRIM(ADJUSTL(cp_to_string(interpolation_method_nr))))
         wf_history%interpolation_method_nr = wfi_use_prev_rho_r_method_nr
      END SELECT
      ALLOCATE (wf_history%past_states(wf_history%memory_depth))

      DO i = 1, SIZE(wf_history%past_states)
         NULLIFY (wf_history%past_states(i)%snapshot)
      END DO

      CALL timestop(handle)
   END SUBROUTINE wfi_create

! **************************************************************************************************
!> \brief ...
!> \param wf_history ...
!> \par History
!>      06.2015 created [jhu]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_create_for_kp(wf_history)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_create_for_kp', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_methods.F",401)
      IF (wf_history%store_rho_ao) THEN
         wf_history%store_rho_ao_kp = .TRUE.
         wf_history%store_rho_ao = .FALSE.
      END IF
      ! Check for KP compatible methods
      IF (wf_history%store_wf) THEN
         CALL cp__b("qs_wf_history_methods.F",408,"WFN based interpolation method not possible for kpoints.")
      END IF
      IF (wf_history%store_frozen_density) THEN
         CALL cp__b("qs_wf_history_methods.F",411,"Frozen density initialization method not possible for kpoints.")
      END IF
      IF (wf_history%store_overlap) THEN
         CALL cp__b("qs_wf_history_methods.F",414,"Inconsistent interpolation method for kpoints.")
      END IF

   END SUBROUTINE wfi_create_for_kp

! **************************************************************************************************
!> \brief returns a string describing the interpolation method
!> \param method_nr ...
!> \retval res ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   FUNCTION wfi_get_method_label(method_nr) RESULT(res)
      INTEGER, INTENT(in)                                :: method_nr
      CHARACTER(len=30)                                  :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_get_method_label', &
         routineP = moduleN//':'//routineN

      res = "unknown"
      SELECT CASE (method_nr)
      CASE (wfi_use_prev_p_method_nr)
         res = "previous_p"
      CASE (wfi_use_prev_wf_method_nr)
         res = "previous_wf"
      CASE (wfi_use_prev_rho_r_method_nr)
         res = "previous_rho_r"
      CASE (wfi_use_guess_method_nr)
         res = "initial_guess"
      CASE (wfi_linear_wf_method_nr)
         res = "mo linear"
      CASE (wfi_linear_p_method_nr)
         res = "P linear"
      CASE (wfi_linear_ps_method_nr)
         res = "PS linear"
      CASE (wfi_ps_method_nr)
         res = "PS Nth order"
      CASE (wfi_frozen_method_nr)
         res = "frozen density approximation"
      CASE (wfi_aspc_nr)
         res = "ASPC"
      CASE default
         CALL cp_abort(cp__l("qs_wf_history_methods.F",457), &
                       "Unknown interpolation method: "// &
                       TRIM(ADJUSTL(cp_to_string(method_nr))))
      END SELECT
   END FUNCTION wfi_get_method_label

! **************************************************************************************************
!> \brief calculates the new starting state for the scf for the next
!>      wf optimization
!> \param wf_history the previous history needed to extrapolate
!> \param qs_env the qs env with the latest result, and that will contain
!>        the new starting state
!> \param dt the time at which to extrapolate (wrt. to the last snapshot)
!> \param extrapolation_method_nr returns the extrapolation method used
!> \param orthogonal_wf ...
!> \par History
!>      02.2003 created [fawzi]
!>      11.2003 Joost VandeVondele : Implemented Nth order PS extrapolation
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_extrapolate(wf_history, qs_env, dt, extrapolation_method_nr, &
                              orthogonal_wf)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(IN)                          :: dt
      INTEGER, INTENT(OUT), OPTIONAL                     :: extrapolation_method_nr
      LOGICAL, INTENT(OUT), OPTIONAL                     :: orthogonal_wf

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_extrapolate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: actual_extrapolation_method_nr, handle, &
                                                            i, img, ispin, k, n, nmo, nvec, &
                                                            output_unit, print_level
      LOGICAL                                            :: do_kpoints, my_orthogonal_wf, use_overlap
      REAL(KIND=dp)                                      :: alpha, t0, t1, t2
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao, rho_frozen_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct, matrix_struct_new
      TYPE(cp_fm_type), POINTER                          :: csc, fm_tmp, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_wf_snapshot_type), POINTER                 :: t0_state, t1_state

      NULLIFY (mos, ao_mo_fm_pools, t0_state, t1_state, mo_coeff, &
               rho, rho_ao, rho_frozen_ao)

      use_overlap = wf_history%store_overlap

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      print_level = logger%iter_info%print_level
      output_unit = cp_print_key_unit_nr(logger, qs_env%input, "DFT%SCF%PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_methods.F",514)
      IF(.NOT.(wf_history%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",515)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_wf_history_methods.F",516)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",517)
      CALL get_qs_env(qs_env, mos=mos, rho=rho, do_kpoints=do_kpoints)
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools)
      ! chooses the method for this extrapolation
      IF (wf_history%snapshot_count < 1) THEN
         actual_extrapolation_method_nr = wfi_use_guess_method_nr
      ELSE
         actual_extrapolation_method_nr = wf_history%interpolation_method_nr
      END IF

      SELECT CASE (actual_extrapolation_method_nr)
      CASE (wfi_linear_wf_method_nr)
         IF (wf_history%snapshot_count < 2) THEN
            actual_extrapolation_method_nr = wfi_use_prev_wf_method_nr
         END IF
      CASE (wfi_linear_p_method_nr)
         IF (wf_history%snapshot_count < 2) THEN
            actual_extrapolation_method_nr = wfi_use_prev_wf_method_nr
         END IF
      CASE (wfi_linear_ps_method_nr)
         IF (wf_history%snapshot_count < 2) THEN
            actual_extrapolation_method_nr = wfi_use_prev_wf_method_nr
         END IF
      END SELECT

      IF (PRESENT(extrapolation_method_nr)) &
         extrapolation_method_nr = actual_extrapolation_method_nr
      my_orthogonal_wf = .FALSE.

      SELECT CASE (actual_extrapolation_method_nr)
      CASE (wfi_frozen_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",548)
         t0_state => wfi_get_snapshot(wf_history, index=1)
         IF(.NOT.(ASSOCIATED(t0_state%rho_frozen)))CALL cp__a("qs_wf_history_methods.F",550)

         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)

         CALL qs_rho_get(t0_state%rho_frozen, rho_ao=rho_frozen_ao)
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         DO ispin = 1, SIZE(rho_frozen_ao)
            CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, &
                               rho_frozen_ao(ispin)%matrix, &
                               keep_sparsity=.TRUE.)
         END DO
         !FM updating rho_ao directly with t0_state%rho_ao would have the
         !FM wrong matrix structure
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

         my_orthogonal_wf = .FALSE.
      CASE (wfi_use_prev_rho_r_method_nr)
         t0_state => wfi_get_snapshot(wf_history, index=1)
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)
         IF (do_kpoints) THEN
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao_kp)))CALL cp__a("qs_wf_history_methods.F",573)
            CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
            DO ispin = 1, SIZE(t0_state%rho_ao_kp, 1)
               DO img = 1, SIZE(t0_state%rho_ao_kp, 2)
                  CALL cp_dbcsr_copy(rho_ao_kp(ispin, img)%matrix, &
                                     t0_state%rho_ao_kp(ispin, img)%matrix, &
                                     keep_sparsity=.TRUE.)
               END DO
            END DO
         ELSE
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao)))CALL cp__a("qs_wf_history_methods.F",583)
            CALL qs_rho_get(rho, rho_ao=rho_ao)
            DO ispin = 1, SIZE(t0_state%rho_ao)
               CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, &
                                  t0_state%rho_ao(ispin)%matrix, &
                                  keep_sparsity=.TRUE.)
            END DO
         END IF
         ! Why is rho_g valid at this point ?
         CALL qs_rho_set(rho, rho_g_valid=.TRUE.)

         ! does nothing
      CASE (wfi_use_prev_wf_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",596)
         my_orthogonal_wf = .TRUE.
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         DO ispin = 1, SIZE(mos)
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            nmo=nmo)
            CALL reorthogonalize_vectors(qs_env, &
                                         v_matrix=mo_coeff, &
                                         n_col=nmo)
            CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                          density_matrix=rho_ao(ispin)%matrix)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      CASE (wfi_use_prev_p_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",614)
         t0_state => wfi_get_snapshot(wf_history, index=1)
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)
         IF (do_kpoints) THEN
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao_kp)))CALL cp__a("qs_wf_history_methods.F",619)
            CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
            DO ispin = 1, SIZE(t0_state%rho_ao_kp, 1)
               DO img = 1, SIZE(t0_state%rho_ao_kp, 2)
                  CALL cp_dbcsr_copy(rho_ao_kp(ispin, img)%matrix, &
                                     t0_state%rho_ao_kp(ispin, img)%matrix, &
                                     keep_sparsity=.TRUE.)
               END DO
            END DO
         ELSE
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao)))CALL cp__a("qs_wf_history_methods.F",629)
            CALL qs_rho_get(rho, rho_ao=rho_ao)
            DO ispin = 1, SIZE(t0_state%rho_ao)
               CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, &
                                  t0_state%rho_ao(ispin)%matrix, &
                                  keep_sparsity=.TRUE.)
            END DO
         END IF
         !FM updating rho_ao directly with t0_state%rho_ao would have the
         !FM wrong matrix structure
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      CASE (wfi_use_guess_method_nr)
         !FM more clean to do it here, but it
         !FM might need to read a file (restart) and thus globenv
         !FM I do not want globenv here, thus done by the caller
         !FM (btw. it also needs the eigensolver, and unless you relocate it
         !FM gives circular dependencies)
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)
      CASE (wfi_linear_wf_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",651)
         t0_state => wfi_get_snapshot(wf_history, index=2)
         t1_state => wfi_get_snapshot(wf_history, index=1)
         IF(.NOT.(ASSOCIATED(t0_state)))CALL cp__a("qs_wf_history_methods.F",654)
         IF(.NOT.(ASSOCIATED(t1_state)))CALL cp__a("qs_wf_history_methods.F",655)
         IF(.NOT.(ASSOCIATED(t0_state%wf)))CALL cp__a("qs_wf_history_methods.F",656)
         IF(.NOT.(ASSOCIATED(t1_state%wf)))CALL cp__a("qs_wf_history_methods.F",657)
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)

         my_orthogonal_wf = .TRUE.
         t0 = 0.0_dp
         t1 = t1_state%dt
         t2 = t1+dt
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         DO ispin = 1, SIZE(mos)
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            nmo=nmo)
            CALL cp_fm_scale_and_add(alpha=0.0_dp, &
                                     matrix_a=mo_coeff, &
                                     matrix_b=t1_state%wf(ispin)%matrix, &
                                     beta=(t2-t0)/(t1-t0))
            ! this copy should be unnecessary
            CALL cp_fm_scale_and_add(alpha=1.0_dp, &
                                     matrix_a=mo_coeff, &
                                     beta=(t1-t2)/(t1-t0), matrix_b=t0_state%wf(ispin)%matrix)
            CALL reorthogonalize_vectors(qs_env, &
                                         v_matrix=mo_coeff, &
                                         n_col=nmo)
            CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                          density_matrix=rho_ao(ispin)%matrix)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)

         CALL qs_ks_did_change(qs_env%ks_env, &
                               rho_changed=.TRUE.)
      CASE (wfi_linear_p_method_nr)
         t0_state => wfi_get_snapshot(wf_history, index=2)
         t1_state => wfi_get_snapshot(wf_history, index=1)
         IF(.NOT.(ASSOCIATED(t0_state)))CALL cp__a("qs_wf_history_methods.F",690)
         IF(.NOT.(ASSOCIATED(t1_state)))CALL cp__a("qs_wf_history_methods.F",691)
         IF (do_kpoints) THEN
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao_kp)))CALL cp__a("qs_wf_history_methods.F",693)
            IF(.NOT.(ASSOCIATED(t1_state%rho_ao_kp)))CALL cp__a("qs_wf_history_methods.F",694)
         ELSE
            IF(.NOT.(ASSOCIATED(t0_state%rho_ao)))CALL cp__a("qs_wf_history_methods.F",696)
            IF(.NOT.(ASSOCIATED(t1_state%rho_ao)))CALL cp__a("qs_wf_history_methods.F",697)
         END IF
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         CALL wfi_set_history_variables(qs_env=qs_env, nvec=nvec)

         t0 = 0.0_dp
         t1 = t1_state%dt
         t2 = t1+dt
         IF (do_kpoints) THEN
            CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
            DO ispin = 1, SIZE(rho_ao_kp, 1)
               DO img = 1, SIZE(rho_ao_kp, 2)
                  CALL cp_dbcsr_add(rho_ao_kp(ispin, img)%matrix, t1_state%rho_ao_kp(ispin, img)%matrix, &
                                    alpha_scalar=0.0_dp, beta_scalar=(t2-t0)/(t1-t0)) ! this copy should be unnecessary
                  CALL cp_dbcsr_add(rho_ao_kp(ispin, img)%matrix, t0_state%rho_ao_kp(ispin, img)%matrix, &
                                    alpha_scalar=1.0_dp, beta_scalar=(t1-t2)/(t1-t0))
               END DO
            END DO
         ELSE
            CALL qs_rho_get(rho, rho_ao=rho_ao)
            DO ispin = 1, SIZE(rho_ao)
               CALL cp_dbcsr_add(rho_ao(ispin)%matrix, t1_state%rho_ao(ispin)%matrix, &
                                 alpha_scalar=0.0_dp, beta_scalar=(t2-t0)/(t1-t0)) ! this copy should be unnecessary
               CALL cp_dbcsr_add(rho_ao(ispin)%matrix, t0_state%rho_ao(ispin)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=(t1-t2)/(t1-t0))
            END DO
         END IF
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

         ! wf not calculated, extract with PSC renormalized?
         ! use wf_linear?
      CASE (wfi_linear_ps_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",730)
         t0_state => wfi_get_snapshot(wf_history, index=2)
         t1_state => wfi_get_snapshot(wf_history, index=1)
         IF(.NOT.(ASSOCIATED(t0_state)))CALL cp__a("qs_wf_history_methods.F",733)
         IF(.NOT.(ASSOCIATED(t1_state)))CALL cp__a("qs_wf_history_methods.F",734)
         IF(.NOT.(ASSOCIATED(t0_state%wf)))CALL cp__a("qs_wf_history_methods.F",735)
         IF(.NOT.(ASSOCIATED(t1_state%wf)))CALL cp__a("qs_wf_history_methods.F",736)
         IF (wf_history%store_overlap) THEN
            IF(.NOT.(ASSOCIATED(t0_state%overlap)))CALL cp__a("qs_wf_history_methods.F",738)
            IF(.NOT.(ASSOCIATED(t1_state%overlap)))CALL cp__a("qs_wf_history_methods.F",739)
         END IF
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         IF (nvec >= wf_history%memory_depth) THEN
            IF ((qs_env%scf_control%max_scf_hist .NE. 0) .AND. (qs_env%scf_control%eps_scf_hist .NE. 0)) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%max_scf_hist .NE. 0) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%eps_scf_hist .NE. 0) THEN
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
            END IF
         END IF

         my_orthogonal_wf = .TRUE.
         ! use PS_2=2 PS_1-PS_0
         ! C_2 comes from using PS_2 as a projector acting on C_1
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         DO ispin = 1, SIZE(mos)
            NULLIFY (mo_coeff, matrix_struct, matrix_struct_new, csc)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_get_info(mo_coeff, nrow_global=n, ncol_global=k, &
                                matrix_struct=matrix_struct)
            CALL cp_fm_struct_create(matrix_struct_new, template_fmstruct=matrix_struct, &
                                     nrow_global=k, ncol_global=k)
            CALL cp_fm_create(csc, matrix_struct_new)
            CALL cp_fm_struct_release(matrix_struct_new)

            IF (use_overlap) THEN
               CALL cp_dbcsr_sm_fm_multiply(t0_state%overlap, t1_state%wf(ispin)%matrix, mo_coeff, k)
               CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, mo_coeff, 0.0_dp, csc)
            ELSE
               CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, &
                            t1_state%wf(ispin)%matrix, 0.0_dp, csc)
            END IF
            CALL cp_gemm('N', 'N', n, k, k, 1.0_dp, t0_state%wf(ispin)%matrix, csc, 0.0_dp, mo_coeff)
            CALL cp_fm_release(csc)
            CALL cp_fm_scale_and_add(-1.0_dp, mo_coeff, 2.0_dp, t1_state%wf(ispin)%matrix)
            CALL reorthogonalize_vectors(qs_env, &
                                         v_matrix=mo_coeff, &
                                         n_col=k)
            CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                          density_matrix=rho_ao(ispin)%matrix)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      CASE (wfi_ps_method_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",789)
         ! figure out the actual number of vectors to use in the extrapolation:
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         IF(.NOT.(nvec .GT. 0))CALL cp__a("qs_wf_history_methods.F",792)
         IF (nvec >= wf_history%memory_depth) THEN
            IF ((qs_env%scf_control%max_scf_hist .NE. 0) .AND. (qs_env%scf_control%eps_scf_hist .NE. 0)) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%max_scf_hist .NE. 0) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%eps_scf_hist .NE. 0) THEN
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
            END IF
         END IF

         my_orthogonal_wf = .TRUE.
         DO ispin = 1, SIZE(mos)
            NULLIFY (mo_coeff, matrix_struct, matrix_struct_new, csc, fm_tmp)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_get_info(mo_coeff, nrow_global=n, ncol_global=k, &
                                matrix_struct=matrix_struct)
            CALL cp_fm_create(fm_tmp, matrix_struct)
            CALL cp_fm_struct_create(matrix_struct_new, template_fmstruct=matrix_struct, &
                                     nrow_global=k, ncol_global=k)
            CALL cp_fm_create(csc, matrix_struct_new)
            CALL cp_fm_struct_release(matrix_struct_new)
            ! first the most recent
            t1_state => wfi_get_snapshot(wf_history, index=1)
            CALL cp_fm_to_fm(t1_state%wf(ispin)%matrix, mo_coeff)
            alpha = nvec
            CALL cp_fm_scale(alpha, mo_coeff)
            CALL qs_rho_get(rho, rho_ao=rho_ao)
            DO i = 2, nvec
               t0_state => wfi_get_snapshot(wf_history, index=i)
               IF (use_overlap) THEN
                  CALL cp_dbcsr_sm_fm_multiply(t0_state%overlap, t1_state%wf(ispin)%matrix, fm_tmp, k)
                  CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, fm_tmp, 0.0_dp, csc)
               ELSE
                  CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, &
                               t1_state%wf(ispin)%matrix, 0.0_dp, csc)
               END IF
               CALL cp_gemm('N', 'N', n, k, k, 1.0_dp, t0_state%wf(ispin)%matrix, csc, 0.0_dp, fm_tmp)
               alpha = -1.0_dp*alpha*REAL(nvec-i+1, dp)/REAL(i, dp)
               CALL cp_fm_scale_and_add(1.0_dp, mo_coeff, alpha, fm_tmp)
            ENDDO

            CALL cp_fm_release(csc)
            CALL cp_fm_release(fm_tmp)
            CALL reorthogonalize_vectors(qs_env, &
                                         v_matrix=mo_coeff, &
                                         n_col=k)
            CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                          density_matrix=rho_ao(ispin)%matrix)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      CASE (wfi_aspc_nr)
         IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_wf_history_methods.F",849)
         CALL cite_reference(Kolafa2004)
         ! figure out the actual number of vectors to use in the extrapolation:
         nvec = MIN(wf_history%memory_depth, wf_history%snapshot_count)
         IF(.NOT.(nvec .GT. 0))CALL cp__a("qs_wf_history_methods.F",853)
         IF (nvec >= wf_history%memory_depth) THEN
            IF ((qs_env%scf_control%max_scf_hist .NE. 0) .AND. &
                (qs_env%scf_control%eps_scf_hist .NE. 0)) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%max_scf_hist .NE. 0) THEN
               qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
               qs_env%scf_control%outer_scf%have_scf = .FALSE.
            ELSE IF (qs_env%scf_control%eps_scf_hist .NE. 0) THEN
               qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
            END IF
         END IF

         my_orthogonal_wf = .TRUE.
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         DO ispin = 1, SIZE(mos)
            NULLIFY (mo_coeff, matrix_struct, matrix_struct_new, csc, fm_tmp)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_get_info(mo_coeff, &
                                nrow_global=n, &
                                ncol_global=k, &
                                matrix_struct=matrix_struct)
            CALL cp_fm_create(fm_tmp, matrix_struct)
            CALL cp_fm_struct_create(matrix_struct_new, &
                                     template_fmstruct=matrix_struct, &
                                     nrow_global=k, &
                                     ncol_global=k)
            CALL cp_fm_create(csc, matrix_struct_new)
            CALL cp_fm_struct_release(matrix_struct_new)
            ! first the most recent
            t1_state => wfi_get_snapshot(wf_history, &
                                         index=1)
            CALL cp_fm_to_fm(t1_state%wf(ispin)%matrix, mo_coeff)
            alpha = REAL(4*nvec-2, KIND=dp)/REAL(nvec+1, KIND=dp)
            IF ((output_unit > 0) .AND. (print_level > low_print_level)) THEN
               WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T3,A,I0,/,/,T3,A2,I0,A4,F10.6)") &
                  "Parameters for the always stable predictor-corrector (ASPC) method:", &
                  "ASPC order: ", MAX(nvec-2, 0), &
                  "B(", 1, ") = ", alpha
            END IF
            CALL cp_fm_scale(alpha, mo_coeff)

            DO i = 2, nvec
               t0_state => wfi_get_snapshot(wf_history, index=i)
               IF (use_overlap) THEN
                  CALL cp_dbcsr_sm_fm_multiply(t0_state%overlap, t1_state%wf(ispin)%matrix, fm_tmp, k)
                  CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, fm_tmp, 0.0_dp, csc)
               ELSE
                  CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, t0_state%wf(ispin)%matrix, &
                               t1_state%wf(ispin)%matrix, 0.0_dp, csc)
               END IF
               CALL cp_gemm('N', 'N', n, k, k, 1.0_dp, t0_state%wf(ispin)%matrix, csc, 0.0_dp, fm_tmp)
               alpha = (-1.0_dp)**(i+1)*REAL(i, KIND=dp)* &
                       binomial(2*nvec, nvec-i)/binomial(2*nvec-2, nvec-1)
               IF ((output_unit > 0) .AND. (print_level > low_print_level)) THEN
                  WRITE (UNIT=output_unit, FMT="(T3,A2,I0,A4,F10.6)") &
                     "B(", i, ") = ", alpha
               END IF
               CALL cp_fm_scale_and_add(1.0_dp, mo_coeff, alpha, fm_tmp)
            END DO
            CALL cp_fm_release(csc)
            CALL cp_fm_release(fm_tmp)
            CALL reorthogonalize_vectors(qs_env, &
                                         v_matrix=mo_coeff, &
                                         n_col=k)
            CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                          density_matrix=rho_ao(ispin)%matrix)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      CASE default
         CALL cp_abort(cp__l("qs_wf_history_methods.F",927), &
                       "Unknown interpolation method: "// &
                       TRIM(ADJUSTL(cp_to_string(wf_history%interpolation_method_nr))))
      END SELECT
      IF (PRESENT(orthogonal_wf)) orthogonal_wf = my_orthogonal_wf
      CALL cp_print_key_finished_output(output_unit, logger, qs_env%input, &
                                        "DFT%SCF%PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)
   END SUBROUTINE wfi_extrapolate

! **************************************************************************************************
!> \brief Decides if scf control variables has to changed due
!>      to using a WF extrapolation.
!> \param qs_env The QS environment
!> \param nvec ...
!> \par History
!>      11.2006 created [TdK]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE wfi_set_history_variables(qs_env, nvec)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: nvec

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_set_history_variables', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_wf_history_methods.F",957)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",958)

      IF (nvec >= qs_env%wf_history%memory_depth) THEN
         IF ((qs_env%scf_control%max_scf_hist .NE. 0) .AND. (qs_env%scf_control%eps_scf_hist .NE. 0)) THEN
            qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
            qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
            qs_env%scf_control%outer_scf%have_scf = .FALSE.
         ELSE IF (qs_env%scf_control%max_scf_hist .NE. 0) THEN
            qs_env%scf_control%max_scf = qs_env%scf_control%max_scf_hist
            qs_env%scf_control%outer_scf%have_scf = .FALSE.
         ELSE IF (qs_env%scf_control%eps_scf_hist .NE. 0) THEN
            qs_env%scf_control%eps_scf = qs_env%scf_control%eps_scf_hist
            qs_env%scf_control%outer_scf%eps_scf = qs_env%scf_control%eps_scf_hist
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE wfi_set_history_variables

! **************************************************************************************************
!> \brief updates the snapshot buffer, taking a new snapshot
!> \param wf_history the history buffer to update
!> \param qs_env the qs_env we get the info from
!> \param dt ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_update(wf_history, qs_env, dt)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(in)                          :: dt

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_update', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_methods.F",994)
      IF(.NOT.(wf_history%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",995)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_wf_history_methods.F",996)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_wf_history_methods.F",997)

      wf_history%snapshot_count = wf_history%snapshot_count+1
      IF (wf_history%memory_depth > 0) THEN
         wf_history%last_state_index = MODULO(wf_history%snapshot_count, &
                                              wf_history%memory_depth)+1
         CALL wfs_update(snapshot=wf_history%past_states &
                         (wf_history%last_state_index)%snapshot, wf_history=wf_history, &
                         qs_env=qs_env, dt=dt)
      END IF
   END SUBROUTINE wfi_update

! **************************************************************************************************
!> \brief reorthogonalizes the mos
!> \param qs_env the qs_env in which to orthogonalize
!> \param v_matrix the vectors to orthogonalize
!> \param n_col number of column of v to orthogonalize
!> \par History
!>      04.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE reorthogonalize_vectors(qs_env, v_matrix, n_col)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_type), POINTER                          :: v_matrix
      INTEGER, INTENT(in), OPTIONAL                      :: n_col

      CHARACTER(len=*), PARAMETER :: routineN = 'reorthogonalize_vectors', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, my_n_col
      LOGICAL                                            :: has_unit_metric, &
                                                            ortho_contains_cholesky, &
                                                            smearing_is_used
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_pool_type), POINTER                     :: maxao_maxmo_fm_pool
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control

      NULLIFY (scf_env, scf_control, maxao_maxmo_fm_pool, matrix_s, mpools, dft_control)
      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_wf_history_methods.F",1040)
      IF(.NOT.(ASSOCIATED(v_matrix)))CALL cp__a("qs_wf_history_methods.F",1041)

      CALL cp_fm_get_info(v_matrix, ncol_global=my_n_col)
      IF (PRESENT(n_col)) my_n_col = n_col
      CALL get_qs_env(qs_env, mpools=mpools, &
                      scf_env=scf_env, &
                      scf_control=scf_control, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control)
      CALL mpools_get(mpools, maxao_maxmo_fm_pool=maxao_maxmo_fm_pool)
      IF (ASSOCIATED(scf_env)) THEN
         ortho_contains_cholesky = (scf_env%method /= ot_method_nr) .AND. &
                                   (scf_env%cholesky_method > 0) .AND. &
                                   ASSOCIATED(scf_env%ortho)
      ELSE
         ortho_contains_cholesky = .FALSE.
      END IF

      CALL get_qs_env(qs_env, has_unit_metric=has_unit_metric)
      smearing_is_used = .FALSE.
      IF (dft_control%smear) THEN
         smearing_is_used = .TRUE.
      END IF

      IF (has_unit_metric) THEN
         CALL make_basis_simple(v_matrix, my_n_col)
      ELSE IF (smearing_is_used) THEN
         CALL make_basis_lowdin(vmatrix=v_matrix, ncol=my_n_col, &
                                matrix_s=matrix_s(1)%matrix)
      ELSE IF (ortho_contains_cholesky) THEN
         CALL make_basis_cholesky(vmatrix=v_matrix, ncol=my_n_col, &
                                  ortho=scf_env%ortho)
      ELSE
         CALL make_basis_sm(v_matrix, my_n_col, matrix_s(1)%matrix)
      END IF
      CALL timestop(handle)
   END SUBROUTINE reorthogonalize_vectors

END MODULE qs_wf_history_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (30.11.2001) : new entries in setup_parameters_type
!>                         change name from input_file_name to coord_...
!>                         added topology file
!>                         added atom_names
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
!> \author CJM & JGH
! **************************************************************************************************
MODULE topology_types
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type
   USE colvar_types,                    ONLY: colvar_p_type,&
                                              colvar_release
   USE input_constants,                 ONLY: do_bondparm_covalent,&
                                              do_conn_generate,&
                                              do_constr_none,&
                                              do_skip_13
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

! **************************************************************************************************
   TYPE atom_info_type
      INTEGER, DIMENSION(:), POINTER         :: id_molname
      INTEGER, DIMENSION(:), POINTER         :: id_resname
      INTEGER, DIMENSION(:), POINTER         :: id_atmname
      INTEGER, DIMENSION(:), POINTER         :: id_atom_names
      INTEGER, DIMENSION(:), POINTER         :: id_element
      INTEGER, POINTER                       :: resid(:)
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: r
      INTEGER, POINTER                       :: map_mol_typ(:)
      INTEGER, POINTER                       :: map_mol_num(:)
      INTEGER, POINTER                       :: map_mol_res(:)
      REAL(KIND=dp), POINTER                 :: atm_charge(:)
      REAL(KIND=dp), POINTER                 :: atm_mass(:)
      REAL(KIND=dp), POINTER                 :: occup(:)
      REAL(KIND=dp), POINTER                 :: beta(:)
   END TYPE atom_info_type

! **************************************************************************************************
   TYPE connectivity_info_type
      INTEGER, POINTER :: bond_a(:), bond_b(:), bond_type(:)
      INTEGER, POINTER :: ub_a(:), ub_b(:), ub_c(:)
      INTEGER, POINTER :: theta_a(:), theta_b(:), theta_c(:), theta_type(:)
      INTEGER, POINTER :: phi_a(:), phi_b(:), phi_c(:), phi_d(:), phi_type(:)
      INTEGER, POINTER :: impr_a(:), impr_b(:), impr_c(:), impr_d(:), impr_type(:)
      INTEGER, POINTER :: onfo_a(:), onfo_b(:)
      INTEGER, POINTER :: c_bond_a(:), c_bond_b(:), c_bond_type(:)
   END TYPE connectivity_info_type

! **************************************************************************************************
   TYPE constraint_info_type
      ! Bonds involving Hydrogens
      LOGICAL                                       :: hbonds_restraint ! Restraints control
      REAL(KIND=dp)                                 :: hbonds_k0 ! Restraints control
      ! Fixed Atoms
      INTEGER                                       :: nfixed_atoms
      INTEGER, POINTER                              :: fixed_atoms(:), fixed_type(:), fixed_mol_type(:)
      LOGICAL, POINTER                              :: fixed_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: fixed_k0(:) ! Restraints control
      ! Freeze QM or MM
      INTEGER                                       :: freeze_qm, freeze_mm, freeze_qm_type, freeze_mm_type
      LOGICAL                                       :: fixed_mm_restraint, fixed_qm_restraint ! Restraints control
      REAL(KIND=dp)                                 :: fixed_mm_k0, fixed_qm_k0 ! Restraints control
      ! Freeze with molnames
      LOGICAL, POINTER                              :: fixed_mol_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: fixed_mol_k0(:) ! Restraints control
      CHARACTER(LEN=default_string_length), POINTER :: fixed_molnames(:)
      LOGICAL, POINTER, DIMENSION(:)                :: fixed_exclude_qm, fixed_exclude_mm
      ! Collective constraints
      INTEGER                                       :: nconst_colv
      INTEGER, POINTER                              :: const_colv_mol(:)
      CHARACTER(LEN=default_string_length), POINTER :: const_colv_molname(:)
      REAL(KIND=dp), POINTER                        :: const_colv_target(:)
      REAL(KIND=dp), POINTER                        :: const_colv_target_growth(:)
      TYPE(colvar_p_type), POINTER, DIMENSION(:)    :: colvar_set
      LOGICAL, POINTER                              :: colv_intermolecular(:)
      LOGICAL, POINTER                              :: colv_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: colv_k0(:) ! Restraints control
      LOGICAL, POINTER, DIMENSION(:)                :: colv_exclude_qm, colv_exclude_mm
      ! G3x3
      INTEGER                                       :: nconst_g33
      INTEGER, POINTER                              :: const_g33_mol(:)
      CHARACTER(LEN=default_string_length), POINTER :: const_g33_molname(:)
      INTEGER, POINTER                              :: const_g33_a(:)
      INTEGER, POINTER                              :: const_g33_b(:)
      INTEGER, POINTER                              :: const_g33_c(:)
      REAL(KIND=dp), POINTER                        :: const_g33_dab(:)
      REAL(KIND=dp), POINTER                        :: const_g33_dac(:)
      REAL(KIND=dp), POINTER                        :: const_g33_dbc(:)
      LOGICAL, POINTER                              :: g33_intermolecular(:)
      LOGICAL, POINTER                              :: g33_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: g33_k0(:) ! Restraints control
      LOGICAL, POINTER, DIMENSION(:)                :: g33_exclude_qm, g33_exclude_mm
      ! G4x6
      INTEGER                                       :: nconst_g46
      INTEGER, POINTER                              :: const_g46_mol(:)
      CHARACTER(LEN=default_string_length), POINTER :: const_g46_molname(:)
      INTEGER, POINTER                              :: const_g46_a(:)
      INTEGER, POINTER                              :: const_g46_b(:)
      INTEGER, POINTER                              :: const_g46_c(:)
      INTEGER, POINTER                              :: const_g46_d(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dab(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dac(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dbc(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dad(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dbd(:)
      REAL(KIND=dp), POINTER                        :: const_g46_dcd(:)
      LOGICAL, POINTER                              :: g46_intermolecular(:)
      LOGICAL, POINTER                              :: g46_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: g46_k0(:) ! Restraints control
      LOGICAL, POINTER, DIMENSION(:)                :: g46_exclude_qm, g46_exclude_mm
      ! virtual_site
      INTEGER                                       :: nconst_vsite
      INTEGER, POINTER                              :: const_vsite_mol(:)
      CHARACTER(LEN=default_string_length), POINTER :: const_vsite_molname(:)
      INTEGER, POINTER                              :: const_vsite_a(:)
      INTEGER, POINTER                              :: const_vsite_b(:)
      INTEGER, POINTER                              :: const_vsite_c(:)
      INTEGER, POINTER                              :: const_vsite_d(:)
      REAL(KIND=dp), POINTER                        :: const_vsite_wbc(:)
      REAL(KIND=dp), POINTER                        :: const_vsite_wdc(:)
      LOGICAL, POINTER                              :: vsite_intermolecular(:)
      LOGICAL, POINTER                              :: vsite_restraint(:) ! Restraints control
      REAL(KIND=dp), POINTER                        :: vsite_k0(:) ! Restraints control
      LOGICAL, POINTER, DIMENSION(:)                :: vsite_exclude_qm, vsite_exclude_mm
   END TYPE constraint_info_type

! **************************************************************************************************
   TYPE topology_parameters_type
      TYPE(atom_info_type), POINTER             :: atom_info
      TYPE(connectivity_info_type), POINTER     :: conn_info
      TYPE(constraint_info_type), POINTER       :: cons_info
      TYPE(cell_type), POINTER                 :: cell, cell_ref, cell_muc
      INTEGER                                   :: conn_type
      INTEGER                                   :: coord_type
      INTEGER                                   :: exclude_vdw
      INTEGER                                   :: exclude_ei
      INTEGER                                   :: bondparm_type
      !TRY TO REMOVE THIS FIVE VARIABLE IN THE FUTURE
      INTEGER                                   :: natoms, natom_type
      INTEGER                                   :: nmol, nmol_type, nmol_conn
      !TRY TO REMOVE THIS FIVE VARIABLE IN THE FUTURE
      LOGICAL                                   :: aa_element
      LOGICAL                                   :: molname_generated
      REAL(KIND=dp)                             :: bondparm_factor
      LOGICAL                                   :: create_molecules
      LOGICAL                                   :: reorder_atom
      LOGICAL                                   :: molecules_check
      LOGICAL                                   :: coordinate
      LOGICAL                                   :: use_g96_velocity
      CHARACTER(LEN=default_path_length)       :: coord_file_name
      CHARACTER(LEN=default_path_length)       :: conn_file_name
      LOGICAL                                   :: const_atom
      LOGICAL                                   :: const_hydr
      LOGICAL                                   :: const_colv
      LOGICAL                                   :: const_33
      LOGICAL                                   :: const_46
      LOGICAL                                   :: const_vsite
      LOGICAL                                   :: charge_occup
      LOGICAL                                   :: charge_beta
      LOGICAL                                   :: charge_extended
      LOGICAL                                   :: para_res
   END TYPE topology_parameters_type

! **************************************************************************************************
   TYPE constr_list_type
      INTEGER, DIMENSION(:), POINTER :: constr
   END TYPE constr_list_type

   PUBLIC :: atom_info_type, &
             connectivity_info_type, &
             constraint_info_type, &
             topology_parameters_type, &
             constr_list_type

   PUBLIC :: init_topology, &
             deallocate_topology, &
             pre_read_topology

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_types'
   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief 1. Just NULLIFY and zero all the stuff
!> \param topology ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE init_topology(topology)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_topology', routineP = moduleN//':'//routineN

!-----------------------------------------------------------------------------
! 1. Nullify and allocate things in topology
!-----------------------------------------------------------------------------

      ALLOCATE (topology%atom_info)
      ALLOCATE (topology%conn_info)
      ALLOCATE (topology%cons_info)
      !-----------------------------------------------------------------------------
      ! 2. Initialize and Nullify things in topology
      !-----------------------------------------------------------------------------
      NULLIFY (topology%cell, topology%cell_ref, topology%cell_muc)
      topology%natoms = 0
      topology%natom_type = 0
      topology%nmol = 0
      topology%nmol_type = 0
      topology%nmol_conn = 0
      topology%bondparm_type = do_bondparm_covalent
      topology%reorder_atom = .FALSE.
      topology%create_molecules = .FALSE.
      topology%molecules_check = .FALSE.
      topology%coordinate = .FALSE.
      topology%use_g96_velocity = .FALSE.
      topology%coord_type = -1
      topology%coord_file_name = ''
      topology%conn_type = do_conn_generate
      topology%conn_file_name = 'OFF'
      topology%const_atom = .FALSE.
      topology%const_hydr = .FALSE.
      topology%const_colv = .FALSE.
      topology%const_33 = .FALSE.
      topology%const_46 = .FALSE.
      topology%const_vsite = .FALSE.
      topology%charge_occup = .FALSE.
      topology%charge_beta = .FALSE.
      topology%charge_extended = .FALSE.
      topology%para_res = .FALSE.
      topology%molname_generated = .FALSE.
      topology%aa_element = .FALSE.
      topology%exclude_vdw = do_skip_13
      topology%exclude_ei = do_skip_13
      !-----------------------------------------------------------------------------
      ! 3. Initialize and Nullify things in topology%atom_info
      !-----------------------------------------------------------------------------
      NULLIFY (topology%atom_info%id_molname)
      NULLIFY (topology%atom_info%id_resname)
      NULLIFY (topology%atom_info%resid)
      NULLIFY (topology%atom_info%id_atmname)
      NULLIFY (topology%atom_info%id_atom_names)
      NULLIFY (topology%atom_info%r)
      NULLIFY (topology%atom_info%map_mol_typ)
      NULLIFY (topology%atom_info%map_mol_num)
      NULLIFY (topology%atom_info%map_mol_res)
      NULLIFY (topology%atom_info%atm_charge)
      NULLIFY (topology%atom_info%atm_mass)
      NULLIFY (topology%atom_info%occup)
      NULLIFY (topology%atom_info%beta)
      NULLIFY (topology%atom_info%id_element)
      !-----------------------------------------------------------------------------
      ! 4. Initialize and Nullify things in topology%conn_info
      !-----------------------------------------------------------------------------
      NULLIFY (topology%conn_info%bond_a)
      NULLIFY (topology%conn_info%bond_b)
      NULLIFY (topology%conn_info%bond_type)
      NULLIFY (topology%conn_info%ub_a)
      NULLIFY (topology%conn_info%ub_b)
      NULLIFY (topology%conn_info%ub_c)
      NULLIFY (topology%conn_info%theta_a)
      NULLIFY (topology%conn_info%theta_b)
      NULLIFY (topology%conn_info%theta_c)
      NULLIFY (topology%conn_info%theta_type)
      NULLIFY (topology%conn_info%phi_a)
      NULLIFY (topology%conn_info%phi_b)
      NULLIFY (topology%conn_info%phi_c)
      NULLIFY (topology%conn_info%phi_d)
      NULLIFY (topology%conn_info%phi_type)
      NULLIFY (topology%conn_info%impr_a)
      NULLIFY (topology%conn_info%impr_b)
      NULLIFY (topology%conn_info%impr_c)
      NULLIFY (topology%conn_info%impr_d)
      NULLIFY (topology%conn_info%impr_type)
      NULLIFY (topology%conn_info%onfo_a)
      NULLIFY (topology%conn_info%onfo_b)
      NULLIFY (topology%conn_info%c_bond_a)
      NULLIFY (topology%conn_info%c_bond_b)
      NULLIFY (topology%conn_info%c_bond_type)
      !-----------------------------------------------------------------------------
      ! 5. Initialize and Nullify things in topology%cons_info
      !-----------------------------------------------------------------------------
      CALL init_constraint(topology%cons_info)
   END SUBROUTINE init_topology

! **************************************************************************************************
!> \brief 1. Just NULLIFY and zero all the stuff
!> \param constraint_info ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE init_constraint(constraint_info)
      TYPE(constraint_info_type), POINTER                :: constraint_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_constraint', &
         routineP = moduleN//':'//routineN

! Bonds involving Hydrogens

      constraint_info%hbonds_restraint = .FALSE.
      ! Fixed Atoms
      constraint_info%nfixed_atoms = 0
      constraint_info%freeze_mm = do_constr_none
      constraint_info%freeze_qm = do_constr_none
      NULLIFY (constraint_info%fixed_atoms)
      NULLIFY (constraint_info%fixed_type)
      NULLIFY (constraint_info%fixed_mol_type)
      NULLIFY (constraint_info%fixed_molnames)
      NULLIFY (constraint_info%fixed_restraint)
      NULLIFY (constraint_info%fixed_k0)
      NULLIFY (constraint_info%fixed_mol_restraint)
      NULLIFY (constraint_info%fixed_mol_k0)
      NULLIFY (constraint_info%fixed_exclude_qm, constraint_info%fixed_exclude_mm)
      ! Collective Constraints
      constraint_info%nconst_colv = 0
      NULLIFY (constraint_info%colvar_set)
      NULLIFY (constraint_info%const_colv_mol)
      NULLIFY (constraint_info%const_colv_molname)
      NULLIFY (constraint_info%const_colv_target)
      NULLIFY (constraint_info%const_colv_target_growth)
      NULLIFY (constraint_info%colv_intermolecular)
      NULLIFY (constraint_info%colv_restraint)
      NULLIFY (constraint_info%colv_k0)
      NULLIFY (constraint_info%colv_exclude_qm, constraint_info%colv_exclude_mm)
      ! G3x3
      constraint_info%nconst_g33 = 0
      NULLIFY (constraint_info%const_g33_mol)
      NULLIFY (constraint_info%const_g33_molname)
      NULLIFY (constraint_info%const_g33_a)
      NULLIFY (constraint_info%const_g33_b)
      NULLIFY (constraint_info%const_g33_c)
      NULLIFY (constraint_info%const_g33_dab)
      NULLIFY (constraint_info%const_g33_dac)
      NULLIFY (constraint_info%const_g33_dbc)
      NULLIFY (constraint_info%g33_intermolecular)
      NULLIFY (constraint_info%g33_restraint)
      NULLIFY (constraint_info%g33_k0)
      NULLIFY (constraint_info%g33_exclude_qm, constraint_info%g33_exclude_mm)
      ! G4x6
      constraint_info%nconst_g46 = 0
      NULLIFY (constraint_info%const_g46_mol)
      NULLIFY (constraint_info%const_g46_molname)
      NULLIFY (constraint_info%const_g46_a)
      NULLIFY (constraint_info%const_g46_b)
      NULLIFY (constraint_info%const_g46_c)
      NULLIFY (constraint_info%const_g46_d)
      NULLIFY (constraint_info%const_g46_dab)
      NULLIFY (constraint_info%const_g46_dac)
      NULLIFY (constraint_info%const_g46_dbc)
      NULLIFY (constraint_info%const_g46_dad)
      NULLIFY (constraint_info%const_g46_dbd)
      NULLIFY (constraint_info%const_g46_dcd)
      NULLIFY (constraint_info%g46_intermolecular)
      NULLIFY (constraint_info%g46_restraint)
      NULLIFY (constraint_info%g46_k0)
      NULLIFY (constraint_info%g46_exclude_qm, constraint_info%g46_exclude_mm)
      ! virtual_site
      constraint_info%nconst_vsite = 0
      NULLIFY (constraint_info%const_vsite_mol)
      NULLIFY (constraint_info%const_vsite_molname)
      NULLIFY (constraint_info%const_vsite_a)
      NULLIFY (constraint_info%const_vsite_b)
      NULLIFY (constraint_info%const_vsite_c)
      NULLIFY (constraint_info%const_vsite_d)
      NULLIFY (constraint_info%const_vsite_wbc)
      NULLIFY (constraint_info%const_vsite_wdc)
      NULLIFY (constraint_info%vsite_intermolecular)
      NULLIFY (constraint_info%vsite_restraint)
      NULLIFY (constraint_info%vsite_k0)
      NULLIFY (constraint_info%vsite_exclude_qm, constraint_info%vsite_exclude_mm)

   END SUBROUTINE init_constraint

! **************************************************************************************************
!> \brief 1. Just DEALLOCATE all the stuff
!> \param topology ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE deallocate_topology(topology)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_topology', &
         routineP = moduleN//':'//routineN

!-----------------------------------------------------------------------------
! 1. DEALLOCATE things in topology%atom_info
!-----------------------------------------------------------------------------

      IF (ASSOCIATED(topology%atom_info%id_molname)) THEN
         DEALLOCATE (topology%atom_info%id_molname)
      END IF
      IF (ASSOCIATED(topology%atom_info%id_resname)) THEN
         DEALLOCATE (topology%atom_info%id_resname)
      END IF
      IF (ASSOCIATED(topology%atom_info%resid)) THEN
         DEALLOCATE (topology%atom_info%resid)
      END IF
      IF (ASSOCIATED(topology%atom_info%id_atmname)) THEN
         DEALLOCATE (topology%atom_info%id_atmname)
      END IF
      IF (ASSOCIATED(topology%atom_info%id_atom_names)) THEN
         DEALLOCATE (topology%atom_info%id_atom_names)
      END IF
      IF (ASSOCIATED(topology%atom_info%r)) THEN
         DEALLOCATE (topology%atom_info%r)
      END IF
      IF (ASSOCIATED(topology%atom_info%map_mol_typ)) THEN
         DEALLOCATE (topology%atom_info%map_mol_typ)
      END IF
      IF (ASSOCIATED(topology%atom_info%map_mol_num)) THEN
         DEALLOCATE (topology%atom_info%map_mol_num)
      END IF
      IF (ASSOCIATED(topology%atom_info%map_mol_res)) THEN
         DEALLOCATE (topology%atom_info%map_mol_res)
      END IF
      IF (ASSOCIATED(topology%atom_info%atm_charge)) THEN
         DEALLOCATE (topology%atom_info%atm_charge)
      END IF
      IF (ASSOCIATED(topology%atom_info%atm_mass)) THEN
         DEALLOCATE (topology%atom_info%atm_mass)
      END IF
      IF (ASSOCIATED(topology%atom_info%occup)) THEN
         DEALLOCATE (topology%atom_info%occup)
      END IF
      IF (ASSOCIATED(topology%atom_info%beta)) THEN
         DEALLOCATE (topology%atom_info%beta)
      END IF
      IF (ASSOCIATED(topology%atom_info%id_element)) THEN
         DEALLOCATE (topology%atom_info%id_element)
      END IF
      !-----------------------------------------------------------------------------
      ! 2. DEALLOCATE things in topology%conn_info
      !-----------------------------------------------------------------------------
      IF (ASSOCIATED(topology%conn_info%bond_a)) THEN
         DEALLOCATE (topology%conn_info%bond_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%bond_b)) THEN
         DEALLOCATE (topology%conn_info%bond_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%bond_type)) THEN
         DEALLOCATE (topology%conn_info%bond_type)
      END IF
      IF (ASSOCIATED(topology%conn_info%ub_a)) THEN
         DEALLOCATE (topology%conn_info%ub_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%ub_b)) THEN
         DEALLOCATE (topology%conn_info%ub_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%ub_c)) THEN
         DEALLOCATE (topology%conn_info%ub_c)
      END IF
      IF (ASSOCIATED(topology%conn_info%theta_a)) THEN
         DEALLOCATE (topology%conn_info%theta_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%theta_b)) THEN
         DEALLOCATE (topology%conn_info%theta_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%theta_c)) THEN
         DEALLOCATE (topology%conn_info%theta_c)
      END IF
      IF (ASSOCIATED(topology%conn_info%theta_type)) THEN
         DEALLOCATE (topology%conn_info%theta_type)
      END IF
      IF (ASSOCIATED(topology%conn_info%phi_a)) THEN
         DEALLOCATE (topology%conn_info%phi_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%phi_b)) THEN
         DEALLOCATE (topology%conn_info%phi_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%phi_c)) THEN
         DEALLOCATE (topology%conn_info%phi_c)
      END IF
      IF (ASSOCIATED(topology%conn_info%phi_d)) THEN
         DEALLOCATE (topology%conn_info%phi_d)
      END IF
      IF (ASSOCIATED(topology%conn_info%phi_type)) THEN
         DEALLOCATE (topology%conn_info%phi_type)
      END IF
      IF (ASSOCIATED(topology%conn_info%impr_a)) THEN
         DEALLOCATE (topology%conn_info%impr_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%impr_b)) THEN
         DEALLOCATE (topology%conn_info%impr_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%impr_c)) THEN
         DEALLOCATE (topology%conn_info%impr_c)
      END IF
      IF (ASSOCIATED(topology%conn_info%impr_d)) THEN
         DEALLOCATE (topology%conn_info%impr_d)
      END IF
      IF (ASSOCIATED(topology%conn_info%impr_type)) THEN
         DEALLOCATE (topology%conn_info%impr_type)
      END IF
      IF (ASSOCIATED(topology%conn_info%onfo_a)) THEN
         DEALLOCATE (topology%conn_info%onfo_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%onfo_b)) THEN
         DEALLOCATE (topology%conn_info%onfo_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%c_bond_a)) THEN
         DEALLOCATE (topology%conn_info%c_bond_a)
      END IF
      IF (ASSOCIATED(topology%conn_info%c_bond_b)) THEN
         DEALLOCATE (topology%conn_info%c_bond_b)
      END IF
      IF (ASSOCIATED(topology%conn_info%c_bond_type)) THEN
         DEALLOCATE (topology%conn_info%c_bond_type)
      END IF
      !-----------------------------------------------------------------------------
      ! 3. DEALLOCATE things in topology%cons_info
      !-----------------------------------------------------------------------------
      IF (ASSOCIATED(topology%cons_info)) &
         CALL deallocate_constraint(topology%cons_info)
      !-----------------------------------------------------------------------------
      ! 4. DEALLOCATE things in topology
      !-----------------------------------------------------------------------------
      CALL cell_release(topology%cell)
      CALL cell_release(topology%cell_ref)
      CALL cell_release(topology%cell_muc)
      IF (ASSOCIATED(topology%atom_info)) THEN
         DEALLOCATE (topology%atom_info)
      END IF
      IF (ASSOCIATED(topology%conn_info)) THEN
         DEALLOCATE (topology%conn_info)
      END IF
      IF (ASSOCIATED(topology%cons_info)) THEN
         DEALLOCATE (topology%cons_info)
      END IF

   END SUBROUTINE deallocate_topology

! **************************************************************************************************
!> \brief 1. Just DEALLOCATE all the stuff
!> \param constraint_info ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE deallocate_constraint(constraint_info)
      TYPE(constraint_info_type), POINTER                :: constraint_info

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_constraint', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

! Fixed Atoms

      IF (ASSOCIATED(constraint_info%fixed_atoms)) THEN
         DEALLOCATE (constraint_info%fixed_atoms)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_type)) THEN
         DEALLOCATE (constraint_info%fixed_type)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_molnames)) THEN
         DEALLOCATE (constraint_info%fixed_molnames)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_mol_type)) THEN
         DEALLOCATE (constraint_info%fixed_mol_type)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_restraint)) THEN
         DEALLOCATE (constraint_info%fixed_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_k0)) THEN
         DEALLOCATE (constraint_info%fixed_k0)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_mol_restraint)) THEN
         DEALLOCATE (constraint_info%fixed_mol_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_mol_k0)) THEN
         DEALLOCATE (constraint_info%fixed_mol_k0)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_exclude_qm)) THEN
         DEALLOCATE (constraint_info%fixed_exclude_qm)
      END IF
      IF (ASSOCIATED(constraint_info%fixed_exclude_mm)) THEN
         DEALLOCATE (constraint_info%fixed_exclude_mm)
      END IF
      ! Collective Constraint
      IF (ASSOCIATED(constraint_info%colvar_set)) THEN
         DO i = 1, SIZE(constraint_info%colvar_set)
            IF (ASSOCIATED(constraint_info%colvar_set(i)%colvar)) THEN
               CALL colvar_release(constraint_info%colvar_set(i)%colvar)
               NULLIFY (constraint_info%colvar_set(i)%colvar)
            END IF
         END DO
         DEALLOCATE (constraint_info%colvar_set)
      END IF
      IF (ASSOCIATED(constraint_info%const_colv_mol)) THEN
         DEALLOCATE (constraint_info%const_colv_mol)
      END IF
      IF (ASSOCIATED(constraint_info%const_colv_molname)) THEN
         DEALLOCATE (constraint_info%const_colv_molname)
      END IF
      IF (ASSOCIATED(constraint_info%const_colv_target)) THEN
         DEALLOCATE (constraint_info%const_colv_target)
      END IF
      IF (ASSOCIATED(constraint_info%const_colv_target_growth)) THEN
         DEALLOCATE (constraint_info%const_colv_target_growth)
      END IF
      IF (ASSOCIATED(constraint_info%colv_intermolecular)) THEN
         DEALLOCATE (constraint_info%colv_intermolecular)
      END IF
      IF (ASSOCIATED(constraint_info%colv_restraint)) THEN
         DEALLOCATE (constraint_info%colv_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%colv_k0)) THEN
         DEALLOCATE (constraint_info%colv_k0)
      END IF
      IF (ASSOCIATED(constraint_info%colv_exclude_qm)) THEN
         DEALLOCATE (constraint_info%colv_exclude_qm)
      END IF
      IF (ASSOCIATED(constraint_info%colv_exclude_mm)) THEN
         DEALLOCATE (constraint_info%colv_exclude_mm)
      END IF
      ! G3x3
      IF (ASSOCIATED(constraint_info%const_g33_mol)) THEN
         DEALLOCATE (constraint_info%const_g33_mol)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_molname)) THEN
         DEALLOCATE (constraint_info%const_g33_molname)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_a)) THEN
         DEALLOCATE (constraint_info%const_g33_a)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_b)) THEN
         DEALLOCATE (constraint_info%const_g33_b)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_c)) THEN
         DEALLOCATE (constraint_info%const_g33_c)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_dab)) THEN
         DEALLOCATE (constraint_info%const_g33_dab)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_dac)) THEN
         DEALLOCATE (constraint_info%const_g33_dac)
      END IF
      IF (ASSOCIATED(constraint_info%const_g33_dbc)) THEN
         DEALLOCATE (constraint_info%const_g33_dbc)
      END IF
      IF (ASSOCIATED(constraint_info%g33_intermolecular)) THEN
         DEALLOCATE (constraint_info%g33_intermolecular)
      END IF
      IF (ASSOCIATED(constraint_info%g33_restraint)) THEN
         DEALLOCATE (constraint_info%g33_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%g33_k0)) THEN
         DEALLOCATE (constraint_info%g33_k0)
      END IF
      IF (ASSOCIATED(constraint_info%g33_exclude_qm)) THEN
         DEALLOCATE (constraint_info%g33_exclude_qm)
      END IF
      IF (ASSOCIATED(constraint_info%g33_exclude_mm)) THEN
         DEALLOCATE (constraint_info%g33_exclude_mm)
      END IF
      ! G4x6
      IF (ASSOCIATED(constraint_info%const_g46_mol)) THEN
         DEALLOCATE (constraint_info%const_g46_mol)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_molname)) THEN
         DEALLOCATE (constraint_info%const_g46_molname)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_a)) THEN
         DEALLOCATE (constraint_info%const_g46_a)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_b)) THEN
         DEALLOCATE (constraint_info%const_g46_b)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_c)) THEN
         DEALLOCATE (constraint_info%const_g46_c)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_d)) THEN
         DEALLOCATE (constraint_info%const_g46_d)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dab)) THEN
         DEALLOCATE (constraint_info%const_g46_dab)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dac)) THEN
         DEALLOCATE (constraint_info%const_g46_dac)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dbc)) THEN
         DEALLOCATE (constraint_info%const_g46_dbc)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dad)) THEN
         DEALLOCATE (constraint_info%const_g46_dad)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dbd)) THEN
         DEALLOCATE (constraint_info%const_g46_dbd)
      END IF
      IF (ASSOCIATED(constraint_info%const_g46_dcd)) THEN
         DEALLOCATE (constraint_info%const_g46_dcd)
      END IF
      IF (ASSOCIATED(constraint_info%g46_intermolecular)) THEN
         DEALLOCATE (constraint_info%g46_intermolecular)
      END IF
      IF (ASSOCIATED(constraint_info%g46_restraint)) THEN
         DEALLOCATE (constraint_info%g46_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%g46_k0)) THEN
         DEALLOCATE (constraint_info%g46_k0)
      END IF
      IF (ASSOCIATED(constraint_info%g46_exclude_qm)) THEN
         DEALLOCATE (constraint_info%g46_exclude_qm)
      END IF
      IF (ASSOCIATED(constraint_info%g46_exclude_mm)) THEN
         DEALLOCATE (constraint_info%g46_exclude_mm)
      END IF
      ! virtual_site
      IF (ASSOCIATED(constraint_info%const_vsite_mol)) THEN
         DEALLOCATE (constraint_info%const_vsite_mol)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_molname)) THEN
         DEALLOCATE (constraint_info%const_vsite_molname)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_a)) THEN
         DEALLOCATE (constraint_info%const_vsite_a)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_b)) THEN
         DEALLOCATE (constraint_info%const_vsite_b)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_c)) THEN
         DEALLOCATE (constraint_info%const_vsite_c)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_d)) THEN
         DEALLOCATE (constraint_info%const_vsite_d)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_wbc)) THEN
         DEALLOCATE (constraint_info%const_vsite_wbc)
      END IF
      IF (ASSOCIATED(constraint_info%const_vsite_wdc)) THEN
         DEALLOCATE (constraint_info%const_vsite_wdc)
      END IF
      IF (ASSOCIATED(constraint_info%vsite_intermolecular)) THEN
         DEALLOCATE (constraint_info%vsite_intermolecular)
      END IF
      IF (ASSOCIATED(constraint_info%vsite_restraint)) THEN
         DEALLOCATE (constraint_info%vsite_restraint)
      END IF
      IF (ASSOCIATED(constraint_info%vsite_k0)) THEN
         DEALLOCATE (constraint_info%vsite_k0)
      END IF
      IF (ASSOCIATED(constraint_info%vsite_exclude_qm)) THEN
         DEALLOCATE (constraint_info%vsite_exclude_qm)
      END IF
      IF (ASSOCIATED(constraint_info%vsite_exclude_mm)) THEN
         DEALLOCATE (constraint_info%vsite_exclude_mm)
      END IF
   END SUBROUTINE deallocate_constraint

! **************************************************************************************************
!> \brief Deallocate possibly allocated arrays before reading topology
!> \param topology ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE pre_read_topology(topology)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology

      CHARACTER(len=*), PARAMETER :: routineN = 'pre_read_topology', &
         routineP = moduleN//':'//routineN

      TYPE(atom_info_type), POINTER                      :: atom_info

      atom_info => topology%atom_info

      IF (ASSOCIATED(atom_info%id_molname)) THEN
         DEALLOCATE (atom_info%id_molname)
      END IF

      IF (ASSOCIATED(atom_info%resid)) THEN
         DEALLOCATE (atom_info%resid)
      END IF

      IF (ASSOCIATED(atom_info%id_resname)) THEN
         DEALLOCATE (atom_info%id_resname)
      END IF

      IF (ASSOCIATED(atom_info%id_atmname)) THEN
         DEALLOCATE (atom_info%id_atmname)
      END IF

      IF (ASSOCIATED(atom_info%atm_charge)) THEN
         DEALLOCATE (atom_info%atm_charge)
      END IF

      IF (ASSOCIATED(atom_info%atm_mass)) THEN
         DEALLOCATE (atom_info%atm_mass)
      END IF

   END SUBROUTINE pre_read_topology

END MODULE topology_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief generates a unique id number for a string (str2id) that can be used
!>      two compare two strings. I.e.
!>      if (str1==str2) => str2id(str1)==str2id(str2)
!>      if (str1.NE.str2) => str2id(str1).NE.str2id(str2)
!>      and the other way around. Given an id, the string can be retrieved.
!> \note
!>      the purpose of this routine is to speed up string handling,
!>      string searching, ... as an operation on an int is much faster than an
!>      operation on a long string.
!> \par History
!>      9.2006 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE string_table

   USE kinds,                           ONLY: default_string_length,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   ! user functions
   PUBLIC :: str2id, id2str, s2s

   ! setup function
   PUBLIC :: string_table_allocate, string_table_deallocate

   PRIVATE
   ! For good performance, the hash table should be larger than the largest number
   ! of strings that will be saved, but the memory for an empty table is 16*hash_table_size
   ! the string_table should remain functional for up to ~ 2**32 strings
   INTEGER, PARAMETER :: Nbit = 16
   INTEGER, PARAMETER :: hash_table_size = 2**Nbit

   ! actual elements in the hash table
   INTEGER, SAVE      :: actual_strings
   INTEGER, SAVE      :: inserted_strings

   ! an element of the linked list of hashed strings
! **************************************************************************************************
   TYPE hash_element_type
      CHARACTER(LEN=default_string_length), POINTER :: str => NULL()
      TYPE(hash_element_type), POINTER :: next => NULL()
   END TYPE

   ! the array of linked lists of hashed strings
   TYPE(hash_element_type), SAVE, ALLOCATABLE, TARGET, DIMENSION(:) :: hash_table

CONTAINS

! **************************************************************************************************
!> \brief returns a unique id for a given string, and stores the string for
!>      later retrieval using the id.
!> \param str the string to be stored (default_string_length)
!> \retval id ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      pass literal strings using the s2s function,
!>      which converts strings of any length to default_string_length
!>      id=str2id(s2s("my short string"))
! **************************************************************************************************
   FUNCTION str2id(str) RESULT(id)
      CHARACTER(LEN=*)                                   :: str
      INTEGER                                            :: id

      INTEGER                                            :: index, ipos
      TYPE(hash_element_type), POINTER                   :: this

      inserted_strings = inserted_strings+1
      ! index is the index in the array, ipos is the Nth element of the linked list
      index = joaat_hash(str)
      ipos = 0
      this => hash_table(index)
      DO ! walk the list
         IF (.NOT. ASSOCIATED(this%str)) THEN
            ! str was not in the linked list, add it now
            ALLOCATE (this%str)
            this%str = str
            actual_strings = actual_strings+1
            EXIT
         ELSE
            IF (this%str == str) THEN
               ! str is in the list already
               EXIT
            ELSE
               IF (.NOT. ASSOCIATED(this%next)) ALLOCATE (this%next)
               ipos = ipos+1
               this => this%next
            ENDIF
         ENDIF
      ENDDO
      id = IOR(index, ISHFT(ipos, Nbit))
   END FUNCTION str2id

! **************************************************************************************************
!> \brief returns the string associated with a given id
!> \param id the id to be converted into a string
!> \retval str ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      only id's of previously 'registered' strings (str2id) should be passed,
!>      otherwise things crash
! **************************************************************************************************
   FUNCTION id2str(id) RESULT(str)
      INTEGER                                            :: id
      CHARACTER(LEN=default_string_length)               :: str

      INTEGER                                            :: i, index, ipos
      TYPE(hash_element_type), POINTER                   :: this

      index = IAND(id, 2**Nbit-1)
      ipos = ISHFT(id, -Nbit)
      this => hash_table(index)
      DO i = 1, ipos
         this => this%next
      ENDDO
      str = this%str
   END FUNCTION id2str

! **************************************************************************************************
!> \brief converts a string in a string of default_string_length
!> \param str ...
!> \retval res ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      useful to pass a literal string to str2id
!>      i.e. id=str2id(s2s("X"))
! **************************************************************************************************
   FUNCTION s2s(str) RESULT(res)
      CHARACTER(LEN=*)                                   :: str
      CHARACTER(LEN=default_string_length)               :: res

      res = str
   END FUNCTION s2s

! **************************************************************************************************
!> \brief allocates the string table
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      this needs to be done only once at program startup, before any use
!>      of other procedures of this module. The scope of this table is global
! **************************************************************************************************
   SUBROUTINE string_table_allocate()
      ALLOCATE (hash_table(0:hash_table_size-1))
      actual_strings = 0
      inserted_strings = 0
   END SUBROUTINE string_table_allocate

! **************************************************************************************************
!> \brief deallocates the string table
!> \param iw a unit to which some info about the table usage can be printed
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      This should be done before program termination, all associated ids become meaningless
! **************************************************************************************************
   SUBROUTINE string_table_deallocate(iw)
      INTEGER, INTENT(IN)                                :: iw

      INTEGER                                            :: i, ilist, ipos, ipos_max
      TYPE(hash_element_type), POINTER                   :: next, this

! clean up all the linked lists of entries

      ipos_max = 0
      ilist = 0
      DO i = 0, hash_table_size-1
         ipos = 1
         IF (ASSOCIATED(hash_table(i)%str)) THEN
            DEALLOCATE (hash_table(i)%str)
            ilist = ilist+1
         ENDIF
         this => hash_table(i)%next
         DO WHILE (ASSOCIATED(this))
            ipos = ipos+1
            next => this%next
            IF (ASSOCIATED(this%str)) DEALLOCATE (this%str)
            DEALLOCATE (this)
            this => next
         ENDDO
         ipos_max = MAX(ipos_max, ipos)
      ENDDO
      DEALLOCATE (hash_table)
      IF (iw > 0) THEN
         WRITE (iw, *) "string table: # inserted str = ", inserted_strings
         WRITE (iw, *) "              # actual       = ", actual_strings
         WRITE (iw, *) "              # lists        = ", ilist, " / ", hash_table_size
         WRITE (iw, *) "              longest list   = ", ipos_max
      ENDIF
      actual_strings = 0
      inserted_strings = 0
   END SUBROUTINE string_table_deallocate

! **************************************************************************************************
!> \brief generates the hash of a string and the index in the table
!> \param key a string of any length
!> \retval hash_index ...
!> \par History
!>       09.2006 created [Joost VandeVondele]
!> \note
!>       http://en.wikipedia.org/wiki/Hash_table
!>       http://www.burtleburtle.net/bob/hash/doobs.html
!>       However, since fortran doesn't have an unsigned 4 byte int
!>       we compute it using an integer with the appropriate range
!>       we return already the index in the table as a final result
! **************************************************************************************************
   FUNCTION joaat_hash(key) RESULT(hash_index)
      CHARACTER(LEN=*), INTENT(IN)                       :: key
      INTEGER                                            :: hash_index

      INTEGER(KIND=int_8), PARAMETER                     :: b32 = 2_int_8**32-1_int_8

      INTEGER                                            :: i
      INTEGER(KIND=int_8)                                :: hash

      hash = 0_int_8
      DO i = 1, LEN(key)
         hash = IAND(hash+ICHAR(key(i:i)), b32)
         hash = IAND(hash+IAND(ISHFT(hash, 10), b32), b32)
         hash = IAND(IEOR(hash, IAND(ISHFT(hash, -6), b32)), b32)
      ENDDO
      hash = IAND(hash+IAND(ISHFT(hash, 3), b32), b32)
      hash = IAND(IEOR(hash, IAND(ISHFT(hash, -11), b32)), b32)
      hash = IAND(hash+IAND(ISHFT(hash, 15), b32), b32)
      ! hash is the real 32bit hash value of the string,
      ! hash_index is an index in the hash_table
      hash_index = INT(MOD(hash, INT(hash_table_size, KIND=int_8)))
   END FUNCTION joaat_hash
END MODULE string_table
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of subroutine needed for topology related things
!> \par History
!>     jgh (23-05-2004) Last atom of molecule information added
! **************************************************************************************************

MODULE topology_connectivity_util
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE force_field_kind_types,          ONLY: do_ff_charmm,&
                                              do_ff_harmonic
   USE input_constants,                 ONLY: do_conn_g87,&
                                              do_conn_g96,&
                                              do_conn_user
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length
   USE memory_utilities,                ONLY: reallocate
   USE molecule_kind_types,             ONLY: &
        allocate_molecule_kind_set, atom_type, bend_type, bond_type, get_molecule_kind, impr_type, &
        molecule_kind_type, opbend_type, set_molecule_kind, torsion_type, ub_type
   USE molecule_types_new,              ONLY: allocate_molecule_set,&
                                              get_molecule,&
                                              local_molecule_type,&
                                              molecule_type,&
                                              set_molecule,&
                                              set_molecule_set
   USE string_table,                    ONLY: id2str
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_connectivity_util'

   PRIVATE
   PUBLIC :: topology_connectivity_pack, topology_conn_multiple

CONTAINS

! **************************************************************************************************
!> \brief topology connectivity pack
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param topology ...
!> \param subsys_section ...
!> \par History 11/2009 (Louis Vanduyhuys): added Out of Plane bends based on
!>                                      impropers in topology
! **************************************************************************************************
   SUBROUTINE topology_connectivity_pack(molecule_kind_set, molecule_set, &
                                         topology, subsys_section)
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_connectivity_pack', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      INTEGER :: counter, first, handle, handle2, i, ibend, ibond, idim, iimpr, ikind, imol, &
         inter_bends, inter_bonds, inter_imprs, inter_torsions, inter_ubs, intra_bends, &
         intra_bonds, intra_imprs, intra_torsions, intra_ubs, inum, ires, istart_mol, istart_typ, &
         itorsion, ityp, iub, iw, j, j1, j2, j3, j4, jind, last, min_index, natom, nelectron, &
         nsgf, nval_tot1, nval_tot2, nvar1, nvar2, output_unit, stat
      INTEGER, DIMENSION(:), POINTER :: c_var_a, c_var_b, c_var_c, c_var_d, c_var_type, &
         first_list, last_list, map_atom_mol, map_atom_type, map_cvar_mol, map_cvars, map_var_mol, &
         map_vars, molecule_list
      INTEGER, DIMENSION(:, :), POINTER                  :: bnd_ctype, bnd_type
      LOGICAL                                            :: found, found_last
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(bend_type), DIMENSION(:), POINTER             :: bend_list
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(impr_type), DIMENSION(:), POINTER             :: impr_list
      TYPE(local_molecule_type), POINTER                 :: lmi
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(opbend_type), DIMENSION(:), POINTER           :: opbend_list
      TYPE(torsion_type), DIMENSION(:), POINTER          :: torsion_list
      TYPE(ub_type), DIMENSION(:), POINTER               :: ub_list

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      atom_info => topology%atom_info
      conn_info => topology%conn_info
      ALLOCATE (map_atom_mol(topology%natoms))
      ALLOCATE (map_atom_type(topology%natoms))
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Set the topology%[nmol_type,nmol,nmol_conn]
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_1", handle2)
      natom = topology%natoms
      topology%nmol = 1
      topology%nmol_type = 1
      topology%nmol_conn = 0
      map_atom_mol = -1
      map_atom_type = -1
      map_atom_mol(1) = 1
      map_atom_type(1) = 1
      DO i = 1, natom-1
         IF ((atom_info%map_mol_typ(i+1) /= atom_info%map_mol_typ(i)) .OR. &
             ((atom_info%map_mol_res(i+1) /= atom_info%map_mol_res(i)) .AND. &
              (.NOT. (topology%conn_type == do_conn_user)))) THEN
            topology%nmol_type = topology%nmol_type+1
         END IF
         map_atom_type(i+1) = topology%nmol_type
         IF ((atom_info%map_mol_typ(i+1) /= atom_info%map_mol_typ(i)) .OR. &
             (atom_info%map_mol_num(i+1) /= atom_info%map_mol_num(i)) .OR. &
             (atom_info%map_mol_res(i+1) /= atom_info%map_mol_res(i))) THEN
            topology%nmol = topology%nmol+1
         END IF
         map_atom_mol(i+1) = topology%nmol
         IF ((atom_info%map_mol_typ(i+1) == atom_info%map_mol_typ(i)) .AND. &
             (atom_info%map_mol_num(i+1) == atom_info%map_mol_num(i)) .AND. &
             (atom_info%map_mol_res(i+1) /= atom_info%map_mol_res(i))) THEN
            topology%nmol_conn = topology%nmol_conn+1
         END IF
      END DO
      IF (iw > 0) WRITE (iw, *) "topology%nmol ::", topology%nmol
      IF (iw > 0) WRITE (iw, *) "topology%nmol_type ::", topology%nmol_type
      IF (iw > 0) WRITE (iw, *) "topology%nmol_conn ::", topology%nmol_conn
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1.1 Clean the temporary arrays to avoid quadratic loops around
      !     after this fix all topology_pack will be linear scaling
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_1.1", handle2)
      istart_mol = map_atom_mol(1)
      istart_typ = map_atom_type(1)
      DO i = 2, natom
         IF ((map_atom_mol(i) /= istart_mol) .AND. (map_atom_type(i) == istart_typ)) THEN
            map_atom_mol(i) = -map_atom_mol(i)
         ELSE IF (map_atom_type(i) /= istart_typ) THEN
            istart_mol = map_atom_mol(i)
            istart_typ = map_atom_type(i)
         END IF
      END DO
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Allocate the molecule_kind_set
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_2", handle2)
      IF (topology%nmol_type <= 0) THEN
         CALL cp__b("topology_connectivity_util.F",166,"No molecule kind defined")
      ELSE
         NULLIFY (molecule_kind_set)
         i = topology%nmol_type
         CALL allocate_molecule_kind_set(molecule_kind_set, i)
         IF (iw > 0) WRITE (iw, *) "    Allocated molecule_kind_set, Dimenstion of ", &
            SIZE(molecule_kind_set)
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Allocate the molecule_set
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_3", handle2)
      IF (topology%nmol <= 0) THEN
         CALL cp__b("topology_connectivity_util.F",181,"No molecule defined")
      ELSE
         NULLIFY (molecule_set)
         i = topology%nmol
         CALL allocate_molecule_set(molecule_set, i)
         IF (iw > 0) WRITE (iw, *) "    Allocated molecule_set, dimension of ", &
            topology%nmol
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Set the molecule_kind_set%[kind_number,name,nsgf,nelectron]
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_4", handle2)
      natom = topology%natoms
      ikind = -1
      DO i = 1, natom
         IF (ikind /= map_atom_type(i)) THEN
            ikind = map_atom_type(i)
            molecule_kind => molecule_kind_set(ikind)
            nsgf = 0
            nelectron = 0
            name = TRIM(id2str(atom_info%id_molname(i)))
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   kind_number=ikind, &
                                   molname_generated=topology%molname_generated, &
                                   name=TRIM(name), &
                                   nsgf=nsgf, &
                                   nelectron=nelectron)
         END IF
      END DO
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Set the molecule_list for molecule_kind in molecule_kind_set
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_5", handle2)
      natom = topology%natoms
      ikind = map_atom_type(1)
      imol = ABS(map_atom_mol(1))
      counter = 0
      DO i = 1, natom-1
         IF (ikind /= map_atom_type(i+1)) THEN
            found = .TRUE.
            found_last = .FALSE.
            imol = ABS(map_atom_mol(i))
         ELSEIF (ikind == topology%nmol_type) THEN
            found = .TRUE.
            found_last = .TRUE.
            imol = ABS(map_atom_mol(natom))
         ELSE
            found = .FALSE.
            found_last = .FALSE.
         END IF

         IF (found) THEN
            ALLOCATE (molecule_list(imol-counter))
            DO j = 1, SIZE(molecule_list)
               molecule_list(j) = j+counter
            END DO
            molecule_kind => molecule_kind_set(ikind)
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   molecule_list=molecule_list)
            IF (iw > 0) WRITE (iw, *) "      molecule_list", ikind, molecule_list(:)
            IF (found_last) EXIT
            counter = imol
            ikind = map_atom_type(i+1)
         END IF
      END DO
      ! Treat separately the case in which the last atom is also a molecule
      IF (i == natom) THEN
         imol = ABS(map_atom_mol(natom))
         ! Last atom is also a molecule by itself
         ALLOCATE (molecule_list(imol-counter))
         DO j = 1, SIZE(molecule_list)
            molecule_list(j) = j+counter
         END DO
         molecule_kind => molecule_kind_set(ikind)
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list)
         IF (iw > 0) WRITE (iw, *) "      molecule_list", ikind, molecule_list(:)
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Set the molecule_set(imol)%molecule_kind via set_molecule
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_6", handle2)
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list)
         DO i = 1, SIZE(molecule_list)
            molecule => molecule_set(molecule_list(i))
            CALL set_molecule(molecule, molecule_kind=molecule_kind)
         END DO
      END DO
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Set the molecule_set(imol)%[first_atom,last_atom] via set_molecule_set
      !-----------------------------------------------------------------------------
      ALLOCATE (first_list(SIZE(molecule_set)))
      ALLOCATE (last_list(SIZE(molecule_set)))
      CALL timeset(routineN//"_7", handle2)
      first_list(:) = 0
      last_list(:) = 0
      ityp = atom_info%map_mol_typ(1)
      inum = atom_info%map_mol_num(1)
      ires = atom_info%map_mol_res(1)
      imol = 1
      first_list(1) = 1
      DO j = 2, natom
         IF ((atom_info%map_mol_typ(j) /= ityp) .OR. &
             (atom_info%map_mol_num(j) /= inum) .OR. &
             (atom_info%map_mol_res(j) /= ires)) THEN
            ityp = atom_info%map_mol_typ(j)
            inum = atom_info%map_mol_num(j)
            ires = atom_info%map_mol_res(j)
            imol = imol+1
            first_list(imol) = j
         END IF
      END DO
      IF(.NOT.(imol == topology%nmol))CALL cp__a("topology_connectivity_util.F",304)
      DO ikind = 1, topology%nmol-1
         last_list(ikind) = first_list(ikind+1)-1
      END DO
      last_list(topology%nmol) = topology%natoms
      CALL set_molecule_set(molecule_set, first_list, last_list)
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 8. Set and NULLIFY the molecule_set(imol)%lmi via set_molecule_set
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_8", handle2)
      DO imol = 1, SIZE(molecule_set)
         molecule => molecule_set(imol)
         NULLIFY (lmi)
         CALL set_molecule(molecule, lmi=lmi)
      END DO
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 9. Set the atom_list for molecule_kind in molecule_kind_set (PART 1)
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_9", handle2)
      counter = 0
      DO imol = 1, SIZE(molecule_set)
         molecule => molecule_set(imol)
         molecule_kind => molecule_set(imol)%molecule_kind
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                kind_number=i)
         IF (counter /= i) THEN
            counter = i
            CALL get_molecule(molecule=molecule, &
                              first_atom=first, last_atom=last)
            natom = 0
            IF (first /= 0 .AND. last /= 0) natom = last-first+1
            ALLOCATE (atom_list(natom))
            DO i = 1, natom
               !Atomic kind information will be filled in (PART 2)
               NULLIFY (atom_list(i)%atomic_kind)
               atom_list(i)%id_name = atom_info%id_atmname(i+first-1)
               IF (iw > 0) WRITE (iw, '(5X,A,3I5,1X,A5)') "atom_list ", &
                  imol, counter, i, TRIM(id2str(atom_list(i)%id_name))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, atom_list=atom_list)
         END IF
      END DO
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 10. Set the molecule_kind%[nbond,bond_list] via set_molecule_kind
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_10", handle2)
      ! First map bonds on molecules
      nvar1 = 0
      nvar2 = 0
      NULLIFY (map_vars, map_cvars, bnd_type, bnd_ctype)
      IF (ASSOCIATED(conn_info%bond_a)) nvar1 = SIZE(conn_info%bond_a)
      IF (ASSOCIATED(conn_info%c_bond_a)) nvar2 = SIZE(conn_info%c_bond_a)
      nval_tot1 = nvar1
      nval_tot2 = 0
      ALLOCATE (map_var_mol(nvar1))
      ALLOCATE (map_cvar_mol(nvar2))
      map_var_mol = -1
      map_cvar_mol = -1
      DO i = 1, nvar1
         j1 = map_atom_mol(conn_info%bond_a(i))
         j2 = map_atom_mol(conn_info%bond_b(i))
         IF (j1 == j2) THEN
            IF (j1 > 0) map_var_mol(i) = map_atom_type(conn_info%bond_a(i))
         END IF
      END DO
      DO i = 1, nvar2
         min_index = MIN(conn_info%c_bond_a(i), conn_info%c_bond_b(i))
         j1 = map_atom_mol(min_index)
         IF (j1 > 0) map_cvar_mol(i) = map_atom_type(min_index)
      END DO
      CALL find_bnd_typ(topology%nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      CALL find_bnd_typ(topology%nmol_type, map_cvars, map_cvar_mol, bnd_ctype, nvar2)
      DO i = 1, topology%nmol_type
         intra_bonds = 0
         inter_bonds = 0
         IF (ALL(bnd_type(:, i) > 0)) THEN
            intra_bonds = bnd_type(2, i)-bnd_type(1, i)+1
         END IF
         IF (ALL(bnd_ctype(:, i) > 0)) THEN
            inter_bonds = bnd_ctype(2, i)-bnd_ctype(1, i)+1
         END IF
         ibond = intra_bonds+inter_bonds
         IF (iw > 0) THEN
            WRITE (iw, *) "    Total number bonds for molecule type ", i, " :", ibond
            WRITE (iw, *) "    intra (bonds inside  molecules) :: ", intra_bonds
            WRITE (iw, *) "    inter (bonds between molecules) :: ", inter_bonds
         END IF
         molecule_kind => molecule_kind_set(i)
         nval_tot2 = nval_tot2+ibond*SIZE(molecule_kind%molecule_list)

         ALLOCATE (bond_list(ibond))
         ibond = 0
         DO j = bnd_type(1, i), bnd_type(2, i)
            IF (j == 0) CYCLE
            ibond = ibond+1
            jind = map_vars(j)
            first = first_list(map_atom_mol(conn_info%bond_a(jind)))
            bond_list(ibond)%a = conn_info%bond_a(jind)-first+1
            bond_list(ibond)%b = conn_info%bond_b(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            bond_list(ibond)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               bond_list(ibond)%itype = conn_info%bond_type(jind)
            END IF
            !point this to the right bond_kind_type if using force field
            NULLIFY (bond_list(ibond)%bond_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,1X,A,I5,I5)') "molecule_kind", &
                  i, "  intra bond", &
                  conn_info%bond_a(jind), &
                  conn_info%bond_b(jind), &
                  "offset number at", &
                  conn_info%bond_a(jind)-first+1, &
                  conn_info%bond_b(jind)-first+1
            END IF
         END DO
         DO j = bnd_ctype(1, i), bnd_ctype(2, i)
            IF (j == 0) CYCLE
            ibond = ibond+1
            jind = map_cvars(j)
            min_index = MIN(conn_info%c_bond_a(jind), conn_info%c_bond_b(jind))
            first = first_list(map_atom_mol(min_index))
            bond_list(ibond)%a = conn_info%c_bond_a(jind)-first+1
            bond_list(ibond)%b = conn_info%c_bond_b(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            bond_list(ibond)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               bond_list(ibond)%itype = conn_info%c_bond_type(jind)
            END IF
            !point this to the right bond_kind_type if using force field
            NULLIFY (bond_list(ibond)%bond_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,1X,A,I5,I5)') "molecule_kind", &
                  i, " inter bond", &
                  conn_info%c_bond_a(jind), &
                  conn_info%c_bond_b(jind), &
                  "offset number at", &
                  conn_info%c_bond_a(jind)-first+1, &
                  conn_info%c_bond_b(jind)-first+1
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                nbond=SIZE(bond_list), bond_list=bond_list)
      END DO
      IF ((nval_tot1 /= nval_tot2) .AND. (output_unit > 0)) THEN
         WRITE (output_unit, '(/)')
         WRITE (output_unit, '(T5,A)') "ERROR| Mismatching found between the total number  of atoms"
         WRITE (output_unit, '(T5,A)') "ERROR| and the number of atoms computed multiplying the Nr."
         WRITE (output_unit, '(T5,A)') "ERROR| of molecules by the  number of atoms  building  that"
         WRITE (output_unit, '(T5,A)') "ERROR| kind of molecule."
         WRITE (output_unit, '(T5,A)') "ERROR| This happens when the connectivity is wrongly  built"
         WRITE (output_unit, '(T5,A)') "ERROR| One example could be two same kind of molecules have"
         WRITE (output_unit, '(T5,A)') "ERROR| a different number of atoms. Check the connectivity!"
      END IF
      IF(.NOT.(nval_tot1 == nval_tot2))CALL cp__a("topology_connectivity_util.F",464)
      DEALLOCATE (map_var_mol)
      DEALLOCATE (map_cvar_mol)
      DEALLOCATE (map_vars)
      DEALLOCATE (map_cvars)
      DEALLOCATE (bnd_type)
      DEALLOCATE (bnd_ctype)
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 11. Set the molecule_kind%[nbend,bend_list] via set_molecule_kind
      !-----------------------------------------------------------------------------
      ! Allocate c_var_a, c_var_b, c_var_c, c_var_type
      CALL timeset(routineN//"_11_pre", handle2)
      idim = 0
      ALLOCATE (c_var_a(idim))
      ALLOCATE (c_var_b(idim))
      ALLOCATE (c_var_c(idim))
      found = ASSOCIATED(conn_info%theta_type)
      IF (found) THEN
         ALLOCATE (c_var_type(idim))
      END IF
      IF (ASSOCIATED(conn_info%c_bond_a) .AND. ASSOCIATED(conn_info%theta_a)) THEN
         DO j = 1, SIZE(conn_info%theta_a)
            j1 = map_atom_mol(conn_info%theta_a(j))
            j2 = map_atom_mol(conn_info%theta_b(j))
            j3 = map_atom_mol(conn_info%theta_c(j))
            IF (j1 /= j2 .OR. j2 /= j3) THEN
               idim = idim+1
            END IF
         END DO
         CALL reallocate(c_var_a, 1, idim)
         CALL reallocate(c_var_b, 1, idim)
         CALL reallocate(c_var_c, 1, idim)
         IF (found) THEN
            CALL reallocate(c_var_type, 1, idim)
         END IF
         idim = 0
         DO j = 1, SIZE(conn_info%theta_a)
            j1 = map_atom_mol(conn_info%theta_a(j))
            j2 = map_atom_mol(conn_info%theta_b(j))
            j3 = map_atom_mol(conn_info%theta_c(j))
            IF (j1 /= j2 .OR. j2 /= j3) THEN
               idim = idim+1
               c_var_a(idim) = conn_info%theta_a(j)
               c_var_b(idim) = conn_info%theta_b(j)
               c_var_c(idim) = conn_info%theta_c(j)
               IF (found) THEN
                  c_var_type(idim) = conn_info%theta_type(j)
               END IF
            END IF
         END DO
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_11", handle2)
      ! map bends on molecules
      nvar1 = 0
      nvar2 = 0
      NULLIFY (map_vars, map_cvars, bnd_type, bnd_ctype)
      IF (ASSOCIATED(conn_info%theta_a)) nvar1 = SIZE(conn_info%theta_a)
      IF (ASSOCIATED(c_var_a)) nvar2 = SIZE(c_var_a)
      nval_tot1 = nvar1
      nval_tot2 = 0
      ALLOCATE (map_var_mol(nvar1))
      ALLOCATE (map_cvar_mol(nvar2))
      map_var_mol = -1
      map_cvar_mol = -1
      DO i = 1, nvar1
         j1 = map_atom_mol(conn_info%theta_a(i))
         j2 = map_atom_mol(conn_info%theta_b(i))
         j3 = map_atom_mol(conn_info%theta_c(i))
         IF (j1 == j2 .AND. j2 == j3) THEN
            IF (j1 > 0) map_var_mol(i) = map_atom_type(conn_info%theta_a(i))
         END IF
      END DO
      DO i = 1, nvar2
         min_index = MIN(c_var_a(i), c_var_b(i), c_var_c(i))
         j1 = map_atom_mol(min_index)
         IF (j1 > 0) map_cvar_mol(i) = map_atom_type(min_index)
      END DO
      CALL find_bnd_typ(topology%nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      CALL find_bnd_typ(topology%nmol_type, map_cvars, map_cvar_mol, bnd_ctype, nvar2)
      DO i = 1, topology%nmol_type
         intra_bends = 0
         inter_bends = 0
         IF (ALL(bnd_type(:, i) > 0)) THEN
            intra_bends = bnd_type(2, i)-bnd_type(1, i)+1
         END IF
         IF (ALL(bnd_ctype(:, i) > 0)) THEN
            inter_bends = bnd_ctype(2, i)-bnd_ctype(1, i)+1
         END IF
         ibend = intra_bends+inter_bends
         IF (iw > 0) THEN
            WRITE (iw, *) "    Total number of angles for molecule type ", i, " :", ibend
            WRITE (iw, *) "    intra (angles inside  molecules) :: ", intra_bends
            WRITE (iw, *) "    inter (angles between molecules) :: ", inter_bends
         END IF
         molecule_kind => molecule_kind_set(i)
         nval_tot2 = nval_tot2+ibend*SIZE(molecule_kind%molecule_list)
         ALLOCATE (bend_list(ibend))
         ibend = 0
         DO j = bnd_type(1, i), bnd_type(2, i)
            IF (j == 0) CYCLE
            ibend = ibend+1
            jind = map_vars(j)
            first = first_list(map_atom_mol(conn_info%theta_a(jind)))
            bend_list(ibend)%a = conn_info%theta_a(jind)-first+1
            bend_list(ibend)%b = conn_info%theta_b(jind)-first+1
            bend_list(ibend)%c = conn_info%theta_c(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            bend_list(ibend)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               bend_list(ibend)%itype = conn_info%theta_type(jind)
            END IF
            !point this to the right bend_kind_type if using force field
            NULLIFY (bend_list(ibend)%bend_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,I5,1X,A,I5,I5,I5)') &
                  "molecule_kind", ikind, "intra bend", &
                  conn_info%theta_a(jind), &
                  conn_info%theta_b(jind), &
                  conn_info%theta_c(jind), &
                  "offset number at", &
                  conn_info%theta_a(jind)-first+1, &
                  conn_info%theta_b(jind)-first+1, &
                  conn_info%theta_c(jind)-first+1
            END IF
         END DO
         DO j = bnd_ctype(1, i), bnd_ctype(2, i)
            IF (j == 0) CYCLE
            ibend = ibend+1
            jind = map_cvars(j)
            min_index = MIN(c_var_a(jind), c_var_b(jind), c_var_c(jind))
            first = first_list(map_atom_mol(min_index))
            bend_list(ibend)%a = c_var_a(jind)-first+1
            bend_list(ibend)%b = c_var_b(jind)-first+1
            bend_list(ibend)%c = c_var_c(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            bend_list(ibend)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               bend_list(ibend)%itype = c_var_type(jind)
            END IF
            !point this to the right bend_kind_type if using force field
            NULLIFY (bend_list(ibend)%bend_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,I5,1X,A,I5,I5,I5)') &
                  "molecule_kind", ikind, "inter bend", &
                  c_var_a(jind), &
                  c_var_b(jind), &
                  c_var_c(jind), &
                  "offset number at", &
                  c_var_a(jind)-first+1, &
                  c_var_b(jind)-first+1, &
                  c_var_c(jind)-first+1
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                nbend=SIZE(bend_list), bend_list=bend_list)
      END DO
      IF(.NOT.(nval_tot1 == nval_tot2))CALL cp__a("topology_connectivity_util.F",623)
      DEALLOCATE (map_var_mol)
      DEALLOCATE (map_cvar_mol)
      DEALLOCATE (map_vars)
      DEALLOCATE (map_cvars)
      DEALLOCATE (bnd_type)
      DEALLOCATE (bnd_ctype)
      DEALLOCATE (c_var_a)
      DEALLOCATE (c_var_b)
      DEALLOCATE (c_var_c)
      IF (found) THEN
         DEALLOCATE (c_var_type)
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 12. Set the molecule_kind%[nub,ub_list] via set_molecule_kind
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_12_pre", handle2)
      idim = 0
      ALLOCATE (c_var_a(idim))
      ALLOCATE (c_var_b(idim))
      ALLOCATE (c_var_c(idim))
      IF (ASSOCIATED(conn_info%c_bond_a) .AND. ASSOCIATED(conn_info%ub_a)) THEN
         DO j = 1, SIZE(conn_info%ub_a)
            j1 = map_atom_mol(conn_info%ub_a(j))
            j2 = map_atom_mol(conn_info%ub_b(j))
            j3 = map_atom_mol(conn_info%ub_c(j))
            IF (j1 /= j2 .OR. j2 /= j3) THEN
               idim = idim+1
            END IF
         END DO
         CALL reallocate(c_var_a, 1, idim)
         CALL reallocate(c_var_b, 1, idim)
         CALL reallocate(c_var_c, 1, idim)
         idim = 0
         DO j = 1, SIZE(conn_info%ub_a)
            j1 = map_atom_mol(conn_info%ub_a(j))
            j2 = map_atom_mol(conn_info%ub_b(j))
            j3 = map_atom_mol(conn_info%ub_c(j))
            IF (j1 /= j2 .OR. j2 /= j3) THEN
               idim = idim+1
               c_var_a(idim) = conn_info%ub_a(j)
               c_var_b(idim) = conn_info%ub_b(j)
               c_var_c(idim) = conn_info%ub_c(j)
            END IF
         END DO
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_12", handle2)
      ! map UBs on molecules
      nvar1 = 0
      nvar2 = 0
      NULLIFY (map_vars, map_cvars, bnd_type, bnd_ctype)
      IF (ASSOCIATED(conn_info%ub_a)) nvar1 = SIZE(conn_info%ub_a)
      IF (ASSOCIATED(c_var_a)) nvar2 = SIZE(c_var_a)
      nval_tot1 = nvar1
      nval_tot2 = 0
      ALLOCATE (map_var_mol(nvar1))
      ALLOCATE (map_cvar_mol(nvar2))
      map_var_mol = -1
      map_cvar_mol = -1
      DO i = 1, nvar1
         j1 = map_atom_mol(conn_info%ub_a(i))
         j2 = map_atom_mol(conn_info%ub_b(i))
         j3 = map_atom_mol(conn_info%ub_c(i))
         IF (j1 == j2 .AND. j2 == j3) THEN
            IF (j1 > 0) map_var_mol(i) = map_atom_type(conn_info%ub_a(i))
         END IF
      END DO
      DO i = 1, nvar2
         min_index = MIN(c_var_a(i), c_var_b(i), c_var_c(i))
         j1 = map_atom_mol(min_index)
         IF (j1 > 0) map_cvar_mol(i) = map_atom_type(min_index)
      END DO
      CALL find_bnd_typ(topology%nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      CALL find_bnd_typ(topology%nmol_type, map_cvars, map_cvar_mol, bnd_ctype, nvar2)
      DO i = 1, topology%nmol_type
         intra_ubs = 0
         inter_ubs = 0
         IF (ALL(bnd_type(:, i) > 0)) THEN
            intra_ubs = bnd_type(2, i)-bnd_type(1, i)+1
         END IF
         IF (ALL(bnd_ctype(:, i) > 0)) THEN
            inter_ubs = bnd_ctype(2, i)-bnd_ctype(1, i)+1
         END IF
         iub = intra_ubs+inter_ubs
         IF (iw > 0) THEN
            WRITE (iw, *) "    Total number of Urey-Bradley for molecule type ", i, " :", iub
            WRITE (iw, *) "    intra (UB inside  molecules) :: ", intra_ubs
            WRITE (iw, *) "    inter (UB between molecules) :: ", inter_ubs
         END IF
         molecule_kind => molecule_kind_set(i)
         nval_tot2 = nval_tot2+iub*SIZE(molecule_kind%molecule_list)
         ALLOCATE (ub_list(iub))
         iub = 0
         DO j = bnd_type(1, i), bnd_type(2, i)
            IF (j == 0) CYCLE
            iub = iub+1
            jind = map_vars(j)
            first = first_list(map_atom_mol(conn_info%ub_a(jind)))
            ub_list(iub)%a = conn_info%ub_a(jind)-first+1
            ub_list(iub)%b = conn_info%ub_b(jind)-first+1
            ub_list(iub)%c = conn_info%ub_c(jind)-first+1
            ub_list(iub)%id_type = do_ff_charmm
            !point this to the right ub_kind_type if using force field
            NULLIFY (ub_list(iub)%ub_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,I5,1X,A,I5,I5,I5)') &
                  "molecule_kind", i, "intra UB", &
                  conn_info%ub_a(jind), &
                  conn_info%ub_b(jind), &
                  conn_info%ub_c(jind), &
                  "offset number at", &
                  conn_info%ub_a(jind)-first+1, &
                  conn_info%ub_b(jind)-first+1, &
                  conn_info%ub_c(jind)-first+1
            END IF
         END DO
         DO j = bnd_ctype(1, i), bnd_ctype(2, i)
            IF (j == 0) CYCLE
            iub = iub+1
            jind = map_cvars(j)
            min_index = MIN(c_var_a(jind), c_var_b(jind), c_var_c(jind))
            first = first_list(map_atom_mol(min_index))
            ub_list(iub)%a = c_var_a(jind)-first+1
            ub_list(iub)%b = c_var_b(jind)-first+1
            ub_list(iub)%c = c_var_c(jind)-first+1
            ub_list(iub)%id_type = do_ff_charmm
            !point this to the right ub_kind_type if using force field
            NULLIFY (ub_list(iub)%ub_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I5,I5,I5,1X,A,I5,I5,I5)') &
                  "molecule_kind", i, "inter UB", &
                  c_var_a(jind), &
                  c_var_b(jind), &
                  c_var_c(jind), &
                  "offset number at", &
                  c_var_a(jind)-first+1, &
                  c_var_b(jind)-first+1, &
                  c_var_c(jind)-first+1
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                nub=SIZE(ub_list), ub_list=ub_list)
      END DO
      IF(.NOT.(nval_tot1 == nval_tot2))CALL cp__a("topology_connectivity_util.F",769)
      DEALLOCATE (map_var_mol)
      DEALLOCATE (map_cvar_mol)
      DEALLOCATE (map_vars)
      DEALLOCATE (map_cvars)
      DEALLOCATE (bnd_type)
      DEALLOCATE (bnd_ctype)
      DEALLOCATE (c_var_a)
      DEALLOCATE (c_var_b)
      DEALLOCATE (c_var_c)
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 13. Set the molecule_kind%[ntorsion,torsion_list] via set_molecule_kind
      !-----------------------------------------------------------------------------
      ! Allocate c_var_a, c_var_b, c_var_c, c_var_d, c_var_type
      CALL timeset(routineN//"_13_pre", handle2)
      idim = 0
      ALLOCATE (c_var_a(idim))
      ALLOCATE (c_var_b(idim))
      ALLOCATE (c_var_c(idim))
      ALLOCATE (c_var_d(idim))
      found = ASSOCIATED(conn_info%phi_type)
      IF (found) THEN
         ALLOCATE (c_var_type(idim))
      END IF
      IF (ASSOCIATED(conn_info%c_bond_a) .AND. ASSOCIATED(conn_info%phi_a)) THEN
         DO j = 1, SIZE(conn_info%phi_a)
            j1 = map_atom_mol(conn_info%phi_a(j))
            j2 = map_atom_mol(conn_info%phi_b(j))
            j3 = map_atom_mol(conn_info%phi_c(j))
            j4 = map_atom_mol(conn_info%phi_d(j))
            IF (j1 /= j2 .OR. j2 /= j3 .OR. j3 /= j4) THEN
               idim = idim+1
            END IF
         END DO
         CALL reallocate(c_var_a, 1, idim)
         CALL reallocate(c_var_b, 1, idim)
         CALL reallocate(c_var_c, 1, idim)
         CALL reallocate(c_var_d, 1, idim)
         IF (found) THEN
            CALL reallocate(c_var_type, 1, idim)
         END IF
         idim = 0
         DO j = 1, SIZE(conn_info%phi_a)
            j1 = map_atom_mol(conn_info%phi_a(j))
            j2 = map_atom_mol(conn_info%phi_b(j))
            j3 = map_atom_mol(conn_info%phi_c(j))
            j4 = map_atom_mol(conn_info%phi_d(j))
            IF (j1 /= j2 .OR. j2 /= j3 .OR. j3 /= j4) THEN
               idim = idim+1
               c_var_a(idim) = conn_info%phi_a(j)
               c_var_b(idim) = conn_info%phi_b(j)
               c_var_c(idim) = conn_info%phi_c(j)
               c_var_d(idim) = conn_info%phi_d(j)
               IF (found) THEN
                  c_var_type(idim) = conn_info%phi_type(j)
               END IF
            END IF
         END DO
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_13", handle2)
      ! map torsions on molecules
      nvar1 = 0
      nvar2 = 0
      NULLIFY (map_vars, map_cvars, bnd_type, bnd_ctype)
      IF (ASSOCIATED(conn_info%phi_a)) nvar1 = SIZE(conn_info%phi_a)
      IF (ASSOCIATED(c_var_a)) nvar2 = SIZE(c_var_a)
      nval_tot1 = nvar1
      nval_tot2 = 0
      ALLOCATE (map_var_mol(nvar1))
      ALLOCATE (map_cvar_mol(nvar2))
      map_var_mol = -1
      map_cvar_mol = -1
      DO i = 1, nvar1
         j1 = map_atom_mol(conn_info%phi_a(i))
         j2 = map_atom_mol(conn_info%phi_b(i))
         j3 = map_atom_mol(conn_info%phi_c(i))
         j4 = map_atom_mol(conn_info%phi_d(i))
         IF (j1 == j2 .AND. j2 == j3 .AND. j3 == j4) THEN
            IF (j1 > 0) map_var_mol(i) = map_atom_type(conn_info%phi_a(i))
         END IF
      END DO
      DO i = 1, nvar2
         min_index = MIN(c_var_a(i), c_var_b(i), c_var_c(i), c_var_d(i))
         j1 = map_atom_mol(min_index)
         IF (j1 > 0) map_cvar_mol(i) = map_atom_type(min_index)
      END DO
      CALL find_bnd_typ(topology%nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      CALL find_bnd_typ(topology%nmol_type, map_cvars, map_cvar_mol, bnd_ctype, nvar2)
      DO i = 1, topology%nmol_type
         intra_torsions = 0
         inter_torsions = 0
         IF (ALL(bnd_type(:, i) > 0)) THEN
            intra_torsions = bnd_type(2, i)-bnd_type(1, i)+1
         END IF
         IF (ALL(bnd_ctype(:, i) > 0)) THEN
            inter_torsions = bnd_ctype(2, i)-bnd_ctype(1, i)+1
         END IF
         itorsion = intra_torsions+inter_torsions
         IF (iw > 0) THEN
            WRITE (iw, *) "    Total number of torsions for molecule type ", i, " :", itorsion
            WRITE (iw, *) "    intra (torsions inside  molecules) :: ", intra_torsions
            WRITE (iw, *) "    inter (torsions between molecules) :: ", inter_torsions
         END IF
         molecule_kind => molecule_kind_set(i)
         nval_tot2 = nval_tot2+itorsion*SIZE(molecule_kind%molecule_list)
         ALLOCATE (torsion_list(itorsion))
         itorsion = 0
         DO j = bnd_type(1, i), bnd_type(2, i)
            IF (j == 0) CYCLE
            itorsion = itorsion+1
            jind = map_vars(j)
            first = first_list(map_atom_mol(conn_info%phi_a(jind)))
            torsion_list(itorsion)%a = conn_info%phi_a(jind)-first+1
            torsion_list(itorsion)%b = conn_info%phi_b(jind)-first+1
            torsion_list(itorsion)%c = conn_info%phi_c(jind)-first+1
            torsion_list(itorsion)%d = conn_info%phi_d(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            torsion_list(itorsion)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               torsion_list(itorsion)%itype = conn_info%phi_type(jind)
            END IF
            !point this to the right torsion_kind_type if using force field
            NULLIFY (torsion_list(itorsion)%torsion_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "intra TOR", &
                  conn_info%phi_a(jind), &
                  conn_info%phi_b(jind), &
                  conn_info%phi_c(jind), &
                  conn_info%phi_d(jind), &
                  "offset number at", &
                  conn_info%phi_a(jind)-first+1, &
                  conn_info%phi_b(jind)-first+1, &
                  conn_info%phi_c(jind)-first+1, &
                  conn_info%phi_d(jind)-first+1
            END IF
         END DO
         DO j = bnd_ctype(1, i), bnd_ctype(2, i)
            IF (j == 0) CYCLE
            itorsion = itorsion+1
            jind = map_cvars(j)
            min_index = MIN(c_var_a(jind), c_var_b(jind), c_var_c(jind), c_var_d(jind))
            first = first_list(map_atom_mol(min_index))
            torsion_list(itorsion)%a = c_var_a(jind)-first+1
            torsion_list(itorsion)%b = c_var_b(jind)-first+1
            torsion_list(itorsion)%c = c_var_c(jind)-first+1
            torsion_list(itorsion)%d = c_var_d(jind)-first+1
            ! Set by default id_type to charmm and modify when handling the forcefield
            torsion_list(itorsion)%id_type = do_ff_charmm
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               torsion_list(itorsion)%itype = c_var_type(jind)
            END IF
            !point this to the right torsion_kind_type if using force field
            NULLIFY (torsion_list(itorsion)%torsion_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "inter TOR", &
                  c_var_a(jind), &
                  c_var_b(jind), &
                  c_var_c(jind), &
                  c_var_d(jind), &
                  "offset number at", &
                  c_var_a(jind)-first+1, &
                  c_var_b(jind)-first+1, &
                  c_var_c(jind)-first+1, &
                  c_var_d(jind)-first+1
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                ntorsion=SIZE(torsion_list), torsion_list=torsion_list)
      END DO
      IF(.NOT.(nval_tot1 == nval_tot2))CALL cp__a("topology_connectivity_util.F",943)
      DEALLOCATE (map_var_mol)
      DEALLOCATE (map_cvar_mol)
      DEALLOCATE (map_vars)
      DEALLOCATE (map_cvars)
      DEALLOCATE (bnd_type)
      DEALLOCATE (bnd_ctype)
      DEALLOCATE (c_var_a)
      DEALLOCATE (c_var_b)
      DEALLOCATE (c_var_c)
      DEALLOCATE (c_var_d)
      IF (found) THEN
         DEALLOCATE (c_var_type)
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 14. Set the molecule_kind%[nimpr,impr_list] via set_molecule_kind
      !     Also set the molecule_kind%[nopbend,opbend_list]
      !-----------------------------------------------------------------------------
      ! Allocate c_var_a, c_var_b, c_var_c, c_var_d, c_var_type
      CALL timeset(routineN//"_14_pre", handle2)
      idim = 0
      ALLOCATE (c_var_a(idim))
      ALLOCATE (c_var_b(idim))
      ALLOCATE (c_var_c(idim))
      ALLOCATE (c_var_d(idim))
      found = ASSOCIATED(conn_info%impr_type)
      IF (found) THEN
         ALLOCATE (c_var_type(idim))
      END IF
      IF (ASSOCIATED(conn_info%c_bond_a) .AND. ASSOCIATED(conn_info%impr_a)) THEN
         DO j = 1, SIZE(conn_info%impr_a)
            j1 = map_atom_mol(conn_info%impr_a(j))
            j2 = map_atom_mol(conn_info%impr_b(j))
            j3 = map_atom_mol(conn_info%impr_c(j))
            j4 = map_atom_mol(conn_info%impr_d(j))
            IF (j1 /= j2 .OR. j2 /= j3 .OR. j3 /= j4) THEN
               idim = idim+1
            END IF
         END DO
         CALL reallocate(c_var_a, 1, idim)
         CALL reallocate(c_var_b, 1, idim)
         CALL reallocate(c_var_c, 1, idim)
         CALL reallocate(c_var_d, 1, idim)
         IF (found) THEN
            CALL reallocate(c_var_type, 1, idim)
         END IF
         idim = 0
         DO j = 1, SIZE(conn_info%impr_a)
            j1 = map_atom_mol(conn_info%impr_a(j))
            j2 = map_atom_mol(conn_info%impr_b(j))
            j3 = map_atom_mol(conn_info%impr_c(j))
            j4 = map_atom_mol(conn_info%impr_d(j))
            IF (j1 /= j2 .OR. j2 /= j3 .OR. j3 /= j4) THEN
               idim = idim+1
               c_var_a(idim) = conn_info%impr_a(j)
               c_var_b(idim) = conn_info%impr_b(j)
               c_var_c(idim) = conn_info%impr_c(j)
               c_var_d(idim) = conn_info%impr_d(j)
               IF (found) THEN
                  c_var_type(idim) = conn_info%impr_type(j)
               END IF
            END IF
         END DO
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_14", handle2)
      ! map imprs on molecules
      nvar1 = 0
      nvar2 = 0
      NULLIFY (map_vars, map_cvars, bnd_type, bnd_ctype)
      IF (ASSOCIATED(conn_info%impr_a)) nvar1 = SIZE(conn_info%impr_a)
      IF (ASSOCIATED(c_var_a)) nvar2 = SIZE(c_var_a)
      nval_tot1 = nvar1
      nval_tot2 = 0
      ALLOCATE (map_var_mol(nvar1))
      ALLOCATE (map_cvar_mol(nvar2))
      map_var_mol = -1
      map_cvar_mol = -1
      DO i = 1, nvar1
         j1 = map_atom_mol(conn_info%impr_a(i))
         j2 = map_atom_mol(conn_info%impr_b(i))
         j3 = map_atom_mol(conn_info%impr_c(i))
         j4 = map_atom_mol(conn_info%impr_d(i))
         IF (j1 == j2 .AND. j2 == j3 .AND. j3 == j4) THEN
            IF (j1 > 0) map_var_mol(i) = map_atom_type(conn_info%impr_a(i))
         END IF
      END DO
      DO i = 1, nvar2
         min_index = MIN(c_var_a(i), c_var_b(i), c_var_c(i), c_var_d(i))
         j1 = map_atom_mol(min_index)
         IF (j1 > 0) map_cvar_mol(i) = map_atom_type(min_index)
      END DO
      CALL find_bnd_typ(topology%nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      CALL find_bnd_typ(topology%nmol_type, map_cvars, map_cvar_mol, bnd_ctype, nvar2)
      DO i = 1, topology%nmol_type
         intra_imprs = 0
         inter_imprs = 0
         IF (ALL(bnd_type(:, i) > 0)) THEN
            intra_imprs = bnd_type(2, i)-bnd_type(1, i)+1
         END IF
         IF (ALL(bnd_ctype(:, i) > 0)) THEN
            inter_imprs = bnd_ctype(2, i)-bnd_ctype(1, i)+1
         END IF
         iimpr = intra_imprs+inter_imprs
         IF (iw > 0) THEN
            WRITE (iw, *) "    Total number of imprs for molecule type ", i, " :", iimpr
            WRITE (iw, *) "    intra (imprs inside  molecules) :: ", intra_imprs
            WRITE (iw, *) "    inter (imprs between molecules) :: ", inter_imprs
            WRITE (iw, *) "    Total number of opbends for molecule type ", i, " :", iimpr
            WRITE (iw, *) "    intra (opbends inside  molecules) :: ", intra_imprs
            WRITE (iw, *) "    inter (opbends between molecules) :: ", inter_imprs
         END IF
         molecule_kind => molecule_kind_set(i)
         nval_tot2 = nval_tot2+iimpr*SIZE(molecule_kind%molecule_list)
         ALLOCATE (impr_list(iimpr), STAT=stat)
         ALLOCATE (opbend_list(iimpr), STAT=stat)
         IF(.NOT.(stat == 0))CALL cp__a("topology_connectivity_util.F",1061)
         iimpr = 0
         DO j = bnd_type(1, i), bnd_type(2, i)
            IF (j == 0) CYCLE
            iimpr = iimpr+1
            jind = map_vars(j)
            first = first_list(map_atom_mol(conn_info%impr_a(jind)))
            impr_list(iimpr)%a = conn_info%impr_a(jind)-first+1
            impr_list(iimpr)%b = conn_info%impr_b(jind)-first+1
            impr_list(iimpr)%c = conn_info%impr_c(jind)-first+1
            impr_list(iimpr)%d = conn_info%impr_d(jind)-first+1
            ! Atom sequence for improper is A B C D in which A is central atom,
            ! B is deviating atom and C & D are secondairy atoms. Atom sequence for
            ! opbend is B D C A in which A is central atom, B is deviating. Hence
            ! to create an opbend out of an improper, B and D need to be interchanged.
            opbend_list(iimpr)%a = conn_info%impr_b(jind)-first+1
            opbend_list(iimpr)%b = conn_info%impr_d(jind)-first+1
            opbend_list(iimpr)%c = conn_info%impr_c(jind)-first+1
            opbend_list(iimpr)%d = conn_info%impr_a(jind)-first+1
            ! Set by default id_type of improper to charmm and modify when handling the forcefield
            impr_list(iimpr)%id_type = do_ff_charmm
            ! Set by default id_type of opbend to harmonic and modify when handling the forcefield
            opbend_list(iimpr)%id_type = do_ff_harmonic
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               impr_list(iimpr)%itype = conn_info%impr_type(jind)
            END IF
            !point this to the right impr_kind_type if using force field
            NULLIFY (impr_list(iimpr)%impr_kind)
            NULLIFY (opbend_list(iimpr)%opbend_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "intra IMPR", &
                  conn_info%impr_a(jind), &
                  conn_info%impr_b(jind), &
                  conn_info%impr_c(jind), &
                  conn_info%impr_d(jind), &
                  "offset number at", &
                  conn_info%impr_a(jind)-first+1, &
                  conn_info%impr_b(jind)-first+1, &
                  conn_info%impr_c(jind)-first+1, &
                  conn_info%impr_d(jind)-first+1
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "intra OPBEND", &
                  conn_info%impr_b(jind), &
                  conn_info%impr_d(jind), &
                  conn_info%impr_c(jind), &
                  conn_info%impr_a(jind), &
                  "offset number at", &
                  conn_info%impr_b(jind)-first+1, &
                  conn_info%impr_d(jind)-first+1, &
                  conn_info%impr_c(jind)-first+1, &
                  conn_info%impr_a(jind)-first+1
            END IF
         END DO
         DO j = bnd_ctype(1, i), bnd_ctype(2, i)
            IF (j == 0) CYCLE
            iimpr = iimpr+1
            jind = map_cvars(j)
            min_index = MIN(c_var_a(jind), c_var_b(jind), c_var_c(jind), c_var_d(jind))
            first = first_list(map_atom_mol(min_index))
            impr_list(iimpr)%a = c_var_a(jind)-first+1
            impr_list(iimpr)%b = c_var_b(jind)-first+1
            impr_list(iimpr)%c = c_var_c(jind)-first+1
            impr_list(iimpr)%d = c_var_d(jind)-first+1
            opbend_list(iimpr)%a = c_var_b(jind)-first+1
            opbend_list(iimpr)%b = c_var_d(jind)-first+1
            opbend_list(iimpr)%c = c_var_c(jind)-first+1
            opbend_list(iimpr)%d = c_var_a(jind)-first+1
            ! Set by default id_type of improper to charmm and modify when handling the forcefield
            impr_list(iimpr)%id_type = do_ff_charmm
            ! Set by default id_type of opbend to harmonic and modify when handling the forcefield
            opbend_list(iimpr)%id_type = do_ff_harmonic
            IF ((topology%conn_type == do_conn_g96) .OR. (topology%conn_type == do_conn_g87)) THEN
               impr_list(iimpr)%itype = c_var_type(jind)
            END IF
            !point this to the right impr_kind_type and opbend_kind_type if using force field
            NULLIFY (impr_list(iimpr)%impr_kind)
            NULLIFY (opbend_list(iimpr)%opbend_kind)
            IF (iw > 0) THEN
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "inter IMPR", &
                  c_var_a(jind), &
                  c_var_b(jind), &
                  c_var_c(jind), &
                  c_var_d(jind), &
                  "offset number at", &
                  c_var_a(jind)-first+1, &
                  c_var_b(jind)-first+1, &
                  c_var_c(jind)-first+1, &
                  c_var_d(jind)-first+1
               WRITE (iw, '(7X,A,I3,1X,A,I4,I4,I4,I4,1X,A,I4,I4,I4,I4)') &
                  "molecule_kind", i, "inter OPBEND", &
                  c_var_b(jind), &
                  c_var_d(jind), &
                  c_var_c(jind), &
                  c_var_a(jind), &
                  "offset number at", &
                  c_var_b(jind)-first+1, &
                  c_var_d(jind)-first+1, &
                  c_var_c(jind)-first+1, &
                  c_var_a(jind)-first+1
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                nimpr=SIZE(impr_list), impr_list=impr_list)
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                nopbend=SIZE(opbend_list), opbend_list=opbend_list)
      END DO
      IF(.NOT.(nval_tot1 == nval_tot2))CALL cp__a("topology_connectivity_util.F",1169)
      DEALLOCATE (map_var_mol)
      DEALLOCATE (map_cvar_mol)
      DEALLOCATE (map_vars)
      DEALLOCATE (map_cvars)
      DEALLOCATE (bnd_type)
      DEALLOCATE (bnd_ctype)
      DEALLOCATE (c_var_a)
      DEALLOCATE (c_var_b)
      DEALLOCATE (c_var_c)
      DEALLOCATE (c_var_d)
      IF (found) THEN
         DEALLOCATE (c_var_type)
      END IF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! Finally deallocate some stuff.
      !-----------------------------------------------------------------------------
      DEALLOCATE (first_list)
      DEALLOCATE (last_list)
      DEALLOCATE (map_atom_mol)
      DEALLOCATE (map_atom_type)
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
   END SUBROUTINE topology_connectivity_pack

! **************************************************************************************************
!> \brief used to achieve linear scaling in the connectivity_pack
!> \param nmol_type ...
!> \param map_vars ...
!> \param map_var_mol ...
!> \param bnd_type ...
!> \param nvar1 ...
!> \par History
!>      Teodoro Laino
! **************************************************************************************************
   SUBROUTINE find_bnd_typ(nmol_type, map_vars, map_var_mol, bnd_type, nvar1)
      INTEGER, INTENT(IN)                                :: nmol_type
      INTEGER, DIMENSION(:), POINTER                     :: map_vars, map_var_mol
      INTEGER, DIMENSION(:, :), POINTER                  :: bnd_type
      INTEGER, INTENT(IN)                                :: nvar1

      CHARACTER(len=*), PARAMETER :: routineN = 'find_bnd_typ', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ibond, j

      ALLOCATE (map_vars(nvar1))
      CALL sort(map_var_mol, nvar1, map_vars)
      ALLOCATE (bnd_type(2, nmol_type))
      bnd_type = 0
      IF (nvar1 == 0) RETURN
      DO j = 1, nvar1
         IF (map_var_mol(j) /= -1) EXIT
      END DO
      IF (j == nvar1+1) RETURN
      i = map_var_mol(j)
      bnd_type(1, i) = j
      DO ibond = j, nvar1
         IF (map_var_mol(ibond) /= i) THEN
            bnd_type(2, i) = ibond-1
            i = map_var_mol(ibond)
            bnd_type(1, i) = ibond
         END IF
      END DO
      bnd_type(2, i) = nvar1

   END SUBROUTINE find_bnd_typ

! **************************************************************************************************
!> \brief   Handles the multiple unit cell option for the connectivity
!> \param topology ...
!> \param subsys_section ...
!> \author  Teodoro Laino [tlaino] - 06.2009
! **************************************************************************************************
   SUBROUTINE topology_conn_multiple(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_conn_multiple', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, fac, i, ind, j, k, m, natoms_orig, &
                                                            nbond, nbond_c, nimpr, nonfo, nphi, &
                                                            ntheta, nub
      INTEGER, DIMENSION(:), POINTER                     :: multiple_unit_cell
      TYPE(connectivity_info_type), POINTER              :: conn_info

      NULLIFY (multiple_unit_cell)
      CALL section_vals_val_get(subsys_section, "TOPOLOGY%MULTIPLE_UNIT_CELL", &
                                i_vals=multiple_unit_cell)
      IF (ANY(multiple_unit_cell /= 1)) THEN
         fac = PRODUCT(multiple_unit_cell)
         conn_info => topology%conn_info

         nbond = 0
         IF (ASSOCIATED(conn_info%bond_a)) THEN
            nbond = SIZE(conn_info%bond_a)
            CALL reallocate(conn_info%bond_a, 1, nbond*fac)
            CALL reallocate(conn_info%bond_b, 1, nbond*fac)
         END IF

         ntheta = 0
         IF (ASSOCIATED(conn_info%theta_a)) THEN
            ntheta = SIZE(conn_info%theta_a)
            CALL reallocate(conn_info%theta_a, 1, ntheta*fac)
            CALL reallocate(conn_info%theta_b, 1, ntheta*fac)
            CALL reallocate(conn_info%theta_c, 1, ntheta*fac)
         END IF

         nphi = 0
         IF (ASSOCIATED(conn_info%phi_a)) THEN
            nphi = SIZE(conn_info%phi_a)
            CALL reallocate(conn_info%phi_a, 1, nphi*fac)
            CALL reallocate(conn_info%phi_b, 1, nphi*fac)
            CALL reallocate(conn_info%phi_c, 1, nphi*fac)
            CALL reallocate(conn_info%phi_d, 1, nphi*fac)
         END IF

         nimpr = 0
         IF (ASSOCIATED(conn_info%impr_a)) THEN
            nimpr = SIZE(conn_info%impr_a)
            CALL reallocate(conn_info%impr_a, 1, nimpr*fac)
            CALL reallocate(conn_info%impr_b, 1, nimpr*fac)
            CALL reallocate(conn_info%impr_c, 1, nimpr*fac)
            CALL reallocate(conn_info%impr_d, 1, nimpr*fac)
         END IF

         nbond_c = 0
         IF (ASSOCIATED(conn_info%c_bond_a)) THEN
            nbond_c = SIZE(conn_info%c_bond_a)
            CALL reallocate(conn_info%c_bond_a, 1, nbond_c*fac)
            CALL reallocate(conn_info%c_bond_b, 1, nbond_c*fac)
         END IF

         nub = 0
         IF (ASSOCIATED(conn_info%ub_a)) THEN
            nub = SIZE(conn_info%ub_a)
            CALL reallocate(conn_info%ub_a, 1, nub*fac)
            CALL reallocate(conn_info%ub_b, 1, nub*fac)
            CALL reallocate(conn_info%ub_c, 1, nub*fac)
         END IF

         nonfo = 0
         IF (ASSOCIATED(conn_info%onfo_a)) THEN
            nonfo = SIZE(conn_info%onfo_a)
            CALL reallocate(conn_info%onfo_a, 1, nonfo*fac)
            CALL reallocate(conn_info%onfo_b, 1, nonfo*fac)
         END IF

         natoms_orig = topology%natoms/fac
         ind = 0
         DO k = 1, multiple_unit_cell(3)
            DO j = 1, multiple_unit_cell(2)
               DO i = 1, multiple_unit_cell(1)
                  ind = ind+1
                  IF (ind == 1) CYCLE
                  a = (ind-1)*natoms_orig

                  ! Bonds
                  IF (nbond > 0) THEN
                     m = (ind-1)*nbond
                     conn_info%bond_a(m+1:m+nbond) = conn_info%bond_a(1:nbond)+a
                     conn_info%bond_b(m+1:m+nbond) = conn_info%bond_b(1:nbond)+a
                  END IF
                  ! Theta
                  IF (ntheta > 0) THEN
                     m = (ind-1)*ntheta
                     conn_info%theta_a(m+1:m+ntheta) = conn_info%theta_a(1:ntheta)+a
                     conn_info%theta_b(m+1:m+ntheta) = conn_info%theta_b(1:ntheta)+a
                     conn_info%theta_c(m+1:m+ntheta) = conn_info%theta_c(1:ntheta)+a
                  END IF
                  ! Phi
                  IF (nphi > 0) THEN
                     m = (ind-1)*nphi
                     conn_info%phi_a(m+1:m+nphi) = conn_info%phi_a(1:nphi)+a
                     conn_info%phi_b(m+1:m+nphi) = conn_info%phi_b(1:nphi)+a
                     conn_info%phi_c(m+1:m+nphi) = conn_info%phi_c(1:nphi)+a
                     conn_info%phi_d(m+1:m+nphi) = conn_info%phi_d(1:nphi)+a
                  END IF
                  ! Impropers
                  IF (nimpr > 0) THEN
                     m = (ind-1)*nimpr
                     conn_info%impr_a(m+1:m+nimpr) = conn_info%impr_a(1:nimpr)+a
                     conn_info%impr_b(m+1:m+nimpr) = conn_info%impr_b(1:nimpr)+a
                     conn_info%impr_c(m+1:m+nimpr) = conn_info%impr_c(1:nimpr)+a
                     conn_info%impr_d(m+1:m+nimpr) = conn_info%impr_d(1:nimpr)+a
                  END IF
                  ! Para_res
                  IF (nbond_c > 0) THEN
                     m = (ind-1)*nbond_c
                     conn_info%c_bond_a(m+1:m+nbond_c) = conn_info%c_bond_a(1:nbond_c)+a
                     conn_info%c_bond_b(m+1:m+nbond_c) = conn_info%c_bond_b(1:nbond_c)+a
                  END IF
                  ! Urey-Bradley
                  IF (nub > 0) THEN
                     m = (ind-1)*nub
                     conn_info%ub_a(m+1:m+nub) = conn_info%ub_a(1:nub)+a
                     conn_info%ub_b(m+1:m+nub) = conn_info%ub_b(1:nub)+a
                     conn_info%ub_c(m+1:m+nub) = conn_info%ub_c(1:nub)+a
                  END IF
                  ! ONFO
                  IF (nonfo > 0) THEN
                     m = (ind-1)*nonfo
                     conn_info%onfo_a(m+1:m+nonfo) = conn_info%onfo_a(1:nonfo)+a
                     conn_info%onfo_b(m+1:m+nonfo) = conn_info%onfo_b(1:nonfo)+a
                  END IF
               END DO
            END DO
         END DO
      END IF

   END SUBROUTINE topology_conn_multiple

END MODULE topology_connectivity_util
! **************************************************************************************************
MODULE fparser
   !------- -------- --------- --------- --------- --------- --------- --------- -------
   ! Fortran 90 function parser v1.1
   !------- -------- --------- --------- --------- --------- --------- --------- -------
   !
   ! This public domain function parser module is intended for applications
   ! where a set of mathematical expressions is specified at runtime and is
   ! then evaluated for a large number of variable values. This is done by
   ! compiling the set of function strings into byte code, which is interpreted
   ! very efficiently for the various variable values.
   !
   ! The source code is available from:
   ! http://www.its.uni-karlsruhe.de/~schmehl/opensource/fparser-v1.1.tar.gz
   !
   ! Please send comments, corrections or questions to the author:
   ! Roland Schmehl <Roland.Schmehl@mach.uni-karlsruhe.de>
   !
   !------- -------- --------- --------- --------- --------- --------- --------- -------

   USE kinds,                           ONLY: default_string_length,&
                                              rn => dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fparser'

   !------- -------- --------- --------- --------- --------- --------- --------- -------
   PUBLIC :: initf, & ! Initialize function parser for n functions
             parsef, & ! Parse single function string
             evalf, & ! Evaluate single function
             finalizef, & ! Finalize the function parser
             evalfd
   INTEGER, PUBLIC            :: EvalErrType ! =0: no error occured, >0: evaluation error
   !------- -------- --------- --------- --------- --------- --------- --------- -------
   PRIVATE
   SAVE
   INTEGER, PARAMETER, PRIVATE :: is = SELECTED_INT_KIND(1) !Data type of bytecode
   INTEGER(is), PARAMETER :: cImmed = 1, &
                             cNeg = 2, &
                             cAdd = 3, &
                             cSub = 4, &
                             cMul = 5, &
                             cDiv = 6, &
                             cPow = 7, &
                             cAbs = 8, &
                             cExp = 9, &
                             cLog10 = 10, &
                             cLog = 11, &
                             cSqrt = 12, &
                             cSinh = 13, &
                             cCosh = 14, &
                             cTanh = 15, &
                             cSin = 16, &
                             cCos = 17, &
                             cTan = 18, &
                             cAsin = 19, &
                             cAcos = 20, &
                             cAtan = 21, &
                             VarBegin = 22
   CHARACTER(LEN=1), DIMENSION(cAdd:cPow), PARAMETER :: Ops = (/'+', &
                                                                '-', &
                                                                '*', &
                                                                '/', &
                                                                '^'/)
   CHARACTER(LEN=5), DIMENSION(cAbs:cAtan), PARAMETER :: Funcs = (/'abs  ', &
                                                                   'exp  ', &
                                                                   'log10', &
                                                                   'log  ', &
                                                                   'sqrt ', &
                                                                   'sinh ', &
                                                                   'cosh ', &
                                                                   'tanh ', &
                                                                   'sin  ', &
                                                                   'cos  ', &
                                                                   'tan  ', &
                                                                   'asin ', &
                                                                   'acos ', &
                                                                   'atan '/)
! **************************************************************************************************
   TYPE tComp
      INTEGER(is), DIMENSION(:), POINTER :: ByteCode
      INTEGER                            :: ByteCodeSize
      REAL(rn), DIMENSION(:), POINTER :: Immed
      INTEGER                            :: ImmedSize
      REAL(rn), DIMENSION(:), POINTER :: Stack
      INTEGER                            :: StackSize, &
                                            StackPtr
   END TYPE tComp
   TYPE(tComp), DIMENSION(:), POINTER :: Comp ! Bytecode
   INTEGER, DIMENSION(:), ALLOCATABLE :: ipos ! Associates function strings
   !
CONTAINS
   !
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE finalizef()
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Finalize function parser
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(len=*), PARAMETER :: routineN = 'finalizef', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

!----- -------- --------- --------- --------- --------- --------- --------- -------

      DO i = 1, SIZE(Comp)
         IF (ASSOCIATED(Comp(i)%ByteCode)) THEN
            DEALLOCATE (Comp(i)%ByteCode)
         END IF
         IF (ASSOCIATED(Comp(i)%Immed)) THEN
            DEALLOCATE (Comp(i)%Immed)
         END IF
         IF (ASSOCIATED(Comp(i)%Stack)) THEN
            DEALLOCATE (Comp(i)%Stack)
         END IF
      END DO

      DEALLOCATE (Comp)

   END SUBROUTINE finalizef
   !
! **************************************************************************************************
!> \brief ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE initf(n)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Initialize function parser for n functions
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: n

      INTEGER                                            :: i

! Number of functions
!----- -------- --------- --------- --------- --------- --------- --------- -------

      ALLOCATE (Comp(n))
      DO i = 1, n
         NULLIFY (Comp(i)%ByteCode, Comp(i)%Immed, Comp(i)%Stack)
      END DO
   END SUBROUTINE initf
   !
! **************************************************************************************************
!> \brief Parse ith function string FuncStr and compile it into bytecode
!> \param i Function identifier
!> \param FuncStr Function string
!> \param Var Array with variable names
! **************************************************************************************************
   SUBROUTINE parsef(i, FuncStr, Var)
      INTEGER, INTENT(in)                                :: i
      CHARACTER(LEN=*), INTENT(in)                       :: FuncStr
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var

      CHARACTER(len=*), PARAMETER :: routineN = 'parsef', routineP = moduleN//':'//routineN

      CHARACTER(LEN=LEN(FuncStr))                        :: Func

! Function string, local use
!----- -------- --------- --------- --------- --------- --------- --------- -------

      IF (i < 1 .OR. i > SIZE(Comp)) THEN
         CALL cp__b("common/fparser.F",165,"Function number is out of range")
      END IF
      IF (SIZE(var) > HUGE(0_is)) THEN
         CALL cp__b("common/fparser.F",168,"Too many variables")
      END IF
      ALLOCATE (ipos(LEN_TRIM(FuncStr))) ! Char. positions in orig. string
      Func = FuncStr ! Local copy of function string
      CALL Replace('**', '^ ', Func) ! Exponent into 1-Char. format
      CALL RemoveSpaces(Func) ! Condense function string
      CALL CheckSyntax(Func, FuncStr, Var)
      DEALLOCATE (ipos)
      CALL Compile(i, Func, Var) ! Compile into bytecode

   END SUBROUTINE parsef
   !
! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param Val ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION evalf(i, Val) RESULT(res)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Evaluate bytecode of ith function for the values passed in array Val(:)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: i
      REAL(rn), DIMENSION(:), INTENT(in)                 :: Val
      REAL(rn)                                           :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'evalf', routineP = moduleN//':'//routineN
      REAL(rn), PARAMETER                                :: zero = 0._rn

      INTEGER                                            :: DP, IP, ipow, SP

! Function identifier
! Variable values
! Result
! Instruction pointer
! Data pointer
! Stack pointer
!----- -------- --------- --------- --------- --------- --------- --------- -------

      DP = 1
      SP = 0
      DO IP = 1, Comp(i)%ByteCodeSize
         SELECT CASE (Comp (i)%ByteCode (IP))

         CASE (cImmed); SP = SP+1; Comp(i)%Stack(SP) = Comp(i)%Immed(DP); DP = DP+1
         CASE (cNeg); Comp(i)%Stack(SP) = -Comp(i)%Stack(SP)
         CASE (cAdd); Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)+Comp(i)%Stack(SP); SP = SP-1
         CASE (cSub); Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)-Comp(i)%Stack(SP); SP = SP-1
         CASE (cMul); Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)*Comp(i)%Stack(SP); SP = SP-1
         CASE (cDiv); IF (Comp(i)%Stack(SP) == 0._rn) THEN; EvalErrType = 1; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)/Comp(i)%Stack(SP); SP = SP-1
         CASE (cPow)
            ! Fixing for possible Negative floating-point value raised to a real power
            IF (Comp(i)%Stack(SP-1) < 0.0_rn) THEN
               ipow = FLOOR(Comp(i)%Stack(SP))
               IF (MOD(Comp(i)%Stack(SP), REAL(ipow, KIND=rn)) == 0.0_rn) THEN
                  Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)**ipow
               ELSE
                  CALL cp__b("common/fparser.F",226,"Negative floating-point value raised to a real power!")
               END IF
            ELSE
               Comp(i)%Stack(SP-1) = Comp(i)%Stack(SP-1)**Comp(i)%Stack(SP)
            END IF
            SP = SP-1
         CASE (cAbs); Comp(i)%Stack(SP) = ABS(Comp(i)%Stack(SP))
         CASE (cExp); Comp(i)%Stack(SP) = EXP(Comp(i)%Stack(SP))
         CASE (cLog10); IF (Comp(i)%Stack(SP) <= 0._rn) THEN; EvalErrType = 3; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP) = LOG10(Comp(i)%Stack(SP))
         CASE (cLog); IF (Comp(i)%Stack(SP) <= 0._rn) THEN; EvalErrType = 3; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP) = LOG(Comp(i)%Stack(SP))
         CASE (cSqrt); IF (Comp(i)%Stack(SP) < 0._rn) THEN; EvalErrType = 3; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP) = SQRT(Comp(i)%Stack(SP))
         CASE (cSinh); Comp(i)%Stack(SP) = SINH(Comp(i)%Stack(SP))
         CASE (cCosh); Comp(i)%Stack(SP) = COSH(Comp(i)%Stack(SP))
         CASE (cTanh); Comp(i)%Stack(SP) = TANH(Comp(i)%Stack(SP))
         CASE (cSin); Comp(i)%Stack(SP) = SIN(Comp(i)%Stack(SP))
         CASE (cCos); Comp(i)%Stack(SP) = COS(Comp(i)%Stack(SP))
         CASE (cTan); Comp(i)%Stack(SP) = TAN(Comp(i)%Stack(SP))
         CASE (cAsin); IF ((Comp(i)%Stack(SP) < -1._rn) .OR. (Comp(i)%Stack(SP) > 1._rn)) THEN
               EvalErrType = 4; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP) = ASIN(Comp(i)%Stack(SP))
         CASE (cAcos); IF ((Comp(i)%Stack(SP) < -1._rn) .OR. (Comp(i)%Stack(SP) > 1._rn)) THEN
               EvalErrType = 4; res = zero; RETURN; ENDIF
            Comp(i)%Stack(SP) = ACOS(Comp(i)%Stack(SP))
         CASE (cAtan); Comp(i)%Stack(SP) = ATAN(Comp(i)%Stack(SP))
         CASE DEFAULT; SP = SP+1; Comp(i)%Stack(SP) = Val(Comp(i)%ByteCode(IP)-VarBegin+1)
         END SELECT
      END DO
      EvalErrType = 0
      res = Comp(i)%Stack(1)
   END FUNCTION evalf
   !
! **************************************************************************************************
!> \brief ...
!> \param Func ...
!> \param FuncStr ...
!> \param Var ...
! **************************************************************************************************
   SUBROUTINE CheckSyntax(Func, FuncStr, Var)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Check syntax of function string,  returns 0 if syntax is ok
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: Func, FuncStr
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var

      INTEGER                                            :: ib, in, j, lFunc, ParCnt
      CHARACTER(LEN=1)                                   :: c
      INTEGER(is)                                        :: n
      LOGICAL                                            :: err
      REAL(rn)                                           :: r

! Function string without spaces
! Original function string
! Array with variable names
! Parenthesis counter
!----- -------- --------- --------- --------- --------- --------- --------- -------

      j = 1
      ParCnt = 0
      lFunc = LEN_TRIM(Func)
      step: DO
         IF (j > lFunc) CALL ParseErrMsg(j, FuncStr)
         c = Func(j:j)
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         ! Check for valid operand (must appear)
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         IF (c == '-' .OR. c == '+') THEN ! Check for leading - or +
            j = j+1
            IF (j > lFunc) CALL ParseErrMsg(j, FuncStr, 'Missing operand')
            c = Func(j:j)
            IF (ANY(c == Ops)) CALL ParseErrMsg(j, FuncStr, 'Multiple operators')
         END IF
         n = MathFunctionIndex(Func(j:))
         IF (n > 0) THEN ! Check for math function
            j = j+LEN_TRIM(Funcs(n))
            IF (j > lFunc) CALL ParseErrMsg(j, FuncStr, 'Missing function argument')
            c = Func(j:j)
            IF (c /= '(') CALL ParseErrMsg(j, FuncStr, 'Missing opening parenthesis')
         END IF
         IF (c == '(') THEN ! Check for opening parenthesis
            ParCnt = ParCnt+1
            j = j+1
            CYCLE step
         END IF
         IF (SCAN(c, '0123456789.') > 0) THEN ! Check for number
            r = RealNum(Func(j:), ib, in, err)
            IF (err) CALL ParseErrMsg(j, FuncStr, 'Invalid number format:  '//Func(j+ib-1:j+in-2))
            j = j+in-1
            IF (j > lFunc) EXIT
            c = Func(j:j)
         ELSE ! Check for variable
            n = VariableIndex(Func(j:), Var, ib, in)
            IF (n == 0) CALL ParseErrMsg(j, FuncStr, 'Invalid element: '//Func(j+ib-1:j+in-2))
            j = j+in-1
            IF (j > lFunc) EXIT
            c = Func(j:j)
         END IF
         DO WHILE (c == ')') ! Check for closing parenthesis
            ParCnt = ParCnt-1
            IF (ParCnt < 0) CALL ParseErrMsg(j, FuncStr, 'Mismatched parenthesis')
            IF (Func(j-1:j-1) == '(') CALL ParseErrMsg(j-1, FuncStr, 'Empty parentheses')
            j = j+1
            IF (j > lFunc) EXIT
            c = Func(j:j)
         END DO
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         ! Now, we have a legal operand: A legal operator or end of string must follow
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         IF (j > lFunc) EXIT
         IF (ANY(c == Ops)) THEN ! Check for multiple operators
            IF (j+1 > lFunc) CALL ParseErrMsg(j, FuncStr)
            IF (ANY(Func(j+1:j+1) == Ops)) CALL ParseErrMsg(j+1, FuncStr, 'Multiple operators')
         ELSE ! Check for next operand
            CALL ParseErrMsg(j, FuncStr, 'Missing operator')
         END IF
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         ! Now, we have an operand and an operator: the next loop will check for another
         ! operand (must appear)
         !-- -------- --------- --------- --------- --------- --------- --------- -------
         j = j+1
      END DO step
      IF (ParCnt > 0) CALL ParseErrMsg(j, FuncStr, 'Missing )')
   END SUBROUTINE CheckSyntax
   !
! **************************************************************************************************
!> \brief ...
!> \retval msg ...
! **************************************************************************************************
   FUNCTION EvalErrMsg() RESULT(msg)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Return error message
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), DIMENSION(4), PARAMETER :: m = (/'Division by zero                ', &
         'Argument of SQRT negative       ', 'Argument of LOG negative        ', &
         'Argument of ASIN or ACOS illegal'/)
      CHARACTER(LEN=LEN(m))                              :: msg

!----- -------- --------- --------- --------- --------- --------- --------- -------

      IF (EvalErrType < 1 .OR. EvalErrType > SIZE(m)) THEN
         msg = ''
      ELSE
         msg = m(EvalErrType)
      ENDIF
   END FUNCTION EvalErrMsg
   !
! **************************************************************************************************
!> \brief ...
!> \param j ...
!> \param FuncStr ...
!> \param Msg ...
! **************************************************************************************************
   SUBROUTINE ParseErrMsg(j, FuncStr, Msg)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Print error message and terminate program
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: j
      CHARACTER(LEN=*), INTENT(in)                       :: FuncStr
      CHARACTER(LEN=*), INTENT(in), OPTIONAL             :: Msg

      CHARACTER(len=*), PARAMETER :: routineN = 'ParseErrMsg', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: message

! Original function string
!----- -------- --------- --------- --------- --------- --------- --------- -------

      IF (PRESENT(Msg)) THEN
         WRITE (UNIT=message, FMT="(A)") "Syntax error in function string: "//Msg
      ELSE
         WRITE (UNIT=message, FMT="(A)") "Syntax error in function string"
      END IF
      WRITE (*, '(/,T2,A)') TRIM(FuncStr)
      IF ((j > LBOUND(ipos, DIM=1)) .AND. (j <= UBOUND(ipos, DIM=1))) THEN
         WRITE (*, '(A)') REPEAT(" ", ipos(j))//"?"
      ELSE
         WRITE (*, '(A)') REPEAT(" ", SIZE(ipos)+1)//"?"
      END IF
      CALL cp__b("common/fparser.F",406,TRIM(message))

   END SUBROUTINE ParseErrMsg
   !
! **************************************************************************************************
!> \brief ...
!> \param c ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION OperatorIndex(c) RESULT(n)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Return operator index
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=1), INTENT(in)                       :: c
      INTEGER(is)                                        :: n

      INTEGER(is)                                        :: j

!----- -------- --------- --------- --------- --------- --------- --------- -------

      n = 0
      DO j = cAdd, cPow
         IF (c == Ops(j)) THEN
            n = j
            EXIT
         END IF
      END DO
   END FUNCTION OperatorIndex
   !
! **************************************************************************************************
!> \brief ...
!> \param str ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION MathFunctionIndex(str) RESULT(n)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Return index of math function beginnig at 1st position of string str
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: str
      INTEGER(is)                                        :: n

      CHARACTER(LEN=LEN(Funcs))                          :: fun
      INTEGER                                            :: k
      INTEGER(is)                                        :: j

!----- -------- --------- --------- --------- --------- --------- --------- -------

      n = 0
      DO j = cAbs, cAtan ! Check all math functions
         k = MIN(LEN_TRIM(Funcs(j)), LEN(str))
         CALL LowCase(str(1:k), fun)
         IF (fun == Funcs(j)) THEN ! Compare lower case letters
            n = j ! Found a matching function
            EXIT
         END IF
      END DO
   END FUNCTION MathFunctionIndex
   !
! **************************************************************************************************
!> \brief ...
!> \param str ...
!> \param Var ...
!> \param ibegin ...
!> \param inext ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION VariableIndex(str, Var, ibegin, inext) RESULT(n)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Return index of variable at begin of string str (returns 0 if no variable found)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: str
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var
      INTEGER, INTENT(out), OPTIONAL                     :: ibegin, inext
      INTEGER(is)                                        :: n

      INTEGER                                            :: ib, in, j, lstr

! String
! Array with variable names
! Index of variable
! Start position of variable name
! Position of character after name
!----- -------- --------- --------- --------- --------- --------- --------- -------

      n = 0
      lstr = LEN_TRIM(str)
      IF (lstr > 0) THEN
         DO ib = 1, lstr ! Search for first character in str
            IF (str(ib:ib) /= ' ') EXIT ! When lstr>0 at least 1 char in str
         END DO
         DO in = ib, lstr ! Search for name terminators
            IF (SCAN(str(in:in), '+-*/^) ') > 0) EXIT
         END DO
         DO j = 1, SIZE(Var)
            IF (str(ib:in-1) == Var(j)) THEN
               n = INT(j, KIND=is) !  Variable name found
               EXIT
            END IF
         END DO
      END IF
      IF (PRESENT(ibegin)) ibegin = ib
      IF (PRESENT(inext)) inext = in
   END FUNCTION VariableIndex
   !
! **************************************************************************************************
!> \brief ...
!> \param str ...
! **************************************************************************************************
   SUBROUTINE RemoveSpaces(str)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Remove Spaces from string, remember positions of characters in old string
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(inout)                    :: str

      INTEGER                                            :: k, lstr

!----- -------- --------- --------- --------- --------- --------- --------- -------

      lstr = LEN_TRIM(str)
      ipos(:) = (/(k, k=1, lstr)/)
      k = 1
      DO WHILE (str(k:lstr) /= ' ')
         IF (str(k:k) == ' ') THEN
            str(k:lstr) = str(k+1:lstr)//' ' ! Move 1 character to left
            ipos(k:lstr) = (/ipos(k+1:lstr), 0/) ! Move 1 element to left
            k = k-1
         END IF
         k = k+1
      END DO
   END SUBROUTINE RemoveSpaces
   !
! **************************************************************************************************
!> \brief ...
!> \param ca ...
!> \param cb ...
!> \param str ...
! **************************************************************************************************
   SUBROUTINE Replace(ca, cb, str)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Replace ALL appearances of character set ca in string str by character set cb
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: ca
      CHARACTER(LEN=LEN(ca)), INTENT(in)                 :: cb
      CHARACTER(LEN=*), INTENT(inout)                    :: str

      INTEGER                                            :: j, lca

! LEN(ca) must be LEN(cb)
!----- -------- --------- --------- --------- --------- --------- --------- -------

      lca = LEN(ca)
      DO j = 1, LEN_TRIM(str)-lca+1
         IF (str(j:j+lca-1) == ca) str(j:j+lca-1) = cb
      END DO
   END SUBROUTINE Replace
   !
! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param F ...
!> \param Var ...
! **************************************************************************************************
   SUBROUTINE Compile(i, F, Var)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Compile i-th function string F into bytecode
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: i
      CHARACTER(LEN=*), INTENT(in)                       :: F
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var

      CHARACTER(len=*), PARAMETER :: routineN = 'Compile', routineP = moduleN//':'//routineN

! Function identifier
! Function string
! Array with variable names
!----- -------- --------- --------- --------- --------- --------- --------- -------

      IF (ASSOCIATED(Comp(i)%ByteCode)) DEALLOCATE (Comp(i)%ByteCode, &
                                                    Comp(i)%Immed, &
                                                    Comp(i)%Stack)
      Comp(i)%ByteCodeSize = 0
      Comp(i)%ImmedSize = 0
      Comp(i)%StackSize = 0
      Comp(i)%StackPtr = 0
      CALL CompileSubstr(i, F, 1, LEN_TRIM(F), Var) ! Compile string to determine size
      ALLOCATE (Comp(i)%ByteCode(Comp(i)%ByteCodeSize), &
                Comp(i)%Immed(Comp(i)%ImmedSize), &
                Comp(i)%Stack(Comp(i)%StackSize))
      Comp(i)%ByteCodeSize = 0
      Comp(i)%ImmedSize = 0
      Comp(i)%StackSize = 0
      Comp(i)%StackPtr = 0
      CALL CompileSubstr(i, F, 1, LEN_TRIM(F), Var) ! Compile string into bytecode
      !
   END SUBROUTINE Compile
   !
! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param b ...
! **************************************************************************************************
   SUBROUTINE AddCompiledByte(i, b)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Add compiled byte to bytecode
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: i
      INTEGER(is), INTENT(in)                            :: b

! Function identifier
! Value of byte to be added
!----- -------- --------- --------- --------- --------- --------- --------- -------

      Comp(i)%ByteCodeSize = Comp(i)%ByteCodeSize+1
      IF (ASSOCIATED(Comp(i)%ByteCode)) Comp(i)%ByteCode(Comp(i)%ByteCodeSize) = b
   END SUBROUTINE AddCompiledByte
   !
! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param F ...
!> \param Var ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION MathItemIndex(i, F, Var) RESULT(n)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Return math item index, if item is real number, enter it into Comp-structure
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: i
      CHARACTER(LEN=*), INTENT(in)                       :: F
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var
      INTEGER(is)                                        :: n

! Function identifier
! Function substring
! Array with variable names
! Byte value of math item
!----- -------- --------- --------- --------- --------- --------- --------- -------

      n = 0
      IF (SCAN(F(1:1), '0123456789.') > 0) THEN ! Check for begin of a number
         Comp(i)%ImmedSize = Comp(i)%ImmedSize+1
         IF (ASSOCIATED(Comp(i)%Immed)) Comp(i)%Immed(Comp(i)%ImmedSize) = RealNum(F)
         n = cImmed
      ELSE ! Check for a variable
         n = VariableIndex(F, Var)
         IF (n > 0) n = VarBegin+n-1_is
      END IF
   END FUNCTION MathItemIndex
   !
! **************************************************************************************************
!> \brief ...
!> \param F ...
!> \param b ...
!> \param e ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION CompletelyEnclosed(F, b, e) RESULT(res)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Check if function substring F(b:e) is completely enclosed by a pair of parenthesis
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: F
      INTEGER, INTENT(in)                                :: b, e
      LOGICAL                                            :: res

      INTEGER                                            :: j, k

! Function substring
! First and last pos. of substring
!----- -------- --------- --------- --------- --------- --------- --------- -------

      res = .FALSE.
      IF (F(b:b) == '(' .AND. F(e:e) == ')') THEN
         k = 0
         DO j = b+1, e-1
            IF (F(j:j) == '(') THEN
               k = k+1
            ELSEIF (F(j:j) == ')') THEN
               k = k-1
            END IF
            IF (k < 0) EXIT
         END DO
         IF (k == 0) res = .TRUE. ! All opened parenthesis closed
      END IF
   END FUNCTION CompletelyEnclosed
   !
! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param F ...
!> \param b ...
!> \param e ...
!> \param Var ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE CompileSubstr(i, F, b, e, Var)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Compile i-th function string F into bytecode
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: i
      CHARACTER(LEN=*), INTENT(in)                       :: F
      INTEGER, INTENT(in)                                :: b, e
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: Var

      CHARACTER(LEN=*), PARAMETER :: &
         calpha = 'abcdefghijklmnopqrstuvwxyz'//'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

      INTEGER                                            :: b2, io, j, k
      INTEGER(is)                                        :: n

! Function identifier
! Function substring
! Begin and end position substring
! Array with variable names
!----- -------- --------- --------- --------- --------- --------- --------- -------
! Check for special cases of substring
!----- -------- --------- --------- --------- --------- --------- --------- -------

      IF (F(b:b) == '+') THEN ! Case 1: F(b:e) = '+...'
!      WRITE(*,*)'1. F(b:e) = "+..."'
         CALL CompileSubstr(i, F, b+1, e, Var)
         RETURN
      ELSEIF (CompletelyEnclosed(F, b, e)) THEN ! Case 2: F(b:e) = '(...)'
!      WRITE(*,*)'2. F(b:e) = "(...)"'
         CALL CompileSubstr(i, F, b+1, e-1, Var)
         RETURN
      ELSEIF (SCAN(F(b:b), calpha) > 0) THEN
         n = MathFunctionIndex(F(b:e))
         IF (n > 0) THEN
            b2 = b+INDEX(F(b:e), '(')-1
            IF (CompletelyEnclosed(F, b2, e)) THEN ! Case 3: F(b:e) = 'fcn(...)'
!            WRITE(*,*)'3. F(b:e) = "fcn(...)"'
               CALL CompileSubstr(i, F, b2+1, e-1, Var)
               CALL AddCompiledByte(i, n)
               RETURN
            END IF
         END IF
      ELSEIF (F(b:b) == '-') THEN
         IF (CompletelyEnclosed(F, b+1, e)) THEN ! Case 4: F(b:e) = '-(...)'
!         WRITE(*,*)'4. F(b:e) = "-(...)"'
            CALL CompileSubstr(i, F, b+2, e-1, Var)
            CALL AddCompiledByte(i, cNeg)
            RETURN
         ELSEIF (SCAN(F(b+1:b+1), calpha) > 0) THEN
            n = MathFunctionIndex(F(b+1:e))
            IF (n > 0) THEN
               b2 = b+INDEX(F(b+1:e), '(')
               IF (CompletelyEnclosed(F, b2, e)) THEN ! Case 5: F(b:e) = '-fcn(...)'
!               WRITE(*,*)'5. F(b:e) = "-fcn(...)"'
                  CALL CompileSubstr(i, F, b2+1, e-1, Var)
                  CALL AddCompiledByte(i, n)
                  CALL AddCompiledByte(i, cNeg)
                  RETURN
               END IF
            END IF
         ENDIF
      END IF
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Check for operator in substring: check only base level (k=0), exclude expr. in ()
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      DO io = cAdd, cPow ! Increasing priority +-*/^
         k = 0
         DO j = e, b, -1
            IF (F(j:j) == ')') THEN
               k = k+1
            ELSEIF (F(j:j) == '(') THEN
               k = k-1
            END IF
            IF (k == 0 .AND. F(j:j) == Ops(io) .AND. IsBinaryOp(j, F)) THEN
               IF (ANY(F(j:j) == Ops(cMul:cPow)) .AND. F(b:b) == '-') THEN ! Case 6: F(b:e) = '-...Op...' with Op > -
!               WRITE(*,*)'6. F(b:e) = "-...Op..." with Op > -'
                  CALL CompileSubstr(i, F, b+1, e, Var)
                  CALL AddCompiledByte(i, cNeg)
                  RETURN
               ELSE ! Case 7: F(b:e) = '...BinOp...'
!               WRITE(*,*)'7. Binary operator',F(j:j)
                  CALL CompileSubstr(i, F, b, j-1, Var)
                  CALL CompileSubstr(i, F, j+1, e, Var)
                  CALL AddCompiledByte(i, OperatorIndex(Ops(io)))
                  Comp(i)%StackPtr = Comp(i)%StackPtr-1
                  RETURN
               END IF
            END IF
         END DO
      END DO
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Check for remaining items, i.e. variables or explicit numbers
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      b2 = b
      IF (F(b:b) == '-') b2 = b2+1
      n = MathItemIndex(i, F(b2:e), Var)
!   WRITE(*,*)'8. AddCompiledByte ',n
      CALL AddCompiledByte(i, n)
      Comp(i)%StackPtr = Comp(i)%StackPtr+1
      IF (Comp(i)%StackPtr > Comp(i)%StackSize) Comp(i)%StackSize = Comp(i)%StackSize+1
      IF (b2 > b) CALL AddCompiledByte(i, cNeg)
   END SUBROUTINE CompileSubstr
   !
! **************************************************************************************************
!> \brief ...
!> \param j ...
!> \param F ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION IsBinaryOp(j, F) RESULT(res)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Check if operator F(j:j) in string F is binary operator
      ! Special cases already covered elsewhere:              (that is corrected in v1.1)
      ! - operator character F(j:j) is first character of string (j=1)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      INTEGER, INTENT(in)                                :: j
      CHARACTER(LEN=*), INTENT(in)                       :: F
      LOGICAL                                            :: res

      INTEGER                                            :: k
      LOGICAL                                            :: Dflag, Pflag

! Position of Operator
! String
! Result
!----- -------- --------- --------- --------- --------- --------- --------- -------

      res = .TRUE.
      IF (F(j:j) == '+' .OR. F(j:j) == '-') THEN ! Plus or minus sign:
         IF (j == 1) THEN ! - leading unary operator ?
            res = .FALSE.
         ELSEIF (SCAN(F(j-1:j-1), '+-*/^(') > 0) THEN ! - other unary operator ?
            res = .FALSE.
         ELSEIF (SCAN(F(j+1:j+1), '0123456789') > 0 .AND. & ! - in exponent of real number ?
                 SCAN(F(j-1:j-1), 'eEdD') > 0) THEN
            Dflag = .FALSE.; Pflag = .FALSE.
            k = j-1
            DO WHILE (k > 1) !   step to the left in mantissa
               k = k-1
               IF (SCAN(F(k:k), '0123456789') > 0) THEN
                  Dflag = .TRUE.
               ELSEIF (F(k:k) == '.') THEN
                  IF (Pflag) THEN
                     EXIT !   * EXIT: 2nd appearance of '.'
                  ELSE
                     Pflag = .TRUE. !   * mark 1st appearance of '.'
                  ENDIF
               ELSE
                  EXIT !   * all other characters
               END IF
            END DO
            IF (Dflag .AND. (k == 1 .OR. SCAN(F(k:k), '+-*/^(') > 0)) res = .FALSE.
         END IF
      END IF
   END FUNCTION IsBinaryOp
   !
! **************************************************************************************************
!> \brief ...
!> \param str ...
!> \param ibegin ...
!> \param inext ...
!> \param error ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION RealNum(str, ibegin, inext, error) RESULT(res)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Get real number from string - Format: [blanks][+|-][nnn][.nnn][e|E|d|D[+|-]nnn]
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: str
      INTEGER, INTENT(out), OPTIONAL                     :: ibegin, inext
      LOGICAL, INTENT(out), OPTIONAL                     :: error
      REAL(rn)                                           :: res

      INTEGER                                            :: ib, in, istat
      LOGICAL                                            :: Bflag, DInExp, DInMan, Eflag, err, &
                                                            InExp, InMan, Pflag

! String
! Real number
! Start position of real number
! 1st character after real number
! Error flag
! .T. at begin of number in str
! .T. in mantissa of number
! .T. after 1st '.' encountered
! .T. at exponent identifier 'eEdD'
! .T. in exponent of number
! .T. if at least 1 digit in mant.
! .T. if at least 1 digit in exp.
! Local error flag
!----- -------- --------- --------- --------- --------- --------- --------- -------

      Bflag = .TRUE.; InMan = .FALSE.; Pflag = .FALSE.; Eflag = .FALSE.; InExp = .FALSE.
      DInMan = .FALSE.; DInExp = .FALSE.
      ib = 1
      in = 1
      DO WHILE (in <= LEN_TRIM(str))
         SELECT CASE (str (in:in))
         CASE (' ') ! Only leading blanks permitted
            ib = ib+1
            IF (InMan .OR. Eflag .OR. InExp) EXIT
         CASE ('+', '-') ! Permitted only
            IF (Bflag) THEN
               InMan = .TRUE.; Bflag = .FALSE. ! - at beginning of mantissa
            ELSEIF (Eflag) THEN
               InExp = .TRUE.; Eflag = .FALSE. ! - at beginning of exponent
            ELSE
               EXIT ! - otherwise STOP
            ENDIF
         CASE ('0':'9') ! Mark
            IF (Bflag) THEN
               InMan = .TRUE.; Bflag = .FALSE. ! - beginning of mantissa
            ELSEIF (Eflag) THEN
               InExp = .TRUE.; Eflag = .FALSE. ! - beginning of exponent
            ENDIF
            IF (InMan) DInMan = .TRUE. ! Mantissa contains digit
            IF (InExp) DInExp = .TRUE. ! Exponent contains digit
         CASE ('.')
            IF (Bflag) THEN
               Pflag = .TRUE. ! - mark 1st appearance of '.'
               InMan = .TRUE.; Bflag = .FALSE. !   mark beginning of mantissa
            ELSEIF (InMan .AND. .NOT. Pflag) THEN
               Pflag = .TRUE. ! - mark 1st appearance of '.'
            ELSE
               EXIT ! - otherwise STOP
            END IF
         CASE ('e', 'E', 'd', 'D') ! Permitted only
            IF (InMan) THEN
               Eflag = .TRUE.; InMan = .FALSE. ! - following mantissa
            ELSE
               EXIT ! - otherwise STOP
            ENDIF
         CASE DEFAULT
            EXIT ! STOP at all other characters
         END SELECT
         in = in+1
      END DO
      err = (ib > in-1) .OR. (.NOT. DInMan) .OR. ((Eflag .OR. InExp) .AND. .NOT. DInExp)
      IF (err) THEN
         res = 0.0_rn
      ELSE
         READ (str(ib:in-1), *, IOSTAT=istat) res
         err = istat /= 0
      END IF
      IF (PRESENT(ibegin)) ibegin = ib
      IF (PRESENT(inext)) inext = in
      IF (PRESENT(error)) error = err
   END FUNCTION RealNum
   !
! **************************************************************************************************
!> \brief ...
!> \param str1 ...
!> \param str2 ...
! **************************************************************************************************
   SUBROUTINE LowCase(str1, str2)
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      ! Transform upper case letters in str1 into lower case letters, result is str2
      !----- -------- --------- --------- --------- --------- --------- --------- -------
      CHARACTER(LEN=*), INTENT(in)                       :: str1
      CHARACTER(LEN=*), INTENT(out)                      :: str2

      CHARACTER(LEN=*), PARAMETER :: lc = 'abcdefghijklmnopqrstuvwxyz', &
         uc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

      INTEGER                                            :: j, k

!----- -------- --------- --------- --------- --------- --------- --------- -------

      str2 = str1
      DO j = 1, LEN_TRIM(str1)
         k = INDEX(uc, str1(j:j))
         IF (k > 0) str2(j:j) = lc(k:k)
      END DO
   END SUBROUTINE LowCase

! **************************************************************************************************
!> \brief Evaluates derivatives
!> \param id_fun ...
!> \param ipar ...
!> \param vals ...
!> \param h ...
!> \param err ...
!> \retval derivative ...
!> \author Main algorithm from Numerical Recipes
!>      Ridders, C.J.F. 1982 - Advances in Engineering Software, Vol.4, no. 2, pp. 75-76
! **************************************************************************************************
   FUNCTION evalfd(id_fun, ipar, vals, h, err) RESULT(derivative)
      INTEGER, INTENT(IN)                                :: id_fun, ipar
      REAL(KIND=rn), DIMENSION(:), INTENT(INOUT)         :: vals
      REAL(KIND=rn), INTENT(IN)                          :: h
      REAL(KIND=rn), INTENT(OUT)                         :: err
      REAL(KIND=rn)                                      :: derivative

      CHARACTER(len=*), PARAMETER :: routineN = 'evalfd', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: ntab = 10
      REAL(KIND=rn), PARAMETER                           :: big_error = 1.0E30_rn, con = 1.4_rn, &
                                                            con2 = con*con, safe = 2.0_rn

      INTEGER                                            :: i, j
      REAL(KIND=rn)                                      :: a(ntab, ntab), errt, fac, funcm, funcp, &
                                                            hh, xval

      derivative = HUGE(0.0_rn)
      IF (h /= 0._rn) THEN
         xval = vals(ipar)
         hh = h
         vals(ipar) = xval+hh
         funcp = evalf(id_fun, vals)
         vals(ipar) = xval-hh
         funcm = evalf(id_fun, vals)
         a(1, 1) = (funcp-funcm)/(2.0_rn*hh)
         err = big_error
         DO i = 2, ntab
            hh = hh/con
            vals(ipar) = xval+hh
            funcp = evalf(id_fun, vals)
            vals(ipar) = xval-hh
            funcm = evalf(id_fun, vals)
            a(1, i) = (funcp-funcm)/(2.0_rn*hh)
            fac = con2
            DO j = 2, i
               a(j, i) = (a(j-1, i)*fac-a(j-1, i-1))/(fac-1.0_rn)
               fac = con2*fac
               errt = MAX(ABS(a(j, i)-a(j-1, i)), ABS(a(j, i)-a(j-1, i-1)))
               IF (errt .LE. err) THEN
                  err = errt
                  derivative = a(j, i)
               ENDIF
            END DO
            IF (ABS(a(i, i)-a(i-1, i-1)) .GE. safe*err) RETURN
         END DO
      ELSE
         CALL cp__b("common/fparser.F",1031,"DX provided equals zero!")
      END IF
      vals(ipar) = xval
   END FUNCTION evalfd

END MODULE fparser
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  set of type/routines to handle the storage of results in force_envs
!> \author fschiff (12.2007)
!> \par    History
!>         - 10.2008 Teodoro Laino [tlaino] - University of Zurich
!>                   major rewriting:
!>                   - information stored in a proper type (not in a character!)
!>                   - module more lean
! **************************************************************************************************
MODULE cp_result_methods
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: &
        cp_result_clean, cp_result_copy, cp_result_create, cp_result_release, cp_result_type, &
        cp_result_value_copy, cp_result_value_create, cp_result_value_init, &
        cp_result_value_p_reallocate, result_type_integer, result_type_logical, result_type_real
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_bcast
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_result_methods'

   PUBLIC :: put_results, &
             test_for_result, &
             get_results, &
             cp_results_erase, &
             cp_results_mp_bcast

   INTERFACE put_results
      MODULE PROCEDURE put_result_r1, put_result_r2
   END INTERFACE

   INTERFACE get_results
      MODULE PROCEDURE get_result_r1, get_result_r2, get_nreps
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Store a 1D array of reals in result_list
!> \param results ...
!> \param description ...
!> \param values ...
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE put_result_r1(results, description, values)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      REAL(KIND=dp), DIMENSION(:)                        :: values

      CHARACTER(len=*), PARAMETER :: routineN = 'put_result_r1', routineP = moduleN//':'//routineN

      INTEGER                                            :: isize, jsize
      LOGICAL                                            :: check

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",68)
      IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",69)
      check = SIZE(results%result_label) == SIZE(results%result_value)
      IF(.NOT.(check))CALL cp__a("common/cp_result_methods.F",71)
      isize = SIZE(results%result_label)
      jsize = SIZE(values)

      CALL reallocate(results%result_label, 1, isize+1)
      CALL cp_result_value_p_reallocate(results%result_value, 1, isize+1)

      results%result_label(isize+1) = description
      CALL cp_result_value_init(results%result_value(isize+1)%value, result_type_real, jsize)
      results%result_value(isize+1)%value%real_type = values

   END SUBROUTINE put_result_r1

! **************************************************************************************************
!> \brief Store a 2D array of reals in result_list
!> \param results ...
!> \param description ...
!> \param values ...
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE put_result_r2(results, description, values)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      REAL(KIND=dp), DIMENSION(:, :)                     :: values

      CHARACTER(len=*), PARAMETER :: routineN = 'put_result_r2', routineP = moduleN//':'//routineN

      INTEGER                                            :: isize, jsize
      LOGICAL                                            :: check

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",104)
      IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",105)
      check = SIZE(results%result_label) == SIZE(results%result_value)
      IF(.NOT.(check))CALL cp__a("common/cp_result_methods.F",107)
      isize = SIZE(results%result_label)
      jsize = SIZE(values, 1)*SIZE(values, 2)

      CALL reallocate(results%result_label, 1, isize+1)
      CALL cp_result_value_p_reallocate(results%result_value, 1, isize+1)

      results%result_label(isize+1) = description
      CALL cp_result_value_init(results%result_value(isize+1)%value, result_type_real, jsize)
      results%result_value(isize+1)%value%real_type = RESHAPE(values, (/jsize/))

   END SUBROUTINE put_result_r2

! **************************************************************************************************
!> \brief test for a certain result in the result_list
!> \param results ...
!> \param description ...
!> \retval res_exist ...
!> \par History
!>      10.2013
!> \author Mandes
! **************************************************************************************************
   FUNCTION test_for_result(results, description) RESULT(res_exist)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      LOGICAL                                            :: res_exist

      CHARACTER(len=*), PARAMETER :: routineN = 'test_for_result', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nlist

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",139)
      nlist = SIZE(results%result_value)
      res_exist = .FALSE.
      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
            res_exist = .TRUE.
            EXIT
         END IF
      END DO

   END FUNCTION test_for_result

! **************************************************************************************************
!> \brief gets the required part out of the result_list
!> \param results ...
!> \param description ...
!> \param values ...
!> \param nval      : if more than one entry for a given description is given you may choose
!>                    which entry you want
!> \param n_rep     : integer indicating how many times the section exists in result_list
!> \param n_entries : gets the number of lines used for a given description
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE get_result_r1(results, description, values, nval, n_rep, n_entries)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      REAL(KIND=dp), DIMENSION(:)                        :: values
      INTEGER, OPTIONAL                                  :: nval, n_rep, n_entries

      CHARACTER(len=*), PARAMETER :: routineN = 'get_result_r1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, nlist, nrep, size_res, size_values

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",175)
      nlist = SIZE(results%result_value)
      IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",177)
      IF(.NOT.(SIZE(results%result_label) == nlist))CALL cp__a("common/cp_result_methods.F",178)
      nrep = 0
      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) nrep = nrep+1
      END DO

      IF (PRESENT(n_rep)) THEN
         n_rep = nrep
      END IF

      IF (nrep .LE. 0) &
         CALL cp_abort(cp__l("common/cp_result_methods.F",189), &
                       " Trying to access result ("//TRIM(description)//") which was never stored!")

      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
            IF (results%result_value(i)%value%type_in_use /= result_type_real) &
               CALL cp__b("common/cp_result_methods.F",195,"Attempt to retrieve a RESULT which is not a REAL!")

            size_res = SIZE(results%result_value(i)%value%real_type)
            EXIT
         END IF
      END DO
      IF (PRESENT(n_entries)) n_entries = size_res
      size_values = SIZE(values, 1)
      IF (PRESENT(nval)) THEN
         IF(.NOT.(size_res == size_values))CALL cp__a("common/cp_result_methods.F",204)
      ELSE
         IF(.NOT.(nrep*size_res == size_values))CALL cp__a("common/cp_result_methods.F",206)
      END IF
      k = 0
      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
            k = k+1
            IF (PRESENT(nval)) THEN
               IF (k == nval) THEN
                  values = results%result_value(i)%value%real_type
                  EXIT
               END IF
            ELSE
               values((k-1)*size_res+1:k*size_res) = results%result_value(i)%value%real_type
            END IF
         END IF
      END DO

   END SUBROUTINE get_result_r1

! **************************************************************************************************
!> \brief gets the required part out of the result_list
!> \param results ...
!> \param description ...
!> \param values ...
!> \param nval      : if more than one entry for a given description is given you may choose
!>                    which entry you want
!> \param n_rep     : integer indicating how many times the section exists in result_list
!> \param n_entries : gets the number of lines used for a given description
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE get_result_r2(results, description, values, nval, n_rep, n_entries)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      REAL(KIND=dp), DIMENSION(:, :)                     :: values
      INTEGER, OPTIONAL                                  :: nval, n_rep, n_entries

      CHARACTER(len=*), PARAMETER :: routineN = 'get_result_r2', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, nlist, nrep, size_res, size_values

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",249)
      nlist = SIZE(results%result_value)
      IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",251)
      IF(.NOT.(SIZE(results%result_label) == nlist))CALL cp__a("common/cp_result_methods.F",252)
      nrep = 0
      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) nrep = nrep+1
      END DO

      IF (PRESENT(n_rep)) THEN
         n_rep = nrep
      END IF

      IF (nrep .LE. 0) &
         CALL cp_abort(cp__l("common/cp_result_methods.F",263), &
                       " Trying to access result ("//TRIM(description)//") which was never stored!")

      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
            IF (results%result_value(i)%value%type_in_use /= result_type_real) &
               CALL cp__b("common/cp_result_methods.F",269,"Attempt to retrieve a RESULT which is not a REAL!")

            size_res = SIZE(results%result_value(i)%value%real_type)
            EXIT
         END IF
      END DO
      IF (PRESENT(n_entries)) n_entries = size_res
      size_values = SIZE(values, 1)*SIZE(values, 2)
      IF (PRESENT(nval)) THEN
         IF(.NOT.(size_res == size_values))CALL cp__a("common/cp_result_methods.F",278)
      ELSE
         IF(.NOT.(nrep*size_res == size_values))CALL cp__a("common/cp_result_methods.F",280)
      END IF
      k = 0
      DO i = 1, nlist
         IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
            k = k+1
            IF (PRESENT(nval)) THEN
               IF (k == nval) THEN
                  values = RESHAPE(results%result_value(i)%value%real_type, (/SIZE(values, 1), SIZE(values, 2)/))
                  EXIT
               END IF
            ELSE
               values((k-1)*size_res+1:k*size_res, :) = RESHAPE(results%result_value(i)%value%real_type, &
                                                                (/SIZE(values, 1), SIZE(values, 2)/))
            END IF
         END IF
      END DO

   END SUBROUTINE get_result_r2

! **************************************************************************************************
!> \brief gets the required part out of the result_list
!> \param results ...
!> \param description ...
!> \param n_rep     : integer indicating how many times the section exists in result_list
!> \param n_entries : gets the number of lines used for a given description
!> \param type_in_use ...
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE get_nreps(results, description, n_rep, n_entries, type_in_use)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length)               :: description
      INTEGER, OPTIONAL                                  :: n_rep, n_entries, type_in_use

      CHARACTER(len=*), PARAMETER :: routineN = 'get_nreps', routineP = moduleN//':'//routineN

      INTEGER                                            :: I, nlist

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",321)
      nlist = SIZE(results%result_value)
      IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",323)
      IF(.NOT.(SIZE(results%result_label) == nlist))CALL cp__a("common/cp_result_methods.F",324)
      IF (PRESENT(n_rep)) THEN
         n_rep = 0
         DO i = 1, nlist
            IF (TRIM(results%result_label(i)) == TRIM(description)) n_rep = n_rep+1
         END DO
      END IF
      IF (PRESENT(n_entries)) THEN
         n_entries = 0
         DO i = 1, nlist
            IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
               SELECT CASE (results%result_value (i)%value%type_in_use)
               CASE (result_type_real)
                  n_entries = n_entries+SIZE(results%result_value(i)%value%real_type)
               CASE (result_type_integer)
                  n_entries = n_entries+SIZE(results%result_value(i)%value%integer_type)
               CASE (result_type_logical)
                  n_entries = n_entries+SIZE(results%result_value(i)%value%logical_type)
               CASE DEFAULT
                  ! Type not implemented in cp_result_type
                  CALL cp__b("common/cp_result_methods.F",344,"")
               END SELECT
               EXIT
            END IF
         END DO
      END IF
      IF (PRESENT(type_in_use)) THEN
         DO i = 1, nlist
            IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
               type_in_use = results%result_value(i)%value%type_in_use
               EXIT
            END IF
         END DO
      END IF
   END SUBROUTINE get_nreps

! **************************************************************************************************
!> \brief erase a part of  result_list
!> \param results ...
!> \param description ...
!> \param nval : if more than one entry for a given description is given you may choose
!>               which entry you want to delete
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE cp_results_erase(results, description, nval)
      TYPE(cp_result_type), POINTER                      :: results
      CHARACTER(LEN=default_string_length), OPTIONAL     :: description
      INTEGER, OPTIONAL                                  :: nval

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_results_erase', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: entry_deleted, i, k, new_size, nlist, &
                                                            nrep
      TYPE(cp_result_type), POINTER                      :: clean_results

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",383)
      new_size = 0
      IF (PRESENT(description)) THEN
         IF(.NOT.(description(1:1) == '['))CALL cp__a("common/cp_result_methods.F",386)
         nlist = SIZE(results%result_value)
         nrep = 0
         DO i = 1, nlist
            IF (TRIM(results%result_label(i)) == TRIM(description)) nrep = nrep+1
         END DO
         IF (nrep .NE. 0) THEN
            k = 0
            entry_deleted = 0
            DO i = 1, nlist
               IF (TRIM(results%result_label(i)) == TRIM(description)) THEN
                  k = k+1
                  IF (PRESENT(nval)) THEN
                     IF (nval == k) THEN
                        entry_deleted = entry_deleted+1
                        EXIT
                     END IF
                  ELSE
                     entry_deleted = entry_deleted+1
                  END IF
               END IF
            END DO
            IF(.NOT.(nlist-entry_deleted >= 0))CALL cp__a("common/cp_result_methods.F",408)
            new_size = nlist-entry_deleted
            NULLIFY (clean_results)
            CALL cp_result_create(clean_results)
            CALL cp_result_clean(clean_results)
            ALLOCATE (clean_results%result_label(new_size))
            ALLOCATE (clean_results%result_value(new_size))
            DO i = 1, new_size
               NULLIFY (clean_results%result_value(i)%value)
               CALL cp_result_value_create(clean_results%result_value(i)%value)
            END DO
            k = 0
            DO i = 1, nlist
               IF (TRIM(results%result_label(i)) /= TRIM(description)) THEN
                  k = k+1
                  clean_results%result_label(k) = results%result_label(i)
                  CALL cp_result_value_copy(clean_results%result_value(k)%value, &
                                            results%result_value(i)%value)
               END IF
            END DO
            CALL cp_result_copy(clean_results, results)
            CALL cp_result_release(clean_results)
         END IF
      ELSE
         CALL cp_result_clean(results)
         ALLOCATE (results%result_label(new_size))
         ALLOCATE (results%result_value(new_size))
      END IF
   END SUBROUTINE cp_results_erase

! **************************************************************************************************
!> \brief broadcast results type
!> \param results ...
!> \param source ...
!> \param para_env ...
!> \author  10.2008 Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE cp_results_mp_bcast(results, source, para_env)
      TYPE(cp_result_type), POINTER                      :: results
      INTEGER, INTENT(IN)                                :: source
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_results_mp_bcast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nlist
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: size_value, type_in_use

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_methods.F",456)
      nlist = 0
      IF (para_env%mepos == source) nlist = SIZE(results%result_value)
      CALL mp_bcast(nlist, source, para_env%group)

      ALLOCATE (size_value(nlist))
      ALLOCATE (type_in_use(nlist))
      IF (para_env%mepos == source) THEN
         DO i = 1, nlist
            CALL get_nreps(results, description=results%result_label(i), &
                           n_entries=size_value(i), type_in_use=type_in_use(i))
         END DO
      END IF
      CALL mp_bcast(size_value, source, para_env%group)
      CALL mp_bcast(type_in_use, source, para_env%group)

      IF (para_env%mepos /= source) THEN
         CALL cp_result_clean(results)
         ALLOCATE (results%result_value(nlist))
         ALLOCATE (results%result_label(nlist))
         DO i = 1, nlist
            results%result_label(i) = ""
            NULLIFY (results%result_value(i)%value)
            CALL cp_result_value_create(results%result_value(i)%value)
            CALL cp_result_value_init(results%result_value(i)%value, &
                                      type_in_use=type_in_use(i), size_value=size_value(i))
         END DO
      END IF
      DO i = 1, nlist
         CALL mp_bcast(results%result_label(i), source, para_env%group)
         SELECT CASE (results%result_value (i)%value%type_in_use)
         CASE (result_type_real)
            CALL mp_bcast(results%result_value(i)%value%real_type, source, para_env%group)
         CASE (result_type_integer)
            CALL mp_bcast(results%result_value(i)%value%integer_type, source, para_env%group)
         CASE (result_type_logical)
            CALL mp_bcast(results%result_value(i)%value%logical_type, source, para_env%group)
         CASE DEFAULT
            CALL cp__b("common/cp_result_methods.F",494,"Type not implemented in cp_result_type")
         END SELECT
      END DO
      DEALLOCATE (type_in_use)
      DEALLOCATE (size_value)
   END SUBROUTINE cp_results_mp_bcast

END MODULE cp_result_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Util mixed_environment
!> \author Teodoro Laino [tlaino] - 02.2011
! **************************************************************************************************
MODULE mixed_environment_utils

   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              get_results,&
                                              put_results,&
                                              test_for_result
   USE cp_result_types,                 ONLY: cp_result_p_type,&
                                              cp_result_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mixed_energy_types,              ONLY: mixed_force_type
   USE particle_list_types,             ONLY: particle_list_type
   USE virial_types,                    ONLY: virial_p_type,&
                                              virial_type,&
                                              zero_virial
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mixed_environment_utils'

   PUBLIC :: mixed_map_forces, &
             get_subsys_map_index

CONTAINS

! **************************************************************************************************
!> \brief Maps forces between the different force_eval sections/environments
!> \param particles_mix ...
!> \param virial_mix ...
!> \param results_mix ...
!> \param global_forces ...
!> \param virials ...
!> \param results ...
!> \param factor ...
!> \param iforce_eval ...
!> \param nforce_eval ...
!> \param map_index ...
!> \param mapping_section ...
!> \param overwrite ...
!> \author Teodoro Laino - University of Zurich [tlaino] - 05.2007
! **************************************************************************************************
   SUBROUTINE mixed_map_forces(particles_mix, virial_mix, results_mix, global_forces, &
                               virials, results, factor, iforce_eval, nforce_eval, map_index, &
                               mapping_section, overwrite)

      TYPE(particle_list_type), POINTER                  :: particles_mix
      TYPE(virial_type), POINTER                         :: virial_mix
      TYPE(cp_result_type), POINTER                      :: results_mix
      TYPE(mixed_force_type), DIMENSION(:), POINTER      :: global_forces
      TYPE(virial_p_type), DIMENSION(:), POINTER         :: virials
      TYPE(cp_result_p_type), DIMENSION(:), POINTER      :: results
      REAL(KIND=dp), INTENT(IN)                          :: factor
      INTEGER, INTENT(IN)                                :: iforce_eval, nforce_eval
      INTEGER, DIMENSION(:), POINTER                     :: map_index
      TYPE(section_vals_type), POINTER                   :: mapping_section
      LOGICAL, INTENT(IN)                                :: overwrite

      CHARACTER(len=*), PARAMETER :: routineN = 'mixed_map_forces', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      INTEGER                                            :: iparticle, jparticle, natom, nres
      LOGICAL                                            :: dip_exists
      REAL(KIND=dp), DIMENSION(3)                        :: dip_mix, dip_tmp

! Get Mapping index array

      natom = SIZE(global_forces(iforce_eval)%forces, 2)
      CALL get_subsys_map_index(mapping_section, natom, iforce_eval, nforce_eval, map_index)
      DO iparticle = 1, natom
         jparticle = map_index(iparticle)
         IF (overwrite) THEN
            particles_mix%els(jparticle)%f(:) = factor*global_forces(iforce_eval)%forces(:, iparticle)
         ELSE
            particles_mix%els(jparticle)%f(:) = particles_mix%els(jparticle)%f(:)+ &
                                                factor*global_forces(iforce_eval)%forces(:, iparticle)
         END IF
      END DO
      ! Mixing Virial
      IF (virial_mix%pv_availability) THEN
         IF (overwrite) CALL zero_virial(virial_mix, reset=.FALSE.)
         virial_mix%pv_total = virial_mix%pv_total+factor*virials(iforce_eval)%virial%pv_total
         virial_mix%pv_kinetic = virial_mix%pv_kinetic+factor*virials(iforce_eval)%virial%pv_kinetic
         virial_mix%pv_virial = virial_mix%pv_virial+factor*virials(iforce_eval)%virial%pv_virial
         virial_mix%pv_xc = virial_mix%pv_xc+factor*virials(iforce_eval)%virial%pv_xc
         virial_mix%pv_fock_4c = virial_mix%pv_fock_4c+factor*virials(iforce_eval)%virial%pv_fock_4c
         virial_mix%pv_constraint = virial_mix%pv_constraint+factor*virials(iforce_eval)%virial%pv_constraint
      END IF
      ! Deallocate map_index array
      IF (ASSOCIATED(map_index)) THEN
         DEALLOCATE (map_index)
      END IF

      ! Collect Requested Results info
      description = '[DIPOLE]'
      IF (overwrite) CALL cp_results_erase(results_mix)

      dip_exists = test_for_result(results=results(iforce_eval)%results, description=description)
      IF (dip_exists) THEN
         CALL get_results(results=results_mix, description=description, n_rep=nres)
         IF(.NOT.(nres <= 1))CALL cp__a("mixed_environment_utils.F",117)
         dip_mix = 0.0_dp
         IF (nres == 1) CALL get_results(results=results_mix, description=description, values=dip_mix)
         CALL get_results(results=results(iforce_eval)%results, description=description, n_rep=nres)
         CALL get_results(results=results(iforce_eval)%results, description=description, &
                          values=dip_tmp, nval=nres)
         dip_mix = dip_mix+factor*dip_tmp
         CALL cp_results_erase(results=results_mix, description=description)
         CALL put_results(results=results_mix, description=description, values=dip_mix)
      END IF

   END SUBROUTINE mixed_map_forces

! **************************************************************************************************
!> \brief performs mapping of the subsystems of different force_eval
!> \param mapping_section ...
!> \param natom ...
!> \param iforce_eval ...
!> \param nforce_eval ...
!> \param map_index ...
!> \author Teodoro Laino - University of Zurich [tlaino] - 05.2007
! **************************************************************************************************
   SUBROUTINE get_subsys_map_index(mapping_section, natom, iforce_eval, nforce_eval, map_index)

      TYPE(section_vals_type), POINTER                   :: mapping_section
      INTEGER, INTENT(IN)                                :: natom, iforce_eval, nforce_eval
      INTEGER, DIMENSION(:), POINTER                     :: map_index

      CHARACTER(len=*), PARAMETER :: routineN = 'get_subsys_map_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, ival, j, jval, k, n_rep, &
                                                            n_rep_loc, n_rep_map, n_rep_sys, tmp
      INTEGER, DIMENSION(:), POINTER                     :: index_glo, index_loc, list
      LOGICAL                                            :: check, explicit
      TYPE(section_vals_type), POINTER                   :: fragments_loc, fragments_sys, &
                                                            map_force_ev, map_full_sys

      IF(.NOT.(.NOT. ASSOCIATED(map_index)))CALL cp__a("mixed_environment_utils.F",155)
      ALLOCATE (map_index(natom))
      CALL section_vals_get(mapping_section, explicit=explicit)
      IF (.NOT. explicit) THEN
         ! Standard Mapping.. subsys are assumed to have the same structure
         DO i = 1, natom
            map_index(i) = i
         END DO
      ELSE
         ! Mapping systems with different structures
         map_full_sys => section_vals_get_subs_vals(mapping_section, "FORCE_EVAL_MIXED")
         map_force_ev => section_vals_get_subs_vals(mapping_section, "FORCE_EVAL")
         CALL section_vals_get(map_full_sys, explicit=explicit)
         IF(.NOT.(explicit))CALL cp__a("mixed_environment_utils.F",168)
         CALL section_vals_get(map_force_ev, explicit=explicit, n_repetition=n_rep)
         IF(.NOT.(explicit))CALL cp__a("mixed_environment_utils.F",170)
         IF(.NOT.(n_rep == nforce_eval))CALL cp__a("mixed_environment_utils.F",171)
         DO i = 1, n_rep
            CALL section_vals_val_get(map_force_ev, "_SECTION_PARAMETERS_", i_rep_section=i, i_val=ival)
            IF (ival == iforce_eval) EXIT
         END DO
         IF(.NOT.(i <= nforce_eval))CALL cp__a("mixed_environment_utils.F",176)
         fragments_sys => section_vals_get_subs_vals(map_full_sys, "FRAGMENT")
         fragments_loc => section_vals_get_subs_vals(map_force_ev, "FRAGMENT", i_rep_section=i)
         !Perform few check on the structure of the input mapping section. as provided by the user
         CALL section_vals_get(fragments_loc, n_repetition=n_rep_loc)
         CALL section_vals_get(fragments_sys, explicit=explicit, n_repetition=n_rep_sys)
         IF(.NOT.(explicit))CALL cp__a("mixed_environment_utils.F",182)
         IF(.NOT.(n_rep_sys >= n_rep_loc))CALL cp__a("mixed_environment_utils.F",183)
         IF (n_rep_loc == 0) THEN
            NULLIFY (list)
            ! We expect an easier syntax in this case..
            CALL section_vals_val_get(map_force_ev, "DEFINE_FRAGMENTS", i_rep_section=i, n_rep_val=n_rep_map)
            check = (n_rep_map /= 0)
            IF(.NOT.(check))CALL cp__a("mixed_environment_utils.F",189)
            CALL section_vals_val_get(map_force_ev, "DEFINE_FRAGMENTS", i_rep_section=i, i_vals=list)
            IF(.NOT.(SIZE(list) > 0))CALL cp__a("mixed_environment_utils.F",191)
            iatom = 0
            DO i = 1, SIZE(list)
               jval = list(i)
               DO j = 1, n_rep_sys
                  CALL section_vals_val_get(fragments_sys, "_SECTION_PARAMETERS_", i_rep_section=j, i_val=tmp)
                  IF (tmp == jval) EXIT
               END DO
               CALL section_vals_val_get(fragments_sys, "_DEFAULT_KEYWORD_", i_rep_section=j, i_vals=index_glo)
               DO k = 0, index_glo(2)-index_glo(1)
                  iatom = iatom+1
                  IF(.NOT.(iatom <= natom))CALL cp__a("mixed_environment_utils.F",202)
                  map_index(iatom) = index_glo(1)+k
               END DO
            END DO
            check = (iatom == natom)
            IF(.NOT.(check))CALL cp__a("mixed_environment_utils.F",207)
         ELSE
            ! General syntax..
            !Loop over the fragment of the force_eval
            DO i = 1, n_rep_loc
               CALL section_vals_val_get(fragments_loc, "_SECTION_PARAMETERS_", i_rep_section=i, i_val=ival)
               CALL section_vals_val_get(fragments_loc, "MAP", i_rep_section=i, i_val=jval)
               ! Index corresponding to the mixed_force_eval fragment
               DO j = 1, n_rep_sys
                  CALL section_vals_val_get(fragments_sys, "_SECTION_PARAMETERS_", i_rep_section=j, i_val=tmp)
                  IF (tmp == jval) EXIT
               END DO
               IF(.NOT.(j <= n_rep_sys))CALL cp__a("mixed_environment_utils.F",219)
               CALL section_vals_val_get(fragments_loc, "_DEFAULT_KEYWORD_", i_rep_section=i, i_vals=index_loc)
               CALL section_vals_val_get(fragments_sys, "_DEFAULT_KEYWORD_", i_rep_section=j, i_vals=index_glo)
               check = ((index_loc(2)-index_loc(1)) == (index_glo(2)-index_glo(1)))
               IF(.NOT.(check))CALL cp__a("mixed_environment_utils.F",223)
               ! Now let's build the real mapping
               DO k = 0, index_loc(2)-index_loc(1)
                  map_index(index_loc(1)+k) = index_glo(1)+k
               END DO
            END DO
         END IF
      END IF

   END SUBROUTINE get_subsys_map_index

END MODULE mixed_environment_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines functions to perform rmsd in 3D
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
MODULE rmsd

   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all,&
                                              matvec_3x3
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rmsd'
   REAL(KIND=dp), PARAMETER, PRIVATE    :: Epsi = EPSILON(0.0_dp)*1.0E4_dp

   PUBLIC :: rmsd3

CONTAINS

! **************************************************************************************************
!> \brief Computes the RMSD in 3D. Provides also derivatives.
!> \param particle_set ...
!> \param r ...
!> \param r0 ...
!> \param output_unit ...
!> \param weights ...
!> \param my_val ...
!> \param rotate ...
!> \param transl ...
!> \param rot ...
!> \param drmsd3 ...
!> \author Teodoro Laino 08.2006
!> \note
!>      Optional arguments:
!>          my_val -> gives back the value of the RMSD
!>          transl -> provides the translational vector
!>          rotate -> if .true. gives back in r the coordinates rotated
!>                    in order to minimize the RMSD3
!>          rot    -> provides the rotational matrix
!>          drmsd3 -> derivatives of RMSD3 w.r.t. atomic positions
! **************************************************************************************************
   SUBROUTINE rmsd3(particle_set, r, r0, output_unit, weights, my_val, &
                    rotate, transl, rot, drmsd3)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: r, r0
      INTEGER, INTENT(IN)                                :: output_unit
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: weights
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: my_val
      LOGICAL, INTENT(IN), OPTIONAL                      :: rotate
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT), OPTIONAL :: transl
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: rot, drmsd3

      CHARACTER(len=*), PARAMETER :: routineN = 'rmsd3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ix, j, k, natom
      LOGICAL                                            :: my_rotate
      REAL(KIND=dp)                                      :: dm_r(4, 4, 3), lambda(4), loc_tr(3), &
                                                            M(4, 4), mtot, my_rot(3, 3), Q(0:3), &
                                                            rr(3), rr0(3), rrsq, s, xx, yy, &
                                                            Z(4, 4), zz
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r0p, rp

      IF(.NOT.(SIZE(r) == SIZE(r0)))CALL cp__a("rmsd.F",73)
      natom = SIZE(particle_set)
      my_rotate = .FALSE.
      IF (PRESENT(rotate)) my_rotate = rotate
      ALLOCATE (w(natom))
      IF (PRESENT(weights)) THEN
         w(:) = weights
      ELSE
         ! All atoms have a weight proportional to their mass in the RMSD unless
         ! differently requested
         DO i = 1, natom
            w(i) = particle_set(i)%atomic_kind%mass
         END DO
         mtot = MINVAL(w)
         IF (mtot /= 0.0_dp) w(:) = w(:)/mtot
      END IF
      ALLOCATE (rp(3, natom))
      ALLOCATE (r0p(3, natom))
      ! Molecule given by coordinates R
      ! Find COM and center molecule in COM
      xx = 0.0_dp
      yy = 0.0_dp
      zz = 0.0_dp
      mtot = 0.0_dp
      DO i = 1, natom
         mtot = mtot+particle_set(i)%atomic_kind%mass
         xx = xx+r((i-1)*3+1)*particle_set(i)%atomic_kind%mass
         yy = yy+r((i-1)*3+2)*particle_set(i)%atomic_kind%mass
         zz = zz+r((i-1)*3+3)*particle_set(i)%atomic_kind%mass
      END DO
      xx = xx/mtot
      yy = yy/mtot
      zz = zz/mtot
      DO i = 1, natom
         rp(1, i) = r((i-1)*3+1)-xx
         rp(2, i) = r((i-1)*3+2)-yy
         rp(3, i) = r((i-1)*3+3)-zz
      END DO
      IF (PRESENT(transl)) THEN
         transl(1) = xx
         transl(2) = yy
         transl(3) = zz
      END IF
      ! Molecule given by coordinates R0
      ! Find COM and center molecule in COM
      xx = 0.0_dp
      yy = 0.0_dp
      zz = 0.0_dp
      DO i = 1, natom
         xx = xx+r0((i-1)*3+1)*particle_set(i)%atomic_kind%mass
         yy = yy+r0((i-1)*3+2)*particle_set(i)%atomic_kind%mass
         zz = zz+r0((i-1)*3+3)*particle_set(i)%atomic_kind%mass
      END DO
      xx = xx/mtot
      yy = yy/mtot
      zz = zz/mtot
      DO i = 1, natom
         r0p(1, i) = r0((i-1)*3+1)-xx
         r0p(2, i) = r0((i-1)*3+2)-yy
         r0p(3, i) = r0((i-1)*3+3)-zz
      END DO
      loc_tr(1) = xx
      loc_tr(2) = yy
      loc_tr(3) = zz
      ! Give back the translational vector
      IF (PRESENT(transl)) THEN
         transl(1) = transl(1)-xx
         transl(2) = transl(2)-yy
         transl(3) = transl(3)-zz
      END IF
      M = 0.0_dp
      !
      DO i = 1, natom
         IF (w(i) == 0.0_dp) CYCLE
         rr(1) = rp(1, I)
         rr(2) = rp(2, I)
         rr(3) = rp(3, I)
         rr0(1) = r0p(1, I)
         rr0(2) = r0p(2, I)
         rr0(3) = r0p(3, I)
         rrsq = w(I)*(rr0(1)**2+rr0(2)**2+rr0(3)**2+rr(1)**2+rr(2)**2+rr(3)**2)
         rr0(1) = w(I)*rr0(1)
         rr0(2) = w(I)*rr0(2)
         rr0(3) = w(I)*rr0(3)
         M(1, 1) = M(1, 1)+rrsq+2.0_dp*(-rr0(1)*rr(1)-rr0(2)*rr(2)-rr0(3)*rr(3))
         M(2, 2) = M(2, 2)+rrsq+2.0_dp*(-rr0(1)*rr(1)+rr0(2)*rr(2)+rr0(3)*rr(3))
         M(3, 3) = M(3, 3)+rrsq+2.0_dp*(rr0(1)*rr(1)-rr0(2)*rr(2)+rr0(3)*rr(3))
         M(4, 4) = M(4, 4)+rrsq+2.0_dp*(rr0(1)*rr(1)+rr0(2)*rr(2)-rr0(3)*rr(3))
         M(1, 2) = M(1, 2)+2.0_dp*(-rr0(2)*rr(3)+rr0(3)*rr(2))
         M(1, 3) = M(1, 3)+2.0_dp*(rr0(1)*rr(3)-rr0(3)*rr(1))
         M(1, 4) = M(1, 4)+2.0_dp*(-rr0(1)*rr(2)+rr0(2)*rr(1))
         M(2, 3) = M(2, 3)-2.0_dp*(rr0(1)*rr(2)+rr0(2)*rr(1))
         M(2, 4) = M(2, 4)-2.0_dp*(rr0(1)*rr(3)+rr0(3)*rr(1))
         M(3, 4) = M(3, 4)-2.0_dp*(rr0(2)*rr(3)+rr0(3)*rr(2))
      END DO
      ! Symmetrize
      M(2, 1) = M(1, 2)
      M(3, 1) = M(1, 3)
      M(3, 2) = M(2, 3)
      M(4, 1) = M(1, 4)
      M(4, 2) = M(2, 4)
      M(4, 3) = M(3, 4)
      ! Solve the eigenvalue problem for M
      Z = M
      CALL diamat_all(Z, lambda)
      ! Pick the correct eigenvectors
      S = 1.0_dp
      IF (Z(1, 1) .LT. 0.0_dp) S = -1.0_dp
      Q(0) = S*Z(1, 1)
      Q(1) = S*Z(2, 1)
      Q(2) = S*Z(3, 1)
      Q(3) = S*Z(4, 1)
      IF (PRESENT(my_val)) THEN
         IF (ABS(lambda(1)) < epsi) THEN
            my_val = 0.0_dp
         ELSE
            my_val = lambda(1)/REAL(natom, KIND=dp)
         ENDIF
      END IF
      IF (ABS(lambda(1)-lambda(2)) < epsi) THEN
         IF (output_unit > 0) WRITE (output_unit, FMT='(T2,"RMSD3|",A)') &
            'NORMAL EXECUTION, NON-UNIQUE SOLUTION'
      END IF
      ! Computes derivatives w.r.t. the positions if requested
      IF (PRESENT(drmsd3)) THEN
         DO I = 1, natom
            IF (W(I) == 0.0_dp) CYCLE
            rr(1) = W(I)*2.0_dp*rp(1, I)
            rr(2) = W(I)*2.0_dp*rp(2, I)
            rr(3) = W(I)*2.0_dp*rp(3, I)
            rr0(1) = W(I)*2.0_dp*r0p(1, I)
            rr0(2) = W(I)*2.0_dp*r0p(2, I)
            rr0(3) = W(I)*2.0_dp*r0p(3, I)

            dm_r(1, 1, 1) = (rr(1)-rr0(1))
            dm_r(1, 1, 2) = (rr(2)-rr0(2))
            dm_r(1, 1, 3) = (rr(3)-rr0(3))

            dm_r(1, 2, 1) = 0.0_dp
            dm_r(1, 2, 2) = rr0(3)
            dm_r(1, 2, 3) = -rr0(2)

            dm_r(1, 3, 1) = -rr0(3)
            dm_r(1, 3, 2) = 0.0_dp
            dm_r(1, 3, 3) = rr0(1)

            dm_r(1, 4, 1) = rr0(2)
            dm_r(1, 4, 2) = -rr0(1)
            dm_r(1, 4, 3) = 0.0_dp

            dm_r(2, 2, 1) = (rr(1)-rr0(1))
            dm_r(2, 2, 2) = (rr(2)+rr0(2))
            dm_r(2, 2, 3) = (rr(3)+rr0(3))

            dm_r(2, 3, 1) = -rr0(2)
            dm_r(2, 3, 2) = -rr0(1)
            dm_r(2, 3, 3) = 0.0_dp

            dm_r(2, 4, 1) = -rr0(3)
            dm_r(2, 4, 2) = 0.0_dp
            dm_r(2, 4, 3) = -rr0(1)

            dm_r(3, 3, 1) = (rr(1)+rr0(1))
            dm_r(3, 3, 2) = (rr(2)-rr0(2))
            dm_r(3, 3, 3) = (rr(3)+rr0(3))

            dm_r(3, 4, 1) = 0.0_dp
            dm_r(3, 4, 2) = -rr0(3)
            dm_r(3, 4, 3) = -rr0(2)

            dm_r(4, 4, 1) = (rr(1)+rr0(1))
            dm_r(4, 4, 2) = (rr(2)+rr0(2))
            dm_r(4, 4, 3) = (rr(3)-rr0(3))

            DO ix = 1, 3
               dm_r(2, 1, ix) = dm_r(1, 2, ix)
               dm_r(3, 1, ix) = dm_r(1, 3, ix)
               dm_r(4, 1, ix) = dm_r(1, 4, ix)
               dm_r(3, 2, ix) = dm_r(2, 3, ix)
               dm_r(4, 2, ix) = dm_r(2, 4, ix)
               dm_r(4, 3, ix) = dm_r(3, 4, ix)
            END DO
            !
            DO ix = 1, 3
               drmsd3(ix, I) = 0.0_dp
               DO k = 1, 4
                  DO j = 1, 4
                     drmsd3(ix, i) = drmsd3(ix, i)+Q(K-1)*Q(j-1)*dm_r(j, k, ix)
                  END DO
               END DO
               drmsd3(ix, I) = drmsd3(ix, I)/REAL(natom, KIND=dp)
            END DO
         END DO
      END IF
      ! Computes the rotation matrix in terms of quaternions
      my_rot(1, 1) = -2.0_dp*Q(2)**2-2.0_dp*Q(3)**2+1.0_dp
      my_rot(1, 2) = 2.0_dp*(-Q(0)*Q(3)+Q(1)*Q(2))
      my_rot(1, 3) = 2.0_dp*(Q(0)*Q(2)+Q(1)*Q(3))
      my_rot(2, 1) = 2.0_dp*(Q(0)*Q(3)+Q(1)*Q(2))
      my_rot(2, 2) = -2.0_dp*Q(1)**2-2.0_dp*Q(3)**2+1.0_dp
      my_rot(2, 3) = 2.0_dp*(-Q(0)*Q(1)+Q(2)*Q(3))
      my_rot(3, 1) = 2.0_dp*(-Q(0)*Q(2)+Q(1)*Q(3))
      my_rot(3, 2) = 2.0_dp*(Q(0)*Q(1)+Q(2)*Q(3))
      my_rot(3, 3) = -2.0_dp*Q(1)**2-2.0_dp*Q(2)**2+1.0_dp
      IF (PRESENT(rot)) rot = my_rot
      ! Give back coordinates rotated in order to minimize the RMSD
      IF (my_rotate) THEN
         DO i = 1, natom
            CALL matvec_3x3(r((i-1)*3+1:(i-1)*3+3), TRANSPOSE(my_rot), rp(:, i))
            r((i-1)*3+1:(i-1)*3+3) = r((i-1)*3+1:(i-1)*3+3)+loc_tr
         END DO
      END IF
      DEALLOCATE (w)
      DEALLOCATE (rp)
      DEALLOCATE (r0p)
   END SUBROUTINE rmsd3

END MODULE rmsd
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>    cjm, Feb-20-2001 : added all the extended variables to
!>    system_type
!>    gt 23-09-2002 : major changes. Pointer part is allocated/deallocated
!>                    and initialized here. Atomic coordinates can now be
!>                    read also from &COORD section in the input file.
!>                    If &COORD is not found, .dat file is read.
!>                    If & coord is found and .NOT. 'INIT', parsing of the .dat
!>                    is performed to get the proper coords/vel/eta variables
!>     CJM 31-7-03  : Major rewrite.  No more atype
! **************************************************************************************************
MODULE atoms_input
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc,&
                                              scaled_to_real
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE cp_parser_methods,               ONLY: read_float_object
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_list_get,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: particle_type
   USE shell_potential_types,           ONLY: shell_kind_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE string_utilities,                ONLY: uppercase
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: read_atoms_input, read_shell_coord_input
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atoms_input'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param topology ...
!> \param overwrite ...
!> \param subsys_section ...
!> \param save_mem ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE read_atoms_input(topology, overwrite, subsys_section, save_mem)

      TYPE(topology_parameters_type)                     :: topology
      LOGICAL, INTENT(IN), OPTIONAL                      :: overwrite
      TYPE(section_vals_type), POINTER                   :: subsys_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'read_atoms_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2*default_string_length)             :: line_att
      CHARACTER(len=default_string_length)               :: error_message, my_default_index, strtmp, &
                                                            unit_str
      INTEGER                                            :: default_id, end_c, handle, iatom, j, &
                                                            natom, output_unit, start_c, wrd
      LOGICAL                                            :: explicit, is_ok, my_overwrite, &
                                                            my_save_mem, scaled_coordinates
      REAL(KIND=dp)                                      :: r0(3), unit_conv
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(section_vals_type), POINTER                   :: coord_section
      TYPE(val_type), POINTER                            :: val

      my_overwrite = .FALSE.
      my_save_mem = .FALSE.
      error_message = ""
      output_unit = cp_logger_get_default_io_unit()
      IF (PRESENT(overwrite)) my_overwrite = overwrite
      IF (PRESENT(save_mem)) my_save_mem = save_mem
      NULLIFY (coord_section)
      coord_section => section_vals_get_subs_vals(subsys_section, "COORD")
      CALL section_vals_get(coord_section, explicit=explicit)
      IF (.NOT. explicit) RETURN

      CALL timeset(routineN, handle)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. get cell and topology%atom_info
      !-----------------------------------------------------------------------------
      atom_info => topology%atom_info
      cell => topology%cell_muc
      CALL section_vals_val_get(coord_section, "UNIT", c_val=unit_str)
      CALL section_vals_val_get(coord_section, "SCALED", l_val=scaled_coordinates)
      unit_conv = cp_unit_to_cp2k(1.0_dp, TRIM(unit_str))

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Read in the coordinates from &COORD section in the input file
      !-----------------------------------------------------------------------------
      CALL section_vals_val_get(coord_section, "_DEFAULT_KEYWORD_", &
                                n_rep_val=natom)
      topology%natoms = natom
      IF (my_overwrite) THEN
         IF(.NOT.(SIZE(atom_info%r, 2) == natom))CALL cp__a("atoms_input.F",121)
         CALL cp_warn(cp__l("atoms_input.F",122), &
                      "Overwriting coordinates. Active coordinates read from &COORD section."// &
                      " Active coordinates READ from &COORD section ")
         CALL section_vals_list_get(coord_section, "_DEFAULT_KEYWORD_", list=list)
         DO iatom = 1, natom
            is_ok = cp_sll_val_next(list, val)
            CALL val_get(val, c_val=line_att)
            ! Read name and atomic coordinates
            start_c = 1
            DO wrd = 1, 4
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) /= ' ') THEN
                     start_c = j
                     EXIT
                  END IF
               END DO
               end_c = LEN(line_att)+1
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) == ' ') THEN
                     end_c = j
                     EXIT
                  END IF
               END DO
               IF (LEN_TRIM(line_att(start_c:end_c-1)) == 0) &
                  CALL cp__b("atoms_input.F",146,"incorrectly formatted line in coord section'"//line_att//"'")
               IF (wrd == 1) THEN
                  atom_info%id_atmname(iatom) = str2id(s2s(line_att(start_c:end_c-1)))
               ELSE
                  READ (line_att(start_c:end_c-1), *) atom_info%r(wrd-1, iatom)
               END IF
               start_c = end_c
            END DO
         END DO
      ELSE
         ! Element is assigned on the basis of the atm_name
         topology%aa_element = .TRUE.

         CALL reallocate(atom_info%id_molname, 1, natom)
         CALL reallocate(atom_info%id_resname, 1, natom)
         CALL reallocate(atom_info%resid, 1, natom)
         CALL reallocate(atom_info%id_atmname, 1, natom)
         CALL reallocate(atom_info%id_element, 1, natom)
         CALL reallocate(atom_info%r, 1, 3, 1, natom)
         CALL reallocate(atom_info%atm_mass, 1, natom)
         CALL reallocate(atom_info%atm_charge, 1, natom)

         CALL section_vals_list_get(coord_section, "_DEFAULT_KEYWORD_", list=list)
         DO iatom = 1, natom
            ! we use only the first default_string_length characters of each line
            is_ok = cp_sll_val_next(list, val)
            CALL val_get(val, c_val=line_att)
            default_id = str2id(s2s(""))
            atom_info%id_molname(iatom) = default_id
            atom_info%id_resname(iatom) = default_id
            atom_info%resid(iatom) = 1
            atom_info%id_atmname(iatom) = default_id
            atom_info%id_element(iatom) = default_id
            topology%molname_generated = .TRUE.
            ! Read name and atomic coordinates
            start_c = 1
            DO wrd = 1, 6
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) /= ' ') THEN
                     start_c = j
                     EXIT
                  END IF
               END DO
               end_c = LEN(line_att)+1
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) == ' ') THEN
                     end_c = j
                     EXIT
                  END IF
               END DO
               IF (LEN_TRIM(line_att(start_c:end_c-1)) == 0) &
                  CALL cp_abort(cp__l("atoms_input.F",197), &
                                "Incorrectly formatted input line for atom "// &
                                TRIM(ADJUSTL(cp_to_string(iatom)))// &
                                " found in COORD section. Input line: <"// &
                                TRIM(line_att)//"> ")
               SELECT CASE (wrd)
               CASE (1)
                  atom_info%id_atmname(iatom) = str2id(s2s(line_att(start_c:end_c-1)))
               CASE (2:4)
                  CALL read_float_object(line_att(start_c:end_c-1), &
                                         atom_info%r(wrd-1, iatom), error_message)
                  IF (LEN_TRIM(error_message) /= 0) &
                     CALL cp_abort(cp__l("atoms_input.F",209), &
                                   "Incorrectly formatted input line for atom "// &
                                   TRIM(ADJUSTL(cp_to_string(iatom)))// &
                                   " found in COORD section. "//TRIM(error_message)// &
                                   " Input line: <"//TRIM(line_att)//"> ")
               CASE (5)
                  READ (line_att(start_c:end_c-1), *) strtmp
                  atom_info%id_molname(iatom) = str2id(strtmp)
                  atom_info%id_resname(iatom) = atom_info%id_molname(iatom)
                  topology%molname_generated = .FALSE.
               CASE (6)
                  READ (line_att(start_c:end_c-1), *) strtmp
                  atom_info%id_resname(iatom) = str2id(strtmp)
               END SELECT
               start_c = end_c
               IF (start_c > LEN_TRIM(line_att)) EXIT
            END DO
            IF (topology%molname_generated) THEN
               ! Use defaults, if no molname was specified
               WRITE (my_default_index, '(I0)') iatom
               atom_info%id_molname(iatom) = str2id(s2s(TRIM(id2str(atom_info%id_atmname(iatom)))//TRIM(my_default_index)))
               atom_info%id_resname(iatom) = atom_info%id_molname(iatom)
            END IF
            atom_info%id_element(iatom) = atom_info%id_atmname(iatom)
            atom_info%atm_mass(iatom) = 0.0_dp
            atom_info%atm_charge(iatom) = -HUGE(0.0_dp)
         END DO
      END IF
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Convert coordinates into internal cp2k coordinates
      !-----------------------------------------------------------------------------
      DO iatom = 1, natom
         IF (scaled_coordinates) THEN
            r0 = atom_info%r(:, iatom)
            CALL scaled_to_real(atom_info%r(:, iatom), r0, cell)
         ELSE
            atom_info%r(:, iatom) = atom_info%r(:, iatom)*unit_conv
         END IF
      END DO
      IF (my_save_mem) CALL section_vals_remove_values(coord_section)

      CALL timestop(handle)
   END SUBROUTINE read_atoms_input

! **************************************************************************************************
!> \brief ...
!> \param particle_set ...
!> \param shell_particle_set ...
!> \param cell ...
!> \param subsys_section ...
!> \param core_particle_set ...
!> \param save_mem ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE read_shell_coord_input(particle_set, shell_particle_set, cell, &
                                     subsys_section, core_particle_set, save_mem)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, shell_particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: subsys_section
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: core_particle_set
      LOGICAL, INTENT(IN), OPTIONAL                      :: save_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'read_shell_coord_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2*default_string_length)             :: line_att
      CHARACTER(len=default_string_length)               :: name_kind, unit_str
      CHARACTER(len=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: at_name, at_name_c
      INTEGER                                            :: end_c, handle, ishell, j, nshell, &
                                                            output_unit, sh_index, start_c, wrd
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: at_index, at_index_c
      LOGICAL                                            :: core_scaled_coordinates, explicit, &
                                                            is_ok, is_shell, my_save_mem, &
                                                            shell_scaled_coordinates
      REAL(KIND=dp)                                      :: dab, mass_com, rab(3), unit_conv_core, &
                                                            unit_conv_shell
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r, rc
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(section_vals_type), POINTER                   :: core_coord_section, shell_coord_section
      TYPE(shell_kind_type), POINTER                     :: shell
      TYPE(val_type), POINTER                            :: val

      my_save_mem = .FALSE.
      NULLIFY (atomic_kind, list, shell_coord_section, shell, val)
      output_unit = cp_logger_get_default_io_unit()

      IF (PRESENT(save_mem)) my_save_mem = save_mem
      NULLIFY (shell_coord_section, core_coord_section)
      shell_coord_section => section_vals_get_subs_vals(subsys_section, "SHELL_COORD")
      CALL section_vals_get(shell_coord_section, explicit=explicit)
      IF (.NOT. explicit) RETURN

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(particle_set)))CALL cp__a("atoms_input.F",307)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Read in the coordinates from &SHELL_COORD section in the input file
      !-----------------------------------------------------------------------------
      CALL section_vals_val_get(shell_coord_section, "UNIT", c_val=unit_str)
      CALL section_vals_val_get(shell_coord_section, "SCALED", l_val=shell_scaled_coordinates)
      unit_conv_shell = cp_unit_to_cp2k(1.0_dp, TRIM(unit_str))
      CALL section_vals_val_get(shell_coord_section, "_DEFAULT_KEYWORD_", &
                                n_rep_val=nshell)

      IF (ASSOCIATED(shell_particle_set)) THEN
         IF(.NOT.((SIZE(shell_particle_set, 1) == nshell)))CALL cp__a("atoms_input.F",319)
         ALLOCATE (r(3, nshell), at_name(nshell), at_index(nshell))
         CALL cp_warn(cp__l("atoms_input.F",321), &
                      "Overwriting shell coordinates. "// &
                      "Active coordinates READ from &SHELL_COORD section. ")
         CALL section_vals_list_get(shell_coord_section, "_DEFAULT_KEYWORD_", list=list)
         DO ishell = 1, nshell
            ! we use only the first default_string_length characters of each line
            is_ok = cp_sll_val_next(list, val)
            CALL val_get(val, c_val=line_att)
            start_c = 1
            DO wrd = 1, 5
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) /= ' ') THEN
                     start_c = j
                     EXIT
                  END IF
               END DO
               end_c = LEN(line_att)+1
               DO j = start_c, LEN(line_att)
                  IF (line_att(j:j) == ' ') THEN
                     end_c = j
                     EXIT
                  END IF
               END DO
               IF (wrd /= 5 .AND. end_c >= LEN(line_att)+1) &
                  CALL cp__b("atoms_input.F",345,"incorrectly formatted line in coord section'"//line_att//"'")
               IF (wrd == 1) THEN
                  at_name(ishell) = line_att(start_c:end_c-1)
                  CALL uppercase(at_name(ishell))
               ELSE IF (wrd == 5) THEN
                  READ (line_att(start_c:end_c-1), *) at_index(ishell)
               ELSE
                  READ (line_att(start_c:end_c-1), *) r(wrd-1, ishell)
               END IF
               start_c = end_c
            END DO
         END DO

         IF (PRESENT(core_particle_set)) THEN
            IF(.NOT.(ASSOCIATED(core_particle_set)))CALL cp__a("atoms_input.F",359)
            core_coord_section => section_vals_get_subs_vals(subsys_section, "CORE_COORD")
            CALL section_vals_get(core_coord_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(core_coord_section, "UNIT", c_val=unit_str)
               CALL section_vals_val_get(core_coord_section, "SCALED", l_val=core_scaled_coordinates)
               unit_conv_core = cp_unit_to_cp2k(1.0_dp, TRIM(unit_str))
               CALL section_vals_val_get(core_coord_section, "_DEFAULT_KEYWORD_", &
                                         n_rep_val=nshell)

               IF(.NOT.((SIZE(core_particle_set, 1) == nshell)))CALL cp__a("atoms_input.F",369)
               ALLOCATE (rc(3, nshell), at_name_c(nshell), at_index_c(nshell))
               CALL cp_warn(cp__l("atoms_input.F",371), &
                            "Overwriting cores coordinates. "// &
                            " Active coordinates READ from &CORE_COORD section. ")
               CALL section_vals_list_get(core_coord_section, "_DEFAULT_KEYWORD_", list=list)
               DO ishell = 1, nshell
                  ! we use only the first default_string_length characters of each line
                  is_ok = cp_sll_val_next(list, val)
                  CALL val_get(val, c_val=line_att)
                  start_c = 1
                  DO wrd = 1, 5
                     DO j = start_c, LEN(line_att)
                        IF (line_att(j:j) /= ' ') THEN
                           start_c = j
                           EXIT
                        END IF
                     END DO
                     end_c = LEN(line_att)+1
                     DO j = start_c, LEN(line_att)
                        IF (line_att(j:j) == ' ') THEN
                           end_c = j
                           EXIT
                        END IF
                     END DO
                     IF (wrd /= 5 .AND. end_c >= LEN(line_att)+1) &
                        CALL cp__b("atoms_input.F",395,"incorrectly formatted line in coord section'"//line_att//"'")
                     IF (wrd == 1) THEN
                        at_name_c(ishell) = line_att(start_c:end_c-1)
                        CALL uppercase(at_name_c(ishell))
                     ELSE IF (wrd == 5) THEN
                        READ (line_att(start_c:end_c-1), *) at_index_c(ishell)
                     ELSE
                        READ (line_att(start_c:end_c-1), *) rc(wrd-1, ishell)
                     END IF
                     start_c = end_c
                  END DO
               END DO
               IF (my_save_mem) CALL section_vals_remove_values(core_coord_section)
            END IF ! explicit
         END IF ! core_particle_set

         !-----------------------------------------------------------------------------
         ! 3. Check corrispondence and convert coordinates into internal cp2k coordinates
         !-----------------------------------------------------------------------------
         DO ishell = 1, nshell
            atomic_kind => particle_set(at_index(ishell))%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_kind, shell_active=is_shell, mass=mass_com, shell=shell)
            CALL uppercase(name_kind)
            IF ((TRIM(at_name(ishell)) == TRIM(name_kind)) .AND. is_shell) THEN
               sh_index = particle_set(at_index(ishell))%shell_index
               IF (shell_scaled_coordinates) THEN
                  CALL scaled_to_real(r(:, ishell), shell_particle_set(sh_index)%r(:), cell)
               ELSE
                  shell_particle_set(sh_index)%r(:) = r(:, ishell)*unit_conv_shell
               END IF
               shell_particle_set(sh_index)%atom_index = at_index(ishell)

               IF (PRESENT(core_particle_set) .AND. .NOT. explicit) THEN
                  core_particle_set(sh_index)%r(1) = (mass_com*particle_set(at_index(ishell))%r(1)- &
                                                      shell%mass_shell*shell_particle_set(sh_index)%r(1))/shell%mass_core
                  core_particle_set(sh_index)%r(2) = (mass_com*particle_set(at_index(ishell))%r(2)- &
                                                      shell%mass_shell*shell_particle_set(sh_index)%r(2))/shell%mass_core
                  core_particle_set(sh_index)%r(3) = (mass_com*particle_set(at_index(ishell))%r(3)- &
                                                      shell%mass_shell*shell_particle_set(sh_index)%r(3))/shell%mass_core
                  core_particle_set(sh_index)%atom_index = at_index(ishell)
                  rab = pbc(shell_particle_set(sh_index)%r, core_particle_set(sh_index)%r, cell)
               ELSE IF (explicit) THEN
                  IF (core_scaled_coordinates) THEN
                     CALL scaled_to_real(rc(:, ishell), core_particle_set(sh_index)%r(:), cell)
                  ELSE
                     core_particle_set(sh_index)%r(:) = rc(:, ishell)*unit_conv_core
                  END IF
                  core_particle_set(sh_index)%atom_index = at_index_c(ishell)
                  rab = pbc(shell_particle_set(sh_index)%r, core_particle_set(sh_index)%r, cell)
                  IF(.NOT.(TRIM(at_name(ishell)) == TRIM(at_name_c(ishell))))CALL cp__a("atoms_input.F",445)
                  IF(.NOT.(at_index(ishell) == at_index_c(ishell)))CALL cp__a("atoms_input.F",446)
               ELSE
                  rab = pbc(shell_particle_set(sh_index)%r, particle_set(at_index(ishell))%r, cell)
               END IF

               dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))
               IF (shell%max_dist > 0.0_dp .AND. shell%max_dist < dab) THEN
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, *) "WARNING : shell and core for atom ", at_index(ishell), " seem to be too distant. "
                  END IF
               END IF

            ELSE
               CALL cp__b("atoms_input.F",459,"shell coordinate assigned to the wrong atom. check the shell indexes in the input")
            END IF
         END DO
         DEALLOCATE (r, at_index, at_name)
         DEALLOCATE (rc, at_index_c, at_name_c)

      END IF

      IF (my_save_mem) CALL section_vals_remove_values(shell_coord_section)

      CALL timestop(handle)

   END SUBROUTINE read_shell_coord_input

END MODULE atoms_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE pair_potential_coulomb

   USE erf_fn,                          ONLY: erf,&
                                              erfc
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: oorootpi
   USE pw_poisson_types,                ONLY: do_ewald_none
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pair_potential_coulomb'

   PUBLIC :: potential_coulomb

CONTAINS

! **************************************************************************************************
!> \brief Evaluates the electrostatic energy and force
!> \param r2 ...
!> \param fscalar ...
!> \param qfac ...
!> \param ewald_type ...
!> \param alpha ...
!> \param beta ...
!> \param interaction_cutoff ...
!> \retval potential_coulomb ...
!> \author Toon.Verstraelen@gmail.com
! **************************************************************************************************
   FUNCTION potential_coulomb(r2, fscalar, qfac, ewald_type, alpha, beta, &
                              interaction_cutoff)

      REAL(KIND=dp), INTENT(IN)                          :: r2
      REAL(KIND=dp), INTENT(INOUT)                       :: fscalar
      REAL(KIND=dp), INTENT(IN)                          :: qfac
      INTEGER, INTENT(IN)                                :: ewald_type
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta, interaction_cutoff
      REAL(KIND=dp)                                      :: potential_coulomb

      REAL(KIND=dp), PARAMETER :: two_over_sqrt_pi = 2.0_dp*oorootpi

      REAL(KIND=dp)                                      :: r, x1, x2

      r = SQRT(r2)
      IF (beta > 0.0_dp) THEN
         IF (ewald_type == do_ewald_none) THEN
            x2 = r*beta
            potential_coulomb = erf(x2)/r
            fscalar = fscalar+qfac*(potential_coulomb- &
                                    two_over_sqrt_pi*EXP(-x2*x2)*beta)/r2
         ELSE
            x1 = alpha*r
            x2 = r*beta
            potential_coulomb = (erf(x2)-erf(x1))/r
            fscalar = fscalar+qfac*(potential_coulomb+ &
                                    two_over_sqrt_pi*(EXP(-x1*x1)*alpha-EXP(-x2*x2)*beta))/r2
         END IF
      ELSE
         IF (ewald_type == do_ewald_none) THEN
            potential_coulomb = 1.0_dp/r
            fscalar = fscalar+qfac*potential_coulomb/r2
         ELSE
            x1 = alpha*r
            potential_coulomb = erfc(x1)/r
            fscalar = fscalar+qfac*(potential_coulomb+ &
                                    two_over_sqrt_pi*EXP(-x1*x1)*alpha)/r2
         END IF
      END IF

      potential_coulomb = qfac*(potential_coulomb-interaction_cutoff)

   END FUNCTION potential_coulomb

END MODULE pair_potential_coulomb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines for handling splines
!> \par History
!>      2001-09-21-HAF added this doc entry and changed formatting
!> \author various
! **************************************************************************************************
MODULE splines_methods

   USE cp_log_handling,                 ONLY: cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE splines_types,                   ONLY: spline_data_p_type,&
                                              spline_data_type,&
                                              spline_factor_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'splines_methods'

   PUBLIC :: init_splinexy ! allocates x and y vectors for splines
   PUBLIC :: init_spline ! generate table for spline (allocates y2)
   PUBLIC :: potential_s ! return value of spline and 1. derivative
   ! without checks (fast routine for pair_potential)
   PUBLIC :: spline_value ! return value of spline and 1. derivative
   ! without check (without assumption of 1/x^2 grid)

CONTAINS

! **************************************************************************************************
!> \brief allocates storage for function table to be interpolated
!>      both x and y are allocated
!> \param spl spline_data structure to be initialized
!> \param nn integer number of datapoints, that the function table will hold
!> \par History
!>      2001-09-21-HAF added this doc entry and changed formatting
!> \author unknown
! **************************************************************************************************
   SUBROUTINE init_splinexy(spl, nn)

      TYPE(spline_data_type), POINTER                    :: spl
      INTEGER, INTENT(IN)                                :: nn

      CHARACTER(len=*), PARAMETER :: routineN = 'init_splinexy', routineP = moduleN//':'//routineN

      INTEGER                                            :: err

      spl%n = nn

      IF (ASSOCIATED(spl%y)) THEN
         DEALLOCATE (spl%y)
         NULLIFY (spl%y)
      END IF

      IF (ASSOCIATED(spl%y2)) THEN
         DEALLOCATE (spl%y2)
         NULLIFY (spl%y2)
      END IF

      ALLOCATE (spl%y(1:nn), STAT=err)
      IF (err /= 0) CALL stop_memory(routineN, moduleN, 69, &
                                     "spl%y", dp_size*nn)

      ALLOCATE (spl%y2(1:nn), STAT=err)
      IF (err /= 0) CALL stop_memory(routineN, moduleN, 73, &
                                     "spl%y2", dp_size*nn)

   END SUBROUTINE init_splinexy

! **************************************************************************************************
!> \brief allocates storage for y2 table
!>      calculates y2 table and other spline parameters
!> \param spl spline_data structure to be initialized
!>                       spl%y() must hold the function values
!>                       spl%x() must hold the absissa values in increasing
!>                       order OR if dx (below) is given, spl%x(1) must hold
!>                       the starting (left-most) point.
!> \param dx x(i) are assumed to be x(1)+dx*(i-1)
!>                       (spline evaluations will also be faster)
!>      y1a : (OPTIONAL) if present, the 1-deriv of the left endpoint
!>                       if not present, natural spline condition at this end
!>                       (2-deriv == 0)
!>      y1b : (OPTIONAL) if present, the 1-deriv of the right endpoint
!>                       if not present, natural spline condition at this end
!>                       (2-deriv == 0)
!> \param y1a ...
!> \param y1b ...
!> \par Examples
!>      CALL init_spline(spline,dx=0.1_dp)
!>      CALL init_spline(spline,y1b=0.0_dp)
!>      CALL init_spline(spline,0.1_dp,0.0_dp,0.0_dp)
!> \par History
!>      2001-09-21-HAF added this doc entry and changed formatting
!>      2001-09-24-HAF changed interface and re-written
!> \author unknown
!> \note
!>      if dx is given, the x array will be used as y2 array instead of
!>      allocating a new array. (y2 will become x, and x will be nullified)
! **************************************************************************************************
   SUBROUTINE init_spline(spl, dx, y1a, y1b)

      TYPE(spline_data_type), POINTER                    :: spl
      REAL(KIND=dp), INTENT(IN)                          :: dx
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: y1a, y1b

      CHARACTER(len=*), PARAMETER :: routineN = 'init_spline', routineP = moduleN//':'//routineN

      INTEGER                                            :: err, i, n
      REAL(KIND=dp)                                      :: p, s
      REAL(KIND=dp), POINTER                             :: ww(:)

      n = spl%n
      spl%xn = spl%x1+(n-1)*dx
      spl%h = dx
      spl%invh = 1.0_dp/dx
      spl%h26 = dx**2/6.0_dp
      ALLOCATE (ww(1:n), STAT=err)
      IF (err /= 0) CALL stop_memory(routineN, moduleN, 126, &
                                     "ww", dp_size*n)
      IF (PRESENT(y1a)) THEN
         spl%y2(1) = -0.5_dp
         ww(1) = 3.0_dp*((spl%y(2)-spl%y(1))/dx-y1a)/dx
      ELSE
         spl%y2(1) = 0.0_dp
         ww(1) = 0.0_dp
      END IF
      DO i = 2, n-1
         s = 0.5_dp
         p = 0.5_dp*spl%y2(i-1)+2.0_dp
         spl%y2(i) = -0.5_dp/p
         ww(i) = (3.0_dp*(spl%y(i+1)-2.0_dp*spl%y(i)+spl%y(i-1))/(dx*dx) &
                  -0.5_dp*ww(i-1))/p
      END DO
      IF (PRESENT(y1b)) THEN
         spl%y2(n) = (3.0_dp*(y1b-(spl%y(n)-spl%y(n-1))/dx)/dx- &
                      0.5_dp*ww(n-1))/(0.5_dp*spl%y2(n-1)+1.0_dp)
      ELSE
         spl%y2(n) = 0.0_dp
      END IF
      DO i = n-1, 1, -1
         spl%y2(i) = spl%y2(i)*spl%y2(i+1)+ww(i)
      END DO
      DEALLOCATE (ww)

   END SUBROUTINE init_spline

! **************************************************************************************************
!> \brief calculates the potential interpolated with splines value at a given point
!>      and the first derivative. Checks included to avoid just segfaulting!!
!> \param spl_p spline_data structure
!> \param xxi absissa value
!> \param y1 1. derivative at xx
!> \param spl_f ...
!> \param logger ...
!> \retval potential_s ...
!> \par Output
!>      spline interpolated value at xx
!> \par History
!>      2001-09-25-HAF added this doc entry and changed formatting
!> \author unknown
!> \note
!>      the spline MUST have uniform x values and xx MUST be
!>      in the interpolation interval. No checks are done to ensure
!>      either condition.
! **************************************************************************************************
   FUNCTION potential_s(spl_p, xxi, y1, spl_f, logger)
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p
      REAL(KIND=dp), INTENT(IN)                          :: xxi
      REAL(KIND=dp), INTENT(OUT)                         :: y1
      TYPE(spline_factor_type), POINTER                  :: spl_f
      TYPE(cp_logger_type), POINTER                      :: logger
      REAL(KIND=dp)                                      :: potential_s

      REAL(KIND=dp), PARAMETER                           :: f13 = 1.0_dp/3.0_dp

      INTEGER                                            :: i, output_unit
      REAL(KIND=dp)                                      :: a, b, h26, invh, x4, xx, xx0, y2hi, &
                                                            y2lo, yhi, ylo, yy

      xx0 = 1.0_dp/xxi
      xx = spl_f%rscale(1)*xx0
      x4 = xx*xx
      h26 = spl_p(1)%spline_data%h26
      invh = spl_p(1)%spline_data%invh
      IF (xx >= spl_p(1)%spline_data%xn) THEN
         ! In case the value is not on the spline let's print a warning and give the value
         ! for the smaller point available in the spline..
         ! This should happen in very few cases though..
         output_unit = cp_logger_get_default_unit_nr(logger)
         yy = spl_p(1)%spline_data%xn-spl_p(1)%spline_data%h
         WRITE (output_unit, FMT='(/,80("*"),/,"*",1X,"Value of r in Input =",F11.6,'// &
                '" not in the spline range. Using =",F11.6,T80,"*",/,80("*"))') SQRT(1.0_dp/xx), SQRT(1.0_dp/yy)
         xx = yy
      END IF
      i = INT((xx-spl_p(1)%spline_data%x1)*invh+1)
      a = (spl_p(1)%spline_data%x1-xx)*invh+REAL(i, kind=dp)
      b = 1.0_dp-a

      ylo = spl_p(1)%spline_data%y(i)
      yhi = spl_p(1)%spline_data%y(i+1)
      y2lo = spl_p(1)%spline_data%y2(i)
      y2hi = spl_p(1)%spline_data%y2(i+1)
      potential_s = (a*ylo+b*yhi-((a+1.0_dp)*y2lo+(b+1.0_dp)*y2hi)*a*b*h26)*spl_f%fscale(1)
      y1 = invh*((yhi-ylo)+((f13-a*a)*y2lo-(f13-b*b)*y2hi)*3.0_dp*h26)
      y1 = 2.0_dp*y1*x4*spl_f%dscale(1)

      potential_s = potential_s+spl_f%cutoff
   END FUNCTION potential_s

! **************************************************************************************************
!> \brief calculates the spline value at a given point
!>        (and possibly the first derivative) WITHOUT checks
!>        and without any funny scaling business, or weird
!>        1/x^2 grid assumptions
!>
!> \param spl ...
!> \param xx ...
!> \param y1 ...
!> \retval spline_value ...
!> \author HAF
! **************************************************************************************************
   FUNCTION spline_value(spl, xx, y1)
      ! Return value
      TYPE(spline_data_type), POINTER                    :: spl
      REAL(KIND=dp), INTENT(IN)                          :: xx
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: y1
      REAL(KIND=dp)                                      :: spline_value

      REAL(KIND=dp), PARAMETER                           :: f13 = 1.0_dp/3.0_dp

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: a, b, h26, invh, y2hi, y2lo, yhi, ylo

      h26 = spl%h26
      invh = spl%invh
      i = INT((xx-spl%x1)*invh+1)

      a = (spl%x1-xx)*invh+REAL(i, kind=dp)
      b = 1.0_dp-a
      ylo = spl%y(i)
      yhi = spl%y(i+1)
      y2lo = spl%y2(i)
      y2hi = spl%y2(i+1)
      spline_value = a*ylo+b*yhi-((a+1.0_dp)*y2lo+(b+1.0_dp)*y2hi)*a*b*h26
      IF (PRESENT(y1)) y1 = invh*((yhi-ylo)+ &
                                  ((f13-a*a)*y2lo-(f13-b*b)*y2hi)*3.0_dp*h26)
   END FUNCTION spline_value

END MODULE splines_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      September 2005 - Introduced the Born-Mayer-Huggins-Fumi-Tosi  Potential (BMHTF)
!>      2006 - Major rewriting of the routines.. Linear scaling setup of splines
!> \author CJM
! **************************************************************************************************
MODULE pair_potential_util

   USE fparser,                         ONLY: EvalErrType,&
                                              evalf
   USE kinds,                           ONLY: dp
   USE pair_potential_types,            ONLY: &
        b4_type, bm_type, ea_type, ft_type, ftd_type, gp_type, gw_type, ip_type, lj_charmm_type, &
        lj_type, not_initialized, pair_potential_single_type, wl_type
   USE physcon,                         ONLY: bohr,&
                                              evolt
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pair_potential_util'
   REAL(KIND=dp), PARAMETER, PRIVATE    :: MIN_HICUT_VALUE = 1.0E-15_dp, &
                                           DEFAULT_HICUT_VALUE = 1.0E3_dp

   PUBLIC :: ener_pot, ener_zbl, zbl_matching_polinomial

CONTAINS

! **************************************************************************************************
!> \brief Evaluates the nonbond potential energy for the implemented FF kinds
!> \param pot ...
!> \param r ...
!> \param energy_cutoff ...
!> \retval value ...
! **************************************************************************************************
   FUNCTION ener_pot(pot, r, energy_cutoff) RESULT(value)
      TYPE(pair_potential_single_type), POINTER          :: pot
      REAL(KIND=dp), INTENT(IN)                          :: r, energy_cutoff
      REAL(KIND=dp)                                      :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'ener_pot', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, index, j, n
      REAL(KIND=dp)                                      :: damp, dampexp, dampsum, f6, f8, &
                                                            factorial, lvalue, pp, qq, scale, xf

      value = 0.0_dp
      DO j = 1, SIZE(pot%type)
         ! A lower boundary for the potential definition was defined
         IF ((pot%set(j)%rmin /= not_initialized) .AND. (r < pot%set(j)%rmin)) CYCLE
         ! An upper boundary for the potential definition was defined
         IF ((pot%set(j)%rmax /= not_initialized) .AND. (r >= pot%set(j)%rmax)) CYCLE
         ! If within limits let's compute the potential...
         IF (pot%type(j) == lj_charmm_type) THEN
            lvalue = &
               4.0_dp*pot%set(j)%lj%epsilon*(pot%set(j)%lj%sigma12*r**(-12)-pot%set(j)%lj% &
                                             sigma6*r**(-6))
         ELSE IF (pot%type(j) == lj_type) THEN
            lvalue = pot%set(j)%lj%epsilon* &
                     (pot%set(j)%lj%sigma12*r**(-12)-pot%set(j)%lj%sigma6*r**(-6))
         ELSE IF (pot%type(j) == ip_type) THEN
            lvalue = 0._dp
            IF (r > pot%set(j)%ipbv%rcore) THEN
               DO i = 2, 15
                  lvalue = lvalue+pot%set(j)%ipbv%a(i)/(r**(i-1)*REAL(i-1, dp))
               END DO
            ELSE
               ! use a linear potential
               lvalue = pot%set(j)%ipbv%m*r+pot%set(j)%ipbv%b
            ENDIF
            lvalue = lvalue
         ELSE IF (pot%type(j) == wl_type) THEN
            lvalue = pot%set(j)%willis%a*EXP(-pot%set(j)%willis%b*r)-pot%set(j)%willis%c/r**6
         ELSE IF (pot%type(j) == gw_type) THEN
            scale = EXP(pot%set(j)%goodwin%m*(-(r/pot%set(j)%goodwin%dc)**pot%set(j)%goodwin%mc+ &
                                              (pot%set(j)%goodwin%d/pot%set(j)%goodwin%dc)**pot%set(j)%goodwin%mc))
            lvalue = scale*pot%set(j)%goodwin%vr0*(pot%set(j)%goodwin%d/r)**pot%set(j)%goodwin%m
         ELSE IF (pot%type(j) == ft_type) THEN
            lvalue = pot%set(j)%ft%a*EXP(-pot%set(j)%ft%b*r)-pot%set(j)%ft%c/r**6-pot%set(j)%ft%d/r**8
         ELSE IF (pot%type(j) == ftd_type) THEN
            dampsum = 1.0_dp
            xf = 1.0_dp
            factorial = 1.0_dp
            damp = pot%set(j)%ftd%bd
            dampexp = dexp(-damp*r)
            DO i = 1, 6
               xf = xf*damp*r
               factorial = factorial*float(i)
               dampsum = dampsum+xf/factorial
            ENDDO
            f6 = 1.0_dp-dampexp*dampsum
            DO i = 7, 8
               xf = xf*damp*r
               factorial = factorial*float(i)
               dampsum = dampsum+xf/factorial
            ENDDO
            f8 = 1.0_dp-dampexp*dampsum
            lvalue = pot%set(j)%ftd%a*EXP(-pot%set(j)%ftd%b*r)-f6*pot%set(j)%ftd%c/r**6-f8*pot%set(j)%ftd%d/r**8
         ELSE IF (pot%type(j) == ea_type) THEN
            index = INT(r/pot%set(j)%eam%drar)+1
            IF (index > pot%set(j)%eam%npoints) THEN
               index = pot%set(j)%eam%npoints
            ELSEIF (index < 1) THEN
               index = 1
            ENDIF
            qq = r-pot%set(j)%eam%rval(index)
            pp = pot%set(j)%eam%phi(index)+ &
                 qq*pot%set(j)%eam%phip(index)
            lvalue = pp
         ELSE IF (pot%type(j) == b4_type) THEN
            IF (r <= pot%set(j)%buck4r%r1) THEN
               pp = pot%set(j)%buck4r%a*EXP(-pot%set(j)%buck4r%b*r)
            ELSEIF (r > pot%set(j)%buck4r%r1 .AND. r <= pot%set(j)%buck4r%r2) THEN
               pp = 0.0_dp
               DO n = 0, pot%set(j)%buck4r%npoly1
                  pp = pp+pot%set(j)%buck4r%poly1(n)*r**n
               END DO
            ELSEIF (r > pot%set(j)%buck4r%r2 .AND. r <= pot%set(j)%buck4r%r3) THEN
               pp = 0.0_dp
               DO n = 0, pot%set(j)%buck4r%npoly2
                  pp = pp+pot%set(j)%buck4r%poly2(n)*r**n
               END DO
            ELSEIF (r > pot%set(j)%buck4r%r3) THEN
               pp = -pot%set(j)%buck4r%c/r**6
            END IF
            lvalue = pp
         ELSE IF (pot%type(j) == bm_type) THEN
            lvalue = pot%set(j)%buckmo%f0*(pot%set(j)%buckmo%b1+pot%set(j)%buckmo%b2)* &
                     EXP((pot%set(j)%buckmo%a1+pot%set(j)%buckmo%a2-r)/(pot%set(j)%buckmo%b1+pot%set(j)%buckmo%b2)) &
                     -pot%set(j)%buckmo%c/r**6 &
                     +pot%set(j)%buckmo%d*(EXP(-2._dp*pot%set(j)%buckmo%beta*(r-pot%set(j)%buckmo%r0))- &
                                           2.0_dp*EXP(-pot%set(j)%buckmo%beta*(r-pot%set(j)%buckmo%r0)))
         ELSE IF (pot%type(j) == gp_type) THEN
            pot%set(j)%gp%values(1) = r
            lvalue = evalf(pot%set(j)%gp%myid, pot%set(j)%gp%values)
            IF (EvalErrType > 0) &
               CALL cp__b("pair_potential_util.F",144,"Error evaluating generic potential energy function")
         ELSE
            lvalue = 0.0_dp
         END IF
         value = value+lvalue
      END DO
      value = value-energy_cutoff
   END FUNCTION ener_pot

! **************************************************************************************************
!> \brief Evaluates the ZBL scattering potential, very short range
!>        Only shell-model for interactions among pairs without repulsive term
!> \param pot ...
!> \param r ...
!> \retval ener_zbl ...
!> \author i soliti ignoti
! **************************************************************************************************
   FUNCTION ener_zbl(pot, r)

      TYPE(pair_potential_single_type), POINTER          :: pot
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ener_zbl

      REAL(KIND=dp)                                      :: au, fac, x

      ener_zbl = 0.0_dp
      IF (r <= pot%zbl_rcut(1)) THEN
         au = 0.88534_dp*bohr/(pot%z1**0.23_dp+pot%z2**0.23_dp)
         x = r/au
         fac = pot%z1*pot%z2/evolt
         ener_zbl = fac/r*(0.1818_dp*EXP(-3.2_dp*x)+0.5099_dp*EXP(-0.9423_dp*x)+ &
                           0.2802_dp*EXP(-0.4029_dp*x)+0.02817_dp*EXP(-0.2016_dp*x))
      ELSEIF (r > pot%zbl_rcut(1) .AND. r <= pot%zbl_rcut(2)) THEN
         ener_zbl = pot%zbl_poly(0)+pot%zbl_poly(1)*r+pot%zbl_poly(2)*r*r+pot%zbl_poly(3)*r*r*r+ &
                    pot%zbl_poly(4)*r*r*r*r+pot%zbl_poly(5)*r*r*r*r*r
      ELSE
         ener_zbl = 0.0_dp
      END IF

   END FUNCTION ener_zbl

! **************************************************************************************************
!> \brief Determine the polinomial coefficients used to set to zero the zbl potential
!>        at the cutoff radius, with continuity in function, first and second derivative
!>        Only shell-model for interactions among pairs without repulsive term
!> \param pot ...
!> \param rcov1 ...
!> \param rcov2 ...
!> \param z1 ...
!> \param z2 ...
!> \author i soliti ignoti
! **************************************************************************************************
   SUBROUTINE zbl_matching_polinomial(pot, rcov1, rcov2, z1, z2)

      TYPE(pair_potential_single_type), POINTER          :: pot
      REAL(KIND=dp), INTENT(IN)                          :: rcov1, rcov2, z1, z2

      REAL(KIND=dp)                                      :: au, d1, d2, dd1, dd2, fac, v1, v2, x, &
                                                            x1, x2

      pot%zbl_rcut(1) = (rcov1+rcov2)*(1.0_dp-0.2_dp)*bohr
      pot%zbl_rcut(2) = (rcov1+rcov2)*bohr
      x1 = pot%zbl_rcut(1)
      x2 = pot%zbl_rcut(2)
      pot%z1 = z1
      pot%z2 = z2

      au = 0.88534_dp*bohr/(z1**0.23_dp+z2**0.23_dp)
      x = x1/au
      fac = z1*z2/evolt
      v1 = fac/x1*(0.1818_dp*EXP(-3.2_dp*x)+0.5099_dp*EXP(-0.9423_dp*x)+ &
                   0.2802_dp*EXP(-0.4029_dp*x)+0.02817_dp*EXP(-0.2016_dp*x))
      d1 = fac/x1/au*(-3.2_dp*0.1818_dp*EXP(-3.2_dp*x)-0.9423_dp*0.5099_dp*EXP(-0.9423_dp*x) &
                      -0.4029_dp*0.2802_dp*EXP(-0.4029_dp*x)-0.2016_dp*0.02817_dp*EXP(-0.2016_dp*x)) &
           -fac/x1/x1*(0.1818_dp*EXP(-3.2_dp*x)+0.5099_dp*EXP(-0.9423_dp*x)+ &
                       0.2802_dp*EXP(-0.4029_dp*x)+0.02817_dp*EXP(-0.2016_dp*x))

      dd1 = 2.0_dp*fac/x1**3*(0.1818_dp*EXP(-0.32E1_dp*x) &
                              +0.5099_dp*EXP(-0.9423_dp*x)+0.2802_dp*EXP(-0.4029_dp*x) &
                              +0.2817E-1_dp*EXP(-0.2016_dp*x)) &
            -0.2E1_dp*fac/x1**2/au*(-0.58176_dp*EXP(-0.32E1_dp*x)-0.48047877_dp*EXP(-0.9423_dp*x) &
                                    -0.11289258_dp*EXP(-0.4029_dp*x)-0.5679072E-2_dp*EXP(-0.2016_dp*x)) &
            +fac/x1/au**2*(0.1861632E1_dp*EXP(-0.32E1_dp*x)+ &
                           0.4527551450_dp*EXP(-0.9423_dp*x)+0.4548442048E-1_dp*EXP(-0.4029_dp*x)+ &
                           0.1144900915E-2_dp*EXP(-0.2016_dp*x))

      v2 = 0.0_dp
      d2 = 0.0_dp
      dd2 = 0.0_dp

      CALL compute_polinomial_5th(x1, v1, d1, dd1, x2, v2, d2, dd2, pot%zbl_poly)

   END SUBROUTINE zbl_matching_polinomial

! **************************************************************************************************
!> \brief ...
!> \param r1 ...
!> \param v1 ...
!> \param d1 ...
!> \param dd1 ...
!> \param r2 ...
!> \param v2 ...
!> \param d2 ...
!> \param dd2 ...
!> \param poly ...
! **************************************************************************************************
   SUBROUTINE compute_polinomial_5th(r1, v1, d1, dd1, r2, v2, d2, dd2, poly)

      REAL(KIND=dp)                                      :: r1, v1, d1, dd1, r2, v2, d2, dd2, &
                                                            poly(0:5)

      REAL(KIND=dp)                                      :: a0, a1, a2, a3, a4, a5

!  5th order

      a0 = .5_dp*(2._dp*r1**5*v2-2._dp*v1*r2**5+10._dp*v1*r2**4*r1-20._dp*v1*r1**2*r2**3-r1**2*dd1*r2**5- &
                  r1**4*r2**3*dd1+20._dp*r1**3*r2**2*v2+2._dp*r1**3*r2**4*dd1+r1**3*r2**4*dd2-8._dp*r1**3*r2**3*d2- &
                  2._dp*r1**4*r2**3*dd2+10._dp*r1**4*r2**2*d2-10._dp*r1**4*r2*v2+2._dp*r1*d1*r2**5-10._dp*r1**2*d1*r2**4+ &
                  8._dp*r1**3*d1*r2**3-2._dp*r1**5*r2*d2+r1**5*r2**2*dd2)/ &
           (10.*r2**2*r1**3-5._dp*r2*r1**4-10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      a1 = -.5_dp*(-4._dp*r2**3*r1**3*dd2+24._dp*r2**2*r1**3*d1+4._dp*r2**3*r1**3*dd1+3._dp*r2**4*r1**2*dd2+ &
                   r2**4*r1**2*dd1-2._dp*r2**5*r1*dd1-10._dp*r2**4*r1*d1+10._dp*r1**4*r2*d2- &
                   r1**4*r2**2*dd2-3._dp*r1**4*r2**2*dd1+ &
                   2._dp*r1**5*r2*dd2-24._dp*r2**3*r1**2*d2-16._dp*r2**3*r1**2*d1+ &
                   16._dp*r2**2*r1**3*d2-2._dp*r1**5*d2+2._dp*r2**5*d1- &
                   60._dp*r1**2*r2**2*v1+60._dp*r1**2*r2**2*v2)/ &
           (10._dp*r2**2*r1**3-5._dp*r2*r1**4-10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      a2 = .5_dp*(60._dp*r1**2*r2*v2-60._dp*v1*r1*r2**2-12._dp*r1**2*r2**2*d2-36._dp*r1*d1*r2**3+3._dp*r2**4*r1*dd2- &
                  24._dp*r2**3*r1*d2-4._dp*r2**4*r1*dd1+12._dp*r1**2*r2**2*d1-8._dp*r1**3*r2**2*dd2+24._dp*r1**3*r2*d1+ &
                  4._dp*r1**4*r2*dd2+36._dp*r1**3*r2*d2-3._dp*r1**4*r2*dd1+8._dp*r2**3*r1**2*dd1+60._dp*r2**2*r1*v2- &
                  60._dp*r1**2*v1*r2+r1**5*dd2-r2**5*dd1)/ &
           (10._dp*r2**2*r1**3-5._dp*r2*r1**4-10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      a3 = -.5_dp*(3._dp*r1**4*dd2-r1**4*dd1+8.*r1**3*d1-4.*r1**3*r2*dd1+ &
                   12._dp*r1**3*d2+32._dp*r1**2*r2*d1-8._dp*r1**2*r2**2*dd2- &
                   20._dp*r1**2*v1+8._dp*r1**2*r2**2*dd1+28._dp*r1**2*r2*d2+ &
                   20._dp*r1**2*v2+80._dp*r1*r2*v2-28._dp*r2**2*r1*d1-80._dp*r1*v1*r2- &
                   32._dp*r2**2*r1*d2+4._dp*r1*r2**3*dd2-8._dp*r2**3*d2-12._dp*r2**3*d1+ &
                   r2**4*dd2-3._dp*r2**4*dd1+20._dp*r2**2*v2-20._dp*r2**2*v1)/ &
           (10._dp*r2**2*r1**3-5._dp*r2*r1**4-10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      a4 = .5_dp*(3._dp*r1**3*dd2-2._dp*r1**3*dd1+r1**2*r2*dd1+14.*r1**2*d1-4._dp*r1**2*r2*dd2+ &
                  16._dp*r1**2*d2-2._dp*r1*r2*d2-r1*r2**2*dd2- &
                  30._dp*r1*v1+30.*r1*v2+2._dp*r1*r2*d1+4._dp*r1*r2**2*dd1-16._dp*r2**2*d1+ &
                  2._dp*r2**3*dd2-14._dp*r2**2*d2+30._dp*r2*v2-30._dp*v1*r2- &
                  3._dp*r2**3*dd1)/(10._dp*r2**2*r1**3-5._dp*r2*r1**4- &
                                    10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      a5 = -.5_dp*(6._dp*r1*d1+2._dp*r2*r1*dd1+6._dp*r1*d2-2.*r2*r1*dd2- &
                   r2**2*dd1-r1**2*dd1-12.*v1+12._dp*v2+r1**2*dd2- &
                   6._dp*r2*d1+r2**2*dd2-6._dp*r2*d2)/ &
           (10._dp*r2**2*r1**3-5._dp*r2*r1**4-10._dp*r1**2*r2**3+5._dp*r2**4*r1-r2**5+r1**5)

      poly(0) = a0
      poly(1) = a1
      poly(2) = a2
      poly(3) = a3
      poly(4) = a4
      poly(5) = a5

   END SUBROUTINE compute_polinomial_5th

END MODULE pair_potential_util

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      September 2005 - Introduced the Born-Mayer-Huggins-Fumi-Tosi  Potential (BMHTF)
!>      2006 - Major rewriting of the routines.. Linear scaling setup of splines
!>      2007 - Teodoro Laino - University of Zurich - Multiple potential
!>             Major rewriting nr.2
!> \author CJM
! **************************************************************************************************
MODULE pair_potential

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE fparser,                         ONLY: finalizef,&
                                              initf,&
                                              parsef
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE pair_potential_types,            ONLY: &
        b4_type, bm_type, compare_pot, ea_type, ft_type, ftd_type, gp_type, gw_type, ip_type, &
        list_pot, lj_charmm_type, lj_type, multi_type, nn_type, pair_potential_pp_type, &
        pair_potential_single_type, potential_single_allocation, quip_type, siepmann_type, &
        tersoff_type, wl_type
   USE pair_potential_util,             ONLY: ener_pot,&
                                              ener_zbl,&
                                              zbl_matching_polinomial
   USE physcon,                         ONLY: bohr,&
                                              evolt,&
                                              kjmol
   USE splines_methods,                 ONLY: init_spline,&
                                              init_splinexy,&
                                              potential_s
   USE splines_types,                   ONLY: spline_data_p_type,&
                                              spline_data_type,&
                                              spline_env_create,&
                                              spline_environment_type,&
                                              spline_factor_create,&
                                              spline_factor_release,&
                                              spline_factor_type
   USE string_table,                    ONLY: str2id
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pair_potential'
   REAL(KIND=dp), PARAMETER, PRIVATE    :: MIN_HICUT_VALUE = 1.0E-15_dp, &
                                           DEFAULT_HICUT_VALUE = 1.0E3_dp
   INTEGER, PARAMETER, PRIVATE          :: MAX_POINTS = 2000000

   PUBLIC :: spline_nonbond_control, &
             get_nonbond_storage

CONTAINS

! **************************************************************************************************
!> \brief Initialize genpot
!> \param potparm ...
!> \param ntype ...
!> \par History
!>      Teo 2007.06 - Zurich University
! **************************************************************************************************
   SUBROUTINE init_genpot(potparm, ntype)
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      INTEGER, INTENT(IN)                                :: ntype

      CHARACTER(len=*), PARAMETER :: routineN = 'init_genpot', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, ngp
      TYPE(pair_potential_single_type), POINTER          :: pot

      CALL timeset(routineN, handle)

      NULLIFY (pot)
      ngp = 0
      ! Prescreen for general potential type
      DO i = 1, ntype ! i:  first  atom type
         DO j = 1, i ! j:  second atom type
            pot => potparm%pot(i, j)%pot
            ngp = ngp+COUNT(pot%type == gp_type)
         END DO
      END DO
      CALL initf(ngp)
      ngp = 0
      DO i = 1, ntype ! i:  first  atom type
         DO j = 1, i ! j:  second atom type
            pot => potparm%pot(i, j)%pot
            DO k = 1, SIZE(pot%type)
               IF (pot%type(k) == gp_type) THEN
                  ngp = ngp+1
                  pot%set(k)%gp%myid = ngp
                  CALL parsef(ngp, TRIM(pot%set(k)%gp%potential), pot%set(k)%gp%parameters)
               END IF
            END DO
         END DO
      END DO
      CALL timestop(handle)

   END SUBROUTINE init_genpot

! **************************************************************************************************
!> \brief creates the splines for the potentials
!> \param spline_env ...
!> \param potparm ...
!> \param atomic_kind_set ...
!> \param eps_spline ...
!> \param max_energy ...
!> \param rlow_nb ...
!> \param emax_spline ...
!> \param npoints ...
!> \param iw ...
!> \param iw2 ...
!> \param iw3 ...
!> \param do_zbl ...
!> \param shift_cutoff ...
!> \param nonbonded_type ...
!> \par History
!>      Teo 2006.05 : Improved speed and accuracy. Linear scaling of the setup
! **************************************************************************************************
   SUBROUTINE spline_nonbond_control(spline_env, potparm, atomic_kind_set, &
                                     eps_spline, max_energy, rlow_nb, emax_spline, npoints, iw, iw2, iw3, do_zbl, &
                                     shift_cutoff, nonbonded_type)

      TYPE(spline_environment_type), POINTER             :: spline_env
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      REAL(KIND=dp), INTENT(IN)                          :: eps_spline, max_energy, rlow_nb, &
                                                            emax_spline
      INTEGER, INTENT(IN)                                :: npoints, iw, iw2, iw3
      LOGICAL, INTENT(IN)                                :: do_zbl, shift_cutoff
      CHARACTER(LEN=*), INTENT(IN)                       :: nonbonded_type

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_nonbond_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ip, j, k, n, ncount, &
                                                            npoints_spline, ntype
      LOGICAL                                            :: found_locut
      REAL(KIND=dp)                                      :: energy_cutoff, hicut, hicut0, locut
      TYPE(pair_potential_single_type), POINTER          :: pot

      n = 0
      ncount = 0

      ntype = SIZE(atomic_kind_set)
      CALL timeset(routineN, handle)
      IF (iw3 > 0) THEN
         WRITE (iw3, "(/,T2,A,I0,A,I0,A)") &
            "SPLINE_INFO| Generating ", (ntype*(ntype+1))/2, " splines for "// &
            TRIM(ADJUSTL(nonbonded_type))//" interactions "
         WRITE (iw3, "(T2,A,I0,A)") &
            "             Due to ", ntype, " different atomic kinds"
      END IF
      CALL init_genpot(potparm, ntype)
      ! Real computation of splines
      ip = 0
      DO i = 1, ntype
         DO j = 1, i
            pot => potparm%pot(i, j)%pot
            IF (iw3 > 0 .AND. iw <= 0) THEN
               IF (MOD(i*(i-1)/2+j, MAX(1, (ntype*(ntype+1))/(2*10))) == 0) THEN
                  WRITE (UNIT=iw3, ADVANCE="NO", FMT='(2X,A3,I0)') '...', i*(i-1)/2+j
                  ip = ip+1
                  IF (ip >= 11) THEN
                     WRITE (iw3, *)
                     ip = 0
                  END IF
               END IF
            ENDIF
            ! Setup of Exclusion Types
            pot%no_pp = .TRUE.
            pot%no_mb = .TRUE.
            DO k = 1, SIZE(pot%type)
               SELECT CASE (pot%type (k))
               CASE (lj_type, lj_charmm_type, wl_type, gw_type, ft_type, ftd_type, ip_type, &
                     b4_type, bm_type, gp_type, ea_type, quip_type)
                  pot%no_pp = .FALSE.
               CASE (tersoff_type)
                  pot%no_mb = .FALSE.
               CASE (siepmann_type)
                  pot%no_mb = .FALSE.
               CASE (nn_type)
                  ! Do nothing..
               CASE DEFAULT
                  ! Never reach this point
                  CALL cp__b("pair_potential.F",197,"")
               END SELECT
               ! Special case for EAM
               SELECT CASE (pot%type (k))
               CASE (ea_type, quip_type)
                  pot%no_mb = .FALSE.
               END SELECT
            END DO

            ! Starting SetUp of splines
            IF (.NOT. pot%undef) CYCLE
            ncount = ncount+1
            n = spline_env%spltab(i, j)
            locut = rlow_nb
            hicut0 = SQRT(pot%rcutsq)
            IF (ABS(hicut0) <= MIN_HICUT_VALUE) hicut0 = DEFAULT_HICUT_VALUE
            hicut = hicut0/SQRT(pot%spl_f%rcutsq_f)

            energy_cutoff = pot%spl_f%cutoff

            ! Find the real locut according emax_spline
            CALL get_spline_cutoff(hicut, locut, found_locut, pot, do_zbl, &
                                   energy_cutoff, emax_spline)
            locut = MAX(locut*SQRT(pot%spl_f%rcutsq_f), rlow_nb)

            ! Real Generation of the Spline
            npoints_spline = npoints
            CALL generate_spline_low(spline_env%spl_pp(n)%spl_p, npoints_spline, locut, &
                                     hicut, eps_spline, iw, iw2, i, j, n, ncount, max_energy, pot, &
                                     energy_cutoff, found_locut, do_zbl, atomic_kind_set, &
                                     nonbonded_type)

            pot%undef = .FALSE.
            ! Unique Spline working only for a pure LJ potential..
            IF (SIZE(pot%type) == 1) THEN
               IF (ANY(potential_single_allocation == pot%type(1))) THEN
                  ! Restoring the proper values for the generating spline pot
                  IF ((pot%type(1) == lj_type) .OR. (pot%type(1) == lj_charmm_type)) THEN
                     pot%set(1)%lj%sigma6 = pot%set(1)%lj%sigma6*pot%spl_f%rscale(1)**3
                     pot%set(1)%lj%sigma12 = pot%set(1)%lj%sigma6**2
                     pot%set(1)%lj%epsilon = pot%set(1)%lj%epsilon*pot%spl_f%fscale(1)
                  END IF
               END IF
            END IF
            ! Correct Cutoff...
            IF (shift_cutoff) THEN
               pot%spl_f%cutoff = pot%spl_f%cutoff*pot%spl_f%fscale(1)- &
                                  ener_pot(pot, hicut0, 0.0_dp)
            END IF
         END DO
      END DO
      CALL finalizef()

      IF (iw > 0) THEN
         WRITE (iw, '(/,T2,A,I0)') &
            "SPLINE_INFO| Number of pair potential splines allocated:   ", MAXVAL(spline_env%spltab)
      END IF
      IF (iw3 > 0) THEN
         WRITE (iw3, '(/,T2,A,I0,/,T2,A)') &
            "SPLINE_INFO| Number of unique splines computed:            ", MAXVAL(spline_env%spltab), &
            "SPLINE_INFO| Done"
      END IF

      CALL timestop(handle)

   END SUBROUTINE spline_nonbond_control

! **************************************************************************************************
!> \brief Finds the cutoff for the generation of the spline
!>      In a two pass approach, first with low resolution, refine in a second iteration
!> \param hicut ...
!> \param locut ...
!> \param found_locut ...
!> \param pot ...
!> \param do_zbl ...
!> \param energy_cutoff ...
!> \param emax_spline ...
!> \par History
!>      Splitting in order to make some season cleaning..
!> \author Teodoro Laino [tlaino] 2007.06
! **************************************************************************************************
   SUBROUTINE get_spline_cutoff(hicut, locut, found_locut, pot, do_zbl, &
                                energy_cutoff, emax_spline)

      REAL(KIND=dp), INTENT(IN)                          :: hicut
      REAL(KIND=dp), INTENT(INOUT)                       :: locut
      LOGICAL, INTENT(OUT)                               :: found_locut
      TYPE(pair_potential_single_type), OPTIONAL, &
         POINTER                                         :: pot
      LOGICAL, INTENT(IN)                                :: do_zbl
      REAL(KIND=dp), INTENT(IN)                          :: energy_cutoff, emax_spline

      CHARACTER(len=*), PARAMETER :: routineN = 'get_spline_cutoff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilevel, jx
      REAL(KIND=dp)                                      :: dx2, e, locut_found, x

      dx2 = (hicut-locut)
      x = hicut
      locut_found = locut
      found_locut = .FALSE.
      DO ilevel = 1, 2
         dx2 = dx2/100.0_dp
         DO jx = 1, 100
            e = ener_pot(pot, x, energy_cutoff)
            IF (do_zbl) THEN
               e = e+ener_zbl(pot, x)
            END IF
            IF (ABS(e) > emax_spline) THEN
               locut_found = x
               found_locut = .TRUE.
               EXIT
            END IF
            x = x-dx2
         END DO
         x = x+dx2
      ENDDO
      locut = locut_found

   END SUBROUTINE get_spline_cutoff

! **************************************************************************************************
!> \brief Real Generation of spline..
!> \param spl_p ...
!> \param npoints ...
!> \param locut ...
!> \param hicut ...
!> \param eps_spline ...
!> \param iw ...
!> \param iw2 ...
!> \param i ...
!> \param j ...
!> \param n ...
!> \param ncount ...
!> \param max_energy ...
!> \param pot ...
!> \param energy_cutoff ...
!> \param found_locut ...
!> \param do_zbl ...
!> \param atomic_kind_set ...
!> \param nonbonded_type ...
!> \par History
!>      Splitting in order to make some season cleaning..
!> \author Teodoro Laino [tlaino] 2007.06
! **************************************************************************************************
   SUBROUTINE generate_spline_low(spl_p, npoints, locut, hicut, eps_spline, &
                                  iw, iw2, i, j, n, ncount, max_energy, pot, energy_cutoff, &
                                  found_locut, do_zbl, atomic_kind_set, nonbonded_type)

      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p
      INTEGER, INTENT(INOUT)                             :: npoints
      REAL(KIND=dp), INTENT(IN)                          :: locut, hicut, eps_spline
      INTEGER, INTENT(IN)                                :: iw, iw2, i, j, n, ncount
      REAL(KIND=dp), INTENT(IN)                          :: max_energy
      TYPE(pair_potential_single_type), POINTER          :: pot
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: energy_cutoff
      LOGICAL, INTENT(IN)                                :: found_locut, do_zbl
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      CHARACTER(LEN=*), INTENT(IN)                       :: nonbonded_type

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_spline_low', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2*default_string_length)             :: message, tmp
      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: ix, jx, mfac, nppa, nx, unit_number
      LOGICAL                                            :: fixed_spline_points
      REAL(KIND=dp)                                      :: df, dg, dh, diffmax, dx, dx2, e, &
                                                            e_spline, f, g, h, r, rcut, x, x2, &
                                                            xdum, xdum1, xsav
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(spline_data_type), POINTER                    :: spline_data
      TYPE(spline_factor_type), POINTER                  :: spl_f

      NULLIFY (logger, spl_f)
      logger => cp_get_default_logger()

      CALL spline_factor_create(spl_f)
      mfac = 5
      IF (npoints > 0) THEN
         fixed_spline_points = .TRUE.
      ELSE
         fixed_spline_points = .FALSE.
         npoints = 20
         IF (.NOT. found_locut) npoints = 2
      ENDIF
      spline_data => spl_p(1)%spline_data
      DO WHILE (.TRUE.)
         CALL init_splinexy(spline_data, npoints+1)
         dx2 = (1.0_dp/locut**2-1.0_dp/hicut**2)/REAL(npoints, KIND=dp)
         x2 = 1.0_dp/hicut**2
         spline_data%x1 = x2
         DO jx = 1, npoints+1
            ! jx: loop over 1/distance**2
            x = SQRT(1.0_dp/x2)
            e = ener_pot(pot, x, energy_cutoff)
            IF (do_zbl) THEN
               e = e+ener_zbl(pot, x)
            END IF
            spline_data%y(jx) = e
            x2 = x2+dx2
         END DO
         CALL init_spline(spline_data, dx=dx2)
         ! This is the check for required accuracy on spline setup
         dx2 = (hicut-locut)/REAL(mfac*npoints+1, KIND=dp)
         x2 = locut+dx2
         diffmax = -1.0_dp
         xsav = hicut
         ! if a fixed number of points is requested, no check on its error
         IF (fixed_spline_points) EXIT
         DO jx = 1, mfac*npoints
            x = x2
            e = ener_pot(pot, x, energy_cutoff)
            IF (do_zbl) THEN
               e = e+ener_zbl(pot, x)
            END IF
            IF (ABS(e) < max_energy) THEN
               xdum1 = ABS(e-potential_s(spl_p, x*x, xdum, spl_f, logger))
               diffmax = MAX(diffmax, xdum1)
               xsav = MIN(x, xsav)
            END IF
            x2 = x2+dx2
            IF (x2 > hicut) EXIT
         END DO
         IF (npoints > MAX_POINTS) THEN
            WRITE (message, '(A,I8,A,G12.6,A)') "SPLINE_INFO| Number of points: ", npoints, &
               " obtained accuracy ", diffmax, ". MM SPLINE: no convergence on required"// &
               " accuracy (adjust EPS_SPLINE and rerun)"
            CALL cp_abort(cp__l("pair_potential.F",426), TRIM(message))
         END IF
         ! accuracy is poor or we have found no points below max_energy, refine mesh
         IF (diffmax > eps_spline .OR. diffmax < 0.0_dp) THEN
            npoints = CEILING(1.2_dp*REAL(npoints, KIND=dp))
         ELSE
            EXIT
         END IF
      END DO
      ! Print spline info to STDOUT if requested
      IF (iw > 0) THEN
         WRITE (UNIT=iw, &
                FMT="(/,A,I0,/,A,I0,/,A,I0,1X,I0,/,A,/,A,I0,2(/,A,ES13.6),2(/,A,2ES13.6))") &
            " SPLINE_INFO| Spline number:                                ", ncount, &
            " SPLINE_INFO| Unique spline number:                         ", n, &
            " SPLINE_INFO| Atomic kind numbers:                          ", i, j, &
            " SPLINE_INFO| Atomic kind names:                            "//TRIM(ADJUSTL(atomic_kind_set(i)%name))//" "// &
            TRIM(ADJUSTL(atomic_kind_set(j)%name)), &
            " SPLINE_INFO| Number of spline points:                      ", npoints, &
            " SPLINE_INFO| Requested accuracy [Hartree]:                ", eps_spline, &
            " SPLINE_INFO| Achieved accuracy [Hartree]:                 ", diffmax, &
            " SPLINE_INFO| Spline range [bohr]:                         ", locut, hicut, &
            " SPLINE_INFO| Spline range used to achieve accuracy [bohr]:", xsav, hicut
         dx2 = (hicut-locut)/REAL(npoints+1, KIND=dp)
         x = locut+dx2
         WRITE (UNIT=iw, FMT='(A,ES17.9)') &
            " SPLINE_INFO| Spline value at RMIN [Hartree]:             ", potential_s(spl_p, x*x, xdum, spl_f, logger), &
            " SPLINE_INFO| Spline value at RMAX [Hartree]:             ", potential_s(spl_p, hicut*hicut, xdum, spl_f, logger), &
            " SPLINE_INFO| Non-bonded energy cutoff [Hartree]:         ", energy_cutoff
      END IF
      ! Print spline data on file if requested
      IF (iw2 > 0) THEN
         ! Set increment to 200 points per Angstrom
         nppa = 200
         dx = bohr/REAL(nppa, KIND=dp)
         nx = NINT(hicut/dx)
         file_name = ""
         tmp = ADJUSTL(cp_to_string(n))
         WRITE (UNIT=file_name, FMT="(A,I0,A)") &
            TRIM(ADJUSTL(nonbonded_type))//"_SPLINE_"//TRIM(tmp)//"_"// &
            TRIM(ADJUSTL(atomic_kind_set(i)%name))//"_"// &
            TRIM(ADJUSTL(atomic_kind_set(j)%name))
         CALL open_file(file_name=file_name, &
                        file_status="UNKNOWN", &
                        file_form="FORMATTED", &
                        file_action="WRITE", &
                        unit_number=unit_number)
         WRITE (UNIT=unit_number, &
                FMT="(2(A,I0,/),A,I0,1X,I0,/,A,/,A,I0,2(/,A,ES13.6),2(/,A,2ES13.6),/,A,ES13.6,/,A,I0,A,/,A)") &
            "# Spline number:                                    ", ncount, &
            "# Unique spline number:                             ", n, &
            "# Atomic kind numbers:                              ", i, j, &
            "# Atomic kind names:                                "//TRIM(ADJUSTL(atomic_kind_set(i)%name))//" "// &
            TRIM(ADJUSTL(atomic_kind_set(j)%name)), &
            "# Number of spline points:                          ", npoints, &
            "# Requested accuracy [eV]:                         ", eps_spline*evolt, &
            "# Achieved accuracy [eV]:                          ", diffmax*evolt, &
            "# Spline range [Angstrom]:                         ", locut/bohr, hicut/bohr, &
            "# Spline range used to achieve accuracy [Angstrom]:", xsav/bohr, hicut/bohr, &
            "# Non-bonded energy cutoff [eV]:                   ", energy_cutoff*evolt, &
            "# Test spline using ", nppa, " points per Angstrom:", &
            "#     Abscissa [Angstrom]              Energy [eV]      Splined energy [eV] Derivative [eV/Angstrom]"// &
            "      |Energy error| [eV]"
         x = 0.0_dp
         DO jx = 0, nx
            IF (x > hicut) x = hicut
            IF (x > locut) THEN
               e = ener_pot(pot, x, energy_cutoff)
               IF (do_zbl) e = e+ener_zbl(pot, x)
               e_spline = potential_s(spl_p, x*x, xdum, spl_f, logger)
               WRITE (UNIT=unit_number, FMT="(5ES25.12)") &
                  x/bohr, e*evolt, e_spline*evolt, -bohr*x*xdum*evolt, ABS((e-e_spline)*evolt)
            END IF
            x = x+dx
         END DO
         CALL close_file(unit_number=unit_number)
         !MK Write table.xvf for GROMACS 4.5.5
         WRITE (UNIT=file_name, FMT="(A,I0,A)") &
            "table_"// &
            TRIM(ADJUSTL(atomic_kind_set(i)%name))//"_"// &
            TRIM(ADJUSTL(atomic_kind_set(j)%name))//".xvg"
         CALL open_file(file_name=file_name, &
                        file_status="UNKNOWN", &
                        file_form="FORMATTED", &
                        file_action="WRITE", &
                        unit_number=unit_number)
         ! Recommended increment for dp is 0.0005 nm = 0.005 Angstrom
         ! which are 200 points/Angstrom
         rcut = 0.1_dp*hicut/bohr
         x = 0.0_dp
         DO jx = 0, nx
            IF (x > hicut) x = hicut
            r = 0.1_dp*x/bohr ! Convert bohr to nm
            IF (x <= locut) THEN
               WRITE (UNIT=unit_number, FMT="(7ES25.12)") &
                  r, (0.0_dp, ix=1, 6)
            ELSE
               e_spline = potential_s(spl_p, x*x, xdum, spl_f, logger)
               f = 1.0_dp/r
               df = -1.0_dp/r**2
               g = -1.0_dp/r**6+1.0_dp/rcut**6
               dg = 6.0_dp/r**7
               h = e_spline*kjmol
               dh = -10.0_dp*bohr*x*xdum*kjmol
               WRITE (UNIT=unit_number, FMT="(7ES25.12)") &
                  r, f, -df, & ! r, f(r), -f'(r) => probably not used
                  g, -dg, & !    g(r), -g'(r) => not used, if C = 0
                  h, -dh !    h(r), -h'(r) => used, if A = 1
            END IF
            x = x+dx
         END DO
         CALL close_file(unit_number=unit_number)
      END IF

      CALL spline_factor_release(spl_f)

   END SUBROUTINE generate_spline_low

! **************************************************************************************************
!> \brief Prescreening of the effective bonds evaluations. linear scaling algorithm
!> \param spline_env ...
!> \param potparm ...
!> \param atomic_kind_set ...
!> \param do_zbl ...
!> \param shift_cutoff ...
!> \author Teodoro Laino [tlaino] 2006.05
! **************************************************************************************************
   SUBROUTINE get_nonbond_storage(spline_env, potparm, atomic_kind_set, do_zbl, &
                                  shift_cutoff)

      TYPE(spline_environment_type), POINTER             :: spline_env
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      LOGICAL, INTENT(IN)                                :: do_zbl, shift_cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'get_nonbond_storage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, idim, iend, istart, j, k, &
                                                            locij, n, ndim, nk, ntype, nunique, &
                                                            nvar, pot_target, tmpij(2), tmpij0(2)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: Iwork1, Iwork2, my_index
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: tmp_index
      LOGICAL                                            :: at_least_one, check
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Cwork, Rwork, wtmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pot_par

      CALL timeset(routineN, handle)

      ntype = SIZE(atomic_kind_set)
      DO i = 1, ntype
         DO j = 1, i
            potparm%pot(i, j)%pot%undef = .FALSE.
         END DO
      END DO
      ALLOCATE (tmp_index(ntype, ntype))
      !
      nunique = 0
      tmp_index = HUGE(0)
      DO pot_target = MINVAL(list_pot), MAXVAL(list_pot)
         ndim = 0
         DO i = 1, ntype
            DO j = 1, i
               IF (SIZE(potparm%pot(i, j)%pot%type) /= 1) CYCLE
               IF (potparm%pot(i, j)%pot%type(1) == pot_target) THEN
                  tmp_index(i, j) = 1
                  tmp_index(j, i) = 1
                  ndim = ndim+1
               END IF
            END DO
         END DO
         IF (ndim == 0) CYCLE ! No potential of this kind found
         nvar = 0
         SELECT CASE (pot_target)
         CASE (lj_type, lj_charmm_type)
            nvar = 3+nvar
         CASE (wl_type)
            nvar = 3+nvar
         CASE (gw_type)
            nvar = 5+nvar
         CASE (ea_type)
            nvar = 4+nvar
         CASE (quip_type)
            nvar = 1+nvar
         CASE (ft_type)
            nvar = 4+nvar
         CASE (ftd_type)
            nvar = 5+nvar
         CASE (ip_type)
            nvar = 3+nvar
         CASE (b4_type)
            nvar = 6+nvar
         CASE (bm_type)
            nvar = 9+nvar
         CASE (gp_type)
            nvar = 2+nvar
         CASE (tersoff_type)
            nvar = 13+nvar
         CASE (siepmann_type)
            nvar = 5+nvar
         CASE (nn_type)
            nvar = nvar
         CASE DEFAULT
            CALL cp__b("pair_potential.F",629,"")
         END SELECT
         ! Setup a table of the indexes..
         ALLOCATE (my_index(ndim))
         n = 0
         nk = 0
         DO i = 1, ntype
            DO j = 1, i
               n = n+1
               IF (SIZE(potparm%pot(i, j)%pot%type) /= 1) CYCLE
               IF (potparm%pot(i, j)%pot%type(1) == pot_target) THEN
                  nk = nk+1
                  my_index(nk) = n
               END IF
            END DO
         END DO
         IF (nvar /= 0) THEN
            ALLOCATE (pot_par(ndim, nvar))
            n = 0
            nk = 0
            DO i = 1, ntype
               DO j = 1, i
                  n = n+1
                  IF (SIZE(potparm%pot(i, j)%pot%type) /= 1) CYCLE
                  IF (potparm%pot(i, j)%pot%type(1) == pot_target) THEN
                     nk = nk+1
                     my_index(nk) = n
                     SELECT CASE (pot_target)
                     CASE (lj_type, lj_charmm_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%lj%epsilon
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%lj%sigma6
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%lj%sigma12
                     CASE (gp_type)
                        pot_par(nk, 1) = str2id(potparm%pot(i, j)%pot%set(1)%gp%potential)
                        pot_par(nk, 2) = str2id(potparm%pot(i, j)%pot%set(1)%gp%variables)
                     CASE (wl_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%willis%a
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%willis%b
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%willis%c
                     CASE (gw_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%goodwin%vr0
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%goodwin%m
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%goodwin%mc
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%goodwin%d
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%goodwin%dc
                     CASE (ea_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%eam%drar
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%eam%drhoar
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%eam%acutal
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%eam%npoints
                     CASE (quip_type)
                        pot_par(nk, 1) = str2id( &
                                         TRIM(potparm%pot(i, j)%pot%set(1)%quip%quip_file_name)// &
                                         TRIM(potparm%pot(i, j)%pot%set(1)%quip%init_args)// &
                                         TRIM(potparm%pot(i, j)%pot%set(1)%quip%calc_args))
                     CASE (ft_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%ft%A
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%ft%B
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%ft%C
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%ft%D
                     CASE (ftd_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%ftd%A
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%ftd%B
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%ftd%C
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%ftd%D
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%ftd%BD
                     CASE (ip_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%ipbv%rcore
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%ipbv%m
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%ipbv%b
                     CASE (b4_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%buck4r%a
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%buck4r%b
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%buck4r%c
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%buck4r%r1
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%buck4r%r2
                        pot_par(nk, 6) = potparm%pot(i, j)%pot%set(1)%buck4r%r3
                     CASE (bm_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%buckmo%f0
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%buckmo%a1
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%buckmo%a2
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%buckmo%b1
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%buckmo%b2
                        pot_par(nk, 6) = potparm%pot(i, j)%pot%set(1)%buckmo%c
                        pot_par(nk, 7) = potparm%pot(i, j)%pot%set(1)%buckmo%d
                        pot_par(nk, 8) = potparm%pot(i, j)%pot%set(1)%buckmo%r0
                        pot_par(nk, 9) = potparm%pot(i, j)%pot%set(1)%buckmo%beta
                     CASE (tersoff_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%tersoff%A
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%tersoff%B
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%tersoff%lambda1
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%tersoff%lambda2
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%tersoff%alpha
                        pot_par(nk, 6) = potparm%pot(i, j)%pot%set(1)%tersoff%beta
                        pot_par(nk, 7) = potparm%pot(i, j)%pot%set(1)%tersoff%n
                        pot_par(nk, 8) = potparm%pot(i, j)%pot%set(1)%tersoff%c
                        pot_par(nk, 9) = potparm%pot(i, j)%pot%set(1)%tersoff%d
                        pot_par(nk, 10) = potparm%pot(i, j)%pot%set(1)%tersoff%h
                        pot_par(nk, 11) = potparm%pot(i, j)%pot%set(1)%tersoff%lambda3
                        pot_par(nk, 12) = potparm%pot(i, j)%pot%set(1)%tersoff%bigR
                        pot_par(nk, 13) = potparm%pot(i, j)%pot%set(1)%tersoff%bigD
                     CASE (siepmann_type)
                        pot_par(nk, 1) = potparm%pot(i, j)%pot%set(1)%siepmann%B
                        pot_par(nk, 2) = potparm%pot(i, j)%pot%set(1)%siepmann%D
                        pot_par(nk, 3) = potparm%pot(i, j)%pot%set(1)%siepmann%E
                        pot_par(nk, 4) = potparm%pot(i, j)%pot%set(1)%siepmann%F
                        pot_par(nk, 5) = potparm%pot(i, j)%pot%set(1)%siepmann%beta
                     CASE (nn_type)
                        ! no checks
                     CASE DEFAULT
                        CALL cp__b("pair_potential.F",739,"")
                     END SELECT
                     IF (ANY(potential_single_allocation == pot_target)) THEN
                        pot_par(nk, :) = REAL(pot_target, KIND=dp)
                     END IF
                  END IF
               END DO
            END DO
            ! Main Sorting Loop
            ALLOCATE (Rwork(ndim))
            ALLOCATE (Iwork1(ndim))
            ALLOCATE (Iwork2(ndim))
            ALLOCATE (wtmp(nvar))
            CALL sort(pot_par(:, 1), ndim, Iwork1)
            ! Sort all the other components of the potential
            DO k = 2, nvar
               Rwork(:) = pot_par(:, k)
               DO i = 1, ndim
                  pot_par(i, k) = Rwork(Iwork1(i))
               END DO
            END DO
            Iwork2(:) = my_index
            DO i = 1, ndim
               my_index(i) = Iwork2(Iwork1(i))
            END DO
            ! Iterative sorting
            DO k = 2, nvar
               wtmp(1:k-1) = pot_par(1, 1:k-1)
               istart = 1
               at_least_one = .FALSE.
               DO j = 1, ndim
                  Rwork(j) = pot_par(j, k)
                  IF (ALL(pot_par(j, 1:k-1) == wtmp(1:k-1))) CYCLE
                  iend = j-1
                  wtmp(1:k-1) = pot_par(j, 1:k-1)
                  ! If the ordered array has no two same consecutive elements
                  ! does not make any sense to proceed ordering the others
                  ! related parameters..
                  idim = iend-istart+1
                  CALL sort(Rwork(istart:iend), idim, Iwork1(istart:iend))
                  Iwork1(istart:iend) = Iwork1(istart:iend)-1+istart
                  IF (idim /= 1) at_least_one = .TRUE.
                  istart = j
               END DO
               iend = ndim
               idim = iend-istart+1
               CALL sort(Rwork(istart:iend), idim, Iwork1(istart:iend))
               Iwork1(istart:iend) = Iwork1(istart:iend)-1+istart
               IF (idim /= 1) at_least_one = .TRUE.
               pot_par(:, k) = Rwork
               IF (.NOT. at_least_one) EXIT
               ! Sort other components
               DO j = k+1, nvar
                  Rwork(:) = pot_par(:, j)
                  DO i = 1, ndim
                     pot_par(i, j) = Rwork(Iwork1(i))
                  END DO
               END DO
               Iwork2(:) = my_index
               DO i = 1, ndim
                  my_index(i) = Iwork2(Iwork1(i))
               END DO
            END DO
            DEALLOCATE (wtmp)
            DEALLOCATE (Iwork1)
            DEALLOCATE (Iwork2)
            DEALLOCATE (Rwork)
            !
            ! Let's determine the number of unique potentials and tag them
            !
            ALLOCATE (Cwork(nvar))
            Cwork(:) = pot_par(1, :)
            locij = my_index(1)
            CALL get_indexes(locij, ntype, tmpij0)
            istart = 1
            DO j = 1, ndim
               ! Special cases for EAM and IPBV
               locij = my_index(j)
               CALL get_indexes(locij, ntype, tmpij)
               SELECT CASE (pot_target)
                  !NB should do something about QUIP here?
               CASE (ea_type, ip_type)
                  ! check the array components
                  CALL compare_pot(potparm%pot(tmpij(1), tmpij(2))%pot, &
                                   potparm%pot(tmpij0(1), tmpij0(2))%pot, &
                                   check)
               CASE (gp_type)
                  check = .TRUE.
                  IF (ASSOCIATED(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%parameters) .AND. &
                      ASSOCIATED(potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%parameters)) THEN
                     IF (SIZE(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%parameters) == &
                         SIZE(potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%parameters)) THEN
                        IF (ANY(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%parameters /= &
                                potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%parameters)) check = .FALSE.
                     END IF
                  END IF
                  IF (ASSOCIATED(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%values) .AND. &
                      ASSOCIATED(potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%values)) THEN
                     IF (SIZE(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%values) == &
                         SIZE(potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%values)) THEN
                        IF (ANY(potparm%pot(tmpij(1), tmpij(2))%pot%set(1)%gp%values /= &
                                potparm%pot(tmpij0(1), tmpij0(2))%pot%set(1)%gp%values)) check = .FALSE.
                     END IF
                  END IF
               CASE default
                  check = .TRUE.
               END SELECT
               IF (ALL(Cwork == pot_par(j, :)) .AND. check) CYCLE
               Cwork(:) = pot_par(j, :)
               nunique = nunique+1
               iend = j-1
               CALL set_potparm_index(potparm, my_index(istart:iend), pot_target, &
                                      ntype, tmpij, atomic_kind_set, shift_cutoff, do_zbl)
               !
               DO i = istart, iend
                  locij = my_index(i)
                  CALL get_indexes(locij, ntype, tmpij)
                  tmp_index(tmpij(1), tmpij(2)) = nunique
                  tmp_index(tmpij(2), tmpij(1)) = nunique
               END DO
               istart = j
               locij = my_index(j)
               CALL get_indexes(locij, ntype, tmpij0)
            END DO
            nunique = nunique+1
            iend = ndim
            CALL set_potparm_index(potparm, my_index(istart:iend), pot_target, &
                                   ntype, tmpij, atomic_kind_set, shift_cutoff, do_zbl)
            DO i = istart, iend
               locij = my_index(i)
               CALL get_indexes(locij, ntype, tmpij)
               tmp_index(tmpij(1), tmpij(2)) = nunique
               tmp_index(tmpij(2), tmpij(1)) = nunique
            END DO
            DEALLOCATE (Cwork)
            DEALLOCATE (pot_par)
         ELSE
            nunique = nunique+1
            CALL set_potparm_index(potparm, my_index, pot_target, ntype, tmpij, &
                                   atomic_kind_set, shift_cutoff, do_zbl)
         END IF
         DEALLOCATE (my_index)
      END DO
      ! Multiple defined potential
      n = 0
      DO i = 1, ntype
         DO j = 1, i
            n = n+1
            IF (SIZE(potparm%pot(i, j)%pot%type) == 1) CYCLE
            nunique = nunique+1
            tmp_index(i, j) = nunique
            tmp_index(j, i) = nunique
            !
            CALL set_potparm_index(potparm, (/n/), multi_type, ntype, tmpij, &
                                   atomic_kind_set, shift_cutoff, do_zbl)
         END DO
      END DO
      ! Concluding the postprocess..
      CALL spline_env_create(spline_env, ntype, nunique)
      spline_env%spltab = tmp_index
      DEALLOCATE (tmp_index)
      CALL timestop(handle)
   END SUBROUTINE get_nonbond_storage

! **************************************************************************************************
!> \brief Trivial for non LJ potential.. gives back in the case of LJ
!>      the potparm with the smallest sigma..
!> \param potparm ...
!> \param my_index ...
!> \param pot_target ...
!> \param ntype ...
!> \param tmpij_out ...
!> \param atomic_kind_set ...
!> \param shift_cutoff ...
!> \param do_zbl ...
!> \author Teodoro Laino [tlaino] 2007.06
! **************************************************************************************************
   SUBROUTINE set_potparm_index(potparm, my_index, pot_target, ntype, tmpij_out, &
                                atomic_kind_set, shift_cutoff, do_zbl)

      TYPE(pair_potential_pp_type), POINTER              :: potparm
      INTEGER, INTENT(IN)                                :: my_index(:), pot_target, ntype
      INTEGER, INTENT(OUT)                               :: tmpij_out(2)
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      LOGICAL, INTENT(IN)                                :: shift_cutoff, do_zbl

      CHARACTER(len=*), PARAMETER :: routineN = 'set_potparm_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, min_val, nvalues, tmpij(2), &
                                                            value, zi, zj
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: wrk
      LOGICAL                                            :: check
      REAL(KIND=dp)                                      :: hicut0, l_epsilon, l_sigma6, m_epsilon, &
                                                            m_sigma6, min_sigma6, rcovi, rcovj
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: sigma6
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(pair_potential_single_type), POINTER          :: pot, pot_ref

      CALL timeset(routineN, handle)

      NULLIFY (pot, pot_ref)
      nvalues = SIZE(my_index)
      IF ((pot_target == lj_type) .OR. (pot_target == lj_charmm_type)) THEN
         ALLOCATE (sigma6(nvalues))
         ALLOCATE (wrk(nvalues))
         min_sigma6 = HUGE(0.0_dp)
         m_epsilon = -HUGE(0.0_dp)
         DO i = 1, nvalues
            value = my_index(i)
            CALL get_indexes(value, ntype, tmpij)
            pot => potparm%pot(tmpij(1), tmpij(2))%pot
            ! Preliminary check..
            check = SIZE(pot%type) == 1
            IF(.NOT.(check))CALL cp__a("pair_potential.F",953)

            sigma6(i) = pot%set(1)%lj%sigma6
            l_epsilon = pot%set(1)%lj%epsilon
            IF (sigma6(i) /= 0.0_dp) min_sigma6 = MIN(min_sigma6, sigma6(i))
            IF (sigma6(i) == 0.0_dp) sigma6(i) = -HUGE(0.0_dp)
            IF (l_epsilon /= 0.0_dp) m_epsilon = MAX(m_epsilon, l_epsilon)
         END DO
         CALL sort(sigma6, nvalues, wrk)
         min_val = my_index(wrk(nvalues))
         m_sigma6 = sigma6(nvalues)
         ! In case there are only zeros.. let's consider them properly..
         IF (m_sigma6 == -HUGE(0.0_dp)) m_sigma6 = 1.0_dp
         IF (m_epsilon == -HUGE(0.0_dp)) m_epsilon = 0.0_dp
         IF (min_sigma6 == HUGE(0.0_dp)) min_sigma6 = 0.0_dp
         DEALLOCATE (sigma6)
         DEALLOCATE (wrk)
      ELSE
         min_val = MINVAL(my_index(:))
      END IF
      CALL get_indexes(min_val, ntype, tmpij)
      tmpij_out = tmpij
      pot => potparm%pot(tmpij(1), tmpij(2))%pot
      pot%undef = .TRUE.
      IF (shift_cutoff) THEN
         hicut0 = SQRT(pot%rcutsq)
         IF (ABS(hicut0) <= MIN_HICUT_VALUE) hicut0 = DEFAULT_HICUT_VALUE
      END IF
      CALL init_genpot(potparm, ntype)

      DO i = 1, nvalues
         value = my_index(i)
         CALL get_indexes(value, ntype, tmpij)
         pot => potparm%pot(tmpij(1), tmpij(2))%pot
         CALL spline_factor_create(pot%spl_f)
         pot%spl_f%rcutsq_f = 1.0_dp
         pot%spl_f%rscale = 1.0_dp
         pot%spl_f%fscale = 1.0_dp
      ENDDO

      IF (ANY(potential_single_allocation == pot_target)) THEN
         DO i = 1, nvalues
            value = my_index(i)
            CALL get_indexes(value, ntype, tmpij)
            pot => potparm%pot(tmpij(1), tmpij(2))%pot

            check = SIZE(pot%type) == 1
            IF(.NOT.(check))CALL cp__a("pair_potential.F",1000)
            ! Undef potential.. this will be used to compute the splines..
            IF ((pot_target == lj_type) .OR. (pot_target == lj_charmm_type)) THEN
               l_sigma6 = pot%set(1)%lj%sigma6
               l_epsilon = pot%set(1)%lj%epsilon
               ! Undef potential.. this will be used to compute the splines..
               IF (pot%undef) THEN
                  pot%set(1)%lj%sigma6 = m_sigma6
                  pot%set(1)%lj%sigma12 = m_sigma6**2
                  pot%set(1)%lj%epsilon = m_epsilon
               END IF
               pot%spl_f%rscale(1) = 1.0_dp
               pot%spl_f%fscale(1) = 0.0_dp
               IF (l_sigma6*l_epsilon /= 0.0_dp) THEN
                  pot%spl_f%rcutsq_f = (min_sigma6/m_sigma6)**(1.0_dp/3.0_dp)
                  pot%spl_f%rscale(1) = (l_sigma6/m_sigma6)**(1.0_dp/3.0_dp)
                  pot%spl_f%fscale(1) = l_epsilon/m_epsilon
               END IF
            END IF
         ENDDO
      END IF

      DO i = 1, nvalues
         value = my_index(i)
         CALL get_indexes(value, ntype, tmpij)
         pot => potparm%pot(tmpij(1), tmpij(2))%pot

         IF (do_zbl) THEN
            atomic_kind => atomic_kind_set(tmpij(1))
            CALL get_atomic_kind(atomic_kind, rcov=rcovi, z=zi)
            atomic_kind => atomic_kind_set(tmpij(2))
            CALL get_atomic_kind(atomic_kind, rcov=rcovj, z=zj)
            CALL zbl_matching_polinomial(pot, rcovi, rcovj, REAL(zi, KIND=dp), &
                                         REAL(zj, KIND=dp))
         END IF
         ! Derivative factors
         pot%spl_f%dscale = pot%spl_f%fscale/pot%spl_f%rscale
         ! Cutoff for the potentials on splines
         IF (shift_cutoff) THEN
            ! Cutoff NonBonded
            pot%spl_f%cutoff = ener_pot(pot, hicut0, 0.0_dp)
         END IF
      END DO

      ! Handle the cutoff
      IF (shift_cutoff) THEN
         pot_ref => potparm%pot(tmpij_out(1), tmpij_out(2))%pot
         DO i = 1, nvalues
            value = my_index(i)
            CALL get_indexes(value, ntype, tmpij)
            pot => potparm%pot(tmpij(1), tmpij(2))%pot
            IF (value == min_val) CYCLE
            ! Cutoff NonBonded
            pot%spl_f%cutoff = pot_ref%spl_f%cutoff*pot%spl_f%fscale(1)-pot%spl_f%cutoff
         END DO
      END IF
      CALL finalizef()

      CALL timestop(handle)

   END SUBROUTINE set_potparm_index

! **************************************************************************************************
!> \brief Gives back the indices of the matrix w.r.t. the collective array index
!> \param Inind ...
!> \param ndim ...
!> \param ij ...
!> \author Teodoro Laino [tlaino] 2006.05
! **************************************************************************************************
   SUBROUTINE get_indexes(Inind, ndim, ij)
      INTEGER, INTENT(IN)                                :: Inind, ndim
      INTEGER, DIMENSION(2), INTENT(OUT)                 :: ij

      INTEGER                                            :: i, tmp

      tmp = 0
      ij = HUGE(0)
      DO i = 1, ndim
         tmp = tmp+i
         IF (tmp >= Inind) THEN
            ij(1) = i
            ij(2) = Inind-tmp+i
            EXIT
         END IF
      END DO
   END SUBROUTINE get_indexes

END MODULE pair_potential

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Lumps all possible extended system variables into one
!>    type for easy access and passing
!> \par History
!>      Teodoro Laino - 09.2007 - University of Zurich
!>        Cleaned the typo.. no need to have an extended
!>        type. Thermostat and Barostat type have been created
!> \author CJM
! **************************************************************************************************
MODULE extended_system_types
   USE bibliography,                    ONLY: Nose1984a,&
                                              Nose1984b,&
                                              cite_reference
   USE input_constants,                 ONLY: do_thermo_no_communication
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: lnhc_dealloc, &
             lnhc_init, &
             lnhc_parameters_type, &
             map_info_type, &
             create_map_info_type, &
             release_map_info_type, &
             npt_info_type

   LOGICAL, PARAMETER, PUBLIC :: debug_isotropic_limit = .FALSE.
   LOGICAL, PARAMETER, PUBLIC :: debug_uniaxial_limit = .FALSE.

! **************************************************************************************************
   TYPE npt_info_type
      REAL(KIND=dp) :: eps
      REAL(KIND=dp) :: v
      REAL(KIND=dp) :: f
      REAL(KIND=dp) :: mass
   END TYPE npt_info_type

! **************************************************************************************************
   TYPE nhc_info_type
      INTEGER        :: degrees_of_freedom
      REAL(KIND=dp) :: eta
      REAL(KIND=dp) :: v
      REAL(KIND=dp) :: f
      REAL(KIND=dp) :: nkt
      REAL(KIND=dp) :: mass
   END TYPE nhc_info_type

! **************************************************************************************************
   TYPE point_info_type
      REAL(KIND=dp), POINTER :: point
   END TYPE point_info_type

! **************************************************************************************************
   TYPE map_info_type
      INTEGER                                         :: dis_type
      INTEGER, POINTER, DIMENSION(:)                  :: index, map_index
      REAL(KIND=dp), POINTER, DIMENSION(:)           :: v_scale
      REAL(KIND=dp), POINTER, DIMENSION(:)           :: s_kin
      TYPE(point_info_type), POINTER, DIMENSION(:, :) :: p_scale
      TYPE(point_info_type), POINTER, DIMENSION(:, :) :: p_kin
   END TYPE map_info_type

! **************************************************************************************************
   TYPE lnhc_parameters_type
      INTEGER                                 :: nyosh, nc, nhc_len
      INTEGER                                 :: glob_num_nhc, loc_num_nhc, region
      REAL(KIND=dp)                           :: tau_nhc, dt_fact
      REAL(KIND=dp), POINTER                  :: dt_yosh(:)
      TYPE(nhc_info_type), POINTER            :: nvt(:, :)
      TYPE(map_info_type), POINTER            :: map_info
   END TYPE lnhc_parameters_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'extended_system_types'

CONTAINS

! **************************************************************************************************
!> \brief Initialize type for Nose-Hoover thermostat
!> \param lnhc ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE lnhc_init(lnhc, section)
      TYPE(lnhc_parameters_type), POINTER                :: lnhc
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lnhc_init', routineP = moduleN//':'//routineN

      NULLIFY (lnhc%dt_yosh)
      NULLIFY (lnhc%nvt)
      NULLIFY (lnhc%map_info)
      lnhc%loc_num_nhc = 0
      lnhc%glob_num_nhc = 0
      lnhc%dt_fact = 1.0_dp
      CALL cite_reference(Nose1984a)
      CALL cite_reference(Nose1984b)
      CALL section_vals_val_get(section, "LENGTH", i_val=lnhc%nhc_len)
      CALL section_vals_val_get(section, "YOSHIDA", i_val=lnhc%nyosh)
      CALL section_vals_val_get(section, "TIMECON", r_val=lnhc%tau_nhc)
      CALL section_vals_val_get(section, "MTS", i_val=lnhc%nc)
      CALL create_map_info_type(lnhc%map_info)

   END SUBROUTINE lnhc_init

! **************************************************************************************************
!> \brief create the map_info type
!> \param map_info ...
! **************************************************************************************************
   SUBROUTINE create_map_info_type(map_info)
      TYPE(map_info_type), POINTER                       :: map_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_map_info_type', &
         routineP = moduleN//':'//routineN

      ALLOCATE (map_info)
      NULLIFY (map_info%index, map_info%map_index)
      NULLIFY (map_info%v_scale)
      NULLIFY (map_info%p_scale)
      NULLIFY (map_info%s_kin)
      NULLIFY (map_info%p_kin)
      map_info%dis_type = do_thermo_no_communication

   END SUBROUTINE create_map_info_type

! **************************************************************************************************
!> \brief release the map_info type
!> \param map_info ...
! **************************************************************************************************
   SUBROUTINE release_map_info_type(map_info)
      TYPE(map_info_type), POINTER                       :: map_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_map_info_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(map_info)) THEN
         IF (ASSOCIATED(map_info%p_kin)) THEN
            DEALLOCATE (map_info%p_kin)
         END IF
         IF (ASSOCIATED(map_info%p_scale)) THEN
            DEALLOCATE (map_info%p_scale)
         END IF
         IF (ASSOCIATED(map_info%v_scale)) THEN
            DEALLOCATE (map_info%v_scale)
         END IF
         IF (ASSOCIATED(map_info%s_kin)) THEN
            DEALLOCATE (map_info%s_kin)
         END IF
         IF (ASSOCIATED(map_info%index)) THEN
            DEALLOCATE (map_info%index)
         END IF
         IF (ASSOCIATED(map_info%map_index)) THEN
            DEALLOCATE (map_info%map_index)
         END IF

         DEALLOCATE (map_info)
      END IF

   END SUBROUTINE release_map_info_type

! **************************************************************************************************
!> \brief Deallocate type for Nose-Hoover thermostat
!> \param lnhc ...
! **************************************************************************************************
   SUBROUTINE lnhc_dealloc(lnhc)
      TYPE(lnhc_parameters_type), POINTER                :: lnhc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lnhc_dealloc', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(lnhc)) THEN
         IF (ASSOCIATED(lnhc%dt_yosh)) THEN
            DEALLOCATE (lnhc%dt_yosh)
         END IF
         IF (ASSOCIATED(lnhc%nvt)) THEN
            DEALLOCATE (lnhc%nvt)
         ENDIF
         CALL release_map_info_type(lnhc%map_info)
         DEALLOCATE (lnhc)
      ENDIF

   END SUBROUTINE lnhc_dealloc

END MODULE extended_system_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to read the binary restart file of CP2K
!> \author Matthias Krack (MK)
!> \par History
!>      - Creation (17.02.2011,MK)
!> \version 1.0
! **************************************************************************************************
MODULE input_cp2k_binary_restarts

   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr,&
                                              debug_print_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE extended_system_types,           ONLY: lnhc_parameters_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_size
   USE message_passing,                 ONLY: mp_bcast
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE termination,                     ONLY: print_message,&
                                              stop_memory
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_binary_restarts'

   PUBLIC :: read_binary_coordinates, &
             read_binary_cs_coordinates, &
             read_binary_thermostats_nose, &
             read_binary_velocities

CONTAINS

! **************************************************************************************************
!> \brief   Read the input section &COORD from an external file written in
!>          binary format.
!> \param topology ...
!> \param root_section ...
!> \param para_env ...
!> \param subsys_section ...
!> \param binary_file_read ...
!> \par History
!>      - Creation (10.02.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_binary_coordinates(topology, root_section, para_env, &
                                      subsys_section, binary_file_read)

      TYPE(topology_parameters_type)                     :: topology
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section
      LOGICAL, INTENT(OUT)                               :: binary_file_read

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_binary_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: binary_restart_file_name
      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: handle, iatom, ikind, input_unit, istat, &
                                                            iw, natom, natomkind, ncore, &
                                                            nmolecule, nmoleculekind, nshell
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ibuf, id_name
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      IF(.NOT.(ASSOCIATED(root_section)))CALL cp__a("input_cp2k_binary_restarts.F",94)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("input_cp2k_binary_restarts.F",95)
      IF(.NOT.(ASSOCIATED(subsys_section)))CALL cp__a("input_cp2k_binary_restarts.F",96)
      logger => cp_get_default_logger()

      binary_file_read = .FALSE.

      CALL section_vals_val_get(root_section, "EXT_RESTART%BINARY_RESTART_FILE_NAME", &
                                c_val=binary_restart_file_name)

      IF (TRIM(ADJUSTL(binary_restart_file_name)) == "") THEN
         CALL timestop(handle)
         RETURN
      END IF

      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XYZ_INFO", &
                                extension=".subsysLog")

      natomkind = 0
      natom = 0
      ncore = 0
      nshell = 0
      nmoleculekind = 0
      nmolecule = 0

      ! Open binary restart file and read number atomic kinds, atoms, etc.
      IF (para_env%ionode) THEN
         CALL open_file(file_name=binary_restart_file_name, &
                        file_status="OLD", &
                        file_form="UNFORMATTED", &
                        file_action="READWRITE", &
                        file_position="REWIND", &
                        unit_number=input_unit, &
                        debug=iw)
         READ (UNIT=input_unit, IOSTAT=istat) &
            natomkind, natom, ncore, nshell, nmoleculekind, nmolecule
         IF (istat /= 0) THEN
            CALL stop_read("natomkind,natom,ncore,nshell,nmoleculekind,nmolecule "// &
                           "(IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                           input_unit)
         END IF
         IF (iw > 0) THEN
            WRITE (UNIT=iw, FMT="(T2,A,T71,I10)") &
               "Number of atomic kinds:", natomkind, &
               "Number of atoms:", natom, &
               "Number of cores (only core-shell model):", ncore, &
               "Number of shells (only core-shell model):", nshell, &
               "Number of molecule kinds:", nmoleculekind, &
               "Number of molecules", nmolecule
         END IF
      END IF

      CALL mp_bcast(natomkind, para_env%source, para_env%group)
      CALL mp_bcast(natom, para_env%source, para_env%group)
      CALL mp_bcast(ncore, para_env%source, para_env%group)
      CALL mp_bcast(nshell, para_env%source, para_env%group)
      CALL mp_bcast(nmoleculekind, para_env%source, para_env%group)
      CALL mp_bcast(nmolecule, para_env%source, para_env%group)

      ALLOCATE (id_name(natomkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 154, &
                                       "id_name", int_size*natomkind)
      ! Read atomic kind names
      DO ikind = 1, natomkind
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) string
            IF (istat /= 0) CALL stop_read("string (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
         END IF
         CALL mp_bcast(string, para_env%source, para_env%group)
         id_name(ikind) = str2id(string)
      END DO

      ! Allocate and initialise atom_info array
      atom_info => topology%atom_info
      ALLOCATE (atom_info%id_molname(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 171, &
                                       "atom_info%id_molname", &
                                       int_size*natom)
      atom_info%id_molname(:) = 0
      ALLOCATE (atom_info%id_resname(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 176, &
                                       "atom_info%id_resname", &
                                       int_size*natom)
      atom_info%id_resname(:) = 0
      ALLOCATE (atom_info%resid(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 181, &
                                       "atom_info%resid", &
                                       int_size*natom)
      atom_info%resid = 1
      ALLOCATE (atom_info%id_atmname(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 186, &
                                       "atom_info%id_atmname", &
                                       int_size*natom)
      atom_info%id_atmname = 0
      ALLOCATE (atom_info%r(3, natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 191, &
                                       "atom_info%r", &
                                       int_size*3*natom)
      atom_info%r(:, :) = 0.0_dp
      ALLOCATE (atom_info%atm_mass(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 196, &
                                       "atom_info%atm_mass", &
                                       dp_size*natom)
      atom_info%atm_mass(:) = HUGE(0.0_dp)
      ALLOCATE (atom_info%atm_charge(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 201, &
                                       "atom_info%atm_charge", &
                                       dp_size*natom)
      atom_info%atm_charge(:) = -HUGE(0.0_dp)
      ALLOCATE (atom_info%occup(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 206, &
                                       "atom_info%occup", &
                                       dp_size*natom)
      atom_info%occup(:) = 0.0_dp
      ALLOCATE (atom_info%beta(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 211, &
                                       "atom_info%id_beta", &
                                       dp_size*natom)
      atom_info%beta(:) = 0.0_dp
      ALLOCATE (atom_info%id_element(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 216, &
                                       "atom_info%id_element", &
                                       int_size*natom)
      atom_info%id_element(:) = 0
      ALLOCATE (ibuf(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 221, &
                                       "ibuf", int_size*natom)

      ! Read atomic kind number of each atom
      IF (para_env%ionode) THEN
         READ (UNIT=input_unit, IOSTAT=istat) ibuf(1:natom)
         IF (istat /= 0) CALL stop_read("ibuf (IOSTAT = "// &
                                        TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                        input_unit)
      END IF
      CALL mp_bcast(ibuf, para_env%source, para_env%group)
      DO iatom = 1, natom
         ikind = ibuf(iatom)
         atom_info%id_atmname(iatom) = id_name(ikind)
         atom_info%id_element(iatom) = id_name(ikind)
      END DO
      DEALLOCATE (id_name)

      ! Read atomic coordinates
      IF (para_env%ionode) THEN
         READ (UNIT=input_unit, IOSTAT=istat) atom_info%r(1:3, 1:natom)
         IF (istat /= 0) CALL stop_read("atom_info%r(1:3,1:natom) (IOSTAT = "// &
                                        TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                        input_unit)
      END IF
      CALL mp_bcast(atom_info%r, para_env%source, para_env%group)

      ! Read molecule information if available
      IF (nmolecule > 0) THEN
         ALLOCATE (id_name(nmoleculekind), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 251, &
                                          "id_name", nmoleculekind*int_size)
         ! Read molecule kind names
         DO ikind = 1, nmoleculekind
            IF (para_env%ionode) THEN
               READ (UNIT=input_unit, IOSTAT=istat) string
               IF (istat /= 0) CALL stop_read("string (IOSTAT = "// &
                                              TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                              input_unit)
            END IF
            CALL mp_bcast(string, para_env%source, para_env%group)
            id_name(ikind) = str2id(string)
         END DO
         ! Read molecule kind numbers
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) ibuf(1:natom)
            IF (istat /= 0) CALL stop_read("ibuf(1:natom) (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
         END IF
         CALL mp_bcast(ibuf, para_env%source, para_env%group)
         DO iatom = 1, natom
            ikind = ibuf(iatom)
            atom_info%id_molname(iatom) = id_name(ikind)
         END DO
         DEALLOCATE (id_name)
         ! Read molecule index which is used also as residue id
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) atom_info%resid(1:natom)
            IF (istat /= 0) CALL stop_read("atom_info%resid(1:natom) (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
         END IF
         CALL mp_bcast(atom_info%resid, para_env%source, para_env%group)
         DO iatom = 1, natom
            atom_info%id_resname(iatom) = str2id(s2s(cp_to_string(atom_info%resid(iatom))))
         END DO
      END IF
      DEALLOCATE (ibuf)

      !MK to be checked ...
      topology%aa_element = .TRUE.
      topology%molname_generated = .FALSE.
      topology%natoms = natom

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "BEGIN of COORD section data [Angstrom] read in binary format from file "// &
            TRIM(binary_restart_file_name)
         DO iatom = 1, natom
            WRITE (UNIT=iw, FMT="(T2,A2,3(1X,ES25.16),2(1X,A))") &
               TRIM(ADJUSTL(id2str(atom_info%id_atmname(iatom)))), &
               atom_info%r(1:3, iatom)*angstrom, &
               TRIM(ADJUSTL(id2str(atom_info%id_molname(iatom)))), &
               TRIM(ADJUSTL(id2str(atom_info%id_resname(iatom))))
         END DO
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "END of COORD section data [Angstrom] read from binary restart file "// &
            TRIM(binary_restart_file_name)
      END IF

      IF (para_env%ionode) CALL close_file(unit_number=input_unit, &
                                           keep_preconnection=.TRUE.)

      binary_file_read = .TRUE.

      CALL timestop(handle)

   END SUBROUTINE read_binary_coordinates

! **************************************************************************************************
!> \brief   Read the input section &CORE_COORD or &SHELL_COORD from an external
!>          file written in binary format.
!> \param prefix ...
!> \param particle_set ...
!> \param root_section ...
!> \param subsys_section ...
!> \param binary_file_read ...
!> \par History
!>      - Creation (17.02.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_binary_cs_coordinates(prefix, particle_set, root_section, &
                                         subsys_section, binary_file_read)

      CHARACTER(LEN=*), INTENT(IN)                       :: prefix
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: root_section, subsys_section
      LOGICAL, INTENT(OUT)                               :: binary_file_read

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_binary_cs_coordinates', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: binary_restart_file_name, message
      CHARACTER(LEN=default_string_length)               :: section_label, section_name
      INTEGER                                            :: handle, input_unit, iparticle, istat, &
                                                            iw, nbuf, nparticle
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ibuf
      LOGICAL                                            :: exit_routine
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rbuf
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      IF(.NOT.(ASSOCIATED(root_section)))CALL cp__a("input_cp2k_binary_restarts.F",358)
      IF(.NOT.(ASSOCIATED(subsys_section)))CALL cp__a("input_cp2k_binary_restarts.F",359)
      logger => cp_get_default_logger()
      para_env => logger%para_env

      binary_file_read = .FALSE.

      IF (ASSOCIATED(particle_set)) THEN
         exit_routine = .FALSE.
         nparticle = SIZE(particle_set)
      ELSE
         exit_routine = .TRUE.
         nparticle = 0
      END IF

      CALL section_vals_val_get(root_section, "EXT_RESTART%BINARY_RESTART_FILE_NAME", &
                                c_val=binary_restart_file_name)

      IF (TRIM(ADJUSTL(binary_restart_file_name)) == "") THEN
         CALL timestop(handle)
         RETURN
      END IF

      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XYZ_INFO", &
                                extension=".subsysLog")

      section_name = prefix//" COORDINATES"

      ! Open binary restart file at last position
      IF (para_env%ionode) THEN
         CALL open_file(file_name=TRIM(binary_restart_file_name), &
                        file_status="OLD", &
                        file_form="UNFORMATTED", &
                        file_action="READWRITE", &
                        file_position="ASIS", &
                        unit_number=input_unit, &
                        debug=iw)
         READ (UNIT=input_unit, IOSTAT=istat) section_label, nbuf
         IF (istat /= 0) CALL stop_read("section_label, nbuf -> "//TRIM(section_label)//", "// &
                                        TRIM(ADJUSTL(cp_to_string(nbuf)))// &
                                        " (IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//"). "// &
                                        "Section "//TRIM(ADJUSTL(section_name))//" was expected.", &
                                        input_unit)
         IF (TRIM(section_label) == TRIM(section_name)) THEN
            IF (nbuf /= nparticle) THEN
               IF (iw > 0) THEN
                  message = "INFO: The requested number of "//TRIM(section_name)//" ("// &
                            TRIM(ADJUSTL(cp_to_string(nparticle)))//") does not match the "// &
                            "number ("//TRIM(ADJUSTL(cp_to_string(nbuf)))//") available from the "// &
                            "binary restart file <"//TRIM(binary_restart_file_name)// &
                            ">. The restart file information is ignored."
                  CALL print_message(message, iw, 1, 1, 1)
               END IF
               ! Ignore this section
               IF (nbuf > 0) THEN
                  ! Perform dummy read
                  ALLOCATE (rbuf(3, nbuf), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 415, &
                                                   "rbuf", dp_size*3*nbuf)
                  READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:3, 1:nbuf)
                  IF (istat /= 0) CALL stop_read("rbuf(1:3,1:nbuf) -> "//prefix// &
                                                 " coordinates (IOSTAT = "// &
                                                 TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                                 input_unit)
                  DEALLOCATE (rbuf)
                  ALLOCATE (ibuf(nbuf), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 424, &
                                                   "ibuf", int_size*nbuf)
                  READ (UNIT=input_unit, IOSTAT=istat) ibuf(1:nbuf)
                  IF (istat /= 0) CALL stop_read("ibuf(1:nparticle) -> atomic indices of the "// &
                                                 TRIM(section_name)//" (IOSTAT = "// &
                                                 TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                                 input_unit)
                  DEALLOCATE (ibuf)
               END IF
               exit_routine = .TRUE.
            ELSE
               IF (iw > 0) THEN
                  WRITE (UNIT=iw, FMT="(T2,A,T71,I10)") &
                     "Number of "//prefix//" particles:", nparticle
               END IF
               IF (nparticle == 0) exit_routine = .TRUE.
            END IF
         ELSE
            CALL cp_abort(cp__l("input_cp2k_binary_restarts.F",442), &
                          "Section label <"//TRIM(section_label)//"> read from the "// &
                          "binary restart file <"//TRIM(binary_restart_file_name)// &
                          "> does not match the requested section name <"// &
                          TRIM(section_name)//">.")
         END IF
      END IF

      CALL mp_bcast(exit_routine, para_env%source, para_env%group)
      IF (exit_routine) THEN
         IF (para_env%ionode) CALL close_file(unit_number=input_unit, &
                                              keep_preconnection=.TRUE.)
         CALL timestop(handle)
         RETURN
      END IF

      IF(.NOT.(nparticle > 0))CALL cp__a("input_cp2k_binary_restarts.F",458)

      ALLOCATE (rbuf(3, nparticle), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 461, &
                                       "rbuf", dp_size*3*nparticle)

      IF (para_env%ionode) THEN
         READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:3, 1:nparticle)
         IF (istat /= 0) CALL stop_read("rbuf(1:3,1:nparticle) -> "//prefix// &
                                        " coordinates (IOSTAT = "// &
                                        TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                        input_unit)
      END IF
      CALL mp_bcast(rbuf, para_env%source, para_env%group)

      DO iparticle = 1, nparticle
         particle_set(iparticle)%r(1:3) = rbuf(1:3, iparticle)
      END DO

      DEALLOCATE (rbuf)

      ALLOCATE (ibuf(nparticle), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 480, &
                                       "ibuf", int_size*nparticle)

      IF (para_env%ionode) THEN
         READ (UNIT=input_unit, IOSTAT=istat) ibuf(1:nparticle)
         IF (istat /= 0) CALL stop_read("ibuf(1:nparticle) -> atomic indices of the "// &
                                        TRIM(section_name)//" (IOSTAT = "// &
                                        TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                        input_unit)
      END IF

      CALL mp_bcast(ibuf, para_env%source, para_env%group)

      DO iparticle = 1, nparticle
         particle_set(iparticle)%atom_index = ibuf(iparticle)
      END DO

      DEALLOCATE (ibuf)

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "BEGIN of "//TRIM(ADJUSTL(section_name))// &
            " section data [Angstrom] read in binary format from file "// &
            TRIM(binary_restart_file_name)
         DO iparticle = 1, nparticle
            WRITE (UNIT=iw, FMT="(T2,A2,3(1X,ES25.16),1X,I0)") &
               TRIM(ADJUSTL(particle_set(iparticle)%atomic_kind%name)), &
               particle_set(iparticle)%r(1:3)*angstrom, &
               particle_set(iparticle)%atom_index
         END DO
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "END of "//TRIM(ADJUSTL(section_name))// &
            " section data [Angstrom] read from binary restart file "// &
            TRIM(binary_restart_file_name)
      END IF

      IF (para_env%ionode) CALL close_file(unit_number=input_unit, &
                                           keep_preconnection=.TRUE.)

      binary_file_read = .TRUE.

      CALL timestop(handle)

   END SUBROUTINE read_binary_cs_coordinates

! **************************************************************************************************
!> \brief   Read the input section &VELOCITY, &CORE_VELOCITY, or
!>          &SHELL_VELOCITY from an external file written in binary format.
!> \param prefix ...
!> \param particle_set ...
!> \param root_section ...
!> \param para_env ...
!> \param subsys_section ...
!> \param binary_file_read ...
!> \par History
!>      - Creation (17.02.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_binary_velocities(prefix, particle_set, root_section, para_env, &
                                     subsys_section, binary_file_read)

      CHARACTER(LEN=*), INTENT(IN)                       :: prefix
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section
      LOGICAL, INTENT(OUT)                               :: binary_file_read

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_binary_velocities', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: binary_restart_file_name, message
      CHARACTER(LEN=default_string_length)               :: section_label, section_name
      INTEGER                                            :: handle, i, input_unit, iparticle, istat, &
                                                            iw, nbuf, nparticle
      LOGICAL                                            :: have_velocities
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rbuf
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      IF(.NOT.(ASSOCIATED(root_section)))CALL cp__a("input_cp2k_binary_restarts.F",563)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("input_cp2k_binary_restarts.F",564)
      IF(.NOT.(ASSOCIATED(subsys_section)))CALL cp__a("input_cp2k_binary_restarts.F",565)
      logger => cp_get_default_logger()

      binary_file_read = .FALSE.

      CALL section_vals_val_get(root_section, "EXT_RESTART%BINARY_RESTART_FILE_NAME", &
                                c_val=binary_restart_file_name)

      IF (TRIM(ADJUSTL(binary_restart_file_name)) == "") THEN
         CALL timestop(handle)
         RETURN
      END IF

      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XYZ_INFO", &
                                extension=".subsysLog")

      IF (LEN_TRIM(prefix) == 0) THEN
         section_name = "VELOCITIES"
      ELSE
         section_name = prefix//" VELOCITIES"
      END IF

      have_velocities = .FALSE.

      IF (ASSOCIATED(particle_set)) THEN
         nparticle = SIZE(particle_set)
      ELSE
         nparticle = 0
      END IF

      ! Open binary restart file at last position and check if there are
      ! velocities available
      IF (para_env%ionode) THEN
         CALL open_file(file_name=binary_restart_file_name, &
                        file_status="OLD", &
                        file_form="UNFORMATTED", &
                        file_action="READWRITE", &
                        file_position="ASIS", &
                        unit_number=input_unit, &
                        debug=iw)
         DO
            READ (UNIT=input_unit, IOSTAT=istat) section_label, nbuf
            IF (istat /= 0) CALL stop_read("section_label, nbuf -> "//TRIM(section_label)//", "// &
                                           TRIM(ADJUSTL(cp_to_string(nbuf)))// &
                                           " (IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//"). "// &
                                           "Section "//TRIM(ADJUSTL(section_name))//" was expected.", &
                                           input_unit)
            IF (INDEX(section_label, "THERMOSTAT") > 0) THEN
               IF (nbuf > 0) THEN
                  ! Ignore thermostat information
                  ALLOCATE (rbuf(nbuf, 1), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 616, &
                                                   "rbuf", dp_size*nbuf)
                  ! Perform dummy read
                  DO i = 1, 4
                     READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:nbuf, 1)
                     IF (istat /= 0) CALL stop_read("rbuf(1:nbuf,1) -> "// &
                                                    TRIM(ADJUSTL(section_label))// &
                                                    " (IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                                    input_unit)
                  END DO
                  DEALLOCATE (rbuf)
                  IF (iw > 0) THEN
                     message = "INFO: Ignoring section <"//TRIM(ADJUSTL(section_label))// &
                               "> from binary restart file <"//TRIM(binary_restart_file_name)//">."
                     CALL print_message(message, iw, 1, 1, 1)
                  END IF
               END IF
               CYCLE
            ELSE IF (INDEX(section_label, "VELOCIT") == 0) THEN
               CALL cp_abort(cp__l("input_cp2k_binary_restarts.F",635), &
                             "Section label <"//TRIM(section_label)//"> read from the "// &
                             "binary restart file <"//TRIM(binary_restart_file_name)// &
                             "> does not match the requested section name <"// &
                             TRIM(section_name)//">.")
            ELSE
               IF (nbuf > 0) have_velocities = .TRUE.
               EXIT
            END IF
         END DO
      END IF

      CALL mp_bcast(nbuf, para_env%source, para_env%group)
      CALL mp_bcast(have_velocities, para_env%source, para_env%group)

      IF (have_velocities) THEN

         ALLOCATE (rbuf(3, nbuf), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 653, &
                                          "rbuf", dp_size*3*nbuf)

         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:3, 1:nbuf)
            IF (istat /= 0) CALL stop_read("rbuf(1:3,1:nbuf) -> "// &
                                           TRIM(ADJUSTL(section_name))// &
                                           " (IOSTAT = "//TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
         END IF

         IF (nbuf == nparticle) THEN
            CALL mp_bcast(rbuf, para_env%source, para_env%group)
            DO iparticle = 1, nparticle
               particle_set(iparticle)%v(1:3) = rbuf(1:3, iparticle)
            END DO
         ELSE
            IF (iw > 0) THEN
               message = "INFO: The requested number of "//TRIM(ADJUSTL(section_name))// &
                         " ("//TRIM(ADJUSTL(cp_to_string(nparticle)))//") does not match the "// &
                         "number ("//TRIM(ADJUSTL(cp_to_string(nbuf)))//") available from the "// &
                         "binary restart file <"//TRIM(binary_restart_file_name)// &
                         ">. The restart file information is ignored."
               CALL print_message(message, iw, 1, 1, 1)
            END IF
         END IF

         DEALLOCATE (rbuf)

      END IF

      IF (nbuf == nparticle) THEN
         IF (iw > 0) THEN
            WRITE (UNIT=iw, FMT="(T2,A)") &
               "BEGIN of "//TRIM(ADJUSTL(section_name))// &
               " section data [a.u.] read in binary format from file "// &
               TRIM(binary_restart_file_name)
            IF (have_velocities) THEN
               DO iparticle = 1, nparticle
                  WRITE (UNIT=iw, FMT="(T2,A2,3(1X,ES25.16))") &
                     TRIM(ADJUSTL(particle_set(iparticle)%atomic_kind%name)), &
                     particle_set(iparticle)%v(1:3)
               END DO
            ELSE
               WRITE (UNIT=iw, FMT="(A)") &
                  "# No "//TRIM(ADJUSTL(section_name))//" available"
            END IF
            WRITE (UNIT=iw, FMT="(T2,A)") &
               "END of "//TRIM(ADJUSTL(section_name))// &
               " section data [a.u.] read from binary restart file "// &
               TRIM(binary_restart_file_name)
         END IF
         binary_file_read = .TRUE.
      END IF

      IF (para_env%ionode) CALL close_file(unit_number=input_unit, &
                                           keep_preconnection=.TRUE.)

      CALL timestop(handle)

   END SUBROUTINE read_binary_velocities

! **************************************************************************************************
!> \brief   Read the input section &THERMOSTAT for Nose thermostats from an
!>          external file written in binary format.
!> \param prefix ...
!> \param nhc ...
!> \param binary_restart_file_name ...
!> \param restart ...
!> \param para_env ...
!> \par History
!>      - Creation (28.02.2011,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_binary_thermostats_nose(prefix, nhc, binary_restart_file_name, &
                                           restart, para_env)

      CHARACTER(LEN=*), INTENT(IN)                       :: prefix
      TYPE(lnhc_parameters_type), POINTER                :: nhc
      CHARACTER(LEN=*), INTENT(IN)                       :: binary_restart_file_name
      LOGICAL, INTENT(OUT)                               :: restart
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_binary_thermostats_nose', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: section_label, section_name
      INTEGER                                            :: handle, i, idx, input_unit, istat, j, &
                                                            nhc_size, output_unit
      LOGICAL                                            :: debug
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: rbuf
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(nhc)))CALL cp__a("input_cp2k_binary_restarts.F",749)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("input_cp2k_binary_restarts.F",750)

      ! Set to .TRUE. for debug mode, i.e. all data read are written to stdout
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF (logger%iter_info%print_level >= debug_print_level) THEN
         debug = .TRUE.
      ELSE
         debug = .FALSE.
      END IF

      restart = .FALSE.

      section_name = prefix//" THERMOSTATS"

      ! Open binary restart file at last position
      IF (para_env%ionode) THEN
         CALL open_file(file_name=binary_restart_file_name, &
                        file_status="OLD", &
                        file_form="UNFORMATTED", &
                        file_action="READWRITE", &
                        file_position="ASIS", &
                        unit_number=input_unit)
         READ (UNIT=input_unit, IOSTAT=istat) section_label, nhc_size
         IF (istat /= 0) CALL stop_read("nhc_size (IOSTAT = "// &
                                        TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                        input_unit)
         IF (INDEX(section_label, "THERMOSTAT") == 0) THEN
            CALL cp_abort(cp__l("input_cp2k_binary_restarts.F",780), &
                          "Section label <"//TRIM(section_label)//"> read from the "// &
                          "binary restart file <"//TRIM(binary_restart_file_name)// &
                          "> does not match the requested section name <"// &
                          TRIM(section_name)//">.")
         END IF
         IF (debug .AND. output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,/,T2,A,I0)") &
               "BEGIN of "//TRIM(ADJUSTL(section_label))// &
               " section data read in binary format from file "// &
               TRIM(binary_restart_file_name), &
               "# nhc_size = ", nhc_size
         END IF
      END IF

      CALL mp_bcast(nhc_size, para_env%source, para_env%group)

      IF (nhc_size > 0) THEN

         ALLOCATE (rbuf(nhc_size), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 800, &
                                          "rbuf", dp_size*nhc_size)
         rbuf(:) = 0.0_dp

         ! Read NHC section &COORD
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:nhc_size)
            IF (istat /= 0) CALL stop_read("eta -> rbuf (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
            IF (debug .AND. output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(4(1X,ES25.16)))") &
                  "&COORD", rbuf(1:nhc_size)
            END IF
         END IF
         CALL mp_bcast(rbuf, para_env%source, para_env%group)
         DO i = 1, SIZE(nhc%nvt, 2)
            idx = (nhc%map_info%index(i)-1)*nhc%nhc_len
            DO j = 1, SIZE(nhc%nvt, 1)
               idx = idx+1
               nhc%nvt(j, i)%eta = rbuf(idx)
            END DO
         END DO

         ! Read NHC section &VELOCITY
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:nhc_size)
            IF (istat /= 0) CALL stop_read("veta -> rbuf (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
            IF (debug .AND. output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(4(1X,ES25.16)))") &
                  "&VELOCITY", rbuf(1:nhc_size)
            END IF
         END IF
         CALL mp_bcast(rbuf, para_env%source, para_env%group)
         DO i = 1, SIZE(nhc%nvt, 2)
            idx = (nhc%map_info%index(i)-1)*nhc%nhc_len
            DO j = 1, SIZE(nhc%nvt, 1)
               idx = idx+1
               nhc%nvt(j, i)%v = rbuf(idx)
            END DO
         END DO

         ! Read NHC section &MASS
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:nhc_size)
            IF (istat /= 0) CALL stop_read("mnhc -> rbuf (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
            IF (debug .AND. output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(4(1X,ES25.16)))") &
                  "&MASS:", rbuf(1:nhc_size)
            END IF
         END IF
         CALL mp_bcast(rbuf, para_env%source, para_env%group)
         DO i = 1, SIZE(nhc%nvt, 2)
            idx = (nhc%map_info%index(i)-1)*nhc%nhc_len
            DO j = 1, SIZE(nhc%nvt, 1)
               idx = idx+1
               nhc%nvt(j, i)%mass = rbuf(idx)
            END DO
         END DO

         ! Read NHC section &FORCE
         IF (para_env%ionode) THEN
            READ (UNIT=input_unit, IOSTAT=istat) rbuf(1:nhc_size)
            IF (istat /= 0) CALL stop_read("fnhc -> rbuf (IOSTAT = "// &
                                           TRIM(ADJUSTL(cp_to_string(istat)))//")", &
                                           input_unit)
            IF (debug .AND. output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(4(1X,ES25.16)))") &
                  "&FORCE", rbuf(1:nhc_size)
            END IF
         END IF
         CALL mp_bcast(rbuf, para_env%source, para_env%group)
         DO i = 1, SIZE(nhc%nvt, 2)
            idx = (nhc%map_info%index(i)-1)*nhc%nhc_len
            DO j = 1, SIZE(nhc%nvt, 1)
               idx = idx+1
               nhc%nvt(j, i)%f = rbuf(idx)
            END DO
         END DO

         DEALLOCATE (rbuf)

         restart = .TRUE.

      END IF

      IF (para_env%ionode) THEN
         IF (debug .AND. output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "END of"//TRIM(ADJUSTL(section_label))// &
               " section data read in binary format from file "// &
               TRIM(binary_restart_file_name)
         END IF
         CALL close_file(unit_number=input_unit, &
                         keep_preconnection=.TRUE.)
      END IF

      CALL timestop(handle)

   END SUBROUTINE read_binary_thermostats_nose

! **************************************************************************************************
!> \brief Print an error message and stop the program execution in case of a
!>        read error.
!> \param object ...
!> \param unit_number ...
!> \par History
!>      - Creation (15.02.2011,MK)
!> \author Matthias Krack (MK)
!> \note
!>      object     : Name of the data object for which I/O operation failed
!>      unit_number: Logical unit number of the file read from
! **************************************************************************************************
   SUBROUTINE stop_read(object, unit_number)
      CHARACTER(LEN=*), INTENT(IN)                       :: object
      INTEGER, INTENT(IN)                                :: unit_number

      CHARACTER(LEN=2*default_path_length)               :: message
      CHARACTER(LEN=default_path_length)                 :: file_name
      LOGICAL                                            :: file_exists

      IF (unit_number >= 0) THEN
         INQUIRE (UNIT=unit_number, EXIST=file_exists)
      ELSE
         file_exists = .FALSE.
      ENDIF
      IF (file_exists) THEN
         INQUIRE (UNIT=unit_number, NAME=file_name)
         WRITE (UNIT=message, FMT="(A)") &
            "An error occurred reading data object <"//TRIM(ADJUSTL(object))// &
            "> from file <"//TRIM(ADJUSTL(file_name))//">"
      ELSE
         WRITE (UNIT=message, FMT="(A,I0,A)") &
            "Could not read data object <"//TRIM(ADJUSTL(object))// &
            "> from logical unit ", unit_number, ". The I/O unit does not exist."
      END IF

      CALL cp__b("input_cp2k_binary_restarts.F",941,message)

   END SUBROUTINE stop_read

END MODULE input_cp2k_binary_restarts
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Splitting and cleaning the original force_field_pack - May 2007
!>      Teodoro Laino - Zurich University
!> \author CJM
! **************************************************************************************************
MODULE force_fields_all

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set,&
                                              set_atomic_kind
   USE atoms_input,                     ONLY: read_shell_coord_input
   USE cell_types,                      ONLY: cell_type
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE damping_dipole_types,            ONLY: damping_p_create,&
                                              damping_p_type,&
                                              tang_toennies
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_env_set,&
                                              ewald_environment_type
   USE external_potential_types,        ONLY: fist_potential_type,&
                                              get_potential,&
                                              set_potential
   USE force_field_kind_types,          ONLY: &
        allocate_bend_kind_set, allocate_bond_kind_set, allocate_impr_kind_set, &
        allocate_opbend_kind_set, allocate_torsion_kind_set, allocate_ub_kind_set, bend_kind_type, &
        bond_kind_type, do_ff_amber, do_ff_charmm, do_ff_g87, do_ff_g96, do_ff_undef, &
        impr_kind_type, opbend_kind_type, torsion_kind_type, ub_kind_type
   USE force_field_types,               ONLY: amber_info_type,&
                                              charmm_info_type,&
                                              force_field_type,&
                                              gromos_info_type,&
                                              input_info_type
   USE input_constants,                 ONLY: do_qmmm_none
   USE input_cp2k_binary_restarts,      ONLY: read_binary_cs_coordinates
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_list_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: sqrthalf
   USE memory_utilities,                ONLY: reallocate
   USE molecule_kind_types,             ONLY: &
        bend_type, bond_type, get_molecule_kind, impr_type, molecule_kind_type, opbend_type, &
        set_molecule_kind, shell_type, torsion_type, ub_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE pair_potential,                  ONLY: get_nonbond_storage,&
                                              spline_nonbond_control
   USE pair_potential_coulomb,          ONLY: potential_coulomb
   USE pair_potential_types,            ONLY: &
        ea_type, lj_charmm_type, lj_type, nn_type, nosh_nosh, nosh_sh, pair_potential_lj_create, &
        pair_potential_pp_create, pair_potential_pp_type, pair_potential_single_add, &
        pair_potential_single_clean, pair_potential_single_copy, pair_potential_single_type, &
        quip_type, sh_sh, siepmann_type, tersoff_type
   USE particle_types,                  ONLY: allocate_particle_set,&
                                              particle_type
   USE physcon,                         ONLY: bohr
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE shell_potential_types,           ONLY: shell_create,&
                                              shell_kind_type,&
                                              shell_release
   USE splines_types,                   ONLY: spline_data_p_release,&
                                              spline_data_p_retain,&
                                              spline_data_p_type,&
                                              spline_env_release,&
                                              spline_environment_type
   USE string_utilities,                ONLY: compress,&
                                              integer_to_string,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_fields_all'

   PRIVATE
   PUBLIC :: force_field_unique_bond, &
             force_field_unique_bend, &
             force_field_unique_ub, &
             force_field_unique_tors, &
             force_field_unique_impr, &
             force_field_unique_opbend, &
             force_field_pack_bond, &
             force_field_pack_bend, &
             force_field_pack_ub, &
             force_field_pack_tors, &
             force_field_pack_impr, &
             force_field_pack_opbend, &
             force_field_pack_charge, &
             force_field_pack_charges, &
             force_field_pack_radius, &
             force_field_pack_pol, &
             force_field_pack_shell, &
             force_field_pack_nonbond14, &
             force_field_pack_nonbond, &
             force_field_pack_splines, &
             force_field_pack_eicut, &
             force_field_pack_damp

CONTAINS

! **************************************************************************************************
!> \brief Determine the number of unique bond kind and allocate bond_kind_set
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_bond(particle_set, &
                                      molecule_kind_set, molecule_set, ff_type)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_bond', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2
      INTEGER                                            :: atm_a, atm_b, counter, first, handle2, &
                                                            i, j, k, last, natom, nbond
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_bond_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(bond_kind_type), DIMENSION(:), POINTER        :: bond_kind_set
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nbond=nbond, bond_list=bond_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (nbond > 0) THEN
            ALLOCATE (map_bond_kind(nbond))
            counter = 0
            IF ((ff_type%ff_type == do_ff_g96) .OR. (ff_type%ff_type == do_ff_g87)) THEN
               DO j = 1, nbond
                  map_bond_kind(j) = j
               END DO
               counter = nbond
            ELSE
               DO j = 1, nbond
                  atm_a = bond_list(j)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a)
                  atm_b = bond_list(j)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b)
                  found = .FALSE.
                  DO k = 1, j-1
                     atm_a = bond_list(k)%a
                     atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_a2)
                     atm_b = bond_list(k)%b
                     atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_b2)
                     IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2))) .OR. &
                         (((name_atm_a) == (name_atm_b2)) .AND. &
                          ((name_atm_b) == (name_atm_a2)))) THEN
                        found = .TRUE.
                        map_bond_kind(j) = map_bond_kind(k)
                        EXIT
                     END IF
                  END DO
                  IF (.NOT. found) THEN
                     counter = counter+1
                     map_bond_kind(j) = counter
                  END IF
               END DO
            END IF
            NULLIFY (bond_kind_set)
            CALL allocate_bond_kind_set(bond_kind_set, counter)
            DO j = 1, nbond
               bond_list(j)%bond_kind => bond_kind_set(map_bond_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   bond_kind_set=bond_kind_set, bond_list=bond_list)
            DEALLOCATE (map_bond_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_bond

! **************************************************************************************************
!> \brief Determine the number of unique bend kind and allocate bend_kind_set
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_bend(particle_set, &
                                      molecule_kind_set, molecule_set, ff_type)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_bend', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2, name_atm_c, name_atm_c2
      INTEGER                                            :: atm_a, atm_b, atm_c, counter, first, &
                                                            handle2, i, j, k, last, natom, nbend
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_bend_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(bend_kind_type), DIMENSION(:), POINTER        :: bend_kind_set
      TYPE(bend_type), DIMENSION(:), POINTER             :: bend_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nbend=nbend, bend_list=bend_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (nbend > 0) THEN
            ALLOCATE (map_bend_kind(nbend))
            counter = 0
            IF ((ff_type%ff_type == do_ff_g96) .OR. (ff_type%ff_type == do_ff_g87)) THEN
               DO j = 1, nbend
                  map_bend_kind(j) = j
               END DO
               counter = nbend
            ELSE
               DO j = 1, nbend
                  atm_a = bend_list(j)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a)
                  atm_b = bend_list(j)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b)
                  atm_c = bend_list(j)%c
                  atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_c)
                  found = .FALSE.
                  DO k = 1, j-1
                     atm_a = bend_list(k)%a
                     atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_a2)
                     atm_b = bend_list(k)%b
                     atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_b2)
                     atm_c = bend_list(k)%c
                     atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_c2)
                     IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_c2))) .OR. &
                         (((name_atm_a) == (name_atm_c2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_a2)))) THEN
                        found = .TRUE.
                        map_bend_kind(j) = map_bend_kind(k)
                        EXIT
                     END IF
                  END DO
                  IF (.NOT. found) THEN
                     counter = counter+1
                     map_bend_kind(j) = counter
                  END IF
               END DO
            END IF
            NULLIFY (bend_kind_set)
            CALL allocate_bend_kind_set(bend_kind_set, counter)
            DO j = 1, nbend
               bend_list(j)%bend_kind => bend_kind_set(map_bend_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   bend_kind_set=bend_kind_set, bend_list=bend_list)
            DEALLOCATE (map_bend_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_bend

! **************************************************************************************************
!> \brief Determine the number of unique Urey-Bradley kind and allocate ub_kind_set
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_ub(particle_set, &
                                    molecule_kind_set, molecule_set)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_ub', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2, name_atm_c, name_atm_c2
      INTEGER                                            :: atm_a, atm_b, atm_c, counter, first, &
                                                            handle2, i, j, k, last, natom, nub
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_ub_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(ub_kind_type), DIMENSION(:), POINTER          :: ub_kind_set
      TYPE(ub_type), DIMENSION(:), POINTER               :: ub_list

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nub=nub, ub_list=ub_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (nub > 0) THEN
            ALLOCATE (map_ub_kind(nub))
            counter = 0
            DO j = 1, nub
               atm_a = ub_list(j)%a
               atomic_kind => particle_set(atm_a+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_a)
               atm_b = ub_list(j)%b
               atomic_kind => particle_set(atm_b+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_b)
               atm_c = ub_list(j)%c
               atomic_kind => particle_set(atm_c+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_c)
               found = .FALSE.
               DO k = 1, j-1
                  atm_a = ub_list(k)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a2)
                  atm_b = ub_list(k)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b2)
                  atm_c = ub_list(k)%c
                  atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_c2)
                  IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                       ((name_atm_b) == (name_atm_b2)) .AND. &
                       ((name_atm_c) == (name_atm_c2))) .OR. &
                      (((name_atm_a) == (name_atm_c2)) .AND. &
                       ((name_atm_b) == (name_atm_b2)) .AND. &
                       ((name_atm_c) == (name_atm_a2)))) THEN
                     found = .TRUE.
                     map_ub_kind(j) = map_ub_kind(k)
                     EXIT
                  END IF
               END DO
               IF (.NOT. found) THEN
                  counter = counter+1
                  map_ub_kind(j) = counter
               END IF
            END DO
            CALL allocate_ub_kind_set(ub_kind_set, counter)
            DO j = 1, nub
               ub_list(j)%ub_kind => ub_kind_set(map_ub_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   ub_kind_set=ub_kind_set, ub_list=ub_list)
            DEALLOCATE (map_ub_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_ub

! **************************************************************************************************
!> \brief Determine the number of unique torsion kind and allocate torsion_kind_set
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_tors(particle_set, &
                                      molecule_kind_set, molecule_set, ff_type)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_tors', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2, name_atm_c, name_atm_c2, &
                                                            name_atm_d, name_atm_d2
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, counter, &
                                                            first, handle2, i, j, k, last, natom, &
                                                            ntorsion
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_torsion_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(torsion_kind_type), DIMENSION(:), POINTER     :: torsion_kind_set
      TYPE(torsion_type), DIMENSION(:), POINTER          :: torsion_list

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                ntorsion=ntorsion, torsion_list=torsion_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (ntorsion > 0) THEN
            ALLOCATE (map_torsion_kind(ntorsion))
            counter = 0
            IF ((ff_type%ff_type == do_ff_g96) .OR. (ff_type%ff_type == do_ff_g87)) THEN
               DO j = 1, ntorsion
                  map_torsion_kind(j) = j
               END DO
               counter = ntorsion
            ELSE
               DO j = 1, ntorsion
                  atm_a = torsion_list(j)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a)
                  atm_b = torsion_list(j)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b)
                  atm_c = torsion_list(j)%c
                  atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_c)
                  atm_d = torsion_list(j)%d
                  atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_d)
                  found = .FALSE.
                  DO k = 1, j-1
                     atm_a = torsion_list(k)%a
                     atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_a2)
                     atm_b = torsion_list(k)%b
                     atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_b2)
                     atm_c = torsion_list(k)%c
                     atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_c2)
                     atm_d = torsion_list(k)%d
                     atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_d2)
                     IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_c2)) .AND. &
                          ((name_atm_d) == (name_atm_d2))) .OR. &
                         (((name_atm_a) == (name_atm_d2)) .AND. &
                          ((name_atm_b) == (name_atm_c2)) .AND. &
                          ((name_atm_c) == (name_atm_b2)) .AND. &
                          ((name_atm_d) == (name_atm_a2)))) THEN
                        found = .TRUE.
                        map_torsion_kind(j) = map_torsion_kind(k)
                        EXIT
                     END IF
                  END DO
                  IF (.NOT. found) THEN
                     counter = counter+1
                     map_torsion_kind(j) = counter
                  END IF
               END DO
            END IF
            NULLIFY (torsion_kind_set)
            CALL allocate_torsion_kind_set(torsion_kind_set, counter)
            DO j = 1, ntorsion
               torsion_list(j)%torsion_kind => torsion_kind_set(map_torsion_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   torsion_kind_set=torsion_kind_set, torsion_list=torsion_list)
            DEALLOCATE (map_torsion_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_tors

! **************************************************************************************************
!> \brief Determine the number of unique impr kind and allocate impr_kind_set
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_impr(particle_set, &
                                      molecule_kind_set, molecule_set, ff_type)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_impr', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2, name_atm_c, name_atm_c2, &
                                                            name_atm_d, name_atm_d2
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, counter, &
                                                            first, handle2, i, j, k, last, natom, &
                                                            nimpr
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_impr_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(impr_kind_type), DIMENSION(:), POINTER        :: impr_kind_set
      TYPE(impr_type), DIMENSION(:), POINTER             :: impr_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nimpr=nimpr, impr_list=impr_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (nimpr > 0) THEN
            ALLOCATE (map_impr_kind(nimpr))
            counter = 0
            IF ((ff_type%ff_type == do_ff_g96) .OR. (ff_type%ff_type == do_ff_g87)) THEN
               DO j = 1, nimpr
                  map_impr_kind(j) = j
               END DO
               counter = nimpr
            ELSE
               DO j = 1, nimpr
                  atm_a = impr_list(j)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a)
                  atm_b = impr_list(j)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b)
                  atm_c = impr_list(j)%c
                  atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_c)
                  atm_d = impr_list(j)%d
                  atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_d)
                  found = .FALSE.
                  DO k = 1, j-1
                     atm_a = impr_list(k)%a
                     atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_a2)
                     atm_b = impr_list(k)%b
                     atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_b2)
                     atm_c = impr_list(k)%c
                     atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_c2)
                     atm_d = impr_list(k)%d
                     atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_d2)
                     IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_c2)) .AND. &
                          ((name_atm_d) == (name_atm_d2))) .OR. &
                         (((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_d2)) .AND. &
                          ((name_atm_d) == (name_atm_c2)))) THEN
                        found = .TRUE.
                        map_impr_kind(j) = map_impr_kind(k)
                        EXIT
                     END IF
                  END DO
                  IF (.NOT. found) THEN
                     counter = counter+1
                     map_impr_kind(j) = counter
                  END IF
               END DO
            END IF
            NULLIFY (impr_kind_set)
            CALL allocate_impr_kind_set(impr_kind_set, counter)
            DO j = 1, nimpr
               impr_list(j)%impr_kind => impr_kind_set(map_impr_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   impr_kind_set=impr_kind_set, impr_list=impr_list)
            DEALLOCATE (map_impr_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_impr

! **************************************************************************************************
!> \brief Determine the number of unique opbend kind and allocate opbend_kind_set
!>        based on the present impropers. With each improper, there also
!>        corresponds a opbend
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE force_field_unique_opbend(particle_set, &
                                        molecule_kind_set, molecule_set, ff_type)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_unique_opbend', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a2, name_atm_b, &
                                                            name_atm_b2, name_atm_c, name_atm_c2, &
                                                            name_atm_d, name_atm_d2
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, counter, &
                                                            first, handle2, i, j, k, last, natom, &
                                                            nopbend
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      INTEGER, POINTER                                   :: map_opbend_kind(:)
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(opbend_kind_type), DIMENSION(:), POINTER      :: opbend_kind_set
      TYPE(opbend_type), DIMENSION(:), POINTER           :: opbend_list

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nopbend=nopbend, opbend_list=opbend_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         IF (nopbend > 0) THEN
            ALLOCATE (map_opbend_kind(nopbend))
            counter = 0
            IF ((ff_type%ff_type == do_ff_g96) .OR. (ff_type%ff_type == do_ff_g87)) THEN
               DO j = 1, nopbend
                  map_opbend_kind(j) = j
               END DO
               counter = nopbend
            ELSE
               DO j = 1, nopbend
                  atm_a = opbend_list(j)%a
                  atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_a)
                  atm_b = opbend_list(j)%b
                  atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_b)
                  atm_c = opbend_list(j)%c
                  atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_c)
                  atm_d = opbend_list(j)%d
                  atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       name=name_atm_d)
                  found = .FALSE.
                  DO k = 1, j-1
                     atm_a = opbend_list(k)%a
                     atomic_kind => particle_set(atm_a+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_a2)
                     atm_b = opbend_list(k)%b
                     atomic_kind => particle_set(atm_b+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_b2)
                     atm_c = opbend_list(k)%c
                     atomic_kind => particle_set(atm_c+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_c2)
                     atm_d = opbend_list(k)%d
                     atomic_kind => particle_set(atm_d+first-1)%atomic_kind
                     CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                          name=name_atm_d2)
                     IF ((((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_b2)) .AND. &
                          ((name_atm_c) == (name_atm_c2)) .AND. &
                          ((name_atm_d) == (name_atm_d2))) .OR. &
                         (((name_atm_a) == (name_atm_a2)) .AND. &
                          ((name_atm_b) == (name_atm_c2)) .AND. &
                          ((name_atm_c) == (name_atm_b2)) .AND. &
                          ((name_atm_d) == (name_atm_d2)))) THEN
                        found = .TRUE.
                        map_opbend_kind(j) = map_opbend_kind(k)
                        EXIT
                     END IF
                  END DO
                  IF (.NOT. found) THEN
                     counter = counter+1
                     map_opbend_kind(j) = counter
                  END IF
               END DO
            END IF
            NULLIFY (opbend_kind_set)
            CALL allocate_opbend_kind_set(opbend_kind_set, counter)
            DO j = 1, nopbend
               opbend_list(j)%opbend_kind => opbend_kind_set(map_opbend_kind(j))
            END DO
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   opbend_kind_set=opbend_kind_set, opbend_list=opbend_list)
            DEALLOCATE (map_opbend_kind)
         END IF
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_unique_opbend

! **************************************************************************************************
!> \brief Pack in bonds information needed for the force_field
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param fatal ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
!> \param amb_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_bond(particle_set, molecule_kind_set, molecule_set, &
                                    fatal, Ainfo, chm_info, inp_info, gro_info, amb_info)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      LOGICAL                                            :: fatal
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(amber_info_type), POINTER                     :: amb_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_bond', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b
      INTEGER                                            :: atm_a, atm_b, first, handle2, i, itype, &
                                                            j, k, last, natom, nbond
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)

      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nbond=nbond, bond_list=bond_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, nbond
            atm_a = bond_list(j)%a
            atomic_kind => particle_set(atm_a+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_a)
            atm_b = bond_list(j)%b
            atomic_kind => particle_set(atm_b+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_b)
            found = .FALSE.
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)

            ! loop over params from GROMOS
            IF (ASSOCIATED(gro_info%bond_k)) THEN
               k = SIZE(gro_info%bond_k)
               itype = bond_list(j)%itype
               IF (itype <= k) THEN
                  bond_list(j)%bond_kind%k(1) = gro_info%bond_k(itype)
                  bond_list(j)%bond_kind%r0 = gro_info%bond_r0(itype)
               ELSE
                  itype = itype-k
                  bond_list(j)%bond_kind%k(1) = gro_info%solvent_k(itype)
                  bond_list(j)%bond_kind%r0 = gro_info%solvent_r0(itype)
               END IF
               bond_list(j)%bond_kind%id_type = gro_info%ff_gromos_type
               bond_list(j)%id_type = gro_info%ff_gromos_type
               found = .TRUE.
            END IF

            ! loop over params from CHARMM
            IF (ASSOCIATED(chm_info%bond_a)) THEN
               DO k = 1, SIZE(chm_info%bond_a)
                  IF ((((chm_info%bond_a(k)) == (name_atm_a)) .AND. &
                       ((chm_info%bond_b(k)) == (name_atm_b))) .OR. &
                      (((chm_info%bond_a(k)) == (name_atm_b)) .AND. &
                       ((chm_info%bond_b(k)) == (name_atm_a)))) THEN
                     bond_list(j)%bond_kind%id_type = do_ff_charmm
                     bond_list(j)%bond_kind%k(1) = chm_info%bond_k(k)
                     bond_list(j)%bond_kind%r0 = chm_info%bond_r0(k)
                     CALL issue_duplications(found, "Bond", name_atm_a, name_atm_b)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            ! loop over params from AMBER
            IF (ASSOCIATED(amb_info%bond_a)) THEN
               DO k = 1, SIZE(amb_info%bond_a)
                  IF ((((amb_info%bond_a(k)) == (name_atm_a)) .AND. &
                       ((amb_info%bond_b(k)) == (name_atm_b))) .OR. &
                      (((amb_info%bond_a(k)) == (name_atm_b)) .AND. &
                       ((amb_info%bond_b(k)) == (name_atm_a)))) THEN
                     bond_list(j)%bond_kind%id_type = do_ff_amber
                     bond_list(j)%bond_kind%k(1) = amb_info%bond_k(k)
                     bond_list(j)%bond_kind%r0 = amb_info%bond_r0(k)
                     CALL issue_duplications(found, "Bond", name_atm_a, name_atm_b)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%bond_a)) THEN
               DO k = 1, SIZE(inp_info%bond_a)
                  IF ((((inp_info%bond_a(k)) == (name_atm_a)) .AND. &
                       ((inp_info%bond_b(k)) == (name_atm_b))) .OR. &
                      (((inp_info%bond_a(k)) == (name_atm_b)) .AND. &
                       ((inp_info%bond_b(k)) == (name_atm_a)))) THEN
                     bond_list(j)%bond_kind%id_type = inp_info%bond_kind(k)
                     bond_list(j)%bond_kind%k(:) = inp_info%bond_k(:, k)
                     bond_list(j)%bond_kind%r0 = inp_info%bond_r0(k)
                     bond_list(j)%bond_kind%cs = inp_info%bond_cs(k)
                     CALL issue_duplications(found, "Bond", name_atm_a, name_atm_b)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            IF (.NOT. found) CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                                       atm2=TRIM(name_atm_b), &
                                                       fatal=fatal, &
                                                       type_name="Bond", &
                                                       array=Ainfo)
            ! QM/MM modifications
            IF (only_qm) THEN
               bond_list(j)%id_type = do_ff_undef
               bond_list(j)%bond_kind%id_type = do_ff_undef
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                bond_list=bond_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_bond

! **************************************************************************************************
!> \brief Pack in bends information needed for the force_field
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param fatal ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
!> \param amb_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_bend(particle_set, molecule_kind_set, molecule_set, &
                                    fatal, Ainfo, chm_info, inp_info, gro_info, amb_info)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      LOGICAL                                            :: fatal
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(amber_info_type), POINTER                     :: amb_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_bend', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c
      INTEGER                                            :: atm_a, atm_b, atm_c, first, handle2, i, &
                                                            itype, j, k, l, last, natom, nbend
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(bend_type), DIMENSION(:), POINTER             :: bend_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)

      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nbend=nbend, bend_list=bend_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, nbend
            atm_a = bend_list(j)%a
            atomic_kind => particle_set(atm_a+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_a)
            atm_b = bend_list(j)%b
            atomic_kind => particle_set(atm_b+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_b)
            atm_c = bend_list(j)%c
            atomic_kind => particle_set(atm_c+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_c)
            found = .FALSE.
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b, id3=name_atm_c)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            CALL uppercase(name_atm_c)

            ! loop over params from GROMOS
            IF (ASSOCIATED(gro_info%bend_k)) THEN
               k = SIZE(gro_info%bend_k)
               itype = bend_list(j)%itype
               IF (itype > 0) THEN
                  bend_list(j)%bend_kind%k = gro_info%bend_k(itype)
                  bend_list(j)%bend_kind%theta0 = gro_info%bend_theta0(itype)
               ELSE
                  bend_list(j)%bend_kind%k = gro_info%bend_k(itype/k)
                  bend_list(j)%bend_kind%theta0 = gro_info%bend_theta0(itype/k)
               END IF
               bend_list(j)%bend_kind%id_type = gro_info%ff_gromos_type
               bend_list(j)%id_type = gro_info%ff_gromos_type
               found = .TRUE.
            END IF

            ! loop over params from CHARMM
            IF (ASSOCIATED(chm_info%bend_a)) THEN
               DO k = 1, SIZE(chm_info%bend_a)
                  IF ((((chm_info%bend_a(k)) == (name_atm_a)) .AND. &
                       ((chm_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((chm_info%bend_c(k)) == (name_atm_c))) .OR. &
                      (((chm_info%bend_a(k)) == (name_atm_c)) .AND. &
                       ((chm_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((chm_info%bend_c(k)) == (name_atm_a)))) THEN
                     bend_list(j)%bend_kind%id_type = do_ff_charmm
                     bend_list(j)%bend_kind%k = chm_info%bend_k(k)
                     bend_list(j)%bend_kind%theta0 = chm_info%bend_theta0(k)
                     CALL issue_duplications(found, "Bend", name_atm_a, name_atm_b, &
                                             name_atm_c)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            ! loop over params from AMBER
            IF (ASSOCIATED(amb_info%bend_a)) THEN
               DO k = 1, SIZE(amb_info%bend_a)
                  IF ((((amb_info%bend_a(k)) == (name_atm_a)) .AND. &
                       ((amb_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((amb_info%bend_c(k)) == (name_atm_c))) .OR. &
                      (((amb_info%bend_a(k)) == (name_atm_c)) .AND. &
                       ((amb_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((amb_info%bend_c(k)) == (name_atm_a)))) THEN
                     bend_list(j)%bend_kind%id_type = do_ff_amber
                     bend_list(j)%bend_kind%k = amb_info%bend_k(k)
                     bend_list(j)%bend_kind%theta0 = amb_info%bend_theta0(k)
                     CALL issue_duplications(found, "Bend", name_atm_a, name_atm_b, &
                                             name_atm_c)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%bend_a)) THEN
               DO k = 1, SIZE(inp_info%bend_a)
                  IF ((((inp_info%bend_a(k)) == (name_atm_a)) .AND. &
                       ((inp_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((inp_info%bend_c(k)) == (name_atm_c))) .OR. &
                      (((inp_info%bend_a(k)) == (name_atm_c)) .AND. &
                       ((inp_info%bend_b(k)) == (name_atm_b)) .AND. &
                       ((inp_info%bend_c(k)) == (name_atm_a)))) THEN
                     bend_list(j)%bend_kind%id_type = inp_info%bend_kind(k)
                     bend_list(j)%bend_kind%k = inp_info%bend_k(k)
                     bend_list(j)%bend_kind%theta0 = inp_info%bend_theta0(k)
                     bend_list(j)%bend_kind%cb = inp_info%bend_cb(k)
                     bend_list(j)%bend_kind%r012 = inp_info%bend_r012(k)
                     bend_list(j)%bend_kind%r032 = inp_info%bend_r032(k)
                     bend_list(j)%bend_kind%kbs12 = inp_info%bend_kbs12(k)
                     bend_list(j)%bend_kind%kbs32 = inp_info%bend_kbs32(k)
                     bend_list(j)%bend_kind%kss = inp_info%bend_kss(k)
                     bend_list(j)%bend_kind%legendre%order = inp_info%bend_legendre(k)%order
                     IF (bend_list(j)%bend_kind%legendre%order /= 0) THEN
                        IF (ASSOCIATED(bend_list(j)%bend_kind%legendre%coeffs)) THEN
                           DEALLOCATE (bend_list(j)%bend_kind%legendre%coeffs)
                        END IF
                        ALLOCATE (bend_list(j)%bend_kind%legendre%coeffs(bend_list(j)%bend_kind%legendre%order))
                        DO l = 1, bend_list(j)%bend_kind%legendre%order
                           bend_list(j)%bend_kind%legendre%coeffs(l) = inp_info%bend_legendre(k)%coeffs(l)
                        END DO
                     END IF
                     CALL issue_duplications(found, "Bend", name_atm_a, name_atm_b, &
                                             name_atm_c)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            IF (.NOT. found) CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                                       atm2=TRIM(name_atm_b), &
                                                       atm3=TRIM(name_atm_c), &
                                                       fatal=fatal, &
                                                       type_name="Angle", &
                                                       array=Ainfo)
            ! QM/MM modifications
            IF (only_qm) THEN
               bend_list(j)%id_type = do_ff_undef
               bend_list(j)%bend_kind%id_type = do_ff_undef
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                bend_list=bend_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_bend

! **************************************************************************************************
!> \brief Pack in Urey-Bradley information needed for the force_field
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_ub(particle_set, molecule_kind_set, molecule_set, &
                                  Ainfo, chm_info, inp_info, iw)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      INTEGER                                            :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_ub', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c
      INTEGER                                            :: atm_a, atm_b, atm_c, first, handle2, i, &
                                                            j, k, last, natom, nub
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(ub_type), DIMENSION(:), POINTER               :: ub_list

      CALL timeset(routineN, handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nub=nub, ub_list=ub_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, nub
            atm_a = ub_list(j)%a
            atomic_kind => particle_set(atm_a+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_a)
            atm_b = ub_list(j)%b
            atomic_kind => particle_set(atm_b+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_b)
            atm_c = ub_list(j)%c
            atomic_kind => particle_set(atm_c+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_c)
            found = .FALSE.
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b, id3=name_atm_c)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            CALL uppercase(name_atm_c)

            ! loop over params from GROMOS
            ! ikuo - None that I know...

            ! loop over params from CHARMM
            IF (ASSOCIATED(chm_info%ub_a)) THEN
               DO k = 1, SIZE(chm_info%ub_a)
                  IF ((((chm_info%ub_a(k)) == (name_atm_a)) .AND. &
                       ((chm_info%ub_b(k)) == (name_atm_b)) .AND. &
                       ((chm_info%ub_c(k)) == (name_atm_c))) .OR. &
                      (((chm_info%ub_a(k)) == (name_atm_c)) .AND. &
                       ((chm_info%ub_b(k)) == (name_atm_b)) .AND. &
                       ((chm_info%ub_c(k)) == (name_atm_a)))) THEN
                     ub_list(j)%ub_kind%id_type = do_ff_charmm
                     ub_list(j)%ub_kind%k(1) = chm_info%ub_k(k)
                     ub_list(j)%ub_kind%r0 = chm_info%ub_r0(k)
                     IF (iw > 0) WRITE (iw, *) "    Found UB ", TRIM(name_atm_a), " ", &
                        TRIM(name_atm_b), " ", TRIM(name_atm_c)
                     CALL issue_duplications(found, "Urey-Bradley", name_atm_a, &
                                             name_atm_b, name_atm_c)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            ! loop over params from AMBER
            ! teo - None that I know...

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%ub_a)) THEN
               DO k = 1, SIZE(inp_info%ub_a)
                  IF ((((inp_info%ub_a(k)) == (name_atm_a)) .AND. &
                       ((inp_info%ub_b(k)) == (name_atm_b)) .AND. &
                       ((inp_info%ub_c(k)) == (name_atm_c))) .OR. &
                      (((inp_info%ub_a(k)) == (name_atm_c)) .AND. &
                       ((inp_info%ub_b(k)) == (name_atm_b)) .AND. &
                       ((inp_info%ub_c(k)) == (name_atm_a)))) THEN
                     ub_list(j)%ub_kind%id_type = inp_info%ub_kind(k)
                     ub_list(j)%ub_kind%k(:) = inp_info%ub_k(:, k)
                     ub_list(j)%ub_kind%r0 = inp_info%ub_r0(k)
                     CALL issue_duplications(found, "Urey-Bradley", name_atm_a, &
                                             name_atm_b, name_atm_c)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            IF (.NOT. found) THEN
               CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                         atm2=TRIM(name_atm_b), &
                                         atm3=TRIM(name_atm_c), &
                                         type_name="Urey-Bradley", &
                                         array=Ainfo)
               ub_list(j)%id_type = do_ff_undef
               ub_list(j)%ub_kind%id_type = do_ff_undef
               ub_list(j)%ub_kind%k = 0.0_dp
               ub_list(j)%ub_kind%r0 = 0.0_dp
            END IF

            ! QM/MM modifications
            IF (only_qm) THEN
               ub_list(j)%id_type = do_ff_undef
               ub_list(j)%ub_kind%id_type = do_ff_undef
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                ub_list=ub_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_ub

! **************************************************************************************************
!> \brief Pack in torsion information needed for the force_field
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
!> \param amb_info ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_tors(particle_set, molecule_kind_set, molecule_set, &
                                    Ainfo, chm_info, inp_info, gro_info, amb_info, iw)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(amber_info_type), POINTER                     :: amb_info
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_tors', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: ldum, name_atm_a, name_atm_b, &
                                                            name_atm_c, name_atm_d
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, first, &
                                                            handle2, i, imul, itype, j, k, last, &
                                                            natom, ntorsion
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(torsion_type), DIMENSION(:), POINTER          :: torsion_list

      CALL timeset(routineN, handle2)

      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                ntorsion=ntorsion, torsion_list=torsion_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, ntorsion
            IF (torsion_list(j)%torsion_kind%id_type == do_ff_undef) THEN
               atm_a = torsion_list(j)%a
               atomic_kind => particle_set(atm_a+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_a)
               atm_b = torsion_list(j)%b
               atomic_kind => particle_set(atm_b+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_b)
               atm_c = torsion_list(j)%c
               atomic_kind => particle_set(atm_c+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_c)
               atm_d = torsion_list(j)%d
               atomic_kind => particle_set(atm_d+first-1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=name_atm_d)
               found = .FALSE.
               only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b, id3=name_atm_c, id4=name_atm_d)
               CALL uppercase(name_atm_a)
               CALL uppercase(name_atm_b)
               CALL uppercase(name_atm_c)
               CALL uppercase(name_atm_d)

               ! loop over params from GROMOS
               IF (ASSOCIATED(gro_info%torsion_k)) THEN
                  k = SIZE(gro_info%torsion_k)
                  itype = torsion_list(j)%itype
                  IF (itype > 0) THEN
                     CALL reallocate(torsion_list(j)%torsion_kind%k, 1, 1)
                     CALL reallocate(torsion_list(j)%torsion_kind%m, 1, 1)
                     CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, 1)
                     torsion_list(j)%torsion_kind%nmul = 1
                     torsion_list(j)%torsion_kind%m(1) = gro_info%torsion_m(itype)
                     torsion_list(j)%torsion_kind%k(1) = gro_info%torsion_k(itype)
                     torsion_list(j)%torsion_kind%phi0(1) = gro_info%torsion_phi0(itype)
                  ELSE
                     CALL reallocate(torsion_list(j)%torsion_kind%k, 1, 1)
                     CALL reallocate(torsion_list(j)%torsion_kind%m, 1, 1)
                     CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, 1)
                     torsion_list(j)%torsion_kind%nmul = 1
                     torsion_list(j)%torsion_kind%m(1) = gro_info%torsion_m(itype/k)
                     torsion_list(j)%torsion_kind%k(1) = gro_info%torsion_k(itype/k)
                     torsion_list(j)%torsion_kind%phi0(1) = gro_info%torsion_phi0(itype/k)
                  END IF
                  torsion_list(j)%torsion_kind%id_type = gro_info%ff_gromos_type
                  torsion_list(j)%id_type = gro_info%ff_gromos_type
                  found = .TRUE.
                  imul = torsion_list(j)%torsion_kind%nmul
               END IF

               ! loop over params from CHARMM
               IF (ASSOCIATED(chm_info%torsion_a)) THEN
                  DO k = 1, SIZE(chm_info%torsion_a)
                     IF ((((chm_info%torsion_a(k)) == (name_atm_a)) .AND. &
                          ((chm_info%torsion_b(k)) == (name_atm_b)) .AND. &
                          ((chm_info%torsion_c(k)) == (name_atm_c)) .AND. &
                          ((chm_info%torsion_d(k)) == (name_atm_d))) .OR. &
                         (((chm_info%torsion_a(k)) == (name_atm_d)) .AND. &
                          ((chm_info%torsion_b(k)) == (name_atm_c)) .AND. &
                          ((chm_info%torsion_c(k)) == (name_atm_b)) .AND. &
                          ((chm_info%torsion_d(k)) == (name_atm_a)))) THEN
                        imul = torsion_list(j)%torsion_kind%nmul+1
                        CALL reallocate(torsion_list(j)%torsion_kind%k, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%m, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, imul)
                        torsion_list(j)%torsion_kind%id_type = do_ff_charmm
                        torsion_list(j)%torsion_kind%k(imul) = chm_info%torsion_k(k)
                        torsion_list(j)%torsion_kind%m(imul) = chm_info%torsion_m(k)
                        torsion_list(j)%torsion_kind%phi0(imul) = chm_info%torsion_phi0(k)
                        torsion_list(j)%torsion_kind%nmul = imul
                        found = .TRUE.
                     END IF
                  END DO

                  IF (.NOT. found) THEN
                     DO k = 1, SIZE(chm_info%torsion_a)
                        IF ((((chm_info%torsion_a(k)) == ("X")) .AND. &
                             ((chm_info%torsion_b(k)) == (name_atm_b)) .AND. &
                             ((chm_info%torsion_c(k)) == (name_atm_c)) .AND. &
                             ((chm_info%torsion_d(k)) == ("X"))) .OR. &
                            (((chm_info%torsion_a(k)) == ("X")) .AND. &
                             ((chm_info%torsion_b(k)) == (name_atm_c)) .AND. &
                             ((chm_info%torsion_c(k)) == (name_atm_b)) .AND. &
                             ((chm_info%torsion_d(k)) == ("X")))) THEN
                           imul = torsion_list(j)%torsion_kind%nmul+1
                           CALL reallocate(torsion_list(j)%torsion_kind%k, 1, imul)
                           CALL reallocate(torsion_list(j)%torsion_kind%m, 1, imul)
                           CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, imul)
                           torsion_list(j)%torsion_kind%id_type = do_ff_charmm
                           torsion_list(j)%torsion_kind%k(imul) = chm_info%torsion_k(k)
                           torsion_list(j)%torsion_kind%m(imul) = chm_info%torsion_m(k)
                           torsion_list(j)%torsion_kind%phi0(imul) = chm_info%torsion_phi0(k)
                           torsion_list(j)%torsion_kind%nmul = imul
                           found = .TRUE.
                        END IF
                     END DO
                  END IF
               END IF

               ! loop over params from AMBER
               IF (ASSOCIATED(amb_info%torsion_a)) THEN
                  DO k = 1, SIZE(amb_info%torsion_a)
                     IF ((((amb_info%torsion_a(k)) == (name_atm_a)) .AND. &
                          ((amb_info%torsion_b(k)) == (name_atm_b)) .AND. &
                          ((amb_info%torsion_c(k)) == (name_atm_c)) .AND. &
                          ((amb_info%torsion_d(k)) == (name_atm_d))) .OR. &
                         (((amb_info%torsion_a(k)) == (name_atm_d)) .AND. &
                          ((amb_info%torsion_b(k)) == (name_atm_c)) .AND. &
                          ((amb_info%torsion_c(k)) == (name_atm_b)) .AND. &
                          ((amb_info%torsion_d(k)) == (name_atm_a)))) THEN
                        imul = torsion_list(j)%torsion_kind%nmul+1
                        CALL reallocate(torsion_list(j)%torsion_kind%k, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%m, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, imul)
                        torsion_list(j)%torsion_kind%id_type = do_ff_amber
                        torsion_list(j)%torsion_kind%k(imul) = amb_info%torsion_k(k)
                        torsion_list(j)%torsion_kind%m(imul) = amb_info%torsion_m(k)
                        torsion_list(j)%torsion_kind%phi0(imul) = amb_info%torsion_phi0(k)
                        torsion_list(j)%torsion_kind%nmul = imul
                        found = .TRUE.
                     END IF
                  END DO

                  IF (.NOT. found) THEN
                     DO k = 1, SIZE(amb_info%torsion_a)
                        IF ((((amb_info%torsion_a(k)) == ("X")) .AND. &
                             ((amb_info%torsion_b(k)) == (name_atm_b)) .AND. &
                             ((amb_info%torsion_c(k)) == (name_atm_c)) .AND. &
                             ((amb_info%torsion_d(k)) == ("X"))) .OR. &
                            (((amb_info%torsion_a(k)) == ("X")) .AND. &
                             ((amb_info%torsion_b(k)) == (name_atm_c)) .AND. &
                             ((amb_info%torsion_c(k)) == (name_atm_b)) .AND. &
                             ((amb_info%torsion_d(k)) == ("X")))) THEN
                           imul = torsion_list(j)%torsion_kind%nmul+1
                           CALL reallocate(torsion_list(j)%torsion_kind%k, 1, imul)
                           CALL reallocate(torsion_list(j)%torsion_kind%m, 1, imul)
                           CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, imul)
                           torsion_list(j)%torsion_kind%id_type = do_ff_amber
                           torsion_list(j)%torsion_kind%k(imul) = amb_info%torsion_k(k)
                           torsion_list(j)%torsion_kind%m(imul) = amb_info%torsion_m(k)
                           torsion_list(j)%torsion_kind%phi0(imul) = amb_info%torsion_phi0(k)
                           torsion_list(j)%torsion_kind%nmul = imul
                           found = .TRUE.
                        END IF
                     END DO
                  END IF
               END IF

               ! always have the input param last to overwrite all the other ones
               IF (ASSOCIATED(inp_info%torsion_a)) THEN
                  DO k = 1, SIZE(inp_info%torsion_a)
                     IF ((((inp_info%torsion_a(k)) == (name_atm_a)) .AND. &
                          ((inp_info%torsion_b(k)) == (name_atm_b)) .AND. &
                          ((inp_info%torsion_c(k)) == (name_atm_c)) .AND. &
                          ((inp_info%torsion_d(k)) == (name_atm_d))) .OR. &
                         (((inp_info%torsion_a(k)) == (name_atm_d)) .AND. &
                          ((inp_info%torsion_b(k)) == (name_atm_c)) .AND. &
                          ((inp_info%torsion_c(k)) == (name_atm_b)) .AND. &
                          ((inp_info%torsion_d(k)) == (name_atm_a)))) THEN
                        imul = torsion_list(j)%torsion_kind%nmul+1
                        CALL reallocate(torsion_list(j)%torsion_kind%k, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%m, 1, imul)
                        CALL reallocate(torsion_list(j)%torsion_kind%phi0, 1, imul)
                        torsion_list(j)%torsion_kind%id_type = inp_info%torsion_kind(k)
                        torsion_list(j)%torsion_kind%k(imul) = inp_info%torsion_k(k)
                        torsion_list(j)%torsion_kind%m(imul) = inp_info%torsion_m(k)
                        torsion_list(j)%torsion_kind%phi0(imul) = inp_info%torsion_phi0(k)
                        torsion_list(j)%torsion_kind%nmul = imul
                        found = .TRUE.
                     END IF
                  END DO
               END IF

               IF (.NOT. found) THEN
                  CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                            atm2=TRIM(name_atm_b), &
                                            atm3=TRIM(name_atm_c), &
                                            atm4=TRIM(name_atm_d), &
                                            type_name="Torsion", &
                                            array=Ainfo)
                  torsion_list(j)%torsion_kind%id_type = do_ff_undef
                  torsion_list(j)%id_type = do_ff_undef
               ELSE
                  ldum = cp_to_string(imul)
                  IF ((imul /= 1) .AND. (iw > 0)) &
                     WRITE (iw, '(/,2("UTIL_INFO| ",A,/))') &
                     "Multiple Torsion declarations: "//TRIM(name_atm_a)// &
                     ","//TRIM(name_atm_b)//","//TRIM(name_atm_c)//" and "//TRIM(name_atm_d), &
                     "Number of defined torsions: "//TRIM(ldum)//" ."
               END IF
               !
               ! QM/MM modifications
               !
               IF (only_qm) THEN
                  IF (iw > 0) WRITE (iw, *) "    Torsion PARAM between QM atoms ", j, " : ", &
                     TRIM(name_atm_a), " ", &
                     TRIM(name_atm_b), " ", &
                     TRIM(name_atm_c), " ", &
                     TRIM(name_atm_d), " ", &
                     torsion_list(j)%a, &
                     torsion_list(j)%b, &
                     torsion_list(j)%c, &
                     torsion_list(j)%d
                  torsion_list(j)%torsion_kind%id_type = do_ff_undef
                  torsion_list(j)%id_type = do_ff_undef
               END IF
            END IF
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                torsion_list=torsion_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_tors

! **************************************************************************************************
!> \brief Pack in impropers information needed for the force_field
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_impr(particle_set, molecule_kind_set, molecule_set, &
                                    Ainfo, chm_info, inp_info, gro_info)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_impr', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c, &
                                                            name_atm_d
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, first, &
                                                            handle2, i, itype, j, k, last, natom, &
                                                            nimpr
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(impr_type), DIMENSION(:), POINTER             :: impr_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle2)

      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nimpr=nimpr, impr_list=impr_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, nimpr
            atm_a = impr_list(j)%a
            atomic_kind => particle_set(atm_a+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_a)
            atm_b = impr_list(j)%b
            atomic_kind => particle_set(atm_b+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_b)
            atm_c = impr_list(j)%c
            atomic_kind => particle_set(atm_c+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_c)
            atm_d = impr_list(j)%d
            atomic_kind => particle_set(atm_d+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_d)
            found = .FALSE.
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b, id3=name_atm_c, id4=name_atm_d)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            CALL uppercase(name_atm_c)
            CALL uppercase(name_atm_d)

            ! loop over params from GROMOS
            IF (ASSOCIATED(gro_info%impr_k)) THEN
               k = SIZE(gro_info%impr_k)
               itype = impr_list(j)%itype
               IF (itype > 0) THEN
                  impr_list(j)%impr_kind%k = gro_info%impr_k(itype)
                  impr_list(j)%impr_kind%phi0 = gro_info%impr_phi0(itype)
               ELSE
                  impr_list(j)%impr_kind%k = gro_info%impr_k(itype)
                  impr_list(j)%impr_kind%phi0 = gro_info%impr_phi0(itype)
               END IF
               found = .TRUE.
               impr_list(j)%impr_kind%id_type = gro_info%ff_gromos_type
               impr_list(j)%id_type = gro_info%ff_gromos_type
            END IF

            ! loop over params from CHARMM
            IF (ASSOCIATED(chm_info%impr_a)) THEN
               DO k = 1, SIZE(chm_info%impr_a)
                  IF ((((chm_info%impr_a(k)) == (name_atm_a)) .AND. &
                       ((chm_info%impr_b(k)) == (name_atm_b)) .AND. &
                       ((chm_info%impr_c(k)) == (name_atm_c)) .AND. &
                       ((chm_info%impr_d(k)) == (name_atm_d))) .OR. &
                      (((chm_info%impr_a(k)) == (name_atm_d)) .AND. &
                       ((chm_info%impr_b(k)) == (name_atm_c)) .AND. &
                       ((chm_info%impr_c(k)) == (name_atm_b)) .AND. &
                       ((chm_info%impr_d(k)) == (name_atm_a)))) THEN
                     impr_list(j)%impr_kind%id_type = do_ff_charmm
                     impr_list(j)%impr_kind%k = chm_info%impr_k(k)
                     impr_list(j)%impr_kind%phi0 = chm_info%impr_phi0(k)
                     CALL issue_duplications(found, "Impropers", name_atm_a, name_atm_b, &
                                             name_atm_c, name_atm_d)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. found) THEN
                  DO k = 1, SIZE(chm_info%impr_a)
                     IF ((((chm_info%impr_a(k)) == (name_atm_a)) .AND. &
                          ((chm_info%impr_b(k)) == ("X")) .AND. &
                          ((chm_info%impr_c(k)) == ("X")) .AND. &
                          ((chm_info%impr_d(k)) == (name_atm_d))) .OR. &
                         (((chm_info%impr_a(k)) == (name_atm_d)) .AND. &
                          ((chm_info%impr_b(k)) == ("X")) .AND. &
                          ((chm_info%impr_c(k)) == ("X")) .AND. &
                          ((chm_info%impr_d(k)) == (name_atm_a)))) THEN
                        impr_list(j)%impr_kind%id_type = do_ff_charmm
                        impr_list(j)%impr_kind%k = chm_info%impr_k(k)
                        impr_list(j)%impr_kind%phi0 = chm_info%impr_phi0(k)
                        CALL issue_duplications(found, "Impropers", name_atm_a, name_atm_b, &
                                                name_atm_c, name_atm_d)
                        found = .TRUE.
                        EXIT
                     END IF
                  END DO
               END IF
            END IF

            ! loop over params from AMBER not needed since impropers in AMBER
            ! are treated like standard torsions

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%impr_a)) THEN
               DO k = 1, SIZE(inp_info%impr_a)
                  IF (((inp_info%impr_a(k)) == (name_atm_a)) .AND. &
                      ((inp_info%impr_b(k)) == (name_atm_b)) .AND. &
                      ((((inp_info%impr_c(k)) == (name_atm_c)) .AND. &
                        ((inp_info%impr_d(k)) == (name_atm_d))) .OR. &
                       (((inp_info%impr_c(k)) == (name_atm_d)) .AND. &
                        ((inp_info%impr_d(k)) == (name_atm_c))))) THEN
                     impr_list(j)%impr_kind%id_type = inp_info%impr_kind(k)
                     impr_list(j)%impr_kind%k = inp_info%impr_k(k)
                     IF (((inp_info%impr_c(k)) == (name_atm_c)) .AND. &
                         ((inp_info%impr_d(k)) == (name_atm_d))) THEN
                        impr_list(j)%impr_kind%phi0 = inp_info%impr_phi0(k)
                     ELSE
                        impr_list(j)%impr_kind%phi0 = -inp_info%impr_phi0(k)
                        ! alternative solutions:
                        !  - swap impr_list(j)%c with impr_list(j)%d and
                        !    name_atom_c with name_atom_d and
                        !    atm_c with atm_d
                        !  - introduce impr_list(j)%impr_kind%sign. if one, the
                        !    sign of phi is not changed in mol_force.f90. if minus
                        !    one, the sign of phi is changed in mol_force.f90
                        ! similar problems with parameters from charmm pot file
                        ! above
                     END IF
                     CALL issue_duplications(found, "Impropers", name_atm_a, name_atm_b, &
                                             name_atm_c, name_atm_d)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            IF (.NOT. found) THEN
               CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                         atm2=TRIM(name_atm_b), &
                                         atm3=TRIM(name_atm_c), &
                                         atm4=TRIM(name_atm_d), &
                                         type_name="Improper", &
                                         array=Ainfo)
               impr_list(j)%impr_kind%k = 0.0_dp
               impr_list(j)%impr_kind%phi0 = 0.0_dp
               impr_list(j)%impr_kind%id_type = do_ff_undef
               impr_list(j)%id_type = do_ff_undef
            END IF
            !
            ! QM/MM modifications
            !
            IF (only_qm) THEN
               impr_list(j)%impr_kind%id_type = do_ff_undef
               impr_list(j)%id_type = do_ff_undef
            END IF

         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, impr_list=impr_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_impr

! **************************************************************************************************
!> \brief Pack in opbend information needed for the force_field.
!>        No loop over params for charmm, amber and gromos since these force
!>        fields have no opbends
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param Ainfo ...
!> \param inp_info ...
!> \author Louis Vanduyfhuys
! **************************************************************************************************
   SUBROUTINE force_field_pack_opbend(particle_set, molecule_kind_set, molecule_set, &
                                      Ainfo, inp_info)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_opbend', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c, &
                                                            name_atm_d
      INTEGER                                            :: atm_a, atm_b, atm_c, atm_d, first, &
                                                            handle2, i, j, k, last, natom, nopbend
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: found, only_qm
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(opbend_type), DIMENSION(:), POINTER           :: opbend_list

      CALL timeset(routineN, handle2)

      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                molecule_list=molecule_list, &
                                natom=natom, &
                                nopbend=nopbend, opbend_list=opbend_list)
         molecule => molecule_set(molecule_list(1))

         CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
         DO j = 1, nopbend
            atm_a = opbend_list(j)%a
            atomic_kind => particle_set(atm_a+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_a)
            atm_b = opbend_list(j)%b
            atomic_kind => particle_set(atm_b+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_b)
            atm_c = opbend_list(j)%c
            atomic_kind => particle_set(atm_c+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_c)
            atm_d = opbend_list(j)%d
            atomic_kind => particle_set(atm_d+first-1)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=name_atm_d)
            found = .FALSE.
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b, id3=name_atm_c, id4=name_atm_d)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            CALL uppercase(name_atm_c)
            CALL uppercase(name_atm_d)

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%opbend_a)) THEN
               DO k = 1, SIZE(inp_info%opbend_a)
                  IF (((inp_info%opbend_a(k)) == (name_atm_a)) .AND. &
                      ((inp_info%opbend_d(k)) == (name_atm_d)) .AND. &
                      ((((inp_info%opbend_c(k)) == (name_atm_c)) .AND. &
                        ((inp_info%opbend_b(k)) == (name_atm_b))) .OR. &
                       (((inp_info%opbend_c(k)) == (name_atm_b)) .AND. &
                        ((inp_info%opbend_b(k)) == (name_atm_c))))) THEN
                     opbend_list(j)%opbend_kind%id_type = inp_info%opbend_kind(k)
                     opbend_list(j)%opbend_kind%k = inp_info%opbend_k(k)
                     IF (((inp_info%opbend_c(k)) == (name_atm_c)) .AND. &
                         ((inp_info%opbend_b(k)) == (name_atm_b))) THEN
                        opbend_list(j)%opbend_kind%phi0 = inp_info%opbend_phi0(k)
                     ELSE
                        opbend_list(j)%opbend_kind%phi0 = -inp_info%opbend_phi0(k)
                        ! alternative solutions:
                        !  - swap opbend_list(j)%c with opbend_list(j)%b and
                        !    name_atom_c with name_atom_b and
                        !    atm_c with atm_b
                        !  - introduce opbend_list(j)%opbend_kind%sign. if one, the
                        !    sign of phi is not changed in mol_force.f90. if minus
                        !    one, the sign of phi is changed in mol_force.f90

                     END IF
                     CALL issue_duplications(found, "Out of plane bend", name_atm_a, name_atm_b, &
                                             name_atm_c, name_atm_d)
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF

            IF (.NOT. found) THEN
               CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                         atm2=TRIM(name_atm_b), &
                                         atm3=TRIM(name_atm_c), &
                                         atm4=TRIM(name_atm_d), &
                                         type_name="Out of plane bend", &
                                         array=Ainfo)
               opbend_list(j)%opbend_kind%k = 0.0_dp
               opbend_list(j)%opbend_kind%phi0 = 0.0_dp
               opbend_list(j)%opbend_kind%id_type = do_ff_undef
               opbend_list(j)%id_type = do_ff_undef
            END IF
            !
            ! QM/MM modifications
            !
            IF (only_qm) THEN
               opbend_list(j)%opbend_kind%id_type = do_ff_undef
               opbend_list(j)%id_type = do_ff_undef
            END IF

         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, opbend_list=opbend_list)
      END DO
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_opbend

! **************************************************************************************************
!> \brief Set up array of full charges
!> \param charges ...
!> \param charges_section ...
!> \param particle_set ...
!> \param my_qmmm ...
!> \param qmmm_env ...
!> \param inp_info ...
!> \param iw4 ...
!> \date 12.2010
!> \author Teodoro Laino (teodoro.laino@gmail.com)
! **************************************************************************************************
   SUBROUTINE force_field_pack_charges(charges, charges_section, particle_set, &
                                       my_qmmm, qmmm_env, inp_info, iw4)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      TYPE(section_vals_type), POINTER                   :: charges_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      LOGICAL                                            :: my_qmmm
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      TYPE(input_info_type), POINTER                     :: inp_info
      INTEGER                                            :: iw4

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atmname
      INTEGER                                            :: handle, iatom, ilink, j, nval
      LOGICAL                                            :: found_p, is_link_atom, is_ok, &
                                                            only_manybody, only_qm
      REAL(KIND=dp)                                      :: charge, charge_tot, rval, scale_factor
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(val_type), POINTER                            :: val

      CALL timeset(routineN, handle)
      charge_tot = 0.0_dp
      NULLIFY (list)

      ! Not implemented for core-shell
      IF (ASSOCIATED(inp_info%shell_list)) THEN
         CALL cp__b("force_fields_all.F",1875,"Array of charges not implemented for CORE-SHELL model!!")
      END IF

      ! Allocate array to particle_set size
      IF(.NOT.(.NOT. (ASSOCIATED(charges))))CALL cp__a("force_fields_all.F",1879)
      ALLOCATE (charges(SIZE(particle_set)))

      ! Fill with input values
      CALL section_vals_val_get(charges_section, "_DEFAULT_KEYWORD_", n_rep_val=nval)
      IF(.NOT.(nval == SIZE(charges)))CALL cp__a("force_fields_all.F",1884)
      CALL section_vals_list_get(charges_section, "_DEFAULT_KEYWORD_", list=list)
      DO iatom = 1, nval
         ! we use only the first default_string_length characters of each line
         is_ok = cp_sll_val_next(list, val)
         CALL val_get(val, r_val=rval)
         ! assign values
         charges(iatom) = rval

         ! Perform a post-processing
         atomic_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              fist_potential=fist_potential, &
                              name=atmname)
         CALL get_potential(potential=fist_potential, qeff=charge)

         only_qm = qmmm_ff_precond_only_qm(id1=atmname, is_link=is_link_atom)
         CALL uppercase(atmname)
         IF (charge /= -HUGE(0.0_dp)) &
            CALL cp_warn(cp__l("force_fields_all.F",1903), &
                         "The charge for atom index ("//cp_to_string(iatom)//") and atom name ("// &
                         TRIM(atmname)//") was already defined. The charge associated to this kind"// &
                         " will be set to an uninitialized value and only the atom specific charge will be used! ")
         charge = -HUGE(0.0_dp)

         ! Check if the potential really requires the charge definition..
         IF (ASSOCIATED(inp_info%nonbonded)) THEN
            IF (ASSOCIATED(inp_info%nonbonded%pot)) THEN
               ! Let's find the nonbonded potential where this atom is involved
               only_manybody = .TRUE.
               found_p = .FALSE.
               DO j = 1, SIZE(inp_info%nonbonded%pot)
                  IF (atmname == inp_info%nonbonded%pot(j)%pot%at1 .OR. &
                      atmname == inp_info%nonbonded%pot(j)%pot%at2) THEN
                     SELECT CASE (inp_info%nonbonded%pot (j)%pot%type (1))
                     CASE (ea_type, tersoff_type, siepmann_type)
                        ! Charge is zero for EAM, TERSOFF and SIEPMANN  type potential
                        ! Do nothing..
                     CASE DEFAULT
                        only_manybody = .FALSE.
                        EXIT
                     END SELECT
                     found_p = .TRUE.
                  END IF
               END DO
               IF (only_manybody .AND. found_p) THEN
                  charges(iatom) = 0.0_dp
               END IF
            END IF
         END IF
         !
         ! QM/MM modifications
         !
         IF (only_qm .AND. my_qmmm) THEN
            IF (qmmm_env%qmmm_coupl_type /= do_qmmm_none) THEN
               scale_factor = 0.0_dp
               IF (is_link_atom) THEN
                  !
                  ! Find the scaling factor...
                  !
                  DO ilink = 1, SIZE(qmmm_env%mm_link_atoms)
                     IF (iatom == qmmm_env%mm_link_atoms(ilink)) EXIT
                  END DO
                  IF(.NOT.(ilink <= SIZE(qmmm_env%mm_link_atoms)))CALL cp__a("force_fields_all.F",1947)
                  scale_factor = qmmm_env%fist_scale_charge_link(ilink)
               END IF
               charges(iatom) = charges(iatom)*scale_factor
            END IF
         END IF
      END DO
      ! Sum up total charges for IO
      charge_tot = SUM(charges)
      ! Print Total Charge of the system
      IF (iw4 > 0) THEN
         WRITE (iw4, FMT='(T2,"CHARGE_INFO| Total Charge of the Classical System: ",T69,F12.6)') charge_tot
      END IF
      CALL timestop(handle)
   END SUBROUTINE force_field_pack_charges

! **************************************************************************************************
!> \brief Set up atomic_kind_set()%fist_potentail%[qeff]
!>      and shell potential parameters
!> \param atomic_kind_set ...
!> \param qmmm_env ...
!> \param fatal ...
!> \param iw ...
!> \param iw4 ...
!> \param Ainfo ...
!> \param my_qmmm ...
!> \param inp_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_charge(atomic_kind_set, qmmm_env, fatal, iw, iw4, &
                                      Ainfo, my_qmmm, inp_info)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      LOGICAL                                            :: fatal
      INTEGER                                            :: iw, iw4
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      LOGICAL                                            :: my_qmmm
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_charge', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atmname
      INTEGER                                            :: handle, i, ilink, j
      INTEGER, DIMENSION(:), POINTER                     :: my_atom_list
      LOGICAL                                            :: found, found_p, is_link_atom, is_shell, &
                                                            only_manybody, only_qm
      REAL(KIND=dp)                                      :: charge, charge_tot, cs_charge, &
                                                            scale_factor
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fist_potential_type), POINTER                 :: fist_potential

      CALL timeset(routineN, handle)
      charge_tot = 0.0_dp
      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              fist_potential=fist_potential, &
                              atom_list=my_atom_list, &
                              name=atmname)
         CALL get_potential(potential=fist_potential, qeff=charge)

         is_shell = .FALSE.
         found = .FALSE.
         only_qm = qmmm_ff_precond_only_qm(id1=atmname, is_link=is_link_atom)
         CALL uppercase(atmname)
         IF (charge /= -HUGE(0.0_dp)) found = .TRUE.

         ! Always have the input param last to overwrite all the other ones
         IF (ASSOCIATED(inp_info%charge_atm)) THEN
            DO j = 1, SIZE(inp_info%charge_atm)
               IF (iw > 0) WRITE (iw, *) "Charge Checking ::", TRIM(inp_info%charge_atm(j)), atmname
               IF ((inp_info%charge_atm(j)) == atmname) THEN
                  charge = inp_info%charge(j)
                  CALL issue_duplications(found, "Charge", atmname)
                  found = .TRUE.
               END IF
            END DO
         END IF
         ! Check if the ATOM type has a core-shell associated.. In this case
         ! print a warning: the CHARGE will not be used if defined.. or we can
         ! even skip its definition..
         IF (ASSOCIATED(inp_info%shell_list)) THEN
            DO j = 1, SIZE(inp_info%shell_list)
               IF ((inp_info%shell_list(j)%atm_name) == atmname) THEN
                  is_shell = .TRUE.
                  cs_charge = inp_info%shell_list(j)%shell%charge_core+ &
                              inp_info%shell_list(j)%shell%charge_shell
                  charge = 0.0_dp
                  IF (found) THEN
                     IF (found) &
                        CALL cp_warn(cp__l("force_fields_all.F",2039), &
                                     "CORE-SHELL model defined for KIND ("//TRIM(atmname)//")"// &
                                     " ignoring charge definition! ")
                  ELSE
                     found = .TRUE.
                  END IF
               END IF
            END DO
         END IF
         ! Check if the potential really requires the charge definition..
         IF (ASSOCIATED(inp_info%nonbonded)) THEN
            IF (ASSOCIATED(inp_info%nonbonded%pot)) THEN
               ! Let's find the nonbonded potential where this atom is involved
               only_manybody = .TRUE.
               found_p = .FALSE.
               DO j = 1, SIZE(inp_info%nonbonded%pot)
                  IF (atmname == inp_info%nonbonded%pot(j)%pot%at1 .OR. &
                      atmname == inp_info%nonbonded%pot(j)%pot%at2) THEN
                     SELECT CASE (inp_info%nonbonded%pot (j)%pot%type (1))
                     CASE (ea_type, tersoff_type, siepmann_type, quip_type)
                        ! Charge is zero for EAM, TERSOFF and SIEPMANN type potential
                        ! Do nothing..
                     CASE DEFAULT
                        only_manybody = .FALSE.
                        EXIT
                     END SELECT
                     found_p = .TRUE.
                  END IF
               END DO
               IF (only_manybody .AND. found_p) THEN
                  charge = 0.0_dp
                  found = .TRUE.
               END IF
            END IF
         END IF
         IF (.NOT. found) THEN
            ! Set the charge to zero anyway in case the user decides to ignore
            ! missing critical parameters.
            charge = 0.0_dp
            CALL store_FF_missing_par(atm1=TRIM(atmname), &
                                      fatal=fatal, &
                                      type_name="Charge", &
                                      array=Ainfo)
         END IF
         !
         ! QM/MM modifications
         !
         IF (only_qm .AND. my_qmmm) THEN
            IF (qmmm_env%qmmm_coupl_type /= do_qmmm_none) THEN
               scale_factor = 0.0_dp
               IF (is_link_atom) THEN
                  !
                  ! Find the scaling factor...
                  !
                  DO ilink = 1, SIZE(qmmm_env%mm_link_atoms)
                     IF (ANY(my_atom_list == qmmm_env%mm_link_atoms(ilink))) EXIT
                  END DO
                  IF(.NOT.(ilink <= SIZE(qmmm_env%mm_link_atoms)))CALL cp__a("force_fields_all.F",2096)
                  scale_factor = qmmm_env%fist_scale_charge_link(ilink)
               END IF
               charge = charge*scale_factor
            END IF
         END IF

         CALL set_potential(potential=fist_potential, qeff=charge)
         ! Sum up total charges for IO
         IF (found) THEN
            IF (is_shell) THEN
               charge_tot = charge_tot+atomic_kind%natom*cs_charge
            ELSE
               charge_tot = charge_tot+atomic_kind%natom*charge
            END IF
         END IF
      END DO
      ! Print Total Charge of the system
      IF (iw4 > 0) THEN
         WRITE (iw4, FMT='(T2,"CHARGE_INFO| Total Charge of the Classical System: ",T69,F12.6)') charge_tot
      END IF
      CALL timestop(handle)
   END SUBROUTINE force_field_pack_charge

! **************************************************************************************************
!> \brief Set up the radius of the electrostatic multipole in Fist
!> \param atomic_kind_set ...
!> \param iw ...
!> \param subsys_section ...
!> \author Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE force_field_pack_radius(atomic_kind_set, iw, subsys_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER, INTENT(IN)                                :: iw
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_radius', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: inp_kind_name, kind_name
      INTEGER                                            :: handle, i, i_rep, n_rep
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: mm_radius
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(section_vals_type), POINTER                   :: kind_section

      CALL timeset(routineN, handle)

      kind_section => section_vals_get_subs_vals(subsys_section, "KIND")
      CALL section_vals_get(kind_section, n_repetition=n_rep)

      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              fist_potential=fist_potential, name=kind_name)
         CALL uppercase(kind_name)
         found = .FALSE.

         ! Try to find a matching KIND section in the SUBSYS section and read the
         ! MM_RADIUS field if it is present. In case the kind section is never
         ! encountered, the mm_radius remains zero.
         mm_radius = 0.0_dp
         DO i_rep = 1, n_rep
            CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                      c_val=inp_kind_name, i_rep_section=i_rep)
            CALL uppercase(inp_kind_name)
            IF (iw > 0) WRITE (iw, *) "Matching kinds for MM_RADIUS :: '", &
               TRIM(kind_name), "' with '", TRIM(inp_kind_name), "'"
            IF (TRIM(kind_name) == TRIM(inp_kind_name)) THEN
               CALL section_vals_val_get(kind_section, i_rep_section=i_rep, &
                                         keyword_name="MM_RADIUS", r_val=mm_radius)
               CALL issue_duplications(found, "MM_RADIUS", kind_name)
               found = .TRUE.
            END IF
         END DO

         CALL set_potential(potential=fist_potential, mm_radius=mm_radius)
      END DO
      CALL timestop(handle)
   END SUBROUTINE force_field_pack_radius

! **************************************************************************************************
!> \brief Set up the polarizable FF parameters
!> \param atomic_kind_set ...
!> \param iw ...
!> \param inp_info ...
!> \author Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE force_field_pack_pol(atomic_kind_set, iw, inp_info)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER, INTENT(IN)                                :: iw
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_pol', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: kind_name
      INTEGER                                            :: handle, i, j
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: apol, cpol
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fist_potential_type), POINTER                 :: fist_potential

      CALL timeset(routineN, handle)

      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              fist_potential=fist_potential, &
                              name=kind_name)
         CALL get_potential(potential=fist_potential, apol=apol, cpol=cpol)
         CALL uppercase(kind_name)
         found = .FALSE.

         ! Always have the input param last to overwrite all the other ones
         IF (ASSOCIATED(inp_info%apol_atm)) THEN
            DO j = 1, SIZE(inp_info%apol_atm)
               IF (iw > 0) WRITE (iw, *) "Matching kinds for APOL :: '", &
                  TRIM(kind_name), "' with '", TRIM(inp_info%apol_atm(j)), "'"
               IF ((inp_info%apol_atm(j)) == kind_name) THEN
                  apol = inp_info%apol(j)
                  CALL issue_duplications(found, "APOL", kind_name)
                  found = .TRUE.
               END IF
            END DO
         END IF

         IF (ASSOCIATED(inp_info%cpol_atm)) THEN
            DO j = 1, SIZE(inp_info%cpol_atm)
               IF (iw > 0) WRITE (iw, *) "Matching kinds for CPOL :: '", &
                  TRIM(kind_name), "' with '", TRIM(inp_info%cpol_atm(j)), "'"
               IF ((inp_info%cpol_atm(j)) == kind_name) THEN
                  cpol = inp_info%cpol(j)
                  CALL issue_duplications(found, "CPOL", kind_name)
                  found = .TRUE.
               END IF
            END DO
         END IF

         CALL set_potential(potential=fist_potential, apol=apol, cpol=cpol)
      END DO
      CALL timestop(handle)
   END SUBROUTINE force_field_pack_pol

! **************************************************************************************************
!> \brief Set up damping parameters
!> \param atomic_kind_set ...
!> \param iw ...
!> \param inp_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_damp(atomic_kind_set, &
                                    iw, inp_info)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER                                            :: iw
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_damp', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: atm_name1, atm_name2, my_atm_name1, &
                                                            my_atm_name2
      INTEGER                                            :: handle2, i, j, k, nkinds
      LOGICAL                                            :: found
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind, atomic_kind2
      TYPE(damping_p_type), POINTER                      :: damping

      CALL timeset(routineN, handle2)
      NULLIFY (damping)
      nkinds = SIZE(atomic_kind_set)

      DO j = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(j)
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              name=atm_name1)
         CALL UPPERCASE(atm_name1)
         IF (ASSOCIATED(inp_info%damping_list)) THEN
            DO i = 1, SIZE(inp_info%damping_list)
               my_atm_name1 = inp_info%damping_list(i)%atm_name1
               my_atm_name2 = inp_info%damping_list(i)%atm_name2

               IF (iw > 0) WRITE (iw, *) "Damping Checking ::", TRIM(my_atm_name1), &
                  TRIM(atm_name1)
               IF (my_atm_name1 == atm_name1) THEN
                  IF (.NOT. ASSOCIATED(damping)) THEN
                     CALL damping_p_create(damping, nkinds)
                  END IF

                  found = .FALSE.
                  DO k = 1, SIZE(atomic_kind_set)
                     atomic_kind2 => atomic_kind_set(k)
                     CALL get_atomic_kind(atomic_kind=atomic_kind2, &
                                          name=atm_name2)
                     CALL UPPERCASE(atm_name2)

                     IF (my_atm_name2 == atm_name2) THEN
                        IF (damping%damp(k)%bij /= HUGE(0.0_dp)) found = .TRUE.
                        CALL issue_duplications(found, "Damping", atm_name1)
                        found = .TRUE.

                        SELECT CASE (TRIM (inp_info%damping_list (i)%dtype))
                        CASE ('TANG-TOENNIES')
                           damping%damp(k)%itype = tang_toennies
                        CASE DEFAULT
                           CALL cp__b("force_fields_all.F",2303,"Unknown damping type.")
                        END SELECT
                        damping%damp(k)%order = inp_info%damping_list(i)%order
                        damping%damp(k)%bij = inp_info%damping_list(i)%bij
                        damping%damp(k)%cij = inp_info%damping_list(i)%cij
                     ENDIF

                  END DO
                  IF (.NOT. found) &
                     CALL cp_warn(cp__l("force_fields_all.F",2312), &
                                  "Atom "//TRIM(my_atm_name2)// &
                                  " in damping parameters for atom "//TRIM(my_atm_name1)// &
                                  " not found! ")
               ENDIF
            END DO

         ENDIF

         CALL set_atomic_kind(atomic_kind=atomic_kind, damping=damping)
         NULLIFY (damping)
      END DO

      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_damp

! **************************************************************************************************
!> \brief Set up shell potential parameters
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param root_section ...
!> \param subsys_section ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param cell ...
!> \param iw ...
!> \param inp_info ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_shell(particle_set, atomic_kind_set, &
                                     molecule_kind_set, molecule_set, root_section, subsys_section, &
                                     shell_particle_set, core_particle_set, cell, iw, inp_info)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(section_vals_type), POINTER                   :: root_section, subsys_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_particle_set, core_particle_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER                                            :: iw
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_shell', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atmname
      INTEGER                                            :: counter, first, first_shell, handle2, i, &
                                                            j, last, last_shell, n, natom, nmol, &
                                                            nshell_tot
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list, shell_list_tmp
      LOGICAL :: core_coord_read, found_shell, is_a_shell, is_link_atom, null_massfrac, only_qm, &
         save_mem, shell_adiabatic, shell_coord_read
      REAL(KIND=dp)                                      :: atmmass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(section_vals_type), POINTER                   :: global_section
      TYPE(shell_kind_type), POINTER                     :: shell
      TYPE(shell_type), DIMENSION(:), POINTER            :: shell_list

      CALL timeset(routineN, handle2)
      nshell_tot = 0
      n = 0
      first_shell = 1
      null_massfrac = .FALSE.
      core_coord_read = .FALSE.
      shell_coord_read = .FALSE.

      NULLIFY (global_section)
      global_section => section_vals_get_subs_vals(root_section, "GLOBAL")
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)

      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              name=atmname)

         found_shell = .FALSE.
         only_qm = qmmm_ff_precond_only_qm(id1=atmname, is_link=is_link_atom)
         CALL uppercase(atmname)

         ! The shell potential can be defined only from input
         IF (ASSOCIATED(inp_info%shell_list)) THEN
            DO j = 1, SIZE(inp_info%shell_list)
               IF (iw > 0) WRITE (iw, *) "Shell Checking ::", TRIM(inp_info%shell_list(j)%atm_name), atmname
               IF ((inp_info%shell_list(j)%atm_name) == atmname) THEN
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       shell=shell, mass=atmmass, natom=natom)
                  IF (.NOT. ASSOCIATED(shell)) THEN
                     CALL shell_create(shell)
                  END IF
                  nshell_tot = nshell_tot+natom
                  shell%charge_core = inp_info%shell_list(j)%shell%charge_core
                  shell%charge_shell = inp_info%shell_list(j)%shell%charge_shell
                  shell%massfrac = inp_info%shell_list(j)%shell%massfrac
                  IF (shell%massfrac < EPSILON(1.0_dp)) null_massfrac = .TRUE.
                  shell%k2_spring = inp_info%shell_list(j)%shell%k2_spring
                  shell%k4_spring = inp_info%shell_list(j)%shell%k4_spring
                  shell%max_dist = inp_info%shell_list(j)%shell%max_dist
                  shell%shell_cutoff = inp_info%shell_list(j)%shell%shell_cutoff
                  shell%mass_shell = shell%massfrac*atmmass
                  shell%mass_core = atmmass-shell%mass_shell
                  CALL issue_duplications(found_shell, "Shell", atmname)
                  found_shell = .TRUE.
                  CALL set_atomic_kind(atomic_kind=atomic_kind, &
                                       shell=shell, shell_active=.TRUE.)
                  CALL shell_release(shell)
               END IF
            END DO ! j shell kind
         END IF ! associated shell_list
      END DO ! i atomic kind

      IF (iw > 0) WRITE (iw, *) "Total number of particles with a shell :: ", nshell_tot
      ! If shell-model is present: Create particle_set of shells (coord. vel. force)
      NULLIFY (shell_particle_set)
      NULLIFY (core_particle_set)
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, shell_adiabatic=shell_adiabatic)
      IF (nshell_tot > 0) THEN
         IF (shell_adiabatic .AND. null_massfrac) THEN
            CALL cp__b("force_fields_all.F",2434,"Shell-model adiabatic: at least one shell_kind has mass zero")
         END IF
         CALL allocate_particle_set(shell_particle_set, nshell_tot)
         CALL allocate_particle_set(core_particle_set, nshell_tot)
         counter = 0
         ! Initialise the shell (and core) coordinates with the particle (atomic) coordinates,
         ! count the shell and set pointers
         DO i = 1, SIZE(particle_set)
            NULLIFY (atomic_kind)
            NULLIFY (shell)
            atomic_kind => particle_set(i)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, shell_active=is_a_shell)
            IF (is_a_shell) THEN
               counter = counter+1
               particle_set(i)%shell_index = counter
               shell_particle_set(counter)%shell_index = counter
               shell_particle_set(counter)%atomic_kind => particle_set(i)%atomic_kind
               shell_particle_set(counter)%r(1:3) = particle_set(i)%r(1:3)
               shell_particle_set(counter)%atom_index = i
               core_particle_set(counter)%shell_index = counter
               core_particle_set(counter)%atomic_kind => particle_set(i)%atomic_kind
               core_particle_set(counter)%r(1:3) = particle_set(i)%r(1:3)
               core_particle_set(counter)%atom_index = i
            ELSE
               particle_set(i)%shell_index = 0
            END IF
         END DO
         IF(.NOT.(counter == nshell_tot))CALL cp__a("force_fields_all.F",2461)
      END IF

      ! Read the shell (and core) coordinates from the restart file, if available
      CALL read_binary_cs_coordinates("SHELL", shell_particle_set, root_section, &
                                      subsys_section, shell_coord_read)
      CALL read_binary_cs_coordinates("CORE", core_particle_set, root_section, &
                                      subsys_section, core_coord_read)

      IF (nshell_tot > 0) THEN
         ! Read the shell (and core) coordinates from the input, if no coordinates were found
         ! in the restart file
         IF (shell_adiabatic) THEN
            IF (.NOT. (core_coord_read .AND. shell_coord_read)) THEN
               CALL read_shell_coord_input(particle_set, shell_particle_set, cell, &
                                           subsys_section, core_particle_set, &
                                           save_mem=save_mem)
            END IF
         ELSE
            IF (.NOT. shell_coord_read) THEN
               CALL read_shell_coord_input(particle_set, shell_particle_set, cell, &
                                           subsys_section, save_mem=save_mem)
            END IF
         END IF
         ! Determine the number of shells per molecule kind
         n = 0
         DO i = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(i)
            CALL get_molecule_kind(molecule_kind=molecule_kind, molecule_list=molecule_list, &
                                   natom=natom, nmolecule=nmol)
            molecule => molecule_set(molecule_list(1))
            CALL get_molecule(molecule=molecule, first_atom=first, last_atom=last)
            ALLOCATE (shell_list_tmp(natom))
            counter = 0
            DO j = first, last
               atomic_kind => particle_set(j)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, shell_active=is_a_shell)
               IF (is_a_shell) THEN
                  counter = counter+1
                  shell_list_tmp(counter) = j-first+1
                  first_shell = MIN(first_shell, MAX(1, particle_set(j)%shell_index))
               END IF
            END DO ! j atom in molecule_kind i, molecule 1 of the molecule_list
            IF (counter /= 0) THEN
               ! Setup of fist_shell and last_shell for all molecules..
               DO j = 1, SIZE(molecule_list)
                  last_shell = first_shell+counter-1
                  molecule => molecule_set(molecule_list(j))
                  molecule%first_shell = first_shell
                  molecule%last_shell = last_shell
                  first_shell = last_shell+1
               END DO
               ! Setup of shell_list
               CALL get_molecule_kind(molecule_kind=molecule_kind, shell_list=shell_list)
               IF (ASSOCIATED(shell_list)) THEN
                  DEALLOCATE (shell_list)
               END IF
               ALLOCATE (shell_list(counter))
               DO j = 1, counter
                  shell_list(j)%a = shell_list_tmp(j)
                  atomic_kind => particle_set(shell_list_tmp(j)+first-1)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, name=atmname, shell=shell)
                  CALL uppercase(atmname)
                  shell_list(j)%name = atmname
                  shell_list(j)%shell_kind => shell
               END DO
               CALL set_molecule_kind(molecule_kind=molecule_kind, nshell=counter, shell_list=shell_list)
            END IF
            DEALLOCATE (shell_list_tmp)
            n = n+nmol*counter
         END DO ! i molecule kind
      END IF
      IF(.NOT.(first_shell-1 == nshell_tot))CALL cp__a("force_fields_all.F",2533)
      IF(.NOT.(n == nshell_tot))CALL cp__a("force_fields_all.F",2534)
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_shell

! **************************************************************************************************
!> \brief Assign input and potential info to potparm_nonbond14
!> \param atomic_kind_set ...
!> \param ff_type ...
!> \param qmmm_env ...
!> \param iw ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
!> \param amb_info ...
!> \param potparm_nonbond14 ...
!> \param ewald_env ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_nonbond14(atomic_kind_set, ff_type, qmmm_env, iw, &
                                         Ainfo, chm_info, inp_info, gro_info, amb_info, potparm_nonbond14, ewald_env)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      INTEGER                                            :: iw
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(amber_info_type), POINTER                     :: amb_info
      TYPE(pair_potential_pp_type), POINTER              :: potparm_nonbond14
      TYPE(ewald_environment_type), POINTER              :: ewald_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_nonbond14', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a_local, &
                                                            name_atm_b, name_atm_b_local
      INTEGER                                            :: handle2, i, ii, j, jj, k, match_names
      LOGICAL                                            :: found, found_a, found_b, only_qm, &
                                                            use_qmmm_ff
      REAL(KIND=dp)                                      :: epsilon0, epsilon_a, epsilon_b, &
                                                            ewald_rcut, rmin, rmin2_a, rmin2_b
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(pair_potential_single_type), POINTER          :: pot

      use_qmmm_ff = qmmm_env%use_qmmm_ff
      NULLIFY (pot)
      CALL ewald_env_get(ewald_env, rcut=ewald_rcut)
      CALL timeset(routineN, handle2)
      CALL pair_potential_pp_create(potparm_nonbond14, SIZE(atomic_kind_set))
      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, name=name_atm_a_local)
         DO j = i, SIZE(atomic_kind_set)
            atomic_kind => atomic_kind_set(j)
            CALL get_atomic_kind(atomic_kind=atomic_kind, name=name_atm_b_local)
            found = .FALSE.
            found_a = .FALSE.
            found_b = .FALSE.
            name_atm_a = name_atm_a_local
            name_atm_b = name_atm_b_local
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            pot => potparm_nonbond14%pot(i, j)%pot

            ! loop over params from GROMOS
            IF (ASSOCIATED(gro_info%nonbond_a_14)) THEN
               ii = 0
               jj = 0
               DO k = 1, SIZE(gro_info%nonbond_a_14)
                  IF (TRIM(name_atm_a) == TRIM(gro_info%nonbond_a_14(k))) THEN
                     ii = k
                     found_a = .TRUE.
                     EXIT
                  END IF
               END DO
               DO k = 1, SIZE(gro_info%nonbond_a_14)
                  IF (TRIM(name_atm_b) == TRIM(gro_info%nonbond_a_14(k))) THEN
                     jj = k
                     found_b = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (ii /= 0 .AND. jj /= 0) THEN
                  CALL pair_potential_lj_create(pot%set(1)%lj)
                  pot%type = lj_type
                  pot%at1 = name_atm_a
                  pot%at2 = name_atm_b
                  pot%set(1)%lj%epsilon = 1.0_dp
                  pot%set(1)%lj%sigma6 = gro_info%nonbond_c6_14(ii, jj)
                  pot%set(1)%lj%sigma12 = gro_info%nonbond_c12_14(ii, jj)
                  pot%rcutsq = (10.0_dp*bohr)**2
                  CALL issue_duplications(found, "Lennard-Jones", name_atm_a, name_atm_b)
                  found = .TRUE.
               END IF
            END IF

            ! loop over params from CHARMM
            ii = 0
            jj = 0
            IF (ASSOCIATED(chm_info%nonbond_a_14)) THEN
               DO k = 1, SIZE(chm_info%nonbond_a_14)
                  IF ((name_atm_a) == (chm_info%nonbond_a_14(k))) THEN
                     ii = k
                     rmin2_a = chm_info%nonbond_rmin2_14(k)
                     epsilon_a = chm_info%nonbond_eps_14(k)
                     found_a = .TRUE.
                  END IF
               END DO
               DO k = 1, SIZE(chm_info%nonbond_a_14)
                  IF ((name_atm_b) == (chm_info%nonbond_a_14(k))) THEN
                     jj = k
                     rmin2_b = chm_info%nonbond_rmin2_14(k)
                     epsilon_b = chm_info%nonbond_eps_14(k)
                     found_b = .TRUE.
                  END IF
               END DO
            END IF
            IF (ASSOCIATED(chm_info%nonbond_a)) THEN
               IF (.NOT. found_a) THEN
                  DO k = 1, SIZE(chm_info%nonbond_a)
                     IF ((name_atm_a) == (chm_info%nonbond_a(k))) THEN
                        ii = k
                        rmin2_a = chm_info%nonbond_rmin2(k)
                        epsilon_a = chm_info%nonbond_eps(k)
                     END IF
                  END DO
               END IF
               IF (.NOT. found_b) THEN
                  DO k = 1, SIZE(chm_info%nonbond_a)
                     IF ((name_atm_b) == (chm_info%nonbond_a(k))) THEN
                        jj = k
                        rmin2_b = chm_info%nonbond_rmin2(k)
                        epsilon_b = chm_info%nonbond_eps(k)
                     END IF
                  END DO
               END IF
            END IF
            IF (ii /= 0 .AND. jj /= 0) THEN
               rmin = rmin2_a+rmin2_b
               ! ABS to allow for mixing the two different sign conventions for epsilon
               epsilon0 = SQRT(ABS(epsilon_a*epsilon_b))
               CALL pair_potential_lj_create(pot%set(1)%lj)
               pot%type = lj_charmm_type
               pot%at1 = name_atm_a
               pot%at2 = name_atm_b
               pot%set(1)%lj%epsilon = epsilon0
               pot%set(1)%lj%sigma6 = 0.5_dp*rmin**6
               pot%set(1)%lj%sigma12 = 0.25_dp*rmin**12
               pot%rcutsq = (10.0_dp*bohr)**2
               CALL issue_duplications(found, "Lennard-Jones", name_atm_a, name_atm_b)
               found = .TRUE.
            END IF

            ! loop over params from AMBER
            IF (ASSOCIATED(amb_info%nonbond_a)) THEN
               ii = 0
               jj = 0
               IF (.NOT. found_a) THEN
                  DO k = 1, SIZE(amb_info%nonbond_a)
                     IF ((name_atm_a) == (amb_info%nonbond_a(k))) THEN
                        ii = k
                        rmin2_a = amb_info%nonbond_rmin2(k)
                        epsilon_a = amb_info%nonbond_eps(k)
                     END IF
                  END DO
               END IF
               IF (.NOT. found_b) THEN
                  DO k = 1, SIZE(amb_info%nonbond_a)
                     IF ((name_atm_b) == (amb_info%nonbond_a(k))) THEN
                        jj = k
                        rmin2_b = amb_info%nonbond_rmin2(k)
                        epsilon_b = amb_info%nonbond_eps(k)
                     END IF
                  END DO
               END IF
               IF (ii /= 0 .AND. jj /= 0) THEN
                  rmin = rmin2_a+rmin2_b
                  ! ABS to allow for mixing the two different sign conventions for epsilon
                  epsilon0 = SQRT(ABS(epsilon_a*epsilon_b))
                  CALL pair_potential_lj_create(pot%set(1)%lj)
                  pot%type = lj_charmm_type
                  pot%at1 = name_atm_a
                  pot%at2 = name_atm_b
                  pot%set(1)%lj%epsilon = epsilon0
                  pot%set(1)%lj%sigma6 = 0.5_dp*rmin**6
                  pot%set(1)%lj%sigma12 = 0.25_dp*rmin**12
                  pot%rcutsq = (10.0_dp*bohr)**2
                  CALL issue_duplications(found, "Lennard-Jones", name_atm_a, &
                                          name_atm_b)
                  found = .TRUE.
               END IF
            END IF

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%nonbonded14)) THEN
               DO k = 1, SIZE(inp_info%nonbonded14%pot)
                  IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                     " with ", TRIM(inp_info%nonbonded14%pot(k)%pot%at1), &
                     TRIM(inp_info%nonbonded14%pot(k)%pot%at2)
                  IF ((((name_atm_a) == (inp_info%nonbonded14%pot(k)%pot%at1)) .AND. &
                       ((name_atm_b) == (inp_info%nonbonded14%pot(k)%pot%at2))) .OR. &
                      (((name_atm_b) == (inp_info%nonbonded14%pot(k)%pot%at1)) .AND. &
                       ((name_atm_a) == (inp_info%nonbonded14%pot(k)%pot%at2)))) THEN
                     IF (ff_type%multiple_potential) THEN
                        CALL pair_potential_single_add(inp_info%nonbonded14%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",2745), &
                                        "Multiple ONFO declaration: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" ADDING! ")
                        potparm_nonbond14%pot(i, j)%pot => pot
                        potparm_nonbond14%pot(j, i)%pot => pot
                     ELSE
                        CALL pair_potential_single_copy(inp_info%nonbonded14%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",2753), &
                                        "Multiple ONFO declarations: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" OVERWRITING! ")
                     END IF
                     IF (iw > 0) WRITE (iw, *) "    FOUND ", TRIM(name_atm_a), " ", TRIM(name_atm_b)
                     found = .TRUE.
                  END IF
               END DO
            END IF
            ! at the very end we offer the possibility to overwrite the parameters for QM/MM
            ! nonbonded interactions
            IF (use_qmmm_ff) THEN
               match_names = 0
               IF ((name_atm_a) == (name_atm_a_local)) match_names = match_names+1
               IF ((name_atm_b) == (name_atm_b_local)) match_names = match_names+1
               IF (match_names == 1) THEN
                  IF (ASSOCIATED(qmmm_env%inp_info%nonbonded14)) THEN
                     DO k = 1, SIZE(qmmm_env%inp_info%nonbonded14%pot)
                        IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                           " with ", TRIM(qmmm_env%inp_info%nonbonded14%pot(k)%pot%at1), &
                           TRIM(qmmm_env%inp_info%nonbonded14%pot(k)%pot%at2)
                        IF ((((name_atm_a) == (qmmm_env%inp_info%nonbonded14%pot(k)%pot%at1)) .AND. &
                             ((name_atm_b) == (qmmm_env%inp_info%nonbonded14%pot(k)%pot%at2))) .OR. &
                            (((name_atm_b) == (qmmm_env%inp_info%nonbonded14%pot(k)%pot%at1)) .AND. &
                             ((name_atm_a) == (qmmm_env%inp_info%nonbonded14%pot(k)%pot%at2)))) THEN
                           IF (qmmm_env%multiple_potential) THEN
                              CALL pair_potential_single_add(qmmm_env%inp_info%nonbonded14%pot(k)%pot, pot)
                              IF (found) &
                                 CALL cp_warn(cp__l("force_fields_all.F",2781), &
                                              "Multiple ONFO declaration: "//TRIM(name_atm_a)// &
                                              " and "//TRIM(name_atm_b)//" ADDING QM/MM forcefield specifications! ")
                              potparm_nonbond14%pot(i, j)%pot => pot
                              potparm_nonbond14%pot(j, i)%pot => pot
                           ELSE
                              CALL pair_potential_single_copy(qmmm_env%inp_info%nonbonded14%pot(k)%pot, pot)
                              IF (found) &
                                 CALL cp_warn(cp__l("force_fields_all.F",2789), &
                                              "Multiple ONFO declaration: "//TRIM(name_atm_a)// &
                                              " and "//TRIM(name_atm_b)//" OVERWRITING QM/MM forcefield specifications! ")
                           END IF
                           IF (iw > 0) WRITE (iw, *) "    FOUND ", TRIM(name_atm_a), &
                              " ", TRIM(name_atm_b)
                           found = .TRUE.
                        END IF
                     END DO
                  END IF
               END IF
            END IF

            IF (.NOT. found) THEN
               CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                         atm2=TRIM(name_atm_b), &
                                         type_name="Spline_Bond_Env", &
                                         array=Ainfo)
               CALL pair_potential_single_clean(pot)
               pot%type = nn_type
               pot%at1 = name_atm_a
               pot%at2 = name_atm_b
            END IF
            ! If defined global RCUT let's use it
            IF (ff_type%rcut_nb > 0.0_dp) THEN
               pot%rcutsq = ff_type%rcut_nb*ff_type%rcut_nb
            END IF
            ! Cutoff is defined always as the maximum between the FF and Ewald
            pot%rcutsq = MAX(pot%rcutsq, ewald_rcut*ewald_rcut)
            IF (only_qm) THEN
               CALL pair_potential_single_clean(pot)
            END IF
         END DO ! atom kind j
      END DO ! atom kind i
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_nonbond14

! **************************************************************************************************
!> \brief Assign input and potential info to potparm_nonbond
!> \param atomic_kind_set ...
!> \param ff_type ...
!> \param qmmm_env ...
!> \param fatal ...
!> \param iw ...
!> \param Ainfo ...
!> \param chm_info ...
!> \param inp_info ...
!> \param gro_info ...
!> \param amb_info ...
!> \param potparm_nonbond ...
!> \param ewald_env ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_nonbond(atomic_kind_set, ff_type, qmmm_env, fatal, &
                                       iw, Ainfo, chm_info, inp_info, gro_info, amb_info, potparm_nonbond, &
                                       ewald_env)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      LOGICAL                                            :: fatal
      INTEGER                                            :: iw
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(amber_info_type), POINTER                     :: amb_info
      TYPE(pair_potential_pp_type), POINTER              :: potparm_nonbond
      TYPE(ewald_environment_type), POINTER              :: ewald_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_nonbond', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_a_local, &
                                                            name_atm_b, name_atm_b_local
      INTEGER                                            :: handle2, i, ii, j, jj, k, match_names
      LOGICAL                                            :: found, is_a_shell, is_b_shell, only_qm, &
                                                            use_qmmm_ff
      REAL(KIND=dp)                                      :: epsilon0, ewald_rcut, rmin
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(pair_potential_single_type), POINTER          :: pot

      CALL timeset(routineN, handle2)
      use_qmmm_ff = qmmm_env%use_qmmm_ff
      NULLIFY (pot)
      CALL ewald_env_get(ewald_env, rcut=ewald_rcut)
      CALL pair_potential_pp_create(potparm_nonbond, SIZE(atomic_kind_set))
      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind, name=name_atm_a_local, &
                              shell_active=is_a_shell)
         DO j = i, SIZE(atomic_kind_set)
            atomic_kind => atomic_kind_set(j)
            CALL get_atomic_kind(atomic_kind=atomic_kind, name=name_atm_b_local, &
                                 shell_active=is_b_shell)
            found = .FALSE.
            name_atm_a = name_atm_a_local
            name_atm_b = name_atm_b_local
            only_qm = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b)
            CALL uppercase(name_atm_a)
            CALL uppercase(name_atm_b)
            pot => potparm_nonbond%pot(i, j)%pot

            ! loop over params from GROMOS
            IF (ASSOCIATED(gro_info%nonbond_a)) THEN
               ii = 0
               jj = 0
               DO k = 1, SIZE(gro_info%nonbond_a)
                  IF (TRIM(name_atm_a) == TRIM(gro_info%nonbond_a(k))) THEN
                     ii = k
                     EXIT
                  END IF
               END DO
               DO k = 1, SIZE(gro_info%nonbond_a)
                  IF (TRIM(name_atm_b) == TRIM(gro_info%nonbond_a(k))) THEN
                     jj = k
                     EXIT
                  END IF
               END DO

               IF (ii /= 0 .AND. jj /= 0) THEN
                  CALL pair_potential_lj_create(pot%set(1)%lj)
                  pot%type = lj_type
                  pot%at1 = name_atm_a
                  pot%at2 = name_atm_b
                  pot%set(1)%lj%epsilon = 1.0_dp
                  pot%set(1)%lj%sigma6 = gro_info%nonbond_c6(ii, jj)
                  pot%set(1)%lj%sigma12 = gro_info%nonbond_c12(ii, jj)
                  pot%rcutsq = (10.0_dp*bohr)**2
                  CALL issue_duplications(found, "Lennard-Jones", name_atm_a, name_atm_b)
                  found = .TRUE.
               END IF
            END IF

            ! loop over params from CHARMM
            IF (ASSOCIATED(chm_info%nonbond_a)) THEN
               ii = 0
               jj = 0
               DO k = 1, SIZE(chm_info%nonbond_a)
                  IF ((name_atm_a) == (chm_info%nonbond_a(k))) THEN
                     ii = k
                  END IF
               END DO
               DO k = 1, SIZE(chm_info%nonbond_a)
                  IF ((name_atm_b) == (chm_info%nonbond_a(k))) THEN
                     jj = k
                  END IF
               END DO

               IF (ii /= 0 .AND. jj /= 0) THEN
                  rmin = chm_info%nonbond_rmin2(ii)+chm_info%nonbond_rmin2(jj)
                  epsilon0 = SQRT(chm_info%nonbond_eps(ii)* &
                                  chm_info%nonbond_eps(jj))
                  CALL pair_potential_lj_create(pot%set(1)%lj)
                  pot%type = lj_charmm_type
                  pot%at1 = name_atm_a
                  pot%at2 = name_atm_b
                  pot%set(1)%lj%epsilon = epsilon0
                  pot%set(1)%lj%sigma6 = 0.5_dp*rmin**6
                  pot%set(1)%lj%sigma12 = 0.25_dp*rmin**12
                  pot%rcutsq = (10.0_dp*bohr)**2
                  CALL issue_duplications(found, "Lennard-Jones", name_atm_a, name_atm_b)
                  found = .TRUE.
               END IF
            END IF

            ! loop over params from AMBER
            IF (ASSOCIATED(amb_info%nonbond_a)) THEN
               ii = 0
               jj = 0
               DO k = 1, SIZE(amb_info%nonbond_a)
                  IF ((name_atm_a) == (amb_info%nonbond_a(k))) THEN
                     ii = k
                  END IF
               END DO
               DO k = 1, SIZE(amb_info%nonbond_a)
                  IF ((name_atm_b) == (amb_info%nonbond_a(k))) THEN
                     jj = k
                  END IF
               END DO

               IF (ii /= 0 .AND. jj /= 0) THEN
                  rmin = amb_info%nonbond_rmin2(ii)+amb_info%nonbond_rmin2(jj)
                  epsilon0 = SQRT(amb_info%nonbond_eps(ii)*amb_info%nonbond_eps(jj))
                  CALL pair_potential_lj_create(pot%set(1)%lj)
                  pot%type = lj_charmm_type
                  pot%at1 = name_atm_a
                  pot%at2 = name_atm_b
                  pot%set(1)%lj%epsilon = epsilon0
                  pot%set(1)%lj%sigma6 = 0.5_dp*rmin**6
                  pot%set(1)%lj%sigma12 = 0.25_dp*rmin**12
                  pot%rcutsq = (10.0_dp*bohr)**2
                  CALL issue_duplications(found, "Lennard-Jones", name_atm_a, name_atm_b)
                  found = .TRUE.
               END IF
            END IF

            ! always have the input param last to overwrite all the other ones
            IF (ASSOCIATED(inp_info%nonbonded)) THEN
               DO k = 1, SIZE(inp_info%nonbonded%pot)
                  IF ((TRIM(inp_info%nonbonded%pot(k)%pot%at1) == "*") .OR. &
                      (TRIM(inp_info%nonbonded%pot(k)%pot%at2) == "*")) CYCLE

                  IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                     " with ", TRIM(inp_info%nonbonded%pot(k)%pot%at1), &
                     TRIM(inp_info%nonbonded%pot(k)%pot%at2)
                  IF ((((name_atm_a) == (inp_info%nonbonded%pot(k)%pot%at1)) .AND. &
                       ((name_atm_b) == (inp_info%nonbonded%pot(k)%pot%at2))) .OR. &
                      (((name_atm_b) == (inp_info%nonbonded%pot(k)%pot%at1)) .AND. &
                       ((name_atm_a) == (inp_info%nonbonded%pot(k)%pot%at2)))) THEN
                     IF (ff_type%multiple_potential) THEN
                        CALL pair_potential_single_add(inp_info%nonbonded%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",3003), &
                                        "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" ADDING! ")
                        potparm_nonbond%pot(i, j)%pot => pot
                        potparm_nonbond%pot(j, i)%pot => pot
                     ELSE
                        CALL pair_potential_single_copy(inp_info%nonbonded%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",3011), &
                                        "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" OVERWRITING! ")
                     END IF
                     IF (iw > 0) WRITE (iw, *) "    FOUND ", TRIM(name_atm_a), " ", TRIM(name_atm_b)
                     found = .TRUE.
                  END IF
               END DO
               ! Check for wildcards for one of the two types (if not associated yet)
               IF (.NOT. found) THEN
                  DO k = 1, SIZE(inp_info%nonbonded%pot)
                     IF ((TRIM(inp_info%nonbonded%pot(k)%pot%at1) == "*") .EQV. &
                         (TRIM(inp_info%nonbonded%pot(k)%pot%at2) == "*")) CYCLE

                     IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                        " with ", TRIM(inp_info%nonbonded%pot(k)%pot%at1), &
                        TRIM(inp_info%nonbonded%pot(k)%pot%at2)

                     IF ((name_atm_a == inp_info%nonbonded%pot(k)%pot%at1) .OR. &
                         (name_atm_b == inp_info%nonbonded%pot(k)%pot%at2) .OR. &
                         (name_atm_b == inp_info%nonbonded%pot(k)%pot%at1) .OR. &
                         (name_atm_a == inp_info%nonbonded%pot(k)%pot%at2)) THEN
                        IF (ff_type%multiple_potential) THEN
                           CALL pair_potential_single_add(inp_info%nonbonded%pot(k)%pot, pot)
                           IF (found) &
                              CALL cp_warn(cp__l("force_fields_all.F",3036), &
                                           "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                           " and "//TRIM(name_atm_b)//" ADDING! ")
                           potparm_nonbond%pot(i, j)%pot => pot
                           potparm_nonbond%pot(j, i)%pot => pot
                        ELSE
                           CALL pair_potential_single_copy(inp_info%nonbonded%pot(k)%pot, pot)
                           IF (found) &
                              CALL cp_warn(cp__l("force_fields_all.F",3044), &
                                           "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                           " and "//TRIM(name_atm_b)//" OVERWRITING! ")
                        END IF
                        IF (iw > 0) WRITE (iw, *) "    FOUND (one WILDCARD)", TRIM(name_atm_a), " ", TRIM(name_atm_b)
                        found = .TRUE.
                     END IF
                  END DO
               END IF
               ! Check for wildcards for both types (if not associated yet)
               IF (.NOT. found) THEN
                  DO k = 1, SIZE(inp_info%nonbonded%pot)
                     IF ((TRIM(inp_info%nonbonded%pot(k)%pot%at1) /= "*") .OR. &
                         (TRIM(inp_info%nonbonded%pot(k)%pot%at2) /= "*")) CYCLE

                     IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                        " with ", TRIM(inp_info%nonbonded%pot(k)%pot%at1), &
                        TRIM(inp_info%nonbonded%pot(k)%pot%at2)

                     IF (ff_type%multiple_potential) THEN
                        CALL pair_potential_single_add(inp_info%nonbonded%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",3066), &
                                        "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" ADDING! ")
                        potparm_nonbond%pot(i, j)%pot => pot
                        potparm_nonbond%pot(j, i)%pot => pot
                     ELSE
                        CALL pair_potential_single_copy(inp_info%nonbonded%pot(k)%pot, pot)
                        IF (found) &
                           CALL cp_warn(cp__l("force_fields_all.F",3074), &
                                        "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                        " and "//TRIM(name_atm_b)//" OVERWRITING! ")
                     END IF
                     IF (iw > 0) WRITE (iw, *) "    FOUND (both WILDCARDS)", TRIM(name_atm_a), " ", TRIM(name_atm_b)
                     found = .TRUE.
                  END DO
               END IF
            END IF

            ! at the very end we offer the possibility to overwrite the parameters for QM/MM
            ! nonbonded interactions
            IF (use_qmmm_ff) THEN
               match_names = 0
               IF ((name_atm_a) == (name_atm_a_local)) match_names = match_names+1
               IF ((name_atm_b) == (name_atm_b_local)) match_names = match_names+1
               IF (match_names == 1) THEN
                  IF (ASSOCIATED(qmmm_env%inp_info%nonbonded)) THEN
                     DO k = 1, SIZE(qmmm_env%inp_info%nonbonded%pot)
                        IF (iw > 0) WRITE (iw, *) "    TESTING ", TRIM(name_atm_a), TRIM(name_atm_b), &
                           " with ", TRIM(qmmm_env%inp_info%nonbonded%pot(k)%pot%at1), &
                           TRIM(qmmm_env%inp_info%nonbonded%pot(k)%pot%at2)
                        IF ((((name_atm_a) == (qmmm_env%inp_info%nonbonded%pot(k)%pot%at1)) .AND. &
                             ((name_atm_b) == (qmmm_env%inp_info%nonbonded%pot(k)%pot%at2))) .OR. &
                            (((name_atm_b) == (qmmm_env%inp_info%nonbonded%pot(k)%pot%at1)) .AND. &
                             ((name_atm_a) == (qmmm_env%inp_info%nonbonded%pot(k)%pot%at2)))) THEN
                           IF (qmmm_env%multiple_potential) THEN
                              CALL pair_potential_single_add(qmmm_env%inp_info%nonbonded%pot(k)%pot, pot)
                              IF (found) &
                                 CALL cp_warn(cp__l("force_fields_all.F",3103), &
                                              "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                              " and "//TRIM(name_atm_b)//" ADDING QM/MM forcefield specifications! ")
                              potparm_nonbond%pot(i, j)%pot => pot
                              potparm_nonbond%pot(j, i)%pot => pot
                           ELSE
                              CALL pair_potential_single_copy(qmmm_env%inp_info%nonbonded%pot(k)%pot, pot)
                              IF (found) &
                                 CALL cp_warn(cp__l("force_fields_all.F",3111), &
                                              "Multiple NONBONDED declaration: "//TRIM(name_atm_a)// &
                                              " and "//TRIM(name_atm_b)//" OVERWRITING QM/MM forcefield specifications! ")
                           END IF
                           IF (iw > 0) WRITE (iw, *) "    FOUND ", TRIM(name_atm_a), " ", TRIM(name_atm_b)
                           found = .TRUE.
                        END IF
                     END DO
                  END IF
               END IF
            END IF
            IF (.NOT. found) THEN
               CALL store_FF_missing_par(atm1=TRIM(name_atm_a), &
                                         atm2=TRIM(name_atm_b), &
                                         type_name="Spline_Non_Bond_Env", &
                                         fatal=fatal, &
                                         array=Ainfo)
            END IF
            ! If defined global RCUT let's use it
            IF (ff_type%rcut_nb > 0.0_dp) THEN
               pot%rcutsq = ff_type%rcut_nb*ff_type%rcut_nb
            END IF
            ! Cutoff is defined always as the maximum between the FF and Ewald
            pot%rcutsq = MAX(pot%rcutsq, ewald_rcut*ewald_rcut)
            ! Set the shell type
            IF ((is_a_shell .AND. .NOT. is_b_shell) .OR. (is_b_shell .AND. .NOT. is_a_shell)) THEN
               pot%shell_type = nosh_sh
            ELSE IF (is_a_shell .AND. is_b_shell) THEN
               pot%shell_type = sh_sh
            ELSE
               pot%shell_type = nosh_nosh
            END IF
            IF (only_qm) THEN
               CALL pair_potential_single_clean(pot)
            END IF
         END DO ! jkind
      END DO ! ikind
      CALL timestop(handle2)
   END SUBROUTINE force_field_pack_nonbond

! **************************************************************************************************
!> \brief create the pair potential spline environment
!> \param atomic_kind_set ...
!> \param ff_type ...
!> \param iw2 ...
!> \param iw3 ...
!> \param iw4 ...
!> \param potparm ...
!> \param do_zbl ...
!> \param nonbonded_type ...
! **************************************************************************************************
   SUBROUTINE force_field_pack_splines(atomic_kind_set, ff_type, iw2, iw3, iw4, &
                                       potparm, do_zbl, nonbonded_type)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      INTEGER                                            :: iw2, iw3, iw4
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      LOGICAL, INTENT(IN)                                :: do_zbl
      CHARACTER(LEN=*), INTENT(IN)                       :: nonbonded_type

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_splines', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle2, ikind, jkind, n
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p
      TYPE(spline_environment_type), POINTER             :: spline_env

      CALL timeset(routineN, handle2)
      ! Figure out which nonbonded interactions happen to be indentical, and
      ! prepare storage for these, avoiding duplicates.
      NULLIFY (spline_env)
      CALL get_nonbond_storage(spline_env, potparm, atomic_kind_set, &
                               do_zbl, shift_cutoff=ff_type%shift_cutoff)
      ! Effectively compute the spline data.
      CALL spline_nonbond_control(spline_env, potparm, &
                                  atomic_kind_set, eps_spline=ff_type%eps_spline, &
                                  max_energy=ff_type%max_energy, rlow_nb=ff_type%rlow_nb, &
                                  emax_spline=ff_type%emax_spline, npoints=ff_type%npoints, iw=iw2, iw2=iw3, iw3=iw4, &
                                  do_zbl=do_zbl, shift_cutoff=ff_type%shift_cutoff, &
                                  nonbonded_type=nonbonded_type)
      ! Let the pointers on potparm point to the splines generated in
      ! spline_nonbond_control.
      DO ikind = 1, SIZE(potparm%pot, 1)
         DO jkind = ikind, SIZE(potparm%pot, 2)
            n = spline_env%spltab(ikind, jkind)
            spl_p => spline_env%spl_pp(n)%spl_p
            CALL spline_data_p_retain(spl_p)
            CALL spline_data_p_release(potparm%pot(ikind, jkind)%pot%pair_spline_data)
            potparm%pot(ikind, jkind)%pot%pair_spline_data => spl_p
         END DO
      END DO
      CALL spline_env_release(spline_env)
      CALL timestop(handle2)

   END SUBROUTINE force_field_pack_splines

! **************************************************************************************************
!> \brief Compute the electrostatic interaction cutoffs
!> \param atomic_kind_set ...
!> \param ff_type ...
!> \param potparm_nonbond ...
!> \param ewald_env ...
!> \author Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE force_field_pack_eicut(atomic_kind_set, ff_type, &
                                     potparm_nonbond, ewald_env)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(force_field_type), INTENT(IN)                 :: ff_type
      TYPE(pair_potential_pp_type), POINTER              :: potparm_nonbond
      TYPE(ewald_environment_type), POINTER              :: ewald_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack_eicut', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, handle, i1, i2, nkinds
      REAL(KIND=dp)                                      :: alpha, beta, mm_radius1, mm_radius2, &
                                                            rcut2, rcut2_ewald, tmp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: interaction_cutoffs
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      CALL timeset(routineN, handle)

      tmp = 0.0_dp
      nkinds = SIZE(atomic_kind_set)
      ! allocate the array with interaction cutoffs for the electrostatics, used
      ! to make the electrostatic interaction continuous at ewald_env%rcut
      ALLOCATE (interaction_cutoffs(3, nkinds, nkinds))
      interaction_cutoffs = 0.0_dp

      ! compute the interaction cutoff if SHIFT_CUTOFF is active
      IF (ff_type%shift_cutoff) THEN
         CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type, &
                            rcut=rcut2_ewald)
         rcut2_ewald = rcut2_ewald*rcut2_ewald
         DO i1 = 1, nkinds
            atomic_kind => atomic_kind_set(i1)
            CALL get_atomic_kind(atomic_kind=atomic_kind, mm_radius=mm_radius1)
            DO i2 = 1, nkinds
               rcut2 = rcut2_ewald
               IF (ASSOCIATED(potparm_nonbond)) THEN
                  rcut2 = MAX(potparm_nonbond%pot(i1, i2)%pot%rcutsq, rcut2_ewald)
               END IF
               IF (rcut2 > 0) THEN
                  atomic_kind => atomic_kind_set(i2)
                  CALL get_atomic_kind(atomic_kind=atomic_kind, mm_radius=mm_radius2)
                  ! cutoff for core-core
                  interaction_cutoffs(1, i1, i2) = potential_coulomb(rcut2, tmp, &
                                                                     1.0_dp, ewald_type, alpha, 0.0_dp, 0.0_dp)
                  ! cutoff for core-shell, core-ion, shell-core or ion-core
                  IF (mm_radius1 > 0.0_dp) THEN
                     beta = sqrthalf/mm_radius1
                  ELSE
                     beta = 0.0_dp
                  END IF
                  interaction_cutoffs(2, i1, i2) = potential_coulomb(rcut2, tmp, &
                                                                     1.0_dp, ewald_type, alpha, beta, 0.0_dp)
                  ! cutoff for shell-shell or ion-ion
                  IF (mm_radius1+mm_radius2 > 0.0_dp) THEN
                     beta = sqrthalf/SQRT(mm_radius1*mm_radius1+mm_radius2*mm_radius2)
                  ELSE
                     beta = 0.0_dp
                  END IF
                  interaction_cutoffs(3, i1, i2) = potential_coulomb(rcut2, tmp, &
                                                                     1.0_dp, ewald_type, alpha, beta, 0.0_dp)
               END IF
            END DO
         END DO
      END IF

      CALL ewald_env_set(ewald_env, interaction_cutoffs=interaction_cutoffs)

      CALL timestop(handle)
   END SUBROUTINE force_field_pack_eicut

! **************************************************************************************************
!> \brief Issues on screen a warning when repetitions are present in the
!>        definition of the forcefield
!> \param found ...
!> \param tag_label ...
!> \param name_atm_a ...
!> \param name_atm_b ...
!> \param name_atm_c ...
!> \param name_atm_d ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE issue_duplications(found, tag_label, name_atm_a, name_atm_b, &
                                 name_atm_c, name_atm_d)

      LOGICAL, INTENT(IN)                                :: found
      CHARACTER(LEN=*), INTENT(IN)                       :: tag_label, name_atm_a
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name_atm_b, name_atm_c, name_atm_d

      CHARACTER(len=*), PARAMETER :: routineN = 'issue_duplications', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: item

      item = "( "//TRIM(name_atm_a)
      IF (PRESENT(name_atm_b)) THEN
         item = TRIM(item)//" , "//TRIM(name_atm_b)
      END IF
      IF (PRESENT(name_atm_c)) THEN
         item = TRIM(item)//" , "//TRIM(name_atm_c)
      END IF
      IF (PRESENT(name_atm_d)) THEN
         item = TRIM(item)//" , "//TRIM(name_atm_d)
      END IF
      item = TRIM(item)//" )"
      IF (found) &
         CALL cp__w("force_fields_all.F",3322,"Multiple "//TRIM(tag_label)//" declarations: "//TRIM(item)//" overwriting! ")

   END SUBROUTINE issue_duplications

! **************************************************************************************************
!> \brief Store informations on possible missing ForceFields parameters
!> \param atm1 ...
!> \param atm2 ...
!> \param atm3 ...
!> \param atm4 ...
!> \param type_name ...
!> \param fatal ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE store_FF_missing_par(atm1, atm2, atm3, atm4, type_name, fatal, array)
      CHARACTER(LEN=*), INTENT(IN)                       :: atm1
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: atm2, atm3, atm4
      CHARACTER(LEN=*), INTENT(IN)                       :: type_name
      LOGICAL, INTENT(INOUT), OPTIONAL                   :: fatal
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: array

      CHARACTER(len=*), PARAMETER :: routineN = 'store_FF_missing_par', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: sfmt
      CHARACTER(LEN=4)                                   :: my_atm1, my_atm2, my_atm3, my_atm4
      CHARACTER(LEN=default_path_length)                 :: my_format
      INTEGER                                            :: fmt, i, nsize
      LOGICAL                                            :: found

      nsize = 0
      fmt = 1
      my_format = '(T2,"FORCEFIELD| Missing ","'//TRIM(type_name)// &
                  '",T40,"(",A4,")")'
      IF (PRESENT(atm2)) fmt = fmt+1
      IF (PRESENT(atm3)) fmt = fmt+1
      IF (PRESENT(atm4)) fmt = fmt+1
      CALL integer_to_string(fmt-1, sfmt)
      IF (fmt > 1) &
         my_format = '(T2,"FORCEFIELD| Missing ","'//TRIM(type_name)// &
                     '",T40,"(",A4,'//TRIM(sfmt)//'(",",A4),")")'
      IF (PRESENT(fatal)) fatal = .TRUE.
      ! Check for previous already stored equal force fields
      IF (ASSOCIATED(array)) nsize = SIZE(array)
      found = .FALSE.
      IF (nsize >= 1) THEN
         DO i = 1, nsize
            SELECT CASE (type_name)
            CASE ("Bond")
               IF (INDEX(array(i) (21:39), "Bond") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2)) .OR. &
                   ((atm1 == my_atm2) .AND. (atm2 == my_atm1))) found = .TRUE.
            CASE ("Angle")
               IF (INDEX(array(i) (21:39), "Angle") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               my_atm3 = array(i) (51:54)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               CALL compress(my_atm3, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm3)) .OR. &
                   ((atm1 == my_atm3) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm1))) &
                  found = .TRUE.
            CASE ("Urey-Bradley")
               IF (INDEX(array(i) (21:39), "Urey-Bradley") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               my_atm3 = array(i) (51:54)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               CALL compress(my_atm3, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm3)) .OR. &
                   ((atm1 == my_atm3) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm1))) &
                  found = .TRUE.
            CASE ("Torsion")
               IF (INDEX(array(i) (21:39), "Torsion") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               my_atm3 = array(i) (51:54)
               my_atm4 = array(i) (56:59)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               CALL compress(my_atm3, .TRUE.)
               CALL compress(my_atm4, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm3) .AND. (atm4 == my_atm4)) .OR. &
                   ((atm1 == my_atm4) .AND. (atm2 == my_atm3) .AND. (atm3 == my_atm2) .AND. (atm4 == my_atm1))) &
                  found = .TRUE.
            CASE ("Improper")
               IF (INDEX(array(i) (21:39), "Improper") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               my_atm3 = array(i) (51:54)
               my_atm4 = array(i) (56:59)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               CALL compress(my_atm3, .TRUE.)
               CALL compress(my_atm4, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm3) .AND. (atm4 == my_atm4)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm3) .AND. (atm3 == my_atm2) .AND. (atm4 == my_atm4)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm3) .AND. (atm3 == my_atm4) .AND. (atm4 == my_atm3)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm4) .AND. (atm3 == my_atm3) .AND. (atm4 == my_atm2)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm4) .AND. (atm3 == my_atm2) .AND. (atm4 == my_atm3)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm4) .AND. (atm4 == my_atm3))) &
                  found = .TRUE.

            CASE ("Out of plane bend")
               IF (INDEX(array(i) (21:39), "Out of plane bend") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               my_atm3 = array(i) (51:54)
               my_atm4 = array(i) (56:59)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               CALL compress(my_atm3, .TRUE.)
               CALL compress(my_atm4, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2) .AND. (atm3 == my_atm3) .AND. (atm4 == my_atm4)) .OR. &
                   ((atm1 == my_atm1) .AND. (atm2 == my_atm3) .AND. (atm3 == my_atm2) .AND. (atm4 == my_atm4))) &
                  found = .TRUE.

            CASE ("Charge")
               IF (INDEX(array(i) (21:39), "Charge") == 0) CYCLE
               my_atm1 = array(i) (41:44)
               CALL compress(my_atm1, .TRUE.)
               IF (atm1 == my_atm1) found = .TRUE.
            CASE ("Spline_Bond_Env", "Spline_Non_Bond_Env")
               IF (INDEX(array(i) (21:39), "Spline_") == 0) CYCLE
               fmt = 0
               my_atm1 = array(i) (41:44)
               my_atm2 = array(i) (46:49)
               CALL compress(my_atm1, .TRUE.)
               CALL compress(my_atm2, .TRUE.)
               IF (((atm1 == my_atm1) .AND. (atm2 == my_atm2)) .OR. &
                   ((atm1 == my_atm2) .AND. (atm2 == my_atm1))) found = .TRUE.
            CASE DEFAULT
               ! Should never reach this point
               CALL cp__b("force_fields_all.F",3462,"")
            END SELECT
            IF (found) EXIT
         END DO
      ENDIF
      IF (.NOT. found) THEN
         nsize = nsize+1
         CALL reallocate(array, 1, nsize)
         SELECT CASE (fmt)
         CASE (1)
            WRITE (array(nsize), FMT=TRIM(my_format)) atm1
         CASE (2)
            WRITE (array(nsize), FMT=TRIM(my_format)) atm1, atm2
         CASE (3)
            WRITE (array(nsize), FMT=TRIM(my_format)) atm1, atm2, atm3
         CASE (4)
            WRITE (array(nsize), FMT=TRIM(my_format)) atm1, atm2, atm3, atm4
         END SELECT
      END IF

   END SUBROUTINE store_FF_missing_par

END MODULE force_fields_all

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Subroutine input_torsions changed (DG) 05-Dec-2000
!>      Output formats changed (DG) 05-Dec-2000
!>      JGH (26-01-2002) : force field parameters stored in tables, not in
!>        matrices. Input changed to have parameters labeled by the position
!>        and not atom pairs (triples etc)
!>      Teo (11.2005) : Moved all information on force field  pair_potential to
!>                      a much lighter memory structure
!> \author CJM
! **************************************************************************************************
MODULE force_fields_util

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE colvar_types,                    ONLY: dist_colvar_id
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_create,&
                                              fist_nonbond_env_set,&
                                              fist_nonbond_env_type
   USE force_field_kind_types,          ONLY: &
        allocate_bend_kind_set, allocate_bond_kind_set, allocate_impr_kind_set, &
        allocate_opbend_kind_set, allocate_torsion_kind_set, allocate_ub_kind_set, bend_kind_type, &
        bond_kind_type, deallocate_bend_kind_set, deallocate_bond_kind_set, do_ff_undef, &
        impr_kind_dealloc_ref, impr_kind_type, opbend_kind_type, torsion_kind_dealloc_ref, &
        torsion_kind_type, ub_kind_dealloc_ref, ub_kind_type
   USE force_field_types,               ONLY: amber_info_type,&
                                              charmm_info_type,&
                                              force_field_type,&
                                              gromos_info_type,&
                                              input_info_type
   USE force_fields_all,                ONLY: &
        force_field_pack_bend, force_field_pack_bond, force_field_pack_charge, &
        force_field_pack_charges, force_field_pack_damp, force_field_pack_eicut, &
        force_field_pack_impr, force_field_pack_nonbond, force_field_pack_nonbond14, &
        force_field_pack_opbend, force_field_pack_pol, force_field_pack_radius, &
        force_field_pack_shell, force_field_pack_splines, force_field_pack_tors, &
        force_field_pack_ub, force_field_unique_bend, force_field_unique_bond, &
        force_field_unique_impr, force_field_unique_opbend, force_field_unique_tors, &
        force_field_unique_ub
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE molecule_kind_types,             ONLY: &
        atom_type, bend_type, bond_type, colvar_constraint_type, g3x3_constraint_type, &
        g4x6_constraint_type, get_molecule_kind, impr_type, molecule_kind_type, opbend_type, &
        set_molecule_kind, torsion_type, ub_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE pair_potential_types,            ONLY: pair_potential_pp_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE shell_potential_types,           ONLY: shell_kind_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_fields_util'

   PRIVATE
   PUBLIC :: force_field_pack, &
             force_field_qeff_output, &
             clean_intra_force_kind, &
             get_generic_info

CONTAINS

! **************************************************************************************************
!> \brief Pack in all the information needed for the force_field
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param ewald_env ...
!> \param fist_nonbond_env ...
!> \param ff_type ...
!> \param root_section ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param mm_section ...
!> \param subsys_section ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param cell ...
! **************************************************************************************************
   SUBROUTINE force_field_pack(particle_set, atomic_kind_set, molecule_kind_set, &
                               molecule_set, ewald_env, fist_nonbond_env, ff_type, root_section, qmmm, &
                               qmmm_env, mm_section, subsys_section, shell_particle_set, core_particle_set, &
                               cell)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(section_vals_type), POINTER                   :: root_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: mm_section, subsys_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: shell_particle_set, core_particle_set
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_pack', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: Ainfo
      INTEGER                                            :: handle, iw, iw2, iw3, iw4, output_unit
      LOGICAL                                            :: do_zbl, explicit, fatal, ignore_fatal, &
                                                            my_qmmm
      REAL(KIND=dp)                                      :: ewald_rcut, verlet_skin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      TYPE(amber_info_type), POINTER                     :: amb_info
      TYPE(charmm_info_type), POINTER                    :: chm_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gromos_info_type), POINTER                    :: gro_info
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(pair_potential_pp_type), POINTER              :: potparm_nonbond, potparm_nonbond14
      TYPE(section_vals_type), POINTER                   :: charges_section

      CALL timeset(routineN, handle)
      fatal = .FALSE.
      ignore_fatal = ff_type%ignore_missing_critical
      NULLIFY (logger, Ainfo, charges_section, charges)
      logger => cp_get_default_logger()
      ! Error unit
      output_unit = cp_logger_get_default_io_unit(logger)

      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")
      iw2 = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO/SPLINE_INFO", &
                                 extension=".mmLog")
      iw3 = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO/SPLINE_DATA", &
                                 extension=".mmLog")
      iw4 = cp_print_key_unit_nr(logger, mm_section, "PRINT%PROGRAM_RUN_INFO", &
                                 extension=".mmLog")
      NULLIFY (potparm_nonbond14, potparm_nonbond)
      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env)) my_qmmm = qmmm
      inp_info => ff_type%inp_info
      chm_info => ff_type%chm_info
      gro_info => ff_type%gro_info
      amb_info => ff_type%amb_info
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Determine the number of unique bond kind and allocate bond_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_bond(particle_set, molecule_kind_set, molecule_set, &
                                   ff_type)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Determine the number of unique bend kind and allocate bend_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_bend(particle_set, molecule_kind_set, molecule_set, &
                                   ff_type)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Determine the number of unique Urey-Bradley kind and allocate ub_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_ub(particle_set, molecule_kind_set, molecule_set)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Determine the number of unique torsion kind and allocate torsion_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_tors(particle_set, molecule_kind_set, molecule_set, &
                                   ff_type)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Determine the number of unique impr kind and allocate impr_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_impr(particle_set, molecule_kind_set, molecule_set, &
                                   ff_type)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Determine the number of unique opbend kind and allocate opbend_kind_set
      !-----------------------------------------------------------------------------
      CALL force_field_unique_opbend(particle_set, molecule_kind_set, molecule_set, &
                                     ff_type)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Bonds
      !-----------------------------------------------------------------------------
      CALL force_field_pack_bond(particle_set, molecule_kind_set, molecule_set, &
                                 fatal, Ainfo, chm_info, inp_info, gro_info, amb_info)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 8. Bends
      !-----------------------------------------------------------------------------
      CALL force_field_pack_bend(particle_set, molecule_kind_set, molecule_set, &
                                 fatal, Ainfo, chm_info, inp_info, gro_info, amb_info)
      ! Give information and abort if any bond or angle parameter is missing..
      CALL release_FF_missing_par(fatal, ignore_fatal, AInfo, output_unit, iw)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 9. Urey-Bradley
      !-----------------------------------------------------------------------------
      CALL force_field_pack_ub(particle_set, molecule_kind_set, molecule_set, &
                               Ainfo, chm_info, inp_info, iw)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 10. Torsions
      !-----------------------------------------------------------------------------
      CALL force_field_pack_tors(particle_set, molecule_kind_set, molecule_set, &
                                 Ainfo, chm_info, inp_info, gro_info, amb_info, iw)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 11. Impropers
      !-----------------------------------------------------------------------------
      CALL force_field_pack_impr(particle_set, molecule_kind_set, molecule_set, &
                                 Ainfo, chm_info, inp_info, gro_info)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 12. Out of plane bends
      !-----------------------------------------------------------------------------
      CALL force_field_pack_opbend(particle_set, molecule_kind_set, molecule_set, &
                                   Ainfo, inp_info)
      ! Give information only if any Urey-Bradley, Torsion, improper or opbend is missing
      ! continue calculation..
      CALL release_FF_missing_par(fatal, ignore_fatal, AInfo, output_unit, iw)

      charges_section => section_vals_get_subs_vals(mm_section, "FORCEFIELD%CHARGES")
      CALL section_vals_get(charges_section, explicit=explicit)
      IF (.NOT. explicit) THEN
         !-----------------------------------------------------------------------------
         !-----------------------------------------------------------------------------
         ! 13.a Set up atomic_kind_set()%fist_potentail%[qeff] and shell
         !      potential parameters
         !-----------------------------------------------------------------------------
         CALL force_field_pack_charge(atomic_kind_set, qmmm_env, fatal, iw, iw4, &
                                      Ainfo, my_qmmm, inp_info)
         ! Give information only if charge is missing and abort..
         CALL release_FF_missing_par(fatal, ignore_fatal, AInfo, output_unit, iw)
      ELSE
         !-----------------------------------------------------------------------------
         !-----------------------------------------------------------------------------
         ! 13.b Setup a global array of classical charges - this avoids the packing and
         !      allows the usage of different charges for same atomic types
         !-----------------------------------------------------------------------------
         CALL force_field_pack_charges(charges, charges_section, particle_set, my_qmmm, &
                                       qmmm_env, inp_info, iw4)
      END IF

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 14. Set up the radius of the electrostatic multipole in Fist
      !-----------------------------------------------------------------------------
      CALL force_field_pack_radius(atomic_kind_set, iw, subsys_section)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 15. Set up the polarizable FF parameters
      !-----------------------------------------------------------------------------
      CALL force_field_pack_pol(atomic_kind_set, iw, inp_info)
      CALL force_field_pack_damp(atomic_kind_set, iw, inp_info)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 16. Set up  Shell potential
      !-----------------------------------------------------------------------------
      CALL force_field_pack_shell(particle_set, atomic_kind_set, &
                                  molecule_kind_set, molecule_set, root_section, subsys_section, &
                                  shell_particle_set, core_particle_set, cell, iw, inp_info)
      IF (ff_type%do_nonbonded) THEN
         !-----------------------------------------------------------------------------
         !-----------------------------------------------------------------------------
         ! 17. Set up potparm_nonbond14
         !-----------------------------------------------------------------------------
         ! Move the data from the info structures to potparm_nonbond
         CALL force_field_pack_nonbond14(atomic_kind_set, ff_type, qmmm_env, iw, Ainfo, &
                                         chm_info, inp_info, gro_info, amb_info, potparm_nonbond14, ewald_env)
         ! Give information if any 1-4 is missing.. continue calculation..
         CALL release_FF_missing_par(fatal, ignore_fatal, AInfo, output_unit, iw)
         ! Create the spline data
         CALL section_vals_val_get(mm_section, "FORCEFIELD%ZBL_SCATTERING", l_val=do_zbl)
         CALL force_field_pack_splines(atomic_kind_set, ff_type, iw2, iw3, iw4, &
                                       potparm_nonbond14, do_zbl, nonbonded_type="NONBONDED14")
         !-----------------------------------------------------------------------------
         !-----------------------------------------------------------------------------
         ! 18. Set up potparm_nonbond
         !-----------------------------------------------------------------------------
         ! Move the data from the info structures to potparm_nonbond
         CALL force_field_pack_nonbond(atomic_kind_set, ff_type, qmmm_env, &
                                       fatal, iw, Ainfo, chm_info, inp_info, gro_info, amb_info, &
                                       potparm_nonbond, ewald_env)
         ! Give information and abort if any pair potential spline is missing..
         CALL release_FF_missing_par(fatal, ignore_fatal, AInfo, output_unit, iw)
         ! Create the spline data
         CALL section_vals_val_get(mm_section, "FORCEFIELD%ZBL_SCATTERING", l_val=do_zbl)
         CALL force_field_pack_splines(atomic_kind_set, ff_type, iw2, iw3, iw4, &
                                       potparm_nonbond, do_zbl, nonbonded_type="NONBONDED")
      END IF
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 19. Create nonbond environment
      !-----------------------------------------------------------------------------
      CALL ewald_env_get(ewald_env, rcut=ewald_rcut)
      CALL section_vals_val_get(mm_section, "NEIGHBOR_LISTS%VERLET_SKIN", &
                                r_val=verlet_skin)
      CALL fist_nonbond_env_create(fist_nonbond_env, atomic_kind_set, &
                                   potparm_nonbond14, potparm_nonbond, ff_type%do_nonbonded, &
                                   verlet_skin, ewald_rcut, ff_type%ei_scale14, &
                                   ff_type%vdw_scale14, ff_type%shift_cutoff)
      CALL fist_nonbond_env_set(fist_nonbond_env, charges=charges)
      ! Compute the electrostatic interaction cutoffs.
      CALL force_field_pack_eicut(atomic_kind_set, ff_type, potparm_nonbond, &
                                  ewald_env)

      CALL cp_print_key_finished_output(iw4, logger, mm_section, "PRINT%PROGRAM_RUN_INFO")
      CALL cp_print_key_finished_output(iw3, logger, mm_section, "PRINT%FF_INFO/SPLINE_DATA")
      CALL cp_print_key_finished_output(iw2, logger, mm_section, "PRINT%FF_INFO/SPLINE_INFO")
      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%FF_INFO")
      CALL timestop(handle)

   END SUBROUTINE force_field_pack

! **************************************************************************************************
!> \brief Store informations on possible missing ForceFields parameters
!> \param fatal ...
!> \param ignore_fatal ...
!> \param array ...
!> \param output_unit ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE release_FF_missing_par(fatal, ignore_fatal, array, output_unit, iw)
      LOGICAL, INTENT(INOUT)                             :: fatal, ignore_fatal
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: array
      INTEGER, INTENT(IN)                                :: output_unit, iw

      CHARACTER(len=*), PARAMETER :: routineN = 'release_FF_missing_par', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(array)) THEN
         IF (output_unit > 0) THEN
            WRITE (output_unit, *)
            WRITE (output_unit, '(T2,"FORCEFIELD|",A)') &
               " WARNING: A non Critical ForceField parameter is missing! CP2K GOES!", &
               " Non critical parameters are:Urey-Bradley,Torsions,Impropers, Opbends and 1-4", &
               " All missing parameters will not contribute to the potential energy!"
            IF (fatal .OR. iw > 0) THEN
               WRITE (output_unit, *)
               DO i = 1, SIZE(array)
                  WRITE (output_unit, '(A)') array(i)
               END DO
            END IF
            IF (.NOT. fatal .AND. iw < 0) THEN
               WRITE (output_unit, '(T2,"FORCEFIELD|",A)') &
                  " Activate the print key FF_INFO to have a list of missing parameters"
               WRITE (output_unit, *)
            END IF
         END IF
         DEALLOCATE (array)
      END IF
      IF (fatal) THEN
         IF (ignore_fatal) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, *)
               WRITE (output_unit, '(T2,"FORCEFIELD|",A)') &
                  " WARNING: Ignoring missing critical FF parameters! CP2K GOES!", &
                  " Critical parameters are: Bonds, Bends and Charges", &
                  " All missing parameters will not contribute to the potential energy!"
            END IF
         ELSE
            CALL cp__b("force_fields_util.F",387,"Missing critical ForceField parameters!")
         END IF
      END IF
   END SUBROUTINE release_FF_missing_par

! **************************************************************************************************
!> \brief Compute the total qeff charges for each molecule kind and total system
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param mm_section ...
!> \param charges ...
! **************************************************************************************************
   SUBROUTINE force_field_qeff_output(particle_set, molecule_kind_set, &
                                      molecule_set, mm_section, charges)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(section_vals_type), POINTER                   :: mm_section
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'force_field_qeff_output', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atmname, molname
      INTEGER                                            :: first, handle, iatom, imol, iw, j, jatom
      LOGICAL                                            :: shell_active
      REAL(KIND=dp)                                      :: mass, mass_mol, mass_sum, qeff, &
                                                            qeff_mol, qeff_sum
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(shell_kind_type), POINTER                     :: shell

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")

      qeff = 0.0_dp
      qeff_mol = 0.0_dp
      qeff_sum = 0.0_dp
      mass_sum = 0.0_dp

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Sum of [qeff,mass] for each molecule_kind
      !-----------------------------------------------------------------------------
      DO imol = 1, SIZE(molecule_kind_set)
         qeff_mol = 0.0_dp
         mass_mol = 0.0_dp
         molecule_kind => molecule_kind_set(imol)

         j = molecule_kind_set(imol)%molecule_list(1)
         molecule => molecule_set(j)
         CALL get_molecule(molecule=molecule, first_atom=first)

         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                name=molname, atom_list=atom_list)
         DO iatom = 1, SIZE(atom_list)
            atomic_kind => atom_list(iatom)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 name=atmname, qeff=qeff, mass=mass, shell_active=shell_active, shell=shell)
            IF (shell_active) qeff = shell%charge_core+shell%charge_shell
            IF (ASSOCIATED(charges)) THEN
               jatom = first-1+iatom
               qeff = charges(jatom)
            END IF
            IF (iw > 0) WRITE (iw, *) "      atom ", iatom, " ", TRIM(atmname), " charge = ", qeff
            qeff_mol = qeff_mol+qeff
            mass_mol = mass_mol+mass
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, charge=qeff_mol, mass=mass_mol)
         IF (iw > 0) WRITE (iw, *) "    Mol Kind ", TRIM(molname), " charge = ", qeff_mol
      END DO

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Sum of [qeff,mass] for particle_set
      !-----------------------------------------------------------------------------
      DO iatom = 1, SIZE(particle_set)
         atomic_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              name=atmname, qeff=qeff, mass=mass, shell_active=shell_active, shell=shell)
         IF (shell_active) qeff = shell%charge_core+shell%charge_shell
         IF (ASSOCIATED(charges)) THEN
            qeff = charges(iatom)
         END IF
         IF (iw > 0) WRITE (iw, *) "      atom ", iatom, " ", TRIM(atmname), &
            " charge = ", qeff
         qeff_sum = qeff_sum+qeff
         mass_sum = mass_sum+mass
      END DO
      IF (iw > 0) WRITE (iw, '(A,F20.10)') "    Total system charge = ", qeff_sum
      IF (iw > 0) WRITE (iw, '(A,F20.10)') "    Total system mass   = ", mass_sum

      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%FF_INFO")
      CALL timestop(handle)
   END SUBROUTINE force_field_qeff_output

! **************************************************************************************************
!> \brief Removes UNSET force field types
!> \param molecule_kind_set ...
!> \param mm_section ...
! **************************************************************************************************
   SUBROUTINE clean_intra_force_kind(molecule_kind_set, mm_section)

      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'clean_intra_force_kind', &
         routineP = moduleN//':'//routineN

      INTEGER :: atm2_a, atm2_b, atm2_c, atm_a, atm_b, atm_c, atm_d, counter, handle, i, ibend, &
         ibond, icolv, ig3x3, ig4x6, iimpr, ikind, iopbend, itorsion, iub, iw, j, k, nbend, nbond, &
         newkind, ng3x3, ng4x6, nimpr, nopbend, ntorsion, nub
      INTEGER, POINTER                                   :: bad1(:), bad2(:)
      LOGICAL                                            :: unsetme, valid_kind
      TYPE(bend_kind_type), DIMENSION(:), POINTER        :: bend_kind_set, new_bend_kind_set
      TYPE(bend_type), DIMENSION(:), POINTER             :: bend_list, new_bend_list
      TYPE(bond_kind_type), DIMENSION(:), POINTER        :: bond_kind_set, new_bond_kind_set
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list, new_bond_list
      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER  :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER  :: g4x6_list
      TYPE(impr_kind_type), DIMENSION(:), POINTER        :: impr_kind_set, new_impr_kind_set
      TYPE(impr_type), DIMENSION(:), POINTER             :: impr_list, new_impr_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(opbend_kind_type), DIMENSION(:), POINTER      :: new_opbend_kind_set, opbend_kind_set
      TYPE(opbend_type), DIMENSION(:), POINTER           :: new_opbend_list, opbend_list
      TYPE(torsion_kind_type), DIMENSION(:), POINTER     :: new_torsion_kind_set, torsion_kind_set
      TYPE(torsion_type), DIMENSION(:), POINTER          :: new_torsion_list, torsion_list
      TYPE(ub_kind_type), DIMENSION(:), POINTER          :: new_ub_kind_set, ub_kind_set
      TYPE(ub_type), DIMENSION(:), POINTER               :: new_ub_list, ub_list

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Lets Tag the unwanted bonds due to the use of distance constraint
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                colv_list=colv_list, &
                                nbond=nbond, &
                                bond_list=bond_list)
         IF (ASSOCIATED(colv_list)) THEN
            DO icolv = 1, SIZE(colv_list)
               IF ((colv_list(icolv)%type_id == dist_colvar_id) .AND. &
                   ((.NOT. colv_list(icolv)%use_points) .OR. (SIZE(colv_list(icolv)%i_atoms) == 2))) THEN
                  atm_a = colv_list(icolv)%i_atoms(1)
                  atm_b = colv_list(icolv)%i_atoms(2)
                  DO ibond = 1, nbond
                     unsetme = .FALSE.
                     atm2_a = bond_list(ibond)%a
                     atm2_b = bond_list(ibond)%b
                     IF (atm2_a == atm_a .AND. atm2_b == atm_b) unsetme = .TRUE.
                     IF (atm2_a == atm_b .AND. atm2_b == atm_a) unsetme = .TRUE.
                     IF (unsetme) bond_list(ibond)%id_type = do_ff_undef
                  END DO
               END IF
            END DO
         END IF
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Lets Tag the unwanted bends due to the use of distance constraint
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                colv_list=colv_list, &
                                nbend=nbend, &
                                bend_list=bend_list)
         IF (ASSOCIATED(colv_list)) THEN
            DO ibend = 1, nbend
               unsetme = .FALSE.
               atm_a = bend_list(ibend)%a
               atm_b = bend_list(ibend)%b
               atm_c = bend_list(ibend)%c
               DO icolv = 1, SIZE(colv_list)
                  IF ((colv_list(icolv)%type_id == dist_colvar_id) .AND. &
                      ((.NOT. colv_list(icolv)%use_points) .OR. (SIZE(colv_list(icolv)%i_atoms) == 2))) THEN
                     atm2_a = colv_list(icolv)%i_atoms(1)
                     atm2_b = colv_list(icolv)%i_atoms(2)
                     ! Check that the bonds we're constraining does not involve atoms defining
                     ! a bend..
                     IF (((atm2_a == atm_a) .AND. (atm2_b == atm_c)) .OR. &
                         ((atm2_a == atm_c) .AND. (atm2_b == atm_a))) THEN
                        unsetme = .TRUE.
                        EXIT
                     END IF
                  END IF
               END DO
               IF (unsetme) bend_list(ibend)%id_type = do_ff_undef
            END DO
         END IF
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Lets Tag the unwanted bonds due to the use of 3x3
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                ng3x3=ng3x3, &
                                g3x3_list=g3x3_list, &
                                nbond=nbond, &
                                bond_list=bond_list)
         DO ig3x3 = 1, ng3x3
            atm_a = g3x3_list(ig3x3)%a
            atm_b = g3x3_list(ig3x3)%b
            atm_c = g3x3_list(ig3x3)%c
            DO ibond = 1, nbond
               unsetme = .FALSE.
               atm2_a = bond_list(ibond)%a
               atm2_b = bond_list(ibond)%b
               IF (atm2_a == atm_a .AND. atm2_b == atm_b) unsetme = .TRUE.
               IF (atm2_a == atm_a .AND. atm2_b == atm_c) unsetme = .TRUE.
               IF (atm2_a == atm_c .AND. atm2_b == atm_c) unsetme = .TRUE.
               IF (unsetme) bond_list(ibond)%id_type = do_ff_undef
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Lets Tag the unwanted bends due to the use of 3x3
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                ng3x3=ng3x3, &
                                g3x3_list=g3x3_list, &
                                nbend=nbend, &
                                bend_list=bend_list)
         DO ig3x3 = 1, ng3x3
            atm_a = g3x3_list(ig3x3)%a
            atm_b = g3x3_list(ig3x3)%b
            atm_c = g3x3_list(ig3x3)%c
            DO ibend = 1, nbend
               unsetme = .FALSE.
               atm2_a = bend_list(ibend)%a
               atm2_b = bend_list(ibend)%b
               atm2_c = bend_list(ibend)%c
               IF (atm2_a == atm_a .AND. atm2_b == atm_b .AND. atm2_c == atm_c) unsetme = .TRUE.
               IF (atm2_a == atm_a .AND. atm2_b == atm_c .AND. atm2_c == atm_b) unsetme = .TRUE.
               IF (atm2_a == atm_b .AND. atm2_b == atm_a .AND. atm2_c == atm_c) unsetme = .TRUE.
               IF (atm2_a == atm_b .AND. atm2_b == atm_c .AND. atm2_c == atm_a) unsetme = .TRUE.
               IF (unsetme) bend_list(ibend)%id_type = do_ff_undef
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Lets Tag the unwanted bonds due to the use of 4x6
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                ng4x6=ng4x6, &
                                g4x6_list=g4x6_list, &
                                nbond=nbond, &
                                bond_list=bond_list)
         DO ig4x6 = 1, ng4x6
            atm_a = g4x6_list(ig4x6)%a
            atm_b = g4x6_list(ig4x6)%b
            atm_c = g4x6_list(ig4x6)%c
            atm_d = g4x6_list(ig4x6)%d
            DO ibond = 1, nbond
               unsetme = .FALSE.
               atm2_a = bond_list(ibond)%a
               atm2_b = bond_list(ibond)%b
               IF (atm2_a == atm_a .AND. atm2_b == atm_b) unsetme = .TRUE.
               IF (atm2_a == atm_a .AND. atm2_b == atm_c) unsetme = .TRUE.
               IF (atm2_a == atm_a .AND. atm2_b == atm_d) unsetme = .TRUE.
               IF (unsetme) bond_list(ibond)%id_type = do_ff_undef
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Lets Tag the unwanted bends due to the use of 4x6
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                ng4x6=ng4x6, &
                                g4x6_list=g4x6_list, &
                                nbend=nbend, &
                                bend_list=bend_list)
         DO ig4x6 = 1, ng4x6
            atm_a = g4x6_list(ig4x6)%a
            atm_b = g4x6_list(ig4x6)%b
            atm_c = g4x6_list(ig4x6)%c
            atm_d = g4x6_list(ig4x6)%d
            DO ibend = 1, nbend
               unsetme = .FALSE.
               atm2_a = bend_list(ibend)%a
               atm2_b = bend_list(ibend)%b
               atm2_c = bend_list(ibend)%c
               IF (atm2_a == atm_b .AND. atm2_b == atm_a .AND. atm2_c == atm_c) unsetme = .TRUE.
               IF (atm2_a == atm_b .AND. atm2_b == atm_a .AND. atm2_c == atm_d) unsetme = .TRUE.
               IF (atm2_a == atm_c .AND. atm2_b == atm_a .AND. atm2_c == atm_d) unsetme = .TRUE.
               IF (unsetme) bend_list(ibend)%id_type = do_ff_undef
            END DO
         END DO
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Count the number of UNSET bond kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                nbond=nbond, &
                                bond_kind_set=bond_kind_set, &
                                bond_list=bond_list)
         IF (nbond > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old BOND Count: ", &
               SIZE(bond_list), SIZE(bond_kind_set)
            IF (iw > 0) WRITE (iw, '(2I6)') (bond_list(ibond)%a, bond_list(ibond)%b, ibond=1, SIZE(bond_list))
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(bond_kind_set)))
            bad1(:) = 0
            DO ibond = 1, SIZE(bond_kind_set)
               unsetme = .FALSE.
               IF (bond_kind_set(ibond)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(bond_list)
                  IF (bond_list(i)%id_type /= do_ff_undef .AND. &
                      bond_list(i)%bond_kind%kind_number == ibond) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(ibond) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(bond_kind_set)-SUM(bad1)
               CALL allocate_bond_kind_set(new_bond_kind_set, counter)
               counter = 0
               DO ibond = 1, SIZE(bond_kind_set)
                  IF (bad1(ibond) == 0) THEN
                     counter = counter+1
                     new_bond_kind_set(counter) = bond_kind_set(ibond)
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(bond_list)))
               bad2(:) = 0
               DO ibond = 1, SIZE(bond_list)
                  unsetme = .FALSE.
                  IF (bond_list(ibond)%bond_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (bond_list(ibond)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(ibond) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(bond_list)-SUM(bad2)
                  ALLOCATE (new_bond_list(counter))
                  counter = 0
                  DO ibond = 1, SIZE(bond_list)
                     IF (bad2(ibond) == 0) THEN
                        counter = counter+1
                        new_bond_list(counter) = bond_list(ibond)
                        newkind = bond_list(ibond)%bond_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_bond_list(counter)%bond_kind => new_bond_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         nbond=SIZE(new_bond_list), &
                                         bond_kind_set=new_bond_kind_set, &
                                         bond_list=new_bond_list)
                  DO ibond = 1, SIZE(new_bond_kind_set)
                     new_bond_kind_set(ibond)%kind_number = ibond
                  END DO
               END IF
               DEALLOCATE (bad2)
               CALL deallocate_bond_kind_set(bond_kind_set)
               DEALLOCATE (bond_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New BOND Count: ", &
                  SIZE(new_bond_list), SIZE(new_bond_kind_set)
               IF (iw > 0) WRITE (iw, '(2I6)') (new_bond_list(ibond)%a, new_bond_list(ibond)%b, &
                                                ibond=1, SIZE(new_bond_list))
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 8. Count the number of UNSET bend kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                nbend=nbend, &
                                bend_kind_set=bend_kind_set, &
                                bend_list=bend_list)
         IF (nbend > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old BEND Count: ", &
               SIZE(bend_list), SIZE(bend_kind_set)
            IF (iw > 0) WRITE (iw, '(3I6)') (bend_list(ibend)%a, bend_list(ibend)%b, &
                                             bend_list(ibend)%c, ibend=1, SIZE(bend_list))
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(bend_kind_set)))
            bad1(:) = 0
            DO ibend = 1, SIZE(bend_kind_set)
               unsetme = .FALSE.
               IF (bend_kind_set(ibend)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(bend_list)
                  IF (bend_list(i)%id_type /= do_ff_undef .AND. &
                      bend_list(i)%bend_kind%kind_number == ibend) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(ibend) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(bend_kind_set)-SUM(bad1)
               CALL allocate_bend_kind_set(new_bend_kind_set, counter)
               counter = 0
               DO ibend = 1, SIZE(bend_kind_set)
                  IF (bad1(ibend) == 0) THEN
                     counter = counter+1
                     new_bend_kind_set(counter) = bend_kind_set(ibend)
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(bend_list)))
               bad2(:) = 0
               DO ibend = 1, SIZE(bend_list)
                  unsetme = .FALSE.
                  IF (bend_list(ibend)%bend_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (bend_list(ibend)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(ibend) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(bend_list)-SUM(bad2)
                  ALLOCATE (new_bend_list(counter))
                  counter = 0
                  DO ibend = 1, SIZE(bend_list)
                     IF (bad2(ibend) == 0) THEN
                        counter = counter+1
                        new_bend_list(counter) = bend_list(ibend)
                        newkind = bend_list(ibend)%bend_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_bend_list(counter)%bend_kind => new_bend_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         nbend=SIZE(new_bend_list), &
                                         bend_kind_set=new_bend_kind_set, &
                                         bend_list=new_bend_list)
                  DO ibend = 1, SIZE(new_bend_kind_set)
                     new_bend_kind_set(ibend)%kind_number = ibend
                  END DO
               END IF
               DEALLOCATE (bad2)
               CALL deallocate_bend_kind_set(bend_kind_set)
               DEALLOCATE (bend_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New BEND Count: ", &
                  SIZE(new_bend_list), SIZE(new_bend_kind_set)
               IF (iw > 0) WRITE (iw, '(3I6)') (new_bend_list(ibend)%a, new_bend_list(ibend)%b, &
                                                new_bend_list(ibend)%c, ibend=1, SIZE(new_bend_list))
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 9. Count the number of UNSET Urey-Bradley kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                nub=nub, &
                                ub_kind_set=ub_kind_set, &
                                ub_list=ub_list)
         IF (nub > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old UB Count: ", &
               SIZE(ub_list), SIZE(ub_kind_set)
            IF (iw > 0) WRITE (iw, '(3I6)') (ub_list(iub)%a, ub_list(iub)%b, &
                                             ub_list(iub)%c, iub=1, SIZE(ub_list))
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(ub_kind_set)))
            bad1(:) = 0
            DO iub = 1, SIZE(ub_kind_set)
               unsetme = .FALSE.
               IF (ub_kind_set(iub)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(ub_list)
                  IF (ub_list(i)%id_type /= do_ff_undef .AND. &
                      ub_list(i)%ub_kind%kind_number == iub) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(iub) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(ub_kind_set)-SUM(bad1)
               CALL allocate_ub_kind_set(new_ub_kind_set, counter)
               counter = 0
               DO iub = 1, SIZE(ub_kind_set)
                  IF (bad1(iub) == 0) THEN
                     counter = counter+1
                     new_ub_kind_set(counter) = ub_kind_set(iub)
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(ub_list)))
               bad2(:) = 0
               DO iub = 1, SIZE(ub_list)
                  unsetme = .FALSE.
                  IF (ub_list(iub)%ub_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (ub_list(iub)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(iub) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(ub_list)-SUM(bad2)
                  ALLOCATE (new_ub_list(counter))
                  counter = 0
                  DO iub = 1, SIZE(ub_list)
                     IF (bad2(iub) == 0) THEN
                        counter = counter+1
                        new_ub_list(counter) = ub_list(iub)
                        newkind = ub_list(iub)%ub_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_ub_list(counter)%ub_kind => new_ub_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         nub=SIZE(new_ub_list), &
                                         ub_kind_set=new_ub_kind_set, &
                                         ub_list=new_ub_list)
                  DO iub = 1, SIZE(new_ub_kind_set)
                     new_ub_kind_set(iub)%kind_number = iub
                  END DO
               END IF
               DEALLOCATE (bad2)
               CALL ub_kind_dealloc_ref(ub_kind_set)
               DEALLOCATE (ub_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New UB Count: ", &
                  SIZE(new_ub_list), SIZE(new_ub_kind_set)
               IF (iw > 0) WRITE (iw, '(3I6)') (new_ub_list(iub)%a, new_ub_list(iub)%b, &
                                                new_ub_list(iub)%c, iub=1, SIZE(new_ub_list))
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 10. Count the number of UNSET torsion kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                ntorsion=ntorsion, &
                                torsion_kind_set=torsion_kind_set, &
                                torsion_list=torsion_list)
         IF (ntorsion > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old TORSION Count: ", &
               SIZE(torsion_list), SIZE(torsion_kind_set)
            IF (iw > 0) WRITE (iw, '(4I6)') (torsion_list(itorsion)%a, torsion_list(itorsion)%b, &
                                             torsion_list(itorsion)%c, torsion_list(itorsion)%d, itorsion=1, SIZE(torsion_list))
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(torsion_kind_set)))
            bad1(:) = 0
            DO itorsion = 1, SIZE(torsion_kind_set)
               unsetme = .FALSE.
               IF (torsion_kind_set(itorsion)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(torsion_list)
                  IF (torsion_list(i)%id_type /= do_ff_undef .AND. &
                      torsion_list(i)%torsion_kind%kind_number == itorsion) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(itorsion) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(torsion_kind_set)-SUM(bad1)
               CALL allocate_torsion_kind_set(new_torsion_kind_set, counter)
               counter = 0
               DO itorsion = 1, SIZE(torsion_kind_set)
                  IF (bad1(itorsion) == 0) THEN
                     counter = counter+1
                     new_torsion_kind_set(counter) = torsion_kind_set(itorsion)
                     i = SIZE(torsion_kind_set(itorsion)%m)
                     j = SIZE(torsion_kind_set(itorsion)%k)
                     k = SIZE(torsion_kind_set(itorsion)%phi0)
                     ALLOCATE (new_torsion_kind_set(counter)%m(i))
                     ALLOCATE (new_torsion_kind_set(counter)%k(i))
                     ALLOCATE (new_torsion_kind_set(counter)%phi0(i))
                     new_torsion_kind_set(counter)%m = torsion_kind_set(itorsion)%m
                     new_torsion_kind_set(counter)%k = torsion_kind_set(itorsion)%k
                     new_torsion_kind_set(counter)%phi0 = torsion_kind_set(itorsion)%phi0
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(torsion_list)))
               bad2(:) = 0
               DO itorsion = 1, SIZE(torsion_list)
                  unsetme = .FALSE.
                  IF (torsion_list(itorsion)%torsion_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (torsion_list(itorsion)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(itorsion) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(torsion_list)-SUM(bad2)
                  ALLOCATE (new_torsion_list(counter))
                  counter = 0
                  DO itorsion = 1, SIZE(torsion_list)
                     IF (bad2(itorsion) == 0) THEN
                        counter = counter+1
                        new_torsion_list(counter) = torsion_list(itorsion)
                        newkind = torsion_list(itorsion)%torsion_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_torsion_list(counter)%torsion_kind => new_torsion_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         ntorsion=SIZE(new_torsion_list), &
                                         torsion_kind_set=new_torsion_kind_set, &
                                         torsion_list=new_torsion_list)
                  DO itorsion = 1, SIZE(new_torsion_kind_set)
                     new_torsion_kind_set(itorsion)%kind_number = itorsion
                  END DO
               END IF
               DEALLOCATE (bad2)
               DO itorsion = 1, SIZE(torsion_kind_set)
                  CALL torsion_kind_dealloc_ref(torsion_kind_set(itorsion))
               END DO
               DEALLOCATE (torsion_kind_set)
               DEALLOCATE (torsion_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New TORSION Count: ", &
                  SIZE(new_torsion_list), SIZE(new_torsion_kind_set)
               IF (iw > 0) WRITE (iw, '(4I6)') (new_torsion_list(itorsion)%a, new_torsion_list(itorsion)%b, &
                                                new_torsion_list(itorsion)%c, new_torsion_list(itorsion)%d, itorsion=1, &
                                                SIZE(new_torsion_list))
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 11. Count the number of UNSET improper kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                nimpr=nimpr, &
                                impr_kind_set=impr_kind_set, &
                                impr_list=impr_list)
         IF (nimpr > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old IMPROPER Count: ", &
               SIZE(impr_list), SIZE(impr_kind_set)
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(impr_kind_set)))
            bad1(:) = 0
            DO iimpr = 1, SIZE(impr_kind_set)
               unsetme = .FALSE.
               IF (impr_kind_set(iimpr)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(impr_list)
                  IF (impr_list(i)%id_type /= do_ff_undef .AND. &
                      impr_list(i)%impr_kind%kind_number == iimpr) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(iimpr) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(impr_kind_set)-SUM(bad1)
               CALL allocate_impr_kind_set(new_impr_kind_set, counter)
               counter = 0
               DO iimpr = 1, SIZE(impr_kind_set)
                  IF (bad1(iimpr) == 0) THEN
                     counter = counter+1
                     new_impr_kind_set(counter) = impr_kind_set(iimpr)
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(impr_list)))
               bad2(:) = 0
               DO iimpr = 1, SIZE(impr_list)
                  unsetme = .FALSE.
                  IF (impr_list(iimpr)%impr_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (impr_list(iimpr)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(iimpr) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(impr_list)-SUM(bad2)
                  ALLOCATE (new_impr_list(counter))
                  counter = 0
                  DO iimpr = 1, SIZE(impr_list)
                     IF (bad2(iimpr) == 0) THEN
                        counter = counter+1
                        new_impr_list(counter) = impr_list(iimpr)
                        newkind = impr_list(iimpr)%impr_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_impr_list(counter)%impr_kind => new_impr_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         nimpr=SIZE(new_impr_list), &
                                         impr_kind_set=new_impr_kind_set, &
                                         impr_list=new_impr_list)
                  DO iimpr = 1, SIZE(new_impr_kind_set)
                     new_impr_kind_set(iimpr)%kind_number = iimpr
                  END DO
               END IF
               DEALLOCATE (bad2)
               DO iimpr = 1, SIZE(impr_kind_set)
                  CALL impr_kind_dealloc_ref() !This Subroutine doesn't deallocate anything, maybe needs to be implemented
               END DO
               DEALLOCATE (impr_kind_set)
               DEALLOCATE (impr_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New IMPROPER Count: ", &
                  SIZE(new_impr_list), SIZE(new_impr_kind_set)
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 11. Count the number of UNSET opbends kinds there are
      !-----------------------------------------------------------------------------
      DO ikind = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                nopbend=nopbend, &
                                opbend_kind_set=opbend_kind_set, &
                                opbend_list=opbend_list)
         IF (nopbend > 0) THEN
            IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") Old OPBEND Count: ", &
               SIZE(opbend_list), SIZE(opbend_kind_set)
            NULLIFY (bad1, bad2)
            ALLOCATE (bad1(SIZE(opbend_kind_set)))
            bad1(:) = 0
            DO iopbend = 1, SIZE(opbend_kind_set)
               unsetme = .FALSE.
               IF (opbend_kind_set(iopbend)%id_type == do_ff_undef) unsetme = .TRUE.
               valid_kind = .FALSE.
               DO i = 1, SIZE(opbend_list)
                  IF (opbend_list(i)%id_type /= do_ff_undef .AND. &
                      opbend_list(i)%opbend_kind%kind_number == iopbend) THEN
                     valid_kind = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. valid_kind) unsetme = .TRUE.
               IF (unsetme) bad1(iopbend) = 1
            END DO
            IF (SUM(bad1) /= 0) THEN
               counter = SIZE(opbend_kind_set)-SUM(bad1)
               CALL allocate_opbend_kind_set(new_opbend_kind_set, counter)
               counter = 0
               DO iopbend = 1, SIZE(opbend_kind_set)
                  IF (bad1(iopbend) == 0) THEN
                     counter = counter+1
                     new_opbend_kind_set(counter) = opbend_kind_set(iopbend)
                  END IF
               END DO
               counter = 0
               ALLOCATE (bad2(SIZE(opbend_list)))
               bad2(:) = 0
               DO iopbend = 1, SIZE(opbend_list)
                  unsetme = .FALSE.
                  IF (opbend_list(iopbend)%opbend_kind%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (opbend_list(iopbend)%id_type == do_ff_undef) unsetme = .TRUE.
                  IF (unsetme) bad2(iopbend) = 1
               END DO
               IF (SUM(bad2) /= 0) THEN
                  counter = SIZE(opbend_list)-SUM(bad2)
                  ALLOCATE (new_opbend_list(counter))
                  counter = 0
                  DO iopbend = 1, SIZE(opbend_list)
                     IF (bad2(iopbend) == 0) THEN
                        counter = counter+1
                        new_opbend_list(counter) = opbend_list(iopbend)
                        newkind = opbend_list(iopbend)%opbend_kind%kind_number
                        newkind = newkind-SUM(bad1(1:newkind))
                        new_opbend_list(counter)%opbend_kind => new_opbend_kind_set(newkind)
                     END IF
                  END DO
                  CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                         nopbend=SIZE(new_opbend_list), &
                                         opbend_kind_set=new_opbend_kind_set, &
                                         opbend_list=new_opbend_list)
                  DO iopbend = 1, SIZE(new_opbend_kind_set)
                     new_opbend_kind_set(iopbend)%kind_number = iopbend
                  END DO
               END IF
               DEALLOCATE (bad2)
               DEALLOCATE (opbend_kind_set)
               DEALLOCATE (opbend_list)
               IF (iw > 0) WRITE (iw, *) "    Mol(", ikind, ") New OPBEND Count: ", &
                  SIZE(new_opbend_list), SIZE(new_opbend_kind_set)
            END IF
            DEALLOCATE (bad1)
         END IF
      END DO

      !---------------------------------------------------------------------------
      !---------------------------------------------------------------------------
      ! 12. Count the number of UNSET NONBOND14 kinds there are
      !-                NEED TO REMOVE EXTRAS HERE   - IKUO
      !---------------------------------------------------------------------------
      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%FF_INFO")
      CALL timestop(handle)
   END SUBROUTINE clean_intra_force_kind

! **************************************************************************************************
!> \brief Reads from the input structure all information for generic functions
!> \param gen_section ...
!> \param func_name ...
!> \param xfunction ...
!> \param parameters ...
!> \param values ...
!> \param var_values ...
!> \param size_variables ...
!> \param i_rep_sec ...
!> \param input_variables ...
! **************************************************************************************************
   SUBROUTINE get_generic_info(gen_section, func_name, xfunction, parameters, values, &
                               var_values, size_variables, i_rep_sec, input_variables)
      TYPE(section_vals_type), POINTER                   :: gen_section
      CHARACTER(LEN=*), INTENT(IN)                       :: func_name
      CHARACTER(LEN=default_path_length), INTENT(OUT)    :: xfunction
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: parameters
      REAL(KIND=dp), DIMENSION(:), POINTER               :: values
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: var_values
      INTEGER, INTENT(IN), OPTIONAL                      :: size_variables, i_rep_sec
      CHARACTER(LEN=*), DIMENSION(:), OPTIONAL           :: input_variables

      CHARACTER(len=*), PARAMETER :: routineN = 'get_generic_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: my_par, my_par_tmp, my_units, &
                                                            my_units_tmp, my_var
      INTEGER                                            :: i, ind, irep, isize, j, mydim, n_par, &
                                                            n_units, n_val, nblank
      LOGICAL                                            :: check
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_val, my_val_tmp

      NULLIFY (my_var, my_par, my_val, my_par_tmp, my_val_tmp)
      NULLIFY (my_units)
      NULLIFY (my_units_tmp)
      IF (ASSOCIATED(parameters)) THEN
         DEALLOCATE (parameters)
      END IF
      IF (ASSOCIATED(values)) THEN
         DEALLOCATE (values)
      END IF
      irep = 1
      IF (PRESENT(i_rep_sec)) irep = i_rep_sec
      mydim = 0
      CALL section_vals_val_get(gen_section, TRIM(func_name), i_rep_section=irep, c_val=xfunction)
      CALL compress(xfunction, full=.TRUE.)
      IF (PRESENT(input_variables)) THEN
         ALLOCATE (my_var(SIZE(input_variables)))
         my_var = input_variables
      ELSE
         CALL section_vals_val_get(gen_section, "VARIABLES", i_rep_section=irep, c_vals=my_var)
      END IF
      IF (ASSOCIATED(my_var)) THEN
         mydim = SIZE(my_var)
      END IF
      IF (PRESENT(size_variables)) THEN
         IF(.NOT.(mydim == size_variables))CALL cp__a("force_fields_util.F",1284)
      END IF
      ! Check for presence of Parameters
      CALL section_vals_val_get(gen_section, "PARAMETERS", i_rep_section=irep, n_rep_val=n_par)
      CALL section_vals_val_get(gen_section, "VALUES", i_rep_section=irep, n_rep_val=n_val)
      check = (n_par > 0) .EQV. (n_val > 0)
      IF(.NOT.(check))CALL cp__a("force_fields_util.F",1290)
      CALL section_vals_val_get(gen_section, "UNITS", i_rep_section=irep, n_rep_val=n_units)
      IF (n_par > 0) THEN
         ! Parameters
         ALLOCATE (my_par(0))
         ALLOCATE (my_val(0))
         DO i = 1, n_par
            isize = SIZE(my_par)
            CALL section_vals_val_get(gen_section, "PARAMETERS", i_rep_section=irep, i_rep_val=i, c_vals=my_par_tmp)
            nblank = COUNT(my_par_tmp == "")
            CALL reallocate(my_par, 1, isize+SIZE(my_par_tmp)-nblank)
            ind = 0
            DO j = 1, SIZE(my_par_tmp)
               IF (my_par_tmp(j) == "") CYCLE
               ind = ind+1
               my_par(isize+ind) = my_par_tmp(j)
            END DO
         END DO
         DO i = 1, n_val
            isize = SIZE(my_val)
            CALL section_vals_val_get(gen_section, "VALUES", i_rep_section=irep, i_rep_val=i, r_vals=my_val_tmp)
            CALL reallocate(my_val, 1, isize+SIZE(my_val_tmp))
            my_val(isize+1:isize+SIZE(my_val_tmp)) = my_val_tmp
         END DO
         IF(.NOT.(SIZE(my_par) == SIZE(my_val)))CALL cp__a("force_fields_util.F",1314)
         ! Optionally read the units for each parameter value
         ALLOCATE (my_units(0))
         IF (n_units > 0) THEN
            DO i = 1, n_units
               isize = SIZE(my_units)
               CALL section_vals_val_get(gen_section, "UNITS", i_rep_section=irep, i_rep_val=i, c_vals=my_units_tmp)
               nblank = COUNT(my_units_tmp == "")
               CALL reallocate(my_units, 1, isize+SIZE(my_units_tmp)-nblank)
               ind = 0
               DO j = 1, SIZE(my_units_tmp)
                  IF (my_units_tmp(j) == "") CYCLE
                  ind = ind+1
                  my_units(isize+ind) = my_units_tmp(j)
               END DO
            END DO
            IF(.NOT.(SIZE(my_units) == SIZE(my_val)))CALL cp__a("force_fields_util.F",1330)
         END IF
         mydim = mydim+SIZE(my_val)
         IF (SIZE(my_val) == 0) THEN
            DEALLOCATE (my_par)
            DEALLOCATE (my_val)
            DEALLOCATE (my_units)
         END IF
      END IF
      ! Handle trivial case of a null function defined
      ALLOCATE (parameters(mydim))
      ALLOCATE (values(mydim))
      IF (mydim > 0) THEN
         parameters(1:SIZE(my_var)) = my_var
         values(1:SIZE(my_var)) = 0.0_dp
         IF (PRESENT(var_values)) THEN
            IF(.NOT.(SIZE(var_values) == SIZE(my_var)))CALL cp__a("force_fields_util.F",1346)
            values(1:SIZE(my_var)) = var_values
         END IF
         IF (ASSOCIATED(my_val)) THEN
            DO i = 1, SIZE(my_val)
               parameters(SIZE(my_var)+i) = my_par(i)
               IF (n_units > 0) THEN
                  values(SIZE(my_var)+i) = cp_unit_to_cp2k(my_val(i), TRIM(ADJUSTL(my_units(i))))
               ELSE
                  values(SIZE(my_var)+i) = my_val(i)
               END IF
            END DO
         END IF
      END IF
      IF (ASSOCIATED(my_par)) THEN
         DEALLOCATE (my_par)
      END IF
      IF (ASSOCIATED(my_val)) THEN
         DEALLOCATE (my_val)
      END IF
      IF (ASSOCIATED(my_units)) THEN
         DEALLOCATE (my_units)
      END IF
      IF (PRESENT(input_variables)) THEN
         DEALLOCATE (my_var)
      END IF
   END SUBROUTINE get_generic_info

END MODULE force_fields_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE constraint_fxd

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: get_atomic_kind_set
   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE colvar_types,                    ONLY: colvar_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              get_molecule_kind,&
                                              local_fixd_constraint_type,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: local_g3x3_constraint_type,&
                                              local_g4x6_constraint_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: fix_atom_control, &
             check_fixed_atom_cns_g3x3, &
             check_fixed_atom_cns_g4x6, &
             check_fixed_atom_cns_colv, &
             create_local_fixd_list, &
             release_local_fixd_list

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_fxd'

CONTAINS

! **************************************************************************************************
!> \brief allows for fix atom constraints
!> \param force_env ...
!> \param w ...
!> \par History
!>      - optionally apply fix atom constraint to random forces (Langevin)
!>        (04.10.206,MK)
! **************************************************************************************************
   SUBROUTINE fix_atom_control(force_env, w)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: w

      CHARACTER(len=*), PARAMETER :: routineN = 'fix_atom_control', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ifixd, ii, ikind, iparticle, iparticle_local, my_atm_fixed, natom, &
         ncore, nfixed_atoms, nkind, nparticle, nparticle_local, nshell, shell_index
      LOGICAL                                            :: shell_present
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: force
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_fixd_constraint_type), POINTER          :: lfixd_list(:)
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kinds)
      NULLIFY (core_particles)
      NULLIFY (particles)
      NULLIFY (shell_particles)
      shell_present = .FALSE.

      NULLIFY (lfixd_list)
      CALL force_env_get(force_env=force_env, &
                         subsys=subsys)
      CALL cp_subsys_get(subsys=subsys, &
                         atomic_kinds=atomic_kinds, &
                         core_particles=core_particles, &
                         local_particles=local_particles, &
                         molecule_kinds_new=molecule_kinds, &
                         natom=natom, &
                         ncore=ncore, &
                         nshell=nshell, &
                         particles=particles, &
                         shell_particles=shell_particles)
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kinds%els, &
                               shell_present=shell_present)

      particle_set => particles%els
      IF(.NOT.((SIZE(particle_set) == natom)))CALL cp__a("constraint_fxd.F",112)
      IF (shell_present) THEN
         core_particle_set => core_particles%els
         IF(.NOT.((SIZE(core_particle_set) == ncore)))CALL cp__a("constraint_fxd.F",115)
         shell_particle_set => shell_particles%els
         IF(.NOT.((SIZE(shell_particle_set) == nshell)))CALL cp__a("constraint_fxd.F",117)
      END IF
      nparticle = natom+nshell
      molecule_kind_set => molecule_kinds%els

      nkind = molecule_kinds%n_els
      my_atm_fixed = 0
      DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, nfixd=nfixed_atoms)
         my_atm_fixed = my_atm_fixed+nfixed_atoms
      END DO

      IF (my_atm_fixed /= 0) THEN
         IF (.NOT. PRESENT(w)) THEN
            ! Allocate scratch array
            ALLOCATE (force(3, nparticle))
            force(:, :) = 0.0_dp
            DO i = 1, SIZE(local_particles%n_el)
               nparticle_local = local_particles%n_el(i)
               DO iparticle_local = 1, nparticle_local
                  iparticle = local_particles%list(i)%array(iparticle_local)
                  shell_index = particle_set(iparticle)%shell_index
                  IF (shell_index == 0) THEN
                     force(:, iparticle) = particle_set(iparticle)%f(:)
                  ELSE
                     force(:, iparticle) = core_particle_set(shell_index)%f(:)
                     force(:, natom+shell_index) = shell_particle_set(shell_index)%f(:)
                  END IF
               END DO
            END DO
         END IF

         ! Create the list of locally fixed atoms
         CALL create_local_fixd_list(lfixd_list, nkind, molecule_kind_set, local_particles)

         ! Apply fixed atom constraint
         DO ifixd = 1, SIZE(lfixd_list)
            ikind = lfixd_list(ifixd)%ikind
            ii = lfixd_list(ifixd)%ifixd_index
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, fixd_list=fixd_list)
            IF (.NOT. fixd_list(ii)%restraint%active) THEN
               iparticle = fixd_list(ii)%fixd
               shell_index = particle_set(iparticle)%shell_index
               ! Select constraint type
               IF (PRESENT(w)) THEN
                  SELECT CASE (fixd_list (ii)%itype)
                  CASE (use_perd_x)
                     w(1, iparticle) = 0.0_dp
                  CASE (use_perd_y)
                     w(2, iparticle) = 0.0_dp
                  CASE (use_perd_z)
                     w(3, iparticle) = 0.0_dp
                  CASE (use_perd_xy)
                     w(1, iparticle) = 0.0_dp
                     w(2, iparticle) = 0.0_dp
                  CASE (use_perd_xz)
                     w(1, iparticle) = 0.0_dp
                     w(3, iparticle) = 0.0_dp
                  CASE (use_perd_yz)
                     w(2, iparticle) = 0.0_dp
                     w(3, iparticle) = 0.0_dp
                  CASE (use_perd_xyz)
                     w(:, iparticle) = 0.0_dp
                  END SELECT
               ELSE
                  SELECT CASE (fixd_list (ii)%itype)
                  CASE (use_perd_x)
                     force(1, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(1, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_y)
                     force(2, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(2, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_z)
                     force(3, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(3, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_xy)
                     force(1, iparticle) = 0.0_dp
                     force(2, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(1, natom+shell_index) = 0.0_dp
                        force(2, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_xz)
                     force(1, iparticle) = 0.0_dp
                     force(3, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(1, natom+shell_index) = 0.0_dp
                        force(3, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_yz)
                     force(2, iparticle) = 0.0_dp
                     force(3, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(2, natom+shell_index) = 0.0_dp
                        force(3, natom+shell_index) = 0.0_dp
                     END IF
                  CASE (use_perd_xyz)
                     force(:, iparticle) = 0.0_dp
                     IF (shell_index /= 0) THEN
                        force(:, natom+shell_index) = 0.0_dp
                     END IF
                  END SELECT
               END IF
            END IF
         END DO
         CALL release_local_fixd_list(lfixd_list)

         IF (.NOT. PRESENT(w)) THEN
            CALL mp_sum(force, force_env%para_env%group)
            DO iparticle = 1, natom
               shell_index = particle_set(iparticle)%shell_index
               IF (shell_index == 0) THEN
                  particle_set(iparticle)%f(:) = force(:, iparticle)
               ELSE
                  core_particle_set(shell_index)%f(:) = force(:, iparticle)
                  shell_particle_set(shell_index)%f(:) = force(:, natom+shell_index)
               END IF
            END DO
            DEALLOCATE (force)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE fix_atom_control

! **************************************************************************************************
!> \brief ...
!> \param imass1 ...
!> \param imass2 ...
!> \param imass3 ...
!> \param index_a ...
!> \param index_b ...
!> \param index_c ...
!> \param fixd_list ...
!> \param lg3x3 ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE check_fixed_atom_cns_g3x3(imass1, imass2, imass3, &
                                        index_a, index_b, index_c, fixd_list, lg3x3)
      REAL(KIND=dp), INTENT(INOUT)                       :: imass1, imass2, imass3
      INTEGER, INTENT(IN)                                :: index_a, index_b, index_c
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_g3x3_constraint_type)                   :: lg3x3

      INTEGER                                            :: i

      IF (lg3x3%init) THEN
         imass1 = lg3x3%imass1
         imass2 = lg3x3%imass2
         imass3 = lg3x3%imass3
      ELSE
         IF (ASSOCIATED(fixd_list)) THEN
            IF (SIZE(fixd_list) > 0) THEN
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_a) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass1 = 0.0_dp
                     EXIT
                  END IF
               END DO
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_b) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass2 = 0.0_dp
                     EXIT
                  END IF
               END DO
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_c) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass3 = 0.0_dp
                     EXIT
                  END IF
               END DO
            END IF
         END IF
         lg3x3%imass1 = imass1
         lg3x3%imass2 = imass2
         lg3x3%imass3 = imass3
         lg3x3%init = .TRUE.
      END IF
   END SUBROUTINE check_fixed_atom_cns_g3x3

! **************************************************************************************************
!> \brief ...
!> \param imass1 ...
!> \param imass2 ...
!> \param imass3 ...
!> \param imass4 ...
!> \param index_a ...
!> \param index_b ...
!> \param index_c ...
!> \param index_d ...
!> \param fixd_list ...
!> \param lg4x6 ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE check_fixed_atom_cns_g4x6(imass1, imass2, imass3, imass4, &
                                        index_a, index_b, index_c, index_d, fixd_list, lg4x6)
      REAL(KIND=dp), INTENT(INOUT)                       :: imass1, imass2, imass3, imass4
      INTEGER, INTENT(IN)                                :: index_a, index_b, index_c, index_d
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_g4x6_constraint_type)                   :: lg4x6

      INTEGER                                            :: i

      IF (lg4x6%init) THEN
         imass1 = lg4x6%imass1
         imass2 = lg4x6%imass2
         imass3 = lg4x6%imass3
         imass4 = lg4x6%imass4
      ELSE
         IF (ASSOCIATED(fixd_list)) THEN
            IF (SIZE(fixd_list) > 0) THEN
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_a) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass1 = 0.0_dp
                     EXIT
                  END IF
               END DO
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_b) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass2 = 0.0_dp
                     EXIT
                  END IF
               END DO
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_c) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass3 = 0.0_dp
                     EXIT
                  END IF
               END DO
               DO i = 1, SIZE(fixd_list)
                  IF (fixd_list(i)%fixd == index_d) THEN
                     IF (fixd_list(i)%itype /= use_perd_xyz) CYCLE
                     IF (.NOT. fixd_list(i)%restraint%active) imass4 = 0.0_dp
                     EXIT
                  END IF
               END DO
            END IF
         END IF
         lg4x6%imass1 = imass1
         lg4x6%imass2 = imass2
         lg4x6%imass3 = imass3
         lg4x6%imass4 = imass4
         lg4x6%init = .TRUE.
      END IF
   END SUBROUTINE check_fixed_atom_cns_g4x6

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param colvar ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE check_fixed_atom_cns_colv(fixd_list, colvar)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(colvar_type), POINTER                         :: colvar

      INTEGER                                            :: i, j, k

      IF (ASSOCIATED(fixd_list)) THEN
         IF (ASSOCIATED(fixd_list)) THEN
            IF (SIZE(fixd_list) > 0) THEN
               DO i = 1, SIZE(colvar%i_atom)
                  j = colvar%i_atom(i)
                  DO k = 1, SIZE(fixd_list)
                     IF (fixd_list(k)%fixd == j) THEN
                        IF (fixd_list(k)%itype /= use_perd_xyz) CYCLE
                        IF (.NOT. fixd_list(k)%restraint%active) &
                           colvar%dsdr(:, i) = 0.0_dp
                        EXIT
                     END IF
                  END DO
               END DO
            END IF
         END IF
      END IF

   END SUBROUTINE check_fixed_atom_cns_colv

! **************************************************************************************************
!> \brief setup a list of local atoms on which to apply constraints/restraints
!> \param lfixd_list ...
!> \param nkind ...
!> \param molecule_kind_set ...
!> \param local_particles ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
   SUBROUTINE create_local_fixd_list(lfixd_list, nkind, molecule_kind_set, &
                                     local_particles)
      TYPE(local_fixd_constraint_type), POINTER          :: lfixd_list(:)
      INTEGER, INTENT(IN)                                :: nkind
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_local_fixd_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, iparticle, &
                                                            iparticle_local, isize, jsize, ncnst, &
                                                            nparticle_local, nparticle_local_all, &
                                                            nsize
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: fixed_atom_all, kind_index_all, &
                                                            local_particle_all, work0, work1, work2
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      CALL timeset(routineN, handle)
      IF(.NOT.(.NOT. ASSOCIATED(lfixd_list)))CALL cp__a("constraint_fxd.F",441)
      nsize = 0
      DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, fixd_list=fixd_list)
         IF (ASSOCIATED(fixd_list)) THEN
            nsize = nsize+SIZE(fixd_list)
         END IF
      END DO
      IF (nsize /= 0) THEN
         ALLOCATE (fixed_atom_all(nsize))
         ALLOCATE (work0(nsize))
         ALLOCATE (work1(nsize))
         ALLOCATE (kind_index_all(nsize))
         nsize = 0
         DO ikind = 1, nkind
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, fixd_list=fixd_list)
            IF (ASSOCIATED(fixd_list)) THEN
               DO i = 1, SIZE(fixd_list)
                  nsize = nsize+1
                  work0(nsize) = i
                  kind_index_all(nsize) = ikind
                  fixed_atom_all(nsize) = fixd_list(i)%fixd
               END DO
            END IF
         END DO
         ! Sort the number of all atoms to be constrained/restrained
         CALL sort(fixed_atom_all, nsize, work1)

         ! Sort the local particles
         nparticle_local_all = 0
         DO i = 1, SIZE(local_particles%n_el)
            nparticle_local_all = nparticle_local_all+local_particles%n_el(i)
         END DO
         ALLOCATE (local_particle_all(nparticle_local_all))
         ALLOCATE (work2(nparticle_local_all))
         nparticle_local_all = 0
         DO i = 1, SIZE(local_particles%n_el)
            nparticle_local = local_particles%n_el(i)
            DO iparticle_local = 1, nparticle_local
               nparticle_local_all = nparticle_local_all+1
               iparticle = local_particles%list(i)%array(iparticle_local)
               local_particle_all(nparticle_local_all) = iparticle
            END DO
         END DO
         CALL sort(local_particle_all, nparticle_local_all, work2)

         ! Count the amount of local constraints/restraints
         ncnst = 0
         jsize = 1
         Loop_count: DO isize = 1, nparticle_local_all
            DO WHILE (local_particle_all(isize) > fixed_atom_all(jsize))
               jsize = jsize+1
               IF (jsize > nsize) THEN
                  jsize = nsize
                  EXIT Loop_count
               END IF
            END DO
            IF (local_particle_all(isize) == fixed_atom_all(jsize)) ncnst = ncnst+1
         END DO Loop_count

         ! Allocate local fixed atom array
         ALLOCATE (lfixd_list(ncnst))

         ! Fill array with constraints infos
         ncnst = 0
         jsize = 1
         Loop_fill: DO isize = 1, nparticle_local_all
            DO WHILE (local_particle_all(isize) > fixed_atom_all(jsize))
               jsize = jsize+1
               IF (jsize > nsize) THEN
                  jsize = nsize
                  EXIT Loop_fill
               END IF
            END DO
            IF (local_particle_all(isize) == fixed_atom_all(jsize)) THEN
               ncnst = ncnst+1
               lfixd_list(ncnst)%ifixd_index = work0(work1(jsize))
               lfixd_list(ncnst)%ikind = kind_index_all(work1(jsize))
            END IF
         END DO Loop_fill

         ! Deallocate working arrays
         DEALLOCATE (local_particle_all)
         DEALLOCATE (work2)
         DEALLOCATE (fixed_atom_all)
         DEALLOCATE (work1)
         DEALLOCATE (kind_index_all)
      ELSE
         ! Allocate local fixed atom array with dimension 0
         ALLOCATE (lfixd_list(0))
      END IF
      CALL timestop(handle)
   END SUBROUTINE create_local_fixd_list

! **************************************************************************************************
!> \brief destroy the list of local atoms on which to apply constraints/restraints
!>      Teodoro Laino [tlaino] - 11.2008
!> \param lfixd_list ...
! **************************************************************************************************
   SUBROUTINE release_local_fixd_list(lfixd_list)
      TYPE(local_fixd_constraint_type), POINTER          :: lfixd_list(:)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_local_fixd_list', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(lfixd_list)))CALL cp__a("constraint_fxd.F",548)
      DEALLOCATE (lfixd_list)
   END SUBROUTINE release_local_fixd_list

END MODULE constraint_fxd
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief defines collective variables s({R}) and the derivative of this variable wrt R
!>      these can then be used in constraints, restraints and metadynamics ...
!> \par History
!>      04.2004 created
!>      01.2006 Refactored [Joost VandeVondele]
!> \author Alessandro Laio,Fawzi Mohamed
! **************************************************************************************************
MODULE colvar_methods

   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE colvar_types,                    ONLY: &
        HBP_colvar_id, Wc_colvar_id, acid_hyd_dist_colvar_id, acid_hyd_shell_colvar_id, &
        angle_colvar_id, colvar_create, colvar_setup, colvar_type, combine_colvar_id, &
        coord_colvar_id, dfunct_colvar_id, dist_colvar_id, distance_from_path_colvar_id, &
        do_clv_fix_point, do_clv_geo_center, do_clv_x, do_clv_xy, do_clv_xz, do_clv_y, do_clv_yz, &
        do_clv_z, eval_point_der, eval_point_mass, eval_point_pos, gyration_colvar_id, &
        hydronium_colvar_id, mindist_colvar_id, plane_def_atoms, plane_def_vec, &
        plane_distance_colvar_id, plane_plane_angle_colvar_id, population_colvar_id, &
        qparm_colvar_id, reaction_path_colvar_id, ring_puckering_colvar_id, rmsd_colvar_id, &
        rotation_colvar_id, torsion_colvar_id, u_colvar_id, xyz_diag_colvar_id, &
        xyz_outerdiag_colvar_id
   USE constraint_fxd,                  ONLY: check_fixed_atom_cns_colv
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_p_type,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              use_mixed_force
   USE force_fields_util,               ONLY: get_generic_info
   USE fparser,                         ONLY: EvalErrType,&
                                              evalf,&
                                              evalfd,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE input_constants,                 ONLY: rmsd_all,&
                                              rmsd_list,&
                                              rmsd_weightlist
   USE input_cp2k_colvar,               ONLY: create_colvar_xyz_d_section,&
                                              create_colvar_xyz_od_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: fac,&
                                              maxfac,&
                                              pi,&
                                              twopi
   USE mathlib,                         ONLY: vector_product
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum,&
                                              mp_sync
   USE mixed_energy_types,              ONLY: mixed_force_type
   USE mixed_environment_utils,         ONLY: get_subsys_map_index
   USE molecule_kind_types,             ONLY: fixd_constraint_type
   USE particle_list_types,             ONLY: particle_list_p_type,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE rmsd,                            ONLY: rmsd3
   USE spherical_harmonics,             ONLY: dlegendre,&
                                              legendre
   USE string_utilities,                ONLY: compress,&
                                              uppercase
   USE wannier_states_types,            ONLY: wannier_centres_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'colvar_methods'
   REAL(KIND=dp), PRIVATE, PARAMETER    :: tolerance_acos = 1.0E-5_dp

   PUBLIC :: colvar_read, &
             colvar_eval_glob_f, &
             colvar_eval_mol_f

CONTAINS

! **************************************************************************************************
!> \brief reads a colvar from the input
!> \param colvar the place where to store what will be read
!> \param icol number of the current colvar (repetition in colvar_section)
!> \param colvar_section the colvar section
!> \param para_env ...
!> \par History
!>      04.2004 created [alessandro laio and fawzi mohamed]
!> \author teo
! **************************************************************************************************
   RECURSIVE SUBROUTINE colvar_read(colvar, icol, colvar_section, para_env)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: icol
      TYPE(section_vals_type), POINTER                   :: colvar_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_read', routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: fmid
      CHARACTER(LEN=7)                                   :: tag, tag_comp, tag_comp1, tag_comp2
      CHARACTER(LEN=default_path_length)                 :: path_function
      CHARACTER(LEN=default_string_length)               :: tmpStr, tmpStr2
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: c_kinds, my_par
      INTEGER                                            :: handle, i, iatm, icomponent, iend, &
                                                            ifunc, ii, isize, istart, iw, iw1, j, &
                                                            k, kk, n_var, n_var_k, ncol, ndim, &
                                                            nr_frame, v_count
      INTEGER, DIMENSION(:), POINTER                     :: iatms
      INTEGER, DIMENSION(:, :), POINTER                  :: p_bounds
      LOGICAL                                            :: check, use_mixed_energy
      LOGICAL, DIMENSION(25)                             :: my_subsection
      REAL(dp), DIMENSION(:), POINTER                    :: s1, wei, weights
      REAL(dp), DIMENSION(:, :), POINTER                 :: p_range, s1v
      REAL(KIND=dp), DIMENSION(1)                        :: my_val
      REAL(KIND=dp), DIMENSION(:), POINTER               :: g_range, grid_point, grid_sp, my_vals, &
                                                            range
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER :: acid_hyd_dist_section, acid_hyd_shell_section, &
         angle_section, colvar_subsection, combine_section, coordination_section, dfunct_section, &
         distance_from_path_section, distance_section, frame_section, gyration_section, &
         HBP_section, hydronium_section, mindist_section, path_section, plane_dist_section, &
         plane_plane_angle_section, plane_sections, point_section, population_section, &
         qparm_section, reaction_path_section, ring_puckering_section, rmsd_section, &
         rotation_section, torsion_section, u_section, Wc_section, wrk_section, xyz_diag_section
      TYPE(section_vals_type), POINTER :: xyz_outerdiag_section

      CALL timeset(routineN, handle)
      NULLIFY (logger, c_kinds, iatms)
      logger => cp_get_default_logger()
      my_subsection = .FALSE.
      distance_section => section_vals_get_subs_vals(colvar_section, "DISTANCE", i_rep_section=icol)
      dfunct_section => section_vals_get_subs_vals(colvar_section, "DISTANCE_FUNCTION", &
                                                   i_rep_section=icol)
      angle_section => section_vals_get_subs_vals(colvar_section, "ANGLE", i_rep_section=icol)
      torsion_section => section_vals_get_subs_vals(colvar_section, "TORSION", i_rep_section=icol)
      coordination_section => section_vals_get_subs_vals(colvar_section, "COORDINATION", i_rep_section=icol)
      plane_dist_section => section_vals_get_subs_vals(colvar_section, "DISTANCE_POINT_PLANE", i_rep_section=icol)
      plane_plane_angle_section &
         => section_vals_get_subs_vals(colvar_section, "ANGLE_PLANE_PLANE", i_rep_section=icol)
      rotation_section => section_vals_get_subs_vals(colvar_section, "BOND_ROTATION", i_rep_section=icol)
      qparm_section => section_vals_get_subs_vals(colvar_section, "QPARM", i_rep_section=icol)
      hydronium_section => section_vals_get_subs_vals(colvar_section, "HYDRONIUM", i_rep_section=icol)
      acid_hyd_dist_section => section_vals_get_subs_vals(colvar_section, "ACID_HYDRONIUM_DISTANCE", i_rep_section=icol)
      acid_hyd_shell_section &
         => section_vals_get_subs_vals(colvar_section, "ACID_HYDRONIUM_SHELL", i_rep_section=icol)
      reaction_path_section => section_vals_get_subs_vals(colvar_section, "REACTION_PATH", i_rep_section=icol, &
                                                          can_return_null=.TRUE.)
      distance_from_path_section &
         => section_vals_get_subs_vals(colvar_section, "DISTANCE_FROM_PATH", &
                                       i_rep_section=icol, can_return_null=.TRUE.)
      combine_section => section_vals_get_subs_vals(colvar_section, "COMBINE_COLVAR", i_rep_section=icol, &
                                                    can_return_null=.TRUE.)
      population_section => section_vals_get_subs_vals(colvar_section, "POPULATION", i_rep_section=icol)
      gyration_section => section_vals_get_subs_vals(colvar_section, "GYRATION_RADIUS", i_rep_section=icol)
      rmsd_section => section_vals_get_subs_vals(colvar_section, "RMSD", i_rep_section=icol)
      xyz_diag_section => section_vals_get_subs_vals(colvar_section, "XYZ_DIAG", i_rep_section=icol)
      xyz_outerdiag_section => section_vals_get_subs_vals(colvar_section, "XYZ_OUTERDIAG", i_rep_section=icol)
      u_section => section_vals_get_subs_vals(colvar_section, "U", i_rep_section=icol)
      Wc_section => section_vals_get_subs_vals(colvar_section, "WC", i_rep_section=icol)
      HBP_section => section_vals_get_subs_vals(colvar_section, "HBP", i_rep_section=icol)
      ring_puckering_section &
         => section_vals_get_subs_vals(colvar_section, "RING_PUCKERING", i_rep_section=icol)
      mindist_section => section_vals_get_subs_vals(colvar_section, "CONDITIONED_DISTANCE", i_rep_section=icol)

      CALL section_vals_get(distance_section, explicit=my_subsection(1))
      CALL section_vals_get(angle_section, explicit=my_subsection(2))
      CALL section_vals_get(torsion_section, explicit=my_subsection(3))
      CALL section_vals_get(coordination_section, explicit=my_subsection(4))
      CALL section_vals_get(plane_dist_section, explicit=my_subsection(5))
      CALL section_vals_get(rotation_section, explicit=my_subsection(6))
      CALL section_vals_get(dfunct_section, explicit=my_subsection(7))
      CALL section_vals_get(qparm_section, explicit=my_subsection(8))
      CALL section_vals_get(hydronium_section, explicit=my_subsection(9))
      ! These are just special cases since they are not present in their own defition of COLVARS
      IF (ASSOCIATED(reaction_path_section)) THEN
         CALL section_vals_get(reaction_path_section, &
                               explicit=my_subsection(10))
      END IF
      IF (ASSOCIATED(distance_from_path_section)) THEN
         CALL section_vals_get(distance_from_path_section, &
                               explicit=my_subsection(16))
      END IF
      IF (ASSOCIATED(combine_section)) THEN
         CALL section_vals_get(combine_section, explicit=my_subsection(11))
      END IF
      CALL section_vals_get(population_section, explicit=my_subsection(12))
      CALL section_vals_get(plane_plane_angle_section, &
                            explicit=my_subsection(13))
      CALL section_vals_get(gyration_section, explicit=my_subsection(14))
      CALL section_vals_get(rmsd_section, explicit=my_subsection(15))
      CALL section_vals_get(xyz_diag_section, explicit=my_subsection(17))
      CALL section_vals_get(xyz_outerdiag_section, explicit=my_subsection(18))
      CALL section_vals_get(u_section, explicit=my_subsection(19))
      CALL section_vals_get(Wc_section, explicit=my_subsection(20))
      CALL section_vals_get(HBP_section, explicit=my_subsection(21))
      CALL section_vals_get(ring_puckering_section, &
                            explicit=my_subsection(22))
      CALL section_vals_get(mindist_section, explicit=my_subsection(23))
      CALL section_vals_get(acid_hyd_dist_section, explicit=my_subsection(24))
      CALL section_vals_get(acid_hyd_shell_section, explicit=my_subsection(25))

      ! Only one colvar can be present
      IF(.NOT.(COUNT(my_subsection) == 1))CALL cp__a("colvar_methods.F",238)
      IF(.NOT.(.NOT. ASSOCIATED(colvar)))CALL cp__a("colvar_methods.F",239)

      IF (my_subsection(1)) THEN
         ! Distance
         wrk_section => distance_section
         CALL colvar_create(colvar, dist_colvar_id)
         CALL colvar_check_points(colvar, distance_section)
         CALL section_vals_val_get(distance_section, "ATOMS", i_vals=iatms)
         colvar%dist_param%i_at = iatms(1)
         colvar%dist_param%j_at = iatms(2)
         CALL section_vals_val_get(distance_section, "AXIS", i_val=colvar%dist_param%axis_id)
      ELSE IF (my_subsection(2)) THEN
         ! Angle
         wrk_section => angle_section
         CALL colvar_create(colvar, angle_colvar_id)
         CALL colvar_check_points(colvar, angle_section)
         CALL section_vals_val_get(angle_section, "ATOMS", i_vals=iatms)
         colvar%angle_param%i_at_angle = iatms
      ELSE IF (my_subsection(3)) THEN
         ! Torsion
         wrk_section => torsion_section
         CALL colvar_create(colvar, torsion_colvar_id)
         CALL colvar_check_points(colvar, torsion_section)
         CALL section_vals_val_get(torsion_section, "ATOMS", i_vals=iatms)
         colvar%torsion_param%i_at_tors = iatms
         colvar%torsion_param%o0 = 0.0_dp
      ELSE IF (my_subsection(4)) THEN
         ! Coordination
         wrk_section => coordination_section
         CALL colvar_create(colvar, coord_colvar_id)
         CALL colvar_check_points(colvar, coordination_section)
         NULLIFY (colvar%coord_param%i_at_from, colvar%coord_param%c_kinds_from)
         NULLIFY (colvar%coord_param%i_at_to, colvar%coord_param%c_kinds_to)
         NULLIFY (colvar%coord_param%i_at_to_b, colvar%coord_param%c_kinds_to_b)
         ! This section can be repeated
         CALL section_vals_val_get(coordination_section, "ATOMS_FROM", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(coordination_section, "ATOMS_FROM", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%coord_param%i_at_from, 1, ndim+SIZE(iatms))
               colvar%coord_param%i_at_from(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%coord_param%n_atoms_from = ndim
            colvar%coord_param%use_kinds_from = .FALSE.
         ELSE
            ! KINDS
            CALL section_vals_val_get(coordination_section, "KINDS_FROM", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",289)
            DO k = 1, n_var
               CALL section_vals_val_get(coordination_section, "KINDS_FROM", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%coord_param%c_kinds_from, 1, ndim+SIZE(c_kinds))
               colvar%coord_param%c_kinds_from(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%coord_param%n_atoms_from = 0
            colvar%coord_param%use_kinds_from = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%coord_param%c_kinds_from(k))
            END DO
         END IF
         ! This section can be repeated
         CALL section_vals_val_get(coordination_section, "ATOMS_TO", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(coordination_section, "ATOMS_TO", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%coord_param%i_at_to, 1, ndim+SIZE(iatms))
               colvar%coord_param%i_at_to(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%coord_param%n_atoms_to = ndim
            colvar%coord_param%use_kinds_to = .FALSE.
         ELSE
            ! KINDS
            CALL section_vals_val_get(coordination_section, "KINDS_TO", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",319)
            DO k = 1, n_var
               CALL section_vals_val_get(coordination_section, "KINDS_TO", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%coord_param%c_kinds_to, 1, ndim+SIZE(c_kinds))
               colvar%coord_param%c_kinds_to(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%coord_param%n_atoms_to = 0
            colvar%coord_param%use_kinds_to = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%coord_param%c_kinds_to(k))
            END DO
         END IF
         ! Let's finish reading the other parameters
         CALL section_vals_val_get(coordination_section, "R0", r_val=colvar%coord_param%r_0)
         CALL section_vals_val_get(coordination_section, "NN", i_val=colvar%coord_param%nncrd)
         CALL section_vals_val_get(coordination_section, "ND", i_val=colvar%coord_param%ndcrd)
         ! This section can be repeated
         CALL section_vals_val_get(coordination_section, "ATOMS_TO_B", n_rep_val=n_var)
         CALL section_vals_val_get(coordination_section, "KINDS_TO_B", n_rep_val=n_var_k)
         ndim = 0
         IF (n_var /= 0 .OR. n_var_k /= 0) THEN
            colvar%coord_param%do_chain = .TRUE.
            IF (n_var /= 0) THEN
               ! INDEX LIST
               DO k = 1, n_var
                  CALL section_vals_val_get(coordination_section, "ATOMS_TO_B", i_rep_val=k, i_vals=iatms)
                  CALL reallocate(colvar%coord_param%i_at_to_b, 1, ndim+SIZE(iatms))
                  colvar%coord_param%i_at_to_b(ndim+1:ndim+SIZE(iatms)) = iatms
                  ndim = ndim+SIZE(iatms)
               END DO
               colvar%coord_param%n_atoms_to_b = ndim
               colvar%coord_param%use_kinds_to_b = .FALSE.
            ELSE
               ! KINDS
               CALL section_vals_val_get(coordination_section, "KINDS_TO_B", n_rep_val=n_var_k)
               IF(.NOT.(n_var_k > 0))CALL cp__a("colvar_methods.F",356)
               DO k = 1, n_var_k
                  CALL section_vals_val_get(coordination_section, "KINDS_TO_B", i_rep_val=k, c_vals=c_kinds)
                  CALL reallocate(colvar%coord_param%c_kinds_to_b, 1, ndim+SIZE(c_kinds))
                  colvar%coord_param%c_kinds_to_b(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
                  ndim = ndim+SIZE(c_kinds)
               END DO
               colvar%coord_param%n_atoms_to_b = 0
               colvar%coord_param%use_kinds_to_b = .TRUE.
               ! Uppercase the label
               DO k = 1, ndim
                  CALL uppercase(colvar%coord_param%c_kinds_to_b(k))
               END DO
            END IF
            ! Let's finish reading the other parameters
            CALL section_vals_val_get(coordination_section, "R0_B", r_val=colvar%coord_param%r_0_b)
            CALL section_vals_val_get(coordination_section, "NN_B", i_val=colvar%coord_param%nncrd_b)
            CALL section_vals_val_get(coordination_section, "ND_B", i_val=colvar%coord_param%ndcrd_b)
         ELSE
            colvar%coord_param%do_chain = .FALSE.
            colvar%coord_param%n_atoms_to_b = 0
            colvar%coord_param%use_kinds_to_b = .FALSE.
            NULLIFY (colvar%coord_param%i_at_to_b)
            NULLIFY (colvar%coord_param%c_kinds_to_b)
            colvar%coord_param%nncrd_b = 0
            colvar%coord_param%ndcrd_b = 0
            colvar%coord_param%r_0_b = 0._dp
         END IF

      ELSE IF (my_subsection(5)) THEN
         ! Distance point from plane
         wrk_section => plane_dist_section
         CALL colvar_create(colvar, plane_distance_colvar_id)
         CALL colvar_check_points(colvar, plane_dist_section)
         CALL section_vals_val_get(plane_dist_section, "ATOMS_PLANE", i_vals=iatms)
         IF(.NOT.(SIZE(iatms) == 3))CALL cp__a("colvar_methods.F",391)
         colvar%plane_distance_param%plane = iatms
         CALL section_vals_val_get(plane_dist_section, "ATOM_POINT", i_val=iatm)
         colvar%plane_distance_param%point = iatm
         CALL section_vals_val_get(plane_dist_section, "PBC", l_val=colvar%plane_distance_param%use_pbc)
      ELSE IF (my_subsection(6)) THEN
         ! Rotation colvar of a segment w.r.t. another segment
         wrk_section => rotation_section
         CALL colvar_create(colvar, rotation_colvar_id)
         CALL colvar_check_points(colvar, rotation_section)
         CALL section_vals_val_get(rotation_section, "P1_BOND1", i_val=colvar%rotation_param%i_at1_bond1)
         CALL section_vals_val_get(rotation_section, "P2_BOND1", i_val=colvar%rotation_param%i_at2_bond1)
         CALL section_vals_val_get(rotation_section, "P1_BOND2", i_val=colvar%rotation_param%i_at1_bond2)
         CALL section_vals_val_get(rotation_section, "P2_BOND2", i_val=colvar%rotation_param%i_at2_bond2)
      ELSE IF (my_subsection(7)) THEN
         ! Difference of two distances
         wrk_section => dfunct_section
         CALL colvar_create(colvar, dfunct_colvar_id)
         CALL colvar_check_points(colvar, dfunct_section)
         CALL section_vals_val_get(dfunct_section, "ATOMS", i_vals=iatms)
         colvar%dfunct_param%i_at_dfunct = iatms
         CALL section_vals_val_get(dfunct_section, "COEFFICIENT", r_val=colvar%dfunct_param%coeff)
         CALL section_vals_val_get(dfunct_section, "PBC", l_val=colvar%dfunct_param%use_pbc)
      ELSE IF (my_subsection(8)) THEN
         ! Q Parameter
         wrk_section => qparm_section
         CALL colvar_create(colvar, qparm_colvar_id)
         CALL colvar_check_points(colvar, qparm_section)
         CALL section_vals_val_get(qparm_section, "RCUT", r_val=colvar%qparm_param%rcut)
         CALL section_vals_val_get(qparm_section, "ALPHA", r_val=colvar%qparm_param%alpha)
         CALL section_vals_val_get(qparm_section, "L", i_val=colvar%qparm_param%l)
         NULLIFY (colvar%qparm_param%i_at_from)
         NULLIFY (colvar%qparm_param%i_at_to)
         CALL section_vals_val_get(qparm_section, "ATOMS_FROM", n_rep_val=n_var)
         ndim = 0
         DO k = 1, n_var
            CALL section_vals_val_get(qparm_section, "ATOMS_FROM", i_rep_val=k, i_vals=iatms)
            CALL reallocate(colvar%qparm_param%i_at_from, 1, ndim+SIZE(iatms))
            colvar%qparm_param%i_at_from(ndim+1:ndim+SIZE(iatms)) = iatms
            ndim = ndim+SIZE(iatms)
         END DO
         colvar%qparm_param%n_atoms_from = ndim
         ! This section can be repeated
         CALL section_vals_val_get(qparm_section, "ATOMS_TO", n_rep_val=n_var)
         ndim = 0
         DO k = 1, n_var
            CALL section_vals_val_get(qparm_section, "ATOMS_TO", i_rep_val=k, i_vals=iatms)
            CALL reallocate(colvar%qparm_param%i_at_to, 1, ndim+SIZE(iatms))
            colvar%qparm_param%i_at_to(ndim+1:ndim+SIZE(iatms)) = iatms
            ndim = ndim+SIZE(iatms)
         END DO
         colvar%qparm_param%n_atoms_to = ndim
      ELSE IF (my_subsection(9)) THEN
         ! Hydronium
         CALL colvar_create(colvar, hydronium_colvar_id)
         NULLIFY (colvar%hydronium_param%i_oxygens)
         NULLIFY (colvar%hydronium_param%i_hydrogens)

         CALL section_vals_val_get(hydronium_section, "OXYGENS", n_rep_val=n_var)
         ndim = 0
         DO k = 1, n_var
            CALL section_vals_val_get(hydronium_section, "OXYGENS", i_vals=iatms)
            CALL reallocate(colvar%hydronium_param%i_oxygens, 1, ndim+SIZE(iatms))
            colvar%hydronium_param%i_oxygens(ndim+1:ndim+SIZE(iatms)) = iatms
            ndim = ndim+SIZE(iatms)
         END DO
         colvar%hydronium_param%n_oxygens = ndim

         CALL section_vals_val_get(hydronium_section, "HYDROGENS", n_rep_val=n_var)
         ndim = 0
         DO k = 1, n_var
            CALL section_vals_val_get(hydronium_section, "HYDROGENS", i_vals=iatms)
            CALL reallocate(colvar%hydronium_param%i_hydrogens, 1, ndim+SIZE(iatms))
            colvar%hydronium_param%i_hydrogens(ndim+1:ndim+SIZE(iatms)) = iatms
            ndim = ndim+SIZE(iatms)
         END DO
         colvar%hydronium_param%n_hydrogens = ndim

         CALL section_vals_val_get(hydronium_section, "ROO", r_val=colvar%hydronium_param%roo)
         CALL section_vals_val_get(hydronium_section, "ROH", r_val=colvar%hydronium_param%roh)
         CALL section_vals_val_get(hydronium_section, "pOH", i_val=colvar%hydronium_param%poh)
         CALL section_vals_val_get(hydronium_section, "qOH", i_val=colvar%hydronium_param%qoh)
         CALL section_vals_val_get(hydronium_section, "pOO", i_val=colvar%hydronium_param%poo)
         CALL section_vals_val_get(hydronium_section, "qOO", i_val=colvar%hydronium_param%qoo)
         CALL section_vals_val_get(hydronium_section, "pM", i_val=colvar%hydronium_param%pm)
         CALL section_vals_val_get(hydronium_section, "qM", i_val=colvar%hydronium_param%qm)
         CALL section_vals_val_get(hydronium_section, "NH", r_val=colvar%hydronium_param%nh)
         CALL section_vals_val_get(hydronium_section, "LAMBDA", r_val=colvar%hydronium_param%lambda)
      ELSE IF (my_subsection(10) .OR. my_subsection(16)) THEN
         !reaction path or distance from reaction path
         IF (my_subsection(10)) THEN
            path_section => reaction_path_section
            CALL colvar_create(colvar, reaction_path_colvar_id)
            fmid = "POS"
            ifunc = 1
         ELSE IF (my_subsection(16)) THEN
            path_section => distance_from_path_section
            CALL colvar_create(colvar, distance_from_path_colvar_id)
            fmid = "DIS"
            ifunc = 2
         END IF
         colvar%use_points = .FALSE.
         CALL section_vals_val_get(path_section, "LAMBDA", r_val=colvar%reaction_path_param%lambda)
         CALL section_vals_val_get(path_section, "DISTANCES_RMSD", l_val=colvar%reaction_path_param%dist_rmsd)
         CALL section_vals_val_get(path_section, "RMSD", l_val=colvar%reaction_path_param%rmsd)
         IF (colvar%reaction_path_param%dist_rmsd .AND. colvar%reaction_path_param%rmsd) THEN
            CALL cp__b("colvar_methods.F",497,"CV REACTION PATH: only one between DISTANCES_RMSD and RMSD can be used ")
         END IF
         IF (colvar%reaction_path_param%dist_rmsd .OR. colvar%reaction_path_param%rmsd) THEN
            NULLIFY (colvar%reaction_path_param%i_rmsd, colvar%reaction_path_param%r_ref)
            frame_section => section_vals_get_subs_vals(path_section, "FRAME")
            CALL section_vals_get(frame_section, n_repetition=nr_frame)

            colvar%reaction_path_param%nr_frames = nr_frame
            CALL read_frames(frame_section, para_env, nr_frame, colvar%reaction_path_param%r_ref, &
                             colvar%reaction_path_param%n_components)
            CALL section_vals_val_get(path_section, "SUBSET_TYPE", i_val=colvar%reaction_path_param%subset)
            IF (colvar%reaction_path_param%subset == rmsd_all) THEN
               ALLOCATE (colvar%reaction_path_param%i_rmsd(colvar%reaction_path_param%n_components))
               DO i = 1, colvar%reaction_path_param%n_components
                  colvar%reaction_path_param%i_rmsd(i) = i
               END DO
            ELSE IF (colvar%reaction_path_param%subset == rmsd_list) THEN
               ! This section can be repeated
               CALL section_vals_val_get(path_section, "ATOMS", n_rep_val=n_var)
               ndim = 0
               IF (n_var /= 0) THEN
                  ! INDEX LIST
                  DO k = 1, n_var
                     CALL section_vals_val_get(path_section, "ATOMS", i_rep_val=k, i_vals=iatms)
                     CALL reallocate(colvar%reaction_path_param%i_rmsd, 1, ndim+SIZE(iatms))
                     colvar%reaction_path_param%i_rmsd(ndim+1:ndim+SIZE(iatms)) = iatms
                     ndim = ndim+SIZE(iatms)
                  END DO
                  colvar%reaction_path_param%n_components = ndim
               ELSE
                  CALL cp__b("colvar_methods.F",527,"CV REACTION PATH: if SUBSET_TYPE=LIST a list of atoms needs to be provided ")
               END IF
            END IF

            CALL section_vals_val_get(path_section, "ALIGN_FRAMES", l_val=colvar%reaction_path_param%align_frames)
         ELSE
            colvar_subsection => section_vals_get_subs_vals(path_section, "COLVAR")
            CALL section_vals_get(colvar_subsection, n_repetition=ncol)
            ALLOCATE (colvar%reaction_path_param%colvar_p(ncol))
            IF (ncol > 0) THEN
               DO i = 1, ncol
                  NULLIFY (colvar%reaction_path_param%colvar_p(i)%colvar)
                  CALL colvar_read(colvar%reaction_path_param%colvar_p(i)%colvar, i, colvar_subsection, para_env)
               ENDDO
            ELSE
               CALL cp__b("colvar_methods.F",542,"CV REACTION PATH: the number of CV to define the path must be >0 ")
            ENDIF
            colvar%reaction_path_param%n_components = ncol
            NULLIFY (range)
            CALL section_vals_val_get(path_section, "RANGE", r_vals=range)
            CALL section_vals_val_get(path_section, "STEP_SIZE", r_val=colvar%reaction_path_param%step_size)
            iend = CEILING(MAX(RANGE(1), RANGE(2))/colvar%reaction_path_param%step_size)
            istart = FLOOR(MIN(RANGE(1), RANGE(2))/colvar%reaction_path_param%step_size)
            colvar%reaction_path_param%function_bounds(1) = istart
            colvar%reaction_path_param%function_bounds(2) = iend
            colvar%reaction_path_param%nr_frames = 2 !iend - istart + 1
            ALLOCATE (colvar%reaction_path_param%f_vals(ncol, istart:iend))
            CALL section_vals_val_get(path_section, "VARIABLE", c_vals=my_par, i_rep_val=1)
            CALL section_vals_val_get(path_section, "FUNCTION", n_rep_val=ncol)
            check = (ncol == SIZE(colvar%reaction_path_param%colvar_p))
            IF(.NOT.(check))CALL cp__a("colvar_methods.F",557)
            CALL initf(ncol)
            DO i = 1, ncol
               CALL section_vals_val_get(path_section, "FUNCTION", c_val=path_function, i_rep_val=i)
               CALL compress(path_function, full=.TRUE.)
               CALL parsef(i, TRIM(path_function), my_par)
               DO j = istart, iend
                  my_val = REAL(j, kind=dp)*colvar%reaction_path_param%step_size
                  colvar%reaction_path_param%f_vals(i, j) = evalf(i, my_val)
               END DO
            END DO
            CALL finalizef()

            iw1 = cp_print_key_unit_nr(logger, path_section, &
                                       "MAP", middle_name=fmid, extension=".dat", file_status="REPLACE")
            IF (iw1 > 0) THEN
               CALL section_vals_val_get(path_section, "MAP%GRID_SPACING", n_rep_val=ncol)
               ALLOCATE (grid_sp(ncol))
               DO i = 1, ncol
                  CALL section_vals_val_get(path_section, "MAP%GRID_SPACING", r_val=grid_sp(i))
               END DO
               CALL section_vals_val_get(path_section, "MAP%RANGE", n_rep_val=ncol)
               IF(.NOT.(ncol == SIZE(grid_sp)))CALL cp__a("colvar_methods.F",579)
               ALLOCATE (p_range(2, ncol))
               ALLOCATE (p_bounds(2, ncol))
               DO i = 1, ncol
                  CALL section_vals_val_get(path_section, "MAP%RANGE", r_vals=g_range)
                  p_range(:, i) = g_range(:)
                  p_bounds(2, i) = CEILING(MAX(p_range(1, i), p_range(2, i))/grid_sp(i))
                  p_bounds(1, i) = FLOOR(MIN(p_range(1, i), p_range(2, i))/grid_sp(i))
               END DO
               ALLOCATE (s1v(2, istart:iend))
               ALLOCATE (s1(2))
               ALLOCATE (grid_point(ncol))
               v_count = 0
               kk = rec_eval_grid(iw1, ncol, colvar%reaction_path_param%f_vals, v_count, &
                                  grid_point, grid_sp, colvar%reaction_path_param%step_size, istart, &
                                  iend, s1v, s1, p_bounds, colvar%reaction_path_param%lambda, ifunc=ifunc, &
                                  nconf=colvar%reaction_path_param%nr_frames)
               DEALLOCATE (grid_sp)
               DEALLOCATE (p_range)
               DEALLOCATE (p_bounds)
               DEALLOCATE (s1v)
               DEALLOCATE (s1)
               DEALLOCATE (grid_point)
            END IF
            CALL cp_print_key_finished_output(iw1, logger, path_section, &
                                              "MAP")
         END IF

      ELSE IF (my_subsection(11)) THEN
         ! combine colvar
         CALL colvar_create(colvar, combine_colvar_id)
         colvar%use_points = .FALSE.
         colvar_subsection => section_vals_get_subs_vals(combine_section, "COLVAR")
         CALL section_vals_get(colvar_subsection, n_repetition=ncol)
         ALLOCATE (colvar%combine_cvs_param%colvar_p(ncol))
         ! In case we need to print some information..
         iw = cp_print_key_unit_nr(logger, colvar_section, &
                                   "PRINT%PROGRAM_RUN_INFO", extension=".colvarLog")
         IF (iw > 0) THEN
            WRITE (iw, '( A )') '          '// &
               '**********************************************************************'
            WRITE (iw, '( A,I8)') ' COLVARS| COLVAR INPUT INDEX: ', icol
            WRITE (iw, '( A,T49,4I8)') ' COLVARS| COMBINATION OF THE FOLOWING COLVARS:'
         END IF
         CALL cp_print_key_finished_output(iw, logger, colvar_section, &
                                           "PRINT%PROGRAM_RUN_INFO")
         ! Parsing the real COLVARs
         DO i = 1, ncol
            NULLIFY (colvar%combine_cvs_param%colvar_p(i)%colvar)
            CALL colvar_read(colvar%combine_cvs_param%colvar_p(i)%colvar, i, colvar_subsection, para_env)
         END DO
         ! Function definition
         CALL section_vals_val_get(combine_section, "FUNCTION", c_val=colvar%combine_cvs_param%function)
         CALL compress(colvar%combine_cvs_param%function, full=.TRUE.)
         ! Variables
         CALL section_vals_val_get(combine_section, "VARIABLES", c_vals=my_par)
         ALLOCATE (colvar%combine_cvs_param%variables(SIZE(my_par)))
         colvar%combine_cvs_param%variables = my_par
         ! Check that the number of COLVAR provided is equal to the number of variables..
         IF (SIZE(my_par) /= ncol) &
            CALL cp_abort(cp__l("colvar_methods.F",639), &
                          "Number of defined COLVAR for COMBINE_COLVAR is different from the "// &
                          "number of variables! It is not possible to define COLVARs in a COMBINE_COLVAR "// &
                          "and avoid their usage in the combininig function!")
         ! Parameters
         ALLOCATE (colvar%combine_cvs_param%c_parameters(0))
         CALL section_vals_val_get(combine_section, "PARAMETERS", n_rep_val=ncol)
         DO i = 1, ncol
            isize = SIZE(colvar%combine_cvs_param%c_parameters)
            CALL section_vals_val_get(combine_section, "PARAMETERS", c_vals=my_par, i_rep_val=i)
            CALL reallocate(colvar%combine_cvs_param%c_parameters, 1, isize+SIZE(my_par))
            colvar%combine_cvs_param%c_parameters(isize+1:isize+SIZE(my_par)) = my_par
         END DO
         ALLOCATE (colvar%combine_cvs_param%v_parameters(0))
         CALL section_vals_val_get(combine_section, "VALUES", n_rep_val=ncol)
         DO i = 1, ncol
            isize = SIZE(colvar%combine_cvs_param%v_parameters)
            CALL section_vals_val_get(combine_section, "VALUES", r_vals=my_vals, i_rep_val=i)
            CALL reallocate(colvar%combine_cvs_param%v_parameters, 1, isize+SIZE(my_vals))
            colvar%combine_cvs_param%v_parameters(isize+1:isize+SIZE(my_vals)) = my_vals
         END DO
         ! Info on derivative evaluation
         CALL section_vals_val_get(combine_section, "DX", r_val=colvar%combine_cvs_param%dx)
         CALL section_vals_val_get(combine_section, "ERROR_LIMIT", r_val=colvar%combine_cvs_param%lerr)
      ELSE IF (my_subsection(12)) THEN
         ! Population
         wrk_section => population_section
         CALL colvar_create(colvar, population_colvar_id)
         CALL colvar_check_points(colvar, population_section)

         NULLIFY (colvar%population_param%i_at_from, colvar%population_param%c_kinds_from)
         NULLIFY (colvar%population_param%i_at_to, colvar%population_param%c_kinds_to)
         ! This section can be repeated

         CALL section_vals_val_get(population_section, "ATOMS_FROM", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(population_section, "ATOMS_FROM", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%population_param%i_at_from, 1, ndim+SIZE(iatms))
               colvar%population_param%i_at_from(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%population_param%n_atoms_from = ndim
            colvar%population_param%use_kinds_from = .FALSE.
         ELSE
            ! KINDS
            CALL section_vals_val_get(population_section, "KINDS_FROM", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",688)
            DO k = 1, n_var
               CALL section_vals_val_get(population_section, "KINDS_FROM", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%population_param%c_kinds_from, 1, ndim+SIZE(c_kinds))
               colvar%population_param%c_kinds_from(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%population_param%n_atoms_from = 0
            colvar%population_param%use_kinds_from = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%population_param%c_kinds_from(k))
            END DO
         END IF
         ! This section can be repeated
         CALL section_vals_val_get(population_section, "ATOMS_TO", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(population_section, "ATOMS_TO", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%population_param%i_at_to, 1, ndim+SIZE(iatms))
               colvar%population_param%i_at_to(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%population_param%n_atoms_to = ndim
            colvar%population_param%use_kinds_to = .FALSE.
         ELSE
            ! KINDS
            CALL section_vals_val_get(population_section, "KINDS_TO", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",718)
            DO k = 1, n_var
               CALL section_vals_val_get(population_section, "KINDS_TO", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%population_param%c_kinds_to, 1, ndim+SIZE(c_kinds))
               colvar%population_param%c_kinds_to(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%population_param%n_atoms_to = 0
            colvar%population_param%use_kinds_to = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%population_param%c_kinds_to(k))
            END DO
         END IF
         ! Let's finish reading the other parameters
         CALL section_vals_val_get(population_section, "R0", r_val=colvar%population_param%r_0)
         CALL section_vals_val_get(population_section, "NN", i_val=colvar%population_param%nncrd)
         CALL section_vals_val_get(population_section, "ND", i_val=colvar%population_param%ndcrd)
         CALL section_vals_val_get(population_section, "N0", i_val=colvar%population_param%n0)
         CALL section_vals_val_get(population_section, "SIGMA", r_val=colvar%population_param%sigma)
      ELSE IF (my_subsection(13)) THEN
         ! Angle between two planes
         wrk_section => plane_plane_angle_section
         CALL colvar_create(colvar, plane_plane_angle_colvar_id)
         CALL colvar_check_points(colvar, plane_plane_angle_section)
         ! Read the specification of the two planes
         plane_sections => section_vals_get_subs_vals(plane_plane_angle_section, "PLANE")
         CALL section_vals_get(plane_sections, n_repetition=n_var)
         IF (n_var /= 2) &
            CALL cp__b("colvar_methods.F",747,"PLANE_PLANE_ANGLE Colvar section: Two PLANE sections must be provided!")
         ! Plane 1
         CALL section_vals_val_get(plane_sections, "DEF_TYPE", i_rep_section=1, &
                                   i_val=colvar%plane_plane_angle_param%plane1%type_of_def)
         IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_vec) THEN
            CALL section_vals_val_get(plane_sections, "NORMAL_VECTOR", i_rep_section=1, &
                                      r_vals=s1)
            colvar%plane_plane_angle_param%plane1%normal_vec = s1
         ELSE
            CALL section_vals_val_get(plane_sections, "ATOMS", i_rep_section=1, &
                                      i_vals=iatms)
            colvar%plane_plane_angle_param%plane1%points = iatms
         END IF

         ! Plane 2
         CALL section_vals_val_get(plane_sections, "DEF_TYPE", i_rep_section=2, &
                                   i_val=colvar%plane_plane_angle_param%plane2%type_of_def)
         IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_vec) THEN
            CALL section_vals_val_get(plane_sections, "NORMAL_VECTOR", i_rep_section=2, &
                                      r_vals=s1)
            colvar%plane_plane_angle_param%plane2%normal_vec = s1
         ELSE
            CALL section_vals_val_get(plane_sections, "ATOMS", i_rep_section=2, &
                                      i_vals=iatms)
            colvar%plane_plane_angle_param%plane2%points = iatms
         END IF
      ELSE IF (my_subsection(14)) THEN
         ! Gyration Radius
         wrk_section => gyration_section
         CALL colvar_create(colvar, gyration_colvar_id)
         CALL colvar_check_points(colvar, gyration_section)

         NULLIFY (colvar%gyration_param%i_at, colvar%gyration_param%c_kinds)

         ! This section can be repeated
         CALL section_vals_val_get(gyration_section, "ATOMS", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(gyration_section, "ATOMS", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%gyration_param%i_at, 1, ndim+SIZE(iatms))
               colvar%gyration_param%i_at(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%gyration_param%n_atoms = ndim
            colvar%gyration_param%use_kinds = .FALSE.
         ELSE
            ! KINDS
            CALL section_vals_val_get(gyration_section, "KINDS", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",797)
            DO k = 1, n_var
               CALL section_vals_val_get(gyration_section, "KINDS", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%gyration_param%c_kinds, 1, ndim+SIZE(c_kinds))
               colvar%gyration_param%c_kinds(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%gyration_param%n_atoms = 0
            colvar%gyration_param%use_kinds = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%gyration_param%c_kinds(k))
            END DO
         END IF
      ELSE IF (my_subsection(15)) THEN
         ! RMSD_AB
         wrk_section => rmsd_section
         CALL colvar_create(colvar, rmsd_colvar_id)

         NULLIFY (colvar%rmsd_param%i_rmsd, colvar%rmsd_param%r_ref, colvar%rmsd_param%weights)

         frame_section => section_vals_get_subs_vals(rmsd_section, "FRAME")
         CALL section_vals_get(frame_section, n_repetition=nr_frame)

         colvar%rmsd_param%nr_frames = nr_frame
         ! Calculation is aborted if reference frame are less than 2
         IF(.NOT.(nr_frame >= 1 .AND. nr_frame <= 2))CALL cp__a("colvar_methods.F",823)
         CALL read_frames(frame_section, para_env, nr_frame, colvar%rmsd_param%r_ref, &
                          colvar%rmsd_param%n_atoms)

         ALLOCATE (colvar%rmsd_param%weights(colvar%rmsd_param%n_atoms))
         colvar%rmsd_param%weights = 0.0_dp

         CALL section_vals_val_get(rmsd_section, "SUBSET_TYPE", i_val=colvar%rmsd_param%subset)
         IF (colvar%rmsd_param%subset == rmsd_all) THEN
            ALLOCATE (colvar%rmsd_param%i_rmsd(colvar%rmsd_param%n_atoms))
            DO i = 1, colvar%rmsd_param%n_atoms
               colvar%rmsd_param%i_rmsd(i) = i
            END DO
         ELSE IF (colvar%rmsd_param%subset == rmsd_list) THEN
            ! This section can be repeated
            CALL section_vals_val_get(rmsd_section, "ATOMS", n_rep_val=n_var)
            ndim = 0
            IF (n_var /= 0) THEN
               ! INDEX LIST
               DO k = 1, n_var
                  CALL section_vals_val_get(rmsd_section, "ATOMS", i_rep_val=k, i_vals=iatms)
                  CALL reallocate(colvar%rmsd_param%i_rmsd, 1, ndim+SIZE(iatms))
                  colvar%rmsd_param%i_rmsd(ndim+1:ndim+SIZE(iatms)) = iatms
                  ndim = ndim+SIZE(iatms)
               END DO
               colvar%rmsd_param%n_atoms = ndim
            ELSE
               CALL cp__b("colvar_methods.F",850,"CV RMSD: if SUBSET_TYPE=LIST a list of atoms needs to be provided ")
            END IF
         ELSE IF (colvar%rmsd_param%subset == rmsd_weightlist) THEN
            CALL section_vals_val_get(rmsd_section, "ATOMS", n_rep_val=n_var)
            ndim = 0
            IF (n_var /= 0) THEN
               ! INDEX LIST
               DO k = 1, n_var
                  CALL section_vals_val_get(rmsd_section, "ATOMS", i_rep_val=k, i_vals=iatms)
                  CALL reallocate(colvar%rmsd_param%i_rmsd, 1, ndim+SIZE(iatms))
                  colvar%rmsd_param%i_rmsd(ndim+1:ndim+SIZE(iatms)) = iatms
                  ndim = ndim+SIZE(iatms)
               END DO
               colvar%rmsd_param%n_atoms = ndim
            ELSE
               CALL cp__b("colvar_methods.F",865,"CV RMSD: if SUBSET_TYPE=WEIGHT_LIST a list of atoms needs to be provided ")
            END IF
            CALL section_vals_val_get(rmsd_section, "WEIGHTS", n_rep_val=n_var)
            ndim = 0
            IF (n_var /= 0) THEN
               ! INDEX LIST
               DO k = 1, n_var
                  CALL section_vals_val_get(rmsd_section, "WEIGHTS", i_rep_val=k, r_vals=wei)
                  CALL reallocate(weights, 1, ndim+SIZE(wei))
                  weights(ndim+1:ndim+SIZE(wei)) = wei
                  ndim = ndim+SIZE(wei)
               END DO
               IF (ndim /= colvar%rmsd_param%n_atoms) &
                  CALL cp_abort(cp__l("colvar_methods.F",878), "CV RMSD: list of atoms and list of "// &
                                "weights need to contain same number of entries. ")
               DO i = 1, ndim
                  ii = colvar%rmsd_param%i_rmsd(i)
                  colvar%rmsd_param%weights(ii) = weights(i)
               END DO
               DEALLOCATE (weights)
            ELSE
               CALL cp__b("colvar_methods.F",886,"CV RMSD: if SUBSET_TYPE=WEIGHT_LIST a list of weights need to be provided. ")
            END IF

         ELSE
            CALL cp__b("colvar_methods.F",890,"CV RMSD: unknown SUBSET_TYPE.")
         END IF

         CALL section_vals_val_get(rmsd_section, "ALIGN_FRAMES", l_val=colvar%rmsd_param%align_frames)
      ELSE IF (my_subsection(17)) THEN
         ! Work on XYZ positions of atoms
         wrk_section => xyz_diag_section
         CALL colvar_create(colvar, xyz_diag_colvar_id)
         CALL colvar_check_points(colvar, wrk_section)
         CALL section_vals_val_get(wrk_section, "ATOM", i_val=iatm)
         CALL section_vals_val_get(wrk_section, "COMPONENT", i_val=icomponent)
         CALL section_vals_val_get(wrk_section, "PBC", l_val=colvar%xyz_diag_param%use_pbc)
         CALL section_vals_val_get(wrk_section, "ABSOLUTE_POSITION", l_val=colvar%xyz_diag_param%use_absolute_position)
         colvar%xyz_diag_param%i_atom = iatm
         colvar%xyz_diag_param%component = icomponent
      ELSE IF (my_subsection(18)) THEN
         ! Work on the outer diagonal (two atoms A,B) XYZ positions
         wrk_section => xyz_outerdiag_section
         CALL colvar_create(colvar, xyz_outerdiag_colvar_id)
         CALL colvar_check_points(colvar, wrk_section)
         CALL section_vals_val_get(wrk_section, "ATOMS", i_vals=iatms)
         colvar%xyz_outerdiag_param%i_atoms = iatms
         CALL section_vals_val_get(wrk_section, "COMPONENT_A", i_val=icomponent)
         colvar%xyz_outerdiag_param%components(1) = icomponent
         CALL section_vals_val_get(wrk_section, "COMPONENT_B", i_val=icomponent)
         colvar%xyz_outerdiag_param%components(2) = icomponent
         CALL section_vals_val_get(wrk_section, "PBC", l_val=colvar%xyz_outerdiag_param%use_pbc)
      ELSE IF (my_subsection(19)) THEN
         ! Energy
         wrk_section => u_section
         CALL colvar_create(colvar, u_colvar_id)
         colvar%u_param%mixed_energy_section => section_vals_get_subs_vals(wrk_section, "MIXED")
         CALL section_vals_get(colvar%u_param%mixed_energy_section, explicit=use_mixed_energy)
         IF (.NOT. use_mixed_energy) NULLIFY (colvar%u_param%mixed_energy_section)
      ELSE IF (my_subsection(20)) THEN
         ! Wc hydrogen bond
         wrk_section => Wc_section
         CALL colvar_create(colvar, Wc_colvar_id)
         CALL colvar_check_points(colvar, Wc_section)
         CALL section_vals_val_get(Wc_section, "ATOMS", i_vals=iatms)
         CALL section_vals_val_get(wrk_section, "RCUT", r_val=my_val(1))
         colvar%Wc%rcut = cp_unit_to_cp2k(my_val(1), "angstrom")
         colvar%Wc%ids = iatms
      ELSE IF (my_subsection(21)) THEN
         ! HBP colvar
         wrk_section => HBP_section
         CALL colvar_create(colvar, HBP_colvar_id)
         CALL colvar_check_points(colvar, HBP_section)
         CALL section_vals_val_get(wrk_section, "NPOINTS", i_val=colvar%HBP%nPoints)
         CALL section_vals_val_get(wrk_section, "RCUT", r_val=my_val(1))
         colvar%HBP%rcut = cp_unit_to_cp2k(my_val(1), "angstrom")
         CALL section_vals_val_get(wrk_section, "RCUT", r_val=colvar%HBP%shift)

         ALLOCATE (colvar%HBP%ids(colvar%HBP%nPoints, 3))
         ALLOCATE (colvar%HBP%ewc(colvar%HBP%nPoints))
         DO i = 1, colvar%HBP%nPoints
            CALL section_vals_val_get(wrk_section, "ATOMS", i_rep_val=i, i_vals=iatms)
            colvar%HBP%ids(i, :) = iatms
         ENDDO
      ELSE IF (my_subsection(22)) THEN
         ! Ring Puckering
         CALL colvar_create(colvar, ring_puckering_colvar_id)
         CALL section_vals_val_get(ring_puckering_section, "ATOMS", i_vals=iatms)
         colvar%ring_puckering_param%nring = SIZE(iatms)
         ALLOCATE (colvar%ring_puckering_param%atoms(SIZE(iatms)))
         colvar%ring_puckering_param%atoms = iatms
         CALL section_vals_val_get(ring_puckering_section, "COORDINATE", &
                                   i_val=colvar%ring_puckering_param%iq)
         ! test the validity of the parameters
         ndim = colvar%ring_puckering_param%nring
         IF (ndim <= 3) &
            CALL cp__b("colvar_methods.F",961,"CV Ring Puckering: Ring size has to be 4 or larger. ")
         ii = colvar%ring_puckering_param%iq
         IF (ABS(ii) == 1 .OR. ii < -(ndim-1)/2 .OR. ii > ndim/2) &
            CALL cp__b("colvar_methods.F",964,"CV Ring Puckering: Invalid coordinate number.")
      ELSE IF (my_subsection(23)) THEN
         ! Minimum Distance
         wrk_section => mindist_section
         CALL colvar_create(colvar, mindist_colvar_id)
         CALL colvar_check_points(colvar, mindist_section)
         NULLIFY (colvar%mindist_param%i_dist_from, colvar%mindist_param%i_coord_from, &
                  colvar%mindist_param%k_coord_from, colvar%mindist_param%i_coord_to, &
                  colvar%mindist_param%k_coord_to)
         CALL section_vals_val_get(mindist_section, "ATOMS_DISTANCE", i_vals=iatms)
         colvar%mindist_param%n_dist_from = SIZE(iatms)
         ALLOCATE (colvar%mindist_param%i_dist_from(SIZE(iatms)))
         colvar%mindist_param%i_dist_from = iatms
         CALL section_vals_val_get(mindist_section, "ATOMS_FROM", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(mindist_section, "ATOMS_FROM", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%mindist_param%i_coord_from, 1, ndim+SIZE(iatms))
               colvar%mindist_param%i_coord_from(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%mindist_param%n_coord_from = ndim
            colvar%mindist_param%use_kinds_from = .FALSE.
         ELSE
            !KINDS
            CALL section_vals_val_get(mindist_section, "KINDS_FROM", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",992)
            DO k = 1, n_var
               CALL section_vals_val_get(mindist_section, "KINDS_FROM", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%mindist_param%k_coord_from, 1, ndim+SIZE(c_kinds))
               colvar%mindist_param%k_coord_from(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%mindist_param%n_coord_from = 0
            colvar%mindist_param%use_kinds_from = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%mindist_param%k_coord_from(k))
            END DO
         END IF

         CALL section_vals_val_get(mindist_section, "ATOMS_TO", n_rep_val=n_var)
         ndim = 0
         IF (n_var /= 0) THEN
            ! INDEX LIST
            DO k = 1, n_var
               CALL section_vals_val_get(mindist_section, "ATOMS_TO", i_rep_val=k, i_vals=iatms)
               CALL reallocate(colvar%mindist_param%i_coord_to, 1, ndim+SIZE(iatms))
               colvar%mindist_param%i_coord_to(ndim+1:ndim+SIZE(iatms)) = iatms
               ndim = ndim+SIZE(iatms)
            END DO
            colvar%mindist_param%n_coord_to = ndim
            colvar%mindist_param%use_kinds_to = .FALSE.
         ELSE
            !KINDS
            CALL section_vals_val_get(mindist_section, "KINDS_TO", n_rep_val=n_var)
            IF(.NOT.(n_var > 0))CALL cp__a("colvar_methods.F",1022)
            DO k = 1, n_var
               CALL section_vals_val_get(mindist_section, "KINDS_TO", i_rep_val=k, c_vals=c_kinds)
               CALL reallocate(colvar%mindist_param%k_coord_to, 1, ndim+SIZE(c_kinds))
               colvar%mindist_param%k_coord_to(ndim+1:ndim+SIZE(c_kinds)) = c_kinds
               ndim = ndim+SIZE(c_kinds)
            END DO
            colvar%mindist_param%n_coord_to = 0
            colvar%mindist_param%use_kinds_to = .TRUE.
            ! Uppercase the label
            DO k = 1, ndim
               CALL uppercase(colvar%mindist_param%k_coord_to(k))
            END DO
         END IF

         CALL section_vals_val_get(mindist_section, "R0", r_val=colvar%mindist_param%r_cut)
         CALL section_vals_val_get(mindist_section, "NN", i_val=colvar%mindist_param%p_exp)
         CALL section_vals_val_get(mindist_section, "ND", i_val=colvar%mindist_param%q_exp)
!       CALL section_vals_val_get(mindist_section,"NC",r_val=colvar%mindist_param%n_cut)
         CALL section_vals_val_get(mindist_section, "LAMBDA", r_val=colvar%mindist_param%lambda)
      ELSE IF (my_subsection(24)) THEN
         ! Distance carboxylic acid and hydronium
         CALL colvar_create(colvar, acid_hyd_dist_colvar_id)
         NULLIFY (colvar%acid_hyd_dist_param%i_oxygens_water)
         NULLIFY (colvar%acid_hyd_dist_param%i_oxygens_acid)
         NULLIFY (colvar%acid_hyd_dist_param%i_hydrogens)
         CALL read_acid_hydronium_colvars(acid_hyd_dist_section, colvar, acid_hyd_dist_colvar_id, &
                                          colvar%acid_hyd_dist_param%n_oxygens_water, &
                                          colvar%acid_hyd_dist_param%n_oxygens_acid, &
                                          colvar%acid_hyd_dist_param%n_hydrogens, &
                                          colvar%acid_hyd_dist_param%i_oxygens_water, &
                                          colvar%acid_hyd_dist_param%i_oxygens_acid, &
                                          colvar%acid_hyd_dist_param%i_hydrogens)
      ELSE IF (my_subsection(25)) THEN
         ! Number of oxygens in 1st shell of hydronium for carboxylic acid / water system
         CALL colvar_create(colvar, acid_hyd_shell_colvar_id)
         NULLIFY (colvar%acid_hyd_shell_param%i_oxygens_water)
         NULLIFY (colvar%acid_hyd_shell_param%i_oxygens_acid)
         NULLIFY (colvar%acid_hyd_shell_param%i_hydrogens)
         CALL read_acid_hydronium_colvars(acid_hyd_shell_section, colvar, acid_hyd_shell_colvar_id, &
                                          colvar%acid_hyd_shell_param%n_oxygens_water, &
                                          colvar%acid_hyd_shell_param%n_oxygens_acid, &
                                          colvar%acid_hyd_shell_param%n_hydrogens, &
                                          colvar%acid_hyd_shell_param%i_oxygens_water, &
                                          colvar%acid_hyd_shell_param%i_oxygens_acid, &
                                          colvar%acid_hyd_shell_param%i_hydrogens)
      END IF
      CALL colvar_setup(colvar)

      iw = cp_print_key_unit_nr(logger, colvar_section, &
                                "PRINT%PROGRAM_RUN_INFO", extension=".colvarLog")
      IF (iw > 0) THEN
         tag = "ATOMS: "
         IF (colvar%use_points) tag = "POINTS:"
         ! Description header
         IF (colvar%type_id /= combine_colvar_id) THEN
            WRITE (iw, '( A )') '          '// &
               '----------------------------------------------------------------------'
            WRITE (iw, '( A,I8)') ' COLVARS| COLVAR INPUT INDEX: ', icol
         END IF
         ! Colvar Description
         SELECT CASE (colvar%type_id)
         CASE (angle_colvar_id)
            WRITE (iw, '( A,T57,3I8)') ' COLVARS| ANGLE          >>> '//tag, &
               colvar%angle_param%i_at_angle
         CASE (dfunct_colvar_id)
            WRITE (iw, '( A,T49,4I8)') ' COLVARS| DISTANCE DIFFERENCE  >>> '//tag, &
               colvar%dfunct_param%i_at_dfunct
         CASE (plane_distance_colvar_id)
            WRITE (iw, '( A,T57,3I8)') ' COLVARS| PLANE DISTANCE - PLANE  >>> '//tag, &
               colvar%plane_distance_param%plane
            WRITE (iw, '( A,T73,1I8)') ' COLVARS| PLANE DISTANCE - POINT  >>> '//tag, &
               colvar%plane_distance_param%point
         CASE (plane_plane_angle_colvar_id)
            IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) THEN
               WRITE (iw, '( A,T57,3I8)') ' COLVARS| PLANE-PLANE ANGLE - PLANE 1  (ATOMS) >>> '//tag, &
                  colvar%plane_plane_angle_param%plane1%points
            ELSE
               WRITE (iw, '( A,T57,3F8.3)') ' COLVARS| PLANE-PLANE ANGLE - PLANE 1 (VECTOR) >>> '//tag, &
                  colvar%plane_plane_angle_param%plane1%normal_vec
            END IF

            IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) THEN
               WRITE (iw, '( A,T57,3I8)') ' COLVARS| PLANE-PLANE ANGLE - PLANE 1  (ATOMS) >>> '//tag, &
                  colvar%plane_plane_angle_param%plane2%points
            ELSE
               WRITE (iw, '( A,T57,3F8.3)') ' COLVARS| PLANE-PLANE ANGLE - PLANE 1 (VECTOR) >>> '//tag, &
                  colvar%plane_plane_angle_param%plane2%normal_vec
            END IF
         CASE (torsion_colvar_id)
            WRITE (iw, '( A,T49,4I8)') ' COLVARS| TORSION       >>> '//tag, &
               colvar%torsion_param%i_at_tors
         CASE (dist_colvar_id)
            WRITE (iw, '( A,T65,2I8)') ' COLVARS| BOND          >>> '//tag, &
               colvar%dist_param%i_at, colvar%dist_param%j_at
         CASE (coord_colvar_id)
            IF (colvar%coord_param%do_chain) THEN
               WRITE (iw, '( A)') ' COLVARS| COORDINATION CHAIN FC(from->to)*FC(to->to_B)>> '
            END IF
            IF (colvar%coord_param%use_kinds_from) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| COORDINATION  >>> FROM KINDS', &
                                           ADJUSTR(colvar%coord_param%c_kinds_from(kk) (1:10)), &
                                           kk=1, SIZE(colvar%coord_param%c_kinds_from))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| COORDINATION  >>> FROM '//tag, &
                                           colvar%coord_param%i_at_from(kk), &
                                           kk=1, SIZE(colvar%coord_param%i_at_from))
            END IF
            IF (colvar%coord_param%use_kinds_to) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| COORDINATION  >>>   TO KINDS', &
                                           ADJUSTR(colvar%coord_param%c_kinds_to(kk) (1:10)), &
                                           kk=1, SIZE(colvar%coord_param%c_kinds_to))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| COORDINATION  >>>   TO '//tag, &
                                           colvar%coord_param%i_at_to(kk), &
                                           kk=1, SIZE(colvar%coord_param%i_at_to))
            END IF
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| R0', colvar%coord_param%r_0
            WRITE (iw, '( A,T71,I10)') ' COLVARS| NN', colvar%coord_param%nncrd
            WRITE (iw, '( A,T71,I10)') ' COLVARS| ND', colvar%coord_param%ndcrd
            IF (colvar%coord_param%do_chain) THEN
               IF (colvar%coord_param%use_kinds_to_b) THEN
                  WRITE (iw, '( A,T71,A10)') (' COLVARS| COORDINATION  >>>   TO KINDS B', &
                                              ADJUSTR(colvar%coord_param%c_kinds_to_b(kk) (1:10)), &
                                              kk=1, SIZE(colvar%coord_param%c_kinds_to_b))
               ELSE
                  WRITE (iw, '( A,T71,I10)') (' COLVARS| COORDINATION  >>>   TO '//tag//' B', &
                                              colvar%coord_param%i_at_to_b(kk), &
                                              kk=1, SIZE(colvar%coord_param%i_at_to_b))
               END IF
               WRITE (iw, '( A,T71,F10.5)') ' COLVARS| R0 B', colvar%coord_param%r_0_b
               WRITE (iw, '( A,T71,I10)') ' COLVARS| NN B', colvar%coord_param%nncrd_b
               WRITE (iw, '( A,T71,I10)') ' COLVARS| ND B', colvar%coord_param%ndcrd_b
            END IF
         CASE (population_colvar_id)
            IF (colvar%population_param%use_kinds_from) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| POPULATION based on coordination >>> FROM KINDS', &
                                           ADJUSTR(colvar%population_param%c_kinds_from(kk) (1:10)), &
                                           kk=1, SIZE(colvar%population_param%c_kinds_from))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| POPULATION based on coordination >>> FROM '//tag, &
                                           colvar%population_param%i_at_from(kk), &
                                           kk=1, SIZE(colvar%population_param%i_at_from))
            END IF
            IF (colvar%population_param%use_kinds_to) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| POPULATION based on coordination >>>   TO KINDS', &
                                           ADJUSTR(colvar%population_param%c_kinds_to(kk) (1:10)), &
                                           kk=1, SIZE(colvar%population_param%c_kinds_to))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| POPULATION based on coordination   >>>   TO '//tag, &
                                           colvar%population_param%i_at_to(kk), &
                                           kk=1, SIZE(colvar%population_param%i_at_to))
            END IF
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| R0', colvar%population_param%r_0
            WRITE (iw, '( A,T71,I10)') ' COLVARS| NN', colvar%population_param%nncrd
            WRITE (iw, '( A,T71,I10)') ' COLVARS| ND', colvar%population_param%ndcrd
            WRITE (iw, '( A,T71,I10)') ' COLVARS| N0', colvar%population_param%n0
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| SIGMA', colvar%population_param%sigma
         CASE (gyration_colvar_id)
            IF (colvar%gyration_param%use_kinds) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| Gyration Radius >>> KINDS', &
                                           ADJUSTR(colvar%gyration_param%c_kinds(kk) (1:10)), &
                                           kk=1, SIZE(colvar%gyration_param%c_kinds))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| Gyration Radius >>> ATOMS '//tag, &
                                           colvar%gyration_param%i_at(kk), &
                                           kk=1, SIZE(colvar%gyration_param%i_at))
            END IF
         CASE (rotation_colvar_id)
            WRITE (iw, '( A,T71,I10)') ' COLVARS| BOND_ROTATION   - POINT 1 LINE 1  >>> '//tag, &
               colvar%rotation_param%i_at1_bond1
            WRITE (iw, '( A,T71,I10)') ' COLVARS| BOND_ROTATION   - POINT 2 LINE 1  >>> '//tag, &
               colvar%rotation_param%i_at2_bond1
            WRITE (iw, '( A,T71,I10)') ' COLVARS| BOND_ROTATION   - POINT 1 LINE 2  >>> '//tag, &
               colvar%rotation_param%i_at1_bond2
            WRITE (iw, '( A,T71,I10)') ' COLVARS| BOND_ROTATION   - POINT 2 LINE 2  >>> '//tag, &
               colvar%rotation_param%i_at2_bond2
         CASE (qparm_colvar_id)
            WRITE (iw, '( A,T71,I10)') (' COLVARS| Q-PARM  >>> FROM '//tag, &
                                        colvar%qparm_param%i_at_from(kk), &
                                        kk=1, SIZE(colvar%qparm_param%i_at_from))
            WRITE (iw, '( A,T71,I10)') (' COLVARS| Q-PARM  >>>   TO '//tag, &
                                        colvar%qparm_param%i_at_to(kk), &
                                        kk=1, SIZE(colvar%qparm_param%i_at_to))
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| RCUT', colvar%qparm_param%rcut
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| ALPHA', colvar%qparm_param%alpha
            WRITE (iw, '( A,T71,I10)') ' COLVARS| L', colvar%qparm_param%l
         CASE (combine_colvar_id)
            WRITE (iw, '( A)') ' COLVARS| COMBINING FUNCTION : '// &
               TRIM(colvar%combine_cvs_param%function)
            WRITE (iw, '( A)', ADVANCE="NO") ' COLVARS| VARIABLES : '
            DO i = 1, SIZE(colvar%combine_cvs_param%variables)
               WRITE (iw, '( A)', ADVANCE="NO") &
                  TRIM(colvar%combine_cvs_param%variables(i))//" "
            END DO
            WRITE (iw, '(/)')
            WRITE (iw, '( A)') ' COLVARS| DEFINED PARAMETERS [label]  [value]:'
            DO i = 1, SIZE(colvar%combine_cvs_param%c_parameters)
               WRITE (iw, '( A,A7,F9.3)') '                            ', &
                  TRIM(colvar%combine_cvs_param%c_parameters(i)), colvar%combine_cvs_param%v_parameters(i)
            END DO
            WRITE (iw, '( A,T71,G10.5)') ' COLVARS| ERROR ON DERIVATIVE EVALUATION', &
               colvar%combine_cvs_param%lerr
            WRITE (iw, '( A,T71,G10.5)') ' COLVARS| DX', &
               colvar%combine_cvs_param%dx
         CASE (reaction_path_colvar_id)
            CALL cp__w("colvar_methods.F",1228,"Description header for REACTION_PATH COLVAR missing!!")
         CASE (distance_from_path_colvar_id)
            CALL cp__w("colvar_methods.F",1230,"Description header for REACTION_PATH COLVAR missing!!")
         CASE (hydronium_colvar_id)
            WRITE (iw, '( A,T71,I10)') ' COLVARS| POH', colvar%hydronium_param%poh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QOH', colvar%hydronium_param%qoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| POO', colvar%hydronium_param%poo
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QOO', colvar%hydronium_param%qoo
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| ROO', colvar%hydronium_param%roo
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| ROH', colvar%hydronium_param%roh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| NH', colvar%hydronium_param%nh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| LAMBDA', colvar%hydronium_param%lambda
         CASE (acid_hyd_dist_colvar_id)
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PAOH', colvar%acid_hyd_dist_param%paoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QAOH', colvar%acid_hyd_dist_param%qaoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PWOH', colvar%acid_hyd_dist_param%pwoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QWOH', colvar%acid_hyd_dist_param%qwoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PCUT', colvar%acid_hyd_dist_param%pcut
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QCUT', colvar%acid_hyd_dist_param%qcut
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| RAOH', colvar%acid_hyd_dist_param%raoh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| RWOH', colvar%acid_hyd_dist_param%rwoh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| NC', colvar%acid_hyd_dist_param%nc
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| LAMBDA', colvar%acid_hyd_dist_param%lambda
         CASE (acid_hyd_shell_colvar_id)
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PAOH', colvar%acid_hyd_shell_param%paoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QAOH', colvar%acid_hyd_shell_param%qaoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PWOH', colvar%acid_hyd_shell_param%pwoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QWOH', colvar%acid_hyd_shell_param%qwoh
            WRITE (iw, '( A,T71,I10)') ' COLVARS| POO', colvar%acid_hyd_shell_param%poo
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QOO', colvar%acid_hyd_shell_param%qoo
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PM', colvar%acid_hyd_shell_param%pm
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QM', colvar%acid_hyd_shell_param%qm
            WRITE (iw, '( A,T71,I10)') ' COLVARS| PCUT', colvar%acid_hyd_shell_param%pcut
            WRITE (iw, '( A,T71,I10)') ' COLVARS| QCUT', colvar%acid_hyd_shell_param%qcut
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| RAOH', colvar%acid_hyd_shell_param%raoh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| RWOH', colvar%acid_hyd_shell_param%rwoh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| ROO', colvar%acid_hyd_shell_param%roo
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| NH', colvar%acid_hyd_shell_param%nh
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| NC', colvar%acid_hyd_shell_param%nc
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| LAMBDA', colvar%acid_hyd_shell_param%lambda
         CASE (rmsd_colvar_id)
            CALL cp__w("colvar_methods.F",1269,"Description header for RMSD COLVAR missing!!")
         CASE (xyz_diag_colvar_id)
            NULLIFY (section, keyword, enum)
            CALL create_colvar_xyz_d_section(section)
            keyword => section_get_keyword(section, "COMPONENT")
            CALL keyword_get(keyword, enum=enum)
            tag_comp = enum_i2c(enum, colvar%xyz_diag_param%component)
            CALL section_release(section)

            WRITE (iw, '( A,T57,3I8)') ' COLVARS| POSITION ('//TRIM(tag_comp) &
               //')  >>> '//tag, colvar%xyz_diag_param%i_atom
         CASE (xyz_outerdiag_colvar_id)
            NULLIFY (section, keyword, enum)
            CALL create_colvar_xyz_od_section(section)
            keyword => section_get_keyword(section, "COMPONENT_A")
            CALL keyword_get(keyword, enum=enum)
            tag_comp1 = enum_i2c(enum, colvar%xyz_outerdiag_param%components(1))
            keyword => section_get_keyword(section, "COMPONENT_B")
            CALL keyword_get(keyword, enum=enum)
            tag_comp2 = enum_i2c(enum, colvar%xyz_outerdiag_param%components(2))
            CALL section_release(section)

            WRITE (iw, '( A,T57,3I8)') ' COLVARS| CROSS TERM POSITION ('//TRIM(tag_comp1) &
               //" * "//TRIM(tag_comp2)//')  >>> '//tag, colvar%xyz_outerdiag_param%i_atoms
         CASE (u_colvar_id)
            WRITE (iw, '( A,T77,A4)') ' COLVARS| ENERGY          >>> '//tag, 'all!'
         CASE (Wc_colvar_id)
            WRITE (iw, '( A,T57,F16.8)') ' COLVARS| Wc          >>> RCUT: ', &
               colvar%Wc%rcut
            WRITE (iw, '( A,T57,3I8)') ' COLVARS| Wc          >>> '//tag, &
               colvar%Wc%ids
         CASE (HBP_colvar_id)
            WRITE (iw, '( A,T57,I8)') ' COLVARS| HBP          >>> NPOINTS', &
               colvar%HBP%nPoints
            WRITE (iw, '( A,T57,F16.8)') ' COLVARS| HBP          >>> RCUT', &
               colvar%HBP%rcut
            WRITE (iw, '( A,T57,F16.8)') ' COLVARS| HBP          >>> RCUT', &
               colvar%HBP%shift
            DO i = 1, colvar%HBP%nPoints
               WRITE (iw, '( A,T57,3I8)') ' COLVARS| HBP          >>> '//tag, &
                  colvar%HBP%ids(i, :)
            ENDDO
         CASE (ring_puckering_colvar_id)
            WRITE (iw, '( A,T57,I8)') ' COLVARS| Ring Puckering      >>> ring size', &
               colvar%ring_puckering_param%nring
            IF (colvar%ring_puckering_param%iq == 0) THEN
               WRITE (iw, '( A,T40,A)') ' COLVARS| Ring Puckering      >>> coordinate', &
                  ' Total Puckering Amplitude'
            ELSEIF (colvar%ring_puckering_param%iq > 0) THEN
               WRITE (iw, '( A,T35,A,T57,I8)') ' COLVARS| Ring Puckering      >>> coordinate', &
                  ' Puckering Amplitude', &
                  colvar%ring_puckering_param%iq
            ELSE
               WRITE (iw, '( A,T35,A,T57,I8)') ' COLVARS| Ring Puckering      >>> coordinate', &
                  ' Puckering Angle', &
                  colvar%ring_puckering_param%iq
            END IF
         CASE (mindist_colvar_id)
            WRITE (iw, '( A)') ' COLVARS| CONDITIONED DISTANCE>> '
            WRITE (iw, '( A,T71,I10)') (' COLVARS| COND.DISTANCE  >>> DISTANCE FROM '//tag, &
                                        colvar%mindist_param%i_dist_from(kk), &
                                        kk=1, SIZE(colvar%mindist_param%i_dist_from))
            IF (colvar%mindist_param%use_kinds_from) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| COND.DIST.  >>> COORDINATION FROM KINDS ', &
                                           ADJUSTR(colvar%mindist_param%k_coord_from(kk) (1:10)), &
                                           kk=1, SIZE(colvar%mindist_param%k_coord_from))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| COND.DIST.  >>> COORDINATION FROM '//tag, &
                                           colvar%mindist_param%i_coord_from(kk), &
                                           kk=1, SIZE(colvar%mindist_param%i_coord_from))
            END IF
            IF (colvar%mindist_param%use_kinds_to) THEN
               WRITE (iw, '( A,T71,A10)') (' COLVARS| COND.DIST.  >>> COORDINATION TO KINDS ', &
                                           ADJUSTR(colvar%mindist_param%k_coord_to(kk) (1:10)), &
                                           kk=1, SIZE(colvar%mindist_param%k_coord_to))
            ELSE
               WRITE (iw, '( A,T71,I10)') (' COLVARS| COND.DIST.  >>> COORDINATION TO '//tag, &
                                           colvar%mindist_param%i_coord_to(kk), &
                                           kk=1, SIZE(colvar%mindist_param%i_coord_to))
            END IF
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| R0', colvar%mindist_param%r_cut
            WRITE (iw, '( A,T71,I10)') ' COLVARS| NN', colvar%mindist_param%p_exp
            WRITE (iw, '( A,T71,I10)') ' COLVARS| ND', colvar%mindist_param%q_exp
            WRITE (iw, '( A,T71,F10.5)') ' COLVARS| LAMBDA', colvar%mindist_param%lambda

         END SELECT
         IF (colvar%use_points) THEN
            WRITE (iw, '( A)') ' COLVARS| INFORMATION ON DEFINED GEOMETRICAL POINTS'
            DO kk = 1, SIZE(colvar%points)
               point_section => section_vals_get_subs_vals(wrk_section, "POINT")
               CALL section_vals_val_get(point_section, "TYPE", i_rep_section=kk, c_val=tmpStr)
               tmpStr2 = cp_to_string(kk)
               WRITE (iw, '( A)') ' COLVARS|  POINT Nr.'//TRIM(tmpStr2)//' OF TYPE: '//TRIM(tmpStr)
               IF (ASSOCIATED(colvar%points(kk)%atoms)) THEN
                  WRITE (iw, '( A)') ' COLVARS|   ATOMS BUILDING THE GEOMETRICAL POINT'
                  WRITE (iw, '( A, I10)') (' COLVARS|   ATOM:', colvar%points(kk)%atoms(k), k=1, SIZE(colvar%points(kk)%atoms))
               ELSE
                  WRITE (iw, '( A,4X,3F12.6)') ' COLVARS|   XYZ POSITION OF FIXED POINT:', colvar%points(kk)%r
               END IF
            END DO
         END IF
         ! Close the description layer
         IF (colvar%type_id /= combine_colvar_id) THEN
            WRITE (iw, '( A )') '          '// &
               '----------------------------------------------------------------------'
         ELSE
            WRITE (iw, '( A )') '          '// &
               '**********************************************************************'
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, colvar_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)
   END SUBROUTINE colvar_read

! **************************************************************************************************
!> \brief read collective variables for the dissociation of a carboxylic acid
!>        in water
!> \param section ...
!> \param colvar collective variable
!> \param colvar_id  ...
!> \param n_oxygens_water number of oxygens of water molecules
!> \param n_oxygens_acid number of oxgyens of carboxyl groups
!> \param n_hydrogens number of hydrogens (water and carboxyl group)
!> \param i_oxygens_water list of oxygens of water molecules
!> \param i_oxygens_acid list of oxygens of carboxyl group
!> \param i_hydrogens list of hydrogens (water and carboxyl group)
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE read_acid_hydronium_colvars(section, colvar, colvar_id, n_oxygens_water, &
                                          n_oxygens_acid, n_hydrogens, i_oxygens_water, &
                                          i_oxygens_acid, i_hydrogens)
      TYPE(section_vals_type), POINTER                   :: section
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: colvar_id
      INTEGER, INTENT(OUT)                               :: n_oxygens_water, n_oxygens_acid, &
                                                            n_hydrogens
      INTEGER, DIMENSION(:), POINTER                     :: i_oxygens_water, i_oxygens_acid, &
                                                            i_hydrogens

      CHARACTER(len=*), PARAMETER :: routineN = 'read_acid_hydronium_colvars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k, n_var, ndim
      INTEGER, DIMENSION(:), POINTER                     :: iatms

      NULLIFY (iatms)

      CALL section_vals_val_get(section, "OXYGENS_WATER", n_rep_val=n_var)
      ndim = 0
      DO k = 1, n_var
         CALL section_vals_val_get(section, "OXYGENS_WATER", i_vals=iatms)
         CALL reallocate(i_oxygens_water, 1, ndim+SIZE(iatms))
         i_oxygens_water(ndim+1:ndim+SIZE(iatms)) = iatms
         ndim = ndim+SIZE(iatms)
      END DO
      n_oxygens_water = ndim

      CALL section_vals_val_get(section, "OXYGENS_ACID", n_rep_val=n_var)
      ndim = 0
      DO k = 1, n_var
         CALL section_vals_val_get(section, "OXYGENS_ACID", i_vals=iatms)
         CALL reallocate(i_oxygens_acid, 1, ndim+SIZE(iatms))
         i_oxygens_acid(ndim+1:ndim+SIZE(iatms)) = iatms
         ndim = ndim+SIZE(iatms)
      END DO
      n_oxygens_acid = ndim

      CALL section_vals_val_get(section, "HYDROGENS", n_rep_val=n_var)
      ndim = 0
      DO k = 1, n_var
         CALL section_vals_val_get(section, "HYDROGENS", i_vals=iatms)
         CALL reallocate(i_hydrogens, 1, ndim+SIZE(iatms))
         i_hydrogens(ndim+1:ndim+SIZE(iatms)) = iatms
         ndim = ndim+SIZE(iatms)
      END DO
      n_hydrogens = ndim

      SELECT CASE (colvar_id)
      CASE (acid_hyd_dist_colvar_id)
         CALL section_vals_val_get(section, "pWOH", i_val=colvar%acid_hyd_dist_param%pwoh)
         CALL section_vals_val_get(section, "qWOH", i_val=colvar%acid_hyd_dist_param%qwoh)
         CALL section_vals_val_get(section, "pAOH", i_val=colvar%acid_hyd_dist_param%paoh)
         CALL section_vals_val_get(section, "qAOH", i_val=colvar%acid_hyd_dist_param%qaoh)
         CALL section_vals_val_get(section, "pCUT", i_val=colvar%acid_hyd_dist_param%pcut)
         CALL section_vals_val_get(section, "qCUT", i_val=colvar%acid_hyd_dist_param%qcut)
         CALL section_vals_val_get(section, "RWOH", r_val=colvar%acid_hyd_dist_param%rwoh)
         CALL section_vals_val_get(section, "RAOH", r_val=colvar%acid_hyd_dist_param%raoh)
         CALL section_vals_val_get(section, "NC", r_val=colvar%acid_hyd_dist_param%nc)
         CALL section_vals_val_get(section, "LAMBDA", r_val=colvar%acid_hyd_dist_param%lambda)
      CASE (acid_hyd_shell_colvar_id)
         CALL section_vals_val_get(section, "pWOH", i_val=colvar%acid_hyd_shell_param%pwoh)
         CALL section_vals_val_get(section, "qWOH", i_val=colvar%acid_hyd_shell_param%qwoh)
         CALL section_vals_val_get(section, "pAOH", i_val=colvar%acid_hyd_shell_param%paoh)
         CALL section_vals_val_get(section, "qAOH", i_val=colvar%acid_hyd_shell_param%qaoh)
         CALL section_vals_val_get(section, "pOO", i_val=colvar%acid_hyd_shell_param%poo)
         CALL section_vals_val_get(section, "qOO", i_val=colvar%acid_hyd_shell_param%qoo)
         CALL section_vals_val_get(section, "pM", i_val=colvar%acid_hyd_shell_param%pm)
         CALL section_vals_val_get(section, "qM", i_val=colvar%acid_hyd_shell_param%qm)
         CALL section_vals_val_get(section, "pCUT", i_val=colvar%acid_hyd_shell_param%pcut)
         CALL section_vals_val_get(section, "qCUT", i_val=colvar%acid_hyd_shell_param%qcut)
         CALL section_vals_val_get(section, "RWOH", r_val=colvar%acid_hyd_shell_param%rwoh)
         CALL section_vals_val_get(section, "RAOH", r_val=colvar%acid_hyd_shell_param%raoh)
         CALL section_vals_val_get(section, "ROO", r_val=colvar%acid_hyd_shell_param%roo)
         CALL section_vals_val_get(section, "NC", r_val=colvar%acid_hyd_shell_param%nc)
         CALL section_vals_val_get(section, "NH", r_val=colvar%acid_hyd_shell_param%nh)
         CALL section_vals_val_get(section, "LAMBDA", r_val=colvar%acid_hyd_shell_param%lambda)
      END SELECT

   END SUBROUTINE read_acid_hydronium_colvars

! **************************************************************************************************
!> \brief Check and setup about the use of geometrical points instead of atoms
!> \param colvar the colvat to initialize
!> \param section ...
!> \author Teodoro Laino, [teo] 03.2007
! **************************************************************************************************
   SUBROUTINE colvar_check_points(colvar, section)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_check_points', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, irep, natoms, npoints, nrep, nweights
      INTEGER, DIMENSION(:), POINTER                     :: atoms
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r, weights
      TYPE(section_vals_type), POINTER                   :: point_sections

      NULLIFY (point_sections)
      NULLIFY (atoms)
      NULLIFY (weights)
      IF(.NOT.(ASSOCIATED(colvar)))CALL cp__a("colvar_methods.F",1502)
      point_sections => section_vals_get_subs_vals(section, "POINT")
      CALL section_vals_get(point_sections, explicit=explicit)
      IF (explicit) THEN
         colvar%use_points = .TRUE.
         CALL section_vals_get(point_sections, n_repetition=npoints)
         ALLOCATE (colvar%points(npoints))
         ! Read points definition
         DO i = 1, npoints
            natoms = 0
            nweights = 0
            NULLIFY (colvar%points(i)%atoms)
            NULLIFY (colvar%points(i)%weights)
            CALL section_vals_val_get(point_sections, "TYPE", i_rep_section=i, i_val=colvar%points(i)%type_id)
            SELECT CASE (colvar%points (i)%type_id)
            CASE (do_clv_geo_center)
               ! Define a point through a list of atoms..
               CALL section_vals_val_get(point_sections, "ATOMS", i_rep_section=i, n_rep_val=nrep, i_vals=atoms)
               DO irep = 1, nrep
                  CALL section_vals_val_get(point_sections, "ATOMS", i_rep_section=i, i_rep_val=irep, i_vals=atoms)
                  natoms = natoms+SIZE(atoms)
               END DO
               ALLOCATE (colvar%points(i)%atoms(natoms))
               natoms = 0
               DO irep = 1, nrep
                  CALL section_vals_val_get(point_sections, "ATOMS", i_rep_section=i, i_rep_val=irep, i_vals=atoms)
                  colvar%points(i)%atoms(natoms+1:) = atoms(:)
                  natoms = natoms+SIZE(atoms)
               END DO
               ! Define weights of the list
               ALLOCATE (colvar%points(i)%weights(natoms))
               colvar%points(i)%weights = 1.0_dp/REAL(natoms, KIND=dp)
               CALL section_vals_val_get(point_sections, "WEIGHTS", i_rep_section=i, n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  DO irep = 1, nrep
                     CALL section_vals_val_get(point_sections, "WEIGHTS", i_rep_section=i, i_rep_val=irep, &
                                               r_vals=weights)
                     colvar%points(i)%weights(nweights+1:) = weights(:)
                     nweights = nweights+SIZE(weights)
                  END DO
                  IF(.NOT.(natoms == nweights))CALL cp__a("colvar_methods.F",1542)
               END IF
            CASE (do_clv_fix_point)
               ! Define the point as a fixed point in space..
               CALL section_vals_val_get(point_sections, "XYZ", i_rep_section=i, r_vals=r)
               colvar%points(i)%r = r
            END SELECT
         END DO
      END IF
   END SUBROUTINE colvar_check_points

! **************************************************************************************************
!> \brief evaluates the derivatives (dsdr) given and due to the given colvar
!>      variables in a molecular environment
!> \param colvar the collective variable to evaluate
!> \param cell ...
!> \param particles ...
!> \param pos ...
!> \param fixd_list ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE colvar_eval_mol_f(colvar, cell, particles, pos, fixd_list)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(kind=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: pos
      TYPE(fixd_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: fixd_list

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_eval_mol_f', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      LOGICAL                                            :: colvar_ok

      colvar_ok = ASSOCIATED(colvar)
      IF(.NOT.(colvar_ok))CALL cp__a("colvar_methods.F",1579)

      IF (PRESENT(pos)) THEN
         DO i = 1, SIZE(colvar%i_atom)
            j = colvar%i_atom(i)
            particles(j)%r = pos(:, j)
         END DO
      END IF
      ! Initialize the content of the derivative
      colvar%dsdr = 0.0_dp
      SELECT CASE (colvar%type_id)
      CASE (dist_colvar_id)
         CALL dist_colvar(colvar, cell, particles=particles)
      CASE (coord_colvar_id)
         CALL coord_colvar(colvar, cell, particles=particles)
      CASE (population_colvar_id)
         CALL population_colvar(colvar, cell, particles=particles)
      CASE (gyration_colvar_id)
         CALL gyration_radius_colvar(colvar, cell, particles=particles)
      CASE (torsion_colvar_id)
         CALL torsion_colvar(colvar, cell, particles=particles)
      CASE (angle_colvar_id)
         CALL angle_colvar(colvar, cell, particles=particles)
      CASE (dfunct_colvar_id)
         CALL dfunct_colvar(colvar, cell, particles=particles)
      CASE (plane_distance_colvar_id)
         CALL plane_distance_colvar(colvar, cell, particles=particles)
      CASE (plane_plane_angle_colvar_id)
         CALL plane_plane_angle_colvar(colvar, cell, particles=particles)
      CASE (rotation_colvar_id)
         CALL rotation_colvar(colvar, cell, particles=particles)
      CASE (qparm_colvar_id)
         CALL qparm_colvar(colvar, cell, particles=particles)
      CASE (hydronium_colvar_id)
         CALL hydronium_colvar(colvar, cell, particles=particles)
      CASE (acid_hyd_dist_colvar_id)
         CALL acid_hyd_dist_colvar(colvar, cell, particles=particles)
      CASE (acid_hyd_shell_colvar_id)
         CALL acid_hyd_shell_colvar(colvar, cell, particles=particles)
      CASE (rmsd_colvar_id)
         CALL rmsd_colvar(colvar, particles=particles)
      CASE (reaction_path_colvar_id)
         CALL reaction_path_colvar(colvar, cell, particles=particles)
      CASE (distance_from_path_colvar_id)
         CALL distance_from_path_colvar(colvar, cell, particles=particles)
      CASE (combine_colvar_id)
         CALL combine_colvar(colvar, cell, particles=particles)
      CASE (xyz_diag_colvar_id)
         CALL xyz_diag_colvar(colvar, cell, particles=particles)
      CASE (xyz_outerdiag_colvar_id)
         CALL xyz_outerdiag_colvar(colvar, cell, particles=particles)
      CASE (ring_puckering_colvar_id)
         CALL ring_puckering_colvar(colvar, cell, particles=particles)
      CASE (mindist_colvar_id)
         CALL mindist_colvar(colvar, cell, particles=particles)
      CASE (u_colvar_id)
         CALL cp__b("colvar_methods.F",1635,"need force_env!")
      CASE (Wc_colvar_id)
         !!! FIXME this is rubbish at the moment as we have no force to be computed on this
         CALL Wc_colvar(colvar, cell, particles=particles)
      CASE (HBP_colvar_id)
         !!! FIXME this is rubbish at the moment as we have no force to be computed on this
         CALL HBP_colvar(colvar, cell, particles=particles)
      CASE DEFAULT
         CALL cp__b("colvar_methods.F",1643,"")
      END SELECT
      ! Check for fixed atom constraints
      IF (PRESENT(fixd_list)) CALL check_fixed_atom_cns_colv(fixd_list, colvar)

   END SUBROUTINE colvar_eval_mol_f

! **************************************************************************************************
!> \brief evaluates the derivatives (dsdr) given and due to the given colvar
!> \param icolvar the collective variable to evaluate
!> \param force_env ...
!> \author Alessandro Laio and fawzi
!> \note
!>      The torsion that generally is defined without the continuity problem
!>      here (for free energy calculations) is defined only for (-pi,pi]
! **************************************************************************************************
   SUBROUTINE colvar_eval_glob_f(icolvar, force_env)
      INTEGER                                            :: icolvar
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_eval_glob_f', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: colvar_ok
      TYPE(cell_type), POINTER                           :: cell
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(qs_environment_type), POINTER                 :: qs_env

      NULLIFY (subsys, cell, colvar, qs_env)
      CALL force_env_get(force_env, subsys=subsys, cell=cell, qs_env=qs_env)
      colvar_ok = ASSOCIATED(subsys%colvar_p)
      IF(.NOT.(colvar_ok))CALL cp__a("colvar_methods.F",1675)

      colvar => subsys%colvar_p(icolvar)%colvar
      ! Initialize the content of the derivative
      colvar%dsdr = 0.0_dp
      SELECT CASE (colvar%type_id)
      CASE (dist_colvar_id)
         CALL dist_colvar(colvar, cell, subsys=subsys)
      CASE (coord_colvar_id)
         CALL coord_colvar(colvar, cell, subsys=subsys)
      CASE (population_colvar_id)
         CALL population_colvar(colvar, cell, subsys=subsys)
      CASE (gyration_colvar_id)
         CALL gyration_radius_colvar(colvar, cell, subsys=subsys)
      CASE (torsion_colvar_id)
         CALL torsion_colvar(colvar, cell, subsys=subsys, no_riemann_sheet_op=.TRUE.)
      CASE (angle_colvar_id)
         CALL angle_colvar(colvar, cell, subsys=subsys)
      CASE (dfunct_colvar_id)
         CALL dfunct_colvar(colvar, cell, subsys=subsys)
      CASE (plane_distance_colvar_id)
         CALL plane_distance_colvar(colvar, cell, subsys=subsys)
      CASE (plane_plane_angle_colvar_id)
         CALL plane_plane_angle_colvar(colvar, cell, subsys=subsys)
      CASE (rotation_colvar_id)
         CALL rotation_colvar(colvar, cell, subsys=subsys)
      CASE (qparm_colvar_id)
         CALL qparm_colvar(colvar, cell, subsys=subsys)
      CASE (hydronium_colvar_id)
         CALL hydronium_colvar(colvar, cell, subsys=subsys)
      CASE (acid_hyd_dist_colvar_id)
         CALL acid_hyd_dist_colvar(colvar, cell, subsys=subsys)
      CASE (acid_hyd_shell_colvar_id)
         CALL acid_hyd_shell_colvar(colvar, cell, subsys=subsys)
      CASE (rmsd_colvar_id)
         CALL rmsd_colvar(colvar, subsys=subsys)
      CASE (reaction_path_colvar_id)
         CALL reaction_path_colvar(colvar, cell, subsys=subsys)
      CASE (distance_from_path_colvar_id)
         CALL distance_from_path_colvar(colvar, cell, subsys=subsys)
      CASE (combine_colvar_id)
         CALL combine_colvar(colvar, cell, subsys=subsys)
      CASE (xyz_diag_colvar_id)
         CALL xyz_diag_colvar(colvar, cell, subsys=subsys)
      CASE (xyz_outerdiag_colvar_id)
         CALL xyz_outerdiag_colvar(colvar, cell, subsys=subsys)
      CASE (u_colvar_id)
         CALL u_colvar(colvar, force_env=force_env)
      CASE (Wc_colvar_id)
         CALL Wc_colvar(colvar, cell, subsys=subsys, qs_env=qs_env)
      CASE (HBP_colvar_id)
         CALL HBP_colvar(colvar, cell, subsys=subsys, qs_env=qs_env)
      CASE (ring_puckering_colvar_id)
         CALL ring_puckering_colvar(colvar, cell, subsys=subsys)
      CASE (mindist_colvar_id)
         CALL mindist_colvar(colvar, cell, subsys=subsys)
      CASE DEFAULT
         CALL cp__b("colvar_methods.F",1732,"")
      END SELECT
      ! Check for fixed atom constraints
      CALL check_fixed_atom_cns_colv(subsys%gci%fixd_list, colvar)
   END SUBROUTINE colvar_eval_glob_f

! **************************************************************************************************
!> \brief evaluates the derivatives (dsdr) given and due to the given colvar
!>        for the specification of a recursive colvar type
!> \param colvar the collective variable to evaluate
!> \param cell ...
!> \param particles ...
!> \author sfchiff
! **************************************************************************************************
   SUBROUTINE colvar_recursive_eval(colvar, cell, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_recursive_eval', &
         routineP = moduleN//':'//routineN

! Initialize the content of the derivative

      colvar%dsdr = 0.0_dp
      SELECT CASE (colvar%type_id)
      CASE (dist_colvar_id)
         CALL dist_colvar(colvar, cell, particles=particles)
      CASE (coord_colvar_id)
         CALL coord_colvar(colvar, cell, particles=particles)
      CASE (torsion_colvar_id)
         CALL torsion_colvar(colvar, cell, particles=particles)
      CASE (angle_colvar_id)
         CALL angle_colvar(colvar, cell, particles=particles)
      CASE (dfunct_colvar_id)
         CALL dfunct_colvar(colvar, cell, particles=particles)
      CASE (plane_distance_colvar_id)
         CALL plane_distance_colvar(colvar, cell, particles=particles)
      CASE (plane_plane_angle_colvar_id)
         CALL plane_plane_angle_colvar(colvar, cell, particles=particles)
      CASE (rotation_colvar_id)
         CALL rotation_colvar(colvar, cell, particles=particles)
      CASE (qparm_colvar_id)
         CALL qparm_colvar(colvar, cell, particles=particles)
      CASE (hydronium_colvar_id)
         CALL hydronium_colvar(colvar, cell, particles=particles)
      CASE (acid_hyd_dist_colvar_id)
         CALL acid_hyd_dist_colvar(colvar, cell, particles=particles)
      CASE (acid_hyd_shell_colvar_id)
         CALL acid_hyd_shell_colvar(colvar, cell, particles=particles)
      CASE (rmsd_colvar_id)
         CALL rmsd_colvar(colvar, particles=particles)
      CASE (reaction_path_colvar_id)
         CALL reaction_path_colvar(colvar, cell, particles=particles)
      CASE (distance_from_path_colvar_id)
         CALL distance_from_path_colvar(colvar, cell, particles=particles)
      CASE (combine_colvar_id)
         CALL combine_colvar(colvar, cell, particles=particles)
      CASE (xyz_diag_colvar_id)
         CALL xyz_diag_colvar(colvar, cell, particles=particles)
      CASE (xyz_outerdiag_colvar_id)
         CALL xyz_outerdiag_colvar(colvar, cell, particles=particles)
      CASE (ring_puckering_colvar_id)
         CALL ring_puckering_colvar(colvar, cell, particles=particles)
      CASE (mindist_colvar_id)
         CALL mindist_colvar(colvar, cell, particles=particles)
      CASE (u_colvar_id)
         CALL cp__b("colvar_methods.F",1799,"need force_env!")
      CASE (Wc_colvar_id)
         CALL Wc_colvar(colvar, cell, particles=particles)
      CASE (HBP_colvar_id)
         CALL HBP_colvar(colvar, cell, particles=particles)
      CASE DEFAULT
         CALL cp__b("colvar_methods.F",1805,"")
      END SELECT
   END SUBROUTINE colvar_recursive_eval

! **************************************************************************************************
!> \brief Get coordinates of atoms or of geometrical points
!> \param colvar ...
!> \param i ...
!> \param ri ...
!> \param my_particles ...
!> \author Teodoro Laino 03.2007 [created]
! **************************************************************************************************
   SUBROUTINE get_coordinates(colvar, i, ri, my_particles)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: ri
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF (colvar%use_points) THEN
         CALL eval_point_pos(colvar%points(i), my_particles, ri)
      ELSE
         ri(:) = my_particles(i)%r(:)
      END IF

   END SUBROUTINE get_coordinates

! **************************************************************************************************
!> \brief Get masses of atoms or of geometrical points
!> \param colvar ...
!> \param i ...
!> \param mi ...
!> \param my_particles ...
!> \author Teodoro Laino 03.2007 [created]
! **************************************************************************************************
   SUBROUTINE get_mass(colvar, i, mi, my_particles)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), INTENT(OUT)                         :: mi
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF (colvar%use_points) THEN
         CALL eval_point_mass(colvar%points(i), my_particles, mi)
      ELSE
         mi = my_particles(i)%atomic_kind%mass
      END IF

   END SUBROUTINE get_mass

! **************************************************************************************************
!> \brief Transfer derivatives to ds/dr
!> \param colvar ...
!> \param i ...
!> \param fi ...
!> \author Teodoro Laino 03.2007 [created]
! **************************************************************************************************
   SUBROUTINE put_derivative(colvar, i, fi)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: fi

      IF (colvar%use_points) THEN
         CALL eval_point_der(colvar%points, i, colvar%dsdr, fi)
      ELSE
         colvar%dsdr(:, i) = colvar%dsdr(:, i)+fi
      END IF

   END SUBROUTINE put_derivative

! **************************************************************************************************
!> \brief  evaluates the force due to the position colvar
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2010 [created]
! **************************************************************************************************
   SUBROUTINE xyz_diag_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'xyz_diag_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp)                                           :: fi(3), r, r0(3), ss(3), xi(3), xpi(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == xyz_diag_colvar_id))CALL cp__a("colvar_methods.F",1898)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",1902)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%xyz_diag_param%i_atom
      ! Atom coordinates
      CALL get_coordinates(colvar, i, xpi, my_particles)
      ! Use the current coordinates as initial coordinates, if no initialization
      ! was performed yet
      IF (.NOT. colvar%xyz_diag_param%use_absolute_position) THEN
         IF (ALL(colvar%xyz_diag_param%r0 == HUGE(0.0_dp))) THEN
            colvar%xyz_diag_param%r0 = xpi
         END IF
         r0 = colvar%xyz_diag_param%r0
      ELSE
         r0 = 0.0_dp
      ENDIF

      IF (colvar%xyz_diag_param%use_pbc) THEN
         ss = MATMUL(cell%h_inv, xpi-r0)
         ss = ss-NINT(ss)
         xi = MATMUL(cell%hmat, ss)
      ELSE
         xi = xpi-r0
      END IF

      IF (.NOT. colvar%xyz_diag_param%use_absolute_position) THEN
         SELECT CASE (colvar%xyz_diag_param%component)
         CASE (do_clv_x)
            xi(2) = 0.0_dp
            xi(3) = 0.0_dp
         CASE (do_clv_y)
            xi(1) = 0.0_dp
            xi(3) = 0.0_dp
         CASE (do_clv_z)
            xi(1) = 0.0_dp
            xi(2) = 0.0_dp
         CASE (do_clv_xy)
            xi(3) = 0.0_dp
         CASE (do_clv_xz)
            xi(2) = 0.0_dp
         CASE (do_clv_yz)
            xi(1) = 0.0_dp
         CASE DEFAULT
            ! do_clv_xyz
         END SELECT

         r = xi(1)**2+xi(2)**2+xi(3)**2
         fi(:) = 2.0_dp*xi
      ELSE
         SELECT CASE (colvar%xyz_diag_param%component)
         CASE (do_clv_x)
            r = xi(1)
            xi(1) = 1.0_dp
            xi(2) = 0.0_dp
            xi(3) = 0.0_dp
         CASE (do_clv_y)
            r = xi(2)
            xi(1) = 0.0_dp
            xi(2) = 1.0_dp
            xi(3) = 0.0_dp
         CASE (do_clv_z)
            r = xi(3)
            xi(1) = 0.0_dp
            xi(2) = 0.0_dp
            xi(3) = 1.0_dp
         CASE DEFAULT
            !Not implemented for anything which is not a single component.
            CALL cp__b("colvar_methods.F",1970,"")
         END SELECT
         fi(:) = xi
      ENDIF

      colvar%ss = r
      CALL put_derivative(colvar, 1, fi)

   END SUBROUTINE xyz_diag_colvar

! **************************************************************************************************
!> \brief  evaluates the force due to the position colvar
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2010 [created]
! **************************************************************************************************
   SUBROUTINE xyz_outerdiag_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'xyz_outerdiag_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, l
      REAL(dp)                                           :: fi(3, 2), r, r0(3), ss(3), xi(3, 2), &
                                                            xpi(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == xyz_outerdiag_colvar_id))CALL cp__a("colvar_methods.F",2006)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2010)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      DO k = 1, 2
         i = colvar%xyz_outerdiag_param%i_atoms(k)
         ! Atom coordinates
         CALL get_coordinates(colvar, i, xpi, my_particles)
         r0 = colvar%xyz_outerdiag_param%r0(:, k)
         IF (ALL(colvar%xyz_outerdiag_param%r0(:, k) == HUGE(0.0_dp))) r0 = xpi

         IF (colvar%xyz_outerdiag_param%use_pbc) THEN
            ss = MATMUL(cell%h_inv, xpi-r0)
            ss = ss-NINT(ss)
            xi(:, k) = MATMUL(cell%hmat, ss)
         ELSE
            xi(:, k) = xpi-r0
         END IF

         SELECT CASE (colvar%xyz_outerdiag_param%components (k))
         CASE (do_clv_x)
            xi(2, k) = 0.0_dp
            xi(3, k) = 0.0_dp
         CASE (do_clv_y)
            xi(1, k) = 0.0_dp
            xi(3, k) = 0.0_dp
         CASE (do_clv_z)
            xi(1, k) = 0.0_dp
            xi(2, k) = 0.0_dp
         CASE (do_clv_xy)
            xi(3, k) = 0.0_dp
         CASE (do_clv_xz)
            xi(2, k) = 0.0_dp
         CASE (do_clv_yz)
            xi(1, k) = 0.0_dp
         CASE DEFAULT
            ! do_clv_xyz
         END SELECT
      END DO

      r = 0.0_dp
      fi = 0.0_dp
      DO i = 1, 3
         DO l = 1, 3
            IF (xi(l, 1) /= 0.0_dp) fi(l, 1) = fi(l, 1)+xi(i, 2)
            r = r+xi(l, 1)*xi(i, 2)
         END DO
         IF (xi(i, 2) /= 0.0_dp) fi(i, 2) = SUM(xi(:, 1))
      END DO

      colvar%ss = r
      CALL put_derivative(colvar, 1, fi(:, 1))
      CALL put_derivative(colvar, 2, fi(:, 2))

   END SUBROUTINE xyz_outerdiag_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the energy as collective variable
!> \param colvar ...
!> \param force_env ...
!> \par History Modified to allow functions of energy in a mixed_env environment
!>              Teodoro Laino [tlaino] - 02.2011
!> \author Sebastiano Caravati
! **************************************************************************************************
   SUBROUTINE u_colvar(colvar, force_env)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'u_colvar', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length)               :: def_error, this_error
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: parameters
      INTEGER                                            :: iatom, iforce_eval, iparticle, &
                                                            jparticle, natom, natom_iforce, &
                                                            nforce_eval
      INTEGER, DIMENSION(:), POINTER                     :: glob_natoms, map_index
      REAL(dp)                                           :: dedf, dx, err, fi(3), lerr, &
                                                            potential_energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: values
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: subsystems
      TYPE(cp_subsys_type), POINTER                      :: subsys_main
      TYPE(mixed_force_type), DIMENSION(:), POINTER      :: global_forces
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles
      TYPE(particle_list_type), POINTER                  :: particles_main
      TYPE(section_vals_type), POINTER                   :: force_env_section, mapping_section, &
                                                            wrk_section

      IF (PRESENT(force_env)) THEN
         NULLIFY (particles_main, subsys_main)
         CALL force_env_get(force_env=force_env, subsys=subsys_main)
         CALL cp_subsys_get(subsys=subsys_main, particles=particles_main)
         natom = SIZE(particles_main%els)
         colvar%n_atom_s = natom
         colvar%u_param%natom = natom
         CALL reallocate(colvar%i_atom, 1, natom)
         CALL reallocate(colvar%dsdr, 1, 3, 1, natom)
         DO iatom = 1, natom
            colvar%i_atom(iatom) = iatom
         ENDDO

         IF (.NOT. ASSOCIATED(colvar%u_param%mixed_energy_section)) THEN
            CALL force_env_get(force_env, potential_energy=potential_energy)
            colvar%ss = potential_energy

            DO iatom = 1, natom
               ! store derivative
               fi(:) = -particles_main%els(iatom)%f
               CALL put_derivative(colvar, iatom, fi)
            ENDDO
         ELSE
            IF (force_env%in_use /= use_mixed_force) &
               CALL cp_abort(cp__l("colvar_methods.F",2123), &
                             'ASSERTION (cond) failed at line '//cp_to_string(2124)// &
                             ' A combination of mixed force_eval energies has been requested as '// &
                             ' collective variable, but the MIXED env is not in use! Aborting.')
            CALL force_env_get(force_env, force_env_section=force_env_section)
            mapping_section => section_vals_get_subs_vals(force_env_section, "MIXED%MAPPING")
            NULLIFY (values, parameters, subsystems, particles, global_forces, map_index, glob_natoms)
            nforce_eval = SIZE(force_env%sub_force_env)
            ALLOCATE (glob_natoms(nforce_eval))
            ALLOCATE (subsystems(nforce_eval))
            ALLOCATE (particles(nforce_eval))
            ! Local Info to sync
            ALLOCATE (global_forces(nforce_eval))

            glob_natoms = 0
            DO iforce_eval = 1, nforce_eval
               NULLIFY (subsystems(iforce_eval)%subsys, particles(iforce_eval)%list)
               IF (.NOT. ASSOCIATED(force_env%sub_force_env(iforce_eval)%force_env)) CYCLE
               ! Get all available subsys
               CALL force_env_get(force_env=force_env%sub_force_env(iforce_eval)%force_env, &
                                  subsys=subsystems(iforce_eval)%subsys)
               ! Get available particles
               CALL cp_subsys_get(subsys=subsystems(iforce_eval)%subsys, &
                                  particles=particles(iforce_eval)%list)

               ! Get Mapping index array
               natom_iforce = SIZE(particles(iforce_eval)%list%els)

               ! Only the rank 0 process collect info for each computation
               IF (force_env%sub_force_env(iforce_eval)%force_env%para_env%mepos == &
                   force_env%sub_force_env(iforce_eval)%force_env%para_env%source) THEN
                  glob_natoms(iforce_eval) = natom_iforce
               END IF
            END DO

            ! Handling Parallel execution
            CALL mp_sync(force_env%para_env%group)
            CALL mp_sum(glob_natoms, force_env%para_env%group)

            ! Transfer forces
            DO iforce_eval = 1, nforce_eval
               ALLOCATE (global_forces(iforce_eval)%forces(3, glob_natoms(iforce_eval)))
               global_forces(iforce_eval)%forces = 0.0_dp
               IF (ASSOCIATED(force_env%sub_force_env(iforce_eval)%force_env)) THEN
                  IF (force_env%sub_force_env(iforce_eval)%force_env%para_env%mepos == &
                      force_env%sub_force_env(iforce_eval)%force_env%para_env%source) THEN
                     ! Forces
                     DO iparticle = 1, glob_natoms(iforce_eval)
                        global_forces(iforce_eval)%forces(:, iparticle) = &
                           particles(iforce_eval)%list%els(iparticle)%f
                     END DO
                  END IF
               END IF
               CALL mp_sum(global_forces(iforce_eval)%forces, force_env%para_env%group)
            END DO

            wrk_section => colvar%u_param%mixed_energy_section
            ! Support any number of force_eval sections
            CALL get_generic_info(wrk_section, "ENERGY_FUNCTION", coupling_function, parameters, &
                                  values, force_env%mixed_env%energies)
            CALL initf(1)
            CALL parsef(1, TRIM(coupling_function), parameters)
            ! Store the value of the COLVAR
            colvar%ss = evalf(1, values)
            IF(.NOT.(EvalErrType <= 0))CALL cp__a("colvar_methods.F",2187)

            DO iforce_eval = 1, nforce_eval
               CALL section_vals_val_get(wrk_section, "DX", r_val=dx)
               CALL section_vals_val_get(wrk_section, "ERROR_LIMIT", r_val=lerr)
               dedf = evalfd(1, iforce_eval, values, dx, err)
               IF (ABS(err) > lerr) THEN
                  WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
                  WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
                  CALL compress(this_error, .TRUE.)
                  CALL compress(def_error, .TRUE.)
                  CALL cp_warn(cp__l("colvar_methods.F",2198), &
                               'ASSERTION (cond) failed at line '//cp_to_string(2199)// &
                               ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                               TRIM(def_error)//' .')
               END IF
               ! General Mapping of forces...
               ! First: Get Mapping index array
               CALL get_subsys_map_index(mapping_section, glob_natoms(iforce_eval), iforce_eval, &
                                         nforce_eval, map_index)

               ! Second: store derivatives
               DO iparticle = 1, glob_natoms(iforce_eval)
                  jparticle = map_index(iparticle)
                  fi = -dedf*global_forces(iforce_eval)%forces(:, iparticle)
                  CALL put_derivative(colvar, jparticle, fi)
               END DO
               ! Deallocate map_index array
               IF (ASSOCIATED(map_index)) THEN
                  DEALLOCATE (map_index)
               END IF
            END DO
            CALL finalizef()
            DO iforce_eval = 1, nforce_eval
               DEALLOCATE (global_forces(iforce_eval)%forces)
            END DO
            DEALLOCATE (glob_natoms)
            DEALLOCATE (values)
            DEALLOCATE (parameters)
            DEALLOCATE (global_forces)
            DEALLOCATE (subsystems)
            DEALLOCATE (particles)
         END IF
      ELSE
         CALL cp__b("colvar_methods.F",2231,"need force_env!")
      ENDIF
   END SUBROUTINE u_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the distance from the plane collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2006 [created]
! **************************************************************************************************
   SUBROUTINE plane_distance_colvar(colvar, cell, subsys, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'plane_distance_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l
      REAL(dp) :: a, b, dsdxpn(3), dxpndxi(3, 3), dxpndxj(3, 3), dxpndxk(3, 3), fi(3), fj(3), &
         fk(3), fl(3), r12, ri(3), rj(3), rk(3), rl(3), ss(3), xpij(3), xpkj(3), xpl(3), xpn(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == plane_distance_colvar_id))CALL cp__a("colvar_methods.F",2262)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2266)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%plane_distance_param%plane(1)
      j = colvar%plane_distance_param%plane(2)
      k = colvar%plane_distance_param%plane(3)
      l = colvar%plane_distance_param%point
      ! Get coordinates of atoms or points
      CALL get_coordinates(colvar, i, ri, my_particles)
      CALL get_coordinates(colvar, j, rj, my_particles)
      CALL get_coordinates(colvar, k, rk, my_particles)
      CALL get_coordinates(colvar, l, rl, my_particles)
      xpij = ri-rj
      xpkj = rk-rj
      xpl = rl-(ri+rj+rk)/3.0_dp
      IF (colvar%plane_distance_param%use_pbc) THEN
         ! xpij
         ss = MATMUL(cell%h_inv, ri-rj)
         ss = ss-NINT(ss)
         xpij = MATMUL(cell%hmat, ss)
         ! xpkj
         ss = MATMUL(cell%h_inv, rk-rj)
         ss = ss-NINT(ss)
         xpkj = MATMUL(cell%hmat, ss)
         ! xpl
         ss = MATMUL(cell%h_inv, rl-(ri+rj+rk)/3.0_dp)
         ss = ss-NINT(ss)
         xpl = MATMUL(cell%hmat, ss)
      END IF
      ! xpn
      xpn(1) = xpij(2)*xpkj(3)-xpij(3)*xpkj(2)
      xpn(2) = xpij(3)*xpkj(1)-xpij(1)*xpkj(3)
      xpn(3) = xpij(1)*xpkj(2)-xpij(2)*xpkj(1)
      a = DOT_PRODUCT(xpn, xpn)
      b = DOT_PRODUCT(xpl, xpn)
      r12 = SQRT(a)
      colvar%ss = b/r12
      dsdxpn(1) = xpl(1)/r12-b*xpn(1)/(r12*a)
      dsdxpn(2) = xpl(2)/r12-b*xpn(2)/(r12*a)
      dsdxpn(3) = xpl(3)/r12-b*xpn(3)/(r12*a)
      !
      dxpndxi(1, 1) = 0.0_dp
      dxpndxi(1, 2) = 1.0_dp*xpkj(3)
      dxpndxi(1, 3) = -1.0_dp*xpkj(2)
      dxpndxi(2, 1) = -1.0_dp*xpkj(3)
      dxpndxi(2, 2) = 0.0_dp
      dxpndxi(2, 3) = 1.0_dp*xpkj(1)
      dxpndxi(3, 1) = 1.0_dp*xpkj(2)
      dxpndxi(3, 2) = -1.0_dp*xpkj(1)
      dxpndxi(3, 3) = 0.0_dp
      !
      dxpndxj(1, 1) = 0.0_dp
      dxpndxj(1, 2) = -1.0_dp*xpkj(3)+xpij(3)
      dxpndxj(1, 3) = -1.0_dp*xpij(2)+xpkj(2)
      dxpndxj(2, 1) = -1.0_dp*xpij(3)+xpkj(3)
      dxpndxj(2, 2) = 0.0_dp
      dxpndxj(2, 3) = -1.0_dp*xpkj(1)+xpij(1)
      dxpndxj(3, 1) = -1.0_dp*xpkj(2)+xpij(2)
      dxpndxj(3, 2) = -1.0_dp*xpij(1)+xpkj(1)
      dxpndxj(3, 3) = 0.0_dp
      !
      dxpndxk(1, 1) = 0.0_dp
      dxpndxk(1, 2) = -1.0_dp*xpij(3)
      dxpndxk(1, 3) = 1.0_dp*xpij(2)
      dxpndxk(2, 1) = 1.0_dp*xpij(3)
      dxpndxk(2, 2) = 0.0_dp
      dxpndxk(2, 3) = -1.0_dp*xpij(1)
      dxpndxk(3, 1) = -1.0_dp*xpij(2)
      dxpndxk(3, 2) = 1.0_dp*xpij(1)
      dxpndxk(3, 3) = 0.0_dp
      !
      fi(:) = MATMUL(dsdxpn, dxpndxi)-xpn/(3.0_dp*r12)
      fj(:) = MATMUL(dsdxpn, dxpndxj)-xpn/(3.0_dp*r12)
      fk(:) = MATMUL(dsdxpn, dxpndxk)-xpn/(3.0_dp*r12)
      fl(:) = xpn/r12
      ! Transfer derivatives on atoms
      CALL put_derivative(colvar, 1, fi)
      CALL put_derivative(colvar, 2, fj)
      CALL put_derivative(colvar, 3, fk)
      CALL put_derivative(colvar, 4, fl)

   END SUBROUTINE plane_distance_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the angle between two planes.
!>        plane-plane angle collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2009 [created]
! **************************************************************************************************
   SUBROUTINE plane_plane_angle_colvar(colvar, cell, subsys, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'plane_plane_angle_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, j1, j2, k1, k2, np
      LOGICAL                                            :: check
      REAL(dp) :: a1, a2, d, dnorm_dxpn(3), dprod12_dxpn(3), dsdxpn(3), dt_dxpn(3), dxpndxi(3, 3), &
         dxpndxj(3, 3), dxpndxk(3, 3), fi(3), fj(3), fk(3), fmod, norm1, norm2, prod_12, ri1(3), &
         ri2(3), rj1(3), rj2(3), rk1(3), rk2(3), ss(3), t, xpij1(3), xpij2(3), xpkj1(3), xpkj2(3), &
         xpn1(3), xpn2(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      check = colvar%type_id == plane_plane_angle_colvar_id
      IF(.NOT.(check))CALL cp__a("colvar_methods.F",2382)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2386)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      ! Plane 1
      IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) THEN
         i1 = colvar%plane_plane_angle_param%plane1%points(1)
         j1 = colvar%plane_plane_angle_param%plane1%points(2)
         k1 = colvar%plane_plane_angle_param%plane1%points(3)

         ! Get coordinates of atoms or points
         CALL get_coordinates(colvar, i1, ri1, my_particles)
         CALL get_coordinates(colvar, j1, rj1, my_particles)
         CALL get_coordinates(colvar, k1, rk1, my_particles)

         ! xpij
         ss = MATMUL(cell%h_inv, ri1-rj1)
         ss = ss-NINT(ss)
         xpij1 = MATMUL(cell%hmat, ss)

         ! xpkj
         ss = MATMUL(cell%h_inv, rk1-rj1)
         ss = ss-NINT(ss)
         xpkj1 = MATMUL(cell%hmat, ss)

         ! xpn
         xpn1(1) = xpij1(2)*xpkj1(3)-xpij1(3)*xpkj1(2)
         xpn1(2) = xpij1(3)*xpkj1(1)-xpij1(1)*xpkj1(3)
         xpn1(3) = xpij1(1)*xpkj1(2)-xpij1(2)*xpkj1(1)
      ELSE
         xpn1 = colvar%plane_plane_angle_param%plane1%normal_vec
      END IF
      a1 = DOT_PRODUCT(xpn1, xpn1)
      norm1 = SQRT(a1)
      IF(.NOT.(norm1 /= 0.0_dp))CALL cp__a("colvar_methods.F",2421)

      ! Plane 2
      IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) THEN
         i2 = colvar%plane_plane_angle_param%plane2%points(1)
         j2 = colvar%plane_plane_angle_param%plane2%points(2)
         k2 = colvar%plane_plane_angle_param%plane2%points(3)

         ! Get coordinates of atoms or points
         CALL get_coordinates(colvar, i2, ri2, my_particles)
         CALL get_coordinates(colvar, j2, rj2, my_particles)
         CALL get_coordinates(colvar, k2, rk2, my_particles)

         ! xpij
         ss = MATMUL(cell%h_inv, ri2-rj2)
         ss = ss-NINT(ss)
         xpij2 = MATMUL(cell%hmat, ss)

         ! xpkj
         ss = MATMUL(cell%h_inv, rk2-rj2)
         ss = ss-NINT(ss)
         xpkj2 = MATMUL(cell%hmat, ss)

         ! xpn
         xpn2(1) = xpij2(2)*xpkj2(3)-xpij2(3)*xpkj2(2)
         xpn2(2) = xpij2(3)*xpkj2(1)-xpij2(1)*xpkj2(3)
         xpn2(3) = xpij2(1)*xpkj2(2)-xpij2(2)*xpkj2(1)
      ELSE
         xpn2 = colvar%plane_plane_angle_param%plane2%normal_vec
      END IF
      a2 = DOT_PRODUCT(xpn2, xpn2)
      norm2 = SQRT(a2)
      IF(.NOT.(norm2 /= 0.0_dp))CALL cp__a("colvar_methods.F",2453)

      ! The value of the angle is defined only between 0 and Pi
      prod_12 = DOT_PRODUCT(xpn1, xpn2)

      d = norm1*norm2
      t = prod_12/d
      t = MIN(1.0_dp, ABS(t))*SIGN(1.0_dp, t)
      colvar%ss = ACOS(t)

      IF ((ABS(colvar%ss) .LT. tolerance_acos) .OR. (ABS(colvar%ss-pi) .LT. tolerance_acos)) THEN
         fmod = 0.0_dp
      ELSE
         fmod = -1.0_dp/SIN(colvar%ss)
      ENDIF
      ! Compute derivatives
      np = 0
      ! Plane 1
      IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) THEN
         dprod12_dxpn = xpn2
         dnorm_dxpn = 1.0_dp/norm1*xpn1
         dt_dxpn = (dprod12_dxpn*d-prod_12*dnorm_dxpn*norm2)/d**2

         dsdxpn(1) = fmod*dt_dxpn(1)
         dsdxpn(2) = fmod*dt_dxpn(2)
         dsdxpn(3) = fmod*dt_dxpn(3)
         !
         dxpndxi(1, 1) = 0.0_dp
         dxpndxi(1, 2) = 1.0_dp*xpkj1(3)
         dxpndxi(1, 3) = -1.0_dp*xpkj1(2)
         dxpndxi(2, 1) = -1.0_dp*xpkj1(3)
         dxpndxi(2, 2) = 0.0_dp
         dxpndxi(2, 3) = 1.0_dp*xpkj1(1)
         dxpndxi(3, 1) = 1.0_dp*xpkj1(2)
         dxpndxi(3, 2) = -1.0_dp*xpkj1(1)
         dxpndxi(3, 3) = 0.0_dp
         !
         dxpndxj(1, 1) = 0.0_dp
         dxpndxj(1, 2) = -1.0_dp*xpkj1(3)+xpij1(3)
         dxpndxj(1, 3) = -1.0_dp*xpij1(2)+xpkj1(2)
         dxpndxj(2, 1) = -1.0_dp*xpij1(3)+xpkj1(3)
         dxpndxj(2, 2) = 0.0_dp
         dxpndxj(2, 3) = -1.0_dp*xpkj1(1)+xpij1(1)
         dxpndxj(3, 1) = -1.0_dp*xpkj1(2)+xpij1(2)
         dxpndxj(3, 2) = -1.0_dp*xpij1(1)+xpkj1(1)
         dxpndxj(3, 3) = 0.0_dp
         !
         dxpndxk(1, 1) = 0.0_dp
         dxpndxk(1, 2) = -1.0_dp*xpij1(3)
         dxpndxk(1, 3) = 1.0_dp*xpij1(2)
         dxpndxk(2, 1) = 1.0_dp*xpij1(3)
         dxpndxk(2, 2) = 0.0_dp
         dxpndxk(2, 3) = -1.0_dp*xpij1(1)
         dxpndxk(3, 1) = -1.0_dp*xpij1(2)
         dxpndxk(3, 2) = 1.0_dp*xpij1(1)
         dxpndxk(3, 3) = 0.0_dp
         !
         fi = MATMUL(dsdxpn, dxpndxi)
         fj = MATMUL(dsdxpn, dxpndxj)
         fk = MATMUL(dsdxpn, dxpndxk)

         ! Transfer derivatives on atoms
         CALL put_derivative(colvar, np+1, fi)
         CALL put_derivative(colvar, np+2, fj)
         CALL put_derivative(colvar, np+3, fk)
         np = 3
      END IF

      ! Plane 2
      IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) THEN
         dprod12_dxpn = xpn1
         dnorm_dxpn = 1.0_dp/norm2*xpn2
         dt_dxpn = (dprod12_dxpn*d-prod_12*dnorm_dxpn*norm1)/d**2

         dsdxpn(1) = fmod*dt_dxpn(1)
         dsdxpn(2) = fmod*dt_dxpn(2)
         dsdxpn(3) = fmod*dt_dxpn(3)
         !
         dxpndxi(1, 1) = 0.0_dp
         dxpndxi(1, 2) = 1.0_dp*xpkj1(3)
         dxpndxi(1, 3) = -1.0_dp*xpkj1(2)
         dxpndxi(2, 1) = -1.0_dp*xpkj1(3)
         dxpndxi(2, 2) = 0.0_dp
         dxpndxi(2, 3) = 1.0_dp*xpkj1(1)
         dxpndxi(3, 1) = 1.0_dp*xpkj1(2)
         dxpndxi(3, 2) = -1.0_dp*xpkj1(1)
         dxpndxi(3, 3) = 0.0_dp
         !
         dxpndxj(1, 1) = 0.0_dp
         dxpndxj(1, 2) = -1.0_dp*xpkj1(3)+xpij1(3)
         dxpndxj(1, 3) = -1.0_dp*xpij1(2)+xpkj1(2)
         dxpndxj(2, 1) = -1.0_dp*xpij1(3)+xpkj1(3)
         dxpndxj(2, 2) = 0.0_dp
         dxpndxj(2, 3) = -1.0_dp*xpkj1(1)+xpij1(1)
         dxpndxj(3, 1) = -1.0_dp*xpkj1(2)+xpij1(2)
         dxpndxj(3, 2) = -1.0_dp*xpij1(1)+xpkj1(1)
         dxpndxj(3, 3) = 0.0_dp
         !
         dxpndxk(1, 1) = 0.0_dp
         dxpndxk(1, 2) = -1.0_dp*xpij1(3)
         dxpndxk(1, 3) = 1.0_dp*xpij1(2)
         dxpndxk(2, 1) = 1.0_dp*xpij1(3)
         dxpndxk(2, 2) = 0.0_dp
         dxpndxk(2, 3) = -1.0_dp*xpij1(1)
         dxpndxk(3, 1) = -1.0_dp*xpij1(2)
         dxpndxk(3, 2) = 1.0_dp*xpij1(1)
         dxpndxk(3, 3) = 0.0_dp
         !
         fi = MATMUL(dsdxpn, dxpndxi)
         fj = MATMUL(dsdxpn, dxpndxj)
         fk = MATMUL(dsdxpn, dxpndxk)

         ! Transfer derivatives on atoms
         CALL put_derivative(colvar, np+1, fi)
         CALL put_derivative(colvar, np+2, fj)
         CALL put_derivative(colvar, np+3, fk)
      END IF

   END SUBROUTINE plane_plane_angle_colvar

! **************************************************************************************************
!> \brief Evaluates the value of the rotation angle between two bonds
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2006 [created]
! **************************************************************************************************
   SUBROUTINE rotation_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rotation_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idum
      REAL(dp)                                           :: a, b, fmod, t0, t1, t2, t3, xdum(3), &
                                                            xij(3), xkj(3)
      REAL(KIND=dp)                                      :: dp1b1(3), dp1b2(3), dp2b1(3), dp2b2(3), &
                                                            ss(3), xp1b1(3), xp1b2(3), xp2b1(3), &
                                                            xp2b2(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == rotation_colvar_id))CALL cp__a("colvar_methods.F",2602)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2606)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%rotation_param%i_at1_bond1
      CALL get_coordinates(colvar, i, xp1b1, my_particles)
      i = colvar%rotation_param%i_at2_bond1
      CALL get_coordinates(colvar, i, xp2b1, my_particles)
      i = colvar%rotation_param%i_at1_bond2
      CALL get_coordinates(colvar, i, xp1b2, my_particles)
      i = colvar%rotation_param%i_at2_bond2
      CALL get_coordinates(colvar, i, xp2b2, my_particles)
      ! xij
      ss = MATMUL(cell%h_inv, xp1b1-xp2b1)
      ss = ss-NINT(ss)
      xij = MATMUL(cell%hmat, ss)
      ! xkj
      ss = MATMUL(cell%h_inv, xp1b2-xp2b2)
      ss = ss-NINT(ss)
      xkj = MATMUL(cell%hmat, ss)
      ! evaluation of the angle..
      a = SQRT(DOT_PRODUCT(xij, xij))
      b = SQRT(DOT_PRODUCT(xkj, xkj))
      t0 = 1.0_dp/(a*b)
      t1 = 1.0_dp/(a**3.0_dp*b)
      t2 = 1.0_dp/(a*b**3.0_dp)
      t3 = DOT_PRODUCT(xij, xkj)
      colvar%ss = ACOS(t3*t0)
      IF ((ABS(colvar%ss) .LT. tolerance_acos) .OR. (ABS(colvar%ss-pi) .LT. tolerance_acos)) THEN
         fmod = 0.0_dp
      ELSE
         fmod = -1.0_dp/SIN(colvar%ss)
      ENDIF
      dp1b1 = xkj(:)*t0-xij(:)*t1*t3
      dp2b1 = -xkj(:)*t0+xij(:)*t1*t3
      dp1b2 = xij(:)*t0-xkj(:)*t2*t3
      dp2b2 = -xij(:)*t0+xkj(:)*t2*t3

      xdum = dp1b1*fmod
      idum = colvar%rotation_param%i_at1_bond1
      CALL put_derivative(colvar, idum, xdum)
      xdum = dp2b1*fmod
      idum = colvar%rotation_param%i_at2_bond1
      CALL put_derivative(colvar, idum, xdum)
      xdum = dp1b2*fmod
      idum = colvar%rotation_param%i_at1_bond2
      CALL put_derivative(colvar, idum, xdum)
      xdum = dp2b2*fmod
      idum = colvar%rotation_param%i_at2_bond2
      CALL put_derivative(colvar, idum, xdum)

   END SUBROUTINE rotation_colvar

! **************************************************************************************************
!> \brief evaluates the force due to the function of two distances
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2006 [created]
!> \note modified Florian Schiffmann 08.2008
! **************************************************************************************************
   SUBROUTINE dfunct_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'dfunct_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l
      REAL(dp)                                           :: fi(3), fj(3), fk(3), fl(3), r12, r34, &
                                                            ss(3), xij(3), xkl(3), xpi(3), xpj(3), &
                                                            xpk(3), xpl(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == dfunct_colvar_id))CALL cp__a("colvar_methods.F",2686)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2690)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%dfunct_param%i_at_dfunct(1)
      j = colvar%dfunct_param%i_at_dfunct(2)
      ! First bond
      CALL get_coordinates(colvar, i, xpi, my_particles)
      CALL get_coordinates(colvar, j, xpj, my_particles)
      IF (colvar%dfunct_param%use_pbc) THEN
         ss = MATMUL(cell%h_inv, xpi-xpj)
         ss = ss-NINT(ss)
         xij = MATMUL(cell%hmat, ss)
      ELSE
         xij = xpi-xpj
      END IF
      r12 = SQRT(xij(1)**2+xij(2)**2+xij(3)**2)
      ! Second bond
      k = colvar%dfunct_param%i_at_dfunct(3)
      l = colvar%dfunct_param%i_at_dfunct(4)
      CALL get_coordinates(colvar, k, xpk, my_particles)
      CALL get_coordinates(colvar, l, xpl, my_particles)
      IF (colvar%dfunct_param%use_pbc) THEN
         ss = MATMUL(cell%h_inv, xpk-xpl)
         ss = ss-NINT(ss)
         xkl = MATMUL(cell%hmat, ss)
      ELSE
         xkl = xpk-xpl
      END IF
      r34 = SQRT(xkl(1)**2+xkl(2)**2+xkl(3)**2)
      !
      colvar%ss = r12+colvar%dfunct_param%coeff*r34
      fi(:) = xij/r12
      fj(:) = -xij/r12
      fk(:) = colvar%dfunct_param%coeff*xkl/r34
      fl(:) = -colvar%dfunct_param%coeff*xkl/r34
      CALL put_derivative(colvar, 1, fi)
      CALL put_derivative(colvar, 2, fj)
      CALL put_derivative(colvar, 3, fk)
      CALL put_derivative(colvar, 4, fl)

   END SUBROUTINE dfunct_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the distance from the plane collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino 02.2006 [created]
! **************************************************************************************************
   SUBROUTINE angle_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'angle_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      REAL(dp)                                           :: a, b, fi(3), fj(3), fk(3), fmod, ri(3), &
                                                            rj(3), rk(3), ss(3), t0, t1, t2, t3, &
                                                            xij(3), xkj(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == angle_colvar_id))CALL cp__a("colvar_methods.F",2759)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2763)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%angle_param%i_at_angle(1)
      j = colvar%angle_param%i_at_angle(2)
      k = colvar%angle_param%i_at_angle(3)
      CALL get_coordinates(colvar, i, ri, my_particles)
      CALL get_coordinates(colvar, j, rj, my_particles)
      CALL get_coordinates(colvar, k, rk, my_particles)
      ! xij
      ss = MATMUL(cell%h_inv, ri-rj)
      ss = ss-NINT(ss)
      xij = MATMUL(cell%hmat, ss)
      ! xkj
      ss = MATMUL(cell%h_inv, rk-rj)
      ss = ss-NINT(ss)
      xkj = MATMUL(cell%hmat, ss)
      ! Evaluation of the angle..
      a = SQRT(DOT_PRODUCT(xij, xij))
      b = SQRT(DOT_PRODUCT(xkj, xkj))
      t0 = 1.0_dp/(a*b)
      t1 = 1.0_dp/(a**3.0_dp*b)
      t2 = 1.0_dp/(a*b**3.0_dp)
      t3 = DOT_PRODUCT(xij, xkj)
      colvar%ss = ACOS(t3*t0)
      IF ((ABS(colvar%ss) .LT. tolerance_acos) .OR. (ABS(colvar%ss-pi) .LT. tolerance_acos)) THEN
         fmod = 0.0_dp
      ELSE
         fmod = -1.0_dp/SIN(colvar%ss)
      ENDIF
      fi(:) = xkj(:)*t0-xij(:)*t1*t3
      fj(:) = -xkj(:)*t0+xij(:)*t1*t3-xij(:)*t0+xkj(:)*t2*t3
      fk(:) = xij(:)*t0-xkj(:)*t2*t3
      fi = fi*fmod
      fj = fj*fmod
      fk = fk*fmod
      CALL put_derivative(colvar, 1, fi)
      CALL put_derivative(colvar, 2, fj)
      CALL put_derivative(colvar, 3, fk)

   END SUBROUTINE angle_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the distance collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Alessandro Laio, Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE dist_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'dist_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(dp)                                           :: fi(3), fj(3), r12, ss(3), xij(3), &
                                                            xpi(3), xpj(3)
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)

      IF(.NOT.(colvar%type_id == dist_colvar_id))CALL cp__a("colvar_methods.F",2831)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2835)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      i = colvar%dist_param%i_at
      j = colvar%dist_param%j_at
      CALL get_coordinates(colvar, i, xpi, my_particles)
      CALL get_coordinates(colvar, j, xpj, my_particles)
      ss = MATMUL(cell%h_inv, xpi-xpj)
      ss = ss-NINT(ss)
      xij = MATMUL(cell%hmat, ss)
      SELECT CASE (colvar%dist_param%axis_id)
      CASE (do_clv_x)
         xij(2) = 0.0_dp
         xij(3) = 0.0_dp
      CASE (do_clv_y)
         xij(1) = 0.0_dp
         xij(3) = 0.0_dp
      CASE (do_clv_z)
         xij(1) = 0.0_dp
         xij(2) = 0.0_dp
      CASE (do_clv_xy)
         xij(3) = 0.0_dp
      CASE (do_clv_xz)
         xij(2) = 0.0_dp
      CASE (do_clv_yz)
         xij(1) = 0.0_dp
      CASE DEFAULT
         !do_clv_xyz
      END SELECT
      r12 = SQRT(xij(1)**2+xij(2)**2+xij(3)**2)

      colvar%ss = r12
      fi(:) = xij/r12
      fj(:) = -xij/r12

      CALL put_derivative(colvar, 1, fi)
      CALL put_derivative(colvar, 2, fj)

   END SUBROUTINE dist_colvar

! **************************************************************************************************
!> \brief evaluates the force due to the torsion collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \param no_riemann_sheet_op ...
!> \author Alessandro Laio, Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE torsion_colvar(colvar, cell, subsys, particles, no_riemann_sheet_op)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles
      LOGICAL, INTENT(IN), OPTIONAL                      :: no_riemann_sheet_op

      CHARACTER(len=*), PARAMETER :: routineN = 'torsion_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii
      LOGICAL                                            :: no_riemann_sheet
      REAL(dp) :: angle, cosine, dedphi, dedxia, dedxib, dedxic, dedxid, dedxt, dedxu, dedyia, &
         dedyib, dedyic, dedyid, dedyt, dedyu, dedzia, dedzib, dedzic, dedzid, dedzt, dedzu, dt, &
         e, ftmp(3), o0, rcb, rt2, rtmp(3), rtru, ru2, sine, ss(3), xba, xca, xcb, xdb, xdc, xt, &
         xtu, xu, yba, yca, ycb, ydb, ydc, yt, ytu, yu, zba, zca, zcb, zdb, zdc, zt, ztu, zu
      REAL(dp), DIMENSION(3, 4)                          :: rr
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == torsion_colvar_id))CALL cp__a("colvar_methods.F",2907)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",2911)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      no_riemann_sheet = .FALSE.
      IF (PRESENT(no_riemann_sheet_op)) no_riemann_sheet = no_riemann_sheet_op
      DO ii = 1, 4
         i = colvar%torsion_param%i_at_tors(ii)
         CALL get_coordinates(colvar, i, rtmp, my_particles)
         rr(:, ii) = rtmp(1:3)
      ENDDO
      o0 = colvar%torsion_param%o0
      ! ba
      ss = MATMUL(cell%h_inv, rr(:, 2)-rr(:, 1))
      ss = ss-NINT(ss)
      ss = MATMUL(cell%hmat, ss)
      xba = ss(1)
      yba = ss(2)
      zba = ss(3)
      ! cb
      ss = MATMUL(cell%h_inv, rr(:, 3)-rr(:, 2))
      ss = ss-NINT(ss)
      ss = MATMUL(cell%hmat, ss)
      xcb = ss(1)
      ycb = ss(2)
      zcb = ss(3)
      ! dc
      ss = MATMUL(cell%h_inv, rr(:, 4)-rr(:, 3))
      ss = ss-NINT(ss)
      ss = MATMUL(cell%hmat, ss)
      xdc = ss(1)
      ydc = ss(2)
      zdc = ss(3)
      !
      xt = yba*zcb-ycb*zba
      yt = zba*xcb-zcb*xba
      zt = xba*ycb-xcb*yba
      xu = ycb*zdc-ydc*zcb
      yu = zcb*xdc-zdc*xcb
      zu = xcb*ydc-xdc*ycb
      xtu = yt*zu-yu*zt
      ytu = zt*xu-zu*xt
      ztu = xt*yu-xu*yt
      rt2 = xt*xt+yt*yt+zt*zt
      ru2 = xu*xu+yu*yu+zu*zu
      rtru = SQRT(rt2*ru2)
      IF (rtru .NE. 0.0_dp) THEN
         rcb = SQRT(xcb*xcb+ycb*ycb+zcb*zcb)
         cosine = (xt*xu+yt*yu+zt*zu)/rtru
         sine = (xcb*xtu+ycb*ytu+zcb*ztu)/(rcb*rtru)
         cosine = MIN(1.0_dp, MAX(-1.0_dp, cosine))
         angle = ACOS(cosine)
         IF (sine .LT. 0.0_dp) angle = -angle
         !
         dt = angle ! [rad]
         dt = MOD(2.0E4_dp*pi+dt-o0, 2.0_dp*pi)
         IF (dt .GT. pi) dt = dt-2.0_dp*pi
         dt = o0+dt
         colvar%torsion_param%o0 = dt
         !
         !     calculate improper energy and master chain rule term
         !
         e = dt
         dedphi = 1.0_dp
         !
         !     chain rule terms for first derivative components
         !
         ! ca
         ss = MATMUL(cell%h_inv, rr(:, 3)-rr(:, 1))
         ss = ss-NINT(ss)
         ss = MATMUL(cell%hmat, ss)
         xca = ss(1)
         yca = ss(2)
         zca = ss(3)
         ! db
         ss = MATMUL(cell%h_inv, rr(:, 4)-rr(:, 2))
         ss = ss-NINT(ss)
         ss = MATMUL(cell%hmat, ss)
         xdb = ss(1)
         ydb = ss(2)
         zdb = ss(3)
         !
         dedxt = dedphi*(yt*zcb-ycb*zt)/(rt2*rcb)
         dedyt = dedphi*(zt*xcb-zcb*xt)/(rt2*rcb)
         dedzt = dedphi*(xt*ycb-xcb*yt)/(rt2*rcb)
         dedxu = -dedphi*(yu*zcb-ycb*zu)/(ru2*rcb)
         dedyu = -dedphi*(zu*xcb-zcb*xu)/(ru2*rcb)
         dedzu = -dedphi*(xu*ycb-xcb*yu)/(ru2*rcb)
         !
         !     compute first derivative components for this angle
         !
         dedxia = zcb*dedyt-ycb*dedzt
         dedyia = xcb*dedzt-zcb*dedxt
         dedzia = ycb*dedxt-xcb*dedyt
         dedzia = ycb*dedxt-xcb*dedyt
         dedxib = yca*dedzt-zca*dedyt+zdc*dedyu-ydc*dedzu
         dedyib = zca*dedxt-xca*dedzt+xdc*dedzu-zdc*dedxu
         dedzib = xca*dedyt-yca*dedxt+ydc*dedxu-xdc*dedyu
         dedxic = zba*dedyt-yba*dedzt+ydb*dedzu-zdb*dedyu
         dedyic = xba*dedzt-zba*dedxt+zdb*dedxu-xdb*dedzu
         dedzic = yba*dedxt-xba*dedyt+xdb*dedyu-ydb*dedxu
         dedxid = zcb*dedyu-ycb*dedzu
         dedyid = xcb*dedzu-zcb*dedxu
         dedzid = ycb*dedxu-xcb*dedyu
      ENDIF
      !
      colvar%ss = e
      IF (no_riemann_sheet) colvar%ss = ATAN2(SIN(e), COS(e))
      ftmp(1) = dedxia
      ftmp(2) = dedyia
      ftmp(3) = dedzia
      CALL put_derivative(colvar, 1, ftmp)
      ftmp(1) = dedxib
      ftmp(2) = dedyib
      ftmp(3) = dedzib
      CALL put_derivative(colvar, 2, ftmp)
      ftmp(1) = dedxic
      ftmp(2) = dedyic
      ftmp(3) = dedzic
      CALL put_derivative(colvar, 3, ftmp)
      ftmp(1) = dedxid
      ftmp(2) = dedyid
      ftmp(3) = dedzid
      CALL put_derivative(colvar, 4, ftmp)
   END SUBROUTINE torsion_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the Q PARM collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
! **************************************************************************************************
   SUBROUTINE qparm_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'qparm_colvar', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: ylm, ylm_calc, ylm_calc2
      INTEGER                                            :: i, ii, j, jj, l, m, n_atoms_from, &
                                                            n_atoms_to
      REAL(KIND=dp) :: alpha, bond, costheta, dplm, dylm, exp_fac, fact, fi, ftmp(3), &
         inv_n_atoms_from, nbond, plm, pre_fac, qlm, qlm2, qparm, r12, r_tmp, rcut, ss(3), theta, &
         x_tmp, xij(3), y_tmp, ymag, ymag1, z_tmp
      REAL(KIND=dp), DIMENSION(3)                        :: grad, grad_nb, grad_nb_tot, grad_tot, &
                                                            xpi, xpj
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      n_atoms_to = colvar%qparm_param%n_atoms_to
      n_atoms_from = colvar%qparm_param%n_atoms_from
      rcut = colvar%qparm_param%rcut
      l = colvar%qparm_param%l
      alpha = colvar%qparm_param%alpha
      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == qparm_colvar_id))CALL cp__a("colvar_methods.F",3070)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3074)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      qparm = 0.0_dp
      inv_n_atoms_from = 1.0_dp/REAL(n_atoms_from)
      DO ii = 1, n_atoms_from
         i = colvar%qparm_param%i_at_from(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         qlm = 0.0_dp
         qlm2 = 0.0_dp
         grad_tot(:) = 0.0_dp
         grad_nb_tot(:) = 0.0_dp
         DO m = 0, l
            ylm = 0.0_dp
            grad(:) = 0.0_dp
            grad_nb(:) = 0.0_dp
            nbond = 0
            DO jj = 1, n_atoms_to
               j = colvar%qparm_param%i_at_to(jj)
               IF (i == j) CYCLE
               CALL get_coordinates(colvar, j, xpj, my_particles)
               ss = MATMUL(cell%h_inv, xpj(:)-xpi(:))
               ss = ss-NINT(ss)
               xij = MATMUL(cell%hmat, ss)
               r12 = SQRT(DOT_PRODUCT(xij, xij))
               IF (r12 < rcut) nbond = nbond+1.0_dp
               x_tmp = xij(1)
               y_tmp = xij(2)
               z_tmp = xij(3)
               r_tmp = r12
               IF (ABS(x_tmp) .GT. 1.0E-8_dp) THEN
                  fi = ATAN(y_tmp/x_tmp)
               ELSE
                  fi = 0.5_dp*pi
               ENDIF
               costheta = z_tmp/r_tmp
               IF (costheta > 1.0_dp) costheta = 1.0_dp
               IF (costheta < -1.0_dp) costheta = -1.0_dp
               theta = ACOS(costheta)
               plm = legendre(costheta, l, m)
               dplm = dlegendre(costheta, l, m)
               IF ((l+m) > maxfac) THEN
                  CALL cp__b("colvar_methods.F",3118,"(l-m) > maxfac")
               END IF
               bond = 1.0_dp/(1.0_dp+EXP(alpha*(r_tmp-rcut)))
               IF (bond > 1.0_dp) THEN
                  CALL cp__b("colvar_methods.F",3122,"bond > 1.0_dp")
               END IF
               pre_fac = bond*SQRT(((2*l+1)*fac(l-m))/(4*pi*fac(l+m)))
               IF (plm < 0.0_dp) THEN
                  dylm = -pre_fac*dplm
               ELSE
                  dylm = pre_fac*dplm
               ENDIF
               ylm_calc = pre_fac*plm*(CMPLX(COS(m*fi), SIN(m*fi), KIND=dp))
               ylm = ylm+ylm_calc
               ylm_calc2 = pre_fac/bond*plm*(CMPLX(COS(m*fi), SIN(m*fi), KIND=dp))
               ymag = (SQRT(REAL(ylm_calc2*CONJG(ylm_calc2), KIND=dp)))
               ! Fi = -dUi/dr, but r_tmp = rj - ri, so have double negative
               grad(1) = grad(1)-dylm*(z_tmp*x_tmp)/(r_tmp*r_tmp*r_tmp)
               grad(2) = grad(2)-dylm*(z_tmp*y_tmp)/(r_tmp*r_tmp*r_tmp)
               grad(3) = grad(3)+dylm*((1.0_dp/r_tmp)-(z_tmp**2/r_tmp**3))
               exp_fac = alpha*EXP(alpha*(r_tmp-rcut))/((1.0_dp+EXP(alpha*(r_tmp-rcut)))**2)
               grad_nb(1) = grad_nb(1)-ymag*exp_fac*(x_tmp/r_tmp)
               grad_nb(2) = grad_nb(2)-ymag*exp_fac*(y_tmp/r_tmp)
               grad_nb(3) = grad_nb(3)-ymag*exp_fac*(z_tmp/r_tmp)
            ENDDO
            IF (m .GT. 0) THEN
               fact = 2.0_dp
            ELSE
               fact = 1.0_dp
            ENDIF
            IF (nbond < 0.0001_dp) THEN
               CALL cp__b("colvar_methods.F",3149,"NBOND = 0.0")
            END IF
            ylm = ylm/nbond
            grad(:) = grad(:)/nbond
            grad_nb(:) = grad_nb(:)/nbond
            ymag1 = REAL(ylm*CONJG(ylm), KIND=dp)
            qlm = qlm+fact*ymag1
            grad_tot(:) = grad_tot(:)+fact*grad(:)
            grad_nb_tot(:) = grad_nb_tot(:)+fact*grad_nb(:)
         ENDDO
         pre_fac = (4.0_dp*pi)/(2.0_dp*l+1)
         qlm = SQRT(pre_fac*qlm)
         qlm2 = SQRT(pre_fac*qlm2)
         qparm = qparm+qlm
         ftmp(:) = SQRT(pre_fac)*(grad_tot(:)+grad_nb_tot(:))
         CALL put_derivative(colvar, ii, ftmp)
      ENDDO
      colvar%ss = qparm*inv_n_atoms_from
      colvar%dsdr(:, :) = colvar%dsdr(:, :)*inv_n_atoms_from

   END SUBROUTINE qparm_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the hydronium collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Marcel Baer
!> \note This function needs to be extended to the POINT structure!!
!>       non-standard conform.. it's a breach in the colvar module.
! **************************************************************************************************
   SUBROUTINE hydronium_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'hydronium_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, j, jj, n_hydrogens, n_oxygens, &
                                                            pm, poh, poo, qm, qoh, qoo
      REAL(dp)                                           :: drji, fscalar, invden, lambda, nh, num, &
                                                            qtot, rji(3), roh, roo, rrel
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: M, noh, noo, qloc
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: dM, dnoh, dnoo
      REAL(dp), DIMENSION(3)                             :: rpi, rpj
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      n_oxygens = colvar%hydronium_param%n_oxygens
      n_hydrogens = colvar%hydronium_param%n_hydrogens
      nh = colvar%hydronium_param%nh
      poh = colvar%hydronium_param%poh
      qoh = colvar%hydronium_param%qoh
      poo = colvar%hydronium_param%poo
      qoo = colvar%hydronium_param%qoo
      roo = colvar%hydronium_param%roo
      roh = colvar%hydronium_param%roh
      lambda = colvar%hydronium_param%lambda
      pm = colvar%hydronium_param%pm
      qm = colvar%hydronium_param%qm

      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == hydronium_colvar_id))CALL cp__a("colvar_methods.F",3215)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3219)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      ALLOCATE (dnoh(3, n_hydrogens, n_oxygens))
      ALLOCATE (noh(n_oxygens))
      ALLOCATE (M(n_oxygens))
      ALLOCATE (dM(3, n_hydrogens, n_oxygens))

      ALLOCATE (dnoo(3, n_oxygens, n_oxygens))
      ALLOCATE (noo(n_oxygens))

      ALLOCATE (qloc(n_oxygens))

      ! Zero Arrays:
      dnoh = 0._dp
      dnoo = 0._dp
      M = 0._dp
      dM = 0._dp
      noo = 0._dp
      qloc = 0._dp
      noh = 0._dp
      DO ii = 1, n_oxygens
         i = colvar%hydronium_param%i_oxygens(ii)
         rpi(:) = my_particles(i)%r(1:3)
         ! Computing M( n ( ii ) )
         DO jj = 1, n_hydrogens
            j = colvar%hydronium_param%i_hydrogens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rji = pbc(rpj, rpi, cell)
            drji = SQRT(SUM(rji**2))
            rrel = drji/roh
            num = (1.0_dp-rrel**poh)
            invden = 1.0_dp/(1.0_dp-rrel**qoh)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               noh(ii) = noh(ii)+num*invden
               fscalar = ((-poh*(rrel**(poh-1))*invden) &
                          +num*(invden)**2*qoh*(rrel**(qoh-1)))/(drji*roh)
               dnoh(1:3, jj, ii) = rji(1:3)*fscalar
            ELSE
               !correct limit if rji --> roh
               noh(ii) = noh(ii)+REAL(poh, dp)/REAL(qoh, dp)
               fscalar = REAL(poh*(poh-qoh), dp)/(REAL(2*qoh, dp)*roh*drji)
               dnoh(1:3, jj, ii) = rji(1:3)*fscalar
            ENDIF
         END DO
         M(ii) = 1.0_dp-(1.0_dp-(noh(ii)/nh)**pm)/ &
                 (1.0_dp-(noh(ii)/nh)**qm)

         ! Computing no ( ii )
         DO jj = 1, n_oxygens
            IF (ii == jj) CYCLE
            j = colvar%hydronium_param%i_oxygens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rji = pbc(rpj, rpi, cell)
            drji = SQRT(SUM(rji**2))
            rrel = drji/roo
            num = (1.0_dp-rrel**poo)
            invden = 1.0_dp/(1.0_dp-rrel**qoo)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               noo(ii) = noo(ii)+num*invden
               fscalar = ((-poo*(rrel**(poo-1))*invden) &
                          +num*(invden)**2*qoo*(rrel**(qoo-1)))/(drji*roo)
               dnoo(1:3, jj, ii) = rji(1:3)*fscalar
            ELSE
               !correct limit if rji --> roo
               noo(ii) = noo(ii)+REAL(poo, dp)/REAL(qoo, dp)
               fscalar = REAL(poo*(poo-qoo), dp)/(REAL(2*qoo, dp)*roo*drji)
               dnoo(1:3, jj, ii) = rji(1:3)*fscalar
            ENDIF
         END DO
      END DO

      ! computing qloc and Q
      qtot = 0._dp
      DO ii = 1, n_oxygens
         qloc(ii) = EXP(lambda*M(ii)*noo(ii))
         qtot = qtot+qloc(ii)
      END DO
      ! compute forces
      DO ii = 1, n_oxygens
         ! Computing f_OH
         DO jj = 1, n_hydrogens
            dM(1:3, jj, ii) = (pm*((noh(ii)/nh)**(pm-1))*dnoh(1:3, jj, ii))/nh/ &
                              (1.0_dp-(noh(ii)/nh)**qm)- &
                              (1.0_dp-(noh(ii)/nh)**pm)/ &
                              ((1.0_dp-(noh(ii)/nh)**qm)**2)* &
                              qm*dnoh(1:3, jj, ii)*(noh(ii)/nh)**(qm-1)/nh

            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii)+qloc(ii)*dM(1:3, jj, ii)*noo(ii)/qtot
            colvar%dsdr(1:3, n_oxygens+jj) = colvar%dsdr(1:3, n_oxygens+jj) &
                                             -qloc(ii)*dM(1:3, jj, ii)*noo(ii)/qtot
         END DO
         ! Computing f_OO
         DO jj = 1, n_oxygens
            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii)+qloc(ii)*M(ii)*dnoo(1:3, jj, ii)/qtot
            colvar%dsdr(1:3, jj) = colvar%dsdr(1:3, jj) &
                                   -qloc(ii)*M(ii)*dnoo(1:3, jj, ii)/qtot
         END DO
      END DO

      colvar%ss = LOG(qtot)/lambda
      DEALLOCATE (dnoh)
      DEALLOCATE (noh)
      DEALLOCATE (M)
      DEALLOCATE (dM)
      DEALLOCATE (dnoo)
      DEALLOCATE (noo)
      DEALLOCATE (qloc)
   END SUBROUTINE hydronium_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the acid-hydronium-distance
!>        collective variable. Colvar: distance between carboxy group and
!>        hydronium ion.
!> \param colvar collective variable
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Dorothea Golze
!> \note this function does not use POINTS, not reasonable for this colvar
! **************************************************************************************************
   SUBROUTINE acid_hyd_dist_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'acid_hyd_dist_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, j, jj, k, kk, n_hydrogens, &
                                                            n_oxygens_acid, n_oxygens_water, &
                                                            offsetH, offsetO, paoh, pcut, pwoh, &
                                                            qaoh, qcut, qwoh
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dexpfac, expfac, nwoh
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: dexpfac_rik
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ddist_rik, dnaoh, dnwoh
      REAL(KIND=dp) :: dfcut, drik, drji, drjk, fbrace, fcut, fscalar, invden, invden_cut, lambda, &
         naoh, nc, num, num_cut, raoh, rik(3), rion, rion_den, rion_num, rji(3), rjk(3), rpi(3), &
         rpj(3), rpk(3), rrel, rwoh
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (my_particles, particles_i)

      n_oxygens_water = colvar%acid_hyd_dist_param%n_oxygens_water
      n_oxygens_acid = colvar%acid_hyd_dist_param%n_oxygens_acid
      n_hydrogens = colvar%acid_hyd_dist_param%n_hydrogens
      pwoh = colvar%acid_hyd_dist_param%pwoh
      qwoh = colvar%acid_hyd_dist_param%qwoh
      paoh = colvar%acid_hyd_dist_param%paoh
      qaoh = colvar%acid_hyd_dist_param%qaoh
      pcut = colvar%acid_hyd_dist_param%pcut
      qcut = colvar%acid_hyd_dist_param%qcut
      rwoh = colvar%acid_hyd_dist_param%rwoh
      raoh = colvar%acid_hyd_dist_param%raoh
      nc = colvar%acid_hyd_dist_param%nc
      lambda = colvar%acid_hyd_dist_param%lambda
      ALLOCATE (expfac(n_oxygens_water))
      ALLOCATE (nwoh(n_oxygens_water))
      ALLOCATE (dnwoh(3, n_hydrogens, n_oxygens_water))
      ALLOCATE (dnaoh(3, n_hydrogens, n_oxygens_acid))
      ALLOCATE (dexpfac(n_oxygens_water))
      ALLOCATE (ddist_rik(3, n_oxygens_water, n_oxygens_acid))
      ALLOCATE (dexpfac_rik(n_oxygens_water, n_oxygens_acid))
      rion_den = 0._dp
      rion_num = 0._dp
      nwoh(:) = 0._dp
      naoh = 0._dp
      dnaoh(:, :, :) = 0._dp
      dnwoh(:, :, :) = 0._dp
      ddist_rik(:, :, :) = 0._dp
      dexpfac(:) = 0._dp
      dexpfac_rik(:, :) = 0._dp

      IF(.NOT.(colvar%type_id == acid_hyd_dist_colvar_id))CALL cp__a("colvar_methods.F",3397)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3401)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      ! Calculate coordination functions nwoh(ii) and denominator of rion
      DO ii = 1, n_oxygens_water
         i = colvar%acid_hyd_dist_param%i_oxygens_water(ii)
         rpi(:) = my_particles(i)%r(1:3)
         DO jj = 1, n_hydrogens
            j = colvar%acid_hyd_dist_param%i_hydrogens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rji = pbc(rpj, rpi, cell)
            drji = SQRT(SUM(rji**2))
            rrel = drji/rwoh
            num = 1.0_dp-rrel**pwoh
            invden = 1.0_dp/(1.0_dp-rrel**qwoh)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               nwoh(ii) = nwoh(ii)+num*invden
               fscalar = (-pwoh*(rrel**(pwoh-1))*invden &
                          +num*(invden**2)*qwoh*(rrel**(qwoh-1)))/(drji*rwoh)
               dnwoh(1:3, jj, ii) = rji(1:3)*fscalar
            ELSE
               !correct limit if rji --> rwoh
               nwoh(ii) = nwoh(ii)+REAL(pwoh, dp)/REAL(qwoh, dp)
               fscalar = REAL(pwoh*(pwoh-qwoh), dp)/(REAL(2*qwoh, dp)*rwoh*drji)
               dnwoh(1:3, jj, ii) = rji(1:3)*fscalar
            ENDIF
         ENDDO
         expfac(ii) = EXP(lambda*nwoh(ii))
         dexpfac(ii) = lambda*expfac(ii)
         rion_den = rion_den+expfac(ii)
      ENDDO

      ! Calculate nominator of rion
      DO kk = 1, n_oxygens_acid
         k = colvar%acid_hyd_dist_param%i_oxygens_acid(kk)
         rpk(:) = my_particles(k)%r(1:3)
         DO ii = 1, n_oxygens_water
            i = colvar%acid_hyd_dist_param%i_oxygens_water(ii)
            rpi(:) = my_particles(i)%r(1:3)
            rik = pbc(rpi, rpk, cell)
            drik = SQRT(SUM(rik**2))
            rion_num = rion_num+drik*expfac(ii)
            ddist_rik(1:3, ii, kk) = rik(1:3)/drik
            dexpfac_rik(ii, kk) = drik*dexpfac(ii)
         ENDDO
      ENDDO

      !Calculate cutoff function
      DO kk = 1, n_oxygens_acid
         k = colvar%acid_hyd_dist_param%i_oxygens_acid(kk)
         rpk(:) = my_particles(k)%r(1:3)
         DO jj = 1, n_hydrogens
            j = colvar%acid_hyd_dist_param%i_hydrogens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rjk = pbc(rpj, rpk, cell)
            drjk = SQRT(SUM(rjk**2))
            rrel = drjk/raoh
            num = 1.0_dp-rrel**paoh
            invden = 1.0_dp/(1.0_dp-rrel**qaoh)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               naoh = naoh+num*invden
               fscalar = (-paoh*(rrel**(paoh-1))*invden &
                          +num*(invden**2)*qaoh*(rrel**(qaoh-1)))/(drjk*raoh)
               dnaoh(1:3, jj, kk) = rjk(1:3)*fscalar
            ELSE
               !correct limit if rjk --> raoh
               naoh = naoh+REAL(paoh, dp)/REAL(qaoh, dp)
               fscalar = REAL(paoh*(paoh-qaoh), dp)/(REAL(2*qaoh, dp)*raoh*drjk)
               dnaoh(1:3, jj, kk) = rjk(1:3)*fscalar
            ENDIF
         ENDDO
      ENDDO
      num_cut = 1.0_dp-(naoh/nc)**pcut
      invden_cut = 1.0_dp/(1.0_dp-(naoh/nc)**qcut)
      fcut = num_cut*invden_cut

      !Final distance acid - hydronium
      fbrace = rion_num/rion_den/2.0_dp
      rion = fcut*fbrace
      colvar%ss = rion

      !Derivatives of fcut
      dfcut = ((-pcut*(naoh/nc)**(pcut-1)*invden_cut) &
               +num_cut*(invden_cut**2)*qcut*(naoh/nc)**(qcut-1))/nc
      offsetO = n_oxygens_water
      offsetH = n_oxygens_water+n_oxygens_acid
      DO kk = 1, n_oxygens_acid
         DO jj = 1, n_hydrogens
            colvar%dsdr(1:3, offsetO+kk) = colvar%dsdr(1:3, offsetO+kk) &
                                           +dfcut*dnaoh(1:3, jj, kk)*fbrace
            colvar%dsdr(1:3, offsetH+jj) = colvar%dsdr(1:3, offsetH+jj) &
                                           -dfcut*dnaoh(1:3, jj, kk)*fbrace
         ENDDO
      ENDDO

      !Derivatives of fbrace
      !***nominator
      DO kk = 1, n_oxygens_acid
         DO ii = 1, n_oxygens_water
            colvar%dsdr(1:3, offsetO+kk) = colvar%dsdr(1:3, offsetO+kk) &
                                           +fcut*ddist_rik(1:3, ii, kk)*expfac(ii)/rion_den/2.0_dp
            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii) &
                                   -fcut*ddist_rik(1:3, ii, kk)*expfac(ii)/rion_den/2.0_dp
            DO jj = 1, n_hydrogens
               colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii) &
                                      +fcut*dexpfac_rik(ii, kk)*dnwoh(1:3, jj, ii)/rion_den/2.0_dp
               colvar%dsdr(1:3, offsetH+jj) = colvar%dsdr(1:3, offsetH+jj) &
                                              -fcut*dexpfac_rik(ii, kk)*dnwoh(1:3, jj, ii)/rion_den/2.0_dp
            ENDDO
         ENDDO
      ENDDO
      !***denominator
      DO ii = 1, n_oxygens_water
         DO jj = 1, n_hydrogens
            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii) &
                                   -fcut*rion_num*dexpfac(ii)*dnwoh(1:3, jj, ii)/2.0_dp/(rion_den**2)
            colvar%dsdr(1:3, offsetH+jj) = colvar%dsdr(1:3, offsetH+jj) &
                                           +fcut*rion_num*dexpfac(ii)*dnwoh(1:3, jj, ii)/2.0_dp/(rion_den**2)
         ENDDO
      ENDDO

   END SUBROUTINE acid_hyd_dist_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the acid-hydronium-shell
!>        collective variable. Colvar: number of oxygens in 1st shell of the
!>        hydronium.
!> \param colvar collective variable
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Dorothea Golze
!> \note this function does not use POINTS, not reasonable for this colvar
! **************************************************************************************************
   SUBROUTINE acid_hyd_shell_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'acid_hyd_shell_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER :: i, ii, j, jj, k, kk, n_hydrogens, n_oxygens_acid, n_oxygens_water, offsetH, &
         offsetO, paoh, pcut, pm, poo, pwoh, qaoh, qcut, qm, qoo, qwoh, tt
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dM, M, noo, nwoh, qloc
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: dnaoh, dnoo, dnwoh
      REAL(KIND=dp) :: dfcut, drji, drjk, drki, fcut, fscalar, invden, invden_cut, lambda, naoh, &
         nc, nh, num, num_cut, qsol, qtot, raoh, rji(3), rjk(3), rki(3), roo, rpi(3), rpj(3), &
         rpk(3), rrel, rwoh
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (my_particles, particles_i)

      n_oxygens_water = colvar%acid_hyd_shell_param%n_oxygens_water
      n_oxygens_acid = colvar%acid_hyd_shell_param%n_oxygens_acid
      n_hydrogens = colvar%acid_hyd_shell_param%n_hydrogens
      pwoh = colvar%acid_hyd_shell_param%pwoh
      qwoh = colvar%acid_hyd_shell_param%qwoh
      paoh = colvar%acid_hyd_shell_param%paoh
      qaoh = colvar%acid_hyd_shell_param%qaoh
      poo = colvar%acid_hyd_shell_param%poo
      qoo = colvar%acid_hyd_shell_param%qoo
      pm = colvar%acid_hyd_shell_param%pm
      qm = colvar%acid_hyd_shell_param%qm
      pcut = colvar%acid_hyd_shell_param%pcut
      qcut = colvar%acid_hyd_shell_param%qcut
      rwoh = colvar%acid_hyd_shell_param%rwoh
      raoh = colvar%acid_hyd_shell_param%raoh
      roo = colvar%acid_hyd_shell_param%roo
      nc = colvar%acid_hyd_shell_param%nc
      nh = colvar%acid_hyd_shell_param%nh
      lambda = colvar%acid_hyd_shell_param%lambda
      ALLOCATE (nwoh(n_oxygens_water))
      ALLOCATE (dnwoh(3, n_hydrogens, n_oxygens_water))
      ALLOCATE (dnaoh(3, n_hydrogens, n_oxygens_acid))
      ALLOCATE (M(n_oxygens_water))
      ALLOCATE (dM(n_oxygens_water))
      ALLOCATE (noo(n_oxygens_water))
      ALLOCATE (dnoo(3, n_oxygens_water+n_oxygens_acid, n_oxygens_water))
      ALLOCATE (qloc(n_oxygens_water))
      nwoh(:) = 0._dp
      naoh = 0._dp
      noo = 0._dp
      dnaoh(:, :, :) = 0._dp
      dnwoh(:, :, :) = 0._dp
      dnoo(:, :, :) = 0._dp
      M = 0._dp
      dM = 0._dp
      qtot = 0._dp

      IF(.NOT.(colvar%type_id == acid_hyd_shell_colvar_id))CALL cp__a("colvar_methods.F",3596)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3600)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      ! Calculate coordination functions nwoh(ii) and the M function
      DO ii = 1, n_oxygens_water
         i = colvar%acid_hyd_shell_param%i_oxygens_water(ii)
         rpi(:) = my_particles(i)%r(1:3)
         DO jj = 1, n_hydrogens
            j = colvar%acid_hyd_shell_param%i_hydrogens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rji = pbc(rpj, rpi, cell)
            drji = SQRT(SUM(rji**2))
            rrel = drji/rwoh
            num = 1.0_dp-rrel**pwoh
            invden = 1.0_dp/(1.0_dp-rrel**qwoh)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               nwoh(ii) = nwoh(ii)+num*invden
               fscalar = (-pwoh*(rrel**(pwoh-1))*invden &
                          +num*(invden**2)*qwoh*(rrel**(qwoh-1)))/(drji*rwoh)
               dnwoh(1:3, jj, ii) = rji(1:3)*fscalar
            ELSE
               !correct limit if rji --> rwoh
               nwoh(ii) = nwoh(ii)+REAL(pwoh, dp)/REAL(qwoh, dp)
               fscalar = REAL(pwoh*(pwoh-qwoh), dp)/(REAL(2*qwoh, dp)*rwoh*drji)
               dnwoh(1:3, jj, ii) = rji(1:3)*fscalar
            ENDIF
         ENDDO
      ENDDO

      ! calculate M function
      DO ii = 1, n_oxygens_water
         num = 1.0_dp-(nwoh(ii)/nh)**pm
         invden = 1.0_dp/(1.0_dp-(nwoh(ii)/nh)**qm)
         M(ii) = 1.0_dp-num*invden
         dM(ii) = (pm*(nwoh(ii)/nh)**(pm-1)*invden-qm*num*(invden**2)* &
                   (nwoh(ii)/nh)**(qm-1))/nh
      ENDDO

      ! Computing noo(i)
      DO ii = 1, n_oxygens_water
         i = colvar%acid_hyd_shell_param%i_oxygens_water(ii)
         rpi(:) = my_particles(i)%r(1:3)
         DO kk = 1, n_oxygens_water+n_oxygens_acid
            IF (ii == kk) CYCLE
            IF (kk <= n_oxygens_water) THEN
               k = colvar%acid_hyd_shell_param%i_oxygens_water(kk)
               rpk(:) = my_particles(k)%r(1:3)
            ELSE
               tt = kk-n_oxygens_water
               k = colvar%acid_hyd_shell_param%i_oxygens_acid(tt)
               rpk(:) = my_particles(k)%r(1:3)
            ENDIF
            rki = pbc(rpk, rpi, cell)
            drki = SQRT(SUM(rki**2))
            rrel = drki/roo
            num = 1.0_dp-rrel**poo
            invden = 1.0_dp/(1.0_dp-rrel**qoo)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               noo(ii) = noo(ii)+num*invden
               fscalar = (-poo*(rrel**(poo-1))*invden &
                          +num*(invden**2)*qoo*(rrel**(qoo-1)))/(drki*roo)
               dnoo(1:3, kk, ii) = rki(1:3)*fscalar
            ELSE
               !correct limit if rki --> roo
               noo(ii) = noo(ii)+REAL(poo, dp)/REAL(qoo, dp)
               fscalar = REAL(poo*(poo-qoo), dp)/(REAL(2*qoo, dp)*roo*drki)
               dnoo(1:3, kk, ii) = rki(1:3)*fscalar
            ENDIF
         ENDDO
      ENDDO

      !Calculate cutoff function
      DO kk = 1, n_oxygens_acid
         k = colvar%acid_hyd_shell_param%i_oxygens_acid(kk)
         rpk(:) = my_particles(k)%r(1:3)
         DO jj = 1, n_hydrogens
            j = colvar%acid_hyd_shell_param%i_hydrogens(jj)
            rpj(:) = my_particles(j)%r(1:3)
            rjk = pbc(rpj, rpk, cell)
            drjk = SQRT(SUM(rjk**2))
            rrel = drjk/raoh
            num = 1.0_dp-rrel**paoh
            invden = 1.0_dp/(1.0_dp-rrel**qaoh)
            IF (ABS(1.0_dp-rrel) > 1.0E-6_dp) THEN
               naoh = naoh+num*invden
               fscalar = (-paoh*(rrel**(paoh-1))*invden &
                          +num*(invden**2)*qaoh*(rrel**(qaoh-1)))/(drjk*raoh)
               dnaoh(1:3, jj, kk) = rjk(1:3)*fscalar
            ELSE
               !correct limit if rjk --> raoh
               naoh = naoh+REAL(paoh, dp)/REAL(qaoh, dp)
               fscalar = REAL(paoh*(paoh-qaoh), dp)/(REAL(2*qaoh, dp)*raoh*drjk)
               dnaoh(1:3, jj, kk) = rjk(1:3)*fscalar
            ENDIF
         ENDDO
      ENDDO
      num_cut = 1.0_dp-(naoh/nc)**pcut
      invden_cut = 1.0_dp/(1.0_dp-(naoh/nc)**qcut)
      fcut = num_cut*invden_cut

      ! Final value: number of oxygens in 1st shell of hydronium
      DO ii = 1, n_oxygens_water
         qloc(ii) = EXP(lambda*M(ii)*noo(ii))
         qtot = qtot+qloc(ii)
      ENDDO
      qsol = LOG(qtot)/lambda
      colvar%ss = fcut*qsol

      ! Derivatives of fcut
      dfcut = ((-pcut*(naoh/nc)**(pcut-1)*invden_cut) &
               +num_cut*(invden_cut**2)*qcut*(naoh/nc)**(qcut-1))/nc
      offsetO = n_oxygens_water
      offsetH = n_oxygens_water+n_oxygens_acid
      DO kk = 1, n_oxygens_acid
         DO jj = 1, n_hydrogens
            colvar%dsdr(1:3, offsetO+kk) = colvar%dsdr(1:3, offsetO+kk) &
                                           +dfcut*dnaoh(1:3, jj, kk)*qsol
            colvar%dsdr(1:3, offsetH+jj) = colvar%dsdr(1:3, offsetH+jj) &
                                           -dfcut*dnaoh(1:3, jj, kk)*qsol
         ENDDO
      ENDDO

      ! Derivatives of qsol
      !*** M derivatives
      DO ii = 1, n_oxygens_water
         fscalar = fcut*qloc(ii)*dM(ii)*noo(ii)/qtot
         DO jj = 1, n_hydrogens
            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii) &
                                   +fscalar*dnwoh(1:3, jj, ii)
            colvar%dsdr(1:3, offsetH+jj) = colvar%dsdr(1:3, offsetH+jj) &
                                           -fscalar*dnwoh(1:3, jj, ii)
         ENDDO
      ENDDO
      !*** noo derivatives
      DO ii = 1, n_oxygens_water
         fscalar = fcut*qloc(ii)*M(ii)/qtot
         DO kk = 1, n_oxygens_water+n_oxygens_acid
            IF (ii == kk) CYCLE
            colvar%dsdr(1:3, ii) = colvar%dsdr(1:3, ii)+fscalar*dnoo(1:3, kk, ii)
            colvar%dsdr(1:3, kk) = colvar%dsdr(1:3, kk)-fscalar*dnoo(1:3, kk, ii)
         ENDDO
      ENDDO

   END SUBROUTINE acid_hyd_shell_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) the coordination-chain collective variable
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author MI
!> \note When the third set of atoms is not defined, this variable is equivalent
!>       to the simple coordination number.
! **************************************************************************************************
   SUBROUTINE coord_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'coord_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, j, jj, k, kk, n_atoms_from, &
                                                            n_atoms_to_a, n_atoms_to_b, p_a, p_b, &
                                                            q_a, q_b
      REAL(dp) :: dfunc_ij, dfunc_jk, func_ij, func_jk, func_k, inv_n_atoms_from, invden_ij, &
         invden_jk, ncoord, num_ij, num_jk, r_0_a, r_0_b, rdist_ij, rdist_jk, rij, rjk
      REAL(dp), DIMENSION(3)                             :: ftmp_i, ftmp_j, ftmp_k, ss, xij, xjk, &
                                                            xpi, xpj, xpk
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

! If we defined the coordination number with KINDS then we have still
! to fill few missing informations...

      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == coord_colvar_id))CALL cp__a("colvar_methods.F",3780)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3784)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      n_atoms_to_a = colvar%coord_param%n_atoms_to
      n_atoms_to_b = colvar%coord_param%n_atoms_to_b
      n_atoms_from = colvar%coord_param%n_atoms_from
      p_a = colvar%coord_param%nncrd
      q_a = colvar%coord_param%ndcrd
      r_0_a = colvar%coord_param%r_0
      p_b = colvar%coord_param%nncrd_b
      q_b = colvar%coord_param%ndcrd_b
      r_0_b = colvar%coord_param%r_0_b

      ncoord = 0.0_dp
      inv_n_atoms_from = 1.0_dp/REAL(n_atoms_from)
      DO ii = 1, n_atoms_from
         i = colvar%coord_param%i_at_from(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         DO jj = 1, n_atoms_to_a
            j = colvar%coord_param%i_at_to(jj)
            CALL get_coordinates(colvar, j, xpj, my_particles)
            ! define coordination of atom A with itself to be 0. also fixes rij==0 for the force calculation
            IF (i .EQ. j) CYCLE
            ss = MATMUL(cell%h_inv, xpi(:)-xpj(:))
            ss = ss-NINT(ss)
            xij = MATMUL(cell%hmat, ss)
            rij = SQRT(xij(1)**2+xij(2)**2+xij(3)**2)
            IF (rij < 1.0e-8_dp) CYCLE
            rdist_ij = rij/r_0_a
            IF (ABS(1.0_dp-rdist_ij) > EPSILON(0.0_dp)*1.0E+4_dp) THEN
               num_ij = (1.0_dp-rdist_ij**p_a)
               invden_ij = 1.0_dp/(1.0_dp-rdist_ij**q_a)
               func_ij = num_ij*invden_ij
               IF (rij < 1.0E-8_dp) THEN
                  ! provide the correct limit of the derivative
                  dfunc_ij = 0.0_dp
               ELSE
                  dfunc_ij = (-p_a*rdist_ij**(p_a-1)*invden_ij &
                              +num_ij*(invden_ij)**2*q_a*rdist_ij**(q_a-1))/(rij*r_0_a)
               END IF
            ELSE
               ! Provide the correct limit for function value and derivative
               func_ij = REAL(p_a, KIND=dp)/REAL(q_a, KIND=dp)
               dfunc_ij = REAL(p_a, KIND=dp)*REAL((-q_a+p_a), KIND=dp)/(REAL(2*q_a, KIND=dp)*r_0_a)
            END IF
            IF (n_atoms_to_b /= 0) THEN
               func_k = 0.0_dp
               DO kk = 1, n_atoms_to_b
                  k = colvar%coord_param%i_at_to_b(kk)
                  IF (k .EQ. j) CYCLE
                  CALL get_coordinates(colvar, k, xpk, my_particles)
                  ss = MATMUL(cell%h_inv, xpj(:)-xpk(:))
                  ss = ss-NINT(ss)
                  xjk = MATMUL(cell%hmat, ss)
                  rjk = SQRT(xjk(1)**2+xjk(2)**2+xjk(3)**2)
                  IF (rjk < 1.0e-8_dp) CYCLE
                  rdist_jk = rjk/r_0_b
                  IF (ABS(1.0_dp-rdist_jk) > EPSILON(0.0_dp)*1.0E+4_dp) THEN
                     num_jk = (1.0_dp-rdist_jk**p_b)
                     invden_jk = 1.0_dp/(1.0_dp-rdist_jk**q_b)
                     func_jk = num_jk*invden_jk
                     IF (rjk < 1.0E-8_dp) THEN
                        ! provide the correct limit of the derivative
                        dfunc_jk = 0.0_dp
                     ELSE
                        dfunc_jk = (-p_b*rdist_jk**(p_b-1)*invden_jk &
                                    +num_jk*(invden_jk)**2*q_b*rdist_jk**(q_b-1))/(rjk*r_0_b)
                     END IF
                  ELSE
                     ! Provide the correct limit for function value and derivative
                     func_jk = REAL(p_b, KIND=dp)/REAL(q_b, KIND=dp)
                     dfunc_jk = REAL(p_b, KIND=dp)*REAL((-q_b+p_b), KIND=dp)/(REAL(2*q_b, KIND=dp)*r_0_b)
                  ENDIF
                  func_k = func_k+func_jk
                  ftmp_k = -func_ij*dfunc_jk*xjk
                  CALL put_derivative(colvar, n_atoms_from+n_atoms_to_a+kk, ftmp_k)

                  ftmp_j = -dfunc_ij*xij*func_jk+func_ij*dfunc_jk*xjk
                  CALL put_derivative(colvar, n_atoms_from+jj, ftmp_j)
               END DO
            ELSE
               func_k = 1.0_dp
               dfunc_jk = 0.0_dp
               ftmp_j = -dfunc_ij*xij
               CALL put_derivative(colvar, n_atoms_from+jj, ftmp_j)
            END IF
            ncoord = ncoord+func_ij*func_k
            ftmp_i = dfunc_ij*xij*func_k
            CALL put_derivative(colvar, ii, ftmp_i)
         ENDDO
      ENDDO
      colvar%ss = ncoord*inv_n_atoms_from
      colvar%dsdr(:, :) = colvar%dsdr(:, :)*inv_n_atoms_from
   END SUBROUTINE coord_colvar

! **************************************************************************************************
!> \brief ...
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
! **************************************************************************************************
   SUBROUTINE mindist_colvar(colvar, cell, subsys, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'mindist_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, j, jj, n_coord_from, n_coord_to, &
                                                            n_dist_from, p, q
      REAL(dp) :: den_n, den_Q, fscalar, ftemp_i(3), inv_den_n, inv_den_Q, lambda, num_n, num_Q, &
         Qfunc, r12, r_cut, rfact, rij(3), rpi(3), rpj(3)
      REAL(dp), DIMENSION(:), POINTER                    :: dqfunc_dnL, expnL, nLcoord, sum_rij
      REAL(dp), DIMENSION(:, :, :), POINTER              :: dnLcoord, dqfunc_dr
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

! If we defined the coordination number with KINDS then we have still
! to fill few missing informations...

      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == mindist_colvar_id))CALL cp__a("colvar_methods.F",3910)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",3914)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      n_dist_from = colvar%mindist_param%n_dist_from
      n_coord_from = colvar%mindist_param%n_coord_from
      n_coord_to = colvar%mindist_param%n_coord_to
      p = colvar%mindist_param%p_exp
      q = colvar%mindist_param%q_exp
      r_cut = colvar%mindist_param%r_cut
      lambda = colvar%mindist_param%lambda

      NULLIFY (nLcoord, dnLcoord, dqfunc_dr, dqfunc_dnL, expnL, sum_rij)
      ALLOCATE (nLcoord(n_coord_from))
      ALLOCATE (dnLcoord(3, n_coord_from, n_coord_to))
      ALLOCATE (expnL(n_coord_from))
      ALLOCATE (sum_rij(n_coord_from))
      ALLOCATE (dqfunc_dr(3, n_dist_from, n_coord_from))
      ALLOCATE (dqfunc_dnL(n_coord_from))

      ! coordination numbers
      nLcoord = 0.0_dp
      dnLcoord = 0.0_dp
      expnL = 0.0_dp
      den_Q = 0.0_dp
      DO i = 1, n_coord_from
         ii = colvar%mindist_param%i_coord_from(i)
         rpi = my_particles(ii)%r(1:3)
         DO j = 1, n_coord_to
            jj = colvar%mindist_param%i_coord_to(j)
            rpj = my_particles(jj)%r(1:3)
            rij = pbc(rpj, rpi, cell)
            r12 = SQRT(rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3))
            rfact = r12/r_cut
            num_n = 1.0_dp-rfact**p
            den_n = 1.0_dp-rfact**q
            inv_den_n = 1.0_dp/den_n
            IF (ABS(inv_den_n) < 1.e-10_dp) THEN
               inv_den_n = 1.e-10_dp
               num_n = ABS(num_n)
            END IF

            fscalar = (-p*rfact**(p-1)+num_n*q*rfact**(q-1)*inv_den_n)*inv_den_n/(r_cut*r12)

            dnLcoord(1, i, j) = rij(1)*fscalar
            dnLcoord(2, i, j) = rij(2)*fscalar
            dnLcoord(3, i, j) = rij(3)*fscalar

            nLcoord(i) = nLcoord(i)+num_n*inv_den_n
         END DO
         expnL(i) = EXP(lambda*nLcoord(i))
!dbg
! write(*,*) ii,nLcoord(i),expnL(i)
!dbg
         den_Q = den_Q+expnL(i)
      END DO
      inv_den_Q = 1.0_dp/den_Q

      qfunc = 0.0_dp
      dqfunc_dr = 0.0_dp
      dqfunc_dnL = 0.0_dp
      num_Q = 0.0_dp
      sum_rij = 0.0_dp
      DO i = 1, n_dist_from
         ii = colvar%mindist_param%i_dist_from(i)
         rpi = my_particles(ii)%r(1:3)
         DO j = 1, n_coord_from
            jj = colvar%mindist_param%i_coord_from(j)
            rpj = my_particles(jj)%r(1:3)
            rij = pbc(rpj, rpi, cell)
            r12 = SQRT(rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3))

!dbg
! write(*,*) ii,jj,rpi(1:3),rpj(1:3),rij(1:3),r12
!dbg
            num_Q = num_Q+r12*expnL(j)

            sum_rij(j) = sum_rij(j)+r12
            dqfunc_dr(1, i, j) = expnL(j)*rij(1)/r12
            dqfunc_dr(2, i, j) = expnL(j)*rij(2)/r12
            dqfunc_dr(3, i, j) = expnL(j)*rij(3)/r12

         END DO

      END DO

      ! Function and derivatives
      qfunc = num_Q*inv_den_Q
      dqfunc_dr = dqfunc_dr*inv_den_Q
      colvar%ss = qfunc
!dbg
!      write(*,*) ' ss ',  colvar%ss
! stop
!dbg

      DO i = 1, n_coord_from
         dqfunc_dnL(i) = lambda*expnL(i)*inv_den_Q*(sum_rij(i)-num_Q*inv_den_Q)
      END DO

      !Compute Forces
      DO i = 1, n_dist_from
         DO j = 1, n_coord_from
            ftemp_i(1) = dqfunc_dr(1, i, j)
            ftemp_i(2) = dqfunc_dr(2, i, j)
            ftemp_i(3) = dqfunc_dr(3, i, j)

            CALL put_derivative(colvar, i, ftemp_i)
            CALL put_derivative(colvar, j+n_dist_from, -ftemp_i)

         END DO
      END DO
      DO i = 1, n_coord_from
         DO j = 1, n_coord_to
            ftemp_i(1) = dqfunc_dnL(i)*dnLcoord(1, i, j)
            ftemp_i(2) = dqfunc_dnL(i)*dnLcoord(2, i, j)
            ftemp_i(3) = dqfunc_dnL(i)*dnLcoord(3, i, j)

            CALL put_derivative(colvar, i+n_dist_from, ftemp_i)
            CALL put_derivative(colvar, j+n_dist_from+n_coord_from, -ftemp_i)

         END DO
      END DO

      DEALLOCATE (nLcoord)
      DEALLOCATE (dnLcoord)
      DEALLOCATE (expnL)
      DEALLOCATE (dqfunc_dr)
      DEALLOCATE (sum_rij)
      DEALLOCATE (dqfunc_dnL)

   END SUBROUTINE mindist_colvar

! **************************************************************************************************
!> \brief  evaluates function and forces due to a combination of COLVARs
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE combine_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'combine_colvar', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: def_error, this_error
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: my_par
      INTEGER                                            :: i, ii, j, ncolv, ndim
      REAL(dp)                                           :: err
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dss_vals, my_val, ss_vals
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: fi
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF(.NOT.(colvar%type_id == combine_colvar_id))CALL cp__a("colvar_methods.F",4074)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",4078)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      ncolv = SIZE(colvar%combine_cvs_param%colvar_p)
      ALLOCATE (ss_vals(ncolv))
      ALLOCATE (dss_vals(ncolv))

      ! Evaluate the individual COLVARs
      DO i = 1, ncolv
         CALL colvar_recursive_eval(colvar%combine_cvs_param%colvar_p(i)%colvar, cell, my_particles)
         ss_vals(i) = colvar%combine_cvs_param%colvar_p(i)%colvar%ss
      ENDDO

      ! Evaluate the combination of the COLVARs
      CALL initf(1)
      ndim = SIZE(colvar%combine_cvs_param%c_parameters)+ &
             SIZE(colvar%combine_cvs_param%variables)
      ALLOCATE (my_par(ndim))
      my_par(1:SIZE(colvar%combine_cvs_param%variables)) = colvar%combine_cvs_param%variables
      my_par(SIZE(colvar%combine_cvs_param%variables)+1:) = colvar%combine_cvs_param%c_parameters
      ALLOCATE (my_val(ndim))
      my_val(1:SIZE(colvar%combine_cvs_param%variables)) = ss_vals
      my_val(SIZE(colvar%combine_cvs_param%variables)+1:) = colvar%combine_cvs_param%v_parameters
      CALL parsef(1, TRIM(colvar%combine_cvs_param%function), my_par)
      colvar%ss = evalf(1, my_val)
      DO i = 1, ncolv
         dss_vals(i) = evalfd(1, i, my_val, colvar%combine_cvs_param%dx, err)
         IF ((ABS(err) > colvar%combine_cvs_param%lerr)) THEN
            WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
            WRITE (def_error, "(A,G12.6,A)") "(", colvar%combine_cvs_param%lerr, ")"
            CALL compress(this_error, .TRUE.)
            CALL compress(def_error, .TRUE.)
            CALL cp_warn(cp__l("colvar_methods.F",4112), &
                         'ASSERTION (cond) failed at line '//cp_to_string(4113)// &
                         ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                         TRIM(def_error)//' . ')
         END IF
      END DO
      DEALLOCATE (my_val)
      DEALLOCATE (my_par)
      CALL finalizef()

      ! Evaluate forces
      ALLOCATE (fi(3, colvar%n_atom_s))
      ii = 0
      DO i = 1, ncolv
         DO j = 1, colvar%combine_cvs_param%colvar_p(i)%colvar%n_atom_s
            ii = ii+1
            fi(:, ii) = colvar%combine_cvs_param%colvar_p(i)%colvar%dsdr(:, j)*dss_vals(i)
         END DO
      END DO

      DO i = 1, colvar%n_atom_s
         CALL put_derivative(colvar, i, fi(:, i))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (ss_vals)
      DEALLOCATE (dss_vals)
   END SUBROUTINE combine_colvar

! **************************************************************************************************
!> \brief evaluates the force due (and on) reaction path collective variable
!>             ss(R) = [\sum_i i*dt exp{-\lambda \sum_a(S_a(R)-f_a(i))^2}]/
!>                     [\sum_i exp{-\lambda \sum_a(S_a(R)-f_a(i))^2}]
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \par History
!>      extended MI 01.2010
!> \author fschiff
!> \note the system is still able to move in the space spanned by the CV
!>       perpendicular to the path
! **************************************************************************************************
   SUBROUTINE reaction_path_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'reaction_path_colvar', &
         routineP = moduleN//':'//routineN

      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF(.NOT.(colvar%type_id == reaction_path_colvar_id))CALL cp__a("colvar_methods.F",4168)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",4172)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      IF (colvar%reaction_path_param%dist_rmsd) THEN
         CALL rpath_dist_rmsd(colvar, my_particles)
      ELSEIF (colvar%reaction_path_param%rmsd) THEN
         CALL rpath_rmsd(colvar, my_particles)
      ELSE
         CALL rpath_colvar(colvar, cell, my_particles)
      END IF

   END SUBROUTINE reaction_path_colvar

! **************************************************************************************************
!> \brief  position along the path calculated using selected colvars
!>         as compared to functions describing the variation of these same colvars
!>         along the path given as reference
!> \param colvar ...
!> \param cell ...
!> \param particles ...
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE rpath_colvar(colvar, cell, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rpath_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iend, ii, istart, j, k, ncolv, nconf
      REAL(dp)                                           :: lambda, step_size
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: s1, ss_vals
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ds1, f_vals, fi, s1v
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ds1v

      istart = colvar%reaction_path_param%function_bounds(1)
      iend = colvar%reaction_path_param%function_bounds(2)

      nconf = colvar%reaction_path_param%nr_frames
      step_size = colvar%reaction_path_param%step_size
      ncolv = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      ALLOCATE (f_vals(ncolv, istart:iend))
      f_vals(:, :) = colvar%reaction_path_param%f_vals
      ALLOCATE (ss_vals(ncolv))

      DO i = 1, ncolv
         CALL colvar_recursive_eval(colvar%reaction_path_param%colvar_p(i)%colvar, cell, particles)
         ss_vals(i) = colvar%reaction_path_param%colvar_p(i)%colvar%ss
      ENDDO

      ALLOCATE (s1v(2, istart:iend))
      ALLOCATE (ds1v(ncolv, 2, istart:iend))

      ALLOCATE (s1(2))
      ALLOCATE (ds1(ncolv, 2))

      DO k = istart, iend
         s1v(1, k) = REAL(k, kind=dp)*step_size*EXP(-lambda*DOT_PRODUCT(ss_vals(:)-f_vals(:, k), ss_vals(:)-f_vals(:, k)))
         s1v(2, k) = EXP(-lambda*DOT_PRODUCT(ss_vals(:)-f_vals(:, k), ss_vals(:)-f_vals(:, k)))
         DO j = 1, ncolv
            ds1v(j, 1, k) = f_vals(j, k)*s1v(1, k)
            ds1v(j, 2, k) = f_vals(j, k)*s1v(2, k)
         END DO
      END DO
      DO i = 1, 2
         s1(i) = accurate_sum(s1v(i, :))
         DO j = 1, ncolv
            ds1(j, i) = accurate_sum(ds1v(j, i, :))
         END DO
      END DO

      colvar%ss = s1(1)/s1(2)/REAL(nconf-1, dp)

      ALLOCATE (fi(3, colvar%n_atom_s))

      ii = 0
      DO i = 1, ncolv
         DO j = 1, colvar%reaction_path_param%colvar_p(i)%colvar%n_atom_s
            ii = ii+1
            fi(:, ii) = colvar%reaction_path_param%colvar_p(i)%colvar%dsdr(:, j)*lambda* &
                        (ds1(i, 1)/s1(2)/REAL(nconf-1, dp)-colvar%ss*ds1(i, 2)/s1(2))*2.0_dp
         END DO
      END DO

      DO i = 1, colvar%n_atom_s
         CALL put_derivative(colvar, i, fi(:, i))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (f_vals)
      DEALLOCATE (ss_vals)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (s1)
      DEALLOCATE (ds1)

   END SUBROUTINE rpath_colvar

! **************************************************************************************************
!> \brief  position along the path calculated from the positions of a selected list of
!>         atoms as compared to the same positions in reference
!>         configurations belonging to the given path.
!> \param colvar ...
!> \param particles ...
!> \date  01.2010
!> \author MI
! **************************************************************************************************
   SUBROUTINE rpath_dist_rmsd(colvar, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rpath_dist_rmsd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ii, ik, natom, nconf, rmsd_atom
      INTEGER, DIMENSION(:), POINTER                     :: iatom
      REAL(dp)                                           :: lambda, my_rmsd, s1(2), sum_exp
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: r, r0, vec_dif
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: dvec_dif, fi, riat, s1v
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ds1
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :, :)       :: ds1v
      REAL(dp), DIMENSION(:, :), POINTER                 :: path_conf

      nconf = colvar%reaction_path_param%nr_frames
      rmsd_atom = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      path_conf => colvar%reaction_path_param%r_ref
      iatom => colvar%reaction_path_param%i_rmsd

      natom = SIZE(particles)

      ALLOCATE (r0(3*natom))
      ALLOCATE (r(3*natom))
      ALLOCATE (riat(3, rmsd_atom))
      ALLOCATE (vec_dif(rmsd_atom))
      ALLOCATE (dvec_dif(3, rmsd_atom))
      ALLOCATE (s1v(2, nconf))
      ALLOCATE (ds1v(3, rmsd_atom, 2, nconf))
      ALLOCATE (ds1(3, rmsd_atom, 2))
      DO i = 1, natom
         ii = (i-1)*3
         r0(ii+1) = particles(i)%r(1)
         r0(ii+2) = particles(i)%r(2)
         r0(ii+3) = particles(i)%r(3)
      END DO

      DO iat = 1, rmsd_atom
         ii = iatom(iat)
         riat(:, iat) = particles(ii)%r
      END DO

      DO ik = 1, nconf
         DO i = 1, natom
            ii = (i-1)*3
            r(ii+1) = path_conf(ii+1, ik)
            r(ii+2) = path_conf(ii+2, ik)
            r(ii+3) = path_conf(ii+3, ik)
         END DO

         CALL rmsd3(particles, r, r0, output_unit=-1, my_val=my_rmsd, rotate=.TRUE.)

         sum_exp = 0.0_dp
         DO iat = 1, rmsd_atom
            i = iatom(iat)
            ii = (i-1)*3
            vec_dif(iat) = (riat(1, iat)-r(ii+1))**2+(riat(2, iat)-r(ii+2))**2 &
                           +(riat(3, iat)-r(ii+3))**2
            sum_exp = sum_exp+vec_dif(iat)
         END DO

         s1v(1, ik) = REAL(ik-1, dp)*EXP(-lambda*sum_exp)
         s1v(2, ik) = EXP(-lambda*sum_exp)
         DO iat = 1, rmsd_atom
            i = iatom(iat)
            ii = (i-1)*3
            ds1v(1, iat, 1, ik) = r(ii+1)*s1v(1, ik)
            ds1v(1, iat, 2, ik) = r(ii+1)*s1v(2, ik)
            ds1v(2, iat, 1, ik) = r(ii+2)*s1v(1, ik)
            ds1v(2, iat, 2, ik) = r(ii+2)*s1v(2, ik)
            ds1v(3, iat, 1, ik) = r(ii+3)*s1v(1, ik)
            ds1v(3, iat, 2, ik) = r(ii+3)*s1v(2, ik)
         END DO

      END DO
      s1(1) = accurate_sum(s1v(1, :))
      s1(2) = accurate_sum(s1v(2, :))
      DO i = 1, 2
         DO iat = 1, rmsd_atom
            ds1(1, iat, i) = accurate_sum(ds1v(1, iat, i, :))
            ds1(2, iat, i) = accurate_sum(ds1v(2, iat, i, :))
            ds1(3, iat, i) = accurate_sum(ds1v(3, iat, i, :))
         END DO
      END DO

      colvar%ss = s1(1)/s1(2)/REAL(nconf-1, dp)

      ALLOCATE (fi(3, rmsd_atom))

      DO iat = 1, rmsd_atom
         fi(1, iat) = 2.0_dp*lambda/s1(2)/REAL(nconf-1, dp)*(ds1(1, iat, 1)-ds1(1, iat, 2)*s1(1)/s1(2))
         fi(2, iat) = 2.0_dp*lambda/s1(2)/REAL(nconf-1, dp)*(ds1(2, iat, 1)-ds1(2, iat, 2)*s1(1)/s1(2))
         fi(3, iat) = 2.0_dp*lambda/s1(2)/REAL(nconf-1, dp)*(ds1(3, iat, 1)-ds1(3, iat, 2)*s1(1)/s1(2))
         CALL put_derivative(colvar, iat, fi(:, iat))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (r0)
      DEALLOCATE (r)
      DEALLOCATE (riat)
      DEALLOCATE (vec_dif)
      DEALLOCATE (dvec_dif)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (ds1)

   END SUBROUTINE rpath_dist_rmsd

! **************************************************************************************************
!> \brief ...
!> \param colvar ...
!> \param particles ...
! **************************************************************************************************
   SUBROUTINE rpath_rmsd(colvar, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rpath_rmsd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ii, ik, natom, nconf, rmsd_atom
      INTEGER, DIMENSION(:), POINTER                     :: iatom
      REAL(dp)                                           :: lambda, my_rmsd, s1(2)
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: r, r0
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: fi, riat, s1v
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ds1
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :, :)       :: ds1v
      REAL(dp), DIMENSION(:, :), POINTER                 :: path_conf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: drmsd

      nconf = colvar%reaction_path_param%nr_frames
      rmsd_atom = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      path_conf => colvar%reaction_path_param%r_ref
      iatom => colvar%reaction_path_param%i_rmsd

      natom = SIZE(particles)

      ALLOCATE (r0(3*natom))
      ALLOCATE (r(3*natom))
      ALLOCATE (riat(3, rmsd_atom))
      ALLOCATE (s1v(2, nconf))
      ALLOCATE (ds1v(3, rmsd_atom, 2, nconf))
      ALLOCATE (ds1(3, rmsd_atom, 2))
      ALLOCATE (drmsd(3, natom))
      drmsd = 0.0_dp
      ALLOCATE (weight(natom))

      DO i = 1, natom
         ii = (i-1)*3
         r0(ii+1) = particles(i)%r(1)
         r0(ii+2) = particles(i)%r(2)
         r0(ii+3) = particles(i)%r(3)
      END DO

      DO iat = 1, rmsd_atom
         ii = iatom(iat)
         riat(:, iat) = particles(ii)%r
      END DO

! set weights of atoms in the rmsd list
      weight = 0.0_dp
      DO iat = 1, rmsd_atom
         i = iatom(iat)
         weight(i) = 1.0_dp
      END DO

      DO ik = 1, nconf
         DO i = 1, natom
            ii = (i-1)*3
            r(ii+1) = path_conf(ii+1, ik)
            r(ii+2) = path_conf(ii+2, ik)
            r(ii+3) = path_conf(ii+3, ik)
         END DO

         CALL rmsd3(particles, r0, r, output_unit=-1, weights=weight, my_val=my_rmsd, &
                    rotate=.FALSE., drmsd3=drmsd)

         s1v(1, ik) = REAL(ik-1, dp)*EXP(-lambda*my_rmsd)
         s1v(2, ik) = EXP(-lambda*my_rmsd)
         DO iat = 1, rmsd_atom
            i = iatom(iat)
            ds1v(1, iat, 1, ik) = drmsd(1, i)*s1v(1, ik)
            ds1v(1, iat, 2, ik) = drmsd(1, i)*s1v(2, ik)
            ds1v(2, iat, 1, ik) = drmsd(2, i)*s1v(1, ik)
            ds1v(2, iat, 2, ik) = drmsd(2, i)*s1v(2, ik)
            ds1v(3, iat, 1, ik) = drmsd(3, i)*s1v(1, ik)
            ds1v(3, iat, 2, ik) = drmsd(3, i)*s1v(2, ik)
         END DO
      END DO ! ik

      s1(1) = accurate_sum(s1v(1, :))
      s1(2) = accurate_sum(s1v(2, :))
      DO i = 1, 2
         DO iat = 1, rmsd_atom
            ds1(1, iat, i) = accurate_sum(ds1v(1, iat, i, :))
            ds1(2, iat, i) = accurate_sum(ds1v(2, iat, i, :))
            ds1(3, iat, i) = accurate_sum(ds1v(3, iat, i, :))
         END DO
      END DO

      colvar%ss = s1(1)/s1(2)/REAL(nconf-1, dp)

      ALLOCATE (fi(3, rmsd_atom))

      DO iat = 1, rmsd_atom
         fi(1, iat) = -lambda/s1(2)/REAL(nconf-1, dp)*(ds1(1, iat, 1)-ds1(1, iat, 2)*s1(1)/s1(2))
         fi(2, iat) = -lambda/s1(2)/REAL(nconf-1, dp)*(ds1(2, iat, 1)-ds1(2, iat, 2)*s1(1)/s1(2))
         fi(3, iat) = -lambda/s1(2)/REAL(nconf-1, dp)*(ds1(3, iat, 1)-ds1(3, iat, 2)*s1(1)/s1(2))
         CALL put_derivative(colvar, iat, fi(:, iat))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (r0)
      DEALLOCATE (r)
      DEALLOCATE (riat)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (ds1)
      DEALLOCATE (drmsd)
      DEALLOCATE (weight)

   END SUBROUTINE rpath_rmsd

! **************************************************************************************************
!> \brief evaluates the force due (and on) distance from reaction path collective variable
!>             ss(R) = -1/\lambda \log[\sum_i exp{-\lambda \sum_a(S_a(R)-f_a(i))^2}]
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \date 01.2010
!> \author MI
! **************************************************************************************************
   SUBROUTINE distance_from_path_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'distance_from_path_colvar', &
         routineP = moduleN//':'//routineN

      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF(.NOT.(colvar%type_id == distance_from_path_colvar_id))CALL cp__a("colvar_methods.F",4531)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",4535)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      IF (colvar%reaction_path_param%dist_rmsd) THEN
         CALL dpath_dist_rmsd(colvar, my_particles)
      ELSEIF (colvar%reaction_path_param%rmsd) THEN
         CALL dpath_rmsd(colvar, my_particles)
      ELSE
         CALL dpath_colvar(colvar, cell, my_particles)
      END IF

   END SUBROUTINE distance_from_path_colvar

! **************************************************************************************************
!> \brief  distance from path calculated using selected colvars
!>         as compared to functions describing the variation of these same colvars
!>         along the path given as reference
!> \param colvar ...
!> \param cell ...
!> \param particles ...
!> \date  01.2010
!> \author MI
! **************************************************************************************************
   SUBROUTINE dpath_colvar(colvar, cell, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'dpath_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iend, ii, istart, j, k, ncolv
      REAL(dp)                                           :: lambda, s1
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: ds1, s1v, ss_vals
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ds1v, f_vals, fi

      istart = colvar%reaction_path_param%function_bounds(1)
      iend = colvar%reaction_path_param%function_bounds(2)

      ncolv = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      ALLOCATE (f_vals(ncolv, istart:iend))
      f_vals(:, :) = colvar%reaction_path_param%f_vals
      ALLOCATE (ss_vals(ncolv))

      DO i = 1, ncolv
         CALL colvar_recursive_eval(colvar%reaction_path_param%colvar_p(i)%colvar, cell, particles)
         ss_vals(i) = colvar%reaction_path_param%colvar_p(i)%colvar%ss
      ENDDO

      ALLOCATE (s1v(istart:iend))
      ALLOCATE (ds1v(ncolv, istart:iend))
      ALLOCATE (ds1(ncolv))

      DO k = istart, iend
         s1v(k) = EXP(-lambda*DOT_PRODUCT(ss_vals(:)-f_vals(:, k), ss_vals(:)-f_vals(:, k)))
         DO j = 1, ncolv
            ds1v(j, k) = f_vals(j, k)*s1v(k)
         END DO
      END DO

      s1 = accurate_sum(s1v(:))
      DO j = 1, ncolv
         ds1(j) = accurate_sum(ds1v(j, :))
      END DO
      colvar%ss = -1.0_dp/lambda*LOG(s1)

      ALLOCATE (fi(3, colvar%n_atom_s))

      ii = 0
      DO i = 1, ncolv
         DO j = 1, colvar%reaction_path_param%colvar_p(i)%colvar%n_atom_s
            ii = ii+1
            fi(:, ii) = colvar%reaction_path_param%colvar_p(i)%colvar%dsdr(:, j)* &
                        2.0_dp*(ss_vals(i)-ds1(i)/s1)
         END DO
      END DO

      DO i = 1, colvar%n_atom_s
         CALL put_derivative(colvar, i, fi(:, i))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (f_vals)
      DEALLOCATE (ss_vals)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (ds1)

   END SUBROUTINE dpath_colvar

! **************************************************************************************************
!> \brief  distance from path calculated from the positions of a selected list of
!>         atoms as compared to the same positions in reference
!>         configurations belonging to the given path.
!> \param colvar ...
!> \param particles ...
!> \date  01.2010
!> \author MI
! **************************************************************************************************
   SUBROUTINE dpath_dist_rmsd(colvar, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'dpath_dist_rmsd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ii, ik, natom, nconf, rmsd_atom
      INTEGER, DIMENSION(:), POINTER                     :: iatom
      REAL(dp)                                           :: lambda, s1, sum_exp
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: r, r0, s1v, vec_dif
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ds1, dvec_dif, fi, riat
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ds1v
      REAL(dp), DIMENSION(:, :), POINTER                 :: path_conf

      nconf = colvar%reaction_path_param%nr_frames
      rmsd_atom = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      path_conf => colvar%reaction_path_param%r_ref
      iatom => colvar%reaction_path_param%i_rmsd

      natom = SIZE(particles)

      ALLOCATE (r0(3*natom))
      ALLOCATE (r(3*natom))
      ALLOCATE (riat(3, rmsd_atom))
      ALLOCATE (vec_dif(rmsd_atom))
      ALLOCATE (dvec_dif(3, rmsd_atom))
      ALLOCATE (s1v(nconf))
      ALLOCATE (ds1v(3, rmsd_atom, nconf))
      ALLOCATE (ds1(3, rmsd_atom))
      DO i = 1, natom
         ii = (i-1)*3
         r0(ii+1) = particles(i)%r(1)
         r0(ii+2) = particles(i)%r(2)
         r0(ii+3) = particles(i)%r(3)
      END DO

      DO iat = 1, rmsd_atom
         ii = iatom(iat)
         riat(:, iat) = particles(ii)%r
      END DO

      DO ik = 1, nconf
         DO i = 1, natom
            ii = (i-1)*3
            r(ii+1) = path_conf(ii+1, ik)
            r(ii+2) = path_conf(ii+2, ik)
            r(ii+3) = path_conf(ii+3, ik)
         END DO

         CALL rmsd3(particles, r, r0, output_unit=-1, rotate=.TRUE.)

         sum_exp = 0.0_dp
         DO iat = 1, rmsd_atom
            i = iatom(iat)
            ii = (i-1)*3
            vec_dif(iat) = (riat(1, iat)-r(ii+1))**2+(riat(2, iat)-r(ii+2))**2+(riat(3, iat)-r(ii+3))**2
            sum_exp = sum_exp+vec_dif(iat)
            dvec_dif(1, iat) = r(ii+1)
            dvec_dif(2, iat) = r(ii+2)
            dvec_dif(3, iat) = r(ii+3)
         END DO
         s1v(ik) = EXP(-lambda*sum_exp)
         DO iat = 1, rmsd_atom
            ds1v(1, iat, ik) = dvec_dif(1, iat)*s1v(ik)
            ds1v(2, iat, ik) = dvec_dif(2, iat)*s1v(ik)
            ds1v(3, iat, ik) = dvec_dif(3, iat)*s1v(ik)
         END DO
      END DO

      s1 = accurate_sum(s1v(:))
      DO iat = 1, rmsd_atom
         ds1(1, iat) = accurate_sum(ds1v(1, iat, :))
         ds1(2, iat) = accurate_sum(ds1v(2, iat, :))
         ds1(3, iat) = accurate_sum(ds1v(3, iat, :))
      END DO
      colvar%ss = -1.0_dp/lambda*LOG(s1)

      ALLOCATE (fi(3, rmsd_atom))

      DO iat = 1, rmsd_atom
         fi(:, iat) = 2.0_dp*(riat(:, iat)-ds1(:, iat)/s1)
         CALL put_derivative(colvar, iat, fi(:, iat))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (r0)
      DEALLOCATE (r)
      DEALLOCATE (riat)
      DEALLOCATE (vec_dif)
      DEALLOCATE (dvec_dif)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (ds1)
   END SUBROUTINE dpath_dist_rmsd

! **************************************************************************************************
!> \brief ...
!> \param colvar ...
!> \param particles ...
! **************************************************************************************************
   SUBROUTINE dpath_rmsd(colvar, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'dpath_rmsd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ii, ik, natom, nconf, rmsd_atom
      INTEGER, DIMENSION(:), POINTER                     :: iatom
      REAL(dp)                                           :: lambda, my_rmsd, s1
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: r, r0, s1v
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ds1, fi, riat
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: ds1v
      REAL(dp), DIMENSION(:, :), POINTER                 :: path_conf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: drmsd

      nconf = colvar%reaction_path_param%nr_frames
      rmsd_atom = colvar%reaction_path_param%n_components
      lambda = colvar%reaction_path_param%lambda
      path_conf => colvar%reaction_path_param%r_ref
      iatom => colvar%reaction_path_param%i_rmsd

      natom = SIZE(particles)

      ALLOCATE (r0(3*natom))
      ALLOCATE (r(3*natom))
      ALLOCATE (riat(3, rmsd_atom))
      ALLOCATE (s1v(nconf))
      ALLOCATE (ds1v(3, rmsd_atom, nconf))
      ALLOCATE (ds1(3, rmsd_atom))
      ALLOCATE (drmsd(3, natom))
      drmsd = 0.0_dp
      ALLOCATE (weight(natom))

      DO i = 1, natom
         ii = (i-1)*3
         r0(ii+1) = particles(i)%r(1)
         r0(ii+2) = particles(i)%r(2)
         r0(ii+3) = particles(i)%r(3)
      END DO

      DO iat = 1, rmsd_atom
         ii = iatom(iat)
         riat(:, iat) = particles(ii)%r
      END DO

! set weights of atoms in the rmsd list
      weight = 0.0_dp
      DO iat = 1, rmsd_atom
         i = iatom(iat)
         weight(i) = 1.0_dp
      END DO

      DO ik = 1, nconf
         DO i = 1, natom
            ii = (i-1)*3
            r(ii+1) = path_conf(ii+1, ik)
            r(ii+2) = path_conf(ii+2, ik)
            r(ii+3) = path_conf(ii+3, ik)
         END DO

         CALL rmsd3(particles, r0, r, output_unit=-1, weights=weight, my_val=my_rmsd, &
                    rotate=.FALSE., drmsd3=drmsd)

         s1v(ik) = EXP(-lambda*my_rmsd)
         DO iat = 1, rmsd_atom
            i = iatom(iat)
            ds1v(1, iat, ik) = drmsd(1, i)*s1v(ik)
            ds1v(2, iat, ik) = drmsd(2, i)*s1v(ik)
            ds1v(3, iat, ik) = drmsd(3, i)*s1v(ik)
         END DO
      END DO

      s1 = accurate_sum(s1v(:))
      DO iat = 1, rmsd_atom
         ds1(1, iat) = accurate_sum(ds1v(1, iat, :))
         ds1(2, iat) = accurate_sum(ds1v(2, iat, :))
         ds1(3, iat) = accurate_sum(ds1v(3, iat, :))
      END DO
      colvar%ss = -1.0_dp/lambda*LOG(s1)

      ALLOCATE (fi(3, rmsd_atom))

      DO iat = 1, rmsd_atom
         fi(:, iat) = ds1(:, iat)/s1
         CALL put_derivative(colvar, iat, fi(:, iat))
      END DO

      DEALLOCATE (fi)
      DEALLOCATE (r0)
      DEALLOCATE (r)
      DEALLOCATE (riat)
      DEALLOCATE (s1v)
      DEALLOCATE (ds1v)
      DEALLOCATE (ds1)
      DEALLOCATE (drmsd)
      DEALLOCATE (weight)

   END SUBROUTINE dpath_rmsd

! **************************************************************************************************
!> \brief evaluates the force due to population colvar
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \date  01.2009
!> \author fsterpone
! **************************************************************************************************
   SUBROUTINE population_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'population_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, jj, n_atoms_from, n_atoms_to, &
                                                            ndcrd, nncrd
      REAL(dp) :: dfunc, dfunc_coord, ftmp(3), func, func_coord, inv_n_atoms_from, invden, n_0, &
         ncoord, norm, num, population, r12, r_0, rdist, sigma, ss(3), xij(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ftmp_coord
      REAL(dp), DIMENSION(3)                             :: xpi, xpj
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

! If we defined the coordination number with KINDS then we have still
! to fill few missing informations...

      NULLIFY (particles_i)
      IF(.NOT.(colvar%type_id == population_colvar_id))CALL cp__a("colvar_methods.F",4872)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",4876)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      n_atoms_to = colvar%population_param%n_atoms_to
      n_atoms_from = colvar%population_param%n_atoms_from
      nncrd = colvar%population_param%nncrd
      ndcrd = colvar%population_param%ndcrd
      r_0 = colvar%population_param%r_0
      n_0 = colvar%population_param%n0
      sigma = colvar%population_param%sigma

      ALLOCATE (ftmp_coord(3, n_atoms_to))
      ftmp_coord = 0.0_dp

      ncoord = 0.0_dp
      population = 0.0_dp

      colvar%dsdr = 0.0_dp
      inv_n_atoms_from = 1.0_dp/REAL(n_atoms_from)

      norm = SQRT(pi*2.0_dp)*sigma
      norm = 1/norm

      DO ii = 1, n_atoms_from
         i = colvar%population_param%i_at_from(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         DO jj = 1, n_atoms_to
            i = colvar%population_param%i_at_to(jj)
            CALL get_coordinates(colvar, i, xpj, my_particles)
            ss = MATMUL(cell%h_inv, xpi(:)-xpj(:))
            ss = ss-NINT(ss)
            xij = MATMUL(cell%hmat, ss)
            r12 = SQRT(xij(1)**2+xij(2)**2+xij(3)**2)
            IF (r12 < 1.0e-8_dp) CYCLE
            rdist = r12/r_0
            num = (1.0_dp-rdist**nncrd)
            invden = 1.0_dp/(1.0_dp-rdist**ndcrd)
            func_coord = num*invden
            dfunc_coord = (-nncrd*rdist**(nncrd-1)*invden &
                           +num*(invden)**2*ndcrd*rdist**(ndcrd-1))/(r12*r_0)

            ncoord = ncoord+func_coord
            ftmp_coord(1, jj) = dfunc_coord*xij(1)
            ftmp_coord(2, jj) = dfunc_coord*xij(2)
            ftmp_coord(3, jj) = dfunc_coord*xij(3)
         END DO

         func = EXP(-(ncoord-n_0)**2/(2.0_dp*sigma*sigma))
         dfunc = -func*(ncoord-n_0)/(sigma*sigma)

         population = population+norm*func
         DO jj = 1, n_atoms_to
            ftmp(1) = ftmp_coord(1, jj)*dfunc
            ftmp(2) = ftmp_coord(2, jj)*dfunc
            ftmp(3) = ftmp_coord(3, jj)*dfunc
            CALL put_derivative(colvar, ii, ftmp)
            ftmp(1) = -ftmp_coord(1, jj)*dfunc
            ftmp(2) = -ftmp_coord(2, jj)*dfunc
            ftmp(3) = -ftmp_coord(3, jj)*dfunc
            CALL put_derivative(colvar, n_atoms_from+jj, ftmp)
         ENDDO
         ncoord = 0.0_dp
      ENDDO
      colvar%ss = population
   END SUBROUTINE population_colvar

! **************************************************************************************************
!> \brief evaluates the force due to the gyration radius colvar
!>        sum_i (r_i-rcom)^2/N
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \date  03.2009
!> \author MI
! **************************************************************************************************
   SUBROUTINE gyration_radius_colvar(colvar, cell, subsys, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'gyration_radius_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, n_atoms
      REAL(dp)                                           :: dri2, func, gyration, inv_n, mass_tot, mi
      REAL(dp), DIMENSION(3)                             :: dfunc, dxi, ftmp, ss, xpcom, xpi
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (particles_i, my_particles)
      IF(.NOT.(colvar%type_id == gyration_colvar_id))CALL cp__a("colvar_methods.F",4971)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",4975)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      n_atoms = colvar%gyration_param%n_atoms
      inv_n = 1.0_dp/n_atoms

      !compute COM position
      xpcom = 0.0_dp
      mass_tot = 0.0_dp
      DO ii = 1, n_atoms
         i = colvar%gyration_param%i_at(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         CALL get_mass(colvar, i, mi, my_particles)
         xpcom(:) = xpcom(:)+xpi(:)*mi
         mass_tot = mass_tot+mi
      END DO
      xpcom(:) = xpcom(:)/mass_tot

      func = 0.0_dp
      ftmp = 0.0_dp
      dfunc = 0.0_dp
      DO ii = 1, n_atoms
         i = colvar%gyration_param%i_at(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         ss = MATMUL(cell%h_inv, xpi(:)-xpcom(:))
         ss = ss-NINT(ss)
         dxi = MATMUL(cell%hmat, ss)
         dri2 = (dxi(1)**2+dxi(2)**2+dxi(3)**2)
         func = func+dri2
         dfunc(:) = dfunc(:)+dxi(:)
      END DO
      gyration = SQRT(inv_n*func)

      DO ii = 1, n_atoms
         i = colvar%gyration_param%i_at(ii)
         CALL get_coordinates(colvar, i, xpi, my_particles)
         CALL get_mass(colvar, i, mi, my_particles)
         ss = MATMUL(cell%h_inv, xpi(:)-xpcom(:))
         ss = ss-NINT(ss)
         dxi = MATMUL(cell%hmat, ss)
         ftmp(1) = dxi(1)-dfunc(1)*mi/mass_tot
         ftmp(2) = dxi(2)-dfunc(2)*mi/mass_tot
         ftmp(3) = dxi(3)-dfunc(3)*mi/mass_tot
         ftmp(:) = ftmp(:)*inv_n/gyration
         CALL put_derivative(colvar, ii, ftmp)
      END DO
      colvar%ss = gyration

   END SUBROUTINE gyration_radius_colvar

! **************************************************************************************************
!> \brief evaluates the force due to the rmsd colvar
!> \param colvar ...
!> \param subsys ...
!> \param particles ...
!> \date  12.2009
!> \author MI
!> \note  could be extended to be used with more than 2 reference structures
! **************************************************************************************************
   SUBROUTINE rmsd_colvar(colvar, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rmsd_colvar', routineP = moduleN//':'//routineN

      CALL rmsd_colvar_low(colvar, subsys, particles)
   END SUBROUTINE rmsd_colvar

! **************************************************************************************************
!> \brief  evaluates the force due to the rmsd colvar
!>        ss = (RMSDA-RMSDB)/(RMSDA+RMSDB)
!>        RMSD is calculated with respect to two reference structures, A and B,
!>        considering all the atoms of the system or only a subset of them,
!>        as selected by the input keyword LIST
!> \param colvar ...
!> \param subsys ...
!> \param particles ...
!> \date  12.2009
!> \par History TL 2012 (generalized to any number of frames)
!> \author MI
! **************************************************************************************************
   SUBROUTINE rmsd_colvar_low(colvar, subsys, particles)

      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'rmsd_colvar_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, natom, nframes
      REAL(kind=dp)                                      :: cv_val, f1, ftmp(3)
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: der, r, rmsd
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: r0
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: drmsd
      REAL(kind=dp), DIMENSION(:), POINTER               :: weights
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      NULLIFY (my_particles, particles_i, weights)
      IF(.NOT.(colvar%type_id == rmsd_colvar_id))CALL cp__a("colvar_methods.F",5079)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",5083)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      natom = SIZE(my_particles)
      nframes = colvar%rmsd_param%nr_frames
      ALLOCATE (drmsd(3, natom, nframes))
      drmsd = 0.0_dp

      ALLOCATE (r0(3*natom, nframes))
      ALLOCATE (rmsd(nframes))
      ALLOCATE (der(nframes))
      ALLOCATE (r(3*natom))

      weights => colvar%rmsd_param%weights
      DO i = 1, natom
         ii = (i-1)*3
         r(ii+1) = my_particles(i)%r(1)
         r(ii+2) = my_particles(i)%r(2)
         r(ii+3) = my_particles(i)%r(3)
      END DO
      r0(:, :) = colvar%rmsd_param%r_ref
      rmsd = 0.0_dp

      CALL rmsd3(my_particles, r, r0(:, 1), output_unit=-1, weights=weights, my_val=rmsd(1), rotate=.FALSE., drmsd3=drmsd(:, :, 1))

      IF (nframes == 2) THEN
         CALL rmsd3(my_particles, r, r0(:, 2), output_unit=-1, weights=weights, &
                    my_val=rmsd(2), rotate=.FALSE., drmsd3=drmsd(:, :, 2))

         f1 = 1.0_dp/(rmsd(1)+rmsd(2))
         ! (rmsdA-rmsdB)/(rmsdA+rmsdB)
         cv_val = (rmsd(1)-rmsd(2))*f1
         ! (rmsdA+rmsdB)^-1-(rmsdA-rmsdB)/(rmsdA+rmsdB)^2
         der(1) = f1-cv_val*f1
         ! -(rmsdA+rmsdB)^-1-(rmsdA-rmsdB)/(rmsdA+rmsdB)^2
         der(2) = -f1-cv_val*f1

         DO i = 1, colvar%rmsd_param%n_atoms
            ii = colvar%rmsd_param%i_rmsd(i)
            IF (weights(ii) > 0.0_dp) THEN
               ftmp(1) = der(1)*drmsd(1, ii, 1)+der(2)*drmsd(1, ii, 2)
               ftmp(2) = der(1)*drmsd(2, ii, 1)+der(2)*drmsd(2, ii, 2)
               ftmp(3) = der(1)*drmsd(3, ii, 1)+der(2)*drmsd(3, ii, 2)
               CALL put_derivative(colvar, i, ftmp)
            END IF
         END DO
      ELSE IF (nframes == 1) THEN
         ! Protect in case of numerical issues (for two identical frames!)
         rmsd(1) = ABS(rmsd(1))
         cv_val = SQRT(rmsd(1))
         f1 = 0.0_dp
         IF (cv_val /= 0.0_dp) f1 = 0.5_dp/cv_val
         DO i = 1, colvar%rmsd_param%n_atoms
            ii = colvar%rmsd_param%i_rmsd(i)
            IF (weights(ii) > 0.0_dp) THEN
               ftmp(1) = f1*drmsd(1, ii, 1)
               ftmp(2) = f1*drmsd(2, ii, 1)
               ftmp(3) = f1*drmsd(3, ii, 1)
               CALL put_derivative(colvar, i, ftmp)
            END IF
         END DO
      ELSE
         CALL cp__b("colvar_methods.F",5147,"RMSD implemented only for 1 and 2 reference frames!")
      END IF
      colvar%ss = cv_val

      DEALLOCATE (der)
      DEALLOCATE (r0)
      DEALLOCATE (r)
      DEALLOCATE (drmsd)
      DEALLOCATE (rmsd)

   END SUBROUTINE rmsd_colvar_low

! **************************************************************************************************
!> \brief evaluates the force from ring puckering collective variables
!>   Cramer and Pople, JACS 97 1354 (1975)
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \date 08.2012
!> \author JGH
! **************************************************************************************************
   SUBROUTINE ring_puckering_colvar(colvar, cell, subsys, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'ring_puckering_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, j, jj, m, nring
      REAL(KIND=dp)                                      :: a, at, b, da, db, ds, kr, rpxpp, svar
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cosj, sinj, z
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: nforce, zforce
      REAL(KIND=dp), DIMENSION(3)                        :: ftmp, nv, r0, rp, rpp, uv
      REAL(KIND=dp), DIMENSION(3, 3)                     :: dnvp, dnvpp
      TYPE(particle_list_type), POINTER                  :: particles_i
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particles

      IF(.NOT.(colvar%type_id == ring_puckering_colvar_id))CALL cp__a("colvar_methods.F",5189)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",5193)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF

      nring = colvar%ring_puckering_param%nring
      ALLOCATE (r(3, nring), z(nring), cosj(nring), sinj(nring))
      ALLOCATE (nforce(3, 3, nring), zforce(nring, nring, 3))
      DO ii = 1, nring
         i = colvar%ring_puckering_param%atoms(ii)
         CALL get_coordinates(colvar, i, r(:, ii), my_particles)
      END DO
      ! get all atoms within PBC distance of atom 1
      r0(:) = r(:, 1)
      DO ii = 1, nring
         r(:, ii) = pbc(r(:, ii), r0, cell)
      END DO
      !compute origin position
      r0 = 0.0_dp
      DO ii = 1, nring
         r0(:) = r0(:)+r(:, ii)
      END DO
      kr = 1._dp/REAL(nring, KIND=dp)
      r0(:) = r0(:)*kr
      DO ii = 1, nring
         r(:, ii) = r(:, ii)-r0(:)
      END DO
      ! orientation vectors
      rp = 0._dp
      rpp = 0._dp
      DO ii = 1, nring
         cosj(ii) = COS(twopi*(ii-1)*kr)
         sinj(ii) = SIN(twopi*(ii-1)*kr)
         rp(:) = rp(:)+r(:, ii)*sinj(ii)
         rpp(:) = rpp(:)+r(:, ii)*cosj(ii)
      END DO
      nv = vector_product(rp, rpp)
      nv = nv/SQRT(SUM(nv**2))

      ! derivatives of normal
      uv = vector_product(rp, rpp)
      rpxpp = SQRT(SUM(uv**2))
      DO i = 1, 3
         uv = 0._dp
         uv(i) = 1._dp
         uv = vector_product(uv, rpp)/rpxpp
         dnvp(:, i) = uv-nv*SUM(uv*nv)
         uv = 0._dp
         uv(i) = 1._dp
         uv = vector_product(rp, uv)/rpxpp
         dnvpp(:, i) = uv-nv*SUM(uv*nv)
      END DO
      DO ii = 1, nring
         nforce(:, :, ii) = dnvp(:, :)*sinj(ii)+dnvpp(:, :)*cosj(ii)
      END DO

      ! molecular z-coordinate
      DO ii = 1, nring
         z(ii) = SUM(r(:, ii)*nv(:))
      END DO
      ! z-force
      DO ii = 1, nring
         DO jj = 1, nring
            IF (ii == jj) THEN
               zforce(ii, jj, :) = nv
            ELSE
               zforce(ii, jj, :) = 0._dp
            END IF
            DO i = 1, 3
               DO j = 1, 3
                  zforce(ii, jj, i) = zforce(ii, jj, i)+r(j, ii)*nforce(j, i, jj)
               END DO
            END DO
         END DO
      END DO

      IF (colvar%ring_puckering_param%iq == 0) THEN
         ! total puckering amplitude
         svar = SQRT(SUM(z**2))
         DO ii = 1, nring
            ftmp = 0._dp
            DO jj = 1, nring
               ftmp(:) = ftmp(:)+zforce(jj, ii, :)*z(jj)
            END DO
            ftmp = ftmp/svar
            CALL put_derivative(colvar, ii, ftmp)
         END DO
      ELSE
         m = ABS(colvar%ring_puckering_param%iq)
         IF(.NOT.(m /= 1))CALL cp__a("colvar_methods.F",5282)
         IF (MOD(nring, 2) == 0 .AND. colvar%ring_puckering_param%iq == nring/2) THEN
            ! single puckering amplitude
            svar = 0._dp
            DO ii = 1, nring
               IF (MOD(ii, 2) == 0) THEN
                  svar = svar-z(ii)
               ELSE
                  svar = svar+z(ii)
               END IF
            END DO
            svar = svar*SQRT(kr)
            DO ii = 1, nring
               ftmp = 0._dp
               DO jj = 1, nring
                  IF (MOD(jj, 2) == 0) THEN
                     ftmp(:) = ftmp(:)-zforce(jj, ii, :)*SQRT(kr)
                  ELSE
                     ftmp(:) = ftmp(:)+zforce(jj, ii, :)*SQRT(kr)
                  END IF
               END DO
               CALL put_derivative(colvar, ii, -ftmp)
            END DO
         ELSE
            IF(.NOT.(m <= (nring-1)/2))CALL cp__a("colvar_methods.F",5306)
            a = 0._dp
            b = 0._dp
            DO ii = 1, nring
               a = a+z(ii)*COS(twopi*m*(ii-1)*kr)
               b = b-z(ii)*SIN(twopi*m*(ii-1)*kr)
            END DO
            a = a*SQRT(2._dp*kr)
            b = b*SQRT(2._dp*kr)
            IF (colvar%ring_puckering_param%iq > 0) THEN
               ! puckering amplitude
               svar = SQRT(a*a+b*b)
               da = a/svar
               db = b/svar
            ELSE
               ! puckering phase angle
               at = ATAN2(a, b)
               IF (at > pi/2._dp) THEN
                  svar = 2.5_dp*pi-at
               ELSE
                  svar = 0.5_dp*pi-at
               END IF
               da = -b/(a*a+b*b)
               db = a/(a*a+b*b)
            END IF
            DO jj = 1, nring
               ftmp = 0._dp
               DO ii = 1, nring
                  ds = da*COS(twopi*m*(ii-1)*kr)
                  ds = ds-db*SIN(twopi*m*(ii-1)*kr)
                  ftmp(:) = ftmp(:)+ds*SQRT(2._dp*kr)*zforce(ii, jj, :)
               END DO
               CALL put_derivative(colvar, jj, ftmp)
            END DO
         END IF
      END IF

      colvar%ss = svar

      DEALLOCATE (r, z, cosj, sinj, nforce, zforce)

   END SUBROUTINE ring_puckering_colvar

! **************************************************************************************************
!> \brief used to print reaction_path function values on an arbitrary dimensional grid
!> \param iw1 ...
!> \param ncol ...
!> \param f_vals ...
!> \param v_count ...
!> \param gp ...
!> \param grid_sp ...
!> \param step_size ...
!> \param istart ...
!> \param iend ...
!> \param s1v ...
!> \param s1 ...
!> \param p_bounds ...
!> \param lambda ...
!> \param ifunc ...
!> \param nconf ...
!> \retval k ...
!> \author fschiff
! **************************************************************************************************
   RECURSIVE FUNCTION rec_eval_grid(iw1, ncol, f_vals, v_count, &
                                    gp, grid_sp, step_size, istart, iend, s1v, s1, p_bounds, lambda, ifunc, nconf) RESULT(k)
      INTEGER                                            :: iw1, ncol
      REAL(dp), DIMENSION(:, :), POINTER                 :: f_vals
      INTEGER                                            :: v_count
      REAL(dp), DIMENSION(:), POINTER                    :: gp, grid_sp
      REAL(dp)                                           :: step_size
      INTEGER                                            :: istart, iend
      REAL(dp), DIMENSION(:, :), POINTER                 :: s1v
      REAL(dp), DIMENSION(:), POINTER                    :: s1
      INTEGER, DIMENSION(:, :), POINTER                  :: p_bounds
      REAL(dp)                                           :: lambda
      INTEGER                                            :: ifunc, nconf, k

      INTEGER                                            :: count1, i

      k = 1
      IF (v_count .LT. ncol) THEN
         count1 = v_count+1
         DO i = p_bounds(1, count1), p_bounds(2, count1)
            gp(count1) = REAL(i, KIND=dp)*grid_sp(count1)
            k = rec_eval_grid(iw1, ncol, f_vals, count1, gp, grid_sp, step_size, &
                              istart, iend, s1v, s1, p_bounds, lambda, ifunc, nconf)
         END DO
      ELSE IF (v_count == ncol .AND. ifunc == 1) THEN
         DO i = istart, iend
            s1v(1, i) = REAL(i, kind=dp)*step_size*EXP(-lambda*DOT_PRODUCT(gp(:)-f_vals(:, i), &
                                                                           gp(:)-f_vals(:, i)))
            s1v(2, i) = EXP(-lambda*DOT_PRODUCT(gp(:)-f_vals(:, i), gp(:)-f_vals(:, i)))
         END DO
         DO i = 1, 2
            s1(i) = accurate_sum(s1v(i, :))
         END DO
         WRITE (iw1, '(5F10.5)') gp(:), s1(1)/s1(2)/REAL(nconf-1, dp)
      ELSE IF (v_count == ncol .AND. ifunc == 2) THEN
         DO i = istart, iend
            s1v(1, i) = EXP(-lambda*DOT_PRODUCT(gp(:)-f_vals(:, i), gp(:)-f_vals(:, i)))
         END DO
         s1(1) = accurate_sum(s1v(1, :))

         WRITE (iw1, '(5F10.5)') gp(:), -lambda*LOG(s1(1))
      END IF
   END FUNCTION rec_eval_grid

! **************************************************************************************************
!> \brief  Reads the coordinates of reference configurations given in input
!>         either as xyz files or in &COORD section
!> \param frame_section ...
!> \param para_env ...
!> \param nr_frames ...
!> \param r_ref ...
!> \param n_atoms ...
!> \date 01.2010
!> \author MI
! **************************************************************************************************
   SUBROUTINE read_frames(frame_section, para_env, nr_frames, r_ref, n_atoms)

      TYPE(section_vals_type), POINTER                   :: frame_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: nr_frames
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_ref
      INTEGER, INTENT(OUT)                               :: n_atoms

      CHARACTER(len=*), PARAMETER :: routineN = 'read_frames', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: dummy_char
      INTEGER                                            :: i, j, natom
      LOGICAL                                            :: explicit, my_end
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rptr
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(section_vals_type), POINTER                   :: coord_section

      NULLIFY (rptr)

      DO i = 1, nr_frames
         coord_section => section_vals_get_subs_vals(frame_section, "COORD", i_rep_section=i)
         CALL section_vals_get(coord_section, explicit=explicit)
         ! Cartesian Coordinates
         IF (explicit) THEN
            CALL section_vals_val_get(coord_section, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=natom)
            IF (i == 1) THEN
               ALLOCATE (r_ref(3*natom, nr_frames))
               n_atoms = natom
            ELSE
               IF(.NOT.(3*natom == SIZE(r_ref, 1)))CALL cp__a("colvar_methods.F",5455)
            END IF
            DO j = 1, natom
               CALL section_vals_val_get(coord_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=j, r_vals=rptr)
               r_ref((j-1)*3+1:(j-1)*3+3, i) = rptr(1:3)
            END DO ! natom
         ELSE
            CALL section_vals_val_get(frame_section, "COORD_FILE_NAME", i_rep_section=i, c_val=filename)
            IF(.NOT.(TRIM(filename) /= ""))CALL cp__a("colvar_methods.F",5464)
            NULLIFY (parser)
            ALLOCATE (rptr(3))
            CALL parser_create(parser, filename, para_env=para_env, parse_white_lines=.TRUE.)
            CALL parser_get_next_line(parser, 1)
            ! Start parser
            CALL parser_get_object(parser, natom)
            CALL parser_get_next_line(parser, 1)
            IF (i == 1) THEN
               ALLOCATE (r_ref(3*natom, nr_frames))
               n_atoms = natom
            ELSE
               IF(.NOT.(3*natom == SIZE(r_ref, 1)))CALL cp__a("colvar_methods.F",5476)
            END IF
            DO j = 1, natom
               ! Atom coordinates
               CALL parser_get_next_line(parser, 1, at_end=my_end)
               IF (my_end) &
                  CALL cp_abort(cp__l("colvar_methods.F",5482), &
                                "Number of lines in XYZ format not equal to the number of atoms."// &
                                " Error in XYZ format for COORD_A (CV rmsd). Very probably the"// &
                                " line with title is missing or is empty. Please check the XYZ file and rerun your job!")
               READ (parser%input_line, *) dummy_char, rptr(1:3)
               r_ref((j-1)*3+1, i) = cp_unit_to_cp2k(rptr(1), "angstrom")
               r_ref((j-1)*3+2, i) = cp_unit_to_cp2k(rptr(2), "angstrom")
               r_ref((j-1)*3+3, i) = cp_unit_to_cp2k(rptr(3), "angstrom")
            END DO ! natom
            CALL parser_release(parser)
            DEALLOCATE (rptr)
         END IF
      END DO ! nr_frames

   END SUBROUTINE read_frames

! **************************************************************************************************
!> \brief evaluates the collective variable associated with a hydrogen bond
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \param qs_env should be removed
!> \author alin m elena
! **************************************************************************************************
   SUBROUTINE Wc_colvar(colvar, cell, subsys, particles, qs_env)
      TYPE(colvar_type), POINTER               :: colvar
      TYPE(cell_type), POINTER                 :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER  :: subsys
      TYPE(particle_type), DIMENSION(:), &
         OPTIONAL, POINTER                      :: particles
      TYPE(qs_environment_type), POINTER, OPTIONAL       :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'Wc_colvar', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: Od, H, Oa
      REAL(dp)                                 :: rOd(3), rOa(3), rH(3), &
                                                  x, y, s(3), xv(3), dmin, amin
      INTEGER                                  :: idmin, iamin, i, j
      TYPE(particle_list_type), POINTER        :: particles_i
      TYPE(particle_type), DIMENSION(:), &
         POINTER                                :: my_particles
      TYPE(wannier_centres_type), DIMENSION(:), POINTER :: wc
      INTEGER, ALLOCATABLE                     :: wcai(:), wcdi(:)
      INTEGER                                  :: nwca, nwcd
      REAL(dp)                                 :: rcut

      NULLIFY (particles_i, wc)

      IF(.NOT.(colvar%type_id == Wc_colvar_id))CALL cp__a("colvar_methods.F",5532)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",5536)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      CALL get_qs_env(qs_env, WannierCentres=wc)
      rcut = colvar%Wc%rcut ! distances are in bohr as far as I remember
      Od = colvar%Wc%ids(1)
      H = colvar%Wc%ids(2)
      Oa = colvar%Wc%ids(3)
      CALL get_coordinates(colvar, Od, rOd, my_particles)
      CALL get_coordinates(colvar, H, rH, my_particles)
      CALL get_coordinates(colvar, Oa, rOa, my_particles)
      ALLOCATE (wcai(SIZE(wc(1)%WannierHamDiag)))
      ALLOCATE (wcdi(SIZE(wc(1)%WannierHamDiag)))
      nwca = 0
      nwcd = 0
      DO j = 1, SIZE(wc(1)%WannierHamDiag)
         x = distance(rOd-wc(1)%centres(:, j))
         y = distance(rOa-wc(1)%centres(:, j))
         IF (x < rcut) THEN
            nwcd = nwcd+1
            wcdi(nwcd) = j
            CYCLE
         ENDIF
         IF (y < rcut) THEN
            nwca = nwca+1
            wcai(nwca) = j
         ENDIF
      ENDDO

      dmin = distance(rH-wc(1)%centres(:, wcdi(1)))
      amin = distance(rH-wc(1)%centres(:, wcai(1)))
      idmin = wcdi(1)
      iamin = wcai(1)
      !dmin constains the smallest numer, amin the next smallest
      DO i = 2, nwcd
         x = distance(rH-wc(1)%centres(:, wcdi(i)))
         IF (x < dmin) THEN
            dmin = x
            idmin = wcdi(i)
         ENDIF
      ENDDO
      DO i = 2, nwca
         x = distance(rH-wc(1)%centres(:, wcai(i)))
         IF (x < amin) THEN
            amin = x
            iamin = wcai(i)
         ENDIF
      ENDDO
!     zero=0.0_dp
!     CALL put_derivative(colvar, 1, zero)
!     CALL put_derivative(colvar, 2,zero)
!     CALL put_derivative(colvar, 3, zero)

!     write(*,'(2(i0,1x),4(f16.8,1x))')idmin,iamin,wc(1)%WannierHamDiag(idmin),wc(1)%WannierHamDiag(iamin),dmin,amin
      colvar%ss = wc(1)%WannierHamDiag(idmin)-wc(1)%WannierHamDiag(iamin)
      DEALLOCATE (wcai)
      DEALLOCATE (wcdi)

   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param rij ...
!> \retval distance ...
! **************************************************************************************************
      REAL(dp) FUNCTION distance(rij)
         REAL(dp), INTENT(in) :: rij(3)

         s = MATMUL(cell%h_inv, rij)
         s = s-NINT(s)
         xv = MATMUL(cell%hmat, s)
         distance = SQRT(DOT_PRODUCT(xv, xv))
      END FUNCTION distance

   END SUBROUTINE Wc_colvar

! **************************************************************************************************
!> \brief evaluates the collective variable associated with a hydrogen bond wire
!> \param colvar ...
!> \param cell ...
!> \param subsys ...
!> \param particles ...
!> \param qs_env ...
!> \author alin m elena
! **************************************************************************************************
   SUBROUTINE HBP_colvar(colvar, cell, subsys, particles, qs_env)
      TYPE(colvar_type), POINTER               :: colvar
      TYPE(cell_type), POINTER                 :: cell
      TYPE(cp_subsys_type), OPTIONAL, POINTER  :: subsys
      TYPE(particle_type), DIMENSION(:), &
         OPTIONAL, POINTER                      :: particles
      TYPE(qs_environment_type), POINTER, OPTIONAL       :: qs_env ! optional just because I am lazy... but I should get rid of it...

      CHARACTER(len=*), PARAMETER :: routineN = 'HBP_colvar', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: Od, H, Oa
      REAL(dp)                                 :: rOd(3), rOa(3), rH(3), &
                                                  x, y, s(3), xv(3), dmin, amin
      INTEGER                                  :: idmin, iamin, i, j, il, output_unit
      TYPE(particle_list_type), POINTER        :: particles_i
      TYPE(particle_type), DIMENSION(:), &
         POINTER                                :: my_particles
      TYPE(wannier_centres_type), &
         DIMENSION(:), POINTER :: wc
      INTEGER, ALLOCATABLE                     :: wcai(:), wcdi(:)
      INTEGER                                  :: nwca, nwcd
      REAL(dp)                                 :: rcut

      NULLIFY (particles_i, wc)
      output_unit = cp_logger_get_default_io_unit()

      IF(.NOT.(colvar%type_id == HBP_colvar_id))CALL cp__a("colvar_methods.F",5648)
      IF (PRESENT(particles)) THEN
         my_particles => particles
      ELSE
         IF(.NOT.(PRESENT(subsys)))CALL cp__a("colvar_methods.F",5652)
         CALL cp_subsys_get(subsys, particles=particles_i)
         my_particles => particles_i%els
      END IF
      CALL get_qs_env(qs_env, WannierCentres=wc)
      rcut = colvar%HBP%rcut ! distances are in bohr as far as I remember
      ALLOCATE (wcai(SIZE(wc(1)%WannierHamDiag)))
      ALLOCATE (wcdi(SIZE(wc(1)%WannierHamDiag)))
      colvar%ss = 0.0_dp
      DO il = 1, colvar%HBP%nPoints
         Od = colvar%HBP%ids(il, 1)
         H = colvar%HBP%ids(il, 2)
         Oa = colvar%HBP%ids(il, 3)
         CALL get_coordinates(colvar, Od, rOd, my_particles)
         CALL get_coordinates(colvar, H, rH, my_particles)
         CALL get_coordinates(colvar, Oa, rOa, my_particles)
         nwca = 0
         nwcd = 0
         DO j = 1, SIZE(wc(1)%WannierHamDiag)
            x = distance(rOd-wc(1)%centres(:, j))
            y = distance(rOa-wc(1)%centres(:, j))
            IF (x < rcut) THEN
               nwcd = nwcd+1
               wcdi(nwcd) = j
               CYCLE
            ENDIF
            IF (y < rcut) THEN
               nwca = nwca+1
               wcai(nwca) = j
            ENDIF
         ENDDO

         dmin = distance(rH-wc(1)%centres(:, wcdi(1)))
         amin = distance(rH-wc(1)%centres(:, wcai(1)))
         idmin = wcdi(1)
         iamin = wcai(1)
         !dmin constains the smallest numer, amin the next smallest
         DO i = 2, nwcd
            x = distance(rH-wc(1)%centres(:, wcdi(i)))
            IF (x < dmin) THEN
               dmin = x
               idmin = wcdi(i)
            ENDIF
         ENDDO
         DO i = 2, nwca
            x = distance(rH-wc(1)%centres(:, wcai(i)))
            IF (x < amin) THEN
               amin = x
               iamin = wcai(i)
            ENDIF
         ENDDO
         colvar%HBP%ewc(il) = colvar%HBP%shift+wc(1)%WannierHamDiag(idmin)-wc(1)%WannierHamDiag(iamin)
         colvar%ss = colvar%ss+colvar%HBP%shift+wc(1)%WannierHamDiag(idmin)-wc(1)%WannierHamDiag(iamin)
      ENDDO
      IF (output_unit > 0) THEN
         DO il = 1, colvar%HBP%nPoints
            WRITE (output_unit, '(a,1(f16.8,1x))') "HBP| = ", colvar%HBP%ewc(il)
         ENDDO
         WRITE (output_unit, '(a,1(f16.8,1x))') "HBP|\theta(x) = ", colvar%ss
      ENDIF
      DEALLOCATE (wcai)
      DEALLOCATE (wcdi)

   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param rij ...
!> \retval distance ...
! **************************************************************************************************
      REAL(dp) FUNCTION distance(rij)
         REAL(dp), INTENT(in) :: rij(3)

         s = MATMUL(cell%h_inv, rij)
         s = s-NINT(s)
         xv = MATMUL(cell%hmat, s)
         distance = SQRT(DOT_PRODUCT(xv, xv))
      END FUNCTION distance

   END SUBROUTINE HBP_colvar

END MODULE colvar_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles the multiple unit cell option regarding atomic coordinates
!> \author Teodoro Laino [tlaino] - 05.2009
! **************************************************************************************************
MODULE topology_multiple_unit_cell
   USE cell_types,                      ONLY: cell_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE topology_types,                  ONLY: topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_multiple_unit_cell'

   PRIVATE

! *** Public parameters ***
   PUBLIC :: topology_muc

CONTAINS

! **************************************************************************************************
!> \brief Handles the multiple_unit_cell for the atomic coordinates..
!> \param topology ...
!> \param subsys_section ...
!> \author Teodoro Laino [tlaino] - 05.2009
! **************************************************************************************************
   SUBROUTINE topology_muc(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_muc', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: unit_str
      INTEGER                                            :: handle, i, ind, j, k, m, n, natoms, nrep
      INTEGER, DIMENSION(:), POINTER                     :: iwork, multiple_unit_cell
      LOGICAL                                            :: check, explicit, scale
      REAL(KIND=dp), DIMENSION(3)                        :: trsl, trsl_i, trsl_j, trsl_k
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: work_section

      CALL timeset(routineN, handle)
      NULLIFY (multiple_unit_cell, iwork, cell)
      CALL section_vals_val_get(subsys_section, "TOPOLOGY%MULTIPLE_UNIT_CELL", &
                                i_vals=multiple_unit_cell)
      ! Fail is one of the value is set to zero..
      IF (ANY(multiple_unit_cell <= 0)) &
         CALL cp_abort(cp__l("topology_multiple_unit_cell.F",61), "SUBSYS%TOPOLOGY%MULTIPLE_UNIT_CELL accepts "// &
                       "only integer values larger than 0! A value of 0 or negative is meaningless!")
      IF (ANY(multiple_unit_cell /= 1)) THEN
         ! Check that the setup between CELL and TOPOLOGY is the same..
         CALL section_vals_val_get(subsys_section, "CELL%MULTIPLE_UNIT_CELL", &
                                   i_vals=iwork)
         IF (ANY(iwork /= multiple_unit_cell)) &
            CALL cp_abort(cp__l("topology_multiple_unit_cell.F",68), "SUBSYS%TOPOLOGY%MULTIPLE_UNIT_CELL and "// &
                          "SUBSYS%CELL%MULTIPLE_UNIT_CELL have been "// &
                          "setup to two different values!! Correct this error!!")
         cell => topology%cell_muc
         natoms = topology%natoms*PRODUCT(multiple_unit_cell)

         ! Check, if velocities are provided, that they are consistent in number with the atoms...
         work_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(work_section, '_DEFAULT_KEYWORD_', n_rep_val=nrep)
            check = nrep == natoms
            IF (.NOT. check) &
               CALL cp_abort(cp__l("topology_multiple_unit_cell.F",81), &
                             "Number of available entries in VELOCITY section is not compatible with the number of atoms!")
         END IF

         CALL reallocate(topology%atom_info%id_molname, 1, natoms)
         CALL reallocate(topology%atom_info%id_resname, 1, natoms)
         CALL reallocate(topology%atom_info%resid, 1, natoms)
         CALL reallocate(topology%atom_info%id_atmname, 1, natoms)
         CALL reallocate(topology%atom_info%r, 1, 3, 1, natoms)
         CALL reallocate(topology%atom_info%atm_mass, 1, natoms)
         CALL reallocate(topology%atom_info%atm_charge, 1, natoms)
         CALL reallocate(topology%atom_info%occup, 1, natoms)
         CALL reallocate(topology%atom_info%beta, 1, natoms)
         CALL reallocate(topology%atom_info%id_element, 1, natoms)
         ind = 0
         DO k = 1, multiple_unit_cell(3)
            trsl_k = cell%hmat(:, 3)*REAL(k-1, KIND=dp)
            DO j = 1, multiple_unit_cell(2)
               trsl_j = cell%hmat(:, 2)*REAL(j-1, KIND=dp)
               DO i = 1, multiple_unit_cell(1)
                  trsl_i = cell%hmat(:, 1)*REAL(i-1, KIND=dp)
                  trsl = trsl_i+trsl_j+trsl_k
                  ind = ind+1
                  IF (ind == 1) CYCLE

                  ! loop over atoms
                  n = (ind-1)*topology%natoms
                  DO m = 1, topology%natoms
                     topology%atom_info%id_atmname(n+m) = topology%atom_info%id_atmname(m)
                     topology%atom_info%r(1, n+m) = topology%atom_info%r(1, m)+trsl(1)
                     topology%atom_info%r(2, n+m) = topology%atom_info%r(2, m)+trsl(2)
                     topology%atom_info%r(3, n+m) = topology%atom_info%r(3, m)+trsl(3)
                     topology%atom_info%id_molname(n+m) = topology%atom_info%id_molname(m)
                     topology%atom_info%id_resname(n+m) = topology%atom_info%id_resname(m)
                     topology%atom_info%resid(n+m) = topology%atom_info%resid(m)
                     topology%atom_info%id_element(n+m) = topology%atom_info%id_element(m)
                     topology%atom_info%atm_mass(n+m) = topology%atom_info%atm_mass(m)
                     topology%atom_info%atm_charge(n+m) = topology%atom_info%atm_charge(m)
                  END DO
               END DO
            END DO
         END DO
         topology%natoms = natoms

         ! Deallocate the coordinate section (will be rebuilt later with the whole atomic set)
         work_section => section_vals_get_subs_vals(subsys_section, "COORD")
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(work_section, "UNIT", c_val=unit_str)
            CALL section_vals_val_get(work_section, "SCALED", l_val=scale)
         END IF
         CALL section_vals_remove_values(work_section)
         IF (explicit) THEN
            CALL section_vals_val_set(work_section, "UNIT", c_val=unit_str)
            CALL section_vals_val_set(work_section, "SCALED", l_val=scale)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE topology_muc

END MODULE topology_multiple_unit_cell
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief uses a combination of graphs and hashing to determine if two molecules
!>      are topologically equivalent, and if so, finds the one by one mapping
!> \note
!>      the graph isomorphism being solved is a computationally hard one
!>      and can not be solved in polynomial time in the general case
!>      http://mathworld.wolfram.com/IsomorphicGraphs.html
!>      the problem arises if many atoms are topologically equivalent
!>      the current algorithm is able to solve the problem for benzene (C6H6)
!>      but not for a fullerene (C60). Large systems are not really a problem (JAC).
!>      as almost all atoms are topologically unique.
!> \par History
!>      09.2006 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE graphcon
  
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

    IMPLICIT NONE

    PRIVATE
    PUBLIC :: vertex, graph_type, reorder_graph, hash_molecule

    ! a molecule is an array of vertices, each vertex has a kind
    ! and a list of edges (bonds).
    ! (the number is the index of the other vertex in the array that builds the molecule)
! **************************************************************************************************
    TYPE graph_type
       TYPE(vertex), POINTER, DIMENSION(:) :: graph
    END TYPE graph_type

! **************************************************************************************************
    TYPE vertex
       INTEGER :: kind
       INTEGER, POINTER, DIMENSION(:) :: bonds
    END TYPE vertex

! **************************************************************************************************
    TYPE class
       INTEGER, DIMENSION(:), POINTER :: reference
       INTEGER, DIMENSION(:), POINTER :: unordered
       INTEGER :: kind
       INTEGER :: Nele
       LOGICAL :: first
       INTEGER, DIMENSION(:), POINTER :: order
       INTEGER, DIMENSION(:), POINTER :: q
    END TYPE class

! **************************************************************************************************
    TYPE superclass
       INTEGER :: Nele
       INTEGER, DIMENSION(:), POINTER :: classes
    END TYPE

CONTAINS

! **************************************************************************************************
!> \brief hashes a molecule to a number. Molecules that are the (topologically) the same
!>      have the same hash. However, there is a small chance that molecules with the same hash
!>      are different
!> \param reference IN  : molecule with atomic kinds and bonds
!> \param kind_ref OUT : an atomic hash which is the same for topologically equivalent atoms
!> \param hash OUT : a hash which is the same for topologically equivalent molecules
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      Although relatively fast in general, might be quadratic with molecule size for
!>      some systems (e.g. linear alkanes)
! **************************************************************************************************
    SUBROUTINE hash_molecule(reference,kind_ref,hash)
      TYPE(vertex), DIMENSION(:), INTENT(IN)             :: reference
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: kind_ref
      INTEGER, INTENT(OUT)                               :: hash

      INTEGER                                            :: I, Ihash, N, Nclasses, Nclasses_old, &
                                                            old_class
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index, kind_new

        N=SIZE(kind_ref)
        ALLOCATE(kind_new(N),INDEX(N))
        kind_ref=reference%kind
        Nclasses_old=0
        DO Ihash=1,N
           ! generate a hash based on the the kind of each atom and the kind of its bonded atoms
           DO I=1,N
              kind_new(I)=hash_kind(kind_ref(I),kind_ref(reference(I)%bonds))
           ENDDO
           kind_ref=kind_new
           ! find the number of equivalent atoms
           CALL sort(kind_new,N,index)
           Nclasses=1
           old_class=kind_new(1)
           DO i=2,N
              IF (kind_new(I).NE.old_class) THEN
                 Nclasses=Nclasses+1
                 old_class=kind_new(I)
              ENDIF
           ENDDO
           ! if we have not generated new classes, we have presumably found all equivalence classes
           IF (Nclasses==Nclasses_old) EXIT
           Nclasses_old=Nclasses
           ! write(*,*) "Classes",Ihash, Nclasses
        ENDDO
        ! hash (sorted) kinds to a molecular hash
        hash=joaat_hash_i(kind_new)
        DEALLOCATE(kind_new,index)
    END SUBROUTINE hash_molecule

! **************************************************************************************************
!> \brief If two molecules are topologically the same, finds the ordering that maps
!>      the unordered one on the ordered one.
!> \param reference molecular description (see type definition)
!> \param unordered molecular description (see type definition)
!> \param order the mapping reference=order(unordred) if matches=.TRUE.
!>                             undefined if matches=.FALSE.
!> \param matches .TRUE. = the ordering was found
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>      See not at the top of the file about why this algorithm might consider
!>      molecules with a large number of equivalent atoms as different
!>      despite the fact that an ordering could exist for which they are the same
! **************************************************************************************************
    SUBROUTINE reorder_graph(reference, unordered, order, matches)
      TYPE(vertex), DIMENSION(:), INTENT(IN)             :: reference, unordered
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: order
      LOGICAL, INTENT(OUT)                               :: matches

      INTEGER, PARAMETER                                 :: max_tries = 1000000

      INTEGER                                            :: hash_re, hash_un, I, Iclass, iele, &
                                                            isuperclass, itries, J, N, Nclasses, &
                                                            Nele, old_class
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: class_of_atom, index_ref, index_un, &
                                                            kind_ref, kind_ref_ordered, kind_un, &
                                                            kind_un_ordered, superclass_of_atom
      TYPE(class), ALLOCATABLE, DIMENSION(:)             :: classes
      TYPE(superclass), ALLOCATABLE, DIMENSION(:)        :: superclasses

! allows for worst case matching of two benzenes ... (6!)*(6!)/6=86400
! with some margin for other molecules
! molecules with no symmetry e.g. JAC need less than 500 tries
! catch the cases where the molecules are trivially different

        IF (SIZE(reference).NE.SIZE(unordered)) THEN
           matches=.FALSE.
           RETURN
        ENDIF

        ! catch the case where the molecules are already in the right order
        N=SIZE(order)
        order=(/(i,i=1,N)/)
        IF (matrix_equal(reference,unordered,order)) THEN
           matches=.TRUE.
           RETURN
        ENDIF

        ! determine the kind of each atom, and the hash of the whole molecule
        ALLOCATE(kind_ref(N),kind_un(N),index_ref(N),index_un(N), &
                 kind_ref_ordered(N),kind_un_ordered(N), &
                 class_of_atom(N),superclass_of_atom(N))
        CALL hash_molecule(reference,kind_ref,hash_re)
        CALL hash_molecule(unordered,kind_un,hash_un)
        IF (hash_re .NE. hash_un) THEN
           matches=.FALSE.
           RETURN
        ENDIF

        ! generate the classes of equivalent atoms, i.e. the groups of atoms of the same topological kind
        kind_ref_ordered(:)=kind_ref
        CALL sort(kind_ref_ordered,N,index_ref)
        kind_un_ordered(:)=kind_un
        CALL sort(kind_un_ordered,N,index_un)
        IF (ANY(kind_ref_ordered.NE.kind_un_ordered)) THEN
           matches=.FALSE.
           RETURN
        ENDIF

        ! count different classes, assign their kinds, and the number of elements
        Nclasses=1
        old_class=kind_ref_ordered(1)
        DO i=2,N
           IF (kind_ref_ordered(I).NE.old_class) THEN
              Nclasses=Nclasses+1
              old_class=kind_ref_ordered(I)
           ENDIF
        ENDDO
        ALLOCATE(classes(Nclasses))
        classes(1)%kind=kind_ref_ordered(1)
        Nclasses=1
        classes(1)%Nele=1
        DO i=2,N
           IF (kind_ref_ordered(I).NE.classes(Nclasses)%kind) THEN
              Nclasses=Nclasses+1
              classes(Nclasses)%kind=kind_ref_ordered(I)
              classes(Nclasses)%Nele=1
           ELSE
              classes(Nclasses)%Nele=classes(Nclasses)%Nele+1
           ENDIF
        ENDDO

        ! assign the atoms to their classes
        iele=0
        DO I=1,Nclasses
           Nele=classes(I)%Nele
           ALLOCATE(classes(I)%reference(Nele))
           ALLOCATE(classes(I)%unordered(Nele))
           DO J=1,Nele
              iele=iele+1
              classes(I)%reference(J)=index_ref(iele)
              classes(I)%unordered(J)=index_un(iele)
           ENDDO
           class_of_atom(classes(I)%reference)=I
           ALLOCATE(classes(I)%order(Nele))
           ALLOCATE(classes(I)%q(Nele))
           classes(I)%order=(/(J,J=1,Nele)/)
           classes(I)%first=.TRUE.
        ENDDO

        ! find which groups of classes (superclasses) that can be solved independently.
        ! only classes with more than one element that are connected need to be reordered simultaniously

        ! find these connected components in a recursive way
        superclass_of_atom=-1
        isuperclass=0
        DO I=1,N
           ! this atom belongs to a class with several equivalent atoms, and has not yet been found
           IF (superclass_of_atom(I).EQ.-1 .AND. classes(class_of_atom(I))%Nele>1) THEN
              isuperclass=isuperclass+1
              CALL spread_superclass(I,isuperclass,superclass_of_atom,class_of_atom,classes,reference)
           ENDIF
        ENDDO

        ! put classes into superclasses
        ALLOCATE(superclasses(isuperclass))
        superclasses%Nele=0
        DO I=1,Nclasses
           J=superclass_of_atom(classes(I)%reference(1))
           IF (J>0) superclasses(J)%Nele=superclasses(J)%Nele+1
        ENDDO
        DO I=1,isuperclass
           ALLOCATE(superclasses(I)%classes(superclasses(I)%Nele))
           superclasses(I)%Nele=0
        ENDDO
        DO I=1,Nclasses
           J=superclass_of_atom(classes(I)%reference(1))
           IF (J>0) THEN
               superclasses(J)%Nele=superclasses(J)%Nele+1
               superclasses(J)%classes(superclasses(J)%Nele)=I
           ENDIF
        ENDDO

        ! write(*,*) "Class generation time",t2-t1
        ! WRITE(*,*) "Nclasses, max size, total-non-1 ",Nclasses,MAXVAL(classes%Nele),COUNT(classes%Nele>1)
        ! write(*,*) "isuperclass ",isuperclass

        ! assign the order array to their initial value
        DO Iclass=1,Nclasses
           order(classes(Iclass)%unordered)=classes(Iclass)%reference(classes(Iclass)%order)
        ENDDO

        ! reorder the atoms superclass after superclass
        itries=0
        DO I=1,isuperclass
          DO
            itries=itries+1

            ! assign the current order
            DO iclass=1,superclasses(I)%Nele
               J=superclasses(I)%classes(iclass)
               order(classes(J)%unordered)=classes(J)%reference(classes(J)%order)
            ENDDO

            ! check for matches within this superclass only, be happy if we have a match
            matches =  matrix_superclass_equal(reference,unordered,order,superclasses(I),classes)
            IF (itries>max_tries) THEN
               WRITE(*,*) "Could not find the 1-to-1 mapping to prove graph isomorphism"
               WRITE(*,*) "Reordering failed, assuming these molecules are different"
               EXIT
            ENDIF
            IF (matches) EXIT

            ! generate next permutation within this superclass
            DO iclass=1,superclasses(I)%Nele
               J=superclasses(I)%classes(iclass)
               CALL all_permutations(classes(J)%order,classes(J)%Nele, &
                                     classes(J)%q,classes(J)%first)
               IF (.NOT. classes(J)%first) EXIT
            ENDDO

            ! we are back at the original permutation so we're unable to match this superclass.
            IF (iclass.EQ.superclasses(I)%Nele .AND. &
                classes(superclasses(I)%classes(superclasses(I)%Nele))%first) EXIT
          ENDDO
          ! failed in this superblock, can exit now
          IF (.NOT. matches) EXIT
        ENDDO

        ! the final check, just to be sure
        matches=matrix_equal(reference,unordered,order)

        DO Iclass=1,Nclasses
           DEALLOCATE(classes(Iclass)%reference)
           DEALLOCATE(classes(Iclass)%unordered)
           DEALLOCATE(classes(Iclass)%order)
           DEALLOCATE(classes(Iclass)%q)
        ENDDO
        DEALLOCATE(classes)
        DO I=1,isuperclass
           DEALLOCATE(superclasses(I)%classes)
        ENDDO
        DEALLOCATE(superclasses)
    END SUBROUTINE reorder_graph

! **************************************************************************************************
!> \brief spreads the superclass over all atoms of this class and all their bonded atoms
!>      provided that the latter belong to a class which contains more than one element
!> \param I ...
!> \param isuperclass ...
!> \param superclass_of_atom ...
!> \param class_of_atom ...
!> \param classes ...
!> \param reference ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
! **************************************************************************************************
    RECURSIVE SUBROUTINE spread_superclass(I,isuperclass,superclass_of_atom,class_of_atom, &
                                             classes, reference)
      INTEGER, INTENT(IN)                                :: i, isuperclass
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: superclass_of_atom
      INTEGER, DIMENSION(:), INTENT(IN)                  :: class_of_atom
      TYPE(class), DIMENSION(:), INTENT(IN)              :: classes
      TYPE(vertex), DIMENSION(:), INTENT(IN)             :: reference

      INTEGER                                            :: J

        IF (superclass_of_atom(I).EQ.-1 .AND. classes(class_of_atom(I))%Nele>1) THEN
           superclass_of_atom(I)=isuperclass
           DO J=1, classes(class_of_atom(I))%Nele
              CALL spread_superclass(classes(class_of_atom(I))%reference(J), isuperclass, &
                                     superclass_of_atom,class_of_atom,classes,reference)
           ENDDO
           DO J=1,SIZE(reference(I)%bonds)
              CALL spread_superclass(reference(I)%bonds(J),isuperclass, &
                                     superclass_of_atom,class_of_atom,classes,reference)
           ENDDO
        ENDIF
    END SUBROUTINE spread_superclass

! **************************************************************************************************
!> \brief determines of the vertices of this superclass have the same edges
!> \param reference ...
!> \param unordered ...
!> \param order ...
!> \param super ...
!> \param classes ...
!> \retval res ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
! **************************************************************************************************
    FUNCTION matrix_superclass_equal(reference,unordered,order,super,classes) RESULT(res)
      TYPE(vertex), DIMENSION(:), INTENT(IN)             :: reference, unordered
      INTEGER, DIMENSION(:), INTENT(IN)                  :: order
      TYPE(superclass), INTENT(IN)                       :: super
      TYPE(class), DIMENSION(:), INTENT(IN)              :: classes
      LOGICAL                                            :: res

      INTEGER                                            :: I, iclass, iele, J

! I is the atom in the unordered set

loop:   DO iclass=1,super%Nele
         DO iele=1,classes(super%classes(iclass))%Nele
           I=classes(super%classes(iclass))%unordered(iele)
           res=( reference(order(I))%kind == unordered(I)%kind .AND. &
               SIZE(reference(order(I))%bonds) == SIZE(unordered(I)%bonds) )
           IF (res) THEN
              DO J=1,SIZE(reference(order(I))%bonds)
                IF (ALL(reference(order(I))%bonds(:).NE.order(unordered(I)%bonds(J)))) THEN
                    res=.FALSE.
                    EXIT loop
                ENDIF
              ENDDO
           ELSE
              EXIT loop
           ENDIF
         ENDDO
        ENDDO loop
    END FUNCTION matrix_superclass_equal

! **************************************************************************************************
!> \brief determines of the vertices of the full set is equal, i.e.
!>      we have the same connectivity graph
!> \param reference ...
!> \param unordered ...
!> \param order ...
!> \retval res ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
! **************************************************************************************************
    FUNCTION matrix_equal(reference,unordered,order) RESULT(res)
      TYPE(vertex), DIMENSION(:), INTENT(IN)             :: reference, unordered
      INTEGER, DIMENSION(:), INTENT(IN)                  :: order
      LOGICAL                                            :: res

      INTEGER                                            :: I, J

loop:   DO I=1,SIZE(reference)
           res=( reference(order(I))%kind == unordered(I)%kind .AND. &
               SIZE(reference(order(I))%bonds) == SIZE(unordered(I)%bonds) )
           IF (res) THEN
              DO J=1,SIZE(reference(order(I))%bonds)
                IF (ALL(reference(order(I))%bonds(:).NE.order(unordered(I)%bonds(J)))) THEN
                    res=.FALSE.
                    EXIT loop
                ENDIF
              ENDDO
           ELSE
              EXIT loop
           ENDIF
        ENDDO loop
    END FUNCTION matrix_equal

! **************************************************************************************************
!> \brief creates a hash for an atom based on its own kind and on the kinds
!>       of its bonded neighbors
!> \param me ...
!> \param bonds ...
!> \retval res ...
!> \par History
!>      09.2006 created [Joost VandeVondele]
!> \note
!>       bonds are sorted so that the order of neighbors appearing in the bonded list
!>       is not important
! **************************************************************************************************
    FUNCTION hash_kind(me,bonds) RESULT(res)
      INTEGER, INTENT(IN)                                :: me
      INTEGER, DIMENSION(:), INTENT(IN)                  :: bonds
      INTEGER                                            :: res

      INTEGER                                            :: I, N
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index, ordered_bonds

       N=SIZE(bonds)
       ALLOCATE(ordered_bonds(N+1),INDEX(N))
       DO I=1,N
          ordered_bonds(I)=bonds(I)
       ENDDO
       ordered_bonds(N+1)=me
       ! N: only sort the bonds, not me
       CALL sort(ordered_bonds,N,index)
       res=joaat_hash_i(ordered_bonds)
    END FUNCTION hash_kind

! **************************************************************************************************
!> \brief generates the hash of an array of integers and the index in the table
!> \param key an integer array of any length
!> \retval hash_index ...
!> \par History
!>       09.2006 created [Joost VandeVondele]
!> \note
!>       http://en.wikipedia.org/wiki/Hash_table
!>       http://www.burtleburtle.net/bob/hash/doobs.html
!>       However, since fortran doesn't have an unsigned 4 byte int
!>       we compute it using an integer with the appropriate range
!>       we return already the index in the table as a final result
! **************************************************************************************************
  FUNCTION joaat_hash_i(key) RESULT(hash_index)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: key
      INTEGER                                            :: hash_index

      INTEGER, PARAMETER                                 :: k64 = SELECTED_INT_KIND(10)
      INTEGER(KIND=k64), PARAMETER                       :: b32 = 2_k64**32-1_k64

      INTEGER                                            :: i
      INTEGER(KIND=k64)                                  :: hash

    hash=0_k64
    DO i=1,SIZE(key)
       hash=IAND(hash+IBITS(key(i),0,8)                 ,b32)
       hash=IAND(     hash+IAND(ISHFT(hash,10),b32)     ,b32)
       hash=IAND(IEOR(hash,IAND(ISHFT(hash,-6),b32))    ,b32)
       hash=IAND(hash+IBITS(key(i),8,8)                 ,b32)
       hash=IAND(     hash+IAND(ISHFT(hash,10),b32)     ,b32)
       hash=IAND(IEOR(hash,IAND(ISHFT(hash,-6),b32))    ,b32)
       hash=IAND(hash+IBITS(key(i),16,8)                ,b32)
       hash=IAND(     hash+IAND(ISHFT(hash,10),b32)     ,b32)
       hash=IAND(IEOR(hash,IAND(ISHFT(hash,-6),b32))    ,b32)
       hash=IAND(hash+IBITS(key(i),24,8)                ,b32)
       hash=IAND(     hash+IAND(ISHFT(hash,10),b32)     ,b32)
       hash=IAND(IEOR(hash,IAND(ISHFT(hash,-6),b32))    ,b32)
    ENDDO
    hash=IAND(     hash+IAND(ISHFT(hash,  3),b32)  ,b32)
    hash=IAND(IEOR(hash,IAND(ISHFT(hash,-11),b32)) ,b32)
    hash=IAND(     hash+IAND(ISHFT(hash, 15),b32)  ,b32)
    ! hash is the real 32bit hash value of the string,
    ! hash_index is an index in the hash_table
    hash_index=INT(MOD(hash,INT(HUGE(hash_index),KIND=k64)),KIND=KIND(hash_index))
  END FUNCTION joaat_hash_i

!===ACM Algorithm 323, Generation of Permutations in Lexicographic
!   Order (G6) by R. J. Ord-Smith, CACM 11 (Feb. 1968):117
!   Original Algorithm modified via Certification by I.M. Leitch,
!   17 March 1969.
! Algol to Fortran 77 by H.D.Knoble <hdkLESS at SPAM psu dot edu>,
!                                          May 1995.
!   x = initial values (/1...n/), first=.TRUE.
!   q = scratch
!   first = .TRUE. if you're back at the original order
! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param n ...
!> \param q ...
!> \param first ...
! **************************************************************************************************
        SUBROUTINE all_permutations(x,n,q,first)
      INTEGER                                            :: n, x(n), q(n)
      LOGICAL                                            :: first

      INTEGER                                            :: k, m, t

        IF (n==1) RETURN
        IF (first) THEN
          first=.FALSE.
          DO m=1,n-1
            q(m)=n
          END DO
        ENDIF
        IF(q(n-1).eq.n) THEN
          q(n-1)=n-1
          t=x(n)
          x(n)=x(n-1)
          x(n-1)=t
          RETURN
        ENDIF
        DO k=n-1,1,-1
          IF(q(k).eq.k) THEN
            q(k)=n
          ELSE
            go to 1
          ENDIF
        END DO
        first=.TRUE.
        k=1
        GOTO 2
1       m=q(k)
        t=x(m)
        x(m)=x(k)
        x(k)=t
        q(k)=m-1
        k=k+1
2       m=n
        t=x(m)
        x(m)=x(k)
        x(k)=t
        m=m-1
        k=k+1
        DO WHILE (k.lt.m)
          t=x(m)
          x(m)=x(k)
          x(k)=t
          m=m-1
          k=k+1
        END DO
        END SUBROUTINE
END MODULE graphcon
